# 
#  Copyright (C) 2007  Smithsonian Astrophysical Observatory
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from cratedict import *
from pycrates import CrateData
from pycrates import CrateKey
from string import split
import pycrates as pycr
import hashlib
import warnings

class Crate:
    def __init__(self):
        """
        Initializes the base Crate class.
        """
        self.__clear__()

        self.__keylist = CrateDict()
        self.__datalist = CrateDict()
        self.__axes = CrateDict()


    def __clear__(self):
        """
        Clears the Crate.
        """
        # public attributes
        self.name  = "" 

        # protected and private attributes
        self._crtype     = ""
        self._number     = 0
        self.__signature = ""

        if hasattr(self, "_Crate__keylist"):
            self.__keylist.clear()

        if hasattr(self, "_Crate__datalist"):
            self.__datalist.clear()

        self.__subspace = []

        self.__key_flag = False
        self.__data_flag = False

        self.__key_trash = []
        self.__data_trash = []


        ## TABLE
        self.__nrows = 0
        self.__ncols = 0


        ## IMAGE
        self.__axes = None
        self.__primary_null = False


    def __del__(self):
        """
        Clears and closes the Crate.
        """
        self.__clear__()


    def __str__(self):
        """
        Returns a formatted string representation of the Crate.
        """
        return self.__repr__()


    def __repr__(self):
        """
        Returns a formatted string representation of the Crate.        
        """
        retstr = ""

        if self._crtype == "IMAGE":
            retstr  = "   Crate Type:        <IMAGECrate>\n"
            retstr += "   Crate Name:        " + self.name + "\n"
            if len(self.__axes) > 0:
                retstr += "   Image Axes:        " + str(self.get_axisnames()) + "\n"

        else:
            retstr  = "   Crate Type:        <TABLECrate>\n"
            retstr += "   Crate Name:        " + self.name + "\n"
            retstr += "   Ncols:             " + str(self.__ncols) + "\n"
            retstr += "   Nrows:             " + str(self.__nrows) + "\n"
            
        return retstr
    

    def get_number(self):
        """
        Returns the crate number in the dataset.
        """
        return self._number


    def _set_number(self, val):
        """
        Sets the crate number in the dataset.
        """
        self._number = val


    def is_image(self):
        """
        Returns True if the crate is contains image data, False if not.
        """
        if self._crtype == "IMAGE":
            return True
        return False


    ## SUBSPACE functions ###############################################

    def _set_subspace(self, sslist):
        """
        Sets/replaces the subspace component list.
        """
        self.__subspace = sslist


    def _get_subspace(self):
        """
        Retrieves the list of subspace components.
        """
        return self.__subspace
        

    def get_subspace_data(self, cptnum, item):
        """
        Retrieves the subspace column information for a specified
        subspace component.  Component index number is one-based.
        """
        if (cptnum < 1) or (cptnum > len(self.__subspace)):
            raise IndexError("Subspace component number is out of range.")

        try: 
            ssdata = self.__subspace[cptnum-1].get_subspace_data(item)
        except KeyError:
            raise IndexError("Cannot find subspace data item " + str(item) )
        except:
            raise

        return ssdata


    def print_subspace(self):
        """
        Prints all of the data contained each of the subspace components.
        """
        retstr = ""
        for ii in range( len(self.__subspace) ):
            retstr = retstr + "Component " + str(ii+1) + "\n"
            retstr = retstr + str(self.__subspace[ii])

        return retstr


    def print_subspace_component(self, cptnum):
        """
        Prints data contained the specified subspace component.   
        Component index number is one-based.
        """
        if (cptnum < 1) or (cptnum > len(self.__subspace)):
            raise IndexError("Subspace component number is out of range.")

        retstr = ""
        retstr = retstr + "Component " + str(cptnum) + "\n"
        retstr = retstr + str(self.__subspace[cptnum-1])

        return retstr
                

    ## HEADER KEYWORD functions #########################################

    def key_exists(self, keyname):
        """
        Searches header list for specified keyword.  If it exists, 
        returns True; if not, returns False.
        """
        if keyname.upper() in self.__keylist:
            return True
        return False


    def get_key(self, key):
        """
        Searches header list for keyword by name or number and returns the 
        CrateKey if it exists.

        key - the name or number (zero-based) of the keyword to retrieve

        """
        if not isinstance( key, str):
            keylist = self._get_keylist().keys()
            
            if len(keylist) == 0:
                return

            try:
                keyname = keylist[key]
            except:
                raise IndexError("Invalid key index number.")
        else:
            keyname = key

        if self.key_exists(keyname):
            return self.__keylist[keyname.upper()]
        return None


    def get_key_value(self, key):
        """
        Finds the value of a keyword and returns it.  If the keyword 
        does not exist, returns None.
        """
        keyobj = self.get_key(key)
        
        if keyobj is not None:
            return keyobj.value
        return None


    def add_key(self, key, index=None):
        """
        Appends a key to the end of header list or replaces an 
        already existing key in the list.
        """
        if isinstance(key, pycr.CrateKey):
            if index and not isinstance(index, str):
                self.__keylist.insert( key.name.upper(), key, index )
            else:
                self.__keylist[key.name.upper()] = key
        else:
            raise TypeError("Input must be a CrateKey.")
            

    def delete_key(self, key):
        """
        Removes a keyword from the header list.
        """
        if not isinstance( key, str):
            keylist = self._get_keylist().keys()
            
            if len(keylist) == 0:
                return

            try:
                keyname = keylist[key]
            except:                
                return
        else:
            keyname = key

        if self.key_exists(keyname):
            self.__keylist.remove(keyname.upper())
            self.__key_trash.append(keyname.upper())


    def get_nkeys(self):
        """
        Returns the number of keys read in from the header.
        """
        return self.__keylist.nkeys()


    def get_keynames(self):
        """
        Returns a list of key names in the header list.
        """
        return self.__keylist.keys()


    def _get_keylist(self):
        """
        Retrieves the list containing the header keywords.
        """
        return self.__keylist


    def _copy_keylist(self):
        return self.__keylist.copy()


    def _set_keylist(self, in_list):
        """
        Replaces the keyword list.
        """
        self.__keylist = in_list
        self._set_key_flag(True)


    def _get_key_flag(self):
        """
        Returns the key flag.  'True' means that the keywords 
        have been read in from the input file.
        """
        return self.__key_flag


    def _set_key_flag(self, flag):
        """
        Sets the key flag.
        """
        self.__key_flag = flag


    def _get_key_trash(self):
        return self.__key_trash[:]


    ## COLUMN functions #################################################

    def get_colnames(self, vectors=True, rawonly=True):
        """
        Outputs as an array of the column names found matching the 
        vectors and rawonly criteria.

        vectors - flag to control format for vector columns
                     True:  shows vector notation
                     False: displays individual component columns
                  default = True

        rawonly - flag for including/excluding virtual columns
                     True:  exclude virtual columns
                     False: include virtual columns
                  default = False

        """
        return self._get_datalist_itemnames(vectors, rawonly)


    def print_colnames(self, vectors=True, rawonly=True):
        """
        Prints to screen the column names found matching the 
        vectors and rawonly criteria.

        vectors - flag to control format for vector columns
                     True:  shows vector notation
                     False: displays individual component columns
                  default = True

        rawonly - flag for including/excluding virtual columns
                     True:  exclude virtual columns
                     False: include virtual columns
                  default = False

        """
        print_list = self._get_datalist_itemnames(vectors, rawonly)

        print_str = "        Colname\n"
        for ii in range( len(print_list) ):
            print_str += ' {0:3d})   {1!s}\n'.format(ii, print_list[ii])

        return print_str


    def column_exists(self, colname):
        """
        Checks to see if the input column name exists.  Returns 'True' 
        if column exists, 'False' if it does not.
        """
        dlist = self._get_datalist_itemnames(vectors=True, rawonly=False)

        for item in dlist:
            items_lc =  item.lower()

            # tokenize
            items_lc = items_lc.replace('(', ' ')
            items_lc = items_lc.replace(", ", ' ')
            items_lc = items_lc.replace(')', '')
            items_lc = items_lc.split(' ')
            
            # search
            for name_lc in items_lc:
                if colname.lower() == name_lc:
                    return True

        return False
    

    def add_column(self, col, index=None):
        """
        Adds a column to the TABLECrate.

        col   - the column (CrateData object) to add to the crate
        index - numerical position of where to insert the column
        """

        if isinstance(col, CrateData ):
            # add name to the datalist
            if col.name == "":
                ii = 1
                while True:
                    tmpname = "COLUMN" + str(self.__ncols+ii)
                    ii = ii+1
                    if not self.column_exists(tmpname):
                        col.name = tmpname
                        break

            if '(' in col.name:
                colname = split(col.name, '(')
                if len(colname) > 1:
                    colname = colname[0]
                warnings.warn("Invalid column name.  '"+ col.name + "' has been changed to '" + 
                              colname+ "'.")
                col.name = colname

            self._append_to_datalist( col, index=index )

            self.__ncols = self.__datalist.nkeys()
            if col.get_nsets() > self.__nrows:
                self.__nrows = col.get_nsets()

        else:
            raise TypeError("Input must be a column.")



    def delete_column(self, col):
        """
        Removes a column from the Crate.
        
        col - the name or number (zero-based) of the column to delete

        """
        if not isinstance( col, str):
            if self.__datalist.nkeys() == 0:
                return
            
            try:
                colname = datalist[col]
            except:
                raise IndexError("Invalid column number.")
        else:
            colname = col

        if self.column_exists(colname):
            datalist = self._get_datalist().keys()

            for name in (datalist):
                if colname.lower() == name.lower():
                    # remove input column name from datalist
                    self._remove_from_datalist( name )
                    break
                    
            self.__ncols = self.__datalist.nkeys()

            if self.__ncols == 0:
                self.__nrows = 0
        else:
            raise ValueError("Column '"+ colname + "' does not exist.")


    def get_column(self, col):    
        """
        Finds and returns a column from the TABLECrate.
        
        col - the name or number (zero-based) of the column to retrieve

        """
        if not isinstance( col, str):
            if self.__datalist.nkeys() == 0:
                return
            
            try:
                datalist = self._get_datalist().keys()
                colname = datalist[col]
            except:
                raise IndexError("Invalid column index number " + str(col) + ".")
        else:
            colname = col

        if self.column_exists(colname):
            return self._retrieve_from_datalist( colname )
        else:
            raise ValueError("Column '"+ colname + "' does not exist.")


    def _get_datalist(self):
        """
        Retrieves the python dictionary containing regular column names.
        """
        return self.__datalist


    def _get_datalist_itemnames(self, vectors=True, rawonly=True):
        """
        Retrieves the list containing regular column names plus
        vector component names and virtual columns.
        """
        retlist = []
        
        for item_name in self.__datalist.keys():
            item = self.__datalist[item_name]

            if (rawonly == False and item.is_virtual()) or item.is_virtual() == False:

                if vectors == True:
                    if item.is_vector():
                        retlist.append( item._get_full_name() )
                    else:
                        retlist.append( item_name )
                else:
                    if item.is_vector():
                        cptslist = item.get_cptslist()
                        for ii in range(item.vdim):
                            retlist.append( cptslist[ii] )
                    else:
                        retlist.append( item_name )

        return retlist


    def _copy_datalist(self):
        """
        Returns a copy of the data list.
        """

        return self.__datalist.copy()


    def _set_datalist(self, in_dict):
        """
        Replaces the data list with input dictionary.
        """
        if isinstance(in_dict, pycr.CrateDict):
            self.__datalist = in_dict
            self._set_data_flag(True)
        else:
            raise TypeError("Input must be a CrateDict.")


    def _get_data_flag(self):
        """
        Returns the data flag.  'True' means that column data 
        has been read in from the input file.
        """
        return self.__data_flag


    def _set_data_flag(self, flag):
        """
        Sets the data flag.
        """
        self.__data_flag = flag


    def _append_to_datalist( self, dataitem, name=None, index=None ):
        """
        Adds a new column name to the data list.
        """
        if not isinstance(dataitem, pycr.CrateData):
            raise TypeError("Item must be a CrateData.")

        if name is not None and isinstance(name, str):
            itemname = name
        else:
            itemname = dataitem.name

        if index and not isinstance(index, str):
            self.__datalist.insert( itemname, dataitem, index )
        else:
            self.__datalist[ itemname ] = dataitem


    def _remove_from_datalist( self, itemname):
        """
        Removes a column name from the data list.
        """
        if itemname in self.__datalist:
            if itemname == "image":
                self.__data_trash.append( self.__datalist["image"].name )
            else:
                self.__data_trash.append( itemname.lower() )

            self.__datalist.remove( itemname )


    def _retrieve_from_datalist( self, itemname ):
        """
        Searches the datalist for the input item name and 
        returns the item if found.
        """
        for colname in self.__datalist.keys():
            if colname.lower() == itemname.lower():
                return self.__datalist[ colname ]

            # loop through items in datalist for vector columns
            if self.__datalist[colname].is_vector():
                cpts = self.__datalist[colname].get_cptslist()
                for cptname in cpts:
                    if cptname.lower() == itemname.lower():
                        return self.__datalist[colname].get_vector_component(cptname)
   

    def _get_data_trash(self):
        return self.__data_trash[:]


    ## SIGNATURE functions ##############################################

    def is_modified(self):
        """
        Returns True is the Crate information has been changed or 
        updated, False if not.
        """
        current_sig = self.__calculate_signature()

        if current_sig != self.get_signature():
            return True
        return False


    def __calculate_signature(self):
        """
        Calculates checksum of the Crate in it's current state.
        """
        sigstr = self.__str__()
        sigstr += str(self.__keylist.values())
        sigstr += str(self.__datalist.values())

        try: 
            for item_name in self.__datalist.keys():
                vals = self.__datalist[item_name].values
                if vals is not None:
                    sigstr += vals.tostring()
        except MemoryError:
            sigstr = None
            return None

        return hashlib.sha256( sigstr ).hexdigest()


    def get_signature(self):
        """
        Retrieves the stored checksum of the Crate.
        """

        if len(self.__signature) == 0:
            self.update_signature()
        return self.__signature


    def update_signature(self):
        """
        Recalculates and stores the new checksum.
        """
        self.__signature = self.__calculate_signature()
        

    ## TABLE related funcions ###########################################

    def get_ncols(self):
        """
        Retrieves the number of columns in the TABLECrate.
        """
        return self.__ncols


    def get_nrows(self):
        """
        Retrieves the number of rows in the TABLECrate.
        """
        return self.__nrows


    def _set_ncols(self, val):
        """
        Sets the number of columns.
        """
        self.__ncols = val


    def _set_nrows(self, val):
        """
        Sets the number of rows.
        """
        self.__nrows = val


    ## IMAGE related funcions ###########################################

    def get_axis(self, axis_name):
        """
        Returns the requested axis.
        """
        retaxis = None

        axislist = self.__axes.keys()
        for name in (axislist):
            if axis_name.lower() == name.lower():
                retaxis = self.__axes[ name ]
                break

        if retaxis:
            return retaxis
        else:
            raise KeyError(axis_name + " not found in list of image axes.")


    def get_axisnames(self):
        """
        Returns an array of the names of the image axes.
        """ 
        return self.__axes.keys()


    def _set_axes(self, axis_dict):
        """
        Sets the Crate's axes.
        """
        if isinstance(axis_dict, pycr.CrateDict):
            self.__axes = axis_dict
        else:
            raise TypeError("Input must be a CrateDict.")
        

    def _get_axes(self):
        """
        Returns the python dictionary of coordinate transforms associated 
        with the image. 

        """
        return self.__axes
    

    def get_ncoords(self):
        """
        Returns the number of coordinate transforms associated 
        with the image. 
        """
        return len(self.__axes)


    def get_image(self):
        """
        Returns the image in the Crate.
        """
        if "image" in self.__datalist:
            return self._retrieve_from_datalist( "image" )
        else:
            raise LookupError("Image data does not exist.")


    def add_image(self, in_image):
        """
        Adds an image to the Crate or overwrites the existing image.
        """
        if isinstance(in_image, pycr.CrateData):
            in_image._image = True
            self._append_to_datalist(in_image, name="image")
            self._set_data_flag(True)
        else:
            raise TypeError("Input must be a CrateData.")


    def delete_image(self):
        """
        Removes image from the Crate.
        """
        if "image" in self.__datalist:
            self._remove_from_datalist("image")
        else:
            raise LookupError("Image data does not exist.")


    def _set_primary_null(self, flag):
        """
        Sets flag indicating if image is Primary Null.
        """
        self.__primary_null = flag


    def _get_primary_null(self):
        """
        Returns True if image is Primary Null, False if not.
        """
        return self.__primary_null


    ## TRANSFORM related funcions #######################################

    def get_transform(self, trname):
        """
        Searches the axes or columns lists for named transform.  
        If found, the transform is returned, else returns None.
        """

        item = None
        
        if self.is_image():
            item = self.get_axis(trname)
        else:        
            item = self.get_column(trname)

        if item is not None:
            return item.get_transform()
        else:
            raise StandardError("Unable to find transform named " + trname +".")

