# 
#  Copyright (C) 2007  Smithsonian Astrophysical Observatory
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


from pytransform import *
from cratedict import *
from numpy import array
from numpy import ascontiguousarray
import hashlib

# Define 'constants' for REGULAR and VIRTUAL
REGULAR = 0
VIRTUAL = 1

class CrateData(object):

    def _get_values(self):
        vals = self._values

        # get VIRTUAL data
        if self.is_virtual() and self.parent is None:
            if self.source is not None:
                vals = self.__transform.apply( self.source.values )
                return vals
            else:
                return None
        else:
            # for self.values containing a single element
            if vals is not None:
                if vals.size == 1:
                    vals = vals.flatten()

        # get VECTOR OR COMPONENT values
        if self.parent is not None:
            ii = 0
            cpts = self.parent.get_cptslist()

            while self.name != cpts[ii] and ii < self.vdim:
                ii += 1

            vals = self.parent.values[::, ii::self.parent.vdim]
            vals_shape = list(vals.shape)

            if len(vals_shape) > 2:
                if vals_shape[1] == 1:
                    vals_shape.pop(1)
                    vals = vals.reshape( vals_shape )
            else:
                vals = vals.reshape( vals_shape[0] )

        return vals


    def _set_values(self, invals):
        if self.is_virtual():
            return
        else:
            if self._image and "array" not in str(type(invals)):
                invals = array(invals)
            elif "array" not in str(type(invals)):
                invals = array(invals).flatten()

            self._values = ascontiguousarray(invals)

    values = property(_get_values, _set_values)


    def __init__(self):
        """
        Initializes the CrateData object.
        """
        self.__clear__()
        self._values = None


    def __clear__(self):
        """
        Clears the CrateData object.
        """
        self.name = ""
        self.desc = ""
        self.unit = ""
        self._values = None
        self.vdim = 0
        self.source = None       # for virtual columns; points to it's source CrateData object
        self.parent = None       # for vector columns; points to it's parent CrateData object

        if hasattr(self, "_CrateData__cptslist") and len(self.__cptslist) != 0:
            for item in (self.__cptslist):
                del self.__dict__[item]
            self.__cptslist.clear()
        else:
            self.__cptslist = CrateDict()

        self.__eltype = REGULAR
        self.__tlmin = None
        self.__tlmax = None
        self.__nullval = None
        self.__transform = None
        self.__crateref = None   # refers to the Crate this object belongs to
        self.__signature = ""
        self._image = False

    def __del__(self):
        """
        Clears and closes the CrateData object.
        """
        self.__clear__()


    def __str__(self):
        """
        Returns a formatted string representation of the CrateData object.
        """
        return self.__repr__()


    def __repr__(self):
        """
        Returns a formatted string representation of the CrateData object.
        """
        dtype = ""

        retstr =  "  Name:     " + self.name  + "\n"

        if not self._image:
            retstr += "  Shape:    " + str(self.get_shape()) + "\n"

        if not self.is_virtual():
            if self._values is not None:
                dtype = str(self._values.dtype.name)
                retstr += "  Datatype: " + dtype + "\n"
                if not self._image:
                    retstr += "  Nsets:    " + str(self.get_nsets()) + "\n"
          

        retstr += "  Unit:     " + self.unit + "\n"
        retstr += "  Desc:     " + self.desc + "\n"

        ndim = self.get_ndim()        
        retstr += "  Eltype:   " + self._get_eltype_str() + "\n"

        if self.is_vector():
            retstr += "     NumCpts:   " + str(self.vdim) + "\n"
            retstr += "     Cpts:      " + str(self.__cptslist.keys()) + "\n"

        if ndim > 0 and (self.is_virtual() == False or (self.is_virtual() and self.source is not None)):
            retstr += "     Ndim:     " + str(ndim) + "\n"
            retstr += "     Dimarr:   " + str(self.get_dimarr())  + "\n"

        if not self.is_virtual() and not dtype.find("str"):
            retstr += "  Range:    "  + "\n"
            retstr += "     Min:   " + str(self.__tlmin) + "\n"
            retstr += "     Max:   " + str(self.__tlmax) + "\n"

        return retstr


    def _get_eltype_str(self):
        """
        Returns a string definition of the element type, which can be 
        one of the following:
           Scalar
           Array
           Vector
           Virtual
           Vector Array
           Virtual Array
           Virtual Vector Array
        """
        retstr = ""
        
        ndim = self.get_ndim()
        
        if self.is_vector():
            if self.is_virtual():
                retstr += "Virtual Vector " 
            else: 
                retstr += "Vector " 
                      
            if ndim > 0:
                retstr += "Array"

        elif self.is_virtual():
            retstr += "Virtual " 

            if ndim > 0:
                retstr += "Array"

        else:
            if ndim > 0:
                retstr += "Array"
            else:
                retstr += "Scalar"

        return retstr


    def get_eltype(self):
        """
        Returns the element type; either REGULAR or VIRTUAL.
        """
        return self.__eltype


    def _set_eltype(self, intype):
        """
        Assigns value (REGULAR or VIRTUAL) to eltype field. 
        """
        if (intype == REGULAR) or (intype ==VIRTUAL):
            self.__eltype = intype
        else:
            self.__eltype = REGULAR


    def get_shape(self):
        """
        Retrieves the shape of the data values in the format: (rows, [vdim], [dimarr]).
        """
        vals = None

        if self.is_virtual() and (self.source is not None) and (self.parent is None):
            vals = self.source.values
        elif (self._values is not None) or (self.parent is not None):
            vals = self.values

        if vals is not None:
            return vals.shape

        return None



    def get_size(self):
        """
        Returns the number of elements per row.
        """
        shape = self.get_shape()
        if shape is None:
            return 0

        vals = None
        if self.is_virtual() and (self.source is not None) and (self.parent is None):
            vals = self.source.values
        elif (self._values is not None) or (self.parent is not None):
            vals = self.values
            
        if vals is not None:
            valsize = vals.size
            return valsize/shape[0]

        return 0


    def get_nsets(self):
        """
        Returns the number of rows of data.
        """
        shape = self.get_shape()

        if shape is not None:
            return shape[0]

        return 0


    def get_ndim(self):
        """
        Returns the number of dimensions.
        """
        if self._image == True and (self.values is not None):
            return self.values.ndim
        
        if self.is_virtual() and (self.source is not None) and (self.parent is None):
            if self.is_vector():
                return self.source.values.ndim - 2
            else:
                return self.source.values.ndim - 1
 
        if (self._values is not None) or (self.parent is not None):
            # is a vector
            if self.is_vector():
                if (self.parent is not None):
                    return self.parent.values.ndim - 2
                else:
                    return self.values.ndim - 2
            else:
                return self.values.ndim - 1
            
        return 0


    def get_dimarr(self):
        """
        Returns an array containing size of each dimension.
        """
        
        shape = self.get_shape()

        if shape is None:
            return None

        if self._image == True:
            return shape

        dimarr = None
        length = len(shape)
 
        if length > 1:
            # is a vector
            if self.is_vector():
                dimarr = shape[2: length]
            else:
                dimarr = shape[1: length]

        return dimarr


    def is_virtual(self):
        """
        Returns whether the CrateData object is Virtual.
        """
        if self.__eltype == VIRTUAL:
            return True
        return False


    def is_vector(self):
        """
        Returns whether the CrateData object contains vector data or not.
        """
        if self.vdim > 1:
            return True

        return False


    def _get_full_name(self):
        """
        Returns the 'full' name of a column.

        For non-vector columns, this is just the name.

        For vectors, combines the vector name with the name of its 
        components in the following format:  vector(cpt1, cpt2, etc.)
        """
        retstr = self.name 

        if self.is_vector():
            retstr += "("
            
            cptslist = self.__cptslist.keys()
            for ii in range(self.vdim):
                if ii > 0 and ii < self.vdim:
                    retstr += ", "
                retstr +=  cptslist[ii] 

            retstr += ")"

        return retstr


    def _set_crate_ref(self, crate):
        """
        Assigns the reference to the parent Crate.
        """
        self.__crate_ref = crate

    def _get_crate_ref(self):
        """
        Retrieves the reference to the parent Crate.
        """
        return self.__crate_ref


    def get_vector_component(self, cpt):
        """
        Retrieves vector component specified by name or number (zero-based indexing).
        """

        if self.is_vector() == False:
            raise TypeError("This is not a vector.")

        cptname = None
        cptslist = self.__cptslist.keys()

        if isinstance(cpt, str):
            for name in (cptslist):
                if name.upper() == cpt.upper():
                    cptname = name

            if cptname is None:
                raise Exception("Requested component not found.")
            
        if isinstance(cpt, int) or isinstance(cpt, long):            
            if cpt < 0 or cpt >= self.vdim:
                raise IndexError("Index out of range.")

            cptname = cptslist[cpt]

        if not cptname:
            raise KeyError("Input must be a string or a number.")

        return self.__cptslist[ cptname ]


    def get_cptslist(self):
        """
        Returns an array containing the component names if the
        CrateData object is a vector.
        """
        if self.is_vector():
            return self.__cptslist.keys()

        return None

    def _set_cptslist(self, in_dict):
        """
        Sets the vector components list.
        """
        if isinstance(in_dict, CrateDict):
            self.__cptslist = in_dict
        else:
            raise TypeError("Input must be a CrateDict.")


    def _set_transform(self, in_trans):
        """
        Sets the transform.
        """
        if isinstance(in_trans, Transform):
            self.__transform = in_trans

            if self.get_eltype() == REGULAR:
                self.__eltype = VIRTUAL
        else:
            raise TypeError("Input must be a Transform.")


    def get_transform(self):
        """
        Returns the transform.
        """
        return self.__transform


    def get_transform_matrix(self):
        """
        Returns matrix derived from transform.
        """
        if self.__transform is not None:
            return self.__transform.get_transform_matrix()
        else:
            raise Exception("No transform found.  Unable to retrieve transform matrix.")


    def _set_transform_matrix(self, in_matrix):
        """
        Sets the transform matrix.
        """
        # check that in_matrix in 3X3
        self.__transform.set_transform_matrix( in_matrix )


    def is_modified(self):
        current_sig = self.__calculate_signature()

        if current_sig != self.get_signature():
            return True
        return False


    def __calculate_signature(self):
        sigstr = self.__str__()

        if self.values is not None:
            sigstr += self.values.tostring()
        
        return hashlib.sha256( sigstr ).hexdigest()


    def get_signature(self):
        """
        Retrieves the stored checksum of the CrateData.
        """
        
        if len(self.__signature) == 0:
            self.update_signature()
        return self.__signature


    def update_signature(self):
        """
        Recalculates and stores the new checksum.
        """
        self.__signature = self.__calculate_signature()
        

    def get_tlmin(self):
        return self.__tlmin

    def _set_tlmin(self, in_min):
        self.__tlmin = in_min


    def get_tlmax(self):
        return self.__tlmax

    def _set_tlmax(self, in_max):
        self.__tlmax = in_max


    def get_nullval(self):
        return self.__nullval

    def _set_nullval(self, in_null):
        self.__nullval = in_null
