# 
#  Copyright (C) 2007  Smithsonian Astrophysical Observatory
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


import pycrates.io as io
import hashlib

class CrateKey:

    def __init__(self, input=None):
        """
        Initializes the CrateKey object.
        """
        self.__clear__()

        if input is None:
            return
        
        # if input is a tuple
        if isinstance(input, tuple):
            if len(input) == 4 :
                self.name, self.value, self.unit, self.desc = input
            elif len(input) > 0 :
                for ii in range( len(input) ) :
                    if ii == 0:
                        self.name = input[ii]
                    if ii == 1:
                        self.value = input[ii]
                    if ii == 2:
                        self.unit = input[ii]
                    if ii == 3:
                        self.desc = input[ii]
        else:
            backend = io.select( input )
            if (backend is not None ):
                backend.load(self, input) 
 
            else:
                raise AttributeError("Unable to populate CrateKey object from input.")


    def __clear__(self):
        """
        Clears the CrateKey object.
        """
        self.name = ""
        self.desc = ""
        self.unit = ""
        self.value = 0
        self.__signature = ""

    def __del__(self):
        """
        Clears and closes the CrateKey object.
        """
        self.__clear__()


    def __str__(self):
        """
        Returns a formatted string representation of the CrateKey object.
        """
        return self.__repr__()


    def __repr__(self):
        """
        Returns a formatted string representation of the CrateKey object.
        """
        retstr =  "   Name:   " + self.name  + "\n"
        retstr += "   Value:  " + str(self.value) + "\n"
        retstr += "   Unit:   " + self.unit  + "\n"
        retstr += "   Desc:   " + self.desc  + "\n"
        return retstr


    def is_modified(self):
        current_sig = self.__calculate_signature()

        if current_sig != self.get_signature():
            return True
        return False


    def __calculate_signature(self):
        sigstr = self.__str__()
        return hashlib.sha256( sigstr ).hexdigest()


    def get_signature(self):
        """
        Retrieves the stored checksum of the CrateKey.
        """
        
        if len(self.__signature) == 0:
            self.update_signature()
        return self.__signature


    def update_signature(self):
        """
        Recalculates and stores the new checksum.
        """
        self.__signature = self.__calculate_signature()
        
