# 
#  Copyright (C) 2007  Smithsonian Astrophysical Observatory
#
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


import pycrates.io as io
import numpy as np
import pycrates as pycr
import os.path
import warnings 

class TABLECrate(object):
 
    def _set_name(self, in_name):
        """
        Sets the name of the Crate.
        """
        if self.__base_crate:
            self.__base_crate.name = in_name

    def _get_name(self):
        """
        Returns the name of the Crate.
        """
        if self.__base_crate:
            return self.__base_crate.name

    name = property(_get_name, _set_name)

    def _set_number(self, in_num):
        """
        Sets the number of the Crate in the dataset.
        """
        if self.__base_crate:
            self.__base_crate._number = in_num

    def get_number(self):
        """
        Returns the Crate number in the dataset.
        """
        if self.__base_crate:
            return self.__base_crate._number
        
    _number = property(get_number, _set_number)

    def _set_crtype(self, in_type):
        """
        Sets the Crate type, either 'TABLE' or 'IMAGE'.
        """
        if self.__base_crate:
            self.__base_crate._crtype = in_type

    def _get_crtype(self):
        """
        Returns the Crate type, either 'TABLE' or 'IMAGE'.
        """
        if self.__base_crate:
            return self.__base_crate._crtype

    _crtype = property(_get_crtype, _set_crtype)


    def __init__(self, input=None, mode='rw'):
        """
        Initializes the TABLECrate and chooses the appropriate backend 
        based on the input.
        """
        self.__clear__()

        if ( (mode != "rw") and (mode != "r" ) and (mode != "w") ):
            raise AttributeError("Bad value for argument 'mode': " + mode)
        
        if input and isinstance(input, pycr.Crate):
            self.name = input.name
            self.__base_crate = input
            self._crtype= "TABLE"
            return
            
        self.__base_crate = pycr.Crate()
        self._crtype= "TABLE"
       
        if input is None:
            return

        # Select the backend appropriate for the input
        backend = io.select( input )
        if backend is None:
            raise AttributeError("Can not determine kernel for input: %s" % input )
        else:
            if backend._file_exists( input ) == False:
                raise IOError("File " + input + " does not exist.")

            try:
                rw_mode = backend.open( input, "TABLE", mode=mode )
            except Exception as ex:
                raise ex

            try:
                backend._load_dataset(self)
            except Exception as ex:
                self.__clear__()
                raise ex

            self.__parent_dataset.filename = input
            self.__parent_dataset.set_rw_mode( rw_mode )
            self.update_signature()


    def __clear__(self):
        """
        Clears all TABLECrate attributes.
        """
        self.__parent_dataset = None
        self.__base_crate = None


    def __del__(self):
        """
        Clears and closes the TABLECrate.
        """
        self.__clear__()

 
    def __str__(self):
        """
        Returns a formatted string representation of the TABLECrate.
        """
        return self.__base_crate.__repr__()


    def __repr__(self):
        """
        Returns a formatted string representation of the TABLECrate.        
        """
        return self.__base_crate.__repr__()


    ## CLASS GETTER/SETTER functions ####################################
    
    def _set_parent_dataset(self, ds):
        """
        Sets the Crate's parent dataset.
        """
        if isinstance(ds, pycr.CrateDataset):
            self.__parent_dataset = ds
        else:
            raise TypeError("Input must be a CrateDataset.")


    def  _get_base_crate(self):
        """
        Retrieves the base Crate.
        """
        return self.__base_crate


    ## FILE functions ###################################################
  
    def get_dataset(self):
        """
        Retrieves the Crate's parent dataset.
        """
        return self.__parent_dataset


    def get_filename(self):
        """
        Returns the file name.
        """

        if self.__parent_dataset is not None:
            return self.__parent_dataset.filename


    def get_rw_mode(self):
        """
        Returns whether the dataset file is readable, writeable, or both.
        """
        if self.__parent_dataset is not None:
            return self.__parent_dataset.get_rw_mode()


    def _read_header(self):
        """
        Reads in header keywords.
        """
        backend = None

        # get keys from block and populate header list
        if self.__parent_dataset:
            backend = self.__parent_dataset._get_backend()

        if backend is None:
            # disable lazy initialization
            self.__base_crate._set_key_flag(True)
            self.__base_crate._set_data_flag(True)
            return
        else:
            try:
                backend.read_header(self)
                backend.read_subspace(self)

                if self._get_key_flag() and self._get_data_flag():
                    backend._close_block( self.get_number() )
                    self.update_signature()
            except:
                raise


    def _read_data(self):
        """
        Reads in the columns from the input file.
        """
        backend = None

        if self.__parent_dataset is not None:
            backend = self.__parent_dataset._get_backend()

        if backend is None:
            # disable lazy initialization
            self.__base_crate._set_key_flag(True)
            self.__base_crate._set_data_flag(True)
            return
        else:
            try:
                backend.read_columns(self)

                if self._get_key_flag() and self._get_data_flag():
                    backend._close_block( self.get_number() )
                    self.update_signature()
            except:
                raise


    def write(self, outfile=None, clobber=False):
        """
        Writes the Crate contents out to a file.

        outfile  - name of the output file
                   if None, will write to the input file
        """

        if self.__parent_dataset and outfile == self.get_filename():
            outfile = None

        # if this (current) file is not writeable, raise error
        if outfile is None and self.__parent_dataset and self.__parent_dataset.is_writeable() == False:
            # unable to write file in place
            raise IOError("File is not writeable.")

        # if outfile exists and clobber is set to False, raise error
        if clobber == False and outfile:
            # remove filter if any
            outfilename = outfile.split('[')
            if os.path.exists(outfilename[0]):
                raise IOError("Unable to write to file; clobber is set to False and file " + 
                              outfilename[0] + " exists.")

        # determine backend 
        backend = None
        if outfile is None:
            if self.__parent_dataset is None:
                raise IOError("Unable to write file without a destination.")
            else:
                backend = self.__parent_dataset._get_backend()
        else:
            backend = io.select( outfile )
            backend._clobber( outfile )
            backend.create( input=outfile )
                
        if self._get_key_flag() == False:
            self._read_header()
            
        if self._get_data_flag() == False:
            self._read_data()
            
        backend.write( self, outfile=outfile )
        

    def write_key(self, keyname):
        """
        Writes the keyword in place.
        """
        if self.key_exists(keyname):
            backend = self.__parent_dataset._get_backend()

            if backend is None:
                raise AttributeError("Can not determine kernel for input: %s" % input )
            else:
                try:
                    key = self.get_key( keyname )
                    backend.write_key(key, self.get_number())
                except:
                    raise
        else:
            raise AttributeError("Key " + keyname + " does not exist.")


    def is_image(self):
        """
        Returns True if the crate is contains image data, False if not.
        """
        return False


    ## TABLE related funcions ###########################################

    def get_ncols(self):
        """
        Retrieves the number of columns in the Crate.
        """
        if self.__base_crate:
            return self.__base_crate.get_ncols()

    def _set_ncols(self, ncols):
        """
        Sets the number of columns.
        """
        if self.__base_crate:
            self.__base_crate._set_ncols(ncols)

    def get_nrows(self):
        """
        Retrieves the number of rows in the Crate.
        """
        if self.__base_crate:
            return self.__base_crate.get_nrows()

    def _set_nrows(self, nrows):
        """
        Sets the number of rows.
        """
        if self.__base_crate:
            self.__base_crate._set_nrows(nrows)


   ## SUBSPACE functions ###############################################

    def _get_subspace(self):
        """
        Retrieves the list of subspace components.
        """
        if self.__base_crate:
            if self._get_key_flag() == False:
                self._read_header()
            return self.__base_crate._get_subspace()

    def print_subspace(self):
        """
        Prints all of the data contained each of the subspace components.
        """
        if self.__base_crate:
            if self._get_key_flag() == False:
                self._read_header()
            return self.__base_crate.print_subspace()

    def print_subspace_component(self, cptnum):
        """
        Prints data contained the specified subspace component.   
        Component index number is one-based.
        """
        if self.__base_crate:
            if self._get_key_flag() == False:
                self._read_header()
            return self.__base_crate.print_subspace_component(cptnum)

    def get_subspace_data(self, cptnum, item):
        """
        Retrieves the subspace column information for a specified
        subspace component.  Component index number is one-based.
        """
        if self.__base_crate:
            if self._get_key_flag() == False:
                self._read_header()
            return self.__base_crate.get_subspace_data(cptnum, item)

    def _set_subspace(self, sslist):
        """
        Sets/replaces the subspace component list.
        """
        if self.__base_crate:
            self.__base_crate._set_subspace(sslist)
            

    ## HEADER KEYWORD functions #########################################

    def key_exists(self, keyname):
        """
        Searches header list for specified keyword.  If it exists, 
        returns True; if not, returns False.
        """
        if self.__base_crate:
            if self._get_key_flag() == False:
                self._read_header()
            return self.__base_crate.key_exists(keyname)

    def get_key(self, key):
        """
        Searches header list for keyword by name or number and returns the 
        CrateKey if it exists.

        key - the name or number (zero-based) of the keyword to retrieve

        """
        if self.__base_crate:
            if self._get_key_flag() == False:
                self._read_header()
            return self.__base_crate.get_key(key)

    def get_key_value(self, key):
        """
        Finds the value of a keyword and returns it.  If the keyword 
        does not exist, returns None.
        """
        if self.__base_crate:
            if self._get_key_flag() == False:
                self._read_header()
            return self.__base_crate.get_key_value(key)

    def add_key(self, key, index=None):
        """
        Appends a key to the end of header list or replaces an 
        already existing key in the list.
        """
        if self.__base_crate:
            if self._get_key_flag() == False:
                self._read_header()
            self.__base_crate.add_key(key, index=index)

    def delete_key(self, key):
        """
        Removes a keyword from the header list.
        """
        if self.__base_crate:
            if self._get_key_flag() == False:
                self._read_header()
            self.__base_crate.delete_key(key)


    def get_nkeys(self):
        """
        Returns the number of keys read in from the header.
        """
        if self.__base_crate:
            if self._get_key_flag() == False:
                self._read_header()
            return self.__base_crate.get_nkeys()

    def get_keynames(self):
        """
        Returns a list of key names in the header list.
        """
        if self.__base_crate:
            if self._get_key_flag() == False:
                self._read_header()
            return self.__base_crate.get_keynames()

    def _get_keylist(self):
        """
        Retrieves the list containing the header keywords.
        """
        if self.__base_crate:
            if self._get_key_flag() == False:
                self._read_header()
            return self.__base_crate._get_keylist()

    def _set_keylist(self, klist):
        """
        Replaces the keyword list.
        """
        if self.__base_crate:
            self.__base_crate._set_keylist(klist)

    def _get_key_flag(self):
        """
        Returns the key flag.  'True' means that the keywords 
        have been read in from the input file.
        """
        if self.__base_crate:
            return self.__base_crate._get_key_flag()

    def _set_key_flag(self, flag):
        """
        Sets the key flag.
        """
        if self.__base_crate:
            self.__base_crate._set_key_flag(flag)

    def _get_key_trash(self):
        if self.__base_crate:
            return self.__base_crate._get_key_trash()


    ## SIGNATURE functions ##############################################

    def is_modified(self):
        """
        Returns True is the Crate information has been changed or 
        updated, False if not.
        """
        if self.__base_crate:
            return self.__base_crate.is_modified()

    def get_signature(self):
        """
        Retrieves the stored checksum of the Crate.
        """
        if self.__base_crate:
            return self.__base_crate.get_signature()

    def update_signature(self):
        """
        Recalculates and stores the new checksum.
        """
        if self.__base_crate:
            self.__base_crate.update_signature()


    ## COLUMN functions #################################################

    def get_colnames(self, vectors=True, rawonly=True):
        """
        Outputs as an array of the column names found matching the 
        vectors and rawonly criteria.

        vectors - flag to control format for vector columns
                     True:  shows vector notation
                     False: displays individual component columns
                  default = True

        rawonly - flag for including/excluding virtual columns
                     True:  exclude virtual columns
                     False: include virtual columns
                  default = False

        """
        if self.__base_crate:
            if self._get_data_flag() == False:
                self._read_data()
            return self.__base_crate.get_colnames(vectors=vectors, rawonly=rawonly)

    def print_colnames(self, vectors=True, rawonly=True):
        """
        Prints to screen the column names found matching the 
        vectors and rawonly criteria.

        vectors - flag to control format for vector columns
                     True:  shows vector notation
                     False: displays individual component columns
                  default = True

        rawonly - flag for including/excluding virtual columns
                     True:  exclude virtual columns
                     False: include virtual columns
                  default = False

        """
        if self.__base_crate:
            if self._get_data_flag() == False:
                self._read_data()
            return self.__base_crate.print_colnames(vectors=vectors, rawonly=rawonly)

    def column_exists(self, colname):
        """
        Checks to see if the input column name exists.  Returns 'True' 
        if column exists, 'False' if it does not.
        """
        if self.__base_crate:
            if self._get_data_flag() == False:
                self._read_data()
            return self.__base_crate.column_exists(colname)

    def add_column(self, col, index=None):
        """
        Adds a column to the TABLECrate.

        col   - the column (CrateData object) to add to the crate
        index - numerical position of where to insert the column
        """
        if self.__base_crate:
            if self._get_data_flag() == False:
                self._read_data()
            self.__base_crate.add_column(col, index=index)

            # add column as crate attribute
            if hasattr(self, col.name):
                warnings.warn("Unable to create named-column attribute for column '" + 
                              col.name + "'.  Attribute already exists.")
            else:
                self.__dict__[ col.name ] = col


    def delete_column(self, col):
        """
        Removes a column from the Crate.
        
        col - the name or number (zero-based) of the column to delete

        """
        if self.__base_crate:
            if self._get_data_flag() == False:
                self._read_data()
            self.__base_crate.delete_column(col)

        # remove input column name from class attributes
        if isinstance( col, str) and hasattr(self, col):
            del self.__dict__[ col ] 


    def get_column(self, col):
        """
        Finds and returns a column from the TABLECrate.
        
        col - the name or number (zero-based) of the column to retrieve

        """

        
        if self.__base_crate:
            if self._get_data_flag() == False:
                self._read_data()
            return self.__base_crate.get_column(col)


    def get_transform(self, trname):
        """
        Searches the columns list for named transform.  
        If found, the transform is returned, else returns None.
        """
        if self.__base_crate:
            if self._get_data_flag() == False:
                self._read_data()
            return self.__base_crate.get_transform(trname)
    

    def _get_datalist(self):
        """
        Retrieves the python dictionary containing regular column names.
        """
        if self.__base_crate:
            if self._get_data_flag() == False:
                self._read_data()
            return self.__base_crate._get_datalist()

    def _set_datalist(self, dlist):
        """
        Replaces the data list with input dictionary.
        """
        if self.__base_crate:
            self.__base_crate._set_datalist( dlist )

    def _get_data_flag(self):
        """
        Returns the data flag.  'True' means that column data 
        has been read in from the input file.
        """
        if self.__base_crate:
            return self.__base_crate._get_data_flag()

    def _set_data_flag(self, flag):
        """
        Sets the data flag.
        """
        if self.__base_crate:
            self.__base_crate._set_data_flag(flag)

    def _get_data_trash(self):
        if self.__base_crate:
            return self.__base_crate._get_data_trash()


    ## SELF TEST routine ################################################

    def _self_test(self, filename, debug=False):
        retstat = "PASS"

        if debug:
            print "Running TABLECrate smoketest"

        try:
            cr = TABLECrate()
            key1 = pycr.CrateKey( ("KEY1", 3, "", "Keyword 1") )
            key2 = pycr.CrateKey( ("KEY2", 1423.09, "", "Keyword 2") )
            key3 = pycr.CrateKey( ("KEY3", 23470908.6, "", "Keyword 3") )

            cr.add_key( key1 )
            cr.add_key( key2 )
            cr.add_key( key3 )

            col1 = pycr.CrateData()
            col1.name = "COL1"
            col1.values = np.array( np.arange(10) )
            col1.desc = "Column 1"
            col1.unit = "d"

            cr.add_column(col1)

            cr.write(filename, clobber=True)

            cr2 = TABLECrate( filename )
            keynames = cr2.get_keynames()

            for name in keynames:
                if cr.get_key_value( name ) != cr2.get_key_value( name ):
                    retstat = "FAIL"
                    break

            colvals2 = cr2.get_column(col1.name).values
            
            for ii in range(0, len(col1.values)):
                if col1.values[ii] != colvals2[ii]:
                    retstat = "FAIL"

        except:
            retstat = "FAIL"
            
            if debug:
                print "ERROR in TABLECrate."
            pass

        return retstat


    
