/*                                                                
**  Copyright (C) 2007,2010-2012  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/* H*****************************************************************
 *
 * FILE NAME:  pyparamio.c
 *
 * DEVELOPMENT: tools
 *
 * DESCRIPTION:  Provides a c-extension for the Python language to
 * the paramio functions
 *
 *
 *
 * REVISION HISTORY:
 *
 * Ref. No.         Date
 ----------       -----
 0.1              21May2007 	File Created
 H***************************************************************** */

#include "pyparamio.h"

/* * * * * * * * * * * * * * * * * * * * * * * *
 * Utilities for Python Interaction
 * * * * * * * * * * * * * * * * * * * * * * * */

/*
 * Set up the methods table
 */
static PyMethodDef
    paramioMethods[] =
        {
         /* Each entry in the groupMethods array is a PyMethodDef structure containing
          * 1) the Python name,
          * 2) the C-function that implements the function,
          * 3) flags indicating whether or not keywords are accepted for this function,
          * and 4) The docstring for the function.
          */
          {
           "paramopen",
           (PyCFunction)_paramopen,
           METH_VARARGS,
           "Example:  >>> paramopen(\"dmextract\")\
     >>> paramopen(\"dmextract\", \"rw\") \
     or >>> args = [ \"dmcopy\", \"test.fits\", \"out=out.fits\", \"cl+\" ] >>> paramopen(\"dmextract\", \"rw\", args)"},
          {"paramclose", (PyCFunction)_paramclose, METH_VARARGS,
           "Example:  >>> fp = paramopen(\"dmextract\")\
     >>> paramclose(fp) "},
          {
           "pset",
           (PyCFunction)_pset,
           METH_VARARGS,
           "Example:  >>> pset(\"dmcopy\",\"infile\",\"in.fits\")\
     or >>> pars = dict(infile=\"in.fits\", outfile=\"out.fits\", clobber=\"yes\") >>> pset( \"dmcopy\", pars )"},
          {"pget", (PyCFunction)_pget, METH_VARARGS,
           "Example:  >>> pget(\"dmcopy\", \"infile\")"},
          {"pquery", (PyCFunction)_pquery, METH_VARARGS,
           "Example:  >>> pquery( \"dmcopy\", \"infile\" )"},
          {
           "paccess",
           (PyCFunction)_paccess,
           METH_VARARGS,
           "Example:  >>> paccess(\"dmlist\")\
     >>> paccess(\"dmlist\", \"r\")\
     >>> fp = paramopen(\"dmcopy\") >>> paccess( fp, \"infile\" ) "},
          {
           "paramgetpath",
           (PyCFunction)_paramgetpath,
           METH_VARARGS,
           "Example:  >>> paramgetpath(\"dmlist\")\
     >>> fp = paramopen(\"dmlist\") >>> paramgetpath(fp)"},
          {
           "punlearn",
           (PyCFunction)_punlearn,
           METH_VARARGS,
           "Example:  punlearn(\"dmextract\")\
     >>> punlearn([\"dmextract\",\"dmlist\",\"dmcopy\"])"},
          {"pgeti", (PyCFunction)_pgeti, METH_VARARGS,
           "Example:  >>> vali = pgeti( \"dmextract\", \"verbose\" )"},
          {"pputi", (PyCFunction)_pputi, METH_VARARGS,
           "Example:  >>> pputi(\"dmextract\", \"verbose\", 1)"},
          {"pgetb", (PyCFunction)_pgetb, METH_VARARGS,
           "Example:  >>> valb = pgetb( \"dmextract\", \"verbose\" )"},
          {"pputb", (PyCFunction)_pputb, METH_VARARGS,
           "Example:  >>> pputb(\"dmextract\", \"verbose\", 1)"},
          {"pgetd", (PyCFunction)_pgetd, METH_VARARGS,
           "Example:  >>> vald = pgetd( \"dmextract\", \"verbose\" )"},
          {"pputd", (PyCFunction)_pputd, METH_VARARGS,
           "Example:  >>> pputd(\"dmextract\", \"verbose\", 1)"},
          {"pgetstr", (PyCFunction)_pgetstr, METH_VARARGS,
           "Example:  >>> valstr = pgetstr( \"dmextract\", \"verbose\" )"},
          {"pputstr", (PyCFunction)_pputstr, METH_VARARGS,
           "Example:  >>> pputstr(\"dmextract\", \"clobber\", \"yes\")"},
          {"plist", (PyCFunction)_plist, METH_VARARGS,
           "Example:  >>> list = plist(\"dmextract\")"}, {NULL, NULL, 0, NULL} /* Sentinel - marks the end of this structure */
        };/*end...paramioMethods */

/*
 * New datatype dealloc method
 */
static void paramio_dealloc(Param_File *fptr)
{
  PyObject_Del(fptr);
}/*end...paramio_dealloc */

/*
 * New datatype method to create new region
 */
static Param_File *paramio_new(paramfile fptr)
{
  Param_File *myParam = PyObject_New(Param_File, &Param_File_Type);
  myParam->fptr = fptr;

  return myParam;
}/*end...paramio_new*/

/*
 * New datatype initialization
 */
static PyTypeObject Param_File_Type =
{
  PyObject_HEAD_INIT(NULL)
  0, /*ob_size*/
  "Param_File", /*tp_name*/
  sizeof(Param_File), /*tp_basicsize*/
  0, /*tp_itemsize*/
  (destructor)paramio_dealloc, /*tp_dealloc*/
};

/* Error Message Format Strings */
char parmsg[1024];
#define PARAMIO_GENERIC_MSG        "%s() %s"
#define PARAMIO_TYPE_ERROR_MSG     "%s() Could not parse input arguments, please check input for correct type(s)"
#define PARAMIO_OPEN_FILE_MSG      "%s() Could not open parameter file"
#define PARAMIO_PARAM_MSG          "%s() Parameter not found"

/*
 * Initialize the module
 * */
void initparamio(void);

void initparamio(void)
{
  (void)Py_InitModule("paramio", paramioMethods);
  set_paramerror(0); /* Disable exiting when encounters an error */

  Param_File_Type.ob_type = &PyType_Type;
}/*end... initparamio */

/* * * * * * * * * * * * * * * * * * * * * * * *
 * Function Definitions
 * * * * * * * * * * * * * * * * * * * * * * * */

/*
 * This function will open a parameter file, returning a file pointer
 * which can be used with the other paramio functions.
 */
static PyObject *_paramopen(PyObject *self, /*i: Used by Python */
PyObject *args /*i: Python tuple of the arguments */)
{
  char* funcName = "paramopen";
  paramfile fptr; /* the file pointer */
  char *pfile = NULL; /* parameter file */
  char *pmode = NULL; /* mode of the parameter file */
  char **argv = NULL; /* optional list of parameter files */
  int argc = 0; /* count of optional parameter files */
  int ii; /* loop variable */

  PyObject *py_paramlist = NULL; /* Python List Object of the parameter files */
  PyObject *param = NULL; /* Each parameter file within the py_paramlist list */

  /*'z' = Python object may be a string or None, in which case the C pointer is set to NULL. */
  if (!PyArg_ParseTuple(args, "z|sO", &pfile, &pmode, &py_paramlist))
  {
    sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
    PyErr_SetString(PyExc_TypeError, parmsg);
    return NULL;
  }/*end if... */

  if (pmode == NULL)
  {
    pmode = "r"; /*default value*/
  }/*end if...*/

  if (py_paramlist != NULL)/* if a paramlist is present */
  {
    if (!PyList_Check(py_paramlist) && (PyList_Size(py_paramlist) != 0))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    argc = PyList_GET_SIZE(py_paramlist); /* get the number of params in the list */
    argv = (char **)calloc(argc + 1, sizeof(char *)); /*create the space for the list*/

    for (ii = 0; ii < argc; ii++)
    {
      param = PyList_GetItem(py_paramlist, ii); /*get the iith param in the py_paramlist */
      argv[ii] = PyString_AsString(param); /*add item to the argv array*/
    }/*end for... ii < argc */
  }/*end if... */

  fptr = paramopen(pfile, argv, argc, pmode);
  if (fptr == NULL)
  {
    sprintf(parmsg, PARAMIO_GENERIC_MSG, funcName,
            (char*)"Could not execute paramopen");
    PyErr_SetString(PyExc_IOError, parmsg);
    return NULL;
  }/*end if... */

  free(argv); /*free allocated memory */

  return ((PyObject*)paramio_new(fptr)); /*return the new pointer*/
}/*end... _paramopen */

/*
 * This function will close the parameter file associated with the given
 * Param_File_Type variable.
 */
static PyObject *_paramclose(PyObject *self, /*i: Used by Python */
PyObject *args /*i: Python tuple of the arguments */)
{
  char* funcName = "paramclose";
  PyObject *py_fptr = NULL; /* fileptr python object */

  if (!PyArg_ParseTuple(args, "O!", &Param_File_Type, &py_fptr))
  {
    sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
    PyErr_SetString(PyExc_TypeError, parmsg);
    return NULL;
  }/*end if... */

  if (((Param_File*)(py_fptr))->fptr != NULL)
  {
    (void)paramclose(((Param_File*)(py_fptr))->fptr);
  }/*end if... */

  ((Param_File*)(py_fptr))->fptr = NULL;

  return Py_BuildValue(""); /* return none */
}/*end... paramclose*/

/*
 * This function will set parameters.
 */
static PyObject *_pset(PyObject *self, /*i: Used by Python */
PyObject *args /*i: Python tuple of the arguments */)
{
  char* funcName = "pset";
  char *pfile = NULL; /* parameter file */
  char *parname = NULL; /* name of the parameter */
  char *value = NULL; /* name of the value to set */
  Py_ssize_t pos = 0; /* position in the dictionary*/

  PyObject *dictKey = NULL; /* Used to extract keys from dictionary */
  PyObject *dictValue = NULL; /* Used to extract values from dictionary */
  paramfile *paramFptr = NULL; /* file pointer for dictionary */
  PyObject *py_paramDict = NULL; /* Python Dict Object of the parameters and values */
  PyObject *py_fptr = NULL; /* fileptr python object */

  if (PyTuple_Size(args) == 3)
  {
    /* Must handle two cases for paramFile, it can be a string or a Param_File_Type */
    if (!PyObject_TypeCheck(PyTuple_GET_ITEM(args, 0), &Param_File_Type))
    {
      if (!PyArg_ParseTuple(args, "sss", &pfile, &parname, &value))
      {
        sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
        PyErr_SetString(PyExc_TypeError, parmsg);
        return NULL;
      }/*end if... */

      paramFptr = paramopen(pfile, NULL, 0, "rwL");
      if (paramFptr == NULL)
      {
        sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
        PyErr_SetString(PyExc_IOError, parmsg);
        return NULL;
      }/*end if... */

      (void)pputstr(paramFptr, parname, value);

      if (paramFptr != NULL)
      {
        paramclose(paramFptr);
      }/*end if... */
    }/*end if... */
    else
    {
      if (!PyArg_ParseTuple(args, "O!ss", &Param_File_Type, &py_fptr, &parname,
                            &value))
      {
        sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
        PyErr_SetString(PyExc_TypeError, parmsg);
        return NULL;
      }/*end if... */

      paramFptr = ((Param_File*)(py_fptr))->fptr;
      if (paramFptr == NULL)
      {
        sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
        PyErr_SetString(PyExc_IOError, parmsg);
        return NULL;
      }/*end if... */

      (void)pputstr(paramFptr, parname, value);
    }/*end else...is Param_File_Type*/
  }/*end if... args == 3*/
  else
  {
    /* Must handle two cases for paramFile, it can be a string or a Param_File_Type */
    if (!PyObject_TypeCheck(PyTuple_GET_ITEM(args, 0), &Param_File_Type))
    {
      if (!PyArg_ParseTuple(args, "sO", &pfile, &py_paramDict))
      {
        sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
        PyErr_SetString(PyExc_TypeError, parmsg);
        return NULL;
      }/*end if... */

      if (!PyDict_Check(py_paramDict) && (PyDict_Size(py_paramDict) != 0))
      {
        sprintf(parmsg, PARAMIO_GENERIC_MSG, funcName,
                (char*)"Could not create dictionary");
        PyErr_SetString(PyExc_IOError, parmsg);
        return NULL;
      }/*end if... */

      paramFptr = paramopen(pfile, NULL, 0, "rwL");
      if (paramFptr == NULL)
      {
        sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
        PyErr_SetString(PyExc_IOError, parmsg);
        return NULL;
      }/*end if... */

      while (PyDict_Next(py_paramDict, &pos, &dictKey, &dictValue))
      {
        parname = PyString_AsString(dictKey); /*get the parameter*/
        value = PyString_AsString(dictValue); /*get the value*/
        (void)pputstr(paramFptr, parname, value);
      }/*end while... next value in dictionary */

      if (paramFptr != NULL)
      {
        paramclose(paramFptr);
      }/*end if... */
    }/*end if... is Param_File_Type */
    else
    {
      if (!PyArg_ParseTuple(args, "O!O", &Param_File_Type, &py_fptr,
                            &py_paramDict))
      {
        sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
        PyErr_SetString(PyExc_TypeError, parmsg);
        return NULL;
      }/*end if... */

      paramFptr = ((Param_File*)(py_fptr))->fptr;
      if (paramFptr == NULL)
      {
        sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
        PyErr_SetString(PyExc_IOError, parmsg);
        return NULL;
      }/*end if... */

      while (PyDict_Next(py_paramDict, &pos, &dictKey, &dictValue))
      {
        parname = PyString_AsString(dictKey); /*get the parameter*/
        value = PyString_AsString(dictValue); /*get the value*/
        (void)pputstr(paramFptr, parname, value);
      }/*end while... next value in dictionary */
    }/*end else... is Param_File_Type */
  }/*end else...PyTuple_Size(args) */

  return Py_BuildValue(""); /* return none */
}/*end... pset*/

/*
 * This function will get the value of the specified parameter(s) from the
 * given parameter file.
 */
static PyObject *_pget(PyObject *self, /*i: Used by Python */
PyObject *args /*i: Python tuple of the arguments */)
{
  char* funcName = "pget";
  char* retVal = NULL;
  static char str[PF_MAX_LINE_LEN + 1];
  char *pfile = NULL; /* parameter file */
  char *parname = NULL; /* parameter name */

  PyObject *py_fptr = NULL; /* fileptr python object */
  paramfile *paramFptr = NULL; /* file pointer for dictionary */

  /*if the first argument is not Param_File_Type open the file*/
  if (!PyObject_TypeCheck(PyTuple_GET_ITEM(args, 0), &Param_File_Type))
  {
    if (!PyArg_ParseTuple(args, "ss", &pfile, &parname))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = paramopen(pfile, NULL, 0, "rH");
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    retVal = pgetstr(paramFptr, parname, str, PF_MAX_LINE_LEN);

    if (retVal == NULL)
    {
      sprintf(parmsg, PARAMIO_PARAM_MSG, funcName);
      PyErr_SetString(PyExc_ValueError, parmsg);
      return NULL;
    }

    if (paramFptr != NULL)
    {
      paramclose(paramFptr);
    }/*end if... */
  }/*end if... */
  else
  {
    if (!PyArg_ParseTuple(args, "O!s", &Param_File_Type, &py_fptr, &parname))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = ((Param_File*)(py_fptr))->fptr;
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    retVal = pgetstr(paramFptr, parname, str, PF_MAX_LINE_LEN);

    if (retVal == NULL)
    {
      sprintf(parmsg, PARAMIO_PARAM_MSG, funcName);
      PyErr_SetString(PyExc_ValueError, parmsg);
      return NULL;
    }
  }/*end else... */

  return Py_BuildValue("s", str); /* return string */
}/*end... pget*/

/*
 * This function will get parameter values.
 */
static PyObject *_pquery(PyObject *self, /*i: Used by Python */
PyObject *args /*i: Python tuple of the arguments */)
{
  char* funcName = "pquery";
  char* retVal = NULL;
  static char str[PF_MAX_LINE_LEN + 1];
  char *pfile = NULL; /* parameter file */
  char *parname = NULL; /* parameter name */

  PyObject *py_fptr = NULL; /* fileptr python object */
  paramfile *paramFptr = NULL; /* file pointer for dictionary */

  /*if the first argument is not Param_File_Type open the file*/
  if (!PyObject_TypeCheck(PyTuple_GET_ITEM(args, 0), &Param_File_Type))
  {
    if (!PyArg_ParseTuple(args, "ss", &pfile, &parname))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = paramopen(pfile, NULL, 0, "r");
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    retVal = pgetstr(paramFptr, parname, str, PF_MAX_LINE_LEN);

    if (retVal == NULL)
    {
      sprintf(parmsg, PARAMIO_PARAM_MSG, funcName);
      PyErr_SetString(PyExc_ValueError, parmsg);
      return NULL;
    }

    if (paramFptr != NULL)
    {
      paramclose(paramFptr);
    }/*end if... */
  }/*end if... */
  else
  {
    if (!PyArg_ParseTuple(args, "O!s", &Param_File_Type, &py_fptr, &parname))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = ((Param_File*)(py_fptr))->fptr;
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    retVal = pgetstr(paramFptr, parname, str, PF_MAX_LINE_LEN);

    if (retVal == NULL)
    {
      sprintf(parmsg, PARAMIO_PARAM_MSG, funcName);
      PyErr_SetString(PyExc_ValueError, parmsg);
      return NULL;
    }
  }/*end else... */

  return Py_BuildValue("s", str); /* return string */
}/*end... pquery*/

/*
 * This function will find out whether parameters and parameter files
 * are accessible
 */
static PyObject *_paccess(PyObject *self, /*i: Used by Python */
PyObject *args /*i: Python tuple of the arguments */)
{
  char* funcName = "paccess";
  static char *str = NULL; /* string return from paramgetpath */
  char *pfile = NULL; /* parameter file */
  char *pmode = NULL; /* parameter name */
  char *param = NULL; /* parameter for Param_File arg */

  PyObject *py_fptr = NULL; /* fileptr python object */
  paramfile *paramFptr = NULL; /* file pointer for dictionary */

  /*if the first argument is not Param_File_Type open the file*/
  if (!PyObject_TypeCheck(PyTuple_GET_ITEM(args, 0), &Param_File_Type))
  {
    if (!PyArg_ParseTuple(args, "s|s", &pfile, &pmode))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    if (NULL == pmode)
      pmode = "rw";

    paramFptr = paramopen(pfile, NULL, 0, pmode);
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    str = paramgetpath(paramFptr);
    if (paramFptr != NULL)
    {
      paramclose(paramFptr);
    }/*end if... */

    return Py_BuildValue("s", str);
  }/*end if... */
  else
  {
    if (!PyArg_ParseTuple(args, "O!s", &Param_File_Type, &py_fptr, &param))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = ((Param_File*)(py_fptr))->fptr;
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    return Py_BuildValue("i", paccess(paramFptr, param));
  }/*end else... */
}/*end... paccess*/

/*
 * This function will return the location of the parameter file for
 * the given argument.
 */
static PyObject *_paramgetpath(PyObject *self, /*i: Used by Python */
PyObject *args /*i: Python tuple of the arguments */)
{
  char* funcName = "paramgetpath";
  static char *str = NULL; /* string return from paramgetpath */
  char *pfile = NULL; /* parameter file */

  PyObject *py_fptr = NULL; /* fileptr python object */
  paramfile *paramFptr = NULL; /* file pointer for dictionary */

  /*if the first argument is not Param_File_Type open the file*/
  if (!PyObject_TypeCheck(PyTuple_GET_ITEM(args, 0), &Param_File_Type))
  {
    if (!PyArg_ParseTuple(args, "s", &pfile))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = paramopen(pfile, NULL, 0, "r");
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    str = paramgetpath(paramFptr);
    if (paramFptr != NULL)
    {
      paramclose(paramFptr);
    }/*end if... */
  }/*end if... */
  else
  {
    if (!PyArg_ParseTuple(args, "O!", &Param_File_Type, &py_fptr))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = ((Param_File*)(py_fptr))->fptr;
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    str = paramgetpath(paramFptr);
  }/*end else... */

  return Py_BuildValue("s", str);
}/*end... paramgetpath*/

/*
 * This function will restore the parameter values in the specified
 * parameter file to the system defaults.
 */
static PyObject *_punlearn(PyObject *self, /*i: Used by Python */
PyObject *args /*i: Python tuple of the arguments */)
{
  char* funcName = "punlearn";
  paramfile fptr = NULL; /* the file pointer */
  char *pfile = NULL; /* parameter file */
  int argc = 0; /* count of optional parameter files */
  int ii; /* loop variable */

  PyObject *py_paramlist = NULL; /* Python List Object of the parameter files */
  PyObject *param = NULL; /* Each parameter file within the py_paramlist list */

  if (PyObject_TypeCheck(PyTuple_GET_ITEM(args, 0), &PyList_Type))
  {
    if (!PyArg_ParseTuple(args, "O", &py_paramlist))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    argc = PyList_GET_SIZE(py_paramlist); /* get the number of params in the list */

    for (ii = 0; ii < argc; ii++)
    {
      param = PyList_GetItem(py_paramlist, ii); /*get the iith param in the py_paramlist */
      fptr = paramopen(PyString_AsString(param), NULL, 0, "rwR"); /*open file and reset params*/
    }/*end for... ii < argc */
  }/*end if... */
  else
  {
    if (!PyArg_ParseTuple(args, "s", &pfile))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */
    fptr = paramopen(pfile, NULL, 0, "rwR"); /*open file and reset params*/
  }/*end else...PyList_Type*/

  if (fptr != NULL)
  {
    paramclose(fptr);
  }/*end if... */
  else
  {
    sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
    PyErr_SetString(PyExc_IOError, parmsg);
    return NULL;
  }/*end else...*/

  return Py_BuildValue(""); /* return none */
}/*end... punlearn*/

/*
 * This function will read/write individual parameter values.
 */
static PyObject *_pgeti(PyObject *self, /*i: Used by Python */
PyObject *args /*i: Python tuple of the arguments */)
{
  char* funcName = "pgeti";
  int value = 0; /* return value */
  char *pfile = NULL; /* parameter file */
  char *parname = NULL; /* parameter name */

  PyObject *py_fptr = NULL; /* fileptr python object */
  paramfile *paramFptr = NULL; /* file pointer for dictionary */

  /*if the first argument is not Param_File_Type open the file*/
  if (!PyObject_TypeCheck(PyTuple_GET_ITEM(args, 0), &Param_File_Type))
  {
    if (!PyArg_ParseTuple(args, "ss", &pfile, &parname))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = paramopen(pfile, NULL, 0, "r");
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    value = pgeti(paramFptr, parname);

    if (paramFptr != NULL)
    {
      paramclose(paramFptr);
    }/*end if... */
  }/*end if... */
  else
  {
    if (!PyArg_ParseTuple(args, "O!s", &Param_File_Type, &py_fptr, &parname))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = ((Param_File*)(py_fptr))->fptr;
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    value = pgeti(paramFptr, parname);
  }/*end else... */

  return Py_BuildValue("i", value); /* return int type */
}/*end... pgeti*/

/*
 * This function will provide read and write access and typecast the
 * variable into the given type.
 */
static PyObject *_pputi(PyObject *self, /*i: Used by Python */
PyObject *args /*i: Python tuple of the arguments */)
{
  char* funcName = "pputi";
  int value = 0; /* return value */
  char *pfile = NULL; /* parameter file */
  char *parname = NULL; /* parameter name */

  PyObject *py_fptr = NULL; /* fileptr python object */
  paramfile *paramFptr = NULL; /* file pointer for dictionary */

  /*if the first argument is not Param_File_Type open the file*/
  if (!PyObject_TypeCheck(PyTuple_GET_ITEM(args, 0), &Param_File_Type))
  {
    if (!PyArg_ParseTuple(args, "ssi", &pfile, &parname, &value))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = paramopen(pfile, NULL, 0, "rwL");
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    pputi(paramFptr, parname, value);

    if (paramFptr != NULL)
    {
      paramclose(paramFptr);
    }/*end if... */
  }/*end if... */
  else
  {
    if (!PyArg_ParseTuple(args, "O!si", &Param_File_Type, &py_fptr, &parname,
                          &value))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = ((Param_File*)(py_fptr))->fptr;
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    pputi(paramFptr, parname, value);
  }/*end else... */

  return Py_BuildValue(""); /* return none */
}/*end... pputi*/

/*
 * This function will read/write individual parameter values.
 */
static PyObject *_pgetb(PyObject *self, /*i: Used by Python */
PyObject *args /*i: Python tuple of the arguments */)
{
  char* funcName = "pgetb";
  int value = 0; /* return value */
  char *pfile = NULL; /* parameter file */
  char *parname = NULL; /* parameter name */

  PyObject *py_fptr = NULL; /* fileptr python object */
  paramfile *paramFptr = NULL; /* file pointer for dictionary */

  /*if the first argument is not Param_File_Type open the file*/
  if (!PyObject_TypeCheck(PyTuple_GET_ITEM(args, 0), &Param_File_Type))
  {
    if (!PyArg_ParseTuple(args, "ss", &pfile, &parname))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = paramopen(pfile, NULL, 0, "r");
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    value = pgeti(paramFptr, parname);

    if (paramFptr != NULL)
    {
      paramclose(paramFptr);
    }/*end if... */
  }/*end if... */
  else
  {
    if (!PyArg_ParseTuple(args, "O!s", &Param_File_Type, &py_fptr, &parname))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = ((Param_File*)(py_fptr))->fptr;
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    value = pgeti(paramFptr, parname);
  }/*end else... */

  return Py_BuildValue("i", value); /* return none */
}/*end... pgetb*/

/*
 * This function will provide read and write access and typecast the
 * variable into the given type.
 */
static PyObject *_pputb(PyObject *self, /*i: Used by Python */
PyObject *args /*i: Python tuple of the arguments */)
{
  char* funcName = "pputb";
  int value = 0; /* return value */
  char *pfile = NULL; /* parameter file */
  char *parname = NULL; /* parameter name */

  PyObject *py_fptr = NULL; /* fileptr python object */
  paramfile *paramFptr = NULL; /* file pointer for dictionary */

  /*if the first argument is not Param_File_Type open the file*/
  if (!PyObject_TypeCheck(PyTuple_GET_ITEM(args, 0), &Param_File_Type))
  {
    if (!PyArg_ParseTuple(args, "ssi", &pfile, &parname, &value))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = paramopen(pfile, NULL, 0, "rwL");
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    pputi(paramFptr, parname, value);

    if (paramFptr != NULL)
    {
      paramclose(paramFptr);
    }/*end if... */
  }/*end if... */
  else
  {
    if (!PyArg_ParseTuple(args, "O!si", &Param_File_Type, &py_fptr, &parname,
                          &value))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = ((Param_File*)(py_fptr))->fptr;
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    pputi(paramFptr, parname, value);
  }/*end else... */

  return Py_BuildValue(""); /* return none */
}/*end... pputb*/

/*
 * This function will read/write individual parameter values.
 */
static PyObject *_pgetd(PyObject *self, /*i: Used by Python */
PyObject *args /*i: Python tuple of the arguments */)
{
  char* funcName = "pgetd";
  double value = 0; /* return value */
  char *pfile = NULL; /* parameter file */
  char *parname = NULL; /* parameter name */

  PyObject *py_fptr = NULL; /* fileptr python object */
  paramfile *paramFptr = NULL; /* file pointer for dictionary */

  /*if the first argument is not Param_File_Type open the file*/
  if (!PyObject_TypeCheck(PyTuple_GET_ITEM(args, 0), &Param_File_Type))
  {
    if (!PyArg_ParseTuple(args, "ss", &pfile, &parname))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = paramopen(pfile, NULL, 0, "r");
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    value = pgetd(paramFptr, parname);

    if (paramFptr != NULL)
    {
      paramclose(paramFptr);
    }/*end if... */
  }/*end if... */
  else
  {
    if (!PyArg_ParseTuple(args, "O!s", &Param_File_Type, &py_fptr, &parname))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = ((Param_File*)(py_fptr))->fptr;
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    value = pgetd(paramFptr, parname);
  }/*end else... */

  return Py_BuildValue("d", value); /* return double type */
}/*end... pgetd*/

/*
 * This function will provide read and write access and typecast the
 * variable into the given type.
 */
static PyObject *_pputd(PyObject *self, /*i: Used by Python */
PyObject *args /*i: Python tuple of the arguments */)
{
  char* funcName = "pputd";
  double value = 0; /* return value */
  char *pfile = NULL; /* parameter file */
  char *parname = NULL; /* parameter name */

  PyObject *py_fptr = NULL; /* fileptr python object */
  paramfile *paramFptr = NULL; /* file pointer for dictionary */

  /*if the first argument is not Param_File_Type open the file*/
  if (!PyObject_TypeCheck(PyTuple_GET_ITEM(args, 0), &Param_File_Type))
  {
    if (!PyArg_ParseTuple(args, "ssd", &pfile, &parname, &value))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = paramopen(pfile, NULL, 0, "rwL");
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    pputd(paramFptr, parname, value);

    if (paramFptr != NULL)
    {
      paramclose(paramFptr);
    }/*end if... */
  }/*end if... */
  else
  {
    if (!PyArg_ParseTuple(args, "O!sd", &Param_File_Type, &py_fptr, &parname,
                          &value))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = ((Param_File*)(py_fptr))->fptr;
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    pputd(paramFptr, parname, value);
  }/*end else... */

  return Py_BuildValue(""); /* return none */
}/*end... pputd*/

/*
 * This function will read/write individual parameter values.
 */
static PyObject *_pgetstr(PyObject *self, /*i: Used by Python */
PyObject *args /*i: Python tuple of the arguments */)
{
  char* funcName = "pgetstr";
  char* retVal = NULL;
  static char value[PF_MAX_LINE_LEN + 1]; /* return value */
  char *pfile = NULL; /* parameter file */
  char *parname = NULL; /* parameter name */

  PyObject *py_fptr = NULL; /* fileptr python object */
  paramfile *paramFptr = NULL; /* file pointer for dictionary */

  /*if the first argument is not Param_File_Type open the file*/
  if (!PyObject_TypeCheck(PyTuple_GET_ITEM(args, 0), &Param_File_Type))
  {
    if (!PyArg_ParseTuple(args, "ss", &pfile, &parname))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = paramopen(pfile, NULL, 0, "r");
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    retVal = pgetstr(paramFptr, parname, value, PF_MAX_LINE_LEN);

    if(retVal == NULL)
    {
      sprintf(parmsg, PARAMIO_PARAM_MSG, funcName);
      PyErr_SetString(PyExc_ValueError, parmsg);
      return NULL;
    }

    if (paramFptr != NULL)
    {
      paramclose(paramFptr);
    }/*end if... */
  }/*end if... */
  else
  {
    if (!PyArg_ParseTuple(args, "O!s", &Param_File_Type, &py_fptr, &parname))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = ((Param_File*)(py_fptr))->fptr;
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    retVal = pgetstr(paramFptr, parname, value, PF_MAX_LINE_LEN);

    if(retVal == NULL)
    {
      sprintf(parmsg, PARAMIO_PARAM_MSG, funcName);
      PyErr_SetString(PyExc_ValueError, parmsg);
      return NULL;
    }
  }/*end else... */

  return Py_BuildValue("s", value); /* return string type */
}/*end... pgetstr*/

/*
 * This function will provide read and write access and typecast the
 * variable into the given type.
 */
static PyObject *_pputstr(PyObject *self, /*i: Used by Python */
PyObject *args /*i: Python tuple of the arguments */)
{
  char* funcName = "pputstr";
  char *value = 0; /* return value */
  char *pfile = NULL; /* parameter file */
  char *parname = NULL; /* parameter name */

  PyObject *py_fptr = NULL; /* fileptr python object */
  paramfile *paramFptr = NULL; /* file pointer for dictionary */

  /*if the first argument is not Param_File_Type open the file*/
  if (!PyObject_TypeCheck(PyTuple_GET_ITEM(args, 0), &Param_File_Type))
  {
    if (!PyArg_ParseTuple(args, "sss", &pfile, &parname, &value))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = paramopen(pfile, NULL, 0, "rwL");
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    (void)pputstr(paramFptr, parname, value);

    if (paramFptr != NULL)
    {
      paramclose(paramFptr);
    }/*end if... */
  }/*end if... */
  else
  {
    if (!PyArg_ParseTuple(args, "O!ss", &Param_File_Type, &py_fptr, &parname,
                          &value))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = ((Param_File*)(py_fptr))->fptr;
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */

    (void)pputstr(paramFptr, parname, value);
  }/*end else... */

  return Py_BuildValue(""); /* return none */
}/*end... pputstr*/

/*
 * Returns a list of parameter names in current parameter file
 */
static PyObject *_plist(PyObject *self, /*i: Used by Python */
PyObject *args /*i: Python tuple of the arguments */)
{
  char* funcName = "plist";
  char *pfile = NULL; /* parameter file */

  PyObject *py_fptr = NULL; /* fileptr python object */
  paramfile *paramFptr = NULL; /* file pointer for dictionary */

  if ((NULL == args) || (0 == PyTuple_Size(args)))
  {
    sprintf(parmsg, PARAMIO_GENERIC_MSG, funcName,
            (char*)"Must supply parameter name or parameter object");
    PyErr_SetString(PyExc_TypeError, parmsg);
    return NULL;
  }

  /*if the first argument is not Param_File_Type open the file*/
  if (!PyObject_TypeCheck(PyTuple_GetItem(args, 0), &Param_File_Type))
  {
    if (!PyArg_ParseTuple(args, "s", &pfile))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = paramopen(pfile, NULL, 0, "rH");
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */
  }/*end if... */
  else
  {
    if (!PyArg_ParseTuple(args, "O!", &Param_File_Type, &py_fptr))
    {
      sprintf(parmsg, PARAMIO_TYPE_ERROR_MSG, funcName);
      PyErr_SetString(PyExc_TypeError, parmsg);
      return NULL;
    }/*end if... */

    paramFptr = ((Param_File*)(py_fptr))->fptr;
    if (paramFptr == NULL)
    {
      sprintf(parmsg, PARAMIO_OPEN_FILE_MSG, funcName);
      PyErr_SetString(PyExc_IOError, parmsg);
      return NULL;
    }/*end if... */
  }/*end else... */

  pmatchlist pList;
  if (NULL == (pList = pmatchopen(paramFptr, "*")))
  {
    sprintf(parmsg, PARAMIO_GENERIC_MSG, funcName, (char*)"Parammatch failed");
    PyErr_SetString(PyExc_Exception, parmsg);
    return NULL;
  }

  PyObject *pylist;
  if (NULL == (pylist = PyList_New((Py_ssize_t)0)))
  {
    sprintf(parmsg, PARAMIO_GENERIC_MSG, funcName,
            (char*)"Could not create new list");
    PyErr_SetString(PyExc_Exception, parmsg);
    return NULL;
  }

  char *curPar;
  while (NULL != (curPar = pmatchnext(pList)))
  {
    PyObject *pstr;
    if (NULL == (pstr = PyString_FromString(curPar)))
    {
      sprintf(parmsg, PARAMIO_GENERIC_MSG, funcName,
              (char*)"Could not convert char * to string");
      PyErr_SetString(PyExc_Exception, parmsg);
      return NULL;
    }

    if (0 != PyList_Append(pylist, pstr))
    {
      sprintf(parmsg, PARAMIO_GENERIC_MSG, funcName,
              (char*)"Could not append string to list");
      PyErr_SetString(PyExc_Exception, parmsg);
      return NULL;
    }
  } /* end ... while */

  pmatchclose(pList);

  /*
   * If this routine opened, then needs to close it.
   */
  if (NULL != pfile)
  {
    paramclose(paramFptr);
  }

  return (pylist);
}/*end... plist*/
