/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afblock.c                                                   */
/*                                                               */
/*   Functions that create and work with ASCII file type         */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"
#include "../ftkernel/ftkernelP.h"
#include "ryi.h"

extern dmBool ftp_subspace_is_gti( char* tabname );
/*static void afp_block_write_fixed_keys( dmBlock* block );*/

static int bRowBufOverRun = 0;


/***************************************************************/
/*                                                             */
/* Function:   afp_find_interesting_block_no                   */
/*                                                             */
/* Description:                                                */
/*   This routine encodes the concept of 'interesting block'   */
/*   The rules are:                                            */
/*                                                             */
/*  1) If there are no blocks, return zero.                    */
/*  2) If there is only one block, return one.                 */
/*  3) If the block name contains "GTI", it's not interesting  */
/*      (except for case 2)                                    */
/*  4) If the block name contains "WMAP", it's not interesting */
/*      (except for case 2)                                    */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDataset *ds  - pointer to dataset      */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long   BlockNumber  - block number               */
/*                                                             */
/***************************************************************/
long afp_find_interesting_block_no( dmDataset* ds )
{
  afDataset  *pafAsciiDataSet = NULL;
  long        BlockNumber     = 0;
  long        nblocks         = 0;
  char*       bname           = NULL;
  dmBlockType btype;

  int         null;     /* null image? */
  int         good;     /* good block? */
  long  first_good = 0; /* first good block */

  /* Get the Ascii dataset pointer from within the dataset */
  pafAsciiDataSet = (afDataset*)dmi_dataset_get_kds( ds );
  if ( ! pafAsciiDataSet )
   return BlockNumber;

  nblocks = dmi_dataset_get_n( ds );

  /* if there are no blocks, return         */
  /* if there is only one block.. return it */
  if ( nblocks <= 1 )
    return nblocks;

  /* MCD NOTE: (7/31/2007)  This segment needs expanding to 
     be more equivalent to the FITS counterpart... 
     MCD NOTE: (12/14/2007) Updated to ignore WMAP and GTI
     blocks unless they are the only block that is not the null primary.
     MCD NOTE2:  At this point, there does not seem to be 
     any place storing the info that an image is NULL.
     (blocks are not open yet, just identified).  Set up to 
     consider block1 = dmIMAGE with name == PRIMARY as uninteresting.
  */
  for( BlockNumber = 1; BlockNumber <= nblocks; BlockNumber++ ) 
  {
    bname = dmi_dataset_get_block_name( ds, BlockNumber );
    btype = dmi_dataset_get_block_type( ds, BlockNumber );
    null  = ( BlockNumber == 1 ) && (btype == dmIMAGE ) && (cs_uceq( bname, "PRIMARY"));
    good  = ( btype == dmIMAGE ) || (btype == dmTABLE );

    /* retain first good, non-NULL block in case all are rejected */
    if ( good && !null && first_good==0){ first_good = BlockNumber; }

    /* Screen blocks that should generally not be considered interesting */
    if ( bname && strstr( bname, "GTI" ) != NULL ){ continue;}
    if ( bname && strstr( bname, "WMAP" ) != NULL ){ continue;}

    if ( good && !null )
      return BlockNumber;
  }

  /* We did not find a non-GTI or non-WMAP block.               */ 
  /* Any 'good' block is more interesting than the NULL primary */
  if ( first_good ){ return first_good; }

  /* do not return block number bigger than number of blocks */
  if ( BlockNumber > nblocks )
    BlockNumber = 1;

  return BlockNumber;
}


/****************************************************************************/
/*                                                                          */
/* Function:   afp_block_open                                               */
/*                                                                          */
/* Description:                                                             */
/*   Open a block in the dataset.  The create flag controls whether it      */
/*   is able to create the block if it does not exist.                      */
/*                                                                          */
/*                                                                          */
/*  Input parameters:                                                       */
/*                dmDataset *ds             - pointer to dataset            */
/*                long       blocknum       - block number                  */
/*                dmBool     create         - create the block?             */
/*                dmBool     open_subspace  - true or false                 */
/*                const char *option        - ds option string              */
/*                                                                          */
/*  Output parameters:                                                      */
/*                                                                          */
/*                                                                          */
/*  Returns:                                                                */
/*     dmBlock *pBlockPtr  - pointer to open block                          */
/*                                                                          */
/****************************************************************************/
dmBlock* afp_block_open( dmDataset*  ds,
                         long        blocknum,
                         dmBool      create,
                         dmBool      open_subspace,
                         const char* option )
{
  dmErrCode   status = dmSUCCESS;
  dmBlock*    block  = NULL;

  afBlock    *ablock = NULL;
  afBuffer   *abuf   = NULL;
  afDataset  *ads    = NULL;
  afStyle    *astyle = NULL;
  afFlavour   flav   = AK_FLAV_UNK;
  const char *pFlavorString = NULL;
  long        nrows  = 0;
  char        buf[CARD_SIZE];



  if ( blocknum == 0 )
    blocknum = 1;

  /* Get style info from dataset */
  /* NOTE: this should already be correctly setup from the dataset_open stage */
  ads    = (afDataset*)dmi_dataset_get_kds( ds );
  astyle = afk_dataset_get_style( ads );
  flav   = astyle->flavor;

 
  /* Check if it is OK to open this block */
  if ( blocknum > 1 && !afk_style_get_multiple( astyle ))
  {
    pFlavorString =  afk_style_get_flavorString( astyle );

    if ( pFlavorString )
      strcpy( buf, pFlavorString);

    sprintf( dm_errbuf, 
	    "TEXT kernel specification %s does not support multiple blocks.",
	    buf );
    dmu_err_set( DM_ERR_OPEN_FAIL, dm_errbuf );
    return NULL;
  }
 
  /* Move to requested block */
  status = afk_dataset_moveto_block( ads, blocknum );

  if ( status != dmSUCCESS )
  {
     sprintf( dm_errbuf,
	     "Couldn't move to block %ld in dataset %s.\n",
	      blocknum, ads->name );
     dmu_err_set( DM_ERR_OPEN_FAIL, dm_errbuf );
     return NULL;
  }

  block = dmi_block_alloc( ds, blocknum, create );
  if ( !block )
    return NULL;


  ablock = afk_block_alloc();
  if ( !ablock )
  {
    dmi_block_free (block);
    return NULL;
  }

  ablock->no = blocknum;
  ablock->block_type = dmi_block_get_type( block );

  afk_block_set_kds( ablock, ads );
  dmi_block_set_kb( block, (dmkBlock*)ablock );

  /* Set HINT struct in the block from input options..     */
  /* is this needed?  These fields are all covered by      */
  /* the dataset style and shouldn't change for each block */

  if ( option )
    status = afp_set_hint_list( block, option );

  /* Get the buffer from this block */
  abuf = afk_block_get_buffer( ablock );

  /* Give block buffer access to the dataset style for parsing */
  /* Code Change  Bug 11905 August 21, 2009                    */
  /*  added case for cscrdb format                             */
  /* Code Change  Bug 11940 March 9, 2010                        */
  /*  added case for tsv format                                */

  switch ( flav )
  {
    case AK_FLAV_DTF:
    case AK_FLAV_SIMPLE:
    case AK_FLAV_RAW:
    case AK_FLAV_CSCRDB:
    case AK_FLAV_TSV:
    default:
      afk_buffer_set_style( abuf, astyle );
      break;
  }


  /* NOTE: The above is roughly equivalent to ftp_block_open..  */
  /*       The following combines elements of ftp_block_init, ftp_table_open, ftp_subspace_open */

  if ( create )
  {
    /* Allocate a new keylist */
    ablock->keylist = dmu_keylist_alloc();

    if (ablock->block_type == dmIMAGE )
    {
      /* Generate placeholder descriptors for image 'column' */
      afp_block_init_cols( block );
    }
  }
  else
  {
    /* read in header keywords, comments, and kernel level column defs */
    afp_block_read_header( block );


   /* init DM level column definitions */
   afp_block_init_cols( block );

    // Code Change - Bug 9023  August 11, 2006
    // If file is IMAGE, there is one row and one column
    if ( ablock->block_type == dmIMAGE )
    {
      nrows = 1;
    }
    else
    {
      /* This is hacky and will slow things down significantly,         */
      /* but, currently DM forces us to give it an exact number of rows */
      /* Code Change  Bug 11905 August 21, 2009                         */
      /*  added case for cscrdb format                                  */
      /* Code Change  Bug 11940 March 9, 2010                             */
      /*  added case for tsv format                                     */

      switch ( flav )
      {
        case AK_FLAV_DTF:
          nrows = afk_dtf_get_nrows( ablock );
          break;
        case AK_FLAV_SIMPLE:
        case AK_FLAV_RAW:
        case AK_FLAV_CSCRDB:
        case AK_FLAV_TSV:
          nrows = afk_raw_get_nrows( ablock );
          break;
        default:
          break;
      }
    }

    dmi_block_set_nrows( block, nrows );


    // Code change - Bug 9222  June 8, 2006
    // check here for any buffer overruns encountered on the read

    if ( bRowBufOverRun == 1 )
    {
      bRowBufOverRun = -1;
      dmu_err_set(DM_WARN_DATA_TRUNCATED, "Some data rows exceed length limits.  Their data will be truncated.\n");

    }
  }

  dmi_block_init( block );


  if ( open_subspace )
    afp_block_open_subspace( block, create );

  return block;
}




/***************************************************************/
/*                                                             */
/* Function:   afp_block_read_header                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock *block - pointer to block        */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  ErrorCode                             */
/*                                                             */
/***************************************************************/
dmErrCode afp_block_read_header( dmBlock* block )
{
  dmDataset      *ds = NULL;
  dmDescriptor   *dd = NULL;

  afDataset      *ads    = NULL;
  afFlavour       flav   = AK_FLAV_UNK;
  afStyle        *astyle = NULL;
  afBlock        *ablock = NULL;
  void           *value  = NULL;
  dmuKey         *ukey   = NULL;
 
  long       ii     = 0;
  long       keyno  = 1;
  dmErrCode  status = dmSUCCESS;
  dmuKeyList keylist;



  // Get pointer to dataset from block pointer
  ds  = dmi_block_get_ds( block );

  // Get pointer to Adcii dataset from dataset
  ads = (afDataset*)dmi_dataset_get_kds( ds );

  // Get pointer to style structure from Ascii dataset pointer
  astyle = afk_dataset_get_style( ads );

  // Get pointer to block requested
  ablock = (afBlock*)dmi_block_get_kb( block );


  // Get file flavor
  flav = astyle->flavor;



  // Read header based on file type
  switch ( flav )
  {
    case AK_FLAV_DTF:
    {
      // If the dtf header information is missing, treat
      // file as raw

     if ( astyle->has_header )
        status = afk_dtf_read_header( ablock, astyle );
      else
        status = afk_raw_read_header( ablock, astyle );
      break;
    }

    case AK_FLAV_SIMPLE:
      status = afk_simple_read_header( ablock, astyle );
      break;

    // Code Change  Bug 11905  August 21, 2009
    case AK_FLAV_CSCRDB:
      status = afk_cscrdb_read_header( ablock, astyle );
      break;

    case AK_FLAV_RAW:
    default:
      status = afk_raw_read_header( ablock, astyle );
      break;

    // Code Change  Bug 11940  March 9, 2010
    case AK_FLAV_TSV:
      status = afk_tsv_read_header( ablock, astyle );
      break;
  }


  keylist = ablock->keylist;

  if ( keylist )
  {
    /* now need to allocate the key descriptors */
    for ( ii = 0; ii < keylist->n; ii++ )
    {
      ukey = keylist->data[ii];

      if ( !ukey->mask )
      {
	/* Check if key already exists */
	dd = dmi_key_open( block, ukey->name );
	if (!dd)
	{
	  value = dmu_key_value( ukey );
	  dd = dmi_key_alloc( block, keyno, ukey->name, ukey->datatype, 
			      value, 1, ukey->unit, ukey->desc );

	  dmi_descriptor_set_kdd( dd, (void*)ukey );
	  ++keyno;
	}
	else
	{
	  printf("Key %s present more than once in header.\n", ukey->name );
	}
      }
    }
  }

  return status;
}




/***************************************************************/
/*                                                             */
/* Function:   afp_block_init_cols                             */
/*                                                             */
/* Description:                                                */
/*             use the kernel block information to initialize  */
/*             the block column structure                      */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock *block - pointer to block        */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  ErrorCode                             */
/*                                                             */
/***************************************************************/
dmErrCode afp_block_init_cols( dmBlock* block )
{
  dmDataset  *ds     = NULL;
  dmErrCode   status = dmSUCCESS;

  afDataset  *ads    = NULL;
  afFlavour   flav   = AK_FLAV_UNK;


  // Get dataset pointer from block pointer
  ds  = dmi_block_get_ds( block );

  // Get Ascii dataset pointer from dataset pointer
  ads = (afDataset*)dmi_dataset_get_kds( ds );

  // Get file flavor
  flav   = afk_dataset_get_flavour( ads );


  switch (flav)
  {
    case AK_FLAV_DTF:
      afp_dtf_init_cols( block );
      break;

    case AK_FLAV_SIMPLE:
      afp_simple_init_cols( block );
      break;

    // Code Change  Bug 11905  August 21, 2009
    case AK_FLAV_CSCRDB:
      afp_cscrdb_init_cols( block );
      break;

    case AK_FLAV_RAW:
    default:
       afp_raw_init_cols( block );
       break;

    // Code Change  Bug 11940  March 9, 2010
    case AK_FLAV_TSV:
      afp_tsv_init_cols( block );
      break;
  } 

  return status;  
}




/***************************************************************/
/*                                                             */
/* Function:   afp_block_open_subspace                         */
/*                                                             */
/* Description:                                                */
/*             taken from ftkernel code                        */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock *block - pointer to block        */
/*                    dmBool   create - true or false          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  ErrorCode                             */
/*                                                             */
/***************************************************************/
dmErrCode afp_block_open_subspace( dmBlock* block, dmBool create )
{
  dmDataset      *ds       = NULL;
  dmDescriptor   *dd       = NULL;
  dmDescriptor  **list     = NULL;
  dmSubspace     *subspace = NULL;
  dmuKey         *ukey     = NULL;
  dmuKeyList      keylist;
  dmErrCode       status   = dmSUCCESS;
  dmDataType      type;

  afBlock        *ablock = NULL;
  ftSubspaceCol  *ftss   = NULL;

  long componentNumber = 0;
  long keyNo     = 0;
  long NumKeys   = 0;
  long n         = 0;

  char   typekey[CARD_SIZE]; 
  char   typename[CARD_SIZE];
  char   compNumString[4] = {'\0', '\0', '\0', '\0' };
  char  *pKeyDSVAL = NULL;
  char  *dsval     = NULL;
  char  *tabname   = NULL;
  char  *keyName   = NULL;
  char  *ssname    = NULL;

  long onmax      = 0;      /* Temp for reallocation */
  long nmax       = 0;       /* Actual max entry in list */
  long lnmax      = 0;      /* Space allocated in list */
  long offset     = 0;
  long index      = 0;
  long maxcpt     = 1;
  int  have_time  = UT_FALSE;   /* Is TIME an explicit DSTYP? */
  int  was_table  = UT_FALSE;   /* Was first cpt a table?  */
  int  is_image   = 0;
  int  ii         = 0;
  int  numLength  = 0;
  int  is_compNum = 1;

  /* 
   * The create key is a hack to let us know if we are trying to
   * initialize a subspace on a block we have just created. 
   */

  is_image = dmi_block_get_raw_type( block ) == dmIMAGE;
  ds       = dmi_block_get_ds( block );
  ablock   = (afBlock*)dmi_block_get_kb( block );
  keylist  = ablock->keylist;

  if ( keylist )
  {
    NumKeys = keylist->n;
  }

  // Code Change  Bug 10136  November 7, 2006

  /* Pass 0: count subspace components */
  for ( keyNo = 0; keyNo < NumKeys; keyNo++ )
  {
    ukey    = keylist->data[keyNo];
    keyName = ukey->name;

    // Is this a  DSVAL keyword?
    pKeyDSVAL = strstr(keyName, "DSVAL");
 
    if (pKeyDSVAL)
    {
      // Check for preceeding subspace component number
      // Keyword format is <digits>DSVAL<digits>

      numLength = pKeyDSVAL - keyName;

      if ( numLength )
      {
        strncpy(compNumString, keyName, numLength );
        compNumString[numLength] = '\0';
        is_compNum = 1;

        // Verify that this is truly a number, no alpha chars
        for (ii = 0; ii < numLength; ii++)
          if ( ! isdigit(compNumString[ii] ))
          {
            is_compNum = 0;
            break;
	  }
        
 	if (is_compNum)
	{
          // Extract component number.  If number is larger than
          // current max component number, update max value

          componentNumber = (long) atoi(compNumString);

          if ( componentNumber > maxcpt )
            maxcpt = componentNumber;
        }
      }
    }
  }

  // Limit is 99 for max components
  if (maxcpt > 99)
    maxcpt = 99;       

  // end Code Change  Bug 10136


  /*
   *  Pass 1: Make a subspace descriptor for each raw column
   * If it's an image, we should do it for the axis groups...
   */
  dmi_block_make_default_subspace( block, maxcpt );
  subspace = dmi_block_get_subspace( block );

  /*
   *  Pass 2: Look for DSTYP keywords and assign them
   */
  for ( keyNo = 1; keyNo <= keylist->n; keyNo++ )
  {
    offset  = keyNo - 1;
    ukey    = keylist->data[offset];
    keyName = ukey->name;

    /* Rewrite this to add new descriptors to subspace.
       Preinit subspace with range values */
    /* DSTYPn defines a subspace */

    if ( cs_eq_siz( keyName, "DSTYP", 5 ) && ukey->datatype == DT_CHAR )
    {
      n = atol( keyName + 5 );
      /* Store the n of DSTYPn in a kernel structure */
      ssname = dmu_key_value_c( ukey );
      /* Here should search subspace only */
      /* Watch it: SKY(X,Y) needs to be parsed */
      sprintf( typekey, "DSFORM%ld", n );
      ukey = dmu_keylist_read_value_c( keylist, typekey, typename, CARD_SIZE );

      if ( ukey )
      {
        type = ftp_bt_tform_to_datatype( *typename );
      }
      else
      {
        type = DT_REAL8;
      }

      dd = dmi_subspace_col_init( block, ssname, NULL, type );

      if ( !dd ) /* Internal error */
        return dmFAILURE;

      ftss = afp_get_ftss( block, dd, n ); 
      dmi_descriptor_set_kdd( dd, ftss ); 

      /* Extend list for new entry. Entries may not be in any particular order
       * in the header.
       */
      if ( lnmax < n )
      {
        onmax = lnmax;
        lnmax = 2 * n + 10;
        list = ut_realloc( list, onmax, lnmax, UT_SZ_P );
      } 

      if ( nmax < n )
        nmax = n;
      list[n-1] = dd;
    }
  } // end for key loop

  ablock->max_dstyp = nmax;

  /* Now we have a list of all the DSTYP values ordered by DSTYP index */
  /*
   *  Pass 3: Look for other DS keywords 
   */

  for ( index = 1; index <= nmax; index++ )
  {
    was_table = UT_FALSE;
    dd = list[index-1];

    for ( componentNumber = 1; componentNumber <= maxcpt; componentNumber++ )
    {  
      dmi_block_set_subspace_cpt( block, componentNumber );
      ftp_read_dsval_key( keylist, index, componentNumber, was_table, &dsval, &tabname );

      if ( componentNumber== 1 && tabname )
        was_table = UT_TRUE;

      status = afp_read_subspace_values( dd, componentNumber, dsval, tabname, create );

      if ( status == dmSUCCESS && cs_uceq( dmi_descriptor_get_name( dd ), "TIME" ))
        have_time = UT_TRUE;

      free( tabname );
    }
  }

  free( list ); 

  componentNumber = 1;
  dmi_block_set_subspace_cpt( block, componentNumber );

  /* Remember to look for the default GTI!                                      */
  /*  Since no DSTYP for TIME if we get here, no ftss either, so must create it.*/ 
  /*  Null primary doesn't get one; neither does file without tstart/tstop      */
  /*  Code Change  Bug 10044  June 10, 2010                                                */
  /*    Do not attach default GTI on newly created subspace.                    */
  /*    Do that later using key information                                     */

  if ( !create )
  {
    if ( !is_image && !have_time && maxcpt == 1 && !dmi_block_is_null_primary( block ))
    {
      double tstart  = 0.0;
      char  *tabname = NULL;

      ukey = dmu_keylist_read_typed_value( keylist, "TSTART", DT_REAL8, &tstart );

      if ( ftp_subspace_is_gti( dmi_block_get_name( block )))
      {
        /* A GTI block has itself as a GTI */
        tabname = cs_dup( dmi_block_get_name( block ));
      }
      else
        tabname = ftp_subspace_table_name( ds, NULL );
 
      if ( ( ukey != NULL ) || ( tabname != NULL ) ) 
      {
        double tstop = 0.0;

        dd = dmi_subspace_col_init( block, "TIME", NULL, DT_REAL8 );
        dmi_descriptor_set_unit( dd, "s" );
        dmi_descriptor_clear_update( dd );  /* Don't try flushing this descriptor */
        ftss = afp_get_ftss( block, dd, 0 );
        dmi_descriptor_set_kdd( dd, ftss );
        componentNumber = 1;

        if ( tabname != NULL )
        {
          afp_read_subspace_values( dd, componentNumber, NULL, tabname, create );      
          free( tabname );
        }
        else
          if ( ukey != NULL )
          {
            ukey = dmu_keylist_read_typed_value( keylist, "TSTOP", DT_REAL8, &tstop );

            if ( ukey != NULL )      
              dmi_subspace_col_init_cpt( dd, NULL, 1, DT_REAL8, &tstart, &tstop, 1 );
          }
      } // end if ( ( ukey != NULL ) || ( tabname != NULL ) 
    }// end if ( !is_image && !have_time && maxcpt == 1
  } // end if (!create)

  return dmSUCCESS;
}




/***************************************************************/
/*                                                             */
/* Function:   afp_block_flush                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock *block - pointer to block        */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_block_flush( dmBlock* block )
{
  dmDataset *ds     = NULL;
  afBlock   *ablock = (afBlock*)dmi_block_get_kb( block );
  afBuffer  *abuf   = afk_block_get_buffer( ablock );
  long       moved  = 0; /* need to use this to adjust other blocks */


  afk_buffer_flush( abuf, &moved );

  if ( moved != 0 )
  {
    ds = dmi_block_get_ds( block );
    afp_dataset_notify_moved( ds, block, moved );
  }
}



/***************************************************************/
/*                                                             */
/* Function:   afp_block_print                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    const dmBlock *block - pointer to block  */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_block_print( const dmBlock* block )
{
  afBlock    *ablock = NULL;
  char       *name   = NULL;
  char        dmtype_name[CARD_SIZE];
  dmBlockType type;

  ablock = (afBlock*)dmi_block_get_kb( block );
  name   = dmi_block_get_name( block );

  printf( "Block %ld: %s\n", ablock->no, name );

  type = dmi_block_get_raw_type( block );
  dmu_block_type_name( type, dmtype_name );

  printf( "DM Block type: %s\n", dmtype_name );
  printf( "TEXT HDU type: TABLE\n" );
  printf( "\n" );

  afp_block_print_keys( block );
  afp_block_print_cols( block );
  return;
}




/***************************************************************/
/*                                                             */
/* Function:   afp_block_print_cols                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    const dmBlock *block - pointer to block  */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_block_print_cols( const dmBlock* block )
{
  afBlock *ablock = NULL;
  long     i      = 0;

  ablock = (afBlock*)dmi_block_get_kb( block );

  for ( i = 0; i < ablock->ncols; i++ )
  {
    afk_col_print( ablock->cols[i] );
  }
}



/***************************************************************/
/*                                                             */
/* Function:   afp_block_print_keys                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    const dmBlock *block - pointer to block  */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_block_print_keys( const dmBlock* block )
{
  afBlock    *ablock = NULL;
  dmuKeyList  keylist;

  ablock  = (afBlock*)dmi_block_get_kb( block );
  keylist = afk_block_keylist( ablock );
  dmu_keylist_print( keylist, NULL );
}



/***************************************************************/
/*                                                             */
/* Function:   afp_block_read_col                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock       *block  - pointer to block */
/*                    dmDescriptor  *dd                        */
/*                    long          *llpix  - lower left pt    */
/*                    long          *urpix  - upper right pt   */
/*                    long          *dpix   - number of pts    */
/*                    long           row1                      */
/*                    long           row2                      */
/*                                                             */
/*  Output parameters:                                         */
/*                    void          *value  - output buffer    */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  ErrCode                               */
/*                                                             */
/***************************************************************/
dmErrCode afp_block_read_col( dmBlock      *block,
                              dmDescriptor *dd,
                              long         *llpix,
                              long         *urpix,
                              long         *dpix,
                              long          row1,
                              long          row2,
                              void         *value )
{
  dmErrCode     status = dmSUCCESS;
  dmDataset    *ds     = NULL;
  dmDescriptor *cpt_dd = NULL;
  dmDataType    dataType;
  dmBlockType   blockType;

  afDataset  *ads    = NULL;
  afBlock    *ablock = NULL;
  afCol      *acol   = NULL;
  char       *vptr   = NULL;

  long dim    = 0;
  long cptno  = 0;
  long colLength   = 0;
  long type_nbytes = 0;
  long row_nbytes  = 0;
  long nbytes      = 0;
  long imageOffset = 0;

 
  ds     = dmi_block_get_ds( block );
  ads    = (afDataset*)dmi_dataset_get_kds( ds );
  ablock = (afBlock*)dmi_block_get_kb( block );

  dataType  = dmi_descriptor_get_datatype( dd );
  blockType = ablock->block_type;
  dim       = dmi_descriptor_get_vector_dim( dd );

  if ( dataType != DT_CHAR )
    type_nbytes = ar_mix_type_size( dataType );
  else
    type_nbytes = 1;

  row_nbytes = type_nbytes * ( row2 - row1 + 1 );
  vptr = (char*)value;


  /* loop over vector components */
  for ( cptno = 1; cptno <= dim && status == dmSUCCESS; cptno++ )
  {
    cpt_dd = dmi_descriptor_get_cpt( dd, cptno );

    acol   = (afCol*)dmi_descriptor_get_kdd( cpt_dd );


    // Code Change - Bug 9023 - September 13, 2006
    //  Need to use information in input arguments
    //  llpix and urpix to determine column length
    //  and buffer offset when reading image files
    //  If not an image file, use component descriptor length

    if ( blockType != dmIMAGE )
    {
       colLength = dmi_descriptor_get_length( cpt_dd );

       // The length of character column in ascii kernel is 1024.
       // However, when reading from the dmBuffer, the code expected
       // the column to be 1025 (1024 chars plus null terminator).
       // Adjust for that discrepancy here by calling
       //  dmi_descriptor_get_bytes() instead.

       if ( dataType == DT_CHAR )
         colLength = dmi_descriptor_get_bytes( cpt_dd );
    }
    else
    {
      // Block is type dmIMAGE, use the point values
      //  to determine column length and buffer offset.

      long naxes      =  ablock->naxes;
      long pageAdjust = 0;
      long ii         = 0;

      colLength   = 1;
      imageOffset = 0;

      // Calculate length of image column (i.e. number of values)
      // colLength = Upper right coord - lower left coord + 1

      for (ii = 0; ii < naxes; ii++)
        colLength *= (*(urpix+ii) - *(llpix+ii) + 1);

      // Calculate the offset into the image column for the
      // starting point
      // offset = (lower left coord - 1) * (page adjustment value)
 
      for (ii = 0; ii < naxes; ii++)
      {
        if ( !ii )
          pageAdjust = 1;
        else
          pageAdjust *= ablock->axes[ii-1];
 
        imageOffset += (*(llpix+ii)-1) * pageAdjust;
      }

    }

    status = afk_block_read_col( ablock, acol, dataType, row1, row2, vptr, 
				 colLength, cpt_dd->null, imageOffset );

    /* note: this will not work correctly for a vector-array column
     * the data will come in like  cpt1[[row1[array1]row2[array2]...]cpt2[...
     * May need to re-arrange data.
     */

    if ( colLength > 0 )
    {
      // Code change Bug 12078 December 7, 2009
      // Byte length for Bits is different

      if ( dataType != DT_BIT )
        nbytes = row_nbytes * colLength;
      else
        nbytes = row_nbytes * cpt_dd->size;
    }
    else 
       nbytes = row_nbytes;

    vptr += nbytes;
  }

  return status;
}


/***************************************************************/
/*                                                             */
/* Function:   afp_block_key_write                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock       *block  - pointer to block */
/*                    long           pos                       */
/*                    const char    *name                      */
/*                    dmDataType     type                      */
/*                    const void    *value                     */
/*                    const char    *unit                      */
/*                    const char    *comment                   */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmuKey  *ukey                                    */
/*                                                             */
/***************************************************************/
dmuKey* afp_block_key_write( dmBlock    *block,
                             long        pos,
                             const char *name,
                             dmDataType  type,
                             const void *value,
                             const char *unit,
                             const char *comment )
{
  dmDataset  *ds     = NULL;
  afDataset  *ads    = NULL;
  afFlavour   flav   = 0;
  dmuKey     *ukey   = NULL;
  afBlock    *ablock = NULL;
  dmuKeyList  keylist;
 
  ds  = dmi_block_get_ds( block );
  ads = (afDataset*)dmi_dataset_get_kds( ds );
   
  flav   = afk_dataset_get_flavour( ads );
  ablock = (afBlock*)dmi_block_get_kb( block );

  keylist = ablock->keylist;

  if ( pos == 0 )
  {
    pos = keylist->n + 1;
  }


  switch (flav)
  {
    case AK_FLAV_DTF:
      ukey = afk_dtf_key_write( ablock, pos, name, type, value, unit, comment );
      break;

    case AK_FLAV_SIMPLE:
      ukey = afk_simple_key_write( ablock, pos, name, type, value, unit, comment );
      break;

    // Code Change  Bug 11905  August 21, 2009
    case AK_FLAV_CSCRDB:
      ukey = afk_cscrdb_key_write( ablock, pos, name, type, value, unit, comment );
      break;

    case AK_FLAV_RAW:
    default:
      ukey = afk_raw_key_write( ablock, pos, name, type, value, unit, comment );
      break;

    // Code Change  Bug 11940  March 9, 2010
    case AK_FLAV_TSV:
      ukey = afk_tsv_key_write( ablock, pos, name, type, value, unit, comment );
      break;
  }

 return ukey;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_block_write_comment                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock       *block  - pointer to block */
/*                    long           pos                       */
/*                    const char    *tag                       */
/*                    const char    *comment                   */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmuKey  *ukey                                    */
/*                                                             */
/***************************************************************/
dmErrCode afp_block_write_comment( dmBlock     *block,
                                   long         pos,
                                   const char  *tag,
                                   const char  *comment )
{
  dmDataset  *ds     = NULL;
  afDataset  *ads    = NULL;
  afBlock    *ablock = NULL;
  afFlavour   flav   = AK_FLAV_UNK;
  dmErrCode   status = dmFAILURE;


  ds     = dmi_block_get_ds( block );
  ads    = (afDataset*)dmi_dataset_get_kds( ds );
  flav   = afk_dataset_get_flavour( ads );
  ablock = (afBlock*)dmi_block_get_kb( block );

  switch (flav)
  {
    case AK_FLAV_DTF:
      status = afk_dtf_write_comment( ablock, pos, tag, comment );
      break;
    case AK_FLAV_SIMPLE:
      status = dmSUCCESS;
#if 0
      status = afk_simple_write_comment( ablock, pos, tag, comment );
#endif
      break;

    case AK_FLAV_RAW:
      status = afk_raw_write_comment( ablock, pos, tag, comment );
      break;

    // Code Change  Bug 11905  August 21, 2009 
    case AK_FLAV_CSCRDB:
      status = afk_cscrdb_write_comment( ablock, pos, tag, comment );
      break;

    // Code Change Bug 11940 March 9, 2010
    //  add case for TSV 
    case AK_FLAV_TSV:
      status = afk_tsv_write_comment( ablock, pos, tag, comment );
      break;

    default:
      break;
  }

 return status;
}



/****************************************************************************/
/*                                                                          */
/* Function:   afp_block_insert                                             */
/*                                                                          */
/* Description:                                                             */
/*   Insert a block in the dataset.                                         */
/*                                                                          */
/*  Input parameters:                                                       */
/*                dmDataset *ds             - pointer to dataset            */
/*                const char *name          - block name                    */
/*                long       blocknum       - block number                  */
/*        const dmBlockType  type           - block type                    */
/*                void       *kb            - kernel block ptr.             */
/*                const char *option        - ds option string              */
/*                                                                          */
/*  Output parameters:                                                      */
/*                                                                          */
/*                                                                          */
/*  Returns:                                                                */
/*     dmBlock *pBlockPtr  - pointer to block                               */
/*                                                                          */
/****************************************************************************/
dmBlock* afp_block_insert( dmDataset   *ds,
                           const char  *name,
                           long         blocknum,
                           const dmBlockType type,
                           void        *kb,
                           char        *option )
{
  dmBlock    *block  = NULL;
  afBlock    *ablock = NULL;
  afDataset  *ads    = NULL;
  afFile     *afptr  = NULL;
  afStyle    *astyle = NULL;
  afBuffer   *abuf   = NULL;
  const char *pFlavorString = NULL;

  long nblocks = 0;
  long nrows   = 0;
  long pos     = 0;

  afFlavour flav;
  dmErrCode status = dmSUCCESS;

  char buf[CARD_SIZE];
  (void) kb;


  /* Get style info from dataset */
  /* NOTE: this should already be correctly setup from the dataset_create stage */
  ads     = (afDataset*)dmi_dataset_get_kds( ds );
  astyle  = afk_dataset_get_style( ads );
  nblocks = dmi_dataset_get_n( ds );
  flav    = astyle->flavor;


  /* Check if it is OK to open this block */
  if (( blocknum > 1 || nblocks > 0 ) && !afk_style_get_multiple( astyle ))
  {
    pFlavorString = afk_style_get_flavorString ( astyle);
    if ( pFlavorString )
      strcpy(buf, pFlavorString);

    sprintf( dm_errbuf, 
	    "TEXT kernel specification %s does not support multiple blocks.",
	    buf);
    dmu_err_set( DM_ERR_OPEN_FAIL, dm_errbuf );
    return NULL;
  }
 
  /* Go to correct position in file */
  afptr = afk_dataset_fptr( ads );

  if ( blocknum <= nblocks )
  {
    status = afk_dataset_moveto_block( ads, blocknum );
    if ( status != dmSUCCESS )
    {
      sprintf( dm_errbuf,
	       "Couldn't move to block %ld in dataset %s.\n",
	       blocknum, ads->name );
      dmu_err_set( DM_ERR_OPEN_FAIL, dm_errbuf );
      return NULL;
    }
  }
  else
  {
    afk_file_seek( afptr, 0, SEEK_END );
  }

  pos = afk_file_tell( afptr );
  afp_dataset_add_block( ds, blocknum, type, name, pos );

  block = afp_block_open( ds, blocknum, UT_TRUE, UT_TRUE, option );
  if ( !block )
    return NULL; /* Error already set */

  ablock = (afBlock*)dmi_block_get_kb( block );
  abuf = afk_block_get_buffer( ablock );
  afk_buffer_set_datastart( abuf, pos );

  dmi_block_set_type( block, type );
  dmi_block_set_name( block, name );
  dmi_dataset_set_current( ds, block );
  dmi_block_set_nrows( block, nrows );

  /* Set HINT struct in the block from input options..     */
  /* is this needed?  These fields are all covered by      */
  /* the dataset style and shouldn't change for each block */
  /* Also note that this would already have been done in   */
  /* the afp_block_open routine above.                     */

  if ( option )
    status = afp_set_hint_list( block, option );


  switch (flav)
  {
    case AK_FLAV_DTF:
      afp_dtf_create_block( block );
      break;

    case AK_FLAV_SIMPLE:
      afp_simple_create_block( block );
      break;

    case AK_FLAV_RAW:
      afp_raw_create_block( block );
      break;

    // Code Change  Bug 11905  August 21, 2009
    case AK_FLAV_CSCRDB:
      afp_cscrdb_create_block( block);
      break;

    // Code Change  Bug 11940  march 9, 2010
    case AK_FLAV_TSV:
      afp_tsv_create_block( block);
      break;

    default:
      sprintf( dm_errbuf, "Error: Unknown kernel enumerant: %d", flav );
      dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
      break;
  }
  dmi_block_init( block );

  return block;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_block_notify_moved                          */
/*                                                             */
/* Description:                                                */
/*             Alert the block that its file data has moved.   */
/*             Move all file position holders by moved.        */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock      *block  - ptr to block      */
/*                    dmBlock      *moved_block                */
/*                    long          moved                      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_block_notify_moved( dmBlock* block, dmBlock* moved_block, long moved )
{
  afBlock  *ablock   = NULL;
  afBuffer *abuf     = NULL;
  long      moved_no = 0;
  long      no       = 0;

  moved_no = dmi_block_get_no( moved_block );
  no       = dmi_block_get_no( block );

  if ( no > moved_no )
  {
    ablock = (afBlock*)dmi_block_get_kb( block );
    abuf   = afk_block_get_buffer( ablock );

    afk_buffer_notify_moved( abuf, moved );
  }
}



/***************************************************************/
/*                                                             */
/* Function:   afp_block_close                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock      *block  - ptr to block      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  ErrCode                               */
/*                                                             */
/***************************************************************/
dmErrCode afp_block_close( dmBlock* block )
{
  dmDataset *ds     = NULL;
  afBlock   *ablock = NULL;
  afBuffer  *abuf   = NULL;
  afDataset *ads    = NULL;
  afStyle   *astyle = NULL;
  afFlavour  flav;

  long lmoved = 0;
  long moved  = 0;
  long nrows  = 0;
  long ii;

  ablock = (afBlock*)dmi_block_get_kb( block );

  if ( dmi_block_writeable( block ))
  {
    ads = ablock->ads;
    astyle = afk_dataset_get_style( ads );
    abuf = afk_block_get_buffer( ablock );

    /* align the block's data in case it is fixed width format */

    if ( astyle->fixed )
    {
      nrows = dmi_block_get_nrows( block );
      lmoved = 0;

      flav = astyle->flavor;
      if ( (flav == AK_FLAV_DTF ) && (abuf->ncols > 0 ) && (ablock->block_type == dmTABLE) )
      {

	/* This writes any changes to column defs to the file.               */
	afp_dtf_flush_col_keys( block );

	/* Code Change  Bug 11055  Aug 15, 2008                              */
	/*  Update value for NAXIS1 if style is fixed and type is table.     */
	/*  If copying from non-fixed type to fixed-dtf, NAXIS1 value needs  */
	/*  to be updated before flushing head buffer                        */

	/* Sum the maxwidth of all columns..                                 */
	/* though we should be able to use the prowwidths attribute.         */
	long naxis1 = 0;
	for ( ii = 0; ii < abuf->ncols; ii++)
	  naxis1 += abuf->pmaxwidths[ii];

	afk_dtf_key_write(ablock,0,(char*)"NAXIS1",DT_INT4,&naxis1,NULL,NULL);
      }
    }

    lmoved = 0;
    afk_buffer_flush( abuf, &lmoved );
    moved += lmoved;

    if ( moved )
    {
      ds = dmi_block_get_ds( block );
      afp_dataset_notify_moved( ds, block, moved );
    }
  }

  afk_block_free( ablock );
  dmi_block_set_kb( block, NULL );
  return dmSUCCESS;
}

/***************************************************************/
/*                                                             */
/* Function:   afp_get_RowBufOverRun                           */
/*                                                             */
/* Description:                                                */
/*              read value in bRowBufOverRun                   */
/*              0  means no buffer overrun encountered         */
/*              1  means long row read, truncated, and error   */
/*                 flag set                                    */
/*              -1 means error message pushed onto error stack */
/*                 do not record any more overrun errors       */
/*                                                             */
/*  Input parameters:                                          */
/*                    none                                     */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                    none                                     */
/*  Returns:                                                   */
/*             int  bRowBufOverRun                             */
/*                                                             */
/***************************************************************/
int afp_get_RowBufOverRun()
{
  return bRowBufOverRun;
}

/***************************************************************/
/*                                                             */
/* Function:   afp_set_RowBufOverRun                           */
/*                                                             */
/* Description:                                                */
/*              set value in bRowBufOverRun                    */
/*                                                             */
/*  Input parameters:                                          */
/*                    int  iValue - new value                  */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                    none                                     */
/*  Returns:                                                   */
/*             nothing                                         */
/*                                                             */
/***************************************************************/
void afp_set_RowBufOverRun(int iValue)
{
  bRowBufOverRun = iValue;
}




/***************************************************************/
/*                                                             */
/* Function:   afp_block_write_fixed_keys                      */
/*                                                             */
/* Description:                                                */
/*               THIS FUNCTION IS NOT CURRENTLY USED           */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock      *block  - ptr to block      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*             nothing                                         */
/*                                                             */
/***************************************************************/
#if 0
void afp_block_write_fixed_keys( dmBlock* block )
{
  char       keyname[CARD_SIZE];
  dmuKeyList keylist;
  afBlock   *ablock = NULL;
  afBuffer  *abuf   = NULL;
  afCol     *acol   = NULL;
  long       colndx  = 0; /* column index */
  long       kcolndx = 0;
  long       arndx   = 0; /* array element index */
  long       ncols   = 0;
  long       length  = 0; /* array length */
  long       byte_offset = 0; /* byte offset of each column */
  long       colno   = 0;
  long       keyno   = 0;

  ablock = (afBlock*)dmi_block_get_kb( block );
  abuf = ablock->abuf;

  if ( !abuf->pmaxwidths )
    return;

  keylist = ablock->keylist;
  keyno = keylist->n + 1;
  ncols = ablock->ncols;
  byte_offset = 1; /* TBCOL byte offset starts at 1 */
  kcolndx = 0;

  for ( colndx = 0, colno = 1; colndx < ncols; colndx++, colno++ )
  {
    acol = ablock->cols[colndx];
    length = acol->length;

    if ( length == 0 ) length = 1;
      sprintf( keyname, "TBCOL%ld", colno );
    afk_dtf_key_write( ablock, keyno++, keyname, DT_INT4
                       &byte_offset, NULL,
                       "byte position in row to start of column." );

    for ( arndx=0; arndx < length; arndx++, kcolndx++ )
      byte_offset += abuf->pmaxwidths[kcolndx];
  }
}
#endif
