/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afcol.c c                                                   */
/*                                                               */
/*   Functions that work with columns                            */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"
#include "ryi.h"


/***************************************************************/
/*                                                             */
/* Function:   afp_col_find_coord                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor *col_dd                     */
/*                    long          axisno                     */
/*                    char          sub                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor *coord_dd                           */
/*                                                             */
/***************************************************************/
dmDescriptor* afp_col_find_coord( dmDescriptor* col_dd, long axisno, char sub )
{
  long i     = 0;
  long n     = 0;
  long k     = 0;
  long nc    = 0;
  long colno = 0;
  long dim   = 0;

  dmDescriptor *coord_dd = NULL;
  dmDescriptor *dd       = NULL;
  ftCoord      *ftcoord  = NULL;

 
  if ( !col_dd )
    return NULL;

  colno = dmi_descriptor_get_no( col_dd );

  if ( axisno == 0 )
  {
  
    /* For each coord, check for matching number n and letter subscript c 
      e.g. CTYPEnc  */
  
    n = dmi_descriptor_get_no_coords( col_dd );
    for ( i = 1; i <= n; i++ )
    {
      dd = dmi_descriptor_get_coord_no( col_dd, i );
      ftcoord = (ftCoord*)dmi_descriptor_get_kdd( dd );
      if ( ftcoord && ftcoord->no == colno && ftcoord->sub == sub && ftcoord->axis == axisno )
        return dd; 
    }
  }
  else
  {
    /* For each axis group, check for matching axis number, column number and subscript */
    n = dmi_descriptor_get_no_groups( col_dd );
    for ( i = 1; i <= n; i++ )
    {
      dd = dmi_descriptor_get_group_no( col_dd, i );
      if ( dd )
      {
        dim = dmi_descriptor_get_element_dim( dd );
        ftcoord = (ftCoord*)dmi_descriptor_get_kdd( dd );
    
        if ( ftcoord && ftcoord->no == colno && ftcoord->sub == sub && ftcoord->axis == axisno )
          return dd;
    
        /* Also look on each world coord belonging to the axis group */
        nc = dmi_descriptor_get_no_coords( dd );
        for ( k = 1; k <= nc; k++ )
        {
          coord_dd = dmi_descriptor_get_coord_no( dd, k );
          ftcoord = dmi_descriptor_get_kdd( coord_dd );
     
          if ( ftcoord && ftcoord->no == colno && ftcoord->sub == sub && ftcoord->axis == axisno )
            return coord_dd;
        } 
      }
    }
  }
  return NULL;  
}



/***************************************************************/
/*                                                             */
/* Function:   afp_col_attach_wcs                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor *col_dd                     */
/*                    dmDescriptor *coord_dd                   */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_col_attach_wcs( dmDescriptor* col_dd, dmDescriptor* coord_dd )
{
  dmBool        is_phys = dmFALSE;
  dmDescriptor *phys_dd = NULL;
  ftCoord      *ftcoord = NULL;
 
  long colno  = 0;
  long axisno = 0;

  ftcoord = (ftCoord*)dmi_descriptor_get_kdd( coord_dd );
  colno = ftcoord->no;

  if ( colno == 0 )
    return;

  /* Is this an axis group? */ 
  axisno = ftcoord->axis;
  is_phys = ( ftcoord->sub == 'P' );

 /* Three cases:
  *   Coord system on the descriptor  (axisno=0)
  *   Physical axis group on the descriptor 
  *   World transform on one of the descriptor's physical axis groups 
  */
  if ( axisno == 0 )
    dmi_descriptor_add_coord( col_dd, coord_dd );
  else
    if ( is_phys )
      dmi_descriptor_set_axis( col_dd, axisno, coord_dd );
    else
    {
      phys_dd = dmi_descriptor_get_group_no( col_dd, axisno );

      if ( !phys_dd )
      {
        /* Create phys if needed. 
         * Beware!  This is recursive, calling afp_col_attach_wcs */
        phys_dd = afp_col_create_default_wcs( col_dd, axisno, 'P' );
        phys_dd->flag |= DM_FLAG_TEMP;

        /* Special cases */
        if ( cs_eq( dmi_descriptor_get_name( coord_dd ), "RA" ))
          dmi_descriptor_set_name( phys_dd, "X" );
        else
          if ( cs_eq( dmi_descriptor_get_name( coord_dd ), "DEC" ))
            dmi_descriptor_set_name( phys_dd, "Y" );   
          /* End special cases */
      }
    dmi_descriptor_add_coord( phys_dd, coord_dd );
  }
}



/***************************************************************/
/*                                                             */
/* Function:   afp_col_create_default_wcs                      */
/*                                                             */
/* Description:                                                */
/*             Make a trivial WCS transform for given column,  */
/*             axis,and subscript                              */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor *col_dd                     */
/*                    long          axisno                     */
/*                    char          sub                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor *dd                                 */
/*                                                             */
/***************************************************************/
dmDescriptor* afp_col_create_default_wcs( dmDescriptor* col_dd,
                                          long axisno,
                                          char sub )
{
  char *default_axis_names[3] = { "X", "Y", "Z" };

  char          name[80];
  dmDescriptor *dd = NULL;
  dmBlockType   type;
  long          colno   = 0;

      /* 
         No CTYPE or TCTYP key present, so we invent one 
         and add a new axis to the list
       */

  colno = dmi_descriptor_get_no( col_dd );
  type  = dmi_block_get_raw_type( col_dd->block );

  if ( type == dmTABLE )
  {
    /* Handle case of spuriously copied keywords eg TCROT5 where TFIELDS=3 */
    if ( axisno > 0 )
    {
      sprintf( name, "#COL%ld_AXIS%ld%c", colno, axisno, sub );
    }
    else
    {
      sprintf( name, "COORD%ld%c", colno, sub );
    }
  }
  else
  {
    // Code Change - Bug 9023 - September 14, 2006
    // ascii image file

    if (axisno <= 3 && sub == 'P' )
      strcpy(name, default_axis_names[axisno-1]);
    else
      sprintf( name, "AXIS%ld", axisno);
  }

  dd = afp_col_create_wcs( col_dd, axisno, sub, name, "LINEAR", dmFALSE, name );


  return dd;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_col_create_wcs                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor *col_dd                     */
/*                    long          axisno                     */
/*                    char          sub                        */
/*                    const char   *name                       */
/*                    const char   *transform                  */
/*                    dmBool        composite                  */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor *dd                                 */
/*                                                             */
/***************************************************************/
dmDescriptor* afp_col_create_wcs( dmDescriptor *col_dd, long axisno,
                                  char sub, const char *name,
                                  const char *transform,
                                  dmBool composite, const char *ctype_name )
{
  dmDescriptor *dd    = NULL;
  dmBlock      *block = NULL;
  dmBool        new   = dmFALSE;

  ftCoord *ftcoord = NULL;

  long   colno    = 0;
  long   dim      = 1;
  double val      = 0.0;
  double pix      = 0.0;
  double delt     = 1.0;


  colno     = dmi_descriptor_get_no( col_dd );

  /*
   *  For physical, Check if it already exists by looking through
   *  the list of columns and axes 
   */
  if ( sub == 'P' )
    dd = afp_col_find_coord( col_dd, axisno, sub );

  if ( dd )
  {
    new = dmFALSE;
    ftcoord = (ftCoord*)dmi_descriptor_get_kdd( dd );
    free( ftcoord->ctype_name );
    dmi_descriptor_set_name( dd, (char*)name );
  }
  else
  {
    new = dmTRUE;

    ftcoord  = ftp_coord_alloc();
    block    = dmi_descriptor_get_block( col_dd );
 
    dd = ryi_coord_alloc( block, col_dd, colno, name, transform, &val, &pix,
	 		  &delt, dim, ftcoord );
  }
  ftcoord->axis= axisno;
  ftcoord->no  = colno;
  ftcoord->sub = sub;
  ftcoord->composite = composite;
  ftcoord->pc_row = NULL;
  ftcoord->is_cd = 0;
  ftcoord->is_ltm = 1;
  ftcoord->partner_name = NULL;

  // Code Change  Bug 11488 August 3, 2010
  // Use value of CTYPE keyword for name
  ftcoord->ctype_name = cs_dup( (char*)ctype_name );

  if ( new )
    afp_col_attach_wcs( col_dd, dd );

  return dd;
}


/***************************************************************/
/*                                                             */
/* Function:   afp_col_update_physical                         */
/*                                                             */
/* Description:                                                */
/*      FITS uses logical-to-world where we use                */
/*      physical-to-world.  For all transforms, update their   */
/*      physical-to-world values from their logical-to-world   */
/*      values.  Find the matching physical transform (same    */
/*      axis no, but sub = 'P' ) and use it for the update.    */
/*      Physical tfms themselves do not get updated.           */
/*                                                             */
/*                                                             */
/*      Fill the physical-to-world transform given the         */
/*      logical-to-world and the associated logical-to-physical*/
/*        W = wv + wd * ( P - wp )                             */
/*        P = pv + pd * ( L - pp )                             */
/*        W = lv + ld * ( L - lp )                             */
/*      Then                                                   */
/*        L = pp + ( P - pv ) / pd                             */
/*      so   W = lv + ld * ( pp + (P-pv)/pd - lp )             */
/*            = lv + ( ld / pd ) * ( P - [ pv +  (lp-pp)*pd ] )*/
/*      Hence                                                  */
/*        wv = lv                                              */
/*        wd = ld / pd                                         */
/*        wp = pv + ( lp - pp ) * pd                           */
/*      EG suppose phys-to-world is                            */
/*               W =  1000 + 2.0 * ( P - 4.0 )                 */
/*      and we bin and scale to generate a logical to phys of  */
/*               P =  20 + 4.0 * ( L - 10 )                    */
/*      Then                                                   */
/*        W = 1000 + 2.0 * ( (20-4) - 4 * 10 ) + 2 * 4 $       */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor *col_dd                     */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_col_update_physical( dmDescriptor* col_dd )
{
 dmDescriptor *phys_dd = NULL;
 dmDescriptor *dd      = NULL;

  long groupno = 0;
  long coordno = 0;
  long axisno  = 0;
  long nw      = 0;
  long ng      = 0;
  long naxes   = 0;
  long colno   = 0;

  colno = dmi_descriptor_get_no( col_dd );
  ng = dmi_descriptor_get_no_groups( col_dd );

  for( groupno = 1; groupno <= ng; groupno++ )
  {
    phys_dd = dmi_descriptor_get_group_no( col_dd, groupno );
    nw = dmi_descriptor_get_no_coords( phys_dd );
    for ( coordno = 1; coordno <= nw; coordno++ )
    {
      dd = dmi_descriptor_get_coord_no( phys_dd, coordno );
      dmi_coord_convert_lw_to_pw( dd, phys_dd );
    }
  }

  if ( !dmi_descriptor_check_groups( col_dd ))
  {
    /*
     * We need to add some missing axis groups. Right now we'll assume
     * this is because all of them are missing - might get bitten if
     * some CTYPE are present but not all. 
     */
    naxes = dmi_descriptor_get_array_dim( col_dd );
    for ( axisno = 1; axisno <= naxes; axisno++ )
    {
      phys_dd = afp_col_create_default_wcs( col_dd, axisno, 'P' );
      phys_dd->flag |= DM_FLAG_TEMP;  // added for Bug 10846
      dmi_descriptor_set_axis( col_dd, axisno, phys_dd );
    }  
  }
}



/***************************************************************/
/*                                                             */
/* Function:   afp_create_col                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock      *block                      */
/*                    dmDescriptor *dd                         */
/*                    long          no                         */
/*                    void         *kdd                        */
/*                    const char   *hint                       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_create_col( dmBlock* block, dmDescriptor* dd, long no, void* kdd, const char* hint )
{
 dmDataset *ds     = NULL;
 afDataset *ads    = NULL;
 afBlock   *ablock = NULL;
 afFlavour  flav   = AK_FLAV_UNK;


 (void) hint;
 
 if ( !block || !dd )
   return;
 if ( !dmi_descriptor_check( dd ))
   return;

  ds = dmi_block_get_ds( block );
  ads = (afDataset*)dmi_dataset_get_kds( ds );
  flav = afk_dataset_get_flavour( ads );

  /* 
   * We use the seed kernel descriptor to ensure that vararray columns
   * are copied to other vararray columns. We rely on the calling layer
   * to only pass FTK seeds to the FTK kernel.
   */ 
  ablock = (afBlock*)dmi_block_get_kb( block );

  if ( (block->type == dmTABLE )||(block->type == dmUNKBLOCK ))
  {
    if ( no == 0 )
      no = block->ncols + 1; 
 
    switch ( flav )
    {
      case AK_FLAV_DTF:
        afp_dtf_insert_col( block, dd, no, kdd );
        break;

      case AK_FLAV_SIMPLE:
        afp_simple_insert_col( block, dd, no, kdd );
        break;

      case AK_FLAV_RAW:
        afp_raw_insert_col( block, dd, no, kdd );
        break;

      case AK_FLAV_CSCRDB:
        // Code Change  Bug 11905  August 21, 2009
        afp_cscrdb_insert_col( block, dd, no, kdd );
        break;

      case AK_FLAV_TSV:
        // Code Change  Bug 11940  March 9, 2010
        afp_tsv_insert_col( block, dd, no, kdd );
        break;

      default:
        break;
    }
  }
  else 
  {
    no = 1; 

    /* block type == dmIMAGE */
    switch ( flav )
    {
      case AK_FLAV_DTF:
        afp_dtf_insert_image_col( block, dd, no, kdd );
        break;
      default:
        break;
    }  
  } 

}
