/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afcscrdb.c                                                  */
/*                                                               */
/*   Functions that create and work with ASCII file type         */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"



/***************************************************************/
/*                                                             */
/* Function:   afp_cscrdb_create_block                         */
/*                                                             */
/* Description:                                                */
/*              Create a new block for cscrdb file             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock   *block                         */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/*  Called from:                                               */
/*        afp_block_insert()                                   */
/*                                                             */
/***************************************************************/
 void afp_cscrdb_create_block( dmBlock *block )
 {
   afBlock   *pakBlock  = NULL;
   afDataset *pakDataset = NULL;
   afFile    *pakFile    = NULL;
   afBuffer  *pakBuffer  = NULL;
   afStyle   *pakStyle   = NULL;

   long       pos     = 0;
   long       moved   = 0;
   int        flavorStringLength = 0;

   const char *pFlavorString  = NULL;
   char        flavorString[CARD_SIZE];

   pakBlock   = (afBlock*)dmi_block_get_kb( block );
   pakDataset = pakBlock->ads;
   pakStyle   = pakDataset->astyle;
   pakBuffer  = afk_block_get_buffer( pakBlock );
   pakFile    = afk_dataset_fptr( pakDataset );


   pFlavorString = afk_style_get_flavorString(pakStyle);

   if ( pFlavorString)
     strcpy( flavorString, pFlavorString );
   strcat (flavorString, "\n" );

   flavorStringLength = strlen(flavorString);

   // insert flavor sting into the file and flush buffer
   afk_file_insert_data( pakFile, flavorString );

   afk_buffer_flush( pakBuffer, &moved );

   // reset start of file
   afk_file_seek( pakFile, pakDataset->pos+flavorStringLength, SEEK_SET );
 
   pos = afk_file_tell( pakFile );
   pakDataset->blocks_pos[0] = pos;
   afk_buffer_set_datastart( pakBuffer, pos );

   return;
 }


/***************************************************************/
/*                                                             */
/* Function:   afp_cscrdb_dataset_scan                         */
/*                                                             */
/* Description:                                                */
/*   Read header of input dataset file and determine if it     */ 
/*   has a cxcrdb format style. If so, return flavor type      */ 
/*   AK_FLAV_CXCRDB, otherwise return AK_FLAV_UNK              */
/*                                                             */
/*   Code Change  Bug 11905  August 21, 2009                   */
/*   Added this function                                       */
/*                                                             */
/*  Input parameters:                                          */
/*                       afDataset  *pakAsciiDataSet           */
/*                       dmDataset  *ds                        */
/*                       afStyle    *pakStyle                  */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*              akFlavor  aFlavor                              */
/*                                                             */
/*  Called from:                                               */
/*     afp_dataset_scan()                                      */
/*                                                             */
/***************************************************************/
afFlavour afp_cscrdb_dataset_scan( afDataset *pakAsciiDataSet,
                              dmDataset* ds,
                              afStyle  *pakStyle)
{
  afFile      *pakFilePointer   = NULL;
  const char  *pcLine           = NULL;
  char        *pcLineLookPtr    = NULL;
  afFlavour    akFlavor         = AK_FLAV_UNK;

  dmBool  bHeader     = UT_FALSE;
  dmBool  bDashLine   = UT_FALSE;

  char  *sDefaultSep        = "\t\r";
  char  *sSeparator         = NULL;
  char  *sDefaultHeaderChar = "#";
  char   sHeaderChar[2];

  long   lLineCounter     = 0;
  int    ii               = 0;          // loop counter


  // Check for user defined header char and separator
  // otherwise use defaults

  sHeaderChar[0] = pakStyle->header_char;
  sHeaderChar[1] = '\0';

  if ( sHeaderChar[0] == '\0' )
    strcpy( sHeaderChar, sDefaultHeaderChar );

  if ( pakStyle->sep != NULL )
    sSeparator = cs_dup( pakStyle->sep );
  else
    sSeparator = cs_dup( sDefaultSep );



  // Get file pointer and  file type from the Ascii dataset
  pakFilePointer  = afk_dataset_fptr( pakAsciiDataSet );

  // Get pointer to first line
  pcLine = afk_file_getline_ptr( pakFilePointer );

  if ( !pcLine )
  { 
    // Problem reading line from file.. problem with the file
    //    or maybe the line is too long.
    return dmFAILURE;
  }

  // Header comment lines start with # - skip those lines
  pcLineLookPtr = (char*)pcLine;
    
  // Is the character a comment character?
  //  if so, treat line as comment and skip over it

  do
  {
    if ( pcLineLookPtr )
    {
      ++lLineCounter;

      if (pcLineLookPtr[0] == sHeaderChar[0])
      {
        // Set the header flag
        bHeader  = UT_TRUE;

        // skip - read next line
        pcLine = afk_file_getline_ptr(pakFilePointer); 
        pcLineLookPtr = (char*)pcLine;
      }
      else
      {
        // First non-comment line - break loop and parse line
        bHeader = UT_FALSE;
        break;
      }
    }
  } while (pcLine && bHeader);

  // If no non-comment lines found, return
  if ( bHeader )
    return (akFlavor);


  for (ii = 0; ii < 3 && pcLineLookPtr; ii++)
  {
    // Scan for dash line
    bDashLine = afk_cscrdb_is_dashline( pcLineLookPtr, sSeparator );

    if ( bDashLine )
    {
      akFlavor = AK_FLAV_CSCRDB;
      break;
    }

    // failed to find the dash line
    // read next line in file and try again
    pcLine = afk_file_getline_ptr(pakFilePointer);
    pcLineLookPtr = (char*)pcLine;
  }

  return (akFlavor);

}


/*****************************************************************************/
/*                                                                           */
/*  Function:  afp_cscrdb_init_cols()                                        */
/*                                                                           */
/*  Description:                                                             */
/*      Based on afp_simple_init_cols()                                      */
/*                                                                           */
/*      Calls afp_raw_init_cols() and afp_simple_make_vectors()              */
/*                                                                           */
/*  Input parameters:                                                        */
/*                    dmBlock        *pBlock                                 */
/*                                                                           */
/*  Output parameters:  dmuKey* key                                          */
/*                                                                           */
/*  Returns:                                                                 */
/*     dmErrCode  status                                                     */
/*                                                                           */
/*  Called from:                                                             */
/*        afp_block_init_cols()                                              */
/*                                                                           */
/*****************************************************************************/
dmErrCode afp_cscrdb_init_cols( dmBlock *pBlock)
{
  dmErrCode  status   = dmSUCCESS;
  dmDataset *pDataset = NULL;
  dmDescriptor  **rawcols = NULL;

  afDataset *pakDataset = NULL;
  afStyle   *pakStyle   = NULL;
  afBlock   *pakBlock   = NULL;
  afCol     *pakCol     = NULL;
  afCol   ***ppakCmpnts  = NULL;

  dmBool     bAppend   = dmFALSE;  /* Flag to append vector component */
  dmBool     bUseUnits = dmFALSE;

  char  colName[CARD_SIZE];
  char *pEnd    = NULL;

  long *maxcpts = NULL;   /* number of cpts in each cpts element */
  long *dims    = 0;      /* pointer to number of components for each vector */
  long  numCols = 0;
  long  colincr = 10;     /* amount to increment maxcols when it overflows */
  long  cptincr = 5;      /* amount to increment maxcpts when it overflows */
  long  maxcols = 10;     /* number of cols in cpts and ncpts */
  long  newVal  = 0;
  long  colIndex= 0;
  long  cptNumber = 0;
  long  numKCols = 0;
  long  ii      = 0;
  long  jj      = 0;


  // Get dataset pointer from block pointer
  pDataset = dmi_block_get_ds( pBlock );

  // Get Ascii dataset pointer from dataset pointer
  pakDataset = (afDataset*)dmi_dataset_get_kds( pDataset );

  // Get style structure
  pakStyle = afk_dataset_get_style( pakDataset );

  // if no column names, treat columns same as raw file
  if ( pakStyle && pakStyle->colnames == AK_STYLE_COLNAMES_NONE )
  {  
    status = afp_raw_init_cols( pBlock );
    return (status);
  }

  //  If header has unit names, set flag bUseUnits to true
  if (pakStyle && pakStyle->colnames == AK_STYLE_COLNAMES_LM2)
    bUseUnits = dmTRUE;

  pakBlock = (afBlock*)dmi_block_get_kb( pBlock );

  status = afk_cscrdb_init_cols( pakBlock, pakStyle );

  numKCols   = afk_block_ncols( pakBlock );
  dims     = (long*)calloc( maxcols, sizeof(long) );
  maxcpts  = (long*)calloc( maxcols, sizeof(long) );
  ppakCmpnts = (afCol***)calloc( numKCols, sizeof(afCol**) );

  /* first, loop through and put the afCols into a list of vector component lists */
  for( ii = 0; ii < numKCols; ii++ )
  {
    pakCol = afk_block_get_col( pakBlock, ii+1 );

    bAppend = dmFALSE;
    colIndex = 0;

    /* if this is a vector component, it will be named 'vec(cpt)' */
    pEnd = strchr( pakCol->name, '(' );

    if ( pEnd )
    {
      /* copy vector name + '(' char */
      cs_copy_siz( pakCol->name, colName, pEnd - pakCol->name+1 );

      for ( jj = 0; jj < numCols; jj++ )
      {
        if (  ppakCmpnts[jj] &&  ppakCmpnts[jj][0] && 
	   ( cs_eq_siz( colName,  ppakCmpnts[jj][0]->name, strlen(colName) )))
        {
          /* this vector is already in the list, mark for append */
	  bAppend = dmTRUE;
          colIndex = jj;
          break;
        }
      }
    }

    if ( bAppend )
    {
      /* append this vector component to existing vector */
      dims[colIndex]++;

      if ( dims[colIndex] > maxcpts[colIndex] )
      {
        maxcpts[colIndex] += cptincr;
        ppakCmpnts[colIndex] = (afCol**)realloc( ppakCmpnts[colIndex], sizeof(afCol*) * maxcpts[colIndex] );
      }

      cptNumber = dims[colIndex]-1;
      ppakCmpnts[colIndex][cptNumber] = pakCol;   

    }
    else
    {
       /* create a new column */
       colIndex = numCols;
       ++numCols;

       if ( numCols > maxcols )
       {
         newVal = maxcols + colincr;
         ppakCmpnts = (afCol***)realloc( ppakCmpnts, sizeof(afCol**) * newVal );
         maxcpts = (long*)realloc( maxcpts, sizeof(long) * newVal );
         dims = (long*)realloc( dims, sizeof(long) * newVal );
         memset( &ppakCmpnts[maxcols], 0, colincr*sizeof(afCol**) );
         memset( &maxcpts[maxcols], 0, colincr*sizeof(long) );
         memset( &dims[maxcols], 0, colincr*sizeof(long) );
         maxcols = newVal;
       }

       dims[colIndex]++;

       if ( dims[colIndex] > maxcpts[colIndex] )
       {
         newVal = maxcpts[colIndex] + cptincr;
         ppakCmpnts[colIndex] = (afCol**)realloc( ppakCmpnts[colIndex], 
				      sizeof(afCol*) * newVal );
         memset( &ppakCmpnts[colIndex][maxcpts[colIndex]], 0, cptincr*sizeof(afCol*) );
         maxcpts[colIndex] = newVal;
       }  
       ppakCmpnts[colIndex][0] = pakCol;
    }
  }


  /* use our list of lists to create our final array of dmDescriptors */
  rawcols = afp_simple_make_vectors( pBlock, ppakCmpnts, numCols, dims, bUseUnits );

  for ( ii = 0; ii < numCols; ii++ )
    free( ppakCmpnts[ii] );

  free( ppakCmpnts );
  free( dims );
  free( maxcpts );
  dmi_block_set_raw_cols( pBlock, rawcols, numCols );

  return (status);
}


/*****************************************************************************/
/*                                                                           */
/*  Function:  afp_cscrdb_insert_col()                                       */
/*                                                                           */
/*  Description:                                                             */
/*                                                                           */
/*     Based on afp_simple_insert_col()                                      */
/*                                                                           */
/*  Input parameters:                                                        */
/*                    dmBlock        *pBlock                                 */
/*                    dmDescriptor   *pDescriptor                            */
/*                    long            colNumber                              */
/*                    void           *pKernelDescriptor                      */
/*                                                                           */
/*  Output parameters:                                                       */
/*                                                                           */
/*  Returns:                                                                 */
/*     dmErrCode  status                                                     */
/*                                                                           */
/*  Called from:                                                             */
/*        afp_col_create()                                                   */
/*                                                                           */
/*****************************************************************************/
dmErrCode  afp_cscrdb_insert_col( dmBlock *pBlock, dmDescriptor *pDescriptor,
                                  long colNumber, void *pKernelDescriptor )
{ 
  dmErrCode   status  = dmSUCCESS;
  dmBool      bAppend = dmFALSE;
  dmDataType  type    = DT_UNK;

  dmDescriptor *pColDescriptor      = NULL;
  dmDescriptor *pCmptDescriptor     = NULL;
  dmDescriptor *pUnscaledDescriptor = NULL;

  afBlock  *pakBlock  = NULL;
  afStyle  *pakStyle  = NULL;
  afCol    *pakCol    = NULL;
  afVector *pakVector = NULL;

  long  numColumns = 0;
  long  dim        = 0;
  long  kColNumber = 0;
  long  cptno  = 0;
  long  moved  = 0;


  pakBlock = (afBlock*)dmi_block_get_kb( pBlock );
  pakStyle = pakBlock->ads->astyle;


  numColumns = dmi_block_get_ncols( pBlock );
  bAppend    = ( colNumber == numColumns + 1 );
  dim = dmi_descriptor_get_vector_dim( pDescriptor );

  if ( bAppend )
    kColNumber = pakBlock->ncols+1;
  else
  {
    pColDescriptor  = dmi_block_get_col( pBlock, colNumber );
    pCmptDescriptor = dmi_descriptor_get_cpt( pColDescriptor, 1 );
    pakCol     = (afCol*)dmi_descriptor_get_kdd( pColDescriptor );
    kColNumber = pakCol->no;
  }

  if ( dim > 1 )
    pakVector = afk_vector_alloc( dim );

  for ( cptno = 1; cptno <= dim; cptno++ )
  {
    pCmptDescriptor = dmi_descriptor_get_cpt( pDescriptor, cptno );
    type = dmi_descriptor_get_datatype( pCmptDescriptor );
    pakCol = (afCol*)dmi_descriptor_get_kdd( pCmptDescriptor );

    if ( !pakCol )
    {
      if ( pKernelDescriptor ) /* pass on the kernel descriptor */
      {
        if ( dim == 1 )
          pakCol = afk_col_alloc_copy( (afCol*)pKernelDescriptor );
        else
          if ( dim > 1 )
            pakCol = afk_col_alloc_copy( ((afVector*)pKernelDescriptor)->cpts[cptno-1] );
      }
      else
      {
        pakCol = afk_col_alloc();
        pakCol->name     = cs_dup( dmi_descriptor_get_name( pCmptDescriptor ));
        pakCol->datatype = dmi_descriptor_get_datatype( pCmptDescriptor );
        pakCol->length   = dmi_descriptor_get_array_size( pCmptDescriptor );

        // Code Change  Bug 10784  November 8, 2007
        pakCol->width    = dmi_descriptor_get_length( pCmptDescriptor );

	/* Code change October 28, 2008 */
	/* Bit type stored by DM as array of bytes, but as   */
	/* single bit string in ascii                        */
	if (pakCol->datatype == DT_BIT )
	  pakCol->length = 0;
      }
      dmi_descriptor_set_kdd( pCmptDescriptor, (void*)pakCol );
    }

    pUnscaledDescriptor = dmi_descriptor_get_unscaled( pCmptDescriptor );

    if ( pUnscaledDescriptor )
      dmi_descriptor_set_kdd( pUnscaledDescriptor, (void*)pakCol );
    if ( dim > 1 )
      pakVector->cpts[cptno-1] = pakCol;

    afk_block_insert_col( pakBlock, pakCol, kColNumber );
    afk_buffer_insert_col( pakBlock->abuf, pakCol, kColNumber, &moved );
    ++kColNumber;
  }

  if ( dim > 1 )
  {
    dmi_descriptor_set_kdd( pDescriptor, pakVector );
  }

  // update header keys
  afp_cscrdb_write_col_header( pBlock, pDescriptor, colNumber );

  return (status);
}

/*****************************************************************************/
/*                                                                           */
/*  Function:  afp_cscrdb_key_compose_colnames()                             */
/*                                                                           */
/*  Description:                                                             */
/*       based on afp_simple_write_col_header with changes to accommodate    */
/*       CSCRBD header definitions                                           */
/*                                                                           */
/*  Input parameters:                                                        */
/*                    dmDescriptor   *pDescriptor                            */
/*                    afStyle        *pakStyle,                              */
/*                    char           *currKeyValue                           */
/*                    char           *newEntry                               */
/*                    dmBool          append                                 */
/*                    long            colNumber                              */
/*                                                                           */
/*  Output parameters:                                                       */
/*                    char           *newKeyValue                            */
/*                                                                           */
/*  Returns:                                                                 */
/*     dmErrCode  status                                                     */
/*                                                                           */
/*  Called from:                                                             */
/*        afp_cscrdb_write_col_header()                                      */
/*                                                                           */
/*****************************************************************************/
dmErrCode afp_cscrdb_key_compose_colnames( dmDescriptor *pDescriptor,
                                           afStyle *pakStyle,
                                           char *currKeyValue,
                                           char *newEntry,
                                           char  cSeparator,
                                           dmBool append,
                                           long   colNumber,
					   char *newKeyValue )
{
  char  col_entry[CARD_SIZE];

  dmErrCode status = dmSUCCESS;
  dmDescriptor *cpt_dd  = NULL;
  char     *eptr        = NULL;
  char     *ptr         = NULL;
  char     *name        = NULL;


  long  dim = 0;
   long  len = 0;
  long  entry_len = 0; 
  long  arsize    = 0;
  long  cptno     = 0;
  long  colno     = 0;


  dim = dmi_descriptor_get_vector_dim( pDescriptor );

  eptr = col_entry + CARD_SIZE;
  ptr = col_entry;

  len = strlen( newEntry );

  if ( len + ptr < eptr )
    ptr += sprintf( col_entry, "%s", newEntry );

  arsize = dmi_descriptor_get_array_size( pDescriptor );

  if ( arsize > 1 && ( ptr + (arsize / 10) + 3 < eptr ))
    ptr += sprintf( ptr, "[%ld]", arsize );

  if ( dim > 1 )
  {
    cpt_dd = dmi_descriptor_get_cpt( pDescriptor, 1 );
    name = dmi_descriptor_get_name( cpt_dd );
    len = strlen( name );

    if ( ptr + len + 2 < eptr )
      ptr += sprintf( ptr, "(%s", name);

    for ( cptno = 2; cptno <= dim; cptno++ )
    {
      cpt_dd = dmi_descriptor_get_cpt( pDescriptor, cptno );
      name = dmi_descriptor_get_name( cpt_dd );
      len = strlen( name );

      if ( ptr + len + 1 < eptr )
        ptr += sprintf( ptr, ",%s", name );
    }

    strcat( col_entry, ")" );
    ++ptr;
  }

  entry_len = ptr - col_entry;
  len = strlen( currKeyValue );

  if ( entry_len + len > AK_COMMENTLINE_LEN )
    return (status);

  if ( append )
  {
    if ( *currKeyValue != '\0' )
      sprintf( newKeyValue, "%s%c%s", currKeyValue, cSeparator, col_entry );
    else
      sprintf( newKeyValue, "%s", col_entry );
  }
  else
    if ( colNumber == 1 )
      sprintf( newKeyValue, "%s%c%s", col_entry, cSeparator, currKeyValue );
    else
    {
      dmBool in_paren = UT_FALSE;
      ptr = currKeyValue;
      colno = 1;

      while( *ptr == ' ' || *ptr == '\t' )
        ++ptr;

      while( *ptr != '\0' && colno != colNumber )
      {
        if ( !in_paren && *ptr == '(' )
          in_paren = UT_TRUE;
        else
          if ( in_paren && *ptr == ')' )
            in_paren = UT_FALSE;
          else
            if ( *ptr == ' ' || *ptr == '\t' )
              ++colno;
       ++ptr;
      }

      if ( *ptr != '\0' )
      {
        *ptr = '\0';
        ++ptr;
      }

      sprintf( newKeyValue, "%s%c%s%c%s", currKeyValue, cSeparator, col_entry, cSeparator, ptr );
    }
 
  return (status);
} 


/*****************************************************************************/
/*                                                                           */
/*  Function:  afp_cscrdb_write_col_header()                                 */
/*                                                                           */
/*  Description:                                                             */
/*       based on afp_simple_write_col_header with changes to accommodate    */
/*       CSCRBD header definitions                                           */
/*                                                                           */
/*  Input parameters:                                                        */
/*                    dmBlock        *pBlock                                 */
/*                    dmDescriptor   *pDescriptor                            */
/*                    long            colNumber                              */
/*                                                                           */
/*  Output parameters:                                                       */
/*                                                                           */
/*  Returns:                                                                 */
/*     dmErrCode  status                                                     */
/*                                                                           */
/*  Called from:                                                             */
/*        afp_cscrdb_insert_col()                                            */
/*                                                                           */
/*****************************************************************************/
dmErrCode afp_cscrdb_write_col_header( dmBlock *pBlock,
                                       dmDescriptor *pDescriptor,
                                       long colNumber )
{
  dmErrCode status = dmSUCCESS;
  dmuKeyList keyList;

  dmuKey  *nameKey = NULL;
  dmuKey  *unitKey = NULL;
  dmuKey  *dashKey = NULL;
  dmBool   append  = UT_FALSE;

  afBlock  *pakBlock = NULL;
  afStyle  *pakStyle = NULL;
  afBuffer *pakBuffer = NULL;
 

  char  newLine[AK_COMMENTLINE_LEN];

  char *tag = "dmCOMMENT";
  char *comment = "";
  char *currKeyValue = NULL;
  char *newEntry     = NULL;
  char *sDash        = "-";
  const char *sSeparator = NULL;
  char  cSeparator       = '\0';
  char  defaultSeparator = '\t';
  char  sSep[2] = { 2 * '\0'};
  long  numCols = 0;
  long  mask = UT_TRUE;
  long  len  = 0;
  long  numKeys = 0;

  ftKeyType   fktype  = FTK_COMMENT;

 
  // get pointer, initial values
  pakBlock = (afBlock*)dmi_block_get_kb( pBlock );
  pakStyle = pakBlock->ads->astyle;
  pakBuffer = pakBlock->abuf;

  numCols = dmi_block_get_ncols( pBlock );
  append  = ( colNumber == numCols+1 );
  keyList = pakBlock->keylist;


  //   Get any specified separator characters and use
  //   the first one in the list when forming the
  //   format strings
  //   If NULL, use default (tab)

  sSeparator = afk_style_get_sep((const afStyle *) pakStyle);

  if ( sSeparator )
    cSeparator = sSeparator[0];

  if ( cSeparator == '\0')
     cSeparator = defaultSeparator;

  sSep[0] = cSeparator;
  sSep[1] = '\0';


  // To avoid outputting blank units line, only name and dash lines are
  // inserted into the key list when creating columns.
  // If a unit entry is present, after the original keylist with 2 keys
  // is created, then a third key for units needs to be allocated and
  // initial value put in it.
  // This code section accommodates the condition that some columns have units
  // and some do not.  This code section is invoked if first column has no units
  // but subsequent columns do.
  // Only need to do this check if there is a keylist and the
  // col names style is AK_STYLE_COLNAMES_LM1.

  if ( keyList->data && pakStyle->colnames == AK_STYLE_COLNAMES_LM1 )
  {
    // check if unit info here and change style if necessary
    newEntry = dmi_descriptor_get_unit( pDescriptor );

    if (newEntry )
    {
      // unit name is not null, so create a key for units line
      long ii;
      char *line = NULL;

      // change the col names style
      pakStyle->colnames = AK_STYLE_COLNAMES_LM2;

      numKeys = keyList->n; 
      unitKey  = dmu_key_alloc();
      unitKey->type = FTK_KEY;
      dmu_key_insert( keyList, unitKey, numKeys );

      // create the initial entry allowing for some number of blank columns
      line = (char*)malloc( (numKeys+2) * sizeof(char));

      for (ii = 0; ii < colNumber-2; ii++)
	strcpy(line+ii, sSep );

      dmu_key_set( unitKey, fktype, DT_CHAR, tag, &line, NULL, NULL, mask );

      // put into header buffer
      strcat(line, "\n");
      status = afk_buffer_insert_keyline( pakBuffer, line, unitKey->index );
      free( line );
    }
  }



  // If no keylist, create one with 2 keys for column names and dash
  // Create key for units only if col names style is AK_STYLE_COLNAMES_LM2

  if ( ! keyList->data )
  {
    newLine[0] = '\0';

    if ( pakStyle->colnames == AK_STYLE_COLNAMES_LM1 || 
         pakStyle->colnames == AK_STYLE_COLNAMES_LM2 )
    {
      // Create column names key
      ++numKeys;
      nameKey  = dmu_key_alloc();
      nameKey->type = FTK_KEY;
      dmu_key_insert( keyList, nameKey, numKeys );
      dmu_key_set( nameKey, fktype, DT_CHAR, tag, &comment, NULL, NULL, mask );
      status = afk_buffer_insert_keyline( pakBuffer, newLine, nameKey->index );
    }

    if ( pakStyle->colnames == AK_STYLE_COLNAMES_LM2 )
    {
      // Create units key
      ++numKeys;
      unitKey  = dmu_key_alloc();
      unitKey->type = FTK_KEY;
      dmu_key_insert( keyList, unitKey, numKeys );
      dmu_key_set( unitKey, fktype, DT_CHAR, tag, &comment, NULL, NULL, mask );
      status = afk_buffer_insert_keyline( pakBuffer, newLine, unitKey->index );
    }

    // Create dash line key
    ++numKeys;
    dashKey  = dmu_key_alloc();
    dashKey->type = FTK_KEY;
    dmu_key_insert( keyList, dashKey, numKeys );
    dmu_key_set( dashKey, fktype, DT_CHAR, tag, &comment, NULL, NULL, mask );
    status = afk_buffer_insert_keyline( pakBuffer, newLine, dashKey->index );
  } 
  else
  {
    // keylist exists; add new entries
    // get keys, using known offsets in keylist

    numKeys = keyList->n;

    if ( pakStyle->colnames == AK_STYLE_COLNAMES_LM1 )
      nameKey = keyList->data[numKeys-2];

    if ( pakStyle->colnames == AK_STYLE_COLNAMES_LM2 )
    {
      nameKey = keyList->data[numKeys-3];
      unitKey = keyList->data[numKeys-2];
    }

    dashKey = keyList->data[numKeys-1];
  }


  // Column Names
  if ( nameKey )
  {
    // Insert name of new column into key
    currKeyValue = dmu_key_value_c( nameKey );
    newEntry = dmi_descriptor_get_name( pDescriptor );

    afp_cscrdb_key_compose_colnames( pDescriptor, pakStyle, currKeyValue,
                                     newEntry, cSeparator, append, colNumber, newLine );
    free( currKeyValue );

    // put new key line into column names key 
    *(char**)&nameKey->value = cs_dup( newLine );

    // add newline, insert into buffer
    len = strlen( newLine );
    if ( len <= AK_COMMENTLINE_LEN-1 )
     strcat( newLine, "\n" );

    status = afk_buffer_replace_keyline( pakBuffer, newLine, nameKey->index );
  }

  // Units
  if (unitKey)
  {
    // Insert unit string into key
    currKeyValue = dmu_key_value_c( unitKey);
    newEntry = dmi_descriptor_get_unit( pDescriptor );

    // If not null, add new entry to existing string
    if ( currKeyValue && *currKeyValue != '\0' )
    {
      strcpy( newLine, currKeyValue );

      // don't add separator if last entry is a separator
      // null entries are allowed

      // avoid adding extra separators for null unit names
      if ( currKeyValue[strlen(currKeyValue)-1] != *sSep )
        strcat( newLine, sSep);

      // add new entry
      if ( newEntry )
        strcat(newLine, newEntry );
      else
        strcat(newLine, sSep);
    }
    else
    {
      // current key has null value
      // if new entry is non-null, copy it
      // if new entry is null (allowed), insert separator
      if (newEntry )
        strcpy(newLine, newEntry );
      else
        strcpy(newLine, sSep); 
    }

    free( currKeyValue );
 
    // put new key line into units key 
    *(char**)&unitKey->value = cs_dup( newLine );

    // add newline, insert into buffer
    len = strlen( newLine );
    if ( len <= AK_COMMENTLINE_LEN-1)
      strcat( newLine, "\n" );

    status = afk_buffer_replace_keyline( pakBuffer, newLine, unitKey->index );
  }

  // Dash Line
  // Add a dash for every new column
  currKeyValue = dmu_key_value_c(dashKey);

  if ( currKeyValue && *currKeyValue != '\0' )
  {
    // add separator followed by dash to current value
    strcpy( newLine, currKeyValue );
    strcat(newLine, sSep );
    strcat(newLine, sDash );
  }
  else
  {
    // first entry, just insert dash
    strcpy(newLine, sDash );
  }

  free( currKeyValue );
 
  // put new key line into dash line key 
  *(char**)&dashKey->value = cs_dup( newLine );
 
  // add newline, insert into buffer 
  len = strlen( newLine );
  if ( len <= AK_COMMENTLINE_LEN-1 )
    strcat( newLine, "\n" );

  status = afk_buffer_replace_keyline( pakBuffer, newLine, dashKey->index );


  return (status);
}

