/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afdataset.c                                                 */
/*                                                               */
/*   Functions that create and work with ASCII datasets          */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/
#include "afkernelP.h"
#include <strings.h>
#include <sys/types.h>
#include <dirent.h>




/***************************************************************/
/*                                                             */
/* Function:   afp_dataset_create                              */
/*                                                             */
/* Description:                                                */
/*   Create an ASCII dataset.                                  */
/*   Setup dataset style definitions.                          */
/*                                                             */
/*  Input parameters:                                          */
/*    const char *filename   - dataset filename                */
/*    const char *options    - option list string              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*    dmDataset  *ds         - pointer to created dataset      */
/*                                                             */
/***************************************************************/
dmDataset* afp_dataset_create( const char *filename, const char* options )
{
  dmDataset  *pDataSet         = NULL;
  afDataset  *pakAsciiDataSet  = NULL;
  afFile     *pakFilePointer   = NULL;
  afFlavour   akFlavor;
 
  // open filename
  pakFilePointer = afk_file_open( filename, dmTRUE, dmTRUE );

  if ( pakFilePointer )
  {
    // file open, allocate dataset
    pakAsciiDataSet = afk_dataset_alloc();
  }
  else
  {
   // File open failed, set error
    dmu_err_set( DM_ERR_OPEN_FAIL,  "Error, TEXT Kernel failed to create the file." );
  }

  // did dataset allocation go ok?
  if (! pakAsciiDataSet)
  {
    // Error allocating dataset or opening file
    // Close file and return NULL

    afk_file_close( pakFilePointer);
    return NULL;
  }

  // File open, dataset allocated.. Now set descriptors
  pakAsciiDataSet->name  = cs_dup( filename );
  pakAsciiDataSet->afptr = pakFilePointer;

  pDataSet = dmi_dataset_alloc( filename, (dmkDataset*)pakAsciiDataSet );

  dmi_dataset_set_mode( pDataSet, UT_TRUE );

  /* Set dataset style struct.. */

  /* Load style struct from global hint values. */
  afp_set_style_hints( NULL, pakAsciiDataSet->astyle );

  /* Check if a flavor has been set.. */
  akFlavor = afk_dataset_get_flavour(pakAsciiDataSet);
  if ( akFlavor == AK_FLAV_UNK )
  {
    // No file type set, default it to raw
    akFlavor = AK_FLAV_RAW;
  }

  /* Set ALL style fields to default settings for that flavor */
  afk_style_set( pakAsciiDataSet->astyle, akFlavor );

  /* Reset individual style items from global hints */
  afp_set_style_hints( NULL, pakAsciiDataSet->astyle );

  /* Override style settings from input option list         */
  /*   NOTE: This may reset the entire struct if the kernel */
  /*   is specified here, but that is OK since this is the  */
  /*   ultimate priority method of defining the kernel      */
  /*                                                        */
  /*   The only exception to this is text/cscrdb. If doing  */
  /*   a copy from cscrdb to cscrdb, the necessary colnames */
  /*   value is in the hints. It gets wiped out by this next*/
  /*   call and has to be reset - again!                    */

  afp_dataset_set_hint_list( pDataSet, options );

  akFlavor = pakAsciiDataSet->astyle->flavor;

  if ( akFlavor == AK_FLAV_CSCRDB )
      afp_set_style_hints( NULL, pakAsciiDataSet->astyle );

  /* At this point, the dataset style should be completely defined. */
  /*  even CSCRDB                                                   */

  return pDataSet;
}


/***************************************************************/
/*                                                             */
/* Function:   afp_file_is_binary                              */
/*                                                             */
/* Description:                                                */
/*  Check for binary filetype by searching for non-newline,    */
/*  unprintable characters within the first 1024 bytes of file */
/*                                                             */
/*  Input parameters:                                          */
/*                       afFile  *afptr                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*           dmBool  is_binary  - true or false                */
/*                                                             */
/***************************************************************/
static dmBool afp_file_is_binary( afFile* afptr )
{
#define NCHARS_SCAN 1024

  const char  *pcLine      = NULL;
  dmBool       bIsBinary   = dmFALSE;
  long         StartPos   = afk_file_tell( afptr );
  int          iNumOfChars = 0;


  // initialize file pointer
  afk_file_seek( afptr, 0, SEEK_SET );

  do
  {
    if ( !( pcLine = afk_file_getline_ptr( afptr ) ) )
      break;

    // check for unprintable characters

    while ( *pcLine && ( iNumOfChars < NCHARS_SCAN ) )
    {
      if ( !isprint( *pcLine ) && !isspace( *pcLine ) )
      {
        bIsBinary = dmTRUE;
        break;
      }
     ++pcLine;
     ++iNumOfChars;
    }

  } while ( !bIsBinary && ( iNumOfChars < NCHARS_SCAN ) );


  afk_file_seek( afptr, StartPos, SEEK_SET );

  return bIsBinary;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dataset_scan                                */
/*                                                             */
/* Description:                                                */
/*   scans dataset.  determines flavor if none specified       */
/*                                                             */
/*  Input parameters:                                          */
/*                       dmDataset  *ds - pointer to dataset   */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode   ErrCode - success or failure         */
/*                                                             */
/***************************************************************/
dmErrCode afp_dataset_scan( dmDataset *ds )
{
  afDataset   *pakAsciiDataSet  = NULL;
  afStyle     *pakStyle         = NULL;
  afFile      *pakFilePointer   = NULL;
  const char  *pcLine           = NULL;
  const char  *pcLineLookPtr    = NULL;
  afFlavour    akFlavor         = AK_FLAV_UNK;

  long         NumberOfOldBlocks = 0;
  long         BlockNumber       = 0;
  long         LineOffset        = 0;

  dmBool       bFixed      = UT_FALSE;
  dmBool       bHeader     = UT_FALSE;

  char         cHeaderChar         = '#';
  char        *pcFieldSeparators   = NULL;
  int          iNumberOfFieldSep   = 0;
  int          iNumberOfTextStr    = 0;
  int          ii                  = 0;          // loop counter
  dmBool       bTextStringFound    = UT_FALSE;
  dmBool       bSeparatorFound     = UT_FALSE;

  char        *sDefaultSep = " \t\r";


  // If null pointer, return
  if ( !ds )
    return dmFAILURE;

  // get pointer to the ASCII dataset parameter for the dataset
  pakAsciiDataSet  = (afDataset*)dmi_dataset_get_kds( ds );

  // if null, return
  if ( !pakAsciiDataSet )
    return dmFAILURE;



  /* make sure blocks from any previous scans are deleted */
  NumberOfOldBlocks = dmi_dataset_get_n( ds );

  if (  NumberOfOldBlocks > 0 )
  {
    /* Don't do this if there are open blocks! */
    if ( dmi_dataset_get_nv( ds ) > 0 )
    {
      sprintf( dm_errbuf, "Cannot rescan TEXT dataset.  Blocks have already been accessed.\n" );
      dmu_err_set( DM_ERR_OPEN_FAIL, dm_errbuf );
      return dmFAILURE;
    }

    /* array delete is faster backwards */
    for ( BlockNumber =  NumberOfOldBlocks; BlockNumber >= 1; BlockNumber-- )
       afp_dataset_delete_block( ds, BlockNumber );
  }


  // With old blocks deleted, proceed.

  // Get style definitions from the Ascii dataset
  pakStyle  = afk_dataset_get_style( pakAsciiDataSet );

  // Get file flavor 
  akFlavor  = pakStyle->flavor;

  // Get the header indicator character if not NULL
  if ( pakStyle->header_char != '\0')
    cHeaderChar = pakStyle->header_char;

  // Get pointer to field separators from the ascii style structure
  pcFieldSeparators = pakStyle->sep;

  // How many defined separators does this file use?
  if (pcFieldSeparators == NULL)
    pcFieldSeparators = cs_dup(sDefaultSep);

  iNumberOfFieldSep = strlen(pcFieldSeparators);


  // Get file pointer and  file type from the Ascii dataset
  pakFilePointer  = afk_dataset_fptr( pakAsciiDataSet );

  // Rewind the file position pointer
  afk_file_seek( pakFilePointer, 0, SEEK_SET );

  // Note current file position in style structure
  pakAsciiDataSet->pos = afk_file_tell( pakFilePointer ); 

  // Get pointer to first line
  pcLine = afk_file_getline_ptr( pakFilePointer );

  if ( !pcLine )
  { 
    // Problem reading line from file.. problem with the file
    //    or maybe the line is too long.
    return dmFAILURE;
  }


  // Has the file type been determined prior to entry to this function?
  // If so, akFlavor has a type set to something other than AK_FLAV_UNK
  // However, look for a kernel specification line as the first line of
  // the file, because if it is present, the file position must be moved
  // to the next line before continuing

  // Is this line a kernel specification line?
  // Kernel specification line starts with #TEXT

  if ( cs_eq_siz( pcLine, "#TEXT", 5) )
  {
    // There is a kernel specification line

    if ( akFlavor == AK_FLAV_UNK )
    {
      // No flavor was specified earlier, so use this one.
      // Validate specification and set style to this flavor.

      if (! afp_verify_and_set_flavor ( (pcLine+1), &akFlavor, &bFixed) )
        return dmFAILURE;
    }
 
    // Adjust file position pointer so that it points here..
    // The flavor spec line should be ignored for the rest of the routine.

    if ( akFlavor != AK_FLAV_UNK )
      pakAsciiDataSet->pos = afk_file_tell( pakFilePointer );

  } /* end if (cs_eq_siz( pcLine, "#TEXT", 5)) */   


  // Skip any lines as indicated on command line
  if ( pakStyle->skip > 0 )
  {
    afk_file_seek( pakFilePointer, pakAsciiDataSet->pos, SEEK_SET );

    for ( LineOffset = 0; LineOffset < pakStyle->skip; LineOffset++ )
    {
      if ( !( pcLine = afk_file_getline_ptr( pakFilePointer )))
        break;
    }
    // Note current file position
    pakAsciiDataSet->pos = afk_file_tell( pakFilePointer);
  }


  // if no header type found, try to interpret what is there
  if (akFlavor == AK_FLAV_UNK)
  {
    // Check to see if file is binary.  Function fails if it is.
    if ( afp_file_is_binary( pakFilePointer ) )
      return dmFAILURE;

  
    // Not binary - determine the ASCII file type
    // Return to the start of the file, otherwise the
    //   first line to read gets skipped.

    afk_file_seek(  pakFilePointer, pakAsciiDataSet->pos, SEEK_SET );
    

    // Code Change  Bug 11905  August 21, 2009
    //  At this point, call function afp_dataset_scan_cxcrdb() to 
    //  investigatte if this file is cxcrdb format. This format
    //  does not follow the ordinary ascii rules and is handled
    //  separately

    akFlavor = afp_cscrdb_dataset_scan(pakAsciiDataSet, ds, pakStyle);

    // Put file back where it was before wandering off in cscrdb function
    afk_file_seek(  pakFilePointer, pakAsciiDataSet->pos, SEEK_SET );

   if (akFlavor == AK_FLAV_UNK)
   {
    // If the line starts with a header character and is followed
    // by a mixture of separators and alphanumeric
    // characters, then the line is interpreted as a set of column headings
    // and the flavor is TEXT/SIMPLE.  If it has a structure of
    // "keyword = ", then it is interpreted as TEXT/DTF.
    // Otherwise, it is TEXT/RAW.

    // Get pointer to first line to scan
    pcLine = afk_file_getline_ptr( pakFilePointer );


    // skip blank lines at the start of the file
    while ( pcLine && ( pcLine[0] == '\0') )
      pcLine = afk_file_getline_ptr (pakFilePointer);

    // Look at first non-blank line
    pcLineLookPtr = pcLine;
    
    // Is the character a comment character?
    if ( pcLineLookPtr )
      if (pcLineLookPtr[0] == cHeaderChar)
      {
        // Set the header flag
        bHeader  = UT_TRUE;

        // increment the pointer to look at the next character
        ++pcLineLookPtr;
      }


    // Scan through line to determine if the format is DTF

    while (pcLineLookPtr && *pcLineLookPtr != '\0')
    {
      // Is the character a seperator character?
      bSeparatorFound = UT_FALSE;

      for (ii = 0; ii < iNumberOfFieldSep; ii++)
      {
        if (*pcLineLookPtr == *(pcFieldSeparators + ii))
        {
	  bSeparatorFound = UT_TRUE;

          // if text string found prior to separator, increment text string count
          if (bTextStringFound)
            ++iNumberOfTextStr;

          break;  // break for loop
	}
      }

      // if separator not found, what is it?
      if ( ! bSeparatorFound )
      {
        if ( *pcLineLookPtr ==  '=')
	{
          if (! bTextStringFound )
	  {
            // Malformed expression or bizarre column name
            // set flavor to TEXT/SIMPLE
            akFlavor = AK_FLAV_SIMPLE;
            bHeader  = UT_TRUE;
          }
          else
	  {
	    // The equal sign indicates the end of a text string keyword
            // bTextStringFound is set TRUE

            // Only set to DTF when line is a well formed expression
            // This implies that the number of text strings found before
            //   the equal sign is 1.  Multiple strings imply malformed
            //   expression and the kernel specification defaults to RAW
 
            if (iNumberOfTextStr <= 1)
            {
              // set flavor to DTF
              akFlavor = AK_FLAV_DTF;
              bHeader  = UT_TRUE;

              // check here for fixed format DTF file
              while (pakFilePointer && pcLine)
              {
                // get the next line

                if ( !( pcLine = afk_file_getline_ptr( pakFilePointer )))
                  break;
                    
                // end of header?
	        if ( cs_eq_siz( pcLine, "END", 3 ))
		  break;

		// fixed size column descriptor present?
		if ( cs_eq_siz (pcLine, "TBCOL", 5) )
                {
                  bFixed = UT_TRUE;
                  break;
                }
	      } // end while (pakFilePointer && pcLine) loop
            } // end if (iNumberOfTextStr == 1)
          } // else !bTextStringFound
                   
          break;  // break outer while (pcLineLookPtr... loop

        } // end if ( *pcLineLookPtr ==  '=')
                
        // if character is not an =, then it must be part of a keyword or column header
        bTextStringFound = UT_TRUE;

      } // end if (!bSeparatorFound)

      ++pcLineLookPtr;
 
    } // end while (pcLineLookPtr && *pcLineLookPtr != '\0') loop



    // once through line or header, check for flavor still not set
    if (akFlavor == AK_FLAV_UNK)
    {
      // Code change  Feb 28, 2006
      // adding check for change in column names

      if ( bHeader &&
           ( bTextStringFound ||
             (pakStyle->colnames == AK_STYLE_COLNAMES_FIRST || pakStyle->colnames == AK_STYLE_COLNAMES_LAST)))
        akFlavor = AK_FLAV_SIMPLE; 
      else
        akFlavor = AK_FLAV_RAW;
    }
   } // end if flavor == AK_FLAV_UNK -- not cxcrbd type
  } // end if (akFlavor == AK_FLAV_UNK) -- no header found


  // Set the flavor and default style parameters
  afk_style_set( pakStyle, akFlavor );

  // don't override kernel options!
  afp_set_style_hints( NULL, pakStyle );


  // Code change  Bug 11084 October 7, 2008
  //   header and no column names implies raw
  //   Can't do this check until this point, unfortunately

  if ( (akFlavor == AK_FLAV_SIMPLE) &&
       (pakStyle->colnames == AK_STYLE_COLNAMES_NONE) )
  {
     akFlavor = AK_FLAV_RAW;
     afk_style_set( pakStyle, akFlavor );
     afp_set_style_hints( NULL, pakStyle);
  }


  if ( bFixed )
  {
    pakStyle->fixed = UT_TRUE;

    // Changing the fixed field can change the 
    // string representation..
    afk_style_set_flavor( pakStyle, akFlavor, NULL );
  }

  if ( bHeader )
    pakStyle->has_header = UT_TRUE;


  /* Add the block.. for DTF, scan for other blocks  */
  if ( akFlavor == AK_FLAV_DTF )
    afp_dtf_dataset_scan( ds );
  else
  {
    char* blockname = dmi_dataset_select_block_name(ds, NULL);
    afp_dataset_add_block( ds, 1, dmTABLE, blockname, pakAsciiDataSet->pos );
    free(blockname);
  }


  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_verify_and_set_flavor()                     */
/*                                                             */
/* Description:                                                */
/*   Compare first line of header with expected file type label*/
/*   strings to determine the file type.                       */
/*                                                             */
/*   There was a name change in March 2007, which made "simple"*/
/*   change to "raw" and "sm" change to "simple"               */
/*   Just in case there are some old ascii data files with     */
/*   "#TEXT/SM" in them, check for it and if present, assign   */
/*   it to file type "simple"                                  */ 
/*                                                             */
/*  Input parameters:                                          */
/*                   const char *pcLine  - kernel spec. line   */
/*                                                             */
/*  Output parameters:                                         */
/*                   afFlavour  *akFlavor  - spec. type        */
/*                   dmBool     *bFixed  - for dtf fixed       */
/*                                                             */
/*  Return:                                                    */
/*           dmBool - TRUE or FALSE                            */
/*                                                             */
/***************************************************************/
dmBool afp_verify_and_set_flavor( const char *pcLine,
                                  afFlavour  *akFlavor,
                                  dmBool     *bFixed)
{
  dmBool  bValidity = UT_TRUE;


  // assume non-fixed format until otherwise determined
  *bFixed = UT_FALSE;

  // Scan for TEXT/--- keyword
  // If kernel specification found, set "flavor"  

  if ( cs_uceq_siz(pcLine, "TEXT/DTF", 8) )
  {
    *akFlavor  = AK_FLAV_DTF;

    if ( cs_uceq_siz( pcLine, "TEXT/DTF-FIXED", 14 ) )
    {
      *bFixed = UT_TRUE;
    }
  }
  else
    // accommodate old label, TEXT/SM, as well as new
    // label, TEXT/SIMPLE

    if ( cs_uceq_siz( pcLine, "TEXT/SIMPLE", 11 ) ||
         cs_uceq_siz( pcLine, "TEXT/SM", 7) )
    {
      *akFlavor   = AK_FLAV_SIMPLE;
    }
    else
      if ( cs_uceq_siz( pcLine, "TEXT/RAW", 8 ) )
      {
        *akFlavor    = AK_FLAV_RAW;
      }
      else
        // Code Change  Bug 11905  August 21, 2009
        if ( cs_uceq_siz( pcLine, "TEXT/CSCRDB", 11 ))
        {
          *akFlavor  = AK_FLAV_CSCRDB;
        }
        else
          // Code Change  Bug 11940  March 9, 2010
          if ( cs_uceq_siz( pcLine, "TEXT/TAB-SEPARATED-VALUES", 25 ) ||
               cs_uceq_siz( pcLine, "TEXT/TSV", 8) )
          {
            *akFlavor  = AK_FLAV_TSV;
          }
          else
            if ( cs_uceq( pcLine, "TEXT" ) )
            {
              // Code Change  Bug 10556  November 8, 2007
              // set default to simple
              *akFlavor    = AK_FLAV_SIMPLE;
            }
            else
            {
              // don't know what this is, so return failure
              bValidity = UT_FALSE;
            }

  return bValidity;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dataset_close                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                       dmDataset  *ds      - current dataset */
/*                       dmBool      delete  - delete or not   */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Return:                                                    */
/*           nothing                                           */
/*                                                             */
/***************************************************************/
void afp_dataset_close( dmDataset *ds, dmBool delete )
{
  afDataset  *pakAsciiDataSet = NULL;

  if ( !ds )
    return;
 
  pakAsciiDataSet = (afDataset*)dmi_dataset_get_kds( ds );

  if ( !pakAsciiDataSet )
    return;

  afk_dataset_close( pakAsciiDataSet, delete );
  afk_dataset_free( pakAsciiDataSet );
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dataset_add_block                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                       dmDataset   *ds    - dataset          */
/*                       long         BlockNumber              */
/*                       dmBlockType  dmBlockType              */
/*                       const char  *pcBlockName              */
/*                       long         Position                 */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*              nothing                                        */
/*                                                             */
/***************************************************************/
void afp_dataset_add_block( dmDataset   *ds,
                            long         BlockNumber,
                            dmBlockType  dmBlockType,
                            const char  *pcBlockName,
                            long         Position )
{
  afDataset  *pakAsciiDataSet = NULL;
  long        AllocBlock      = 0;


  // if dataset pointer NULL, return
  if ( !ds )
    return;
 
  // Get pointer to Ascii dataset
  pakAsciiDataSet = (afDataset*)dmi_dataset_get_kds( ds );


  if ( BlockNumber > pakAsciiDataSet->n )
  {
    afk_dataset_resize_blocks( pakAsciiDataSet, BlockNumber );
    pakAsciiDataSet->n = BlockNumber;

    if ( pakAsciiDataSet->blocks[BlockNumber - 1] )
      free( pakAsciiDataSet->blocks[BlockNumber - 1] ); 

    pakAsciiDataSet->blocks[BlockNumber - 1] = cs_dup(pcBlockName);
  }

  pakAsciiDataSet->blocks_pos[BlockNumber - 1] = Position;
  AllocBlock = dmi_dataset_block_alloc( ds, BlockNumber );

  dmi_dataset_set_block_type( ds, BlockNumber, dmBlockType );
  dmi_dataset_set_block_name( ds, BlockNumber, pcBlockName );
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dataset_delete_block                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                       dmDataset   *ds                       */
/*                       long         BlockNumber              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*           nothing                                           */
/*                                                             */
/***************************************************************/
void afp_dataset_delete_block( dmDataset* ds, long BlockNumber )
{
  long        Number = 0;
  afDataset*  pakAsciiDataSet = NULL;

  // Get the Ascii dataset pointer from the dataset
  pakAsciiDataSet = (afDataset*)dmi_dataset_get_kds( ds );


  if ( pakAsciiDataSet && BlockNumber <= pakAsciiDataSet->n && BlockNumber >= 1)
  {
    dmi_block_delete( ds, BlockNumber );

    free( pakAsciiDataSet->blocks[BlockNumber - 1] );

    for ( Number = BlockNumber; Number < pakAsciiDataSet->n; Number++ )
    {
      pakAsciiDataSet->blocks[Number - 1]     = pakAsciiDataSet->blocks[Number];
      pakAsciiDataSet->blocks_pos[Number - 1] = pakAsciiDataSet->blocks_pos[Number];
    }

   --(pakAsciiDataSet->n);
  }
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dataset_notify_moved                        */
/*                                                             */
/* Description:                                                */
/*   notify the dataset that some of its file contents have    */ 
/*   been moved.  Need to adjust all file position markers in  */ 
/*   blocks > blockno by amount 'moved'.                       */
/*                                                             */
/*  Input parameters:                                          */
/*                       dmDataset   *ds                       */
/*                       dmBlock     *moved_block              */
/*                       long         moved                    */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*              nothing                                        */
/*                                                             */
/***************************************************************/
void afp_dataset_notify_moved( dmDataset* ds,
                               dmBlock* moved_block,
                               long moved )
{
  dmBlock   *pdmBlock         = NULL;
  afDataset *pakAsciiDataSet  = NULL;
  long       Offset           = 0;
  long       BlockNumber      = 0;
  long       NumberV          = 0;
  long       MovedBlockNumber = 0;
 

  pakAsciiDataSet  = (afDataset*)dmi_dataset_get_kds( ds );
  MovedBlockNumber = dmi_block_get_no( moved_block );


  Offset = MovedBlockNumber;

  if ( Offset >= pakAsciiDataSet->n )
    return;


  for ( ; Offset < pakAsciiDataSet->n; Offset++ )
  {
    pakAsciiDataSet->blocks_pos[Offset] += moved;
  }

  NumberV = dmi_dataset_get_nv( ds );

  for ( BlockNumber = 1; BlockNumber <= NumberV; BlockNumber++ )
  {
    pdmBlock = dmi_dataset_get_block( ds, BlockNumber );

    if ( pdmBlock )
    {
      afp_block_notify_moved( pdmBlock, moved_block, moved );
    }
  }
}


