/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afdtf.c                                                     */
/*                                                               */
/*   Functions that create and work with dtf type files          */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"
#include "../ftkernel/ftkernelP.h"
#include "ryi.h"


static dmDescriptor** afp_dtf_read_cols( dmBlock* block );
static void afp_dtf_write_col_keys( dmBlock* block, dmDescriptor* dd );
static void afp_dtf_move_col_keys( dmBlock* block, dmDescriptor* dd, long offset );

void afp_init_tnull( dmuKeyList keylist, dmDescriptor *dd, dmBool is_image );



/***************************************************************/
/*                                                             */
/* Function:   afp_dtf_init_cols                               */
/*                                                             */
/* Description:                                                */
/*              Initialize column names                        */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock   *block                         */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode   Status                               */
/*                                                             */
/***************************************************************/
dmErrCode afp_dtf_init_cols( dmBlock *block )
{
  dmErrCode        status  = dmSUCCESS;
  dmuKeyList       keylist;
  dmDescriptor   **rawcols = NULL;
  dmDescriptor   **vectors = NULL;
  dmDataset       *ds      = NULL;
  dmBlockType      blockType = dmUNKBLOCK;
  afBlock         *ablock = NULL;
  afDataset       *ads    = NULL;
  afStyle         *astyle = NULL;
  const char      *name   = NULL;
  long             ncols  = 0;
  long             nvcols = 0;


  ablock = (afBlock*)dmi_block_get_kb( block );
  keylist = afk_block_keylist( ablock );

  /* create the kernel layer columns */
  blockType = dmi_block_get_type( block);

  /* code below changed for Bug 9342  1/5/2006  */
  /* Get a pointer to the key list              */
  /* If the pointer to the data section is NULL */
  /* call afp_raw_init_cols() because there     */
  /* are no column names stored                 */

  if (( blockType != dmIMAGE ) && (!keylist || keylist->data == NULL))
  {
    status = afp_raw_init_cols (block);
    return status;
  }
  /* end code change for Bug 9342 */


  ds     = dmi_block_get_ds( block );
  ads    = (afDataset*)dmi_dataset_get_kds( ds );
  astyle = afk_dataset_get_style( ads );


  // Code Change - Bug 9023  August 11, 2006
  if ( blockType == dmIMAGE )
  {
    name = afk_dataset_block_name( ads, (long)ablock->no );
    ablock->name = cs_dup(name);

    status = afk_dtf_init_image_cols( ablock, astyle );
  }
  else
    status = afk_dtf_init_cols( ablock, astyle );

  /* create the DM layer columns */
  rawcols = afp_dtf_read_cols( block );

  ncols = ablock->ncols;

  afp_gen_find_transforms( rawcols, ncols, keylist );
  ftp_composite_keylist_process( block, keylist );

  vectors = afp_dtf_vectorize( rawcols, ncols, keylist, &nvcols );
  dmi_block_set_raw_cols( block, vectors, nvcols );

  /*  need to free rawcols */
  free (rawcols);
  

  return status;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dtf_read_cols                               */
/*                                                             */
/* Description:                                                */
/*            Use header info and kernel cols to create        */
/*            dmDescriptor columns                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock   *block                         */
/*                                                             */
/*  Output parameters:                                         */
/*                    long      *ncols                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor **cols                              */
/*                                                             */
/***************************************************************/
dmDescriptor** afp_dtf_read_cols( dmBlock *block )
{
  char keyword[CARD_SIZE];
  char buf[CARD_SIZE];
  char tunit[CARD_SIZE];

  dmuKeyList       keylist;
  dmDescriptor   **cols   = NULL;
  dmDescriptor    *dd     = NULL;
  dmDescriptor    *us_dd  = NULL;
  dmuKey          *akey   = NULL;
  afBlock         *ablock = NULL;
  afCol           *acol   = NULL;

  dmDataType  type           = DT_UNK;
  dmBool      force_unscaled = dmFALSE;
  dmBool      is_scaled      = dmFALSE;
  dmBool      is_image       = dmFALSE;

  long  *axes    = NULL;
  long   naxes   = 0;
  long   colno   = 0;
  long   ii      = 0;
  long   nkcols  = 0;
  long   index   = 0;
  long   slength = 0;
  double tscal   = 1.0;
  double tzero   = 0.0;
  double pix     = 0.0;
  long bitpix    = 32;

  char *desc  = NULL;
  char *unit  = NULL;


  ablock = (afBlock*)dmi_block_get_kb( block );
  keylist = afk_block_keylist( ablock );

  nkcols = afk_block_ncols( ablock );

  if ( nkcols > 0 )
    cols = (dmDescriptor**)calloc( nkcols, sizeof(dmDescriptor*) );

  for ( ii = 0; ii < nkcols; ii++ )
  {
    colno   = ii+1;
    acol    = afk_block_get_col( ablock, colno );
    type    = acol->datatype;
    slength = 0;

    /* get the column description */
    sprintf( keyword, "TTYPE%ld", colno );

    akey = dmu_keylist_read_value_c( keylist, keyword, buf, CARD_SIZE );

    // Code Change Bug 9023  May 21, 2007
    // Let all descriptions pass through to output

    if ( akey && akey->desc )
      desc = cs_dup( akey->desc );

    /* get the column units */
    sprintf( keyword, "TUNIT%ld", colno );
    akey = dmu_keylist_read_value_c( keylist, keyword, buf, CARD_SIZE );

    if ( akey && akey->value )
      unit = cs_dup( dmu_key_value_c( akey ) );


    // Code Change - Bug 9023 - September 14, 2006
    is_image = ( block->type == dmIMAGE );

    switch (block->type )
    {
      case dmTABLE:
      default:
        /* get the columns scale / offset */
        tscal = 1.0;
        tzero = 0.0;
        sprintf( keyword, "TSCAL%ld", colno );
        akey = dmu_keylist_read_typed_value( keylist, keyword, DT_REAL8, &tscal );
        sprintf( keyword, "TZERO%ld", colno );
        akey = dmu_keylist_read_typed_value( keylist, keyword, DT_REAL8, &tzero );

        if ( type == DT_INT4 && tscal == 1.0 && tzero == ULONG_TZERO ) 
        {
          type = DT_UINT4;
          tscal = 1.0;
          tzero = 0.0;
        }
        else
          if ( type == DT_INT2 && tscal == 1.0 && tzero == USHORT_TZERO )
          {
            type = DT_UINT2;
            tscal = 1.0;
            tzero = 0.0;
          }

        // Code Change  Bug 11836  July 24, 2009
        if ( type == DT_UINT2 || type == DT_UINT4 )
          is_scaled = dmFALSE;
        else
          is_scaled = ( tscal != 1.0 || tzero != 0.0 );

        /* get column array dimensionality */
        if ( acol->length )
        {
          sprintf( keyword, "TDIM%ld", colno );
          akey = dmu_keylist_read_value_c( keylist, keyword, buf, CARD_SIZE );

          if ( akey )
          {
            ftp_parse_tdim( buf, &axes, &naxes );

            // Code change - Bug 10042  November 3, 2006
            //  Redefine axes/naxes pair to save string length
            //  for char types

            if ( type == DT_CHAR )
            {
              axes[0] = acol->length;
              naxes   = 1;
            }
          }
          else
          {
            // Code Change  Bug 10785  November 7, 2007
            // There is no key to define dimensionality
            //  so set naxes to 1 and axes[0] to length
            //  For char types, do the same to save string

            if ( acol->length > 1 || type == DT_CHAR )
            {
               naxes   = 1;
               axes    = (long*)malloc( sizeof(long) );
               axes[0] = acol->length;
            }
          }
        } /* end if (acol->length) */
        else
        {
	  if ( type == DT_BIT )
	  {
	    /* DM stores array of bytes.. */
	    naxes = 1;
	    axes = (long*)malloc( sizeof(long) );
	    axes[0] = (acol->width-1)/8 + 1;
	  }
	  else
	  {
	    naxes = 1;
	    axes = (long*)malloc( sizeof(long) );
	    axes[0] = 1;
	  }
        }
        break;

      case dmIMAGE:
        // Code Change - Bug 9023

        // Check for scaling keywords for image file
        //  If keys are not present, tzero and tscal retain their default values

        akey = dmu_keylist_read_typed_value( keylist, "BITPIX", DT_INT4,  &bitpix );
        akey = dmu_keylist_read_typed_value( keylist, "BZERO",  DT_REAL8, &tzero);
        akey = dmu_keylist_read_typed_value( keylist, "BSCALE", DT_REAL8, &tscal );

        // Determine data type
        type = ftp_bitpix_to_type( bitpix, tscal, tzero );

        // Code Change - Bug 9023 - September 14, 2006
        // The following section has been modified to match
        // the FITS code.

        if ( type == DT_UINT2 || type == DT_UINT4 )
	{
          is_scaled = dmFALSE;
          tscal = 1.0;
          tzero = 0.0;
	}
        else
          is_scaled = ( tscal != 1.0 || tzero != 0.0 );

	// Set kernel descriptor data type 
        acol->datatype = type;

        akey = dmu_keylist_read_value_c( keylist, "BUNIT", tunit, FITS_MAXLEN );
        if ( akey )
          unit = cs_dup( tunit );

        naxes = ablock->naxes;
	if ( naxes )
	{
	  /* transfer axes info into local variable    */
	  /* also calculate column length at same time */

	  slength = 1;
	  axes = (long*)malloc( naxes * sizeof(long) );
	  for (index = 0; index < naxes; index++ )
	  {
	    axes[index] = ablock->axes[index];
	    slength    *= ablock->axes[index];
	  }
	}
	break;

    } /* end switch */
 

    /* Allow user to globally suppress scaling */
    force_unscaled = dmi_block_is_unscaled( block );

    // Code Change  Bug 11836  July 24, 2009
    //  only set is_scaled to false if user is forcing issue

    if ( force_unscaled && is_scaled )
    {
      is_scaled = dmFALSE;
      tscal = 1.0;
      tzero = 0.0;
    }


    //  Code Change Bug 10312 March 21, 2007
    //    If type is DT_CHAR, set value for
    //    slength in order for DM buffer to be
    //    allocated correctly

    // Get length

    if ( type == DT_CHAR )
    {
      if (acol->width)
        slength = acol->width;
      else
        slength = AK_COLVAL_LEN;
    }
    else
      if ( type == DT_BIT )
      {
        if ( acol->width )
          slength = acol->width;
        else
          slength = 8;
      }
      else
        if ( type == DT_LOGICAL )
	{
          // Code Change Bug 11791  February 25, 2010
          //   Make a note of whether the boolean column uses a single 
          //   character (T or F) or the whole word (TRUE or FALSE) by 
          //   using the value stored in acol->width. This is for future
          //   use to facilitate exact copying of this type of column
          //   from one ascii flavor to another.

          if ( acol->width )
            slength = acol->width;
        }


    /* allocate the column descriptor */

    dd = ryi_column_alloc( block, colno, acol->name, type,
			   slength, unit, desc, 1, NULL,
			   naxes, axes, tzero, pix, tscal, acol );

    if ( acol->disp )
      dd->disp = cs_dup( acol->disp );

    us_dd = dmi_descriptor_get_unscaled( dd );


    /*    --- Get column range --- */
    ftp_init_col_range( keylist, us_dd );

    // get TDBIN and TDNULL values and store in descriptor
    ftp_init_col_bin_null( keylist, us_dd, is_image );

    // Get TNULL value and store in descriptor
    afp_init_tnull(keylist, us_dd, is_image);

    if ( !is_scaled )
    {
      ftp_init_col_range( keylist, dd );
    }

    /* Dubious: null is really for unscaled col           */
    /* get TDBIN, TDNULL and TNULL values and store in descriptor */

    ftp_init_col_bin_null( keylist, dd, is_image );
    afp_init_tnull(keylist, dd, is_image);

    cols[ii] = dd;

    if ( naxes > 0 )
    {
      if (axes)
        free( axes );
      axes = NULL;
      naxes = 0;
    }

    if ( desc )
    {
      free( desc );
      desc = NULL;
    }

    if ( unit )
    {
      free( unit );
      unit = NULL;
    }
  } /* end for(...) */

  return cols;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dtf_vectorize                               */
/*                                                             */
/* Description:                                                */
/*              Group the individual FITS descriptors into     */
/*              DM vector descriptors                          */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor  **cols                     */
/*                    long            ncols                    */
/*                    dmuKeyList      keylist                  */
/*                    long           *nvcols                   */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor  **vectors                          */
/*                                                             */
/***************************************************************/
dmDescriptor** afp_dtf_vectorize( dmDescriptor **cols, long ncols,
                                  dmuKeyList keylist,  long *nvcols )
{
  dmuAliasKeyList  mlist   = NULL; /* List of MTYPE/MFORM aliases */
  dmDescriptor   **vectors = NULL;

  /*  Go through the MTYPE keywords and figure out which are columns */
  /*  and which are coord axes                                       */

  mlist = dmu_alias_key_make_list( keylist, "MTYPE", "MFORM", "METYP",
                                   "MENAM", 5 );
  
 
  /* Apply the MTYPE info to the WCS information                        */
  /* Handle cols at same time (since RA/DEC forces parent to be vector) */

  vectors = afp_dtf_map_vectors( cols, ncols, mlist, nvcols );

  dmu_alias_key_list_free( mlist );
  return vectors;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dtf_map_vectors                             */
/*                                                             */
/* Description:                                                */
/*              Map vector columns                             */
/*                                                             */
/*  Brave reader: this is arguably the most arcane part of the */
/*  code.                                                      */
/*  NoFtAxes is the number of raw axes for an image, the number*/
/*  of raw columns for a table.                                */
/*  We allocate a Vectors array this big to start with.        */
/*  We iterate over the columns (axes) in numerical order.     */
/*  First, we check if it is a pre-declared vector column      */
/*  (with MTYPE).  Next,we check if it has a coord system which*/
/*  is a predeclared MTYPE.  Finally, we check for RA,Dec pairs*/
/*  and treat them as implicitly defined vectors.              */
/*  For any of these cases, we deal with the other components  */
/*  of the column.  If none of these cases, we treat it as a   */
/*  scalar column.                                             */
/*                                                             */  
/*  Now, we mask off all the components we've used and ignore  */
/*  them when we reach them in the iteration.                  */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor  **cols                     */
/*                    long            ncols                    */
/*                    dmuAliasKeylist mlist                    */
/*                    long           *nvcols                   */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor  **vectors                          */
/*                                                             */
/***************************************************************/
dmDescriptor** afp_dtf_map_vectors( dmDescriptor **cols,
                                    long ncols,
                                    dmuAliasKeyList mlist,
                                    long *nvcols )
{
  dmDescriptor  **vectors = NULL;

  dmDescriptor   *vector  = NULL;   /* An individual vector */
  dmDescriptor   *dd      = NULL;   
  dmDescriptor   *parent  = NULL;
 
  long ftaxisno = 0;
  long nvectors = 0;
 
 
  if ( ncols > 0 )
    vectors = calloc( ncols, UT_SZ_P );

  nvectors = 0; 


  for ( ftaxisno = 1; ftaxisno <= ncols; ftaxisno++ )
  {
    dd = cols[ftaxisno-1];
    parent = dmi_descriptor_get_parent( dd );

    if ( !parent  )   /* Not parented yet */
    {
      vector = afp_dtf_column_match( cols, ncols, ftaxisno, mlist );

      if ( vector )
      {
        if ( nvectors > ncols )
          vectors = realloc( vectors, (nvectors+1) * UT_SZ_P );

        vectors[nvectors++] = vector;
        dmi_descriptor_set_no( vector, nvectors );
      }

      /* Now look for axis groups on each column (e.g. for image) */
      afp_dtf_axis_match( cols, ncols, ftaxisno, mlist );
    }
  } /* End loop on rawaxis */

  *nvcols = nvectors;
 
  dmi_make_unscaled_groups( vectors, nvectors );
  return vectors;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dtf_column_match                            */
/*                                                             */
/* Description:                                                */
/*        We look for the column at a given raw number and see */
/*        if it is a vector column. First we look to see if it */
/*        is in the MTYPE/MFORM list, then we look to see if   */
/*        its coordinate transforms are (in which case it      */
/*        inherits their vector-ness).                         */
/*        The order of the FITS columns (as in MFORM) is also  */
/*        the order we store things internally. The remapping  */
/*        is done with the 'mcpt_offset=' line below.          */
/*                                                             */
/*         Example (for the complicated METYP1 case which is   */
/*                  experimental)                              */
/*         MTYPE1 = VEC                                        */
/*         MFORM1 = XLO,YLO,ZLO,XHI,YHI,ZHI                    */
/*         METYP1 = R                                          */
/*         MENAM1 = X,Y,Z                                      */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor  **cols                     */
/*                    long            ncols                    */
/*                    long            ftaxisno                 */
/*                    dmuAliasKeyList mlist                    */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor  *vectors                           */
/*                                                             */
/***************************************************************/
dmDescriptor* afp_dtf_column_match( dmDescriptor  **cols,
                                    long   ncols,
                                    long   ftaxisno,
                                    dmuAliasKeyList   mlist )
{
  dmBlock       *block            = NULL;
  long           coord_no         = 0;
  dmDescriptor  *vector           = NULL;
  dmDescriptor  *transform_dd     = NULL;
  dmDescriptor  *transform_vector = NULL;
  dmDescriptor  *dd               = NULL;
  char          *name             = NULL;
  long           nc               = 0;


 
  /* Get logical (unscaled) column */  
  if ( ftaxisno > ncols )
    return NULL;

  dd = cols[ftaxisno-1];
  block = dmi_descriptor_get_block( dd );

  /* Look in MFORMn for the name of this column, and make a vector
     with the other names in that MFORMn */

  if ( mlist )
    vector = afp_dtf_dd_match( dd, cols, ncols, mlist );

  /* Now look for transform vectors */
  nc = dmi_descriptor_get_no_coords( dd );

  for ( coord_no = 1; coord_no <= nc; coord_no++ )
  {
    transform_dd = dmi_descriptor_get_coord_no( dd, coord_no );

    /* transform_dd is a singlet transform. Find the other components */
    /*  of its vector by name                                         */
    transform_vector = afp_dtf_dd_match( transform_dd, cols, ncols, mlist );

    /* Search for implicit vector transforms (FITS WCS convention) */
    if ( !transform_vector )
      transform_vector = afp_dtf_implicit_vector_transform( transform_dd,
                                                            cols, ncols );

    if ( transform_vector )
    {
      /* Now inherit col vector if needed */
      if ( !vector )
        vector = afp_dtf_dd_inherit( transform_vector );

      /* For subsequent transforms, make sure they are parented */
      dmi_descriptor_add_coord( vector, transform_vector );

      /* Break the links to the component transform parents */
      dmi_coord_detach_cpts( transform_vector );
    }


    /* It would be an error to have different dim transforms on a column. */
    /* We should check for this but don't                                 */

  }

  /* No vector found */

  if ( !vector )
  {
    name = dmi_descriptor_get_name( dd );
    vector = dmi_make_vector( block, name, &dd, 1 );
  }

  return vector;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dtf_dd_match                                */
/*                                                             */
/* Description:                                                */
/*  This routine looks up a name in the MTYPE/MFORM list and   */
/*  makes the corresponding vector descriptor.                 */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor   *col_dd                   */
/*                    dmDescriptor  **cols                     */
/*                    long            ncols                    */
/*                    dmuAliasKeyList mlist                    */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor  *vectors                           */
/*                                                             */
/***************************************************************/
dmDescriptor* afp_dtf_dd_match( dmDescriptor *col_dd,
                                dmDescriptor **cols,
                                long ncols,
                                dmuAliasKeyList mlist )
{
  dmBlock       *block  = NULL;
  dmDescriptor  *vector = NULL;
  char          *name   = NULL;
  char          *desc   = NULL;
  dmElementType  etype  = dmVALUE;
  char         **enames = NULL;
  dmDescriptor **cpts   = NULL;
  afCol         *acol   = NULL;
  afVector      *avec   = NULL;

  long  mtype_offset = 0;
  long  cpt_offset   = 0;       /* Offset of component within element */
  long  mcpt_offset  = 0;      /* Offset of component name in MFORM list */
  long  cptno        = 0;
  long  dim          = 0;
  long  maxlen       = CARD_SIZE;

  char  cptname[CARD_SIZE];
  char  vname[CARD_SIZE];
  

  if ( !mlist )
    return NULL;

  block = dmi_descriptor_get_block( col_dd );
  name  = dmi_descriptor_get_name( col_dd );

  /* Check if this name appears in an MFORM (or MENAM) */

  mtype_offset = ar_alias_search( mlist->cpts, name );

  if( mtype_offset >= 0 )
  {
    /* We have a vector. Let's make up the remaining entries. */
    ar_alias_get_name( mlist->names, mtype_offset, vname, maxlen );

    dim  = ar_alias_get_dim( mlist->cpts,  mtype_offset );
    cpts = calloc( dim, UT_SZ_P );
    avec = afk_vector_alloc( dim );

    for ( cpt_offset = 0; cpt_offset < dim; cpt_offset++ )
    {
      mcpt_offset = cpt_offset;
      cptno       = cpt_offset + 1;

      ar_alias_get_cpt( mlist->names, mtype_offset, mcpt_offset, cptname, maxlen );

      cpts[cpt_offset] =  dmi_descriptor_list_search((const dmDescriptor**)cols,
                                                      ncols, cptname );
      if ( !cpts[cpt_offset] )
      {
        sprintf( dm_errbuf, "Error: token %s in MTYPE list not found in descriptor list", cptname );
        dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
        return NULL;
      }

      acol = dmi_descriptor_get_kdd( cpts[cpt_offset]);

      // Code change  Bug 12080  December 2, 2009
      //   acol points to the kernel data descriptor of this component
      //   Modify this structure only if this descriptor
      //   is a column descriptor. Skip this section if the 
      //   descriptor is a dmCoord descriptor otherwise memory gets
      //   corrupted because the structs are different

      if ( acol && acol->type == FTK_COL )
      {
        acol->mtype_index = mtype_offset + 1;
      }

      avec->cpts[cpt_offset] = acol;
    } /* End for elt */


    desc   = mlist->desc[mtype_offset];
    vector = dmi_make_element_vector( block, vname, cpts, dim, etype, desc );

    dmi_descriptor_set_kdd( vector, avec );
    free( cpts );

    /* Handle element types */
    dim = ar_alias_get_dim( mlist->cpts, mtype_offset );

    ar_mix_get_elt( mlist->etypes, mtype_offset, NULL, &etype );

    enames = ar_alias_get_cpt_list( mlist->cpts, mtype_offset );

    dmi_descriptor_set_element_names( vector, etype, enames, dim );
  }

  return vector;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dtf_implicit_vector_transform               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor   *transform_dd             */
/*                    dmDescriptor  **cols                     */
/*                    long            ncols                    */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor  *vectors                           */
/*                                                             */
/***************************************************************/
dmDescriptor* afp_dtf_implicit_vector_transform( dmDescriptor  *transform_dd,
						 dmDescriptor **cols,
						 long           ncols )
{
  ftCoord       *ftcoord = NULL;
  dmDescriptor  *vector  = NULL;
  dmDescriptor  *cpts[2];
  dmBlock       *block = NULL;
  char alt_name[CARD_SIZE] = " ";


  if ( !transform_dd )
    return NULL;

  ftcoord = dmi_descriptor_get_kdd( transform_dd );
  block = dmi_descriptor_get_block( transform_dd );

  /* We recognize that e.g. GLON--SIN is part of a pair */
  if ( ftcoord && ftcoord->composite )     
  {
    ftp_find_wcs_pair_name( (char*)ftcoord->ctype_name, alt_name );

    if ( alt_name[0] == '\0' ) 
      return NULL;
 
    /* Now look for the paired column */
    cpts[0] = transform_dd;
    cpts[1] = ftp_match_ctype_name( cols, ncols, alt_name );  

    if ( !cpts[1] ) 
      return NULL;

    /* We denote the LONG/NPOL pair with a TAN-P transform type */
    if ( cs_eq( alt_name, "NPOL" ))
    {
      dmi_coord_set_transform_type( cpts[0], "TAN-P" );
      dmi_coord_set_transform_type( cpts[1], "TAN-P" );
    }
 
    /* Map TANP here */

    vector = dmi_make_vector( block, NULL, cpts, 2 );
  }

 return vector; 
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dtf_dd_inherit                              */
/*                                                             */
/* Description:                                                */
/*      For the vector transform, we make a vector column      */
/*      consisting of the parents of each component of the     */
/*      vector transform                                       */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor   *transform_vector         */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor  *vectors                           */
/*                                                             */
/***************************************************************/
dmDescriptor* afp_dtf_dd_inherit( dmDescriptor *transform_vector )
{
  dmBlock       *block  = NULL;
  dmDescriptor  *vector = NULL;
  long           cptno  = 0;
  long           dim    = 0;
  dmDescriptor  *transform_cpt_dd = NULL; 
  dmDescriptor **cpt_dd = NULL;


  dim    = dmi_descriptor_get_element_dim( transform_vector );
  cpt_dd = calloc( dim, UT_SZ_P );

  for  ( cptno = 1; cptno <= dim; cptno++ )
  {
    transform_cpt_dd = dmi_descriptor_get_cpt( transform_vector, cptno );
    cpt_dd[cptno-1]  = dmi_coord_get_parent( transform_cpt_dd );

    if ( !cpt_dd[cptno-1] )
      printf( "INTERNAL ERROR: LOOSE COORD CPT %ld %s\n", cptno,
              dmi_descriptor_get_name( transform_cpt_dd ));
  }

  block  = dmi_descriptor_get_block( transform_vector );
  vector = dmi_make_vector( block,  NULL, cpt_dd, dim );

  free( cpt_dd );
  return vector; 
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dtf_axis_match                              */
/*                                                             */
/* Description:                                                */
/*      Now we look at each column which is an array (image),  */
/*      to see if any of the axes on this image are grouped    */
/*      (vectorized).                                          */
/*      Note that grouping axes doesn't affect the vector      */
/*      dimension of the column itself.                        */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor  **cols                     */
/*                    long            ncols                    */
/*                    long            ftaxisno                 */
/*                    dmuAliasKeyList mlist                    */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_dtf_axis_match( dmDescriptor **cols, long ncols,
                         long ftaxisno,
                         dmuAliasKeyList mlist )
{
  dmBlock       *block            = NULL;
  dmDescriptor  *vector           = NULL;
  dmDescriptor **vectors          = NULL;
  dmDescriptor **ddlist           = NULL;
  dmDescriptor  *transform_dd     = NULL;
  dmDescriptor  *transform_vector = NULL;
  dmDescriptor  *dd               = NULL;   
  dmDescriptor  *col              = NULL;
  dmDescriptor  *parent           = NULL;

  char          *name = NULL;

  long ng       = 0;
  long coord_no = 0;
  long nc       = 0;
  long nwc      = 0;
  long group    = 0;



   /* Get physical (scaled) column */  

   if ( ftaxisno > ncols )
      return;

   col = cols[ftaxisno-1];

   if ( !col )
   {
     sprintf( dm_errbuf, "No physical column on axis %ld", ftaxisno );
     dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
     return;
   }

   block = dmi_descriptor_get_block( col );

   /* What if it's a vector column? */
   /* Find axis groups */

   nc = dmi_descriptor_get_no_groups( col );
   ddlist = calloc( nc, UT_SZ_P );
   vectors = calloc( nc, UT_SZ_P );

   for ( group = 1; group <= nc; group++ )
   {
     ddlist[group-1] = dmi_descriptor_get_group_no( col, group );
   }

   /* Now we have saved the raw groups we can reorder them */
   /* Maybe attach scalar groups to raw col?               */

   for ( group = 1; group <= nc; group++ )
   {
     vector = NULL;
     dd = ddlist[group-1];  
     parent = dmi_descriptor_get_parent( dd );
  
     if ( dd && !parent )
     {
       if ( mlist )
         vector = afp_dtf_dd_match( dd, cols, ncols, mlist );

       /* Now look for world transform vectors */
       nwc = dmi_descriptor_get_no_coords( dd );

       for ( coord_no = 1; coord_no <= nwc; coord_no++ )
       {
         transform_dd = dmi_descriptor_get_coord_no( dd, coord_no );
         transform_vector = afp_dtf_dd_match( transform_dd, cols, ncols,
                                             mlist );

         /* Search for implicit vector transforms (FITS WCS convention) */
         if ( !transform_vector )
           transform_vector = afp_dtf_implicit_vector_transform( transform_dd,
							         cols, ncols );

         /* transform_dd is a singlet transform. Find the other components */
         /* of its vector by name                                          */

         if ( transform_vector )
         {
           /* Now inherit col vector if needed */
           if ( !vector )
             vector = afp_dtf_dd_inherit( transform_vector );

           /* For subsequent transforms, make sure they are parented */
           dmi_descriptor_add_coord( vector, transform_vector );
           dmi_coord_detach_cpts( transform_vector );
         }
       } /* End loop world coords */

       /*  No vector found*/   
       if ( !vector )
       {
         name = dmi_descriptor_get_name( dd );
         vector = dmi_make_vector( block, name, &dd, 1 );
       }

       vectors[ng++] = vector;
    
      } /* End no parent */
    } /* End group */

  free( ddlist );
  dmi_descriptor_set_groups( col, vectors, ng );
  free( vectors );
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dtf_dataset_scan                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDataset  *ds                           */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afp_dtf_dataset_scan( dmDataset *ds )
{
  char tmpstr[CARD_SIZE];
  char kname[CARD_SIZE];
  char comment[CARD_SIZE];
  char unit[CARD_SIZE];

  char hduname[CARD_SIZE];
  char extname[CARD_SIZE];
  char verstr[CARD_SIZE];
  long extver = 1;

  dmErrCode status      = dmSUCCESS;

  dmBool found_xtension = dmFALSE;
  dmBool found_hduname  = dmFALSE;
  dmBool found_extname  = dmFALSE;
  dmBool found_extver  = dmFALSE;

  dmBlockType  blockType = dmTABLE;

  afFile     *afptr  = NULL;
  afDataset  *ads    = NULL;
  afStyle    *astyle = NULL;

  const char  *line = NULL;
  char        *pTmpLine = NULL;
  char        *tptr = NULL;

  long  nblocks   = 0;
  long  block_pos = 0;
  long  pos       = 0;
  int   index     = 0;

  ads = (afDataset*)dmi_dataset_get_kds( ds );
  astyle = afk_dataset_get_style( ads );
  afptr = afk_dataset_fptr( ads );

  /* Go to the start of the dataset  */
  afk_file_seek( afptr, ads->pos, SEEK_SET );

  pos = afk_file_tell( afptr );

  /* first XTENSION keyword is optional */
  block_pos = pos;
  found_xtension = dmTRUE;

  while( NULL != ( line = afk_file_getline_ptr( afptr ) ) )
  {
    if ( cs_eq_siz( line, "XTENSION", 8 ))
    {
      // Code Change - Bug 9023
      // Check here for "IMAGE" type

      pTmpLine = strchr( line, '=');

      // parse line - look at what is after the equals
      if (pTmpLine != NULL)
      {
        ++pTmpLine;

        afk_keyline_read_value( pTmpLine, tmpstr, CARD_SIZE);

        /* Code Change Bug 10313 May 2, 2007            */
        /*   Remove quotes from string values           */
	/*   10/02/2008: tweak to not write over itself */
	tptr = tmpstr;
        if (*tptr == '\'' || *tptr == '\"' )
	{
          tmpstr[(strlen(tmpstr))-1] = '\0';
	  tptr++;
        }
	strcpy(comment, tptr); 
 
        if (cs_eq( comment, "IMAGE"))
	{
           blockType = dmIMAGE;
	}
	else if (cs_eq( comment, "TABLE"))
	{
           blockType = dmTABLE;
	}
	else
	{
	  blockType = dmUNKBLOCK;
	}
      }
      // end Code Change Bug 9023


      if ( !found_xtension )
      {
        block_pos      = pos;
        found_xtension = dmTRUE;
      }
    }
    else if ( cs_eq_siz( line, "HDUNAME", 7 ))
    {
      if ( !found_hduname && found_xtension )
      {
	afk_gen_read_keyline( astyle, line,
			      kname, CARD_SIZE,
			      tmpstr, CARD_SIZE,
			      comment, CARD_SIZE,
                              unit, CARD_SIZE );

	found_hduname = dmTRUE;

	/* Code Change Bug 10313 May 2, 2007            */
	/*   Remove quotes if present                   */
	/*   10/02/2008: tweak to not write over itself */

	tptr = tmpstr;

	if (*tptr == '\'' || *tptr == '\"' )
  	{
	  tmpstr[(strlen(tmpstr))-1] = '\0';
	  tptr++;
	}
	strcpy(hduname, tptr); 
      }
    }
    else if ( cs_eq_siz( line, "EXTNAME", 7 ))
    {
      if ( !found_extname && found_xtension )
      {
	afk_gen_read_keyline( astyle, line,
			      kname, CARD_SIZE,
			      tmpstr, CARD_SIZE,
			      comment, CARD_SIZE,
                              unit, CARD_SIZE );

	found_extname = dmTRUE;

        /* Code Change Bug 10313 May 2, 2007            */
        /*   Remove quotes from string values           */
	/*   10/02/2008: tweak to not write over itself */

	tptr = tmpstr;

	if (*tptr == '\'' || *tptr == '\"' )
  	{
	  tmpstr[(strlen(tmpstr))-1] = '\0';
	  tptr++;
	}
	strcpy(extname, tptr); 
      }
    }
    else if ( cs_eq_siz( line, "EXTVER", 6 ))
    {
      if ( !found_extver && found_xtension )
      {
	afk_gen_read_keyline( astyle, line,
			      kname, CARD_SIZE,
			      verstr, CARD_SIZE,
			      comment, CARD_SIZE,
                              unit, CARD_SIZE );
	afk_key_parse_val( verstr, DT_INT4, &extver );

	found_extver = dmTRUE;

      }
    }
    else if ( cs_eq( line, "END" ) )
    {
      if ( found_xtension )
      {
	++nblocks;

	/* Define Block Name */
	/* + if HDUNAME is defined.. use it                          */
	/* + else if EXTNAME but not EXTVER.. use EXTNAME            */
	/* + else if EXTNAME and EXTVER.. HDUNAME = EXTNAME + EXTVER */
	/* + else use default name HDU#                              */

	if ( found_hduname )
	{	  
	  strcpy(kname, hduname );
	}
	else if ( found_extname && ! found_extver )
	{
	  strcpy(kname, extname );
	}
	else if ( found_extname && found_extver && (extver != 1))
	{
	  sprintf(kname, "%s%ld", extname, extver );
	}
	else
	{
	  sprintf( kname, "HDU%ld", nblocks );
	}

        // Code Change October 17, 2007
	// Strip off any trailing blanks in block name before storing

        while ( ((index = (strlen(kname) - 1)) >= 0 ) &&  (kname[index] == ' ') )
          kname[index] = '\0';

        // end code change

	afp_dataset_add_block( ds, nblocks, blockType, kname, block_pos );
      }

      /* setup for next loop */
      found_xtension = dmFALSE;
      found_hduname  = dmFALSE;
      found_extname  = dmFALSE;
      found_extver   = dmFALSE;
    }

    pos = afk_file_tell( afptr );

  } // end while loop


  if ( !nblocks && pos > 0 )
  {
    /* File has something, but no DTF header, set  */
    /* the block, it will be defined from the data */
    /* Set style element "has_header" to FALSE     */

    astyle->has_header = UT_FALSE;

    /* add block name */
    nblocks = 1;
    sprintf( hduname, ds->name, nblocks );
    afp_dataset_add_block( ds, nblocks, dmTABLE, hduname, ads->pos );
  }

   afk_file_seek( afptr, ads->pos, SEEK_SET );
 
  return status;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dtf_insert_col                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock      *block                      */
/*                    dmDescriptor *dd                         */
/*                    long          no                         */
/*                    void         *ikdd                       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_dtf_insert_col( dmBlock *block, dmDescriptor *dd,
                         long no, void *ikdd )
{
  dmDataset      *ds     = NULL;
  dmDescriptor   *col_dd = NULL;
  dmDescriptor   *cpt_dd = NULL;
  dmDescriptor   *us_dd  = NULL;
  afCol          *acol   = NULL;
  afBlock        *ablock = NULL;
  afVector       *avec   = NULL;
 
  long  moved_total = 0;
  long  moved       = 0;
  long  dim         = 0;
  long  col_dim     = 0;
  long  cptno       = 0;
  long  colno       = 0;
  long  ncols       = 0;
  long  kcolno      = 0; /* kernel column number */

  dmDataType  type   = DT_UNK;
  dmBool      append = dmFALSE;

  char        tfields[] = "TFIELDS";
  long        index = 3;

 
  ablock = (afBlock*)dmi_block_get_kb( block );
  ncols = dmi_block_get_ncols( block );
  append = ( no == ncols + 1 );
  dim = dmi_descriptor_get_vector_dim( dd );

  for ( colno = ncols; colno >= no; colno-- )
  {
    col_dd  = dmi_block_get_col( block, colno );
    col_dim = dmi_descriptor_get_vector_dim( col_dd );

    for ( cptno = 1; cptno <= col_dim; cptno++ )
    {
      cpt_dd = dmi_descriptor_get_cpt( dd, cptno );
      afp_dtf_move_col_keys( block, cpt_dd, dim );
    }
  }


  if ( append )
    kcolno = ablock->ncols+1;
  else
  {
    col_dd = dmi_block_get_col( block, no );
    cpt_dd = dmi_descriptor_get_cpt( col_dd, 1 );
    acol = (afCol*)dmi_descriptor_get_kdd( cpt_dd );
    kcolno = acol->no;
  }

  if ( dim > 1 )
    avec = afk_vector_alloc( dim );

  for ( cptno = 1; cptno <= dim; cptno++ )
  {
    cpt_dd = dmi_descriptor_get_cpt( dd, cptno );
    type = dmi_descriptor_get_datatype( cpt_dd );
    acol = (afCol*)dmi_descriptor_get_kdd( cpt_dd );

    if ( !acol )
    {
      if ( ikdd ) /* pass on the afCol */
      {
        if ( dim == 1 )
          acol = afk_col_alloc_copy( (afCol*)ikdd ); 
        else
          if ( dim > 1 )
            acol = afk_col_alloc_copy( ((afVector*)ikdd)->cpts[cptno-1] );
      }
      else
      {
        acol = afk_col_alloc();
        acol->name     = cs_dup( dmi_descriptor_get_name( cpt_dd ));
        acol->datatype = dmi_descriptor_get_datatype( cpt_dd );
        acol->length   = dmi_descriptor_get_array_size( cpt_dd );

        // Code Change  Bug 10784  November 8, 2007
        acol->width    = dmi_descriptor_get_length( cpt_dd );

	/* Code change October 28, 2008 */
	/* Bit type stored by DM as array of bytes, but as   */
	/* single bit string in ascii                        */
	if (acol->datatype == DT_BIT )
	  acol->length = 0;
      }
      dmi_descriptor_set_kdd( cpt_dd, (void*)acol );
    }

    us_dd = dmi_descriptor_get_unscaled( cpt_dd );

    if ( us_dd )
      dmi_descriptor_set_kdd( us_dd, (void*)acol );

    if ( dim > 1 )
      avec->cpts[cptno-1] = acol;

    afk_block_insert_col( ablock, acol, kcolno );
    afk_buffer_insert_col( ablock->abuf, acol, kcolno, &moved );
    moved_total += moved;
    afp_dtf_write_col_keys( block, cpt_dd );
  
    ++kcolno;
  }

  if ( dim > 1 ) 
  {
    dmi_descriptor_set_kdd( dd, avec );
  }

  if ( moved_total != 0 )
  {
    ds = dmi_block_get_ds( block );
    afp_dataset_notify_moved( ds, block, moved_total );
  }

  if ( dim > 1 )
    afp_dtf_write_col_keys( block, dd );

  /* Update TFIELDS keyword.. should already exist.                   */
  ncols   = ablock->ncols;
  afk_dtf_key_write( ablock, index, tfields, DT_INT4, &ncols, NULL, NULL );

}


/***************************************************************/
/*                                                             */
/* Function:   afp_dtf_insert_image_col                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock      *block                      */
/*                    dmDescriptor *dd                         */
/*                    long          no                         */
/*                    void         *ikdd                       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_dtf_insert_image_col( dmBlock *block, dmDescriptor *dd,
                         long no, void *ikdd )
{
  dmDescriptor   *us_dd  = NULL;
  afCol          *acol   = NULL;
  afBlock        *ablock = NULL;

  long  moved   = 0; 
  long  kcolno  = 1; /* kernel column number */
  char* bunit   = NULL;
  int   naxes   = dd->naxes;
  long  *axes   = dd->axes;

  bunit   = dmi_descriptor_get_unit( dd );
 
  ablock = (afBlock*)dmi_block_get_kb( block );

  acol = afk_col_alloc();
  acol->name     = cs_dup( dmi_descriptor_get_name( dd ));
  acol->datatype = dmi_descriptor_get_datatype( dd );
  acol->length   = dmi_descriptor_get_array_size( dd );

  dmi_descriptor_set_kdd( dd, (void*)acol );

  us_dd = dmi_descriptor_get_unscaled( dd );
  if ( us_dd )
    dmi_descriptor_set_kdd( us_dd, (void*)acol );

  afk_block_insert_col( ablock, acol, kcolno );
  afk_buffer_insert_col( ablock->abuf, acol, kcolno, &moved );

  afp_dtf_image_set( block, acol->datatype, bunit, axes, naxes );

}


/***************************************************************/
/*                                                             */
/* Function:   afp_dtf_write_col_keys                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock      *block                      */
/*                    dmDescriptor *dd                         */
/*                    long          no                         */
/*                    void         *ikdd                       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_dtf_write_col_keys( dmBlock *block, dmDescriptor *dd )
{
  char keyname[CARD_SIZE];
  char buf[CARD_SIZE];
  char tval[CARD_SIZE];

  char   *tptr = NULL;
  char   *disp = NULL;
  char   *name = NULL;
  char   *desc = NULL;
  char   *unit = NULL;
  char   *ptr  = NULL;
 
  double  one      = 1.0;
  double  tzero_ul = 2147483648.0;
  double  tzero_us = 32768.0;

  dmDataType    type = DT_UNK;
  dmuKeyList    keylist;
  dmDescriptor *cpt_dd = NULL;

  afBuffer   *abuf   = NULL;
  afCol      *acol   = NULL;
  afBlock    *ablock = NULL;
  afStyle    *astyle = NULL;
  afDataset  *ads    = NULL;
  long       *axes   = NULL;

  long  keyno  = 0;
  long  dim    = 0;
  long  size   = 0;
  long  length = 0;
  long  width  = 0;
  long  naxes  = 0;
  long  i      = 0;
  long  no     = 0;
  long  cptno  = 0;
  long  lval   = 0;
  long  arlen  = 0;

  dmBool fixed = UT_FALSE;


  dim = dmi_descriptor_get_vector_dim( dd );
  ablock = (afBlock*)dmi_block_get_kb( block );
  abuf = ablock->abuf;
  ads = ablock->ads;
  astyle = afk_dataset_get_style( ads );
  fixed = astyle->fixed;

  keylist = ablock->keylist;
  keyno = keylist->n + 1;

  name = dmi_descriptor_get_name( dd );
  tptr = &tval[0];

  if ( dim == 1 )
  {
    type  = dmi_descriptor_get_datatype( dd );
    disp  = dmi_descriptor_get_disp( dd );
    size  = dmi_descriptor_get_array_size( dd );
    length = dmi_descriptor_get_length( dd );
    desc  = dmi_descriptor_get_desc( dd );
    unit  = dmi_descriptor_get_unit( dd );
    acol  = (afCol*)dmi_descriptor_get_kdd( dd );
    no    = acol->no;
    arlen = acol->length;

    if ( arlen < 1 )
      arlen = 1;

    if ( fixed )
    {
      width = afk_buffer_get_col_maxwidth( abuf, no, arlen );
      if (type == DT_CHAR)
	length = width-1;  /* this holds the maximum string length */
    }
    else
      width = 0;


    if ( type == DT_UINT4 )
    {
      sprintf( keyname, "TZERO%ld", no );
      afk_dtf_key_write( ablock, keyno++, keyname, DT_REAL8, &tzero_ul, NULL,
		      "offset for unsigned integers" );
      sprintf( keyname, "TSCAL%ld", no );
      afk_dtf_key_write( ablock, keyno++, keyname, DT_REAL8, &one, NULL,
		      "data are not scaled" );
    }
    else
      if ( type == DT_UINT2 )
      {
        sprintf( keyname, "TZERO%ld", no );
        afk_dtf_key_write( ablock, keyno++, keyname, DT_REAL8, &tzero_us, NULL,
		      "offset for unsigned integers" );
   
        sprintf( keyname, "TSCAL%ld", no );
        afk_dtf_key_write( ablock, keyno++, keyname, DT_REAL8, &one, NULL,
		      "data are not scaled" );
      }

      sprintf( keyname, "TTYPE%ld", no );
      strcpy( tval, name );
      afk_dtf_key_write( ablock, keyno++, keyname, DT_CHAR, &tptr, NULL, desc );

      sprintf( keyname, "TFORM%ld", no );
      afp_compose_tform( type, size, length, width, UT_FALSE, UT_FALSE, tval );
      afk_dtf_key_write( ablock, keyno++, keyname, DT_CHAR, &tptr,
                         NULL, "data format of field." );

      if ( unit )
      {
        sprintf( keyname, "TUNIT%ld", no );
        strcpy( tval, unit );
        afk_dtf_key_write( ablock, keyno++, keyname, DT_CHAR, &tptr,
                           NULL, "physical unit of field." );
      }

      naxes = dmi_descriptor_get_array_dimensions( dd, &axes );

      if ( naxes > 1 )
      {
        sprintf( keyname, "TDIM%ld", no );
        ptr = tval;
        ptr += sprintf( ptr, "(%ld", axes[0] );

        for ( i = 1; i < naxes; i++ )
          ptr += sprintf( ptr, ",%ld", axes[i] );

        strcat( tval, ")" );
        afk_dtf_key_write( ablock, keyno++, keyname, DT_CHAR, &tptr, NULL,
                          "dimensions of field." );
      }
      if ( axes )
      {
	free( axes );
	axes = NULL;
      }

      if ( fixed )
      {
        sprintf( keyname, "TBCOL%ld", no );
        lval = afk_buffer_get_col_byteoffset( abuf, no );
        afk_dtf_key_write( ablock, keyno++, keyname, DT_INT4, &lval,
                           NULL, "byte position in row to start of column." );
      }

      if ( disp )
      {
        sprintf( keyname, "TDISP%ld", no );
        afk_dtf_key_write( ablock, keyno++, keyname, DT_CHAR, &disp, NULL,
                           "Display format" );
      }
      else
        if ( fixed )
        {
          buf[0] = '\0';
          afk_buffer_get_col_cfmt( abuf, no, buf );
          if ( buf[0] )
            afk_cfmt_to_tdisp( buf, ( width / arlen ) - 1, tval );
          else
            afp_compose_tdisp( type, ( width / arlen ) - 1, tval );

          sprintf( keyname, "TDISP%ld", no );
          disp = &tval[0];
          afk_dtf_key_write( ablock, keyno++, keyname, DT_CHAR, &disp,
                             NULL, "Display format" );
        }
  } /* end if (dim == 1) */
  else
  {
    no   = afk_dtf_next_mtype_index( ablock );
    desc = dmi_descriptor_get_desc( dd );

    sprintf( keyname, "MTYPE%ld", no );
    strcpy( tval, name );

    afk_dtf_key_write( ablock, keyno++, keyname, DT_CHAR, &tptr, NULL, desc );

    sprintf( keyname, "MFORM%ld", no );
    cpt_dd = dmi_descriptor_get_cpt( dd, 1 );
    strcpy( tval, dmi_descriptor_get_name( cpt_dd ));

    for ( cptno = 2; cptno <= dim; cptno++ )
    {
      cpt_dd = dmi_descriptor_get_cpt( dd, cptno );
      strcat( tval, "," );
      strcat( tval, dmi_descriptor_get_name( cpt_dd ));
    }

    afk_dtf_key_write( ablock, keyno++, keyname, DT_CHAR, &tptr, NULL, NULL);
  }
}


/***************************************************************/
/*                                                             */
/* Function:   afp_dtf_move_col_keys                           */
/*                                                             */
/* Description:                                                */
/*              Move the keywords for a column + or -.         */
/*              ie. if offset = 3, TTYPE3 becomes TTYPE6.      */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock      *block                      */
/*                    dmDescriptor *dd                         */
/*                    long          offset                     */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_dtf_move_col_keys( dmBlock *block, dmDescriptor *dd, long offset )
{
  char keyname[CARD_SIZE];
  char buf[CARD_SIZE];

  afBlock    *ablock = NULL;
  afCol      *acol   = NULL;
  dmuKey     *ukey   = NULL;
  void       *keyval = NULL;
  dmuKeyList  keylist;

  long dim   = 0;
  long colno = 0;
  long newno = 0;
 
  ablock = (afBlock*)dmi_block_get_kb( block );
  keylist = ablock->keylist;
  dim = dmi_descriptor_get_vector_dim( dd );

  if ( dim == 1 )
  {
    acol = (afCol*)dmi_descriptor_get_kdd( dd );
    colno = acol->no;
    newno = colno + offset;
    sprintf( keyname, "TZERO%ld", colno );
    ukey = dmu_keylist_read_value_c( keylist, keyname, buf, CARD_SIZE );

    if ( ukey )
    {
      sprintf( keyname, "TZERO%ld", newno );
      keyval = dmu_key_value( ukey );
      afk_dtf_key_write( ablock, ukey->index, keyname, ukey->datatype,
                         keyval, ukey->unit, ukey->desc );
      afk_key_delete( ablock, ukey );
    }

    sprintf( keyname, "TSCAL%ld", colno );
    ukey = dmu_keylist_read_value_c( keylist, keyname, buf, CARD_SIZE );

    if ( ukey )
    {
      sprintf( keyname, "TSCAL%ld", newno );
      keyval = dmu_key_value( ukey );
      afk_dtf_key_write( ablock, ukey->index, keyname, ukey->datatype,
                         keyval, ukey->unit, ukey->desc );
      afk_key_delete( ablock, ukey );
    }

    sprintf( keyname, "TTYPE%ld", colno );
    ukey = dmu_keylist_read_value_c( keylist, keyname, buf, CARD_SIZE );

    if ( ukey )
    {
      sprintf( keyname, "TTYPE%ld", newno );
      keyval = dmu_key_value( ukey );
      afk_dtf_key_write( ablock, ukey->index, keyname, ukey->datatype,
                         keyval, ukey->unit, ukey->desc );
      afk_key_delete( ablock, ukey );
    }

    sprintf( keyname, "TFORM%ld", colno );
    ukey = dmu_keylist_read_value_c( keylist, keyname, buf, CARD_SIZE );

    if ( ukey )
    {
      sprintf( keyname, "TFORM%ld", newno );
      keyval = dmu_key_value( ukey );
      afk_dtf_key_write( ablock, ukey->index, keyname, ukey->datatype,
                         keyval, ukey->unit, ukey->desc );   
      afk_key_delete( ablock, ukey );
    }

    sprintf( keyname, "TUNIT%ld", colno );
    ukey = dmu_keylist_read_value_c( keylist, keyname, buf, CARD_SIZE );

    if ( ukey )
    {
      sprintf( keyname, "TUNIT%ld", newno );
      keyval = dmu_key_value( ukey );
      afk_dtf_key_write( ablock, ukey->index, keyname, ukey->datatype, keyval,
                         ukey->unit, ukey->desc );   
      afk_key_delete( ablock, ukey );
    }

    sprintf( keyname, "TDIM%ld", colno );
    ukey = dmu_keylist_read_value_c( keylist, keyname, buf, CARD_SIZE );

    if ( ukey )
    {
      sprintf( keyname, "TDIM%ld", newno );
      keyval = dmu_key_value( ukey );
      afk_dtf_key_write( ablock, ukey->index, keyname, ukey->datatype, keyval,
                         ukey->unit, ukey->desc );   
      afk_key_delete( ablock, ukey );
    }
  }
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dtf_create_block                            */
/*                                                             */
/* Description:                                                */
/*              Create the DTF block in the file               */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock      *block                      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_dtf_create_block( dmBlock *block )
{
  char hduname[FITS_MAXLEN]    = "";
  char extname[FITS_MAXLEN]    = "";
  char xtension[FITS_MAXLEN]   = "";
  char flavorString[CARD_SIZE] = "";
  int  flavorStringLength      = 0;
  const char *pFlavorString    = NULL;

  char      *name   = NULL;
  char      *keyval = NULL;
  afBlock   *ablock = NULL;
  afFile    *afptr  = NULL;
  afDataset *ads    = NULL;
  afBuffer  *abuf   = NULL;
  afStyle   *astyle = NULL;

  long extver  = 0;
  long moved   = 0; /* for moving other blocks */
  long spos    = 0;
  long epos    = 0;
  long keyno   = 0;
  long zero    = 0;
  int  have_extver = 0;


  ablock = (afBlock*)dmi_block_get_kb( block );
  ads    = ablock->ads;
  astyle = afk_dataset_get_style( ads );
  name   = dmi_block_get_name( block );
  abuf   = afk_block_get_buffer( ablock );
  afptr  = afk_dataset_fptr( ads );

  pFlavorString = afk_style_get_flavorString(astyle);

  if ( pFlavorString )
    strcpy(flavorString, pFlavorString);
  strcat (flavorString, "\n" );
  flavorStringLength = strlen(flavorString);

  switch ( block->type)
  {
    case dmTABLE:
    default:
      strcpy(xtension,"TABLE");
      break;

    case dmIMAGE:
      strcpy(xtension, "IMAGE");
      break;
  }

  keyno = 0;

  if ( ablock->no == 1 )
  {
    afk_file_seek( afptr, ads->pos, SEEK_SET );
    spos = afk_file_tell( afptr );
    afk_file_seek( afptr, 0, SEEK_END );
    epos = afk_file_tell( afptr );

    if ( spos == epos )
    {
      /* if this is the first and only block in the file, need
       * to add the text-dtf identifier */

      afk_file_insert_data( afptr, flavorString );

      afk_buffer_flush( abuf, &moved );
      afk_file_seek( afptr, ads->pos+flavorStringLength, SEEK_SET );

      spos = afk_file_tell( afptr );
      ads->blocks_pos[0] = spos;
      afk_buffer_set_datastart( abuf, spos );
    }
  }
  else
  {
    /* pad the new block with a blank line */
    afp_comment_write( block, 0, NULL, "" );
    ++keyno;
  }

  /* Convert CXCDM block name to FITS keyword values */
  have_extver = ftp_block_name_parse( name, hduname, extname, &extver );

  keyval = xtension;
  afp_key_write( block, 0, "XTENSION", DT_CHAR, &keyval, 1, NULL, NULL );
  ++keyno;

  keyval = hduname;

  afp_key_write( block, 0, "HDUNAME", DT_CHAR, &keyval, 1, NULL, NULL );
  ++keyno;

  keyval = extname;
  afp_key_write( block, 0, "EXTNAME", DT_CHAR, &keyval, 1, NULL, NULL );
  ++keyno;

  if ( have_extver )
  {
    afp_key_write( block, 0, "EXTVER", DT_INT4, &extver, 1, NULL, NULL );
    ++keyno;
  }

  if ( block->type != dmIMAGE )
  {
    afp_key_write( block, 0, "TFIELDS", DT_INT4, &zero, 1, NULL, NULL );
    ++keyno;

    /* Enh 10847: write NAXIS1 keyword for DTF-FIXED format tables */
    if ( afk_style_get_fixed( astyle ) )
    {
      afp_key_write( block, 0, "NAXIS1", DT_INT4, &zero, 1, NULL, (char*)"Width of table in bytes" );
      ++keyno;
    }
  }

  afk_buffer_insert_keyline( abuf, "END\n\n", keyno+1 );
 
  return;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dtf_flush_col_keys                          */
/*                                                             */
/* Description:                                                */
/*              Currently, this is only used for TBCOL and     */
/*              TFORM changes in fixed width data tables.      */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock      *block                      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_dtf_flush_col_keys( dmBlock *block )
{
  dmDescriptor  *cpt_dd = NULL;
  dmDescriptor  *dd     = NULL;

  long  colno  = 0;
  long  dim    = 0;
  long  ncols  = 0;
  long  cptno  = 0;
 
  /* hrmmm... should we be using rawcols here instead of cols? */
  ncols = dmi_block_get_ncols( block );

  for ( colno = 1; colno <= ncols; colno++ )
  {
    dd = dmi_block_get_col( block, colno );
    dim = dmi_descriptor_get_element_dim( dd );

    for ( cptno = 1; cptno <= dim; cptno++ )
    {
      cpt_dd = dmi_descriptor_get_cpt( dd, cptno );
      afp_dtf_write_col_keys( block, cpt_dd );
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:   afp_dtf_image_set                               */
/*                                                             */
/* Description:                                                */
/*   (Re)set image descriptor fields defining the image        */
/*   datatype and shape.  Define/update relevant keywords.     */
/*                                                             */
/*  Input parameters:                                          */
/*    dmBlock       *block  - DM Block descriptor              */
/*  const dmDataType type   - data type                        */
/*  const char      *unit   - unit string                      */
/*  const long      *axes   - length of axis                   */
/*  const long       naxes  - number of axes                   */
/*                                                             */
/*  Output parameters:                                         */
/*                      none                                   */
/*                                                             */
/*  Returns:                                                   */
/*                      void                                   */
/*                                                             */
/***************************************************************/
void afp_dtf_image_set( dmBlock *block, const dmDataType type,
                        const char *unit, const long *axes, const long naxes )
{
  char keyname[CARD_SIZE];

  dmDescriptor *pImgDescriptor = NULL;
  dmBool        bIsUnsigned    = dmFALSE;
  dmBool        bIsScaled      = dmFALSE;

  afBlock    *ablock = NULL;
  afCol      *acol   = NULL;

  double bscale  = 1.0;
  double bzero   = 0.0;

  long   axlen   = 0;
  long   n_axes  = naxes;
  long   bitpix  = 32;

  int    ii      = 0;
 

  ablock = (afBlock*)dmi_block_get_kb( block );

  if ( !ablock )
    return;


  /* set datatype */
  if ( ablock->cols )
    acol   = ablock->cols[0];

  if ( acol )
    acol->datatype = type;

  /* Set image shape in kernel block */
  ablock->naxes = naxes;

  for ( ii = 0; ii < naxes; ii++ )
    ablock->axes[ii] = axes[ii];

  /* Determine bitpix, bscale, bzero from image descriptor datatype */
  bitpix = ftp_bitpix( type, &bscale, &bzero );

  if ( bscale != 1.0 || bzero != 0.0 )
    bIsUnsigned = dmTRUE;
 
  // Get necessary information from DM descriptor
  pImgDescriptor = dmi_block_get_col(block, 1);

  if ( pImgDescriptor )
  {
    /* Code Change  Bug 12880  September 7, 2011       */
    /* handling of BSCALE/BZERO as reserve keywords    */
    /* Extract scale info from image descriptor        */
    /* Only applies to non-unsigned type images        */

    bIsScaled = (dmi_descriptor_get_flag( pImgDescriptor ) & DM_FLAG_COORD );

    if ( bIsScaled && !bIsUnsigned  )
    {
      // This is a scaled image, get correct scaling values for bzero and bscale
      double crpix = 0.0;
      double crval = 0.0;

      dmi_coord_get_transform_d(pImgDescriptor, &crpix, &crval, &bscale, 1);
      bzero = crval - crpix * bscale;
    }

    // Get unit while here
    unit = dmi_descriptor_get_unit( pImgDescriptor );
  }


  /* Write image keys */

  // First BITPIX
  sprintf( keyname, "BITPIX");
  afp_key_write( block, 0, keyname, DT_INT4, &bitpix, 1, NULL, "number of bits per data pixel" );

  // Next NAXIS
  // FITS Standard Compliance  July 2010
  // If naxis=1 and *axes=0, output only keyword NAXIS=0

  if ( n_axes == 1 && *axes == 0 )
    n_axes = 0;

  sprintf( keyname, "NAXIS");
  afp_key_write( block, 0, keyname, DT_INT4, &n_axes, 1, NULL, "number of data axes" );

  for (ii = 0; ii < n_axes; ii++)
  {
    axlen = (int)axes[ii];

    sprintf( keyname, "NAXIS%d", (ii+1) );
    afp_key_write( block, 0, keyname, DT_INT4, &axlen, 1, NULL, "length of data axis" );
  }


  // NEXT BZERO, BSCALE and BUNIT
  // NOTE: We only write these for Unsigned types.  The Ascii kernel currently only
  //       writes the scaled image (Bug #10782) so we should not write scale keys yet.
  // if ( bIsUnsigned || bIsScaled )

  if ( bIsUnsigned )
  {
    sprintf( keyname, "BZERO");
    afp_key_write( block, 0, keyname, DT_REAL8, &bzero, 1, NULL, "Zero point ");
   
    sprintf( keyname, "BSCALE");
    afp_key_write( block, 0, keyname, DT_REAL8, &bscale, 1, NULL,"Scaling factor" );
  }

  if ( unit && !cs_is_blank( unit ))
  {
    sprintf( keyname, "BUNIT");
    afp_key_write( block, 0, keyname, DT_CHAR, &unit, 1, NULL, "Units after BSCALE applied" );
  }

  return;
}


/***************************************************************/
/* Function: afp_init_tnull                                    */
/*                                                             */
/* Description:                                                */
/*   Check TNULL key and set in descriptor                     */
/*                                                             */
/* Parameters:                                                 */
/*   dmukeylist    keylist  - List of DM keyword descriptors   */
/*   dmDescriptor *dd       - Column descriptor to update      */
/*   dmBool        is_image - type image or table              */
/*                                                             */
/* Returns:                                                    */
/*     nothing                                                 */
/*                                                             */
/***************************************************************/
void afp_init_tnull( dmuKeyList keylist, dmDescriptor *dd, 
                     dmBool is_image )
{
  dmuKey      *pKey  = NULL;
  dmDataType   dataType;               /* Column data type     */
  GenericType  keyValue;

  void  *pKeyValue = NULL;
  char   keyword[CARD_SIZE];            /* Keyword name         */
  long   lColNumber = 0;                /* Column number        */


  lColNumber = dmi_descriptor_get_no( dd );
  dataType   = dmi_descriptor_get_datatype( dd );

  pKeyValue = &keyValue;

  ftp_null_keyword_compose( dataType, lColNumber, keyword, is_image );
  pKey = dmu_keylist_read_typed_value( keylist, keyword, dataType, pKeyValue );

  if ( pKey )
    dmi_descriptor_set_null( dd, dataType, pKeyValue );

}
