/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/***************************************************************/
/*                                                             */
/*  afgen.c                                                    */
/*                                                             */
/*  Functions for ASCII kernel type files                      */
/*                                                             */
/*                                                             */
/***************************************************************/

#include "afkernelP.h"
#include "../ftkernel/ftkernelP.h"

/* from ftwcs.c */
enum special_offsets { FWCS_EQUI, FWCS_LONP, FWCS_LATP, FWCS_RADE };



/***************************************************************/
/*                                                             */
/*  Function:  afp_gen_find_transforms                         */
/*                                                             */
/*  Description:                                               */
/*    Hunt through a FITS HDU header looking for WCS keywords, */
/*    and setting up the scalar coord transform descriptors    */
/*    on each column, as needed.                               */
/*    Called at block open.                                    */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor  **cols                     */
/*                    long            ncols                    */
/*                    dmuKeyList      keylist                  */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:  nothing                                          */
/*                                                             */
/***************************************************************/
void afp_gen_find_transforms( dmDescriptor **cols, long ncols,
                              dmuKeyList keylist )
{
  dmDescriptor *dd = NULL;
  dmuKey       *akey = NULL;
  char         *key  = NULL;           /* Key name */
  char         *name = NULL;          /* Transform axis name */
  char         *ctype = NULL;         /* Value of CTYPE keyword */

  dmBlock      *block = NULL;
  dmBlockType  block_type;
  dmBool       is_image  = dmFALSE;
  dmBool       composite = dmFALSE;

  long   colno   = 0;
  long   axisno  = 0;
  long   axisno2 = 0;
  long   keyno   = 0;         /* Key number */   
  long   offset  = 0;         /* Offset  in keylist array */

  char **wcs_keys   = NULL;
  char   sub        = '\0';
  char  *ctype_name = NULL;
  char  *transform  = NULL;
  char  *pBlank     = NULL;
  char  *pRadesys   = NULL;
  char  *pRade      = NULL;

  double theta     = 0.0;
  double delt      = 0.0;
  double sval[3]   = { 0.0, 0.0, 0.0 };
  double pix       = 0.0;
  double val       = 0.0 ;
  int found[4]     = { 0, 0, 0, 0 };  /* Have we found EQUIn etc */

  // FITS Standard Compliance July 2010
  //  Added *CNA to cover TCNA and iCNA keywords

  char* ImageKeys[] = { "CTYPE", "CUNIT", "CRPIX", "CRVAL", "CDELT", "CROTA", "CNAME", "--------", "CD" };
  char* TableKeys[] = { "*CTYP", "*CUNI", "*CRPX", "*CRVL", "*CDLT", "*CROT", "*CNAM", "*CNA", "*C" };
  enum offsets { FWCS_TYPE, FWCS_UNIT, FWCS_PIX, FWCS_VAL, FWCS_DELT, FWCS_ROT, FWCS_NAME, FWCS_CNA, FWCS_CD };


  /* If no keylist, return */
  if ( !keylist ) 
     return;

  // Code Change  Bug 10799 November 14, 2007
  //   Null column pointer is a problem. If ncols == 0 or
  //   cols == NULL. return
  if ( !ncols || !cols )
    return;


  /* Determine if block is an image or table */
  block = dmi_descriptor_get_block( cols[0] );
  block_type = dmi_block_get_raw_type( block );
  is_image = ( block_type == dmIMAGE );

  /*  Pass 1: Look for CTYPE/TCTYP keywords and count them         */
  /*  to estimate the number of single-axis transforms to allocate */

  wcs_keys = ( block_type == dmTABLE ) ? &TableKeys[0] : &ImageKeys[0];


  /* Pass 2: Look for CTYPE/TCTYP keywords and assign them to transforms */

  for ( keyno = 1; keyno <= keylist->n; keyno++ )
  {
    offset = keyno - 1;
    akey = keylist->data[offset];
    key = akey->name;

    /* Check for TCTYPn and iCTYPn */
    if( ftp_coord_wcskey_match( key, wcs_keys[FWCS_TYPE] ))
    {
      /* Add a new axis */
      /* Initialize it */
      colno = ftp_parse_wcs_key( key, &axisno, &axisno2, &sub, is_image );
      ctype = dmu_key_value_c( akey );
      composite = ftp_ctype_parse( ctype, &ctype_name, &transform );

      // Code Change Bug 11488 August 13, 2010
      // Add 'ctype' as calling parameter to function afp_col_create_wcs()

      if ( colno > 0 )
        dd = afp_col_create_wcs( cols[colno-1], axisno, sub, ctype_name, 
                                 transform, composite, ctype ); 

      if ( ctype_name )
      {
        free( ctype_name );
        ctype_name = NULL;
      }

      if ( transform ) 
      {
        free( transform );
        transform = NULL;
      }
    }
  }

  /*  Pass 3: Look for other coord keywords   */

  for ( keyno = 1; keyno <= keylist->n; keyno++ )
  {
    offset = keyno-1;
    akey = keylist->data[offset];
    key = akey->name;

    if ( cs_eq( key, "EQUINOX" ) )
    {
      dmu_key_typed_value( akey, DT_REAL8, &sval[FWCS_EQUI] );    
      found[FWCS_EQUI] = -1;
    }
    else if ( cs_eq( key, "RADECSYS" ) || cs_eq( key, "RADESYS" ))
    {
      pRadesys = dmu_key_value_c( akey );
      found[FWCS_RADE] = -1;
    }
    else if ( cs_eq( key, "LONPOLE" ) || cs_eq( key, "LONGPOLE" ))
    {
      dmu_key_typed_value( akey, DT_REAL8, &sval[FWCS_LONP] );    
      found[FWCS_LONP] = -1;
    }
    else if ( cs_eq( key, "LATPOLE" ))
    {
      dmu_key_typed_value( akey, DT_REAL8, &sval[FWCS_LATP] );    
      found[FWCS_LATP] = -1;
    }
    else if ( ( dd = afp_coord_find_key( cols, ncols, key,
                         wcs_keys[FWCS_DELT] )) != NULL )
    {
      dmu_key_typed_value( akey, DT_REAL8, &delt );
      dmi_coord_set_delt( dd, &delt, 1 );
    }
    else if ( ( dd = afp_coord_find_key( cols, ncols, key, wcs_keys[FWCS_CD] )) != NULL )
    {
      dmu_key_typed_value( akey, DT_REAL8, &delt );
      dmi_coord_set_delt( dd, &delt, 1 );
    }
    else if ( ( dd = afp_coord_find_key( cols, ncols, key, wcs_keys[FWCS_PIX] )) != NULL )
    {
      dmu_key_typed_value( akey, DT_REAL8, &pix );
      dmi_coord_set_pix( dd, &pix, 1 );
    }
    else if ( ( dd = afp_coord_find_key( cols, ncols, key, wcs_keys[FWCS_UNIT] )) != NULL )
    {
      name = dmu_key_value_c( akey );
      dmi_descriptor_set_unit( dd, name );
    }
    else if ( ( dd = afp_coord_find_key( cols, ncols, key, wcs_keys[FWCS_VAL] )) != NULL )
    {
      dmu_key_typed_value( akey, DT_REAL8, &val );
      dmi_coord_set_val( dd, &val, 1 );
     }    
     else if ( ( dd = afp_coord_find_key( cols, ncols, key, wcs_keys[FWCS_NAME] )) != NULL )
     {
       /* This DM convention overrides dd->name but not ftcoord->ctype_name */
       name = dmu_key_value_c( akey );

       // FITS Standard Compliance  July 2010
       // CNAME value string can have spaces in it. 
       // Substitute space with underscore if present in string
       //  If string is completely blank, skip this section
       if ( !cs_is_blank(name) )
       {
         do
         {
           pBlank = strpbrk(name, " ");
      
           if ( pBlank )
             *pBlank = '_';

         } while (pBlank != NULL); 
       }

       dmi_descriptor_set_name( dd, name );
     }
     else if ( ( dd = afp_coord_find_key( cols, ncols, key, wcs_keys[FWCS_CNA] )) != NULL )
     {
       name = dmu_key_value_c( akey );

       // FITS Standarard Compliance  July 2010
       // TCNA and iCNA value strings can have spaces in them. 
       // Substitute space with underscore if present in string.

       if ( !cs_is_blank(name) )
       {
         do
         {
           pBlank = strpbrk(name, " ");
      
           if ( pBlank )
             *pBlank = '_';

         } while (pBlank != NULL); 
       }

       dmi_descriptor_set_name( dd, name );
     }
     else if ( ( dd = afp_coord_find_key( cols, ncols, key, wcs_keys[FWCS_ROT] )) != NULL )
     {
       dmu_key_typed_value( akey, DT_REAL8, &theta );
       dmi_coord_set_theta( dd, theta );     
     }
     else if ( ( dd= afp_coord_find_key( cols, ncols, key, "LONP" ) ) != NULL )
     {
       dmu_key_typed_value( akey, DT_REAL8, &val);
       dmi_coord_set_longpole( dd, val );
     }
     else if ( ( dd= afp_coord_find_key( cols, ncols, key, "LATP" ) ) != NULL )
     {
       dmu_key_typed_value( akey, DT_REAL8, &val);
       dmi_coord_set_latpole( dd, val );
     }
     else if ( ( dd= afp_coord_find_key( cols, ncols, key, "EQUI" ) ) != NULL )
     {  
       dmu_key_typed_value( akey, DT_REAL8, &val);
       dmi_coord_set_equinox( dd, val );
     }
     else if ( ( dd = afp_coord_find_key( cols, ncols, key, "RADE" ) ) != NULL )
     {
       pRade = dmu_key_value_c( akey );
       dmi_coord_set_radesys( dd, pRade );
     }
     else if ( ( dd= afp_coord_find_key( cols, ncols, key, "WCSTY" ) ) != NULL )
     {
       name = dmu_key_value_c( akey );
       ftp_coord_set_wname( dd, name );
     }
  }  /* End of loop on keys */

  afp_apply_coord_params( cols, ncols, sval, found, pRadesys );

  afp_update_physical_coords( cols, ncols );
}


/***************************************************************/
/*                                                             */
/*  Function:  afp_apply_coord_params                          */
/*                                                             */
/*  Description:                                               */
/*                                                             */   
/* Input parameters:                                           */
/*                    dmDescriptor  **cols                     */
/*                    long            ncols                    */
/*                    double         *vals                     */
/*                    int            *found                    */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:  nothing                                          */
/*                                                             */
/***************************************************************/
void afp_apply_coord_params( dmDescriptor** cols, long ncols, double* vals,
                             int* found, char *pRadesys )
{
  dmDescriptor *dd = NULL;
  dmDescriptor *phys_dd = NULL;
  dmDescriptor *col_dd  = NULL;
  dmCoordTransform *transform = NULL;

  ftCoord      *ftcoord = NULL;

  dmBool  bIsEQorEC = dmFALSE;
  dmBool  bIsICRS   = dmTRUE;

  char   *pRade = NULL;
  char   *pRadeDefault = "ICRS";
  char    cDash = '-';

  double  dValue = 0.0;
  double  B1950  = 1950.0;
  double  J2000  = 2000.0;

  long ii  = 0;
  long jj  = 0;
  long kk  = 0;
  long nCoords  = 0;
  long nGroups  = 0;
  long nwCoords = 0;


  /* Apply LONP LATP etc to ALL 2D WCS coordinate systems */

  for ( ii = 1; ii <= ncols; ii++ )
  {
    col_dd = cols[ii-1];
    nCoords = dmi_descriptor_get_no_coords( col_dd );

    for ( jj = 1; jj <= nCoords; jj++ )
    {
      dd = dmi_descriptor_get_coord_no( col_dd, jj );
      ftcoord = dmi_descriptor_get_kdd( dd );
   
      if ( (strlen( ftcoord->ctype_name ) >= 8) && 
           (ftcoord->ctype_name[4] == cDash) )
      {
        /* WCS key */

        bIsICRS   = dmTRUE; // ICRS is default
        bIsEQorEC = ftp_use_equinox( ftcoord->ctype_name );
        transform =  dmi_descriptor_get_aux( dd );

        // Apply global longpole if no column longpole value
        if ( (found[FWCS_LONP] < 0)  && (dValue = dmi_coord_get_longpole( transform )) == 0.0 )
          dmi_coord_set_longpole( dd, vals[FWCS_LONP] );

        // Apply global latpole if no column latpole value
        if ( (found[FWCS_LATP] < 0)  && (dValue = dmi_coord_get_latpole( transform )) == 0.0 )
          dmi_coord_set_latpole( dd, vals[FWCS_LATP] );

        // Apply global RADESYS if no column RADE# value
        if ( (pRade = dmi_coord_get_radesys( transform )) == NULL) 
	{
          // If global RADESYS, use it, otherwise default is "ICRS"

          if (found[FWCS_RADE] < 0 )
	  {
            dmi_coord_set_radesys( dd, pRadesys );
            pRade = pRadesys;
	  }
          else
	  {
            dmi_coord_set_radesys( dd, pRadeDefault );
            pRade = pRadeDefault;
          }
        }

        // EQUINOX applies only to equitorial & ecliptic coordinate systems

        if ( bIsEQorEC )
        {
          // Check first for global EQUINOX and no column EQUI#

          if ( (found[FWCS_EQUI] < 0) && (dValue = dmi_coord_get_equinox( transform)) == 0.0 )
	  {
            // Before assigning EQUINOX value, consider value of RADESYS
            // EQUINOX not used if RADESYS is ICRS
            // Pointer pRade gets value from call above. If NULL, then no global
            // RADESYS value was found and default frame is ICRS
            // Flag bIsICRS is set to dmTRUE at top  of loop

            if ( pRade ) 
              bIsICRS = (cs_uceq(pRade, "ICRS")) | (cs_uceq(pRade, "GAPPT"));
          
            // Apply global equinox if no column equinox value and RADESYS != ICRS

	    if ( !bIsICRS )
	    {
              // If global EQUINOX, use it, otherwise use default values

              if ( found[FWCS_EQUI] < 0 )
                dmi_coord_set_equinox( dd, vals[FWCS_EQUI] );
              else
	      {
                // use defaults

                if ( cs_uceq(pRade, "FK4") )
		  dmi_coord_set_equinox( dd, B1950 );
                else
                  dmi_coord_set_equinox( dd, J2000 );
              }
            }
          }
        }

        transform = NULL;
      }
    }

    nGroups = dmi_descriptor_get_no_groups( col_dd );

    for ( jj = 1; jj <= nGroups; jj++ )
    {
      phys_dd  = dmi_descriptor_get_group_no( col_dd, jj );
      nwCoords = dmi_descriptor_get_no_coords( phys_dd );

      for ( kk = 1; kk <= nwCoords; kk++ )
      {
        dd = dmi_descriptor_get_coord_no( phys_dd, kk );
        ftcoord = dmi_descriptor_get_kdd( dd );

        /* A FITS WCS CTYPE must be at least 8 chars and have a dash in byte number 5 */

        if ( (strlen( ftcoord->ctype_name ) >= 8) &&
             (ftcoord->ctype_name[4] == cDash) )
        {
          bIsICRS   = dmTRUE;
          bIsEQorEC = ftp_use_equinox( ftcoord->ctype_name );
          transform =  dmi_descriptor_get_aux( dd );

          /* WCS key */

          // Apply global longpole if no column longpole value
          if ( (found[FWCS_LONP] < 0 ) && (dValue = dmi_coord_get_longpole( transform )) == 0.0 )
            dmi_coord_set_longpole( dd, vals[FWCS_LONP] );

          // Apply global latpole if no column latpole value
          if ( (found[FWCS_LATP] < 0 ) && (dValue = dmi_coord_get_latpole( transform )) == 0.0 )
            dmi_coord_set_latpole( dd, vals[FWCS_LATP] );

          // Apply global RADESYS if no column RADE# value
          if ( (pRade = dmi_coord_get_radesys( transform )) == NULL)
	  {
            if ( found[FWCS_RADE] < 0 )
	    {
	      dmi_coord_set_radesys( dd, pRadesys );
              pRade = pRadesys;
            }
            else
	    {
              dmi_coord_set_radesys( dd, pRadeDefault );
              pRade = pRadeDefault;
            }
          }
 
          // EQUINOX applies only to equitorial & ecliptic coordinate systems
          if (bIsEQorEC)
	  {
            // Check first for global EQUINOX and no column EQUI#

            if ( (found[FWCS_EQUI] < 0 ) && (dValue = dmi_coord_get_equinox( transform)) == 0.0 )
	    {
              // Before assigning EQUINOX value, consider value of RADESYS
              // EQUINOX not used if RADESYS is ICRS
              // Pointer pRade gets value from call above. If NULL, then no global
              // RADESYS value was found and default frame is ICRS
              // Flag bIsICRS is set to dmTRUE at top  of loop

              if ( pRade ) 
                bIsICRS = (cs_uceq(pRade, "ICRS")) | (cs_uceq(pRade, "GAPPT"));

              // Apply global equinox if no column equinox value and RADESYS != ICRS

	      if ( !bIsICRS )
	      {
                if ( found[FWCS_EQUI] < 0 )
                  dmi_coord_set_equinox( dd, vals[FWCS_EQUI] );
                else
		{
                  if ( cs_uceq(pRade, "FK4") )
		    dmi_coord_set_equinox( dd, B1950 );
                  else
                    dmi_coord_set_equinox( dd, J2000 );
                }
              }
            }
          }
          transform = NULL;
        }
      }
    }
  }
}



/***************************************************************/
/*                                                             */
/*  Function:  afp_update_physical_coords                      */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/* Input parameters:                                           */
/*                    dmDescriptor  **cols                     */
/*                    long            ncols                    */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:  nothing                                          */
/*                                                             */
/***************************************************************/
void afp_update_physical_coords( dmDescriptor** cols, long ncols )
{
  long i = 0;
 
  for ( i = 0; i < ncols; i++ )
  {
    afp_col_update_physical( cols[i] );
  }
}



/***************************************************************/
/*                                                             */
/*  Function:  afp_coord_find_key                              */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/* Input parameters:                                           */
/*                    dmDescriptor  **cols                     */
/*                    long            ncols                    */
/*                    const char     *key                      */
/*                    const char     *key_prefix               */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:  dmDescriptor  dd                                 */
/*                                                             */
/***************************************************************/
dmDescriptor* afp_coord_find_key( dmDescriptor** cols, long ncols,
                                  const char* key, const char* key_prefix )
{
  long  axisno  = 0;
  long  axisno2 = 0;
  long  colno   = 0;
  char  sub     = ' ';

  dmBlock      *block = NULL;
  dmBlockType  block_type;
  dmBool       is_image = dmFALSE;

  dmDescriptor *dd     = NULL;
  dmDescriptor *col_dd = NULL;


  block = dmi_descriptor_get_block( cols[0] );
  block_type = dmi_block_get_raw_type( block );
  is_image = ( block_type == dmIMAGE );

  if ( ftp_coord_wcskey_match( key, key_prefix ))
  {
   colno = ftp_parse_wcs_key( key, &axisno, &axisno2, &sub, is_image );

   if ( colno == 0 && axisno == 0 )
     return dd;
   if ( colno > ncols ) 
     return dd;

   col_dd = cols[colno-1];

   if ( axisno2 == 0 || axisno2 == colno )
     dd = afp_col_find_coord( col_dd, axisno, sub );

   /* Make a new WCS with this axisno and sub */
   if ( !dd )
     dd = afp_col_create_default_wcs( col_dd, axisno, sub );
  }

  return dd;
}
