/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*  afkblock.c                                                   */
/*                                                               */
/*  Functions that work with ASCII kernel type files             */
/*                                                               */
/*                                                               */
/*****************************************************************/
#include "afkernelP.h"


/*****************************************************************/
/*                                                               */
/*  Function:  afk_block_alloc                                   */
/*                                                               */
/*  Description:                                                 */
/*           Allocate block structure and initialize elements    */ 
/*                                                               */
/*  Input parameters:                                            */
/*                    none                                       */
/*                                                               */
/*  Output parameters:                                           */
/*                                                               */
/*  Returns:                                                     */
/*             afBlock  *block                                   */
/*                                                               */
/*****************************************************************/
afBlock* afk_block_alloc( void )
{
  afBlock* ablock;

  ablock = (afBlock*)malloc( sizeof(afBlock) );

  if ( !ablock )
    return NULL;

  ablock->ads        = NULL;
  ablock->no         = 0;
  ablock->name       = NULL;
  ablock->block_type = dmUNKBLOCK;
  ablock->keylist    = NULL;
  ablock->keyno      = 0;
  ablock->ncols      = 0;
  ablock->nmax       = 0;
  ablock->cols       = NULL;
  ablock->hint       = NULL;
  ablock->max_dstyp  = 0;
  ablock->naxes      = 0;

  ablock->abuf  = afk_buffer_alloc();

  if (ablock->abuf)
    ablock->abuf->ablock = ablock;

  return ablock;
}



/*****************************************************************/
/*                                                               */
/*  Function:  afk_block_resize_cols                             */
/*                                                               */
/*  Description:                                                 */
/*               Make sure there are enough col pointers to      */
/*               hold at least specified number of columns       */
/*                                                               */
/*  Input parameters:                                            */
/*                   afBlock  *ablock  - block pointer           */
/*                   long      no      - number of columns       */
/*                                                               */
/*  Output parameters:                                           */
/*                                                               */
/*  Returns:                                                     */
/*            dmErrCode  status                                  */
/*                                                               */
/*****************************************************************/
dmErrCode afk_block_resize_cols( afBlock* ablock, long no )
{
  long oldmax;
  long incr = 10;

  if ( !ablock )
    return dmFAILURE;

  if ( no > ablock->nmax )
  {
    oldmax = ablock->nmax;
    ablock->nmax += incr;
    ablock->cols = (afCol**)realloc( ablock->cols, 
				     sizeof(afCol*) * ablock->nmax );
    if ( !ablock->cols )
      return dmFAILURE;
    memset( ablock->cols + oldmax, 0, incr * sizeof(afCol*) );
  }

  return dmSUCCESS;
}



/*****************************************************************/
/*                                                               */
/*  Function:  afk_block_free                                    */
/*                                                               */
/*  Description:                                                 */
/*               Free block                                      */
/*                                                               */
/*  Input parameters:                                            */
/*                   afBlock  *ablock  - block pointer           */
/*                                                               */
/*  Output parameters:                                           */
/*                                                               */
/*  Returns:                                                     */
/*           nothing                                             */
/*                                                               */
/*****************************************************************/
void afk_block_free( afBlock *ablock )
{
  if ( ablock )
  {
    /* free keylist structure */
    if ( ablock->keylist )
    {
      dmu_keylist_free( ablock->keylist );
      ablock->keylist = NULL;
    }

    /* free name */
    if ( ablock->name ) 
    {
      free(ablock->name);
      ablock->name = NULL;
    } 

    /* free buffer */
    afk_buffer_free( ablock->abuf );
    ablock->abuf = NULL;

    /* free hint structure */
    if (ablock->hint)
      afk_hint_free( ablock );

    /* free column data */
    /* Actual column descriptors freed at DM level */
    if (ablock->cols)
    {
      free (ablock->cols);
      ablock->cols  = NULL;
      ablock->ncols = 0;
    }

    /* free block */
    free(ablock);
  }
}


/*****************************************************************/
/*                                                               */
/*  Function:  afk_block_ncols                                   */
/*                                                               */
/*  Description:                                                 */
/*               Returns number of columns from block structure  */
/*                                                               */
/*  Input parameters:                                            */
/*                   afBlock  *ablock  - block pointer           */
/*                                                               */
/*  Output parameters:                                           */
/*                                                               */
/*  Returns:                                                     */
/*           long  ncols                                         */
/*                                                               */
/*****************************************************************/
long afk_block_ncols( const afBlock* ablock )
{
  if ( ablock )
     return ablock->ncols;
  return 0;
}



/*****************************************************************/
/*                                                               */
/*  Function:  afk_block_keylist                                 */
/*                                                               */
/*  Description:                                                 */
/*               Returns keylist from block structure            */
/*                                                               */
/*  Input parameters:                                            */
/*                   afBlock  *ablock  - block pointer           */
/*                                                               */
/*  Output parameters:                                           */
/*                                                               */
/*  Returns:                                                     */
/*           dmuKeyList  keylist                                 */
/*                                                               */
/*****************************************************************/
dmuKeyList afk_block_keylist( afBlock* ablock )
{
  if ( ablock )
    return ablock->keylist;
  else
    return NULL;
}


/*****************************************************************/
/*                                                               */
/*  Function:  afk_block_set_row                                 */
/*                                                               */
/*  Description:                                                 */
/*               set buffer pointer to row                       */
/*                                                               */
/*  Input parameters:                                            */
/*                   afBlock  *ablock  - block pointer           */
/*                   long      row     - row number              */
/*                                                               */
/*  Output parameters:                                           */
/*                                                               */
/*  Returns:                                                     */
/*           dmErrCode  status                                   */
/*                                                               */
/*****************************************************************/
dmErrCode afk_block_set_row( afBlock* ablock, long row )
{
  afBuffer* abuf = NULL;

  abuf = afk_block_get_buffer( ablock );
  afk_buffer_moveto_row( abuf, row );

  return dmSUCCESS;
}


/*****************************************************************/
/*                                                               */
/*  Function:  afk_block_get_buffer                              */
/*                                                               */
/*  Description:                                                 */
/*               Returns pointer to block buffer                 */
/*                                                               */
/*  Input parameters:                                            */
/*                   afBlock  *ablock  - block pointer           */
/*                                                               */
/*  Output parameters:                                           */
/*                                                               */
/*  Returns:                                                     */
/*           afBuffer  abuf                                      */
/*                                                               */
/*****************************************************************/
afBuffer* afk_block_get_buffer( afBlock* ablock )
{
  if ( ablock )
    return ablock->abuf;
  return NULL;
}


/*****************************************************************/
/*                                                               */
/*  Function:  afk_block_get_col                                 */
/*                                                               */
/*  Description:                                                 */
/*               Returns pointer to column                       */
/*                                                               */
/*  Input parameters:                                            */
/*                   afBlock  *ablock  - block pointer           */
/*                   long      colno   - column number           */
/*                                                               */
/*  Output parameters:                                           */
/*                                                               */
/*  Returns:                                                     */
/*           afCol  *cols                                        */
/*                                                               */
/*****************************************************************/
afCol* afk_block_get_col( const afBlock* ablock, long colno )
{
  if ( ablock && colno <= ablock->ncols && colno >= 1)
  {
    return ablock->cols[colno-1];
  }
  return NULL;
}



/*****************************************************************/
/*                                                               */
/*  Function:  afk_block_append_col                              */
/*                                                               */
/*  Description:                                                 */
/*                                                               */
/*  Input parameters:                                            */
/*                   afBlock  *ablock  - block pointer           */
/*                   afCol     acol    - pointer to column       */
/*                                                               */
/*  Output parameters:                                           */
/*                                                               */
/*  Returns:                                                     */
/*           nothing                                             */
/*                                                               */
/*****************************************************************/
void afk_block_append_col( afBlock* ablock, afCol* acol )
{
  /*afCol* tmpcol;*/
  if ( ablock && acol )
  {
    ablock->ncols++;
    afk_block_resize_cols( ablock, ablock->ncols );
    ablock->cols[ablock->ncols-1] = acol;
    acol->no = ablock->ncols;
#if 0
  if ( ablock->ncols > 1 )
  {
   tmpcol = ablock->cols[ablock->ncols-2];
   if ( tmpcol->length )
    acol->no = tmpcol->no + tmpcol->length;
   else
    acol->no = tmpcol->no + 1;
  }
#endif
  }
}


/*****************************************************************/
/*                                                               */
/*  Function:  afk_block_delete_col                              */
/*                                                               */
/*  Description:                                                 */
/*              Delete a column from the block.  Be sure to      */
/*              maintain the column's data position              */
/*                                                               */
/*  Input parameters:                                            */
/*                   afBlock  *ablock  - block pointer           */
/*                   long      colno   - column number           */
/*                                                               */
/*  Output parameters:                                           */
/*                                                               */
/*  Returns:                                                     */
/*           nothing                                             */
/*                                                               */
/*****************************************************************/
void afk_block_delete_col( afBlock* ablock, long colno )
{
  afCol *tmpcol = NULL;
  long   colndx = 0;
  long   subno  = 0;
  long   maxndx = 0;

  if ( ablock )
  {
    colndx = colno-1;

    if ( colno == ablock->ncols )
    {
      afk_col_free( ablock->cols[colndx] );
      ablock->cols[colndx] = NULL;
      ablock->ncols--;
    }
    else
    {
      tmpcol = ablock->cols[colndx];
#if 0
   if ( tmpcol->length )
    subno = tmpcol->length;
   else
    subno = 1;
#endif
      subno = 1;
      afk_col_free( tmpcol );
      maxndx = ablock->ncols-1;

      for ( ; colndx < maxndx; colndx++ )
      {
        tmpcol = ablock->cols[colndx+1];
        tmpcol->no -= subno; /* decrement data position */
        ablock->cols[colndx] = tmpcol;
      }
      ablock->cols[maxndx] = NULL;
      ablock->ncols--;
    }
  }
}



/*****************************************************************/
/*                                                               */
/*  Function:  afk_block_insert_col                              */
/*                                                               */
/*  Description:                                                 */
/*              Insert a column into a block.  Move all columns  */
/*              at place of insertion and higher up one position.*/
/*              Of column number > number of columns, the new    */
/*              column will be appended.                         */   
/*                                                               */
/*  Input parameters:                                            */
/*                   afBlock  *ablock  - block pointer           */
/*                   afCol    *acol    - pointer to column       */
/*                   long      colno   - column number           */
/*                                                               */
/*  Output parameters:                                           */
/*                                                               */
/*  Returns:                                                     */
/*           nothing                                             */
/*                                                               */
/*****************************************************************/
void afk_block_insert_col( afBlock* ablock, afCol* acol, long colno )
{
  long  ncols  = 0;
  long  colndx = 0;
  long  addno  = 0;
  long  i      = 0;
  afCol  *tmpcol = NULL;

  if ( !acol || !ablock )
    return;

  ncols = afk_block_ncols( ablock );

  if ( colno > ncols+1 )
  {
     colno = ncols+1;
  }

  if ( colno < 1 )
  {
    colno = 1;
  }

  if ( colno == ncols+1 )
  {
    afk_block_append_col( ablock, acol );
  }
  else
  {
    ncols++;
    ablock->ncols = ncols;
    afk_block_resize_cols( ablock, ablock->ncols );
    colndx = colno-1;
    /* increment the position of columns appropriately */

#if 0

    if ( acol->length )
     addno = acol->length;
    else
     addno = 1;
#endif

    addno = 1;

    for ( i = ncols-1; i >= colndx; i-- )
    {
      tmpcol = ablock->cols[i-1];
      tmpcol->no += addno;
      ablock->cols[i] = tmpcol;
    }

    /* insert the new col */
    ablock->cols[colndx] = acol;

    /* now set the new column's data position */
    acol->no = colno;
  }
}



/*****************************************************************/
/*                                                               */
/*  Function:  afk_block_read_col                                */
/*                                                               */
/*  Description:                                                 */
/*                                                               */
/*  Input parameters:                                            */
/*                   afBlock  *ablock  - block pointer           */
/*                   afCol    *acol    - pointer to column       */
/*                   dmDataType type   -                         */
/*                   long      rowl                              */
/*                   void     *value                             */
/*                   long      slen                              */
/*                   void*    *nullval                           */
/*                   long      imageOffset                       */
/*                                                               */
/*  Output parameters:                                           */
/*                                                               */
/*  Returns:                                                     */
/*          dmErrCode  status                                    */
/*                                                               */
/*****************************************************************/
dmErrCode afk_block_read_col( afBlock* ablock, afCol* acol,
			      dmDataType type,
			      long row1, long row2, void* value,
                              long slen, void* nullval, long imageOffset )
{
  dmErrCode status = dmSUCCESS;

  afBuffer *abuf = NULL;
  dmBlockType blockType  = ablock->block_type;
  long nvals = 1;


  if ( !acol || !ablock || !value ) 
  {
    dmu_err_set( DM_ERR_KERNEL, "Error: NULL passed to afk_block_read_col()." );
    return dmFAILURE;
  }


  // establish how many values to read

  if ( (blockType != dmIMAGE && acol->length > 0) ||
       (blockType == dmIMAGE && slen > 1) )
    nvals = acol->length;


  abuf = afk_block_get_buffer( ablock );
  afk_buffer_sync( abuf, ablock, dmFALSE );

  status = afk_buffer_read_col( abuf, row1, row2, acol->no, 
			        type, value, slen, nvals, nullval,
                                ablock->block_type, imageOffset );
  return status;
}


/*****************************************************************/
/*                                                               */
/*  Function:  afk_block_set_kds                                 */
/*                                                               */
/*  Description:                                                 */
/*               set dataset pointer in block structure          */
/*                                                               */
/*  Input parameters:                                            */
/*                   afBlock  *ablock  - block pointer           */
/*                   afDataset *ads    - dataset pointer         */
/*                                                               */
/*  Output parameters:                                           */
/*                                                               */
/*  Returns:                                                     */
/*           nothing                                             */
/*                                                               */
/*****************************************************************/
void afk_block_set_kds( afBlock* ablock, afDataset* ads )
{
  if ( ablock )
    ablock->ads = ads;
}


/*****************************************************************/
/*                                                               */
/*  Function:  afk_block_get_kds                                 */
/*                                                               */
/*  Description:                                                 */
/*               Returns pointer to dataset from block structure */
/*                                                               */
/*  Input parameters:                                            */
/*                   afBlock  *ablock  - block pointer           */
/*                                                               */
/*  Output parameters:                                           */
/*                                                               */
/*  Returns:                                                     */
/*           afDataset *ads                                      */
/*                                                               */
/*****************************************************************/
afDataset* afk_block_get_kds( const afBlock* ablock )
{
  if ( ablock )
    return ablock->ads;
  else
    return NULL;
}
