/*                                                                
**  Copyright (C) 2003-2012  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afkbuffer.c                                                  */
/*                                                               */
/*   Functions that create and work with buffers held in         */
/*   structures used with ASCII file type                        */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"
#include "utnan.h"
#include "dmi.h"

#define NULLSTR     "INDEF"
#define NULLSTR_LEN  5
#define CFMT_WIDTH   16
#define BYTE_BITS    8

#define NULL_DASH    0x01
#define NULL_NAN     0x02
#define NULL_INDEF   0x04
#define NULL_INF     0x08
#define NULL_NEGINF  0x10
#define NULL_NULLSTR 0x20


extern void dmu_map_bits( unsigned char* buf, long n, char* bits );

static afFile* afk_buffer_get_fptr( afBuffer* abuf );
static const char* afk_buffer_getline_ptr( afBuffer* abuf, long rowno );

static long afk_buffer_adjust( afBuffer* abuf, long startrow, long nrows );

static dmErrCode afk_buffer_parse_val( afBuffer* abuf, const char* strval, 
				       dmDataType type, void *value, long slen, void* nullval );

static void afk_buffer_init_rowbuf( afBuffer* abuf );
static void afk_buffer_init_parsebuf( afBuffer* abuf );
static void afk_buffer_init_writebuf( afBuffer* abuf );
static void afk_buffer_init_headerbuf( afBuffer* abuf );
static void afk_buffer_init_structure( afBuffer* abuf, const afBlock* ablock );

static dmErrCode afk_buffer_resize_rowbuf( afBuffer* abuf, long nrows );
static dmErrCode afk_buffer_resize_parsebuf( afBuffer* abuf, long llen );
static dmErrCode afk_buffer_resize_writebuf( afBuffer* abuf, long newsize, long nrows );
static dmErrCode afk_buffer_resize_headbuf( afBuffer* abuf, long newsize, long nrows );

static dmErrCode afk_buffer_set_parsebuf( afBuffer* abuf, const char* line, dmBool append );
static dmErrCode afk_buffer_process_parsebuf( afBuffer* abuf, long rowno );

static dmErrCode afk_buffer_load_row( afBuffer* abuf, const char* line, long rowno );
static dmErrCode afk_buffer_write_parsebuf( afBuffer* abuf, long rowno );
static dmErrCode afk_buffer_prepare_write( afBuffer* abuf, long rowno, long* moved );

static dmErrCode afk_buffer_flush_headbuf( afBuffer* abuf, long* moved );
static dmErrCode afk_buffer_flush_writebuf( afBuffer* abuf, long* moved );
static void afk_buffer_set_fixed_col_widths( afBuffer* abuf, long** pwidths, long** pmaxwidths, long* nwidths, afCol* acol );
static char* afk_buffer_write_vals( afBlock* ablock, long colno, void* values, void* nullval, char* disp );

static dmErrCode afk_buffer_align( afBuffer* );

/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_alloc                                */
/*                                                             */
/* Description:                                                */
/*             allocates ascii file buffer                     */
/*                                                             */
/*  Input parameters:                                          */
/*                    none                                     */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            afBuffer   abuf - pointer to new buffer          */
/*                                                             */
/***************************************************************/

afBuffer* afk_buffer_alloc( void )
{
  afBuffer *abuf = NULL;

  abuf = (afBuffer*)malloc( sizeof(afBuffer) );

  if ( !abuf )
    return NULL;
 
  abuf->ablock = NULL;

  /* parsing info */
  abuf->astyle = NULL;

  /* file structure */
  abuf->data_start = 0;
  abuf->ncols = 0;
  abuf->ndcols = 0;
  abuf->col_offsets = NULL;

  /* row buffer */
  abuf->rrows = NULL;
  abuf->rrow1 = 0;
  abuf->rrow2 = 0;
  abuf->nrrows = 0;
  abuf->maxrrows = AK_MAX_ROWBUF / sizeof(long);

  /* read buffer */
  abuf->pdata = NULL;
  abuf->parsebuf = NULL;
  abuf->pwidths  = NULL;
  abuf->pmaxwidths = NULL;
  abuf->prowwidth = 0;
  abuf->prow1 = 0;
  abuf->prow2 = 0;
  abuf->pmax = 0;
  abuf->psize = 0;
  abuf->pfmts = NULL;
  abuf->align = UT_FALSE;

  /* write buffer */
  abuf->writebuf = NULL;
  abuf->wrows = NULL;
  abuf->wmaxrows = 0;
  abuf->wrow1 = 0;
  abuf->wrow2 = 0;
  abuf->wsize = 0;
  abuf->wrsize = 0;

  /* header buffer */
  abuf->headbuf = NULL;
  abuf->hmax = 0;
  abuf->hsize = 0;
  abuf->hrsize = 0;
  abuf->hrows = NULL;
  abuf->hrow1 = 0;
  abuf->hrow2 = 0;
  abuf->hupdate = UT_FALSE;

  return abuf;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_free                                 */
/*                                                             */
/* Description:                                                */
/*             frees buffer and its internal structures        */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf  - pointer to buffer     */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_buffer_free( afBuffer* abuf )
{
  long colndx = 0;

  if ( abuf )
  {
    /* file structure */
    if ( abuf->col_offsets )
    {
      free( abuf->col_offsets );
      abuf->col_offsets = NULL;
    }
 
    /* row buffer */
    if ( abuf->rrows )
    {
      free( abuf->rrows );
      abuf->rrows = NULL;
    }

    /* read buffer */
    if ( abuf->pdata )
    {
      free( abuf->pdata );
      abuf->pdata = NULL;
    }

    if ( abuf->pwidths )
    {
      free( abuf->pwidths );
      abuf->pwidths = NULL;
    }
 
    if ( abuf->parsebuf )
    {
      free( abuf->parsebuf );
      abuf->parsebuf = NULL;
    }

    if ( abuf->pmaxwidths )
    {
      free( abuf->pmaxwidths );
      abuf->pmaxwidths = NULL;
    }

    if ( abuf->pfmts )
    {
      for ( colndx = 0; colndx < abuf->ndcols; colndx++ )
      {
        if ( abuf->pfmts[colndx] )
          free( abuf->pfmts[colndx] );
      }
      free( abuf->pfmts );
      abuf->pfmts = NULL;
    }

    /* write buffer */
    if ( abuf->writebuf )
    {
      free( abuf->writebuf );
      abuf->writebuf = NULL;
    }

    if ( abuf->wrows )
    {
      free( abuf->wrows );
      abuf->wrows = NULL;
    }

    if ( abuf->headbuf )
    {
      free( abuf->headbuf );
      abuf->headbuf = NULL;
    }

    if ( abuf->hrows )
    {
      free( abuf->hrows );
      abuf->hrows = NULL;
    }

    free( abuf );
  }
}



/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_set_datastart                        */
/*                                                             */
/* Description:                                                */
/*             sets element data_start and sets row markers to */
/*             zero                                            */  
/*                                                             */
/*  Input parameters:                                          */
/*                   afBuffer  *abuf        - pointer to buffer*/
/*                   long       data_start  - data start point */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*             nothing                                         */
/*                                                             */
/***************************************************************/
void afk_buffer_set_datastart( afBuffer* abuf, long data_start )
{
  if ( abuf )
  {
    abuf->data_start = data_start;
    /* invalidate the row buffer */
    abuf->rrow1 = 0;
    abuf->rrow2 = 0;
  }
}



/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_set_style                            */
/*                                                             */
/* Description:                                                */
/*             sets style pointer in buffer                    */
/*                                                             */
/*  Input parameters:                                          */
/*                   afBuffer  *abuf                           */
/*                   afStyle   *astyle                         */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_buffer_set_style( afBuffer* abuf, afStyle* astyle )
{
  abuf->astyle = astyle;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_get_datastart                        */
/*                                                             */
/* Description:                                                */
/*             returns data start element                      */
/*                                                             */
/*  Input parameters:                                          */
/*                    const afBuffer  *abuf - buffer pointer   */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            long   data_start    - data start location       */
/*                                                             */
/***************************************************************/
long afk_buffer_get_datastart( const afBuffer* abuf )
{
  if ( abuf )
    return abuf->data_start;
 
  return 0;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_sync                                 */
/*                                                             */
/* Description:                                                */
/*   Initialize the buffer for reading the block.              */
/*   This function allocates the buffers based on              */
/*   the structure of input block                              */
/*                                                             */
/*   Set reset to dmTRUE to force a reset of the               */
/*   buffer structure                                          */
/*                                                             */
/* Input parameters:                                           */
/*   afBuffer      *abuf   - buffer pointer                    */
/*   const afBlock *ablock - block pointer                     */
/*   dmBool         reset  - force reset flag                  */
/*                                                             */
/* Output parameters:                                          */
/*   none                                                      */
/*                                                             */
/* Returns:                                                    */
/*   dmErrCode  status                                         */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_sync( afBuffer* abuf, const afBlock* ablock, dmBool reset )
{
  dmErrCode status = dmSUCCESS;

  /* allocate the row buffer */
  if ( !abuf->rrows || reset )
  {
    afk_buffer_init_rowbuf( abuf );

    if ( !abuf->rrows )
    {
      dmu_err_set( DM_ERR_MEM, "TEXT Kernel: couldn't allocate row buffer." );
      return dmFAILURE;
    }
  }

  /* prepare the structural buffer */
  if ( !abuf->col_offsets || reset )
  {
    afk_buffer_init_structure( abuf, ablock );

    if ( !abuf->col_offsets )
    {
      dmu_err_set( DM_ERR_MEM, "TEXT Kernel: Couldn't allocate col offsets." );
      return dmFAILURE;
    }
  }
 
  /* prepare the read buffer */
  if ( !abuf->pdata || reset )
  {
    afk_buffer_init_parsebuf( abuf );

    if ( !abuf->pdata )
    {
      dmu_err_set( DM_ERR_MEM, "TEXT Kernel: Couldn't allocate parse buffer." );
      return dmFAILURE;
    }
  }

  /* don't initialize the write buffer until we try to write to the file */

  return status;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_init_writebuf                        */
/*                                                             */
/* Description:                                                */
/*             initialize buffer elements before writing       */
/*                                                             */
/*  Input parameters:                                          */
/*                   afBuffer  *abuf - buffer pointer          */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_buffer_init_writebuf( afBuffer* abuf )
{
  if ( abuf->writebuf )
    free( abuf->writebuf );

  if ( abuf->wrows )
    free( abuf->wrows );

  abuf->wmax = AK_MIN_WRITEBUF;
  abuf->writebuf = calloc( abuf->wmax, sizeof(char) );
  abuf->wrows = calloc( AK_WRITEBUF_NROWS, sizeof(char*) );
  abuf->wmaxrows = AK_WRITEBUF_NROWS;
  abuf->wrow1 = 0;
  abuf->wrow2 = 0;
  abuf->wsize = 0;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_flush                                */
/*                                                             */
/* Description:                                                */
/*             flush headbuf and writebuf elements             */
/*                                                             */
/*  Input parameters:                                          */
/*                   afBuffer  *abuf - buffer pointer          */
/*                                                             */
/*  Output parameters:                                         */
/*                    long  *moved  - number of bytes flushed  */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_flush( afBuffer* abuf, long* moved )
{
  dmErrCode  status   = dmSUCCESS;

  long diff1 = 0;
  long diff2 = 0;


  status = afk_buffer_flush_headbuf( abuf, &diff1 );

  if ( status != dmSUCCESS )
    return status;

  status = afk_buffer_flush_writebuf( abuf, &diff2 );

  if ( status != dmSUCCESS )
    return status;

  if ( moved )
    *moved = diff1 + diff2;

  return status;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_flush_writebuf                       */
/*                                                             */
/* Description:                                                */
/*             Force a flush of the write buffer.  Parameter   */
/*             moved holds the number of bytes added to the    */
/*             file or the number of bytes removed from the    */
/*             file after the flush                            */
/*                                                             */ 
/*  Input parameters:                                          */
/*                   afBuffer  *abuf - buffer pointer          */
/*                                                             */
/*  Output parameters:                                         */
/*                     long  *moved                            */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_flush_writebuf( afBuffer* abuf, long* moved )
{
  afFile     *afptr = NULL;
  const char *line  = NULL;

  long rowno      = 0;
  long nbytes     = 0;
  long diff       = 0;
  long max_offset = 0;
  long offset     = 0;
  long last_rrow  = 0;
  long woffset    = 0;
  long maxrow     = 0;
 
  /* only need to flush if there is pending data */
  if ( abuf->wsize > 0 )
  {
    diff = abuf->wsize - abuf->wrsize;
    afptr = afk_buffer_get_fptr( abuf );
    rowno = afk_buffer_moveto_row( abuf, abuf->wrow1 );

    /* need to back up,  ...this is kinda bad */
    if ( rowno == ( abuf->wrow1 - 1 ))
    {
      line = afk_file_getline_ptr( afptr );
    }
    else 
      if ( rowno != abuf->wrow1 )
      {
        sprintf( dm_errbuf, "TEXT Kernel: Couldn't move to row %ld", abuf->wrow1 );
        dmu_err_set( DM_ERR_KERNEL, dm_errbuf );

        if ( moved )
          *moved = 0;

       return dmFAILURE;
      }

    nbytes = afk_file_replace_data( afptr, abuf->writebuf, abuf->wsize, abuf->wrsize );

    if ( nbytes != abuf->wsize )
    {
      dmu_err_set( DM_ERR_KERNEL, "TEXT Kernel: Failed to flush write buffer." );

      if ( moved )
        *moved = 0;

      return dmFAILURE;
    }

    if ( diff != 0 && abuf->rrow2 > abuf->wrow1)
    {
      /* need to reset the row position pointers */
      maxrow     = abuf->rrow2 < abuf->wrow2 ? abuf->rrow2 : abuf->wrow2;
      max_offset = maxrow - abuf->rrow1;
      offset     = abuf->wrow1 - abuf->rrow1 + 1;
   
      if ( offset < 1 )
        offset = 1;

      last_rrow = abuf->rrows[offset-1];
      woffset = 1;

      for ( ; offset <= max_offset; offset++, woffset++ )
      {
	/* set this row start to (last row start + size of last row) */
        abuf->rrows[offset] = last_rrow + ( abuf->wrows[woffset] - abuf->wrows[woffset-1] );
        last_rrow = abuf->rrows[offset];
      }

      if ( maxrow < abuf->rrow2 )
      {
	/* remaining rows shift by the net difference */
        offset = max_offset+1;
        max_offset = abuf->rrow2 - abuf->rrow1;

        for ( ; offset <= max_offset; offset++ )
        {
          abuf->rrows[offset] += diff;
        }
      }
    }

    /* reset the write buffer */
    abuf->writebuf[0] = '\0';
    abuf->wrow1 = 0;
    abuf->wrow2 = 0;
    abuf->wsize = 0;
    abuf->wrsize = 0;
  }

  if ( moved )
    *moved = diff;

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_init_rowbuf                          */
/*                                                             */
/* Description:                                                */
/*              Initialize the row bufffer to get ready for    */
/*              reading and/or writing                         */
/*                                                             */
/*  Input parameters:                                          */
/*                   afBuffer  *abuf - buffer pointer          */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_buffer_init_rowbuf( afBuffer* abuf )
{
  long  nrows = AK_MIN_ROWBUF / sizeof(long);
  long *rrows = (long*)calloc( nrows, sizeof(long) );

  if ( abuf->rrows ) 
  {
    free( abuf->rrows );
  }

  if ( rrows )
  {
    abuf->rrows = rrows;
    abuf->nrrows = nrows;
  }
  else
  {
    abuf->rrows = NULL;
    abuf->nrrows = 0;
  }

  abuf->rrow1 = 0;
  abuf->rrow2 = 0; 
}



/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_init_parsebuf                        */
/*                                                             */
/* Description:                                                */
/*             Initialize the read buffer for reading          */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_buffer_init_parsebuf( afBuffer* abuf )
{
  long colndx = 0;

  if ( abuf->pdata )
    free( abuf->pdata );

  abuf->pdata = calloc( abuf->ndcols, sizeof(char*) );

  if ( !abuf->pwidths )
    abuf->pwidths = calloc( abuf->ndcols, sizeof(long) );

  if ( abuf->pfmts )
  {
    for ( colndx = 0; colndx < abuf->ndcols; colndx++ )
      free( abuf->pfmts[colndx] ); 
    free( abuf->pfmts );
  }

  abuf->pfmts = calloc( abuf->ndcols, sizeof(char*) );

  for ( colndx = 0; colndx < abuf->ndcols; colndx++ )
    abuf->pfmts[colndx] = calloc( CFMT_WIDTH, sizeof(char) );

  if ( !abuf->pmaxwidths )
    abuf->pmaxwidths = calloc( abuf->ndcols, sizeof(long) );

  abuf->prow1 = 0;
  abuf->prow2 = 0;
  /* leave the parsebuf, it gets alloc'ed by resize_parsebuf() */
}



/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_init_structure                       */
/*                                                             */
/* Description:                                                */
/*             Initialize the file structure information for   */
/*             reading and/or writing                          */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                    const afBlock *ablock  - block pointer   */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_buffer_init_structure( afBuffer* abuf, const afBlock* ablock )
{
  long  ncols   = 0;
  long  ndcols  = 0;
  long  colno   = 0;
  long  nwidths = 0;

  afCol   *acol   = NULL;
  afStyle *astyle = NULL;
  dmBool fixed;
 
  ncols = afk_block_ncols( ablock );
  abuf->ncols = ncols;

  if ( abuf->col_offsets )
    free( abuf->col_offsets );

  abuf->col_offsets = (long*)calloc( ncols, sizeof(long) );
  astyle = abuf->astyle;
  fixed = astyle->fixed;

  /* MCD 20091112 splitting single loop into components */

  ndcols = 0;
  for ( colno = 1; colno <= ncols; colno++ )
  {
    acol = afk_block_get_col( ablock, colno );
    abuf->col_offsets[colno-1] = ndcols;

    if ( acol->length > 0 )
      ndcols += acol->length;
    else
      ++ndcols;

  }
  abuf->ndcols = ndcols;

  /* allocate memory for widths arrays. */
  abuf->pmaxwidths = calloc( ndcols, sizeof(long) );
  if ( fixed )
  {
    abuf->pwidths = calloc( ndcols, sizeof(long) );
    nwidths = ncols;
  }

  /* If fixed format, need to set column widths. */
  if ( fixed )
  {
    for ( colno = 1; colno <= ncols; colno++ )
    {
      acol = afk_block_get_col( ablock, colno );
      afk_buffer_set_fixed_col_widths( abuf, &abuf->pwidths, &abuf->pmaxwidths, &nwidths, acol );
    }
  }

}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_set_parsebuf                         */
/*                                                             */
/* Description:                                                */
/*   Put provided line at parsebuf pointer                     */
/*                                                             */
/* Input parameters:                                           */
/*   afBuffer  *abuf - buffer pointer                          */
/*                                                             */
/* Output parameters:                                          */
/*   none                                                      */
/*                                                             */
/* Returns:                                                    */
/*   nothing                                                   */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_set_parsebuf( afBuffer* abuf, const char* line, dmBool append )
{
  dmErrCode status = dmSUCCESS;
  long      slen;              /* length of line               */
  long      totlen;            /* total length of new parsebuf */
  char*     bufptr;            /* ptr to copy location         */
  char*     cptr;              /* ptr within parsebuf          */

  if ( !abuf ){ return dmFAILURE; }

  /* line may already be there */
  if ( line == abuf->parsebuf ){ return dmSUCCESS;}

  /* check that parsebuf can handle the line */
  slen = strlen(line);

  /* determine space needed in parsebuf */
  totlen = slen;
  if ( append )
    totlen += abuf->psize;

  /* make sure we have it */
  if (totlen >= abuf->pmax)
  {
    status = afk_buffer_resize_parsebuf( abuf, totlen );
    if ( status == dmFAILURE ){ return dmFAILURE; }
  }

  /* set pointer to write location within parsebuf */
  bufptr = abuf->parsebuf;
  if ( append )
    bufptr += abuf->psize;

  /* Transfer to parsebuf */
  memcpy( bufptr, line, slen );

  /* make sure parsebuf is null terminated */
  cptr  = bufptr + slen;
  *cptr = '\0';

  abuf->psize = totlen;
  abuf->prsize = abuf->psize;

  return status;
}

/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_process_parsebuf                     */
/*                                                             */
/* Description:                                                */
/*   Parse the line at the parsebuf pointer, set column        */
/*   fields pdata[], pwidths[], etc.                           */
/*                                                             */
/* Input parameters:                                           */
/*   afBuffer  *abuf - buffer pointer                          */
/*   long      rowno - row number                              */
/*                                                             */
/* Output parameters:                                          */
/*   none                                                      */
/*                                                             */
/* Returns:                                                    */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_process_parsebuf( afBuffer* abuf, long rowno )
{
  const char  *iptr   = NULL;   /* ptr to walk through input line           */
  char        *optr   = NULL;   /* ptr to end of parsebuf                   */

  long ncols;                   /* # of buffer columns             */
  long colndx;                  /* current column                  */
  long width;                   /* width of current value string   */
  long rowlen;                  /* total length of row             */

  afStyle *style = NULL;
  dmBool   fixed = dmFALSE;

  dmErrCode status = dmSUCCESS;

  if ( !abuf ){ return dmFAILURE; }

  style = abuf->astyle;
  fixed = style->fixed;
  ncols = abuf->ndcols;

  iptr = abuf->parsebuf;
  colndx = 0;
  rowlen = 0;

  if ( fixed )
  {
    /* if the data is fixed width, it's much easier to read */
    /* NOTE: alignment is now maintained during write.      */

    /* clear the data pointers for this row  */
    /* column widths are fixed.. don't clear */
    memset( abuf->pdata, 0, sizeof(char*)*ncols );

    for ( colndx = 0; colndx < ncols; colndx++ )
    {
      abuf->pdata[colndx] = (char*)iptr;
      iptr   += abuf->pmaxwidths[colndx];
      rowlen += abuf->pmaxwidths[colndx];
    }
    rowlen--; /* terminator */
    
    /* The line may end before the full length of the definition...
       the parsebuf is only guaranteed for the size of the line  */
    /* Make sure it can hold the full defined length.            */
    if ( rowlen >= abuf->pmax )
    {
      status = afk_buffer_resize_parsebuf( abuf, rowlen );
      if ( status == dmFAILURE ){ return dmFAILURE; }
    }
  }
  else
  {
    /* clear the data pointers and the column widths for this row */
    memset( abuf->pdata, 0, sizeof(char*)*ncols );
    memset( abuf->pwidths, 0, sizeof(long)*ncols );

    /* set column start pointers and widths */
    while ( iptr && (colndx < ncols) )
    {
      width  = 0;
      iptr = afk_gen_find_next_val( style, iptr, &width );

      if ( iptr )
      {
	/* Have found a new data record */

	/* set pointer to record start. (n chars from beginning) */
	abuf->pdata[colndx] = (char*)iptr;

	/* set width of this value */
	abuf->pwidths[colndx] = width;
	
	colndx++;

	/* shift to next scan start position */
	iptr += width;

	/* skip 1 spacer unless at end. */
	if ( (*iptr != '\0') && (*iptr != '\n'))
	  iptr++;

	/* set length of row.. so far */
	rowlen = iptr - abuf->parsebuf;
      }
    } /* end while */

    /* there may be more chars in the original line to account for. */
    if ( iptr )
    {
      abuf->prsize += strlen( iptr ); 
      
      while ( (*iptr != '\0') && (*iptr != '\n') )
      {
	if ( !afk_style_is_sep( style, *iptr ))
	{
	  /* 'extra' has non-separator chars */
	  status = DM_ERR_PARSE;
	  break;
	}
	iptr++;
      }
    }

    /* make sure parsebuf is null terminated */
    optr  = abuf->parsebuf + rowlen;
    *optr = '\0';

    /* set pointers for any columns with missing data. */
    while ( colndx < ncols )
    {
      if ( abuf->pdata[colndx] )
      {
	/* last column data not assigned?? */
	width = optr - abuf->pdata[colndx];
      }
      else
      {
	/* Missing data.. will be set to NULL location.*/

	/* no data for column.. */
	width = 0;
	abuf->pdata[colndx] = optr;
      }

      /* Set column width */
      abuf->pwidths[colndx] = width;
      colndx++ ;
    }

  } /* end !fixed */

  abuf->psize = rowlen;
  abuf->prsize = abuf->psize; 
  abuf->prow1 = abuf->prow2 = rowno; 


  return status;

}

/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_resize_rowbuf                        */
/*                                                             */
/* Description:                                                */
/*             Resize the row buffer to hold at least nrows.   */
/*             If it already has enough space, do nothing      */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                    long       nrows - number of rows        */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_resize_rowbuf( afBuffer* abuf, long nrows )
{
  long  incr_rows = 0;
  long  new_nrows = 0;
 

  if ( nrows <= abuf->nrrows )
    return dmSUCCESS;

  // Code change - Bug 9222  June 8, 2006
  // removed limit on number of rows
  // code section that did limit test removed

  if ( nrows == abuf->maxrrows)
    new_nrows = nrows;
  else
  {
    // increment number of rows until enough room
    incr_rows = AK_ROWBUF_INCR / sizeof(long);
    new_nrows = abuf->nrrows;

    while( new_nrows < nrows )
    {
      new_nrows += incr_rows;
    }
  }

  // reallocate row buffer
  abuf->rrows = (long*)realloc( abuf->rrows, sizeof(long) * new_nrows );

  if ( !abuf->rrows ) 
  {
    sprintf( dm_errbuf, "TEXT Kernel: Failed to allocate row buffer for %ld rows.", new_nrows );
    dmu_err_set( DM_ERR_MEM, dm_errbuf );
    return dmFAILURE;
  }

  // initialize and set number of rows
  memset( abuf->rrows + abuf->nrrows, 0, 
	  sizeof(long) * ( new_nrows - abuf->nrrows ));
  abuf->nrrows = new_nrows;


  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_resize_parsebuf                      */
/*                                                             */
/* Description:                                                */
/*              Resize the read buffer to handle rows of size  */
/*              llen                                           */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                    long       llen - row size               */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_resize_parsebuf( afBuffer* abuf, long llen )
{
  char  *cptr   = NULL;
  long   diff   = 0;
  long   offset = 0;
  long   ndcols = 0;

  // Code Change Bug 11420 November 18, 2009
  //  Need to amend case where line length == buffer length
  //  Since afk_buffer_prepare_write() puts a null char at end of line,
  //  the buffer needs to be a size at least 1 more byte than line length

  if ( llen >= abuf->pmax )
  {
    while ( abuf->pmax <= llen )
    {
      abuf->pmax += AK_PARSEBUF_INCR;
    }

    cptr = abuf->parsebuf;
    abuf->parsebuf = realloc( abuf->parsebuf, abuf->pmax * sizeof(char) );

    if ( !abuf->parsebuf )
    {
      sprintf( dm_errbuf, "TEXT Kernel: The row length (%ld) is too long for the parse buffer.", llen );
      dmu_err_set( DM_ERR_MEM, dm_errbuf );
      abuf->pmax = 0;
      return dmFAILURE;
    }
    else
    {
      if ( abuf->pdata && abuf->pdata[0] )
      {
        diff = abuf->parsebuf - cptr;
        ndcols = abuf->ndcols;

        for( offset = 0; offset < ndcols; offset++ )
        {
          abuf->pdata[offset] += diff;
        }
      }
    }
  }

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_read_col                             */
/*                                                             */
/* Description:                                                */
/*             Read column                                     */
/*             slen is string length in case the values are    */
/*             strings.                                        */
/*             nvals is how many array values to read          */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                    long       row1                          */
/*                    long       row2                          */
/*                    long       col                           */
/*                    dmDataType  type                         */
/*                    void      *values                        */
/*                    long       slen                          */
/*                    long       nvals                         */
/*                    void      *nullval                       */
/*                    dmBlockType blockType                    */
/*                    long       imageOffset                   */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_read_col( afBuffer* abuf, long row1, long row2,
                               long col, dmDataType type, void* values,
                               long slen, long nvals, void* nullval,
                               dmBlockType blockType, long imageOffset )
{
  char *strval = NULL;
  char *optr   = (char*)values;

  long nrows_left = row2 - row1 + 1;
  long nrows_read = 0; /* just set to one to start the while loop */
  long col_offset = 0;
  long ii         = 0;
  long offset     = 0;
  long nbytes     = 0;
  long minLength  = 0; /* min of width of field and the requested string length */
  long rowno      = 0;

  dmErrCode status = dmSUCCESS;


  if ( type != DT_CHAR )
    nbytes = ar_mix_type_size( type );
  else
    nbytes = 1;

  // Code Change Bug 12078 December 7, 2009
  //  Make correct calculation for type DT_BIT

  if ( blockType != dmIMAGE && slen > 0 )
  {
    if ( type != DT_BIT )
      nbytes *= slen;
    else
      nbytes = (slen-1)/BYTE_BITS +1;
  }

  if ( col < 1 || col > abuf->ncols )
  {
    sprintf( dm_errbuf, "TEXT Kernel: Bad column number %ld.", col );
    dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
    return dmFAILURE;
  }


  col_offset = abuf->col_offsets[col-1];
  rowno = row1;

  while( nrows_left > 0 )
  {
    /* nrows_read is always 1 right now so we don't have to loop over it */
    /* loads row into abuf, sets pdata[] and pwidths[] array for this line */

    /* MCD TEMP:  ALERT!!! This will NOT read the full image any more   */
    /* be sure we have everything covered!!!!                           */
    nrows_read = afk_buffer_adjust( abuf, rowno, nrows_left );
  
    if ( nrows_read < 1 )
    {
      status = dmFAILURE;
      break;
    }


    // Code Change - Bug 9023 - September 13, 2006
    //   Calculate buffer offset using input parameter imageOffset
    //   instead of loop counter, ii.
    //   Use pwidths to determine min byte length if slen < 2
 
    // initialize offset into data buffer
    offset = col_offset + imageOffset;

    for ( ii = 0 ; ii < nvals && status == dmSUCCESS; ii++ )
    {
      /* get copy of value for parsing                            */
      /* Need to isolate the single token for it to work properly */

      minLength = abuf->pwidths[offset];

      if ( minLength )
	strval = cs_dup_siz(abuf->pdata[offset], minLength);
      else
	strval = cs_dup_siz("", 1); /* emptry string */

      // adjust minLength if necessary
      if ( blockType != dmIMAGE && slen > 0 )
        minLength = abuf->pwidths[offset] < slen ? abuf->pwidths[offset] : slen;

      status = afk_buffer_parse_val( abuf, (const char*)strval, type, optr, minLength, nullval );

      free( strval );

      optr += nbytes;
      ++offset;

    } // end for loop


    nrows_left -= nrows_read;
    rowno      += nrows_read;

  } // end while loop



  // Check status before returning.  If problem, set error

  if (status == dmFAILURE)
  {
      sprintf( dm_errbuf, "TEXT Kernel: Failed to load row %ld into the buffer.", row1 );
      dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
  }


  return status;
}


/***************************************************************/
/*                                                             */
/*  Function:   afk_buffer_parse_val                           */
/*                                                             */
/*  Description:                                               */
/*     Parse values.  Translates string rep. of values to      */
/*     their datatype.                                         */
/*                                                             */
/*  Input parameters:                                          */
/*     afBuffer   *abuf    - buffer pointer                    */
/*     const char *strval  - value as string                   */
/*     dmDataType  type    - datatype of value                 */
/*     long        slen    - length of next token              */
/*     void       *nullval - null value                        */
/*                                                             */
/*  Output parameters:                                         */
/*     void       *value   - value in correct datatype         */
/*                                                             */
/*  Returns:                                                   */
/*     dmErrCode  status                                       */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_parse_val( afBuffer* abuf, const char* strval, 
				dmDataType type, void *value,
                                long slen, void* nullval )
{
  char       *pToken  = NULL;
  afStyle    *pakStyle = NULL;
  char       *pDefinedNullStr = NULL;

  long        NullFlag = 0;
  long        length = 0;

  unsigned char mask[8]  = { 1, 2, 4, 8, 16, 32, 64, 128 };


  pToken   =  (char*) strval;

  if ( ! pToken )
    return dmFAILURE;

  pakStyle =  abuf->astyle;

  // Change July, 2008  Bug 10848
  pDefinedNullStr = pakStyle->nullstr;

  // end change

  // Code added for Bug 9343  Feb 22, 2006 
  // Check for Nan equivalent strings which are not 
  // converted automatically by function atof()

  // Code Change  July 9, 2008  Bug 10848
  // Check all types now for defined and default null values
  // If found, set flag

  if ( pDefinedNullStr && cs_uceq(pToken, pDefinedNullStr))
  {
    NullFlag = NULL_NULLSTR;
  }
  else
    if ( type != DT_CHAR)
    {
      if ( cs_uceq(pToken, NULLSTR ) )
      {
         NullFlag = NULL_INDEF;
      }  
      else
        if ( cs_eq(pToken, "-" ) )
        {
          NullFlag = NULL_DASH;
        }
        else
          if ( cs_uceq(pToken, "INF" ) || cs_uceq(pToken, "+Inf" ) )
          {
            NullFlag = NULL_INF;
          }
          else
            if ( cs_uceq(pToken, "-Inf" ))
            {
              NullFlag = NULL_NEGINF;
            }
    } // end if ( type != DT_CHAR)


  // Convert value to select type
  switch( type )
  {
    case DT_INT2:
      if ( !NullFlag && *pToken != '\0' )
        *(short*)value = (short)atol( pToken );
      else if ( nullval )
	*(short*)value = *(short*)nullval;
      else
        *(short*)value = 0;
      break;

    case DT_INT4:
      if (  !NullFlag && *pToken != '\0' )
        *(long*)value = atol( pToken );
      else if ( nullval )
	*(long*)value = *(long*)nullval;
      else
        *(long*)value = 0;
      break;

    case DT_UINT4:
      if (  !NullFlag && *pToken != '\0' )
        *(unsigned long*)value = (unsigned long)atof( pToken );
      else
        *(unsigned long*)value = 0;
      break;

    case DT_UINT2:
      if (  !NullFlag && *pToken != '\0' )
        *(unsigned short*)value = (unsigned short)atol( pToken );
      else
        *(unsigned short*)value = 0;
      break;

    case DT_REAL4:
      // Code Change added for Bug 9343 - September 26, 2007
      // Problems can occur because of fixed length values having 
      // leading or trailing blank or separator character.
      // Remove them for Real values

      afk_strip_value( pToken, pakStyle ); 

      if ( *pToken != '\0'  )
      {
        if ( !NullFlag )
        {
	  /* use function atof() to get default string value */
	  *(float*)value = (float)atof( pToken );
        }
        else
	{
          /* Handle special cases */
	  if ( NullFlag & NULL_NULLSTR)
	  {
            *(float*)value = ut_nan_create_r( UT_NAN_R_CXC);
	  }
          else if ( NullFlag & NULL_INDEF  )
	  {
            *(float*)value = ut_nan_create_r( UT_NAN_R_CXC);
	  }
          else if ( NullFlag & NULL_DASH )
	  {
	    *(float*)value = ut_nan_create_r( UT_NAN_R_CXC);
	  }
	  else if ( NullFlag & NULL_INF )
	  {
	    /* convert "Inf" or "+Inf" to Inf value */
	    *(float*)value = ut_nan_create_r( UT_INF_R);
	  }
	  else if ( NullFlag & NULL_NEGINF )
	  {
	    /* convert "-Inf" to neg. Inf value */
	    *(float*)value = ut_nan_create_r( UT_NINF_R);
	  }
        }
      }
      else
        *(float*)value = ut_nan_create_r( UT_NAN_R_CXC );

      break;

    case DT_REAL8:
      // Code Change added for Bug 9343 - September 26, 2007
      // Problems can occur because of fixed length values having 
      // leading or trailing blank or separator character.
      // Remove them for Real values

      afk_strip_value( pToken, pakStyle ); 

      if ( *pToken != '\0' )
      {
        if ( !NullFlag )
	{
	  /* use function atof() to get default string value */
	  *(double*)value = (double)atof( pToken );
	}
        else
	{
          /* Handle special cases */
          if ( NullFlag & NULL_NULLSTR)
          {
            *(double*)value = ut_nan_create_d( UT_NAN_D_CXC);
          }
          else if ( NullFlag  &NULL_INDEF )
  	  {
            *(double*)value = ut_nan_create_d( UT_NAN_D_CXC);
	  }
          else if ( NullFlag & NULL_DASH )
	  {
            /* catch the single dash and substitute a Nan value for it */
	    *(double*)value = ut_nan_create_d( UT_NAN_D_CXC);
	  }
	  else if ( NullFlag & NULL_INF )
	  {
	    /* convert "Inf" or "+Inf" to Inf value */
	    *(double*)value = ut_nan_create_d( UT_INF_D);
	  }
	  else if ( NullFlag & NULL_NEGINF )
	  {
	    /* convert "-Inf" to neg. Inf value */
	    *(double*)value = ut_nan_create_d( UT_NINF_D);
	  }
        } 
      }
      else
        *(double*)value = ut_nan_create_d( UT_NAN_D_CXC );

      break;

    case DT_CHAR:
      *(char *)value = '\0';

      if ( NullFlag )
        break;          // empty string or nullstr value

      if ( pToken )
      {
        /* Code Change added for Bug 10844 - March 20, 2008             */
	/* Code change added for Bug 10866 - May 29, 2008               */
	/* Fixed format:                                                */
	/*   - quotes and leading whitespace are significant...         */
	/*   - remove only trailing whitespace                          */
	/* otherwise:                                                   */
        /*   - Strip leading and trailing whitespace, and quotes        */
	/*                                                              */

	if ( pakStyle->fixed )
	{
	  /* could be done with afk_strip_value if control brought up.  */

	  char* white = pakStyle->white;       /* white chars to remove */ 
	  long  len   = strlen( pToken );
	  char* cptr  = pToken+(len-1);

	  while ( (cptr >= pToken) && ((strchr(white, *cptr ) != NULL )) )
	  {
	      *cptr = '\0';
	      --cptr;
	  }
	}
	else
	{
	  afk_strip_value( pToken, pakStyle );
	}

        if ( nullval && cs_eq(pToken, (*(char**)nullval)) )
          break;

	/* check size, should be min of column size or value length. */
	if ( (length = strlen(pToken)) > slen )
          length = slen;

        if (length > AK_COLVAL_LEN)
          length = AK_COLVAL_LEN;

        memcpy( (char*)value, pToken, length );
        ((char*)value)[length] = '\0';
      }
      break;

    case DT_BIT:
      /* bit pattern string gets translated to array of bytes */
      if ( *pToken != '\0' && !NullFlag )
      {
	long nbytes = (slen - 1)/BYTE_BITS + 1;
	long offset = 0; 
	long bitno  = 0;
	long byteno = 0; 
	long ii     = 0;

	memset( (unsigned char*)value, 0, nbytes );

	for ( bitno = 1; bitno <= slen; bitno++)
	{
	  offset = (BYTE_BITS - 1) - ((bitno - 1) % BYTE_BITS);
	  byteno = (bitno - 1 )/ BYTE_BITS;

	  if ( pToken[ii] == '1')
            ((unsigned char*)value)[byteno] += mask[offset];
	  ++ii;
	}
      }
      else
      {
        *(unsigned char*)value = 0;
      }
      break;

    // Case added May 4, 2007
    // Code Change  Bug 12078  December 2, 2009 
    case DT_LOGICAL:
      if ( *pToken == 'T' || *pToken == 't' )
        *(dmBool*)value = 1;
      else
        *(dmBool*)value = 0; 
      break;

    // Case added May 4, 2007
    case DT_UINT1:
      if ( !NullFlag)
        *(unsigned char*)value = (unsigned char)atoi( pToken );
      else
        *(unsigned char*)value = 0;
      break;

    case DT_VOID:
    case DT_PTR:
    case DT_UNK:
    default:
      dmu_err_set( DM_ERR_KERNEL, "TEXT Kernel: Failed to parse value." );
      return dmFAILURE;
   }

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/*  Function:   afk_buffer_adjust                              */
/*                                                             */
/*  Description:                                               */
/*    Returns the number of rows actually read.                */
/*                                                             */
/*  Input parameters:                                          */
/*    afBuffer   *abuf    - buffer pointer                     */
/*    long        frow    - first (only) row to read           */
/*    long        nrows   - dummy                              */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*    1 or 0: number of rows read.                             */
/*                                                             */
/***************************************************************/
long afk_buffer_adjust( afBuffer* abuf, long frow, long nrows )
{
  const char *line  = NULL;
  dmErrCode status = dmSUCCESS;
  
  (void) nrows;

  line = afk_buffer_getline_ptr( abuf, frow );
  if ( !line )
    return 0;

  /* load a line into the parse buffer */
  status = afk_buffer_load_row( abuf, line, frow );
  if ( status == dmFAILURE )
    return 0;


  return 1;
}

/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_load_row                             */
/*                                                             */
/* Description:                                                */
/*   Load and parse the line into the parse buffer             */
/*                                                             */
/* Input parameters:                                           */
/*   afBuffer   *abuf   - buffer pointer                       */
/*   const char *line   - string with data values              */
/*   long        rowno  - current row number                   */
/*                                                             */
/* Output parameters:                                          */
/*   none                                                      */
/*                                                             */
/* Returns:                                                    */
/*   dmErrCode  status                                         */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_load_row( afBuffer* abuf, const char* line, long rowno )
{
  dmErrCode status = dmSUCCESS;

  if ( !abuf ){ return dmFAILURE; }

  /* Set input line to parsebuf */
  status = afk_buffer_set_parsebuf( abuf, line, dmFALSE );
  if ( status == dmFAILURE ){ return dmFAILURE; }

  /* Process the line.. providing access to values */
  status = afk_buffer_process_parsebuf( abuf, rowno );

  return status;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_moveto_row                           */
/*                                                             */
/* Description:                                                */
/*              Move to a new row.  If we have not yet visited */
/*              the rows along the way, record their positions */
/*              in the row buffer.                             */
/*              Returns the final row position of the buffer.  */
/*              This is  in case we try to scan past the end of*/
/*              the file                                       */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                    long       row  - row number             */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            long  row_position                               */
/*                                                             */
/***************************************************************/
long afk_buffer_moveto_row( afBuffer* abuf, long row )
{
  const char *line    = NULL;
  long    offset      = 0;
  long    last_row    = 0; /* the row the buffer ends up pointing at */
  long    pos         = 0;


  // Invalid row request returns first row
  if ( row < 1 )
    row = 1;

  // Get file pointer
  afFile *afptr = afk_buffer_get_fptr( abuf );


  // Code change - Bug 9222  June 8, 2006
  // Is the requested row position currently held in the buffer ?
  // It is if it is a row number between first row and last row
  // If the requested row is less than the first row in the buffer,
  // need to reset file position to the beginning.
  if ( row < abuf->rrow1 )
  {
    // Rewind the row buffer.  Same as initial setting. 

    abuf->rrow1 = 1;
    abuf->rrow2 = 0;
  }

 
  // Calculate the offset to the requested position
  offset = row - abuf->rrow1;

  // Mark the new last row number
  last_row = row;

  // Is the row beyond what is already being held in list??
  if ( row > abuf->rrow2 )
  {
    // Yes,  but is it also beyond the last index of the current row array?
    if ( offset >= abuf->nrrows )
    {
      // Yes, but is it less than the max rows allowed?
      if (offset < abuf->maxrrows)
      {
          // Yes, resize buffer slightly
          afk_buffer_resize_rowbuf(abuf, offset+1);
      }
      else
      {
        // No, resize to the max allowed space
        // unless the buffer is already at max

        if ( ! (abuf->nrrows >= abuf->maxrrows ) )
          afk_buffer_resize_rowbuf(abuf, abuf->maxrrows);
      }
    }


    // if this is first time through, need to initialize
    if ( abuf->rrow2 == 0 )
    {
      abuf->rrow1     = 1;
      abuf->rrows[0]  = abuf->data_start;
      offset          = 0;
      last_row        = 1;
    }
    else
    {
      // calculate new offset and last row based on new values

      offset = abuf->rrow2 - abuf->rrow1;
      last_row = abuf->rrow2;
    }
  }
 

  // Advance to last record marked in buffer
  afk_file_seek( afptr, abuf->rrows[offset], SEEK_SET);


  if (last_row < row)
  {
    // needed row is still beyond what is marked in buffer
    // read until target row is reached

    // get current file location
    pos = afk_file_tell( afptr );

    while( (abuf->rrow2 < row) && 
	   ( line = afk_file_getline_ptr( afptr )) != NULL )
    {
      if (offset == abuf->nrrows )
      {
        abuf->rrow1 = abuf->rrow2;
        abuf->rrows[0] = abuf->rrows[offset-1];
        offset = 1;
      }

      abuf->rrows[offset] = pos;
      pos = afk_file_tell( afptr );
      abuf->rrow2 = offset + abuf->rrow1;
      ++offset;

      last_row = abuf->rrow2;
    }

    // Set position at beginning of last row
    afk_file_seek( afptr, abuf->rrows[last_row - abuf->rrow1], SEEK_SET );
  }

  return  last_row;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_get_nrows                            */
/*                                                             */
/* Description:                                                */
/*              Move to the last row in the buffer             */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            long  nrows                                      */
/*                                                             */
/***************************************************************/
long afk_buffer_get_nrows( afBuffer* abuf )
{
  long nrows = 0;

  nrows = afk_buffer_moveto_row( abuf, abuf->maxrrows );

  if ( nrows < 0 )
    nrows = 0;

  return nrows;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_set_kb                               */
/*                                                             */
/* Description:                                                */
/*              Set block pointer                              */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                    afBlock   *ablock - block pointer        */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_buffer_set_kb( afBuffer* abuf, afBlock* ablock )
{
  if ( abuf )
    abuf->ablock = ablock;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_get_fptr                             */
/*                                                             */
/* Description:                                                */
/*              Get the file pointer from the dataset          */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            afFile  *filepointer                             */
/*                                                             */
/***************************************************************/
afFile* afk_buffer_get_fptr( afBuffer* abuf )
{
 afDataset *ads = NULL;

  if ( abuf )
  {
    ads = afk_block_get_kds( abuf->ablock );
    return afk_dataset_fptr( ads );
  }             


  return NULL;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_append_nullrows                      */
/*                                                             */
/* Description:                                                */
/*              Currently unused                               */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                    long       start                         */
/*                    long       nrows                         */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
#if 0
/* Unused currently */
void afk_buffer_append_nullrows( afBuffer* abuf, long start, long nrows )
{
 (void) abuf;
 (void) start;
 (void) nrows;
}
#endif


/***************************************************************/
/*                                                             */
/*  Function:   afk_buffer_prepare_write                       */
/*                                                             */
/*  Description:                                               */
/*    Prepare the buffer to write to rowno                     */
/*    This makes sure rowno is loaded into the read buffer.    */
/*    It also appends blank rows, if it needs to.              */
/*                                                             */
/*  Input parameters:                                          */
/*    afBuffer  *abuf - buffer pointer                         */
/*    long       rowno                                         */
/*    long      *moved                                         */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*    dmErrCode  status                                        */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_prepare_write( afBuffer* abuf, long rowno, long* moved )
{
  long rowndx   = 0;
  long nrows    = 0;
  long next_row = 0;
  long plen     = 0;
  long rowlen   = 0;
  long ndcols   = 0;
  long ii       = 0;
  dmErrCode status = dmSUCCESS;
  char  *row  = NULL;
  char  *cptr = NULL;
  char  *sep  = NULL;
  dmBool fixed;


  ndcols = abuf->ndcols;
  fixed  = abuf->astyle->fixed;      
  sep    = abuf->astyle->sep;

  if ( rowno >= abuf->wrow1 && rowno <= abuf->wrow2 )
  {
    /* The row is already in the write buffer */
    rowndx = rowno - abuf->wrow1;
    row = abuf->wrows[rowndx];

    /* load it into the parse buffer if it isn't already there */
    if ( rowno < abuf->prow1 || rowno > abuf->prow2 )
      afk_buffer_load_row( abuf, row, rowno );

    /* now can write to the row loaded into the parse buffer */
  }
  else 
  {
    /* the row isn't in the write buffer */

    /* ****************************************************************************** */
    /* The following section has to do with preparing the parse buffer..              */
    /* We need a fresh parsebuf, with as good an estimate as we can get for column    */
    /* sized in order to minimize shifting things later on.                           */
    /* ------------------------------------------------------------------------------ */
    /*                                                                                */

    nrows = afk_buffer_adjust( abuf, rowno, 1 );
    if ( nrows < 1 )
    {
      /* Didn't find the row.. need to setup a new/fresh parsebuf line. */

      /* Determine or estimate row width */
      rowlen = 0;
      if ( fixed && (abuf->prowwidth > 0 ) )
      {
	/* Fixed file with established format */
	rowlen = abuf->prowwidth;
      }
      else
      {
	/* estimate from previous rows maximum         */
	/* each column gets a minimum for 1 separator  */
	for ( ii = 0; ii < ndcols; ii++ )
	{
	  if ( !fixed && (abuf->pmaxwidths[ii] == 0) )
	    abuf->pmaxwidths[ii] = 1;

	  rowlen  += abuf->pmaxwidths[ii];
	}

	rowlen--;
      }
      if ( rowlen <= 0 )
      {
	/* Last resort.. empty parsebuf */
	  rowlen = 1;
      }

      /* Check that the parsebuf is big enough */
      status = afk_buffer_resize_parsebuf( abuf, rowlen );
      if ( status != dmSUCCESS )
	return status;

      /* Fill parsebuf with separator chars (empty).    */
      cptr = abuf->parsebuf;
      memset( cptr, sep[0], (sizeof(char) * rowlen) );
      cptr[rowlen] = '\0'; 
    
      /* Set pointers to column data locations.         */
      for ( ii = 0; ii < ndcols; ii++ )
      {
	abuf->pdata[ii]   = cptr;
	abuf->pwidths[ii] = 0;
	cptr += abuf->pmaxwidths[ii];
      }
      abuf->prow1 = abuf->prow2 = rowno;
 
      /* set parsebuf row size fields */
      abuf->psize = abuf->prsize = rowlen;


    } /* end if (nrows < 1) */

    /*                                                                                */
    /* ------------------------------------------------------------------------------ */
    /* The following part has to do with flushing the write buffer..                  */
    /* This is the only place where flushing is triggered (other than at close).      */
    /* I thought that afk_buffer_write_parsebuf() would be good for it, but it        */
    /* has logic to grow the writebuf instead.  The way things are, the initial       */
    /* writebuf is seldom grown, but flushed here when the next row will require      */
    /* more space.                                                                    */
    /*                                                                                */
    if ( nrows > 0 )
    {
      /* +1 for the newline char */
      plen = abuf->prsize + 1;

      if ( rowno == abuf->wrow2 + 1 )
      {
        nrows = rowno - abuf->wrow1 + 1;

        if ( (nrows > abuf->wmaxrows) || (abuf->wsize + plen > abuf->wmax) )
          afk_buffer_flush_writebuf( abuf, moved );
      }
      else
      {
        afk_buffer_flush_writebuf( abuf, moved );
      }

      abuf->wrsize += plen;
    } 
    else 
    {
      if ( abuf->rrow2 > abuf->wrow2 )
        next_row = abuf->rrow2 + 1;
      else
        next_row = abuf->wrow2 + 1;

      /* we are appending a row to the file */
      if (( next_row-1 ) == abuf->wrow2 )
      {
         /* we might be able to keep the current write buffer */
         nrows = next_row - abuf->wrow1 + 1;

         if ( (nrows > abuf->wmaxrows ) || ( abuf->wsize + rowlen > abuf->wmax ))
	 {
           afk_buffer_flush_writebuf( abuf, moved );
	 }
      }
      else
      {
        /* need to flush buffer and start a new one */
        afk_buffer_flush_writebuf( abuf, moved );
      }

    }
  }
  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_write_col                            */
/*                                                             */
/* Description:                                                */
/*   Write data values to the specified rows of the input      */
/*   buffer to the Ascii buffer.                               */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDescriptor  *dd  - DM Column descriptor                */
/*    long          row1 - start row                           */
/*    long          row2 - end row                             */
/*    void       *values - ptr to data, datatype matches dd    */
/*                                                             */
/*  Output parameters:                                         */
/*   long         *moved - no idea                             */
/*                                                             */
/*  Returns:                                                   */
/*    dmErrCode  status                                        */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_write_col( dmDescriptor* dd, long row1, long row2,
                                void* values, long* moved )
{
  dmBlock       *block  = dmi_descriptor_get_block( dd );
  afBlock       *ablock = (afBlock*)dmi_block_get_kb( block );
  afBuffer      *abuf   = afk_block_get_buffer( ablock );

  dmDescriptor  *cpt_dd = NULL;    /* DM cpt descriptor         */
  afCol         *acol   = NULL;    /* Kernel column descriptor  */
  char          *disp   = NULL;    /* col display format        */

  dmBool       is_image = dmFALSE; /* flag for image data          */ 
  long  hold_image_size = 0;

  void  *vptr     = values;        /* pointer to current value  */

  long vdim       = dmi_descriptor_get_vector_dim( dd );
  long rowno      = 0;
  long cptno      = 0;

  dmErrCode  status =  dmSUCCESS;

  is_image = (ablock->block_type == dmIMAGE);

  if ( is_image )
  {
    /* Code Change  October 30, 2009                */
    /*  Avoid dying with SIGFPE - can't divide by 0 */
    /*  This happens when style is not DTF          */

    if ( ! ablock->axes[0] )
      return dmSUCCESS;

    /* Efficiency mods.. modify the image descriptor information */
    /* so that it wraps image values at the first dimension size */
    /* ie: a 2D MxN image is written as N rows of M values..     */
    /*     this will greatly reduce the size of the parsebuf,    */
    /*     the # of values to shift when the value size changes, */
    /*     and take advantage of the prev 'row' size estimates.  */
    /*                                                           */
    /*  NB: assignment to value param variable deliberate        */
    /*      using as local variable                              */

    hold_image_size = abuf->ndcols;
    row2            = hold_image_size / ablock->axes[0];
    abuf->ndcols    = ablock->axes[0];     /* # values per row = dim1 */
  }


  if ( ! abuf->writebuf )
    afk_buffer_init_writebuf( abuf );

  if ( !abuf->writebuf )
  {
    dmu_err_set( DM_ERR_MEM, "Failed to allocate write buffer." );
    return dmFAILURE;
  }

  for( rowno = row1; rowno <= row2; rowno++ )
  {
    /* prepare parsebuf to accept a row of values */
    status = afk_buffer_prepare_write( abuf, rowno, moved );
    if ( status != dmSUCCESS )
      return status;

    /* Loop over column components */
    for ( cptno = 1; cptno <= vdim; cptno++ )
    {
      cpt_dd = dmi_descriptor_get_cpt( dd, cptno );

      /* Get kernel column for this col */
      acol   = (afCol*)dmi_descriptor_get_kdd( cpt_dd );

      if ( is_image )
      {
	/* temporarily change array length field.. to write dim1 values */
	acol->length = abuf->ndcols;

	/* Write value(s) to kernel buffer */
	vptr = afk_buffer_write_vals( abuf->ablock, acol->no, vptr, cpt_dd->null, disp );

	/* restore */
	acol->length = hold_image_size; 
      }
      else
      {
	/* Write value(s) to kernel buffer */
	vptr = afk_buffer_write_vals( abuf->ablock, acol->no, vptr, cpt_dd->null, disp );
      }

    } /* end component loop */

    status = afk_buffer_write_parsebuf( abuf, rowno );
    if ( status != dmSUCCESS )
      return status; 

  }/* end row loop */

  /* restor image descriptor settings */
  if ( is_image )
  {
    abuf->ndcols = hold_image_size;
  }

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_write_parsebuf                       */
/*                                                             */
/* Description:                                                */
/*   Transfer parse buffer to write buffer.                    */
/*   Align data as needed.                                     */ 
/*                                                             */
/*  Input parameters:                                          */
/*    afBuffer  *abuf - buffer pointer                         */
/*    long       rowno - row number                            */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*    dmErrCode  status                                        */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_write_parsebuf( afBuffer* abuf, long rowno )
{
  long slen   = 0;
  long olen   = 0;
  long offset = 0;
  long diff   = 0;
  long nbytes = 0;
  long i      = 0;
  long iend   = 0;

  dmErrCode status = dmSUCCESS;

  /* See if we need to align the existing records */
  if ( abuf->align )
    status = afk_buffer_align( abuf );

  slen = abuf->psize;

  if ( abuf->wrow1 == 0 ) /* write buffer is empty */
  {
    abuf->wrow1 = abuf->wrow2 = rowno;
    abuf->wrows[0] = abuf->writebuf;
  }

  offset = rowno - abuf->wrow1;

  if ( rowno == abuf->wrow2+1 ) /* appending a row */
  {
    // Code Change  Bug 11420  December 8, 2008
    //  Code below modified to allocate extra 2 bytes instead of 1
    //  to allow for new line character and null terminator

    status = afk_buffer_resize_writebuf( abuf, abuf->wsize + slen + 2,
                                         abuf->wrow2 - abuf->wrow1 + 2 );

    if ( status != dmSUCCESS )
      return status;

    abuf->wrow2++;
    abuf->wrows[offset] = abuf->writebuf + abuf->wsize;

    strcpy( abuf->wrows[offset], abuf->parsebuf );
    abuf->wsize += slen + 1; /* +1 for newline char */

    /* make sure the row is terminated with a newline */
    abuf->writebuf[abuf->wsize - 1] = '\n';
    abuf->writebuf[abuf->wsize] = '\0';
  }
  else
    if ( rowno == abuf->wrow2 ) /* The last row in the write buffer */
    {
      olen = abuf->writebuf + abuf->wsize - abuf->wrows[offset];
      diff = slen + 1 - olen;
      status = afk_buffer_resize_writebuf( abuf, abuf->wsize + diff + 1, 0 );

      if ( status != dmSUCCESS )
        return status;

      strcpy( abuf->wrows[offset], abuf->parsebuf );
      abuf->wsize += diff;
      abuf->writebuf[abuf->wsize - 1] = '\n';
      abuf->writebuf[abuf->wsize] = '\0';
    }
    else
      if ( rowno >= abuf->wrow1 && rowno < abuf->wrow2 )
      {
        olen = abuf->wrows[offset+1] - abuf->wrows[offset];

        if ( slen+1 == olen )
        {
          memcpy( abuf->wrows[offset], abuf->parsebuf, slen );
        }
        else
          if ( slen < olen )
          {
	    /* MCD 20070910:  This segment is applied when the write buffer  */
	    /* is already allocated and the current row is smaller than the  */
	    /* existing line.                                                */
	    /* Change from filling with white space at EOL which are not     */
	    /* in any of the other rows (earlier buffer segments), to        */
	    /* shifting the data inward to take up the space.                */
            diff = slen +1 - olen;
            memcpy( abuf->wrows[offset], abuf->parsebuf, slen );
            abuf->wrows[offset][slen] = '\n';
            nbytes = abuf->wsize - ( abuf->wrows[offset+1] - abuf->wrows[0] ) + 1;
            memmove( abuf->wrows[offset+1] + diff, abuf->wrows[offset+1], nbytes );
            iend = abuf->wrow2 - abuf->wrow1;

            for ( i = offset+1; i <= iend; i++ )
            {
              abuf->wrows[i] += diff;
            }

          }
          else
          {
            diff = slen + 1 - olen;
            status = afk_buffer_resize_writebuf( abuf, abuf->wsize + diff, 0 );

            if ( status != dmSUCCESS )
              return status;

            nbytes = abuf->wsize - ( abuf->wrows[offset+1] - abuf->wrows[0] ) + 1;
            memmove( abuf->wrows[offset+1] + diff - 1, abuf->wrows[offset+1] - 1,
                     nbytes );
            memcpy( abuf->wrows[offset], abuf->parsebuf, slen );
            abuf->wsize += diff;
            iend = abuf->wrow2 - abuf->wrow1;

            for ( i = offset+1; i <= iend; i++ )
            {
              abuf->wrows[i] += diff;
            }
          }
      }
      else
      {  
        /* This shouldn't happen.. */
        printf( "DM EMERGENCY, WRITE ERROR!!\n" );
        dmu_err_set( DM_ERR_KERNEL, "TEXT Kernel: Severe write error." );
        return dmFAILURE;
      }

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_add_keyline                          */
/*                                                             */
/* Description:                                                */
/*              Add a keyline that was read from the file to   */
/*              the buffer                                     */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer   *abuf - buffer pointer        */
/*                    const char *keyline - keyline            */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_add_keyline( afBuffer* abuf, const char* keyline )
{
  char *cptr   = NULL;
  long len     = 0;
  long totsize = 0;
  long totrows = 0;
  long status  = dmSUCCESS;

  len = strlen(keyline);

  if ( abuf->hrow1 == 0 )
  {
    afk_buffer_init_headerbuf( abuf );
    abuf->hrow1 = 1;
  }

  totsize = abuf->hsize + len + 2; /* +1 for newline, +1 for \0 */
  totrows = abuf->hrow2 - abuf->hrow1 + 2;
  status = afk_buffer_resize_headbuf( abuf, totsize, totrows );
 
  if ( status != dmSUCCESS )
    return status;

  /* Use cptr so strcat doesn't need to zip through whole header row */
  cptr = abuf->headbuf + abuf->hsize;
  strcat( cptr, keyline );
  cptr += len;
  strcat( cptr, "\n" );
  abuf->hrows[abuf->hrow2] = &abuf->headbuf[abuf->hsize];
  abuf->hsize += len + 1;
  abuf->hrsize += len + 1;
  abuf->hrow2++;

  return status;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_init_headerbuf                       */
/*                                                             */
/* Description:                                                */
/*              Initialize header buffer                       */
/*              llen                                           */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_buffer_init_headerbuf( afBuffer* abuf )
{
  abuf->hmax = AK_MIN_HEADBUF;
  abuf->headbuf = calloc( AK_MIN_HEADBUF, sizeof(char) );

  if ( !abuf->headbuf ) 
  {
    abuf->hmax = 0;
    return;
  }

  abuf->hmaxrows = AK_HEADBUF_MINROWS;
  abuf->hrows = calloc( AK_HEADBUF_MINROWS, sizeof(char*) );

  if ( !abuf->hrows )
  {
    free( abuf->headbuf );
    abuf->hmax = 0;
    abuf->hmaxrows = 0;
    return;
  }
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_resize_headbuf                       */
/*                                                             */
/* Description:                                                */
/*    Resize the header buffer to hold newsize chars           */
/*    and nrows rows                                           */
/*                                                             */
/*  Input parameters:                                          */
/*    afBuffer  *abuf - buffer pointer                         */
/*     ong       newsize - new size                           l*/
/*     ong       nrows - number of rows                       l*/
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*     dmErrCode  status                                       */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_resize_headbuf( afBuffer* abuf, long newsize,
                                     long nrows )
{
  char *cptr = NULL;

  long offset = 0;
  long diff   = 0;
  long hrow2  = 0;

  // Code change - Bug 9222  June 8, 2006
  // removed size limit on headbuf
  // code section for testing limits removed


  // Is the new size greater than the current max?
  if ( newsize > abuf->hmax )
  {
    // increment the max size so that it is greater than request
    while ( abuf->hmax < newsize )
    {
      abuf->hmax += AK_HEADBUF_INCR;
    }

    // reallocate buffer
    cptr = abuf->headbuf;
    abuf->headbuf = realloc( abuf->headbuf, abuf->hmax * sizeof(char) );

    // if reallocation failed, clear elements in abuf, set error
    if ( ! abuf->headbuf )
    {
      free( abuf->hrows );
      abuf->hmax = 0;
      abuf->hmaxrows = 0;
      printf( dm_errbuf,
	     "TEXT Kernel: Failed to resize header buffer to hold %d bytes.",
	     newsize );
      dmu_err_set( DM_ERR_MEM, dm_errbuf );
      return dmFAILURE;
    }

    // mark row position
    if ( cptr )
    {
      diff  = abuf->headbuf - cptr;
      hrow2 = abuf->hrow2;

      for ( offset = abuf->hrow1-1; offset < hrow2; offset++ )
      {
        abuf->hrows[offset] += diff;
      }
    }
  }


  // check that header row buffer has adequate space
  if ( nrows > abuf->hmaxrows )
  {
    // increment hrows buffer
    while( abuf->hmaxrows < nrows )
    {
      abuf->hmaxrows += AK_HEADBUF_ROWINCR;
    }

    abuf->hrows = realloc( abuf->hrows, abuf->hmaxrows * sizeof(char*) );

    // if reallocation failed, set error
    if ( ! abuf->hrows )
    {
      abuf->hmaxrows = 0;
      printf( dm_errbuf,
              "TEXT Kernel: Failed to resize header buffer to hold %d rows.",
              nrows );
      dmu_err_set( DM_ERR_MEM, dm_errbuf );
      return dmFAILURE;
    }
  }

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_resize_writebuf                      */
/*                                                             */
/* Description:                                                */
/*              Resize the write buffer to hold newsize chars  */
/*              and nrows rows.  Only use this if there is     */
/*              NO WAY to flush the buffer at the point where  */
/*              this is called.  The write buffer SHOULD stay  */
/*              the same size throughout the program.          */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                    long       newsize                       */
/*                    long       nrows                         */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_resize_writebuf( afBuffer* abuf, long newsize,
                                      long nrows )
{
 long  diff   = 0;
 long  offset = 0;
 long  eoff   = 0;
 char *oldbuf = NULL;


  if ( newsize > abuf->wmax )
  {
    abuf->wmax = newsize;
    oldbuf = abuf->writebuf;
    abuf->writebuf = realloc( abuf->writebuf, abuf->wmax * sizeof(char) );

    if ( ! abuf->writebuf )
    {
      free( abuf->wrows );
      abuf->wmax = 0;
      abuf->wmaxrows = 0;
      printf( dm_errbuf,
             "TEXT Kernel: Failed to resize writebuf to hold %d bytes.",
              newsize );
      dmu_err_set( DM_ERR_MEM, dm_errbuf );
      return dmFAILURE;
    }

    diff = abuf->writebuf - oldbuf;

    if ( abuf->wrow1 )
    {
      eoff = abuf->wrow2 - abuf->wrow1;

      for ( offset = 0; offset <= eoff; offset++ )
        abuf->wrows[offset] += diff;
    }
  }

  if ( nrows > abuf->wmaxrows )
  {
    abuf->wmaxrows = nrows;
    abuf->wrows = realloc( abuf->wrows, abuf->wmaxrows * sizeof(char*) );

    if ( ! abuf->wrows )
    {
      abuf->wmaxrows = 0;
      printf( dm_errbuf,
              "TEXT Kernel: Failed to resize writebuf to hold %d rows.", nrows );
      dmu_err_set( DM_ERR_MEM, dm_errbuf );
      return dmFAILURE;
    }
  }

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_delete_keyline                       */
/*                                                             */
/* Description:                                                */
/*              Delete keyline                                 */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                    long       lineno - line number          */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_buffer_delete_keyline( afBuffer* abuf, long lineno )
{
  long   size   = 0;  /* the size of the rest of the buffer */
  long   lsize  = 0; /* line size */
  long   offset = 0;
  long   hrow2  = 0;
  char  *line   = NULL;
  char  *ptr    = NULL;

  if ( lineno < 1 || lineno > abuf->hrow2 )
    return;

  line = abuf->hrows[lineno-1]; 

  if ( lineno < abuf->hrow2 )
  {
    lsize = abuf->hrows[lineno] - abuf->hrows[lineno-1];
    size = abuf->headbuf + abuf->hsize + 1 - abuf->hrows[lineno];
    ptr = abuf->hrows[lineno];
    memmove( line, ptr, size );
    hrow2 = abuf->hrow2;

    for ( offset = lineno; offset < hrow2; offset++ )
    {
     abuf->hrows[offset-1] = abuf->hrows[offset] - lsize;
    }
  }
  else /* lineno == abuf->hrow2 */
  {
    line[0] = '\0';
    lsize = abuf->headbuf + abuf->hsize - abuf->hrows[lineno-1];
  }

  abuf->hrow2--;
  abuf->hsize -= lsize;
  abuf->hupdate = UT_TRUE;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_insert_keyline                       */
/*                                                             */
/* Description:                                                */
/*              Insert a new key line at the given header      */
/*              position                                       */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer   *abuf - buffer pointer        */
/*                    const char *keyline                      */
/*                    long        lineno - line number         */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_insert_keyline( afBuffer* abuf, const char* keyline,
                                     long lineno )
{
  long maxrow      = abuf->hrow2+1;
  long llen        = 0; /* line length */
  long offset      = 0;
  long end_offset  = 0;
  long totsize     = 0;
  long totrows     = 0;
  long size        = 0;
  dmErrCode status = dmSUCCESS;


  if ( lineno < 1 || lineno > maxrow )
  {
    sprintf( dm_errbuf, "TEXT Kernel: Bad header line number: %ld.", lineno );
    dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
    return dmFAILURE;
  }

  if ( abuf->hrow1 == 0 )
  {
    afk_buffer_init_headerbuf( abuf );

    if ( !abuf->headbuf )
    {
      dmu_err_set( DM_ERR_MEM, "TEXT Kernel: Failed to allocate header buffer." );
      return dmFAILURE;
    }

    abuf->hrow1 = 1;
  }

  llen = strlen( keyline );
  totsize = abuf->hsize + llen;
  totrows = abuf->hrow2 - abuf->hrow1 + 2;
  status = afk_buffer_resize_headbuf( abuf, totsize+1, totrows );

  if ( status != dmSUCCESS )
    return status;

  if ( lineno < maxrow )
  {
    end_offset = lineno-1;

    for( offset = abuf->hrow2-1; offset >= end_offset; offset-- )
    {
      abuf->hrows[offset+1] = abuf->hrows[offset] + llen;
    }

    offset = lineno - 1;
    size = abuf->headbuf + abuf->hsize - abuf->hrows[offset];
    memmove( abuf->hrows[lineno], abuf->hrows[offset], size );
    memcpy( abuf->hrows[offset], keyline, llen );
  }
  else /* lineno == maxrow */
  {
    abuf->hrows[abuf->hrow2] = &abuf->headbuf[abuf->hsize];
    strcpy( abuf->hrows[lineno-1], keyline );
  }

  abuf->hrow2++;
  abuf->hsize = totsize;
  abuf->hupdate = UT_TRUE;

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_replace_keylline                     */
/*                                                             */
/* Description:                                                */
/*              Replace an existing key line                   */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                    const char *keyline - new key line       */
/*                    long       lineno - line number          */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_replace_keyline( afBuffer* abuf,
                                      const char* keyline, long lineno )
{
  long offset   = 0;
  long diff     = 0;
  long old_llen = 0;
  long llen     = 0;
  long size     = 0;
  long totsize  = 0;
  long totrows  = 0;
  long hrow2    = 0;
  dmErrCode status = dmSUCCESS;


  if ( lineno < 1 || lineno > abuf->hrow2 )
  {
    sprintf( dm_errbuf,
             "TEXT Kernel: Trying to replace keyline at bad line number: %ld.",
              lineno );

    dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
    return dmFAILURE;
  }

  llen = strlen( keyline );
  offset = lineno - 1;

  if ( lineno < abuf->hrow2 )
  {
    old_llen = abuf->hrows[lineno] - abuf->hrows[offset];
    diff = llen - old_llen;
    totsize = diff + abuf->hsize;

    if ( diff > 0 )
    {
      totrows = abuf->hrow2 - abuf->hrow1 + 1;
      status = afk_buffer_resize_headbuf( abuf, totsize, totrows );

      if ( status != dmSUCCESS )
        return status;
    }

    size = abuf->headbuf + abuf->hsize - abuf->hrows[lineno];
    memmove( abuf->hrows[lineno] + diff, abuf->hrows[lineno], size );
    memcpy( abuf->hrows[offset], keyline, llen );
    hrow2 = abuf->hrow2;

    for ( offset = lineno; offset < hrow2; offset++ )
    {
      abuf->hrows[offset] += diff;
    }
  }
  else /* lineno == abuf->hrow2 */
  {
    old_llen = abuf->headbuf + abuf->hsize - abuf->hrows[offset];
    diff = llen - old_llen;
    totsize = diff + abuf->hsize;

    if ( diff > 0 )
    {
      totrows = abuf->hrow2 - abuf->hrow1 + 1;
      status = afk_buffer_resize_headbuf( abuf, totsize, totrows );

      if ( status != dmSUCCESS )
        return status;
    }

    strcpy( abuf->hrows[offset], keyline );
  }

  abuf->hsize = totsize;
  abuf->hupdate = UT_TRUE;

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_flush_headbuf                        */
/*                                                             */
/* Description:                                                */
/*              Flush the header buffer                        */
/*              Parameter moved holds the + number of bytes    */
/*              added to the file or the - number of bytes     */
/*              removed from the file after the flush.         */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                                                             */
/*  Output parameters:                                         */
/*                     long  *moved                            */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_flush_headbuf( afBuffer* abuf, long* moved )
{
  afBlock    *ablock = NULL;
  afDataset  *ads    = NULL;
  afFile     *afptr  = NULL;

  long diff   = 0;
  long offset = 0;
  long max_offset = 0;


  if ( !abuf->hupdate )
  {
    if ( moved )
      *moved = 0;
    return dmSUCCESS;
  }

  ablock = abuf->ablock;
  ads    = ablock->ads;
  diff   = abuf->hsize - abuf->hrsize;
  afk_dataset_moveto_block( ads, ablock->no );
  afptr  = afk_buffer_get_fptr( abuf );
  afk_file_replace_data( afptr, abuf->headbuf, abuf->hsize, abuf->hrsize );

  /* need to update the row positions */
  if ( diff != 0 )
  {
    abuf->data_start += diff;

    if ( abuf->rrows )
    {
      max_offset = abuf->rrow2 - abuf->rrow1;

      for ( offset = 0; offset <= max_offset; offset++ )
      {
        abuf->rrows[offset] += diff;
      }
    }

    abuf->hrsize = abuf->hsize;
  }

  if ( moved )
    *moved = diff;

  abuf->hupdate = UT_FALSE;

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_insert_col                           */
/*                                                             */
/* Description:                                                */
/*              Insert a column                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                     afCol     *acol - column pointer        */
/*                    long       colno - column number         */
/*                                                             */
/*  Output parameters:                                         */
/*                    long      *moved                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_insert_col( afBuffer* abuf, afCol* acol,
                                 long colno, long* moved )
{
  char  *tmp  = NULL;
  char  *cptr = NULL;
  char  *ptr  = NULL;

  long tlen       = 6;
  long rowno      = 1;
  long col_offset = 0;
  long llen       = 0;
  long len        = 0;
  long lmoved     = 0;
  long tmoved     = 0;
  long nbytes     = 0;
  long ar_len     = 0;
  long ncols      = 0;
  long old_ncols  = 0;
  long i          = 0;
  long maxi       = 0;
  long width      = 0;
  long colndx     = 0;
  long colend     = 0;

  dmErrCode status = dmSUCCESS;
  dmBool append = ( colno > abuf->ncols );
  dmBool fixed;

  if ( acol->length == 0 )
    ar_len = 1;
  else 
    ar_len = acol->length;

  tmp = calloc( ar_len * 6 + 2, sizeof(char) );

  if ( !tmp )
  {
    dmu_err_set( DM_ERR_MEM, "TEXT Kernel: Failed to allocate memory." );
    return dmFAILURE;
  }

  tlen = 0;
  ptr = tmp;

  for ( i = 0; i < ar_len; i++ )
  {
    len = sprintf( ptr, "%c%s", abuf->astyle->sep[0], NULLSTR );
    tlen += len;
    ptr += len;
  }

  if ( append )

  {
    col_offset = abuf->ndcols;
  }
  else
  {
    tlen += sprintf( ptr, "%c", abuf->astyle->sep[0] );
    col_offset = abuf->col_offsets[colno-1];
  }

  if ( !abuf->writebuf )
    afk_buffer_init_writebuf( abuf );

  while( afk_buffer_adjust( abuf, rowno, 1 ))
  {
    status = afk_buffer_prepare_write( abuf, rowno, &lmoved );

    if ( status != dmSUCCESS )
    {
      free( tmp );
      return status;
    }

    tmoved += lmoved;
    llen = abuf->psize;
    status = afk_buffer_resize_parsebuf( abuf, llen + tlen );

    if ( status != dmSUCCESS )
    {
      free( tmp );
      return status;
    }

    if ( append )
    {
      /* use cptr so strcat doesn't need to zip through whole row */
      cptr = abuf->parsebuf + llen;
      strcat( cptr, tmp );
    }
    else
    {
      nbytes = llen - ( abuf->pdata[col_offset] - abuf->parsebuf );
      memmove( abuf->pdata[col_offset] + tlen, abuf->pdata[col_offset], nbytes );
      memcpy( abuf->pdata[col_offset], tmp, tlen );
    }

    abuf->psize += tlen;
    status = afk_buffer_write_parsebuf( abuf, rowno );

    if ( status != dmSUCCESS )
    {
      free( tmp );
      return status;
    }

    ++rowno;
  }

  old_ncols = abuf->ncols;
  abuf->ncols++;
  abuf->ndcols += ar_len;
  abuf->col_offsets = realloc( abuf->col_offsets, abuf->ncols*sizeof(long) );

  /* expand col info arrays in parse buffer */
  /* and initialize value for this column.. */
  /* MCD NOTE: this does not appear to shift existing info for current  */
  /*           columns beyond col_offsets.. problem?                    */

  abuf->pdata = realloc( abuf->pdata, abuf->ndcols*sizeof(char*) );
  memset( abuf->pdata + col_offset, 0, sizeof(char*)*ar_len );

  abuf->pwidths = realloc( abuf->pwidths, abuf->ndcols*sizeof(long) );
  memset( abuf->pwidths + col_offset, 0, sizeof(long)*ar_len );

  abuf->pmaxwidths = realloc( abuf->pmaxwidths, abuf->ndcols*sizeof(long) );
  memset( abuf->pmaxwidths + col_offset, 0, sizeof(long)*ar_len );

  abuf->pfmts = realloc( abuf->pfmts, abuf->ndcols*sizeof(char*) );
  colend = col_offset + ar_len-1;
  for ( colndx = col_offset; colndx <= colend; colndx++ )
    abuf->pfmts[colndx] = calloc( CFMT_WIDTH, sizeof(char) );


  fixed = abuf->astyle->fixed;

  if ( fixed )
  {
    maxi = col_offset + ar_len - 1;

    if ( !acol->fwidth )
    {
      abuf->prowwidth = 0;
      abuf->align = UT_TRUE;
    }
    else
    {
      width = acol->fwidth / ar_len;

      for ( i = col_offset; i <= maxi; i++ )
        abuf->pmaxwidths[i] = width;

      if ( !abuf->align )
        abuf->prowwidth += acol->fwidth;
    }

    if ( !acol->width )
      memset( abuf->pwidths + col_offset, 0, sizeof(long)*ar_len );
    else
    {
      width = acol->width;

      for ( i = col_offset; i <= maxi; i++ )
        abuf->pwidths[i] = width;
    }
  }

  if ( !abuf->col_offsets || !abuf->pdata || !abuf->pwidths )
  {
    free( tmp );
    dmu_err_set( DM_ERR_MEM, "TEXT Kernel: Failed to allocate memory." );
    return dmFAILURE;
  }

  if ( !append )
  {
    ncols = abuf->ncols;

    for ( i = col_offset + 1; i < ncols; i++ )
    {
      abuf->col_offsets[i] += ar_len;
    }
  }
  else
    abuf->col_offsets[colno-1] = col_offset;

  abuf->prow1 = 0;
  abuf->prow2 = 0;
  *moved = tmoved;
  free( tmp );

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_notify_moved                         */
/*                                                             */
/* Description:                                                */
/*              Alert the buffer that its data has been        */
/*              moved.  Move all file position markers by      */
/*              the value in parameter moved.                  */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                    long       moved - offset position       */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_buffer_notify_moved( afBuffer* abuf, long moved )
{
  long offset = 0;
  long eoff   = 0;

  abuf->data_start += moved;
 
  if ( abuf->rrow1 > 0 )
  {
    eoff = abuf->rrow2 - abuf->rrow1;

    for ( offset = 0; offset <= eoff; offset++ )
    {
      abuf->rrows[offset] += moved;
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_write_rows                           */
/*                                                             */
/* Description:                                                */
/*   Write data values from the specified rows of the input    */
/*   buffer to the provided Ascii buffer.                      */
/*                                                             */
/*   NOTE:  The buffer input here is NOT guaranteed to be      */
/*     to be from the same kernel as the destination.          */
/*     It IS assumed that the buffer has the same structure    */
/*     and definitions as the destination, but I don't see     */
/*     why that should be expected.                            */
/*                                                             */
/*                                                             */
/* Input parameters:                                           */
/*   afBuffer  *abuf     - Destination Ascii kernel buffer     */
/*   dmBuffer  *buffer   - Source DM buffer                    */
/*   long       row1     - start row                           */
/*   long       row2     - end row                             */
/*                                                             */
/* Output parameters:                                          */
/*   long     *moved     - no idea                             */
/*                                                             */
/* Returns:                                                    */
/*   dmErrCode  status                                         */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_write_rows( afBuffer* abuf, dmBuffer* buffer,
                                 long row1, long row2, long* moved )
{
  dmDescriptor **cols    = NULL;   /* Array of DM column descriptors  */
  dmDescriptor  *dd      = NULL;   /* DM column descriptor            */
  dmDescriptor  *cpt_dd  = NULL;   /* column component                */
  afCol         *acol    = NULL;   /* Ascii column descriptor         */
  afBlock       *ablock  = NULL;
  void          *vptr    = NULL;   /* pointer to values as a void     */

  long cptno       = 0;  /* column component number */
  long rrow1       = 0;  /* first row in DM buffer. */
  long rowno       = 0;  /* current rows            */
  long bcolndx     = 0;  /* DM buffer col. index    */
  long nbcols      = 0;  /* # DM buffer cols        */
  long vdim        = 0;  /* vector dimensionality   */
  long rrow_offset = 0;  /* row offset in DM buffer */

  dmErrCode status = dmSUCCESS;


  if ( ! abuf->writebuf )
    afk_buffer_init_writebuf( abuf );

  if ( !abuf->writebuf )
  {
    dmu_err_set( DM_ERR_MEM, "Failed to allocate write buffer." );
    return dmFAILURE;
  }

  nbcols   = buffer->ncols;
  cols     = buffer->cols;
  rrow1    = buffer->rrow1;
  ablock   = abuf->ablock;


  for( rowno = row1; rowno <= row2; rowno++ )
  {
    status = afk_buffer_prepare_write( abuf, rowno, moved );

    if ( status != dmSUCCESS )
      return status;

    rrow_offset = rowno - rrow1;

    /* Loop DM buffer columns */
    for ( bcolndx = 0; bcolndx < nbcols; bcolndx++ )
    {
      /* Get DM column descriptor */
      dd = cols[bcolndx];

      /* In column based thread..                                  */
      /*   dmp_buffer_flush() screens columns based on flag type   */
      /*      DM_FLAG_BINNED, DM_FLAG_UNBUF, DM_FLAG_AUX_COLUMN    */
      /*      which are either not in the buffer, or shouldn't be  */
      /*      flushed to the kernel level buffer.                  */
      /*   dmp_buffer_write() contains a usage for DM_FLAG_UNBUF   */
      /*      in order to set data to the kernel buffer.  That     */
      /*      usage is not supported here since there is no        */
      /*      provision for supplying input values.  Perhaps we    */
      /*      could extract the values from the input buffer and   */
      /*      write them to the destination block, but that would  */
      /*      be redundant in the first usage since they are the   */
      /*      same.                                                */

      if ( ( dd->flag & DM_FLAG_BINNED ) || 
	   ( dd->flag & DM_FLAG_AUX_COLUMN ) ||
	   ( dd->flag & DM_FLAG_UNBUF ) )
        continue;
      
      /* Data should be in buffer.. set pointer to start position */
      vptr = (char*)buffer->data + dmi_buffer_offset( buffer, bcolndx, rrow_offset );

      /* Loop column components */
      /* Do not have to shift value pointer between components..  */
      /* they should be in order.. thus the loop here.            */
 
      vdim = dmi_descriptor_get_vector_dim( dd );
 
      for ( cptno = 1; cptno <= vdim; cptno++ )
      {
        cpt_dd = dmi_descriptor_get_cpt( dd, cptno );

	/* get the kernel descriptor */
        acol = (afCol*)dmi_descriptor_get_kdd( cpt_dd );

	/* Write value(s) to kernel buffer */
	vptr = afk_buffer_write_vals( abuf->ablock, acol->no, vptr, cpt_dd->null, NULL );

      }/* end cpt loop */
    } /* end column loop */

    /* Transfer parsebuf to write buffer */
    status = afk_buffer_write_parsebuf( abuf, rowno );

    if ( status != dmSUCCESS )
     return status;

  } /* end row loop */

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_write_vals                           */
/*                                                             */
/* Description:                                                */
/*   Write data values to ascii kernel buffer (parsebuf)       */
/*   Will extract value description from kernel structures     */
/*                                                             */
/* Input parameters:                                           */
/*   afBlock *ablock     - Ascii kernel block                  */
/*   long      colno     - Ascii kernel column number          */
/*   void    *values     - values going to buffer              */
/*   void   *nullval     - external value for null (TNULL)     */
/*   char*      disp     - external display format             */
/*                                                             */
/* Output parameters:                                          */
/*                                                             */
/* Returns:                                                    */
/*   dmErrCode  status                                         */
/*                                                             */
/***************************************************************/
char*  afk_buffer_write_vals( afBlock* ablock, long colno, 
			      void* values, void* nullval, char* disp )
{
  afBuffer    *abuf = NULL;   /* Ascii buffer                  */
  afStyle   *astyle = NULL;   /* Ascii kernel style info       */
  afCol       *acol = NULL;   /* Ascii column descriptor       */

  long        nvals = 1;      /* # values this column contains */
  long        start = 0;      /* start pos of column in buffer */
  long          end = 0;      /* end pos of column in buffer   */
  long      kcolndx = 0;      /* counter                       */
  long         size = 0;      /* size of datatype in bytes     */
  long         slen = 0;      /* length of string col. values  */
  long         shft = 0;      /* #bytes to shift values ptr    */
  long        width = 0;      /* Width of values as string     */
  long       owidth = 0;      /* space needed in buffer        */
  long        avail = 0;      /* current space in buffer       */
  long         diff = 0;
  long          len = 0;
  long       nbytes = 0;
  long           ii = 0;
  long     lLogical = 0;
  long       lValue = 0;
  int        iValue = 0;

  float      fValue = 0.0;
  double     dValue = 0.0;
  
  char tmp[AK_COLVAL_LEN];    /* Holds value in string format  */
  char        *vptr = NULL;   /* used with string values       */
  char        *cptr = NULL;   /* used with string values       */
  char      sep[10] = "";     /* Holds field separator string  */
  char      fmt[20] = "";     /* Holds format                  */
  char        *cfmt = fmt;    /* pointer to format string      */
  char *pDefinedNullStr = NULL; /* user defined null string    */

  dmDataType  dtype;           /* datatype of column values    */
  dmBool      fixed = dmFALSE; /* fixed format?                */
  dmErrCode  status = dmSUCCESS;

  unsigned char mask[8]  = { 1, 2, 4, 8, 16, 32, 64, 128 };

  abuf   = ablock->abuf;
  acol   = ablock->cols[colno-1];
  astyle = abuf->astyle;

  dtype = acol->datatype;
  slen  = acol->width;

  pDefinedNullStr = astyle->nullstr;


  if ( acol->length )
    nvals = acol->length;

  size  = ar_mix_type_size( dtype );
  shft  = size;

  /* check if output is fixed format or not */
  fixed = astyle->fixed;

  vptr  = (char*)values;
  start = abuf->col_offsets[colno-1];
  end   = start + nvals;

  for ( kcolndx = start; kcolndx < end; kcolndx++ )
  {
    /* Set field separator string.. if separator is non-white,   */
    /* add pad space for aesthetic appeal as stated in spec.     */

    if ( !fixed && ( kcolndx < (abuf->ndcols-1) ) &&
	 ((astyle->sep[0] != ' ') && (astyle->sep[0] != '\t')) )
    {
      sprintf(sep, "%c ", astyle->sep[0] );
    }
    else
    {
      sprintf(sep, "%c", astyle->sep[0] );
    }

    /* Get the C display format for the column                      */
    /* NOTE: Format not expected to change within loop for elements */
    /*       of an array column, but the default format is dependent*/
    /*       on the data value.. so set here.                       */

    cfmt[0] = '\0';

    if ( disp && *disp == '\0' )
    {
      /* External override */
      strcpy( cfmt, disp );
    }
    else if ( acol->disp )
    {
      /* External (TDISP) format provided for this column         */
      /* Use format string from buffer.                           */
      strcpy(cfmt, abuf->pfmts[start]);

      if ( !cfmt[0] )
      {
	/* format has not been set from disp yet.. do so          */
	/* seed with default to ensure that quote characters      */
	/* are inserted if necessary (Bug 10408 04/24/2007)       */

	afk_default_fmt( dtype, vptr, cfmt, astyle );
	afk_tdisp_to_cfmt( dtype, acol->disp, cfmt );
      }
    }

    if ( !cfmt[0] )
      afk_default_fmt( dtype, vptr, cfmt, astyle );

    width = 0;

    switch( dtype )
    {
      case DT_UINT1: 
        width = sprintf( tmp, cfmt, ((unsigned char*)vptr)[0] );
        break;

      case DT_INT2:
        iValue = ((short*)vptr)[0];

        if ( pDefinedNullStr && 
	     ( (  nullval && ( iValue == ((short*)nullval)[0] )) ||
	       ( !nullval && ( iValue == 0                    )) ))
          width = sprintf( tmp, "%s", pDefinedNullStr);
        else
          width = sprintf( tmp, cfmt, iValue );
        break;

      case DT_UINT2:
        iValue = ((unsigned short*)vptr)[0];

        if ( pDefinedNullStr && 
	     ( (  nullval && ( iValue == ((unsigned short*)nullval)[0] )) ||
	       ( !nullval && ( iValue == 0                             )) ))
          width = sprintf( tmp, "%s", pDefinedNullStr);
        else
          width = sprintf( tmp, cfmt, (unsigned short)iValue );
        break;

      case DT_INT4:
       lValue = ((long*)vptr)[0];

        if ( pDefinedNullStr && 
	     ( (  nullval && ( lValue == ((long*)nullval)[0] )) ||
	       ( !nullval && ( lValue == 0                   )) ))
          width = sprintf( tmp, "%s", pDefinedNullStr);
        else
          width = sprintf( tmp, cfmt, lValue );
        break;

      case DT_UINT4:
       lValue = ((unsigned long*)vptr)[0];

        if ( pDefinedNullStr && 
	     ( (  nullval && ( lValue == ((unsigned long*)nullval)[0] )) ||
	       ( !nullval && ( lValue == 0                            )) ))
          width = sprintf( tmp, "%s", pDefinedNullStr);
        else
          width = sprintf( tmp, cfmt, (unsigned long) lValue );
        break;

      case DT_REAL4:
        fValue = ((float*)vptr)[0];

        if ( pDefinedNullStr && (ut_qnan_r(fValue)))
          width = sprintf( tmp, "%s", pDefinedNullStr);
        else
          width = sprintf( tmp, cfmt,  fValue);
        break;

      case DT_REAL8:
        dValue = ((double*)vptr)[0];

        if ( pDefinedNullStr && (ut_qnan_d(dValue)))
          width = sprintf( tmp, "%s", pDefinedNullStr);
        else
          width = sprintf( tmp, cfmt, dValue );

        break;

      case DT_CHAR:
        cptr = (char*)vptr;
        shft = (slen+1);

        if ( cptr )
        {
	  len = strlen(cptr);
          if ( len >= AK_COLVAL_LEN )
	  {
	    sprintf( dm_errbuf,
		     "TEXT Kernel: string value is too long (%ld).", len );
	    dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
	    return vptr;
	  }

          if ( pDefinedNullStr && (len == 0))
            width = sprintf( tmp, "%s", pDefinedNullStr);
	  else
	    width = sprintf( tmp, "%s", cptr );

        }
        else
        {
          // null string value - use either defined nullstring or INDEF
          if ( pDefinedNullStr )
            width = sprintf( tmp, "%s", pDefinedNullStr);
          else
	    width = sprintf( tmp,"%s", NULLSTR );
        }


        /* process string for output.. */
        cptr  = afk_prepare_string( tmp, cfmt, astyle );
        width = sprintf( tmp, "%s", cptr );

        if ( cptr )
	  free(cptr);
        break;

      /* Case added May 4, 2007 */
      /* Code Change  Bug 12078 December 2, 2009        */
      /*  Access DT_LOGICAL as dmBool instead of long */
      /*  or else size is wrong for 64-bit machines   */

      case DT_LOGICAL:
        lLogical = (long)((dmBool*)vptr)[0];

        if ( lLogical )
	  width = sprintf( tmp, cfmt, "T" );
        else
	  width = sprintf( tmp, cfmt, "F" );                
        break;

      case DT_BIT:
	/* array of bytes gets translated to bit pattern string */
	if ( vptr )
	{
	  long offset = 0; 
	  long bitno = 0;
	  long byteno = 0; 
	  long ii = 0;
	  for ( bitno = 1; bitno <= slen; bitno++)
	  {
	    offset = (BYTE_BITS - 1) - ((bitno - 1) % BYTE_BITS);
	    byteno = (bitno - 1 )/ BYTE_BITS;
	    (((unsigned char*)vptr)[byteno] & mask[offset] )?( tmp[ii] = '1'):( tmp[ii] = '0');
	    ii++;
	  }
	  tmp[ii] = '\0';
	  width = ii;

          // Code Change  Bug 12078 December 4, 2009
          // adjust buffer pointer correctly for bit values written
          shft = (slen-1)/BYTE_BITS+1;
	}
        break;

      default:
        tmp[0] = '\0';
        sprintf( dm_errbuf,
	         "TEXT Kernel: Unable to handle type %d.", dtype );
        dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
        return dmFAILURE;

    } // end switch


    /* shift data pointer to next value */
    vptr = (char*)vptr + shft; 
    
    /* At this point, we have the value represented as a string  */
    /* according to the appropriate display format..             */
    /*    width = length of that string                          */

    /* Determine how much space we need in buffer (parsebuf) for the value */
    if ( fixed )
    {
      if ( abuf->prowwidth )
      {
	/* column sizes are already determined */
	owidth = abuf->pmaxwidths[kcolndx];
	if ( width > owidth )
	{
	  /* value as string is too long for fixed column space.. truncate */
	  width = owidth;
	}
      }
      else if ( width >= abuf->pmaxwidths[kcolndx] )
      {
	/* column sizes not set and have new maximum for this column        */
	/* flag for alignment, abuf->pmaxwidths will be reset when aligned  */
	/* at transfer to writebuf (afk_buffer_write_parsebuf)              */
	owidth = width + strlen(sep);     /* value as string + 1 separator  */
	abuf->align = UT_TRUE;
	strcpy( abuf->pfmts[kcolndx], cfmt );
      }
      else
      {
	/* use previous maximum size for this column */
	owidth = abuf->pmaxwidths[kcolndx];
      }

      /* Set format string if not already done */
      if ( !abuf->pfmts[kcolndx][0] )
      {
	strcpy( abuf->pfmts[kcolndx], cfmt );
      }

    }
    else
    {
      /* Not fixed */
      owidth = width + strlen(sep);

      /* for non-fixed output, there is no need to align, so we need to */
      /* keep track of the maximum here (for setting up future rows)    */
      if ( owidth > abuf->pmaxwidths[kcolndx] )
	abuf->pmaxwidths[kcolndx] = owidth;
    }

    /* Check how much space is available in the buffer */
    if ( kcolndx < abuf->ndcols-1 )
    {
      /* # chars between next column start and this. */
      avail = abuf->pdata[kcolndx+1] - abuf->pdata[kcolndx];
    }
    else
    {
      /* # chars between parsebuf end and this column start. */
      avail = &abuf->parsebuf[abuf->psize] - abuf->pdata[kcolndx] + 1;
    }


    /* Adjust buffer if needed */
    if ( avail != owidth )
    {
      diff = owidth - avail;
      len  = abuf->psize + 1;

      status = afk_buffer_resize_parsebuf( abuf, len + diff );
      if ( status != dmSUCCESS )
	return vptr;

      /* Shift existing data for subsequent columns */
      nbytes = 0;
      if ( kcolndx < abuf->ndcols - 1 )
      {
	nbytes = &abuf->parsebuf[abuf->psize] - abuf->pdata[kcolndx+1] + 1;
      }
      
      if ( nbytes > 0 )
	memmove( abuf->pdata[kcolndx+1] + diff, abuf->pdata[kcolndx+1], nbytes );
      
      /* Shift column start position pointers too */
      for ( ii = kcolndx+1; ii < abuf->ndcols; ii++ )
      {
	abuf->pdata[ii] += diff;
      }

      abuf->psize += diff;               /* update total size of parsebuf */
    }

    /* Transfer the value as string to parse buffer space*/
    memcpy( abuf->pdata[kcolndx], tmp, width );

    /* set used space for this col.  */
    abuf->pwidths[kcolndx] = width;

    /* Add separator char(s) and/or terminate */
    if ( fixed )
    {
      /* pad remainder of field with separator char(s)      */
      /* NOTE: with fixed format, sep is just a single char */

      memset( abuf->pdata[kcolndx] + width, sep[0], owidth - width );
 
     if ( kcolndx == (abuf->ndcols - 1) )
      {
	/* last column.. terminate string */
	abuf->parsebuf[abuf->psize] = '\0';
      }
    }
    else
    {
      if ( kcolndx < abuf->ndcols-1 )
      {
	/* NOTE: sep may be a single charactor or padded as determined above */
	/*       with non-fixed format, there should be only 1 being written */
	memcpy(abuf->pdata[kcolndx] + width, sep, strlen(sep) );
      }
      else
      {
	abuf->pdata[kcolndx][width] = '\0';
      }
    }

  }/* end kcolndx loop */

  return vptr;
}

/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_read_rows                            */
/*                                                             */
/* Description:                                                */
/*   Transfer data from ascii buffer to DM buffer              */
/*                                                             */
/* Input parameters:                                           */
/*   afBuffer  *abuf   - buffer pointer                        */
/*   long       row1   - start row                             */
/*   long       row2   - end row                               */
/*                                                             */
/* Output parameters:                                          */
/*   dmBuffer  *buffer - DM buffer                             */
/*                                                             */
/* Returns:                                                    */
/*   dmErrCode  status                                         */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_read_rows( afBuffer* abuf, dmBuffer* buffer,
                                long row1, long row2 )
{
  const char    *line     = NULL;

  dmDescriptor  **cols_dd = NULL;   // array of col descriptors
  dmDescriptor  *dd       = NULL;   // individ. col descriptor
  dmDescriptor  *cpt_dd   = NULL;   // individ. component descript
  afCol         *acol     = NULL;   // ascii column descriptor
  void          *vptr     = NULL;   // DM buffer data
  char          *strval   = NULL;   // string form of column values
  afBlock       *ablock   = NULL;   // block ptr from buffer var.
  dmBool        is_image  = dmFALSE;// is the block an image?
  dmBool        append   = dmFALSE; // append line to parse buffer?
  dmBool        goodToGo = dmFALSE; // queue to process parsebuf.

  long cptno       = 0;             // component counter
  long dim         = 0;             // dimensionality
  long size        = 0;             // byte size of data value
  long arlen       = 0;             // array size of component
  long rowno       = 0;             // current row counter
  long nrows       = 0;             // num. of rows to read
  long rrow_offset = 0;             // offset in data buffer to row
  long rrow1       = 0;             // beginning row in data buffer
  long bcolndx     = 0;             // DM buffer column counter
  long nbcols      = 0;             // number of DM buffer columns
  long kcol_start  = 0;             // kernel level column  corr. to 1st value of cmp. col.
  long kcol_end    = 0;             // kernel level column corr. to last value of cmp. col.
  long len         = 0;             // length of string or value
  long slen        = 0;             // length of string or value
  long kcolndx     = 0;             // kernel level column counter

  char sep[10]     = " ";           // separator for appending lines

  dmDataType type;                  // column datatype
  dmErrCode status = dmSUCCESS;     // return status

  vptr    = buffer->data;
  cols_dd = buffer->cols;
  rrow1   = buffer->rrow1;
  nbcols  = buffer->ncols;

  ablock   = abuf->ablock;
  is_image = (ablock->block_type == dmIMAGE);

  if ( is_image )
  {
    rowno = 1;  /* should already be so                */
    nrows = 1;  /* for an image.. row1 = row2 = 1      */
  }
  else
  {
    rowno = row1;
    nrows = row2 - row1 + 1;
  }

  /* set flag so first read will reset the parsebuf */
  append = UT_FALSE;

  do
  {
    /* ********** place proper row(s) at parsebuf pointer ********** */

    /* Get the line */
    line = afk_buffer_getline_ptr( abuf, rowno );

    /* Here we check if it is OK to process the line */
    /* For images, we want to read all lines up to   */
    /* eof or the next block as a single unit prior  */
    /* to processing.                                */
    /* For tables, each line is done individually    */

    if ( is_image )
    {
      /* Check for image block end */
      if ( !line || (*line == '\0'))
      {
	/* end */
	goodToGo = UT_TRUE;
      }
      else if ( cs_uceq_siz(line, "XTENSION", 8) )
      {
	/* hit new block */
	line = NULL;
	goodToGo = UT_TRUE;
      }
      if ( !goodToGo ){ nrows++; } /* to read another line */
    }
    else
    {
      if ( !line )
      {
	sprintf( dm_errbuf, "TEXT Kernel: Failed to load row %ld into the buffer.", rowno );
	dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
	return dmFAILURE;
      }

      /* For tables.. parsebuf should contain only 1 row */
      /* process each one individually.                  */
      goodToGo = UT_TRUE;
    }
      
    /* pop it onto the parsebuf pointer */
    /* for images, this should be building parsebuf with the full image array */
    if ( line )
    {
      /* if appending this line.. add a separator between,   */
      /* basically replacing the newline char with separator */
      /* It may be better to have this within set_parsebuf,  */
      /* but it is not clear that we always want it and at   */
      /* this point, I would like to leave it up to the user */
      if ( append )
      {
	if ( abuf->astyle && abuf->astyle->sep ){ sprintf(sep,"%c", abuf->astyle->sep[0]);}
	afk_buffer_set_parsebuf( abuf, sep, append );
      }

      /* Set line to parsebuf pointer. */
      afk_buffer_set_parsebuf( abuf, line, append );

      if ( is_image )
	append = UT_TRUE;
    }

    if ( goodToGo )
    {
      /* process the parse buffer.. setting pdata[], pwidths[] */
      status = afk_buffer_process_parsebuf( abuf, rowno );
      if (status == dmFAILURE){ return dmFAILURE;}


      /* **********  Transfer Parsebuf info to DM Buffer  ********** */
      // calulate offset to correct row in DM buffer
      rrow_offset = rowno - rrow1;
  
      // loop over DM buffer columns
      for ( bcolndx = 0; bcolndx < nbcols; bcolndx++ )
      {
	// get column descriptor
	dd = cols_dd[bcolndx];

	/* FIXME: What if dd is a hidden column?  Is that possible at this stage?*/
	/* How should we handle it?  (See afk_buffer_write_rows(),               */
	/* dmp_buffer_read_filter_cols(), and dmp_buffer_read_main_cols().)      */

	/* set pointer to start position within DM buffer */
	vptr = (char*)buffer->data + dmi_buffer_offset( buffer, bcolndx, rrow_offset );

	dim = dmi_descriptor_get_vector_dim( dd );

	for ( cptno = 1; cptno <= dim; cptno++ )
	{
	  // get component descriptor
	  cpt_dd = dmi_descriptor_get_cpt( dd, cptno );

	  // get kernel descriptor for component
	  acol   = (afCol*)dmi_descriptor_get_kdd( cpt_dd );
 
	  type   = dmi_descriptor_get_datatype( cpt_dd );
	  size   = ar_mix_type_size( type );

	  // transfer values or skip over space for them in buffer
	  if ( acol ) 
	  {
	    // have kernel descriptor, transfer values

	    // NB: can component descriptor really have different type from kernel
	    //     descriptor?

	    type = acol->datatype;

	    // Code Change Bug 10312 - April 10, 2007
	    // Get correct number of bytes for values

	    if ( type == DT_CHAR )
	    {
	      /* number of chars * size of 1 char */
	      /*     (  #chars + \0  )            */
	      size = (acol->width + 1) *  UT_SZ_C;
	    }
	    else
	    {
              // Code change  Bug 12078 December 4, 2009
              //   correct size for bit type which can be multiple bytes

              if ( type == DT_BIT )
		size = cpt_dd->size;
              else
	        /* # bytes for datatype */
	        size = ar_mix_type_size( type );
	    }

	    /* get kernel level column start/stop for this component */
	    kcol_start = abuf->col_offsets[acol->no-1];

	    arlen = 1;
	    if ( acol->length )
	    {
	      /* one kernel column per array element */
	      arlen = acol->length;
	    }

	    kcol_end = kcol_start + arlen;
	    slen     = acol->width;

	    for ( kcolndx = kcol_start; kcolndx < kcol_end; kcolndx++ )
	    {
	      /* get copy of value for parsing                            */
	      /* Need to isolate the single token for it to work properly */

	      if ( abuf->pwidths[kcolndx] )
		strval = cs_dup_siz(abuf->pdata[kcolndx], abuf->pwidths[kcolndx]);
	      else
		strval = cs_dup_siz("", 1); /* empty string */

	      /* set #chars to use for data value */
	      if ( slen > 0 )
	      {
		/* strings use min of column width or actual value length */
		len = abuf->pwidths[kcolndx] < slen ? abuf->pwidths[kcolndx] : slen;
	      }
	      else
	      {
		len = abuf->pwidths[kcolndx];
	      }

	      /* convert value and put into DM buffer */
	      status = afk_buffer_parse_val( abuf, strval, type, vptr, len, cpt_dd->null );
	      if ( status != dmSUCCESS )
	      {
		sprintf( dm_errbuf, "TEXT Kernel: Unable to parse value %s.", strval );
		dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
		free(strval);
		return dmFAILURE;
	      }

	      /* release duplicated value space */
	      free(strval);

	      /* adjust DM buffer pointer for next value */
	      vptr = (char*)vptr + size;

	    } /* end value loop */
	  }
	  else
	  {
	    // if no kernel descriptor, then skip over space
	    // in DM buffer for this component

	    arlen = dmi_descriptor_get_array_size( cpt_dd );
	    if ( arlen > 0 )
	      vptr = (char*)vptr + size * arlen;
	    else
	      vptr = (char*)vptr + size;
	  }

	} /* end component loop */
      } /* end column loop */
  
      goodToGo = UT_FALSE;

    } /* end if goodToGo */

    /* next row */
    rowno++;

  } while ( rowno < row1 + nrows );

  return dmSUCCESS;
}

/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_set_fixed_col_widths                 */
/*                                                             */
/* Description:                                                */
/*              For fixed width data.  Set the max width of a  */
/*              column using acol->fwidth and the data width   */
/*              and the data width of the field to acol->width */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                    long     **pwidths                       */
/*                    long     **pmaxwidths                    */
/*                    long      *nwidths                       */
/*                    afCol     *acol                          */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            nothin                                           */
/*                                                             */
/***************************************************************/
void afk_buffer_set_fixed_col_widths( afBuffer* abuf, long** pwidths,
                                      long **pmaxwidths, long* nwidths,
                                      afCol* acol )
{
  long bcolndx = 0;
  long maxndx = 0;
  long width = 0;
  long new_nwidths = 0;
  long length = 0;

  if ( acol->length > 0 )
    length = acol->length;
  else
    length = 1;

  bcolndx = abuf->col_offsets[acol->no-1];
  new_nwidths = bcolndx + length;

  if ( new_nwidths >= *nwidths )
  {
    *nwidths = new_nwidths;
    *pmaxwidths = realloc( *pmaxwidths, sizeof(long)*(new_nwidths) );
    *pwidths = realloc( *pwidths, sizeof(long)*(new_nwidths) );
  }

  if ( length == 1 )
  {
    (*pmaxwidths)[bcolndx] = acol->fwidth;
    (*pwidths)[bcolndx] = acol->width;
  }
  else
  {
    maxndx = bcolndx + length - 1;
    width = acol->fwidth / length;

    for ( ; bcolndx <= maxndx; bcolndx++ )
    {
      (*pmaxwidths)[bcolndx] = width;
      (*pwidths)[bcolndx] = acol->width;
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_get_col_byteoffset                   */
/*                                                             */
/* Description:                                                */
/*              Get column offset                              */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                    long       colno - column number         */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            long  byte_offset                                */
/*                                                             */
/***************************************************************/
long afk_buffer_get_col_byteoffset( afBuffer* abuf, long colno )
{
  long bcolndx = 0;
  long maxndx = 0;
  long byte_offset = 1;

  maxndx = abuf->col_offsets[colno-1]-1;

  for ( bcolndx=0; bcolndx <= maxndx; bcolndx++ )
  {
    byte_offset += abuf->pmaxwidths[bcolndx];
  }

  return byte_offset;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_get_col_maxwidth                     */
/*                                                             */
/* Description:                                                */
/*              Get column width                               */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                    long       colno                         */
/*                    long       ar_length                     */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            long  width                                      */
/*                                                             */
/***************************************************************/
long afk_buffer_get_col_maxwidth( afBuffer* abuf, long colno, long ar_length )
{
  long bcolndx = 0;
  long width = 0;

  bcolndx = abuf->col_offsets[colno-1];
  width = abuf->pmaxwidths[bcolndx];

  if ( ar_length > 1 )
  {
    width *= ar_length;
  }

  return width;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_get_col_cfmt                         */
/*                                                             */
/* Description:                                                */
/*              Get column format                              */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBuffer  *abuf - buffer pointer         */
/*                    long       colno                         */
/*                                                             */
/*  Output parameters:                                         */
/*                    char      *cfmt                          */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_buffer_get_col_cfmt( afBuffer* abuf, long colno, char* cfmt )
{
  long bcolndx = 0;

  bcolndx = abuf->col_offsets[colno-1];
  strcpy( cfmt, abuf->pfmts[bcolndx] );
}



/***************************************************************/
/*                                                             */
/* Function:   afk_buffer_align                                */
/*                                                             */
/* Description:                                                */
/*   (re)align all columns for fixed width output.             */
/*                                                             */
/*  Input parameters:                                          */
/*    afBuffer  *abuf   - buffer pointer                       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*    dmErrCode  status                                        */
/*                                                             */
/***************************************************************/
dmErrCode afk_buffer_align( afBuffer* abuf )
{
  long  ii,jj;
  long  arlen;
  long  moved;
  long  nbytes;
  long  maxwidth;
  long  curwidth;
  long  rowno;
  long  nrows         = 0;
  long  new_rowlen    = 0;
  long  old_rowlen    = 0;
  long *new_maxwidths = NULL;

  const char  *iptr = NULL;
  char        *optr = NULL;
  char        *sep  = NULL;

  afFile *afptr = NULL;

  dmErrCode status = dmSUCCESS;

  new_maxwidths = (long*)calloc( abuf->ndcols, sizeof(long) );
  if ( !new_maxwidths ) { return dmFAILURE; }

  /* Determine the new maxwidths array from current parsebuf definition   */
  for ( ii = 0; ii < abuf->ncols; ii++ )
  { 
    maxwidth = 0;
    curwidth = 0;

    /* get array length of column (1=scalar) */
    if ( ii == (abuf->ncols - 1) )
      arlen = abuf->ndcols - abuf->col_offsets[ii];
    else
      arlen = abuf->col_offsets[ii+1] - abuf->col_offsets[ii];

    /* loop over array elements, find largest width. */
    for (jj = abuf->col_offsets[ii]; jj < abuf->col_offsets[ii] + arlen; jj++ )
    {
      /* get size of current record in parsebuf */
      if ( jj == (abuf->ndcols - 1) )
	curwidth = &abuf->parsebuf[abuf->psize] - abuf->pdata[jj] + 1;
      else
	curwidth = abuf->pdata[jj+1] - abuf->pdata[jj];

      /* Check against current maximum */
      if ( abuf->pmaxwidths[jj] > maxwidth ){ maxwidth = abuf->pmaxwidths[jj]; }
      if ( curwidth             > maxwidth ){ maxwidth = curwidth; }
    }

    /* loop over array elements again, set new maximum width */
    for (jj = abuf->col_offsets[ii]; jj < abuf->col_offsets[ii] + arlen; jj++ )
    {
      new_maxwidths[jj] = maxwidth;
      new_rowlen += maxwidth;
    }
  }

  /* Check if the writebuf contains the parsebuf row(s) */
  /* we need to account for that space later.           */
  nrows = 0;
  for ( rowno = abuf->prow1; rowno <= abuf->prow2; rowno++)
  {
    if ( (rowno >= abuf->wrow1)&&(rowno <= abuf->wrow2) )
    {
      nrows++;
    }
  }

  /* Flush the write buffer.. any records there follow the old alignment. */
  afk_buffer_flush_writebuf( abuf, &moved );


  /* ---------------------------------------------------------- */
  /*                                                            */
  /*     Adjust all existing records to the new alignment       */
  /*                                                            */
  /* ---------------------------------------------------------- */

  sep = abuf->astyle->sep;      


  /* We're going to use the writebuf space to align each row... */
  /* make sure it is big enough to handle the new alignment.    */
  status = afk_buffer_resize_writebuf( abuf, new_rowlen, 0 );

  /* move to first row */
  rowno = afk_buffer_moveto_row( abuf, 1 );

  afptr = afk_buffer_get_fptr( abuf );

  /* set output pointer to beginning of write buffer */
  abuf->wrow1 = rowno;
  optr = abuf->writebuf;


  /* loop through rows in file.. and update the alignment */
  while ( (iptr = afk_file_getline_ptr( afptr )) != NULL )
  {
    if ( *iptr == '\0' ){ continue; /*blank*/}
    old_rowlen = strlen(iptr) + 1;

    if ( ((abuf->wsize + new_rowlen) > abuf->wmax) ||
	 ((abuf->wrow2 - abuf->wrow1 + 1) > abuf->wmaxrows) )
    {
      /* write buffer is full.. flush it */
      afk_buffer_flush_writebuf( abuf, &moved );

      /* This sets the file pointer back to the beginning of the current */
      /* line.. end of last line held in writebuf.  Move up...           */
      afk_file_getline_ptr( afptr );
	 
      abuf->wrow1 = rowno;
      optr = abuf->writebuf;
    }

    /* clean space */
    memset( optr, sep[0], sizeof(char)*new_rowlen );

    /* Set reference pointer to this row */
    abuf->wrows[rowno - abuf->wrow1] = optr;

    /* Transfer values to writebuf according to new spacing.  */
    /* MCD NOTE: This part can be modified to do fewer copies */
    /* if you scan across columns where the alignment         */
    /* is unchanged.                                          */
    for ( ii = 0; ii < abuf->ndcols; ii++ )
    {
      if ( ii == (abuf->ndcols - 1) )
	nbytes = abuf->pmaxwidths[ii] - 1; /* not terminator  */
      else
	nbytes = abuf->pmaxwidths[ii];

      memcpy(optr, iptr, nbytes);
      iptr += abuf->pmaxwidths[ii];
      optr += new_maxwidths[ii];
    }

    /* end this row */
    optr--;  *optr = '\n';
    optr++;  *optr = '\0';

    /* update writebuf size fields */
    abuf->wsize  += new_rowlen;
    abuf->wrsize += old_rowlen;
    abuf->wrow2 = rowno;

    rowno++;
  }

  if ( abuf->wsize > 0 )
  {
    /* write buffer is full.. flush it */
    afk_buffer_flush_writebuf( abuf, &moved );
  }
   
  /* reset the write buffer */
  abuf->writebuf[0] = '\0';
  abuf->wrow1 = 0;
  abuf->wrow2 = 0;
  abuf->wsize = 0;
  abuf->wrsize = nrows * new_rowlen;

  /* ---------------------------------------------------------- */
  /*                                                            */
  /*     All existing records now have the new alignment        */
  /*                                                            */
  /*   The writebuf is empty, but if any parsebuf row(s) are    */
  /*   in the file, that space is accounted for at abuf->wrsize */
  /*   This does assume that the parsebuf record will be        */
  /*   sent to the writebuf, but since this function is         */
  /*   being called from the write_parsebuf routine, this seems */
  /*   like a safe bet.  Loading those rows back into the       */
  /*   writebuf would be more thorough/complete, but also more  */
  /*   time consuming.                                          */
  /*                                                            */
  /* ---------------------------------------------------------- */

  /* One last pass through columns, to update pmaxwidths */
  for ( ii = 0; ii < abuf->ndcols; ii++ )
  {
    abuf->pmaxwidths[ii] = new_maxwidths[ii];
  }

  abuf->align = UT_FALSE;

  free(new_maxwidths);

  return status;
}





/***************************************************************/
/*  Function:   afk_buffer_getline_ptr                         */
/*                                                             */
/*  Description:                                               */
/*    Gets a specific row from the ascii kernel buffer.        */
/*                                                             */
/*  Input parameters:                                          */
/*    afBuffer   *abuf      - buffer pointer                   */
/*    long        rowno     - row to read                      */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*    const char* line      - pointer to the row               */
/*                            NULL if not found                */
/*                                                             */
/***************************************************************/
const char* afk_buffer_getline_ptr( afBuffer* abuf, long rowno )
{
  const char* line   = NULL;
  afFile     *afptr  = NULL;
  long        rowgot = 0;
  
  if ( rowno >= abuf->prow1 && rowno <= abuf->prow2 )
  {
    /* I have the row */
    line = abuf->parsebuf;
  }
  else if ( rowno >= abuf->wrow1 && rowno <= abuf->wrow2 )
  {
    /* The row is in the write buffer */
    line = abuf->wrows[rowno - abuf->wrow1];
  }
  else
  {
    /* Have to find the row */
    rowgot = afk_buffer_moveto_row( abuf, rowno );
    if ( rowgot == rowno )
    {
      /* found it */
      afptr = afk_buffer_get_fptr( abuf );

      /* Pull the line from the file.. empty ones don't count */
      line = afk_file_getline_ptr( afptr );

      while( line && *line == '\0' )
	line = afk_file_getline_ptr( afptr );

    }
  }

return line;
}
