/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afkcol.c c                                                  */
/*                                                               */
/*   Functions that work with columns                            */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"



/***************************************************************/
/*                                                             */
/* Function:   afk_col_alloc                                   */
/*                                                             */
/* Description:                                                */
/*             Allocate column buffer                          */
/*                                                             */
/*  Input parameters:                                          */
/*                    none                                     */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            afCol  *acol - column pointer                    */
/*                                                             */
/***************************************************************/
afCol* afk_col_alloc( void )
{
  afCol  *acol = NULL;


  acol = (afCol*)malloc( sizeof(afCol) );
  acol->type = FTK_COL;
  acol->no = 0;
  acol->width = 0;
  acol->name = NULL;
  acol->length = 0;
  acol->datatype = DT_UNK;
  acol->mtype_index = 0;
  acol->bytepos = 0;
  acol->disp = NULL;
  acol->fwidth = 0;

  return acol;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_col_alloc_copy                              */
/*                                                             */
/* Description:                                                */
/*              Make a copy of the column information          */
/*                                                             */
/*  Input parameters:                                          */
/*                    const afCol  *acol                       */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            afCol  *copy_col  - pointer to copy column       */
/*                                                             */
/***************************************************************/
afCol* afk_col_alloc_copy( const afCol* acol )
{
  afCol *racol = NULL;


  racol = (afCol*)malloc( sizeof(afCol) );

  racol->type = FTK_COL;
  racol->no = acol->no;
  racol->width = acol->width;

  if ( acol->name )
    racol->name = cs_dup( acol->name );
  else
    racol->name = NULL;

  racol->length = acol->length;
  racol->datatype = acol->datatype;
  racol->mtype_index = acol->mtype_index;
  /* racol->bytepos = acol->bytepos; */
  racol->bytepos = 0; /* bytepos shouldn't get propagated */
  racol->fwidth = acol->fwidth;

  if ( acol->disp )
    racol->disp = cs_dup( acol->disp );
  else
    racol->disp = NULL;

  return racol;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_col_free                                    */
/*                                                             */
/* Description:                                                */
/*              Free column space                              */
/*                                                             */
/*  Input parameters:                                          */
/*                    afCol  *acol  - column pointer           */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_col_free( afCol* acol )
{
  if ( acol )
  {
    if (acol->name)
    {
      free( acol->name );
      acol->name = NULL;
    }

    if ( acol->disp )
    {
      free( acol->disp );
      acol->disp = NULL;
    }

    free( acol );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   afk_col_print                                   */
/*                                                             */
/* Description:                                                */
/*              Print column information                       */
/*                                                             */
/*  Input parameters:                                          */
/*                    afCol  *acol                             */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            afCol  *copy_col                                 */
/*                                                             */
/***************************************************************/
void afk_col_print( afCol* acol )
{
  if ( acol )
  {
    printf( "Col %ld: %s\n", acol->no, acol->name );

    if ( acol->width )
      printf( "\tWidth: %ld\n", acol->width );

    if ( acol->length )
      printf( "\tArray length: %ld\n", acol->length );

    printf( "\tType: " );

    if ( acol->datatype == DT_REAL8 )
    {
      printf( "DT_REAL8" );
    }
    else
      if ( acol->datatype == DT_CHAR )
      {
        printf( "DT_CHAR" );
      }
      else
        if ( acol->datatype == DT_INT4 )
        {
          printf( "DT_INT4" );
        }
        else
        {
          printf( "Unknown: %ld", acol->type );
        }

    printf( "\n" );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   afk_vector_alloc                                */
/*                                                             */
/* Description:                                                */
/*              Create vector column                           */
/*                                                             */
/*  Input parameters:                                          */
/*                    long dim  - vector dimension             */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            afVector  *avec                                  */
/*                                                             */
/***************************************************************/
afVector* afk_vector_alloc( long dim )
{
  afVector *avec = NULL;

  avec = calloc( 1, sizeof(afVector) );
  avec->cpts = calloc( dim, sizeof(afCol*) );
  avec->type = AFK_VECTOR;

  return avec;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_vector_free                                 */
/*                                                             */
/* Description:                                                */
/*              Free vector space                              */
/*                                                             */
/*  Input parameters:                                          */
/*                    afVector  *avec                          */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_vector_free( afVector* avec )
{
  if (avec)
  {
    if ( avec->cpts )
    {
      free( avec->cpts );
      avec->cpts = NULL;
    }

    free( avec );
  }
}
