/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afkcscrdb.c                                                 */
/*                                                               */
/*   Functions that create and work with ASCII file type         */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"
#include "dmi.h"



/***************************************************************/
/*                                                             */
/*  Function:  afk_cscrdb_attach_units()                       */
/*                                                             */
/*  Description:                                               */
/*     Reads the unit names line from keylist and puts the     */
/*     unit names for each column in each raw col descriptor   */ 
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock        *block                    */
/*                    dmDescriptor  **rawcols                  */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*           dmErrCode  status                                 */
/*                                                             */
/*  Called from:                                               */
/*         afp_simple_make_vectors()                           */
/*                                                             */
/***************************************************************/
dmErrCode afk_cscrdb_attach_units( dmBlock *block,
                                   dmDescriptor **rawcols,
                                   long numColumns )
{
  dmErrCode     status = dmSUCCESS;
  dmuKeyList    keylist;
  dmuKey       *key = NULL;
  dmDescriptor *pColDescriptor = NULL;

  afBlock  *pakBlock = NULL;

  char   unitName[AK_COMMENTLINE_LEN];
  char  *pTokenStart = NULL;
  char  *pUnit       = NULL;
  char   cNull       = '\0';


  long   numKeys    = 0;
  long   lenToken   = 0;
  long   ii         = 0;


  // get kernel block
  pakBlock = (afBlock*)dmi_block_get_kb( block );

  // get the keylist
  keylist = pakBlock->keylist;
  numKeys = keylist->n;

  // watch out for empty keylist
  if ( numKeys < 2 )
    return (status);

  // get units key, which is the next to last key in the list ( -2)
  key = keylist->data[numKeys-2];

  // get units key line value
  pUnit = dmu_key_value_c( key );

  if (pUnit)
  {
    // get unit name for each column from key value string

    for (ii = 0; ii < numColumns; ii++)
    {
       pTokenStart = pUnit;
       lenToken = 0;

       // find unit name in string
       while (pUnit && (*pUnit != '\t' && *pUnit != '\r'))
       {
         ++lenToken;
         ++pUnit;
       }

       // don't overrun the buffer on copy
       if ( lenToken >= AK_COMMENTLINE_LEN)
         lenToken = AK_COMMENTLINE_LEN - 1;

      // break apart unit name from value line
       if (lenToken > 0 )
         strncpy(unitName, pTokenStart, lenToken);

       // add terminator
       unitName[lenToken] = cNull;

       // move starting pointer for next pass
       ++pUnit; 

       // update each column descriptor with unit string
       //  if no unit, skip
       
       pColDescriptor = rawcols[ii];

       if ( pColDescriptor && unitName[0] != cNull)
	 dmi_descriptor_set_unit(pColDescriptor, unitName);
    }
  }

  return (status);
}

/*****************************************************************************/
/*                                                                           */
/*  Function:  afk_cscrdb_init_cols()                                        */
/*                                                                           */
/*  Description:                                                             */
/*     Get column information and type                                       */
/*     This function is based on afk_simple_init_cols()                      */
/*                                                                           */
/*     Calls afk_simple_read_cols()                                          */
/*                                                                           */
/*  Input parameters:                                                        */
/*                    afBlock  *ablock                                       */
/*                    afStyle  *astyle                                       */
/*                                                                           */
/*  Output Parameters:  dmErrCode status                                     */
/*                                                                           */
/*****************************************************************************/
dmErrCode afk_cscrdb_init_cols( afBlock* ablock, afStyle* astyle )
{
  dmuKey       *key      = NULL;
  dmuKeyList    keylist;
  dmErrCode     status   = dmSUCCESS;
  dmDataType    type     = DT_UNK;

  afCol        *acol  = NULL;
  afFile       *afptr = NULL;
  afDataset    *ads   = NULL;

  const char*   cols_line = NULL;
  char          colval[AK_COLVAL_LEN];
  const char   *rptr         = NULL;
  long          ncols        = 0;
  long          ii           = 0;
  long          jj           = 0;


  keylist = afk_block_keylist( ablock );


  if (astyle->colnames == AK_STYLE_COLNAMES_LM1)
  {
    key = keylist->data[keylist->n-2];
    cols_line = dmu_key_value_c( key ); 
  }
  else
    if ( astyle->colnames == AK_STYLE_COLNAMES_LM2 )
    {
      key = keylist->data[keylist->n-3];
      cols_line = dmu_key_value_c( key ); 
    }

  if ( cols_line )
  {
    status = afk_simple_read_cols( ablock, astyle, cols_line );

    if ( status != dmSUCCESS )
      return status;
  }

  // Code Change  Bug 11791  February 25, 2010
  //   Now need to get the column types from the first data row

  ads   = afk_block_get_kds( ablock );
  afptr = afk_dataset_fptr( ads );
  ncols = afk_block_ncols( ablock );

  afk_block_set_row( ablock, 1 );

  
  /* loop over all rows to get column data type */

  while ( (rptr = afk_file_getline_ptr( afptr )) )
  {
    for ( ii = 0; ii < ncols && rptr; ii++ )
    {
      acol = afk_block_get_col( ablock, ii+1 );
      rptr = afk_gen_read_next_col( astyle, rptr, colval, AK_COLVAL_LEN );

      afk_gen_determine_col_datatype( astyle, acol, colval, &type);

      if ( rptr )
      {
        /* skip over array elements */

        for ( jj = 1; jj < acol->length; jj++ )
        {
          rptr = afk_gen_read_next_col( astyle, rptr, colval, AK_COLVAL_LEN );

          afk_gen_determine_col_datatype( astyle, acol, colval, &type);
        }
      }
    }
  }

  /* final rectification of column types */
  afk_gen_final_col_datatype( ablock, astyle );

  /* reset to start */
  afk_block_set_row( ablock, 1 );

 return status;
}


/***************************************************************/
/*                                                             */
/*  Function:  afk_cscrdb_is_dashline()                        */
/*                                                             */
/* Description:                                                */
/*    checks comment line value for dash line properties       */
/*                                                             */
/*  Input parameters:                                          */
/*                    const char     *comment                  */
/*                                                             */
/*  Output parameters:  none                                   */
/*                                                             */
/*  Returns:                                                   */
/*            dmBool true or false                             */
/*                                                             */
/*  Called by:                                                 */
/*       afk_cscrdb_read_header()                              */
/*       afk_cscrdb_write_comment()                            */
/*       afp_cscrdb_dataset_scan()                             */
/*                                                             */
/***************************************************************/
dmBool afk_cscrdb_is_dashline( const char *comment, 
                               const char *sSeparator )
{
  char  sLine[AK_COMMENTLINE_LEN];
  char *pComment = sLine;
  char *pToken   = NULL;
  char *pcLineLookPtr = NULL;
  char *pcTokenLookPtr = NULL;
  char  cNull       = '\0';
  char  cDash       = '-';

  dmBool bDashLine = dmFALSE;


  // copy string
  strncpy(pComment, comment,AK_COMMENTLINE_LEN);
  pComment[AK_COMMENTLINE_LEN-1] = cNull;
  pcLineLookPtr = pComment;

  if (pcLineLookPtr && *pcLineLookPtr != cNull)
  {
    // get first token
    pToken = strtok(pcLineLookPtr, sSeparator);

    pcTokenLookPtr = pToken;  
    bDashLine = dmFALSE;

    // scan name; if not dash line, store it as comment

    while ( pcTokenLookPtr )
    {
      while ( pcTokenLookPtr[0] != cNull )
      {
        if ( pcTokenLookPtr[0] == cDash )
        {
          pcTokenLookPtr++;
          bDashLine = dmTRUE;
        }
        else
        {
          bDashLine = dmFALSE;
          break;
        }
      }

      // any non-dash character encounter means token is comment
      if ( !bDashLine )
        break;

      // only dashes - scan next token in this line
      pToken = strtok(NULL, sSeparator);
      pcTokenLookPtr = pToken; 
    }
  }

  return ( bDashLine);
}


/*****************************************************************************/
/*                                                                           */
/*  Function:  afk_cscrdb_key_write()                                        */
/*                                                                           */
/*  Description:                                                             */
/*     Write key information                                                 */
/*     code based on afk_simple_write_key()                                  */
/*                                                                           */
/*  Input parameters:                                                        */
/*                    afBlock        *pakBlock                               */
/*                    long            lKeyPosition                           */
/*                    const char     *sKeyName                               */
/*                    dmDataType      dataType                               */
/*                    const void     *pValue                                 */
/*                    const char     *sUnit                                  */
/*                    const char     *sDescription                           */
/*                                                                           */
/*  Output parameters:  dmuKey* key                                          */
/*                                                                           */
/*  Called by:                                                               */
/*         afp_block_key_write()                                             */
/*                                                                           */
/*****************************************************************************/
dmuKey* afk_cscrdb_key_write( afBlock* pakBlock, long lKeyPosition,
                              const char* sKeyName,
                              dmDataType dataType, const void* pValue,
                              const char* sUnit, const char* sDescription )
{
  afStyle    *pakStyle = NULL;
  dmuKeyList  keyList  = pakBlock->keylist;
  long        firstKey = 0; /* first valid new key position, need to keep colnames line*/
  long        lastKey  = 0; /* last valid new key position, need to keep colnames line*/


  pakStyle = pakBlock->ads->astyle;

  // default positions
  firstKey = 1;
  lastKey  = keyList->n;
                      
  // adjust position so that header information appears before column names

  if ( pakStyle->colnames == AK_STYLE_COLNAMES_LM1 )
    lastKey = keyList->n - 1;
  else
    if ( pakStyle->colnames == AK_STYLE_COLNAMES_LM2 )
      lastKey = keyList->n - 2;

  // verify range
  if ( lKeyPosition < firstKey )
    lKeyPosition = firstKey;
  else
    if ( lKeyPosition > lastKey )
      lKeyPosition = lastKey;

  // write key
  return afk_raw_key_write( pakBlock, lKeyPosition, sKeyName, dataType, pValue, sUnit, sDescription );
}


/***************************************************************/
/*                                                             */
/* Function:   afk_cscrdb_read_header                          */
/*                                                             */
/* Description:                                                */
/*             We expect the file already to be pointing at    */
/*             the beginning of the block.                     */
/*                                                             */
/*  Input parameters:                                          */
/*                   afBlock  *pakBlock - block pointer        */
/*                   afStyle  *pakStyle - style pointer        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/*  Called by:                                                 */
/*         afp_block_read_header()                             */
/*                                                             */
/***************************************************************/
dmErrCode afk_cscrdb_read_header( afBlock* pakBlock, afStyle* pakStyle )
{
  char keyname[AK_KEYNAME_LEN];
  char comment[AK_KEYCOMMENT_LEN];

  dmuKeyList keylist;
  dmDataType type   = DT_CHAR;
  dmBool     mask   = dmTRUE;
  dmuKey    *key    = NULL;
  ftKeyType  ktype  = FTK_COMMENT;
  dmErrCode  status = dmSUCCESS;
  dmBool     bDashLine = UT_FALSE;
  dmBool     bBadHeader = UT_FALSE;

  afBuffer  *pakBuffer  = NULL;
  afFile    *pakFile    = NULL;
  afDataset *pakDataset = NULL;

  const char *pcLine    = NULL;
  char   cHeaderChar    = '\0';
  char   cDefaultHeadChar = '#';
  char  *pcLineLookPtr  = NULL;
  char  *sSeparator     = NULL;
  char  *sDefaultSep    = "\t\r";

  GenericType tmp;
  void  *parsed_value = &tmp;

  long pos     = 0;
  long posDash = 0;
  long posEndHeader = 0; 
  long histlen = 7;
  long commlen = 7;
  int  ii      = 0;


  pakDataset = afk_block_get_kds( pakBlock );
  pakBuffer  = afk_block_get_buffer( pakBlock );
  pakFile    = afk_dataset_fptr( pakDataset );
  keylist    = dmu_keylist_alloc();
  posDash    = pakDataset->pos;

  /* might need to delete a previous keylist? */
  pakBlock->keylist = keylist;

  // Get header char and separators in case user changed
  // from default values
  // If a value is null, use default value

  cHeaderChar = pakStyle->header_char;

  if ( cHeaderChar == '\0')
    cHeaderChar = cDefaultHeadChar;

  if ( pakStyle->sep != NULL )
    sSeparator = cs_dup( pakStyle->sep );
  else
    sSeparator = cs_dup( sDefaultSep );


  // read the first line
  pos    = afk_file_tell( pakFile );
  pcLine = afk_file_getline_ptr( pakFile );

  // skip blank lines
  while ( pcLine && *pcLine == '\0' )
    pcLine = afk_file_getline_ptr( pakFile );


  // process as a comment any line that starts with a header character

  while ( pcLine && *pcLine == cHeaderChar )
  {
    mask = dmFALSE;

    afk_buffer_add_keyline( pakBuffer, pcLine );

    // store the comment line as the key's value
    afk_gen_read_comment( pakStyle, pcLine, comment, AK_KEYCOMMENT_LEN );

    /*
      MCD NOTE: This handling is from the DTF flavor.  Simple/Raw, do 
       a similar check for HISTORY/COMMENT, but do not offset the string
       to remove the tag. (which I think will need to change there too).
    */
      if ( cs_eq_siz( comment, dmHISTORY, histlen ))
      {
        strcpy( keyname, dmHISTORY );
        *(char**)parsed_value = cs_dup( comment + histlen + 1 );
      }
      else if ( cs_eq_siz( comment, dmCOMMENT, commlen ))
      {
        strcpy( keyname, dmCOMMENT );
        *(char**)parsed_value = cs_dup( comment + commlen + 1 );
      }
      else
     {
      strcpy ( keyname, dmCOMMENT );
      *(char**)parsed_value = cs_dup( comment );
      }


    comment[0] = '\0';
    mask = UT_TRUE;

    // add a key to the list
    key = dmu_key_alloc();
 
    dmu_key_set( key, (dmuKeyType)ktype, type, keyname, parsed_value, NULL,
	               comment, mask );
    dmu_key_insert( keylist, key, 0 );

    // clean up
    if ( type == DT_CHAR )
      free( *(char**)parsed_value );

    // read the next comment line
    pos    = afk_file_tell( pakFile );
    pcLine = afk_file_getline_ptr( pakFile );

    // skip blank lines
    while( pcLine && *pcLine == '\0' )
    {
      pos    = afk_file_tell( pakFile );
      pcLine = afk_file_getline_ptr( pakFile );
    }
  }

  // Finished reading comment lines
  // Next 3 lines will contain column names, unit names, and a dash line
  // that separates header from data. Unit names line is optional. Dash
  // line is always last line before data. Allow for blank column names.
  // Any file without column names is processed as "raw"

  pakStyle->colnames = AK_STYLE_COLNAMES_NONE;

  if ( !pcLine )
    return status; /* no data in file */


  // Get first non-comment line
  pcLineLookPtr = (char*)pcLine;

  /* Mark the end of the comment section in case file is non-conforming           */
  /* ??? MCD NOTE:  I think this position is at the end of the line following the */
  /*                comment segment.  But the intent is that it is the beginning  */
  /*                of that line.  I think 'pos' is the value wanted.             */

  posEndHeader = afk_file_tell( pakFile );


  /* The next 2-3 lines should contain the column specification info             */
  /*   - line containing column names (separator delimited)                      */
  /*   - optional line containing units for each column (separator delimited)    */
  /*   - line of 'dash' characters.. which ends the column specification segment.*/
  /*                                                                             */
  /* NOTE: at this point, pcLine should contain the first of these lines.        */

  for ( ii = 0; ii < 3; ii++)
  {
    bDashLine = afk_cscrdb_is_dashline( pcLineLookPtr, sSeparator);

    // store line as comment
    //    afk_file_seek( pakFile, pos, SEEK_SET );
    // pcLine = afk_file_getline_ptr( pakFile );

    // be sure that there is data at this point and not EOF
    if ( !pcLine )
    {
      bBadHeader = UT_TRUE;
      break;
    }

    /* add line to ascii buffer */
    afk_buffer_add_keyline( pakBuffer, pcLine );

    /* process the line as a comment ??? */
    afk_gen_read_comment( pakStyle, pcLine, comment, AK_KEYCOMMENT_LEN );

    /* assign keyname with line content as value */
    /* MCD: Changed to assign an actual key name for these colspec lines */
    /*      masked as private..                                          */
    sprintf( keyname,"_COLSPEC%1d", ii );
    *(char**)parsed_value = cs_dup( comment );

    /* reset comment variable for next line. */ 
    comment[0] = '\0';

   /* Add the key to dm keylist */
    ktype = FTK_STRING;
    mask  = dmTRUE;

    key = dmu_key_alloc();
    dmu_key_set( key, (dmuKeyType)ktype, type, keyname, parsed_value, NULL,
                 comment, mask );

    dmu_key_insert( keylist, key, 0 );

    // clean up
    if ( type == DT_CHAR )
      free( *(char**)parsed_value );

    // if the line just handled was not the dash, it is either a column
    // name line or unit name line. change style struct accordingly
    // unit names line always follows column names line
    // colname style value indicates if only column names or column names
    // plus unit names

   if ( ! bDashLine )
   {
      if ( ii == 0 )
        pakStyle->colnames = AK_STYLE_COLNAMES_LM1;
      else
        if ( ii == 1 )
          pakStyle->colnames = AK_STYLE_COLNAMES_LM2;
   }
   else
   {
      /* Read dashed line.. done reading column spec. segment */
       break;
   }

    // read the next line
    pos    = afk_file_tell( pakFile );
    pcLine = afk_file_getline_ptr( pakFile );

    // skip blank lines
    while( pcLine && *pcLine == '\0' )
    {
      pos = afk_file_tell( pakFile );
      pcLine = afk_file_getline_ptr( pakFile );
    }
  }

  // check for read err
  if ( bBadHeader )
  {
    // move to position after last comment
    // file has bad cscrdb format
    afk_file_seek( pakFile, posEndHeader, SEEK_SET );    
  }

  // set data start position
  pos = afk_file_tell( pakFile );
  afk_buffer_set_datastart( pakBuffer, pos );

  // record in global hint structure the colname style
  // when doing a cscrdb to cscrdb copy, this information is necessary
  afp_set_colnames_hint(pakStyle->colnames, dmTRUE);
  
  // housekeeping
  free (sSeparator );

  return status;

}

/***************************************************************/
/*                                                             */
/*  Function:  afk_cscrdb_write_comment()                      */
/*                                                             */
/* Description:                                                */
/*    puts comments into header buffer                         */
/*    code based on  afk_simple_write_comment()                */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBlock        *pakBlock                 */
/*                    long            pos                      */
/*                    const char     *tag                      */
/*                    const char     *comment                  */
/*                                                             */
/*  Output parameters:  none                                   */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/*  Called by:                                                 */
/*         afk_cscrdb_write_comment()                          */
/*                                                             */
/***************************************************************/
dmErrCode afk_cscrdb_write_comment( afBlock* pakBlock,
                                    long pos, const char* tag,
                                    const char* comment )
{
  char  cbuf[AK_COMMENTLINE_LEN];
  char  prefix[4];
  char *sSeparator = NULL;
  char *sDefaultSep = "\t\r";
  char  cHeaderChar = '\0';
  char  cDefaultHeadChar = '#';
  char  cNull  = '\0';
  char  cBlank = ' ';

  dmuKeyList keylist;
  dmErrCode  status    = dmSUCCESS;
  dmuKey    *key       = NULL;

  afBuffer  *pakBuffer = NULL;
  afStyle   *pakStyle  = NULL;
  ftKeyType  fktype    = FTK_UNKNOWN;

  long       mask   = UT_TRUE;
  long       keyno  = 0;
  long       keyOffset = -1;
  long       numKeys   = 0;


  keylist   = pakBlock->keylist;
  numKeys   = keylist->n;
  pakBuffer = afk_block_get_buffer( pakBlock );
  fktype    = FTK_COMMENT;

  // Get header char and separator
  // if either value is null, use default

  pakStyle = pakBlock->ads->astyle;
  cHeaderChar = afk_style_get_header_char( pakStyle );

  if ( cHeaderChar == '\0' )
    cHeaderChar = cDefaultHeadChar;

  if ( pakStyle->sep != NULL )
    sSeparator = cs_dup( pakStyle->sep );
  else
    sSeparator = cs_dup( sDefaultSep );

  // determine fudge factor for keylist having 2 or 3 keys already entered in it
  if ( pakStyle->colnames == AK_STYLE_COLNAMES_LM1 )
    keyOffset = -2;
  else
    if ( pakStyle->colnames == AK_STYLE_COLNAMES_LM2 )
      keyOffset = -3;



  // HISTORY goes after the comments and before the name/units/dash entries
  // adjust position accordingly

  if ( ! strcmp(tag,"HISTORY" ))
    keyno = keylist->n+keyOffset+1;
  else
  {
    for ( keyno = pos+keyOffset+1; keyno <= keylist->n; keyno++ ) /* was keyno=pos+1 */
    {
      key = keylist->data[keyno-1];
 
      if ( key->keytype != FTK_COMMENT )
        break;
    }

    // COMMENTS go before name/unit/dash entries, adjust accordingly
    keyno += keyOffset;
  }

  // don't put out header char on dash line
  //  bDashLine = afk_cscrdb_is_dashline(comment, sSeparator );

 
  // set up comment line
  //  if ( !bDashLine )
  //  {
    key = dmu_key_alloc();
    key->type = FTK_KEY;
    dmu_key_insert( keylist, key, keyno );
    dmu_key_set( key, fktype, DT_CHAR, tag, &comment, NULL, NULL, mask );

    prefix[0] = cHeaderChar;
    prefix[1] = cBlank; //sSeparator[0];
    prefix[2] = cNull;

   if ( strcmp( tag, AK_SPECIFICATION) == 0)
     sprintf (cbuf, "%s", comment);
   else
   {
      if ( strcmp( tag, dmCOMMENT ) == 0 )
        sprintf( cbuf, "%s%s\n", prefix, comment );
      else
        sprintf( cbuf, "%s%s%c%s\n", prefix, tag, cBlank, comment );
   }

    status = afk_buffer_insert_keyline( pakBuffer, cbuf, keyno );
 
    if ( status != dmSUCCESS )
    {
      dmu_key_delete( keylist, key );
      key = NULL;
    }
    //  }


  // housekeeping
  free (sSeparator);

  return status;
}


