/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afdataset.c                                                 */
/*                                                               */
/*   Functions that create and work with ASCIII type files       */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"

static dmErrCode afk_dataset_open_fcreate( afDataset* ads, const char* filename );



/***************************************************************/
/*                                                             */
/* Function:   afp_dataset_alloc                               */
/*                                                             */
/* Description:                                                */
/*              Allocates dataset structure                    */
/*                                                             */
/*  Input parameters:                                          */
/*                    none                                     */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            afDataset  *DatasetPointer                       */
/*                                                             */
/***************************************************************/
afDataset* afk_dataset_alloc( void )
{
  afDataset *ads = NULL;

  ads = malloc( sizeof(afDataset) );

  if ( ads )
  {
   ads->name = NULL;
   ads->n = 0;
   ads->nmax = 0;
   ads->blocks = NULL;
   ads->afptr = NULL;
   ads->astyle = afk_style_alloc();
   ads->blocks_pos = NULL;
   ads->pos = 0;
  }

  return ads;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dataset_open                                */
/*                                                             */
/* Description:                                                */
/*              Open dataset                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                   afDataset  *ads       - dataset pointer   */
/*                   const char *filename  - file pointer      */
/*                   dmBool      mode      - fcreate?          */
/*                   dmBool      create    - create?           */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode   Status                               */
/*                                                             */
/***************************************************************/
dmErrCode afk_dataset_open( afDataset* ads, const char* filename,
                            dmBool mode, dmBool create )
{
  if ( ads->afptr ) 
    afk_dataset_close( ads, dmFALSE ); 

  if ( !mode && cs_uceq_siz( "FCREATE(", filename, 8 ))
    afk_dataset_open_fcreate( ads, filename );
  else 
    ads->afptr = afk_file_open( filename, mode, create );

  if ( afk_file_okay( ads->afptr ))
    return dmSUCCESS;

 return dmFAILURE; 
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dataset_open_fcreate                        */
/*                                                             */
/* Description:                                                */
/*              Open dataset from fcreate parameters           */
/*                                                             */
/*  Input parameters:                                          */
/*                   afDataset  *ads    - dataset pointer      */
/*                   const char *filename   - file pointer     */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode   Status                               */
/*                                                             */
/***************************************************************/
dmErrCode afk_dataset_open_fcreate( afDataset* ads, const char* filename )
{
  afFile  *afptr      = NULL;
  afFile  *cols_afptr = NULL;
  afFile  *data_afptr = NULL;
  afStyle *astyle     = NULL;

  char    keyline[80];
  char    tdisp_val[16];
  char    flavorBuffer[CARD_SIZE];
  char   *fcreate_fnames[3];
  const char *line_ptr = NULL;
  char   *col_line     = NULL;
  char   *coldef[5];
  char   *cptr1        = NULL;
  char   *cptr2        = NULL;
  char   *tmpstring    = NULL;

  long   colno  = 0;
  long   arsize = 0;
  long   length = 0;
  long   width  = 0;
  int     flavorBufferLength = 0;

  dmDataType  type;
  dmBool      isvar   = UT_FALSE;
  dmBool      isfixed = UT_FALSE;


  cptr1 = strchr( filename, '(' );

  if ( !cptr1 )
    return dmFAILURE;

  ++cptr1;
  cptr2 = strchr( cptr1, ')' );

  /* allow escaping right paren */
  while( cptr2 && *(cptr2-1) == '\\' )
    cptr2 = strchr( cptr2, ')' );

  if ( !cptr2 )
    return dmFAILURE;
 
  tmpstring = cs_dup_siz( cptr1, cptr2-cptr1 );
  fcreate_fnames[0] = strtok( tmpstring, "," );
  fcreate_fnames[1] = strtok( NULL, "," );
  fcreate_fnames[2] = strtok( NULL, "," );

  /* load the header file */
  if ( fcreate_fnames[1] )
    afptr = afk_file_load( fcreate_fnames[1] );

  if ( !afk_file_okay( afptr ))
  {
    if ( fcreate_fnames[1] )
      sprintf( dm_errbuf, "Error, TEXT kernel failed to open header file: %s", fcreate_fnames[1] );
    else
      sprintf( dm_errbuf, "Error, TEXT kernel failed to open header file." );
    dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
  }

  /* load the cols file and create the corrrect keywords */
  colno = 0;

  if ( fcreate_fnames[0] )
    cols_afptr = afk_file_open( fcreate_fnames[0], UT_FALSE, UT_FALSE );
  if ( !afk_file_okay( cols_afptr ))
  {
    if ( fcreate_fnames[0] )
      sprintf( dm_errbuf, "Error, TEXT kernel failed to open column definition file: %s", fcreate_fnames[0] );
    else
      sprintf( dm_errbuf, "Error, TEXT kernel failed to open column definition file." );
    dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
    afk_file_close( afptr );
    afptr = NULL;
  }

  if ( afk_file_okay( afptr ))
  {
    afk_file_seek( afptr, 0, SEEK_END );
    while( NULL != ( line_ptr = afk_file_getline_ptr( cols_afptr ) ) )
    {
      col_line = cs_dup( line_ptr );
      coldef[0] = strtok( col_line, " \t" );
      coldef[1] = strtok( NULL, " \t" );

      if ( coldef[0] && coldef[1] )
      {
         ++colno;

         sprintf( keyline, "TTYPE%ld = %s", colno, coldef[0] );
         afk_file_puts( afptr, keyline );
         sprintf( keyline, "TFORM%ld = %s", colno, coldef[1] );
         afk_file_puts( afptr, keyline );
         coldef[2] = strtok( NULL, " \t" );

         if ( coldef[2] )
         {
           if ( !cs_uceq( coldef[2], "NONE" ))
           {
             sprintf( keyline, "TUNIT%ld = %s", colno, coldef[2] );
             afk_file_puts( afptr, keyline );
           }

           coldef[3] = strtok( NULL, " \t" );
           coldef[4] = strtok( NULL, " \t" );

           if ( coldef[3] && coldef[4] )
           {
             sprintf( keyline, "TBCOL%ld = %s", colno, coldef[3] );
             afk_file_puts( afptr, keyline );
      
             type = afp_parse_tform( coldef[1], UT_FALSE, &arsize, &length,
			             &width, &isvar );
      
             width = atoi( coldef[4] );
             afp_compose_tdisp( type, width, tdisp_val );
             sprintf( keyline, "TDISP%ld = %s", colno, tdisp_val );
             afk_file_puts( afptr, keyline );
             isfixed = UT_TRUE;
           }
           else
             isfixed = UT_FALSE;
         }
      }
      free( col_line );
    }
  }

  afk_file_close( cols_afptr );
  afk_file_puts( afptr, "END" );
  afk_file_seek( afptr, 0, SEEK_SET );

  /* Get Flavor string and write to file */
  astyle = afk_style_alloc();
  afk_style_set_fixed( astyle, isfixed );
  afk_style_set_flavor( astyle, AK_FLAV_DTF, NULL );
  strcpy( flavorBuffer, afk_style_get_flavorString(astyle) );

  strcat (flavorBuffer, "\n");
  flavorBufferLength = strlen(flavorBuffer);

  if ( isfixed )
    afk_file_replace_data(afptr, flavorBuffer, flavorBufferLength, 0 );
  else
    afk_file_replace_data( afptr, flavorBuffer, flavorBufferLength, 0 );

  if ( fcreate_fnames[2] )
    data_afptr = afk_file_load( fcreate_fnames[2] );

  if ( !afk_file_okay( data_afptr ))
  {
    if ( fcreate_fnames[2] )
      sprintf( dm_errbuf, "Error, TEXT kernel failed to open data file: %s", fcreate_fnames[2] );
    else
      sprintf( dm_errbuf, "Error, TEXT kernel failed to open data file." );

    dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
    afk_file_close( afptr );
    afptr = NULL;
  }

  if ( afk_file_okay( afptr ))
  {
    afk_file_seek( afptr, 0, SEEK_END );
    afk_file_replace_data( afptr, data_afptr->membuf, data_afptr->bufsize, 0 );
  
    afk_file_seek( afptr, 0, SEEK_SET );
  }

  afk_file_close( data_afptr );

  free( tmpstring );
  afk_style_free(astyle);

  ads->afptr = afptr;
 
  if ( !afk_file_okay( ads->afptr ))
    return dmFAILURE;
 
  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dataset_fptrs                               */
/*                                                             */
/* Description:                                                */
/*             Returns file pointer for dataset                */
/*                                                             */
/*  Input parameters:                                          */
/*                   afDataset  *ads  - dataset pointer        */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            afFile    *FilePointer                           */
/*                                                             */
/***************************************************************/
afFile* afk_dataset_fptr( const afDataset* ads )
{
  if ( ads )
  {
    return ads->afptr;
  }

 return NULL;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dataset_close                               */
/*                                                             */
/* Description:                                                */
/*              Close dataset                                  */
/*                                                             */
/*  Input parameters:                                          */
/*                   afDataset  *ads      - dataset pointer    */
/*                   dmBool      delete                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_dataset_close( afDataset* ads, dmBool delete )
{
  char *comstr = NULL;
  long  len    = 0;

  if ( ads && ads->afptr )
  {
    afk_file_close( ads->afptr );
    ads->afptr = NULL;

    if ( delete )
    {
      len = strlen(ads->name)+1;
      len += 6;

      comstr = (char*)malloc(len*sizeof(char));
      sprintf( comstr, "rm -f %s", ads->name );
   
      system(comstr); /* ryan, need to replace with something safer than system */
      free(comstr);
    }
  } 
}



/***************************************************************/
/*                                                             */
/* Function:   afk_dtaset_free                                 */
/*                                                             */
/* Description:                                                */
/*              free dataset storage                           */
/*                                                             */
/*  Input parameters:                                          */
/*                   afDataset  *ads- pointer to ascii dataset */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_dataset_free( afDataset *ads )
{
  long i = 0;

  if ( ads )
  {
    if ( ads->name )
    {
      free( ads->name );
      ads->name = NULL;
    }

    if ( ads->blocks )
    {
      for ( i=0; i<ads->nmax; i++ )
      {
        if ( ads->blocks[i] )
          free( ads->blocks[i] );
      }

      free( ads->blocks );
      ads->blocks = NULL;
    }

    if ( ads->blocks_pos )
    {
      free( ads->blocks_pos );
      ads->blocks_pos = NULL;
    }

    if ( ads->astyle )
    {
      afk_style_free( ads->astyle );
      ads->astyle = NULL;
    }

    free( ads );
  }
}



/***************************************************************/
/*                                                             */
/* Function:   afk_dataset_block_name                          */
/*                                                             */
/* Description:                                                */
/*              Returns dataset block name                     */
/*                                                             */
/*  Input parameters:                                          */
/*                   const afDataset *ads  - dataset pointer   */
/*                   long             no   - block number      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            const char  *DatasetName                         */
/*                                                             */
/***************************************************************/
const char* afk_dataset_block_name( const afDataset* ads, long no )
{
  char *name = NULL;

  if ( ads )
  {
    if ( no <= ads->n )
      name = ads->blocks[no-1];
  }

  return name;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dataset_resize_blocks                       */
/*                                                             */
/* Description:                                                */
/*             Make sure there are enough blocks to hold at    */
/*             least a minimum block indices                   */
/*                                                             */
/*  Input parameters:                                          */
/*                   afDataset  *ads  - dataset pointer        */
/*                   long        no   - number of indices      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode   Status                               */
/*                                                             */
/***************************************************************/
dmErrCode afk_dataset_resize_blocks( afDataset* ads, long no )
{
  long oldmax = 0;
  long incr   = 10;

  if ( !ads )
    return dmFAILURE;

  if ( no > ads->nmax )
  {
    oldmax = ads->nmax;
    ads->nmax += incr;
    ads->blocks = (char**)realloc( ads->blocks, 
				 sizeof(char*) * ads->nmax );
    if ( !ads->blocks )
      return dmFAILURE;

   memset( ads->blocks + oldmax, 0, incr * sizeof(char*) );

   ads->blocks_pos = (long*)realloc( ads->blocks_pos,
				      sizeof(long) * ads->nmax );
   if ( !ads->blocks_pos )
     return dmFAILURE;

   memset( ads->blocks_pos + oldmax, 0, incr * sizeof(long) );
  }

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dataset_set_flavour                         */
/*                                                             */
/* Description:                                                */
/*              Set style structure's flavor element           */
/*                                                             */
/*  Input parameters:                                          */
/*                   afDataset  *ads   - dataset pointer       */
/*                   afFlavour   flav  - kernel spec type      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_dataset_set_flavour( afDataset* ads, afFlavour flav )
{
  afStyle *astyle = NULL;

  if ( ads )
  {
    astyle = ads->astyle;
    astyle->flavor = flav;
    afk_style_set( astyle, flav );
  }
}



/***************************************************************/
/*                                                             */
/* Function:   afp_dataset_get_flavour                         */
/*                                                             */
/* Description:                                                */
/*              Returns style structure's flavor element       */
/*                                                             */
/*  Input parameters:                                          */
/*                   afDataset  *ads  - Dataset pointer        */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            afFlavour  Flavors                               */
/*                                                             */
/***************************************************************/
afFlavour afk_dataset_get_flavour( afDataset* ads )
{
  afStyle *astyle = NULL;

  if ( ads )
  {
    astyle = ads->astyle;
    return astyle->flavor;
  }
  else
    return AK_FLAV_UNK;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_dataset_get_style                           */
/*                                                             */
/* Description:                                                */
/*              Returns pointer to style structure             */
/*                                                             */
/*  Input parameters:                                          */
/*                   const afDataset *ads  - dataset pointer   */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            afStyle  *StylePointer                           */
/*                                                             */
/***************************************************************/
afStyle* afk_dataset_get_style( const afDataset* ads )
{
  if ( ads )
    return ads->astyle;
  else
    return NULL;
}



/***************************************************************/
/*                                                             */
/* Function:   afkp_dataset_moveto_block                        */
/*                                                             */
/* Description:                                                */
/*              Move to a block in a dataset                   */
/*                                                             */
/*  Input parameters:                                          */
/*                   afDataset  *ads  - dataset pointer        */
/*                   long        no   - block number           */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode   Status                               */
/*                                                             */
/***************************************************************/
dmErrCode afk_dataset_moveto_block( afDataset* ads, long no )
{
  afFile *afptr = NULL;

  afptr = afk_dataset_fptr( ads );

  if ( no > ads->n || no < 1 )
    return dmFAILURE;

  return afk_file_seek( afptr, ads->blocks_pos[no-1], SEEK_SET );
}
