/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afkdft.c                                                    */
/*                                                               */
/*   Functions that create and work with dtf type files          */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"
#include "../ftkernel/ftkernelP.h"

/* local functions */
static afCol* afk_dtf_col_read( afBlock* ablock, long colno, afCol* acol );
static void afk_dtf_guess_coltypes( afBlock* ablock, afStyle* astyle );
static dmErrCode afk_dtf_compose_keyline( const char* name, const void* value, dmDataType type, const char* unit, const char* desc, char** keyline, long klen, afStyle* astyle );

static int afp_key_continue( char** pptr, const char* text, const char* desc );
static int afp_key_ends_in_amp( const char* value );


/***************************************************************/
/*                                                             */
/* Function:   afk_dtf_read_header                             */
/*                                                             */
/* Description:                                                */
/*              The file is already pointing at the beginning  */
/*              of the block.                                  */
/*                                                             */ 
/*              This can be merged with afk_raw_read_header    */
/*              to create afk_gen_read_header when I can get   */
/*              around the difference between '#' style headers*/
/*              and "END" style headers.                       */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock   *block                         */
/*                    afStyle   *astyle                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode   Status                               */
/*                                                             */
/***************************************************************/
dmErrCode afk_dtf_read_header( afBlock* ablock, afStyle* astyle )
{
  char keyname[AK_KEYNAME_LEN];
  char keyvalue[AK_KEYVAL_LEN];
  char comment[AK_KEYCOMMENT_LEN];
  char unit[AK_KEYNAME_LEN];

  dmuKey     *key = NULL;
  dmuKeyList  keylist;
  dmuKeyType  ktype  = FTK_UNKNOWN;
  dmDataType  dtype  = DT_UNK;
  dmBool      iskey  = dmFALSE;
  dmBool      mask   = dmFALSE;
  dmErrCode   status = dmSUCCESS;

  afBuffer   *abuf  = NULL;        /* Ascii buffers      */
  afDataset  *ads   = NULL;        /* Ascii dataset      */
  afFile     *afptr = NULL;        /* Ascii file pointer */

  GenericType  tmp;
  void        *parsed_value = &tmp;
  const char  *line     = NULL;
  const char  *pline    = NULL;

  long         pos = 0;
  const char   endstr[] = "END";
  dmBool       endofhdr = UT_FALSE;
  dmBool       haveNextLine = UT_FALSE;
 
  long  histlen = strlen(dmHISTORY); 
  long  commlen = strlen(dmCOMMENT);


  abuf  = afk_block_get_buffer( ablock );
  ads   = afk_block_get_kds( ablock );
  afptr = afk_dataset_fptr( ads );


  /* Allocate keyword list for the block */
  keylist = dmu_keylist_alloc();
  ablock->keylist = keylist;

  /* Get next non-empty line */
  pos = afk_file_tell( afptr );
  line = afk_file_getline_ptr( afptr );

  while( line && !endofhdr )
  {
    mask = dmFALSE;
    haveNextLine = UT_FALSE;

    /* Add line to ascii buffer */
    afk_buffer_add_keyline( abuf, line );

    if ( cs_eq( line, endstr ) )
    {
      /* END key found.. done */
      endofhdr = UT_TRUE;
      continue;
    }

    /* Split line into keyword components if we can */
    iskey = ( afk_gen_read_keyline( astyle, line,
				    keyname, AK_KEYNAME_LEN,
				    keyvalue, AK_KEYVAL_LEN,
				    comment, AK_KEYCOMMENT_LEN,
                                    unit, AK_KEYNAME_LEN ) == dmSUCCESS );
    if ( iskey )   
    {
      mask  = afk_key_private( keyname );

      dtype = afk_key_get_type( keyvalue, AK_KEYVAL_LEN );
      ktype = afk_datatype_to_keytype( dtype );
      afk_key_parse_val( &keyvalue[0], dtype, parsed_value );

      /* Check for long keyword.. CONTINUE key handling */
      if ( ktype == FTK_STRING )
      {
	/* Set flag for continue key search */
	dmBool done = UT_FALSE;

	while ( !done )
	{
	  if ( afp_key_ends_in_amp( *(char**)parsed_value ) )
	  {
	    /* Get next line */
	    pos = afk_file_tell( afptr );
	    line = afk_file_getline_ptr( afptr );
	    if ( cs_eq_siz( line, "CONTINUE  ", 10 ))
	    {
	      /* Add line to buffer and append value */
	      afk_buffer_add_keyline( abuf, line );
	      afp_key_continue( (char**)parsed_value, line+10, comment );
	    }
	    else
	    {
	      /* Set flag to NOT read the next line.. already have it */
	      haveNextLine = UT_TRUE;

	      /* Following keyword is NOT a CONTINUE. nothing else to append.. go on. */
	      done = UT_TRUE;
	    }
	  }
	  else
	  {
	    /* Not a continued line */
	    done = UT_TRUE;
	  }
	} // end while
      } // end if STRING
    }
    else /* Read it as a comment instead */
    {
      mask = UT_TRUE;

      afk_gen_read_comment( astyle, line, comment, AK_KEYCOMMENT_LEN );
      dtype = DT_CHAR;
      ktype = FTK_COMMENT;

      if ( cs_eq_siz( comment, dmHISTORY, histlen ))
      {
        strcpy( keyname, dmHISTORY );
        *(char**)parsed_value = cs_dup( comment + histlen + 1 );
      }
      else if ( cs_eq_siz( comment, dmCOMMENT, commlen ))
      {
	strcpy( keyname, dmCOMMENT );
	*(char**)parsed_value = cs_dup( comment + commlen + 1 );
      }
      else
      {
	strcpy ( keyname, dmCOMMENT );
	*(char**)parsed_value = cs_dup( comment );
      }

      comment[0] = '\0';
    }

    /* Allocate and set keyword descriptor */
    key = dmu_key_alloc();
    dmu_key_set( key, ktype, dtype, keyname, parsed_value, unit, comment, mask );

    dmu_key_insert( keylist, key, 0 );

    if ( dtype == DT_CHAR )
      free( *(char**)parsed_value );

    /* Get next non-empty line */
    /* May have already read it in CONTINUE segement. */
    if ( !haveNextLine )
    {
      do
      {
	pos = afk_file_tell( afptr );
	line = afk_file_getline_ptr( afptr );
      } while( line && *line == '\0' );
    }

  } /* end while header line reading */

  /* We've now read the header.  Now we need to find the data. */
  /* Move past blank lines */
  do
  {
    pos  = afk_file_tell( afptr );
    line = afk_file_getline_ptr( afptr );

    if ( ! line )
    {
      /* no data in file */
      afk_buffer_set_datastart( abuf, pos );
      return status;
    }

    /* get rid of leading separator chars */
    pline = line;
    while ( *pline != '\0' && afk_style_is_sep( astyle, *pline ))
    {
      ++pline;
    }

  } while ( *pline == '\0' );

  /* Set current file position in buffer structure */
  afk_buffer_set_datastart( abuf, pos );

  return status;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_dtf_init_cols                               */
/*                                                             */
/* Description:                                                */
/*              Initialize column names                        */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBlock   *ablock                        */
/*                    afStyle   *astyle                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode   Status                               */
/*                                                             */
/***************************************************************/
dmErrCode afk_dtf_init_cols( afBlock* ablock, afStyle* astyle )
{
  long  ncols    = 0;
  long  colno    = 0;
  long  maxcolno = 0;

  afCol  *acol    = NULL;
  afCol  *nextcol = NULL;

  dmuKeyList keylist;
  dmErrCode  status    = dmSUCCESS;
  dmuKey    *akey       = NULL;
  dmBool     have_types = dmTRUE;


  keylist = afk_block_keylist( ablock );

  if (keylist)
  {
    akey = dmu_keylist_read_typed_value( keylist, "TFIELDS", DT_INT4,
                                         (void*)&ncols );
  }

  if ( akey )
  {
    for ( colno = 1; colno <= ncols; colno++ )
    {
      acol = afk_col_alloc();
      afk_dtf_col_read( ablock, colno, acol );
      afk_block_append_col( ablock, acol );
    }
  }
  else
  {
    if ( !astyle->fixed )
    {
      afk_raw_init_cols( ablock, astyle );
      ncols = afk_block_ncols( ablock );

      for ( colno = 1; colno <= ncols; colno++ )
      {
        acol = afk_block_get_col( ablock, colno );
        afk_dtf_col_read( ablock, colno, acol );
      }
    }
    else
    {
      colno = 1;

      while( NULL != ( acol = afk_dtf_col_read( ablock, colno, NULL ) ) )
      {
        afk_block_append_col( ablock, acol );
        ++colno;
      }
    }
  }


  ncols = afk_block_ncols( ablock );

  /* make sure all columns have types */
  for ( colno = 1; colno <= ncols && have_types; colno++ )
  {
    acol = afk_block_get_col( ablock, colno );

    if ( acol->datatype == DT_UNK )
      have_types = dmFALSE;
  }

  if ( !have_types )
  {
    afk_dtf_guess_coltypes( ablock, astyle );
  }

  /* If fixed, use TBCOL to determine width of the field and make */
  /* sure all columns have a byteposition and a width             */

  if ( astyle->fixed )
  {
    maxcolno = ncols - 1;
    nextcol = NULL;

    for ( colno = 1; colno <= maxcolno; colno++ )
    {
      acol = afk_block_get_col( ablock, colno );
      nextcol = afk_block_get_col( ablock, colno+1 );

      if ( !acol->width || !acol->bytepos )
      {
        printf( "DM EMERGENCY Error with bytepositions / widths\n" );
        break;
      }
      acol->fwidth = nextcol->bytepos - acol->bytepos;
    }

    if ( nextcol )
    {
      if ( !nextcol->width || !nextcol->bytepos )
      {
        printf( "DM EMERGENCY Error in last byteposition / width\n" );
      }
      /* Last column.. can only use display width, but need to */
      /* include the terminator.                               */
      nextcol->fwidth = nextcol->width + 1;
    }
  }
  
  return status;
}

/***************************************************************/
/*                                                             */
/* Function:   afk_dtf_init_image_cols                         */
/*                                                             */
/* Description:                                                */
/*    Use the kernel block information to initialize           */
/*    the block column structure for image blocks              */
/*    Added for Bug 9023 - ascii image support                 */
/*    NOTE: Keywords for image structure are created later.    */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock *block - pointer to block        */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  ErrorCode                             */
/*                                                             */
/***************************************************************/
dmErrCode afk_dtf_init_image_cols( afBlock* ablock, afStyle* astyle )
{
  dmuKeyList    keylist;
  dmuKey       *akey   = NULL;
  afCol        *acol   = NULL;

  char          sNaxis[CARD_SIZE]; 
  char          ttype[CARD_SIZE];
  long          naxes   = 2;
  long          axlen   = 0;
  int           ii      = 0;

  // Get keylist values for NAXIS
  keylist = afk_block_keylist( ablock );

  if (keylist)
  {
    akey = dmu_keylist_read_typed_value( keylist, "NAXIS", DT_INT4, (void*)&naxes );
  }
  if ( akey )
  {
    /* NOTE: null PRIMARY defines naxes as 0 */
   ablock->naxes = naxes;
  }
  else
  {
     ablock->naxes = 2;
     naxes = 2;
  }

  for (ii = 0; ii < naxes; ii++)
  {
    // read array dimensions

    sprintf(sNaxis, "NAXIS%d", (ii+1) );
    akey = dmu_keylist_read_typed_value( keylist, sNaxis, DT_INT4, &axlen );

    if ( akey )
    {
      // set axis
      ablock->axes[ii] = axlen;
    }
    else
    {
      // Key does not exist. Use number of columns in first data row
      // for NAXIS1. Use number of rows in file for Naxis2
      // Value not defined for axis 3 or higher.

      if ( !ii )
        axlen = afk_dtf_imag_get_ncols(ablock, astyle);
      else
        axlen = afk_dtf_get_nrows(ablock);

      ablock->axes[ii] = axlen;
    }
  }

  /* Get IMAGE name.. if keyword not yet set, try block name.. */
  ttype[0] = '\0';
  if (keylist)
    akey = dmu_keylist_read_value_c( keylist, "EXTNAME", ttype, CARD_SIZE );

  if (!akey && ablock->name )
    strcpy( ttype, ablock->name );

  if (ttype[0] == '\0' || ttype[0] == ' ')
    strcpy( ttype, "COL1");

  /* allocate column */
  acol = afk_col_alloc();

  if (acol)
  {
    acol->name = cs_dup( ttype );

    /* adjust length */
    if ( naxes )
    {
      acol->length = 1;

      for (ii = 0; ii < naxes; ii++)
	acol->length *= ablock->axes[ii];
    }
  }
  

  afk_block_append_col( ablock, acol );

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_dtf_guess_coltypes                          */
/*                                                             */
/* Description:                                                */
/*             evaluate column type and sets datatype element  */
/*             in the afCol structure contained in the block.  */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBlock   *ablock                        */
/*                    afStyle   *astyle                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*           nothing                                           */
/*                                                             */
/***************************************************************/
void afk_dtf_guess_coltypes( afBlock* ablock, afStyle* astyle )
{
  char value[AK_COLVAL_LEN];
  const char* line  = NULL;
  const char* pline = NULL;

  afCol       *acol  = NULL;  /* Ascii column descriptor  */
  afFile      *afptr = NULL;  /* Ascii file pointer       */
  afDataset   *ads   = NULL;  /* Ascii dataset descriptor */

  long colUnknowns[99];       /* list of col numbers with unknown type */
  long nUnknowns = 0;         /* number of cols with unknown type      */
  long colno = 1;             /* column number                         */
  long ncols = 0;             /* number of columns                     */
  long ii    = 0;
  long jj    = 0;
  long uu    = 0;

  dmDataType dtype = DT_UNK;


  ncols = afk_block_ncols( ablock );

  /* there is at least one column that doesn't have type info */ 
  /*  which ones? Loop over columns and store column number   */
  /*  of any column that has unknown datatype                 */

  for ( ii = 1; ii <= ncols; ii++)
  {
    acol = afk_block_get_col(ablock, ii);

    if (acol->datatype == DT_UNK )
      colUnknowns[nUnknowns++] = ii;
  }

  /* set block to first row, get file and dataset pointers */

  afk_block_set_row( ablock, 1 );
  ads   = afk_block_get_kds( ablock );
  afptr = afk_dataset_fptr( ads );


  // Code Change  Bug 11791  February 25, 2010
  // try to determine datatype from the data
  // loop over rows

  while ( (line  = afk_file_getline_ptr( afptr )) )
  {
    /* uu indexes unknown column numbers in colUnknowns */
    uu = 0;

    /* local pointer to line */
    pline = line;

    /* loop over columns looking at the unknown columns */

    for ( colno = 1; (colno <= ncols) && (uu < nUnknowns) && line; colno++)
    {
      pline = afk_gen_read_next_col( astyle, pline, value, AK_COLVAL_LEN );

      /* is this a column of interest? */

      if ( colno == colUnknowns[uu] )
      {
        /* yes, this is an unknown data type column                    */
        /* increment index now so that next time it will look for next */
        /* unknown column                                              */
        ++uu;

        /* get column descriptor */
        acol = afk_block_get_col( ablock, colno );

        /* Set descriptor type */
        afk_gen_determine_col_datatype( astyle, acol, value, &dtype);


        /* is this an array */
        if ( acol->length )
        {
	  /* need to check data for all array elements */

	  for ( jj = 1; (jj <= acol->length-1) && pline; jj++)
          {
	    /* increment the column number */
            ++colno;

            pline = afk_gen_read_next_col( astyle, pline, value, AK_COLVAL_LEN );
              
            /* Set descriptor type */
            afk_gen_determine_col_datatype( astyle, acol, value, &dtype);
	  }
        }
      }
    } /* end for */
  } /* end while */
}



/***************************************************************/
/*                                                             */
/* Function:   afk_dtf_col_read                                */
/*                                                             */
/* Description:                                                */
/*              Read column                                    */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBlock   *ablock                        */
/*                    long       colno                         */
/*                    afCol     *acol                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode   Status                               */
/*                                                             */
/***************************************************************/
afCol* afk_dtf_col_read( afBlock* ablock, long colno, afCol* acol )
{
  char keyword[CARD_SIZE];
  char keyval[CARD_SIZE];
  char ttype[CARD_SIZE];

  dmuKeyList keylist;
  dmuKey    *akey = NULL;

  afDataset  *ads    = NULL;
  afStyle    *astyle = NULL;
  long       *axes   = NULL;

  long width     = 0;
  long precision = 0;
  long exponent  = 0;
  long naxes     = 0;
  long size      = 0;
  long i         = 0;
  long length    = 0;
  long bytepos   = 0;
  long arlen     = 0;
  long dispWidth = 0;

  dmBool varlen     = dmFALSE;
  dmBool bHaveTFORM = dmFALSE;
  dmDataType type   = DT_UNK;

  char typecode = '\0';


  keylist = afk_block_keylist( ablock );

  if ( acol )
    type = acol->datatype;
  else
    type = DT_UNK;



  /* read column type and such - TFORM */
  sprintf( keyword, "TFORM%ld", colno );

  if (keylist)
    akey = dmu_keylist_read_value_c( keylist, keyword, keyval, CARD_SIZE );

  if ( akey )
  {
    bHaveTFORM = dmTRUE;

    if ( !acol )
      acol = afk_col_alloc();

    type = afp_parse_tform( keyval, dmFALSE, &size, &length, &width, &varlen );
  }
  else
  {
    /* if no TFORM  specified, give default length based on type */

    if ( type == DT_CHAR )
    {
      size   = 1;
      length = AK_COLVAL_LEN;
      width  = AK_COLVAL_LEN;
    }
    else
    {
      size   = 0;
      length = 0;
      width  = 0;
    }
  }


  /* read column dimensionality - TDIM */
  sprintf( keyword, "TDIM%ld", colno );
 
  if (keylist)
    akey = dmu_keylist_read_value_c( keylist, keyword, keyval, CARD_SIZE );

  if ( akey )
  {
    if ( !acol ) 
      acol = afk_col_alloc();

    ftp_parse_tdim( keyval, &axes, &naxes );

    /* Code change - Bug 10042  November 3, 2006             */
    /* For DT_CHAR types only - special case                 */
    /*   Adjust variable "length" based on TDIM keyword      */
    /*   values to imitate action based on older format      */

    if ( type == DT_CHAR )
    {
      if ( naxes > 0 && axes[0] > 0 )
      {
         size    = length / axes[0];
         length  = axes[0];
         width   = length;
         axes[0] = size;
         naxes   = 1;
      }
    }
  }
  else
    /* No TDIM statement                               */
    /*   If column is array set up axes/naxes pair     */
    /*   If column is type is char, set up the         */
    /*   axex/naxes pairs as well so as to no loose    */
    /*   the size information which is needed for      */
    /*   correct buffer allocation and dm buffer access*/
    /*   Code Change Bug 10312 April 10, 2007          */
 
    if ( (type == DT_CHAR && !naxes) ||
         (size > 1) )
    {
      naxes   = 1;
      axes    = malloc( sizeof(long) );
      axes[0] = size;
    }

  /* end Code Changes Bugs 10312, 10042 */


  /* This section calculates the array length */
  if ( naxes )
  {
    arlen = 1;

    for ( i = 0; i < naxes; i++ )
    {
      arlen *= axes[i];
    }

    free( axes );
    axes = NULL;
  }


  /* Get column name - TTYPE */
  sprintf( keyword, "TTYPE%ld", colno );
  if (keylist)
    akey = dmu_keylist_read_value_c( keylist, keyword, ttype, CARD_SIZE );

  if ( !akey )
  {
    if ( !acol || !acol->name )
      sprintf( ttype, "COL%ld", colno );
    else
      ttype[0] = '\0';
  }
  else
    if ( !acol )
      acol = afk_col_alloc();


  /* If fixed width data, read TBCOL to get data alignment */
  ads    = ablock->ads;
  astyle = afk_dataset_get_style( ads );

  if ( astyle->fixed )
  {
    sprintf( keyword, "TBCOL%ld", colno );
    if (keylist)
      akey = dmu_keylist_read_typed_value( keylist, keyword, DT_INT4,
                                           &bytepos );

    if ( akey && !acol )
      acol = afk_col_alloc();
  }


  /* get display parameters, TDISP */

  sprintf( keyword, "TDISP%ld", colno );

  if (keylist)
    akey = dmu_keylist_read_value_c( keylist, keyword, keyval, CARD_SIZE );

  if ( akey )
  {
    if ( !acol )
      acol = afk_col_alloc();

    afp_parse_tdisp( keyval, &typecode, &dispWidth, &precision, &exponent );

    acol->disp = cs_dup( keyval );

    // If data type is unknown, get it from the TDISP parameter value

    if ( type == DT_UNK )
    {
      type = afk_tdisp_code_to_datatype( typecode );
    }

    // Code Change Bug 10408 April 17, 2007
    // If there is no TFORM given, use TDISP to define column size

    if ( ! bHaveTFORM )
    {
      if ( type == DT_CHAR )
          arlen = 1;

      if ( arlen )
        width = dispWidth * arlen;
      else
        width = dispWidth;
    }

    if (astyle->fixed && !width)
      width = dispWidth;

    // end Code Change
  }


  /* Set final values for ascii column descriptor */

  if ( acol )
  {
    if ( acol->name )
      free( acol->name );

    if ( ttype[0] != '\0')
      acol->name = cs_dup( ttype );
    else
      acol->name = NULL;

    acol->datatype = type;
    acol->length   = arlen;
    acol->width    = width;
    acol->bytepos  = bytepos;
  }

  return acol;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_dtf_key_write                               */
/*                                                             */
/* Description:                                                */
/*              Write key                                      */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBlock    *ablock                       */
/*                    long        pos                          */
/*                    const char *name                         */
/*                    dmDataType  type                         */
/*                    const void *value                        */
/*                    const char *unit                         */
/*                    const char *desc                         */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmuKey     *ukey                                 */
/*                                                             */
/***************************************************************/
dmuKey* afk_dtf_key_write( afBlock* ablock, long pos, const char* name,
                           dmDataType type, const void* value,
                           const char* unit, const char* desc )
{
  char       *keyline = NULL;

  dmuKeyList keylist;
  dmuKey     *ukey     = NULL;
  ftKeyType  fktype    = FTK_UNKNOWN; 

  afBuffer  *abuf      = NULL;
  afStyle   *astyle    = NULL;
  long       old_entry = 0;
  dmBool     mask      = dmFALSE;
  dmErrCode  status    =dmSUCCESS;

  keylist = ablock->keylist;
  abuf    = afk_block_get_buffer( ablock );
  astyle  = ablock->ads->astyle;
  fktype  = ftp_datatype_to_keytype( type, name );


  /* If the style doesn't have a header, quit without actually writing the
   * key to the buffer */
  if ( !afk_style_has_header( astyle ) )
    return ukey;

  if ( fktype == FTK_COMMENT )
  {
    old_entry = 0;
  }
  else
  {
    /* Check if the keyword already exists */
    if (keylist)
    {
      ukey = dmu_key_find( keylist, (char*)name );
      if ( ukey )
        old_entry = ukey->index;
    }
  }

  /* New key */
  if ( !ukey )
  {
    ukey = dmu_key_alloc();
    ukey->type = FTK_KEY;
    pos = dmu_key_insert( keylist, ukey, pos );
  }

  mask = afk_key_private( name );
  dmu_key_set( ukey, fktype, type, name, value, unit, desc, mask );

  /* dmu_key_set handles the setting/replacing of fields in ukey.    */
  /* input NULL for unit and desc do not replace any existing value. */
  /* We should compose the keyword line from ukey elements.          */


  /* Generate the key record */
  status = afk_dtf_compose_keyline( name, value, type, ukey->unit, ukey->desc,
				      &keyline, AK_COMMENTLINE_LEN, astyle );

  if ( keyline && (status == dmSUCCESS) )
  {
    if ( old_entry == 0 )
    {
      status = afk_buffer_insert_keyline( abuf, keyline, pos );
    }
    else
    {
      status = afk_buffer_replace_keyline( abuf, keyline, old_entry );
    }
  }
  if ( status != dmSUCCESS )
  {
    char errmsg[dmMAXSTR];
    sprintf( errmsg, " Failed to write keyword %s, keyword deleted.", name );

    strcat( dm_errbuf, errmsg );
    dmu_err_set( DM_ERR_KERNEL, dm_errbuf );

    dmu_key_delete( keylist, ukey );
    ukey = NULL;
  }

  if ( keyline )
    free(keyline);

  return ukey;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_dtf_compose_keyline                         */
/*                                                             */
/* Description:                                                */
/*    Generate the full keyword line(s) from the inputs        */
/*                                                             */
/*  Input parameters:                                          */
/*    const char *name     = keyword name                      */
/*    const void *value    = keyword value                     */
/*    dmDataType  type     = keyword datatype                  */
/*    const char *unit     = keyword unit                      */
/*    const char *desc     = keyword description/comment       */
/*    char      **keyline  = output ptr for full key line(s)   */
/*    long        klimit   = size limit for keyword line       */
/*    afStyle    *astyle   = kernel style settings             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*    dmErrCode  status                                        */
/*                                                             */
/***************************************************************/
dmErrCode afk_dtf_compose_keyline( const char* name, const void* value,
                                   dmDataType type, const char* unit,
                                   const char* desc, char** keyline,
                                   long klimit, afStyle* astyle )
{
  long  len         = 0;          /* Length of string                 */
  long  keylen      = 0;          /* Length of output line            */

  char  fmt[16];                  /* format for field                 */
  char  wfmt[16];                 /* format for wrapped lines         */
  char  tmpstr[FITS_MAXLEN];      /* temporary string variable        */

  char  *optr       = NULL;       /* start of full output line        */
  char  *tptr       = NULL;       /* start of temporary string        */

  long  minNameLen     = 8;       /* Minimum length of name segment   */
  long  minStrValueLen = 8;       /* Minimum length of string value   */

  /* Check for valid name */
  len = strlen( name );
  if ( len > AK_KEYNAME_LEN )
  {
    sprintf( dm_errbuf, 
	     "TEXT Kernel:  Keyword name exceeds limit of %ld characters.", (long)AK_KEYNAME_LEN );
    dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
    return dmFAILURE;
  }

  /* Allocate space for the output line (there will be at least one). */
  optr = calloc( klimit, UT_SZ_C );
  if ( !optr ){return dmFAILURE;}

  memset((void*)tmpstr, '\0', FITS_MAXLEN);

  /* Determine format for name segment of output line         */
  /* Requirement:                                             */
  /*    Name segment should be a minimum of 8 characters long */
  /*    Name should be left justified in segment              */
  if ( len < minNameLen )
  {
    strcpy( fmt, "%-8s" );
  }
  else
  {
    strcpy( fmt, "%s" );
  }

  /* write name to output line */
  sprintf( optr, fmt, name );
  keylen = strlen(optr);

  /* Add name/value separator */
  strcat( optr, "= ");
  keylen += 2;

  /* Get value as a string using default formats based on datatype and value  */
  /* Non string type values are written to strings of no more than 20 chars   */
  /* and are written to a temporary string.                                   */
  /* String values must be at least 8 chars long (not including quotes)       */
  /* and left justified within the quotes.                                    */
  afk_key_default_fmt( type, value, fmt, astyle );

  if ( type != DT_CHAR )
  {
    tptr = &tmpstr[0];
    switch ( type )
    {
    case DT_INT2:
      sprintf( tptr, fmt, *((short*)value) );
      break;
    case DT_UINT2:
      sprintf( tptr, fmt, *((unsigned short*)value) );
      break;
    case DT_UINT1: 
      sprintf( tptr, fmt, *((unsigned char*)value) );
      break;
    case DT_INT4:
      sprintf( tptr, fmt, *((long*)value) );
      break;
    case DT_UINT4:
      sprintf( tptr, fmt, *((unsigned long*)value) );
      break;
    case DT_REAL4:
      sprintf( tptr, fmt, *((float*)value) );
      break;
    case DT_REAL8:
      sprintf( tptr, fmt, *((double*)value) );
      break;
    case DT_LOGICAL:
      if ( *(dmBool*)value == UT_TRUE )
        sprintf( tptr, "T" );
      else
        sprintf( tptr, "F" );
      break;
    default:
      sprintf( dm_errbuf, "TEXT header can't handle type: %d", type );
      dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
      return dmFAILURE;
    }

    /* set value as string within the value segment of line          */
    /* Requirements:                                                 */
    /*   non-string types are right justified ending at char 30      */
    /*   unless the keyword name length forces extension beyond      */
    len = strlen(tptr);
    if (( keylen + len ) <= 30 )
    {
      /* Value fits within standard string segment */
      sprintf(fmt,"%c%ds", '%', (30 - (int)keylen));
    }
    else
    {
      strcpy(fmt,"%s");
    }

    /* Add to output line */
    sprintf( optr+keylen, fmt, tptr );
    keylen = strlen(optr);

  }
  else
  {
    /* value is already a string.. */
    /* Requirements:                                                 */
    /*   quoted string value should be minimum of 10 chars long      */
    /*   string value should be left justified within value segment  */
    /*   string values which cause the line to exceed the limit      */
    /*    should wrap to additional CONTINUE keywords.               */
    /*   string value segment ends at a minimum of char 30           */

    tptr = *(char**)value;

    len = strlen(tptr);
    if ( len <  minStrValueLen ) 
    {
      /* write value to temporary variable to bring up size */
      tptr = &tmpstr[0];
      sprintf(tptr, "%-8s", *(char**)value);
      len = strlen(tptr);
    }

    len += 2;   /* for quotes */
    if (( keylen + len ) <= klimit )
    {
      /* Full value will fit in the line.. write to output line      */
      /* using the default format (which defines the quotes)         */
      sprintf( optr+keylen, fmt, tptr );
      keylen = strlen(optr);

    }
    else
    {
      /* Value will need to wrap with CONTINUE keys                  */
      /*   each continue line needs space for 'CONTINUE  ','&','\n' and quotes */
      long offset = klimit - 3;                     /* 3 = '&','\n'  */
      long nextra = (keylen + len) - offset;        /* # extra chars */
      long ncont  = nextra/(klimit - 15) + 1;       /* # extra lines */
      long namelen = keylen + 1;                    /* include first quote */

      /* Reallocate space for the output line to allow for continues */
      optr = realloc( optr, (1+ncont)*klimit*UT_SZ_C );
      if ( !optr ){return dmFAILURE;}

      /* Write full value  */
      sprintf( optr+keylen, fmt, tptr );

      /* Define format for continue lines */
      sprintf(wfmt,"&%c\nCONTINUE  %s", fmt[0], fmt);

      char* cptr = optr;
      do 
      {
	tptr += offset - namelen;  /* shift position in input line  */
	cptr += offset;            /* shift position in output line */
	namelen = 14;              /* "&'\nCONTINUE  "              */

	len = sprintf(cptr, wfmt, tptr );
	keylen = len - 3;          /* length from 'CONTINUE'        */

      }while ( keylen > klimit );

    }

    while ( keylen < 30 )
    {
      /* Value is short.. pad with spaces */
      /* NOTE: using strcat so that I don't have to know what the   */
      /*  offset to the current line is if it wraps.                */
      strcat(optr, " ");
      keylen++;
    }      

  }    

  /* Generate comment segment from input */
  // Code Change July 10, 2009  Bug 10802
  //   Subtract 1 from FITS_MAXLEN because copy functions need one byte
  //   free at end of buffer for null terminator. (81 char buffer - 80 char limit)

  tptr = ftp_key_comment_compose( name, type, desc, unit, tmpstr, (FITS_MAXLEN-1) );

  /* Add comment to output line IFF it does not make it too long. */
  if ( tptr )
  {
    len = strlen(tptr);
    len += 3;           /* for ' / ' separator */
    if ( (keylen + len) < klimit )
    {
      strcat( optr, " / ");
      strcat( optr, tptr );
      keylen = strlen(optr);
    }
  }

  /* pop a return on the line */
  strcat( optr, "\n");
  keylen = strlen(optr);


  /* set line to output variable */
  *keyline = optr;

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_dtf_next_mtype_index                        */
/*                                                             */
/* Description:                                                */
/*             This is the slow way to do this                 */
/*                                                             */
/*  Input parameters:                                          */
/*                   afBlock   *ablock                         */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            long   Number                                    */
/*                                                             */
/***************************************************************/
long afk_dtf_next_mtype_index( afBlock* ablock )
{
  char  mtype[CARD_SIZE];
  char  buf[CARD_SIZE];
  dmuKeyList  keylist;

  dmuKey  *ukey = NULL;
  long     no   = 1;

  keylist = ablock->keylist;
  sprintf( mtype, "MTYPE%ld", no );
  ukey = dmu_keylist_read_value_c( keylist, mtype, buf, CARD_SIZE );

  while ( ukey )
  {
    ++no;
    sprintf( mtype, "MTYPE%ld", no );
    ukey = dmu_keylist_read_value_c( keylist, mtype, buf, CARD_SIZE );
  }

 return no;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_dtf_write_comment                           */
/*                                                             */
/* Description:                                                */
/*              Write comment                                  */
/*                                                             */
/*  Input parameters:                                          */
/*                   afBlock     *ablock                       */
/*                   long         pos                          */
/*                   const char  *tag                          */
/*                   const char  *comment                      */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afk_dtf_write_comment( afBlock* ablock, long pos,
                                 const char* tag, const char* comment )
{
  char cbuf[AK_COMMENTLINE_LEN];
  afBuffer  *abuf   = NULL;
  dmuKey    *ukey   = NULL;
  dmuKeyList keylist;
  long       mask   = UT_TRUE;
  long       keyno  = 0;
  ftKeyType  fktype = FTK_COMMENT;
  dmErrCode  status = dmFALSE;


  keylist = ablock->keylist;
  abuf    = afk_block_get_buffer( ablock );

  /* find end of COMMENT segment beginning at pos */
  for ( keyno = pos; keyno <= keylist->n; keyno++ )
  {
    ukey = keylist->data[keyno-1];

    if ( ukey->keytype != FTK_COMMENT )
      break;
  }

  /* define keyword descriptor for key */
  ukey = dmu_key_alloc();
  ukey->type = FTK_KEY;
  dmu_key_insert( keylist, ukey, keyno );
  dmu_key_set( ukey, fktype, DT_CHAR, tag, &comment, NULL, NULL, mask );

  if ( tag )
    sprintf( cbuf, "%s %s\n", tag, comment );
  else
    sprintf( cbuf, "%s\n", comment );


  /* Write line to ascii buffer */
  status = afk_buffer_insert_keyline( abuf, cbuf, keyno );
  if ( status != dmSUCCESS )
  {
    dmu_key_delete( keylist, ukey );
    ukey = NULL;
  }

  return status;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_dtf_get_nrows                               */
/*                                                             */
/* Description:                                                */
/*             Count the number of rows in the block as        */
/*             quickly as possible.                            */
/*                                                             */
/*  Input parameters:                                          */
/*                   afBlock     *ablock                       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
long afk_dtf_get_nrows( afBlock* ablock )
{
  afBuffer   *abuf  = NULL;
  afFile     *afptr = NULL;
  const char *line  = NULL;

  long nrows = 0;

  abuf = ablock->abuf;
  afptr = afk_dataset_fptr( ablock->ads );
  afk_buffer_moveto_row( abuf, 1 );

  while(( line = afk_file_getline_ptr( afptr )) != NULL )
  {
    if ( *line == '\0' )
      break;
    else
      if ( *line == 'X' )
      {
        if ( cs_uceq_siz( line, "XTENSION", 8 ))
          break;
      }

    ++nrows;
  }

  return nrows;
}

/***************************************************************/
/*                                                             */
/* Function:   afp_key_ends_in_amp                             */
/*                                                             */
/* Description:                                                */
/*    Check a string for an end character of & or &'           */
/*    DUPLICATE OF ftp_key_ends_in_amp IN FITS KERNEL          */
/*                                                             */
/*  Input parameters:                                          */
/*    const char *value                                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/***************************************************************/
static int afp_key_ends_in_amp( const char* value )
{
 char* sval = NULL;
 char* ptr  = NULL;

 sval = (char*)value;


 if ( sval && *sval )
 {
  ptr = sval + strlen(sval)-1;

  if ( *ptr == '\'' && ptr > sval )
    ptr--;

  if ( *ptr == '&' )  /* CONTINUATION? */
   return 1;
 }
 return 0;
}

/***************************************************************/
/*                                                             */
/*  Function:  afp_key_continue                                */
/*                                                             */
/*  Description:                                               */
/*     Process continue statements as a special case           */
/*                                                             */
/*  Input parameters:                                          */
/*                      char ** pptr                           */
/*                      char *  text                           */
/*                                                             */
/*  Output parameters:  char ** pptr                           */
/*                      char *  desc                           */
/*                                                             */
/*                                                             */
/***************************************************************/
static int afp_key_continue( char** pptr, const char* text, const char* desc )
{
 int   flag = UT_FALSE;
 char* ptr  = *pptr;
 char* tmp  = NULL;

 long  n    = 0;
 long  m    = 0;
 int   loop = 1;


 n = strlen( ptr );

 /* Lose trailing ampersand on previous line (GSFC FITS continue convention) */
 if ( ptr[n-1] == '&' )
   n--;

 /* Scan past any leading white space */
 while ( *text == ' ' )
   text++;

 /* Lose leading quote of new line */
 if (( *text == '\'' )||( *text == '"' ))
  text++;

 /* Seek trailing quote of new line */
 while ( loop )
 {
   /* Text ends on null */
   if ( text[m] == '\0' )
   {
      loop = UT_FALSE;
   }
   else
   {
     if (( text[m] == '\'' )||( text[m] == '"' ))
     {
       /* Found trailing quote.. exit scan */
       loop = UT_FALSE;
     }
   }

   if (loop)
     m++;

 } /* end while */


 /* Now m is the first char past the trailing quote or is the term null */

 /* allocate new buffer */
 tmp = malloc( n + m + 1 );

 /* copy text from two input buffers into new buffer  */
 cs_copy_siz( ptr, tmp, n );
 cs_copy_siz( text, tmp+n, m );

 /* clean up and set pptr to new buffer area */
 free( ptr );
 *pptr = tmp;


 /* Scan for comment segment */
 if ( flag )
 {
   while ( text[m] == ' ' )
     m++;

  if ( text[m] == '/' )
  {
   m++;

   while( text[m] == ' ' )
     m++;

   cs_copy( text+m, (char*)desc );
  }
 }

 return 1;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_dtf_imag_get_ncols                          */
/*                                                             */
/* Description:                                                */
/*   This function returns the number of raw data columns in   */
/*   an  image file. The number is to be used for NAXIS1 when  */
/*   there is no keyword in header                             */
/*                                                             */
/*   This function is a combined and stripped down version of  */
/*   afk_raw_init_cols() and afk_raw_read_cols()               */
/*                                                             */
/*   Added for Bug 9023 - Ascii Image support                  */
/*                                                             */
/*  Input parameters:                                          */
/*                   afBlock  *ablock - block pointer          */
/*             const afStyle  *astyle - style pointer          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
long afk_dtf_imag_get_ncols( afBlock* ablock, const afStyle *astyle )
{
  long         ncols = 0;
  const char  *line  = NULL;
  const char  *iptr  = NULL;

  char colval[AK_COLVAL_LEN];

  afFile   *afptr = afk_dataset_fptr( ablock->ads );
  afBuffer *abuf  = ablock->abuf;
 

  afk_buffer_moveto_row( abuf, 1 );
 
  /* Get first row of data file */
  line = afk_file_getline_ptr( afptr );
  iptr = line;

  /* Count the columns */
  while ((iptr = afk_gen_read_next_col( astyle, iptr, colval,
					AK_COLVAL_LEN )))
    ++ncols;

  /* return the number of columns found */
  return ncols;
}
