/*                                                                
**  Copyright (C) 2003-2012  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afkernel.c                                                  */
/*                                                               */
/*   Functions for ASCII kernel interface                        */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"
#include "afkernel.h"
#include "../ftkernel/ftkernelP.h"



/***************************************************************/
/*                                                             */
/* Function:   afInit                                          */
/*                                                             */
/* Description:                                                */
/*              Initialization                                 */
/*                                                             */
/*  Input parameters:                                          */
/*                    none                                     */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afInit( void )
{
  return;
}



/***************************************************************/
/*                                                             */
/* Function:   afDatasetAccess                                 */
/*                                                             */
/* Description:                                                */
/*              Test filename for existence, access permission,*/
/*              and valid TEXT file.  Return TRUE if all are   */
/*              true.  If mode = "RW", test for read-write     */
/*              access.                                        */
/*                                                             */
/*  Input parameters:                                          */
/*                    const  char  *filename                   */
/*                    const  char  *mode                       */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmBool  retval                                   */
/*                                                             */
/***************************************************************/
dmBool afDatasetAccess( const char* filename, const char* mode )
{
  dmBool retval = UT_FALSE;
  int    openmode = 0;

  if ( !filename )
  {
    dmu_err_set( DM_ERR_NULL, "[afDatasetAccess]: null filename" );
    return retval;
  }

  openmode = fio_access( filename );

  if ( cs_uceq( mode, "RW" ))
  {
    if ( openmode == FIO_READWRITE )
    {
      retval = UT_TRUE;
    }
  }
  else
  {
    if ( openmode >= FIO_READWRITE )
      retval = UT_TRUE;
    else
    {
      sprintf( dm_errbuf, 
	      "[afDatasetAccess]: File %s does not exist or is unreadable",
	       filename );
      dmu_err_set( DM_ERR_NO_FILE, dm_errbuf );
     }
  }

 return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_reject_by_name                              */
/*                                                             */
/* Description:                                                */
/*              Temporary hack:                                */
/*              At the moment, we don't handle RDB ir CXC      */
/*              parameter files correctly, so just reject any  */
/*              file with the extension ".rdb" or ".par"       */
/*                                                             */
/*  Input parameters:                                          */
/*                    const  char  *filename                   */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            static dmBool  status                            */
/*                                                             */
/***************************************************************/
static dmBool afk_reject_by_name( const char* filename )
{
  char *ext = NULL;

  ext = strrchr( filename, '.' );

  if ( ext )
  {
    if ( cs_uceq( ext, ".rdb" ) || cs_uceq( ext, ".par" ) )
      return dmTRUE;
  }

  return dmFALSE;
}




/***************************************************************/
/*                                                             */
/* Function:   afDatasetOpen                                   */
/*                                                             */
/* Description:                                                */
/*              Open a TEXT Dataset and return a dataset       */
/*              handle to it.                                  */
/*                                                             */
/*  Input parameters:                                          */
/*                    const  char  *filename                   */
/*                    const  char  *hint                       */
/*                    const  char  *mode                       */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmDataset *ds                                    */
/*                                                             */
/***************************************************************/
dmDataset* afDatasetOpen( const char* filename, const char* hint,
                          const char* mode )
{
  afDataset  *ads = NULL;
  dmDataset  *ds  = NULL;
  int         dsmode = 0;

  if ( !filename )
  {
    dmu_err_set( DM_ERR_NULL, "[afDatasetOpen]: null filename" );
    return NULL;
  }

  /* TEMPORARY HACK */
  if ( afk_reject_by_name( filename ) )
    return NULL;

  dsmode = cs_uceq( mode, "RW" );

  /* Allocate the kernel dataset */
  ads = afk_dataset_alloc();

  if ( !ads )
    return NULL;

  ads->name = cs_dup( filename );

  if ( afk_dataset_open( ads, filename, dsmode, dmFALSE ) != dmSUCCESS )
  {
    afk_dataset_free( ads );
    return NULL;
  }

  /* Allocate the main dataset */
  if (!( ds = dmi_dataset_alloc( filename, (dmkDataset*)ads )))
  {
    afk_dataset_close( ads, dmFALSE );
    afk_dataset_free( ads );
    return NULL;
  }

  dmi_dataset_set_mode( ds, dsmode ); 

 
  /* If global hints have already been passed to the kernel, */
  /* put them into this dataset's style structure now        */
  afp_set_style_hints( NULL, ads->astyle );

  /* now parse command line option input and set style values*/
  afp_dataset_set_hint_list( ds, hint );


  /* Scan the dataset and set up the list of blocks             */
  /* NB:  This routine may very well clobber some of the values */
  /*      in the dataset's style structure.  Temporarily put    */
  /*      those values into the kernel hint structure for safe  */
  /*      keeping. Those values will  be replaced after this    */
  /*      call and kernel hints reinitialized.                  */


  /* store style values in global hint structure */
  afp_set_kernel_hints( ads->astyle, NULL);
 
  /*  Scan dataset */

  if ( afp_dataset_scan( ds ) != dmSUCCESS )
  {
    afk_dataset_close( ads, dmFALSE );
    afk_dataset_free( ads );
    dmi_dataset_close( ds );
    return NULL;
  }

  /*  replace parsed style values */
  afp_set_style_hints( NULL, ads->astyle );

  /* clear kernel hint struct */
  afp_clear_kernel_hints(NULL);


  return ds;
}


/***************************************************************/
/*                                                             */
/* Function:   afDatasetCreate                                 */
/*                                                             */
/* Description:                                                */
/*              Create a dataset.  Makes an empty TEXT file.   */
/*              Calls function afp_dataset_create() to do      */
/*              actual work.                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                    const  char  *filename                   */
/*                    const  char  *hint                       */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmDataset *ds                                    */
/*                                                             */
/***************************************************************/
dmDataset* afDatasetCreate( const char* filename, const char* hint )
{
  return  afp_dataset_create( filename, hint );
}


/***************************************************************/
/*                                                             */
/* Function:   afDatasetClose                                  */
/*                                                             */
/* Description:                                                */
/*              Close a dataset.                               */
/*              Calls function afp_dataset_close() to do the   */
/*              actual work.                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDataset  *ds                           */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afDatasetClose( dmDataset* ds )
{
  afp_dataset_close( ds, dmFALSE );
}


/***************************************************************/
/*                                                             */
/* Function:   afDatasetDelete                                 */
/*                                                             */
/* Description:                                                */
/*              Delete a dataset.                              */
/*              Calls function afp_dataset_close() to do the   */
/*              actual work.                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDataset  *ds                           */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afDatasetDelete( dmDataset* ds )
{
  afp_dataset_close( ds, dmTRUE );
}


/***************************************************************/
/*                                                             */
/* Function:   afDatasetPrint                                  */
/*                                                             */
/* Description:                                                */
/*              Debugging routine to display DM knowledge of   */
/*              the dataset.                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                    const dmDataset  *ds                     */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afDatasetPrint( const dmDataset* ds )
{
  afDataset *ads   = NULL;
  long       ii    = 0;
  dmBlock   *block = 0;
  long        n    = 0;

  ads = (afDataset*)dmi_dataset_get_kds( ds );

  if ( !ads )
  {
    printf( "Null dataset\n" );
    return;
  }

  n = dmi_dataset_get_nv( ds );

  for ( ii = 1; ii <= n; ii++ )
  {
    block = dmi_dataset_get_block( ds, ii );

    if ( block )
      afp_block_print( block );
  }
}




/***************************************************************/
/*                                                             */
/* Function:   afBlockOpen                                     */
/*                                                             */
/* Description:                                                */
/*              Open a block by number.                        */
/*              Special case: no = 0 does the 'most interesting*/
/*              block' of type 'type'.  The asp_block_open()   */
/*              routine does the major work.                   */
/*                                                             */
/*              Currently, TEXT tables only have one block.    */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDataset  *ds                           */
/*                    long        no                           */
/*                    dmBlockType type                         */
/*                    char       *hint                         */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmBlock  *block                                  */
/*                                                             */
/***************************************************************/
dmBlock* afBlockOpen( dmDataset* ds, long no, dmBlockType type,
                      char* hint )
{
  afDataset *ads   = NULL;
  dmBlock   *block = NULL;


  ads = (afDataset*)dmi_dataset_get_kds( ds );

  if ( !ads )
    return NULL;

  if ( no == 0 )
    no = afp_find_interesting_block_no( ds );

  if ( no > ads->n || no <= 0 || !ads->blocks[no-1] )
  {
    sprintf( dm_errbuf, "Error: Block number %ld does not exist in TEXT file.",
             no );
    dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
    return  NULL;
  }
 
  block = afp_block_open( ds, no, dmFALSE, UT_TRUE, hint );

  return  block;
}



/***************************************************************/
/*                                                             */
/* Function:   afBlockCreate                                   */
/*                                                             */
/* Description:                                                */
/*              Create a block.                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDataset  *ds                           */
/*                    long        no                           */
/*                    const char *name                         */
/*                    dmBlockType type                         */
/*                    void       *kb                           */
/*                    char       *hint                         */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmBlock  *block                                  */
/*                                                             */
/***************************************************************/
dmBlock* afBlockCreate( dmDataset* ds, long no, const char* name,
                        dmBlockType type, void* kb, char* hint )
{
  dmBlock *block = NULL;


  block = afp_block_insert( ds, name, no, type, kb, hint );

  return  block;
}



/***************************************************************/
/*                                                             */
/* Function:   afBlockFlush                                    */
/*                                                             */
/* Description:                                                */
/*              Flush a block from memory.  Calls function     */
/*              afp_block_flush() to do the actual work        */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock  *block                          */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  dmSUCCESS                             */
/*                                                             */
/***************************************************************/
dmErrCode afBlockFlush( dmBlock* block )
{
  afp_block_flush( block );
  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   afBlockClose                                    */
/*                                                             */
/* Description:                                                */
/*              Close a block.  This does not destroy the      */
/*              block pointer, but releases the structure      */
/*              hanging off of it.                             */
/*              Calls function afp_block_close() to do the     */
/*              actual work.                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock  *block                          */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afBlockClose( dmBlock* block )
{
  return afp_block_close( block );
}



/***************************************************************/
/*                                                             */
/* Function:   afBlockDelete                                   */
/*                                                             */
/* Description:                                                */
/*              Delete a Block.                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDataset *ds                            */
/*                    long       no                            */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  dmSUCCESS                             */
/*                                                             */
/***************************************************************/
dmErrCode afBlockDelete( dmDataset* ds, long no )
{
  (void) ds;
  (void) no;

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   afBlockSetPref                                  */
/*                                                             */
/* Description:                                                */
/*              Set block preferences                          */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock    *block                        */
/*                    char      **prefcols                     */
/*                    long        npref                        */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afBlockSetPref( dmBlock* block, char** prefcols, long npref )
{
  int    is_default = UT_FALSE;
  char  *value      = NULL;

  char   buf[CARD_SIZE];

  dmi_block_set_pref( block, (char **)prefcols, npref, is_default );
  cs_compose_paren_list ( prefcols, npref, buf, CARD_SIZE );

  value = buf;

  if ( dmi_block_writeable( block ))
    afp_key_write( block, 0, "CPREF", DT_CHAR, &value, npref, NULL, NULL );
}



/***************************************************************/
/*                                                             */
/* Function:   afBlockUpdateFilter                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock    *block                        */
/*                    dmSubspace *subspace                     */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afBlockUpdateFilter( dmBlock* block, dmSubspace* subspace )
{
  (void) block;
  (void) subspace;
}



/***************************************************************/
/*                                                             */
/* Function:   afImageSetAxes                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock          *block - block pointer         */
/*             const dmDataType  type  - data type             */
/*             const long       *axlen - axis length(s)        */
/*             const long        naxes - number of axis        */ 
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afImageSetAxes( dmBlock* block, const dmDataType type,
                     const long* axlen, const long naxes )
{
  dmDataset    *ds     = NULL;
  afDataset    *ads    = NULL;
  afBlock      *ablock = NULL;
  afCol        *acol   = NULL;
  afBuffer     *abuf   = NULL;
  dmDescriptor *dd     = NULL;
  char         *unit   = NULL;
  afFlavour     flav   = AK_FLAV_UNK;
  long          ii;

  /* Determine ascii flavor */
  ds = dmi_block_get_ds( block );

  if ( !ds )
    return;

  ads = (afDataset*)dmi_dataset_get_kds( ds );

  if ( !ads )
    return;

  flav = afk_dataset_get_flavour( ads );

  ablock = (afBlock*)dmi_block_get_kb( block );

  if ( !ablock )
    return;

  if ( ablock->cols )
    acol   = ablock->cols[0];

  /* adjust length */
  if ( acol && naxes )
  {
    acol->length = 1;

    for (ii = 0; ii < naxes; ii++)
      acol->length *= axlen[ii];
  }

  switch ( flav )
  {
    case AK_FLAV_DTF:
      afp_dtf_image_set( block, type, unit, axlen, naxes );
      break;
    default:
      break;
  }

  /* (re)set dm layer descriptor info */
  dd = dmi_block_get_col( block, 1 );

  if ( dd )
  {
    dmi_descriptor_set_axes( dd, axlen, naxes );
    dmi_descriptor_set_datatype( dd, type );
  }

  /* (re)set buffer struct.  Unsure if reset s/b TRUE or FALSE here.*/
  /* Functional difference at this point is resetting the rrows ptr */
  /* Mode check? */
  abuf   = afk_block_get_buffer( ablock );
  afk_buffer_sync( abuf, ablock, dmFALSE );
}



/***************************************************************/
/*                                                             */
/* Function:   afKeyWrite                                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock     *block                       */
/*                    long         keyno                       */
/*                    const char  *name                        */
/*                    const void  *value                       */
/*                    long         nvals                       */
/*                    const char  *unit                        */
/*                    const char  *comment                     */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor  *dd                                */
/*                                                             */
/***************************************************************/
dmDescriptor* afKeyWrite( dmBlock* block, long keyno, const char* name,
                          dmDataType type, const void* value,
                          long nvals, const char* unit, const char* comment )
{
  dmDescriptor *dd = NULL;


  /* Catch erroneous COMMENT and HISTORY requests.  */
  /* Should use dmBlockWriteComment for these       */
  if ( cs_uceq( name, "COMMENT" ) || cs_uceq( name, "HISTORY" ))
  {
    afWriteComment( block, keyno, name, *(char**)value );
    return NULL;
  }

  /* Write keyword */
  dd = afp_key_write( block, keyno, name, type, value, nvals,
                        unit, comment );

  return dd;
}


/***************************************************************/
/*                                                             */
/* Function:   afKeyDelete                                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor   *dd                       */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afKeyDelete( dmDescriptor* dd )
{
  dmBlock *block = NULL;
  dmuKey  *ukey  = NULL;
  afBlock *ablock = NULL;

  ukey   = (dmuKey*)dmi_descriptor_get_kdd( dd );
  block  = dmi_descriptor_get_block( dd );
  ablock = (afBlock*)dmi_block_get_kb( block );

  afk_key_delete( ablock, ukey );
  dmi_descriptor_set_kdd( dd, NULL );
  dmi_key_delete( dd );
}



/***************************************************************/
/*                                                             */
/* Function:   afReadComment                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock   *block                         */
/*                    long       keyno                         */
/*                    char     **tag                           */
/*                    char     **comment                       */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmBool  status                                   */
/*                                                             */
/***************************************************************/
dmBool afReadComment( dmBlock* block, long keyno, char** tag,
                      char** comment )
{
  dmuKeyList  keylist;
  dmuKeyType  ktype;
  dmuKey     *akey   = NULL;
  afBlock    *ablock = NULL;

  long  nkeys   = 0;
  long  akeyno  = 0;
  long  koffset = 0;

  long  start_offset = 0;
  long  end_offset   = 0;


  nkeys = dmi_block_get_nkeys( block );
  ablock = (afBlock*)dmi_block_get_kb( block );
  keylist = afk_block_keylist( ablock );
  *tag = NULL;
  *comment = NULL;
 
  if ( nkeys == 0 )
    nkeys = 1;
 
  if ( !ablock || !keylist )
    return dmFALSE;

  if ( keyno < 0 || keyno > nkeys )
    return dmFALSE;
 
  akeyno = afp_key_pos( block, keyno );

  /* The next DM level key is the next key with mask not set       */
  /* Associated keys can be in the range start_offset:end_offset-1 */
 
  koffset = ablock->keyno;      /* last one tried plus 1  (keyno -1 + 1 ) */
  start_offset = akeyno;        /* akeyno - 1 + 1 */
  end_offset = start_offset;

  while( end_offset < keylist->n && keylist->data[end_offset]->mask ) 
    ++end_offset;

 
  /* koffset saved from last trip; if it's outside the range, we must have   */
  /* been looking at another key, so we reset.                               */
  /* that is:                                                                */
  /*   KEY1  = VAL1                                                          */
  /*   COMMENT  k1.c1                                                        */
  /*   COMMENT  k1.c2                                                        */
  /*   COMMENT  k1.c3                                                        */
  /*   KEY2  = VAL2                                                          */
  /*   COMMENT  k2.c1                                                        */
  /*                                                                         */
  /*  In the DM we provide access to k1.c1 to k1.c3 via KEY1, considering    */
  /*  them as comments associated with the preceding key                     */
  /*  We keep track of which the next comment is. If we get to KEY2 we       */
  /*  reset the counter.                                                     */

  if ( koffset < start_offset || koffset > end_offset ) 
    koffset = start_offset;

  /* Now we look for a comment */
  while( koffset < end_offset )
  {
    akey = keylist->data[koffset];
    ktype = akey->keytype;

    if ( ktype == FTK_COMMENT )
    {
      *tag = cs_dup( akey->name );
      *comment = cs_dup( *(char**)&akey->value );
      ablock->keyno = koffset+1;
      return UT_TRUE;
    }

    ++koffset;
  }

  /* No more comments for this key */
  ablock->keyno = -1; 

  return dmFALSE; 
}



/***************************************************************/
/*                                                             */
/* Function:   afWriteComment                                  */
/*                                                             */
/* Description:                                                */
/*             Calls function afp_comment_write() to do the    */
/*             actual work.                                    */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock    *block                        */
/*                    long        keyno                        */
/*                    const char *tag                          */
/*                    const char *comment                      */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmBool  status                                   */
/*                                                             */
/***************************************************************/
dmBool afWriteComment( dmBlock* block, long keyno, const char* tag,
                       const char* comment )
{
  return afp_comment_write( block, keyno, tag, comment );
}



/***************************************************************/
/*                                                             */
/* Function:   afWriteComment                                  */
/*                                                             */
/* Description:                                                */
/*             Calls function afp_comment_write() to do the    */
/*             actual work.                                    */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock      *block                      */
/*                    dmDescriptor *dd                         */
/*                    void         *kdd                        */
/*                    char         *hint                       */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afColCreate( dmBlock* block, dmDescriptor* dd, void* kdd, char* hint )
{
  long no = 0;
 
  afp_create_col( block, dd, no, kdd, hint );
}



/***************************************************************/
/*                                                             */
/* Function:   afColSetRanget                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor *dd                         */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afColSetRange( dmDescriptor* dd )
{
  char name[FITS_MAXLEN];
  GenericType buf[2];
  GenericType cpt_buf[2];
 
  dmDescriptor *cpt_dd = NULL;
  dmBlock      *block  = NULL;
  afBlock      *ablock = NULL;
  afCol        *acol   = NULL;
  afDataset    *ads    = NULL;
  afStyle      *astyle = NULL;
  void         *tlmin  = NULL;
  void         *tlmax  = NULL;
  void         *cpt_tlmin = NULL;
  void         *cpt_tlmax = NULL;
  long          no     = 0;
  long          dim    = 0;
  long          cptNo  = 0;
  int           is_time = 0;
  dmDataType    ddtype;

 
  block  = dmi_descriptor_get_block( dd );
  ablock = (afBlock*)dmi_block_get_kb( block );

  if ( !ablock )
    return dmSUCCESS;

  ads    = ablock->ads;

  if ( !ads )
    return dmSUCCESS;
 
  astyle = ads->astyle;

  if ( !afk_style_use_fits( astyle ))
    return dmSUCCESS;

  ddtype = dmi_descriptor_get_datatype( dd );
  is_time = cs_uceq( dmi_descriptor_get_name( dd ), "TIME" );

  if ( is_time ) 
    return dmSUCCESS;

  tlmin = &buf[0];
  tlmax = &buf[1];

  if ( !dmi_descriptor_get_range( dd, tlmin, tlmax ))
  {
    tlmin = NULL;
    tlmax = NULL;
  }

  dim = dmi_descriptor_get_element_dim( dd );

  for ( cptNo = 1; cptNo <= dim; cptNo++ )
  {
    cpt_dd = dmi_descriptor_get_cpt( dd, cptNo );
    cpt_tlmin = &cpt_buf[0];
    cpt_tlmax = &cpt_buf[1];

    if ( !dmi_descriptor_get_range( cpt_dd, cpt_tlmin, cpt_tlmax ))
    {
      cpt_tlmin = tlmin;
      cpt_tlmax = tlmax;
    }

    acol = dmi_descriptor_get_kdd( cpt_dd );
    no = acol->no;

    if ( cpt_tlmin && !ar_mix_is_min( ddtype, cpt_tlmin ))
    {
      sprintf( name, "TLMIN%ld", no );
      afp_key_write( block, 0, name, ddtype, cpt_tlmin, 1, NULL, NULL );
    }

    if ( cpt_tlmax && !ar_mix_is_max( ddtype, cpt_tlmax ))
    {
      sprintf( name, "TLMAX%ld", no );
      afp_key_write( block, 0, name, ddtype, cpt_tlmax, 1, NULL, NULL );
    }
  }
 
  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   afColSetNull                                    */
/*                                                             */
/* Description:                                                */
/*  Write a null value for the column.                         */
/*  Note that the DM supports the ability to define a null     */
/*  value even for floating point columns, while FITS usually  */
/*  uses NaN to denote null values for floating point columns. */
/*  It is recommended that NaN be used as the null value       */
/*  normally, for compatibility with normal FITS usage - only  */
/*  use non-NaN null values for a floating point column if you */
/*  really need to.  On the other hand, there's still lots of  */
/*  code that chokes on NaN, so there are times when it makes  */
/*  sense.                                                     */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor *dd                         */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afColSetNull( dmDescriptor* dd )
{
  char name[FITS_MAXLEN];
  GenericType buf;
  void         *value  = NULL;
  dmBlock      *block  = NULL;
  dmDescriptor *cpt_dd = NULL;
  afBlock      *ablock = NULL;
  afDataset    *ads    = NULL;
  afStyle      *astyle = NULL;
  afCol        *acol   = NULL;
  long          dim    = 0;
  long          cptNo  = 0;
  long          no     = 0;
  dmDataType    ddtype;

  block  = dmi_descriptor_get_block( dd );
  ablock = (afBlock*)dmi_block_get_kb( block );

  if ( !ablock )
    return dmSUCCESS;

  ads    = ablock->ads;

  if ( !ads )
    return dmSUCCESS;

  astyle = ads->astyle;

  if ( !afk_style_use_fits( astyle ))
    return dmSUCCESS;

  value = &buf;

  if ( !dmi_descriptor_get_null( dd, value ))
    return dmSUCCESS;

  ddtype = dmi_descriptor_get_datatype( dd );
  dim = dmi_descriptor_get_element_dim( dd );

  for ( cptNo = 1; cptNo <= dim; cptNo ++ )
  {
    cpt_dd = dmi_descriptor_get_cpt( dd, cptNo );
    acol = (afCol*)dmi_descriptor_get_kdd( cpt_dd );
    no = acol->no;
    ftp_null_keyword_compose( ddtype, no, name, UT_FALSE );
    afp_key_write( block, 0, name, ddtype, value, 1, NULL, NULL );
  }
 
  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   afColSetBin                                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor *dd                         */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afColSetBin( dmDescriptor* dd )
{
  char name[FITS_MAXLEN];
  GenericType    buf;
  void          *value  = NULL;
  dmBlock       *block  = NULL;
  dmDescriptor  *cpt_dd = NULL;
  afBlock       *ablock = NULL;
  afDataset     *ads    = NULL;
  afStyle       *astyle = NULL;
  afCol         *acol   = NULL;
  long           dim    = 0;
  long           cptNo  = 0;
  long           no     = 0;
  dmDataType     ddtype;
 
 
  block  = dmi_descriptor_get_block( dd );
  ablock = (afBlock*)dmi_block_get_kb( block );

  if ( !ablock )
    return dmSUCCESS;

  ads    = ablock->ads;

  if ( !ads )
    return dmSUCCESS;

  astyle = ads->astyle;

  if ( !afk_style_use_fits( astyle ))
    return dmSUCCESS;

  value = &buf;

  if ( !dmi_descriptor_get_bin( dd, value ))
    return dmSUCCESS;
 
  /* Set TDBIN for descriptor and each component */
  dim = dmi_descriptor_get_element_dim( dd );
  ddtype = dmi_descriptor_get_datatype( dd );

  for ( cptNo = 1; cptNo <= dim; cptNo++ )
  {
    cpt_dd = dmi_descriptor_get_cpt( dd, cptNo );
    acol = (afCol*)dmi_descriptor_get_kdd( cpt_dd );
    no = acol->no;
    sprintf( name, "TDBIN%ld", no );
    afp_key_write( block, 0, name, ddtype, value, 1, NULL, NULL );
  }

  return dmSUCCESS;
}


#if 0
/***************************************************************/
/*                                                             */
/* Function:   afColRead                                       */
/*                                                             */
/* Description:                                                */
/*             llpix, urpix, and dpix are ignored for now      */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor *dd                         */
/*                    long         *llpix                      */
/*                    long         *urpix                      */
/*                    long         *dpix                       */
/*                    long          row1                       */
/*                    long          row2                       */
/*                    void         *value                      */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afColRead( dmDescriptor* dd, long* llpix, long* urpix,
                     long* dpix, long row1, long row2, void* value )
{
  dmErrCode  status = dmSUCCESS;
  dmBlock   *block  = NULL;

  block  = dmi_descriptor_get_block( dd );

  status = afp_block_read_col( block, dd, llpix, urpix, dpix, row1,
                               row2, value);

  return status;
}
#endif


/***************************************************************/
/*                                                             */
/* Function:   afReadRows                                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock   *block                         */
/*                    dmBuffer  *buffer                        */
/*                    long       row1                          */
/*                    long       row2                          */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afReadRows( dmBlock* block, dmBuffer* buffer,
                      long row1, long row2 )
{
  afBuffer  *abuf   = NULL;
  afBlock   *ablock = NULL;
  dmErrCode  status = dmSUCCESS;

  ablock = (afBlock*)dmi_block_get_kb( block );
  abuf   = afk_block_get_buffer( ablock );

  afk_buffer_sync( abuf, ablock, dmFALSE );
  status = afk_buffer_read_rows( abuf, buffer, row1, row2 );

  return status;
}



/***************************************************************/
/*                                                             */
/* Function:   afWriteRows                                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock   *block                         */
/*                    dmBuffer  *buffer                        */
/*                    long       row1                          */
/*                    long       row2                          */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afWriteRows( dmBlock* block, dmBuffer* buffer, long row1, long row2 )
{
  afBuffer  *abuf   = NULL;
  afBlock   *ablock = NULL;
  long       moved  = 0;
  dmErrCode  status = dmSUCCESS;

  ablock = (afBlock*)dmi_block_get_kb( block );
  abuf   = afk_block_get_buffer( ablock );

  status = afk_buffer_write_rows( abuf, buffer, row1, row2, &moved );

  return status;
}



/***************************************************************/
/*                                                             */
/* Function:   afColWrite                                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor  *dd                        */
/*                    long          *llpix                     */
/*                    long          *urpix                     */
/*                    long          *dpix                      */
/*                    long           row1                      */
/*                    long           row2                      */
/*                    void          *value                     */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afColWrite( dmDescriptor* dd, long* llpix, long* urpix,
                      long* dpix, long row1, long row2, void* values )
{
  long           moved = 0; 
  dmErrCode     status = dmSUCCESS;

  /* to avoid compiler warnings */
  (void) llpix;
  (void) urpix;
  (void) dpix;


  status = afk_buffer_write_col( dd, row1, row2, values, &moved );

  return status;
}



/***************************************************************/
/*                                                             */
/* Function:   afColFlush                                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor  *dd                        */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afColFlush( dmDescriptor* dd )
{
  (void) dd;
}




/***************************************************************/
/*                                                             */
/* Function:   afbufferSize                                    */
/*                                                             */
/* Description:                                                */
/*             What is a sensible number of rows to try and    */
/*             read at once?                                   */
/*             Kernel routine                                  */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock  *block                          */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            long  1024                                       */
/*                                                             */
/***************************************************************/
long afBufferSize( dmBlock* block )
{
  (void) block;
  return 1024;
}



/***************************************************************/
/*                                                             */
/* Function:   afSetHint                                       */
/*                                                             */
/* Description:                                                */
/*    Kernel call                                              */
/*                                                             */
/*  Input parameters:                                          */
/*    dmBlock    *block                                        */
/*    const char *name                                         */
/*    const char *value                                        */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*    dmSUCCESS  - option was recognized                       */
/*     other     - option was not recognized                   */
/*                                                             */
/***************************************************************/
dmErrCode afSetHint( dmBlock* block, const char* name, const char* value )
{
  dmErrCode status = dmSUCCESS;

  status = afp_set_hint( block, name, value );

  return status;
}



/***************************************************************/
/*                                                             */
/* Function:   afDescriptorFree                                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor  *dd                        */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afDescriptorFree( dmDescriptor* dd )
{
  dmBlock* block   = NULL;    /* DM block pointer           */
  afBlock* ablock  = NULL;    /* Ascii block pointer        */
  ftAux*   ftaux   = NULL;
  long     flag    = 0;

  ftaux = (ftAux*)dmi_descriptor_get_kdd( dd );

  if ( ftaux )
  {
    switch( ftaux->type )
    {
      case FTK_COL:
      flag = dmi_descriptor_get_flag( dd );

      if ( !( flag & DM_FLAG_COORD ))
        afk_col_free( (afCol*)ftaux );

      /* Special code to handle case of deleting trivial image   */
      /* descriptor at the kernel level.                         */
      block = dmi_descriptor_get_block( dd );
      ablock = (afBlock*)dmi_block_get_kb( block );
      if ( ablock &&  ablock->block_type == dmIMAGE && ablock->ncols == 1 )
      {
	free( ablock->cols );
	ablock->cols = NULL;
	ablock->nmax = 0;
	ablock->ncols = 0;
      }

      break;

      case FTK_COORD:
        ftp_coord_free( (ftCoord*)ftaux );
        break;

      case FTK_SUBSPACE:
        ftp_ftss_free( (ftSubspaceCol*)ftaux );
        break;

      case AFK_VECTOR:
        // Code Change  October 18, 2012
        //  Do not call afk_vector_free() if coordinate
        //  Can result in double free error

	flag = dmi_descriptor_get_flag( dd );

	if ( !( flag & DM_FLAG_COORD ))
	  afk_vector_free( (afVector*)ftaux );

        break;

      default:
        break;
    }

    dmi_descriptor_set_kdd( dd, NULL );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   afSubspaceDescriptorWrite                       */
/*                                                             */
/* Description:                                                */
/*             from ftkernel                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor  *dd                        */
/*                    long           cptno                     */
/*                    char          *value                     */
/*                    char          *tabname                   */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afSubspaceDescriptorWrite( dmDescriptor* dd, long cptno,
                                char* value, char* tabname )
{
  dmBlock    *block = NULL;
  dmDataType  type;
  dmDescriptorType  dtype;

  afBlock        *ablock = NULL;
  ftSubspaceCol  *ftss   = NULL;

  char keyname[CARD_SIZE];
  char tabval[TEXT_SIZE];
  char fullname[CARD_SIZE];
  char tvalue[2] ="I"; 
  char *desc    = " ";
  char *valdesc = "DM Keyword: Descriptor value.";
  char *fdesc   = "DM Keyword: Descriptor datatype.";
  char *name    = NULL;
  char *valuep  = value;
  char *table_value = "TABLE";
  char *tvaluep  = NULL;
  char *unitp    = NULL;
  int   do_table = 0;


  if ( !dd )
    return;

  type =  dmi_descriptor_get_datatype( dd );
  dtype = dmi_descriptor_get_type( dd );

  if ( dtype != dmSUBSPACE )
  {
    sprintf( dm_errbuf, "Descriptor %s is not a subspace descriptor",
             dmi_descriptor_get_name( dd ));
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return;
  }

  do_table = !cs_is_blank( tabname );
  block    = (dmBlock*)dmi_descriptor_get_block( dd );
  ablock   = (afBlock*)dmi_block_get_kb( block );

  if ( !ablock )
    return;

  ftss = afp_get_ftss( block, dd, -1 );

  if ( ftss->index == 0 )
  {
    /* We already had an ftss and its index is zero */
    /* No DSTYP keyword used? EG Default time       */  
  

    if ( cptno == 1 )
      ftss->sum++;

    if ( do_table && ftss->sum > 1 )
      ftss->index = ++ablock->max_dstyp;
    else
      return;
  } 


  /* Have we already written DSTYPn? */
  if ( !ftss->written )
  {
    sprintf( keyname, "DSTYP%ld", ftss->index );
    dmi_descriptor_get_name_with_cpts( dd, fullname, CARD_SIZE );
    name = fullname;

    afp_key_write( block, 0, keyname, DT_CHAR, &name, 1, NULL, 
                   "DM Keyword: Descriptor name." );
    ftss->written = UT_TRUE;  
  }
 
  unitp = NULL;

 
 /* The first subspace component has DSVALn, the second has 2DSVALn etc*/ 
 /* If it's a range, we do  DSVALn = 'min:max'                         */
 /* If it's a table, we do  DSVALn = 'TABLE'                           */
 /*                         DSREFn = 'GTI4:'  etc                      */

  if ( cptno == 1 )
  {
    sprintf( keyname, "DSVAL%ld", ftss->index );
    desc = valdesc;

    if (( cs_is_blank( value ) && do_table ) || cs_eq( value, "TABLE" ))
    {
      valuep = table_value;   
      desc = NULL;
    }
  }
  else
  {
    sprintf( keyname, "%ldDSVAL%ld", cptno, ftss->index );

    if ( cs_is_blank( value ) && do_table )
      valuep = NULL;
  }

  unitp = dmi_descriptor_get_unit( dd );

  if (( dm_global & DM_GLOBAL_OLD ) && ( unitp && *unitp ))
    desc = " ";

  if ( valuep )
    afp_key_write( block, 0, keyname, DT_CHAR, &valuep, 1, unitp, desc ); 

  /* Write the data type and unit */
  if ( cptno == 1 )
  {
    tvalue[0] = ftp_datatype_to_bt_tform( type );  
    tvaluep = &tvalue[0];
    sprintf( keyname, "DSFORM%ld", ftss->index );  
    afp_key_write( block, 0, keyname, DT_CHAR, &tvaluep, 1, NULL, fdesc ); 
    sprintf( keyname, "DSUNIT%ld", ftss->index );  
    tvaluep = dmi_descriptor_get_unit( dd );

    if ( !cs_is_blank( tvaluep ))
      afp_key_write( block, 0, keyname, DT_CHAR, &tvaluep, 1, NULL,
                    "DM Keyword: Descriptor unit."); 
  }

  if ( tabname && *tabname )
  {
    if ( cptno == 1 )
      sprintf( keyname, "DSREF%ld", ftss->index );
    else
      sprintf( keyname, "%ldDSREF%ld", cptno, ftss->index );

    /* HEASARC Grouping convention: add colon. Not widely adopted, may */
    /* remove someday                                                  */

    sprintf( tabval, "%s", tabname );
    tvaluep = &tabval[0];
    afp_key_write( block, 0, keyname, DT_CHAR, &tvaluep, 1, NULL, NULL ); 
  }
}


/***************************************************************/
/*                                                             */
/* Function:   afSubspaceTableWrite                            */
/*                                                             */
/* Description:                                                */
/*             from ftkernel                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor  *dd                        */
/*                    char          *tabname                   */
/*                    void          *val1                      */
/*                    void          *val2                      */
/*                    long           nvals                     */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afSubspaceTableWrite( dmDescriptor* dd, char* tabname,
                           void* val1, void* val2, long nvals )
{
  dmDataset    *ds           = NULL;  /* DataSet pointer       */
  dmBlock      *parent_block = NULL;  /* Block we are writing  */
  dmBlock      *block        = NULL;  /* New GTI block we create here */
  dmDescriptor *new_dd       = NULL;  
  dmDescriptor *cols[2];

  dmuKeyList    keylist;
  dmuKey       *ukey   = NULL;

  afBlock      *ablock = NULL;  /* Kernel block pointer        */
  afBuffer     *abuf   = NULL;  /* Kernel buffer               */
  afDataset    *ads    = NULL;  /* Kernel dataset pointer      */
  afStyle      *astyle = NULL;  /* Ascii style structure       */

  char *name = NULL;  /* Block name                            */

  long ii  = 0; /* counter */
  long no  = 0; /* Subspace table block # */

  //dmErrCode status = dmSUCCESS;


  /* Keys to copy from parent header if present.                       */
  /* This breaks the DM paradigm a bit but is included to keep         */
  /* SDS/Archives (Arnold) happy, so that the block has a minimal      */
  /* CXC header if the parent block also has a CXC (or HEASARC) header.*/

  #define NCXC 8


  dmuKey *ukeys[NCXC];
  char   *keyname[NCXC] = { "MISSION", "TELESCOP", "INSTRUME", "MJDREF",
                            "TSTART",  "TSTOP", "TIMESYS", "TIMEUNIT" };
  char   keytype[NCXC+1]  = "cccdddcc";
  double keydval[NCXC];
  long   keyival[NCXC];
  char  *keyval[NCXC] = { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL };
  char  *keycmt[NCXC];
  long   nkeys = NCXC;
  long   ncols = 0;


  if ( !dd || !tabname || !(*tabname))
    return;

  /* Get pointer to DM dataset */
  parent_block = (dmBlock*)dmi_descriptor_get_block( dd );
  ds = dmi_block_get_ds( parent_block );

  /* Check write mode */
  if ( !dmi_dataset_mode( ds ))
    return;

  ablock = (afBlock*)dmi_block_get_kb( parent_block );
  if ( !ablock )
    return;

  ads    = ablock->ads;
  if ( !ads )
    return;

  astyle = ads->astyle;
  if ( !afk_style_use_fits( astyle ))
    return;

  /* Obtain values for minimal keyword set from current block */
  keylist = ablock->keylist;
  for ( ii = 0; ii < nkeys; ii++ )
  {
    keycmt[ii] = calloc( FLEN_CARD, UT_SZ_C );

    switch( keytype[ii] )
    {
    case 'c' : 
      keyval[ii] = calloc( FLEN_CARD, UT_SZ_C );
      ukey = dmu_keylist_read_value_c( keylist, keyname[ii], keyval[ii], FLEN_CARD );
      break;

    case 'i' :
      ukey = dmu_keylist_read_typed_value( keylist, keyname[ii], DT_INT4, &keyival[ii] );
      if ( !ukey )
	keyival[ii] = 0;

      break;
    case 'd' :
    default:
      ukey = dmu_keylist_read_typed_value( keylist, keyname[ii], DT_REAL8, &keydval[ii] );
      if ( !ukey )
	keydval[ii] = 0.0;
      break;
    }

    if ( ukey )
      strcpy( keycmt[ii], ukey->desc );

    ukeys[ii] = ukey;

  } /* end key loop */

  name = ftp_subspace_table_name( ds, tabname ); 

  /* Is this name an existing block or not? */
  no = dmi_block_search( ds, name );

  if ( no > 0 )
  {
    /* Handle case of multiply defined block? This is the raw block; */
    /* we might be in syncing trouble if it's already open. We don't */
    /* make it read its own subspace,too!                            */

    block  = afp_block_open( ds, no, UT_FALSE, UT_FALSE, NULL ); 
    ablock = (afBlock*)dmi_block_get_kb( block );
    abuf   = afk_block_get_buffer( ablock );
    ncols  = dmi_block_get_ncols( block );

    if ( ncols < 2 )
    {
      sprintf( dm_errbuf, "Subspace table %s in %s has less than two columns",
              name, dmi_dataset_get_name( ds ) );
      dmu_err_set( DM_ERR_SUBSPACE, dm_errbuf );

      for ( ii = 0; ii < nkeys; ii++ )
      {
        free( keycmt[ii] );

        if ( keytype[ii] == 'c' )
          free( keyval[ii] );
      }

      free( name );
      return;
    }

    afk_buffer_sync( abuf, ablock, dmFALSE );

    /* Convention: GTI is first two scalar cols whatever they are */
    cols[0] = dmi_block_get_col( block, 1 );
    cols[1] = dmi_block_get_col( block, 2 );  
  }
  else
  {
    /* Create  block: Always a dmTABLE block */
    no = dmi_dataset_get_n( ds ) + 1;
    block = afBlockCreate( ds, no, name, dmTABLE, NULL, NULL );

    if ( !block ) 
    {
      dmu_err_set( DM_ERR_INTERNAL, "Failed to make subspace table in TEXT kernel" );
      for ( ii = 0; ii < nkeys; ii++ )
      {
        free( keycmt[ii] );

        if ( keytype[ii] == 'c' )
          free( keyval[ii] );
      }

      free( name );
      return;
    }


    /* If we use the table mechanism for other than TIME, we'll pass the   */
    /* column names (equiv to START, STOP) in as dd->eltNames. Recall that */
    /* subspaces are dmRANGE element type so dd->edim will be 2 * dd->dim  */
    /* when all is working. At this time dd->edim/eltNames stuff is not    */
    /* implemented properly.                                               */


    /* Create the columns */

    new_dd = dmi_column_create_subtable( block, dd );
 
 
    /* In the old DM we make two scalar columns; in the new we support      */
    /* making a single vector column with two components, which is cleaner. */
    /* Alas, this would change behaviour in certain cases: namely, if       */
    /* a CIAO tool tries to read the GTI extensions directly                */
    /* The chances that this happens and would mess things up is too great, */
    /* and I regretfully retain the old behaviour by default.               */
#if 1
    if ( dm_global & DM_GLOBAL_OLD )
#endif
    {
      cols[0] = dmi_descriptor_get_cpt( new_dd, 1 );
      cols[1] = dmi_descriptor_get_cpt( new_dd, 2 );
      afColCreate( block, cols[0], 0, NULL );
      afColCreate( block, cols[1], 0, NULL );
    }
    else
    {
      afColCreate( block, new_dd, 0, NULL );
      cols[0] = dmi_descriptor_get_cpt( new_dd, 1 );
      cols[1] = dmi_descriptor_get_cpt( new_dd, 2 );
    }
  } /* end else */

 
  /* Copy keys */
  for ( ii = 0; ii < nkeys; ii++ )
  {
    //status = dmSUCCESS;

    /* NOTE: write keys to subspace table block IFF they are in parent */
    if ( ukeys[ii] )
    {
      if ( keytype[ii] == 'c' ) 
        afp_key_write( block, 0, keyname[ii], DT_CHAR,
                       &keyval[ii], 1, NULL, keycmt[ii] );
      else
        if ( keytype[ii] == 'i' )
          afp_key_write( block, 0, keyname[ii], DT_INT4,
                         &keyival[ii], 1, NULL, keycmt[ii] );
        else
          if ( keytype[ii] == 'd' )
            afp_key_write( block, 0, keyname[ii], DT_REAL8,
                           &keydval[ii], 1, NULL, keycmt[ii] );
    }

    if ( keytype[ii] == 'c' ) 
      free( keyval[ii] );
    free( keycmt[ii] );
  }


  /* Write the actual data */

   afColWrite( cols[0], NULL, NULL, NULL, 1, nvals, val1 );
   afColWrite( cols[1], NULL, NULL, NULL, 1, nvals, val2 );

   /* Update DM number of rows - if zero at block close, that would mess */
   /* up NAXIS2                                                          */
   dmi_block_set_nrows( block, nvals );

   afBlockClose( block );

   /* We close and free the block descriptors */
   dmi_block_close_descriptors( block, afDescriptorFree );
   free( name );
}


/***************************************************************/
/*                                                             */
/* Function:   afCoordWrite                                    */
/*                                                             */
/* Description:                                                */
/*             from ftkernel                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor  *dd                        */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afCoordWrite( dmDescriptor* dd )
{
  dmBlock       *block  = NULL;
  dmDescriptor  *dd_lp  = NULL;  /* Logical-to-physical descriptor */
  dmDescriptor  *cpt_dd = NULL;
  dmDescriptorType dtype;

  long   dim   = 0;
  char  *ttype = NULL;
  long   cptno = 0;
  double crval[dmMAXDIM];
  double cdelt[dmMAXDIM];
  double crpix[dmMAXDIM];
  int    ok    = 0;
  char  *ctype[dmMAXDIM];

  long  size = 80;
//  long  tno  = 0;
  char *cname[dmMAXDIM];
  long  flag = 0;
  int   is_physical = 0;
  dmDataset *ds = NULL;


  if ( !dd )
    return;

  block = dmi_descriptor_get_block( dd );
  ds = dmi_block_get_ds( block );

  if ( !dmi_dataset_mode( ds ))  /* WRITE TO FILE */
  {
    dmu_err_set( DM_ERR_INTERNAL, "Attempt to write coord to readonly file" );
    return;
  }

  dmi_descriptor_clear_update( dd );
  dtype = dmi_descriptor_get_type( dd );

  if ( dtype != dmCOORD )
  {
    sprintf( dm_errbuf, "Descriptor %s is not a coord descriptor",
            dmi_descriptor_get_name( dd ));
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return;
  }

  dim = dmi_descriptor_get_element_dim( dd );
//  tno = dmi_coord_get_transform_no( dd );

  flag = dmi_descriptor_get_flag( dd );

  /* Physical flag may not be correctly propagated to cpts - oops; */
  /* check it here                                                 */
  is_physical = ( flag & DM_FLAG_PHYS );

  if ( !is_physical )
    dd_lp = dmi_coord_get_parent( dd );

  if ( dd_lp )
    dmi_coord_convert_pw_to_lw( dd, dd_lp, crpix, crval, cdelt, dim );
  else
    dmi_coord_get_transform_d( dd, crpix, crval, cdelt, dim );

  ttype = dmi_coord_get_transform_type( dd );

  if ( cs_eq( ttype, "LINEAR" ))
    ttype = NULL;

  ok = UT_TRUE;

  for ( cptno = 1; cptno <= dim; cptno++ )
  {
    cpt_dd = dmi_descriptor_get_cpt( dd, cptno );
    cname[cptno-1] = dmi_descriptor_get_name( cpt_dd );
    ctype[cptno-1] = calloc( size, UT_SZ_C );
    ok &= dmi_wcs_make_ctype( cname[cptno-1], ttype, cptno,
                             ctype[cptno-1], UT_FALSE );
  }

  if ( dim > 1 )
    afp_coord_write_mtype( block, dd );

  for ( cptno = 1; cptno <= dim; cptno++ )
  {
    cpt_dd = dmi_descriptor_get_cpt( dd, cptno );
    afp_scalar_coord_write( cpt_dd, crpix[cptno-1], crval[cptno-1],
                            cdelt[cptno-1], ctype[cptno-1], cptno,
                            is_physical, ok );
    free( ctype[cptno-1] );
  } 
}


/***************************************************************/
/*                                                             */
/* Function:   afGetKernelKey                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock  *block                          */
/*                    long      no                             */
/*                    char     *buf                            */
/*                    long      maxlen                         */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afGetKernelKey( dmBlock* block, long no, char* buf, long maxlen )
{
  afBlock *ablock = NULL;

  char buf1[256];
  char m = ' ';
  char ct[CARD_SIZE] = " ";

  dmuKeyList keylist;
  dmuKey *ukey = NULL;
 
  ablock = (afBlock*)dmi_block_get_kb( block );

  cs_copy( " ", buf );

  if ( ablock )
  {
    keylist = afk_block_keylist( ablock );

    if ( no <= keylist->n )
    {
      ukey = keylist->data[no-1];
      ftp_keytype_name( (ftKeyType)ukey->keytype, ct, 80 );

      if ( ukey->mask )
        m = '*';

      dmu_key_print( ukey, NULL, dmFALSE, buf1, maxlen );
      sprintf( buf, "Key %4ld: %-14s%c%s", no, ct, m, buf1 );
    }
  }
}



/***************************************************************/
/*                                                             */
/* Function:   afGetNoKernelKeys                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock  *block                          */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            long nkeys                                       */
/*                                                             */
/***************************************************************/
long afGetNoKernelKeys( dmBlock* block )
{
  afBlock    *ablock = NULL;
  dmuKeyList  keylist;
  long       nkeys   = 0;

  ablock = (afBlock*)dmi_block_get_kb( block );
 
  if ( ablock )
  {
    keylist = ablock->keylist;
    nkeys = keylist->n;
  }

  return nkeys;
}


