/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afkernelP.h                                                 */
/*                                                               */
/*   Definitions and function prototypes for ASCII file type     */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#ifndef AFKERNELP_H
#define AFKERNELP_H

#include <stdio.h>
#include "dmilib.h"
#include "../ftkernel/ftkernelP.h"


#define AK_COMMENTLINE_LEN 1024
#define AK_KEYNAME_LEN     80
#define AK_KEYVAL_LEN      AK_COMMENTLINE_LEN
#define AK_KEYCOMMENT_LEN  AK_COMMENTLINE_LEN
#define AK_LINE_LEN        5096
#define AK_COLVAL_LEN      1024

#define AK_SUBSPACE  50

#define AK_SPECIFICATION  "KERNEL"

typedef enum { AK_FLAV_UNK=0, 
	       AK_FLAV_RAW, 
	       AK_FLAV_SIMPLE, 
	       AK_FLAV_DTF, 
	       AK_FLAV_RDB,
               AK_FLAV_CSCRDB,
               AK_FLAV_TSV,
               AK_FLAV_USER } afFlavour;





/* AK structure types */

typedef struct afDataset_s afDataset;
typedef struct afBlock_s   afBlock;
typedef struct afCol_s     afCol;
typedef struct afFile_s    afFile;
typedef struct afBuffer_s  afBuffer;
typedef struct afVector_s  afVector;

typedef struct afStyle_s afStyle;
typedef struct afHintData_s afHintData;


/* A file interface built on top of stdio.
 * So I can read/write to either a file or a memory buffer, and handle
 * seeking pitfalls if they appear.
 */
#define AK_FILE_LINEBUF_INCR 1024
#define AK_FILE_LINE_MINBUF  1024
#define AK_FILE_LINE_MAXBUF  10240


struct afFile_s {
 /* The actual file */
 FILE* fptr;  /* the file pointer if there is a file */

 /* the memory buffer if the input is not a seekable interface */
 char* membuf; /* the memory buffer */
 long bufmax; /* allocated size of the membuf */
 long bufsize; /* size of the data in the membuf */
 long blksize; /* size of a block */
 long bufpos;  /* current position in the memory buffer */

 /* the last line read */
 char* line; /* Pointer to the current line in the file */
 long llen;  /* size of memory allocated for line buffer */
};


/*
 * This is an attempt to make TEXT table reading as generic 
 * as possible.  I am trying to enumerate the aspects that
 * make TEXT files different.  There may be too many 
 * non-qualifiable differences for this to be possible
 */
#define AK_STYLE_COLNAMES_NONE 0
#define AK_STYLE_COLNAMES_FIRST 1
#define AK_STYLE_COLNAMES_LAST 2
#define AK_STYLE_COLNAMES_LM1 3
#define AK_STYLE_COLNAMES_LM2 4
#define AK_STYLE_COLNAMES_TSV 5
#define AK_STYLE_COLNAMES_UNKNOWN -1


/*   TEXT KERNEL hints   */
struct afHintData_s 
{
  afFlavour  flav;
  dmBool     set_flavor;
  dmBool     has_header;
  dmBool     set_header;
  char       header_char;
  dmBool     set_header_char;
  dmBool     fixed;
  dmBool     set_fixed;
  dmBool     multisep;
  dmBool     set_multisep;
  char      *quotes;
  dmBool     set_quotes;
  dmBool     use_fits;
  dmBool     set_fits;
  char      *sep;
  dmBool     set_sep;
  int        colnames;
  dmBool     set_colnames;
  int        skip;
  dmBool     set_skip;
  char      *nullstr;
  dmBool     set_nullstr;
};


struct afStyle_s {
  afFlavour   flavor;         /* kernel specification  (enumerated)*/
  char        flavorString[CARD_SIZE];   /* kernel specification text string */
   dmBool      has_header;    /* does this flavour have a header? */
   char        header_char;   /* the character specifying a header line if there is one */
   char       *key_flag;      /* how do I know if a header line is a keyword line? */
   dmBool      multiple;      /* does this flavour have multiple blocks? */
   dmBool      fixed;         /* Is this flavour a fixed width format? */
   char       *white;         /* whitespace characters */
   char       *sep;           /* field seperator characters */
   dmBool      multisep;      /* flavor allows multiple seperators between fields? */
   char       *quotes;        /* quote characters for string columns */
   dmBool      use_fits;      /* use FITS style keywords */
   long        skip;          /* skip this many lines until the text dataset begins */
   int         colnames;      /* Should I look for SIMPLE style column names and where? */
   char       *nullstr;       /* used as replacement value for null values */
};


/*
 * The kernel's view of a table column
 */
struct afCol_s {
 long type;     /* a hack to determine if this is a col, coord, or subspace */
 long no;       /* The TEXT data column number, 
		   this is managed by the afBlock */

 long width;    /* Column data width / string size */
 long fwidth;   /* The column's field width ( for fixed width data ) */
 char* name;    /* Column name */
 long length;   /* array length */
 long bytepos;  /* byte position of column in row (for fixed width data ) */
 dmDataType datatype; /* The column's data type */

 long mtype_index; /* mtype index for this vector, only used by fits style
		    * files */
 char* disp;   /* the display format of the column */
};

/*
 * Specifically for passing a kdd into kernel->ColCreate and
 * have access to the actual col kdd's.
 */
#define AFK_VECTOR 100
struct afVector_s {
 long type;
 long dim;
 afCol** cpts;
};

/* 
 * The kernel's view of the block 
 */
struct afBlock_s {
 afDataset* ads;       /* the dataset I belong to
		        * Please, don't change the dataset from the block. */
 /* --------------- Basic block info ------------------------------------- */
 long         no;              /* TEXT block number (1-based) */
 char*        name;            /* name of the block, usually just the filename */
 dmBlockType  block_type;      /* unknown, table or image */
 /* --------------- Header data ------------------------------------------ */
 dmuKeyList keylist;         /* Cached list of TEXT header keys */
 long       keyno;           /* last key read/written */
 /* --------------- Column data ------------------------------------------ */
 long      ncols;           /* Number of cols */
 long      nmax;            /* Cols allocated */
 afCol**   cols;            /* Column data */
 /* --------------- Misc. ------------------------------------------------ */ 
 afBuffer *abuf;            /* The block's buffer: the high level file interface */
 void*     hint;            /* hints for the block */
 long      max_dstyp;       /* Max index for DSTYPn keyword */
 long      naxes;           /* Number of axes for image */
 long      axes[FITS_MAXDIM];
};

/* 
 * The kernel's view of the dataset.
 *
 * The file pointer "belongs" to the dataset.  No other
 * structure should have a file pointer.
 */

struct afDataset_s {
 char* name;          /* File name */
 char** blocks;       /* Names of blocks */
 long* blocks_pos;  /* positions of blocks in file */
 long pos;            /* where does the dataset start in the file? */

 long n;              /* number of blocks */
 long nmax;           /* amount of space allocated for blocks */
 
 afStyle* astyle;      /* describes the flavour in programmer friendly way */
 afFile* afptr;       /* the file pointer for this dataset */
};


/*
 * A (somewhat) simple buffer.
 *
 * Note that the file pointer "belongs" to the dataset, but
 * can be accessed from the buffer through ablock->ads->afptr.
 * Same goes for the astyle pointer.
 */
#define AK_MAX_ROWBUF  200000 /* process 50 000 row file before wrapping*/
#define AK_MIN_ROWBUF  1024   /* start with 1024/sizeof(long) rows */
#define AK_ROWBUF_INCR 1024   /* Row buffer increment */
#define AK_WRITEBUF_ROW 128   /* a guess at the row size, 
			       * doesn't need to bound the actual row sizes */
#define AK_WRITEBUF_NROWS 1024 /* A maximum number of rows for the write */
/* the write buffer size, 
 * enough to hold 1024 rows each of 128 bytes in length,
 * This better be bigger than the size of one row... */
#define AK_MIN_WRITEBUF (AK_WRITEBUF_ROW * AK_WRITEBUF_NROWS)

#define AK_MIN_HEADBUF 1024
#define AK_HEADBUF_INCR 80
#define AK_HEADBUF_MINROWS 10
#define AK_HEADBUF_ROWINCR 10

#define AK_PARSEBUF_INCR 80

struct afBuffer_s {
 afBlock* ablock; /* the block I belong to.  
		   * Please, don't change the block from the buffer */

 /* --------------- Info to help parse the data -------------------------- */
 afStyle* astyle;

 /* --------------- Basic file structure info ---------------------------- */
 long data_start;
 long ncols;        /* number of table columns */
 long ndcols;       /* number of data columns */
 long* col_offsets; /* data col number for each table column */

 /* --------------- Row position buffer ---------------------------------- */
 long *rrows; /* row positions */
 long   rrow1; /* first valid row */
 long   rrow2; /* last valid row */
 long   nrrows; /* number of rows allocated */
 long   maxrrows; /* maximum number of rows that can be allocated */

 /* --------------- The Parse Buffer ------------------------------------- */
 /* Note: For now the read buffer only holds one row at a time.
  *       If this turns out to be insufficient, a new scheme must
  *       me designed.
  */
 char** pdata;   /* pointer to the beginning of each data column in each
		   row */
 long*  pwidths; /* the string width of each data col */
 long   prow1;     /* number of first row in buffer */
 long   prow2;     /* number of last row in buffer */
 char  *parsebuf; /* raw data buffer */
 long   psize; /* length of the row contained in the parse buffer */
 long   prsize; /* size of the row as it existed before parsing */
 long   pmax;      /* number of bytes allocated to the parse buffer */
 long  *pmaxwidths; /* maximum width of each column (if fixed width) */
 long   prowwidth; /* the width of a row / sum of pmaxwidths (if fixed width) */
 char** pfmts; /* the c display format for each column */
 dmBool align; /* if fixed width, does the file need to be re-aligned? */

 /* --------------- The Data Write Buffer -------------------------------- */
 char*  writebuf; /* the write buffer */
 long wmax;     /* memory allocated for write buffer */
 long wsize;    /* the number of bytes pending in the write buffer */
 long wrsize;   /* the number of bytes in the file that the pending buffer will
		 * replace */
 char** wrows;   /* pointers to each row in the write buffer */
 long wmaxrows;  /* number of rows allocated to wrows */
 long wrow1;    /* first row in write buffer */
 long wrow2;    /* last row in write buffer */

 /* --------------- The Header Buffer ------------------------------------ */
 char* headbuf;  /* the raw buffer */
 long hmax;      /* memory allocated for header buffer */
 long hsize;     /* number of bytes pending in header buffer */
 long hrsize;    /* number of bytes that head buffer will replace */
 char** hrows;   /* pointers to each row in the header buffer */
 long hmaxrows;  /* number of rows allocated to hrows */
 long hrow1;     /* first row in header buffer */
 long hrow2;     /* last row in header buffer */
 dmBool hupdate; /* Does the header need to be updated? */
};



/* Dataset routines */
dmDataset* afp_dataset_create( const char* filename, const char* hint );

dmErrCode afp_dataset_scan( dmDataset* ds );
void afp_dataset_close( dmDataset* ds, dmBool delete );
void afp_dataset_add_block( dmDataset* ds, long no, dmBlockType type, const char *name, long pos );
void afp_dataset_delete_block( dmDataset* ds, long blockno );
dmBool  afp_verify_and_set_flavor( const char *pcLine, afFlavour *akFlavor, dmBool *bFixed);


/* Kernel dataset routines */
afDataset* afk_dataset_alloc( void );
dmErrCode afk_dataset_open( afDataset* ads, const char* filename, dmBool mode, dmBool create );
void afk_dataset_close( afDataset* ads, dmBool delete );
void afk_dataset_free( afDataset* ads );
const char* afk_dataset_block_name( const afDataset* ads, long no );
dmErrCode afk_dataset_resize_blocks( afDataset* ads, long no );
void afk_dataset_set_flavour( afDataset* ads, afFlavour flav );
afFlavour afk_dataset_get_flavour( afDataset* ads );
afFile* afk_dataset_fptr( const afDataset* ads );
afStyle* afk_dataset_get_style( const afDataset* ads );
dmErrCode afk_dataset_moveto_block( afDataset* ads, long no );
void afp_dataset_notify_moved( dmDataset* ds, dmBlock* block, long moved );

/* Block public routines */
dmBlock* afp_block_open( dmDataset* ds, long no, dmBool create, dmBool open_subspace, const char* hint );
dmErrCode afp_block_close( dmBlock* block );
dmErrCode afp_block_read_header( dmBlock* block );
dmErrCode afp_block_init_cols( dmBlock* block );
long afp_find_interesting_block_no( dmDataset* ds );
void afp_block_flush( dmBlock* block );
dmErrCode afp_set_hint_list( dmBlock* block, const char* hint );
void afp_dataset_set_hint_list( dmDataset* ds, const char* hint );
dmErrCode afp_set_hint( dmBlock* block, const char* name, const char* value );
void afp_dataset_set_hint( dmDataset* ds, const char* name, const char* value );
afFlavour afp_hint_kernel( dmBlock* block );
void afp_set_style_hints( dmBlock* block, afStyle* astyle );
void afk_hint_free( afBlock* ablock );
void afp_block_print( const dmBlock* block );
void afp_block_print_cols( const dmBlock* block );
void afp_block_print_keys( const dmBlock* block );
dmErrCode afp_block_read_col( dmBlock* block, dmDescriptor* dd, long* llpix, long* urpix, long* dpix, long row1, long row2, void* value );
dmuKey* afp_block_key_write( dmBlock* block, long pos, const char* name, dmDataType type, const void* value, const char* unit, const char* comment );
dmErrCode afp_block_write_comment( dmBlock *block, long keyno, const char* tag, const char* comment );
dmBlock* afp_block_insert( dmDataset* ds, const char* name, long no, const dmBlockType type, void* kb, char* hint );
void afp_block_notify_moved( dmBlock* block, dmBlock* moved_block, long moved );
/* functions taken from ftkernel */
dmErrCode afp_block_open_subspace( dmBlock* block, dmBool create );

/* Kernel block routines */
afBlock* afk_block_alloc( void );
void afk_block_free( afBlock* ablock );
dmuKeyList afk_block_keylist( afBlock* ablock );
dmErrCode afk_block_resize_cols( afBlock* ablock, long no );
afBuffer* afk_block_get_buffer( afBlock* ablock );
dmErrCode afk_block_set_row( afBlock* ablock, long row );
afCol* afk_block_get_col( const afBlock* ablock, long colno );
long afk_block_ncols( const afBlock* ablock );
void afk_block_append_col( afBlock* ablock, afCol* acol );
void afk_block_insert_col( afBlock* ablock, afCol* acol, long colno );
dmErrCode afk_block_read_col( afBlock* ablock, afCol* acol, 
			      dmDataType type, long row1, long row2, 
			      void* value, long slen, void* nullval, long imageOffset );
afDataset* afk_block_get_kds( const afBlock* ablock );
void afk_block_set_kds( afBlock* ablock, afDataset* ads );

/* ryan, need to implement */
void afk_block_delete_col( afBlock* ablock, long colno );

/* Public column routines */
dmDescriptor* afp_col_match_axis( dmDescriptor* col_dd, long axisno, char sub );
void afp_col_attach_wcs( dmDescriptor* col_dd, dmDescriptor* coord_dd );
dmErrCode afp_col_read( dmDescriptor* col, const long* llpix, const long *urpix, const long *dpix, long row1, long row2, void* data );
void afp_create_col( dmBlock* block, dmDescriptor* dd, long no, void* kdd, const char* hint );

/* Column routines that should replace ftkernel routines */

/* this one's called ftp_coord_match_col() or something */
dmDescriptor* afp_col_find_coord( dmDescriptor* col_dd, long axisno, char sub );
void afp_col_attach_wcs( dmDescriptor* col_dd, dmDescriptor* coord_dd );
dmDescriptor* afp_col_create_default_wcs( dmDescriptor* col_dd, long axisno, char sub );
dmDescriptor* afp_col_create_wcs( dmDescriptor* col_dd, long axisno, char sub, const char* name, const char* transform, dmBool composite, const char *ctype_name );
void afp_col_update_physical( dmDescriptor* col_dd );


/* KERNEL COLUMN ROUTINES */
afCol* afk_col_alloc( void );
afCol* afk_col_alloc_copy( const afCol* acol );
void afk_col_free( afCol* acol );
void afk_col_print( afCol* col );

afVector* afk_vector_alloc( long dim );
void afk_vector_free( afVector* avec );

/* Coord Routines */
/* taken from ftkernel */
void afp_coord_write_mtype( dmBlock* block, dmDescriptor* dd );
void afp_scalar_coord_write( dmDescriptor* dd, double crpix, double crval, double cdelt, char* ctype, long cptno, int is_physical, int ok );

/* KERNEL COORD ROUTINES */


/* Public subspace routines */
/* functions taken from ftkernel */
ftSubspaceCol* afp_get_ftss( dmBlock* block, dmDescriptor* dd, long n );
ftSubspaceCol* afp_ftss_alloc( afBlock* ablock, long n );
dmErrCode afp_read_subspace_values( dmDescriptor* dd, long subspace_cptno, char* dsval, char* tabname, int create );
dmErrCode afp_subspace_read_table( dmDescriptor* dd, const char* tabname, dmBool create, void** val1, void** val2, long* nvals, regRegion** ss_region );
regRegion* afp_block_read_region( dmBlock* block );
long afp_block_read_gti( dmBlock* block, void** val1, dmDataType *type1, void** val2, dmDataType* type2 );

/* ------------------ Buffer kernel routines --------------------------- */
/* buffer allocation/deallocation */
afBuffer* afk_buffer_alloc( void );
void afk_buffer_free( afBuffer* abuf );
/* data parsing information */
void afk_buffer_set_style( afBuffer* abuf, afStyle* astyle );

/* Buffer routines */
dmErrCode afk_buffer_sync( afBuffer* abuf, const afBlock* ablock, dmBool reset );
void afk_buffer_set_datastart( afBuffer* abuf, long data_start );
long afk_buffer_get_datastart( const afBuffer* abuf );
void afk_buffer_set_kb( afBuffer* abuf, afBlock* ablock );
dmErrCode afk_buffer_read_col( afBuffer* abuf, long row1, long row2, long col, dmDataType type, 
			       void* values, long slen, long nvals, void* nullval, dmBlockType blockType, long imageOffset );
dmErrCode afk_buffer_write_col( dmDescriptor* dd, long row1, long row2, void* values, long* moved );
long afk_buffer_get_nrows( afBuffer* abuf );
dmErrCode afk_buffer_append_col( afBuffer* abuf, afCol* acol );
dmErrCode afk_buffer_insert_col( afBuffer* abuf, afCol* acol, long colno, long* moved );
dmErrCode afk_buffer_flush( afBuffer* abuf, long* moved );
long afk_buffer_get_col_byteoffset( afBuffer* abuf, long colno );
long afk_buffer_get_col_maxwidth( afBuffer* abuf, long colno, long ar_length );
void afk_buffer_get_col_cfmt( afBuffer* abuf, long colno, char* cfmt );

/* header routines */
dmErrCode afk_buffer_add_keyline( afBuffer* abuf, const char* keyline );
void afk_buffer_delete_keyline( afBuffer* abuf, long lineno );
dmErrCode afk_buffer_insert_keyline( afBuffer* abuf, const char* keyline, long lineno );
dmErrCode afk_buffer_replace_keyline( afBuffer* abuf, const char* keyline, long lineno );
void afk_buffer_notify_moved( afBuffer* abuf, long moved );
dmErrCode afk_buffer_write_rows( afBuffer* abuf, dmBuffer* buffer, long row1, long row2, long* moved );
dmErrCode afk_buffer_read_rows( afBuffer* abuf, dmBuffer* buffer, long row1, long row2 );



/* should this be public? */
long afk_buffer_moveto_row( afBuffer* abuf, long row );

/* TEXT file routines */
afFile* afk_file_open( const char* filename, dmBool mode, dmBool create );
afFile* afk_file_load( const char* filename );
dmErrCode afk_file_seek( afFile* afptr, long offset, int whence );
long afk_file_tell( afFile* afptr );
const char* afk_file_getline_ptr( afFile* afptr );
long afk_file_delete_size( afFile* afptr, long size );
long afk_file_insert_data( afFile* afptr, char* data );
long afk_file_replace_data( afFile* afptr, char* data, long dsize, long size );
int afk_file_puts( afFile* afptr, char* s );
void afk_file_close( afFile* afptr );
dmBool afk_file_okay( afFile* afptr );
void afk_file_flush( afFile* afptr );
long afk_file_get_llen( afFile* afptr );

/* Public Descriptor Routines */

/* Kernel keyword functions */

/* is the keyword private? */
dmBool afk_key_private( const char* name );
void afk_key_parse_val( const char* valstr, dmDataType type, void* val );
dmDataType afk_key_get_type( const char* keyval, long vlen );
void afk_key_delete( afBlock* ablock, dmuKey* ukey );
void afk_key_default_fmt( dmDataType type, const void* value, char* val, afStyle* astyle );
ftKeyType afk_datatype_to_keytype( dmDataType type );

/* KERNEL GENERIC ROUTINES */
void afk_gen_read_comment( const afStyle* astyle, const char* line, char* comment, long clen);

/* read from a keyword line.  Return failure if not a valid keyword line */
dmErrCode afk_gen_read_keyline( const afStyle* astyle, const char* line, char* name, long nlen, char* value, long vlen, char* comment, long clen,  char* unit, long ulen );
const char* afk_gen_read_next_col( const afStyle* astyle, const char* lbegin,
				   char* colval, long vlen);
const char* afk_gen_find_next_val( afStyle* astyle, const char* line, long* vlen );



/* GENERIC routines that should replace ftkernel routines */
void afp_gen_find_transforms( dmDescriptor** cols, long ncols, dmuKeyList keylist );
void afp_apply_coord_params( dmDescriptor** cols, long ncols, double* vals, int* found, char *pRadesys );
void afp_update_physical_coords( dmDescriptor** cols, long ncols );
dmDescriptor* afp_coord_find_key( dmDescriptor** cols, long ncols, const char* key, const char* key_prefix );
void afp_compose_tform( dmDataType type, long isize, long length, long width, dmBool ascii, dmBool variable, char* tform );
void afp_compose_tdisp( dmDataType type, long width, char* tdisp );
dmDataType afp_parse_tform( const char* tform, dmBool ascii, long* arraysize, long* length, long* width, dmBool* variable );
void afp_parse_tdisp( const char* tdisp, char* typecode, long* width, long* precision, long* exponent );
void afk_gen_determine_col_datatype(  const afStyle* astyle, afCol *acol, char *colval, dmDataType *type);
void afk_gen_final_col_datatype( afBlock *ablock, afStyle *astyle);

/* KERNEL RAW ROUTINES */
void afp_raw_create_block( dmBlock* block );
dmErrCode afp_raw_init_cols( dmBlock* block );
dmErrCode afk_raw_read_header( afBlock* ablock, afStyle* astyle );
dmErrCode afk_raw_init_cols( afBlock* ablock, const afStyle* astyle );
dmErrCode afk_raw_read_cols( afBlock* ablock, const afStyle* astyle, const char* line );
dmuKey* afk_raw_key_write( afBlock* ablock, long pos, const char* name, dmDataType type, const void* value, const char* unit, const char* desc );
dmErrCode afk_raw_write_comment( afBlock* ablock, long pos, const char* tag, const char* comment );
long afk_raw_get_nrows( afBlock* ablock );

/* KERNEL SIMPLE ROUTINES */
void afp_simple_create_block( dmBlock* block );
dmErrCode afk_simple_read_header( afBlock* ablock, afStyle* astyle );
dmErrCode afp_simple_init_cols( dmBlock* ablock );
dmErrCode afk_simple_init_cols( afBlock* ablock, afStyle* astyle );
dmErrCode afk_simple_read_cols( afBlock* ablock, const afStyle* astyle, const char* line );
dmuKey* afk_simple_key_write( afBlock* ablock, long pos, const char* name, dmDataType type, const void* value, const char* unit, const char* desc );
dmErrCode afk_simple_write_comment( afBlock* ablock, long pos, const char* tag, const char* comment );
long afk_simple_get_nrows( afBlock* ablock );
dmDescriptor** afp_simple_make_vectors( dmBlock* block, afCol*** acpts,
                                        long ncols, long *dims, dmBool bUseUnits );


/* KERNEL DTF ROUTINES */
void afp_dtf_create_block( dmBlock* block );
dmErrCode afp_dtf_dataset_scan( dmDataset* ds );
dmErrCode afp_dtf_init_cols( dmBlock* block );
dmErrCode afk_dtf_read_header( afBlock* ablock, afStyle* astyle );
dmErrCode afk_dtf_init_cols( afBlock* ablock, afStyle* astyle );
dmErrCode afk_dtf_init_image_cols(afBlock* ablock, afStyle* astyle );
dmuKey* afk_dtf_key_write( afBlock* ablock, long pos, const char* name, dmDataType type, const void* value, const char* unit, const char* desc );
void afp_dtf_insert_col( dmBlock* block, dmDescriptor* dd, long no, void* ikdd );
void afp_dtf_insert_image_col( dmBlock* block, dmDescriptor* dd, long no, void* ikdd );
void afp_simple_insert_col( dmBlock* block, dmDescriptor* dd, long no, void* ikdd );
void afp_raw_insert_col( dmBlock* block, dmDescriptor* dd, long no, void* ikdd );
long afk_dtf_next_mtype_index( afBlock* ablock );
dmErrCode afk_dtf_write_comment( afBlock* ablock, long pos, const char* tag, const char* comment );
long afk_dtf_get_nrows( afBlock* ablock );
void afp_dtf_flush_col_keys( dmBlock* block );
void afp_dtf_image_set( dmBlock* block, const dmDataType type, const char* unit, const long* axes, const long naxes );

/* DTF routines that should replace ftkernel routines */
dmDescriptor** afp_dtf_vectorize( dmDescriptor** cols, long ncols, dmuKeyList keylist, long* nvcols );
dmDescriptor** afp_dtf_map_vectors( dmDescriptor** cols, long ncols, dmuAliasKeyList mlist, long* nvcols );
dmDescriptor* afp_dtf_column_match( dmDescriptor** cols, long ncols, long ftaxisno, dmuAliasKeyList mlist );
dmDescriptor* afp_dtf_dd_match( dmDescriptor* col_dd, dmDescriptor** cols, long ncols, dmuAliasKeyList mlist );
dmDescriptor* afp_dtf_implicit_vector_transform( dmDescriptor* transform_dd,
						 dmDescriptor** cols,
						 long ncols );
dmDescriptor* afp_dtf_dd_inherit( dmDescriptor* transform_vector );
void afp_dtf_axis_match( dmDescriptor** cols, long ncols, long ftaxisno,
			 dmuAliasKeyList mlist );
long afk_dtf_imag_get_ncols( afBlock* ablock, const afStyle *astyle );


/* KERNEL CSCRDB ROUTINES */

void  afp_cscrdb_create_block( dmBlock *block );
afFlavour  afp_cscrdb_dataset_scan( afDataset *pakAsciiDataSet,
                                    dmDataset* ds,
                                    afStyle  *pakStyle);
dmErrCode  afp_cscrdb_init_cols( dmBlock *pBlock);
dmErrCode  afp_cscrdb_insert_col( dmBlock *pBlock, dmDescriptor *pDescriptor,
                                  long colNumber, void *pKernelDescriptor );

dmErrCode  afp_cscrdb_key_compose_colnames( dmDescriptor *pDescriptor, afStyle *pakStyle,
                                           char *currKeyValue, char *newEntry, char  cSeparator,
					   dmBool append, long colNumber, char *newKeyValue );
dmErrCode  afp_cscrdb_write_col_header( dmBlock *pBlock, dmDescriptor *pDescriptor,long colNumber );

dmErrCode afk_cscrdb_attach_units( dmBlock *block, dmDescriptor **rawcols, long numColumns);
dmErrCode afk_cscrdb_init_cols( afBlock* ablock, afStyle* astyle );
dmBool afk_cscrdb_is_dashline( const char * comment, const char *sSeparator );
dmuKey* afk_cscrdb_key_write( afBlock* pakBlock, long lKeyPosition,
                              const char* sKeyName,
                              dmDataType dataType, const void* pValue,
                              const char* sUnit, const char* sDescription );
dmErrCode afk_cscrdb_read_header( afBlock* pakBlock, afStyle* pakStyle );
dmErrCode afk_cscrdb_write_comment( afBlock* pakBlock, long pos, const char* tag, const char* comment );



/* KERNEL TSV ROUTINES */
void afp_tsv_create_block( dmBlock *block );
dmErrCode afp_tsv_init_cols( dmBlock *pBlock);
dmErrCode  afp_tsv_insert_col( dmBlock *pBlock, dmDescriptor *pDescriptor,
			       long colNumber, void *pKernelDescriptor );
dmErrCode afp_tsv_key_compose_colnames( dmDescriptor *pDescriptor, afStyle *pakStyle, char *currKeyValue,
                                        char *newEntry, char  cSeparator, dmBool append,
                                        long   colNumber, char *newKeyValue );
dmErrCode afp_tsv_write_col_header( dmBlock *pBlock, dmDescriptor *pDescriptor, long colNumber );
dmErrCode afk_tsv_init_cols( afBlock* ablock, afStyle* astyle, char ***ppUnitStrings, char ***ppDescriptions );
dmuKey* afk_tsv_key_write( afBlock* pakBlock, long lKeyPosition, const char* sKeyName, dmDataType dataType,
                           const void* pValue, const char* sUnit, const char* sDescription );
dmErrCode afk_tsv_read_header( afBlock* pakBlock, afStyle* pakStyle );
dmErrCode afk_tsv_write_comment( afBlock* pakBlock, long pos, const char* tag, const char* comment );




/* KEYWORD routines */
long afp_key_pos( dmBlock* block, long keyno ); /* taken straight from ftkernel */
dmDescriptor* afp_key_write( dmBlock* block, long keyno, const char* name, dmDataType type, const void* value, long nvals, const char* unit, const char* comment );
dmBool afp_comment_write( dmBlock* block, long keyno, const char* tag, const char* comment );


/* KERNEL UTILITY ROUTINES */

const char* afk_keyline_read_name( const char* ifirst, 
				   char* ofirst, long len );
const char* afk_keyline_read_value( const char* ifirst,
			      char* ofirst, long len );
const char* afk_keyline_read_comment( const char* ifirst, char* ofirst, long len, char *unit, long ulen );
long afk_count_words( const char* buf, const char* sep );
void afk_default_fmt( dmDataType type, const void* value, char* val, afStyle* astyle );
void afk_tdisp_to_cfmt( dmDataType dataType, const char* tdisp, char* fmt );
void afk_cfmt_to_tdisp( const char* cfmt, long col_width, char* tdisp );
dmDataType afk_tdisp_code_to_datatype( char tchar );
void afk_strip_value( char *pToken, afStyle *pakStyle );
char* afk_prepare_string( char* istr, char* cfmt, afStyle* astyle );



/* TEXT kernel style functions */
afStyle* afk_style_alloc( void );
void afk_style_free( afStyle* astyle );
void afk_style_set( afStyle* astyle, afFlavour flav );
void afk_style_set_header( afStyle* astyle, dmBool has_header );
void afk_style_set_header_char( afStyle* astyle, char header_char );
void afk_style_set_key_flag( afStyle* astyle, const char* key_flag );
void afk_style_set_colspec( afStyle* astyle, dmBool colspec );
void afk_style_set_multiple( afStyle* astyle, dmBool multiple );
void afk_style_set_fixed( afStyle* astyle, dmBool fixed );
void afk_style_set_sep( afStyle* astyle, const char* sep );
void afk_style_set_multisep( afStyle* astyle, dmBool multisep );
void afk_style_set_quotes( afStyle* astyle, const char* quotes );
void afk_style_set_use_fits( afStyle* astyle, dmBool use_fits );
void afk_style_set_skip( afStyle* astyle, long skip );
void afk_style_set_colnames( afStyle* astyle, int colnames );
void afk_style_set_nullstr( afStyle* astyle, const char* nulltext );
int afk_style_get_colnames( afStyle* astyle, int colnames );
long afk_style_get_skip( afStyle* astyle );

dmBool afk_style_has_header( const afStyle* astyle );
char afk_style_get_header_char( const afStyle* astyle );
const char* afk_style_get_key_flag( const afStyle* astyle );
dmBool afk_style_get_multiple( const afStyle* astyle );
dmBool afk_style_get_fixed( const afStyle* astyle );
const char* afk_style_get_sep( const afStyle* astyle );
dmBool afk_style_get_multisep( const afStyle* astyle );
const char* afk_style_get_quotes( const afStyle* astyle );
const char* afk_style_get_nullstr( const afStyle* astyle );

dmBool afk_style_is_sep( const afStyle* astyle, char c );
dmBool afk_style_is_quote( const afStyle* astyle, char c );
dmBool afk_style_use_fits( const afStyle* astyle );

void afk_style_set_flavor ( afStyle *astyle, afFlavour flavor, const char *flavorString);
int  afk_style_get_flavor ( afStyle *astyle );

const char  *afk_style_get_flavorString( afStyle* astyle );

void afk_style_set_white( afStyle* astyle, char *white );
const char* afk_style_get_white( const afStyle* astyle );

void afp_set_kernel_hints( afStyle *pStyle, afHintData *pInHints );
void  afp_set_colnames_hint(long lColNamesStyle, dmBool bSetColNames);
void afp_clear_kernel_hints(afHintData  *pInHints);

int  afp_get_RowBufOverRun( void );
void afp_set_RowBufOverRun(int iValue);

#endif





