/*                                                                
**  Copyright (C) 2004,2006-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afkey.c                                                     */
/*                                                               */
/*   Functions that handle keylists in in ASCII type files       */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/


#include "afkernelP.h"



/***************************************************************/
/*                                                             */
/* Function:   afp_key_write                                   */
/*                                                             */
/* Description:                                                */
/*              Write keys                                     */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock    *block - block pointer        */
/*                    long        keyno - key number           */
/*                    const char *name - key name              */
/*                    dmDataType  type                         */
/*                    const void *value                        */
/*                    long        nvals                        */
/*                    const char *unit                         */
/*                    const char *comment                      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor  *dd                                */
/*                                                             */
/***************************************************************/
dmDescriptor* afp_key_write( dmBlock* block, long keyno, const char *name,
                             dmDataType type, const void* value,
                             long nvals, const char* unit,
                             const char* comment )
{
  long new_keyno = 0;
  long nkeys = 0;
  long pos   = 0;

  dmDescriptor  *dd = NULL;
  afBlock       *ablock = NULL;
  dmuKey        *ukey   = NULL;
  dmuKey        *old_ukey = NULL;

  nkeys = dmi_block_get_nkeys( block );
  ablock = (afBlock*)dmi_block_get_kb( block );
 
  if ( keyno > 0 )
    new_keyno = keyno + 1;

  dd = dmi_key_alloc( block, new_keyno, name, type, value, nvals, unit, comment );
  new_keyno = dmi_descriptor_get_no( dd );
  old_ukey = (dmuKey*)dmi_descriptor_get_kdd( dd );

  if ( old_ukey )
  {
    pos = old_ukey->index;
  }
  else
    if ( keyno == 0 )
      pos = 0;
    else
      pos = afp_key_pos( block, keyno ) + 1;
 

  ukey = afp_block_key_write( block, pos, name, type, value, unit, comment );

  if ( old_ukey && old_ukey != ukey )
  {
    /* We've changed ftkey.. hmm. Ignore this for now but perhaps worth */
    /* checking at verbose = n?                                         */  
  }

  if ( !ukey )
  {
    if ( !old_ukey )
      dmi_key_delete( dd );

    dd = NULL;
  }
  else
  {
    dmi_descriptor_set_kdd( dd, (void*)ukey );
    dmi_block_set_keyno( block, new_keyno );
  } 

  return dd;
}


/***************************************************************/
/*                                                             */
/* Function:   afp_key_pos                                     */
/*                                                             */
/* Description:                                                */
/*              Convert current DM key position to current     */
/*              FITS key positioin                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock    *block - block pointer        */
/*                    long        keyno - key number           */
/*                    const char *name - key name              */
/*                    dmDataType  type                         */
/*                    const void *value                        */
/*                    long        nvals                        */
/*                    const char *unit                         */
/*                    const char *comment                      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor  *dd                                */
/*                                                             */
/***************************************************************/
long afp_key_pos( dmBlock* block, long keyno )
{
  long kkeyno        = 0;
  dmDescriptor *dd   = NULL;
  dmuKey       *kkey = NULL;


  if ( keyno > 0 )
  {
    dd = dmi_block_get_key( block, keyno );

    if ( dd )
      kkey = dmi_descriptor_get_kdd( dd );

    if ( kkey )
      kkeyno = kkey->index;
  }

  return kkeyno;
}


/***************************************************************/
/*                                                             */
/* Function:   afp_comment_write                               */
/*                                                             */
/* Description:                                                */
/*              Write keys                                     */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock    *block - block pointer        */
/*                    long        keyno - key number           */
/*                    const char *tag                          */
/*                    const char *comment                      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor  *dd                                */
/*                                                             */
/***************************************************************/
dmBool afp_comment_write( dmBlock* block, long keyno,
                          const char* tag, const char* comment )
{
  afBlock    *ablock = NULL;
  afStyle    *astyle = NULL;
  afDataset  *ads    = NULL;
  dmuKeyList keylist;
  long ukeyno = 0;
  long offset = 0;


  ablock = (afBlock*)dmi_block_get_kb( block );

  if ( !ablock )
    return UT_FALSE;

  ads = ablock->ads;
  astyle = afk_dataset_get_style( ads );

  if ( !afk_style_has_header( astyle ))
    return UT_TRUE;

  keylist = ablock->keylist;

  ukeyno = afp_key_pos( block, keyno );

  if ( ukeyno > 0 )
  {
    offset = ukeyno - 1;

    if ( keylist )
      while ( offset < keylist->n && keylist->data[offset]->mask )
      {
        ++offset;
      }

    ukeyno = offset+1;
  }
  else
    if ( keylist  && keylist->n > 0 )
      ukeyno = keylist->n;
    else
      ukeyno = 1;

  afp_block_write_comment( block, ukeyno, tag, comment );

  return UT_TRUE;
}
