/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afkfile.c                                                   */
/*                                                               */
/*   Functions that create and work with ASCII file type         */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"
#include <stdio.h>
#include <unistd.h>



/* Local Functions */
static long afk_file_read( afFile* afptr, char* buf, size_t size, size_t nitems );
static long afk_file_write( afFile* afptr, char* buf, size_t size, size_t nitems );
static long afk_file_move_data( afFile* afptr, long offset );
static void afk_file_truncate( afFile* afptr );


/***************************************************************/
/*                                                             */
/* Function:   afk_file_open                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              const char  *filename                          */
/*              dmBool       mode                              */
/*              dmBool       create                            */
/*                                                             */
/*  Output parameters:                                         */
/*               none                                          */
/*                                                             */
/*  Returns:                                                   */
/*           afFile  *afptr  - file pointer                    */
/*                                                             */
/***************************************************************/
afFile* afk_file_open( const char* filename, dmBool mode, dmBool create )
{
  afFile *afptr  = NULL;
  FILE   *fptr   = NULL;
  char   *membuf = NULL;
  long    bufmax = 0;
  long    bufsize = 0;
  long    blksize = 2880;
  long    nread  = 0;
  long    nblk   = 0;

  afptr = (afFile*)malloc( sizeof(afFile) );

  if ( !afptr )
    return NULL;

  if ( mode )
  {
    if ( create )
    {
      if ( !( fptr = fopen( filename, "w" )))
      {
        free( afptr );
        return NULL;
      }

      fclose( fptr );

      if ( !( fptr = fopen( filename, "r+" )))
      {
        free( afptr );
        return NULL;
      }
    }
    else
    {
      /* writing to stdout, use memory buffer */
      if ( strcmp( filename, "-" ) == 0 )
      {
        bufmax = blksize;
        membuf = (char*)malloc( bufmax*sizeof(char) );

        if ( membuf )
          membuf[0] = '\0';
      }
      else
      {
        if ( !( fptr = fopen( filename, "r+" )))
        {
          free( afptr );
          return NULL;
        }
      }
    }
  }
  else
  {
    /* reading from stdin, use memory buffer */
    if ( strcmp( filename, "-" ) == 0 )
    {
      bufmax = blksize;
      membuf = (char*)malloc( bufmax*sizeof(char) );

      while( membuf && 
	   ((nread = fread( membuf + blksize*nblk, sizeof(char), blksize,
             stdin ))   == blksize ))
      {
        ++nblk;
        bufsize += nread;
        bufmax += blksize;
        membuf = (char*)realloc( membuf, bufmax );
      }

      if ( !membuf )
      {
        free( afptr );
        return NULL;
      }

      if ( nread && nread < blksize )
      {
        membuf[nblk*blksize + nread] = '\0';
        bufsize += nread;
      }
    }
    else
    {
      fptr = fopen( filename, "r" );
    }
  }

  afptr->fptr    = fptr;
  afptr->membuf  = membuf;
  afptr->bufmax  = bufmax;
  afptr->bufsize = bufsize;
  afptr->blksize = blksize;
  afptr->bufpos  = 0;


  if ( fptr || membuf )
  {
    /* allocate the line buffer */
    afptr->line = calloc( AK_FILE_LINE_MINBUF, sizeof(char) );

    if ( afptr->line )
      afptr->llen = AK_FILE_LINE_MINBUF;
    else
      afptr->llen = 0;
  }

  return afptr;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_file_load                                   */
/*                                                             */
/* Description:                                                */
/*     Load an entire file into the afFile* membuf for reading */
/*                                                             */
/*  Input parameters:                                          */
/*              const char  *filename                          */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*               afFile  *afptr  - file pointer                */
/*                                                             */
/***************************************************************/
afFile* afk_file_load( const char* filename )
{
  afFile  *afptr  = NULL;
  FILE    *fptr   = NULL;
  char    *membuf = NULL;
  long     bufmax = 0;
  long     blksize = 2880;
  long     bufsize = 0;
  long     nread  = 0;
  long     nblk   = 0;
 
 
  fptr = fopen( filename, "r" );

  if ( !fptr )
    return NULL;

  afptr = (afFile*)malloc( sizeof(afFile) );

  if ( !afptr ) 
    return NULL;

  bufmax = blksize;
  membuf = (char*)malloc( bufmax*sizeof(char) );

  while( membuf && 
	((nread = fread( membuf + blksize*nblk, sizeof(char), blksize, fptr ))
	 == blksize ))
  {
    ++nblk;
    bufsize += nread;
    bufmax += blksize;
    membuf = (char*)realloc( membuf, bufmax );
  }

  if ( !membuf )
  {
    free( afptr );
    return NULL;
  }

  if ( nread && nread < blksize )
  {
    membuf[nblk*blksize + nread] = '\0';
    bufsize += nread;
  }

  afptr->fptr    = NULL;
  afptr->membuf  = membuf;
  afptr->bufmax  = bufmax;
  afptr->bufsize = bufsize;
  afptr->blksize = blksize;
  afptr->bufpos  = 0;

  if ( membuf )
  {
    /* allocate the line buffer */
    afptr->line = calloc( AK_FILE_LINE_MINBUF, sizeof(char) );

    if ( afptr->line )
      afptr->llen = AK_FILE_LINE_MINBUF;
    else
      afptr->llen = 0;
  }

  return afptr;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_file_close                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              afFile  *afptr                                 */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*               afFile  *afptr  - file pointer                */
/*                                                             */
/***************************************************************/
void afk_file_close( afFile* afptr )
{
  if ( afptr )
  {
    if ( afptr->fptr ) 
    {
      fclose( afptr->fptr );
      afptr->fptr = NULL;
    }

    if ( afptr->membuf ) 
    {
      free( afptr->membuf );
      afptr->membuf = NULL;
    }

    if ( afptr->line ) 
    {
      free( afptr->line );
      afptr->line = NULL;
    }

    free( afptr );
  }
}



/***************************************************************/
/*                                                             */
/* Function:   afk_file_okay                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              afFile  *afptr                                 */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*              dmBool  status                                 */
/*                                                             */
/***************************************************************/
dmBool afk_file_okay( afFile* afptr )
{
  if ( afptr && ( afptr->fptr || afptr->membuf ))
    return dmTRUE;

  return dmFALSE;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_file_seek                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              afFile  *afptr                                 */
/*              long     offset                                */
/*              int      whence                                */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*             dmErrCode  status                               */
/*                                                             */
/***************************************************************/
dmErrCode afk_file_seek( afFile* afptr, long offset, int whence )
{
  dmErrCode status = dmFAILURE;

  if ( afptr )
  {
    if ( afptr->fptr )
    {
      if ( !fseek( afptr->fptr, offset, whence ))
        status = dmSUCCESS;
    }
    else
      if ( afptr->membuf )
      {
        switch ( whence )
        {
          case SEEK_SET:
            break;
          case SEEK_CUR:
            offset = afptr->bufpos + offset;
            break;
          case SEEK_END:
            offset = afptr->bufsize + offset;
            break;
        }

        if ( offset <= afptr->bufsize && offset >= 0)
        {
          afptr->bufpos = offset;
          status = dmSUCCESS;
        }
      }
  }

  return status;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_file_tell                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              afFile  *afptr                                 */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*             long  offset                                    */
/*                                                             */
/***************************************************************/
long afk_file_tell( afFile* afptr )
{
  long offset = -1;

  if ( afptr )
  {
    if ( afptr->fptr )
    {
      offset = ftell( afptr->fptr );
    }
    else
      if ( afptr->membuf )
      {
        offset = afptr->bufpos;
      }
  }
  return offset;
}

/***************************************************************/
/*                                                             */
/* Function:   afk_file_getline_ptr                            */
/*                                                             */
/* Description:                                                */
/* return a pointer to the next line in the file.              */
/* Uses an internal line buffer.  A pointer to this            */
/* buffer is returned to the user.                             */
/* The user should not attempt to modify this line.            */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*    afFile      *afptr                                       */
/*    dmBlockType  blockType                                   */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*    const char  *ret                                         */
/*                                                             */
/***************************************************************/
const char* afk_file_getline_ptr( afFile* afptr )
{

  const char *ret   = NULL;      /* return                             */
  char  *buf        = NULL;      /* pointer to beginning of line       */
  char  *bptr       = NULL;      /* return from fgets (== buf || NULL) */
  long   fpos       = 0;         /* current position in file           */
  long   lBufLength = 0;         /* length of buffer to hold line      */
  long   lStrLength = 0;         /* length of line read                */
  long   ii         = 0;         /* index                              */

  char  *iptr     = NULL;
  char  *eiptr    = NULL;

  char  *optr     = NULL;
  char  *eoptr    = NULL;

  int    bDone      = 0;         /* flag when have full line           */

  if ( ! afptr )
    return NULL;

  /* get buffer pointer and length of current buffer */
  buf = afptr->line;
  lBufLength = afptr->llen;

  if ( afptr->fptr )
  {
    /* get file position pointer */
    fpos = afk_file_tell( afptr );

    /* read from the file */
    bptr = fgets( buf, lBufLength, afptr->fptr );

    /* remove newline char */
    if ( bptr )
    {
      /* fgets() returns a NULL terminated string, so if full line is read the  */
      /* newline character will be last char in string.  Replace newline with   */
      /* null terminator and be done.                                           */

      lStrLength = strlen( bptr );
      ii         = lStrLength - 1;

      if ( bptr[ii] == '\n' )
      {
	bptr[ii] = '\0';
	ret = buf;
	bDone = 1;
      }
      else if ( lStrLength < (lBufLength - 1) )
      {
	/* Tweek above: lBufLength includes terminator, lStrLength */
        /*  does not, a full initial buf would be 1023 chars long. */

	/* Code Change  Bug 11278  September 25, 2008             */
	/* If a file is not terminated with a new line character, */
	/* this function goes into an infinite read loop looking  */
	/* for EOF.                                               */
	/* If the string returned has few characters than the     */
	/* buffer length and has no new line character, assume it */
	/* has reached the end.                                   */

	ret = buf;
	bDone = 1;
      }

    } /* end if bptr */
    else
    {
      /* bptr returned from fget() with invalid value.. */
      /* check for EOF; if it is, clear error           */
      if ( feof( afptr->fptr ))
      {
	clearerr( afptr->fptr );
      }
      bDone = 1;
    }

  }
  else if ( afptr->membuf )
  {
    /* read a line from memory buffer in same way as above */
    fpos = afptr->bufpos;

    iptr  = afptr->membuf + afptr->bufpos;
    eiptr = afptr->membuf + afptr->bufsize;

    optr  = buf;
    eoptr = buf + lBufLength;

    for ( ; iptr != eiptr && optr != eoptr && *iptr !='\n' && *iptr !='\0';
	  iptr++, optr++, afptr->bufpos++ )
    {
      *optr = *iptr;
    }

    if ( iptr == eiptr )
    {
      /* empty input (end of input) */
      bDone = 1;
    }

    if ( iptr != eiptr && *iptr == '\n' )
    { 
      /* end of line */
      afptr->bufpos++;
      bDone = 1;
    }

    if ( lBufLength )
      *optr = '\0';

    if ( optr != buf )
    {
      /* non-empty line.. set return ptr */
      ret = buf;
      bDone = 1;
    }

    if ( optr == eoptr && iptr != eiptr && *iptr != '\n' )
    {
      /* didn't fine end of line.. enlarge and try again */
      bDone = 0;
    }

  } /* end if ( afptr->membuf ) */

  if ( !bDone )
  {
    /* Didn't get the full line, expand buffer and try again */

    lBufLength += AK_FILE_LINEBUF_INCR;

    buf = realloc( buf, lBufLength*sizeof(char));
    if ( !buf )
      return NULL;

    afptr->line = buf;
    afptr->llen = lBufLength;

    afk_file_seek( afptr, fpos, SEEK_SET );

    ret = afk_file_getline_ptr( afptr );

  } /* end if !bDone */


  return ret;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_file_puts                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              afFile  *afptr                                 */
/*              char    *s                                     */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*             int  nchar                                      */
/*                                                             */
/***************************************************************/
int afk_file_puts( afFile* afptr, char* s )
{
  int   nchar = EOF;
  int   len   = 0;
  char *optr  = NULL;


  if ( afptr )
  {
    if ( afptr->fptr )
    {
      nchar = fputs( s, afptr->fptr );
    }
    else
      if ( afptr->membuf )
      {
        len = strlen(s) + 2;

        if ( len + afptr->bufpos > afptr->bufmax )
        {
          afptr->bufmax = len + afptr->bufpos;
          afptr->membuf = realloc( afptr->membuf, afptr->bufmax*sizeof(char) );
        }

        if ( afptr->membuf )
        {
          optr = afptr->membuf + afptr->bufpos;
          memcpy( optr, s, (len-2)*sizeof(char) );
          optr[len-2] = '\n';
          optr[len-1] = '\0';
          afptr->bufpos += len-1;
          afptr->bufsize += len-1;
          nchar = len-2;
        }
      }
  }

 return nchar;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_file_flush                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              afFile  *afptr                                 */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*             nothing                                         */
/*                                                             */
/***************************************************************/
void afk_file_flush( afFile* afptr )
{
  if ( afptr )
  {
    fflush( afptr->fptr );
  }
}



/***************************************************************/
/*                                                             */
/* Function:   afk_file_delete_size                            */
/*                                                             */
/* Description:                                                */
/*           Remove "size" bytes from the file.                */
/*                                                             */
/*  Input parameters:                                          */
/*              afFile  *afptr                                 */
/*              long     size                                  */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*             long  deleted                                   */
/*                                                             */
/***************************************************************/
long afk_file_delete_size( afFile* afptr, long size )
{
  long  pos = 0;
  long  deleted = 0; /* number of bytes deleted */

  if ( !afptr || size <= 0 )
    return 0;

  pos = afk_file_tell( afptr );
  afk_file_seek( afptr, size, SEEK_CUR );
  deleted = afk_file_move_data( afptr, -size );

  if ( deleted > 0 )
    deleted = size;
  else
    deleted = 0;

  return deleted;
}




/***************************************************************/
/*                                                             */
/* Function:   afk_file_insert_data                            */
/*                                                             */
/* Description:                                                */
/* Insert null terminated data at the current position in file.*/
/* Moving all data that was at the current position down as    */
/* many bytes as are in parameter data.                        */
/* Leave file pointer pointing after data just written.        */ 
/*                                                             */
/*  Input parameters:                                          */
/*              afFile  *afptr                                 */
/*              char    *data                                  */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*             long  size                                      */
/*                                                             */
/***************************************************************/
long afk_file_insert_data( afFile* afptr, char* data )
{
  long  size = 0; /* number of bytes in data */
  long  pos  = 0;
 
  if ( !afptr || !data )
    return 0;

  size = strlen( data );
  pos = afk_file_tell( afptr );
 
  afk_file_move_data( afptr, size );
  afk_file_seek( afptr, pos, SEEK_SET );
  afk_file_write( afptr, data, size, 1 );

  return size;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_file_replace_data                           */
/*                                                             */
/* Description:                                                */
/*   Replace size bytes of data with dsize bytes of the null   */
/*   terminated data string. Leave file pointer pointing right */
/*   after bytes just written.                                 */
/*   Return the number of bytes written to the file.           */
/*                                                             */
/*  Input parameters:                                          */
/*              afFile  *afptr                                 */
/*              char    *data                                  */
/*              long     dsize                                 */
/*              long     size                                  */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*             long  outbytes                                  */
/*                                                             */
/***************************************************************/
long afk_file_replace_data( afFile* afptr, char* data,
                            long dsize, long size )
{
  long  diff = 0;
  long  pos  = 0;
  long  outbytes = 0;

  if ( afptr )
  {
    if ( size == dsize )
    {
      afk_file_write( afptr, data, dsize, 1 );
    }
    else
      if ( size > dsize )
      {
        diff = dsize - size; /* will be negative */

        afk_file_write( afptr, data, dsize, 1 );

        pos = afk_file_tell( afptr );

        afk_file_seek( afptr, -diff, SEEK_CUR );
        afk_file_move_data( afptr, diff ); 
        afk_file_seek( afptr, pos, SEEK_SET );
      }
      else /* size < dsize */
      {
        diff = dsize - size; /* will be positive */
        pos  = afk_file_tell( afptr );

        afk_file_seek( afptr, size, SEEK_CUR );
        afk_file_move_data( afptr, diff );
        afk_file_seek( afptr, pos, SEEK_SET );
        afk_file_write( afptr, data, dsize, 1 );
      }

      if ( afptr->fptr )
        fflush( afptr->fptr );

      outbytes = dsize;
  }

  return outbytes;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_file_move_data                              */
/*                                                             */
/* Description:                                                */
/*        Move all data following the current file position    */
/*        position the specified signed offset amount          */
/*                                                             */
/*  Input parameters:                                          */
/*              afFile  *afptr                                 */
/*              long     offset                                */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*             long  total                                     */
/*                                                             */
/***************************************************************/
long afk_file_move_data( afFile* afptr, long offset )
{
  char *buf     = NULL;
  long  blksize = afptr->blksize;
  long  nbytes  = 0;
  long  total   = 0; /* total number of bytes moved */
  long  pos     = 0;
  long  end_pos = 0;
  long  nbytes_left = 0;


  if ( offset == 0 || ! afptr ) 
    return 0;

  buf = malloc( sizeof(char) * blksize );

  if ( ! buf ) 
    return total;

  if ( offset < 0 ) /* moving data backwards */
  {
    while ( ( nbytes = afk_file_read( afptr, buf, 1, blksize )) > 0 )
    {
      afk_file_seek( afptr, offset-nbytes, SEEK_CUR );
      total += afk_file_write( afptr, buf, 1, nbytes );
      afk_file_seek( afptr, -offset, SEEK_CUR );
    }

    /* now must shorten the file */
    afk_file_seek( afptr, offset, SEEK_CUR );
    afk_file_truncate( afptr );
  }
  else /* moving data forwards */
  {
   pos = afk_file_tell( afptr );
   afk_file_seek( afptr, 0, SEEK_END );
   end_pos = afk_file_tell( afptr );
   nbytes_left = end_pos - pos;

   while( nbytes_left > blksize )
   {
     afk_file_seek( afptr, -blksize, SEEK_CUR );
     afk_file_read( afptr, buf, 1, blksize );
     afk_file_seek( afptr, -blksize + offset, SEEK_CUR );
     total += afk_file_write( afptr, buf, 1, blksize );
     afk_file_seek( afptr, -blksize - offset, SEEK_CUR );
     nbytes_left -= blksize;
   }

   afk_file_seek( afptr, -nbytes_left, SEEK_CUR );
   afk_file_read( afptr, buf, 1, nbytes_left );
   afk_file_seek( afptr, -nbytes_left + offset, SEEK_CUR );
   total += afk_file_write( afptr, buf, 1, nbytes_left );
  }

  if ( buf )
    free( buf );

  return total;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_file_read                                   */
/*                                                             */
/* Description:                                                */
/*        Like fread except never set the eof flag             */
/*                                                             */
/*  Input parameters:                                          */
/*              afFile  *afptr                                 */
/*              char    *buf                                   */
/*              size_t   size                                  */
/*              size_t   nitems                                */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*             long  nbytes                                    */
/*                                                             */
/***************************************************************/
long afk_file_read( afFile* afptr, char* buf, size_t size, size_t nitems )
{
  long nbytes = 0;

  if ( !afptr ) 
    return 0;

  if ( afptr->fptr )
  {
    nbytes = fread( buf, size, nitems, afptr->fptr );

    if ( feof( afptr->fptr ))
      clearerr( afptr->fptr );
  }
  else
    if ( afptr->membuf )
    {
      if ( afptr->bufpos + (long)nitems > afptr->bufmax )
        nitems = afptr->bufmax - afptr->bufpos;

      memcpy( buf, afptr->membuf+afptr->bufpos, nitems );
      afptr->bufpos += nitems;
      nbytes = nitems;
    }

 return nbytes;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_file_write                                  */
/*                                                             */
/* Description:                                                */
/*        Like fwrite                                          */
/*                                                             */
/*  Input parameters:                                          */
/*              afFile  *afptr                                 */
/*              char    *buf                                   */
/*              size_t   size                                  */
/*              size_t   nitems                                */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*             long  nbytes                                    */
/*                                                             */
/***************************************************************/
long afk_file_write( afFile* afptr, char* buf, size_t size, size_t nitems )
{
  long nbytes = 0;

  if ( !afptr )
    return nbytes;
 
  if ( afptr->fptr )
  {
    nbytes = fwrite( buf, size, nitems, afptr->fptr );
  }
  else
    if ( afptr->membuf )
    {
      nbytes = size * nitems;

      while ( afptr->bufpos + nbytes > afptr->bufmax )
      {
        afptr->bufmax += afptr->blksize;
        afptr->membuf = realloc( afptr->membuf, afptr->bufmax );

        if ( !afptr->membuf )
        {
          afptr->bufmax = 0;
          return 0;
        }
      }

      memcpy( afptr->membuf + afptr->bufpos, buf, nbytes );
      afptr->bufpos += nbytes;

      if ( afptr->bufpos > afptr->bufsize )
        afptr->bufsize = afptr->bufpos;
    }

  return nbytes;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_file_get_llen                               */
/*                                                             */
/* Description:                                                */
/*       Return the maximum allowed line length.  This can     */
/*       change with time so should be checked and re-checked  */
/*       if needed.                                            */
/*                                                             */
/*  Input parameters:                                          */
/*              afFile  *afptr                                 */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*             long  length                                    */
/*                                                             */
/***************************************************************/
long afk_file_get_llen( afFile* afptr )
{
  if ( afptr )
    return afptr->llen;
  else
    return 0;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_file_truncate                               */
/*                                                             */
/* Description:                                                */
/*       Set the current psition in the file as the end of     */
/*       the file.                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              afFile  *afptr                                 */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*             long  length                                    */
/*                                                             */
/***************************************************************/
void afk_file_truncate( afFile* afptr )
{
  long pos = 0;
  long fd  = 0;
 
  if ( afptr->fptr )
  {
    /* ryan, is this efficient?  Doubt it.  Might want         */
    /* to keep track of eof location and perform truncation at */
    /* close time                                              */

    pos = ftell( afptr->fptr );
    fd = fileno( afptr->fptr );
    ftruncate( fd, pos );
  } 
  else
    if ( afptr->membuf )
    {
      afptr->membuf[afptr->bufpos] = '\0';
      afptr->bufsize = afptr->bufpos;
    }
}
