/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afkgen.c                                                    */
/*                                                               */
/*   Functions that create and work with ASCII file type         */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/
#include "afkernelP.h"


/***************************************************************/
/*                                                             */
/* Function:   afp_gen_read_keyline                            */
/*                                                             */
/* Description:                                                */
/*        Read the name, value, and comment from a keyword     */
/*        line.  Return failure if the line is not a keyword   */
/*        line.                                                */
/*                                                             */
/*  Input parameters:                                          */
/*              const afStyle  *astyle                         */
/*              const char     *line                           */
/*              char           *name                           */
/*              long            nlen                           */
/*              char           *value                          */
/*              long            vlen                           */
/*              char           *comment                        */
/*              long            clen                           */
/*              char           *unit                           */
/*              long            ulen                           */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           dmErrCode  ret                                    */
/*                                                             */
/***************************************************************/
dmErrCode afk_gen_read_keyline( const afStyle* astyle, const char* line,
                                char* name, long nlen, char* value,
                                long vlen, char* comment, long clen,
                                char *unit, long ulen )
{
  const char *iptr     = line;
  const char *key_flag = NULL;
  char        cchar    = '\0';
 

  if ( !line || !name || !value || !comment )
    return dmFAILURE;
 
  // clear values before proceeding
  name[0]    = '\0';
  value[0]   = '\0';
  comment[0] = '\0';
  unit[0]    = '\0';

  if ( !afk_style_use_fits( astyle ))
    return dmFAILURE;

  if ( '\0' != ( cchar = afk_style_get_header_char( astyle ) ) )
  {
    if ( *iptr != cchar )
      return dmFAILURE;
  }

  /* move past comment chars */
  if ( cchar )
  {
    while ( *iptr != '\0' && *iptr == cchar )
      ++iptr;
  }

  /* skip whitespace */
  while ( *iptr != '\0' && *iptr == ' ' )
    ++iptr;
 
  if ( *iptr == '\0' )
    return dmFAILURE;

  /* check for the key flag */
  if ( NULL != ( key_flag = afk_style_get_key_flag( astyle ) ) )
  {
    if ( !cs_uceq_siz( key_flag, iptr, strlen(key_flag) ))
      return dmFAILURE;
  }

  iptr = afk_keyline_read_name( iptr, name, nlen );

  if ( !iptr || *iptr == '\0' )
    return dmFAILURE;

  iptr = afk_keyline_read_value( iptr, value, vlen );
  iptr = afk_keyline_read_comment( iptr, comment, clen, unit, ulen );

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_gen_read_comment                            */
/*                                                             */
/* Description:                                                */
/*        Read a comment line in the file header.              */
/*                                                             */
/*  Input parameters:                                          */
/*              const afStyle  *astyle                         */
/*              const char     *line                           */
/*              char           *comment                        */ 
/*              long            clen                           */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           nothing                                           */
/*                                                             */
/***************************************************************/
void afk_gen_read_comment( const afStyle* astyle, const char* line,
                           char* comment, long clen)
{
  const char *iptr = line;
  char       *optr = NULL;
  char        cchar = '\0';
  long        len  = 0;
  long        size = 0;

  if ( !iptr )
    return;

  cchar = afk_style_get_header_char( astyle );

  /* Move past the comment chars */
  while ( *iptr != '\0' && *iptr == cchar )
    ++iptr;

  /* Move past whitespace */
  while( *iptr != '\0' && *iptr == ' ' )
    ++iptr;

  len  = strlen(line)+1;
  size = len < clen ? len : clen;
  memcpy( comment, iptr, size );
  optr = &comment[size-1];

  if ( optr != comment)
  {
    --optr;

    for( ; optr != comment && *optr == ' '; optr-- )
      *optr = '\0';
  }
}



/***************************************************************/
/*                                                             */
/* Function:   afk_gen_read_next_col                           */
/*                                                             */
/* Description:                                                */
/*    Read the value of the next column in the row pointed to  */
/*    by line into colval. Returns NULL when there is no more  */
/*    to read.                                                 */
/*                                                             */
/*    Code Change  Bug 11791  February 25, 2010                */
/*    Functionality to determine data type was moved to new    */
/*    function afk_gen_determine_col_datatype()                */
/*                                                             */
/*  Input parameters:                                          */
/*              const afStyle  *astyle                         */
/*              const char     *lbegin                         */
/*              char           *colval                         */
/*              long            vlen                           */
/*              dmDataType     *type                           */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           const char *lptr                                  */
/*                                                             */
/***************************************************************/
const char* afk_gen_read_next_col( const afStyle* astyle, const char* lbegin, 
				   char* colval, long vlen )
{
  const char *pValueStart  = lbegin;   /* value start pointer */
  const char *pValueEnd    = NULL;     /* value end pointer  */
  long        ValueLength  = 0;        /* length of value */
  long        width        = 0;        /* length of value before stripping */


  if ( !pValueStart )
    return NULL;
 
  /* clear data from previous pass through */
  *colval = '\0';

  pValueStart = afk_gen_find_next_val( (afStyle*)astyle, lbegin, &width );

  if (!pValueStart || *pValueStart == '\0' )
    return NULL;

  /* Transfer up to vlen chars to output string */
  ValueLength = width;

  if ( ValueLength >= vlen )
    ValueLength = vlen -1;

  strncpy( colval, pValueStart, ValueLength );

  /* Terminate the string */
  colval[ValueLength] = '\0';

  // Code Change  Bug 10844  January 9, 2008
  // Remove leading and trailing blanks in column data
  afk_strip_value( colval, (afStyle *)astyle );


  pValueEnd = pValueStart + width;

  /* move past one separator if not at end */
  if ( *pValueEnd != '\0' )
    ++pValueEnd;

  return pValueEnd;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_gen_determine_col_datatype                  */
/*                                                             */
/* Description:                                                */
/*    Determine column data type based on string value         */
/*                                                             */
/*    Function added for Bug 11791, February 25, 2010          */
/*    Much of its code comes from code formerly found in       */
/*    function afk_gen_read_next_col()                         */
/*                                                             */
/*  Input parameters:                                          */
/*              char           *colval                         */
/*              dmDataType     *type                           */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           none                                              */
/*                                                             */
/***************************************************************/
void afk_gen_determine_col_datatype(  const afStyle* astyle,
                                      afCol *acol,
                                      char *colval, dmDataType *type)
{
  const char *pForwardScan = NULL;     /* scan pointer */
  dmBool      quoted       = dmFALSE;  /* quoted value? */


  if ( !colval || *colval == '\0' )
    return;


  /* See if value is quoted */
  if ( afk_style_is_quote( astyle, *colval ))
  {
    quoted = dmTRUE;
  }

  /* Determine data type of this column entry */
  if ( type )
  {
    if ( quoted )
    {
      *type = DT_CHAR;
    }
    else if ( cs_nancode( colval ) || !(strcmp (colval, "nan")) ||
            !(strcmp(colval, "INDEF")) || !(strcmp(colval, "INF")) ||
            !(strcmp(colval, "Inf"))   || !(strcmp(colval, "+Inf")) ||
            !(strcmp(colval,"-Inf")) )
    {
      /* Special cases of string reps of numerical values */
      *type = DT_REAL8;
    }
    else if ( cs_numeric_char( *colval ) )
    {
      *type = DT_INT4;

      /* scan through value to see if it remains numerical */
      pForwardScan = colval;

      while ( (*pForwardScan != '\0') && ( *type != DT_CHAR ) )
      {
	if ( ! cs_numeric_char( *pForwardScan ) ) /* #, '+', '-', '.' */
	{
	  /* found non-numeric.. change to char */
	  *type = DT_CHAR;

	  /* maybe exponential format */
	  if ( *pForwardScan == 'e' || *pForwardScan == 'E' )
          {
            ++pForwardScan;

            if ( (*pForwardScan != '\0') && ( cs_numeric_char(*pForwardScan) ))
	    {
	      /* OK.. keep it real  */
              *type = DT_REAL8;
	    }
	  }
	}
        else if ( *pForwardScan == '.' )
          *type = DT_REAL8;

	++pForwardScan;

      }/* end while */

      /* verify INT range */
      if ( *type == DT_INT4 )
      {
        if ( abs((atol(colval))) >= LONG_MAX )
	  *type = DT_REAL8;
      }
    }
    else 
    {
      /* not a number, check for possible logical */
      *type = DT_CHAR;

      if ( *colval == 'T' || *colval == 'F' )
      {
        long len = strlen(colval);

        if ( len == 1 )
          *type = DT_LOGICAL;
        else
        {
          if ( len == 4 || len == 5 )
            if ( cs_uceq(colval, "TRUE") || cs_uceq(colval, "FALSE"))
              *type = DT_LOGICAL;
        }
      }
    }

    /* Determine if type needs to be changed by comparing it with descriptor entry */
    if ( acol->datatype != *type )
    {
      switch ( acol->datatype )
      {
        case DT_UNK:
          acol->datatype = *type;
          break;

        case DT_CHAR:
        default:
          /* once identified as char, always leave as char */
          *type = DT_CHAR;
          break;

        case DT_REAL8:
          /* if once identified as real, leave as real if numeric */
          /* if char, set column to char                          */
          /* if type is logical, set column to char               */

          if ( *type == DT_INT4 )
            *type = DT_REAL8;
          else
            if (*type == DT_LOGICAL )
              *type = DT_CHAR;

          acol->datatype = *type;
          break;

        case DT_INT4:
          /* set datatype to new type, unless logical */
          /* if logical, set column to char           */

          if ( *type == DT_LOGICAL )
            *type = DT_CHAR;
 
          acol->datatype = *type;
          break;

        case DT_LOGICAL:
          /* any datatype but logical, set column to char */
          *type = DT_CHAR;
          acol->datatype = *type;
          break;
      }
    }

    /* keep track of char string length */
    /* keep track of logicals too, indicating if full word or just 1 char */

    if ( acol->datatype == DT_CHAR || acol->datatype == DT_LOGICAL )
    {
      long len = 0;
      len = strlen(colval);
      if ( len > acol->width )
        acol->width = len;
    }
  }
}

/***************************************************************/
/*                                                             */
/* Function:   afk_gen_final_col_datatype                      */
/*                                                             */
/* Description:                                                */
/*    Rectify any datatype and descriptor conflicts here       */
/*                                                             */
/*     Function added for Bug 11791, February 25, 2010         */
/*                                                             */
/*  Input parameters:                                          */
/*              afBlock      *ablock - block                   */
/*              afStyle      *astyle - style struct            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           none                                              */
/*                                                             */
/***************************************************************/
void afk_gen_final_col_datatype( afBlock *ablock, afStyle *astyle)
{
  long ncols = 0;
  long ii    = 0;

  afCol *acol = NULL;


  // Each flavor handles data types slightly differently
  // In order to keep it straight and make it easy to modify,
  // switch on style flavor, then switch on data type under that
  // flavor case

  switch( astyle->flavor )
  {
    case AK_FLAV_RAW:
    case AK_FLAV_SIMPLE:
    default:
    {
      ncols = afk_block_ncols( ablock );

      for ( ii = 0; ii < ncols; ii++)
      {
        acol = afk_block_get_col( ablock, ii+1 );

        // RAW and SIMPLE have only two data types, CHAR and REAL

        switch ( acol->datatype )
        {
          case DT_CHAR:
            break;

          case DT_INT4:
            acol->width = 0;
            acol->datatype = DT_REAL8;
            break;
      
          case DT_REAL8:
            acol->width = 0;
            break;

          case DT_LOGICAL:
            acol->datatype = DT_CHAR;
            break;

	  case DT_UNK:
	  default:
            // Code Change  Bug 11791  March 17, 2010
            //  Define default datatype as DT_CHAR width=1
            //  in case no data exists to define it

            acol->datatype = DT_CHAR;
            acol->width = 1;
            break;
        }
      }
      break;
    }

    case AK_FLAV_CSCRDB:
    case AK_FLAV_DTF:
    case AK_FLAV_TSV:
    {
      // DTF, CSCRDB and TSV support all data types
    
      ncols = afk_block_ncols( ablock );

      for ( ii = 0; ii < ncols; ii++)
      {
        acol = afk_block_get_col( ablock, ii+1 );

        switch ( acol->datatype )
        {
          case DT_CHAR:
  	  case DT_LOGICAL:
          case DT_INT4:
          case DT_REAL8:
          default:
            break;

          case DT_UNK:
          // Code Change  Bug 11791  March 17, 2010
          //  Define default datatype as DT_CHAR width=1
          //  in case no data exists to define it 
          acol->datatype = DT_CHAR;
          acol->width = 1;
          break;
        }
      }
      break;
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:   afk_gen_find_next_val                           */
/*                                                             */
/* Description:                                                */
/*   Utility function to contain the logic for scanning        */
/*   to and through a data value from the input string         */
/*                                                             */
/*  Input parameters:                                          */
/*    afStyle  *astyle - Ascii style structure                 */
/*    char*       line - Null or '\n' terminated string        */
/*                       containing data value(s)              */
/*                                                             */
/*  Output parameters:                                         */
/*    long       vlen  - length of value in chars              */
/*                                                             */
/*  Returns:                                                   */
/*    char*      sptr  - pointer to start char of value        */
/*                                                             */
/***************************************************************/
const char* afk_gen_find_next_val( afStyle* astyle, const char* line, long* vlen )
{

  const char  *sptr   = NULL;   /* ptr to start char of value               */
  const char  *eptr   = NULL;   /* ptr to end char of value                 */

  char *seps   = NULL;          /* list of separator chars                  */
  char *quotes = NULL;          /* list of quote chars                      */
  char *white  = NULL;          /* default whitespace                       */
  char qchar  = '\0';

  dmBool multisep;              /* allows multiple separators?              */
  dmBool inquotes = dmFALSE;    /* within quoted segment?                   */
  dmBool extquotes = dmFALSE;   /* is external quoted segment?              */
  dmBool clquotes = dmFALSE;    /* have closed external quotes?             */
  dmBool escaped  = dmFALSE;    /* have an escaped char?                    */
  dmBool done     = dmFALSE;    /* done scanning line?                      */

  *vlen = 0;

  if ( !astyle ){ return sptr; }
  if ( !line ){ return sptr; }

  seps     = astyle->sep;
  white    = astyle->white;
  quotes   = astyle->quotes;
  multisep = astyle->multisep;

  /* scan to value start position */
  sptr = line;
  if ( multisep )
  {
    /* scan through separator group */
    while ((*sptr != '\0') && (*sptr != '\n') && strchr(seps, *sptr))
    { sptr++; }
  }

  /* check for end of line.. return NULL */
  if ( (*sptr == '\0') || (*sptr == '\n') )
    return NULL; 

  /* set pointer for end scan */
  eptr = sptr;
  *vlen = 0;

  /* scan to value end */
  while ( !done )
  {
    if ( (*eptr == '\0') || (*eptr == '\n') )
    {
      /* hit end of line */
      done = dmTRUE; 
    }
    // Code Change  Bug 11905  August 21, 2009
    //  Check validity of quotes pointer before use
    //  For CSCRDB, quotes is NULL
    else if ( quotes && strchr( quotes, *eptr ) && !escaped )
    {
      /* found quotes */
      if (!inquotes)
      {
	/* entering quoted segment.. set which kind of quote. */
	inquotes = dmTRUE;
	qchar    = *eptr;

	if ( eptr == sptr )
	{
	  /* These are external quotes.. */
	  /* There are special rules for external quotes w.r.t. */
	  /* ending the field.                                  */
	  extquotes = dmTRUE;
	}
      }
      else if ( *eptr == qchar )
      {
	/* matching non-escaped quote.. leaving quoted segment  */
	inquotes = dmFALSE;
	qchar    = '\0';

	if ( extquotes ){ clquotes = dmTRUE;}
      }
    }
    else if ( !inquotes && strchr( seps, *eptr ) && !escaped )
    {
      done = dmTRUE;
    }
    else if ( clquotes && !(strchr( white, *eptr )) )
    {
      /* Found non-white char following external quotes close */
      /* and before next separator.. this is a problem!       */
      return NULL;
    }

    /* check if this is an escape char  */
    /*   handles escaped backslash VVVV */
    ( *eptr == '\\' )?(escaped = !escaped):(escaped = dmFALSE);

    if ( !done )
    {
      // Code Change  Bug 11905  August 21, 2009
      //  removed * in front of eptr - increment pointer, not value in pointer
      eptr++;      /* move to next char */
    }

  } /* end while */

  /* set length of value in chars  */
  *vlen = (long)(eptr - sptr);

  return sptr;
}
