/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afkkey.c                                                    */
/*                                                               */
/*   Functions that handle keylists in in ASCII type files       */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"
#include <stdlib.h>
#include "../ftkernel/ftkernelP.h"


/***************************************************************/
/*                                                             */
/* Function:   afk_key_private                                 */
/*                                                             */
/* Description:                                                */
/*   Determines if the keyword is private                      */
/*                                                             */
/* Input parameters:                                           */
/*   const char *name                                          */
/*                                                             */
/* Output parameters:                                          */
/*                                                             */
/* Returns:                                                    */
/*   dmBool  bool                                              */
/*                                                             */
/***************************************************************/
dmBool afk_key_private( const char* name )
{
  dmBool priv = dmFALSE;

  /* use same set as FITS kernel */
  priv = ftp_key_private_fits(name);

  return priv;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_datatype_to_keytype                         */
/*                                                             */
/* Description:                                                */
/*              Returns key type given the data type           */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDataType type                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            ftKeyType   keytype                              */
/*                                                             */
/***************************************************************/
ftKeyType afk_datatype_to_keytype( dmDataType type )
{
  switch ( type )
  {
   case DT_CHAR:     return FTK_STRING;
   case DT_INT4:     return FTK_INT;
   case DT_INT2:     return FTK_INT;
   case DT_UINT1:    return FTK_INT;
   case DT_BIT:      return FTK_INT;
   case DT_UINT2:    return FTK_INT;
   case DT_UINT4:    return FTK_UINT;
   case DT_REAL4:    return FTK_REAL;
   case DT_REAL8:    return FTK_REAL;
   case DT_LOGICAL:  return FTK_LOGICAL;
   case DT_UNK:      return FTK_UNKNOWN;
   default:          return FTK_UNKNOWN;
  }
}



/***************************************************************/
/*                                                             */
/* Function:   afp_key_get_type                                */
/*                                                             */
/* Description:                                                */
/*              Returns data type given key                    */
/*                                                             */
/*  Input parameters:                                          */
/*                    const char  *keyval                      */
/*                    long         vlen                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDataType  type                                 */
/*                                                             */
/***************************************************************/
dmDataType afk_key_get_type( const char* keyval, long vlen )
{
  dmDataType  type = DT_UNK;
  double      dval = 0.0;
  const char *iptr = keyval;
  const char *niptr = NULL;
  char        c = '\0';     /* a character value */


  if ( !keyval || vlen <= 0 )
    return type;

  if ( *iptr == '\'' || *iptr == '"' )
    return DT_CHAR;

  niptr = iptr+1;
 
  if ( cs_digit_char( *iptr ))
    type = DT_INT4;
  else
    if ( cs_numeric_char( *iptr ) &&
	  *iptr != '\0' && cs_digit_char( *niptr ))
      type = DT_INT4;
    else
    {
      if ( *iptr == 'T' || *iptr == 'F' )
      {
        ++iptr;

        for ( ; *iptr != '\0'; iptr++ )
        {
          if ( *iptr != ' ' && *iptr != '\t' )
            return DT_CHAR;
        }

        return DT_LOGICAL;
      }

      return DT_CHAR;
    }

  for( ; *iptr != '\0'; iptr++ )
  {
    if ( type == DT_INT4 || type == DT_REAL8 )
    {
       if ( *iptr == '.' )
         type = DT_REAL8;
       else
         if ( *iptr == 'e' || *iptr == 'E' )
         {
           c = *(iptr+1); /* make sure this isn't just a dangling 'e' */

           if ( c == '+' || c == '-' || isdigit(c) )
             type = DT_REAL8;
           else
             type = DT_CHAR;  
         }
         else
           if ( !cs_numeric_char( *iptr ) && *iptr != ' ' )
           {
             type = DT_CHAR;
             return type;
           }
    }
  }

  if ( type == DT_INT4 )
  {
    dval = atof( keyval );

    if ( dval > DT_INT4_MAX )
    {
      if ( dval > DT_UINT4_MAX )
        type = DT_REAL8;
      else
        type = DT_UINT4;
    }
  }

  return type;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_key_parse_val                               */
/*                                                             */
/* Description:                                                */
/*              Parse key value                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    const char *valstr                       */
/*                    dmDataType  type                         */
/*                    void       *val                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_key_parse_val( const char* valstr, dmDataType type, void* val )
{
  char sValueString[AK_KEYVAL_LEN];

  switch ( type )
  {
    case DT_REAL8:
      *(double*)val = atof(valstr);
      break;

    case DT_INT4:
      *(long*)val = atol(valstr);
      break;

    case DT_UINT4:
      *(unsigned long*)val = (unsigned long)atof(valstr);
      break;

    case DT_CHAR:
      // Code Change Bug 10313 April 24, 2007
      // Remove quotes from value here

      if ( *valstr == '\'' || *valstr == '\"' )
      {
        strcpy(sValueString, (valstr+1));
        sValueString[(strlen(sValueString)-1)] = '\0';
        *(char**)val = cs_dup(sValueString);
      }
      else
      {
        *(char**)val = cs_dup(valstr);
      }
      break;

   case DT_LOGICAL:
     if ( *valstr == 'T' )
       *(dmBool*)val = UT_TRUE;
     else
       *(dmBool*)val = UT_FALSE;
     break;

   default:
     *(double*)val = atof(valstr);
     break;
  }
}


/***************************************************************/
/*                                                             */
/* Function:   afp_key_delete                                  */
/*                                                             */
/* Description:                                                */
/*              Delete key                                     */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBlock  *ablock                         */
/*                    dmuKey   *ukey                           */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_key_delete( afBlock* ablock, dmuKey* ukey )
{
  dmuKeyList keylist;

  keylist = ablock->keylist;
  afk_buffer_delete_keyline( ablock->abuf, ukey->index );
  dmu_key_delete( keylist, ukey );
}



/***************************************************************/
/*                                                             */
/* Function:   afp_key_default_fmt                             */
/*                                                             */
/* Description:                                                */
/*              Set default key format                         */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDataType  type                         */
/*                    const void *value                        */
/*                    char       *val                          */
/*                    afStyle    *astyle                       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_key_default_fmt( dmDataType type, const void* value,
                          char* val, afStyle* astyle )
{
  const char *quotes = afk_style_get_quotes( astyle );

  if ( type == DT_REAL8 )
  {
    strcpy( val, "%.13E" );
  }
  else
    if ( type == DT_CHAR && quotes )
    {
      val[0] = quotes[0];
      val[1] = '%';
      val[2] = 's';
      val[3] = quotes[0];
      val[4] = '\0';
    }
    else
      afk_default_fmt( type, value, val, astyle );
}
