/*                                                                
**  Copyright (C) 2003-2013  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afklib.c                                                    */
/*                                                               */
/*   Functions that work with ASCII file type                    */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"


/***************************************************************/
/*                                                             */
/* Function:   afk_count_words                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    const char  *buf                         */
/*                    const char  *sep                         */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            long   NumberOfWords                             */
/*                                                             */
/***************************************************************/
long afk_count_words( const char* buf, const char* sep )
{
  char  *copy = NULL;
  char  *ptr  = NULL;
  long   len  = 0;
  long   n    = 0;

  len = strlen( buf )+1;

  if ( len <= 1 )
    return 0;

  copy = malloc( len*sizeof(char) );
  strcpy( copy, buf );

  ptr = strtok( copy, sep );

  while( ptr )
  {
    ++n;
    ptr = strtok( NULL, sep );
  }

  free( copy );

  return n;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_keyline_read_name                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    const char  *ifirst                      */
/*                    char        *ofirst                      */
/*                    long         len                         */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            const char  *pointerToName                       */
/*                                                             */
/***************************************************************/
const char* afk_keyline_read_name( const char* ifirst, char* ofirst,
                                   long len )
{
  const char  *iptr  = ifirst;
  char        *olast = ofirst + len;
  char        *optr  = ofirst;
  dmBool       white = dmFALSE;

  /* move past whitespace */
  if ( !iptr )
    return iptr;

  while( *iptr != '\0' && *iptr == ' ' )
    ++iptr;

  for( ; *iptr != '\0' && ofirst != olast && *iptr != '='; iptr++, optr++ )
  {
    if ( !white && *iptr == ' ' )
      white = dmTRUE;
    else
      if ( white && *iptr != ' ' )
        return NULL;
    *optr = *iptr;
  }

  /* make sure name is null terminated */
  if ( optr == olast && optr != ofirst )
    --optr;

  *optr = '\0';

  if ( optr != ofirst )
    --optr;

  /* remove trailing whitespace */
  for ( ; optr != ofirst && *optr == ' '; optr--)
  {
   *optr = '\0';
  }
 
  /* move past '=' */
  if ( iptr != '\0' )
    ++iptr;

  return iptr;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_keyline_read_value                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    const char  *ifirst                      */
/*                    char        *ofirst                      */
/*                    long         len                         */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            const char  *pointerToValue                      */
/*                                                             */
/***************************************************************/
const char* afk_keyline_read_value( const char* ifirst, 
				    char* ofirst, long len )
{
  const char  *iptr = ifirst;
  char        *olast = ofirst + len;
  char        *optr  = ofirst;
  char         end_char = '\0';

  if ( !iptr )
    return NULL;

  /* move past whitespace */
  while( *iptr != '\0' && *iptr == ' ' )
    ++iptr;

 
  /*  Code Change  Bug 10313  April 25, 2007        */
  /*   Check for string valued keyword              */
  /*   Set end_char to char that marks end of value */

  if ( *iptr != '\0' )
  {
    if ( *iptr == '\'' ) 
    {
      // yes, found, output them to value
      // and set end character

      *optr++ = *iptr++;
      end_char = '\'';
    }
    else
      if ( *iptr == '"') 
      {
      // yes, found, output them to value
      // and set end character

        *optr++ = *iptr++;
        end_char = '"';
      }
      else
        end_char = '/'; /* stop when we hit the description */
  }

  /* read value and transfer into output variable */

  for ( ; *iptr != '\0' && optr != olast && *iptr != end_char;
       iptr++,optr++ )
  {
    *optr = *iptr;
  }

  /* make sure the value retains the quotes if it is a string valued keyword */

  if ( end_char != '/' && *iptr == end_char)
    *optr++ = *iptr++;

  /* End Code Change */



  /* make sure value is null terminated */
  if ( optr == olast && optr != ofirst ) 
    --optr;

  *optr = '\0';
 
  /* remove trailing whitespace */
  for ( ; optr != ofirst && *optr == ' '; optr--)
  {
    *optr = '\0';
  }

  /* move iptr up to comment or end of line */
  while ( *iptr != '\0' && *iptr != '/' )
  {
    ++iptr;
  }

  /* move past '/' */
  if ( *iptr != '\0' )
    ++iptr;

  return iptr;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_keyline_read_comment                        */
/*                                                             */
/* Description:                                                */
/*         Read the keyword commont on a keyword line.         */
/*                                                             */
/*  Input parameters:                                          */
/*                    const char  *ifirst                      */
/*                    char        *ofirst                      */
/*                    long         len                         */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            const char  *pointerToComment                    */
/*                                                             */
/***************************************************************/
const char* afk_keyline_read_comment( const char* ifirst,
				      char* ofirst, long len,
                                      char *unit, long ulen )
{
  char        *olast  = ofirst + len;
  const char  *iptr   = ifirst;
  char        *optr   = NULL;
  long         nbytes = 0;

  if ( !iptr )
    return NULL;

  /* move past whitespace */
  while( *iptr != '\0' && *iptr == ' ' )
    ++iptr;

  // Code Change  Bug 10896  January 31, 2008
  //  Code segment taken from fits kernel code
  //  Strip off unit information if present

 if ( *iptr == '[' )
 {
   optr = (char *)++iptr;
   while( *iptr != ']' && *iptr != '\0' ) 
     ++iptr;

  nbytes = iptr - optr;

  if ( nbytes > ulen )
   nbytes = ulen;

  cs_copy_siz( optr, unit, nbytes );

  if ( *iptr == ']' )
    ++iptr;
  while( isspace( *iptr ) )
    ++iptr;
 }


 // read comment string

  optr = ofirst;

  for ( ; *iptr != '\0' && optr != olast; iptr++,optr++ )
  {
    *optr = *iptr;
   }

  if ( optr == olast )
    --optr;

  *optr = '\0';

  /* do not back up if still at beginning */
  if ( optr != ofirst )
    --optr;

  /* remove trailing whitespace */
  for ( ; optr != ofirst && *optr == ' '; optr--)
  {
    *optr = '\0';
  }

  return iptr;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_default_fmt                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDataType    dtype       - datatype of value            */
/*    const void   *value       - data value                   */
/*    char         *ofmt        - resulting fformat string     */
/*    afStyle      *astyle      - style structure.             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_default_fmt( dmDataType dtype, const void* value, 
		      char* ofmt, afStyle* astyle )
{
  double   tmp;
  double   frac;
  long     whole = 0;

 
  switch ( dtype )
  {
    case DT_INT2:
      strcpy( ofmt, "%d" );
      break;
    case DT_INT4:
      strcpy( ofmt, "%ld" );
      break;
    case DT_UINT2:
      strcpy( ofmt, "%u" );
      break;
    case DT_UINT4:
      strcpy( ofmt, "%lu" );
      break;
    case DT_UINT1: 
      strcpy( ofmt, "%u" );
      break;
    case DT_BIT:
      strcpy( ofmt, "%c" );
      break;
    case DT_REAL4:
    {
      if ( value )
      {
        tmp = fabs((double)*((float*)value));
        whole = floor(tmp);
        frac = tmp - whole;

        if ( tmp >= 100000.0 )
          strcpy( ofmt, "%.8e" );
        else if ( frac == 0.0 )
	  strcpy( ofmt, "%.1f" );
	else if ( tmp < .00001 )
	  strcpy( ofmt, "%.8e" );
	else if ( tmp < 0.0001 )
	  strcpy( ofmt, "%.11f" );
	else if ( tmp < 0.001 )
	  strcpy( ofmt, "%.10f" );
	else if ( tmp < 0.01 )
	  strcpy( ofmt, "%.9f" );
	else if ( tmp < 0.1 )
	  strcpy( ofmt, "%.8f" );
	else if ( tmp < 1.0 )
	  strcpy( ofmt, "%.7f" );
	else if ( tmp < 10.0 )
	  strcpy( ofmt, "%.6f" );
	else if ( tmp < 100.0 )
	  strcpy( ofmt, "%.5f" );
	else if ( tmp < 1000.0 )
	  strcpy( ofmt, "%.4f" );
	else if ( tmp < 10000.0 )
	  strcpy( ofmt, "%.3f" );
	else
	  strcpy( ofmt, "%.8e" );
      }
      else
        strcpy( ofmt, "%.8e" );
      break;
    }

    case DT_REAL8:
    {
      if ( value )
      {
        tmp = fabs(*((double*)value));
        whole = floor(tmp);
        frac = tmp - whole;

        if ( tmp >= 100000.0 )
          strcpy( ofmt, "%.13e" );
        else if ( frac == 0.0 )
	  strcpy( ofmt, "%.1f" );
	else if ( tmp < 1.0 )
	  strcpy( ofmt, "%.13e" );
	else if ( tmp < 10 )
	  strcpy( ofmt, "%.12f" );
	else if ( tmp < 100 )
	  strcpy( ofmt, "%.11f" );
	else if ( tmp < 1000 )
	  strcpy( ofmt, "%.10f" );
	else if ( tmp < 10000 )
	  strcpy( ofmt, "%.9f" );
	else if ( tmp < 100000 )
	  strcpy( ofmt, "%.8f" );
	else
	  strcpy( ofmt, "%.13e" );
      }
      else
        strcpy( ofmt, "%.13e" );
      break;
    }

    case DT_CHAR:
    {
      strcpy( ofmt, "%s" );
      break;
    }

    case DT_LOGICAL:
      strcpy( ofmt, "%s" );  /* T or F */
      break;

    default:
      ofmt[0] = '\0';
      break;
  }/* end switch (dtype) */
}



/***************************************************************/
/*                                                             */
/* Function:   afk_tdisp_to_cmft                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDataType   dataType                    */
/*                    const char  *tdisp                       */
/*                    char        *fmt                         */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_tdisp_to_cfmt( dmDataType dataType, const char* tdisp, char* fmt )
{
  char  *ptr       = NULL;
  char   typecode  = '\0';
  long   width     = 0;
  long   precision = 0;
  long   exponent  = 0;
  dmBool bUseQuotes = dmFALSE;
  char   cQuoteChar = '\0';


  afp_parse_tdisp( tdisp, &typecode, &width, &precision, &exponent );


  ptr  = fmt;

  // Code Change Bug 10408 April 24, 2007
  //   Some character strings need to be quoted
  //   If input parametet fmt is not null and type
  //   is character, check first position for quote char
  //   If quote char, save it and put it at beginning 
  //   and end of format string being formed

  if (typecode == 'A')
  {
    if ( *ptr != '\0' && *ptr != '%')
    {
      // Assuming that char is quote char

      bUseQuotes = dmTRUE;
      cQuoteChar = *ptr;
      ++ptr;
    }
  }
  // end code change 

  *ptr = '%';
  ++ptr;

  if (typecode == 'A')
  {
    /* Left justify strings */
    *ptr = '-';
    ++ptr;
  }

  if ( width > 0 )
  {
    ptr += sprintf( ptr, "%ld", width );
  }


  switch( typecode )
  {
    case 'A':
      *ptr = 's';
      ++ptr;

      // add quote char for strings if needed

      if ( bUseQuotes )
      {
        *ptr = cQuoteChar;
        ++ptr;
      }

      *ptr = '\0';
      break;


    case 'I':
      switch( dataType )
      {
        case DT_INT2:
	{
          *ptr = 'd';
          ++ptr;
          break;
        }
        case DT_INT4:
        default:
        {
          *ptr = 'l';
          ++ptr;
          *ptr = 'd';
          ++ptr;
          break;
        }
        case DT_UINT2:
        case DT_UINT1:
	{
          *ptr = 'u';
          ++ptr;
          break;
        }
        case DT_UINT4:
        {
          *ptr = 'l';
          ++ptr;
          *ptr = 'u';
          ++ptr;
          break;
        }
      }
      break;

    case 'B':
      *ptr = 'c';
      ++ptr;
      break;

    case 'F':
      /* set precision */
      ptr += sprintf( ptr, ".%ld", precision );

      *ptr = 'f';
      ++ptr;
      break;


    // Code Change  Bug 13511  March 6, 2013
    //   Added 'G' - tread as DT_REAL8 for TSV

    case 'E':
    case 'G':
      /* set precision */
      ptr += sprintf( ptr, ".%ld", precision );

      *ptr = 'e';
      ++ptr;
      break;

   case 'D':
      /* set precision */
     ptr += sprintf( ptr, ".%ld", precision );

     *ptr = 'e';
     ++ptr;
     break;
  }

  *ptr = '\0';
}



/***************************************************************/
/*                                                             */
/* Function:   afk_cfmt_to_tdisp                               */
/*                                                             */
/* Description:                                                */
/*            Convert a C format string to a Fortran style     */
/*            display string                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                    const char  *cfmt                        */
/*                    long         col_width                   */
/*                    char        *tdisp                       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_cfmt_to_tdisp( const char* cfmt, long col_width, char* tdisp )
{
  const char  *iptr = NULL;
  char        *optr = NULL;
  long         width = 0;
  long         precision = 0;
  char         typecode = '\0';

  iptr = cfmt;
  optr = tdisp;

  if ( *iptr != '%' )
  {
    *optr = '\0';
    return;
  }

  ++iptr;

  if ( *iptr == '-' || *iptr == '+' )
    ++iptr;

  if ( isdigit( *iptr ))
  {
    width = atoi( iptr );
    ++iptr;

    while( isdigit( *iptr ))
      ++iptr;
  }
 

  if ( *iptr == '.' )
  {
    ++iptr;

    if ( isdigit( *iptr ))
    {
      precision = atoi( iptr );
      ++iptr;

      while( isdigit( *iptr ))
        ++iptr;
    }
  }

  switch( *iptr )
  {
    case 'c':
      typecode = 'A';

      if ( !width )
        width = 1;
      precision = 0;
      break;

    case 'l':
      ++iptr;
      if ( *iptr != 'd' && *iptr != 'u' )
        break;

    case 'd':
    case 'u':
      typecode = 'I';
      precision = 0;
      if ( !width )
        width = col_width;
      break;

    case 'f':
      typecode = 'F';
      if ( !width )
        width = col_width;
      if ( !precision )
        precision = 1;
      break;

    case 'e':
    case 'g': /* not sure what to do with g.... */
      typecode = 'E';
      if ( !width )
        width = col_width;
      if ( !precision )
      {
        precision = width - 6;
        if ( precision < 0 )
          precision = width - 1;
      }
      break;

    case 's':
      default:
      typecode = 'A';
      if ( !width )
        width = col_width;
      precision = 0;
     break;
  } // end switch

  if ( width && precision )
    sprintf( tdisp, "%c%ld.%ld", typecode, width, precision );
  else
    if ( width )
      sprintf( tdisp, "%c%ld", typecode, width );
    else
      if ( precision )
        sprintf( tdisp, "%c.%ld", typecode, precision );
      else
        sprintf( tdisp, "%c", typecode ); 
}



/***************************************************************/
/*                                                             */
/* Function:   afk_tdisp_code_to_datatype                      */
/*                                                             */
/* Description:                                                */
/*            Convert a C format string to a Fortran style     */
/*            display string                                   */
/*                                                             */
/*            Code Change  Bug 13511  March 6, 2013            */
/*            Added 'G' as DT_REAL8 for TSV                    */
/*                                                             */
/*  Input parameters:                                          */
/*                    const char  tchar                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDataType  DataType                             */
/*                                                             */
/***************************************************************/
dmDataType afk_tdisp_code_to_datatype( const char tchar )
{
  switch ( tchar )
  {
    case 'A':
      return DT_CHAR;
    case 'F':
      return DT_REAL8;
    case 'E':
      return DT_REAL8;
    case 'G':
      return DT_REAL8;
    case 'I':
      return DT_INT4;
    case 'D':
      return DT_REAL8;
    default:
     return DT_UNK;
  }
}


/***************************************************************/
/*                                                             */
/* Function:   afk_strip_value                                 */
/*                                                             */
/* Description:                                                */
/*   Cleans up the input string:                               */
/*     + Remove any trailing whitespace characters             */
/*     + Remove any leading whitespace characters              */
/*     + Remove edge-quotes                                    */
/*     + unescape internal quotes (other?)                     */
/*                                                             */
/*   Reference Bug number 9343.                                */
/*   Modified and moved here for Bug 10844                     */
/*   Modified more for Bug 11086                               */
/*                                                             */
/* Input parameters:                                           */
/*   char      *pToken       - value string                    */
/*   afStyle   *pakStyle     - Style structure                 */
/*                                                             */
/* Output parameters:                                          */
/*   char      *pToken       - (modified in place)             */
/*                                                             */
/* Returns:                                                    */
/*   nothing                                                   */
/*                                                             */
/***************************************************************/
void afk_strip_value( char *pToken, afStyle *pakStyle )
{
  const char *quotes     = NULL;     /* quote char set         */
  const char *seps       = NULL;     /* separator char set     */
  char       *white      = NULL;     /* default whitespace     */
  char       *pQuoteChar = NULL;     /* quote char             */
  char       *pStart     = NULL;     /* start of good string   */
  char       *pEnd       = NULL;     /* end of good string     */
  long        length     = 0;        /* length of value string */


  if (!pakStyle)
    return;

  length = strlen(pToken);

  if ( (length == 0) || (*pToken == '\0') )
    return;

  /* Strip Edge whitespace from input string. */

  /* Remove trailing whitespace characters.                 */
  /* Scan backward through string.. match char to white     */
  /* replace with null.                                     */

  white = pakStyle->white;

  for ( pEnd = (pToken+(length-1)); pEnd >= pToken; pEnd--)
  {
    if ( strchr( white, *pEnd ) != NULL )
      *pEnd = '\0';
    else
      break;
  }

  /* get new length of token string and check for null string */
  length = strlen(pToken);

  if ( length == 0 )
    return;

  /* Remove leading separator/whitespace chars   */
  /* Scan forward through string.. match char in sep list. */
  /* Shift remaining string forward.                       */

  for ( pStart = pToken; pStart < pToken+length; pStart++)
  {
    if ( strchr( white, *pStart) == NULL )
      break;
  }
  length = strlen(pStart);

  seps   = afk_style_get_sep(pakStyle);
  quotes = afk_style_get_quotes(pakStyle);

  /* Strip Edge quotes if present */
  if ( quotes )
  {
    if ( (pQuoteChar = strchr( quotes, *pStart )) && (*pEnd == *pQuoteChar) )
    {
      pStart++;         /* move past quote start       */
      *pEnd = '\0';     /* replace end quote with null */
      pEnd--;           /* move up end pointer         */
     }

    length = strlen(pStart);
  }

  /* Scan backward through string again..                               */
  /*   check for escaped quotes and separators                          */
  char* cptr; 
 
 for ( cptr = (pStart+(length-1)); cptr >= pStart; cptr--)
  {
    if ( *cptr == '\\' )
    {
      /* Have an escape char.. see if we want to remove it. */
      if ( (quotes && ( strchr( quotes, *(cptr+1) ) != NULL )) ||
	   (seps   && ( strchr( seps, *(cptr+1) ) != NULL ))  )
      {
	// Code Change  Bug 13332 July 20, 2012
        // 64-bit Linux library function behavior issue
        // Call to strcpy failing when copying involves shifting
        // bytes within a string (destination and source overlap)
        // Transfer using memmove

        length = strlen(cptr+1);
        memmove(cptr, (cptr+1), (length+1));
      }
    }
  }

  /* If temp pointer is same as pToken, then there are no leading         */
  /* separators, and the string pToken is complete. If the temp           */
  /* pointer is not the same as pToken, then need to strip off leading    */
  /* separators. Do this by copying the string, starting at the first     */
  /* non-separator character, into the same location. Remember, there are */
  /* nulls following, so no need to worry about length here.              */

  if (pStart > pToken)
  {
    // Code Change  Bug 13332 July 20, 2012
    // 64-bit Linux library function behavior issue
    // Call to strcpy failing when copying involves shifting
    // bytes within a string (destination and source overlap)
    // Transfer stripped value string using memmove

    length = strlen(pStart);
    memmove(pToken, pStart, (length+1));
  }
}


/***************************************************************/
/*                                                             */
/* Function:   afk_prepare_string                              */
/*                                                             */
/* Description:                                                */
/*    Handles stuff like escaping internal quotes              */
/*                                                             */
/*  Input parameters:                                          */
/*    char* istr               - string value                  */
/*    char* cfmt               - output format for string      */
/*    afStyle* astyle          - ascii style structure         */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*    char* ostr - modified output string                      */
/*                                                             */
/***************************************************************/
char* afk_prepare_string( char* istr, char* cfmt, afStyle* astyle )
{
  char* iptr = NULL;
  char* cptr = NULL;
  char* tptr = NULL;

  char* tstr = NULL;
  char* ostr = NULL;

  const char*  seps   = NULL;
  const char*  quotes = NULL;
  const char*  white  = NULL;

  char  qchar;
  char  default_fmt[5] = "%s";
  char  tmp[AK_COLVAL_LEN];

  dmBool add_quotes = UT_FALSE;
  dmBool ext_quotes = UT_FALSE;

  long  ilen;

  if (!istr)
    return ostr;

  if (!cfmt)
    cfmt = default_fmt;

  /* Get list of quote chars */
  quotes = afk_style_get_quotes( astyle );

  /* Get list of separator chars */
  seps = afk_style_get_sep( astyle );

  /* Get list of whitespace chars */
  white = astyle->white;

  /* Get length of input string */
  ilen = strlen(istr);

  if ( quotes && !astyle->fixed )
  {
    /* Check if we need to ADD quotes */

    /* On output, strings are quoted only if they contain a  */
    /* separator char or leading/trailing  whitespace.       */
    if ( cs_is_blank( istr ) )
    {
      /* Empty string gets quoted if using white separator */
      if ( (strchr(seps, white[0]) || strchr(seps, white[1])) )
      {
	add_quotes = UT_TRUE;
      }
    }
    else if ( strchr( white, *istr ) )
    {
      /* value has significant leading whitespace */
      add_quotes = UT_TRUE;
    }
    else if (UT_FALSE)
    {
      /* if ( strchr( white, istr[ilen-1] ) */
      /* value has significant trailing whitespace */
      /* TEMPORARILY DISABLE THIS ONE.. any value with TDISP */
      /* defined can come through with trailing whitespace.. */
      add_quotes = UT_FALSE;
    }
    else
    {
      /* Check for separators within the value */
      if ( NULL != ( cptr = strpbrk( (char*)istr, seps ) ) )
      {
	++cptr;
	while( *cptr != '\0' )
	{
	  if ( !afk_style_is_sep( astyle, *cptr ))
	  {
	    /* non-separator.. therefore multi-word */
	    add_quotes = UT_TRUE;
	    break;
	  }
	  ++cptr;
	}
      }
    }

    /* Check if we need to escape significant external quotes */
    /* To provide a string value which includes the quotes,   */
    /*  the quotes must be escaped.                           */
    if ( strchr( quotes, *istr ) )
    {
      /* value has significant external quotes..              */
      /* However, we do not need to escape it if wrapped by   */
      /* a different quote character.                         */
      /*    "'quoted value'"  '\'quoted value\''              */ 

      qchar = *istr;
      if ( add_quotes && ( *istr != quotes[0] ) )
      {
	ext_quotes = UT_FALSE;
      }
      else
      {
	ext_quotes = UT_TRUE;
      }
    }
     
    /* Generate string value with any needed quotes and escapes */
    iptr = istr;
    if ( add_quotes || ext_quotes )
    {
      /* Allocate holding space.. worst case        */
      /* escape every char + add quotes + terminate */
      long tlen = 2 * strlen( istr ) + 2 + 1;
      tstr = calloc( tlen, sizeof(char) );

      tptr = tstr;
      if ( add_quotes ){ *tptr = quotes[0]; tptr++; }

      while ( *iptr != '\0' )
      {
	if ( ext_quotes && (*iptr == qchar) )
	{
	  /* if first char, or non-escaped (closing) instance */
	  if ( (iptr == istr ) || (*(iptr-1) != '\\' ) )
	  {
	    *tptr = '\\'; tptr++;
	  }
	}
	else if ( add_quotes && (*iptr == quotes[0]) )
	{
	  if ( (iptr != istr ) && (*(iptr-1) != '\\' ) )
	  {
	    /* escape any internal quotes matching the one being added. */
	    *tptr = '\\'; tptr++;
	  }
	}

	*tptr = *iptr;
	tptr++;
	iptr++;
      }

      if ( add_quotes ){ *tptr = quotes[0]; tptr++; }

      *tptr = '\0'; /* terminate */

      /* re-set pointer to temp space */
      iptr = tstr;

    } /* end needs be modified */

  } /* end if quotes && !fixed */


  if ( iptr)
  {
    ilen = strlen( iptr ); /* may have changed */
    istr = iptr;
  }

  if ( ilen < AK_COLVAL_LEN )
  {
    /* Transfer to output format */
    ilen = sprintf(tmp, cfmt, istr );

    /* Make copy for output */
    ostr = cs_dup(tmp);
  }

  if ( tstr ){ free(tstr); }

  return ostr;
}
