/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afkraw.c                                                    */
/*                                                               */
/*   Functions that create and work with ASCII file type         */
/*                                                               */
/*  Copyright 1999-2007  Smithsonian Astrophysical Observatory   */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"

static dmErrCode afk_raw_compose_keyline( const char* name, const void* value, dmDataType type, const char* unit, const char* desc, char* keyline, long klen, afStyle* astyle );




/***************************************************************/
/*                                                             */
/* Function:   afk_raw_read_header                             */
/*                                                             */
/* Description:                                                */
/*             We expect the file already to be pointing at    */
/*             the beginning of the block.                     */
/*                                                             */
/*  Input parameters:                                          */
/*                   afBlock  *ablock - block pointer          */
/*                   afStyle  *astyle - style pointer          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afk_raw_read_header( afBlock* ablock, afStyle* astyle )
{
  char keyname[AK_KEYNAME_LEN];
  char comment[AK_KEYCOMMENT_LEN];

  dmuKeyList keylist;
  dmDataType type  = DT_UNK;
  dmBool     mask  = dmTRUE;
  dmuKey    *key   = NULL;
  ftKeyType  ktype = FTK_UNKNOWN;
  dmErrCode  status = dmSUCCESS;

  afBuffer  *abuf  = NULL;
  afFile    *afptr = NULL;
  afDataset *ads   = NULL;

  GenericType tmp;
  void  *parsed_value = &tmp;

  const char *line     = NULL;
  const char *line_ptr = NULL;
  char  cchar = '\0';

  long pos     = 0;
  long histlen = 7;
  long commlen = 7;
 

  ads = afk_block_get_kds( ablock );
  abuf = afk_block_get_buffer( ablock );
  afptr = afk_dataset_fptr( ads );
  keylist = dmu_keylist_alloc();

  /* ryan, might need to delete a previous keylist? */
  ablock->keylist = keylist;

  cchar = afk_style_get_header_char( astyle );

  /* read the next comment line */
  pos = afk_file_tell( afptr );
  line = afk_file_getline_ptr( afptr );

  while ( line && *line == '\0' )
    line = afk_file_getline_ptr( afptr );

  if ( !line || (cchar == '\0') || ( cchar && *line != cchar ))
  {
    afk_style_set_header( astyle, dmFALSE );
    afk_buffer_set_datastart( abuf, pos );
    return dmSUCCESS;
  }
  else
    afk_style_set_header( astyle, dmTRUE );

  while ( line && cchar && *line == cchar )
  {
    mask = dmFALSE;

    afk_buffer_add_keyline( abuf, line );

    afk_gen_read_comment( astyle, line, comment, AK_KEYCOMMENT_LEN );

    afk_gen_read_comment( astyle, line,
			  comment, AK_KEYCOMMENT_LEN );

    type = DT_CHAR;
    ktype = FTK_COMMENT;

    if ( cs_eq_siz( comment, dmHISTORY, histlen ))
      strcpy( keyname, dmHISTORY );
    else
      if ( cs_eq_siz( comment, dmCOMMENT, commlen ))
        strcpy( keyname, dmCOMMENT );
      else
        strcpy ( keyname, dmCOMMENT );
 
    *(char**)parsed_value = cs_dup( comment );
    comment[0] = '\0';
    mask = UT_TRUE;

    key = dmu_key_alloc();
 
    /* ryan, do I need to get the keyword units? */
    dmu_key_set( key, (dmuKeyType)ktype, type, keyname, parsed_value, NULL,
	               comment, mask );
    dmu_key_insert( keylist, key, 0 );

    if ( type == DT_CHAR )
      free( *(char**)parsed_value );

    /* read the next comment line */
    pos = afk_file_tell( afptr );
    line = afk_file_getline_ptr( afptr );

    while( line && *line == '\0' )
    {
      pos = afk_file_tell( afptr );
      line = afk_file_getline_ptr( afptr );
    }
  }

  /* We've now read the header.  Now we need to find the columns. */

  if ( !line )
    return status; /* no data in file */

  /* get rid of spaces at front of line */
  line_ptr = line;

  while ( *line_ptr != '\0' && afk_style_is_sep( astyle, *line_ptr ))
    ++line_ptr;

  /* Move past blank lines */
  while ( *line_ptr == '\0' )
  {
    pos = afk_file_tell( afptr );
    line_ptr = afk_file_getline_ptr( afptr );  

    if ( ! line_ptr )
      return status; /* no data in file */

    /* get rid of spaces at front of line */
    while ( *line_ptr != '\0' && afk_style_is_sep( astyle, *line_ptr ))
     ++line_ptr;
  }

  afk_buffer_set_datastart( abuf, pos );

 return status;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_raw_init_cols                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                   afBlock  *ablock - block pointer          */
/*             const afStyle  *astyle - style pointer          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afk_raw_init_cols( afBlock* ablock, const afStyle *astyle )
{
  afDataset  *ads   = afk_block_get_kds( ablock );
  afFile     *afptr = afk_dataset_fptr( ads );
  afCol      *acol  = NULL;

  dmErrCode   status = dmSUCCESS;
  dmDataType  type   = DT_UNK;

  const char  *line      = NULL;
  const char  *rptr      = NULL;
  char         colval[AK_COLVAL_LEN];

  long  ncols = 0;
  long  ii    = 0;


  afk_block_set_row( ablock, 1 );
  line = afk_file_getline_ptr( afptr );

  if ( line )
    afk_raw_read_cols( ablock, astyle, line );

  // Code Change  Bug 11791  February 25, 2010
  // now need to get the column types from the data

  ads   = afk_block_get_kds( ablock );
  afptr = afk_dataset_fptr( ads );
  ncols = afk_block_ncols( ablock );

  afk_block_set_row( ablock, 1 );

  /* loop over all rows to get column data type */

  while ( (rptr = afk_file_getline_ptr( afptr )) )
  {
    for ( ii = 0; ii < ncols && rptr; ii++ )
    {
      acol = afk_block_get_col( ablock, ii+1 );
      rptr = afk_gen_read_next_col( astyle, rptr, colval, AK_COLVAL_LEN );

      afk_gen_determine_col_datatype( astyle, acol, colval, &type);
    }
  }

  /* final rectification of column types */
  afk_gen_final_col_datatype( ablock, (afStyle*) astyle );

  /* reset to start */
  afk_block_set_row( ablock, 1 );

  return status;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_raw_read_cols                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                      afBlock  *ablock - block pointer       */
/*               const  afStyle  *astyle - style pointer       */
/*               const  char     *line                         */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afk_raw_read_cols( afBlock* ablock,
                             const afStyle* astyle, const char* line )
{
  afCol *acol = NULL;
  long   ncols  = 0;

  const char  *iptr      = line;
  const char  *last_iptr = iptr;

  char colval[AK_COLVAL_LEN];
  char colnam[CARD_SIZE];

  dmDataType type  = DT_UNK;
  dmErrCode status = dmSUCCESS;


  while ( status && ( iptr = afk_gen_read_next_col( astyle, iptr, colval,
                                                    AK_COLVAL_LEN )))
  {
    ++ncols;
    last_iptr = iptr;
    sprintf( colnam, "col%ld", ncols );
    acol = afk_col_alloc();

    if ( !acol )
    {
      status = dmFAILURE;
      break;
    }

    acol->name = cs_dup( colnam );
    acol->no = ncols;
    acol->datatype = type;
    acol->width = 0;
    afk_block_append_col( ablock, acol );
  }

 return status;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_raw_key_write                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                      afBlock  *ablock - block pointer       */
/*               const  afStyle  *astyle - style pointer       */
/*               const  char     *line                         */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmuKey* afk_raw_key_write( afBlock* ablock, long pos, const char* name,
                           dmDataType type, const void* value,
                           const char* unit, const char* desc )
{
  char keyline[AK_COMMENTLINE_LEN];

  dmuKeyList keylist;

  dmuKey    *ukey      = NULL;
  dmuKey    *old_ukey  = NULL;
  afBuffer  *abuf      = NULL;
  afStyle   *astyle    = NULL;
  long       old_entry = 0;
  ftKeyType  fktype    = FTK_UNKNOWN; 
  dmBool     mask      = dmFALSE;
  dmErrCode  status    = dmSUCCESS;


  keylist = ablock->keylist;
  abuf    = afk_block_get_buffer( ablock );
  astyle  = ablock->ads->astyle;
  fktype  = ftp_datatype_to_keytype( type, name );

  if ( fktype == FTK_COMMENT )
  {
    old_entry = 0;
  }
  else
  {
    old_ukey = dmu_key_find( keylist, (char*)name );
    ukey = old_ukey;

    if ( ukey )
      old_entry = ukey->index;
  }

  if ( !ukey )
  {
    ukey = dmu_key_alloc();
    ukey->type = FTK_KEY;
    dmu_key_insert( keylist, ukey, pos );
  }
 
  mask = afk_key_private( name );
  dmu_key_set( ukey, fktype, type, name, value, unit, desc, mask );

  /* If the style doesn't have a header, quit without actually writing */
  /* the key to the buffer                                             */
  if ( !afk_style_has_header( astyle ) )
    return ukey;
 
  status = afk_raw_compose_keyline( name, value, type, unit, desc,
                                    keyline, AK_COMMENTLINE_LEN, astyle );

  if ( old_entry == 0 && status == dmSUCCESS )
  {
    status = afk_buffer_insert_keyline( abuf, keyline, pos );
  }
  else
    if ( status == dmSUCCESS )
    {
      status = afk_buffer_replace_keyline( abuf, keyline, old_entry );
    }

  if ( status != dmSUCCESS )
  {
    /* Note: key specific message already in buffer */
    strcat( dm_errbuf, "  Keyword not written.");
    dmu_err_set( DM_ERR_KERNEL, dm_errbuf );

    dmu_key_delete( keylist, ukey );
    ukey = NULL;
  }

  return ukey;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_raw_compose_keyline                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*               const  char     *name                         */
/*               const  void     *value                        */
/*               dmDataType       type                         */
/*               const  char     *unit                         */
/*               const  char     *desc                         */
/*               long             klen                         */
/*               afStyle         *astyle                       */
/*                                                             */
/*  Output parameters:                                         */
/*               char            *keyline                      */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afk_raw_compose_keyline( const char* name, const void* value,
                                   dmDataType type, const char* unit,
                                   const char* desc, char* keyline,
                                   long klen, afStyle* astyle )
{
  char  prefix[4];
  char  fmt[16];
  char  comment_buf[FITS_MAXLEN];
  char *comment = NULL;
  char *cptr    = NULL;
  long  nlen    = 0;
  long  vlen    = 0;
  long  clen    = 0;
  long  len     = 0;
  char  hchar   = '\0';


  nlen = strlen( name );

  if ( nlen + 16 > klen )
  {
    keyline[0] = '\0';
    sprintf( dm_errbuf, "TEXT Kernel:  Keyword name exceeds line limit of %ld bytes.", klen );
    dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
    return dmFAILURE;
  }

  hchar = afk_style_get_header_char( astyle );

  if ( hchar )
  {
   prefix[0] = hchar;
   prefix[1] = ' ';
   prefix[2] = '\0';
  }
  else
    prefix[0] = '\0';

  len += sprintf( keyline, "%s%s = ", prefix, name );
  cptr = keyline + len;

  afk_key_default_fmt( type, value, fmt, astyle );

  switch ( type )
  {
    case DT_INT2:
      len += sprintf( cptr, fmt, *((short*)value) );
      break;
    case DT_UINT2:
      len += sprintf( cptr, fmt, *((unsigned short*)value) );
       break;
    case DT_UINT1: 
      len += sprintf( cptr, fmt, *((unsigned char*)value) );
      break;
    case DT_INT4:
      len += sprintf( cptr, fmt, *((long*)value) );
      break;
    case DT_UINT4:
      len += sprintf( cptr, fmt, *((unsigned long*)value) );
      break;
    case DT_REAL4:
      len += sprintf( cptr, fmt, *((float*)value) );
      break;
    case DT_REAL8:
      len += sprintf( cptr, fmt, *((double*)value) );
      break;
    case DT_CHAR:
      vlen = strlen( *((char**)value) );
 
      if ( nlen + 4 + vlen > klen )
      {
	sprintf( dm_errbuf, "TEXT Kernel:  Keyword %s exceeds line limit of %ld bytes.",
		 name, klen );
        dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
        keyline[0] = '\0';
        return dmFAILURE;
      }

      len += sprintf( cptr, fmt, *((char**)value) );
      break;
    case DT_LOGICAL:
      if ( *(dmBool*)value == UT_TRUE )
        len += sprintf( cptr, "T" );
      else
        len += sprintf( cptr, "F" );
      break;
    default:
      keyline[0] = '\0';
      sprintf( dm_errbuf, "TEXT header can't handle type: %d", type );
      dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
      return dmFAILURE;
  } // end switch

  // Code Change July 10, 2009  Bug 10802
  //   Subtract 1 from FITS_MAXLEN because copy functions need one byte
  //   free at end of buffer for null terminator. (81 char buffer - 80 char limit)

  comment = ftp_key_comment_compose( name, type, desc, unit, comment_buf, 
				     (FITS_MAXLEN-1) );
  if ( comment )
  {
    clen = strlen( comment );
  
    if ( clen + len + 5 > klen )
    {
      /* It's okay to just ignore the comment if it's too long */
      strcat( keyline, "\n" );
      return dmSUCCESS;
    }
  
    cptr = keyline + len;
    sprintf( cptr, " / %s\n", comment );
  }
  else
    strcat( keyline, "\n" );

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_raw_write_comment                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*               afBlock       *block                          */
/*               long           pos                            */
/*               const  char   *tag                            */
/*               const  char   *comment                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afk_raw_write_comment( afBlock* ablock, long pos,
                                 const char* tag, const char* comment )
{
  char cbuf[AK_COMMENTLINE_LEN];
  char prefix[4];

  dmuKeyList keylist;

  afBuffer  *abuf   = NULL;
  afStyle   *astyle = NULL;
  dmuKey    *ukey   = NULL;
  ftKeyType  fktype = FTK_UNKNOWN;
  dmErrCode  status = dmSUCCESS;
  long       mask   = UT_TRUE;
  long       keyno  = 0;
  char       hchar  = '\0';

  keylist = ablock->keylist;
  abuf = afk_block_get_buffer( ablock );
  fktype = FTK_COMMENT;

  for ( keyno = pos; keyno <= keylist->n; keyno++ ) /* was keyno=pos+1 */
  {
    ukey = keylist->data[keyno-1];
 
    if ( ukey->keytype != FTK_COMMENT )
      break;
  }


  ukey = dmu_key_alloc();
  ukey->type = FTK_KEY;
  dmu_key_insert( keylist, ukey, keyno );
  dmu_key_set( ukey, fktype, DT_CHAR, tag, &comment, NULL, NULL, mask );

  astyle = ablock->ads->astyle;
  hchar = afk_style_get_header_char( astyle );

  if ( hchar )
  {
    prefix[0] = hchar;
    prefix[1] = ' ';
    prefix[2] = '\0';
  }
  else
    prefix[0] = '\0';

  if ( strcmp( tag, dmCOMMENT ) == 0 )
    sprintf( cbuf, "%s%s\n", prefix, comment );
  else
    sprintf( cbuf, "%s%s %s\n", prefix, tag, comment );

  status = afk_buffer_insert_keyline( abuf, cbuf, keyno );
 
  if ( status != dmSUCCESS )
  {
    dmu_key_delete( keylist, ukey );
    ukey = NULL;
  }

  return status;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_raw_get_nrows                               */
/*                                                             */
/* Description:                                                */
/*         Count the number of rows in the block as quickly    */
/*         as possible                                         */
/*                                                             */
/*  Input parameters:                                          */
/*               afBlock       *block                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long  nrows                                      */
/*                                                             */
/***************************************************************/
long afk_raw_get_nrows( afBlock* ablock )
{
  const char *line  = NULL;
  afBuffer   *abuf  = NULL;
  afFile     *afptr = NULL;
  long        nrows = 0;

  abuf  = ablock->abuf;
  afptr = afk_dataset_fptr( ablock->ads );

  afk_buffer_moveto_row( abuf, 1 );

  while(( line = afk_file_getline_ptr( afptr )) != NULL )
  {
    if ( *line == '\0' )
      break;

    ++nrows;
  }

  return nrows;
}
