/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************************/
/*                                                                           */
/*                                                                           */
/*  File:  afksimple.c                                                       */
/*                                                                           */
/*   Functions that create and work with ASCII file type simple              */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/

#include "afkernelP.h"
 
const char* afk_simple_parse_next_col( afBlock* ablock, const afStyle* astyle,
                                      const char* lbegin, long ColumnNumber );





/*****************************************************************************/
/*                                                                           */
/*  Function:  afk_simple_read_header()                                      */
/*                                                                           */
/*  Input parameters:                                                        */
/*                    afBlock  *ablock                                       */
/*                    afStyle  *astyle                                       */
/*                                                                           */
/*  OutputpParameters:  none                                                 */
/*                                                                           */
/*****************************************************************************/
dmErrCode afk_simple_read_header( afBlock* ablock, afStyle* astyle )
{
  return afk_raw_read_header( ablock, astyle );
}


/*****************************************************************************/
/*                                                                           */
/*  Function:  afk_simple_init_cols()                                        */
/*                                                                           */
/*  Description:                                                             */
/*     Get column information and type                                       */
/*     This function is used for both simple files and cscrdb files          */
/*                                                                           */
/*  Input parameters:                                                        */
/*                    afBlock  *ablock                                       */
/*                    afStyle  *astyle                                       */
/*                                                                           */
/*  OutputpParameters:  dmErrCode status                                     */
/*                                                                           */
/*****************************************************************************/
dmErrCode afk_simple_init_cols( afBlock* ablock, afStyle* astyle )
{
  dmuKey       *key      = NULL;
  dmuKeyList    keylist;
  dmErrCode     status   = dmSUCCESS;
  dmDataType    type     = DT_UNK;

  afCol        *acol  = NULL;
  afFile       *afptr = NULL;
  afDataset    *ads   = NULL;

  const char*   cols_line = NULL;
  char          colval[AK_COLVAL_LEN];
  const char   *rptr         = NULL;
  long          ncols        = 0;
  long          ii           = 0;
  long          jj           = 0;


  keylist = afk_block_keylist( ablock );


  /* ryan, this should be a dmu keylist function                          */
  /* get the first or last key, it should be the columns description line */

  if ( astyle->colnames == AK_STYLE_COLNAMES_FIRST )
  {
    key = keylist->data[0];
    cols_line = dmu_key_value_c( key ); 
  }
  else
    if ( astyle->colnames == AK_STYLE_COLNAMES_LAST )
    {
      key = keylist->data[keylist->n-1];
      cols_line = dmu_key_value_c( key ); 
    }
  
  /* set up column descriptors */
  if ( cols_line )
  {
    status = afk_simple_read_cols( ablock, astyle, cols_line );

    if ( status != dmSUCCESS )
      return status;
  }

  /* now need to get the column types from the data  */

  ads   = afk_block_get_kds( ablock );
  afptr = afk_dataset_fptr( ads );
  ncols = afk_block_ncols( ablock );

  afk_block_set_row( ablock, 1 );

  // Code Change  Bug 11791  February 25, 2010
  // loop over all rows to get column data type

  while ( (rptr = afk_file_getline_ptr( afptr )) )
  {
    for ( ii = 0; ii < ncols && rptr; ii++ )
    {
      acol = afk_block_get_col( ablock, ii+1 );
      rptr = afk_gen_read_next_col( astyle, rptr, colval, AK_COLVAL_LEN );

      afk_gen_determine_col_datatype( astyle, acol, colval, &type);

      if ( rptr )
      {
        /* skip over array elements */

        for ( jj = 1; jj < acol->length; jj++ )
        {
          rptr = afk_gen_read_next_col( astyle, rptr, colval, AK_COLVAL_LEN );

          afk_gen_determine_col_datatype( astyle, acol, colval, &type);
        }
      }
    }
  }

  /* final rectification of column types */
  afk_gen_final_col_datatype( ablock, astyle );


  /* reset to start */
  afk_block_set_row( ablock, 1 );

 return status;
}


/*****************************************************************************/
/*                                                                           */
/*  Function:  afk_simple_read_cols()                                        */
/*                                                                           */
/*  Input parameters:                                                        */
/*                    afBlock        *ablock                                 */
/*                    const afStyle  *astyle                                 */
/*                    const char     *line                                   */
/*                                                                           */
/*  OutputpParameters:  none                                                 */
/*                                                                           */
/*****************************************************************************/
dmErrCode afk_simple_read_cols( afBlock* ablock, const afStyle* astyle,
                                const char* line )
{
  dmErrCode    status = dmSUCCESS;
  const char  *iptr   = line;
  long         ColumnNumber = 1;
 
  do
  {
    iptr = afk_simple_parse_next_col( ablock, astyle, iptr, ColumnNumber );
    ++ColumnNumber;
  } while( iptr );

  return status;
}



/*****************************************************************************/
/*                                                                           */
/*  Function:  afk_simple_parse_next_col()                                   */
/*                                                                           */
/*             Expects acol to be last col read.  Will overwrite values with */
/*             new column values.                                            */
/*                                                                           */
/*  Input parameters:                                                        */
/*                    afBlock        *ablock                                 */
/*                    const afStyle  *astyle                                 */
/*                    cpmst char     *lbegin                                 */
/*                                                                           */
/*  Output parameters:  none                                                 */
/*                                                                           */
/*****************************************************************************/
const char* afk_simple_parse_next_col( afBlock* ablock, const afStyle* astyle,
                                       const char* lbegin, long ColumnNumber )
{
  dmBool  bMultiSep   = dmFALSE;
  dmBool  bIsVector   = dmFALSE;
  dmBool  bIsArray    = dmFALSE;
  dmBool  bStartQuote = dmFALSE;

  const char  *pInPtr     = lbegin;
  const char  *separators = NULL;

  afCol  *pACol  = NULL;

  char  *pBeginToken = NULL;
  char  *pEndToken   = NULL;
  char  *pTmpString  = NULL;
  char  *pLBracket   = NULL;
  char  *pRBracket   = NULL;
  char  *pLParen     = NULL;
  char  *pRParen     = NULL;

  char   cQuote      = '\"';
  char   cCopyBuffer[AK_LINE_LEN];
  char   cToken[AK_LINE_LEN];
  char   cGenericColName[CARD_SIZE];
  char   cColName[CARD_SIZE];

  long   lTokenLength = 0;
  long   lArraySize   = 0;
  int    ii           = 0;

  // if null string, return
  if ( !pInPtr )
    return NULL;

  // are multi-separators allowed?
  bMultiSep = afk_style_get_multisep( astyle );

  // find first non-separator character
  // This routine leaves the pointer at the end of the previous token.
  // Check for and skip single separator, then.. if multi any additional ones.

  if ( *pInPtr != '\0' && afk_style_is_sep( astyle, *pInPtr ))
    ++pInPtr;

  if ( bMultiSep )
  {
    while ( *pInPtr != '\0' && afk_style_is_sep( astyle, *pInPtr ))
      ++pInPtr;
  }

  // if null string, return
  if ( *pInPtr == '\0' )
    return NULL;


  // is string in quotes?
  if ( *pInPtr == cQuote )
  {
    bStartQuote = dmTRUE;
    ++pInPtr;
  }

  // set start of token search
  pBeginToken = (char *) pInPtr;

  // get separator character(s)
  separators = afk_style_get_sep(astyle);

  // select first token by scanning until first separator character
  // if this column name is in quotes, scan until end quote

  if ( ! bStartQuote )
    pEndToken = strpbrk(pInPtr, separators);
  else
    pEndToken = strchr(pInPtr, cQuote );


  // if not null pointer, get the token length
  // otherwise, token length is entire string

  if ( pEndToken )
    lTokenLength = pEndToken - pBeginToken;
  else
    lTokenLength = strlen(pBeginToken);

  // Enforce length maximum.
  if (lTokenLength >= AK_LINE_LEN)
    lTokenLength = AK_LINE_LEN - 1;

  // get the token from the input string
  strncpy(cToken, pBeginToken, lTokenLength );
  cToken[lTokenLength] = '\0';

  // move input pointer to end of token.
  pInPtr = pEndToken;

  if ( bStartQuote )
    ++pInPtr;

  // Code Change  Bug 11390 November 21, 2008
  // Remove leading and trailing white space
  afk_strip_value(cToken, (afStyle *)astyle);
  lTokenLength = strlen(cToken);


  // mark array and vector symbols in the token

  pLBracket = strchr(cToken, '[');

  if (pLBracket)
  {
    pRBracket = strchr(cToken, ']');
    if (pRBracket > pLBracket )
      bIsArray = dmTRUE;
  }

  pLParen   = strchr(cToken, '(');
  if (pLParen)
  {
    pRParen = strchr(cToken, ')');
    if (pRParen > pLParen)
      bIsVector = dmTRUE;
  }


  // if the column name does not indicate that it
  // is an array or a vector, then parsing is complete
  // allocate column space, set name, and append to block

  if ( !bIsArray && !bIsVector)
  {
    pACol = afk_col_alloc();

    if ( lTokenLength )
    {
      // enforce column name length limit.
      if ( lTokenLength >= CARD_SIZE )
	lTokenLength = CARD_SIZE - 1;

      pACol->name = cs_dup_siz( cToken, lTokenLength );
    }
    else
    {
      // No column name specified, so make a generic one
      sprintf( cGenericColName, "col%ld", ColumnNumber );
      pACol->name = cs_dup( cGenericColName );
    }

    afk_block_append_col( ablock, pACol );

    // done with this column
  }
  else
  {
    /* Code section below originally modified for Bug 9531  October 31, 2005    */
    /* Further modification made January 2006                                   */
    /*                                                                          */
    /* The column name of array or vector can be null string, so insert generic */
    /* column name before the vector or array specification                     */
 
    /* parse string for column name */
    /* break when NULL or seperator character encountered */


      /* Character encountered while parsing column name is either a ( or a ] */
      /* If length of column name greater than 0, then break and treat what   */
      /* remains in buffer as either a vector or an array information.        */
      /* If length of column name is 0, use ( or [ as the start of the        */
      /* column name.  Continue with parsing until next seperator or NULL is  */
      /* encountered.                                                         */

    if (bIsArray && bIsVector)
    {
      if (pLBracket < pLParen)
        lTokenLength = pLBracket - cToken;
      else
        lTokenLength = pLParen - cToken;
    }
    else
    {
      if (bIsArray)
        lTokenLength = pLBracket - cToken;
      else
        if (bIsVector)
          lTokenLength = pLParen - cToken;
    }

    if ( !lTokenLength )
    {
      // No column name specified, so Make generic column number
      sprintf( cGenericColName, "col%ld", ColumnNumber );
    }
    /*  end new code section for Bug 9531 */


    /* at this point, have column name   */
    /* enforce column name length limit. */
    if ( lTokenLength >= CARD_SIZE )
      lTokenLength = CARD_SIZE - 1;

    if ( lTokenLength )
      strncpy(cColName, cToken, lTokenLength );
    else
    {
      lTokenLength = strlen(cGenericColName);
      strncpy(cColName, cGenericColName, lTokenLength);
    }

    cColName[lTokenLength]='\0';

    /* test if the column is an array or a vector or a vector array*/
    if ( bIsArray )
    {
      // get the array length

      lTokenLength = pRBracket - pLBracket - 1;

      if ( lTokenLength )
      {
        pTmpString = cs_dup_siz( pLBracket+1, lTokenLength );
        lArraySize = atol( pTmpString );
        free( pTmpString );
      }
      else
        lArraySize = 1;

    } /* end if bIsArray*/


    if (!bIsVector )
    {
      // if not a vector, this column is only an array column
      // size and name are determined at this point
      // allocate column space, set name and length, append to block

      pACol = afk_col_alloc();

      pACol->name = cs_dup( cColName );

      pACol->length = lArraySize;
      afk_block_append_col( ablock, pACol );

      // done with this column
    }
    else
    {
      // it's a vector column

       pBeginToken = pLParen+1;
    
      // get names held between the parenthesis

      while (pBeginToken < pRParen)
      {
        // move past any white space
        while( *pBeginToken != '\0' && *pBeginToken == ' ')
           ++pBeginToken;

        // get the vector column name
        pEndToken = (char *)pBeginToken;

        // look for end or separator comma
        while ( *pEndToken != '\0' && *pEndToken != ',' && *pEndToken != ')' )
          ++pEndToken;

        // found vector name
        // append to column name

        lTokenLength = pEndToken - pBeginToken;


        // get rid of training blanks in string
        for ( ii = 0; ii < lTokenLength; ii++)
          if (pBeginToken[ii] == ' ')
            break;

        if ( ii != lTokenLength )
          lTokenLength = ii;


        pACol = afk_col_alloc();
        pTmpString = cs_dup_siz( pBeginToken, lTokenLength );
        sprintf( cCopyBuffer, "%s(%s)", cColName, pTmpString );
        free( pTmpString );

	lTokenLength = strlen(cCopyBuffer);
	/* enforce column name length limit. */
	if ( lTokenLength >= CARD_SIZE )
	  lTokenLength = CARD_SIZE - 1;

        pACol->name   = cs_dup_siz( cCopyBuffer, lTokenLength );

        // the vector may be in an array
        // if so, set vector column length to same as array length

        if ( bIsArray )
          pACol->length = lArraySize;


        // append column to block
        afk_block_append_col( ablock, pACol );

	// increment pointer to continue search   
        pBeginToken = pEndToken + 1;

      } /* end while( pBeginToken <  pRParen ) */

    } /* end else if bIsVector */

  }  // end else if !bIsArray && !bIsVector


  // return pointer to next token in input string
  return pInPtr;
}


/*****************************************************************************/
/*                                                                           */
/*  Function:  afk_simple_key_write()                                        */
/*                                                                           */
/*                                                                           */
/*  Input parameters:                                                        */
/*                    afBlock        *ablock                                 */
/*                    long            pos                                    */
/*                    const char     *name                                   */
/*                    dmDataType      type                                   */
/*                    const void     *value                                  */
/*                    const char     *unit                                   */
/*                    const char     *desc                                   */
/*                                                                           */
/*  Output parameters:  none                                                 */
/*                                                                           */
/*****************************************************************************/
dmuKey* afk_simple_key_write( afBlock* ablock, long pos, const char* name,
                              dmDataType type, const void* value,
                              const char* unit, const char* desc )
{
  afStyle    *astyle  = NULL;
  dmuKeyList  keylist = ablock->keylist;
  long        first   = 0; /* first valid new key position, need to keep colnames line*/
  long        last    = 0; /* last valid new key position, need to keep colnames line*/
  afFlavour   flavor  = AK_FLAV_UNK;

 
  astyle = ablock->ads->astyle;

  flavor = astyle->flavor;

  /* default values */
  first = 1;
  last  = keylist->n+1;

  /* adjust default values based on col names type */

  if ( astyle->colnames == AK_STYLE_COLNAMES_FIRST )
  {
    first = 2;
    last = keylist->n+1;
  }
  else
    if ( astyle->colnames == AK_STYLE_COLNAMES_LAST )
    {
      first = 1;
      last = keylist->n;
    }
 
  /* verify range */
  if ( pos < first )
    pos = first;
  else
    if ( pos > last )
      pos = last;

  /* write key info */
  return afk_raw_key_write( ablock, pos, name, type, value, unit, desc );
}


/*****************************************************************************/
/*                                                                           */
/*  Function:  afk_simple_write_comment()                                    */
/*                                                                           */
/*                                                                           */
/*  Input parameters:                                                        */
/*                    afBlock        *ablock                                 */
/*                    long            pos                                    */
/*                    const char     *tag                                    */
/*                    const char     *comment                                */
/*                                                                           */
/*  Output parameters:  none                                                 */
/*                                                                           */
/*****************************************************************************/
dmErrCode afk_simple_write_comment( afBlock* ablock, long pos, const char* tag,
                                    const char* comment )
{
 char       cbuf[AK_COMMENTLINE_LEN];
 char       prefix[4];
 afBuffer  *abuf   = NULL;
 afStyle  * astyle = NULL;
 dmuKey    *ukey   = NULL;
 dmuKeyList keylist;
 long       mask   = UT_TRUE;
 long       keyno  = 0;
 ftKeyType  fktype = FTK_UNKNOWN;
 dmErrCode  status = dmSUCCESS;
 char       hchar  = '\0';

 keylist = ablock->keylist;
 abuf    = afk_block_get_buffer( ablock );
 fktype  = FTK_COMMENT;

 // initialize prefix to null
 prefix[0] = hchar;

 for ( keyno = pos; keyno < keylist->n; keyno++ ) /* was keyno=pos+1 */
 {
   ukey = keylist->data[keyno-1];

   if ( ukey->keytype != FTK_COMMENT )
     break;
 } 

 ukey       = dmu_key_alloc();
 ukey->type = FTK_KEY;

 dmu_key_insert( keylist, ukey, pos );
 dmu_key_set( ukey, fktype, DT_CHAR, tag, &comment, NULL, NULL, mask );

 astyle = ablock->ads->astyle;
 hchar  = afk_style_get_header_char( astyle );

 if ( hchar )
 {
   prefix[0] = hchar;
   prefix[1] = '\0';
   prefix[2] = '\0';
 }

 if ( strcmp( tag, AK_SPECIFICATION) == 0)
   sprintf (cbuf, "%s", comment);
 else
 {
 if ( strcmp( tag, dmCOMMENT ) == 0 )
   sprintf( cbuf, "%s%s\n", prefix, comment );
 else
   sprintf( cbuf, "%s%s %s\n", prefix, tag, comment );
 }

 status = afk_buffer_insert_keyline( abuf, cbuf, pos );
 
 if ( status != dmSUCCESS )
 {
   dmu_key_delete( keylist, ukey );
   ukey = NULL;
 }

 return status;
}



//This function is identical to afk_raw_get_nrows()
//Not needed
#if 0
/*****************************************************************************/
/*                                                                           */
/*  Function:  afk_simple_get_nrows                                          */
/*                                                                           */
/*             count the number of rows in the block as quickly as possible  */
/*                                                                           */
/*  Input parameters:                                                        */
/*                    afBlock        *ablock                                 */
/*                                                                           */
/*  Output parameters:  none                                                 */
/*                                                                           */
/*****************************************************************************/
long afk_simple_get_nrows( afBlock* ablock )
{
  const char *line  = NULL;
  afBuffer   *abuf  = NULL;
  afFile     *afptr = NULL;

  long nrows = 0;

  abuf = ablock->abuf;
  afptr = afk_dataset_fptr( ablock->ads );
  afk_buffer_moveto_row( abuf, 1 );

  while(( line = afk_file_getline_ptr( afptr, dmUNKBLOCK )) != NULL )
  {
    if ( *line == '\0' )
      break;

    ++nrows;
  }

 return nrows;
}
#endif
