/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afkstyle.c                                                  */
/*                                                               */
/*   Functions that create and work with ASCII datasets          */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/
#include "afkernelP.h"

const char textStringKernelUnknown[21]   = "";
const char textStringKernelRaw[21]       = "#TEXT/RAW";
const char textStringKernelSimple[21]    = "#TEXT/SIMPLE";
const char textStringKernelDTF[21]       = "#TEXT/DTF";
const char textStringKernelDTFFIXED[21]  = "#TEXT/DTF-FIXED";
const char textStringKernelRDB[21]       = "#TEXT/RDB";
const char textStringKernelCSCRDB[21]    = "#TEXT/CSCRDB";
const char textStringKernelTSV[27]       = "#TEXT/TAB-SEPARATED-VALUES";
const char textStringKernelUser[21]      = "#TEXT/USER";


/***************************************************************/
/*                                                             */
/* Function:   afk_style_alloc                                 */
/*                                                             */
/* Description:                                                */
/*  Allocate dataset's style structure                         */
/*                                                             */
/*  Input parameters:                                          */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            afStyle  *astyle - pointer to style structure    */
/*                                                             */
/***************************************************************/
afStyle* afk_style_alloc( void )
{
  afStyle *astyle = NULL;

  astyle = malloc( sizeof(afStyle) );

  if ( astyle )
  {
    astyle->has_header   = dmFALSE;
    astyle->header_char  = '#';
    astyle->multiple     = dmFALSE;
    astyle->fixed        = dmFALSE;
    astyle->multisep     = dmTRUE;
    astyle->use_fits     = dmFALSE;
    astyle->skip         = 0;
    astyle->colnames     = AK_STYLE_COLNAMES_UNKNOWN;
 
    astyle->key_flag = NULL;
    astyle->sep      = NULL;
    astyle->quotes   = NULL;
    astyle->nullstr  = NULL;

    afk_style_set_flavor (astyle, AK_FLAV_UNK, NULL);
    afk_style_set_key_flag( astyle, "Key" );
    afk_style_set_sep( astyle, " \t\r" );
    afk_style_set_quotes( astyle, "\"'" );

    astyle->white = cs_dup(" \t\r");
  }

 return astyle;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_free                                  */
/*                                                             */
/* Description:                                                */
/*  Deallocate dataset's style structure                       */
/*                                                             */
/*  Input parameters:                                          */
/*    afStyle  *astyle  -- style struct pointer                */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_style_free( afStyle* astyle )
{
  if ( astyle )
  {
    if ( astyle->sep && *astyle->sep != '\0')
      free( astyle->sep );

    if ( astyle->key_flag && *astyle->key_flag != '\0')
      free( astyle->key_flag );

    if ( astyle->quotes  && *astyle->quotes != '\0')
      free( astyle->quotes );

    if ( astyle->white  && *astyle->white != '\0')
      free( astyle->white );

    if ( astyle->nullstr  && *astyle->nullstr != '\0')
      free( astyle->nullstr );
    
    free( astyle );

    astyle = NULL;
  }
}


/***************************************************************/
/*                                                             */
/* Function:   afk_style_set_quotes                            */
/*                                                             */
/* Description:                                                */
/*     set file quote element                                  */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                   const char * quotes -- ptr to chars       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_style_set_quotes( afStyle* astyle, const char* quotes )
{
  if ( astyle ) 
  {
    if ( astyle->quotes && *astyle->quotes != '\0')
      free( astyle->quotes );

    if ( quotes )
      astyle->quotes = cs_dup( quotes );
    else
      astyle->quotes = NULL;
  }
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_get_quotes                            */
/*                                                             */
/* Description:                                                */
/*     get quote element                                       */
/*                                                             */
/*  Input parameters:                                          */
/*                   const afStyle  *astyle - style pointer    */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            const char *quotes                               */
/*                                                             */
/***************************************************************/
const char* afk_style_get_quotes( const afStyle* astyle )
{
  if ( astyle )
    return astyle->quotes;
  return NULL;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_set_header                            */
/*                                                             */
/* Description:                                                */
/*     set file has header element                             */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                   dmBool    has_header                      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_style_set_header( afStyle* astyle, dmBool has_header )
{
  if ( astyle )
    astyle->has_header = has_header;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_has_header                            */
/*                                                             */
/* Description:                                                */
/*     returns has_header element                              */
/*                                                             */
/*  Input parameters:                                          */
/*                   const afStyle  *astyle -   struct pointer */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmBool True or False                             */
/*                                                             */
/***************************************************************/
dmBool afk_style_has_header( const afStyle* astyle )
{
  if ( astyle )
    return astyle->has_header;
  else
    return dmFALSE;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_set_header_char                       */
/*                                                             */
/* Description:                                                */
/*     set file has header element r                           */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                   dmBool    header_char                     */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_style_set_header_char( afStyle* astyle, char header_char )
{
  if ( astyle )
    astyle->header_char = header_char;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_get_header_char                       */
/*                                                             */
/* Description:                                                */
/*     get header char element                                 */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            char  header_char                                */
/*                                                             */
/***************************************************************/
char afk_style_get_header_char( const afStyle* astyle )
{
  if ( astyle )
    return astyle->header_char;
  else 
    return '\0';
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_set_key_flag                          */
/*                                                             */
/* Description:                                                */
/*     set file key_flag element                               */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                   const char * key_flag                     */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_style_set_key_flag( afStyle* astyle, const char* key_flag )
{
  if ( astyle )
  {
    if ( astyle->key_flag && *astyle->key_flag != '\0')
      free( astyle->key_flag );
 
   if ( key_flag )
      astyle->key_flag = cs_dup( key_flag );
    else
      astyle->key_flag = NULL;
  }
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_get_key_flag                          */
/*                                                             */
/* Description:                                                */
/*     return key flag element                                 */
/*                                                             */
/*  Input parameters:                                          */
/*                   const afStyle  *astyle  -e struct pointer */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            const char *key_flag                             */
/*                                                             */
/***************************************************************/
const char* afk_style_get_key_flag( const afStyle* astyle )
{
  if ( astyle )
    return astyle->key_flag;
  else
    return NULL;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_set_multiple                          */
/*                                                             */
/* Description:                                                */
/*     set multiple element                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_style_set_multiple( afStyle* astyle, dmBool multiple )
{
  if ( astyle )
    astyle->multiple = multiple;
}




/***************************************************************/
/*                                                             */
/* Function:   afk_style_get_multiple                          */
/*                                                             */
/* Description:                                                */
/*     returns multiple elemente                               */
/*                                                             */
/*  Input parameters:                                          */
/*                   const afStyle  *astyle  -  struct pointer */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmBool  true or false                            */
/*                                                             */
/***************************************************************/
dmBool afk_style_get_multiple( const afStyle* astyle )
{
  if ( astyle )
    return astyle->multiple;
  else
    return dmFALSE;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_set_fixed                             */
/*                                                             */
/* Description:                                                */
/*     set fixed element                                       */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                   dmBool    fixed                           */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_style_set_fixed( afStyle* astyle, dmBool fixed )
{
  if ( astyle )
    astyle->fixed = fixed;
}




/***************************************************************/
/*                                                             */
/* Function:   afk_style_get_fixed                             */
/*                                                             */
/* Description:                                                */
/*     get kernel spec. fixed elemente                         */
/*                                                             */
/*  Input parameters:                                          */
/*                   const afStyle  *astyle  -  struct pointer */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmBool  true or false                            */
/*                                                             */
/***************************************************************/
dmBool afk_style_get_fixed( const afStyle* astyle )
{
  if ( astyle )
    return astyle->fixed;
  else
   return dmFALSE;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_set_sep                               */
/*                                                             */
/* Description:                                                */
/*     set sep element                                         */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                   const char * sep                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_style_set_sep( afStyle* astyle, const char* sep )
{
  if ( astyle )
  {
    if ( astyle->sep && *astyle->sep != '\0')
      free( astyle->sep );

    if ( sep )
      astyle->sep = cs_dup( sep );
    else
      astyle->sep = NULL;
  } 
}




/***************************************************************/
/*                                                             */
/* Function:   afk_style_get_sep                               */
/*                                                             */
/* Description:                                                */
/*     returns seperator element                               */
/*                                                             */
/*  Input parameters:                                          */
/*                   const afStyle  *astyle  -  struct pointer */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            const char  *sep - seperator char                */
/*                                                             */
/***************************************************************/
const char* afk_style_get_sep( const afStyle* astyle )
{
  if ( astyle )
    return astyle->sep;
  else
    return NULL;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_set_multisep                          */
/*                                                             */
/* Description:                                                */
/*     set multisep element                                    */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                   dmBool    multisep                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_style_set_multisep( afStyle* astyle, dmBool multisep )
{
  if ( astyle )
    astyle->multisep = multisep;
}


/***************************************************************/
/*                                                             */
/* Function:   afk_style_get_multisep                          */
/*                                                             */
/* Description:                                                */
/*     returns multisep element                                */
/*                                                             */
/*  Input parameters:                                          */
/*                   const afStyle  *astyle - e struct pointer */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmBool  True or False                            */
/*                                                             */
/***************************************************************/
dmBool afk_style_get_multisep( const afStyle* astyle )
{
 if ( astyle ) 
  return astyle->multisep;
 return dmFALSE;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_set_use_fits                          */
/*                                                             */
/* Description:                                                */
/*     set use fits element                                    */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                   dmBool    use_fits                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_style_set_use_fits( afStyle* astyle, dmBool use_fits )
{
  if ( astyle )
    astyle->use_fits = use_fits;
}





/***************************************************************/
/*                                                             */
/* Function:   afk_style_use_fits                              */
/*                                                             */
/* Description:                                                */
/*     set use fits element                                    */
/*                                                             */
/*  Input parameters:                                          */
/*                   const afStyle  *astyle  -  struct pointer */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*           dmBool True or False                              */
/*                                                             */
/***************************************************************/
dmBool afk_style_use_fits( const afStyle* astyle )
{
  if ( astyle ) 
    return astyle->use_fits;
 
 return dmFALSE;
}




/***************************************************************/
/*                                                             */
/* Function:   afk_style_set_skip                              */
/*                                                             */
/* Description:                                                */
/*     set skip element                                        */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                   long      skip                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_style_set_skip( afStyle* astyle, long skip )
{
  astyle->skip = skip;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_get_skip                              */
/*                                                             */
/* Description:                                                */
/*     returns skip element                                    */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            long  skip                                       */
/*                                                             */
/***************************************************************/
long afk_style_get_skip( afStyle* astyle )
{
  return astyle->skip;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_set_colnames                          */
/*                                                             */
/* Description:                                                */
/*     set column names element                                */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                   int       colnames                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_style_set_colnames( afStyle* astyle, int colnames )
{
  astyle->colnames = colnames;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_get_colnames                          */
/*                                                             */
/* Description:                                                */
/*     teturns colnames element                                */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                   int       colnames                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            int  colnames                                    */
/*                                                             */
/***************************************************************/
int afk_style_get_colnames( afStyle* astyle, int colnames )
{
  (void) colnames; 
  return astyle->colnames;
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_set_flavor                            */
/*                                                             */
/* Description:                                                */
/*     sets kernel specification element                       */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle   *astyle  -- style struct pointer*/
/*                   afFlavour  flavor                         */
/*                   cont char *flavorString                   */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void  afk_style_set_flavor( afStyle    *astyle,
                            afFlavour   flavor,
                            const char *flavorString )
{
  char  flavorTextString[CARD_SIZE];

  if ( astyle )
  {
    astyle->flavor = flavor;  

    if (flavorString != NULL)
     strcpy( astyle->flavorString, flavorString );
    else
    {
      switch (flavor)
      {
	case  AK_FLAV_UNK:
        default:
          strcpy(flavorTextString, textStringKernelUnknown) ;
          break;
        case  AK_FLAV_RAW:
          strcpy(flavorTextString, textStringKernelRaw );
	  break;
        case  AK_FLAV_SIMPLE:
          strcpy( flavorTextString, textStringKernelSimple);
          break;
        case  AK_FLAV_DTF:
          if (astyle->fixed)
            strcpy( flavorTextString, textStringKernelDTFFIXED);
          else
            strcpy(flavorTextString, textStringKernelDTF);
          break;
        case  AK_FLAV_RDB:
          strcpy( flavorTextString, textStringKernelRDB);
          break;
        case  AK_FLAV_CSCRDB:
          strcpy( flavorTextString, textStringKernelCSCRDB);
          break;
        case  AK_FLAV_TSV:
          strcpy( flavorTextString, textStringKernelTSV);
          break;
      }
      strcpy(astyle->flavorString, flavorTextString );
    }
  }  
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_get_flavor                            */
/*                                                             */
/* Description:                                                */
/*     returns kernel specification element                    */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            afFlavor  flavor                                 */
/*                                                             */
/***************************************************************/
int  afk_style_get_flavor( afStyle* astyle )
{
  if ( astyle )
    return astyle->flavor;

  return AK_FLAV_UNK;  
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_get_flavorString                      */
/*                                                             */
/* Description:                                                */
/*     returns kernel specification text string element        */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            const char  *flavorString                        */
/*                                                             */
/***************************************************************/
const char  *afk_style_get_flavorString( afStyle *astyle )
{
  if ( astyle )
    return astyle->flavorString;

   return NULL;  
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_set_nullstr                           */
/*                                                             */
/* Description:                                                */
/*     set nullstr element                                     */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                   const char *nullstr - replacement string  */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_style_set_nullstr( afStyle* astyle, const char* nulltext )
{
  if ( astyle )
  {
    if ( astyle->nullstr && *astyle->nullstr != '\0')
      free( astyle->nullstr );

    if ( nulltext )
      astyle->nullstr = cs_dup( nulltext );
    else
      astyle->nullstr = NULL;
  } 
}




/***************************************************************/
/*                                                             */
/* Function:   afk_style_get_nullstr                           */
/*                                                             */
/* Description:                                                */
/*     returns null value replacement string                   */
/*                                                             */
/*  Input parameters:                                          */
/*                   const afStyle  *astyle  -  struct pointer */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            const char  *sep - seperator char                */
/*                                                             */
/***************************************************************/
const char* afk_style_get_nullstr( const afStyle* astyle )
{
  if ( astyle )
    return astyle->nullstr;
  else
    return NULL;
}

/***************************************************************/
/*                                                             */
/* Function:   afk_style_set                                   */
/*                                                             */
/* Description:                                                */
/*     set defaults for style structure based on kernel spec.  */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                   afFlavour flav   -- kernel specification  */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_style_set( afStyle* astyle, afFlavour flavor )
{
  // set flavor and flavor string upon entry
  afk_style_set_flavor ( astyle, flavor, NULL );

  switch ( flavor )
  {
    case AK_FLAV_DTF:
      afk_style_set_header( astyle, dmTRUE );
      afk_style_set_header_char( astyle, '\0' );
      afk_style_set_key_flag( astyle, NULL );
      afk_style_set_multiple( astyle, dmTRUE );
      afk_style_set_fixed( astyle, dmFALSE );
      afk_style_set_sep( astyle, " \t\r" );
      afk_style_set_multisep( astyle, dmTRUE );
      afk_style_set_quotes( astyle, "\"'" );
      afk_style_set_use_fits( astyle, dmTRUE );
      afk_style_set_skip( astyle, 0 );
      afk_style_set_colnames( astyle, AK_STYLE_COLNAMES_NONE );
      break;

    case AK_FLAV_SIMPLE:
      afk_style_set_header( astyle, dmFALSE );
      afk_style_set_header_char( astyle, '#' );
      afk_style_set_key_flag( astyle, "Key" );
      afk_style_set_multiple( astyle, dmFALSE );
      afk_style_set_fixed( astyle, dmFALSE );
      afk_style_set_sep( astyle, " \t\r" );
      afk_style_set_multisep( astyle, dmTRUE );
      afk_style_set_quotes( astyle, "\"'" );
      afk_style_set_use_fits( astyle, dmFALSE );
      afk_style_set_skip( astyle, 0 );
      afk_style_set_colnames( astyle, AK_STYLE_COLNAMES_FIRST );
      break;

    case AK_FLAV_RDB:
      afk_style_set_header( astyle, dmTRUE );
      afk_style_set_header_char( astyle, '\0' );
      afk_style_set_key_flag( astyle, "Key" );
      afk_style_set_multiple( astyle, dmFALSE );
      afk_style_set_fixed( astyle, dmFALSE );
      afk_style_set_sep( astyle, "\t\r" );
      afk_style_set_multisep( astyle, dmFALSE );
      afk_style_set_quotes( astyle, NULL );
      afk_style_set_use_fits( astyle, dmTRUE );
      afk_style_set_skip( astyle, 0 );
      afk_style_set_colnames( astyle, AK_STYLE_COLNAMES_LAST );
      break;

    case AK_FLAV_RAW:
      afk_style_set_header( astyle, dmFALSE );
      afk_style_set_header_char( astyle, '#' );
      afk_style_set_key_flag( astyle, "Key" );
      afk_style_set_multiple( astyle, dmFALSE );
      afk_style_set_fixed( astyle, dmFALSE );
      afk_style_set_sep( astyle, " \t\r" );
      afk_style_set_multisep( astyle, dmTRUE );
      afk_style_set_quotes( astyle, "\"'" );
      afk_style_set_use_fits( astyle, dmFALSE );
      afk_style_set_skip( astyle, 0 );
      afk_style_set_colnames( astyle, AK_STYLE_COLNAMES_NONE );
      break;

    case AK_FLAV_CSCRDB:
      afk_style_set_header( astyle, dmTRUE );
      afk_style_set_header_char( astyle, '#' );
      afk_style_set_key_flag( astyle, "Key" );
      afk_style_set_multiple( astyle, dmFALSE );
      afk_style_set_fixed( astyle, dmFALSE );
      afk_style_set_sep( astyle, "\t\r" );
      afk_style_set_multisep( astyle, dmFALSE );
      afk_style_set_quotes( astyle, NULL );
      afk_style_set_use_fits( astyle, dmFALSE );
      afk_style_set_skip( astyle, 0 );
      afk_style_set_colnames( astyle, AK_STYLE_COLNAMES_LM1 );
      afk_style_set_white( astyle, "\t");
      break;

    case AK_FLAV_TSV:
      afk_style_set_header( astyle, dmTRUE );
      afk_style_set_header_char( astyle, '#' );
      afk_style_set_key_flag( astyle, "Key" );
      afk_style_set_multiple( astyle, dmFALSE );
      afk_style_set_fixed( astyle, dmFALSE );
      afk_style_set_sep( astyle, "\t\r" );
      afk_style_set_multisep( astyle, dmFALSE );
      afk_style_set_quotes( astyle,  "\"'" );
      afk_style_set_use_fits( astyle, dmFALSE );
      afk_style_set_skip( astyle, 0 );
      afk_style_set_colnames( astyle, AK_STYLE_COLNAMES_TSV );
      break;

    case AK_FLAV_UNK:
    default:
      afk_style_set_header( astyle, dmFALSE );
      afk_style_set_header_char( astyle, '#' );
      afk_style_set_key_flag( astyle, "Key" );
      afk_style_set_multiple( astyle, dmFALSE );
      afk_style_set_fixed( astyle, dmFALSE );
      afk_style_set_sep( astyle, " \t\r" );
      afk_style_set_multisep( astyle, dmTRUE );
      afk_style_set_quotes( astyle, "\"'" );
      afk_style_set_use_fits( astyle, dmFALSE );
      afk_style_set_skip( astyle, 0 );
      afk_style_set_colnames( astyle, AK_STYLE_COLNAMES_UNKNOWN );
      break;
  }
}


/***************************************************************/
/*                                                             */
/* Function:   afk_style_is_sep                                */
/*                                                             */
/* Description:                                                */
/*     determines if character is separator                    */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                   char      c                               */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmBool TorF -- true or false                     */
/*                                                             */
/***************************************************************/
dmBool afk_style_is_sep( const afStyle* astyle, char c )
{
 return ( astyle && astyle->sep && strchr( astyle->sep, c ));
}



/***************************************************************/
/*                                                             */
/* Function:   afk_style_is_quote                              */
/*                                                             */
/* Description:                                                */
/*     determines if character is used as a quote character    */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                   char      c                               */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmBool  TorF   - true or false                   */
/*                                                             */
/***************************************************************/
dmBool afk_style_is_quote( const afStyle* astyle, char c )
{
 return ( astyle && astyle->quotes && strchr( astyle->quotes, c ));
}


/***************************************************************/
/*                                                             */
/* Function:   afk_style_set_white                             */
/*                                                             */
/* Description:                                                */
/*     set white  space string                                 */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                   char     *c                               */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afk_style_set_white( afStyle* astyle, char *white )
{
  if (astyle )
  {
    if ( astyle->white && *astyle->white != '\0' )
      free (astyle->white);

    if ( white )
      astyle->white = cs_dup(white);
    else
      astyle->white = cs_dup(" \t"); // default
  }
}

/***************************************************************/
/*                                                             */
/* Function:   afk_style_get_white                             */
/*                                                             */
/* Description:                                                */
/*     returns white space string                              */
/*                                                             */
/*  Input parameters:                                          */
/*                   afStyle  *astyle  -- style struct pointer */
/*                   char     *c                               */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
const char* afk_style_get_white( const afStyle* astyle )
{
  if (astyle )
    return (astyle->white);

  return NULL;
}

