/*                                                                
**  Copyright (C) 2003-2013  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afktsv.c                                                    */
/*                                                               */
/*   Functions that create and work with ASCII file type         */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"
#include "dmi.h"


void afk_tsv_column_key_parse( const char *psSeparators, const char *sKeyLine,
                               char *sColName, dmDataType *dataType, long *width,
                               long *length, long *arraysize, char *sDisplayFormat,
                               char *sUnit, char *sDescription );



/*****************************************************************************/
/*                                                                           */
/*  Function:  afk_tsv_column_key_parse                                      */
/*                                                                           */
/*  Description:                                                             */
/*     Take as input one comment line from file header (stored as comment    */
/*     key ) that contains column information and parse the key value        */
/*     according to the following format:                                    */
/*                                                                           */
/*     Column <name> (<datatype>) <description>  ["<units>"] [<comment>]     */
/*                                                                           */
/*     Function only called with non-null key line string                    */
/*                                                                           */
/*  Input parameters:                                                        */
/*              const char       *psSeparators                               */
/*              const char       *sKeyline                                   */
/*                                                                           */
/*  Output Parameters:                                                       */
/*                    char       *sColName                                   */
/*                    dmDataType *dataType                                   */
/*                    long       *width                                      */
/*                    long       *length                                     */
/*                    long       *arraysize                                  */
/*                    char       *unit                                       */
/*                                                                           */
/*  Return:                                                                  */
/*                    void                                                   */
/*                                                                           */
/*  Called by:                                                               */
/*      afp_tsv_init_cols()                                                  */
/*                                                                           */
/*****************************************************************************/
void afk_tsv_column_key_parse( const char *psSeparators, const char *sKeyLine,
                               char *sColName, dmDataType *dataType, long *width,
                               long *length, long *arraysize,
                               char *sDisplayFormat, char *sUnit, char *sDescription )
{
  dmBool      bVariable = dmFALSE;

  char *pTokenStart  = NULL;
  char *pTokenEnd    = NULL;

  char *sLParen      = "(";
  char *sRParen      = ")";
  char *sLBracket    = "[";
  char *sRBracket    = "]";
  char  cQuote       = '\"';

  char  cNull        = '\0';

  long  tokenSize = 0;

  // initialize output vars
  *dataType    = DT_UNK;
  *width       = 0;
  *length      = 0;
  *arraysize   = 0;
  sColName[0]       = cNull;
  sUnit[0]          = cNull;
  sDisplayFormat[0] = cNull;
 
 
  // assign local pointer
  pTokenStart = (char *)sKeyLine;

  // First, move past word "Column"
  pTokenEnd = strpbrk(pTokenStart, psSeparators);

  if ( !pTokenEnd || *pTokenEnd == cNull )
    return;

  // Move past separator
  pTokenStart = pTokenEnd;
 
  // Get column name
  pTokenEnd = strpbrk(++pTokenStart, psSeparators);

  // If no separator char, take everything until end of line
  // as token

  if ( !pTokenEnd )
    pTokenEnd = pTokenStart + (strlen(pTokenStart)-1);

  tokenSize = pTokenEnd - pTokenStart;

  // enforce col name length limit
  if ( tokenSize >= CARD_SIZE)
    tokenSize = CARD_SIZE - 1;

  strncpy(sColName, pTokenStart, tokenSize);
  sColName[tokenSize] = cNull;

  pTokenStart = pTokenEnd;

  // Find parenthesis indicating format code
  pTokenStart = strpbrk(++pTokenStart, sLParen);

  // no parenthesis means nothing more to see here
  if ( !pTokenStart || *pTokenStart == cNull )
    return;

  // Move past parenthisis
  ++pTokenStart;

  // Begin to get format string
 
  pTokenEnd = strpbrk(pTokenStart, sRParen);

  if ( !pTokenEnd || *pTokenEnd == cNull )
    return;

  tokenSize = pTokenEnd - pTokenStart;

  if ( tokenSize >= AK_COLVAL_LEN)
    tokenSize = AK_COLVAL_LEN - 1;

  strncpy(sDisplayFormat, pTokenStart, tokenSize);
  sDisplayFormat[tokenSize] = cNull;

  // Get arraysize, datatype, width and length using parsing method
  // used for keyword TFORM
  
  *dataType = afp_parse_tform( sDisplayFormat, dmTRUE, arraysize, length, width, &bVariable );

  // Store next token as description
  pTokenStart = pTokenEnd;
  pTokenStart = strpbrk(pTokenStart, psSeparators);

  if ( !pTokenStart )
    return;

  ++pTokenStart;
  pTokenEnd = strpbrk(pTokenStart, psSeparators);

  if ( pTokenEnd && (pTokenEnd > pTokenStart))
  {
    tokenSize = pTokenEnd - pTokenStart;

    if ( tokenSize >= AK_COLVAL_LEN)
      tokenSize = AK_COLVAL_LEN - 1;

    strncpy( sDescription, pTokenStart, tokenSize);

    sDescription[tokenSize] = cNull;
  }

  pTokenStart = pTokenEnd;

  // move past description string, stop at opening [
  pTokenStart = strpbrk(pTokenStart, sLBracket);

  if ( !pTokenStart || *pTokenStart == cNull )
    return;

  ++pTokenStart;;

  // go to closing bracket
  pTokenEnd = strpbrk(pTokenStart, sRBracket);

  if ( !pTokenEnd || *pTokenEnd == cNull )
    return;

  --pTokenEnd;

  // if unit string surrounded by quotes, strip them off
  if ( *pTokenEnd == cQuote )
    --pTokenEnd;

  if ( pTokenEnd <= pTokenStart )
    return;

  if (*pTokenStart == cQuote)
    ++pTokenStart;

  tokenSize = pTokenEnd - pTokenStart + 1;

  if ( tokenSize >= CARD_SIZE)
    tokenSize = CARD_SIZE - 1;

  strncpy(sUnit, pTokenStart, tokenSize);
  sUnit[tokenSize] = cNull;


  // return
  return;

}


/*****************************************************************************/
/*                                                                           */
/*  Function:  afk_tsv_init_cols()                                           */
/*                                                                           */
/*  Description:                                                             */
/*     Get column information and type                                       */
/*     This function is based on afk_simple_init_cols()                      */
/*                                                                           */
/*     Calls afk_simple_read_cols()                                          */
/*                                                                           */
/*  Input parameters:                                                        */
/*                    afBlock  *pakBlock                                     */
/*                    afStyle  *pakStyle                                     */
/*                                                                           */
/*  Output Parameters:  dmErrCode status                                     */
/*                                                                           */
/*  Called by:                                                               */
/*      afp_tsv_init_cols()                                                  */
/*                                                                           */
/*****************************************************************************/
dmErrCode afk_tsv_init_cols( afBlock *pakBlock, afStyle *pakStyle,
                             char ***ppUnitStrings, char ***ppDescriptions )
{
  dmuKeyList    keylist;
  dmuKey       *pKey      = NULL;
  dmErrCode     status   = dmSUCCESS;
  dmDataType    dataType = DT_UNK;

  afCol        *pakCol     = NULL;
  afFile       *pakFile    = NULL;
  afDataset    *pakDataset = NULL;

  const char   *psColumnNames = NULL;
  const char   *psRow         = NULL;
  const char   *psSeparators  = NULL;
  char        **ppUnits       = NULL;
  char        **ppComments    = NULL;
  char         *psColSpec     = "_COLSPC";
  char          sColValues[AK_COLVAL_LEN];
  char          sColName[CARD_SIZE];
  char          sKeyLine[AK_COLVAL_LEN];
  char          sDisplayFormat[AK_COLVAL_LEN];
  char          sDescription[AK_COLVAL_LEN];
  char          sUnit[CARD_SIZE];

  long   width        = 0;
  long   length       = 0;
  long   arraysize    = 0;
  long   ncols        = 0;
  long   ii           = 0;
  long   jj           = 0;
  long   uu           = 0;
  long   nUnknowns    = 0;      /* number of cols with unknown type      */
  long   colUnknowns[99];       /* list of col numbers with unknown type */


  keylist = afk_block_keylist( pakBlock );

  if ( keylist )
  {
    pKey = keylist->data[keylist->n-1];
    psColumnNames = dmu_key_value_c( pKey ); 
  }

  if ( psColumnNames )
  {
    status = afk_simple_read_cols( pakBlock, pakStyle, psColumnNames );

    if ( status != dmSUCCESS )
      return status;
  }

  //   Need to get the column types from the header information

  pakDataset = afk_block_get_kds( pakBlock );
  pakFile    = afk_dataset_fptr( pakDataset );
  ncols      = afk_block_ncols( pakBlock );
  psSeparators = afk_style_get_sep(pakStyle);

  // make a list to keep unit strings so that the
  // keylist column keys do not have to be parsed twice
  // function afk_tsv_attach_units() will release memory

  ppUnits = (char**)calloc( ncols, sizeof(char**));
  *ppUnitStrings = ppUnits;

  ppComments = (char**)calloc( ncols, sizeof(char**));
  *ppDescriptions = ppComments;

  for (ii = 0; ii < ncols; ii++)
  {
    ppUnits[ii]= NULL;
    ppComments[ii] = NULL;
  }

  if ( keylist && psColumnNames )
  {
    /* MCD Mod. 20100917:                            */
    /* Loop over all ColSpec keys in keylist..       */
    /* Allow for there to not be same number         */

    for ( ii = 0; ii < keylist->n; ii++)
    {
      pKey = keylist->data[ii];

      /* Check if this is a _COLSPC key */
      if ( cs_eq_siz( pKey->ucname, psColSpec, strlen(psColSpec) ))
      {
	dmu_key_read_value_c( pKey, AK_COLVAL_LEN, sKeyLine);
	
        // get column information from key
        afk_tsv_column_key_parse( psSeparators, sKeyLine, sColName,
                                  &dataType, &width, &length, &arraysize,
                                  sDisplayFormat, sUnit, sDescription );

        // match this info with the column
        // store in column descriptor

        for (jj = 1; jj <= ncols; jj++)
	{
	  pakCol =  afk_block_get_col( pakBlock, jj);

          if ( cs_eq(pakCol->name, sColName) )
	  {
            pakCol->datatype = dataType;      // data type
            pakCol->fwidth   = width;         // width or string length
            pakCol->length   = arraysize;     // array size
            pakCol->width    = length;
            pakCol->disp     = cs_dup(sDisplayFormat);

	    // Units and descriptions are not a part of the kernel
            // descriptor. Keep track of them and add them when
            // dmDescriptor is made
  
            if (sUnit[0] != '\0')
              ppUnits[jj-1] = cs_dup(sUnit);

            ppComments[jj-1] = cs_dup(sDescription);

            break;
          }
	} // for column loop (jj)
      } // end if colspec line
    } // end keylist loop (ii)
  }

  // Get list of any columns that are not assigned a data type
  for ( ii = 1; ii <= ncols; ii++)
  {
    pakCol = afk_block_get_col(pakBlock, ii);

    if (pakCol->datatype == DT_UNK )
      colUnknowns[nUnknowns++] = ii;
  }

  
  // Check for any loose ends

  if ( nUnknowns )
  {
    afk_block_set_row( pakBlock, 1 );

    // If there are any columns that do not have a type assigned
    // use data to determine that
  
    /* loop over all rows to get column data type */

    while ( (psRow = afk_file_getline_ptr( pakFile )) )
    {
      /* uu indexes unknown column numbers in colUnknowns */
      uu = 0;

      for ( ii = 1; (ii <= ncols) && psRow && (uu < nUnknowns); ii++ )
      {
        /* is this a column of interest? */
        if ( ii == colUnknowns[uu] )
        {
          /* yes, this is an unknown data type column                    */
          /* increment index now so that next time it will look for next */
          /* unknown column                                              */
          ++uu;

          pakCol = afk_block_get_col( pakBlock, ii );
          psRow  = afk_gen_read_next_col( pakStyle, psRow, sColValues, AK_COLVAL_LEN );
 
          afk_gen_determine_col_datatype( pakStyle, pakCol, sColValues, &dataType);

          if ( psRow )
          {
            /* skip over array elements */

            for ( jj = 1; jj < pakCol->length; jj++ )
            {
              psRow = afk_gen_read_next_col( pakStyle, psRow, sColValues, AK_COLVAL_LEN );
              afk_gen_determine_col_datatype( pakStyle, pakCol, sColValues, &dataType);
            }
          }
        }
        else
         // Code Change  Bug 13511  March 6, 2013
  	 //   Need to call afk_gen_read_next_col() if
         //   not desired column in order to set ptr to
	 //   next column

	 // skip this column
         psRow  = afk_gen_read_next_col( pakStyle, psRow, sColValues, AK_COLVAL_LEN );

      } // end for loop
    } // end while loop
  } // if (nUnknowns )

  /* final rectification of column types */
  afk_gen_final_col_datatype( pakBlock, pakStyle );

  /* reset to start */
  afk_block_set_row( pakBlock, 1 );

 return status;
}


/*****************************************************************************/
/*                                                                           */
/*  Function:  afk_tsv_key_write()                                           */
/*                                                                           */
/*  Description:                                                             */
/*     Write key information                                                 */
/*     code based on afk_simple_write_key()                                  */
/*                                                                           */
/*  Input parameters:                                                        */
/*                    afBlock        *pakBlock                               */
/*                    long            lKeyPosition                           */
/*                    const char     *sKeyName                               */
/*                    dmDataType      dataType                               */
/*                    const void     *pValue                                 */
/*                    const char     *sUnit                                  */
/*                    const char     *sDescription                           */
/*                                                                           */
/*  Output parameters:  dmuKey* key                                          */
/*                                                                           */
/*  Called by:                                                               */
/*         afp_block_key_write()                                             */
/*                                                                           */
/*****************************************************************************/
dmuKey* afk_tsv_key_write( afBlock *pakBlock, long lKeyPosition,
                              const char *sKeyName,
                              dmDataType dataType, const void *pValue,
                              const char *sUnit, const char *sDescription )
{
  dmuKeyList  keyList  = pakBlock->keylist;
  long        firstKey = 0; /* first valid new key position, need to keep colnames line*/
  long        lastKey  = 0; /* last valid new key position, need to keep colnames line*/


  // default positions
  firstKey = 1;
  lastKey  = keyList->n;

                      
  // verify range
  if ( lKeyPosition < firstKey )
    lKeyPosition = firstKey;
  else
    if ( lKeyPosition > lastKey )
      lKeyPosition = lastKey;

  // write key
  return afk_raw_key_write( pakBlock, lKeyPosition, sKeyName, dataType, pValue, sUnit, sDescription );
}


/***************************************************************/
/*                                                             */
/* Function:   afk_tsv_read_header                             */
/*                                                             */
/* Description:                                                */
/*             We expect the file already to be pointing at    */
/*             the beginning of the block.                     */
/*                                                             */
/*  Input parameters:                                          */
/*                   afBlock  *pakBlock - block pointer        */
/*                   afStyle  *pakStyle - style pointer        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/*  Called by:                                                 */
/*         afp_block_read_header()                             */
/*                                                             */
/***************************************************************/
dmErrCode afk_tsv_read_header( afBlock *pakBlock, afStyle *pakStyle )
{
  GenericType tmp;
  void  *parsed_value = &tmp;

  dmuKeyList  keylist;
  dmDataType  dataType = DT_CHAR;
  dmBool      bMask   = dmTRUE;
  dmuKey     *pKey    = NULL;
  ftKeyType   keyType = FTK_COMMENT;
  dmErrCode   status  = dmSUCCESS;

  afBuffer  *pakBuffer  = NULL;
  afFile    *pakFile    = NULL;
  afDataset *pakDataset = NULL;

  const  char *pcLine     = NULL;
   char  *psSeparator      = NULL;
  char  *psDefaultSep     = "\t\r";
  char  *psColSpec        = "_COLSPC";
  char  *psColumn         = "COLUMN";

  char   cHeaderChar      = '\0';
  char   cDefaultHeadChar = '#';
  char sKeyName[AK_KEYNAME_LEN];
  char sComment[AK_KEYCOMMENT_LEN];

  long pos     = 0;
  long histlen = 7;
  long commlen = 7;
  long colNum  = 0;
  long columnlen = 6;
  int  ii      = 0;


  pakDataset = afk_block_get_kds( pakBlock );
  pakBuffer  = afk_block_get_buffer( pakBlock );
  pakFile    = afk_dataset_fptr( pakDataset );
  keylist    = dmu_keylist_alloc();

  /* might need to delete a previous keylist? */
  pakBlock->keylist = keylist;

  // Get header char and separators in case user changed
  // from default values
  // If a value is null, use default value

  cHeaderChar = pakStyle->header_char;

  if ( cHeaderChar == '\0')
    cHeaderChar = cDefaultHeadChar;

  if ( pakStyle->sep != NULL )
    psSeparator = cs_dup( pakStyle->sep );
  else
    psSeparator = cs_dup( psDefaultSep );


  // read the first line
  pos    = afk_file_tell( pakFile );
  pcLine = afk_file_getline_ptr( pakFile );

  // skip blank lines
  while ( pcLine && *pcLine == '\0' )
    pcLine = afk_file_getline_ptr( pakFile );


  // process as a comment any line that starts with a header character

  while ( pcLine && *pcLine == cHeaderChar )
  {
    bMask = dmFALSE;

    afk_buffer_add_keyline( pakBuffer, pcLine );

    // store the comment line as the key's value
    afk_gen_read_comment( pakStyle, pcLine, sComment, AK_KEYCOMMENT_LEN );

    /*  MCD NOTE: This handling is from the DTF flavor.  Simple/Raw, do */
    /*  a similar check for HISTORY/COMMENT, but do not offset the string */
    /*  to remove the tag. (which I think will need to change there too).  */

    if ( cs_eq_siz( sComment, dmHISTORY, histlen ))
    {
      strcpy( sKeyName, dmHISTORY );
      *(char**)parsed_value = cs_dup( sComment + histlen + 1 );
    }
    else if ( cs_eq_siz( sComment, dmCOMMENT, commlen ))
    {
      strcpy( sKeyName, dmCOMMENT );
      *(char**)parsed_value = cs_dup( sComment + commlen + 1 );
    }
    else if (cs_uceq_siz( sComment, psColumn, columnlen))
    {
      // If column info line, make a special key with  keyname
      // composed of string _COLSPEC and the item number

      sprintf(sKeyName, "%s%d", psColSpec, (int)(++colNum));
      *(char**)parsed_value = cs_dup( sComment );

      // store column comment line as string type
      keyType = FTK_STRING;
    }
    else
    {
      strcpy ( sKeyName, dmCOMMENT );
      *(char**)parsed_value = cs_dup( sComment );
    }

    sComment[0] = '\0';
    bMask = dmTRUE;

    // add a key to the list
    pKey = dmu_key_alloc();
 
    dmu_key_set( pKey, (dmuKeyType)keyType, dataType, sKeyName, parsed_value, NULL,
	               sComment, bMask );
    dmu_key_insert( keylist, pKey, 0 );

    // clean up
    if ( dataType == DT_CHAR )
      free( *(char**)parsed_value );

    // read the next comment line
    pos    = afk_file_tell( pakFile );
    pcLine = afk_file_getline_ptr( pakFile );

    // skip blank lines
    while( pcLine && *pcLine == '\0' )
    {
      pos    = afk_file_tell( pakFile );
      pcLine = afk_file_getline_ptr( pakFile );
    }

    // reset for next loop in case it was changed
    keyType = FTK_COMMENT;
  }

  // Finished reading comment lines
  // Current line contains column names

  if ( !pcLine )
    return status; /* no data in file */


  /* process the line as a comment  */
  afk_gen_read_comment( pakStyle, pcLine, sComment, AK_KEYCOMMENT_LEN );

  /* assign keyname with line content as value */
  /* MCD: Changed to assign an actual key name for these colspec lines */
  /*      masked as private..                                          */

  sprintf( sKeyName,"_COLSPEC%1d", ii );
  *(char**)parsed_value = cs_dup( sComment );

  /* reset comment variable for next line. */ 
  sComment[0] = '\0';

  /* Add the key to dm keylist */
  keyType = FTK_STRING;
  bMask   = dmTRUE;

  pKey = dmu_key_alloc();
  dmu_key_set( pKey, (dmuKeyType)keyType, dataType, sKeyName, parsed_value, NULL,
               sComment, bMask );

  dmu_key_insert( keylist, pKey, 0 );

  // clean up
  if ( dataType == DT_CHAR )
    free( *(char**)parsed_value );

  // read the next line
  pos    = afk_file_tell( pakFile );
  pcLine = afk_file_getline_ptr( pakFile );

  // skip blank lines
  while( pcLine && *pcLine == '\0' )
  {
    pos = afk_file_tell( pakFile );
    pcLine = afk_file_getline_ptr( pakFile );
  }

  // set data start position, already held in var pos
  afk_buffer_set_datastart( pakBuffer, pos );

  // housekeeping
  free (psSeparator );

  return status;

}


/***************************************************************/
/*                                                             */
/*  Function:  afk_tsv_write_comment()                         */
/*                                                             */
/* Description:                                                */
/*    puts comments into header buffer                         */
/*    code based on  afk_simple_write_comment()                */
/*                                                             */
/*  Input parameters:                                          */
/*                    afBlock        *pakBlock                 */
/*                    long            pos                      */
/*                    const char     *tag                      */
/*                    const char     *comment                  */
/*                                                             */
/*  Output parameters:  none                                   */
/*                                                             */
/*  Returns:                                                   */
/*           dmErrCode  status                                 */
/*                                                             */
/*  Called by:                                                 */
/*         afp_block_write_comment()                           */
/*                                                             */
/***************************************************************/
dmErrCode afk_tsv_write_comment( afBlock *pakBlock,
                                 long pos, const char *psTag,
                                 const char *psComment )
{
  char  sBuf[AK_COMMENTLINE_LEN];
  char  sPrefix[4];
  char *psSeparator = NULL;
  char *psDefaultSep = "\t\r";
  char  cHeaderChar = '\0';
  char  cDefaultHeadChar = '#';
  char  cNull  = '\0';

  dmuKeyList keylist;
  dmErrCode  status  = dmSUCCESS;
  dmuKey    *pKey    = NULL;
  dmBool     bMask   = dmTRUE;

  afBuffer  *pakBuffer = NULL;
  afStyle   *pakStyle  = NULL;
  ftKeyType  fkeyType  = FTK_UNKNOWN;

  long       keyno  = 0;
  long       keyOffset = -1;
 

  keylist   = pakBlock->keylist;
  pakBuffer = afk_block_get_buffer( pakBlock );
  fkeyType  = FTK_COMMENT;

  // Get header char and separator
  // if either value is null, use default

  pakStyle = pakBlock->ads->astyle;
  cHeaderChar = afk_style_get_header_char( pakStyle );

  if ( cHeaderChar == '\0' )
    cHeaderChar = cDefaultHeadChar;

  if ( pakStyle->sep != NULL )
    psSeparator = cs_dup( pakStyle->sep );
  else
    psSeparator = cs_dup( psDefaultSep );

  // HISTORY goes after the comments and before the column name entries
  // adjust position accordingly

  if ( ! strcmp(psTag,"HISTORY" ))
    keyno = keylist->n;
  else
  {
    for ( keyno = pos+1; keyno <= keylist->n; keyno++ ) /* was keyno=pos+1 */
    {
      pKey = keylist->data[keyno-1];
 
      if ( pKey->keytype != FTK_COMMENT )
        break;
    }

    // COMMENTS go before name entry, adjust accordingly
    keyno += keyOffset;
  }


  pKey = dmu_key_alloc();
  pKey->type = FTK_KEY;
  dmu_key_insert( keylist, pKey, keyno );
  dmu_key_set( pKey, fkeyType, DT_CHAR, psTag, &psComment, NULL, NULL, bMask );

  sPrefix[0] = cHeaderChar;
  sPrefix[1] = cNull;
  sPrefix[2] = cNull;

  if ( strcmp( psTag, AK_SPECIFICATION) == 0)
    sprintf (sBuf, "%s", psComment);
  else
  {
    if ( strcmp( psTag, dmCOMMENT ) == 0 )
      sprintf( sBuf, "%s%s\n", sPrefix, psComment );
    else
      sprintf( sBuf, "%s%s%c%s\n", sPrefix, psTag, psSeparator[0], psComment );
  }

  status = afk_buffer_insert_keyline( pakBuffer, sBuf, keyno );

  if ( status != dmSUCCESS )
  {
    dmu_key_delete( keylist, pKey );
    pKey = NULL;
  }

  // housekeeping
  free (psSeparator);

  return status;
}


