/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   aflib.c                                                     */
/*                                                               */
/*   Functions that create and work with "hint" structure used   */
/*   with ASCII kernel file type                                 */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"



/* Global for back compatibility */
static afHintData afHints = { 
  AK_FLAV_UNK, 0,
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

/* Suppress some compiler warnings */
afHintData* afp_hint_data( dmBlock* block );
afHintData* afp_hint_alloc( void );



/***************************************************************/
/*                                                             */
/* Function:   afp_hint_alloc                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*      afHintData  *hints - pointer to hint structure         */
/*                                                             */
/***************************************************************/
afHintData* afp_hint_alloc( void )
{
  afHintData *pHints = NULL;

  pHints = calloc( 1, sizeof( struct afHintData_s ));

  // initialize hint structure to nulls
  pHints->sep    = NULL;
  pHints->quotes = NULL;

  // Code Change  Bug 10848 July 2008
  pHints->nullstr = NULL;

  afp_clear_kernel_hints( pHints );

  return pHints;
}

/***************************************************************/
/*                                                             */
/* Function: afk_hint_free                                     */
/*                                                             */
/* Description:                                                */
/*    Free memory allocated for hint structure in Ascii Block  */
/*                                                             */
/*  Input parameters:                                          */
/*      afBlock ablock   - Ascii block pointer                 */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*                                                             */
/***************************************************************/
void afk_hint_free( afBlock* ablock )
{
  afHintData *pHints  = NULL;

  if ( ablock->hint )
  {
    pHints = ablock->hint;

    // Elements sep and quotes are pointers to strings
    // Free strings first

    if ( pHints->sep)
      free ( pHints->sep );

    if ( pHints->quotes)
      free ( pHints->quotes );

    // Code Change  Bug 10848 July 2008
    if ( pHints->nullstr)
      free ( pHints->nullstr );

    free( ablock->hint );

    ablock->hint = NULL;
  }
}

/***************************************************************/
/*                                                             */
/* Function:  afp_hint_data                                    */
/*                                                             */
/* Description:                                                */
/*    Return a Hint structure for the given block              */
/*                                                             */
/*  Input parameters:                                          */
/*    dmBlock block      -  DM Block pointer                   */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*    afHintData hints   -  Hint structure                     */
/*                                                             */
/***************************************************************/
afHintData* afp_hint_data( dmBlock* block )
{
  afHintData *hints  = NULL;
  afBlock*    ablock = NULL;

  if ( block )
  {
    /* Get Ascii block struct from DM block */
    ablock = (afBlock*)dmi_block_get_kb( block );

    if ( !ablock )
      return  NULL;

    /* Create a new hint struct if needed */
    if ( !ablock->hint )
      ablock->hint = afp_hint_alloc();

    /* Set return pointer */
    hints = ablock->hint;
  }
  else
  {
    /* Back compat for global variable use */
    hints = &afHints;
  }

  return hints;
}

/***************************************************************/
/*                                                             */
/* Function:  afp_set_hint_list                                */
/*                                                             */
/* Description:                                                */
/*    Parse option line and set hint fields of block           */
/*                                                             */
/*  Input parameters:                                          */
/*      dmBlock* block     - DM Block pointer                  */
/*         char* option    - Option line                       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*      dmSUCCESS if OK                                        */
/*      dmFAILURE if problem was encountered                   */
/*                                                             */
/***************************************************************/
dmErrCode afp_set_hint_list( dmBlock* block, const char* option )
{
  char   namebuf[CARD_SIZE] = "";
  char  *name  = NULL;
  char  *value = NULL;
  char **list  = NULL;

  long nitems   = 0;
  long ii       = 0;
  long iDoFirst = -1;

  int  bDoFirstFlag = 0;
  dmErrCode status  = dmSUCCESS;


  if ( cs_is_blank( option )) 
    return status;

  name = namebuf;
  list = cs_parse_comma_list( option, &nitems );

  if ( !nitems )
    return status;


  // Code change   February 28, 2006
  // If a kernel option is specified, it needs to 
  // be processed first.  Otherwise, style parameters
  // are likely to be lost along the processing line.
  // First, determine if a kernel option has been specified

  for ( ii = 0; ii < nitems; ii++ )
  {
    if (cs_uceq_siz( list[ii], "KERNEL", 6))
    {
      iDoFirst = ii;
      bDoFirstFlag = 1;
      break;
    }
  }


  /* Code change  Bug 9053  March 15, 2006                                */
  /*  Process option list                                                 */
  /*  Do the kernel option first, if one present (bDoFirstFlag == dmTRUE) */
  /*  then proceed with list in order.                                    */
  /*  Otherwise, just process items in order in list                      */
  ii = -1;
  do
  {
    char* tmpstr = NULL;

    if (! bDoFirstFlag)
    {
      /* skip if alreaady done */
      if ( ++ii == iDoFirst )
        continue;
  
      tmpstr = list[ii];

    }
    else
    {
      /* process kernel option first */
      tmpstr = list[iDoFirst];
      bDoFirstFlag = 0;
    }
    /* Split token into name-value pair */
    value = dmu_parse_token( tmpstr, '=', &name, CARD_SIZE );
    
    // check for correctness
    if ( afp_set_hint( block, name, value ) != dmSUCCESS )
    {
      /* Set error message - cannot parse option */
      /* Set status to failure */
      sprintf( dm_errbuf, "Unrecognized option %s ignored.", tmpstr );

      dmu_err_set( DM_WARN_PARSE_OPTION, dm_errbuf );
      status = dmFAILURE;
    }
  }  while ( ii < (nitems-1) );

  // end code change Bug 9053


  ar_free_cp( list, nitems );

  return status;
}

/***************************************************************/
/*                                                             */
/* Function:  afp_set_hint                                     */
/*                                                             */
/* Description:                                                */
/*   Set field of Hint structure within the given block        */
/*                                                             */
/*  Input parameters:                                          */
/*      dmBlock* block     - DM Block pointer                  */
/*      char*    name      - name of field to set              */
/*      char*    value     - value for field                   */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*      dmSUCCESS if OK                                        */
/*      dmFAILURE if problem was encountered                   */
/*                                                             */
/***************************************************************/
dmErrCode afp_set_hint( dmBlock* block, const char* name, const char* value )
{
  dmErrCode   status = dmSUCCESS;
  afHintData *hints  = NULL;
  dmBool      bFixed = UT_FALSE;


  /* Get Hint structure from block */
  hints = afp_hint_data( block );

  if ( name && *name )
  {
    if ( !value )
    {
      /* No VALUE provided for field..  */

      if ( cs_uceq_siz( name, "DTF", 3 ))
      {
	hints->flav = AK_FLAV_DTF;
	hints->set_flavor = UT_TRUE;
	if ( cs_uceq_siz( name+3, "-FIXED", 6 ))
	{
	  hints->set_fixed = UT_TRUE;
	  hints->fixed = UT_TRUE;
	}
      }
      else if ( cs_uceq_siz( name, "SIMPLE", 6 ))
      {
	hints->flav = AK_FLAV_SIMPLE;
	hints->set_flavor = UT_TRUE;
	if ( cs_uceq_siz( name+6, "-FIXED", 6 ))
	{
	  hints->set_fixed = UT_TRUE;
	  hints->fixed = UT_TRUE;
	}
      }
      else if ( cs_uceq_siz( name, "RAW", 3 ))
      {
	hints->flav = AK_FLAV_RAW;
	hints->set_flavor = UT_TRUE;
	if ( cs_uceq_siz( name+3, "-FIXED", 6 ))
	{
	  hints->set_fixed = UT_TRUE;
	  hints->fixed = UT_TRUE;
	}
      }
      // Code Change  Bug 11905  August 21, 2009
      else if ( cs_uceq_siz( name, "CSCRDB", 6 ))
      {
	hints->flav = AK_FLAV_CSCRDB;
	hints->set_flavor = UT_TRUE;
	if ( cs_uceq_siz( name+6, "-FIXED", 6 ))
	{
	  hints->set_fixed = UT_TRUE;
	  hints->fixed = UT_TRUE;
	}
      }
      else if ( cs_uceq( name, "FITS" ))
      {
	hints->set_fits = UT_TRUE;
	hints->use_fits = UT_TRUE;
      }
      else if ( cs_uceq( name, "FIXED" ))
      {
	hints->set_fixed = UT_TRUE;
	hints->fixed = UT_TRUE;
      }
      else if ( cs_uceq( name, "WHITE" ))
      {
	/* Treat separators as white space */
	hints->set_multisep = UT_TRUE;
	hints->multisep = UT_TRUE;
      }
      else 
      {
	/* Dont recognize this option.. pass it on */

        // Code change  Bug 9053  March 7, 2006

	status = dmi_block_set_hint( block, name, value );

        //  end code change

      }
    }
    else if ( name && *name && value && *value )
    {
      /* Name-value pair provided */

      if ( cs_uceq( name, "KERNEL" ))
      {
	if ( afp_verify_and_set_flavor ( value, &hints->flav, &bFixed) )
	{
          hints->set_flavor = UT_TRUE;

	  if (bFixed)
	  {
            hints->set_fixed = UT_TRUE;
	    hints->fixed     = UT_TRUE;
          }
        }
	else
        {
	  sprintf( dm_errbuf, "Error: Unknown TEXT kernel %s", value );
	  dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
	  status = dmFAILURE;
	}
      }
      else if ( cs_uceq( name, "QUOTES" ))
      {
	if ( hints->set_quotes && hints->quotes && *hints->quotes != '\0')
	{
	  /* Field already set.. prep to override */
	  free( hints->quotes );
	}

	hints->set_quotes = UT_TRUE;
	hints->quotes     = cs_dup( value );
      }
      else if ( cs_uceq( name, "SEP" ))
      {
	/* Set Separator characters */
	if ( hints->set_sep && hints->sep && *hints->sep != '\0')
	{
	  /* Field already set.. prep to override */
	  free( hints->sep );
	}

	hints->set_sep = UT_TRUE;
	hints->sep     = cs_dup( value );

	/* Override default mulitsep value if a separator is specified */
	/* If the multi flag is already set, it was also specified     */
	if ( hints->set_multisep == UT_FALSE )
	{
	  hints->multisep = UT_FALSE;
	}
      }
      else if ( cs_uceq( name, "WHITE" ))
      {
	/* Treat separators as whitespace */
	hints->set_multisep = UT_TRUE;
	hints->multisep     = UT_TRUE;
      }
      else if ( cs_uceq( name, "HEAD" ))
      {
	if ( cs_uceq( value, "YES" ) )
	{
	  hints->has_header = UT_TRUE;
	}
	else
	{
	  hints->has_header = UT_FALSE;
	}
	hints->set_header = UT_TRUE;

      }
      else if ( cs_uceq( name, "COMMENT" ))
      {
	hints->set_header_char = UT_TRUE;
	hints->header_char     = value[0];
	hints->set_header      = UT_TRUE;
	hints->has_header      = UT_TRUE;
      }
      else if ( cs_uceq( name, "COLNAMES" ))
      {
	if ( cs_uceq( value, "FIRST" ))
	{
	  hints->set_colnames = UT_TRUE;
	  hints->colnames     = AK_STYLE_COLNAMES_FIRST;
	}
	else if ( cs_uceq( value, "LAST" ))
	{
	  hints->set_colnames = UT_TRUE;
	  hints->colnames     = AK_STYLE_COLNAMES_LAST;
	}
	else if ( cs_uceq( value, "NONE" ))
	{
	  hints->set_colnames = UT_TRUE;
	  hints->colnames     = AK_STYLE_COLNAMES_NONE;
	}
      }
     // Code change  Bug 9053  March 7, 2006
      else
        if ( cs_uceq (name, "SKIP"))
        {
          hints->set_skip = UT_TRUE;
          hints->skip = atoi(value);
        }
        // Code change  Bug   July, 2008
        else
          if ( cs_uceq (name, "NULLSTR"))
          {
	    if ( hints->set_nullstr && hints->nullstr && *hints->nullstr != '\0')
	    {
	      /* Field already set.. prep to override */
	      free( hints->nullstr );
	    }

            hints->set_nullstr = UT_TRUE;
            hints->nullstr     = cs_dup( value );
          }
          else
          {
	    /* Dont recognize this option.. pass it on */

 	    status = dmi_block_set_hint( block, name, value );

            //  end code change
          }
    }
  }
  else
  {
    dmu_err_set( DM_ERR_KERNEL, "Error: Option passed to kernel with no name" );
    status = dmFAILURE;
  }

  return status;
}

/***************************************************************/
/*                                                             */
/* Function:  afp_hint_kernel                                  */
/*                                                             */
/* Description:                                                */
/*   Return the kernel Flavor indicated in the Hint struct     */
/*   of the given block                                        */
/*                                                             */
/*  Input parameters:                                          */
/*      dmBlock* block     - DM Block pointer                  */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*     Kernel flavor                                           */
/*                                                             */
/***************************************************************/
afFlavour afp_hint_kernel( dmBlock* block )
{
  afHintData *hints = NULL;

  hints = afp_hint_data( block );

  if ( hints )
    return hints->flav;
  else
    return AK_FLAV_UNK;
}

/***************************************************************/
/*                                                             */
/* Function:  afp_set_style_hints                              */
/*                                                             */
/* Description:                                                */
/*    Transfer Hint information from block to the provided     */
/*    style structure                                          */
/*                                                             */
/*  Input parameters:                                          */
/*      dmBlock* block     - DM Block pointer                  */
/*                                                             */
/*  Output parameters:                                         */
/*      afStyle* astyle    - Ascii style struct                */
/*                                                             */
/*  Returns:                                                   */
/*                                                             */
/***************************************************************/
void afp_set_style_hints( dmBlock* block, afStyle* astyle )
{
  afHintData *hints = NULL;


  // If block is NULL, use static kernel hint structure
  if ( block )
  {
    hints = afp_hint_data( block );

    if ( !hints )
      return;
  }
  else
    hints = &afHints;


  // set style elements from hints

  if ( hints->set_flavor )
    astyle->flavor = hints->flav;

  if ( hints->set_header )
    astyle->has_header = hints->has_header;

  if ( hints->set_header_char )
    astyle->header_char = hints->header_char;

  if ( hints->set_fixed )
    astyle->fixed = hints->fixed;

  if ( hints->set_sep && hints->sep )
  {
    if ( astyle->sep && *astyle->sep != '\0')
      free( astyle->sep );

    astyle->sep = cs_dup( hints->sep );
  }

  if ( hints->set_multisep )
    astyle->multisep = hints->multisep;

  if ( hints->set_quotes )
  {
    if ( hints->quotes )
    {
      if ( astyle->quotes && *astyle->quotes != '\0')
        free ( astyle->quotes );

      astyle->quotes = cs_dup( hints->quotes );
    }
    else
    {
      astyle->quotes = NULL;
    }
  }

  // Code Change  Bug 10848 July 2008
  if ( hints->set_nullstr && hints->nullstr )
  {
    if ( astyle->nullstr && *astyle->nullstr != '\0')
      free( astyle->nullstr );

    astyle->nullstr = cs_dup( hints->nullstr );
  }

  if ( hints->set_fits )
    astyle->use_fits = hints->use_fits;

  if ( hints->set_colnames )
    astyle->colnames = hints->colnames;

  if ( hints->set_skip )
    astyle->skip = hints->skip;
}


/***************************************************************/
/*                                                             */
/* Function:  afp_dataset_set_hint_list                        */
/*                                                             */
/* Description:                                                */
/*    Parse option line and call apf_dataset_set_hint() to     */
/*    set style fields in style structure belonging to dataset */
/*                                                             */
/* Input parameters:                                           */
/*      dmBlock* block     - DM Block pointer                  */
/*         char* option    - Option line                       */
/*                                                             */
/* Output parameters:                                          */
/*                                                             */
/* Returns:                                                    */
/*      dmSUCCESS if OK                                        */
/*      dmFAILURE if problem was encountered                   */
/*                                                             */
/***************************************************************/
void afp_dataset_set_hint_list( dmDataset* ds, const char* option )
{
  char   namebuf[CARD_SIZE] = "";
  char  *name  = NULL;
  char  *value = NULL;
  char **list  = NULL;

  long nitems = 0;
  long ii     = 0;
  long iDoFirst = -1;
 
  if ( cs_is_blank( option ))
    return;

  name = namebuf;
  list = cs_parse_comma_list( option, &nitems );

  // Code change   February 28, 2006
  // If a kernel option is specified, it needs to 
  // be processed first.  Otherwise, style parameters
  // are likely to be lost along the processing line.
  // First, determine if a kernel option has been specified

  for ( ii = 0; ii < nitems; ii++ )
  {
    if (cs_uceq_siz( list[ii], "KERNEL", 6))
    {
      iDoFirst = ii;
      break;
    }
  }

  // if a kernel option was found, process it first
  if (iDoFirst >= 0 )
  {   
    value = dmu_parse_token( list[iDoFirst], '=', &name, CARD_SIZE );

    afp_dataset_set_hint( ds, name, value );
  }


  // process remaining options as usual
  for ( ii= 0; ii < nitems; ii++ )
  {
    if ( ii == iDoFirst )
      continue;
 
    value = dmu_parse_token( list[ii], '=', &name, CARD_SIZE );
    afp_dataset_set_hint( ds, name, value );
  }

  ar_free_cp( list, nitems );

}


/***************************************************************/
/*                                                             */
/* Function:  afp_dataset_set_hint                             */
/*                                                             */
/* Description:                                                */
/*   Sets style values for dataset with values passed in from  */
/*   apf_dataset_set_hint_list()                               */
/*                                                             */
/*  Input parameters:                                          */
/*      dmDataset* ds      - DM Dataset pointer                */
/*      char*    name      - name of field to set              */
/*      char*    value     - value for field                   */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*                                                             */
/***************************************************************/
void afp_dataset_set_hint( dmDataset* ds, const char* name, const char* value )
{
  dmErrCode  status   = dmSUCCESS;
  afDataset *ads      = NULL;
  afStyle   *astyle   = NULL;
  afFlavour  akFlavor = AK_FLAV_UNK;
  dmBool     bFixed   = UT_FALSE;

 
  /* Get Style structure from dataset */
  ads    = (afDataset*)dmi_dataset_get_kds( ds );
  astyle = ads->astyle;

  if ( name && *name )
  {
    if ( !value )
    {
      /* No VALUE provided for field..  */

      if ( cs_uceq_siz( name, "DTF", 3 ))
      {
	afk_style_set( astyle, AK_FLAV_DTF );

	if ( cs_uceq_siz( name+3, "-FIXED", 6 ))
	  astyle->fixed = UT_TRUE;
      }
      else
        if ( cs_uceq_siz( name, "SIMPLE", 6 ))
        {
	  afk_style_set( astyle, AK_FLAV_SIMPLE );

          if ( cs_uceq_siz( name+6, "-FIXED", 6 ))
	    astyle->fixed = UT_TRUE;
        }
        else
          if ( cs_uceq_siz( name, "RAW", 3 ))
          {
	    afk_style_set( astyle, AK_FLAV_RAW );

            if ( cs_uceq_siz( name+3, "-FIXED", 6 ))
	      astyle->fixed = UT_TRUE;
          }
          else
            if ( cs_uceq_siz( name, "CSCRDB", 6 ))
            {
	      afk_style_set( astyle, AK_FLAV_CSCRDB );

              if ( cs_uceq_siz( name+6, "-FIXED", 6 ))
	        astyle->fixed = UT_TRUE;
            }
            else
               if ( cs_uceq( name, "WHITE" ))
               {
	         /* Treat separators as whitespace */
                 astyle->multisep = UT_TRUE;
               }
               else
	          status = dmFAILURE;

    } // endif name and no value
    else
    {
      /* Name-value pair provided */

      if ( cs_uceq( name, "KERNEL" ))
      {
	if ( afp_verify_and_set_flavor ( value, &akFlavor, &bFixed) )
	{
	  /* Set style to defaults for this flavor */
	  afk_style_set( astyle, akFlavor );

	  if ( bFixed )
	  {
	    /* Set Fixed field */
	    afk_style_set_fixed( astyle, bFixed );

	    /* Modifying fixed field can effect flavor string.. update it. */
	    afk_style_set_flavor( astyle, akFlavor, NULL );
	  }
	}
	else
	{
	  status = dmFAILURE;
	}
      } // end if name = KERNEL
      else
        if ( cs_uceq( name, "SKIP" ))
        {
	  astyle->skip = atoi( value );
        }
        else if ( cs_uceq( name, "QUOTES" ))
        {
	  if ( astyle->quotes && *astyle->quotes != '\0' )
	  {
	    /* Field already set.. prep to override */
	    free( astyle->quotes );
	  }
	  astyle->quotes = cs_dup( value );

	} /* end if name = QUOTES */
	else if ( cs_uceq( name, "SEP" ))
	{
	  /* User may have entered "white" after sep= */
	  /* This is a format error, but easy enough  */
	  /* to catch and correct.  Do it here.       */
	  if ( cs_uceq( value, "WHITE"))
	  {
	    /* "white" entered as value  */
	    /* set multisep element true */
	    astyle->multisep = UT_TRUE;
	  }     
	  else
	  {     
	    /* Set Separator characters */
	    char* tptr = cs_dup( value );

	    /* strip quotes from value */
	    /* check for command line quotes.. not file style quotes.      */
	    if ( ( *tptr == '\'') || (*tptr == '\"') )
	    {
	      tptr[strlen(tptr)-1] = '\0';
	      memmove((void*)tptr,(void*)tptr+1, strlen(tptr));
	    }
	    if (*tptr != '\0')
	    {
	      if ( astyle->sep && *astyle->sep != '\0' )
	      {
		free( astyle->sep );
	      }
	      astyle->sep = tptr;
	    }

	    /* Declaration of SEP defines a single separator */
	    /* turn off multisep.                            */
	    /* NOTE: This requires white to be set after sep */
	    /*       which is sub-optimal                    */
	    astyle->multisep = UT_FALSE;
	    
	  }
	} /* end if name == SEP */
	else if ( cs_uceq( name, "COMMENT" ))
	{
	  astyle->header_char = value[0];
	  astyle->has_header = UT_TRUE;
	}
	else if ( cs_uceq( name, "COLNAMES" ))
	{
	  if ( cs_uceq( value, "FIRST"))
            astyle->colnames = AK_STYLE_COLNAMES_FIRST;
	  else
	    if ( cs_uceq( value, "LAST" ))
	      astyle->colnames = AK_STYLE_COLNAMES_LAST;
            // Code Change  Bug 11905  August 21, 2009
            else
#if 0
              if ( cs_uceq( value, "LM1" ))
	        astyle->colnames = AK_STYLE_COLNAMES_LM1;
	      else
                if ( cs_uceq( value, "LM2" ))
	          astyle->colnames  = AK_STYLE_COLNAMES_LM2;
            // End code change
	        else
#endif
	          astyle->colnames = AK_STYLE_COLNAMES_NONE;
	} /* end of COLNAMES  */
        // Code Change  Bug 10848 July 2008
        else if ( cs_uceq( name, "NULLSTR" ))
        {
	  if ( astyle->nullstr && *astyle->nullstr != '\0' )
	  {
	    /* Field already set.. prep to override */
	    free( astyle->nullstr );
	  }
	  astyle->nullstr = cs_dup( value );

	} /* end if name = NULL */

    } /* end if have name-value pair */
  } /* end if name provided */ 
}



/***************************************************************/
/*                                                             */
/* Function:  afp_set_kernel_hints                             */
/*                                                             */
/* Description:                                                */
/*    Transfer Style structure information into kernel hint    */
/*    structure or another hint structure whose pointer is     */
/*    an input parameter.                                      */
/*                                                             */
/*  Input parameters:                                          */
/*        afStyle     *astyle  - style pointer                 */
/*        afHintData  *pInHints - hint data structure          */
/*                                                             */
/*  Output parameters:                                         */
/*        none                                                 */
/*                                                             */
/*  Returns: nothing                                           */
/*                                                             */
/***************************************************************/
void afp_set_kernel_hints( afStyle *pStyle, afHintData *pInHints )
{
  afHintData  *pHints        = NULL;
  afStyle     *pStyleDefault = NULL;
  //  afFlavour    akFlavor      = pStyle->flavor;

  
  // if pInHints is NULL, use the static global hint structure
  // otherwise, use the input hint structure

  if (pInHints)
    pHints = pInHints;
  else
    pHints = &afHints;


  // load default style elements for input flavor
  pStyleDefault = afk_style_alloc();
  afk_style_set ( pStyleDefault, pStyle->flavor);


  // put style elements into hint structure if there
  // are any differences between style and hint structures

  if (pStyle->has_header != pStyleDefault->has_header)
  {
    pHints->has_header = pStyle->has_header;
    pHints->set_header = 1;
  }
  else
  {
    pHints->has_header = 0;
    pHints->set_header = 0;
  }


  if ( pStyle->header_char != pStyleDefault->header_char )
  {
    pHints->header_char = pStyle->header_char;   
    pHints->set_header_char = 1;
  }
  else
  {
    pHints->header_char = '\0';
    pHints->set_header_char = 0;
  }
 

  if ( pStyle->fixed != pStyleDefault->fixed)
  {
    pHints->fixed = pStyle->fixed;
    pHints->set_fixed = 1;
  }
  else
  {
    pHints->fixed = 0;
    pHints->set_fixed = 0;
  }


  if ( cs_eq(pStyle->sep, pStyleDefault->sep) )
  {
    pHints->sep = NULL;
    pHints->set_sep = 0;
  }
  else
  {
    if ( pHints->sep && *pHints->sep != '\0' )
      free( pHints->sep );

    pHints->sep = cs_dup( pStyle->sep );

    pHints->set_sep = 1;
  }


  if ( pStyle->multisep != pStyleDefault->multisep)
  {
    pHints->multisep = pStyle->multisep;
    pHints->set_multisep = 1;
  }
  else
  {
    pHints->multisep = 0;
    pHints->set_multisep = 0;
  }


  if ( cs_eq(pStyle->quotes, pStyleDefault->quotes) )
  {
    pHints->quotes = NULL;
    pHints->set_quotes = 0;
  }
  else
  {
    if ( pHints->quotes && *pHints->quotes != '\0' )
      free ( pStyle->quotes );

    pHints->quotes = cs_dup( pStyle->quotes );

    pHints->set_quotes = 1;
  }


  if ( pStyle->use_fits != pStyleDefault->use_fits)
  {
    pHints->use_fits = pStyle->use_fits;
    pHints->set_fits = 1;
  }
  else
  {
    pHints->use_fits = 0;
    pHints->set_fits = 0;
  }


  if ( pStyle->skip != pStyleDefault->skip)
  {
    pHints->set_skip = 1;
    pHints->skip = pStyle->skip;
  }
  else
  {
    pHints->set_skip = 0;
    pHints->skip = 0;
  }


  if ( pStyle->colnames != pStyleDefault->colnames )
  {
    pHints->colnames = pStyle->colnames;
    pHints->set_colnames = 1;
  }
  else
  {
    pHints->colnames = 0;
    pHints->set_colnames = 0;
  }


  // Code Change  Bug 10848 July 2008
  if ( cs_eq(pStyle->nullstr, pStyleDefault->nullstr) )
  {
    pHints->nullstr = NULL;
    pHints->set_nullstr = 0;
  }
  else
  {
    if ( pHints->nullstr && *pHints->nullstr != '\0' )
      free( pHints->nullstr );

    pHints->nullstr = cs_dup( pStyle->nullstr );

    pHints->set_nullstr = 1;
  }

  // default structure no longer necessary - free memory
  afk_style_free(pStyleDefault);
}



/***************************************************************/
/*                                                             */
/* Function:  afp_clear_hints                                  */
/*                                                             */
/* Description:                                                */
/*               Clear elements of hint structure.  If the     */
/*               input pointer is NULL, use the static kernel  */
/*               hint structure.                               */ 
/*                                                             */
/*  Input parameters:                                          */
/*                     afHintData  *pInHints - pointer to hint */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns: nothing                                           */
/*                                                             */
/***************************************************************/
void  afp_clear_kernel_hints(afHintData  *pInHints)
{
  afHintData  *pHints = NULL;


  // if input pointer is NULL, use static kernel hint structure

  if (pInHints)
    pHints = pInHints;
  else
    pHints = &afHints;


  // Return values to initial state

  pHints->flav = AK_FLAV_UNK;

  pHints->set_flavor = 0;

  pHints->set_header = 0;

  pHints->set_header_char = 0;

  pHints->set_fixed = 0;

  pHints->set_multisep = 0;

  pHints->set_quotes = 0;

  pHints->set_fits = 0;

  pHints->set_sep = 0;

  pHints->has_header = 0;

  pHints->set_colnames = 0;

  pHints->header_char = '\0';

  pHints->fixed = 0;

  pHints->multisep = 0;

  if ( pHints->quotes )
  {
    if ( *pHints->quotes != '\0' )
      free (pHints->quotes);

    pHints->quotes = NULL;
  } 

  pHints->use_fits = 0;

  if ( pHints->sep )
  {
    if ( *pHints->sep != '\0' )
      free( pHints->sep );

    pHints->sep = NULL;
  }

  pHints->colnames = 0;

  pHints->set_skip = 0;

  pHints->skip = 0;

  // Code Change  Bug 10848 July 2008
  if ( pHints->nullstr )
  {
    if ( *pHints->nullstr != '\0' )
      free( pHints->nullstr );

    pHints->nullstr = NULL;
  }
  pHints->set_nullstr = 0;
} 

/***************************************************************/
/*                                                             */
/* Function:  afp_set_colnames_hint                            */
/*                                                             */
/* Description:                                                */
/*           For CSCRDB to CSCRDB copy, set source file's      */
/*           header type here, so that it can be used in       */
/*           the dest copy. Otherwise, there is a problem      */
/*           with HISTORY lines being placed between the       */
/*           colnames line and the units name line or dash line*/
/*                                                             */
/*  Input parameters:                                          */
/*            long   lColNamesStyle - enum of colname style    */
/*            dmBool bSetColNames   - flag indicating hint     */
/*                                    is set                   */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns: nothing                                           */
/*                                                             */
/***************************************************************/
void  afp_set_colnames_hint(long lColNamesStyle, dmBool bSetColNames)
{
  afHintData  *pHints = &afHints;

  pHints->colnames = lColNamesStyle;
  pHints->set_colnames = bSetColNames;

}
