/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afraw.c                                                     */
/*                                                               */
/*   Functions that create and work with raw type files          */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"
#include "ryi.h"



/***************************************************************/
/*                                                             */
/* Function:   afp_raw_init_cols                               */
/*                                                             */
/* Description:                                                */
/*              Initialize column names                        */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock   *block                         */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode   Status                               */
/*                                                             */
/***************************************************************/
dmErrCode afp_raw_init_cols( dmBlock* block )
{
  dmErrCode    status = dmSUCCESS;

  dmDataset       *ds      = NULL;
  dmDescriptor   **rawcols = NULL;
  dmDescriptor    *dd      = NULL;
  dmDataType       type;

  afBlock         *ablock = NULL;
  afDataset       *ads    = NULL;
  afStyle         *astyle = NULL;
  afCol           *acol   = NULL;

  long nkcols = 0; 
  long ncols  = 0;
  long i      = 0;
  long colno  = 0;
  long length = 0;
  long naxes  = 1;
  long axes[] = { 1 };
 

  ablock = (afBlock*)dmi_block_get_kb( block );
  ds     = dmi_block_get_ds( block );
  ads    = (afDataset*)dmi_dataset_get_kds( ds );

  astyle = afk_dataset_get_style( ads );

  status = afk_raw_init_cols( ablock, astyle );
  nkcols = afk_block_ncols( ablock );
  ncols  = nkcols;

  if ( nkcols )
  {
    rawcols = (dmDescriptor**)malloc( nkcols * sizeof(dmDescriptor*) );
  }

  if ( rawcols )
  {
    for ( i = 0; i < nkcols; i++ )
    {
      colno = i+1;
      acol = afk_block_get_col( ablock, colno );
      type = acol->datatype;

      if ( type == DT_CHAR )
      {
        if ( acol->width )
          length = acol->width;
        else
          length = AK_COLVAL_LEN;
      }
      else
        length = 0;

      dd = ryi_column_alloc( block, colno, acol->name, type, length,
  			     NULL, NULL, 1, NULL, naxes, axes,
 			     0.0, 0.0, 1.0, acol );
      rawcols[i] = dd;
    }
  }

  dmi_block_set_raw_cols( block, rawcols, ncols );

  return status;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_raw_insert_col                              */
/*                                                             */
/* Description:                                                */
/*              Insert a new column                            */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock       *block                     */
/*                    dmDescriptor  *dd                        */
/*                    long           no                        */
/*                    void          *ikdd                      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_raw_insert_col( dmBlock* block, dmDescriptor* dd,
                         long no, void* ikdd )
{
  dmDescriptor    *cpt_dd = NULL;
  dmDescriptor    *col_dd = NULL;
  dmDescriptor    *us_dd  = NULL;
  afBlock         *ablock = NULL;
  afCol           *acol   = NULL;
  afVector        *avec   = NULL;

  long dim    = 0;
  long ncols  = 0;
  long kcolno = 0;
  long cptno  = 0;
  long moved  = 0;

  dmDataType   type   = DT_UNK;
  dmBool       append = dmSUCCESS;


  ablock = (afBlock*)dmi_block_get_kb( block );
  ncols  = dmi_block_get_ncols( block );

  append = ( no == ncols + 1 );
  dim    = dmi_descriptor_get_vector_dim( dd );

  if ( append )
    kcolno = ablock->ncols+1;
  else
  {
    col_dd = dmi_block_get_col( block, no );
    cpt_dd = dmi_descriptor_get_cpt( col_dd, 1 );
    acol   = (afCol*)dmi_descriptor_get_kdd( cpt_dd );
    kcolno = acol->no;
  }

  if ( dim > 1 )
    avec = afk_vector_alloc( dim );

  for ( cptno = 1; cptno <= dim; cptno++ )
  {
    cpt_dd = dmi_descriptor_get_cpt( dd, cptno );
    type   = dmi_descriptor_get_datatype( cpt_dd );
    acol   = (afCol*)dmi_descriptor_get_kdd( cpt_dd );

    if ( !acol )
    {
      if ( ikdd )
      {
        if ( dim == 1 )
          acol = afk_col_alloc_copy( (afCol*)ikdd );
        else
          if ( dim > 1 )
            acol = afk_col_alloc_copy( ((afVector*)ikdd)->cpts[cptno-1] );
      }
      else
      {
        acol = afk_col_alloc();
        acol->name     = cs_dup( dmi_descriptor_get_name( cpt_dd ));
        acol->datatype = dmi_descriptor_get_datatype( cpt_dd );
        acol->length   = dmi_descriptor_get_array_size( cpt_dd );

        // Code Change  Bug 10784  November 8, 2007
        acol->width    = dmi_descriptor_get_length( cpt_dd );

	/* Code change October 28, 2008 */
	/* Bit type stored by DM as array of bytes, but as   */
	/* single bit string in ascii                        */
	if (acol->datatype == DT_BIT )
	  acol->length = 0;
      }
      dmi_descriptor_set_kdd( cpt_dd, (void*)acol );
    }

    us_dd = dmi_descriptor_get_unscaled( cpt_dd );

    if ( us_dd )
      dmi_descriptor_set_kdd( us_dd, (void*)acol );
    if ( dim > 1 )
      avec->cpts[cptno-1] = acol;

    afk_block_insert_col( ablock, acol, kcolno );
    afk_buffer_insert_col( ablock->abuf, acol, kcolno, &moved );
    ++kcolno;
  }

  if ( dim > 1 )
  {
    dmi_descriptor_set_kdd( dd, avec );
    us_dd = dmi_descriptor_get_unscaled( dd );

    if ( us_dd )
      dmi_descriptor_set_kdd( us_dd, avec );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   afp_raw_create_block                            */
/*                                                             */
/* Description:                                                */
/*              Create a new block for file                    */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock   *block                         */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
 void afp_raw_create_block( dmBlock *block )
 {
   afBlock   *ablock = NULL;
   afDataset *ads    = NULL;
   afFile    *afptr  = NULL;
   afBuffer  *abuf   = NULL;
   afStyle   *astyle = NULL;
   long       spos   = 0;
   long       moved  = 0;

   int         flavorStringLength = 0;
   const char *pFlavorString      = NULL;
   char        flavorString[CARD_SIZE];
 

   ablock = (afBlock*)dmi_block_get_kb( block );
   ads    = ablock->ads;
   astyle = ads->astyle;
   abuf   = afk_block_get_buffer( ablock );
   afptr  = afk_dataset_fptr( ads );


   pFlavorString = afk_style_get_flavorString(astyle);
   if ( pFlavorString)
     strcpy( flavorString, pFlavorString );
   strcat (flavorString, "\n" );

   flavorStringLength = strlen(flavorString);

   afk_file_insert_data( afptr, flavorString );

   afk_buffer_flush( abuf, &moved );

   afk_file_seek( afptr, ads->pos+flavorStringLength, SEEK_SET );
 
   spos = afk_file_tell( afptr );
   ads->blocks_pos[0] = spos;
   afk_buffer_set_datastart( abuf, spos );
 
   return;
 }
