/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afsimple.c                                                  */
/*                                                               */
/*   Functions that create and work with SIMPLE kernel type files*/
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"
#include "ryi.h"

//static dmDescriptor** afp_simple_make_vectors( dmBlock* block, afCol*** acpts,
//                                               long ncols, long *dims,
//                                               dmBool bUseUnits );

static void afp_simple_write_col_header( dmBlock* block, dmDescriptor* dd,
                                         long no );



/***************************************************************/
/*                                                             */
/* Function:   afp_simple_init_cols                            */
/*                                                             */
/* Description:                                                */
/*              Initialize column names                        */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock   *block  - block pointer        */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode   Status                               */
/*                                                             */
/***************************************************************/
dmErrCode afp_simple_init_cols( dmBlock* block )
{
  dmErrCode       status  = dmSUCCESS;
  dmDataset      *ds      = NULL;
  dmDescriptor  **rawcols = NULL;
  dmuKeyList      KeyList;
  dmBool          doappend  = UT_FALSE;  /* Flag to append vector component */
  dmBool          bUseUnits = dmFALSE;

  afBlock   *ablock = NULL;
  afDataset *ads    = NULL;
  afStyle   *astyle = NULL;
  afCol     *acol   = NULL;
  afCol   ***acpts  = NULL;

  char  colname[CARD_SIZE];
  char *nend   = NULL; 

  long  nkcols  = 0;
  long *maxcpts = NULL;   /* number of cpts in each cpts element */
  long *dims    = 0;      /* pointer to number of components for each vector */
  long  ncols   = 0;
  long  colincr = 10;     /* amount to increment maxcols when it overflows */
  long  cptincr = 5;      /* amount to increment maxcpts when it overflows */
  long  maxcols = 10;     /* number of cols in cpts and ncpts */
  long  newVal  = 0;
  long  colndx  = 0;
  long  cptno   = 0;
  long  ii      = 0;
  long  jj      = 0;


  ablock = (afBlock*)dmi_block_get_kb( block );

  // Code Change  Bug 9342  1/5/2006
  // Get a pointer to the key list
  // If the poitner to the data section is NULL
  // call afp_raw_init_cols() because there
  // are no column names stored

  KeyList = afk_block_keylist (ablock);

  if (!KeyList  || KeyList->data == NULL)
  {
    status = afp_raw_init_cols (block);
    return status;
  }

  // end code change for Bug 9342



  ds = dmi_block_get_ds( block );
  ads = (afDataset*)dmi_dataset_get_kds( ds );

  astyle = afk_dataset_get_style( ads );

  status = afk_simple_init_cols( ablock, astyle );
  nkcols = afk_block_ncols( ablock );
 

  dims = (long*)calloc( maxcols, sizeof(long) );
  maxcpts = (long*)calloc( maxcols, sizeof(long) );
  acpts = (afCol***)calloc( nkcols, sizeof(afCol**) );

  /* first, loop through and put the afCols into a list of vector component lists */
  for( ii = 0; ii < nkcols; ii++ )
  {
    acol = afk_block_get_col( ablock, ii+1 );

    doappend = UT_FALSE;
    colndx = 0;

    /* if this is a vector component, it will be named 'vec(cpt)' */
    nend = strchr( acol->name, '(' );

    if ( nend )
    {
      /* copy vector name + '(' char */
      cs_copy_siz( acol->name, colname, nend - acol->name+1 );

      for ( jj = 0; jj < ncols; jj++ )
      {
        if ( acpts[jj] && acpts[jj][0] && 
	   ( cs_eq_siz( colname, acpts[jj][0]->name, strlen(colname) )))
        {
          /* this vector is already in the list, mark for append */
	  doappend = UT_TRUE;
          colndx = jj;
          break;
        }
      }
    }

    if ( doappend )
    {
      /* append this vector component to existing vector */
      dims[colndx]++;

      if ( dims[colndx] > maxcpts[colndx] )
      {
        maxcpts[colndx] += cptincr;
        acpts[colndx] = (afCol**)realloc( acpts[colndx], sizeof(afCol*) * maxcpts[colndx] );
      }

      cptno = dims[colndx]-1;
      acpts[colndx][cptno] = acol;   

    }
    else
    {
       /* create a new column */
       colndx = ncols;
       ++ncols;

       if ( ncols > maxcols )
       {
         newVal = maxcols + colincr;
         acpts = (afCol***)realloc( acpts, sizeof(afCol**) * newVal );
         maxcpts = (long*)realloc( maxcpts, sizeof(long) * newVal );
         dims = (long*)realloc( dims, sizeof(long) * newVal );
         memset( &acpts[maxcols], 0, colincr*sizeof(afCol**) );
         memset( &maxcpts[maxcols], 0, colincr*sizeof(long) );
         memset( &dims[maxcols], 0, colincr*sizeof(long) );
         maxcols = newVal;
       }

       dims[colndx]++;

       if ( dims[colndx] > maxcpts[colndx] )
       {
         newVal = maxcpts[colndx] + cptincr;
         acpts[colndx] = (afCol**)realloc( acpts[colndx], 
				      sizeof(afCol*) * newVal );
         memset( &acpts[colndx][maxcpts[colndx]], 0, cptincr*sizeof(afCol*) );
         maxcpts[colndx] = newVal;
       }  
       acpts[colndx][0] = acol;
    }
  }

  /* use our list of lists to create our final array of dmDescriptors */
  rawcols = afp_simple_make_vectors( block, acpts, ncols, dims, bUseUnits );

  for ( ii = 0; ii < ncols; ii++ )
  {
   free( acpts[ii] );
  }

  free( acpts );
  free( dims );
  free( maxcpts );
  dmi_block_set_raw_cols( block, rawcols, ncols );

  return status;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_simple_make_vectors                         */
/*                                                             */
/* Description:                                                */
/*              Initialize column names                        */
/*                                                             */
/*   CSCRDB processing uses this function also                 */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock   *block  - block pointer        */
/*                    afCol    **acpts                         */
/*                    long       ncols                         */
/*                    long      *dims                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor  **Descriptor                       */
/*                                                             */
/***************************************************************/
dmDescriptor** afp_simple_make_vectors( dmBlock* block, afCol*** acpts,
                                        long ncols, long* dims, dmBool bUseUnits)
{
  dmDescriptor **rawcols = NULL;
  dmDescriptor **cpts    = NULL;
  dmDescriptor  *dd      = NULL;
  afCol         *acol    = NULL;
  afVector      *avec    = NULL;

  long i       = 0;
  long j       = 0;
  long length  = 0;
  long naxes   = 1; /* currently we only deal with 1-dim arrays in SIMPLE */
  long axes[1]; 
  long colno   = 0;
  char* nend   = NULL;
  char* ptr    = NULL;
  char cptname[CARD_SIZE];
  char colname[CARD_SIZE];

  if ( ncols )
  {
    rawcols = (dmDescriptor**)malloc( sizeof(dmDescriptor*) * ncols );
  }

  /* loop through my vector component lists to construct my DM columns */
  if ( rawcols )
  {
    for ( i = 0; i < ncols; i++ )
    {
      colno = i+1;
      axes[0] = 1;
      naxes = 1; /* currently, SIMPLE is restricted to 1-dim arrays */
      length = 0;

      if ( dims[i] > 1 ) /* this is a vector column */
      {
        avec = afk_vector_alloc( dims[i] );
        cpts = (dmDescriptor**)malloc( sizeof(dmDescriptor*) * dims[i] );
        acol = acpts[i][0];
        nend = strchr( acol->name, '(' );
        cs_copy_siz( acol->name, colname, nend - acol->name );

       for ( j = 0; j < dims[i]; j++ )
       {
         acol = acpts[i][j];
         avec->cpts[j] = acol;
         nend = strchr( acol->name, '(' );
         ptr = nend+1;
         nend = strpbrk( ptr, " ,)" );

         if ( nend )
           cs_copy_siz( ptr, cptname, nend - ptr );
         else
           cs_copy( ptr, cptname );

         if ( acol->datatype == DT_CHAR )
         {
           if ( acol->width > 0 )
             length = acol->width;
           else
             length = AK_COLVAL_LEN;

           // Code Change  Bug 10784  November 8, 2007
           // Be sure to set column width in acol structure
           // String length info gets into descriptor
           // but not to the acol structure.
           acol->width = length;

         }
	 else
           // Code Change  Bug 11791 February 25, 2010
           //   Make a note of whether the boolean column uses a single 
           //   character (T or F) or the whole word (TRUE or FALSE) by 
           //   using the value stored in acol->width. This is for future
           //   use to facilitate exact copying of this type of column
           //   from one ascii flavor to another.

           if ( acol->datatype == DT_LOGICAL && acol->width > 0 )
             length = acol->width;    
           else
             length = 0;


         // Code change April 27, 2006
         // corrects parse error for format <name>[##](X,Y)

         if (acol->length > 1)
           axes[0] = acol->length;

         // end change


         dd = ryi_column_alloc( block, colno, cptname, acol->datatype, length,
		       NULL, NULL, 1, NULL, naxes, axes, 0.0, 0.0, 1.0, acol );
         cpts[j] = dd;
       }

       dd = ryi_column_alloc( block, colno, colname, DT_UNK, 0,
	 		      NULL, NULL, dims[i], cpts, 0, NULL, 
			      0.0, 0.0, 0.0, avec );
      }
      else
        if ( dims[i] == 1 ) /* this is a scalar column */
        {
          acol = acpts[i][0];
          strcpy( colname, acol->name );

          if ( acol->length > 0 )
          {
            axes[0] = acol->length;
          }

          if ( acol->datatype == DT_CHAR )
          {
            if ( acol->width > 0 )
              length = acol->width;
            else
              length = AK_COLVAL_LEN;
          }
	  else
            // Code Change  Bug 11791 February 25, 2010
            //   Make a note of whether the boolean column uses a single 
            //   character (T or F) or the whole word (TRUE or FALSE) by 
            //   using the value stored in acol->width. This is for future
            //   use to facilitate exact copying of this type of column
            //   from one ascii flavor to another.

            if ( acol->datatype == DT_LOGICAL && acol->width > 0 )
              length = acol->width;    

          dd = ryi_column_alloc( block, colno, colname, acol->datatype, length,
			         NULL, NULL, 1, NULL, naxes, axes, 0.0, 0.0, 1.0,
			         acol );
        }

      rawcols[i] = dd;
    }

    // Code Change  Bug 11905  August 21, 2009
    //  If CSCRDB file header with unit names, get them now
    if (bUseUnits)
      afk_cscrdb_attach_units(block, rawcols, ncols);
  }
  return rawcols;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_simple_make_vectors2                        */
/*                                                             */
/* Description:                                                */
/*              Function Currently Unused                      */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock   *block  - block pointer        */
/*                    afCol    **acpts                         */
/*                    long       ncols                         */
/*                    long      *dims                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor  **Descriptor                       */
/*                                                             */
/***************************************************************/
#if 0
dmDescriptor** afp_simple_make_vectors2( dmBlock* block,
                                         afCol*** acpts, long ncols,
                                         long *dims )
{
  dmDescriptor  *us_dd   = NULL;
  dmDescriptor  *dd      = NULL;
  dmDescriptor **cpts    = NULL;
  dmDescriptor **rawcols = NULL;
  dmDataType     type    = DT_UNK;

  char  colname[CARD_SIZE];
  char  rawname[CARD_SIZE];
  char  cptname[CARD_SIZE];
  char *ptr  = NULL;
  char *nend = NULL;
  long  i = 0;
  long  j = 0;
  long  colno = 0;
  long  naxes  = 0;

  long  *axes = NULL;
  afCol *acol = NULL;

  if ( ncols )
  {
    rawcols = (dmDescriptor**)malloc( sizeof(dmDescriptor*) * ncols );
  }

  /* loop through my vector component lists to construct my DM columns */
  if ( rawcols )
  {
    for ( i = 0; i < ncols; i++ )
    {
      colno = i+1;

      if ( dims[i] > 1 ) /* this is a vector column */
      {
        cpts = (dmDescriptor**)malloc( sizeof(dmDescriptor*) * dims[i] );
        acol = acpts[i][0];
        nend = strchr( acol->name, '(' );
        cs_copy_siz( acol->name, colname, nend - acol->name );

        for ( j = 0; j < dims[i]; j++ )
        {
          acol = acpts[i][j];
          nend = strchr( acol->name, '(' );
          ptr = nend+1;
          nend = strpbrk( ptr, " ,)" );

          if ( nend )
            cs_copy_siz( ptr, cptname, nend - ptr );
          else
             cs_copy( ptr, cptname );

          dmi_column_make_raw_name( cptname, colno, rawname );
          us_dd = dmi_descriptor_alloc( block, dmCOLUMN, colno, rawname,
				        acol->datatype, NULL, NULL, 1, dmVALUE, 0,
				        NULL, DM_FLAG_RAW | DM_FLAG_LOGICAL );
          type = dmi_descriptor_get_datatype( us_dd );

          if ( type == DT_CHAR )
          {
            if ( acol->width )
              dmi_descriptor_set_length( us_dd, acol->width );
            else
              dmi_descriptor_set_length( us_dd, CARD_SIZE );
          }
          dmi_descriptor_set_kdd( us_dd, acol );
          dd =  dmi_create_linear_wcs( block, us_dd, cptname, acol->datatype, 
				       0.0, 0.0, 1.0 );
          type = dmi_descriptor_get_datatype( dd );

          if ( type == DT_CHAR )
          {
            if ( acol->width )
              dmi_descriptor_set_length( dd, acol->width );
            else
              dmi_descriptor_set_length( dd, CARD_SIZE );
          }
          dmi_descriptor_set_kdd( dd, acol );
          cpts[j] = dd;
        }
        dd = dmi_make_vector( block, colname, cpts, dims[i] );
      }
      else
        if ( dims[i] == 1 ) /* this is a scalar column */
        {
          acol = acpts[i][0];
          strcpy( colname, acol->name );
          if ( acol->length > 0 )
          {
            naxes = 1;
            axes = malloc( sizeof(long) );
            axes[0] = acol->length;
          } 
          else
          {
            naxes = 0;
            axes = NULL;
          }
          dmi_column_make_raw_name( colname, colno, rawname );
          us_dd = dmi_descriptor_alloc( block, dmCOLUMN, colno, rawname,
				        acol->datatype, NULL, NULL, 1, dmVALUE, naxes,
				        axes, DM_FLAG_RAW | DM_FLAG_LOGICAL );
          type = dmi_descriptor_get_datatype( us_dd );

          if ( type == DT_CHAR )
          {
            if ( acol->width )
              dmi_descriptor_set_length( us_dd, acol->width );
            else
              dmi_descriptor_set_length( us_dd, CARD_SIZE );
          }

          dmi_descriptor_set_kdd( us_dd, acol );
          dd =  dmi_create_linear_wcs( block, us_dd, colname, acol->datatype, 
				       0.0, 0.0, 1.0 );
          type = dmi_descriptor_get_datatype( dd );

          if ( type == DT_CHAR )
          {
            if ( acol->width )
              dmi_descriptor_set_length( dd, acol->width );
            else
              dmi_descriptor_set_length( dd, CARD_SIZE );
          }
          dmi_descriptor_set_kdd( dd, acol );
          dd = dmi_make_vector( block, colname, &dd, 1 );
          /* ryan, I may need to initialize the column range and the null value */
        }
        if ( axes ) free( axes );
          rawcols[i] = dd;
      }
      dmi_make_unscaled_groups( rawcols, ncols );
    }

  return rawcols;
}
#endif



/***************************************************************/
/*                                                             */
/* Function:   afp_simple_insert_col                           */
/*                                                             */
/* Description:                                                */
/*              Insert column                                  */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock   *block  - block pointer        */
/*                    dmDescriptor  * dd - descriptor pointer  */
/*                    long       no      - number              */
/*                    void      *ikdd                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_simple_insert_col( dmBlock* block, dmDescriptor* dd,
                            long no, void* ikdd )
{
  dmDescriptor *cpt_dd  = NULL;
  dmDescriptor *col_dd  = NULL;
  dmDescriptor *us_dd   = NULL;
  afBlock      *ablock  = NULL;
  afCol        *acol    = NULL;
  afVector     *avec    = NULL;

  long dim    = 0;
  long ncols  = 0;
  long kcolno = 0;
  long cptno  = 0;
  long moved  = 0;
  dmDataType type   = DT_UNK;
  dmBool     append = dmFALSE;

  ablock = (afBlock*)dmi_block_get_kb( block );
  ncols = dmi_block_get_ncols( block );

  append = ( no == ncols + 1 );
  dim = dmi_descriptor_get_vector_dim( dd );

  if ( append )
    kcolno = ablock->ncols+1;
  else
  {
    col_dd = dmi_block_get_col( block, no );
    cpt_dd = dmi_descriptor_get_cpt( col_dd, 1 );
    acol = (afCol*)dmi_descriptor_get_kdd( cpt_dd );
    kcolno = acol->no;
  }

  if ( dim > 1 )
    avec = afk_vector_alloc( dim );

  for ( cptno = 1; cptno <= dim; cptno++ )
  {
    cpt_dd = dmi_descriptor_get_cpt( dd, cptno );
    type = dmi_descriptor_get_datatype( cpt_dd );
    acol = (afCol*)dmi_descriptor_get_kdd( cpt_dd );

    if ( !acol )
    {
      if ( ikdd ) /* pass on the kernel descriptor */
      {
        if ( dim == 1 )
          acol = afk_col_alloc_copy( (afCol*)ikdd );
        else
          if ( dim > 1 )
            acol = afk_col_alloc_copy( ((afVector*)ikdd)->cpts[cptno-1] );
      }
      else
      {
        acol = afk_col_alloc();
        acol->name     = cs_dup( dmi_descriptor_get_name( cpt_dd ));
        acol->datatype = dmi_descriptor_get_datatype( cpt_dd );
        acol->length   = dmi_descriptor_get_array_size( cpt_dd );

        // Code Change  Bug 10784  November 8, 2007
        acol->width    = dmi_descriptor_get_length( cpt_dd );

	/* Code change October 28, 2008 */
	/* Bit type stored by DM as array of bytes, but as   */
	/* single bit string in ascii                        */
	if (acol->datatype == DT_BIT )
	  acol->length = 0;
      }
      dmi_descriptor_set_kdd( cpt_dd, (void*)acol );
    }

    us_dd = dmi_descriptor_get_unscaled( cpt_dd );

    if ( us_dd )
      dmi_descriptor_set_kdd( us_dd, (void*)acol );
    if ( dim > 1 )
      avec->cpts[cptno-1] = acol;

    afk_block_insert_col( ablock, acol, kcolno );
    afk_buffer_insert_col( ablock->abuf, acol, kcolno, &moved );
    ++kcolno;
  }

  if ( dim > 1 )
  {
    dmi_descriptor_set_kdd( dd, avec );
  }

  afp_simple_write_col_header( block, dd, no );
}



/***************************************************************/
/*                                                             */
/* Function:   afp_simple_write_col_header                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  This function is called from afp_simpe_make_vectors(),     */
/*  which is used for both SIMPLE and CSCRDB processing.       */
/*  If using CSCRDB, return from this function immediately     */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock   *block  - block pointer        */
/*                    dm Descriptor  * dd                      */
/*                    long       no                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_simple_write_col_header( dmBlock* block, dmDescriptor* dd,
                                  long no )
{
  char col_entry[CARD_SIZE];
  char colline[AK_COMMENTLINE_LEN];

  dmuKeyList keylist;

  dmuKey       *ukey        = NULL;
  dmDescriptor *cpt_dd      = NULL;
  afBlock      *ablock      = NULL;
  afBuffer     *abuf        = NULL;
  afStyle      *astyle      = NULL;
  char         *old_colline = NULL; 
  char         *ptr         = NULL;
  char         *eptr        = NULL;
  char         *name        = NULL;
  const char   *sep         = NULL;
  char          sep_char    = ' ';

  long   dim       = 0;
  long   arsize    = 0;
  long   cptno     = 0;
  long   len       = 0;
  long   entry_len = 0;
  long   colno     = 0;
  long   ncols     = 0;
  dmBool append    = UT_FALSE;
  char   hchar     = '\0';
 

  ablock = (afBlock*)dmi_block_get_kb( block );
  astyle = ablock->ads->astyle;
  ncols  = dmi_block_get_ncols( block );
  append = ( no == ncols+1 );


  // Code Change  Bug 10752  September 26, 2007
  //   Get any specified separator characters and use
  //   the first one in the list when forming the
  //   format strings below that write out
  //   the header information 

  sep = afk_style_get_sep((const afStyle *) astyle);
  sep_char = sep[0];


  keylist = ablock->keylist;
  dim = dmi_descriptor_get_vector_dim( dd );

  if ( astyle->colnames == AK_STYLE_COLNAMES_FIRST )
  {
    ukey = keylist->data[0];
    old_colline = dmu_key_value_c( ukey );
  }
  else
    if ( astyle->colnames == AK_STYLE_COLNAMES_LAST )
    {
      ukey = keylist->data[keylist->n];
      old_colline = dmu_key_value_c( ukey );  
    }

  eptr = col_entry + CARD_SIZE;
  ptr = col_entry;
  name = dmi_descriptor_get_name( dd );
  len = strlen( name );

  if ( len + ptr < eptr )
    ptr += sprintf( col_entry, "%s", name );

  arsize = dmi_descriptor_get_array_size( dd );

  if ( arsize > 1 && ( ptr + (arsize / 10) + 3 < eptr ))
    ptr += sprintf( ptr, "[%ld]", arsize );

  if ( dim > 1 )
  {
    cpt_dd = dmi_descriptor_get_cpt( dd, 1 );
    name = dmi_descriptor_get_name( cpt_dd );
    len = strlen( name );

    if ( ptr + len + 2 < eptr )
      ptr += sprintf( ptr, "(%s", dmi_descriptor_get_name( cpt_dd ));

    for ( cptno = 2; cptno <= dim; cptno++ )
    {
      cpt_dd = dmi_descriptor_get_cpt( dd, cptno );
      name = dmi_descriptor_get_name( cpt_dd );
      len = strlen( name );

      if ( ptr + len + 1 < eptr )
        ptr += sprintf( ptr, ",%s", name );
    }

    strcat( col_entry, ")" );
    ++ptr;
  }

  entry_len = ptr - col_entry;
  len = strlen( old_colline );

  if ( entry_len + len > AK_COMMENTLINE_LEN )
    return;

  if ( append )
    sprintf( colline, "%s%c%s", old_colline, sep_char, col_entry );
  else
    if ( no == 1 )
      sprintf( colline, "%s%c%s", col_entry, sep_char, old_colline );
    else
    {
      dmBool in_paren = UT_FALSE;
      ptr = old_colline;
      colno = 1;

      while( *ptr == ' ' || *ptr == '\t' )
        ++ptr;

      while( *ptr != '\0' && colno != no )
      {
        if ( !in_paren && *ptr == '(' )
          in_paren = UT_TRUE;
        else
          if ( in_paren && *ptr == ')' )
            in_paren = UT_FALSE;
          else
            if ( *ptr == ' ' || *ptr == '\t' )
              ++colno;
       ++ptr;
      }

      if ( *ptr != '\0' )
      {
        *ptr = '\0';
        ++ptr;
      }

      sprintf( colline, "%s%c%s%c%s", old_colline, sep_char, col_entry, sep_char, ptr );
    }
 
    free( old_colline );
 
    *(char**)&ukey->value = cs_dup( colline );
    len = strlen( colline ) + 1;

    if ( len + 3 > AK_COMMENTLINE_LEN )
      return;

    memmove( colline + 1, colline, len );

    hchar = afk_style_get_header_char( astyle );

    if ( hchar )
    {
      colline[0] = hchar;
    }

    strcat( colline, "\n" );
    abuf = ablock->abuf;
    afk_buffer_replace_keyline( abuf, colline, ukey->index );
}



/***********************************************************************/
/*                                                                     */
/*  Function:  afp_simple_create_block()                               */
/*              Create the SM block in the file.  Need to initialize   */
/*              the cols definition line.                              */
/*                                                                     */
/*  Input parameters:                                                  */
/*                     dmBlock   *block                                */
/*                                                                     */
/*  Output parameters:                                                 */
/*                                                                     */
/*  Returns:  nothing                                                  */
/*                                                                     */
/***********************************************************************/
void afp_simple_create_block( dmBlock* block )
{
  afBlock   *ablock = NULL;
  afDataset *ads    = NULL;
  afFile    *afptr  = NULL;
  afBuffer  *abuf   = NULL;
  afStyle   *astyle = NULL;
  long       spos   = 0;
  long       moved  = 0;
  int        flavorStringLength = 0;
  char       flavorString[CARD_SIZE];
  const char *pFlavorString = NULL;


  ablock = (afBlock*)dmi_block_get_kb( block );
  ads    = ablock->ads;
  astyle = ads->astyle;
  abuf   = afk_block_get_buffer( ablock );
  afptr  = afk_dataset_fptr( ads );


  pFlavorString = afk_style_get_flavorString(astyle);
  if ( pFlavorString )
    strcpy(flavorString, pFlavorString);
  strcat (flavorString, "\n" );
  flavorStringLength = strlen(flavorString);

  afk_file_insert_data( afptr, flavorString );

  afk_buffer_flush( abuf, &moved );

  afk_file_seek( afptr, ads->pos+flavorStringLength, SEEK_SET );
 
  spos = afk_file_tell( afptr );
  ads->blocks_pos[0] = spos;
  afk_buffer_set_datastart( abuf, spos );

  afk_simple_write_comment( ablock, 1, dmCOMMENT, " " );
 
}
