/*                                                                
**  Copyright (C) 2004-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   afsubspace.c                                                */
/*                                                               */
/*   Functions that create and work with ASCII file type         */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"
#include "../ftkernel/ftkernelP.h"
#include "afkernel.h"



/***************************************************************/
/*                                                             */
/* Function:   afp_get_ftss                                    */
/*                                                             */
/* Description:                                                */
/*       From ftkernel code.                                   */
/*       Get the kernel-specific subspace col for a descriptor,*/
/*       creating it if needed.  Manages the index of DSTYPn   */
/*       keywords.  If n is negative, increment the DSTYP      */
/*       index; otherwise use n as the index.                  */
/*                                                             */
/*  Input parameters:                                          */
/*                dmBlock       *block                         */
/*                dmDescriptor  *dd                            */
/*                long           n                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            ftSubspaceCol  *ftss                             */
/*                                                             */
/***************************************************************/
ftSubspaceCol* afp_get_ftss( dmBlock* block, dmDescriptor* dd, long n )
{
  ftSubspaceCol  *ftss    = NULL;
  afBlock        *ablock  = (afBlock*)dmi_block_get_kb( block );

  ftss = dmi_descriptor_get_kdd( dd );

  if ( ftss && ftss->type != FTK_SUBSPACE )
  {
    sprintf( dm_errbuf, "INTERNAL ERROR FTK SUB %s %ld \n",
             dmi_descriptor_get_name( dd ), ftss->type );
    dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
  } 
 
  if ( !ftss )
  {
    ftss = afp_ftss_alloc( ablock, n );

    dmi_descriptor_set_kdd( dd, ftss );
  }
 
 return  ftss;
}


/***************************************************************/
/*                                                             */
/* Function:   afp_ftss_alloc                                  */
/*                                                             */
/* Description:                                                */
/*       From ftkernel code.                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                dmBlock       *block                         */
/*                long           n                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            ftSubspaceCol  *ftss                             */
/*                                                             */
/***************************************************************/
ftSubspaceCol* afp_ftss_alloc( afBlock* ablock, long n )
{
  ftSubspaceCol *ftss = NULL;


  ftss = calloc( 1, sizeof( ftSubspaceCol ));
  ftss->type = FTK_SUBSPACE;

  if ( n < 0 )
  {
    ftss->index = ++ablock->max_dstyp;  
  }
  else 
  {
    ftss->index = n;

    if ( n > ablock->max_dstyp ) 
      ablock->max_dstyp = n;
  }

 return ftss;
}


/***************************************************************/
/*                                                             */
/* Function:   afp_read_subspace_values                        */
/*                                                             */
/* Description:                                                */
/*       From ftkernel code.                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                dmDescriptor  *dd                            */
/*                long           subspace_cptno                */
/*                char           dsval                         */
/*                char          *tabname                       */
/*                int            create                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afp_read_subspace_values( dmDescriptor* dd,
                                    long subspace_cptno, char* dsval,
                                    char* tabname, int create )
{
  dmDataset  *ds = NULL;
  dmBlock    *block = NULL;
  regRegion  *local_region = NULL;
  char       *new_tabname = NULL;
  void       *value1 = NULL;
  void       *value2 = NULL;
  void       *v1     = NULL;
  void       *v2     = NULL;
  char       *name   = NULL;
  dmDataType  type   = DT_REAL8;
  long        vcpt    = 0;
  long        minvcpt = 1;
  long        maxvcpt = 1;
  long        nvals   = 0;
  long        dim     = 0;
  int         status  = 0;
  int         null    = UT_FALSE;
  int         is_alloced = UT_TRUE;
  int         incnull    = UT_FALSE;



 if ( subspace_cptno < 0 )
   return dmFAILURE;

  block = dmi_descriptor_get_block( dd );
  ds = dmi_block_get_ds( block );
  dim = dmi_descriptor_get_element_dim( dd );
  name = dmi_descriptor_get_name( dd );

  /* Case of a table: GTI or REGION */
  if ( tabname )
  {
    new_tabname = ftp_subspace_table_name( ds, tabname );
    /* returns errcode */
    status = afp_subspace_read_table( dd, new_tabname, create,
                      &value1, &value2, &nvals, &local_region ); 

    if ( status != dmSUCCESS )
    {
      free( new_tabname );
      return status;
    }
  }
  else
  {
    /* Case of a DSVAL filter string  - parse it.       */
    /* We recognize the following cases:                */
    /*  - String contains a "(". It must be a region.   */
    /*  - String is "NULL" denoting the null subspace.  */
    /*  - Otherwise, string must be a simple range list.*/

    if ( dsval )
    {
      dim = 1;
      null = dmi_parse_filter_string( dsval, dd, &value1, &value2,
                                      &nvals, &local_region, &incnull );

      if ( null ) 
        new_tabname = cs_dup( "NULL" );

      if ( !local_region ) 
        type = dmi_descriptor_get_datatype( dd );
    }
    else
     return dmFAILURE;
  }

  if ( new_tabname ) 
    name = new_tabname;
  else
    name = tabname;

  if ( local_region )
  {
    maxvcpt = 1;
    dim = 2;  
    dmi_subspace_col_init_region( dd, name, local_region );
  }
  else
  {
    if ( nvals > 0 || null )
    {
      for ( vcpt = minvcpt; vcpt <= maxvcpt; vcpt++ )
      {
        if ( dim ==2 ) 
        {
          v1 = ar_gen_ptr( value1, type, 0, vcpt-1 );
          v2 = ar_gen_ptr( value2, type, 0, vcpt-1 );
          /* Array mode */
        }
        else
        {
          v1 = value1;
          v2 = value2;
        }
        dmi_subspace_col_init_cpt( dd, name, vcpt, type, value1, value2, nvals );
      }
    }
    else
      if ( incnull )
      {
        for ( vcpt = minvcpt; vcpt <= maxvcpt; vcpt++ )
        dmi_subspace_col_init_cpt_incnull( dd, name, vcpt, type, incnull );
      }
  }

  if ( nvals > 0 && is_alloced )
  {
    free( value1 );
    free( value2 ); 
  }

  if ( new_tabname )
    free( new_tabname );

 return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_subspace_read_table                         */
/*                                                             */
/* Description:                                                */
/*       From ftkernel code.                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                dmDescriptor  *dd                            */
/*                const char    *tabname                       */
/*                dmBool         create                        */
/*                void         **val1                          */
/*                void         **val2                          */
/*                                                             */
/*  Output parameters:                                         */
/*                long          *nvals                         */
/*                regRegion    **ss_region                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode afp_subspace_read_table( dmDescriptor* dd, const char* tabname,
                                   dmBool create, void** val1, void** val2,
                                   long* nvals, regRegion** ss_region )
{
  dmDataset  *ds           = NULL;
  dmBlock    *parent_block = NULL;
  dmBlock    *ss_block     = NULL;
  void       *vval1        = NULL;
  void       *vval2        = NULL;

  dmDataType  dd_type;
  dmDataType  type1;
  dmDataType  type2;
  dmErrCode   status = dmFAILURE;

  long  dim  = 0;
  long  no   = 0;
  long  size = 0;


  *ss_region = NULL;
  *nvals = 0;

  parent_block = dmi_descriptor_get_block( dd );

  ds  = dmi_block_get_ds( parent_block );
  dim = dmi_descriptor_get_element_dim( dd );


  /* A block should not have itself as a subspace.   */
  /* This is arguable, but saves some problems...    */
  if ( create && cs_eq( tabname, dmi_block_get_name( parent_block )))
  {
    return dmSUCCESS; /* Smile */
  }

  /* Is this name an existing block or not? */
  no = dmi_block_search( ds, tabname );

  if ( no > 0 )
  {
    /* We don't use ftBlockOpen here to avoid creating the */
    /* subspace and getting in an infinite loop.           */
    ss_block = afp_block_open( ds, no, UT_FALSE, UT_FALSE, NULL );

    if ( dim == 2 )
    {
      /* try reading as a region first */
      *ss_region = afp_block_read_region( ss_block );
    }

    if ( !*ss_region )
    {
      /* try reading as a GTI block */
      *nvals = afp_block_read_gti( ss_block, val1, &type1, val2, &type2 );

      if ( *nvals > 0 )
      {
        /* make sure we return range values of the same type as the col */
        dd_type = dmi_descriptor_get_datatype( dd );

        if ( dd_type != type1 )
        {
          size = ar_mix_type_size( dd_type );
          vval1 = calloc( *nvals, size );
          ar_mix_cast_array( *val1, type1, dd_type, vval1, *nvals );
          free( ar_gen_ptr( *val1, type1, 0, 0 ));
          *val1 = vval1;
        }

        if ( dd_type != type2 )
        {
          size = ar_mix_type_size( dd_type );
          vval2 = calloc( *nvals, size );
          ar_mix_cast_array( *val2, type2, dd_type, vval2, *nvals );
          free( ar_gen_ptr( *val2, type2, 0, 0 ));
          *val2 = vval2;
        }
      }
    }

    afp_block_close( ss_block );
    /* We close and free the block descriptors */
    dmi_block_close_descriptors( ss_block, afDescriptorFree );

    if ( *nvals > 0 || *ss_region )
      status = dmSUCCESS;
   }
   else
     if ( tabname && !cs_is_blank( tabname ))
     {
       sprintf( dm_errbuf, "NO SUBSPACE TABLE %s\n", tabname );
       dmu_err_set( DM_ERR_SUBSPACE, dm_errbuf );
     }

 return status;
}


/***************************************************************/
/*                                                             */
/* Function:   afp_block_read_region                           */
/*                                                             */
/* Description:                                                */
/*       From ftkernel code.                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                dmBlock     *block                           */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            regRegion  region                                */
/*                                                             */
/***************************************************************/
regRegion* afp_block_read_region( dmBlock* block )
{
  dmDescriptor  **cols   = NULL;
  regRegion      *region = NULL;
  dmDataType     *types  = NULL;
  double         *pos  = NULL;
  double         *x    = NULL;
  double         *y    = NULL;
  long           *dims = NULL;
  void           *ptr  = NULL;
  long            maxpoints = 0; /* Max number of points for position vector */
  long            n       = 0;
  long            i       = 0;
  long            ncols   = 0;
  long            vdim    = 2;
  long            offset  = 0;
  int             include = 1;
  int             orFlag  = 0;
  int             start   = 1;
  short           oldCpt  = 1;
  short           cpt     = 1;

  char shape_name[FLEN_CARD] = "";
  char shape_type[FLEN_CARD] = "";
  double r[2] = { 0.0, 0.0 };
  double angle[2] = { 0.0, 0.0 };


  if ( !block )
    return NULL;
 
  /* Which columns are the interesting ones? */
  ncols = dmi_region_file_cols( block, &cols, &dims );

  if ( !ncols )
    return NULL;

  types = (dmDataType*)calloc( ncols, UT_SZ_I );

  for ( i = 0; i < ncols; i++ )
    types[i] = dmi_descriptor_get_datatype( cols[i] );

  maxpoints = dims[REG_POS_COL];
  ptr = (double*)calloc( vdim * maxpoints, UT_SZ_D );
  pos = (double*)calloc( vdim * maxpoints, UT_SZ_D );
  x = &pos[0];  
  y = &pos[maxpoints];


  region = regCreateRegion( NULL, NULL );


  /*  Now we have the columns. */
  n = dmi_block_get_nrows( block );

  for ( i = 0; i < n; i++ )
  {
    if ( cols[REG_POS_COL] )
    {
      if ( types[REG_POS_COL] == DT_REAL8 )
        afp_block_read_col( block, cols[REG_POS_COL], NULL, NULL, NULL,
                            i+1, i+1, pos );
      else
      {
        afp_block_read_col( block, cols[REG_POS_COL], NULL, NULL, NULL,
                            i+1, i+1, ptr );
        ar_mix_numeric_cast_array( ptr, types[REG_POS_COL], DT_REAL8, pos,
                                   vdim * maxpoints );
      }
    }

    if ( cols[REG_SHAPE_COL] )
      afp_block_read_col( block, cols[REG_SHAPE_COL], NULL, NULL, NULL,
                          i+1, i+1, shape_name );
  
    if ( cols[REG_R_COL] )
    {
      if ( types[REG_R_COL] == DT_REAL8 )
        afp_block_read_col( block, cols[REG_R_COL], NULL, NULL, NULL,
                            i+1, i+1, r );
      else
      {
        afp_block_read_col( block, cols[REG_R_COL], NULL, NULL, NULL,
                            i+1, i+1, ptr );
        ar_mix_numeric_cast_array( ptr, types[REG_R_COL], DT_REAL8, r, 2 );
      }
    }

    if ( cols[REG_ROT_COL] )
    {
      if ( types[REG_ROT_COL] == DT_REAL8 )
        afp_block_read_col( block, cols[REG_ROT_COL], NULL, NULL, NULL,
                            i+1, i+1, angle );
      else
      {
        afp_block_read_col( block, cols[REG_ROT_COL], NULL, NULL, NULL,
                            i+1, i+1, ptr );
        ar_mix_numeric_cast_array( ptr, types[REG_ROT_COL], DT_REAL8, angle, 2);
      }
    }

    if ( cols[REG_CPT_COL] )
    {
      if ( types[REG_CPT_COL] == DT_INT2 )
        afp_block_read_col( block, cols[REG_CPT_COL], NULL, NULL, NULL,
                            i+1, i+1, &cpt );
      else
      {
        afp_block_read_col( block, cols[REG_CPT_COL], NULL, NULL, NULL,
                            i+1, i+1, ptr );
        ar_mix_numeric_cast( ptr, types[REG_CPT_COL], DT_INT2, &cpt );
      }
    }

    include = (shape_name[0] != '!' );
    offset = include ? 0 : 1;
    strcpy( shape_type, shape_name+offset );
 
    if ( start )
      start = 0;
    else 
      orFlag = ( cpt != oldCpt );

    oldCpt = cpt;
    regAppendShape( region, shape_type, include, orFlag, x, y, maxpoints,
                    r, angle, UT_FALSE, UT_FALSE );
  }

  free( types );
  free( pos );
  free( cols );
  free( dims );
  free( ptr );

  return region;
}



/***************************************************************/
/*                                                             */
/* Function:   afp_block_read_region                           */
/*                                                             */
/* Description:                                                */
/*       From ftkernel code.                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                dmBlock     *block                           */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            regRegion  region                                */
/*                                                             */
/***************************************************************/
long afp_block_read_gti( dmBlock* block, void** val1, dmDataType *type1,
                         void** val2, dmDataType* type2 )
{
 dmDescriptor *cols[2];
 dmDataType    ltype1;   /* local type1 */
 dmDataType    ltype2;   /* local type2 */

 long nrows = 0;
 long size  = 0;
 long ncols = 0;

#if 0
 afBlock* ablock = NULL;

 ablock = (ftBlock*)dmi_block_get_kb( block );
#endif


  ncols = dmi_block_get_ncols( block );

  /* Convention: GTI is first two scalar cols whatever they are */
  if ( ncols < 2 )
  {
    return 0;
  }

  cols[0] = dmi_block_get_col( block, 1 );
  cols[1] = dmi_block_get_col( block, 2 );
  nrows   = dmi_block_get_nrows( block );
  ltype1  = dmi_descriptor_get_datatype( cols[0] );
  ltype2  = dmi_descriptor_get_datatype( cols[1] );

  if ( nrows > 0 && ar_mix_is_numeric( ltype1 ) && ar_mix_is_numeric( ltype2 ))
  {
    *type1 = ltype1;
    *type2 = ltype2;
    size = ar_mix_type_size( *type1 );
    *val1 = calloc( nrows, size );
    size = ar_mix_type_size( *type2 );
    *val2 = calloc( nrows, size );
    afp_block_read_col( block, cols[0], NULL, NULL, NULL, 1, nrows, *val1 );
    afp_block_read_col( block, cols[1], NULL, NULL, NULL, 1, nrows, *val2 );
  }
  else
    nrows = 0;

  return nrows;
}
