/*                                                                
**  Copyright (C) 2004-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   aftform.c                                                   */
/*                                                               */
/*   Functions that create and work with ASCII file type         */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"


/*----------------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------------*/
/*  The binary table TFORM keyword is of the form   nCm                             */
/*  where n is the array size, C is the datatype, and m is extra info.              */
/*  For string cols, the string length is m (default 1) and the array length is n/m */
/*  For BIT columns, the length is n and the array length is the next multiple of 8.*/
/*  For ASCII tables, the array size is always 1 and the string length is m.        */
/*  For variable length arrays, the max size of the array can be returned.          */
/*----------------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------------*/



/***************************************************************/
/*                                                             */
/* Function:   afp_compose_tform                               */
/*                                                             */
/* Description:                                                */
/*             Here we encode the TFORM convention which tells */
/*             FITS about the column properties (data type,    */
/*             size, etc).                                     */
/*                                                             */
/*             TEXT kernel uses TFORM almost exactly like FITS */
/*             Binary TFORMS, but all dtatypes can have a      */
/*             width suffix.                                   */
/*                                                             */
/*             width is the display width (only usefule for    */
/*             fixed width tables).  For strings, length is    */
/*             used instead.                                   */ 
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDataType  type                         */
/*                    long        isize                        */
/*                    long        width                        */
/*                    dmBool      ascii                        */
/*                    dmBool      variable                     */
/*                    char       *tform                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_compose_tform( dmDataType type, long isize, long length,
                        long width, dmBool ascii, dmBool variable,
                        char* tform )
{
  char multiplier[CARD_SIZE]="";
  char suffix[CARD_SIZE] ="";
  char letter  = ' ';
  long size    = isize;
  long type_id = 0;
  long ntypes  = 10;
  long types[] = { DT_BIT, DT_UINT1, DT_INT2, DT_INT4, DT_REAL4, DT_REAL8, DT_LOGICAL, DT_UINT2, DT_UINT4, DT_CHAR };
  char bletters[] = "XBIJEDLIJA";
  char aletters[] = "IIIIFFIIIA";
  char *letters = NULL;

  /* use ascii for binary data types */
  if ( ascii )
    letters = aletters;
  else 
    letters = bletters;

  /* Force scalars to be of size 1 */
  if ( !variable && size == 0 )
    size = 1;

  if ( type == DT_BIT )
  {
    sprintf( multiplier, "%ld", length );

  }
  else
    if ( variable )
      cs_copy( "1P", multiplier );
    else
      if ( type == DT_CHAR ) 
      {
        sprintf( multiplier, "%ld", length * size );

        if ( size > 1 )
          sprintf( suffix, "%ld", length );
      }
      else
      {
        sprintf( multiplier, "%ld", size );
        suffix[0] = '\0';
#if 0
   if ( width > 0 )
     sprintf( suffix, "%ld", width );
#endif
      }

   type_id = ar_match_i( (long)type, types, ntypes );

   if ( type_id > 0 )
   {
     letter = letters[type_id-1];
   }
   else
     dmu_err_set( DM_ERR_INTERNAL, "FTK: Bad datatype in TFORM" );

   sprintf( tform, "%s%c%s", multiplier, letter, suffix );

   if( variable && size > 0 )
   {
     sprintf( suffix, "(%ld)", size );
     strcat( tform, suffix );
   }
}



/***************************************************************/
/*                                                             */
/* Function:   afp_parse_tform                                 */
/*                                                             */
/* Description:                                                */
/*             TEXT kernel uses TFORM almost exactly like FITS */
/*             Binary TFORMS, but all dtatypes can have a      */
/*             width suffix.                                   */
/*                                                             */
/*             width is the display width (only usefule for    */
/*             fixed width tables).  For strings, length is    */
/*             used instead.                                   */ 
/*                                                             */
/*  Input parameters:                                          */
/*                    const char *tform                        */
/*                    dmBool      ascii                        */
/*                    long       *arraysize                    */
/*                    long       *length                       */
/*                    long       *widthble                     */
/*                    dmBool     *variable                     */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDataType  type                                 */
/*                                                             */
/***************************************************************/
dmDataType afp_parse_tform( const char* tform, dmBool ascii,
                          long* arraysize, long* length,
			  long* width,
                          dmBool* variable )
{
  char  *tptr = (char*)tform;
  long   trailer = 1;
  long   multiplier = 1;
  char  *ptr   = NULL;
  char  *final = NULL;
  char   code;
  dmDataType type;


  if ( !tform )
    return dmUNKNOWNTYPE;

  /* Strip off leading multiplier */
  while( isspace( *tptr ) )
    ++tptr;

  if ( isdigit( *tptr ) )
    multiplier = strtol( tptr, &ptr, 10 );
  else
    ptr = (char*)tptr;  /* Discard const char* */

  *length    = 0;
  *arraysize = multiplier; 

  if ( !ptr ) 
    return dmUNKNOWNTYPE;

  /*-----------------------------------------------*/
  /*    Parse the character code for the data type.*/
  /*   Check for variable length columns.          */
  /*-----------------------------------------------*/
  code = *ptr;

  if ( code )
    ++ptr;

  if ( ascii )
  {
    type = ftp_at_tform_to_datatype( code );
    *variable = UT_FALSE;
  }
  else
  {
    *variable = ( code == 'P' ); /* Variable length */

    if ( *variable )
    {
      code = *ptr++;

      if ( *ptr == '(' )
        ++ptr;
    }

    /* Provisional type - must still check for unsigned case */
    type = ftp_bt_tform_to_datatype( code );
  }

  if ( ptr && isdigit( *ptr ) ) 
    trailer = strtol( ptr, &final, 10 );
  else
    trailer = 0;

  /*------------------------------------------*/
  /* Set the final array size and length.     */
  /* String and bit columns are special cases.*/
  /*------------------------------------------*/
  if ( *variable )
  {
    if ( final && *final == ')' )
    {
      if ( trailer == 0 )
        trailer = 1;

      *arraysize = trailer;
    }
    else
    {
      *arraysize = 0;
    }
  } 
  else
    if ( type == DT_CHAR )
    {
      *length    = trailer;

      if ( !ascii )
      {
        if ( trailer > 0 )
        {
          *length    = trailer;
          *arraysize = multiplier/trailer; 
        }
        else
        {
          *arraysize = 1;
	  trailer    = multiplier;
          *length    = multiplier;
        }
      }
    } 
    else
      if ( type == DT_BIT )
      {
	/* set up as array of chars */
        *arraysize = 1;
        trailer    = multiplier;
        *length    = multiplier;
      }
  
  *width = trailer;
  return type;
 }


/***************************************************************/
/*                                                             */
/* Function:   afp_parse_tdisp                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    const char *tdisp                        */
/*                    char       *typecode                     */
/*                    long       *width                        */
/*                    long       *precision                    */
/*                    long       *exponent                     */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_parse_tdisp( const char* tdisp, char* typecode, long* width,
                      long* precision, long* exponent )
{
  const char *ptr;
 
  // Initialize precision and exponent to 0
  // Width and typecode are always set no matter what the datatype

  *exponent  = 0;
  *precision = 0;

  ptr = tdisp;

  // Get the data type, 1 char
  *typecode = *ptr;
  ++ptr;
 
  // width is any number following the type code
  *width = atoi( ptr );

  // Check for a period. It indicates precision value
  ptr = strchr( ptr, '.' );

  if ( ptr )
  {
    // copy precision value. it follows the dot, so increment ptr
    ++ptr;
    *precision = atoi( ptr );

    // There may be an optional exponent value. Search for 'E'

    ptr = strchr( ptr, 'E' );
 
    if ( ptr )
    {
      // Exponent value comes after 'E'. Increment ptr. Get value

      ++ptr;
      *exponent = atoi( ptr );
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:   afp_compose_tdisp                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDataType  type                         */
/*                    long        width                        */
/*                    char       *tdisp                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void afp_compose_tdisp( dmDataType type, long width, char* tdisp )
{
  long precision = 0;
  char typecode;


  switch ( type )
  {
    case DT_CHAR:
      typecode = 'A';
      break;
    case DT_UINT1:
    case DT_UINT2:
    case DT_UINT4:
    case DT_INT2:
    case DT_INT4:
      typecode = 'I';
      break;
    case DT_REAL4:
      typecode = 'F';
      precision = width - 6;
      break;
    case DT_REAL8:
      typecode = 'E';
      precision = width - 6;
      break;
    case DT_BIT:
      typecode = 'I';
      break;
    case DT_LOGICAL:
      typecode = 'A';
      break;
    default:
      typecode = '\0';
      break;
  }

 
  if ( precision > 0 )
    sprintf( tdisp, "%c%ld.%ld", typecode, width, precision );
  else
    sprintf( tdisp, "%c%ld", typecode, width );
 
}
