/*                                                                
**  Copyright (C) 2003-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   aftsv.c                                                     */
/*                                                               */
/*   Functions that create and work with ASCII file type         */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "afkernelP.h"
#include "ryi.h"


dmDescriptor** afp_tsv_make_vectors( dmBlock* block, afCol*** acpts,
                                     long ncols, long* dims,
                                     char **ppUnitStrings,
                                     char **ppDescriptions );


/***************************************************************/
/*                                                             */
/* Function:   afp_tsv_create_block                            */
/*                                                             */
/* Description:                                                */
/*              Create a new block for tsv file                */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock   *block                         */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/*  Called from:                                               */
/*        afp_block_insert()                                   */
/*                                                             */
/***************************************************************/
 void afp_tsv_create_block( dmBlock *block )
 {
   afBlock   *pakBlock  = NULL;
   afDataset *pakDataset = NULL;
   afFile    *pakFile    = NULL;
   afBuffer  *pakBuffer  = NULL;
   afStyle   *pakStyle   = NULL;

   long       pos     = 0;
   long       moved   = 0;
   int        flavorStringLength = 0;

   const char *pFlavorString  = NULL;
   char        flavorString[CARD_SIZE];

   pakBlock   = (afBlock*)dmi_block_get_kb( block );
   pakDataset = pakBlock->ads;
   pakStyle   = pakDataset->astyle;
   pakBuffer  = afk_block_get_buffer( pakBlock );
   pakFile    = afk_dataset_fptr( pakDataset );


   pFlavorString = afk_style_get_flavorString(pakStyle);

   if ( pFlavorString)
     strcpy( flavorString, pFlavorString );
   strcat (flavorString, "\n" );

   flavorStringLength = strlen(flavorString);

   // insert flavor string into the file and flush buffer
   afk_file_insert_data( pakFile, flavorString );

   afk_buffer_flush( pakBuffer, &moved );

   // reset start of file
   afk_file_seek( pakFile, pakDataset->pos+flavorStringLength, SEEK_SET );
 
   pos = afk_file_tell( pakFile );
   pakDataset->blocks_pos[0] = pos;
   afk_buffer_set_datastart( pakBuffer, pos );

   return;
 }


/*****************************************************************************/
/*                                                                           */
/*  Function:  afp_tsv_init_cols()                                           */
/*                                                                           */
/*  Description:                                                             */
/*      Based on afp_simple_init_cols()                                      */
/*                                                                           */
/*      Calls dmi_block_set_raw_cols() and afp_tsv_make_vectors()            */
/*                                                                           */
/*  Input parameters:                                                        */
/*                    dmBlock   *pBlock                                      */
/*                                                                           */
/*  Output parameters:                                                       */
/*                    dmuKey    *key                                         */
/*                                                                           */
/*  Returns:                                                                 */
/*     dmErrCode  status                                                     */
/*                                                                           */
/*  Called from:                                                             */
/*        afp_block_init_cols()                                              */
/*                                                                           */
/*****************************************************************************/
dmErrCode afp_tsv_init_cols( dmBlock *pBlock)
{
  dmErrCode      status   = dmSUCCESS;
  dmDataset     *pDataset = NULL;
  dmDescriptor **rawcols  = NULL;

  afDataset *pakDataset = NULL;
  afStyle   *pakStyle   = NULL;
  afBlock   *pakBlock   = NULL;
  afCol     *pakCol     = NULL;
  afCol   ***ppakCmpnts = NULL;

  dmBool     bAppend    = dmFALSE;  /* Flag to append vector component */

  char    colName[CARD_SIZE];
  char   *pEnd           = NULL;
  char  **ppUnitStrings  = NULL;
  char  **ppDescriptions = NULL;

  long *maxcpts = NULL;   /* number of cpts in each cpts element */
  long *dims    = 0;      /* pointer to number of components for each vector */
  long  numCols = 0;
  long  colincr = 10;     /* amount to increment maxcols when it overflows */
  long  cptincr = 5;      /* amount to increment maxcpts when it overflows */
  long  maxcols = 10;     /* number of cols in cpts and ncpts */
  long  newVal  = 0;
  long  colIndex= 0;
  long  cptNumber = 0;
  long  numKCols = 0;
  long  ii      = 0;
  long  jj      = 0;


  // Get dataset pointer from block pointer
  pDataset = dmi_block_get_ds( pBlock );

  // Get Ascii dataset pointer from dataset pointer
  pakDataset = (afDataset*)dmi_dataset_get_kds( pDataset );

  // Get style structure
  pakStyle = afk_dataset_get_style( pakDataset );
  pakBlock = (afBlock*)dmi_block_get_kb( pBlock );

  // Get the basic column information
  status = afk_tsv_init_cols( pakBlock, pakStyle, &ppUnitStrings, &ppDescriptions );

  // Get vector information
  numKCols   = afk_block_ncols( pakBlock );
  dims       = (long*)calloc( maxcols, sizeof(long) );
  maxcpts    = (long*)calloc( maxcols, sizeof(long) );
  ppakCmpnts = (afCol***)calloc( numKCols, sizeof(afCol**) );

  /* first, loop through and put the afCols into a list of vector component lists */
  for( ii = 0; ii < numKCols; ii++ )
  {
    pakCol = afk_block_get_col( pakBlock, ii+1 );

    bAppend = dmFALSE;
    colIndex = 0;

    /* if this is a vector component, it will be named 'vec(cpt)' */
    pEnd = strchr( pakCol->name, '(' );

    if ( pEnd )
    {
      /* copy vector name + '(' char */
      cs_copy_siz( pakCol->name, colName, pEnd - pakCol->name+1 );

      for ( jj = 0; jj < numCols; jj++ )
      {
        if (  ppakCmpnts[jj] &&  ppakCmpnts[jj][0] && 
	   ( cs_eq_siz( colName,  ppakCmpnts[jj][0]->name, strlen(colName) )))
        {
          /* this vector is already in the list, mark for append */
	  bAppend = dmTRUE;
          colIndex = jj;
          break;
        }
      }
    }

    if ( bAppend )
    {
      /* append this vector component to existing vector */
      dims[colIndex]++;

      if ( dims[colIndex] > maxcpts[colIndex] )
      {
        maxcpts[colIndex] += cptincr;
        ppakCmpnts[colIndex] = (afCol**)realloc( ppakCmpnts[colIndex], sizeof(afCol*) * maxcpts[colIndex] );
      }

      cptNumber = dims[colIndex]-1;
      ppakCmpnts[colIndex][cptNumber] = pakCol;   

    }
    else
    {
       /* create a new column */
       colIndex = numCols;
       ++numCols;

       if ( numCols > maxcols )
       {
         newVal = maxcols + colincr;
         ppakCmpnts = (afCol***)realloc( ppakCmpnts, sizeof(afCol**) * newVal );
         maxcpts = (long*)realloc( maxcpts, sizeof(long) * newVal );
         dims = (long*)realloc( dims, sizeof(long) * newVal );
         memset( &ppakCmpnts[maxcols], 0, colincr*sizeof(afCol**) );
         memset( &maxcpts[maxcols], 0, colincr*sizeof(long) );
         memset( &dims[maxcols], 0, colincr*sizeof(long) );
         maxcols = newVal;
       }

       dims[colIndex]++;

       if ( dims[colIndex] > maxcpts[colIndex] )
       {
         newVal = maxcpts[colIndex] + cptincr;
         ppakCmpnts[colIndex] = (afCol**)realloc( ppakCmpnts[colIndex], 
				      sizeof(afCol*) * newVal );
         memset( &ppakCmpnts[colIndex][maxcpts[colIndex]], 0, cptincr*sizeof(afCol*) );
         maxcpts[colIndex] = newVal;
       }  
       ppakCmpnts[colIndex][0] = pakCol;
    }
  }


  // Use our list of lists to create our final array of dmDescriptors
  // Also use list of units and descriptions to add onto the dmDescriptors
  //   when they are created

  rawcols = afp_tsv_make_vectors( pBlock, ppakCmpnts, numCols, dims, ppUnitStrings, ppDescriptions );

  // Clean-up all this memory

  for ( ii = 0; ii < numCols; ii++ )
    free( ppakCmpnts[ii] );

  free( ppakCmpnts );
  free( dims );
  free( maxcpts );
  free (ppUnitStrings);
  free (ppDescriptions);

  // set raw cols
  dmi_block_set_raw_cols( pBlock, rawcols, numCols );

  return (status);
}


/*****************************************************************************/
/*                                                                           */
/*  Function:  afp_tsv_insert_col()                                          */
/*                                                                           */
/*  Description:                                                             */
/*                                                                           */
/*     Based on afp_simple_insert_col()                                      */
/*                                                                           */
/*  Input parameters:                                                        */
/*                    dmBlock        *pBlock                                 */
/*                    dmDescriptor   *pDescriptor                            */
/*                    long            colNumber                              */
/*                    void           *pKernelDescriptor                      */
/*                                                                           */
/*  Output parameters:                                                       */
/*                                                                           */
/*  Returns:                                                                 */
/*     dmErrCode  status                                                     */
/*                                                                           */
/*  Called from:                                                             */
/*        afp_col_create()                                                   */
/*                                                                           */
/*****************************************************************************/
dmErrCode  afp_tsv_insert_col( dmBlock *pBlock, dmDescriptor *pDescriptor,
                                long colNumber, void *pKernelDescriptor )
{ 
  dmErrCode   status  = dmSUCCESS;
  dmBool      bAppend = dmFALSE;
  dmDataType  type    = DT_UNK;

  dmDescriptor *pColDescriptor      = NULL;
  dmDescriptor *pCmptDescriptor     = NULL;
  dmDescriptor *pUnscaledDescriptor = NULL;

  afBlock  *pakBlock  = NULL;
  afStyle  *pakStyle  = NULL;
  afCol    *pakCol    = NULL;
  afVector *pakVector = NULL;

  long  numColumns = 0;
  long  dim        = 0;
  long  kColNumber = 0;
  long  cptno  = 0;
  long  moved  = 0;


  pakBlock = (afBlock*)dmi_block_get_kb( pBlock );
  pakStyle = pakBlock->ads->astyle;

  numColumns = dmi_block_get_ncols( pBlock );
  bAppend    = ( colNumber == numColumns + 1 );
  dim = dmi_descriptor_get_vector_dim( pDescriptor );

  if ( bAppend )
    kColNumber = pakBlock->ncols+1;
  else
  {
    pColDescriptor  = dmi_block_get_col( pBlock, colNumber );
    pCmptDescriptor = dmi_descriptor_get_cpt( pColDescriptor, 1 );
    pakCol     = (afCol*)dmi_descriptor_get_kdd( pColDescriptor );
    kColNumber = pakCol->no;
  }

  if ( dim > 1 )
    pakVector = afk_vector_alloc( dim );

  for ( cptno = 1; cptno <= dim; cptno++ )
  {
    pCmptDescriptor = dmi_descriptor_get_cpt( pDescriptor, cptno );
    type   = dmi_descriptor_get_datatype( pCmptDescriptor );
    pakCol = (afCol*)dmi_descriptor_get_kdd( pCmptDescriptor );

    if ( !pakCol )
    {
      if ( pKernelDescriptor ) /* pass on the kernel descriptor */
      {
        if ( dim == 1 )
          pakCol = afk_col_alloc_copy( (afCol*)pKernelDescriptor );
        else
          if ( dim > 1 )
            pakCol = afk_col_alloc_copy( ((afVector*)pKernelDescriptor)->cpts[cptno-1] );
      }
      else
      {
        pakCol = afk_col_alloc();
        pakCol->name     = cs_dup( dmi_descriptor_get_name( pCmptDescriptor ));
        pakCol->datatype = dmi_descriptor_get_datatype( pCmptDescriptor );
        pakCol->length   = dmi_descriptor_get_array_size( pCmptDescriptor );
        pakCol->width    = dmi_descriptor_get_length( pCmptDescriptor );

	/* Bit type stored by DM as array of bytes, but as   */
	/* single bit string in ascii                        */

	if (pakCol->datatype == DT_BIT )
	  pakCol->length = 0;
      }

      dmi_descriptor_set_kdd( pCmptDescriptor, (void*)pakCol );
    }

    pUnscaledDescriptor = dmi_descriptor_get_unscaled( pCmptDescriptor );

    if ( pUnscaledDescriptor )
      dmi_descriptor_set_kdd( pUnscaledDescriptor, (void*)pakCol );

    if ( dim > 1 )
      pakVector->cpts[cptno-1] = pakCol;

    afk_block_insert_col( pakBlock, pakCol, kColNumber );
    afk_buffer_insert_col( pakBlock->abuf, pakCol, kColNumber, &moved );
    ++kColNumber;
  }

  if ( dim > 1 )
    dmi_descriptor_set_kdd( pDescriptor, pakVector );

  // update header keys
  afp_tsv_write_col_header( pBlock, pDescriptor, colNumber );

  return (status);
}


/*****************************************************************************/
/*                                                                           */
/*  Function:  afp_tsv_key_compose_colnames()                                */
/*                                                                           */
/*  Description:                                                             */
/*       based on afp_simple_write_col_header with changes to accommodate    */
/*       TSV header definitions                                              */
/*                                                                           */
/*       Reads column name on descriptor and adds it onto the value of a key */
/*       that has all column names separated by tab. Key is rewritten to     */
/*       key list with composite value                                       */
/*                                                                           */
/*  Input parameters:                                                        */
/*                    dmDescriptor   *pDescriptor                            */
/*                    afStyle        *pakStyle,                              */
/*                    char           *currKeyValue                           */
/*                    char           *newEntry                               */
/*                    dmBool          append                                 */
/*                    long            colNumber                              */
/*                                                                           */
/*  Output parameters:                                                       */
/*                    char           *newKeyValue                            */
/*                                                                           */
/*  Returns:                                                                 */
/*     dmErrCode  status                                                     */
/*                                                                           */
/*  Called from:                                                             */
/*        afp_tsv_write_col_header()                                         */
/*                                                                           */
/*****************************************************************************/
dmErrCode afp_tsv_key_compose_colnames( dmDescriptor *pDescriptor,
                                        afStyle *pakStyle,
                                        char *currKeyValue,
                                        char *newEntry,
                                        char  cSeparator,
                                        dmBool append,
                                        long   colNumber,
			                char *newKeyValue )
{
  dmDescriptor *pCpntDescriptor = NULL;
  dmErrCode     status = dmSUCCESS;

  char  *pEndNames   = NULL;
  char  *pStartNames = NULL;
  char  *name = NULL;
  char   columnNames[CARD_SIZE];

  long  dim = 0;
  long  len = 0;
  long  entry_len  = 0; 
  long  arraySize  = 0;
  long  cmptNumber = 0;
  long  colNo      = 0;


  dim = dmi_descriptor_get_vector_dim( pDescriptor );

  pEndNames = columnNames + CARD_SIZE;
  pStartNames  = columnNames;

  len = strlen( newEntry );

  if ( len + pStartNames < pEndNames )
    pStartNames += sprintf( columnNames, "%s", newEntry );

  arraySize = dmi_descriptor_get_array_size( pDescriptor );

  if ( arraySize > 1 && ( pStartNames + (arraySize / 10) + 3 < pEndNames ))
    pStartNames += sprintf( pStartNames, "[%ld]", arraySize );

  if ( dim > 1 )
  {
    pCpntDescriptor = dmi_descriptor_get_cpt( pDescriptor, 1 );
    name = dmi_descriptor_get_name( pCpntDescriptor );
    len  = strlen( name );

    if ( pStartNames + len + 2 < pEndNames )
      pStartNames += sprintf( pStartNames, "(%s", name);

    for ( cmptNumber = 2; cmptNumber <= dim; cmptNumber++ )
    {
      pCpntDescriptor = dmi_descriptor_get_cpt( pDescriptor, cmptNumber );
      name = dmi_descriptor_get_name( pCpntDescriptor );
      len = strlen( name );

      if ( pStartNames + len + 1 < pEndNames )
        pStartNames += sprintf( pStartNames, ",%s", name );
    }

    strcat( columnNames, ")" );
    ++pStartNames;
  }

  entry_len = pStartNames - columnNames;
  len = strlen( currKeyValue );

  if ( entry_len + len > AK_COMMENTLINE_LEN )
    return (status);

  if ( append )
  {
    if ( *currKeyValue != '\0' )
      sprintf( newKeyValue, "%s%c%s", currKeyValue, cSeparator, columnNames );
    else
      sprintf( newKeyValue, "%s", columnNames );
  }
  else
    if ( colNumber == 1 )
      sprintf( newKeyValue, "%s%c%s", columnNames, cSeparator, currKeyValue );
    else
    {
      dmBool in_paren = UT_FALSE;
      pStartNames = currKeyValue;
      colNo = 1;

      while( *pStartNames == '\t' || *pStartNames == ' ' )
        ++pStartNames;

      while( *pStartNames != '\0' && colNo != colNumber )
      {
        if ( !in_paren && *pStartNames == '(' )
          in_paren = UT_TRUE;
        else
          if ( in_paren && *pStartNames == ')' )
            in_paren = UT_FALSE;
          else
            if ( *pStartNames == '\t' || *pStartNames == ' ' )
              ++colNo;
       ++pStartNames;
      }

      if ( *pStartNames != '\0' )
      {
        *pStartNames = '\0';
        ++pStartNames;
      }

      sprintf( newKeyValue, "%s%c%s%c%s", currKeyValue, cSeparator, columnNames, cSeparator, pStartNames );
    }
 
  return (status);
} 


/***************************************************************/
/*                                                             */
/* Function:   afp_tsv_make_vectors                            */
/*                                                             */
/* Description:                                                */
/*              Initialize column names and call function      */
/*              ryi_column_alloc() to create dmDescriptors     */
/*              for ascii kernel descriptors                   */
/*                                                             */
/*   base on afp_simple_make_vectors                           */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock   *pBlock  - block pointer       */
/*                    afCol   ***ppakComponents                */
/*                    long       numCols                       */
/*                    long      *dims                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor  **Descriptor                       */
/*                                                             */
/***************************************************************/
dmDescriptor** afp_tsv_make_vectors( dmBlock *pBlock, 
                                     afCol ***pppakComponents,
                                     long numCols, long *dims,
                                     char **ppUnitStrings,
                                     char **ppDescriptions )
{
  dmDescriptor **rawcols = NULL;
  dmDescriptor **ppComponents = NULL;
  dmDescriptor  *pDescriptor = NULL;

  afCol         *pakCol    = NULL;
  afVector      *pakVector = NULL;

  char *pUnit      = NULL;
  char *pDesc      = NULL;
  char *pNameEnd   = NULL;
  char *pNameStart = NULL;

  char  sCptName[CARD_SIZE];
  char  sColName[CARD_SIZE];
  char  sUnit[CARD_SIZE];
  char  sDesc[AK_COLVAL_LEN];
  char  sDisp[CARD_SIZE];

  long  ii      = 0;
  long  jj      = 0;
  long  length  = 0;
  long  naxes   = 1; /* currently we only deal with 1-dim arrays in SIMPLE */
  long  axes[1]; 
  long  colNumber = 0;


  if ( numCols )
    rawcols = (dmDescriptor**)malloc( sizeof(dmDescriptor*) * numCols );

  /* loop through vector component lists to construct DM columns */

  if ( rawcols )
  {
    for ( ii = 0; ii < numCols; ii++ )
    {
      sUnit[0] = '\0';
      sDisp[0] = '\0';
      sDesc[0] = '\0';
      pUnit = (char *)ppUnitStrings[ii];
      pDesc = (char *)ppDescriptions[ii];

      colNumber = ii + 1;
      axes[0]   = 1;
      naxes     = 1; /* currently, SIMPLE is restricted to 1-dim arrays */
      length    = 0;


      // Vector columns have dim > 1

      if ( dims[ii] > 1 )
      {
        // This is a vector column
        // Get vector name

        pakVector      = afk_vector_alloc( dims[ii] );
        ppComponents = (dmDescriptor**)malloc( sizeof(dmDescriptor*) * dims[ii] );
        pakCol       = pppakComponents[ii][0];
        pNameEnd     = strchr( pakCol->name, '(' );
        cs_copy_siz( pakCol->name, sColName, pNameEnd - pakCol->name );

        // Get components names

        for ( jj = 0; jj < dims[ii]; jj++ )
        {
          pakCol     = pppakComponents[ii][jj];
          pakVector->cpts[jj] = pakCol;
          pNameEnd   = strchr( pakCol->name, '(' );
          pNameStart = pNameEnd+1;
          pNameEnd   = strpbrk( pNameStart, " ,)" );

          if ( pNameEnd )
            cs_copy_siz( pNameStart, sCptName, pNameEnd - pNameStart );
          else
            cs_copy( pNameStart, sCptName );

          // If string type, need to set length element to size

          if ( pakCol->datatype == DT_CHAR )
          {
            if ( pakCol->width > 0 )
              length = pakCol->width;
            else
              length = AK_COLVAL_LEN;

           // Be sure to set column width in afCol structure
           // String length info gets into descriptor
           // but not to the afCol structure.

           pakCol->width = length;
         }
	 else
	 {
           //   Make a note of whether the boolean column uses a single 
           //   character (T or F) or the whole word (TRUE or FALSE) by 
           //   using the value stored in pakCol->width. This is for future
           //   use to facilitate exact copying of this type of column
           //   from one ascii flavor to another.

           if ( pakCol->datatype == DT_LOGICAL && pakCol->width > 0 )
             length = pakCol->width;    
           else
             length = 0;
         }

         // The following corrects parse error for format <name>[##](X,Y)

         if (pakCol->length > 1)
           axes[0] = pakCol->length;

         // Get unit and  description strings
         if ( pUnit )
           strcpy(sUnit, pUnit);

         if ( pDesc )
           strcpy( sDesc, pDesc);

         if ( pakCol->disp )
           strcpy(sDisp, pakCol->disp);

         // Make dmDescriptor for this component column
         pDescriptor = ryi_column_alloc( pBlock, colNumber, sCptName, pakCol->datatype, length,
				sUnit, sDesc, 1, NULL, naxes, axes, 0.0, 0.0, 1.0, pakCol );

         // Set display string now
         if ( sDisp[0] != '\0' )
           dmi_descriptor_set_disp(pDescriptor, sDisp);

         ppComponents[jj] = pDescriptor;
       }

       // Make dmDescriptor for this vector column

       pDescriptor = ryi_column_alloc( pBlock, colNumber, sColName, DT_UNK, 0,
	 		      sUnit, sDesc, dims[ii], ppComponents, 0, NULL, 
			      0.0, 0.0, 0.0, pakVector );

       // Set display width now
       if ( sDisp[0] != '\0' )
	 dmi_descriptor_set_disp(pDescriptor, sDisp);
      
      }
      else
        if ( dims[ii] == 1 )
        {
          // This is a scalar column

          // Get column name
          pakCol = pppakComponents[ii][0];
          strcpy( sColName, pakCol->name );

          // If array, length is non-zero

          if ( pakCol->length > 0 )
            axes[0] = pakCol->length;

          // If string, set string length

          if ( pakCol->datatype == DT_CHAR )
          {
            if ( pakCol->width > 0 )
              length = pakCol->width;
            else
              length = AK_COLVAL_LEN;
          }
	  else
	  {
            // Code Change  Bug 11791 February 25, 2010
            //   Make a note of whether the boolean column uses a single 
            //   character (T or F) or the whole word (TRUE or FALSE) by 
            //   using the value stored in pakCol->width. This is for future
            //   use to facilitate exact copying of this type of column
            //   from one ascii flavor to another.

            if ( pakCol->datatype == DT_LOGICAL && pakCol->width > 0 )
              length = pakCol->width;    
	  }

          // Get unit and  description strings
          if ( pUnit )
            strcpy(sUnit, pUnit);

          if ( pDesc )
            strcpy( sDesc, pDesc);

          if ( pakCol->disp )
            strcpy(sDisp, pakCol->disp);

          // Make dmDescriptor for this column

          pDescriptor = ryi_column_alloc( pBlock, colNumber, sColName, pakCol->datatype, length,
	  		                  sUnit, sDesc, 1, NULL, naxes, axes, 0.0, 0.0, 1.0,
		                          pakCol );

          // Set display format string

          if ( sDisp[0] != '\0' )
            dmi_descriptor_set_disp(pDescriptor, sDisp);
        }

      // store column descriptor
      rawcols[ii] = pDescriptor;
    }
  }
  return rawcols;
}


/*****************************************************************************/
/*                                                                           */
/*  Function:  afp_tsv_write_col_header()                                    */
/*                                                                           */
/*  Description:                                                             */
/*       based on afp_simple_write_col_header with changes to accommodate    */
/*       TSV header definitions                                              */
/*                                                                           */
/*       This function generates the column information comment lines and    */
/*       the column names line in the TSV header                             */
/*                                                                           */
/*  Input parameters:                                                        */
/*                    dmBlock        *pBlock                                 */
/*                    dmDescriptor   *pDescriptor                            */
/*                    long            colNumber                              */
/*                                                                           */
/*  Output parameters:                                                       */
/*                                                                           */
/*  Returns:                                                                 */
/*     dmErrCode  status                                                     */
/*                                                                           */
/*  Called from:                                                             */
/*        afp_tsv_insert_col()                                               */
/*                                                                           */
/*****************************************************************************/
dmErrCode afp_tsv_write_col_header( dmBlock *pBlock,
                                    dmDescriptor *pDescriptor,
                                    long colNumber )
{
  ftKeyType   fktype = FTK_COMMENT;

  dmuKey    *nameKey = NULL;
  dmuKey    *pKey    = NULL;
  dmBool     bAppend = dmFALSE;
  dmBool     bFirst  = dmFALSE;
  dmErrCode  status  = dmSUCCESS;
  dmuKeyList keyList;

  afBlock  *pakBlock  = NULL;
  afStyle  *pakStyle  = NULL;
  afBuffer *pakBuffer = NULL;
 
  const char *sSeparator = NULL;
  char *tag          = "dmCOMMENT";
  char *comment      = "";
  char *currKeyValue = NULL;
  char *newEntry     = NULL;
  char *pDesc        = NULL;
  char *pDisp        = NULL;
  char *pUnit        = NULL;

  char  cHeaderChar       = '\0';
  char  cDefaultHeadChar  = '#';
  char  cSeparator        = '\0';
  char  cDefaultSeparator = '\t';
  char  cNull             = '\0';
  char  sSep[2]           = { 2 * '\0'};
  char  sDisplay[10]      = { 10 * '\0' };
  char  newLine[AK_COMMENTLINE_LEN];

  long  numCols = 0;
  long  mask    = UT_TRUE;
  long  len     = 0;
  long  numKeys = 0;
  long  ColNum  = 0;


  // get pointer, initial values
  pakBlock  = (afBlock*)dmi_block_get_kb( pBlock );
  pakStyle  = pakBlock->ads->astyle;
  pakBuffer = pakBlock->abuf;

  numCols = dmi_block_get_ncols( pBlock );
  bAppend  = ( colNumber == numCols+1 );
  keyList = pakBlock->keylist;

  bFirst = (!keyList->data);

  //   Get any specified separator characters and use
  //   the first one in the list when forming the
  //   format strings
  //   If NULL, use default (tab)

  sSeparator = afk_style_get_sep((const afStyle *) pakStyle);

  if ( sSeparator )
    cSeparator = sSeparator[0];

  if ( cSeparator == cNull)
     cSeparator = cDefaultSeparator;

  sSep[0] = cSeparator;
  sSep[1] = cNull;

  cHeaderChar = afk_style_get_header_char((const afStyle *) pakStyle);

  if (cHeaderChar == cNull )
    cHeaderChar = cDefaultHeadChar;


  // Compose header column comment line
  // Information is contained in the descriptor elements

  newEntry = dmi_descriptor_get_name( pDescriptor );
  ColNum   = dmi_descriptor_get_no( pDescriptor );
  pDesc    = dmi_descriptor_get_desc( pDescriptor );
  pUnit    = dmi_descriptor_get_unit( pDescriptor );
  pDisp    = dmi_descriptor_get_disp( pDescriptor );

  // fuss with strings

  // If no display format, compose one
  if ( !pDisp )
  {
    dmDataType type   = dmi_descriptor_get_datatype(pDescriptor);
    long       length = dmi_descriptor_get_length(pDescriptor);

    afp_compose_tdisp(type, length, sDisplay);
    pDisp = sDisplay;
  }

  // Start formin column comment line
  sprintf(newLine, "%c%s%c%s%c(%s)%c",cHeaderChar,"Column",cSeparator,newEntry,cSeparator,pDisp,cSeparator);

  // add description string
  if ( pDesc )
    sprintf(newLine, "%s%s%c( )%c", newLine, pDesc, cSeparator,cSeparator);
  else
    sprintf(newLine, "%s%c( )%c", newLine, cSeparator,cSeparator);

  // add unit string
  if ( pUnit )
    sprintf(newLine, "%s[\"%s\"]%c[ ]\n", newLine, pUnit,cSeparator);
  else
    sprintf(newLine, "%s[\"\"]%c[ ]\n", newLine, cSeparator);
    
  // make comment key

  pKey = dmu_key_alloc();
  dmu_key_insert( keyList, pKey, ColNum );
  comment = newLine;
  dmu_key_set( pKey, FTK_COMMENT, DT_CHAR, tag, &comment, NULL, NULL, dmTRUE );

  status = afk_buffer_insert_keyline( pakBuffer, newLine, pKey->index );


  // Form column names line
  // If no line exists, start one

  if ( bFirst )
  {
    newLine[0] = '\0';

    //  Create column names line key
    numKeys = keyList->n;
    nameKey = dmu_key_alloc();
    nameKey->type = FTK_KEY;
    dmu_key_insert( keyList, nameKey, ++numKeys );
    dmu_key_set( nameKey, fktype, DT_CHAR, tag, &comment, NULL, NULL, mask );
    status = afk_buffer_insert_keyline( pakBuffer, newLine, nameKey->index );
  }
  else
  {
    // key exists; append new name entries onto its value string
    // get key, using known offsets in keylist

    numKeys = keyList->n;
    nameKey = keyList->data[numKeys-1];
  }

  // Column Names line

  if ( nameKey )
  {
    // Insert name of new column into key
    currKeyValue = dmu_key_value_c( nameKey );
    newEntry = dmi_descriptor_get_name( pDescriptor );

    afp_tsv_key_compose_colnames( pDescriptor, pakStyle, currKeyValue,
                                  newEntry, cSeparator, bAppend, colNumber, newLine );
    free( currKeyValue );

    // put new key line into column names key 
    *(char**)&nameKey->value = cs_dup( newLine );

    // add newline char to end of string, insert into buffer
    len = strlen( newLine );

    if ( len <= AK_COMMENTLINE_LEN-1 )
      strcat( newLine, "\n" );

    status = afk_buffer_replace_keyline( pakBuffer, newLine, nameKey->index );
  }

  return (status);
}

