/*                                                                
**  Copyright (C) 2003-2004,2006-2009  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ryi.c                                                       */
/*                                                               */
/*   Functions that create and work with ASCII file type         */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"
#include "ryi.h"


/***************************************************************/
/*                                                             */
/* Function:   ryi_column_alloc                                */
/*                                                             */
/* Description:                                                */
/*             For scalar columns, all arguments are used.     */
/*             For vector columns, only block, no, name, dim,  */
/*             and cpts are used.                              */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock    *block - block pointer        */
/*                    long        no    - column number        */
/*                    const char *name  - column name          */
/*                    dmDataType  type                         */
/*                    long        width                        */
/*                    const char *unit                         */
/*                    const char *desc                         */
/*                    long        dim                          */
/*                    dmDescriptor **cpts                      */
/*                    long        naxes                        */
/*                    const long *axes                         */
/*                    double      val                          */
/*                    double      pix                          */
/*                    double      delt                         */
/*                    void       *kdd                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor  *dd                                */
/*                                                             */
/***************************************************************/
dmDescriptor* ryi_column_alloc( dmBlock* block, long no, 
				const char* name, 
				dmDataType type, long width,
				const char* unit, const char* desc, 
				long dim, dmDescriptor** cpts, 
				long naxes, const long* axes, 
				double val, double pix, double delt, 
				void* kdd )
{
  dmDescriptor   *us_dd   = NULL;
  dmDescriptor   *dd      = NULL;
  dmDescriptor  **us_cpts = NULL;

  char rawname[CARD_SIZE];
  long   i        = 0;
  long   flag     = DM_FLAG_RAW;
  dmBool is_image = dmFALSE;

  // Code Change - Bug 9023  Aug. 16, 2006
  // remove hard coded "flag" paramter from call
  //  to dmi_descriptor_alloc()

  // set up descriptor flag here
  is_image = (block->type == dmIMAGE);

  if (is_image)
    flag |= DM_FLAG_IMG;

  flag |= DM_FLAG_LOGICAL;


  if ( dim == 1)
  {
    dmi_column_make_raw_name( (char*)name, no, rawname );

    us_dd = dmi_descriptor_alloc( block, dmCOLUMN, no, rawname,
				  type, unit, desc, dim, dmVALUE,
				  naxes, axes, flag );

   dmi_descriptor_set_length( us_dd, width );
   dmi_descriptor_set_kdd( us_dd, kdd );
   dmi_make_vector( block, rawname, &us_dd, 1 );

   dd = dmi_create_linear_wcs( block, us_dd, (char*)name, type,
			       val, pix, delt );

   dmi_descriptor_set_length( dd, width );
   dmi_descriptor_set_kdd( dd, kdd );
   dmi_make_vector( block, (char*)name, &dd, 1 );
  }
  else
    if ( dim > 1 )
    {
      us_cpts = malloc( sizeof(dmDescriptor*) * dim );

      for ( i = 0; i < dim; i++ )
      {
        dd = cpts[i];
        us_dd = dmi_descriptor_get_unscaled( dd );
        us_cpts[i] = us_dd;
      }

      dmi_column_make_raw_name( (char*)name, no, rawname );
      us_dd = dmi_make_vector( block, rawname, us_cpts, dim );
      free( us_cpts );
      dd = dmi_make_vector( block, (char*)name, cpts, dim );
      dmi_descriptor_add_coord( us_dd, dd );
      dmi_descriptor_set_kdd( dd, kdd );
    }

  return dd;
}


/***************************************************************/
/*                                                             */
/* Function:   ryi_coord_alloc                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock    *block                        */
/*                    long        no                           */
/*                    const char *name                         */
/*                    const char *transform                    */
/*                    double     *val                          */
/*                    double     *pix                          */
/*                    double     *delt                         */
/*                    long        dim                          */
/*                    void       *kdd                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor  *dd                                */
/*                                                             */
/***************************************************************/
dmDescriptor* ryi_coord_alloc( dmBlock* block, dmDescriptor *col_dd,
                               long no, 
			       const char* name, 
			       const char* transform,
			       double *val, double *pix, double *delt, 
			       long dim,  void* kdd )
{
  dmDescriptor     *dd    = NULL;
  dmElementType     etype = dmVALUE;
  dmCoordTransform *ct    = NULL;

  char  *unit  = NULL;
  char  *desc  = NULL;
  long  *axes  = NULL;
  long   naxes = 1;
  long   vdim  = 1;


  // Code Change Bug 10319 May 14, 2007
  //   Set WCS descriptor size based on source descriptor

  if ( col_dd && (block->type != dmIMAGE) )
  {
    naxes = col_dd->naxes;
    axes  = col_dd->axes;
  }
 
  dd = dmi_descriptor_alloc( block, dmCOORD, no, name, DT_REAL8, unit, desc,
			    vdim, etype, naxes, axes, DM_FLAG_COORD );

  ct = dmi_coord_transform_alloc( dd, (char*)transform, val, pix, delt, dim );
  dmi_descriptor_set_kdd( dd, kdd );

  return dd;
}

