/*                                                                
**  Copyright (C) 1997-2003,2005-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmblock.c                                                   */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmlibP.h"

/***************************************************************/
/*                                                             */
/* Function:   dmBlockCreate                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDataset  *ds                                  */
/*             char       *name                                */
/*             dmBlockType type                                */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmBlock  *block                                 */
/*                                                             */
/***************************************************************/
dmBlock* dmBlockCreate(dmDataset *ds, char *name, dmBlockType type)
{
  dmBlock* block = NULL;

  /* Special hack to let by the old bad way of opening primary headers */

  if ( type == dmIMAGE && name && !strcmp( name, "PRIMARY" ) && ds->nblocks > 1 )
    block = dmp_block_open_mode( ds, name, 0, dmIMAGE, NULL, "RW" );

  /* End of special hack */
  else
    block = dmp_block_open_mode( ds, name, 0, type, NULL, "W" );

  return block;
}


/***************************************************************/
/*                                                             */
/* Function:   dmBlockFlush                                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock    *block                               */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             nothing                                         */
/*                                                             */
/***************************************************************/
void dmBlockFlush( dmBlock* block )
{
  dmBool flush_tables = UT_TRUE;  /* When users call, flush everything */

  dmi_block_set_update( block );
  dmp_block_flush( block, flush_tables );
}


/***************************************************************/
/*                                                             */
/* Function:   dmBlockGetType                                  */
/*                                                             */
/* Description:                                                */
/*            Return the block type                            */
/*                                                             */
/*  Input parameters:                                          */
/*             const dmBlock *block                            */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmBlockType type                                */
/*                                                             */
/***************************************************************/
dmBlockType dmBlockGetType( const dmBlock* block )
{
  return dmi_block_get_type( block );
}



/***************************************************************/
/*                                                             */
/* Function:   dmBlockGetDataset                               */
/*                                                             */
/* Description:                                                */
/*            Return a pointer to the block's parent dataset   */
/*                                                             */
/*  Input parameters:                                          */
/*             const dmBlock *block                            */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmDataset  *ds                                  */
/*                                                             */
/***************************************************************/
dmDataset* dmBlockGetDataset( const dmBlock* block )
{
  return dmi_block_get_ds( block );
}



/***************************************************************/
/*                                                             */
/* Function:   dmBlockGetName                                  */
/*                                                             */
/* Description:                                                */
/*            Return the block name                            */
/*                                                             */
/*  Input parameters:                                          */
/*             const dmBlock *block                            */
/*             const long     maxlen                           */
/*                                                             */
/*  Output parameters:                                         */
/*             char  *name                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmErrCode  err_status                           */
/*                                                             */
/***************************************************************/
dmErrCode dmBlockGetName( const dmBlock* block, char* name, 
                          const long maxlen )
{
  char *bname = NULL;

  bname  = dmi_block_get_name( block );

  if ( bname )
    cs_copy_siz( bname, name, maxlen );
  else
  {
    strcpy( name, "" );
    dmu_err_set( DM_ERR_NULL, "[dmBlockGetName]: No block" );
   }

  return dmu_err_get();
}


/***************************************************************/
/*                                                             */
/* Function:   dmBlockGetNo                                    */
/*                                                             */
/* Description:                                                */
/*            Return the block number of this block in the     */
/*            dataset, counting starts at 1                    */
/*                                                             */
/*  Input parameters:                                          */
/*             const dmBlock *block                            */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             long   block_number                             */
/*                                                             */
/***************************************************************/
long dmBlockGetNo( const dmBlock* block )
{
  if ( block )
    return block->no;
  else
    return 0;
}



/***************************************************************/
/*                                                             */
/* Function:   dmBlockDelete                                   */
/*                                                             */
/* Description:                                                */
/*            Physically delete a block from the file          */
/*            All the blocks get renumbered                    */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock *block                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmErrorCode  err_status                         */
/*                                                             */
/***************************************************************/
dmErrCode dmBlockDelete( dmBlock* block )
{
  dmKernel    kernel;
  dmDataset  *ds      =  NULL;
  long        BlockNo = 0;

  if ( !block )
    return dmSUCCESS;

  ds      = block->ds;
  kernel  = ds->kernel;
  BlockNo = block->no;

  dmp_block_close( block );

  kernel->BlockDelete( ds, BlockNo );

  return dmu_err_get();
}



/***************************************************************/
/*                                                             */
/* Function:   dmBlockClose                                    */
/*                                                             */
/* Description:                                                */
/*            Wrapper for internal block free                  */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock *block                                  */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmErrorCode  err_status                         */
/*                                                             */
/***************************************************************/
dmErrCode dmBlockClose( dmBlock* block )
{
  dmp_block_close( block );
  return dmu_err_get();
}



/***************************************************************/
/*                                                             */
/* Function:   dmBlockOpen                                     */
/*                                                             */
/* Description:                                                */
/*            Open block, return pointer to it                 */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDataset  *ds                                  */
/*             const char  name                                */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmBlock  *block                                 */
/*                                                             */
/***************************************************************/
dmBlock* dmBlockOpen( dmDataset* ds, const char* name )
{
  dmBlock *block = NULL;

  block = dmp_block_open_mode( ds, name, 0, dmUNKNOWNBLOCK, NULL, "R" );
  return block;
}



/***************************************************************/
/*                                                             */
/* Function:   dmBlockOpenMode                                 */
/*                                                             */
/* Description:                                                */
/*            Open block as either RW or R according to        */
/*            parameter 'update'                               */
/*                                                             */
/*  Input parameters:                                          */
/*             char  *vspec                                    */
/*             dmBool update                                   */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmBlock  *block                                 */
/*                                                             */
/***************************************************************/
dmBlock* dmBlockOpenMode( char* vspec, dmBool update )
{
  dmBlock *block =  NULL;
  char mode[4];

  if ( update )
    cs_copy( "RW", mode );
  else
    cs_copy( "R", mode );

  block = dmp_block_open_mode( NULL, vspec, 0, dmUNKBLOCK, NULL, mode );

  return block;
}



/***************************************************************/
/*                                                             */
/* Function:   dmBlockGetNoKeys                                */
/*                                                             */
/* Description:                                                */
/*            Return the number of block keys                  */
/*                                                             */
/*  Input parameters:                                          */
/*             const dmBlock *block                            */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             long  number_keys                               */
/*                                                             */
/***************************************************************/
long  dmBlockGetNoKeys( const dmBlock* block )
{ 
  return dmi_block_get_nkeys( block );
}



/***************************************************************/
/*                                                             */
/* Function:   dmBlockListKeys                                 */
/*                                                             */
/* Description:                                                */
/*            Note that the current key number is changed      */
/*            since BlockReadComment needs it                  */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock  *block                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmErrCode  err_status                           */
/*                                                             */
/***************************************************************/
dmErrCode dmBlockListKeys( dmBlock* block )
{
  long   keyno   = 0;
  char  *tag     = NULL;     /* Comment tag */
  char  *comment = NULL;     /* Comment string */

  dmDescriptor *dd = NULL;
  char buf[TEXT_SIZE];

  if ( !block || !block->keys )
    return dmSUCCESS;

  if ( block->nkeys == 0 )
    return dmSUCCESS;

  for ( keyno = 1; keyno <= block->nkeys; keyno++ )
  {
    dmi_block_set_keyno( block, keyno );
    dd = block->keys[keyno-1];

    if ( dd )
    {
      dmi_key_print( dd, buf );
      printf( "%s\n", buf );

      /* Reads from kernel */
      while( dmBlockReadComment( block, &tag, &comment ) )
      {
        printf( "    %-20s %s\n", tag, comment );
        free( tag );
        free( comment );
      }
    }
  }

  keyno = 1;
  dmi_block_set_keyno( block, keyno );
  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmBlockGetTypeStr                               */
/*                                                             */
/* Description:                                                */
/*            Return the block type as string value            */
/*                                                             */
/*  Input parameters:                                          */
/*             const dmBlock *block                            */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             char *type_name                                 */
/*                                                             */
/***************************************************************/
char* dmBlockGetTypeStr( const dmBlock* block )
{
  char *name = NULL;
  dmBlockType type;

  type = dmi_block_get_type( block );
  name = dmi_block_type_name( type );
  return name;
}




/***************************************************************/
/*                                                             */
/* Function:   dmBlockGetPref                                  */
/*                                                             */
/* Description:                                                */
/*           Note that the weighting dd is not yet supported   */
/*                                                             */
/*  Input parameters:                                          */
/*             const dmBlock  *block                           */
/*             dmDescriptor  **weightp                         */
/*             long           *nprefp                          */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmDescriptor **dd                               */
/*                                                             */
/***************************************************************/ 
dmDescriptor**  dmBlockGetPref( const dmBlock* block,
                                dmDescriptor** weightp, long* nprefp )
{
  return dmi_block_get_pref_list( block, weightp, nprefp );
}



/***************************************************************/
/*                                                             */
/* Function:   dmBlockSetPref                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock *block                                  */
/*             char    *buffer                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmErrCode  err_status                           */
/*                                                             */
/***************************************************************/
dmErrCode   dmBlockSetPref( dmBlock* block, char* buffer )
{
  char     **prefcols = NULL;
  long       npref    = 0;
  dmKernel   kernel;

  dmi_block_set_update( block );
  kernel = dmi_block_kernel( block );
  prefcols = cs_parse_paren_list( buffer, &npref );

  if ( kernel )
    kernel->BlockSetPref( block, prefcols, npref );

  free( prefcols );
  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmBlockSetBufferMaxSize                         */
/*                                                             */
/* Description:                                                */
/*            Set maximum number of rows to keep in buffer     */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock *block                                  */
/*             long     size                                   */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             nothing                                         */
/*                                                             */
/***************************************************************/
void dmBlockSetBufferMaxSize( dmBlock* block, long size )
{
  dmi_block_set_buffer_maxsize( block, size );
}

