/*                                                                
**  Copyright (C) 2002-2004,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmblocklib.c                                                */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmnutils.h"
#include "dmlibP.h"

void dmp_block_flush_keys( dmBlock* block );



/***************************************************************/
/*                                                             */
/* Function:   dmp_block_flush                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock  *block                            */
/*                  dmBool    flush_tables                     */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                    nothing                                  */
/*                                                             */
/***************************************************************/
void dmp_block_flush( dmBlock* block, dmBool flush_tables )
{
  dmDataset  *ds     = NULL;
  dmKernel    kernel = NULL;

  if ( !dmi_block_check_update( block ))
    return;


  /* First we flush the keywords, then the data buffer.                  */  
  /* This ensures that in FITS-like kernels the data doesn't have to     */
  /* be moved down later when the header gets bigger.                    */
  /* At block close, flush_tables is true and the GTIs get written out.  */
  /* We call flush dss twice - once to flush the header keywords, and    */
  /* once to flush the GTIs. There is a slight overhead here because the */
  /* TIME dss keywords will get written to the header multiple times;    */
  /* but I think this is a small overhead compared to the risk of moving */
  /* whole GTI tables. All the other (non-table) dss keywords only get   */
  /* written once as the update flag is cleared.                         */

  ds = block->ds;

  if ( ds )
    kernel = ds->kernel;

  dmp_flush_cols( block );
  dmp_block_flush_keys( block );
 
  if ( ds->current != block )
  {
    if ( ds->current )
      dmp_buffer_flush( ds->current ); 
    dmi_dataset_set_current( ds, block );
  }

  dmp_buffer_flush( block );

  if ( kernel )
    kernel->BlockFlush( block ); 

  dmp_block_flush_subspace( block, UT_FALSE );

  if ( dm_global & DM_GLOBAL_OLD ) /* Fudge for right order of DSVAL keys */
    dmp_block_flush_subspace( block, UT_FALSE );

  dmp_block_flush_coords( block );

  if ( kernel )
    kernel->BlockFlush( block ); 

  if ( flush_tables )
    dmp_block_flush_subspace( block, flush_tables );

  if ( kernel )
    kernel->BlockFlush( block );

  /* Only clear the update flag if the block is completely updated; */
  /* we don't want to think it's up to date when the GTIs haven't   */
  /* been flushed                                                   */
  if ( flush_tables )
    dmi_block_clear_update( block );
}


/***************************************************************/
/*                                                             */
/* Function:   dmp_block_flush_keys                            */
/*                                                             */
/* Description:                                                */
/*            Users can call e.g. dmSetDesc on a key so we     */
/*            may need to rewrite them                         */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock  *block                            */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                    nothing                                  */
/*                                                             */
/***************************************************************/
void dmp_block_flush_keys( dmBlock* block )
{
  long           keyno = 0;
  long           nvals = 1;
  dmDescriptor  *dd    = NULL;

  for ( keyno = 1; keyno <= block->nkeys; keyno++ )
  {
    dd = block->keys[keyno-1];

    if ( dd && dmi_descriptor_check_update( dd ))
    {
      nvals = dmi_descriptor_get_nvals( dd );
      dmp_key_write( block, dd->name, dd->type, dd->aux, nvals,
                     dd->unit, dd->desc, UT_FALSE );
      dmi_descriptor_clear_update( dd );
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmp_block_open_mode                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmDataset    *ds                           */
/*                  const char   *vspec                        */
/*                  const long    bno                          */
/*                  dmBlocktype   btype                        */
/*                  void         *kb                           */
/*                  const char   *mode                         */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                  dmblock  *block                            */
/*                                                             */
/***************************************************************/
dmBlock* dmp_block_open_mode( dmDataset* ds, const char* vspec,
                              const long bno, dmBlockType btype,
                              void* kb, const char* mode )
{
  dmBlock        *block     = NULL;
  dmDescriptor   *image_dd  = NULL;
  char           *select    = NULL;
  char           *hint      = NULL;  /*  [opt... ] parsing for kernel block hints */
  long            no        = bno;
  char           *gen_spec  = (char*)vspec;
  dmParseList    *vparse    = NULL;
  char           *vpname    = NULL;  /* Block name from parse */
  char           *name      = NULL;    /* Corrected block name */
  int             create    = UT_FALSE;
  dmErrCode       status    = dmFAILURE;
  int             is_filtered_image = 0;
  int             is_local_dataset  = 0;
  dmDataType      image_bin_type;


  /* For debugging output, make sure we can print vspec */
  if ( !vspec )
    gen_spec = "(Null)";

  /* Need to call this here so that dm_error gets initialized before */
  /* we call dmi_block_parse_syntax().  Otherwise, we won't catch    */
  /* any errors that are thrown in that function.                    */
  dmInit(0,NULL);

  /* Kill previous errors */
  dmu_err_clear();

  is_local_dataset = ( ds == NULL );

  /* translate vspec into dataset and block names */

  vparse = dmi_block_parse_syntax( ds, vspec, btype );

  /* Open local dataset */
  if ( dmu_err_get() != dmSUCCESS )
  {
    dmi_parse_list_free( vparse );
    return NULL;
  }

  if ( is_local_dataset && vparse )
    ds = dmp_dataset_open_mode( vparse->dsname, vparse->option, mode );

  if ( !ds ) 
    return NULL;

  /* If we are binning a table to an image, the input (raw) block type */
  /* is not the same as the requested output (filtered) block type.    */
  create = cs_uceq( mode, "W" );

  /* What is the raw name of the block? */

  if ( vparse )
  {
     if ( create && vparse->outname )
       vpname = vparse->outname;
     else
       vpname = vparse->block;
  }
  else
  {
     vpname = (char*)vspec;
  }

  if ( create )
  {
    no = dmi_block_search( ds, vpname );

    if ( no > 0 && vpname ) /* Non-fatal warning */
    {
      sprintf( dm_errbuf, "Error in DM block create: block called %s:%s already exists", ds->name, vpname );
      dmu_err_set( DM_WARN_NAME_CLASH, dm_errbuf );
    }

    no = ds->nblocks + 1;
    name = dmi_dataset_select_block_name( ds, vpname );
  }
  else
  {
    if ( vpname )
      no =   dmi_block_search( ds, vpname );
   }

  if ( no < 0 )
  {
    if ( vpname )
      sprintf( dm_errbuf, "Failed to find block %s in dataset %s", vpname, ds->name );
    else
      sprintf( dm_errbuf, "Failed to find block %ld in dataset %s", no, ds->name );

    if ( is_local_dataset )
      dmDatasetClose( ds );
    dmu_err_set( DM_ERR_OPEN_FAIL, dm_errbuf );
    return NULL; 
  }

  /* Open Raw Block */
  if ( vparse && vparse->option )
    hint = vparse->option;
  else
    hint = dmKernelGetOption();

  block = dmp_block_open_raw( ds, no, name, btype, kb, hint, mode );

  /* If can't open update try open readonly */
  if ( !block && cs_uceq( mode, "RW" ))
  {
    block = dmp_block_open_raw( ds, no, name, btype, kb, hint, "R" );
  }

  if ( !block )
  {
    if ( name )
      sprintf( dm_errbuf, "Failed to open block %s\n", name );
    else
      sprintf( dm_errbuf, "Failed to open block in %s\n", ds->name  );

    dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );

    if ( is_local_dataset )
      dmDatasetClose( ds );

    return NULL; 
  }
  else
  {
    // Code change - Bug 9053  March 7, 2006

    if (dmu_err_get() != dmSUCCESS )
    {
      // Need to put out warning messages here.
      // The function below prints only warning messages
      // for parsing errors that are now in the stack

      // code change April 12, 2006
      dmu_err_disp_select_msgs( DM_WARN_PARSE_OPTION );

      // Code change - BUf 9222 June 8, 2006
      // need to output warnings about buffer overrun and
      // data will be truncated (part of Bug 9222)
      dmu_err_disp_select_msgs( DM_WARN_DATA_TRUNCATED );

    }
    // end code change for Bug 9053 and 9222


    // clear error stack and continue processing
    dmu_err_clear();
  }


  dmi_block_open_set_binning( block, vparse, btype );


  /* Parse the cols filter. Add the visible cols to the buffer; */
  /* then add all operand cols needed to calculate them.        */
  if ( vparse )
    select = vparse->select;
  else
    select = NULL;

  dmi_block_apply_select( block, select );

  if ( dmu_err_get() != dmSUCCESS )
  {
    dmi_parse_list_free( vparse );

    if ( is_local_dataset )
      dmDatasetClose( ds );
    else
     dmBlockClose( block );
    return NULL;
  }


  // Code Change - October 5, 2006
  // subspace editing

  if ( vparse && vparse->subspace )
  {
    dmi_block_edit_subspace( block, vparse->subspace );

    if ( dmu_err_get() != dmSUCCESS )
    {
      dmi_parse_list_free( vparse);

      if ( is_local_dataset )
        dmDatasetClose( ds );
      else
        dmBlockClose( block );
      return NULL;
    }
  }



  /* Parse the bin filter. Add the binning cols and the result buffer to the buffer list. */


  /* Parse filters. Add filtered columns to buffer list, then add #row column */
  if ( vparse && vparse->filter )
  {
    block->virtual = UT_TRUE;
    dmp_block_process_range_filter( block, vparse->filter );

    if ( dmu_err_get() != dmSUCCESS )
    {
      dmi_parse_list_free( vparse );

      if ( is_local_dataset )
        dmDatasetClose( ds );
      else
        dmBlockClose( block );
      return NULL;
    }
  }

  /* check for an image cast (test for !create because we obviously can't */
  /* cast a nonexistent image)                                            */
  if ( !create && block->type == dmIMAGE && block->cols && vparse && !vparse->binning && !block->virtual )
  {
    /* need to force the image to get filtered if the user selected */
    /* to change the image type                                     */
    image_dd = block->cols[0];
    image_bin_type = block->image_bin_type;

    if (image_bin_type != DT_UNK && image_bin_type != image_dd->type )
      block->virtual = UT_TRUE;
  }

  is_filtered_image =  ( block->type == dmIMAGE ) && block->virtual;

  if ( ( vparse && vparse->binning ) || is_filtered_image )
  {
    block->virtual = UT_TRUE;
    dmi_parse_bin_filter( block, vparse->binning );
    status = dmu_err_get();

    /* the DM_ERR_MEM check is a hack to make sure that */ 
    /* "image too large" isn't enough to break binning  */
    if ( status != dmSUCCESS && status != DM_ERR_MEM )
    {
      dmi_parse_list_free( vparse );

      if ( is_local_dataset )
        dmDatasetClose( ds );
      else
        dmBlockClose( block );
      return NULL;
    }
  }

  dmi_block_reset_rows( block );

 /* Special routine to update dynamic keywords */
  if ( vparse && vparse->filter )
    dmp_block_update_special( block );


  /* output name */
  if ( vparse && vparse->outname )
    dmi_block_set_name( block, vparse->outname );

  dmi_parse_list_free( vparse );


  /* Check for bogus files */
  if ( !cs_uceq( mode, "W" ))
    dmi_block_check_tstart( block );


  if ( dmverbose > 0 )
  {
    int ic = 0;

    for ( ic = 0; ic < block->ncols; ic++ )
      dmi_col_print2( block->cols[ic] );

    dmi_buffer_print( block );
  }

  if ( name )
    free( name );

  if ( dmi_dataset_mode( ds ))
    block->keyno = block->nkeys;
  else
    block->keyno = 1;

  status = dmu_err_get();

  /* DM_ERR_MEM check for binning warning as above */
  if ( block && status != dmSUCCESS && status != DM_ERR_MEM )
  {
    dmu_err_clear();
  }

  return block;
}


dmBlock* dmp_block_open_raw( dmDataset* ds, long no, char* name, dmBlockType type, void* kb, char* hint, const char* mode )
{
 dmBlock* rblock = NULL;
 dmKernel kernel;

 kernel = ds->kernel;
 if ( !kernel )
  return NULL;

 if ( cs_uceq( mode, "W" ))
 {
/* Flush whatever we're doing */
  if ( ds->current )
   dmp_block_flush( ds->current, UT_FALSE ); 

  rblock = kernel->BlockCreate( ds, no, name, type, kb, hint );
 }  else {
  if ( cs_uceq( mode, "RW" ))
  {
   if ( ds->current )
    dmp_block_flush( ds->current, UT_FALSE );
  }
  rblock = kernel->BlockOpen( ds, no, type, hint );  
 }

 return rblock;
}




/***************************************************************/
/*                                                             */
/* Function:   dmp_block_close                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock  *block                            */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                  int  int                                   */
/*                                                             */
/***************************************************************/
int dmp_block_close( dmBlock* block )
{
  dmBool flush_tables = UT_TRUE;  /* At block close, flush everything */

  return dmp_block_close_mode( block, flush_tables );
}



/***************************************************************/
/*                                                             */
/* Function:   dmp_block_close_mode                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock  *block                            */
/*                  dmBool    flush_tables                     */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                  int  dmSUCCESS                             */
/*                                                             */
/***************************************************************/
int dmp_block_close_mode( dmBlock* block, dmBool flush_tables )
{
  dmDataset  *ds = NULL;
  dmKernel    kernel;

 
  if ( !block )
   return dmFAILURE;

  ds = block->ds;

  if ( dmverbose > 2 )
    dmi_buffer_status( block );

  if ( dmi_dataset_mode( ds ))
  {
    dmp_block_flush( block, flush_tables );
    dmi_dataset_set_current( ds, NULL );
  }

  if ( ds && ds->kernel )
  {
    kernel = ds->kernel;
    kernel->BlockClose( block );

    dmi_block_close_descriptors( block, kernel->DescriptorFree );
  }

  return dmSUCCESS;
}
