/*                                                                
**  Copyright (C) 1997-2004,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmblock.c                                                   */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmlibP.h"

long   dmp_live_offset( long id, long* idlist, long n );
dmBool dmp_block_copy_comments( dmBlock* srcBlock, dmBlock* destBlock );
void   dmp_livetime_free( LivetimeData* ltdata );
dmBool dmp_block_copy_keys(dmBlock *destBlock, dmBlock *srcBlock);



/***************************************************************/
/*                                                             */
/* Function:   dmBlockCopy                                     */
/*                                                             */
/* Description:                                                */
/*    Initial version of dmBlockCopy, to copy header keys only */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock  *srcBlockk                             */
/*             dmBlock  *destBlock                             */
/*             char     *option                                */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             int    status                                   */
/*                                                             */
/***************************************************************/
int dmBlockCopy( dmBlock* srcBlock, dmBlock* destBlock, char* option )
{
  dmBool retval = dmSUCCESS;

  if ( cs_uceq( option, "HEADER" ) )
  {
    retval = dmp_block_copy_keys( destBlock, srcBlock );

    if ( retval == UT_TRUE )
      return dmSUCCESS;   
  }
  else
    if ( cs_uceq( option, "SUBSPACE" ) )
    {
      retval = dmi_block_copy_subspace( srcBlock, destBlock);

      if ( retval == UT_TRUE )
        return dmSUCCESS;   
    }

  return dmFAILURE;

}



/***************************************************************/
/*                                                             */
/* Function:   dmBlockCopyCol                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock      *block2                            */
/*             dmDescriptor *parent_dd                         */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor  dd                                  */
/*                                                             */
/***************************************************************/     
dmDescriptor* dmBlockCopyCol( dmBlock* block2, dmDescriptor* parent_dd )
{
  dmDescriptor *dd        = NULL;
  dmDescriptor *new_us_dd = NULL;
  dmDescriptor *us_dd     = NULL;
  dmDataset    *ds1       = NULL;
  dmDataset    *ds2       = NULL;
  dmBlock      *block1    = NULL;
  dmKernel      kernel2;

  char *hint = NULL;
  void *kdd  = NULL;

  int same_kernel = 0;


  block1 = (dmBlock*)parent_dd->block;
  ds1    = block1->ds;
  ds2    = block2->ds;
  same_kernel = dmk_kernel_eq( ds1, ds2 );

  if ( same_kernel )
   kdd = dmi_descriptor_get_kdd( parent_dd );

  kernel2  = ds2->kernel;
  us_dd    = dmi_descriptor_get_unscaled( parent_dd );

  if ( !us_dd )
  {
    sprintf( dm_errbuf, "dmBlockCreateCopy: No unscaled column for %s", parent_dd->name );
    dmu_err_set( dmFAILURE, dm_errbuf );
    return NULL;
  }

  new_us_dd = dmi_descriptor_copy( block2, NULL, us_dd );

  if ( !new_us_dd )
  {
    sprintf( dm_errbuf, "dmBlockCreateCopy: Failed to create column %s", parent_dd->name );
    dmu_err_set( dmFAILURE, dm_errbuf );
    return NULL;
  }

  if ( block2->type == dmIMAGE )
  {
    dmi_image_descriptor_clear( block2, kernel2->DescriptorFree );
  }

  dd = dmi_descriptor_get_scaled( new_us_dd );


  /* This step is needed for virtual columns:                                     */
  /*                                                                              */
  /* If parent_dd is a component of a vector virtual column (e.g. DEC of EQPOS),  */
  /* then us_dd and new_us_dd point to the real column on which the virtual       */
  /* column is a transform (e.g. SKY), and dmi_descriptor_get_scaled( new_us_dd ) */
  /* returns the full vector column rather than the component (e.g. EQPOS rather  */
  /* than DEC).  This call to dmi_descriptor_list_search_vector() will pull the   */
  /* component out so it gets added to block2 rather than the full vector.        */
  /*                                                                              */
  /* For all other cases, dmi_descriptor_list_search_vector() will just return    */
  /* the same dd we pass in, so it's safe to always call it.                      */

  dd = dmi_descriptor_list_search_vector( (const dmDescriptor**)(&dd), 1, parent_dd->name );

  if ( !dd )
  {
    sprintf( dm_errbuf, "dmBlockCreateCopy: Failed to create column %s", parent_dd->name );
    dmu_err_set( dmFAILURE, dm_errbuf );
    return NULL;
  }

  dmi_column_add_to_list( block2, dd );
  new_us_dd->no = dd->no;

  /* Add to buffer */
  dmi_block_add_col( block2, dd, 0 );

  /* Add to subspace */
  dmi_subspace_col_init(block2, NULL, dd, dd->type );
  kernel2->ColCreate( block2, dd, kdd, hint );
  kernel2->ColSetRange( dd );
  kernel2->ColSetBin( dd );
  kernel2->ColSetNull( dd );

  if ( block2->type == dmIMAGE )
  {
     dd->flag |= DM_FLAG_UNBUF;
  }

  /* turn off filtering and binning for the output col */
  dd->flag &= ~( DM_FLAG_FILT | DM_FLAG_RESULT | DM_FLAG_BIN |
		 DM_FLAG_OPERAND | DM_FLAG_BINNED );

  return dd;
}



/***************************************************************/
/*                                                             */
/* Function:   dmBlockCopyData                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock  *srcBlock                              */
/*             dmBlock  *destBlock                             */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             int  1                                          */
/*                                                             */
/***************************************************************/        
int dmBlockCopyData( dmBlock* srcBlock, dmBlock* destBlock )
{
  dmBool flush_tables = UT_FALSE;  /* Don't flush the GTIs until block close time */

  /* Make sure header is filled */
  dmp_block_flush( destBlock, flush_tables );

  if ( !dmi_block_is_null_primary( srcBlock ))
    dmp_block_copy_buffer( srcBlock, destBlock );

  return 1;
}



/***************************************************************/
/*                                                             */
/* Function:   dmBlockCreateCopy                               */
/*                                                             */
/* Description:                                                */
/*       calls dmBlockCreateGenCopy() to do work               */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDataset  *destDataset                         */
/*             char       *destBlockName                       */
/*             dmBlock    *srcBlock                            */
/*             dmBool      copyData                            */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmBlock  *block                                 */
/*                                                             */
/***************************************************************/
dmBlock *dmBlockCreateCopy(dmDataset *destDataset, char *destBlockName,
                           dmBlock *srcBlock, dmBool copyData)
{
  return dmBlockCreateGenCopy( destDataset, destBlockName, srcBlock,
                               NULL, copyData, UT_TRUE );
}


/***************************************************************/
/*                                                             */
/* Function:   dmBlockCreateGenCopy                            */
/*                                                             */
/* Description:                                                */
/*       sets up loop to copy files                            */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDataset  *destDataset                         */
/*             char       *destBlockName                       */
/*             dmBlock    *srcBlock                            */
/*             char       *hist                                */
/*             dmBool      copyData                            */
/*             dmBool      copySubspace                        */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmBlock  *block                                 */
/*                                                             */
/***************************************************************/
dmBlock *dmBlockCreateGenCopy(dmDataset *destDataset, char *destBlockName,
                              dmBlock *srcBlock, char* hist,
                              dmBool copyData, dmBool copySubspace )
{
  dmDescriptor  **cols      = NULL;
  dmBlockType     blockType = dmUNKNOWNBLOCK;
  dmBlock        *destBlock = NULL;
  dmDataset      *ds1       = NULL;
  dmDataset      *ds2       = destDataset;

  char  destName[dmMAXSTR] = " ";
  void *kb = NULL;

  long  i     = 0;
  long  ncols = 0;

  int  same_kernel = UT_FALSE;  

  /* Validate input */
  if (!destDataset || !srcBlock)
    return NULL;


  /* Create destination block */
  blockType = dmBlockGetType(srcBlock);

  if ( !destBlockName || !*destBlockName )
  {
    dmBlockGetName( srcBlock, destName, dmMAXSTR );
    destBlockName = destName;
  }

  /* Are the kernels the same? */
  ds1 = srcBlock->ds;

  same_kernel = dmk_kernel_eq( ds1, ds2 );

  if ( same_kernel )
   kb = dmi_block_get_kb( srcBlock );
 
  if ( dmverbose >= 3 )
    fprintf( stderr, "Copying block at step 1\n" );

  dmp_buffer_flush( srcBlock );
  destBlock = dmp_block_open_mode( destDataset, destBlockName, 0, blockType, kb, "W" );

  if (!(destBlock))
    return NULL;  

  cols = dmTableOpenColumnList(srcBlock, &ncols);

  if ( cols == NULL && ncols != 0 )
    return NULL;

  if ( dmverbose >= 3 )
    fprintf( stderr, "Copying block at step 2\n" );

  for (i=0 ; i<ncols ; i++)
  {
    if ( !dmBlockCopyCol( destBlock, cols[i] ) )
      return NULL;
  }

  free( cols );
  dmp_buffer_flush( destBlock );

  if ( dmverbose >= 3 )
    fprintf( stderr, "Copying block at step 3\n" );

  /* Init dmTableCopyRow filters (R2) when dmTableCopyRow is written */
      
  /* copy key descriptors */
  if (dmp_block_copy_keys(destBlock, srcBlock) == UT_FALSE)
  {
    dmu_err_set( DMU_WARN, "Key copy failed" );
  }

  if ( hist && *hist )
  {    
     if ( dmverbose >= 1 )
       fprintf( stderr, "HIST = %s\n", hist );

     dmBlockWriteComment( destBlock, "HISTORY", hist );  
  }

  if ( dmverbose >= 3 )
    fprintf( stderr, "Copying block at step 4\n" );
  
  if ( copySubspace )
  {
    /* Copy data subspace and runtime filter information */

    if (!dmi_block_copy_subspace(srcBlock, destBlock))
    {
      dmu_err_set( DMU_WARN, "DSS copy failed" );
    }
  }

  /* Copy the data by request */
  if ( dmverbose >= 3 )
    fprintf( stderr, "Copying block at step 5\n" );

  if (copyData)
  {
    if ( dmBlockCopyData( srcBlock, destBlock ) != dmSUCCESS )
      return NULL;
  }
               
  if ( dmverbose >= 3 )
    fprintf( stderr, "Copying block complete at step 6\n" );

  return destBlock;
}
          


/***************************************************************/
/*                                                             */
/* Function:   dmp_block_copy_comments                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock  *srcBlockk                             */
/*             dmBlock  *destBlock                             */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmBool   TrueOrFalse                            */
/*                                                             */
/***************************************************************/
dmBool dmp_block_copy_comments( dmBlock* srcBlock, dmBlock* destBlock )
{
  char 	*tag     = NULL;
  char	*comment = NULL;

  /* FITS tag to skip */
  char *fits_tag = "  FITS (Flexible Image Transport System) format defined in Astronomy and";
  char *long_tag = "  This FITS file may contain long string keyword values that are";

  long skip = 0;

  while ( dmBlockReadComment(srcBlock, &tag, &comment) )
  {
    if ( !strcmp( comment, fits_tag ) )
      skip = 4;

    if ( !strcmp( comment, long_tag ) )
      skip = 4;

    if ( skip )
      --skip;
    else
    {
      if (!dmBlockWriteComment(destBlock, tag, comment))
      {
        dmu_err_set(dmFAILURE, "dmp_block_copy_keys: failed to write comment");
        return UT_FALSE;
      }
    }

    free(tag);
    free(comment);
  }

  return  UT_TRUE;
  
}



/***************************************************************/
/*                                                             */
/* Function:   dmp_block_copy_keys                             */
/*                                                             */
/* Description:                                                */
/*           copy keys for the block                           */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock  *destBlock                             */
/*             dmBlock  *srcBlock                              */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmBool   TrueOrFalse                            */
/*                                                             */
/***************************************************************/
dmBool  dmp_block_copy_keys(dmBlock *destBlock, dmBlock *srcBlock)
{
  dmDescriptor **keyList = NULL;
  dmDescriptor *dd       = NULL;
  dmBool        special  = UT_TRUE;

  long keyno   = 0;
  long numKeys = 0;

  if ( dmverbose >= 3 )
    fprintf( stderr, "Copying keys at step 1\n" );

  dmBlockMoveToKeyNo(srcBlock, 0);
  dmResetErrorState();

  if ( dmi_block_is_null_primary( srcBlock ))
    special = UT_FALSE;

  if ( !dmp_block_copy_comments( srcBlock, destBlock ) )
    return UT_FALSE;


  if ( dmverbose >= 3 )
    fprintf( stderr, "Copying keys at step 2\n" );

  keyList = srcBlock->keys;
  numKeys = srcBlock->nkeys;

  for (keyno =1 ; keyno <=numKeys; keyno++ )
  {
    dd = srcBlock->keys[keyno-1];
    srcBlock->keyno = keyno;
    dmp_key_write( destBlock, dd->name, dd->type, dd->aux, dd->size, dd->unit, dd->desc, special );

    if ( !dmp_block_copy_comments( srcBlock, destBlock ) )
      return UT_FALSE;
  }

  if ( dmverbose >= 3 )
   fprintf( stderr, "Copying keys at step 3\n" );

  /* Reset srcblock. Leave destBlock at end */
  srcBlock->keyno = 1;

  return  UT_TRUE;
}



/***************************************************************/
/*                                                             */
/* Function:   dmp_block_update_special                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  We define the ONTIMEn keyword to be the time for the CCD   */
/*  with ID=n.                                                 */
/*  This only is meaningful with tables with column CCD_ID.    */
/*  Otherwise n = the subspace cpt no.                         */
/*                                                             */ 
/*  In general the updating of time ONTIME, LIVETIME only makes*/
/*  sense if the data has a TIME column that can be filtered   */
/*  on.                                                        */
/*  We can have broken files where TSTOP - TSTART != ONTIME and*/
/*  sum(GTIs) != ONTIME. What should we do with these?         */
/*  One case is a PHA extraction where external software has   */
/*  time filtered and updated the ONTIME keyword but not       */
/*  propagated a GTI extension.                                */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock  *block                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             nothing                                         */
/*                                                             */
/***************************************************************/
void dmp_block_update_special( dmBlock* block )
{
  LivetimeData *ltdata = NULL;

  dmDescriptor *time_ss = NULL;
  dmDescriptor *dd      = NULL;
  dmDescriptor *id_ss   = NULL;

  char* key = NULL;

  double *tstart = NULL;
  double *tstop  = NULL;
  double  dtcor  = 0.0; 
  double  sum    = 0.0;

  long *idmin      = NULL;
  long *idmax      = NULL;
  long  key_offset = 0;
  long  offset     = 0;
  long  n          = 0;
  long  i          = 0;
  long  cpt        = 0;
  long  id         = 0;
  long  nid        = 0;
  long  nt         = 0;

  int   null = 0;


  /* Allocate the structure */
  time_ss = dmSubspaceColOpen( block, "TIME" );

  if ( !time_ss )
    return;

  dd = dmi_descriptor_list_search( (const dmDescriptor**)block->rawcols, block->nrcols, "TIME" );

  if ( !dd )  /* No raw time column, supress dynamic time updating */
    return;

  ltdata = calloc( 1, sizeof( LivetimeData ));

  n = dmBlockGetNoSubspaceCpts( block );
  ltdata->n = n;
  ltdata->livetime = calloc( n, UT_SZ_D);
  ltdata->ontime = calloc( n, UT_SZ_D);
  ltdata->id = calloc( n, UT_SZ_I );

  if (! dmKeyRead_d( block, "DTCOR", &dtcor ))
    dtcor = 1.0;

  null =  ( n == 1 && dmi_subspace_is_null( block->subspace ) );

  id_ss = dmSubspaceColOpen( block, "CCD_ID" );

  /* Find the values */
  for ( cpt = 1; cpt <= n; cpt++ )
  {
    dmBlockSetSubspaceCpt( block, cpt );
    dmSubspaceColGet_d( time_ss, &tstart, &tstop, &nt );
    ltdata->id[cpt-1] = cpt;

    if ( id_ss )
    {
      dmSubspaceColGet_l( id_ss, &idmin, &idmax, &nid );

      if ( nid > 0 )
        ltdata->id[cpt-1]  = idmin[0];   

      free( idmin );
      free( idmax );
    }

    sum = 0.0;

    for ( i = 0; i < nt; i++ )
      sum += tstop[i] - tstart[i];

    if (null )
    {
      if ( nt > 0 && tstart )
        dmSubspaceColSet_d( time_ss, tstart, tstart, 1 );
    }
    else
    {
      ltdata->ontime[cpt-1] += sum;
      ltdata->livetime[cpt-1] += sum * dtcor;
    }
  
    free( tstop );
    free( tstart );
  } 

 
  /* Now update the keywords - note we don't access the kernel or change the file */

  /* In the old DM, when filtering on CCD_ID=5 the other ONTIMEn got set to zero. */
  /* In the new DM, it would be nicer to delete these other keys entirely.        */
  /* At the moment this behaviour is disabled to support regression testing.      */

  for ( key_offset = 0; key_offset < block->nkeys; key_offset++ )
  {
    dd = block->keys[key_offset];
    key = dd->name;

    /* Enumerate special keys here. We do this in a painful hard coded way. */
    if ( cs_eq( key, "ONTIME" ))
    {
      ar_mix_copy_value( &ltdata->ontime[0], DT_REAL8, dd->aux, UT_TRUE, 0 );
    }
    else
      if ( cs_eq( key, "LIVETIME" ))
      {
        ar_mix_copy_value( &ltdata->livetime[0], DT_REAL8, dd->aux, UT_TRUE, 0 );
      }
      else
        if ( cs_eq( key, "EXPOSURE" ))
        {
          ar_mix_copy_value( &ltdata->livetime[0], DT_REAL8, dd->aux, UT_TRUE, 0 );
        }
        else
          if ( cs_eq_siz( key, "ONTIME", 6 ))
          {
            id = atol( dd->name + 6 );
            offset = dmp_live_offset( id, ltdata->id, ltdata->n );

            if ( offset >= 0 )
              ar_mix_copy_value( &ltdata->ontime[offset], DT_REAL8, dd->aux, UT_TRUE, 0 );
            else
              if ( id >= 0 )
              {
                if ( dm_global & DM_GLOBAL_OLD )
                {
                  double zero = 0.0;
                  ar_mix_copy_value( &zero, DT_REAL8, dd->aux, UT_TRUE, 0 );
                }
                else
                {
                  dmDescriptorDelete( dd );
                  --key_offset;
                }
              }
          }
          else
            if ( cs_eq_siz( key, "LIVTIME", 7 ) || cs_eq_siz( key, "EXPOSUR", 7 ))
            {
              id = atol( dd->name + 7 );
              offset = dmp_live_offset( id, ltdata->id, ltdata->n );

              if ( offset >= 0 )
                ar_mix_copy_value( &ltdata->livetime[offset], DT_REAL8, dd->aux, UT_TRUE, 0 );
              else
                if ( id >= 0 )
                {
                  if ( dm_global & DM_GLOBAL_OLD )
                  {
                    double zero = 0.0;
                    ar_mix_copy_value( &zero, DT_REAL8, dd->aux, UT_TRUE, 0 );
                  }
                  else
                  {
                    dmDescriptorDelete( dd );
                    --key_offset;
                  }
                }
            }
            else
              if ( cs_eq_siz( key, "LIVETI", 6 ))
              {
                /* XMM case */
                id = atol( dd->name + 6 );
                offset = dmp_live_offset( id, ltdata->id, ltdata->n );

                if ( offset >= 0 )
                  ar_mix_copy_value( &ltdata->livetime[offset], DT_REAL8, dd->aux, UT_TRUE, 0 );
                else
                  if ( id >= 0 )
                  {
                    if ( dm_global & DM_GLOBAL_OLD )
                    {
                      double zero = 0.0;
                      ar_mix_copy_value( &zero, DT_REAL8, dd->aux, UT_TRUE, 0 );
                    }
                    else
                    {
                      dmDescriptorDelete( dd );
                      --key_offset;
                    }
                  }
              } 
  } // end for loop

  dmp_livetime_free( ltdata );
}



/***************************************************************/
/*                                                             */
/* Function:   dmp_livetime_free                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             LivetimeData *ltdata                            */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                    nothing                                  */
/*                                                             */
/***************************************************************/
void dmp_livetime_free( LivetimeData* ltdata )
{
  if ( !ltdata )
    return;

  free( ltdata->livetime );
  free( ltdata->ontime );
  free( ltdata->id );
  free( ltdata );
}



/***************************************************************/
/*                                                             */
/* Function:   dmp_live_offset                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             long  id                                        */
/*             long *idlist                                    */
/*             long  n                                         */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                long   i                                     */
/*                                                             */
/***************************************************************/
long dmp_live_offset( long id, long* idlist, long n )
{
  long i = 0;

  for ( i = 0; i < n; i++ )
  {
    if ( id == idlist[i] )
      return i;
  }

  return -1;
}






