/*                                                                
**  Copyright (C) 1997-2000,2003,2005,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmimage.c                                                   */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmlibP.h"



/***************************************************************/
/*                                                             */
/* Function:   dmArrayCreateAxisGroup                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/* Note:                                                       */
/*   This routine assumes that the image has been created      */
/*   wioth the right number of total axes.                     */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor *image                             */
/*             char         *name                              */
/*             dmDataType    dtype                             */
/*             char         *unit                              */
/*             char        **cptNames                          */
/*             long          dim                               */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmDescriptor *dd                                */
/*                                                             */
/***************************************************************/
dmDescriptor* dmArrayCreateAxisGroup(dmDescriptor *image, char *name,
                                     dmDataType dtype, char *unit, 
                                     char **cptNames, long dim)
{
  double crpix[dmMAXDIM];
  double crval[dmMAXDIM];
  double cdelt[dmMAXDIM];

  dmKernel      kernel;
  dmDataType    type;
  dmDataset    *ds    = NULL;
  dmBlock      *block = NULL;
  dmDescriptor *dd    = NULL;

  long ii      = 0;
  long groupno = 0;


  if ( !image )
  {
    dmu_err_set( DM_ERR_NULL, "[dmArrayCreateAxisGroup]: Null image" );
    return NULL;
  }

  block = (dmBlock*)image->block;

  if ( !image->declared )
  {
    ds = block->ds;
    kernel = ds->kernel;

    /* Delete groups. Note that image->axes and image->naxes are not affected */
    for ( ii = 0; ii < image->ngroups; ii++ )
    {
      dd = image->groups[ii];
      dmi_descriptor_kernel_free( dd, kernel->DescriptorFree );
    }

    free( image->groups );
    image->groups = calloc( image->naxes, UT_SZ_P );
    image->ngroups = image->naxes - dim - 1;
    image->declared = UT_TRUE;
  }

  for ( ii = 0; ii < dim; ii++ )
  {
    cdelt[ii] = 1.0;
    crpix[ii] = 0.0;
    crval[ii] = 0.0;
  }

  /* Next unused group */
  groupno = 1;
  while( groupno <= image->ngroups && image->groups[groupno-1] )
    groupno++;

  type = dtype;
  type = DT_REAL8;  /* Ignore data type arg for now */

  dd = dmi_coord_init( (dmBlock*)image->block, name, unit, cptNames, dim,
		       "LINEAR", crpix, crval, cdelt, type, groupno, NULL );

  dmi_descriptor_set_update( dd );

  dmi_descriptor_set_axis( image, groupno, dd );

  return dd;
}



/***************************************************************/
/*                                                             */
/* Function:   dmArrayGetAxisGroup                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*       const dmDescriptor *dd                                */
/*       const long          no                                */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmDescriptor *dd                                */
/*                                                             */
/***************************************************************/
dmDescriptor* dmArrayGetAxisGroup( const dmDescriptor* dd, const long no )
{
  return dmi_descriptor_get_group_no( dd, no );
}



/***************************************************************/
/*                                                             */
/* Function:   dmArrayGetNoAxisGroups                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*       const dmDescriptor *dd                                */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             long ngroups                                    */
/*                                                             */
/***************************************************************/
long dmArrayGetNoAxisGroups( const dmDescriptor* dd )
{
  long ngroups = 0;

  /* Suppress #COL groups */
  if ( dmi_descriptor_silent_groups( (dmDescriptor*)dd ))
  {
    ngroups = 0;
  }
  else
  {
    ngroups = dmi_descriptor_get_no_groups( dd );
  }

  return ngroups;
}



/***************************************************************/
/*                                                             */
/* Function:   dmDatasetCreateImage                            */
/*                                                             */
/* Description:                                                */
/*  Create an image with given data type and dimensions        */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDataset  *ds                                  */
/*       const char       *name                                */
/*             dmDataType  type                                */
/*             long       *axlen                               */
/*       const long        naxes                               */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmBlock *block                                  */
/*                                                             */
/***************************************************************/
dmBlock* dmDatasetCreateImage( dmDataset* ds, const char* name, const dmDataType type,
                                   long* axlen, const long naxes )
{
  dmBlock* block = NULL;

  block = dmp_block_open_mode( ds, name, 0, dmIMAGE, NULL, "W" );

  if ( block )
    dmp_image_set_axes( block, type, axlen, naxes );

  return block;
}



/***************************************************************/
/*                                                             */
/* Function:   dmImageClose                                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock  *pImage                                */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             int dmSUCCESS                                   */
/*                                                             */
/***************************************************************/
int dmImageClose(dmBlock* pImage)
{
  dmDataset *pDs = NULL;

  if (pImage)
  {
    pDs = pImage->ds;

    dmp_block_close(pImage);
    dmDatasetClose(pDs);
  }

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmImageCreate                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            char      *vspec                                 */
/*            dmDataType dType                                 */
/*            long      *axisLens                              */
/*            long       naxes                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmBlock *block                                  */
/*                                                             */
/***************************************************************/
dmBlock* dmImageCreate(char* vspec, dmDataType dType, long *axisLens, long naxes)   
{ 
  return dmDatasetCreateImage( NULL, vspec, dType,axisLens,naxes);
}



/***************************************************************/
/*                                                             */
/* Function:   dmImageOpen                                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            char      *vspec                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmBlock *block                                  */
/*                                                             */
/***************************************************************/
dmBlock* dmImageOpen(char* vspec)
{  
  return dmp_block_open_mode( NULL, vspec, 0, dmIMAGE, NULL, "R" );
}



/***************************************************************/
/*                                                             */
/* Function:   dmImageOpenUpdate                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            char      *vspec                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmBlock *block                                  */
/*                                                             */
/***************************************************************/
dmBlock* dmImageOpenUpdate(char* vspec)
{
  return dmp_block_open_mode( NULL, vspec, 0, dmIMAGE, NULL, "RW" );
}



/***************************************************************/
/*                                                             */
/* Function:   dmp_image_set_axes                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock   *block                                 */
/*            dmDataType type                                  */
/*            long      *axlen                                 */
/*            long       naxes                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             nothing                                         */
/*                                                             */
/***************************************************************/
void dmp_image_set_axes( dmBlock* block, dmDataType type, long* axlen, long naxes )
{
  dmKernel   kernel;
  dmDataset *ds = NULL;

  /* Check NULL input */
  if ( !block ) 
    return;

  ds = block->ds;
  kernel = ds->kernel;
  kernel->ImageSetAxes( block, type, axlen, naxes );

  /* IMPLEMENT update block subspace here */

}
