/*                                                                
**  Copyright (C) 2004,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmrowfilt.c                                                 */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmlibP.h"

static int dmp_filter_row( dmBuffer* buffer, dmRowFilter* rfilter, void* row );
static dmDescriptor* dmi_subspace_col_get_filtered_col( dmDescriptor* sdd );
static dmBool dmp_rowfilter_check_syntax( dmParseList* vparse );


dmBool dmi_check_null_values(dmDescriptor* dd, dmRangeList* range, void* ptr, double* dpos, double* df, dmBool* paccept );


/***************************************************************/
/*                                                             */
/* Function:   dmTableAllocRowFilter                           */
/*                                                             */
/* Description:                                                */
/*    Allocate a row filter object                             */
/*                                                             */
/*    Note: When I say "independent" filter, I mean a filter   */
/*    on a column that exists in the buffer by itself.  If a   */
/*    full vector in the buffer is filtered, then the multi-dim*/
/*    filter is the independent filter.  If only components of */
/*    that column is are the buffer then the components are    */
/*    filtered "independently" of the other components         */
/*                                                             */ 
/*  Input parameters:                                          */
/*             dmBlock    *block                               */
/*             const char *vspec                               */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmRowFilter  *rts                               */
/*                                                             */
/***************************************************************/
dmRowFilter* dmTableAllocRowFilter( dmBlock* block, const char* vspec )
{
 dmParseList  *vparse        = NULL; /* parse list to aid DM syntax parsing */
 dmRowFilter  *rts           = NULL; /* the row filter aka generic real-time subspace */
 dmDescriptor *rts_dd        = NULL; /* a subspace column descriptor */
 dmDescriptor *rts_cptdd     = NULL; /* a vector subspace col's component's descriptor */
 dmDescriptor *dd            = NULL; /* a buffer column descriptor */
 dmDescriptor *cptdd         = NULL; /* a buffer column component descriptor */
 dmDescriptor **ddlist       = NULL; /* list of independently filtered buffer cols */
 dmDescriptor **sub_ddlist   = NULL; /* list of independently filtered subspace cols */
 dmDescriptor *rts_dd_parent = NULL; /* parent descriptor for the subspace col */
 dmBuffer     *buffer        = NULL; /* the buffer */
 dmRTSCache   *cache         = NULL; /* the real-time subspace cache */
 dmFilter     *filter        = NULL; /* a filter */
 dmFilter   ***filter_list   = NULL; /* list of independent filters */

 long *col_offsets   = NULL; /* filter to buffer->cols mapping */
 long *total_filters = NULL; /* +1 for each 1-dim filter, +2 for each 2-dim filter */
 long *filter_count  = NULL; /* count number of filters who have a column in the buffer*/
 long  colOffset   = 0; /* subspace col offset */
 long  bColOffset  = 0; /* buffer col offset */
 long  ncpts       = 0; /* number of subspace components */
 long  ncols       = 0; /* number of subspace columns */
 long  bncols      = 0; /* number of buffer columns */
 long  vcptno      = 0; /* vector component number */
 long  dim         = 0; /* vector dimensionality */
 long  cptno       = 0; /* subspace component number */
 long  n_filt      = 0; /* number of independent filters */
 long  fOffset     = 0; /* filter offset */
 long  offset      = 0; /* an offset into an array */
 int   isnull      = 0; /* is this a null subspace? */
 long  ii;              /* counter */

  if ( block->type != dmTABLE )
  {
    dmu_err_set( DM_ERR_BAD_TYPE, "Error: Row filters can only be applied to tables." );
    return NULL;
  }

  vparse = dmi_block_parse_syntax( NULL, vspec, dmTABLE );
 
  if ( !vparse )
    return NULL;

  if ( dmu_err_get() != dmSUCCESS )
  {
    dmi_parse_list_free( vparse );
    return NULL;
  }

  if ( dmp_rowfilter_check_syntax( vparse ))
  {
    dmi_parse_list_free( vparse );
    return NULL;
  }
 
 /* 
  * Parse the filter and create the runtime subspace 
  */

  if ( vparse->filter )
  {
    rts = dmp_block_parse_filter( block, vparse->filter );
    dmi_parse_list_free( vparse );

    if ( dmu_err_get() != dmSUCCESS )
    {
      dmi_subspace_free( rts, NULL );
     return NULL;
    }

    ncols = rts->ncols;
    filter_count = calloc( ncols, UT_SZ_I );
    total_filters = calloc( ncols, UT_SZ_I );

    for ( colOffset = 0; colOffset < ncols; colOffset++ )
    {
      rts_dd = rts->cols[colOffset];

      if ( !strcmp( rts_dd->name, "#row" ))
      {
        dmu_err_set( DM_ERR_PARSE, "Error: Row filter does not accept DM #row filter syntax." );
        free( filter_count );
        free( total_filters );
        dmi_subspace_free( rts, NULL );
        return NULL;
      }

      dmi_filter_check_regions( rts_dd, rts->ncpts );

      /* tally up the number of actual filters */
      rts_cptdd = dmi_subspace_col_get_filtered_col( rts_dd );
      total_filters[colOffset] += rts_cptdd->dim;
    }
  }
  else
    rts = dmi_subspace_alloc( 0, 0 );


  if ( !rts )
    return NULL;


  cache = calloc( 1, sizeof(dmRTSCache) );
  rts->cache = cache;

  n_filt = rts->ncols;
  ncpts  = rts->ncpts;
  buffer = block->buffer;
  bncols = buffer->ncols;

  cache->accept = calloc( ncpts, UT_SZ_Q );
  cache->null   = UT_FALSE;
  col_offsets   = calloc( n_filt, UT_SZ_I );
  filter_list   = calloc( n_filt, UT_SZ_P );
  ddlist        = calloc( n_filt, UT_SZ_P );
  sub_ddlist    = calloc( n_filt, UT_SZ_P );

  ncols   = rts->ncols;

  /* We need to collect all of the cache information based on the */
  /* structure of the buffer.                                     */
  /* This means finding the filtered columns in the buffer which  */
  /* can be tricky if vectors are split up in the buffer using    */
  /* column select syntax on the input block.                     */

  for ( colOffset = 0; colOffset < ncols; colOffset++ )
  {
    rts_dd     = rts->cols[colOffset];
    rts_cptdd  = NULL;

    /* Find the filtered column(s) in the buffer                     */
    /*   First, check if any of the buffer cols directly match the   */
    /*    subspace column.                                           */
    /*   Second, check if any of the buffer cols match the           */ 
    /*    subspace vector components                                 */


    rts_dd_parent = rts_dd->parent;
    for ( bColOffset = 0; bColOffset < bncols; bColOffset++ )
    {
      dd = buffer->cols[bColOffset];

      if ( rts_dd_parent == dd )
      {
	/* The subspace column is found in the buffer, count only those   */
	/* components which are being filtered.                           */
        rts_cptdd = dmi_subspace_col_get_filtered_col( rts_dd );
        filter_count[colOffset] += rts_cptdd->dim;

        /* store all information on hand in structures        */
        /*  put data descriptor of filter into ddlist         */
        /*  put buffer column offset into col_offsets         */
        /*  put subspace data descriptor into subspace list   */
        /*  allocate space for filter in a filter list        */

        if ( fOffset >= n_filt )
        {
          n_filt = fOffset+1;
          ddlist = realloc( ddlist, UT_SZ_P * n_filt );
          sub_ddlist = realloc( sub_ddlist, UT_SZ_P * n_filt );
          filter_list = realloc( filter_list, UT_SZ_P * n_filt );
          col_offsets = realloc( col_offsets, UT_SZ_P * n_filt );
        }

        ddlist[fOffset]      = dd;
        col_offsets[fOffset] = bColOffset;
        sub_ddlist[fOffset]  = rts_dd;
        filter_list[fOffset] = calloc( ncpts, UT_SZ_P );
        isnull = UT_TRUE;

        /* Add filter(s) to filter list */
        for ( cptno = 1; cptno <= ncpts; cptno++ )
        {
          filter = dmi_subspace_col_get_filter( rts_dd, cptno );
          filter_list[fOffset][cptno-1] = filter;

          if ( !filter->null )
            isnull = UT_FALSE;
        }


        /* If the filter is null over all components..   */
        /* flag the entire filter as null                */
        if ( isnull )
          cache->null = UT_TRUE;

        fOffset++;
        break;
      }
    } /* end loop over buffer cols */

    if ( !rts_cptdd && rts_dd->dim > 1  )
    {

      /* Did not find the full vector col in the buffer..       */
      /* Look for any components which are filtered             */
      rts_cptdd = dmi_subspace_col_get_filtered_col( rts_dd );

      dim = rts_cptdd->dim;
      vcptno = rts_cptdd->cptno;
      if (vcptno == 0){ vcptno = 1;}

      for ( ii = 0; ii < dim; ii++ )
      {
	vcptno += ii;
	
	cptdd = rts_dd_parent->cpts[vcptno-1];
        for ( bColOffset = 0; bColOffset < bncols; bColOffset++ )
        {
          /* get buffer column descriptor */
          dd = buffer->cols[bColOffset];

          /* SHORTCUT: only looking for vector components, so can  */
          /* skip descriptors not part of a vector                 */
          if ( dd->cptno == 0 ){ continue; }

          if ( cptdd == dd )
          {
            /* The component is found in the buffer                */
	    /* Already know it is filtered, so count it.           */
	    filter_count[colOffset] += dd->dim;

	    /* store all information on hand in structures        */
	    /*  put data descriptor of filter into ddlist         */
	    /*  put buffer column offset into col_offsets         */
	    /*  put subspace data descriptor into subspace list   */
	    /*  allocate space for filter in a filter list        */

	    if ( fOffset >= n_filt )
	    {
	      n_filt = fOffset+1;
	      ddlist = realloc( ddlist, UT_SZ_P * n_filt );
	      sub_ddlist = realloc( sub_ddlist, UT_SZ_P * n_filt );
	      filter_list = realloc( filter_list, UT_SZ_P * n_filt );
	      col_offsets = realloc( col_offsets, UT_SZ_P * n_filt );
	    }
	      
	    ddlist[fOffset]      = dd;
	    col_offsets[fOffset] = bColOffset;
	    sub_ddlist[fOffset]  = rts_dd;
	    filter_list[fOffset] = calloc( ncpts, UT_SZ_P );
	    isnull = UT_TRUE;

	    /* Add filter(s) to filter list */
            for ( cptno = 1; cptno <= ncpts; cptno++ )
            {
              filter = dmi_subspace_col_get_filter( rts_dd, cptno );
	      filter_list[fOffset][cptno-1] = filter;

	      if ( !filter->null )
		isnull = UT_FALSE;
	    }

	    /* If the filter is null over all components..   */
	    /* flag the entire filter as null                */
	    if ( isnull )
	      cache->null = UT_TRUE;

	    ++fOffset;

	    break; /* break out of loop over subspacee cpts */

          }
        } /* end buffer column loop */
      } /* end filtered vector component loop */
    }

    /* Check that we found all needed components  */
    if ( filter_count[colOffset] != total_filters[colOffset] )
    {
      sprintf( dm_errbuf,
               "Error: Could not find all filtered components of %s in block.",
                rts_dd->name );

      dmu_err_set( DM_ERR_PARSE, dm_errbuf );
      dmi_subspace_free( rts, NULL );
      free( filter_count );
      free( total_filters );
      free( ddlist );
      free( sub_ddlist );
      free( col_offsets );

      for ( offset = 0; offset < fOffset; offset++ )
        free( filter_list[offset] );

      free( filter_list );
      return NULL;
    }
  } /* end loop over subspace cols */


  // Clean-up before exit
  free( filter_count );
  free( total_filters );


  // these values are returned as part of 'rts'
  cache->n_filt = n_filt;             // number of filters
  cache->ddlist = ddlist;             // list of filtered columns
  cache->filter_list = filter_list;   // list of pts to filters list
  cache->sub_ddlist = sub_ddlist;     // list of filtered subspace columns
  cache->col_offsets = col_offsets;   // mapping of filter to buffer->cols
 
  return rts;
}



/***************************************************************/
/*                                                             */
/* Function:   dmTableFilterRow                                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock     *block                              */
/*             dmrowFilter *rts                                */
/*                                                             */
/*  Output parameters:                                         */
/*             void        *row                                */
/*                                                             */
/*  Returns:                                                   */
/*             int   status                                    */
/*                                                             */
/***************************************************************/
int dmTableFilterRow( dmBlock* block, dmRowFilter* rts, void* row )
{
  return dmp_filter_row( block->buffer, rts, row );
}


/***************************************************************/
/*                                                             */
/* Function:   dmRowFilterFree                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRowFilter *rts                                */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             nothing                                         */
/*                                                             */
/***************************************************************/
void dmRowFilterFree( dmRowFilter* rts )
{
  dmi_subspace_free( rts, NULL );
}



/***************************************************************/
/*                                                             */
/* Function:   dmp_filter_row                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBuffer    *buffer                             */
/*             dmRowFilter *rts                                */
/*             void        *row                                */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             int     accept                                  */
/*                                                             */
/***************************************************************/
int dmp_filter_row( dmBuffer* buffer, dmRowFilter* rts, void* row )
{
  double minval = 0.0;                 // range limit
  double maxval = 0.0;                 // range limit
  double ypos   = 0.0;                 // y value
  double xpos   = 0.0;                 // x value
  double df     = 0.0;

  dmRangeList   *xrange      = NULL;
  dmRangeList   *yrange      = NULL;
  dmDescriptor  *sub_dd      = NULL;   // subspace descriptor
  dmDescriptor  *sub_fltdd   = NULL;   // subspace filtered col. descriptor
  dmDescriptor  *xdd         = NULL;   // column descriptor
  dmDescriptor  *ydd         = NULL;   // column descriptor
  dmFilter      *filter      = NULL;   // filter
  dmDescriptor **ddlist      = NULL;   // descriptor list
  dmDescriptor **sub_ddlist  = NULL;   // subspace descriptor list
  dmFilter    ***filter_list = NULL;   // filter list
  dmBool        *paccept     = NULL;   // list of pass/fail results for datum
  dmRTSCache    *cache       = NULL;   // storage
  regRegion     *region      = NULL;   // region

  void *xptr    = NULL;                // pointer to x column data
  void *yptr    = NULL;                // pointer to y colum data
  void *nullval = NULL;                // null value

  long *offsets     = NULL;            // array of buffer offsets to column data
  long  bcolno      = 0;               // buffer column offset
  long  cpt_offset  = 0;               // index
  long  ncpts       = 0;               // number of components
  long  range_no    = 0;               // range number
  long  index       = 0;               // loop index
  long  nfilt       = 0;               // number of filters in filter_list
  long  vcptno      = 0;               // vector component number

  dmBool accept = dmFALSE;             // does this datum meet criterea?
  dmBool xeval  = dmFALSE;             // result 
  dmBool yeval  = dmFALSE;             // result
  int    invert = 0;                   // flag



 
  cache = rts->cache;

  if ( cache->null )
    return 0;

  ncpts  = rts->ncpts;
  invert = rts->invert;

  nfilt       = cache->n_filt;
  ddlist      = cache->ddlist;
  sub_ddlist  = cache->sub_ddlist;
  filter_list = cache->filter_list;
  offsets     = cache->col_offsets;
  paccept     = cache->accept;


  /* Here we do a minor optimization to put all the categories of filtered   */
  /* descriptor in one list; this removes one loop and one if statement from */
  /* the inner loops, and removes calls to dmi_subspace_col_get_filter.      */

  for ( cpt_offset = 0; cpt_offset < ncpts; cpt_offset++ )
  {
    paccept[cpt_offset] = 0;

    for ( index = 0; index < nfilt; index++ )
    {
      filter = filter_list[index][cpt_offset];
      region = filter->region;

      if ( filter->null )
      {
        paccept[cpt_offset] = 0;
        break;
      }

      /* Get the filtered subspace descriptor, which will tell us */
      /* which component(s) of the column need to be checked.     */
      sub_dd     = sub_ddlist[index];
      sub_fltdd  = dmi_subspace_col_get_filtered_col( sub_dd );
      vcptno     = sub_fltdd->cptno;

      bcolno     = offsets[index];

      /* Set up pointers to data */
      /*   X data descriptor is next in list */
      /*   Y data is initialized to NULL     */
      xdd     = ddlist[index];
      nullval = xdd->null;

      ydd     = NULL;
      yptr    = NULL;
      yrange  = NULL;


      if ( sub_dd->dim <= 1 )
      {
	/* We are filtering a single descriptor.. not a vector */
	/* Only need X info                                    */
	xptr = (char*)row + buffer->offsets[bcolno];
        xrange = &filter->ranges[0];

      }
      else
      {
	/* Subspace descriptor is a vector                     */
	/* We are either filtering the full vector OR          */
	/* a component thereof..                               */
	if ( sub_fltdd->dim >= xdd->dim )
	{
	  /* subspace desc. matches data descriptor (single or vector) OR        */
	  /* multiple column descriptors will combine to match a subspace vector */

	  xptr   = (char*)row + buffer->offsets[bcolno];

	  /* set range using component number to catch instance where we */
	  /* are filtering a component of a vector WITH a column filter  */
	  /* that isolates that component in the column list             */
	  /*  ie: [cols time,y][y=2900:3000] (y is part of 'sky' vector) */
	  if ( vcptno <= 0)
	    xrange = &filter->ranges[0];
	  else
	    xrange = &filter->ranges[vcptno-1];


	  if ( sub_fltdd->dim > 1 )
	  {
	    /* Filtering a vector.. setup Y info */
	    if ( sub_fltdd->dim == xdd->dim )
	    {
	      /* True vector column.. components should be adjacent in buffer */
	      ydd  = xdd;
	      yptr = (char*)xptr + ar_mix_type_size( xdd->type );
	    }
	    else if ( sub_fltdd->dim > xdd->dim  )
	    {
	      /* virtual vector column.. components are in separate buffer columns       */
	      /* The lists are setup so that the next record contains the next component */

	      ++index;
	      bcolno = offsets[index];
	    
	      ydd  = ddlist[index];
	      yptr = (char*)row + buffer->offsets[bcolno];
	    
	    }
	    yrange = &filter->ranges[1];

	  } /* end if filtering full vector */
	}
	else /* (sub_fltdd->dim < xdd->dim ) */
	{
	  /* We are filtering a component of a vector.. */
	  /* load X info according to component #       */
	  if ( vcptno <= 1)
	  {
	    xptr   = (char*)row + buffer->offsets[bcolno];
	    xrange = &filter->ranges[0];
	  }
	  else
	  {
	    xptr   = (char*)row + buffer->offsets[bcolno]+ ar_mix_type_size( xdd->type );
	    xrange = &filter->ranges[vcptno-1];
	  }
	}
      }

      /* Extract X value and perform NULL screen */
      xeval = dmi_check_null_values(xdd, xrange, xptr, &xpos, &df, &paccept[cpt_offset] );

      if ( ydd )
      {
	/* Extract Y value and perform NULL screen */
        yeval = dmi_check_null_values(ydd, yrange, yptr, &ypos, &df, &paccept[cpt_offset] );
      }

      /* Check that value lies within range, if there is a range given */
      if ( xrange )
      {
        if (xrange->n > 0 )
	{
          /* Must be inside one of the ranges or row fails */
          if ( xeval )
          {
            /* For numeric types, eval now; other types already eval'ed */
            paccept[cpt_offset] = 0;
	
            for ( range_no = 0; range_no < xrange->n; range_no++ ) 
            {
              minval = ((double*)xrange->mins)[range_no];
              maxval =((double*)xrange->maxes)[range_no] + df;

              if ( xpos == minval || ( xpos > minval && xpos < maxval ))
              {
                paccept[cpt_offset] = 1;
	        break;
              }
            }
          }
 
          /* This test to be done even if xeval is false */
          if ( !paccept[cpt_offset] )
            break;
        }
        else
          if ( xrange->incnull && !paccept[cpt_offset] )
            break;
      } // end if range
   

      /* If there is a y column, check range also */
      if ( yrange && yrange->n && yeval )
      {
        paccept[cpt_offset] = 0;
	
        for ( range_no = 0; range_no < yrange->n; range_no++ )
        {
          minval = ((double*)yrange->mins)[range_no];
          maxval =((double*)yrange->maxes)[range_no] + df;

          if ( ypos == minval || ( ypos > minval && ypos < maxval ))
          {
            paccept[cpt_offset] = 1; break;
          }
        }

        /* This check can be inside the if(yeval) test since */
        /*  yeval==NULL is caught by checks above            */
        if ( !paccept[cpt_offset] )
          break;

      } /* End if yrange && yeval */       

      /* If there is a region, check against region */
      if ( region )
      {
        paccept[cpt_offset] = regInsideRegion( region, xpos, ypos );
 
        if ( !paccept[cpt_offset] )
          break;
      }
    } /* End for(index;...      */
  } /* End  for(cpt_offset;... */


  /* Check all acceptance values to determine final outcome */
  for ( cpt_offset = 0; cpt_offset < ncpts; cpt_offset++ )
    if ( paccept[cpt_offset] )
      accept = 1;
  
  /* Check for invert flag */
  if ( invert )
    accept = !accept;


  /* Final determination of this point */
  return accept;
}


/***************************************************************/
/*                                                             */
/* Function:   dmp_rowfilter_check_syntax                      */
/*                                                             */
/* Description:                                                */
/*    Note: returns UT_TRUE on error                           */
/*                                                             */
/*  Input parameters:                                          */
/*             dmParseList *vparse                             */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             dmBool  true-or-false                           */
/*                                                             */
/***************************************************************/
dmBool dmp_rowfilter_check_syntax( dmParseList* vparse )
{
 if ( vparse->binning )
 {
  dmu_err_set( DM_ERR_PARSE, "Error: Row filter does not accept DM binning syntax." );
  return UT_TRUE;
 }

 if ( vparse->select )
 {
  dmu_err_set( DM_ERR_PARSE, "Error: Row filter does not accept DM select syntax." );
  return UT_TRUE;
 }

 if ( vparse->option )
 {
  dmu_err_set( DM_ERR_PARSE, "Error: Row filter does not accept DM option syntax." );
  return UT_TRUE;
 }

 if ( vparse->block )
 {
  dmu_err_set( DM_ERR_PARSE, "Error: Row filter does not accept DM block selection syntax." );
  return UT_TRUE;
 }

 return UT_FALSE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_col_get_filtered_col               */
/*                                                             */
/* Description:                                                */
/*   This function determines if the filter is on the whole    */
/*    vector column or just a component of the column          */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor *sdd                               */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             dmDescriptor  *sdd                              */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_subspace_col_get_filtered_col( dmDescriptor* sdd )
{
 dmFilter *filter     = NULL; 
 long      dim        = 0;   /* dimensionality of the filter */
 long      vcptOffset = 0;   /* vector component offset */
 long      fvcptno    = 0;   /* the vector cpt number of the filter */

 filter = dmi_subspace_col_get_filter( sdd, 1 );
 dim = filter->dim;

 if ( filter->region || dim <= 1 )
  return sdd;

 fvcptno = 0;

 for ( vcptOffset = 0; vcptOffset < dim; vcptOffset++ )
 {
  if ( filter->ranges[vcptOffset].n || filter->ranges[vcptOffset].incnull )
  {
   if ( !fvcptno )
    fvcptno = vcptOffset+1;
   else
    return sdd;
  }
 }

 if ( fvcptno )
  return sdd->cpts[fvcptno-1];

 return sdd;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_check_null_values                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor  *dd                               */
/*             dmRangeList   *range                            */
/*             void          *ptr                              */
/*                                                             */
/*  Output parameters:                                         */
/*             double        *dpos                             */
/*             double        *df                               */
/*             dmBool        paccept                           */
/*                                                             */
/*  Returns:                                                   */
/*             dmBool  eval                                    */
/*                                                             */
/***************************************************************/
dmBool dmi_check_null_values( dmDescriptor* dd, dmRangeList* range,
                              void* ptr, double* dpos, double* df,
                              dmBool* paccept )
{
  void  *nullval = dd->null;
  dmBool isnull  = UT_FALSE;        /* is the value null? */
  long   ii      = 0;               /* counter */
  dmBool eval    = UT_TRUE;


  /// Return if no data row
  if ( ptr == NULL )
    return eval;

  // Check data against null values based on data type

  switch ( dd->type )
  {
    case DT_INT2:
      *dpos = (double)*((short*)ptr);
      *df   = 0.1;

      if ( range && range->incnull && nullval )
      {
        if ( *((short*)ptr) == *((short*)nullval) )
        {
	  if ( range->incnull == 1 )
	  {
	    *paccept = 1;
	     eval    = UT_FALSE;
	  }
	  else
	  {
	    *paccept = 0;
	     eval    = UT_FALSE;
	  }
        }
        else
          if ( range->incnull == -1 )
          {
	    *paccept = 1;
          }
      }
      break;

    case DT_INT4:
      *dpos = (double)*((long*)ptr);
      *df   = 0.1;

      if ( range && range->incnull && nullval )
      {
        if ( *((long*)ptr) == *((long*)nullval) )
        {
          if ( range->incnull == 1 )
          {
            *paccept = 1;
             eval    = UT_FALSE;
          }
          else
          {
            *paccept = 0;
             eval = 0;
          }
        }
        else
          if ( range->incnull == -1 )
          {
            *paccept = 1;
          }
      }
      break;

    case DT_REAL4:
      *dpos = (double)*((float*)ptr);
      *df   = 0.0;

      if ( range && range->incnull )
      {
        if ( isnan( *((float*)ptr) ))
          isnull = UT_TRUE;
        else
          if ( nullval && *((float*)ptr) == *((float*)nullval) )
            isnull = UT_TRUE;
          else 
            isnull = UT_FALSE;

        if ( isnull )
        {
          if ( range->incnull == 1 )
          {
            *paccept = 1;
             eval    = UT_FALSE;
          }
          else
          {
            *paccept = 0;
             eval    = UT_FALSE;
          }
        }
        else
          if ( range->incnull == -1 )
          {
            *paccept = 1;
          }
      }
      break;

    case DT_REAL8:
      *dpos = (double)*((double*)ptr);
      *df   = 0.0;
 
      if ( range && range->incnull )
      {
        if ( isnan( *((double*)ptr) ))
          isnull = UT_TRUE;
	else
          if ( nullval && *((double*)ptr) == *((double*)nullval) )
            isnull = UT_TRUE;
          else 
            isnull = UT_FALSE;

        if ( isnull )
        {
          if ( range->incnull == 1 )
          {
            *paccept = 1;
             eval    = UT_FALSE;
          }
          else
          {
            *paccept = 0;
             eval    = UT_FALSE;
          }
        }
        else
          if ( range->incnull == -1 )
          {
            *paccept = 1;
          }
      }
      break;

    case DT_UINT1:
      *dpos = (double)*((unsigned char*)ptr);
      *df   = 0.1;

      if ( range && range->incnull && nullval )
      {
        if ( *((unsigned char*)ptr) == *((unsigned char*)nullval) )
        {
          if ( range->incnull == 1 )
          {
            *paccept = 1;
             eval    = UT_FALSE;
          }
          else
          {
            *paccept = 0;
             eval    = UT_FALSE;
          }
        }
        else
          if ( range->incnull == -1 )
          {
            *paccept = 1;
          }
      }
      break;

    case DT_UINT2:
      *dpos = (double)*((unsigned short*)ptr);
      *df   = 0.1;

      if ( range && range->incnull && nullval )
      {
        if ( *((unsigned short*)ptr) == *((unsigned short*)nullval) )
        { 
          if ( range->incnull == 1 )
          {
            *paccept = 1;
             eval    = UT_FALSE;
          }
          else
          {
            *paccept = 0;
             eval    = UT_FALSE;
          }
        }
        else
          if ( range->incnull == -1 )
          {
            *paccept = 1;
          }
      }
      break;

    case DT_UINT4:
      *dpos = (double)*((unsigned long*)ptr);
      *df   = 0.1;

      if ( range && range->incnull && nullval )
      {
        if ( *((unsigned long*)ptr) == *((unsigned long*)nullval) )
        {
          if ( range->incnull == 1 )
          {
            *paccept = 1;
             eval    = UT_FALSE;
          }
          else
          {
            *paccept = 0;
             eval    = UT_FALSE;
          }
        }
        else
          if ( range->incnull == -1 )
          {
            *paccept = 1;
          }
      }
      break;

    case DT_LOGICAL:
      *dpos = (double)*((dmBool*)ptr);
      *df   = 0.1;

      if ( range && range->incnull && nullval )
      {
        if ( *((dmBool*)ptr) == *((dmBool*)nullval) )
        {
          if ( range->incnull == 1 )
          {
            *paccept = 1;
             eval    = UT_FALSE;
          }
          else
          {
            *paccept = 0;
             eval    = UT_FALSE;
          }
        }
        else
          if ( range->incnull == -1 )
            *paccept = 1;
      }
      break;

    case DT_CHAR:
      /* This is not yet implemented correctly. Of course, we don't */
      /* currently have any files with string vector columns in.    */
      /* You'd need to offset ptr with the string length.           */

      if ( range )
      {
        if ( range->incnull )
        {
          if ( *((char*)ptr) == '\0' )
            isnull = UT_TRUE;
          else
            if ( nullval && !strcmp(((char*)ptr), ((char*)nullval) ))
              isnull = UT_TRUE;
            else 
              isnull = UT_FALSE;

          if ( isnull )
          {
            if ( range->incnull == 1 )
            {
	      *paccept = 1;
	       eval    = UT_FALSE;
            }
            else
            {
	     *paccept = 0;
	      eval    = UT_FALSE;
            }
          }
          else
            if ( range->incnull == -1 )
              *paccept = 1;
        }

        if ( eval )
        {
          *paccept = 0; /* Will fail unless one matches */

          for ( ii = 0; ii < range->n; ii++ )
          {
            if ( dmu_text_in_range( ptr,
                                   ((char**)range->mins)[ii], ((char**)range->maxes)[ii] ))
            {
	      *paccept = 1;
              break;  
            }
          }
        }

        eval = UT_FALSE;
      }
      break;

    case DT_BIT:
      /* This is not yet implemented correctly. Of course, we don't  */
      /* Currently have any files with bit vector columns in.        */
      /* You'd need to offset ptr with the byte length of a field    */

      if ( range ) 
      {
        if ( range->incnull && nullval )
        {
          long nbytes = dd->length >> 3;

          if ( !memcmp( ptr, nullval, nbytes ))
          {
            if ( range->incnull == 1 )
            {
	      *paccept = 1;
	       eval    = UT_FALSE;
            }
            else
            {
	      *paccept = 0;
	       eval    = UT_FALSE;
            }
          }
          else
            if ( range->incnull == -1 )
              *paccept = 1;
        }
     
        if ( eval )
        {
          *paccept = 0; /* Will fail unless one matches */

          for ( ii = 0; ii < range->n; ii++ )
          {
            if ( dmu_bit_match( ptr, dd->length, ((char**)range->mins)[ii] ))
            {
	      *paccept = 1;
              break;
            }
          }
        }
      }

      eval = UT_FALSE;
      break;

    default:
      printf( "NO DT support\n" );
  }    

  return eval;
}
