/*                                                                
**  Copyright (C) 1997-2005,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/*****************************************************************/
/*                                                               */
/*   dmtable.c                                                   */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmlibP.h"



/***************************************************************/
/*                                                             */
/* Function:  dmColumnGetRowOffset                             */
/*                                                             */
/* Description:                                                */
/*   Determine and return the offset into a row (in bytes)     */
/*   to the beginning of the value(s) for the given column.    */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor  *dd                                 */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/* Returns:                                                    */
/*   long  offset  - byte offset to data value(s)              */
/*                                                             */
/***************************************************************/
long dmColumnGetRowOffset( dmDescriptor* dd )
{
  long col_offset  = 0; 
  long item_offset = 0;
  long offset      = 0;
  long cptOffset   = 0;
  long cptMax      = 0;
  long nvals       = 0;
  long nbytes      = 0;

  dmBlock       *block = NULL;
  dmBuffer      *buffer = NULL;
  dmDescriptor **cpts   = NULL;

  if ( !dd )
    return -1;

  block  = (dmBlock*)dd->block;
  buffer = block->buffer;
  col_offset = dd->bufno-1;

  if ( col_offset >= 0 )
  {
    /* We're dealing with visible columns only, so use voffsets */
    /* rather* than offsets                                     */

    /* Offset to beginning of the column */
    offset = buffer->voffsets[col_offset];
  
    if ( dd->cptno > 0 && dd != buffer->cols[col_offset] )
    {
      /* This is not the first component of the column  */
      /* add offset to value(s) for this component.     */

      cptMax = dd->cptno-1;
      item_offset = 0;
      cpts = dd->parent->cpts;

      for ( cptOffset = 0; cptOffset < cptMax; cptOffset++ )
      {
	/* number of values   */
	nvals  = dmi_descriptor_get_nvals( cpts[cptOffset] );

	/* size of each value */
	nbytes = dmi_descriptor_get_bytes( cpts[cptOffset] );

	/* add offset for this component */
	item_offset += nvals * nbytes ;
      }
      offset += item_offset;
    }
  }
  else
    offset = -1;

  return offset;
}



/***************************************************************/
/*                                                             */
/* Function:   dmColumnGetRowPointer                           */
/*                                                             */
/* Description:                                                */
/*   Determine and return pointer to start of data value(s)    */
/*   within provided row, for the given column.                */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor  *dd                                 */
/*           void          *row                                */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/* Returns:                                                    */
/*   (void*) pointer to data value(s) within row               */
/*                                                             */
/***************************************************************/
void* dmColumnGetRowPointer(dmDescriptor* dd, void *row )
{
  long offset = 0;
 
  offset = dmColumnGetRowOffset( dd );

  if ( offset >= 0 )
    return ((char*)row + offset );

  return NULL; 
}



/***************************************************************/
/*                                                             */
/* Function:   dmDatasetCreateTable                            */
/*                                                             */
/* Description:                                                */
/*   Create a table in an existing dataset                     */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDataset *ds                                     */
/*     const char      *name                                   */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/* Returns:                                                    */
/*   dmBlock* descriptor for the table.                        */
/*                                                             */
/***************************************************************/
dmBlock* dmDatasetCreateTable( dmDataset* ds, const char* name )
{
  return dmp_block_open_mode( ds, name, 0, dmTABLE, NULL, "W" );
}



/***************************************************************/
/*                                                             */
/* Function:   dmTableAllocRow                                 */
/*                                                             */
/* Description:                                                */
/*   Allocate memory large enough to hold one row.             */  
/*                                                             */
/*  Input parameters:                                          */
/*           dmBlock *tab                                      */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/* Returns:                                                    */
/*          void *RowPointer                                   */
/*                                                             */
/***************************************************************/
void* dmTableAllocRow(dmBlock *tab)
{                
  long size = dmTableGetRowSize(tab);
  
  if (size > 0)
     return malloc(size);
  else
     return NULL;
}



/***************************************************************/
/*                                                             */
/* Function:  dmTableClose                                     */
/*                                                             */
/* Description:                                                */
/*   Close provided DM block and associated dataset.           */
/*                                                             */
/*  Input parameters:                                          */
/*      const dmBlock  *block                                  */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/* Returns:                                                    */
/*   dmErrCode  status  - DM Error status                      */
/*                                                             */
/***************************************************************/
dmErrCode dmTableClose( dmBlock* block )
{
  dmDataset *ds = NULL;

  int  status = 0; 

  ds = dmi_block_get_ds( block );

  status = dmp_block_close( block );

  if ( status == dmSUCCESS )
    status = dmDatasetClose( ds );

 return status;
}



/***************************************************************/
/*                                                             */
/* Function:   dmTableCreate                                   */
/*                                                             */
/* Description:                                                */
/*   Create the specified table                                */
/*                                                             */
/*  Input parameters:                                          */
/*           char  *vspec                                      */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/* Returns:                                                    */
/*   dmBlock* descriptor for the table.                        */
/*                                                             */
/***************************************************************/
dmBlock* dmTableCreate(char* vspec)
{
  return dmp_block_open_mode( NULL, vspec, 0, dmTABLE, NULL, "W" );
}


/***************************************************************/
/*                                                             */
/* Function:  dmTableCreateColumns                             */
/*                                                             */
/* Description:                                                */
/*   Create and return list of DM Table columns based on       */
/*   provided specifications.                                  */
/*                                                             */
/*  Input parameters:                                          */
/*           dmBlock     *table                                */
/*           char       **names                                */
/*           dmDataType   types                                */
/*           long        *strLens                              */
/*           char       **units                                */
/*           char       **descs                                */
/*           long         numCols                              */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/* Returns:                                                    */
/*            dmDescriptor **list                              */
/*                                                             */
/***************************************************************/
dmDescriptor**  dmTableCreateColumns(dmBlock *table,
				     char **names,
				     dmDataType* types,
				     long *strLens,
				     char **units,
				     char **descs,
				     long numCols)
{
  long ii = 0;
  dmDescriptor  **list = NULL;

  if ( numCols > 0 )
  {
    list = calloc( numCols, UT_SZ_P );

    for ( ii = 0; ii < numCols; ii++ )
    {
      list[ii] = dmColumnCreate( table, names[ii], types[ii],
				strLens[ii], units[ii], descs[ii]);
    }
  }

  return list;
}



/***************************************************************/
/*                                                             */
/* Function:  dmTableGetNoCols                                 */
/*                                                             */
/* Description:                                                */
/*   Get number of table columns defined in the provided block */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmBlock *block                                    */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/* Returns:                                                    */
/*   long  ncols  - number of table columns                    */
/*                                                             */
/***************************************************************/
long dmTableGetNoCols( const dmBlock* block )
{
  long ncols = 0;

  if ( block )
    ncols = block->ncols;

  return ncols;
}



/***************************************************************/
/*                                                             */
/* Function:  dmTableGetNoRows                                 */
/*                                                             */
/* Description:                                                */
/*   Determine, if necessary, and return the number of actual  */
/*   rows in the table.                                        */
/*                                                             */
/*  Input parameters:                                          */
/*      const dmBlock  *block                                  */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/* Returns:                                                    */
/*   long nvrows  - Number of table rows                       */
/*                                                             */
/***************************************************************/
long dmTableGetNoRows(dmBlock* block)
{
  long nvrows  = 0;
  long row     = 0;
  long old_row = 0;

  if ( !block )
    return 0;

  if ( block->virtual && !block->traversed )
  {
    /* Store current position */
    old_row = dmi_block_get_current_row( block );

    /* Attempt to read the number of raw rows; there may be many fewer */
    row = block->nrows;

    nvrows = dmp_buffer_adjust( block,  1, 1, UT_FALSE );

    if ( nvrows != DM_BUF_BADROW && !block->traversed )
      nvrows = dmp_buffer_adjust( block,  row, 1, UT_FALSE );

    block->traversed = 1;  

    /* Reset table to old position */
    dmTableSetRow( block, old_row );  
  }
 
  return block->nvrows;
}



/***************************************************************/
/*                                                             */
/* Function:  dmTableGetRow                                    */
/*                                                             */
/* Description:                                                */
/*   Load current table row into provided pointer              */
/*                                                             */
/*  Input parameters:                                          */
/*           dmBlock *block                                    */
/*                                                             */
/*  Output parameters:                                         */
/*           void    *row                                      */
/*                                                             */
/* Returns:                                                    */
/*            long rowno                                       */
/*                                                             */
/*   row number   = success                                    */
/*   dmNOMOREROWS = at end of data or problem                  */
/*                                                             */
/***************************************************************/
long dmTableGetRow(dmBlock* block, void *row)
{
  long nvrows = 0;
  long rowno  = 0;
 
  if ( block )
  {
    rowno = dmi_block_get_current_row( block );

    if ( rowno < 0 ) /* Bad row */
      return dmNOMOREROWS;

    if( rowno == 0 )
      rowno = 1;

    nvrows = dmp_buffer_adjust( block, rowno, 1, UT_FALSE );
 
   if ( rowno <= block->nvrows &&  nvrows >= 1 && row )
      dmp_copy_buffer_to_row( block, row, rowno );    
    else
      return dmNOMOREROWS;
  }

  dmi_block_set_current_row( block, rowno+1 );

  return rowno;
}



/***************************************************************/
/*                                                             */
/* Function:  dmTableGetRows                                   */
/*                                                             */
/* Description:                                                */
/*   Allocate and return a list of pointers to table rows      */
/*                                                             */
/*  Input parameters:                                          */
/*           dmBlock *block                                    */
/*                                                             */
/*  Output parameters:                                         */
/*           long    *nrowsp                                   */
/*                                                             */
/* Returns:                                                    */
/*            void **rows                                      */
/*                                                             */
/***************************************************************/
void** dmTableGetRows(dmBlock* block, long* nrowsp)
{
  void **rows = NULL;

  long ii    = 0;
  long size  = dmTableGetRowSize(block);
  long nrows = *nrowsp;

  if ( nrows < 1 )
    nrows = dmi_buffer_eval_size( block, nrows );

  *nrowsp = nrows;

  /* Allocate list array */
  rows = malloc( nrows * UT_SZ_P );

  for ( ii = 0; ii < nrows; ii++ )
  {
    /* Allocate and file table row */
    rows[ii] = malloc( size );
    dmTableGetRow( block, rows[ii] );
  }

 return rows;
}



/***************************************************************/
/*                                                             */
/* Function:  dmTableGetRowNo                                  */
/*                                                             */
/* Description:                                                */
/*   Get the current row number from the given block.          */
/*                                                             */
/*  Input parameters:                                          */
/*           dmBlock *block                                    */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/* Returns:                                                    */
/*   long row - current row number                             */
/*                                                             */
/***************************************************************/
long dmTableGetRowNo(dmBlock* block )
{
  long row;

  row = dmi_block_get_current_row( block );

  if ( row == 0 )
    row = 1;

  return row;
}



/***************************************************************/
/*                                                             */
/* Function:  dmTableGetRowSize                                */
/*                                                             */
/* Description:                                                */
/*   Return size of each table row                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmBlock *block                                    */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/* Returns:                                                    */
/*          long rowsize                                       */
/*                                                             */
/***************************************************************/
long dmTableGetRowSize( dmBlock* block )
{
  dmBuffer *buffer = NULL;

  if ( !block || !block->buffer)
    return 0;

  buffer = block->buffer;

  return buffer->vtotal;
}



/***************************************************************/
/*                                                             */
/* Function:   dmTableOpen                                     */
/*                                                             */
/* Description:                                                */
/*   Open the specified table                                  */
/*                                                             */
/*  Input parameters:                                          */
/*           char  *vspec                                      */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/* Returns:                                                    */
/*   dmBlock* descriptor for the table.                        */
/*                                                             */
/***************************************************************/
dmBlock* dmTableOpen( char* vspec )
{
  return dmp_block_open_mode( NULL, vspec, 0, dmTABLE, NULL, "R" );
}



/***************************************************************/
/*                                                             */
/* Function:   dmTableOpenColumn                               */
/*                                                             */
/*   Extract and return, if avail., the descriptor matching    */
/*   the specified column name  .                              */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmBlock *block                                    */
/*     const char    *colname                                  */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/* Returns:                                                    */
/*   Descriptor* for specified (by name) column                */
/*   NULL if not found.                                        */
/*                                                             */
/***************************************************************/
dmDescriptor* dmTableOpenColumn( const dmBlock* block,
				 const char* colname )
{
  dmDescriptor* dd = NULL;

  if ( !block )
    return NULL;

  /* Optimize efficient search for common case */
  if ( colname && block )
  {
    long ii;
    for ( ii = 0; ii < block->ncols; ii++ )
    {
      dd = block->cols[ii];
      if ( cs_uceq( dd->name, colname ) )
	return dd;
    }
    dd = NULL;  
  }

  /* Default case.. only have one  */
  if ( !colname && block->ncols == 1 )
  {
    dd = block->cols[0];
    return dd;
  }

  /* Execute a more thorough search.. */
  dd = dmi_block_search_public_identifier( block, colname );

  return dd;
}



/***************************************************************/
/*                                                             */
/* Function:  dmTableOpenColumnList                            */
/*                                                             */
/* Description:                                                */
/*   Allocate and return a list of column descriptors from     */
/*   the provided block.                                       */
/*                                                             */
/*  Input parameters:                                          */
/*      const dmBlock  *block                                  */
/*            long     *ncolsp                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/* Returns:                                                    */
/*   dmDescriptor *list                                        */
/*   NULL if no columns found.                                 */
/*                                                             */
/***************************************************************/
dmDescriptor**  dmTableOpenColumnList(const dmBlock* block, long* ncolsp)
{
  long ii    = 0;
  long ncols = 0;

  dmDescriptor **cols = NULL;

  ncols = block->ncols;

  if ( ncols > 0 ) 
  {
    cols = calloc( ncols, UT_SZ_P );

    for ( ii = 0; ii < ncols; ii++ ) 
      cols[ii] = block->cols[ii];
  }

  *ncolsp = ncols;

  return cols;
}



/***************************************************************/
/*                                                             */
/* Function:   dmTableOpenColumnNo                             */
/*                                                             */
/* Description:                                                */
/*   Extract and return, if avail., the descriptor matching    */
/*   the specified column number.                              */
/*                                                             */
/*  Input parameters:                                          */
/*    const dmBlock *block                                     */
/*    const long     no                                        */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/* Returns:                                                    */
/*   Descriptor* for specified (by number) column              */
/*   NULL if not found.                                        */
/*                                                             */
/***************************************************************/
dmDescriptor* dmTableOpenColumnNo( const dmBlock* block, const long no )
{
  dmDescriptor* dd = NULL;

  if ( !block )
    return NULL;

  /* Check if requested column is within range */
  if ( no > 0  && no <= block->ncols )  
    dd = block->cols[no-1];

  return dd;
}



/***************************************************************/
/*                                                             */
/* Function:   dmTableOpenUpdate                               */
/*                                                             */
/* Description:                                                */
/*   Open the specified table, write enabled.                  */
/*                                                             */
/*  Input parameters:                                          */
/*           char  *vspec                                      */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/* Returns:                                                    */
/*   dmBlock* descriptor for the table.                        */
/*                                                             */
/***************************************************************/
dmBlock* dmTableOpenUpdate(char* vspec)
{
 return dmp_block_open_mode( NULL, vspec, 0, dmTABLE, NULL, "RW" );
}  



/***************************************************************/
/*                                                             */
/* Function:  dmTableNextRow                                   */
/*                                                             */
/* Description:                                                */
/*   Set block to next table row                               */
/*                                                             */
/*  Input parameters:                                          */
/*           dmBlock *block                                    */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/* Returns:                                                    */
/*          long retval                                        */
/*                                                             */
/*   row number   = success                                    */
/*   dmNOMOREROWS = at end of data                             */
/*   dmBADROW     = problem                                    */
/*                                                             */
/***************************************************************/
long dmTableNextRow(dmBlock* block)
{
  long rowno  = 0;
  long retval = dmNOMOREROWS;

  if ( !block )
    return retval;

  /* Get current row number */
  dmi_block_reset_current_row( block );
  rowno  = dmi_block_get_current_row( block );

  /* Increment and set block to next row */
  ++rowno;

  retval = dmTableSetRow(block, rowno);

  return retval;
}


  
/***************************************************************/
/*                                                             */
/* Function:  dmTablePutRow                                    */
/*                                                             */
/* Description:                                                */
/*   Write provided row to DM buffer                           */
/*                                                             */
/*  Input parameters:                                          */
/*           dmBlock *block                                    */
/*           void    *row                                      */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/* Returns:                                                    */
/*           long nwrote - row number                          */
/*                                                             */
/***************************************************************/
long dmTablePutRow(dmBlock* block, void *row)
{
  long rowno  = 0;
  long nvrows = 0;
  long nwrote = dmBADROW;

  if ( block )
  {
    rowno = dmi_block_get_current_row( block );
    if( rowno == 0 )
      rowno = 1;

    nvrows = dmp_buffer_adjust( block, rowno, 1, UT_TRUE );
    if ( nvrows >= 1 )
    {
      if ( row ) 
	dmi_copy_row_to_buffer( block, row, rowno );

      nwrote = rowno;
      if ( rowno > block->nrows )
	block->nrows++;

      if ( rowno > block->nvrows )
	block->nvrows++;

      dmi_block_set_current_row( block, rowno+1 );
    } 
    else 
    {
      sprintf( dm_errbuf, "Failed to write row %ld", rowno );
      dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );   
    }
  }
 
  return nwrote;
}



/***************************************************************/
/*                                                             */
/* Function:  dmTableSetRow                                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*           dmBlock *block                                    */
/*           long     row                                      */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/* Returns:                                                    */
/*           long retval                                       */
/*                                                             */
/***************************************************************/
long dmTableSetRow(dmBlock* block, long row)
{

  long nactual = 0;
  int  retval  = dmBADROW;
  long newRow  = 0;
  
  if (!block)
    return retval;


  if (row > block->nrows || row < 0 )
  {
    /* We are allowed to set the row beyond nrows if we are */
    /*  writing, but not if we are reading.                 */

    if( dmi_block_writeable( block ))
      newRow = block->nvrows + 1;
    else
      newRow = dmBADROW;
      
    retval = dmNOMOREROWS;
  }
  else if (row < 1)
  {
    newRow = 1;
    retval = 1;
  }
  else
  {
    /* Here we force  read the row if needed, checking for the
     * end of a filtered file
     */
    nactual= dmp_buffer_adjust( block,  row, 1,UT_FALSE );
      
    if ( nactual > 0 && row <= block->nvrows )
    {
      newRow = row;
      retval = row;
    }
    else 
    {
      newRow = block->nvrows + 1;
      retval = dmNOMOREROWS;
    }  
  }
   
  dmi_block_set_current_row( block, newRow );

  return retval;
}




/****************************************************************************/
/*                          DEPRECATED FUNCTIONS                            */
/****************************************************************************/

/***************************************************************/
/*                                                             */
/* Function:  		dmTableGetColOffset                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*           dmBlock    *tab                                   */
/*           long        colno                                 */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/* Returns:                                                    */
/*            short offset                                     */
/*                                                             */
/***************************************************************/
short  dmTableGetColOffset(dmBlock *tab, long colno)
{
  dmDescriptor *dd = NULL;
  long offset = 0;

  dd = dmi_block_get_scalar_col( tab, colno );
  offset  = dmColumnGetRowOffset( dd );

  return (short)offset;
}



/***************************************************************/
/*                                                             */
/* Function:  dmTableGetColPtr                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*           dmBlock    *tab                                   */
/*           void       *row                                   */
/*           long        colno                                 */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/* Returns:                                                    */
/*          void *ColPointer                                   */
/*                                                             */
/***************************************************************/
void*  dmTableGetColPtr(dmBlock *tab, void *row, long colno)
{
  dmDescriptor *dd = NULL;

  dd = dmi_block_get_scalar_col( tab, colno );

  return dmColumnGetRowPointer( dd, row );
}



/***************************************************************/
/*                                                             */
/* Function:   dmTableOpenSelect                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDataset  *pDs                                   */
/*           char       *tabName                               */
/*           char       *selectCols                            */
/*           dmDataType *castToType                            */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/* Returns:                                                    */
/*          dmBlock *pBl                                       */
/*                                                             */
/***************************************************************/
dmBlock* dmTableOpenSelect(dmDataset *pDs,
			   char *tabName, 
			   char *selectCols,
			   dmDataType *castToType)
{
  dmBlock *pBl  = NULL;
  char    *name = NULL;
  long     m    = 0;

  if ( !pDs || !tabName )
    return NULL;

  if ( tabName && selectCols)                   
  {
    m    = strlen( selectCols ) + strlen( tabName ) + 12;
    name = (char*)malloc( m * UT_SZ_C );

    sprintf( name, "[%s][cols %s]", tabName, selectCols );
  }
  else
  {
    name = dmStrDup( tabName );
  }

  pBl = dmp_block_open_mode( pDs, name, 0, dmTABLE, NULL, "R" );
  free( name );

  if ( castToType ) /* Back compatibility */
  {
    dmu_err_set( DM_ERR_GENERIC, "TableOpenSelect: casting not supported" );
  }

  return pBl;
}
