/*                                                                
**  Copyright (C) 2002-2005,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmcoord.c                                                   */
/*                                                               */
/*   Functions that access dmCoord descriptors                   */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmlibP.h"


/***************************************************************/
/*                                                             */
/* Function:   dmCoordSetParam                                 */
/*                                                             */
/* Description:                                                */
/*     Sets coordinate descriptor elements                     */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *dd   - coordinate descriptor             */
/*     char          *name - element name                      */
/*     double         rot  - value                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void  dmCoordSetParam( dmDescriptor* dd, char* name, double rot )
{
  dmCoordTransform* t;

  if ( !dd ) return;

  if ( cs_uceq( name, "LONGPOLE" ))
  {
    dmi_coord_set_longpole( dd, rot );
  }
  else if ( cs_uceq( name, "ROT" ))
  {
    dmi_coord_set_theta( dd, rot );
  }
  else if ( cs_uceq( name, "LATPOLE" ))
  {
    dmi_coord_set_latpole( dd, rot );
  }
  else if ( cs_uceq( name, "EQUINOX" ))
  {
    dmi_coord_set_equinox( dd, rot );
  }

  /* Reinit wcs */
  t = dmi_descriptor_get_aux( dd );
  dmi_coord_mink_coat( t, dd);
}


/***************************************************************/
/*                                                             */
/* Function:   dmCoordGetParent                                */
/*                                                             */
/* Description:                                                */
/*     Return parent descriptor                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *dd   - coordinate descriptor             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor  *parent                             */
/*                                                             */
/***************************************************************/
dmDescriptor*  dmCoordGetParent( dmDescriptor* dd )
{
  return dmi_coord_get_parent( dd );
}


/***************************************************************/
/*                                                             */
/* Function:   dmCoordGetParams                                */
/*                                                             */
/* Description:                                                */
/*     Returns coordinate parameters                           */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *dd   - coordinate descriptor             */
/*                                                             */
/*  Output parameters:                                         */
/*     long    *nparams    - number of parameters returned     */
/*                                                             */
/*  Returns:                                                   */
/*           double  *params                                   */
/*                                                             */
/***************************************************************/
double*  dmCoordGetParams(dmDescriptor* dd, long* nparams)
{
  long    np = 4;       /* MAX PARAMS*/
  double *params = calloc( np, UT_SZ_D );

  *nparams = dmi_coord_get_params( dd, params );
  return params;
}


/***************************************************************/
/*                                                             */
/* Function:   dmCoordGetTransform_d                           */
/*                                                             */
/* Description:                                                */
/*     Returns coordinate crval, cdelt, and dim                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *dd    - coordinate descriptor            */
/*     long           dim   - length                           */
/*                                                             */
/*  Output parameters:                                         */
/*     double        *crpix - pix                              */
/*     double        *crval - val                              */
/*     double        *cdelt - delta                            */
/*                                                             */
/*  Returns:                                                   */
/*           dmErrCode  - dmSUCCESS or dmFAILURE               */
/*                                                             */
/***************************************************************/
dmErrCode  dmCoordGetTransform_d(dmDescriptor* dd, double* crpix, 
				 double* crval, double* cdelt, long dim)
{
  long i;

  if ( dd && dd->descType == dmCOORD )
  {
    dmi_coord_get_transform_d( dd, crpix, crval, cdelt, dim );
    return dmSUCCESS;
  }

  for ( i = 0; i < dim; i++ )
  {
    crpix[i] = 0.0;
    crval[i] = 0.0;
    cdelt[i] = 1.0;
  }
  return dmFAILURE;  
}


/***************************************************************/
/*                                                             */
/* Function:   dmCoordGetTransformType                         */
/*                                                             */
/* Description:                                                */
/*     Returns transform type                                  */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *dd     - coordinate descriptor           */
/*     long           maxlen - max length of copy string       */
/*                                                             */
/*  Output parameters:                                         */
/*     char          *ttype - transform type                   */
/*     long          *nparams - coord parameters               */
/*                                                             */
/*  Returns:                                                   */
/*           dmErrCode  - dmSUCCESS or dmFAILURE               */
/*                                                             */
/***************************************************************/
dmErrCode   dmCoordGetTransformType(dmDescriptor* dd, char* ttype, 
		 		    long* nparams, long maxlen)
{
  char *dd_ttype = NULL;
  double params[4];

  if ( !dd || dd->descType != dmCOORD )
  {
    cs_copy( " ", ttype );
    return dmFAILURE;
  }

  *nparams = dmi_coord_get_params( dd, params );
  dd_ttype = dmi_coord_get_transform_type( dd );
  cs_copy_siz( dd_ttype, ttype, maxlen );
  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   dmBlockCopyWCS                                  */
/*                                                             */
/* Description:                                                */
/*     Copy WCS information                                    */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmBlock       *block1 - from block                      */
/*                                                             */
/*  Output parameters:                                         */
/*     dmBlock       *block2 - to block                        */
/*                                                             */
/*  Returns:                                                   */
/*           dmErrCode  - dmSUCCESS or dmFAILURE               */
/*                                                             */
/***************************************************************/
int dmBlockCopyWCS(dmBlock* block1, dmBlock* block2 )
{
  long colOffset = 0;

  if ( !block1 || !block2 )
    return dmFAILURE;

  for ( colOffset = 0; colOffset < block1->ncols; colOffset++ )
  {                    
    dmi_coord_list_copy( block1->cols[colOffset], block2->cols[colOffset] );
  }

  /* Will flush at block close time*/
  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   dmCoordGetAxisGroupNo                           */
/*                                                             */
/* Description:                                                */
/*     Return 0 for a world coordinate for compatibility with  */
/*     old DM, used in dmlist                                  */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *dd    - coordinate descriptor            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           long  no   - group number                         */
/*                                                             */
/***************************************************************/
long  dmCoordGetAxisGroupNo(dmDescriptor* dd)
{
  if ( dd && ( dd->flag & DM_FLAG_AXIS ))
  {
    if ( ( dd->flag & DM_FLAG_PHYS ) || ( dd->flag & DM_FLAG_LOGICAL ) )
      return dmi_coord_get_group_no( dd );
  } 
  return 0;
}

/***************************************************************/
/*                                                             */
/* Function:   dmi_block_ckeck_cols_for_vector                 */
/*                                                             */
/* Description:                                                */
/*    Returns dmTRUE if block contains a column with the same  */
/*    name as vector or one of its components, dmFALSE         */
/*    otherwise.  Used below in dmp_block_flush_coords().      */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmBlock      *block  - block                      */
/*     const dmDescriptor *vector - vector                     */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           static dmBool - dmTRUE or dmFALSE                 */
/*                                                             */
/***************************************************************/
static dmBool dmi_block_check_cols_for_vector( const dmBlock* block,
                                               const dmDescriptor* vector )
{
  dmDescriptor *col_dd = NULL;
  long          i = 0;

  if ( !block || !vector )
    return dmFALSE;

  col_dd = dmi_descriptor_list_search_vector( (const dmDescriptor**)block->cols, block->ncols, vector->name );
 
  if ( col_dd )
    return dmTRUE;

  if ( vector->dim > 1 )
  {
    for ( i = 0; i < vector->dim; i++ )
    {
      col_dd = dmi_descriptor_list_search_vector( (const dmDescriptor**)block->cols, block->ncols, vector->cpts[i]->name );

      if ( col_dd )
        return dmTRUE;
    }
  }

  return dmFALSE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_flush_coords                          */
/*                                                             */
/* Description:                                                */
/*    Loop through all the associated descriptors of the       */
/*    input block and close and write and associted WCS's for  */
/*    each descriptor.                                         */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmBlock      *block  - block                            */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           int status    - dmSUCCESS or dmFAILURE            */
/*                                                             */
/***************************************************************/
int dmp_block_flush_coords(dmBlock* block)
{
  dmDescriptor  *axis_dd  = NULL;
  dmDescriptor  *col      = NULL;
  dmDescriptor  *coord_dd = NULL;
  dmKernel       kernel;

  long k = 0;
  long i = 0;
  long j = 0;
  int is_unit = 0;

  if ( !dmi_block_writeable( block ))
    return dmSUCCESS;

  kernel =  dmi_block_kernel( block );

  for ( i = 0; i < block->ncols; i++ )
  {
    col = block->cols[i];
  
    /* We don't write out a coord if there's already a column with the same */
    /* name.  This is needed when writing virtual columns.  Is there a      */
    /* better way (i.e. detaching the transforms from which the virtual     */
    /* columns were calculated)?                                            */

    for ( j = 0; j < col->ncoords; j++ )
    {
      coord_dd = col->coords[j];

      if( dmi_descriptor_check_update( coord_dd ) && !dmi_block_check_cols_for_vector( block, coord_dd ) )
      {
        kernel->CoordWrite( coord_dd );
        dmi_descriptor_clear_update( coord_dd );
      }
    }


    for ( j = 0; j < col->ngroups; j++ )
    {
      axis_dd = dmi_descriptor_get_group_no( col, j+1 );      

      if ( axis_dd )
      {        
        if( dmi_descriptor_check_update( axis_dd ) && !dmi_block_check_cols_for_vector( block, axis_dd ) )
        {
          is_unit = !dmi_coord_is_default_axis( axis_dd );
          is_unit = ( is_unit && dmi_coord_transform_is_unit( axis_dd ));

          /* A lone group descriptor which is the only thing binding
           * multiple axes, needs to remain */

          is_unit = ( is_unit && ( axis_dd->ncoords || axis_dd->dim <= 1 ));

          /* The physical transform on an image should always be flushed
           * so that the name of the axis group (e.g. sky) isn't lost.  However,
           * we don't flush it if it's flagged with DM_FLAG_TEMP, as this means
           * it was automatically added by the DM. */

          is_unit = ( is_unit && !( ( dmi_block_get_type( block ) == dmIMAGE ) && ( axis_dd->flag & DM_FLAG_PHYS ) && ( ~axis_dd->flag & DM_FLAG_TEMP ) ) );

          if ( !is_unit )
          {
            /* This could be a temporary grouping specifically for a filtering
             * or binning operation.  Just flush its children */

            if (!( axis_dd->flag & DM_FLAG_TEMP ))
              kernel->CoordWrite( axis_dd );
            else if ( axis_dd->dim > 1 )
            {
              for ( k = 0; k < axis_dd->dim; k++ )
	      kernel->CoordWrite( axis_dd->cpts[k] );
            }
          }
          dmi_descriptor_clear_update( axis_dd );
        }
    
        for ( k = 0; k < axis_dd->ncoords; k++ )
        {
          coord_dd = axis_dd->coords[k];
     
          if( dmi_descriptor_check_update( coord_dd ) && !dmi_block_check_cols_for_vector( block, coord_dd ) )
          {
            kernel->CoordWrite( coord_dd );
            dmi_descriptor_clear_update( coord_dd );
          }
        }
      } // end if ( axis_dd )
    } // end for ( j = 0; j < col->ngroups; j++ )
  } // end  for ( i = 0; i < block->ncols; i++ )

  return dmSUCCESS;
}


