/*                                                                
**  Copyright (C) 1998,2003-2004,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmcoord.c                                                   */
/*                                                               */
/*   Functions that create dmCoord descriptors                   */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/
#include "dmlibP.h"


/***************************************************************/
/*                                                             */
/* Function:   dmCoordCreate_s                                 */
/*                                                             */
/* Description:                                                */
/*    Create a coordinate system on the input descriptor.      */
/*    Return the coordinate system descriptor.                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *dd        - coordinate descriptor        */
/*     char          *name      - name                         */
/*     char          *unit      - units                        */
/*     char         **cptNames  - component names              */
/*     long           dim       - length                       */
/*     char          *transform -                              */
/*     short         *crpix     -                              */
/*     double        *crval     -                              */
/*     double        *cdelt     -                              */
/*     double        *params    -                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *retval                              */
/*                                                             */
/***************************************************************/
dmDescriptor* dmCoordCreate_s(dmDescriptor* dd, char* name, char* unit,
			      char** cptNames, long dim, char* transform,
			      short* crpix, double* crval, double* cdelt,
			      double* params)
{
  dmDescriptor* retval  = NULL;
  double*  darray = NULL;
     
  darray = malloc( dim * UT_SZ_D );
  ar_mix_cast_array( crpix, DT_INT2, DT_REAL8, darray, dim );
  retval = dmi_coord_create(dd, name, unit, cptNames, dim, transform,
			     darray, crval, cdelt, DT_REAL8, params);
  free(darray);
  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   dmCoordCreate_l                                 */
/*                                                             */
/* Description:                                                */
/*    Create a coordinate system on the input descriptor.      */
/*    Return the coordinate system descriptor.                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *dd        - coordinate descriptor        */
/*     char          *name      - name                         */
/*     char          *unit      - units                        */
/*     char         **cptNames  - component names              */
/*     long           dim       - length                       */
/*     char          *transform -                              */
/*     long          *crpix     -                              */
/*     double        *crval     -                              */
/*     double        *cdelt     -                              */
/*     double        *params    -                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *retval                              */
/*                                                             */
/***************************************************************/
dmDescriptor* dmCoordCreate_l(dmDescriptor* dd, char* name, char* unit,
			      char** cptNames, long dim, char* transform,
			      long* crpix, double* crval, double* cdelt,
			      double* params)
{
  dmDescriptor* retval  = NULL;
  double*  darray = NULL;

  darray = malloc( dim * UT_SZ_D );
  ar_mix_cast_array( crpix, DT_INT4, DT_REAL8, darray, dim );
  retval =  dmi_coord_create(dd, name, unit, cptNames, dim, transform,
			      darray, crval, cdelt, DT_REAL8, params);
  free(darray);
  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   dmCoordCreate_f                                 */
/*                                                             */
/* Description:                                                */
/*    Create a coordinate system on the input descriptor.      */
/*    Return the coordinate system descriptor.                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *dd        - coordinate descriptor        */
/*     char          *name      - name                         */
/*     char          *unit      - units                        */
/*     char         **cptNames  - component names              */
/*     long           dim       - length                       */
/*     char          *transform -                              */
/*     float         *crpix     -                              */
/*     double        *crval     -                              */
/*     double        *cdelt     -                              */
/*     double        *params    -                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *retval                              */
/*                                                             */
/***************************************************************/
dmDescriptor* dmCoordCreate_f(dmDescriptor* dd, char* name, char* unit,
			      char** cptNames, long dim, char* transform,
			      float* crpix, double* crval, double* cdelt,
			      double* params)
{
  dmDescriptor* retval  = NULL;
  double*  darray = NULL;

  darray = malloc( dim * UT_SZ_D );
  ar_mix_cast_array( crpix, DT_REAL4, DT_REAL8, darray, dim );
  retval = dmi_coord_create(dd, name, unit, cptNames, dim, transform,
		            darray, crval, cdelt, DT_REAL8, params);

  free(darray);
  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   dmCoordCreate_d                                 */
/*                                                             */
/* Description:                                                */
/*    Create a coordinate system on the input descriptor.      */
/*    Return the coordinate system descriptor.                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *dd        - coordinate descriptor        */
/*     char          *name      - name                         */
/*     char          *unit      - units                        */
/*     char         **cptNames  - component names              */
/*     long           dim       - length                       */
/*     char          *transform -                              */
/*     double        *crpix     -                              */
/*     double        *crval     -                              */
/*     double        *cdelt     -                              */
/*     double        *params    -                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *retval                              */
/*                                                             */
/***************************************************************/
dmDescriptor* dmCoordCreate_d(dmDescriptor* dd, char* name, char* unit,
			      char** cptNames, long dim, char* transform,
			      double* crpix, double* crval, double* cdelt,
			      double* params)
{
  return dmi_coord_create(dd, name, unit, cptNames, dim, transform,
		          crpix, crval, cdelt, DT_REAL8, params);
}


/***************************************************************/
/*                                                             */
/* Function:   dmCoordCreate_ub                                */
/*                                                             */
/* Description:                                                */
/*    Create a coordinate system on the input descriptor.      */
/*    Return the coordinate system descriptor.                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *dd        - coordinate descriptor        */
/*     char          *name      - name                         */
/*     char          *unit      - units                        */
/*     char         **cptNames  - component names              */
/*     long           dim       - length                       */
/*     char          *transform -                              */
/*     unsigned char *crpix     -                              */
/*     double        *crval     -                              */
/*     double        *cdelt     -                              */
/*     double        *params    -                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *retval                              */
/*                                                             */
/***************************************************************/
dmDescriptor* dmCoordCreate_ub(dmDescriptor* dd, char* name, char* unit,
			      char** cptNames, long dim, char* transform,
			      unsigned char* crpix, double* crval, 
			      double* cdelt, double* params)
{
  dmDescriptor* retval  = NULL;
  double*  darray = NULL;

  darray = malloc( dim * UT_SZ_D );
  ar_mix_cast_array( crpix, DT_UINT1, DT_REAL8, darray, dim );
  retval =  dmi_coord_create(dd, name, unit, cptNames, dim, transform,
			      darray, crval, cdelt, DT_REAL8, params);
  free(darray);
  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   dmCoordCreate_us                                */
/*                                                             */
/* Description:                                                */
/*    Create a coordinate system on the input descriptor.      */
/*    Return the coordinate system descriptor.                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *dd        - coordinate descriptor        */
/*     char          *name      - name                         */
/*     char          *unit      - units                        */
/*     char         **cptNames  - component names              */
/*     long           dim       - length                       */
/*     char          *transform -                              */
/*     unsigned short *crpix    -                              */
/*     double        *crval     -                              */
/*     double        *cdelt     -                              */
/*     double        *params    -                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *retval                              */
/*                                                             */
/***************************************************************/
dmDescriptor* dmCoordCreate_us(dmDescriptor* dd, char* name, char* unit,
			      char** cptNames, long dim, char* transform,
			      unsigned short* crpix, double* crval, 
			      double* cdelt, double* params)
{
  dmDescriptor* retval  = NULL;
  double*  darray = NULL;

  darray = malloc( dim * UT_SZ_D );
  ar_mix_cast_array( crpix, DT_UINT2, DT_REAL8, darray, dim );
  retval =  dmi_coord_create(dd, name, unit, cptNames, dim, transform,
			      darray, crval, cdelt, DT_REAL8, params);
  free(darray);
  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   dmCoordCreate_ul                                */
/*                                                             */
/* Description:                                                */
/*    Create a coordinate system on the input descriptor.      */
/*    Return the coordinate system descriptor.                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *dd        - coordinate descriptor        */
/*     char          *name      - name                         */
/*     char          *unit      - units                        */
/*     char         **cptNames  - component names              */
/*     long           dim       - length                       */
/*     char          *transform -                              */
/*     unsigned long *crpix     -                              */
/*     double        *crval     -                              */
/*     double        *cdelt     -                              */
/*     double        *params    -                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *retval                              */
/*                                                             */
/***************************************************************/
dmDescriptor* dmCoordCreate_ul(dmDescriptor* dd, char* name, char* unit,
			      char** cptNames, long dim, char* transform,
			      unsigned long* crpix, double* crval, 
			      double* cdelt, double* params)
{
  dmDescriptor* retval  = NULL;
  double*  darray = NULL;

  darray = malloc( dim * UT_SZ_D );
  ar_mix_cast_array( crpix, DT_UINT4, DT_REAL8, darray, dim );
  retval = dmi_coord_create(dd, name, unit, cptNames, dim, transform,
			   darray, crval, cdelt, DT_REAL8, params);
  free(darray);
  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   dmCoordCreateLinear_s                           */
/*                                                             */
/* Description:                                                */
/*    Create a coordinate system on the input descriptor.      */
/*    Return the coordinate system descriptor.                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *desc      - coordinate descriptor        */
/*     char          *name      - name                         */
/*     char          *unit      - units                        */
/*     short          crpix     -                              */
/*     double         crval     -                              */
/*     double         cdelt     -                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *coord_dd                            */
/*                                                             */
/***************************************************************/
dmDescriptor* dmCoordCreateLinear_s(dmDescriptor* desc, char* name, char* unit, 
				     short crpix, double crval, double cdelt)
{
  short  tpix  = crpix; 
  double tval  = crval; 
  double tdelt = cdelt; 
     
  return dmCoordCreate_s(desc, name, unit, NULL, (long)1, 
		         NULL, &tpix, &tval, &tdelt, NULL);	  
}

/***************************************************************/
/*                                                             */
/* Function:   dmCoordCreateLinear_l                           */
/*                                                             */
/* Description:                                                */
/*    Create a coordinate system on the input descriptor.      */
/*    Return the coordinate system descriptor.                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *desc      - coordinate descriptor        */
/*     char          *name      - name                         */
/*     char          *unit      - units                        */
/*     long           crpix     -                              */
/*     double         crval     -                              */
/*     double         cdelt     -                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *coord_dd                            */
/*                                                             */
/***************************************************************/
dmDescriptor* dmCoordCreateLinear_l(dmDescriptor* desc, char* name, char* unit, 
				     long crpix, double crval, double cdelt)
{
  long    tpix  = crpix; 
  double  tval  = crval; 
  double  tdelt = cdelt; 
     
  return    dmCoordCreate_l(desc, name, unit, NULL, (long)1, 
		            NULL, &tpix, &tval, &tdelt, NULL);	  
}

/***************************************************************/
/*                                                             */
/* Function:   dmCoordCreateLinear_f                           */
/*                                                             */
/* Description:                                                */
/*    Create a coordinate system on the input descriptor.      */
/*    Return the coordinate system descriptor.                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *desc      - coordinate descriptor        */
/*     char          *name      - name                         */
/*     char          *unit      - units                        */
/*     float          crpix     -                              */
/*     double         crval     -                              */
/*     double         cdelt     -                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *coord_dd                            */
/*                                                             */
/***************************************************************/
dmDescriptor* dmCoordCreateLinear_f(dmDescriptor* desc, char* name, char* unit, 
				     float crpix, double crval, double cdelt)
{
  float   tpix  = crpix; 
  double  tval  = crval; 
  double  tdelt = cdelt; 
     
  return dmCoordCreate_f(desc, name, unit, NULL, (long)1, 
		         NULL, &tpix, &tval, &tdelt, NULL);	  
}

/***************************************************************/
/*                                                             */
/* Function:   dmCoordCreateLinear_d                           */
/*                                                             */
/* Description:                                                */
/*    Create a coordinate system on the input descriptor.      */
/*    Return the coordinate system descriptor.                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *desc      - coordinate descriptor        */
/*     char          *name      - name                         */
/*     char          *unit      - units                        */
/*     double         crpix     -                              */
/*     double         crval     -                              */
/*     double         cdelt     -                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *coord_dd                            */
/*                                                             */
/***************************************************************/
dmDescriptor* dmCoordCreateLinear_d(dmDescriptor* desc, char* name, char* unit, 
				     double crpix, double crval, double cdelt)
{
  double  tpix  = crpix; 
  double  tval  = crval; 
  double  tdelt = cdelt; 
     
  return dmCoordCreate_d(desc, name, unit, NULL, (long)1, 
		         NULL, &tpix, &tval, &tdelt, NULL);	  
}


/***************************************************************/
/*                                                             */
/* Function:   dmCoordCreateLinear_ub                          */
/*                                                             */
/* Description:                                                */
/*    Create a coordinate system on the input descriptor.      */
/*    Return the coordinate system descriptor.                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *desc      - coordinate descriptor        */
/*     char          *name      - name                         */
/*     char          *unit      - units                        */
/*     unsigned char  crpix     -                              */
/*     double         crval     -                              */
/*     double         cdelt     -                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *coord_dd                            */
/*                                                             */
/***************************************************************/
dmDescriptor* dmCoordCreateLinear_ub(dmDescriptor* desc, char* name, char* unit, 
				     unsigned char crpix, double crval, 
				      double cdelt)
{
  unsigned char  tpix  = crpix; 
  double         tval  = crval; 
  double         tdelt = cdelt; 
     
  return dmCoordCreate_ub(desc, name, unit, NULL, (long)1, 
		          NULL, &tpix, &tval, &tdelt, NULL);	  
}


/***************************************************************/
/*                                                             */
/* Function:   dmCoordCreateLinear_us                          */
/*                                                             */
/* Description:                                                */
/*    Create a coordinate system on the input descriptor.      */
/*    Return the coordinate system descriptor.                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *desc      - coordinate descriptor        */
/*     char          *name      - name                         */
/*     char          *unit      - units                        */
/*     unsigned short crpix     -                              */
/*     double         crval     -                              */
/*     double         cdelt     -                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *coord_dd                            */
/*                                                             */
/***************************************************************/
dmDescriptor* dmCoordCreateLinear_us(dmDescriptor* desc, char* name, char* unit, 
				     unsigned short crpix, double crval, 
				      double cdelt)
{
  unsigned short  tpix  = crpix; 
  double          tval  = crval; 
  double          tdelt = cdelt; 
     
  return dmCoordCreate_us(desc, name, unit, NULL, (long)1, 
		          NULL, &tpix, &tval, &tdelt, NULL);	  
}


/***************************************************************/
/*                                                             */
/* Function:   dmCoordCreateLinear_ul                          */
/*                                                             */
/* Description:                                                */
/*    Create a coordinate system on the input descriptor.      */
/*    Return the coordinate system descriptor.                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor  *desc      - coordinate descriptor        */
/*     char          *name      - name                         */
/*     char          *unit      - units                        */
/*     unsigned long  crpix     -                              */
/*     double         crval     -                              */
/*     double         cdelt     -                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *coord_dd                            */
/*                                                             */
/***************************************************************/
dmDescriptor* dmCoordCreateLinear_ul(dmDescriptor* desc, char* name, char* unit, 
				      unsigned long crpix, double crval, 
				      double cdelt)
{
  unsigned long  tpix  = crpix; 
  double         tval  = crval; 
  double         tdelt = cdelt; 
    
  return  dmCoordCreate_ul(desc, name, unit, NULL, (long)1, 
		            NULL, &tpix, &tval, &tdelt, NULL);	  
}
