/*                                                                
**  Copyright (C) 1999,2003,2005,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmtan.c                                                     */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#ifndef QUICK_H
#define QUICK_H

#include <stdio.h>
#include <math.h>
#ifndef PI
#define PI 3.14159265358979
#endif
#ifndef RADIAN
#define RADIAN (PI/180.0)
#endif
void dm_wcs_make_cd( double* cdelt, double crota, double* cd );
void dmTanApplyAspect( double* det, double* nom, double* asol, double* crpix, double* cdelt, double* sky );
void dmTanPixToWorld( double* pix, double* asp, double* crpix, double* cdelt, double* world );
void dmTanWorldToPix( double* world, double* asp, double* crpix, double* cdelt, double* pix );
int dm_wcs_matinv2( double* mat, double* inv);
int dm_wcs_sphfwd( double lng, double lat, double* eul, double* phi, double* theta);
int dm_wcs_sphrev ( double phi, double theta, double* eul, double* lng, double* lat);

#ifndef __STDC__
#ifndef const
#define const
#endif
#endif

#ifndef copysign
#define copysign(X, Y) ((Y) < 0.0 ? -fabs(X) : fabs(X))
#endif

const double sph_tol = 1.0e-5;

#endif



/*
 * Spec for new aspect code:
 * replace
 *
         pix_apply_aspect(fpc, aspect->asp_sol, evt_p->sky);
 *
 * with
 *
 * double nominal_asp[3];
 * double crpix[2];
 * double cdelt[2];
 * double asp[3];
 *	nominal_asp[0] = obs_info_p->ObservKey_ptr.ra_nom;
 *	nominal_asp[1] = obs_info_p->ObservKey_ptr.dec_nom;
 *	nominal_asp[2] = 0.0;
 *	pix_get_fp_image_cntr(crpix);
 *	cdelt[1] = inp_p->fp_scale/3600.0;  
 *	cdelt[0] = -cdelt[1];   
 *	dmTanApplyAspect(  fpc, nominal_asp, asp, crpix, cdelt, evt_p->sky );
 * Where aspect->asp is the RA,DEC,ROLL read from the asol.fits file
 * and NOT the offsets from the aoff.fits file.
 */    
/* Apr 30 1998	Define copysign only if it is not already defined
 * Sep 28 1998	Include wcslib.h instead of wcstrig.h
 */

/***************************************************************/
/*                                                             */
/* Function:   dm_wcs_matinv2                                  */
/*                                                             */
/* Description:                                                */
/*     Simple routine for 2D matrix inversion                  */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     double      *mat  - matrix                              */
/*                                                             */
/*  Output parameters:                                         */
/*     double      *inv  - inverted matrix                     */
/*                                                             */
/*  Returns:                                                   */
/*           int retval  - 0 or 1                              */
/*                                                             */
/***************************************************************/
int dm_wcs_matinv2( double* mat, double* inv)
{
  double det = 0.0;
  double a   = 0.0;
  double b   = 0.0;
  double c   = 0.0;
  double d   = 0.0;

  int retval = 0;

  /* Set matrix elements */
  a = mat[0];
  b = mat[1];
  c = mat[2];
  d = mat[3];

  /* Calc the determinant */
  det = a * d - b * c;

  if ( det == 0 )
  {
    inv[0] = 1.0;
    inv[1] = 0.0;
    inv[2] = 0.0;
    inv[3] = 1.0;
    retval = 0;
  }
  else
  {
    inv[0] =  d / det;
    inv[1] = -b / det;
    inv[2] = -c / det;
    inv[3] =  a / det;
    retval = 1;
  }

  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   dm_wcs_make_cd                                  */
/*                                                             */
/* Description:                                                */
/*     Make forward and reverse rotation matrices              */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     double      *cdelt -                                    */
/*     double       crota -                                    */
/*                                                             */
/*  Output parameters:                                         */
/*     double      *cd    -                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dm_wcs_make_cd( double* cdelt, double crota, double* cd )
{
  long   t1 = 0;
  long   t2 = 0;
  double c  = 0.0;
  double s  = 0.0;

  t1 = (cdelt[0] < 0.0 ) ? -1 : 1;
  t2 = (cdelt[1] < 0.0 ) ?  1 : -1;

  c = cos( RADIAN * crota );
  s = sin( RADIAN * crota );

  if ( cdelt[0] * cdelt[1] > 0 )
    s = -s; 

  cd[0] =           cdelt[0]   * c;
  cd[1] = t1 * fabs(cdelt[1] ) * s;
  cd[2] = t2 * fabs(cdelt[0] ) * s;
  cd[3] =           cdelt[1]   * c;
 
}


/***************************************************************/
/*                                                             */
/* Function:   dmTanPixToWorld                                 */
/*                                                             */
/* Description:                                                */
/*     Change coordinate systems.                              */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     double      *pix   -                                    */
/*     double      *asp   -                                    */
/*     double      *crpix -                                    */
/*     double      *cdelt -                                    */
/*                                                             */
/*  Output parameters:                                         */
/*     double      *world -                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmTanPixToWorld( double* pix, double* asp, double* crpix,
                      double* cdelt, double* world )
{
  double  r     = 0.0;
  double  x     = 0.0;
  double  y     = 0.0;
  double  phi   = 0.0; 
  double  theta = 0.0;

  double  euler[5];
  double  offset[2];
  double  cd[4];

  long ii   = 0;
  long ndim = 2;


  dm_wcs_make_cd( cdelt, asp[2], cd );

  for ( ii = 0; ii < ndim; ii++ )
  {
    offset[ii] = pix[ii] - crpix[ii];
  }

  x = cd[0] * offset[0] + cd[2] * offset[1];
  y = cd[1] * offset[0] + cd[3] * offset[1];
  r = sqrt( x*x + y* y );

  phi = 0.0;

  if ( r > 0.0) 
  {
    phi = atan2( x, -y )/RADIAN;
  }

  theta = atan2( 1.0, r*RADIAN )/RADIAN;

  euler[0] = asp[0];
  euler[1] = 90.0 - asp[1];
  euler[2] = 180.0;
  euler[3] = cos( RADIAN * euler[1] );
  euler[4] = sin( RADIAN * euler[1] );

  dm_wcs_sphrev( phi, theta, euler, &world[0], &world[1] );
}


/***************************************************************/
/*                                                             */
/* Function:   dmTanApplyAspect                                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     double      *det   -                                    */
/*     double      *nom   -                                    */
/*     double      *asol  -                                    */
/*     double      *crpix -                                    */
/*     double      *cdelt -                                    */
/*                                                             */
/*  Output parameters:                                         */
/*     double      *sky   -                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmTanApplyAspect( double* det, double* nom, double* asol, double* crpix, 
		       double* cdelt, double* sky )
{
  double cel[2];

  dmTanPixToWorld( det, asol, crpix, cdelt, cel );
  dmTanWorldToPix( cel, nom, crpix, cdelt, sky );
}


/***************************************************************/
/*                                                             */
/* Function:   dmTanWorldToPix                                 */
/*                                                             */
/* Description:                                                */
/*     Change coordinate systems.                              */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     double      *world -                                    */
/*     double      *asp   -                                    */
/*     double      *crpix -                                    */
/*     double      *cdelt -                                    */
/*                                                             */
/*  Output parameters:                                         */
/*     double      *pix   -                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmTanWorldToPix( double* world, double* asp, double* crpix,
                      double* cdelt, double* pix )
{
  double  r     = 0.0;
  double  x     = 0.0;
  double  y     = 0.0;
  double  s     = 0.0;
  double  phi   = 0.0; 
  double  theta = 0.0;
 
  double euler[5];
  double cd[4];
  double cdf[4];  /* Forward matrix, not needed */


  if ( asp[2] == 0.0 )
  {
    cd[0] = 1/cdelt[0];  
    cd[1] = 0;
    cd[2] = 0; 
    cd[3] = 1/cdelt[1];
  }
  else 
  {
    dm_wcs_make_cd( cdelt, asp[2], cdf );
    dm_wcs_matinv2( cdf, cd );
   }

  euler[0] = asp[0];
  euler[1] = 90.0 - asp[1];
  euler[2] = 180.0;
  euler[3] = cos( RADIAN * euler[1] );
  euler[4] = sin( RADIAN * euler[1] );

  dm_wcs_sphfwd( world[0], world[1], euler, &phi, &theta );
  s = sin( RADIAN * theta );

  if ( s == 0.0 )
  {
    pix[0] = crpix[0];
    pix[1] = crpix[1];
    return;
  }

  r = cos (RADIAN * theta )/ s / RADIAN;
  x =  r * sin ( RADIAN * phi );
  y = -r * cos ( RADIAN * phi );

  pix[0] = cd[0] * x + cd[2] * y + crpix[0];
  pix[1] = cd[1] * x + cd[3] * y + crpix[1];
}


/***************************************************************/
/*                                                             */
/* Function:   dm_wcs_sphfwd                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     double       lng   -                                    */
/*     double       lat   -                                    */
/*     double      *eul   -                                    */
/*                                                             */
/*  Output parameters:                                         */
/*     double      *phi   -                                    */
/*     double      *theta -                                    */
/*                                                             */
/*  Returns:                                                   */
/*           int 0                                             */
/*                                                             */
/***************************************************************/
int dm_wcs_sphfwd( double lng, double lat, double* eul, double* phi, double* theta)
{
  double  coslat = 0.0;
  double  coslng = 0.0;
  double  dlng   = 0.0;
  double  dphi   = 0.0;
  double  sinlat = 0.0;
  double  sinlng = 0.0;
  double  x      = 0.0;
  double  y      = 0.0;
  double  z      = 0.0;

  coslat = cos( RADIAN * lat);
  sinlat = sin( RADIAN * lat);

  dlng = lng - eul[0];
  coslng = cos( RADIAN * dlng);
  sinlng = sin( RADIAN * dlng);

  /* Compute the native longitude. */
  x = sinlat*eul[4] - coslat*eul[3]*coslng;

  if (fabs(x) < sph_tol) 
  {
    /* Rearrange formula to reduce roundoff errors. */
    x = -cos(RADIAN* (lat+eul[1])) + coslat*eul[3]*(1.0 - coslng);
  }

  y = -coslat*sinlng;

  if (x != 0.0 || y != 0.0) 
  {
    dphi = atan2(y, x)/RADIAN;
  } 
  else 
  {
    /* Change of origin of longitude. */
    dphi = dlng - 180.0;
  }
  *phi = eul[2] + dphi;

  /* Normalize the native longitude. */
  if (*phi > 180.0) 
  {
    *phi -= 360.0;
  } 
  else if (*phi < -180.0) 
  {
    *phi += 360.0;
  }

  /* Compute the native latitude. */
  if (fmod(dlng,180.0) == 0.0) 
  {
    *theta = lat + coslng*eul[1];

    if (*theta >  90.0)
      *theta =  180.0 - *theta;

    if (*theta < -90.0)
      *theta = -180.0 - *theta;
  } 
  else 
  {
    z = sinlat*eul[3] + coslat*eul[4]*coslng;

    if (fabs(z) > 0.99) 
    {
      /* Use an alternative formula for greater numerical accuracy. */
      *theta = copysign(acos(sqrt(x*x+y*y))/RADIAN, z);
    }
    else 
    {
      *theta = asin(z)/RADIAN;
    }
  }

  return 0;
}


/***************************************************************/
/*                                                             */
/* Function:   dm_wcs_sphrev                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     double       phi   -                                    */
/*     double       theta -                                    */
/*     double      *eul   -                                    */
/*                                                             */
/*  Output parameters:                                         */
/*     double      *lng   -                                    */
/*     double      *lat   -                                    */
/*                                                             */
/*  Returns:                                                   */
/*           int 0                                             */
/*                                                             */
/***************************************************************/
int dm_wcs_sphrev ( double phi, double theta, double* eul, double* lng, double* lat)
{
  double  cosphi = 0.0;
  double  costhe = 0.0;
  double  dlng   = 0.0;
  double  dphi   = 0.0;
  double  sinphi = 0.0;
  double  sinthe = 0.0;
  double  x      = 0.0;
  double  y      = 0.0;
  double  z      = 0.0;

  costhe = cos(RADIAN*theta);
  sinthe = sin(RADIAN*theta);

  dphi = phi - eul[2];
  cosphi = cos(RADIAN*dphi);
  sinphi = sin(RADIAN*dphi);

  /* Compute the celestial longitude. */
  x = sinthe*eul[4] - costhe*eul[3]*cosphi;

  if (fabs(x) < sph_tol) 
  {
    /* Rearrange formula to reduce roundoff errors. */
    x = -cos(RADIAN*(theta+eul[1])) + costhe*eul[3]*(1.0 - cosphi);
  }
  y = -costhe*sinphi;

  if (x != 0.0 || y != 0.0) 
  {
    dlng = atan2(y, x)/RADIAN;
  } 
  else 
  {
    /* Change of origin of longitude. */
    dlng = dphi + 180.0;
  }

  *lng = eul[0] + dlng;

  /* Normalize the celestial longitude. */
  if (eul[0] >= 0.0) 
  {
    if (*lng < 0.0) *lng += 360.0;
  } 
  else 
  {
    if (*lng > 0.0) *lng -= 360.0;
  }

  if (*lng > 360.0) 
  {
    *lng -= 360.0;
  }
  else if (*lng < -360.0) 
  {
    *lng += 360.0;
  }

  /* Compute the celestial latitude. */
  if (fmod(dphi,180.0) == 0.0) 
  {
    *lat = theta + cosphi*eul[1];

    if (*lat >  90.0)
      *lat =  180.0 - *lat;

    if (*lat < -90.0)
      *lat = -180.0 - *lat;
  } 
  else 
  {
    z = sinthe*eul[3] + costhe*eul[4]*cosphi;

    if (fabs(z) > 0.99) 
    {
      /* Use an alternative formula for greater numerical accuracy. */
      *lat = copysign(acos(sqrt(x*x+y*y))/RADIAN, z);
    }
    else 
    {
      *lat = asin(z)/RADIAN;
    }
  }
  return 0;
}


