/*                                                                
**  Copyright (C) 1997-2001,2003-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmdataset.c                                                 */
/*                                                               */
/*   Functions that create and work with dmDataset structs       */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmlibP.h"
#include <unistd.h>


/***************************************************************/
/*                                                             */
/* Function:   dmDatasetAccess                                 */
/*                                                             */
/* Description:  Test for access.  Deprecated.  If you want to */
/*               test access, just open the darn thing         */
/*                                                             */
/*  Input parameters:                                          */
/*               const char *ddsname                           */
/*               const char *mode                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmBool  retval                                   */
/*                                                             */
/***************************************************************/
dmBool dmDatasetAccess( const char* ddsname, const char* mode )
{
  dmBool     retval = dmSUCCESS;
  dmDataset *ds     = NULL;

  ds = dmp_dataset_open_mode( ddsname, NULL, mode );
  dmu_err_clear();
  retval =  ( ds != NULL );
  dmDatasetClose( ds );

  return retval;
} 


/***************************************************************/
/*                                                             */
/* Function:   dmDatasetCreate                                 */
/*                                                             */
/* Description:  Create a dataset                              */
/*                                                             */
/*  Input parameters:                                          */
/*               const char *ddsname                           */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmDataset *ds                                    */
/*                                                             */
/***************************************************************/
dmDataset *dmDatasetCreate( const char *ddsname )
{
  dmParseList *vparse = NULL;
  dmDataset   *ds     = NULL;

  vparse = dmi_block_parse_syntax( NULL, ddsname, dmUNKBLOCK );

  if ( !vparse )
    return NULL;

  ds = dmp_dataset_open_mode( ddsname, vparse->option, "W" );
  dmi_parse_list_free( vparse );
  return ds;
}



/***************************************************************/
/*                                                             */
/* Function:   dmDatasetOpen                                   */
/*                                                             */
/* Description:  Open a dataset.  There are two options, read  */
/*               and read/write                                */
/*                                                             */
/*  Input parameters:                                          */
/*               const char *dsname                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmDataset *ds                                    */
/*                                                             */
/***************************************************************/
dmDataset *dmDatasetOpen( const char *dsname )
{
  dmParseList *vparse = NULL;
  dmDataset   *ds     = NULL;

  vparse = dmi_block_parse_syntax( NULL, dsname, dmUNKBLOCK );
 
  if ( !vparse )
    return NULL;

  ds = dmp_dataset_open_mode( dsname, vparse->option, "R" );
  dmi_parse_list_free( vparse );
  return ds;
}


/***************************************************************/
/*                                                             */
/* Function:   dmDatasetOpenUpdate                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*               const char *dsname                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmDataset *ds                                    */
/*                                                             */
/***************************************************************/
dmDataset  *dmDatasetOpenUpdate( const char *dsname )
{
  dmParseList *vparse = NULL;
  dmDataset   *ds     = NULL;

  vparse = dmi_block_parse_syntax( NULL, dsname, dmUNKBLOCK );
 
  if ( !vparse )
    return NULL;

  ds = dmp_dataset_open_mode( dsname, vparse->option, "RW" );
  dmi_parse_list_free( vparse );
  return ds;
}



/***************************************************************/
/*                                                             */
/* Function:   dmDatasetRename                                 */
/*                                                             */
/* Description: For now just use POSIX rename()                */
/*                                                             */
/*  Input parameters:                                          */
/*               const char *oldname                           */
/*               const char *newname                           */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  errcode                               */
/*                                                             */
/***************************************************************/
dmErrCode  dmDatasetRename( const char *oldname, const char *newname )
{
  if( rename( oldname, newname ) ) 
  {
    dmu_err_set( DM_ERR_SYS_CALL, "Rename file failed." );
    return dmFAILURE;
  }

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmDatasetPrinte                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*               const dmDataset *ds                           */
/*               char            *options                      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  errcode                               */
/*                                                             */
/***************************************************************/
dmErrCode dmDatasetPrint( const dmDataset* ds, char* options )
{
  dmKernel  kernel;
  dmBlock  *current = NULL;

  if ( !ds )
    return dmSUCCESS;

  kernel = ds->kernel;
  printf( "Dataset: %s\n", ds->name );

  if ( ds->mode )
    printf( "Mode:    RW\n" );
  else
    printf( "Mode:    R \n" ); 

  printf( "Blocks: %ld\n", ds->nblocks );
  current = ds->current;

  if ( current )
    printf( "Current Block: %ld\n", current->no );

  kernel->DatasetPrint( ds );

  /* Suppress compiler warn */
  if ( !options )
    return dmSUCCESS;

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   dmDatasetClose                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*               dmDataset  *ds                                */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode dmDatasetClose( dmDataset* ds )
{
  dmBlock   *block   =  NULL;
  dmKernel   kernel;
  dmErrCode  status  = dmSUCCESS; 
  long       nblocks = 0;
  long       no      = 0;


  if ( !ds )
    return dmFAILURE;

  /* Flush all virtual blocks */
  nblocks = dmi_dataset_get_nv( ds );

  for ( no = 1; no <= nblocks; no++ )
  {
    block = dmi_dataset_get_block( ds, no );

    if ( block )
    {
      dmp_block_close( block );
     }
  }

  kernel = ds->kernel;
  kernel->DatasetClose( ds ); 

  status = dmu_err_get();
  return status; 
}



/***************************************************************/
/*                                                             */
/* Function:   dmDatasetGetKernel                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*             const dmDataset  *ds                            */
/*             char             *kername                       */
/*             const long        maxlen                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode dmDatasetGetKernel( const dmDataset* ds, char* kername,
                              const long maxlen )
{
  dmKernel kernel;

  if ( !kername )
  {
    dmu_err_set( DM_ERR_NULL, "[dmDatasetGetKernel]: Null name" );
    return dmFAILURE;
  }

  if ( !ds )
  {
    dmu_err_set( DM_ERR_NULL, "[dmDatasetGetKernel]: Null dataset" );
    strcpy( kername, "" );
    return dmFAILURE;
  }

  kernel = ds->kernel;
  strncpy( kername, kernel->name, maxlen );
  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   dmDatasetGetName                                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*             const dmDataset  *ds                            */
/*             char             *name                          */
/*             const long        maxlen                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode  dmDatasetGetName( const dmDataset* ds, char* name,
                             const long maxlen )
{
  if ( !name )
  {
    dmu_err_set( DM_ERR_NULL, "[dmDatasetGetName]: Null name" );
    return dmFAILURE;
  }

  if ( !ds )
  {
    dmu_err_set( DM_ERR_NULL, "[dmDatasetGetName]: Null dataset" );
    strcpy( name, "" );
    return dmFAILURE;
  }

  strncpy( name, ds->name, maxlen );
  return dmSUCCESS;
}
 

/***************************************************************/
/*                                                             */
/* Function:   dmDatasetMoveToBlock                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDataset   *ds                                 */
/*             long         no                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmBlock  *block                                  */
/*                                                             */
/***************************************************************/
dmBlock *dmDatasetMoveToBlock( dmDataset *ds, long no )
{
  dmBlock *block = NULL;
 
  if ( !ds )
  {
    dmu_err_set( DM_ERR_NULL, "[dmDatasetMoveToBlock]: Null dataset" );
    return NULL;
  }

  /* CIAO3 functionality: no=0 supported as default block */

  if ( dmi_dataset_mode( ds ))
    block = dmp_block_open_mode( ds, NULL, no, dmUNKNOWNBLOCK, NULL, "RW" );
  else
    block = dmp_block_open_mode( ds, NULL, no, dmUNKNOWNBLOCK, NULL, "R" );

  dmi_block_reset_rows( block );

  return block;
}

/***************************************************************/
/*                                                             */
/* Function:   dmDatasetAdvanceBlocks                          */
/*                                                             */
/* Description: Move ahead given number of blocks (can be a    */
/*              negative number)                               */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDataset   *ds                                 */
/*             long         relNo                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmBlock  *block                                  */
/*                                                             */
/***************************************************************/
dmBlock*    dmDatasetAdvanceBlocks( dmDataset* ds, long relNo )
{
  long blockNo = 0;

  if ( !ds )
  {
    dmu_err_set( DM_ERR_NULL, "[dmDatasetAdvanceBlocks]: Null dataset" );
    return NULL;
  }

  blockNo = ds->current_no + relNo;

  return dmDatasetMoveToBlock( ds, blockNo );
}


/***************************************************************/
/*                                                             */
/* Function:   dmDatasetGetBlockName                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             const dmDataset  *ds                            */
/*             const long        blockNo                       */
/*             char             *name                          */
/*             const long        maxlen                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmBlock  *block                                  */
/*                                                             */
/***************************************************************/
dmErrCode dmDatasetGetBlockName( const dmDataset* ds, const long blockNo,
                                 char* name, const long maxlen )
{
  char *pname =  NULL;

  if ( !ds )
  {
    dmu_err_set( DM_ERR_NULL, "[dmDatasetCurrentBlockNo]: Null dataset" );
    return dmFAILURE;
  }

  if ( blockNo < 1 || blockNo > ds->nblocks )
  {
    sprintf( dm_errbuf, "[dmDatasetGetBlockName]: Bad block number %ld", blockNo );
    dmu_err_set( DM_ERR_INDEX, dm_errbuf );

    return dmFAILURE;
  }

  pname = dmi_dataset_get_block_name( ds, blockNo );
  cs_copy_siz( pname, name, maxlen );

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   dmDatasetGetBlockType                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             const dmDataset  *ds                            */
/*             const long        blockNo                       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmBlockType  blockType                           */
/*                                                             */
/***************************************************************/
dmBlockType dmDatasetGetBlockType( const dmDataset* ds, const long blockNo )
{
  if ( !ds )
  {
    dmu_err_set( DM_ERR_NULL, "[dmDatasetCurrentBlockNo]: Null dataset" );
    return dmUNKNOWNBLOCK;
  }

  if ( blockNo < 1 || blockNo > ds->nblocks )
  {
    sprintf( dm_errbuf, "Bad block number %ld", blockNo );
    dmu_err_set( DM_ERR_INDEX, dm_errbuf );
    return dmUNKNOWNBLOCK;
  }

  return dmi_dataset_get_block_type( ds, blockNo );
}


/***************************************************************/
/*                                                             */
/* Function:   dmDatasetGetNoBlocks                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             const dmDataset  *ds                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long  nblocks                                    */
/*                                                             */
/***************************************************************/
long dmDatasetGetNoBlocks( const dmDataset* ds )
{
  if ( !ds )
  {
    dmu_err_set( DM_ERR_NULL, "Null dataset" );
    return 0;
  }

  return ds->nblocks;
}


/***************************************************************/
/*                                                             */
/* Function:   dmDatasetGetCurrentBlockNo                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             const dmDataset  *ds                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long  blockNumber                                */
/*                                                             */
/***************************************************************/
long dmDatasetGetCurrentBlockNo( const dmDataset* ds )
{
  return  dmi_dataset_get_current( ds );
}


/***************************************************************/
/*                                                             */
/* Function:   dmDatasetNextBlock                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             const dmDataset  *ds                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmBlock  *block                                  */
/*                                                             */
/***************************************************************/
dmBlock* dmDatasetNextBlock( dmDataset* ds )
{
  return  dmDatasetAdvanceBlocks( ds, 1 );
}


/***************************************************************/
/*                                                             */
/* Function:   dmDatasetDelete                                 */
/*                                                             */
/* Description:                                                */
/*             Close and remove from disk                      */
/*                                                             */
/*  Input parameters:                                          */
/*             const dmDataset  *ds                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode dmDatasetDelete( dmDataset* ds )
{
  dmKernel  kernel;
  long      blockNo = 0;


  if ( !ds )
  {
    dmu_err_set( DM_ERR_NULL, "[dmDatasetDelete]: Null dataset" );
  }
  else
  {
    for ( blockNo = 1; blockNo <= ds->nvblocks; blockNo++ )
      dmp_block_close(ds->blocks[blockNo-1]);

    kernel = ds->kernel;
    kernel->DatasetDelete( ds ); 
  }

  return  dmu_err_get();
}


/***************************************************************/
/*                                                             */
/* Function:   dmDatasetDestroy                                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             const char  *dsname                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode  dmDatasetDestroy( const char* dsname )
{
  dmDataset *ds;
  int        status = 0;  
  int        exists = 0;


  exists = fio_access( dsname );
      
  if ( exists == FIO_NOFILE )
    return dmSUCCESS;
        
  if ( exists != FIO_READWRITE )
  {
    sprintf( dm_errbuf, "No permission to delete file %s", dsname );
    dmu_err_set( DM_ERR_NO_PERM, dm_errbuf );
    return dmFAILURE;
  }

  /* This is the problematic bit ... may die if file is partly corrupt.    */
  /* However, can potentially handle datasets consisting of multiple files.*/

  ds = dmp_dataset_open_mode( dsname, NULL, "RW" );

  if ( ds )
  {
    status = dmDatasetDelete( ds );
    return status;
  }

  /* DM_ERR_NO_KERNEL is okay; it just means the DM didn't understand */
  /* the file, which doesn't matter if we only want to delete it      */

  if ( dmu_err_get() == DM_ERR_NO_KERNEL )
    dmu_err_clear();

  /* That didn't work. Try just deleting it with POSIX unlink */

  if ( 0 != unlink( dsname ) )  /* POSIX */
    return dmFAILURE;
      
  return dmSUCCESS;
}

/***************************************************************/
/*                                                             */
/* Function:   dmDatasetWriteable                              */
/*                                                             */
/* Description:                                                */
/*   Check if the dataset is writeable.                        */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDataset *ds                                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*    dmTRUE  - dataset is writeable                           */
/*    dmFALSE - dataset is not writeable                       */
/*                                                             */
/***************************************************************/
dmBool dmDatasetWriteable( dmDataset *ds )
{
  dmBool retval = dmFALSE;

  if ( dmi_dataset_mode( ds ))
    retval = dmTRUE;

  return retval;
} 
