/*                                                                
**  Copyright (C) 1997-2001,2003-2005,2007,2008  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/***********************************************************************/
/*                                                                     */
/*  dmarrayget.c                                                       */
/*                                                                     */
/*                                                                     */
/***********************************************************************/

#include "dmlibP.h" 

long dmp_coord_get_vector( dmDescriptor* dd, void* vals, dmDataType type, long row, long nrows, long start, long nvals );


/***************************************************************/
/* Function:  dmSetVectorArray                                 */
/*                                                             */
/* Description:                                                */
/*    Set values to the provided data descriptor for the       */
/*    current row.                                             */
/*                                                             */
/* Input parameters:                                           */
/*       dmDescriptor  *dd    - descriptor                     */
/*       dmDataType     type  - datatype of values provided    */
/*                              (may be diff. from desc.)      */
/*       void          *vals  - values to assign.              */
/*       long           dim   - number of vector cpts in vals  */
/*       long           start - start index of array to assign */
/*                              values (typcially 1)           */
/*       long           nvals - number of array elements       */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*        dmErrCode  error_code from dmu_err_get()             */
/*                                                             */
/***************************************************************/
dmErrCode dmSetVectorArray( dmDescriptor* dd, dmDataType type, void* vals,
                            long dim, long start, long nvals)
{
  long     row   = 0;
  long     nrows = 1L;
  dmBlock *block = NULL;

  if ( !dd || !vals ) 
  {
    dmu_err_set( DM_ERR_NULL, "Null argument" );
    return dmu_err_get();
  }

  block = (dmBlock*)dd->block;
  row = dmi_block_reset_current_row( block );

  dmSetVectorArrays( dd, type, vals, dim, row, nrows, start, nvals );

  return dmu_err_get();
}


/***************************************************************/
/* Function:  dmGetVectorArray                                 */
/*                                                             */
/* Description:                                                */
/*    Extract values from current row, for provided data       */
/*    descriptor.                                              */
/*                                                             */
/* Input parameters:                                           */
/*       dmDescriptor  *dd    - descriptor                     */
/*       dmDataType     type  - datatype of values provided    */
/*                              (may be diff. from desc.)      */
/*       long           dim   - number of vector cpts in vals  */
/*       long           start - start index of array to assign */
/*                              values (typcially 1)           */
/*       long           nvals - number of array elements       */
/*                                                             */
/* Output parameters:                                          */
/*       void          *vals  - memory for return values       */
/*                                                             */
/* Returns:                                                    */
/*        dmErrCode  error_code from dmu_err_get()             */
/*                                                             */
/***************************************************************/
dmErrCode dmGetVectorArray( dmDescriptor* dd, dmDataType type, void* vals, 
                            long dim, long start, long nvals)
{
  long     row   = 0;
  dmBlock *block = NULL;

  if ( !dd || !vals ) 
  {
    dmu_err_set( DM_ERR_NULL, "Null argument" );
    return dmu_err_get();
  }

  block = (dmBlock*)dd->block;
  row   = dmi_block_reset_current_row( block );

  dmGetVectorArrays( dd, type, vals, dim, row, 1L, start, nvals );

  return dmu_err_get();
}


/***************************************************************/
/* Function:  dmSetVectorArrays                                */
/*                                                             */
/* Description:                                                */
/*     Set values to the provided data descriptor for multiple */
/*     rows.                                                   */  
/*                                                             */
/* Input parameters:                                           */
/*       dmDescriptor  *dd    - descriptor                     */
/*       dmDataType     type  - datatype of values provided    */
/*                              (may be diff. from desc.)      */
/*       void          *vals  - values to assign               */
/*       long           dim   - number of vector cpts in vals  */
/*       long           row   - start row                      */
/*       long           nrows - number of rows of data provided*/
/*                              in vals                        */
/*       long           start - start index of array to assign */
/*                              values (typcially 1)           */
/*       long           nvals - number of array elements       */
/*                                                             */
/* Output parameters:                                          */
/*       none                                                  */
/*                                                             */
/* Returns:                                                    */
/*        long  ngot - number of rows set                      */
/*                                                             */
/***************************************************************/
long dmSetVectorArrays( dmDescriptor* dd, dmDataType type, void* vals, 
                        long dim, long row, long nrows, long start, long nvals)
{
   GenericType tmp;
   void    *iptr  = NULL;
   dmBlock *block = NULL;
   long ngot = 1L;
   int  is_alloced = UT_FALSE;

   if ( !dd || !vals )
   {
     dmu_err_set( DM_ERR_NULL, "Null argument" );
     return dmFAILURE;
   }

   if ( dd->dim != dim )
     return dmFAILURE;

   block = (dmBlock*)dd->block;

   switch( dd->descType )
   {
     case dmCOLUMN:
     case dmCOORD:
       ngot = dmp_buffer_set_cast_data( dd, vals, type, row, nrows, start, nvals );
       break;

     case dmKEY:
     if ( type != dd->type )
     {
        /* need to cast values to descriptor datatype. */
        if ( nvals > 1L ) 
        {
 	  if ( !dd->axes || dd->axes[0] < nvals  )
	  {
	    free( dd->aux );
	    dd->axes[1] = nvals;
	    iptr = malloc( dmi_descriptor_get_bytes( dd ) * nvals );      
	    is_alloced = UT_TRUE;
	  }
        }
        else
        {
	  /* single value, can just use the tmp space */
	  iptr = &tmp;
        }
        ar_mix_cast_array( vals, type, dd->type, iptr, nvals );
     } 
     else 
     {
       /* type matches.. set pointer to values. */
       iptr = vals;
     }

     dmp_key_write( block, dd->name, dd->type, iptr, nvals, dd->unit, dd->desc, UT_TRUE );

     if ( is_alloced )
     {
       free( iptr );
     }
     break;

     default:
       dmu_err_set( DM_ERR_GENERIC, "dmSetVectorArrays, descriptor type not recognized." );
     break;
   }

  return ngot;
}


/***************************************************************/
/* Function:  dmGetVectorArrays                                */
/*                                                             */
/* Description:                                                */
/*    Extract values from multiple rows, for provided data     */
/*    descriptor.                                              */  
/*                                                             */
/* Input parameters:                                           */
/*       dmDescriptor  *dd    - descriptor                     */
/*       dmDataType     type  - datatype of values provided    */
/*                              (may be diff. from desc.)      */
/*       long           dim   - number of vector cpts in vals  */
/*       long           row   - start row                      */
/*       long           nrows - number of rows of data to get  */
/*       long           start - start index of array to extract*/
/*                              (subarray, typically 1)        */
/*       long           nvals - number of array elements       */
/*                                                             */
/* Output parameters:                                          */
/*       void          *vals  - return values                  */
/*                                                             */
/* Returns:                                                    */
/*        long  ngot - number of rows returned                 */
/*                                                             */
/***************************************************************/
long dmGetVectorArrays( dmDescriptor* dd, dmDataType type, void* vals, 
                        long dim, long row, long nrows, long start, long nvals)
{
  long nitems = 0;
  long ngot   = 1L;
  long stat   = DM_FLAG_RESULT | DM_FLAG_VISIBLE | DM_FLAG_PHYS;

  if ( !dd )
  {
    dmu_err_set( DM_ERR_NULL, "Null argument" );
    return dmFAILURE;
  }

  if ( dd->dim != dim ) 
  {
    /* May be asking to use first component.. */
    if ( dd->dim > 1 && dim == 1 )
      dd = dmi_descriptor_get_cpt( dd, 1 ); 
    else
      return dmFAILURE;
  }

  nitems = nvals;

  switch( dd->descType )
  {
    case dmCOLUMN:

    /* pull values from buffer.. cast if needed */
    ngot = dmp_buffer_get_cast_data( dd, vals, type,
			     row, nrows, start, nitems, NULL, NULL, NULL );
    break;

    case dmCOORD:
      /* transformed (coord) values may or may not be in buffer.. */

      if ( dd->flag & stat )
      {
         /* It's in the buffer as a virtual column */
         ngot = dmp_buffer_get_cast_data( dd, vals, type,
			      row, nrows, start, nitems, NULL, NULL, NULL );
      }
      else
      {
        /* We need to calculate it here on the fly */
        if ( type != DT_REAL8 ) /* Only supported for real8 output */
          return 0;

        ngot = dmp_coord_get_vector( dd, vals, type, row, nrows, start, nitems );
      }
      break;

    case dmKEY:
      nitems = 1;

      if ( dd->axes )
        nitems = dd->axes[0]; 

      if ( nvals < nitems )
        nitems = nvals;

      if ( dd->type == DT_CHAR && type == DT_CHAR )
      {
        long ii = 0;
        for ( ii =0; ii < nitems; ii++ )
	  cs_copy( ((char**)dd->aux)[ii], ((char**)vals)[ii] );
      }
      else
        ar_mix_cast_array( dd->aux, dd->type, type, vals, nitems );

      break;

    default:
      dmu_err_set( DM_ERR_GENERIC, "dmGetVectorArrays, descriptor type not recognized." );
      ngot = -1;  /* 20100630: MCD - set return value to indicate a problem occurred */
      break;
  }
  return ngot;
}


/***************************************************************/
/* Function:  dmp_coord_get_vector                             */
/*                                                             */
/* Description:                                                */
/*    Extracts source values for the coordinate descriptor and */
/*    applies the coordinate transform to that data.           */
/*                                                             */
/*    If the descriptor is a vector component, the values for  */
/*    the full vector (parent) must be extracted and           */
/*    transformed.  The appropriate component of the results   */
/*    are returned.                                            */
/*                                                             */
/* Input parameters:                                           */
/*       dmDescriptor  *dd    - coordinate descriptor          */
/*       dmDataType     otype - datatype of values             */
/*                              (may be diff. from desc.)      */
/*       long           start_row - start row                  */
/*       long           nrows - number of rows of data to get  */
/*       long           start_val - start index of array to get*/
/*                              (subarray, typically 1)        */
/*       long           nvals - number of array elements       */
/*                                                             */
/* Output parameters:                                          */
/*       void          *outvals  - return values               */
/*                                                             */
/* Returns:                                                    */
/*        long  ngot - number of rows returned                 */
/*                                                             */
/***************************************************************/
long dmp_coord_get_vector( dmDescriptor* dd, void* outvals, dmDataType otype, 
			   long start_row, long nrows,
			   long start_val, long nvals )
{

  dmDescriptor *vector = NULL;   /* coord vector descriptor */
  dmDescriptor *parent = NULL;   /* src (parent) descriptor */

  long cpt1  = 0;                /* used in selecting output component */
  long ncpts = 0;                /* # of components in output          */  
  long ngot  = 0;                /* # of values extracted              */

                                 /* Source (Parent) info               */
  long nsrc = 0;                 /*   array size of value(s)           */
  dmDataType itype = DT_UNK;     /*   datatype of source values        */
  long isize = 0;                /*   size in bytes of each value      */
  void *srcvals = NULL;          /*   source value storage             */

                                 /* Output (Coord) info                */
  long  nres  = 0;               /*   array size of value(s)           */
  long  osize = 0;               /*   size in bytes of each value      */
  void *resvals = NULL;          /*   xformed value storage            */
  char *iptr    = NULL;          /*   ptr into resvals                 */
  char *optr    = NULL;          /*   ptr into outvals                 */
  long  orow_offset = 0;         /*   shift for output start this row  */

  long ii    = 0;                /* counter                            */
  long irow  = 0;                /* counter                            */
  long cptno = 0;                /* counter                            */
  long rnum  = 0;                /* current row number                 */


  if ( !dd )
    { return 0; }

  if ( dd->cptno > 0 )
  {
    /* input coord is a vector component.. need full vector */
    vector = dd->parent;    /* full vector descr.    */
    cpt1   = dd->cptno;     /* out cpt matchs dd     */
    ncpts  = 1;             /* # output components   */
  }
  else
  {
    vector = dd;            /* input dd is complete  */
    cpt1   = 1;             /* first cpt is first    */
    ncpts = vector->dim;    /* # cts matches dd      */
  }

  /* get source value descriptor.. parent of virtual (coord) vector */
  parent = dmi_coord_get_parent( vector );

  /* setup intermediate space for source and transformed values     */

  nsrc = dmi_descriptor_get_array_size( parent );  /* # src values per     */
  nres = dmi_descriptor_get_array_size( vector );  /* # xformed values per */

  if ( nsrc == 0 )
   { nsrc = 1; }

  if ( nres == 0 )
    { nres = 1; }

  itype = parent->type;                     /* datatype of src values       */
  isize = ar_mix_type_size( itype );        /* size in bytes of src value   */
  osize = ar_mix_type_size( otype );        /* size in bytes of out value   */

  srcvals = calloc( nsrc*parent->dim, isize); /* hold 1 row of vector data  */
  resvals = calloc( nres*vector->dim, osize); /* hold 1 row of vector data  */

  ngot = 0;

  /* loop over specified rows.. */
  for ( irow = 0; irow < nrows; irow++ )
  {
    rnum = start_row + irow;

    /* Extract source values for this row */
    dmp_buffer_get_cast_data( parent, srcvals, itype, rnum, 1L, 
			      1L, nsrc, NULL, NULL, NULL );

    /* Apply coordinate transform.. */
    dmi_coord_calc( vector, srcvals, itype, otype, resvals );

    /* we now have a full set of transformed vector values for this row */
    /* pull out the desired component(s) and/or subarray to output ptr  */

    /* set offset into output ptr for this row of values */
    orow_offset = irow*(nvals*ncpts*osize);

    /* loop over components to be output */
    for ( ii = 1; ii <= ncpts; ii++ )
    {
      cptno = cpt1 + ii - 1 ;

      /* set ptr to this component in xformed values                        */
      /*            start   +    component offset    +  subarray offset     */
      iptr = (char*)resvals + (cptno-1)*(nres*osize) + (start_val-1)*(osize);

      /* set ptr proper location in output                                  */
      /*            start   +    row offset  +    component offset          */
      optr = (char*)outvals + orow_offset    + (ii-1)*(nvals*osize);

      /* transfer values */
      memcpy( optr, iptr, nvals*osize );

      ngot += nvals;
    }

  } /* end row loop */

  free( srcvals );
  free( resvals );

  return ngot;
}


/***************************************************************/
/* Function:  dmCopyGeneric                                    */
/*                                                             */
/* Description:                                                */
/*    Copy data for the current row from dd1 to dd2.           */
/*    They can be in different blocks.                         */
/*                                                             */
/* Input parameters:                                           */
/*       dmDescriptor  *dd1    - source descriptor             */
/*                                                             */
/* Output parameters:                                          */
/*       dmDescriptor  *dd2    - destination descriptor        */
/*                                                             */
/* Returns:                                                    */
/*        int  dmSUCCESS or dmFAILURE                          */
/*                                                             */
/***************************************************************/
int dmCopyGeneric( dmDescriptor* dd1, dmDescriptor* dd2 )
{
  void *val   = NULL;
  long  n     = 0;
  long  size  = 0;
  long  nvals = 0;
  long  row1  = 0;
  long  row2  = 0;
  long  nrows = 1;
  long  start = 1;
  dmDataType type = DT_UNK;

  if ( !dd1 || !dd2 ) 
    return dmFAILURE;

  type = dd1->type;

  nvals = dmi_descriptor_get_array_size( dd1 );
  size  = dmi_descriptor_get_bytes( dd1 );
  size  = size * nvals * dd1->dim;

  val = malloc( size );

  row1 = dmi_block_get_current_row( (dmBlock*)dd1->block );
  row2 = dmi_block_get_current_row( (dmBlock*)dd2->block );

  n = dmp_buffer_get_cast_data( dd1, val, type, row1, nrows, start, nvals, NULL, NULL, NULL );

  if ( n > 0 )
    dmp_buffer_set_cast_data( dd2, val, type, row2, nrows, start, nvals );

  free ( val );

  return dmSUCCESS;
}


/***************************************************************/
/* Function:  dmTableOpenCopyColumnList                        */
/*                                                             */
/* Description:                                                */
/*    Open columns whose descriptors  are in src. Return list  */
/*    of open column descriptor pointers.                      */
/*                                                             */
/* Input parameters:                                           */
/*       dmBlock       *destblock - block                      */
/*       dmDescriptor **src       - input list of columns      */
/*       long           n         - number of columns          */
/*                                                             */
/* Output parameters:                                          */
/*       none                                                  */
/*                                                             */
/* Returns:                                                    */
/*        dmDescriptor **dest - list of open col descriptors   */
/*                                                             */
/***************************************************************/
dmDescriptor** dmTableOpenCopyColumnList( dmBlock* destblock, dmDescriptor** src, long n )
{
  dmDescriptor  **dest = NULL;
  long ii = 0;

  if ( n <= 0 ) 
    return NULL;

  dest = (dmDescriptor**)calloc( n, UT_SZ_P );

  for ( ii = 0; ii < n; ii++ )
    dest[ii] = dmTableOpenColumn( destblock, src[ii]->name );
   
  return dest;
}


/***************************************************************/
/* Function:  dmTableCopyColumnListData                        */
/*                                                             */
/* Description:                                                */
/*    Copy list of column descriptors to new list              */
/*                                                             */
/* Input parameters:                                           */
/*       dmDescriptor **src       - input list of columns      */
/*       long           n         - number of columns          */
/*                                                             */
/* Output parameters:                                          */
/*       dmDescriptor **dest      - return list of copied cols */
/*                                                             */
/* Returns:                                                    */
/*        int  retval = dmSUCCESS or dmFAILURE                 */
/*                                                             */
/***************************************************************/
int dmTableCopyColumnListData( dmDescriptor** src, dmDescriptor** dest, long n )
{
  long ii = 0;
  int retval = dmSUCCESS;

  if ( !src || !dest || n == 0) 
    return retval;

  for ( ii = 0; ii < n; ii++ )
  {
    if ( src[ii] && dest[ii] )
      retval = dmCopyGeneric( src[ii], dest[ii] );

    if ( retval != dmSUCCESS )
      return dmFAILURE;
  }

  return retval;
}

