/*                                                                
**  Copyright (C) 1997-2001,2003-2004,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/* dmcolumn.c                                                    */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmlibP.h"

/***************************************************************/
/* Function:  dmColumnCreateGeneric                            */
/*                                                             */
/* Description:                                                */
/*     Create column given name, type and other elements uesd  */
/*     in data descriptor                                      */
/*                                                             */
/* Input parameters:                                           */
/*       dmBlock      *block    - block pointer                */
/*       char         *name     - column name                  */
/*       dmDataType    type     - data type                    */
/*       long          length   - size                         */
/*       char         *unit     - unit string                  */
/*       char         *desc     - description                  */
/*       dmElementType etype    - element type                 */
/*       char        **cptNames - component names              */
/*       long          vdim     - array dim                    */
/*       long         *axesp    - axes array                   */
/*       long          naxesp   - number of axes               */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*     dmDescriptor *column - pointer to new column descriptor */
/*                                                             */
/***************************************************************/
dmDescriptor* dmColumnCreateGeneric(dmBlock *block, char *name, dmDataType type, 
		      long length, char *unit, char *desc, 
		      dmElementType etype, char **cptNames,
		      long vdim, long* axesp, long naxesp )
{
  char *hint  = NULL;
  long  naxes = naxesp;
  long *axes  = axesp;
  long  daxes = 1;

  /* In old DM, scalars had naxes = 0: this is required for e.g. dmpaste */
  if ( naxes == 0 && block && block->type == dmTABLE )
  {
    naxes = 1;
    axes = &daxes;
   }

  /* Make sure that a BIT column is large enough to hold the given number of bits */
  if ( naxes == 1 && type == DT_BIT && length > 0 )
  {
    long nbytes = ( length - 1 ) / 8 + 1;
    if ( axes && nbytes > axes[0] )
      axes[0] = nbytes;
  }

  return dmColumnCreateGen( block, name, type, length, unit, desc, etype, cptNames, vdim, axes, naxes, dmFALSE, hint );
}

/***************************************************************/
/* Function:  dmp_image_default_col                            */
/*                                                             */
/* Description:                                                */
/*     Create generic image column in block                    */
/*                                                             */
/*                                                             */
/* Input parameters:                                           */
/*       dmBlock      *block    - block pointer                */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*     dmDescriptor *image - pointer to new image   descriptor */
/*                                                             */
/***************************************************************/
dmDescriptor* dmp_image_create_default_col( dmBlock* block )
{
  long *axes = NULL;
  long  naxes = 0;

  if ( block->type != dmIMAGE )
    return NULL;
  return dmColumnCreateGen( block, block->name, DT_INT4, 0, NULL, NULL, dmVALUE, NULL, 1, axes, naxes, dmFALSE, NULL );
}


/***************************************************************/
/* Function:  dmColumnCreateGen                                */
/*                                                             */
/* Description:                                                */
/*     Create column given name, type and other elements uesd  */
/*     in data descriptor                                      */
/*                                                             */
/* Input parameters:                                           */
/*       dmBlock      *block    - block pointer                */
/*       char         *name     - column name                  */
/*       dmDataType    type     - data type                    */
/*       long          length   - size                         */
/*       char         *unit     - unit string                  */
/*       char         *desc     - description                  */
/*       dmElementType etype    - element type                 */
/*       char        **cptNames - component names              */
/*       long          vdim     - array dim                    */
/*       long         *axes     - axes array                   */
/*       long          naxes    - number of axes               */
/*       dmBool        varlen   - variable length flag         */
/*       char         *hint     - hint array                   */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*     dmDescriptor *dd - pointer to new column descriptor     */
/*                                                             */
/***************************************************************/
dmDescriptor* dmColumnCreateGen(dmBlock *block, char *name, dmDataType type, 
		      long length, char *unit, char *desc, 
		      dmElementType etype, char **cptNames,
		      long vdim, long* axes, long naxes, dmBool varlen, char* hint )
{
  dmDescriptor *dd = NULL;
  dmKernel      kernel = NULL;
  dmDataset    *ds = NULL;

  if ( !block )
    return NULL;

  ds = block->ds;
  kernel = ds->kernel;

  if ( !dmi_block_writeable( block ))
  {
    dmu_err_set( DM_ERR_GENERIC, "Block not writeable" );
    return NULL;
  }

  /* Must flush buffer before creating a new column */
  dmp_buffer_flush( block );

  /* Image can have only one 'column' */
  if ( block->type == dmIMAGE && block->nrcols > 0 )
  {
    kernel->DescriptorFree( block->rawcols[0] );
    dmi_delete_col( block->rawcols[0] );
  }

  dd = dmi_column_create( block, name, type, length, unit, desc, etype, cptNames, vdim, axes, naxes, varlen );

  /* Add it to the kernel structures and write out the appropriate kernel info */
  kernel->ColCreate( block, dd, NULL, hint );
  dmi_block_set_update( block );
  dmp_buffer_flush( block );

  return dd;
}


/***************************************************************/
/* Function:  dmColumnGetNo                                    */
/*                                                             */
/* Description:                                                */
/*     Deprecated routine.                                     */
/*     Return the total number of scalar components (raw       */
/*     column number) up to and including the descritptor      */
/*     which is assumed to be a column or column component.    */
/*                                                             */
/* Input parameters:                                           */
/*       dmDescriptor  *dd - descriptor                        */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    long total - number of scalar components                 */
/*                                                             */
/***************************************************************/
long dmColumnGetNo( dmDescriptor* dd )
{
  long no    = 0;
  long total = 0;
  dmBlock      *block  = NULL;
  dmDescriptor *col    = NULL;
  dmDescriptor *vector = NULL;

  if ( !dd ) 
    return 0;

  block = (dmBlock*)dd->block;
 
  if ( dd->cptno > 0 )
  {
    vector = dd->parent;
    total = dd->cptno;
  }
  else
  {
    vector = dd;
    total = 1;
  }

  for ( no = 1; no < vector->no; no++ )
  {
    col = block->cols[no-1];
    total += col->dim;
  }
  return total; 
}

