/*                                                                
**  Copyright (C) 1997-2001,2003-2004,2007-2008  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/* dmdescriptor.c                                                */
/*   Mostly API accessor functions for DM Descriptor.            */
/*                                                               */
/*****************************************************************/
#include "dmlibP.h"


/***************************************************************/
/* Function:  dmp_flush_cols                                   */
/*                                                             */
/* Description:                                                */
/*   Flush table column descriptors                            */
/*                                                             */
/* Input parameters:                                           */
/*    dmBlock* block      - Pointer to current block           */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    none                                                     */
/*                                                             */
/***************************************************************/
void dmp_flush_cols( dmBlock* block )
{
  long ii;
  dmKernel kernel;
  dmDescriptor* dd;

  /* Check that the block can be updated */
  if ( !dmi_block_writeable( block ))
    return;

  /* only need to flush columns for TABLE blocks */
  if ( dmi_block_get_type( block ) == dmIMAGE)
    return;

  kernel = block->ds->kernel;

  /* Flush each column that needs updating */
  /* and reset update flag.                */
  for ( ii = 0; ii < block->ncols; ii++ )
  {
    dd = block->cols[ii];
    if ( dd && dmi_descriptor_check_update( dd ))
    {
      kernel->ColFlush( dd );
      dmi_descriptor_clear_update( dd );
    }
  }
}

/***************************************************************/
/* Function:  dmSetName                                        */
/*                                                             */
/* Description:                                                */
/*   Set descriptor name field from provided value.            */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor to update                */
/*    char* name         = value to assign                     */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*   dmSUCCESS                                                 */
/*                                                             */
/***************************************************************/
dmErrCode dmSetName(dmDescriptor *dd, char* name)
{
  dmi_descriptor_set_name( dd, name );
  dmi_descriptor_set_update( dd );

  return dmSUCCESS;
}

/***************************************************************/
/* Function:  dmSetCptName                                     */
/*                                                             */
/* Description:                                                */
/*   Set name field of a particular component of the input     */
/*   descriptor uing the provided value.                       */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor to update                */
/*    long  cptno        = which component to assign           */
/*    char* name         = value to assign                     */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*   dmSUCCESS                                                 */
/*                                                             */
/***************************************************************/
dmErrCode dmSetCptName(dmDescriptor *dd, long cptno, char* name)
{
  dmElementType etype;
  dmDescriptor* ddcpt;
 
  etype = dmi_descriptor_get_element_type( dd );
  switch ( etype )
  {
  case dmVALUE:
    ddcpt = dmi_descriptor_get_cpt( dd, cptno );
    dmi_descriptor_set_name( ddcpt, name );
    dmi_descriptor_set_update( dd );
    break;
  case dmRANGE:
    /* fill this in once we decide how to handle eltNames,edim,... */
    break;
  case dmINTERVAL:
    /* fill this in once we decide how to handle eltNames,edim,... */
    break;
  }

 return dmSUCCESS;
}

/***************************************************************/
/* Function:  dmSetUnit                                        */
/*                                                             */
/* Description:                                                */
/*   Set descriptor unit field from provided value.            */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor to update                */
/*    char* name         = value to assign                     */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*   dmSUCCESS                                                 */
/*                                                             */
/***************************************************************/
dmErrCode dmSetUnit( dmDescriptor *dd,char* name)
{
  dmi_descriptor_set_unit( dd, name );
  dmi_descriptor_set_update( dd );

  return dmSUCCESS;
}

/***************************************************************/
/* Function:  dmSetDisp                                        */
/*                                                             */
/* Description:                                                */
/*   Set descriptor display field from provided value.         */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor to update                */
/*    char* name         = value to assign                     */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*   dmSUCCESS                                                 */
/*                                                             */
/***************************************************************/
dmErrCode dmSetDisp(dmDescriptor *dd,char* name)
{
  dmi_descriptor_set_disp( dd, name );
  dmi_descriptor_set_update( dd );

  return dmSUCCESS;
}

/***************************************************************/
/* Function:  dmSetDesc                                        */
/*                                                             */
/* Description:                                                */
/*   Set descriptor description from provided value.           */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor to update                */
/*    char* name         = value to assign                     */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*   dmSUCCESS                                                 */
/*                                                             */
/***************************************************************/
dmErrCode dmSetDesc(dmDescriptor *dd,char* name)
{
  dmi_descriptor_set_desc( dd, name );
  dmi_descriptor_set_update( dd );

  return dmSUCCESS;
}

/***************************************************************/
/* Function:  dmDescriptorDelete                               */
/*                                                             */
/* Description:                                                */
/*   Delete a descriptor. Only partially implemented           */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor to update                */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*   dmSUCCESS - descriptor successfully deleted               */
/*   dmFAILURE - unimplemented descriptor type                 */
/*                                                             */
/***************************************************************/
dmErrCode dmDescriptorDelete( dmDescriptor* dd )
{
  dmBlock* block  = NULL;
  dmDataset* ds   = NULL;
  dmKernel kernel = NULL;
  dmErrCode status = dmSUCCESS;

  if ( dd ) 
  {
    block = (dmBlock*)dd->block;

    /* Get kernel */
    if ( block )
    {
      ds = block->ds;
      if ( ds )
	kernel = ds->kernel;
    }
    if ( !kernel ) return dmFAILURE;

    switch( dd->descType )
    {
    case dmKEY:
      kernel->KeyDelete( dd );
      dmi_block_set_update( block );
      break;
    default:
      status = dmFAILURE;
      dmu_err_set( DM_ERR_GENERIC, "Descriptor delete not implemented" );
    }
  }

  return status;
}

/***************************************************************/
/* Function:  dmGetCpt                                         */
/*                                                             */
/* Description:                                                */
/*   Get a particular component descriptor                     */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*    long cptno         = which component to extract          */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*   dmDescriptor*   - requested component descriptor          */
/*                                                             */
/***************************************************************/
dmDescriptor* dmGetCpt(const dmDescriptor *dd, long cptNo)
{
  return dmi_descriptor_get_cpt( dd, cptNo );
}

/***************************************************************/
/* Function:  dmGetName                                        */
/*                                                             */
/* Description:                                                */
/*   Get name field from descriptor.                           */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*    long  maxlen       = maximum length for returned name    */
/*                                                             */
/* Output parameters:                                          */
/*    char* name         = copy of name (user provided memory) */
/*                                                             */
/* Returns:                                                    */
/*   dmSUCCESS -                                               */
/*   dmFAILURE - Descriptor has no name (blank)                */
/*                                                             */
/***************************************************************/
dmErrCode dmGetName( const dmDescriptor* dd, char* name, const long maxlen )
{
  char* dd_name;

  dd_name = dmi_descriptor_get_name( dd );
  cs_copy_siz( dd_name, name, maxlen );

  if ( cs_is_blank( name ))
    return dmFAILURE;

 return dmSUCCESS;
}

/***************************************************************/
/* Function:  dmGetDataType                                    */
/*                                                             */
/* Description:                                                */
/*   Get the datatype of the input descriptor.                 */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmDataType  - numerical DataType key for the descriptor  */
/*                                                             */
/***************************************************************/
dmDataType dmGetDataType( const dmDescriptor* dd )
{
  return dmi_descriptor_get_datatype( dd );
}

/***************************************************************/
/* Function:  dmGetElementType                                 */
/*                                                             */
/* Description:                                                */
/*   Get the element type of the input descriptor.             */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmElementType  - numerical Type key for the descriptor   */
/*                                                             */
/***************************************************************/
dmElementType dmGetElementType( const dmDescriptor* dd )
{
  return dmi_descriptor_get_element_type( dd );
}

/***************************************************************/
/* Function:  dmGetArraySize                                   */
/*                                                             */
/* Description:                                                */
/*   Get the total numer of elements in a descriptor value.    */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    long arrsize - total number of elements                  */
/*                                                             */
/***************************************************************/
long dmGetArraySize( const dmDescriptor* dd )
{
  return dmi_descriptor_get_array_size( dd );
}

/***************************************************************/
/* Function:  dmGetCurrentArraySize                            */
/*                                                             */
/* Description:                                                */
/*   If dd is variable-length, return the number of elements   */
/*   in the current row.                                       */
/*   Otherwise, return dmi_descriptor_get_array_size( dd ).    */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    long arrsize - total number of elements in current row.  */
/*                                                             */
/***************************************************************/
long dmGetCurrentArraySize( const dmDescriptor* dd )
{
  dmBlock* block;
  dmDescriptor* length_dd;

  if ( !dd )
    return 0;

  if ( !dmi_descriptor_is_varlen( dd ) || 
       NULL == ( block = dd->block )   ||
       NULL == ( length_dd = dmi_buffer_get_varlen_length_col( block, (dmDescriptor*)dd ) ) )
    return dmi_descriptor_get_array_size( dd );

 return dmGetScalar_l( length_dd );
}

/***************************************************************/
/* Function:  dmDescriptorGetLength                            */
/*                                                             */
/* Description:                                                */
/*   Get the 'length' of a descriptor.                         */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    long length - descriptor 'length' field                  */
/*                                                             */
/***************************************************************/
long dmDescriptorGetLength( const dmDescriptor* dd )
{
  return dmi_descriptor_get_length( dd );
}

/***************************************************************/
/* Function:  dmGetDesc                                        */
/*                                                             */
/* Description:                                                */
/*   Get a copy of the description from the descriptor.        */
/*   (no more than maxlen chars).                              */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*    long maxlen        =  max length of returned description.*/
/*                                                             */
/* Output parameters:                                          */
/*   char* desc  -  copy of descriptor description             */
/*                   (user provided memory)                    */
/*                                                             */
/* Returns:                                                    */
/*   dmSUCCESS                                                 */
/*                                                             */
/***************************************************************/
dmErrCode dmGetDesc( const dmDescriptor* dd, char* desc, const long maxlen )
{
  char* dd_desc;

  dd_desc = dmi_descriptor_get_desc( dd );
  cs_copy_siz( dd_desc, desc, maxlen );

  return dmSUCCESS;
}

/***************************************************************/
/* Function:  dmGetUnit                                        */
/*                                                             */
/* Description:                                                */
/*   Get a copy of the unit field from the descriptor.         */
/*   (up to maxlen chars)                                      */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*                                                             */
/* Output parameters:                                          */
/*   char* unit  -  copy of descriptor unit field              */
/*                   (user provided memory)                    */
/*                                                             */
/* Returns:                                                    */
/*   dmSUCCESS                                                 */
/*                                                             */
/***************************************************************/
dmErrCode dmGetUnit( const dmDescriptor* dd, char* unit, const long maxlen )
{
  char* dd_unit;

  dd_unit = dmi_descriptor_get_unit( dd );
  cs_copy_siz( dd_unit, unit, maxlen );

  return dmSUCCESS;
}


/***************************************************************/
/* Function:  dmGetDisp                                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*                                                             */
/* Output parameters:                                          */
/*   char* desc  -  copy of descriptor display field           */
/*                   (user provided memory)                    */
/*                                                             */
/* Returns:                                                    */
/*   dmSUCCESS                                                 */
/*                                                             */
/***************************************************************/
dmErrCode dmGetDisp( const dmDescriptor* dd, char* desc, const long maxlen )
{
  char* disp;

  disp = dmi_descriptor_get_disp( dd );
  cs_copy_siz( disp, desc, maxlen );

  return dmSUCCESS;
}

/***************************************************************/
/* Function:  dmGetCptName                                     */
/*                                                             */
/* Description:                                                */
/*   Get name field from a specific component of the           */
/*   provided descriptor. (up to maxlen chars).                */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*    long cptno         = which component to extract from.    */
/*                                                             */
/* Output parameters:                                          */
/*    char* name         = copy of name (user provided memory) */
/*                                                             */
/* Returns:                                                    */
/*   dmSUCCESS                                                 */
/*                                                             */
/***************************************************************/
dmErrCode dmGetCptName( const dmDescriptor* dd, const long cptNo,
			char* cptName, const long maxlen ) 
{
  char* name;

  name = dmi_descriptor_get_cpt_name( dd, cptNo );
  cs_copy_siz( name, cptName, maxlen );

  return dmSUCCESS;
}

/***************************************************************/
/* Function:  dmDescriptorGetBlock                             */
/*                                                             */
/* Description:                                                */
/*   Returns the pointer to the block from which this          */
/* descriptor comes.                                           */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*   dmBlock* block      - parent block                        */
/*                                                             */
/***************************************************************/
dmBlock* dmDescriptorGetBlock ( const dmDescriptor* dd )
{
  return dmi_descriptor_get_block( dd );
}


/***************************************************************/
/* Function:  dmGetElementDim                                  */
/*                                                             */
/* Description:                                                */
/*   Return the combined vector/element dimension              */
/*     = total number of components                            */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*                                                             */
/* Output parameters:                                          */
/*   none                                                      */
/*                                                             */
/* Returns:                                                    */
/*   long ncpts          - number of components                */
/*                                                             */
/***************************************************************/
long dmGetElementDim( const dmDescriptor* dd )
{
  return dmi_descriptor_get_element_dim( dd );
}

/***************************************************************/
/* Function:  dmGetArrayDim                                    */
/*                                                             */
/* Description:                                                */
/*   Inside the new DM we use naxes = 1 for a scalar.          */
/*   In the old DM we used naxes = 0 (now we only use that     */
/*   for a null image ). In this routine, if dmold is set,     */
/*   we reproduce the old behaviour.                           */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*   long naxes          - Dimensionality of value             */
/*                                                             */
/***************************************************************/
long dmGetArrayDim( const dmDescriptor* dd )
{ 
  long naxes;

  naxes =  dmi_descriptor_get_array_dim( dd );
  if (( dm_global & DM_GLOBAL_OLD ) &&
        naxes == 1 && dd->axes[0] <= 1 )
    naxes = 0;

  return naxes;
}

/***************************************************************/
/* Function:  dmGetArrayDimensions                             */
/*                                                             */
/* Description:                                                */
/*   Get the dimensionality array from the descriptor..        */
/*   providing the size of each dimension.                     */
/*                                                             */
/* Input parameters:                                           */
/*   dmDescriptor* dd    = descriptor                          */
/*                                                             */
/* Output parameters:                                          */
/*   long** axes         - array of sizes                      */
/*                                                             */
/* Returns:                                                    */
/*   long naxes          - size of dimarray                    */
/*                                                             */
/***************************************************************/
long dmGetArrayDimensions( const dmDescriptor* dd, long** axisLens )
{
  (void)dmi_descriptor_get_array_dimensions( dd, axisLens );

  return dmGetArrayDim( dd );
}

/***************************************************************/
/* Function:  dmDescriptorGetNoCoords                          */
/*                                                             */
/* Description:                                                */
/*   Get the number of coordinate systems attached to the      */
/*   provided descriptor.                                      */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*   long n               - number of attached coordinates     */
/*                                                             */
/***************************************************************/
long dmDescriptorGetNoCoords( const dmDescriptor* dd )
{
  return dmi_descriptor_get_no_coords( dd );
}

/***************************************************************/
/* Function:  dmDescriptorGetCoord                             */
/*                                                             */
/* Description:                                                */
/*   Get the first (primary) coordinate system assigned to     */
/*   the provided descriptor.                                  */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmDescriptor* coord = coordinate descriptor.             */
/*                                                             */
/***************************************************************/
dmDescriptor* dmDescriptorGetCoord( const dmDescriptor* dd )
{
  return dmi_descriptor_get_coord_no( dd, 1 );
}

/***************************************************************/
/* Function:  dmDescriptorGetCoordNo                           */
/*                                                             */
/* Description:                                                */
/*   Get a particular coordinate system from the provided      */
/*   descriptor.                                               */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*    long no            = which coordinate system.            */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmDescriptor* coord = coordinate descriptor.             */
/*                                                             */
/***************************************************************/
dmDescriptor* dmDescriptorGetCoordNo( const dmDescriptor* dd, const long no )
{
  return dmi_descriptor_get_coord_no( dd, no );
}


/***************************************************************/
/* Function:  dmMapDataTypeName                                */
/*                                                             */
/* Description:                                                */
/*   Get the string equivalent for numerical datatype key.     */
/*                                                             */
/* Input parameters:                                           */
/*    dmDataType type    = numerical datatype key              */
/*                                                             */
/* Output parameters:                                          */
/*    char* pbuf         = String equivalent name of datatype. */
/*                                                             */
/* Returns:                                                    */
/*    none                                                     */
/*                                                             */
/***************************************************************/
void dmMapDataTypeName( dmDataType type, char* pbuf )
{
  char buf[CARD_SIZE];

  /* If OLD behaviour is requested, return "String" for */
  /* what is now called "Text".                         */
  if ( (dm_global & DM_GLOBAL_OLD) && type == DT_CHAR )
    cs_copy( "String", buf );
  else
    ar_mix_type_name( type, buf, CARD_SIZE );

  cs_copy( buf, pbuf );
}

/***************************************************************/
/* Function:  dmGetDataTypeName                                */
/*                                                             */
/* Description:                                                */
/*   Get the string representation of the datatype from the    */
/*   provided descriptor.  Value returned in new memory.       */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    char* dtype        - Datatype of descriptor as string    */
/*                                                             */
/***************************************************************/
char* dmGetDataTypeName( const dmDescriptor* dd )
{
  char buf[CARD_SIZE];
  if ( !dd )
    return NULL;

  dmMapDataTypeName( dd->type, buf );

  return cs_dup( buf );
}

/***************************************************************/
/* Function:  dmDescriptorGetType                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmDescriptorType type  = DM Type of descriptor.          */
/*                              (dmCOLUMN, dmKEY, dmCOORD,...) */
/*                                                             */
/***************************************************************/
dmDescriptorType dmDescriptorGetType( const dmDescriptor* dd )
{
  dmDescriptorType type;

  type = dmi_descriptor_get_type( dd );
  if( type == dmBADDESC ) return type;

  /* 
   * This code fragment for back compatibility with CIAO2.
   * In the true DM model, a dmIMAGEDATA is just a dmCOLUMN;
   * should test for block type being image, not descriptor type 
   * being imagedata. Delete this code when CIAO tools no longer use
   * dmIMAGEDATA.  Alas, dmlist is an example of an offending tool.
   *
   * Also, in the new model, the scaled columns are in fact dmCOORDs
   * (only the unscaled columns, which you don't usually see, are dmCOLUMNs).
   * Should amend this so that for unit scaling they are true dmCOLUMNs.
   */

  if ( dd->flag & DM_FLAG_IMG )
    type = dmIMAGEDATA;
  else if  ( dd->flag & DM_FLAG_VISIBLE )  /* Or DM_FLAG_RAW? */
    type = dmCOLUMN;
  else if ( dd->cptno > 0 )
  {
    dmDescriptor* parent = dd->parent;
    if ( parent &&  ( parent->flag & DM_FLAG_VISIBLE )) /* Component of column */
      type = dmCOLUMN;
  }

 return type;
}

/***************************************************************/
/* Function:  dmDescriptorCompare                              */
/*                                                             */
/* Description:                                                */
/*   Compares two descriptors.                                 */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*    dmDescriptor* dd   = descriptor                          */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*   dmbool equal        = boolean results of compare          */
/*                                                             */
/***************************************************************/
dmBool dmDescriptorCompare( dmDescriptor* dd1, dmDescriptor* dd2, long mask )
{
  int result;

  result = dmi_descriptor_compare( dd1, dd2, mask );

  /* Now test things that need the kernel */
  if ( result )
    result = dmp_descriptor_compare( dd1,dd2, mask );

  return result;
}

/***************************************************************/
/* Function:  dmp_descriptor_compare                           */
/*                                                             */
/* Description:                                                */
/*   Does/invokes kernel level comparisons of descriptors.     */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*    dmDescriptor* dd   = descriptor                          */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*   dmbool equal        = boolean results of compare          */
/*                                                             */
/***************************************************************/
dmBool dmp_descriptor_compare( dmDescriptor* dd1, dmDescriptor* dd2, long mask )
{
  double val1 = 0;
  double val2 = 0;
  long size1;
  long size2;
  char cval1[dmMAXSTR]= "";
  char cval2[dmMAXSTR]= "";


  if ( mask & dmDESCVALUE )
  {
    /* Equivalent descriptors must have the same dimensionality  */
    /* and axis lengths                                          */

    size1 = dmi_descriptor_get_array_size( dd1 );
    size2 = dmi_descriptor_get_array_size( dd2 );

    if ( (dd1->type == DT_CHAR || dd2->type == DT_CHAR) && 
	 (dd1->type != dd2->type) )
      return UT_FALSE;

    if ( size1 == 1 )
    {
      if ( dd1->type == DT_CHAR )
      {
	dmGetScalar_c( dd1, cval1, dmMAXSTR );     
	dmGetScalar_c( dd1, cval2, dmMAXSTR );     
	if ( dmGetError()!= dmSUCCESS || !cs_uceq( cval1, cval2 ))
	  return UT_FALSE;
      } 
      else 
      {
	val1 = dmGetScalar_d( dd1 );
	val2 = dmGetScalar_d( dd2 );
	if ( dmGetError()!= dmSUCCESS || val1 != val2 )
	  return UT_FALSE;
      }
    }
    else 
    {
      /* Could implement array comparison */
      return UT_FALSE;
    }
  }
  return UT_TRUE;
}

/***************************************************************/
/* Function:  dmDescriptorIsVarLength                          */
/*                                                             */
/* Description:                                                */
/*   Is descriptor a variable-length array?                    */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor* dd   = descriptor                          */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*   dmbool res          = dmTRUE if variable length descr.    */
/*                                                             */
/***************************************************************/
dmBool dmDescriptorIsVarLength( const dmDescriptor* dd )
{
  return dmi_descriptor_is_varlen( dd );
}
