/*                                                                
**  Copyright (C) 1997-2001,2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/* dmimagedata.c                                                 */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmlibP.h"


/***************************************************************/
/*                                                             */
/* Function:  dmp_image_set_subarray                           */
/*                                                             */
/* Description:                                                */
/*   Set image subsection (single region)                      */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *dd     - column descriptor                */
/*    long          *llpix  - lower left pixel                 */
/*    long          *urpix  - upper right pixel                */
/*    dmDataType     type   - data type                        */
/*    void          *values - return buffer                    */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    int   status - dmFAILURE or dmSUCCESS                    */
/*                                                             */
/***************************************************************/
int dmp_image_set_subarray(dmDescriptor* dd, 
                             long* llpix, long* urpix, dmDataType type, void* values)
{
  long nrows  = 0;
  long nvrows = 1;
  long vrowno = 1;
  long start  = 0;
  long nitems = 0;

  if (!dd)
   return dmFAILURE;

  start  = dmu_pixelno_from_pixel( llpix, dd->axes, dd->naxes );
  nitems = dmu_pixelno_from_pixel( urpix, dd->axes, dd->naxes ) - start + 1;

  /* Put section with set cast data */  

  nrows = dmp_buffer_set_cast_data( dd, values, type, vrowno, nvrows, start, nitems );

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:  dmp_image_get_subarray                           */
/*                                                             */
/* Description:                                                */
/*   Return image subarray                                     */
/*   This function gets called when reading data from a file   */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *dd     - column descriptor                */
/*    long          *llpix  - lower left pixel                 */
/*    long          *urpix  - upper right pixel                */
/*    dmDataType     type   - data type                        */
/*                                                             */
/* Output parameters:                                          */
/*    void          *vals   - return buffer                    */
/*                                                             */
/* Returns:                                                    */
/*    int   status - dmFAILURE or dmSUCCESS                    */
/*                                                             */
/***************************************************************/
int dmp_image_get_subarray(dmDescriptor* dd, long* llpix, long* urpix, dmDataType type, void* vals)
{
  dmBlock *block;

  long currRow = 1;
  long nrows = 1;
  long *dpix = NULL;

  long nItems    = 0;
  long startCell = 1;

  long ngot = 0;
  long ii = 0;

  // Validity checks first

  if (!dd)
   return dmFAILURE;

  /* This function checks validity and returns the  number of cells to copy */
  nItems = dmi_array_check_range( dd, llpix, urpix );

  if ( nItems < 0 )
    return dmFAILURE;

  /* Set up dpix array for use by lower subroutines */

   dpix = malloc( ( dd->naxes + 1 )* UT_SZ_I );

   for ( ii = 0; ii < dd->naxes; ii++ )
     dpix[ii] = 1;


   /* Get starting cell number */
   startCell = dmu_pixelno_from_pixel( llpix, dd->axes, dd->naxes );

   block = (dmBlock*)dd->block;
   currRow = dmi_block_reset_current_row( block );


   /* Get section */

   ngot = dmp_buffer_get_cast_data( dd, vals, type, currRow, nrows, startCell, nItems,
     llpix, urpix, dpix );

   free( dpix );
   return dmSUCCESS;

}



