/*                                                                
**  Copyright (C) 1997-2001,2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/***********************************************************************/
/*                                                                     */
/*  dmkey.c                                                            */
/*                                                                     */
/*                                                                     */
/***********************************************************************/

#include "dmlibP.h"


/***************************************************************/
/* Function:  dmBlockGetKeyList                                */
/*                                                             */
/* Description:                                                */
/*    Return descriptor list of keys in block                  */
/*                                                             */
/* Input parameters:                                           */
/*       const dmBlock *block - block                          */
/*                                                             */
/* Output parameters:                                          */
/*       long          *nkeys - number of descriptors in list  */
/*                                                             */
/* Returns:                                                    */
/*        dmDescriptor  **list - list of key descriptors       */
/*                                                             */
/***************************************************************/
dmDescriptor** dmBlockGetKeyList( const dmBlock* block, long *nkeys )
{
  dmDescriptor **list = NULL;
  long i = 0;

  if ( !block )
  {
    *nkeys = 0;
  }
  else
  {
    *nkeys = block->nkeys;
    list = calloc( *nkeys, UT_SZ_P );

    for ( i = 0; i < block->nkeys; i++ )
       list[i] = block->keys[i];
  }
  return list;
}


/***************************************************************/
/* Function:  dmBlockGetKey                                    */
/*                                                             */
/* Description:                                                */
/*    Return descriptor for key given key number               */
/*                                                             */
/* Input parameters:                                           */
/*             dmBlock *block - block                          */
/*             long     no    - key number                     */
/*                                                             */
/* Output parameters:                                          */
/*       none                                                  */
/*                                                             */
/* Returns:                                                    */
/*        dmDescriptor  *dd - key descriptor                   */
/*                                                             */
/***************************************************************/
dmDescriptor*	dmBlockGetKey( dmBlock* block, long no )
{
  return dmi_block_get_key( block, no );
}


/***************************************************************/
/* Function:  dmBlockMoveToKey                                 */
/*                                                             */
/* Description:                                                */
/*    Move to key location given key name                      */
/*                                                             */
/* Input parameters:                                           */
/*             dmBlock *block - block                          */
/*             char    *name  - key name                       */
/*                                                             */
/* Output parameters:                                          */
/*       none                                                  */
/*                                                             */
/* Returns:                                                    */
/*        dmDescriptor  *dd - key descriptor                   */
/*                                                             */
/***************************************************************/
dmDescriptor*  dmBlockMoveToKey(dmBlock *block, char *name)
{
  return dmi_key_read( block, name, DT_UNK, NULL );
}


/***************************************************************/
/* Function:  dmKeyOpen                                        */
/*                                                             */
/* Description:                                                */
/*    Find key by name and return descriptor                   */
/*                                                             */
/* Input parameters:                                           */
/*             dmBlock *block - block                          */
/*             char    *name  - key name                       */
/*                                                             */
/* Output parameters:                                          */
/*       none                                                  */
/*                                                             */
/* Returns:                                                    */
/*        dmDescriptor  *dd - key descriptor                   */
/*                                                             */
/***************************************************************/
dmDescriptor* dmKeyOpen( dmBlock *block, char *name)
{
  dmDescriptor *dd = NULL; 

  if ( !block ) 
  {
    dmu_err_set( DM_ERR_NULL, "Null argument" );
    return NULL;
  }

  dd = dmi_key_read( block, name, DT_UNK, NULL );

  if ( !dd ) 
    dd = dmi_block_search_public_identifier( block, name );

  return dd;
}


/***************************************************************/
/* Function:  dmp_key_reead                                    */
/*                                                             */
/* Description:                                                */
/*    Read key or possibly column                              */
/*                                                             */
/* Input parameters:                                           */
/*           dmBlock   *block - block                          */
/*     const char      *name  - key name                       */
/*     const dmDataType type  - data type                      */
/*                                                             */
/* Output parameters:                                          */
/*       void  *value - key value                              */
/*                                                             */
/* Returns:                                                    */
/*        dmDescriptor  *dd - key descriptor                   */
/*                                                             */
/***************************************************************/
dmDescriptor* dmp_key_read( dmBlock* block, const char* name, const dmDataType type, void* value )
{
  dmDescriptor *dd = NULL;

  dd = dmi_key_read( block, name, type, value );

  if ( !dd )
  {
    dd = dmi_block_search_public_identifier( block, name );

    if ( dd )
    {
      long nvals = 1;
      long dim   = 1;
      long start = 1;
      dmGetVectorArray( dd, type, value, dim, start, nvals );
    }
  }
  return dd;
}


/***************************************************************/
/* Function:  dmBlockAdvanceKeys                               */
/*                                                             */
/* Description:                                                */
/*    Advance to key position at current position plus numkeys */
/*                                                             */
/* Input parameters:                                           */
/*           dmBlock   *block - block                          */
/*           long       numkeys - number of keys to advance    */
/*                                                             */
/* Output parameters:                                          */
/*       none                                                  */
/*                                                             */
/* Returns:                                                    */
/*        dmDescriptor  *dd - key descriptor                   */
/*                                                             */
/***************************************************************/
dmDescriptor*	dmBlockAdvanceKeys(dmBlock *block, long numkeys)
{
  long          no = 0;
  dmDescriptor *dd = NULL;

  no = block->keyno + numkeys;
  dd = dmi_block_get_key( block, no );
  return dd;
}


/***************************************************************/
/* Function:  dmBlockMoveToKeyNo                               */
/*                                                             */
/* Description:                                                */
/*    Move to position of key at 'no'                          */
/*                                                             */
/* Input parameters:                                           */
/*           dmBlock   *block - block                          */
/*           long       no    - key number                     */
/*                                                             */
/* Output parameters:                                          */
/*       none                                                  */
/*                                                             */
/* Returns:                                                    */
/*        dmDescriptor  *dd - key descriptor                   */
/*                                                             */
/***************************************************************/
dmDescriptor*  dmBlockMoveToKeyNo(dmBlock *block, long no )
{
  dmDescriptor *dd = NULL;

  dd = dmi_block_get_key( block, no );
  return dd;
}


/***************************************************************/
/* Function:  dmp_key_write                                    */
/*                                                             */
/* Description:                                                */
/*     Update DM level key list (check for match)              */
/*     If the key already exists, the descriptor returned is   */
/*     the old one                                             */
/*                                                             */
/* Input parameters:                                           */
/*           dmBlock   *block - block                          */
/*     const char      *name  - key name                       */
/*     const dmDataType type  - data type                      */
/*     const void      *value - key value                      */
/*           long       nvals - number of values               */
/*     const char      *unit  - unit string                    */
/*     const char      *desc  - description string             */
/*     const dmBool     special - flag                         */
/*                                                             */
/* Output parameters:                                          */
/*       none                                                  */
/*                                                             */
/* Returns:                                                    */
/*        dmDescriptor  *dd - key descriptor                   */
/*                                                             */
/***************************************************************/
dmDescriptor* dmp_key_write( dmBlock* block, const char* name, const dmDataType type,
                             const void* value, long nvals, const char* unit, 
                             const char* desc, const dmBool special )
{
  dmDescriptor *dd = NULL;
  dmKernel      kernel;

  if ( !block )
    return NULL;
 
  /* must update the subspace before we write the key */
  if ( special && type == DT_REAL8 && *name == 'T' )
    dmi_block_update_time_range( block, name, *(double*)value );

  kernel = block->ds->kernel;
  dd = kernel->KeyWrite( block, block->keyno, name, type, value, nvals, unit, desc );
  dmi_block_set_update( block );

  return dd;
}

