/*                                                                
**  Copyright (C) 2002-2005,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/* dmlib.c                                                       */
/*                                                               */
/* This file contains public interface routines for specific     */
/* data types which are simple wrappers to generic-type internal */
/* routines.                                                     */
/*                                                               */
/*****************************************************************/

#include "dmlibP.h"


/***************************************************************/
/* Function:  dmColumnCreate                                   */
/*                                                             */
/* Description:                                                */
/*   Creates a column in table block                           */
/*                                                             */
/* Input parameters:                                           */
/*    dmBlock    *table     - Pointer to block                 */
/*    char       *name      - column name                      */
/*    dmDataType  type      - data type of column data         */
/*    long        desclen   - length                           */
/*    char       *unit      - untis                            */
/*    char       *desc      - description                      */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmDescriptor  *dd                                        */
/*                                                             */
/***************************************************************/
dmDescriptor*
dmColumnCreate(dmBlock *table, char *name, dmDataType type, long desclen,
	       char *unit, char *desc)
{
  return dmColumnCreateVector( table, name, type, desclen, unit, desc, &name, 0L );
}


/***************************************************************/
/* Function:  dmColumnCreateArray                              */
/*                                                             */
/* Description:                                                */
/*   Create a datamodel array column.                          */
/*                                                             */
/* Input parameters:                                           */
/*    dmBlock    *table     - Pointer to block                 */
/*    char       *name      - column name                      */
/*    dmDataType  type      - data type of column data         */
/*    long        desclen   - length                           */
/*    char       *unit      - untis                            */
/*    char       *desc      - description                      */
/*    long        arraySize - array size                       */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmDescriptor  *dd                                        */
/*                                                             */
/***************************************************************/
dmDescriptor*
dmColumnCreateArray(dmBlock *table, char *name, dmDataType type, long desclen,
		    char *unit, char *desc, long arraySize)
{
  return dmColumnCreateNDArray( table, name, type, desclen, unit, desc,
                                &arraySize, 1 );
}


/***************************************************************/
/* Function:  dmColumnCreateNDArray                            */
/*                                                             */
/* Description:                                                */
/*   Create an ND array column.                                */
/*                                                             */
/* Input parameters:                                           */
/*    dmBlock    *table     - Pointer to block                 */
/*    char       *name      - column name                      */
/*    dmDataType  dType     - data type of column data         */
/*    long        descLen   - length                           */
/*    char       *unit      - untis                            */
/*    char       *desc      - description                      */
/*    long       *axlen     - axes length                      */
/*    long        naxes     - number of axes                   */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmDescriptor  *dd                                        */
/*                                                             */
/***************************************************************/
dmDescriptor*
dmColumnCreateNDArray(dmBlock *table, char *name, dmDataType dType,
		      long descLen, char *unit, char *desc, long* axlen,
		      long naxes)
{
  return dmColumnCreateGeneric( table, name, dType, descLen, unit, desc, 
                                dmVALUE, &name, 0L,
                                axlen, naxes );
}


/***************************************************************/
/* Function:  dmColumnCreateVarArray                           */
/*                                                             */
/* Description:                                                */
/*   Create a variable array column.                           */
/*                                                             */
/* Input parameters:                                           */
/*    dmBlock    *table     - Pointer to block                 */
/*    char       *name      - column name                      */
/*    dmDataType  type      - data type of column data         */
/*    long        length    - length                           */
/*    char       *unit      - untis                            */
/*    char       *desc      - description                      */
/*    long        arraysize - array size                       */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmDescriptor  *dd                                        */
/*                                                             */
/***************************************************************/
dmDescriptor*
dmColumnCreateVarArray(dmBlock *table, char *name, dmDataType type,
		       long length, char *unit, char *desc, long arraysize)
{
  return dmColumnCreateGen( table, name, type, length, unit, desc, 
                                dmVALUE, &name, 0L,
                                 &arraysize, 1L, dmTRUE, NULL );

}


/***************************************************************/
/* Function:  dmColumnCreateVector                             */
/*                                                             */
/* Description:                                                */
/*   Create a datamodel vector column.  The vector column      */
/*   descriptor is attached to the block while its component   */
/*   column descriptors are listed within it.  If the name     */
/*   field is empty, we create the vector column name from the */
/*   component names: "(cpt1, cpt2)".  Component column        */
/*   descriptors can be manipulated dirtectly by doing a       */
/*   dmGetCpt on a vector column descriptor.                   */
/*   If vdim = 0, make a scalar column                         */
/*                                                             */
/* Input parameters:                                           */
/*    dmBlock    *table     - Pointer to block                 */
/*    char       *name      - column name                      */
/*    dmDataType  dType     - data type of column data         */
/*    long        descLen   - length                           */
/*    char       *unit      - untis                            */
/*    char       *desc      - description                      */
/*    char      **cptNames  - list of component names          */
/*    long        vdim      - vector dim                       */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmDescriptor  *dd                                        */
/*                                                             */
/***************************************************************/
dmDescriptor*
dmColumnCreateVector(dmBlock* table, char* name, dmDataType dType, long descLen,
		     char* unit, char* desc, char** cptNames, long vdim)
{
  return dmColumnCreateGeneric( table, name, dType, descLen, unit, desc, 
                                dmVALUE, cptNames, vdim,
                                NULL, 1L );
}


/***************************************************************/
/* Function:  dmCoordCalc_*                                    */
/*                                                             */
/* Description:                                                */
/*   Returns calculated value for pixel coordinate data        */
/*   Separate functions for each different incoming data       */
/*     type. Last letter of function name indicates data       */
/*     type.                                                   */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *dd    - Pointer to descriptor             */
/*    <data type>   *value - pointer to pixel coordinate data  */
/*                                                             */
/* Output parameters:                                          */
/*    double  *result  - calculated value of world coord data  */
/*                                                             */
/* Returns:                                                    */
/*    int  status    (dmSUCCES or dmFAILURE)                   */
/*                                                             */
/***************************************************************/
int  dmCoordCalc_s(dmDescriptor* dd, short* value, double* result)
{
  return(dmi_coord_calc(dd, (void*)value, DT_INT2, DT_REAL8, result));
}


int  dmCoordCalc_l(dmDescriptor* dd, long* value, double* result)
{
  return(dmi_coord_calc(dd, (void*)value, DT_INT4, DT_REAL8, result));
}


int  dmCoordCalc_f(dmDescriptor* dd, float* value, double* result)
{
  return(dmi_coord_calc(dd, (void*)value, DT_REAL4, DT_REAL8, result));
}


int  dmCoordCalc_d(dmDescriptor* dd, double* value, double* result)
{
  return(dmi_coord_calc(dd, (void*)value, DT_REAL8, DT_REAL8, result));
}


int  dmCoordCalc_ub(dmDescriptor* dd, unsigned char* value, 
				     double* result)
{
  return(dmi_coord_calc(dd, (void*)value, DT_UINT1, DT_REAL8, result));
}


int  dmCoordCalc_us(dmDescriptor* dd, unsigned short* value, 
				     double* result)
{
  return(dmi_coord_calc(dd, (void*)value, DT_UINT2, DT_REAL8, result));
}


int  dmCoordCalc_ul(dmDescriptor* dd, unsigned long* value, 
				     double* result)
{
  return(dmi_coord_calc(dd, (void*)value, DT_UINT4, DT_REAL8, result));
}


/***************************************************************/
/* Function:  dmCoordGetTransform_*                            */
/*                                                             */
/* Description:                                                */
/*   Get the transformation parameters for the input wcs       */
/*   coordinate descriptor                                     */
/*   Last letter of function name indicates data type.         */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *dd     - Pointer to descriptor            */
/*    <data type>   *crpix  - values                           */
/*    double        *crval  - value                            */
/*    double        *cdelt  -                                  */
/*    long           dim    - dim                              */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmErrCode  status (dmSUCCES or dmFAILURE)                */
/*                                                             */
/***************************************************************/
int  dmCoordGetTransform_s(dmDescriptor* dd, short* crpix, double* crval,
				    double* cdelt, long dim)
{
  return(dmi_coord_get_transform_cast(dd,(void*)crpix,DT_INT2,crval,cdelt,dim));
}

int  dmCoordGetTransform_l(dmDescriptor* dd, long* crpix, double* crval,
				    double* cdelt, long dim)
{
  return(dmi_coord_get_transform_cast(dd,(void*)crpix, DT_INT4,crval,cdelt,dim));
}

int  dmCoordGetTransform_f(dmDescriptor* dd, float* crpix, double* crval,
				    double* cdelt, long dim)
{
  return(dmi_coord_get_transform_cast(dd,(void*)crpix, DT_REAL4,crval,cdelt,dim));
}


int  dmCoordGetTransform_ub(dmDescriptor* dd, unsigned char* crpix, 
				     double* crval, double* cdelt, long dim)
{
  return(dmi_coord_get_transform_cast(dd, (void*)crpix, DT_UINT1,crval,cdelt,dim));
}

int  dmCoordGetTransform_us(dmDescriptor* dd, unsigned short* crpix, 
				     double* crval, double* cdelt, long dim)
{
  return(dmi_coord_get_transform_cast(dd,(void*)crpix, DT_UINT2,crval,cdelt,dim));
}

int  dmCoordGetTransform_ul(dmDescriptor* dd, unsigned long* crpix, 
				     double* crval, double* cdelt, long dim)
{
  return(dmi_coord_get_transform_cast(dd,(void*)crpix, DT_UINT4, crval,cdelt,dim));
}


/***************************************************************/
/* Function:  dmCoordInvert_*                                  */
/*                                                             */
/* Description:                                                */
/*   Return  pixel value corresponding to an input coordinate. */
/*   The coordinate value is always type double. The pixel     */
/*     output type is up to the user.                          */
/*   Last letter of function name indicates data type.         */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *dd    - Pointer to descriptor             */
/*    double        *value - pointer to world coordinate data  */
/*                                                             */
/* Output parameters:                                          */
/*    <data type>   *result  - pixel value of coord data       */
/*                                                             */
/* Returns:                                                    */
/*    int  status    (dmSUCCES or dmFAILURE)                   */
/*                                                             */
/***************************************************************/
int  dmCoordInvert_s(dmDescriptor* dd, double* value, short* result)
{
  return(dmi_coord_invert(dd, value, DT_REAL8, DT_INT2, (void*)result));
}


int  dmCoordInvert_l(dmDescriptor* dd, double* value, long* result)
{
  return(dmi_coord_invert(dd, value, DT_REAL8, DT_INT4, (void*)result));
}


int  dmCoordInvert_f(dmDescriptor* dd, double* value, float* result)
{
  return(dmi_coord_invert(dd, value, DT_REAL8, DT_REAL4, (void*)result));
}


int  dmCoordInvert_d(dmDescriptor* dd, double* value, double* result)
{
  return(dmi_coord_invert(dd, value, DT_REAL8, DT_REAL8, (void*)result));
}


int  dmCoordInvert_ub(dmDescriptor* dd, double* value, 
				     unsigned char* result)
{
  return(dmi_coord_invert(dd, value, DT_REAL8, DT_UINT1, (void*)result));
}


int  dmCoordInvert_us(dmDescriptor* dd, double* value, 
				     unsigned short* result)
{
  return(dmi_coord_invert(dd, value, DT_REAL8, DT_UINT2, (void*)result));
}


int  dmCoordInvert_ul(dmDescriptor* dd, double* value, 
				     unsigned long* result)
{
  return(dmi_coord_invert(dd, value, DT_REAL8, DT_UINT4, (void*)result));
}


/***************************************************************/
/* Function:  dmCoordSetTransform_*                            */
/*                                                             */
/* Description:                                                */
/*   Set the transformation values for the input coordinate    */
/*   descriptor                                                */
/*   Last letter of function name indicates data type.         */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *dd     - Pointer to descriptor            */
/*    <data type>   *crpix  - values                           */
/*    double        *crval  - value                            */
/*    double        *cdelt  -                                  */
/*    long           dim    - dim                              */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmErrCode  status (dmSUCCES or dmFAILURE)                */
/*                                                             */
/***************************************************************/
int  dmCoordSetTransform_s(dmDescriptor* dd, short* crpix, double* crval,
				    double* cdelt, long dim)
{
  return(dmi_coord_set_transform(dd, (void*)crpix, DT_INT2, crval, cdelt, dim));
}

int  dmCoordSetTransform_l(dmDescriptor* dd, long* crpix, double* crval,
				    double* cdelt, long dim)
{
  return(dmi_coord_set_transform(dd, (void*)crpix, DT_INT4, crval, cdelt, dim));
}
int  dmCoordSetTransform_f(dmDescriptor* dd, float* crpix, double* crval,
				    double* cdelt, long dim)
{
  return(dmi_coord_set_transform(dd, (void*)crpix, DT_REAL4, crval, cdelt, dim));
}

int  dmCoordSetTransform_d(dmDescriptor* dd, double* crpix, double* crval,
				    double* cdelt, long dim)
{
  return(dmi_coord_set_transform(dd, (void*)crpix, DT_REAL8, crval, cdelt, dim));
}

int  dmCoordSetTransform_ub(dmDescriptor* dd, unsigned char* crpix, 
				     double* crval, double* cdelt, long dim)
{
  return(dmi_coord_set_transform(dd, (void*)crpix, DT_UINT1, crval, cdelt, dim));
}

int  dmCoordSetTransform_us(dmDescriptor* dd, unsigned short* crpix, 
				     double* crval, double* cdelt, long dim)
{
  return(dmi_coord_set_transform(dd, (void*)crpix, DT_UINT2, crval, cdelt, dim));
}

int  dmCoordSetTransform_ul(dmDescriptor* dd, unsigned long* crpix, 
				     double* crval, double* cdelt, long dim)
{
  return(dmi_coord_set_transform(dd, (void*)crpix, DT_UINT4, crval, cdelt, dim));
}


/***************************************************************/
/* Function:  dmDescriptorGetBin_*                             */
/*                                                             */
/* Description:                                                */
/*   These functions returns bin value for descriptor          */
/*   Last letter of function name indicates data type          */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *pDesc  - Pointer to descriptor            */
/*                                                             */
/* Output parameters:                                          */
/*    <data type >  *value  - bin value                        */
/*                                                             */
/* Returns:                                                    */
/*    int  status  (dmSUCCESS or dmFAILURE)                    */
/*                                                             */
/***************************************************************/
int
dmDescriptorGetBin_s( const dmDescriptor* pDesc, short* value )
{
  return dmi_descriptor_get_cast_bin( pDesc, DT_INT2, (void*)value );
}

int
dmDescriptorGetBin_l( const dmDescriptor* pDesc, long* value )
{
  return dmi_descriptor_get_cast_bin( pDesc, DT_INT4, (void*)value );
}

int
dmDescriptorGetBin_f( const dmDescriptor* pDesc, float* value )
{
  return dmi_descriptor_get_cast_bin( pDesc, DT_REAL4, (void*)value );
}

int
dmDescriptorGetBin_d( const dmDescriptor* pDesc, double* value )
{
  return dmi_descriptor_get_cast_bin( pDesc, DT_REAL8, (void*)value );
}

int
dmDescriptorGetBin_ub( const dmDescriptor* pDesc, unsigned char* value )
{
  return dmi_descriptor_get_cast_bin( pDesc, DT_UINT1, (void*)value );
}

int
dmDescriptorGetBin_us( const dmDescriptor* pDesc, unsigned short* value )
{
  return dmi_descriptor_get_cast_bin( pDesc, DT_UINT2, (void*)value );
}

int
dmDescriptorGetBin_ul( const dmDescriptor* pDesc, unsigned long* value )
{
  return dmi_descriptor_get_cast_bin( pDesc, DT_UINT4, (void*)value );
}


/***************************************************************/
/* Function:  dmDescriptorGetNo                                */
/*                                                             */
/* Description:                                                */
/*   Returns descriptor number                                 */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor *dd    - Pointer to descriptor              */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    long  no                                                 */
/*                                                             */
/***************************************************************/
long dmDescriptorGetNo( const dmDescriptor* dd )
{
 if ( dd )
   return dd->no;
 else
   return 0;
}  


/***************************************************************/
/* Function:  dmDescriptorGetNull_*                            */
/*                                                             */
/* Description:                                                */
/*   These functions returns null value for descriptor         */
/*   Last letter of function name indicates data type          */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *pDesc  - Pointer to descriptor            */
/*                                                             */
/* Output parameters:                                          */
/*    <data type >  *value  - null value                       */
/*                                                             */
/* Returns:                                                    */
/*    int  status  (dmSUCCESS or dmFAILURE)                    */
/*                                                             */
/***************************************************************/
int
dmDescriptorGetNull_c( const dmDescriptor* pDesc, char** value )
{
  *value = (char*)0;
  return dmi_descriptor_get_cast_null( pDesc, DT_CHAR, (void*)value );
}

int
dmDescriptorGetNull_s( const dmDescriptor* pDesc, short* value )
{
  *value = 0;
  return dmi_descriptor_get_cast_null( pDesc, DT_INT2, (void*)value );
}

int
dmDescriptorGetNull_l( const dmDescriptor* pDesc, long* value )
{
  *value = 0;
  return dmi_descriptor_get_cast_null( pDesc, DT_INT4, (void*)value );
}

int
dmDescriptorGetNull_f( const dmDescriptor* pDesc, float* value )
{
  *value = 0.0;
  return dmi_descriptor_get_cast_null( pDesc, DT_REAL4, (void*)value );
}

int
dmDescriptorGetNull_d( const dmDescriptor* pDesc, double* value )
{
  *value = 0.0;
   return dmi_descriptor_get_cast_null( pDesc, DT_REAL8, (void*)value );
}

int
dmDescriptorGetNull_ub( const dmDescriptor* pDesc, unsigned char* value )
{
  *value = (unsigned char)0;
  return dmi_descriptor_get_cast_null( pDesc, DT_UINT1, (void*)value );
}

int
dmDescriptorGetNull_us( const dmDescriptor* pDesc, unsigned short* value )
{
  *value = 0;
  return dmi_descriptor_get_cast_null( pDesc, DT_UINT2, (void*)value );
}

int
dmDescriptorGetNull_ul( const dmDescriptor* pDesc, unsigned long* value )
{
  *value = 0;
  return dmi_descriptor_get_cast_null( pDesc, DT_UINT4, (void*)value );
}


/***************************************************************/
/* Function:  dmDecriptorGetRange_*                            */
/*                                                             */
/* Description:                                                */
/*   Return descriptor range values                            */
/*   Last letter of function name indicates data type          */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor *pDesc   - Pointer to descriptor            */
/*                                                             */
/* Output parameters:                                          */
/*    <data type >  *tlmin  - rante min value                  */
/*    <data type >  *tlmax  - range max value                  */
/*                                                             */
/* Returns:                                                    */
/*    int  status  (dmSUCCESS or dmFAILURE)                    */
/*                                                             */
/***************************************************************/
int
dmDescriptorGetRange_s(dmDescriptor* pDesc, short* tlmin, short* tlmax)
{
  return dmi_descriptor_get_cpt_range( pDesc, 1L, DT_INT2, (void*)tlmin, (void*)tlmax );
}



int
dmDescriptorGetRange_us(dmDescriptor* pDesc, unsigned short* tlmin, 
                        unsigned short* tlmax)
{
  return dmi_descriptor_get_cpt_range( pDesc, 1L, DT_UINT2, (void*)tlmin, (void*)tlmax );
}

int
dmDescriptorGetRange_ub(dmDescriptor* pDesc, unsigned char* tlmin, 
                        unsigned char* tlmax)
{
  return dmi_descriptor_get_cpt_range( pDesc, 1L, DT_UINT1, (void*)tlmin, (void*)tlmax );
}


int
dmDescriptorGetRange_bit(dmDescriptor* pDesc, unsigned char* tlmin, 
                         unsigned char* tlmax)
{
  return dmi_descriptor_get_cpt_range( pDesc, 1L, DT_BIT, (void*)tlmin, (void*)tlmax );
}



int
dmDescriptorGetRange_l(dmDescriptor* pDesc, long* tlmin, long* tlmax)
{
  return dmi_descriptor_get_cpt_range( pDesc, 1L, DT_INT4, (void*)tlmin, (void*)tlmax );
}



int
dmDescriptorGetRange_ul(dmDescriptor* pDesc, unsigned long* tlmin, 
                        unsigned long* tlmax)
{
  return dmi_descriptor_get_cpt_range( pDesc, 1L, DT_UINT4, (void*)tlmin, (void*)tlmax );
}


int
dmDescriptorGetRange_f(dmDescriptor* pDesc, float* tlmin, float* tlmax)
{
  return dmi_descriptor_get_cpt_range( pDesc, 1L, DT_REAL4, (void*)tlmin, (void*)tlmax );
}


int
dmDescriptorGetRange_d(dmDescriptor* pDesc, double* tlmin, double* tlmax)
{
 return dmDescriptorGetRangeCpt_d( pDesc, 1L, tlmin, tlmax );
}


/***************************************************************/
/* Function:  dmDescriptorSetBin_*                             */
/*                                                             */
/* Description:                                                */
/*   These functions set bin value for descriptor              */
/*   Last letter of function name indicates data type          */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *pDesc  - Pointer to descriptor            */
/*    <data type >   value  - bin value                        */
/*                                                             */
/* Output parameters:                                          */
/*  none                                                       */
/*                                                             */
/* Returns:                                                    */
/*    int  status  (dmSUCCESS or dmFAILURE)                    */
/*                                                             */
/***************************************************************/
int
dmDescriptorSetBin_s( dmDescriptor* pDesc, short value )
{
  return dmp_descriptor_set_bin( pDesc, DT_INT2, (void*)&value );
}

int
dmDescriptorSetBin_l( dmDescriptor* pDesc, long value )
{
  return dmp_descriptor_set_bin( pDesc, DT_INT4, (void*)&value );
}

int
dmDescriptorSetBin_f( dmDescriptor* pDesc, float value )
{
  return dmp_descriptor_set_bin( pDesc, DT_REAL4, (void*)&value );
}

int
dmDescriptorSetBin_d( dmDescriptor* pDesc, double value )
{
  return dmp_descriptor_set_bin( pDesc, DT_REAL8, (void*)&value );
}

int
dmDescriptorSetBin_ub( dmDescriptor* pDesc, unsigned char value )
{
  return dmp_descriptor_set_bin( pDesc, DT_UINT1, (void*)&value );
}

int
dmDescriptorSetBin_us( dmDescriptor* pDesc, unsigned short value )
{
  return dmp_descriptor_set_bin( pDesc, DT_UINT2, (void*)&value );
}

int
dmDescriptorSetBin_ul( dmDescriptor* pDesc, unsigned long value )
{
  return dmp_descriptor_set_bin( pDesc, DT_UINT4, (void*)&value );
}


/***************************************************************/
/* Function:  dmDescriptorSetNull_*                            */
/*                                                             */
/* Description:                                                */
/*   These functions set null value for descriptor             */
/*   Last letter of function name indicates data type          */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *pDesc  - Pointer to descriptor            */
/*    <data type >  *value  - null value                       */
/*                                                             */
/* Output parameters:                                          */
/*  none                                                       */
/*                                                             */
/* Returns:                                                    */
/*    int  status  (dmSUCCESS or dmFAILURE)                    */
/*                                                             */
/***************************************************************/
int
dmDescriptorSetNull_c( dmDescriptor* pDesc, char* value )
{
 return dmp_descriptor_set_null( pDesc, DT_CHAR, (void*)&value );
}

int
dmDescriptorSetNull_s( dmDescriptor* pDesc, short value )
{
 return dmp_descriptor_set_null( pDesc, DT_INT2, (void*)&value );
}

int
dmDescriptorSetNull_l( dmDescriptor* pDesc, long value )
{
 return dmp_descriptor_set_null( pDesc, DT_INT4, (void*)&value );
}

int
dmDescriptorSetNull_f( dmDescriptor* pDesc, float value )
{
 return dmp_descriptor_set_null( pDesc, DT_REAL4, (void*)&value );
}

int
dmDescriptorSetNull_d( dmDescriptor* pDesc, double value )
{
 return dmp_descriptor_set_null( pDesc, DT_REAL8, (void*)&value );
}

int
dmDescriptorSetNull_ub( dmDescriptor* pDesc, unsigned char value )
{
 return dmp_descriptor_set_null( pDesc, DT_UINT1, (void*)&value );
}

int
dmDescriptorSetNull_us( dmDescriptor* pDesc, unsigned short value )
{
 return dmp_descriptor_set_null( pDesc, DT_UINT2, (void*)&value );
}

int
dmDescriptorSetNull_ul( dmDescriptor* pDesc, unsigned long value )
{
 return dmp_descriptor_set_null( pDesc, DT_UINT4, (void*)&value );
}

		   
/***************************************************************/
/* Function:  dmDescriptorSetRange_*                           */
/*                                                             */
/* Description:                                                */
/*   These functions set the allowable ranges for a particular */
/*   particular descriptor ( method specific to descriptor     */
/*   type). Currently only columns are supported.              */
/*   Last letter of function name indicates data type          */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *pDesc  - Pointer to descriptor            */
/*    <data type >  *tlmin  - range min value                  */
/*    <data type >  *tlmax  - range max value                  */
/*                                                             */
/* Output parameters:                                          */
/*  none                                                       */
/*                                                             */
/* Returns:                                                    */
/*    int  status  (dmSUCCESS or dmFAILURE)                    */
/*                                                             */
/***************************************************************/
int
dmDescriptorSetRange_s( dmDescriptor* pDesc, short tlmin, short tlmax)
{
  return dmp_descriptor_set_range( pDesc, DT_INT2, (void*)&tlmin, (void*)&tlmax );
}


int
dmDescriptorSetRange_us(dmDescriptor* pDesc, unsigned short tlmin, 
                        unsigned short tlmax)
{
  return dmp_descriptor_set_range( pDesc, DT_UINT2, (void*)&tlmin, (void*)&tlmax );
}

int
dmDescriptorSetRange_ub( dmDescriptor* pDesc, unsigned char tlmin,
                        unsigned char tlmax)
{
  return dmp_descriptor_set_range( pDesc, DT_UINT1, (void*)&tlmin, (void*)&tlmax );
}


int
dmDescriptorSetRange_bit( dmDescriptor* pDesc, unsigned char tlmin,
                         unsigned char tlmax)
{
  return dmp_descriptor_set_range( pDesc, DT_BIT, (void*)&tlmin, (void*)&tlmax );
}



int
dmDescriptorSetRange_l(dmDescriptor* pDesc, long tlmin, long tlmax)
{
  return dmp_descriptor_set_range( pDesc, DT_INT4, (void*)&tlmin, (void*)&tlmax );
}


int
dmDescriptorSetRange_ul(dmDescriptor* pDesc, unsigned long tlmin, 
                        unsigned long tlmax)
{
  return dmp_descriptor_set_range( pDesc, DT_UINT4, (void*)&tlmin, (void*)&tlmax );
}


int
dmDescriptorSetRange_f( dmDescriptor* pDesc, float tlmin, float tlmax)
{

   return dmp_descriptor_set_range( pDesc, DT_REAL4, (void*)&tlmin, (void*)&tlmax );
}

int
dmDescriptorSetRange_d(dmDescriptor* pDesc, double tlmin, double tlmax)
{
  return dmp_descriptor_set_range( pDesc, DT_REAL8, (void*)&tlmin, (void*)&tlmax );
}


/***************************************************************/
/* Function:  dmGetArray_*                                     */
/*                                                             */
/* Description:                                                */
/*   Returns array data from buffer                            */
/*   Last letter of function name indicates data type.         */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *dd    - Pointer to descriptor             */
/*    long           npixels - number of items to retrieve     */
/*                                                             */
/* Output parameters:                                          */
/*   <data type> *val                                          */
/*                                                             */
/* Returns:                                                    */
/*    dmErrCode  status (dmSUCCESS or dmFAILURE)               */
/*                                                             */
/***************************************************************/
dmErrCode   dmGetArray_s(dmDescriptor* dd, short* val,	long npixels)
{
  return dmGetVectorArray( dd, DT_INT2, val, 1L, 1L, npixels );
}

dmErrCode   dmGetArray_l(dmDescriptor* dd, long* val,	long npixels)
{
  return dmGetVectorArray( dd, DT_INT4, val, 1L, 1L, npixels );
}

dmErrCode   dmGetArray_f( dmDescriptor* dd, float* val,	long npixels)
{
  return dmGetVectorArray( dd, DT_REAL4, val, 1L, 1L, npixels );
}

dmErrCode   dmGetArray_d(dmDescriptor* dd, double* val,	long npixels)
{
  return dmGetVectorArray( dd, DT_REAL8, val, 1L, 1L, npixels );
}

dmErrCode   dmGetArray_q(dmDescriptor* dd, dmBool* val,	long npixels)
{
  return dmGetVectorArray( dd, DT_LOGICAL, val, 1L, 1L, npixels );
}

dmErrCode   dmGetArray_ub(dmDescriptor* dd, unsigned char* val,	long npixels)
{
  return dmGetVectorArray( dd, DT_UINT1, val, 1L, 1L, npixels );
}

dmErrCode   dmGetArray_us(dmDescriptor* dd, unsigned short* val,	long npixels)
{
  return dmGetVectorArray( dd, DT_UINT2, val, 1L, 1L, npixels );
}

dmErrCode   dmGetArray_ul(dmDescriptor* dd, unsigned long* val,	long npixels)
{
  return dmGetVectorArray( dd, DT_UINT4, val, 1L, 1L, npixels );
}

dmErrCode   dmGetArray_c(dmDescriptor* dd, char** val,	long npixels)
{
  return dmGetVectorArray( dd, DT_CHAR, val, 1L, 1L, npixels );
}

dmErrCode   dmGetArray_bit(dmDescriptor* dd, void* val,	long npixels)
{
  return dmGetVectorArray( dd, DT_BIT, val, 1L, 1L, npixels );
}

/***************************************************************/
/* Function:  dmGetScalor_*                                    */
/*                                                             */
/* Description:                                                */
/*   Return  scalor data from buffer                           */
/*   Used to retrieve one datum.                               */
/*   Last letter of function name indicates data type.         */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *dd    - Pointer to descriptor             */
/*                                                             */
/* Output parameters:                                          */
/*   none                                                      */
/*                                                             */
/* Returns:                                                    */
/*    <data type> val                                          */
/*                                                             */
/***************************************************************/
short dmGetScalar_s(dmDescriptor* dd )
{
 short val;
 dmGetVectorArray( dd, DT_INT2, &val, 1L, 1L, 1L );
 return val;
}

long dmGetScalar_l(dmDescriptor* dd )
{
 long val;
 dmGetVectorArray( dd, DT_INT4, &val, 1L, 1L, 1L );
 return val;
}

float dmGetScalar_f(dmDescriptor* dd )
{
 float val;
 dmGetVectorArray( dd, DT_REAL4, &val, 1L, 1L, 1L );
 return val;
}

double dmGetScalar_d(dmDescriptor* dd )
{
 double val;
 dmGetVectorArray( dd, DT_REAL8, &val, 1L, 1L, 1L );
 return val;
}

unsigned short dmGetScalar_us(dmDescriptor* dd )
{
 unsigned short val;
 dmGetVectorArray( dd, DT_UINT2, &val, 1L, 1L, 1L );
 return val;
}


unsigned long dmGetScalar_ul(dmDescriptor* dd )
{
 unsigned long val;
 dmGetVectorArray( dd, DT_UINT4, &val, 1L, 1L, 1L );
 return val;
}

int  dmGetScalar_q(dmDescriptor* dd)
{
 dmBool val;
 dmGetVectorArray( dd, DT_LOGICAL, &val, 1L, 1L, 1L );
 return val;
}

unsigned char dmGetScalar_ub(dmDescriptor* dd)
{
 unsigned char val;
 dmGetVectorArray( dd, DT_UINT1, &val, 1L, 1L, 1L );
 return val;
}


/***************************************************************/
/* Function:  dmGetScalor_c                                    */
/*                                                             */
/* Description:                                                */
/*   Return char data from buffer                              */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *dd     - Pointer to descriptor            */
/*    long           maxlen - max length of char string        */
/*                                                             */
/* Output parameters:                                          */
/*   char  *val  - pointer to char buffer to hold data         */
/*                                                             */
/* Returns:                                                    */
/*    none                                                     */
/*                                                             */
/***************************************************************/
void dmGetScalar_c(dmDescriptor* dd,  char *val, long maxlen)
{
 char *valp   = val;
 int   qalloc = 0;

/* If the string is bigger than maxlen (size of the buffer provided)  */
/* (only then) use malloc. After reading the string into this buffer, */
/* we then copy up to maxlen characters into the provided string.     */

 if ( dd && dd->length > maxlen )
 {
   valp = malloc( dd->length + 1 );
   qalloc = 1;
 }

 dmGetVectorArray( dd, DT_CHAR, &valp, 1L, 1L, 1L );

 if ( qalloc )
 {
   cs_copy_siz( valp, val, maxlen );
   free( valp );
 }
}

/***************************************************************/
/* Function:  dmGetScalor_bit                                  */
/*                                                             */
/* Description:                                                */
/*   Return bit data from buffer                               */
/*   Uses call to dmGetVectorArray() to retieve data           */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *dd     - Pointer to descriptor            */
/*                                                             */
/* Output parameters:                                          */
/*   void  *val  - pointer to void buffer to hold data         */
/*                                                             */
/* Returns:                                                    */
/*    dmErrCode  errcode                                       */
/*                                                             */
/***************************************************************/
dmErrCode  dmGetScalar_bit(dmDescriptor* dd, void* val)
{
  long nvals = 1;
  int errcode;

  if ( dd )
    nvals = dmi_descriptor_get_array_size( dd );

   errcode = dmGetVectorArray( dd, DT_BIT, val, 1L, 1L, nvals );
   return errcode;
}


/***************************************************************/
/* Function:  dmGetScalors_*                                   */
/*                                                             */
/* Description:                                                */
/*   Returns rows of data from buffer                          */
/*   Last letter of function name indicates data type.         */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *dd     - Pointer to descriptor            */
/*    long           row    - starting row number              */
/*    long           nrows  - number of rows of data to return */
/*                                                             */
/* Output parameters:                                          */
/*   char  *val  - pointer to char buffer to hold data         */
/*                                                             */
/* Returns:                                                    */
/*    long status  (dmSUCCESS or dmFAILURE)                    */
/*                                                             */
/***************************************************************/
long  dmGetScalars_s(dmDescriptor* dd, short* vals, long row, long nrows)
{
  return dmGetVectorArrays( dd, DT_INT2, vals, 1L, row, nrows, 1L, 1L );
}


long  dmGetScalars_bit(dmDescriptor* dd, void* vals, long row, long nrows)
{
  long nvals = 1;

  if ( dd )
    nvals = dmi_descriptor_get_array_size( dd );

  return dmGetVectorArrays( dd, DT_BIT, vals, 1L, row, nrows, 1L, nvals );
}


long  dmGetScalars_l(dmDescriptor* dd, long* vals, long row, long nrows)
{
  return dmGetVectorArrays( dd, DT_INT4, vals, 1L, row, nrows, 1L, 1L );
}


long  dmGetScalars_f(dmDescriptor* dd, float* vals, long row, long nrows)
{
  return dmGetVectorArrays( dd, DT_REAL4, vals, 1L, row, nrows, 1L, 1L );
}


long  dmGetScalars_d(dmDescriptor* dd, double* vals, long row, long nrows)
{
  return dmGetVectorArrays( dd, DT_REAL8, vals, 1L, row, nrows, 1L, 1L );
}

long  dmGetScalars_c( dmDescriptor* dd, char** vals, long maxlen, long row, long nrows)
{
  (void) maxlen;
  return dmGetVectorArrays( dd, DT_CHAR, vals, 1L, row, nrows, 1L, 1L );
}

long  dmGetScalars_ub(dmDescriptor* dd, unsigned char* vals, long row, long nrows)
{
  return dmGetVectorArrays( dd, DT_UINT1, vals, 1L, row, nrows, 1L, 1L );
}

long  dmGetScalars_us(dmDescriptor* dd, unsigned short* vals, long row, long nrows)
{
  return dmGetVectorArrays( dd, DT_UINT2, vals, 1L, row, nrows, 1L, 1L );
}

long  dmGetScalars_ul( dmDescriptor* dd, unsigned long* vals, long row, long nrows)
{
  return dmGetVectorArrays( dd, DT_UINT4, vals, 1L, row, nrows, 1L, 1L );
}

long  dmGetScalars_q( dmDescriptor* dd, dmBool* vals, long row, long nrows )
{
  return dmGetVectorArrays( dd, DT_LOGICAL, vals, 1L, row, nrows, 1L, 1L );
}


/***************************************************************/
/* Function:  dmGetVector_*                                    */
/*                                                             */
/* Description:                                                */
/*   Return  vector data from buffer                           */
/*   Used to retrieve one datum.                               */
/*   Last letter of function name indicates data type.         */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *dd    - Pointer to descriptor             */
/*    long           dim   - vector dim                        */
/*                                                             */
/* Output parameters:                                          */
/*    <data type>   *vals    - buffer to store retrieved data  */
/*                                                             */
/* Returns:                                                    */
/*    dmErrCode  status (dmSUCCES or dmFAILURE)                */
/*                                                             */
/***************************************************************/
dmErrCode  dmGetVector_q(dmDescriptor* dd, dmBool* vals,  long dim)
{
  return dmGetVectorArray( dd, DT_LOGICAL, vals, dim, 1L, 1L );
}

dmErrCode  dmGetVector_c(dmDescriptor* dd, char** vals,  long dim)
{
  return dmGetVectorArray( dd, DT_CHAR, vals, dim, 1L, 1L );
}

dmErrCode  dmGetVector_ub(dmDescriptor* dd, unsigned char* vals,  long dim)
{
  return dmGetVectorArray( dd, DT_UINT1, vals, dim, 1L, 1L );
}

dmErrCode  dmGetVector_us(dmDescriptor* dd, unsigned short* vals,  long dim )
{
  return dmGetVectorArray( dd, DT_UINT2, vals, dim, 1L, 1L );
}

dmErrCode  dmGetVector_ul(dmDescriptor* dd, unsigned long* vals,  long dim )
{
  return dmGetVectorArray( dd, DT_UINT4, vals, dim, 1L, 1L );
}

dmErrCode  dmGetVector_s(dmDescriptor* dd, short* vals,  long dim )
{
  return dmGetVectorArray( dd, DT_INT2, vals, dim, 1L, 1L );
}

dmErrCode  dmGetVector_l(dmDescriptor* dd, long* vals,  long dim)
{
  return dmGetVectorArray( dd, DT_INT4, vals, dim, 1L, 1L );
}

dmErrCode  dmGetVector_f(dmDescriptor* dd, float* vals,  long dim )
{
  return dmGetVectorArray( dd, DT_REAL4, vals, dim, 1L, 1L );
}

dmErrCode  dmGetVector_d( dmDescriptor* dd, double* vals, long dim )
{ 
  return dmGetVectorArray( dd, DT_REAL8, vals, dim, 1L, 1L );
}


/***************************************************************/
/* Function:  dmGetVectors_*                                   */
/*                                                             */
/* Description:                                                */
/*   Get vector data from buffer                               */
/*   Last letter of function name indicates data type.         */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *pDesc  - Pointer to descriptor            */
/*    long           dim    - vector dim                       */
/*    long           row    - starting row                     */
/*    long           nrows  - number of rows                   */
/*                                                             */
/* Output parameters:                                          */
/*    <data type>   *vals   - values                           */
/*                                                             */
/* Returns:                                                    */
/*    dmErrCode  status (dmSUCCES or dmFAILURE)                */
/*                                                             */
/***************************************************************/
dmErrCode  dmGetVectors_s(dmDescriptor* pDesc, short* vals, long dim, long row, long nrows)
{
  return dmGetVectorArrays( pDesc, DT_INT2, vals, dim, row, nrows, 1L, 1L  );
}

dmErrCode  dmGetVectors_l(dmDescriptor* pDesc, long* vals, long dim, long row, long nrows)
{
  return dmGetVectorArrays( pDesc, DT_INT4, vals, dim, row, nrows, 1L, 1L  );
}

dmErrCode  dmGetVectors_f(dmDescriptor* pDesc, float* vals, long dim, long row, long nrows)
{
  return dmGetVectorArrays( pDesc, DT_REAL4, vals, dim, row, nrows, 1L, 1L  );
}

dmErrCode  dmGetVectors_d(dmDescriptor* pDesc, double* vals, long dim, long row, long nrows)
{
  return dmGetVectorArrays( pDesc, DT_REAL8, vals, dim, row, nrows, 1L, 1L  );
}

dmErrCode  dmGetVectors_q(dmDescriptor* pDesc, dmBool* vals, long dim, long row, long nrows)
{
  return dmGetVectorArrays( pDesc, DT_LOGICAL, vals, dim, row, nrows, 1L, 1L  );
}

dmErrCode  dmGetVectors_c(dmDescriptor* pDesc, char** vals, long dim, long row, long nrows)
{
  return dmGetVectorArrays( pDesc, DT_CHAR, vals, dim, row, nrows, 1L, 1L  );
}

dmErrCode  dmGetVectors_ub(dmDescriptor* pDesc, unsigned char* vals, long dim, long row, long nrows)
{
  return dmGetVectorArrays( pDesc, DT_UINT1, vals, dim, row, nrows, 1L, 1L  );
}

dmErrCode  dmGetVectors_us(dmDescriptor* pDesc, unsigned short* vals, long dim, long row, long nrows)
{
  return dmGetVectorArrays( pDesc, DT_UINT2, vals, dim, row, nrows, 1L, 1L  );
}
dmErrCode  dmGetVectors_ul(dmDescriptor* pDesc, unsigned long* vals, long dim, long row, long nrows)
{
  return dmGetVectorArrays( pDesc, DT_UINT4, vals, dim, row, nrows, 1L, 1L  );
}


/***************************************************************/
/* Function:  dmGetVectorArray_*                               */
/*                                                             */
/* Description:                                                */
/*   Return  vector data from buffer                           */
/*   Used to retrieve one or more data items.                  */
/*   Last letter of function name indicates data type.         */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *dd    - Pointer to descriptor             */
/*    long           dim   - vector dim                        */
/*    long           nvals - number of data items to retrieve  */
/*                                                             */
/* Output parameters:                                          */
/*    <data type>   *vals    - buffer to store retrieved data  */
/*                                                             */
/* Returns:                                                    */
/*    dmErrCode  status (dmSUCCES or dmFAILURE)                */
/*                                                             */
/***************************************************************/
dmErrCode  dmGetVectorArray_q(dmDescriptor* dd, dmBool* vals, long dim, long nvals)
{
  return dmGetVectorArray( dd, DT_LOGICAL, vals, dim, 1L, nvals );
}

dmErrCode  dmGetVectorArray_c(dmDescriptor* dd, char** vals, long dim, long nvals)
{
  return dmGetVectorArray( dd, DT_CHAR, vals, dim, 1L, nvals );
}

dmErrCode  dmGetVectorArray_ub(dmDescriptor* dd, unsigned char* vals, long dim, long nvals)
{
  return dmGetVectorArray( dd, DT_UINT1, vals, dim, 1L, nvals );
}

dmErrCode  dmGetVectorArray_ul(dmDescriptor* dd, unsigned long* vals, long dim, long nvals)
{
  return dmGetVectorArray( dd, DT_UINT4, vals, dim, 1L, nvals );
}

dmErrCode  dmGetVectorArray_us(dmDescriptor* dd, unsigned short* vals, long dim, long nvals)
{
  return dmGetVectorArray( dd, DT_UINT2, vals, dim, 1L, nvals );
}

dmErrCode  dmGetVectorArray_s(dmDescriptor* dd, short* vals, long dim, long nvals)
{
  return dmGetVectorArray( dd, DT_INT2, vals, dim, 1L, nvals );
}

dmErrCode  dmGetVectorArray_l(dmDescriptor* dd, long* vals, long dim, long nvals)
{
  return dmGetVectorArray( dd, DT_INT4, vals, dim, 1L, nvals );
}

dmErrCode  dmGetVectorArray_f(dmDescriptor* dd, float* vals, long dim, long nvals)
{
  return dmGetVectorArray( dd, DT_REAL4, vals, dim, 1L, nvals );
}

dmErrCode  dmGetVectorArray_d(dmDescriptor* dd, double* vals, long dim, long nvals)
{
  return dmGetVectorArray( dd, DT_REAL8, vals, dim, 1L, nvals );
}


/***************************************************************/
/* Function:  dmGetVectorArrays_*                              */
/*                                                             */
/* Description:                                                */
/*   Get vector array data from buffer                         */
/*   Last letter of function name indicates data type.         */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *pDesc  - Pointer to descriptor            */
/*    long           dim    - vector dim                       */
/*    long           row    - starting row                     */
/*    long           nrows  - number of rows                   */
/*                                                             */
/* Output parameters:                                          */
/*    <data type>  * invals - values                           */
/*                                                             */
/* Returns:                                                    */
/*    dmErrCode  status (dmSUCCES or dmFAILURE)                */
/*                                                             */
/***************************************************************/
dmErrCode    dmGetVectorArrays_s(dmDescriptor* pDesc, short* invals, long dim, long nvals, long row, long nrows)
{
  return dmGetVectorArrays( pDesc, DT_INT2, invals, dim, row, nrows, 1, nvals  );
}

dmErrCode    dmGetVectorArrays_l(dmDescriptor* pDesc, long* invals, long dim, long nvals, long row, long nrows)
{
  return dmGetVectorArrays( pDesc, DT_INT4, invals, dim, row, nrows, 1, nvals  );
}

dmErrCode    dmGetVectorArrays_f(dmDescriptor* pDesc, float* invals, long dim, long nvals, long row, long nrows)
{
  return dmGetVectorArrays( pDesc, DT_REAL4, invals, dim, row, nrows, 1, nvals  );
}

dmErrCode    dmGetVectorArrays_d(dmDescriptor* pDesc, double* invals, long dim, long nvals, long row, long nrows)
{
  return dmGetVectorArrays( pDesc, DT_REAL8, invals, dim, row, nrows, 1, nvals  );
}

dmErrCode    dmGetVectorArrays_q(dmDescriptor* pDesc, dmBool* invals, long dim, long nvals, long row, long nrows)
{
  return dmGetVectorArrays( pDesc, DT_LOGICAL, invals, dim, row, nrows, 1, nvals  );
}

dmErrCode    dmGetVectorArrays_c(dmDescriptor* pDesc, char** invals, long dim, long nvals,  long row, long nrows)
{
  return dmGetVectorArrays( pDesc, DT_CHAR, invals, dim, row, nrows, 1, nvals  );
}

dmErrCode    dmGetVectorArrays_ub(dmDescriptor* pDesc, unsigned char* invals, long dim, long nvals, 
		    long row, long nrows)
{
  return dmGetVectorArrays( pDesc, DT_UINT1, invals, dim, row, nrows, 1, nvals  );
}
dmErrCode   dmGetVectorArrays_us(dmDescriptor* pDesc, unsigned short* invals, long dim, long nvals,  long row, 
		long nrows)
{
  return dmGetVectorArrays( pDesc, DT_UINT2, invals, dim, row, nrows, 1, nvals  );
}

dmErrCode   dmGetVectorArrays_ul(dmDescriptor* pDesc, unsigned long* invals, long dim, long nvals, long row, 
		long nrows)
{
  return dmGetVectorArrays( pDesc, DT_UINT4, invals, dim, row, nrows, 1, nvals  );
}


/***************************************************************/
/* Function:  dmImageGetDataDescriptor                         */
/*                                                             */
/* Description:                                                */
/*   Returns first descriptor in cols list                     */
/*                                                             */
/* Input parameters:                                           */
/*    dmBlock *block      - Pointer to block                   */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmDescriptor  *dd                                        */
/*                                                             */
/***************************************************************/
dmDescriptor* dmImageGetDataDescriptor(dmBlock* block )
{
 dmDescriptor* dd = NULL;

 if ( block )
   dd = block->cols[0];
 return dd;
}


/***************************************************************/
/* Function:  dmImageDataGetPixel_*                            */
/*                                                             */
/* Description:                                                */
/*   Return one image data value                               */
/*   Last letter of function name indicates data type          */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor *imdata  - Pointer to descriptor            */
/*    long         *pixelNo - pixel number                     */
/*    long          dim     - dim                              */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    <data type> value                                        */
/*                                                             */
/***************************************************************/
short
dmImageDataGetPixel_s(dmDescriptor *imdata, long *pixelNo, long dim)
{
  short value[1];
  long cdim;
  
  cdim=dmGetArrayDim(imdata);
  if (cdim!=dim) 
  {
     dmu_err_set(dmBADDIM,"DIM DOESN'T MATCH PIXELNO [dmImageDataGetPixel_s]");
     return 0;
  }

  if (dmImageDataGetSubArray_s(imdata,pixelNo,pixelNo,value)==dmSUCCESS)
     return value[0];
  else
     return 0;
}

long 
dmImageDataGetPixel_l(dmDescriptor *imdata, long *pixelNo, long dim)
{
  long value[1];
  long cdim;
  
  cdim=dmGetArrayDim(imdata);
  if (cdim!=dim) 
  {
     dmu_err_set(dmBADDIM,"DIM DOESN'T MATCH PIXELNO [dmImageDataGetPixel_s]");
     return 0;
  }


  if (dmImageDataGetSubArray_l(imdata,pixelNo,pixelNo,value)==dmSUCCESS) 
     return value[0];
  else
     return 0;
}

float 
dmImageDataGetPixel_f(dmDescriptor *imdata, long *pixelNo, long dim)
{
  float value[1];
  long cdim;
  
  cdim=dmGetArrayDim(imdata);
  if (cdim!=dim) 
  {
     dmu_err_set(dmBADDIM,"DIM DOESN'T MATCH PIXELNO [dmImageDataGetPixel_s]");
     return 0;
  }


  if (dmImageDataGetSubArray_f(imdata,pixelNo,pixelNo,value) ==dmSUCCESS) 
     return value[0];
  else
     return 0.;
}

double 
dmImageDataGetPixel_d(dmDescriptor *imdata, long *pixelNo, long dim)
{
  double value[1];
  long cdim;
  
  cdim=dmGetArrayDim(imdata);
  if (cdim!=dim) 
  {
     dmu_err_set(dmBADDIM,"DIM DOESN'T MATCH PIXELNO [dmImageDataGetPixel_s]");
     return 0;
  }


  if (dmImageDataGetSubArray_d(imdata,pixelNo,pixelNo,value) ==dmSUCCESS) 
     return value[0];
  else
     return 0.;
}

unsigned char 
dmImageDataGetPixel_ub(dmDescriptor *imdata, long *pixelNo, long dim)
{
  unsigned char value[1];
  long cdim;
  
  cdim=dmGetArrayDim(imdata);
  if (cdim!=dim) 
  {
     dmu_err_set(dmBADDIM,"DIM DOESN'T MATCH PIXELNO [dmImageDataGetPixel_s]");
     return 0;
  }


  if (dmImageDataGetSubArray_ub(imdata,pixelNo,pixelNo,value) ==dmSUCCESS) 
     return value[0];
  else
     return 0x0;
}

unsigned short
dmImageDataGetPixel_us(dmDescriptor *imdata, long *pixelNo, long dim)
{
  unsigned short value[1];
  long cdim;
  
  cdim=dmGetArrayDim(imdata);
  if (cdim!=dim) 
  {
     dmu_err_set(dmBADDIM,"DIM DOESN'T MATCH PIXELNO [dmImageDataGetPixel_s]");
     return 0;
  }


  if (dmImageDataGetSubArray_us(imdata,pixelNo,pixelNo,value) ==dmSUCCESS) 
     return value[0];
  else
     return 0;
}

unsigned long 
dmImageDataGetPixel_ul(dmDescriptor *imdata, long *pixelNo, long dim)
{
  unsigned long value[1];
  long cdim;
  
  cdim=dmGetArrayDim(imdata);
  if (cdim!=dim) 
  {
     dmu_err_set(dmBADDIM,"DIM DOESN'T MATCH PIXELNO [dmImageDataGetPixel_s]");
     return 0;
  }


  if (dmImageDataGetSubArray_ul(imdata,pixelNo,pixelNo,value) ==dmSUCCESS) 
     return value[0];
  else
     return 0;
}


/***************************************************************/
/* Function:  dmImageDataGetSubArray_*                         */
/*                                                             */
/* Description:                                                */
/*   Get image data                                            */
/*   Last letter of function name indicates data type          */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor *imdata  - Pointer to descriptor            */
/*    long         *pixelLl - pixel lower left                 */
/*    long         *pixelUr - pixel upper right                */
/*                                                             */
/* Output parameters:                                          */
/*    void   *values  - returned value buffer                  */
/*                                                             */
/* Returns:                                                    */
/*    int  status  (dmSUCCESS or dmFAILURE)                    */
/*                                                             */
/***************************************************************/
int  dmImageDataGetSubArray_s(dmDescriptor* imdata, long* pixelLl,
			     long* pixelUr, short* values)
{
  return dmp_image_get_subarray( imdata, pixelLl, pixelUr, DT_INT2, (void*)values );
}

int  dmImageDataGetSubArray_l(dmDescriptor* imdata, long* pixelLl,
			     long* pixelUr, long* values)
{
  return dmp_image_get_subarray( imdata, pixelLl, pixelUr, DT_INT4, (void*)values );
}


int  dmImageDataGetSubArray_f(dmDescriptor* imdata, long* pixelLl,
			     long* pixelUr, float* values)
{
  return dmp_image_get_subarray( imdata, pixelLl, pixelUr, DT_REAL4, (void*)values );
}


int  dmImageDataGetSubArray_d(dmDescriptor* imdata, long* pixelLl,
			     long* pixelUr, double* values)
{
  return dmp_image_get_subarray( imdata, pixelLl, pixelUr, DT_REAL8, (void*)values );
}

int  dmImageDataGetSubArray_ub(dmDescriptor* imdata, long* pixelLl, 
			      long* pixelUr, unsigned char* values)
{
  return dmp_image_get_subarray( imdata, pixelLl, pixelUr, DT_UINT1, (void*)values );
}

int  dmImageDataGetSubArray_us(dmDescriptor* imdata, long* pixelLl,
			      long* pixelUr,unsigned short* values)
{
  return dmp_image_get_subarray( imdata, pixelLl, pixelUr, DT_UINT2, (void*)values );
}

int  dmImageDataGetSubArray_ul(dmDescriptor* imdata, long* pixelLl,
			      long* pixelUr,unsigned long* values)
{
  return dmp_image_get_subarray( imdata, pixelLl, pixelUr, DT_UINT4, (void*)values );
}


/***************************************************************/
/* Function:  dmImageDataSetPixel_*                            */
/*                                                             */
/* Description:                                                */
/*   Store image data value                                    */
/*   Last letter of function name indicates data type          */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor *imdata  - Pointer to descriptor            */
/*    long         *pixelNo - pixel number                     */
/*    <data type>   val     - value to store                   */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    int  status  (dmSUCCESS or dmFAILURE)                    */
/*                                                             */
/***************************************************************/
int
dmImageDataSetPixel_s(dmDescriptor *imdata, long *pixelNo, short val)
{
  short value[1];

  value[0]=val;
  return dmImageDataSetSubArray_s(imdata, pixelNo, pixelNo, value);
}

int
dmImageDataSetPixel_l(dmDescriptor *imdata, long *pixelNo, long val)
{
  long value[1];

  value[0]=val;
  return dmImageDataSetSubArray_l(imdata, pixelNo, pixelNo, value);
}

int
dmImageDataSetPixel_f(dmDescriptor *imdata, long *pixelNo, float val)
{
  float value[1];

  value[0]=val;
  return dmImageDataSetSubArray_f(imdata, pixelNo, pixelNo, value);
}

int
dmImageDataSetPixel_d(dmDescriptor *imdata, long *pixelNo, double val)
{
  double value[1];

  value[0]=val;
  return dmImageDataSetSubArray_d(imdata, pixelNo, pixelNo, value);
}

int
dmImageDataSetPixel_ub(dmDescriptor *imdata, long *pixelNo, unsigned char val)
{
  unsigned char value[1];

  value[0]=val;
  return dmImageDataSetSubArray_ub(imdata, pixelNo, pixelNo, value);
}

int
dmImageDataSetPixel_us(dmDescriptor *imdata, long *pixelNo, unsigned short val)
{
  unsigned short value[1];

  value[0]=val;
  return dmImageDataSetSubArray_us(imdata, pixelNo, pixelNo, value);
}

int
dmImageDataSetPixel_ul(dmDescriptor *imdata, long *pixelNo, unsigned long val)
{
  unsigned long value[1];

  value[0]=val;
  return dmImageDataSetSubArray_ul(imdata, pixelNo, pixelNo, value);
}


/***************************************************************/
/* Function:  dmImageDataSetSubArray_*                         */
/*                                                             */
/* Description:                                                */
/*   Set image data                                            */
/*   Last letter of function name indicates data type          */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor *imdata  - Pointer to descriptor            */
/*    long         *pixelLl - pixel lower left                 */
/*    long         *pixelUr - pixel upper right                */
/*    <datatype>   *values  - values                           */
/*                                                             */
/* Output parameters:                                          */
/*   none                                                      */
/*                                                             */
/* Returns:                                                    */
/*    int  status  (dmSUCCESS or dmFAILURE)                    */
/*                                                             */
/***************************************************************/
int  dmImageDataSetSubArray_s(dmDescriptor* imdata, long* pixelLl,
			     long* pixelUr, short* values)
{
  return dmp_image_set_subarray( imdata, pixelLl, pixelUr, DT_INT2, (void*)values );
}

int  dmImageDataSetSubArray_l(dmDescriptor* imdata, long* pixelLl,
			     long* pixelUr, long* values)
{
  return dmp_image_set_subarray( imdata, pixelLl, pixelUr, DT_INT4, (void*)values );
}

int  dmImageDataSetSubArray_f(dmDescriptor* imdata, long* pixelLl,
			     long* pixelUr, float* values)
{
  return dmp_image_set_subarray( imdata, pixelLl, pixelUr, DT_REAL4, (void*)values );
}

int  dmImageDataSetSubArray_d(dmDescriptor* imdata, long* pixelLl,
			     long* pixelUr, double* values)
{
  return dmp_image_set_subarray( imdata, pixelLl, pixelUr, DT_REAL8, (void*)values );
}

int  dmImageDataSetSubArray_ub(dmDescriptor* imdata, long* pixelLl, 
			      long* pixelUr,unsigned char* values)
{
  return dmp_image_set_subarray( imdata, pixelLl, pixelUr, DT_UINT1, (void*)values );
}

int  dmImageDataSetSubArray_us(dmDescriptor* imdata, long* pixelLl, 
			      long* pixelUr,unsigned short* values)
{
  return dmp_image_set_subarray( imdata, pixelLl, pixelUr, DT_UINT2, (void*)values );
}

int  dmImageDataSetSubArray_ul(dmDescriptor* imdata, long* pixelLl,
			      long* pixelUr,unsigned long* values)
{
  return dmp_image_set_subarray( imdata, pixelLl, pixelUr, DT_UINT4, (void*)values );
}


/***************************************************************/
/* Function:  dmKeyRead_*                                      */
/*                                                             */
/* Description:                                                */
/*   Read key value                                            */
/*   Last letter of function name indicates data type          */
/*                                                             */
/* Input parameters:                                           */
/*    dmBlock      *block   - Pointer to block                 */
/*    const char   *name    - key name                         */
/*                                                             */
/* Output parameters:                                          */
/*    <data type >  *value  - returned key value               */
/*                                                             */
/* Returns:                                                    */
/*    dmDescriptor  *key  - key descriptor                     */
/*                                                             */
/***************************************************************/
dmDescriptor*  dmKeyRead_s( dmBlock* block, const char* name, short* value )
{
  return dmp_key_read( block, name, DT_INT2, value );
} 

dmDescriptor*  dmKeyRead_l( dmBlock* block, const char* name, long* value )
{
  return dmp_key_read( block, name, DT_INT4, value );
} 

dmDescriptor*  dmKeyRead_f( dmBlock* block, const char* name, float* value )
{
  return dmp_key_read( block, name, DT_REAL4, value );
} 

dmDescriptor*  dmKeyRead_d( dmBlock* block, const char* name, double* value )
{
  return dmp_key_read( block, name, DT_REAL8, value );
} 

dmDescriptor*  dmKeyRead_q( dmBlock* block, const char* name, dmBool* value )
{
  return dmp_key_read( block, name, DT_LOGICAL, value );
} 

dmDescriptor*  dmKeyRead_ub( dmBlock* block, const char* name, unsigned char* value )
{
  return dmp_key_read( block, name, DT_UINT1, value );
} 

dmDescriptor*  dmKeyRead_us( dmBlock* block, const char* name, unsigned short* value )
{
  return dmp_key_read( block, name, DT_UINT2, value );
} 

dmDescriptor*  dmKeyRead_ul(dmBlock* block, const char* name, unsigned long* value )
{
  return dmp_key_read( block, name, DT_UINT4, value );
} 

dmDescriptor*  dmKeyRead_c( dmBlock* block, const char* name, char* value, long maxlen )
{
  dmDescriptor* dd;
  char* tmp;

  dd = dmp_key_read( block, name, DT_CHAR, &tmp );
  if ( dd )
  {
    cs_copy_siz( tmp, value, maxlen );
    free( tmp );
  }
  return dd;
} 


/***************************************************************/
/* Function:  dmKeyWrite_*                                     */
/*                                                             */
/* Description:                                                */
/*   Write key value                                           */
/*   Last letter of function name indicates data type          */
/*                                                             */
/* Input parameters:                                           */
/*    dmBlock      *block   - Pointer to block                 */
/*    const char   *name    - key name                         */
/*    <data type>  *val     - value                            */
/*    const char   *unit    - unit                             */
/*    const char   *desc    - description                      */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmDescriptor  *key  - key descriptor                     */
/*                                                             */
/***************************************************************/
dmDescriptor*  dmKeyWrite_s( dmBlock* block, const char* name, const short value, const char* unit, const char* desc )
{
  return dmp_key_write( block, name, DT_INT2, &value, 1L, unit, desc, UT_TRUE );
}

dmDescriptor*  dmKeyWrite_l( dmBlock* block, const char* name, const long value, const char* unit, const char* desc )
{
  return dmp_key_write( block, name, DT_INT4, &value, 1L, unit, desc, UT_TRUE );
}

dmDescriptor*  dmKeyWrite_f( dmBlock* block, const char* name, const float value, const char* unit, const char* desc )
{
  return dmp_key_write( block, name, DT_REAL4, &value, 1L, unit, desc, UT_TRUE );
}

dmDescriptor*  dmKeyWrite_d( dmBlock* block, const char* name, const double value, const char* unit, const char* desc )
{
  return dmp_key_write( block, name, DT_REAL8, &value, 1L, unit, desc, UT_TRUE );
}

dmDescriptor*  dmKeyWrite_us( dmBlock* block, const char* name, const unsigned short value, const char* unit, const char* desc )
{
  return dmp_key_write( block, name, DT_UINT2, &value, 1L, unit, desc, UT_TRUE );
}

dmDescriptor*  dmKeyWrite_ul( dmBlock* block, const char* name, const unsigned long value, const char* unit, const char* desc )
{
  return dmp_key_write( block, name, DT_UINT4, &value, 1L, unit, desc, UT_TRUE );
}

dmDescriptor*  dmKeyWrite_ub( dmBlock* block, const char* name, const unsigned char value, const char* unit, const char* desc )
{
  return dmp_key_write( block, name, DT_UINT1, &value, 1L, unit, desc, UT_TRUE );
}

dmDescriptor*  dmKeyWrite_q( dmBlock* block, const char* name, const dmBool value, const char* unit, const char* desc )
{
  return dmp_key_write( block, name, DT_LOGICAL, &value, 1L, unit, desc, UT_TRUE );
}

dmDescriptor*  dmKeyWrite_c( dmBlock* block, const char* name, const char* value, const char* unit, const char* desc )
{
  return dmp_key_write( block, name, DT_CHAR, &value, 1L, unit, desc, UT_TRUE );
}


/***************************************************************/
/* Function:  dmKeyWriteArray_*                                */
/*                                                             */
/* Description:                                                */
/*   Write key array values                                    */
/*   Last letter of function name indicates data type          */
/*                                                             */
/* Input parameters:                                           */
/*    dmBlock      *block   - Pointer to block                 */
/*    const char   *name    - key name                         */
/*    <data type>  *val     - value                            */
/*    const char   *unit    - unit                             */
/*    const char   *desc    - description                      */
/*    const long    nvals   - number of values in array data   */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmDescriptor  *key  - key descriptor                     */
/*                                                             */
/***************************************************************/
dmDescriptor*  dmKeyWriteArray_s(dmBlock *block, const char *name, const short* val, 
				const char *unit, const char *desc, const long nvals)
{
  return dmp_key_write( block, name, DT_INT2, val, nvals, unit, desc, UT_TRUE );
}

dmDescriptor*  dmKeyWriteArray_l(dmBlock *block, const char *name, const long* val, 
				const char *unit, const char *desc, const long nvals)
{
  return dmp_key_write( block, name, DT_INT4, val, nvals, unit, desc, UT_TRUE );
}

dmDescriptor*  dmKeyWriteArray_f(dmBlock *block, const char *name, const float* val, 
				const char *unit, const char *desc, const long nvals)
{
  return dmp_key_write( block, name, DT_REAL4, val, nvals, unit, desc, UT_TRUE );
}

dmDescriptor*  dmKeyWriteArray_d(dmBlock *block, const char *name, const double* val, 
				const char *unit, const char *desc, const long nvals)
{
  return dmp_key_write( block, name, DT_REAL8, val, nvals, unit, desc, UT_TRUE );
}

dmDescriptor*  dmKeyWriteArray_q(dmBlock *block, const char *name, const dmBool* val,
				const char *unit, const char *desc, const long nvals)
{
  return dmp_key_write( block, name, DT_LOGICAL, val, nvals, unit, desc, UT_TRUE );
}

dmDescriptor*  dmKeyWriteArray_c(dmBlock *block, const char *name, const char* *val, 
				const char *unit, const char *desc, const long nvals)
{
  return dmp_key_write( block, name, DT_CHAR, val, nvals, unit, desc, UT_TRUE );
}

dmDescriptor*  dmKeyWriteArray_ub(dmBlock *block, const char *name, 
				 const unsigned char* val,
				 const char *unit, const char *desc, const long nvals)
{
  return dmp_key_write( block, name, DT_UINT1, val, nvals, unit, desc, UT_TRUE );
}

dmDescriptor*  dmKeyWriteArray_us(dmBlock *block, const char *name, 
				 const unsigned short* val,
				 const char *unit, const char *desc, const long nvals)
{
  return dmp_key_write( block, name, DT_UINT2, val, nvals, unit, desc, UT_TRUE );
}

dmDescriptor*  dmKeyWriteArray_ul(dmBlock *block, const char *name, 
				 const unsigned long* val, const char *unit, const char *desc, 
				 const long nvals)
{
  return dmp_key_write( block, name, DT_UINT4, val, nvals, unit, desc, UT_TRUE );
}


/***************************************************************/
/* Function:  dmSetArray_*                                     */
/*                                                             */
/* Description:                                                */
/*   Inserts rows of array data into buffer                    */
/*   Calls dmSetVectorArray() to do work.                      */
/*   Last letter of function name indicates data type.         */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *dd    - Pointer to descriptor             */
/*    char          *vals  - pointer to data                   */
/*    long           nvals - number of values to insert        */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmErrCode  status  (dmSUCCESS or dmFAILURE)              */
/*                                                             */
/***************************************************************/
dmErrCode  dmSetArray_s(dmDescriptor* dd, short* vals, long nvals )
{
  return dmSetVectorArray( dd, DT_INT2, vals, 1L, 1L, nvals );
}

dmErrCode  dmSetArray_l(dmDescriptor* dd, long* vals, long nvals )
{
  return dmSetVectorArray( dd, DT_INT4, vals, 1L, 1L, nvals );
}

dmErrCode  dmSetArray_f(dmDescriptor* dd, float* vals, long nvals )
{
  return dmSetVectorArray( dd, DT_REAL4, vals, 1L, 1L, nvals );
}

dmErrCode  dmSetArray_d(dmDescriptor* dd, double* vals, long nvals )
{
  return dmSetVectorArray( dd, DT_REAL8, vals, 1L, 1L, nvals );
}

dmErrCode  dmSetArray_c(dmDescriptor* dd, char** vals, long nvals )
{
  return dmSetVectorArray( dd, DT_CHAR, vals, 1L, 1L, nvals );
}

dmErrCode  dmSetArray_us(dmDescriptor* dd, unsigned short* vals, long nvals )
{
  return dmSetVectorArray( dd, DT_UINT2, vals, 1L, 1L, nvals );
}

dmErrCode  dmSetArray_ul(dmDescriptor* dd, unsigned long* vals, long nvals )
{
  return dmSetVectorArray( dd, DT_UINT4, vals, 1L, 1L, nvals );
}

dmErrCode  dmSetArray_ub( dmDescriptor* dd, unsigned char* vals, long nvals )
{
 return  dmSetVectorArray( dd, DT_UINT1, vals, 1L, 1L, nvals );
}

dmErrCode  dmSetArray_q(dmDescriptor* dd, dmBool* vals, long nvals )
{
  return dmSetVectorArray( dd, DT_LOGICAL, vals, 1L, 1L, nvals );
}

dmErrCode  dmSetArray_bit(dmDescriptor* dd, void* vals, long nvals )
{
  return dmSetVectorArray( dd, DT_BIT, vals, 1L, 1L, nvals );
}


/***************************************************************/
/* Function:  dmSetScalor_*                                    */
/*                                                             */
/* Description:                                                */
/*   Inserts one scalar datum into buffer                      */
/*   Last letter of function name indicates data type.         */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *dd    - Pointer to descriptor             */
/*    <data type>    val   - data value                        */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmErrCode  status  (dmSUCCESS or dmFAILURE)              */
/*                                                             */
/***************************************************************/
dmErrCode  dmSetScalar_s(dmDescriptor* dd, short val )
{
  return dmSetVectorArray( dd, DT_INT2, &val, 1L, 1L, 1L );
}

dmErrCode  dmSetScalar_f(dmDescriptor* dd, float val )
{
  return dmSetVectorArray( dd, DT_REAL4, &val, 1L, 1L, 1L );
}

dmErrCode  dmSetScalar_d(dmDescriptor* dd, double val )
{
  return dmSetVectorArray( dd, DT_REAL8, &val, 1L, 1L, 1L );
}

dmErrCode  dmSetScalar_l(dmDescriptor* dd, long val )
{
  return dmSetVectorArray( dd, DT_INT4, &val, 1L, 1L, 1L );
}

dmErrCode  dmSetScalar_c(dmDescriptor* dd, char* val )
{
  return dmSetVectorArray( dd, DT_CHAR, &val, 1L, 1L, 1L );
}

dmErrCode  dmSetScalar_q(dmDescriptor* dd, dmBool val )
{
  return dmSetVectorArray( dd, DT_LOGICAL, &val, 1L, 1L, 1L );
}

dmErrCode  dmSetScalar_bit(dmDescriptor* dd, void*  val )
{
  long nvals = 1L;

  if ( dd )
    nvals = dmi_descriptor_get_array_size( dd );

  return dmSetVectorArray( dd, DT_BIT, val, 1L, 1L, nvals );
}

dmErrCode  dmSetScalar_ub(dmDescriptor* dd, unsigned char val )
{
  return dmSetVectorArray( dd, DT_UINT1, &val, 1L, 1L, 1L );
}

dmErrCode  dmSetScalar_us(dmDescriptor* dd, unsigned short val )
{
  return dmSetVectorArray( dd, DT_UINT2, &val, 1L, 1L, 1L );
}

dmErrCode  dmSetScalar_ul(dmDescriptor* dd, unsigned long val )
{
  return dmSetVectorArray( dd, DT_UINT4, &val, 1L, 1L, 1L );
}


/***************************************************************/
/* Function:  dmSetScalors_*                                   */
/*                                                             */
/* Description:                                                */
/*   Inserts rows of data into buffer                          */
/*   Last letter of function name indicates data type.         */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *dd    - Pointer to descriptor             */
/*    char          *vals  - pointer to data                   */
/*    long           row   - starting row number               */
/*    long           nrows - number of rows of data to insert  */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    long status  (dmSUCCESS or dmFAILURE)                    */
/*                                                             */
/***************************************************************/
long  dmSetScalars_s(dmDescriptor* dd, short* vals, long row, long nrows)
{
  return dmSetVectorArrays( dd, DT_INT2, vals, 1L, row, nrows, 1L, 1L );
}

long  dmSetScalars_l(dmDescriptor* dd, long* vals, long row, long nrows)
{
  return dmSetVectorArrays( dd, DT_INT4, vals, 1L, row, nrows, 1L, 1L );
}

long  dmSetScalars_f(dmDescriptor* dd, float* vals, long row, long nrows)
{
  return dmSetVectorArrays( dd, DT_REAL4, vals, 1L, row, nrows, 1L, 1L );
}

long  dmSetScalars_d(dmDescriptor* dd, double* vals, long row, long nrows)
{
  return dmSetVectorArrays( dd, DT_REAL8, vals, 1L, row, nrows, 1L, 1L );
}

long  dmSetScalars_c(dmDescriptor* dd, char** vals, long row, long nrows)
{
  return dmSetVectorArrays( dd, DT_CHAR, vals, 1L, row, nrows, 1L, 1L );
}

long  dmSetScalars_q(dmDescriptor* dd, dmBool* vals, long row, long nrows)
{
  return dmSetVectorArrays( dd, DT_LOGICAL, vals, 1L, row, nrows, 1L, 1L );
}

long  dmSetScalars_ub(dmDescriptor* dd, unsigned char* vals, long row, long nrows)
{
  return dmSetVectorArrays( dd, DT_UINT1, vals, 1L, row, nrows, 1L, 1L );
}

long  dmSetScalars_bit(dmDescriptor* dd, void* vals, long row, long nrows)
{
  long nvals = 1;

  if ( dd )
    nvals = dmi_descriptor_get_array_size( dd );
  return dmSetVectorArrays( dd, DT_BIT, vals, 1L, row, nrows, 1L, nvals );
}

long  dmSetScalars_us(dmDescriptor* dd, unsigned short* vals, long row, long nrows)
{
  return dmSetVectorArrays( dd, DT_UINT2, vals, 1L, row, nrows, 1L, 1L );
}

long  dmSetScalars_ul(dmDescriptor* dd, unsigned long* vals, long row, long nrows)
{
  return dmSetVectorArrays( dd, DT_UINT4, vals, 1L, row, nrows, 1L, 1L );
}


/***************************************************************/
/* Function:  dmSetVector_*                                    */
/*                                                             */
/* Description:                                                */
/*   Set one data value into buffer                            */
/*   Last letter of function name indicates data type.         */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *dd     - Pointer to descriptor            */
/*    <data type>   *vals   - values                           */
/*    long           dim    - dim                              */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmErrCode  status (dmSUCCES or dmFAILURE)                */
/*                                                             */
/***************************************************************/
dmErrCode  dmSetVector_q(dmDescriptor* dd, dmBool* vals,  long dim)
{
  return dmSetVectorArray( dd, DT_LOGICAL, vals, dim, 1L, 1L );
}

dmErrCode  dmSetVector_c(dmDescriptor* dd, char** vals,  long dim)
{
  return dmSetVectorArray( dd, DT_CHAR, vals, dim, 1L, 1L );
}

dmErrCode  dmSetVector_ub(dmDescriptor* dd, unsigned char* vals,  long dim)
{
  return dmSetVectorArray( dd, DT_UINT1, vals, dim, 1L, 1L );
}

dmErrCode  dmSetVector_ul(dmDescriptor* dd, unsigned long* vals,  long dim)
{
  return dmSetVectorArray( dd, DT_UINT4, vals, dim, 1L, 1L );
}

dmErrCode  dmSetVector_us(dmDescriptor* dd, unsigned short* vals,  long dim)
{
  return dmSetVectorArray( dd, DT_UINT2, vals, dim, 1L, 1L );
}

dmErrCode  dmSetVector_s(dmDescriptor* dd, short* vals,  long dim)
{
  return dmSetVectorArray( dd, DT_INT2, vals, dim, 1L, 1L );
}

dmErrCode  dmSetVector_l(dmDescriptor* dd, long* vals,  long dim)
{
  return dmSetVectorArray( dd, DT_INT4, vals, dim, 1L, 1L );
}

dmErrCode  dmSetVector_f(dmDescriptor* dd, float* vals,  long dim)
{
  return dmSetVectorArray( dd, DT_REAL4, vals, dim, 1L, 1L );
}

dmErrCode  dmSetVector_d( dmDescriptor* dd, double* vals, long dim )
{ 
  return dmSetVectorArray( dd, DT_REAL8, vals, dim, 1L, 1L );
}


/***************************************************************/
/* Function:  dmSetVectors_*                                   */
/*                                                             */
/* Description:                                                */
/*   Set vector data into buffer                               */
/*   Last letter of function name indicates data type.         */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *pDesc  - Pointer to descriptor            */
/*    <data type>   *vals   - values                           */
/*    long           dim    - vector dim                       */
/*    long           row    - starting row                     */
/*    long           nrows  - number of rows                   */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmErrCode  status (dmSUCCES or dmFAILURE)                */
/*                                                             */
/***************************************************************/
dmErrCode  dmSetVectors_s(dmDescriptor* pDesc, short* vals, long dim, long row, long nrows)
{
  return dmSetVectorArrays( pDesc, DT_INT2, vals, dim, row, nrows, 1L, 1L  );
}

dmErrCode  dmSetVectors_l(dmDescriptor* pDesc, long* vals, long dim, long row, long nrows)
{
  return dmSetVectorArrays( pDesc, DT_INT4, vals, dim, row, nrows, 1L, 1L  );
}

dmErrCode  dmSetVectors_f(dmDescriptor* pDesc, float* vals, long dim, long row, long nrows)
{
 return dmSetVectorArrays( pDesc, DT_REAL4, vals, dim, row, nrows, 1L, 1L  );
}

dmErrCode  dmSetVectors_d(dmDescriptor* pDesc, double* vals, long dim, long row,  long nrows)
{
  return dmSetVectorArrays( pDesc, DT_REAL8, vals, dim, row, nrows, 1L, 1L  );
}

dmErrCode  dmSetVectors_q(dmDescriptor* pDesc, dmBool* vals, long dim, long row, long nrows)
{
  return dmSetVectorArrays( pDesc, DT_LOGICAL, vals, dim, row, nrows, 1L, 1L  );
}

dmErrCode  dmSetVectors_c(dmDescriptor* pDesc, char** vals, long dim, long row, long nrows)
{
  return dmSetVectorArrays( pDesc, DT_CHAR, vals, dim, row, nrows, 1L, 1L  );
}

dmErrCode  dmSetVectors_ub(dmDescriptor* pDesc, unsigned char* vals, long dim, long row, long nrows)
{
  return dmSetVectorArrays( pDesc, DT_UINT1, vals, dim, row, nrows, 1L, 1L  );
}

dmErrCode dmSetVectors_us(dmDescriptor* pDesc, unsigned short* vals, long dim, long row, long nrows)
{
  return dmSetVectorArrays( pDesc, DT_UINT2, vals, dim, row, nrows, 1L, 1L  );
}

dmErrCode  dmSetVectors_ul(dmDescriptor* pDesc, unsigned long* vals, long dim, long row, long nrows)
{
  return dmSetVectorArrays( pDesc, DT_UINT4, vals, dim, row, nrows, 1L, 1L  );
}

 
/***************************************************************/
/* Function:  dmSetVectorArray_*                               */
/*                                                             */
/* Description:                                                */
/*   Inserts vector data into buffer                           */
/*   Last letter of function name indicates data type.         */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *dd    - Pointer to descriptor             */
/*    <data type>    val   - data value                        */
/*    long           dim   - vector dim                        */
/*    long           nvals - number of values to insert        */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmErrCode  status  (dmSUCCESS or dmFAILURE)              */
/*                                                             */
/***************************************************************/
dmErrCode  dmSetVectorArray_s(dmDescriptor* dd, short* val, long dim, long nvals)
{
  return dmSetVectorArray( dd, DT_INT2, val, dim, 1L, nvals );
}

dmErrCode  dmSetVectorArray_f(dmDescriptor* dd, float* val, long dim, long nvals)
{
  return dmSetVectorArray( dd, DT_REAL4, val, dim, 1L, nvals );
}

dmErrCode  dmSetVectorArray_d(dmDescriptor* dd, double* val, long dim, long nvals)
{
  return dmSetVectorArray( dd, DT_REAL8, val, dim, 1L, nvals );
}

dmErrCode  dmSetVectorArray_l(dmDescriptor* dd, long* val, long dim, long nvals)
{
  return dmSetVectorArray( dd, DT_INT4, val, dim, 1L, nvals );
}

dmErrCode  dmSetVectorArray_c(dmDescriptor* dd, char** val, long dim, long nvals)
{
  return dmSetVectorArray( dd, DT_CHAR, val, dim, 1L, nvals );
}

dmErrCode  dmSetVectorArray_ub(dmDescriptor* dd, unsigned char* val, long dim, long nvals)
{
  return dmSetVectorArray( dd, DT_UINT1, val, dim, 1L, nvals );
}

dmErrCode  dmSetVectorArray_us(dmDescriptor* dd, unsigned short* val, long dim, long nvals)
{
  return dmSetVectorArray( dd, DT_UINT2, val, dim, 1L, nvals );
}

dmErrCode  dmSetVectorArray_ul(dmDescriptor* dd, unsigned long* val, long dim, long nvals)
{
  return dmSetVectorArray( dd, DT_UINT4, val, dim, 1L, nvals );
}

dmErrCode  dmSetVectorArray_q(dmDescriptor* dd, dmBool* val, long dim, long nvals)
{
  return dmSetVectorArray( dd, DT_LOGICAL, val, dim, 1L, nvals );
}


/***************************************************************/
/* Function:  dmSetVectorArrays_*                              */
/*                                                             */
/* Description:                                                */
/*   Set vector data in buffer                                 */
/*   Last letter of function name indicates data type.         */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *pDesc  - Pointer to descriptor            */
/*    <data type>   *invals - values                           */
/*    long           dim    - vector dim                       */
/*    long           row    - starting row                     */
/*    long           nrows  - number of rows                   */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    dmErrCode  status (dmSUCCES or dmFAILURE)                */
/*                                                             */
/***************************************************************/
dmErrCode    dmSetVectorArrays_s(dmDescriptor* pDesc, short* invals, long dim, long row, long nrows)
{
  long nvals = dmGetArraySize( pDesc );
  return dmSetVectorArrays( pDesc, DT_INT2, invals, dim, row, nrows, 1, nvals  );
}

dmErrCode    dmSetVectorArrays_l(dmDescriptor* pDesc, long* invals, long dim, long row, long nrows)
{
  long nvals = dmGetArraySize( pDesc );
  return dmSetVectorArrays( pDesc, DT_INT4, invals, dim, row, nrows, 1, nvals  );
}

dmErrCode    dmSetVectorArrays_f(dmDescriptor* pDesc, float* invals, long dim, long row, long nrows)
{
  long nvals = dmGetArraySize( pDesc );
  return dmSetVectorArrays( pDesc, DT_REAL4, invals, dim, row, nrows, 1, nvals  );
}

dmErrCode    dmSetVectorArrays_d(dmDescriptor* pDesc, double* invals, long dim, long row, long nrows)
{
  long nvals = dmGetArraySize( pDesc );
  return dmSetVectorArrays( pDesc, DT_REAL8, invals, dim, row, nrows, 1, nvals  );
}

dmErrCode    dmSetVectorArrays_q(dmDescriptor* pDesc, dmBool* invals, long dim, long row, long nrows)
{
  long nvals = dmGetArraySize( pDesc );
  return dmSetVectorArrays( pDesc, DT_LOGICAL, invals, dim, row, nrows, 1, nvals  );
}

dmErrCode    dmSetVectorArrays_c(dmDescriptor* pDesc, char** invals, long dim,  long row, long nrows)
{
  long nvals = dmGetArraySize( pDesc );
  return dmSetVectorArrays( pDesc, DT_CHAR, invals, dim, row, nrows, 1, nvals  );
}

dmErrCode    dmSetVectorArrays_ub(dmDescriptor* pDesc, unsigned char* invals, long dim, 
		    long row, long nrows)
{
  long nvals = dmGetArraySize( pDesc );
  return dmSetVectorArrays( pDesc, DT_UINT1, invals, dim, row, nrows, 1, nvals  );
}
dmErrCode   dmSetVectorArrays_us(dmDescriptor* pDesc, unsigned short* invals, long dim,  long row, 
		long nrows)
{
  long nvals = dmGetArraySize( pDesc );
  return dmSetVectorArrays( pDesc, DT_UINT2, invals, dim, row, nrows, 1, nvals  );
}

dmErrCode   dmSetVectorArrays_ul(dmDescriptor* pDesc, unsigned long* invals, long dim, long row, 
		long nrows)
{
  long nvals = dmGetArraySize( pDesc );
  return dmSetVectorArrays( pDesc, DT_UINT4, invals, dim, row, nrows, 1, nvals  );
}


/***************************************************************/
/* Function:  dmSubspaceColGet_*                               */
/*                                                             */
/* Description:                                                */
/*   These functions returns subspace values                   */
/*   Last letter of function name indicates data type          */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *filter  - Pointer to descriptor           */
/*    long          *nvalues - number of values                */
/*                                                             */
/* Output parameters:                                          */
/*    <data type > **value1  - subspace values                 */
/*    <data type > **value2  - subspace values                 */
/*                                                             */
/* Returns:                                                    */
/*    int  status  (dmSUCCESS or dmFAILURE)                    */
/*                                                             */
/***************************************************************/
int
dmSubspaceColGet_s(dmDescriptor* filter, short** value1, 
				   short** value2, long* nvalues)
{
  return dmi_subspace_col_get_values( filter, 1, DT_INT2, (void**)value1, (void**)value2, nvalues );
}


int
dmSubspaceColGet_l(dmDescriptor* filter, long** value1, 
				   long** value2, long* nvalues)
{
  return dmi_subspace_col_get_values( filter, 1, DT_INT4, (void**)value1, (void**)value2, nvalues );
}

int
dmSubspaceColGet_us(dmDescriptor* filter, unsigned short** value1, 
				   unsigned short** value2, long* nvalues)
{
  return dmi_subspace_col_get_values( filter, 1, DT_UINT2, (void**)value1, (void**)value2, nvalues );
}


int
dmSubspaceColGet_ul(dmDescriptor* filter, unsigned long** value1, 
				   unsigned long** value2, long* nvalues)
{
  return dmi_subspace_col_get_values( filter, 1, DT_UINT4, (void**)value1, (void**)value2, nvalues );
}

int
dmSubspaceColGet_f(dmDescriptor* filter, float** value1, 
				   float** value2, long* nvalues)
{
  return dmi_subspace_col_get_values( filter, 1, DT_REAL4, (void**)value1, (void**)value2, nvalues );
}

int
dmSubspaceColGet_d(dmDescriptor* filter, double** value1, 
				   double** value2, long* nvalues)
{
  return dmi_subspace_col_get_values( filter, 1, DT_REAL8, (void**)value1, (void**)value2, nvalues );
}


/***************************************************************/
/* Function:  dmSubspaceColRead_*                              */
/*                                                             */
/* Description:                                                */
/*   Essentially a combination of dmSubspaceColOpen and        */
/*   dmSubspaceColGet                                          */
/*   Last letter of function name indicates data type          */
/*                                                             */
/* Input parameters:                                           */
/*    dmBlock       *block   - Pointer to block                */
/*    char          *name    - column name                     */
/*                                                             */
/* Output parameters:                                          */
/*    <data type > **value1  - subspace values                 */
/*    <data type > **value2  - subspace values                 */
/*    long          *nvalues - number of values                */
/*                                                             */
/* Returns:                                                    */
/*    dmDescriptor *subspace_dd                                */
/*                                                             */
/***************************************************************/
dmDescriptor*
dmSubspaceColRead_s(dmBlock* block, char* name, short** value1, 
		    short** value2, long* nvalues)
{
  dmDescriptor* dmDSS = NULL;

  *value1 = NULL;
  *value2 = NULL;
  *nvalues = 0;
  if(!(dmDSS = dmSubspaceColOpen(block, name)))
      return NULL;
  if(dmSubspaceColGet_s(dmDSS, value1, value2, nvalues))
      return dmDSS;
  else
      return NULL;
}

dmDescriptor*
dmSubspaceColRead_l(dmBlock* block, char* name, long** value1, 
		    long** value2, long* nvalues)
{
  dmDescriptor* dmDSS = NULL;

  *value1 = NULL;
  *value2 = NULL;
  *nvalues = 0;
  if(!(dmDSS = dmSubspaceColOpen(block, name)))
      return NULL;
  if(dmSubspaceColGet_l(dmDSS, value1, value2, nvalues))
      return dmDSS;
  else
      return NULL;
}

dmDescriptor*
dmSubspaceColRead_us(dmBlock* block, char* name, unsigned short** value1, 
		    unsigned short** value2, long* nvalues)
{
  dmDescriptor* dmDSS = NULL;

  *value1 = NULL;
  *value2 = NULL;
  *nvalues = 0;
  if(!(dmDSS = dmSubspaceColOpen(block, name)))
      return NULL;
  if(dmSubspaceColGet_us(dmDSS, value1, value2, nvalues))
      return dmDSS;
  else
      return NULL;
}

dmDescriptor*
dmSubspaceColRead_ul(dmBlock* block, char* name, unsigned long** value1, 
		    unsigned long** value2, long* nvalues)
{
  dmDescriptor* dmDSS = NULL;

  *value1 = NULL;
  *value2 = NULL;
  *nvalues = 0;
  if(!(dmDSS = dmSubspaceColOpen(block, name)))
      return NULL;
  if(dmSubspaceColGet_ul(dmDSS, value1, value2, nvalues))
      return dmDSS;
  else
      return NULL;
}

dmDescriptor*
dmSubspaceColRead_f(dmBlock* block, char* name, float** value1, 
		    float** value2, long* nvalues)
{
  dmDescriptor* dmDSS = NULL;

  *value1 = NULL;
  *value2 = NULL;
  *nvalues = 0;
  if(!(dmDSS = dmSubspaceColOpen(block, name)))
      return NULL;
  if(dmSubspaceColGet_f(dmDSS, value1, value2, nvalues))
      return dmDSS;
  else
      return NULL;
}

dmDescriptor*
dmSubspaceColRead_d(dmBlock* block, char* name, double** value1, 
		    double** value2, long* nvalues)
{
  dmDescriptor* dmDSS = NULL;
  *value1 = NULL;
  *value2 = NULL;
  *nvalues = 0;
  if(!(dmDSS = dmSubspaceColOpen(block, name)))
      return NULL;
  if(dmSubspaceColGet_d(dmDSS, value1, value2, nvalues))
      return dmDSS;
  else
      return NULL;
}


/***************************************************************/
/* Function:  dmSubspaceColSet_*                               */
/*                                                             */
/* Description:                                                */
/*   These functions sets subspace values                      */
/*   Last letter of function name indicates data type          */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *filter  - Pointer to descriptor           */
/*    <data type >  *value1  - subspace values                 */
/*    <data type >  *value2  - subspace values                 */
/*    long           nvalues - number of values                */
/*                                                             */
/* Output parameters:                                          */
/*     none                                                    */
/*                                                             */
/* Returns:                                                    */
/*    int  status  (dmSUCCESS or dmFAILURE)                    */
/*                                                             */
/***************************************************************/
int             
dmSubspaceColSet_s(dmDescriptor* filter, short* value1,
				   short* value2, long nvalues)
{
  return dmi_subspace_col_update_null( filter, NULL, 1, DT_INT2, (void*)value1, (void*)value2, nvalues, UT_FALSE );
}

int             
dmSubspaceColSet_us(dmDescriptor* filter, unsigned short* value1,
				   unsigned short* value2, long nvalues)
{
  return dmi_subspace_col_update_null( filter, NULL, 1, DT_UINT2, (void*)value1, (void*)value2, nvalues, UT_FALSE );
}

int             
dmSubspaceColSet_ul(dmDescriptor* filter, unsigned long* value1,
				   unsigned long* value2, long nvalues)
{
  return dmi_subspace_col_update_null( filter, NULL, 1, DT_UINT4, (void*)value1, (void*)value2, nvalues, UT_FALSE );
}

int             
dmSubspaceColSet_l(dmDescriptor* filter, long* value1,
				   long* value2, long nvalues)
{
  return dmi_subspace_col_update_null( filter, NULL, 1, DT_INT4, (void*)value1, (void*)value2, nvalues, UT_FALSE );
}

int             
dmSubspaceColSet_f(dmDescriptor* filter, float* value1,
				   float* value2, long nvalues)
{
  return dmi_subspace_col_update_null( filter, NULL, 1, DT_REAL4, (void*)value1, (void*)value2, nvalues, UT_FALSE );
}

int             
dmSubspaceColSet_d(dmDescriptor* filter, double* value1,
				   double* value2, long nvalues)
{
  return dmi_subspace_col_update_null( filter, NULL, 1, DT_REAL8, (void*)value1, (void*)value2, nvalues, UT_FALSE );
}


/***************************************************************/
/* Function:  dmSubspaceColUpdate_*                            */
/*                                                             */
/* Description:                                                */
/*   Update functions intersect input values with existing     */
/*   values, as opposed to overwriting them as in the set      */
/*   routines.                                                 */
/*   Last letter of function name indicates data type          */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor  *filter  - Pointer to descriptor           */
/*    <data type >  *value1  - subspace values                 */
/*    <data type >  *value2  - subspace values                 */
/*    long           nvalues - number of values                */
/*                                                             */
/* Output parameters:                                          */
/*     none                                                    */
/*                                                             */
/* Returns:                                                    */
/*    int  status  (dmSUCCESS or dmFAILURE)                    */
/*                                                             */
/***************************************************************/
int             
dmSubspaceColUpdate_s(dmDescriptor* filter, short* value1,
				   short* value2, long nvalues)
{
  return dmi_subspace_col_update_null( filter, NULL, 1, DT_INT2, (void*)value1, (void*)value2, nvalues, UT_TRUE );
}

int             
dmSubspaceColUpdate_l(dmDescriptor* filter, long* value1,
				   long* value2, long nvalues)
{
  return dmi_subspace_col_update_null( filter, NULL, 1, DT_INT4, (void*)value1, (void*)value2, nvalues, UT_TRUE );
}

int             
dmSubspaceColUpdate_f(dmDescriptor* filter, float* value1,
				   float* value2, long nvalues)
{
  return dmi_subspace_col_update_null( filter, NULL, 1, DT_REAL4, (void*)value1, (void*)value2, nvalues, UT_TRUE );
}

int             
dmSubspaceColUpdate_d(dmDescriptor* filter, double* value1,
				   double* value2, long nvalues)
{
  return dmi_subspace_col_update_null( filter, NULL, 1, DT_REAL8, (void*)value1, (void*)value2, nvalues, UT_TRUE );
}


