/*                                                                
**  Copyright (C) 1998-2001,2003,2006-2007,2013  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/******************************************************************************/
/*  dmrange.c                                                                 */
/*                                                                            */
/*  Get Descriptor Range Functions                                            */
/*                                                                            */
/* These functions attempt to determine the allowable range of values a       */
/* descriptor can have (method specific to descriptor type).                  */
/* Currently only columns are supported.                                      */
/*                                                                            */
/******************************************************************************/

#include "dmlibP.h"


/***************************************************************/
/* Function:  dmp_descriptor_set_range                         */
/*                                                             */
/* Description:                                                */
/*    Set the range on Column descriptor. (TLMIN/TLMAX)        */
/*    For a vector column, all components are set to the same  */
/*    range.                                                   */
/*                                                             */
/* Input parameters:                                           */
/*             dmDescriptor *dd    - descriptor                */
/*       const dmDataType    type  - data type                 */
/*             void         *tlmin - min range value           */
/*             void         *tlmax - max range value           */   
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*     int  dmSUCCESS or dmFAILURE                             */
/*                                                             */
/***************************************************************/
int dmp_descriptor_set_range( dmDescriptor* dd, const dmDataType type,
                              void* tlmin, void* tlmax )
{
  dmKernel   kernel;
  dmDataset *ds = NULL;
  dmBlock   *block = NULL;

  if(!dd )
  {
    dmu_err_set(DM_ERR_NULL, "Null arg" );
    return dmFAILURE;
  }

  /* Get Kernel block. */
  block = (dmBlock*)dd->block;
  ds = block->ds;
  kernel = ds->kernel;

  /* Set descriptor range */
  dmi_descriptor_set_range( dd, type, tlmin, tlmax );

  /* Set at kernel level too */
  kernel->ColSetRange( (dmDescriptor*)dd );

  /* Tag block as updated */
  dmi_block_set_update( block );

  return dmu_err_get();
}


/***************************************************************/
/* Function:  dmp_descriptor_REset_range                       */
/*                                                             */
/* Description:                                                */
/*    Change the range on Column descriptor. (TLMIN/TLMAX)     */
/*    This function enables a post-facto change to a column    */
/*    descriptors legal range.  In addition to changing the    */
/*    range values themselves, it also resets the default      */
/*    subsubspace ranges.                                      */
/*                                                             */
/*  This method is being added in support of dmmerge           */
/*  enhancement #13475.                                        */
/*                                                             */
/* Input parameters:                                           */
/*             dmDescriptor *in_dd - descriptor                */
/*       const dmDataType    type  - data type                 */
/*             void         *tlmin - min range value           */
/*             void         *tlmax - max range value           */   
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*     int  dmSUCCESS or dmFAILURE                             */
/*                                                             */
/***************************************************************/
int dmp_descriptor_reset_range( dmDescriptor* in_dd, const dmDataType type, void* tlmin, void* tlmax )
{
   dmBlock      *block  = NULL;
   dmDescriptor *dd     = in_dd;/* column descriptor */
   dmDescriptor *cd     = NULL; /* vector component descriptor */
   dmDescriptor *sd     = NULL; /* subspace descriptor         */
   dmFilter     *filter = NULL;
   dmBool        isDefault = dmFALSE;

   char *name = NULL;

   long nsscpts = 0;
   long vdim    = 0;
   long vcpt    = 0;
   long ii      = 0;
   long start   = 0;
   long end     = 0;

   if( !dd || !tlmin || !tlmax )
   {
     dmu_err_set(DM_ERR_NULL, "Null argument" );
     return dmFAILURE;
   }

   block = (dmBlock*)dd->block;
   nsscpts = dmBlockGetNoSubspaceCpts( block );

   /* Setup column descriptor for extracting subspace filter     */
   /* if vector component input.. get parent                     */
   /* also sets loop control for updating appropriate components */
   vdim = dd->dim;
 
   if ( vdim > 1 ) /* vector descriptor */
   {
     start = 1; 
     end = dd->dim;
   }
   else
   {
     start = end = dd->cptno;

     if ( dd->cptno > 0 ) /* vector component */
       dd = dd->parent;
   }

   /* Get Subspace descriptor containing filter ranges */
   if ( block )
   {
     name  = dmi_descriptor_get_name( dd ); 
     sd = dmi_subspace_col_open( block->subspace, name );
   }
  
   /* If vector descriptor input.. reset each componet to provide range */
   for ( vcpt = start; vcpt <= end ; vcpt++ )
   { 
     if ( vcpt > 0 )
       cd = dmi_descriptor_get_cpt( dd, vcpt );
     else
       cd = dd;

     /* Reset default subspace first */
     if ( sd )
     {
       for ( ii = 1; ii <= nsscpts; ii++ )
       {
         dmBlockSetSubspaceCpt( block, ii );
  
         filter = dmi_subspace_col_get_current_filter( sd );
         isDefault = dmi_filter_range_is_default( dd, filter, vcpt );

 	 if ( isDefault )
         {
           dmi_subspace_col_update_null( sd, NULL, vcpt, type, (void*)tlmin, (void*)tlmax, 1L, UT_FALSE );
         }
       } /* end subspace component loop */

       dmBlockSetSubspaceCpt( block, 1 );
     }

     /* Update descriptor range (tlmin/max) */
     dmp_descriptor_set_range( cd, type, tlmin, tlmax );

   } /* end vector component loop */

   return dmu_err_get();
}


/***************************************************************/
/* Function:  dmDescriptorGetRangeCpt                          */
/*                                                             */
/* Description:                                                */
/*    Get the range of the specified descriptor component      */
/*     - vcpt identifies which component of a vector is of     */
/*       interest.                                             */
/*     - ranges are returned as doubles, with default of       */
/*       -/+ DT_REAL8_MAX                                      */
/*                                                             */
/*     NOTE: return code indicates dmFAILURE when returning    */
/*           default values                                    */
/*                                                             */
/* Input parameters:                                           */
/*             dmDescriptor *pDesc - descriptor                */
/*       const long          vcpt  - vector component          */
/*                                                             */
/* Output parameters:                                          */
/*             double       *tlmin - min range value           */
/*             double       *tlmax - max range value           */ 
/*                                                             */
/* Returns:                                                    */
/*     int  dmSUCCESS or dmFAILURE                             */
/*                                                             */
/***************************************************************/
int dmDescriptorGetRangeCpt_d(dmDescriptor* pDesc, const long vcpt, double* tlmin, double* tlmax)
{
  int status = 0;

  status = dmi_descriptor_get_cpt_range( pDesc, vcpt, DT_REAL8, (void*)tlmin, (void*)tlmax );

  if ( status == dmFAILURE )
  {
    *tlmin = -DT_REAL8_MAX;
    *tlmax = DT_REAL8_MAX;
  }

  return status;
}


/***************************************************************/
/* Function:  dmp_descriptor_set_null                          */
/*                                                             */
/* Description:                                                */
/*  Set the null value on the specified descriptor.            */
/*                                                             */
/* Input parameters:                                           */
/*             dmDescriptor *dd    - descriptor                */
/*       const dmDataType    type  - data type                 */
/*             void         *value - null value                */
/*                                                             */
/* Output parameters:                                          */
/*         none                                                */
/*                                                             */
/* Returns:                                                    */
/*     int  dmSUCCESS or dmFAILURE                             */
/*                                                             */
/***************************************************************/
int dmp_descriptor_set_null( dmDescriptor* dd, const dmDataType type, void* value )
{
   dmKernel kernel;
   dmDataset *ds    = NULL;
   dmBlock   *block = NULL;

   block = (dmBlock*)dd->block;
   ds = block->ds;
   kernel = ds->kernel;

   /* Set descriptor null attribute */
   dmi_descriptor_set_null( dd, type, value );

   /* Set at kernel level too */
   kernel->ColSetNull( dd );

   /* flag block as updated */
   dmi_block_set_update( block );

   return dmu_err_get();
}


/***************************************************************/
/* Function:  dmp_descriptor_set_bin                           */
/*                                                             */
/* Description:                                                */
/*  Set the null value on the specified descriptor.            */
/*                                                             */
/* Input parameters:                                           */
/*             dmDescriptor *dd    - descriptor                */
/*       const dmDataType    type  - data type                 */
/*             void         *value - null value                */
/*                                                             */
/* Output parameters:                                          */
/*         none                                                */
/*                                                             */
/* Returns:                                                    */
/*     int  dmSUCCESS or dmFAILURE                             */
/*                                                             */
/***************************************************************/
int dmp_descriptor_set_bin( dmDescriptor* dd, const dmDataType type, void* value )
{
   dmKernel kernel;
   dmDataset *ds    = NULL;
   dmBlock   *block = NULL;
 
   block = (dmBlock*)dd->block;
   ds = block->ds;
   kernel = ds->kernel;
 
   /* Set descriptor bin attribute */
   dmi_descriptor_set_bin( dd, type, value );

   /* Set at kernel level too */
   kernel->ColSetBin( dd );

   /* flag block as updated */
   dmi_block_set_update( block );
 
   return dmu_err_get();
}


