/*                                                                
**  Copyright (C) 2002-2012  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmclause.c                                                  */
/*                                                               */
/*   Functions that handle parsing on stacks and clause lists    */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"
#include "dmiparser.h"
#include "kertable.h"
#include <string.h>
#include <stdio.h>

dmBlock* dmBlockOpen( dmDataset*, const char*);
dmBlock* dmDatasetClose( dmDataset* );

#define DEBUG 0
#define XDEBUG 1
#define MAXDEPTH 50

enum { NullOp, OrOp, AndOp, ConstOp, CopyOp } dmi_parse_ops;
enum { ExpectUnknown, ExpectArg1, ExpectArg2, ExpectClose, ExpectOp } dmi_parse_expects;
enum { NullClause, BeginClause, InClause, EndClause } dmi_parse_nexts;


/***************************************************************/
/*                                                             */
/* Function:  dmi_parse_clause_operator                        */
/*                                                             */
/* Description:                                                */
/*   Check if the input string starts with a clause operator   */
/*   Clause operators are '&&' or '||'                         */
/*                                                             */
/*  Input parameters:                                          */
/*     char *ptr            String to check                    */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*     1 = Yes it does                                         */
/*     0 = No it does not                                      */
/*                                                             */
/***************************************************************/
int dmi_parse_clause_operator( char* ptr )
{
  char c1 = 0;
  char c2 = 0;
  int retval = 0;

  /* Check empty input */
  if ( !ptr )
    return retval;

  /* set local variables for first two chars */
  c1 = *ptr;
  if ( c1 )
    c2 = *(ptr+1);

  /* test values of first two chars */
  if (( c1 == '&' && c2 == '&' ) || ( c1 == '|' && c2 == '|' ))
    retval = 1;

  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_parse_operator                               */
/*                                                             */
/* Description:                                                */
/*   Check if the input string starts with an operator         */
/*   Operators are  '=', '<', '>', '!='                        */
/*   False if string is null                                   */
/*                                                             */
/*  Input parameters:                                          */
/*     char *ptr            String to check                    */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*     1 = Yes it does                                         */
/*     0 = No it does not                                      */
/*                                                             */
/***************************************************************/
int dmi_parse_operator( char* ptr )
{
  char c1 = 0;
  char c2 = 0;
  int retval = 0;

  /* check empty input */
  if ( !ptr )
    return retval;

  /* set local variable for first two chars */
  c1 = *ptr;
  if ( c1 )
    c2 = *(ptr+1);

  /* test value(s) for operators */
  switch( c1 )
  {
  case '=':
  case '<':
  case '>':
    retval = 1;
    break;
  case '!':
    if( c2 == '=' )
      retval = 1;
    break;
  default:
    retval = 0;
  }

  return retval;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_parse_test_clause_open                       */
/*                                                             */
/* Description:                                                */
/*   Test Clause open..                                        */
/*   If we are at depth 1 when we see an operator, then we     */
/*   consider the input string a clause.                       */
/*                                                             */
/*   A parenthesis can indicate a clause opening or an         */
/*   'implicit vector' (x,y).  This routine checks that        */
/*   we have a real clause opening.                            */
/*                                                             */
/*  Input parameters:                                          */
/*     char *ptr            String to check                    */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*     1 = Yes it is                                           */
/*     0 = No it is not                                        */
/*                                                             */
/***************************************************************/
int dmi_parse_test_clause_open( char* ptr )
{
  int Clause = 0;
  char *uptr = NULL;
  long depth = 1;

  /* String should start with a parenthesis.. possible clause */
  if ( ptr && *ptr == '(' )
  {
    uptr = ptr + 1;
    while ( depth && !Clause )
    {
      if ( *uptr == '(' )
      {
	/* Found another paren level.. */
	depth++; uptr++;
      }
      else if ( *uptr == ')' ) 
      {
	/* End a paren level.. */
	depth--; uptr++;
      }
      else if ( *uptr == '\0' ) 
      {
	/* end of input */
	depth = 0;
      }
      else if ( dmi_parse_operator( uptr ) || dmi_parse_clause_operator( uptr ) )
      {
	/* Found an operator.. */
	if ( depth == 1 ){ Clause = 1; }
        depth = 0;  /* end loop */
      }
      else
      {
	/* keep checking */
	uptr++;
      }

    } /* end while */
  }

 return Clause;
}

/***************************************************************/
/*                                                             */
/* Function:    dmi_parse_state_disp                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*      dmParseStack *stack                                    */
/*                                                             */
/*  Output parameters:                                         */
/*      none                                                   */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_parse_state_disp( dmParseStack* stack )
{
  char buf[4096];    /* formatted string to display */
  char narg1[80];
  char narg2[80];

  long ii  = 0;           /* counter */
  long arg = 0;          /* index   */
  int disp = 0;

  dmParseStackEntry *entry =  NULL;

  char *opnames[] = { "Null", "Or", "And", "Const", "Copy" };

  for ( ii = 0; ii < stack->ns; ii++ )
  {
    arg = ii+1;
    entry = &stack->entry[ii];
    if ( entry->op == ConstOp )
    {
      dmi_parse_stack_format_name( stack, arg, narg1 );
      sprintf( buf, "#%3ld  = %s %s", arg, opnames[entry->op], narg1 );
      disp = 0;
    }
    else
    {
      disp = 1;
      dmi_parse_stack_format_name( stack, entry->arg[0], narg1 );
      dmi_parse_stack_format_name( stack, entry->arg[1], narg2 );
      sprintf( buf, "#%3ld  = %s( %s, %s )", arg, opnames[entry->op], narg1, narg2 );
    }
    if ( dmverbose >= 5 || disp )
      printf( "%s\n", buf );
  }

}

/***************************************************************/
/*                                                             */
/* Function:    dmi_parse_stack_format_name                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*     dmParseStack  *stack                                    */
/*     long           arg1                                     */
/*                                                             */
/*  Output parameters:                                         */
/*    char* outstr                                             */
/*                                                             */
/*  Returns:                                                   */
/*    void                                                     */
/*                                                             */
/***************************************************************/
void dmi_parse_stack_format_name( dmParseStack* stack, long arg1, char* outstr )
{
  dmParseStackEntry* entry1;

  if ( arg1 == 0  )
  {
    sprintf( outstr, "-" );
  }
  else if ( arg1 > stack->ns )
  {
    sprintf( outstr, "ERROR" );
  }
  else 
  {  
    entry1 = &stack->entry[arg1-1];

    if ( entry1->op == ConstOp && entry1->ptr )
      sprintf( outstr, entry1->ptr );
    else
      sprintf( outstr, "#%ld", arg1 );
  }
}

/***************************************************************/
/*                                                             */
/* Function:   dmi_parse_state_init                            */
/*                                                             */
/* Description:                                                */
/*       Allocate a parse state sturcture which include a new  */
/*       stack                                                 */
/*                                                             */
/*  Input parameters:                                          */
/*        long  n - number of stack entries to allocate        */
/*                                                             */
/*  Output parameters:                                         */
/*        none                                                 */
/*                                                             */
/*  Returns:                                                   */
/*        dmParseState *state                                  */
/*                                                             */
/***************************************************************/
dmParseState* dmi_parse_state_init( long n )
{
  dmParseState *state = NULL;
  dmSubspace   *rts   = NULL;

  state = calloc( 1, sizeof( dmParseState ));
  state->stack = dmi_parse_stack_init( n );

  /* allocate rts space */
  rts = dmi_subspace_alloc( 0, 1 ); 
  state->rts = rts;

  /* Initialize current component to zero */
  rts->cpt = 0;

  return state;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_parse_stack_init                             */
/*                                                             */
/* Description:                                                */
/*     Allocates a new stack and space for n entries           */
/*                                                             */
/*  Input parameters:                                          */
/*       long  n  - number of stack entries to allocate        */
/*                                                             */
/*  Output parameters:                                         */
/*        none                                                 */
/*                                                             */
/*  Returns:                                                   */
/*        dmParseStack *stack                                  */
/*                                                             */
/***************************************************************/
dmParseStack* dmi_parse_stack_init( long n )
{
  dmParseStack *Stack = NULL;

  Stack = calloc( 1, sizeof( dmParseStack ));

  if ( n > 0 )
    Stack->entry = calloc( n, sizeof( dmParseStackEntry ));
  else
    Stack->entry = NULL;

  Stack->ns    = 0;
  Stack->nsmax = n;

  return Stack;
}

/***************************************************************/
/*                                                             */
/* Function: dmi_parse_stack_entry                             */
/*                                                             */
/* Description:                                                */
/*   Returns the specified entry from the stack                */
/*   Will create it if it needs to                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmParseStack *stack                                     */
/*     long          no                                        */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            dmParseStackEntry *entry                         */
/*                                                             */
/***************************************************************/
dmParseStackEntry* dmi_parse_stack_entry( dmParseStack* stack, long no )
{
  long nent;

  if ( no > stack->nsmax )
  {
    nent = no * 2 + 10;
    stack->entry = realloc( stack->entry, nent * sizeof( dmParseStackEntry ));
    memset( &stack->entry[stack->nsmax], 0, (nent - stack->nsmax )* sizeof( dmParseStackEntry ));
    stack->nsmax = nent;
  }

  return &stack->entry[no-1];
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_parse_stack_free                             */
/*                                                             */
/* Description:                                                */
/*   Assumes RTS already free                                  */
/*                                                             */
/*  Input parameters:                                          */
/*      dmParseStack  *stack                                   */
/*                                                             */
/*  Output parameters:                                         */
/*      none                                                   */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_parse_stack_free( dmParseStack* stack )
{
  long id = 0;
  dmParseStackEntry *entry = NULL;

  for ( id = 1; id <= stack->ns; id++ )
  {
    entry = dmi_parse_stack_entry( stack, id );
    if ( entry )
      free( entry->ptr );
  }

  free( stack->id );
  free( stack->expect );
  free( stack->entry );
  free( stack );

}

/***************************************************************/
/*                                                             */
/* Function:  dmi_parse_stack_set                              */
/*                                                             */
/* Description:                                                */
/*   Set the op, arg1 and arg2 values for the specified id     */
/*                                                             */
/*  Input parameters:                                          */
/*      dmParseStack  *stack                                   */
/*      long           id                                      */
/*      long           op                                      */
/*      long           arg1                                    */
/*      long           arg2                                    */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_parse_stack_set( dmParseStack* stack, long id, long op, long arg1, long arg2 )
{
  dmParseStackEntry *entry = NULL;

  entry = dmi_parse_stack_entry( stack, id );

  if ( op   > 0 ) entry->op = op;
  if ( arg1 > 0 ) entry->arg[0] = arg1;
  if ( arg2 > 0 ) entry->arg[1] = arg2;

}

/***************************************************************/
/*                                                             */
/* Function:                                                   */
/*                                                             */
/* Description:                                                */
/*   Adds the values as the next entry on the stack            */
/*                                                             */
/*  Input parameters:                                          */
/*      dmParseStack  *stack                                   */
/*      long           op                                      */
/*      long           arg1                                    */
/*      long           arg2                                    */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*          long  group_no                                     */
/*                                                             */
/***************************************************************/
long dmi_parse_stack_push( dmParseStack* stack, long op, long arg1, long arg2 )
{
  dmParseStackEntry *entry = NULL;
  long group_no = 0;

  group_no = ++stack->ns; 
  entry = dmi_parse_stack_entry( stack, group_no );
  entry->op     = op;
  entry->arg[0] = arg1;
  entry->arg[1] = arg2;
  entry->ptr = NULL;

  return group_no;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_parse_stack_arg_push                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*      dmParseStack  *stack                                   */
/*      long           op                                      */
/*      char          *ptr                                     */
/*                                                             */
/*  Output parameters:                                         */
/*      none                                                   */
/*                                                             */
/*  Returns:                                                   */
/*      long  group_no                                         */
/*                                                             */
/***************************************************************/
long dmi_parse_stack_arg_push( dmParseStack* stack, long op, char* ptr )
{
  long group_no;
  dmParseStackEntry* entry;

  group_no = ++stack->ns; 
  entry = dmi_parse_stack_entry( stack, group_no );
  entry->op     = op;
  entry->arg[0] = 0;
  entry->arg[1] = 0;
  entry->ptr = ptr;

  return group_no;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_parse_stack_arg                             */
/*                                                             */
/* Description:                                                */
/*   Return second argument of operator                        */
/*                                                             */
/*  Input parameters:                                          */
/*        dmParseStack  *stack                                 */
/*        long           id                                    */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*           long retval                                       */
/***************************************************************/
long dmi_parse_stack_arg( dmParseStack* stack, long id )
{
  dmParseStackEntry *entry;
  long retval = 0;

  entry = &stack->entry[id-1];

  switch( entry->op )
  {
  case CopyOp: 
    retval = entry->arg[0];
  case ConstOp:
    retval = id;
  default:
    retval = entry->arg[1];
  }

  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:    dmi_parse_clause_stack                         */
/*                                                             */
/* Description:                                                */
/*   Implement boolean algebra parser                          */
/*                                                             */
/*  Input parameters:                                          */
/*        char          *buf                                   */
/*        dmParseStack *stack                                  */
/*                                                             */
/*  Output parameters:                                         */
/*        none                                                 */
/*                                                             */
/*  Returns:                                                   */
/*        int dmSUCCESS                                        */
/*                                                             */
/***************************************************************/
int dmi_parse_clause_stack( char* buf, dmParseStack* stack )
{

  stack->id    = (long*)malloc( MAXDEPTH * UT_SZ_I );
  stack->expect= (long*)malloc( MAXDEPTH * UT_SZ_I );

  stack->group_no = 0;
  stack->next = BeginClause;
  stack->id[stack->group_no] = dmi_parse_stack_push( stack, CopyOp, 0, 0 );
  stack->expect[stack->group_no] = ExpectArg1;
  stack->clause_depth = 0;
  stack->init = 1;

  if ( dmverbose >= 4 )
    fprintf( stderr, "DM Parser begin: %s\n", buf );

  return dmSUCCESS;
}

/***************************************************************/
/*                                                             */
/* Function: dmi_parse_clause_line                             */
/*                                                             */
/* Description:                                                */
/*   Parse Clause Line takes a single line and increments      */
/*   the parse stack.                                          */
/*      - skips whitespace                                     */
/*      - checks for @file and !@file notation                 */
/*      - checks for a paren clause: (...)                     */
/*                                                             */
/*  Input parameters:                                          */
/*      char          *buf                                     */
/*      dmParseStack  *stack                                   */
/*                                                             */
/*  Output parameters:                                         */
/*      none                                                   */
/*                                                             */
/*  Returns:                                                   */
/*       int retval                                            */
/*                                                             */
/***************************************************************/
int dmi_parse_clause_line( char* buf, dmParseStack* stack )
{

  int loop    = 1;
  int disp    = 0;
  int verb    = 4;
  int exclude = 0;
  int extasc  = 0;
  int retval  = 0;

  long depth = 0;
  long arg   = 0;
  long op    = 0;
  long new   = 0;

  char *clause = NULL;
  char *fptr   = NULL;
  char *ptr    = buf;
  char *start  = buf;

  char *nextnames[] = { "None", "Begin", "In", "End" }; 
  char *expnames[] = { "None", "Arg1", "Arg2", "Close", "Op" };


  /* set pointer to beginning */
  start = cs_advptr( &ptr );

  while ( ptr && loop )
  {
    if ( depth == 0 )
      cs_advptr( &ptr ); /* Skip spaces */
     
    disp = 0;

    /* Check for external file spec                    */
    /* The official syntax is @ or @! but we allow !@  */     
    extasc = 0;
    if(ptr && *ptr == '@' )
    {
      /* Have external file.. check for exclude syntax */
      extasc = 1;
      fptr = ptr+1;
      if (*fptr == '!') 
      {
	exclude = 1; /* Global Exclude */
	fptr++;
      }
    }
    if(ptr && *ptr == '!' && *(ptr+1) == '@' )
    {
      /* Have sloppy external file syntax with exclude.. */
      extasc = 1;
      fptr = ptr + 2;
      exclude = 1; /* Global Exclude */
    }

    /* Handle external file spec */
    if ( extasc )
    {
      char extfile[dmMAXSTR];
      char* eptr = extfile;

      /* Copy just filename segment of ptr string to local variable */
      while ( *fptr && *fptr != ',' && *fptr!= ' ' )
      { *eptr = *fptr++; eptr++; }

      *eptr = 0;

      /* There are two possibilitie with @<file> syntax                         */
      /*   1) Simple text file whose contents are filter directives             */
      /*   2) DM file whose subspace should be used as a filter template        */
      /*                                                                        */
      /* Option 1 is handled here by expanding the file contents into the stack */
      /* Try to open as a DM dataset, if this fails, or DM reports this is of   */
      /* type TEXT, then assume this is option 1...                             */
      /* otherwise, leave it for later in the parsing process where option 2    */
      /* will be handled.                                                       */
      /* NOTE: We would like to allow subflavors of TEXT to pass, so DTF format */
      /*       files could be used as subspace templates, however the auto-     */
      /*       determination logic matches filter syntax to DTF keyword         */
      /*       pattern and so we cannot distinquish between the two :-<         */

      dmBlock*  block;
      dmBool    isText = dmTRUE;
      char*     dsname;

      /* Remove DM syntax from filename for this test */
      /* MCD NOTE: this is what used to happen within dmDatasetAccess,          */
      /*  and is being put here to handle case where the external file has      */
      /*  extended dmsyntax on it ([subspace -expno]).. the loop above splits   */
      /*  the filename at the space within the block, which creates a bad       */
      /*  name for the dmBlockOpen.                                             */
      /*  Alternatively.. the loop above should scan through the block syntax   */
      /*   (using a dm parse routine to get the full filename and open that.)   */
      dsname = dmu_extract_filename( extfile );

      block = dmBlockOpen( NULL, dsname );

      if ( block )
      {
	/* successful open.. check kernel */
	dmKernel kernel = dmi_block_kernel( block );

	if (! cs_uceq( kernel->name, "TEXT" ))
	{
	  isText = dmFALSE;
	}

	dmDatasetClose(block->ds);
	free( dsname );
      }

      if ( isText )
      {
	/* Load contents as a filter spec. */
        dmu_err_clear();

	if (exclude)
          retval = 1;
        ptr = fptr;

	dmi_parse_external_ascii_filter( extfile, stack );

        // Code Change  Bug 9698
        //  If there is another filter clause following the ascii filter file,
        //  change the op to AndOP and push another entry for the next clause.

        if ( *ptr == ',' )
        { 
          // Increment past the comma
          ++ptr;

          // Change op to AndOP
          ((dmParseStackEntry*)(&stack->entry[(stack->group_no-1)]))->op = AndOp;

          // Change expect and next
          stack->expect[stack->group_no] = ExpectOp;
          stack->next = BeginClause;

          // create new entry for next clause
          new = dmi_parse_stack_push( stack, 0, 0, 0 );
	  dmi_parse_stack_set( stack, stack->id[(stack->group_no-1)], 0, 0, new );

          stack->id[stack->group_no] = new;
        }
        // end Code Change Bug 9698

	start = ptr;
      }
    }
    else if ( ptr&& *ptr == '(' ) 
    {
      if ( stack->next == BeginClause && dmi_parse_test_clause_open( ptr ) )
      {
	stack->clause_depth++; 
	start = ptr + 1;
	cs_advptr( &start );
	
	new = dmi_parse_stack_push( stack, 0, 0, 0 );
	if ( stack->expect[stack->group_no] == ExpectArg1 )
	{
	  dmi_parse_stack_set( stack, stack->id[stack->group_no], CopyOp, new, 0 );
	  stack->expect[stack->group_no] = ExpectOp;
	}
	else if ( stack->expect[stack->group_no] == ExpectArg2 )
	{
	  dmi_parse_stack_set( stack, stack->id[stack->group_no], 0, 0, new );    
	  stack->expect[stack->group_no] = ExpectClose;
	}
	else
	  printf( "Unexpected ( in text\n" );

	stack->group_no++;        
	stack->id[stack->group_no] = new; 
	stack->expect[stack->group_no] = ExpectArg1;
	stack->next = InClause;
	if ( dmverbose >= verb )
	  fprintf( stderr, "DM Parser: (    Begin group %ld\n", stack->group_no );

	disp = 1;
      }
      else
      {
	depth++;
	stack->next = InClause;
      }
    }
    else if ( ptr && *ptr == ')' ) 
    {
      depth--;
      if ( depth < 0 )
      {
	disp = 1;
	depth = 0;
	clause = cs_dup_siz( start, ptr-start );
	     
	if ( clause )
	{
	  arg = dmi_parse_stack_arg_push( stack, ConstOp, clause );
	}
	else
	{
	  arg = 0;
	  if ( stack->expect[stack->group_no] != ExpectClose ) 
	    fprintf( stderr, "DM Parser: Clause is empty\n");
	}
	if ( stack->expect[stack->group_no] == ExpectArg2 )
	{
	  dmi_parse_stack_set( stack, stack->id[stack->group_no], 0, 0, arg );
	}
	else if ( stack->expect[stack->group_no] == ExpectOp ) 
	{ /* Single clause in parens */
	  dmi_parse_stack_set( stack, stack->id[stack->group_no], CopyOp, arg, 0 );
	}
	else if ( stack->expect[stack->group_no] == ExpectArg1 ) 
	{
	  dmi_parse_stack_set( stack, stack->id[stack->group_no], CopyOp, arg, 0 );
	}
	else if ( stack->expect[stack->group_no] != ExpectClose ) 
	{
	  fprintf( stderr, "DM Parser: Unexpected clause %s\n", clause);
	}

	if ( stack->clause_depth == 0 && dmverbose >= 1 )
	{
	  fprintf( stderr, "DM Parser: Clause %s has too many closing parens\n", clause );
	}
	else if ( dmverbose >= verb )
	{
	  fprintf( stderr, "DM Parser: )    End group %ld\n", stack->group_no );
	  if ( clause )
	    fprintf( stderr, "DM Parser: Clause %s\n", clause );
	}

	arg = stack->id[stack->group_no];
	stack->group_no--;
	stack->clause_depth--;
	if ( stack->group_no < 0 ) 
	  fprintf( stderr, "DM Parser: Too many closes\n" );

	if ( stack->expect[stack->group_no] == ExpectArg2 )
	{
	  dmi_parse_stack_set( stack, stack->id[stack->group_no], 0, 0, arg );
	  stack->expect[stack->group_no] = ExpectClose;
	}
	else if ( stack->expect[stack->group_no] == ExpectArg1 ) 
	{
	  dmi_parse_stack_set( stack, stack->id[stack->group_no], arg, 0, 0 );
	  stack->expect[stack->group_no] = ExpectOp;
	}
	start = ptr+1;
	cs_advptr( &start );
	stack->next = EndClause;
      }
    }
    else if ( ptr && depth == 0 && dmi_parse_clause_operator( ptr ) ) 
    {
      disp = 1; 
      op = (*ptr== '&' ) ? AndOp : OrOp;
      clause = NULL; 
      arg = 0; 
      if ( stack->next == InClause && ptr > start )
      {
	clause = cs_dup_siz( start, ptr-start );
	arg = dmi_parse_stack_arg_push( stack, ConstOp, clause );       
	if ( stack->expect[stack->group_no] == ExpectArg2 ) 
	{  /* Associative without parens */
	  new = dmi_parse_stack_push( stack, 0, 0, 0 );
	  dmi_parse_stack_set( stack, stack->id[stack->group_no], 0, 0, new );

	  /* Close group and open new one */
	  stack->id[stack->group_no] = new;
	}
      } /* Else we have an operator following a close paren? */
      else if ( stack->expect[stack->group_no] != ExpectOp )     
      {
	if ( stack->expect[stack->group_no] != ExpectClose )
	  fprintf( stderr, "DM Parser error: Found operator, expected arg\n" );            

	arg = dmi_parse_stack_arg( stack, stack->id[stack->group_no] );
	new = dmi_parse_stack_push( stack, 0, 0, 0 );
	dmi_parse_stack_set( stack, stack->id[stack->group_no], 0, 0, new );
	stack->id[stack->group_no] = new;
      }
      stack->expect[stack->group_no] = ExpectArg2;

      dmi_parse_stack_set( stack, stack->id[stack->group_no], op, arg, 0 );
      if ( dmverbose >= verb )
      {
	fprintf( stderr, "DM Parser: %c%c Clause Operator\n", *ptr, *(ptr+1) );
	if ( clause )
	  fprintf( stderr, "DM Parser: Clause %s\n", clause );
      }
      stack->next = BeginClause;
      ptr++;    /* Clause operators &&, || are two bytes long */
      start = ptr+1;
      cs_advptr( &start );

    }
    else if ( ptr && *ptr ) 
    {
      stack->next = InClause;
    }

    /* We are at the beginning of the first clause and there is no paren: */
    if ( stack->group_no == 0 && stack->init )
    { 
      long top;
      disp = 1;

      /* In this group, we expect an operator next */
      top = stack->id[stack->group_no];
      stack->expect[stack->group_no] = ExpectOp;
      stack->group_no++;

      start = ptr;

      stack->next = InClause;
      stack->id[stack->group_no] = dmi_parse_stack_push( stack, 0, 0, 0 );
      dmi_parse_stack_set( stack, top, 0, stack->id[stack->group_no], 0 );
      stack->expect[stack->group_no] = ExpectOp;
      if ( dmverbose >= verb )
	fprintf( stderr, "DM Parser:      Begin group %ld\n", stack->group_no );
    }
     
     stack->init = 0;
     
     if ( !ptr || *ptr== '\0' )
     {
       disp = 1;
       loop = 0;
       if ( ptr > start )
       {
	 clause = cs_dup_siz( start, ptr-start );
	 if ( depth > 0  && dmverbose >= 1 )
	 {
	   fprintf( stderr, "DM Parser: Clause %s has too few closing parens\n", clause );
	 }
	 else 
	 {
	   arg = dmi_parse_stack_arg_push( stack, ConstOp, clause );       
	   if ( stack->expect[stack->group_no] == ExpectArg2 )
	   {
	     dmi_parse_stack_set( stack, stack->id[stack->group_no], 0, 0, arg );
	     stack->expect[stack->group_no] = ExpectClose;
	   }
	   else if ( stack->expect[stack->group_no] == ExpectOp ) 
	   {
	     dmi_parse_stack_set( stack, stack->id[stack->group_no], CopyOp, arg, 0 );
	   }
	   if ( dmverbose >= verb )
	   {
	     fprintf( stderr, "DM Parser: Clause %s\n", clause );
	   }
	   stack->next = EndClause;
	 }    
       }
     }
     else 
     {
       ptr++;
     }
     
     if ( disp && dmverbose == 5 )
     {
       char fmt[800];
       long g;
       printf( "PARSER: %s\n", buf );
       if ( ptr )
       {
	 sprintf( fmt, "PARSER: %%%lds %%s\n", (long)( ptr-buf+1 ));
	 printf( fmt, "^", nextnames[stack->next] );
       }
       for ( g = 0; g<= stack->group_no; g++ )
	 printf( "Group: %ld Entry: %ld Expect: %s\n", g, stack->id[g], expnames[(stack->expect[g])] );
       dmi_parse_state_disp( stack );
     }

  } /* end while loop */
 
  if (dmverbose == 3 )
    dmi_parse_state_disp( stack );
 
  return retval;
}

/***************************************************************/
/*                                                             */
/* Function:   dmi_parse_stack_eval                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*        dmParseStack  *stack                                 */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*        dmClauseList  *clist                                 */
/*                                                             */
/***************************************************************/
dmClauseList* dmi_parse_stack_eval( dmParseStack* stack )
{
  dmClauseList *clist  = NULL;
  long         *estack = NULL;
  dmBool       *eval   = NULL;
  long          depth  = 0;

  if ( !stack || !stack->ns ) 
    return clist;

  eval = calloc( stack->ns, UT_SZ_Q );
  estack= calloc( 2 * stack->ns, UT_SZ_I );

  estack[0] = 1;
  depth = 1;

  clist  = dmi_parse_stack_entry_eval( stack, estack, eval, &depth, estack[0] );

  free( estack );
  free( eval );

 return clist;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_parse_stack_entry_eval                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*      dmParseStack  *stack                                   */
/*      long          *estack                                  */
/*      dmBool        *eval                                    */
/*      long          *depth                                   */
/*      long           id                                      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*           dmClauseList *clist                               */
/*                                                             */
/***************************************************************/
dmClauseList* dmi_parse_stack_entry_eval( dmParseStack* stack, long* estack,
                                         dmBool* eval, long* depth, long id )
{
  dmParseStackEntry *entry  = NULL;
  dmClauseList      *clist  = NULL;
  dmClauseList      *clist1 = NULL;
  dmClauseList      *clist2 = NULL;
  long i,j,k,n;

  if ( id == 0 )
    return NULL;

  if( eval[id-1] == 1 )
    printf( "Note: Repeat exec %ld\n", id );

  estack[*depth] = id;
  (*depth)++;

  entry = &stack->entry[id-1];
  if ( entry->op != ConstOp )
  {
    clist1 = dmi_parse_stack_entry_eval( stack, estack, eval, depth, entry->arg[0] );
    clist2 = dmi_parse_stack_entry_eval( stack, estack, eval, depth, entry->arg[1] );
  } 
  eval[id-1] = 1;

  if ( entry->op == ConstOp )
  {
    clist = calloc( 1, sizeof( dmClauseList ));
 
    clist->clauses = calloc( 1, UT_SZ_P );
    clist->clauses[0] = cs_dup( entry->ptr );
    clist->n = 1;
  } 
  else if ( entry->op == OrOp )
  {
    clist = calloc( 1, sizeof( dmClauseList ));

    n = clist1->n + clist2->n;
    k = 0;
    clist->clauses = calloc( n, UT_SZ_P );
    for ( i = 0; i < clist1->n; i++ )
      clist->clauses[k++] = cs_dup( clist1->clauses[i] );
    for ( j = 0; j < clist2->n; j++ )
      clist->clauses[k++] = cs_dup( clist2->clauses[j] );
    clist->n = n;
  }
  else if ( entry->op == AndOp )
  {
    clist = calloc( 1, sizeof( dmClauseList ));

    n = clist1->n * clist2->n;
    k = 0;
    clist->clauses = calloc( n, UT_SZ_P );
    for ( i = 0; i < clist1->n; i++ )
    {
      for ( j = 0; j < clist2->n; j++ )
      {
	clist->clauses[k++] = cs_dup_cat( clist1->clauses[i], clist2->clauses[j], "," );
      }
    }
    clist->n = n;
  }
  else if ( entry->op == CopyOp )
  {
    clist = clist1;
    clist1 = (dmClauseList*)NULL;
  }
  else
  { /* Dummy case */
    clist = calloc( 1, sizeof( dmClauseList ));
  }

  dmi_clauselist_free( clist1 );
  dmi_clauselist_free( clist2 );
  (*depth)--;

  if ( dmverbose >= 6 )
  {
    printf( "Depth %ld ID = %ld\n", *depth, id );
    for ( i = 0; i < clist->n; i++ )
      printf( "Clause %4ld: %s\n", i+1, clist->clauses[i] );
  }

  return clist;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_clauselist_free                              */
/*                                                             */
/* Description:                                                */
/*       Frees clause list                                     */
/*                                                             */
/*  Input parameters:                                          */
/*      dmClauseList  *c                                       */
/*                                                             */
/*  Output parameters:                                         */
/*      none                                                   */
/*                                                             */
/*  Returns:                                                   */
/*          void                                               */
/*                                                             */
/***************************************************************/
void dmi_clauselist_free( dmClauseList *c )
{
  long ii = 0;

  if ( !c ) 
    return;

  for (  ii = 0; ii < c->n; ii++ )
    free( c->clauses[ii] );

  free( c->clauses );
  free( c );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_parse_filter_assign                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*       char    *itemp                                        */
/*       long     itemno                                       */
/*       long    *axisnop                                      */
/*       char    *name                                         */
/*       char    *op                                           */
/*       char    *ptr                                          */
/*       char    *tptr                                         */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*           char *ptr                                         */
/*                                                             */
/***************************************************************/
char* dmi_parse_filter_assign( char* itemp, long itemno, long* axisnop, 
			       char* name, char* op, char* ptr, char* tptr )
{
  char *sptr    = NULL;
  int   special = 0;

  /* Split token into  <name op spec>  */
  sptr = dmu_split_token( itemp, name, MAX_NAME_LENGTH, op );

  if ( !sptr )
  { 
    if( axisnop )
      special = dmi_parse_special_syntax( name, op, itemno, axisnop );

    if ( !special )
      ptr = tptr; 
  }
  else
    ptr +=  (sptr-itemp );

 return ptr;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_parse_special_syntax                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*   Handle special filter syntax                              */
/*   No '=' sign in the filter... could be                     */
/*   IMAGE SPECIAL SYNTAX case                                 */
/*     e.g.   image.fits[100:300,20:50]                        */
/*     e.g.   image.fits[circle(165,30,10)]                    */
/*   We check that it's an image and that all previous items   */
/*   were in the same format. Then, if it's a number followed  */
/*   by a colon, we have the special case and convert it to    */
/*   effectively                                               */
/*     image.fits[#1=100:300,#2=20:50]                         */
/*                                                             */
/*   Otherwise, we just ignore the token                       */
/*                                                             */
/*  Input parameters:                                          */
/*         char    *name                                       */
/*         char    *op                                         */
/*         long     itemno                                     */
/*                                                             */
/*  Output parameters:                                         */
/*         long    *axisnop                                    */
/*                                                             */
/*  Returns:                                                   */
/*         int  special                                        */
/*                                                             */
/***************************************************************/
int dmi_parse_special_syntax( char* name, char* op, long itemno, long* axisnop )
{
  char  *tmpval = NULL;

  double tmpdval = 0.0;
  long   axisno  = *axisnop;
  int    special = 0;
 
  if ( axisno + 1 == itemno )
  {
    if ( axisno == 0 && strchr( name, '(' ) && strchr( name, ')' ) )
    {
      sprintf( name, "(#1,#2)" );
      strcpy( op, "=" );
      special = 1;
    }
    else 
    {
      dmu_parse_token( name, ':', &tmpval, 0 );

      // Code Change  Bug 12596 (revisited)  April 15, 2011
      //  dmu_parse_token() can return NULL in result buffer
      //  Check before passing onto dmu_lex_number() which
      //  does no check before using pointer

      if ( tmpval )
      {
        if ( dmu_lex_number( tmpval, &tmpdval ))
        {
	  free( tmpval );
	  sprintf( name, "#%ld", ++axisno );
	  *axisnop = axisno;
	  strcpy( op, "=" );
	  special = 1;
        }
      }
    }
  }

  return special;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_subspace_search_and_init                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*       dmBlock    *block                                     */
/*       dmSubspace *rts                                       */
/*       char       *name                                      */
/*       long       *typ                                       */
/*                                                             */
/*  Output parameters:                                         */
/*       long       *vcptp                                     */
/*       dmFilter  **filterp                                   */
/*                                                             */
/*  Returns:                                                   */
/*          dmDescriptor *sub_vector                           */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_subspace_search_and_init( dmBlock* block, dmSubspace* rts, 
					    char* name, long* vcptp, long* typ, dmFilter** filterp )
{
  dmDescriptor *sub_dd     = NULL;
  dmDescriptor *sub_vector = NULL;

  long dim  = 0;
  long vcpt = 0;

  dmDataType type;

  int  err_fail = 1;

  sub_dd = dmi_subspace_search( block, rts, name, typ, err_fail );

  if ( !sub_dd ) /* Error message already set */
    return NULL;

  vcpt = sub_dd->cptno;

  if ( vcpt > 0 )
    sub_vector = sub_dd->parent;
  else
    sub_vector = sub_dd;
    
  dim = sub_vector->dim;
  type = dmi_filter_type( sub_dd->type );

  *filterp = dmi_filter_init( NULL, dim, type );
  *vcptp   = vcpt;

  return sub_vector;
}

/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_apply_filter                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Finding the parent column of a subspace descriptor:        */
/*  If the subdd is actually a vector component, need to find  */
/*  the vector parents first:                                  */
/*                                                             */
/*         vector_sub_dd ----->   vector_dd                    */
/*            |          parent       |        ^               */
/*          ------                  ----       | parent        */
/*          |    |                 |    |      |               */
/*      cpt_sub_dds                cpt_dds                     */
/*                                                             */
/*                                                             */
/*  Apply temporary parsed filter to the descriptor.           */
/*  First, intersect its range with the default range.         */
/*  Then intersect it with the current runtime subspace filter */
/*  on this descriptor.                                        */
/*  Make the result the new runtime subspace filter.           */
/*  Finally, free the temporary filter.                        */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*       dmBlock      *block                                   */
/*       dmSubspace   *rts                                     */
/*       dmDescriptor *sub_vector                              */
/*       long          vcpt                                    */
/*       long          ty                                      */
/*       dmFilter     *filter                                  */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*           dmBool  UT_TRUE or UT_FALSE                       */
/*                                                             */
/***************************************************************/
dmBool dmi_subspace_col_apply_filter( dmBlock* block, dmSubspace* rts, 
				      dmDescriptor* sub_vector, long vcpt, 
				      long ty, int intersect, dmFilter* filter )
{
  dmFilter     *rtfilter = NULL;
  dmDescriptor *var_dd   = NULL;  /* Column descriptor corresponding to subspace descriptor */

  var_dd = dmi_descriptor_get_cpt( sub_vector->parent, vcpt );

  if ( !dmi_subspace_col_map_range( block, var_dd, rts, filter, ty ))
  {
    /* error already set */
    return UT_FALSE;
  }

  rtfilter = dmi_subspace_col_get_filter( sub_vector, rts->cpt );
  rtfilter = dmi_filter_intersect( rtfilter, filter, intersect, vcpt, sub_vector->dim );
  dmi_subspace_col_set_filter( sub_vector, rts->cpt, rtfilter );   

  dmi_filter_free( filter );

  return UT_TRUE;
}

/***************************************************************/
/*                                                             */
/* Function:    dmi_parse_external_ascii_filter                */
/*                                                             */
/* Description:                                                */
/*        Read filter lines from text file                     */
/*                                                             */
/*  Input parameters:                                          */
/*         char         *filename - text filter file           */
/*         dmParseStack *state    - stack state                */ 
/*                                                             */
/*  Output parameters:                                         */
/*         none                                                */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_parse_external_ascii_filter( char* filename, dmParseStack* state )
{
  FioFile fptr;

  char  line[dmMAXSTR] = {dmMAXSTR * '\0' };
  char *buf = NULL;
  char *ptr = line;

  fptr = fio_open_ar( filename );

  if ( !fptr )
  {
    /* Error, couldn't open as text! */

    sprintf( dm_errbuf, "Failed to open external filter file %s", filename );
    dmu_err_set(DM_ERR_NO_FILE,dm_errbuf );
    return;
  }


  while( fio_read_line( fptr, ptr, dmMAXSTR ))
  {
    if ( dmverbose >= 4 )
      fprintf( stderr, "Parsing external filter line %s\n", ptr );

    if ( buf )
    {
      buf = realloc( buf, strlen( buf ) + strlen( ptr ) + 1 );
      strcat( buf, ptr );
    }   
    else
      buf = cs_dup( ptr );
  
  }

  /* Code Change  Bug 12762  March 30, 2011            */
  /* If fio_read_line() fails during the first read    */
  /* then "buf" will be NULL. Don't send NULL ptr onto */
  /* other functions. Set error and return.            */
  /* Check for valid buffer pointer before use         */

  if ( buf )
  {
    dmi_parse_clause_line( buf, state );
    free( buf );
  }
  else
  {
    /* Error return from fio_read_line() */
    /* Set error message                 */

    sprintf( dm_errbuf, "Empty external filter file %s\n", filename );
    dmu_err_set(DM_WARN_PARSE_OPTION, dm_errbuf ); 
  }

  fio_file_close(fptr);
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_subspace_search                              */
/*                                                             */
/* Description:                                                */
/*   Interpret a filter name and return the filter.            */
/*   If it's not already in the list, add it to the list.      */
/*   If it's a component, return a component of the            */
/*   subspace col                                              */
/*                                                             */
/*  Input parameters:                                          */
/*      dmBlock    *block                                      */
/*      dmSubspace *rts                                        */
/*      char       *name                                       */
/*      int         q                                          */
/*                                                             */
/*  Output parameters:                                         */
/*      long       *ty                                         */
/*                                                             */
/*  Returns:                                                   */
/*       dmDescriptor  *sub_dd                                 */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_subspace_search( dmBlock* block, dmSubspace* rts, char* name, long* ty, int q )
{
  dmDescriptor *dd     = NULL;
  dmDescriptor *cpt_dd = NULL;
  dmDescriptor *sub_dd = NULL;

  long cptno = 0;
  long ii    = 0;

  if ( !rts  )
    return NULL;

  cpt_dd = dmi_block_search_raw_col_identifier( block, name );

  if ( !cpt_dd )
  {
    if ( q )
    {
      sprintf( dm_errbuf, "Error: Could not find identifier %s.", name );
      dmu_err_set( DM_ERR_PARSE, dm_errbuf );
    }
    return NULL;
  }

  if ( cpt_dd->flag & DM_FLAG_LOGICAL )
  {
    *ty = DM_FLAG_LOGICAL;
    cpt_dd = dmi_descriptor_get_coord_no( cpt_dd, 1 );
  }
  else
    *ty = DM_FLAG_PHYS;

#if DEBUG
printf( "DMPINTERP %s\n", dmi_descriptor_get_qualname( cpt_dd,dm_errbuf ));
#endif

  /* Force vector */
  dd = cpt_dd;
  cptno = dd->cptno;

  if ( cptno > 0 )
    dd    = dd->parent;

  /* Match on parent name in subspace */
  for ( ii = 0; ii < rts->ncols; ii++ )
  {
    sub_dd = rts->cols[ii];

    if ( cs_uceq( sub_dd->name, dd->name ) )
    {
      /* resolve to component */
      if ( cptno > 0 && sub_dd->cpts )
	sub_dd = dmi_descriptor_get_cpt( sub_dd, cptno );

      return sub_dd;
    }
  }

  /* No subspace descriptor? make one... */
  sub_dd = dmi_subspace_col_create( block, rts, dd->name, dd->cptNames, 
	       dd->dim, dd->type, dd->unit, dd, UT_FALSE );

  if ( cptno > 0 && sub_dd && sub_dd->cpts )
    sub_dd = dmi_descriptor_get_cpt( sub_dd, cptno );

 return sub_dd;
}
