/*                                                                
**  Copyright (C) 2002-2007,2010  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmdsspriv.c                                                 */
/*                                                               */
/*   Functions that work on data subspaces                       */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"
#define MAXCPTS 10

int dmi_filter_compare_tab( dmSubspaceData* ssdata, long ncpts, dmFilter* filter, dmFilter* altfilter, long cptno );

double dmi_array_get_max( double* data, long n );
double dmi_array_get_min( double* data, long n );

char*     dmMaskName(void);                                      /* MCD TEMP */

/***************************************************************/
/*                                                             */
/* Function:  dmi_subspace_col_init                            */
/*                                                             */
/* Description:                                                */
/*   Subspace column initialization                            */
/*                                                             */
/* Input parameters:                                           */
/*   dmBlock      *block         - block to hold subspace col. */
/*   char         *total_name    - full name of column/vector  */
/*   dmDescriptor *col_dd        - coord descriptor            */
/*   dmDataType    type          - datatype of subspace col.   */
/*                                                             */
/* Output parameters:                                          */
/*   none                                                      */
/*                                                             */
/* Returns:                                                    */
/*   dmDescriptor *ssd           - subspace descriptor         */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_subspace_col_init( dmBlock* block, const char* total_name, dmDescriptor* col_dd, dmDataType type )
{
  int image = 0;
  char* name = (char*)total_name;
  char** colnames = NULL;
  char** cptnames = NULL;
  char* unit = NULL;
  long ii;
  long dim = 1;
  dmDescriptor* ssd = NULL; /* The returned descriptor */

  const char* ptr = NULL;
  long namelen = 0;

  if ( !block )
    return NULL;

  if ( col_dd && block->type == dmIMAGE && col_dd->descType == dmCOORD )
    image = 1;

  /* If total_name is provided, use it. It may be in the form  FOO(C1,C2)
   * in which case we interpret it as a vector name with components.
   * Otherwise we use the name of the col_dd if it is provided.
   */
  if ( name && strrchr( name, '(' ) )
  {
    /* MCD NOTE: 20101116  Bug_#12608
     *   cs_parse_paren_list does not handle implicit vector names
     *   [e.g. "(tg_mlam,pi)(tg_mlam,pi)"] as needed by this code segment.
     *   so we now send in just the cpt list and assign the name ourselves.
     */
    ptr = strrchr( name, '(' );
    namelen = (long)(ptr - name);

    cptnames = cs_parse_paren_list( ptr, &dim );

    cptnames[0] = cs_dup_siz( total_name, namelen );  /* add to this list so it gets freed */

    name = cptnames[0];
    colnames = cptnames + 1;
  }
  if ( !name && col_dd )
  {
    name = col_dd->name;
  }
  if ( col_dd )
  {
    unit = col_dd->unit;
    type = col_dd->type;
  }
  if ( !name )
  {
    if ( cptnames )
      ar_free_cp( cptnames, dim+1 );
    return NULL;
  }

  /* See if it is already open */
  ssd = dmi_subspace_col_open( block->subspace, name );

  if ( ssd )
  {
    if ( cptnames )
      ar_free_cp( cptnames, dim+1 );  
    return ssd;
  }

  /* Nope - better create it */

  /* If we came in with no name, we need to deal with other stuff too from coldd */
  if ( !total_name && col_dd )
  {
    if ( cptnames )
    {
      ar_free_cp( cptnames, dim+1 );   
      cptnames = NULL;
      colnames = NULL;
    }
    dim = col_dd->dim;
    if ( dim > 1 ) 
    {
      cptnames = calloc( dim+1, UT_SZ_P );
      cptnames[0] = cs_dup( name );
      for ( ii = 1; ii <= dim; ii++ )
      {
	if ( col_dd->cptNames )
	  cptnames[ii] = cs_dup( col_dd->cptNames[ii-1] );
      }
      colnames = cptnames + 1;
    }
  }

  ssd = dmi_subspace_col_create( block, block->subspace, name, colnames, dim, 
				type, NULL, col_dd, UT_TRUE );
  if ( cptnames )
    ar_free_cp( cptnames, dim+1 );

  return ssd;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_subspace_col_create                          */
/*                                                             */
/* Description:                                                */
/*   Called from col_init                                      */
/*    - have already checked name is non-null                  */
/*    - subspace may be null.. if so, gets from block          */
/*                                                             */
/*   Default filter tells dmi_subspace_col_create whether or   */
/*   not to create the subspace column with the default        */
/*   filter attached or an empty one.                          */
/*                                                             */
/*   Ryan, I think using default filter makes it so we         */
/*   don't need the ignore_time_range_variable.                */
/*                                                             */
/* Input parameters:                                           */
/*   dmBlock*      block         - block to hold subspace col. */
/*   dmSubspace*   subspacep     - pointer to subspace         */
/*   char*         name          - name of subspace col/vector */
/*   char**        cptnames      - vector component names      */
/*   long          dim           - ??vector dim??              */
/*   dmDataType    type          - datatype of subspace col    */
/*   char*         unitp         - unit string                 */
/*   dmDescriptor* col_dd        - associated coord descriptor */
/*   dmBool        default_filter- flag                        */
/*                                                             */
/* Output parameters:                                          */
/*   none                                                      */
/*                                                             */
/* Returns:                                                    */
/*   dmDescriptor* ssd           - subspace descriptor         */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_subspace_col_create( dmBlock* block, dmSubspace* subspacep,
   char* name, char** cptnames, long dim, dmDataType type, char* unitp, dmDescriptor* col_dd, dmBool default_filter )
{
 double dmin[MAXCPTS];
 double dmax[MAXCPTS];
 void* vmin[MAXCPTS];
 void* vmax[MAXCPTS];
 long length[MAXCPTS];
 dmSubspace* subspace = subspacep;
 dmDescriptor* cpt;
 dmDescriptor* dd_cpt;
 dmDescriptor* ssd = NULL;
 dmFilter* filter;
 char* unit = unitp;
 char* desc = NULL;
 long nvalues = 1;
 long i;
 long cptNo;
 long cptdim = 1;
 long naxes = 1;
 long voff;
 long no = 0;
 int ignore_time_range = 0;
 int numeric;

 if ( !block) 
  return NULL;

 numeric = ar_mix_is_numeric( type );
 memset( dmin, 0, UT_SZ_D * MAXCPTS );
 memset( dmax, 0, UT_SZ_D * MAXCPTS );
/* Normally use the default subspace */
 if ( !subspace )
  subspace = block->subspace;

/* If null primary, there ay be no subspace */
 if ( !subspace )
  return NULL;

/* get default ranges */  
 if ( col_dd )
 {
  if ( !name ) name = col_dd->name;
  if ( !cptnames ) cptnames = col_dd->cptNames;
/* 
   Force subspace data type to double for time, even if time column is integer.
   Required for GTI conventions.
 */
  if ( cs_uceq( name, "TIME" ) )
  {
   type = DT_REAL8;
   ignore_time_range = 1;
  }
  else
   type = col_dd->type;
  if ( !unit )
   unit = col_dd->unit;
  desc = col_dd->desc;
  no = col_dd->no;
 }
/* Make name (with cpts in parens for dim>1) */

 ssd = dmi_descriptor_alloc( block, dmSUBSPACE, no, name, type, 
       unit, desc, dim, dmRANGE, naxes, NULL, DM_FLAG_SUBSPACE ); 
#if DEBUG
printf( "CREATE SS COL %s = %s DIM %ld %ld VD %ld\n", name, tmpname,dim, ssd->dim, ssd->vdim );
#endif
 ssd->parent = col_dd;


/* Components: use parent column if available else use provided subspace info */
 if ( dim > 1 ) 
 {
   for ( i = 0; i < dim; i++ )
   {
    cptdim = 1;
    if ( col_dd )
    {
     cpt = col_dd->cpts[i];
     ssd->cptNames[i] = cs_dup( cpt->name );
    } else {
     cpt = ssd;
     ssd->cptNames[i] = cs_dup( cptnames[i] );
    }
    dd_cpt = dmi_descriptor_alloc( block, dmSUBSPACE, 0, ssd->cptNames[i], cpt->type,
       cpt->unit, cpt->desc, cptdim, dmRANGE, naxes, NULL, DM_FLAG_SUBSPACE | DM_FLAG_CPT );
    dd_cpt->parent = ssd;
    dd_cpt->cptno = i+1;
    ssd->cpts[i] = dd_cpt;
   }
 }


 dmi_subspace_add_col( subspace, ssd ); 

/* When initializing the normal subspace, use the TSTART:TSTOP as before.
 * When using a run time filter, don't.
 * It would be better to add an explicit argument to this routine but
 * this avoids altering the .so calling tree for the current (3.0.1) patch
 */    
 if ( subspace == block->subspace )
  ignore_time_range = 0;

 if ( default_filter )
 {
  for ( voff = 0; voff < dim; voff++ )
  {
   dmin[voff] = 0.0;
   dmax[voff] = 0.0;
   length[voff] = 0;
   vmin[voff] = &dmin[voff];
   vmax[voff] = &dmax[voff];
  }
  if ( col_dd && numeric && !ignore_time_range )
  {
   for ( voff = 0; voff < dim; voff++ )
   {
    if ( dmi_descriptor_get_cpt_range( col_dd, voff+1, DT_REAL8, vmin[voff], vmax[voff] ))
     length[voff] = 1;
   }
  }
 }
/* Create filters initialized to TLMIN/TLMAX */


 for( cptNo=1; cptNo <= subspace->ncpts; cptNo++) 
 {
  int update = 0;
  int remove = 0;
  int dirty = 0;
  filter = dmi_subspace_col_reset_filter( ssd, cptNo );
  if ( default_filter )
  {
   for ( voff = 0; voff < dim; voff++ )
   {
    if ( length[voff] && numeric )
    {
     dmi_filter_update( filter, voff+1, NULL, ssd->name, DT_REAL8, &dmin[voff], &dmax[voff], nvalues, 0, update, &remove, &dirty );
    }
   }
  }
 }     
 return ssd;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_subspace_col_set_values                      */
/*                                                             */
/* Description:                                                */
/*   put values into subspace data                             */
/*                                                             */
/*  Input parameters:                                          */
/*                dmDescriptor *dd                             */
/*                char         *tabname                        */
/*                void         *value1                         */
/*                void         *value2                         */
/*                long          nvalues                        */
/*                dmDataType    type                           */
/*                long          dim                            */
/*                int           array                          */
/*                regRegion    *region                         */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*              void                                           */
/*                                                             */
/***************************************************************/
void dmi_subspace_col_set_values( dmDescriptor* dd, char* tabname, 
   void* value1, void* value2, long nvalues, dmDataType type, long dim, int array, regRegion* region )
{
 dmBlock* block;
 dmSubspace* subspace;
 long vcpt = 1;
 void* v1 = value1;
 void* v2 = value2;
 dmFilter* filter;
 
 if ( !dd ) return;
 block = (dmBlock*)dd->block;
 subspace = (dmSubspace*)block->subspace;
 if ( !subspace ) return;

 filter = dmi_subspace_col_reset_filter( dd, subspace->cpt );
 if ( tabname && cs_uceq( tabname, "NULL" ))
 {
   dmi_filter_set_null( filter );
 } else {
   for ( vcpt = 1; vcpt <= dim; vcpt++ )
   {
    if ( array )  /* In array mode, the v1 and v2 arrays carry one entry for each axis */
    {
      v1 = ar_gen_ptr( value1, type, 0, vcpt-1 );
      v2 = ar_gen_ptr( value2, type, 0, vcpt-1 );
    }      
    dmi_filter_setup( dd, filter, vcpt, type, v1, v2, nvalues, tabname, region );
    dmi_descriptor_set_update( dd );
   }
 } 

}


/***************************************************************/
/*                                                             */
/* Function:  dmi_subspace_create_cpt                          */
/*                                                             */
/* Description:                                                */
/*   create subspace components                                */
/*                                                             */
/*  Input parameters:                                          */
/*                dmSubspace   *subspace                       */
/*                int           empty                          */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*              int  dmSUCCESS                                 */
/*                                                             */
/***************************************************************/
int dmi_subspace_create_cpt(dmSubspace* subspace, int empty )
{
  dmDescriptor* dd = NULL;
  long           i = 0;

  dmFilter* previous_filter;
  dmFilter* filter;
  long cpt;

  /* If there aren't any subspace columns open, just keep count of # of 
   * components */
  dmi_subspace_add_cpt( subspace );
  cpt = subspace->ncpts;
  /* If there are open subspace columns, we must fill them in with another
   * row (component) with default (unlimited) range */
  /* !!! If a subspace column is associated, the new component should probably
   * have a range set to the default column range values (TLMIN, TLMAX) 
   * as happens on open */

  for(i=0; i< subspace->ncols; i++)
  {
    dd = subspace->cols[i];
    filter   = dmi_subspace_col_reset_filter( dd, cpt );
    if ( !empty ) 
    {
     previous_filter = dmi_subspace_col_get_filter( dd, cpt-1 );
     dmi_filter_copy( previous_filter, filter );
     if ( !dmi_filter_is_implicit( filter ) ) dmi_descriptor_set_update(dd);
    }

  }
  
  subspace->cpt = subspace->ncpts;

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_subspace_col_check_tabnames                  */
/*                                                             */
/* Description:                                                */
/*   For subspace 'BLOCKS'.. checks the names are distinct.    */
/*   Primarily for GTIs, to ensure a unique block for each     */
/*   subspace component.                                       */
/*                                                             */
/*   For MASK blocks.. do not want to do this. There is only   */
/*   one MASK block, and we may have multiple references to it */
/*                                                             */
/*  Input parameters:                                          */
/*   dmDescriptor *dd                                          */
/*   long          ncpts - number of components                */
/*                                                             */
/*  Output parameters:                                         */
/*   none                                                      */
/*                                                             */
/*  Returns:                                                   */
/*   void                                                      */
/*                                                             */
/***************************************************************/
void dmi_subspace_col_check_tabnames( dmDescriptor* dd, long ncpts )
{
  dmSubspaceData* ssdata;
  long cptno;
  long altcpt;
  dmFilter* filter;
  int loop = 1;
  
  char* mask_name = dmMaskName();
  
  ssdata = dmi_ss_get_data( dd );
  for ( cptno = 2; cptno <= ncpts; cptno++ )
  {
    filter = ssdata->filters[cptno-1];
    if ( filter->name && !cs_uceq_siz(filter->name, mask_name, strlen(mask_name)) )
    {
      loop = 1;
      for( altcpt = 1; altcpt < cptno && loop; altcpt++ )
        loop = dmi_filter_compare_tab( ssdata, ncpts, filter, ssdata->filters[altcpt-1], cptno );
    }
  }
  if ( mask_name )
    free( mask_name );

}


/***************************************************************/
/*                                                             */
/* Function:  dmi_filter_compare_tab                           */
/*                                                             */
/* Description:                                                */
/*   Compares name of filter and alternate filter, and renames */
/*   the filter if there is a conflict (they match).           */
/*                                                             */
/*  Input parameters:                                          */
/*    dmSubspaceData *ssdata                                   */
/*    long            ncpts - number of components             */
/*    dmFilter       *filter                                   */
/*    dmFilter       *altfilter                                */
/*    long            cptno                                    */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*    1  = No conflict with the current table                  */
/*    0  = name changed due to conflict.                       */
/*                                                             */
/***************************************************************/
int dmi_filter_compare_tab( dmSubspaceData* ssdata, long ncpts, dmFilter* filter, dmFilter* altfilter, long cptno )
{
  int match;
  char* name;

  if ( altfilter->name && !strcmp( altfilter->name, filter->name ))
  {
     /* Has the same name! */
     match = dmi_filter_compare( altfilter, filter );
     if ( !match )
     {
       /* We must rename the cpt */
       name = dmi_filter_rename_tab( filter->name, ssdata, ncpts, cptno );
       free( filter->name );
       filter->name = name;
       return 0;
      }
  }
  return 1;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_filter_rename_tab                            */
/*                                                             */
/* Description:                                                */
/*     We create a new unique name (at least unique for this   */
/*     descriptor) for the table.  We check against current    */
/*     names since there's a chance names have recently been   */
/*     changed                                                 */
/*                                                             */
/*  Input parameters:                                          */
/*                char           *old_name                     */
/*                dmSubspaceData *ssdata                       */
/*                long            ncpts                        */
/*                long            cptno                        */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*              char  *name - new name                         */
/*                                                             */
/***************************************************************/
char* dmi_filter_rename_tab( char* old_name, dmSubspaceData* ssdata, long ncpts, long cptno )
{
 long spare = 20;  
 int  loop = 1;
 char*  name;
 long maxcpt;
 long matchcpt;
 dmFilter* matchfilter;

 name = malloc( strlen( old_name ) + spare );
        maxcpt = cptno;
        while( loop )
        {
          loop= 0;
          sprintf( name, "%s_CPT%ld", old_name, maxcpt );
          for ( matchcpt = 1; matchcpt <= ncpts; matchcpt++ )
          {
           matchfilter = ssdata->filters[matchcpt-1];
           if ( matchfilter && matchfilter->name && !strcmp( name, matchfilter->name )) 
           {
            loop = 1;
            maxcpt++;
           }
          }
         }
 return name;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_array_get_min                                */
/*                                                             */
/* Description:                                                */
/*     Return smallest value in array data                     */
/*                                                             */
/*  Input parameters:                                          */
/*                double    *data  - data array                */
/*                long       n - number of data elements       */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*              double  min  - smallest value in array data    */
/*                                                             */
/***************************************************************/
double dmi_array_get_min( double* data, long n )
{
 double x;
 long i;
 if ( n == 0 ) return 0.0;
 x = data[0];
 for ( i = 1; i < n; i++ )
  if ( data[i] < x ) x = data[i];
 return x;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_array_get_max                                */
/*                                                             */
/* Description:                                                */
/*     Return largest value in array data                      */
/*                                                             */
/*  Input parameters:                                          */
/*                double    *data  - data array                */
/*                long       n - number of data elements       */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*              double  max  - greatest value in array data    */
/*                                                             */
/***************************************************************/
double dmi_array_get_max( double* data, long n )
{
 double x;
 long i;
 if ( n == 0 ) return 0.0;
 x = data[0];
 for ( i = 1; i < n; i++ )
  if ( data[i] > x ) x = data[i];
 return x;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_filter_simple_range                          */
/*                                                             */
/* Description:                                                */
/*     Return range values on real time subapace descriptor    */
/*                                                             */
/*  Input parameters:                                          */
/*                dmDeacriptor   *rts_dd                       */
/*                long            vcpt                         */
/*                                                             */
/*  Output parameters:                                         */
/*                 double        *vals                         */
/*                                                             */
/*  Returns:                                                   */
/*              int UT_TRUE or UT_FALSE                        */
/*                                                             */
/***************************************************************/
int dmi_filter_get_simple_range( dmDescriptor* rts_dd, long vcpt, double* vals )
{
  long  rtscpt = 1;
  dmDataType type;
  long dim;
  double* lc_mins = NULL;
  double* lc_maxes=NULL; 
  long nf = 0;
  long n = 0;
  dmRangeList* range = NULL;
  dmFilter* filter = dmi_subspace_col_get_filter( rts_dd, rtscpt );

  if ( !filter )
   return 0;
/* If complicated, bail! */
  type = dmi_filter_type( filter->type );
  if ( type != DT_REAL8 )
   return 0;
  dim = filter->dim;
  if ( !filter->null )
     {  
      range = &filter->ranges[vcpt-1];
      n = dmi_rangelist_get_size( range );
     }
    nf = UT_FALSE; 
    if (  filter->null || (range && n > 0) )
     {
      if ( range )
      {
        lc_mins = range->mins;
        lc_maxes = range->maxes;
      } else {
	lc_mins = NULL;
	lc_maxes = NULL;
      }
      if ( lc_mins )
      {
       vals[0] = dmi_array_get_min( lc_mins, n );
       vals[1] = dmi_array_get_max( lc_maxes, n );
       nf = UT_TRUE;
      }
     }
  return nf;  
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_parse_filter_string                          */
/*                                                             */
/* Description:                                                */
/*   Parse filter string, set elements                         */
/*                                                             */
/*   Expectation is that 'dsval; holds a rangelist or region   */
/*   (shape) string.                                           */
/*                                                             */
/* Input parameters:                                           */
/*   char           *dsval      - subspace value string        */
/*   dmDescriptor   *dd         - subspace col descr.          */
/*                                                             */
/* Output parameters:                                          */
/*   void          **value1     - rangelist mins               */
/*   void          **value2     - rangelist maxs               */
/*   long           *nvalsp     - number of values in arrays   */
/*   regRegion     **regRegion  - region struct                */
/*   int            *pincnull                                  */  
/*                                                             */
/* Returns:                                                    */
/*   int UT_TRUE                                               */
/*       UT_FALSE                                              */
/*                                                             */
/***************************************************************/
int dmi_parse_filter_string( char* dsval, dmDescriptor* dd, void** value1, void** value2,
			       long* nvalsp, regRegion** regionp, int* pincnull )
{

  double mins[2];                 /* used in multi-dim ranges  */
  double maxes[2];                /* used in multi-dim ranges  */
  double xvals[2];                /* x-axis range values from rectangle region */
  double yvals[2];                /* y-axis range values from rectangle region */

  regRegion* local_region = NULL; /* Hold region filter                          */

  char*      ptr;                 /* local pointer into value string             */
  char*      name;                /* name of descriptor                          */
  long       dim;                 /* descriptor dimensionality                   */
  dmDataType type;                /* descriptor datatype                         */
  dmDataType ftype;               /* filter datatype                             */
  dmFilter*  filter = NULL;       /* Temp filter struct to hold parsed rangelist */

  int null = UT_FALSE;            /* Flag empty/NULL filters   */
  int loop = UT_TRUE;
  int incnull;

  int ii;
  int size;

  *nvalsp = 0;

  /* Get data type and dim of descriptor */
  type = dmi_descriptor_get_datatype( dd );
  dim  = dmi_descriptor_get_element_dim( dd );

  if ( strchr( dsval, '(' ))
  {
    /* Filter is some form of region */
    local_region = regParse( dsval );
  } 
  else if ( cs_uceq( dsval, "NULL" ) || cs_uceq( dsval, "EMPTY" ))
  {
    /* Filter is empty */
    null = UT_TRUE;
  }
  else 
  {
    /* Filter is a single rangelist */
    void *tmp1 = NULL;
    void *tmp2 = NULL;

    ptr = dsval;
    name = dmi_descriptor_get_name( dd ); 
    ftype = dmi_filter_type( type );
    filter =  dmi_filter_init( NULL, dim, ftype );
    dmi_filter_parse_typed_rangelist( name, type, 1, dim, &ptr, filter, "=", loop, &incnull );

    /* Pull range values out of filter */
    dmi_filter_extract( filter, 1, type, &tmp1, &tmp2, nvalsp, pincnull );

    if ( dim > 1 )
    {
      /* got a single range for a multi-D descriptor..   */
      /* Use same range on all dims, only 1 set allowed. */

      size = ar_mix_type_size( type );
      *value1 = (void*)calloc( dim, size );
      *value2 = (void*)calloc( dim, size );

      for ( ii = 0; ii < dim; ii++ )
      {
	ar_mix_numeric_cast_elt( tmp1, type, type, *value1, 0, ii );
	ar_mix_numeric_cast_elt( tmp2, type, type, *value2, 0, ii );
      }

      /* indicate one set of ranges */
      *nvalsp = 1;

      free(tmp1);
      free(tmp2);

    }
    else
    {
      *value1 = tmp1;
      *value2 = tmp2;
    }

    dmi_filter_free( filter );
  }

  /* Handle rectangle regions specially.. convert into ranges on component axes */
  /* NOTE: allows only 1 range set                                              */
  if ( dmi_region_is_rectangle( local_region, xvals, yvals ))
  {
    mins[0]  = xvals[0];
    mins[1]  = yvals[0];
    maxes[0] = xvals[1];
    maxes[1] = yvals[1];

    *nvalsp = 1;  /* one set of ranges */
    dmi_range_pair_copy( mins, maxes, 2, type, value1, value2 );

    regFree( local_region );
    local_region = NULL;
  }
 
 *regionp = local_region;
 return null;
}



/*
 *  Update current component of subspace col filter
 */
int  dmi_subspace_col_update_cpt(dmDescriptor* dd, char* tname, long cptNo, dmDataType type, void* value1,
				   void* value2, long nvalues, int incnull, dmBool update )
{
  dmDescriptor* filter_dd = dd;
  dmFilter* filter = NULL;        /* Value group for new values            */
  int retval = dmSUCCESS;
  int remove = 0;
  int dirty = 0;
  int null = 0;
  char* name = tname;


  null  = ( name && cs_uceq( name, "NULL" ));

  if ( nvalues == 0 && update && !null && !incnull )   /* Update with no values */
   return dmSUCCESS;

  if(!filter_dd)
  {
      dmu_err_set( DM_ERR_NULL, "No filter descriptor" );
      return dmFAILURE;
  }



  if(filter_dd->descType != dmSUBSPACE)
  {
      dmu_err_set( DM_ERR_GENERIC, "Wrong type of descriptor" );
      return dmFAILURE;
  }


  /* Check if we have a component of a vector column */
  if ( filter_dd->cptno > 0 )
  {
   cptNo     = filter_dd->cptno;   
   filter_dd = filter_dd->parent;  
  }  

  if ( !type )
   printf( "DM Internal error: Subspace update type unknown for %s\n", filter_dd->name );

  /* Find the value group which corresponds to the current component */
  filter = dmi_subspace_col_get_current_filter( filter_dd );


  retval = dmi_filter_update( filter, cptNo, name, filter_dd->name, type, value1, value2, nvalues, incnull,
                            update, &remove, &dirty );


   if ( dirty )
      dmi_descriptor_set_update(filter_dd);

  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_subspace_col_init_cpt                        */
/*                                                             */
/* Description:                                                */
/*    Initialize current component of subspace col filter.     */
/*    This is the same as dmi_subspace_col_update_cpt() except */
/*    the descriptor won't be marked for update for next block */
/*    flush.  It is intended for use by the kernels when       */
/*    reading the subspace from the file.                      */
/*                                                             */
/*  Input parameters:                                          */
/*                dmDeacriptor   *dd                           */
/*                char           *tname                        */
/*                long            cptNo                        */
/*                dmDataType      type                         */
/*                void           *value1                       */
/*                void           *value1                       */
/*                long            nvalues                      */
/*                                                             */
/*  Output parameters:                                         */
/*                   none                                      */
/*                                                             */
/*  Returns:                                                   */
/*              int retval - dmSUCCESS or dmFAILURE            */
/*                                                             */
/***************************************************************/
int  dmi_subspace_col_init_cpt(dmDescriptor* dd, char* tname, long cptNo, dmDataType type, void* value1,
				   void* value2, long nvalues )
{
  dmDescriptor* filter_dd = dd;
  dmFilter* filter = NULL;        /* Value group for new values            */
  int retval = dmSUCCESS;
  int remove = 0;
  int dirty = 0;
  int null = 0;
  dmBool update = UT_FALSE;
  char* name = tname;


  null  = ( name && cs_uceq( name, "NULL" ));

  if ( nvalues == 0 && update && !null )   /* Update with no values */
   return dmSUCCESS;

  if(!filter_dd)
  {
      dmu_err_set( DM_ERR_NULL, "No filter descriptor" );
      return dmFAILURE;
  }



  if(filter_dd->descType != dmSUBSPACE)
  {
      dmu_err_set( DM_ERR_GENERIC, "Wrong type of descriptor" );
      return dmFAILURE;
  }


  /* Check if we have a component of a vector column */
  if ( filter_dd->cptno > 0 )
  {
   cptNo     = filter_dd->cptno;   
   filter_dd = filter_dd->parent;  
  }  

  if ( !type )
   printf( "DM Internal error: Subspace update type unknown for %s\n", filter_dd->name );

  /* Find the value group which corresponds to the current component */
  filter = dmi_subspace_col_get_current_filter( filter_dd );


  retval = dmi_filter_update( filter, cptNo, name, filter_dd->name, type, value1, value2, nvalues, 0,
                            update, &remove, &dirty );

  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_subspace_col_init_incnull                    */
/*                                                             */
/* Description:                                                */
/*     Initialize current component of subspace col filter.    */
/*     This is the same as dmi_subspace_col_update_cpt() except*/
/*     the descriptor won't be marked for update for next block*/
/*     flush.  It is intended for use by the kernels when      */
/*     reading the subspace from the file.                     */
/*                                                             */
/*     This specifically initializes the filter to explicitly  */
/*     include or exclude NULL values.                         */
/*                                                             */
/*  Input parameters:                                          */
/*                dmDeacriptor   *dd                           */
/*                char           *tname                        */
/*                long            cptNo                        */
/*                dmDataType      type                         */
/*                int             incnull                      */
/*                                                             */
/*  Output parameters:                                         */
/*                   none                                      */
/*                                                             */
/*  Returns:                                                   */
/*              int retval - dmSUCCESS or dmFAILURE            */
/*                                                             */
/***************************************************************/
int  dmi_subspace_col_init_cpt_incnull(dmDescriptor* dd, char* tname, long cptNo, dmDataType type, int incnull )
{
  dmDescriptor* filter_dd = dd;
  dmFilter* filter = NULL;        /* Value group for new values            */
  char* name = tname;
  int retval = dmSUCCESS;
  int remove = 0;
  int dirty = 0;
  dmBool update = UT_FALSE;

  if ( incnull == 0 )   /* Update with no values */
   return dmSUCCESS;

  if(!filter_dd)
  {
      dmu_err_set( DM_ERR_NULL, "No filter descriptor" );
      return dmFAILURE;
  }

  if(filter_dd->descType != dmSUBSPACE)
  {
      dmu_err_set( DM_ERR_GENERIC, "Wrong type of descriptor" );
      return dmFAILURE;
  }

  /* Check if we have a component of a vector column */
  if ( filter_dd->cptno > 0 )
  {
   cptNo     = filter_dd->cptno;   
   filter_dd = filter_dd->parent;  
  }  

  if ( !type )
   printf( "DM Internal error: Subspace update type unknown for %s\n", filter_dd->name );

  /* Find the value group which corresponds to the current component */
  filter = dmi_subspace_col_get_current_filter( filter_dd );


  retval = dmi_filter_update( filter, cptNo, name, filter_dd->name, type, NULL, NULL, 0, incnull,
                            update, &remove, &dirty );

  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_subspace_col_get_values                      */
/*                                                             */
/* Description:                                                */
/*     Get Values for subspace descriptor.  Array elements are */
/*     allocated and must be freed by caller.                  */
/*                                                             */
/*  Input parameters:                                          */
/*                dmDeacriptor   *filter_dd                    */
/*                long            cptNo                        */
/*                dmDataType      type                         */
/*                                                             */
/*  Output parameters:                                         */
/*                void          **value1                       */
/*                void          **value2                       */
/*                long           *nvalues                      */
/*                                                             */
/*  Returns:                                                   */
/*              int retval - dmSUCCESS or dmFAILURE            */
/*                                                             */
/***************************************************************/
int dmi_subspace_col_get_values(dmDescriptor* filter_dd, long cptno, dmDataType type, void** value1, void** value2, long* nvalues )
{
 GenericType nullval;
  dmFilter*  filter = NULL;  /* Value group which contains data             */
  dmDescriptor* col_dd;
  long size;
  int incnull;
  int retval;

/* Set nvals = 0 in case of failure return */
  if ( nvalues ) 
    *nvalues = 0;

  if ( value1 )
    *value1 = NULL;

  if ( value2 )
    *value2 = NULL;

  if(!filter_dd || !value1 || !value2 || !nvalues)
  {
      dmu_err_set(DM_ERR_NULL, "Null arg");
      return dmFAILURE;
  }

  if(filter_dd->descType != dmSUBSPACE)
  {
      dmu_err_set( DM_ERR_INTERNAL, "Not a subspace descriptor" );
      return dmFAILURE;
  }

  if ( filter_dd->cptno > 0 )
  {
    cptno = filter_dd->cptno;
    filter_dd = filter_dd->parent;
  }

  /* Find the value group which corresponds to the current component */
  filter = dmi_subspace_col_get_current_filter( filter_dd );

  retval = dmi_filter_extract( filter, cptno, type, value1, value2, nvalues, &incnull );
  if ( !*nvalues && incnull == 1 )
  {
   col_dd = filter_dd->parent;
   if ( dmi_descriptor_get_nullval( col_dd, &nullval ))
   {
    size = ar_mix_type_size( type );
    *nvalues = 1;
    *value1 = calloc( 1, size );
    *value2 = calloc( 2, size );
    ar_gen_get_cast_elt( &nullval, 0, 0, col_dd->type, type, *value1 );
    ar_gen_get_cast_elt( &nullval, 0, 0, col_dd->type, type, *value2 );
   }
  }
  return retval;
}
