/*                                                                
**  Copyright (C) 2002-2008  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmiblock.c                                                  */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_reset_rows                            */
/*                                                             */
/* Description:                                                */
/* Raw blocks set nvrows = nrows. In block open mode, we do    */
/* this after filtering, so that we know whether it's virtual  */
/* or not.  In dataset move to block, we need to do this too,  */
/* so it's broken out into a subroutine.                       */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock   *block - pointer to block      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_reset_rows( dmBlock* block )
{
  if ( block && !block->virtual )
  {
    block->nvrows    = block->nrows;
    block->traversed = UT_TRUE;
  }
}

/***************************************************************/
/*                                                             */
/* Function:   dmi_block_delete                                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDataset *ds        - dataset pointer   */
/*                    const long block_no  - block number      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            int errorStatus                                  */
/*                                                             */
/***************************************************************/
int dmi_block_delete( dmDataset* ds, const long block_no )
{
  long  no     = 0;
  long  nb     = 0;
  long  old_no = 0;
  long  new_no = 0;

  dmIndexBlock  *index_block = NULL;
  dmBlock       *block       = NULL;


  dmi_dataset_block_free( ds, block_no );

  nb = ds->nblocks;

  /* Update the remaining block numbers in both the block and ds structures */
  for ( old_no = block_no + 1; old_no <= nb; old_no++ )
  {
    new_no = old_no - 1;
    index_block = ds->index_blocks[old_no-1];
    index_block->no--;
    ds->index_blocks[new_no-1] = index_block;
  }

  ds->index_blocks[nb-1] = NULL;
  ds->nblocks--;

  /* Actual blocks, not necessarily in order */
  for ( no = 1; no <= ds->nvblocks; no++ )
  {
    block = ds->blocks[no-1];

    if ( block )
    {
      if ( block->no == block_no )
        dmu_err_set( DM_ERR_GENERIC, "Attempt to delete block which is currently open" );
      else
        if ( block->no > block_no )
          block->no--;
    }
  }

  return dmu_err_get(); 
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_ncols                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock  *block - block pointer          */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long   BlockNCols  - number of columns           */
/*                                                             */
/***************************************************************/
long dmi_block_get_ncols( dmBlock* block )
{
  if ( block )
    return block->ncols;

  return 0;
}
/* Set no for an existing block (used in block delete code) */


/***************************************************************/
/*                                                             */
/* Function:   dmi_alloc_to_list                               */
/*                                                             */
/* Description:                                                */
/*   Append/Insert at position 'no' to a buffered list of      */
/*   current size *np and allocated size *nmaxp.               */
/*                                                             */
/*  Input parameters:                                          */
/*            void  ***listp                                   */
/*            long    *nmaxp                                   */
/*            void    *p                                       */
/*            long     no                                      */
/*                                                             */
/*  Output parameters:                                         */
/*            long    *np                                      */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_alloc_to_list( void*** listp, long* np, long* nmaxp, 
                        void* p, long no )
{
  long   n    = 0;
  void **list = NULL;

  if ( *nmaxp < no )
  {
    *nmaxp = 2 * no + 10;
    *listp = realloc( *listp, *nmaxp * UT_SZ_P );
  }
 
  list = *listp;

  /* Zero out any missing blocks */  
  for ( n = *np + 1; n < no; n++ )
    list[n-1] = NULL;

  list[no-1] = p; 
  *np = no;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_free                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock  *block                                  */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_free( dmBlock* block )
{
  dmDataset  *ds = NULL;
  long        i  = 0;

  if ( !block )  /* No problem if block already null */
    return;  

  if ( block->buffer )
    dmi_buffer_free( block->buffer );

  block->buffer = NULL;

  free( block->name );
  block->name = NULL;

  ds = block->ds;

  if ( ds && ds->blocks )
  {
    for ( i = 0; i < ds->nvblocks; i++ )
    {
      if ( ds->blocks[i] == block )
        ds->blocks[i] = NULL;
    }
  }

  /* set current to NULL but leave ds->current_no the same */
  if ( ds )
    ds->current = NULL;

  /* Free it */
  free( block );

}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_close_descriptors                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock  *block                                  */
/*            void     (*DescriptorFree)(dmDescritor*)         */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_close_descriptors( dmBlock* block, void (*DescriptorFree)( dmDescriptor* ))
{
  long i = 0;
  dmDescriptor *dd      = NULL;
  dmDescriptor *ddp     = NULL;
  dmBinSpec    *binspec = NULL;


  binspec = dmi_block_get_binspec( block );
 
  if ( binspec )
  {
    dmi_binspec_free( binspec );

    /* Cols and rawcols are not the same */
    dd = block->cols[0];
    ddp = dmi_descriptor_get_unscaled( dd );

    if ( ddp )
      dd = ddp;

    dmi_descriptor_kernel_free( dd, DescriptorFree );
  }

  /* Assume cols has no new descriptor */
  free( block->cols );
  block->cols  = NULL;
  block->ncols = 0;

  /* Free raw columns */
  for ( i = 0; i < block->nrcols; i++ )
  {
    dd = block->rawcols[i];
    ddp = dmi_descriptor_get_unscaled( dd );

    if ( ddp )
      dd = ddp;

    dmi_descriptor_kernel_free( dd, DescriptorFree );
  }

  free( block->rawcols );
  block->rawcols = NULL;
  block->nrcols = 0;

  /* Free buffer */
  dmi_buffer_free( block->buffer );
  block->buffer = NULL;


  if ( block->row_dd )
  {
   dmi_descriptor_kernel_free( block->row_dd, DescriptorFree );
   block->row_dd = NULL;
  }

  /* Free keys */
  for ( i = 0; i < block->nkeys; i++ )
  {
    dmi_descriptor_kernel_free( block->keys[i], NULL );  /* kernel keys freed in kernel block close */
  }

  free( block->keys );
  block->keys = NULL;
  block->nkeys = 0;

  /* Free subspace */
  dmi_subspace_free( block->subspace,  DescriptorFree );
  block->subspace = NULL;

  dmi_block_close( block );
  dmi_block_free( block );

  block = NULL;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_alloc                                 */
/*                                                             */
/* Description:                                                */
/*   Allocate enough space for at least 1 more block           */
/*   Return the block pointer to a new block                   */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDataset  *ds                                   */
/*            const long  no                                   */
/*            int         create                               */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmBlock *block                                   */
/*                                                             */
/***************************************************************/
dmBlock* dmi_block_alloc( dmDataset* ds, const long no, int create )
{
  dmBlock     *block = NULL;
  dmBlockType type;
  long        listno = 0;
 
 
  if ( no <= 0 || no > ds->nblocks + 1 )
    return NULL;

  /* Create the index block if needed */
  if ( create && no == ds->nblocks+1 )
    dmi_dataset_block_alloc( ds, no );

  block =(dmBlock*)calloc( 1, sizeof( dmBlock )); 

  if ( !block ) 
    return NULL;

  block->index_block = dmi_ds_index( ds, no );
  listno = ds->nvblocks + 1;

  dmi_alloc_to_list( (void***)&ds->blocks, &ds->nvblocks, &ds->maxvblocks, block, listno );

  block->virtual = UT_FALSE;
  block->ds = ds;
  block->no = no;
  block->name = cs_dup( dmi_dataset_get_block_name( ds, no ) );

  type = dmi_dataset_get_block_type( ds, no );
  block->type = type;
  block->buffer = NULL;

  block->keys = NULL;
  block->cols = NULL;
  block->rawcols = NULL;
  block->nrows = 0;
  block->ncols = 0;
  block->nrcols = 0;
  block->maxmeg = DM_MAX_MEG_DEFAULT;
  block->region_bin = DM_REGION_BIN_DEFAULT;
  block->region_nmax= DM_REGION_NMAX_DEFAULT;
  block->nvrows = 0;
  block->traversed = UT_FALSE;
  block->image_bin_type = DT_UNK;
  block->ssupdate = UT_TRUE;

  if ( create )
    block->update = UT_TRUE;
  else
    block->update = UT_FALSE;

  dmi_block_set_filter_compact( block, UT_TRUE );
  return block;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_block_set_kb                                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock *block                                   */
/*            const dmkBlock  *kb                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_set_kb( dmBlock* block, const dmkBlock* kb )
{
  if ( block )
    block->kernel_block = (dmkBlock*)kb;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_kb                                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            const dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmkBlock *block                                  */
/*                                                             */
/***************************************************************/
dmkBlock* dmi_block_get_kb( const dmBlock* block )
{
  if ( block )
    return block->kernel_block;
  else
    return NULL;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_set_name                              */
/*                                                             */
/* Description:                                                */
/*   Assumes memory is already allocated                       */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock    *block                                */
/*            const char *name                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_set_name( dmBlock* block, const char* name )
{
  if ( block )
  {
     if ( block->name )
       free( block->name );

     block->name  = cs_dup( name );
     dmi_dataset_set_block_name( block->ds, block->no, name );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_type                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            const dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmBlockType  type                                */
/*                                                             */
/***************************************************************/
dmBlockType dmi_block_get_type( const dmBlock* block )
{
  if ( block )
    return block->type;
  else
    return dmUNKNOWNBLOCK;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_raw_type                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            const dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmBlockType  type                                */
/*                                                             */
/***************************************************************/
dmBlockType dmi_block_get_raw_type( const dmBlock* block )
{
  dmIndexBlock* index_block = NULL;

  if ( block )
    index_block = block->index_block;

  if ( index_block )
    return index_block->type;
  else
    return dmUNKNOWNBLOCK;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_set_type                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock           *block                         */
/*            const dmBlockType  type                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_set_type( dmBlock* block, const dmBlockType type )
{
  dmIndexBlock *index_block = NULL;

  if ( block )
  {
    block->type = type;
    index_block = block->index_block;

    if ( index_block->type == dmUNKNOWNBLOCK )
      index_block->type = type;
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_close                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock           *block                         */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_close( dmBlock* block )
{
  if ( !block )
    return;

  dmi_ddlist_free( block->cols, block->ncols );
  block->cols = NULL;

  /* For now, let cols do the freeing */
  block->rawcols = NULL;

  /* Subspace should already be gone */
  dmi_subspace_free( block->subspace, NULL );
  block->subspace = NULL;

  dmi_buffer_free( block->buffer );
  free( block->null );

  ar_free_cp( block->pref, block->npref+1 );

  block->buffer = NULL;
  block->traversed = UT_FALSE;
  block->nvrows = 0;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_ddlist_free                                 */
/*                                                             */
/* Description:                                                */
/*             Free a list of descriptors                      */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor  **list                             */
/*            long            n                                */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_ddlist_free( dmDescriptor** list, long n )
{
  long          kOffset = 0;
  dmDescriptor *dd      = NULL;

  if ( !list )
    return;

  for ( kOffset = 0; kOffset < n; kOffset++ )
  {
    dd = list[kOffset];
    dmi_descriptor_free( dd );
  }

  free( list );
  list = NULL;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_raw_col                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            const dmBlock  *block                            */
/*            const long      colNo                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor *descritor                          */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_block_get_raw_col( const dmBlock* block, const long colNo )
{ 
  dmDescriptor *v = NULL;

  if ( colNo <= 0 || colNo > block->nrcols )
    return NULL;

  v = block->rawcols[colNo-1];

  return v;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_col                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            const dmBlock  *block                            */
/*            const long      colNo                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor *descriptor                         */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_block_get_col( const dmBlock* block, const long colNo )
{ 
  dmDescriptor *v = NULL;

  if ( colNo <= 0 || colNo > block->ncols )
    return NULL;

  v = block->cols[colNo-1];

  return v;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_set_nrows                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock   *block                                 */
/*            long       nrows                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_set_nrows( dmBlock* block, long nrows )
{
  if ( block )
    block->nrows = nrows;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_set_raw_cols                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock       *block                             */
/*            dmDescriptor **rawcols                           */
/*            long           nrcols                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_set_raw_cols( dmBlock* block, dmDescriptor** rawcols, long nrcols )
{
  /* Need to check and free here as required */
  if ( block )
  {
    block->rawcols = rawcols;
    block->nrcols  = nrcols;
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_set_update                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock       *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_set_update( dmBlock* block )
{
  if ( block )
   block->update = UT_TRUE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_clear_update                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock       *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_clear_update( dmBlock* block )
{
  if ( block )
    block->update = UT_FALSE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_check_update                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock       *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmBool  status                                   */
/*                                                             */
/***************************************************************/
dmBool dmi_block_check_update( dmBlock* block )
{
  if ( block )
    return block->update;

  return UT_FALSE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_set_pref                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock       *block                             */
/*            char         **pref                              */
/*            const long     npref                             */
/*            const int      flag                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_set_pref( dmBlock* block, char** pref, const long npref, const int flag )
{
  block->npref = npref;
  block->pref = (char**)pref;
  block->pref_default = flag;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_npref                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            const dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long  npref                                      */
/*                                                             */
/***************************************************************/
long dmi_block_get_npref( const dmBlock* block )
{
  if ( block )
    return block->npref;
  else
    return 0;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_pref_default                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            const dmblock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           int  pref_default                                 */
/*                                                             */
/***************************************************************/
int dmi_block_pref_default( const dmBlock* block )
{
  if ( block )
    return block->pref_default;
  else
    return UT_TRUE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_pref                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            const dmBlock *block                             */
/*            long           no                                */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            char *pref                                       */
/*                                                             */
/***************************************************************/
char* dmi_block_get_pref( const dmBlock* block, long no )
{
  return block->pref[no-1];
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_print                                 */
/*                                                             */
/* Description:                                                */
/*            Diagnostic for block status                      */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock       *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_print( dmBlock* block )
{
  char buf[CARD_SIZE];

  printf( "Block %s (%ld in dataset %s)\n", block->name, block->no, block->ds->name );
  dmu_block_type_name( block->type, buf );
  printf( "Block Type: %s\n", buf );
  printf( "Cols: %ld  Rows: %ld\n", block->ncols, block->nrows );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_type_name                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            const dmBlockType blockType                      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            char *name                                       */
/*                                                             */
/***************************************************************/
char* dmi_block_type_name( const dmBlockType blockType )
{
  const long ntypes = 3;
  char       *name  = NULL;
  long        type  = 0;
  char       *names[] = { "Table", "Image", "Unknown" };
  dmBlockType list[]  = { dmTABLE, dmIMAGE, dmUNKNOWNBLOCK };

  for ( type = 0; type < ntypes; type++ )
  {
    if ( list[type] == blockType )
    {
      name = cs_dup( names[type] );
      return( name );
    }
  }

  name = cs_dup( "?" );
  return name;
}
 

/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_pref_list                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            const dmBlock *block                             */
/*            dmDescriptor **weightp                           */
/*            long           nprefp                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor *descriptor                         */
/*                                                             */
/***************************************************************/
dmDescriptor**  dmi_block_get_pref_list( const dmBlock* block, dmDescriptor** weightp, long* nprefp )
{
  dmDescriptor  *weight = NULL;
  dmDescriptor **pref   = NULL;
  long           npref  = 0;
  long           i      = 0;

  if ( block )
  {
    npref = block->npref;

    if ( npref > 0 )
      pref = calloc( npref, UT_SZ_P );

    for ( i = 0; i < npref; i++ )
      pref[i] = dmi_descriptor_list_search( (const dmDescriptor**)block->cols, block->ncols, block->pref[i+1] );

   weight = dmi_descriptor_list_search( (const dmDescriptor**)block->cols, block->ncols, block->pref[0] );
 }

  *nprefp  = npref;
  *weightp = weight;

  return pref;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_set_name                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            const dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            char *name                                       */
/*                                                             */
/***************************************************************/
char* dmi_block_get_name( const dmBlock* block )
{
  if ( block )
    return block->name;

  return NULL;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_ds                                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            const dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmDataset *ds                                    */
/*                                                             */
/***************************************************************/
dmDataset* dmi_block_get_ds( const dmBlock* block )
{
  if ( block )
    return block->ds;
 else
   return NULL;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_no                                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            const dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long no                                          */
/*                                                             */
/***************************************************************/
long dmi_block_get_no( const dmBlock* block )
{
  if ( !block )
   return 0;

  return block->no;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_block_make_default_subspace                 */
/*                                                             */
/* Description:                                                */
/*            We initialize the subspace by using all raw      */
/*            columns.  If it's an image, we should do it for  */
/*            the axis groups...                               */
/*  Input parameters:                                          */
/*            dmBlock       *block                             */
/*            consst long    maxcpt                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_make_default_subspace( dmBlock* block, const long maxcpt )
{
  long i     = 0;
  long ncols = 0;
  long ng    = 0;

  dmDescriptor **list  = NULL;
  dmSubspace    *subspace = NULL;
  dmDescriptor  *dd    = NULL;
  dmDescriptor  *group = NULL;


  if ( block->type == dmIMAGE  )
  {
    if ( block->rawcols )
    {
      dd = block->rawcols[0];

      /* No subspace for null image */
      if ( dd->naxes == 0 && !dmi_block_is_null_primary( block ))
        return;

      if ( dd )
        ng  = dd->ngroups;
    }

    ncols = ng;
  }
  else
  {
    ncols = block->nrcols;
  }

  if ( !block->subspace )
    block->subspace = dmi_subspace_alloc( ncols, maxcpt );

  subspace = block->subspace;
  list = calloc( ncols, UT_SZ_P );

  if ( dd )
  {
   for ( i = 0 ; i < ng; i++ )
   {
     group = dmi_descriptor_get_scaled( dd->groups[i] );
     list[i] = dmi_descriptor_make_subspace( group, subspace );
     dmi_descriptor_copy_range( dd, list[i] );
   }
  }
  else
  {
    for ( i = 0; i < block->nrcols; i++ )
    {
      ncols = block->nrcols;
      dd = block->rawcols[i];
      list[i+ng] = dmi_descriptor_make_subspace( dd, subspace );
      dmi_descriptor_copy_range( dd, list[i+ng] );
   }
  }

  if ( subspace->cols )
  {
#if DEBUG
    fprintf( stderr, "DELETING %ld SS COLS\n", subspace->ncols );
#endif
    free( subspace->cols );
  }

  subspace->cols    = list;
  subspace->maxcols = ncols;
  subspace->ncols   = ncols;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_is_null_primary                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            int nullPrimary                                  */
/*                                                             */
/***************************************************************/
int dmi_block_is_null_primary( dmBlock* block )
{         
 int  nullPrimary = 0;
 dmDescriptor *image = NULL;

  if ( !block )
    return UT_FALSE;

  nullPrimary = (block->no==1) && (block->type == dmIMAGE);

  if ( nullPrimary )
  {
    if ( block->rawcols )
      image = block->rawcols[0];

    if ( image && image->naxes > 0 )
      nullPrimary = 0;
  }

 return  nullPrimary;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_nrows                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long nrows                                       */
/*                                                             */
/***************************************************************/
long dmi_block_get_nrows( dmBlock* block )
{
  if ( block )
    return block->nrows;
  else
    return 0;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_set_maxmeg                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                  long     maxmeg                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long no                                          */
/*                                                             */
/***************************************************************/
void dmi_block_set_maxmeg( dmBlock* block, long maxmeg )
{
  if ( block )
    block->maxmeg = maxmeg;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_maxmeg                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long maxmeg                                      */
/*                                                             */
/***************************************************************/
long dmi_block_get_maxmeg( dmBlock* block )
{
  if ( block )
    return block->maxmeg;
  else 
    return 0;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_bin                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            const dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            double  region_bin                               */
/*                                                             */
/***************************************************************/
double dmi_region_get_bin( const dmBlock* block )
{
  if ( block )
    return block->region_bin;
  else
    return 1.0;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_region_get_nmax                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            const dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long region_max                                  */
/*                                                             */
/***************************************************************/
long dmi_region_get_nmax( const dmBlock* block )
{
  if ( block )
    return block->region_nmax;
  else
    return 1;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_region_set_bin                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                  double   bin                               */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_region_set_bin( dmBlock* block, double bin )
{
  if ( block )
    block->region_bin = bin;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_region_set_nmax                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                  long     nmax                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_region_set_nmax( dmBlock* block, long nmax )
{
  if ( block )
    block->region_nmax = nmax;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_writable                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            int  mode                                        */
/*                                                             */
/***************************************************************/
int dmi_block_writeable( dmBlock* block )
{
  dmDataset *ds = NULL;

  if ( !block )
    return UT_FALSE;

  ds = block->ds;

  return ( ds->mode && !block->virtual );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_kernel                                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmKernel  kernel                                 */
/*                                                             */
/***************************************************************/
dmKernel dmi_block_kernel( dmBlock* block )
{
  dmDataset *ds = NULL;

  if ( !block )
    return NULL;

  ds = block->ds;
  return ds->kernel;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_alloc_row_dd                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_alloc_row_dd( dmBlock* block )
{
  dmDescriptor *dd    = NULL;
  long          naxes = 1;

  if ( !block )
    return;

  dd = dmi_descriptor_alloc( block, dmCOLUMN, 0, "#row", DT_INT4, NULL,
                            "Table row number", 1, dmVALUE, naxes, NULL, DM_FLAG_ROW | DM_FLAG_RESULT );
  block->row_dd = dd;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_set_current_row                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                  long     row                               */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_set_current_row( dmBlock* block, long row )
{
  block->buffer->vrow = row;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_reset_current_row                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long row                                         */
/*                                                             */
/***************************************************************/
long dmi_block_reset_current_row( dmBlock* block )
{
  long row = 0;

  row = dmi_block_get_current_row( block );

  if ( row == 0 )
  {
    row = 1L;
    dmi_block_set_current_row( block, row );
  }

  return row;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_current_row                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long row                                         */
/*                                                             */
/***************************************************************/
long dmi_block_get_current_row( dmBlock* block )
{
  return block->buffer->vrow;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_block_init                                  */
/*                                                             */
/* Description:                                                */
/*             Called on open or create                        */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_init( dmBlock* block )
{
  long colno = 0;
  dmDescriptor *dd = NULL;

  block->ncols = block->nrcols;
  block->cols = calloc( block->ncols, UT_SZ_P );

  for ( colno = 1; colno <= block->ncols; colno++ )
  {
    dd = block->rawcols[colno-1];
    block->cols[colno-1] = dd;
  }

  block->buffer = dmi_buffer_alloc();
  block->keyno = 0;
  dmi_block_alloc_row_dd( block );
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_block_set_hint                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock    *block                          */
/*                  const char *name                           */
/*                  const char *value                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode dmi_block_set_hint( dmBlock* block, const char* name, const char* value )
{
  long        v = 0;
  double      x = 0;
  int         q = 0;
  dmDataType  t;
  dmErrCode   status = dmSUCCESS;

  /* 
     NOTE:  This function is sometimes called with the block 
     pointer deliberately set to NULL.  Example: dmKernelSetOption
     where it checks to see if the option is recognized as a block
     option.  If so, it is added to an internal list for when the
     block is actually opened.

     It must do all name checks and return dmSUCCESS for matches.
     All block subroutines called should check the block pointer
     themselves.

  */


  if ( cs_uceq( name, "VERBOSE" ))
  {
    v = atol( value );
    dmuSetVerbose( v );
  }
  else if ( cs_uceq( name, "MEM" ))
  {
    v = atol( value );
    dmi_block_set_maxmeg( block, v );
  }
  else if ( cs_uceq( name, "AREA" ))
  {
    x = atof( value );
    dmi_region_set_bin( block, x );
  }
  else if ( cs_uceq( name, "AREAPIX" ))
  {
    v = atol( value );
    dmi_region_set_nmax( block, v );
  }
  else if ( cs_uceq( name, "FULL" ))
  {
    q = cs_uceq( value, "NO" );
    dmi_block_set_filter_compact( block, q );
  }
  else if ( cs_uceq( name, "TYPE" ))
  {
    t = dmi_parse_type( value );
    dmi_block_set_image_bin_type( block, t );
  }
  else if ( cs_uceq( name, "SCALE" ))
  {
    q = cs_uceq( value, "NO" );
    dmi_block_set_unscaled( block, q );
  }
  else if ( cs_uceq( name, "UPDATE" ))
  {
    q = !cs_uceq( value, "NO" );
    dmi_block_set_subspace_update( block, q );
  }
  else if ( cs_uceq( name, "-SSUPDATE" ))
  {
    dmi_block_set_subspace_update( block, UT_FALSE );
  }
  else if ( cs_uceq( name, "+SSUPDATE" ))
  {
    dmi_block_set_subspace_update( block, UT_TRUE );
  }
  else if ( cs_uceq( name, "NULL" ))
  {
    dmi_block_set_image_null( block, value );
    if ( dmu_err_get() != dmSUCCESS )
    {
      dmu_err_disp_select_msgs( DM_WARN_PARSE_OPTION );
      dmu_err_clear();
    }
  }
  else
  {
    /* Code change - Bug 9053  March 7, 2006 */
    /* unrecognized option - set error       */
    status = dmFAILURE;
  }

  return  status;
}

/***************************************************************/
/*                                                             */
/* Function:   dmi_block_set_image_null                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock    *block                          */
/*                  const char *value                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_set_image_null( dmBlock* block, const char* value )
{
  dmDescriptor *dd = NULL;
  GenericType   null;
  int           flag;

  /* Code chanage - BUG ?? June 13, 2008         */
  /*   check pointer block validity before using */
  /*   just in case caller does not              */
  if ( !block )
    return;

  if (block->cols && block->type == dmIMAGE )
  {
    dd = block->cols[0];
    flag = dmu_parse_null( value, dd->type, &null );
    if ( flag == DM_NULLTYPE_NONE )
    {
      dmu_err_set(DM_WARN_PARSE_OPTION, 
		  "WARNING: Empty NULL value specified for image, will use default (0).\n");
    }
    else if ( flag == DM_NULLTYPE_BAD )
    {
      dmu_err_set(DM_WARN_PARSE_OPTION, 
       "WARNING: Specified NULL value is inconsistent with image datatype.\nWill use truncated or default value (0).");
    }
    else
    {
      dmi_descriptor_set_null( dd, dd->type, &null );

      if ( block->rawcols[0] != dd && dmi_block_get_raw_type( block ) == dmIMAGE )
	dmi_descriptor_set_null( block->rawcols[0], dd->type, &null );
    }
  }
  else
  {
    block->null = cs_dup( value );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_subspace_update                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            int  update                                      */
/*                                                             */
/***************************************************************/
int dmi_block_get_subspace_update( dmBlock* block )
{
  if ( block )
    return block->ssupdate;

  return UT_FALSE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_set_subspace_update                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                  int      q                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long no                                          */
/*                                                             */
/***************************************************************/
void dmi_block_set_subspace_update( dmBlock* block, int q )
{
  if ( block )
    block->ssupdate = q;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_filter_is_compact                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            int  compact                                     */
/*                                                             */
/***************************************************************/
int dmi_block_filter_is_compact( dmBlock* block )
{
  if ( block )
    return block->compact;

  return UT_FALSE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_set_filter_compact                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                  int      q                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_set_filter_compact( dmBlock* block, int q )
{
  if ( block )
    block->compact = q;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_image_bin_type                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmDataType  image_bin_type                       */
/*                                                             */
/***************************************************************/
dmDataType dmi_block_get_image_bin_type( dmBlock *block )
{
  if ( block )
    return block->image_bin_type;

  return dmUNKNOWNTYPE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_set_image_bin_type                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock   *block                           */
/*                  dmDataType t                               */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_set_image_bin_type( dmBlock *block, dmDataType t )
{
  if ( block )
    block->image_bin_type = t;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_is_unscaled                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            int  unscaled                                    */
/*                                                             */
/***************************************************************/
int dmi_block_is_unscaled( dmBlock* block )
{
  if ( block )
    return block->unscaled;

  return UT_FALSE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_set_unscaled                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                  int      q                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_set_unscaled( dmBlock* block, int q )
{
  if ( block )
    block->unscaled = q;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_parse_type                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            const char  *value                               */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmDataType  type                                 */
/*                                                             */
/***************************************************************/
dmDataType dmi_parse_type( const char* value )
{
  dmDataType type = DT_INT2;

  if ( cs_uceq( value, "I2" ))
    type = DT_INT2;
  else
    if ( cs_uceq( value, "I4" ))
      type = DT_INT4;
    else
      if ( cs_uceq( value, "R4" ))
        type = DT_REAL4;
      else
        if ( cs_uceq( value, "R8" ))
          type = DT_REAL8;

  return type;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_check_tstart                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_check_tstart( dmBlock* block )
{
  double tstart        = 0.0;
  double tstop         = 0.0;
  dmDescriptor *sub_dd = NULL;
  double       *v1     = NULL;
  double       *v2     = NULL;
  long          nv     = 0;
       
  if ( ( dmi_key_read( block, "TSTART", DT_REAL8, &tstart ) != NULL )
     &&  ( dmi_key_read( block, "TSTOP", DT_REAL8, &tstop )   != NULL ) )
  {
    if ( tstop == 0.0 )
      return;

    sub_dd = dmi_subspace_col_open( block->subspace, "TIME" );

    if ( sub_dd )
      dmi_subspace_col_get_values( sub_dd, 1, DT_REAL8, (void**)&v1, (void**)&v2, &nv );

    if ( nv > 0 )
    {
      if (( tstart > v2[nv-1] || tstop < v1[0] ) && dmverbose >= 0 )
      {
        fprintf( stderr, "- WARNING: TSTART and TSTOP do not match GTIs in %s[%s]\n",
                 block->ds->name, block->name );
        fprintf( stderr, "- TSTART:TSTOP = %20.4f:%20.4f\n", tstart, tstop );
        fprintf( stderr, "- GTIs         = %20.4f:%20.4f\n", v1[0], v2[nv-1] );         
      }
    }

    if ( v1 )
      free( v1 );

    if ( v2 )
      free( v2 );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_update_time_range                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock    *block                          */
/*                  const char *name                           */
/*                  double      value                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            int retval                                       */
/*                                                             */
/***************************************************************/
int dmi_block_update_time_range( dmBlock* block, const char* name, double value )
{
  double range[2];
  long   nvalues = 1;

  int    retval = dmSUCCESS;
  double old_range[2];
  double minval = 0.0;
  double maxval = 0.0;

  dmDescriptor *tstart_dd = NULL;
  dmDescriptor *tstop_dd  = NULL;
  dmDescriptor *sub_dd    = NULL;
  dmFilter     *filter    = NULL;
  dmDataType    type;
  dmRangeList  *range_lst = NULL;
  dmSubspace   *ss        = NULL;

  long ncpts = 0;
  long cptno = 0;
  long i     = 0;

  if ( cs_uceq( name, "TSTART" ) )
  {
    range[0] = value;
    range[1] = DT_REAL8_MAX;
    tstart_dd = dmi_key_read( block, "TSTART",  DT_REAL8, &old_range[0] );
  } 
  else
    if ( cs_uceq( name, "TSTOP" ) )
    {
      range[0] = -DT_REAL8_MAX;
      range[1] = value;
      tstop_dd  = dmi_key_read( block, "TSTOP",  DT_REAL8, &old_range[1] );
    }
    else
    {
      return retval;
    }

  ss = dmi_block_get_subspace( block );

  if ( !(sub_dd = dmi_subspace_col_open( ss, "TIME" )))
  {
    dmu_err_clear();
    return dmFAILURE;
  }

  ncpts = dmi_subspace_get_ncpts( ss );
  cptno = ss->cpt;

  if ( tstart_dd || tstop_dd )
  {
    for ( i = 1; i <= ncpts; i++ )
    {
      ss->cpt = i;
      filter = dmi_subspace_col_get_filter( sub_dd, i );
      type = filter->type;
      range_lst = &filter->ranges[0];

      if ( range_lst->n > 0 ) 
      {
        ar_mix_numeric_cast( range_lst->mins, type, DT_REAL8, &minval );
        ar_mix_numeric_cast( range_lst->maxes, type, DT_REAL8, &maxval );
      }

      if ( tstart_dd && range_lst->n == 1 && minval == old_range[0] )
      {
        /* existing range is the old tstart, replace it */
        ar_mix_numeric_cast( &range[0], DT_REAL8, type, range_lst->mins );    
      } 
      else
        if( tstop_dd && range_lst->n == 1 && maxval == old_range[1] )
        {
          /* existing range is old tstop, replace it */
          ar_mix_numeric_cast( &range[1], DT_REAL8, type, range_lst->maxes );    
        }
        else 
        {
          /* the existing range list is now a mixed bag, the best we can do is */ 
          /* intersect it with what's there (BANDAID)                          */
          if ( retval == dmSUCCESS )
            retval = dmi_subspace_col_update_null( sub_dd, NULL, 1, DT_REAL8, (void*)&range[0], (void*)&range[1], nvalues, UT_TRUE );
        }
    }
  }
  else
  {
    /* there wasn't a TSTART/TSTOP to begin with, so we can safely intersect */ 
    /* the new one with what's there                                         */
    for ( i = 1; i <= ncpts; i++ )
    {
      ss->cpt = i;

      if ( retval == dmSUCCESS )
        retval = dmi_subspace_col_update_null( sub_dd, NULL, 1, DT_REAL8, (void*)&range[0], (void*)&range[1], nvalues, UT_TRUE );
    }
  }
  ss->cpt = cptno;

  return retval;
}
  

/***************************************************************/
/*                                                             */
/* Function:   dmi_image_descriptor_clear                      */
/*                                                             */
/* Description:                                                */
/* This ugly procedure is because of the following problem.    */
/* When we create an image, we don't know the data type and    */
/* axes of the image descriptor. So perhaps we shouldn't       */
/* allocate an image descriptor yet at all.                    */
/* But then, the paradigm that a null image still has an image */
/* descriptor is broken - perhaps that's good but it would     */
/* break parts of the current code.  So, e.g. when doing an    */
/* image dmcopy, we create a new image, make a trivial image   */
/* descriptor, then delete the trivial image descriptor (in    */
/* this here routine) and create a new image descriptor        */
/* reflecting the actual properties of the image we are copying*/
/*                                                             */
/*  Input parameters:                                          */
/*            const dmBlock *block                             */
/*            void  (*DescriptorFree)(dmDescriptor)            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_image_descriptor_clear( dmBlock* block, void (*DescriptorFree)( dmDescriptor* ))
{
  dmDescriptor *dd  = NULL;
  dmDescriptor *ddp = NULL;
 
  if( block->nrcols >0 )
  {
    dd = block->rawcols[0];
    ddp = dmi_descriptor_get_unscaled( dd );

    if ( ddp )
      dd = ddp;

    dmi_descriptor_kernel_free( dd, DescriptorFree );

    block->nrcols = 0;
    block->rawcols[0] = NULL;
    block->ncols = 0;
    block->cols[0] = NULL;

    dmi_buffer_free( block->buffer );

    block->buffer = dmi_buffer_alloc();
  }
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_block_set_buffer_maxsize                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock *block                             */
/*                  long     size                              */   
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_set_buffer_maxsize( dmBlock* block, long size )
{
  /* Trigger a buffer resync only if the max size is being changed */
  if ( block && block->buffer && ( block->buffer->max_size != size ) )
  {
    block->buffer->max_size = size;
    block->buffer->resync   = UT_TRUE;
  }
}

