/*                                                                
**  Copyright (C) 2002-2005,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmicolumn.c                                                 */
/*                                                               */
/*   Functions for columns                                       */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"

char* dmi_column_make_name( char* name, char** cptNames, long dim );
dmDescriptor* dmi_dd_reset( dmDescriptor* dd, char* name, dmDataType type, char* unit,
 char* desc, long dim, dmElementType etype, long naxes, long* axes, long flag );


/***************************************************************/
/*                                                             */
/*  Function: dmi_column_make_name                             */
/*                                                             */
/*  Description:                                               */
/*     create column name                                      */
/*                                                             */
/*  Input parameters:                                          */
/*            char     *name                                   */
/*            char    **cptNames                               */
/*            long      dim                                    */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*          char *usedName                                     */
/*                                                             */
/***************************************************************/
char* dmi_column_make_name( char* name, char** cptNames, long dim )
{
  char *usedName = NULL;
  long  i = 0;
  long  size = 2;

  if(cs_is_blank( name ))
  {
     for ( i = 0; i < dim; i++ )
       size += strlen( cptNames[i] ) + 2;
     usedName = malloc( size * UT_SZ_C );     
     strcpy( usedName, "(" );

     for ( i = 0; i < dim; i++ )
     {
       strcat( usedName, cptNames[i] );
       if ( i < dim-1 )
         strcat( usedName, "," );
     }
     strcat( usedName, ")" );
  }
  else
  {
     usedName = cs_dup(name);
  }
  return usedName;
}

/***************************************************************/
/*                                                             */
/*  Function: dmi_dd_reset                                     */
/*                                                             */
/*  Description:                                               */
/*     give descriptor, empty it, creates new elements         */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor  *dd                                */
/*            char          *name                              */
/*            dmDataType     type                              */
/*            char          *unit                              */
/*            char          *desc                              */
/*            long           dim                               */
/*            dmElementType  etype                             */
/*            long           naxes                             */
/*            long          *axes                              */
/*            long           flag                              */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*          char *dd  - empty descriptor with reallocated space*/
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_dd_reset( dmDescriptor* dd, char* name, dmDataType type, char* unit,
 char* desc, long dim, dmElementType etype, long naxes, long* axes, long flag )
{
  long i = 0;

  dd->type = type;

  if ( naxes != dd->naxes )
  {
    dd->naxes = naxes;
    free( dd->axes );
    if ( naxes > 0 )
      dd->axes = calloc( dd->naxes, UT_SZ_I );
  }

  if ( axes )
  {
    for ( i = 0; i < naxes; i++ )
      dd->axes[i] = axes[i];
   }

  if ( unit )
  {
    free( dd->unit );
    dd->unit = cs_dup( unit );
  }

  if ( name && !cs_eq( dd->name, name ))
  {
    free( dd->name );
    dd->name = cs_dup( name );
  }

  if ( desc )
  {
    free( dd->desc );
    dd->desc = cs_dup( desc );
  }

  if ( dd->dim && dd->dim != dim )
  {
    sprintf( dm_errbuf, "Error: Changing descriptor dimension for %s from %ld to %ld is not supported\n", dd->name, dd->dim, dim );
    dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
  }

  dd->dim = dim;
  dd->flag = flag; 
  dd->etype = etype;
  return dd;
}

/***************************************************************/
/*                                                             */
/*  Function: dmi_column_create                                */
/*                                                             */
/*  Description:                                               */
/*     create new column descriptor                            */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock       *block                             */
/*            char          *name                              */
/*            dmDataType     type                              */
/*            long           length                            */
/*            char          *unit                              */
/*            char          *desc                              */
/*            dmElementType  etype                             */
/*            char         **cptNames                          */
/*            long           vdim                              */
/*            long          *axes                              */
/*            long           naxes                             */
/*            dmBool         varlen                            */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*          dmDescriptor  *dd - column descriptor              */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_column_create(dmBlock *block, char *name, dmDataType type, 
		      long length, char *unit, char *desc, 
		      dmElementType etype, char **cptNames,
		      long vdim, long* axes, long naxes, dmBool varlen )
{
  dmDescriptor *dd=NULL;      /* Return value          */
  dmDescriptor *cpt_dd;       /* component descriptor */
  dmDescriptor *us_dd;        /* Unscaled column */

  char *usedName = NULL;   /* Name of vector column */
  char  rawname[80];
  char  rawcpt[80];

  double tscal = 1; /* Default scaling */
  double tzero = 0.0;
  double pix   = 0.0;
 
  long i = 0;           /* Loop variable         */
  long dim = 1;
  long flag = DM_FLAG_VISIBLE | DM_FLAG_RAW | DM_FLAG_VECTOR | DM_FLAG_LOGICAL;
  long no = 0;
  long size = 0;

  if ( !block )
    return NULL;

  if ( vdim == 0 )
     dim = 1;
  else
    dim = vdim;

  if (!cptNames )
  {
    cptNames = &name;

    if ( dim > 1 )
    {
      sprintf( dm_errbuf, "No component names for %ld-dim vector column %s", dim, name );
      dmu_err_set( DM_ERR_VECTOR, dm_errbuf );
      return NULL;
     }   
   }

  if(dim < 1)
  {
    sprintf( dm_errbuf, "Bad vector dimension %ld for column %s", dim, name );
    dmu_err_set( DM_ERR_VECTOR, dm_errbuf );
    return NULL;
  }

  usedName = dmi_column_make_name( name, cptNames, dim );

  size = 1;

  if ( axes )
  {
    for ( i = 0; i < naxes; i++ )
      size *= axes[i];
  }          

  /* Adjust array size for bits */
  if ( type == DT_BIT )
  {
    if ( length )
    {
      size = (length + 7)/8;

      if ( size > 1 && !axes )
      {
        naxes = 1;
        axes=calloc(1,UT_SZ_I);
        axes[0] = size;
      }
    }
    else
      length = 8 * size;
  }
	
  /* There is only one image column. We just redefine the current one. */
  if ( block->type == dmIMAGE && block->nrcols == 1 )
    dmi_delete_col( block->rawcols[0] );

  no = block->ncols + 1;          
  dmi_column_make_raw_name( usedName, no, rawname );
  dd = dmi_descriptor_alloc( block, dmCOLUMN, no, rawname, type, unit, 
                            desc, dim, etype, naxes, axes, flag ); 
  dmi_descriptor_set_length( dd, length );
  dmi_descriptor_set_varlen( dd, varlen );

  if ( cptNames && dim > 1 )
  {         
    if ( !dd->cptNames )
      dd->cptNames = calloc( dim, UT_SZ_P );

    for ( i = 0; i < dim; i++ )
    {
      dmi_column_make_raw_name( cptNames[i], no, rawcpt );
      dd->cptNames[i] = cs_dup( rawcpt );
    }
  }

  us_dd = dd;
  dmi_vector_make_cpts( us_dd, dim );

  dd = dmi_create_linear_wcs( block, us_dd, usedName, us_dd->type, tzero, pix, tscal );
  dmi_descriptor_set_length( dd, length );
  dmi_descriptor_set_varlen( dd, varlen );

  if ( cptNames && dim > 1 )
  {         
    if ( !dd->cptNames )
      dd->cptNames = calloc( dim, UT_SZ_P );

    if ( !dd->cpts )
      dd->cpts  = calloc( dim, UT_SZ_P );

    for ( i = 0; i < dim; i++ )
    {
      dd->cptNames[i] = cs_dup( cptNames[i] );
      dd->cpts[i] = dmi_create_linear_wcs( block, us_dd->cpts[i], cptNames[i], us_dd->type, tzero, pix, tscal );             
      cpt_dd = dd->cpts[i];
      cpt_dd->parent = dd;          
      cpt_dd->cptno  = i+1;
    }
  }

  /* Add dd to rawcols and cols */
  dmi_column_add_to_list( block, dd );

  free(usedName);

  /* Add to buffer */
  dmi_block_add_col( block, dd, flag );
  /* Add to subspace */
  dmi_subspace_col_init(block, NULL, dd, dd->type );
  return dd;
}

void dmi_column_add_to_list( dmBlock* block, dmDescriptor* dd )
{
  long no;
  no = block->ncols + 1;
  dd->no = no;
  block->nrcols++;

  block->rawcols = realloc( block->rawcols, block->nrcols * UT_SZ_P );
  block->rawcols[no-1] = dd;
  block->ncols++;
  block->cols = realloc( block->cols, block->ncols * UT_SZ_P );
  block->cols[no-1] = dd;
}

dmDescriptor* dmi_column_create_subtable( dmBlock* block, dmDescriptor* parent )
{
 dmDescriptor* dd;
 dmElementType etype = dmRANGE;
 long vdim = 2;
 long* axes = NULL;
 long naxes = 1;
 char** cptnames= NULL;
 char* time_cptnames[2] = { "START", "STOP" };
 char** alloc_cptnames= NULL;
 long size;
 if ( parent->etype == dmRANGE && parent->eltNames )
 {
  cptnames = parent->cptNames;
 }
 else if ( cs_uceq( parent->name, "TIME" ))
 {
  cptnames = time_cptnames;
 } else {
  alloc_cptnames = calloc( 2, UT_SZ_P );
  size =  strlen( parent->name ) + 5;
  alloc_cptnames[0] = malloc( size );
  alloc_cptnames[1] = malloc( size );
  sprintf( alloc_cptnames[0], "%s_MIN", parent->name );
  sprintf( alloc_cptnames[1], "%s_MAX", parent->name );
  cptnames = alloc_cptnames;
 }
 dd= dmi_column_create( block, parent->name, parent->type, parent->length, parent->unit, parent->desc, etype, cptnames, vdim, axes, naxes, dmFALSE );
 if ( alloc_cptnames ) free( alloc_cptnames );
 return dd;
}

/***************************************************************/
/*                                                             */
/*  Function: dmi_column_make_raw_name                         */
/*                                                             */
/*  Description:                                               */
/*     create new column descriptor                            */
/*                                                             */
/*  Input parameters:                                          */
/*            char         *name                               */
/*            long          no                                 */
/*            char         *rawname                            */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*          void                                               */
/*                                                             */
/***************************************************************/
void dmi_column_make_raw_name( char* name, long no, char* rawname )
{
#if OLD_METHOD
   sprintf( rawname, "#COL%ld", no );
#else
  if ( name && *name )
   sprintf( rawname, "#U_%s", name );
  else
   sprintf( rawname, "#U_COL%ld", no );
#endif
}


/***************************************************************/
/*                                                             */
/*  Function: dmi_de;ete_col                                   */
/*                                                             */
/*  Description:                                               */
/*     remove a column descriptor from block                   */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor *dd                                 */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*          void                                               */
/*                                                             */
/***************************************************************/
void dmi_delete_col( dmDescriptor* dd )
{
  dmDescriptor *ddp = NULL;
  dmBlock      *block = NULL;
  dmBuffer     *buffer = NULL;
  long no     = 0;
  long offset = 0;
  long bufno  = 0;


  if ( !dd )
    return;

  block = (dmBlock*)dd->block;
  no = dd->no;

  /* First check in buffer */
  buffer = block->buffer;
  bufno = dd->bufno;

  if ( bufno > 0 )
    buffer->cols[bufno-1] = NULL; /* Rely here on hope that scaled dd is what is in buffer  - could do better */

  dmi_descriptor_free( dd );
  block->nrcols--;

  for ( offset = no-1; offset < block->nrcols;  offset++ )
  {
    ddp  = block->rawcols[offset+1];
    ddp->no--;
    block->rawcols[offset] = ddp;
  }

  block->ncols--;

  for ( offset = no-1; offset < block->ncols;  offset++ )
  {
    ddp  = block->cols[offset+1];
    ddp->no--;
    block->cols[offset] = ddp;
  }
}
