/*                                                                
**  Copyright (C) 2002-2012  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmicoord.c                                                  */
/*                                                               */
/*   Basic routines to allocat, free, delete transforms          */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"

void dmi_descriptor_list_delete( dmDescriptor** list, long* np, dmDescriptor* dd );

dmDescriptor* dmi_coord_make_logical_group( dmDescriptor* dd, long groupno, dmDescriptor* axis_dd );


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_transform_copy                        */
/*                                                             */
/* Description:                                                */
/*   Copy one transform to another                             */
/*                                                             */
/*  Input parameters:                                          */
/*      dmDescriptor     *dd            - descriptor           */
/*      dmCoordTransform *old_transform - transform descriptor */
/*      long              dim           - length               */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      dmCoordTransform *new_transform                        */
/*                                                             */
/***************************************************************/
dmCoordTransform* dmi_coord_transform_copy( dmDescriptor *dd,
                                            dmCoordTransform *old_transform,
                                            long dim )
{
  dmCoordTransform* new_transform = NULL;

  if ( old_transform )
  {
    new_transform = dmi_coord_transform_alloc( dd, old_transform->type, old_transform->val,
                                               old_transform->pix, old_transform->delt, dim );

    new_transform->theta    = old_transform->theta;
    new_transform->longpole = old_transform->longpole;
    new_transform->latpole  = old_transform->latpole;
    new_transform->equinox  = old_transform->equinox;
    new_transform->linear   = old_transform->linear;
    new_transform->scaled   = old_transform->scaled;
    new_transform->physical = old_transform->physical;
  }

  dd->aux = new_transform;
  return new_transform;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_transform_alloc                       */
/*                                                             */
/* Description:                                                */
/*   Allocate transform object and initialize fields           */
/*   Does not set longpole, etc.                               */
/*                                                             */
/*   NB: descriptor element aux set with new  pointer value    */
/*                                                             */
/*  Input parameters:                                          */
/*      dmDescriptor   *dd             - descriptor            */
/*      char           *transform_type - transform type        */
/*      double         *val            -                       */
/*      double         *pix            - pixels                */
/*      double         *delt           -                       */  
/*      long            dim            - length                */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      dmCoordTransform *transform                            */
/*                                                             */
/***************************************************************/
dmCoordTransform* dmi_coord_transform_alloc( dmDescriptor *dd,
                        char *transform_type, double *val,
                        double *pix, double *delt, long dim )
{
  long  ii = 0;
  dmCoordTransform *transform = NULL;


  transform = calloc( 1, sizeof( struct dmCoordTransform_s ));

  transform->auxtype = DM_AUX_COORD;

  // Code Change June 23, 2010
  //  Force type to upper case
  transform->type    = cs_dup( transform_type );
  cs_upper( transform->type );

  transform->scaled  = 0;
  transform->linear  = cs_eq( transform->type, "LINEAR" );
  transform->no      = 0;

  for ( ii = 0; ii < dim; ii++ )
  {
    transform->val[ii]  = val[ii];
    transform->pix[ii]  = pix[ii];
    transform->delt[ii] = delt[ii];

    if( val[ii] != pix[ii] || delt[ii] != 1.0 )
    {
      transform->scaled = UT_TRUE;
    }
  }

  dmi_descriptor_set_aux( dd, transform );

  return transform;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_add_coord                        */
/*                                                             */
/* Description:                                                */
/*   Attach a coord descriptor to another descriptor           */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*      dmDescriptor   *dd             - descriptor            */
/*      dmDescriptor   *coord_dd       - coordinate descriptor */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_descriptor_add_coord( dmDescriptor *dd, dmDescriptor *coord_dd )
{
  dmCoordTransform *transform = NULL;

  if ( !dd || !coord_dd ) return;

  dd->ncoords++;
  dd->coords = realloc( dd->coords, dd->ncoords * UT_SZ_P );
  dd->coords[dd->ncoords-1] = coord_dd;

  if ( dd->flag & DM_FLAG_AXIS ) /* This is a coordinate system on an image axis */
    coord_dd->flag |= DM_FLAG_AXIS;

  transform = dmi_descriptor_get_aux( coord_dd );

  if ( transform )
  {
    transform->parent = dd;
    transform->no = dd->ncoords;
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_remove_coord                     */
/*                                                             */
/* Description:                                                */
/*   Detach the specified coordinate from descriptor dd, free  */
/*   the coordinate descriptor, and decrement dd's coordinate  */
/*   count.  If the detached coordinate is not the last in dd's*/
/*   coord list, the last coord in the list is moved to the    */
/*   newly vacated spot.  This means that if the calling       */
/*   code is iterating over the coordinate list, it will have  */
/*   to re-check the current slot after calling                */
/*   dmi_descriptor_remove_coord().                            */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*      dmDescriptor   *dd             - descriptor            */
/*      long            coord_no       - coordinate number     */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_descriptor_remove_coord( dmDescriptor *dd, long coord_no )
{
  if ( !dd || coord_no < 1 || coord_no > dd->ncoords )
    return;
 
  dmi_descriptor_free( dd->coords[coord_no-1] );

  if ( coord_no != dd->ncoords )
  {
    dd->coords[coord_no-1] = dd->coords[dd->ncoords-1];
    /* Don't bother with a realloc() */
  }

  dd->ncoords--;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_transform_type                    */
/*                                                             */
/* Description:                                                */
/*   Returns transform type from coordinate descriptor         */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*      const  dmDescriptor   *dd      - descriptor            */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      char *transform_type                                   */
/*                                                             */
/***************************************************************/
char* dmi_coord_get_transform_type( const dmDescriptor *dd )
{
  dmCoordTransform *transform = NULL;
  dmDescriptor     *ddp       = (dmDescriptor*)dd;
  static char      *null      = " ";

  if ( ddp->cptno > 0 )
    ddp = ddp->parent;

  transform = dmi_descriptor_get_aux( ddp );

  if ( transform && transform->auxtype == DM_AUX_COORD )
    return transform->type;
  else
    return null;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_parent                            */
/*                                                             */
/* Description:                                                */
/*   Returns transform parent descriptor                       */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*      const  dmDescriptor   *dd      - descriptor            */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      dmDescriptor *parent_dd                                */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_coord_get_parent( const dmDescriptor *dd )
{
  dmCoordTransform *transform = NULL;

  if ( !dd )
    return NULL;

  transform = dmi_descriptor_get_aux( dd );

  if ( transform && transform->auxtype == DM_AUX_COORD )
    return transform->parent;
  else if ( dd->flag & DM_FLAG_LOGICAL  && dd->flag & DM_FLAG_AXIS )
    return dd->parent;

  return NULL;
 
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_parent                            */
/*                                                             */
/* Description:                                                */
/*   Set and get transform values. These are accessed through  */
/*   the descriptor object rather than the transform object,   */
/*   which we hide from the API                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*      dmDescriptor   *dd      - descriptor                   */
/*      const double   *val     - value                        */
/*      const long      dim     - length                       */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      int  dmSUCCESS                                         */
/*                                                             */
/***************************************************************/
int dmi_coord_set_val( dmDescriptor *dd, const double *val, const long dim )
{
  long i = 0;
  dmCoordTransform *transform = NULL;

  if ( !dd || !dd->aux )
    return dmFAILURE;

  transform = dmi_descriptor_get_aux( dd );

  for ( i = 0; i < dim; i++ )
  {
    transform->val[i] = val[i];

    if ( val[i] != transform->pix[i] ) 
    {
      transform->scaled = 1;
      dd->ref = 0;
    }
  }
  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_set_pix                               */
/*                                                             */
/* Description:                                                */
/*   Set pix values in transform descriptor.                   */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*      dmDescriptor   *dd      - descriptor                   */
/*      const double   *pix     - value                        */
/*      const long      dim     - length                       */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      int  dmSUCCESS                                         */
/*                                                             */
/***************************************************************/
int dmi_coord_set_pix( dmDescriptor *dd, const double *pix, const long dim )
{
  long i = 0;
  dmCoordTransform *transform = NULL;

  if ( !dd )
    return dmFAILURE;

  transform = dmi_descriptor_get_aux( dd );

  if ( !transform )
    return dmFAILURE;


  for ( i = 0; i < dim; i++ )
  {
    transform->pix[i] = pix[i];

    if ( pix[i] != transform->val[i] ) 
    {
      transform->scaled = 1;
      dd->ref = 0;
    }
 }

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_set_delt                              */
/*                                                             */
/* Description:                                                */
/*   Set delt values in transform descriptor.                  */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*      dmDescriptor   *dd      - descriptor                   */
/*      const double   *delt    - value                        */
/*      const long      dim     - length                       */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      int  dmSUCCESS                                         */
/*                                                             */
/***************************************************************/
int dmi_coord_set_delt( dmDescriptor *dd, const double *delt, const long dim )
{
  long i = 0;
  dmCoordTransform *transform = NULL;

  transform = dmi_descriptor_get_aux( dd );

  if ( !dd || !transform )
    return dmFAILURE;

  for ( i = 0; i < dim; i++ )
  {
    transform->delt[i] = delt[i];

    if ( delt[i] != 1.0 ) 
    {
      transform->scaled = 1;
      dd->ref = 0;
    }
 }

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_set_delt                              */
/*                                                             */
/* Description:                                                */
/*    Get logical-to-world transform values                    */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor *dd    - descriptor                 */
/*      const dmDescriptor *dd_lp - descriptor                 */
/*            double       *crpix - pixels                     */
/*            double       *crval - values                     */
/*            double       *cdelt -                            */
/*            long          dim   -                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_coord_convert_pw_to_lw( dmDescriptor *dd, const dmDescriptor *dd_lp,
                      double *crpix, double *crval, double *cdelt, long dim )
{
  long   i    = 0;
  double delt = 0.0;

  dmCoordTransform *transform = NULL;
  dmCoordTransform *phys      = NULL;

  if ( !dd || !dd_lp ) return;

  transform = dmi_descriptor_get_aux( dd );
  phys      = dmi_descriptor_get_aux( dd_lp );

  if ( transform && phys && ( dd->dim == dd_lp->dim ) )
  {
    /* Use dim argument, normally equal to dd->dim */

    for ( i = 0; i < dim; i++ )
    {
      /* Fudge to avoid division by zero in the pathological broken case */
      delt     = ( phys->delt[i] != 0.0 ) ? phys->delt[i]: 1.0;
      cdelt[i] = transform->delt[i] * delt;
      crpix[i] = phys->pix[i] + ( transform->pix[i] - phys->val[i] ) / delt;
      crval[i] = transform->val[i];
    }
  } 
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_convert_lw_to_pw                      */
/*                                                             */
/* Description:                                                */
/*    Rescale a transform dd by the linear transform in ddp.   */
/*    Used in the FTK to map  FITS logical-to-world to DM      */
/*    physical-to-world                                        */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor *dd    - descriptor                 */
/*      const dmDescriptor *dd_lp - descriptor                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_coord_convert_lw_to_pw( dmDescriptor *dd, const dmDescriptor *dd_lp )
{
  double delt = 0.0;
  long   i    = 0;

  dmCoordTransform *transform = NULL;
  dmCoordTransform *phys      = NULL;

  if ( !dd || !dd_lp ) return;

  transform = dmi_descriptor_get_aux( dd );
  phys      = dmi_descriptor_get_aux( dd_lp );
  transform->scaled = 1;
  dd->ref = 0;

  if ( transform && phys && ( dd->dim == dd_lp->dim ) )
  {
    for ( i = 0; i < dd->dim; i++ )
    {
      /* Fudge to avoid division by zero in the pathological broken case */
      delt               = ( phys->delt[i] != 0.0 ) ? phys->delt[i]: 1.0;
      transform->delt[i] = transform->delt[i] / delt;
      transform->pix[i]  = ( transform->pix[i] - phys->pix[i] ) * delt + phys->val[i];
    }
  } 
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_make_transform_vector                 */
/*                                                             */
/* Description:                                                */
/*    NB: pointer to new transform vector set in dd->aux       */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor *dd    - descriptor                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_coord_make_transform_vector( dmDescriptor *dd )
{
  dmCoordTransform *cpt_coords[2];
  dmCoordTransform *t    = NULL;
  char             *type = NULL;
  long              dim  = 2;
  long              i    = 0;

  double val[2];
  double pix[2];
  double delt[2];


  if ( dd->dim != 2 )
    return;

  cpt_coords[0] = dmi_descriptor_get_aux( dd->cpts[0] );
  cpt_coords[1] = dmi_descriptor_get_aux( dd->cpts[1] );

  if ( !cpt_coords[0] || !cpt_coords[1] )
    return;

  for ( i = 0; i < dim; i++ )
  {
    val[i]  = cpt_coords[i]->val[0];
    pix[i]  = cpt_coords[i]->pix[0];
    delt[i] = cpt_coords[i]->delt[0]; 
  }

  if ( dim == 2 && cs_eq( cpt_coords[1]->type, "TAN-P" ))
    type = cpt_coords[1]->type;
  else
    type = cpt_coords[0]->type;

  t = dmi_coord_transform_alloc( dd, type, val, pix, delt, dim );

  if ( t )
  {
    /* theta wasn't necessarily set for the all coordinates, but it's */
    /* the same for all of them, so take the first non-zero value     */
    /* we find                                                        */

    t->theta = 0.0;
 
    for ( i = 0; i < dim; i++ )
      if ( fabs( cpt_coords[i]->theta ) > DBL_EPSILON )
      {
        t->theta = cpt_coords[i]->theta;
         break;
       }

    t->longpole = cpt_coords[0]->longpole;
    t->latpole  = cpt_coords[0]->latpole;
    t->equinox  = cpt_coords[0]->equinox;
    t->radesys  = cs_dup(cpt_coords[0]->radesys);
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_transform_d                       */
/*                                                             */
/* Description:                                                */
/*    Returns values crpix, crval, and cdelt from transform    */
/*    in dd->aux                                               */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor *dd    - descriptor                 */
/*            long          dim   -                            */
/*                                                             */
/*  Output parameters:                                         */
/*            double       *crpix -                            */
/*            double       *crval -                            */
/*            double       *cdelt -                            */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_coord_get_transform_d( dmDescriptor *dd, double *crpix,
                        double *crval, double *cdelt, long dim )
{
  dmCoordTransform *transform = NULL;

  long vdim = 0;
  long i    = 0; 
  int  ok   = 0;


  if ( !dd ) return;

  vdim = dd->dim;

  transform = dmi_descriptor_get_aux( dd );
  ok        = ( transform && transform->auxtype == DM_AUX_COORD );

  for ( i = 0; i < dim; i++ )
  {
    if ( ok && i < vdim )
    {
      crpix[i] = transform->pix[i];
      crval[i] = transform->val[i];
      cdelt[i] = transform->delt[i];
    }
    else
    {
      /* Handle case where we ask for 2D and it's only 1D */
      crpix[i] = 0.0;
      crval[i] = 0.0;
      cdelt[i] = 1.0;
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_set_transform_type                    */
/*                                                             */
/* Description:                                                */
/*    Set transform type in dmCoordDescriptor stored in        */
/*    dd->aux                                                  */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor *dd    - descriptor                 */
/*            char         *type  - transform type             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_coord_set_transform_type( dmDescriptor *dd, char *type )
{
  dmCoordTransform *transform = NULL;

  transform = dmi_descriptor_get_aux( dd );

  if ( !transform ) return;

  free( transform->type );
  transform->type = cs_dup( type ); 
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_create_unit_wcs                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock      *block = block pointer              */
/*            dmDescriptor *col   - descriptor                 */
/*            char         *name  - name                       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      dmDescriptor *dd                                       */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_create_unit_wcs( dmBlock *block, dmDescriptor *col, char *name )
{
  double val = 0.0;
  double pix = 0.0;
  double delt = 1.0;

  return dmi_create_linear_wcs( block, col, name, col->type, val, pix, delt );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_create_linear_wcs                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock      *block = block pointer              */
/*            dmDescriptor *col   - descriptor                 */
/*            char         *name  - name                       */
/*            dmDataType    type  - transform type             */
/*            double        val   -                            */
/*            double        pix   -                            */
/*            double        delt  -                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      dmDescriptor *dd                                       */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_create_linear_wcs( dmBlock *block, dmDescriptor *col,
      char *name, dmDataType type, double val, double pix, double delt )
{
  dmDescriptor *dd = NULL;

 double vals[dmMAXDIM];
 double vpix[dmMAXDIM];
 double vdlt[dmMAXDIM];

 long i    = 0;
 long dim  = 1;
 long flag = DM_FLAG_COORD;


 if ( !col )
   return NULL;


  /* Coordinate systems directly applied to image pixels are flaged as being image */

  if ( col->flag & DM_FLAG_IMG ) 
  {
    /* By default we mark images as unbuffered */
    flag |= DM_FLAG_IMG | DM_FLAG_UNBUF;
    col->flag |= DM_FLAG_UNBUF;
  } 
  else if ( col->flag & DM_FLAG_LOGICAL )
  {
    flag |= DM_FLAG_PHYS;
  }

  dd = dmi_descriptor_alloc( block, dmCOORD, col->no, name, type, col->unit, 
                             col->desc, col->dim, col->etype, col->naxes, col->axes, flag );
  dmi_descriptor_set_length( dd, col->length );

  if ( dim < dd->dim )
  {
    for ( i = 0; i < dd->dim; i++ )
    {
      vals[i] = val;
      vpix[i] = pix;
      vdlt[i] = delt;
    }
    dmi_coord_transform_alloc( dd, "LINEAR", vals, vpix, vdlt, dd->dim );
  }
  else 
    dmi_coord_transform_alloc( dd, "LINEAR", &val, &pix, &delt, dim );

  if ( type != col->type )
    dmi_coord_transform_set_scaled( (dmCoordTransform*)dd->aux, dmTRUE);

  dmi_descriptor_add_coord( col, dd );
  return dd;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_no_groups                    */
/*                                                             */
/* Description:                                                */
/*   Returns number of groups attached to a descriptor         */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const  dmDescriptor *dd    - descriptor                 */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      long  number_of_groups                                 */
/*                                                             */
/***************************************************************/
long dmi_descriptor_get_no_groups( const dmDescriptor *dd )
{
  if ( dd )
   return dd->ngroups;

  return 0;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_no_groups                    */
/*                                                             */
/* Description:                                                */
/*   Return physical group n, which is the first coord on      */
/*   logical group n.                                          */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const  dmDescriptor *dd    - descriptor                 */
/*            long          no    - group number               */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      dmDescriptor *group                                    */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_get_group_no( const dmDescriptor *dd, long no )
{
  dmDescriptor *group  = NULL; /* Logical group */
  dmDescriptor *sgroup = NULL;

  if ( dd && dd->groups && no > 0 && no <= dd->ngroups )
  {
    group = dd->groups[no-1];

    if ( group && group->coords )
    {
      sgroup =  group->coords[0];
      return sgroup;
    }
  }

  return NULL;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_set_groups                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor  *col     - descriptor              */
/*            dmDescriptor **vectors -                         */
/*            long            ng     - number of groups        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_descriptor_set_groups( dmDescriptor *col, dmDescriptor **vectors, long ng )
{
  long i = 0;

  free( col->groups );
  col->ngroups = ng;

  if ( !ng ) return;

  col->groups = calloc( ng, UT_SZ_P );

  for ( i =0; i < ng; i++ )
  {
    dmi_descriptor_set_axis( col, i+1, vectors[i] );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_group_group_no                   */
/*                                                             */
/* Description:                                                */
/*   Here's some pretty confusing code....                     */
/*                                                             */
/*   If ddp is an image axis, this function will return the    */
/*   axis' parent image, number of the group containing the    */
/*   axis, and the axis vector component number within the     */
/*   group.                                                    */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*      const dmDescriptor  *ddp      - descriptor             */
/*            dmDescriptor **parentp  -                        */
/*            long          *groupnop - number of groups       */
/*            long          *vcptp    -                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_descriptor_get_group_group_no( const dmDescriptor *ddp,
                 dmDescriptor **parentp, long *groupnop, long *vcptp )
{
  dmDescriptor *dd     = (dmDescriptor*)ddp;
  dmDescriptor *parent = NULL;

  long vcpt    = 0;
  long groupno = 0;


  if ( !dd )
    return;

  if ( parentp )
    *parentp = NULL;

  vcpt = dd->cptno;

  if ( vcpt > 0 )
    dd = dd->parent;
  else
    vcpt = 1;

  if ( vcptp )
    *vcptp = vcpt;

  while ( dd &&  (dd->flag & DM_FLAG_COORD) && !( dd->flag & DM_FLAG_LOGICAL ))
  {
    dd = dmi_coord_get_parent( dd );
  }

  if ( dd &&  (dd->flag & DM_FLAG_COORD) && ( dd->flag & DM_FLAG_LOGICAL ))
  {
    parent = dmi_coord_get_parent( dd );

    if (parentp) 
      *parentp = parent;

    if ( parent && parent->groups )
    {
      groupno = dd->no;

      if ( dd == parent->groups[groupno-1] )
      {
        if ( groupnop )
          *groupnop = groupno;
        return;
      }
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_group_axis_no                */
/*                                                             */
/* Description:                                                */
/*    This returns the axis number of a descriptor or 0 if it  */
/*    isn't an image axis.                                     */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*      const dmDescriptor  *ddp      - descriptor             */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      long axis_number                                       */
/*                                                             */
/***************************************************************/
long dmi_descriptor_get_group_axis_no( const dmDescriptor *ddp )
{
  dmDescriptor *parent = NULL;

  long vcpt    = 1;
  long groupno = 0;


  if ( !ddp )
    return 0;

  dmi_descriptor_get_group_group_no( ddp, &parent, &groupno, &vcpt );
 
  if ( groupno > 0 )
    return dmi_descriptor_get_group_axis( parent, groupno, vcpt );

  return 0;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_groupno_of_axis              */
/*                                                             */
/* Description:                                                */
/*    This function returns the group number given an axis     */
/*    number. It also sets the vector component of the axis    */
/*    within the group.                                        */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor  *dd       - descriptor             */
/*            long           t_axisno -                        */
/*            long          *vcpt     -                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      long axis_number                                       */
/*                                                             */
/***************************************************************/
long dmi_descriptor_get_groupno_of_axis( dmDescriptor *dd, long t_axisno, long *vcpt )
{
  dmDescriptor *group = NULL;

  long dim          = 0;
  long group_offset = 0;
  long axisno       = 0;
  long i            = 0;


  if ( !dd )
    return 0;

  *vcpt = 0;

  if ( dd->groups )
  {
    for ( group_offset = 0; group_offset < dd->ngroups; group_offset++ )
    {
      group = dd->groups[group_offset];
      dim = group? group->dim : 1;

      for ( i = 0; i < group->dim; i++ )
      {
        axisno++;

        if ( t_axisno == axisno )
        {
          *vcpt = dim > 1 ? i+1 : 0;
          return group_offset+1;
        }
      }
    }
  }

  return 0;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_group_axis                   */
/*                                                             */
/* Description:                                                */
/*    Given an image descriptor, a group number, and a vector  */
/*    component number to index into the group, return the     */
/*     axis number.                                            */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor  *dd       - descriptor             */
/*            long           t_axisno -                        */
/*            long           vcpt     -                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      long axis_number                                       */
/*                                                             */
/***************************************************************/
long dmi_descriptor_get_group_axis( dmDescriptor *dd, long groupno, long vcpt )
{
  long axisno = 1;
  long group_offset;
  dmDescriptor* group = NULL;

  if ( !dd )
    return 0;

  if ( groupno > dd->ngroups )
    return 0;

  if ( dd->groups )
  {
    for ( group_offset = 0; group_offset <= groupno-1; group_offset++ )
    {
      group = dd->groups[group_offset];

      if ( group_offset == groupno-1 )
        return axisno + vcpt - 1;
      else
        if ( group )
          axisno += group->dim;
        else
          axisno += 1;
    }
  }

  axisno += vcpt - 1;
  return axisno;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_set_axis                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor  *dd       - descriptor             */
/*            long           groupno  -                        */
/*            dmDescriptor  *axis_dd  -                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_descriptor_set_axis( dmDescriptor *dd, long groupno, dmDescriptor *axis_dd )
{
  dmCoordTransform *transform =  NULL;
  dmDescriptor     *logical_dd = NULL;

  char lname[80];
  long dim = 0;
  long n   = dd->ngroups;

  if ( groupno <= 0 || groupno > dd->naxes )
    return;

  if ( groupno > n || !dd->groups )
  {
    dmDescriptor **groups;
    long i    = 0;
    long nmax = n;

    /* Start off with naxes */

    dim = axis_dd->dim;

    if ( nmax == 0 )
      nmax = dd->naxes + 1 - dim;

    if ( groupno > nmax )
      nmax = groupno;

    /* Realloc groups ensuring remaining ones null */

    groups = calloc( nmax, UT_SZ_P );
    dd->ngroups = nmax;

    for ( i = 0; i < n; i++ )
      groups[i] = dd->groups[i];

    free( dd->groups );
    dd->groups = groups;
  }

  axis_dd->no = groupno;

  if ( !( axis_dd->flag & DM_FLAG_LOGICAL ) )
  { 
    axis_dd->flag |= DM_FLAG_AXIS | DM_FLAG_PHYS;
    logical_dd = dmi_coord_get_parent( axis_dd );

    /* EG: dd = IMAGE axis_dd = SKY */
    /* Make "#G1" logical group descriptor */

    if ( !logical_dd )
    {
      logical_dd = dmi_coord_make_logical_group( dd, groupno, axis_dd );
      /* #G1 is group on IMAGE or #U_IMAGE */  
    }
    else
    {
      logical_dd->no = groupno;

      if ( cs_eq_siz( logical_dd->name, "#G", 2 ))
      {
        free( logical_dd->name );
        sprintf( lname, "#G%ld", groupno );
        logical_dd->name = cs_dup( lname );
      }
    }
    /* SKY is coord on #G1 */
  }
  else
  {
    logical_dd = axis_dd;
  }
 
  dd->groups[groupno-1] = logical_dd;
  logical_dd->parent = dd;
 
  transform = dmi_descriptor_get_aux( logical_dd );

  if ( transform )
  {
    transform->parent = dd;
    transform->no = 0;
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_group_no                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor  *dd       - descriptor             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      long  group_no                                         */
/*                                                             */
/***************************************************************/
long dmi_coord_get_group_no(dmDescriptor *dd)
{
  dmDescriptor *parent = NULL;
  long no = 0;


  if ( dd && ( dd->flag & DM_FLAG_AXIS ))
  {
    if (dd->flag & DM_FLAG_CPT )
    {
      parent = dd->parent;
      no = parent->no;
    } 
    else 
    {
      no = dd->no;
    }
  }
  return no;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_make_logical_group                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor  *dd       - descriptor             */
/*            long           groupno  -                        */
/*            dmDescriptor  *axis_dd  -                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      dmDescriptor *logical_dd                               */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_coord_make_logical_group( dmDescriptor *dd,
                            long groupno, dmDescriptor *axis_dd )
{
  dmDescriptor *logical_dd = NULL;
  dmDescriptor *cpt        = NULL;
  dmBlock *block           = (dmBlock*)dd->block;

  char lname[80];

  long i      = 0;
  long axisno = 0;
 

  sprintf( lname, "#G%ld", groupno );

  logical_dd = dmi_descriptor_alloc( block, dmCOORD, groupno, lname, axis_dd->type,
                                      "pixel", "Logical axis", axis_dd->dim, 
                                      axis_dd->etype, axis_dd->naxes, axis_dd->axes, 
                                      DM_FLAG_COORD | DM_FLAG_AXIS | DM_FLAG_LOGICAL );

  if( axis_dd->dim > 1 && axis_dd->cpts )
  {
    if ( !logical_dd->cpts )
      logical_dd->cpts = calloc( axis_dd->dim, UT_SZ_P );

    for ( i = 0; i < axis_dd->dim; i++ )
    {
      /* Coord parent of x,y should be #1, #2 */
      cpt = dmi_coord_get_parent( axis_dd->cpts[i] );
      axisno = dmi_descriptor_get_group_axis( dd, groupno, i+1 );
      sprintf( lname, "#%ld", axisno );

      if ( !cpt )
      {
        cpt = dmi_descriptor_alloc( block, dmCOORD, groupno, lname, axis_dd->type,
                                    "pixel", "Logical axis", 1, 
                                     axis_dd->etype, axis_dd->naxes, axis_dd->axes, 
                       DM_FLAG_CPT | DM_FLAG_COORD | DM_FLAG_AXIS | DM_FLAG_LOGICAL );

        dmi_descriptor_add_coord( cpt, axis_dd->cpts[i] );
      }
      else
      { /* Make sure cpt is marked as such */
        cpt->flag |= DM_FLAG_CPT;
        /* Rename */

        if( cs_eq_siz( cpt->name, "#G", 2 ))
        {
          free( cpt->name );
          cpt->name = cs_dup( lname );     
        }
      }

      /* Should be "x" */
      logical_dd->cpts[i] = cpt;    
      cpt->cptno = i+1;
      logical_dd->cptNames[i] = cs_dup( cpt->name );
    }
  }

  dmi_descriptor_add_coord( logical_dd, axis_dd );
  return logical_dd;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_detach_cpts                           */
/*                                                             */
/* Description:                                                */
/*   For each component coord, we detach it from the component */
/*   col and reattach it to the vector col.                    */
/*   Thus, RA and DEC are removed from X and Y and attached    */
/*   to SKY.                                                   */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor  *dd       - descriptor             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_coord_detach_cpts( dmDescriptor *dd )
{
  dmDescriptor    *cpt                   = NULL;
  dmDescriptor     *transform_parent     = NULL;
  dmDescriptor     *cpt_transform_parent =  NULL;
  dmCoordTransform *ct                   = NULL;
  dmDescriptor     *lparent              = NULL;

  long i = 0;
 
  transform_parent = dmi_coord_get_parent( dd );

  if ( !transform_parent )
    return;

  lparent = dmi_coord_get_parent( transform_parent );

  for ( i = 0; i < dd->dim; i++ )
  {
    cpt = dd->cpts[i];
    ct  = cpt->aux;
    cpt_transform_parent = ct->parent;
    dmi_descriptor_list_delete( cpt_transform_parent->coords, &cpt_transform_parent->ncoords, cpt );
    ct->parent = transform_parent;  
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_list_delete                      */
/*                                                             */
/* Description:                                                */
/*   Remove a descriptor from a descriptor list                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor **list   - descriptor list          */
/*            dmDescriptor  *dd     - descriptor to delete     */
/*                                                             */
/*  Output parameters:                                         */
/*            long          *np     - number of pointers in    */
/*                                    list                     */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_descriptor_list_delete( dmDescriptor **list, long *np,
                                 dmDescriptor *dd )
{
  long j = 0;
  long k = 0;
  long n = *np;

  for ( j = 0; j < n; j++ )
  {
    if ( list[j] == dd )
    {
      for ( k = j+1; k < n; k++ )
      {
        list[k-1] = list[k];
      }
      *np = n-1;
      return;
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_theta                             */
/*                                                             */
/* Description:                                                */
/*   Return value of theta in transform                        */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmCoordTransform *transform - transform descriptor  */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      double theta                                           */
/*                                                             */
/***************************************************************/
double dmi_coord_get_theta( dmCoordTransform *transform )
{
  if ( transform )
    return transform->theta;

  return 0.0;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_set_theta                             */
/*                                                             */
/* Description:                                                */
/*   Set value of theta in transform                           */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor  *dd    - descriptor                */
/*            double         theta - value                     */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_coord_set_theta( dmDescriptor *dd, double theta )
{
  dmCoordTransform *transform = NULL;

  if ( dd && dd->descType == dmCOORD )
  {
    transform = dd->aux;

    if ( transform )
      transform->theta = theta;
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_poles                             */
/*                                                             */
/* Description:                                                */
/*   Return values of longpole and latpole from transform      */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmCoordTransform *transform - descriptor         */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*            double  *poles  - values                         */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_coord_get_poles( dmCoordTransform *transform, double *poles )
{
  if ( transform )
  {
    poles[0] = transform->longpole;
    poles[1] = transform->latpole;
  }
  else
  {
    poles[0] = 0.0;
    poles[1] = 0.0;
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_equinox                           */
/*                                                             */
/* Description:                                                */
/*   Return values of equinox from transform                   */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmCoordTransform *transform - descriptor         */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      double  equinox_value                                  */
/*                                                             */
/***************************************************************/
double dmi_coord_get_equinox( dmCoordTransform *transform )
{
  if ( transform )
    return transform->equinox;
  else
    return 0.0;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_set_equinox                           */
/*                                                             */
/* Description:                                                */
/*   Set equinox element in transform descriptor               */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor  *dd    - descriptor                   */
/*         double         value - equinox value                */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_coord_set_equinox( dmDescriptor *dd, double value )
{
  dmCoordTransform *transform = NULL;

  if ( dd && dd->descType == dmCOORD )
  {
    transform = dd->aux;

    if ( transform )
      transform->equinox = value;
  }
}

/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_longpole                          */
/*                                                             */
/* Description:                                                */
/*   Return values of longpole from transform                  */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmCoordTransform *transform - descriptor         */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      double  longpole_value                                 */
/*                                                             */
/***************************************************************/
double dmi_coord_get_longpole( dmCoordTransform *transform )
{
  if ( transform )
    return transform->longpole;
  else
    return 0.0;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_set_longpole                          */
/*                                                             */
/* Description:                                                */
/*   Set longpole value in transform descriptor                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor  *dd    - descriptor                   */
/*         double         value - longpole value               */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_coord_set_longpole( dmDescriptor *dd, double value )
{
  dmCoordTransform *transform = NULL;

  if ( dd && dd->descType == dmCOORD )
  {
    transform = dd->aux;

    if ( transform )
      transform->longpole = value;
  }
}

/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_latpole                           */
/*                                                             */
/* Description:                                                */
/*   Return values of latpole from transform                   */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmCoordTransform *transform - descriptor         */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      double  latpole_value                                  */
/*                                                             */
/***************************************************************/
double dmi_coord_get_latpole( dmCoordTransform *transform )
{
  if ( transform )
    return transform->latpole;
  else
    return 0.0;
}

/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_set_latpole                           */
/*                                                             */
/* Description:                                                */
/*   Set latpole value in transform descriptor                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor  *dd    - descriptor                   */
/*         double         value - latpole value                */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_coord_set_latpole( dmDescriptor *dd, double value )
{
  dmCoordTransform *transform = NULL;

  if ( dd && dd->descType == dmCOORD )
  {
    transform = dd->aux;

    if ( transform )
      transform->latpole = value;
  }
}

/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_radesys                           */
/*                                                             */
/* Description:                                                */
/*   Return values of radesys from transform                   */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmCoordTransform *transform - descriptor         */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      char  *radesys_value                                   */
/*                                                             */
/***************************************************************/
char* dmi_coord_get_radesys( dmCoordTransform *transform )
{
  if ( transform )
    return transform->radesys;
  else
    return NULL;
}

/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_set_radesys                           */
/*                                                             */
/* Description:                                                */
/*   Set radesys value in transform descriptor                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor  *dd      - descriptor                 */
/*         char          *radesys - radesys value string       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_coord_set_radesys( dmDescriptor *dd, char *radesys )
{
  dmCoordTransform *transform = NULL;

  if ( dd && dd->descType == dmCOORD )
  {
    transform = dd->aux;

    if ( transform )
      transform->radesys = cs_dup(radesys);
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_transform_no                      */
/*                                                             */
/* Description:                                                */
/*   Returns transform number from descriptor                  */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor  *dd    - descriptor                   */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      long  no                                               */
/*                                                             */
/***************************************************************/
long dmi_coord_get_transform_no( dmDescriptor *dd )
{
  long              no        = 0;
  dmCoordTransform *transform = NULL;

  if ( dd && dd->descType == dmCOORD )
  {
    transform = dd->aux;

    if ( transform )
      no = transform->no;
  }
  return no;
}




/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_image                             */
/*                                                             */
/* Description:                                                */
/*   Returns image descriptor                                  */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor  *vector - descriptor                  */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      dmDescriptor *image                                    */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_coord_get_image( dmDescriptor *vector )
{
  dmDescriptor *image    = NULL;
  dmDescriptor *group_dd = vector;

  while ( group_dd && !(group_dd->flag & DM_FLAG_LOGICAL ))
    group_dd = dmi_coord_get_parent( group_dd );

  image = dmi_coord_get_parent( group_dd );
  return image;
}

