/*                                                                
**  Copyright (C) 2002-2012  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/*****************************************************************/
/*                                                               */
/*   dmicoordlib.c                                               */
/*                                                               */
/*   Coordinate functions                                        */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"
#include "wcs.h"
#include <limits.h>
#include <ctype.h>

/*-------------------------------------------------------------*/
/* 		    Private Functions 	                       */
/*-------------------------------------------------------------*/
static int dmi_wcs_standard_axis_name( const char* name );


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_create                                */
/*                                                             */
/* Description:                                                */
/*   Create a coordinate descriptor and attach it to the input */
/*   parent descriptor.                                        */
/*                                                             */
/*   The parameters cptNames, params, and ttype may be NULL.   */
/*     cptNames = NULL implies a linear transform              */
/*     ttype    = NULL implies a linear transform also         */
/*     params   = NULL usually                                 */
/*                                                             */
/*   Allowed component pair names for 2D transforms:           */
/*      RA--  DEC-                                             */
/*      GLON  GLAT                                             */
/*      ELON  ELAT                                             */
/*      HLON  HLAT                                             */
/*      SLON  SLAT                                             */
/*      PLON  PLAT                                             */
/*      XLON  XLAT                                             */
/*                                                             */
/*  Input parameters:                                          */
/*      dmDescriptor  *dd        - parent descriptor           */
/*      char          *name      - coord descritpor name       */
/*      char          *unit      - coord units                 */
/*      char         **cptNames  - coord component names       */
/*      long           dim       - dim                         */
/*      char          *ttype     -                             */
/*      void          *pix       -                             */
/*      void          *val       -                             */
/*      void          *delt      -                             */
/*      dmDataType     dtype     -  data type                  */
/*      double        *params    -                             */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      dmDescriptor  retval - output coordinate descriptor    */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_coord_create(dmDescriptor *dd, char *name, char *unit,
			     char **cptNames, long dim, char *ttype, 
			     void *pix, void *val, void *delt, 
			     dmDataType dtype, double *params)
{
  dmDescriptor *retval = NULL;     /* Output coordinate descriptor     */
  dmBlock      *block  = NULL;


  if( !dd || (dim > 1 && ( !cptNames || !ttype ))) 
  {
    dmu_err_set(DM_ERR_NULL, "Null arg");
    return NULL;
  }

  if( dd->descType == dmCOLUMN && dim > dd->dim) 
  {
    sprintf( dm_errbuf, "Coord dim %ld is greater than descriptor dim %ld", 
	     dim, dd->dim );
    dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
    return NULL;
  }

  block = (dmBlock*)dd->block;

  retval = dmi_coord_init( block, name, unit, cptNames, dim, ttype, pix, 
			   val, delt, dtype, 0, params );

  /* Attach coord descriptor to parent descriptor */
  dmi_descriptor_add_coord( dd, retval );

  if ( dim > 1 && ttype && !cs_uceq( ttype, "LINEAR" ))     
    dmi_coord_mink_coat( retval->aux, retval );

  dmi_descriptor_set_update(retval);

  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_init                                  */
/*                                                             */
/* Description:                                                */
/*   Initialize coordinate descriptor                          */
/*                                                             */
/*  Input parameters:                                          */
/*      dmBlock       *block     - block                       */
/*      char          *name      - coord descritpor name       */
/*      char          *unit      - coord units                 */
/*      char         **cptNames  - coord component names       */
/*      long           dim       - lenth of input arrays       */
/*      char          *ttype     -                             */
/*      void          *pix       - pixels                      */
/*      void          *val       - values                      */
/*      void          *delt      - scaling                     */
/*      dmDataType     dtype     - data type                   */
/*      long           groupno   - group number                */
/*      double        *params    -                             */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      dmDescriptor  retval - output coordinate descriptor    */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_coord_init(dmBlock *block, char *name, char *unit,
			     char **cptNames, long dim, char *ttype, 
			     void *pix, void *val, void *delt, 
			     dmDataType dtype,  long groupno, double *params)
{
  dmDescriptor     *retval  = NULL;    /* Output coordinate descriptor   */
  dmCoordTransform *wcsvals = NULL;    /* Value structure for WCS desc   */

  /* Unit values */
  double dpix[2] = { 0.0, 0.0 };
  double dval[2] = { 0.0, 0.0 };
  double ddlt[2] = { 1.0, 1.0 };

  /* Groupno not currently used */
  if( (dim > 1 && ( !cptNames || !ttype )) || groupno < 0 )
  {
    dmu_err_set(DM_ERR_NULL, "Null arg");
    return NULL;
  }

  /* Set default values */
  if( !pix  ) pix = dpix;
  if( !val  ) val = dval;
  if( !delt ) delt = ddlt;

  /* Create and populate descriptor */
  retval = dmi_coord_make_descriptor(block, name, dtype, unit, cptNames, dim);

  if (!retval)
    return NULL;

  /* Create and populate WCS value structure */
  wcsvals = dmi_coord_transform_make( NULL, ttype, pix, val, delt, dim, params);

  if (!wcsvals)
    return NULL;

  /* Attach value structure to descriptor */
  retval->aux = wcsvals;

  dmi_descriptor_clear_update(retval);

  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_set_transform                         */
/*                                                             */
/* Description:                                                */
/*   Set the values for the given coordinate descriptor.       */
/*                                                             */
/*  Input parameters:                                          */
/*      dmDescriptor  *dd        - descriptor                  */
/*      void          *crpix     - pixel coordinates           */
/*      dmDataType     dType     - data type of crpix array    */
/*      double        *crval     - coordinate value array      */
/*      double        *cdelt     - coordinate scaling array    */
/*      long           dim       - length of input arrays      */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      int  dmSUCCESS or dmFAILURE                            */
/*                                                             */
/***************************************************************/
int dmi_coord_set_transform(dmDescriptor *dd,   /* Coordinate descriptor    */
			    void *crpix,        /* Pixel coordinates        */
			    dmDataType dType,   /* Datatype of crpix array  */ 
			    double *crval,      /* Coordinate value array   */
			    double *cdelt,      /* Coordinate scaling array */
			    long dim)           /* Length of input arrays   */
{
  dmCoordTransform *wcsvals = NULL;
  double            dpix[2];


  if(!dd || !crpix || !crval || !cdelt)
  {
    dmu_err_set(DM_ERR_NULL, "Null arg");
    return dmFAILURE;
  }

  /* Perhaps we should set a warning here if dimensions don't match */
  /*  (once we get the warning mechanism in place...).              */

  if (dim < 1)
  {
    dmu_err_set(DM_ERR_BAD_DIM, "Coordinate system dimension is out of legal range [dmi_coord_set_transform].");
    return dmFAILURE;
  }

  if (dd->descType != dmCOORD)
  {
    dmu_err_set(DM_ERR_BAD_DESC_TYPE, "Not a coordinate descriptor" );
    return dmFAILURE;
  }

  if (!(wcsvals = dd->aux))
  {
    dmu_err_set(DM_ERR_INTERNAL, "No coord transform on descriptor" );
    return dmFAILURE;
  }

  /* Allocate a double array for pix array buffering */
  ar_mix_numeric_cast_array( crpix, dType, DT_REAL8, dpix, dim );
    
  /* Attach new cordinate values. */
  ar_copy_d(dpix, wcsvals->pix, dim);
  ar_copy_d(crval, wcsvals->val, dim);
  ar_copy_d(cdelt, wcsvals->delt, dim);

  dmi_coord_transform_check_is_unit( dd, 0 );
  dmi_coord_mink_coat(wcsvals, dd);

  dmi_descriptor_set_update(dd);

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_calc                                  */
/*                                                             */
/* Description:                                                */
/*   Apply transform to coordinate descriptor.                 */
/*                                                             */
/*  Input parameters:                                          */
/*  const dmDescriptor  *dd      - descriptor                  */
/*        void          *invals  - pixel coordinate data       */
/*        dmDataType     iType   - data type of input value    */
/*        dmDataType     oType   - data type of output value   */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*        void          *outvals - World Coord data            */
/*                                                             */
/*  Returns:                                                   */
/*      int  dmSUCCESS or dmFAILURE                            */
/*                                                             */
/***************************************************************/
int dmi_coord_calc( const dmDescriptor *dd,    /* Coordinate descriptor    */
		    void *invals,              /* Pixel coordinate data    */
		    dmDataType iType,          /* Input  value datatype    */ 
		    dmDataType oType,          /* Output value datatype    */
		    void *outvals )            /* World Coord data         */
{
  long    ii     = 0;              /* counter                          */
  long    dim    = 0;              /* counter                          */
  long    ndim   = 0;              /* dimensionality of input          */
  long    nvals  = 0;              /* total number of values           */
  long    nsets  = 0;              /* number of value sets             */
  long    offset = 0;              /* shift into array                 */

  double  dvalue[dmMAXDIM];        /* single input value set           */
  double  rvalue[dmMAXDIM];        /* single output value set          */

  dmDescriptor *mydd = NULL;      /* non-constant ptr                 */


  /* Check input validity */
  if (!dd || !invals || !outvals )
  {
    dmu_err_set(DM_ERR_NULL, "Null arg  [dmi_coord_calc]");
    return dmFAILURE;
  }

  mydd  = (dmDescriptor*)dd;
  ndim  = mydd->dim;
  nvals = dmi_descriptor_get_nvals( mydd );
  nsets = dmi_descriptor_get_array_size( mydd );

  /* Check easy case..                        */
  /* Non coords always are the unit transform */
  if (dd->descType != dmCOORD)
  {
    ar_mix_numeric_cast_array( invals, iType, oType, outvals, nvals );
    return dmSUCCESS;
  }

  /* Determine the number of value 'sets' to be transformed.           */
  /* It is assumed that each dimension is the same length,             */
  /*  e.g. vector of scalars => 1 set of (x,y) pairs                   */
  /*       vector of arrays  => N set of ( x[N], y[N] )                */
  /*                                                                   */
  /* NOTE..array  of vectors => N set of (x,y)[N] is NOT supported..   */
  /*                                                                   */

  for (ii = 0; ii < nsets; ii++ )
  {
    /* Xfer set from invals to dvalue                                  */
    /* provides a set of values of type double required by transform   */

    for ( dim = 0; dim < ndim; dim++)
    {
      offset = ii + dim*nsets;
      ar_mix_numeric_cast_elt( invals, iType, DT_REAL8, dvalue, offset, dim );
    }
  
    /*  Apply transform on this set. dvalue[] => rvalue[]              */
    dmi_coord_transform_apply( mydd, dvalue, rvalue, dim );

    /*  Xfer results from rvalue to outvals */

    for ( dim = 0; dim < ndim; dim++)
    {
      offset = ii + dim*nsets;
      ar_mix_numeric_cast_elt( rvalue, DT_REAL8, oType, outvals, dim, offset );
    }

  } /* end loop over sets */

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_calc_d                                */
/*                                                             */
/* Description:                                                */
/*   Apply transform to coordinate descriptor with double      */
/*   type values and results                                   */
/*   Calls dmi_coord_calc() to do work.                        */
/*                                                             */
/*  Input parameters:                                          */
/*      dmDescriptor  *dd      - descriptor                    */
/*      double        *values  - input values                  */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*        double      *results  -  output values               */
/*                                                             */
/*  Returns:                                                   */
/*      int  dmSUCCESS or dmFAILURE                            */
/*                                                             */
/***************************************************************/
int dmi_coord_calc_d(  dmDescriptor *dd, double *value, double *result )
{
  return dmi_coord_calc( dd, value, DT_REAL8, DT_REAL8, result );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_transform_apply                       */
/*                                                             */
/* Description:                                                */
/*    This function is very much like dmCoordCalc used to be   */
/*    in that it will apply the transform to the the provided  */
/*    data with the expectation that the input is a SINGLE set */
/*    of values.                                               */
/*                                                             */
/*  Input parameters:                                          */
/*      dmDescriptor  *dd      - descriptor                    */
/*      double        *invals  - input values                  */
/*      long           dim     - length of array values        */
/*                                                             */
/*  Output parameters:                                         */
/*        double      *outvals  -  output values               */
/*                                                             */
/*  Returns:                                                   */
/*      int  status  - dmSUCCESS or dmFAILURE                  */
/*                                                             */
/***************************************************************/
int dmi_coord_transform_apply(  dmDescriptor *dd, double *invals,
				double *outvals, long dim )
{
  long ii    = 0;
  int status = dmSUCCESS;
  dmCoordTransform *trans = NULL;

  /* Extract transform to apply */
  trans = (dmCoordTransform*)dd->aux;

  if (!trans)
  {
    dmu_err_set( DM_ERR_INTERNAL, "No coord transform" );
    return dmFAILURE;
  }

  if ( trans->linear )
  {
    for ( ii = 0; ii < dim; ii++ )
    {
      /* linear transform */
      outvals[ii] = trans->val[ii] +
                   trans->delt[ii] * ( invals[ii] - trans->pix[ii] );
    }
  }
  else
  {
    if (!trans->wcs )
    {
      /* Initialize */
      dmi_coord_mink_coat(trans, dd);

      if ( !trans->wcs )
      {
	dmu_err_set( DM_ERR_COORD,"Illegal projection [dmi_coord_calc]");
	return dmFAILURE;
      }
    }

    /* NOTE: dim should be 2 */
    pix2wcs(trans->wcs, invals[0], invals[1], &outvals[0], &outvals[1] );   /* WCSLIB */
  }

  return status;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_invert_d                              */
/*                                                             */
/* Description:                                                */
/*    Calls dmi_coord_invert() to perform work.                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*      dmDescriptor  *dd      - descriptor                    */
/*      double        *values  - input values                  */
/*                                                             */
/*  Output parameters:                                         */
/*        double      *result  -  output values                */
/*                                                             */
/*  Returns:                                                   */
/*      int  status  - dmSUCCESS or dmFAILURE                  */
/*                                                             */
/***************************************************************/
int dmi_coord_invert_d( dmDescriptor *dd, double *value, double *result )
{
  return dmi_coord_invert( dd, value, DT_REAL8, DT_REAL8, result );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_invert                                */
/*                                                             */
/* Description:                                                */
/*   Apply transform to coordinate descriptor.                 */
/*                                                             */
/*  Input parameters:                                          */
/*  const dmDescriptor  *dd      - descriptor                  */
/*        void          *invals  - World Coordinates           */
/*        dmDataType     iType   - data type of input value    */
/*        dmDataType     oType   - data type of output value   */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*        void          *outvals - resulting pixel coordinates */
/*                                                             */
/*  Returns:                                                   */
/*      int  dmSUCCESS or dmFAILURE                            */
/*                                                             */
/***************************************************************/
int dmi_coord_invert(const dmDescriptor *dd,    /* Coordinate descriptor   */
		     void *invals,              /* World Coordinates       */
		     dmDataType iType,          /* Datatype of value array */ 
                     dmDataType oType,          
		     void *outvals)             /* Resulting Pixel coords  */
{
  long    ii     = 0;              /* counter                          */
  long    dim    = 0;              /* counter                          */
  long    ndim   = 0;              /* dimensionality of input          */
  long    nvals  = 0;              /* total number of values           */
  long    nsets  = 0;              /* number of value sets             */
  long    offset = 0;              /* shift into array                 */
  int     offscl = 0;

  double  dvalue[dmMAXDIM];        /* single input value set           */
  double  rvalue[dmMAXDIM];        /* single output value set          */

  dmCoordTransform *t    = NULL;   /* Tranform                         */
  dmDescriptor     *mydd = NULL;   /* non-constant ptr                 */

  int retval = dmSUCCESS;


  if (!dd || !invals || !outvals )
  {
    dmu_err_set(DM_ERR_NULL, "Null arg"" [dmPCoordInvert]");
    return dmFAILURE;
  }

  offscl = 0;
  mydd   = (dmDescriptor*)dd;
  ndim   = mydd->dim;
  nvals  = dmi_descriptor_get_nvals( mydd );
  nsets  = dmi_descriptor_get_array_size( mydd );

  /* Check easy case..                        */
  /* Non coords always are the unit transform */

  if (dd->descType != dmCOORD)
  {
    ar_mix_numeric_cast_array( invals, iType, oType, outvals, ndim );
    return dmSUCCESS;
  }

  /* Extract transform to apply */
  t = (dmCoordTransform*)dd->aux;

  if (!t)
  {
    dmu_err_set(DM_ERR_COORD, "Problem in coord invert" );
    return dmFAILURE;
  }

  /* Determine the number of value 'sets' to be transformed.           */
  /* It is assumed that each dimension is the same length,             */
  /*  e.g. vector of scalars => 1 set of (x,y) pairs                   */
  /*       vector of arrays  => N set of ( x[N], y[N] )                */
  /*                                                                   */
  /* NOTE..array  of vectors => N set of (x,y)[N] is NOT supported..   */
  /*                                                                   */

  for (ii = 0; ii < nsets; ii++ )
  {
    /* Xfer set from invals to dvalue                                  */
    /* provides a set of values of type double required by transform   */

    for ( dim = 0; dim < ndim; dim++)
    {
      offset = ii + dim*nsets;
      ar_mix_numeric_cast_elt( invals, iType, DT_REAL8, dvalue, offset, dim );
    }

    /*  Apply transform on this set. dvalue[] => rvalue[]              */
    if ( t->linear )
    {
      for ( dim = 0; dim < ndim; dim++ )
      {
	if ( t->delt[dim] == 0.0 )
	{
	  rvalue[dim] = 0.0;
	}
	else
	{
	  rvalue[dim] = t->pix[dim] + ( dvalue[dim] - t->val[dim] )/(t->delt[dim]);
	}
      }
    }
    else
    {
      /* Initialize transform */
      if ( !t->wcs )
	dmi_coord_mink_coat(t, dd);

      if( !t->wcs )
      {
	dmu_err_set( DM_ERR_COORD,"Illegal Projection [dmi_coord_invert]");
	return dmFAILURE;
      }

      wcs2pix(t->wcs, dvalue[0], dvalue[1], &rvalue[0], &rvalue[1], &offscl); /* WCSLIB */

      if ( offscl )
      {
	retval = dmFAILURE;
      }

    } /* end non-linear transform */

    /* If converting to integers correct for truncation */
    if ( (iType != DT_REAL4) && (iType != DT_REAL8) )
    {
      for ( dim = 0; dim < ndim; dim++ )
      {
	rvalue[dim] += 0.5;
      }
    }

    /*  Xfer results from rvalue to outvals */
    for ( dim = 0; dim < ndim; dim++)
    {
      offset = ii + dim*nsets;
      ar_mix_numeric_cast_elt( rvalue, DT_REAL8, oType, outvals, dim, offset );
    }

  } /* end loop over sets */

  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_mink_coat                             */
/*                                                             */
/* Description:                                                */
/*   Code to wrap Doug Mink's WCS library init call.           */
/*                                                             */
/*  Input parameters:                                          */
/*         dmCoordTransform *wcsvals - wcs vals                */
/*   const dmDescriptor     *dd      - descripor               */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      int  dmSUCCESS or dmFAILURE                            */
/*                                                             */
/***************************************************************/
int dmi_coord_mink_coat(dmCoordTransform *wcsvals, const dmDescriptor *dd)
{
  struct WorldCoor *wcs = NULL;	/* World coordinate system structure */

  char   c1type[dmMAXSTR];
  char   c2type[dmMAXSTR];

  double *dcrval = NULL;
  double *dcrpix = NULL;
  double *dcdelt = NULL;
  double *cd     = NULL;
  double  rotate = 0.0;
  double  epoch  = 0.0;

  int equinox = 0;
  int nxpix   = 0;
  int nypix   = 0;


  if ( wcsvals->wcs ) 
  {
    wcsfree( wcsvals->wcs );
    wcsvals->wcs = NULL;
  }

  if ( wcsvals->linear )
    return dmSUCCESS;

  if ( dd->dim != 2 ) 
    return dmSUCCESS;

  /* NOTE:                                                     */
  /* The nxpix and nypix are not used by the routines we call, */
  /* so their values are not crucial. However, here is a       */
  /* 'best effort' to get reasonable values.                   */

  rotate  = wcsvals->theta;

  equinox = wcsvals->equinox;

  epoch=0.0; /* We don't bother with this */
  
  dcrpix = (double *)wcsvals->pix;
  dcrval = (double *)wcsvals->val;
  dcdelt = (double *)wcsvals->delt;

  nxpix = INT_MAX;
  nypix = INT_MAX; /* We don't care */

  dmi_wcs_make_ctype(  dd->cptNames[0], wcsvals->type, 1, c1type, UT_TRUE );
  dmi_wcs_make_ctype(  dd->cptNames[1], wcsvals->type, 2, c2type, UT_TRUE );

  // Function wcskinit() sets up a basic WorldCoord structure
  // It auto-guesses the RADESYS frame based on the value of EQUINOX
  // If EQUINOX < 1981, RADESYS is set as FK4
  // If EQUINOX >= 1981, RADESYS is set as FK5
  // if EQUINOX == 0, RADESYS set as FK4
  // If the input transform descriptor has RADESYS = "ICRS", the structure
  //   values need to be changed below. Also, longpole and latpole values
  //   need to be manually set.

  wcs = wcskinit(nxpix,nypix,c1type,c2type,dcrpix[0],dcrpix[1],dcrval[0],
                 dcrval[1],cd,dcdelt[0],dcdelt[1],rotate,equinox,epoch);

  // add WorldCoord structure pointer to the dmCoordTransform
  wcsvals->wcs  = wcs;

  // Manually set values of longpole and latpole in WCS structure

  wcs->longpole = wcsvals->longpole;
  wcs->latpole  = wcsvals->latpole;

  if ( wcsvals->longpole != 0.0 )
    wcs->cel.ref[2] = wcsvals->longpole;

  if ( wcsvals->latpole != 0.0 )
    wcs->cel.ref[3] = wcsvals->latpole;

  // Insure correct ICRS frame set up, since wcskinit() doesn't
  //  seem to recognize this

  if (cs_uceq(wcsvals->radesys, "ICRS"))
  {
    // set system frame number
    wcs->syswcs = WCS_ICRS;
    wcs->sysin  = WCS_ICRS;
    wcs->sysout = WCS_ICRS;

    // set equinox to 0
    wcs->equinox = 0.0;
    wcs->eqin    = 0.0;
    wcs->eqout   = 0.0;

    // set radesys string values
    cs_copy( wcsvals->radesys, wcs->radecsys );
    cs_copy( wcsvals->radesys, wcs->radecin );
    cs_copy( wcsvals->radesys, wcs->radecout );
  }

  // Other frames should be determined ok with supplied equinox values
  // However, double check that the RADESYS selected by wcskinit()
  // matches the transform descriptor RADESYS. If not, manually correct
  // wcs element values.

  if (! cs_uceq(wcsvals->radesys, wcs->radecsys) )
  {
    // RADESYS value differ, adjust according to frame type

    // set radesys string values
    cs_copy( wcsvals->radesys, wcs->radecsys );
    cs_copy( wcsvals->radesys, wcs->radecin );
    cs_copy( wcsvals->radesys, wcs->radecout );

    // wcs->equinox is assigned the value of variable "equinox"
    // just adjust value of "eqin" to be the same
    // "eqout" is 0

    wcs->eqin = wcs->equinox;

    // insure correct frame number
    if ( cs_uceq(wcsvals->radesys, "FK5"))
    {
      wcs->syswcs = WCS_J2000;
      wcs->sysin  = WCS_J2000;
      wcs->sysout = WCS_J2000;
    }
    else
      if (cs_uceq_siz(wcsvals->radesys, "FK4", 3))
      {
        wcs->syswcs = WCS_B1950;
        wcs->sysin  = WCS_B1950;
        wcs->sysout = WCS_B1950;
      }
  }

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_wcs_make_ctype                              */
/*                                                             */
/* Description:                                                */
/*   This routine constructs the CTYPE argument expected by    */
/*   wcslib.                                                   */
/*                                                             */
/*  Input parameters:                                          */
/*         char    *name                                       */
/*         char    *ttype                                      */
/*         long     subaxis                                    */
/*         int      wcslib                                     */
/*                                                             */
/*  Output parameters:                                         */
/*         char    *ctype                                      */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      int  used_name  -  dmSUCCESS or dmFAILURE              */
/*                                                             */
/***************************************************************/
int dmi_wcs_make_ctype( char *name, char *ttype, long subaxis, 
			char *ctype, int wcslib )
{
  char  transform[CARD_SIZE];
  char *tanp_name[2]     = { "LONG", "NPOL" };
  char *ctype_default[2] = { "XLON", "XLAT" };

  long  k = 0;
  int   wtype     = UT_FALSE;
  int   used_name = UT_TRUE;


  /* For FITS, all linear transforms just use their name */
  strcpy( ctype, name );

  if ( ttype && cs_eq( ttype, "TAN-P" ))
  {
    strcpy( ctype, tanp_name[subaxis-1] );
    wtype = UT_TRUE;
    used_name = UT_FALSE;
  } 
  else if ( ttype && strlen( ttype ) == 3 ) 
  {
    /* We recognize WCS projection transforms from the */
    /* fact that all have names of length 3 bytes.     */
    /* Only certain names are allowed.                 */

    if ( !dmi_wcs_standard_axis_name( ctype ) ) 
    {
      strcpy( ctype, ctype_default[subaxis-1] );  
      used_name = UT_FALSE;
    }
    wtype = UT_TRUE;
  }

  if ( wtype ) 
  {
    cs_copy_siz( ttype, transform, 3 );

    k = strlen( ctype );

    while ( k < 5 ) 
      ctype[k++] = '-';

    ctype[k] = '\0';
 
   strcat( ctype, transform );

  }
  else if ( wcslib ) 
  {
    /* For WCSLIB, all linear transforms are renamed LINEAR */
    strcpy( ctype, "LINEAR" );
  }

 return used_name;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_wcs_standard_axis_name                      */
/*                                                             */
/* Description:                                                */
/*   Enumerates the recognized WCS axis name.                  */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         const char  *name                                   */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      static int  - enumerator                               */
/*                                                             */
/***************************************************************/
static int dmi_wcs_standard_axis_name( const char *name )
{
  if ( !strcmp( name, "RA" )) return 1;
  if ( !strcmp( name, "DEC")) return 1;
  if ( !strcmp( name+1, "LON" )) return 1;
  if ( !strcmp( name+1, "LAT" )) return 1; 

  return 0;
}   


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_make_descriptor                       */
/*                                                             */
/* Description:                                                */
/*   Create a coordinate descriptor and populate all the       */
/*   generic descriptor fields.                                */
/*                                                             */
/*   Generic descriptor fields are populated separately from   */
/*   descriptor specific info with the hope that we will one   */
/*   day unite all the dmPMake[..]Descriptor functions into    */
/*   one.                                                      */
/*                                                             */
/*  Input parameters:                                          */
/*         char        *name     - descriptor name             */
/*         dmDataType   dtype    - data type, double or long   */
/*         char        *unit     - unit                        */
/*         char       **cptNames - coordinate axes names       */
/*         long         dim      - transform dimension         */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      dmDescriptor *dd                                       */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_coord_make_descriptor(dmBlock *block, /* Parent block    */
		       char *name,         /* Descriptor name              */
		       dmDataType dtype,   /* Datatype, double or long     */
		       char *unit,         /* Unit                         */
		       char **cptNames,    /* Names of coordinate axes     */
		       long dim)           /* Dimension of transform       */
{

  long i  = 0;
  long no = 0;
  long naxes = 1;

  dmDescriptor *dd = NULL;


  dd = dmi_descriptor_alloc( block, dmCOORD, no, name, dtype, 
		     unit, NULL, dim, dmVALUE, naxes, NULL, DM_FLAG_COORD ); 

  if ( dim == 1 && !cptNames )
    cptNames = &name;

  for (i = 0; i < dim; i++)
    dd->cptNames[i] = cs_dup(cptNames[i]);

  dmi_vector_make_cpts( dd, dim );

  return dd;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_transform_make                        */
/*                                                             */
/* Description:                                                */
/*   Create and populate a WCS value structure, which contains */
/*   information specific to coordinate type descriptors.      */
/*   Input descriptor is parent descriptor of coordinate       */
/*   descriptor.                                               */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor  *dd     - descriptor                  */
/*         char          *ttype  - data type                   */
/*         double        *pix    - pixels                      */
/*         void          *val    - values                      */
/*         void          *delt   -                             */
/*         long           dim    - transform dimension         */
/*         double        *params -                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      dmCoordTransform  *transform                           */
/*                                                             */
/***************************************************************/
dmCoordTransform* dmi_coord_transform_make(dmDescriptor *dd, char *ttype, 
					   double *pix, void *val, void *delt,
					   long dim, double *params)
{
 dmCoordTransform *transform = NULL;

 transform = dmi_coord_transform_alloc( dd, ttype, val, pix, delt, dim );

 if ( params && !cs_uceq( ttype, "LINEAR" ))
 {
   transform->theta    = params[0];
   transform->longpole = params[1];
   transform->latpole  = params[2];
   transform->equinox  = params[3];
 }

 if ( dd )
   dmi_coord_mink_coat(transform, dd);

 return transform;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_transform_free                        */
/*                                                             */
/* Description:                                                */
/*   Free a WCS value structure.                               */
/*   Free char strings and use wcsfree() to release memory     */
/*   used in WorldCoord struct wcs                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmCoordTransform  *transform                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      void                                                   */
/*                                                             */
/***************************************************************/
void dmi_coord_transform_free( dmCoordTransform *transform )
{
  if ( transform )
  {
    // free char strings
    if ( transform->type )
      free( transform->type );
    if ( transform->name )
      free( transform->name );
    if ( transform->radesys )
      free( transform->radesys );

    // free WorldCoord structure
    if (transform->wcs)
      wcsfree( transform->wcs );

    free( transform );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_transform_is_unit                     */
/*                                                             */
/* Description:                                                */
/*   Is the coordinate a unit transform? If empty, answer yes  */
/*   (e.g. case of image logical descriptor)                   */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor  *dd                                   */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      int  dmTRUE or dmFALSE                                 */
/*                                                             */
/***************************************************************/
int dmi_coord_transform_is_unit( dmDescriptor *dd )
{
  dmCoordTransform *transform = NULL;
 
  transform = dmi_descriptor_get_aux( dd );

  if ( transform )
    return (transform->scaled == 0);
 
  return 1;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_transform_check_is_unit               */
/*                                                             */
/* Description:                                                */
/*   Checks if the coord is a unit (identity) physical         */
/*   transform.                                                */
/*   Used to suppress output of trivial phys transforms.       */
/*   Note:  sets value of dd->aux->scaled                      */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor  *dd  - column descriptor              */
/*         long           dim - coponent number                */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      int  dmTRUE or dmFALSE                                 */
/*                                                             */
/***************************************************************/
int dmi_coord_transform_check_is_unit( dmDescriptor *dd, long dim )
{
  dmDescriptor     *cpt_dd  = NULL;
  dmCoordTransform *wcsvals = NULL;

  long i = 0;
  int  isUnit = 1;

  // Code Change  Bug 12880  January 10, 2012
  //  Use component descriptor
  cpt_dd  = dd->cpts[dim];

  if (!cpt_dd )
    return isUnit;

  wcsvals = cpt_dd->aux;

  if ( wcsvals != NULL)
  {
    for ( i = 0; i < cpt_dd->dim; i++ )
    {
      if ( ((double*)wcsvals->delt)[i] != 1.0 )
      {
        isUnit = 0;
        break;
      }

      if ( ((double*)wcsvals->val)[i] != ((double*)wcsvals->pix)[i]) 
      {
        isUnit = 0;
        break;
      }
    }

    wcsvals->scaled = isUnit ? UT_FALSE : UT_TRUE;
  }

  return isUnit;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descrirptor_silent_groups                   */
/*                                                             */
/* Description:                                                */
/*   Do we have only silent (i.e. default) groups?             */
/*   We return yes if all groups are default, are unit         */
/*   transforms, and have no attached coordinate systems.      */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor  *dd                                   */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      int  dmTRUE or dmFALSE                                 */
/*                                                             */
/***************************************************************/
int dmi_descriptor_silent_groups( dmDescriptor *dd )
{
  long i = 0;
  dmDescriptor *group = NULL;

  if ( dd->ngroups == 0 )
    return UT_TRUE;

  for ( i = 1; i <= dd->ngroups; i++ )
  {
    group = dmi_descriptor_get_group_no( dd, i );

    if ( group && group->ncoords > 0 )
      return UT_FALSE;

    /* Right now, suppress only the ones called "#COL..." */
    if ( group && !cs_eq_siz( "#COL", group->name, 4 ))
      return UT_FALSE;
  }

  return UT_TRUE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_is_default_axis                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor  *dd                                   */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      int  dmTRUE or dmFALSE                                 */
/*                                                             */
/***************************************************************/
int dmi_coord_is_default_axis( dmDescriptor *dd ) 
{    
  int ok = UT_FALSE;

  ok = dmi_coord_is_group( dd );

  if ( ok )
  {
    ok = !strncmp( "AXIS", dd->name, 4 )  ||
         !strncmp( "#AXIS", dd->name, 5 ) || 
         !strcmp( "X", dd->name )   || 
         !strcmp( "Y", dd->name )   || 
         !strcmp( "Z", dd->name )   || 
         !strcmp( "POS", dd->name ) ||
         !strncmp( "#COL", dd->name, 4 );
  }

 return ok;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_is_group                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor  *dd                                   */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*      int  dmTRUE or dmFALSE                                 */
/*                                                             */
/***************************************************************/
int dmi_coord_is_group( dmDescriptor *dd ) 
{
  dmCoordTransform *transform = NULL;
  int retval = UT_FALSE;

  if ( dd && dd->descType == dmCOORD )
  {    
    transform = dmi_descriptor_get_aux( dd );
    retval    = (transform && transform->no == 0 && dd->no > 0);
  }

  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_params                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor  *dd                                   */
/*                                                             */
/*  Output parameters:                                         */
/*         double        *params                               */
/*                                                             */
/*  Returns:                                                   */
/*      int  np  - number of params                            */
/*                                                             */
/***************************************************************/
long dmi_coord_get_params( dmDescriptor *dd, double *params )
{
  dmCoordTransform* t;
  long np = 0;

  if ( dd && dd->descType == dmCOORD )
  {
    t = dd->aux;
    if ( t )
    {
      params[0] = t->theta;
      params[1] = t->longpole;
      params[2] = t->latpole;
      params[3] = t->equinox;
      np = 4;
    }
  }

  return np; 
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_transform_cast                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor  *dd                                   */
/*         dmDataType     type                                 */
/*         double        *crval                                */
/*         double        *cdelt                                */
/*         long           dim                                  */
/*                                                             */
/*  Output parameters:                                         */
/*         void          *crpix                                */
/*                                                             */
/*  Returns:                                                   */
/*      int     dmSUCCESS                                      */
/*                                                             */
/***************************************************************/
int dmi_coord_get_transform_cast( dmDescriptor *dd,  void *crpix, 
 	        dmDataType type, double *crval, double *cdelt, long dim)
{
  double dcrpix[2];

  dmi_coord_get_transform_d( dd, dcrpix, crval, cdelt, dim );

  ar_mix_numeric_cast_array( dcrpix, DT_REAL8, type, crpix, dim );

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_list_copy                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor  *col1 - descriptor                    */
/*         dmDescriptor  *col2 - parent                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*      int     dmSUCCESS                                      */
/*                                                             */
/***************************************************************/
int dmi_coord_list_copy( dmDescriptor *col1, dmDescriptor *col2 )
{
  long coordNo = 0;
  long groupNo = 0;
  dmDescriptor  *dd = NULL;


  for ( groupNo = 1; groupNo <= col1->ngroups; groupNo++ )
  {
    dd = dmi_descriptor_copy( (dmBlock*)col2->block, col2, 
			                           col1->groups[groupNo-1] );
    if ( !dd )
       return dmFAILURE;

    dmi_descriptor_set_axis( col2, groupNo, dd );
  }

  for ( coordNo = 1; coordNo <= col1->ncoords; coordNo++ )
  {
    dd  = dmi_descriptor_copy( (dmBlock*)col2->block, col2, 
 			                           col1->coords[coordNo-1] );
    if ( !dd )
      return dmFAILURE;

    dmi_descriptor_set_coord( col2, coordNo, dd );
  }

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_transform_set_scaled                  */
/*                                                             */
/* Description:                                                */
/*        Used to set or clear transform scaled flag element   */
/*                                                             */
/*  Input parameters:                                          */
/*         dmCoordTransform  *transform - descriptor           */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*              void                                           */
/*                                                             */
/***************************************************************/
void dmi_coord_transform_set_scaled( dmCoordTransform *transform, long flag )
{
  transform->scaled = flag;
   
}
