/*                                                                
**  Copyright (C) 2002-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmidd.c                                                     */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"
#include "dmlibP.h"

#define DEBUG 0
#define XDEBUG 1

/* Basic descriptor routines */
dmDescriptor* dmi_parent_coord_cpt( dmDescriptor* dd, long coordno );
long dmi_descriptor_calc_size( long* axes, long naxes );


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_alloc                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmBlock          *block                           */
/*     const dmDescriptorType  descType                        */
/*     const long              no                              */
/*     const char             *name                            */
/*     const dmDataType        type                            */
/*     const char             *unit                            */
/*     const char             *desc                            */
/*     const long              dim                             */
/*     const dmElementType     etype                           */
/*     const long              naxes                           */
/*     const long             *axes                            */
/*           long              flag                            */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*              dmDescriptor *dd                               */
/*                                                             */
/***************************************************************/
dmDescriptor*  dmi_descriptor_alloc( dmBlock* block,
                                     const dmDescriptorType descType,
                                     const long no, const char* name,
                                     const dmDataType type,
                                     const char* unit, const char* desc,
                                     const long dim,
                                     const dmElementType etype,
                                     const long naxes,
                                     const long* axes, long flag )
{
  dmDescriptor *dd = NULL;
  long          m  = 0;


  dd = (dmDescriptor*)calloc( 1, sizeof( dmDescriptor ));

  if ( dd )
  {
    dd->block = (dmBlock*)block;  /* discard const */
    dd->descType = descType;
    dd->no = no;
    dd->bufno = 0;
    dd->name = cs_dup( name );
    dd->eltNames = NULL;
    dd->cptNames = calloc( dim, UT_SZ_P );
    dd->cpts =     calloc( dim, UT_SZ_P );
    dd->type = type;
    dd->etype = etype; 
    dd->cptno = 0;
    dd->edim = dmu_etype_dim( etype );
    dd->kernel_dd = NULL;  
    dd->disp = NULL;
    dd->desc = NULL;

    if ( unit )
      dd->unit = cs_dup( unit );

    if ( desc )
      dd->desc = cs_dup( desc );

    dd->length = 0;
    dd->dim = dim;
    dd->vdim = dim/dd->edim;
    dd->naxes = naxes;
    dd->varlen = dmFALSE;
    dd->ncoords = 0; 
    dd->ngroups = 0; 
    dd->groups = NULL;
    dd->parent = NULL;
    dd->coords = NULL;
    dd->aux = NULL;
    dd->flag = flag;
    dd->range = NULL;


    if ( naxes > 0 )
    {
      long i = 0;

      dd->axes = (long*)calloc( naxes, UT_SZ_I );

      for ( i = 0; i<naxes; i++ )
      {
        if ( axes )
        {
          m = axes[i];
        }
        else
        {
          m = 1;
        }

        dd->axes[i] = m;
      }

      dd->size = dmi_descriptor_calc_size( dd->axes, dd->naxes );
    }
  }
  return dd;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_free                             */
/*                                                             */
/* Description:                                                */
/*   Free a descriptor. We assume that its components and      */
/*   other descendants are already freed.  We check that its   */
/*   ancestors no longer point to it.                          */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor  *dd                                 */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_descriptor_free( dmDescriptor* dd )
{
  dmCoordTransform *t      = NULL;
  dmDescriptor     *parent = NULL;

  if ( !dd )
    return;

  switch( dd->descType )
  {
    case dmKEY:
      dmi_key_free_aux( dd );
      break;

    case dmCOORD: 
      if ( dd->aux )
      {
        t = dd->aux;
        parent = t->parent;

        if ( t->no > 0 && t->no < parent->ncoords && parent->coords[t->no-1] == dd )
          parent->coords[t->no-1] = NULL;

        parent = dd->parent;

        if ( dd->no > 0 && ( dd->flag & DM_FLAG_AXIS ) && ( dd->flag | DM_FLAG_LOGICAL) 
             && parent &&  parent->groups && parent->groups[dd->no-1] == dd )
          parent->groups[dd->no-1] = NULL;

        dmi_coord_transform_free( dd->aux );
      }

    case dmCOLUMN:
    default:
      break;
  }

  parent = dd->parent;

  if ( parent && dd->cptno > 0 && dd->cptno <= parent->dim )
  {
    parent->cpts[dd->cptno-1] = NULL;
  }

  if ( dd->vdim > 0 && dd->eltNames )
    ar_free_cp( dd->eltNames, dd->vdim );  

  if( dd->dim > 0 && dd->cptNames )
  {
    ar_free_cp( dd->cptNames, dd->dim );
  }

  free( dd->coords );
  free( dd->groups );
  free( dd->cpts );
  free( dd->bin );
  free( dd->null );
  free( dd->axes );

  if ( dd->range )
    free( dd->range );

  free( dd->name );
  free( dd->disp );
  free( dd->desc );
  free( dd->unit ); 
  free( dd );
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_calc_size                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           long   *axes                                      */
/*           long    naxes                                     */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            long  size                                       */
/*                                                             */
/***************************************************************/
long dmi_descriptor_calc_size( long* axes, long naxes )
{
  long i    = 0;
  long size = 1;

  for ( i = 0; i < naxes; i++ )
    size *= axes[i];

  return size;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_kernel_free                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor  *dd                                 */
/*           void           (*kernel_free)(dmDescriptor*)      */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_descriptor_kernel_free( dmDescriptor* dd,
                                 void (*kernel_free)( dmDescriptor* ))
{
  long i = 0;

  if ( !dd )
    return;

  for ( i = 0; i < dd->dim; i++ )
  {
    if ( dd->cpts[i] && dd->cpts[i] != dd )
    {
      dmi_descriptor_kernel_free( dd->cpts[i], kernel_free );
    }
  }

  for ( i = 0; i < dd->ncoords; i++ )
  {
    if ( dd->coords[i] && dd->coords[i] != dd )
      dmi_descriptor_kernel_free(  dd->coords[i], kernel_free );
  }

  for ( i = 0; i < dd->ngroups; i++ )
  {
    if ( dd->groups[i] && dd->groups[i] != dd )
      dmi_descriptor_kernel_free( dd->groups[i], kernel_free );
  }

  if ( kernel_free != NULL )
    kernel_free( dd );

  dmi_descriptor_free( dd );
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_check_groups                     */
/*                                                             */
/* Description:                                                */
/*     Check we have the right groups on a nontrivial image    */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor  *dd                                 */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            int True or False                                */
/*                                                             */
/***************************************************************/
int dmi_descriptor_check_groups( dmDescriptor* dd )
{
  dmDescriptor *group = NULL;
  long          naxes = 0;
  long          size  = 0;
  long          i     = 0;

  if ( !dd || dd->naxes == 0 )
    return UT_TRUE;

  for ( i = 0; i < dd->ngroups; i++ )
  {
    group = dd->groups[i];

    if ( group )
      naxes += group->dim;
  }

  if ( naxes == dd->naxes )
    return UT_TRUE;

  size = dmi_descriptor_calc_size( dd->axes, dd->naxes );

  if ( size == 1 )
    return UT_TRUE;

  return UT_FALSE;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_copy                             */
/*                                                             */
/* Description:                                                */
/*    Recursive copy                                           */
/*                                                             */
/*  Input parameters:                                          */
/*           dmBlock       *blockp                             */ 
/*           dmDescriptor  *nparent                            */
/*           dmDescriptor  *dd                                 */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor *vector                             */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_copy( dmBlock* blockp,
                                   dmDescriptor* nparent,
                                   dmDescriptor* dd )
{
  dmBlock      *block  = blockp;
  dmDescriptor *cpt_dd = NULL;
  dmDescriptor *vector = NULL;
  dmDescriptor *ccoord = NULL;

  double tmp[2];

  void *tlmin = &tmp[0];
  void *tlmax = &tmp[1];
  void *val   = &tmp[0];

  long i    = 0;
  long flag = 0;


  if ( !dd )
    return NULL;

  if ( !block ) 
    block = (dmBlock*)dd->block;

  flag = dd->flag | DM_FLAG_COPY;

#if DEBUG
 printf( "DDC copy %s\n", dd->name );
#endif

  vector = dmi_descriptor_alloc( block, dd->descType, dd->no, dd->name, dd->type, dd->unit, 
                            dd->desc, dd->dim, dd->etype, dd->naxes, dd->axes, flag ); 

  dmi_descriptor_set_update( vector );
  vector->parent = nparent;
  vector->disp = cs_dup( dd->disp );
  vector->length = dd->length;
  vector->varlen = dd->varlen;

  if ( nparent )
    vector->cptno = dd->cptno;
  else
    vector->cptno = 0;

  if ( !vector->cptNames )
    vector->cptNames = calloc( dd->dim, UT_SZ_P );


  for ( i = 0; i < dd->dim; i++ )
    vector->cptNames[i] = cs_dup( dd->cptNames[i] );

  if ( dd->edim > 1 && dd->eltNames )
  {
    vector->eltNames = calloc( dd->vdim, UT_SZ_P );

    for ( i = 0; i < dd->vdim; i++ )
      vector->eltNames[i] = cs_dup( dd->eltNames[i] );
  }

  /* Copy coords */
  for ( i = 0; i < dd->ncoords; i++ )
  {
     cpt_dd = NULL;
     ccoord = dmi_parent_coord_cpt( dd, i+1 );   

     if ( ccoord && ccoord == dd->coords[i] ) 
       cpt_dd = dmi_parent_coord_cpt( vector, i+1 );

     if ( !cpt_dd )
     {
#if DEBUG
 printf( "DDC %s coord %ld = %s\n", dd->name, i+1, dd->coords[i]->name );
#endif
       cpt_dd = dmi_descriptor_copy( block, NULL, dd->coords[i] );   
     }
      
     dmi_descriptor_add_coord( vector, cpt_dd );
   }

  /* Copy cpts */
  if ( !vector->cpts )
    vector->cpts = calloc( dd->dim, UT_SZ_P );

  if ( dd->dim > 1 )
  {
    for ( i = 0; i < dd->dim; i++ )
    {
#if DEBUG
 if ( dd->cpts[i])
 printf( "DDC %s cpt %ld = %s\n", dd->name, i+1, dd->cpts[i]->name );
else
 printf( "NO CPT %ld FOR DDC %s\n", i+1, dd->name );
#endif
      cpt_dd = dmi_descriptor_copy( block, vector, dd->cpts[i] );
      vector->cpts[i] = cpt_dd;
    }
  }
  else
  {
    vector->cpts[0] = vector;
  }

  /* Copy groups */
  vector->ngroups = dd->ngroups;

  if ( dd->ngroups > 0 )
    vector->groups = calloc( vector->ngroups, UT_SZ_P );

  for ( i = 1; i <= dd->ngroups; i++ )
  {
#if DEBUG
 printf( "DDC %s group %ld = %s\n", dd->name, i, dd->groups[i-1]->name );
#endif
    if ( dd->groups[i-1] )
      cpt_dd = dmi_descriptor_copy( block, NULL, dd->groups[i-1] );
    else
     cpt_dd = NULL;

    dmi_descriptor_set_axis( vector, i, cpt_dd );
  }

  /* Copy ranges and nulls - does not flush to file */
  if ( dmi_descriptor_get_range( dd, tlmin, tlmax ) )
    dmi_descriptor_set_cpt_range( vector, 0, dd->type, tlmin, tlmax );

  if ( dmi_descriptor_get_null( dd, val ))
    dmi_descriptor_set_null( vector, dd->type, val );

  if ( dmi_descriptor_get_bin( dd, val ))
    dmi_descriptor_set_bin( vector, dd->type, val );

  /* Copy kernel dd */
  dmi_descriptor_copy_kdd( dd, vector );

  /* Copy transform */
  if ( vector->descType == dmCOORD )
    dmi_coord_transform_copy( vector, dd->aux, dd->dim );

#if DEBUG
printf( "SET UPDATE FOR %p %s\n", vector, vector->name );
#endif

  dmi_descriptor_set_update( vector );
  return vector;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_copy_kdd                         */
/*                                                             */
/* Description:                                                */
/*     Copy kernel descriptor of dmDescriptor dd1 to           */
/*     dmDescriptor dd2, but only if the kernels are the same. */
/*     Note that, depending on the kernel and descriptor type, */
/*     this function may do nothing at all, even if the source */
/*     kernel dd is non-NULL.                                  */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor  *dd1                                */
/*           dmDescriptor  *dd2                                */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_descriptor_copy_kdd( dmDescriptor* dd1, dmDescriptor* dd2 )
{
 dmDataset *ds1  = NULL;
 dmDataset *ds2  = NULL;
 dmKernel   kernel2;

  ds1 = dd1->block->ds;
  ds2 = dd2->block->ds;

  if ( !dmk_kernel_eq( ds1, ds2 ) )
    return;

  kernel2 = ds2->kernel;

  if ( kernel2->DescriptorCopy )
    kernel2->DescriptorCopy( dd1, dd2 );
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_name                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            char *name                                       */
/*                                                             */
/***************************************************************/
char* dmi_descriptor_get_name( const dmDescriptor* dd )
{
  static char* null = " ";

  if ( !dd )
  {
    return null;
  }

  return dd->name;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_datatype                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            dmDataType type                                  */
/*                                                             */
/***************************************************************/
dmDataType dmi_descriptor_get_datatype( const dmDescriptor* dd )
{
  if ( !dd )
    return dmUNKNOWNTYPE;

  return dd->type;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_element_type                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            dmElementType type                               */
/*                                                             */
/***************************************************************/
dmElementType dmi_descriptor_get_element_type( const dmDescriptor* dd )
{
  if ( !dd )
   return dmVALUE;

  return dd->etype;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_element_dim                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            long  dim                                        */
/*                                                             */
/***************************************************************/
long dmi_descriptor_get_element_dim( const dmDescriptor* dd )
{
  if ( !dd )
    return 0;

  return dd->dim;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_vector_dim                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            long  vdim                                       */
/*                                                             */
/***************************************************************/
long dmi_descriptor_get_vector_dim( const dmDescriptor* dd )
{
  if ( !dd )
    return 0;

  return dd->vdim;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_array_dim                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            long  naxes                                      */
/*                                                             */
/***************************************************************/
long dmi_descriptor_get_array_dim( const dmDescriptor* dd )
{
  if ( !dd )
    return 0;

  return dd->naxes;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_array_dimensions             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*           long         **axisLens                           */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            long  naxes                                      */
/*                                                             */
/***************************************************************/
long dmi_descriptor_get_array_dimensions( const dmDescriptor* dd, long** axisLens )
{
  long i = 0;

  if ( !axisLens )
    return 0;

  *axisLens = NULL;

  if ( !dd )
    return 0;
 
  if ( dd->naxes < 1 )
    return 0;

  *axisLens = (long*)calloc( dd->naxes, UT_SZ_I );

  if ( !(*axisLens) )
  { 
    dmu_err_set( DM_ERR_MEM, "Memory alloc failed" );
    return 0;
  }

  for ( i = 0; i < dd->naxes; i++ )
    (*axisLens)[i] = dd->axes[i];

  return dd->naxes;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_array_size                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            long  size                                       */
/*                                                             */
/***************************************************************/
long dmi_descriptor_get_array_size( const dmDescriptor* dd )
{
  long i    = 0;
  long size = 1;

  if ( !dd )
    return 0;

  if ( dd->naxes == 0 )
    return 0;

  for ( i = 0; i < dd->naxes; i++ )
    size *= dd->axes[i];

  return size;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_length                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            long  length                                     */
/*                                                             */
/***************************************************************/
long dmi_descriptor_get_length( const dmDescriptor* dd )
{
  if ( !dd )
    return 0;

  return dd->length;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_desc                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            char *desc                                       */
/*                                                             */
/***************************************************************/
char* dmi_descriptor_get_desc( const dmDescriptor* dd )
{
  if ( !dd )
    return NULL;

  return dd->desc;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_disp                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            char *disp                                       */
/*                                                             */
/***************************************************************/
char* dmi_descriptor_get_disp( const dmDescriptor* dd )
{
  if ( !dd )
    return NULL;

  return dd->disp;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_unit                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            char *unit                                       */
/*                                                             */
/***************************************************************/
char* dmi_descriptor_get_unit( const dmDescriptor* dd )
{
  if ( !dd )
    return NULL;

  return dd->unit;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_check                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            int  status                                      */
/*                                                             */
/***************************************************************/
int dmi_descriptor_check( const dmDescriptor* dd )
{
  int status = UT_TRUE;

  if ( dd->length < 0 )
  {
    sprintf( dm_errbuf, "%s: Negative descriptor length not allowed", dd->name );
    dmu_err_set( DM_ERR_INTERNAL, dm_errbuf ); 
    status = UT_FALSE;
  }

  /* Zero size is allowed for variable-length arrays: no DMP level check for this? */
  if ( dd->size < 0 )
  {
    sprintf( dm_errbuf, "%s: Negative or zero array length not allowed", dd->name );
    dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
    status = UT_FALSE;
  }

  if ( dd->type == DT_CHAR && dd->length <= 0 )
  {
    sprintf( dm_errbuf, "%s: String length must be at least 1", dd->name );
    dmu_err_set( DM_ERR_INTERNAL, dm_errbuf ); 
    status = UT_FALSE;
  }

  return status;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_cpt_name                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            char *cptNames                                   */
/*                                                             */
/***************************************************************/
char* dmi_descriptor_get_cpt_name( const dmDescriptor* dd, const long cptNo )
{
  if ( !dd || !dd->cptNames )
    return NULL;

  if ( cptNo < 1 || cptNo > dd->dim )
  {
    dmu_err_set( DM_ERR_RANGE, "[dmGetCptName]: range error" );
    return NULL;
  }

 return dd->cptNames[cptNo-1];
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_block                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            dmBlock *block                                   */
/*                                                             */
/***************************************************************/
dmBlock* dmi_descriptor_get_block ( const dmDescriptor* dd )
{
  if ( !dd )
    return NULL;
  else
    return (dmBlock*)dd->block;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_cpt                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*           long           no                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor *cpt                                */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_get_cpt( const dmDescriptor* dd, long no )
{
  dmDescriptor *ddc = NULL;

  /* Trap scalar case */
  if ( no == 0 )
    return (dmDescriptor*)dd;

  if ( no > 0 && no <= dd->dim )
  {
    ddc = dd->cpts[no-1];

    if ( !ddc && dd->dim == 1 ) ddc = (dmDescriptor*)dd;
     return ddc;
  }
  else
    return NULL;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_type                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptorType type                            */
/*                                                             */
/***************************************************************/
dmDescriptorType dmi_descriptor_get_type( const dmDescriptor* dd )
{
  if ( !dd )
  {
    dmu_err_set( DM_ERR_NULL, "[dmDescriptorGetType]: null argument" );
    return dmBADDESC;
  }

  return dd->descType;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_no_coords                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            long ncoords                                     */
/*                                                             */
/***************************************************************/
long dmi_descriptor_get_no_coords( const dmDescriptor* dd )
{
  if ( !dd )
    return 0;

  return dd->ncoords;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_cast_bin                     */
/*                                                             */
/* Description:                                                */
/*    We provide this routine to allow access to the           */
/*    information before column creation                       */
/*    This is needed for default binning.                      */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*     const dmDataType     type                               */
/*                                                             */
/*  Output parameters:                                         */
/*          void   *value                                      */
/*                                                             */
/*  Returns:                                                   */
/*            int dmSUCCESS or dmFAILURE                       */
/*                                                             */
/***************************************************************/
int dmi_descriptor_get_cast_bin( const dmDescriptor* dd,
                                 const dmDataType type,
                                 void* value )
{
  long unit = 1;

  if ( dd && dd->bin )
  {
    ar_gen_get_cast_elt( dd->bin, 0, 1, dd->type, type, value );
    return dmSUCCESS;
  }
  else
  {
    ar_mix_numeric_cast( &unit, DT_INT4, type, value );
    return dmFAILURE;
  }
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_cast_null                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*     const dmDataType     type                               */
/*                                                             */
/*  Output parameters:                                         */
/*          void   *value                                      */
/*                                                             */
/*  Returns:                                                   */
/*            int dmSUCCESS or dmFAILURE                       */
/*                                                             */
/***************************************************************/
int dmi_descriptor_get_cast_null( const dmDescriptor* dd, 
                                  const dmDataType type,
                                  void* value )
{
  if ( !dd || !dd->null )
    return dmFAILURE;

  ar_gen_get_cast_elt( dd->null, 0, 1, dd->type, type, value );

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_cast_null                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*  Output parameters:                                         */
/*          void   *value                                      */
/*                                                             */
/*  Returns:                                                   */
/*         dmBool UT_TRUE or UT_FALSE                          */
/*                                                             */
/***************************************************************/
dmBool dmi_descriptor_get_null( const dmDescriptor* dd, void* null )
{
  if ( dd && dd->null )
  {
    ar_gen_get_elt( dd->null, 0, dd->type, null );
    return UT_TRUE;
  }

  return UT_FALSE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_nullval                      */
/*                                                             */
/* Description:                                                */
/*   Same as dmi_descriptor_get_null() except returns the      */
/*   default NULL value for cols of type DT_REAL4 and DT_REAL8 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*  Output parameters:                                         */
/*          void   *null                                       */
/*                                                             */
/*  Returns:                                                   */
/*           dmBool UT_TRUE or UT_FALSE                        */
/*                                                             */
/***************************************************************/
dmBool dmi_descriptor_get_nullval( const dmDescriptor* dd, void* null )
{
  if ( !dd )
    return UT_FALSE;

  if ( dd->null )
  {
    ar_gen_get_elt( dd->null, 0, dd->type, null );
    return UT_TRUE;
  }
  else
    if ( dd->type == DT_REAL4 )
    {
      *((float*)null) = ut_nan_r();
      return UT_TRUE;
    }
    else
      if ( dd->type == DT_REAL8 )
      {
        *((double*)null) = ut_nan_d();
        return UT_TRUE;
      }

  return UT_FALSE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_has_nullval                      */
/*                                                             */
/* Description:                                                */
/*   Test whether the descriptor has a null value defined      */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          dmBool UT_TRUE or UT_FALSE                         */
/*                                                             */
/***************************************************************/
dmBool dmi_descriptor_has_nullval( const dmDescriptor* dd )
{
  if ( !dd )
    return UT_FALSE;

  return ( dd->null || dd->type == DT_REAL4 || dd->type == DT_REAL8 );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_bin                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*  Output parameters:                                         */
/*          void *bin                                          */
/*                                                             */
/*  Returns:                                                   */
/*          dmBool UT_TRUE or UT_FALSE                         */
/*                                                             */
/***************************************************************/
dmBool dmi_descriptor_get_bin( const dmDescriptor* dd, void* bin )
{
  if ( dd && dd->bin )
  {
    ar_gen_get_elt( dd->bin, 0, dd->type, bin );
    return UT_TRUE;
  }

  return UT_FALSE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_no                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*  Output parameters:                                         */
/*          void *bin                                          */
/*                                                             */
/*  Returns:                                                   */
/*          long  no                                           */
/*                                                             */
/***************************************************************/
long dmi_descriptor_get_no( dmDescriptor* dd )
{
  if ( dd )
    return dd->no;
  else
    return 0; 
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_bytes                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          long  size                                         */
/*                                                             */
/***************************************************************/
long dmi_descriptor_get_bytes( dmDescriptor* dd )
{
  long size = 0;

  if ( !dd )
    return 0;

  size = ar_mix_type_size( dd->type );

  if ( dd->type == DT_CHAR )
    size = UT_SZ_C * (dd->length + 1); 

  return size;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_scaled                       */
/*                                                             */
/* Description:                                                */
/*    A raw table column is the 'logical column'. It is        */
/*    possible to define a 'physical column' which is a linear */
/*    scaling on these values. This is stored as a coord system*/
/*    on the logical column.                                   */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          dmDescriptor scaled_dd                             */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_get_scaled( const dmDescriptor* dd )
{
  dmDescriptor *scaled_dd = NULL;

  scaled_dd = dmi_descriptor_get_coord_no( dd, 1 );
  return scaled_dd;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_unscaled                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          dmDescriptor unscaled_dd                           */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_get_unscaled( const dmDescriptor* dd )
{
  dmDescriptor *unscaled_dd = NULL;

  if ( dd->flag & DM_FLAG_COORD )
    unscaled_dd = dmi_coord_get_parent( dd );

  return unscaled_dd;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_coord_no                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*           long           no                                 */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          dmDescriptor *dd_coord                             */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_get_coord_no( const dmDescriptor* dd, long no )
{
  if ( !dd )
    return NULL;

  if ( dd->ncoords == 0 || !dd->coords || no < 1  || dd->ncoords < no )
    return NULL;

  return dd->coords[no-1];
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_parent                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          dmDescriptor *parent                               */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_get_parent( const dmDescriptor* dd )
{
  if ( dd )
    return dd->parent;
  else
    return NULL;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_aux                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          dmDescriptor *aux                                  */
/*                                                             */
/***************************************************************/
void* dmi_descriptor_get_aux( const dmDescriptor* dd )
{
  if ( !dd )
    return NULL;

  return dd->aux;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_kdd                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          dmDescriptor *kernel_dd                            */
/*                                                             */
/***************************************************************/
void* dmi_descriptor_get_kdd( const dmDescriptor* dd )
{
  if ( !dd )
   return NULL;

  return dd->kernel_dd;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_index                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*           long           i                                  */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          long  cpt_no                                       */
/*                                                             */
/***************************************************************/
long dmi_descriptor_get_index( const dmDescriptor* dd, long i )
{
  dmDescriptor *cpt = NULL;

  if ( dd->cpts )
    cpt = (dmDescriptor*)dd->cpts[i-1];
  else
    cpt = (dmDescriptor*)dd;

  if ( cpt )
    return cpt->no;
  else
    return 0;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_convert_type                     */
/*                                                             */
/* Description:                                                */
/*    COnvert a descriptor to a new datatype                   */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor *dd                                  */
/*           dmDataType    type                                */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          nothing                                            */
/*                                                             */
/***************************************************************/
void dmi_descriptor_convert_type( dmDescriptor* dd, dmDataType type )
{
  double     tmp[2];
  dmDataType old_type    = dd->type;
  long       dim         = 0;
  long       vcpt_offset = 0;

  void *null  = NULL;
  void *range = NULL;
  void *bin   = NULL;
  void *tlmin = &tmp[0];
  void *tlmax = &tmp[1];

  if ( old_type == type ) 
    return;

  dim = dd->dim;
 
  if ( dim > 1 )
  {
    for ( vcpt_offset = 0; vcpt_offset < dim; vcpt_offset++ )
      dmi_descriptor_convert_type( dd->cpts[vcpt_offset], type );
  }

  range = dd->range;
  null = dd->null;
  bin = dd->bin;  
  dd->range = NULL;
  dd->null = NULL;
  dd->bin = NULL;

  if ( ar_mix_is_numeric( type ) && ar_mix_is_numeric( old_type ))
  {
    if ( range )
      dmi_descriptor_get_range( dd, tlmin, tlmax );

    dd->type = type;

    if ( type != DT_REAL4 && type != DT_REAL8 )
    {
      if ( null )
        dmi_descriptor_set_null( dd, old_type, null );
    }

    if ( range )
      dmi_descriptor_set_cpt_range( dd, 0, old_type, tlmin, tlmax );
  }
  else
    dd->type = type;

  if ( range )
    free( range );

  if ( null )
    free( null );

  if ( bin )
    free( bin );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_is_varlen                        */
/*                                                             */
/* Description:                                                */
/*    COnvert a descriptor to a new datatype                   */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor *dd                                  */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          dmBool varlen                                      */
/*                                                             */
/***************************************************************/
dmBool dmi_descriptor_is_varlen( const dmDescriptor* dd )
{
  if ( !dd )
    return dmFALSE;

  return dd->varlen;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_parent_coord_cpt                            */
/*                                                             */
/* Description:                                                */
/*     Avoid multiple inheritance: Up, across, down* e.g.      */
/*        EQPOS(RA,DEC), SKY(X,Y)                              */
/*                                                             */
/*  dd = Y  (dd->cptno = 2)                                    */
/*  parent = SKY                                               */
/*  vector_coord = parent coords no 1 = EQPOS                  */
/*  cpt_coord = vector_coord component no 2 = DEC              */
/*                                                             */
/* It is not necessarily true that DEC = coord-parent of Y     */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor  *dd                                 */
/*           long           coordno                            */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor *cpt_coord                          */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_parent_coord_cpt( dmDescriptor* dd, long coordno )
{
  dmDescriptor *vector       = dd->parent;
  dmDescriptor *vector_coord = NULL;
  dmDescriptor *cpt_coord    = NULL;
  long          dim          = 0;

  /* Check for multiple inheritance */
  if ( vector && vector->coords && vector->ncoords >= coordno )
  {
    vector_coord = vector->coords[coordno-1];

    if ( vector_coord )
    {
      dim = vector_coord->dim;

      if ( dd->cptno > 0 && dim >= dd->cptno  )
        cpt_coord = vector_coord->cpts[dd->cptno-1];
    }
   }

  return cpt_coord;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_range_dim_recurse                 */
/*                                                             */
/* Description:                                                */
/*   This is logically part of dmi_coord_get_range().  It's    */
/*   written as a separate function so we can call it          */
/*   recursively.  See the comments in dmi_coord_get_range()   */
/*   for the full story on what's going on here.               */
/*                                                             */
/*   Basically, we're implementing a set of nested, two-pass   */
/*   "for" loops dim layers deep.  The purpose is to call      */
/*   dmi_coord_calc() on every vertex of the parent            */
/*   descriptor's subspace.  To cover every vertex, we combine */
/*   the minimum and maximum value of each axis with the       */
/*   min/max of every other axis, getting us 2^dim vertices.   */
/*   For two dimensions (e.g. X,Y), the 2^2==4 iterations are: */
/*                                                             */
/*   Xmin, Ymin                                                */
/*   Xmin, Ymax                                                */
/*   Xmax, Ymin                                                */
/*   Xmax, Ymax                                                */
/*                                                             */
/*   Given that dim (at least currently) is always 1 or 2, this*/
/*   is probably overkill.  However, it does at least give us  */
/*   the potential to support 3 or more dimensions.            */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor  *coord_dd                           */
/*           double         minvals[][dmMAXDIM]                */
/*           double         maxvals[][dmMAXDIM]                */
/*           double         pvals[dmMAXDIM]                    */
/*           double         cvals[dmMAXDIM]                    */
/*           long           cddn                               */
/*           long           dim                                */
/*           long           cptno                              */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          static int dmSUCCESS or dmFAILURE                  */
/*                                                             */
/***************************************************************/
static int dmi_coord_get_range_dim_recurse( dmDescriptor* coord_dd,
            double minvals[][dmMAXDIM], double maxvals[][dmMAXDIM],
            double pvals[dmMAXDIM], double cvals[dmMAXDIM],
            long cddn, long dim, long cptno )
{
  long i = 0;
  long j = 0;

  for ( i = 0; i < 2; i++ )
  {
    /* We do two loops so that we use both the minimum and maximum value for
     * axis/component cptno-1.  This statement selects the minval for the first
     * pass and the maxval for the second pass. */

    pvals[cptno-1] = i ? minvals[cddn][cptno-1] : maxvals[cddn][cptno-1];

    if ( cptno < dim )
    {
      /* There are more axes to set after this one, so run this function on cptno+1.
       * Return immediately if something fails, since there's no point in
       * continuing. */

      if ( !dmi_coord_get_range_dim_recurse( coord_dd, minvals, maxvals, pvals, cvals, cddn, dim, cptno+1 ) )
        return dmFAILURE;
    }
    else
    {
      /*
       * We've set values for all the axes, so it's time to calculate the child
       * descriptor's coordinates for this vertex of the parent descriptor's
       * subspace.
       */

      /* Quit if the coord calculation fails */
      if ( !dmi_coord_transform_apply( coord_dd, pvals, cvals, dim ) )
        return dmFAILURE;

      /* Check for new minima and maxima for each axis */
      for ( j = 0; j < dim; j++ )
      {
        if ( cvals[j] < minvals[cddn+1][j] )
          minvals[cddn+1][j] = cvals[j];

        if ( cvals[j] > maxvals[cddn+1][j] )
          maxvals[cddn+1][j] = cvals[j];
      }
    }
  }
  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_cpt_range                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *ddp                                */
/*     const long           vcpt                               */
/*     const dmDataType     type                               */
/*           void          *tlmin                              */
/*           void          *tlmax                              */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          int status                                         */
/*                                                             */
/***************************************************************/
int dmi_descriptor_get_cpt_range( const dmDescriptor* ddp,
                                  const long vcpt,
                                  const dmDataType type,
                                  void* tlmin, void* tlmax )
{
  GenericType   tmp;
  dmDescriptor *dd = NULL;
  int           status = dmSUCCESS; 

  dd = (dmDescriptor*)ddp;

  if ( !dd )
  {
    dmu_err_set(DM_ERR_NULL, "Null arg" );
    return dmFAILURE;
  }

  if ( dd->descType == dmSUBSPACE )
  {
    /* Get column corresponding to subspace */

    dd = dmi_block_search_identifier( dd->block, dd->name );

    if ( !dd ) 
    {
      dd = (dmDescriptor*)ddp;
      ar_mix_type_min( dd->type, &tmp );
      ar_mix_numeric_cast( &tmp, dd->type, type, tlmin );
      ar_mix_type_max( dd->type, &tmp );
      ar_mix_numeric_cast( &tmp, dd->type, type, tlmax );   
      return dmFAILURE;
    }
  }
 
  if ( vcpt != 0 && dd->dim > 1 )
    dd = dd->cpts[vcpt-1];

  if ( !dd )
  {
    sprintf( dm_errbuf, "Problem finding range for descriptor %s", ddp->name );
    dmu_err_set(DM_ERR_NULL, dm_errbuf );
    return dmFAILURE;
  }

  /* the axis range takes precedence over the tlmin/tlmax range */

  if ( dd->flag & DM_FLAG_AXIS && dd->descType == dmCOORD )
  {
    status = dmi_coord_get_range( dd, 0, type, tlmin, tlmax );

    if ( status == dmSUCCESS )
     return status;
  }
  else
  {
    if ( dd->range )
    {
      ar_gen_get_cast_elt( dd->range, 0, 2, dd->type, type, tlmin );
      ar_gen_get_cast_elt( dd->range, 1, 2, dd->type, type, tlmax );
      return dmSUCCESS;
    } 
    else 
    {
      if ( dd->descType == dmCOORD )
      {
        /* Note dd is now the component so vcpt = 0 here */

        status = dmi_coord_get_range( dd, 0, type, tlmin, tlmax );

        if ( status == dmSUCCESS )
          return status;
      }
    }
  
    ar_mix_type_min( dd->type, &tmp );
    ar_mix_numeric_cast( &tmp, dd->type, type, tlmin );
    ar_mix_type_max( dd->type, &tmp );
    ar_mix_numeric_cast( &tmp, dd->type, type, tlmax );
  
    return dmFAILURE;
  }

  return dmFAILURE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_range                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*           void          *tlmin                              */
/*           void          *tlmax                              */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          dmBool UT_TRUE or UT_FALSE                         */
/*                                                             */
/***************************************************************/
dmBool dmi_descriptor_get_range( const dmDescriptor* dd,
                                 void* tlmin, void* tlmax )
{
  if ( dd && dd->range )
  {
    ar_gen_get_elt( dd->range, 0, dd->type, tlmin );
    ar_gen_get_elt( dd->range, 1, dd->type, tlmax );
    return UT_TRUE;
  }

  return UT_FALSE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_make_subspace                    */
/*                                                             */
/* Description:                                                */
/*   Make the subspace equivalent of a column and attach to the*/
/*   given subspace                                            */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor  *col_dd                             */
/*           dmSubspace    *subspace                           */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          dmDsecriptor subspace_dd                           */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_make_subspace( dmDescriptor* col_dd, dmSubspace* subspace )
{
  if ( !col_dd )
    return NULL;

  return dmi_subspace_col_create( (dmBlock*)col_dd->block, subspace,
             col_dd->name, col_dd->cptNames, col_dd->dim, col_dd->type,
             col_dd->unit, col_dd, UT_TRUE );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_copy_range                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor  *dd                                 */
/*           dmDescriptor  *new_dd                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          nothing                                            */
/*                                                             */
/***************************************************************/
void dmi_descriptor_copy_range( dmDescriptor* dd, dmDescriptor* new_dd )
{
  GenericType tlmin;
  GenericType tlmax;
  long        vcpt  = 0;

  if ( dmi_descriptor_get_cpt_range( dd, 0, dd->type, (void*)&tlmin, (void*)&tlmax ) )
    dmi_descriptor_set_range( new_dd, dd->type, &tlmin, &tlmax );

  if ( dd->dim > 1 )
  { 
    for ( vcpt = 1; vcpt <= dd->dim; vcpt++ )
    {
      if ( dmi_descriptor_get_cpt_range( dd, vcpt, dd->type, (void*)&tlmin, (void*)&tlmax ))
        dmi_descriptor_set_range( new_dd->cpts[vcpt-1], dd->type, &tlmin, &tlmax );  
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_nvals                        */
/*                                                             */
/* Description:                                                */
/*     Return total items for a descriptor                     */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor  *dd                                 */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          long nvals                                         */
/*                                                             */
/***************************************************************/
long dmi_descriptor_get_nvals( dmDescriptor* dd )
{
  long nvals = 0;

  if ( !dd ) 
    return 0;

  dd->size = dmi_descriptor_calc_size( dd->axes, dd->naxes );
  nvals = dd->dim * dd->size; 
  return nvals;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_scalar_coord_no                   */
/*                                                             */
/* Description:                                                */
/*   Add up total coords up to this raw column                 */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor  *dd                                 */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          long total                                         */
/*                                                             */
/***************************************************************/
long dmi_coord_get_scalar_coord_no( dmDescriptor* dd )
{
  dmDescriptor *parent = 0;
  dmDescriptor *ddc    = 0;
  dmBlock      *block  = 0;

  long no    = 0; 
  long i     = 0;
  long cpt   = 0;
  long total = 0;

  if ( !dd )
    return 0;

  cpt = dd->cptno;

  if ( cpt > 0 )
    dd = dd->parent;

  no = dd->no;
  block = (dmBlock*)dd->block;
  parent = dmi_coord_get_parent( dd );

  for ( i = 0; i < no-1; i++ )
  {
    ddc = parent->coords[i];
    total+= ddc->dim;
  }

  total += cpt;
  return total; 
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_scalar_col_no                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor  *dd                                 */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          long col_no                                        */
/*                                                             */
/***************************************************************/
long dmi_coord_get_scalar_col_no( dmDescriptor* dd )
{
  dmBlock      *block  = NULL;
  dmDescriptor *parent = NULL;
  dmDescriptor *ddc    = NULL;

  long no    = 0; 
  long i     = 0;
  long cpt   = 0;
  long total = 0;

  if ( !dd )
    return 0;

  cpt = dd->cptno;

  if ( cpt > 0 )
    dd = dd->parent;
  else
    cpt = 1;

  parent = dmi_coord_get_parent( dd );

  if ( !parent )
    return 0;

  while ( parent && !(parent->flag & DM_FLAG_LOGICAL) )
    parent = dmi_coord_get_parent( parent );

  no = parent->no;
  block = (dmBlock*)parent->block;

  if ( block->type == dmIMAGE )
    return 1;

  if ( no > block->ncols )
  {
    sprintf( dm_errbuf, "Inconsistent column number in %s (%ld but ncols=%ld)", parent->name, no, block->ncols );
    dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
    return 0;
  }

  for ( i = 0; i < no-1; i++ )
  {
    ddc = block->cols[i];
    total+= ddc->dim;
  }
 
  total += cpt;
  return total; 
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_flag                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor  *dd                                 */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          long flag                                          */
/*                                                             */
/***************************************************************/
long dmi_descriptor_get_flag( dmDescriptor* dd )
{
  if ( dd )
    return dd->flag;
  return 0;
}




/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_name_with_cpts                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor  *dd                                 */
/*           char          *fullname                           */
/*           const long     maxlen                             */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          nothing                                            */
/*                                                             */
/***************************************************************/
void dmi_descriptor_get_name_with_cpts( const dmDescriptor* dd,
                                        char* fullname, const long maxlen )
{
  long   ncpts    = 0;
  long   dim      = 0;
  char **cptnames = NULL;

  if ( !dd )
  {
    cs_copy( " ", fullname );
    return;
  }

  dim = dd->dim;
  ncpts =  ( dim > 1 ) ? dim : 0;

  if ( ncpts )
    cptnames = dd->cptNames;

  if ( !cptnames ) 
    cs_copy( dd->name, fullname );
  else
  {
    long i = 0;
    char **pcptnames = calloc( dim + 1, UT_SZ_P );

    pcptnames[0] = dd->name;

    for ( i = 1; i <= dim; i++ )
      pcptnames[i]=cptnames[i-1];

    cs_compose_paren_list( pcptnames, ncpts, fullname, maxlen );
    free( pcptnames );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_get_range                             */
/*                                                             */
/* Description:                                                */
/*  Get the range for a coord. Works specifically for an       */
/*  image axis or axis group.                                  */
/*                                                             */
/*  For a general coord, it gets the coord values corresponding*/
/*  to the range of the parent.  However, since the coord      */
/*  transform may not be monotonic, this doesn't always work.  */
/*  (See detailed comments below.)                             */
/*  So yes, it's flaky, but it's the best we're going to do    */
/*  right now.                                                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *ddp                                */
/*     const long           vcptp                              */
/*     const dmDataType     type                               */
/*           double        *tlmin                              */
/*           double        *tlmax                              */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          int  dmSUCCESS  or dmFAILURE                       */
/*                                                             */
/***************************************************************/
int dmi_coord_get_range( const dmDescriptor* ddp, const long vcptp,
                         const dmDataType type, double* tlmin,
                         double* tlmax )
{
#define  DM_C_N        4
#define  DM_C_IMAGE    0
#define  DM_C_LOGICAL  1
#define  DM_C_PHYSICAL 2
#define  DM_C_WORLD    3

  dmBlock      *block       = NULL;
  dmDescriptor *image_dd    = NULL;
  dmDescriptor *dd          = NULL;
  dmDescriptor *coord_dd    = NULL;
  dmDescriptor *range_dd    = NULL;
  dmDescriptor *cdd[DM_C_N] = { NULL, NULL, NULL, NULL };

  double minvals[DM_C_N][dmMAXDIM];
  double maxvals[DM_C_N][dmMAXDIM];
  double pvals[dmMAXDIM];
  double cvals[dmMAXDIM];
 
  void *pmin = NULL;
  void *pmax = NULL;

  long *axlen    = NULL;
  long  vcpt     = (long)vcptp;
  long  groupno  = 0;
  long  psize    = 0;
  long  dim      = 0;
  long  index    = 0;
  long  subaxis  = 0;
  long  naxes    = 0;
  long  axoffset = 0;
  long  base     = 0;
  long  i        = 0;
  long  j        = 0;
  int   ok       = 0;



  if ( !ddp || type != DT_REAL8 )
    return dmFAILURE;

  dd = (dmDescriptor*)ddp;

  if( vcpt == 0 && dd->cptno > 0 )
  {
    vcpt = dd->cptno;
    dd = dd->parent;
  }

  block = (dmBlock*)dd->block;
  dim   = dd->dim;

  if ( dd->flag & DM_FLAG_LOGICAL )
    base = DM_C_LOGICAL;
  else
    if ( dd->flag & DM_FLAG_PHYS )
      base = DM_C_PHYSICAL;
    else 
     base = DM_C_WORLD;
 
  index = base;
  cdd[index] = dd;
 
  while ( index > DM_C_IMAGE )
  {
    cdd[index-1] = dmi_coord_get_parent( cdd[index] );
    --index;
  }

  if ( dd->flag & DM_FLAG_AXIS )
  {
    /* OK, we have a group descriptor */
    groupno = cdd[DM_C_LOGICAL]->no;
    image_dd = cdd[DM_C_IMAGE];
    axoffset = dmi_descriptor_get_group_axis( cdd[DM_C_IMAGE], groupno, 1 ) - 1;
    naxes = image_dd->naxes;
    axlen = image_dd->axes;

    for ( subaxis = 0; subaxis < dim; subaxis++ )
    {
      minvals[DM_C_LOGICAL][subaxis] = 0.5;
      maxvals[DM_C_LOGICAL][subaxis] = axlen[axoffset] + 0.5;
      ++axoffset;
    }

    for ( index = DM_C_LOGICAL; index < base; index++ )
    {
      dmi_coord_calc( cdd[index+1], minvals[index], DT_REAL8, DT_REAL8, minvals[index+1] );
      dmi_coord_calc( cdd[index+1], maxvals[index], DT_REAL8, DT_REAL8, maxvals[index+1] );
    }
  } 
  else 
  {
    /*
     * Here we deal with virtual table columns, i.e. those "columns" which are
     * actually coordinate transforms on other columns.  Since they aren't
     * real columns in the input file, they won't have ranges associated with
     * them, so we have to calculate the range based on the range of the
     * coordinate parent descriptor.
     *
     * The method is to work up the coord parent stack (contained in the cdd[]
     * array) until we find a descriptor that has range values associated with
     * it.  We then work back down, calculating the ranges for each child by
     * applying the appropriate coordinate transform to its parent's ranges.
     * This is complicated by the fact that the coord transform may include a
     * rotation, which means we have to map each vertex of the parent's
     * subspace (e.g. a rectangle for a 2-D transform) to the child's axes
     * and find from that set of points the minimum and maximum value for each
     * axis.  Since the number of dimensions is arbitrary, we have to do
     * this recursively.
     *
     * !!! IMPORTANT !!!
     * Since this method assumes that every coordinate transform is monotonic,
     * it doesn't always work.  For example, it will break badly with EQPOS
     * around the celestial poles or at Ra=0, or with MSC(THETA,PHI) anywhere.
     * We really, really need a better and more general solution, but this is
     * at least partially useful in the meantime.
     */


    /* It doesn't make sense to ask for the range of a vector.  You have to
     * ask for the ranges of its components instead. */

    if ( vcpt == 0 && dim > 1 )  return dmFAILURE;


    /*
     * Work up the coord-parent stack until we find a descriptor with ranges.
     */ 

    ok = 0;

    for ( index = base; index >= DM_C_LOGICAL; index-- )
    {
      /* No descriptor, so quit */
      if ( !cdd[index] )
        break;

      /* Check if the descriptor itself has a range (i.e. for 1-D transforms) */
      if ( cdd[index]->range )
      {
        ok = 1;
        break;
      }

      /* Check if descriptor's components have ranges.  If any are missing,
       * stop checking and move up the stack. */

      if ( cdd[index]->cpts && cdd[index]->dim == dim )
      {
        ok = 1;

        for ( i = 0; i < dim; i++ )
        {
          if ( !cdd[index]->cpts[i] || !cdd[index]->cpts[i]->range )
          {
            ok = 0;
            break;
          }
        }
      }

      if ( ok )
        break;
    }

    /* Quit if we didn't find a descriptor with ranges */
    if ( !ok )
      return dmFAILURE;


    /*
     * Convert the range values we found to type DT_REAL8.
     */

    psize = ar_mix_type_size( cdd[index]->type );
    pmin = (void*)malloc( psize );
    pmax = (void*)malloc( psize );

    for ( i = 0; i < dim ; i++ )
    {
      if ( cdd[index]->range )
        range_dd = cdd[index];
      else
        range_dd = cdd[index]->cpts[i];

     ok = dmi_descriptor_get_range( range_dd, pmin, pmax );

     if ( !ok )
       return dmFAILURE;

     ar_mix_numeric_cast( pmin, cdd[index]->type, DT_REAL8, &(minvals[index][i]) );
     ar_mix_numeric_cast( pmax, cdd[index]->type, DT_REAL8, &(maxvals[index][i]) );
    }

    free( pmin );
    free( pmax );


    /*
     * Work back up, calculating ranges for each child from the ranges of its
     * coord parent.  As noted above, we have to do this recursively.
     */

    for ( i = index; i < base; i++ )
    {
      /* Find the coord transform that will take us from the current parent to
       * it's child.  We search for it by name in case the parent has multiple
       * coordinate systems. */
       coord_dd = dmi_descriptor_list_search_descriptor( (const dmDescriptor**)(cdd[i]->coords), cdd[i]->ncoords, cdd[i+1]->name );

      if ( !coord_dd )
        return dmFAILURE;

      /* Initialize the minvals for each axis of the child to DBL_MAX, the maxvals to
       * -DBL_MAX */
      for ( j = 0; j < dim; j++ )
      {
        minvals[i+1][j] =  DT_REAL8_MAX;
        maxvals[i+1][j] = -DT_REAL8_MAX;
      }

      /* Map every vertex of the parent's subspace to the child's axes, and determine
       * from that set of points the minimum and maximum value for each of the child's
       * axes.  Quit if any error occurs. */

      ok = dmi_coord_get_range_dim_recurse( coord_dd, minvals, maxvals, pvals, cvals, i, dim, 1 );

      if ( !ok )
        return dmFAILURE;
    }
  }

  if ( vcpt == 0 )
  {
    *tlmin = minvals[base][0];
    *tlmax = maxvals[base][0];
  }
  else
  {
    *tlmin = minvals[base][vcpt-1]; 
    *tlmax = maxvals[base][vcpt-1]; 
  }

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_scalar_col                        */
/*                                                             */
/* Description:                                                */
/*    Map  scalar column number to descriptor                  */
/*                                                             */
/*  Input parameters:                                          */
/*           dmBlock   *block                                  */
/*           long       scalar_no                              */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*          dmDescriptor *dd                                   */
/*                                                             */
/***************************************************************/
dmDescriptor *dmi_block_get_scalar_col( dmBlock* block, long scalar_no )
{
  long col_offset = 0; 
  long no = 0;
  long dim;
  dmDescriptor* dd;

  for ( col_offset = 0; col_offset < block->ncols; col_offset++ )
  {
    dd = block->cols[col_offset];
    dim = dd->dim;

    if ( scalar_no <= no + dim )
      return dmi_descriptor_get_cpt( dd, scalar_no - no );

    no += dim;
  }
  return NULL;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_copy_kdd                         */
/*                                                             */
/* Description:                                                */
/*     Diagnostic routine                                      */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor  *image                              */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_descriptor_dump_axes( dmDescriptor* image )
{
  dmDescriptor *group = NULL;

  char *name = NULL;
  char *null = "(none)";

  long ng      = image->ngroups;
  long n       = image->naxes;
  long axisno  = 0;
  long groupno = 0;
  long dim     = 1;
  long vcpt    = 1;

  printf( "------------NAX = %ld NG = %ld --\n", n, ng );

  for ( axisno = 1; axisno <= n; axisno++ )
  {
     ++vcpt;

     if ( vcpt > dim )
     {
       ++groupno;
       vcpt = 1;

       if ( groupno <= ng )
         group = image->groups[groupno-1];
       else
         group = NULL;

       if ( group )
       {
         dim  = group->dim;
         name = group->name;
       }
       else
       {
         dim = 1;
         name = null;
       }
     }

   printf( "Axis %4ld Length %10ld Group %4ld %s\n", axisno, image->axes[axisno-1], groupno, name );

  }  
  printf( "-------------\n" );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_get_qualname                     */
/*                                                             */
/* Description:                                                */
/*     Diagnostic routine useful in debugging                  */
/*                                                             */
/*  Input parameters:                                          */
/*     const dmDescriptor  *dd                                 */
/*           char          *buf                                */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            char *buf                                        */
/*                                                             */
/***************************************************************/
char* dmi_descriptor_get_qualname( const dmDescriptor* dd, char* buf )
{
  dmBlock   *block = NULL;
  dmDataset *ds    = NULL;

  block = (dmBlock*)dd->block;

  if ( block ) 
  {
    ds = block->ds;

    if ( dd->descType == dmSUBSPACE )
      sprintf( buf, "%s:%s.#SUBSPACE.%s", ds->name, block->name, dd->name );
    else
      sprintf( buf, "%s:%s.%s", ds->name, block->name, dd->name );
  }
  else
  {
    sprintf( buf, ":.%s", dd->name );
  }

  return buf;
}



