/*                                                                
**  Copyright (C) 2002-2003,2005,2008  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmiddlib.c                                                  */
/*                                                               */
/*       Functions that work with dmDescriptors                  */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"


/* Other descriptor methods */

char* dmi_vector_make_default_name( dmDescriptor **list, long dim );
void dmi_col_print2_loop( const dmDescriptor *dd, char *prefix, int down );


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_match                            */
/*                                                             */
/* Description:                                                */
/*   Find descriptor in descriptor list                        */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*          dmDescriptor  *dd    - descriptor to find          */
/*          dmDescriptor **list  - list of descriptors         */
/*          long           n     - number of descriptors in    */
/*                                 list                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*          long index-of-descriptor-in-list                   */
/*                                                             */
/***************************************************************/
long dmi_descriptor_match( dmDescriptor *dd, dmDescriptor **list, long n )
{
  long ii = 0;

  for ( ii = 0; ii < n; ii++ )
  {
    if ( dd == list[ii] )
      return ii+1;
  }

  return 0;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_alloc_vec                        */
/*                                                             */
/* Description:                                                */
/*   Allocate a descriptor for a vector column                 */
/*                                                             */
/*  Input parameters:                                          */
/*          long          dim   - dimension of column          */
/*          long          vdim  - dimention of vector column   */
/*          dmElementType etype - element type                 */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*          dmDescriptor *vector                               */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_alloc_vec( long dim, long vdim,
                                        dmElementType etype )
{
  long naxes           = 1;
  dmDescriptor *vector = NULL;    
  dmBlock      *null   = NULL;

  vector = dmi_descriptor_alloc( null, dmCOLUMN, 0, NULL, dmUNKNOWNTYPE,
                     NULL, NULL, dim, etype, naxes, NULL, DM_FLAG_VECTOR );

  vector->dim  = dim;
  vector->vdim = vdim;
  vector->edim = dim / vdim;

  vector->eltNames = (char**)calloc( vector->vdim, UT_SZ_P );
  vector->cptNames = (char**)calloc( vector->dim, UT_SZ_P);

  return vector;
}


/***************************************************************/
/*                                                             */
/* Function:   fdmi_col_alloc                                  */
/*                                                             */
/* Description:                                                */
/*   Allocat column descriptor                                 */
/*   kernel.                                                   */
/*                                                             */
/*  Input parameters:                                          */
/*       dmBlock            *block                             */
/*       const long          no                                */
/*       const char         *name                              */
/*       const dmDataType    type                              */
/*       const char         *unit                              */
/*       const char         *desc                              */
/*       const long          vdim                              */
/*       const dmElementType etype                             */
/*       const long          naxes                             */
/*       const long         *axes                              */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*          dmDescriptor *dd                                   */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_col_alloc( dmBlock *block, const long no, const char *name,
                             const dmDataType type, const char *unit,
                             const char *desc, const long vdim,
                             const dmElementType etype, const long naxes,
                             const long *axes )
{
  dmDescriptor *dd = NULL;

  dd = dmi_descriptor_alloc( block, dmCOLUMN, no, name, type, unit, 
                            desc, vdim, etype, naxes, axes, DM_FLAG_VISIBLE ); 
  if ( !dd )
    return NULL;

  block->cols = realloc( block->cols, (block->ncols+1)* UT_SZ_P );
  block->cols[block->ncols] = dd;
  block->ncols++;

  return dd;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_make_vector                                 */
/*                                                             */
/* Description:                                                */
/*   Calls function dmi_make_element_vector() to do work       */
/*                                                             */
/*  Input parameters:                                          */
/*          dmBlock       *block                               */
/*          char          *name                                */
/*          dmDescriptor **ddlist                              */
/*          long           dim                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*          dmDescriptor  *dd                                  */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_make_vector( dmBlock *block, char *name,
                               dmDescriptor **ddlist, long dim )
{
 return dmi_make_element_vector( block, name, ddlist, dim, dmVALUE, NULL );
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_make_element_vector                         */
/*                                                             */
/* Description:                                                */
/*   Create a vector descriptor                                */
/*                                                             */
/*  Input parameters:                                          */
/*          dmBlock       *block                               */
/*          char          *name                                */
/*          dmDescriptor **ddlist                              */
/*          long           dim                                 */
/*          dmElementType  etype                               */
/*          char          *descp                               */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*          dmDescriptor  *vector                              */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_make_element_vector( dmBlock *block, char *name,
                                       dmDescriptor **ddlist, long dim,
                                       dmElementType etype, char *descp )
{
  dmDescriptor *dd     = NULL;
  dmDescriptor *ddc    = NULL;
  dmDescriptor *vector = NULL;
  char         *vname  = (char*)name;
  char         *desc   = descp;

  long  cptno    = 0;
  long  flag     = 0;
  long  no       = 0;
  int   is_alloc = 0;


  dd = ddlist[0];

  if ( !dd ) 
  {
    sprintf( dm_errbuf, "Attempt to vectorize null component" );
    dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
    return NULL;
  }

  flag = dd->flag;

  if ( dim == 1 )
  {
    dd->cptNames[0] = cs_dup( dd->name );
    return dd;
  }


  /* Consistency checks - CXC DM requires like-typed aggregation */
  for ( cptno = 1; cptno <= dim; cptno++ )
  {
    ddc = ddlist[cptno-1];

    if ( !ddc )
    {
      sprintf( dm_errbuf, "Vector descriptor build list is inconsistent for %s", dd->name );
      dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );

      return NULL;
    }

    if ( dd->descType != ddc->descType )
      dmu_err_set( DM_ERR_INTERNAL, "Vector column type conflict" );

    if ( dd->type != ddc->type )
      dmu_err_set( DM_ERR_INTERNAL, "Vector column datatype conflict" );

    ddc->flag |= DM_FLAG_CPT;
  }

  if ( dim == 1 || !desc )
  {
    desc = dd->desc;
  }
 
  if ( !vname) 
  {
    vname = dmi_vector_make_default_name( ddlist, dim );
    is_alloc = 1;
  }

  /* Make the vector */
  vector = dmi_descriptor_alloc( block, dd->descType, no, vname, dd->type, dd->unit, 
                            desc, dim, etype, dd->naxes, dd->axes, flag | DM_FLAG_VECTOR ); 

  if ( is_alloc )
    free( vname );

  /* Assign the components */
  for  ( cptno = 1; cptno <= dim; cptno++ )
  {
    dmi_descriptor_set_cpt( vector, cptno, ddlist[cptno-1] );  
  }
 
  if ( vector->descType == dmCOORD )
  {
    dmi_coord_make_transform_vector( vector );
  }

  /* Note: the multi-dim vector does NOT get the range from its children */

  /* Copy nulls */
  dmi_descriptor_set_null( vector, dd->type, dd->null );
  dmi_descriptor_set_bin( vector, dd->type, dd->bin );

  // Code Change Bug 11332 October 22, 2008
  //  Strings need length element in order to allocate buffer space
  //  when reading or writing

  if (dd->length)
    vector->length = dd->length;

  return vector;
}



/***************************************************************/
/*                                                             */
/* Function:  dmi_vector_make_default_name                     */
/*                                                             */
/* Description:                                                */
/*  Default names for a vector.                                */
/*     CptName[elt][cpt] = raw col name                        */
/*     EltName[elt]      = CptName[elt][0]                     */
/*                                                             */
/*     If dim > 1:                                             */
/*       Name  = (CptName,CptName,CptName..)                   */
/*      else                                                   */
/*       Name  = CptName                                       */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*          dmDescriptor **list                                */
/*          long           dim                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*          char *buf      - default name string               */
/*                                                             */
/***************************************************************/
char* dmi_vector_make_default_name( dmDescriptor **list, long dim )
{
   dmDescriptor *dd = NULL;

   char  buf[CARD_SIZE]="";   /* Buffer to compose vector name */
   char *ptr   = NULL;
   char *cname = NULL;

   long  cpt_offset = 0;


  /* Default cases */
  if ( dim == 2 )
  {
    if ( cs_uceq( list[0]->name, "X" ) && cs_uceq( list[1]->name, "Y" ))
    {
      return cs_dup( "POS" );
    }

    if ( cs_uceq( list[0]->name, "RA" ) && cs_uceq( list[1]->name, "DEC" ))
    {
      return cs_dup( "EQPOS" );
    }

     // FITS Standard Compliance  July 2010
     //  Added the following 4 new default cases

    if ( cs_uceq( list[0]->name, "GLON" ) && cs_uceq( list[1]->name, "GLAT" ))
    {
      return cs_dup( "GAL" );
    }

    if ( cs_uceq( list[0]->name, "ELON" ) && cs_uceq( list[1]->name, "ELAT" ))
    {
      return cs_dup( "EC" );
    }

    if ( cs_uceq( list[0]->name, "HLON" ) && cs_uceq( list[1]->name, "HLAT" ))
    {
      return cs_dup( "HEC" );
    }

    if ( cs_uceq( list[0]->name, "SLON" ) && cs_uceq( list[1]->name, "SLAT" ))
    {
      return cs_dup( "SGAL" );
    }
  }

  /* Standard case */
  ptr = buf;

   if( dim > 1 )
    *ptr++ = '(';

  for ( cpt_offset = 0; cpt_offset < dim; cpt_offset++ )
  {
    dd = list[cpt_offset];
    cname = dd->name;

    if ( cpt_offset < dim -1 )
      ptr += sprintf( ptr, "%s,", cname );
    else
      if ( dim == 1 )
        strcpy( buf, cname );
      else
        sprintf( ptr, "%s)", cname );
  }

  return cs_dup( buf );  
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_tstart                            */
/*                                                             */
/* Description:                                                */
/*      Get TSTART and TSTOP value for this block              */
/*                                                             */
/*  Input parameters:                                          */
/*          dmBlock   *block                                   */
/*                                                             */
/*  Output parameters:                                         */
/*          double    *tstart                                  */
/*          double    *tstop                                   */
/*                                                             */
/*  Returns:                                                   */
/*          int   retval                                       */
/*                                                             */
/***************************************************************/
int dmi_block_get_tstart( dmBlock *block, double *tstart, double *tstop )
{  
 dmDescriptor *key_dd = NULL;
 int           retval = 0;

  /* Get TSTART and TSTOP keys */

  *tstart = 0.0;
  *tstop = 0.0;

   key_dd =  dmi_key_read( block, "TSTART", DT_REAL8, tstart );

   if ( key_dd )
     key_dd = dmi_key_read( block, "TSTOP", DT_REAL8, tstop );

  retval = ( key_dd != NULL );

  if ( retval )
    retval = ( *tstop != 0.0 );

 return retval;
}                              



/***************************************************************/
/*                                                             */
/* Function:   dmi_vector_make_cpts                            */
/*                                                             */
/* Description:                                                */
/*   Allocate and populate a vector component descriptor       */
/*                                                             */
/*  Input parameters:                                          */
/*          dmDescriptor  *dd                                  */
/*          long           trueDim                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*          void                                               */
/*                                                             */
/***************************************************************/
void dmi_vector_make_cpts( dmDescriptor *dd, long trueDim )
{
  dmDescriptor *ddcpt = NULL;
  long          elem  = 0;

  if ( trueDim <= 1 || !dd->cptNames )
  {
    dd->cpts        = (dmDescriptor**)calloc(1, UT_SZ_P );
    dd->cpts[0]     = dd;
    dd->cptNames    = calloc( 1, UT_SZ_P );
    dd->cptNames[0] = cs_dup( dd->name );      
  }
  else
  {
    if ( !dd->cpts )
      dd->cpts    = (dmDescriptor**)calloc(trueDim, UT_SZ_P );

    for ( elem = 0; elem < trueDim; elem++ )
    {
      ddcpt = dmi_descriptor_alloc( (dmBlock*)dd->block, dd->descType, 0, dd->cptNames[elem], 
                                    dd->type, dd->unit, dd->desc, 1, dmVALUE, dd->naxes,
                                    dd->axes, dd->flag | DM_FLAG_CPT ); 

      ddcpt->cptno   = elem + 1;
      ddcpt->parent  = dd;
      dd->cpts[elem] = ddcpt;
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_axis_bounds                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Define an image subarray by calculating the lower left and */
/*  upper right pixel coords                                   */
/*  (1) if *size is zero on entry, return the entire image     */
/*      and set *size to the number of pixels in the image.    */
/*  (2) otherwise, return a subarray starting at the first     */
/*      pixel and extending (*size) pixels in all.             */  
/*                                                             */
/*  Input parameters:                                          */
/*          dmDescriptor  *imagedata                           */
/*          long          *size                                */
/*                                                             */
/*  Output parameters:                                         */
/*          long         **ll                                  */
/*          long         **ur                                  */
/*                                                             */
/*  Returns:                                                   */
/*          void                                               */
/*                                                             */
/***************************************************************/
void dmi_descriptor_axis_bounds( dmDescriptor *imagedata, long **ll,
                                 long **ur, long *size )
{
  long *lowerLeft  = NULL;
  long *upperRight = NULL;
  long *axlens     = NULL;
  long 	axes       = 0;
  long  imageBufSize = 1;
  long  ii         = 0;


  if ((axes = dmi_descriptor_get_array_dimensions( imagedata, &axlens)) == dmBADDIM)
    return;
  
  if (axes > 0)
  {
    /* initialize corners */
    lowerLeft  = (long *)malloc( UT_SZ_I * axes);
    upperRight = (long *)malloc( UT_SZ_I * axes);

    for ( ii = 0; ii < axes; ii++ )
      imageBufSize *= axlens[ii];

    if ( *size <= 0 || *size >= imageBufSize )
      *size = imageBufSize;

    if ( imageBufSize == 0 )
    {
      *ll = NULL;
      *ur = NULL;
      return;
    }
	      
    dmu_pixel_from_pixelno( 1,     axlens, axes, lowerLeft );
    dmu_pixel_from_pixelno( *size, axlens, axes, upperRight );
    free( axlens ); 
  }

  *ll = lowerLeft;
  *ur = upperRight;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_image_col_create                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*          dmBlock       *blockr                              */
/*          dmDataType     type                                */
/*          long          *axes                                */
/*          long           naxes                               */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*          dmDescriptor  *dd                                  */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_image_col_create( dmBlock *block, dmDataType type,
                                    long *axes, long naxes )
{
  dmDescriptor  *dd        = NULL; 
  dmDescriptor  *scaled_dd = NULL;
  dmElementType  etype     = dmVALUE;

  char *name    = dmi_block_get_name( block ); 
  char *unit    = NULL;
  char *desc    = NULL;
  char *rawname = "#1";

  double  pix   = 0.0;
  double  tscal = 1.0;
  double  tzero = 0.0;

  long  dim = 1;
  long  no  = 1;
  long  flag = DM_FLAG_IMG | DM_FLAG_RAW | DM_FLAG_LOGICAL;

  dd = dmi_descriptor_alloc( block, dmCOLUMN, no, rawname, type, unit, 
                            desc, dim, etype, naxes, axes, flag ); 

  scaled_dd = dmi_create_linear_wcs( block, dd, name, type, tzero, pix, tscal );

  return dd;
}

/***************************************************************/
/*                                                             */
/* Function:   dmi_make_unscaled_groups                        */
/*                                                             */
/* Description:                                                */
/*   make unscaled vector groups                               */
/*                                                             */
/*  Input parameters:                                          */
/*          dmDescriptor **vectors                             */
/*          long           nvectors                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*          void                                               */
/*                                                             */
/***************************************************************/
void dmi_make_unscaled_groups( dmDescriptor **vectors, long nvectors )
{
  dmDescriptor *vector = NULL;
  dmDescriptor *group  = NULL;
  dmDescriptor *dd     = NULL;
  dmDescriptor *us_dd[dmMAXDIM];

  char name[dmMAXLEN];

  long ii = 0;
  long jj = 0;

  for ( ii =0; ii < nvectors; ii++ )
  {
    vector = vectors[ii];

    if ( vector->dim > 1 )
    {
      for ( jj = 0; jj < vector->dim; jj++ )
      {
        dd = vector->cpts[jj];
        us_dd[jj] = dmi_descriptor_get_unscaled( dd );
      }

      dmi_column_make_raw_name( vector->name, vector->no, name );
      group = dmi_make_vector( (dmBlock*)vector->block, name, us_dd, vector->dim );   
      dmi_descriptor_add_coord( group, vector );
    }
  }   
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_col_print2                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*    const dmDescriptor  *dd                                  */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*          void                                               */
/*                                                             */
/***************************************************************/
void dmi_col_print2( const dmDescriptor *dd )
{
  char prefix[TEXT_SIZE] = "DD";
  fio_dashline( 80 );
  dmi_col_print2_loop( dd, prefix, -1 );
  fio_dashline( 80 );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_col_print2_loop                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*    const dmDescriptor  *dd                                  */
/*          char          *prefix                              */
/*          int            down                                */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*          void                                               */
/*                                                             */
/***************************************************************/
void dmi_col_print2_loop( const dmDescriptor *dd, char *prefix, int down )
{
  dmDescriptor *ddp = NULL;

  char buf[TEXT_SIZE];
  char suffix[TEXT_SIZE];

  long maxlen = TEXT_SIZE;
  long ii     = 0;


  if ( dd && dd->parent && down == -1 )
  {
    sprintf( suffix, "%s[P]", prefix );
   dmi_col_print2_loop( dd->parent, suffix, 0 );
  }

  if ( dd && dd->aux && down == -1 )
  { 
    ddp = (dmDescriptor*)dmi_coord_get_parent( dd );

    if ( ddp )
    {
      sprintf( suffix, "%s[TP]", prefix );
      dmi_col_print2_loop( ddp, suffix, 0 );  
    }
  }

  dmi_col_print( dd, buf, maxlen );
  printf( "%-20s %s\n", prefix, buf );

  if ( !dd || down == 0 )
    return;

  for ( ii = 1; ii <= dd->dim; ii++ )
  {
    sprintf( suffix, "%s[%ld]", prefix, ii );

    if ( dd->dim > 1 || dd->cpts[ii-1] != NULL )
      dmi_col_print2_loop( dd->cpts[ii-1], suffix, 1 );
  }

  for ( ii = 1; ii <= dd->ncoords; ii++ )
  {
    sprintf( suffix, "%s[C%ld]", prefix, ii );
    dmi_col_print2_loop( dd->coords[ii-1], suffix, 1 );
  }

  for ( ii = 1; ii <= dd->ngroups; ii++ )
  {
    sprintf( suffix, "%s[G%ld]", prefix, ii );
    dmi_col_print2_loop( dd->groups[ii-1], suffix, 1 );
  }
}

/***************************************************************/
/*                                                             */
/* Function:   dmi_descriptor_compare                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*          dmDescriptor  *dd1                                 */
/*          dmDescriptor  *dd2                                 */
/*          long           mask                                */
/*                                                             */
/*  Output parameters:                                         */
/*          void          *data                                */
/*                                                             */
/*  Returns:                                                   */
/*          dmBool         status  - UT_TRUE or UT_FALSE       */
/*                                                             */
/***************************************************************/
dmBool dmi_descriptor_compare( dmDescriptor *dd1, dmDescriptor *dd2,
                              long mask )
{
  long size1 = 0;
  long size2 = 0;

  if ( !dd1 || !dd2 )
  {
    dmu_err_set( DM_ERR_NULL, "DescriptorCompare on null" );
    return UT_FALSE;
  }
 
  size1 = dmi_descriptor_get_array_size( dd1 );
  size2 = dmi_descriptor_get_array_size( dd2 );

  if ( mask & dmDESCNAME )
  {
    if ( !cs_uceq( dd1->name, dd2->name ))
      return UT_FALSE;
  }

  if ( mask & dmDESCUNIT )
  {
    if ( !cs_uceq( dd1->unit, dd2->unit ))
      return UT_FALSE;
  }

  if ( mask & dmDESCKIND )
  {
    if ( dd1->descType != dd2->descType )
      return UT_FALSE;
  }

  if ( mask & dmDESCTYPE )
  {
     if ( dd1->type != dd2->type )
      return UT_FALSE;
  }

  if ( mask & dmDESCELMTYPE )
  {
    if ( dd1->etype != dd2->etype )
      return UT_FALSE;
  }

  if ( mask & dmDESCDIM )
  {
    if ( dd1->naxes != dd2->naxes )
      return UT_FALSE;
  }

  if (  mask & ( dmDESCELMDIM | dmDESCCPTNAMES | dmDESCVALUE ) )
  {
    if ( dd1->dim != dd2->dim )
      return UT_FALSE;
  }

  if ( mask & dmDESCCMNT )
  {
    if ( !cs_uceq( dd1->desc, dd2->desc ))
      return UT_FALSE;
  }

  if ( mask & ( dmDESCLENS | dmDESCVALUE) )
  {
    if ( size1 != size2 )
      return UT_FALSE;
  }

  if ( mask & dmDESCCPTNAMES )
  {
    dmDescriptor *cpt1  = NULL;
    dmDescriptor *cpt2  = NULL;
    long          cptno = 0;

    for ( cptno = 1; cptno <= dd1->dim; cptno++ )
    {
      cpt1 = dmi_descriptor_get_cpt( dd1, cptno );
      cpt2 = dmi_descriptor_get_cpt( dd2, cptno );

      if ( !cs_uceq( cpt1->name, cpt2->name ))
        return UT_FALSE;
    }
  }
  return UT_TRUE;
}

/***************************************************************/
/*                                                             */
/* Function:   dmi_array_check_range                           */
/*                                                             */
/* Description:                                                */
/*   Check input pixel coordinate subarray is within           */
/*   the input image descriptor bounds.                        */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDescriptor  *dd       Image data descriptor            */
/*    long          *llpix    Lower Left subarray pixel coord. */
/*    long          *urpix    Upper Rght subarray pixel coord. */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*    long nitems   Total number of pixels in subarray         */
/*                  -1 if subarray extends beyond image bounds */
/*                                                             */
/***************************************************************/
long dmi_array_check_range( dmDescriptor *dd, long *llpix, long *urpix )
{
  long *axes   = NULL;
  long  nitems = 0;
  long  bad    = -1;
  long  naxes  = 0;
  long  ii     = 0;

  naxes = dmi_descriptor_get_array_dimensions( dd, &axes );
  nitems = 1;

  for ( ii = 0; ii < naxes; ii++ )
  {
    nitems *= urpix[ii] - llpix[ii] + 1;

    if ( ( llpix[ii] < 1 ) || 
         ( llpix[ii] > axes[ii] ) || 
         ( urpix[ii] < 1 ) || 
         ( urpix[ii] > axes[ii] ) )
    {
      free( axes );
      return bad;
    }
  }

  free( axes );
  return nitems;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_calc_image_weights                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*          long       *axes                                   */
/*          long        naxes                                  */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*          long  *weight                                      */
/*                                                             */
/***************************************************************/
long* dmi_calc_image_weights( long *axes, long naxes )
{
  long  axis_offset = 0;
  long *weight      = calloc( naxes, UT_SZ_I );

  weight[0] = 1;

 for ( axis_offset = 1; axis_offset < naxes; axis_offset++ )
   weight[axis_offset] = weight[axis_offset-1] * axes[axis_offset-1];

 return weight;
}


