/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmiddprint.c                                                */
/*                                                               */
/*   Print column descriptor information                         */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/
#include "dmi.h"

int dmi_descriptor_print_coord( dmDescriptor* wcs, char* name, char* list, char* ilist );

/***************************************************************/
/*                                                             */
/* Function:  dmi_col_print                                    */
/*                                                             */
/* Description:                                                */
/*   Print column data                                         */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDescriptor *col                           */
/*                 char         *bufp                          */
/*           const long          maxlen                        */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_col_print( const dmDescriptor* col, char* bufp, const long maxlen )
{
  char name[CARD_SIZE] = "";
  char array[TEXT_SIZE] = "";
  char type[WORD_SIZE] = "";
  char dlen[WORD_SIZE] = "";
  char range[CARD_SIZE] = "";
  char disp[2*WORD_SIZE] = "";
  char unit[2*WORD_SIZE] = "";

  char bin[CARD_SIZE] = "";
  char null[CARD_SIZE] = "";
  char buf1[TEXT_SIZE];
  char buf2[TEXT_SIZE]; 
  char fbuf[TEXT_SIZE]=""; 
  char buf[dmMAXLEN];
  long i = 0;
  char *ptr  = NULL;
  void *vptr = NULL;

  if ( !col )
  {
    sprintf( bufp, "Null column" );
    return;
  }

  ar_mix_type_name( col->type, type, WORD_SIZE );

  if ( col->length > 0 )
  {
    sprintf( dlen, "[%ld]", col->length );
    strcat( type, dlen );
  }
 
  ptr = name;
  ptr+= sprintf( ptr, "%s", col->name );

  if ( col->size > 0 )
  {
    ptr+= sprintf( ptr, "[%ld]", col->size );
  }

  if ( col->dim > 0 && col->axes )
  {
    ptr = array;
    ptr += sprintf( ptr, "(%ld", col->axes[0] );

    for ( i = 1; i <= col->dim; i++ )
    {
      ptr += sprintf( ptr, "x%ld", col->axes[i] );
    }
    strcat( array, ")" );
  }

  if ( col->range )
  {
    vptr = ar_gen_ptr( col->range, col->type, 0, 0 );
    ar_mix_print_generic( vptr, col->type, buf1, TEXT_SIZE );
    vptr = ar_gen_ptr( col->range, col->type, 0, 1 );
    ar_mix_print_generic( vptr, col->type, buf2, TEXT_SIZE );
    sprintf( range, "%s:%s", buf1, buf2 );
  }

  if ( col->bin )
  {
    ar_mix_print_generic( col->bin, col->type, buf1, TEXT_SIZE );
    sprintf( bin, "BIN=%s", buf1 );
  }

  if ( col->null ) 
  {
    ar_mix_print_generic( col->null, col->type, buf1, TEXT_SIZE );
    sprintf( null, "NULL=%s", buf1 );
  }

  if ( col->disp )
    sprintf( disp, "(%s)", col->disp );

  if ( col->unit )
    sprintf( unit, "%s", col->unit );
  else
    sprintf( unit, " " );

#if 0
{
 char desc[TEXT_SIZE] = "";
 if ( col->desc )
  sprintf( desc, "%s", col->desc );
 else
  sprintf( desc, " " );
}
#endif

  dmi_dd_print_flag( col, fbuf );
  sprintf( buf, "%12ld %4ld %-20s %s %s [%s] %s %s %s / %s %s",
          (long)col, col->no, name, type, array, unit, 
           range, bin, null, disp,  fbuf );

  cs_copy_siz( buf, bufp, maxlen );
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_dd_effective_type                            */
/*                                                             */
/* Description:                                                */
/*    We implement (scaled) physical table columns as dmCOORD  */
/*    descriptors, as opposed to the logical dmCOLUMN column   */
/*    descriptors. There are some occasions when we want to    */
/*    know that the physical descriptors are columns, and      */
/*    this routine handles that.                               */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDescriptor *dd                            */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptorType  descType                       */
/*                                                             */
/***************************************************************/
dmDescriptorType dmi_dd_effective_type( dmDescriptor* dd )
{
  dmBlock *block = NULL;

  if ( dd->descType == dmCOORD )
  {
     block = dd->block;

     if ( block->type == dmTABLE && ( dd->flag & (DM_FLAG_VISIBLE | DM_FLAG_PHYS )))
       return dmCOLUMN;
  }
  return dd->descType;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_dd_print_flag                                */
/*                                                             */
/* Description:                                                */
/*    WPrint value of descriptor flag                          */
/*                                                             */
/*  Input parameters:                                          */
/*           const  dmDescriptor *dd                           */
/*                                                             */
/*  Output parameters:                                         */
/*                  char   *buf                                */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_dd_print_flag( const dmDescriptor* dd, char* buf )
{
  long flag = 0;
  cs_copy( " ", buf );

  if ( !dd )
    return;

  flag = dd->flag;

  if ( flag & DM_FLAG_COPY ) strcat( buf, "copy of " );
  if ( flag & DM_FLAG_RESULT ) strcat( buf, "virtual " );
  if ( flag & DM_FLAG_BIN ) strcat( buf, "binning " );
  if ( flag & DM_FLAG_BINNED ) strcat( buf, "binned " );
  if ( flag & DM_FLAG_UNBUF ) strcat( buf, "unbuffered " );
  if ( flag & DM_FLAG_FILT ) strcat( buf, "filtered " );
  if ( flag & DM_FLAG_OPERAND ) strcat( buf, "operand " );
  if ( flag & DM_FLAG_ROW ) strcat( buf, "row descriptor " );
  if ( flag & DM_FLAG_LOGICAL ) 
    strcat( buf, "logical " );
  else
    if ( flag & DM_FLAG_COORD )
    {
      if ( flag & (DM_FLAG_VISIBLE | DM_FLAG_PHYS ))
        strcat( buf, " physical " );
      else
        strcat( buf, " coord " );
    } 

  if ( flag & DM_FLAG_AXIS )
    strcat( buf, " axis " );
  if ( flag & DM_FLAG_VECTOR ) strcat( buf, "vector " );
  if ( flag & DM_FLAG_CPT ) strcat( buf, "component " );
  if ( flag & DM_FLAG_IMG ) strcat( buf, " image " );

  if ( flag & DM_FLAG_SUBSPACE )strcat( buf, "subspace " );

  if ( flag & DM_FLAG_KEY ) strcat( buf, "key " );
  if ( flag & ( DM_FLAG_VISIBLE | DM_FLAG_RAW ) ) strcat( buf, "column" );
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_print_coord                       */
/*                                                             */
/* Description:                                                */
/*    Print coordinate descriptor information                  */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmDescriptor *wcs                          */
/*                                                             */
/*  Output parameters:                                         */
/*                  char         *name                         */
/*                  char         *list                         */
/*		    char         *ilist                        */
/*                                                             */
/*  Returns:                                                   */
/*           int  retval                                       */
/*                                                             */
/***************************************************************/ 
int dmi_descriptor_print_coord( dmDescriptor* wcs, char* name, char* list, char* ilist )
{
  char c[2];
  int retval = 0;

  strcpy( name, "" );

  if ( wcs )
  {
    if ( wcs->name )
      strcpy( name, wcs->name );

    if ( wcs->eltNames )
    {
      long ii = 0;

      sprintf( c, "%c", '[' );

      for( ii = 0; ii < wcs->vdim; ii++ )
      {
        strcat( name, c );
        strcat( name, wcs->eltNames[ii] ); 
        sprintf( c, "%c", ',' );
      }
      strcat( name, "]" );
    }

    strcpy( list, "" );
    strcpy( ilist, "" );
    retval = 1;
  }
  else
  {
    strcpy( list, "" );
    strcpy( ilist, "" );
   }
   return retval;
}

