/*                                                                
**  Copyright (C) 2002-2005,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmiddsec.c                                                  */
/*                                                               */
/*   Functions that handle descriptor get/set                    */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"


/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_set_name                          */
/*                                                             */
/* Description:                                                */
/*   Set descriptor name                                       */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor  *dd                                */
/*            char          *name                              */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_descriptor_set_name( dmDescriptor* dd, char* name )
{
  dmDescriptor *parent = NULL;

  if ( dd && name )
  {
    if ( dmverbose > 0 && dd->name )
      printf( "RENAME DD %s TO %s\n", dd->name, name );

    free( dd->name );
    dd->name = cs_dup( name );

    if ( dd->cptno > 0 )
    {
      parent = dd->parent;
      if ( parent && parent->cptNames )
      {
        free( parent->cptNames[dd->cptno-1] );
        parent->cptNames[dd->cptno-1] = cs_dup( name );
      }
    }
  }
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_set_element_names                 */
/*                                                             */
/* Description:                                                */
/*   Set descriptor element names                              */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor  *vector                            */
/*            dmElementType  etype                             */
/*            char         **enames                            */
/*            long           dim                               */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_descriptor_set_element_names( dmDescriptor* vector, dmElementType etype, char** enames, long dim )
{
  long elt_offset = 0;

  if ( !vector )
    return;

  if ( dim != vector->vdim )
  {
    dmu_err_set( DM_ERR_INTERNAL, "Element dimension mismatch" );
    return;
  }

  vector->etype = etype;

  if ( vector->eltNames )
  {
    ar_free_cp( vector->eltNames, vector->vdim );
  }

  vector->eltNames = calloc( dim, UT_SZ_P );

  for( elt_offset = 0; elt_offset < dim; elt_offset++ )
  {
    vector->eltNames[elt_offset] = cs_dup( enames[elt_offset] );
  }
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_set_range                         */
/*                                                             */
/* Description:                                                */
/*   Set descriptor range values                               */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor  *dd                                */
/*            dmDataTYpe     type  - data type                 */
/*            void          *tlmin - min value                 */
/*            void          *tlman - max value                 */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_descriptor_set_range( dmDescriptor* dd, dmDataType type, void* tlmin, void* tlmax )
{
  long vcpt_offset = 0;

  if ( !dd )
    return;

  if ( dd->dim > 1 )
  {
    for ( vcpt_offset = 0; vcpt_offset < dd->dim; vcpt_offset++ )
    {
      dmi_descriptor_set_cpt_range( dd, vcpt_offset+1, type, tlmin, tlmax );
    }
  }
  else /* dd->dim == 1 */
  {
    dmi_descriptor_set_cpt_range( dd, 0, type, tlmin, tlmax );
  }
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_set_cpt_range                     */
/*                                                             */
/* Description:                                                */
/*   Set descriptor component range values                     */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDescriptor  *ddp                          */
/*           const long           vcpt                         */
/*           const dmDataTYpe     type                         */
/*                 void          *tlmin - min value            */
/*                 void          *tlman - max value            */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
int dmi_descriptor_set_cpt_range( const dmDescriptor* ddp, const long vcpt, 
                           const dmDataType type, void* tlmin, void* tlmax )
{
  dmDataType dtype;
  void *ptr[2];
  GenericType tmp[2];
  dmDescriptor *dd = NULL;

  dd = (dmDescriptor*)ddp;

  if ( !dd )
    return dmFAILURE;
 
  if ( dd->descType == dmSUBSPACE )
  {
    /* Get column corresponding to subspace */
    dd = dmi_block_search_identifier( dd->block, dd->name );

    if ( !dd ) 
      return dmFAILURE;
  }
 
  if ( vcpt != 0 && dd->dim > 1 )
    dd = dd->cpts[vcpt-1];

  if ( !dd )
    return dmFAILURE;

  dtype = dd->type;

  if ( type != dtype )
  {
    ptr[0] = &tmp[0];
    ptr[1] = &tmp[1];
    ar_mix_numeric_cast( tlmin, type, dtype, ptr[0] );
    ar_mix_numeric_cast( tlmax, type, dtype, ptr[1] );
  }
  else 
  {
    ptr[0] = tlmin;
    ptr[1] = tlmax;
  }

  if ( !dd->range )
    dd->range = ar_gen_alloc( dtype, 2);
  if ( ptr[0] )
    ar_gen_set_elt( dd->range, 0, dtype, ptr[0] );
  if ( ptr[1] )
    ar_gen_set_elt( dd->range, 1, dtype, ptr[1] );

  return dmSUCCESS;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_set_length                        */
/*                                                             */
/* Description:                                                */
/*   Set descriptor length                                     */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDescriptor  *dd                           */
/*           const long           length                       */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_descriptor_set_length( dmDescriptor* dd, const long length )
{
  if ( dd )
    dd->length = length;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_set_no                            */
/*                                                             */
/* Description:                                                */
/*   Set descriptor no                                         */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDescriptor  *dd                           */
/*           const long           no                           */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_descriptor_set_no( dmDescriptor* dd, const long no )
{
  if ( dd )
    dd->no = no;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_set_disp                          */
/*                                                             */
/* Description:                                                */
/*   Set descriptor disp                                       */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDescriptor  *dd                           */
/*           const char          *disp                         */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_descriptor_set_disp( dmDescriptor* dd, const char* disp )
{
  if ( dd )
  {
    if ( dd->disp )
      free( dd->disp );
    dd->disp = cs_dup( disp );
  }
}

void dmi_descriptor_set_desc( dmDescriptor* dd, const char* desc )
{
 if ( dd )
 {
   if ( dd->desc ) free( dd->desc );
   dd->desc = cs_dup( desc );
 }
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_set_unit                          */
/*                                                             */
/* Description:                                                */
/*   Set descriptor unit string                                */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDescriptor  *dd                           */
/*           const char          *unit                         */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_descriptor_set_unit( dmDescriptor* dd, const char* unit )
{
  if ( dd )
  {
    if ( dd->unit )
      free( dd->unit );
    dd->unit = cs_dup( unit );
  }
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_set_null                          */
/*                                                             */
/* Description:                                                */
/*   Set descriptor null value                                 */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDescriptor  *dd                           */
/*                 dmDataType     type                         */
/*                 void          *value                        */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           int  0 or 1                                       */
/*                                                             */
/***************************************************************/
int dmi_descriptor_set_null( dmDescriptor* dd, dmDataType type, void* value )
{
  dmDescriptor *cpt_dd = NULL;
  dmDataType    dtype;
  GenericType   tmp;
  void         *ptr = NULL;
  char         *ccopy = NULL;
  long          vcpt_offset = 0;

  if ( !value )
    return 0;

  dtype = dd->type;

  /* Is desired type the intrinsic type of the column? */

  if ( type != dtype )
  {
    ptr = &tmp;
    ar_mix_numeric_cast( value, type, dtype, ptr );   
  }
  else
    if ( type == DT_CHAR )
    {
      ccopy = cs_dup( *(char**)value );
      ptr = &ccopy;
    }
    else
    {
      ptr = value;
    }

  if ( !dd->null )
    dd->null = ar_gen_alloc( dtype, 1 );

  ar_gen_set_elt( dd->null, 0, dtype, ptr );

  if ( dd->dim > 1 )
  {
    for ( vcpt_offset = 0; vcpt_offset < dd->dim; vcpt_offset++ )
    {
      cpt_dd = dd->cpts[vcpt_offset];

      if ( !cpt_dd->null )
        cpt_dd->null = ar_gen_alloc( dd->type, 1 );
      ar_gen_set_elt( cpt_dd->null, 0, dtype, ptr );
    }
  }
  return 1;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_set_bin                           */
/*                                                             */
/* Description:                                                */
/*   Set descriptor null value                                 */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDescriptor  *dd                           */
/*                 dmDataType     type                         */
/*                 void          *value                        */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           int  0 or 1                                       */
/*                                                             */
/***************************************************************/
int dmi_descriptor_set_bin( dmDescriptor* dd, dmDataType type, void* value )
{
  dmDescriptor *cpt_dd = NULL;
  dmDataType    dtype;
  GenericType   tmp;
  void         *ptr = NULL;
  long          vcpt_offset;

  if ( !value )
    return 0;

  dtype = dd->type;


  /* Is desired type the intrinsic type of the column? */

  if ( type != dtype )
  {
    ptr = &tmp;
    ar_mix_numeric_cast( value, type, dtype, ptr );   
  }
  else
  {
    ptr = value;
  }

  if ( !dd->bin )
    dd->bin = ar_gen_alloc( dtype, 1 );
  ar_gen_set_elt( dd->bin, 0, dtype, ptr );

  if ( dd->dim > 1 )
  {
    for ( vcpt_offset = 0; vcpt_offset < dd->dim; vcpt_offset++ )
    {
      cpt_dd = dd->cpts[vcpt_offset];
      if ( !cpt_dd->null )
        cpt_dd->null = ar_gen_alloc( dd->type, 1 );
      ar_gen_set_elt( cpt_dd->null, 0, dtype, ptr );
    }
  }

  return 1;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_set_cpt                           */
/*                                                             */
/* Description:                                                */
/*   Set vector names                                          */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDescriptor  *vector                       */
/*                 long           cptno                        */
/*                 dmDescriptor  *ddc                          */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_descriptor_set_cpt( dmDescriptor* vector, long cptno, dmDescriptor* ddc )
{
  if ( !vector )
   return;

  vector->cptNames[cptno-1] = cs_dup( ddc->name );
  vector->cpts[cptno-1] = ddc; 
  ddc->cptno = cptno;
  ddc->parent = vector;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_set_coord                         */
/*                                                             */
/* Description:                                                */
/*   Temporary                                                 */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDescriptor  *w                            */
/*           const long           no                           */
/*                 dmDescriptor  *p                            */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_descriptor_set_coord( dmDescriptor* w, const long no, dmDescriptor* p )
{
  if ( w && w->coords )
    w->coords[no-1] = p;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_set_axes                          */
/*                                                             */
/* Description:                                                */
/*    Set axes length values                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDescriptor  *col                          */
/*           const long           axlen                        */
/*           const long           naxes                        */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_descriptor_set_axes( dmDescriptor* col, const long* axlen, const long naxes )
{
  long ii = 0;

  if ( col->axes )
    free( col->axes );

  col->naxes = naxes; 

  if ( naxes > 0 )
  {
    col->axes = (long*)calloc( naxes, UT_SZ_I );

    for ( ii = 0; ii < naxes; ii++ )
    {
      col->axes[ii] = axlen[ii];
    }
  }
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_set_datatype                      */
/*                                                             */
/* Description:                                                */
/*    Set descriptor data type                                 */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDescriptor  *dd                           */
/*           const dmDataType     typen                        */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_descriptor_set_datatype( dmDescriptor* dd, const  dmDataType type)
{
  dd->type = type;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_set_aux                           */
/*                                                             */
/* Description:                                                */
/*    Set descriptor aux element                               */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDescriptor  *dd                           */
/*           const void          *aux                          */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_descriptor_set_aux( dmDescriptor* dd, const void* aux )
{
  if ( dd )
    dd->aux = (void*)aux;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_set_kdd                           */
/*                                                             */
/* Description:                                                */
/*    Set descriptor kernel descriptor                         */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDescriptor  *dd                           */
/*           const void          *kernel_dd                    */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_descriptor_set_kdd( dmDescriptor* dd, const void* kernel_dd )
{
  if ( dd ) 
    dd->kernel_dd = (void*)kernel_dd;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_check_update                      */
/*                                                             */
/* Description:                                                */
/*    Set descriptor update flag. If descriptor is NULL,       */
/*    return false                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDescriptor  *dd                           */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           dmBol  UT_TRUE or UT_FALSE                        */
/*                                                             */
/***************************************************************/
dmBool dmi_descriptor_check_update( dmDescriptor* dd )
{
  if ( dd )
    return dd->update;
  else
    return UT_FALSE;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_set_update                        */
/*                                                             */
/* Description:                                                */
/*    Set descriptor update flag. Update dd's status in block  */
/*    return false                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDescriptor  *dd                           */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_descriptor_set_update( dmDescriptor* dd )
{
  if ( dd )
  {
    dd->update = UT_TRUE;
    dmi_block_set_update( dd->block );
  }
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_clear_update                      */
/*                                                             */
/* Description:                                                */
/*    Set descriptor update flag to false                      */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDescriptor  *dd                           */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_descriptor_clear_update( dmDescriptor* dd )
{
  if ( dd )
    dd->update = UT_FALSE;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_clear_update                      */
/*                                                             */
/* Description:                                                */
/*    Set descriptor varlen flag                               */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDescriptor  *dd                           */
/*                 dmBool         varlen                       */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_descriptor_set_varlen( dmDescriptor* dd, dmBool varlen )
{
  if ( dd )
    dd->varlen = varlen;
}

