/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmids.c                                                     */
/*                                                               */
/*   Functions that handle datasets                              */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/
#include "dmi.h"


/***************************************************************/
/*                                                             */
/* Function:  dmi_dataset_alloc                                */
/*                                                             */
/* Description:                                                */
/*   Allocate memory for a dm level dataset given the kernel   */
/*   ds pointer                                                */
/*                                                             */
/*  Input parameters:                                          */
/*           const char       *dsname     k                    */
/*                 dmkDataset *kds                             */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           dmDataset *ds                                     */
/*                                                             */
/***************************************************************/
dmDataset* dmi_dataset_alloc( const char* dsname, dmkDataset* kds )
{
  dmDataset* ds = NULL;

  ds = calloc( 1, sizeof( dmDataset ) );

  if ( ds )
  {
    ds->name = cs_dup( dsname );
    ds->kernel_ds = kds;
    ds->current = NULL;
    ds->current_no = 0;

    ds->index_blocks = NULL;
    ds->nblocks = 0;
    ds->maxblocks = 0;

    ds->blocks = NULL;
    ds->nvblocks = 0;
    ds->maxvblocks = 0;
  }
  return ds;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_dataset_get_kds                              */
/*                                                             */
/* Description:                                                */
/* Manage the kernel-dependent dataset flag.                   */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDataset  *ds                              */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           dmkDataset *kds                                   */
/*                                                             */
/***************************************************************/
dmkDataset* dmi_dataset_get_kds( const dmDataset* ds )
{
  if ( ds )
    return (void*)ds->kernel_ds;
  else
    return NULL;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_dataset_set_current                          */
/*                                                             */
/* Description:                                                */
/*    Set the current block number                             */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDataset  *ds                              */
/*                 dmBlock    *block                           */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_dataset_set_current( dmDataset* ds, dmBlock* block )
{
  if ( ds )
  {
    ds->current = block;

    if ( block )
      ds->current_no = block->no;
    else 
      ds->current_no = 0;
  }
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_dataset_get_current                          */
/*                                                             */
/* Description:                                                */
/*    Get the current block number                             */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDataset  *ds                              */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           long block_no                                     */
/*                                                             */
/***************************************************************/
long dmi_dataset_get_current( const dmDataset* ds )
{
  if ( !ds )
  {
    dmu_err_set( DM_ERR_NULL, "Null dataset" );
    return 0;
  }
  return ds->current_no;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_dataset_select_block_name                    */
/*                                                             */
/* Description:                                                */
/*    Get the current block name                               */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDataset  *ds                              */
/*                 char       *name                            */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           char *name                                        */
/*                                                             */
/***************************************************************/
char* dmi_dataset_select_block_name( dmDataset* ds, char* name )
{
  char *ptr = NULL;

  if ( !name )
  {
    /* Default block name is chosen to be same as dsname without path. */
    ptr = strrchr( ds->name, '/' );

    if ( ptr )
      name = ++ptr;
    else
      name = ds->name;
   }

  return cs_dup( name );
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_dataset_mode                                 */
/*                                                             */
/* Description:                                                */
/*    Get dataset mode                                         */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDataset  *ds                              */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           int UT_TRUE or UT_FALSE                           */
/*                                                             */
/***************************************************************/
int dmi_dataset_mode( dmDataset* ds )
{
  if ( ds )
    return ds->mode;
  else
    return UT_FALSE;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_dataset_set_mode                             */
/*                                                             */
/* Description:                                                */
/*    Set dataset mode                                         */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDataset  *ds                              */
/*           const int         mode                            */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_dataset_set_mode( dmDataset* ds, const int mode )
{
  if ( ds )
    ds->mode = mode;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_dataset_close                                */
/*                                                             */
/* Description:                                                */
/*    Close dataset                                            */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDataset  *ds                              */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           dmErrCode err                                     */
/*                                                             */
/***************************************************************/
dmErrCode dmi_dataset_close( dmDataset* ds )
{
  long blockNo = 0;
  dmBlock *block = NULL;

  if ( !ds )
    return dmFAILURE;

  if ( ds->blocks )
  {
    for ( blockNo = 1; blockNo <= ds->nvblocks; blockNo++ )
    {
      block = ds->blocks[blockNo-1];
      dmi_block_free( block );
    }
    free( ds->blocks );
  }

  if ( ds->index_blocks )
  {
    for ( blockNo = 1; blockNo <= ds->nblocks; blockNo++ )
    {
      dmi_dataset_block_free( ds, blockNo );
    }
    free( ds->index_blocks );
  }

  free( ds->name );
  free( ds );

  return dmu_err_get();
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_dataset_get_n                                */
/*                                                             */
/* Description:                                                */
/*    How many blocks in the dataset?                          */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDataset  *ds                              */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           long   n                                          */
/*                                                             */
/***************************************************************/
long dmi_dataset_get_n( const  dmDataset* ds )
{
  if ( ds )
    return ds->nblocks;
  else
   return 0;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_dataset_get_name                             */
/*                                                             */
/* Description:                                                */
/*    Return dataset name                                      */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDataset  *ds                              */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           char *dname                                       */
/*                                                             */
/***************************************************************/
char* dmi_dataset_get_name( const dmDataset* ds )
{
  static char *dname=" ";

  if ( ds )
    return ds->name;
  else
    return dname;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_dataset_get_nv                               */
/*                                                             */
/* Description:                                                */
/*    Return dataset nvblocks                                  */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDataset  *ds                              */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           long nvblocks                                     */
/*                                                             */
/***************************************************************/
long dmi_dataset_get_nv( const dmDataset* ds )
{
  if ( ds )
    return ds->nvblocks;
  return 0;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_dataset_get_block                            */
/*                                                             */
/* Description:                                                */
/*    Return a block given the block number                    */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDataset  *ds  dataset                     */
/*           const long        no - block number               */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           dmBlock *block                                    */
/*                                                             */
/***************************************************************/
dmBlock* dmi_dataset_get_block( const dmDataset* ds, const long no )
{
 if ( ds && no <= ds->nvblocks && no> 0  )
   return ds->blocks[no-1];
 else
   return NULL;
}

