/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmidslib.c                                                  */
/*                                                               */
/*   Functions that work with datasets                           */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"

/***************************************************************/
/*                                                             */
/* Function:  dmi_dataset_block_freetor                        */
/*                                                             */
/* Description:                                                */
/*   Free block from dataset                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDataset  *ds                              */
/*           const long        no - block number               */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*              void                                           */
/*                                                             */
/***************************************************************/
void dmi_dataset_block_free( dmDataset* ds, const long no )
{
  dmIndexBlock *index_block = dmi_ds_index( ds, no );

  if ( index_block )
  {  
    free( index_block->name );
    free( index_block );
    ds->index_blocks[no-1] = NULL;
  }
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_dataset_block_alloc r                        */
/*                                                             */
/* Description:                                                */
/*   Allocate memory for block in index block                  */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDataset  *ds                              */
/*           const long        no - block number               */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*              long    bloc_type                              */
/*                                                             */
/***************************************************************/
long dmi_dataset_block_alloc( dmDataset* ds, const long no )
{
  dmIndexBlock *index_block = NULL;
  dmIndexBlock *tblock = NULL;
  long nmax = 0;
  long i = 0;

  if ( no <= 0 )
    return 0;
  index_block =(dmIndexBlock*)calloc( 1, sizeof( dmIndexBlock )); 

  if ( !index_block ) 
     return 0;
 
  if ( no <= ds->nblocks )
  {
    /* Insert: Shift all blocks up one */
    nmax = ds->nblocks+1;
    dmi_alloc_to_list( (void***)&ds->index_blocks, &ds->nblocks, &ds->maxblocks, index_block, nmax );

    /* ds->nblocks has increased: shift blocks up as required */  
    for ( i = ds->nblocks-1; i >= no; i-- )
    {
      tblock =  ds->index_blocks[i-1];
      ds->index_blocks[i] = tblock;
      tblock->no++;
    } 
  }
  else
  {
    /* Allocate blocks: note ds->blocks may be realloced */
    dmi_alloc_to_list( (void***)&ds->index_blocks, &ds->nblocks, &ds->maxblocks, index_block, no );
  }

  /* Tag the new block to the list */
  ds->index_blocks[no-1] = index_block;
  index_block->ds = ds;
  index_block->no = no;
  return no;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_dataset_get_block_type                       */
/*                                                             */
/* Description:                                                */
/*   return block type given block number                      */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDataset  *ds                              */
/*           const long        no - block number               */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*             dmBlockType  type                               */
/*                                                             */
/***************************************************************/
dmBlockType dmi_dataset_get_block_type( const dmDataset* ds, const long no )
{
  dmIndexBlock *index_block = dmi_ds_index( ds, no );

  if ( index_block )
    return index_block->type;
  return dmUNKNOWNBLOCK;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_dataset_set_block_type                       */
/*                                                             */
/* Description:                                                */
/*   set block type given type and block number                */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDataset  *ds                              */
/*                 long        no - block number               */
/*           const dmBlockType type - block type               */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_dataset_set_block_type( dmDataset* ds, long no, const dmBlockType type )
{
  dmIndexBlock *index_block = dmi_ds_index( ds, no );

  if ( index_block )
    index_block->type = type;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_ds_index                                     */
/*                                                             */
/* Description:                                                */
/*    Return block index number give dataset and block number  */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDataset  *ds                              */
/*           const long        no - block number               */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*            dmIndexBlock *index_block                        */
/*                                                             */
/***************************************************************/
dmIndexBlock* dmi_ds_index( const dmDataset* ds, const long no )
{
  dmIndexBlock *index_block = NULL;

  if ( ds && no > 0 && no <= ds->nblocks )
    index_block = ds->index_blocks[no-1];
  return index_block;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_block_search                                 */
/*                                                             */
/* Description:                                                */
/*    This is the way we figure out which block we want to     */
/*    open when  given its name.                               */
/*    The routine returns                                      */
/*      -1 for failure                                         */
/*       0 for default block                                   */
/*      >0 for specific block number                           */
/*    It supports                                              */
/*      1) Numeric block ID       data.fits[3]                 */
/*      2) Block name             data.fits[events]            */
/*      3) Default block          data.fits                    */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDataset  *ds                              */
/*           const char       *name - block name               */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*            long  block number or error                      */
/*                                                             */
/***************************************************************/
long dmi_block_search( const dmDataset* ds, const char* name )
{
  char *tail = NULL;
  dmIndexBlock *index_block = NULL;
  long no = 0;

  if (  !ds || !name || cs_is_blank( name ))
    return 0;  /* Use most interesting block */

  /* Fail silently if no blocks yet */   
  if ( ds->nblocks == 0 )
    return -1;

  /* Check for a numeric block ID */

  no = strtol( name, &tail, 10 );

  if ( !tail || !(*tail) )
  {
    if ( no > 0 && no <= ds->nblocks )
      return no;
   }

  /* Check for an explicit block name */
  for ( no = 1; no <= ds->nblocks; no++ )
  { 
    index_block = ds->index_blocks[no-1];
    if ( index_block->name && cs_uceq( index_block->name, name ))
      return no;
  } 

  /* Fail silently */
   return -1;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_dataset_get_block_name                       */
/*                                                             */
/* Description:                                                */
/*    Return block name given number                           */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDataset  *ds                              */
/*           const long        no  - block number              */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*            char *name                                       */
/*                                                             */
/***************************************************************/
char* dmi_dataset_get_block_name( const dmDataset* ds, const long no )
{
  dmIndexBlock *index_block = dmi_ds_index( ds, no );

  if ( index_block )
     return index_block->name;
  return NULL;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_dataset_set_block_name                       */
/*                                                             */
/* Description:                                                */
/*    Set block name given number and name string              */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDataset  *ds                              */
/*           const long        no  - block number              */
/*           const char       *name -name string               */ 
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_dataset_set_block_name( dmDataset* ds, const long no, const char* name )
{
  dmIndexBlock *index_block = dmi_ds_index( ds, no );

  if ( index_block )
  {
     if ( index_block->name )
       free( index_block->name );
     index_block->name  = cs_dup( name );
  }
}


