/*                                                                
**  Copyright (C) 2003-2005,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmifiltrows.c                                               */
/*                                                               */
/*   Functions that handle row filtering                         */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"

/* Since these two functions are guaranteed to operate on every
 * row, they need to be as optimal as possible.  This is why
 * they were split into two functions.  We don't want normal 
 * filtering bogged down with the col=NULL test code when
 * there is no col=NULL filter in place 
 */

extern long dmi_filter_rows_simple( dmBuffer* buffer, void* data );
extern long dmi_filter_rows_with_nulltest( dmBuffer* buffer, void* data );


/***************************************************************/
/*                                                             */
/*  Function:  dmi_parse_clause_operator                       */
/*                                                             */
/*  Description:                                               */
/*     We have now read raw_row1 to raw_row2 into the raw row  */
/*    buffer.  Now mark which rows pass the row filter.        */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBuffer  *buffer                                */
/*            void      *data                                  */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*              long ngood - number of good rows               */
/*                                                             */
/***************************************************************/
long dmi_filter_rows_simple( dmBuffer* buffer, void* data )
{
  /* To optimize, we don't allow general ND DM-vectors here: only */
  /* 1D and 2D DM-vectors are currently supported. */

  double minval = 0.0;
  double maxval = 0.0;
  double ypos   = 0.0;
  double df     = 0.0;
  double xpos   = 0.0;

  dmSubspace  *rts = NULL;
  dmRangeList *range = NULL;
  dmRangeList *yrange = NULL;  /* For 2D vectors */
  dmDescriptor *sub_dd = NULL;
  dmDescriptor *dd = NULL;
  dmFilter     *filter = NULL;
  dmDescriptor **ddlist = NULL;
  dmDescriptor **sub_ddlist = NULL;
  dmFilter    ***filter_list = NULL;
  dmBool        *paccept = NULL;
  dmRTSCache    *cache = NULL;

  regRegion *region = NULL;

  void *ptr = NULL;
  long *offsets = NULL;

  long nraw = 0;
  long rrow_offset = 0;   /* Extra offset in raw row buffer relative to rrow_start_offset */
  long col_offset = 0;
  long cpt_offset = 0;
  long last_contiguous = -1;
  long ncpts = 0;
  long ngood = 0;
  long range_no = 0;
  long index = 0;
  long vrow_start_offset = 0;
  long nfilt = 0; /* number of filters in filter_list */

  dmBool accept = UT_FALSE;
  dmBool eval = UT_FALSE;
  int invert = 0;
  int yeval = 0;

  rts = buffer->subspace;
  cache = rts->cache;

  if ( cache->null )
    return 0;

  nraw = buffer->rrow_seg2 - buffer->rrow_seg1 + 1;

  /* 
   * On entry to this routine, the vrow1:vrow2 range gives the virtual rows
   * already read. If soffset > 0 we are reading another segment 
   * into an already partially full buffer.
   */
  if ( buffer->lastrow > 0 && buffer->soffset > 0 )
    vrow_start_offset = buffer->lastrow - buffer->vrow1 + 1;
  else
    vrow_start_offset = 0;

  ncpts = rts->ncpts;
  invert = rts->invert;

/* 
 * Here we do a minor optimization to put all the categories of filtered
 * descriptor in one list; this removes one loop and one if statement from
 * the inner loops, and removes calls to dmi_subspace_col_get_filter.
 */
  nfilt = cache->n_filt;
  ddlist = cache->ddlist;
  sub_ddlist = cache->sub_ddlist;
  filter_list = cache->filter_list;
  offsets = cache->col_offsets;
  paccept = cache->accept;

   /* Loop over rows just read and filter them */
  for ( rrow_offset = 0; rrow_offset < nraw; rrow_offset++ )
  {
    /* Absolute raw row is         buffer->rrow_seg1 + rrow_offset    *
     * This maps to buffer offset  buffer->soffset+rrow_offset    *
    */

    /* For each component */
    for ( cpt_offset = 0; cpt_offset < ncpts; cpt_offset++ )
    {
      /* For each filtered column */
      paccept[cpt_offset] = 0;	

      for ( index = 0; index < nfilt; index++ )
      {
        filter = filter_list[index][cpt_offset];

        if ( filter->null )
        {
          paccept[cpt_offset] = 0;
          break;
        }
        yrange = NULL;
        dd = ddlist[index];
        col_offset = offsets[index];
        sub_dd = sub_ddlist[index];
        eval = 1;
        /* HERE WE LOOK FOR THE DATA IN THE BUFFER */
        ptr = (char*)data + dmi_buffer_offset( buffer, col_offset, rrow_offset + buffer->soffset );

        /* IMPLEMENT vdim rather than dim here when sub_dd correctly has dim = 4, vdim = 2 for sky */
        range = &filter->ranges[0];

        if ( sub_dd->dim > 1 ) 
          yrange = &filter->ranges[1];
        region = filter->region;


        /*
         * NOTE: df is intentionally set to 0.0 for real types because range
         * limits are interpreted differently for integers and reals.
         * Specifically, x=a:b is interpreted as x in [a,b] (lower and upper
         * bound closed) for integer types and x in [a,b) (lower bound closed,
         * upper bound open) for real types.  However, we currently aren't
         * getting the lower boundary quite right for reals, since we test
         * for xpos == minval when we really want something like
         * abs(xpos - minval) <= DBL_EPSILON.
         */

        switch(  dd->type )
        {
           case DT_INT2:
             xpos = (double)*((short*)ptr);
             df = 0.1;
             break;
           case DT_INT4:
             xpos = (double)*((long*)ptr);
             df = 0.1;
             break;
           case DT_REAL4:
             xpos = (double)*((float*)ptr);
             df = 0.0;
             break;
           case DT_REAL8:
             xpos = (double)*((double*)ptr);
             df = 0.0;
             break;
           case DT_UINT1:
             xpos = (double)*((unsigned char*)ptr);
             df = 0.1;
             break;
           case DT_UINT2:
             xpos = (double)*((unsigned short*)ptr);
             df = 0.1;
             break;
           case DT_UINT4:
             xpos = (double)*((unsigned long*)ptr);
             df = 0.1;
             break;
           case DT_LOGICAL:
             xpos = (double)*((dmBool*)ptr);
             df = 0.1;
             break;
           case DT_CHAR:
             if ( range )
             {
               paccept[cpt_offset] = 0; /* Will fail unless one matches */
               for ( range_no = 0; range_no < range->n; range_no++ )
               {
                 if ( dmu_text_in_range( ptr, ((char**)range->mins)[range_no], ((char**)range->maxes)[range_no] ))
                 {
                   paccept[cpt_offset] = 1; break;
                 }
               }
             }
             eval = 0;
             break;
           case DT_BIT:
             if ( range ) 
             {
               paccept[cpt_offset] = 0; /* Will fail unless one matches */
               for ( range_no = 0; range_no < range->n; range_no++ )
               {
                 if ( dmu_bit_match( ptr, dd->length, ((char**)range->mins)[range_no] ))
                 {
    	           paccept[cpt_offset] = 1; break;
                 }
               }
             } 
             eval = 0;
             break;
           default:
             printf( "NO DT support\n" );
         }    

   if ( yrange || region )
   {
     yeval = 1;
     switch(  dd->type )
     {
     case DT_INT2:
      ypos = (double)*((short*)ptr+1);
      df = 0.1;
      break;
     case DT_INT4:
      ypos = (double)*((long*)ptr+1);
      df = 0.1;
      break;
     case DT_REAL4:
      ypos = (double)*((float*)ptr+1);
      df = 0.0;
      break;
     case DT_REAL8:
      ypos = (double)*((double*)ptr+1);
      df = 0.0;
      break;
     case DT_UINT1:
      ypos = (double)*((unsigned char*)ptr+1);
      df = 0.1;
      break;
     case DT_UINT2:
      ypos = (double)*((unsigned short*)ptr+1);
      df = 0.1;
      break;
     case DT_UINT4:
      ypos = (double)*((unsigned long*)ptr+1);
      df = 0.1;
      break;
     case DT_LOGICAL:
      ypos = (double)*((dmBool*)ptr+1);
      df = 0.1;
      break;
     case DT_CHAR:
/* This is not yet implemented correctly. Of course, we don't
   currently have any files with string vector columns in.
   You'd need to offset ptr with the string length.
*/
      if ( yrange )
      {
       paccept[cpt_offset] = 0;
       for ( range_no = 0; range_no < yrange->n; range_no++ )
       {
	if ( dmu_text_in_range( ptr, ((char**)range->mins)[range_no], ((char**)range->maxes)[range_no] ))
	{
	 paccept[cpt_offset] = 1; break;
	}
       }
      }
      yeval = 0;
      break;
     case DT_BIT:
/* This is not yet implemented correctly. Of course, we don't
   currently have any files with bit vector columns in.
   You'd need to offset ptr with the byte length of a field
       */
      if ( yrange )
      {
       paccept[cpt_offset] = 0;
       for ( range_no = 0; range_no < yrange->n; range_no++ )
       {
	if ( dmu_bit_match( ptr, dd->length * 8, ((char**)yrange->mins)[range_no] ))
	{
	 paccept[cpt_offset] = 1; break;
	}
       }
      } 
      yeval = 0;
      break;
     default:
      printf( "NO DT support\n" );
     }          
    }
    
    if ( range && range->n   )
    {
/* Must be inside one of the ranges or row fails */
     if ( eval )
     {
/* For numeric types, eval now; other types already eval'ed */
      paccept[cpt_offset] = 0;	
      for ( range_no = 0; range_no < range->n; range_no++ ) 
      {
       minval = ((double*)range->mins)[range_no] - df;
       maxval =((double*)range->maxes)[range_no] + df;
       if ( xpos == minval || ( xpos > minval && xpos < maxval ))
       {
	paccept[cpt_offset] = 1;
	break;
       }
      }
     }
/* This test to be done even if eval is false */
     if ( !paccept[cpt_offset] ) break;
    }

    if ( yrange && yrange->n && yeval )
    {
     paccept[cpt_offset] = 0;	
     for ( range_no = 0; range_no < yrange->n; range_no++ )
     {
      minval = ((double*)yrange->mins)[range_no] - df;
      maxval =((double*)yrange->maxes)[range_no] + df;
      if ( ypos == minval || ( ypos > minval && ypos < maxval ))
      {
       paccept[cpt_offset] = 1; break;
      }
     }
/* This one can be inside the if(yeval) test since if !yeval we already caught it above */
     if ( !paccept[cpt_offset] ) break;
    } /* End eval */       

    if ( region )
    {
     paccept[cpt_offset] = regInsideRegion( region, xpos, ypos );
     if ( !paccept[cpt_offset] ) break;
    }
   } /* End col */
  } /* End cpt */
  accept = 0;
 
  for ( cpt_offset = 0; cpt_offset < ncpts; cpt_offset++ )
   if ( paccept[cpt_offset] ) accept = 1;

  if ( invert )
   accept = !accept;
  
  if ( accept ) 
  {
/* 
   What is the raw buffer position for a given virtual row ?
   How many virtual rows are contiguous? (so we can optimize the memcpy)
 */
   if ( last_contiguous < 0 )
    last_contiguous = ngood+vrow_start_offset;
   buffer->index[ngood+vrow_start_offset] = rrow_offset + buffer->soffset;
   buffer->contiguous[last_contiguous]++;
   buffer->contig_start[ngood+vrow_start_offset] = last_contiguous;
   ngood++;
  } else {
   last_contiguous = -1;
  }
 } /* End row */

 return ngood;
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_filter_rows_with_nulltest                   */
/*                                                             */
/*  Description:                                               */
/*     We have now read raw_row1 to raw_row2 into the raw row  */
/*      buffer.   Now mark which rows pass the row filter.     */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBuffer  *buffer                                */
/*            void      *data                                  */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*              long ngood - number of good rows               */
/*                                                             */
/***************************************************************/
long dmi_filter_rows_with_nulltest( dmBuffer* buffer, void* data )
{
/* To optimize, we don't allow general ND DM-vectors here: only 1D and 2D DM-vectors
are currently supported.
 */
 double minval, maxval;
 double ypos;
 double df;
 double xpos;

 dmSubspace* rts;
 dmRangeList* range;
 dmRangeList* yrange;  /* For 2D vectors */
 dmDescriptor* sub_dd;
 dmDescriptor* dd;
 dmFilter* filter;
 dmDescriptor** ddlist;
 dmDescriptor** sub_ddlist;
 dmFilter*** filter_list;
 dmBool* paccept;
 dmRTSCache* cache;

 regRegion* region;

 void* ptr;
 void* nullval;
 long* offsets;

 long nraw;
 long rrow_offset;   /* Extra offset in raw row buffer relative to rrow_start_offset */
 long col_offset;
 long cpt_offset;
 long last_contiguous = -1;
 long ncpts;
 long ngood = 0;
 long range_no;
 long index;
 long vrow_start_offset;
 long nfilt; /* number of filters in filter_list */
 long nbytes;

 dmBool accept;
 dmBool eval;
 int invert;
 int yeval;
 int isnull;

 rts = buffer->subspace;
 cache = rts->cache;

 if ( cache->null )
  return 0;

 nraw = buffer->rrow_seg2 - buffer->rrow_seg1 + 1;

/* 
 * On entry to this routine, the vrow1:vrow2 range gives the virtual rows
 * already read. If soffset > 0 we are reading another segment 
 * into an already partially full buffer.
 */
 if ( buffer->lastrow > 0 && buffer->soffset > 0 )
  vrow_start_offset = buffer->lastrow - buffer->vrow1 + 1;
 else
  vrow_start_offset = 0;


 ncpts = rts->ncpts;
 invert = rts->invert;

/* 
 * Here we do a minor optimization to put all the categories of filtered
 * descriptor in one list; this removes one loop and one if statement from
 * the inner loops, and removes calls to dmi_subspace_col_get_filter.
 */
 nfilt = cache->n_filt;
 ddlist = cache->ddlist;
 sub_ddlist = cache->sub_ddlist;
 filter_list = cache->filter_list;
 offsets = cache->col_offsets;
 paccept = cache->accept;

/* Loop over rows just read and filter them */
 for ( rrow_offset = 0; rrow_offset < nraw; rrow_offset++ )
 {
/* 
 * Absolute raw row is         buffer->rrow_seg1 + rrow_offset    *
 * This maps to buffer offset  buffer->soffset+rrow_offset    *
 */

/* For each component */
  for ( cpt_offset = 0; cpt_offset < ncpts; cpt_offset++ )
  {
/* For each filtered column */
   paccept[cpt_offset] = 0;	
   for ( index = 0; index < nfilt; index++ )
   {
    filter = filter_list[index][cpt_offset];
    if ( filter->null )
    {
     paccept[cpt_offset] = 0;
     break;
    }
    dd = ddlist[index];
    nullval = dd->null;
    col_offset = offsets[index];
    sub_dd = sub_ddlist[index];
    eval = 1;
/* HERE WE LOOK FOR THE DATA IN THE BUFFER */
    ptr = (char*)data + dmi_buffer_offset( buffer, col_offset, rrow_offset + buffer->soffset );

/* IMPLEMENT vdim rather than dim here when sub_dd correctly has dim = 4, vdim = 2 for sky */
    range = &filter->ranges[0];
    if ( sub_dd->dim > 1 ) 
     yrange = &filter->ranges[1];
    else 
     yrange = NULL;
    region = filter->region;


    /*
     * NOTE: df is intentionally set to 0.0 for real types because range
     * limits are interpreted differently for integers and reals.
     * Specifically, x=a:b is interpreted as x in [a,b] (lower and upper
     * bound closed) for integer types and x in [a,b) (lower bound closed,
     * upper bound open) for real types.  However, we currently aren't
     * getting the lower boundary quite right for reals, since we test
     * for xpos == minval when we really want something like
     * abs(xpos - minval) <= DBL_EPSILON.
     */

    switch(  dd->type )
    {
    case DT_INT2:
     xpos = (double)*((short*)ptr);
     df = 0.1;
     if ( range && range->incnull && nullval )
     {
      if ( *((short*)ptr) == *((short*)nullval) )
      {
       if ( range->incnull == 1 )
       {
	paccept[cpt_offset] = 1;
	eval = 0;
       }
       else
       {
	paccept[cpt_offset] = 0;
	eval = 0;
       }
      }
      else if ( range->incnull == -1 )
      {
       paccept[cpt_offset] = 1;
      }
     }
     break;
    case DT_INT4:
     xpos = (double)*((long*)ptr);
     df = 0.1;
     if ( range && range->incnull && nullval )
     {
      if ( *((long*)ptr) == *((long*)nullval) )
      {
       if ( range->incnull == 1 )
       {
	paccept[cpt_offset] = 1;
	eval = 0;
       }
       else
       {
	paccept[cpt_offset] = 0;
	eval = 0;
       }
      }
      else if ( range->incnull == -1 )
      {
       paccept[cpt_offset] = 1;
      }
     }
     break;
    case DT_REAL4:
     xpos = (double)*((float*)ptr);
     df = 0.0;
     if ( range && range->incnull )
     {
      if ( isnan( *((float*)ptr) ))
       isnull = UT_TRUE;
      else if ( nullval && *((float*)ptr) == *((float*)nullval) )
       isnull = UT_TRUE;
      else 
       isnull = UT_FALSE;
      if ( isnull )
      {
       if ( range->incnull == 1 )
       {
	paccept[cpt_offset] = 1;
	eval = 0;
       }
       else
       {
	paccept[cpt_offset] = 0;
	eval = 0;
       }
      }
      else if ( range->incnull == -1 )
      {
       paccept[cpt_offset] = 1;
      }
     }
     break;
    case DT_REAL8:
     xpos = (double)*((double*)ptr);
     df = 0.0;
     if ( range && range->incnull )
     {
      if ( isnan( *((double*)ptr) ))
       isnull = UT_TRUE;
      else if ( nullval && *((double*)ptr) == *((double*)nullval) )
       isnull = UT_TRUE;
      else 
       isnull = UT_FALSE;
      if ( isnull )
      {
       if ( range->incnull == 1 )
       {
	paccept[cpt_offset] = 1;
	eval = 0;
       }
       else
       {
	paccept[cpt_offset] = 0;
	eval = 0;
       }
      }
      else if ( range->incnull == -1 )
      {
       paccept[cpt_offset] = 1;
      }
     }
     break;
    case DT_UINT1:
     xpos = (double)*((unsigned char*)ptr);
     df = 0.1;
     if ( range && range->incnull && nullval )
     {
      if ( *((unsigned char*)ptr) == *((unsigned char*)nullval) )
      {
       if ( range->incnull == 1 )
       {
	paccept[cpt_offset] = 1;
	eval = 0;
       }
       else
       {
	paccept[cpt_offset] = 0;
	eval = 0;
       }
      }
      else if ( range->incnull == -1 )
      {
       paccept[cpt_offset] = 1;
      }
     }
     break;
    case DT_UINT2:
     xpos = (double)*((unsigned short*)ptr);
     df = 0.1;
     if ( range && range->incnull && nullval )
     {
      if ( *((unsigned short*)ptr) == *((unsigned short*)nullval) )
      {
       if ( range->incnull == 1 )
       {
	paccept[cpt_offset] = 1;
	eval = 0;
       }
       else
       {
	paccept[cpt_offset] = 0;
	eval = 0;
       }
      }
      else if ( range->incnull == -1 )
      {
       paccept[cpt_offset] = 1;
      }
     }
     break;
    case DT_UINT4:
     xpos = (double)*((unsigned long*)ptr);
     df = 0.1;
     if ( range && range->incnull && nullval )
     {
      if ( *((unsigned long*)ptr) == *((unsigned long*)nullval) )
      {
       if ( range->incnull == 1 )
       {
	paccept[cpt_offset] = 1;
	eval = 0;
       }
       else
       {
	paccept[cpt_offset] = 0;
	eval = 0;
       }
      }
      else if ( range->incnull == -1 )
      {
       paccept[cpt_offset] = 1;
      }
     }
     break;
    case DT_LOGICAL:
     xpos = (double)*((dmBool*)ptr);
     df = 0.1;
     if ( range && range->incnull && nullval )
     {
      if ( *((dmBool*)ptr) == *((dmBool*)nullval) )
      {
       if ( range->incnull == 1 )
       {
	paccept[cpt_offset] = 1;
	eval = 0;
       }
       else
       {
	paccept[cpt_offset] = 0;
	eval = 0;
       }
      }
      else if ( range->incnull == -1 )
       paccept[cpt_offset] = 1;
     }
     break;
    case DT_CHAR:
     if ( range )
     {
      if ( range->incnull )
      {
       if ( *((char*)ptr) == '\0' )
	isnull = UT_TRUE;
       else if ( nullval && !strcmp(((char*)ptr), ((char*)nullval) ))
	isnull = UT_TRUE;
       else 
	isnull = UT_FALSE;
       if ( isnull )
       {
	if ( range->incnull == 1 )
	{
	 paccept[cpt_offset] = 1;
	 eval = 0;
	}
	else
	{
	 paccept[cpt_offset] = 0;
	 eval = 0;
	}
       }
       else if ( range->incnull == -1 )
	paccept[cpt_offset] = 1;
      }
      if ( eval )
      {
       paccept[cpt_offset] = 0; /* Will fail unless one matches */
       for ( range_no = 0; range_no < range->n; range_no++ )
       {
	if ( dmu_text_in_range( ptr, ((char**)range->mins)[range_no], ((char**)range->maxes)[range_no] ))
	{
	 paccept[cpt_offset] = 1; break;
	}
       }
      }
      eval = 0;
     }
     break;
    case DT_BIT:
     if ( range ) 
     {
      if ( range->incnull && nullval )
      {
       nbytes = dd->length >> 3;
       if ( !memcmp( ptr, nullval, nbytes ))
       {
	if ( range->incnull == 1 )
	{
	 paccept[cpt_offset] = 1;
	 eval = 0;
	}
	else
	{
	 paccept[cpt_offset] = 0;
	 eval = 0;
	}
       }
       else if ( range->incnull == -1 )
	paccept[cpt_offset] = 1;
      }

      if ( eval )
      {
       paccept[cpt_offset] = 0; /* Will fail unless one matches */
       for ( range_no = 0; range_no < range->n; range_no++ )
       {
	if ( dmu_bit_match( ptr, dd->length, ((char**)range->mins)[range_no] ))
	{
	 paccept[cpt_offset] = 1; break;
	}
       }
      }
     }
     eval = 0;
     break;
    default:
     printf( "NO DT support\n" );
    }    
    
    if ( yrange || region )
    {
     yeval = 1;
     switch(  dd->type )
     {
     case DT_INT2:
      ypos = (double)*((short*)ptr+1);
      df = 0.1;
      if ( yrange && yrange->incnull && nullval )
      {
       if ( *((short*)ptr) == *((short*)nullval) )
       {
	if ( yrange->incnull == 1 )
	{
	 paccept[cpt_offset] = 1;
	 yeval = 0;
	}
	else
	{
	 paccept[cpt_offset] = 0;
	 yeval = 0;
	}
       }
       else if ( range->incnull == -1 )
       {
	paccept[cpt_offset] = 1;
       }
      }
      break;
     case DT_INT4:
      ypos = (double)*((long*)ptr+1);
      df = 0.1;
      if ( yrange && yrange->incnull && nullval )
      {
       if ( *((long*)ptr) == *((long*)nullval) )
       {
	if ( yrange->incnull == 1 )
	{
	 paccept[cpt_offset] = 1;
	 yeval = 0;
	}
	else
	{
	 paccept[cpt_offset] = 0;
	 yeval = 0;
	}
       }
       else if ( range->incnull == -1 )
       {
	paccept[cpt_offset] = 1;
       }
      }
      break;
     case DT_REAL4:
      ypos = (double)*((float*)ptr+1);
      df = 0.0;
      if ( yrange && yrange->incnull )
      {
       if ( isnan( *((float*)ptr) ))
	isnull = UT_TRUE;
       else if ( nullval && *((float*)ptr) == *((float*)nullval) )
	isnull = UT_TRUE;
       else 
	isnull = UT_FALSE;
       if ( isnull )
       {
	if ( yrange->incnull == 1 )
	{
	 paccept[cpt_offset] = 1;
	 yeval = 0;
	}
	else
	{
	 paccept[cpt_offset] = 0;
	 yeval = 0;
	}
       }
       else if ( range->incnull == -1 )
       {
	paccept[cpt_offset] = 1;
       }
      }
      break;
     case DT_REAL8:
      ypos = (double)*((double*)ptr+1);
      df = 0.0;
      if ( yrange && yrange->incnull )
      {
       if ( isnan( *((double*)ptr) ))
	isnull = UT_TRUE;
       else if ( nullval && *((double*)ptr) == *((double*)nullval) )
	isnull = UT_TRUE;
       else 
	isnull = UT_FALSE;
       if ( isnull )
       {
	if ( yrange->incnull == 1 )
	{
	 paccept[cpt_offset] = 1;
	 yeval = 0;
	}
	else
	{
	 paccept[cpt_offset] = 0;
	 yeval = 0;
	}
       }
       else if ( range->incnull == -1 )
       {
	paccept[cpt_offset] = 1;
       }
      }
      break;
     case DT_UINT1:
      ypos = (double)*((unsigned char*)ptr+1);
      df = 0.1;
      if ( yrange && yrange->incnull && nullval )
      {
       if ( *((unsigned char*)ptr) == *((unsigned char*)nullval) )
       {
	if ( yrange->incnull == 1 )
	{
	 paccept[cpt_offset] = 1;
	 yeval = 0;
	}
	else
	{
	 paccept[cpt_offset] = 0;
	 yeval = 0;
	}
       }
       else if ( range->incnull == -1 )
       {
	paccept[cpt_offset] = 1;
       }
      }
      break;
     case DT_UINT2:
      ypos = (double)*((unsigned short*)ptr+1);
      df = 0.1;
      if ( yrange && yrange->incnull && nullval )
      {
       if ( *((unsigned short*)ptr) == *((unsigned short*)nullval) )
       {
	if ( yrange->incnull == 1 )
	{
	 paccept[cpt_offset] = 1;
	 yeval = 0;
	}
	else
	{
	 paccept[cpt_offset] = 0;
	 yeval = 0;
	}
       }
       else if ( range->incnull == -1 )
       {
	paccept[cpt_offset] = 1;
       }
      }
      break;
     case DT_UINT4:
      ypos = (double)*((unsigned long*)ptr+1);
      df = 0.1;
      if ( yrange && yrange->incnull && nullval )
      {
       if ( *((unsigned long*)ptr) == *((unsigned long*)nullval) )
       {
	if ( yrange->incnull == 1 )
	{
	 paccept[cpt_offset] = 1;
	 yeval = 0;
	}
	else
	{
	 paccept[cpt_offset] = 0;
	 yeval = 0;
	}
       }
       else if ( range->incnull == -1 )
       {
	paccept[cpt_offset] = 1;
       }
      }
      break;
     case DT_LOGICAL:
      ypos = (double)*((dmBool*)ptr+1);
      df = 0.1;
      if ( yrange && yrange->incnull && nullval )
      {
       if ( *((dmBool*)ptr) == *((dmBool*)nullval) )
       {
	if ( yrange->incnull == 1 )
	{
	 paccept[cpt_offset] = 1;
	 yeval = 0;
	}
	else
	{
	 paccept[cpt_offset] = 0;
	 yeval = 0;
	}
       }
       else if ( range->incnull == -1 )
       {
	paccept[cpt_offset] = 1;
       }
      }
      break;
     case DT_CHAR:
/* This is not yet implemented correctly. Of course, we don't
   currently have any files with string vector columns in.
   You'd need to offset ptr with the string length.
 */
      if ( yrange )
      {
       if ( yrange->incnull )
       {
	if ( *((char*)ptr) == '\0' )
	 isnull = UT_TRUE;
	else if ( nullval && !strcmp(((char*)ptr), ((char*)nullval) ))
	 isnull = UT_TRUE;
	else 
	 isnull = UT_FALSE;
	if ( isnull )
	{
	 if ( yrange->incnull == 1 )
	 {
	  paccept[cpt_offset] = 1;
	  yeval = 0;
	 }
	 else
	 {
	  paccept[cpt_offset] = 0;
	  yeval = 0;
	 }
	}
       }
       else if ( range->incnull == -1 )
       {
	paccept[cpt_offset] = 1;
       }
       if ( yeval )
       {
	paccept[cpt_offset] = 0;
	for ( range_no = 0; range_no < yrange->n; range_no++ )
	{
	 if ( dmu_text_in_range( ptr, ((char**)range->mins)[range_no], ((char**)range->maxes)[range_no] ))
	 {
	  paccept[cpt_offset] = 1; break;
	 }
	}
       }
       yeval = 0;
      }
      break;
     case DT_BIT:
/* This is not yet implemented correctly. Of course, we don't
   currently have any files with bit vector columns in.
   You'd need to offset ptr with the byte length of a field
 */
      if ( yrange )
      {
       if ( yrange->incnull && nullval )
       {
	nbytes = dd->length >> 3;
	if ( !memcmp( ptr, nullval, nbytes ))
	{
	 if ( yrange->incnull == 1 )
	 {
	  paccept[cpt_offset] = 1;
	  yeval = 0;
	 }
	 else
	 {
	  paccept[cpt_offset] = 0;
	  yeval = 0;
	 }
	}
	else if ( yrange->incnull == -1 )
	 paccept[cpt_offset] = 1;
       }
       if ( yeval )
       {
	paccept[cpt_offset] = 0;
	for ( range_no = 0; range_no < yrange->n; range_no++ )
	{
	 if ( dmu_bit_match( ptr, dd->length, ((char**)yrange->mins)[range_no] ))
	 {
	  paccept[cpt_offset] = 1; break;
	 }
	}
       } 
      }
      yeval = 0;
      break;
     default:
      printf( "NO DT support\n" );
     }          
    }
    
    if ( range )
    {
     if ( range->n > 0 )
     {
      /* Must be inside one of the ranges or row fails */
      if ( eval )
      {
       /* For numeric types, eval now; other types already eval'ed */
       paccept[cpt_offset] = 0;	
       for ( range_no = 0; range_no < range->n; range_no++ ) 
       {
	minval = ((double*)range->mins)[range_no] - df;
	maxval =((double*)range->maxes)[range_no] + df;
	if ( xpos == minval || ( xpos > minval && xpos < maxval ))
	{
	 paccept[cpt_offset] = 1;
	 break;
	}
       }
      }
      /* This test to be done even if eval is false */
      if ( !paccept[cpt_offset] ) break;
     }
     else if ( range->incnull && !paccept[cpt_offset] )
      break;
    }
    
    if ( yrange && yrange->n && yeval )
    {
     paccept[cpt_offset] = 0;	
     for ( range_no = 0; range_no < yrange->n; range_no++ )
     {
      minval = ((double*)yrange->mins)[range_no] - df;
      maxval =((double*)yrange->maxes)[range_no] + df;
      if ( ypos == minval || ( ypos > minval && ypos < maxval ))
      {
       paccept[cpt_offset] = 1; break;
      }
     }
/* This one can be inside the if(yeval) test since if !yeval we already caught it above */
     if ( !paccept[cpt_offset] ) break;
    } /* End eval */       

    if ( region )
    {
     paccept[cpt_offset] = regInsideRegion( region, xpos, ypos );
     if ( !paccept[cpt_offset] ) break;
    }
   } /* End col */
  } /* End cpt */
  accept = 0;
 
  for ( cpt_offset = 0; cpt_offset < ncpts; cpt_offset++ )
   if ( paccept[cpt_offset] ) accept = 1;
  
  if ( invert )
   accept = !accept;

  if ( accept ) 
  {
/* 
   What is the raw buffer position for a given virtual row ?
   How many virtual rows are contiguous? (so we can optimize the memcpy)
 */
   if ( last_contiguous < 0 )
    last_contiguous = ngood+vrow_start_offset;
   buffer->index[ngood+vrow_start_offset] = rrow_offset + buffer->soffset;
   buffer->contiguous[last_contiguous]++;
   buffer->contig_start[ngood+vrow_start_offset] = last_contiguous;
   ngood++;
  } else {
   last_contiguous = -1;
  }
 } /* End row */

 return ngood;
}

