/*                                                                
**  Copyright (C) 2002-2004,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmigroup.c                                                  */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"
static dmDescriptor*  dmi_descriptor_get_axis( const dmDescriptor* dd, const long axnum_target );


/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_get_axis                          */
/*                                                             */
/* Description:                                                */
/*    Find the group descriptor which has a given scalar axis  */
/*    number.  Return the descriptor which matches that axis   */
/*    numbers.                                                 */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDescriptor *dd                            */
/*           const long          axnum_target                  */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*     dmDescriptor *cpt_dd                                    */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_get_axis( const dmDescriptor* dd, const long axnum_target )
{
  const dmDescriptor *group_dd = NULL;
  const dmDescriptor *cpt_dd   = NULL;
  long group_offset = 0;
  long cptOffset = 0;
  long axnum = 0;
  long ngroups = 0;
  long dim = 0;

  ngroups = dd->ngroups;

  for (  group_offset = 0; group_offset < ngroups; group_offset++ )
  {
    group_dd = dd->groups[group_offset];
    dim = group_dd->dim;

    for ( cptOffset = 0; cptOffset < dim; cptOffset++ )
    {
      ++axnum;

      if ( axnum == axnum_target )
      {
        /* the group_dd is the axis for a 1-dim group */
        if ( !group_dd->cpts[cptOffset] && dim == 1 )
          return (dmDescriptor*)group_dd;
        cpt_dd = group_dd->cpts[cptOffset];
        return (dmDescriptor*)cpt_dd;
      }
    }
  }
  return NULL;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_get_logical_axis                  */
/*                                                             */
/* Description:                                                */
/*    Find the logical group descriptor which has a given      */
/*    scalar axis number.  Return the descriptor which matches */
/*    that axis numbers.                                       */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDescriptor *dd                            */
/*           const long          axnum_target                  */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*     dmDescriptor *rdd                                       */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_get_logical_axis( const dmDescriptor* dd, const long axnum_target )
{
  dmDescriptor *rdd = NULL;

  rdd = dmi_descriptor_get_axis( dd, axnum_target );

  if ( rdd && rdd->flag & DM_FLAG_PHYS )
    rdd =  dmi_coord_get_parent( rdd );

  return rdd;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_get_physical_axis                 */
/*                                                             */
/* Description:                                                */
/*    Find the physical group descriptor which has a given     */
/*    scalar axis number.                                      */ 
/*    Return the descriptor which matches that axis numbers.   */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDescriptor *dd                            */
/*           const long          axnum_target                  */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*     dmDescriptor *rdd                                       */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_get_physical_axis( const dmDescriptor* dd, const long axnum_target )
{
 dmDescriptor *rdd = NULL;

  rdd = dmi_descriptor_get_axis( dd, axnum_target );

  if ( rdd && rdd->flag & DM_FLAG_LOGICAL )
    rdd =  dmi_descriptor_get_scaled( rdd );
  return rdd;
}

