/*                                                                
**  Copyright (C) 2002-2005,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmikey  .c                                                  */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"

void kl( dmBlock* block );


/***************************************************************/
/*                                                             */
/* Function:   k1                                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock   *block - pointer to block      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void kl( dmBlock* block )
{
  long i;
  dmDescriptor *ddp;

  printf( "----------- NK = %ld KN = %ld\n", block->nkeys, block->keyno );

  for ( i = 0; i < block->nkeys; i++ )
  {
    ddp = block->keys[i];
    printf( "%4ld %-20s %4ld %p\n", i+1, ddp->name, ddp->no, (void*)ddp );
  }
  printf( "-----------\n" );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_key                               */
/*                                                             */
/* Description:                                                */
/*    Note that no = 0 is valid but returns null; it sets the  */
/*    current position to zero                                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmBlock  *block                                   */
/*           long      no                                      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *dd                                  */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_block_get_key( dmBlock *block, const long no )
{
  dmDescriptor *dd = NULL;

  if ( block && block->keys )
  {
    if ( no > 0 && no <= block->nkeys )
    {
      dd = block->keys[no-1];
    }

    if ( dd || no == 0 )
      dmi_block_set_keyno( block, no );
  }
  return dd;
}

/***************************************************************/
/*                                                             */
/* Function:   dmi_key_delete                                  */
/*                                                             */
/* Description:                                                */
/*    Delete key that corresponds to descriptor from block     */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor *dd                                  */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_key_delete( dmDescriptor *dd )
{
  dmBlock      *block = (dmBlock*)dd->block;
  dmDescriptor *ddp   = NULL;

  long keyno = 0;
  long ii = 0;

  if ( block->keys[dd->no-1] != dd )
  {
    printf( "INTERNAL KEY NO ERR KEY %s HAS NO %ld BUT ENTRY IS %s\n", dd->name, dd->no, block->keys[dd->no-1]->name );

    for ( ii = 0; ii < block->nkeys; ii++ )
    {
      ddp = block->keys[ii];
      printf( "%4ld %-20s %4ld\n", ii+1, ddp->name, ddp->no );
    }
  }

  block->keys[dd->no-1] = NULL;
  keyno = dd->no;

  for ( ii = dd->no; ii < block->nkeys; ii++ )
  {
    ddp =  block->keys[ii];
    block->keys[ii-1] = ddp;
    ddp->no = ii;
  }

  block->keys[block->nkeys-1] = NULL;
  block->nkeys--;

  if ( block->keyno >=keyno && block->keyno > 0 ) 
    block->keyno--;

  /* Free descriptor */
  dmi_descriptor_free( dd );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_set_keyno                             */
/*                                                             */
/* Description:                                                */
/*    Set key number in block struct element keyno             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmBlock *block                                    */
/*           long     no                                       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_block_set_keyno( dmBlock *block, long no )
{
  block->keyno = no;
}

/***************************************************************/
/*                                                             */
/* Function:   dmi_key_print                                   */
/*                                                             */
/* Description:                                                */
/*    Print key nformation                                     */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor *dd                                  */
/*                                                             */
/*  Output parameters:                                         */
/*           char         *pbuf                                */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_key_print( dmDescriptor *dd, char *pbuf )
{
  long keyno = dd->no;
  char buf[TEXT_SIZE];
  char tbuf[TEXT_SIZE];
  char typename[WORD_SIZE];

  ar_mix_type_name( dd->type, typename, WORD_SIZE );
  ar_mix_print_generic( dd->aux, dd->type, buf, TEXT_SIZE );  

  if ( dd->unit && *dd->unit )
  { 
    sprintf( tbuf, "%s [%s]", buf, dd->unit );
  }
  else
  {
    strcpy( tbuf, buf );
  }

  sprintf( pbuf, "Key %4ld: %-20s = %-40s (%-6s) / %s\n", keyno, dd->name, tbuf, typename, dd->desc );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_nkeys                             */
/*                                                             */
/* Description:                                                */
/*    Returns the number of keys in the block                  */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmBlock *block                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           long  numberOfKeys                                */
/*                                                             */
/***************************************************************/
long dmi_block_get_nkeys( const dmBlock *block )
{
  if ( !block )
    return 0;

  return block->nkeys;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_key_alloc                                   */
/*                                                             */
/* Description:                                                */
/*     Allocate a key descriptor, or update it if the key      */
/*     already exists.                                         */
/*     Does not handle array keys... need to realloc dd->aux   */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmBlock    *block                                 */
/*     const long        keyno                                 */
/*     const char       *name                                  */
/*     const dmDataType  type                                  */
/*     const void       *value                                 */
/*           long        nvals                                 */
/*     const char       *unit                                  */
/*     const char       *desc                                  */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*          dmDescriptor  *dd                                  */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_key_alloc( dmBlock *block, const long keyno,
                             const char *name, const dmDataType type,
                             const void *value, long nvals,
                             const char *unit, const char *desc )
{
  dmDescriptor *dd  = NULL;
  dmDescriptor *ddp = NULL;

  long dim   = 1;
  long ii    = 0;
  long naxes = 1;
  long no    = 0;
  long nkeys = 100;  /* Initial minimum */


  if ( !block->keys )
  {
    block->keys  = calloc( nkeys, UT_SZ_P );
    block->nkeys = 0; 
    block->nkmax = nkeys;
  }

  // Does this key exist already? Seach list first
  dd = dmi_descriptor_list_search_descriptor( (const dmDescriptor**)block->keys, block->nkeys, name );

  // If key exists, update the key infomration
  if ( dd ) 
  {
    if ( type != dd->type )
    {
      dmi_key_free_aux( dd );
      dd->type = type;
    }

    if ( unit && unit != dd->unit )
    {
      free( dd->unit );
      dd->unit = cs_dup( unit );
    }

    if ( desc && desc != dd->desc )
    {
      free( dd->desc );
      dd->desc = cs_dup( desc );
    }
  }
  else
  {
    // key does not exist, so generate a new one

    no = keyno;

    // a key number of 0 means add onto end of key list
    if ( no == 0 )
      no = block->nkeys + 1;

    // allocate descriptor
    dd = dmi_descriptor_alloc( block, dmKEY, no, name, type, unit, desc, dim, dmVALUE, naxes, &nvals, DM_FLAG_KEY ); 

    // check on block's key storage
    if ( block->nkeys >= block->nkmax )
    {
      block->nkmax = ( block->nkmax  + block->nkeys + 10 ) * 2;
      block->keys  = realloc( block->keys, block->nkmax * UT_SZ_P );
    }

    /* Shift up keys above insertion */ 

    for ( ii = block->nkeys; ii >= no; ii-- )
    {
      ddp = block->keys[ii-1];
      ddp->no++;
      block->keys[ii] = ddp;
    }

    // increment number of keys
    block->nkeys++;

    if ( no > block->nkeys )
    {
      printf( "DM Internal Key Error for key %s NO=%ld N=%ld\n", name, no, block->nkeys );
      no = block->nkeys;
      dd->no = no;
    }

    // add descriptor to block's list
    block->keys[no-1] = dd ;
  }

  /* allocate and set dd->aux to hold the key value */
  dmi_key_alloc_aux( dd, value, nvals );

  return dd;

}


/***************************************************************/
/*                                                             */
/* Function:   dmi_key_open                                    */
/*                                                             */
/* Description:                                                */
/*     Find a key by name in the block                         */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmBlock    *block                                 */
/*     const char       *name                                  */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*          dmDescriptor  *dd                                  */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_key_open( dmBlock* block, const char* name )
{ 
  dmDescriptor *dd = NULL;

  dd = dmi_descriptor_list_search_descriptor( (const dmDescriptor**)block->keys, block->nkeys, name );
  return dd;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_key_read                                    */
/*                                                             */
/* Description:                                                */
/*     Read a key given the name. Return the value, if it      */
/*     isn't void. Set the current key number (used even if    */
/*     value is void).                                         */
/*                                                             */
/*  Input parameters:                                          */
/*           dmBlock    *block                                 */
/*     const char       *name                                  */
/*     const dmDataType  type                                  */
/*                                                             */
/*  Output parameters:                                         */
/*           void       *value                                 */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*          dmDescriptor  *dd                                  */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_key_read( dmBlock *block, const char *name, 
                            const dmDataType type, void *value )
{
  dmDescriptor *dd = NULL;

#if 0
 if ( value )
  ar_mix_type_zero( type, value );
#endif

  if ( !block )
    return NULL;

  dd = dmi_key_open( block, name );

  if ( dd )
    dmi_block_set_keyno( block, dd->no );

  if ( !dd )
    return NULL;

  if ( value )
    ar_mix_cast_array( dd->aux, dd->type, type, value, 1 );

  return dd;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_key_free_aux                                */
/*                                                             */
/* Description:                                                */
/*     Free 'aux' pointer from descriptor                      */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor  *dd                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*          void                                               */
/*                                                             */
/***************************************************************/
void dmi_key_free_aux( dmDescriptor *dd )
{
  long   nvals = 1;
  long   i     = 0;
  char **pptr  = NULL;

  if ( !dd || !dd->aux )
    return;

  if ( dd->type == DT_CHAR )
  {
    /* May be an array key so free all entries */
    if ( dd->axes )
      nvals = dd->axes[0];
    else
      nvals = 1;

    pptr = (char**)dd->aux;

    for ( i= 0; i< nvals; i++ )
      free( pptr[i] );
  }

  free( dd->aux );
  dd->aux = NULL;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_key_alloc_aux                               */
/*                                                             */
/* Description:                                                */
/*     Allocate 'aux; for descriptor and set values            */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           dmDescriptor  *dd                                 */
/*           const void    *value                              */
/*           const long     nvals                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*          void                                               */
/*                                                             */
/***************************************************************/
void dmi_key_alloc_aux( dmDescriptor *dd, const void *value, const long nvals )
{
  char **cptr = NULL;
  long   size = 0;
  long   ii   = 0;

  /* Protect against calling key_write with dd->aux as the value argument */
  /* - could result in freeing the memory before you copy it!             */

  if ( !dd || dd->aux == value )
    return;

  size = ar_mix_type_size( dd->type );

  if ( dd->axes && dd->axes[0] != nvals )
    dd->axes[0] = nvals;

  free( dd->aux );

  if ( dd->type == DT_CHAR )
  {
    dd->length = strlen( *(char**)value );
    size = UT_SZ_P;
    cptr =malloc( size * nvals );

    for ( ii =0; ii< nvals; ii++ )
      cptr[ii] = cs_dup( ((char**)value)[ii] );

    dd->aux = cptr;
  }
  else
  {
    dd->aux = malloc( size * nvals );
    memcpy( dd->aux, value, size* nvals );
  } 
}


