/*                                                                
**  Copyright (C) 2002-2012  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/*****************************************************************/
/*                                                               */
/*   dmiparse.c                                                  */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"
#include "dmiparser.h"


long dmi_parse_op_range_c( int bit, char* op, char* spec, char** specvals, int *pincnull );
long dmi_parse_op_range_n( char* op, char* spec, double* specvals, dmDataType type, int *pincnull );
dmBool dmi_parse_filt_spec( char* spec, double* specvals );
int dmi_parse_filter_item( char** pptr,  long item_no, long dim, char* spec );
void dmi_parse_filt_spec_c( char* spec, char** specvals );

dmBool dmi_strtod( char *pStart, double *dValue);

/***************************************************************/
/*                                                             */
/* Function:   dmi_parse_filter_item                           */
/*                                                             */
/* Description:                                                */
/*    Read the next comma separated token.                     */
/*    Check it's not a new  varname.op.spec                    */
/*    Check it's not a @ (but if region file?)                 */
/* Some syntax depends on whether this is the first item (n=1) */
/*    Read the next comma separated token.                     */
/*                                                             */
/*  Input parameters:                                          */
/*           char  **pptr                                      */
/*           long    item_no                                   */
/*           long    dim                                       */
/*                                                             */
/*  Output parameters:                                         */
/*           char   *spec                                      */
/*                                                             */
/*  Returns:                                                   */
/*            int  retval                                      */
/*                                                             */
/***************************************************************/
int dmi_parse_filter_item( char** pptr,  long item_no, long dim, char* spec )
{
  char  op[40];
  char *spec_ptr = NULL;
  char *item     = NULL;
  char *tmp      = NULL;
  char *nptr     = NULL;
  char *ptr      = *pptr;

  long  maxlen   = 0;
  int   retval   = 0;

  if ( ! ptr || ! (*ptr) || dim < 0 ) 
    return retval;
 
 
  /* Only one @ region filter allowed, must be a new external filter */
  if ( *ptr == '@' && item_no > 1 )
     return retval;
 
  strcpy( spec, " " );

  /* nptr is provisionally the new rest-of-string */

  nptr   = dmu_parse_token( ptr, ',', &item, 0 );

  /* Code Change  Bug 12596 (revisited)  April 15, 2011 */
  /*   dmu_parse_token() can return NULL in item  */
  /*   check pointer before use                   */

  if ( item )
  {
    maxlen = strlen( item );
    tmp    = calloc( maxlen+1, UT_SZ_C );
 
    /* Is it name=val? */
    spec_ptr = dmu_split_token( item, tmp, maxlen, op );

    if ( !spec_ptr )
    {
      /* OK, we like this: it's not name=val */
      *pptr = nptr;
      strcpy( spec, tmp );
      retval = 1;
    }

    free( item );
    free( tmp );
  }
  else
  {
    *pptr = nptr;
  }

  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_parse_op_range_c                            */
/*                                                             */
/* Description:                                                */
/*    Parse the  >4, =4:6, etc.                                */
/*                                                             */
/*  Input parameters:                                          */
/*           int   **bit                                       */
/*           char   *op                                        */
/*           char   *spec                                      */
/*                                                             */
/*  Output parameters:                                         */
/*           char  **specvals                                  */
/*           int    *pincnull                                  */
/*                                                             */
/*  Returns:                                                   */
/*            long n                                           */
/*                                                             */
/***************************************************************/
long dmi_parse_op_range_c( int bit, char* op, char* spec, char** specvals, int *pincnull )
{
  long  n    = 1;
  char *tmp  = NULL;


  specvals[0] = NULL;
  specvals[1] = NULL;
  
  if( !strcmp( op, "<" ) )
  {
    specvals[1] = cs_dup( spec );
  }
  else if ( !strcmp( op, "<=" ) )
  {
    specvals[1] = cs_dup( spec );
  }
  else if ( !strcmp( op, ">" ) )
  {
    specvals[0] = cs_dup( spec );
  }
  else if ( !strcmp( op, ">=" ) )
  {
    specvals[0] = cs_dup( spec );
  }
  else
    if ( !strcmp( op, "!=" ) )
    {
      if ( cs_uceq( spec, "NULL" ))
      {
        *pincnull = -1;
        n = 0;
      }
      else
        if ( bit ) 
        {
          tmp = malloc( strlen(spec) + 2 );
          sprintf( tmp, "!%s", spec );
          specvals[0] = tmp;
          n =  1;
        }
        else
        {
          specvals[0] = dmu_string_delta( spec, 1 );
          specvals[1] = dmu_string_delta( spec, -1 );
          n = -1;
        }
    } 
    else
      if ( !strcmp( op, "=" ) ) 
      {
        if ( cs_uceq_siz( spec, "NULL", 4 )
	     && ( cs_uceq( spec, "NULL" ) || cs_uceq( spec, "NULL:NULL" )))
	{
	  *pincnull = 1;
	  n = 0;
	}
        else
	  dmi_parse_filt_spec_c( spec, specvals );
      }
      else
      {
        n = dmFBAD_PARSE;
      }

  return n;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_parse_op_range_n                            */
/*                                                             */
/* Description:                                                */
/*    Parse the  >4, =4:6, etc.                                */
/*                                                             */
/*  Input parameters:                                          */
/*           char        *op                                   */
/*           char        *spec                                 */
/*           dmDataType   type                                 */
/*                                                             */
/*  Output parameters:                                         */
/*           char       **specvals                             */
/*           int         *pincnull                             */
/*                                                             */
/*  Returns:                                                   */
/*            long n                                           */
/*                                                             */
/***************************************************************/
long dmi_parse_op_range_n( char* op, char* spec, double* specvals, dmDataType type, int *pincnull )
{
  dmBool bIsValid = dmTRUE;

  double dval = 0.0;
  long   n    = 1;

  specvals[0] = -DT_REAL8_MAX;
  specvals[1] =  DT_REAL8_MAX;

  // Code Change Bug 13286 June 28, 2012
  // Insure numeric string before calling atof
  //  otherwise, range may become open
  // Check these cases here, do = and != in their own code sections

  if ( (!strcmp( op, "<" )) || (!strcmp( op, "<=" )) ||
       (!strcmp( op, ">")) || (!strcmp(op, ">=")) )
  {
    bIsValid = dmi_strtod(spec, &dval);

    if ( !bIsValid )
    {
      /* UNABLE TO PARSE */
      n = dmFBAD_PARSE;
      return n;
    }
  }

    
  /*  LESS THAN */  
  if ( !strcmp( op, "<" ) )
  {
    if ( type == DT_REAL8 )
    {
      if ( dval != 0.0 )
	specvals[1] = dval - 2 * DBL_EPSILON * fabs( dval );
      else
	specvals[1] = dval - 2 * DBL_MIN;
    }
    else
      if ( type == DT_REAL4 )
      {
        if ( dval != 0.0 )
	  specvals[1] = dval - FLT_EPSILON * fabs( dval );
        else
	  specvals[1] = dval - FLT_MIN;
      }
      else
        specvals[1] = dval - 1.0;
    }
    else
      if ( !strcmp( op, "<=" ) )
      {
        /* LESS THAN OR EQUAL TO */

        if ( type == DT_REAL8 )
        {
          if ( dval != 0.0 )
	    specvals[1] = dval + 2 * DBL_EPSILON * fabs( dval );
          else
	    specvals[1] = dval + 2 * DBL_MIN;
        }
        else
          if ( type == DT_REAL4 )
          {
            if ( dval != 0.0 )
	      specvals[1] = dval + FLT_EPSILON * fabs( dval );
            else
	      specvals[1] = dval + FLT_MIN;
          }
          else
            specvals[1] = dval;
      }
      else
        if ( !strcmp( op, ">" ) )
        {
          /* GREATER THAN */
 
         if ( type == DT_REAL8 )
          {
            if ( dval != 0.0 )
	      specvals[0] = dval + 2 * DBL_EPSILON * fabs( dval );
            else
	      specvals[0] = dval + 2 * DBL_MIN;
          }
          else
            if ( type == DT_REAL4 )
            {
              if ( dval != 0.0 )
	        specvals[0] = dval + FLT_EPSILON * fabs( dval );
              else
	        specvals[0] = dval + FLT_MIN;
            }
            else
              specvals[0] = dval + 1.0;
        }
        else
        if ( !strcmp( op, ">=" ) )
        {
          /* GREATER THAN OR EQUAL TO */

          if ( type == DT_REAL8 )
          {
            if ( dval != 0.0 )
	      specvals[0] = dval - 2 * DBL_EPSILON * fabs( dval );
            else
	      specvals[0] = dval - 2 * DBL_MIN;
          }
          else
            if ( type == DT_REAL4 )
            {
              if ( dval != 0.0 )
	        specvals[0] = dval - FLT_EPSILON * fabs( dval );
              else
	        specvals[0] = dval - FLT_MIN;
            }
            else
              specvals[0] = dval;
        }
        else
          if ( !strcmp( op, "!=" ) )
          {
            /* NOT EQUAL TO */
            if ( *spec == '(' )
            {
              n = -dmVECPARSE;
            }
            else
            {
              if ( cs_uceq( spec, "NULL" ))
              {
	        *pincnull = -1;
	         n = 0;
              }
              else
              {
                // Code Change Bug 13286 June 28, 2012
                // Insure numeric string

                bIsValid  = dmi_strtod( spec, &dval);

                if ( !bIsValid )
                {
                  /* UNABLE TO PARSE */
                  n = dmFBAD_PARSE;
                  return n;
                }

	        specvals[0] = dval;
	
	        if ( type == DT_REAL8 )
                {
	          if ( dval != 0.0 )
                  {
	            specvals[1] = dval + 2 * fabs( dval ) * DBL_EPSILON;
	          }
                  else
                  {
	            specvals[1] = dval + 2 * DBL_MIN;
	          }
	        }
                else
                  if ( type == DT_REAL4 )
                  {
	            if ( dval != 0.0 )
                    {
	              specvals[1] = dval + fabs( dval ) * FLT_EPSILON;
	            }
                    else
                    {
	              specvals[1] = dval + FLT_MIN;
	            }
	          }
                  else
                  {
	            specvals[0] = dval - 1.0;
	            specvals[1] = dval + 1.0;
	          }
	        n = -1;
              }
            }
          }
          else
           if ( !strcmp( op, "=" ) )
           {
             /*EQUAL TO */
             if ( cs_uceq_siz( spec, "NULL", 4 ) &&
                  ( cs_uceq( spec, "NULL" ) ||
                    cs_uceq( spec, "NULL:NULL" )))
             {
               *pincnull = 1;
               n = 0;
             }
             else
               if ( type == DT_LOGICAL )
               {
                 n = 1;
                 if ( cs_index_char( "TtYy1", *spec ))
                 {
	           specvals[0] = UT_TRUE;
	           specvals[1] = UT_TRUE;
                 }
                 else
                   if ( cs_index_char( "FfNn0", *spec ))
                   {
	             specvals[0] = UT_FALSE;
	             specvals[1] = UT_FALSE;
                   }
                   else
                   {
	             specvals[0] = UT_FALSE;
	             specvals[1] = UT_FALSE;
	             sprintf( dm_errbuf, "Value %s is not a valid logical (e.g. T, F)", spec );
	             dmu_err_set( DM_ERR_PARSE, dm_errbuf );
	             n = dmFBAD_PARSE;
                   }
               }
	       else
                 if (strchr( spec, '(' ) )
                 {
                   /* We have a region */
                   if ( *spec == '(')
                   {
                     /* We have a vector filter */
	             n = dmVECPARSE;
                   }
                   else
	            n = 0;
                 }
                 else
                 {
                   if (dmi_parse_filt_spec( spec, specvals ) != dmSUCCESS )
		   {
                     n = dmFBAD_PARSE;
                     return n;
                   }

                   if ( ( type == DT_REAL8 ) && ( fabs(specvals[0]-specvals[1]) < DBL_EPSILON ) )
                   {
	             if ( specvals[1] != 0 ) 
	               specvals[1] = specvals[1] + 2 * fabs( specvals[1] ) * DBL_EPSILON;
	             else 
	               specvals[1] = specvals[1] + 2 * DBL_MIN;
                   }
                   else
                     if ( (type == DT_REAL4 ) && ( fabs(specvals[0]-specvals[1]) < FLT_EPSILON ) )
                     {
	               if ( specvals[1] != 0 ) 
	                 specvals[1] = specvals[1] + fabs( specvals[1] ) * FLT_EPSILON;
	               else 
	                 specvals[1] = specvals[1] + FLT_MIN;
                     }
                 }
           }
           else
           {
             /* UNABLE TO PARSE */
             n = dmFBAD_PARSE;
           }

  return n;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_parse_rangelist                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*        dmDescriptor   *vector                               */
/*        long            vcpt                                 */
/*        char          **pptrvals                             */
/*        dmFilter       *filter                               */
/*        char           *op                                   */
/*        int             loop                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            int  retval                                      */
/*                                                             */
/***************************************************************/
int dmi_parse_rangelist( dmDescriptor* vector, long vcpt, char** pptr,
                         dmFilter* filter, char* op, int loop )
{
  dmDescriptor *dd     = NULL;
  dmDescriptor *col_dd = NULL;
  dmRangeList  *range  = NULL;
  dmDataType    type;

  char         *name   = NULL;

  double *value1 = NULL;
  double *value2 = NULL;

  long maxcpt  = filter->dim;
  long mincpt  = 1;
  long no      = 0;
  long nvalues = 0;
  int  incnull = 0;
  int  retval  = 0;

  
  dd   = dmi_descriptor_get_cpt( vector, vcpt );
  name = dd->name;
  type = dd->type;

  if ( vcpt > 0 )
  {
    mincpt = vcpt;
    maxcpt = vcpt;
  }

  retval = dmi_filter_parse_typed_rangelist( name, type, mincpt, maxcpt, pptr, filter, op, loop, &incnull );
  
  /* added to tie in the range to datatype boundaries*/
  dmi_filter_extract( filter, 1, type, (void*)&value1, (void*)&value2, &nvalues, &incnull );
  dmi_rangelist_set( filter->ranges, type, value1, value2, nvalues );
  
  /* if the filter calls for a null test, need to make sure there is  */ 
  /* a null value defined for the column, if not, get rid of the test */

  col_dd = vector->parent;

  if ( incnull && !dmi_descriptor_has_nullval( col_dd ))
  {
    for ( no = mincpt; no <= maxcpt; no++ )
    {
      range = &filter->ranges[no-1];

      if ( range->incnull == 1 )
      {
        if ( range->n == 0 )
        {
          filter->null = UT_TRUE;
	   range->incnull = 0;
	}
      }
      else
        if ( range->incnull == -1 )
	  range->incnull = 0;
    }
  }

  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_parse_typed_rangelist                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*        char         *name                                   */
/*        dmDataType    type                                   */
/*        long          mincpt                                 */
/*        long          maxcpt                                 */
/*        char        **pptr                                   */
/*        dmFilter      filter                                 */
/*        char         *op                                     */
/*        int           loop                                   */
/*                                                             */
/*  Output parameters:                                         */
/*        int          *pincnull                               */
/*                                                             */
/*  Returns:                                                   */
/*            int intersect                                    */
/*                                                             */
/***************************************************************/
int dmi_filter_parse_typed_rangelist( char* name, dmDataType type,
                        long mincpt, long maxcpt, char** pptr,
                       dmFilter* filter, char* op, int loop, int *pincnull )
{
  char  sptmp[MAX_BUF_LENGTH];  /* Buffer for special vector col notation support */
  char  errname[]    = "UNKNOWN";
  char *cspecvals[2];
  char *spec         =  NULL;
  char *start        = *pptr;
  char *ptr          = *pptr;

  double specvals[2];

  long i       = 0;
  long n       = 0;
  long dim     = 0;
  long item_no = 1;

  int text      = 0;
  int exclude   = 0;
  int bit       = 0;
  int intersect = 0;
  int ok        = 1;
  int incnull   = 0;
  
  spec = malloc( strlen(ptr) + MAX_BUF_LENGTH );  /* Generous space */

  if ( !name )
    name = errname;
  
  dim = dmi_filter_get_dim( filter );

  if ( maxcpt == 0 ) 
    maxcpt = dim;
  
  while ( ok && dmi_parse_filter_item( &ptr, item_no, dim, spec ))
  {
    if ( !loop )
      ok = 0;  /* Terminate loop if an image: multiple ranges disallowed */

    ++item_no;

    intersect = 1;
    incnull   = 0;

    if ( !strcmp( op, "=" ) )
      intersect = 0;
      
    if ( type == DT_CHAR || type == DT_BIT )
    {
      text = 1;
      bit = ( type == DT_BIT );
      n = dmi_parse_op_range_c( bit, op, spec, cspecvals, &incnull );
    }
    else     
      n = dmi_parse_op_range_n( op, spec, specvals, type, &incnull );


    if ( n == 0 )
    {
      if ( incnull )
      {
        *pincnull = incnull;

        for ( i = mincpt; i <= maxcpt; i++ )
          dmi_filter_set_incnull( filter, i, incnull );
      }
      else
      {
        dmi_filter_set_region_string( filter, spec );

        if ( dim < 2 )
	{
	  sprintf( dm_errbuf, "DM syntax error: 1-dim variable %s with 2-dim region %s\n",
		   name, spec );
	  dmu_err_set( DM_ERR_PARSE, dm_errbuf );
	  ok = 0;
	}
      }
    }
    else
    {
      if ( n == dmFBAD_PARSE )
      {
	sprintf( dm_errbuf, "Syntax error (no operator in filter): %s", start );
	  
        dmu_err_set( DM_ERR_PARSE, dm_errbuf );
        ok = 0;
      }
      else
      {
	exclude = ( n < 0 );

	if ( exclude && item_no > 2 )
        {
          sprintf( dm_errbuf, "DM syntax error (%s%s%s) \nRemember != takes only one value (use explicit ranges or [exclude %s=...] instead?) ",
          name, op, start, name );

          dmu_err_set( DM_ERR_PARSE, dm_errbuf );
          ok = 0;
        }
        else
	{
          if ( text )
          {
            for ( i = mincpt; i <= maxcpt; i++ )
              dmi_filter_add_range( filter, i, cspecvals, exclude );

            for ( i = 0; i < 2; i++ )
              free( cspecvals[i] );
	  }
	  else
	  {
            if ( n == dmVECPARSE || n == -dmVECPARSE )
            {
              /* Support the notation   (chipx,chipy)=(246:500,960:980)   */
              /* Allow one range per axis only */

              char *sptr    = spec + 1;
              char *sbufptr = sptmp;
              char *eptr    = NULL;
	      
              for ( i = mincpt; i <= maxcpt; i++ )
              {
                sbufptr = sptmp;

                if ( sptr )
                {
                  sptr = dmu_parse_token( sptr, ',', &sbufptr, MAX_BUF_LENGTH ); 
    
                  // Code Change Bug 12596 (revisited)  April 15, 2011
                  //  Function dmu_parse_token() can return NULL in result buffer
                  //  Check before using

                  if ( *sbufptr == '\0' )
		  {
                    // Most likely result of bad input string format, e.g. two commas in a row
                    // punt
                    continue;
                  }

		  if ( (eptr = strchr( sbufptr, ')' )) != NULL )
                     *eptr = '\0';

                   if ( cs_uceq( sbufptr, "NULL" ))
                   {
                     dmi_filter_include_null( filter, i );
                     *pincnull = 1;
                   }

                   dmi_parse_filt_spec( sbufptr, specvals );
                   dmi_filter_add_range( filter, i, specvals, exclude );
                }
              }
            }
            else
            {
              /* The standard case      pha = 40:60,80:100,...  */
              for ( i = mincpt; i <= maxcpt; i++ )
                dmi_filter_add_range( filter, i, specvals, exclude );
            }
          }
        }
      }
    }
  } // end while loop

  *pptr = ptr;
  free( spec );

  return intersect;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_parse_filt_spec                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*        char    *spec                                        */
/*                                                             */
/*  Output parameters:                                         */
/*        double  *specvals                                    */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
dmBool dmi_parse_filt_spec( char* spec, double* specvals)
{  
  dmBool errRet = dmSUCCESS; 
  dmBool bIsValid = dmTRUE;
  char  *item[3] = { NULL, NULL, NULL };
  char  *ptr     = spec;

  double val     = 0.0;
 
  long   nitems  = 0;
  long   i       = 0; 


  /* Cut into colon substrings */
  if ( !spec || *spec == '\0' )
    return errRet;

  /* Special case: one value only is treated as last item */
  if ( !strchr( ptr, ':' ) )
  {
    // Code Change Bug 13286 June 28, 2012
    // Insure numeric string

    bIsValid = dmi_strtod(spec, &val);
   
    if ( !bIsValid )
      return dmFAILURE;

    specvals[0] = val;
    specvals[1] = val;
    return dmSUCCESS;
  }
  
  while( ptr && *ptr && nitems < 3 )
    ptr = dmu_parse_token( ptr, ':', &item[nitems++], 0 );
  
  if ( nitems > 2 )
  {
    sprintf( dm_errbuf, "DM parse error: too many ranges in filter %s", spec );
    dmu_err_set( DM_ERR_PARSE, dm_errbuf );
    nitems = 2;
  } 
  
  for( i = 0; i < nitems; i++ )
  {
    if ( item[i] && *(item[i]) )
    {
      // Code Change Bug 13286
      // Insure numeric string
 
      bIsValid = dmi_strtod(item[i], &specvals[i] );

       if ( !bIsValid )
         errRet = dmFAILURE;

       free( item[i] );
     }
  }

  return errRet;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_parse_filt_spec_c                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*        char    *spec                                        */
/*                                                             */
/*  Output parameters:                                         */
/*        char   **specvals                                    */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_parse_filt_spec_c( char* spec, char** specvals )
{ 
  char *item[3] = { NULL, NULL, NULL };
  char *ptr     = spec;
  long  nitems  = 0;
  long  i       = 0;

  /* Cut into colon substrings */
  specvals[0] = NULL;
  specvals[1] = NULL;

  if ( !spec || *spec == '\0' )
    return;

  /* Special case: one value only is treated as last item */
  if ( !strchr( ptr, ':' ) )
  {
    specvals[0] = cs_dup( ptr );
    specvals[1] = cs_dup( ptr );
    return;
  }
  
  while( ptr && *ptr && nitems < 3 )
    ptr = dmu_parse_token( ptr, ':', &item[nitems++], 0 );
  
  if ( nitems > 2 )
  {
    sprintf( dm_errbuf, "DM parse error: too many ranges in string filter %s", spec );
    dmu_err_set( DM_ERR_PARSE, dm_errbuf );
    nitems = 2;
  } 
  
  for( i = 0; i < nitems; i++ )
  {
    if ( item[i] && *(item[i]) )
      specvals[i] = item[i];
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_strtod                                      */
/*                                                             */
/* Description:                                                */
/*       Convert string operand into number. Use strtod() in   */
/*       order to trap invalid number string.                  */
/*                                                             */
/*       Created for Bug 13286                                 */
/*                                                             */
/*  Input parameters:                                          */
/*        char    *pSpec                                       */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*        double  *dValue                                      */
/*                                                             */
/*  Returns:                                                   */
/*            dmTRUE or dmFALSE                                */
/*                                                             */
/***************************************************************/
dmBool dmi_strtod(char *pSpec, double *dValue)
{
  dmBool  bIsValid      = dmTRUE;
  char  *pConvertedSpec = NULL;

  *dValue = strtod(pSpec, &pConvertedSpec);

  if ( strlen(pConvertedSpec) != 0 )
    bIsValid = dmFALSE;

  return bIsValid;
}
