/*                                                                
**  Copyright (C) 2002-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/* *************************************************************** */
/* 								   */
/*    dmisearch.c                                                  */
/* 								   */
/* 								   */
/* 								   */
/* 								   */
/*  This library of functions is used to search through blocks,    */
/*  descriptor lists, and descriptors for named descriptors.       */
/* 								   */
/*  These are the public searching functions:                      */
/* 								   */
/*  dmi_descriptor_list_search_descriptor()                        */
/*    Search a list of descriptors for a named descriptor.         */
/*    Just search the descriptors themselves, not their            */
/*    vector components, coords, or groups.                        */
/*  dmi_descriptor_list_search()                                   */
/*    Search a list of descriptors for a named descriptor.         */
/*    Will search the descriptors', their vector components,       */
/*    their coords, and their groups.                              */
/*  dmi_descriptor_search()                                        */
/*    Search a descriptor, its vector components, its coords,      */
/*    and its groups for a named descriptor.                       */
/*  dmi_block_search_col_identifier()                              */
/*    Searches a block's cols then rawcols for a named descriptor. */
/*    Understands '#' and (a,b) vector syntax.                     */
/*    Will search the block's descriptors, their vector components */
/*    their coords, and their groups.                              */
/*    Used by bin parsing to find binned descriptors.              */
/*  dmi_block_search_raw_col_identifier()                          */
/*    Same as dmi_block_search_col_identifier except only searches */
/*    rawcols.  Used by filter parsing since filters are attached  */
/*    to raw columns.                                              */
/*  dmi_descriptor_list_search_vector()                            */
/*    Search a descriptor list for a named descriptor.             */
/*    Search the descriptors and their vector components, but      */
/*    not their coords or groups.                                  */
/* 								   */
/* *************************************************************** */
#include "dmi.h"


static dmDescriptor* dmi_descriptor_search_vector( const dmDescriptor* dd, const char* name );
static dmDescriptor* dmi_descriptor_search_children( const dmDescriptor* dd, const char* name );
static dmDescriptor* dmi_block_search_hash_identifier( const dmBlock* block, const dmDescriptor** dds, const long n, const char* name );
static dmDescriptor* dmi_block_accumulate_hash_identifiers( const dmBlock* block, const dmDescriptor** dds, const long n, const char** names, dmDescriptor** cpt_dds, const long target, long* pmatch );
static dmDescriptor* dmi_descriptor_accumulate_vector( const dmDescriptor* dd, const char** names, dmDescriptor** cpt_dds, const long target, long* pmatch );
static dmDescriptor* dmi_descriptor_accumulate_children( const dmDescriptor* dd, const char** names, dmDescriptor** cpt_dds, const long target, long* pmatch );
static dmDescriptor* dmi_descriptor_list_accumulate( const dmDescriptor** dds, const long n, const char** names, dmDescriptor** cpt_dds, const long target, long* pmatch );

static dmDescriptor* dmi_descriptor_accumulate( const dmDescriptor* dd, const char** name, dmDescriptor** cpt_dds, const long target, long* pmatch );
static dmDescriptor* dmi_descriptor_get_group_of_axis( const dmDescriptor* image_dd, const dmDescriptor* axis_dd );


/***************************************************************/
/*                                                             */
/*  Function:  dmi_descriptor_search                           */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/* search a vector, its vector components, its coords, and     */
/* its groups for a named descriptor.  Return the descriptor   */
/* if found and NULL otherwise.                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*          const dmDescriptor   *dd                           */
/*          const char           *name                         */
/*                                                             */
/*  Output parameters:                                         */
/*            none                                             */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *rdd  - returned descriptor pointer  */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_search( const dmDescriptor* dd, const char* name )
{
  dmDescriptor *rdd = NULL;  /* result descriptor */

  if ( !dd )
     return NULL;

  rdd = dmi_descriptor_search_vector( dd, name );

  if ( !rdd )
    rdd = dmi_descriptor_search_children( dd, name );

  return rdd;
}


/***************************************************************/
/*                                                             */
/* Function: dmi_dsscriptor_search_vector                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/* Search a vector itself and its components for a named       */
/* descriptor.                                                 */
/* Return the descriptor itself if found and NULL otherwise.   */
/*                                                             */
/*  Input parameters:                                          */
/*            const dmDescriptor   *dd                         */
/*            const char           *name                       */
/*                                                             */
/*  Output parameters:                                         */
/*                none                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor *cpt_dd                             */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_search_vector( const dmDescriptor* dd, const char* name )
{
  dmDescriptor  *cpt_dd = NULL;
  dmDescriptor **cpts   = NULL;
  long dim = 0;
  long cptOffset = 0;


  if ( cs_uceq( dd->name, name ))
    return (dmDescriptor*)dd;
 
  dim = dd->dim;

  if ( dim > 1 )
  {
    cpts = dd->cpts;

    for ( cptOffset = 0; cptOffset < dim; cptOffset++ )
    {
      cpt_dd = cpts[cptOffset];
      if ( cpt_dd && cs_uceq( cpt_dd->name, name ))
        return (dmDescriptor*)cpt_dd;
    }
  }

  return NULL;
}

/***************************************************************/
/*                                                             */
/*  Function:  dmi_descriptor_search_children                  */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/* Search through a descriptor's coordinates and groups for a  */
/* named descriptor.  Return the descriptor if found and NULL  */
/* otherwise.                                                  */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDescriptor  *dd                           */
/*           const char          *name                         */
/*                                                             */
/*  Output parameters:                                         */
/*                   none                                      */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *rdd                                 */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_search_children( const dmDescriptor* dd, const char* name )
{
  dmDescriptor  *rdd = NULL; /* returned descriptor */
  dmDescriptor **dds = NULL; /* list of descriptors */
  long n = 0;                /* number of coords or groups */
  long offset = 0;           /* offset into coords or groups */

  /* search coords */
  n = dd->ncoords;

  if ( n > 0 )
  {
    dds = dd->coords;

    for ( offset = 0; offset < n; offset++ )
    {
      rdd = dmi_descriptor_search( dds[offset], name );
      if ( rdd )
        return rdd;
    }
  }

  /* search groups */
  n = dd->ngroups;

  if ( n > 0 )
  {
    dds = dd->groups;
    for ( offset = 0; offset < n; offset++ )
    {
      rdd = dmi_descriptor_search( dds[offset], name );
      if ( rdd )
        return rdd;
    }  
  }
  return NULL;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_list_search_descriptor            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/* Search a descriptor list for a named descriptor.  Do not    */
/* search the descriptors' vector components, coords, or       */
/* groups.                                                     */
/* Return the descriptor if found and NULL otherwise.          */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDescriptor  **dds                         */
/*                 long            n                           */
/*           const char           *name                        */
/*                                                             */
/*  Output parameters:                                         */
/*                 none                                        */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor *dd                                 */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_list_search_descriptor( const dmDescriptor** dds, long n, const char* name )
{
  long offset = 0;
  const dmDescriptor *dd = NULL;
 
  for ( offset = 0; offset < n; offset++ )
  {
    dd = dds[offset];
    if ( cs_uceq( dd->name, name ))
      return (dmDescriptor*)dd;
   }
  return NULL;
}

/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_list_search_vector                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/* Search a descriptor list for a named descriptor.            */
/* Search the descriptors' themselves and their vector         */
/* components but not their coords or groups.                  */
/* Special Case Note:                                          */
/*   If name == NULL and n == 1, returns the first             */
/*   and only descriptor in the list.  This is for image       */
/*   searches which should always return the image descriptor. */
/* Return the descriptor if found and NULL otherwise.          */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDescriptor  **dds                         */
/*           const long            n                           */
/*           const char           *name                        */
/*                                                             */
/*  Output parameters:                                         */
/*                none                                         */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *dd                                  */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_list_search_vector( const dmDescriptor** dds, const long n, const char* name )
{
  long offset = 0;
  dmDescriptor *dd = NULL;

  /* Special case: null string */
  if ( !name )
  {
    if ( n == 1 )  /* For IMAGE case, null name gives the image descriptor */
      return (dmDescriptor*)dds[0];
    return NULL;
  }

  for ( offset = 0; offset < n; offset++ )
  {
    dd = dmi_descriptor_search_vector( dds[offset], name );
    if ( dd )
      return dd;
  }
  return NULL;
}

/***************************************************************/
/*                                                             */
/*  Function:  dmi_descriptor_list_search                      */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/* Search a descriptor list for a named descriptor.  Search    */
/* of the the descriptors' themselves, their vector components,*/
/* their coords, and their groups.                             */
/* Return the descriptor if found and NULL otherwise.          */
/*                                                             */
/*  Input parameters:                                          */
/*            const dmDescriptor  **dds                        */
/*            const long            n                          */
/*            const char           *name                       */
/*                                                             */
/*  Output parameters:                                         */
/*                 none                                        */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor *dd                                 */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_list_search( const dmDescriptor** dds, const long n, const char* name )
{
  dmDescriptor *dd = NULL; /* descriptor to return */
  long offset = 0;

  for ( offset = 0; offset < n; offset++ )
  {
    dd = dmi_descriptor_search( dds[offset], name );
    if ( dd )
      return dd;
  }
  return NULL;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_block_search_identifier                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/* Search a block for a named descriptor.                      */
/* Search the block's public then raw cols then its keys.      */
/* Search the cols' themselves, their vector components, their */
/* coords, and their groups.                                   */
/* Return the descriptor if found and NULL otherwise.          */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmBlock  *block                             */
/*           const char     *name                              */
/*                                                             */
/*  Output parameters:                                         */
/*                 none                                        */
/*                                                             */
/*  Returns:                                                   */
/*             dmDescriptor *rdd                               */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_block_search_identifier( const dmBlock* block, const char* name )
{
  dmDescriptor *rdd = NULL; /* descriptor to return */

  rdd = dmi_descriptor_list_search( (const dmDescriptor**)block->cols, block->ncols, name );

  if ( !rdd )
  {
    rdd = dmi_descriptor_list_search( (const dmDescriptor**)block->rawcols, block->nrcols, name );
    if ( !rdd )
      rdd = dmi_descriptor_list_search( (const dmDescriptor**)block->keys, block->nkeys, name );
  }
  return rdd;
}

/***************************************************************/
/*                                                             */
/*  Function:   dmi_block_search_public_identifier             */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/* Search a block for a named descriptor.                      */
/* Search the block's public cols then its keys for the desc.  */
/* Search the cols' themselves, their vector components,       */
/* their coords, and groups.                                   */
/* Return the descriptor if found and NULL otherwise.          */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmBlock  *block                             */
/*           const char     *name                              */
/*                                                             */
/*  Output parameters:                                         */
/*               none                                          */
/*                                                             */
/*  Returns:                                                   */
/*             dmDescriptor *dd                                */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_block_search_public_identifier( const dmBlock* block, const char* name )
{
  dmDescriptor *dd = NULL; /* descriptor to return */
  dd = dmi_descriptor_list_search( (const dmDescriptor**)block->cols, block->ncols, name );

  if ( !dd )
    dd = dmi_descriptor_list_search( (const dmDescriptor**)block->keys, block->nkeys, name );

  return dd;
}

/***************************************************************/
/*                                                             */
/*  Function:                                                  */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/* Search a block for a named descriptor.                      */
/* Search the block's raw cols for the descriptor.             */
/* Search the cols' themselves, their vector components,       */
/* their coords, and groups.                                   */
/* Handles all '#' and (a,b) syntax.                           */
/* Will build a vector if needed to satisfy (a,b) syntax.      */
/* Return the descriptor if found and NULL otherwise.          */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmBlock  *block                             */
/*           const char     *name                              */
/*                                                             */
/*  Output parameters:                                         */
/*                none                                         */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *dd                                  */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_block_search_raw_col_identifier( dmBlock* block, const char* name )
{
  dmDescriptor   *dd         = NULL;
  dmDescriptor  **cpt_dds    = NULL;
  dmDescriptor   *image_dd   = NULL;
  dmDescriptor  **new_groups = NULL;
  dmBool          is_logical = UT_FALSE;
  dmBool          make_group = UT_TRUE;
  char          **tcptNames  = NULL;
  char          **cptNames   = NULL;
  char            c          = *name;

  long ncpts     = 0;
  long cptOffset = 0;
  long match     = 0;
  long groupno   = 0;
  long new_ng    = 0;
  long ng        = 0;


  /* First, try '#' syntax */
  if ( c == '#' )
  {
    dd = dmi_block_search_hash_identifier( block, (const dmDescriptor**)block->rawcols, block->nrcols, name );
    if ( dd ) 
      return dd;
  }

  /* Second, try a straight comparison of all vectors, coords, and groups */
  dd = dmi_descriptor_list_search( (const dmDescriptor**)block->rawcols, block->nrcols, name );

  if ( dd )
    return dd;

  /* Third, check vector '(' syntax */
  if ( c == '(' )
  {
    tcptNames = cs_parse_paren_list( name+1, &ncpts );

    if ( ncpts == 0 || ncpts > 2 )
    {
      /* this code can handle ncpts > 2, but later code can't. */
      /* might as well not go on... */
      ar_free_cp( tcptNames, ncpts+1 );
      return NULL;
    }

    cptNames = calloc( ncpts, UT_SZ_P );

    for ( cptOffset = 0; cptOffset < ncpts; cptOffset++ )
       cptNames[cptOffset] = tcptNames[cptOffset+1];

    free( tcptNames );
    cpt_dds = calloc( ncpts, UT_SZ_P );
    match = 0;
    dd = dmi_block_accumulate_hash_identifiers( block, (const dmDescriptor**)block->rawcols, block->nrcols, (const char**)cptNames, cpt_dds, ncpts, &match );

    if ( dd )
    {
      ar_free_cp( cptNames, ncpts );
      free( cpt_dds );
      return dd;
    }

    /* now search for vector components the normal way */
   if ( match != ncpts )
   {
     dd = dmi_descriptor_list_accumulate( (const dmDescriptor**)block->rawcols, block->nrcols, (const char**)cptNames, cpt_dds, ncpts, &match );

     if ( dd )
     {
       ar_free_cp( cptNames, ncpts );
       free( cpt_dds );
       return dd;
     }
   }

   ar_free_cp( cptNames, ncpts );
  
   if ( match == ncpts )
   { 
     is_logical = cpt_dds[0]->flag & DM_FLAG_LOGICAL;
     /* If these are image axes, we must make a new group */
     groupno = cpt_dds[0]->no;

     for ( cptOffset = 0; cptOffset < ncpts; cptOffset++ )
     {
       if ( cpt_dds[cptOffset]->flag & DM_FLAG_AXIS )
       {
         /* can't mix image axes with normal columns                  */
         /* can't regroup an already grouped axis                     */
         /* can't group logical with physical                         */
         /* the new grouping needs to be attached to consecutive axes */
         if ( ! make_group || cpt_dds[cptOffset]->cptno > 0 || cpt_dds[cptOffset]->no != groupno )
         {
           free( cpt_dds );
           return NULL;
         }

         if ( is_logical )
         {
           if ( cpt_dds[cptOffset]->flag & DM_FLAG_LOGICAL )
             cpt_dds[cptOffset] = dmi_descriptor_get_coord_no( cpt_dds[cptOffset], 1 );
           else
           {
             free( cpt_dds );
             return NULL;
           }
         }
         ++groupno;
       }
       else
         make_group = UT_FALSE;
     }

     if ( make_group )
       image_dd = dmi_coord_get_image( cpt_dds[0] );

      /* Let's make a temporary vector for the user to filter/bin on */
      dd = dmi_make_vector( (dmBlock*)block, (char*)name, cpt_dds, ncpts );
      dd->flag &= ~DM_FLAG_VISIBLE;
      dd->flag |= DM_FLAG_TEMP;

      if ( make_group )
      {
        /* need to regroup the image */
        ng = dmi_descriptor_get_no_groups( image_dd );
        new_ng = ng - ncpts + 1;
        new_groups = calloc( new_ng, UT_SZ_P );
        new_groups[cpt_dds[0]->no-1] = dd;

        for ( groupno = 1; groupno < cpt_dds[0]->no; groupno++ )
          new_groups[groupno-1] = dmi_descriptor_get_group_no( image_dd, groupno );

        for ( groupno = cpt_dds[ncpts-1]->no+1; groupno <= image_dd->ngroups; groupno++ )
          new_groups[groupno - ncpts] = dmi_descriptor_get_group_no( image_dd, groupno );

        dmi_descriptor_set_groups( image_dd, new_groups, new_ng );

        /* make sure we return the logical group, if the user requested it */
        if ( is_logical )
          dd = dmi_coord_get_parent( dd );
      }

      free( cpt_dds );
      return dd;
    }
  }

  /* Descriptor could not be found */
  return NULL;
}

/***************************************************************/
/*                                                             */
/*  Function:   dmi_block_search_col_identifier                */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/* Search a block for a named descriptor.                      */
/* Search the block's cols first then rawcols for the descr.   */
/* Search the cols' themselves, their vector components,       */
/* their coords, and groups.                                   */
/* Handles all '#' and (a,b) syntax.                           */
/* Will build a vector if needed to satisfy (a,b) syntax.      */
/* Return the descriptor if found and NULL otherwise.          */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmBlock  *block                             */
/*           const char     *name                              */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor  *dd                                */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_block_search_col_identifier( dmBlock* block, const char* name )
{
  dmDescriptor  *dd       = NULL;
  dmDescriptor **cpt_dds  = NULL;
  dmDescriptor  *image_dd = NULL;
  dmDescriptor **new_groups = NULL;
  dmBool         is_logical = UT_FALSE;
  dmBool         make_group = UT_TRUE;

  char **tcptNames = NULL;
  char **cptNames  = NULL;
  char   c = *name; 

  long ncpts     = 0;
  long cptOffset = 0;
  long match     = 0;
  long groupno   = 0;
  long new_ng    = 0;
  long ng        = 0;
 
 
  /* First, try '#' syntax */
  if ( c == '#' )
  {
    dd = dmi_block_search_hash_identifier( block, (const dmDescriptor**)block->cols, block->ncols, name );
    if ( dd ) 
      return dd;
   }

  /* Second, try a straight comparison of all vectors, coords, and groups */
  dd = dmi_descriptor_list_search( (const dmDescriptor**)block->cols, block->ncols, name );

  if ( dd )
    return dd;
 
  dd = dmi_descriptor_list_search( (const dmDescriptor**)block->rawcols, block->nrcols, name );
  if ( dd )
    return dd;

  /* Third, check vector '(' syntax */
  if ( c == '(' )
  {
    tcptNames = cs_parse_paren_list( name+1, &ncpts );

    if ( ncpts == 0 || ncpts > 2 )
    {
      /* this code can handle ncpts > 2, but later code can't. */
      /* might as well not go on... */
      ar_free_cp( tcptNames, ncpts+1 );
      return NULL;
    }
    cptNames = calloc( ncpts, UT_SZ_P );

    for ( cptOffset = 0; cptOffset < ncpts; cptOffset++ )
       cptNames[cptOffset] = tcptNames[cptOffset+1];

    free( tcptNames );
    cpt_dds = calloc( ncpts, UT_SZ_P );
    match = 0;
    dd = dmi_block_accumulate_hash_identifiers( block, (const dmDescriptor**)block->cols, block->ncols, (const char**)cptNames, cpt_dds, ncpts, &match );

    if ( dd )
    {
      ar_free_cp( cptNames, ncpts );
      free( cpt_dds );
      return dd;
    }

    /* now search for vector components the normal way */
    if ( match != ncpts )
    {
      dd = dmi_descriptor_list_accumulate( (const dmDescriptor**)block->cols, block->ncols, (const char**)cptNames, cpt_dds, ncpts, &match );
      if ( dd )
      {
        ar_free_cp( cptNames, ncpts );
        free( cpt_dds );
        return dd;
      }
    }

    if ( match != ncpts )
    {
      dd = dmi_descriptor_list_accumulate( (const dmDescriptor**)block->rawcols, block->nrcols, (const char**)cptNames, cpt_dds, ncpts, &match );
      if ( dd )
      {
        ar_free_cp( cptNames, ncpts );
        free( cpt_dds );
        return dd;
      }
    }  

    ar_free_cp( cptNames, ncpts );

    if ( match == ncpts )
    {
      is_logical = cpt_dds[0]->flag & DM_FLAG_LOGICAL;
      /* If these are image axes, we must make a new group */
      groupno = cpt_dds[0]->no;

      for ( cptOffset = 0; cptOffset < ncpts; cptOffset++ )
      {
        if ( cpt_dds[cptOffset]->flag & DM_FLAG_AXIS )
        {
          /* can't mix image axes with normal columns                  */
          /* can't regroup an already grouped axis                     */
          /* can't group logical with physical                         */
          /* the new grouping needs to be attached to consecutive axes */
          if ( ! make_group || cpt_dds[cptOffset]->cptno > 0 || cpt_dds[cptOffset]->no != groupno )
          {
            free( cpt_dds );
            return NULL;
          }

          if ( is_logical )
          {
            if ( cpt_dds[cptOffset]->flag & DM_FLAG_LOGICAL )
              cpt_dds[cptOffset] = dmi_descriptor_get_coord_no( cpt_dds[cptOffset], 1 );
            else
            {
              free( cpt_dds );
              return NULL;
            }
          }
          ++groupno;
        }
        else
          make_group = UT_FALSE;
      }

      if ( make_group )
        image_dd = dmi_coord_get_image( cpt_dds[0] );

      /* Let's make a temporary vector for the user to filter/bin on */
      dd = dmi_make_vector( block, (char*)name, cpt_dds, ncpts );
      dd->flag &= ~DM_FLAG_VISIBLE;
      dd->flag |= DM_FLAG_TEMP;

      if ( make_group )
      {
        /* need to regroup the image */
        ng = dmi_descriptor_get_no_groups( image_dd );
        new_ng = ng - ncpts + 1;
        new_groups = calloc( new_ng, UT_SZ_P );
        new_groups[cpt_dds[0]->no-1] = dd;

        for ( groupno = 1; groupno < cpt_dds[0]->no; groupno++ )
          new_groups[groupno-1] = dmi_descriptor_get_group_no( image_dd, groupno );

        for ( groupno = cpt_dds[ncpts-1]->no+1; groupno <= image_dd->ngroups; groupno++ )
          new_groups[groupno - ncpts] = dmi_descriptor_get_group_no( image_dd, groupno );

        dmi_descriptor_set_groups( image_dd, new_groups, new_ng );

        /* make sure we return the logical group, if the user requested it */
        if ( is_logical )
          dd = dmi_coord_get_parent( dd );
      }

      free( cpt_dds );
      return dd;
    }
  }
 
  /* Descriptor could not be found */
  return NULL;
}

/***************************************************************/
/*                                                             */
/*  Function:  dmi_block_search_hash_identifier                */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/* return the numbered logical column or axis depending on     */
/* block type.                                                 */
/* Use dds as the block's columns as they might be either      */
/* public or raw columns.                                      */
/* accepts #row, #col, and #1, #2 syntaxes.                    */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmBlock       *block\                       */
/*           const dmDescriptor **dds                          */
/*           const long         *name                          */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*               dmDescriptor *dd                              */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_block_search_hash_identifier( const dmBlock* block, const dmDescriptor** dds, const long n, const char* name )
{
  const char *cptr = NULL;
  long  cnum = 0;

  if ( cs_uceq( name, "#row" ))
    return (dmDescriptor*)block->row_dd;

  if ( cs_uceq( name, "#col" ))
  {
    cptr = name+4;
    cnum = atol( cptr );
    if ( cnum > 0 && cnum <= n )
      return (dmDescriptor*)dds[cnum-1];
  }

  cptr = name+1;

  if ( isdigit( *cptr ))
  {
    cnum = atol( cptr );
    if ( block->type == dmTABLE )
    {
      if ( cnum <= n && cnum > 0 )
        return (dmDescriptor*)dds[cnum-1];
     }
    else
      if ( block->type == dmIMAGE )
      {
        if ( cnum == 0 ) 
          return (dmDescriptor*)dds[0];
        return dmi_descriptor_get_logical_axis( dds[0], cnum );
      }
   }
  return NULL;
}

/***************************************************************/
/*                                                             */
/*  Function:  dmi_descriptor_list_accumulate                  */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/* Search a descriptor list for a list of descriptors.         */
/* If the full list of sought descriptors are just             */
/* the components of a vector or group, return the parent      */
/* vector.  Otherwise return NULL.                             */
/* If each sought descriptor is found individually,            */
/* add the found descriptor to cpt_dds, and increment          */
/* pmatch until *pmatch == target.                             */
/* Search of the the descriptors' themselves, their vector     */
/* components, their coords, and their groups.                 */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDescriptor  **dds                         */
/*           const long            n                           */
/*           const char          **names                       */
/*                 dmDescriptor  **cpt_dds                     */
/*          const  long            target                      */
/*                 long           *pmatch                      */     
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *dd                                  */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_list_accumulate( const dmDescriptor** dds, const long n, const char** names, dmDescriptor** cpt_dds, const long target, long* pmatch )
{
  dmDescriptor *dd = NULL; /* descriptor to return */
  long offset = 0;

  for ( offset = 0; offset < n; offset++ )
  {
    dd = dmi_descriptor_accumulate( dds[offset], names, cpt_dds, target, pmatch );

    if ( dd || *pmatch == target )
      return dd;
  }
  return dd;
}

/***************************************************************/
/*                                                             */
/*  Function: dmi_descriptor_accumulate                        */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/* Search a descriptor for a list of descriptors.              */
/* If the full list of sought descriptors are just             */
/* the components of a vector or group, return the parent      */
/* vector.  Otherwise return NULL.                             */
/* If each sought descriptor is found individually,            */
/* add the found descriptor to cpt_dds, and increment          */
/* pmatch until *pmatch == target.                             */
/* Search of the the descriptor itself, its vector components, */
/* its coords, and its groups.                                 */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDescriptor  *dd                           */
/*           const char         **names                        */
/*                 dmDescriptor **cpt_dds                      */
/*           const long           target                       */
/*                 long          *pmatch                       */
/*                                                             */
/*  Output parameters:                                         */
/*                none                                         */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *rdd                                 */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_accumulate( const dmDescriptor* dd, const char** names, dmDescriptor** cpt_dds, const long target, long* pmatch )
{
  dmDescriptor *rdd = NULL;  /* result descriptor */
 
  rdd = dmi_descriptor_accumulate_vector( dd, names, cpt_dds, target, pmatch );

  if ( rdd || *pmatch == target )
    return rdd;

  rdd = dmi_descriptor_accumulate_children( dd, names, cpt_dds, target, pmatch );
  return rdd;
}

/***************************************************************/
/*                                                             */
/*  Function:  dmi_descriptor_accumulate_vector                */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/* Search a descriptor for a list of descriptors.              */
/* If the full list of sought descriptors are just             */
/* the components of the vector, dd, return dd, o              */
/* therwise return NULL.                                       */
/* If each sought descriptor is found individually,            */
/* add the found descriptor to cpt_dds, and increment          */
/* pmatch until *pmatch == target.                             */
/* Search of the the descriptor itself and its vector          */
/* components.                                                 */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDescriptor  *dd                           */
/*           const char         **names                        */
/*                 dmDescriptor   cpt_dds                      */
/*           const long           target                       */
/*                 long          *pmatch                       */
/*                                                             */
/*  Output parameters:                                         */
/*                  none                                       */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *dd                                  */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_accumulate_vector( const dmDescriptor* dd, const char** names, dmDescriptor** cpt_dds, const long target, long* pmatch )
{
  dmDescriptor  *cpt_dd = NULL;
  dmDescriptor **cpts   = NULL;
  dmBool         all_match = UT_FALSE;
  char          *dd_name = NULL;

  long dim = 0;
  long cptOffset = 0;
  long nameOffset = 0;

  for ( nameOffset = 0; nameOffset < target; nameOffset++ )
  {
    if ( cs_uceq( dd->name, names[nameOffset] ))
    {
      cpt_dds[nameOffset] = (dmDescriptor*)dd;
      (*pmatch)++;

      if ( target == 1 )
        return (dmDescriptor*)dd;
      if ( *pmatch == target )
        return NULL;
    }
  }
 
  dim = dd->dim;

  if ( dim > 1 )
  {
    cpts = dd->cpts;
    /* remember, in order to consolidate the vector, the components must */
    /* be listed in the exact order */
    if ( target == dim )
      all_match = UT_TRUE;
    else
      all_match = UT_FALSE;

    for ( cptOffset = 0; cptOffset < dim; cptOffset++ )
    {
      cpt_dd = cpts[cptOffset];
      dd_name = cpt_dd->name;

      for ( nameOffset = 0; nameOffset < target; nameOffset++ )
      {
        if ( cs_uceq( dd_name, names[nameOffset] ))
        {
          cpt_dds[nameOffset] = cpt_dd;
          (*pmatch)++;
          break;
        }
        else
          if ( all_match && nameOffset == cptOffset )
            all_match = UT_FALSE;
      }

      if ( !all_match && *pmatch == target )
        return NULL;
    }
    if ( all_match )
      return (dmDescriptor*)dd;
  }
  return NULL; 
}

/***************************************************************/
/*                                                             */
/*  Function:  dmi_descriptor_accumulate_children              */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/* Search a descriptor for a list of descriptors.              */
/* If the full list of sought descriptors are just             */
/* the components of a group, return the parent                */
/* group.  Otherwise return NULL.                              */
/* If each sought descriptor is found individually,            */
/* add the found descriptor to cpt_dds, and increment          */
/* pmatch until *pmatch == target.                             */
/* Search the descriptor's coords and groups.                  */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDescriptor  *dd                           */
/*           const char         **names                        */
/*                 dmDescriptor **cpt_dds                      */
/*           const long           target                       */
/*                 long          *pmatch                       */
/*                                                             */
/*  Output parameters:                                         */
/*                 none                                        */
/*                                                             */
/*  Returns:                                                   */
/*             dmDescriptor *rdd                               */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_accumulate_children( const dmDescriptor* dd, const char** names, dmDescriptor** cpt_dds, const long target, long* pmatch )
{
  dmDescriptor  *rdd = NULL; /* returned descriptor */
  dmDescriptor **dds = NULL; /* list of descriptors */
  long n      = 0 ;   /* number of coords or groups */
  long offset = 0;    /* offset into coords or groups */

  /* search coords */
  n = dd->ncoords;

  if ( n > 0 )
  {
    dds = dd->coords;

    for ( offset = 0; offset < n; offset++ )
    {
       rdd = dmi_descriptor_accumulate( dds[offset], names, cpt_dds, target, pmatch );

       if ( rdd || *pmatch == target )
         return rdd;
    }
  }

  /* search groups */
  n = dd->ngroups;

  if ( n > 0 )
  {
    dds = dd->groups;

    for ( offset = 0; offset < n; offset++ )
    {
      rdd = dmi_descriptor_accumulate( dds[offset], names, cpt_dds, target, pmatch );

      if ( rdd || *pmatch == target )
        return rdd;
    }  
  }
  return NULL;
}


/***************************************************************/
/*                                                             */
/*  Function:   dmi_block_accumulate_hash_identifiers          */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/* Search for the numbered logical columns or axes depending   */
/* on block type listed in names.                              */
/* Use dds as the block's columns as they might be either      */
/* public or raw columns.                                      */
/* If the full list of sought descriptors are just             */
/* the components of a vector or group, return the parent      */
/* vector.  Otherwise return NULL.                             */
/* If each sought descriptor is found individually,            */
/* add the found descriptor to cpt_dds, and increment          */
/* pmatch until *pmatch == target.                             */
/* accepts #row, #col, and #1, #2 syntaxes.                    */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmBlock       *block                        */
/*           const dmDescriptor **dds                          */
/*           const long           n                            */
/*           const char         **names                        */
/*                 dmDescriptor **cpt_dds                      */
/*           const long           target                       */
/*                 long          *pmatch                       */
/*                                                             */
/*  Output parameters:                                         */
/*                none                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor *dd                                 */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_block_accumulate_hash_identifiers( const dmBlock* block, const dmDescriptor** dds, const long n, const char** names, dmDescriptor** cpt_dds, const long target, long* pmatch )
{
  dmDescriptor  *dd    = NULL;
  dmDescriptor *pdd    = NULL;
  dmDescriptor *parent = NULL;  /* another parent descriptor to compare to pdd */

  long cptOffset = 0;
  int  all_match = 0;

  /* look for '#' named components */
  for ( cptOffset = 0; cptOffset < target; cptOffset++ )
  {
    if ( *names[cptOffset] == '#' )
    {
      dd = dmi_block_search_hash_identifier( block, dds, n, names[cptOffset] );

      if ( dd )
      {
        cpt_dds[cptOffset] = dd;
        (*pmatch)++;
        if ( *pmatch == target )
          break;
      }
    }
  }

  /* if we found all of them, see if they all belong to the same vector */
  if ( *pmatch == target )
  {
    /* if we only needed one descriptor, return it */
    if ( target == 1 )
      return cpt_dds[0];

    /* Now check if we got all components of a vector */
    dd = cpt_dds[0];
    if ( dd->flag & DM_FLAG_AXIS )
      pdd = dmi_descriptor_get_group_of_axis( dds[0], dd );
    else
      pdd = dd->parent;

    if ( pdd )
    {
      /* remember, in order to consolidate the vector, the components must */
      /* be listed in the exact order */
      all_match = UT_TRUE;

      for ( cptOffset = 0; cptOffset < target; cptOffset++ )
      {
        dd = cpt_dds[cptOffset];
        if ( dd->flag & DM_FLAG_AXIS )
          parent = dmi_descriptor_get_group_of_axis( dds[0], dd );
        else
          parent = dd->parent;

        if ( parent != pdd || dd != parent->cpts[cptOffset] )
          all_match = UT_FALSE;
      }

      if ( all_match )
      {
        return pdd;
      }
    } /* end if parent */
  } /* end if found all cpts */ 

  return NULL;
}

/***************************************************************/
/*                                                             */
/*  Function:  dmi_descriptor_get_group_of_axis                */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/* The only way to get the group of an axis is to search       */
/* through the image groups for the axis.  Ugh.                */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDescriptor  *image_dd                     */
/*           const dmDescriptor  *axis_dd                      */
/*                                                             */
/*  Output parameters:                                         */
/*                  none                                       */
/*                                                             */
/*  Returns:                                                   */
/*           dmDescriptor *group_dd                            */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_descriptor_get_group_of_axis( const dmDescriptor* image_dd, const dmDescriptor* axis_dd )
{
  dmDescriptor *group_dd = NULL;
  long ngroups     = 0;
  long groupOffset = 0;
  long cptOffset   = 0;
  long dim         = 0;

  ngroups = image_dd->ngroups;

  for ( groupOffset = 0; groupOffset < ngroups; groupOffset++ )
  {
    group_dd = image_dd->groups[groupOffset];

    if ( axis_dd == group_dd ) 
      return (dmDescriptor*)group_dd;

    dim = group_dd->dim;

    if ( dim > 1 )
    {
      for ( cptOffset = 0; cptOffset < dim; cptOffset++ )
      {
	if ( axis_dd == group_dd->cpts[cptOffset] )
	  return (dmDescriptor*)group_dd;
      }
    }
  }
  return NULL;
}
