/*                                                                
**  Copyright (C) 2002-2012  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmparselib.c                                                */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include <string.h>
#include <stdio.h>
#include <float.h>
#include "dmi.h"

#define PT_BLOCK    0
#define PT_FILTER   1
#define PT_BIN      2
#define PT_SELECT   3
#define PT_EXT      4
#define PT_RENAME   5
#define PT_EXCLUDE  6
#define PT_OPTION   7
#define PT_SUBSPACE 8



/***************************************************************/
/*                                                             */
/* Function:   dmi_parse_list_free                             */
/*                                                             */
/* Description:                                                */
/*             Free a parse list object                        */
/*                                                             */
/*  Input parameters:                                          */
/*              dmParseList  *list                             */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            none                                             */
/*                                                             */
/***************************************************************/
void dmi_parse_list_free( dmParseList* list )
{
  if ( !list )
    return;

  free( list->block );
  free( list->select  );
  free( list->binning );
  free( list->filter );
  free( list->outname );   
  free( list->dsname );
  free( list->option );
  free( list->subspace);
  free( list );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_setup_runtime_filter                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock  *block                                */
/*              dmBuffer *buffer                               */
/*              dmSubspace *rts                                */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            none                                             */
/*                                                             */
/***************************************************************/
void dmi_block_setup_runtime_filter( dmBlock* block, dmBuffer* buffer, dmSubspace* rts )
{
  dmDescriptor  *parent       = NULL;
  dmDescriptor  *rts_dd       =  NULL;
  dmDescriptor  *coord_parent = NULL;

  long  i       = 0;
  int   is_unit = 0;
  long  cptno   = 0;
  long  errCode = dmSUCCESS;



 /* Here's how we pass the filters to the buffer, even if we don't call ApplySubspace */
  buffer->subspace = rts;

  for ( i = 0; i < rts->ncols; i++ )
  {
    rts_dd = rts->cols[i];
    parent = rts_dd->parent;
    if ( dmi_block_get_raw_type(block) == dmTABLE || parent->flag & DM_FLAG_IMG )  /* Don't buffer filters on image axes */
    {
      dmi_buffer_add_col( buffer, parent, DM_FLAG_FILT );

      /*
       * If parent is a WCS virtual column, then we need to mark the parent of
       * its transform (i.e. the real column it's derived from) as a filter
       * operand so that it's copied to the buffer before filtering.
       *
       * The procedure here mirrors the one in dmi_block_apply_select().
       */

      if ( dmi_descriptor_get_type( parent ) == dmCOORD )
      {
	is_unit = dmi_coord_transform_is_unit( parent );
	coord_parent = dmi_coord_get_parent( parent );

	if ( is_unit )
	{
	  if ( coord_parent ) parent->flag |= DM_FLAG_RAW & coord_parent->flag;
	}
	else
	{
	  if ( coord_parent )
	    dmi_buffer_add_col( buffer, coord_parent, DM_FLAG_OPERAND|DM_FLAG_FILT_OPERAND );
          else 
          {
           sprintf( dm_errbuf, "No coord parent found for %s", dmi_descriptor_get_name( parent ));
           dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
          } 

          parent->flag |= DM_FLAG_RESULT;

	  /* If parent is a vector, also flag its components as results so
	   * dmGetVectorArrays() sees that they're in the buffer */
	  if ( parent->dim > 1 )
	  {
	   for ( cptno = 0; cptno < parent->dim; cptno++ )
            parent->cpts[cptno]->flag |= DM_FLAG_RESULT;
	  }
	}
      }
    }
    dmi_filter_check_regions( rts_dd, rts->ncpts );
  }

  /* 
   If we are filtering, add the #row pseudo-column to the buffer.
   Should check if this is really required.
  */

  if ( rts->ncols > 0 )
   dmi_buffer_add_col( buffer, block->row_dd, DM_FLAG_RESULT | DM_FLAG_ROW );

  if ( (errCode = dmu_err_get()) != dmSUCCESS )
  {
     fprintf( stderr, "Error in filter setup\n" );
     dmu_err_get_message( dm_errbuf, dmMAXLEN );
     fprintf( stderr, "%s\n", dm_errbuf );
  }

}



/***************************************************************/
/*                                                             */
/* Function:   dmi_block_update_runtime_subspace               */
/*                                                             */
/* Description:                                                */
/*             Apply the subspace if required, and invert it   */
/*             if required - in that case returning the new    */
/*             subspace                                        */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock    *block                              */
/*              dmSubspace *rts                                */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            dmSubspace *subspace                             */
/*                                                             */
/***************************************************************/
dmSubspace* dmi_block_update_runtime_subspace( dmBlock* block, dmSubspace* rts )
{
  dmSubspace  *subspace = NULL;
  dmSubspace  *irts     = NULL;
  dmBuffer    *buffer   = block->buffer;
  dmBool       copyDefault  = UT_TRUE;
  dmBool       requireTable = UT_TRUE;


  if ( rts->invert )
  {
    subspace = block->subspace;
    subspace->invert = rts->invert;

    if ( dmi_block_get_subspace_update( block ))   
    {
      irts = dmi_subspace_invert( rts );
      dmi_subspace_free( rts, NULL );
      rts = irts;

      if ( dmverbose >= 1 )
        dmi_subspace_print( rts );
      rts->invert = 0;
    }
  }

  /* Reattach the new rts to the buffer */
  buffer->subspace = rts;

  if ( dmi_block_get_subspace_update( block ))   
  {
    /* Intersect filter with block subspace. Suppressed by  [opt -ssupdate] */
    dmi_subspace_apply( rts, block->subspace, block, copyDefault, requireTable );  
  }

  return rts;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_parse_syntax                          */
/*                                                             */
/* Description:                                                */
/*  Open the dataset if it is not already open,                */
/*                                                             */
/*   If vspec is NULL we pass it down                          */
/*   If we have a dataset and vspec doesn't begin with [, it's */
/*   just a block name and we pass it on down.                 */
/*   Even if we have a dataset, if vspec begins with [ we must */
/*   parse it. If we don't have a dataset, the non-[ bit of    */
/*   vspec will be the dataset name.                           */
/*   If vspec and ds are NULL, throw an error.                 */
/*                                                             */
/*  Input parameters:                                          */
/*              dmDataset  *dd                                 */
/*              const char *vspec                              */
/*              dmBlockType blockType                          */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            dmParseList *list                                */
/*                                                             */
/***************************************************************/
dmParseList* dmi_block_parse_syntax( dmDataset* ds, const char* vspec, dmBlockType blockType )
{
  dmParseList *list = NULL;

  /* Handle null arguments */
  if ( !ds && !vspec )
  {
    dmu_err_set( DM_ERR_NULL, "[dmi_block_parse_syntax]: null arguments" );
    return NULL;
  }

  if ( !ds  || (vspec && *vspec == '[' ))
  {
    list = dmi_parse_split( vspec );

    if ( !list )
    {
      if ( dmu_err_get() == dmSUCCESS ) /* Unknown error */
      {
        sprintf( dm_errbuf, "DM syntax parsing failed: %s", vspec );
        dmu_err_set(DM_ERR_PARSE,dm_errbuf );
      }
      return NULL;
    }

    /* disallow bin specifications on table opens (only for images)*/
    list->type = blockType;

    if( list->binning ) 
    {
      if ( blockType == dmTABLE )
      {
        sprintf( dm_errbuf, "Table open was requested but binning syntax implies image: %s", vspec );
        dmu_err_set( DM_ERR_PARSE, dm_errbuf );
        return NULL;
      }
      list->type = dmIMAGE;
    }
  }

  return list;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_parse_split                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/* DM PHASE I PARSER                                           */
/*                                                             */
/*   Top level parse a DM specification separating it into     */
/*   its component parts. The syntax is                        */
/*   filename[block][filter <filter>][bin <bin>][cols <cols>][opt options][name newname] */
/*   We allow multiple occurrences of each type now, simply    */
/*   appending the lists.                                      */
/*                                                             */
/*  Input parameters:                                          */
/*            const char *buf                                  */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            dmParseList *list                                */
/*                                                             */
/***************************************************************/
dmParseList* dmi_parse_split( const char* buf )
{
  dmParseList *list = NULL;
  dmBool       bExclude = dmFALSE;
  dmBool       bFilter  = dmFALSE;

  char  *ptr = (char*)buf;
  char  *command = NULL;
  char  *select  = NULL;
  char  *bin     = NULL;
  char  *tptr    = NULL;
  char  *cptr    = NULL;
  char  *subspace = NULL;
  char  *exclude_prefix = NULL;

  long   type    = 0;
  int    compact = 1;
  int    got_block    = 0;
  int    parse_status = 0;

  if ( !buf )
    return NULL;

  list = (dmParseList*)calloc( 1, sizeof( dmParseList ));

  if ( !list )
    return NULL;

  /* Special cases for interactive demo parser */
  while ( *ptr != '#' && ( *ptr == ' ' || *ptr == '\t' ) )
    ++ptr;

  if ( *ptr == '#' || *ptr == '\0' )
    return NULL;


  /* The DS name is everything up to the first []. */
  while( *ptr == ' ' )
    ++ptr;

  list->dsname = cs_parse_term_token( &ptr, '[' );

  while( ptr && *ptr )
  {
    /* Skip ahead to value */
    cptr = cs_parse_paren_token_validate( &ptr, '[', ']', NULL, 0, &parse_status );

    if ( !parse_status )
    {
      sprintf( dm_errbuf, "DM Parse error: unmatched [ or (: %s\n", ptr );
      dmu_err_set( DM_ERR_PARSE, dm_errbuf );
      free( list );
      return NULL;
    }

    /* Only process non-empty components */
    if ( cptr && *cptr )
    {
      tptr = cptr;
      type = dmi_parse_token_type( &tptr, &compact );
      command = cs_dup( tptr );
      free( cptr );
      ++got_block;

      switch( type ) 
      {
        case PT_SELECT: 
          cs_append( &select, ",", command );
          break;

        case PT_BLOCK:
	  // Code Change Bug 10558 July 5, 2007
          //  Remove restriction of block names should not have
	  //  embedded spaces, e.g. "SPECRESP MATRIX" should be 
          //  acceptable block name

#if 0
          c = strchr( command, ' ' ); /* Does it have a space? */
          if ( c )
          {
            sprintf( dm_errbuf, "DM Parse error: bad format for syntax: %s\n", command );
            dmu_err_set( DM_ERR_PARSE, dm_errbuf );
            free( command );     
            free( list );
            return NULL;
          }
          else
#endif

          if ( got_block == 1 ) 
          {
            list->block = command;
          }
          else
            if ( list->outname )
            {
              sprintf( dm_errbuf, "DM Parse error: too many block name entries: %s\n", command );
              dmu_err_set( DM_ERR_PARSE, dm_errbuf );
              free( command );
              free( list );
              return NULL;
            }
            else
            {
              list->outname = command;
            }
            break;

        case PT_BIN:
          if ( bin )
          {
            sprintf( dm_errbuf, "DM Parse error: only one bin command allowed" );
            dmu_err_set( DM_ERR_PARSE, dm_errbuf );
            free( command );
            free( list );
            return NULL;
          } 
          else
            cs_append( &bin, ",", command );
          break;

        case PT_OPTION:
          if ( dmverbose >= 3 )
            fprintf( stderr, "Set kernel option %s\n", command );
          list->option = command;
          break;

        case PT_EXCLUDE:
	  // Code Change Bug 12745 January 18, 2012
          // Check for filter flag before doing exclude

          if ( bFilter )
          {
            sprintf( dm_errbuf, "DM Parse error: cannot mix EXCLUDE and FILTER" );
            dmu_err_set( DM_ERR_PARSE, dm_errbuf );
            free( command );
            free( list );
            return NULL;
          }

          bExclude = dmTRUE;
          exclude_prefix = cs_dup( "#!" );

          cs_append( &list->filter, ",", exclude_prefix );
          cs_append( &list->filter, " ", command );
          break;
 
       case PT_FILTER:
	  // Code Change Bug 12745 January 18, 2012
          // Check for exclude flag before doing filter
          //  If ok, set filter flag

          if ( bExclude )
          {
            sprintf( dm_errbuf, "DM Parse error: cannot mix EXCLUDE and FILTER" );
            dmu_err_set( DM_ERR_PARSE, dm_errbuf );
            free( command );
            free( list );
            return NULL;
          }

          bFilter = dmTRUE;
          cs_append( &list->filter, ",", command );
          break;

        case PT_SUBSPACE: 
          cs_append( &subspace, ",", command );
          break;

        default:
          sprintf( dm_errbuf, "DM Parse error: unknown syntax: %s", command );
          dmu_err_set( DM_ERR_PARSE, dm_errbuf );
          free( command );
          free( list );
          return NULL;

      } //end switch
    }

    /* If characters remain in ptr, insist that the next one is '[' */
    if ( ptr && *ptr )
    {
      if ( *ptr != '[' )
      {
        sprintf( dm_errbuf, "DM Parse error: unknown syntax: %s", ptr );
        dmu_err_set( DM_ERR_PARSE, dm_errbuf );
        free( list );
        return NULL;
      }
    }
  }

  /* Parse the [cols a,b,c] list */
  list->select   = select;
  list->subspace = subspace;

  /* Parse the [bin x,y] list */
  list->binning = bin;
  list->compact = compact;
  return list;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_parse_token_type                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Here we recognize the different kinds of [] section in the */
/*  DM syntax                                                  */
/*  In general a space denotes a command                       */
/*  but we remove spaces if we don't get a command we recognize*/
/*                                                             */
/*  Input parameters:                                          */
/*              char  **ptr                                    */
/*              int    *compact                                */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            int token_type                                   */
/*                                                             */
/***************************************************************/
int dmi_parse_token_type( char** ptr, int* compact )
{
  char  commandbuf[TEXT_SIZE];
  char *command = commandbuf;
  char *tptr    = *ptr;
  char *pptr    = *ptr;
  char  opset[] = "=<>:(,";  

  /* @ denotes an external filter */

  if ( !tptr )
    return PT_BLOCK;

  if ( *tptr == '@' )
    return PT_FILTER;

  /* If it has a space, it's a command word */
  cs_parse_paren_token( &tptr, UT_NULL_CHAR, ' ', command, TEXT_SIZE );

  if ( *tptr == '\0')
    tptr = NULL;

  if ( tptr )
  {
    while( *tptr == ' ' || *tptr == '\t' )
      ++tptr;

    *ptr = tptr;
 
     if ( cs_uceq( command, "filter" ) )
       return PT_FILTER;
     else
       if ( cs_uceq( command, "exclude" ) )
         return PT_EXCLUDE;
       else
         if ( cs_uceq( command, "option" )|| cs_uceq( command, "options" ) ||
              cs_uceq( command, "opt" ))
            return PT_OPTION;
         else
           if ( cs_uceq( command, "bin" ) ) 
             return PT_BIN;
           else
             if ( cs_uceq( command, "binf" ) || cs_uceq( command,"bin/full" ) 
                  ||  cs_uceq( command,"bin/f" ) )
             {
               *compact = 0;
               return PT_BIN;
             }
             else
               if (   cs_uceq( command, "cols" ) || cs_uceq( command, "col" ) ) 
                  return PT_SELECT;
               else
                 if (   cs_uceq( command, "rename" ) )
                   return PT_BLOCK;
                 else
                   if (   cs_uceq( command, "columns" ) ) 
                     return PT_SELECT;
                   else
                     if (   cs_uceq( command, "subspace" ) ) 
                       return PT_SUBSPACE;
                     else 
		     /* Not a known command, must be user error */
                     /*  - ignore extraneous space */
                     {
                       tptr = NULL;
                       *ptr = pptr;  
                     }
  }

  /* Not a recognized command word */
  /* If it contains =<>, must be a filter not a blockname */
  if ( !tptr ) 
    tptr = strpbrk( *ptr, opset );

  if ( tptr )
    return PT_FILTER;
  else
    if ( *ptr && **ptr == '(' )  /* Compound filter syntax */
      return PT_FILTER;

  return PT_BLOCK;
}




/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_check_regions                        */
/*                                                             */
/* Description:                                                */
/* Force the filters to be created and check the regions and   */
/* their string versions are in sync                           */
/*                                                             */
/*  Input parameters:                                          */
/*              dmDescriptor  *rts_dd                          */
/*              long           ncpts                           */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            none                                             */
/*                                                             */
/***************************************************************/
void dmi_filter_check_regions( dmDescriptor* rts_dd, long ncpts )
{
  dmFilter *filter = NULL;
  long      rtscpt = 0;


  /* Get size for first cpt */
  while ( !filter && rtscpt < ncpts )
  {
    ++rtscpt;
    filter = dmi_subspace_col_get_filter( rts_dd, rtscpt );
  }

  for ( rtscpt = 1; rtscpt <= ncpts; rtscpt++ )
  {
    filter = dmi_subspace_col_get_filter( rts_dd, rtscpt );

    if ( filter )
    {
      /* Convert copy of region to one in pixel space only */
      if ( filter->region && !filter->regstring )
      {
        filter->regstring = malloc( (dmMAXSTR + 1 ) * UT_SZ_C );
        regComposeRegion( filter->region, filter->regstring, dmMAXSTR );
      }

      if ( filter->regstring && !filter->region )
      {
        /* No support for reading region files at this layer, so hope we don't get here */
        filter->region = dmi_region_resolve( filter->regstring, NULL, NULL, NULL );

        if ( filter->region )
        {
          fprintf( stderr, "DM INTERNAL ERROR: LATE REGION PARSE\n" );
        }
        else
        {
          dmu_err_set( DM_ERR_PARSE, "Region syntax error" );
         return;
        }
      }
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_parse_bin_spec                              */
/*                                                             */
/* Description:                                                */
/*    Receives from caller a string that needs to broken into  */
/*    three segments: lower bound, upper bound, and a number   */
/*    that is either bin size or number of bins.               */
/*    The string can omit any ora all of the input values.  t  */
/*    A single token value as the input string is considered to*/
/*    either the number of bins or the bin size.               */
/*    This function allows for the following syntax forms:     */
/*                                                             */
/*          lr : ur : n                                        */
/*          lr : ur :                                          */
/*          lr : ur                                            */
/*          lr :    : n                                        */
/*          lr :    :                                          */
/*             : ur : n                                        */
/*             : ur :                                          */
/*             :    : n                                        */
/*                    n                                        */
/*             :    :                                          */
/*                                                             */
/*     where lr is lower range limit, ur is upper range limit, */
/*     and n is defined as either                              */
/*           n  - bin size                                     */
/*       or  #n - number of bins                               */
/*                                                             */
/*     Code Change  Bugs 11788, 11800, 8756  October 13, 2011  */
/*     Function rewritten to correct these problems            */
/*                                                             */
/*  Input parameters:                                          */
/*       char  *spec       - input bin string                  */
/*       double *specvals  - contains default values for       */
/*                           min, max range and bin size       */
/*       long   specFlag   - flag indicating compact, full     */
/*                           logarithmic bins or none          */
/*       int    is_integer - integer data type col flag        */
/*                                                             */
/*  Output parameters:                                         */
/*             double *specvals - return parsed values for min,*/
/*                                max, and bin size            */
/*                                                             */
/*  Returns:                                                   */
/*            long  retval - modified value of specFlag        */
/*                                                             */
/***************************************************************/
long dmi_parse_bin_spec( char *spec, double *specvals, long specFlag, int is_integer )
{
  char  *pItem  = NULL;
  char  *pSpec  = spec;
  char  *pTemp  = spec;
  char  *pRetPtr = NULL;
  char   cColon = ':';
  char   cOctothorpe = '#';
  char   cNull  = '\0';

  double dItem  = 0.0;

  long   nItems = 0;
  long   retval = specFlag;
  long   ivals  = 0;
  long   nBins  = 0;
  long   ii     = 0;

  dmBool bError = dmFALSE;

  enum  specParts { LRNG, URNG, NBINS };


  /* null string, just return */
  if ( !spec || *spec == cNull )
    return retval;

  /* Special case: grid file */
  if ( cs_uceq_siz( spec, "grid(", 5 ))
  {
    retval |=  DM_BINS_FILE;
   /* Special case: one value only is treated as last item */
  }
  else
  {
    /* Determine the number of colon delimiated segments, max is 3 */
    /* Added for Bug 11800                                         */

    while ( (pTemp = strchr( pTemp, cColon )) != NULL )
    {
      ++nItems;
      ++pTemp;
    }

    if ( nItems > 2 )
    {
      sprintf( dm_errbuf, "DM parse error, invalid binning syntax, too many terms: %s", spec );
      dmu_err_set( DM_ERR_PARSE, dm_errbuf );
      return retval;
    }

    /* Adjust item count to 1s based value*/
    ++nItems;



    /* Cut segments and store values */

    for( ii = 0; ii < nItems; ii++ )
    {
       pRetPtr = NULL;
       pItem   = NULL;
       dItem   = 0.0;

       if ( nItems == 1 )
         ii = NBINS;

       if ( pSpec && *pSpec )
         pSpec = dmu_parse_token( pSpec, cColon, &pItem, 0 );

       if ( pItem && *pItem )
         dItem = strtod( pItem, &pRetPtr );
       else
         continue;

       /* if conversion ok, store value */
       if ( pRetPtr && *pRetPtr == cNull )
          specvals[ii] = dItem;
       else
       {
         // check the invalid character triggering conversion error

         if ( *pItem != cOctothorpe )
           bError = dmTRUE;
         else
         {
           /* first char is #                  */
           /*  only allowed for number of bins */

           if ( ii != NBINS )
  	   {
             // error condition - invalid char
             bError = dmTRUE;
           }
           else
           {
             /* working with #n */
             /*  move past # char, set flag */  
             pTemp = pItem+1;
             retval |= DM_BINS_N;

             nBins  = (strtol( pTemp, &pRetPtr, 10 ));
 
             if ( pRetPtr && *pRetPtr != cNull )
               // conversion error
               bError = dmTRUE;

           } // end else
         } // end pItem != #
       } // end else  pRetPtr == cNull

       /* Clean Up */
       if (pItem && *pItem )
         free(pItem);

       /* Verify number conversion ok*/
       if ( bError ) 
       {
         sprintf( dm_errbuf, "DM parse error, invalid binning syntax: %s", spec );
         dmu_err_set( DM_ERR_PARSE, dm_errbuf );
         return retval;
       }
     } // end for loop


     /* integer values require extra handling to move boundary off */
     /* even integer value                                         */

     if ( is_integer )
     {
        ivals = (long)specvals[LRNG];

        if ( (double)ivals == specvals[LRNG] )
          specvals[LRNG] -= 0.5;

        ivals = (long)specvals[URNG];

        if ( (double)ivals == specvals[URNG] )
          specvals[URNG] += 0.5;
     }


     if (retval & DM_BINS_LOG) 
     {
       if ( specvals[URNG] <= 0 )
         specvals[URNG] = 1.0;

       if ( specvals[LRNG] <= 0 )
         specvals[LRNG] = ( specvals[URNG] > 1.0 ) ? 1.0 : 0.1 * specvals[URNG];

       if ( nBins )              
         specvals[NBINS] = ( log10( specvals[URNG] ) - log10( specvals[LRNG] ) ) / nBins;
     }
     else
     {
       if ( nBins )
         specvals[NBINS] = ( specvals[URNG] - specvals[LRNG] ) / nBins;
     }


     /* catch zero bin number or bin size */

     if ( specvals[NBINS] == 0.0 )
       specvals[NBINS] = 1.0;

  }

  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_parse_bin_token                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock  *block                                 */
/*             char     *itemp                                 */
/*             long      nitems                                */
/*             char    **specp                                 */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor *col                                */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_parse_bin_token( dmBlock* block, char* itemp, long nitems, char** specp )
{
  dmDescriptor *col  = NULL;
  dmDescriptor *dd   = NULL;
  char         *spec = NULL;
  char         *item = itemp;
  char         *name = NULL;
  long          groupno = 0;
  long          vcpt    = 0;
  double        dval    = 0;
  long          axis_no = 0;


  /*
   *  Read up to equals sign
   *Cases:       x=4:4          name = x, spec = 4:4
   *             x              name = x, spec =NULL
   *             4              name = NULL, spec = 4    (special case)
   */

  cs_advptr( &item );  /* Skip whitespace */

  spec = dmu_parse_token( item, '=', &name, 0 );

  // Code Change  Bug 12596 (revisited)  April 15, 2011
  //  Check var 'name' for possible NULL value returned from dmu_parse_token()

  if ( !spec  && name && dmu_lex_number( name, &dval ) && nitems == 0 )
  {
    /* There's no bin = ::; we just keep name unless name is actually
     * a number, when we invoke the special case
     *   [bin 4]  ==>  [bin #G1=4]
     */
    spec = name;
    name = NULL;
    axis_no = nitems + 1;

    if ( block->type == dmIMAGE )
    {
      if ( block->cols && block->cols[0] )
        dd = block->cols[0];
      else
        dd = block->rawcols[0];

      groupno = dmi_descriptor_get_groupno_of_axis( dd, axis_no, &vcpt );
      col = dmi_descriptor_get_group_no( dd, groupno );
      col = dmi_descriptor_get_unscaled( col );  /* Logical axis for default binning */
    }
    else
      col = block->rawcols[axis_no-1];
  }
  else
  {
    col = dmi_block_search_col_identifier( block, name );
  }

  if ( col && dmverbose > 0 )
    printf( "Bin on %s\n", col->name );
#if DEBUG
   printf( "Bin on %s item=%s %ld\n", dmi_descriptor_get_qualname( col, dm_errbuf ), item, nitems );
#endif

  if ( !col )
  {
    if ( name )
      sprintf( dm_errbuf, "DM error: attempt to bin on unknown identifier %s", name );
    else
      sprintf( dm_errbuf, "DM error: bin on null identifier failed" );

    dmu_err_set( DM_ERR_PARSE, dm_errbuf );
  }
  else
    if ( block->type == dmIMAGE && !( col->flag & DM_FLAG_AXIS ))
    {
      sprintf( dm_errbuf, "DM Error: Attempt to bin in invalid desctiptor %s", name );
      dmu_err_set( DM_ERR_PARSE, dm_errbuf );
      col = NULL;
    }

    free( name );
    *specp = spec;

    return col;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_parse_bin_weight                            */
/*                                                             */
/* Description:                                                */
/*   Given an as-yet unparsed binning spec, extract the        */
/*   weight-column specifier and return the corresponding      */
/*   descriptor.  The weight spec remains in the bin_spec      */
/*   string, but it is separated from the rest of the bin spec */
/*   with a null char.  This allows the bin spec to later be   */
/*   parsed without concern for the weight spec.  If there is  */
/*   no weight spec (or it's empty), the function will         */
/*   return NULL and set *weight_spec to NULL.  If there is a  */
/*   weight spec, but it indicates a non-existent or invalid   */
/*   column, the return value will be NULL, but*weight_spec    */
/*   will still point to the weight spec.                      */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock  *block                                 */
/*             char     *bin-spec                              */
/*             char    **weight_spec                           */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor *weight_dd                          */
/*                                                             */
/***************************************************************/
static dmDescriptor* dmi_parse_bin_weight( dmBlock* block, char* bin_spec, char** weight_spec )
{
  dmDescriptor *weight_dd = NULL;
  dmDataType    type;


  /* Weight spec is separated from rest of bin spec by a semi-colon */
  *weight_spec = strchr( bin_spec, ';' );

  if ( *weight_spec )
  {
    /* Mark start of weight spec */
    **weight_spec = '\0';
    (*weight_spec)++;

    /* Trim leading and trailing spaces */
    cs_advptr( weight_spec );
    cs_trim( *weight_spec );

    if ( !(**weight_spec) )
    {
      /* Empty weight spec -- ignore */
      *weight_spec = NULL;
    }
    else
    {
      if ( block->type == dmIMAGE )
      {
        sprintf( dm_errbuf, "DM error: weighted binning not supported on images" );
        dmu_err_set( DM_ERR_PARSE, dm_errbuf );
        return NULL;
      }

      weight_dd = dmi_block_search_col_identifier( block, *weight_spec );

      if ( !weight_dd )
      {
        sprintf( dm_errbuf, "DM error: attempt to bin with unknown weight identifier %s",
        *weight_spec );
        dmu_err_set( DM_ERR_PARSE, dm_errbuf );
        return NULL;
      }

      type = weight_dd->type;

      if ( type == dmTEXT || type == dmBIT || type == dmUNKNOWNTYPE )
      {
        sprintf( dm_errbuf, "DM error: invalid data type for bin weight descriptor %s",
        weight_dd->name );
        dmu_err_set( DM_ERR_PARSE, dm_errbuf );
        return NULL;
      }

      if ( weight_dd->dim > 1 || dmi_descriptor_get_array_size( weight_dd ) > 1 )
      {
        sprintf( dm_errbuf, "DM error: non-scalar bin weight descriptor %s", weight_dd->name );
        dmu_err_set( DM_ERR_PARSE, dm_errbuf );
        return NULL;
      }
    }
  }

  return weight_dd;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_binning_is_one                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor  *dd                               */
/*             char          *spec                             */
/*             char          *name                             */
/*             char          *filter                           */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            int ok                                           */
/*                                                             */
/***************************************************************/
int dmi_binning_is_one( dmDescriptor* dd, char* spec, char* name, char* filter )
{
  dmDescriptor *cpt_dd = NULL;
  char   result[256];
  double specvals[3] = { 0.0, 0.0, 0.0 };
  double bounds[2];
  long   retval = 0;
  long   dim    = 0;
  long   cptno  = 0;
  int    isint  = 0;
  int    ok     = 0;


  dmi_parse_bin_spec( spec, specvals, retval, isint );
  ok = ( specvals[2] == 1.0 || specvals[2] == 0.0 )  
       && specvals[0] != 0.0 && specvals[1] != 0.0;

  /* Need to make sure we are binning inside the image field.
   * Otherwise, we still need to bin. */
  if ( ok )
  {
    if ( dd->flag & DM_FLAG_AXIS )
    {
      dim = dd->dim;

      for ( cptno = 1; cptno <= dim && ok; cptno++ )
      {
        cpt_dd = dmi_descriptor_get_cpt( dd, cptno );
        dmi_coord_get_range( cpt_dd, cptno, DT_REAL8, &bounds[0], &bounds[1] );
        ok = ( specvals[0] >= bounds[0] && specvals[1] <= bounds[1] );
      }
    }
  }

  if ( ok )
  {
    if ( *filter )
      strcat( filter, "," );
    sprintf( result, "%s=%g:%g", name, specvals[0], specvals[1] );
    strcat( filter, result );
  }

  return ok;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_check_binning_special_case                  */
/*                                                             */
/* Description:                                                */
/*  First pass through binning spec to check for special case: */
/*   when rebinning an image with pixel size 1, use filtering  */
/*   instead  i.e.  "img.fits[bin x=1:100,y=1:100]"            */
/*   is treated as                                             */
/*        "img.fits[filter x=1:100,y=1:100]"                   */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock    *block                               */
/*             const char *bin_spec                            */
/*             int        *specialp                            */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            char *spec                                       */
/*                                                             */
/***************************************************************/
char* dmi_check_binning_special_case( dmBlock* block, const char* bin_spec, int* specialp )
{
  dmDescriptor *col = NULL;
  char   *ptr  = cs_dup( bin_spec );
  char   *item = NULL;
  char    col_buf[1024] = "";
  char    filter[256]   = "";
  char   *spec          = NULL;
  char   *weight_spec   = NULL;
  long     n            = 0;
  long     nitems       = 0;
  int      special      = 0;

  if ( !bin_spec || *bin_spec == '\0' )
    return NULL;

  /* Get name of weight column (if any) */
  col = dmi_parse_bin_weight( block, ptr, &weight_spec );

  if ( weight_spec )
  {
    if ( !col )
      return NULL;
    strcat( col_buf, col->name );
    strcat( col_buf, "," );
  }

  n = dmu_count_commas( bin_spec );

  if ( block->type == dmIMAGE )
    special = 1;

  while( ptr && *ptr && nitems < n )
  {
    ptr = dmu_parse_token( ptr, ',', &item, 0 );

    // Code Change  Bug 12596 (revisited)  April 15, 2011
    //  Check var 'item' for null - meaning bad input string
    //  If null, continue. Var 'ptr' already incremented by function

    if ( !item )
      continue;

    col = dmi_parse_bin_token( block, item, nitems, &spec );

    if ( !col )
      return NULL;

    if ( special )
      special = dmi_binning_is_one( col, spec, col->name, filter );

    if ( nitems > 0 )
      strcat( col_buf, "," );

    strcat( col_buf, col->name );
    ++nitems;
    free( item );
  }

  /* If we're rebinning an existing image with weights (which we
   * can't yet, but it's conceivable), then we can't convert
   * the bin spec to a filter */

  if ( weight_spec )
    *specialp = 0;
  else
    *specialp = special;

  free( ptr );

  if ( special )
    return cs_dup( filter );

  return cs_dup( col_buf );
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_block_open_set                              */
/*                                                             */
/* Description:                                                */
/*  The default way to bin an image if no [bin ...] directive  */
/*  is present is                                              */
/*  (1) use the selected columns from [cols ...]               */
/*  (2) use the CPREFn keywords (FITS) or other kernel         */
/*      dependent hint                                         */
/*  (3) use the first two columns (per FITS kernel choice)     */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock     *block                              */
/*             dmParseList *list                               */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            none                                             */
/*                                                             */
/***************************************************************/
void dmi_block_open_set_default_select(  dmBlock* block, dmParseList* list )
{
  char pbuf[512] = "";
  long i         = 0;

  if ( !list->select )
  {
    if ( block->npref > 0 )
    {
      for ( i = 1; i <= block->npref; i++ )
      {
        if ( block->pref[i] )
        {
          if ( i > 1 )
            strcat( pbuf, "," );

          strcat( pbuf, block->pref[i] );
        }
      }

      list->select = cs_dup( pbuf );
    }



    if ( dmverbose > 0 )
      printf( "Set default select: [cols %s]\n", list->select );
  }

  if ( !list->binning )
  {
    if ( list->select)
      list->binning = cs_dup( list->select );

    if ( dmverbose > 0 )
      printf( "Set default binning: [bin %s]\n", list->binning );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_open_set_binning                      */
/*                                                             */
/* Description:                                                */
/* This rearranges the binning strings to handle special cases */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock     *block                              */
/*             dmParseList *list                               */
/*             dmBlockType  btype                              */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            none                                             */
/*                                                             */
/***************************************************************/
void dmi_block_open_set_binning( dmBlock* block, dmParseList* list, dmBlockType btype )
{
  int   special = 0;
  char *col_buf = NULL;

  if ( !list )
    return;

  if ( block->type == dmTABLE && btype == dmIMAGE && !list->binning )
    dmi_block_open_set_default_select( block, list );

  if ( list->binning )
  {
    col_buf = dmi_check_binning_special_case( block, list->binning, &special );

    if ( special )
    {
      if ( dmverbose >= 1 )
        fprintf( stderr, "Syntax %s interpreted as filter %s\n", list->binning, col_buf );
      free( list->filter );
      free( list->binning );
      list->binning = NULL;
      list->filter = col_buf;
    }
    else
      if ( block->type == dmTABLE )
      {
        /* Don't add bin cols to select if already image */
        if ( list->select )
        free( list->select );

       list->select = col_buf;
      }
  }

  if ( list->select && list->select[0] == '\0' )
  {
    free( list->select );
    list->select = NULL;
  }
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_block_apply_select                          */
/*                                                             */
/* Description:                                                */
/*   Parse the syntax                                          */
/*   cols  name1,name2,name3...    List of selected columns    */
/*   cols  name1,name2,*...        List of selected columns,   */
/*                                 append remainder in standard*/
/*                                 order                       */
/*   cols  -name1                  Deleted column              */
/*   cols  foo=bar                 Renamed column              */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock  *block                                 */
/*             char     *select                                */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            none                                             */
/*                                                             */
/***************************************************************/
void dmi_block_apply_select( dmBlock* block, char* select )
{
  dmDescriptor   *dd      = NULL;
  dmDescriptor   *prev_dd = NULL;
  dmDescriptor   *ddc     = NULL;
  dmDescriptor   *ddp     = NULL;
  dmDescriptor  **cols    = NULL;  /* The filtered column list */
  dmDescriptor   *parent  = NULL;

  char  *colname  = NULL;
  char  *rename   = NULL;
  char **entries  = NULL;
  char  *tcolname = NULL;
  char  *affix    = NULL;
  short *flags    = NULL;

  long no          = 0;
  long colno       = 0;
  long offset      = 0;
  long nraw        = 0;
  long ncols       = 0;
  long i           = 0;
  long EntryOffset = 0;
  long nmax        = 0;
  long n           = 0;
  long cptno       = 0;
  int  append      = 0;
  int  is_unit     = 0;
  int  state       = 1;


  /* We replace the raw column list with a filtered column list */
  nmax = 3 * block->nrcols; /* Allow for components */
  nraw = block->nrcols;
  ncols = nraw;
  cols = calloc( nmax, UT_SZ_P );
  flags= calloc( nmax, UT_SZ_S );

  for ( i = 0; i < ncols; i++ )
  {
    cols[i] = block->rawcols[i];
    flags[i] = DM_FLAG_VISIBLE;
  }

  if ( select )
  {
    if ( block->type == dmIMAGE )
    {
      sprintf( dm_errbuf, "DM column select syntax not valid with image.");
      dmu_err_set( DM_ERR_PARSE, dm_errbuf );
    }

    entries = cs_parse_comma_list( select, &n );

    for ( EntryOffset = 0; EntryOffset < n; EntryOffset++ )
    {
      if ( ncols > nmax )
      {
        nmax = 2 * nmax + 10;
        cols = realloc( cols, nmax * UT_SZ_P );
        flags=realloc( flags, nmax  * UT_SZ_I );
      }

      tcolname = entries[EntryOffset];
 
      if ( !tcolname )
        continue;

      while( *tcolname == ' ' )
        ++tcolname;

      colname = cs_dup( tcolname );
      rename = NULL;
      affix = strchr( colname, '=' );

      if ( affix )  /* We have a string  "time=sc_time" */
      {
        *affix = '\0'; /* Now we have two strings, "time" and "sc_time" */
        ++affix;
        rename = colname;
        colname = affix;
      }

      /* Copy remainder of columns to output */
      /* Thus  "time,pha,*"  maps "x,y,pha,det,time" to "time,pha,x,y,det" */

      /* Special case ALL */
      if (cs_eq(colname,"*") || cs_uceq( colname, "#all" ))
      {
        if ( !state )  
        {
          /* Copy remainder of columns to output; skip ones already in list */
          /* Thus  "time,pha,*"  maps "x,y,pha,det,time" to "time,pha,x,y,det" */ 

          for ( offset = 0; offset < nraw; offset++ )
          {
            dd = block->rawcols[offset];
            colno = dmi_descriptor_match( dd, cols, ncols );

            if ( dd->dim > 1 && colno == 0 && ncols > 0 ) 
            {
              for ( i=1; i <= dd->dim; i++ )
	      {
	        ddc = dmi_descriptor_get_cpt( dd, i );

	        if ( ddc && ( no = dmi_descriptor_match( ddc, cols, ncols )))
                {
	          if ( colno == 0 )
	          {
	            colno = no;
	            cols[no-1] = dd;
	          }
	          else
	          {
	            cols[no-1] = NULL;
	            flags[no-1] = 0;
	          }
	        }
	      }
            } // end if (dd->dim > 1...

            if ( colno == 0 )
            {
              flags[ncols] = DM_FLAG_VISIBLE;
              cols[ncols++] = dd;    
            }
          } // end for( offset...
        } // end if !state

        state = 0;
      }// end special case ALL...

      /* Special case NONE; not very useful */
      else
        if ( cs_eq( colname, "-" ) || cs_uceq( colname, "#none" ))
        {
          state = 0;
          ncols = 0;
        }
        /* Delete a column. Note this doesn't work for a cpt of a vector col */
        else
          if ( colname[0] == '-' || colname[0] == '!' )
          {
            dd = dmi_block_search_public_identifier( block, colname+1 );

            if ( dd && state )
            {
              no = dmi_descriptor_match( dd, cols, ncols );

              if ( no > 0 ) 
              {
                flags[no-1] = 0;
                cols[no-1]  = NULL;
              }
            }
          }
          /* Explicit column choice */
          else 
          {
            if ( state )
              ncols = 0;

            state = 0;
            dd = dmi_block_search_col_identifier( block, colname );

            if ( dd && dd->descType == dmCOORD )
            {
              append = 1;      

              if ( rename )
                dmi_descriptor_set_name( dd, rename );

              /* make sure the column and its components aren't duplicated in the
                 list */

              if ( ncols > 0 )
              {
                /* make sure column isn't already in list */
                colno = dmi_descriptor_match( dd, cols, ncols );

                /* if this is a vector remove its components from the list */
                if ( dd->dim > 1 && colno == 0 ) 
                {
                  for ( i=1; i <= dd->dim; i++ )
	          {
                    ddc = dmi_descriptor_get_cpt( dd, i );

                    if ( ddc && ( no = dmi_descriptor_match( ddc, cols, ncols )))
                    {
	              if ( colno == 0 )
	              {
	                colno = no;
	                cols[no-1] = dd;
	              }
	              else
	              {
	                cols[no-1] = NULL;
	                flags[no-1] = 0;
	              }
	            }
	          }
                } // end if ( dd->dim...

                if ( colno > 0 )
	          append = 0;

                /* also need to make sure column isn't one of the lists' components */
                if ( append )
                {
	          for ( i=1; i <= ncols; i++ )
	          {
	            ddp = cols[i-1];

                    if ( ddp && ddp->dim > 1 )
	            {
	               if ( dmi_descriptor_match( dd, ddp->cpts, ddp->dim ))
	               {
	                 append = 0;
	               }
	            }
                  }
                }
	      } // if (ncols > 0 )

              /* If two cols are specified in their vector order, replace them
	         with their vector parent */
              if ( dd->cptno > 1 && ncols > 0 )
              {
                prev_dd = cols[ncols-1];

                /* Is this the second component of a vector we just had the first cpt of? */
                if ( prev_dd && prev_dd->parent == dd->parent && prev_dd->cptno == 1 )
                {
                  /* Replace the previous entry in the list by its parent */ 
                  /* Only works if dim = 2 */

                  cols[ncols-1] = dd->parent;

                  /* Flags stays the same */
                  append = 0;
                }       
              }
 
              // Code Change  Bug 12880  January 10, 2012
              //  If unscaled descriptor, point parent to same buffer space

              is_unit = dmi_coord_transform_is_unit( dd );
              parent  = dmi_coord_get_parent ( dd);

              if ( is_unit && parent && ( dd->type == parent->type))
	      {
                // Unscaled column
                flags[ncols] |= DM_FLAG_RAW & parent->flag;
	        parent->bufno = dd->bufno;
              }
	      // else
	      //{
                // Scaled column, add parent column to buffer since 
                // it is needed to calculate result
                //if ( parent )
		//dmi_block_add_col( block, parent, DM_FLAG_OPERAND);
	      // }

             /* A new column or component */
              if ( append )
              {
                flags[ncols] |= DM_FLAG_VISIBLE; 
                cols[ncols++] = dd;
              }

            } // end if ( dd && dd->descType == dmCOORD )
            else
            {
              /* DD not found */
              sprintf( dm_errbuf, "Column select syntax: unknown column %s", colname );
              dmu_err_set( DM_ERR_PARSE, dm_errbuf );      
            }
          } // end else Explicit column choice

        if ( rename )
          free( rename );
        else
          if ( colname )
            free( colname );
      } /* end for loop offset */

      /* 
        QQ: WHat about  "[cols pha,foo=pha]"
        Do we get two copies? Right now that doesn't work...
        eventually want   [cols pha, foo=pha, goo= pha+pi]
      */
      ar_free_cp( entries, n );

  } // end if ( select )


  free( block->cols );
  block->cols = calloc( ncols, UT_SZ_P );
  block->ncols = 0;

#if DEBUG
printf( "NCOLS = %ld\n", ncols );
#endif

  /* Add all visible cols to buffer first */
  for ( i = 0; i < ncols; i++ )
  {
    if ( cols[i] )
      dmi_block_add_col( block, cols[i], flags[i] );
  }

  for ( i = 0; i < ncols; i++ )
  {
    if ( cols[i] )
    {
      dd = cols[i];

      if ( dd->descType == dmCOORD )
      {
        is_unit =dmi_coord_transform_is_unit( dd );
        parent = dmi_coord_get_parent( dd );

        if ( is_unit )
        {
          if ( parent )
            flags[i] |= DM_FLAG_RAW & parent->flag;
        }
        else 
        {
          /* Scaled column. Add parent column to buffer since we need it to calculate result */
          if ( parent )
            dmi_block_add_col( block, parent, DM_FLAG_OPERAND );
          else 
          {
            sprintf( dm_errbuf, "No coord parent found for %s", dmi_descriptor_get_name( dd ));
            dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
          }
 
          flags[i] |= DM_FLAG_RESULT;

          /* If dd is a vector, also flag its components as results so
           * dmGetVectorArrays() sees that they're in the buffer */
          if ( dd->dim > 1 )
          {
            for ( cptno = 0; cptno < dd->dim; cptno++ )
              dd->cpts[cptno]->flag |= DM_FLAG_RESULT;
          }
        }
      }
      
      /* Visible column is already in buffer but this updates its flags. */
      dmi_block_add_col( block, cols[i], flags[i] );
      block->cols[block->ncols++] = cols[i];
      dmi_descriptor_set_no( cols[i], block->ncols );
    }
  }   

  /* end if select */
  free( cols );
  free( flags );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_map_image_filter_bounds                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor   *image                           */
/*             dmFilter      **filter_list                     */
/*             long           *image_mins                      */
/*             long           *image_maxes                     */
/*             long            naxes                           */
/*             int             compact                         */
/*             int             reset                           */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            none                                             */
/*                                                             */
/***************************************************************/
void dmi_map_image_filter_bounds( dmDescriptor* image, dmFilter** filter_list,
                                  long* image_mins, long* image_maxes,
                                  long naxes, int compact, int reset )
{
  dmFilter      *filter = NULL;
  dmDescriptor  *colcpt = NULL;
  dmDescriptor  *colp   = NULL;

  double  bounds[2];
  double  axis_bounds[2];
  double  logical_bounds[2];
  double  cdelt[dmMAXDIM];
  double  crpix[dmMAXDIM];
  double  crval[dmMAXDIM];
  long    axis_offset = 0;
  long    groupno     = 0;
  long    axisno      = 0;
  long    vcpt        = 0;
  long    ng          = 0;
  long    i           = 0;

  /* 
   * Reset=1 does two things: it initializes the ranges, and it sets up the
   * new log-to-phys transform. 
   */

  if ( reset )
  {
    for ( i = 0; i < naxes; i++ )
    {
      image_mins[i] = 1;
      image_maxes[i] = image->axes[i];
    }
  }

  if ( !compact )
    return;

  /* Map the runtime filters to image groups */    
  /* Reset the image bounds */
  ng = image->ngroups;
  axisno = 1;


  for ( groupno = 1; groupno <= ng; groupno++ )
  {
    vcpt = 1;
    colp = dmi_descriptor_get_group_no( image, groupno );
    filter = filter_list[groupno-1];

    if ( filter )
    {
      if ( colp->dim != filter->dim )
      {
        sprintf( dm_errbuf, "DM interal error: Filter and descriptor dimensions do not match (%s)", colp->name );
        dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
      }

      dmi_coord_get_transform_d( colp, crpix, crval, cdelt, colp->dim );

      for ( vcpt = 1; vcpt <= filter->dim; vcpt++ )
      { 
        axis_offset =  ( axisno - 1 ) + ( vcpt - 1 );

        dmi_filter_get_cpt_bounds( filter, vcpt, bounds );

        colcpt = dmi_descriptor_get_cpt( colp, vcpt );
      
        /* bounds[1] == DT_REAL8_MAX means that there is no filter applied to
         * this component of the vector and we needn't adjust the image bounds 
         */

	if ( bounds[1] != DT_REAL8_MAX )
       	{
          /* bound the filter by the axis range */
             dmi_coord_get_range( colcpt, 0, DT_REAL8, &axis_bounds[0], &axis_bounds[1] );

       	     if((bounds[1]< axis_bounds[0])||(bounds[0] > axis_bounds[1]))
             {
//The image_mins and image_maxes are set to -1 so that dmi_parse_bin_filter() would set the no image
//is overlapped with the filter
			image_mins[axis_offset] = -1;
			image_maxes[axis_offset] = -1;		
			continue;
	     }

            if ( bounds[0] < axis_bounds[0] )
	       bounds[0] = axis_bounds[0];

            if ( bounds[1] > axis_bounds[1] )
	       bounds[1] = axis_bounds[1];

            dmi_coord_invert_d( colcpt, &bounds[0], &logical_bounds[0] );
            image_mins[axis_offset]  = (long)(logical_bounds[0]+0.5+DM_FILT_EPSILON);
            dmi_coord_invert_d( colcpt, &bounds[1], &logical_bounds[1] );
            image_maxes[axis_offset] = (long)(logical_bounds[1]+0.5-DM_FILT_EPSILON);
        }

        /* Now we go back the other way. Why? Our user filter in physical coords
         * may not have started on a pixel boundary. Actual image filtering must happen on pixel
         * boundaries, though. We just ensured this by rounding logical_bounds to get integer
         * logical pixel numbers; now we see
         * what that rounded value is in physical coords.
         * Suppose our image is 256 x 256 and our
         *  old transform was           x = 4000 + 32 ( #1 - 100.0 )
         * Then the old bounds were  logical(0.5:256.5) = physical(816.0:9008.0)  (corners of pixels)
         * If the filter is             x = 1000:2000
         * Then the logical bounds are  physical(1000.0:2000.0) = logical(6.25:37.5)
         * which is rounded to pixel numbers logical#(6:37) 
         * which in coords is           logical(5.5:37.5)   including the edges of the pixels
         * which maps back to           physical(976.0:2000.0) 
         * - that's as close as we can get to physical 1000:2000 in whole binned logical pixels -
         * The new mapping is
         *                        x = 976.0 + 32 ( #1' - 0.5 )
         * Note that there is a degeneracy in the choice of crpix and crval for these linear scalings.
         * We could choose to keep the old crpix and have the equivalent mapping
         *                        x = 4160.0 + 32 ( #1 - 100.0 )
         * but this puts crpix off the actual resulting image, which is a bit inelegant.
         * It's a matter of taste really.
         */

         logical_bounds[0] = image_mins[axis_offset]-0.5;   /* Left edge of first pixel */

         dmi_coord_calc_d( colcpt, &logical_bounds[0], &bounds[0] );

         crpix[vcpt-1] = 0.5;
         crval[vcpt-1] = bounds[0];     
      } // end for(vpct...) loop
 
      if ( reset )
      {
        dmi_coord_set_pix( colp, crpix, colp->dim );
        dmi_coord_set_val( colp, crval, colp->dim );
      }
    } // end if (filter)

    axisno += colp->dim;

  } // end for (groupno...) loop
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_get_binning_defaults                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock        *block                           */
/*             dmDescriptor   *col                             */
/*             long            retval                          */
/*                                                             */
/*  Output parameters:                                         */
/*             double         *vals                            */
/*                                                             */
/*  Returns:                                                   */
/*            none                                             */
/*                                                             */
/***************************************************************/
void dmi_get_binning_defaults( dmBlock* block, dmDescriptor* col,
                               long retval, double* vals )
{
  dmDescriptor *image = NULL;

  double delt   = 0.0;
  double pix    = 0.0;
  double val    = 0.0;
  long   ivals  = 0;
  long   vcpt   = 0;
  long   axisno = 0;
  long   pre_groupno = 0;
  long   groupOffset = 0;
  int    is_image    = 0;


  is_image = ( col && block && block->type == dmIMAGE );

  vals[0] = -DT_REAL8_MAX;
  vals[1] =  DT_REAL8_MAX;
  vals[2] = 0.0;

  if ( !block || !col )
  {
    if ( dmverbose >= 4 ) fprintf( stderr, "Using no-default mode\n" );
  }
  else
    if ( is_image && ( col->flag & DM_FLAG_LOGICAL ))
    {
       image = block->cols[0];
       vcpt = col->cptno;

       if ( vcpt == 0 )
         vcpt = 1;

       pre_groupno = dmi_coord_get_group_no( col )-1;
       axisno = 0;

       for ( groupOffset = 0; groupOffset < pre_groupno; groupOffset++ )
          axisno += image->groups[groupOffset]->dim;

       axisno += vcpt;

      /* Here we want to allow for image filters. Note the vals are a logical range here */
      vals[0] = 0.5;
      vals[1] = image->axes[axisno-1]+0.5;

      if ( dmverbose >=4 )
        fprintf( stderr, "Using axis length image mode\n" );

      vals[2] = 1.0;
    }   
    else
      if ( retval & DM_BINS_COMPACT )
      {   
        if ( dmverbose >=4 ) fprintf( stderr, "Using compact image mode\n" );
        dmi_column_extremes( block, col, vals );
      }
      else
      {
        if ( dmverbose >=4 )
          fprintf( stderr, "Using full image mode\n" );
        dmi_descriptor_get_cpt_range( col, 1, DT_REAL8, (void*)&vals[0], (void*)&vals[1] );
      }

      /* Round up integral TLMIN/MAX to full range */

      /* Modify TLMIN/TLMAX,
       * This is to defend against old files that put the subspace range
       * at the integer boundary when in fact they should be +0.5
       * and -0.5.  This bit of code should be removed when these old
       * files get fixed. */

      if ( ! is_image )
      {
        ivals = (long)vals[0];

        if ( (double)ivals == vals[0] )
          vals[0] -= 0.5;

        ivals = (long)vals[1];

        if ( (double)ivals == vals[1] )
          vals[1] += 0.5;
      }

      if ( !col )
        vals[2] = 1.0;
      else
        if ( is_image && !( col->flag & DM_FLAG_LOGICAL ))
        {
          dmi_coord_get_transform_d( col, &pix, &val, &delt, 1 );
          vals[2] = delt;
        }
        else
          dmi_descriptor_get_cast_bin( col, DT_REAL8, (void*)&vals[2] );
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_adjust_binning_edges                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             char    *name                                   */
/*             long     vcpt                                   */
/*             int      is_integer                             */
/*                                                             */
/*  Output parameters:                                         */
/*             double         *vals                            */
/*                                                             */
/*  Returns:                                                   */
/*            none                                             */
/*                                                             */
/***************************************************************/
void dmi_adjust_binning_edges( char* name, long vcpt, double* vals, int is_integer )
{
  long  ivals = 0;

  if ( vals[2] == 0.0 )
    vals[2] = 1.0;

  if ( is_integer )
  {
    ivals = (long)vals[0];

    if ( (double)ivals == vals[0] )
      vals[0] -= 0.5;

    ivals = (long)vals[1];

    if ( (double)ivals == vals[1] )
      vals[1] += 0.5;
   }

   if ( dmverbose >= 1 )
   {
     char cvals0[80];    
     char cvals1[80];
     char cvals2[80];
     dmu_format_number_d( vals[0], cvals0 );
     dmu_format_number_d( vals[1], cvals1 );
     dmu_format_number_d( vals[2], cvals2 );
     fprintf( stderr, "DM binning: %s %ld Min,Max,Step = %-20s : %-20s : %-20s\n",
                  name, vcpt, cvals0, cvals1, cvals2 );
   }
} 



/***************************************************************/
/*                                                             */
/* Function:   dmi_preparse_bin_spec                           */
/*                                                             */
/* Description:                                                */
/*  Support the syntax of trailing 'L' and 'C', 'F' letters to */
/*  control logarithmic bins and compact/full defaulting mode. */
/*                                                             */
/*  Input parameters:                                          */
/*             char    *inspec                                 */
/*             int      compact                                */
/*             char   **spec                                   */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            long retval                                      */
/*                                                             */
/***************************************************************/
long dmi_preparse_bin_spec( char* inspec, int compact, char** spec )
{
  long  retval = 0;
  char *ptr    = NULL;
  long  slen   = 0;
  char  c      = '\0';
 
  if ( inspec && *inspec )
  {
    ptr = cs_dup( inspec );
    slen = strlen( ptr );
  }

  *spec = ptr;

  while( slen > 0 && isalpha( ptr[slen-1] ))
  {
    c = ptr[slen-1];

    if ( c == 'L' || c == 'l' )
      retval |= DM_BINS_LOG;
    else
      if ( c == 'F' || c == 'f' )
        compact = 0;
      else
        if ( c == 'C' || c == 'c' )
          compact = 1;

    ptr[slen-1] = '\0';
    --slen;
  }

  if ( compact )
    retval |= DM_BINS_COMPACT;

 return retval;
}
 


/***************************************************************/
/*                                                             */
/* Function:   dmi_eval_bin_spec                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock      *block                             */
/*             dmDescriptor *col                               */
/*             char         *inspec                            */
/*             double      **vvals                             */
/*             long         *gdim                              */
/*             int           compact                           */
/*             int           is_integer                        */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            long retval                                      */
/*                                                             */
/***************************************************************/
long dmi_eval_bin_spec( dmBlock* block, dmDescriptor* col, char* inspec,
                        double** vvals, long* gdim, int compact, int is_integer )
{
  dmDescriptor *ccol = NULL;    

  char    defname[] = "None";
  char   *spec = NULL;
  char   *name = defname;
  double *vals = NULL;
  long    dim  = 1;
  long    vcpt = 0;
  int     image  = 0; 
  long    retval = 0;


  image = ( col && block && block->type == dmIMAGE );

  if ( col )
  {
    dim = col->dim;
    name = col->name;

    if ( !is_integer )
      is_integer = ! ( col->type == DT_REAL8 || col->type == DT_REAL4 );
  }  

  if ( dmverbose >= 4 ) fprintf( stderr, "Parse binning %s\n", inspec );

  retval = dmi_preparse_bin_spec( inspec, compact, &spec );
 
  *gdim = dim;
  ccol  = col;

  for ( vcpt = 0; vcpt < dim; vcpt++ )
  {
    if ( col && col->cpts )
      ccol = dmi_descriptor_get_cpt( col, vcpt+1 );

    vvals[vcpt] = calloc( 3, UT_SZ_D );
    vals = vvals[vcpt];
    dmi_get_binning_defaults( block, ccol, retval, vals );
    retval = dmi_parse_bin_spec( spec, vals, retval, is_integer );
    dmi_adjust_binning_edges( name, vcpt, vals, is_integer );
  }

  free( spec );
  return retval; 
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_binspec                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock      *block                             */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            dmBinSpec *binspec                               */
/*                                                             */
/***************************************************************/
dmBinSpec* dmi_block_get_binspec( dmBlock* block )
{
  dmBinSpec     *binspec = NULL;
  dmDescriptor  *dd      = NULL;

  if ( block->type != dmIMAGE || block->ncols == 0 )
    return NULL;

  dd = block->cols[0];

  if ( dd == block->rawcols[0] )  /* Binned images always have different cols and rawcols */
    return NULL;

  dd = dmi_descriptor_get_unscaled( dd );

  if ( dd )
    binspec = (dmBinSpec*)dd->aux;

  return binspec;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_binspec_free                                */
/*                                                             */
/* Description:                                                */
/*   Free a binspec. Note the freeing of the filtered image    */
/*   descriptor here.                                          */ 
/*   Cases:                                                    */
/*     Image, no filter:    no binspec.  cols[0] = rawcols[0]. */
/*     Image, filter but no bin:  binspec present. Need to free*/
/*                                rawcols[0].                  */
/*     Image, filter and bin: binspec present. Need to free    */
/*                      rawcols[0] and binspec->filtered_image.*/
/*     Table, bin:  binspec present. Need to free all rawcols. */
/*     No filtered image.                                      */
/*                                                             */
/*   If there is a binspec, it is aux for logical-cols[0].     */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBinSpec  *binspec                             */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            none                                             */
/*                                                             */
/***************************************************************/
void dmi_binspec_free( dmBinSpec* binspec )
{
 long  i = 0;
 long  k = 0;
 dmDescriptor *dd = NULL;
 dmBinAxis    *axis = NULL;

/* 
 *  Free a binning axis. Don't do a free(axis) because the
 *  space is allocated as a single array.
 */

  for ( i = 0; i < binspec->n; i++ )
  {
    axis= &binspec->axis[i];

    if (axis )
    {
      dd = axis->dd;

      for ( k = 0; k < dd->dim; k++ )
        free( axis->vals[k] );

      free( axis->vals );
    }
  }

  free( binspec->axis );

  if ( binspec->is_binned )
    dmi_descriptor_kernel_free( binspec->filtered_image, NULL );

  free( binspec->offsets );
  free( binspec->weights );
  free( binspec->vcpts );
  free( binspec->groups );
  free( binspec->axis_map );
  free( binspec->image_mins );
  free( binspec->image_maxes );
  free( binspec->min );
  free( binspec->max );
  free( binspec->axes );
  free( binspec->start );
  free( binspec->step );
  free( binspec );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_axis_steps                                  */
/*                                                             */
/* Description:                                                */
/*     Number of pixels given a linear binning.                */
/*     No support yet for other types of binning.              */
/*                                                             */
/*  Input parameters:                                          */
/*             double  *vals                                   */
/*             long      type                                  */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            long  nsteps                                     */
/*                                                             */
/***************************************************************/
long dmi_axis_steps( double* vals, long type )
{
  double steps  = 0.0;
  long   nsteps = 0;

  if ( vals[2] == 0.0 )
    return 1;

  if ( type == DM_BINS_NONE )
    return 0;  

  /* calculate the number of steps */
  steps = ( vals[1] - vals[0] ) / vals[2];

  /* make sure its positive */
  if ( steps < 0 )
    steps = -steps;

  /* convert to integer */
  nsteps = steps;

  if ( type & DM_BINS_N )
  {
    /* ENH 10227:                                        */
    /* number of bins were specified, so this should be  */
    /* VERY close to an integer                          */
    if ( (steps - (double)nsteps) > FLT_EPSILON )
      ++nsteps;
  }
  else
  {
    if ( steps > (double)nsteps )
      ++nsteps;
  }

  /* IMPLEMENT (other) binning types */

  return nsteps;
}



/**************************************************************************************************/
/*                                                                                                */
/* Function:   dmi_parse_bin_filter                                                               */
/*                                                                                                */
/* Description:                                                                                   */
/*                                                                                                */
/* Rules for making image grid:                                                                   */
/*                                                                                                */
/* Type           Example                             Grid                                        */
/* ==========================================================================================     */
/* rebin explicit           im.fits[x=...][bin x=1:100:2,y=1:100:2]    From bin                   */
/* rebin implicit           im.fits[x=...][bin sky=2]                  From filter                */
/* image filter             im.fits[x=...]                             From filter                */
/* image region             im.fits[circle()]                          From filter full or compact*/
/* table implicit           ev.fits[circle()][bin sky]                 From filter full or compact*/
/* table explicit           ev.fits[circle()][bin x=1:100:2,...]       From bin                   */
/* image slice              im.fits[z=1][bin x,y]                      From bin/filter            */
/* So the grid is determined by the filter unless there is an explicit                            */
/* bin. In particular, bin in compact image mode looks at the filter bounds.                      */
/*                                                                                                */
/*                                                                                                */
/* When rebinning image, world coords stay the same (but may have different number of axes)       */
/* phys coords may change transform and dimensionality.                                           */
/* When binning table, must make new coord descriptors if axis list is changed.                   */
/* When just filtering image, phys coords change transform only.                                  */
/*                                                                                                */
/*                                                                                                */
/*  Input parameters:                                                                             */
/*             dmBlock    *block                                                                  */
/*             const char *bin_spec                                                               */
/*                                                                                                */
/*  Output parameters:                                                                            */
/*             none                                                                               */
/*                                                                                                */
/*  Returns:                                                                                      */
/*            dmDescriptor *image                                                                 */
/*                                                                                                */
/**************************************************************************************************/
dmDescriptor* dmi_parse_bin_filter( dmBlock* block, const char* bin_spec )
{
  dmDescriptor  *phys_group  = NULL;
  dmDescriptor  *world_group = NULL;
  dmDescriptor  *col_coord   = NULL;
  dmDescriptor  *col         = NULL;
  dmDescriptor  *image       = NULL;
  dmDescriptor  *logical_image = NULL;
  dmDescriptor  *cpt_dd      = NULL;
  dmDescriptor  *dd          = NULL;
  dmDescriptor  *phys_col    = NULL;
  dmDescriptor  *old_col     = NULL;
  dmDescriptor  *raw_image   = NULL;
  dmDescriptor  *logical_raw_image = NULL;
  dmDescriptor **groups      = NULL;
  dmDataType     type        = DT_INT2;
  dmElementType  etype       =  dmVALUE;
  dmFilter     **filter_list = NULL;
  dmDataType     scaled_type;
  dmDataType     image_bin_type;
  dmBool         is_scaled   = UT_FALSE;
  dmBool         is_image_cast = UT_FALSE;
  dmBool         bIsNull     = UT_FALSE;
  dmBinSpec     *binspec     = NULL;
  dmBinAxis     *binaxis     = NULL;
  dmSubspace    *rts         = NULL;
  dmBuffer      *buffer      = NULL;

  double    logical_mins[dmMAXDIM];
  double    logical_maxes[dmMAXDIM];
  double    physical_mins[dmMAXDIM];
  double    physical_maxes[dmMAXDIM];
  double    cdelt[dmMAXDIM];
  double    crpix[dmMAXDIM];
  double    crval[dmMAXDIM];
  double  **valsp = NULL;
  double   *vals = NULL;
  double    tzero =0.0;
  double    pix = 0.0;
  double    tscal = 1.0;
  double    scale = 0.0;

  char *ptr = cs_dup( bin_spec );
  char *spec = NULL;
  char *item = NULL;
  char *weight_spec = NULL;
  char *unit = NULL;
  char *desc = NULL;
  char *scaledUnit = NULL;
  char  imagename[1024] = "IMAGE";

  long *axes  = NULL;
  long  no    = 1;
  long  n     = 0;
  long  imdim = 1; /* Vector dimension of resulting image pixels is always 1 */
  long  nmax  = 0;
  long  cdim  = 1;
  long  vcpt  = 0;
  long  naxes = 0;
  long  dim   = 1;
  long  flag  = 0;
  long  i     = 0;
  long  nvals = 0;
  long  size  = 0;
  long  maxmeg  = 0;
  long  groupno = 0;
  long  coordno = 0;
  long  ngroups = 0;
  long  cptflag = 0;
  long  total_size = 0;

  int  compact = 1;
  int  is_binned = 0;
  int  force_integer = 0;
  int  is_logical = UT_FALSE;
  int  is_image = 0;


  is_binned = !( cs_is_blank( bin_spec ));
  is_image  = (block->type == dmIMAGE );
  buffer    = block->buffer;

  if ( buffer )
    rts = buffer->subspace;

  compact = dmi_block_filter_is_compact( block );

  binspec = calloc( 1, sizeof( dmBinSpec ));
  binspec->auxtype = DM_AUX_BIN;
  binspec->compact = compact;
  binspec->is_binned = is_binned;
  image_bin_type = block->image_bin_type;

  if ( is_image )
  {
    raw_image = block->rawcols[0];
    logical_raw_image = dmi_descriptor_get_unscaled( raw_image );

    if ( image_bin_type != DT_UNK && logical_raw_image->type != image_bin_type )
      is_image_cast = UT_TRUE;

    logical_image = dmi_descriptor_copy( block, NULL, logical_raw_image );
    image = dmi_descriptor_get_scaled( logical_image );
    dmi_coord_get_transform_d( image, &pix, &tzero, &tscal, 1 );

    // Code Change  Bug 11271  September 28, 2011
    //  get scaled unit value for later use in creating image descriptor
    if ( image->unit )
      scaledUnit = cs_dup( image->unit);

    if ( tscal != 1.0 || tzero != 0.0 )
      is_scaled = UT_TRUE;

    binspec->filtered_image = image;  
    binspec->image_mins = calloc( image->naxes, UT_SZ_I );
    binspec->image_maxes = calloc( image->naxes, UT_SZ_I );

    filter_list = dmi_map_filters_to_image_groups( raw_image, rts, compact );
    dmi_map_image_filter_bounds( image, filter_list, binspec->image_mins, binspec->image_maxes, image->naxes, compact, UT_TRUE );
    free( filter_list );
  }

  if ( image )
  {
    for ( i = 0; i < raw_image->naxes; i++ )
    {
// null image is indicated by the check with the filter bounds returned fromm dmi_map_image_filter_bounds()
	if(binspec->image_maxes[i] == -1 && binspec->image_mins[i] == -1)
	{
		bIsNull = UT_TRUE;
		break;
	}
    }

    if(!bIsNull)
    {
    	for ( i = 0; i < raw_image->naxes; i++ )
    	{
      		image->axes[i] = binspec->image_maxes[i] - binspec->image_mins[i] + 1;
      		logical_image->axes[i] = image->axes[i];
	}
    }
    else
    {
//filtered image is NULL image
        for ( i = 0; i < raw_image->naxes; i++ )
        {
		image->axes[i] = 0;
		logical_image->axes[i] = 0;
	}
	image->naxes = 0;
	image->size = 0;
    }  

    dmi_block_set_type( block, dmIMAGE );
    free( block->cols );
    block->ncols = 1;
    block->cols = calloc( 1, UT_SZ_P );
    block->cols[0] = image;
  }

  /* If it's not binned, the number of out axes is the number of in axes */
  if ( !is_binned )
  {
    image->flag &= ~DM_FLAG_UNBUF;
    flag = DM_FLAG_FILT | DM_FLAG_BINNED; /* Same as original image: we don't mention that it's filtered */

    if ( is_image_cast )
    {
      dmi_descriptor_convert_type( logical_image, image_bin_type );

      if ( is_scaled )
        scaled_type = (( dm_global & DM_GLOBAL_OLD ) && is_image ) ? DT_REAL4 : DT_REAL8;
      else
        scaled_type = image_bin_type;

      dmi_descriptor_convert_type( image, image_bin_type );
    }

    dmi_block_add_col( block, image, flag );
    dmi_descriptor_set_aux( logical_image, binspec );
    return image;
  }

  /* DM default binned-image descriptor name and block name */
  if ( !is_binned )
  { 
    cs_copy( raw_image->name, imagename );
  }
  else
    if ( !cs_ss( block->name, "_IMAGE" ))
    {
      sprintf( imagename, "%s_IMAGE", block->name );
      dmi_block_set_name( block, imagename );
    }
    else
    {
      cs_copy( block->name, imagename );
    }


  if ( is_binned )
  {
    n = dmu_count_commas( bin_spec );

    /* We'll be conservative and allocate the index arrays twice as big
       as they need to be, in case a later syntax change makes it possible
       for there to be more groups than entries in the spec.
    */

    nmax = n * 2;
    binspec->axis = calloc( nmax, sizeof( dmBinAxis ));
    groups = calloc( nmax, UT_SZ_P );   /* List of phys axis group descriptors */
    axes = calloc( nmax, UT_SZ_I );     /* Lengths of axes */
    binaxis = NULL;
    old_col = NULL;


    /* If there's a weight spec (e.g. [bin ...;pha]), parse it and store the
     * weight-column descriptor */
    col = dmi_parse_bin_weight( block, ptr, &weight_spec );

    if ( !weight_spec )
      binspec->weight_dd = NULL;
    else
    {
      if ( !col )
        return NULL;
      binspec->weight_dd = col;
    }

    while( ptr && *ptr && ngroups < n )
    {
      /*
       *  Read up to next comma.  Syntax example:   [bin x=4,y=4]
       */  
      ptr = dmu_parse_token( ptr, ',', &item, 0 );

      // Code Change  Bug 12596 (revisited)  April 15, 2011
      //  Check var 'item' for null - meaning bad input string
      //  If null, continue. Var 'ptr' already incremented by function

      if ( !item )
        continue;

      col = dmi_parse_bin_token( block, item, ngroups, &spec );

      if ( !col )
        return NULL;

      is_logical = ( col->flag & DM_FLAG_LOGICAL ) != 0;
      force_integer = is_logical;

      /*
       * Care: when binning [bin x,y] need to pick up the fact that it's
       * really [bin sky]. When we get to y, we reuse the x binaxis and make it sky.
       * This is why we can't set ngroups right away.
       */

       if ( binaxis ) 
         old_col = binaxis->dd;

       if ( col->cptno > 1 && binaxis &&
         old_col->cptno == col->cptno - 1 && old_col->parent == col->parent )
       {
         dim = dmi_descriptor_get_vector_dim( col->parent );
         binaxis->vals = (double**)realloc( binaxis->vals, dim * UT_SZ_P );
         cdim = 1;
         vcpt = col->cptno;
         valsp = (double**)&binaxis->vals[vcpt-1];
         binaxis->type = dmi_eval_bin_spec( block, col, spec,  valsp, 
                                     &cdim, compact, force_integer );
         col = col->parent; /* Must come after line above */
       }
       else
       {
         dim = dmi_descriptor_get_vector_dim( col );
         binaxis = &binspec->axis[ngroups++];
         binaxis->vals = (double**)calloc( dim, UT_SZ_P );   

         /*
           Parse the spec, applying defaults where needed.
           Note: dim is not changed when col is non null (this arg is used elsewhere
           when calling routine with col=NULL)
        */

         binaxis->type = dmi_eval_bin_spec( block, col, spec, (double**)binaxis->vals, 
                                            &dim, compact, force_integer );
         cdim = dim;
       }

       binaxis->dd = col;
       binspec->n = ngroups;
       free( item );

    } // end while( ptr &...

    free( ptr );

    /* Now we loop again, having grouped vector binning axes as required */

    for ( groupno = 1; groupno <= ngroups; groupno++ ) 
    {
      binaxis = &binspec->axis[groupno-1];
      col = binaxis->dd;  
      dim = col->vdim;


      /* if rebinning align the bin spec */

      if ( is_image )
      {
        if ( ! is_logical )
        {
          dmi_coord_get_transform_d( col, crpix, crval, cdelt, dim );

          /* force the bin range to the pixel boundaries for image rebinning
           * do this by performing the transformation forwards to an integer 
           * then back again to a double 
           */

          for ( vcpt = 1; vcpt <= dim; vcpt++ )
          {
            physical_mins[vcpt-1] = binaxis->vals[vcpt-1][0];
            physical_maxes[vcpt-1] = binaxis->vals[vcpt-1][1];
          }

          dmi_coord_invert_d( col, &physical_mins[0], &logical_mins[0] );
          dmi_coord_invert_d( col, &physical_maxes[0], &logical_maxes[0] );

          for ( vcpt = 1; vcpt <= dim; vcpt++ )
          {
            logical_mins[vcpt-1] = ((long)(logical_mins[vcpt-1]+0.5+FLT_EPSILON))-0.5;
            logical_maxes[vcpt-1] = ((long)(logical_maxes[vcpt-1]+0.5-FLT_EPSILON))+0.5;
          }

          dmi_coord_calc_d( col, &logical_mins[0], &physical_mins[0] );
          dmi_coord_calc_d( col, &logical_maxes[0], &physical_maxes[0] );

          for ( vcpt = 1; vcpt <= dim; vcpt++ )
          {
            binaxis->vals[vcpt-1][0] = physical_mins[vcpt-1];
            binaxis->vals[vcpt-1][1] = physical_maxes[vcpt-1];
          }
        }

        /* make sure that new pixels are integral size of old pixels */
        for ( vcpt = 1; vcpt <= dim; vcpt++ )
        {
          vals = binaxis->vals[vcpt-1];

          if ( is_logical )
            cdelt[vcpt-1] = 1.0;

          scale = binaxis->vals[vcpt-1][2] / cdelt[vcpt-1];

          if ( fabs( scale - floor(scale) ) > FLT_EPSILON )
          {
            sprintf( dm_errbuf, "Error: Bin step (%g) is not a multiple of pixel size (%g).  Try dmregrid.", binaxis->vals[vcpt-1][2], cdelt[vcpt-1] );
            dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
          }
        }
      } // end if (image)...

  
      /* Set the physical col descriptor and default coord transform */
      if ( is_logical )
      {
        phys_col = dmi_descriptor_get_scaled( col );
        dmi_coord_get_transform_d( phys_col, crpix, crval, cdelt, dim );
      }
      else
      {
        phys_col = col;

        for ( vcpt = 1; vcpt <= dim; vcpt ++ )
        {
          crpix[vcpt-1] = 0.5;
          crval[vcpt-1] = 0.5;
          cdelt[vcpt-1] = 1.0;
        }
      }


      /* Check the axes array is big enough - just in case */
      dim = col->vdim;

      if ( naxes + dim > nmax )
      {
        nmax += dim;
        axes = realloc( axes, nmax * UT_SZ_I );
      }

     /* Calculate the physical transform and the axis lengths. */
     for ( vcpt = 1; vcpt <= dim; vcpt ++ )
     {
        vals = binaxis->vals[vcpt-1];

          if ( vals )
          {
            crval[vcpt-1] = crval[vcpt-1] +  cdelt[vcpt-1] * ( vals[0] - crpix[vcpt-1] );
            crpix[vcpt-1] = 0.5;
            cdelt[vcpt-1] = vals[2] * cdelt[vcpt-1];

            /* The bin spec doesn't have proper limits */
            if ( vals[0] == -DT_REAL8_MAX || vals[1] == DT_REAL8_MAX )
            {
              dmu_err_set( DM_ERR_PARSE, "Error: Please supply bin range, file has no default." );
              return NULL;
            }

            axes[naxes+vcpt-1] =  dmi_axis_steps( vals, binaxis->type );

            if ( axes[naxes+vcpt-1] < 0 )
            {
              dmu_err_set( DM_ERR_PARSE, "Error: Invalid bin range would cause negative axis length." );
            }
          }
          else
          {
            sprintf( dm_errbuf, "Error parsing binning spec %s", bin_spec );
            dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
          } 
       }

       naxes += dim; 

      /* Allocate the coordinates that will hang off the image */
      /* 
       * The physical axis group is the binning column.
       * The binning spec gives the logical-to-physical coordinate system too.

       * The world coord on the binning column must be transferred too.
       */

      flag = DM_FLAG_PHYS | DM_FLAG_AXIS | DM_FLAG_COORD | ( DM_FLAG_TEMP & phys_col->flag );

      cptflag = dim > 1 ?  DM_FLAG_CPT : 0;

      phys_group = dmi_descriptor_alloc( block, dmCOORD, groupno, phys_col->name, DT_REAL8,
                                         phys_col->unit, phys_col->desc, phys_col->vdim,
                                         dmVALUE, phys_col->naxes, phys_col->axes, flag );
 
      dmi_coord_transform_alloc( phys_group, "LINEAR", crval, crpix, cdelt, dim );

      /* Copy components */
      if ( cptflag )
      {
        for ( vcpt = 1; vcpt <= dim; vcpt++ )
        {
          cpt_dd = dmi_descriptor_get_cpt( phys_col, vcpt );
          dd = dmi_descriptor_alloc( block, dmCOORD, vcpt, cpt_dd->name, DT_REAL8, cpt_dd->unit,
                   cpt_dd->desc, cpt_dd->vdim, dmVALUE, cpt_dd->naxes, cpt_dd->axes, flag | cptflag );
          dmi_descriptor_set_cpt( phys_group, vcpt, dd );
          dmi_coord_transform_alloc( dd, "LINEAR", &crval[vcpt-1], &crpix[vcpt-1], &cdelt[vcpt-1], 1 );
        }
      }

      /* Copy world coords */
      for ( coordno = 1; coordno <= phys_col->ncoords; coordno++ )
      {
        col_coord = phys_col->coords[coordno-1];
        world_group = dmi_descriptor_copy( block, NULL, col_coord );
        flag = DM_FLAG_AXIS | DM_FLAG_COORD;      
        world_group->flag = flag;    

        for ( vcpt = 1; vcpt <= dim; vcpt++ )
        {
          dd = dmi_descriptor_get_cpt( world_group, vcpt );
          dd->flag = flag | cptflag;
        }

        dmi_descriptor_add_coord( phys_group, world_group );
      }

      groups[groupno-1] = phys_group;


      /* 
         Mark the column for use in the buffer (so that its values
         will be read during the computation). Only if we are binning a table.
       */

      if ( block->type != dmIMAGE )
        dmi_block_add_col( block, col, DM_FLAG_OPERAND | DM_FLAG_BIN );

    } // end for (groupno...
  } // end if (is_binned)


  /* Make image descriptor */

  flag = DM_FLAG_IMG | DM_FLAG_BINNED | DM_FLAG_RESULT;

  if ( image_bin_type != DT_UNK )
  {
    /* Type set by [opt type=...] */
    type = image_bin_type;
  }
  else
    if ( binspec->weight_dd )
    {
      /* Weighted bin; use type of weight col
       *
       * We can only create short, long, float, and double images,
       * so find the closest match */

      switch ( binspec->weight_dd->type )
      {
        case DT_REAL8:
          type = DT_REAL8;
          break;
        case DT_REAL4:
          type = DT_REAL4;
          break;
        case DT_INT4:
        case DT_UINT4:
          type = DT_INT4;
          break;
        default:
          type = DT_INT2;
          break;
      }

      // Code Change  Bug 11271  September 28, 2011
      //  assign scaled unit value in weight_dd for use in creating image descriptor
      if ( scaledUnit )
        free (scaledUnit);

      scaledUnit = cs_dup( binspec->weight_dd->unit );
     }
    else
      if ( block->type == dmIMAGE )
      {
        /* Image rebin; use existing image type */
        type = logical_image->type;
      }
      else
      {
         /* Default */
         type = DT_INT2;
      }

  logical_image = dmi_descriptor_alloc( block, dmCOLUMN, no, "#COL1", type, unit, 
                            desc, imdim, etype, naxes, axes, flag | DM_FLAG_LOGICAL ); 

  logical_image->flag = flag | DM_FLAG_LOGICAL;

  if ( is_scaled )
    scaled_type = (( dm_global & DM_GLOBAL_OLD ) && is_image ) ? DT_REAL4 : DT_REAL8;
  else
    scaled_type = type;

  image = dmi_create_linear_wcs( block, logical_image, imagename, scaled_type, tzero, pix, tscal );
  image->flag = flag | DM_FLAG_COORD | DM_FLAG_PHYS;

  // Code Change  Bug 11271  September 28, 2011
  //   Set unit value for new image descriptor
  //   This variable may be null or may contain units 
  //   from input image or weighted column

  if ( scaledUnit )
  {
    image->unit = cs_dup( scaledUnit );
    free (scaledUnit);
  }

  /* Attach the bin spec to the logical image, so it knows how to compute itself */
  dmi_descriptor_set_aux( logical_image, binspec );

  /* 
     Mark the image for the buffer - this means that the image memory
     will be allocated and the image will appear in buffer lists;
     we make it a visible column in the output block elsewhere
     (in block open routine). Right now the virtual block columns are still
     the table columns.
   */

  dmi_block_add_col( block, image, flag );
  nvals = dmi_descriptor_get_nvals( image );
  size = dmi_descriptor_get_bytes( image );
  maxmeg = dmi_block_get_maxmeg( block );
  total_size = nvals * size;

  if ( total_size > maxmeg * 1000000 )
  {
    if ( dmu_err_get() == dmSUCCESS )
    {
      size = (long)(((double)( nvals * size )) / (double)1000000 + 0.5);
      sprintf( dm_errbuf, "WARNING: Creating large image: %ld MB. Current max set at %ld MB.\nIncrease maximum using [opt mem=n] or increase blocking to reduce size.", size, maxmeg );
      dmu_err_set( DM_ERR_MEM, dm_errbuf );
    }
  }

#if 0
  /* FIXME: removed this as a temporary fix for bug #9021 */
  else
    if ( total_size < 0 )
    {
      sprintf( dm_errbuf, "Error: Image too large.  Size wraps type long." );
      dmu_err_set( DM_ERR_PARSE, dm_errbuf );
    }
#endif

  /* Attach the coordinate systems to the image */
  dmi_descriptor_set_groups( image, groups, ngroups );
  free( groups );
  free( axes );

  /* Overwrite previous filtered image descriptor - should free it really, or cache it */
  dmi_block_set_type( block, dmIMAGE );


  free( block->cols );
  block->ncols = 1;
  block->cols = calloc( 1, UT_SZ_P );
  block->cols[0] = image;

  /* Late parse of "opt null=val" directive */
  if ( block->null )
    dmi_block_set_image_null( block, block->null );

  return image;
}






/***************************************************************/
/*                                                             */
/* Function:   dmi_parse_subspace_line                         */
/*                                                             */
/* Description:                                                */
/*         taken from dmi_parse_filter_line()                  */
/*         Parses subspace command strings and then takes      */
/*         appropriate actions                                 */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock    *block                          */
/*                  dmSubspace *pSubspace                       */
/*                  char       *line                           */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                  nothing                                    */
/*                                                             */
/***************************************************************/
void dmi_parse_subspace_line( dmBlock *block, dmSubspace *pSubspace,
                              char *line)
{
  char  *pInString    = line;
  char  *pToken       = NULL;
  char  *pNext        = NULL;
  char  *pTokenName   = NULL;

  long   minRange        = 1;

  /* if no editing commands, return */
  if ( !line || *line == '\0' ) 
    return;

  do
  {
    /* get the first segment, everything up to the first comma */
    pNext = dmu_parse_token ( pInString, ',', &pToken, 0 );

    if ( !pToken )
      break;

    if (*pToken == '-')
    {
      /* Remove/Reset subspace directive */
       
      /* name starts after the dash, increment pointer */
      pTokenName = pToken + 1;

      /* remove any blanks between dash and name */
      while (*pTokenName == ' ')
	pTokenName += 1;

      /* is there really a name here? */
      if (*pTokenName == '\0')
	break;

      /* reset subspace for column */
      dmi_subspace_col_reset(block, pSubspace, pTokenName, minRange, pSubspace->ncpts);

    }
    else
    {
      /* syntax error.. only reset is supported at present. */
      printf("Incorrect subspace edit syntax.. ignoring  %s\n", pToken);

    }

    pInString = pNext;

  } while (pInString != NULL && *pInString != '\0' );
}


