/*                                                                
**  Copyright (C) 2002-2012  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmreglib.c                                                  */
/*                                                               */
/*   Functions working iwth filter regions                       */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"

#define MAXLS 81
#define DCSIZE 80
#define XDEBUG 1

static void dmi_image_filter_regions_2d( dmDescriptor* raw_image, long* v_llpix, long* v_urpix, regRegion* region, int invert, void* null, void* values );
static void  dmi_image_filter_regions_gen( dmDescriptor* image, dmDescriptor* raw_image, long xaxis_offset, long yaxis_offset, long* v_llpix, long* v_urpix, regRegion* region, int invert, void* null, void* values );
void dmpRegionArea( dmDescriptor* dd, regRegion* region, double* rrarea, double* rfarea, int verbose );

typedef struct
{
 long offset;
 long xmin;
 long xmax;
 long ymin;
 long ymax;
 long xpix;
 long ypix;
 regRegion* region;
 double xregoffset;
 double yregoffset;
 int invert;
} dmRegionIterator;


static void dmRegionIterateInit( dmBlock* block, dmRegionIterator* iter,
  long xmin, long ymin, long xmax, long ymax, regRegion* region, int invert );
static long dmNextOffset( long offset, long xmin, long ymin, long xmax, long ymax, long* xpix, long* ypix,
 regRegion* region, int invert );
static long dmi_region_iterate_2d( dmRegionIterator* iter );
static dmBool dmNextPosition( dmRegionIterator* iter );
dmBool dmi_region_iterate_gen( dmRegionIterator* iter, long* xpix, long* ypix );

/* From filterlib2.c */
int dmi_coord_fix_region_filter( dmDescriptor* dd, dmFilter* filter, int is_image, long ty );

/***************************************************************/
/*                                                             */
/* Function:  dmRegionIterateInit                              */
/*                                                             */
/* Description:                                                */
/*   Function to clarify the region iteration code             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     dmBlock          *block                                 */
/*     long              xmin                                  */
/*     long              xmax                                  */
/*     long              ymin                                  */
/*     long              ymax                                  */
/*     regRegion        *region                                */
/*     int               invert                                */
/*                                                             */
/*  Output parameters:                                         */
/*     dmRegionIterator *iter                                  */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmRegionIterateInit( dmBlock *block, dmRegionIterator* iter,
                          long xmin, long ymin, long xmax, long ymax,
                          regRegion* region, int invert )
{
  double field[2]      = { 0.5, DT_REAL8_MAX };
  double xregbounds[2] = { 0.5, DT_REAL8_MAX };
  double yregbounds[2] = { 0.5, DT_REAL8_MAX };

  if ( !iter )
    return;

  iter->offset = 0;
  iter->xmin = ( ( xmin > 0 ) ? xmin : 1 );
  iter->xmax = xmax;
  iter->ymin = ( ( ymin > 0 ) ? ymin : 1 );
  iter->ymax = ymax;
  iter->xpix = iter->xmin;
  iter->ypix = iter->ymin;

  iter->region = region;

  if ( block->compact )
    regExtent( region, field, field, xregbounds, yregbounds );

  iter->xregoffset = xregbounds[0];
  iter->yregoffset = yregbounds[0];

  iter->invert = invert;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_region_iterate_2d                            */
/*                                                             */
/* Description:                                                */
/*   Iterate over the pixels in a 2D image that fail a region  */
/*   filter. Returns an offset in the pixel data to each pixel */
/*   that fails the filter or a value < 0 when finished.       */
/*                                                             */
/*  Input parameters:                                          */
/*     dmRegionIterator *iter                                  */
/*                                                             */
/*  Output parameters:                                         */
/*      none                                                   */
/*                                                             */
/*  Returns:                                                   */
/*      long   offset                                          */
/*                                                             */
/***************************************************************/
long dmi_region_iterate_2d( dmRegionIterator* iter )
{  
  long offset = 0;

  iter->offset = dmNextOffset( iter->offset,
                   iter->xmin, iter->ymin,
                   iter->xmax, iter->ymax,
                   &iter->xpix, &iter->ypix,
                   iter->region, iter->invert );
  offset = iter->offset++;
  return offset;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_region_iterate_2d                            */
/*                                                             */
/* Description:                                                */
/*   Iterate over the pixels in a N-dim image that fail a      */
/*   region filter.  Returns the 2D position of each failing   */
/*   pixel in xpix and ypix.  Returns UT_TRUE if there is more */
/*   image to filter. Returns UT_FALSE when we are finished    */
/*   iterating over the image.                                 */
/*                                                             */
/*  Input parameters:                                          */
/*     dmRegionIterator *iter                                  */
/*                                                             */
/*  Output parameters:                                         */
/*     long             *xpix                                  */
/*     long             *ypix                                  */
/*                                                             */
/*  Returns:                                                   */
/*      dmBool   dmSUCCESS or dmFAILURE                        */
/*                                                             */
/***************************************************************/
dmBool dmi_region_iterate_gen( dmRegionIterator* iter, long* xpix, long* ypix )
{
  dmBool retval = dmSUCCESS;

  retval = dmNextPosition( iter );
  *xpix = iter->xpix;
  *ypix = iter->ypix;
  iter->xpix++;
  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:  dmNextOffset                                     */
/*                                                             */
/* Description:                                                */
/*   Routine to find the next pixel NOT in the region.         */
/*   Returns the offset to the next pixel or < 0 if            */
/*   there are no more pixels.                                 */
/*                                                             */
/*  Input parameters:                                          */
/*     long       offset                                       */
/*     long       xmin                                         */
/*     long       ymin                                         */
/*     long       xmax                                         */
/*     long       ymax                                         */
/*     long      *xpix                                         */
/*     long      *ypix                                         */
/*     regRegion *region                                       */
/*     long       invert                                       */
/*                                                             */
/*  Output parameters:                                         */
/*      none                                                   */
/*                                                             */
/*  Returns:                                                   */
/*      long  offset                                           */
/*                                                             */
/***************************************************************/
long dmNextOffset( long offset, long xmin, long ymin, long xmax, long ymax, long* xpix, long* ypix,
 regRegion* region, int invert )
{
  double xpos = 0.0;
  double ypos = ymin;  /* Initialization to avoid compiler warning - ymin not used yet */
  int    accept = 0;

  while( *ypix <= ymax )
  {
    ypos = (double) *ypix;

    while( *xpix <= xmax )
    {
      xpos = (double) *xpix;  
      (*xpix)++;
    
      accept = regInsideRegion( region, xpos, ypos );

#if DEBUG
{ 
 static int flag = 1;
 if (flag < 10 )
 {
   printf( "LOGICAL RANGE REGFILT %ld,%ld : %ld,%ld\n", xmin, xmax, ymin, ymax );
   regComposeRegion( region, dm_errbuf, 1024 );
   printf( "LOGICAL RANGE REGFILT %f,%f in %s = %d\n", xpos, ypos, dm_errbuf, accept );
   flag++;
 }
}
#endif

      if ( invert )
      {
        if ( accept )
          return offset;
      }
      else
      {
        if ( !accept )
          return offset;
      }
      ++offset;
    }

    *xpix = xmin;
    (*ypix)++;
  }

  offset = -1;
  return offset;  
}


/***************************************************************/
/*                                                             */
/* Function:  dmNextPosition                                   */
/*                                                             */
/* Description:                                                */
/*   Routine to find the next pixel NOT in the region.         */
/*   Returns the 2D position of each pixel in xpix and ypix.   */
/*   returns UT_TRUE if there is more image to iterate over,   */
/*   or UT_FALSE if we are done searching the image.           */
/*                                                             */
/*  Input parameters:                                          */
/*    dmRegionIterator  *iter                                  */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*      dmBool  UT_TRUE or UT_FALSE                            */
/*                                                             */
/***************************************************************/
dmBool dmNextPosition( dmRegionIterator* iter )
{
  double xpos = 0.0;
  double ypos = 0.0;
  int accept  = 0;

  while( iter->ypix <= iter->ymax )
  {
    ypos = (double)iter->ypix + iter->yregoffset - 0.5;

    while( iter->xpix <= iter->xmax )
    {
      xpos = (double)iter->xpix + iter->xregoffset - 0.5;

      accept = regInsideRegion( iter->region, xpos, ypos );

      if ( iter->invert )
      {
        if ( accept )
         return UT_TRUE;
      }
      else
      {
        if ( !accept )
          return UT_TRUE;
      }

      (iter->xpix)++;
    }
 
    iter->xpix = iter->xmin;
    (iter->ypix)++;
  }

  return UT_FALSE;
}


static dmDescriptor* dmRegCoord;   /* For region invert routine */

/***************************************************************/
/*                                                             */
/* Function:  dmi_subspace_col_create_region                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*    dmBlock          *block                                  */
/*    char             *name                                   */
/*    char            **colNames                               */
/*    dmDataType        type                                   */
/*    char             *unit                                   */
/*    regRegion        *region                                 */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*      dmDescriptor *dd                                       */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_subspace_col_create_region(dmBlock *block, char* name,
			   char** colNames, dmDataType type, char* unit, regRegion* region )
{
  dmDescriptor *dd = NULL;
  long dim     = 2;
  long val_dim = 1;

  if ( !region )
    return NULL;

  dd =  dmi_subspace_col_create( block, block->subspace, name, colNames, dim, type, unit, NULL, UT_TRUE );
  dmi_subspace_col_set_values( dd, name, NULL, NULL, 0, type, val_dim, UT_FALSE, region );
  return dd;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_subspace_col_get_region_pointer              */
/*                                                             */
/* Description:                                                */
/*   Return a direct pointer to the region for the current     */
/*   subspace col.                                             */
/*   Careful not to modify the returned value.                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDescriptor     *dd                                     */
/*    dmSubspace       *subspace                               */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*     regRegion       *filter                                 */
/*                                                             */
/***************************************************************/
regRegion* dmi_subspace_col_get_region_pointer( dmDescriptor* dd, dmSubspace* subspace )
{
  dmFilter *filter = NULL;

  if ( !dd ||  dd->dim < 2 )
    return NULL;
 
  filter = dmi_subspace_col_reset_filter( dd, subspace->cpt );
  return dmi_filter_get_region( filter );
}



/***************************************************************/
/*                                                             */
/* Function:  dmi_subspace_col_get_region_pointer              */
/*                                                             */
/* Description:                                                */
/*   Supports correct BACKSCAL when filtering on only          */
/*   spatial ranges                                            */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDescriptor     *dd                                     */
/*    dmSubspace       *subspace                               */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*     regRegion       *filter                                 */
/*                                                             */
/***************************************************************/
regRegion* dmi_subspace_col_alt_get_region_pointer ( dmDescriptor* dd,
                                                     dmSubspace* subspace )
{
  dmFilter *filter = NULL;

  if ( !dd ||  dd->dim < 2 )
    return NULL;
 
  filter = dmi_subspace_col_reset_filter( dd, subspace->cpt );
  return dmi_filter_alt_get_region( filter );
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_subspace_col_update_region                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDescriptor     *dmSS                                   */
/*    const char       *tname                                  */
/*    regRegion        *region                                 */
/*    dmBool            update                                 */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_subspace_col_update_region( dmDescriptor* dmDSS, const char* tname,
                                     regRegion* region, dmBool update )
{
  dmFilter *filter = NULL;
  filter = dmi_subspace_col_get_current_filter( dmDSS );

  if ( !filter)
    return;
     
  dmi_filter_update_region( dmDSS, tname, filter, region, update );
  dmi_descriptor_set_update( dmDSS );
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_subspace_col_init_region                     */
/*                                                             */
/* Description:                                                */
/*   same as dmi_subspace_col_update_region() except don't set */
/*   the descriptor to update on next flush                    */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDescriptor     *dmSS                                   */
/*    const char       *tname                                  */
/*    regRegion        *region                                 */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_subspace_col_init_region( dmDescriptor* dmDSS, const char* tname, regRegion* region )
{
  dmFilter *filter = NULL;

  filter = dmi_subspace_col_get_current_filter( dmDSS );

  if ( !filter)
    return;

  dmi_filter_update_region( dmDSS, tname, filter, region, UT_FALSE );
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_filter_update_region                         */
/*                                                             */
/* Description:                                                */
/*   Same as dmi_subspace_col_update_region() except don't set */
/*   the descriptor to update on next flush                    */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDescriptor     *dmSS                                   */
/*    const char       *tname                                  */
/*    dmFilter         *filter                                 */
/*    regRegion        *region                                 */
/*    dmBool            update                                 */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_filter_update_region( dmDescriptor* dmDSS, const char* tname,
                               dmFilter* filter, regRegion* region, dmBool update )
{
  double fieldx[2] = {0.0, 0.0};
  double fieldy[2] = {0.0, 0.0};

  regRegion *logical_region = NULL;

  if ( dmDSS->dim == 1 )  /* Problem */
    return;

  dmi_descriptor_get_cpt_range( dmDSS, 1, DT_REAL8, &fieldx[0], &fieldx[1] );
  dmi_descriptor_get_cpt_range( dmDSS, 2, DT_REAL8, &fieldy[0], &fieldy[1] );
  dmi_filter_update_region_with_field( filter, tname, region, logical_region, fieldx, fieldy, update );
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_descriptor_get_area                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDescriptor     *dd                                     */
/*    double           *field_area                             */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           double  area                                      */
/*                                                             */
/***************************************************************/
double dmi_descriptor_get_area( const dmDescriptor* dd, double* field_area )
{
  dmDescriptor *ss       = NULL;
  regRegion    *region   = NULL;
  dmSubspace   *subspace = NULL;

  double ffarea = 0.0;
  double area   = 0.0;
  int verbose   = 0;

  if ( !dd )
  {
    if ( field_area )
      *field_area = 1.0;
    return 0.0;
  }

  subspace = dd->block->subspace;
  ss = dmi_subspace_col_open( subspace, dd->name );
  region = dmi_subspace_col_alt_get_region_pointer( ss, subspace );

  if (dmverbose >= 4 )
    verbose = 1;

   dmi_subspace_col_region_area( ss, region, &area, &ffarea, verbose );

  if ( field_area )
    *field_area = ffarea;

  return area;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_subspace_col_region_area                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDescriptor     *dd                                     */
/*    regRegion        *region                                 */
/*    int               verbose                                */
/*                                                             */
/*  Output parameters:                                         */
/*    double           *rrarea                                 */
/*    double           *rfarea                                 */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmi_subspace_col_region_area( dmDescriptor* dd, regRegion* region,
                                   double* rrarea, double* rfarea, int verbose )
{
  double fieldx[2]  = { 0.0, 0.0 };
  double fieldy[2]  = { 0.0, 0.0 };
  double xbounds[2] = { 0.0, 0.0 };
  double ybounds[2] = { 0.0, 0.0 };
  double farea     = 0.0;
  double rarea     = 0.0;
  double barea     = 0.0;
  double bin_min   = 0.0;
  double bin       = 0.0;
  double nmax      = 1.0E6;  /* Max number of pixels to iterate over to calc area */

  bin = dmi_region_get_bin( dd->block );
  nmax = dmi_region_get_nmax( dd->block );

  if ( dd->dim >= 2 )
  {
    dmi_descriptor_get_cpt_range( dd, 1, DT_REAL8, &fieldx[0], &fieldx[1] );
    dmi_descriptor_get_cpt_range( dd, 2, DT_REAL8, &fieldy[0], &fieldy[1] );
  }
  else
  {
    dmi_descriptor_get_cpt_range( dd, 0, DT_REAL8, &fieldx[0], &fieldx[1] );
    fieldy[0] = fieldx[0];
    fieldy[1] = fieldx[1];
  }

  if ( verbose )
    fprintf( stderr, "FIELD %s %g:%g %g:%g  AREA = %f BIN=%f\n",dd->name, fieldx[0], fieldx[1],
             fieldy[0], fieldy[1], farea, bin );
  farea = regArea( NULL, fieldx, fieldy, bin );

  if ( verbose )
    fprintf( stderr, "FIELD %g:%g %g:%g  AREA = %f\n", fieldx[0], fieldx[1],
             fieldy[0], fieldy[1], farea );

  regExtent( region, fieldx, fieldy, xbounds, ybounds );
  barea = regArea( NULL, xbounds, ybounds, bin );

  if ( verbose )
    fprintf( stderr, "REGION BOUNDS %f:%f x %f:%f  AREA=%g\n", xbounds[0], xbounds[1],
              ybounds[0], ybounds[1], barea );

  bin_min = sqrt( barea / nmax );

  if ( bin < bin_min )
    bin = bin_min;

  rarea = regArea( region, fieldx, fieldy, bin );

  if ( verbose )
    fprintf( stderr, "REGION AREA = %g BIN=%g BMIN=%g NMAX=%g\n", rarea, bin, bin_min, nmax );

  *rrarea = rarea;
  *rfarea = farea;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_filter_compare_region                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*    dmFilter         *src                                    */
/*    dmFiltern        *cmp  n                                 */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*           dmBool  UT_TRUE or UT_FALSE                       */
/*                                                             */
/***************************************************************/
dmBool dmi_filter_compare_region( dmFilter* src, dmFilter* cmp )
{
  /* Compare regions here */
  if ( !cmp->region && !src->region )
    return UT_TRUE;

  if ( !cmp->region || !src->region )
    return UT_FALSE;

  return (dmBool)regCompareRegion( src->region, cmp->region );
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_filter_get_region                            */
/*                                                             */
/* Description:                                                */
/*   return pointer to filter's region element                 */
/*                                                             */
/*  Input parameters:                                          */
/*    dmFilter         *filter                                 */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*           regRegion *region                                 */
/*                                                             */
/***************************************************************/
regRegion* dmi_filter_get_region( dmFilter* filter )
{
  if ( !filter )
    return NULL;

  return filter->region;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_filter_alt_get_region                        */
/*                                                             */
/* Description:                                                */
/*   Support returning a bounds region (rectangle from ranges) */
/*   if no explicit region                                     */
/*                                                             */
/*  Input parameters:                                          */
/*    dmFilter         *filter                                 */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*           regRegion *region                                 */
/*                                                             */
/***************************************************************/
regRegion* dmi_filter_alt_get_region( dmFilter* filter )
{
  regRegion *bounds_region = NULL;
  char       BoundsRegionString[512] = " ";
  int        HaveBoundsRegion = 0;

  if ( !filter )
     return NULL;

  if ( filter->region )
    return filter->region;

  if ( filter->dim == 2 && ar_mix_is_numeric( filter->type ))
  {
    HaveBoundsRegion = dmi_rangelist_to_rectangle( &filter->ranges[0], &filter->ranges[1], BoundsRegionString );

    if ( HaveBoundsRegion )
    {
#if DEBUG
printf( "REG PARSE 1 %s\n", BoundsRegionString );
#endif

      bounds_region = regParse( BoundsRegionString );
      return bounds_region;
    }
  }

  return NULL;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_region_is_rectangle                          */
/*                                                             */
/* Description:                                                */
/*   Determine is a regions shape is rectangle and returns     */
/*   its X and Y values                                        */
/*                                                             */
/*  Input parameters:                                          */
/*    regRegion        *region                                 */
/*                                                             */
/*  Output parameters:                                         */
/*    double           *x                                      */
/*    double           *y                                      */
/*                                                             */
/*  Returns:                                                   */
/*           int   0 if false, 1 if true                       */
/*                                                             */
/***************************************************************/
int dmi_region_is_rectangle( regRegion* region, double* x, double* y )
{
  regShape *shape = NULL;
  char      name[80] = {80 * '\0'};
  long      n = 0;

  if ( !region )
    return 0;

  n = regGetNoShapes( region );

  if ( n > 1 )
    return 0;

  shape = regGetShapeNo( region, 1 );
  regShapeGetName( shape, name, 80 );

  if ( cs_uceq( name, "Rectangle" ))
  {
    regShapeGetPoints( shape, x, y, 2 );
    return 1;
  }
 
  return 0;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_region_is_inside_bounds                      */
/*                                                             */
/* Description:                                                */
/*   Check if bounds of region entirely within rectangle       */
/*                                                             */
/*  Input parameters:                                          */
/*    regRegion        *region                                 */
/*    double           *x                                      */
/*    double           *y                                      */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           int   0 if false, 1 if true                       */
/*                                                             */
/***************************************************************/
int dmi_region_inside_bounds( regRegion* region, double* x, double* y )
{
  double xbounds[2] = { 0.0, 0.0 };
  double ybounds[2] = { 0.0, 0.0 };

  if ( dmi_region_is_rectangle( region, xbounds, ybounds ))
  {
    if ( !regInsideRegion( region, x[0], y[0] ))
      return 0;

    if ( !regInsideRegion( region, x[0], y[1] ))
      return 0;

    if ( !regInsideRegion( region, x[1], y[0] ))
      return 0;

    if ( !regInsideRegion( region, x[1], y[1] ))
      return 0;

    return 1;
  }

  return 0;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_filter_update_region_with_field              */
/*                                                             */
/* Description:                                                */
/*    Combine regions                                          */
/*                                                             */
/*  Input parameters:                                          */
/*    dmFilter         *filter                                 */
/*    const char       *tname                                  */
/*    regRegion        *region                                 */
/*    regRegion        *logical_region                         */
/*    double           *fieldx                                 */
/*    double           *fieldy                                 */
/*    dmBool            update                                 */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*           int   0 if false, 1 if true                       */
/*                                                             */
/***************************************************************/
void dmi_filter_update_region_with_field( dmFilter* filter, const char* tname,
                            regRegion* region,  regRegion* logical_region,
                            double* fieldx, double* fieldy, dmBool update )
{
  regRegion *newRegion = NULL;
  regRegion *newLogicalRegion = NULL;

  char   *oname = NULL;

  double  xbounds[2] = { 0.0, 0.0 };
  double  ybounds[2] = { 0.0, 0.0 };
  int     apply      = 0; /* Apply empty filters? */
  if ( !filter )
    return;

  if( update && filter->region )
  {
    regExtent( region, fieldx, fieldy, xbounds, ybounds );

    if ( dmi_region_inside_bounds( filter->region, xbounds, ybounds ))
    {
      newRegion = regCopyRegion( region );

      if ( logical_region ) 
        newLogicalRegion = regCopyRegion( logical_region );
    }
    else 
    {
      newRegion = regIntersectRegion( filter->region, region );

      if ( logical_region ) 
        newLogicalRegion = regIntersectRegion( filter->logical_region, logical_region );
    }

    free( filter->regstring );
    filter->regstring= NULL;
    regFree( filter->region );
    regFree( filter->logical_region );
    filter->region = newRegion;
    filter->logical_region = newLogicalRegion;
  }
  else
  {
    free( filter->regstring );
    filter->regstring = NULL;
    regFree( filter->region );
    regFree( filter->logical_region );
    filter->region = regCopyRegion( region );
    filter->logical_region = regCopyRegion( logical_region );
  }

  if ( filter->region )
  {
    regExtent( filter->region, fieldx, fieldy, xbounds, ybounds );

    /* Should check filter->n here; also need proper support for 2D filters */  

    if ( filter->dim == 2 )
    {
      dmi_rangelist_intersect( &filter->ranges[0], DT_REAL8, &xbounds[0], &xbounds[1], 1, UT_FALSE, apply );
      dmi_rangelist_intersect( &filter->ranges[1], DT_REAL8, &ybounds[0], &ybounds[1], 1, UT_FALSE, apply );
    }
  }

 /*
  *  For [Set], only keep the old name if the new name is null
  *             (user just entered the new values but wants to keep the table name )
  *  For [Update], always keep the old name unless it is null
  *             (case of an intersection of two subspaces)
  */

  oname = filter->name;

  if ( tname && *tname && ( !update || !oname ) )
    dmi_filter_set_tabname( filter, tname );
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_coord_region_convert_init                    */
/*                                                             */
/* Description:                                                */
/*   Make a function to pass to the region library to convert  */
/*   world to physical pixels.                                 */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDescriptor     *dd                                     */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*          void                                               */
/*                                                             */
/***************************************************************/
void dmi_coord_region_convert_init( dmDescriptor* dd )
{
  dmRegCoord = dd;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_coord_region_convert                         */
/*                                                             */
/* Description:                                                */
/*   Convert world to physical pixels.                         */
/*                                                             */
/*  Input parameters:                                          */
/*    double           *world                                  */
/*                                                             */
/*  Output parameters:                                         */
/*    double           *pix                                    */
/*                                                             */
/*  Returns:                                                   */
/*          void                                               */
/*                                                             */
/***************************************************************/
void dmi_coord_region_convert( double* world, double* pix )
{
  dmi_coord_invert_d( dmRegCoord, world, pix );
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_coord_convert_region_phys                    */
/*                                                             */
/* Description:                                                */
/*   Make a function to pass to the region library to convert  */
/*   logical to physical pixels.                               */
/*                                                             */
/*  Input parameters:                                          */
/*    double           *logic                                  */
/*                                                             */
/*  Output parameters:                                         */
/*    double           *pix                                    */
/*                                                             */
/*  Returns:                                                   */
/*          void                                               */
/*                                                             */
/***************************************************************/
void dmi_coord_convert_region_phys( double* logic, double* pix )
{
  dmi_coord_calc_d( dmRegCoord, logic, pix );
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_coord_convert_world_region                   */
/*                                                             */
/* Description:                                                */
/*   Convert world region                                      */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDescriptor     *vcoord                                 */
/*                                                             */
/*  Output parameters:                                         */
/*    regRegion        *region                                 */
/*                                                             */
/*  Returns:                                                   */
/*          void                                               */
/*                                                             */
/***************************************************************/
int dmi_coord_convert_world_region( dmDescriptor* vcoord, regRegion* region )
{
  double crpix[2] = { 0.0, 0.0 };
  double crval[2] = { 0.0, 0.0 };
  double cdelt[2] = { 0.0, 0.0 };
  double scale    = 0;
  long   dim       = 2;

  dmi_coord_get_transform_d( vcoord, crpix, crval, cdelt, dim );
  scale = fabs( cdelt[0] );
  dmi_coord_region_convert_init( vcoord );
  regConvertWorldRegion( region, scale, dmi_coord_region_convert );

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_region_file_cols                             */
/*                                                             */
/* Description:                                                */
/*   See dmerror.h for REGNCOL symbols                         */
/*                                                             */
/*  Input parameters:                                          */
/*    dmBlock          *block                                  */
/*                                                             */
/*  Output parameters:                                         */
/*    dmDescriptor   ***colsp                                  */
/*    long            **dimsp                                  */
/*                                                             */
/*  Returns:                                                   */
/*          long  REGNCOLS                                     */
/*                                                             */
/***************************************************************/
long dmi_region_file_cols( dmBlock* block, dmDescriptor*** colsp, long** dimsp )
{
  dmDescriptor  **cols = NULL;
  dmDescriptor  **list = NULL;
  dmDescriptor   *default_pos_cols[2] = { NULL, NULL };

  char  *default_pos_names[2] = { "X", "Y" };
  char  *colname = NULL;

  long *dims      = 0;
  long  ncols     = 0;
  long  maxpoints = 0;
  long  i         = 0;
  long  vdim      = 0;


  if ( !block || !colsp || !dimsp )
    return 0;

  list = block->cols;

  ncols = block->ncols;

  if ( ncols == 0 )
    return 0;

  cols = calloc( REGNCOLS, UT_SZ_P );
  dims = calloc( REGNCOLS, UT_SZ_I );

  for ( i = 0; i < ncols; i++ )
  {
    /*
     *  The first vector column is required to be position
     *  It does not have to be called anything in particular.
     */
    vdim = dmi_descriptor_get_element_dim( list[i] );

    if ( vdim == 2 && !cols[REG_POS_COL] )
      cols[REG_POS_COL] = list[i];

    colname = dmi_descriptor_get_name( list[i] );

    if ( cs_uceq( colname, "SHAPE" ) )
      cols[REG_SHAPE_COL] = list[i];
    else
      if ( cs_uceq( colname, "R" ) )
      {
        cols[REG_R_COL] = list[i];
        dims[REG_R_COL] = dmi_descriptor_get_array_size( cols[REG_R_COL] );
      }
      else
        if ( cs_uceq( colname, "ROTANG" ))
        {
          cols[REG_ROT_COL] = list[i];
          dims[REG_ROT_COL] = dmi_descriptor_get_array_size( cols[REG_ROT_COL] );
        }
        else
          if ( cs_uceq( colname, "COMPONENT" ))
            cols[REG_CPT_COL] = list[i];
  }

  if ( !cols[REG_SHAPE_COL] )
    return 0;

  if ( !cols[REG_POS_COL] )
  {
    /* Didn't find a 2-D vector to use as POS, so look for X and Y */
    /* columns.  If we find them, make them into a POS vector and  */
    /* use that.                                                   */

    for ( i = 0; i < 2; i++ )
    {
      default_pos_cols[i] = dmi_descriptor_list_search_descriptor(
              (const dmDescriptor**)list, ncols, default_pos_names[i] );

      if ( !default_pos_cols[i] )
        return 0;
    }

    cols[REG_POS_COL] = dmi_make_vector( block, "POS", default_pos_cols, 2 );

    if ( !cols[REG_POS_COL] )
      return 0;

    /* NOTE:  Most likely, this is safe only if the buffer contains no data! */
    /* Currently, this function is called only at block open and in          */
    /* dmTableReadRegion(), so it's okay.                                    */

    dmi_block_add_col( block, cols[REG_POS_COL], DM_FLAG_AUX_COLUMN );
  }

  maxpoints = dmi_descriptor_get_array_size( cols[REG_POS_COL] );

  if ( maxpoints < 1 )
   maxpoints = 1;

  dims[REG_POS_COL] = maxpoints;

  *colsp = cols;
  *dimsp = dims;
  return REGNCOLS;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_image_filter_regions                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDescriptor     *image                                  */
/*    dmDescriptor     *raw_image                              */
/*    long             *v_llpix                                */
/*    long             *v_urpix                                */
/*    long             *dpix                                   */
/*    void             *values                                 */
/*    dmSubspace       *rts                                    */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*          void                                               */
/*                                                             */
/***************************************************************/
void  dmi_image_filter_regions( dmDescriptor* image, dmDescriptor* raw_image,
                                long* v_llpix, long* v_urpix, long* dpix,
                                void* values, dmSubspace* rts )
{
  regRegion    *region      = NULL;
  dmDescriptor *dd          = NULL;
  dmFilter    **filter_list = NULL;
  dmFilter     *filter      = NULL;
  dmBool        local_llpix = UT_FALSE;

  GenericType tmp = 0.0;

  void *default_null = &tmp;
  void *null         = default_null;

  long axis_offset = 0;
  long i           = 0;
  int invert       = 0;
  int compact     = UT_TRUE;


  /* Here we filter any regions */

  if ( rts )
  {
    invert = rts->invert;
    filter_list = dmi_map_filters_to_image_groups( raw_image, rts, compact );
    axis_offset = 0;

    for (i = 0; i < raw_image->ngroups; i++ )
    {
      filter = filter_list[i];
      dd = dmi_descriptor_get_group_no( raw_image, i+1 );

      if ( filter && filter->logical_region )
      {
        region = filter->logical_region;
        break;
      }
      else
      {
        if ( filter && filter->region )
        {   
          dmi_coord_fix_region_filter( dd, filter, UT_TRUE, DM_FLAG_PHYS );
          region = filter->logical_region;

          if ( !region )
          {
             sprintf( dm_errbuf, "Can't convert region filter using unequal scaling factors.\nPixels outside region will not be set to null." );
             dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
          }
          break;
        }
      }

      axis_offset += dd->dim;

    } // end for loop  

    free( filter_list );
  }
 
  if ( !region )
    return;

   if ( raw_image->null )
    null = raw_image->null;

   if ( !v_llpix )
   {
      v_llpix = calloc( raw_image->naxes, UT_SZ_I );
      v_urpix = calloc( raw_image->naxes, UT_SZ_I );
      local_llpix = UT_TRUE;

      for ( i = 0; i < raw_image->naxes; i++ )
      {
        v_llpix[i] = 1;
        v_urpix[i] = raw_image->axes[i];
      }  
   }

   /* We have separate code specifically for dealing with >2 dim images. */

   if ( raw_image->naxes == 2 )
   {
     dmi_image_filter_regions_2d( raw_image, v_llpix, v_urpix, region, invert, null, values );
   } 
   else
   {
     dmi_image_filter_regions_gen( image, raw_image, axis_offset, axis_offset+1, v_llpix, v_urpix, region, invert, null, values );
   }

   /* Suppress compiler warning - dpix not used yet */

   if ( local_llpix )
   {
     free( v_llpix );
     free( v_urpix );
   }

   if ( !dpix )
     return;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_image_filter_regions_2d                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDescriptor     *raw_image                              */
/*    long             *v_llpix                                */
/*    long             *v_urpix                                */
/*    regRegion        *region                                 */
/*    int               invert                                 */
/*    void             *null                                   */
/*    void             *values                                 */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*          void                                               */
/*                                                             */
/***************************************************************/
void dmi_image_filter_regions_2d( dmDescriptor* raw_image, long* v_llpix,
                                 long* v_urpix, regRegion* region, int invert,
                                 void* null, void* values )
{
  dmRegionIterator iter;

  unsigned char  *ucvals = NULL;
  unsigned char   ucnull = '\0';
  unsigned short *usvals = NULL;
  unsigned short  usnull = 0;
  unsigned long  *ulvals = NULL;
  unsigned long   ulnull = 0;

  float  *fvals = NULL;
  float   fnull = 0.0;
  double *dvals = NULL;
  double  dnull = 0.0;
  long   *lvals = NULL;
  long    lnull = 0;
  short  *svals = 0;
  short   snull = 0;

  long    offset = 0;


  dmRegionIterateInit( raw_image->block, &iter, v_llpix[0], v_llpix[1], v_urpix[0], v_urpix[1], region, invert );

  switch( raw_image->type )
  {
    case DT_UINT1:
      ucvals = (unsigned char*)values;
      ucnull = *(unsigned char*)null;

      while(  ( offset = dmi_region_iterate_2d( &iter ) ) >= 0 )
        ucvals[offset] = ucnull;
      break;

    case DT_UINT2:
      usvals = (unsigned short*)values;
      usnull = *(unsigned short*)null;

      while(  ( offset = dmi_region_iterate_2d( &iter ) ) >= 0 )
        usvals[offset] = usnull;
     break;

    case DT_UINT4:
      ulvals = (unsigned long*)values;
      ulnull = *(unsigned long*)null;

      while(  ( offset = dmi_region_iterate_2d( &iter ) ) >= 0 )
        ulvals[offset] = ulnull;
     break;

    case DT_INT2:
      svals = (short*)values;
      snull = *(short*)null;

      while(  ( offset = dmi_region_iterate_2d( &iter ) ) >= 0 )
        svals[offset] = snull;
      break;

    case DT_INT4:
      lvals = (long*)values;
      lnull = *(long*)null;

      while(  ( offset = dmi_region_iterate_2d( &iter ) ) >= 0 )
        lvals[offset] = lnull;
      break;

    case DT_REAL4:
      fvals = (float*)values;
      fnull = *(float*)null;

      while(  ( offset = dmi_region_iterate_2d( &iter ) ) >= 0 )
        fvals[offset] = fnull;
      break;

     case DT_REAL8:
       dvals = (double*)values;
       dnull = *(double*)null;

       while(  ( offset = dmi_region_iterate_2d( &iter ) ) >= 0 )
         dvals[offset] = dnull;
       break;
     default:
       break;
  }
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_image_filter_regions_gen                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDescriptor     *image                                  */
/*    dmDescriptor     *raw_image                              */
/*    long              xaxis_offset                           */
/*    long              yaxis_offset                           */
/*    long             *v_llpix                                */
/*    long             *v_urpix                                */
/*    regRegion        *region                                 */
/*    int               invert                                 */
/*    void             *null                                   */
/*    void             *values                                 */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*          void                                               */
/*                                                             */
/***************************************************************/
void  dmi_image_filter_regions_gen( dmDescriptor* image, dmDescriptor* raw_image,
            long xaxis_offset, long yaxis_offset, long* v_llpix, long* v_urpix,
            regRegion* region, int invert, void* null, void* values )
{
  dmRegionIterator iter;

  float  *fvals = NULL;
  float   fnull = 0.0;
  double *dvals = NULL;
  double  dnull = 0.0;
  long   *lvals = NULL;
  long    lnull = 0;
  short  *svals = NULL;
  short   snull = 0;

  unsigned char  *ucvals = NULL;
  unsigned char   ucnull = '\0';
  unsigned long  *ulvals = NULL;
  unsigned long   ulnull = 0;
  unsigned short *usvals = NULL;
  unsigned short  usnull = 0;

  long *weight = NULL;
  long *opixel = NULL;
  long *llpix  = NULL;
  long *urpix  = NULL;
  long *axes   = NULL;
  long  axis_ndx   = 0;
  long  offset     = 0;
  long  xpix       = 0;
  long  ypix       = 0;
  long  min_axisoffset = 0;
  long  max_axisoffset = 0;
  long  naxes      = 0;
  long  first_axis = 0;
  long  last_axis  = 0;
  long  inc_axis   = 0;


  naxes = image->naxes;
  axes = image->axes;
  weight = dmi_calc_image_weights( axes, naxes );
  opixel = malloc( naxes * UT_SZ_I );
  llpix = malloc( naxes * UT_SZ_I );
  urpix = malloc( naxes * UT_SZ_I );

  for ( axis_ndx = 0; axis_ndx < naxes; axis_ndx++ )
  {
    llpix[axis_ndx] = v_llpix[axis_ndx];
    if ( ( axis_ndx != xaxis_offset ) && ( axis_ndx != yaxis_offset ) &&
       ( llpix[axis_ndx] > 0 ) )
    {
       llpix[axis_ndx]--;
    }
    urpix[axis_ndx] = v_urpix[axis_ndx];
  }

  dmRegionIterateInit( image->block, &iter, v_llpix[xaxis_offset], v_llpix[yaxis_offset], v_urpix[xaxis_offset], v_urpix[yaxis_offset], region, invert );

  min_axisoffset = ( xaxis_offset < yaxis_offset ? xaxis_offset : yaxis_offset );
  max_axisoffset = ( min_axisoffset == xaxis_offset ? yaxis_offset : xaxis_offset );
 
  first_axis = 0;

  while( first_axis == xaxis_offset || first_axis == yaxis_offset )
    ++first_axis;

  if ( first_axis >= naxes )
    first_axis = naxes-1;

  last_axis = naxes-1;

  switch( raw_image->type )
  {
    case DT_UINT1:
      ucvals = (unsigned char*)values;
      ucnull = *(unsigned char*)null;

      while( dmi_region_iterate_gen( &iter, &xpix, &ypix ))
      {
        llpix[xaxis_offset] = xpix - 1;
        llpix[yaxis_offset] = ypix - 1;
        urpix[xaxis_offset] = xpix;
        urpix[yaxis_offset] = ypix;
        offset = 0;

        for ( axis_ndx = 0; axis_ndx < naxes; axis_ndx++ )
        {
          opixel[axis_ndx] = llpix[axis_ndx];
          offset += llpix[axis_ndx] * weight[axis_ndx];
        }

        while( opixel[last_axis] < urpix[last_axis] )
        {
          ucvals[offset] = ucnull;
          opixel[first_axis]++;
          offset += weight[first_axis];
          inc_axis = first_axis;

          while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
          {
            offset += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );
            opixel[inc_axis] = llpix[inc_axis];
            inc_axis++;
            opixel[inc_axis]++;
          }
        }
      }
      break;

    case DT_UINT2:
      usvals = (unsigned short*)values;
      usnull = *(unsigned short*)null;

      while( dmi_region_iterate_gen( &iter, &xpix, &ypix ))
      {
        llpix[xaxis_offset] = xpix-1;
        llpix[yaxis_offset] = ypix-1;
        urpix[xaxis_offset] = xpix;
        urpix[yaxis_offset] = ypix;
        offset = 0;

        for ( axis_ndx = 0; axis_ndx < naxes; axis_ndx++ )
        {
          opixel[axis_ndx] = llpix[axis_ndx];
          offset += llpix[axis_ndx] * weight[axis_ndx];
        }

        while( opixel[last_axis] < urpix[last_axis] )
        {
          usvals[offset] = usnull;
          opixel[first_axis]++;
          offset += weight[first_axis];
          inc_axis = first_axis;

          while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
          {
             offset += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );
             opixel[inc_axis] = llpix[inc_axis];
             inc_axis++;
             opixel[inc_axis]++;
          }
        }
      }
      break;

    case DT_UINT4:
      ulvals = (unsigned long*)values;
      ulnull = *(unsigned long*)null;

      while( dmi_region_iterate_gen( &iter, &xpix, &ypix ))
      {
        llpix[xaxis_offset] = xpix-1;
        llpix[yaxis_offset] = ypix-1;
        urpix[xaxis_offset] = xpix;
        urpix[yaxis_offset] = ypix;
        offset = 0;

        for ( axis_ndx = 0; axis_ndx < naxes; axis_ndx++ )
        {
          opixel[axis_ndx] = llpix[axis_ndx];
          offset += llpix[axis_ndx] * weight[axis_ndx];
        }

        while( opixel[last_axis] < urpix[last_axis] )
        {
          ulvals[offset] = ulnull;
          opixel[first_axis]++;
          offset += weight[first_axis];
          inc_axis = first_axis;

          while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
          {
             offset += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );
             opixel[inc_axis] = llpix[inc_axis];
             inc_axis++;
             opixel[inc_axis]++;
          }
        }
      }
      break;

    case DT_INT2:
      svals = (short*)values;
      snull = *(short*)null;

      while( dmi_region_iterate_gen( &iter, &xpix, &ypix ))
      {
        llpix[xaxis_offset] = xpix-1;
        llpix[yaxis_offset] = ypix-1;
        urpix[xaxis_offset] = xpix;
        urpix[yaxis_offset] = ypix;
        offset = 0;

        for ( axis_ndx = 0; axis_ndx < naxes; axis_ndx++ )
        { 
          opixel[axis_ndx] = llpix[axis_ndx];
          offset += llpix[axis_ndx] * weight[axis_ndx];
        }

        while( opixel[last_axis] < urpix[last_axis] )
        {
          svals[offset] = snull;
          opixel[first_axis]++;
          offset += weight[first_axis];
          inc_axis = first_axis;

          while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
          {
             offset += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );
             opixel[inc_axis] = llpix[inc_axis];
             inc_axis++;
             opixel[inc_axis]++;
          }
        }
      }
      break;

    case DT_INT4:
      lvals = (long*)values;
      lnull = *(long*)null;

      while( dmi_region_iterate_gen( &iter, &xpix, &ypix ))
      {
        llpix[xaxis_offset] = xpix-1;
        llpix[yaxis_offset] = ypix-1;
        urpix[xaxis_offset] = xpix;
        urpix[yaxis_offset] = ypix;
        offset = 0;

        for ( axis_ndx = 0; axis_ndx < naxes; axis_ndx++ )
        {
          opixel[axis_ndx] = llpix[axis_ndx];
          offset += llpix[axis_ndx] * weight[axis_ndx];
        }

        while( opixel[last_axis] < urpix[last_axis] )
        {
          lvals[offset] = lnull;
          opixel[first_axis]++;
          offset += weight[first_axis];
          inc_axis = first_axis;

          while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
          {
            offset += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );
            opixel[inc_axis] = llpix[inc_axis];
            inc_axis++;
            opixel[inc_axis]++;
          }
        }
      }
      break;

    case DT_REAL4:
      fvals = (float*)values;
      fnull = *(float*)null;

      while( dmi_region_iterate_gen( &iter, &xpix, &ypix ))
      {
        llpix[xaxis_offset] = xpix-1;
        llpix[yaxis_offset] = ypix-1;
        urpix[xaxis_offset] = xpix;
        urpix[yaxis_offset] = ypix;
        offset = 0;

        for ( axis_ndx = 0; axis_ndx < naxes; axis_ndx++ )
        {
          opixel[axis_ndx] = llpix[axis_ndx];
          offset += llpix[axis_ndx] * weight[axis_ndx];
        }

        while( opixel[last_axis] < urpix[last_axis] )
        {
          fvals[offset] = fnull;
          opixel[first_axis]++;
          offset += weight[first_axis];
          inc_axis = first_axis;

          while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
          {
            offset += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );
            opixel[inc_axis] = llpix[inc_axis];
            inc_axis++;
            opixel[inc_axis]++;
          }
        }
      }
      break;

    case DT_REAL8:
      dvals = (double*)values;
      dnull = *(double*)null;

      while( dmi_region_iterate_gen( &iter, &xpix, &ypix ))
      {
        llpix[xaxis_offset] = xpix-1;
        llpix[yaxis_offset] = ypix-1;
        urpix[xaxis_offset] = xpix;
        urpix[yaxis_offset] = ypix;
        offset = 0;

        for ( axis_ndx = 0; axis_ndx < naxes; axis_ndx++ )
        {
          opixel[axis_ndx] = llpix[axis_ndx];
          offset += llpix[axis_ndx] * weight[axis_ndx];
        }

        while( opixel[last_axis] < urpix[last_axis] )
        {
          dvals[offset] = dnull;
          opixel[first_axis]++;
          offset += weight[first_axis];
          inc_axis = first_axis;

          while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
          {
            offset += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );
            opixel[inc_axis] = llpix[inc_axis];
            inc_axis++;
            opixel[inc_axis]++;
          }
        }
      }
      break;
 
    default:
     break;
  }

  free( weight );
  free( opixel );
  free( llpix );
  free( urpix );
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_region_check_bounds_syntax                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*    char             *ptr                                    */
/*                                                             */
/*  Output parameters:                                         */
/*    int              *extent                                 */
/*    int              *is_reg                                 */
/*                                                             */
/*  Returns:                                                   */
/*          char  *retval                                      */
/*                                                             */
/***************************************************************/
char* dmi_region_check_bounds_syntax( char* ptr, int* extent, int* is_reg )
{
  char *string  = ptr;
  char *tail    = NULL;
  char *regfile = NULL;
  char *retval  = NULL;

  *is_reg = 0;
  *extent = 0;

  if ( cs_uceq_siz( ptr, "bounds(", 7 ) )
  {
    *extent = 1;
    string = cs_dup( ptr + 7 );
     tail = string + strlen( string ) - 1;

    if ( *tail == ')' )
      *tail = '\0';      
  }
  else
  {
    ;
  }

  /* Support  "@my.reg" and "region(my.reg)" */
  if ( *string == '@' )
  {
    *is_reg = 1;
    regfile = cs_dup( string+1 );
  }
  else
    if ( cs_uceq_siz( string, "region(", 7 ) )
    {
      *is_reg = 1;
      regfile=cs_dup( string+7 );
    }

  if ( *is_reg ) 
  {
    retval = regfile;
  }
  else
  {
    if ( *extent )
      retval = string;
    else
      retval = cs_dup( string );
  }

  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_region_bounds                                */
/*                                                             */
/* Description:                                                */
/*   Support "bounds(my.reg)". WARNING: Frees the input region.*/
/*   If ifieldx and ifieldy are not NULL they will be treated  */
/*   as the hard boundaries of the region.  Otherwise,         */
/*   dmi_region_bound() will do its best to not return an      */
/*   infinite boundary.                                        */
/*                                                             */
/*  Input parameters:                                          */
/*    regRegion        *region                                 */
/*    double           *ifieldx                                */
/*    double           *ifieldy                                */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*          regRegion  *region                                 */
/*                                                             */
/***************************************************************/
regRegion* dmi_region_bound( regRegion* region, double* ifieldx, double* ifieldy )
{
  double default_fieldx[2] = { 0.5, 32768.5 }; /* Default field boundary */
  double default_fieldy[2] = { 0.5, 32768.5 }; /* Default field boundary */
  double fieldx[2]         = { DT_REAL8_MIN, DT_REAL8_MAX };
  double fieldy[2]         = { DT_REAL8_MIN, DT_REAL8_MAX };
  double xpos[2]           = { 0.0, 0.0 };
  double ypos[2]           = { 0.0, 0.0 };
  char buf[1024]           = { 1028 * '\0' };

  if ( ifieldx )
    memcpy( fieldx, ifieldx, sizeof(double)*2 );

  if ( ifieldy )
    memcpy( fieldy, ifieldy, sizeof(double)*2 );

  regExtent( region, fieldx, fieldy, xpos, ypos );
  regFree( region );

  /* make sure we don't end up with an infinite rectangle */
  if ( xpos[0] == DT_REAL8_MIN )
    xpos[0] = default_fieldx[0];

  if ( xpos[1] == DT_REAL8_MAX )
    xpos[1] = default_fieldx[1];

  if ( ypos[0] == DT_REAL8_MIN )
    ypos[0] = default_fieldy[0];

  if ( ypos[1] == DT_REAL8_MAX )
    ypos[1] = default_fieldy[1];

  sprintf( buf, "rectangle(%g,%g,%g,%g)", xpos[0], ypos[0], xpos[1], ypos[1] );
  region = regParse( buf );
  return region;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_region_resolve                               */
/*                                                             */
/* Description:                                                */
/*   Support "bounds()" and "region()" syntax. No error on     */
/*   failed parse                                              */
/*                                                             */
/*  Input parameters:                                          */
/*    char             *ptr                                    */
/*    regRegion        *(*region_read)(char*)                  */
/*    double           *fieldx                                 */
/*    double           *fieldy                                 */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*          regRegion  *region                                 */
/*                                                             */
/***************************************************************/
regRegion* dmi_region_resolve( char* ptr, regRegion* (*region_read)(char*),
                               double* fieldx, double* fieldy )
{
  regRegion *region       = NULL;
  regRegion *bound_region = NULL;
  regRegion *region_ptr   = NULL;

  char *string = ptr;

  int extent = 0;
  int is_reg = 0;


  /* Support "bounds(my.reg)" */
  string = dmi_region_check_bounds_syntax( ptr, &extent, &is_reg );

  /* Support region file read */
  if ( is_reg && region_read )
    region = region_read( string );         
  else
    region = regParse( string );

  free( string );

  /* If we don't have a region at this point, then there must be */
  /* a syntax error in ptr, so return NULL to indicate an error  */

  if ( !region )
    return NULL;

  if ( extent )
    region = dmi_region_bound( region, fieldx, fieldy );

  if ( fieldx && fieldy && !extent )
  {
    bound_region = regCreateRectangle( fieldx, fieldy );
    region_ptr = region;
    region = regIntersectRegion( region_ptr, bound_region );
    regFree( region_ptr );
    regFree( bound_region );
  }

  return region;
}


/***************************************************************/
/*                                                             */
/* Function:  dmi_region_resolve                               */
/*                                                             */
/* Description:                                                */
/*   Retain this for dmlist hack compatibility only            */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDescriptor     *dd                                     */
/*    regRegion        *region                                 */
/*    double           *rrarea                                 */
/*    double           *rfarea                                 */
/*    int               verbose                                */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*          void                                               */
/*                                                             */
/***************************************************************/
void dmpRegionArea( dmDescriptor* dd, regRegion* region, double* rrarea, double* rfarea, int verbose )
{
  dmi_subspace_col_region_area( dd, region, rrarea, rfarea, verbose );
}
