/*                                                                
**  Copyright (C) 2002-2012  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmsubspace.c                                                */
/*                                                               */
/*   Functions to allocate and maintain the subspace structure   */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"
#include "dmiparser.h"

#define MAX_DCPTS 10


static int dmi_subspace_col_intersect(dmBlock* destBlock, dmSubspace* subspace,
                                      dmDescriptor* srcDesc, dmBool copyDefault,
                                      dmBool requireTable );

static int dmi_subspace_intersect_cpt( dmBlock* destBlock, dmSubspace* subspace,
                                       dmDescriptor** ddlist, long numDescs,
                                       dmBool copyDefault, dmBool requireTable );

static int dmi_subspace_clone_cpt( dmSubspace* subspace, long CptNo );



/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_col_update                         */
/*                                                             */
/* Description:                                                */
/*             Code fragment to update an existing DSS with    */
/*             new values and possibly providing a table       */
/*             name if it doesn't have one                     */
/*                                                             */
/*  Input parameters:                                          */
/*              dmDescriptor *indd                             */
/*              dmFilter     *filter                           */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            int    return_value                              */
/*                                                             */
/***************************************************************/
int dmi_subspace_col_update_from_filter( dmDescriptor* indd, dmFilter* filter )
{
  dmDescriptor *dd = indd;
  dmSubspace   *subspace = NULL;
  dmRangeList  *range    = NULL;
  regRegion    *region   = NULL;
  void         *rmins    = NULL;
  void         *rmaxes   = NULL;

  long  elemNo  = 1;
  long  rn      = 0;
  long  elemMin = 1;
  long  elemMax = 1;
  int   retval  = 0;


  subspace = dd->block->subspace;
  elemMax  = filter->dim;

  if ( dd->cptno > 1 && dd->parent && dd->cptno <= filter->dim )
  {
    elemMin = elemMax = dd->cptno;
    dd = dd->parent;
  }

  if ( filter->null )
  {
    dmi_subspace_col_make_null( dd );
    return dmSUCCESS;
  }

  for ( elemNo = elemMin; elemNo <= elemMax; elemNo++ )
  {
    range = &filter->ranges[elemNo-1];

    if ( range )
    {
      dmi_rangelist_pointers( range, &rmins, &rmaxes, &rn );

      retval = dmi_subspace_col_update_cpt( dd, filter->name, elemNo, filter->type,
                                       rmins, rmaxes, rn, range->incnull, UT_TRUE );

      if ( retval != dmSUCCESS ) 
        return retval;
    }
  }

  region = dmi_filter_get_region( filter );

  if ( region )
  {
    dmi_subspace_col_update_region( dd, NULL, region, UT_TRUE );
    retval = dmSUCCESS;
  }

  return retval;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_get_ncptse                         */
/*                                                             */
/* Description:                                                */
/*             Return number of components                     */
/*                                                             */
/*  Input parameters:                                          */
/*              dmSubspace   *subspace                         */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            int    numberOfComponents                        */
/*                                                             */
/***************************************************************/
long dmi_subspace_get_ncpts( dmSubspace* subspace )
{
  if ( subspace )
    return subspace->ncpts;
  else
    return 0;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_get_ncols                          */
/*                                                             */
/* Description:                                                */
/*             Return number of columns                        */
/*                                                             */
/*  Input parameters:                                          */
/*              dmSubspace   *subspace                         */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            int    numberOfColumns                           */
/*                                                             */
/***************************************************************/
long dmi_subspace_get_ncols( dmSubspace* subspace )
{
  if ( subspace )
    return subspace->ncols;

 return 0;
}




/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_alloc                              */
/*                                                             */
/* Description:                                                */
/*             Allocate a subspace structure and initialize it */
/*                                                             */
/*  Input parameters:                                          */
/*              long  ncols                                    */
/*              long  cpts                                     */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            dmSubspace  *subspace                            */
/*                                                             */
/***************************************************************/
dmSubspace* dmi_subspace_alloc( long ncols, long ncpts )
{
  dmSubspace *subspace = NULL;

  subspace = (dmSubspace*)calloc( 1, sizeof( dmSubspace ));

  if ( ncols > 0 )
    subspace->cols = (dmDescriptor**)calloc( ncols, UT_SZ_P );
  else
    subspace->cols = NULL;

  subspace->ncpts   = ncpts;
  subspace->ncols   =  0;
  subspace->maxcpts = ncpts;
  subspace->maxcols =  ncols;
  subspace->cpt     = 1;
  subspace->invert  = 0;
  subspace->cache =  NULL;

  return subspace;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_add_col                            */
/*                                                             */
/* Description:                                                */
/*             Add column to subspace                          */
/*                                                             */
/*  Input parameters:                                          */
/*              dmSubspace   *subspace                         */
/*              dmDescriptor *dd                               */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_subspace_add_col( dmSubspace* subspace, dmDescriptor* dd )
{
  dmSubspaceData *ssdata = NULL;

  if ( !subspace )
   return;

  subspace->ncols++;

  if ( subspace->ncols >= subspace->maxcols )
  {
    subspace->maxcols = 2 * ( subspace->maxcols + 10 );
    subspace->cols = (dmDescriptor**)realloc( subspace->cols, subspace->maxcols * UT_SZ_P );
  }

  subspace->cols[subspace->ncols-1] = dd; 
  dd->no = subspace->ncols;
  ssdata = dmi_ss_get_data( dd );
  ssdata->subspace = subspace;

  ssdata->filters = dmi_filter_alloc_array( NULL, dd->dim, 0, subspace->maxcpts );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_ss_get_data                                 */
/*                                                             */
/* Description:                                                */
/*             Get subspace data given DM  descriptor          */
/*                                                             */
/*  Input parameters:                                          */
/*              dmDescriptor *dd                               */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            dmSubspaceData  *data                            */
/*                                                             */
/***************************************************************/
dmSubspaceData* dmi_ss_get_data( dmDescriptor* dd )
{
  dmSubspaceData *ssdata = NULL;

  if ( dd->descType != dmSUBSPACE )
  {
    sprintf( dm_errbuf, "Attempt to access subspace data for non subspace descriptor %s",
             dd->name );
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf ); 
    return NULL;
  }

  ssdata = dmi_descriptor_get_aux( dd );

  if ( !ssdata ) 
  {
    ssdata = calloc( 1, sizeof( dmSubspaceData )); 
    ssdata->auxtype = DM_AUX_SS;
    dmi_descriptor_set_aux( dd, ssdata );
  }

  return ssdata;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_add_cpt                            */
/*                                                             */
/* Description:                                                */
/*             Add a component to the subspace                 */
/*                                                             */
/*  Input parameters:                                          */
/*              dmSubspace   *subspace                         */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_subspace_add_cpt( dmSubspace* subspace )
{
  long  colOffset = 0;
  long  start     = 0;

  dmDescriptor   *dd     = NULL;
  dmSubspaceData *ssdata = NULL;

  subspace->ncpts++;

  if ( subspace->ncpts >= subspace->maxcpts )
  {
    start = subspace->maxcpts;
    subspace->maxcpts = 10 * ( subspace->maxcpts + 1 );

    for ( colOffset = 0; colOffset < subspace->ncols; colOffset++ )
    {
      dd = subspace->cols[colOffset];
      ssdata = dmi_ss_get_data( dd );
      ssdata->filters = dmi_filter_alloc_array( ssdata->filters, dd->dim, start, subspace->maxcpts );
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_free_cache                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmSubspace   *subspace                         */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_subspace_free_cache( dmSubspace* subspace )
{
  dmRTSCache *cache = NULL;
  long        offset = 0;

  if ( subspace )
  {
    cache = subspace->cache;

    if ( cache )
    {
      if ( cache->ddlist ) free( cache->ddlist );
      if ( cache->sub_ddlist ) free( cache->sub_ddlist );
      if ( cache->col_offsets ) free( cache->col_offsets );
      if ( cache->accept ) free( cache->accept );
      if ( cache->filter_list )
      {
        for ( offset = 0; offset < cache->n_filt; offset++ )
        {
          free( cache->filter_list[offset] );
        }
        free( cache->filter_list );
      }
      free( cache );
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_free                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmSubspace   *subspace                         */
/*              void         (*kernel_free)(dmDescriptor)      */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_subspace_free( dmSubspace* subspace, void (*kernel_free)( dmDescriptor* ) )
{
  dmDescriptor   *dd     = NULL;
  dmFilter      **flist  = NULL;
  dmSubspaceData *ssdata = NULL;

  long  colOffset = 0;

  if ( !subspace ) 
    return;

  for ( colOffset = 0; colOffset < subspace->ncols; colOffset++ )
  {
    dd = subspace->cols[colOffset];
    ssdata = dmi_descriptor_get_aux( dd );

    if ( ssdata )
    {
      flist = ssdata->filters;
      dmi_filterlist_free( flist, subspace->ncpts );
      free( ssdata->mins );
      free( ssdata->maxes );
      free( ssdata );
      dmi_descriptor_set_aux( dd, NULL );
    }

    dmi_descriptor_kernel_free( dd, kernel_free );
  }
  free( subspace->cols );

  /* Free real-time subspace info */
  dmi_subspace_free_cache( subspace );

  /* Free the subspace pointer */
  free( subspace );
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_filterlist_free                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmFilter  **flist                              */
/*              long        n                                  */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_filterlist_free( dmFilter** flist, long n )
{
  long cpt = 0;

  for ( cpt = 0; cpt < n; cpt++ )
  {   
    dmi_filter_free( flist[cpt] );
  }
  free( flist );
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_col_set_filter                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmDecriptor *dd                                */
/*              long         cpt                               */
/*              dmFilter    *f                                 */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
int dmi_subspace_col_set_filter( dmDescriptor* dd, long cpt, dmFilter* f )
{
  dmFilter       **flist  = NULL;
  dmSubspaceData  *ssdata = NULL;

  if ( cpt < 1 )
    return 0;

  ssdata = dmi_ss_get_data( dd );

  if ( ssdata )
    flist = ssdata->filters;

  if ( !flist )
    return 0;

  flist[cpt-1] = f;

  return 1;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_col_get_filter                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmDecriptor *dd                                */
/*              long         cpt                               */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            dmFilter  *filter                                */
/*                                                             */
/***************************************************************/
dmFilter* dmi_subspace_col_get_filter( dmDescriptor* dd, long cpt )
{
  dmFilter       **flist    = NULL;
  dmFilter        *f        =  NULL;
  dmSubspaceData  *ssdata   = NULL;
  dmSubspace      *subspace = NULL;

  if ( !dd )
    return NULL;

  ssdata = dmi_ss_get_data( dd );
  subspace = ssdata->subspace;

  /* Need to do this if filtering on newly forced vector column */
  if ( !subspace )
  {
    subspace = dd->block->subspace;
  }
 
  if ( cpt < 1 || cpt > subspace->ncpts)
    return NULL;

  if ( ssdata )
    flist = ssdata->filters;

  if ( flist )
    f = flist[cpt-1];
  else
    f = NULL;

  return f;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_col_reset_filter                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmDecriptor *dd                                */
/*              long         cpt                               */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            dmFilter  *filter                                */
/*                                                             */
/***************************************************************/
dmFilter* dmi_subspace_col_reset_filter( dmDescriptor* dd, long cpt )
{
  dmFilter        *f      = NULL;
  dmFilter       **flist  = NULL;
  dmSubspaceData  *ssdata = NULL;
  long             fdim   = 0;
  long             dim    = 0;  
 
  dim = dd->dim;
  ssdata = dmi_ss_get_data( dd );

  if ( ssdata )
    flist = ssdata->filters;

  if ( !flist )
    return NULL;

  f = flist[cpt-1];

  if ( f ) 
  {
    fdim = dmi_filter_get_dim( f );
  
    if ( dim != fdim )  
    {
#if 0
      fprintf( stderr, "DM INTERNAL ERROR: Filter dimension mismatch\n" );
#endif
      dmi_filter_clear( f );
      dmi_filter_init( f, dim, dd->type );
    }

  }
  else 
  {
#if 0
    fprintf( stderr, "DM INTERNAL ERROR: Filter not allocated\n");
#endif

    f = dmi_filter_alloc( dim, dd->type );

    flist[cpt-1] = f;
  }

  return f;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_block_get_subspace                          */
/*                                                             */
/* Description:                                                */
/*             Returns subspace pointer                        */
/*                                                             */
/*  Input parameters:                                          */
/*              const dmBlock *block                           */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            dmSubspace  *subspace                            */
/*                                                             */
/***************************************************************/
dmSubspace* dmi_block_get_subspace( const dmBlock* block )
{
  if ( block )
    return block->subspace;
  else
    return NULL;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_remove_col_cpt                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmSubspace  *subspace                          */
/*              long         i                                 */
/*              long         CptNo                             */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            int  status                                      */
/*                                                             */
/***************************************************************/
int dmi_subspace_remove_col_cpt( dmSubspace* subspace, long i, long CptNo  )
{
  dmFilter      **flist  = NULL;
  dmFilter       *filter = NULL;
  dmDescriptor   *dd     = NULL;
  dmSubspaceData *ssdata = NULL;

  long j = 0;

  /* Find the value group which corresponds to the current component */
  dd = subspace->cols[i];
  ssdata = dmi_ss_get_data( dd );

  if ( ssdata )
    flist = ssdata->filters;

  if ( flist )
    filter = flist[CptNo-1];

  if ( !filter )
    return dmFAILURE;

  dmi_filter_free( filter );
  flist[CptNo-1] = NULL;

  for ( j = CptNo; j < subspace->ncpts; j++ )
    flist[j-1] = flist[j];

  flist[subspace->ncpts-1] = NULL;
  dmi_descriptor_set_update(subspace->cols[i]);

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_collapse                           */
/*                                                             */
/* Description:                                                */
/*   dmPCollapseDSSCpt: See if we can combine the given        */
/*   component # with any other component in the block.  This  */
/*   is true if no more than one cell in the virtual DSS table */
/*   differs for a given compoent.  In "implementation lingo", */
/*   this means that for each value group in all the subspace  */
/*   columns for the input component #, it is collapsable if   */
/*   only one (or none) or its value groups differ from those  */
/*   of any other component. Algorithm takes a given component,*/
/*   compares it in order to all other components (except      */
/*   itself), and, for each "mergeable" compoent found it      */
/*   merges (OR's) it with the source component and removes it.*/
/*                                                             */
/*  Now modified (Jan 2000) to not collapse if any descriptor  */
/*  has named extension                                        */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock    *block                              */
/*              dmSubspace *subspace                           */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            int  status                                      */
/*                                                             */
/***************************************************************/
int dmi_subspace_collapse( dmBlock* block, dmSubspace* subspace )
{
  long cpt1  = 0;
  long cpt2  = 0;  
  long n     = 0;
  int remove = 0;


  if (  !subspace )
    return dmSUCCESS;

  /* Collapse starting from the top */ 
  n = subspace->ncpts;

  for( cpt1 = n; cpt1 > 1; cpt1-- )
  {
    remove = 0;

    /* Compare this component with all below it */
    for ( cpt2 = 1; cpt2 < cpt1 && !remove; cpt2++ )
      remove = dmi_subspace_collapse_cpt( block, subspace, cpt1, cpt2 );   
  }

  cpt1 = 1;
  cpt2 = 2;

  if ( subspace->ncpts > 1 )  /* Check for null cpt 1 */
    dmi_subspace_collapse_cpt( block, subspace, cpt1, cpt2 );

  return dmSUCCESS;
}




/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_collapse_cpt                       */
/*                                                             */
/* Description:                                                */
/*    The block argument is used only to delete disappearing   */
/*    header keys and may be null.                             */
/*    The current implementation gets rid of the header keys   */
/*    another way.                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock     *block                             */
/*              dmSubspace  *subspace                          */
/*              long         CptNo                             */
/*              long         CompareCptNo                      */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            int  status                                      */
/*                                                             */
/***************************************************************/
int dmi_subspace_collapse_cpt(dmBlock* block, dmSubspace* subspace,
                              long CptNo, long CompareCptNo )
{
  double fieldx[2];
  double fieldy[2];


  dmFilter     *CptFilter          = NULL;  /* Value group of source component  */
  dmFilter     *CompareCptFilter   = NULL;  /* Value group of component to compare to */
  dmFilter     *SavedCptFilter     = NULL;  /* Different source cpt value group  */
  dmFilter     *SavedCompareFilter = NULL;  /* Different compare cpt value group */
  dmDescriptor *dd                 = NULL;
  char         *ddname             = "UNK";

  long lNumDiffs = 0;      /* # of value groups discrepancies between cpts*/
  long colOffset = 0;
  int  null1     = 0;
  int  null2     = 0;
  int  can_combine   = 0;
  int CollapseAnyway = 0;
  int Collapse       = 0;

  dmBool match = dmFALSE;


  /* Check cpt # validity */
  if ( !block )
    { ; }

  /* Loop over subspace columns */
  for(colOffset=0; colOffset< subspace->ncols; colOffset++)
  {
    /* For each subspace column, see if the value group at lCptNo matches the
     * one at i */
    /* Get the value group for the source component */

    dd = subspace->cols[colOffset];
    ddname = dd->name;
    CompareCptFilter = dmi_subspace_col_get_filter( dd, CompareCptNo);

    /* Get the value group for the compare component */
    CptFilter = dmi_subspace_col_get_filter( dd, CptNo);

    if ( !CompareCptFilter || !CptFilter )
    {
     return 0;
    }

    /* 
     * If CptNo is 1 or 2, we are comparing 2 with 1. If 2 is null, we must collapse
     * it whether 1 is null or not. If CptNo is 1, 2 is nonnull and we delete 1.
     */
      
    null1 = dmi_filter_is_null( CompareCptFilter );
    null2 = dmi_filter_is_null( CptFilter );

    if ( null2 )
    {
      if ( null1 )
      {
        lNumDiffs = 1;  /* Force collapse */

        if ( dmverbose >= 5 )
          fprintf( stderr, "Null filter component %ld collapsed to component %ld\n",
                   CptNo, CompareCptNo );

        SavedCompareFilter = CompareCptFilter;
        SavedCptFilter = CptFilter;
      }
      else
      {
        lNumDiffs = 2;  /* Prevent collapse */      

        if ( CptNo <= CompareCptNo + 1 ) 
          CollapseAnyway = 1;
      }
      break;
    }
    else
      if ( null1 )
      {
        lNumDiffs = 2;  /* Prevent collapse */
        break;
      }
  }

  if ( lNumDiffs == 0 && CptNo > 1 )
    for(colOffset=0; colOffset< subspace->ncols; colOffset++)
    {
      dd = subspace->cols[colOffset];
      CompareCptFilter = dmi_subspace_col_get_filter( dd, CompareCptNo);

      /* Get the value group for the compare component */
      CptFilter = dmi_subspace_col_get_filter( dd, CptNo);

      /* need field to see if two filters with regions may combine */
      if ( CompareCptFilter->region && CptFilter->region )
      {
        dmi_descriptor_get_cpt_range( dd, 1, DT_REAL8, &fieldx[0], &fieldx[1] );
        dmi_descriptor_get_cpt_range( dd, 2, DT_REAL8, &fieldy[0], &fieldy[1] );
      }

      match = dmi_filter_compare_for_combine( CompareCptFilter, CptFilter, fieldx,
                                              fieldy, &can_combine );
    
      if( !match )
      {
        if ( can_combine )
        {
          lNumDiffs += 1;
          SavedCompareFilter = CompareCptFilter;
          SavedCptFilter = CptFilter;
        }
        else
          lNumDiffs = 2;
      }

      if (lNumDiffs > 1)
        break;
    }

    /* If we looped through each subspace column and found no more than one
     * discrepancy we can merge the components */

    if(lNumDiffs <2 && CptNo > 1 )
    {
      if ( dmverbose >= 5 && !null1 )
      {
        fprintf( stderr, "Component %ld collapsed to component %ld on var %s\n",
                CptNo, CompareCptNo, ddname );
        dmi_subspace_print_cpt( subspace, CptNo );
        dmi_subspace_print_cpt( subspace, CompareCptNo );
      }

      dmi_filter_combine( SavedCompareFilter, SavedCptFilter );

      if ( dmverbose >= 5 && !null1  )
      {
        fprintf( stderr, "Post combine \n" );
        dmi_subspace_print_cpt( subspace, CompareCptNo );
      }

      dmi_subspace_combine_tabnames( subspace, CptNo, CompareCptNo );
      /* Remove component which was just merged with source component */
      Collapse = 1;
    }

   if ( Collapse || CollapseAnyway )
   {
      dmi_subspace_remove_cpt( subspace, CptNo );
      return 1;
   }

  return 0;
}




/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_remove_cpt                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmSubspace  *subspace                          */
/*              long         CptNo                             */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            int  status                                      */
/*                                                             */
/***************************************************************/
int dmi_subspace_remove_cpt( dmSubspace* subspace, long CptNo )
{
  long  i = 0;

  /* If no subspace to remove components from, no work to do */
  if (!subspace)
    return dmSUCCESS;

  if (CptNo<1 || CptNo> subspace->ncpts)
  {
    sprintf( dm_errbuf, "Bad subspace component %ld", CptNo );
    dmu_err_set( DM_ERR_SUBSPACE, dm_errbuf );
    return dmFAILURE;
  }   

  /* Delete component from the subspace structure for each sscol */

  for (i=0; i< subspace->ncols; i++)
    dmi_subspace_remove_col_cpt( subspace, i, CptNo );

  /* Update total number of components in block  and current cpt no */
  if ( subspace->ncpts > 1 ) 
    subspace->ncpts--;

  if ( subspace->cpt >= CptNo && subspace->cpt > 1 ) 
    subspace->cpt--;

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_combine_tabnames                   */
/*                                                             */
/* Description:                                                */
/*   Added by JCM 1998 Sep: if the disappearing component has  */
/*   a name and the surviving ones dont, we copy the name over */
/*                                                             */
/*  Input parameters:                                          */
/*              dmSubspace  *subspace                          */
/*              long         CptNo                             */
/*              long         lCptNo                            */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_subspace_combine_tabnames( dmSubspace* subspace, long CptNo,
                                    long lCptNo )
{
  long            colOffset = 0;
  dmFilter      **filters   = NULL; 
  dmDescriptor   *dd        = NULL;
  dmSubspaceData *ssdata    = NULL;

  for(colOffset=0; colOffset<subspace->ncols; colOffset++)
  {
    dd = subspace->cols[colOffset];
    ssdata = dmi_ss_get_data( dd );
    filters = ssdata->filters;

    if ( dd->dim == 1 && filters && *filters )
      dmi_filter_copy_tabname( filters[CptNo-1], filters[lCptNo-1] );
  }
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_col_no                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*        const dmSubspace  *subspace                          */
/*        const long         no                                */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor *dd                                 */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_subspace_col_no(      const dmSubspace* subspace, const long no )
{
  if ( !subspace )
    return NULL;

  if ( no < 1 || no > subspace->ncpts )
    return NULL;

  return subspace->cols[no-1];
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_col_open                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*        const dmSubspace  *subspace                          */
/*        const char        *name                              */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor *dd                                 */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_subspace_col_open( const dmSubspace* subspace,
                                     const char* name )
{
  dmDescriptor *dd = NULL;

  if ( subspace )
    dd = dmi_descriptor_list_search_vector( (const dmDescriptor**)subspace->cols,
                                             subspace->ncols, name );

  return dd;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_block_copy_subspace                         */
/*                                                             */
/* Description:                                                */
/*    Copy the DSS from one block into the DSS structure of    */
/*    another any make any runtime filter of the original DSS  */
/*    part of the permanent data subspace. The copied DSS is   */
/*    ANDed with anny existing DSS in the block.               */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock  *srcBlock                             */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*              dmBlock  *destBlock                            */
/*                                                             */
/*  Returns:                                                   */
/*              int  status                                    */
/*                                                             */
/***************************************************************/
int dmi_block_copy_subspace(dmBlock *srcBlock, dmBlock *destBlock)
{
  dmBool copyDefault = UT_TRUE;
  return dmi_block_apply_subspace( srcBlock, destBlock, copyDefault );
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_block_apply_subspace                        */
/*                                                             */
/* Description:                                                */
/*   Apply the permanent portion of the subspace to the input  */
/*   block, ANDing it with the existing subspace if necessary. */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock  *srcBlock                             */
/*              dmBool    copyDefault                          */
/*                                                             */
/*  Output parameters:                                         */
/*              dmBlock  *destBlock                            */
/*                                                             */
/*  Returns:                                                   */
/*              int  status                                    */
/*                                                             */
/***************************************************************/
int dmi_block_apply_subspace(dmBlock *srcBlock, dmBlock *destBlock,
                             dmBool copyDefault )
{
  dmBool requireTable = UT_FALSE;

  dmSubspace  *srcSubspace  = NULL;
  dmSubspace  *destSubspace = NULL;

  /* Apply source subspace to destination block */
  /* Get array of DSS column descriptors */
  if ( !srcBlock || !destBlock ) 
  {
    dmu_err_set( DM_ERR_NULL, "No block" );
    return dmFAILURE;
  }

  if ( dmverbose >= 4 )
    fprintf( stderr, "DM Subspace: Apply %s %s to %s %s\n", 
             srcBlock->ds->name, srcBlock->name,
             destBlock->ds->name, destBlock->name );


  if ( !srcBlock->subspace ) /* Empty subspace to apply */
    return dmSUCCESS;

  srcSubspace = srcBlock->subspace;

  if ( !destBlock->subspace )
    destBlock->subspace = dmi_subspace_alloc( srcSubspace->ncols, srcSubspace->ncpts );

  destSubspace = destBlock->subspace;
 
  return dmi_subspace_apply( srcSubspace, destSubspace, destBlock, copyDefault, requireTable );
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_apply                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmSubspace   *srcSubspace                      */
/*              dmBool        copyDefault                      */
/*              dmBool        requireTable                     */
/*                                                             */
/*  Output parameters:                                         */
/*              dmSubspace   *destSubspace                     */
/*              dmBlock      *destBlock                        */
/*                                                             */
/*  Returns:                                                   */
/*              int  status                                    */
/*                                                             */
/***************************************************************/
int dmi_subspace_apply( dmSubspace* srcSubspace, dmSubspace* destSubspace,
                        dmBlock* destBlock, dmBool copyDefault, dmBool requireTable )
{
  long  destCurrCpt   = 0;
  long  srcCurrCpt    = 0;
  long  destCpt       = 1;
  long  srcCpt        = 0;
  long  destNumCpts   = 0;
  long  srcNumCpts    = 0;
  long  destCptOffset = 0;
  int   retval        = dmSUCCESS;
  int   status        = 0;

  /* Do we have a non-trivial DSS List in the DMFS? */
  if ( !srcSubspace || srcSubspace->ncols == 0 )
    return retval;

  destSubspace->invert = srcSubspace->invert;

  /*  Save current components */
  srcCurrCpt  = srcSubspace->cpt;
  destCurrCpt = destSubspace->cpt;

  /*  For each component in source and destination */
  srcNumCpts  = srcSubspace->ncpts;
  destNumCpts = destSubspace->ncpts;

  /* Clone the dest cpts BEFORE we alter them! */
  if ( dmverbose >= 5 )
    fprintf( stderr, "Reduce product subspace from %ld components\n",
             srcNumCpts * destNumCpts );

  /* Create clones of each component of the existing (dest) subspace.  */
  /* Generate a clone for each component of the new (src) subspace and */
  /* arrange so all instances of a component are grouped together..    */
  /* this will preserve the sequence of components as much as possible */

  for ( destCpt = 1; destCpt <= destNumCpts; destCpt++ )
  {
    for ( srcCpt = 1; srcCpt <= srcNumCpts; srcCpt++ )
    {
      /* Create new subspace component as a clone from the current one */
      dmi_subspace_clone_cpt( destSubspace, destCpt );
    }
  }

  /* Remove the original set of dest. subspace.. leaving the grouped clones */
  for ( destCpt = 1; destCpt <= destNumCpts; destCpt++ )
  {
    dmi_subspace_remove_cpt( destSubspace, 1 );
  }


  /* Intersect components */ 
  for( destCpt = 1; destCpt <= destNumCpts; destCpt++ )
  {
     if ( dmverbose >= 5 )
       fprintf( stderr, "Process subspace component %ld\n", destCpt );

     destCptOffset = ( destCpt - 1 ) * srcNumCpts;

     for ( srcCpt = 1; srcCpt <= srcNumCpts; srcCpt++ ) 
     {
       srcSubspace->cpt = srcCpt;
       destSubspace->cpt = srcCpt + destCptOffset;

       if ( destSubspace->cpt > destSubspace->ncpts )
         fprintf( stderr, "DM INTERNAL ERROR IN APPLY SUBSPACE: CPT (%ld %ld) %ld NCPTS %ld S=%ld D=%ld\n", 
                  destCpt, srcCpt,  destSubspace->cpt, destSubspace->ncpts, srcNumCpts, destNumCpts );

       /* If this intersection is null, return dmSUCCESS and continue, but deletes */
       /*  the - possibly just created - component                                 */
       status = dmi_subspace_intersect_cpt( destBlock, destSubspace, srcSubspace->cols,
                                           srcSubspace->ncols, copyDefault, requireTable );

       if ( !status )
       { 
         fprintf( stderr, "DM INTERNAL ERROR: SUBSPACE INTERSECT FAIL\n" );
 	 retval = dmFAILURE;
 	 goto exit;
       }
     }
   }

#if 1
   if ( dmverbose >= 6 )
#endif
   {   
     dmi_subspace_print( destSubspace );
   }

   if ( !dmi_subspace_collapse( destBlock, destSubspace ))
     return dmFAILURE;

   if ( dmverbose >= 6 )
   {   
     dmi_subspace_print( destSubspace );
   }

exit:  
  srcSubspace->cpt = srcCurrCpt;
  destSubspace->cpt = destCurrCpt;

 return retval;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_clone_cpt                          */
/*                                                             */
/* Description:                                                */
/*   Make a new subspace component which is a copy of another  */
/*   one. This routine creates a cpt, which becomes the current*/
/*   one, and loops through the descriptors copying value      */
/*   groups.                                                   */
/*                                                             */
/*  Input parameters:                                          */
/*              dmSubspace   *subspace                         */
/*              long          CptNo                            */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*              int  status                                    */
/*                                                             */
/***************************************************************/
static int dmi_subspace_clone_cpt( dmSubspace* subspace, long CptNo )
{
  dmFilter     *filter = NULL;
  dmDescriptor *dd     = NULL;
  long          i      = 0;
  int           retval = dmFAILURE;
  int           empty  = 1;

  if (!subspace )
    return retval;

  if (!dmi_subspace_create_cpt( subspace, empty ))
    return retval;

  /* This alloc is within the component loop since new descriptors */
  /*   may have been added on the last iteration                   */ 
  for ( i = 0; i < subspace->ncols; i++ )
  {
    dd = subspace->cols[i];
    filter = dmi_subspace_col_reset_filter( dd, CptNo );

    if ( !filter )
     goto error;

    if ( !dmi_subspace_col_update_from_filter( dd, filter ) )
      goto error;
  }  

  retval = dmSUCCESS;

error:
  return retval;
}




/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_intersect_cpt                      */
/*                                                             */
/* Description:                                                */
/*   Intersect current DSS cpt of source with current DSS cpt  */
/*   of  destination.                                          */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock       *destBlock                       */
/*              dmSubspace    *subspace                        */
/*              dmDescriptor **ddlist                          */
/*              long           numDescs                        */
/*              dmBool         copyDefault                     */
/*              dmBool         requireTable                    */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*              int  status                                    */
/*                                                             */
/***************************************************************/
int dmi_subspace_intersect_cpt( dmBlock* destBlock, dmSubspace* subspace,
                                dmDescriptor** ddlist, long numDescs,
                                dmBool copyDefault, dmBool requireTable )
{
  long i      = 0;
  int  status = 0;
 
  /* Add each DSS column to the destination block */
  for (i=0; i<numDescs; i++)
  {
    status = dmi_subspace_col_intersect(destBlock, subspace, ddlist[i], copyDefault, requireTable );
    if (!status )
      return dmFAILURE;
  }

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_intersect_cpt                      */
/*                                                             */
/* Description:                                                */
/*    Add the input column descriptor to the input block,      */
/*    ANDing it with a subspace column on the same variable in */
/*    the block if one exists.                                 */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock       *destBlock                       */
/*              dmSubspace    *subspace                        */
/*              dmDescriptor  *srcDesc                         */
/*              dmBool         copyDefault                     */
/*              dmBool         requireTable                    */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*              int  status                                    */
/*                                                             */
/***************************************************************/
int dmi_subspace_col_intersect(dmBlock* destBlock, dmSubspace* subspace,
                               dmDescriptor* srcDesc, dmBool copyDefault,
                               dmBool requireTable )
{
  dmDescriptor *destDesc = NULL;  /* Descriptor for current DSS column       */
  dmFilter     *filter   = NULL;  /* Value group for data in etList          */
  dmBlock      *srcBlock = NULL;
  dmDescriptor *colDesc  = NULL;

  if ( !srcDesc || !destBlock ) 
    return dmFAILURE;

  srcBlock = (dmBlock*)srcDesc->block;

  /* If datatypes of DSS are different but both numeric, should write code to */
  /* force them to be the same by promoting the smaller type. We don't do this yet.*/

  /* Get the vals for a component of the srcDesc */
  filter =  dmi_subspace_col_get_current_filter( srcDesc );

  /* Enable skipping the filter if it is default */
  /*  May be a problem with multiple cpts here   */

  colDesc = dmi_descriptor_list_search_vector( (const dmDescriptor**)srcBlock->cols, srcBlock->ncols, srcDesc->name );

  if ( !copyDefault )
  {
    // NOTE: regarding Bug 9606
    //       if "filter" is NULL, the following function traps it

    if( dmi_filter_is_default( colDesc, filter ))
      return dmSUCCESS;  

    if ( destBlock && destBlock->type == dmTABLE )
    {
      destDesc = dmi_descriptor_list_search_vector( (const dmDescriptor**)destBlock->cols,
                                                    destBlock->ncols, srcDesc->name );

      /* Ignore filters on columns which are not present */
      if ( requireTable && !destDesc ) 
        return dmSUCCESS;

      if ( dmi_filter_is_default( destDesc, filter ))
        return dmSUCCESS;  
    }
    else
      if ( dmverbose >= 4 && !destBlock )
      {
        fprintf( stderr, "No dest block for %s\n", srcDesc->name );
      }
  }

  /* See if we need to intersect DSS col with one in block */
  destDesc = dmi_subspace_col_open( subspace, srcDesc->name);
  
  if( destDesc ) 
  {
    // Code Change Bug 9696 January 30, 2012
    //  Insure valid filter pointer from call above
    //  to dmi_subspace_col_get_current_filter()

    if ( !filter || !dmi_subspace_col_update_from_filter(destDesc, filter ) )
    {
       printf( "DM INTERNAL ERROR: Subspace update failed for %s\n", destDesc->name );
       return dmFAILURE;
    }
  }
  else
  {
    /* Creates a destDesc. Sets vals for the current destBlock component. */
    /*  Sets vals for other destBlock components to empty.                */
    /* Parent? Must be from the correct block! */
    if ( destBlock != srcBlock )
    {
      colDesc = dmi_descriptor_list_search_vector( (const dmDescriptor**)destBlock->rawcols, destBlock->nrcols, srcDesc->name );
    }

    if ( !colDesc  )
      colDesc = srcDesc;

    destDesc = dmi_subspace_col_create( destBlock, subspace, srcDesc->name,
                      srcDesc->cptNames, srcDesc->dim, 
                      srcDesc->type, srcDesc->unit, colDesc, UT_TRUE );

    /* Catch the case where blocks are different and we had to set colDesc=srcDesc above;*/
    /* mustn't point the parent to a different block */

    if ( colDesc && colDesc->block != destBlock )
      destDesc->parent = NULL;

    // Code Change Bug 9696 January 30, 2012
    //  Insure valid filter pointer from call above
    //  to dmi_subspace_col_get_current_filter()
 
    if ( destDesc && filter )
      dmi_subspace_col_init_filters( subspace, destDesc, filter );
    else
      return dmFAILURE;
  }

  dmi_descriptor_set_update( destDesc );
  return dmSUCCESS;
}




/***************************************************************/
/*                                                             */
/* Function:   dmi_block_merge_subspace                        */
/*                                                             */
/* Description:                                                */
/*    Merge (logical OR) the subspaces of two blocks, appending*/
/*    the subspace of block1 to the subspace of block0 and     */
/*    collapsing it so that the minimum # of components are    */
/*    used to represent the subspace.                          */
/*                                                             */
/*    dmPCombineDSS: Take the data subspace from block1 and    */
/*    tack it onto the end of the datasubspace for block0.  If */
/*    there exists a subspace column in block1 which is not in */
/*    block0, create the subspace column in block0 and then put*/
/*    the subspace column from block1 on the end of this new   */
/*     column.                                                 */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock       *block0                          */
/*              dmBlock       *block1                          */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*              int  status                                    */
/*                                                             */
/***************************************************************/
int dmi_block_merge_subspace(dmBlock* block0, dmBlock* block1)
{
  dmDescriptor *dmDSS     = NULL;   /* Subspace column descriptor */
  dmDescriptor *dd        = NULL;
  dmSubspace   *subspace0 = NULL;
  dmSubspace   *subspace1 = NULL;
  dmFilter     *filter0   = NULL;
  dmFilter     *filter1   = NULL;

  long  lTotCpts   = 0;     /* Total # of components in both blocks */
  long  lNum0Cpts  = 0;     /* # of initial components in block0    */
  long  lNum1Cpts  = 0;     /* # of initial components in block1    */
  long  lNumDescs  = 0;     /* # of subspace columns in block1      */
  long  i          = 0;     /* Loop counter                         */
  long  cptNo      = 0;
  long  newCptNo   = 0;
  long  dd_no      = 0;
  long  vector_cpt = 0;
  int   empty      = 1;


  subspace0 = block0->subspace;
  subspace1 = block1->subspace;

  /* Determine the size of each DSS */
  lNum0Cpts = subspace0->ncpts;
  lNum1Cpts = subspace1->ncpts;
  lTotCpts = (lNum0Cpts + lNum1Cpts);

  /* Create enough components in block0 to hold both subspaces */
  while( subspace0->ncpts < lTotCpts)
    dmi_subspace_create_cpt( subspace0, empty );


  lNumDescs = block1->subspace->ncols;
  for (i=0; i < lNumDescs; i++)
  {
    dd = block1->subspace->cols[i];

    /* If a matching subspace column exists in block0, append to it. */  
    /* Otherwise create a new subspace column in block0              */
    if (!(dmDSS = dmi_subspace_col_open(block0->subspace, dd->name)))
    {
      dmDSS = dmi_subspace_col_create( block0, NULL, dd->name, dd->cptNames, dd->dim, 
                                       dd->type, dd->unit, dd, UT_TRUE );

      if ( !dmDSS )
        return dmFAILURE;
    }

    if ( dmDSS->cptno > 0 )
    {
      vector_cpt = dmDSS->cptno;
      dmDSS = dmDSS->parent;
    }
    else
      vector_cpt = 0;

    dd_no = dmDSS->no;

    if ( dd_no == 0) 
      printf( "Error\n" );

    for( cptNo = 1; cptNo <= lNum1Cpts; cptNo++ )
    {
      newCptNo = cptNo + lNum0Cpts;
      filter1 = dmi_subspace_col_reset_filter( dd, cptNo );
      filter0 = dmi_subspace_col_reset_filter( dmDSS, newCptNo );
      dmi_filter_copy_element( filter1, filter0, vector_cpt );
    }

    dmi_descriptor_set_update( dmDSS );
  }

  if ( !dmi_subspace_collapse( block0, block0->subspace ))
    return dmFAILURE;

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_print                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmSubspace  *subspace                          */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*              int  status                                    */
/*                                                             */
/***************************************************************/
int dmi_subspace_print( dmSubspace* subspace )
{
  long current = 0;
  long ncpts   = 0;
  long cptNo   = 0;

  if (!subspace)
  {
    printf( "Empty\n" );
    printf( "------------------------------------------\n" );  
   return dmFAILURE;
  }

  current = subspace->cpt;
  ncpts   = subspace->ncpts;
 
  printf( "%ld Subspace Components (Current Cpt = %ld)\n", ncpts, current );

  if ( ncpts == 0 )
    ncpts = 1;

  printf( "%ld Subspace Columns\n\n", subspace->ncols );

  for ( cptNo = 1; cptNo <= ncpts; cptNo++ ) 
  {
    dmi_subspace_print_cpt( subspace, cptNo );
  }

  printf( "------------\n" );

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_print_cpt                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmSubspace  *subspace                          */
/*              long         cptNo                             */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*              nothing                                        */
/*                                                             */
/***************************************************************/
void dmi_subspace_print_cpt( dmSubspace* subspace, long cptNo )
{
  char name[80];
  long i = 0;
  dmDescriptor *dd = NULL;
  dmFilter     *filter = NULL;

  printf( "--- Component %ld:\n", cptNo );

  for(i=0; i<subspace->ncols; i++)
  {
    dd = subspace->cols[i];
    filter = dmi_subspace_col_get_filter( dd, cptNo );
    strcpy( name, dd->name );
    dmi_filter_print( name, filter );
  }
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_compose_total_name                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmDescriptor *dd                               */
/*              char         *buf                              */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*              nothing                                        */
/*                                                             */
/***************************************************************/
void dmi_compose_total_name( dmDescriptor* dd, char* buf )
{
  if ( dd->dim == 2 )
    sprintf( buf, "%s(%s,%s)", dd->name, dd->cptNames[0], dd->cptNames[1] );
  else
    cs_copy( dd->name, buf );
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_col_extremes                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmSubspace   *subspace                         */
/*              dmDescriptor *dd                               */
/*              long          cptno                            */
/*              dmDataType    type                             */
/*              double       *vals                             */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*              nothing                                        */
/*                                                             */
/***************************************************************/
void dmi_subspace_col_extremes( dmSubspace* subspace, dmDescriptor* dd,
                                long cptno,  dmDataType type, double* vals )
{
  long      ncpts  = subspace->ncpts;
  long      subcpt = 0;
  dmFilter *filter = NULL;

  vals[0] = -DT_REAL8_MAX;
  vals[1] =  DT_REAL8_MAX;

  for ( subcpt = 1; subcpt <= ncpts; subcpt++ )
  {
    filter = dmi_subspace_col_get_filter( dd, subcpt );
    dmi_filter_extremes( filter, cptno, type, vals );
  }
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_column_extremes                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock      *pB1                              */
/*              dmDescriptor *col                              */
/*              double       *vals                             */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*              nothing                                        */
/*                                                             */
/***************************************************************/
void dmi_column_extremes(dmBlock *pBl, dmDescriptor* col, double* vals )
{
  double tlmin = 0;
  double tlmax = 0;
  long   cptno = 0;
 
  dmDescriptor *dd = dmi_column_get_ss( pBl, col, &cptno );

  dmi_subspace_col_extremes( pBl->subspace, dd, cptno, col->type, vals );

  /* need to intersect with the column's range */
  if ( dmi_descriptor_get_cpt_range( col, 1, DT_REAL8, &tlmin, &tlmax ))
  {
    vals[0] = tlmin <= vals[0] ? vals[0] : tlmin;
    vals[1] = tlmax >= vals[1] ? vals[1] : tlmax;
  }
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_column_get_ss                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock      *pB1                              */
/*              dmDescriptor *col                              */
/*              long         *cpt                              */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*             dmDescriptor *dd                                */
/*                                                             */
/***************************************************************/
dmDescriptor* dmi_column_get_ss( dmBlock* pBl, dmDescriptor* col, long* cpt )
{
  dmDescriptor *dssDesc = NULL;
  dmFilter     *filter  = NULL;
  dmDescriptor *dd      = NULL;
  long          cptno   = 1;

  dd = col;

  if ( col->cptno > 0 )
  {
    dd = col->parent;
    cptno = col->cptno;
  }

  dssDesc =dmi_subspace_col_open(pBl->subspace, dd->name);

  /* This is a hack :(                                                     */
  /* If this is a temporary vector, its subspace columns might             */
  /* be the components and not the vector, so if we fail to get the vector */
  /* try getting the component.                                            */

  if ( !dssDesc && col->cptno > 0 )
  {
    dd = col;
    cptno = 1;
    dssDesc = dmi_subspace_col_open(pBl->subspace, dd->name);
  }

  /* If col is a virtual column and dmi_subspace_col_open() didn't find */
  /* its subspace col, it must not have one yet, so create one.         */

  if ( !dssDesc && dd->descType == dmCOORD )
  {
    /* If this is a component column, we want to add its parent */
    /* vector to the subspace                                   */
    if ( col->cptno > 0 )
    {
      dd = col->parent;
      cptno = col->cptno;
    }

    /* Need to check flags of both parent and column, since which one */
    /* gets set to DM_FLAG_RESULT depends on the filter/binspec       */
    if ( ( dd->flag & DM_FLAG_RESULT ) || ( col->flag & DM_FLAG_RESULT ) )
      dssDesc = dmi_subspace_col_init( pBl, NULL, dd, dd->type );
  }

  if (dssDesc )
  {
    if ( dssDesc->cptno > 0 )
    {
      cptno = dssDesc->cptno;
      dssDesc = dssDesc->parent;
    }
  
    filter = dmi_subspace_col_get_filter( dssDesc, 1 );    
  }

  *cpt = cptno;

  return dssDesc;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_block_set_subspace_cpt                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock   *block                               */
/*              long       cptnot                              */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*             nothing                                         */
/*                                                             */
/***************************************************************/
void dmi_block_set_subspace_cpt( dmBlock* block, long cptno )
{
  dmSubspace *subspace = 0;

  subspace = block->subspace;

  if ( subspace )
    subspace->cpt = cptno;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_map_filters_to_image_groups                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock   *block                               */
/*              long       cptnot                              */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*             nothing                                         */
/*                                                             */
/***************************************************************/
dmFilter** dmi_map_filters_to_image_groups( dmDescriptor* raw_image,
                                   dmSubspace* rts, int compact )
{
  dmDescriptor *p            = NULL;
  dmDescriptor *colp         = NULL;
  dmDescriptor *image_parent = NULL;
  dmFilter     *filter       = NULL;
  dmDescriptor *gp           = NULL;
  dmFilter    **filter_list  = NULL;

  long  i       = 0;
  long  vcpt    = 0;
  long  groupno = 0;
  long  j       = 0;
  long  ng      = 0;


  ng = raw_image->ngroups;
  filter_list = calloc( ng, UT_SZ_P );

  if ( compact && rts )
  {
    for ( i = 0; i < rts->ncols; i++ )
    {
      p = rts->cols[i];
      colp = p->parent;
   

      /* The following code, which matches on name, handles the case of binned images  */
      /* where colp is the original table column but now we want the image group descriptor */
      if ( colp->cptno > 0 )
      {
        vcpt = colp->cptno;
	colp = colp->parent;
      }

      /* Match on group name */
      groupno = 0;

      for ( j = 0; j < ng; j++ )
      {
        gp = raw_image->groups[j];
        gp = gp->coords[0];

        if ( cs_eq( gp->name, colp->name ))
        {
          colp = gp;

          groupno = j+1;
          break;
        }
      }

      if ( !groupno &&  ( colp->flag & DM_FLAG_AXIS ) )
      {
        dmi_descriptor_get_group_group_no( colp, &image_parent, &groupno, &vcpt );   

        if ( image_parent != raw_image )
        {
          sprintf( dm_errbuf, "DM internal error: Image filter parent is not correct image (%s)", colp->name );
          dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
        }
	else
        {
          gp = colp;
        }
      }

      if ( groupno )
      {
        filter = dmi_subspace_col_get_current_filter( p );
        filter_list[groupno-1] = filter;
      }
    }
  }

  return filter_list;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_is_null                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmSubspace *subspace                           */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*             int  TrueOrFalse                                */
/*                                                             */
/***************************************************************/
int dmi_subspace_is_null( dmSubspace* subspace )
{
  dmFilter     *filter = NULL;
  dmDescriptor *dd     = NULL;
  long  i     = 0;
  long  cptNo = 1;


  if (!subspace)
   return UT_TRUE;

  cptNo = 1;  /* If ss is null, it only has one cpt (by reduction) */

  for (i = 0; i < subspace->ncols; i++)
  {
    dd = subspace->cols[i];
    filter = dmi_subspace_col_get_filter( dd, cptNo );

    if ( dmi_filter_is_null( filter ) ) 
      return UT_TRUE;
  }

  return UT_FALSE;
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_rangelist_adjust_bounds                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock      *raw                              */
/*              dmRangeList  *range                            */
/*              dmDescriptor *dd                               */
/*              long          vcpt                             */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*             nothing                                         */
/*                                                             */
/***************************************************************/
void dmi_rangelist_adjust_bounds( dmBlock* raw, dmRangeList* range, dmDescriptor* dd, long vcpt )
{
  double vals[2];

  if ( !dd )
    return;

  dmi_subspace_col_extremes( raw->subspace, dd, vcpt, dd->type,  vals );
  dmi_rangelist_adjust( range, vals );
}




/***************************************************************/
/*                                                             */
/* Function:   dmi_image_subspace_set_bounds                   */
/*                                                             */
/* Description:                                                */
/*             Store image range                               */
/*                                                             */
/*  Input parameters:                                          */
/*              dmDescriptor *ss_imdata                        */
/*              long          axno                             */
/*              double        xmin                             */
/*              double        xmax                             */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*             nothing                                         */
/*                                                             */
/***************************************************************/
void  dmi_image_subspace_set_bounds( dmDescriptor* ss_imdata, long axno,
                                     double xmin, double xmax )
{
  dmSubspaceData *ssdata = NULL;
  dmDescriptor   *imdata = NULL;
  long            dim    = 0;

  ssdata = dmi_ss_get_data( ss_imdata );
  imdata = ss_imdata->parent;

  if ( imdata )
    dim = imdata->naxes;
  else
    dim = ss_imdata->naxes;

  if ( !ssdata->mins )
  {
    long i = 0;

    ssdata->mins = calloc( dim, UT_SZ_I );
    ssdata->maxes= calloc( dim, UT_SZ_I );

    for ( i = 0; i < dim; i++ )
    {
      ssdata->mins[i] = 1.0;
      ssdata->maxes[i] = imdata->axes[i];
    }
  }

  if ( axno > dim )
  {
    sprintf( dm_errbuf, "Incorrect axis number %ld for descriptor %s of dimension %ld", axno, ss_imdata->name, dim );
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
  }
  else
  {
    /* Round values to integers for image pixel subsection */
    /* Ignore -inf, inf which just means the user did not specify those bounds. */

    if ( xmin > -0.5 * DT_REAL8_MAX )
      ssdata->mins[axno-1] = (long) (xmin+0.5+DM_FILT_EPSILON);

    if ( xmax <  0.5 * DT_REAL8_MAX )
      ssdata->maxes[axno-1]= (long) (xmax+0.5-DM_FILT_EPSILON);
  }
}




/***************************************************************/
/*                                                             */
/* Function:   dmi_block_edit_subspace                         */
/*                                                             */
/* Description:                                                */
/*         taken from dmp_block_parse_filter()                 */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock  *block                            */
/*                  char     *comstr                           */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                  nothing                                    */
/*                                                             */
/***************************************************************/
void dmi_block_edit_subspace( dmBlock *block,  char *comstr)
{
  dmParseStack  *parseStack;
  dmClauseList  *clauseList = NULL;

  long   NumberOfClauses = 0;
  long   stackSize = 10;
  int    ii        = 0;
  int    retval    = 0; 

  // if no editing commands, return
  if ( !comstr || *comstr == '\0' ) 
    return;

  // initialize parse stack values
  parseStack = dmi_parse_stack_init( stackSize );

  // get command string into parse stack
  dmi_parse_clause_stack( comstr, parseStack);

  // parse commands into its parts
  retval = dmi_parse_clause_line( comstr, parseStack);

  // place each command clause as an entry into a list
  clauseList = dmi_parse_stack_eval(parseStack);


  NumberOfClauses = clauseList->n;


  // process each directive in the list

  for ( ii = 0; ii < NumberOfClauses; ii++)
  {
    dmi_parse_subspace_line( block, block->subspace, clauseList->clauses[ii] );
  }

  // Reduce subspace to minimal components
  dmi_subspace_collapse(block, block->subspace);


  // Trans
  // clean up
  dmi_parse_stack_free( parseStack );
  dmi_clauselist_free( clauseList );

  return;
}




/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_col_reset                          */
/*                                                             */
/* Description:                                                */
/*         Resets the subspace filters for the specified       */
/*         component range to the default values.  This should */
/*         be either the datatype range or, if a corresponding */
/*         column descriptor can be found, the TLMIN:TLMAX     */
/*         range.                                              */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock    *block                          */
/*                  dmSubspace *pSubspace                      */
/*                  char       *columnName                     */
/*                  long        minCpt                         */
/*                  long        maxCpt                         */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                  nothing                                    */
/*                                                             */
/***************************************************************/
void dmi_subspace_col_reset( dmBlock *block, dmSubspace *pInSubspace,
                        char *columnName, long min, long max)
{
  dmSubspace   *pSubspace     = pInSubspace;
  dmDescriptor *pSubspaceDesc = NULL;
  dmFilter     *pFilter       = NULL;

  double  dMinCpt[MAX_DCPTS];
  double  dMaxCpt[MAX_DCPTS];
  void   *vMinCpt;
  void   *vMaxCpt;

  long  minCpt  = min;
  long  maxCpt  = max;
  long  NumberOfValues = 1;
  int   ii           = 0;
  int   componentNum = 0;
  int   remove       = 0;
  int   dirty        = 0;

  /* check for valid pointers and range */
  if (!pSubspace)
  {
    if ( !block->subspace )
      return;

    pSubspace = block->subspace;
  }


  if (min < 0 )
    minCpt = 1;

  if (max < 0 )
    maxCpt = pSubspace->ncpts;

  if (minCpt > pSubspace->ncpts || maxCpt > pSubspace->ncpts)
     return;

  /* initialize column range arrays */
  memset( dMinCpt, 0, UT_SZ_D * MAX_DCPTS );
  memset( dMaxCpt, 0, UT_SZ_D * MAX_DCPTS );

  /* get subspace column descriptor */
  pSubspaceDesc = dmi_subspace_col_open( pSubspace, columnName );

  if ( !pSubspaceDesc)
    return;

  /* Get descriptor column range (TLMIN:TLMAX) */
  for (ii = 0; ii < pSubspaceDesc->dim; ii++)
  {
    vMinCpt = (void*)&dMinCpt[ii];
    vMaxCpt = (void*)&dMaxCpt[ii];
    dmi_descriptor_get_cpt_range( pSubspaceDesc, (ii+1), DT_REAL8, vMinCpt, vMaxCpt );
  }

  for (componentNum = minCpt; componentNum <= maxCpt; componentNum++)
  {
    pFilter = dmi_subspace_col_get_filter( pSubspaceDesc, componentNum);

    if ( ! pFilter )
      pFilter = dmi_filter_alloc( pSubspaceDesc->dim, pSubspaceDesc->type );


    if ( pFilter && pFilter->region )
    {
      dmi_filter_update_region( pSubspaceDesc, NULL, pFilter, NULL, UT_FALSE );
    }

    // Code Change   April 6, 2010
    // for dmappend 
    //   Before updating filter for this column, need to clear the pFiler->name
    //   field so that filter is not associated with any table name

    if (pFilter->name)
    {  
      free (pFilter->name);
      pFilter->name = NULL;
    }

    // NB:  MAX_DCPTS is set to 10 to correspond to MAXCPTS that is defined
    //      in file ./dmi/dmdsspriv.c.  If pSubspaceDesc->dim > 10, this will 
    //      cause an out of bounds memory error.

    for ( ii = 0; ii < pSubspaceDesc->dim; ii++)
    {
      dmi_filter_update( pFilter, (ii+1), NULL, columnName, DT_REAL8, &dMinCpt[ii],
                         &dMaxCpt[ii], NumberOfValues, 0, FALSE, &remove, &dirty);
    }

  }  
   
}
