/*                                                                
**  Copyright (C) 2002-2004,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   execute.c                                                   */
/*                                                               */
/* The execute stack - used for compiling dynamically defined    */
/* (DMTCALC-style) columns.                                      */
/*  Right now only coordinate columns are supported, e.g.        */
/*  [cols dec]                                                   */
/* But what the heck, let's put the machinery in place to do     */
/* more general things.                                          */
/*                                                               */
/*****************************************************************/

#include "dmi.h"

/* Max number of arguments (input and output) to an execute macro */
#define DM_EXE_ARGMAX 4
#define DM_EXE_STACK_SIZE 20


/***********************************************************************/
/*                                                                     */
/*  Function:  dmi_execute_stack_alloc                                 */
/*                                                                     */
/*  Description:                                                       */
/*    Allocate an execute stack.                                       */
/*    Preallocate space for nmax functions, extend in chunks           */
/*    to minimize mallocs                                              */
/*                                                                     */
/*  Input Parameters:                                                  */
/*              long nmax - max size                                   */
/*                                                                     */
/*  Output Parameters:                                                 */
/*                none                                                 */
/*                                                                     */
/*  Returns:                                                           */
/*      dmExecute execute - Allocated execute stack or NULL if         */
/*                          allocate fails                             */
/*                                                                     */
/***********************************************************************/
dmExecute dmi_execute_stack_alloc( long nmax )
{
   dmExecute execute;

   execute = (dmExecute)calloc( 1, sizeof( struct dmExecute_s ));

   if ( execute != NULL )
   {
     execute->n    = 0;
     execute->nmax = nmax;  
     if ( nmax > 0 )
       execute->stack = (dmExecuteFunc*)calloc( nmax, UT_SZ_P );
   }

   return execute;
}


/***********************************************************************/
/*                                                                     */
/*  Function:  dmi_execute_stack_add                                   */
/*                                                                     */
/*  Description:                                                       */
/*    Add function to execute stack                                    */
/*                                                                     */
/*  Input Parameters:                                                  */
/*              dmExecute      execute                                 */
/*              dmExecuteFunc  func                                    */
/*                                                                     */
/*  Output Parameters:                                                 */
/*                none                                                 */
/*                                                                     */
/*  Returns:                                                           */
/*              void                                                   */
/*                                                                     */
/***********************************************************************/
void dmi_execute_stack_add( dmExecute execute, dmExecuteFunc func )
{
  if ( !execute ){ return; }

  if ( execute->n >= execute->nmax )
  {
    execute->nmax = 2 * execute->nmax + DM_EXE_STACK_SIZE;
    execute->stack = realloc( execute->stack, execute->nmax * UT_SZ_P );
  }
  execute->stack[execute->n++] = func;
}


/***********************************************************************/
/*                                                                     */
/*  Function:  dmi_execute_stack_free                                  */
/*                                                                     */
/*  Description:                                                       */
/*       Free execute stack                                            */
/*                                                                     */
/*  Input Parameters:                                                  */
/*              dmExecute      execute                                 */
/*                                                                     */
/*  Output Parameters:                                                 */
/*                none                                                 */
/*                                                                     */
/*  Returns:                                                           */
/*              void                                                   */
/*                                                                     */
/***********************************************************************/
void dmi_execute_stack_free( dmExecute execute )
{
  long ii = 0;

  if ( !execute ){ return; }

  for ( ii=0; ii < execute->n; ii++ )
  {
    dmi_execute_func_free( execute->stack[ii] );  
  }

  free( execute->stack );
  free( execute );
  execute = NULL;
}


/***********************************************************************/
/*                                                                     */
/*  Function:  dmi_execute_func_free                                   */
/*                                                                     */
/*  Description:                                                       */
/*      Free execute function                                          */
/*                                                                     */
/*  Input Parameters:                                                  */
/*              dmExecute      execute                                 */
/*                                                                     */
/*  Output Parameters:                                                 */
/*                none                                                 */
/*                                                                     */
/*  Returns:                                                           */
/*              void                                                   */
/*                                                                     */
/***********************************************************************/
void dmi_execute_func_free( dmExecuteFunc func )
{
  if ( !func ){ return; }

  free( func->args );
  free( func );
  func = NULL;
}


/***********************************************************************/
/*                                                                     */
/*  Function:  dmi_execute_func_alloc                                  */
/*                                                                     */
/*  Description:                                                       */
/*     Allocate a compiled function for the execute stack              */
/*                                                                     */
/*  Input Parameters:                                                  */
/*              long       id                                          */
/*              long      *args                                        */
/*              long       nargs                                       */
/*                                                                     */
/*  Output Parameters:                                                 */
/*                none                                                 */
/*                                                                     */
/*  Returns:                                                           */
/*           dmExecuteFunc  func - allocated execute func  or NULL     */
/*                                 of allocation fails                 */
/*                                                                     */
/***********************************************************************/
dmExecuteFunc dmi_execute_func_alloc( long id, long* args, long nargs )
{
  long ii = 0;
  dmExecuteFunc func;

  func = (dmExecuteFunc)calloc( 1, sizeof ( struct dmExecuteFunc_s ));
  if ( func != NULL )
  {
    func->id    = id;
    func->nargs = nargs;
    func->args  = calloc( nargs, UT_SZ_I );
    for ( ii=0; ii < nargs; ii++ )
    {
      func->args[ii] = args[ii];
    }
  }

  return func;
}


/***********************************************************************/
/*                                                                     */
/*  Function:  dmi_buffer_add_execute_func                             */
/*                                                                     */
/*  Description:                                                       */
/*     Here we wrap func_alloc and just add a new func to the given    */
/*     execute stack, creating it as needed.                           */
/*                                                                     */
/*     Used as:                                                        */
/*    dmi_buffer_add_execute_func( &buffer->execute, id, args, nargs );*/
/*                                                                     */ 
/*    This way we don't have to create either of the execute stacks    */
/*    unless they are needed                                           */
/*                                                                     */
/*  Input Parameters:                                                  */
/*              dmExecute  *executep                                   */
/*              long        id                                         */
/*              long       *args                                       */
/*              long        nargs                                      */
/*                                                                     */
/*  Output Parameters:                                                 */
/*                none                                                 */
/*                                                                     */
/*  Returns:                                                           */
/*           void                                                      */
/*                                                                     */
/***********************************************************************/
void dmi_buffer_add_execute_func( dmExecute* executep, long id, long* args, long nargs )
{
  dmExecuteFunc func;

  if ( !( *executep ) )
  {
    *executep = dmi_execute_stack_alloc( DM_EXE_STACK_SIZE );
  }
  func = dmi_execute_func_alloc( id, args, nargs );

  dmi_execute_stack_add( *executep, func );
}


/***********************************************************************/
/*                                                                     */
/*  Function:  dmi_buffer_execute_stack                                */
/*                                                                     */
/*  Description:                                                       */
/*     Evaluate the defined columns in a buffer nvrows long.           */
/*    If filtered is true, only evaluate the rows which pass filtering */
/*                                                                     */
/*  Input Parameters:                                                  */
/*              dmBuffer  *buffer                                      */
/*              dmExecute  execute                                     */
/*              long       nvrows                                      */
/*              int        filtered                                    */
/*                                                                     */
/*  Output Parameters:                                                 */
/*                none                                                 */
/*                                                                     */
/*  Returns:                                                           */
/*          void                                                       */
/*                                                                     */
/***********************************************************************/
void dmi_buffer_execute_stack( dmBuffer* buffer, dmExecute execute, long nvrows, int filtered )
{
  long ii = 0;
  dmExecuteFunc func;

  if ( !buffer || !execute ){ return;}

  for ( ii = 0; ii < execute->n; ii++ )
  {
    func = execute->stack[ii];
    dmi_buffer_execute_eval( buffer, func->id, func->args, 
			           func->nargs, nvrows, filtered );
  }
}


/***********************************************************************/
/*                                                                     */
/*  Function:  dmi_buffer_execute_eval                                 */
/*                                                                     */
/*  Description:                                                       */
/*     Execute a buffer function for each row in the buffer            */
/*                                                                     */
/*  Input Parameters:                                                  */
/*              dmBuffer  *buffer                                      */
/*              long       func                                        */
/*              long      *arga                                        */
/*              long       nargs                                       */
/*              long       nvrows                                      */
/*              int        filtered                                    */
/*                                                                     */
/*  Output Parameters:                                                 */
/*                none                                                 */
/*                                                                     */
/*  Returns:                                                           */
/*          void                                                       */
/*                                                                     */
/***********************************************************************/
void dmi_buffer_execute_eval( dmBuffer* buffer, long func, long* args, 
			      long nargs, long nvrows, int filtered )
{
  dmDescriptor* ddargs[DM_EXE_ARGMAX]; /* Descs. of each column             */
  void*         data[DM_EXE_ARGMAX];   /* ptr to col. data start            */
  long          size[DM_EXE_ARGMAX];   /* size of col. in bytes             */

  dmDescriptor *coord_dd = NULL;  /* Desc. of result                         */
  int is_cpt = UT_FALSE;      /* is this a component of a vector?           */

  char *bdata = NULL;         /* Ptr to buffer data start                   */
  char *iptr  = NULL;         /* Ptr to input value(s)                      */
  char *optr  = NULL;         /* Ptr for transformed/output value(s)        */
  char *cptr  = NULL;         /* Ptr to desired cpt of transformed vector   */
  char *tmp   = NULL;         /* Temporary space for transformed vector     */

  long ii          = 0;       /* counter                                    */
  long cpt_offset  = 0;
  long col_offset  = 0;       /* index of desired column                    */
  long vrow        = 0;       /* row counter                                */
  long minrow      = 0;       /* start of row loop                          */
  long maxrow      = 0;       /* end of row loop                            */
  long rrow_offset = 0;       /* Note that we use rrow_offset for both 
		 	         input and output since we are reading 
				 and writing to the same buffer.  Only 
				 lines which pass the filter are calculated. */


  /* Set pointer to buffer data */
  if ( buffer->in_vdata && filtered )
  {
    bdata = buffer->vdata;
  }
  else 
  {
    bdata = buffer->data;
  }
  tmp = NULL;

  /* will loop through rows in current buffer segment */
  minrow = buffer->rrow_seg1 - buffer->rrow1; 
  maxrow = minrow + nvrows;

  /* Load arrays for each column. */
  for ( ii = 0; ii < nargs; ii++ )
  {
    col_offset = args[ii];
  
    data[ii]  = (char*)bdata + dmi_buffer_offset( buffer, col_offset, 0 );
    ddargs[ii]= buffer->cols[col_offset];
    size[ii]  = buffer->sizes[col_offset];
  }

  switch( func )
  {
  case DM_EXEC_COORD:
    coord_dd = ddargs[1];

    is_cpt = ( coord_dd->cptno > 0 && coord_dd->parent );
    if ( is_cpt )
    {
      /* Result is a component of a vector.. transform the vector */
      /* and return the appropriate component                     */
      cpt_offset = coord_dd->cptno - 1;
      coord_dd   = coord_dd->parent;

      /* Allocate space for one set of the full transformed vector */
      tmp  = calloc( coord_dd->dim*size[1], UT_SZ_C );

      /* Set pointer to the desired output component */
      cptr = (char*)tmp + size[1] * cpt_offset;
    }
    for ( vrow = minrow; vrow < maxrow; vrow++ )
    {
      rrow_offset = filtered? buffer->index[vrow] : vrow;
      iptr = (char*)data[0] + size[0] * rrow_offset;  
      optr = (char*)data[1] + size[1] * rrow_offset;
      if ( is_cpt )
      {
	/* When calculating a component of a vector, write the vector to tmp
	   then copy the desired component to output */
	dmi_coord_calc( coord_dd, iptr, ddargs[0]->type, ddargs[1]->type, tmp );
	memcpy( optr, cptr, size[1] );
      }
      else 
      {
       dmi_coord_calc( coord_dd, iptr, ddargs[0]->type, ddargs[1]->type, optr );
      }
    }
    break;
  case DM_EXEC_ROW:

    for ( vrow = minrow; vrow < maxrow; vrow++ )
    {
      rrow_offset = filtered? buffer->index[vrow] : vrow;
      iptr = (char*)data[0] + size[0] * rrow_offset;  

      /* True raw offset - even if recopied to vdata */
      rrow_offset = ( buffer->in_vdata && filtered ) ? buffer->index[vrow] : vrow;
      ((long*)iptr)[0] = rrow_offset + buffer->rrow1;

    }
    break;
  default:
    printf( "FUNCTION NOT IMPLEMENTED\n" );
    break;

  } /* end switch */ 

  /* Clean up temporary memory if used */
  if ( tmp )
  {
    free(tmp);
    tmp = NULL;
  }

}


/***********************************************************************/
/*                                                                     */
/*  Function:  dmi_buffer_add_execute_coord                            */
/*                                                                     */
/*  Description:                                                       */
/*    Compile a buffer entry which is a coordinate transform           */
/*    e.g. [cols dec]                                                  */
/*                                                                     */
/*  Input Parameters:                                                  */
/*              dmBuffer  *buffer                                      */
/*              dmExecute  execute                                     */
/*              long       id                                          */
/*                                                                     */
/*  Output Parameters:                                                 */
/*                none                                                 */
/*                                                                     */
/*  Returns:                                                           */
/*          void                                                       */
/*                                                                     */
/***********************************************************************/
void dmi_buffer_add_execute_coord( dmBuffer* buffer, dmExecute* execute, long id )
{
  dmDescriptor *parent = NULL;
  dmDescriptor *dd     = NULL;

  long parent_id = 0; 
  long func_id = DM_EXEC_COORD;
  long nargs = 2;
  long args[2];

  dd = buffer->cols[id];

  parent = dmi_coord_get_parent( dd );
  parent_id = parent->bufno-1;
  if ( parent && parent_id >= 0 )
  {
    args[0] = parent_id;
    args[1] = id;
    dmi_buffer_add_execute_func( execute, func_id, args, nargs );
  }
  else
  {
    sprintf( dm_errbuf, "DM buffer error: parent %s of virtual column %s is not in buffer",
	        parent->name, dd->name );
    dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
  }

}


