/*                                                                
**  Copyright (C) 2002-2012  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   filterlib.c                                                 */
/*                                                               */
/*      Routines to access members of the Filter object          */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"
#include <float.h>

#define XDEBUG 1

static long increment_index( long* pixel, long* pcpt, long* pdim, long dim, long ndim, long null );


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_get_dim                              */
/*                                                             */
/* Description:                                                */
/*             Return filter dim element                       */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter  *pFilter  - filter object             */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             long  dim                                       */
/*                                                             */
/***************************************************************/
long dmi_filter_get_dim( dmFilter* pFilter )
{
  return pFilter->dim;
}

/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_datatype                             */
/*                                                             */
/* Description:                                                */
/*             Return filter data type element                 */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter  *pFilter  - filter object             */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             dmDataType  type                                */
/*                                                             */
/***************************************************************/
dmDataType dmi_filter_datatype( dmFilter* pFilter )
{
  return pFilter->type;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_get_tabname                          */
/*                                                             */
/* Description:                                                */
/*             Return filter table name element                */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter  *pFilter  - filter object             */
/*             long       maxlen   - max string size           */
/*                                                             */
/*  Output parameters:                                         */
/*             char      *sName    - table name                */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_get_tabname( dmFilter* pFilter, char* sName, long maxlen )
{
  strcpy(sName, "");

  if ( pFilter && pFilter->name )
  {
    strncpy(sName, pFilter->name, maxlen);
  }
 
  return;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_copy_tabname                         */
/*                                                             */
/* Description:                                                */
/*             Copy table name if it does not exist already    */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter  *pFilter1 - input filter pointer      */
/*                                                             */
/*  Output parameters:                                         */
/*             dmFilter  *pFilter2 - output filter pointer     */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_copy_tabname( dmFilter *pFilter1, dmFilter *pFilter2 )
{
  if ( pFilter2 && !pFilter2->name ) 
    dmi_filter_set_tabname( pFilter2, pFilter1->name );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_set_tabname                          */
/*                                                             */
/* Description:                                                */
/*             Set table name. If input is NULL, return error  */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter   *pFilter - input filter pointer      */
/*             const char *sName   - name string               */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             int  dmSUCCESS or dmFAILURE                     */
/*                                                             */
/***************************************************************/
int dmi_filter_set_tabname( dmFilter *pFilter, const char *sName )
{
 int retval = dmSUCCESS;

 if ( pFilter && sName && *sName )
 {
   if ( pFilter->name )
     free( pFilter->name );

   pFilter->name = cs_dup(sName);
 }
 else
   retval = dmFAILURE;

 return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_set_region_string                    */
/*                                                             */
/* Description:                                                */
/*             Copy region string name into filter element     */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter  *pFilter   - input filter pointer     */
/*             char      *regstring - region name              */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_set_region_string( dmFilter *pFilter, char *regstring )
{
  if ( pFilter->regstring )
    free( pFilter->regstring );

  pFilter->regstring = cs_dup( regstring );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_print_ranges                         */
/*                                                             */
/* Description:                                                */
/*             Display Filter coponents and ranges             */
/*                                                             */
/*  Input parameters:                                          */
/*             char      *sName   - name                       */ 
/*             dmFilter  *pFilter - input filter pointer       */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_print_ranges( char *sName, dmFilter *pFilter )
{
  long ii;
  long n;

  if ( pFilter->name )
  {
    n = dmi_rangelist_get_size( &pFilter->ranges[0] );
    printf( "%-20s: Table %s with %ld ranges \n", sName, pFilter->name, n );
  }

  for ( ii = 0; ii < pFilter->dim; ii++ )
  {
    if ( pFilter->dim > 1 )  
      printf( "%-16s[%ld]: ", sName, ii+1 );
    else
      printf( "%-20s:  ", sName );

    dmi_rangelist_print( &pFilter->ranges[ii], pFilter->type );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_print_region                         */
/*                                                             */
/* Description:                                                */
/*             Display Filter region elements                  */
/*                                                             */
/*  Input parameters:                                          */
/*             char      *sName   - name                       */ 
/*             dmFilter  *pFilter - input filter pointer       */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_print_region( char *sName, dmFilter *pFilter )
{
  char *buf     = NULL;
  char *pRegion = NULL;
  char  null[] = "";

  if( pFilter->region )
  {
    pRegion = regAllocComposeRegion( pFilter->region );
    buf = pRegion;
  }
  else
    buf = null;

  if ( pFilter->name ) 
    printf( "%-20s: Region %s = %s\n", sName, pFilter->name, buf );
  else
    printf( "%-20s: Region %s\n", sName, buf );

  if ( pRegion )
    free( pRegion );

  if( pFilter->logical_region )
  {
    pRegion = regAllocComposeRegion( pFilter->logical_region );
    buf = pRegion;

    if ( pFilter->name ) 
      printf( "%-20s: LRegion %s = %s\n", sName, pFilter->name, buf );
    else
      printf( "%-20s: LRegion %s\n", sName, buf );

    free( pRegion );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_print                                */
/*                                                             */
/* Description:                                                */
/*             Display Filter elements                         */
/*                                                             */
/*  Input parameters:                                          */
/*             char      *sName   - name                       */ 
/*             dmFilter  *pFilter - input filter pointer       */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_print( char *sName, dmFilter *pFilter )
{
  if ( pFilter ) 
  {
    if ( pFilter->null)
       printf( "%-20s: Null\n", sName );
    else 
    {
      if ( pFilter->region )
        dmi_filter_print_region( sName, pFilter );

      dmi_filter_print_ranges( sName, pFilter );
    }
  }
  else
  {
    printf( "%-20s: No ranges\n", sName );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_print2                               */
/*                                                             */
/* Description:                                                */
/*             Display Filter with dim > 1                     */
/*                                                             */
/*  Input parameters:                                          */
/*             char      *sName   - name                       */ 
/*             dmFilter  *pFilter - input filter pointer       */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_print2( char *sName, dmFilter *pFilter )
{
  char type[CARD_SIZE];
  dmRangeList *pRange = NULL;

  long dim = 0;
  long ii  = 0;

  ar_mix_type_name( pFilter->type, type, CARD_SIZE );
  dim = dmi_filter_get_dim( pFilter );
  printf( "Filter Dim = %ld Type = %s\n", dim, type );

  if ( pFilter->null )
  {
    printf( "Null filter\n" );
    return;
  }

  if ( pFilter->ranges )
  {
    for ( ii = 0; ii < pFilter->dim; ii++ )
    {
      pRange = &pFilter->ranges[ii];
      printf( "Range %ld: %ld entries\n", ii+1, dmi_rangelist_get_size( pRange ) );
      dmi_rangelist_print2( sName, pRange, pFilter->type );
    }
  }

  if ( pFilter->regstring )
  {
    printf( "%s = %s\n", sName, pFilter->regstring );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_setup                                */
/*                                                             */
/* Description:                                                */
/*             Construct a filter, setting the table name, the */
/*             region, and the ranges for a specified          */
/*             component.                                      */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor  *dd      - component              */ 
/*             long           vcpt    - component number       */
/*             dmDataType     vtype   - data type              */
/*             void           value1  - range list of values   */
/*             void           value2  - range list of values   */
/*             long           nvalues - number of values       */
/*             char          *tabname - table name             */
/*             regRegion     *region  - region struct          */
/*                                                             */
/*  Output parameters:                                         */
/*             dmFilter      *pFilter - filter pointer         */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_setup( dmDescriptor *dd, dmFilter *pFilter, long vcpt, dmDataType vtype,
 void *value1, void *value2, long nvalues, char *tabname, regRegion *region )
{
  dmDataType ftype = DT_UNK;

#if DEBUG
printf( "FILTER SETUP %p %s\n", filter, dd->name );
#endif

  if ( region )
  {
    dmi_filter_update_region( dd, tabname, pFilter, region, 0 );
  }
  else
    pFilter->region = NULL;

  ftype = dmi_filter_type( vtype );

  if ( !pFilter->type )
    pFilter->type = ftype;

  if ( nvalues > 0 && pFilter->type != ftype )
  {
    dmu_err_set( DM_ERR_GENERIC, "DM Internal error: Subspace filter type mismatch" );
    return;
  }

  if ( pFilter->ranges && (nvalues > 0 || !region ))
  {
    dmi_rangelist_set( &pFilter->ranges[vcpt-1], vtype, value1, value2, nvalues ); 
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_compare                              */
/*                                                             */
/* Description:                                                */
/*        Compare a filter. If a region exists, ignore the     */
/*        ranges.                                              */
/*        If it is a multi-dim filter, it can have multiple    */
/*        diffs in the ranges.  The range filters for each dim */
/*        are considered separately as long as there is no     */
/*        region attached to the filter.                       */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter  *pFilter1 - input filter 1            */ 
/*             dmFilter  *pFilter2 - input filter 2            */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             int match -  UT_TRUE or UT_FALSE                */
/*                                                             */
/***************************************************************/
int dmi_filter_compare( dmFilter *pFilter1, dmFilter *pFilter2 )
{
  int isRegion = 0;
  int match    = UT_TRUE;

  isRegion = ( pFilter1->region || pFilter2->region );

  if ( isRegion )
    match = dmi_filter_compare_region( pFilter1, pFilter2 );

  if ( match )
    match = dmi_filter_compare_ranges( pFilter1, pFilter2 );

 return match;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_compare_for_combine                  */
/*                                                             */
/* Description:                                                */
/*        This comparison routine is a specialized routine to  */
/*        avoid combining filters which differ across multiple */
/*        dimensions (Bug #8522).                              */
/*                                                             */
/*        Two filters are combinable under the following       */
/*        circumstances:                                       */
/*            dim == 1, or                                     */
/*            dim == 2, and both are region filters, or        */
/*            dim > 1,  and there is one or less differences   */
/*                      between the ranges across dimensions   */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter  *pFilter1 - input filter 1            */ 
/*             dmFilter  *pFilter2 - input filter 2            */
/*             double    *fieldx   -                           */
/*             double    *fieldy   -                           */
/*             int       *can_combine                          */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             int   UT_TRUE or UT_FALSE                       */
/*                                                             */
/***************************************************************/
int dmi_filter_compare_for_combine( dmFilter *pFilter1, dmFilter *pFilter2, 
				    double *fieldx, double *fieldy,
                                    int *can_combine)
{
  double xbounds[2];
  double ybounds[2];
  dmRangeList range;

  long vcpt_offset = 0;
  long dim         = 0;
  int  diff        = 0;
 
  if ( pFilter1->type != pFilter2->type )
  {
   *can_combine = UT_FALSE;
   return UT_FALSE;
  }

  if ( pFilter1->region && pFilter2->region )
  {
    if ( !dmi_filter_compare_region( pFilter1, pFilter2 ))
    {
      /* can only combine two different regions if we can  */
      /* be sure that both range lists are just the region */
      /* bounds                                            */

      dmi_rangelist_init( &range );
      *can_combine = UT_FALSE;
      regExtent( pFilter1->region, fieldx, fieldy, xbounds, ybounds );
      dmi_rangelist_set( &range, DT_REAL8, &xbounds[0], &xbounds[1], 1 );

      if ( dmi_rangelist_compare( &pFilter1->ranges[0], &range, DT_REAL8 ))
      {
        dmi_rangelist_set( &range, DT_REAL8, &ybounds[0], &ybounds[1], 1 );

        if ( dmi_rangelist_compare( &pFilter1->ranges[1], &range, DT_REAL8 ))
        {
          regExtent( pFilter2->region, fieldx, fieldy, xbounds, ybounds );
          dmi_rangelist_set( &range, DT_REAL8, &xbounds[0], &xbounds[1], 1 );

          if ( dmi_rangelist_compare( &pFilter2->ranges[0], &range, DT_REAL8 ))
          {
            dmi_rangelist_set( &range, DT_REAL8, &ybounds[0], &ybounds[1], 1 );

            if ( dmi_rangelist_compare( &pFilter2->ranges[1], &range, DT_REAL8 ))
              *can_combine = UT_TRUE;
          }
        }
      }
      return UT_FALSE;
    }
  }

  dim = pFilter1->dim;
  diff = 0;

  for ( vcpt_offset = 0; vcpt_offset < dim; vcpt_offset++ )
  {
    if ( !dmi_rangelist_compare( &pFilter1->ranges[vcpt_offset], &pFilter2->ranges[vcpt_offset], pFilter1->type ))
    {
      if ( diff )
      {
        *can_combine = UT_FALSE;
        return UT_FALSE;
      }
      else
        diff = UT_TRUE;
    }
  }

  if ( diff )
  {
    *can_combine = UT_TRUE;
    return UT_FALSE;
  }

  *can_combine = UT_TRUE;

  if ( pFilter1->name || pFilter2->name )
  { 
    return pFilter1->name && pFilter2->name && !strcmp( pFilter1->name, pFilter2->name );
  }

  return UT_TRUE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_compare_ranges                       */
/*                                                             */
/* Description:                                                */
/*        Compare two filter ranges. They are different if the */
/*        name or types are different, or they have different  */
/*        range numbers.                                       */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter  *pFilter1 - input filter 1            */ 
/*             dmFilter  *pFilter2 - input filter 2            */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             dmBool -  UT_TRUE or UT_FALSE                   */
/*                                                             */
/***************************************************************/
dmBool dmi_filter_compare_ranges( dmFilter *pFilter1, dmFilter *pFilter2 )
{           
  long cpt = 0;
  int  ok  = 0;

  /* They are different if they have different names or types */
  if ( pFilter1->name || pFilter2->name )
  {
    ok = pFilter1->name && pFilter2->name && !strcmp( pFilter1->name, pFilter2->name );

    if ( !ok )
     return UT_FALSE;
  }

  if ( pFilter1->type != pFilter2->type )
    return UT_FALSE;

  /* They are different if they have different numbers of ranges for */ 
  /* any vector component */

  for ( cpt = 0; cpt < pFilter1->dim; cpt++ )
  {
    if ( !dmi_rangelist_compare( &pFilter1->ranges[cpt], &pFilter2->ranges[cpt], pFilter1->type ))
      return UT_FALSE;
  }

  return UT_TRUE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_combine                              */
/*                                                             */
/* Description:                                                */
/*        Combine two filter regions and ranges. Combined      */
/*        filter information is placed in pFilter1.            */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter  *pFilter1 - input filter 1            */ 
/*             dmFilter  *pFilter2 - input filter 2            */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_combine( dmFilter *pFilter1, dmFilter *pFilter2 )
{
  long  filterCpt = 0;
  regRegion *newRegion = NULL;

  if ( !pFilter2 || !pFilter1 )
    return;

  if ( pFilter2->null )
    return;

  if ( pFilter2->region || pFilter1->region )
  {
    if ( !pFilter1->region )
      pFilter1->region = regCopyRegion( pFilter2->region );
    else
    {
      newRegion = regUnionRegion( pFilter1->region, pFilter2->region );
      free( pFilter1->regstring );
      pFilter1->regstring = NULL;
      regFree( pFilter1->region );
      pFilter1->region = newRegion;
    }

#if DEBUG
printf( "COMBINE REG %p %p \n", newRegion, pFilter1 );
#endif
  }

  /* regions have ranges, too! */
  if ( pFilter1->type == pFilter2->type )
  {
    /* Should really be union */
    for ( filterCpt = 0; filterCpt < pFilter2->dim; filterCpt++ )
    {
      dmi_rangelist_combine( &pFilter2->ranges[filterCpt], &pFilter1->ranges[filterCpt], pFilter2->type );
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_extract                              */
/*                                                             */
/* Description:                                                */
/*        Returns values of ranges from filter.                */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter  *pFilter  - input filter              */ 
/*             long       cptno    - component number          */
/*             dmDataType type     - data type                 */
/*             int       *pincnull - flag                      */
/*                                                             */
/*  Output parameters:                                         */
/*             void     **value1   - range list values         */
/*             void     **value2   - range list values         */
/*             long      *nvalues  - number of values returned */ 
/*                                                             */
/*  Returns:                                                   */
/*             int  dmSUCCESS or dmFAILURE                     */
/*                                                             */
/***************************************************************/
int dmi_filter_extract( dmFilter *pFilter, long cptno, dmDataType type,
             void** value1, void** value2, long* nvalues, int* pincnull )
{
  int ctype = 0;

  if ( !pFilter || !pFilter->ranges )
    return dmFAILURE;

  ctype = ar_mix_is_numeric( type );

  if ( dmverbose > 0 )
  {
    if ( pFilter->type == DT_CHAR && ctype )
      printf( "DM Internal Error: Type mismatch in filter extract\n" );
  }

  dmi_rangelist_extract( &pFilter->ranges[cptno-1], type, value1, value2, nvalues, pincnull );

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_free                                 */
/*                                                             */
/* Description:                                                */
/*        Free space associated with filter sturcture          */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter  *pFilter - input filter               */ 
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_free( dmFilter *pFilter )
{
  if ( !pFilter )
    return;

  dmi_filter_clear( pFilter );
  free( pFilter->ranges );
  regFree( pFilter->region );
  regFree( pFilter->logical_region );
  free( pFilter );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_clear                                */
/*                                                             */
/* Description:                                                */
/*        Remove current information from filter               */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter  *pFilter - input filter               */ 
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_clear( dmFilter *pFilter )
{
  if ( !pFilter )
    return;

  if ( pFilter->name )
    free( pFilter->name );

  dmi_filter_zero( pFilter );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_zero                                 */
/*                                                             */
/* Description:                                                */
/*        Clears range values and regstring in filter          */
/*        Called by dmi_filter_clear() above                   */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter  *pFilter - input filter               */ 
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_zero( dmFilter *pFilter )
{
  long ii      = 0;
  int  isChar  = 0;

  if ( !pFilter )
    return;

  isChar = !ar_mix_is_numeric( pFilter->type );

  for ( ii = 0; ii < pFilter->dim; ii++ )
  {
    dmi_rangelist_free( &pFilter->ranges[ii], isChar );
  }

  if ( pFilter->regstring )
    free( pFilter->regstring );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_col_get_current_filter             */
/*                                                             */
/* Description:                                                */
/*        Get the filter for the current component             */
/*        Only if descriptor col is member of block subspace   */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor *pColDescriptor - col descriptor   */ 
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             dmFilter *pFilter                               */
/*                                                             */
/***************************************************************/
dmFilter* dmi_subspace_col_get_current_filter( dmDescriptor *pColDescriptor )
{
  dmSubspace     *pSubspace = NULL;
  dmSubspaceData *pSSdata   = NULL;
  long            cpt       = 1;

  pSSdata   = dmi_ss_get_data( pColDescriptor );

  // Code Change Bug 9606 January 30, 2012
  //  Function dmi_ss_get_data() can return NULL

  if ( !pSSdata )
    return NULL;

  pSubspace = pSSdata->subspace;

  if ( !pSubspace )
  {
#if 0
  if ( verbose > 0 )
    printf( "DM INTERNAL ERROR: SUBSPACE CROSS INDEX NOT INIT %s\n", pColDescriptor->name );
#endif

  // Code Change Bug 9606 January 30, 2012
  //   Check for NULL pointer before use

    if ( pColDescriptor->block )
      pSubspace = pColDescriptor->block->subspace;

    if ( !pSubspace )
      return NULL;
  }

  // Code Change Bug 9606 January 30, 2012
  //   Check for uninitialed structure values

  cpt = pSubspace->cpt;
  if ( cpt <= 0 )
    cpt = 1;
  
  return dmi_subspace_col_reset_filter( pColDescriptor, cpt );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_col_init_filters                   */
/*                                                             */
/* Description:                                                */
/*        Create all the filters for a given subspace col.     */
/*        Fill the current one with the input filter values.   */
/*        Descriptor must be part of the subspace?             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmSubspace   *pSubspace      - subspace ptr     */
/*             dmDescriptor *pColDescriptor - col descriptor   */ 
/*                                                             */
/*  Output parameters:                                         */
/*             dmFilter     *pFilter   - created filter        */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_subspace_col_init_filters( dmSubspace *pSubspace, 
                              dmDescriptor *pColDescriptor, dmFilter *pFilter )
{
  dmFilter *pCptFilter = NULL;
  long      cptNo = 0;
  int       Duplicate = 0;       /* Tweak this if needed to copy all cpts */


  /* Fill the descriptor structure.  Note that the value field is */ 
  /* set separately (not within dmPMakeSubspace)                  */

  /* Create each component within subspace column and write data  */
  /* to current component                                         */

  for( cptNo = 1; cptNo <= pSubspace->ncpts; cptNo++)
  {
    pCptFilter = dmi_subspace_col_get_filter( pColDescriptor, cptNo );

    if ( !pCptFilter )
    {
      pCptFilter = dmi_subspace_col_reset_filter( pColDescriptor, cptNo );

      if (!pCptFilter)
        return;

      if ( cptNo == pSubspace->cpt || Duplicate )
        dmi_filter_copy( pFilter, pCptFilter );
    }
    else
    {
      if ( !dmi_subspace_col_update_from_filter( pColDescriptor, pFilter ))
      {
        sprintf( dm_errbuf, "DM ERROR: Subspace update failed for %s", dmi_descriptor_get_name( pColDescriptor ) );
        dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
      }
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_copy                                 */
/*                                                             */
/* Description:                                                */
/*        Copy a filter. Copy its ranges and its region.       */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter    *pFilter      - input filter        */
/*                                                             */
/*  Output parameters:                                         */
/*             dmFilter    *pFilterCopy  - filter copy         */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_copy( dmFilter *pFilter, dmFilter *pFilterCopy )
{
  long vector_cpt = 0;

  if ( !pFilterCopy->ranges || pFilterCopy->dim != pFilter->dim )
  {
    pFilterCopy->dim    = pFilter->dim;
    pFilterCopy->ranges = dmi_rangelist_alloc_array( NULL, pFilterCopy->dim );
  }

  pFilterCopy->null = pFilter->null;

  for ( vector_cpt = 0; vector_cpt < pFilter->dim; vector_cpt++ )
  {
    dmi_rangelist_copy( &pFilter->ranges[vector_cpt], &pFilterCopy->ranges[vector_cpt], pFilter->type );
  }

  pFilterCopy->name      = cs_dup( pFilter->name );
  pFilterCopy->regstring = cs_dup( pFilter->regstring );
  pFilterCopy->type      = pFilter->type;

  if ( pFilter->region )
    pFilterCopy->region = regCopyRegion( pFilter->region );

  if ( pFilter->logical_region )
     pFilterCopy->logical_region = regCopyRegion( pFilter->logical_region );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_copy_cpt                             */
/*                                                             */
/* Description:                                                */
/*        Copy a filter component                              */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter    *pFilter      - input filter        */
/*             long         cpt          - component number    */
/*                                                             */
/*  Output parameters:                                         */
/*             dmFilter    *pFilterCopy  - filter copy         */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_copy_cpt( dmFilter *pFilter, dmFilter *pFilterCopy, long cpt )
{
  long vector_cpt = 0;

  if ( !pFilterCopy->ranges )
    pFilterCopy->ranges = dmi_rangelist_alloc_array( NULL, pFilterCopy->dim );

  pFilterCopy->null = pFilter->null;

  vector_cpt = cpt-1;
  dmi_rangelist_copy( &pFilter->ranges[0], &pFilterCopy->ranges[vector_cpt], pFilter->type );
  pFilterCopy->name = cs_dup( pFilter->name );
  pFilterCopy->regstring=cs_dup( pFilter->regstring );
  pFilterCopy->type = pFilter->type;
 
  if ( pFilter->region )
    pFilterCopy->region = regCopyRegion( pFilter->region );

  if ( pFilter->logical_region )
    pFilterCopy->logical_region = regCopyRegion( pFilter->logical_region );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_copy_element                         */
/*                                                             */
/* Description:                                                */
/*        Copy a filter. Support the case of copying a 1D      */
/*        filter to a component of a 2D filter.                */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter    *pInFilter    - input filter        */
/*             long         output_vector_cpt = cpt number     */
/*                                                             */
/*  Output parameters:                                         */
/*             dmFilter    *pOutFilter   - filter out          */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_copy_element( dmFilter *pInFilter, dmFilter *pOutFilter, long output_vector_cpt )
{
  long vector_cpt = 0;
  long f_vector_cpt = 0;

  if ( !pOutFilter->ranges )
  {
    pOutFilter->dim = pInFilter->dim;
    pOutFilter->ranges = dmi_rangelist_alloc_array( NULL, pOutFilter->dim );
  }

  pOutFilter->name = cs_dup( pInFilter->name );
  pOutFilter->regstring = cs_dup( pInFilter->regstring );
  pOutFilter->type = pInFilter->type;
  pOutFilter->null = pInFilter->null;

  if ( pOutFilter->null )
    return;

  for ( vector_cpt = 1; vector_cpt <= pInFilter->dim; vector_cpt++ )
  {
    if ( output_vector_cpt > 0 )
      f_vector_cpt = output_vector_cpt;  
    else
      f_vector_cpt = vector_cpt;
 
    dmi_rangelist_copy( &pInFilter->ranges[vector_cpt-1], &pOutFilter->ranges[f_vector_cpt-1], pInFilter->type );
  }

  if ( pInFilter->region )
    pOutFilter->region = regCopyRegion( pInFilter->region );

  if ( pInFilter->logical_region )
    pOutFilter->logical_region = regCopyRegion( pInFilter->logical_region );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_init                                 */
/*                                                             */
/* Description:                                                */
/*        Initialize a filter. If input filter pointer is      */
/*        NULL, allocate filter structure. If non-NULL, set    */
/*        all values to zero/null and set dim.                 */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter    *pFilter      - input filter ptr    */
/*             long         dim          - dim                 */
/*             dmDataType   type         - data type           */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             dmFilter *pInitFilter                           */
/*                                                             */
/***************************************************************/
dmFilter* dmi_filter_init( dmFilter *pFilter, long dim, dmDataType type )
{
  dmFilter *pInitFilter = pFilter;

  if ( pInitFilter )
  {
    if ( dim != pInitFilter->dim )
    {
      dmi_filter_zero( pInitFilter );
      pInitFilter->dim = dim;
    }
  }
  else
  {
    pInitFilter = (dmFilter*)calloc( 1, sizeof( dmFilter ));
    pInitFilter->dim = dim;
  }

  if ( !pInitFilter->ranges )
    pInitFilter->ranges = dmi_rangelist_alloc_array( NULL, pInitFilter->dim );

  pInitFilter->null = 0;
  pInitFilter->name = NULL;
  pInitFilter->regstring = NULL;

  if ( pInitFilter->type == 0 && type != 0 )
    pInitFilter->type = type;

  return pInitFilter;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_is_implicit                          */
/*                                                             */
/* Description:                                                */
/*        Is the filter trivial? (Not the same as NULL)        */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter    *pFilter      - input filter ptr    */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             int  UT_TRUE or UT_FALSE                        */
/*                                                             */
/***************************************************************/
int dmi_filter_is_implicit( dmFilter *pFilter )
{
  long ii = 0;

  if ( !pFilter )
    return UT_TRUE;

  if ( pFilter->null )
    return UT_FALSE;

  for ( ii = 0; ii < pFilter->dim; ii++ )
    if ( dmi_rangelist_get_size( &pFilter->ranges[ii] ) > 0 ) 
      return UT_FALSE;

  if ( pFilter->region ) 
    return UT_FALSE;

 return UT_TRUE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_is_default                           */
/*                                                             */
/* Description:                                                */
/*        Is the filter the default                            */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor *pColDesc   - col descriptor       */
/*             dmFilter    *pFilter     - filter               */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             dmBOOL  UT_TRUE or UT_FALSE                     */
/*                                                             */
/***************************************************************/
dmBool dmi_filter_is_default( dmDescriptor *pColDesc, dmFilter *pFilter )
{
  long vcpt   = 0;
  int  retval = UT_FALSE;

  if ( !pFilter )
    return UT_TRUE;

  if ( pFilter->null )
    return retval;
 
  for ( vcpt = 1; vcpt <= pFilter->dim; vcpt++ )
  {
    retval = dmi_filter_range_is_default( pColDesc, pFilter, vcpt );

    /* All components must be true to pass */
    if ( retval == UT_FALSE )
      return retval;
  }

  if ( retval == UT_TRUE )
    retval = dmi_filter_region_is_default( pFilter );

  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_range_is_default                     */
/*                                                             */
/* Description:                                                */
/*        Does the filter range have only the default values?  */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor *pColDesc   - col descriptor       */
/*             dmFilter     *pFilter    - filter               */
/*             long         vcpt        - component number     */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             dmBOOL  UT_TRUE or UT_FALSE                     */
/*                                                             */
/***************************************************************/
dmBool dmi_filter_range_is_default( dmDescriptor *pColDesc, dmFilter *pFilter, long vcpt )
{
  dmDescriptor *pCptDesc = NULL;
  dmRangeList  *pRange   = NULL;
  long          size     = 0;
  dmBool        retval   = UT_FALSE;

  if ( !pColDesc )
    return retval;

  if ( pFilter->name && *pFilter->name && pFilter->dim == 1 )
    return retval;

  pRange = &pFilter->ranges[vcpt-1];
  size = dmi_rangelist_get_size( pRange );

  if ( size > 1 ) 
    retval = UT_FALSE;
  else
  {
    /* Ruthlessly suppress any attempt to provide filters on array columns */

    if ( pColDesc && dmi_descriptor_get_array_size( pColDesc ) > 1 )
      retval = UT_TRUE;
    else
      if ( size == 1 && pColDesc )
      {
        pCptDesc = dmi_descriptor_get_cpt ( pColDesc, vcpt );
        retval   = dmi_rangelist_is_default( pRange, pCptDesc, pFilter->type );
      }
  }

  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_region_is_default                    */
/*                                                             */
/* Description:                                                */
/*        Does the filter region have only the default values? */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter     *pFilter    - filter               */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             dmBOOL  UT_TRUE or UT_FALSE                     */
/*                                                             */
/***************************************************************/
dmBool dmi_filter_region_is_default( dmFilter *pFilter )
{
  if ( !pFilter->region )
    return UT_TRUE;

  if ( pFilter->dim == 2 && ar_mix_is_numeric( pFilter->type ) )
  {
    char buf[512];

    if ( dmi_rangelist_to_rectangle( &pFilter->ranges[0], &pFilter->ranges[1], buf ))
    { 
      if ( !pFilter->regstring )
      {
        pFilter->regstring = regAllocComposeRegion( pFilter->region );
      }

      if ( dmverbose >= 5 )
        fprintf(stderr, "Comparing default region %s with region %s\n", buf, pFilter->regstring );

      if ( cs_uceq( pFilter->regstring, buf ) )
      {
        /* We have a default rectangle filter */
        regFree (pFilter->region );
        pFilter->region = NULL;
        free( pFilter->regstring );
        pFilter->regstring = NULL;
        return UT_TRUE;
      }
    }
  }
  return UT_FALSE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_set_null                             */
/*                                                             */
/* Description:                                                */
/*        Set all filter elements to null/zero                 */
/*        Calls function dmi_filter_zero()                     */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter     *pFilter    - filter               */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_set_null( dmFilter *pFilter )
{
  dmi_filter_zero( pFilter );
  pFilter->null = UT_TRUE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_is_null                              */
/*                                                             */
/* Description:                                                */
/*        Returns null element of filter                       */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter     *pFilter    - filter               */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             int  null                                       */
/*                                                             */
/***************************************************************/
int dmi_filter_is_null( dmFilter *pFilter )
{
 return pFilter->null;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_col_make_null                      */
/*                                                             */
/* Description:                                                */
/*        Sets a subspace column's filter to null/zero         */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor *pColDescriptor - col descriptor   */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_subspace_col_make_null( dmDescriptor *pColDescriptor )
{
  dmFilter *pDestFilter = NULL;

  pDestFilter = dmi_subspace_col_get_current_filter( pColDescriptor );
  dmi_filter_set_null( pDestFilter );
  dmi_descriptor_set_update( pColDescriptor );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_col_is_gti                         */
/*                                                             */
/* Description:                                                */
/*       Look at a descriptor and its value group.             */
/*       Check if it's a double-type descriptor called TIME.   */
/*       If it has no table name, or if its table name is      */
/*       STDGTI, ALLGTI, or GTI, then recognize it as a GTI.   */
/*       Promoted from macro to real routine by JCM 1998 Oct   */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor *pColDescriptor - col descriptor   */
/*             dmFilter     *pFilter        - filter           */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             dmBOOL  UT_TRUE or UT_FALSE                     */
/*                                                             */
/***************************************************************/
dmBool dmi_subspace_col_is_gti( dmDescriptor *pColDescriptor, dmFilter *pFilter )
{
  if ( cs_uceq(pColDescriptor->name, "TIME")  && pColDescriptor->type == DT_REAL8)
  {
    if ( !pFilter->name || !(*pFilter->name) )
      pFilter->name = cs_dup( "GTI" );

     return UT_TRUE;
  }
  return UT_FALSE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_get_cpt_bounds                       */
/*                                                             */
/* Description:                                                */
/*      For a 2D filter, check for existing rectangular bounds */
/*      which will later be intersected with the region.       */
/*      Otherwise, the field is assumed to be infinite in      */
/*      this case.  Perhaps there's a way to pass in an        */
/*      appropriate tlmin/tlmax - this routine is just a fudge.*/
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter     *pFilter   - filter                */
/*             long          vcpt      - component number      */
/*                                                             */
/*  Output parameters:                                         */
/*             double       *fieldx    - values                */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_get_cpt_bounds( dmFilter *pFilter, long vcpt, double *fieldx )
{
  dmRangeList *pRange = NULL;

  fieldx[0] = 0.0;
  fieldx[1] = DT_REAL8_MAX;

  if ( vcpt > pFilter->dim )
    return;

  pRange = &pFilter->ranges[vcpt-1];

  if ( pRange && pRange->n > 0 )
  {
    if ( !pRange->sorted )
      dmi_rangelist_reduce( pRange, DT_REAL8 );

    fieldx[0] = ((double*)pRange->mins)[0];
    fieldx[1] = ((double*)pRange->maxes)[pRange->n - 1];
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_get_region_bounds                    */
/*                                                             */
/* Description:                                                */
/*      Get the boundary of the filter                         */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter     *pFilter   - filter                */
/*                                                             */
/*  Output parameters:                                         */
/*             double       *fieldx    - values                */
/*             double       *fieldy    - values                */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_get_region_bounds( dmFilter *pFilter, double *fieldx, double *fieldy )
{
  dmi_filter_get_cpt_bounds( pFilter, 1, fieldx );
  dmi_filter_get_cpt_bounds( pFilter, 2, fieldy );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_intersect                            */
/*                                                             */
/* Description:                                                */
/*      Handle case where first filter is null                 */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter     *pFilter1  - filter 1              */
/*             dmFilter     *pFilter2  - filter 2              */
/*             int           intersect - flag                  */
/*             long          cpt       - component number      */
/*             long          vdim      - dim                   */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             dmFilter pFilter1                               */
/*                                                             */
/***************************************************************/
dmFilter* dmi_filter_intersect( dmFilter *pFilter1, dmFilter *pFilter2,
                                int intersect, long cpt, long vdim )
{
  dmDataType type  = DT_UNK;
  long       dim   = 0;
  int        apply_empty = 0;
 
  if ( !pFilter1 )
  {
    dim  = dmi_filter_get_dim( pFilter2 );
    type = dmi_filter_datatype( pFilter2 );
    pFilter1 = dmi_filter_init( NULL, vdim, type );

    if ( dim < vdim && cpt > 0 )
      dmi_filter_copy_cpt( pFilter2, pFilter1, cpt );
    else
      dmi_filter_copy( pFilter2, pFilter1 );
  }
  else
  {
    type = dmi_filter_datatype( pFilter1 );
    dmi_filter_intersect_update( pFilter1, pFilter2, type, intersect, cpt, apply_empty );
  }

  return pFilter1;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_intersect_update                     */
/*                                                             */
/* Description:                                                */
/*      Update filter by intersecting with the contents of a   */
/*      second filter. Results stored in filter 1              */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter     *pFilter1  - filter 1              */
/*             dmFilter     *pFilter2  - filter 2              */
/*             dmDataType    type      - data type             */
/*             int           intersect - flag                  */
/*             long          cpt       - component number      */
/*             int           apply_empty - flag                */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_intersect_update( dmFilter *pFilter1, dmFilter *pFilter2,
             dmDataType type, int intersect, long cpt, int apply_empty )
{
  dmRangeList *pRange1 = NULL;
  dmRangeList *pRange2 = NULL;

  double fieldx[2] = { 0.0, 0.0 };
  double fieldy[2] = { 0.0, 0.0 };

  long ii     = 0;
  long jj     = 0;
  long mincpt = 0;
  long maxcpt = 0;
  long omincpt = 0;
  long omaxcpt = 0;

  int  update = 1;
  int  null   = UT_FALSE; 

  if ( dmverbose > 0 )
  {
    if ( pFilter2->regstring && !pFilter2->region )
      fprintf( stderr, "DM INTERNAL ERROR: REGION %s NOT RESOLVED\n", pFilter2->regstring );

    if ( pFilter1->regstring && !pFilter1->region )
      fprintf( stderr, "DM INTERNAL ERROR: REGION %s NOT RESOLVED\n", pFilter1->regstring );
  }

  if ( pFilter2->type != pFilter1->type )
    return;

  /* intersection with NULL is always NULL */
  if ( pFilter2->null || pFilter1->null )
  {
    pFilter1->null = UT_TRUE;
    return;
  }

  if ( pFilter2->region )
  {
    dmi_filter_get_region_bounds( pFilter2, fieldx, fieldy );
    dmi_filter_update_region_with_field( pFilter1, NULL, pFilter2->region, pFilter2->logical_region, fieldx, fieldy, update );
  }

  if ( cpt > 0 )
  {
    mincpt = maxcpt = cpt;
  }
  else 
  {
    mincpt = 1;
    maxcpt = pFilter1->dim;
   }

  if ( pFilter2->dim == 1 )
  {
    omincpt = omaxcpt = 1;
  }
  else
  {
    omincpt = mincpt;
    omaxcpt = maxcpt;
   }

#if DEBUG
dmi_filter_print( "FILTER1", pFilter1 );
dmi_filter_print( "FILTER2", pFilter2 );
printf( "CPT = %ld:%ld OCPT = %ld:%ld\n", mincpt, maxcpt, omincpt, omaxcpt );
#endif

  for ( ii = mincpt, jj = omincpt; ii <= maxcpt && jj <= omaxcpt; ii++, jj++ )
  {  
    pRange1 = &pFilter1->ranges[ii-1];
    pRange2 = &pFilter2->ranges[jj-1];

    /* When the filter is initialized, it has the TLMIN:TLMAX range.
     * The first user range is intersected with this.
     * Subsequent user ranges are unioned... e.g.
     *    [pha=20:40,100:200,300:400]
     * is parsed as:   initial  0:36855  (TLMIN:TLMAX)
     *                 1st ran  20:40    (after intersection of 20:40 with 0:36855)
     *                 2nd ran  20:40,100:200 (after union of 100:200 with previous)
     * (etc)
     * We use the flag field to let this happen.
     * However, must be careful with vector components: x and y handled separately here.
     */

    /* This is the only use of this flag,
     * there must be a better way... ryan */

    if ( !pRange1->flag )
    {
      pRange1->flag = UT_TRUE;
      intersect = 1;
    }

    null =dmi_rangelist_intersect_mode( pRange2, pRange1, type, intersect, apply_empty );

    if ( null )
      pFilter1->null = UT_TRUE;
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_alloc_array                          */
/*                                                             */
/* Description:                                                */
/*      Allocate or reallocate a filter list                   */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter    **ppFilterArray - filter list       */
/*             long          dim       - dim                   */
/*             long          start     - start index           */
/*             long          arraySize - array size            */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             dmFilter **ppFilterList                         */
/*                                                             */
/***************************************************************/
dmFilter** dmi_filter_alloc_array( dmFilter **ppFilterArray, long dim,
                                   long start, long arraySize )
{
  dmFilter **ppFilterList = NULL;
  long       index = 0;

  if ( arraySize == 0 || dim == 0 )
    return NULL;

  if ( ppFilterArray )
  {
    /* Reallocate space for existing filter list */

    ppFilterList = (dmFilter**) realloc( ppFilterArray, arraySize * UT_SZ_P );

    /* Clear list entries */
    for ( index = start; index < arraySize; index++ )
      ppFilterList[index] = NULL;
  }
  else 
  {
    /* Allocate new list */

    ppFilterList =  (dmFilter**)calloc( arraySize, UT_SZ_P );
  }

 return ppFilterList;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_type                                 */
/*                                                             */
/* Description:                                                */
/*      Return filter data type                                */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDataType    type          - data type         */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             dmDataType  filterDataType                      */
/*                                                             */
/***************************************************************/
dmDataType dmi_filter_type( dmDataType type )
{
  dmDataType filterDataType = DT_UNK;

  if ( type == DT_BIT )
    filterDataType = DT_BIT;
  else 
    filterDataType = ( ar_mix_is_numeric( type ) ) ? DT_REAL8 : DT_CHAR;

  return filterDataType;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_alloc                                */
/*                                                             */
/* Description:                                                */
/*      Allocate and initialize filter of particular data type */
/*                                                             */
/*  Input parameters:                                          */
/*             long          dim       - dim                   */
/*             dmDataType    type      - data type             */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             dmFilter  *pFilter                              */
/*                                                             */
/***************************************************************/
dmFilter* dmi_filter_alloc( long dim, dmDataType type )
{
  dmFilter   *pFilter = NULL;
  dmDataType  filterDataType = DT_UNK;

  filterDataType = dmi_filter_type( type );

  pFilter = (dmFilter*)calloc( 1, sizeof( dmFilter ));

  dmi_filter_init( pFilter, dim, filterDataType );

  return pFilter;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_set_incnull                          */
/*                                                             */
/* Description:                                                */
/*      Set null in component range                            */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter  *pFilter       - filter               */
/*             long       cpt           - component number     */
/*             int        incnull       - flag                 */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_set_incnull( dmFilter *pFilter, long cpt, int incnull )
{
  dmi_rangelist_set_incnull( &pFilter->ranges[cpt-1], incnull );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_include_null                         */
/*                                                             */
/* Description:                                                */
/*      Include null in range                                  */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter  *pFilter       - filter               */
/*             long       cpt           - component number     */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_include_null( dmFilter *pFilter, long cpt )
{
  dmi_rangelist_include_null( &pFilter->ranges[cpt-1] );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_exclude_null                         */
/*                                                             */
/* Description:                                                */
/*      Exclude null in range                                  */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter  *pFilter       - filter               */
/*             long       cpt           - component number     */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_exclude_null( dmFilter *pFilter, long cpt )
{
  dmi_rangelist_exclude_null( &pFilter->ranges[cpt-1] );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_add_range                            */
/*                                                             */
/* Description:                                                */
/*      Add range values to filter                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter  *pFilter       - filter               */
/*             long       cpt           - component number     */
/*             void      *specvals      - values               */
/*             int        exclude       - flag                 */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_add_range( dmFilter *pFilter, long cpt, void *specvals, int exclude )
{
  dmRangeList *pTmpRange = NULL;
  dmDataType   type = DT_UNK;    /* type of range for rangelist */
  void        *min  = NULL;
  void        *max  = NULL;

  long         size    = 0;
  int          is_char = 0;

  if ( ar_mix_is_numeric( pFilter->type ))
  {
    type    = DT_REAL8;
    is_char = UT_FALSE;
    size    = ar_mix_type_size( type );
    min     = ar_gen_ptr( specvals, type, size, 0 );
    max     = ar_gen_ptr( specvals, type, size, 1 );
  }
  else
  {
    /* ar_gen_ptr doesn't seem to  work for char types since it */ 
    /* returns a char** instead of a char*                      */

    type    = pFilter->type;
    is_char = UT_TRUE;
    min     = ((char**)specvals)[0];
    max     = ((char**)specvals)[1];
  }

  if ( exclude )
  {
    pTmpRange = dmi_rangelist_alloc_array( NULL, 1 );

    dmi_rangelist_invert_range( pTmpRange, type, min, max, NULL, NULL );
    dmi_rangelist_add_rangelist( &pFilter->ranges[cpt-1], type, pTmpRange );

    dmi_rangelist_free( pTmpRange, is_char );
    free( pTmpRange );
  }
  else
  {
    dmi_rangelist_add( &pFilter->ranges[cpt-1], type, min, max );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_get_range                            */
/*                                                             */
/* Description:                                                */
/*      Returns filter component's range values                */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter  *pFilter       - filter               */
/*             long       cptno         - component number     */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             dmRangeList  *pRange                            */
/*                                                             */
/***************************************************************/
dmRangeList* dmi_filter_get_range( dmFilter *pFilter, long cptno )
{
  if ( !pFilter )
    return NULL;

  return &pFilter->ranges[cptno-1];
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_extremes                             */
/*                                                             */
/* Description:                                                */
/*      Returns component's extreme range values               */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter  *pFilter       - filter               */
/*             long       cptno         - component number     */
/*             dmDataType type          - data type            */
/*                                                             */
/*  Output parameters:                                         */
/*             double    *vals          - range values         */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_extremes( dmFilter *pFilter, long cptno, dmDataType type, double *vals )
{
  dmRangeList *pRange = NULL;

  if ( !pFilter )
    return;

  pRange = dmi_filter_get_range( pFilter, cptno );

  /* Override type if filter is not numeric; will return +/- Inf */

  if ( !ar_mix_is_numeric( pFilter->type ))
    type = DT_CHAR;

  dmi_rangelist_extremes( pRange, type, vals );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_transform_image_bounds                */
/*                                                             */
/* Description:                                                */
/*      Returns component's extreme range values               */
/*                                                             */
/*  Input parameters:                                          */
/*             dmCoordTransform *wcsvals - coord transform     */
/*             long              cptno   - component number    */
/*             int               invert  - flag                */
/*                                                             */
/*  Output parameters:                                         */
/*             double    *bounds        - bound values         */
/*                                                             */
/*  Returns:                                                   */
/*             int  1                                          */
/*                                                             */
/***************************************************************/
int dmi_coord_transform_image_bounds( double *bounds,  dmCoordTransform *wcsvals, long cptno, int invert )
{

  double cdelt = 0.0;
  double crpix = 0.0;
  double crval = 0.0;

  crpix = ((double*)wcsvals->pix)[cptno-1];
  cdelt = ((double*)wcsvals->delt)[cptno-1];
  crval = ((double*)wcsvals->val)[cptno-1];

  if (  cdelt == 1.0 && crpix == crval )
    return 1;

  if ( invert )
  {
     bounds[0] = crpix + ( bounds[0] - crval ) / cdelt;
     bounds[1] = crpix + ( bounds[1]- crval ) / cdelt;
  }
  else
  {
     bounds[0] = crval + cdelt * ( bounds[0] - crpix );
     bounds[1] = crval + cdelt * ( bounds[1] - crpix );
  }

  return 1;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_update                               */
/*                                                             */
/* Description:                                                */
/*      Update filter values                                   */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter     *pFilter    - filter               */
/*             long          cptno      - component number     */
/*             char         *name       - filter name          */
/*             char         *ddname     - component name       */
/*             dmDataType    type       - data type            */
/*             void         *value1     - values               */
/*             void         *value2     - values               */
/*             long          nvalues    - number of values     */
/*             int           incnull    - flag                 */
/*             dmBool        update     - flag                 */
/*             int          *remove     - state flag           */
/*             int          *dirty      - state flag           */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             int  retval  - dmSUCCESS or dmFAILURE           */
/*                                                             */
/***************************************************************/
int dmi_filter_update( dmFilter *pFilter, long cptNo, char *name, char *ddname,
		       dmDataType type, void *value1, void *value2, 
		       long nvalues, int incnull, dmBool update,
                       int *remove, int *dirty )
{
  dmDataType   filterType = DT_UNK;
  dmRangeList *pRange = NULL;

  void *vvals1 = NULL;
  void *vvals2 = NULL;
  (void) ddname;

  char *origName = NULL;

  long cptOffset  = cptNo-1;
  long null       = UT_FALSE;
  int  retval     = dmSUCCESS;
  int  is_alloced = 0;

#if DEBUG
  dmi_filter_print( "FILTER UPDATE", pFilter );
#endif


  if ( !pFilter )
    return dmFAILURE;
 
  if ( name && cs_uceq( name, "NULL" ))
  {
    if ( dmverbose >= 5 )
      fprintf( stderr, "Update to set null filter\n" );
    dmi_filter_set_null( pFilter );
    *dirty = 1;
    return retval;
  }

  if ( pFilter->ranges )
    pRange = &pFilter->ranges[cptOffset];
 
  /* Get the old values */
  if ( update ) 
  {
    if ( pFilter->null )
    {
      /* Intersect with null is always null */
 
      if ( dmverbose >= 5 )
        fprintf( stderr, "Update to null filter\n" );
 
     return dmSUCCESS;
    }

    origName = pFilter->name;
  } 

  if ( update )
    null = dmi_rangelist_intersect( pRange, type, value1, value2, nvalues, incnull, UT_TRUE );
  else
  {
    dmi_rangelist_set( pRange, type, value1, value2, nvalues );

    if ( incnull )
      dmi_rangelist_set_incnull( pRange, incnull );
  }
 
  if ( null )
  {
    pFilter->null = UT_TRUE;
    *remove = 1;
  }

  if ( name && *name && ( !update || !origName ) )
    dmi_filter_set_tabname( pFilter, name );

  if ( pFilter->type == 0 )
  {
    if ( ar_mix_is_numeric( type ))
      filterType = DT_REAL8;
    else
      filterType = type;

    pFilter->type = filterType;
  }

  *dirty = 1;
  
  if ( is_alloced )
  {
    free( vvals1 ); 
    free( vvals2 );
  }

#if DEBUG
  dmi_filter_print( "FUF", pFilter );
#endif 

  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_invert                             */
/*                                                             */
/* Description:                                                */
/*    Invert a set of filters                                  */
/*    Let the input subspace have variables A1,...An,...       */
/*    and cpts 1,...m                                          */
/*    So the ranges on A1 in cpt 1 are A11.                    */
/*    Then the input subspace is                               */
/*     (A11*A21*..An1) u (A12*A22*...An2) ... u (A1m*A2m..*Anm)*/
/*             =   u(j) (A1j*A2j..*Anj)                        */
/*    where * represents intersection.                         */
/*                                                             */
/*    The invert subspace inverts the individual ranges and    */
/*     swaps u and *                                           */
/*       (A11'uA21'u...An1') * (A12'uA22'u...An2') *...        */
/*              (A1m' u A2m'... uAnm')                         */
/*    This has n**m cpts of the form                           */
/*        u ( A1j'*A2k'...*Anp')                               */
/*    An intermediate computational subspace with no physical  */
/*    significance is the nm-cpt subspace                      */
/*       (A11'uA21'u...An1') u (A12'uA22'u...An2') u...        */
/*       (A1m' u A2m'... uAnm')                                */
/*    We use this as storage to calculate the final subspace.  */
/*                                                             */
/*  Input parameters:                                          */
/*             dmSubspace   *rts        - subspace             */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             dmSubspace *irts                                */
/*                                                             */
/***************************************************************/
dmSubspace* dmi_subspace_invert( dmSubspace *rts )
{
 dmDescriptor   *dd     = NULL;
 dmDescriptor   *new_dd = NULL;
 dmDescriptor   *parent = NULL;
 dmFilter     ***fdir   = NULL;
 dmSubspace     *irts   = NULL;

 long cpt     = 0;
 long col     = 0;
 long trueDim = 0;
 long totCpts = 0;
 long ncols   = 0;
 long ncpts   = 0;


  if ( !rts )
    return irts;

  totCpts = rts->ncols * rts->ncpts;

  if ( totCpts == 0 )
    return irts; 

  irts = dmi_subspace_alloc( rts->ncols, 1 );
  fdir = (dmFilter***)calloc( rts->ncols, UT_SZ_P );

  ncols = rts->ncols;
  ncpts = rts->ncpts;

  if ( dmverbose >= 1 )
  {
    fprintf( stderr, "Invert subspace with %ld cols %ld cpts\n", ncols, ncpts );
  }

  /* Invert a variable */
  for ( col = 0; col < ncols; col++ )
  {
    /* Invert a vector cpt */
    /* Each new dd is a new cpt */

    dd = rts->cols[col];
    parent = dd->parent;
    fdir[col] = dmi_filter_alloc_array( NULL, dd->dim, 0, ncpts );
    trueDim = dd->dim;
    new_dd = dmi_subspace_col_open( irts, dd->name);

    if ( !new_dd )
      new_dd = dmi_descriptor_make_subspace( parent, irts );

    for ( cpt = 0; cpt < rts->ncpts; cpt++ )
    {
      dmi_subspace_col_invert_entry( dd, fdir, irts, col, cpt );
    }
  }
 
  /* Now we have a subspace with all the inverted components. */
  /* Next we need to intersect them                           */

  if ( dmverbose >= 1 )
  {
    fprintf( stderr, "Intersect entries in invert subspace %ld %ld\n", ncols, ncpts );
  }

  dmi_subspace_invert_intersect( irts, fdir, ncols, ncpts );
  dmi_filter_free_array( fdir, ncols, ncpts );

  /* Finally collapse the invert subspace */

  if ( dmverbose >= 1 )
  {
    fprintf( stderr, "Collapse invert subspace\n" );
   }
 
  dmi_subspace_collapse( NULL, irts );

  if ( dmverbose >= 1 )
  {
    fprintf( stderr, "Subspace invert complete\n" );
   }
  return irts;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_col_invert_entry                   */
/*                                                             */
/* Description:                                                */
/*      Invert values for one entry                            */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor *pColDesc   - col descriptor       */
/*             dmFilter   ***pppFilterList - filter list       */
/*             dmSubspace   *rts        - subspace ptr         */
/*             long          col        - col number           */
/*             long          cpt        - component number     */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_subspace_col_invert_entry( dmDescriptor *pColDesc,
                                    dmFilter ***pppFilterList,
                                    dmSubspace *rts, long col, long cpt )
{
  dmFilter *pFilter    = NULL;
  dmFilter *pNewFilter = NULL;

  long vcpt = 0;

  pFilter = dmi_subspace_col_get_filter( pColDesc, cpt+1 );

  if ( pFilter )
  {
    pNewFilter = pppFilterList[col][cpt];

    if ( !pNewFilter )
    {
      pNewFilter = dmi_filter_alloc( pColDesc->dim, pColDesc->type );
      pppFilterList[col][cpt] = pNewFilter;
    }

    if ( !pFilter->region )
    {
      for ( vcpt = 0; vcpt < pColDesc->dim; vcpt++ )
      {
        dmi_subspace_invert_entry_vcpt( pColDesc, vcpt+1, &pFilter->ranges[vcpt], &pNewFilter->ranges[vcpt] );
      }
    }
    else
    {
      dmBlock *pBlock = NULL;
      dmDescriptor *pCptDesc = NULL;
      char *name = NULL;
      long cptType = 0;
  
      pNewFilter->region = regInvert( pFilter->region );

      /* need to update the ranges for the region filter */
      pBlock = dmi_descriptor_get_block( pColDesc );

      if ( !rts )
      {
        rts = dmi_block_get_subspace( pBlock );
      }  

      name = dmi_descriptor_get_name( pColDesc );
      pCptDesc = dmi_block_search_identifier( pBlock, name );

      if ( pCptDesc )
      {
        if ( pCptDesc->flag & DM_FLAG_LOGICAL )
        {
          cptType = DM_FLAG_LOGICAL;
          pCptDesc = dmi_descriptor_get_coord_no( pCptDesc, 1 );
        }
        else
          cptType = DM_FLAG_PHYS;

        dmi_subspace_col_map_range( pBlock, pCptDesc, rts, pNewFilter, cptType );
      }
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_invert_entry_vcpt                  */
/*                                                             */
/* Description:                                                */
/*      Invert entry for one component of the vector structure */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor *pColDesc   - col descriptor       */
/*             long          vcpt       - component number     */
/*             dmRangeList  *pRange     - range                */
/*                                                             */
/*  Output parameters:                                         */
/*             dmRangeList  *pNewRange  - inverted range       */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_subspace_invert_entry_vcpt( dmDescriptor *pColDesc, long vcpt,
                        dmRangeList *pRange, dmRangeList *pNewRange )
{
  double tlmin = 0.0;
  double tlmax = 0.0;
  void  *tmin  = NULL;
  void  *tmax  = NULL;

  if ( pColDesc->type == DT_CHAR || pColDesc->type == DT_BIT )
    tmin = tmax = NULL;
  else
  { 
    dmi_descriptor_get_cpt_range( pColDesc, vcpt, DT_REAL8, &tlmin, &tlmax );
    tmin = &tlmin;
    tmax = &tlmax;
  }

  dmi_rangelist_invert_copy( pRange, pNewRange, pColDesc->type, tmin, tmax );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_free_array                           */
/*                                                             */
/* Description:                                                */
/*      Free the temporary filter storage                      */
/*                                                             */
/*  Input parameters:                                          */
/*             dmFilter   ***pppFilterList - filter list       */
/*             long          ncols      - number of columns    */
/*             long          ncpts      - number of components */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_filter_free_array( dmFilter ***pppFilterList, long ncols, long ncpts )
{
  long col = 0;
  long cpt = 0;

  dmFilter *pNewFilter = NULL;

  for ( col = 0; col < ncols; col++ )
  {
    for ( cpt = 0; cpt < ncpts; cpt++ )
    {
      pNewFilter = pppFilterList[col][cpt];
      dmi_filter_free( pNewFilter );    
    }
    free( pppFilterList[col] );
  }
  free( pppFilterList );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_invert_intersect                   */
/*                                                             */
/* Description:                                                */
/*      This function uses a wacky iteration method, but it    */
/*      works...                                               */
/*                                                             */
/*  Input parameters:                                          */
/*             dmSubspace   *irts       - subspace ptr         */
/*             dmFilter   ***pppFilterList - filter list       */
/*             long          ncols      - number of columns    */
/*             long          ncpts      - number of components */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_subspace_invert_intersect( dmSubspace *irts, dmFilter ***pppFilterList,
                                    long ncols, long ncpts )
{
  dmRangeList    *range1      = NULL;
  dmRangeList    *range2      = NULL;
  dmDescriptor   *dd          = NULL;
  dmFilter       *new_filter  = NULL;
  dmFilter       *irts_filter = NULL;
  dmFilter      **flist       = NULL;
  dmSubspaceData *ssdata      = NULL;
  dmDataType      type        = DT_UNK;

  regRegion      *tmp_region  = NULL;

  char  ttbuf[512];
  char *ttptr = ttbuf;

  long *index      = NULL;
  long *first_used = NULL; 
  long *vdims      = NULL; /* vector dimensionality */
  long *v_index    = NULL; /* vector index */
  long  nullcpt    = 0;
  long  offset     = 0;
  long  icpt       = 0;
  long  cpt        = 0;  /* subspace component */
  long  col        = 0;
  long  nvals      = 0;
  long  ii         = 0;
  long  vcpt       = 0; /* vector component */

  int intersect = 1; 
  int apply_empty = 1;
  int null  = 0;
  int loop  = 0;
  int ok    = 0;
  int inull = 0;

  if ( ncols == 0 || ncpts == 0 )
    return;

  index      = (long*)calloc( ncpts, UT_SZ_I );
  v_index    = (long*)calloc( ncpts, UT_SZ_I );
  first_used = (long*)calloc( ncols, UT_SZ_I );
  vdims      = (long*)calloc( ncols, UT_SZ_I );
  flist      = (dmFilter**)calloc( ncols, UT_SZ_P );
  offset = 0;
  nvals  = 1;

  for ( ii = 0; ii < ncpts; ii++ )
    nvals *= ncols;

  for ( col = 0; col < ncols; col++ )
  {
    dd = irts->cols[col];
    flist[col] = dmi_filter_alloc( dd->dim, dd->type );
    vdims[col] = dd->dim;
  }

  if ( dmverbose >= 4 )
    printf( "COLS=%ld CPTS=%ld Number of combinations = %ld\n", ncols,ncpts,nvals );

  loop = 1;
  icpt = 0;

  while ( loop > 0 )
  {
    ttptr = ttbuf;
    ok = 1;

    for ( ii = 0; ii < ncols; ii++ )
      first_used[ii] = 0;
    null = 0;
    nullcpt = 0;

    for ( cpt = 1; cpt <= ncpts && !null; cpt++ )
    {
      col = index[cpt-1];
      vcpt = v_index[cpt-1];
      dd = irts->cols[col];
      type = dd->type;
      irts_filter = flist[col];
      new_filter = pppFilterList[col][cpt-1];

      if ( dmverbose >= 4 )
        ttptr +=sprintf( ttptr, "%c%1ld", 'A' + (int)col, cpt );

      if ( new_filter )
      {
        /* Need region support here */
        if ( new_filter->region || irts_filter->region )
        {
          if ( !irts_filter->region )
            irts_filter->region = regCopyRegion( new_filter->region );
          else 
          {
            tmp_region = regIntersectRegion( irts_filter->region, new_filter->region );
            regFree( irts_filter->region );
            irts_filter->region = tmp_region;
          }

          for ( ii = 0; ii < new_filter->dim; ii++ )
          {     
            range1 = &new_filter->ranges[ii];
            range2 = &irts_filter->ranges[ii];
            inull =dmi_rangelist_intersect_mode( range1, range2, type, intersect, apply_empty );

            if ( inull )
              null = 1;
          }

          /* set vector cpt index to last cpt so we don't continue to loop on 
           * this column in this subspace cpt */
          v_index[cpt-1] = new_filter->dim-1;
        }
        else
        {
          range1 = &new_filter->ranges[vcpt];
          range2 = &irts_filter->ranges[vcpt];
          inull =dmi_rangelist_intersect_mode( range1, range2, type, intersect, apply_empty );

          if ( inull )
            null = 1;
        }
      }
      else
      {
        null = 1;
      }

      if ( null )
        nullcpt = cpt;

      if ( first_used[col] == 0 )
        first_used[col] = cpt;

    } // end while loop

    /* this intersect component is non-null. Set the filters for it. */

    if ( !null ) 
    {
      ++icpt;

      if ( icpt > irts->ncpts )
        dmi_subspace_add_cpt( irts );

      for ( col = 0; col < ncols; col++ )
      {
        dd = irts->cols[col];
        ssdata = dmi_ss_get_data( dd );

        /*--------- copy irts_filter to actual ssdata->filters --------*/

        ssdata->filters[icpt-1] = flist[col];
        flist[col] = dmi_filter_alloc( dd->dim, dd->type );
      }
    }
    else
    {
      for ( col = 0; col < ncols; col++ )
      {
        if ( first_used[col] > 0 )
        {
          irts_filter =flist[col];
          dmi_filter_clear( irts_filter );
          dmi_filter_init( irts_filter, irts_filter->dim, 0 );
        }
      }
    }

    loop = increment_index( index, v_index, vdims, ncols, ncpts, nullcpt );
  }

  free( index );
  free( first_used );
  free( v_index );
  free( vdims );
}


/***************************************************************/
/*                                                             */
/* Function:   increment_index                                 */
/*                                                             */
/* Description:                                                */
/*  Increment to next pixel in an ndim-cube of side dim        */
/*  If null is nonzero, increment that axis since all the      */
/*  rest of the (dim-null)-d subspace is zero.  Also           */
/*  allow a variable level of dimensionality for each          */
/*  pixel using pcpt, and pdim.                                */
/*  all = 5                                                    */
/*  1 5 4 2                                                    */
/*  0 4 4 2                                                    */
/*  4 4 4 4   4                                                */
/*  0 4 4 3                                                    */
/*  0 4 4 4                                                    */
/*  0 4 4 0                                                    */
/*  0 4 0 0                                                    */
/*  1 0 0 0                                                    */
/*  1 2 4 0                                                    */
/*  1 5 4 5                                                    */
/*  2 4 3 1                                                    */
/*                                                             */
/*  Input parameters:                                          */
/*             long       *pixel      - pixel number           */
/*             long       *pcpt       - component number       */
/*             long       *pdim       - dim                    */
/*             long        dim        - dim                    */
/*             long        ndim       - number of dim          */
/*             long        null       - flag                   */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             static long new_offset                          */
/*                                                             */
/***************************************************************/
static long increment_index( long* pixel, long* pcpt, long *pdim, long dim, long ndim, long null )
{
  long new_offset = 1;
  long offset     = ndim-1;
  long clear     = null-1;

  if ( null == 0 ) /* Else Clear this row */
    clear = ndim;

  while ( offset >= 0 && 
        (( pixel[offset] >= dim-1 && pcpt[offset] >= pdim[pixel[offset]]-1 )
	   || offset > clear ))
  {
    pixel[offset] = 0;
    pcpt[offset] = 0;
    --offset;
  }

  if ( offset >= 0 )
  {
    pcpt[offset]++;

    if ( pcpt[offset] >= pdim[pixel[offset]] )
    {
      pixel[offset]++;
      pcpt[offset] = 0;
    }
  }
  else
    new_offset = 0;

  return new_offset;
}


/***************************************************************/
/*                                                             */
/* Function:   increment_index                                 */
/*                                                             */
/* Description:                                                */
/*   Map logical pixel range from raw to filtered image        */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor *image    - image descriptor       */
/*             long         *llpix    - lower left             */
/*                                                             */
/*  Output parameters:                                         */
/*             long         *v_llpix  - virtual lower left     */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_map_filtered_image_subset( dmDescriptor *image, long *llpix,  long *v_llpix )
{
  long ii = 0;

  dmBinSpec     *binspec = NULL;
  dmDescriptor  *logical_image = NULL;

  logical_image = dmi_descriptor_get_unscaled( image );
  binspec = dmi_descriptor_get_aux( logical_image );

  if ( binspec && binspec->image_mins )
  {
    for ( ii = 0; ii < image->naxes; ii++ )
      v_llpix[ii] = llpix[ii] + binspec->image_mins[ii];  
  }
  else
  {
    for ( ii = 0; ii < image->naxes; ii++ )
      v_llpix[ii] = llpix[ii];
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_range_is_tstart                      */
/*                                                             */
/* Description:                                                */
/*   Determines if range is tstart                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor *pColDesc - col descriptor         */
/*             dmFilter     *pFilter  - filter                 */
/*             long          vcpt     - component number       */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
dmBool dmi_filter_range_is_tstart( dmDescriptor *pColDesc, dmFilter *pFilter, long vcpt )
{
  dmRangeList  *pRange = NULL;
  dmDescriptor *dd    = NULL;
  dmBlock      *pBlock = NULL;
  dmDataType    type  = DT_UNK;
  dmBool        retval = UT_FALSE;

  char   *name = NULL;

  double  tstart = 0.0;
  double  tstop  = 0.0;
  double  val    = 0.0;


  if ( !pColDesc )
    return retval;

  name = dmi_descriptor_get_name( pColDesc );

  if ( !cs_uceq( name, "TIME" ) )
    return retval;

  pRange = &pFilter->ranges[vcpt-1];

  if ( pRange->n != 1 )
    return retval;
 
  type   = dmi_descriptor_get_datatype( pColDesc );
  pBlock = dmi_descriptor_get_block( pColDesc );
  dd     = dmi_key_read( pBlock, "TSTART", DT_REAL8, &tstart );
  ar_mix_numeric_cast( pRange->mins, type, DT_REAL8, &val );

  if( !dd || val != tstart )
    return retval;

  dd = dmi_key_read( pBlock, "TSTOP", DT_REAL8, &tstop );
  ar_mix_numeric_cast( pRange->maxes, type, DT_REAL8, &val );

  if( !dd || val != tstop )
    return retval;

  return UT_TRUE;
}



