/*                                                                
**  Copyright (C) 2002-2012  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   filterlib2.c                                                */
/*                                                               */
/*   Functions working with subspace filters                     */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"
#include <float.h>

#define XDEBUG 1

long dmi_subspace_make_cpt_list( dmSubspace* subspace, dmDescriptor* dd, dmDescriptor** list, long* vcpts );
void dmi_filter_trim_region( dmDescriptor* dd, dmFilter* filter );
void dmi_subspace_adjust_image_bounds( dmBlock* block, dmSubspace* rts, dmDescriptor* vector, long cptNo, dmDescriptor* image, dmRangeList* range, long ty, int fixed );
int dmi_coord_fix_region_filter( dmDescriptor* dd, dmFilter* filter, int is_image, long ty );


/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_col_map_range                      */
/*                                                             */
/* Description:                                                */
/*             This function trims the filter bounds to the    */
/*             data space and insures that both physical and   */
/*             logical regions are defined when required.      */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock      *block                            */
/*              dmDescriptor *dd                               */
/*              dmSubspace   *rts                              */
/*              dmFilter     *filter                           */
/*              long          ty      - type                   */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            dmBool  dmSUCCESS or dmFAILURE                   */
/*                                                             */
/***************************************************************/
dmBool dmi_subspace_col_map_range( dmBlock* block, dmDescriptor* dd,
                         dmSubspace* rts, dmFilter* filter, long ty ) 
{
  dmDescriptor *ddlist[dmMAXDIM];
  dmDescriptor *vector = NULL;
  dmDescriptor *image  = NULL;
  dmRangeList  *range  = NULL;

  long vcptlist[dmMAXDIM];
  long vmin  = 0;
  long vmax  = 0;
  long cptNo = 1;
  long vcpt  = 0;
  long n     = 0;
  long dim   = 0;

  int is_image = (block->type == dmIMAGE);
  int fixed = 0;


  if ( !filter || filter->null ) 
    return UT_TRUE;

  if ( !ar_mix_is_numeric( filter->type ) ) 
    return UT_TRUE;

  memset(ddlist, 0, sizeof(dmDescriptor*)*dmMAXDIM);
  memset(vcptlist, 0, sizeof(long)*dmMAXDIM);

  /* Hack till we figure out the right way */
  if ( cptNo == 0 )
    cptNo = 1;

  dim = dmi_subspace_make_cpt_list( block->subspace, dd, ddlist, vcptlist );

  if ( dim > filter->dim )
  {
    sprintf( dm_errbuf, "Error: Filter and descriptor dimension mismatch for %s %ld vs %ld\n", dd->name, dim, filter->dim );
    dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
    return UT_FALSE;
  }

  if ( is_image ) 
  {
    vector = dd;

    if ( dd->cptno > 0 )
      vector = dd->parent;

    image = dmi_coord_get_image( vector );

    if ( !image )
    {
      sprintf( dm_errbuf, "Error: Descriptor %s is not a valid descriptor for filtering\n", dd->name );
      dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
      return UT_FALSE;
    }
  }

  if( filter->region && dim == 2 )
  {
    fixed = dmi_coord_fix_region_filter( dd, filter, is_image, ty );
    dmi_filter_trim_region( dd, filter );
   }

  if ( dd->cptno == 0 && cptNo > 0 )
  {
    dim = 1;
    vmin = cptNo;
    vmax = cptNo;
  }
  else
  {
   vmin = 1;
   vmax = dim;
  }

  for ( vcpt = vmin; vcpt <= vmax; vcpt++ )
  {
    cptNo = vcpt;
    range = dmi_filter_get_range( filter, cptNo );
    n = dmi_rangelist_get_size( range );

    /* Update the ifilter range */

    if ( is_image && range && n > 0 )
    {
      if ( n != 1 )
      {
        sprintf( dm_errbuf, "Attempt to use multiple ranges in image subsection on axis %s", dd->name );
        dmu_err_set( DM_ERR_PARSE,  dm_errbuf );
        return UT_FALSE;
      }

      dmi_subspace_adjust_image_bounds( block, rts, vector, cptNo, image, range, ty, fixed );
    }

    dmi_rangelist_adjust_bounds( block, range, ddlist[vcpt-1], vcptlist[vcpt-1] );
  }

  return UT_TRUE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_make_cpt_list                      */
/*                                                             */
/* Description:                                                */
/*             Make a list for each component, forced to do    */
/*             this in order to support implicit vector cols.  */
/*                                                             */
/*  Input parameters:                                          */
/*              dmSubspace    *subspace                        */
/*              dmDescriptor  *dd                              */
/*              dmDescriptor **list                            */
/*              long           vcpts                           */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            long  dim                                        */
/*                                                             */
/***************************************************************/
long dmi_subspace_make_cpt_list( dmSubspace* subspace, dmDescriptor* dd,
                                 dmDescriptor** list, long* vcpts )
{
  dmDescriptor *vector = NULL;

  long dim  = 0;
  long vcpt = 0;


  list[0] =dmi_subspace_col_open( subspace, dd->name);

  if ( list[0] )
  {
    vector = list[0];

    if ( vector->cptno > 0 && vector->parent )
      vector = vector->parent;

    dim = vector->dim;

    for ( vcpt = 1; vcpt <= vector->dim; vcpt++ )
    {
      list[vcpt-1] = vector;
      vcpts[vcpt-1] = vcpt;
    }
  }
  else
    if ( dd->name[0] == '(' ) /* implicit vector */
    {
      dim = dd->dim;

      for ( vcpt = 1; vcpt <= dd->dim; vcpt++ )
      {
        list[vcpt-1] = dmi_subspace_col_open( subspace, dd->cptNames[vcpt-1] );
        vcpts[vcpt-1] = 1;
      } 
    }

  if ( !list[0] && dd->descType == dmCOORD && !dmi_coord_transform_is_unit( dd ) ) 
  {
    /*
     * If dd is a virtual column and dmi_subspace_col_open() didn't find
     * its subspace col, it must not have one yet, so create one.
     */

    /* If this is a component coord, add its parent to the subspace */

    if ( dd->cptno > 0 )
      dd = dd->parent;

    vector = dmi_subspace_col_init( dd->block, NULL, dd, dd->type );

    if ( vector )
    {
      dim = vector->dim;

      for ( vcpt = 1; vcpt <= vector->dim; vcpt++ )
      {
        list[vcpt-1] = vector;
        vcpts[vcpt-1] = vcpt;
      }
    }
  }

  if ( !list[0] && !cs_eq( dd->name, "#row" )) 
  {
    sprintf( dm_errbuf, "DM INTERNAL ERROR: FAILED TO OPEN SUBSPACE FOR %s\n", dd->name );
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return 0;
  }

  return dim;
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_subspace_adjust_image_bounds                */
/*                                                             */
/* Description:                                                */
/*             Adjust bounds to insure image is withing        */
/*             bounds                                          */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock       *block                           */
/*              dmSubspace r  *rts                             */
/*              dmDescriptor  *vector                          */
/*              long           cptNo                           */
/*              dmDescriptor  *image                           */
/*              dmRangeList   *range                           */
/*              long           ty                              */
/*              int            fixed                           */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_subspace_adjust_image_bounds( dmBlock* block, dmSubspace* rts,
                                       dmDescriptor* vector, long cptNo,
                                       dmDescriptor* image, dmRangeList* range,
                                       long ty, int fixed )
{
  dmDescriptor     *image_ss = NULL;
  dmCoordTransform *wcsvals  = NULL;

  double bounds[2];
  double tbounds[2];
  double amax = 0;

  long axno = 0;
  long ty1  = 0;


  bounds[0] = ((double*)range->mins)[0];
  bounds[1] = ((double*)range->maxes)[0];

  image_ss = dmi_subspace_search( block, rts, "#0", &ty1, 0 );
  axno = dmi_descriptor_get_group_axis( image, vector->no, cptNo );

#if DEBUG
printf( "AXNO = %ld RAWNO = %ld VCPT = %ld IMDATE = %s\n", axno, vector->no, vcpt, dmi_descriptor_get_qualname( image, dm_errbuf ));
#endif

  wcsvals = dmi_descriptor_get_aux( vector );

  if ( ty == DM_FLAG_PHYS || fixed )
  {
    amax = image->axes[axno-1];  /* +0.5? */
    dmi_coord_transform_image_bounds( bounds, wcsvals, cptNo, 1 );

    /* In this direction, we are always going to require that the logical bounds are within the image. */
    if ( bounds[0] < 0.5 )  bounds[0] = 0.5;
    if ( bounds[1] < 0.5 )  bounds[1] = 0.5;
    if ( bounds[0] > amax ) bounds[0] = amax;
    if ( bounds[1] > amax ) bounds[1] = amax;
  }
  else
  {
    tbounds[0] = bounds[0];
    tbounds[1] = bounds[1];

    /* Forward transform from logical to physical */
    dmi_coord_transform_image_bounds( tbounds, wcsvals, cptNo, 0 );

    ((double*)range->mins)[0]  = tbounds[0];
    ((double*)range->maxes)[0] = tbounds[1];
  }

  dmi_image_subspace_set_bounds( image_ss, axno, bounds[0], bounds[1] );
}


/***************************************************************/
/*                                                             */
/* Function:   dmi_coord_fix_region_filter                     */
/*                                                             */
/* Description:                                                */
/*             Insure coordinates are in same system           */
/*                                                             */
/*  Input parameters:                                          */
/*              dmDescriptor  *dd                              */
/*              dmFilter      *filter                          */
/*              int            is_image                        */
/*              long           ty                              */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            int  fixed                                       */
/*                                                             */
/***************************************************************/
int dmi_coord_fix_region_filter( dmDescriptor* dd, dmFilter* filter,
                                 int is_image, long ty )
{
  dmDescriptor *vcoord = NULL;

  double crpix[2];
  double crval[2];
  double cdelt[2];
  double scale = 0.0;

  int fixed = 0;

  /* Handle case of world coord region: convert to physical coord region */

   vcoord = dmi_descriptor_get_coord_no( dd, 1 );

   if ( vcoord )
     dmi_coord_convert_world_region( vcoord, filter->region );

   if (  is_image )
   {
     vcoord = dd;  /* Physical transform */
     dmi_coord_region_convert_init( vcoord );
     dmi_coord_get_transform_d( vcoord, crpix, crval, cdelt, vcoord->dim );
     scale = fabs( cdelt[0] );

     if ( vcoord->dim > 1 && cdelt[0] != cdelt[1] )
       return fixed;

     filter->logical_region = regCopyRegion( filter->region );

     if(  ty == DM_FLAG_PHYS )
     {
       /* We are physical, make a logical region */
       regConvertRegion( filter->logical_region, scale, dmi_coord_region_convert, 1 ); 
     }
     else
     {
       /* We are logical, convert the  region to physical because it is now logical */

      regConvertRegion( filter->region, 1.0/scale, dmi_coord_convert_region_phys, 1 ); 

      if ( filter->regstring )
        free( filter->regstring );

      filter->regstring = malloc( (dmMAXSTR + 1 ) * UT_SZ_C );
      regComposeRegion( filter->region, filter->regstring, dmMAXSTR );
      fixed = 1;
     }
   }
   return fixed;
}

/***************************************************************/
/*                                                             */
/* Function:   dmi_filter_trim_region                          */
/*                                                             */
/* Description:                                                */
/*             Intersect input region with regExtent bounds    */
/*             to eliminate spurious negative regions          */
/*                                                             */
/*  Input parameters:                                          */
/*              dmDescriptor  *dd                              */
/*              dmFilter      *filter                          */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_filter_trim_region( dmDescriptor* dd, dmFilter* filter )
{
  regRegion *field  = NULL;
  regRegion *region = NULL;

  double fieldx[2]  = { 2 * 0.0 };
  double fieldy[2]  = { 2 * 0.0 };
  double xbounds[2] = { 2 * 0.0 };
  double ybounds[2] = { 2 * 0.0 };


  /* Here we intersect the user region with the regExtent bounds to make
   * spurious negative regions go away through the magic of regIntersectRegion */

  dmi_descriptor_get_cpt_range( dd, 1, DT_REAL8, &fieldx[0], &fieldx[1] );
  dmi_descriptor_get_cpt_range( dd, 2, DT_REAL8, &fieldy[0], &fieldy[1] );

  /* Convert FIELD() to RECTANGLE(X1,X2,Y1,Y2) */
  regResolveField( filter->region, fieldx, fieldy );

  /* Trim to bounds */
  regExtent( filter->region, fieldx, fieldy, xbounds, ybounds );
  field = regCreateRectangle( xbounds,ybounds );
  region = regIntersectRegion( field, filter->region );

  free( filter->regstring );
  regFree( filter->region);
  regFree( field );

  filter->region = region;
  filter->regstring = malloc( (dmMAXSTR + 1 ) * sizeof( char ));
  regComposeRegion( filter->region, filter->regstring, dmMAXSTR );

  if ( dmverbose >= 4 )
  {
    printf( "Region bounds for %s: %g:%g x %g:%g\n", dd->name, xbounds[0], xbounds[1], ybounds[0], ybounds[1] );
  }

  dmi_rangelist_set( &filter->ranges[0], DT_REAL8, &xbounds[0], &xbounds[1], 1 );
  dmi_rangelist_set( &filter->ranges[1], DT_REAL8, &ybounds[0], &ybounds[1], 1 );

}

