/*                                                                
**  Copyright (C) 2002-2012  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/*****************************************************************/
/*                                                               */
/*   rangelist.c                                                 */
/*                                                               */
/*   Functions to generate and verify range lists                */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"
#include <math.h>
#include <float.h>

/* Local function prototypes */
void dmi_rangelist_add_bit( dmRangeList* range, char* vmin );
void dmi_rangelist_add_c( dmRangeList* range, char* vmin, char* vmax );
void dmi_rangelist_add_d( dmRangeList* range, double vmin, double vmax );
void dmi_rangelist_clear( dmRangeList* range, dmDataType type );
void dmi_rangelist_expand_inverted_bitfield( dmRangeList* orange, const char* cval );
int  dmi_rangelist_intersect_bit( dmRangeList* oranges, char** cvals, long n, int incnull, int apply_empty );
int  dmi_rangelist_intersect_c( dmRangeList* oranges, char** mins, char** maxes, long n, int incnull, int apply_empty );
int  dmi_rangelist_intersect_d( dmRangeList* oranges, double* mins, double* maxes, long n, int incnull, int apply_empty );
void dmi_rangelist_invert_copy_bit( dmRangeList* range, dmRangeList* new_range );
void dmi_rangelist_invert_copy_c( dmRangeList* range, dmRangeList* new_range, char *tmin, char *tmax );
void dmi_rangelist_invert_copy_d( dmRangeList* range, dmRangeList* new_range, double *tmin, double *tmax, double step );
void dmi_rangelist_invert_range_bit( dmRangeList* orange, char* val );
void dmi_rangelist_invert_range_c( dmRangeList* orange, char* min, char* max, char* tmin, char* tmax );
void dmi_rangelist_invert_range_d( dmRangeList* orange, double min, double max, double *tmin, double *tmax, double step );
void dmi_rangelist_reduce_bit( dmRangeList* range );
void dmi_rangelist_reduce_c( dmRangeList* range );
void dmi_rangelist_reduce_d( dmRangeList* range );


/* ========================================================================== */
/*                                                                            */
/*                       "Public"  Functions                                  */
/*                                                                            */
/* ========================================================================== */

/***************************************************************/
/*                                                             */
/*  Function:  dmi_range_pair_copy                             */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             void       *mins                                */
/*             void       *maxes                               */
/*             dmDataType  type                                */
/*                                                             */
/*  Output parameters:                                         */
/*             void      **value1                              */
/*             void      **value2                              */
/*                                                             */
/*  Returns:                                                   */
/*              void                                           */
/*                                                             */
/***************************************************************/
void dmi_range_pair_copy( void* mins, void* maxes, long n, dmDataType type, void** value1, void** value2 )
{
 int ctype = ar_mix_is_numeric( type );
 long bytes = 0;

 if ( !n ) 
 {
  *value1 = NULL;
  *value2 = NULL;
 }

 if ( ctype )
 {
   bytes   = n * UT_SZ_D;
   *value1 = malloc( bytes );
   *value2 = malloc( bytes );
   if ( type == DT_REAL8 )
   {
     ar_copy_d( mins, *value1, n );
     ar_copy_d( maxes, *value2, n );
   } 
   else 
   {
     ar_mix_numeric_cast_array( mins,  DT_REAL8, type, *value1, n );
     ar_mix_numeric_cast_array( maxes, DT_REAL8, type, *value2, n );
   }
 } 
 else 
 {
   *value1 = cs_dup_array( mins, n );
   *value2 = cs_dup_array( maxes, n );
 }
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_add                               */
/*                                                             */
/*  Description:                                               */
/*   Append a range to the rangelist.                          */
/*   Does not sort and reduce after adding the range.          */                                               
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *range                             */
/*             dmDataType    type                              */
/*             void         *vmin                              */
/*             void         *vmax                              */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*              void                                           */
/*                                                             */
/***************************************************************/
void dmi_rangelist_add( dmRangeList* range, dmDataType type, void* vmin, void* vmax )
{
  double dmin;
  double dmax;

  switch ( type )
  {
  case DT_CHAR:
    dmi_rangelist_add_c( range, vmin, vmax );
    break;
  case DT_BIT:
    dmi_rangelist_add_bit( range, vmin );
    break;
  case DT_REAL8:
    dmi_rangelist_add_d( range, *(double*)vmin, *(double*)vmax );
    break;
  default:
    ar_mix_numeric_cast( vmin, type, DT_REAL8, &dmin );
    ar_mix_numeric_cast( vmax, type, DT_REAL8, &dmax );
    dmi_rangelist_add_d( range, dmin, dmax );
    break;
  }
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_add_rangelist                     */
/*                                                             */
/*  Description:                                               */
/*   Append the ranges in irange those of output range.        */                                           
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *orange                            */
/*             dmDataType    type                              */
/*             dmRangeList  *irange                            */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*              void                                           */
/*                                                             */
/***************************************************************/
void dmi_rangelist_add_rangelist( dmRangeList* orange, dmDataType type, dmRangeList* irange )
{
  long ii;

  void* min = NULL;
  void* max = NULL;

  if ( irange->incnull )
  {
    if ( irange->incnull == 1 )
      orange->incnull = 1;
    else if ( orange->n == 0 && irange->n == 0 )
      orange->incnull = -1;
  }

  for ( ii = 0; ii < irange->n; ii++ )
  {
    dmi_rangelist_range_pointers( irange, type, &min, &max, ii );
    dmi_rangelist_add( orange, type, min, max );
  }

  if ( orange->incnull == -1 && orange->n > 0 )
    orange->incnull = 0;
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_adjust                            */
/*                                                             */
/*  Description:                                               */
/*   It only adjusts the ranges in values if the ranges are    */
/*   set to  DBL_MIN and DT_REAL8_MAX.                         */
/*                                                             */
/*   Especially for images, it can be meaningful for the range */
/*   to go outside of the given bounds.  This is why we leave  */
/*   them as the userexplicitly sets them.                     */                                      
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *vals                              */
/*             double       *extremes                          */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*              void                                           */
/*                                                             */
/***************************************************************/
void dmi_rangelist_adjust( dmRangeList* vals, double* extremes )
{
  long ii;
  double tmin;
  double tmax;
  double* min = NULL;
  double* max = NULL;
 
  tmin = extremes[0];
  tmax = extremes[1];
 
  for ( ii = 0; ii < vals->n; ii++)
  {
    min = (double*)vals->mins + ii;
    max = (double*)vals->maxes+ ii;
  
    /* Use DSS for lower bound, if necessary */
    if (*min == -DT_REAL8_MAX ) 
      *min = tmin;
  
    /* Use DSS for upper bound, if necessary  */
    if (*max == DT_REAL8_MAX )
      *max = tmax;

    if ( *min > *max )
      *min = *max; 
  }
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_alloc_array                       */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *array                             */
/*             long          n                                 */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             dmRangeList *outArr                             */
/*                                                             */
/***************************************************************/
dmRangeList* dmi_rangelist_alloc_array( dmRangeList* array, long n )
{
 dmRangeList* outArr = NULL;

 if ( n > 0 )
 {
   if ( array )
     outArr = (dmRangeList*)realloc( array, n * sizeof( dmRangeList ));
   else
     outArr = (dmRangeList*)calloc( n, sizeof( dmRangeList ));
 }

 return outArr;
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_combine                           */
/*                                                             */
/*  Description:                                               */
/*   Get the union of two rangelists.                          */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *v                                 */
/*             dmRangeList  *u                                 */
/*             dmDataType   type                               */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_rangelist_combine( dmRangeList* v, dmRangeList* u, dmDataType type )
{
  long n;
  long k;
  int ctype = ar_mix_is_numeric( type );

  n = u->n + v->n;
  if ( v->incnull || u->incnull )
  {
    if ( v->incnull == 1 || u->incnull == 1 )
      u->incnull = 1;
    else if ( n == 0 )
      u->incnull = -1;
    else 
      u->incnull = 0;
  }
  if ( n == 0 )
    return;

  if ( ctype )
  {
    u->mins = (double*)realloc( u->mins, n * UT_SZ_D);
    u->maxes= (double*)realloc( u->maxes, n * UT_SZ_D);

    for(k = 0; k < v->n; k++)
    {
      ((double*)u->mins)[u->n+k]  = ((double*)v->mins)[k];
      ((double*)u->maxes)[u->n+k] = ((double*)v->maxes)[k];
    }
    u->n = n;
    u->nmax = n;
  } 
  else
  {
    u->mins = (char**)realloc( u->mins, n * UT_SZ_P );
    u->maxes= (char**)realloc( u->maxes, n * UT_SZ_P );
    for(k = 0; k < v->n; k++)
    {
      ((char**)u->mins)[u->n+k] = cs_dup(((char**)v->mins)[k]);
      ((char**)u->maxes)[u->n+k] = cs_dup(((char**)v->maxes)[k]);
    }
    u->n = n;
    u->nmax = n;
  }

  dmi_rangelist_reduce( u, type );
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_compare                           */
/*                                                             */
/*  Description:                                               */
/*   Compare two rangelists to see if they are equal.          */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *v                                 */
/*             dmRangeList  *u                                 */
/*             dmDataType   type                               */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*             int  retval - 1 if equal, 0 if not equal        */
/*                                                             */
/***************************************************************/
int dmi_rangelist_compare( dmRangeList* srcVals, dmRangeList* cmpVals, dmDataType type )
{
  long ii;
  int ctype;
  int retval = 1;

  if ( !srcVals->sorted ) 
    dmi_rangelist_reduce( srcVals, type );

  if ( !cmpVals->sorted )
    dmi_rangelist_reduce( cmpVals, type );

  ctype = ar_mix_is_numeric( type );

  if(srcVals->n != cmpVals->n )
  {
    retval = 0;
  }
  else if ( ctype ) 
  {
    if ( memcmp(srcVals->mins, cmpVals->mins, (srcVals->n)*UT_SZ_D) ||
	 memcmp(srcVals->maxes, cmpVals->maxes, (srcVals->n)*UT_SZ_D))
    {
      retval = 0;
    }
  }
  else 
  {
    for ( ii = 0; ii < srcVals->n; ii++ )
    {
      if ( !cs_eq( ((char**)srcVals->mins)[ii], ((char**)cmpVals->mins)[ii] ) ||
	   !cs_eq( ((char**)srcVals->maxes)[ii], ((char**)cmpVals->maxes)[ii] ) )
      {
	retval = 0;
      }
    }
  }

  if ( srcVals->incnull != cmpVals->incnull )
  {
    retval = 0;
  }

 return (retval);
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_compose_string                    */
/*                                                             */
/*  Description:                                               */
/*  Note that internally range->mins is double if it is numeric*/
/*  here we cast back to the original type when formatting     */
/*  string                                                     */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *range                             */
/*             dmDataType    type                              */
/*                                                             */
/*  Output parameters:                                         */
/*             char        **filter_buf                        */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void dmi_rangelist_compose_string( dmRangeList* range, dmDataType type, char** filter_buf )
{
 dmBool     ctype;
 dmDataType etype; 

 *filter_buf = NULL;

 if ( range && ( range->n > 0 || range->incnull == 1))
 {
   ctype = ar_mix_is_numeric( type );
   etype = type;

   *filter_buf = dmu_make_range_string( range->mins, range->maxes, range->n,
                                         range->incnull, etype, ctype, NULL );
 }
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_copy                              */
/*                                                             */
/*  Description:                                               */
/*   Copy a range list                                         */
/*   Assumes range lists are of type either double or char*    */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *a                                 */
/*             dmDataType    type                              */
/*                                                             */
/*  Output parameters:                                         */
/*             dmRangeList  *b                                 */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_rangelist_copy( dmRangeList* a, dmRangeList* b, dmDataType type )
{
 long range_no;
 long size;
 dmBool ctype;
  if ( b->n > 0 )
  {
   free( b->mins );
   free( b->maxes );
   b->mins = NULL;
   b->maxes = NULL;
  }

  b->n = a->n;
  b->sorted = a->sorted;
  b->incnull = a->incnull;
  if ( a->n > 0 )
  {
   b->nmax = b->n;
   ctype = ar_mix_is_numeric( type );
   if ( ctype )
    size = UT_SZ_D;
   else
    size = UT_SZ_P;

   b->mins = (void*)malloc( b->n * size );
   b->maxes= (void*)malloc( b->n * size );
   for ( range_no = 0; range_no < b->n; range_no++ )
   {
    if ( ctype )  
    {  
     ((double*)b->mins)[range_no]  = ((double*)a->mins)[range_no];
     ((double*)b->maxes)[range_no] = ((double*)a->maxes)[range_no];
    } else {
     ((char**)b->mins)[range_no]  = cs_dup(((char**)a->mins)[range_no]);
     ((char**)b->maxes)[range_no] = cs_dup(((char**)a->maxes)[range_no]);
    } 
   }
 }
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_exclude_null                      */
/*                                                             */
/*  Description:                                               */
/*       Set incnull flag in range struct                      */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *range                             */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_rangelist_exclude_null( dmRangeList* range )
{
  range->incnull = -1;
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_extract                           */
/*                                                             */
/*  Description:                                               */
/*       Copy range pair and incnull flag                      */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *vals                              */
/*             dmDataType    type                              */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*             void        **value1                            */
/*             void        **value2                            */
/*             long         *length                            */
/*             long         *pincnull                          */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_rangelist_extract( dmRangeList* vals, dmDataType type, void** value1, void** value2, long* length, int* pincnull )
{
  *length = vals->n;

  dmi_range_pair_copy( vals->mins, vals->maxes, vals->n, type, value1, value2 );

  *pincnull = vals->incnull;
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_extremes                          */
/*                                                             */
/*  Description:                                               */
/*   For numeric range list ONLY                               */
/*   Adjusted to assume vals is already initialized to -inf,   */
/*   inf for multiple calls over several components.           */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *range                             */
/*             dmDataType    type                              */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*             double       *vals                              */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_rangelist_extremes( dmRangeList* range, dmDataType type, double* vals )
{
  double tmin;
  double tmax;

  if ( !range || type == DT_CHAR || type == DT_BIT )
    return;

  if ( range->n == 0 )
    return;

  /* make sure rangelist is sorted before assuming extremes are at ends */
  if ( !range->sorted )
    dmi_rangelist_reduce_d( range );

  tmin = ((double*)range->mins)[0];
  tmax = ((double*)range->maxes)[range->n-1];

  /* Round up integer bins */
  if ( type != DT_REAL4 && type != DT_REAL8 )
  {
    if ( tmin == (long) tmin )
      tmin = floor_d( tmin + 0.5 ) - 0.5;

    if ( tmax == (long) tmax )
      tmax = floor_d( tmax + 0.5 ) + 0.5;
  }

  if ( tmin < vals[0] || vals[0] == -DT_REAL8_MAX )
    vals[0] = tmin;

  if ( tmax > vals[1] || vals[1] ==  DT_REAL8_MAX )
    vals[1] = tmax;

}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_free                              */
/*                                                             */
/*  Description:                                               */
/*   ZeroFilter doesn't clear the table name                   */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *r                                 */
/*             int           is_char                           */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_rangelist_free( dmRangeList* r, int is_char )
{
  long ii;

  if ( is_char )
  {
    for( ii = 0; ii < r->n; ii++ )
    {
      free( ((char**)r->mins)[ii] );
      free( ((char**)r->maxes)[ii] );
    }
  }

  if ( r->mins )  free( r->mins );
  if ( r->maxes ) free( r->maxes );

  r->mins   = NULL;
  r->maxes  = NULL;
  r->n      = 0;
  r->nmax   = 0;
  r->sorted = UT_TRUE;
  r->flag   = UT_FALSE;

}

/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_get_incnull                       */
/*                                                             */
/*  Description:                                               */
/*   return incnull flag                                       */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *range                             */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            int  incnull                                     */
/*                                                             */
/***************************************************************/
int dmi_rangelist_get_incnull( dmRangeList* range )
{
  return range->incnull;
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_get_size                          */
/*                                                             */
/*  Description:                                               */
/*   return incnull flag                                       */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *range                             */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            long retval -  number in range                   */
/*                                                             */
/***************************************************************/
long dmi_rangelist_get_size( dmRangeList* range )
{ 
  long retval = 0;

  if ( range )
    retval = range->n;

  return retval;
}

/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_ignore_null                       */
/*                                                             */
/*  Description:                                               */
/*     set incnull to 0                                        */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *range                             */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_rangelist_ignore_null( dmRangeList* range )
{
  range->incnull = 0;
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_imclude_null                      */
/*                                                             */
/*  Description:                                               */
/*     set incnull to 1                                        */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *range                             */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_rangelist_include_null( dmRangeList* range )
{
  range->incnull = 1;
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_init                              */
/*                                                             */
/*  Description:                                               */
/*   This is for initializing a stack rangelist                */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *range                             */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_rangelist_init( dmRangeList* range )
{

  range->mins    = NULL;
  range->maxes   = NULL;
  range->n       = 0;
  range->nmax    = 0;
  range->incnull = 0;
  range->sorted  = UT_TRUE;
  range->flag    = 0;

}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_intersect                         */
/*                                                             */
/*  Description:                                               */
/*   Type refers to the types of the arguments which should    */
/*    be compatible  with the type of the rangelist.           */
/*                                                             */
/*   apply_empty:                                              */
/*     if apply_empty == UT_TRUE:                              */
/*        set output ranges to null                            */
/*     if apply_empty== UT_FALSE:                              */
/*        leave oranges alone.                                 */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *oranges                           */
/*             dmDataType    type                              */
/*             void         *mins                              */
/*             void         *maxes                             */
/*             long          n                                 */
/*             int           incnull                           */
/*             int           apply_enpty                       */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            int retval                                       */
/*                                                             */
/***************************************************************/
int dmi_rangelist_intersect( dmRangeList* oranges, dmDataType type, void* mins, void* maxes, long n, int incnull, int apply_empty )
{
  int retval;
  double* dmins  = NULL;
  double* dmaxes = NULL;

  switch ( type )
  {
  case DT_BIT:
    retval = dmi_rangelist_intersect_bit( oranges, (char**)mins, 
					             n, incnull, apply_empty );
    break; 

  case DT_CHAR:
    retval = dmi_rangelist_intersect_c( oranges, (char**)mins, (char**)maxes,
					             n, incnull, apply_empty );
    break;

  case DT_REAL8:
    retval = dmi_rangelist_intersect_d( oranges, (double*)mins, (double*)maxes, 
					             n, incnull, apply_empty );
    break;

  default:
    dmins  = (double*)malloc( n * sizeof(double) );
    dmaxes = (double*)malloc( n * sizeof(double) );

    ar_mix_numeric_cast_array( mins,  type, DT_REAL8, dmins,  n );
    ar_mix_numeric_cast_array( maxes, type, DT_REAL8, dmaxes, n );
    retval = dmi_rangelist_intersect_d( oranges, dmins, dmaxes, 
					              n, incnull, apply_empty );

    free( dmins );
    free( dmaxes );
    break;
  }

  return retval;
}


/***************************************************************/
/*                                                             */
/*  Function: dmi_rangelist_intersect_mode                     */
/*                                                             */
/* Description:                                                */
/*   apply_empty:                                              */
/*     if apply_empty == UT_TRUE:                              */
/*        set output ranges to null,                           */
/*     if apply_empty== UT_FALSE:                              */
/*        leave output ranges alone.                           */
/*                                                             */
/*   intersect:                                                */
/*     != 0: get intersection of ranges and output ranges.     */
/*     == 0: get union of ranges and output ranges.            */
/*                                                             */
/*   type:                                                     */
/*     data type of BOTH ranges.                               */
/*                                                             */
/* Note:                                                       */
/*   should change the name of this to dmi_rangelist_merge()   */
/*   since it either intersects or unions.                     */
/*                                                             */
/*   Input Parameters:                                         */
/*             dmRangeList  *ranges                            */
/*             dmRangeList  *oranges                           */
/*             dmDataType    type                              */
/*             int           intersect                         */
/*             int           apply_empty                       */
/*                                                             */
/*    Output Parameters:                                       */
/*         none                                                */
/*                                                             */
/*   Returns:  int UT_TRUE or UT_FALSE                         */
/*                                                             */
/***************************************************************/
int dmi_rangelist_intersect_mode( dmRangeList* ranges, dmRangeList* oranges, dmDataType type, int intersect, int apply_empty )
{
  int null = UT_FALSE;

  if ( ar_mix_is_numeric( type ))
    type = DT_REAL8;

  if ( intersect )
  {
    null = dmi_rangelist_intersect( oranges, type, ranges->mins, ranges->maxes, 
				      ranges->n, ranges->incnull, apply_empty );
  }
  else /* union */
  {
    dmi_rangelist_combine( ranges, oranges, type );

    if ( oranges->n == 0 && !oranges->incnull )
      null = UT_TRUE;
  }

  return null;
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_intersect_self                    */
/*                                                             */
/*  Description:                                               */
/*   Intersect all of the ranges in my rangelist together.     */
/*   As if all ranges are AND'ed together.                     */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *range                             */
/*             dmDataType    type                              */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_rangelist_intersect_self( dmRangeList* range, dmDataType type )
{
  dmRangeList* new_range = NULL;
  void* min = NULL;
  void* max = NULL;

  long ii;
  int ctype = (( type == DT_CHAR ) || ( type == DT_BIT ));

  dmDataType ptype;                             /* type for ar_gen_ptr() */


  if ( ctype )
    ptype = DT_CHAR;
  else
    ptype = DT_REAL8;
  
  if ( !range || range->n < 2 )
    return;

  /* intersecting will be easier with a sorted range list */
  new_range = dmi_rangelist_alloc_array( NULL, 1 );
  dmi_rangelist_range_pointers( range, type, &min, &max, 0 );
  dmi_rangelist_add( new_range, type, min, max );
  new_range->incnull = range->incnull;

  for ( ii = 1; ii < range->n; ii++ )
  {
    /* use ar_gen_ptr instead of dmi_rangelist_range_pointers() since we 
     * want either (char**) or (double*) but not (char*) */
    min = ar_gen_ptr( range->mins, ptype, 0, ii );
    max = ar_gen_ptr( range->maxes, ptype, 0, ii );
    dmi_rangelist_intersect( new_range, type, min, max, 
			                    1, new_range->incnull, UT_FALSE );
  }

  dmi_rangelist_free( range, ctype );

  range->mins    = new_range->mins;
  range->maxes   = new_range->maxes;
  range->n       = new_range->n;
  range->nmax    = new_range->nmax;
  range->maxes   = new_range->maxes;
  range->incnull = new_range->incnull;

  free( new_range );
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_invert                            */
/*                                                             */
/*  Description:                                               */
/*   Invert the rangelist in place using DeMorgens Theorem.    */
/*   !(A + B) = (!A & !B)                                      */
/*                                                             */
/*   tmin are optional type minimum and maximum to give        */
/*   absolute endpoints for the ranges.                        */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *range                             */
/*             dmDataType    type                              */
/*             void         *tmin                              */
/*             void         *tmax                              */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_rangelist_invert( dmRangeList* range, dmDataType type, void* tmin, void* tmax )
{
  dmRangeList* new_range = NULL;
  
  new_range = dmi_rangelist_alloc_array( NULL, 1 );

  /* first invert all of the ranges, and put them in new_range */
  dmi_rangelist_invert_copy( range, new_range, type, tmin, tmax );
  dmi_rangelist_copy( new_range, range, type );

}

/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_is_default                        */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *range                             */
/*             dmDescriptor  dd                                */
/*             dmDataType    type                              */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            int  UT_TRUE or UT_FALSE                         */
/*                                                             */
/***************************************************************/
int dmi_rangelist_is_default( dmRangeList* range, dmDescriptor* dd, dmDataType type )
{
  GenericType dtlmin;
  GenericType dtlmax;

  void* tlmin = &dtlmin;
  void* tlmax = &dtlmax;

  short  smin, smax;
  float  fmin, fmax;
  double vmin, vmax;
  double dmin, dmax;
  double tdmin,tdmax;

  char* cmin  = NULL;
  char* cmax  = NULL;

  int q = UT_FALSE;
  dmDataType range_type;

  /* If no dd, we're just checking if the filter is implicit */
  if ( !dd )
    return UT_TRUE;

  if ( range->incnull )
    return UT_FALSE;

  if ( type != DT_CHAR )
  {
    if ( !( dd->flag & DM_FLAG_AXIS ))
    {
      range_type = dd->type;
      q = dmi_descriptor_get_range( dd, tlmin, tlmax );
    }
    else
    {
      range_type = DT_REAL8;
      q = dmi_coord_get_range( dd, 0, DT_REAL8, tlmin, tlmax );
    }
    if ( !q )
    {
      range_type = dd->type;
      ar_mix_type_min( dd->type, tlmin );
      ar_mix_type_max( dd->type, tlmax );
    }
  }
  switch( type )
  {
  case DT_REAL8:
    vmin = ((double*)range->mins)[0];
    vmax = ((double*)range->maxes)[0];
    switch( range_type ) 
    {
    case DT_REAL8: 
      dmin = *(double*)tlmin;
      dmax = *(double*)tlmax;
      tdmin = ( dmin - vmin );
      tdmax = ( dmax - vmax );

      if ( dmin != 0.0 )
	tdmin /= dmin;

      if ( dmax != 0.0 )
	tdmax /= dmax;

      if( (fabs( tdmin ) < DM_FILT_EPSILON ) &&
	  (fabs( tdmax ) < DM_FILT_EPSILON ) )
	return( UT_TRUE );
      break;
    case DT_INT2:
      smin = *(short*)tlmin;
      smax = *(short*)tlmax;
      dmin = fabs( vmin - (double)smin);
      dmax = fabs( vmax - (double)smax);
      if ( dmin < 0.5 && dmax < 0.5 )
	return( UT_TRUE );
      break;
    case DT_INT4:
      dmin = fabs( vmin - *(long*)tlmin );
      dmax = fabs( vmax - *(long*)tlmax );
      if ( dmin < 0.5 && dmax < 0.5 )
	return( UT_TRUE );
      break;
    case DT_REAL4:
      fmin = *(float*)tlmin;
      fmax = *(float*)tlmax;
      tdmin = ( fmin - vmin );
      tdmax = ( fmax - vmax );
      if ( fmin != 0.0 ) tdmin /= fmin;
      if ( fmax != 0.0 ) tdmax /= fmax;
      if( fabs( tdmin ) < DM_FILT_EPSILON 
	  && fabs( tdmax ) < DM_FILT_EPSILON )
	return( UT_TRUE );
      break;
    case DT_UINT2:
      dmin = fabs( vmin - *(unsigned short*)tlmin );
      dmax = fabs( vmax - *(unsigned short*)tlmax );
      if ( dmin < 0.5 && dmax < 0.5 )
	return( UT_TRUE );
      break;
    case DT_UINT4:
      dmin = fabs( vmin - *(unsigned long*)tlmin );
      dmax = fabs( vmax - *(unsigned long*)tlmax );
      if ( dmin < 0.5 && dmax < 0.5 )
	return( UT_TRUE );
      break;
    case DT_UINT1:
      dmin = fabs( vmin - *(unsigned char*)tlmin );
      dmax = fabs( vmax - *(unsigned char*)tlmax );
      if ( dmin < 0.5 && dmax < 0.5 )
	return UT_TRUE;
      break;
    // Code Change Bugs 9106/8852  january 30, 2012
    //   These case are undefined, just exit from switch
    case DT_BIT:
    case DT_LOGICAL:
      break;
    default:
      sprintf( dm_errbuf, "Filter data type mismatch for %s", dd->name );
      dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
      break;
    }
    break;
  case DT_CHAR:
    if ( dd->type != DT_CHAR && dd->type != DT_BIT )
    {
      sprintf( dm_errbuf, "Filter data type mismatch for text variable %s", dd->name );
      dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
      return UT_FALSE;
    } 
    else 
    {
      cmin = ((char**)range->mins)[0];
      cmax = ((char**)range->maxes)[0];
      if ( (!cmin || !(*cmin)) && (!cmax || !(*cmax)) )
	return UT_TRUE;
    }
    break;
  default:
    dmu_err_set( DM_ERR_INTERNAL, "Filter data type invalid" );
    break;
  }

  return( UT_FALSE );
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_is_empty                          */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *range                             */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            int  UT_TRUE or UT_FALSE                         */
/*                                                             */
/***************************************************************/
int dmi_rangelist_is_empty( dmRangeList* range )
{
  return  ( !range || (( !range->mins || range->n == 0 ) && !range->incnull ));
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_invert_copy                       */
/*                                                             */
/*  Description:                                               */
/*   Place an inverted version of range in new_range           */
/*   for numeric types, tmin and tmax are always considered    */
/*   double values.                                            */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *range                             */
/*             dmRangeLIst  *new_range                         */
/*             dmDataType    type                              */
/*             void         *tmin                              */
/*             void         *tmax                              */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_rangelist_invert_copy( dmRangeList* range, dmRangeList* new_range, dmDataType type, void* tmin, void* tmax )
{
  double step;

  switch ( type )
  {
  case DT_CHAR:
    dmi_rangelist_invert_copy_c( range, new_range, (char*)tmin, (char*)tmax );
    break;
  case DT_BIT:
    dmi_rangelist_invert_copy_bit( range, new_range );
    break;
  case DT_REAL8:
    dmi_rangelist_invert_copy_d( range, new_range, (double*)tmin, (double*)tmax, DBL_EPSILON );
    break;
  case DT_REAL4:
    dmi_rangelist_invert_copy_d( range, new_range, (double*)tmin, (double*)tmax, DBL_EPSILON );
    break;
  default:
    /* integer types */
    step = 1.0;
    dmi_rangelist_invert_copy_d( range, new_range, (double*)tmin, (double*)tmax, step );
    break;
  }
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_invert_range                      */
/*                                                             */
/*  Description:                                               */
/*   Invert a range.                                           */
/*   Return the inverted range in the rangelist orange, since  */
/*   an inverted range almost always leads to multiple ranges. */
/*   tmin and tmax are the optional min/max allowed for the    */
/*   inverted range. if NULL, the type's min/max values will   */
/*   be used. For numeric types, tmin and tmax are always      */
/*   considered double values.                                 */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *orange                            */
/*             dmDataType    type                              */
/*             void         *vmin                              */
/*             void         *vmax                              */
/*             void         *tmin                              */
/*             void         *tmax                              */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_rangelist_invert_range( dmRangeList* orange, dmDataType type,
                         void* vmin, void* vmax, void* tmin, void* tmax )
{
  double dmin;
  double dmax;
  double step = DBL_EPSILON;

  switch ( type )
  {
  case DT_REAL8:
    dmi_rangelist_invert_range_d( orange, *(double*)vmin, *(double*)vmax,
				  (double*)tmin, (double*)tmax, step );
    break;

  case DT_CHAR:
    dmi_rangelist_invert_range_c( orange, (char*)vmin, (char*)vmax,
				  (char*)tmin, (char*)tmax );
    break;

  case DT_BIT:
    dmi_rangelist_invert_range_bit( orange, vmin );
    break;

  default:
    /* need to convert input values to doubles */
    ar_mix_numeric_cast( vmin, type, DT_REAL8, &dmin );
    ar_mix_numeric_cast( vmax, type, DT_REAL8, &dmax );
    if ( type != DT_REAL4 ) /* must be integer type */
      step = 1.0;

    dmi_rangelist_invert_range_d( orange, dmin, dmax, 
				  (double*)tmin, (double*)tmax, step );
  }
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_pointers                          */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *vals                              */
/*                                                             */
/*  Output parameters:                                         */
/*             void        **value1                            */
/*             void        **value2                            */
/*             long          n                                 */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_rangelist_pointers( dmRangeList* vals, void** value1, void** value2, long* n )
{
  *n      = vals->n;
  *value1 = vals->mins;
  *value2 = vals->maxes;
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_print                             */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *range                             */
/*             dmDataType    type                              */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_rangelist_print( dmRangeList* range, dmDataType type )
{
  long jj;
  int ctype = ar_mix_is_numeric( type );

  for ( jj = 0; jj < range->n; jj++ )
  {
    if ( ctype )
      printf( "%g:%g ", ((double*)range->mins)[jj], ((double*)range->maxes)[jj] );
    else
      printf( "%s:%s ", ((char**)range->mins)[jj], ((char**)range->maxes)[jj] );
    
    if ( (jj+1) % 10 == 0 ) printf( "\n%-23s", " " );
  }
  printf( "\n" );
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_print2                            */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             char         *name                              */
/*             dmRangeList  *range                             */
/*             dmDataType    type                              */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_rangelist_print2( char* name, dmRangeList* range, dmDataType type )
{
  long jj;
  int ctype;

  ctype = ar_mix_is_numeric( type );
  if ( range->n < 15 ) 
  {
    printf( "%s=", name );
    for ( jj = 0; jj < range->n; jj++ )
    {
      if ( ctype )
	printf( "%g:%g ", ((double*)range->mins)[jj], ((double*)range->maxes)[jj] );  
      else
	printf( "%s:%s ", ((char**)range->mins)[jj], ((char**)range->maxes)[jj] );  
    }

    if ( range->incnull )
    {
      if ( range->incnull > 0 )
	printf( "NULL:NULL" );
      else 
	printf( "!NULL" );
    }
    printf( "\n" );
   }
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_range_pointers                    */
/*                                                             */
/*  Description:                                               */
/*   Return direct pointers to individual range values         */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *range                             */
/*             dmDataType    type                              */
/*                                                             */
/*  Output parameters:                                         */
/*             void        **values1                           */
/*             void        **values2                           */
/*             long          n                                 */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_rangelist_range_pointers( dmRangeList* range, dmDataType type, void** value1, void** value2, long n )
{
  if ( n >= 0 && n < range->n )
  {
    if ( type == DT_BIT || type == DT_CHAR )
    {
      *value1 = ((char**)range->mins)[n];
      *value2 = ((char**)range->maxes)[n];
    }
    else
    {
      *value1 = &((double*)range->mins)[n];
      *value2 = &((double*)range->maxes)[n];
    }
  }
  else
  {
    *value1 = NULL;
    *value2 = NULL;
  }
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_reduce                            */
/*                                                             */
/*  Description:                                               */
/*   Combine the ranges in the range list and sort them.       */
/*   Ranges in a range list are OR'ed together (union).        */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *range                             */
/*             dmDataType    type                              */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_rangelist_reduce( dmRangeList* range, dmDataType type )
{
  if ( type == DT_BIT )
  {
    dmi_rangelist_reduce_bit( range );
  }
  else if ( type == DT_CHAR )
  {
    dmi_rangelist_reduce_c( range );
  }
  else 
  {
    dmi_rangelist_reduce_d( range );
  }
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_set                               */
/*                                                             */
/*  Description:                                               */
/*   Set the array of ranges to be the rangelist.              */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *vals                              */
/*             dmDataType    type                              */
/*                                                             */
/*  Output parameters:                                         */
/*             void         *value1                            */
/*              void        *value1                            */
/*              long         nvalues                           &/
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmi_rangelist_set( dmRangeList* vals, dmDataType type, void* value1, void* value2,   long nvalues )
{
  size_t bytes;
  long ii;

  /* Handle case of bogus data */
  if ( value1 == NULL || value2 == NULL )
    nvalues = 0;

  if ( vals->mins ) 
  {
    if ( type == DT_CHAR || type == DT_BIT )
    {
      for ( ii = 0; ii < vals->n; ii++ )
      {
	if ( ((char**)vals->mins)[ii] )
	  free( ((char**)vals->mins)[ii] );
      }
    }
    free( (char**)vals->mins );
    vals->mins = NULL;
  }

  if ( vals->maxes )
  {
    if ( type == DT_CHAR || type == DT_BIT )
    {
      for ( ii = 0; ii < vals->n; ii++ )
      {
	if ( ((char**)vals->maxes)[ii] )
	  free( ((char**)vals->maxes)[ii] );
      }
    }
    free( (char**)vals->maxes );
    vals->maxes = NULL;
  }

  vals->n = nvalues;
  vals->nmax = nvalues;

  if ( nvalues == 0 )
  {
    return;
  } 

  if ( type == DT_CHAR || type == DT_BIT )
  {
    bytes = nvalues * UT_SZ_P;
    vals->mins = (char**)malloc( bytes );
    vals->maxes = (char**)malloc( bytes );
    for ( ii = 0; ii < nvalues; ii++ )
    {
      ((char**)vals->mins)[ii] = cs_dup( ((char**)value1)[ii] );
      ((char**)vals->maxes)[ii] = cs_dup( ((char**)value2)[ii] );
    }
  }
  else if ( type == DT_REAL8 )
  {
    bytes = nvalues * UT_SZ_D;
    vals->mins = (double*)malloc( bytes );
    vals->maxes= (double*)malloc( bytes );
    memcpy( vals->mins,  value1, bytes );
    memcpy( vals->maxes, value2, bytes );
  }
  else
  {
    bytes = nvalues * UT_SZ_D;
    vals->mins = (double*)malloc( bytes );
    vals->maxes= (double*)malloc( bytes );
    ar_mix_numeric_cast_array( value1, type, DT_REAL8, vals->mins, nvalues );
    ar_mix_numeric_cast_array( value2, type, DT_REAL8, vals->maxes, nvalues );
  }

  if ( nvalues > 1 )
    dmi_rangelist_reduce( vals, type );

}


/* ========================================================================== */
/* Function: dmi_rangelist_set_incnull                                        */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* ========================================================================== */
void dmi_rangelist_set_incnull( dmRangeList* range, int incnull )
{
  range->incnull = incnull;
}


/***************************************************************/
/*                                                             */
/*  Function:  dmi_rangelist_to_rectangle                      */
/*                                                             */
/*  Description:                                               */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmRangeList  *range1                            */
/*             dmRangeList  *range2                            */
/*             char         *buf                               */
/*                                                             */
/*  Output parameters:                                         */
//*              none                                          */
/*                                                             */
/*  Returns:                                                   */
/*            int  retval                                      */
/*                                                             */
/***************************************************************/
int dmi_rangelist_to_rectangle( dmRangeList* range1, dmRangeList* range2, char* buf )
{
  int retval = UT_FALSE;

  if (range1 && range2 && range1->n == 1 && range2->n == 1 )
  {
    sprintf( buf, "rectangle(%g,%g,%g,%g)", 
	     ((double*)range1->mins)[0], 
	     ((double*)range2->mins)[0],
	     ((double*)range1->maxes)[0], 
	     ((double*)range2->maxes)[0] );

    retval = UT_TRUE;
  }

  return retval;
}


/* ========================================================================== */
/*                                                                            */
/*                         Local Functions                                    */
/*                                                                            */
/* ========================================================================== */

/* ========================================================================== */
/* Function: dmi_rangelist_add_bit                                            */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* ========================================================================== */
void dmi_rangelist_add_bit( dmRangeList* range, char* vmin )
{
  if ( range->n >= range->nmax )
  {
    range->nmax = 10 + 2 * range->nmax;
    range->mins = realloc( range->mins, range->nmax * UT_SZ_P );
    range->maxes= realloc( range->maxes, range->nmax * UT_SZ_P );
    memset( range->maxes, 0, range->nmax * sizeof(char*) );
  }
  ((char**)range->mins)[range->n] = cs_dup( vmin );
  range->n++;

  if ( range->n > 1 )
    range->sorted = UT_FALSE;
}


/* ========================================================================== */
/* Function: dmi_rangelist_add_c                                              */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* ========================================================================== */
void dmi_rangelist_add_c( dmRangeList* range, char* vmin, char* vmax )
{
  if ( range->n >= range->nmax )
  {
    range->nmax = 10 + 2 * range->nmax;
    range->mins = realloc( range->mins, range->nmax * UT_SZ_P );
    range->maxes= realloc( range->maxes, range->nmax * UT_SZ_P );
  }
  ((char**)range->mins)[range->n] = cs_dup( vmin );
  ((char**)range->maxes)[range->n] = cs_dup( vmax );
  range->n++;

  if ( range->n > 1 ) 
    range->sorted = UT_FALSE;
}

/* ========================================================================== */
/* Function: dmi_rangelist_add_d                                              */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* ========================================================================== */
void dmi_rangelist_add_d( dmRangeList* range, double vmin, double vmax )
{
  if ( range->n >= range->nmax )
  {
    range->nmax = 10 + 2 * range->nmax;
    range->mins = realloc( range->mins, range->nmax * UT_SZ_D );
    range->maxes= realloc( range->maxes, range->nmax * UT_SZ_D );
  }
 
  ((double*)range->mins)[range->n] = vmin;
  ((double*)range->maxes)[range->n] = vmax;

  range->n++;
  if ( range->n > 1 ) 
    range->sorted = UT_FALSE;
}


/* ========================================================================== */
/* Function: dmi_rangelist_clear                                              */
/*                                                                            */
/* Description:                                                               */
/*   Just remove the ranges in the list, don't remove the min/max arrays      */
/*   or change the nmax var.                                                  */
/*                                                                            */
/* ========================================================================== */
void dmi_rangelist_clear( dmRangeList* range, dmDataType type )
{
  long ii;
  char** cmins;
  char** cmaxes;

  if ( range )
  {
    if ( type == DT_CHAR || type == DT_BIT )
    {
      cmins  = (char**)range->mins;
      cmaxes = (char**)range->maxes;
      for ( ii = 0; ii < range->n; ii++ )
      {
	if ( cmins[ii] )
	  free( cmins[ii] );

	if ( cmaxes[ii] )
	  free( cmaxes[ii] );
      }
      memset( cmins, 0, sizeof(char*)*range->n );
      memset( cmaxes, 0, sizeof(char*)*range->n );
    }
    else
    {
      memset( range->mins, 0, sizeof(double*)*range->n );
      memset( range->maxes, 0, sizeof(double*)*range->n );
    }

    range->n       = 0;
    range->sorted  = UT_TRUE;
    range->flag    = UT_FALSE;
    range->incnull = 0;
  }
}


/* ========================================================================== */
/* Function: dmi_rangelist_expand_inverted_bitfield                           */
/*                                                                            */
/* Description:                                                               */
/*   This actually expands the bitfield using DeMorgens theorem.              */
/*   Unfortunately, this can lead to a lot of bitfields.                      */
/*                                                                            */
/* ========================================================================== */
void dmi_rangelist_expand_inverted_bitfield( dmRangeList* orange, const char* cval )
{
  const char* cptr = NULL;
  const char* bptr = cval;
  char** cvals;
  char* val  = NULL;
  char* tmp  = NULL;
  char* ptmp = NULL;

  long ii;
  long jj;
  long n = 0;
 
  if ( !bptr || *bptr != '!' )
    return;

  bptr++;
  cptr = bptr;

  /* check how big the new rangelist will be */
  while( *cptr != '\0' )
  {
    if ( *cptr == '0' || *cptr == '1' )
    {
      n++;
    }
    cptr++;
  }

  dmi_rangelist_free( orange, UT_TRUE );

  orange->n    = n;
  orange->nmax = n;
  orange->maxes = calloc( n, sizeof(char*) );

  cvals = malloc( n * sizeof(char*) );
  tmp = cs_dup(bptr);
  ptmp = tmp;

  while( *ptmp != '\0' )
  {
    *ptmp = 'x';
    ptmp++;
  }

  jj = 0;
  for ( ii = 0; ii < n; ii++ )
  {
    val = cs_dup( tmp );
    while( bptr[jj] != '0' && bptr[jj] != '1' )
      jj++;

    if ( bptr[jj] == '0' ) 
      val[jj] = '1';
    else
      val[jj] = '0';

    cvals[ii] = val;
    jj++;
 }
 orange->mins = cvals;

}


/* ========================================================================== */
/* Function: dmi_rangelist_intersect_bit                                      */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/*  Intersect the bitfield rangelist with the input list of bitfields.        */
/*                                                                            */
/*  This will currently not work if any of the bitfields are inverted         */
/*     ( ie. they start with a '!'.)                                          */
/*  This SHOULD BE okay for now.                                              */
/*                                                                            */
/* Parameters:                                                                */
/*    apply_empty:                                                            */
/*      if apply_empty == UT_TRUE:                                            */
/*        set output ranges to null,                                          */
/*      if apply_empty== UT_FALSE:                                            */
/*        leave oranges alone.                                                */
/*                                                                            */
/* ========================================================================== */
int dmi_rangelist_intersect_bit( dmRangeList* oranges, char** cvals, long n, dmBool incnull, int apply_empty )
{
  int null = UT_FALSE;
  long ii;
  long jj;
  long len;        /* length of the bit fields */
  char* buf;       /* a bit field buffer */
  char** ovals;
  char** new_cvals = NULL;

  dmBool intersect        = UT_FALSE;
  dmRangeList* tmp_range1 = NULL;
  dmRangeList* tmp_range2 = NULL;
  dmRangeList* new_range  = NULL;
 
 
  if ( n == 0 )
  {
    if ( !incnull )
    {
      if ( !apply_empty )
	return null;
      
      null = UT_TRUE;
    }
    else if ( incnull == -1 )
    {
      if ( oranges->n == 0 )
	oranges->incnull = -1;

      return null;
    }
    else /* incnull == 1 */
    {
      if ( oranges->incnull >= 0 )
      {
	oranges->n = 0;
	oranges->incnull = 1;

	return null;
      }
      else /* oranges->incnull == -1 */
      {
	oranges->n = 0;
	oranges->incnull = 0;
	null = UT_TRUE;

	return null;
      }
    }
  }

  /* intersecting should be faster on sorted range list */
  if ( !oranges->sorted )
    dmi_rangelist_reduce_bit( oranges );

  new_range  = dmi_rangelist_alloc_array( NULL, 1 );
  tmp_range1 = dmi_rangelist_alloc_array( NULL, 1 );
  tmp_range2 = dmi_rangelist_alloc_array( NULL, 1 );
  if ( oranges->n > 0 )
  {
    if ( incnull || oranges->incnull )
    {
      if ( incnull == 1 && oranges->incnull != -1 )
	oranges->incnull = 1;
      else if ( oranges->incnull == 1 && incnull != -1 )
	oranges->incnull = 1;
      else 
	oranges->incnull = 0;
    }

    if ( n > 0 )
    {
      /* intersect the two bit fields */
      ovals = (char**)oranges->mins;
      len   = strlen(ovals[0]) + 2;          /* +2 to allow for leading '!' */

      buf = (char*)malloc( len * sizeof(char) );
      for ( ii = 0; ii < oranges->n; ii++ )
      {
	for ( jj = 0; jj < n; jj++ )
	{
	  intersect = dmu_bit_intersect( ovals[ii], cvals[jj], buf );
	  if ( intersect )
	  {
	    /* make sure we don't add any null filters */
	    if ( strcmp( buf, "!" ))
	      dmi_rangelist_add_bit( new_range, buf );  
	  }
	  else
	  {
	    /* This means at least one of the input bit fields is inverted.
	     * We must expand the inverted fields, and intersect the resulting
	     * range lists. Then add the intersection to the new range list */
	    if ( ovals[ii][0] == '!' )
	      dmi_rangelist_expand_inverted_bitfield( tmp_range1, ovals[ii] );
	    else
	      dmi_rangelist_copy( oranges, tmp_range1, DT_BIT );

	    if ( cvals[jj][0] == '!' )
	      dmi_rangelist_expand_inverted_bitfield( tmp_range2, cvals[jj] );
	    else
	    {
	      dmi_rangelist_free( tmp_range2, UT_TRUE );
	      tmp_range2->mins = cvals;
	      tmp_range2->n = n;
	      tmp_range2->nmax = n;
	    }

	    /* a wee bit of recursion */
	    dmi_rangelist_intersect_bit( tmp_range1, tmp_range2->mins, tmp_range2->n, incnull, apply_empty );
	    dmi_rangelist_combine( tmp_range1, new_range, DT_BIT );
	  }
	}
      }
      if ( new_range->n == 0 && !oranges->incnull )
	null = UT_TRUE;

      free( buf );
    }
  }
  else
  {
    oranges->incnull = incnull;

    if ( n > 0 )
    {
      new_range->n = n;
      new_range->nmax = n;
      new_cvals = (char**)calloc( n, sizeof( char* ));
      new_range->maxes = (char**)calloc( n, sizeof( char* ));

      for ( ii = 0; ii < n; ii++ )
	new_cvals[ii] = cs_dup( cvals[ii] );
      
      new_range->mins = new_cvals;
    }
    else if ( !oranges->incnull )
      null = UT_TRUE;
  }

  /* put the new ranges in oranges */
  dmi_rangelist_free( oranges, UT_TRUE );

  oranges->mins  = new_range->mins;
  oranges->maxes = new_range->maxes;
  oranges->nmax  = new_range->nmax;
  oranges->n     = new_range->n;

  free( new_range ); /* don't free new_range's mins/maxes */
  dmi_rangelist_free( tmp_range1, UT_TRUE );
  dmi_rangelist_free( tmp_range2, UT_TRUE );
  free( tmp_range1 );
  free( tmp_range2 );

  /* the result isn't necessarily sorted */
  if ( oranges->n > 1 )
    oranges->sorted = UT_FALSE;

  return null;
}

/* ========================================================================== */
/* Function: dmi_rangelist_intersect_c                                        */
/*                                                                            */
/* Description:                                                               */
/*   apply_empty:                                                             */
/*      if apply_empty == UT_TRUE:                                            */
/*         set oranges to null,                                               */
/*      if apply_empty== UT_FALSE:                                            */
/*         leave oranges alone.                                               */
/*                                                                            */
/* ========================================================================== */
int dmi_rangelist_intersect_c( dmRangeList* oranges, char** mins, char** maxes, long n, int incnull, int apply_empty )
{
  int    null   = UT_FALSE;
  char** cvals1 = NULL;
  char** cvals2 = NULL;

  long ii;
  long nvals = 0;

  if ( n == 0 )
  {
    if ( !incnull )
    {
      if ( !apply_empty )
	return null;

      null = UT_TRUE;
    }
    else if ( incnull == -1 )
    {
      if ( oranges->n == 0 )
	oranges->incnull = -1 ;

      return null;
    }
    else /* incnull == 1 */
    {
      if ( oranges->incnull >= 0 )
      {
	oranges->n = 0;
	oranges->incnull = 1;
	return null;
      }
      else /* oranges->incnull == -1 */
      {
	oranges->n = 0;
	oranges->incnull = 0;
	null = UT_TRUE;
	return null;
      }
    }
  }/* end if n==0 */

  /* intersecting will be faster if range list is sorted */
  if ( !oranges->sorted )
    dmi_rangelist_reduce_c( oranges );

  if ( oranges->n > 0 )
  {
    if ( incnull || oranges->incnull )
    {
      if ( incnull == 1 && oranges->incnull != -1 )
	oranges->incnull = 1;
      else if ( oranges->incnull == 1 && incnull != -1 )
	oranges->incnull = 1;
      else 
	oranges->incnull = 0;
    }

    if ( n > 0 )
    {
      ar_mix_intersect(  oranges->mins, oranges->maxes, oranges->n,
			 mins, maxes, n, DT_CHAR,
			 (void**)&cvals1, (void**)&cvals2, &nvals );
      if ( !nvals && oranges->incnull != 1 )
      {
	null = UT_TRUE;
	oranges->incnull = 0;
      }
    }
  }
  else
  {
    oranges->incnull = incnull;
    if ( n > 0 )
    {
      cvals1 = (char**)calloc( n, sizeof( char* ));
      cvals2 = (char**)calloc( n, sizeof( char* ));
      nvals = n;
      for ( ii = 0; ii < n; ii++ )
      {
	cvals1[ii] = cs_dup( mins[ii] );
	cvals2[ii] = cs_dup( maxes[ii] );
      }
    }
    else if ( !oranges->incnull )
    {
      null = UT_TRUE;
    }
  }

  /* put the new ranges in oranges */
  for ( ii = 0; ii < oranges->n; ii++ )
  {
    if ( ((char**)oranges->mins)[ii] ) 
      free( ((char**)oranges->mins)[ii] );

    if ( ((char**)oranges->maxes)[ii] ) 
      free( ((char**)oranges->maxes)[ii] );
  }

  if ( oranges->mins )
    free( (char**)oranges->mins );

  if ( oranges->maxes ) 
    free( (char**)oranges->maxes );

  oranges->mins  = cvals1;
  oranges->maxes = cvals2;
  oranges->n     = nvals;
  oranges->nmax  = nvals; 

  if ( oranges->n > 1 )
    oranges->sorted = UT_FALSE;

 return null;
}


/* ========================================================================== */
/* Function: dmi_rangelist_intersect_d                                        */
/*                                                                            */
/* Description:                                                               */
/*   Intersect the rangelist with the input array of ranges.                  */
/*                                                                            */
/*   apply_empty:                                                             */
/*     if apply_empty == UT_TRUE:                                             */
/*        set output ranges to null,                                          */
/*     if apply_empty== UT_FALSE:                                             */
/*        leave output ranges alone.                                          */
/*                                                                            */
/* ========================================================================== */
int dmi_rangelist_intersect_d( dmRangeList* oranges, double* mins, double* maxes, long n, dmBool incnull, int apply_empty )
{
  long ii;
  long nvals;
  int  null = UT_FALSE;

  double* dvals1 = NULL;
  double* dvals2 = NULL;

  if ( n == 0 )
  {
    if ( !incnull )
    {
      if ( !apply_empty )
	return null;
      
      null = UT_TRUE;
    }
    else if ( incnull == -1 )
    {
      if ( oranges->n == 0 )
	oranges->incnull = -1 ;

      return null;
    }
    else /* incnull == 1 */
    {
      if ( oranges->incnull >= 0 )
      {
	oranges->n = 0;
	oranges->incnull = 1;
	return null;
      }
      else /* oranges->incnull == -1 */
      {
	oranges->n = 0;
	oranges->incnull = 0;
	null = UT_TRUE;
	return null;
      }
    }
  }

  /* intersecting should be faster with sorted range list */
  if ( !oranges->sorted )
    dmi_rangelist_reduce_d( oranges );

  if ( oranges->mins )
  {
    if ( incnull || oranges->incnull )
    {
      if ( incnull == 1 && oranges->incnull != -1 )
	oranges->incnull = 1;
      else if ( oranges->incnull == 1 && incnull != -1 )
	oranges->incnull = 1;
      else 
	oranges->incnull = 0;
    }

    if ( n > 0 )
    {
      ar_mix_intersect(  oranges->mins, oranges->maxes, oranges->n,
			 mins, maxes, n, DT_REAL8,
			 (void**)&dvals1, (void**)&dvals2, &nvals );
   
      if( !nvals && oranges->incnull != 1 )
      {
	null = UT_TRUE;
	oranges->incnull = 0;
      }
    }

    free( oranges->mins );
    free( oranges->maxes );

  }
  else 
  {
    nvals = n;
    oranges->incnull = incnull;
    if ( n > 0 ) 
    {
      dvals1 = (double*)calloc( n, UT_SZ_D );
      dvals2 = (double*)calloc( n, UT_SZ_D );

      for ( ii = 0; ii < n; ii++ )
      {
	dvals1[ii] = mins[ii];
	dvals2[ii] = maxes[ii];
      }    
    }
    else if ( !oranges->incnull )
      null = UT_TRUE;
  }

  oranges->mins  = dvals1;
  oranges->maxes = dvals2;
  oranges->n     = nvals;
  oranges->nmax  = nvals;

  /* result isn't necessarily sorted */
  if ( oranges->n > 1 )
   oranges->sorted = UT_FALSE;

  return null;
}


/* ========================================================================== */
/* Function: dmi_rangelist_invert_copy_bit                                    */
/*                                                                            */
/* Description:                                                               */
/*   Place an inverted version of range in new_range                          */
/*                                                                            */
/* ========================================================================== */
void dmi_rangelist_invert_copy_bit( dmRangeList* range, dmRangeList* new_range )
{
  long   ii;
  char** vals;
  dmRangeList* tmp_range = NULL;

  dmi_rangelist_free( new_range, UT_TRUE );

  /* inverting will be easier with a sorted range list */
  if ( !range->sorted )
    dmi_rangelist_reduce_bit( range );
  
  vals = (char**)range->mins;

  /* first invert all of the ranges, and put them in new_range */
  tmp_range = dmi_rangelist_alloc_array( NULL, 1 );
  for ( ii = 0; ii < range->n; ii++ )
  {
    dmi_rangelist_invert_range_bit( tmp_range, vals[ii] );
    dmi_rangelist_add_rangelist( new_range, DT_BIT, tmp_range );
  }

  /* then, AND the new ranges together */
  dmi_rangelist_intersect_self( new_range, DT_BIT );
  if ( range->incnull == -1 && range->n == 0 )
  {
    new_range->incnull = 1;
  }

}

/* ========================================================================== */
/* Function: dmi_rangelist_invert_copy_c                                      */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* ========================================================================== */
void dmi_rangelist_invert_copy_c( dmRangeList* range, dmRangeList* new_range, char *tmin, char *tmax )
{
  char* cmin = NULL;
  char* cmax = NULL;
  char** cmins;
  char** cmaxes;

  long ii;
  int null = UT_TRUE;

  dmi_rangelist_free( new_range, 1 );

  if ( !range->sorted )
    dmi_rangelist_reduce_c( range );

  cmins  = (char**)range->mins;
  cmaxes = (char**)range->maxes;

  cmin = tmin;
  cmax = tmax;

  for ( ii = 0; ii < range->n; ii++ )
  {
    if ( cmax && ( strcmp( cmins[ii], cmax ) > 0 ))
      break;

    if ( !cmin || ( strcmp( cmins[ii], cmin ) > 0 ))
    {
      dmi_rangelist_add_c( new_range, cmin, cmins[ii] );
      null = UT_FALSE;
    }
    if ( !cmin || ( strcmp( cmaxes[ii], cmin ) >= 0 ))
    {
      cmin = cmaxes[ii];
    }
  }
 
  if (( !cmax || ( strcmp( cmin, cmax ) < 0 )) && !null )
    dmi_rangelist_add_c( new_range, cmin, cmax );

  new_range->sorted = UT_TRUE;
  if ( range->incnull == -1 && range->n == 0 )
  {
    new_range->incnull = 1;
  }

}


/* ========================================================================== */
/* Function: dmi_rangelist_invert_copy_d                                      */
/*                                                                            */
/* Description:                                                               */
/*   Note that here, n == 0 means an all inclusive range, not a null range    */
/*                                                                            */
/* ========================================================================== */
void dmi_rangelist_invert_copy_d( dmRangeList* range, dmRangeList* new_range, double *tmin, double *tmax, double step )
{
  double dmin = -DT_REAL8_MAX;
  double dmax = DT_REAL8_MAX;
  double* dmins  = NULL;
  double* dmaxes = NULL;

  long ii;
  dmBool null = UT_TRUE;

  dmi_rangelist_free( new_range, 0 );

  if ( !range->sorted )
    dmi_rangelist_reduce_d( range );

  dmins  = (double*)range->mins;
  dmaxes = (double*)range->maxes;

  if ( tmin )
    dmin = *tmin;

  if ( tmax )
    dmax = *tmax;

  for ( ii = 0; ii < range->n; ii++ )
  {
    if ( dmins[ii] > dmax )
      break;

    if ( dmins[ii] > dmin )
    {
      dmi_rangelist_add_d( new_range, dmin, dmins[ii] - step );
      null = UT_FALSE;
    }
    if ( dmaxes[ii] >= dmin )
    {
      dmin = dmaxes[ii] + step;
      null = UT_FALSE;
    }
  }
 
  if ( dmin <= dmax && !null )
    dmi_rangelist_add_d( new_range, dmin, dmax );

  new_range->sorted = UT_TRUE;
  if ( range->incnull == -1 && range->n == 0 )
  {
    new_range->incnull = 1;
  }
}


/* ========================================================================== */
/* Function: dmi_rangelist_invert_range_bit                                   */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* ========================================================================== */
void dmi_rangelist_invert_range_bit( dmRangeList* orange, char* val )
{
  char** cvals;
  char*  inv_cval = NULL;

  dmi_rangelist_free( orange, UT_TRUE );
  cvals = (char**)malloc( sizeof(char*) );

  if ( *val == '!' )
    inv_cval = cs_dup( (val+1) );
  else
  {
    inv_cval = malloc( sizeof(char)*(strlen(val)+2) );
    inv_cval[0] = '!';
    strcpy( inv_cval+1, val );
  }

 cvals[0] = inv_cval;
 orange->n     = 1;
 orange->nmax  = 1;
 orange->mins  = cvals;
 orange->maxes = calloc( 1, sizeof(char*) );
}


/* ========================================================================== */
/* Function: dmi_rangelist_invert_range_c                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* ========================================================================== */
void dmi_rangelist_invert_range_c( dmRangeList* orange, char* min, char* max, char* tmin, char* tmax )
{
  char *cmins[2];
  char *cmaxes[2];

  cmins[0]  = tmin;
  cmaxes[0] = min;
  cmins[1]  = max;
  cmaxes[1] = tmax;

  dmi_rangelist_set( orange, DT_CHAR, cmins, cmaxes, 2 );
}


/* ========================================================================== */
/* Function: dmi_rangelist_invert_range_d                                     */
/*                                                                            */
/* Description:                                                               */
/*                                                                            */
/* ========================================================================== */
void dmi_rangelist_invert_range_d( dmRangeList* orange, double min, double max, double *tmin, double *tmax, double step )
{
  double omins[2];
  double omaxes[2];

  if ( tmin )
    omins[0] = *tmin;
  else
    omins[0] = -DT_REAL8_MAX;

  if ( tmax )
    omaxes[1] = *tmax;
  else
    omaxes[1] = DT_REAL8_MAX;

  omaxes[0] = min - step;
  omins[1]  = max + step;

  dmi_rangelist_set( orange, DT_REAL8, omins, omaxes, 2 );
}


/* ========================================================================== */
/* Function: dmi_rangelist_reduce_bit                                         */
/*                                                                            */
/* Description:                                                               */
/*   Combine the bit fields in the range list.                                */
/*   Bit fields in a range list are OR'ed together (union).                   */
/*                                                                            */
/* ========================================================================== */
void dmi_rangelist_reduce_bit( dmRangeList* range )
{
  long imax;
  long jmax;
  long ii;
  long jj;
  char* bufA;
  char* bufB;
  char** mins;
  dmBool merged = dmFALSE; /* did any get merged? */

  if ( !range )
    return;

  range->sorted = UT_TRUE;

  if ( range->incnull == -1 && range->n > 0 )
    range->incnull = 0;

  if ( range->n < 2 )
    return;

  imax = range->n - 2;
  jmax = range->n - 1;
  mins = (char**)range->mins;
  for ( ii = 0; ii <= imax; ii++ )
  {
    bufB = mins[ii];
    jj = ii + 1;
    while( jj <= jmax )
    {
      bufA = mins[jj];
      if ( dmu_bit_union( bufA, bufB, bufB ))
      {
	free( bufA );
	mins[jj] = NULL;
	merged = dmTRUE;

	/* start over with new bit field in bufB */
	jj = ii + 1;
      }
      else
	jj++;
    }
  }

  /* now remove the merged bit fields, they were set to NULL above */
  if ( merged )
  {
    ii = 0;
    for ( jj = 0; jj <= jmax; jj++ )
    {
      if ( mins[jj] != NULL )
      {
	mins[ii] = mins[jj];
	ii++;
      }
    }
    range->n = ii;
  }

  range->sorted = UT_TRUE;
}


/* ========================================================================== */
/* Function: dmi_rangelist_reduce_c                                           */
/*                                                                            */
/* Description:                                                               */
/*   Sort and reduce string ranges.                                           */
/*                                                                            */
/* ========================================================================== */
void dmi_rangelist_reduce_c( dmRangeList* range )
{
  char** mins;
  char** maxes;
  char** rmins;
  char** rmaxes;

  char* tmp = NULL;
  int* mask = NULL;
  long nmask;

  long sorted_offset;
  long last_offset;
  long nreduce;
  int resort;

  long ii;
  long n;
  long* index = NULL;
 
  if ( !range )
    return;

  range->sorted = UT_TRUE;

  if ( range->incnull == -1 && range->n > 0 )
    range->incnull = 0;

  if ( range->n < 2 )
    return;

  rmins  = (char**)range->mins;
  rmaxes = (char**)range->maxes;

  /* First make min < max */
  for ( ii = 0; ii < range->n; ii++ )
  {
    if ( strcmp( rmins[ii], rmaxes[ii] ) > 0 )
    {
      tmp        = rmins[ii];
      rmins[ii]  = rmaxes[ii];
      rmaxes[ii] = tmp;
    }
  } 

  /* Now sort and reduce */
  index = (long*)malloc( range->n * sizeof(long) );
  mask  = (int*)calloc( range->n, sizeof(int) );

  nmask = 0;
  ar_sort_c( rmins, range->n, index );
  for ( ii = 0; ii < range->n; ii++ )
  {
    sorted_offset = index[ii]-1;
    if ( sorted_offset != ii )
      resort = 1;

    if ( ii == 0 )
      last_offset = sorted_offset;
    else
    {
      if ( strcmp( rmins[sorted_offset], rmaxes[last_offset] ) <= 0 )
      {
	if ( strcmp( rmaxes[sorted_offset], rmaxes[last_offset]) > 0 )
	  rmaxes[last_offset] = rmaxes[sorted_offset];

	mask[sorted_offset] = 1;
	nmask++;
      }
      else
	last_offset = sorted_offset;
    }
  }
 
  /* Reduce */
  if ( resort || nmask > 0 )
  {
    n = 0;
    nreduce = range->n - nmask;
    if ( nreduce > 0 )
    {
      mins  = (char**)malloc( nreduce * sizeof(char*) );
      maxes = (char**)malloc( nreduce * sizeof(char*) );
      for ( ii = 0; ii < range->n; ii++ )
      {
	sorted_offset = index[ii]-1;
	if ( !mask[sorted_offset] )
	{
	  mins[n]  = cs_dup( rmins[sorted_offset] );
	  maxes[n] = cs_dup( rmaxes[sorted_offset] );
	  n++;
	}
      }
    }
    for ( ii = 0; ii < range->n; ii++ )
    {
      free( rmins[ii] );
      free( rmaxes[ii] );
    }

    free( rmins );
    free( rmaxes );

    range->mins  = mins;
    range->maxes = maxes;
    range->n     = n;
    range->nmax  = n;
  }

  free( index );
  free( mask );
}


/* ========================================================================== */
/* Function: dmi_rangelist_reduce_d                                           */
/*                                                                            */
/* Description:                                                               */
/*   Combine the ranges in the range list.                                    */
/*   Ranges in a range list are OR'ed together.                               */
/*                                                                            */
/* ========================================================================== */
void dmi_rangelist_reduce_d( dmRangeList* range )
{
  double tmp;

  long* index;
  int* mask;

  long nreduce = 0;
  long nmask = 0;
  int resort = 0;

  double* mins  = NULL;
  double* maxes = NULL;
  double* rmins = NULL;
  double* rmaxes= NULL;

  long ii;
  long n = 0;
  long sorted_offset;
  long last_offset;

  if ( !range )
    return;

  if ( range->incnull == -1 && range->n > 0 )
    range->incnull = 0;

  range->sorted = UT_TRUE;

  if ( range->n < 2 )
    return;

  rmins = range->mins;
  rmaxes= range->maxes;

  /* First make min < max */
  for ( ii = 0; ii < range->n; ii++ )
  {
    if ( rmins[ii] > rmaxes[ii] )
    {
      tmp = rmins[ii];
      rmins[ii] = rmaxes[ii];
      rmaxes[ii] = tmp;
    }
  }

  /* Now sort and reduce */
  index = (long*)malloc( range->n * UT_SZ_I );
  mask  = (int*)calloc( range->n, UT_SZ_Q );

  nmask = 0;
  ar_sort_d( rmins, range->n, index );

  for ( ii = 0; ii < range->n; ii++ )
  {
    sorted_offset = index[ii]-1;
    if ( sorted_offset != ii )
      resort = 1;

    if ( ii == 0 )
    {
      last_offset = sorted_offset;
    } 
    else 
    {
      if ( rmins[sorted_offset] <= rmaxes[last_offset] )
      {
	if ( rmaxes[sorted_offset] > rmaxes[last_offset] )
	  rmaxes[last_offset] = rmaxes[sorted_offset];

	mask[sorted_offset] = 1;
	nmask++;
      } 
      else 
      {
	last_offset = sorted_offset;
      }
    }
  }

  /* Reduce */
  if ( resort || nmask > 0 )
  {
    n = 0;
    nreduce = range->n - nmask;
    if ( nreduce > 0 )
    {
      mins = (double*)malloc(( range->n - nmask )* UT_SZ_D );
      maxes= (double*)malloc(( range->n - nmask )* UT_SZ_D );

      for ( ii = 0; ii < range->n; ii++ )
      {
	sorted_offset = index[ii]-1;
	if ( !mask[sorted_offset] )
	{
	  mins[n] = rmins[sorted_offset];
	  maxes[n]= rmaxes[sorted_offset];
	  n++;
	}
      }
    }

    free( rmins );
    free( rmaxes);

    range->mins  = mins;
    range->maxes = maxes;
    range->n     = n;
    range->nmax  = n;
  }

  free( index );
  free( mask );

  range->sorted = UT_TRUE;
}
