/*                                                                
**  Copyright (C) 2003,2005,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmrangetest.c                                               */
/*                                                               */
/*   Functions for testing rangelist values                      */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/
#include "dmi.h"

/* local testing functions */
dmBool dmi_rangelist_test_d( dmRangeList* range, double val, long start, long* indx );
dmBool dmi_rangelist_binary_test_d( dmRangeList* range, long first, long size, double val, long* indx );
dmBool dmi_rangelist_test_c( dmRangeList* range, char* val, long start, long* indx );
dmBool dmi_rangelist_binary_test_c( dmRangeList* range, long first, long size, char* val, long* indx );
dmBool dmi_rangelist_test_bit( dmRangeList* range, unsigned char* val, long length, long start, long* indx );
dmBool dmi_rangelist_test_l( dmRangeList* range, long val, long start, long* indx );
dmBool dmi_rangelist_binary_test_l( dmRangeList* range, long first, long size, long val, long* indx );
dmBool dmi_rangelist_test_ul( dmRangeList* range, unsigned long val, long start, long* indx );
dmBool dmi_rangelist_binary_test_ul( dmRangeList* range, long first, long size, unsigned long val, long* indx );
int dmi_rangelist_compare_range_d( dmRangeList* range, int rangeno, double val );
int dmi_rangelist_compare_range_ul( dmRangeList* range, int rangeno, unsigned long val );
int dmi_rangelist_compare_range_c( dmRangeList* range, int rangeno, char* val );
int dmi_rangelist_compare_range_l( dmRangeList* range, int rangeno, long val );

/* external local functions */
extern void dmi_rangelist_reduce_d( dmRangeList* range );
extern void dmi_rangelist_reduce_bit( dmRangeList* range );
extern void dmi_rangelist_reduce_c( dmRangeList* range );


/***************************************************************/
/*                                                             */
/*  Function: dmi_rangelist_test                               */
/*                                                             */
/* Description:                                                */
/*  test if a generic value is inside a range in the RangeList.*/
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmRangelist  *range -  range to test against     */
/*            dmDataType    type  -  type of the input value   */ 
/*                  (range is expected to be a compatible type)*/
/*            void         *val   -  the value to test         */
/*            long          length - only matters for bit type */
/*            long          start  -  set start >= to tell     */
/*                           algorithm where to start testing  */
/*            long         *indx - if indx is non-null and test*/
/*                                 succeeds, set indx to the   */
/*                                index of the successful range*/
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*             dmBool pass   -  dmTRUE, if value is inside a   */
/*                              range in the RangeList.        */
/*                              dmFALSE, if value is outside of*/
/*                              all ranges                     */
/*                                                             */
/***************************************************************/
dmBool dmi_rangelist_test( dmRangeList* range, dmDataType type, void* val, long length, long start, long* indx )
{
 dmBool pass = dmFALSE;

 switch (type)
 {
 case DT_CHAR:
  pass = dmi_rangelist_test_c( range, (char*)val, start, indx );
  break;
 case DT_BIT:
  pass = dmi_rangelist_test_bit( range, (unsigned char*)val, length, start, indx );
  break;
 case DT_REAL8:
  pass = dmi_rangelist_test_d( range, *(double*)val, start, indx );
  break;
 case DT_REAL4:
  pass = dmi_rangelist_test_d( range, (double)(*(float*)val), start, indx );
  break;
 case DT_LOGICAL:
  pass = dmi_rangelist_test_l( range, (long)(*(int*)val), start, indx );
  break;
 case DT_INT4:
  pass = dmi_rangelist_test_l( range, *(long*)val, start, indx );
  break;
 case DT_INT2:
  pass = dmi_rangelist_test_l( range, (long)(*(short*)val), start, indx );
  break;
 case DT_UINT4:
  pass = dmi_rangelist_test_ul( range, *(unsigned long*)val, start, indx );
  break;
 case DT_UINT2:
  pass = dmi_rangelist_test_l( range, (long)(*(unsigned short*)val), start, indx );
  break;
 case DT_UINT1:
  pass = dmi_rangelist_test_l( range, (long)(*(unsigned char*)val), start, indx );
  break;

 default:
  break;
 }

 return pass;
}


/***************************************************************/
/*                                                             */
/*  Function: dmi_rangelist_test                               */
/*                                                             */
/* Description:                                                */
/*   Test if a double value is inside a range in the RangeList.*/
/*   set start >= to try tell algorithm where to start testing.*/
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmRangelist  *range -  range to test against     */
/*            double        val                                */
/*            long          start                              */
/*            long         *indx                               */ 
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*             dmBool pass   -  dmTRUE, if value is inside a   */
/*                              range in the RangeList.        */
/*                              dmFALSE, if value is outside of*/
/*                              all ranges                     */
/*                                                             */
/***************************************************************/
dmBool dmi_rangelist_test_d( dmRangeList* range, double val, long start, long* indx )
{
 int cmp = 0;
 long i = 0;

 /* rangelist must be sorted for testing */
 if ( !range->sorted )
  dmi_rangelist_reduce_d( range );

 if ( start >= 0 )
 {
  cmp = dmi_rangelist_compare_range_d( range, start, val );
  if ( cmp > 0 )
  {
   return dmi_rangelist_binary_test_d( range, start+1, range->n - start, val, indx );
  }
  else if( cmp < 0 )
  {
   return dmi_rangelist_binary_test_d( range, 0, start, val, indx );
  }
  if ( indx )
   *indx = start;
  return dmTRUE;
 }

 return dmi_rangelist_binary_test_d( range, 0, range->n, val, indx );
}


/***************************************************************/
/*                                                             */
/*  Function: dmi_rangelist_binary_test                        */
/*                                                             */
/* Description:                                                */
/*    Use binary search to see if a double value d is inside   */
/*    a range in the Range List.                               */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmRangelist  *range -  range to test against     */
/*            long          first                              */
/*            long          size                               */
/*            double        val                                */
/*            long         *indx                               */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*             dmBool pass   -  dmTRUE, if value is inside a   */
/*                              range in the RangeList.        */
/*                              dmFALSE, if value is outside of*/
/*                              all ranges                     */
/*                                                             */
/***************************************************************/
dmBool dmi_rangelist_binary_test_d( dmRangeList* range, long first, long size, double val, long* indx )
{
 long middle = 0;
 int cmp = 0;

 while ( size > 0 )
 {
  middle = first + size / 2;
  cmp = dmi_rangelist_compare_range_d( range, middle, val );
  if ( cmp < 0 )
  {
   size = size / 2;
  }
  else if ( cmp > 0 )
  {
   size = (size-1) / 2;
   first = middle + 1;
  }
  else
  {
   if ( indx )
    *indx = middle;
   return dmTRUE;
  }
 }

 return dmFALSE;
}
 if ( size > 0 )
 {
  middle = first + size / 2;
  cmp = dmi_rangelist_compare_range_d( range, middle, val );
  if ( cmp < 0 )
  {
   return dmi_rangelist_binary_test_d( range, first, size/2, val, indx );
  }
  else if ( cmp > 0 )
  {
   return 
    dmi_rangelist_binary_test_d( range, middle+1, (size-1)/2, val, indx );
  }
  if ( indx )
   *indx = middle;
  return dmTRUE;
 }

 return dmFALSE;
}


/***************************************************************/
/*                                                             */
/*  Function: dmi_rangelist_test_c                             */
/*                                                             */
/* Description:                                                */
/*     Check to see if a string value is inside a range in     */
/*     the RangeList. Set start >= to try tell algorithm       */
/*     where to start testing.                                 */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmRangelist  *range -  range to test against     */
/*            char         *val   - string                     */
/*            long          start                              */
/*            long         *indx                               */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*             dmBool pass   -  dmTRUE, if value is inside a   */
/*                              range in the RangeList.        */
/*                              dmFALSE, if value is outside of*/
/*                              all ranges                     */
/*                                                             */
/***************************************************************/
dmBool dmi_rangelist_test_c( dmRangeList* range, char* val, long start, long* indx )
{
 int cmp;

 /* range list must be sorted for testing */
 if ( !range->sorted )
  dmi_rangelist_reduce_c( range );

 if ( start >= 0 )
 {
  cmp = dmi_rangelist_compare_range_c( range, start, val );
  if ( cmp > 0 )
  {
   return dmi_rangelist_binary_test_c( range, start+1, range->n - start, val, indx );
  }
  else if ( cmp < 0 )
  {
   return dmi_rangelist_binary_test_c( range, 0, start, val, indx );
  }
  if ( indx )
   *indx = start;
  return dmTRUE;
 }

 return dmi_rangelist_binary_test_c( range, 0, range->n, val, indx );
}

/***************************************************************/
/*                                                             */
/*  Function: dmi_rangelist_binary_test_c                      */
/*                                                             */
/* Description:                                                */
/*     Use binary search to see if a string value is inside    */
/*     a range in the RangeList. Set start >= to try tell      */
/*     algorithm where to start testing.                       */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmRangelist  *range -  range to test against     */
/*            long          first                              */
/*            long          size                               *.
/*            char         *val   - string                     */
/*            long         *indx                               */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*             dmBool pass   -  dmTRUE, if value is inside a   */
/*                              range in the RangeList.        */
/*                              dmFALSE, if value is outside of*/
/*                              all ranges                     */
/*                                                             */
/***************************************************************/
dmBool dmi_rangelist_binary_test_c( dmRangeList* range, long first, long size, char* val, long* indx )
{
 long middle;
 int cmp;

 if ( size > 0 )
 {
  middle = first + size / 2;
  cmp = dmi_rangelist_compare_range_c( range, middle, val );
  if ( cmp < 0 )
  {
   return dmi_rangelist_binary_test_c( range, first, size/2, val, indx );
  }
  else if ( cmp > 0 )
  {
   return 
    dmi_rangelist_binary_test_c( range, middle+1, (size-1)/2, val, indx );
  }
  if ( indx )
   *indx = middle;
  return dmTRUE;
 }

 return dmFALSE;
}

/***************************************************************/
/*                                                             */
/*  Function: dmi_rangelist_test_bit                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmRangelist   *range - range to test against     */
/*            unsigned char *val                               */
/*            long           length                            *.
/*            long           start                             */
/*            long          *indx                              */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*             dmBool pass   -  dmTRUE, if value is inside a   */
/*                              range in the RangeList.        */
/*                              dmFALSE, if value is outside of*/
/*                              all ranges                     */
/*                                                             */
/***************************************************************/
dmBool dmi_rangelist_test_bit( dmRangeList* range, unsigned char* val, long length, long start, long* indx )
{
 long i;

 /* range list must be sorted for testing */
 if ( !range->sorted )
  dmi_rangelist_reduce_bit( range );

 if ( start < 0 )
  start = 0;

 for ( i = start; i < range->n; i++ )
 {
  if ( dmu_bit_match( val, length, ((char**)range->mins)[i] ))
  {
   if ( indx )
    *indx = i;
   return dmTRUE;
  }
 }
 for ( i = 0; i < start; i++ )
 {
  if ( dmu_bit_match( val, length, ((char**)range->mins)[i] ))
  {
   if ( indx )
    *indx = i;
   return dmTRUE;
  }  
 }

 return dmFALSE;
}


/***************************************************************/
/*                                                             */
/*  Function: dmi_rangelist_test_l                             */
/*                                                             */
/* Description:                                                */
/*   Test if an integer value is inside any range in the       */
/*   RangeList.                                                */
/*                                                             */
/*   Set start >= 0 to try tell algorithm where to start       */
/*   testing.                                                  */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmRangelist   *range - range to test against     */
/*            long           val                               */
/*            long           start                             */
/*            long          *indx                              */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*             dmBool pass   -  dmTRUE, if value is inside a   */
/*                              range in the RangeList.        */
/*                              dmFALSE, if value is outside of*/
/*                              all ranges                     */
/*                                                             */
/***************************************************************/
dmBool dmi_rangelist_test_l( dmRangeList* range, long val, long start, long* indx )
{
 double dx = 0.1;
 int cmp = 0;

 /* rangelist must be sorted for testing */
 if ( !range->sorted )
  dmi_rangelist_reduce_d( range );

 if ( start >= 0 )
 {
  cmp = dmi_rangelist_compare_range_l( range, start, val );
  if ( cmp > 0 )
  {
   return dmi_rangelist_binary_test_l( range, start+1, range->n - start, val, indx );
  }
  else if ( cmp < 0 )
  {
   return dmi_rangelist_binary_test_l( range, 0, start, val, indx );
  }
  else
  {
   if ( indx )
    *indx = start;
   return dmTRUE;
  }
 }

 return dmi_rangelist_binary_test_l( range, 0, range->n, val, indx );
}


/***************************************************************/
/*                                                             */
/*  Function: dmi_rangelist_binary_test_l                      */
/*                                                             */
/* Description:                                                */
/*    Use binary search to see if an integer value is inside   */
/*   a range in the Range List.                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmRangelist   *range - range to test against     */
/*            long           first                             */
/*            long           size                              */
/*            long           val                               */
/*            long          *indx                              */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*             dmBool pass   -  dmTRUE, if value is inside a   */
/*                              range in the RangeList.        */
/*                              dmFALSE, if value is outside of*/
/*                              all ranges                     */
/*                                                             */
/***************************************************************/
dmBool dmi_rangelist_binary_test_l( dmRangeList* range, long first, 
                                    long size, long val, long* indx )
{
 long middle = 0;
 int cmp = 0;

 if ( size > 0 )
 {
  middle = first + size / 2;
  cmp = dmi_rangelist_compare_range_l( range, middle, val );
  if ( cmp < 0 )
  {
   return dmi_rangelist_binary_test_l( range, first, size/2, val, indx );
  }
  else if ( cmp > 0 )
  {
   return 
    dmi_rangelist_binary_test_l( range, middle+1, (size-1)/2, val, indx );
  }
  if ( indx )
   *indx = middle;
  return dmTRUE;
 }

 return dmFALSE;
}


/***************************************************************/
/*                                                             */
/*  Function: dmi_rangelist_test_ul                            */
/*                                                             */
/* Description:                                                */
/*    Test if an unsigned integer value is inside any range in */
/*    the RangeList.                                           */
/*                                                             */
/*    Set start >= 0 to try tell algorithm where to start      */
/*     testing.                                                */ 
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmRangelist   *range - range to test against     */
/*            unsigned long  val                               */
/*            long           start                             */
/*            long          *indx                              */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*             dmBool pass   -  dmTRUE, if value is inside a   */
/*                              range in the RangeList.        */
/*                              dmFALSE, if value is outside of*/
/*                              all ranges                     */
/*                                                             */
/***************************************************************/
dmBool dmi_rangelist_test_ul( dmRangeList* range, unsigned long val, long start, long* indx )
{
 double dx = 0.1;
 int cmp;

 /* range list must be sorted for testing */
 if ( !range->sorted )
  dmi_rangelist_reduce_d( range );

 if ( start >= 0 )
 {
  cmp = dmi_rangelist_compare_range_ul( range, start, val );
  if ( cmp > 0 )
  {
   return dmi_rangelist_binary_test_ul( range, start+1, range->n - start, val, indx );
  }
  else if ( cmp < 0 )
  {
   return dmi_rangelist_binary_test_ul( range, 0, start, val, indx );
  }
  else
  {
   if ( indx )
    *indx = start;
   return dmTRUE;
  }
 }

 return dmi_rangelist_binary_test_ul( range, 0, range->n, val, indx );
}


/***************************************************************/
/*                                                             */
/*  Function: dmi_rangelist_binary_test_ul                     */
/*                                                             */
/* Description:                                                */
/*    Use binary search to see if an unsigned integer value is */
/*    inside a range in the Range List.                        */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmRangelist   *range - range to test against     */
/*            long           first                             */
/*            long           size                              */
/*            unsigned long  val                               */
/*            long          *indx                              */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*             dmBool pass   -  dmTRUE, if value is inside a   */
/*                              range in the RangeList.        */
/*                              dmFALSE, if value is outside of*/
/*                              all ranges                     */
/*                                                             */
/***************************************************************/
dmBool dmi_rangelist_binary_test_ul( dmRangeList* range, long first, long size, unsigned long val, long* indx )
{
 long middle = 0;
 int cmp = 0;

 if ( size > 0 )
 {
  middle = first + size / 2;
  cmp = dmi_rangelist_compare_range_ul( range, middle, val );
  if ( cmp < 0 )
  {
   return dmi_rangelist_binary_test_ul( range, first, size/2, val, indx );
  }
  else if ( cmp > 0 )
  {
   return 
    dmi_rangelist_binary_test_ul( range, middle+1, (size-1)/2, val, indx );
  }
  if ( indx )
   *indx = middle;
  return dmTRUE;
 }

 return dmFALSE;
}

/***************************************************************/
/*                                                             */
/*  Function: dmi_rangelist_compare_range_d                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmRangelist   *range - range to test against     */
/*            int            rangeno                           */
/*            double         val                               */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*          int  -1 or 0  or 1                                 */
/*                                                             */
/***************************************************************/
int dmi_rangelist_compare_range_d( dmRangeList* range, int rangeno, double val )
{
  double minval = ((double*)range->mins)[rangeno];
  double maxval = ((double*)range->maxes)[rangeno];

  if ( val >= maxval )
    return 1;
  else if ( val != minval && val < minval )
   return -1;
  return 0;
}

/***************************************************************/
/*                                                             */
/*  Function: dmi_rangelist_compare_range_l                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmRangelist   *range - range to test against     */
/*            int            rangeno                           */
/*            long           val                               */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*          int  -1 or 0  or 1                                 */
/*                                                             */
/***************************************************************/
int dmi_rangelist_compare_range_l( dmRangeList* range, int rangeno, long val )
{
  double dx = 0.1;
  double minval = ((double*)range->mins)[rangeno];
  double maxval = ((double*)range->maxes)[rangeno] + dx;

  if ( val >= maxval )
    return 1;
  else
    if ( val != minval && val < minval )
      return -1;
  return 0;
}

/***************************************************************/
/*                                                             */
/*  Function: dmi_rangelist_compare_range_ul                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmRangelist   *range - range to test against     */
/*            int            rangeno                           */
/*            unsigned long  val                               */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*          int  -1 or 0  or 1                                 */
/*                                                             */
/***************************************************************/
int dmi_rangelist_compare_range_ul( dmRangeList* range, int rangeno, unsigned long val )
{
  double dx = 0.1;
  double minval = ((double*)range->mins)[rangeno];
  double maxval = ((double*)range->maxes)[rangeno] + dx;

  if ( val >= maxval )
    return 1;
  else
    if ( val != minval && val < minval )
      return -1;
  return 0;
}

/***************************************************************/
/*                                                             */
/*  Function: dmi_rangelist_compare_range_c                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmRangelist   *range - range to test against     */
/*            int            rangeno                           */
/*            char          *val                               */
/*                                                             */
/*  Output parameters:                                         */
/*     none                                                    */
/*                                                             */
/*  Returns:                                                   */
/*          int  -1 or 0  or 1                                 */
/*                                                             */
/***************************************************************/
int dmi_rangelist_compare_range_c( dmRangeList* range, int rangeno, char* val )
{
  double dx = 0.0;
  int cmp = 0;
  char *minval = ((char**)range->mins)[rangeno];
  char *maxval = ((char**)range->maxes)[rangeno];

  if (( cmp = strcmp( val, maxval )) > 0 )
    return 1;
  if ( cmp && strcmp( val, minval ) < 0 )
    return -1;
  return 0;
}
