/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmintersect.c                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmulib.h"


/***************************************************************/
/*                                                             */
/* Function:   dmu_intersect_array_d                           */
/*                                                             */
/* Description:                                                */
/*    Intersects [xmin:xmax] array with [ymin:ymax] array,     */
/*    reallocating [ymin:ymax] array for answers. ny is size   */
/*    of array, nymax is allocated size.                       */
/*                                                             */
/*  Input parameters:                                          */
/*             double    *xmin                                 */
/*             double    *xmax                                 */
/*             long       n                                    */
/*                                                             */
/*  Output parameters:                                         */
/*             double   **dmin                                 */
/*             double   **dmax                                 */
/*             long      *nvals                                */
/*             long      *nmax                                 */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmu_intersect_array_d( double* xmin, double* xmax, long n, 
           double** dmin, double** dmax, long* nvals, long* nmax )
{
  double *rmin = NULL;
  double *rmax = NULL;
  double *ymin = *dmin;
  double *ymax = *dmax;

  long size = 0;
  long insize = *nvals;
  long rsize = 0;
  long count = 0;
  long i = 0;
 
  if ( insize == 0 )
    return;

  if ( n == 0 )
  {
    *nvals = 0;
    return;
  }

  for ( i = 0; i < n; i++ )
  {
    if ( count + insize > rsize )
    {
      rsize += insize;
      rmin = realloc( rmin, rsize* UT_SZ_D );
      rmax = realloc( rmax, rsize* UT_SZ_D );
      if ( !rmin || !rmax )
        return;
    }
    size = insize;
    dmu_intersect_d( xmin[i], xmax[i], ymin, ymax, &rmin[count], &rmax[count], &size );
    count += size;  
  }

  free( ymin );
  free( ymax );
 
 *nvals = count;
  *dmin = rmin;
  *dmax = rmax;
  *nmax = rsize;
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_intersect_d                                 */
/*                                                             */
/* Description:                                                */
/*     Given a range array allocated with nvals filled of nmax */
/*     allocated,   intersect (xmin,xmax) and update nvals     */ 
/*                                                             */
/*  Input parameters:                                          */
/*             double     xmin                                 */
/*             double     xmax                                 */
/*             double    *dmin                                 */
/*             double    *dmax                                 */
/*                                                             */
/*  Output parameters:                                         */
/*             double    *rmin                                 */
/*             double    *rmax                                 */
/*             long      *nvals                                */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmu_intersect_d( double xmin, double xmax, double* dmin, double* dmax, 
                      double* rmin, double* rmax, long* nvals  )
{
  long n          = *nvals;
  long old_offset = 0;
  long new_offset = 0;
  long delete    = 0;

  double ymin = 0.0;
  double ymax = 0.0;

  for ( old_offset = 0; old_offset < n; old_offset++ )
  {
    ymin = dmin[old_offset];
    ymax = dmax[old_offset];
    delete = 0;

    if ( xmin <= ymax )
    {
      if ( xmax >= ymin )
      {
        if ( xmax < ymax ) /*   xmin:dmin:xmax:dmax */
          ymax = xmax;
      }
      else                /* xmin:xmax:dmin:dmax */
        delete = 1;
    }
    else
    {  /*  xmin > dmin[j] */
      if ( xmin < ymax )
      {
        if ( xmax < ymax )
        {
          /* dmin:xmin:xmax:dmax */
          ymin = xmin;
          ymax = xmax;
        }
        else
        {     /*  dmin:xmin:dmax:xmax */
          ymin = xmin;
        }
      } 
      else    /*  dmin:dmax:xmin:xmax */
        delete = 1;
    }

    if ( !delete )
    {
      rmin[new_offset] = ymin;
      rmax[new_offset]= ymax;
      ++new_offset;
    }
  }

  *nvals = new_offset; 

  /* Zero old values to avoid confusion */
  for ( old_offset = new_offset; old_offset < n; old_offset++ )
  {
    rmin[old_offset] = 0.0;
    rmax[old_offset] = 0.0;
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_union_d                                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */ 
/*                                                             */
/*  Input parameters:                                          */
/*             double     xmin                                 */
/*             double     xmax                                 */
/*                                                             */
/*  Output parameters:                                         */
/*             double    *amin                                 */
/*             double    *amax                                 */
/*             long      *nvals                                */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmu_union_d( double xmin, double xmax, double* amin, double* amax, long* nvals )
{
  amin[*nvals] = xmin;
  amax[*nvals] = xmax;
  (*nvals)++;
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_collapse_ranges_d                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */ 
/*                                                             */
/*  Input parameters:                                          */
/*             double   **dmin                                 */
/*             double   **dmax                                 */
/*             long      *nvals                                */
/*                                                             */
/*  Output parameters:                                         */
/*             lnone                                           */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmu_collapse_ranges_d( double** dmin, double** dmax, long* nvals )
{
   dmBool overlap = UT_TRUE;

  double *rmin = NULL;
  double *rmax = NULL;
  double *xmin = *dmin;
  double *xmax = *dmax;

  long *sort_index   = NULL;
  long  n            = *nvals;
  long  offset       = 0;
  long  SortedOffset = 0;
  long  LastOffset   = 0; 
  long  NewOffset    = 0;
  

  if ( n <= 1 ) 
    return;

  /* First sort on xmin */
  sort_index = (long*)calloc( n, UT_SZ_I );
  rmin = (double*)calloc( n, UT_SZ_D );
  rmax = (double*)calloc( n, UT_SZ_D );
  ar_csort_d( xmin, n, sort_index );
  overlap = UT_FALSE;
  NewOffset = -1;

  for ( offset = 0; offset < n; offset++ )
  {
    SortedOffset = sort_index[offset];

    if ( offset > 0 ) 
      overlap = ( xmax[LastOffset] >= xmin[SortedOffset] );

    if ( overlap )
    {
      /* Have an overlap*/
      if ( xmax[LastOffset] < xmax[SortedOffset] )
        rmax[NewOffset] = xmax[SortedOffset];
    }
    else 
    {
      NewOffset++;
      rmin[NewOffset] = xmin[SortedOffset];
      rmax[NewOffset] = xmax[SortedOffset];
      LastOffset = SortedOffset;
    }
  }

  free( sort_index );
  free( xmin );
  free( xmax );
 
  *dmin  = rmin;
  *dmax  = rmax;
  *nvals = NewOffset+1;
}
