/*                                                                
**  Copyright (C) 2002-2012  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmstring.c                                                  */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmulib.h"
#include <limits.h>

void dmu_map_bits( unsigned char* buf, long n, char* bits );

#define BYTE_BITS 8


/* Change this value to clip data to given no of significant digits */
#define SIG_DIGITS 0
#define XDEBUG 0


/***************************************************************/
/*                                                             */
/* Function:   dmu_format_number_d                             */
/*                                                             */
/* Description:                                                */
/*             Format a double value into a char string.       */
/*                                                             */
/*  Input parameters:                                          */
/*             const double  val -  input value                */
/*                                                             */
/*  Output parameters:                                         */
/*                   char   *buf - string form of value        */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmu_format_number_d( const double val, char *buf )
{
  if ( val == DT_REAL8_MAX )
    strcpy( buf, "+Inf" );
  else if ( val == -DT_REAL8_MAX )
    strcpy( buf, "-Inf" );
  else
    sprintf( buf, "%g", val );
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_print_number                                */
/*                                                             */
/* Description:                                                */
/*           This routine is used to format the values in a    */
/*           subspace filter string. It trims leading zeroes   */
/*                                                             */
/*  Input parameters:                                          */
/*             const double  x      -  input value             */
/*                   long    maxlen - max numbers in string    */
/*                                                             */
/*  Output parameters:                                         */
/*                   char   *buf - string form of value        */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmu_print_number( const double x, char *buf, const long maxlen )
{
  cs_copy( " ", buf ); /* Initialize */
  cs_writef_clean_d( x, "%30.15g", UT_TRUE, SIG_DIGITS, buf, maxlen );
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_lex_number                                  */
/*                                                             */
/* Description:                                                */
/*           Check for integer, real, or other                 */
/*                                                             */
/*  Input parameters:                                          */
/*             char   *buf  -  input string                    */
/*                                                             */
/*  Output parameters:                                         */
/*             double *dval -  output value                    */
/*                                                             */
/*  Returns:                                                   */
/*            int  UT_TRUE or UT_FALSE                         */
/*                                                             */
/***************************************************************/
int dmu_lex_number( char* buf, double *dval )
{
  char *ptr  = NULL;
  long  ival = 0;
 
  *dval = 0.0;
  
  ival = strtol( buf, &ptr, 10 );

  if ( ! *ptr )
  {
    *dval = (double)ival;
    return UT_TRUE;
  }
   
  *dval = strtod( buf, &ptr );

  if ( *ptr )
    return UT_FALSE;
  else
    return UT_TRUE;
}

/***************************************************************/
/*                                                             */
/* Function:   dmu_count_commas                                */
/*                                                             */
/* Description:                                                */
/*   Count commas in a string                                  */
/*   Calling functions,                                        */
/*   dmi/dmparlib.c::dmi_check_binning_special_case() and      */      
/*   dmi/dmparselib.c::dmi_parse_bin_filter() expect the return*/
/*   value to be at least 1. That is why the variable "n"      */
/*   has an initial value of 1                                 */
/*                                                             */
/*   It is not counting commas, so much as trying to determine */
/*   the number of tokens in a comma separated list.           */
/*   If there are two or more adjacent commas in the input     */
/*   string (typos, most likely), the count returned will be   */
/*   incorrect for the number of tokens in the string          */
/*                                                             */
/*  Input parameters:                                          */
/*        const char  *buf  -  input string                    */
/*                                                             */
/*  Output parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*         long number-of-commas                               */
/*                                                             */
/***************************************************************/
long dmu_count_commas( const char *buf )
{
  long  n   = 1;
  char *ptr = (char*)buf;

  if ( !ptr ) return 0;

  while ( *ptr++ )
  {
    if ( *ptr == ',' ) n++;
  }
   
  return n;
} 


/***************************************************************/
/*                                                             */
/* Function:   dmu_parse_name_value                            */
/*                                                             */
/* Description:                                                */
/*     Parse a token of the form                               */
/*         name op value                                       */
/*     where op is at most 2 bytes and can be                  */
/*           =  !=                                             */
/*           >= <= > <                                         */
/*     but not  =!  (to exclude  pos=!circle(foo))             */
/*                                                             */
/*  Input parameters:                                          */
/*           char  *buf  - input string                        */
/*                                                             */
/*  Output parameters:                                         */
/*           char  *name - output name, predeclared string of  */
/*                         length maxlen                       */
/*           char  *op   - output operator, predeclared of     */
/*                         length at least 2                   */
/*  Returns:                                                   */
/*           char *ptr   - pointer to start of value string    */
/*                                                             */
/***************************************************************/
char* dmu_parse_name_value( char *buf, char *name, long maxlen, char *op )
{
  char *iptr    = buf;
  char *optr    = name;
  char *op_ptr  = op;
  char *tptr    = NULL;
  char  opset[] = "!=<>";

  *op_ptr = '\0';
  *optr   = '\0';

  strcpy( op, "" );
  strcpy( name, "" );

  if ( !buf || !*buf )
    return NULL;

  tptr = strpbrk( buf, opset );

  if ( tptr )
  {
    /* Copy characters to name string */

    while( iptr < tptr && maxlen-- )
    { *optr = *iptr++; optr++; }

    *optr = '\0';

    /* Copy operator characters to op string. Use opset+1 to exclude ! as a trailing character. */
    *op_ptr = *tptr++; op_ptr++;

    if ( strchr( opset+1, *tptr ) )
    {  *op_ptr = *tptr++; op_ptr++; }

    *op_ptr = '\0';

    /* Copy remaining characters to value string by returning tptr */
  }
  else 
  {
    strncpy( name, buf, maxlen );
  }

  return tptr;
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_parse_token                                 */
/*                                                             */
/* Description:                                                */
/*             Return the first token terminated by character  */
/*             sep in *result, the rest of the string in the   */
/*             function return value.                          */
/*                                                             */
/*  Input parameters:                                          */
/*             char  *ptr                                      */
/*             char   sep                                      */
/*             char **result                                   */
/*             long   maxlen                                   */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            char   pointer  - pointer to char after seperator*/
/*                              in input buffer                */
/*                                                             */
/***************************************************************/
char*  dmu_parse_token( char* ptr, char sep, char** result, long maxlen )
{
  return dmu_parse_paren_token( ptr, '\0', sep, result, maxlen );
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_parse_paren_token                           */
/*                                                             */
/* Description:                                                */
/*  The ParseParen Token routine returns a token from the      */
/*  string pointer. It places the result in *result.           */
/*  If maxlen > 0, *result is taken to be a user-preallocated  */
/*  string of length maxlen                                    */
/*  If maxlen = 0, *result is allocated within the routine.    */
/*  The lsep and rsep arguments give left and right separator  */
/*  characters.                                                */
/*                                                             */
/*  (1) The routine finds the first token bounded by these     */
/*  characters. For example, if lsep = [ and rsep = ], and     */
/*  ptr =  aaa[bbb,ccc][ggg]eee, we find the result  [bbb,ccc] */
/*                                                             */
/*  (2) However, the separator characters are ignored if they  */
/*  are inside parentheses:                                    */
/*        ptr = aaa[bbb,c(ggg]h)][fff]                         */
/*  returns  [bbb,c(ggg]h)]  not [bbb,c(ggg]                   */
/*                                                             */
/*  (3) Note that lsep may be the null character, so that      */
/*     result is any string terminated by rsep.                */
/*                                                             */
/*  (4) The return value of the function is a pointer to the   */
/*  first character following rsep.                            */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             char   *ptr   - pointer to input buffer         */
/*             char    lsep  - left separator character        */
/*             char    rsep  - right separator character       */
/*             char  **result - token                          */
/*             long    maxlen - max token length               */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*             char  *pointer - pointer to next char in input  */
/*                              buffer after separator         */
/*                                                             */
/***************************************************************/
char*  dmu_parse_paren_token( char* ptr, char lsep, char rsep,
                              char** result, long maxlen )
{
  char    buf[dmMAXLEN];
  char   *iptr   = ptr;
  char   *optr   = buf;
  int     fixed  = 0;
  long    size   = dmMAXLEN;
  long    depth  = 1;
  long    pdepth = 0;
  dmBool  copy   = dmFALSE;


  /* Is the result buffer preallocated or do we need to allocate */
  /* it inside the routine?                                      */

  if ( maxlen ) 
  { 
    fixed = 1;

    if ( maxlen < dmMAXLEN )
      size = maxlen;
    strcpy( *result, "" );
  }
  else
  {
    *result = NULL;
  }


  // Watch for NULL pointer - just return now
  if ( !ptr )
    return NULL;


  copy = UT_FALSE;   /* Start copying after one iteration */

  while ( depth > 0 && *iptr != '\0' && size-- ) 
  {
    if ( copy )
      {
        *optr = *iptr++;
        ++optr;
      }

    if ( *iptr == '('  )
      ++pdepth;

    if ( *iptr == ')'  )
      --pdepth;

    if ( pdepth <= 0 ) 
    {
      if ( *iptr == rsep )
        --depth;

      if ( *iptr == lsep )
        ++depth;
    }
    copy = UT_TRUE;
  }

  *optr = '\0';

  if ( *buf ) 
  {
    if ( fixed )
      strncpy( *result, buf, maxlen );
    else
      *result = cs_dup( buf );
  }

  if ( *iptr )
    return ++iptr;
  else
    return NULL;
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_make_range_string                           */
/*                                                             */
/* Description:                                                */
/*  Format a string of   min1:max1,min2:max2,....              */
/*  Leave of min1 and maxn if equal to the type limits         */
/*  cast_limits Special case: data are double but limits are   */
/*  of given type                                              */
/*                                                             */
/*  Input parameters:                                          */
/*           void      *mins                                   */
/*           void      *maxes                                  */
/*           long       n                                      */
/*           int        incnull                                */
/*           dmDataType type                                   */
/*           dmBool     cast_limits                            */
/*           char      *fmt                                    */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*          char *result - string containg range               */
/*                                                             */
/***************************************************************/
char* dmu_make_range_string( void* mins, void* maxes, long n, 
                             int incnull, dmDataType type, 
                             dmBool cast_limits, char* fmt )
{
  char   stmp1[CARD_SIZE];
  char   stmp2[CARD_SIZE];
  char   buf[512]; 
  char  *result = NULL;
  char  *format = NULL;
  char **cvals1 = NULL;
  char **cvals2 = NULL;

  double *dvals1 = (double*)mins;
  double *dvals2 = (double*)maxes;
  double  dmin   = 0;
  double  dmax   = 0;
  double  vmin   = 0;
  double  vmax   = 0; /* Generic type buffers */

  long i = 0;
  long size = 0;


  /* Hopefully ample size */
  if ( n )
    size = 3 + 200 * n;
  else 
    size = 10;
  
  result = calloc( size, UT_SZ_C );
  strcpy( buf, "" );

  if ( type == DT_REAL8 )
  {
    ar_mix_type_min( type, &dmin );
    ar_mix_type_max( type, &dmax );
  }
  else if ( cast_limits )
  {
    ar_mix_type_min( type, &vmin );
    ar_mix_numeric_cast( &vmin, type, DT_REAL8, &dmin );
    ar_mix_type_max( type, &vmax );
    ar_mix_numeric_cast( &vmax, type, DT_REAL8, &dmax );
  }

  switch(type)
  {
    case DT_REAL4:
    case DT_REAL8: 
    case DT_INT2: 
    case DT_INT4: 
    case DT_UINT1:
    case DT_UINT2:
    case DT_UINT4:
      for(i =0; i < n; i++)
      {
	if ((dvals1[i] >= dmax) && (dvals1[i] == dvals2[i]))
	  continue;

	if ( ((i > 0) && ( i < n )) && (strcmp(result, "") != 0) )
	  strcat( result,",");
	  
	if ( fmt )
	  sprintf( buf, fmt, dvals1[i], dvals2[i] );
	else if ( i == 0 && dvals1[i] <= dmin )
	{
	  dmu_print_number( dvals2[i], stmp2, CARD_SIZE );
	  sprintf( buf,":%s", stmp2 );
	}
	else if ( i == n-1 && dvals2[i] >= dmax )
	{
	  dmu_print_number( dvals1[i], stmp2, CARD_SIZE );
	  sprintf( buf,"%s:", stmp2 );
	}
	else
        {
          dmu_print_number( dvals1[i], stmp1, CARD_SIZE );
	  dmu_print_number( dvals2[i], stmp2, CARD_SIZE );
	  sprintf( buf, "%s:%s", stmp1, stmp2 );
	}

        strcat(result, buf );	  
      }
      break;

    case DT_CHAR:
      cvals1 = (char**)mins;
      cvals2 = (char**)maxes;
      format = "%s:%s";

      for(i = 0; i < n; i++)
      {
	sprintf( buf, format, cvals1[i], cvals2[i] );
        strcat(result, buf );

        if ( i < n-1 )
	  strcat( result,",");
      }
      break;

    case DT_BIT:
      cvals1 = (char**)mins;
      format = "%s";

      for ( i=0; i<n; i++ )
      {
	sprintf( buf, format, cvals1[i] );
	strcat( result, buf );

	if ( i < n-1 )
          strcat( result, "," );
      }
      break;

    // Code Change  Bug 9106  January 31, 2012
    //   Added case to prevent falling onto error message below
  
    case DT_LOGICAL:
      break;

    default:
      sprintf( dm_errbuf, "Bad data type in filter string formatting" );
      dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
      free( result );
      return NULL;
  }

  if ( n == 0 && incnull == 1 )
  {
    strcat( result, "NULL:NULL" );
  }

  return result;
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_split_token                                 */
/*                                                             */
/* Description:                                                */
/*   Parse a token of the form                                 */
/*       name op value                                         */
/*   where op is at most 2 bytes and can be:                   */
/*       =  !=                                                 */
/*       >=  <=  <                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             char  *buf   - input string value               */
/*             long   maxlen                                   */
/*                                                             */
/*  Output parameters:                                         */
/*             char  *name                                     */
/*             char  *op                                       */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*        char *iptr - pointer to remainder of string in buf   */
/*                                                             */
/***************************************************************/
char* dmu_split_token( char *buf, char *name, long maxlen, char *op )
{ 
  char opset[] = "!=<>";
  return dmu_split_token_chars( buf, name, maxlen, op, opset ); 
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_split_token_chars                           */
/*                                                             */
/* Description:                                                */
/*   Split string into  <name><op><remainder>                  */
/*   where <op> is one of the characters in opset.             */
/*   Ignore everything within parentheses, except if the (     */
/*   character is part of opset.                               */
/*                                                             */
/*  Input parameters:                                          */
/*             char  *buf    - input string value              */
/*             long   maxlen - max len of name                 */
/*             char  *opset  - string of operator characters   */
/*                                                             */
/*  Output parameters:                                         */
/*             char  *name   - name token                      */
/*             char  *op     - op token                        */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*        char *iptr - pointer to remainder of string in buf   */
/*                                                             */
/***************************************************************/
char* dmu_split_token_chars( char *buf, char *name, long maxlen,
                             char *op, char *opset )
{
  char *iptr   = buf;
  char *optr   = name;
  char *op_ptr = op;
  char  c      = '\0';

  long depth  = 0;
  long pdepth = 0;
  long size   = maxlen;

  int seek     = 1;
  int doparens = 1;

  *op_ptr = '\0';
  *optr   = '\0';

  if ( strchr( opset, '(' ) ) doparens = 0;

  while( *iptr == ' ' )
    ++iptr;

  while ( seek && *iptr != '\0' && size-- )
  {
    c = *iptr;

    if ( c == '('  && doparens ) ++pdepth;
    if ( c == ')'  && doparens ) --pdepth;
    if ( c == '['  ) ++depth;
    if ( c == ']'  ) --depth;

    if ( pdepth == 0 && depth == 0 )
    {
      if ( strchr( opset, c ))
      {
	if ( c == '!' && *(iptr+1) != '=' )  /* ! must be part of != */
	  seek = 1;
	else
	  seek = 0;
      }
    }

    if ( seek )
      *optr++ = *iptr++; 
  }

  *optr = '\0';

  /* Delete trailing spaces */
  --optr;

  while( *optr == ' ' )
    *optr-- = '\0';

  if ( !*iptr )
    return NULL;

  /* Copy operator characters to op string; allow != but disallow =!         */
  /* since in chip=!circle(..) the ! belongs to the region not the operator  */

  if ( !seek )
  {
    *op_ptr++ = *iptr++;

    if ( strchr( opset, *iptr ) && *iptr!= '!' )
      *op_ptr++ = *iptr++;

    *op_ptr = '\0';
   }

  /* Skip spaces */
  while( *iptr && *iptr == ' ' )
    ++iptr;

  /* Copy remaining characters to value string by returning tptr */
  return iptr;
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_string_delta                                */
/*                                                             */
/* Description:                                                */
/*   Copy string                                               */
/*   If n = 1, append "!" to string                            */
/*   If n =-1, decrement last character and append "~"         */
/*   to string                                                 */
/*                                                             */
/*  Input parameters:                                          */
/*             char  *buf    - input string value              */
/*             long   n      - length                          */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*        char *nbuf - pointer to newly allocated string       */
/*                                                             */
/***************************************************************/
char* dmu_string_delta( char* buf, long n )
{
  char *nbuf = NULL;
  long  m    = 0;

  m    = strlen( buf );
  nbuf = malloc( (m+1)* UT_SZ_C );

  strcpy( nbuf, buf );

  if ( n > 0 )
    nbuf[m] = '!';
  else
  {
    nbuf[m-1]--;
    nbuf[m] = '~';
  }

  return nbuf; 
} 


/***************************************************************/
/*                                                             */
/* Function:   dmu_string_delta                                */
/*                                                             */
/* Description:                                                */
/*   Return true if buf is in the closed interval start:end    */
/*   This is a case sensitive match                            */
/*                                                             */
/*  Input parameters:                                          */
/*             char  *buf    - input string value              */
/*             char  *start  - start string value              */
/*             char  *end    - end string value                */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*        int        TRUE or FALSE                             */
/*                                                             */
/***************************************************************/
int dmu_text_in_range( char *buf, char *start, char *end )
{              
  if ( start && *start ) 
  {
    if ( cs_lt( buf, start ))
      return 0;
  }

  if ( end && *end )
  {
    if ( cs_gt( buf, end ))
      return 0;
  }
  return 1;
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_bit_match                                   */
/*                                                             */
/* Description:                                                */
/*   We test only the 'start' pattern.                         */
/*   Explicit 0 and 1 in the pattern must match.               */
/*                                                             */
/*  Input parameters:                                          */
/*    unsigned char  *buf     - input string                   */
/*             long   n       - length                         */
/*             char  *pattern - pattern match                  */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*        int  TRUE or FALSE                                   */
/*                                                             */
/***************************************************************/
int dmu_bit_match( unsigned char *buf, long n, char *pattern )
{
  char  bits[1000];
  char  c     = '\0';
  char  bit   = '\0';        
  char *start = pattern;

  long m       = 0;
  long i       = 0;
  long offset  = 0;
  long soffset = 0;

  int true  = 1;
  int false = 0;


  if ( !start )
    return true;
  
  if ( *start == '!' ) /* Negation */
  {
    true =  0;
    false = 1;
    ++start;
  }

  m = strlen( start );

  if ( n > 1000 )
    n = 1000;

  dmu_map_bits( buf, n, bits );

#if DEBUG
  printf( "BUF = %ld PATTERN = %s\n", n, pattern );
  printf( "BITS = " );
  for ( i = 0; i < n; i++ )
    printf("%d", bits[i] );
  printf( "\n" );
#endif

  for (i = 0; i < n; i++ )
  {
    offset  = n-1-i;
    bit     = bits[offset];
    soffset = m-1-i;

    if ( soffset < 0 )
    {
      if ( bit != '0' )
        return false;
    } 
    else 
    {
      c = start[m-1-i];

      switch( c )
      {
        case 'x': 
	  break;   
        case '0':  
        case '1':
	  if ( bit != c )
            return false;
	  break;
        default:
	  break;
      }
    }
  }
 return true;
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_bit_union                                   */
/*                                                             */
/* Description:                                                */
/*   Attempt to merge two bit fields that are OR'ed together   */
/*   ( the union ).                                            */
/*   If they can be merged, bufB is set to the merged value.   */
/*                                                             */
/*   Note: It is okay for bufA or bufB to be the exact same    */
/*         chunk of memory as obuf.                            */
/*                                                             */
/*   There are 3 possibilities for merge.                      */
/*     1. A == B, return B                                     */
/*     2. The bits set by both A and B are the same, but       */
/*        either A or B have less bits set.  Return the        */
/*        one with less bits set.                              */
/*     3. A and B differ by one bit.  Set that bit to 'x'      */
/*        and return that bit field.                           */
/*                                                             */
/*  Input parameters:                                          */
/*       const char  *bufA    - input string A                 */
/*       const char  *bufB    - input string B                 */
/*                                                             */
/*  Output parameters:                                         */
/*         char  *obuf - resulting string                      */
/*                                                             */
/*  Returns:                                                   */
/*         dmBool dmTRUE  if they could be merged.             */
/*                dmFALSE if they couldn't be merged.          */
/*                                                             */
/***************************************************************/
dmBool dmu_bit_union( const char* bufA, const char* bufB, char* obuf )
{
  long num_diff  = 0; /* number of non-'x' bit diffs between A and B */
  long last_diff = 0; /* position of the last diff */
  long numA      = 0; /* number of non 'x' bits in A */
  long numB      = 0; /* number of non 'x' bits in B */
  long pos       = 0; /* position in the bit strings */

  dmBool merged = dmFALSE;

  const char *Aptr = bufA;
  const char *Bptr = bufB;
  char       *optr = obuf;


  if ( !Aptr || !Bptr || !optr )
    return dmFALSE;
 
  if ( *Aptr == '!' && *Bptr == '!' )
  {
    /* use De Morgens */
    ++Aptr;
    ++Bptr;
    *optr = '!';
    ++optr;

    if ( dmu_bit_intersect( Aptr, Bptr, optr ))
    {
      if ( *optr == '!' ) 
      {
        /*  double negative makes positive.                        */
        /*  this should only happen if we get back the null filter */ 
        /*  in which case we set the allow everything filter.      */

        while( *optr != '\0' )
        { 
          ++optr;
          *(optr-2) = *optr;
        }
      }
      return dmTRUE;
    }
    else
     return dmFALSE;
  }
  else if ( *Aptr == '!' || *Bptr == '!' )
    return dmFALSE;
  else if ( *Aptr == '\0' || *Bptr == '\0' )
  { 
    *optr = '\0';
    return dmTRUE;
  }

  while( *Aptr != '\0' && *Bptr != '\0' )
  {
    if ( *Aptr == '1' || *Aptr == '0' )
      ++numA;

    if ( *Bptr == '1' || *Bptr == '0' )
     ++numB;

    if ( *Aptr != *Bptr )
    {
      ++num_diff;
      last_diff = pos;
    }

    if ( num_diff > 1 )
      return dmFALSE;

    ++Aptr;
    ++Bptr;
    ++pos;
  }
 
  /* don't try to merge if the two bit fields are of different length */
  if ( num_diff == 0 )
  {
    if ( numA == numB ) /* A == B */
    {
      if ( obuf != bufA && obuf != bufB )
        strcpy( obuf, bufA );
      merged = dmTRUE;
    }
    else if ( numB < numA ) /*  B overrules A, return B */
    {
      if ( bufB != obuf )
        strcpy( obuf, bufB );
      merged = dmTRUE;
    }
    else if ( numA < numB ) /* A overrules B */
    {
      if ( bufA != obuf )
        strcpy( obuf, bufA );
      merged = dmTRUE;
    }
  }
  else if ( num_diff == 1 )
  {
    /* one difference, set it to 'x' */
    if ( obuf == bufA || obuf == bufB )
      obuf[last_diff] = 'x';
    else
    {
      strcpy( obuf, bufA );
      obuf[last_diff] = 'x';
    }
    merged = dmTRUE;
  }
  return merged;
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_bit_intersect                               */
/*                                                             */
/* Description:                                                */
/*   Two bit fields that are AND'ed together ( intersection ). */
/*   Place result in B.                                        */
/*                                                             */
/*   For each bit,                                             */
/*     if both A and B are non 'x' and they differ,            */
/*       return false                                          */
/*     if A is non 'x' and B is 'x', set B to A.               */
/*                                                             */
/*   Note: It is okay for obuf to be the exact same chunk of   */
/*          memory as bufA or bufB.                            */
/*                                                             */
/*  Input parameters:                                          */
/*       const char  *bufA    - input string A                 */
/*       const char  *bufB    - input string B                 */
/*                                                             */
/*  Output parameters:                                         */
/*         char  *obuf - resulting string                      */
/*                                                             */
/*  Returns:                                                   */
/*         dmBool dmTRUE  if there is a valid intersection     */
/*                dmFALSE if they can't be intersected         */
/*                                                             */
/***************************************************************/
dmBool dmu_bit_intersect( const char* bufA, const char* bufB, char* obuf )
{
  const char *Aptr = bufA;
  const char *Bptr = bufB;
  char       *optr = obuf;

  dmBool null = dmFALSE;
 
  if ( !bufA || !bufB )
    return dmFALSE;
 
  /* use De Morgans */
  if ( *Aptr == '!' && *Bptr == '!' )
  {
    *optr = '!';
    ++optr;
    ++Aptr;
    ++Bptr;

    if ( dmu_bit_union( Aptr, Bptr, optr ))
      return dmTRUE;
    return dmFALSE;
  }
  else if ( *Aptr == '!' || *Bptr == '!' )
    return dmFALSE;

  while ( *Aptr != '\0' && *Bptr != '\0' )
  {
    if ( *Aptr == '1' || *Aptr == '0' )
    {
      if ( *Bptr == '1' || *Bptr == '0' )
      {
        if ( *Bptr != *Aptr )
        {
          null = dmTRUE;
          break;
        }
        else
          *optr = *Aptr;
      }
      else 
      {
        *optr = *Aptr;
      }
    }
    else
      *optr = *Bptr;

    ++Aptr;
    ++Bptr;
    ++optr;
  }

  /* set the null filter */
  if ( null )
  {
   optr  = obuf;
   *optr = '!';
   ++optr;
  }

  *optr = '\0';

  return dmTRUE;
}



/***************************************************************/
/*                                                             */
/* Function:   dmu_map_bits                                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*    unsigned char  *buf     - input string                   */
/*             long   n       - length                         */
/*                                                             */
/*  Output parameters:                                         */
/*          char  *bits       - output string                  */
/*                                                             */
/*  Returns:                                                   */
/*        void                                                 */
/*                                                             */
/***************************************************************/  
void dmu_map_bits( unsigned char* buf, long n, char* bits )
{
  long bit_no  = 0;   
  long offset  = 0;
  long byte_no = 0;

  unsigned char mask[BYTE_BITS]  = { 1, 2, 4, 8, 16, 32, 64, 128 };

  for ( bit_no = 1; bit_no <= n; bit_no++ )
  {
    offset = BYTE_BITS - 1 -  (bit_no-1) % BYTE_BITS;
    byte_no = (bit_no -1) / BYTE_BITS;
    bits[bit_no-1] = ( buf[byte_no] & mask[offset] ) ? '1' : '0';
  }

  bits[n] = 0;
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_print_pixel                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*          long  *pixel  - pixel number                       */
/*          long   naxes  - number of axis                     */
/*                                                             */
/*  Output parameters:                                         */
/*          char  *buf    - output sting                       */
/*                                                             */
/*  Returns:                                                   */
/*        void                                                 */
/*                                                             */
/***************************************************************/ 
void dmu_print_pixel( long* pixel, long naxes, char* buf )
{
  long  ii  = 0;
  char *ptr = buf;

  *ptr = '('; ptr++;

  for ( ii = 0; ii < naxes; ii++ )
  {
    ptr += sprintf( ptr, "%ld", pixel[ii] );

    if ( ii < naxes-1 ) 
    {
      *ptr = ',';
      ++ptr; 
    }
  }

  *ptr= ')';
  ++ptr;
  *ptr = '\0';
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_parse_null                                  */
/*                                                             */
/* Description:                                                */
/*   Parse string value and convert to the specified           */
/*   datatype, set converted value to provided pointer.        */
/*     val = "NAN"  => IEEE NaN for real types                 */
/*                  => error (DM_NULLTYPE_NONE) non-real types */
/*                                                             */
/*  Input parameters:                                          */
/*    const char *value   = Null value as string               */
/*    dmDatatType type    = output datatype for value          */
/*                                                             */
/*  Output parameters:                                         */
/*    void*       null    = ptr for converted value.           */
/*                                                             */
/*  Returns:                                                   */
/*    int         flag    = DM_NULLTYPE_NONE == Problem        */
/*                        = DM_NULLTYPE_NAN  == NaN input      */
/*                        = DM_NULLTYPE_VAL  == OK             */
/*                                                             */
/***************************************************************/
int dmu_parse_null( const char* val, dmDataType type, GenericType* null )
{

  dmBool is_nan = dmFALSE;
  dmBool is_flt = dmFALSE;
  int    flag   = DM_NULLTYPE_VAL;

  /* Zero out */
  *null = 0;

  /* Check for empty input */
  if ( !val || *val == '\0')
  {
    flag = DM_NULLTYPE_NONE;
  }
  else
  {
    is_nan = cs_uceq( val, "NAN");
    if ( strchr(val,'.') != NULL)
      is_flt = dmTRUE;

    switch( type )
    {
    case DT_UINT1: 
      if( is_nan || is_flt )
      {
	flag = DM_NULLTYPE_BAD;
      }
      else
      {
	*(unsigned char*)null = (unsigned char)atol(val);
      }
      break;
    case DT_UINT2: 
      if( is_nan || is_flt )
      {
	flag = DM_NULLTYPE_BAD;
      }
      else
      {
	*(unsigned short*)null = (unsigned short)atol(val);
      }
      break;
    case DT_UINT4: 
      if( is_nan || is_flt )
      {
	flag = DM_NULLTYPE_BAD;
      }
      else
      {
	*(unsigned long*)null = (unsigned long)atol(val);
      }
      break;
    case DT_INT2: 
      if( is_nan || is_flt )
      {
	flag = DM_NULLTYPE_BAD;
      }
      else
      {
	*(short*)null = (short)atol(val);
      }
      break;
    case DT_INT4: 
      if( is_nan || is_flt )
      {
	flag = DM_NULLTYPE_BAD;
      }
      else
      {
	*(long*)null = (long)atol(val);
      }
      break;
    case DT_REAL4: 
      if( is_nan )
      {
	*(float*)null = ut_nan_r();
	flag = DM_NULLTYPE_NAN;
      }
      else
      {
	*(float*)null = (float)atof(val);
      }
      break;
    case DT_REAL8: 
      if( is_nan )
      {
	*(double*)null = ut_nan_d();
	flag = DM_NULLTYPE_NAN;
      }
      else
      {
	*(double*)null = (double)atof(val);
      }
      break;
    default:
      break;

    } /* end switch */
  } /* end else */

 return flag;
}
