/*                                                                
**  Copyright (C) 2002-2003,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   keylib.c                                                    */
/*                                                               */
/*   Functions that create and work with key list structures     */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmulib.h"

/* The key list structure allows us to maintain                     */
/*  a list of key-value(-unit-desc) sets. It is intended to support */
/*  a typical header key implementation, of which DM-enhanced FITS  */
/*  is the most complex I have come across to date.                 */
/*  We use it for the internal cache of the `native' or `raw' keys, */
/*  before mapping them to the dmDescriptor keys.                   */



/***************************************************************/
/*                                                             */
/* Function:   dmu_key_free                                    */
/*                                                             */
/* Description:                                                */
/*              Free key storage                               */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmuKey  *ftkey - key pointer             */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmu_key_free( dmuKey* ftkey )
{
  char *ptr = NULL;

  if ( ftkey != NULL)
  {
    if ( ftkey->name )
      free( ftkey->name );

    if ( ftkey->ucname )
      free( ftkey->ucname );

    if ( ftkey->desc )
      free( ftkey->desc );

    if ( ftkey->unit )
      free( ftkey->unit );

    if ( ftkey->datatype == DT_CHAR )
    {
       ptr = dmu_key_value_c( ftkey );

       if  (ptr )
         free( ptr );
    }

    free( ftkey );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_key_alloc                                   */
/*                                                             */
/* Description:                                                */
/*              Allocate key storage                           */
/*                                                             */
/*  Input parameters:                                          */
/*                    none                                     */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            dmuKey  *key                                     */
/*                                                             */
/***************************************************************/
dmuKey* dmu_key_alloc( void )
{
  dmuKey  *ftkey = calloc( 1, sizeof( dmuKey ));

  return ftkey;
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_keylist_n                                   */
/*                                                             */
/* Description:                                                */
/*              Returns keylist element n.  If keylist pointer */
/*              is NULL, returns 0                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    const dmuKeyList keylist                 */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
long dmu_keylist_n( const dmuKeyList keylist )
{
  if ( !keylist )
    return 0;

  return keylist->n;
}



/***************************************************************/
/*                                                             */
/* Function:   dmu_key_match                                   */
/*                                                             */
/* Description:                                                */
/*              Match two names.  The pattern blah* will match */
/*              blah or blah4 or blah44 but not blahg          */
/*              Trailing * in pattern matches trailing digits  */
/*              or trailing *, or nothing.                     */
/*                                                             */
/*  Input parameters:                                          */
/*                    const char *name                         */
/*                    const char *pattern                      */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            logical  result                                  */
/*                                                             */
/***************************************************************/
logical dmu_key_match( const char* name, const char* pattern )
{ 
  long m = 0;
  long n = 0;
 
  n = strlen( pattern );
  m = strlen( name );
  
  if ( pattern[n-1] == '*' && name[m-1] != '*' && isdigit( name[m-1] ) )
  {
    --n;

    while( isdigit( name[m-1] ) && m > 1 )
      --m;

    if ( m != n )
      return UT_FALSE;

    return cs_uceq_siz( name, pattern, n );
  }
  else
    return cs_uceq( name, pattern );

}  



/***************************************************************/
/*                                                             */
/* Function:   dmu_key_find                                    */
/*                                                             */
/* Description:                                                */
/*              Find the first key with this name in the list. */
/*              Caution: recall COMMENT and HISTORY may be     */
/*                       repeated, and although it's very bad  */
/*                       FITS style, it does happen, but if so,*/
/*                       tough, we won't handle it.  Other     */
/*                       repeated keys get the same treatment. */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmuKeylist  keylist - keylist            */
/*                    char       *name    - key name           */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            dmuKey  *ftkey                                   */
/*                                                             */
/***************************************************************/
dmuKey* dmu_key_find( dmuKeyList keylist, char* name )
{
  long    keyNo  = 0;
  long    offset = 0;
  dmuKey *ftkey  = NULL;
  char   *ucname = NULL;

  if ( keylist && keylist->data )
  {
    /* convert name into upper case string */
    ucname = cs_dup( name );
    cs_upper( ucname );

    for ( keyNo = 1; keyNo <= keylist->n; keyNo++ )
    {
      offset = keyNo - 1;
      ftkey = keylist->data[offset];

      if ( cs_eq( ftkey->ucname, ucname ))
      {
        free ( ucname );
        return ftkey;
      }
    }
  }

  free ( ucname );
  return NULL;
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_key_set                                     */
/*                                                             */
/* Description:                                                */
/*              Set key                                        */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmuKey           *ftkey - key pointer    */
/*                    const dmuKeyType  fktype                 */
/*                    const dmDataType  datatype               */
/*                    const char       *name                   */
/*                    const void       *value                  */
/*                    const char       *unit                   */
/*                    const char       *desc                   */
/*                    cpmst dmBool      mask                   */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmu_key_set( dmuKey* ftkey, const dmuKeyType fktype,
                  const dmDataType datatype, const char* name,
                  const void* value, const char* unit, const char* desc,
                  const dmBool mask )
{
  char *ptr = NULL;

  if ( !ftkey )
    return;

  ftkey->keytype = fktype;

  if ( ftkey->datatype == DT_CHAR )
  {
    ptr = dmu_key_value_c( ftkey );

    if ( ptr )
      free ( ptr );
  }

  ftkey->datatype = datatype;
 
  if ( ftkey->name )
    free( ftkey->name );

  if ( ftkey->ucname )
    free( ftkey->ucname );

  if ( name )
  {
    ftkey->name = cs_dup( name );

    /* make a copy of name in upper case */
    ftkey->ucname = cs_dup( name );
    cs_upper( ftkey->ucname);
  }

  if ( unit )
  {
    if ( ftkey->unit )
      free( ftkey->unit );
    ftkey->unit = cs_dup( unit );
  }

  if ( desc )
  {
    if ( ftkey->desc )
      free( ftkey->desc );
    ftkey->desc = cs_dup( desc );
  }

  ftkey->mask = mask;

  ar_mix_copy_value( value, datatype, &ftkey->value, UT_TRUE, 0 );

  if ( ftkey->datatype == DT_CHAR && ftkey->value )
  {
    long i;

    ptr = dmu_key_value_c( ftkey );
    i = strlen(ptr) - 1;  

    /* remove trailing blanks and set string to empty if it is all blanks */
    while ( i >= 0 && ptr[i] == ' ' ) 
    {
      ptr[i] = '\0';
      --i;
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_key_insert                                  */
/*                                                             */
/* Description:                                                */
/*              Maintain list positions                        */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmuKeyList  keylist                      */
/*                    dmuKey     *ftkey                        */
/*                    long        posp                         */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            long  keyno                                      */
/*                                                             */
/***************************************************************/
long dmu_key_insert( dmuKeyList keylist, dmuKey* ftkey, long posp )
{
  long    keyno = 0;
  dmuKey *ftkeyp = NULL;
  long    pos = posp;


  if (keylist)
  {
    keylist->n++;

    if ( pos <= 0 ) 
      pos = keylist->n;

    if ( keylist->n > keylist->nmax )
    {
      keylist->nmax = 2 * keylist->n + 20;
      keylist->data = realloc( keylist->data, keylist->nmax * UT_SZ_P );
    }

    /* Move keys up */

    if ( keylist->data )
    {
      for ( keyno = keylist->n-1;  keyno >= pos; keyno-- )
      {
        ftkeyp = keylist->data[keyno-1];
        ftkeyp->index = keyno+1;
        keylist->data[keyno] = ftkeyp;
      }

      ftkey->index = pos;

      keylist->data[pos-1] = ftkey;
    }
  }

  return pos;
}



/***************************************************************/
/*                                                             */
/* Function:   dmu_key_delete                                  */
/*                                                             */
/* Description:                                                */
/*              Delete key from list                           */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmuKeyList  keylist                      */
/*                    dmuKey     *ftkey                        */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            dmBool  result                                   */
/*                                                             */
/***************************************************************/
dmBool dmu_key_delete(  dmuKeyList keylist, dmuKey* ftkey )
{
  dmuKey *ftkeyp = NULL;
  long    keyno  = 0;
  long    deleted_keyno = 0;


  if ( !keylist || !keylist->data || !ftkey )
    return UT_FALSE;

  deleted_keyno = ftkey->index;

  if ( deleted_keyno == 0 )
    deleted_keyno = 1;  /* To catch corrupt data */


  ftkeyp = keylist->data[deleted_keyno-1];

  if ( ftkeyp != ftkey )
  {
    fprintf( stderr, "INDEX SYNC ERROR IN FTKERNEL KEY %s\n", ftkey->name );
    return UT_FALSE;
  }

  dmu_key_free( ftkey );

  for ( keyno = deleted_keyno+1; keyno <= keylist->n; keyno++ )
  {
    ftkeyp = keylist->data[keyno-1];
    ftkeyp->index = keyno-1;
    keylist->data[keyno-2] = ftkeyp;  
  }

  keylist->n--;

  return UT_TRUE;
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_keylist_alloc                               */
/*                                                             */
/* Description:                                                */
/*              Allocate a keylist                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    none                                     */
/*                                                             */
/*  Output parameters:                                         */
/*                   none                                      */
/*                                                             */
/*  Returns:                                                   */
/*            dmuKeyList  keylist                              */
/*                                                             */
/***************************************************************/
dmuKeyList dmu_keylist_alloc( void )
{
  dmuKeyList keylist =  calloc( 1, sizeof( struct dmuKeyList_s ));

  if ( keylist)
  {
    keylist->n = 0;
    keylist->nmax = 0;
  }

  return keylist;
}



/***************************************************************/
/*                                                             */
/* Function:   dmu_keylist_free                                */
/*                                                             */
/* Description:                                                */
/*              Free key list storage                          */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmuKeyList  keylist                      */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmu_keylist_free( dmuKeyList keylist )
{
  long offset = 0;

  if ( keylist )
  {
    if (keylist->data)
    {
      for ( offset = 0; offset < keylist->n; offset++ )
        dmu_key_free( keylist->data[offset] );

      free( keylist->data );
    }

    free( keylist );
  }
}



/***************************************************************/
/*                                                             */
/* Function:   dmu_keyword_max_index                           */
/*                                                             */
/* Description:                                                */
/*              For indexed keyword FOOn, what is the maximum  */
/*              index?  We us this to find the next index to   */
/*              use, e.g. for MTYPEn                           */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmuKeyList  keylist                      */
/*                    char       *pattern                      */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            long  max_index                                  */
/*                                                             */
/***************************************************************/
long dmu_keyword_max_index( dmuKeyList keylist, char* pattern )
{
  long     keyNo     = 0;
  long     offset    = 0;
  long     n         = 0;
  long     index     = 0;
  long     max_index = 0;
  char    *key       = NULL;
  dmuKey  *ftkey     = NULL;


  if ( keylist && keylist->data)
  {
    n = cs_ends( pattern );

    for ( keyNo = 1; keyNo <= keylist->n; keyNo++ )
    {
      offset = keyNo - 1;
      ftkey  = keylist->data[offset];

      if (ftkey)
        key = ftkey->name;

      if ( key)
      {
        if ( cs_uceq_siz( key, pattern, n ) && isdigit( key[n] ) )
        {
          index = atol( key+n );

          if ( index > max_index )
            max_index = index;
        }
      }
    }  // end for loop
  }

  return max_index;
}



/***************************************************************/
/*                                                             */
/* Function:   dmu_keylist_read_value_c                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmuKeyList  keylist                      */
/*                    char       *name                         */
/*                    char       *value                        */
/*                    long        maxlen                       */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            dmuKey *ftkey                                    */
/*                                                             */
/***************************************************************/
dmuKey* dmu_keylist_read_value_c( dmuKeyList keylist, char* name,
                                  char* value, long maxlen )
{
  char   *ptr = NULL;
  dmuKey *ftkey = NULL;


  ftkey = dmu_keylist_read_typed_value( keylist, name, DT_CHAR, (void*)&ptr );

  if ( ftkey )
    cs_copy_siz( ptr, value, maxlen );
  else
    cs_copy( " ", value );

  return ftkey;
}



/***************************************************************/
/*                                                             */
/* Function:   dmu_keylist_read_typed_value                    */
/*                                                             */
/* Description:                                                */
/*              Returns pointer to stored string, not copy of  */
/*              string                                         */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmuKeyList  keylist                      */
/*                    char       *name                         */
/*                    dmDataType  type                         */
/*                    void       *value                        */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            dmuKey *ftkey                                    */
/*                                                             */
/***************************************************************/
dmuKey* dmu_keylist_read_typed_value( dmuKeyList keylist, char* name,
                                      dmDataType type, void* value )
{
  dmuKey *ftkey = NULL;

  if ( keylist )
    ftkey = dmu_key_find( keylist, name );

  if ( !ftkey )
    return NULL;

  dmu_key_typed_value( ftkey, type, value );

  return ftkey;
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_keylist_read_value_c                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmuKey     *ftkey                        */
/*                    dmDataType  type                         */
/*                                                             */
/*  Output parameters:                                         */
/*                    void       *value                        */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmu_key_typed_value( dmuKey* ftkey, dmDataType type, void* value )
{
  long  size = 0;
  void *ptr  = NULL;

  if ( !ftkey )
    return;

  if ( ftkey->datatype == DT_CHAR )
  {
    if ( type != DT_CHAR )
     *(char**)value = NULL;
    else
     *(char**)value = dmu_key_value_c( ftkey );
  }
  else
    if ( ftkey->datatype != type )
    {
      ptr = dmu_key_value( ftkey );
      ar_mix_numeric_cast( ptr, ftkey->datatype, type, value );
    }
    else 
    {
      size = ar_mix_type_size( type );
      ptr = dmu_key_value( ftkey );
      ar_mix_copy_value( ptr, type, value, UT_FALSE, size );
    }

}




/***************************************************************/
/*                                                             */
/* Function:   dmu_keylist_check_index                         */
/*                                                             */
/* Description:                                                */
/*              Look for cards of the form indexName = 'name'  */
/*              Return its index                               */
/*              Example:                                       */
/*           n = ftpKeyListCheckIndex( keylist, "DTYPE", "CCD")*/
/*              returns 4 if there is a card                   */
/*           DTYPE4 = 'CCD'                                    */
/*              and 0 if ther is no n for which DTYPEn is CCD. */
/*              If name is null, return the largest index      */
/*              associated with indexName                      */
/*              (this is a slightly ugly way of putting two    */
/*              functions in one, I'm afraid.                  */
/*                                                             */
/*  Input parameters:                                          */
/*                    const dmuKeyList keylist                 */
/*                    const char      *index_name              */
/*                    const char      *name                    */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            long n                                           */
/*                                                             */
/***************************************************************/
long dmu_keylist_check_index( const dmuKeyList keylist,
                              const char* index_name, const char* name )
{
  dmuKey *ftkey     = NULL;
  char   *key_name  = NULL;
  char   *ptr       = NULL;
  long    keyNo     = 0;
  long    size      = 0;
  long    key_index = 0;
  long    n         = 0;
  long    offset    = 0;
 
  if ( keylist  && keylist->data )
  {
    size = strlen( index_name );

    if ( index_name[size-1] == '#' )
      --size;   /* Omit trailing wild card */
     
    /* Loop over all kernel keys looking for DTYPEn (or DSTYPn, etc) */
    /* Only do string type keys                                      */

    for ( keyNo = 1; keyNo <= keylist->n; keyNo++ )
    {
      key_name = NULL;
      offset = keyNo -1;
      ftkey  = keylist->data[offset];

      if ( ftkey )
        key_name = ftkey->name;

      if ( key_name && cs_eq_siz( key_name, index_name, size ) &&
           ftkey->datatype == DT_CHAR )
      {  
        /* NULL name, accumulate largest DTYPEn */
        if ( !name )
        {
          key_index = atol( key_name+size );
          n = ( key_index > n ) ? key_index : n;
        }
        else
        {
          /* See if value of DTYPEn is equal to target name, if so return n */
          ptr = dmu_key_value_c( ftkey );

          if ( dmu_key_match( ptr, name ) )
          {     
             key_index = atol( key_name+size ); 
             return key_index;
          }
        }
      }
    }  // end for loop
  }

  return n;

} 




/***************************************************************/
/*                                                             */
/* Function:   dmu_keylist_print                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                  const dmuKeyList  keylist                  */
/*          void (*NativeTypeName) (dmuKeyType, char *, long ) */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmu_keylist_print( const dmuKeyList keylist, 
                       void (*NativeTypeName)( dmuKeyType, char*, long ) )
{
  long    maxlen = 1024; 
  char    buf[1024];
  long    offset = 0;
  long    keyno  = 0;
  dmuKey *ftkey  = 0;
  int     opts   = UT_TRUE;  /* Print fancy preamble for each key */


  if ( keylist )
  {
    printf( "KEY LIST: %ld keys (allocated space for %ld)\n", keylist->n, keylist->nmax );
    printf( "\n" );
    printf( "--------------------\n" );

    if (keylist->data)
    {
      for ( offset = 0; offset < keylist->n; offset++ )
      {
        keyno = offset + 1;
        ftkey = keylist->data[offset];

        if ( ftkey )
	{
          dmu_key_print( ftkey, NativeTypeName, opts, buf, maxlen );
          printf( "%s\n", buf );
        }
      }

      printf( "--------------------\n" );
    }
  }
}



/***************************************************************/
/*                                                             */
/* Function:   dmu_key_print                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmuKey     *ftkey                        */
/*          void (*NativeTYpeName)( dmuKeyType, char *, long ) */
/*                    int         opts                         */
/*                    long        maxlen                       */
/*                                                             */
/*  Output parameters:                                         */
/*                    char       *buf                          */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmu_key_print( dmuKey* ftkey, void (*NativeTypeName)( dmuKeyType, char*, long ),
                    int opts, char* buf, long maxlen )
{
  char   tag[10];
  char   tmpBuf[CARD_SIZE];
  char   tbuf[1024];
  char   pbuf1[256];
  char   pbuf2[256];
  char   xunit[80];
 
  char  *null = " ";
  char  *desc = NULL;
  char  *unit = NULL;
  char  *name = NULL;
  char   ct[4] = " ";
  char   c;  /* mask value */

  long   index  = 0;
  long   offset = 0;
  long   keyno  = 0;


  if ( !ftkey )
    return;

  keyno = ftkey->index;
  offset = keyno - 1;

  desc = ftkey->desc;

  if ( !desc || !(*desc) )
    desc = null;

  unit = ftkey->unit;

  if ( !unit || !(*unit ))
    unit = null;

  if ( !cs_is_blank( unit ))
    sprintf( xunit, "[%s]", unit );
  else
    cs_copy( " ", xunit );

  name =   ftkey->name;

  ar_mix_print_generic( &ftkey->value, ftkey->datatype, tmpBuf, CARD_SIZE );

  if ( dm_global & DM_GLOBAL_OLD )
  {
    if ( !cs_is_blank( xunit ))
      sprintf( xunit, "[%s] %s", unit, desc );
    else 
      cs_copy( desc, xunit );

    sprintf( pbuf2, "%-12s = %-20s / %-20s", name, tmpBuf, xunit );
  }
  else
    sprintf( pbuf2, "%-12s = %-20s %s / %-20s", name,tmpBuf, xunit, desc );


  if( opts )
  {
   if ( NativeTypeName )
     NativeTypeName( ftkey->keytype, ct, 4 );

    c = ftkey->mask ? '*' : ' ';
    index =  ftkey->group;

    if ( index )
      sprintf( tag, "DTYPE%ld", index );
    else
      strcpy( tag, " ");   
    sprintf( pbuf1, "Key %4ld: %-4s %-8s %c", keyno, ct, tag, c );
    sprintf( tbuf, "%s%s", pbuf1, pbuf2 );
    cs_copy_siz( tbuf, buf, maxlen );
   }
   else
   {
     cs_copy_siz( pbuf2, buf, maxlen );
   }
}



/***************************************************************/
/*                                                             */
/* Function:   dmu_key_value                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmuKey     *ftkey                        */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            void *value                                      */
/*                                                             */
/***************************************************************/
void* dmu_key_value( dmuKey* ftkey )
{
  if ( ftkey )
    return( (void*)&ftkey->value );

  return NULL;
}



/***************************************************************/
/*                                                             */
/* Function:   dmu_key_value_c                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmuKey     *ftkey                        */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            char *value                                      */
/*                                                             */
/***************************************************************/
char* dmu_key_value_c( dmuKey* ftkey )
{
  if ( ftkey && ftkey->datatype == DT_CHAR )
    return( *(char**)&ftkey->value );

  return NULL;
}

/***************************************************************/
/*                                                             */
/* Function:   dmu_key_read_value_c                            */
/*                                                             */
/* Description:                                                */
/*   Function to read DT_CHAR value from key found in          */
/*   column keys keylist                                       */
/*   Take this supplied index and look up the key in the list. */
/*   Return key as return value and key value in output param. */
/*                                                             */
/*   Code Change  Bug 12317  May 27, 2010                      */
/*   Added as part of aspect performance enhancement           */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmuKey  *pKey                            */
/*                    long     maxLen                          */
/*                                                             */
/*  Output parameters:                                         */
/*                    char *value                              */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmu_key_read_value_c( dmuKey *pKey, long maxLen, char *value  )
{
  char *sValue = NULL;

  sValue = dmu_key_value_c(pKey);

  if (sValue)
    cs_copy_siz(sValue, value, maxLen );
  else
    cs_copy(" ", value);

  return;
}


/***************************************************************/
/*                                                             */
/* Function:   dx                                              */
/*                                                             */
/* Description:                                                */
/*                Unused function                              */
/*  Input parameters:                                          */
/*                    dmuKeyList  keylist                      */
/*                                                             */
/*  Output parameters:                                         */
/*                     none                                    */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
#if 0
void dx( dmuKeyList keylist )
{
  dmuKey *ftkey = NULL;
  long    index = -1;

  ftkey = dmu_key_find( keylist, "HDUCLAS2" );

  if ( ftkey )
  {
    index = ftkey->index;
   printf( "FTKEY %p HDUCLAS2 INDEX = %ld\n", ftkey, index );
  }
}
#endif
