/*                                                                
**  Copyright (C) 2002-2004,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   vkey.c                                                      */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmulib.h"
#include "dmnutils.h"


/***************************************************************/
/*                                                             */
/* Function:   dmu_vector_key_init                             */
/*                                                             */
/* Description:                                                */
/*     Initialize dmuVectorKey struct                          */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmuVectorKey  v                            */
/*                                                             */
/*  Output parameters:                                         */
/*                  none                                       */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmu_vector_key_init( dmuVectorKey v )
{
  v->name = NULL;
  v->val = NULL;
  v->form= NULL;
  v->unit = NULL;
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_vector_key_make_list                        */
/*                                                             */
/* Description:                                                */
/*     This routine supports the TTYPEn/TFORMn/TUNITn style    */
/*     convention                                              */
/*                                                             */
/*     tagsize = 5 usually                                     */
/*                                                             */
/*  Input parameters:                                          */
/*            dmuKeyList    keylist                            */
/*            char         *name_tag                           */
/*            char         *form_tag                           */
/*            char         *unit_tag                           */
/*            char         *val_tag                            */
/*            long          tagsize                            */
/*                                                             */
/*  Output parameters:                                         */
/*                  none                                       */
/*                                                             */
/*  Returns:                                                   */
/*            MixArray  list                                   */
/*                                                             */
/***************************************************************/
MixArray dmu_vector_key_make_list( dmuKeyList keylist, char* name_tag, 
                                   char* form_tag, char* unit_tag, 
                                   char* val_tag, long tagsize )
{
  MixArray  list = NULL;
  TextCard  key;
  dmuKey   *ftkey = NULL;
  dmuKey   *cpt_key = NULL;
  dmuVectorKey entry;

  char *keyname = NULL;
  char *name    = NULL;
  long  n       = 0;
  long  size    = 0;
  long  offset  = 0;
  long  keyno   = 0;
  long  index   = 0;

  /* How many XTYPE keys? */
  n = dmu_keylist_check_index( keylist, name_tag, NULL );

  if ( n == 0 )
    return NULL;

  size = sizeof( struct dmuVectorKey_s );
  list = ar_mix_alloc_type( DT_PTR, size, n );

  if ( dmverbose )
    fio_tmsg( "Make list of vector keys" );

  for ( keyno = 1; keyno <= keylist->n; keyno++ )
  {
    offset = keyno-1;
    ftkey = keylist->data[offset];
    keyname = ftkey->name;

    /* Check for XTYPE and that it is a string key */
    if ( cs_eq_siz( keyname, name_tag, tagsize ) && ftkey->datatype == DT_CHAR )
    {
      index = atol( keyname + tagsize );
      list->n++;
      entry = ar_mix_ptr( list, index-1 );
      dmu_vector_key_init( entry );
      entry->name = ftkey;
      name = dmu_key_value_c( ftkey );

      /* Check for XFORMn */
      sprintf( key, "%s%ld", form_tag, index );
      cpt_key = dmu_key_find( keylist, key );

      if ( cpt_key && cpt_key->datatype == DT_CHAR )
        entry->form = cpt_key;

      /* Check for XUNITn */
      sprintf( key, "%s%ld", unit_tag, index );
      cpt_key = dmu_key_find( keylist, key );

      if ( cpt_key && cpt_key->datatype == DT_CHAR )
        entry->unit = cpt_key;

      /* Check for XVALn */
      if ( val_tag )
      {
        sprintf( key, "%s%ld", val_tag, index );
        cpt_key = dmu_key_find( keylist, key );

        /* Support  XTYPEn = 'FOO',  FOO = 'VAL'  convention */
        if ( !cpt_key )
          cpt_key = dmu_key_find( keylist, name );

        if ( cpt_key )
        {
          entry->val = cpt_key;
          cpt_key->group = index;
        } 
      }
    }
  }

  return list;
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_alias_key_make_list                         */
/*                                                             */
/* Description:                                                */
/*     This routine supports the TTYPEn/TFORMn/TUNITn style    */
/*     convention                                              */
/*                                                             */
/*     tagsize = 5 usually                                     */
/*                                                             */
/*  Input parameters:                                          */
/*            dmuKeyList    keylist                            */
/*            char         *name_tag                           */
/*            char         *form_tag                           */
/*            char         *etype_tag                          */
/*            char         *ename_tag                          */
/*            long          tagsize                            */
/*                                                             */
/*  Output parameters:                                         */
/*                  none                                       */
/*                                                             */
/*  Returns:                                                   */
/*            dmuAliasKeyList  list                            */
/*                                                             */
/***************************************************************/
dmuAliasKeyList dmu_alias_key_make_list( dmuKeyList keylist, char* name_tag, 
            char* form_tag, char* etype_tag, char* ename_tag, long tagsize )
{
  AliasList array;
  AliasList cpts;
  /* How many XTYPE keys? */
  MixArray etype_array;
  TextCard key;

  dmuAliasKeyList list;
  dmElementType *etypes  = NULL;
  dmuKey        *ftkey   = NULL;
  dmuKey        *cpt_key = NULL;

  char **cptnames = NULL;
  char **kindex   = NULL;
  char  *name     = NULL;
  char  *cptlist  = NULL;
  char  *keyname  = NULL;
  char  *ename    = NULL;

  long n         = 0;
  long esize     = 0;
  long keyno     = 0;
  long entry     = 0;
  long cpt_entry = 0;
  long dim       = 0;
  long offset    = 0;
  long index     = 0;
  long edim      = 0;


  n = dmu_keylist_check_index( keylist, name_tag, NULL );

  if ( n == 0 )
    return NULL;

  kindex = calloc( n, UT_SZ_P );
  array = ar_alias_alloc( n );
  cpts  = ar_alias_alloc( n );
  esize = sizeof( dmElementType );
  etype_array = ar_mix_alloc_type( DT_PTR, esize, n );
  etypes = (dmElementType*)etype_array->data;

  list = calloc( 1, sizeof( struct dmuAliasKeyList_s ));
  list->names = array;
  list->cpts = cpts;
  list->etypes = etype_array;
  list->desc = kindex;
  list->n = n;

  for ( keyno = 1; keyno <= keylist->n; keyno++ )
  {
    offset = keyno - 1;
    ftkey = keylist->data[offset];
    keyname = ftkey->name;

    if ( cs_eq_siz( keyname, name_tag, tagsize ) && ftkey->datatype == DT_CHAR )
    {
      /* We found an MTYPE key */
      index = atol( keyname + tagsize );    
      sprintf( key, "%s%ld", form_tag, index );
      cpt_key = dmu_key_find( keylist, key );

      if ( cpt_key   && cpt_key->datatype == DT_CHAR )
      {
        name = dmu_key_value_c( ftkey );
        cptlist = dmu_key_value_c( cpt_key );     
        entry = ar_alias_set_name( array, name );     
        cpt_entry = ar_alias_set_name( cpts, name );
        cptnames = cs_parse_comma_list( cptlist, &dim );
        ar_alias_set_cpts( array, entry, cptnames, dim );
        kindex[entry-1] = cs_dup( ftkey->desc );
      }
      else
      {
        sprintf( dm_errbuf, "Missing %s keyword", key );
        ut_error_set( dm_error, dm_errbuf, dm_errmask | DM_ERR_NO_MFORM, 0 );     
        continue;
      }

      /* METYP */
      sprintf( key, "%s%ld", etype_tag, index );
      cpt_key = dmu_key_find( keylist, key );

      /* SUPPRESS SUPPORT FOR METYP KEYS! */
      cpt_key = NULL;

      if ( cpt_key  && cpt_key->datatype == DT_CHAR )
      {
        ename = dmu_key_value_c( cpt_key );
        etypes[entry-1] = dmu_parse_etype_code( ename );
      }
      else
      {
        etypes[entry-1] = dmVALUE;
      }
   
      esize = dmu_etype_dim( etypes[entry-1] );
      dim = array->dim[entry-1];

      if ( esize == 0 )
        esize = 1;

      edim = dim / esize;

      if ( dim % esize != 0 )
      {
        sprintf( dm_errbuf, "%s: %s element dim is %ld but MFORM= %s is not a multiple",
                 name, ename, esize, cptlist );
        ut_error_set( dm_error, dm_errbuf, dm_errmask | DM_ERR_BAD_METYP_DIM, 0 );
      }
    
      /* MENAM */
      sprintf( key, "%s%ld", ename_tag, index );
      cpt_key= dmu_key_find( keylist, key );

      if ( cpt_key  && cpt_key->datatype == DT_CHAR )
      {
        cptlist = dmu_key_value_c( cpt_key );
        cptnames = cs_parse_paren_list( cptlist, &dim );
        etypes[entry-1] = dmu_parse_etype_code( ename );
      }
      else
      {
        /* Otherwise, cptnames are same as above: must make NEW COPIES */
        cptnames = cs_parse_comma_list( cptlist, &dim );
      }

      if ( edim == dim )
      {
        ar_alias_set_cpts( cpts, cpt_entry, cptnames, edim );
      }
      else
      {
        sprintf( dm_errbuf, "%s has wrong number of cpts %ld not %ld", key, dim, edim );
        ut_error_set( dm_error, dm_errbuf, dm_errmask | DM_ERR_BAD_METYP_DIM, 0 );
      }      
    }
  }
  return list;
}


/***************************************************************/
/*                                                             */
/* Function:   dmu_alias_key_list_free                         */
/*                                                             */
/* Description:                                                */
/*     Release memory used for alias key list                  */
/*                                                             */
/*  Input parameters:                                          */
/*            dmuAliasKeyList  list                            */
/*                                                             */
/*  Output parameters:                                         */
/*                  none                                       */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmu_alias_key_list_free( dmuAliasKeyList list )
{
  long ii = 0;

  if ( !list )
    return;

  for ( ii =0; ii < list->n; ii++ )
    free( list->desc[ii] );

  free( list->desc );
  ar_alias_free( list->names );
  ar_alias_free( list->cpts );
  ar_mix_free( list->etypes );
  free( list );
}


