/*                                                                
**  Copyright (C) 1998-2001,2003,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmdss.c                                                     */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmlibP.h"


/********************************************************************/
/*                                                                  */
/* Function:   dmSubspaceColOpen                                    */
/*                                                                  */
/* Description:                                                     */
/*      Open subspace column                                        */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmBlock      *block                                  */
/*             char         *name                                   */
/*                                                                  */
/*  Output parameters:                                              */
/*              none                                                */
/*                                                                  */
/*  Returns:                                                        */
/*           dmDescriptor *data_descriptor                          */
/*                                                                  */
/********************************************************************/
dmDescriptor*  dmSubspaceColOpen(dmBlock* block, char* name)
{ 
  if ( block )
    return dmi_subspace_col_open( block->subspace, name );
  else
    return NULL;
}


/********************************************************************/
/*                                                                  */
/* Function:   dmBlockGetNoSubspaceCols                             */
/*                                                                  */
/* Description:                                                     */
/*      Return number of subspace columns                           */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmBlock      *block                                  */
/*                                                                  */
/*  Output parameters:                                              */
/*              none                                                */
/*                                                                  */
/*  Returns:                                                        */
/*           long  number_columns                                   */
/*                                                                  */
/********************************************************************/
long  dmBlockGetNoSubspaceCols(dmBlock* block)
{ 
  if ( block )
    return dmi_subspace_get_ncols( block->subspace );
  return 0;
}


/********************************************************************/
/*                                                                  */
/* Function:   dmBlockGetNoSubspaceCpts                             */
/*                                                                  */
/* Description:                                                     */
/*      Return number of subspace components                        */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmBlock      *block                                  */
/*                                                                  */
/*  Output parameters:                                              */
/*              none                                                */
/*                                                                  */
/*  Returns:                                                        */
/*           long  number_components                                */
/*                                                                  */
/********************************************************************/
long dmBlockGetNoSubspaceCpts(dmBlock *block)
{ 
  dmSubspace *subspace = NULL;

  if ( block && block->subspace )
  {
    subspace = block->subspace;
    return subspace->ncpts;
  }

  /* Back compat */
  if ( (dm_global & DM_GLOBAL_OLD ) && block && !block->subspace )
    return 1;

  return 0;
}


/********************************************************************/
/*                                                                  */
/* Function:   dmBlockSetSubspaceCpt                                */
/*                                                                  */
/* Description:                                                     */
/*      Set the currnt subspace component for the block.  Must be a */
/*      component which already exists (0 < cptNo <= maxCpt ).      */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmBlock      *block - block pointer                  */
/*             long          cptNo - component number to get        */
/*                                                                  */
/*  Output parameters:                                              */
/*              none                                                */
/*                                                                  */
/*  Returns:                                                        */
/*           dmErrCode  dmSUCCESS or dmFAILURE                      */
/*                                                                  */
/********************************************************************/
dmErrCode dmBlockSetSubspaceCpt(dmBlock *block, long cptNo)
{
  dmSubspace *subspace = NULL;

  if ( block && block->subspace )
  {
    subspace = block->subspace;
    if ( cptNo >= 1 && cptNo <= subspace->ncpts )
    {
      subspace->cpt = cptNo;
      return dmSUCCESS;
    }
  }
  return dmFAILURE;
}


/********************************************************************/
/*                                                                  */
/* Function:   dmBlockGetCurrSubspaceCpt                            */
/*                                                                  */
/* Description:                                                     */
/*     Get the current subspace component for the block.  If no     */
/*     subspace exists, we return 1 since this is where data gets   */
/*     written on a dmSubspaceColCreate routine when no subspace has*/
/*     been initialized.                                            */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmBlock      *block - block pointer                  */
/*                                                                  */
/*  Output parameters:                                              */
/*              none                                                */
/*                                                                  */
/*  Returns:                                                        */
/*           long  component_number                                 */
/*                                                                  */
/********************************************************************/
long dmBlockGetCurrSubspaceCpt(dmBlock *block)
{
  if(!block)
  {
      dmu_err_set(DM_ERR_NULL, "No block");
      return (long)0;
  }
  return block->subspace->cpt;
}


/********************************************************************/
/*                                                                  */
/* Function:   dmBlockCreateSubspaceCpt                             */
/*                                                                  */
/* Description:                                                     */
/*     Make a new component (virtual row in the DSS table) for the  */
/*     block.  Set the current component to be this new component.  */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmBlock      *block - block pointer                  */
/*                                                                  */
/*  Output parameters:                                              */
/*              none                                                */
/*                                                                  */
/*  Returns:                                                        */
/*           int  dmSUCCESS or dmFAILURE                            */
/*                                                                  */
/********************************************************************/
int dmBlockCreateSubspaceCpt(dmBlock* block)
{
  if ( block )
    return dmi_subspace_create_cpt( block->subspace, 0 );
  else
    return dmFAILURE;
}


/********************************************************************/
/*                                                                  */
/* Function:   dmBlockCreateEmptySubspaceCpt                        */
/*                                                                  */
/* Description:                                                     */
/*     Make a new empty component for the block                     */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmBlock      *block - block pointer                  */
/*                                                                  */
/*  Output parameters:                                              */
/*              none                                                */
/*                                                                  */
/*  Returns:                                                        */
/*           int  dmSUCCESS or dmFAILURE                            */
/*                                                                  */
/********************************************************************/
int dmBlockCreateEmptySubspaceCpt(dmBlock* block)
{
  if ( block )
    return dmi_subspace_create_cpt( block->subspace, 1 );
  else
    return dmFAILURE;
}



/*----------------------------------------------------------------------------
                  MISCELLANEOUS DATA SUBSPACE API FUNCTIONS
----------------------------------------------------------------------------*/

/********************************************************************/
/*                                                                  */
/* Function:   dmSubspaceColSetTableName                            */
/*                                                                  */
/* Description:                                                     */
/*     Set the table name for the current component of the input    */
/*     DSS descriptor                                               */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmDescriptor *dmDSS - subspace descriptor            */
/*             char         *name  - component name                 */
/*                                                                  */
/*  Output parameters:                                              */
/*              none                                                */
/*                                                                  */
/*  Returns:                                                        */
/*           int  dmSUCCESS or dmFAILURE                            */
/*                                                                  */
/********************************************************************/
int dmSubspaceColSetTableName(dmDescriptor* dmDSS, char* name)
{
  dmFilter *filter = NULL;
  int       retval = 0;

  if(!dmDSS || !dmDSS->block || dmDSS->dim > 1 )
  {
      dmu_err_set( DM_ERR_NULL, "No subspace descriptor" );
      return dmFAILURE;
  }

  filter = dmi_subspace_col_get_current_filter( dmDSS );
  retval = dmi_filter_set_tabname( filter, name );
  dmi_descriptor_set_update(dmDSS);
  return retval;  
}


/********************************************************************/
/*                                                                  */
/* Function:   dmBlockGetSubspace                                   */
/*                                                                  */
/* Description:                                                     */
/*     Return list of subspace column descriptors in block          */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmBlock      *block - block pointer                  */
/*             long         *ncols - number of columns              */
/*             long         *ncpts - number of components           */
/*                                                                  */
/*  Output parameters:                                              */
/*              none                                                */
/*                                                                  */
/*  Returns:                                                        */
/*           dmDescriptor  **ddlist                                 */
/*                                                                  */
/********************************************************************/
dmDescriptor** dmBlockGetSubspace( dmBlock* block, long* ncols, long* ncpts )
{
  dmDescriptor **ddlist = NULL;
  long           Ncols  = 0;
  long           col    = 0;

  if ( block )
  {
    Ncols = dmBlockGetNoSubspaceCols( block );
    *ncols = Ncols;
    *ncpts = dmBlockGetNoSubspaceCpts( block );

    if ( Ncols > 0 )
      ddlist = (dmDescriptor**)calloc( Ncols, UT_SZ_P );

    for ( col = 0; col < Ncols; col++ )
    {
      ddlist[col] = dmBlockGetSubspaceColNo( block, col+1 );
    }
  }
  return ddlist;
}


/********************************************************************/
/*                                                                  */
/* Function:   dmBlockGetSubspaceColNo                              */
/*                                                                  */
/* Description:                                                     */
/*     For input n, get the nth descriptor on our list of open      */
/*     subspace columns.  If it doesn't exist, return NULL but      */
/*     don't set error.                                             */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmBlock      *block - block pointer                  */
/*             long          colNum - column number                 */
/*                                                                  */
/*  Output parameters:                                              */
/*              none                                                */
/*                                                                  */
/*  Returns:                                                        */
/*           dmDescriptor  *col_descriptor                          */
/*                                                                  */
/********************************************************************/
dmDescriptor* dmBlockGetSubspaceColNo(dmBlock* block, long colNum)
{
  if(!block)
  {
      dmu_err_set(DM_ERR_NULL, "Null arg");
      return NULL;
  }

  if( !block->subspace || colNum <= 0 || colNum > block->subspace->ncols )
  {
      return NULL;
  }

  return block->subspace->cols[colNum-1];
}


/********************************************************************/
/*                                                                  */
/* Function:   dmBlockGetSubspaceColNo                              */
/*                                                                  */
/* Description:                                                     */
/*     Intersect the subspaces from block1 and block2 and copy the  */
/*     result to block0.  Runtime filters become permanent part of  */
/*     DSS on copy. Blocks needn't be unique.                       */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmBlock      *block1 - block pointer                 */
/*             dmBlock      *block2 - block pointer                 */
/*                                                                  */
/*  Output parameters:                                              */
/*             dmBlock      *block0 - intersected subspace block    */
/*                                                                  */
/*  Returns:                                                        */
/*           int  dmSUCCESS or dmFAILURE                            */
/*                                                                  */
/********************************************************************/
int dmBlockIntersectSubspace(dmBlock* block1, dmBlock* block2, dmBlock* block0)
{
  if(!block1 || !block2 || !block0)
  {
    dmu_err_set(DM_ERR_NULL, "Null arg");
    return dmFAILURE;
  }

  /* If blocks 1 & 0 are distinct, copy the DSS of block1 to block0 */
  if(block1 != block0)
  {
    if((!dmi_block_copy_subspace(block1, block0)))
      return dmFAILURE;
  }
  /* If blocks 1 & 0 the same, no effect */

  if(block2 != block0)
  {
    if((!dmi_block_copy_subspace(block2, block0)))
      return dmFAILURE;
  }
  return dmSUCCESS;
}


/********************************************************************/
/*                                                                  */
/* Function:   dmBlockGetSubspaceColNo                              */
/*                                                                  */
/* Description:                                                     */
/*     Take the subspaces of blocks 1 & 2 and merge (logical OR)    */
/*     them into block0.  Blocks needn't be unique.                 */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmBlock      *block1 - block pointer                 */
/*             dmBlock      *block2 - block pointer                 */
/*                                                                  */
/*  Output parameters:                                              */
/*             dmBlock      *block0 - merged subspace block         */
/*                                                                  */
/*  Returns:                                                        */
/*           int  dmSUCCESS or dmFAILURE                            */
/*                                                                  */
/********************************************************************/
int dmBlockMergeSubspace(dmBlock* block1, dmBlock* block2, dmBlock* block0)
{
  if(!block1 || !block2 || !block0)
  {
    dmu_err_set( DM_ERR_NULL, "No block" );
    return dmFAILURE;
  }

  /* If blocks 0 & 1 aren't the same and block1 has a subspace, */
  /* append the subspace of block1 to block0                    */

  if(block1 != block0 && block1->subspace->ncols > 0 )
  {
    if(!dmi_block_merge_subspace(block0, block1))
      return dmFAILURE;
  }

  /* If blocks 2 & 0 aren't the same and blocks 2 & 1 aren't the same  */
  /* (since we just appended 1 to 0) and block2 has a subspace, append */
  /*  the subspace of block2 to block0                                 */

  if(block2 != block0 && block2 != block1 && block2->subspace)
  {
    if(!dmi_block_merge_subspace(block0, block2))
      return dmFAILURE;
  }
  return dmSUCCESS;
}


/********************************************************************/
/*                                                                  */
/* Function:   dmBlockPrintSubspace                                 */
/*                                                                  */
/* Description:                                                     */
/*     Dump the DSS. This routine lists the current state of the    */
/*     DSS and is mainly intended for use within the debugger.      */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmBlock      *block - block pointer                  */
/*                                                                  */
/*  Output parameters:                                              */
/*             none                                                 */
/*                                                                  */
/*  Returns:                                                        */
/*           int  dmSUCCESS or dmFAILURE                            */
/*                                                                  */
/********************************************************************/
int dmBlockPrintSubspace(dmBlock* block )
{
  if ( !block)
    return dmFAILURE;

  printf( "----- Subspace for block %s -----\n", block->name );
  return dmi_subspace_print( block->subspace );
}


/********************************************************************/
/*                                                                  */
/* Function:   dmp_block_flush_subspace                             */
/*                                                                  */
/* Description:                                                     */
/*     Write out subspace                                           */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmBlock      *block - block pointer                  */
/*             dmBool        flush_tables - flag                    */
/*                                                                  */
/*  Output parameters:                                              */
/*             none                                                 */
/*                                                                  */
/*  Returns:                                                        */
/*           int  dmSUCCESS or dmFAILURE                            */
/*                                                                  */
/********************************************************************/
int dmp_block_flush_subspace(dmBlock* pBlock, dmBool flush_tables )
{
  dmDescriptor *dd       = NULL;
  dmSubspace   *subspace = NULL;
  dmBool        retval   = dmFALSE;
  long          ii       = 0;

  if(!pBlock || !pBlock->subspace)
  {
    return dmSUCCESS;
  }

  subspace = pBlock->subspace;

  if ( dmi_block_is_null_primary( pBlock ))
    return dmSUCCESS; 

  /* If no filters or DSS's, our work is done */
  /* Virtual blocks are not writeable */
  if ( dmi_block_writeable( pBlock ))
  {
    /* Loop over descriptors */
    for(ii=0; ii< subspace->ncols; ii++)
    {
      dd = subspace->cols[ii];
      /* Skip descriptors which haven't been updated */
      if(!(dmi_descriptor_check_update(dd)))
	continue;

      retval = dmp_subspace_col_write_values( pBlock, dd, subspace->ncpts, flush_tables );
      if ( !retval )
       goto error;
    }
  }
  error:
    if(dmu_err_get() == dmSUCCESS)
      return dmSUCCESS;
   else
      return dmFAILURE;
}


/********************************************************************/
/*                                                                  */
/* Function:   dmSubspaceColGetTableName                            */
/*                                                                  */
/* Description:                                                     */
/*     Return column filter table name                              */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmDescriptor *dd    - column descriptor              */
/*             long          maxlen - name buffer length            */
/*                                                                  */
/*  Output parameters:                                              */
/*             char          *name - table name                     */
/*                                                                  */
/*  Returns:                                                        */
/*           long  dmSUCCESS or dmFAILURE                           */
/*                                                                  */
/********************************************************************/
long dmSubspaceColGetTableName(dmDescriptor *dd, char *name, long maxlen)
{
  dmFilter *filter = NULL;

  if ( dd )
  {
    filter = dmi_subspace_col_get_current_filter( dd );
    dmi_filter_get_tabname( filter, name, maxlen );
    return dmSUCCESS;
  }
  else
    return dmFAILURE;
}


/********************************************************************/
/*                                                                  */
/* Function:   dmSubspaceColSetRegion                               */
/*                                                                  */
/* Description:                                                     */
/*     Attach region pointer to column descriptor                   */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmDescriptor *dd     - column descriptor             */
/*             regRegion    *region - region struct                 */
/*                                                                  */
/*  Output parameters:                                              */
/*             none                                                 */
/*                                                                  */
/*  Returns:                                                        */
/*           int   dmSUCCESS or dmFAILURE                           */
/*                                                                  */
/********************************************************************/
int dmSubspaceColSetRegion( dmDescriptor* dd, regRegion* region )
{
  if ( !dd || !region )
    return dmFAILURE;
  dmi_subspace_col_update_region( dd, NULL, region, UT_FALSE );
  return dmSUCCESS;
}

/********************************************************************/
/*                                                                  */
/* Function:   dmSubspaceColCreateRegionFromString                  */
/*                                                                  */
/* Description:                                                     */
/*     Create and populate region struct from input string          */
/*     Attach region to col descriptor                              */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmBlock      *block   - block pointer                */
/*             char         *name    - column name                  */
/*             char        **cptNames - component names             */
/*             dmDataType    type     - data type                   */
/*             char         *unit     - unit string                 */
/*             char         *region_string - region string          */
/*                                                                  */
/*  Output parameters:                                              */
/*             none                                                 */
/*                                                                  */
/*  Returns:                                                        */
/*           dmDescriptor  col_descriptor                           */
/*                                                                  */
/********************************************************************/
dmDescriptor* dmSubspaceColCreateRegionFromString( dmBlock* block, char* name,
 char** cptNames, dmDataType type, char* unit, char* region_string )
{
  regRegion *region = NULL;

  if ( block )
  {
    region = dmRegParse( region_string );    /* User create */
    return dmi_subspace_col_create_region( block, name, cptNames, type, unit, region );
  }
  else
    return NULL;
}



