/*                                                                
**  Copyright (C) 1997-2001,2003-2004,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmfilter.c                                                  */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include <string.h>
#include <stdio.h>
#include <float.h>
#include "dmlibP.h"
#include "cxcregion.h"

/* MCD TEMP - these go to dmilib.h */
dmBlock*  dmi_mask_find_block( regRegion* region );              /* MCD TEMP */
dmErrCode dmMaskWrite( dmDataset* ds, char* name, dmBlock* img );/* MCD TEMP */
char*     dmMaskName(void);                                      /* MCD TEMP */
int       dmi_mask_remove_blkname(char **ss_string);             /* MCD TEMP */

/********************************************************************/
/*                                                                  */
/* Function:   dmp_subspace_col_write_values                        */
/*                                                                  */
/* Description:                                                     */
/*    Write the subspace data for the input column.                 */
/*      - evaluates region and range values against defaults        */
/*      - consolidates if possible (absorb range into region)       */
/*      - generates string representation of subspace               */
/*        e.g. range of the form '12:30,50:60,200:203'.             */
/*             region of form 'Circle(4800,4800,100)'               */
/*      - Handles writing of subspace to subspace block             */
/*         GTI Table, REGION Table, MASK Image                      */
/*      - writes Subspace keywords to file                          */
/*                                                                  */
/*  Input parameters:                                               */
/*    dmBlock      *block         - Output block pointer            */
/*    dmDescriptor *pDesc         - Subspace column descriptor      */
/*    long          ncpts         - Number of subspace components   */
/*    dmBool        flush_tables  - Flag to flush subspace tables   */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    dmBool        retval        - TRUE  = Success (no error)      */
/*                                - FALSE = Error                   */
/*                                                                  */
/********************************************************************/
dmBool dmp_subspace_col_write_values( dmBlock* block, dmDescriptor* pDesc, long ncpts, dmBool flush_tables )
{
  dmBool  bWriteData     = UT_TRUE;    /* Flag to write ranges */
  dmBool  WriteRanges    = UT_TRUE;    /* Flag to write ranges */
  dmBool  WriteRegion    = UT_FALSE;   /* Flag to write region */
  dmBool  flush_complete = UT_TRUE;    /* Flag indicating all info flushed */
  dmBool  retval         = UT_FALSE;   /* Return code */

  dmFilter       *filter;
  dmFilter      **filters;
  dmDescriptor  **cpts;                /* subspace col components */
  dmDescriptor   *cptDesc;             /* subspace col component  */
  dmDescriptor   *colDesc = NULL;      /* coordinate descriptor */
  dmRangeList    *range;
  dmRangeList    *reg_ranges;
  dmKernel        kernel;
  dmDataset      *ds;
  dmSubspaceData *ssdata;
  dmBlock        *ssblock = NULL;      /* subspace block pointer       */
  regRegion      *region = NULL;
  regRegion      *bounds_region = NULL;

  char  *strFilter = NULL;             /* string rep of ranges         */
  char  *TabName   = NULL;             /* subspace block (table) name  */
  char  *FilterRegionString = NULL;    /* string rep of region         */
  char  *ReferenceTableCode = NULL;    /* subspace block ref key value */
  char   BoundsRegionString[512] = " ";/* string rep of bounds region  */

  long  vcpt  = 0;                     /* vector component             */
  long  cptno = 1;                     /* subspace component number    */
  long  dim   = 0;                     /* number of vector cpts        */

  int  *WroteTable       = NULL;       /* Flags subspace block written */
  int   HaveBoundsRegion = 0;          /* Flag that range can be expressed as a bounds Rectangle */

  double fieldx[2];
  double fieldy[2];
  double xbounds[2];
  double ybounds[2];

  static int ctrl_mask_blk = 0;        /* control to write only ONE MASK block    */
  char *mask_name = dmMaskName();      /* mask shape name              */

  if ( !pDesc )
    return UT_TRUE;

  /* Don't write BIT type subspace values */
  if ( pDesc->type == DT_BIT )
    return UT_TRUE;

  ds = block->ds;
  kernel = ds->kernel;

  /* Extract filters from subspace column descriptor */
  ssdata = dmi_ss_get_data( pDesc );
  filters = ssdata->filters;
  if ( !filters )
    return UT_TRUE;  /* nothing to do */

  /* Make sure all existing subspace block names are unique */
  dmi_subspace_col_check_tabnames( pDesc, ncpts );

  /* Get coordinate descriptor corresponding to subspace column, if exists. */
  /* NOTE: Not used until checking default ranges, but done outside loop    */
  colDesc = dmi_block_search_public_identifier( block, pDesc->name );

  /* Subspace descriptor vector dimension is on 'dim' element, not 'vdim' */
  dim = pDesc->dim;
  if ( dim > 1 )
  {
    cpts = pDesc->cpts;
    WroteTable = calloc( dim, UT_SZ_Q );
  }
  else
  {
    cpts = &pDesc;
    WroteTable = calloc( 1, UT_SZ_Q );
  }

  /* Loop over subspace components */
  for ( cptno = 1; cptno <= ncpts; cptno++ )
  {

    /* Get filter for this subspace component */
    filter = filters[cptno-1];

    if ( !filter ) 
      continue;  /* Skip to next cpt */

    /* Initialize Write Flags (only region?) */
    WriteRegion = ( filter->region != NULL );

    /* Handle null filter */
    if ( filter->null )
    {
      if ( filter->name )
      { /* not sure why this is here.. memory leak? */
        free( filter->name );
        filter->name = NULL;
      }

      WriteRegion = 0;  /* Turn off writing region data for this subspace component            */
      WriteRanges = 0;  /* Turn off writing range data for ALL remaining subspace components?  */

      retval = dmp_subspace_col_write_keys( block, pDesc, cptno, "NULL", NULL );
      if ( !retval )
        return retval;
    } 

    if ( WriteRegion )
    {
      region = filter->region;

      if ( filter->dim == 2 && ar_mix_is_numeric( filter->type ) )
      {
        /* Make string representation of region */
        FilterRegionString = regAllocComposeRegion( region );

        /* This is an attempt to merge the rangelist with the region. */
        HaveBoundsRegion = dmi_rangelist_to_rectangle( &filter->ranges[0], &filter->ranges[1], BoundsRegionString );
        if( HaveBoundsRegion )
        {
          /* Default region, ignore */
          if ( cs_uceq( FilterRegionString, BoundsRegionString )) 
            WriteRegion = 0;     
          else
          {
            /* compare the region bounds with the rangelist to see if we need to */
            /* append it to the region                                           */

            dmi_descriptor_get_cpt_range( pDesc, 1, DT_REAL8, &fieldx[0], &fieldx[1] );
            dmi_descriptor_get_cpt_range( pDesc, 2, DT_REAL8, &fieldy[0], &fieldy[1] );
            regExtent( filter->region, fieldx, fieldy, xbounds, ybounds ); 
            reg_ranges = dmi_rangelist_alloc_array( NULL, 2 );
            dmi_rangelist_set( &reg_ranges[0], DT_REAL8, &xbounds[0], &xbounds[1], 1 );
            dmi_rangelist_set( &reg_ranges[1], DT_REAL8, &ybounds[0], &ybounds[1], 1 );
      
            if ( !dmi_rangelist_compare( &filter->ranges[0], &reg_ranges[0], DT_REAL8 ) ||
	         !dmi_rangelist_compare( &filter->ranges[1], &reg_ranges[1], DT_REAL8 ))
            {
              /* append the rangelist to the region */
              bounds_region = regParse( BoundsRegionString );   /* From rectangle */
              region = regIntersectRegion( bounds_region, filter->region );
              regFree( bounds_region );
            }
            free(reg_ranges);
          }
        }
        if ( FilterRegionString )
        {
          free( FilterRegionString );
          FilterRegionString = NULL;
        }

      } /* end if 2D numeric region */

      if ( WriteRegion )
      {
        /* Make string rep of region, possibly modified above. */
        FilterRegionString = regAllocComposeRegion( region );

	/* If the region includes MASK shape, ensure that the filter          */
	/* has a block name defined.  analagous to dmi_subspace_col_is_gti()  */
        /* Assumes Mask is first shape on first component as per requirement  */
        if ( cs_uceq_siz( FilterRegionString, mask_name, strlen(mask_name) ) ||
             cs_uceq_siz( FilterRegionString+1, mask_name, strlen(mask_name)  ) )
        {
           if (!filter->name || !(*filter->name) )
            filter->name = cs_dup( mask_name );
        }

        TabName = filter->name;
        if ( !cs_is_blank( TabName ) )
        {
          /* Writing Region info to subspace block.. 2 possibilities          */
          /*   1. MASK image with possible region syntax as well              */
          /*   2. REGION Table - can not include MASK shape                   */
          /* Currently destinquished by MASK shape in region.                 */

          /* Set reference keyword (DSREFn) value */
          ReferenceTableCode = (char*)malloc((strlen(TabName)+3)* UT_SZ_C );
          sprintf( ReferenceTableCode, ":%s", TabName );

          /* See if there is a Mask in the region */
          ssblock = dmi_mask_find_block( region );
          if ( ssblock != NULL )
          {
            /* Adjust DSVALn content as needed (remove block name)  */
            /*  - handles multiple occurrences within expression    */
            dmi_mask_remove_blkname( &FilterRegionString );

            if ( flush_tables )
            {
              /* MCD NOTE: At this point, mask block is on an external dataset,  */
              /* this method transfers the block to the destination dataset.     */
              /* IF/when it moves to the input dataset, this will need to change */
              /* so that the ssblock is flushed instead.                         */
              /* The need for the control flag will go a way too, since block    */
              /* will not show need for update after the first time.             */
              if ( ctrl_mask_blk == 0 )
              {
                 /* Transfer block to working dataset */
                 if ( dmMaskWrite( block->ds, TabName, ssblock ) == dmFAILURE )
                   return UT_FALSE;

                 ctrl_mask_blk++;
              }

            }
            else
            {
              flush_complete = UT_FALSE;
            }
          }
          else
          {
            /* Assume REGION Block */
            if ( flush_tables )
            {
              free( FilterRegionString );
              FilterRegionString = cs_dup( "TABLE" );

              ssblock = dmTableWriteRegion( block->ds, TabName, pDesc, region );
  
              if ( !ssblock )
                return UT_FALSE;
  
              /* Close the block!  Otherwise, things get confused at flush time.*/
              dmBlockClose( ssblock );
            }
            else
            {
              flush_complete = UT_FALSE;
            }
          }
        } /* end if TabName */

        /* Write subspace keys */
        retval = dmp_subspace_col_write_keys( block, pDesc, cptno, FilterRegionString,
                                              ReferenceTableCode );

        if ( !retval )
          return retval;

      } /* end inner if WriteRegion */

      if ( ReferenceTableCode )
      {
        free( ReferenceTableCode );    
        ReferenceTableCode = NULL;
      }

      if ( FilterRegionString )
      {
        free( FilterRegionString );
        FilterRegionString = NULL;
      }

      if ( region != filter->region )
        regFree( region );

    } /* end if WriteRegion */


    /* MCD NOTE: This should not write to stderr.. add to DM error stack */
    if ( filter->dim != dim )
      fprintf( stderr, "DM INTERNAL ERROR: Mismatch of filter and descriptor dimensions for %s\n", pDesc->name );


    retval = UT_TRUE;

    /* the region may have absorbed the ranges so don't bother. */
    if ( !WriteRegion || !HaveBoundsRegion )
    {
      for ( vcpt = 1; vcpt <= filter->dim; vcpt++ )
      { 
        cptDesc = cpts[vcpt-1];
        range = &filter->ranges[vcpt-1];

        /* If the filter is equal to TLMIN/TLMAX, omit it */
        if ( WriteRanges ) /* If was null, false for all cpts */
        {
          bWriteData = UT_TRUE;

          if ( dmi_rangelist_is_empty( range ))
            bWriteData = UT_FALSE;

          if ( colDesc )
          {
            /* Check the range against column descriptor to see if is default.. */
            if ( dmi_filter_range_is_default( colDesc, filter, vcpt )) 
              bWriteData = UT_FALSE;
          }
          else
	  {
            /* use subspace descriptor when there is no column descriptor */
            if ( dmi_filter_range_is_default( pDesc, filter, vcpt )) 
              bWriteData = UT_FALSE;
	  }

          if ( dmi_filter_range_is_tstart( cptDesc, filter, vcpt )) 
            bWriteData = UT_FALSE;
        }
        else
        {
          bWriteData = UT_FALSE;
        }
    
        if ( bWriteData )
        {
          /* Make sure GTI subspace has table name */
          dmi_subspace_col_is_gti(cptDesc, filter );

          TabName = filter->name;
          if ( !cs_is_blank( TabName ) && !WriteRegion )
          {
            /* We are writing a TABLE, and do NOT have a region. */

            /* Create DSREFn keyword value.. */
            ReferenceTableCode = (char*)malloc((strlen(TabName)+3)* UT_SZ_C );
            sprintf( ReferenceTableCode, ":%s", TabName );    

            /* Delay flushing the tables till block close */
            if ( flush_tables )
            {
              kernel->SubspaceTableWrite( cptDesc, TabName, range->mins, range->maxes, range->n );
              dmi_block_update_gti_ccd_key( block, TabName, cptno );
            }
            else
            {
              flush_complete = UT_FALSE;
            }

            /* Define DSVALn keyword value.. first subspace component only.. others NULL */
            if ( !WroteTable[vcpt-1] ) 
            {
              strFilter = cs_dup( "TABLE" );

              if ( cptno == 1 )
                WroteTable[vcpt-1] = UT_TRUE;
            }
            else
            {
              strFilter = NULL;
            }
          }
          else
          {
            /* Define DSVALn keyword value for range lists. */
            ReferenceTableCode = NULL;
            dmi_rangelist_compose_string( range, pDesc->type, &strFilter );
            WroteTable[vcpt-1] = UT_FALSE;      
          }

          /* Write subspace keys  */
          retval = dmp_subspace_col_write_keys( block, cptDesc, cptno, strFilter,
                                                 ReferenceTableCode );
          if ( ReferenceTableCode )
          {
            free( ReferenceTableCode );
            ReferenceTableCode = NULL;
          }

          if ( strFilter )
          {
            free( strFilter );
            strFilter = NULL;
          }

        } /* end if bWriteData  */
      } /* end for vcpt */
    } /* end  if ( !WriteRegion || !HaveBoundsRegion ) */
  } /* end for cptno */
 
  if ( WroteTable )
    free( WroteTable );
 
  if ( flush_complete )
    dmi_descriptor_clear_update( pDesc );

  if(mask_name)
    free(mask_name);

  return retval;
}




/********************************************************************/
/*                                                                  */
/* Function:   dmi_block_update_gti_ccd_key                         */
/*                                                                  */
/* Description:                                                     */
/*    CXC requires the GTIs to have keywords for the CCD they       */
/*    represent. This breaks the DM paradigm (should go to the event*/
/*    block and read the DSS; once users filter the data may end up */
/*    with multiple GTIs per CCD anyway in messy cases).            */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmBlock  *block                                      */
/*             char     *tabname                                    */
/*             long      cpts                                       */
//*                                                                 */
/*  Output parameters:                                              */
/*              none                                                */
/*                                                                  */
/*  Returns:                                                        */
/*           nothing                                                */
/*                                                                  */
/********************************************************************/
void dmi_block_update_gti_ccd_key( dmBlock* block, char* tabname, long cpt )
{
  dmDescriptor    *ccd    = NULL;
  dmSubspaceData  *ssdata = NULL;
  dmFilter        *filter = NULL;
  dmRangeList     *range  = NULL;
  dmBlock         *gti    = NULL;
  dmDataset       *ds     = NULL;

  double id = 0;
  long   no = 0;
  long   idval = 0;

  dmBool flush_gtis = UT_FALSE; /* Prevent infinite recursion */


  if ( !block || !block->ds )
    return;

  ds = block->ds;

  if ( ds->nblocks <= 1 )
   return;

  ccd = dmSubspaceColOpen( block, "CCD_ID" );

  if ( !ccd )
    return;

  ssdata = dmi_ss_get_data( ccd );

  if ( ssdata )
    filter = ssdata->filters[cpt-1];

  if ( filter )
    range = &filter->ranges[0];

  if ( !range || range->n != 1 )
    return;

  id = ((double*)range->mins)[0];

  if ( id != ((double*)range->maxes)[0] )
    return;

  idval = id;

  /* Only open the block if it exists; fail silently */
  no = dmi_block_search( block->ds, tabname );

  if ( no > 0 )
    gti = dmBlockOpen( block->ds, tabname );

  if ( gti )
  {
    dmDescriptor* key = dmi_key_open( gti, "TIMESYS" );  /* Position for writing key */

    if ( key )
      dmi_block_set_keyno( block, key->no );

    dmKeyWrite_l( gti, "CCD_ID", idval, " ", "CCD for this table" );

    dmp_block_close_mode( gti, flush_gtis );
  }
  else
  {
    sprintf( dm_errbuf, "GTI table %s not found in CCD key hack", tabname );
    dmu_err_set( 0, dm_errbuf );
  }
}


/********************************************************************/
/*                                                                  */
/* Function:   dmp_subspace_col_write_keys                          */
/*                                                                  */
/* Description:                                                     */
/*                                                                  */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmBlock      *block                                  */
/*             dmDescriptor *dd                                     */
/*             long          count                                  */
/*             char         *value                                  */
/*             char         *TabName                                */
//*                                                                 */
/*  Output parameters:                                              */
/*              none                                                */
/*                                                                  */
/*  Returns:                                                        */
/*           int status                                             */
/*                                                                  */
/********************************************************************/
int dmp_subspace_col_write_keys( dmBlock* block, dmDescriptor* dd, long count,
                                 char* value, char* TabName )
{
  dmDescriptor  *vector = NULL;
  dmDataset     *ds     = block->ds;
  dmKernel       kernel = ds->kernel;
 
  /* Note that the update flag is set for the vector parent only and not for the components */
  vector = (dd->cptno>0) ? dd->parent : dd;

  if ( dmi_descriptor_check_update( vector ) ) 
    kernel->SubspaceDescriptorWrite( dd, count, value, TabName );

  return 1;
}




