/*                                                                
**  Copyright (C) 1999-2001,2003-2005,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmfilteropen.c                                              */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include <string.h>
#include <math.h>
#include <stdio.h>
#include <float.h>
#include "dmlibP.h"
#define XDEBUG 1


void dmp_read_ascii_bin_file( char* gridfile, double** mins, double** maxes, long* n );
void dmp_read_bin_file( char* gridfile, double** mins, double** maxes, long* n );


/********************************************************************/
/*                                                                  */
/* Function:   dmBinningDisp                                        */
/*                                                                  */
/* Description:                                                     */
/*      Display binning information                                 */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmBlock      *block                                  */
/*             char         *name                                   */
/*             char         *spec                                   */
/*             long          minrow                                 */
/*             long          maxrow                                 */
/*                                                                  */
/*  Output parameters:                                              */
/*              none                                                */
/*                                                                  */
/*  Returns:                                                        */
/*           void                                                   */
/*                                                                  */
/********************************************************************/
void dmBinningDisp( dmBlock* block, char* name, char* spec, long minrow, long maxrow )
{
  char typename[80];
  char tag[80] = "";

  double *mins = NULL;
  double *maxes = NULL;

  long type   = 0;
  long n      = 0;
  long imin   = 1;
  long imax   = 1;
  long maxlen = 80;
  long ii     = 0;
  int compact = 1;
  int partial = 0;

  type = dmBinningParse( block, name, spec, compact, &mins, &maxes, &n, &partial );
  dmBinningTypeName( type, typename, maxlen );
  printf( "Binning variable: %s\nSpecification: %s\nType: %s  Bins: %ld\n", name,
          spec, typename, n );
 
  if ( minrow > 0 )
    imin = minrow;

  imax = n;

  if ( maxrow > 0 && n > maxrow )
    imax = maxrow;

  printf( "Bin            Min               Max                 Size \n" );

  for ( ii = imin-1; ii < imax; ii++ )
  {
    if ( ii == n-1 && partial )
      strcpy( tag, "(partial bin)" );

    printf( "%4ld %20.12g:%20.12g %20.12g %s\n", ii+1, mins[ii], maxes[ii], maxes[ii]-mins[ii], tag );
  }

  if ( n > 0 )
  {
    free( mins );
    free( maxes );
  }
}


/********************************************************************/
/*                                                                  */
/* Function:   dmBinningTypeName                                    */
/*                                                                  */
/* Description:                                                     */
/*       Map binning return values to binning types                 */
/*                                                                  */
/*  Input parameters:                                               */
/*             long          type                                   */
/*             long          maxlen    - max length of name buffer  */
/*                                                                  */
/*  Output parameters:                                              */
/*             char         *ttypename                              */
/*                                                                  */
/*  Returns:                                                        */
/*           void                                                   */
/*                                                                  */
/********************************************************************/
void dmBinningTypeName( long type, char* ttypename, long maxlen )
{
  char typename[512];

  if( type < 0 )
  {
    strcpy( ttypename, "None" );
    return;
  }

  if ( type & DM_BINS_N )
    strcpy( typename, "Fixed number " );
  else
    strcpy( typename, "Fixed size " );

  if ( type & DM_BINS_FILE )
    strcat( typename, "tabulated bins [Unsupported in DMCOPY]" );
  else
    if ( type & DM_BINS_LOG )
    {
      strcat( typename, "logarithmic bins [Unsupported in DMCOPY]" );
    }
    else
      strcat( typename, "linear bins " );

  if (( type & DM_BINS_COMPACT ) && !(type & DM_BINS_FILE))
    strcat( typename, "(compact default)" );

  strncpy( ttypename, typename, maxlen );
  ttypename[maxlen-1] = 0;
}


/********************************************************************/
/*                                                                  */
/* Function:   dmp_read_bin_file                                    */
/*                                                                  */
/* Description:                                                     */
/*     Read bin data from DM file                                   */
/*                                                                  */
/*  Input parameters:                                               */
/*             char       *gridfile- DM bin file                    */
/*                                                                  */
/*  Output parameters:                                              */
/*             double    **mins                                     */
/*             double    **maxes                                    */
/*             long       *n     - number of bin entries            */
/*                                                                  */
/*  Returns:                                                        */
/*           void                                                   */
/*                                                                  */
/********************************************************************/
void dmp_read_bin_file( char* gridfile, double** mins, double** maxes, long* n )
{
  char blockname[TEXT_SIZE];
  dmBlock      *block  = NULL;
  dmDescriptor *xmin   = NULL;
  dmDescriptor *xmax   = NULL;
  double       *vmins  = NULL;
  double       *vmaxes = NULL;
  long ncols = 0;
  long nrows = 0;
 
  block = dmTableOpen( gridfile );
  *n = 0;
  *maxes= NULL;
  *mins = NULL;

  if ( !block )
  {
    if ( dmverbose >=4 )
    {
      fprintf( stderr, "Grid file %s is not DM; try ASCII\n", gridfile );
      dmResetErrorState();
    }
    dmp_read_ascii_bin_file( gridfile, mins, maxes, n );
    return;
  } 

  if ( !strchr( gridfile, '[' ) )  /* A bare filename */
  {
    dmBlockGetName( block, blockname, TEXT_SIZE );

    if ( cs_uceq( blockname, "EBOUNDS" ))
    {
      /* We have an RMF with EBOUNDS first. Default to using the MATRIX information */
      dmTableClose( block );
      sprintf( blockname, "%s[MATRIX][cols ENERG_LO,ENERG_HI]", gridfile );
      block = dmTableOpen( blockname );

      if ( !block )
      {
        sprintf( dm_errbuf, "Grid file %s not found\n", blockname );
        dmu_err_set( DM_ERR_PARSE, dm_errbuf );
      }
    }
  }

  nrows = dmTableGetNoRows( block );
  ncols = dmTableGetNoCols( block );

  if ( ncols < 2 )
  {
    sprintf( dm_errbuf, "Less than 2 columns in grid file %s", gridfile );
    dmu_err_set( DM_ERR_PARSE, dm_errbuf );
    return;
  }

  xmin = dmTableOpenColumnNo( block, 1 );   
  xmax = dmTableOpenColumnNo( block, 2 );

  if ( !xmin || !xmax )
  {
    sprintf( dm_errbuf, "Less than 2 columns in grid file %s", gridfile );
    dmu_err_set( DM_ERR_PARSE, dm_errbuf );
    return;
  }

  if ( nrows > 0 ) 
  {
    vmins = calloc( nrows, UT_SZ_D);
    vmaxes = calloc( nrows, UT_SZ_D);
    dmGetScalars_d( xmin, vmins, 1, nrows );
    dmGetScalars_d( xmax, vmaxes, 1, nrows );
  }

  dmTableClose( block );
  *n = nrows;
  *mins = vmins;
  *maxes= vmaxes;
}


/********************************************************************/
/*                                                                  */
/* Function:   dmp_read_ascii_bin_file                              */
/*                                                                  */
/* Description:                                                     */
/*     Read bin data from ascii file                                */
/*                                                                  */
/*  Input parameters:                                               */
/*             char       *gridfile- ascii bin file                 */
/*                                                                  */
/*  Output parameters:                                              */
/*             double    **mins                                     */
/*             double    **maxes                                    */
/*             long       *n     - number of bin entries            */
/*                                                                  */
/*  Returns:                                                        */
/*           void                                                   */
/*                                                                  */
/********************************************************************/
void dmp_read_ascii_bin_file( char* gridfile, double** mins, double** maxes, long* n )
{
  char  buf[dmMAXLEN];
  char   *ptr    = buf;
  char   *rem    = NULL;
  double *vmins  = NULL;
  double *vmaxes = NULL;
  double  xmin   = 0.0;
  double  xmax   = 0.0;
  long    nmax   = 100;
  long    m      = 0;

  FioFile fptr;

  fptr = fio_open_ar( gridfile );
  if ( !fptr )
    return;

  vmins  = calloc( nmax, UT_SZ_D);
  vmaxes = calloc( nmax, UT_SZ_D);

  while( fio_read_line( fptr, buf, dmMAXLEN ))
  {
    ptr = buf;

    while( *ptr == ' ' || *ptr == '\t' )
      ++ptr;

    if ( *ptr != '\0' && *ptr != '#' )
    {
      xmin = strtod( buf, &ptr );

      if ( ptr )
        xmax = strtod( ptr, &rem );
      else
        xmax = xmin;

      if ( m >= nmax ) 
      {
        nmax *= 5;
        vmins  = realloc( vmins,  nmax * UT_SZ_D);
        vmaxes = realloc( vmaxes, nmax * UT_SZ_D);
      }

      vmins[m] = xmin;
      vmaxes[m] = xmax;
      ++m;
    }
  }

  *mins = vmins;
  *maxes = vmaxes;
  *n = m;
  fio_file_close( fptr );
}


/********************************************************************/
/*                                                                  */
/* Function:   dmBinningParse                                       */
/*                                                                  */
/* Description:                                                     */
/*                                                                  */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmBlock    *block                                    */
/*             char       *name                                     */
/*             char       *spec                                     */
/*             int         compact                                  */
/*                                                                  */
/*  Output parameters:                                              */
/*             double    **mins                                     */
/*             double    **maxes                                    */
/*             long       *n                                        */
/*             int        *partial                                  */
/*                                                                  */
/*  Returns:                                                        */
/*           long  retval                                           */
/*                                                                  */
/********************************************************************/
long dmBinningParse( dmBlock *block, 
		     char *name,
		     char *spec,
		     int   compact,
		     double **mins,
		     double **maxes,
		     long *n,
		     int *partial )
{
  dmDescriptor *fcol = NULL;

  double *vals   = NULL;  /* Only support 1 dim */
  double *vmins  = NULL;
  double *vmaxes = NULL;
  double *vvals[2];
  double  BIN_EPSILON = 0.000001;
  double  tmax  = 0.0;
  double  tstep = 0.0;
  double  t     = 0.0;
  double  xmin  = 0.0;
  double  xmax  = 0.0;
  double  dx    = 0.0;
  double  tmin  = 0.0;
  long    retval   = DM_BINS_NONE;
  long    i        = 0;
  long    dim      = 0; 
  long    imax     = 0;
  int     grid     = 0;
  int     logbins  = 0;
  int    force_integer = 0;


  if ( name && cs_uceq( name, "#int" ))
    force_integer = 1;
  else if ( block && name )
    fcol = dmi_block_search_col_identifier( block, name );

  if (partial )
    *partial = 0;

  *n = 0;
  *mins  = NULL;
  *maxes = NULL;

  if (!fcol && dmverbose > 1 && name )
    fprintf( stderr, "No match for variable :%s:\n", name );

  retval = dmi_eval_bin_spec( block, fcol, spec, vvals, &dim, compact, force_integer );

  grid = ( retval & DM_BINS_FILE );

  if ( grid )
  {
    char *fptr = NULL;
    char *gridfile = malloc( strlen( spec ));
    char *optr = gridfile;

    fptr = spec+5;

    while( *fptr && *fptr != ')' )
    {
      *optr = *fptr++;
      ++optr; 
    }

    *optr = '\0';
    dmp_read_bin_file( gridfile, mins, maxes, n );

    free( gridfile );
    return retval;
  }
 
  logbins = ( retval & DM_BINS_LOG );
  vals= vvals[0];
  dx   = vals[2];

  if ( logbins )
  {
    xmin = log10( vals[0] );
    xmax = log10( vals[1] );
    tstep = dx;
  }
  else 
  {
    xmin =  vals[0];
    xmax =  vals[1];
    tstep =  ( xmin < xmax ) ? dx : -dx;
  }

  tmin = 0.0;
  tmax = ( xmax - xmin ) / tstep;
  imax = (long) (tmax - tmin + 1.0 - BIN_EPSILON);
  *n   = imax;

  if ( dmverbose >= 4 )
    fprintf( stderr, "MAX BINS = %ld\n", imax );

  if ( imax <= 0 )
  {
    retval = DM_BINS_NONE;
  }
  else 
  {
    vmins  = calloc( imax, UT_SZ_D);
    vmaxes = calloc( imax, UT_SZ_D);
    *mins  = vmins;
    *maxes = vmaxes;

    if ( ! (vmins) || ! (vmaxes ))
    {
      fprintf( stderr, "Failed to allocate memory for data\n" );
      return retval;
    }
  }

  for ( i = 0; i < imax; i++ )
  {
    t = xmin + i * tstep;

    if ( logbins )
    {
      (vmins)[i]  = pow( 10.0, t );
      (vmaxes)[i] = pow( 10.0, t + tstep );
    }
    else
    {
      (vmins)[i]  = t;
      (vmaxes)[i] = t + tstep;
    }
  }  

  /* How full is the top bin? */
  if ( imax > 0 && vmaxes )
    vmaxes[imax-1] = vals[1];

  if ( imax - tmax > BIN_EPSILON && partial )
    *partial = 1;

  for ( i = 0; i < dim; i++ )
    free( vvals[i] ); 

  return retval;
}

