/*****************************************************************/
/*                                                               */
/*   dmmask.c                                                    */
/*                                                               */
/*   Functions that work with PixelMask                          */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmlibP.h"
#include "../../region/region_priv.h"

/* Element to hold 'external' block references for MASK blocks */
/* Not sure when it is safe to close them yet...               */
/* MCD NOTE: To be removed when issue is resolved.             */
#define DM_MASK_NMAX_DEFAULT 20
struct dmMaskBlockLog_s{
  dmBlock* blocks[DM_MASK_NMAX_DEFAULT];  /* arry of block pointers */
  long     numblks;                       /* number of the array <MASK_NMAX */
};
typedef struct dmMaskBlockLog_s dmMaskBlockLog;
dmMaskBlockLog* dm_maskblklog;

/**************************************/
/* function called in DM modules      */
/**************************************/

regRegion* dmMaskParse( char* regstring);
dmErrCode  dmMaskWrite( dmDataset *dsbyte, const char *tabname, dmBlock *msk_blk);
char*      dmMaskName(void);
dmBlock*   dmi_mask_find_block( regRegion* region );
dmErrCode  dmi_mask_set_coord(  regShape *shape, dmBlock *flt_blk, char *cptname);
int        dmi_mask_remove_blkname(char **ss_string);
dmBool     dmMaskIsValidMask(dmBlock* msk_blk, dmBlock *flt_blk,  char *cptname, int lvl);
void       dmi_mask_free_blocks(void);

regRegion* dmi_mask_region_from_string( dmDescriptor* ssd, const char* dsref, const char* dsval );

/**************************************/
/* bridge functions to region-lib     */
/**************************************/

int       dmi_mask_inside(  regShape *shape, double xpos, double ypos);
int       dmi_mask_extent(  regShape *shape, double *xpos, double *ypos );
regShape* dmi_mask_copy(    regShape *inshp);
double    dmi_mask_area(    regShape *shape );
int       dmi_mask_isEqual( regShape *shp1, regShape *shp2);
char*     dmi_mask_toString(regShape *shape);
void      dmi_mask_free(    regShape *shape);

/**************************************/
/* all local (Mask) functions      */
/**************************************/

regRegion*  dmi_mask_read_mask( dmBlock* msk_blk );
regShape*   dmi_mask_create(void *ptr);

dmBool      dmi_mask_isvalid_shape(regShape *shape);
dmBool      dmi_mask_isvalid_mask(   dmBlock* msk_blk );
dmBool      dmi_mask_isvalid_content(dmBlock* msk_blk, long *nvals, unsigned char **vbytes);
dmBool      dmi_mask_isvalid_coord(  dmBlock *msk_blk, dmBlock *flt_blk, char *cptname, int *cflag);

char*       dmi_mask_check_syntax( char* ptr, int *incl );
void*       dmi_mask_get_image_array(dmBlock *msk_blk,  dmDataType *dtype, long *imgsz);
void*       dmi_mask_convert_values(void *values, dmDataType dtype, long nvals );
void        dmi_mask_set_includeflag(regShape *shape, int fincl);
int         dmi_mask_get_coordflag(dmBlock *blk1, dmBlock *blk2, char *cptname);

dmBlock*    dmi_mask_get_block(regShape* shape);
void        dmi_mask_log_blocks(dmBlock *block);

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_log_blocks                                   */
/*                                                                  */
/* Description:                                                     */
/*    record filter-Mask block pointer to the default arraies       */
/*                                                                  */
/*  Input parameters:                                               */
/*    Mask block pointer                                            */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
//*  Returns:                                                       */
/*    none                                                          */
/*                                                                  */
/********************************************************************/
void  dmi_mask_log_blocks(dmBlock *block)
{
  if(block==NULL)
    return;
  
  if(dm_maskblklog==NULL){
    dm_maskblklog = (dmMaskBlockLog*) malloc( sizeof(dmMaskBlockLog) );
    dm_maskblklog->numblks = 0;
    long ii;
    for(ii=0;ii<DM_MASK_NMAX_DEFAULT; ii++) 
      dm_maskblklog->blocks[ii] = NULL;
  } else if( dm_maskblklog->numblks > DM_MASK_NMAX_DEFAULT )     
    return;
  
  dm_maskblklog->blocks[dm_maskblklog->numblks] = block; 
  dm_maskblklog->numblks++;
  
  return;
}
/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_free_blocks                                  */
/*                                                                  */
/* Description:                                                     */
/*    free filter-Mask block pointers which are logged previously by*/
/*    function, dmi_mask_log_blocks                                 */
/*                                                                  */
/*  Input parameters:                                               */
/*    none                                                          */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
//*  Returns:                                                       */
/*    none                                      e                   */
/*                                                                  */
/********************************************************************/
void  dmi_mask_free_blocks()
{
  long ii;

  if(dm_maskblklog==NULL)
    return;
 
  for(ii=0;ii < dm_maskblklog->numblks; ii++){
    if(dm_maskblklog->blocks[ii] )
      dmBlockClose(dm_maskblklog->blocks[ii]);
  }
  free(dm_maskblklog);

  return;
}
/********************************************************************/
/*                                                                  */
/* Function:  dmMaskName                                            */
/*                                                                  */
/* Description:                                                     */
/*   Provides the 'name' of Mask region shape.                      */
/*    for use in region string                                      */
/*        MASK(blkname)                                             */
/*      or subspace block name                                      */
/* Note:                                                            */
/*   User must free the returned memory.                            */
/*                                                                  */
/*  Input parameters:                                               */
/*    none                                                          */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    char* name      - name of mask region shape                   */
/*                                                                  */
/********************************************************************/
char* dmMaskName(void)
{
  char *buff=calloc(5, sizeof(char) );
  strcpy(buff, "MASK");
  return buff;
}

/********************************************************************/
/*                                                                  */
/* Function:   dmMaskParse                                          */
/*                                                                  */
/* Description:                                                     */
/*   Parse a mask spec of the form                                  */
/*     sky=mask(file)                                               */
/*                                                                  */
/*   Note: dmRegParse handles 'region' spec                         */
/*                                                                  */
/*  Input parameters:                                               */
/*    char* regstring    Filter region expression for mask          */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    regRegion *region  Region struct with Mask shape              */
/*                                                                  */
/********************************************************************/
regRegion* dmMaskParse( char* regstring )
{
  dmErrCode  status = dmSUCCESS; 
  char* maskfile;
  dmBlock *block  = NULL;	
  int incl;

  if(regstring==NULL)
    return NULL;
  
  /* check and split mask invocation syntax */
  maskfile = dmi_mask_check_syntax( regstring, &incl );
  if(!maskfile)
    return NULL;

  block = dmImageOpen( maskfile );
  
  if (!block){
    if ( fio_access( maskfile ) )
      {
	status = DM_ERR_GENERIC;
	sprintf( dm_errbuf, "Failed to parse FITS mask file %s", maskfile );
      }
    else
     {
       status = DM_ERR_OPEN_FAIL;
       sprintf( dm_errbuf, "Mask file not found; %s", maskfile );
     }
    
    dmu_err_set( status, dm_errbuf );
    return NULL;
  }
  free( maskfile );

  regRegion* reg=dmi_mask_read_mask( block );
  if(reg==NULL){
    dmBlockClose((dmBlock*)block);
    return NULL;
  }

  regShape *shape=regGetShapeNo(reg, 1);
  if (incl) 
    dmi_mask_set_includeflag(shape, regInclude);
  else
    dmi_mask_set_includeflag(shape, regExclude);

  /*******************************************************************-*/
  /* log the open Block pointer and free it at the end of each program */
  /* this call only store the Mask from filter string but not          */
  /* the existing Mask extension of the event-list                     */
  /*******************************************************************-*/
  dmi_mask_log_blocks(block);

  return reg;
}

/********************************************************************/
/*                                                                  */
/* Function:   dmi_mask_find_block                                  */
/*                                                                  */
/* Description:                                                     */
/*   Find and return a Mask block from region structure             */
/*                                                                  */
/*  Input parameters:                                               */
/*    regRegion*  region   Region struct to scan                    */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    dmBlock*  mask       Block from first MASK shape.             */
/*                         (NULL if none found)                     */
/*                                                                  */
/********************************************************************/
dmBlock* dmi_mask_find_block( regRegion* region )
{
  dmBlock *block  = NULL;

  regShape *shape = NULL;

  if (region == NULL)
    return NULL;

  shape = region->shape;
  if (shape == NULL)
    return NULL;

  while ( (shape != NULL) && (block == NULL) )
  {
    if ( shape->type == regMASK )
    {
      block = (dmBlock*)shape->user->spec;
    }
    shape = shape->next;
  }

  return block;
}
/***************************************************************/
/*                                                             */
/* Function:  dmi_mask_check_syntax                            */
/*                                                             */
/* Description:                                                */
/*   Check string for appropriate mask invocation syntax.      */
/*   Extracts mask filename from spec.                         */
/*                                                             */
/*   spec can allow                                            */
/*     mask(<maskfile>)     standard invocation                */
/*     -mask(<maskfile>)    exclude syntax                     */
/*     @maskfile            stack syntax?  TBD                 */
/*                                                             */
/* Note:                                                       */
/*   it is not working for spec='mask(file)+circle(a,b,r)'     */
/*   MCD NOTE: Isn't this illegal syntax?                       HHE: leagal, it works with non-mask shapes */
/*   MCD NOTE: I see no requirement to support '-mask' syntax.  HHE  yes, not in the requirement */
/*                                                             */
/*  Input parameters:                                          */
/*    char* spec         filter spec segment                   */
/*                                                             */
/*  Output parameters:                                         */
/*    int*  incl         include/exclude flag                  */
/*                                                             */
/*  Returns:                                                   */
/*    char* filename     mask filename                         */
/*                        NULL if not mask syntax              */
/*                                                             */
/*  Error:                                                     */
/*    On error, sets DM error state                            */
/*                                                             */
/***************************************************************/
char* dmi_mask_check_syntax( char* spec, int *incl )
{
  char *regfile = NULL;

  if ( spec[0] == '@' )
  {
    /* regfile = cs_dup( spec+1 );*/
    sprintf( dm_errbuf, "Warning file-directive, '@', is not supported yet, '%s;", spec );
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return regfile;
  }

  /* Check for supported invocation tags */
  long pos_l,pos_r = 0;
  if ( cs_uceq_siz( spec, "-mask(", 6 ) )
  {
    pos_l = 6;
    *incl = 0;  /* regExclude */
  }
  else if ( cs_uceq_siz( spec, "mask(", 5 ) )
  {
    pos_l = 5;
    *incl = 1;  /* regInclude */
  }
  else { /* everything else error */
    pos_l=0;
  }

  /* find the first ")" */
  pos_r=cs_index(spec, ")");
  if( !pos_r               ||   /* if spec has no closing bracket, pos_r is 0 */
      strlen(spec) > pos_r ||   /* if spec is Union: mask()+circle(), the length is beyond the 1st ')' */
      !pos_l               )    /* if 'mask' tag is missing, pos_l=0  */
  {
    sprintf( dm_errbuf, "Syntax error or not supported syntax, '%s'.", spec );
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return NULL;
  }

  regfile=cs_dup( spec+pos_l );
  *(regfile+pos_r-pos_l-1)='\0';
  
  /*******************************************-*/
  /*check one more time for wrong syntax for   */
  /* mask(&&circle()                           */
  /*               ^                           */
  /*******************************************-*/
  pos_r=cs_index(regfile, ")");
  if( pos_r != 0 ) {  /* if a compound filter spec has no closing bracket in mask but right one in shape, the pos_r is not 0 */
    sprintf( dm_errbuf, "Syntax error or not supported syntax, '%s'.", spec );
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return NULL;
  }

  return regfile;
}

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_isvalid_mask                                 */
/*                                                                  */
/* Description:                                                     */
/*   Mask block validation for usability.                           */
/*                                                                  */
/*  Input parameters:                                               */
/*    dmBlock* block     Block to examine                           */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    dmBool             dmTRUE == valid ; dmFALSE == not valid     */
/*                                                                  */
/*                                                                  */
/*   NOTE: This routine is only checking USABILITY, see the sections*/
/*   The full Mask block requirements checking should be done in    */
/*   all 3 levels                                                   */
/*   validation levels:                                             */
/*     level-1: Mask validation,    dmi_mask_isvalid_mask()         */
/*     level-2: content validation, dmi_mask_isvalid_content()      */
/*     level-3: coords validation,  dmi_mask_isvalid_coord()        */
/*                                                                  */
/* Requirement in section 3.1.1, 3.2.1, 4.2, 4.2.2                  */
/*    3.1.1   two dimension                                         */
/*    3.2.1   physical or world coords                              */
/*    4.2     image file                                            */
/*    4.2.2   supported data types                                  */
/*                                                                  */
/********************************************************************/
dmBool dmi_mask_isvalid_mask( dmBlock* block )
{
  dmDescriptor *dd;
  dmDescriptor *dd_phys;
  dmDescriptor *dd_logi;
  dmDescriptor *dd_world;
  long naxes;
  long *urpix;

  if(!block)
    return dmFALSE;
  
  /* S4.2,  an image file */
  if( dmi_block_get_raw_type(block )  != dmIMAGE ){
    sprintf( dm_errbuf, "Mask block is not image.");
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return dmFALSE;
  }

  dd = dmImageGetDataDescriptor(block);  
  if(!dd){
    sprintf( dm_errbuf, "Cannot get Mask DM-descriptor." );
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return dmFALSE;
  }
 
    
  /* S3.1.1   2-d */
  naxes = dmi_descriptor_get_array_dimensions(dd, &urpix);
  if( naxes != 2){
    sprintf( dm_errbuf, "Mask image is not 2-D but %ld-D", naxes );
    dmu_err_set( DM_ERR_BAD_DIM, dm_errbuf );
    return dmFALSE;
  }

  if ( urpix )
    free(urpix);

  /* 3.2.1  phys and world ?             */
  if( dmArrayGetNoAxisGroups( dd ) != 0) {
    dd_phys = dmArrayGetAxisGroup( dd, 1 ); 
    if ( dd_phys != NULL ) {
      long ncoords = dmDescriptorGetNoCoords( dd_phys );
      if ( ncoords > 0 ) {
	// extract next coordinate level (World)    
	dd_world = dmDescriptorGetCoord( dd_phys );
	// extract Parent of the Physical axis set, Logical coordinate set
	dd_logi = dmCoordGetParent( dd_phys );
      }
    }
  } 
  if(  !dd_phys && !dd_logi && !dd_world)
  { 
    sprintf( dm_errbuf, "Mask image has no coords defined");
    dmu_err_set( DM_ERR_COORD, dm_errbuf );
    return dmFALSE;
  }
  else if( (!dd_phys && !dd_logi)  ||
           (!dd_phys && !dd_world) || 
           (!dd_world && !dd_logi) )
  {
    sprintf( dm_errbuf, "Unknown Mask image coords");
    dmu_err_set( DM_ERR_COORD, dm_errbuf );
    return dmFALSE;
  }
  
  /* 9.2.1.4 subspace   ? */
  dmSubspace *subspace = dmi_block_get_subspace( block ); /* TBD check */

  /* S4.2.2 supported datatypes */
  /* valid datatype - convertable to byte (numeric)   */ 
  dmDataType dtype = dmGetDataType(dd);
  switch(dtype){
  case dmDOUBLE:
  case dmFLOAT:
  case dmLONG:
  case dmULONG:
  case dmSHORT:
  case dmUSHORT:
  case dmBYTE:
    break;
  default:
    sprintf( dm_errbuf, "Not valid Mask image data type.");
    dmu_err_set(DM_ERR_BAD_TYPE, dm_errbuf );
    return dmFALSE;
  }

  return dmTRUE;
}
/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_isvalid_content                              */
/*                                                                  */
/* Description:                                                     */
/*   Extracts mask image data, validate it for all supported        */
/*   numerical datatypes. Convert the data to byte if not           */ 
/*   output the data array                                          */
/*                                                                  */
/*  Input parameters:                                               */
/*    dmBlock* block     Mask block                                 */
/*                                                                  */
/*  Output parameters:                                              */
/*    long* npix            Number of values returned.              */
/*    unsigned char** vbytes data pointer as 1D pixel byte type     */
/*                                                                  */
/*  Returns:                                                        */
/*     dmBool: true if everything is right otherwise false          */
/*                                                                  */
/*                                                                  */
/*   NOTE: This routine is only valid data, see the sections        */
/*   The full Mask block requirements checking should be done in    */
/*   all 3 levels                                                   */
//*                                                                 */
/*   validation levels:                                             */
/*     level-1: Mask validation,    dmi_mask_isvalid_mask()         */
/*     level-2: content validation, dmi_mask_isvalid_content()      */
/*     level-3: coords validation,  dmi_mask_isvalid_coord()        */
/*                                                                  */
/********************************************************************/
dmBool dmi_mask_isvalid_content(dmBlock* mskimg, long *npix, unsigned char **vbytes )
{
  void *values;
  dmDataType dtype;
  
  if( !mskimg ){
    sprintf( dm_errbuf, "Null Mask block pointer.");
    dmu_err_set(DM_ERR_GENERIC, dm_errbuf );
    return dmFAILURE;
  }

  /* --------------------------------------------- */
  /* get image data array and                      */
  /* convert the input mask data to byte type      */
  /* --------------------------------------------- */
  values = dmi_mask_get_image_array( mskimg, &dtype, npix) ;

  if(values == NULL )
    return dmFALSE;
  
  *vbytes=dmi_mask_convert_values( values, dtype, *npix);
  free(values);  
 
  if( *vbytes == NULL )
    return dmFALSE;
  
  return dmTRUE;  
}
/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_get_image_array                              */
/*                                                                  */
/* Description:                                                     */
/*   Extracts the bulk data array from the provided Mask block.     */
/*   Values are returned in thier 'native' data type.               */
/*                                                                  */
/*  Input parameters:                                               */
/*    dmBlock* block     Mask block                                 */
/*                                                                  */
/*  Output parameters:                                              */
/*    long* arraysz      Number of values returned.                 */
/*                                                                  */
/*  Returns:                                                        */
/*    void* values       image values as 1D pixel list.             */
/*                                                                  */
/*  MCD NOTE:                                                       */
/*    wouldn't it be better to return it?                           */
/*    dmi_mask_get_image_array( *block, *dtype, *nvals, *vals )    HHE:  return dtype now */
/*    this API may be confusing, most 'get' routines, you tell it   */
/*    which datatype you want rather than what are you getting..    */
/*                                                                  */
/********************************************************************/
void* dmi_mask_get_image_array(dmBlock *block, dmDataType *dtype, long *arraysz)
{  
  dmDescriptor *dd = dmImageGetDataDescriptor(block);
  long npix, naxes, dim=2;
  long status;
  long llpix[2]={1,1};
  long *urpix; 
  void *values;
  

  if(!block || !dd)
    return NULL;

  *dtype = dmGetDataType(dd);
  /* determine image size */
  naxes = dmi_descriptor_get_array_dimensions(dd, &urpix);
  if(naxes != dim || !urpix ) {
    sprintf( dm_errbuf, "Failed to get Mask image dimensions.");
    dmu_err_set(DM_ERR_GENERIC, dm_errbuf );
    return NULL;
  }

  npix  = urpix[0] * urpix[1];
  *arraysz=npix;
  switch( *dtype ) {
  case dmBYTE: 
    values = (unsigned char*)malloc(npix*sizeof(unsigned char));
    status = dmp_image_get_subarray(dd, llpix, urpix, dmBYTE, (void*)values);
    break;
  case dmSHORT: 
    values = (short*)malloc(npix*sizeof(short));
    status = dmp_image_get_subarray(dd, llpix, urpix, dmSHORT, (void*)values);
    break;
  case dmUSHORT: 
    values = (unsigned short*)malloc(npix*sizeof(unsigned short));
    status = dmp_image_get_subarray(dd, llpix, urpix, dmUSHORT, (void*)values);
    break;
  case dmLONG: 
    values = (long*)malloc(npix*sizeof(long));
    status = dmp_image_get_subarray(dd, llpix, urpix, dmLONG, (void*)values);
    break;
  case dmULONG: 
    values = (unsigned long*)malloc(npix*sizeof(unsigned long));
    status = dmp_image_get_subarray(dd, llpix, urpix, dmULONG, (void*)values);
   break;
  case dmFLOAT:
    values = (float*)malloc(npix*sizeof(float));
    status = dmp_image_get_subarray(dd, llpix, urpix, dmFLOAT, (void*)values);
    break;
  case dmDOUBLE:
    values = (double*)malloc(npix*sizeof(double));
    status = dmp_image_get_subarray(dd, llpix, urpix, dmDOUBLE, (void*)values);
    break;
  default:
    sprintf( dm_errbuf, "Not valid Mask image data type.");
    dmu_err_set(DM_ERR_BAD_TYPE, dm_errbuf );
    return NULL;
  }
  free(urpix);

  if(status == dmFAILURE){
    sprintf( dm_errbuf, "Failed to read Mask image data array.");
    dmu_err_set(DM_ERR_GENERIC, dm_errbuf );
    free(values);
    values=NULL;
  }

  return values;
}

/***************************************************************/
/*                                                             */
/* Function:  dmMaskWrite                                      */
/*                                                             */
/* Description:                                                */
/*   Create MASK block on provided dataset derived from        */
/*   the provided dmIMAGE block.                               */
/*                                                             */
/*                                                             */
/* Requirement in section 3.1.2-3, 4.2.2-4                     */
/*    3.1.2  type Byte                                         */
/*    3.1.3  value 1 or 0                                      */
/*    4.2.2  valid datatypes converted to byte                 */
/*    4.2.3  conversion 0==0, !0==1                            */
/*    4.2.4  IEEE NULL,NaN,Inf -> 0                            */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDataset*  ds     Dataset to get the MASK block         */
/*    char*       name   Name of the MASK block                */
/*    dmBlock*    img    Source 2D Image block                 */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*    dmErrCode   status dmSUCCESS/dmFAILURE                   */
/*                                                             */
/***************************************************************/
dmErrCode dmMaskWrite( dmDataset* ds, const char* name, dmBlock* img)
{
  dmDescriptor *dd;
  long npix;
  long status;
  unsigned char* vbytes; /* byte data type */;

  if(!img ){
    sprintf( dm_errbuf, "null Mask block pointer.");
    dmu_err_set(DM_ERR_GENERIC, dm_errbuf );
    return dmFAILURE;
  }
  if(!ds ){
    sprintf( dm_errbuf, "null dataset pointer.");
    dmu_err_set(DM_ERR_GENERIC, dm_errbuf );
    return dmFAILURE;
  }
  if( !name || strlen(name)==0 ){
    sprintf( dm_errbuf, "null or empty block name string.");
    dmu_err_set(DM_ERR_GENERIC, dm_errbuf );
    return dmFAILURE;
  }

  /* --------------------------------------------- */
  /* convert the input mask data to byte type      */
  /* --------------------------------------------- */
  vbytes=NULL;
  if( !dmi_mask_isvalid_content(img, &npix, &vbytes )){
    return dmFAILURE;
  }

  /* --------------------------------------------- */
  /* create new Mask block in the input dataset    */
  /* --------------------------------------------- */
  dmDescriptor *imData;
  long  dim, *axlens,*urpix;
  long  llpix[2]={1,1};
  dmBlock *mask;

  dd     = dmImageGetDataDescriptor(img);  
  dim    = dmi_descriptor_get_array_dimensions(dd, &axlens);
  mask   = dmDatasetCreateImage( ds, name, dmBYTE, axlens, dim);
  status = dmFAILURE;
  if (mask) {
    if ((imData = dmImageGetDataDescriptor(mask))){
      status = dmBlockCopyWCS(img, mask);

      if(status == dmFAILURE)
	sprintf( dm_errbuf, "Failed to copy Mask image WCS.");
      else {
	urpix=axlens;
	status=dmImageDataSetSubArray_ub(imData, llpix, urpix, vbytes);
	free(vbytes);  
     }
      if(status == dmFAILURE)
	sprintf( dm_errbuf, "Failed to convert Mask image to byte datatype.");
      else
	dmBlockFlush(mask);
    } else {
      sprintf( dm_errbuf, "Failed to get Mask byte-type descriptor.");
    }
  } else 
    sprintf( dm_errbuf, "Failed to create Mask byte-type block.");
  
  if(status == dmFAILURE)
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );   

  return status;
}

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_convert_values                               */
/*                                                                  */
/* Description:                                                     */
/*   Convert values from any compatible datatype to byte            */
/*   type, applying rules for MASK value interpretation.            */
/*                                                                  */
/* Requirement in section 3.1.2-3, 4.2.2-4                          */
/*   3.1.2  type Byte                                               */
/*   3.1.3  value 1 or 0                                            */
/*   4.2.2  valid datatypes converted to byte                       */
/*   4.2.3  conversion 0==0, !0==1                                  */
/*   4.2.4  IEEE NULL,NaN,Inf -> 0                                  */
/*                                                                  */
/*  Input parameters:                                               */
/*    void* values       Value array                                */
/*    dmDataType dtype   Datatype of input values.                  */
/*    long  nvals        Number of values in array                  */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    uc* bytearr        Byte array of Mask values                  */
/*                                                                  */
/********************************************************************/
void* dmi_mask_convert_values( void *values, dmDataType dtype, long nvals )
{
  long ii;
  unsigned char * vbytes;

  vbytes=(unsigned char*)malloc(nvals*sizeof(unsigned char));
  /* MCD NOTES:                                                   */
  /*   - not sure what ut_q* routines do.. not used within dm     */
  /*   - what is INF for non-floating point?                      */
  /*   - do not see NAN handled.                                  */
  /*   * no value conversion for dmBYTE type..       HHE: added   */
  /*      a byte array may have values 0 -> 255                   */
  /*                                                              */
  
  /* converting .....*/
  /* S4.2.2 - S4.2.4 */
  switch( dtype ) {      
  case dmBYTE:
    for (ii=0; ii < nvals; ii++){
      short v= ((uchar*)values)[ii]; /* don't have byte-type ut_ func avaiable */
      if( ut_qinf_s(v) || ut_qnull_s(v) || v==0)  /* S4.2.4 */
	vbytes[ii]=0;
      else
	vbytes[ii]= ((uchar*)values)[ii];
    }
    break;
  case dmSHORT:
    for (ii=0; ii < nvals; ii++){
      short v= ((short*)values)[ii];
      if( ut_qinf_s(v) || ut_qnull_s(v) || v==0)  /* S4.2.4 */
	vbytes[ii]=0;
      else 
	vbytes[ii]=1; 
    }
    break;
  case dmUSHORT:
    for (ii=0; ii < nvals; ii++){
      short v= ((unsigned short*)values)[ii];
      if( ut_qinf_s(v) || ut_qnull_s(v) || v==0)  /* S4.2.4 */
	vbytes[ii]=0;
      else 
	vbytes[ii]=1; 
    }
    break;
  case dmLONG:
    for (ii=0; ii < nvals; ii++){
      long v= ((long*)values)[ii];
      if( ut_qinf_i(v) || ut_qnull_i(v) || v==0 )  /* S4.2.4 */
	vbytes[ii]=0;
      else 
	vbytes[ii]=1; 
    }
    break;
  case dmULONG:
    for (ii=0; ii < nvals; ii++){
      long v= ((unsigned long*)values)[ii];
      if( ut_qinf_i(v) || ut_qnull_i(v) || v==0 )  /* S4.2.4 */
	vbytes[ii]=0;
      else 
	vbytes[ii]=1; 
    }
    break;
  case dmFLOAT:
    for (ii=0; ii < nvals; ii++){
      float v = ((float*)values)[ii];
      if( ut_qinf_r(v) || ut_qnull_r(v) )  /* S4.2.4 */
	vbytes[ii]=0;
      else if( fabs(v) >= FLT_EPSILON )
	vbytes[ii]=1; 
      else
	vbytes[ii]=0; 
    }
    break;
  case dmDOUBLE:
    for (ii=0; ii < nvals; ii++){
      double v= ((double*)values)[ii];
      if( ut_qinf_d(v) || ut_qnull_d(v) )  /* S4.2.4 */
	vbytes[ii]=0;
      else if( fabs(v) >= DBL_EPSILON )
	vbytes[ii]=1; 
      else
	vbytes[ii]=0; 
    }
    break;
  default:
    sprintf( dm_errbuf, "Not valid Mask image data type.");
    dmu_err_set(DM_ERR_BAD_TYPE, dm_errbuf );
    free(vbytes);
    return NULL;
  }; /* end of convertion */


  return vbytes;
}
 
/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_read_mask                                    */
/*                                                                  */
/* Description:                                                     */
/*   Validate and create Mask region/shape                          */
/*                                                                  */
/*  Input parameters:                                               */
/*    dmBlock  *block    Image block for use as Mask                */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    regRegion *region                                             */
/*                                                                  */
/********************************************************************/
regRegion* dmi_mask_read_mask( dmBlock* block )
{ 
  regRegion *myRegion = NULL;
  regShape *myShape;

  /* validate the block is appropraite Mask image block*/
  if( !dmi_mask_isvalid_mask(block) )
    return NULL;

  /* create MASK region shape attributes */ 
  myShape = dmi_mask_create(block);
  if(!myShape)
    return NULL;
  
  myRegion  = regCreateRegion(NULL,NULL);
  regAddShape(myRegion,1,myShape);

  return myRegion; 
}   

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_set_includeflag                              */
/*                                                                  */
/* Description:                                                     */
/*   Set include flag attribute of Mask Shape                       */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape* shape    Mask Shape                                 */
/*    int       fincl    Include flag                               */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    none                                                          */
/*                                                                  */
/********************************************************************/
void dmi_mask_set_includeflag(regShape *shape, int fincl)
{
  if ( shape != NULL )
    shape->include = fincl;
}

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_free                                         */
/*                                                                  */
/* Description:                                                     */
/*   Bridge function for regUSER shape.                             */
/*   Free memory allocated to members of struct regUSERSHAPE        */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape shape     Mask region shape                          */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    none                                                          */
/*                                                                  */
/********************************************************************/
void dmi_mask_free(regShape *shape)
{
  if(!shape )
    return;

  if ( shape->xpos ) free( shape->xpos );
  if ( shape->ypos ) free( shape->ypos );


  if(!shape->user )
    return;

  if( (shape->user)->name ){
    free((shape->user)->name);
    (shape->user)->name=NULL;
  } 

  /* Note:                                                              */
  /* the user->spec MASK block will be closed by dmi_mask_free_blocks() */
  if( (shape->user)->spec)
    (shape->user)->spec=NULL;

  free((shape->user));
  shape->user=NULL;

}
/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_isvalid_shape                                */
/*                                                                  */
/* Description:                                                     */
/*   validates attributes of Mask "shape" , including               */
/*     shape struct pointer, shape type, User-struct pointer,       */
/*     Mask name, and/or                                            */
/*     mask dm-block pointer                                        */
/*                                                                  */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape *shape     Mask region shape                         */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    dmBool    True if all are valid otherwise False               */
/*                                                                  */
/********************************************************************/
dmBool dmi_mask_isvalid_shape(regShape *shape)
{

  if( !shape ) {
    sprintf(dm_errbuf, "Null shape pointer.");
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return dmFALSE;
  } 

  if ( shape->type != regMASK ){
    sprintf(dm_errbuf, "Not Mask shape type.");
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return dmFALSE;
  } 
  
  if ( !shape->user ) {
    sprintf(dm_errbuf, "Null Mask pointer.");
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return dmFALSE;
  }  

  if ( !shape->user->spec ) {
    sprintf(dm_errbuf, "Null Mask block pointer.");
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return dmFALSE;
  }
  if( !shape->user->name || strlen(shape->user->name)==0 ) {
    sprintf(dm_errbuf, "Null or empty Mask name string."); 
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return dmFALSE;
  } 

  return dmTRUE;
}
/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_toString                                     */
/*                                                                  */
/* Description:                                                     */
/*   Bridge function for regUSER shape.                             */
/*   Generates string representation of the shape.                  */
/*     Spec: 10.1.1.1                                               */
/*                                                                  */
/* Note:                                                            */
/*   User must free the returned memory.                            */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape *shape    Mask region shape                          */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    char*  buff        String rep of Mask shape.                  */
/*                                                                  */
/********************************************************************/
char* dmi_mask_toString(regShape *shape)
{
  dmBlock *blk;
  char blkname[TEXT_SIZE];
  int  buff_sz;
  char *buff = NULL;

  if( !dmi_mask_isvalid_shape(shape) )
    return NULL;
 
  blk=shape->user->spec;
  dmBlockGetName( blk, blkname, TEXT_SIZE );
  if( strlen(blkname)==0)
    strcpy(blkname, "UNKNOWN");

  /* allocate for format "<shapename>(<blockname>)\0" */    
  buff_sz = strlen(shape->user->name) + 1 + strlen(blkname)+ 1 + 1;

  buff=calloc(buff_sz, sizeof(char));
  sprintf(buff, "%s(%s)", shape->user->name, blkname);

  return buff;
}

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_area                                         */
/*                                                                  */
/* Description:                                                     */
/*   Bridge function for regUSER shape.                             */
/*   Determines area of Mask shape.                                 */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape shape     Mask region shape                          */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    double area        Area of Mask shape                         */
/*                                                                  */
/********************************************************************/
double dmi_mask_area( regShape *shape )
{
  dmBlock *blk;
  dmDataType dtype;
  dmDescriptor *phys_dd ;
  long dim=2;
  long npix, ii, count;
  double area=0.0;
  void *values;
  unsigned char *vbytes;

  if( !dmi_mask_isvalid_shape(shape) )
    return area;

  blk=shape->user->spec;
  /* extract pixel data array */
  values = dmi_mask_get_image_array(blk, &dtype, &npix);
  if(values == NULL)
    return area;


  /* interpret values as MASK array */
  vbytes = dmi_mask_convert_values( values, dtype, npix);
  free(values);
  if(vbytes == NULL)
    return area;

  /* MCD NOTE:                                                         */
  /*   - number of 'on' pixels is a good cache candidate..             */
  /*     it will not change unless we are combining Masks (I think)    */

  /* count 'on' pixels (value==1) */
  count = 0; 
  for (ii=0; ii < npix; ii++)
    if ( vbytes[ii] == 1 ) count++; 
  free(vbytes);

  area = (double)count;  
  if( shape->user->flag_coord ==  DM_FLAG_LOGICAL )
    ;
  else if ( shape->user->flag_coord == DM_FLAG_PHYS )
  {
    phys_dd = dmArrayGetAxisGroup( dmImageGetDataDescriptor(blk), 1 ); 
    double crpix[2]; /* =dim */
    double crval[2];
    double cdelt[2];

    dmCoordGetTransform_d( phys_dd, crpix, crval, cdelt, dim );
    /* scale the counts by pixel size */
    area *= cdelt[0] * cdelt[1];    
  }
  else
  {
    sprintf(dm_errbuf, "Unsupported coordinate flag (%d).", shape->user->flag_coord );
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    area=0.0;
  } 

  return area;
}

/********************************************************************/
/*                                                                  */
/* Function: dmi_mask_extent                                        */
/*                                                                  */
/* Description:                                                     */
/*   Bridge function for regUSER shape.                             */
/*   Determines the bounds of the Mask shape                        */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape shape     Mask region shape                          */
/*                                                                  */
/*  Output parameters:                                              */
/*    double*  xpos      X-axis bounds [min,max]                    */
/*    double*  ypos      Y-axis bounds [min,max]                    */
/*                                                                  */
/*  Returns:                                                        */
/*    int        ok      Error flag; 1-success, 0-problem           */
/*                                                                  */
/********************************************************************/
int dmi_mask_extent( regShape *shape, double *xpos, double *ypos )
{
  long ii, dim; 
  int ok=1;
  int nok=0;
  dmBlock *blk;
  dmDescriptor *dd;
 
  if( !dmi_mask_isvalid_shape(shape) )
    return nok;

  blk = shape->user->spec;
  dd  = dmImageGetDataDescriptor(blk);  
  dim = dmi_descriptor_get_array_dim(dd);

  dmDescriptor *coord_dd;
  dmDescriptor *cpt_dd;
  double tlmin, tlmax;
  dmDataType dtype = dmDOUBLE;
  long cptno;
 
  coord_dd = dmArrayGetAxisGroup( dd, 1 ); /* physical coordinates */
  if ( shape->user->flag_coord == DM_FLAG_LOGICAL )
    coord_dd = dmi_coord_get_parent( coord_dd );
    
  if(!coord_dd){
    sprintf(dm_errbuf, "Failed to get Mask's coords.");
    dmu_err_set( DM_ERR_COORD, dm_errbuf );
    return nok;
  }

 for ( cptno = 1; cptno <= dim; cptno++ ) 
 {
   cpt_dd = dmi_descriptor_get_cpt(coord_dd, cptno ); 
   /* the above call is equivalent to call 
      axis1_dd=dmi_descriptor_get_physical_axis(dd,cptno);  
   */

    tlmin=0.0;
    tlmax=0.0;
    static int status;
    if(cpt_dd) {
      status=dmi_coord_get_range( cpt_dd,    /* coordinate descriptor         */
				  0,         /* cpt no, 0 if dd is vector cpt */
				  dtype,     /* MUST be DT_REAL8              */
				  &tlmin,    /* minima                        */
				  &tlmax );  /* maxima                        */
      if(status == dmFAILURE){
	sprintf(dm_errbuf, "Failed to get Mask's coords range.");
	dmu_err_set( DM_ERR_RANGE, dm_errbuf );
    	return  nok;
      }
 
      ii = cptno-1;
      if(cptno==1){     
	xpos[ii]   = tlmin;
	xpos[ii+1] = tlmax;
      } else {
	ypos[ii-1] = tlmin;
	ypos[ii]   = tlmax;  
      }
    }
 }

 return ok;
}

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_isEqual                                      */
/*                                                                  */
/* Description:                                                     */
/*   Bridge function for regUSER shape.                             */
/*   Checks content of Mask Shapes for equality.                    */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape shape1    Mask region shape                          */
/*    regShape shape2    Mask region shape                          */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    int result         dmTRUE(1)  = Masks are equal               */
/*                       dmFALSE(0) = Masks are not equal           */
/*                                                                  */
/********************************************************************/
int dmi_mask_isEqual(regShape *shape1, regShape *shape2)
{
  long ii; 
  
  if( !dmi_mask_isvalid_shape(shape1) || !dmi_mask_isvalid_shape(shape2) )
    return  dmFALSE;

  if(shape1->type != shape2->type           ||
     shape1->include != shape2->include     ||
     shape1->nPoints != shape2->nPoints     ||
     shape1->world_coord != shape2->world_coord ||
     shape1->world_size != shape2->world_size ||
     shape1->user->flag_coord != shape2->user->flag_coord ||
     !cs_eq(shape1->user->name,shape2->user->name) 
     )
    return dmFALSE;

  if(shape1->nPoints !=0 ){
    for(ii=0;ii< shape1->nPoints;ii++){
      if( !shape1->xpos || !shape2->xpos || !shape1->ypos|| !shape2->ypos){
	return dmFALSE;
      }
      if(shape1->xpos[ii] != shape2->xpos[ii] ||
	 shape1->ypos[ii] != shape2->ypos[ii])
	return dmFALSE;
    }     
  }

  dmBlock *blk;
  dmDescriptor *dd1, *dd2;
  long llpix1[2]={1,1};
  long *urpix1=NULL;
  long llpix2[2]={1,1};
  long *urpix2=NULL;
  long naxes1, naxes2;
  char name1[dmMAXSTR-1];
  char name2[dmMAXSTR-1];
  long npix;

  if( !(shape1->user->spec==shape2->user->spec))
    return dmFALSE;
  dmDatasetGetName( dmBlockGetDataset(shape1->user->spec),name1,dmMAXSTR);
  dmDatasetGetName( dmBlockGetDataset(shape2->user->spec),name2,dmMAXSTR);
  if( !cs_eq(name1,name2) ) 
    return dmFALSE;
 
  blk = shape1->user->spec;
  dd1  = dmImageGetDataDescriptor(blk);  
  dmBlockGetName( blk, name1, dmMAXSTR );
  blk = shape2->user->spec;
  dd2  = dmImageGetDataDescriptor(blk);  
  dmBlockGetName( blk, name2, dmMAXSTR );

  if( !cs_eq(name1, name2) )
    return dmFALSE;

  /* check  image size */
  naxes1 = dmi_descriptor_get_array_dimensions(dd1, &urpix1);
  naxes2 = dmi_descriptor_get_array_dimensions(dd2, &urpix2);

  if(urpix1[0] != urpix1[0] ||
     urpix1[1] != urpix1[1] ||
     naxes1 != naxes2 )
    return dmFALSE;


  npix=urpix1[0]*urpix1[1];
  short *val1 = (short*)malloc(npix*sizeof(short));
  int stat1 = dmp_image_get_subarray(dd1, llpix1, urpix1, dmSHORT, (void*)val1);
  short *val2 = (short*)malloc(npix*sizeof(short));
  int stat2 = dmp_image_get_subarray(dd2, llpix2, urpix2, dmSHORT, (void*)val2);
  if(stat1 == dmFAILURE || stat2 == dmFAILURE ){
    sprintf( dm_errbuf, "Failed to read Mask image array.");
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    free(val1);
    free(val2);
    return dmFALSE;
  }

  stat1=dmTRUE;
  for(ii=0;ii<npix;ii++){
    if(val1[ii] != val2[ii]){
      stat1=dmFALSE;
      ii=npix;
    }
  }
  
  free(val1);
  free(val2);

  return stat1;
}

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_copy                                         */
/*                                                                  */
/* Description:                                                     */
/*   Bridge function for regUSER shape.                             */
/*   Make a copy of the provide Mask Shape                          */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape shape     Mask region shape to copy                  */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    regShape shape     resulting Mask shape                       */
/*                                                                  */
/********************************************************************/
regShape* dmi_mask_copy(regShape *inshp)
{
  regShape *myShape =NULL;
  long ii;

  if( !dmi_mask_isvalid_shape(inshp) )
    return NULL;
 

  /* MCD NOTE:  Should we check the shape is a regMASK type? */

  myShape = ( regShape *) calloc ( 1, sizeof( regShape ) );
  myShape->user = (struct regUSERSHAPE *) calloc ( 1, sizeof(struct regUSERSHAPE) );

  myShape->type        = inshp->type;
  myShape->include     = inshp->include;
  myShape->nPoints     = inshp->nPoints;
  myShape->world_coord = inshp->world_coord; 
  myShape->world_size  = inshp->world_size;
  
  if(inshp->xpos && inshp->ypos && myShape->nPoints )
  {
    myShape->xpos=(double *)calloc( myShape->nPoints, sizeof(double));
    myShape->ypos=(double *)calloc( myShape->nPoints, sizeof(double));

    for(ii=0;ii<inshp->nPoints;ii++){
      myShape->xpos[ii]=inshp->xpos[ii];
      myShape->ypos[ii]=inshp->ypos[ii];
    }
  }
  
  myShape->user->flag_coord = inshp->user->flag_coord; 
  myShape->user->name  = calloc( strlen(inshp->user->name)+1, sizeof(char) );
  strcpy(myShape->user->name, inshp->user->name);
  myShape->inside          = inshp->inside;
  myShape->user->copy      = inshp->user->copy;
  myShape->user->isEqual   = inshp->user->isEqual;
  myShape->user->toString  = inshp->user->toString;
  myShape->user->calcExtent= inshp->user->calcExtent;
  myShape->user->calcArea  = inshp->user->calcArea;
  myShape->user->free      = inshp->user->free;
  
  /* Each MASK shape shares the block pointer  */
  myShape->user->spec = inshp->user->spec;
  
  return myShape; 
}

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_create                                       */
/*                                                                  */
/* Description:                                                     */
/*   Create Mask region shape from Mask image dmBlock pointer       */
/*                                                                  */
/*  Input parameters:                                               */
/*    void* ptr          Mask image (dmBlock pointer)               */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    regShape shape     resulting Mask region shape                */
/*                        NULL if problem                           */
/*                                                                  */
/********************************************************************/
regShape* dmi_mask_create(void *ptr)
{
  regShape *myShape =NULL;
  dmBlock *block= ptr;

  if(!block)
    return NULL;

  myShape       =           ( regShape *) calloc ( 1, sizeof( regShape ) );
  myShape->user = (struct regUSERSHAPE *) calloc ( 1, sizeof(struct regUSERSHAPE) );
 
  myShape->type       = regMASK;
  myShape->include    = regInclude;       /* updated somewhere else */
  myShape->user->spec = block;
 
  /* xpos and ypos cache mask extent.. values assigned later      */
  /*   right now, we do not know which coordinate system to use.  */
  myShape->nPoints   = 2;
  myShape->xpos      = calloc( 2, sizeof(double) );
  myShape->ypos      = calloc( 2, sizeof(double) );
  myShape->world_coord=0; 
  myShape->world_size =0;  /* radius scale flag, no for non-world coords */

  myShape->user->name = dmMaskName();

  /* Assign Bridge funcions */
  myShape->inside          = dmi_mask_inside;
  myShape->user->copy      = dmi_mask_copy;
  myShape->user->isEqual   = dmi_mask_isEqual;
  myShape->user->toString  = dmi_mask_toString;
  myShape->user->calcExtent= dmi_mask_extent;
  myShape->user->calcArea  = dmi_mask_area;
  myShape->user->free      = dmi_mask_free;

  return myShape;
}

/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_get_coordflag                                */
/*                                                                  */
/* Description:                                                     */
/*   Get coords flag if blocks are                                  */
/*   validated both coords matching                                 */
/*   validated both coord transformations matching                  */
/*                                                                  */
/*  Input parameters:                                               */
/*    dmBlock*          Blcok-1 pointer (mask or filter )           */
/*    dmBlock*          Blcok-1 pointer (filter or mask )           */
/*    char* cptname     Name of descriptor being filtered           */
/*                                                                  */
/*                                                                  */
/*  Returns:                                                        */
/*    int             Enum indicating coordinate sysy or -1 if error*/
/*                                                                  */
/********************************************************************/
int dmi_mask_get_coordflag(dmBlock *blk_1, dmBlock *blk_2, char *cptname)
{
  dmDescriptor *dd_1;
  dmDescriptor *dd_2;
  int not_ok=-1;
  int cflag = not_ok;

  if(!blk_1 || !blk_2 ){
    sprintf(dm_errbuf, "Null block pointer.");
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return not_ok;
  } 

 
  dd_1 = dmi_block_search_raw_col_identifier( blk_1, cptname );
  dd_2 = dmi_block_search_raw_col_identifier( blk_2, cptname );
  
  if ( !dd_1 || !dd_2 ) {
    sprintf( dm_errbuf, "Could not find identifier %s.", cptname );
    dmu_err_set( DM_ERR_PARSE, dm_errbuf );
    return  cflag;
  }

  if ( (dd_1->flag & DM_FLAG_LOGICAL) &&  (dd_2->flag & DM_FLAG_LOGICAL) ) {
    cflag = DM_FLAG_LOGICAL;
    return cflag;
  }
  else if ( dd_1->flag & DM_FLAG_PHYS && dd_2->flag & DM_FLAG_PHYS) {
    cflag = DM_FLAG_PHYS;
    /* continue to validate coords transformation parameters bewlo */
  } else {
    sprintf( dm_errbuf, "Could not find coords for identifier, '%s'.", cptname );
    dmu_err_set( DM_ERR_PARSE, dm_errbuf );
    return cflag;
    /* not supported WORLD coords at thei time */
  }

  if(blk_1==blk_2) /* do nothing  */
    return cflag;
 
  /* ------------------------------------------------------*- */
  /* validate coords transformation paramters, if not LOGICAL */
  /* between Mask and filter descriptors                      */
  /*                                                          */
  /* Linear Coords transform from x to y:                     */ 
  /* y = a + b * (x-c)                                        */
  /* a: offset  b: sacle   c: reference                       */
  /* ------------------------------------------------------*- */    
  double crpix[2]   = { 0.0, 0.0 };   /* reference */
  double crval[2]   = { 0.0, 0.0 };   /* offset    */
  double cdelt[2]   = { 0.0, 0.0 };   /* scale     */
  double crpix_2[2] = { 0.0, 0.0 };
  double crval_2[2] = { 0.0, 0.0 };
  double cdelt_2[2] = { 0.0, 0.0 };
  int dim = 2;
  int ii;
  float diff_val, diff_dlt;
  float torl=1.e-5;   //FLT_EPSILON    TBdecided
 
  dd_1 = dmi_descriptor_get_coord_no( dd_1, 1 );
  dd_2 = dmi_descriptor_get_coord_no( dd_2, 1 );

  /* get coords transformation */
  dmi_coord_get_transform_d( dd_1, crpix,   crval,   cdelt,   dim );
  dmi_coord_get_transform_d( dd_2, crpix_2, crval_2, cdelt_2, dim );
  
  /* validate the same coords transformation params between Mask and filter descriptors */
  /* but ignore checking the reference, crpix  */
  for( ii = 0; ii < dim; ii++ ) {
    diff_val = fabs(crval[ii] - crval_2[ii]) ;
    diff_dlt = fabs(cdelt[ii] - cdelt_2[ii]);
    if( diff_val >= torl || diff_dlt >= FLT_EPSILON  ) {
      if( diff_val >= torl )
	sprintf( dm_errbuf, "Unmatched Mask and filter coords transform (crval=%g) at identifier, '%s'.\n", diff_val,cptname);
      else
	sprintf( dm_errbuf, "Unmatched Mask and filter coords transform at identifier, '%s'.\n", cptname);
      dmu_err_set( DM_ERR_COORD, dm_errbuf );
      return not_ok;
    }
  }

  return cflag;
}
/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_isvalid_coord                                */
/*                                                                  */
/* Description:                                                     */
/*   Validate Mask DM-descriptor name matches filter's, cptname     */
/*   Validate Mask coords matching to event-list coords,            */
/*   Validated Mask DM-descriptor coords transformation matching    */
/*             filter DM-descriptor coords transformation           */
/*                                                                  */
/*   Assign coordinate system flag on shape.                        */
/*     ( for logical, physical, or world coord system )             */
/*                                                                  */
/*   populate xy-bounds for the shape                               */
/*                                                                  */
/*  Input parameters:                                               */
/*    dmBlock* msk_blk   Mask dm-block pointer                      */
/*    dmBlock* flt_blk   Filter block pointer                       */
/*    char *cptname      Filter description name string             */
/*                                                                  */
/*                                                                  */
/*  Output parameters:                                              */
/*    int * cflag      Enumnator indicating the coordinate sys      */
/*                                                                  */
/*  Returns:                                                        */
/*    dmBool            TRUE if coords matching or FALSE if not     */
/*                                                                  */
/*  NOTE:                                                           */
/*   validation levels:                                             */
/*     level-1: Mask validation,    dmi_mask_isvalid_mask()         */
/*     level-2: content validation, dmi_mask_isvalid_coord()        */
/*     level-3: coords validation,  dmi_mask_isvalid_content()      */
/*                                                                  */
/* Requirement in section 4.3.3                                     */
/*   4.3.3  projection type and (not reference point)               */
/*                                                                  */
/********************************************************************/
dmBool dmi_mask_isvalid_coord(dmBlock *msk_blk, dmBlock *flt_blk, char *cptname, int *cflag)
{
 
  if(!msk_blk ){
    sprintf(dm_errbuf, "Null Mask block pointer.");
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return dmFALSE;
  } 

  if(!flt_blk ){
    sprintf(dm_errbuf, "Null filter block pointer.");
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return dmFALSE;
  } 

  /* ------------------------------------------------------*- */
  /* retrieve descriptors and coords of Mask and filter       */
  /* validate the same coords of Mask and filter descriptor   */
  /* validate coords transformation paramters, if not LOGICAL */
  /* realize Requirement section 4.3.3                        */
  /* ------------------------------------------------------*- */    
  *cflag = dmi_mask_get_coordflag(msk_blk, flt_blk, cptname);
  if(*cflag == -1)
    return  dmFALSE;
  else
    return dmTRUE;
}

/********************************************************************/
/*                                                                  */
/* Function:   dmi_mask_get_block                                   */
/*                                                                  */
/* Description:                                                     */
/*   return a Mask block from region shape structure                */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape*  shape   Mask Region shape struct                   */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    dmBlock*  mask       Block from the  MASK shape.              */
/*                         (NULL if invalid mask shape)             */
/*                                                                  */
/********************************************************************/
dmBlock *dmi_mask_get_block(regShape *myShape)
{
  if( dmi_mask_isvalid_shape(myShape) )
    return myShape->user->spec;
  else 
    return NULL;
}
/********************************************************************/
/*                                                                  */
/* Function:  dmi_mask_set_coord                                    */
/*                                                                  */
/* Description:                                                     */
/*   Validate Mask DM-descriptor name matches filter's, cptname     */
/*   Validate Mask coords matching to event-list coords,            */
/*   Validated Mask DM-descriptor coords transformation matching    */
/*             filter DM-descriptor coords transformation           */
/*                                                                  */
/*   Assign coordinate system flag on shape.                        */
/*     ( for logical, physical, or world coord system )             */
/*   and                                                            */
/*   populate xy-bounds for the shape                               */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape shape     Mask region shape (updated at return)      */
/*    dmBlock*           Filter blcok pointer                       */
/*    char* cptname      Name of filter descriptor                  */
/*                                                                  */
/*  Output parameters:                                              */
/*    regShape shape     Mask region shape is updated               */
/*                                                                  */
/*  Returns:                                                        */
/*    dmBool            TRUE if coords matching or FALSE if not     */
/*                                                                  */
/*  NOTE:                                                           */
/*   validation levels:                                             */
/*     level-1: Mask validation,    dmi_mask_isvalid_mask()         */
/*     level-2: coords validation,  dmi_mask_isvalid_content()      */
/*     level-3: content validation, dmi_mask_isvalid_coord()        */
/*                                                                  */
/* Requirement in section 4.3.3                                     */
/*   4.3.3  projection type and (not reference point)               */
/*                                                                  */
/********************************************************************/
dmErrCode dmi_mask_set_coord(regShape *myShape, dmBlock *flt_block, char *cptname)
{
  int cflag;
  dmBlock *msk_block;

  if( !dmi_mask_isvalid_shape(myShape) )
    return dmFAILURE;

  if(!flt_block ){
    sprintf(dm_errbuf, "Null filter block pointer.");
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return dmFAILURE;
  } 

  /* ------------------------------------------------------*- */
  /* retrieve descriptors and coords of Mask and filter       */
  /* validate the same coords of Mask and filter descriptor   */
  /* validate coords transformation paramters, if not LOGICAL */
  /* realize Requirement section 4.3.3                        */
  /* ------------------------------------------------------*- */    
  msk_block=dmi_mask_get_block(myShape);
  if( dmi_mask_isvalid_coord(msk_block, flt_block, cptname, &cflag) == dmFALSE )
    return  dmFAILURE;

  /* ------------------------------------------------------- */
  /* assign the Mask shape coords flag */
  /* ------------------------------------------------------- */  
  myShape->user->flag_coord=cflag;

  /* ------------------------------------------------------- */
  /* populate Mask bounds                                    */
  /* ------------------------------------------------------- */
  if( myShape->xpos != NULL)
    free(myShape->xpos);
  myShape->xpos = (double *)calloc( myShape->nPoints, sizeof(double));
  if( myShape->ypos != NULL )
    free( myShape->ypos );
  myShape->ypos = (double *)calloc( myShape->nPoints, sizeof(double));

  if( !dmi_mask_extent(myShape,  myShape->xpos, myShape->ypos) ){
    dmi_mask_free(myShape);
    return  dmFAILURE;
  }
  
  return dmSUCCESS;
}

/********************************************************************/
/*                                                                  */
/* Function:    dmi_mask_inside                                     */
/*                                                                  */
/* Description:                                                     */
/*   Bridge function for regUSER shape.                             */
/*   Determine if the input position is inside Mask shape "bounds"  */
/*   and is in an 'on' area.                                        */
/*                                                                  */
/*   Positions are assumed to be in coordinate system of Mask.      */
/*                                                                  */
/*  Input parameters:                                               */
/*    regShape shape     Mask region shape                          */
/*    double   xpos      X-axis position coordinate                 */
/*    double   ypos      Y-axis position coordinate                 */
/*                                                                  */
/*  Output parameters:                                              */
/*    none                                                          */
/*                                                                  */
/*  Returns:                                                        */
/*    int result         dmTRUE(1) = 'inside' ; dmFALSE(0) = not    */
/*                                                                  */
/********************************************************************/
int dmi_mask_inside(regShape* shape, double xpos, double ypos)
{
  int retval=0;
  dmBlock      *block;
  dmDescriptor *dd;
  dmDescriptor *dd_phys;
  double inpos[2];
  long dim, pixno[2];  /* byte datatype in image coords */
  long *axlens;
  int stat1, stat2;

  if( !dmi_mask_isvalid_shape(shape) )
    return 0;


  block = shape->user->spec;
  dd    = dmImageGetDataDescriptor(block);
  dim   = dmi_descriptor_get_array_dimensions( dd, &axlens );

  stat1=stat2=dmSUCCESS;
  inpos[0]=xpos;
  inpos[1]=ypos;
  dd_phys=dmArrayGetAxisGroup(dd, 1 ); 
  if(!dd_phys)
    return 0;

  if( shape->user->flag_coord == DM_FLAG_PHYS  ) {
    double *dpix=(double *)malloc(dim * sizeof(double) );
    stat1=dmi_coord_invert_d( dd_phys, &inpos[0], dpix);         /* physical to logical */   
    pixno[0]=(long)dpix[0];
    pixno[1]=(long)dpix[1];  
    free(dpix);  
  } else  if( shape->user->flag_coord == DM_FLAG_LOGICAL ) {
    pixno[0]=(long)inpos[0];
    pixno[1]=(long)inpos[1]; 
  } else { /* may not happen */
    double *dpix=(double *)malloc(dim * sizeof(double) );
    double phypos[2];
    stat1=dmi_coord_invert_d( dd, inpos, phypos);           /* world to physical */ 
    stat2=dmi_coord_invert_d( dd_phys, &phypos[0], dpix);   /* physical to logical */   
    pixno[0]=(long)dpix[0];
    pixno[1]=(long)dpix[1];  
    free(dpix);
  }

  if(stat1==dmFAILURE || stat2==dmFAILURE ){
    sprintf(dm_errbuf, "Failed to invert descriptor coord to Mask logical coords.");
    dmu_err_set( DM_ERR_COORD, dm_errbuf );
    return 0;
  }

  /* Simple bounds check pixel axis indexing starts at 1              */
  /* MCD NOTE:                                                        */
  /*  dmImageDataGetPixel_<t> does a range check a few layers down,   */
  /*  returning a pixel value of 0.  We add this check here to avoid  */
  /*  traveling the layers, and any confusion over the pixel value    */
  /*  since '0' is a valid value.  Alternatively, we could go down    */
  /*  a layer to dmImageDataGetSubArray_<t>/dmp_image_get_subarray    */
  /*  which returns an error status.                                  */
  /*                                                                  */
  /*  This check does not seem to significantly effect performance..  */
  /*  presumably because range filtering should eliminate most cases. */
  if ( (pixno[0] < 1) || (pixno[0] > axlens[0]) || 
       (pixno[1] < 1) || (pixno[1] > axlens[1]) )
  {
    retval = 0;
  }
  else
  {
    /* get image pixel data at the pixel-no */
    dmDataType dtype = dmGetDataType(dd);
    unsigned char* vbytes;
    void *values;
    long npix=1;
    
    switch(dtype){
    case dmDOUBLE:
      values  = (double*)malloc(npix*sizeof(double));
      ((double*)values)[0] = dmImageDataGetPixel_d( dd, pixno, dim);
      break;
    case dmFLOAT:
      values  = (float*)malloc(npix*sizeof(float));
      ((float*)values)[0] = dmImageDataGetPixel_f( dd, pixno, dim);
     break;
    case dmLONG:
      values  = (long*)malloc(npix*sizeof(long));
      ((long*)values)[0] = dmImageDataGetPixel_l( dd, pixno, dim);
      break;
    case dmULONG:
      values  = (unsigned long*)malloc(npix*sizeof(unsigned long));
      ((unsigned long*)values)[0] = dmImageDataGetPixel_ul( dd, pixno, dim);
      break;
    case dmSHORT:
      values  = (short*)malloc(npix*sizeof(short));
      ((short*)values)[0] = dmImageDataGetPixel_us( dd, pixno, dim);
      break;
    case dmUSHORT:
      values  = (unsigned short*)malloc(npix*sizeof(unsigned short));
      ((unsigned short*)values)[0] = dmImageDataGetPixel_us( dd, pixno, dim);
      break;
    case dmBYTE:
      vbytes  = (unsigned char*)malloc(npix*sizeof(unsigned char));
      *vbytes = dmImageDataGetPixel_ub(dd, pixno, dim);
      break;
    default:  /* UNKNOWN */
      sprintf( dm_errbuf, "Not valid Mask image data type.");
      dmu_err_set(DM_ERR_BAD_TYPE, dm_errbuf );
      return 0;
    }
    
    if( dtype != dmBYTE) {
      vbytes = dmi_mask_convert_values( values, dtype, npix );
      free(values);
    }
    if(vbytes==NULL){
      sprintf( dm_errbuf, "Failed to get Mask image byte type data.");
      dmu_err_set(DM_ERR_BAD_TYPE, dm_errbuf );
      return 0;
    }
    
    /* return pixel byte, 1|0 */
    retval = (*vbytes==1 ? 1 : 0) ;
    free(vbytes);
  }
  free(axlens);

  if ( shape->include == regInclude ){
    return( retval );
  }

  return( 1 - retval ); 
}
/********************************************************************/
/*                                                                  */
/* Function:  dmMaskIsValidMask                                     */
/*                                                                  */
/* Description:                                                     */
/*   Validate Mask usability                                        */
/*   Validate Mask data contents appropriation                      */
/*   Validated Mask coords matching filter/target's                 */
/*                                                                  */
/*                                                                  */
/*  Input parameters:                                               */
/*    dmBlock* msk_blk   Mask dm-block pointer                      */
/*    dmBlock* flt_blk   Filter block pointer                       */
/*    char *cptname      Filter description name string             */
/*    int lvlflag        validation level flag                      */
/*                                                                  */
/*   validation levels:                                             */
/*     level-1: Mask self usability validation,                     */
/*     level-2: Mask self content validation,                       */
/*     level-3: Mask coords validation against a target block       */
/*                                                                  */
/*     allowed #s for lvlfalg: 1 for level-1                        */
/*                             2 for level-2                        */
/*                             3 for level-3                        */
/*                            12 for both level-1 and -2            */
/*                            13 for both level-1 and -3            */
/*                      other #s for all, level-1 and -2 and -3     */
//*                                                                 */
/*  Returns:                                                        */
/*    dmBool            TRUE if validated or FALSE                  */
/*                                                                  */
/*                                                                  */
/********************************************************************/
dmBool dmMaskIsValidMask(dmBlock* msk_blk, dmBlock *flt_blk,  char *cptname, int lvlflag)
{
  unsigned char *vbytes;
  long nvals;
  int cflag;
  dmBool rtn;
  
  if(!msk_blk)
    return dmFALSE;
  
  if(lvlflag==1)         /* level-1 on;y */
    rtn = dmi_mask_isvalid_mask(msk_blk);
  else if( lvlflag==2){  /* level-2 only */
    rtn = dmi_mask_isvalid_content(msk_blk, &nvals, &vbytes);
    if (rtn) 
      free(vbytes);
  } else if( lvlflag == 3 ){  /* level-3 only */
    if( !flt_blk && (!cptname || strlen(cptname)==0) )
      rtn = dmFALSE;
    else
      rtn = dmi_mask_isvalid_coord(msk_blk, flt_blk, cptname, &cflag);
  } else if( lvlflag == 12 ){ /* level-1 + level-2,  */
    rtn = dmi_mask_isvalid_mask(msk_blk);
    if(rtn==dmTRUE){
      rtn=dmi_mask_isvalid_content(msk_blk, &nvals, &vbytes);
      if(rtn) 
	free(vbytes);
    }
  } else if( lvlflag == 13 ){ /* level-1 + level-3 */
    if(!flt_blk && (!cptname || strlen(cptname)==0) )
      rtn = dmFALSE;
    else {
      rtn = dmi_mask_isvalid_mask(msk_blk);
      if(rtn==dmTRUE)
 	rtn= dmi_mask_isvalid_coord(msk_blk, flt_blk, cptname, &cflag);
    }
  } else { /* all, level-1 + level-2 + level-3  */
    if(!flt_blk && (!cptname || strlen(cptname)==0) )
      rtn = dmFALSE;
    else {
      rtn = dmi_mask_isvalid_mask(msk_blk);
      if(rtn==dmTRUE){
	rtn=dmi_mask_isvalid_content(msk_blk, &nvals, &vbytes);
	if(rtn) 
	  free(vbytes);
      } 
      if(rtn==dmTRUE)
	rtn= dmi_mask_isvalid_coord(msk_blk, flt_blk, cptname, &cflag);
    }    
  }
  
  return rtn;
}

/********************************************************************/
/*                                                                  */
/* Function:   dmi_mask_remove_blkname                              */
/*                                                                  */
/* Description:                                                     */
/*   Remove the block-name from the string contaning MASK(blkname)  */
/*                                                                  */
/*   Note: MASK(blkname) is formed by dmi_mask_toString()           */
/*                                                                  */
/*  Input parameters:                                               */
/*    char** ss_string   Subspace string expression                 */
/*                       which is case (capital) sensitive          */
/*                                                                  */
/*  Output parameters:                                              */
/*    updated ss_string from "MASK(blkname)' to 'MASK'              */
/*                                                                  */
/*  Returns:                                                        */
/*    int:  #occurence (>=0) of updates or failure (-1)             */
/*                                                                  */
/********************************************************************/
int dmi_mask_remove_blkname(char **ss_string)
{
  char *ss_str = *ss_string;
  char *masktag_e=")";          /* ending */
  char *maskname, *masktag_s;   /* 'MASK', 'MASK(' for start */
  
  char *ptr_l, *ptr_r;
  long noccur=0;
  long str_len, mname_len;

  if( !ss_str )
    return -1;

  
  maskname  = dmMaskName();
  mname_len = strlen(maskname);
  if( strlen(ss_str) < mname_len ){
    free(maskname);
    return 0;
  }
  
  masktag_s = (char*)malloc( (mname_len+2)*sizeof(char));
  strcat( strcpy(masktag_s,maskname), "(" );

  /* MCD NOTE: would a walking scan would be more efficient? */
  /*   Here we scan the full string each iteration.          */
  while( (ptr_l=strstr(ss_str, masktag_s) )  != NULL){
    /* pass the mask-name string but the left parenthesis, '(' */
    ptr_l += mname_len;

    /* find pointer to the ending of the tag, excluding the right parenthesis, ')'*/
    ptr_r  = strstr(ptr_l, masktag_e) + 1;

    if( !ptr_r)
      return -1; /* syntax error */
    
    /* copy the overlap the src-str to dest-str by omitting the string between '(' and ')' and including the parenthesises */
    memmove(ss_str+strlen(ss_str)-strlen(ptr_l),  ptr_r, strlen(ptr_r) );

    /* marking the end of the updated string */
    str_len=strlen(ss_str) - ( strlen(ptr_l)-strlen(ptr_r) );
    ss_str[str_len]='\0';
    
    /* record the occurance of the removings */
    noccur++;
  }

  free(masktag_s);
  free(maskname);

  return noccur;
}

/***************************************************************/
/*                                                             */
/* Function:   dmi_mask_region_from_string                     */
/*                                                             */
/* Description:                                                */
/*   Generates a Region object from region syntax string       */
/*   which includes Mask shape.  The region library does       */
/*   not recognize user shapes, so must be handled separately  */
/*                                                             */
/* Input parameters:                                           */
/*   dmDescriptor *ssd           - subspace descriptor         */
/*   const char   *dsref         - MASK reference block name   */
/*   const char   *dsval         - Region string               */
/*                                                             */
/* Output parameters:                                          */
/*   regRegion   *region         - region struct               */
/*                                                             */
/* Returns:                                                    */
/*   dmErrCode   error_code                                    */
/*                                                             */
/***************************************************************/
regRegion* dmi_mask_region_from_string( dmDescriptor* ssd, const char* dsref, const char* dsval )
{
  dmDataset* ds = NULL;
  dmBlock*   block = NULL;

  const char *pstart;              /* start of clause     */
  const char *pstop;               /* end of clause       */
  char *tmpstr;                    /* copy of clause      */
  char *ptr;                       /* ptr into copy       */
  long  size;                      /* size of clause      */
  dmBool done = UT_FALSE;          /* Loop control flag   */
  char *maskname = dmMaskName();   /* Mask shape name     */

  regRegion* region = NULL;        /* Output region       */
  regShape*  mask   = NULL;        /* Subspace mask shape */

  regShape *shape = NULL;          /* shape ptr.          */
  regRegion *tmp_region = NULL;    /* temp. region struct */
  regMath  op;                     /* glue operation      */

  int no;                          /* block number      */
  long component;                  /* region cpt number */

  /* Get current dataset */
  block = dmi_descriptor_get_block( ssd );
  ds    = dmi_block_get_ds( block );

  /* Find referenced subspace block in dataset */
  no = dmi_block_search( ds, dsref );

  if ( no <= 0 )
  {
    sprintf( dm_errbuf, "No subspace block, %s\n", dsref );
    dmu_err_set( DM_ERR_SUBSPACE, dm_errbuf );
    return region;
  }


  /* Open the block.. this is expected to be a MASK image block */
  block = dmp_block_open_mode( ds, NULL, no, dmIMAGE, NULL, "R" );

  if ( !block || (dmu_err_get() != dmSUCCESS ) )
  {
    sprintf( dm_errbuf, "Problem reading subspace block, %s\n", dsref );
    dmu_err_set( DM_ERR_SUBSPACE, dm_errbuf );
    return region;
  }

  /* Load MASK block to regShape */
  mask = dmi_mask_create( (void*)block );
  if ( mask == NULL )
  {
    sprintf( dm_errbuf, "Problem interpreting subspace block, %s\n", dsref );
    dmu_err_set( DM_ERR_SUBSPACE, dm_errbuf );
    return region;
  }

  /* Assign coord flag for mask shape */
  dmDescriptor* dd = NULL;
  if ( ssd->descType == dmSUBSPACE ) /* should be */
  {
    /* Get column corresponding to subspace */
    dd = dmi_block_search_identifier( block, ssd->name );
  }
  if ( dd )
  {
    if ( dd->flag & DM_FLAG_LOGICAL )
    {
      mask->user->flag_coord = DM_FLAG_LOGICAL;
    }
    else if ( dd->flag & DM_FLAG_PHYS )
    {
      mask->user->flag_coord = DM_FLAG_PHYS;
    }
  }

  /* ---------------------------------------------------------------------- */
  /* Process dsval string.. using Region parser when possible               */
  /* Expectations:                                                          */
  /*   - string is simple list of Region shapes connected by '&','|' tokens */
  /*   - '|' tokens separate clauses                                        */
  /*   - "MASK" shape, if present is first in the clause (Mask requirement) */
  /*                                                                        */
  /* ---------------------------------------------------------------------- */

  /* Create resulting region structure */
  /*   - we will add shapes to this    */
  region = regCreateRegion(NULL,NULL);

  pstart = dsval;
  while ( !done )
  {
     op = regOR;

     pstop = strchr( pstart, '|' );

     if ( !pstop || *pstop == '\0' )
     {
       done = UT_TRUE;
       size = strlen( pstart );
     }
     else
     {
       size = pstop - pstart;
     }

     /* extract this clause */
     tmpstr = cs_dup_siz( pstart, size );
     ptr = tmpstr;


     /* If clause has MASK node (at start), add it to region */
     if ( (cs_uceq_siz( ptr, maskname, strlen(maskname)) ) ||
          (cs_uceq_siz( ptr+1, maskname, strlen(maskname)) ) )
     {
       /* Mask node found.. put copy of Mask shape to region */

       shape = regCopyShape( mask );

       /* check exclude syntax */
       if (*ptr == '!') 
       {
         shape->include = regExclude;
         ptr++;
       }

       regAddShape(region, op, shape);
       op = regAND;

       /* skip over mask node */
       ptr = ptr + strlen(maskname);
       if ( strlen(ptr) > 0 ){ ptr++;} /* skip operator, if any */

     }
     if ( *ptr != '\0' )
     {
       /* There is more to the clause, this should be regular Region syntax */
       /* Parse it using region parser and add results to our region.       */

       tmp_region = regParse( ptr );

       if ( !tmp_region )
       {
         sprintf( dm_errbuf, "Problem parsing subspace region string\n" );
         dmu_err_set( DM_ERR_SUBSPACE, dm_errbuf );
         return NULL;
       }
       else
       {
         /* Transfer shapes.. do not need anything as fancy as */
         /* the regIntersect/regUnion functions.               */

         /* cut shapes from temporary region struct */
         shape = tmp_region->shape;
         tmp_region->shape = NULL;
         regFree( tmp_region );

         /* Add the chain to our region */
         regAddShape(region, op, shape );
         op = regAND;

         /* update component numbers of trailing shapes. */
         component = shape->component;
         while ( shape != NULL )
         {
            shape->component = component;
            shape = shape->next;
         }
       }
     } /* end process Region portion of clause */

     free(tmpstr);
     if ( !done ){ pstart = ++pstop; }
  }

  return region;
}
