/*                                                                
**  Copyright (C) 2002-2004,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmparser.c                                                  */
/*                                                               */
/*   Functions that work with regions                            */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmlibP.h"
#include "dmiparser.h"


/********************************************************************/
/*                                                                  */
/* Function:   dmp_block_parse_filter                               */
/*                                                                  */
/* Description:                                                     */
/*  Parse a spec of the form                                        */
/* time=100:200,300:400.3,600:,(detx,dety)=circle(300,400,5),pha>22 */
/*                                                                  */
/*  Input parameters:                                               */
/*             dmBlock  *RawBlock                                   */
/*             char     *buf   -                                    */
/*                                                                  */
/*  Output parameters:                                              */
/*              none                                                */
/*                                                                  */
/*  Returns:                                                        */
/*           dmSubspace  *dmSubspace                                */
/*                                                                  */
/********************************************************************/
dmSubspace* dmp_block_parse_filter( dmBlock* RawBlock, char* buf )
{
  dmParseState  *state = NULL;
  dmSubspace    *rts   = NULL;
  long           clause_no = 0;
  dmBool         empty = UT_TRUE;
  dmClauseList  *clist = NULL;
  int            invert = 0;
  long           subcpt = 0; 
  long           n      = 10;  /* Number of clauses to start with */
  long           errCode = dmSUCCESS;


  if ( !buf || *buf == '\0' ) 
    return NULL;

  state = dmi_parse_state_init( n );
  rts = state->rts;
  rts->cpt = 1;

  /* Parse the #! syntax, which is prepended by the EXCLUDE directive */

  if ( strlen( buf ) > 3 && !strncmp( buf, "#! ", 3 ) )
  {
    rts->invert = 1;
    buf += 3;
  }

  dmi_parse_clause_stack( buf, state->stack );
  invert = dmi_parse_clause_line( buf, state->stack );

  if ( invert )
    rts->invert = 1;


  clist = dmi_parse_stack_eval( state->stack );

  for ( clause_no = 0; clause_no < clist->n; clause_no++ )
  {
    subcpt = clause_no+1;

    if ( subcpt > rts->ncpts )
      dmi_subspace_create_cpt( rts, empty );
    rts->cpt = subcpt;

    if ( dmverbose >= 2 )
     fprintf( stderr, "DM parse clause %s\n", clist->clauses[clause_no] );

    dmp_parse_filter_line( RawBlock, clist->clauses[clause_no], state ); 

    errCode = dmu_err_get();

    if ( errCode != dmSUCCESS )
      return rts;
  }


  dmi_parse_stack_free( state->stack );
  dmi_clauselist_free( clist );
  state->stack = NULL;
  rts->cpt = 1;
  free( state );
  return rts;
}


/***************************************************************/
/*                                                             */
/* Function:   dmp_parse_ext_filter                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock  *RawBlock                              */
/*             char     *buf   -                               */
/*             dmParseState  *state                            */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            int  retval                                      */
/*                                                             */
/***************************************************************/
int dmp_parse_ext_filter( dmBlock* RawBlock, char* buf, dmParseState* state )
{
  dmBlock    *ExternalBlock = NULL;
  int         retval = 0;
  int         invert = 0;
  dmSubspace *subspace = NULL;
  dmSubspace *rts      = state->rts;
  dmBool      copyDefault  = UT_FALSE;
  dmBool      requireTable = UT_TRUE;

  /* Open the specfied external dataset/block */
  ExternalBlock = (dmBlock*)dmTableOpen(buf);

  if ( !ExternalBlock )
    ExternalBlock = (dmBlock*)dmImageOpen(buf);

  if ( !ExternalBlock)
  {
    dmu_err_clear();
    dmp_apply_external_ascii_filter( RawBlock, buf, state );
    return retval;
  }

  invert = rts->invert;

  /* Get the list of DSS descriptors from the block */
  subspace = ExternalBlock->subspace;

  if ( dmverbose >=4 )
  {
    printf( "External subspace:\n" );
    dmi_subspace_print( subspace );
    printf( "RTS subspace:\n" );
    dmi_subspace_print( rts );
    printf( "---------------- \n" );
  }

  dmi_subspace_apply( subspace, rts, RawBlock, copyDefault, requireTable );

  /* Save invert from overwrite */
  rts->invert = invert;
 
  if ( dmverbose >=4 )
  {
    printf( "Post External subspace:\n" );
    dmi_subspace_print( rts );
    printf( "---------------- \n" );
  }

  retval = 1;
  return retval;
}



/***************************************************************/
/*                                                             */
/* Function:   dmp_parse_filter_line                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/* WARNING: This routine is called recursively when @ syntax   */
/*          is used                                            */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock  *block - table block                   */
/*             char     *buf                                   */
/*             dmParseState  *state                            */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmp_parse_filter_line( dmBlock* block, char* buf, dmParseState* state )
{
  char          *ptr        = buf;
  dmSubspace    *rts        = state->rts;
  char          *tptr       = NULL;
  char          *itemp      = NULL;
  long          *axisnop    = NULL;
  dmDescriptor  *sub_vector = NULL;
  dmFilter      *filter     = NULL;
  dmBool         bAndOr     = dmFALSE; // false == OR, true == AND
  dmParseStack  *pStack     = state->stack;
  dmParseStackEntry *pEntry = NULL;;

  char  op[MAX_NAME_LENGTH]; 
  char  name[MAX_NAME_LENGTH];

  int   intersect  = 0;
  int   is_image   = 0;
  long  vcpt       = 0;
  long  ty         = 0;
  long  axisno     = 0;
  long  itemno     = 0;
  long  errCode    = dmSUCCESS;


  if ( !buf || *buf == '\0' ) 
    return;

  is_image = ( block->type == dmIMAGE );

  if ( is_image )
    axisnop = &axisno;

  strcpy( name, "" );

  pEntry = &pStack->entry[1];

  // Code Change bug 8721  January 4, 2012
  //  Logical AND can be lost when filter contains
  //  a column that is filtered as part of virtual
  //  vector and filtered a second time as real column
  //  Set flag, and use it later for applying col filter.

  if (pEntry->op == 2) // AND op
    bAndOr = dmTRUE;

  itemno = 0;

  /*
   * Get the next comma separated token.
   * We need to support @ syntax here
   * Split the token into  <name><op><spec>
   * Example:   pha=4:6,20:30,time>5
   *  First token:  pha=4:6  a new filter
   *  Second token:  20:30   recognize as belonging to the first filter
   *  Third token   time>5   recognize > operator so is a new filter
   */


  while( ptr && *ptr ) 
  {
    /* Get comma separated token */
    /* Cases:
     *
     *      pha=....,
     *      (x,y)=...,
     *      pha=.. && ...
     *      (pha=...,g=..)||(...)
     *
     * First, search for an open-clause parenthesis
     */


    tptr = dmu_parse_token( ptr, ',', &itemp, 0 );

    // Code Change  Bug 12596  November 5, 2010
    // When two separators appear together, the return varialbe, itemp,
    //  can be NULL. Need to check pointer before using it. If itemp == NULL,
    //  assume that it is due to a typo and just ignore double separators.
    //  Continue while loop

    if ( ! itemp )
    {
      // Code Change  Bug 12596 (revisited)  April 15, 2011
      //  Just continuing did not help in a difference filter sequence, 
      //  (reported in Bug 12767) so be sure to adjust the buffer pointer too

      ptr = tptr;
      continue;
    }

    // Have token. Increment item counter.
    ++itemno;

    if ( *itemp == '@' )
    {
      char *extfile = itemp;

      ++extfile;

      if ( *extfile == '!' || *extfile == '-' )
      {
        rts->invert = 1;
	++extfile;
      }

      dmp_parse_ext_filter( block, extfile, state );
      ptr = tptr;
    }
    else
    {
      ptr = dmi_parse_filter_assign( itemp, itemno, axisnop, name, op, ptr, tptr );
 
      // Code Change  Bug 13182  March 12, 2012
      //  If syntax is in error, such as qualifier term in filter, there is 
      //  the possibility of having null in "op".
      //     eg: file'[counts, bin_low, bin_high]', missing term "cols" in filter string

      if ( *op == '\0' )
      {
        dmu_err_set( DM_ERR_PARSE, "Invalid filter syntax. Operand is null." );
        return;
      }

      if ( is_image && ( *op == '>' || *op == '<' ))
      {
        dmu_err_set( DM_ERR_PARSE, "Error: '>', '<' syntax not allowed for images." );
        return;
      }


      sub_vector = dmi_subspace_search_and_init( block, rts, name, &vcpt, &ty, &filter );

      if ( !sub_vector )
        return;

      intersect = dmi_parse_rangelist( sub_vector, vcpt, &ptr, filter, op, !is_image ); 

      /* Must be in DMP layer */
      if ( filter->regstring )
      {
        /* MCD NOTE: Failed dmRegParse due to bad region file was not caught correctly */
        /*    by the previous code logic.. this is kind-of ugly here, would prefer     */
        /*    just dmRegParse() which chooses Region or Mask.                          */
        /*                                                                             */
        if ( cs_index(filter->regstring, "mask(" ) != 0 )
        {
          filter->region = dmMaskParse(filter->regstring);
	  if( !filter->region)
	    return;
	  if( dmi_mask_set_coord( regGetShapeNo(filter->region, 1), block, name) != dmSUCCESS)
            return;
        }
        else
        {
          filter->region = dmRegParse( filter->regstring );    /* Filter parse */
        }
      }

      // Code change - Bug 9189
      // Check for null region error set by dmRegParse()
      // If an error, display warning message and clear error stack before
      // proceeding

      if ( (errCode = dmu_err_get()) == DM_WARN_NULL_REGION )
      {
        dmu_err_disp_select_msgs(DM_WARN_NULL_REGION);
        dmu_err_clear();
      }

      if (bAndOr)
        intersect = 1;

      dmi_subspace_col_apply_filter( block, rts, sub_vector, vcpt, ty, intersect, filter );
    }

    free( itemp );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmp_apply_external_ascii_filter                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock  *RawBlock                              */
/*             char     *filename                              */
/*             dmParseState  *state                            */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmp_apply_external_ascii_filter( dmBlock* RawBlock, char* filename, dmParseState* state )
{
  FioFile fptr;
  char    line[dmMAXSTR];
  char   *ptr = line;


  fptr = fio_open_ar( filename );

  if ( !fptr )
  {
    /* Error, couldn't open as text! */
    sprintf( dm_errbuf, "Failed to open external filter file %s", filename );
    dmu_err_set(DM_ERR_NO_FILE,dm_errbuf );
    return;
  }

  while( fio_read_line( fptr, ptr, dmMAXSTR ))
  {
    if ( dmverbose >= 4 )
      fprintf( stderr, "Parsing external filter line %s\n", ptr );
    dmp_parse_filter_line( RawBlock, ptr, state ); 
  }

  fio_file_close(fptr);
}



/***************************************************************/
/*                                                             */
/* Function:   dmp_block_process_range_filterr                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock  *block                                 */
/*             char     *spec                                  */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmp_block_process_range_filter( dmBlock* block,  char* spec )
{
  dmSubspace  *rts = NULL;
  dmDataset   *ds;
  dmKernel     kernel;


  rts = dmp_block_parse_filter( block, spec );

  if ( dmGetError()  != dmSUCCESS )
    return;

  if(dmverbose >= 4 )
  {
    fprintf( stderr, "\n" );
    fprintf( stderr, "--------------------------------------------\n" );
    fprintf( stderr, "\n" );
    fprintf( stderr, "DM Block Open: parse level 2 filter < %s > with %ld filters\n", spec, rts->ncols ); 

    dmi_subspace_print( rts );
    fprintf( stderr, "\n" );
    fprintf( stderr, "--------------------------------------------\n" );
    fprintf( stderr, "\n" );
  }

  dmi_block_setup_runtime_filter( block, block->buffer, rts );

  /* RTS may change if invert is enabled */
  rts = dmi_block_update_runtime_subspace( block, rts );

  /* Hint to the kernel in case it has its own filtering */
  if ( !rts->invert )
  {
    ds = block->ds;
    kernel = ds->kernel;
    kernel->BlockUpdateFilter( block, rts );
  }


#if DEBUG
  if (1)
#else
  if ( dmverbose >=4 )
#endif
  {
    fprintf( stderr, "Filtered Subspace:\n" );
    dmi_subspace_print( block->subspace);
    fprintf( stderr, "Runtime Subspace:\n" );
    dmi_subspace_print( rts );
  }
}




