/*                                                                
**  Copyright (C) 2002-2004,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmpfilter.c                                                 */
/*                                                               */
/*   Functions that work with regions                            */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/


#include "dmlibP.h"


/***************************************************************/
/*                                                             */
/* Function:   dmRegParse                                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              char  *buf - buffer containing region name     */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            regRegion  *region  - parsed region structure    */
/*                                                             */
/***************************************************************/
regRegion* dmRegParse( char* buf )
{
  regRegion  *region = NULL;
  dmErrCode   errorCode = 0;

  region = dmi_region_resolve( buf, dmp_region_read_file, NULL, NULL );   /* dmRegParse */


  // Code change - Bug 9189

  errorCode = dmu_err_get();

  if ( errorCode != dmSUCCESS  || !region )
  {
    /* don't overflow the error buffer here, regions can be huge */
    long lLengthRegionName = strlen(buf);
    long lLengthFormat     = 0;
    long len               = 0;
    char *pErrParseSyntax = "Bad syntax for region %s";
    char *pErrNullRegion  = "WARNING: no rows in %s\nThe region is null.";


    if (errorCode == DM_WARN_NULL_REGION)
    {
      // error that has been set has no text messsage
      // reset error with full text message, clear stack first
      dmu_err_clear();

      lLengthFormat = strlen(pErrNullRegion);
      len = lLengthFormat + lLengthRegionName + 1;

      if (len > TEXT_SIZE)
      {
        char *tmp_errbuf = (char*)malloc( UT_SZ_C * len );

        if (buf && *buf)
          sprintf( tmp_errbuf, pErrNullRegion, buf );
        dmu_err_set( DM_WARN_NULL_REGION, tmp_errbuf );
        free( tmp_errbuf );
      }
      else
      {
        if (buf && *buf)
          sprintf( dm_errbuf, pErrNullRegion, buf ); 
        dmu_err_set( DM_WARN_NULL_REGION, dm_errbuf );
      }

      // end code change - Bug 9189
    }
    else if ( errorCode == DM_ERR_OPEN_FAIL )
    {
      /* message is already set.. */
    }
    else
    {
      if (buf && *buf)
      {
        lLengthFormat = strlen(pErrParseSyntax);
        len = lLengthFormat + lLengthRegionName + 1;

        if (len > TEXT_SIZE )
        {
          char *tmp_errbuf = (char*)malloc( UT_SZ_C * len );
       
          sprintf( tmp_errbuf, "Bad syntax for region %s", buf );
          dmu_err_set( DM_ERR_PARSE, tmp_errbuf );
          free( tmp_errbuf );
        }
        else
        {
          sprintf( dm_errbuf, "Bad syntax for region %s", buf ); 
          dmu_err_set( DM_ERR_PARSE, dm_errbuf );
        }
      }
    }
  }

  return region;
}


/***************************************************************/
/*                                                             */
/* Function:   dmp_region_read_file                            */
/*                                                             */
/* Description:                                                */
/*            Read a region file from disk                     */
/*                                                             */
/*  Input parameters:                                          */
/*              char  *regfile - region file name              */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            regRegion  *region  - read region structure      */
/*                                                             */
/***************************************************************/
regRegion* dmp_region_read_file( char* regfile )
{
  int         verbose = dmverbose;
  dmBlock    *block   = NULL;
  char       *ptr     = NULL;
  regRegion  *region  = NULL;
  long        validRow = 0;
  dmErrCode   status = dmSUCCESS;

  ptr = regfile + strlen(regfile)-1;

  if ( *ptr == ')' )
    *ptr = '\0';

  /* First try ASCII */
  region = regReadAsciiRegion( regfile, verbose );

  if (!region )
  {
    block = dmTableOpen( regfile );

    if ( block )
    {

      // check for non-null region
      validRow = dmp_buffer_adjust(block, 1, 1, 0);

      region = dmTableReadRegion ( block );

      dmTableClose( block );

      if ( !region )
      {
        sprintf( dm_errbuf, "Failed to parse FITS region file %s", regfile );
        dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
      }

      // Code change - Bug 9189
      if (validRow == DM_BUF_BADROW)
      {
        // can't read first row of region - set error
        dmu_err_set( DM_WARN_NULL_REGION, "");
      } 
    }
    else
    {
      if ( fio_access( regfile ) )
      {
	status = DM_ERR_GENERIC;
        sprintf( dm_errbuf, "Failed to parse FITS region file %s", regfile );
      }
      else 
      {
	status = DM_ERR_OPEN_FAIL;
        sprintf( dm_errbuf, "Region file not found; %s", regfile ); 
      }

      dmu_err_set( status, dm_errbuf );
    }
  }

  return region;
}

