/*                                                                
**  Copyright (C) 1999-2001,2003,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmregion.c                                                  */
/*                                                               */
/*   Functions that work with regions                            */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include <string.h>
#include <stdio.h>
#include <float.h>
#include "dmlibP.h"
#include "utnan.h"

#define MAXLS 81
#define DCSIZE 80
#define dw_timeCol  0
#define dw_shapeCol 1
#define dw_rCol     2
#define dw_rotangCol 3
#define dw_cptCol    4
#define dw_statusCol   5
#define dw_posCol    6
#define BADPIX_USER_STATUS 0x40




/***************************************************************/
/*                                                             */
/* Function:   dmTableReadRegion                               */
/*                                                             */
/* Description:                                                */
/*             Read a region from a region block.  Try to      */
/*             interpret any block as a region block.          */
/*                                                             */
/*  Input parameters:                                          */
/*              dmblock  *block - block pointer                */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            regRegion  *region  -  region structure pointer  */
/*                                                             */
/***************************************************************/
regRegion* dmTableReadRegion( dmBlock* block )
{
 char shapeName[MAXLS] = "";
 char shapeType[MAXLS] = "";

 regRegion     *region = NULL;
 dmDescriptor **cols   = NULL;

 double   r[2]     = { 0.0, 0.0 };
 double   angle[2] = { 0.0, 0.0 };
 double  *pos = NULL;
 double  *x   = NULL;
 double  *y   = NULL;
 long    *dims   = NULL;
 long     ncols  = 0;
 long     maxlen = MAXLS - 1;
 long     offset = 0;
 long     maxpoints = 0; /* Max number of points for position vector */
 long     vdim      = 2;
 int      include   = 1;
 int      orFlag    = 0;
 int      start     = 1;
 short    oldCpt    = 0;
 short    cpt       = 0;


 if ( !block )
  return NULL;

 /* Which columns are the interesting ones? */
 ncols = dmi_region_file_cols( block, &cols, &dims );

 if ( !ncols )
   return NULL;


 maxpoints = dims[REG_POS_COL];
 vdim = 2;
 pos  = (double*)calloc( vdim * maxpoints, UT_SZ_D );
 x    = &pos[0];  
 y    = &pos[maxpoints];
 oldCpt = 1;
 cpt    = 1;  


 // this call creates a region with shape=NULL and xcols={0x0,0x0}
 // xregbounds and yregbounds are undefined

 region = regCreateRegion( NULL, NULL );


/* 
 *  Now we have the columns.
 */
  do
  {
    if ( cols[REG_POS_COL] )
      dmGetVectorArray_d( cols[REG_POS_COL], pos, vdim,  dims[REG_POS_COL] );

    if ( cols[REG_SHAPE_COL] )
      dmGetScalar_c( cols[REG_SHAPE_COL], shapeName, maxlen );

    if ( cols[REG_R_COL] )
      dmGetArray_d( cols[REG_R_COL], r, dims[REG_R_COL] );

    if ( cols[REG_ROT_COL] )
      dmGetArray_d( cols[REG_ROT_COL], angle, dims[REG_ROT_COL] );

    if ( cols[REG_CPT_COL] )
      cpt = dmGetScalar_s( cols[REG_CPT_COL] );

    include = (shapeName[0] != '!' );
    offset = include ? 0 : 1;
    strcpy( shapeType, shapeName+offset );

    if ( start )
      start = 0;
    else 
      orFlag = ( cpt != oldCpt );

    oldCpt = cpt;
    regAppendShape( region, shapeType, include, orFlag, x, y,
                    maxpoints, r, angle, UT_FALSE, UT_FALSE );

  } while( dmTableNextRow( block ) != dmNOMOREROWS );

  free( pos );
  free( cols );
  free( dims );

  dmTableSetRow( block, 1 );
  return region;
}




/***************************************************************/
/*                                                             */
/* Function:   dmTableWriteRegion                              */
/*                                                             */
/* Description:                                                */
/*  Prototype to write a region.                               */
/*  The new version will update the block in place if it exists*/
/*  and update time and status columns if they are present.    */
/*  This is for compatibility with BADPIX files.               */
/*                                                             */
/*  Input parameters:                                          */
/*       dmDataset    *ds       - destination dataset          */
/*       char         *name     - block name, usually REGION   */
/*       dmDescriptor *dd       - position column descriptor   */
/*                                containing WCS and name info,*/
/*                                can be NULL                  */
/*                                                             */
/*  Output parameters:                                         */
/*       regRegion    *inRegion - region to write to           */
/*                                                             */
/*  Returns:                                                   */
/*       dmblock *block                                        */
/*                                                             */
/***************************************************************/
dmBlock* dmTableWriteRegion( dmDataset* ds, char* name,
                             dmDescriptor* dd, regRegion* inRegion )
{
 regRegion    *region = NULL;
 dmDescriptor *vcoord = NULL;
 dmBlock      *block  = NULL;
 regShape     *Shape  = NULL;
 dmDescriptor *cols[10];
 dmDataType    type = DT_REAL8;
 char          ddname[DCSIZE];
 unsigned char status[2] = { 0, BADPIX_USER_STATUS };

 /* Pos is not used; arbitrary name position col recognized by dimension 2 */
 char* cnames[] = { "TIME", "SHAPE", "R", "ROTANG", "COMPONENT", "STATUS", "POS" };

 char      *colname = NULL;
 char     **cptNames = NULL;
 char       defaultName[] = "POS";
 char      *defaultCpts[] = { "X", "Y" };
 char       tabname[] = "REGION";
 char       tmpname[20];
 char       shapeName[20];
 double    *x       = NULL;
 double    *y       = NULL;
 double    *pos     = NULL;
 double     time    = 0.0;
 double     r[2]; 
 double     angle[2];
 long      *p_axlen = NULL;
 long       npoints = 0;
 long       offset  = 0;
 long       nnames = 6;
 long       nrows   = 0;
 long       n       = 0;
 long       no      = 0;
 long       arrayDim = 0;
 long       vectorDim = 2;
 long       naxes  = 1;
 long       axlen[1];
 long       p_arrayDim= 1;
 long       maxlen = 20;
 long       nradii = 0;
 long       nangles = 0;
 long       id      = 0;
 int        include = 0;
 short      cpt     = 1;
 short     *spos    = NULL;
 int        update = UT_FALSE;
 long       ii;


  for ( id = 0; id < nnames+1; id++ )
    cols[id] = NULL;

  region = regCopyRegion ( inRegion );
  arrayDim = regGetMaxPoints( region );

  if ( arrayDim < 1 )
    arrayDim = 1;


  if ( !name )
    name = tabname;

  if ( dd && dd->dim > 1 )
  {
    vcoord = dmDescriptorGetCoord( dd );
    colname = dd->name;
    cptNames = dd->cptNames;
  }

  if ( !colname )
    colname = defaultName;

  if ( !cptNames )
    cptNames = defaultCpts;

  if ( vcoord )
  {
     double crpix[2];
     double crval[2];
     double cdelt[2];
     double scale;
     long dim = 2;

     dmCoordGetTransform_d( vcoord, crpix, crval, cdelt, dim );
     scale = fabs( cdelt[0] );
     dmi_coord_region_convert_init( vcoord );
     regConvertWorldRegion( region, scale, dmi_coord_region_convert );
  }

  n   = regGetNoShapes( region );

  block = dmBlockOpen( ds, name );

  if ( block )
  {
    long  dim = 1;
    long  no  = 1;
    long  ncols = dmTableGetNoCols( block );
    dmDescriptor *dd = NULL;

    update = UT_TRUE;

    for ( no = 1; no <= ncols; no++ )
    {
      dd = dmTableOpenColumnNo( block, no );
      dim = dmGetElementDim( dd );
      dmGetName( dd, ddname, DCSIZE );

      if ( !cols[dw_posCol] && dim == 2 )
      {
        cols[dw_posCol] = dd;
        type = dmGetDataType( dd );
        p_arrayDim = dmGetArrayDimensions( cols[dw_posCol], &p_axlen );  

        if ( p_arrayDim == 0 )
          arrayDim = 1;
        else 
        {
          if ( p_axlen[0] < arrayDim || p_arrayDim != 1 )
          {
            printf( "Bad region array dim AD = %ld LOCAL = %ld %ld\n", arrayDim, p_arrayDim, p_axlen[0] );
          }
          arrayDim = p_axlen[0];
          free( p_axlen );
        }
      }
      else 
      {
        for ( id = 0; id < nnames; id++ )
        {
          if ( cs_uceq( cnames[id], ddname  ))
          {
            cols[id] = dd;      
            break;
          }
        }
      }
    }

    if ( !cols[dw_posCol] )
      return NULL;
    if ( !cols[dw_shapeCol] )
      return NULL;

    nrows = dmTableGetNoRows( block );
    dmTableSetRow( block, nrows );

    if ( cols[dw_timeCol] )
      time = dmGetScalar_d( cols[dw_timeCol] );
    offset = dmGetScalar_s( cols[dw_cptCol] );
    dmTablePutRow( block, NULL );
  }
  else
  {
    block = dmDatasetCreateTable( ds, name );
    axlen[0] = arrayDim;
    cols[dw_posCol]    = dmColumnCreateGeneric( block, colname, 
                            DT_REAL8, 0, "pixel", "Position", dmVALUE,
                            cptNames, vectorDim, axlen, naxes );

    cols[dw_shapeCol]  = dmColumnCreate( block, "SHAPE", DT_CHAR, 16, " ", "Region shape type" );
    cols[dw_rCol]      = dmColumnCreateArray( block, "R", DT_REAL8, 0, "pixel", "Radius", 2 );
    cols[dw_rotangCol] = dmColumnCreateArray( block, "ROTANG", DT_REAL8, 0, "pixel", "Angle", 2 );
    cols[dw_cptCol]    = dmColumnCreate( block, "COMPONENT", DT_INT2, 0, "", "Component number" ); 

    dmKeyWrite_c( block, "HDUCLASS", "ASC", "", "Region extension" );
    dmKeyWrite_c( block, "HDUCLAS1", "REGION", "", "Region extension" );
    dmKeyWrite_c( block, "HDUCLAS2", "STANDARD", "", "Region extension" );
    dmKeyWrite_c( block, "CONTENT", "REGION", "", "CXC Content key" );
  }


  pos = (double*)calloc( 2* arrayDim, UT_SZ_D );
  x = &pos[0];
  y = &pos[arrayDim];
  axlen[0] = arrayDim;


  for ( no = 1; no <=n; no++ )
  {
    /* initialize radii and rotang values to NaN */
    r[0] = ut_nan_create_d(UT_NAN_D_CXC);
    r[1] = ut_nan_create_d(UT_NAN_D_CXC);
    angle[0] = ut_nan_create_d(UT_NAN_D_CXC);
    angle[1] = ut_nan_create_d(UT_NAN_D_CXC);

    Shape =  regGetShapeNo( region, no );
    if ( Shape )
    {
      include = regShapeGetName( Shape, tmpname, maxlen );

      if ( !include )
        sprintf( shapeName, "!%s", tmpname );
      else
        strcpy( shapeName, tmpname );

      npoints = regShapeGetPoints( Shape, x, y, arrayDim );
      nradii  = regShapeGetRadii( Shape, r );
      nangles = regShapeGetAngles( Shape, angle );
      cpt     = regShapeGetComponent( Shape )+offset;

      switch( type )
      {
        case DT_INT2: /* Special case for BADPIX support */
          if ( !spos )
            spos = (short*)calloc( 2* arrayDim, UT_SZ_S );

          for ( id = 0; id < 2 * arrayDim; id++ ) 
          {
            spos[id] = (short)pos[id];
          }

          dmSetVectorArray_s( cols[dw_posCol], spos, vectorDim, arrayDim );
          break;
        case DT_REAL8: 
	  /* pad pos values outside of shape with NaN rather than 0.0 */
	  for (ii = npoints ; ii < arrayDim; ii++)
	  {
	    x[ii] = ut_nan_create_d(UT_NAN_D_CXC);
	    y[ii] = ut_nan_create_d(UT_NAN_D_CXC);
	  }
          dmSetVectorArray_d( cols[dw_posCol], pos, vectorDim, arrayDim );
          break;
        default:
          break;
      }

      dmSetScalar_c( cols[dw_shapeCol], shapeName );

      if ( cols[dw_rCol] )
         dmSetArray_d( cols[dw_rCol], r, 2 );
      else
        if ( nradii > 0 )
        {
           sprintf( dm_errbuf, "Shape %s cannot be stored in region block %s: needs radius column\n", 
                    shapeName, name );
           dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
           return NULL;
        }

      if ( cols[dw_rotangCol] )
        dmSetArray_d( cols[dw_rotangCol], angle, 2 );
      else
        if ( nangles > 0 && angle[0] != 0.0 )
        {
          sprintf( dm_errbuf, "Shape %s cannot be stored in region block %s: needs ROTANG column\n", 
                   shapeName, name );
          dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
          return NULL;
        }

      dmSetScalar_s( cols[dw_cptCol], cpt );

      if ( cols[dw_timeCol] )
        dmSetScalar_d( cols[dw_timeCol], time );

      if ( cols[dw_statusCol] )
        dmSetArray_bit( cols[dw_statusCol], status, 2 );
      dmTablePutRow( block, NULL );
    }
  }

  regFree( region );

  if ( spos )
    free( spos );

  if ( pos ) free( pos );
    dmBlockFlush( block );

  return block;
}



/***************************************************************/
/*                                                             */
/* Function:   dmSubspaceColGetRegionString                    */
/*                                                             */
/* Description:                                                */
/*             User call to return region string given region  */
/*             subspace descriptor.  Returns region for the    */
/*             current component.                              */
/*                                                             */
/*  Input parameters:                                          */
/*              dmDescriptor  *dd                              */
/*              long           maxlen                          */
/*                                                             */
/*  Output parameters:                                         */
/*              char          *region_string                   */
/*                                                             */
/*  Returns:                                                   */
/*            int  status                                      */
/*                                                             */
/***************************************************************/
int dmSubspaceColGetRegionString( dmDescriptor* dd, char* region_string, long maxlen )
{
  regRegion *region = NULL;
  dmFilter  *filter = NULL; 

  strcpy( region_string, "" );

  if ( dd )
  {
    filter =  dmi_subspace_col_get_current_filter( dd );

    if ( dmi_filter_is_null( filter ))
    {
      strcpy( region_string, "NULL" );
      return dmSUCCESS;
    }

    if ( dd->dim >= 2 )
      region = dmi_filter_get_region( filter );
  }

  if ( !region )
  {
    strcpy( region_string, "DEFAULT" );
    return dmFAILURE;
  }

  regComposeRegion( region, region_string, maxlen );
  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   dmSubspaceColGetRegion                          */
/*                                                             */
/* Description:                                                */
/*             Return a copy of the region for the current     */
/*             subspace.                                       */
/*                                                             */
/*  Input parameters:                                          */
/*              dmDescriptor  *dd                              */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            regRegion  *region                               */
/*                                                             */
/***************************************************************/
regRegion* dmSubspaceColGetRegion( dmDescriptor* dd )
{
  regRegion *region = NULL;

  if ( !dd ) 
  {
    dmu_err_set(DM_ERR_NULL, "[dmSubspaceColGetRegion] Null argument");
    return NULL;
  }

  region = dmi_subspace_col_get_region_pointer( dd, dd->block->subspace );
  return regCopyRegion( region );
}



/***************************************************************/
/*                                                             */
/* Function:   dmSubspaceColCreateRegion                       */
/*                                                             */
/* Description:                                                */
/*           Calls dmi_subspace_col_create_region() to do work */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock    *dd                                 */
/*              char       *name                               */
/*              char      **colNames                           */
/*              dmDataType  type                               */
/*              char       *unit                               */
/*              regRegion  *region                             */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            dmDescriptor  *dd                                */
/*                                                             */
/***************************************************************/
dmDescriptor* dmSubspaceColCreateRegion(dmBlock *block, char* name,
			   char** colNames, dmDataType type, char* unit, regRegion* region )
{
  return dmi_subspace_col_create_region( block, name, colNames, type, unit, region );
}



/***************************************************************/
/*                                                             */
/* Function:   dmRegConvertWorldRegion                         */
/*                                                             */
/* Description:                                                */
/*             Calls dmi_coord_convert_world_region() to do    */
/*             work.                                           */
/*                                                             */
/*  Input parameters:                                          */
/*              dmDescriptor  *vcoord                          */
/*              regRegion     *region                          */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            int  status                                      */
/*                                                             */
/***************************************************************/
int dmRegConvertWorldRegion( dmDescriptor* vcoord, regRegion* region )
{
 return dmi_coord_convert_world_region( vcoord, region );
}


/***************************************************************/
/*                                                             */
/* Function:   dmDescriptorGetArea                             */
/*                                                             */
/* Description:                                                */
/*             Calls dmi_descriptor_get_area() to do work.     */
/*                                                             */
/*  Input parameters:                                          */
/*              dmDescriptor  *dd                              */
/*              double        *field_area                      */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            double  area                                     */
/*                                                             */
/***************************************************************/
double dmDescriptorGetArea( const dmDescriptor* dd, double* field_area )
{
 return dmi_descriptor_get_area( dd, field_area );
}


/***************************************************************/
/*                                                             */
/* Function:   dmRegConvertInit                                */
/*                                                             */
/* Description:                                                */
/*             Public Interface                                */
/*             Calls dmi_coord_region_convert_init() to do work*/
/*                                                             */
/*  Input parameters:                                          */
/*              dmDescriptor  *dd                              */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmRegConvertInit( dmDescriptor* dd )
{
  dmi_coord_region_convert_init( dd );
}


/***************************************************************/
/*                                                             */
/* Function:   dmRegConvert                                    */
/*                                                             */
/* Description:                                                */
/*             Public Interface                                */
/*             Calls dmi_coord_region_convert() to do work     */
/*                                                             */
/*  Input parameters:                                          */
/*              double  *world                                 */
/*              double  *pix                                   */
/*                                                             */
/*  Output parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmRegConvert( double* world, double* pix )
{
  dmi_coord_region_convert( world, pix );
}
