/*                                                                
**  Copyright (C) 2002-2003,2005-2008  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftblocklib.c                                                */
/*                                                               */
/*   Functions that access blocks                                */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include <unistd.h>
#include "dmilib.h"
#include "ftkernelP.h"

#define XDEBUG 1



/***************************************************************/
/*                                                             */
/* Function:   ftp_block_alloc                                 */
/*                                                             */
/* Description:                                                */
/*            Add a new block to the ftds block list.          */
/*            If we need more space, add more blocks,          */
/*            initially 10, then 60, 310, ...                  */
/*                                                             */
/*  Input parameters:                                          */
/*                    none                                     */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                    ftblock  *block                          */
/*                                                             */
/***************************************************************/
ftBlock* ftp_block_alloc( void )
{
  ftBlock *ftblock = NULL;
 
 ftblock = (ftBlock*)calloc( 1, sizeof( ftBlock ) );

  if ( !ftblock )
    return NULL;

  return ftblock;
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_create_dummy_image                          */
/*                                                             */
/* Description:                                                */
/*     Make an empty image with the given name at the current  */
/*     location.  The image has NAXES = 0 and BITPIX = 16.     */
/*     If no > 1, insert at current position and give it the   */
/*     name. If no= 1 , the name is ignored.                   */
/*                                                             */
/*  Input parameters:                                          */
/*              fitsfile   *fptr                               */
/*              const long  no                                 */
/*              const char *extname                            */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                    nothing                                  */
/*                                                             */
/***************************************************************/
void ftp_create_dummy_image( fitsfile* fptr, const long no,
                             const char* extname )
{

  int     ftstatus = 0;
  int     bitpix   = 0 ;   /* FITS data type code */
  long    naxes    = 0;           /* Default image parameters */
  long    axes[2]  = { 0, 0 };
  double  bscale   = 1.0;
  double  bzero    = 0.0;  
  double  imdatatype = DT_INT2;  /* DM data type code */

 
  /* Old DM had resize_img here for HDU1, and flush_file, but */
  /* this seems superfluous                                   */
  /* Empty HDU1 is created at dataset create time             */

  bitpix = ftp_bitpix( imdatatype, &bscale, &bzero );

  if ( no > 1 )
  {
    fits_insert_img( fptr, bitpix, naxes, axes, &ftstatus ); /* CFITSIO */
    fits_update_key_str( fptr, "EXTNAME", (char*)extname, "Block name", &ftstatus );            /* CFITSIO */
  }
  else
  {
    fits_create_img( fptr, bitpix, naxes, axes, &ftstatus ); /* CFITSIO */
  }
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_block_insert                                */
/*                                                             */
/* Description:                                                */
/*              Create a new FITS block.                       */
/*      1. Position FITS file                                  */
/*      2. Parse block name into FITS keywords EXTNAME, EXTVER,*/
/*         HDUNAME                                             */
/*      3. Insert new HDU of required type.                    */
/*      4. Allocate and initialize new ftBlock internal        */
/*         structure, and tag it to the dataset.               */
/*      5. Add EXTNAME, EXTVER, HDUNAME keywords as required.  */
/*      6. Call BlockOpen routine.                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmDataset  *ds                                 */
/*              const char *name                               */
/*              long        no                                 */
/*              const dmBlockType  type                        */
/*              void       *kb                                 */
/*              char       *hint                               */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*               dmBlock  *block                               */
/*                                                             */
/***************************************************************/
dmBlock* ftp_block_insert( dmDataset* ds, const char* name,
                           long no, const dmBlockType type,
                           void* kb, char* hint )
{
  ftBlock    *seed_ftblock = (ftBlock*)kb;  /* For copying blocks */

  ftDataset  *ftds        = NULL;
  ftBlock    *ftblock     = NULL;
  fitsfile   *fptr        = NULL;
  dmBlockType btype;
  dmBlock    *block       = NULL;
  dmBool      ascii       = dmFALSE;  /* ASCII or Binary FITS table? */
  dmBool      have_extver = UT_FALSE;  /* EXTVER keyword to be added */
  dmBool      create      = UT_TRUE;   /* Pass CREATE flag to BlockOpen */
  dmErrCode   status      = dmSUCCESS;
  long        blockno     = 0; /* Ignored */
  long        nmax        = 0;
  long        extver      = 0;
  long        hdu         = 0;
  long        nrows       = 0;
  int         ft_hdutype  = 0;            /* CFITS local HDU type */
  int         ftstatus    = FITS_KERNEL_OK;
  char       *desc        = NULL;
  char       *old_desc    = "ASCDM block name";   /* If Back compat flag */
  char       *new_desc    = "CXCDM block name";

  char        hduname[FITS_MAXLEN] = "";
  char        extname[FITS_MAXLEN] = "";



  ftds = (ftDataset*)dmi_dataset_get_kds( ds ); 
  nmax = dmi_dataset_get_n( ds );

  /* Reposition CFITSIO within the file */
  hdu = no;

  if ( hdu <= 0 )
    hdu = ftds->current + 1;
  else
    if (hdu > 1 ) /* Force move to previous HDU */
      ftds->current = hdu-1;
  /*-- Old ftpCreateTable -- */

  if ( !ftp_move_to_current_hdu( ds ))
    return NULL;


  fptr = ftds->fptr;

  /* Create new ftBlock - care that null primary created first! */
  blockno = ftp_raw_block_alloc( ds, hdu, type, name );
  block   = ftp_block_open( ds, hdu, create, hint );
 
  ftblock = (ftBlock*)dmi_block_get_kb( block );

  dmi_block_set_type( block, type );
  dmi_block_set_name( block, name );
  dmi_dataset_set_current( ds, block );

  btype = dmi_block_get_raw_type( block );
  nrows = ( btype == dmTABLE ) ? 0 : 1;

  dmi_block_set_nrows( block, nrows );

  /* Set user defined parameters */
  if ( hint )
     status =  ftp_set_hint_list( block, hint, dmFALSE );


  /* This forces making an ASCII table copying from one. May want to do */
  /* something more flexible...                                         */
  if ( seed_ftblock )
    ascii = ( seed_ftblock->hdutype == ASCII_TBL );
  else 
    ascii   = ftp_make_ascii_tables( block ); /* Use HINT now */

  /* Convert CXCDM block name to FITS keyword values */
  have_extver = ftp_block_name_parse( name, hduname, extname, &extver );

  /* Add the empty binary table; it is created with NCOLS = 0 and NROWS = 1*/
  /* If the final NROWS = 0 this must be reset later.                      */

  if ( type == dmIMAGE )
  {
    ftp_create_dummy_image( fptr, hdu, extname );
    ft_hdutype = IMAGE_HDU;  
  }
  else
    if ( ascii )
    {
      fits_insert_atbl( fptr, 0, 0, 0, NULL, NULL, NULL, NULL, extname, &ftstatus ); /* CFITSIO */
      ft_hdutype = ASCII_TBL;
    }
    else
    {
      fits_insert_btbl( fptr, 0, 0, NULL, NULL, NULL, extname, 0, &ftstatus ); /* CFITSIO */
      ft_hdutype = BINARY_TBL;
    }

  if ( !ftp_ds_status( ds, ftstatus, "[ftpBlockCreate]" ) ) 
    return NULL;



  if ( type != ftp_block_type( ft_hdutype ) )
  {
    dmu_err_set( DM_ERR_INTERNAL, "[ftpBlockCreate]: Internal block type mismatch" );
    return NULL;
  }


  /* Record the block in the ftds object */
  ftp_dataset_add_block( ds, hdu, ft_hdutype );

  /* Move to newly created HDU */
  ftds->current = hdu;

  if (  !ftp_move_to_current_hdu( ds ) )
    return NULL;


  if ( have_extver )
    fits_update_key_lng( fptr, "EXTVER", extver, NULL, &ftstatus );  /* CFITSIO */

  desc = new_desc;

  if ( dm_global & DM_GLOBAL_OLD )
  {
    if ( cs_eq( hduname, "PRIMARY" ))
      desc = NULL;
    else
     desc = old_desc;
  }


  fits_update_key_str( fptr, "HDUNAME", hduname, desc, &ftstatus );    /* CFITSIO */



  // Code change - Bug 9522-A  June 12, 2006
  if (ftstatus == FITSERR207)
  {
    // Before failing, check strings to be sure there are no invalid
    // characters and retry update.

    ftstatus = FITS_KERNEL_OK;
    ftp_write_remove_invalid_chars(hduname);
    ftp_write_remove_invalid_chars(desc);
    fits_update_key_str( fptr, "HDUNAME", hduname, desc, &ftstatus );    /* CFITSIO */
    // end code change
  }


  ftp_ds_status( ds, ftstatus, "[ftpCreateBlock]: Problem writing keywords" );

  fits_flush_file( fptr, &ftstatus ); /* CFITSIO */

  ftp_block_init( block, create );
  ftp_subspace_open( block, create );
  ftp_block_reset( block );

  return block;
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_block_open                                  */
/*                                                             */
/* Description:                                                */
/*              Block open routine.                            */
/*              The create argument flags a newly created HDU. */
/*                                                             */
/*  Input parameters:                                          */
/*              dmDataset    *ds                               */
/*              long          no                               */
/*              const dmBool  create                           */
/*              char         *hint                             */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*               dmBlock  *block                               */
/*                                                             */
/***************************************************************/
dmBlock* ftp_block_open( dmDataset* ds, long no, const dmBool create,
                         char* hint )
{
  dmBlock   *block   = NULL;
  ftBlock   *ftblock = NULL;
  dmErrCode  status  = dmSUCCESS;


  /* Alloc block here */
  block = dmi_block_alloc( ds, no, create );

  if ( !block )
    return NULL;

  /*-- Old ftpOpenObject -- */
  ftblock = ftp_block_alloc( );

  if ( !ftblock ) 
    return NULL;

  dmi_block_set_kb( block, (dmkBlock*)ftblock );

  ftblock->hdu     = no;
  ftblock->cols    = NULL;

  if ( hint )
    status = ftp_set_hint_list( block, hint, dmTRUE );

  return block;
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_block_init                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock      *block                            */
/*              const dmBool  create                           */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*               nothing                                       */
/*                                                             */
/***************************************************************/
void ftp_block_init( dmBlock* block, const dmBool create )
{
  int       status  = 0;
  ftBlock  *ftblock = NULL;

  /* We duplicate the hdutype info in ftblock and ftds for convenience */

  ftblock = (ftBlock*) dmi_block_get_kb( block );
  ftblock->hdutype = ftp_block_hdutype( block );

  ftp_block_reset( block );

  /* Sync up header keys, even for newly created HDU */
  ftp_block_read_header( block );

  /* -- Old ftpOpenTab ftpOpenImg -- */
  status = ftp_table_open( block, create );

  if ( dmverbose > 4 )
    printf( "***** Find Transforms \n" );

  ftp_find_transforms( block );


  /*  Map vector columns and transform */

  ftp_composite_keylist_process( block, ftblock->keylist );

  if ( dmverbose > 4 )
    printf( "***** Find Vectors \n" );

  ftp_vectorize( block );
  ftp_block_read_pref( block );

  if ( dmverbose > 4 )
    ftp_block_print_cols( block );

  dmi_block_init( block );

  return;
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_block_flush                                 */
/*                                                             */
/* Description:                                                */
/*       Flush block structure to disk if dataset is writeable.*/
/*       (1) In OLD mode, flush MTYPE keys so they appear in   */
/*           compatible order.                                 */
/*       (2) Flush image structure                             */
/*       (3) DO NOT Update NAXIS2 key                          */ 
/*       (4) Update CHECKSUM/DATASUM if present                */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock      *block                            */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*               nothing                                       */
/*                                                             */
/***************************************************************/
void ftp_block_flush( dmBlock* block )
{
  dmDataset     *ds       = NULL;
  dmDescriptor  *dd       = NULL;
  dmBlockType    btype;
  fitsfile      *fptr     = NULL;
  ftCol         *ftcol    = NULL;
  int            ftstatus = FITS_KERNEL_OK;
  long           ii       = 0;
  long           ncols    = 0;
  char           checksum[80];


  ds = dmi_block_get_ds( block );

  if ( !dmi_dataset_mode( ds) )
    return;

  ftp_block_reset( block );

  if ( dm_global & DM_GLOBAL_OLD )
  {
    ncols = dmi_block_get_ncols( block );

    for ( ii = 1; ii <= ncols; ii++ )
    {
      dd = dmi_block_get_col( block, ii );

      ftcol = (ftCol*)dmi_descriptor_get_kdd( dd );

      if ( ftcol && ftcol->update )
      {
        ftp_col_write_mtype( block, dd, 0 );
        ftcol->update = UT_FALSE;
      }
    }
  }

  btype = dmi_block_get_raw_type( block );

  if ( btype == dmIMAGE )
    ftp_flush_image_struct( block );


  fptr = ftp_fptr( block );

#if 0 /* we should never have to manually modify NAXIS2 */
  nrows = dmi_block_get_nrows( block );

  if ( btype == dmTABLE )
    fits_update_key_lng( fptr, "NAXIS2", nrows, NULL, &ftstatus );  /* CFITSIO */
#endif


  /* Last of all: If file has CHECKSUM, update it */
  if ( fits_read_key_str( fptr, "CHECKSUM", checksum, NULL, &ftstatus ) != KEY_NO_EXIST ) /* CFITSIO */
    fits_write_chksum( fptr, &ftstatus );  /* CFITSIO */
  else
    ftstatus = FITS_KERNEL_OK;

  /* And flush again */
  fits_flush_file( fptr, &ftstatus );
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_flush_image_struct                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock      *block                            */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*               nothing                                       */
/*                                                             */
/***************************************************************/
void ftp_flush_image_struct( dmBlock* block )
{
  long           naxes = 0;
  long          *axlen = NULL;
  dmDescriptor  *dd    = NULL;
  dmDescriptor  *us_dd = NULL;
  dmDataType     type;
 

  /* Make sure image parameters are up to date */
  dd = dmi_block_get_col( block, 1 );

  if ( dd && dmi_descriptor_check_update( dd ) )
  {
    naxes = dmi_descriptor_get_array_dimensions( dd, &axlen );
    us_dd = dmi_descriptor_get_unscaled( dd );
    type  = dmi_descriptor_get_datatype( us_dd );

    ftp_image_reset( block, type, axlen, naxes );  
    free( axlen );

    dmi_descriptor_clear_update( dd );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_move_to_current_hdu                         */
/*                                                             */
/* Description:                                                */
/*           Force move to start of current HDU, given dataset */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock      *block                            */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*               dmBool status                                 */
/*                                                             */
/***************************************************************/
dmBool ftp_move_to_current_hdu( const dmDataset* ds )
{
  ftDataset  *ftds     = NULL;
  int         ftstatus = FITS_KERNEL_OK;
  int         htype    = 0;
  int         hdu      = 0;

  ftds = (ftDataset*)dmi_dataset_get_kds( ds );
  hdu = ftds->current;

  if ( hdu < 1 )
    hdu = 1;

  fits_movabs_hdu( ftds->fptr, hdu, &htype, &ftstatus );  /* CFITSIO */
  ftds->current = hdu;

  return ftp_ds_status( ds, ftstatus, "[ftp_move_to_current_hdu]" );
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_block_reset                                 */
/*                                                             */
/* Description:                                                */
/*       Move to start of HDU for given block; only move if    */
/*       not at current position.  Flush file if writing.      */
/*                                                             */
/*  Input parameters:                                          */
/*                const dmBlock  *block                        */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*               dmBool status                                 */
/*                                                             */
/***************************************************************/
dmBool ftp_block_reset( const dmBlock* block )
{
  ftBlock    *ftblock = NULL;
  dmDataset  *ds      = NULL;
  ftDataset  *ftds    = NULL;
  int        ftstatus = FITS_KERNEL_OK;
  int        hdutype  = 0;
  dmBool     status   = dmFALSE;
  fitsfile   *fptr    = NULL;


  if ( !block ) 
    return UT_FALSE;

  ds = dmi_block_get_ds( block );

  if ( !ds )
  {
    printf( "NO DS FOR BLOCK %s\n", dmi_block_get_name( block ));
    return UT_FALSE;
  }

  ftds = (ftDataset*) dmi_dataset_get_kds( ds );
  fptr = ftds->fptr;
  ftblock = (ftBlock*)dmi_block_get_kb( block );

  if ( !ftblock )
  {
    sprintf( dm_errbuf, "No kernel block for %s", dmi_block_get_name( block ));
    dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
    return  UT_FALSE;
  }

#if 1
 {
   int truehdu = 0;
   fits_get_hdu_num( fptr, &truehdu );

   if ( truehdu != ftds->current )
   {
      sprintf( dm_errbuf, "HDU SYNC ERROR %d %ld\n", truehdu, ftds->current );
      dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
      return UT_FALSE;
   }
 }
#endif

  /* If already at right position, do nothing */
  if ( ftblock->hdu == ftds->current )
  {
    return UT_TRUE;
  }

  /* If writeable, flush */
  /* What is the block pointer for the currently open block? */

  if ( dmi_dataset_mode( ds ) )
  {
    fits_flush_file( fptr, &ftstatus );          /* CFITSIO */
  }

  /* Do the move and record it */

  fits_movabs_hdu( fptr, (int)ftblock->hdu, &hdutype, &ftstatus );  /* CFITSIO */

  ftds->current = ftblock->hdu;
  status        = ftp_ds_status( ds, ftstatus, "[ftp_block_reset]" );

  return status;
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_block_close                                 */
/*                                                             */
/* Description:                                                */
/*       Free memory associated with an open ftBlock, but      */
/*       retain the basic index info                           */
/*                                                             */
/*  Input parameters:                                          */
/*                const dmBlock  *block                        */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*               nothing                                       */
/*                                                             */
/***************************************************************/
void ftp_block_close( dmBlock* block )
{
  /* Free keylist */
  ftBlock *ftblock = NULL;

  ftblock = (ftBlock*)dmi_block_get_kb( block );

  if ( !ftblock || !ftblock->keylist )
    return;

  if ( ftblock->keylist ) 
    ftp_keylist_free( block, ftblock->keylist );

  ftblock->keylist = NULL;
  ftp_hint_free( ftblock );
  free( ftblock->filter );

  ftblock->filter = NULL;
  free( ftblock->vector_buffer );

  ftblock->vector_buffer_size= 0;
  ftblock->vector_buffer = NULL;

  /* The individual descriptors are freed at the DM level */
  free ( ftblock->cols );

  ftblock->cols  = NULL;
  ftblock->ncols = 0;


  /* Close block and free ftblock. Do not free block, as we use */
  /* block number information in dmi_block_delete routine:      */
  /* must therefore be called together                          */
  /* with either dmi_block_free or dmi_block_delete.            */

  /* Free the block */
  free( ftblock );

  dmi_block_set_kb( block, NULL );
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_keylist_free                                */
/*                                                             */
/* Description:                                                */
/*       Zero out references to the dmu Keys in the dm Keys    */
/*                                                             */
/*  Input parameters:                                          */
/*                const dmBlock  *block                        */
/*                dmuKeyList      keylist                      */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*               nothing                                       */
/*                                                             */
/***************************************************************/
void ftp_keylist_free( dmBlock* block, dmuKeyList keylist )
{
  long   no = 0;
  long   n  = 0;

  dmDescriptor *dd = NULL;


  n = dmi_block_get_nkeys( block );

  for ( no = 1; no <= n; no++ )
  {
    dd = dmi_block_get_key( block, no );
    dmi_descriptor_set_kdd( dd, NULL );
  }

  /* Free the dmu Keys */
  dmu_keylist_free( keylist );
}



