/*                                                                
**  Copyright (C) 2002-2004,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftblockprint.c                                              */
/*                                                               */
/*   Functions that print block elements and columns             */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmilib.h"
#include "ftkernelP.h"

/***************************************************************/
/*                                                             */
/* Function:   ftp_block_hdutype                               */
/*                                                             */
/* Description:                                                */
/*      Returns block hdu type                                 */
/*                                                             */
/*  Input parameters:                                          */
/*           const  dmBlock  *block                            */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                 int  hdutype                                */
/*                                                             */
/***************************************************************/
int ftp_block_hdutype( const dmBlock* block )
{
  ftDataset *ftds    = NULL;
  ftBlock   *ftblock = NULL;
  dmDataset *ds      = NULL;

  long hdu     = 0;
  int  hdutype = -1;

  ftblock = (ftBlock*)dmi_block_get_kb( block );

  if ( !ftblock )
    return hdutype;

  hdu = ftblock->hdu;
  ds = dmi_block_get_ds( block );
  ftds = (ftDataset*)dmi_dataset_get_kds( ds );

  if ( hdu > 0 && hdu <= ftds->n )
  {
    hdutype = ftds->hdutype[hdu-1]; 
  }
  return hdutype;
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_block_print_cols                            */
/*                                                             */
/* Description:                                                */
/*      Raw block column print                                 */
/*                                                             */
/*  Input parameters:                                          */
/*           const  dmBlock  *block                            */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                 void                                        */
/*                                                             */
/***************************************************************/
void ftp_block_print_cols( const dmBlock* block )
{
  ftBlock *ftblock = NULL;
  long     ii      = 0;

  ftblock = (ftBlock*)dmi_block_get_kb( block );

  for ( ii = 0; ii < ftblock->ncols; ii++ )
  {
    ftp_col_print( ftblock->cols[ii] );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_col_print                                   */
/*                                                             */
/* Description:                                                */
/*      Raw column print                                       */
/*                                                             */
/*  Input parameters:                                          */
/*                dmDescriptor  *col                           */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                 void                                        */
/*                                                             */
/***************************************************************/
void ftp_col_print( dmDescriptor* col )
{
  ftCol *ftcol = NULL; 

  char buf[TEXT_SIZE];
  char fbuf[CARD_SIZE];

  if ( col )
  {
    ftcol = (ftCol*)dmi_descriptor_get_kdd( col );

    if ( ftcol && dmi_descriptor_is_varlen( col ) )
      cs_copy( "v", fbuf );
  }

  dmi_col_print( col, buf, TEXT_SIZE );
  printf( "%s\n", buf );
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_block_print                                 */
/*                                                             */
/* Description:                                                */
/*      Print the raw block properties                         */
/*                                                             */
/*  Input parameters:                                          */
/*                dmDescriptor  *col                           */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                 void                                        */
/*                                                             */
/***************************************************************/
void ftp_block_print( const dmBlock* block )
{
  ftBlock     *ftblock = NULL;
  dmBlockType  type    = dmUNKBLOCK;

  char hdutype_name[CARD_SIZE];
  char dmtype_name[CARD_SIZE];
  char *name = NULL;


  if ( !block )
    return;

  ftblock = (ftBlock*)dmi_block_get_kb( block );
  name = dmi_block_get_name( block );
  printf( "Block %ld: %s\n", ftblock->hdu, name );

  type = dmi_block_get_raw_type( block );
  dmu_block_type_name( type, dmtype_name );
  printf( "DM Block type: %s\n", dmtype_name );
  ftp_hdutype_name(  ftblock->hdutype, hdutype_name );
  printf( "FITS HDU type: %s \n", hdutype_name );
  printf( "\n" );

  ftp_block_print_cols( block );
  return;
}



