/*                                                                
**  Copyright (C) 2002-2003,2005,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftckey.c                                                    */
/*                                                               */
/*   Functions that work with keys list                          */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmilib.h"
#include "ftkernelP.h"
#include "ftkernel.h"

long ftp_key_pos( dmBlock* block, long keyno );


/***************************************************************/
/* Function:   ftGetKernelKey                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/* Input parameters:                                           */
/*             dmBlock  *block                                 */
/*             long      no     - key number                   */
/*             long      maxlen - buf length                   */
/*                                                             */
/* Output parameters:                                          */
/*             char     *buf - kernel key                      */
/*                                                             */
/* Returns:                                                    */
/*           void                                              */
/*                                                             */
/***************************************************************/
void ftGetKernelKey( dmBlock* block, long no, char* buf, long maxlen )
{
  char buf1[256];
  char m = ' ';
  char ct[80] = " ";

  dmuKey*    ftkey   = NULL;
  ftBlock*   ftblock = (ftBlock*)dmi_block_get_kb( block );
  dmuKeyList keylist;

  cs_copy( " ", buf );
  if ( ftblock )
  {
    keylist = ftblock->keylist;
    ftkey = keylist->data[no-1];
    ftp_keytype_name( (ftKeyType)ftkey->keytype, ct, 80 );

    if ( ftkey->mask ) { m = '*'; }

    dmu_key_print( ftkey, NULL, UT_FALSE, buf1, maxlen );
    sprintf( buf, "Key %4ld: %-14s%c%s", no, ct, m, buf1 );
  }
}


/***************************************************************/
/* Function:  ftGetNoKernelKeys                                */
/*                                                             */
/* Description:                                                */
/*   Returns the number of keys on the kernel keylist          */
/*                                                             */
/* Input parameters:                                           */
/*            dmBlock  *block - DM block                       */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*    long nkeys         Number of keys in kernel keylist      */
/*                                                             */
/***************************************************************/
long ftGetNoKernelKeys( dmBlock* block )
{
  ftBlock* ftblock = (ftBlock*)dmi_block_get_kb( block );
  dmuKeyList keylist;

  if ( !ftblock )
    return 0;

  keylist = ftblock->keylist;
  if ( !keylist )
    return 0;

  return keylist->n;
}


/***************************************************************/
/* Function:  ftWriteComment                                   */
/*                                                             */
/* Description:                                                */
/*   Write a COMMENT after key at specified position           */
/*                                                             */
/* Input parameters:                                           */
/*              dmBlock  *block                                */
/*        const long      keyno - key number                   */
/*        const char     *tag   - key tag                      */
/*        const char     *comment - comment string             */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*        dmBool  dmSUCCESS or dmFAILURE                       */
/*                                                             */
/***************************************************************/
dmBool ftWriteComment( dmBlock* block, const long keyno, const char* tag, const char* comment )
{

  ftBlock*   ftblock = NULL;
  dmuKeyList keylist;

  long ftkeyno = 0;
  long offset = 0;
  long nkeys  = 0;

  nkeys   = dmi_block_get_nkeys( block );
  ftblock = (ftBlock*)dmi_block_get_kb( block );
  if ( !ftblock )
    return UT_FALSE;

  keylist = ftblock->keylist;

  if ( keyno >= nkeys )
  {
    ftkeyno = 0; /* append */
  }
  else
  {
    ftkeyno = ftp_key_pos( block, keyno );
  }
  if ( ftkeyno > 0 ) 
  {
    offset = ftkeyno-1;
    while( offset < keylist->n && keylist->data[offset]->mask )    
    {
      offset++;
    }
    ftkeyno = offset+1;
  }
  ftp_key_write_generic( block, ftkeyno, FTK_COMMENT, DT_CHAR, tag, &comment, NULL, NULL );

  return UT_TRUE;
}


/***************************************************************/
/* Function:  ftp_key_pos                                      */
/*                                                             */
/* Description:                                                */
/*   Convert current DM key position to current FITS           */
/*   key position.                                             */
/*                                                             */
/* Input parameters:                                           */
/*           dmBlock  *block                                   */
/*           long      keyno                                   */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*          long  ftkeyno - current key position               */
/*                                                             */
/***************************************************************/
long ftp_key_pos( dmBlock* block, long keyno )
{
  dmDescriptor *dd    = NULL;
  dmuKey       *ftkey = NULL;

  long ftkeyno = 0;

  if ( keyno > 0 )
  {
    dd = dmi_block_get_key( block, keyno );
    if (dd )
      ftkey = dmi_descriptor_get_kdd( dd );

    if ( ftkey )
      ftkeyno = ftkey->index;

  }
  return ftkeyno;
}


/***************************************************************/
/* Function: ftReadComment                                     */
/*                                                             */
/* Description:                                                */
/*   Read the next comment following specifed key number       */
/*                                                             */
/* Input parameters:                                           */
/*            dmBlock  *block                                  */
/*            long      keyno - key number                     */
/*                                                             */
/* Output parameters:                                          */
/*             char   **tag - key tag                          */
/*             char   **comment - key comment                  */
/*                                                             */
/* Returns:                                                    */
/*          dmBool UT_FALSE or UT_TRUE                         */
/*                                                             */
/***************************************************************/
dmBool ftReadComment( dmBlock* block, long keyno, char** tag, char** comment )
{
  dmuKey  *ftkey = NULL;
  ftBlock *ftblock = NULL;

  dmuKeyList keylist;
  dmuKeyType ktype;

  long ftkeyno      = 0;
  long start_offset = 0;
  long end_offset   = 0;
  long koffset      = 0;
  long nkeys        = 0;

  nkeys = dmi_block_get_nkeys( block );
  ftblock = (ftBlock*)dmi_block_get_kb( block );

  *tag = NULL;
  *comment = NULL;

  if ( !ftblock || !ftblock->keylist )
    return UT_FALSE;
  if ( keyno < 0 || keyno > nkeys )
    return UT_FALSE;

  keylist = ftblock->keylist;
  ftkeyno = ftp_key_pos( block, keyno );

  /* The next DM level key is the next key with mask not set     */
  /* Associated keys can be in the range start_offset:end_offset-1 */
  koffset = ftblock->keyno;      /* last one tried plus 1  (keyno -1 + 1 ) */
  start_offset = ftkeyno;        /* ftkeyno - 1 + 1 */
  end_offset = start_offset;

  while( end_offset < keylist->n && keylist->data[end_offset]->mask ) 
    end_offset++;

  /* 
   * koffset saved from last trip; if it's outside the range, we must have been
   * looking at another key, so we reset.
   * that is: 
   *   KEY1  = VAL1
   *   COMMENT  k1.c1
   *   COMMENT  k1.c2
   *   COMMENT  k1.c3
   *   KEY2  = VAL2
   *   COMMENT  k2.c1
   *
   *  In the DM we provide access to k1.c1 to k1.c3 via KEY1, considering them
   *  as comments associated with the preceding key
   *  We keep track of which the next comment is. If we get to KEY2 we
   *  reset the counter.
   */
  if ( koffset < start_offset || koffset > end_offset ) 
    koffset = start_offset;

  /* Now we look for a comment */
  while( koffset < end_offset )
  {
    ftkey = keylist->data[koffset];
    ktype = ftkey->keytype;
    if ( ktype == FTK_COMMENT )
    {
      *tag = cs_dup( ftkey->name );
      *comment = cs_dup( *(char**)&ftkey->value );
      ftblock->keyno = koffset+1;
      return UT_TRUE;
    }
    koffset++;
  }

  /* No more comments for this key */
  ftblock->keyno = -1; 

  return UT_FALSE;
}


/***************************************************************/
/* Function:  ftKeyWrite                                       */
/*                                                             */
/* Description:                                                */
/*   Write a key.                                              */
/*   Update both the keylist and the file on disk.             */
/*   Write the key just before composite key [keyno+1]         */
/*   except if the key already exists, in which case overwrite */
/*   it in place.                                              */
/*   Similarly, dmi_key_alloc returns the old updated          */
/*   descriptor if it exists.                                  */
/*                                                             */
/* Input parameters:                                           */
/*                dmBlock    *block                            */
/*                long        keyno  - key number              */
/*          const char       *name   - key name                */
/*          const dmDataType  type   - data type               */
/*          const void       *value  - key value               */
/*                long        nvals  - number of values        */
/*          const char       *unit   - unit string             */
/*          const char       *comment - comment string         */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*         dmDescriptor  *dd - key descriptor                  */
/*                                                             */
/***************************************************************/
dmDescriptor* ftKeyWrite( dmBlock* block, const long keyno, 
			  const char* name, const dmDataType type, 
			  const void* value, long nvals, 
			  const char* unit, const char* comment )
{
  ftBlock      *ftblock   = NULL;
  dmuKey       *ftkey     = NULL;
  dmuKey       *old_ftkey = NULL;
  dmDescriptor *dd        = NULL;

  long nkeys = 0;
  long new_keyno = (long)keyno;
  long pos = 0; /* Raw FITS key number */

  /* Catch erroneous COMMENT and HISTORY in FITS.  */
  /* You should use dmBlockWriteComment for these  */
  if ( cs_eq( name, "HISTORY" ) || cs_eq( name, "COMMENT" ))
  {
    /* Naughty user! Comments are not DM keywords */
    ftWriteComment( block, keyno, name, *(char**)value );
    return NULL;
  }

  nkeys   = dmi_block_get_nkeys( block );
  ftblock = (ftBlock*)dmi_block_get_kb( block );
  if ( !ftblock )
    return NULL;

  if ( keyno > 0 )
    new_keyno = keyno + 1;

  dd = dmi_key_alloc( block, new_keyno, name, type, value, nvals, unit, comment );

  /* If keyword is preexisting, we can reset the number here */
  new_keyno = dmi_descriptor_get_no( dd );

  old_ftkey = (dmuKey*)dmi_descriptor_get_kdd( dd );  
  if ( old_ftkey )
  {
    pos = old_ftkey->index;
  } 
  else if ( keyno == 0 || ( keyno >= nkeys ) ) 
  {
    pos = 0;
  }
  else 
  {
    /* New key */
    pos = ftp_key_pos( block, keyno ) + 1;
  }

  ftkey = ftp_composite_key_write( block, pos, name, type, value, nvals, unit, comment );
  if ( old_ftkey && old_ftkey != ftkey )
  {
    /* We've changed ftkey.. hmm. Ignore this for now but perhaps worth
       checking at verbose = n? */
  }

  if ( !ftkey )
  {
    if ( !old_ftkey )
    {
      dmi_key_delete( dd );
    }
    dd = NULL;
  }
  else
  {
    dmi_descriptor_set_kdd( dd, (void*)ftkey );
    dmi_block_set_keyno( block, new_keyno );
  }

  return dd;
}


/***************************************************************/
/* Function:  ftKeyDelete                                      */
/*                                                             */
/* Description:                                                */
/*       Remove key from key list given key descriptor         */
/*                                                             */
/* Input parameters:                                           */
/*            dmDescriptor  *dd - key to delete                */
/*                                                             */
/* Output parameters:                                          */
/*    none                                                     */
/*                                                             */
/* Returns:                                                    */
/*             void                                            */
/*                                                             */
/***************************************************************/
void ftKeyDelete( dmDescriptor* dd )
{
  fitsfile  *fptr    = NULL;
  dmBlock   *block   = NULL;
  ftBlock   *ftblock = NULL;

  dmuKey    *ftkey      = NULL;
  dmuKey    *main_ftkey = NULL;
  dmuKeyList keylist;

  char *name = NULL;
  char  keyname[FITS_MAXLEN];
  char  genname[FITS_MAXLEN]; /* Key name, possibly amended with trailing *  */
  char  valkeyname[FITS_MAXLEN];

  int  verbose_header = 0;
  int  use_dval = 0;
  long nvals = 1;
  long index = 0;
  long ii = 0;
  int ftstatus = FITS_KERNEL_OK;


  main_ftkey = (dmuKey*)dmi_descriptor_get_kdd( dd );
  block = dmi_descriptor_get_block( dd );
  ftblock = (ftBlock*)dmi_block_get_kb( block );
  if ( !ftblock )
    return;

  fptr = ftp_fptr( block );
  name =  dmi_descriptor_get_name( dd );

  cs_copy( name, genname );
  nvals = dmi_descriptor_get_array_size( dd );
  if( nvals > 1 )
    strcat( genname, "*" );

  keylist = ftblock->keylist;

  verbose_header = ftp_verbose_header( block );
  if ( strlen( name ) > 8  || verbose_header || nvals > 1 )
  {
    index = dmu_keylist_check_index( keylist, "DTYPE", genname ); 
    if ( index > 0 )
    {
      sprintf( keyname, "DTYPE%ld", index );
      fits_delete_key( fptr, keyname, &ftstatus ); /* CFITSIO */
      ftkey = dmu_key_find( keylist, keyname );
      if ( ftkey )
	dmu_key_delete( keylist, ftkey );
      if ( verbose_header )
      {
	sprintf( keyname, "DFORM%ld", index );
	fits_delete_key( fptr, keyname, &ftstatus ); /* CFITSIO */
	ftkey = dmu_key_find( keylist, keyname );
	if ( ftkey )
	  dmu_key_delete( keylist, ftkey );
	sprintf( keyname, "DUNIT%ld", index );
	fits_delete_key( fptr, keyname, &ftstatus ); /* CFITSIO */
	ftkey = dmu_key_find( keylist, keyname );
	if ( ftkey )
	  dmu_key_delete( keylist, ftkey );
      }
      if ( strlen( name ) > 8 )
      {
	sprintf( valkeyname, "DVAL%ld", index );
	name = valkeyname;
	use_dval = 1;
      }
    }
  }

  if ( nvals <= 1 )
  {
    /* Might be writing  name  or DVALn */
    fits_delete_key( fptr, name, &ftstatus ); /* CFITSIO */
  }
  else
  {
    /* Support for array keywords, not used in practice so far */
    for ( ii = 1; ii <= nvals; ii++ )
    {
      if ( use_dval )
	sprintf( valkeyname, "%ldDVAL%ld", ii, index );
      else
	sprintf( valkeyname, "%s%ld", name, ii );

      fits_delete_key( fptr, valkeyname, &ftstatus ); /* CFITSIO */
      ftkey = dmu_key_find( keylist, valkeyname );
      if ( ftkey && ftkey != main_ftkey )
	dmu_key_delete( keylist, ftkey );
    }
  }

  /* Free kdd */
  if ( main_ftkey )
    dmu_key_delete( keylist, main_ftkey );
  dmi_descriptor_set_kdd( dd, NULL );

  /* Renumber keys */ 
  dmi_key_delete( dd );

}


