/*                                                                
**  Copyright (C) 2003,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftckeylib.c                                                 */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/
#include "dmilib.h"
#include "ftkernelP.h"

/***************************************************************/
/*                                                             */
/* Function: ftp_composite_key_write                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*   Write the individual raw FT keys to the file.             */
/*   Returns main raw FT key                                   */
/*   corresponding to the composite DM key.                    */
/*                                                             */
/* Input parameters:                                           */
/*                                                             */
/* Output parameters:                                          */
/*                                                             */
/* Returns:                                                    */
/*                                                             */
/***************************************************************/
dmuKey* ftp_composite_key_write( dmBlock* block, long pos, const char* name, const dmDataType type, const void* value, long nvals,
  const char* unit, const char* comment )
{
  ftBlock* ftblock;

  dmuKeyList keylist;
  dmuKey*    ftkey      = NULL;
  dmuKey*    this_ftkey = NULL;
  ftKeyType  fktype;

  long maxvlen;
  long size;        /* Bytes per item */
  long entry;       /* The FTK key number (one per FITS keyword) */
  long this_entry;  /* FTK key number for array keyword element  */

  char* ptr;
  char* val_name = (char*)name;     /* Discard  const */
  char valkeyname[FITS_MAXLEN];     /*                                             */
  char keyname[FITS_MAXLEN];        /* Key name                                    */
  char genname[FITS_MAXLEN];        /* Key name, possibly amended with trailing *  */
  char typename[4];

  void* elt_value;

  long ii;
  long index = 0;
  int verbose_header;
  
  if ( !block ) 
    return 0;
  
  ftblock = (ftBlock*)dmi_block_get_kb( block );
  if ( !ftblock || !ftblock->keylist )
    return 0;
  
  keylist = ftblock->keylist;

  ftp_block_reset( block );
  fktype = ftp_datatype_to_keytype( type, name );

  /* Handle DTYPE keys here */

  /* MCD 20090615: modified behavior.  DTYPE# keys are made for
     single valued (non-array) keyword longer than 8 chars.. 
     Array keys are sent to the FITS layer as is, and will be
     converted to the HIERARCH convention if too long.
  */

  cs_copy( name, genname );

  verbose_header = ftp_verbose_header( block );

  entry = pos;
  maxvlen = 8;

  if ( nvals <= 1 )
  {
    if ( (strlen( name ) > maxvlen)  || verbose_header )
    {
      /* Look for DTYPEn keyword. If not present, use new DTYPEn */
      index = dmu_keylist_check_index( keylist, "DTYPE", genname ); 
      if ( index == 0 )
	index = dmu_keylist_check_index( keylist, "DTYPE", NULL ) + 1; 
      
      sprintf( keyname, "DTYPE%ld", index );
      ptr = genname;
      ftkey = ftp_key_write_generic( block, entry, FTK_STRING, DT_CHAR, keyname, &ptr, NULL, 
				     "Long key name - value in DVALn" );
      if ( !ftkey )
	return NULL;

      entry = ftkey->index + 1;
      if ( verbose_header )
      {
	sprintf( keyname, "DFORM%ld", index );
	sprintf( typename, "%c", ftp_datatype_to_bt_tform( type ) );
	ptr = typename;
	ftkey = ftp_key_write_generic( block, entry, FTK_STRING, DT_CHAR, keyname, &ptr, NULL,
				       "Data type for DTYPEn" );
	if ( !ftkey )
	  return NULL;

	entry = ftkey->index + 1;
	sprintf( keyname, "DUNIT%ld", index );
	ftkey = ftp_key_write_generic( block, entry, FTK_STRING, DT_CHAR, keyname, &unit, NULL,
				       "Unit for DTYPEn" );
	if ( !ftkey )
	  return NULL;

	entry = ftkey->index + 1;
      }

      if ( (long)strlen( name ) > maxvlen )
      {
	sprintf( valkeyname, "DVAL%ld", index );
	val_name = valkeyname;
      }
    }

    /* Might be writing  name  or DVALn */
    ftkey = ftp_key_write_generic( block, entry, fktype, type, val_name, value, unit, comment );
    if ( !ftkey )
      return NULL;
    
    entry = ftkey->index;

  }
  else
  {
    /* Support for array keywords, not used in practice so far        */
    /* MCD NOTE: 20090615 - aspect code has use case                  */
    /* converting to use FITS lib to generate HIERARCH keys as needed */
    size = ar_mix_type_size( type );  
    this_entry = entry;
    for ( ii = 1; ii <= nvals; ii++ )
    {
      sprintf( valkeyname, "%s%ld", name, ii );

      elt_value = (char*)value + (ii-1)*size;

      this_ftkey = ftp_key_write_generic( block, this_entry, fktype, type, valkeyname, elt_value, unit, comment );
      if ( !this_ftkey ) 
	return NULL;

      this_entry = this_ftkey->index;
      if ( ii == 1 ) 
	ftkey = this_ftkey;

      this_entry++;  
    }
  }

  return ftkey;
}


/***************************************************************/
/*                                                             */
/* Function: ftp_key_write_generic                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*   Write a key to the FITS key list and to the file.         */
/*   If the key name already exists, just update that key and  */
/*   return its number.  Doesn't affect the high level         */
/*   composite key list, so can use this routine for all       */
/*   FITS keywords.                                            */
/*   The key is written at key position number pos;            */
/*   except for pos=0, when it is appended to the end.         */
/*                                                             */
/* Input parameters:                                           */
/*                                                             */
/* Output parameters:                                          */
/*                                                             */
/* Returns:                                                    */
/*                                                             */
/***************************************************************/
dmuKey* ftp_key_write_generic( dmBlock* block, long pos, const ftKeyType ifktype, const dmDataType type, 
                         const char* keyname, const void* value, const char* unit, const char* desc )
{
  int writeable;

  dmDataset* ds;
  dmuKey*    ftkey = NULL;
  dmuKey*    old_ftkey = NULL;
  dmuKeyList keylist;

  ftBlock*  ftblock;
  fitsfile* fptr = NULL;
  
  char* name = (char*)keyname;      /* Discard const since CFITSIO not declared const */
  char  comment_buf[FITS_MAXLEN];
  char* comment;                    /* pointer to Comment */
  int status = FITS_KERNEL_OK;
  ftKeyType fktype;

  long old_entry = 0;    /* Old entry for this key name, may be 0 */
  dmBool mask;
  
  fptr = ftp_fptr( block );
  ftblock = (ftBlock*)dmi_block_get_kb( block );
  keylist = ftblock->keylist;

  /* Calculate FITS key type for caching */
  if ( ifktype == FTK_UNKNOWN )
    fktype = ftp_datatype_to_keytype( type, name );
  else
    fktype = ifktype;

  if ( fktype == FTK_COMMENT )
  {
    old_entry =0;
  }
  else  
  {
    old_ftkey = dmu_key_find( keylist, name );
    ftkey = old_ftkey;
    if ( ftkey )
      old_entry = ftkey->index;
  }

  // Code Change July 10, 2009  Bug 10802
  //   Subtract 1 from FITS_MAXLEN because copy functions need one byte
  //   free at end of buffer for null terminator. (81 char buffer - 80 char limit)

  comment = ftp_key_comment_compose( name, type, desc, unit, comment_buf, (FITS_MAXLEN-1) );
  ftp_block_reset( block );
  
  ds = dmi_block_get_ds( block );
  writeable = dmi_dataset_mode( ds );
  
  if ( !ftkey )
  {
    ftkey  = dmu_key_alloc();
    ftkey->type = FTK_KEY;
    dmu_key_insert( keylist, ftkey, pos );
  }
  mask = ftp_key_private_fits( name );
  
  dmu_key_set( ftkey, fktype, type, name, value, unit, desc, mask );
  
  /* Move to key position if needed */
  if ( old_entry == 0 && pos > 0 && pos < keylist->n )
  {
    if ( writeable )
      status = ftp_insert_key_generic( fptr, pos, name, type, value, comment );
  }
  else
  {
    if ( writeable )
      status = ftp_update_key_generic( fptr, name, type, value, comment );
  }
  
  if ( status != FITS_KERNEL_OK )
  {
    sprintf( dm_errbuf, "FITS error %d writing key %s", status, name );
    dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
    if ( !old_ftkey ) 
    {
      dmu_key_delete( keylist, ftkey );
    }
    ftkey = NULL;
  }

  return ftkey;
}

/***************************************************************/
/*                                                             */
/* Function: ftp_key_delete                                    */
/*                                                             */
/* Description:                                                */
/*   Removes the keyword from FITS file                        */
/*                                                             */
/* Input parameters:                                           */
/*                                                             */
/* Output parameters:                                          */
/*                                                             */
/* Returns:                                                    */
/*                                                             */
/***************************************************************/
void ftp_key_delete( dmBlock* block, const char* keyname )
{
  dmDataset* ds      = NULL; /* Dataset the block belongs to */
  ftBlock*  ftblock  = NULL; /* FITS kernel block            */
  fitsfile* fptr     = NULL; /* FITS file pointer            */

  char*      name  = NULL;
  dmuKey*    ftkey = NULL;
  dmuKeyList keylist;

  int writeable;
  int status = 0;

  name = (char*)keyname;

  /* Check that the block is writeable. */
  ds = dmi_block_get_ds( block );
  writeable = dmi_dataset_mode( ds );

  if ( !writeable )
    return;

  /* Get file pointer */
  fptr = ftp_fptr( block );

  /* Call FITS routine to delete the key from the file */
  fits_delete_key( fptr, name, &status );

  /* Check if the keyword is also in the block keyword list */
  ftblock = (ftBlock*)dmi_block_get_kb( block );
  if ( ftblock )
  {
    keylist = ftblock->keylist;
    ftkey   = dmu_key_find( keylist, name );
    if ( ftkey )
    {
      dmu_key_delete( keylist, ftkey );
    }
  }

}
