/*                                                                
**  Copyright (C) 2002-2012  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/***************************************************************/
/*                                                             */
/*  ftcol.c                                                    */
/*                                                             */
/*   Functions for columns                                     */
/*                                                             */
/*                                                             */
/*                                                             */
/*                                                             */
/***************************************************************/

#include "dmilib.h"
#include "ftkernelP.h"

void ftColFlush( dmDescriptor* dd );  /* Suppress a compiler warning */


/***************************************************************/
/*                                                             */
/* Function: ftp_col_alloc                                     */
/*                                                             */
/* Description:                                                */
/*   Allocate memory for an ftCol object.                      */
/*   Initialize type value for FITS kernel column FTK_COL      */
/*                                                             */
/*  Input Parameters:                                          */
/*           none                                              */
/*                                                             */
/*  Output Parameters:                                         */
/*           none                                              */
/* Returns:                                                    */
/*          ftCol  *col - pointer to col                       */
/*                                                             */
/***************************************************************/
ftCol* ftp_col_alloc( void )
{
  ftCol *col = NULL;

  col = (ftCol*)calloc( 1, sizeof( ftCol ) );
  col->type = FTK_COL;

  return col;
}


/***************************************************************/
/*                                                             */
/* Function: ftp_col_free                                      */
/*                                                             */
/* Description:                                                */
/*   Free ftCol memory                                         */
/*                                                             */
/* Input Parameters:                                           */
/*          ftCol   *col - pointer to ftCol object to free     */
/*                                                             */
/*  Output Parameters:                                         */
/*          none                                               */
/*                                                             */
/* Returns:                                                    */
/*          void                                               */
/*                                                             */
/***************************************************************/
void ftp_col_free( ftCol* col )
{
  if ( !col ) return;

  free( col );
}


/***************************************************************/
/* Function: ftp_col_read                                      */
/*                                                             */
/* Description:                                                */
/*   Low level read properties of column from FITS raw block   */
/*   Creates the unscaled and scaled descriptor pair and       */
/*   returns the unscaled one.                                 */
/*                                                             */
/*   Code Change Bug 12317 May 27, 2010 for aspect performance */
/*     Instead of searching the keylist for column description */
/*     keys with each pass, generate a temporary list of all   */
/*     keys associated with column definition. Access this     */
/*     list when creating columns. When all columns have been  */
/*     defined, delete the temporary key list                  */
/*                                                             */
/*      Note that the temporary list is a list of dmuKey* and  */
/*      not a dmuKeyList structure                             */
/*                                                             */
/*  Input Parameters:                                          */
/*                 dmBlock   *block - DM Block descriptor      */
/*           const long       no    - Column number to read    */
/*                 dmuKey   **ppColKeysList                    */
/*                                                             */
/* Returns:                                                    */
/*          dmDescriptor *dd - DM Column descriptor pointer    */
/*                             NULL if problem defining column */
/*                                                             */
/***************************************************************/
dmDescriptor* ftp_col_read( dmBlock* block, const long no,
                            dmuKey **ppColKeysList )
{
  ftBlock*      ftblock = NULL;    /* FITS block pointer         */
  fitsfile*     fptr    = NULL;    /* FITS file pointer          */
  ftCol*        ftcol   = NULL;    /* FITS column pointer        */

  dmuKey       *ftkey   = NULL;    /* DM Key descriptor          */
  dmDescriptor *dd      = NULL;    /* New scaled column descr.   */
  dmDescriptor *us_dd   = NULL;    /* unscaled partner           */
  dmDataType    type    = DT_UNK;  /* Data type of raw column    */
  dmDataType    scaled_type = DT_UNK; /* Data type of scaled column */
  dmElementType etype   = dmVALUE; /* Element type               */
  dmuKeyList    keylist;           /* DM Keyword list            */


  char ttype[FITS_MAXLEN];         /* Buffers for FITS key values */
  char tunit[FITS_MAXLEN];
  char comment[FITS_MAXLEN];
  char tform[FITS_MAXLEN];
  char tdim[FITS_MAXLEN];
  char tdisp[FITS_MAXLEN] = " ";
  char rawname[CARD_SIZE];

  char *name = NULL;               /* Column name                */
  char *desc = NULL;               /* Column description         */
  char *unit = NULL;               /* Column units               */
  char *us_unit = NULL;            /* Unit for unscaled descriptor */

  long *axes   = NULL;             /* Pointer to axis lengths    */
  long  naxes  = 1;                /* Dimensionality of data     */
  long  size   = 0;                /* Column array size          */
  long  length = 0;                /* Size of each array cell?   */
  long  dim    = 1;
  long  bitpix = 0; 
  long  flag = DM_FLAG_RAW;
  long  numColKeys = 12;
  long  ii = 0;

  int ftstatus  = FITS_KERNEL_OK;
  int is_var    = UT_FALSE;        /* Flag for variable length array */
  int is_image  = UT_FALSE;        /* Flag for image block           */
  int is_scaled = UT_FALSE;;
  int force_unscaled = UT_FALSE;
  int ascii = UT_FALSE;            /* Flag for ascii table           */

  double pix   = 0.0;
  double tscal = 1.0;
  double tzero = 0.0;


  /* Get keylist from block */
  ftblock = (ftBlock*)dmi_block_get_kb( block);
  keylist = ftblock->keylist;

  /* Check block type */
  is_image = ( dmi_block_get_raw_type( block ) == dmIMAGE );
  ascii =  ftblock->hdutype == ASCII_TBL;
  
  /* Allocate a column descriptor */
  ftcol = ftp_col_alloc();
  ftcol->no = no;   
  
  if ( is_image )
  {
    flag |= DM_FLAG_IMG;
    name = cs_dup( dmi_block_get_name( block ) );
    
    naxes = ftblock->naxes;
    if ( naxes > 0 )
    {
      axes  = malloc( naxes * UT_SZ_I );
  
      for ( ii = 0; ii < naxes; ii++ )
	axes[ii] = ftblock->axes[ii];
    }

    /* Check for scaling keywords */
    ftkey = dmu_keylist_read_typed_value( keylist, "BZERO", DT_REAL8, &tzero);

    if ( !ftkey )
    {
      tzero = 0.0;
      ftstatus = FITS_KERNEL_OK;
    }

    ftkey = dmu_keylist_read_typed_value( keylist, "BSCALE", DT_REAL8, &tscal );
    if ( !ftkey )
    {
      tscal = 1.0;
      ftstatus = FITS_KERNEL_OK;
    }

    /* Determine data type */
    ftkey= dmu_keylist_read_typed_value( keylist, "BITPIX", DT_INT4, &bitpix );
    type = ftp_bitpix_to_type( bitpix, tscal, tzero );
    
    /* Let FITS do the unsigned offset for images */
    if ( (type == DT_UINT4) && (tscal == 1.0) && (tzero == ULONG_TZERO) )
    {
      tscal = 1.0;
      tzero = 0.0;
    } 
    else if ( type == DT_UINT2 && tscal == 1.0 && tzero == USHORT_TZERO )
    {
      tscal = 1.0;
      tzero = 0.0;
    }

    ftkey = dmu_keylist_read_value_c( keylist, "BUNIT", tunit, FITS_MAXLEN );
    if ( ftkey )
      unit = cs_dup( tunit );

    ftstatus = UT_TRUE;
    
  }
  else 
  {
    // Code Changes  Bug 12317  May 27, 2010
    //  The following code section uses column keys list, ppColKeysList, to
    //  find column information keys instead of linearly searching the full 
    //  keylist for needed column information keys
    //  Change is for table blocks only

    if ( ! ppColKeysList )
      return NULL;

    /*   --- Get column format ---     */
   
    ii = (no-1)*numColKeys+2;
    ftkey = ppColKeysList[ii];

    if ( ! ftkey )
      return NULL;

    dmu_key_read_value_c(ftkey, FITS_MAXLEN, tform );


    type = ftp_parse_tform( tform, ascii, &size, &length, &is_var );

    /* Handle case for VLAs where max size is unspecified */
    if ( is_var && size == 0 )
    {
      size = ftp_col_read_max_varsize( block, no );
    }


    ii = (no-1)*numColKeys+3;
    ftkey = ppColKeysList[ii];

    if ( ftkey )
    {
      dmu_key_read_value_c(ftkey, FITS_MAXLEN, tdim );

      /* Get TDIM information */
      ftp_parse_tdim( tdim, &axes, &naxes );

      if ( type == DT_CHAR )
      {
        /* For DT_CHAR types only - special case                 */
        /* Adjust "length" based on TDIM keyword values          */
        /*   to imitate action based on standard format ( wAr )  */
        /* NOTE: supported by cfitsio version >= 3.04            */

        if ( (naxes > 0) && (axes[0] > 0) )
        {
          size    = length / axes[0];
          length  = axes[0];
          axes[0] = size;
          naxes   = 1;
        }
      }
    }
    else
      // No TDIM keyword
      //  If size is greater than one, set axes value
      if ( size > 1 )
      {
        naxes = 1;
        axes  = calloc( 1, UT_SZ_D );
        axes[0] = size;
      }
    
    /*   --- Get column name ---     */
    
    ii = (no-1)*numColKeys+4;
    ftkey = ppColKeysList[ii];

    if ( ftkey )
    {
      dmu_key_read_value_c(ftkey, FITS_MAXLEN, ttype );
      cs_copy( ftkey->desc, comment );
    }
    else
    {
      sprintf( ttype, "COL%ld", no );
    }

    ftstatus = UT_TRUE;
    
    name = cs_dup( ttype );
    if ( strcmp( comment, "label for field" ) )  /* CF's default string */
    {
      desc = cs_dup( comment );
    }
    else
    {
      desc = NULL;
    }
    
    /*   --- Get column units ---     */
  
    ii = (no-1)*numColKeys+5; 
    ftkey = ppColKeysList[ii];

    if ( ftkey )
    {
      dmu_key_read_value_c(ftkey, FITS_MAXLEN, tunit );
      unit = cs_dup( tunit );
      ftstatus = UT_TRUE;
    }


    /*   --- Get scaling ---     */

    ii = (no-1)*numColKeys+6;
    ftkey = ppColKeysList[ii];

    if ( ftkey )
      dmu_key_typed_value(ftkey, DT_REAL8, (void*)&tscal);

    ii = (no-1)*numColKeys+7;
    ftkey = ppColKeysList[ii];

    if (ftkey )
      dmu_key_typed_value(ftkey, DT_REAL8, (void*)&tzero);


    /* Let FITS do the unsigned offset for images */
    if( (type == DT_INT4) && (tscal == 1.0) && (tzero == ULONG_TZERO) ) 
    {
      type  = DT_UINT4;
      tscal = 1.0;
      tzero = 0.0;
    }
    else if( type == DT_INT2 && tscal == 1.0 && tzero == USHORT_TZERO ) 
    {
      type  = DT_UINT2;
      tscal = 1.0;
      tzero = 0.0;
    }


    /*    --- Get column format --- */

    ii = (no-1)*numColKeys+8;
    ftkey = ppColKeysList[ii];

    if ( ftkey )
      dmu_key_read_value_c(ftkey, FITS_MAXLEN, tdisp );
   
    ftstatus = UT_TRUE;
  }

  // Set value of is_scaled according to values of BZERO/TZERO and BSCALE/TSCALE
  is_scaled = ( tscal != 1.0 || tzero != 0.0 );

  // If unsigned type, set is_scaled to false and set defaults explicitly
  // Also check for user's option to globally suppress scaling
  
  force_unscaled = dmi_block_is_unscaled( block );

  if (( force_unscaled && is_scaled ) || (type == DT_UINT2 || type == DT_UINT4 ))
  {
    is_scaled = 0;
    tscal     = 1.0;
    tzero     = 0.0;
  }
  
  /* Always suppress FITS autoscaling */
  /* Always create scaled descriptor */

  if ( 1 )
  {
    dmi_column_make_raw_name( name, no, rawname );

    /* Allocate unscaled descriptor */
    us_unit = unit;

    if ( is_scaled ) 
      us_unit = NULL;

    us_dd = dmi_descriptor_alloc( block, dmCOLUMN, no, rawname, type, us_unit, 
				    desc, dim, etype, naxes, axes, flag | DM_FLAG_LOGICAL ); 

    dmi_descriptor_set_length( us_dd, length );
    dmi_descriptor_set_kdd( us_dd, (void*)ftcol );
    dmi_descriptor_set_varlen( us_dd, is_var );

    /* Allocate scaled descriptor */
    if ( is_scaled )
    {
      scaled_type =  (( dm_global & DM_GLOBAL_OLD ) && is_image ) ? DT_REAL4: DT_REAL8;
    }
    else
    {
      scaled_type = type;
    }
    dd = dmi_create_linear_wcs( block, us_dd, name, scaled_type, tzero, pix, tscal );

    if ( is_scaled )
    {
      ftstatus = 0;
      fptr = ftp_fptr( block );

      /* Have to call set_tscale after every file flush, so this one is superfluous? */
      fits_set_tscale( fptr, no, 1.0, 0.0, &ftstatus );  /* CFITSIO */

      dmi_descriptor_set_unit( dd, unit );
      ftcol->scaled = UT_TRUE;
    }
    else 
    {
      ftcol->scaled = UT_FALSE;
    }
  }
  else 
  {
    us_dd = dmi_descriptor_alloc( block, dmCOLUMN, no, name, type, unit, 
				  desc, dim, etype, naxes, axes, flag | DM_FLAG_LOGICAL ); 
    dd = us_dd;
  }
  
  dmi_descriptor_set_length( dd, length );
  dmi_descriptor_set_kdd( dd, (void*)ftcol );
  dmi_descriptor_set_varlen( dd, is_var );

  if ( !cs_is_blank( tdisp ))
    dmi_descriptor_set_disp( dd, tdisp );
  
  /*    --- Get column range --- */

  if ( is_image )
  {
    // Image
    ftp_init_col_range( keylist, us_dd );
    ftp_init_col_bin_null( keylist, us_dd, is_image );

    if ( !is_scaled )
    {
      ftp_init_col_range( keylist, dd );
    }
  
    /* Dubious: null is really for unscaled col */
    ftp_init_col_bin_null( keylist, dd, is_image );
  }
  else
  {
    // Table Column
    //  Code Change Bug 12317  May 27, 2010
    //  Call functions that use the colums keys list, ppColKeysList
    //  instead of those that linearly search the full keylist

    ftp_init_col_range_colkeys(ppColKeysList, us_dd );
    ftp_init_col_bin_null_colkeys( ppColKeysList, us_dd );

    if ( !is_scaled )
    {
      ftp_init_col_range_colkeys( ppColKeysList, dd );
    }
  
    /* Dubious: null is really for unscaled col */
    ftp_init_col_bin_null_colkeys( ppColKeysList, dd);
  }

  if ( axes )
    free( axes );

  free( name ); 
  free( unit );
  free( desc );
  
  return us_dd;
}


/***************************************************************/
/*                                                             */
/*  Function: ftp_init_col_range                               */
/*                                                             */
/*  Description:                                               */
/*   Initialize the value range for a column                   */
/*                                                             */
/*  Input Parameters:                                          */
/*                dmuKeyList keylist - List of DM keyword      */
/*                                     descriptors             */
/*  Output Paramters:                                          */
/*                dmDescriptor *dd - updated column descriptor */
/*                                                             */
/* Returns:                                                    */
/*            void                                             */
/*                                                             */
/***************************************************************/
void ftp_init_col_range( dmuKeyList keylist, dmDescriptor* dd )
{
  long  no = 0;                  /* Column number        */
  char  keyword[FLEN_KEYWORD];   /* Keyword name         */

  void       *ptr       = NULL;
  dmuKey     *ftkey     = NULL;
  dmDataType  type;              /* Column data type     */
  GenericType tmp[2];            /* Hold min/max value   */

  int   is_time;                 /* Flag for time column */

  long  no_missing = 0;


  no   = dmi_descriptor_get_no( dd );
  type = dmi_descriptor_get_datatype( dd );

  /* If the appropriate global flag is set, use TSTART/TSTOP for the TIME range.
     There is an argument that instead TSTART/TSTOP should be considered a 
     data subspace filter, but CIAO2 did it this way.   */

  is_time = cs_uceq( dmi_descriptor_get_name( dd ), "TIME" ) && 
                                           ( dm_global & DM_GLOBAL_TSTART_RANGE );

  /* Determine minimum */ 
  ptr = &tmp[0];

  sprintf( keyword, "TLMIN%ld", no );
  ftkey = dmu_keylist_read_typed_value( keylist, keyword, type, ptr );

  if ( is_time && !ftkey )
  {
    ftkey = dmu_keylist_read_typed_value( keylist, "TSTART", type, ptr );
  }
 
  if ( !ftkey )
  {
    /* No minimum from keywords, set by data type */
    no_missing++;  
    ar_mix_type_min( type, ptr );
  }

  
  /* Determine maximum */ 
  ptr = &tmp[1];
  ftkey = NULL;

  sprintf( keyword, "TLMAX%ld", no );
  ftkey = dmu_keylist_read_typed_value( keylist, keyword, type, ptr );

  if ( is_time && !ftkey )
  {
    ftkey = dmu_keylist_read_typed_value( keylist, "TSTOP", type, ptr );
  }

  /* Suppress case TSTOP = 0 which is usually bogus */
  if( ftkey && type == DT_REAL8 )
  {
    double tstop = *(double*)ptr;

    if ( tstop == 0.0 )
    {
      ftkey = NULL;
      no_missing++;    /* <== BUG?  */
    }
  }


  if ( !ftkey )
  {
    /* No maximum from keywords, set by data type */
    no_missing++;  
    ar_mix_type_max( type, ptr );
  } 

  /* Set range in descriptor */
  if ( no_missing < 2 )
    dmi_descriptor_set_range( dd, type, &tmp[0], &tmp[1] );

}

/***************************************************************/
/*                                                             */
/*  Function: ftp_init_col_bin_null                            */
/*                                                             */
/*  Description:                                               */
/*     Check TDBIN and TDNULL keys and set in descriptor       */
/*                                                             */
/*  Input Parameters:                                          */
/*       dmuKeyList  keylist  - List of DM keyword descriptors */
/*       int         is_image - flag                           */
/*                                                             */
/*  Output Parameters:                                         */
/*       dmDescriptor  *dd    - Column descriptor to update    */
/*                                                             */
/* Returns:                                                    */
/*          void                                               */
/*                                                             */
/***************************************************************/
void ftp_init_col_bin_null( dmuKeyList keylist, dmDescriptor* dd, int is_image )
{
  dmuKey      *ftkey = NULL; 
  dmDataType   type;              /* Column data type     */
  GenericType  tmp;

  char  keyword[FLEN_KEYWORD];   /* Keyword name         */

  void  *ptr   = NULL;
  long   no;                      /* Column number        */

  no   = dmi_descriptor_get_no( dd );
  type = dmi_descriptor_get_datatype( dd );
  ptr = &tmp;


  // Get TDBIN
  sprintf( keyword, "TDBIN%ld", no );
  ftkey = dmu_keylist_read_typed_value( keylist, keyword, type, ptr );

  if ( ftkey )
    dmi_descriptor_set_bin( dd, type, ptr );
  

  // Get TNULL or TDNULL 
  ftkey = NULL;

  ftp_null_keyword_compose( type, no, keyword, is_image );
  ftkey = dmu_keylist_read_typed_value( keylist, keyword, type, ptr );

  if ( ftkey )
    dmi_descriptor_set_null( dd, type, ptr );

}


/***************************************************************/
/*                                                             */
/*  Function: ftp_create_col                                   */
/*                                                             */
/*  Description:                                               */
/*    Write a column to the FITS file.                         */
/*    We write each of its scalar components and then the      */
/*    MTYPE keys if needed.                                    */
/*                                                             */
/*  Input Parameters:                                          */
/*          dmBlock      *block  -  DM Block descriptor        */
/*          dnDescriptor *dd     -  DM Column descriptor       */
/*          long          no     -  Column number              */
/*          void          kdd    -  kernel descriptor          */
/*          char         *hint   -  hint structure             */
/*                                                             */
/*  Output Parameters:                                         */
/*          none                                               */
/*                                                             */
/* Returns:                                                    */
/*           void                                              */
/*                                                             */
/***************************************************************/
void ftp_create_col( dmBlock* block, dmDescriptor* dd, long no, void* kdd, char* hint )
{
  long dim   = 0;                    /* Dimensions of column           */
  long cptno = 0;                    /* Component counter              */
  
  ftCol*        ftcol  = NULL;   /* FITS column descriptor         */
  dmDescriptor* cpt_dd = NULL;   /* DM column component descriptor */
  dmDescriptor* us_dd  = NULL;   /* DM unscaled column descriptor  */

  dim = dmi_descriptor_get_element_dim( dd );
  for ( cptno = 1; cptno <= dim; cptno++ )
  {
    cpt_dd = dmi_descriptor_get_cpt( dd, cptno );
    ftp_create_scalar_col( block, cpt_dd, no, kdd, hint );
  }
  
  /* Get kernel descriptor for this column */
  ftcol = (ftCol*)dmi_descriptor_get_kdd( dd );
  if ( !ftcol )
  {
    /* Allocate one and set in column descriptor(s) */
    ftcol = ftp_col_alloc();
#if DEBUG
    printf( "ALLOC FOR %s FTCOL %p\n", dmi_descriptor_get_qualname( dd,dm_errbuf), ftcol );
#endif
    ftcol->no = no;   
    dmi_descriptor_set_kdd( dd, (void*)ftcol );

    us_dd = dmi_descriptor_get_unscaled( dd );
    if ( us_dd )
      dmi_descriptor_set_kdd( us_dd, (void*)ftcol );
  }

  /* Newly created columns flush MTYPEs at block close time */
  if ( dm_global & DM_GLOBAL_OLD )
  {
    ftcol->update = UT_TRUE;
  }
  else
  {
    ftcol->mtype_index = ftp_col_write_mtype( block, dd, 0 );
  }
}


/***************************************************************/
/*                                                             */
/*  Function: ftp_col_write_mtype                              */
/*                                                             */
/*  Description:                                               */
/*   Write the MTYPE keys for a descriptor.                    */
/*   If DM_GLOBAL_OLD this is delayed to block flush time for  */
/*   compatibility of output key order.                        */
/*   We return the n in MTYPEn for later use.                  */
/*                                                             */
/*  Input Parameters:                                          */
/*             dmBlock      *block  -  DM Block descriptor     */
/*             dmDescriptor *dd     -  DM Column descriptor    */
/*             long          n0     -  Starting MTYPEn index   */
/*                                     value                   */
/*                                                             */
/*  Output Parameters:                                         */
/*              none                                           */
/*                                                             */
/* Returns:                                                    */
/*        long n -Index of MTYPE keys for this column          */
/*                                                             */
/***************************************************************/
long ftp_col_write_mtype( dmBlock* block, dmDescriptor* dd, long n0 )
{
  long dim   = 0;                /* Dimensions of column           */
  long cptno = 0;                /* Component counter              */
  long n = n0;                   /* MTYPE keyword index            */

  char* ptr  = NULL;             /* Pointer to keyword value       */
  char* unit = NULL;             /* Keyword units                  */
  char* desc = NULL;             /* Keyword description            */
  char* old_desc = "DM Keyword: Descriptor name.";
  dmElementType etype;

  //fitsfile* fptr    = NULL;      /* FITS file pointer              */
  ftBlock*  ftblock = NULL;      /* FITS block pointer             */

  char keyname[CARD_SIZE];       /* Name of keyword to write       */
  char mform[CARD_SIZE];         /* Keyword values                 */
  char mtype[CARD_SIZE];
  char metyp[CARD_SIZE];
  
  dim = dmi_descriptor_get_element_dim( dd );
  if ( dim > 1 )
  { 
    //fptr = ftp_fptr( block );
    ftblock = (ftBlock*)dmi_block_get_kb( block );
    if ( n == 0 )
    {
      /* Get current max MFORM index */
      n = dmu_keyword_max_index( ftblock->keylist, "MFORM" );
      n++;
    }

    /* Write MTYPE keyword */
    sprintf( keyname, "MTYPE%ld", n );
    cs_copy( dmi_descriptor_get_name( dd ), mtype );  
    ptr = mtype;
    desc = dmi_descriptor_get_desc( dd );
    if ( dm_global & DM_GLOBAL_OLD )
      desc = old_desc;
    ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &ptr, NULL, desc );
    
    /* Write MFORM keyword */
    sprintf( keyname, "MFORM%ld", n );
    ptr = mform;
    cptno = 1;
    ptr += sprintf( ptr, "%s", dmi_descriptor_get_cpt_name( dd, cptno ) );
    for ( cptno = 2 ; cptno <= dim; cptno++ )
      ptr += sprintf( ptr, ",%s", dmi_descriptor_get_cpt_name( dd, cptno ));

    ptr  = mform;
    unit = dmi_descriptor_get_unit( dd );
    ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &ptr, unit, NULL );
    
    if ( dm_global & DM_GLOBAL_OLD )
      return n;
    
    /* Write METYP keyword  - not used yet */
    etype = dmi_descriptor_get_element_type( dd );
    if ( etype != dmVALUE )
    {
      sprintf( keyname, "METYP%ld", n );
      dmu_etype_name( etype, metyp, CARD_SIZE );
      ptr = metyp;
      ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &ptr, 
                                                   NULL, "DM Element Type" );
    }    
  }

  return n;
}


/***************************************************************/
/*                                                             */
/*  Function: ftp_create_scalar_col                            */
/*                                                             */
/*  Description:                                               */
/*     Write a scalar column.                                  */
/*                                                             */
/*  Note:                                                      */
/*   The kdd argument is a ftcol from another descriptor.      */
/*   It is used so that when copying from one FITS file        */
/*   to another we can propagate FITS specific attributes      */
/*   not forming part of the DM.  The DMP layer should ensure  */
/*   that this argument is NULL if the parent descriptor is    */
/*   from another kernel.                                      */
/*                                                             */
/*  Code Change  Bug 11882  April 27, 2011                     */
/*    This function now inserts place holder keys into the     */
/*    keylist when it writes to the FITS file to preserve      */
/*    keyword order in header.                                 */
/*                                                             */
/*  Input Parameters:                                          */
/*            dmBlock       *block -  DM Block descriptor      */
/*            dmDescriptor  *dd    -  DM Column descriptor     */
/*            long           no    -  Column number            */
/*            void          *kdd   -  kernel descriptor        */
/*            char          *hint  -  hint structure           */
/*                                                             */
/*  Output Parameters:                                         */
/*              none                                           */
/*                                                             */
/* Returns:                                                    */
/*           void                                              */
/*                                                             */
/***************************************************************/
void ftp_create_scalar_col( dmBlock* block, dmDescriptor* dd, long no, void* kdd, char* hint )
{

  ftBlock      *ftblock = NULL;    /* FITS block pointer         */
  fitsfile     *fptr    = NULL;    /* FITS file pointer          */
  ftCol        *ftcol   = NULL;    /* FITS column pointer        */

  dmDescriptor *us_dd   = NULL;    /* Unscaled column descriptor */
  dmDataType    type    = DT_UNK;  /* Column data type           */

  char* name = NULL;               /* Column name                */
  char* desc = NULL;               /* Column description         */
  char* unit = NULL;               /* Column units               */
  char* disp = NULL;               /* Column display format      */

  long* axes = NULL;               /* Pointer to axis lengths    */
  long  naxes;                     /* Dimensionality of data     */
  long  size;                      /* Column array size          */
  long  length;                    /* Size of each array cell?   */

  char* ptr  = NULL;
  char keyname[CARD_SIZE];
  char tform[CARD_SIZE];           /* Used for TFORM and TDIM    */

  int var   = UT_FALSE; /* Flag for variable array column */
  int ascii = UT_FALSE; /* Flag for ascii table           */

  int ftstatus = 0; 

  long ii;
  unsigned long one   = 1;
  unsigned long tzero = 0;
  long buffer_nrows   = 1;
  
  (void) kdd;
  (void) hint;


  /* return immediately if nothing to work with */

  if ( !block || !dd || ( !dmi_descriptor_check( dd )) )
    return;

  /* Check if this is a variable length column */
  var = dmi_descriptor_is_varlen( dd );
  
  /* Get FITS file info */
  ftp_block_reset( block ); 
  ftblock = (ftBlock*)dmi_block_get_kb( block );
  fptr    = ftp_fptr( block );

  if ( ftblock->hdutype == ASCII_TBL) 
    ascii = UT_TRUE;

  /* increment number of columns and make a space for it */
  ftblock->ncols++;

  if ( no == 0 )
    no = ftblock->ncols;

  ftblock->cols = realloc( ftblock->cols, ftblock->ncols * UT_SZ_P );
  
  /* Put new column in its place. */ 
  for ( ii = ftblock->ncols-1; ii > no; ii-- )
  {
    ftblock->cols[ii] = ftblock->cols[ii-1];
  }
  ftblock->cols[no-1] = dd;
  
  /* Get or create FITS column descriptor */
  ftcol = (ftCol*)dmi_descriptor_get_kdd( dd );

  if ( !ftcol )
  {
    ftcol = ftp_col_alloc();
    ftcol->no = no;   
    dmi_descriptor_set_kdd( dd, (void*)ftcol );
  }
  
  us_dd = dmi_descriptor_get_unscaled( dd );

  if ( us_dd )
    dmi_descriptor_set_kdd( us_dd, (void*)ftcol);
  

  /* Get Column information */
  name   = dmi_descriptor_get_name( dd );
  type   = dmi_descriptor_get_datatype( dd );
  size   = dmi_descriptor_get_array_size( dd );
  length = dmi_descriptor_get_length( dd );
  disp   = dmi_descriptor_get_disp( dd );
  
  if ( ascii )
    ftp_compose_at_tform( type, size, length, disp, tform );
  else
    ftp_compose_bt_tform( type, size, length, var, tform );
  
  if ( dmverbose > 1 && var )
  {
    fprintf( stderr, "Creating variable array column %s\n", name );
  }

  /* The following keywords support unsigned integers */
  switch ( type )
  {
  case DT_UINT4:
    tzero = (unsigned long)INT_MAX + 1;
    break;
  case  DT_UINT2:
    tzero = (unsigned long)SHRT_MAX + 1;
    break;
  default:
    tzero = 0;
    break;
  }

  if ( tzero > 0 )
  {
    /* Write the TZERO,TSCAL keys */
    sprintf( keyname, "TZERO%ld", no );

    ftp_key_write_generic( block, 0, FTK_UNKNOWN, DT_UINT4, keyname, (void*) &tzero, NULL, "offset for unsigned integers" );

    sprintf( keyname, "TSCAL%ld", no );
    ftp_key_write_generic( block, 0, FTK_UNKNOWN, DT_UINT4, keyname, (void*) &one, NULL, "data are not scaled"  );

  }

  /* Insert the column into FITS file */
  fits_insert_col( fptr, no, name, tform, &ftstatus );     /* CFITSIO */
  
  /* Add the comment and put key on kernel keylist */
  desc   = dmi_descriptor_get_desc( dd );

  sprintf( keyname, "TTYPE%ld", no );
  ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &name, NULL, desc );

  sprintf( keyname, "TFORM%ld", no );
  ptr = tform;
  ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &ptr, NULL, NULL );

  /* Add the units keyword */
  unit = dmi_descriptor_get_unit( dd );

  if ( unit && *unit )
  {
    sprintf( keyname, "TUNIT%ld", no );
    ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &unit, NULL, NULL );
  }

  /* Add TDIM keywords */
  naxes = dmi_descriptor_get_array_dimensions( dd, &axes );

  if ( naxes > 1 )
  {
    sprintf( keyname, "TDIM%ld", no );
    ptr = tform;
    ptr += sprintf( ptr, "(%ld", axes[0] );
    for ( ii = 1; ii < naxes; ii++ )
      ptr += sprintf( ptr, ",%ld", axes[ii] );
    strcat( tform, ")" );
    ptr = tform;

    ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &ptr, NULL, NULL );
  
  }

  /* Add TDISP keyword */

  if ( disp )
  {
    sprintf( keyname, "TDISP%ld", no );

    strcpy( tform, disp);
    ptr = tform;
    ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &ptr, NULL, NULL );
  }


  if ( ftstatus != FITS_KERNEL_OK )
  {
    sprintf( dm_errbuf, "FITS error %d returned while creating column", ftstatus );
    dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
  } 

  if ( axes )
    free( axes );

  fits_flush_file( fptr, &ftstatus );                           /* CFITSIO */
  
  /* Revise recommended buffer size based on new col */
  fits_get_rowsize( fptr, &buffer_nrows, &ftstatus );           /* CFITSIO */
  ftblock->buffer_nrows = buffer_nrows;

}


/***************************************************************/
/*                                                             */
/*  Function: ftp_image_col_create                             */
/*                                                             */
/*  Description:                                               */
/*    Creates a column descriptor for an image block.          */
/*                                                             */
/*  Note:                                                      */
/*    The NAXIS keys are flushed in ftp_image_reset, called in */
/*    block flush after column create is complete.             */
/*                                                             */
/*  Input Parameters:                                          */
/*           dmBLock       *block -  DM Block descriptor       */
/*           dmDescriptor  *dd    -  DM Column descriptor      */
/*           long           no    -  Column number             */
/*           void          *kdd   -  kernel descriptor         */
/*           char          *hint  -  hint structure            */
/*                                                             */
/*  Output Parameters:                                         */
/*             none                                            */
/*                                                             */
/*  Returns:                                                   */
/*              void                                           */
/*                                                             */
/***************************************************************/
void ftp_image_col_create( dmBlock* block, dmDescriptor* dd, void* kdd, char* hint )
{
  ftBlock      *ftblock = NULL;    /* FITS block pointer         */
  ftCol        *ftcol   = NULL;    /* FITS column pointer        */
  dmDescriptor *scaled_dd = NULL;  /* Scaled column descriptor   */

  long no = 1;                     /* Column number              */

  ftblock = (ftBlock*)dmi_block_get_kb( block );

  /* Force unscaled and scaled dd */
  scaled_dd = dd;
  dd = dmi_descriptor_get_unscaled( scaled_dd );
  if ( !dd ) 
    dd = scaled_dd;
  scaled_dd = dmi_descriptor_get_scaled( dd );

  /* Allocate and set a FITS column descriptor */
  ftcol = ftp_col_alloc();
  ftcol->no = no;   
  dmi_descriptor_set_kdd( dd, (void*)ftcol );
  dmi_descriptor_set_kdd( scaled_dd, (void*)ftcol );

  /* Should be already freed? */
  if ( ftblock->ncols > 0 )
    ftp_descriptor_free( ftblock->cols[0] );

  ftblock->ncols = 1;
  if ( !ftblock->cols )
    ftblock->cols = calloc( 1, UT_SZ_P );

  ftblock->cols[0] = dd;

  ftp_flush_image_struct( block );

  /* Hint not used  - suppress compiler warning */
  if ( hint || kdd ) return;

}


/***************************************************************/
/*                                                             */
/*  Function: ftp_descriptor_copy                              */
/*                                                             */
/*  Description:                                               */
/*    Copy kernel descriptor from dmDescriptor srcdd to        */
/*    dmDescriptor destdd.  Currently, this function doe       */
/*    nothing, since no type-specific copy functions have been */
/*    implemented, but it may be useful sometime.              */
/*                                                             */
/*  Input Parameters:                                          */
/*           dmDescriptor *srcdd - source DM Column descriptor */
/*           dmDescriptor *destdd -  destination col descriptor*/
/*                                                             */
/*  Output Parameters:                                         */
/*              none                                           */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void ftp_descriptor_copy( dmDescriptor* srcdd, dmDescriptor* destdd )
{
  void*  kdd1  = NULL;
  void*  kdd2  = NULL;
  ftAux* ftaux = NULL;

  kdd1 = (void*)dmi_descriptor_get_kdd( srcdd );

  if ( kdd1 )
  {
    ftaux = (ftAux*)kdd1;

    switch( ftaux->type )
    {
    case FTK_COL: 
      break;
    case FTK_SUBSPACE:
      break;
    case FTK_COORD:
      break;
    case FTK_KEY:
      break;
    default:
      break;
    } 
  }

  dmi_descriptor_set_kdd( destdd, kdd2 );

}

/***************************************************************/
/*                                                             */
/*  Function: ftp_descriptor_free                              */
/*                                                             */
/*  Description:                                               */
/*    Free descriptor memory                                   */
/*                                                             */
/*  Note:                                                      */
/*   ftcol is pointed to by both unscaled and scaled           */
/*   descriptors.  Have to be careful to free the ftcol only   */
/*   once.  We use the fact that the unscaled dd's have no     */
/*   'coord parent'.                                           */
/*                                                             */
/*  Input Parameters:                                          */
/*         dmDescriptor  *dd  -  DM descriptor                 */
/*                                                             */
/*  Output Parameters:                                         */
/*            none                                             */
/*                                                             */
/* Returns:                                                    */
/*            void                                             */
/*                                                             */
/***************************************************************/
void ftp_descriptor_free( dmDescriptor* dd )
{
  dmBlock *block   = NULL;    /* DM block pointer           */
  ftBlock *ftblock = NULL;    /* FITS block pointer         */
  ftAux   *ftaux   = NULL;
  void    *kdd     = NULL;    /* Kernel level descriptor    */

  long flag = 0;
  
  kdd= (void*)dmi_descriptor_get_kdd( dd );
#if DEBUG
  printf( "FTCLOSING DD %s FTC %p\n", dmi_descriptor_get_qualname( dd, dm_errbuf ), kdd);
#endif
  if ( !kdd ) return;
  
  ftaux = (ftAux*)kdd;
  switch( ftaux->type )
  {
  case FTK_COL: 
    /* This is a column descriptor */

    /* scaled and unscaled column descriptors share the same ftcol. */
    /* Free ftcol only on the unscaled descriptor.                  */
    flag = dmi_descriptor_get_flag( dd );
#if DEBUG
    printf( "DD %s FLAG %d FTC %p\n", dmi_descriptor_get_qualname( dd, dm_errbuf ),
	                                                flag & DM_FLAG_COORD, ftcol );
#endif
    if ( !( flag & DM_FLAG_COORD ))
      ftp_col_free( (ftCol*)ftaux );    
      
    /* Special code to handle case of deleting trivial image descriptor at the kernel level */
    block = dmi_descriptor_get_block( dd );
    ftblock = (ftBlock*)dmi_block_get_kb( block );
    if ( ftblock && ftblock->ncols == 1 && ftblock->cols[0] == dd )
    {
      free( ftblock->cols );
      ftblock->cols = NULL;
      ftblock->ncols = 0;
    }
    break;
  case FTK_SUBSPACE:
    ftp_ftss_free( (ftSubspaceCol*)ftaux );
    break;
  case FTK_COORD:
    ftp_coord_free( (ftCoord*)ftaux );
    break;
  case FTK_KEY:
    /* Keys are freed in the block keylist free */
    break;
  default:
    break;
  } 

  dmi_descriptor_set_kdd( dd, NULL );
}


/***************************************************************/
/*                                                             */
/*  Function: ftColFlush                                       */
/*                                                             */
/*  Description:                                               */
/*       Flush column data to file                             */
/*                                                             */
/*  Input Parameters:                                          */
/*            dmDescriptor  *dd - DM column descriptor         */
/*                                                             */
/*  Output Parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void ftColFlush( dmDescriptor* dd )
{
  ftBlock      *ftblock = NULL;    /* FITS block pointer         */
  fitsfile     *fptr    = NULL;    /* FITS file pointer          */
  ftCol        *ftcol   = NULL;    /* FITS column pointer        */

  dmBlock      *block   = NULL;    /* DM block descriptor        */
  dmuKey       *ftkey   = NULL;    /* DM Key descriptor          */
  dmuKeyList    keylist = NULL;    /* DM Keyword list            */
  dmDescriptor *cpt_dd  = NULL;    /* DM component descriptor    */

  char *name    = NULL;            /* Column name                */
  //char *vname   = NULL;            /* Column name                */
  char *comment = NULL;            /* Column description         */
  char *unit    = NULL;            /* Column units               */
  char key[CARD_SIZE];

  long colno    = 0;               /* Column number              */
  long dim      = 0;               /* Column dimensionality      */
  long cptno    = 0;               /* Column component number    */
  long mtype_index = 0;
  int  ftstatus = 0;

  /* Leave if this is in an image block */
  block = dmi_descriptor_get_block( dd );
  if ( dmi_block_get_type( block ) == dmIMAGE )
    return;

  /* Setup for this column */
  ftp_block_reset( block ); 
  ftblock = (ftBlock*)dmi_block_get_kb( block );
  fptr = ftp_fptr( block );
  
  dim = dmi_descriptor_get_vector_dim( dd );

  /* Sync up descriptor attributes with file on disk */
  for ( cptno = 1; cptno <= dim; cptno++ )
  {
    /* Get component descriptor */
    cpt_dd = dmi_descriptor_get_cpt( dd, cptno );
    ftcol = (ftCol*)dmi_descriptor_get_kdd( cpt_dd );
    if ( !ftcol ) 
      return;

    colno = ftcol->no;
    sprintf( key, "TTYPE%ld", colno );
    name = dmi_descriptor_get_name( cpt_dd );
    comment = dmi_descriptor_get_desc( cpt_dd );
    ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, key, &name, NULL, comment );
    mtype_index = ftcol->mtype_index;
    
    /* If the unit is blank, we need to check in case it was non-blank before, and delete it
     * in that case. If it is not blank, we don't need to check, just overwrite if already there.
     */
    sprintf( key, "TUNIT%ld", colno );
    unit = dmi_descriptor_get_unit( dd );
    if ( cs_is_blank( unit ))
    {
      keylist = ftblock->keylist;
      ftkey = dmu_key_find( keylist, key );
      if ( ftkey )
	dmu_key_delete( keylist, ftkey );
    }
    else 
    {
      ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, key, &unit, NULL, NULL );
    }
  }

  if ( dim > 1 && mtype_index > 0 )
  {
    //vname = dmi_descriptor_get_name( dd );   
    ftp_col_write_mtype( block, dd, mtype_index );  
  }

  /* Flush file */
  fits_flush_file( fptr, &ftstatus );  /* CFITSIO */

}


/***************************************************************/
/*                                                             */
/*  Function: ftp_col_read_max_varsize                         */
/*                                                             */
/*  Description:                                               */
/*    Determine the maximum size of a variable length array    */
/*    column from the data records                             */
/*                                                             */
/*  Input Parameters:                                          */
/*            dmBlock  *block  -  DM Block descriptor          */
/*          const long  no     -  Column number                */
/*                                                             */
/*  Output Parameters:                                         */
/*                  none                                       */
/*                                                             */
/*  Returns:                                                   */
/*            long maxsize - Maximum size of column            */
/*                                                             */
/***************************************************************/
long ftp_col_read_max_varsize( dmBlock *block, const long no )
{
  long *repeats = NULL;
  long *offsets = NULL;
  long 	maxsize = 1;
  long  maxrow  = 0;
  long 	ii      = 0;

  int ftstatus = FITS_KERNEL_OK; 
  fitsfile *fptr = ftp_fptr( block );

  /* Get number of rows in file */
  maxrow = dmi_block_get_nrows( block );

  /* Allocate space for size info */
  repeats = calloc( maxrow, sizeof(long) );
  offsets = calloc( maxrow, sizeof(long) );
  if( !repeats || !offsets ) 
    return maxsize;

  /* Read size info from file */
  fits_read_descripts( fptr, no, 1, maxrow, repeats, offsets, &ftstatus);  /* CFITSIO */
  if ( ftstatus == FITS_KERNEL_OK )
  {
    for ( ii = 0; ii < maxrow; ii++ )
    {
      if ( repeats[ii] > maxsize )
	maxsize = repeats[ii];
    }
  }

  /* Cleanup */
  free( repeats );
  free( offsets );

  return maxsize;
}


/***************************************************************/
/*                                                             */
/*  Function: ftp_init_col_range_colkeys                       */
/*                                                             */
/*  Description:                                               */
/*    Initialize the value range for a column using column     */
/*    keys list instead of standard keylist                    */
/*                                                             */
/*    Created as part of Bug 12317, aspect performance         */
/*    enhancement.                                             */
/*                                                             */
/*  Input Parameters:                                          */
/*       dmuKey   **ppColKeysList - sorted column keys list    */
/*                                                             */
/*  Output Parameters:                                         */
/*        dmDescriptor  *dd - updated column descriptor        */
/*                                                             */
/*  Returns:                                                   */
/*                                                             */
/*        void                                                 */
/*                                                             */
/***************************************************************/
void ftp_init_col_range_colkeys( dmuKey **ppColKeysList, dmDescriptor* dd )
{
  long  no = 0;                  /* Column number        */

  void       *ptr       = NULL;
  dmuKey     *ftkey     = NULL;
  dmuKey     *tstartKey = NULL;
  dmuKey     *tstopKey  = NULL;
  dmDataType  type;              /* Column data type     */
  GenericType tmp[2];            /* Hold min/max value   */

  int   is_time;                 /* Flag for time column */

  long  no_missing = 0;
  long  ii = 0;
  long  numColKeys = 12;


  no   = dmi_descriptor_get_no( dd );
  type = dmi_descriptor_get_datatype( dd );

  if (ppColKeysList)
  {
    tstartKey = ppColKeysList[0];
    tstopKey  = ppColKeysList[1];
  }

  /* If the appropriate global flag is set, use TSTART/TSTOP for the TIME range.*/
  /* There is an argument that instead TSTART/TSTOP should be considered a      */ 
  /* data subspace filter, but CIAO2 did it this way.                           */

  is_time = cs_uceq( dmi_descriptor_get_name( dd ), "TIME" ) && 
                                           ( dm_global & DM_GLOBAL_TSTART_RANGE );

  /* Determine minimum */ 
  ptr = &tmp[0];

  ii = (no-1)*numColKeys+9; // offset 9 is TLMIN
  ftkey = ppColKeysList[ii];

  if ( ftkey )
    dmu_key_typed_value(ftkey, type, ptr);

  if ( is_time && !ftkey )
  {
    if ( tstartKey )
    {
      ftkey = tstartKey;
      dmu_key_typed_value(ftkey, type, ptr);
    }
  }
 
  if ( !ftkey )
  {
    /* No minimum from keywords, set by data type */
    no_missing++;  
    ar_mix_type_min( type, ptr );
  }

  
  /* Determine maximum */ 
  ptr = &tmp[1];
  ftkey = NULL;


  ii = (no-1)*numColKeys+10; // offset 10 is TLMAX
  ftkey = ppColKeysList[ii];

  if ( ftkey )
   dmu_key_typed_value(ftkey, type, ptr);


  if ( is_time && !ftkey )
  {
    if ( tstopKey )
    {
      ftkey = tstopKey;
      dmu_key_typed_value(ftkey, type, ptr);
    }
  }

  /* Suppress case TSTOP = 0 which is usually bogus */
  if( ftkey && type == DT_REAL8 )
  {
    double tstop = *(double*)ptr;

    if ( tstop == 0.0 )
    {
      ftkey = NULL;
      no_missing++;    /* <== BUG?  */
    }
  }


  if ( !ftkey )
  {
    /* No maximum from keywords, set by data type */
    no_missing++;  
    ar_mix_type_max( type, ptr );
  } 

  /* Set range in descriptor */
  if ( no_missing < 2 )
    dmi_descriptor_set_range( dd, type, &tmp[0], &tmp[1] );

}


/***************************************************************/
/*                                                             */
/*  Function: ftp_init_col_bin_null_colkeys                    */
/*                                                             */
/*  Description:                                               */
/*    Check TDBIN and TDNULL keys and set in descriptor using  */
/*    sorted column keys list instead of standard keylist      */
/*                                                             */
/*    Created as part of Bug 12317, aspect performance         */
/*    enhancement.                                             */ 
/*                                                             */
/*  Input Parameters:                                          */
/*       dmuKey   **ppColKeysList - sorted column keys list    */
/*                                                             */
/*  Output Parameters:                                         */
/*       dmDescriptor  *dd - updated column descriptor         */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void ftp_init_col_bin_null_colkeys( dmuKey **ppColKeysList, dmDescriptor* dd )
{
  dmuKey      *ftkey = NULL; 
  dmDataType   type  = DT_UNK;    /* Column data type     */
  GenericType  tmp;

  void  *ptr   = NULL;
  long   no = 0;                  /* Column number        */
  long   ii = 0;
  long   numColKeys = 12;
 

  no   = dmi_descriptor_get_no( dd );
  type = dmi_descriptor_get_datatype( dd );
  ptr = &tmp;


  // Get TDBIN
  ii = (no-1)*numColKeys+11;
  ftkey = ppColKeysList[ii];

  if ( ftkey )
   dmu_key_typed_value(ftkey, type, ptr);

  if ( ftkey )
    dmi_descriptor_set_bin( dd, type, ptr );
  

  // Get TNULL or TDNULL 
  ftkey = NULL;

  ii = (no-1)*numColKeys+12;
  if ( type == DT_CHAR || type == DT_REAL4 || type == DT_REAL8 )
    ++ii;

  ftkey = ppColKeysList[ii];

  if ( ftkey )
   dmu_key_typed_value(ftkey, type, ptr);

  if ( ftkey )
    dmi_descriptor_set_null( dd, type, ptr );

}
