/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftcolset.c                                                  */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmilib.h"
#include "ftkernel.h"
#include "ftkernelP.h"


/***************************************************************/
/*                                                             */
/*  Function:   ftColSetRange                                  */
/*                                                             */
/*  Description:                                               */
/*        Write column range keys                              */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor  *ddp - col descriptor             */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*              dmErrCode  dmSUCCESS or dmFAILURE              */
/*                                                             */
/***************************************************************/
dmErrCode ftColSetRange( dmDescriptor* ddp )
{
  dmDescriptor *dd     = NULL;
  dmBlock      *block  = NULL;
  dmDataType    ddtype = DT_UNK;
  GenericType   buf[2];
  GenericType cpt_buf[2];

  char name[FITS_MAXLEN];
  void *tlmin     = NULL;
  void *tlmax     = NULL; 
  void *cpt_tlmin = NULL;
  void *cpt_tlmax = NULL; 
 
  long no  = 0;
  long dim = 0;
  long i   = 0;
  int  is_time = 0;
 

  if ( !ddp )
    return dmFAILURE;

  ddtype = dmi_descriptor_get_datatype( ddp );

  is_time = cs_uceq( dmi_descriptor_get_name( ddp ), "TIME" );
  if ( is_time ) return dmSUCCESS;

  tlmin = &buf[0];
  tlmax = &buf[1];

  if ( !dmi_descriptor_get_range( ddp, tlmin, tlmax ))
  {
    tlmin = NULL;
    tlmax = NULL;
  }

  block = dmi_descriptor_get_block( ddp );
  ftp_block_reset( block );

  if ( dmi_block_get_raw_type( block ) == dmIMAGE )
    return dmSUCCESS;

  dim = dmi_descriptor_get_element_dim( ddp );

  for ( i = 1; i <= dim; i++ )
  {
    dd = dmi_descriptor_get_cpt( ddp, i );
    cpt_tlmin = &cpt_buf[0];
    cpt_tlmax = &cpt_buf[1];

    if ( !dmi_descriptor_get_range( dd, cpt_tlmin, cpt_tlmax ))
    {
      cpt_tlmin = tlmin;
      cpt_tlmax = tlmax;  
    }
 
    no = ftp_col_no( dd );

    if ( cpt_tlmin && !ar_mix_is_min( ddtype, cpt_tlmin ))
    {
      sprintf( name, "TLMIN%ld", no );
      ftp_key_write_generic( block, 0, FTK_UNKNOWN, ddtype, name, cpt_tlmin, NULL, NULL );
    }

    if ( cpt_tlmax && !ar_mix_is_max( ddtype, cpt_tlmax ))
    {
      sprintf( name, "TLMAX%ld", no );
      ftp_key_write_generic( block, 0, FTK_UNKNOWN, ddtype, name, cpt_tlmax, NULL, NULL );
    }
  }
  return 0;
}


/***************************************************************/
/*                                                             */
/*  Function:   ftColSetNulle                                  */
/*                                                             */
/*  Description:                                               */
/*     Write a null value for the column.                      */
/*     Note that the DM supports the ability to define a null  */
/*     value  even for floating point columns, while FITS      */
/*     usually uses NaN to denote null values for floating     */
/*     point columns. It is recommended that NaN be used as    */
/*     the null value normally, for compatibility with normal  */
/*     FITS usage - only use non-NaN null values for a         */
/*     floating point column if you really need to.            */
/*     On the other hand, there's still lots of code that      */
/*     chokes on NaN, so there are times when it makes sense.  */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor  *ddp - col descriptor             */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*              dmErrCode  dmSUCCESS or dmFAILURE              */
/*                                                             */
/***************************************************************/
dmErrCode ftColSetNull( dmDescriptor* dd )
{
  dmDescriptor *ddc = NULL;
  dmBlock      *block = NULL;
  dmDataType    ddtype = DT_UNK;
  GenericType   buf;
  void         *value = &buf; 
  char          name[FITS_MAXLEN];

  long i   = 0;
  long no  = 0;
  long dim = 0;
  int  is_image = 0;


  /* Only write it out if the value has been set */
  if (!dmi_descriptor_get_null( dd, value ))
    return dmSUCCESS;

  block = dmi_descriptor_get_block( dd );
  /* Set null for descriptor and each component */
  ftp_block_reset( block );
  ddtype = dmi_descriptor_get_datatype( dd );
  dim = dmi_descriptor_get_element_dim( dd );
  is_image = dmi_block_get_raw_type( block ) == dmIMAGE;

  for ( i = 1; i <= dim; i++ )
  {
    ddc = dmi_descriptor_get_cpt( dd, i );
    /* Make null keyword here */
    no = ftp_col_no( ddc );
    ftp_null_keyword_compose( ddtype, no, name, is_image );
    ftp_key_write_generic( block, 0, FTK_UNKNOWN, ddtype, name, value, NULL, NULL );   
  }
  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/*  Function:   ftColSetBin                                    */
/*                                                             */
/*  Description:                                               */
/*        Write column bin key                                 */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor  *dd - col descriptor              */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*              dmErrCode  dmSUCCESS or dmFAILURE              */
/*                                                             */
/***************************************************************/
dmErrCode ftColSetBin( dmDescriptor* dd )
{
  dmBlock      *block  = NULL;
  dmDescriptor *ddc    = NULL; 
  dmDataType    ddtype = DT_UNK;
  GenericType   buf;
  void         *value = &buf;
  char          name[FITS_MAXLEN];
  long i   = 0;
  long dim = 0;
  long no  = 0;


  /* Only write it out if the value has been set */
  if ( !dmi_descriptor_get_bin( dd, value ))
    return dmSUCCESS;

  block = dmi_descriptor_get_block ( dd );
  /* Set null for descriptor and each component */
  ftp_block_reset( block );
  dim = dmi_descriptor_get_element_dim( dd );
  ddtype = dmi_descriptor_get_datatype( dd );

  for ( i = 1; i <= dim; i++ )
  {
    ddc = dmi_descriptor_get_cpt( dd, i );
    no = ftp_col_no( ddc );
    sprintf( name, "TDBIN%ld", no );
    ftp_key_write_generic( block, 0, FTK_UNKNOWN, ddtype, name, value, NULL, NULL );   
  }
  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_col_no                                     */
/*                                                             */
/*  Description:                                               */
/*      What is the raw FITS column number of this             */
/*      (assumed scalar) descriptor?                           */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor  *dd - col descriptor              */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*              long  no  - column number                      */
/*                                                             */
/***************************************************************/
long ftp_col_no( dmDescriptor* dd )
{
  ftCol *ftcol = NULL;
  long   no = 0;

  ftcol = (ftCol*)dmi_descriptor_get_kdd( dd );

  if (!ftcol)
  {
    sprintf( dm_errbuf, "NO FTCOL FOR %s\n", dmi_descriptor_get_name( dd ));
    dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
  }
  else
  {
    no = ftcol->no;
  }
  return no;
}

