/*                                                                
**  Copyright (C) 2002-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftcoord.c                                                   */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmilib.h"
#include "ftkernel.h"
#include "ftkernelP.h"


/***************************************************************/
/*                                                             */
/* Function:  ftCoordWrite                                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/* Input parameters:                                           */
/*           dmDescriptor  *dd                                 */
/*                                                             */
/* Output parameters:                                          */
/*           none                                              */
/*                                                             */
/* Returns:                                                    */
/*         void                                                */
/*                                                             */
/***************************************************************/
void ftCoordWrite( dmDescriptor *dd )
{
  dmDataset        *ds     = NULL;
  dmBlock          *block  = NULL;
  dmDescriptor     *dd_lp  = NULL;  /* Logical-to-physical descriptor */
  dmDescriptor     *cpt_dd = NULL;
  dmDescriptorType  dtype;
  
  char *ctype[dmMAXDIM];
  char *cname[dmMAXDIM];
  char *ttype = NULL;

  double crval[dmMAXDIM];
  double cdelt[dmMAXDIM];
  double crpix[dmMAXDIM];

  long dim   = 0;
  long cptno = 0;
  long size  = 80;
  long tno   = 0;
  long flag  = 0;

  int ok          = 0;
  int is_physical = 0;


  if ( !dd )
    return;

  block = dmi_descriptor_get_block( dd );
  ds = dmi_block_get_ds( block );

  if ( !dmi_dataset_mode( ds ))
  {
    dmu_err_set( DM_ERR_INTERNAL, "Attempt to write coord to readonly file" );
    return;
  }

  dmi_descriptor_clear_update( dd );

  dtype = dmi_descriptor_get_type( dd );

  if ( dtype != dmCOORD )
  {
    sprintf( dm_errbuf, "Descriptor %s is not a coord descriptor", 
	                                      dmi_descriptor_get_name( dd ));
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return;
  }

  dim = dmi_descriptor_get_element_dim( dd );
  tno = dmi_coord_get_transform_no( dd );

  /* Physical flag may not be correctly propagated to cpts - oops; check it here */
  flag = dmi_descriptor_get_flag( dd );
  is_physical = ( flag & DM_FLAG_PHYS );
 
 if ( !is_physical )
    dd_lp = dmi_coord_get_parent( dd );

  if ( dd_lp )
    dmi_coord_convert_pw_to_lw( dd, dd_lp, crpix, crval, cdelt, dim );
  else
    dmi_coord_get_transform_d( dd, crpix, crval, cdelt, dim );

  ttype = dmi_coord_get_transform_type( dd );

  if ( cs_eq( ttype, "LINEAR" )) 
    ttype = NULL;

  ok = UT_TRUE;

  for ( cptno = 1; cptno <= dim; cptno++ )
  {
    cpt_dd = dmi_descriptor_get_cpt( dd, cptno );
    cname[cptno-1] = dmi_descriptor_get_name( cpt_dd );
    ctype[cptno-1] = calloc( size, UT_SZ_C );
    ok &= dmi_wcs_make_ctype( cname[cptno-1], ttype, cptno, ctype[cptno-1], UT_FALSE );
  }

  if ( dim > 1 )
    ftp_coord_write_mtype( block, dd );

  for ( cptno = 1; cptno <= dim; cptno++ )
  {
    cpt_dd = dmi_descriptor_get_cpt( dd, cptno );
    ftp_scalar_coord_write( cpt_dd, crpix[cptno-1], crval[cptno-1], 
                     cdelt[cptno-1], ctype[cptno-1], cptno, is_physical, ok );
    free( ctype[cptno-1] );
  }
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_coord_write_mtype                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/* Input parameters:                                           */
/*           dmBlock       *block                              */
/*           dmDescriptor  *dd                                 */
/*                                                             */
/* Output parameters:                                          */
/*           none                                              */
/*                                                             */
/* Returns:                                                    */
/*         void                                                */
/*                                                             */
/***************************************************************/
void ftp_coord_write_mtype( dmBlock *block, dmDescriptor *dd )
{
  fitsfile *fptr    = NULL;
  ftBlock  *ftblock = NULL;

  char  *value = NULL;
  char  *ptr   = NULL;
  char  *unit  = NULL; 
  char   mform[CARD_SIZE];
  char   mtype[CARD_SIZE];
  char   keyname[CARD_SIZE];

  long n     = 0;
  long dim   = 0;
  long cptno = 0;


  dim = dmi_descriptor_get_element_dim( dd );
  cs_copy( dmi_descriptor_get_name( dd ), mtype );

  fptr = ftp_fptr( block );
  ftblock = (ftBlock*)dmi_block_get_kb( block );

  /* Check if this MTYPE already exists.. */
  n = dmu_keylist_check_index( ftblock->keylist, "MTYPE", mtype );
  
  if ( n == 0 )
  {
    /* There are no keywords with this for this system.. create them */
    n = dmu_keyword_max_index( ftblock->keylist, "MFORM" );
    ++n;

    /* Write MTYPE keyword */
    sprintf( keyname, "MTYPE%ld", n );
    value = &mtype[0];

    ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &value, 
			               NULL, "DM Keyword: Descriptor name." );

    /* Write MFORM keyword */
    sprintf( keyname, "MFORM%ld", n );
    cptno = 1;
    ptr  = mform;
    ptr += sprintf( ptr, "%s", dmi_descriptor_get_cpt_name( dd, cptno ) );

    for ( cptno = 2 ; cptno <= dim; cptno++ )
      ptr += sprintf( ptr, ",%s", dmi_descriptor_get_cpt_name( dd, cptno ));

    value = &mform[0];
    unit  = dmi_descriptor_get_unit( dd );

    if ( cs_is_blank(unit))
    {
      /* Back compat for regression */
      ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &value, 
			              unit, "DM Keyword: Descriptor value." );
    }
    else
    {
      ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &value, 
			                                      unit, NULL );
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_scalar_coord_write                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/* Input parameters:                                           */
/*           dmDescriptor  *dd                                 */
/*           double         crpix                              */
/*           double         crval                              */
/*           double         crdelt                             */
/*           char          *ctype                              */
/*           long           cptno                              */
/*           int            is_physical                        */
/*           int            ok                                 */
/*                                                             */
/* Output parameters:                                          */
/*           none                                              */
/*                                                             */
/* Returns:                                                    */
/*         void                                                */
/*                                                             */
/***************************************************************/
void ftp_scalar_coord_write( dmDescriptor* dd, double crpix, double crval,
                             double cdelt, char* ctype, long cptno,
                             int is_physical, int ok )
{
  ftCoord  *ftcoord  = NULL;
  fitsfile *fptr     = NULL;

  dmBlock           *block      = NULL;
  dmDescriptor      *pdd        = NULL; /* the parent descriptor */
  dmCoordTransform  *transform  = NULL;
  dmCoordTransform  *ptransform = NULL; /* the parent descriptor's transform */
  dmBlockType        type;
  dmBool             bIsICRS = dmFALSE;

  char *name   = NULL;
  char *wcstyp = "PHYSICAL";
  char *unitp  = NULL;
  char *desc   = NULL;
  char *unit   = NULL;
  char  keyname[CARD_SIZE];

  double poles[2]  = { 2 * 0.0 };
  double pcrpix[2] = { 2 * 0.0 };
  double pcrval[2] = { 2 * 0.0 };
  double pcdelt[2] = { 2 * 0.0 };
  double equinox  = 0.0;
  double theta    = 0.0;
  double pcelem   = 0.0;
  double sintheta = 0.0;
  double costheta = 0.0;

  long axisno   = 0;
  long no       = 0;
  long subno    = 0;
  long flag     = 0; 
  int  is_image = 0;
  int  axis     = 0;


  if ( !dd ) 
    return;
  
  block     = dmi_descriptor_get_block( dd );
  type      = dmi_block_get_raw_type( block );
  ftcoord   = dmi_descriptor_get_kdd( dd );
  transform = dmi_descriptor_get_aux( dd );
  pdd       = dmi_descriptor_get_parent( dd );

  if ( pdd )
    ptransform = dmi_descriptor_get_aux( pdd );
  else
    ptransform = NULL;
  
  /* If coord is part of a vector, the subno is stored with the parent */
  if ( ptransform )
    subno = dmi_coord_get_transform_no( pdd );
  else
    subno = dmi_coord_get_transform_no( dd );
  
  no     = dmi_coord_get_scalar_col_no( dd );
  axisno = dmi_descriptor_get_group_axis_no( dd );
  flag   = dmi_descriptor_get_flag( dd );

  if ( !is_physical )
    is_physical = ( flag & DM_FLAG_PHYS ) || ( ftcoord && ftcoord->sub == 'P' );

  is_image = ( type == dmIMAGE );
  
  if ( !ftcoord )
  {
    ftcoord = ftp_coord_alloc();
    dmi_descriptor_set_kdd( dd, (void*)ftcoord );
 
   if ( is_physical )
      ftcoord->sub = 'P';
    else if ( subno > 1 )
      ftcoord->sub =  'A' + subno - 2;
    else
      ftcoord->sub = ' ';
    
    ftcoord->axis = axisno;
 
   if ( is_image && axisno > 0 )
      ftcoord->no = 1;
    else
      ftcoord->no = no;   
  }

  axis = ( ftcoord && ftcoord->axis > 0 );

  fptr = ftp_fptr( block );
  ftp_format_ctype( "CTYPE", "TCTYP", "TCTYP", ftcoord->no, ftcoord->axis, ftcoord->sub, is_image, axis, keyname );
  
  desc = dmi_descriptor_get_desc( dd );
  unit = dmi_descriptor_get_unit( dd ); 
  unitp = unit;

  if ( dm_global & DM_GLOBAL_OLD )
    unitp = NULL;
  
  ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &ctype, unitp, desc );
  
  ftp_format_ctype( "CRVAL", "TCRVL", "TCRVL", ftcoord->no, ftcoord->axis, ftcoord->sub, is_image, axis, keyname );
  ftp_key_write_generic( block, 0, FTK_REAL, DT_REAL8, keyname, &crval, NULL, NULL );
  ftp_format_ctype( "CRPIX", "TCRPX", "TCRPX", ftcoord->no, ftcoord->axis, ftcoord->sub, is_image, axis, keyname );
  ftp_key_write_generic( block, 0, FTK_REAL, DT_REAL8, keyname, &crpix, NULL, NULL );
  ftp_format_ctype( "CDELT", "TCDLT", "TCDLT", ftcoord->no, ftcoord->axis, ftcoord->sub, is_image, axis, keyname );
  ftp_key_write_generic( block, 0, FTK_REAL, DT_REAL8, keyname, &cdelt, NULL, NULL );
  
  if ( !ok )
  {
    // FITS Standard Compliance  July 2010
    // Output keyword TCNA instead of TCNAM

    ftp_format_ctype( "CNAME", "TCNA", "TCNA", ftcoord->no, ftcoord->axis, ftcoord->sub, is_image, axis, keyname );
    name = dmi_descriptor_get_name( dd );
    ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &name, NULL, NULL );
  }

  if ( is_physical && is_image )
  {
    ftp_format_ctype( "WCSTY", "WCSTY", "WCSTY", ftcoord->no, ftcoord->axis, ftcoord->sub, is_image, axis, keyname );
    ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &wcstyp, NULL, NULL );
  }

  if ( unit && *unit )
  {
    ftp_format_ctype( "CUNIT", "TCUNI", "TCUNI", ftcoord->no, ftcoord->axis, ftcoord->sub, is_image, axis, keyname );
    ok = ! ( ( dm_global & DM_GLOBAL_OLD ) && is_physical && cs_eq( unit, "pixel" ));

    if ( ok )
      ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &unit, NULL, NULL );
  }
  
  if ( ptransform )
    theta = dmi_coord_get_theta( ptransform );
  else
    theta = 0.0;
  
  /* IRAF and DS9 compatibility */
  if ( is_image && is_physical && cdelt != 0.0 )
  {
    double ltv = 0.0;
    double ltm = 0.0;

    ltv = crpix - crval / cdelt;
    sprintf( keyname, "LTV%ld", ftcoord->axis );  
    ftp_key_write_generic( block, 0, FTK_REAL, DT_REAL8, keyname, &ltv, NULL, NULL );  
    
    /* If there's a rotation, we write out the keywords for the full */
    /* LTM matrix along with PC below                                */

    if ( fabs(theta) <= DBL_EPSILON )
    {
      ltm = 1.0 / cdelt;
      sprintf( keyname, "LTM%ld_%ld", ftcoord->axis, ftcoord->axis );  
      ftp_key_write_generic( block, 0, FTK_REAL, DT_REAL8, keyname, &ltm, NULL, NULL );  
    }
  }

  /* the equinox, longpole, latpole, crota, ... transforms don't officially */
  /* live in the vector component descriptors, but rather the parent vector */
  /* descriptor.                                                            */

  if ( cptno == 1 )
  {
    equinox = dmi_coord_get_equinox( ptransform );

    // Code Change  FITS Standards  September 15, 2010
    // Need to check value of RADESYS in order before outputting EQUI# key

    bIsICRS = ftp_radesys_equinox_validation( block, ftcoord->no );


    if ( !bIsICRS && equinox != 0 )
    {
      ftp_format_ctype( "EQUINOX", "EQUI", "EQUI", ftcoord->no, -1, ftcoord->sub, is_image, axis, keyname );
      ftp_key_write_generic( block, 0, FTK_REAL, DT_REAL8, keyname, &equinox, NULL, NULL );
    }


    dmi_coord_get_poles( ptransform, poles );

    if ( poles[0] != 0.0 )
    {
      ftp_format_ctype( "LONGPOLE", "LONP", "LONP", ftcoord->no, -1, ftcoord->sub, is_image, axis, keyname );
      ftp_key_write_generic( block, 0, FTK_REAL, DT_REAL8, keyname, &poles[0], NULL, NULL );
    }

    if ( poles[1] != 0.0 )
    {
      ftp_format_ctype( "LATPOLE", "LATP", "LATP", ftcoord->no, -1, ftcoord->sub, is_image, axis, keyname );
      ftp_key_write_generic( block, 0, FTK_REAL, DT_REAL8, keyname, &poles[1], NULL, NULL );
    }
  }

  if ( cptno == 2 && ptransform )
  {
    /* If there's a rotation angle for this coord's parent vector, */
    /* write it out in PC format                                   */

    if ( fabs(theta) > DBL_EPSILON )
    {
      /* Degrees to radians */
      theta = UT_DEG_RAD * theta;

      costheta = dmcos(theta);
      sintheta = dmsin(theta);

      /* Only need parent's cdelts, but it doesn't seem worth the trouble */
      /* to write a separate accessor function                            */

      dmi_coord_get_transform_d( pdd, pcrpix, pcrval, pcdelt, 2 );
      
      /* Using ftcoord->no to get the keyword subscripts only works if       */
      /*   (1) the columns/axes are adjacent in the FITS file, and           */
      /*   (2) component 2 comes after component 1 in the FITS file.         */
      /* For a vector column (which is what we're dealing with), these seem  */
      /* like sensible assumptions.                                          */
      
      /* PC1_1 */
      ftp_format_pc( "PC", "PC", "TP", ftcoord->no-1, ftcoord->no-1, 
	            ftcoord->axis-1, ftcoord->axis-1, ftcoord->sub, is_image, axis, keyname );
      pcelem = costheta;
      ftp_key_write_generic( block, 0, FTK_REAL, DT_REAL8, keyname, &pcelem, NULL, NULL );
      
      /* PC1_2 */
      ftp_format_pc( "PC", "PC", "TP", ftcoord->no-1, ftcoord->no, 
		     ftcoord->axis-1, ftcoord->axis, ftcoord->sub, is_image, axis, keyname );
      pcelem = -( pcdelt[1] / pcdelt[0] ) * sintheta;
      ftp_key_write_generic( block, 0, FTK_REAL, DT_REAL8, keyname, &pcelem, NULL, NULL );

      /* PC2_1 */
      ftp_format_pc( "PC", "PC", "TP", ftcoord->no, ftcoord->no-1, 
		     ftcoord->axis, ftcoord->axis-1, ftcoord->sub, is_image, axis, keyname );
      pcelem = ( pcdelt[0] / pcdelt[1] ) * sintheta;
      ftp_key_write_generic( block, 0, FTK_REAL, DT_REAL8, keyname, &pcelem, NULL, NULL );

      /* PC2_2 */
      ftp_format_pc( "PC", "PC", "TP", ftcoord->no, ftcoord->no, 
		     ftcoord->axis, ftcoord->axis, ftcoord->sub, is_image, axis, keyname );
      pcelem = costheta;
      ftp_key_write_generic( block, 0, FTK_REAL, DT_REAL8, keyname, &pcelem, NULL, NULL );

      /* FITS standards state that a valid file can not contain BOTH PC and CD style */
      /* matrix keywords.  Actively delete CD style keywords now that we've written  */
      /* the PC versions.  NOTE, this is primarily for updating files in place since */
      /* the DM does not create or carry over CD style keywords                      */

      strcpy( keyname, "CD1_1" );
      ftp_key_delete( block, keyname );

      strcpy( keyname, "CD1_2" );
      ftp_key_delete( block, keyname );

      strcpy( keyname, "CD2_1" );
      ftp_key_delete( block, keyname );

      strcpy( keyname, "CD2_2" );
      ftp_key_delete( block, keyname );


      /* If this is a physical transform on image axes, also write out the   */
      /* rotation as an LTM matrix.                                          */

      if ( is_image && is_physical )
      {
	/* LTM1_1 */
	ftp_format_pc( "LTM", "LTM", "LTM", ftcoord->no-1, ftcoord->no-1, ftcoord->axis-1, 
		       ftcoord->axis-1, ' ', is_image, axis, keyname );
	pcelem = costheta / pcdelt[0];
	ftp_key_write_generic( block, 0, FTK_REAL, DT_REAL8, keyname, &pcelem, NULL, NULL );

	/* LTM1_2 */
	ftp_format_pc( "LTM", "LTM", "LTM", ftcoord->no-1, ftcoord->no, ftcoord->axis-1, 
		       ftcoord->axis, ' ', is_image, axis, keyname );
	pcelem = sintheta / pcdelt[0];
	ftp_key_write_generic( block, 0, FTK_REAL, DT_REAL8, keyname, &pcelem, NULL, NULL );

	/* LTM2_1 */
	ftp_format_pc( "LTM", "LTM", "LTM", ftcoord->no, ftcoord->no-1, ftcoord->axis, 
		       ftcoord->axis-1, ' ', is_image, axis, keyname );
	pcelem = -sintheta / pcdelt[1];
	ftp_key_write_generic( block, 0, FTK_REAL, DT_REAL8, keyname, &pcelem, NULL, NULL );

	/* LTM2_2 */
	ftp_format_pc( "LTM", "LTM", "LTM", ftcoord->no, ftcoord->no, ftcoord->axis, 
		       ftcoord->axis, ' ', is_image, axis, keyname );
	pcelem = costheta / pcdelt[1];
	ftp_key_write_generic( block, 0, FTK_REAL, DT_REAL8, keyname, &pcelem, NULL, NULL );
      }
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_format_ctype                                 */
/*                                                             */
/* Description:                                                */
/*  The WCS convention has keywords of the form:               */
/*    EQUINOX   (global, we flag this with axis=0)             */
/*    CTYPE4A   (image; axis 4, multiplicity tag A)            */
/*    TCTYP4A   (table; column 4, multiplicity tag A)          */
/*    4TCTY3A   (image in table: axis 4 of column 3, m. tag A) */
/*                                                             */
/* Input parameters:                                           */
/*           char   *imtext                                    */
/*           char   *timtext                                   */
/*           char   *ttext                                     */
/*           long    no                                        */
/*           long    axisno                                    */
/*           char    sub                                       */
/*           int     image                                     */
/*           int     axis                                      */
/*                                                             */
/* Output parameters:                                          */
/*           char   *keyname                                   */
/*                                                             */
/* Returns:                                                    */
/*         void                                                */
/*                                                             */
/***************************************************************/
void ftp_format_ctype( char *imtext, char *timtext, char *ttext,  long no,
                       long axisno, char sub, int image, int axis, char *keyname )
{
  if ( image )
  {
    if ( axisno < 0 )
      cs_copy( imtext, keyname );
    else
      sprintf( keyname, "%s%ld%c", imtext, axisno, sub );
  }
  else if ( axis )
    sprintf( keyname, "%ld%s%ld%c", axisno, timtext, no, sub );
  else
    sprintf( keyname, "%s%ld%c", ttext, no, sub );
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_format_ctype                                 */
/*                                                             */
/* Description:                                                */
/*   Format WCS rotation-matrix keywords of the following      */
/*   forms:                                                    */
/*    PC1_2A    (image; axes 1 and 2, multiplicity tag A)      */
/*    TP1_2A    (table; columns 1 and 2, multiplicity tag A)   */
/*    12PC7A    (image in table; axes 1 and 2 of column 7,     */
/*                multiplicity tag A)                          */
/*                                                             */
/* Input parameters:                                           */
/*           char   *imtext                                    */
/*           char   *timtext                                   */
/*           char   *ttext                                     */
/*           long    no                                        */
/*           long    no2                                       */
/*           long    axisno                                    */
/*           long    axisno2                                   */
/*           char    sub                                       */
/*           int     image                                     */
/*           int     axis                                      */
/*                                                             */
/* Output parameters:                                          */
/*           char   *keyname                                   */
/*                                                             */
/* Returns:                                                    */
/*         void                                                */
/*                                                             */
/***************************************************************/
void ftp_format_pc( char *imtext, char *timtext, char *ttext, long no,
                    long no2, long axisno, long axisno2, char sub,
                    int image, int axis, char *keyname )
{
  if ( image )
    sprintf( keyname, "%s%ld_%ld%c", imtext, axisno, axisno2, sub );
  else if ( axis )
    sprintf( keyname, "%ld%ld%s%ld%c", axisno, axisno2, timtext, no, sub );
  else
    sprintf( keyname, "%s%ld_%ld%c", ttext, no, no2, sub );
}

