/*                                                                
**  Copyright (C) 2002-2005,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftdataset.c                                                 */
/*                                                               */
/*   Functions to access datasets                                */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include <unistd.h>
#include "dmilib.h"
#include "ftkernel.h"
#include "ftkernelP.h"


/***************************************************************/
/*                                                             */
/*  Function:   ftDatasetAccess                                */
/*                                                             */
/*  Description:                                               */
/*      Test filename for existence, access perm, and valid    */
/*      FITS file.  Return true if all are true.               */
/*      If mode = "RW", test for read-write access             */
/*                                                             */
/*  Input parameters:                                          */
/*           const char  *filename - fits file                 */
/*           const char  *mode     - access mode               */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmBool  UT_TRUE or UT_FALSE                     */
/*                                                             */
/***************************************************************/
dmBool ftDatasetAccess( const char* filename, const char* mode )
{
  int       ftstatus = FITS_KERNEL_OK;
  dmBool    retval = UT_FALSE;
  fitsfile *fptr = NULL;  /* Temp handle for FITS file */
  int       ftopenmode = 0;

  if ( !filename )
  {
    dmu_err_set( DM_ERR_NULL, "[ftAccessDataset]: null filename" );
  }
  else
    if ( fio_access( filename ) >= FIO_READONLY ) /* Can read file */
    {
      ftopenmode = (cs_uceq(mode, "RW")) ? READWRITE : READONLY;
      fits_open_file( &fptr, filename, ftopenmode, &ftstatus );  /* CFITSIO */

      if ( ftstatus == FITS_KERNEL_OK )
      {
        fits_close_file( fptr, &ftstatus );                 /* CFITSIO */
        retval = UT_TRUE;
      }
    } 
    else 
    {
      sprintf( dm_errbuf, "[ftAccessDataset]: File %s does not exist or is not readable",
                filename );
      dmu_err_set( DM_ERR_NO_FILE, dm_errbuf );
    }
  return retval;
}


/***************************************************************/
/*                                                             */
/*  Function:   ftDatasetCreate                                */
/*                                                             */
/*  Description:                                               */
/*     Create a dataset. Makes an empty FITS file.             */
/*                                                             */
/*  Input parameters:                                          */
/*           const char  *filename - fits file                 */
/*           const char  *hint     - hint structure            */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmDataSet  *ds - dataset pointer                */
/*                                                             */
/***************************************************************/
dmDataset* ftDatasetCreate( const char* filename, const char* hint )
{
  dmDataset *ds   = NULL;
  ftDataset *ftds = NULL;
  fitsfile *fptr  = NULL;

  (void) hint;

  fptr = ftp_create_fits( filename );

  if ( fptr )
    ftds = ftp_dataset_alloc();

  if ( ftds )
  {
    ftds->fptr = fptr;
    ftds->name = cs_dup( filename );
    ds = dmi_dataset_alloc( filename, (dmkDataset*)ftds ); 
    dmi_dataset_set_mode( ds, UT_TRUE );
  }
  return ds;
}


/***************************************************************/
/*                                                             */
/*  Function:   ftDatasetOpen                                  */
/*                                                             */
/*  Description:                                               */
/*     Open a FITS file and return a kernel dataset handle     */
/*     to it.                                                  */
/*                                                             */
/*  Input parameters:                                          */
/*           const char  *filename - fits file                 */
/*           const char  *hint     - hint structure            */
/*           const char  *mode     - file access mode          */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             dmDataSet  *ds - dataset pointer                */
/*                                                             */
/***************************************************************/
dmDataset* ftDatasetOpen(
      const char* filename,           /* i    Name of FITS file */
      const char* hint,               /* i    Dataset hints */
      const char* mode                /* i    Mode = "R" or "RW" */
)           
{
  ftDataset *ftds = NULL;
  dmDataset *ds = NULL;
  int ftstatus = FITS_KERNEL_OK;
  int ftopenmode = READONLY;
  int status = 0;
  int dsmode = 0;

  (void) hint;

  if ( !filename )
  {
    dmu_err_set( DM_ERR_NULL, "[ftOpenDataset]: null filename" );
    return NULL;
  }
  dsmode = cs_uceq( mode, "RW" );

  /* Allocate the kernel dataset */

  ftds = ftp_dataset_alloc();
  if ( !ftds ) return NULL;
  ftds->name = cs_dup( filename );

  /* Allocate the main dataset */
  ds = dmi_dataset_alloc( filename, (dmkDataset*)ftds ); 

  /* Open the FITS file */
  dmi_dataset_set_mode( ds, dsmode );

  ftopenmode = dsmode ? READWRITE : READONLY;

  fits_open_file( &(ftds->fptr), filename, ftopenmode, &ftstatus );  /* CFITSIO */

  if ( ftstatus != FITS_KERNEL_OK )
  {
     sprintf( dm_errbuf, "[ftOpenDataset]: fits_open_file error %d on file %s", 
              ftstatus, filename );
     dmu_err_set( DM_ERR_KERNEL, dm_errbuf );

     /* check if the problem was a simple file access problem */
     status = fio_access( filename );

     switch( status )
     {
       case FIO_READWRITE:
         break;
       case FIO_READONLY:
         if ( dsmode )
         {
           strcpy( dm_errbuf, "[DM FITS kernel]: File " );
           strcat( dm_errbuf, filename );
           strcat( dm_errbuf, " is not writeable" );
           dmu_err_set( DM_ERR_NO_FILE, dm_errbuf );
         }
         dsmode = UT_FALSE;  /* Revert to readonly even if readwrite requested */    
         break;
       case FIO_NOPERM:
         strcpy( dm_errbuf, "[DM FITS kernel]: No read permission for file " );
         strcat( dm_errbuf, filename );
         dmu_err_set( DM_ERR_NO_FILE, dm_errbuf );
         return NULL;   
       case FIO_NOFILE:
       default:
         strcpy( dm_errbuf, "[DM FITS kernel]: File " );
         strcat( dm_errbuf, filename );
         strcat( dm_errbuf, " does not exist" );
         dmu_err_set( DM_ERR_NO_FILE, dm_errbuf );
         return NULL;
     }

    free( ftds->name );
    free( ftds );
    free( ds );
    return NULL;
  } 

  /* Success - scan the file and set up the list of blocks */
  ftp_dataset_scan( ds );

  return ds;
}


/***************************************************************/
/*                                                             */
/*  Function:   ftDatasetPrint                                 */
/*                                                             */
/*  Description:                                               */
/*     Debugging routine to display DM knowledge of the dataset*/
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDataset  *ds  - dataset                   */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void ftDatasetPrint( const dmDataset* ds )
{
  ftDataset *ftds = NULL;
  long       ftBlockNo;
  dmBlock    *block = NULL;
  long        n = 0;

  ftds = (ftDataset*)dmi_dataset_get_kds( ds );

  if ( !ftds )
  {
    printf( "Null dataset\n" );
    return;
   }

  printf( "Dataset %s\n", ftds->name );
  printf( "FITS file pointer: %p\n", (void*)ftds->fptr );
  printf( "Blocks: \n" );
  n = dmi_dataset_get_nv( ds );
  
  for ( ftBlockNo = 1; ftBlockNo <= n; ftBlockNo++ )
  {
    block = dmi_dataset_get_block( ds, ftBlockNo );
    if ( block )
     ftp_block_print( block );
  }
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_dataset_close                              */
/*                                                             */
/*  Description:                                               */
/*     Close and free dataset. Close kernel stuff first then DM*/
/*      stuff                                                  */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDataset  *ds  - dataset                   */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void ftp_dataset_close( dmDataset* ds, dmBool delete )
{
  ftDataset *ftds  = NULL;
  dmBlock   *block = NULL;
  long       nblocks = 0;
  long       no      = 0;
  int        ftstatus = FITS_KERNEL_OK;

  ftds = (ftDataset*)dmi_dataset_get_kds( ds );

  if ( !ds || !ftds )
    return;

  /* Flush all virtual blocks */
  nblocks = dmi_dataset_get_nv( ds );

  for ( no = 1; no <= nblocks; no++ )
  {
    block = dmi_dataset_get_block( ds, no );

    if( block )
    {
      ftp_block_close( block );
    }
  }

  fits_close_file( ftds->fptr, &ftstatus );   /* CFITSIO */

  if ( delete )
    unlink( ftds->name );  /* POSIX */

  ftp_dataset_free( ftds );
  dmi_dataset_close( ds );
}



/*
 * 
/***************************************************************/
/*                                                             */
/*  Function:   ftp_dataset_find_interesting_block_no          */
/*                                                             */
/*  Description:                                               */
/*      This routine encodes the concept of the 'interesting   */
/*      block'                                                 */
/*      Here are the rules:                                    */
/*                                                             */
/*      1) If there are no blocks, return zero.                */
/*      2) If there is only one block, try that one.           */
/*         NB [Qpoe kernel]: This lets us open a .qp as an     */
/*            image even though its block type is wrong.       */
/*      3) If the block has "GTI" in the name, it's not        */
/*         interesting (except for case 2)                     */
/*      4) If the block has "WMAP" in the name, it's not       */
/*         interesting (except for case 2)                     */
/*      5) If the block is a null primary header, it's not     */
/*         interesting.                                        */
/*      6) If we are looking for a table and the block is an   */
/*         image, it's not interesting.                        */
/*         (if we are looking for an image and the block is a  */
/*         table,  we'll do a default binning.)                */
/*      7) Otherwise, the block is interesting and we return   */
/*         the first such block.                               */
/*      8) If there are multiple blocks and none of them are   */
/*         interesting, return the first one.                  */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDataset  *ds  - dataset                   */
/*           const dmBlockType type - block type               */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             long  no - block number                         */
/*                                                             */
/***************************************************************/
long ftp_dataset_find_interesting_block_no( dmDataset* ds, const dmBlockType type )
{
  dmBlockType btype;
  ftDataset *ftds = (ftDataset*)dmi_dataset_get_kds( ds );
  char *name = NULL;
  long n   = 0;
  long no  = 0;
  int null = 0;
  int good = 0;
  int gti  = 0;
  int wmap = 0;

  n = dmi_dataset_get_n( ds );

  if ( n <= 1 )
    return n;

  for ( no = 1; no <= n; no++ )
  {
    name = dmi_dataset_get_block_name( ds, no );
    btype = dmi_dataset_get_block_type( ds, no );
    null = ( no == 1 ) && ftds->null;
    good = (  type == dmUNKNOWNBLOCK || type == dmIMAGE || type == btype );
    gti  = ( name && strstr( name, "GTI" ) != NULL );
    wmap = ( name && strstr( name, "WMAP" ) != NULL );

   if ( !gti && !wmap && good && !null )
     return no;
  }

  /* If we didn't find a non-GTI or non-WMAP block,             */ 
  /* any 'good' block is more interesting than the NULL primary */
  if ( ftds->null )
  {
    for ( no = 2; no <= n; no++ )
    {
      btype = dmi_dataset_get_block_type( ds, no );
      good = ( type == dmUNKNOWNBLOCK || type == dmIMAGE || type == btype );
      if ( good )
        return no;
    }
  }

  /* Maybe our file is only boring blocks, send the null one */
  no = 1;
  return no;
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_delete_primary_hdu                         */
/*                                                             */
/*  Description:                                               */
/*     Special routine to delete the first HDU and replace it  */
/*      by an empty primary HDU. Based loosely on the ETOOLS   */
/*      algorithm.                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDataset  *ds  - dataset                   */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            int  error_status                                */
/*                                                             */
/***************************************************************/
int ftp_delete_primary_hdu( dmDataset* ds )
{
  dmBlockType type2;
  dmBlock   *block = NULL;
  ftDataset *ftds = NULL;
  fitsfile  *fptr = NULL; 
  char      *tmpname = NULL;
  char      *name    = NULL;

  long ftblockno = 0;
  long n = 0;
  int  ftstatus = FITS_KERNEL_OK;
  int  ft_hdutype = 0;
  int  decrement = UT_TRUE;
 
 
  if ( !ds )
    return FITS_KERNEL_OK;
 
  ftds = (ftDataset*)dmi_dataset_get_kds( ds );
  n = dmi_dataset_get_n( ds );
   if ( !ftds || n == 0 )
    return FITS_KERNEL_OK;

  type2 = dmi_dataset_get_block_type( ds, 2 );
  tmpname = fio_tmpname();
  fptr = ftp_create_fits( tmpname );

  if ( type2 != dmIMAGE )
  {
    /* Correct name of deleted block */
    decrement = UT_FALSE;
    ftblockno = 1;
    name = ftp_default_hdu_name( ftblockno );
    dmi_dataset_set_block_name( ds, ftblockno, name );

    /* Create scratch dataset */
    ftp_create_dummy_image( fptr, ftblockno, "PRIMARY" );
    block = ftBlockOpen( ds, ftblockno, dmIMAGE, NULL );
    ftBlockClose( block );
  }

  ftblockno = 2;

  while( ftstatus == FITS_KERNEL_OK )
  {
    fits_movabs_hdu( fptr, ftblockno, &ft_hdutype, &ftstatus ); /* CFITSIO */
    if ( ftstatus == FITS_KERNEL_OK )
      fits_copy_hdu( ftds->fptr, fptr, 0, &ftstatus );      	/* CFITSIO */
    ++ftblockno;
  } 

  if ( ftstatus != END_OF_FILE )
    ftp_ds_status( ds, ftstatus, "[ftBlockDelete]" );
  else
    ftstatus = FITS_KERNEL_OK;

  fits_flush_file( fptr, &ftstatus );				/* CFITSIO */

  /* Delete old file */   
  fits_close_file( ftds->fptr, &ftstatus );   				/* CFITSIO */

  if ( !unlink( ftds->name ) )   /* POSIX */
  {
    sprintf( dm_errbuf, "[ftp_delete_primary_hdu]: Failed to delete %s", ftds->name );
    dmu_err_set( DM_ERR_SYS_CALL, dm_errbuf );
  }

  /* Physically copy new file to old */
  if ( !rename( tmpname, ftds->name ) )   /* POSIX */
  {
    sprintf( dm_errbuf, "[ftp_delete_primary_hdu]: Failed to rename %s to %s",
             tmpname, ftds->name );
    dmu_err_set( DM_ERR_SYS_CALL, dm_errbuf );
  }

  /* Now copy new info to original ftds and free temporary memory */
  ftds->fptr = fptr;
  ftds->current = 0;
  free( tmpname );

  return decrement; 
}

