/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftdatasetlib.c                                              */
/*                                                               */
/*   Functions that allocate, free, and create datasets          */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include <unistd.h>
#include "dmilib.h"
#include "ftkernelP.h"


/***************************************************************/
/*                                                             */
/*  Function:   ftp_dataset_alloc                              */
/*                                                             */
/*  Description:                                               */
/*            Allocate a new dataset                           */
/*                                                             */
/*  Input parameters:                                          */
/*                    none                                     */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*           ftDataset *ftds - dataset struct pointer          */
/*                                                             */
/***************************************************************/
ftDataset* ftp_dataset_alloc( void )
{
  ftDataset *ftds = NULL;

  /* Allocate the ftds and initialize it */
  ftds = (ftDataset*)calloc( 1, sizeof( ftDataset ) );

  if ( !ftds )
    return NULL;

  ftds->fptr      = NULL;
  ftds->current   = 0;
  ftds->null      = UT_FALSE;
  ftds->n = 0;
  ftds->nmax = 0;
  return ftds;
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_fptr                                       */
/*                                                             */
/*  Description:                                               */
/*            Return kernel dataset pointer given block        */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock  *block                                */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*           fitsfile  *ftds - fits file pointer               */
/*                                                             */
/***************************************************************/
fitsfile* ftp_fptr( const dmBlock* block )
{
 dmDataset *ds   = NULL;
 ftDataset *ftds = NULL;

 ds   = (dmDataset*)dmi_block_get_ds( block );
 ftds = (ftDataset*)dmi_dataset_get_kds( ds );
 if ( ftds )
  return ftds->fptr;
 else
  return NULL;
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_dataset_free                               */
/*                                                             */
/*  Description:                                               */
/*            Release dataset allocated memory                 */
/*                                                             */
/*  Input parameters:                                          */
/*              ftDataset *ftds                                */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void ftp_dataset_free( ftDataset* ftds )
{
  free( ftds->hdutype );
  free( ftds->name );
  free( ftds );
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_create_fitse                               */
/*                                                             */
/*  Description:                                               */
/*            Create fits file dataset                         */
/*                                                             */
/*  Input parameters:                                          */
/*           const  char *filename                             */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*           fitsfile * fptr - fits file pointer               */
/*                                                             */
/***************************************************************/
fitsfile* ftp_create_fits( const char* filename )
{
  fitsfile *fptr  = NULL;
  int       ftstatus = FITS_KERNEL_OK;

  if ( !filename )
  {
    dmu_err_set( DM_ERR_NULL, "[ftOpenDataset]: null filename" );
  }
  else
    if ( fio_access( filename ) >= FIO_READONLY ) 
    {
      strcpy( dm_errbuf, "[ftOpenDataset]: File " );
      strcat( dm_errbuf, filename );
      strcat( dm_errbuf, " already exists: set dmClobber" );
      dmu_err_set( DM_ERR_CLOBBER, dm_errbuf );
    }
    else
    {
      fits_create_file( &fptr, filename, &ftstatus );     /* CFITSIO */

      if ( ftstatus != FITS_KERNEL_OK )
      {
        sprintf( dm_errbuf, "[ftDatasetCreate]: FITS error %d", ftstatus );
        dmu_err_set( DM_ERR_KERNEL, dm_errbuf );
        return NULL;
      }
    }
  return fptr;
}


