/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftdtype.c                                                   */
/*                                                               */
/*   Functions that work with DTYPE style keys                   */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmilib.h"
#include "ftkernelP.h"


/***************************************************************/
/*                                                             */
/*  Function:  ftp_composite_keylist_process                   */
/*                                                             */
/*  Description:                                               */
/*     The DM level key list maps the kernel keys with the     */
/*     following extra info: long name, unit, and format.      */
/*     The keylist is updated and the processed composite key  */
/*     index is returned                                       */
/*                                                             */
/*     We support:                                             */
/*         simple key        NAME = VALUE / [UNIT]             */
/*                                                             */
/*      composite key        DTYPE1 = NAME                     */
/*                           DVAL1  = VALUE                    */
/*                           DUNIT1 = UNIT                     */
/*   and also (obsolete)                                       */
/*      short composite                                        */
/*                          DTYPE1 = NAME                      */
/*                          NAME   = VALUE                     */
/*                          DUNIT1 = UNIT                      */
/*                                                             */
/*    Since we use the "/ [unit]" convention the need for      */
/*    DUNIT1, DFORM1 has pretty much gone away, but we may     */
/*    later find other composite attributes we want to support */
/*    so the code is still here.                               */
/*                                                             */
/*    CFITSIO has since adopted an alternate long keyword name */
/*    convention using ESO's HIERARCH.                         */
/*                                                             */
/*  Input parameters:                                          */
/*               dmBlock    *block                             */
/*               dmuKeyList  keylist                           */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                dmBool  UT_TRUE or UT_FALSE                  */
/*                                                             */
/***************************************************************/
dmBool ftp_composite_keylist_process( dmBlock* block, dmuKeyList keylist )
{
  dmDescriptor *dd = NULL;
  dmuKey       *ftkey = NULL;
  dmBool        match = UT_TRUE;
  dmuVectorKey entry;
  MixArray dlist;   /* List of pointers to DTYPE-style vector keys */

  void *value   = NULL;
  char *comment = NULL;
  char *unit    = NULL;
  char *name    = NULL;
 
  long dtype_index = 0;
  long koffset = 0;
  long index = 0;
  long keyno = 0;
  int  mask  = 0;
 

  if ( !keylist )
    return UT_FALSE;

  if ( dmverbose )
    fio_tmsg( "Processing composite keys" );

  dlist = dmu_vector_key_make_list( keylist, "DTYPE", "DFORM", "DUNIT", "DVAL", 5 );

  /* Loop through FITS keys */
  for ( koffset = 0; koffset < keylist->n; koffset++ )
  {
    /* Is it indexed as a DTYPE? */
    ftkey = keylist->data[koffset];
    dtype_index = ftkey->group;
    match = UT_FALSE;
    mask = ftkey->mask;
    name = ftkey->name;

    if ( dtype_index > 0 ) 
    {
      /* Yes! (probably) - we earlier marked this key as having a DTYPEn,*/
      /* so look up the cached key number of the DTYPEn                   */
      entry = ar_mix_ptr( dlist, dtype_index-1 );

      if ( entry ) 
      {
        /* entry->name is the key number of the name key. Look it up, and look */
        /* up entry->val which is the key number of the value key.              */
        name = dmu_key_value_c( entry->name );
        ftkey = entry->val;
        index = ftkey->index;    

        if ( name )
        {
          ftp_dtype_update( entry );
          match = UT_TRUE;
        }
        else
        {  /* DTYPE must be a string key */
          dmu_err_set( DM_ERR_INTERNAL, "DTYPE mismatch" );
        }
      }
      else
      {
         sprintf( dm_errbuf, "DTYPE problem with key DTYPE%ld", dtype_index );
         dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );   
       }
    }

    /* OK, so it's not a DTYPE composite key. */
    /* Is it masked? If so, we ignore it entirely. */
    /* If not, it's a regular DM key and we add it to the list */
    /* as a simple key whose index is just the current offset.  */
    else
      if ( !mask )
      {
        match = UT_TRUE;
        index = koffset+1;

        if ( !name )
        {
          dmu_err_set( DM_ERR_INTERNAL, "Null keyword name" );
          return UT_FALSE;
        }
      }

    if ( match  )
    {
      dd = dmi_key_open( block, name );

      if ( !dd )
      {
        comment = ftkey->desc;
        unit = ftkey->unit;
        value = dmu_key_value( ftkey );
        /* Now we have all we need, so instantiate the key descriptor */
        dd = dmi_key_alloc( block, keyno, name, ftkey->datatype, value, 1, ftkey->unit, ftkey->desc );

        if ( !dd )
        {
          printf( "ERROR KEY ALLOC FAIL\n" );
        }
        dmi_descriptor_set_kdd( dd, (void*)ftkey );
      }
      else
      {
        sprintf( dm_errbuf, "Key %s present more than once in header", name );
        if ( !(dm_global & DM_GLOBAL_OLD) )  /* Don't want this warning in old regression tests */
          dmu_err_set( DM_WARN_NAME_CLASH, dm_errbuf );
      }
    }
  }

  ar_mix_free( dlist );
  return UT_TRUE;
}


/***************************************************************/
/*                                                             */
/*  Function:  ftp_composite_keylist_process                   */
/*                                                             */
/*  Description:                                               */
/*      FITS specific updates from DTYPE style keys            */
/*      This is a CXC DM convention not used elsewhere         */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmuVectorKey  entry                         */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                void                                         */
/*                                                             */
/***************************************************************/
void ftp_dtype_update(  const dmuVectorKey entry )
{
  dmDataType  ktype;
  dmDataType  type;
  dmuKey     *ftkey = NULL; 
  char       *name  = NULL;
  void       *value = NULL;
  GenericType nvalue;

  /* DFORM overrides the data type */
  if ( entry->form  ) 
  {
    name = dmu_key_value_c( entry->form );
    type = ftp_bt_tform_to_datatype( *name );
    ftkey = entry->val;
    ktype = ftkey->datatype;

    if ( type != ktype )   
    {
      value = dmu_key_value( ftkey );
      ar_mix_numeric_cast( value, ktype, type, &nvalue );    
      ar_mix_copy_value( &nvalue, type, value, UT_TRUE, 0 );
      ftkey->datatype = type;
    }
  }

  /* DUNIT overrides the unit */
  if ( entry->unit )
  {
    ftkey = entry->val;
    free( ftkey->unit );
    ftkey->unit = cs_dup( dmu_key_value_c( entry->unit ));
  }
}



