/*                                                                
**  Copyright (C) 2002-2012  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftimage.c                                                   */
/*                                                               */
/*   Routines to access images                                   */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmilib.h"
#include "ftkernelP.h"


/***************************************************************/
/*                                                             */
/* Function:   ftp_image_reset                                 */
/*                                                             */
/* Description:                                                */
/*      Call CFITSIO function fits_resize_img() to set         */
/*      CFITSIO internal size parameters. Update the BZERO     */
/*      and BSCALE keywords, if needed.                        */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock     *block  -                           */
/*      const  dmDataType   type   -                           */
/*      const  long        *axlen  -                           */
/*      const  long         naxis  -                           */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                    none                                     */
/*                                                             */
/***************************************************************/
void ftp_image_reset( dmBlock *block, const dmDataType type, 
                      const long *axlen, const long naxes )
{
  fitsfile *fptr = ftp_fptr( block );

  dmDescriptor *dd   = NULL;
  char         *unit = NULL;

  double bscale = 0.0;
  double bzero  = 0.0;

  long *lAxes    = (long *)axlen;
  int   bitpix   = 0;
  int   naxis    = (int) naxes;
  int   ftstatus = FITS_KERNEL_OK;


  // FITS Standards Compliance  July 2010
  // If file has keywords NAXIS =1, and NAXIS1 =0,
  // output keyword is NAXIS = 0 instead.
  // Do this by setting naxis = 0 and axes = NULL as
  // input to cfitsio functin fits_resize_image()
 
  if ( naxis == 1  && *axlen == 0 )
  {
    naxis = 0;
    lAxes = NULL;
  }

  /* This function returns the bitpix type to use in the CFITSIO call */
  /* along with the BSCALE/BZERO offsets needed if the image is an    */
  /* unsigned integer type.  For other types, the default is returned */
  /*                                                                  */

  bitpix = ftp_bitpix( type, &bscale, &bzero );

  /* Extract info from image descriptor */
  dd = dmi_block_get_col( block, 1 );

  if ( dd )
  {
    dmBool is_scaled = UT_FALSE;
    double crpix   = 0.0;
    double crval   = 0.0;

    /* Code Change  Bug 12880  September 7, 2011       */
    /* handling of BSCALE/BZERO as reserve keywords    */
    /* Extract scale info from image descriptor        */
    /* Only applies to non-unsigned type images        */

    /* NOTE: This is the scaled image descriptor..     */
    /*       so I think this flag is always true.      */

    is_scaled = (dmi_descriptor_get_flag( dd ) & DM_FLAG_COORD );

    if ( is_scaled && (type != dmUSHORT) && (type != dmULONG) )
    {
      /* Read any transform info from descriptor */
      dmi_coord_get_transform_d(dd, &crpix, &crval, &bscale, 1);
      bzero = crval - crpix * bscale;
    }

    /* get units string too. */
    unit = dmi_descriptor_get_unit( dd );
  }
   
  /* Resize the image */
  fits_resize_img( fptr, bitpix, naxis, lAxes, &ftstatus ); /* CFITSIO */

  /* Update scale keywords as needed */

  if ( (type == dmUSHORT) || (type == dmULONG) )
  {
    fits_update_key_fixdbl( fptr, "BZERO", bzero, 0, "Zero point for unsigned integers", &ftstatus ); /* CFITSIO */

    fits_update_key_fixdbl( fptr, "BSCALE", bscale, 0, "Scaling factor", &ftstatus );  /* CFITSIO */
  }
  else if ( ( fabs(bscale - 1.0) > DBL_EPSILON ) ||
	    ( fabs(bzero  - 0.0) > DBL_EPSILON ) )
  {
    /* non-default scale on signed type image.. */ 

    fits_update_key_fixdbl( fptr, "BZERO", bzero, FITSK_DOUBLE_DECIMALS, "Zero point ", &ftstatus ); /* CFITSIO */

    fits_update_key_fixdbl( fptr, "BSCALE", bscale, FITSK_DOUBLE_DECIMALS, "Scaling factor", &ftstatus );  /* CFITSIO */
  }

  if ( ((type == dmUSHORT) || (type == dmULONG)) &&  naxis > 0 )
  {
    /* This routine wasn't checking for a NULL pointer when           */ 
    /* dealing with NULL primary images.. causing bad memory access  */
    /* so only calling if there are axes.                            */

    fits_set_bscale( fptr, bscale, bzero, &ftstatus );  /* CFITSIO */
  }

  if ( unit && !cs_is_blank( unit ))
    fits_update_key_longstr( fptr, "BUNIT", unit, "Units after BSCALE applied", &ftstatus ); /* CFITSIO */
}




