/*                                                                
**  Copyright (C) 2002-2004,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftkernel.c                                                  */
/*                                                               */
/*   Functions that form FITS kernel interface routines          */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include <unistd.h>
#include "dmilib.h"
#include "dmkernel.h"
#include "ftkernel.h"
#include "ftkernelP.h"



/***************************************************************/
/*                                                             */
/* Function:   ftInit                                          */
/*                                                             */
/* Description:                                                */
/*             Initialization                                  */
/*                                                             */
/*  Input parameters:                                          */
/*                    none                                     */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                    nothing                                  */
/*                                                             */
/***************************************************************/
void ftInit( void )
{
  return;
}

/***************************************************************/
/*                                                             */
/* Function:   ftDatasetClose                                  */
/*                                                             */
/* Description:                                                */
/*             close dataset                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmDataset  *ds  - ptr to dataset           */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                    nothing                                  */
/*                                                             */
/***************************************************************/
void ftDatasetClose( dmDataset* ds )
{
  dmBool delete = UT_FALSE;
  ftp_dataset_close( ds, delete );
}

/***************************************************************/
/*                                                             */
/* Function:   ftDatasetDelete                                 */
/*                                                             */
/* Description:                                                */
/*             delete dataset                                  */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmDataset  *ds  - pointer to dataset       */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                    nothing                                  */
/*                                                             */
/***************************************************************/
void ftDatasetDelete( dmDataset* ds )
{
  dmBool delete = UT_TRUE;
  ftp_dataset_close( ds, delete );
}


/***************************************************************/
/*                                                             */
/* Function:   ftBlockOpen                                     */
/*                                                             */
/* Description:                                                */
/*     Open a block by number.  Special case: number = 0 does  */
/*     the 'most interesting block' of type 'type'.  The       */
/*     function ftp_block_open() does the work.                */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDataset  *ds   - pointer to dataset       */
/*                 long        no    - block number            */
/*                 dmBlockType type  - block type              */
/*                 char       *hint  - hint structure pointer  */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                 dmBlock  *block                             */
/*                                                             */
/***************************************************************/
dmBlock* ftBlockOpen( dmDataset* ds, long no, dmBlockType type, char* hint )
{
  return ftp_block_open_mode( ds, no, type, UT_TRUE, hint );
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_block_open_mode                             */
/*                                                             */
/* Description:                                                */
/*         When creating a subspace table, we open the subspace*/
/*         table block, e.g. "GTI".  That GTI block open tried */
/*         to open its own subspace, eating its own tail.  To  */
/*         avoid this, we suppress opening subspaces when      */
/*         creating a subspace block.                          */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmDataset  *ds             - dataset       */
/*                  long        no             - block number  */
/*                  dmBlockType type           - block type    */
/*                  int         open_subspace                  */
/*                  char       *hint           - hint struct   */ 
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                    nothing                                  */
/*                                                             */
/***************************************************************/
dmBlock* ftp_block_open_mode( dmDataset* ds, long no, dmBlockType type,
                              int open_subspace, char* hint )
{
  dmBool       create = UT_FALSE;
  dmBlock     *block  = NULL;
  long         n      = 0;
  dmBlockType  btype;

  if ( !ds )
    return NULL;

  if ( no == 0 )
    no = ftp_dataset_find_interesting_block_no( ds, type );

  n = dmi_dataset_get_n( ds );

  if ( no <= 0 || no > n )
  {
    sprintf( dm_errbuf, "[block open]: Bad block number %ld Nblocks=%ld", no, n );
    dmu_err_set( DM_ERR_INDEX, dm_errbuf );
    return NULL;
  }

  /* Fail if we wanted a table but didn't get it */
  btype = dmi_dataset_get_block_type( ds, no );

  if( type == dmTABLE && btype != dmTABLE )
  {
    sprintf( dm_errbuf, "[block open]: Block is not a table %ld", no );
    dmu_err_set( DM_ERR_INDEX, dm_errbuf );
    return NULL;  
  }

  /* Allocate the block here */
  block = ftp_block_open( ds, no, create, hint );

  if ( !block )
    return NULL;

  ftp_block_init( block, create );

  if ( open_subspace )
    ftp_subspace_open( block, create );

  dmi_dataset_set_current( ds, block );

  /* set hints that may affect columns here */
  if ( hint )
    ftp_set_hint_list( block, hint, dmFALSE );

  return block;
}


/***************************************************************/
/*                                                             */
/* Function:   ftBlockFlush                                    */
/*                                                             */
/* Description:                                                */
/*             flush block data                                */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock  *block - pointer to blocket       */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                dmErrCode  dmSUCCESS                         */
/*                                                             */
/***************************************************************/
dmErrCode ftBlockFlush( dmBlock* block )
{
  ftp_block_flush( block );
  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   ftBlockClose                                    */
/*                                                             */
/* Description:                                                */
/*             Close a block.  This does not destroy the       */
/*             block pointer, but releases the structures      */
/*             hanging off it.                                 */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock    *block - pointer to block       */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                  dmErrCode  dmSUCCESS                       */
/*                                                             */
/***************************************************************/
dmErrCode ftBlockClose( dmBlock* block )
{
  ftBlock *ftblock = NULL;

  if ( !block )
    return dmSUCCESS;

  ftblock = (ftBlock*)dmi_block_get_kb( block );
  ftp_block_close( block );

  return dmSUCCESS;
}




/***************************************************************/
/*                                                             */
/* Function:   ftBufferSize                                    */
/*                                                             */
/* Description:                                                */
/*             What is a sensible number of rows to try to     */
/*             read at once?  Kernel routine                   */
/*                                                             */
/*  Input parameters:                                          */
/*                  dmBlock    *block - pointer to block       */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                  long  buffer_nrows                         */
/*                                                             */
/***************************************************************/
long ftBufferSize( dmBlock* block )
{
  ftBlock *ftblock = NULL;

  ftblock = (ftBlock*) dmi_block_get_kb( block );
  return ftblock->buffer_nrows;
}


/***************************************************************/
/*                                                             */
/* Function:   ftSetHint                                       */
/*                                                             */
/* Description:                                                */
/*    Kernel call                                              */
/*                                                             */
/*  Input parameters:                                          */
/*    dmBlock    *block - pointer to block                     */
/*    const char *name  - hint element name                    */
/*    const char *value - hint element value                   */
/*                                                             */
/*  Output parameters:                                         */
/*      none                                                   */
/*                                                             */
/*  Returns:                                                   */
/*    dmSUCCESS   - option was recognized                      */
/*    other       - option was not recognized                  */
/*                                                             */
/***************************************************************/
dmErrCode ftSetHint( dmBlock* block, const char* name, const char* value )
{
  dmErrCode status = dmSUCCESS;

  status = ftp_set_hint( block, name, value );

  return status;
}


/***************************************************************/
/*                                                             */
/* Function:   ftDescriptorCopy                                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor  *srcdd  -  source descriptor      */
/*             dmDescriptor  *destdd -  destination descriptor */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                  nothing                                    */
/*                                                             */
/***************************************************************/
void ftDescriptorCopy( dmDescriptor* srcdd, dmDescriptor* destdd )
{
  ftp_descriptor_copy( srcdd, destdd );
}


/***************************************************************/
/*                                                             */
/* Function:   ftDescriptorFree                                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor  *dd  - pointer to descriptor      */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                  nothing                                    */
/*                                                             */
/***************************************************************/
void ftDescriptorFree( dmDescriptor* dd )
{
  ftp_descriptor_free( dd );
}


/***************************************************************/
/*                                                             */
/* Function:   ftImageSetAxes                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock          *block - block pointer         */
/*             const dmDataType  type  - data type             */
/*             const long       *axlen - axis length           */
/*             const long        naxes - number of axis        */ 
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                    nothing                                  */
/*                                                             */
/***************************************************************/
void ftImageSetAxes( dmBlock* block, const dmDataType type,
                     const long* axlen, const long naxes )
{
  dmDescriptor *col     = NULL;
  dmDataset    *ds      = NULL;
  ftBlock      *ftblock = NULL;

  ftblock = (ftBlock*)dmi_block_get_kb( block );
  col     = ftblock->cols[0]; /* Unscaled column */

  ftblock->naxes = naxes;  /* Save it here too for WCS mapping */

  dmi_descriptor_set_axes( col, axlen, naxes );
  dmi_descriptor_set_datatype( col, type );

  /* Scaled col */
  col = dmi_descriptor_get_scaled( col );

  if ( col )
  {
    dmi_descriptor_set_axes( col, axlen, naxes );
    dmi_descriptor_set_datatype( col, type );
  }

  ds = dmi_block_get_ds( block );

  if ( dmi_dataset_mode( ds )) /* WRITE TO FILE */
    ftp_image_reset( block, type, axlen, naxes );
}


/***************************************************************/
/*                                                             */
/* Function:   ftBLockCreate                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDataset    *ds    - dataset pointer           */
/*             long          no    - block number              */
/*             const char   *name  - block name                */
/*             dmBlockType   types - block type                */
/*             void         *kb                                */
/*             char         *hint  - hint structure pointer    */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                dmBlock  *block                              */
/*                                                             */
/***************************************************************/
dmBlock* ftBlockCreate( dmDataset* ds, long no, const char* name,
                        dmBlockType type, void* kb, char* hint )
{
  dmBlock    *block = NULL;
  ftDataset  *ftds  = NULL;
  long        nb    = 0;
  long        blockno = no;


  nb = dmi_dataset_get_n( ds );
 
  if ( nb == 0 && type == dmTABLE )
  {
    blockno = 1;

    /* Spontaneously create null primary block */
    block = ftp_block_insert( ds, "PRIMARY", blockno, dmIMAGE, NULL, NULL );
    ftds  = (ftDataset*)dmi_dataset_get_kds( ds );
    ftds->null = UT_TRUE;
    ++blockno;

    /* Now close this block */
    ftBlockClose( block );
    dmi_block_close_descriptors( block, ftDescriptorFree );
  }

  block = ftp_block_insert( ds, name, blockno, type, kb, hint );

  return block;
}


/***************************************************************/
/*                                                             */
/* Function:   ftColCreate                                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock      *block - block pointer             */
/*             dmDescriptor *dd    - descriptor pointer        */
/*             void         *kdd                               */
/*             char         *hint  - hint structure pointer    */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                nothing                                      */
/*                                                             */
/***************************************************************/
void ftColCreate( dmBlock* block, dmDescriptor* dd, void* kdd, char* hint )
{
  long         no = 0;
  dmBlockType  type;


  type = dmi_block_get_raw_type( block );

  if ( type == dmIMAGE )
  {
    ftp_image_col_create( block, dd, kdd, hint );
  }
  else
  {
    /* Call ftp_create_col with no = 0 causes it to be appended */
    ftp_create_col( block, dd, no, kdd, hint );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   ftCBlockSetPref                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock      *block - block pointer             */
/*             char        **prefcols - column pointer         */
/*             const long    npref                             */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                nothing                                      */
/*                                                             */
/***************************************************************/
void ftBlockSetPref( dmBlock* block, char** prefcols, const long npref )
{
  int    is_default = UT_FALSE;
  char  *value      = NULL;

  char   buf[CARD_SIZE];


  dmi_block_set_pref( block, (char**)prefcols, npref, is_default );
  cs_compose_paren_list( prefcols, npref, buf, CARD_SIZE );

  value = buf;

  if ( dmi_block_writeable( block ))
    ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, "CPREF", &value, NULL, NULL );
}


/***************************************************************/
/*                                                             */
/* Function:   ftBlockUpdateFilter                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmBlock      *block - block pointer             */
/*             dmSubspace   *rts   - subspace pointer          */
//*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                nothing                                      */
/*                                                             */
/***************************************************************/
void ftBlockUpdateFilter( dmBlock* block, dmSubspace* rts )
{
  long          i     = 0;
  dmDescriptor *dd    = NULL;
  dmDataset    *ds    = NULL;
  long          ncols = 0;
  long          ncpts = 0;


  ds    = (dmDataset*)dmi_block_get_ds( block );
  ncpts = dmi_subspace_get_ncpts( rts );

  if ( ncpts != 1 ) /* No rts or complicated rts */
    return;

  /* For each column in the subspace, add a filter string */
  ncols = dmi_subspace_get_ncols( rts );

  for ( i = 1; i <= ncols; i++ )
  {
    dd = dmi_subspace_col_no( rts, i );
    ftp_build_cfitsio_spec( block, dd );
  }

  ftp_block_reopen_with_filter( block );
}


