/*                                                                
**  Copyright (C) 2002-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftkernel.h                                                  */
/*                                                               */
/*   Function prototypes and definitions                         */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#ifndef FTKERNELP_H
#define FTKERNELP_H
#include "fitsio.h"
#include "cxcregion.h"


#define FITS_KERNEL_OK     0
#define FITS_COMMENT_SIZE 72
/* Default BITPIX for initial image create */
#define FITS_BITPIX       16
#define FITS_MAXDIM       99
#define FITS_MAXLEN       81
#define FITS_MAXSTRKEY    62
#define FT_SPARE_KEYS     20
#define FTK_BAD_KEYNO     -1
#define FITSK_FLOAT_DECIMALS   7
#define FITSK_DOUBLE_DECIMALS 13
#define USHORT_TZERO      32768.0
#define ULONG_TZERO  2147483648.0

/* FTK descriptor types */

#define FTK_RAW_COL   10
#define FTK_COL       20
#define FTK_IMAGE_DD  30
#define FTK_COORD     40
#define FTK_SUBSPACE  50
#define FTK_KEY       60
/*
 *  Symbols for matching MTYPE with CTYPE/TTYPE
 */
#define MT_AXIS_COL   10
#define MT_AXIS_COORD 20


#define  FITSERR207  207

/* FTK structure types */

typedef struct ftDataset_s ftDataset;
typedef struct ftBlock_s   ftBlock;
typedef struct ftCoord_s ftCoord;

typedef struct ftCol_s     ftCol;
typedef struct ftSubspaceCol_s ftSubspaceCol;

typedef struct ftAux_s {
 long type;  /* Dummy structure to allow polymorphism */
} ftAux;

/*
 *  FITS keyword types
 */
typedef enum { FTK_INT, FTK_UINT, FTK_REAL, FTK_STRING, 
               FTK_LOGICAL, FTK_COMMENT, FTK_UNKNOWN } ftKeyType;



struct ftCol_s {
 long type;
 long no;       /* FITS column no */
 int update;    /* Need to flush mtype */
 int scaled;    /* Is a scaled column: suppress TSCAL on read */
 long mtype_index; /* If a vector col cpt, what value of mtype? */
};

struct ftCoord_s {
 long type;         /* Mark as ftCoord */
 long no;          /* FITS column/axis no */
 char* wname;      /* Value of WCSTYP (local convention) */
 char* ctype_name; /* Value of FITS CTYPE */
 long axis;        /* Axis number on image */
 char sub;         /* Label for nth coord system on this axis */
 int composite;    /* Member of a vector coord system? */
 double* pc_row;   /* Row for this (transformed) column/axis in PC/CD matrix */
 int is_cd;        /* Is pc_row from a CD matrix? */
 int is_ltm;       /* Is pc_row from a LTM matrix? */
 char* partner_name;   /* Name of partner column/axis in PC/CD matrix */
};

struct ftSubspaceCol_s {
 long type;       /* Mark as ftSubspaceCol */
 long index;     /* DSTYP index number */
 long hdu;       /* Which (e.g. GTI) table is attached? */
 long written;   /* DSTYP is written? */
 long sum; /* Hack counter */
};



struct ftBlock_s {
 long hdu;             /* FITS HDU number (1-based) */
 int hdutype;         /* CFITSIO HDU type code */

/*
 *  Variables set when block is open 
 */
 dmuKeyList keylist;  /* Cached list of FITS header keys */
 long nkeys;
 long nkmax;

 double bytes;         /* Number of bytes in block data */

 long ncols;           /* Number of cols */
 long nmax;            /* Cols allocated */
 dmDescriptor** cols;           /* Column data */

 long axes[FITS_MAXDIM];    /* Temp storage for passing axes to ftpColRead */
 long naxes;           /* Number of axes for image */

 long keyno;          /* Last key number read */

 char* filter;       /* FITS open filter */
 long max_dstyp;      /* Max index for DSTYPn keyword */

 long buffer_nrows;   /* Good number of rows to read from table */
 char* vector_buffer;   /* Temp storage for reordering vectors */
 long vector_buffer_size;
 void* hint;
};



struct ftDataset_s {
 char* name;          /* Filename */
 fitsfile* fptr;      /* CFITSIO structure pointer */
 long  current;       /* Current HDU number for CFITSIO */ 
 int null;            /* Is HDU 1 a null primary? */
 int* hdutype;        /* Type of HDU */
 long n;
 long nmax;
};


/* Dataset methods */

void       ftp_dataset_close( dmDataset* ds, dmBool delete );
int        ftp_move_to_current_hdu( const dmDataset* ds );
int        ftp_delete_primary_hdu( dmDataset* ds );
void       ftp_dataset_scan( dmDataset* ds );

long       ftp_dataset_find_interesting_block_no( dmDataset* ds, const dmBlockType type );

dmBlock*   ftp_block_insert( dmDataset* ds, const char* name, long no, const dmBlockType type, void* kb, char* hint );

/* ftDataset methods */
ftDataset* ftp_dataset_alloc( void );
void       ftp_dataset_free( ftDataset* ftds );

/* Block methods */
fitsfile*  ftp_fptr( const dmBlock* block );
int ftp_block_hdutype( const dmBlock* block );
void       ftp_block_flush( dmBlock* block );
void ftp_block_init( dmBlock* block, const dmBool create );
void       ftp_block_free( dmBlock* block );
int        ftp_block_reset( const dmBlock* block );
dmErrCode  ftp_set_hint( dmBlock* block, const char* name, const char* value );
dmErrCode  ftp_set_hint_list( dmBlock* block, const char* option, dmBool bPrintErrFlag );
void       ftp_image_reset( dmBlock* block, const dmDataType type, const long* axlen, const long naxes );
void       ftp_block_close( dmBlock* block );
void       ftp_block_print_cols( const dmBlock* block );
void       ftp_block_print( const dmBlock* block );

void       ftp_block_reopen_with_filter( dmBlock* block );
void       ftp_build_cfitsio_spec( dmBlock* block, dmDescriptor* rts_dd );
void       ftp_fits_range_spec( dmBlock* block, char* name, double tmin, double tmax, int incnull );
regRegion* ftp_block_read_region( dmBlock* block );
long       ftp_block_read_gti( dmBlock* block, void** val1, dmDataType* type1, void** val2, dmDataType* type2 );


/* Called at block open */
void ftp_dataset_add_block( dmDataset* ds, long hdu, int ftHduType );
long ftp_raw_block_alloc( dmDataset* ds, const long no, const dmBlockType type, const char* name );

dmBlock* ftp_block_open( dmDataset* ds, long no, const dmBool create, char* hint );
void     ftp_block_read_header( dmBlock* block );
int      ftp_verbose_header( dmBlock* block );
int      ftp_make_ascii_tables( dmBlock* block );
int      ftp_table_open( dmBlock* block, const int create );
void     ftp_find_transforms( dmBlock* block );
void     ftp_vectorize( dmBlock* block );
dmDescriptor** ftp_map_vectors(  dmBlock* block, dmuAliasKeyList mlist, long* nrcols );
void     ftp_axis_match( dmBlock* block, dmDescriptor** cols, long ncols, long ftAxisNo, dmuAliasKeyList mlist );
void     ftp_block_read_pref( dmBlock* block );
void     ftp_subspace_open( dmBlock* block, int create );

/* ftBlock methods */
ftBlock* ftp_block_alloc( void );

/* Descriptor - Column functions */
dmDescriptor* ftp_col_read( dmBlock* block, const long colno, dmuKey **ppColKeysList );
long          ftp_col_read_max_varsize( dmBlock *block, const long no );
void          ftp_create_scalar_col( dmBlock* block, dmDescriptor* dd, long no, void* kdd, char* hint );
void          ftp_create_col( dmBlock* block, dmDescriptor* dd, long no, void* kdd, char* hint );
void          ftp_image_col_create( dmBlock* block, dmDescriptor* dd, void* kdd, char* hint );
long          ftp_col_no( dmDescriptor* dd );
void          ftp_init_col_range( dmuKeyList k, dmDescriptor* col );
void          ftp_init_col_bin_null( dmuKeyList k, dmDescriptor* col, int is_image );
long          ftp_col_write_mtype( dmBlock* block, dmDescriptor* dd, long n );
void          ftp_col_print( dmDescriptor* col );
void          ftp_descriptor_copy( dmDescriptor* srcdd, dmDescriptor* destdd );
void          ftp_descriptor_free( dmDescriptor* dd );
void          ftp_init_col_range_colkeys( dmuKey **ppColKeysList, dmDescriptor* col );
void          ftp_init_col_bin_null_colkeys( dmuKey **ppColKeysList, dmDescriptor* col );

/* Vectorization routines */

dmDescriptor* ftp_match_ctype_name( dmDescriptor** cols, long ncols, char* name );
dmDescriptor* ftp_dd_match( dmBlock* block, char* name,    dmuAliasKeyList mlist );
dmDescriptor* ftp_wcs_pair_match( dmBlock* block, dmDescriptor* transform_dd );
dmDescriptor* ftp_implicit_vector_transform( dmBlock* block, dmDescriptor* transform_dd );
dmDescriptor* ftp_dd_inherit( dmBlock* block, dmDescriptor* transform_vector );
dmDescriptor* ftp_column_match( dmBlock* block, dmDescriptor** cols, long ncols, long ftAxisNo, dmuAliasKeyList mlist );


/* ftCol methods */
ftCol*        ftp_col_alloc( void );
void          ftp_col_free( ftCol* col );


/* Coord routines */
void          ftp_apply_coord_params( dmBlock* block, double* val, int* found, char *pRadesys );
void          ftp_coords_update_physical( dmBlock* block );
int           ftp_coord_wcskey_match( const char* key, const char* keyPrefix );
void          ftp_attach_wcs( dmBlock* block, dmDescriptor* dd );
void          ftp_coord_set_wname( dmDescriptor* dd, const char* wname );
void          ftp_find_col_transforms( dmBlock* block );
dmDescriptor* ftp_coord_match_col_axis( const dmBlock* block, const long colno, const long axisno, const char sub );
void          ftp_coord_write_mtype( dmBlock* block, dmDescriptor* dd );
void          ftp_scalar_coord_write( dmDescriptor* dd, double crpix, double crval, double cdelt, char* ttype, long cptno, int p, int ok );
void          ftp_coord_set_rotmat_element( dmBlock* block, dmuKey* ftkey, int is_cd, int is_ltm );
void          ftp_coords_update_rotation( dmBlock* block );
void          ftp_coord_apply_rotmat( dmBlock* block, dmDescriptor* dd1 );
int           ftp_coord_is_ltm( const dmDescriptor* dd );
void          ftp_coord_clear_ltm( dmDescriptor* dd );


/* Coord create routines */
dmDescriptor* ftp_coord_find_key( dmBlock* block, const char* key, const char* keyPrefix );
dmDescriptor* ftp_create_wcs( dmBlock* block, const char* key, const char* ctype );
dmDescriptor* ftp_create_default_wcs( dmBlock* block, const long colno, const long axisno, const char sub );
dmDescriptor* ftp_create_unit_phys_wcs( dmBlock* block, dmDescriptor* col );

/* ftCoord routines */
ftCoord*      ftp_coord_alloc( void );
void          ftp_coord_free( ftCoord* ftcoord );

/* Subspace descriptors */
dmErrCode      ftp_subspace_read_table( dmDescriptor* dd, const char* tabname, dmBool create, void** val1, void** val2, long* nvals, regRegion** ss_region );
dmErrCode      ftp_read_subspace_values( dmDescriptor* dd, long cptno, char* dsval, char* tabname, int create );
void           ftp_read_dsval_key( dmuKeyList keylist, long index, long cptno, int was_table, char** dsvalp, char** tabnamep );
ftSubspaceCol* ftp_ftss_alloc( ftBlock* ftblock, long n );
ftSubspaceCol* ftp_get_ftss( dmBlock* block, dmDescriptor* dd, long n );
void           ftp_ftss_free( ftSubspaceCol* ftss );

/* Composite Keylist functions */
dmBool   ftp_composite_keylist_process( dmBlock* block, dmuKeyList keylist );



/* Raw Keylist functions */
void        ftp_keylist_print( const dmBlock* block );

void        ftp_dtype_update( const dmuVectorKey entry );

/* Keytype functions */
ftKeyType   ftp_cfitsio_key_type_to_ftk( const char* name, const char* value, const char fctype );
ftKeyType   ftp_datatype_to_keytype( const dmDataType type, const char* name );
dmDataType  ftp_keytype_to_datatype( const ftKeyType keytype );
void        ftp_keytype_name( const dmuKeyType type, char* buf, const long maxlen );

/* Key parsing */
void        ftp_key_parse( const char* buf, const ftKeyType fktype, void* value );
void        ftp_key_string_to_cstring( const char *buf, char *value );
int         ftp_key_is_comment( const char* name );
dmDataType  ftp_key_comment_parse( const char* cmt, dmDataType intype, char* desc, char* unit, long maxlen );
char*       ftp_key_comment_compose( const char* name, dmDataType type, const char* desc, const char* unit, char* comment, long maxlen );

/* Key functions */
dmuKey*      ftp_composite_key_write( dmBlock* block, long pos, const char* name, const dmDataType type, const void* value, long nvals, const char* unit, const char* comment );
dmuKey*      ftp_key_write_generic( dmBlock* block, long pos, const ftKeyType fktype, const dmDataType type, const char* name, const void* value, const char* unit, const char* desc );
void        ftp_key_delete( dmBlock* block, const char* name );
int         ftp_update_key_generic( fitsfile* fptr, char* name, dmDataType type, const void* value, char* comment );
int         ftp_insert_key_generic( fitsfile* fptr, long pos, char* name, dmDataType type, const void* value, char* comment );
int         ftp_write_comment( fitsfile* fptr, const char* tag, const char* value );


/* TFORM and other parsing */
void        ftp_compose_at_tform( DataType type, long isize, long length, char* format, char* tform );
void        ftp_compose_bt_tform( DataType type, long isize, long length, dmBool variable, char* tform );
dmDataType  ftp_at_tform_to_datatype( const char c );
dmDataType  ftp_bt_tform_to_datatype( const char c );
char        ftp_datatype_to_bt_tform( const dmDataType type );
dmDataType  ftp_parse_tform( const char* tform, int ascii, long* multiplier, long* trailer, int* variable );

int         ftp_bitpix( dmDataType type, double* bscale, double* bzero );
dmDataType  ftp_bitpix_to_type( long bitpix, double bscale, double bzero );

dmBlockType ftp_block_type( const int ftHduType );
void        ftp_hdutype_name( const int type, char* name );

int         ftp_block_name_parse( const char* name, char* hduname, char* extname, long* extver );
char*       ftp_read_block_name( fitsfile* fptr, const long ftBlockNo );
char*       ftp_default_hdu_name( const long ftBlockNo );

int         ftp_key_private_fits( const char* name );
void        ftp_parse_tdim( char* tdim, long** axesp, long* naxesp );
void        ftp_null_keyword_compose( const dmDataType type, const long colno, char* buf, int is_image );
int            ftp_recognize_spatial( char* name );
char*          ftp_subspace_table_name( dmDataset* ds, char* tabname );

int         ftp_ctype_parse( const char* ctype_in, char** nameptr, char** transform );
dmBool      ftp_wcsname_parse( const char* key, const char* keyPrefix, long* colno, char* sub, int is_image );
long        ftp_parse_wcs_key( const char* name, long* axisno, long* axisno2, char* sub, int is_image );
int         ftp_make_ctype_name( char** cname, char* ttype, char** ctype, long dim );
void        ftp_format_ctype( char* imtext, char* timtext, char* ttext,  long no, long axisno, char sub, int image, int axis, char* keyname );
void        ftp_format_pc( char* imtext, char* timtext, char* ttext, long no, long no2, long axisno, long axisno2, char sub, int image, int axis, char* keyname );
void        ftp_find_wcs_pair_name( const char* name, char* alt_name );

/* Low level routines */
void        ftp_create_dummy_image( fitsfile* fptr, const long no, const char* extname );
fitsfile*   ftp_create_fits( const char* filename );
int         ftp_ds_status( const dmDataset* ds, const int ftstatus, const char* tag );
void ftp_keylist_free( dmBlock* block, dmuKeyList keylist );

void ftp_hint_free( ftBlock* ftblock );
dmBlock* ftp_block_open_mode( dmDataset* ds, long no, dmBlockType type, int open_subspace, char* hint );


void ftp_write_remove_invalid_chars(const char *cTmpValue );

/* Keys */
void ftp_key_dump( dmBlock* block, char* tag);

void ftp_flush_image_struct( dmBlock* block );

dmBool ftp_radesys_equinox_validation( dmBlock *pBlock, long colNumber );
dmBool ftp_use_equinox( char *ctype_name);

#endif





