/*                                                                
**  Copyright (C) 2002-2003,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftkeylib.c                                                  */
/*                                                               */
/*   Functions that work with key type                           */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmulib.h"
#include "ftkernelP.h"
#include <limits.h>


/***************************************************************/
/*                                                             */
/*  Function:   ftp_cfitsio_key_type_to_ftk                    */
/*                                                             */
/*  Description:                                               */
/*    Convert the CFITSIO keytype to DM internal FITS key type */
/*                                                             */
/*  Input parameters:                                          */
/*           const char  *name - key name                      */
/*           const char  *value - key value                    */
/*           const char   fctype - type                        */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            ftKeyType  type                                  */
/*                                                             */
/***************************************************************/
ftKeyType ftp_cfitsio_key_type_to_ftk( const char* name, const char* value, const char fctype )
{
  double test = 0.0;

  if ( !strcmp( name, "COMMENT" ) || !strcmp( name, "HISTORY" ) )
    return FTK_COMMENT;

  switch( fctype )
  {
    case 'C':   return FTK_STRING;
    case 'L':   return FTK_LOGICAL;
    case 'I':
 
    if ( value )
    {
      test = atof( value );

      if ( test > DT_INT4_MAX )
      {
        if ( test > DT_UINT4_MAX )
          return FTK_REAL;
        else
          return FTK_UINT;
      }
    }   
    return FTK_INT;

    case 'F':   return FTK_REAL;
    default:    return FTK_UNKNOWN;
  }
}   


/***************************************************************/
/*                                                             */
/*  Function:   ftp_keytype_to_datatype                        */
/*                                                             */
/*  Description:                                               */
/*     Map FITS keyword types to DM data types                 */
/*                                                             */
/*  Input parameters:                                          */
/*           const  ftKeyType keytype  - key type              */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmDataType  type                                 */
/*                                                             */
/***************************************************************/
dmDataType ftp_keytype_to_datatype( const ftKeyType keytype )
{
  switch( keytype )
  {
    case FTK_COMMENT: return DT_CHAR;
    case FTK_STRING:  return DT_CHAR;
    case FTK_INT:     return DT_INT4;
    case FTK_UINT:    return DT_UINT4;
    case FTK_REAL:    return DT_REAL8;
    case FTK_LOGICAL: return DT_LOGICAL;
    case FTK_UNKNOWN: return DT_INT4;
    default:
     break;
  }
  return dmUNKNOWNTYPE;
}

 
/***************************************************************/
/*                                                             */
/*  Function:   ftp_keytype_name                               */
/*                                                             */
/*  Description:                                               */
/*      Map ftKeyType to a string for printing; used by        */
/*      KeyListPrint                                           */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmuKeyType type     - key type              */
/*                 long       maxlen - buffer length           */
/*                                                             */
/*  Output parameters:                                         */
/*                 char      *buf  - output buffer for string  */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/  
void ftp_keytype_name( const dmuKeyType type, char* buf, const long maxlen )
{
  if ( !buf || maxlen < 4 )
    return;

  switch( (ftKeyType)type )
  {
    case FTK_STRING:
      strcpy( buf, "C" ); return;
    case FTK_INT:
    case FTK_UINT:
      strcpy( buf, "I" ); return;
    case FTK_REAL:
      strcpy( buf, "R" ); return;
    case FTK_LOGICAL:
      strcpy( buf, "L" ); return;
    case FTK_COMMENT:
      strcpy( buf, "CMT" ); return;
    case FTK_UNKNOWN:
      strcpy( buf, "UNK" ); return;
  }
}
    

/***************************************************************/
/*                                                             */
/*  Function:   ftp_datatype_to_keytype                        */
/*                                                             */
/*  Description:                                               */
/*     Map DM data types to FITS key types                     */
/*                                                             */
/*  Input parameters:                                          */
/*           const dmDataType type  - data type                */
/*           const char      *name  = key name                 */
/*                                                             */
/*  Output parameters:                                         */
/*               none                                          */
/*                                                             */
/*  Returns:                                                   */
/*            ftKeyType key_type                               */
/*                                                             */
/***************************************************************/
ftKeyType ftp_datatype_to_keytype( const dmDataType type, const char* name )
{
  if ( ftp_key_is_comment( name ) )
    return FTK_COMMENT;
 
  switch( type )
  {
    case DT_CHAR:  return FTK_STRING;
    case DT_INT4:  return FTK_INT;
    case DT_INT2:  return FTK_INT;
    case DT_UINT1:   return FTK_INT;
    case DT_BIT:   return FTK_INT;
    case DT_UINT2:  return FTK_INT;
    case DT_UINT4:  return FTK_UINT;
    case DT_REAL4:  return FTK_REAL;
    case DT_REAL8: return FTK_REAL;
    case DT_LOGICAL:  return FTK_LOGICAL;
    case DT_UNK: return FTK_UNKNOWN;
    default:   return FTK_UNKNOWN;
  }
}
  

/***************************************************************/
/*                                                             */
/*  Function:   ftp_key_is_comment                             */
/*                                                             */
/*  Description:                                               */
/*    Test for the 'COMMENT' class of keywords. In the current */
/*    implementation we consider only COMMENT and HISTORY to   */
/*    be comments when reading the file. Another approach would*/
/*    be to consider any key without an equal sign in cols     */
/*    9-10 to be a comment.                                    */
/*                                                             */
/*  Input parameters:                                          */
/*           const char      *name  = key name                 */
/*                                                             */
/*  Output parameters:                                         */
/*               none                                          */
/*                                                             */
/*  Returns:                                                   */
/*            int  0 if false, 1 if is comment                 */
/*                                                             */
/***************************************************************/
int ftp_key_is_comment( const char* name )
{
  if ( !strcmp( name, "COMMENT" ) ) return 1;
  if ( !strcmp( name, "HISTORY" ) ) return 1;
  return 0;
}

