/*                                                                
**  Copyright (C) 2002-2003,2005-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftkeylist.c                                                 */
/*                                                               */
/*   Functions that parse FITS keywords                          */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/


#include "dmilib.h"
#include "ftkernelP.h"

static int ftp_key_continue( char** pptr, char* text, char* desc );
static int ftp_key_ends_in_amp( const char* value );


/***************************************************************/
/*                                                             */
/*  Function:   ftp_block_read_header                          */
/*                                                             */
/*  Description:                                               */
/*  Read the header keys. Ignore CFITSIO errors, since we      */
/*  want to carry on even if a few keys are corrupt.           */
/*                                                             */
/*  Input parameters:                                          */
/*          dmBlock  *block                                    */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void ftp_block_read_header( dmBlock *block )
{
  fitsfile    *fptr = NULL;

  ftBlock     *ftblock = NULL;
  ftKeyType    fktype;

  dmDataset   *ds = NULL;
  dmDataType   type;
  dmBool       mask;
  dmuKeyList   keylist;
  dmuKey      *ftkey   = NULL;

  GenericType         tmp;
  void *parsed_value = &tmp;

  char  fctype        = '\0';   /* CFITSIO card type */
  char  descbuf[FITS_MAXLEN];
  char  unit[FITS_MAXLEN];
  char  name[FLEN_KEYWORD];     /* CFITSIO: FLEN_KEYWORD defined in fitsio.h */
  char  value[FITS_MAXLEN];
  char  card[FITS_MAXLEN];
  char  comment[FITS_MAXLEN];

  long  ncont         = 0;
  long  nkeys         = 0;
  int   keyno         = 0;      /* CFITSIO Key No */
  int   fnkeys        = 0;      /* CFITSIO number of keys */
  int   ftmp          = 0;      /* CFITSIO scratch variable */
  int   ftstatus      = FITS_KERNEL_OK;


  ftblock = (ftBlock*)dmi_block_get_kb( block );

  if ( !ftblock )
    return;

  fptr = ftp_fptr( block );

  ds = dmi_block_get_ds( block );

  if ( dmi_dataset_mode( ds ))
    fits_flush_file( fptr, &ftstatus );         /* CFITSIO */

  fits_get_hdrspace( fptr, &fnkeys, &ftmp, &ftstatus );    /* CFITSIO */

  if ( !ftp_ds_status( ds, ftstatus, "[ftpReadHeaderKeys]" ) )
    return;

  if ( ftblock->keylist )
    dmu_keylist_free( ftblock->keylist );

  nkeys = fnkeys + FT_SPARE_KEYS;
  keylist = dmu_keylist_alloc( );

  if ( !keylist )
    return;

  ftblock->keylist = keylist;
  ftstatus = FITS_KERNEL_OK;

  for ( keyno = 1; keyno <= fnkeys; keyno++ )
  {
    /* Read the next key */
    fits_read_keyn( fptr, keyno, name, value, comment, &ftstatus );  /* CFITSIO */

    ftstatus = FITS_KERNEL_OK;

    /* Determine its FITS key type */
    fits_get_keytype( value, &fctype, &ftstatus );                       /* CFITSIO */

    /* Work around bug in ffdtyp */ 
    if ( fctype == 'I' && strchr( value, 'E' ))
      fctype = 'F';

    ftstatus = FITS_KERNEL_OK;

    /* Map to FTKERNEL key type */
    fktype = ftp_cfitsio_key_type_to_ftk( name, value, fctype );

    type = ftp_keytype_to_datatype( fktype );
    mask = ftp_key_private_fits( name );

    /* Parse key and add to list */

    if ( fktype == FTK_COMMENT )
    {
      ftp_key_parse( comment, fktype, (void*)parsed_value );
      strcpy( unit, "");
      strcpy( descbuf, "" );
    }
    else
    {
      ftp_key_parse( value, fktype, (void*)parsed_value );

      /* Special handling for CONTINUE convention. In this convention, */
      /* keys of type string (only) have their values continued on the */
      /*  next card if their value ends in "&".                        */

      ncont =0;

      if ( fktype == FTK_STRING )
      {
        /* Set flag for continue key search */
        dmBool done = UT_FALSE;

        while ( !done && ftp_key_ends_in_amp( *(char**)parsed_value ) && (keyno < fnkeys) )
        {
          ++ncont;

          fits_read_record( fptr, keyno+1, card, &ftstatus );     /* CFITSIO */

          if ( cs_eq_ss( card, 1, 10, "CONTINUE  " ))
          {
            ftp_key_continue( (char**)parsed_value, card+10, comment );
            ++keyno;
          }
	  else
	  {
	    /* Following keyword is NOT a CONTINUE. nothing else to append.. go on. */
	    done = UT_TRUE;
	  }
         } // end while
      } // end if fktype == FTK_STRING

      type = ftp_key_comment_parse( comment, type, descbuf, unit, FITS_MAXLEN );

    } // end else

    ftkey = dmu_key_alloc();

    ftkey->type = FTK_KEY;

    dmu_key_set( ftkey, (dmuKeyType)fktype, type, name, parsed_value, unit, descbuf, mask );
    dmu_key_insert( keylist, ftkey, 0 );  

    if ( type == DT_CHAR )
      free( *(char**)parsed_value);

  } // end for loop
}



/***************************************************************/
/*                                                             */
/*  Function:  ftp_key_continue                                */
/*                                                             */
/*  Description:                                               */
/*     Process continue statements as a special case           */
/*                                                             */
/*  Input parameters:                                          */
/*           char  **pptr                                      */
/*           char   *text                                      */
/*                                                             */
/*  Output parameters:                                         */
/*           char  **pptr                                      */
/*           char   *desc                                      */
/*                                                             */
/*  Returns:                                                   */
/*           static int 1                                      */
/*                                                             */
/***************************************************************/
static int ftp_key_continue( char **pptr, char *text, char *desc )
{
  char *ptr  = *pptr;
  char *tmp  = NULL;

  long  n    = 0;
  long  m    = 0;

  int   flag = UT_FALSE;
  int   loop = 1;
  int   iToIndex   = 0;
  int   iFromIndex = 0;


  n = strlen( ptr );

  /* Lose trailing ampersand on previous line (GSFC FITS continue convention) */
  if ( ptr[n-1] == '&' )
    --n;

  /* Lose leading quote of new line */
  if ( *text == '\'' )
    ++text;

  /* Seek trailing quote of new line */

  while ( loop )
  { 
    /* Text ends on null */
    if ( text[m] == '\0' )
    {
      loop = UT_FALSE;
    }
    else
    {
      if ( flag ) 
      {
        if ( text[m] == '\'' )
        {
          flag = UT_FALSE;

          /* the following code section added for Bug 9650 10-24-2005     */
          /* remove the double quote from key value string                */
          /*  NB: the following code section expects the string being     */
          /*      seached to be NULL terminated.  If the buffer is        */
          /*      not NULL terminated, this code will go into an infinite */
          /*      loop.  But then again, so will the outer loop that it   */
          /*      lies within.                                            */

          iToIndex   = m-1;
          iFromIndex = m+1;
 
	  do
	  {
 	    /* shift all remaining chars down one location in buffer */
            text[++iToIndex] = text[iFromIndex++];

	  } while (text[iToIndex] != '\0');

	  /* decrement string index so that don't skip the character */
	  /* that we just moved into this place.                     */
	  --m;

        }       /* end of code section added for Bug 9650 10-24-2005 */
        else
        {
          loop = UT_FALSE;
        }

      } /* end if (flag) */
      else
      {
        if ( text[m] == '\'' )
        {
	  /* Quote: this might end the string (or be a quoted quote) */
	  flag = UT_TRUE; 
        }
      }
    }

    if (loop)
      ++m;

  } /* end while */


  /* Now m is the first char past the trailing quote or is the term null */

  /* allocate new buffer */
  tmp = malloc( n + m );

  /* copy text from two input buffers into new buffer  */
  cs_copy_siz( ptr, tmp, n );
  cs_copy_siz( text, tmp+n, m-1 );

  /* clean up and set pptr to new buffer area */
  free( ptr );
  *pptr = tmp;


  if ( flag )
  {
    while ( text[m] == ' ' )
      ++m;

    if ( text[m] == '/' )
    {
      ++m;

      while( text[m] == ' ' )
        ++m;

      cs_copy( text+m, desc );
    }
  }

  return 1;
}



/***************************************************************/
/*                                                             */
/* Function:   ftp_key_ends_in_amp                             */
/*                                                             */
/* Description:                                                */
/*    Check a string for an end character of & or &'           */
/*                                                             */
/*  Input parameters:                                          */
/*                    const char *value                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*          static int 0 or 1                                  */
/*                                                             */
/***************************************************************/
static int ftp_key_ends_in_amp( const char *value )
{
  char *sval = NULL;
  char *ptr  = NULL;

  sval = (char*)value;

  if ( sval && *sval )
  {
    ptr = sval + strlen(sval)-1;

    if ( *ptr == '\'' && ptr > sval )
      --ptr;

    if ( *ptr == '&' )  /* CONTINUATION? */
      return 1;
  }

  return 0;
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_key_comment_parse                           */
/*                                                             */
/* Description:                                                */
/*    Strip leading [unit]{type} information from FITS key     */
/*    comment line                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           const char  *cmt                                  */
/*           dmDataType   intype                               */
/*           char        *desc                                 */
/*           char        *unit                                 */
/*           long         maxlen                               */    
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*           dmDataType  type                                  */
/*                                                             */
/***************************************************************/
dmDataType ftp_key_comment_parse( const char *cmt, dmDataType intype,
                                char *desc, char *unit, long maxlen )
{
  const char *ptr = cmt;
  const char *optr = NULL;
  char  typecode[WORD_SIZE];

  long  nbytes = 0;

  dmDataType type = intype;


  strcpy( desc, "" );
  strcpy( unit, "" );

  if ( !cmt )
    return type;

  while( isspace( *ptr ) )
    ++ptr;
 
  if ( *ptr == '[' )
  {
    optr = ++ptr;

    while( *ptr != ']' && *ptr != '\0' ) 
      ++ptr;

   nbytes = ptr - optr;

   if ( nbytes > maxlen )
     nbytes = maxlen;

   cs_copy_siz( optr, unit, nbytes );

   if ( *ptr == ']' ) 
     ++ptr;

   while( isspace( *ptr ) )
     ++ptr;
  }
 
  if ( *ptr == '{' )
  {
    optr = ++ptr;

    while( *ptr != '}' && *ptr != '\0' ) 
      ++ptr;

    nbytes = ptr - optr;

    if ( nbytes > maxlen )
      nbytes = maxlen;

    cs_copy_siz( optr, typecode, nbytes );

    if ( *ptr == '}' )
      ++ptr;

    while( isspace( *ptr ) )
      ++ptr;

    type =  ftp_bt_tform_to_datatype( typecode[0] );
  }

  cs_copy_siz( ptr, desc, maxlen );

  return type;
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_key_comment_compose                         */
/*                                                             */
/* Description:                                                */
/*    Implement the [unit]{type} comment syntax on write       */
/*    If the comment is blank, return a null pointer, else     */
/*    return the comment pointer.                              */
/*                                                             */
/*  Input parameters:                                          */
/*           const char  *name                                 */
/*           dmDataType   type                                 */
/*           const char  *desc                                 */
/*           const char  *unit                                 */
/*           long         maxlen                               */
/*                                                             */
/*  Output parameters:                                         */
/*           char        *comment                              */
/*                                                             */
/*  Returns:                                                   */
/*           char *comment  or NULL                            */
/*                                                             */
/***************************************************************/
char* ftp_key_comment_compose( const char *name, dmDataType type,
       const char *desc, const char *unit, char *comment, long maxlen )
{
  dmDataType  default_type;
  ftKeyType   fktype;
  char       *ptr = NULL;

  fktype = ftp_datatype_to_keytype( type, name );

  strcpy( comment, "" );

  /* Get rid of COMMENT keys */
  if ( fktype == FTK_COMMENT )
    return NULL; 

  ptr = comment;

  if ( !cs_is_blank( unit ))
    ptr += sprintf( ptr, "[%s]", unit );

  default_type = ftp_keytype_to_datatype( fktype );

  if ( !( dm_global & DM_GLOBAL_OLD ) && default_type != type )
  {
    ptr += sprintf( ptr, "{%c}", ftp_datatype_to_bt_tform( type ) );
  }

  // Add a space for clarity
  if ( ptr > comment ) 
  {
    *ptr++ = ' ';
    *ptr = 0;
  }
 
  /* Return null if there is no comment and no unit/type */

  if( ptr == comment && ( !desc || *desc == '\0' ) )
    return NULL;

  maxlen -= ptr - comment;

  cs_copy_siz( desc, ptr, maxlen );

  return comment;
}


/***************************************************************/
/*                                                             */
/*  Function:  ftp_key_parse                                   */
/*                                                             */
/*  Description:                                               */
/*      Returns the value of a key                             */
/*                                                             */
/*  Input parameters:                                          */
/*           const char      *buf                              */
/*           const ftKeyType  fktype                           */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*           void  *value                                      */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void ftp_key_parse( const char *buf, const ftKeyType fktype, void *value )
{
  char valstring[FITS_MAXLEN];

  switch( fktype )
  {
    case FTK_COMMENT:
      *(char**)value = cs_dup( buf );
      break;

    case FTK_STRING:
      if ( *buf == '\'' )    /* Starts with a quote */
      {
        ftp_key_string_to_cstring( buf, valstring );
        *(char**)value = cs_dup( valstring );
      }
      else
      {
        *(char**)value = cs_dup( buf );
      }
      break;

    case FTK_INT:
      *(long*)value = atol( buf );
      break;

    case FTK_UINT:
      *(unsigned long*)value = (unsigned long)(atof(buf) + 0.1);
      break;

    case FTK_REAL:
      *(double*)value = atof( buf );
      break;

    case FTK_LOGICAL:
      while( *buf == ' ' )
        ++buf;

      *(dmBool*)value = (*buf == 'T') ? UT_TRUE : UT_FALSE;
      break;

    case FTK_UNKNOWN:
    default:
      value = NULL;
      break;
   }
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_key_string_to_cstring                        */
/*                                                             */
/* Description:                                                */
/*            Remove leading and trailing quotes and trailing  */
/*            spaces. Replace FITS style double quotes with    */
/*            single quotes.                                   */
/*                                                             */
/* Input parameters:                                           */
/*          const char *buf                                    */
/*                                                             */
/* Output parameters:                                          */
/*          char *value                                        */
/*                                                             */
/* Returns:                                                    */
/*          void                                               */
/*                                                             */
/***************************************************************/
void ftp_key_string_to_cstring( const char *buf, char *value )
{
  char  valstring[FITS_MAXLEN]; 
  char *iptr = NULL;
  char *optr = NULL;
  long  ij   = 0;

  cs_remove_quotes_and_trail( buf, valstring, FITS_MAXLEN );

  iptr = valstring;

  optr = value;

  while  (ij < FITS_MAXLEN-1 && iptr[ij] != '\0' )
  {
    *optr = iptr[ij];
    ++optr;

    if ( iptr[ij] == '\'' && iptr[ij+1] == '\'' )
      ij += 2;
    else
      ++ij;
   }

   *optr = '\0';
 }


/***************************************************************/
/*                                                             */
/* Function:  ftp_radesys_equinox_validation                   */
/*                                                             */
/* Description:                                                */
/*       Insure that values for keywords RADESYS and EQUINOX   */
/*       are compatible with each other.                       */
/*                                                             */
/*       Presently, only check for RADESYS = ICRS              */
/*       In future, do more validation of values               */
/*                                                             */
/* Input Parameters:                                           */
/*       dmBlock   *pBlock                                     */
/*       long       colNumber                                  */
/*                                                             */
/* Output Parameters:                                          */
/*                                                             */
/*                                                             */
/* Returns:                                                    */
/*          dmBool status  - either dmTRUE or dmFALSE          */
/*                                                             */
/* Called from:                                                */
/*         ftp_scalar_coord_write()                            */
/*                                                             */
/***************************************************************/
dmBool ftp_radesys_equinox_validation(dmBlock *pBlock, long colNumber )
{
  dmuKey      *pKeyRadesys = NULL;
  dmuKeyList   keylist;

  dmBool  bIsICRS  = dmFALSE;
  dmBool  bIsGAPPT = dmFALSE;

  ftBlock  *pftBlock   = NULL;

  char   *psRadesys = NULL;
  char    sRadesys[CARD_SIZE];
 
//double  equinox_number = 0.0;
//double  EQ1984         = 1984.0;
//double  EQ2000         = 2000.0;

    
  // get keylist
  pftBlock = (ftBlock*) dmi_block_get_kb(pBlock);
  keylist  = pftBlock->keylist;

  // get column specific key name RADE#
  sprintf(sRadesys, "%s%d","RADE", (int)colNumber);
  pKeyRadesys = dmu_key_find(keylist, sRadesys);

  if ( !pKeyRadesys )
  {
    // no specific key, get global RADESYS key
    pKeyRadesys = dmu_key_find(keylist, "RADESYS");

    // try RADECSYS key
    if ( !pKeyRadesys )
     pKeyRadesys = dmu_key_find(keylist, "RADECSYS");
  }

  // If no RADESYS key, system is assumed to be "ICRS"
  if ( !pKeyRadesys )
    return dmTRUE;


  // get value of RADE#/RADESYS key
  psRadesys = dmu_key_value_c( pKeyRadesys);

  // is ICRS?
  bIsICRS  = cs_uceq(psRadesys, "ICRS");

  // if not, check for GAPPT
  if ( !bIsICRS )
    bIsGAPPT = cs_uceq(psRadesys, "GAPPT");


  // EQUINOX not applicable for ICRS or GAPPT
  if ( bIsICRS || bIsGAPPT )
    return dmTRUE;

  return dmFALSE;
 }
