/*                                                                
**  Copyright (C) 2002-2003,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftmisc.c                                                    */
/*                                                               */
/*   Functions that create and work with FITS kernel files       */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmilib.h"
#include "ftkernelP.h"

/*                      */
/*  FITS kernel hints   */
/*                      */

typedef struct ftHintData_s
{
  dmBool ascii_table;   /* Next table create is an ASCII table */
  dmBool verbose_hdr;     /* Use verbose DTYPE/DFORM keys */
} ftHintData;

/* Global for back compatibility */
static ftHintData ftHints = { UT_FALSE, UT_FALSE };
ftHintData* ftp_hint_data( dmBlock * block );



/***************************************************************/
/*                                                             */
/* Function:   ftp_hint_free                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              ftBlock  *ftblock                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void ftp_hint_free( ftBlock* ftblock )
{
  if ( ftblock->hint )
  {
    free( ftblock->hint );
    ftblock->hint = NULL;
  }
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_hint_data                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              ftBlock  *ftblock                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            ftHintData  *hints                               */
/*                                                             */
/***************************************************************/
ftHintData* ftp_hint_data( dmBlock * block )
{
  ftHintData *hints   = NULL;
  ftBlock    *ftblock = NULL;

  if ( block )
  { 
    ftblock = (ftBlock*)dmi_block_get_kb( block );

    if ( !ftblock->hint )
      ftblock->hint = calloc( 1, sizeof( struct ftHintData_s ));

    hints = ftblock->hint;
  }
  else
    hints = &ftHints;  /* Back compat for global variable use */

  return hints;
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_set_hint                                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock    *block                              */
/*              const char *name                               */
/*              const char *value                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode ftp_set_hint( dmBlock* block, const char* name, 
                        const char* value )
{
  dmErrCode  status = dmSUCCESS;


  ftHintData  *hints = ftp_hint_data( block );

  if ( cs_uceq( name, "TABLE" ) )
  {
    if ( cs_uceq( value, "ASCII" ) )
      hints->ascii_table = UT_TRUE;
    else
      if ( cs_uceq( value, "BINTABLE" ) )

    hints->ascii_table = UT_FALSE;
    /* All other values ignored */
  }
  else if ( cs_uceq( name, "HEADER" ) )
  {
    if ( cs_uceq( value, "VERBOSE" ) )
      hints->verbose_hdr = UT_TRUE;
    else /* "DEFAULT" */
      hints->verbose_hdr = UT_FALSE;
  }
  else if ( cs_uceq( name, "KERNEL" ) && cs_uceq( value, "FITS" ) )
  {
    /* Nothing to set.. just clears it from list */
  }
  else
    status = dmi_block_set_hint( block, name, value );

 return status;
}

/***************************************************************/
/*                                                             */
/* Function:   ftp_set_hint_list                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock    *block                              */
/*              const char *option                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmErrCode ftp_set_hint_list( dmBlock* block, const char* option,
                             dmBool bPrintErrFlag )
{
  char   namebuf[80] = "";
  char  *name = NULL;
  char **list = NULL;

  long   nitems = 0;
  long   ii     = 0;
  char  *value  = NULL;

  dmErrCode  status = dmSUCCESS;


  if ( cs_is_blank( option )) 
    return status;

  name = namebuf;
  list =  cs_parse_comma_list( option, &nitems );


  if ( !nitems  || !list)
    return status;


  for (ii = 0; ii < nitems; ii++)
  {
    /* Split token into name-value pair */
    value = dmu_parse_token( list[ii], '=', &name, CARD_SIZE );

    if ( ftp_set_hint( block, name, value ) != dmSUCCESS )
    {
      if ( bPrintErrFlag )
      {
	/* Set error message - cannot parse option */
	/* Set status to failure */
	sprintf( dm_errbuf, "Unrecognized option %s ignored.", list[ii] );

	dmu_err_set( DM_WARN_PARSE_OPTION, dm_errbuf );
	status = dmFAILURE;
      }
    }
  }

  ar_free_cp( list, nitems );

  return status;
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_verbose_header                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock    *block                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  status                                */
/*                                                             */
/***************************************************************/
dmBool ftp_verbose_header( dmBlock* block )
{
  ftHintData* hints = ftp_hint_data( block );

  return hints->verbose_hdr;
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_make_ascii_tables                           */
/*                                                             */
/* Description:                                                */
/*             Check what kind of tables we're making today    */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock    *block                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmBool  ascii_table                              */
/*                                                             */
/***************************************************************/
dmBool ftp_make_ascii_tables( dmBlock* block )
{
  ftHintData* hints = ftp_hint_data( block );

  return hints->ascii_table;
} 
  

/***************************************************************/
/*                                                             */
/* Function:   ftp_ds_status                                   */
/*                                                             */
/* Description:                                                */
/*             Check the FITS error status                     */
/*                                                             */
/*  Input parameters:                                          */
/*              const dmDataset  *ds                           */
/*              const int         ftstatus                     */
/*              const char       *tag                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmBool  status                                   */
/*                                                             */
/***************************************************************/
dmBool ftp_ds_status( const dmDataset* ds, const int ftstatus, const char* tag )
{
  char  ftbuf[FITS_MAXLEN] = " " ;
  int   status   = ftstatus;
  char *dsname   = NULL;
  char * name    = NULL;
  long   current = 0;

  if ( ftstatus == FITS_KERNEL_OK )
    return UT_TRUE;
 
  fits_get_errstatus( status, ftbuf );   /* CFITSIO */

  current = dmi_dataset_get_current( ds );
  dsname  = dmi_dataset_get_name( ds );
  name    = dmi_dataset_get_block_name( ds, current );

  if ( !name )
    name = "?";

  if ( !dsname )
    dsname = "?";

  sprintf( dm_errbuf, "%s: FITS error %d %s in dataset %s Block %ld %s", 
                               tag, ftstatus, ftbuf, dsname, current, name );
  dmu_err_set( DM_ERR_KERNEL, dm_errbuf );

  return UT_FALSE;
}

