/*                                                                
**  Copyright (C) 2002-2003,2005,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftname.c                                                    */
/*                                                               */
/*   Functions to construct block name                           */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmulib.h"
#include "ftkernelP.h"

/***************************************************************/
/*                                                             */
/*  Function:   ftp_block_name_parse                           */
/*                                                             */
/*  Description:                                               */
/*      The block name is equal to the HDUNAME keyword.        */
/*      The EXTNAME keyword is generated by stripping off any  */
/*      trailing digits and treating them as EXTVER.           */
/*      Example:                                               */
/*          EVENTS32                                           */
/*              -> HDUNAME = 'EVENTS32'                        */
/*                 EXTNAME = 'EVENTS'                          */
/*                 EXTVER  =  32                               */
/*                                                             */
/*  Input parameters:                                          */
/*           const char  *name - input string                  */
/*                                                             */
/*  Output parameters:                                         */
/*                 char  *hduname                              */
/*                 char  *extname                              */
/*                 long  *extver                               */
/*                                                             */
/*  Returns:                                                   */
/*             dmBool have_extver  - flag for extver           */
/*                                                             */
/***************************************************************/
dmBool ftp_block_name_parse( const char* name, char* hduname, char* extname, long* extver )
{
  char  *tmp = NULL;
  dmBool have_extver = UT_FALSE;
  long   j = 0;
  long maxlen = FITS_MAXLEN - 13;  /* 13 for terminating '\0' and "KEYNAMEx= ''" */
  long slen = 0;

  if ( !name )
  {
    cs_copy( "", hduname );
    cs_copy( "", extname );
    *extver = 0;
  }
  else
  {
    /* Make sure length of name is no greater than maxlen by first stripping
     * any leading path component and then taking at most maxlen characters
     * from the end of the string (the assumption being that the end of the
     * filename will have more useful information than the beginning). */
    if ( strlen( name ) > (size_t)maxlen )
    {
      while ( NULL != ( tmp = strchr( name, '/' ) ) )
        name = tmp + 1;

      slen = strlen( name );

      if ( slen > maxlen )
        name += slen - maxlen;
    }

    j = strlen( name )-1;
    strcpy( hduname, name );  /* We make a separate variable for HDUNAME
                              in case the rule for it gets changed later
                              */
    if ( j >= 0 && isdigit( name[j] ) )
    {
      while( j >= 0 && isdigit( name[j] )  )
        j--;

      if ( j < 0 )  /* If all digits, special case - treat as a name */
      {
        strcpy( extname, name );
        *extver = 0;
      }
      else
      {      /* Strip trailing digits to extver */
        cs_copy_siz( name, extname, j+1 );
        *extver = atol( name+j+1 );
        have_extver = UT_TRUE;
      }
    }
    else
    {    /* No trailing digits */
      strcpy( extname, name );
      *extver = 0;
    } 
  }
  return have_extver;
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_read_block_name                            */
/*                                                             */
/*  Description:                                               */
/*     Here we set the FITS HDU block name. The rule is:       */
/*     - if key HDUNAME is present, its value is the block name*/
/*     - else if keys EXTNAME and EXTVER are present,          */
/*       the name is EXTNAME concatenated with EXTVER if       */
/*       EXTVER is not 1                                       */
/*     - if there is EXTNAME but EXTVER is absent or equal to  */
/*       1, just use EXTNAME                                   */
/*     - Default name for HDU1 is PRIMARY                      */
/*     - if all else fails, or if there is a FITS error, use   */
/*       the string HDUnn where nn is the FITS HDU number      */
/*       counting from 1.                                      */
/*                                                             */
/*  Input parameters:                                          */
/*           fitsfile  *fptr  - file pointer                   */
/*           const long ftblockno - block number               */
/*                                                             */
/*  Output parameters:                                         */
/*                 none                                        */
/*                                                             */
/*  Returns:                                                   */
/*             char *name - block name string                  */
/*                                                             */
/***************************************************************/
char* ftp_read_block_name( fitsfile* fptr, const long ftblockno )
{
  char *name = NULL;
  char  ftcomment[FITS_COMMENT_SIZE];
  char  tmp_name[TEXT_SIZE];
  long  extver = 1;
  int   ftstatus = FITS_KERNEL_OK;

  /* First try HDUNAME */
  fits_read_key_longstr(fptr, "HDUNAME", &name, ftcomment, &ftstatus ); /* CFITSIO */

  /* Second try EXTNAME/EXTVER */
  /* MCD - 20090615: if HDUNAME exists, but is blank.. try EXTNAME */
 
  if ( (ftstatus == KEY_NO_EXIST) || cs_is_blank( name ) )
  {
    ftstatus = FITS_KERNEL_OK;

    /* clean up from first attempt */
    if (name)
    {
      free(name);
      name = NULL;
    }
    fits_read_key_longstr(fptr,"EXTNAME", &name, ftcomment, &ftstatus );  /* CFITSIO */

    /* Check for EXTVER*/
    if ( ftstatus == FITS_KERNEL_OK ) 
    {
      fits_read_key_lng( fptr, "EXTVER", &extver, ftcomment, &ftstatus );  /* CFITSIO */

      if ( ftstatus == FITS_KERNEL_OK && extver != 1 )
      {
        sprintf( tmp_name, "%s%ld", name, extver );
        free( name );
        name = cs_dup( tmp_name );
      }
      else
        if ( ftstatus == KEY_NO_EXIST ) 
        {
          ftstatus = FITS_KERNEL_OK;
        }
    }
  }

  /* Third, set it to HDUn and handle errors */
 
  if ( ftstatus != FITS_KERNEL_OK )
  {
    if ( ftstatus != KEY_NO_EXIST )
      dmu_err_set( DM_ERR_KERNEL, "[ftp_read_block_name]: Problem reading HDU name" );

    name = ftp_default_hdu_name( ftblockno );
  }

  return name; 
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_default_hdu_name                           */
/*                                                             */
/*  Description:                                               */
/*      Rule for default name for FITS HDU:                    */
/*        "HDUn", where n is block number                      */
/*                                                             */
/*  Input parameters:                                          */
/*           const long ftblockno - block number               */
/*                                                             */
/*  Output parameters:                                         */
/*                 none                                        */
/*                                                             */
/*  Returns:                                                   */
/*             char *name - block name string                  */
/*                                                             */
/***************************************************************/
char* ftp_default_hdu_name( const long ftblockno )
{
  char tmp_name[TEXT_SIZE];

  if ( ftblockno == 1 )
    cs_copy( "PRIMARY", tmp_name );
  else
    sprintf( tmp_name, "HDU%ld", ftblockno );

  return cs_dup( tmp_name );
}


