/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftpref.c                                                    */
/*                                                               */
/*   Functions that handle preferred axes                        */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmilib.h"
#include "ftkernelP.h"

char** ftp_block_get_pref_default( dmBlock* block, long* nprefp );

/***************************************************************/
/*                                                             */
/*  Function:   ftp_block_read_pref                            */
/*                                                             */
/*  Description:                                               */
/*      The keyword                                            */
/*         CPREF = 'DETX,DETY'                                 */ 
/*      or                                                     */
/*         CPREF = 'PHA(DETX,DETY)'                            */
/*      denotes preferred axes and (optionally) weighting.     */
/*                                                             */
/*  Input parameters:                                          */
/*               dmBlock  *block - block; prefs set in block   */
/*                                 structure                   */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*               void                                          */
/*                                                             */
/***************************************************************/
void ftp_block_read_pref( dmBlock* block )
{
  ftBlock    *ftblock  = NULL;
  dmuKeyList  keylist;
  dmuKey     *ftkey    = NULL;
  char       *buf      = NULL;
  char      **prefcols = NULL;

  long npref = 0;
  /* Look for CPREF keyword */
  int is_default = UT_TRUE;


  ftblock = (ftBlock*)dmi_block_get_kb( block );
  if ( !ftblock )
    return;

  keylist = ftblock->keylist;
  ftkey = dmu_key_find( keylist, "CPREF" );

  if ( ftkey && ftkey->datatype == DT_CHAR )
  {
    buf = dmu_key_value_c( ftkey );
    prefcols = cs_parse_paren_list( buf, &npref );
    is_default = UT_FALSE;
    /*
     *  prefcols[0] is the weight; ignore for now
     */
  }
  else
  {
    prefcols = ftp_block_get_pref_default( block, &npref );
  }

  dmi_block_set_pref( block, prefcols, npref, is_default );
  return;
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_block_get_pref_default                     */
/*                                                             */
/*  Description:                                               */
/*     If there are table columns called X and Y, use them.    */
/*      Otherwise use the first two axes.                      */
/*                                                             */
/*  Input parameters:                                          */
/*               dmBlock  *block - block; prefs set in block   */
/*                                 structure                   */
/*                                                             */
/*  Output parameters:                                         */
/*               long  *nprefp - number of preferred cols      */
/*                                                             */
/*  Returns:                                                   */
/*               char **prefcols - list of preferred col names */
/*                                                             */
/***************************************************************/
char** ftp_block_get_pref_default( dmBlock* block, long* nprefp )
{
  dmDescriptor *dd = NULL;
  ftBlock      *ftblock  = NULL;
  char        **prefcols = NULL; 
  char         *name     = NULL;
  long          npref = 2;
  long          i = 0;
 

  ftblock = (ftBlock*)dmi_block_get_kb( block );

  if ( ftblock->ncols < 2 )
    npref = ftblock->ncols;

  if ( npref > 0 )
    prefcols = calloc( npref+1, UT_SZ_P );

  for ( i = 0; i < ftblock->ncols; i++ )
  {  
    dd = ftblock->cols[i];
    dd = dmi_descriptor_get_scaled( dd );
    name = dmi_descriptor_get_name( dd );

    if ( cs_uceq( "X", name ) )
      prefcols[1] = cs_dup( name );
    
    if ( cs_uceq( "Y", name ) )
      prefcols[2] = cs_dup( name );
  }

  /* OK use first two */
  if ( npref == 2 && !prefcols[1] ) 
  {
    dd = dmi_descriptor_get_scaled( ftblock->cols[0] );
    name = dmi_descriptor_get_name( dd );
    prefcols[1] = cs_dup( name );
  }

  if ( npref == 2 && !prefcols[2] ) 
  {
    dd = dmi_descriptor_get_scaled( ftblock->cols[1] );
    name = dmi_descriptor_get_name( dd );
    prefcols[2] = cs_dup( name );
  }

  *nprefp = npref;
  return prefcols;
}



