/*                                                                
**  Copyright (C) 2002-2004,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftresv.c                                                    */
/*                                                               */
/*    Matching reserved FITS keywords                            */
/*                                                               */
/*****************************************************************/

#include "dmulib.h"
#include "ftkernelP.h"
#include <ctype.h>


/* Rewritten JCM 1998 Oct to explicitly specify where numbers can appear */
/* in the indexed keywords - so that CDE3LT is now visible but CDELT3 is */
/* not.   * # is a digit, . matches any letter                           */

// FITS Standard Compliance  July 2010
//  Added TCNA#, BLANK keywords

static char *FITS_WCS_KEYWORDS[]=
{
	"CD#_#",       /* WCS axis rotation matrix key  */
	"CDELT#",      /* WCS axis delta key */
	"CNAME#",      /* WCS axis name key  */
	"CROTA#",      /* WCS axis rotation key */
	"CRPIX#",      /* WCS axis reference pixel key */
	"CRVAL#",      /* WCS axis ref pixel value key */
	"CTYPE#",      /* WCS axis type key  */
	"LTM#_#",      /* IRAF WCS axis rotation matrix key  */
	"LTV#",        /* IRAF WCS axis reference pixel key */
	"PC#_#",       /* WCS axis rotation matrix key  */
	"TC#_#",       /* WCS column rotation matrix key */
	"TCDLT#",      /* WCS column delta key */
        "TCNA#",       /* WCS column name key */
	"TCNAM#",      /* WCS column name key */
	"TCROT#",      /* WCS column rotation key */
	"TCRPX#",      /* WCS column reference pixel */
	"TCRVL#",      /* WCS column ref pixel value */
	"TCTYP#",      /* WCS column type key */
	"TCUNI#",      /* WCS column ref pixel value */
	"TP#_#",       /* WCS column rotation matrix key */
	"TWCS#",       /* WCS column coordinate name key */
	"WCSNAME"      /* WCS coordinate name key */
};


/* The following FITS keys do not appear at the DM level.  */
/* There are other candidates for this list, e.g. EQUINOX .*/
/* The keys must be in alphabetical order.                 */
/* WARNING: careful of using full word keywords in this    */
/*          list because they may refer only to header     */
/*          info, and not a value used in processing       */

static char *FITS_PROHIBITED_KEYWORDS[]=
{
        "#DSFORM#",   /* DSS keys */
        "#DSREF#",
        "#DSTYP#",
        "#DSUNIT#",
        "#DSVAL#",
	"BITPIX",		
        "BLANK",
	"BLOCKED",
        "BSCALE",
        "BUNIT",
        "BZERO",
	"CD#_#",       /* WCS axis rotation matrix key  */
	"CDELT#",      /* WCS axis delta key */
	"CNAME#",      /* WCS axis name key  */
        "COMMENT",
        "CPREF",
	"CROTA#",      /* WCS axis rotation key */
	"CRPIX#",      /* WCS axis reference pixel key */
	"CRVAL#",      /* WCS axis ref pixel value key */
	"CTYPE#",      /* WCS axis type key  */
        "CUNIT#",     
        "DFORM#",
        "DSFORM#",
        "DSREF#",
        "DSTYP#",
        "DSUNIT#",
        "DSVAL#",
        "DTYPE#",
        "DUNIT#",
        "DVAL#",
	"END",
        "EQUI#",                /* EQUINOX on column */
	"EXTEND",
	"EXTNAME",
	"EXTVER",
	"GCOUNT",		/* RANDOM GROUPS are not supported */
	"GROUPS",		/* RANDOM GROUPS are not supported */
	"HDUNAME",
        "HISTORY",
        "LATP#",                /* WCS parameter list key */
	"LONGSTRN", 
	"LONP#",                /* WCS parameter list key */
	"LTM#_#",               /* IRAF WCS axis rotation matrix key  */
	"LTV#",                 /* IRAF WCS axis reference pixel key */
        "MENAM#",               /* DM grouping key */
        "METYP#",               /* DM grouping key */
        "MFORM#",               /* DM grouping key */
        "MTYPE#",               /* DM grouping key */
	"NAXIS",		/* regular & sequenced keyword */
	"NAXIS#",		/* regular & sequenced keyword */
	"PC#_#",		/* WCS axis rotation matrix key  */
	"PCOUNT",
	"PPRO#",                /* WCS parameter list key */
	"PROP#",                /* WCS parameter list key */
        "RADE#",                /* WCS parameter list key */ 
	"SIMPLE",
	"TBCOL#",		/* sequenced keyword */
	"TC#_#",                 /* WCS column rotation matrix key */
	"TCDLT#",                /* WCS column delta key */
	"TCNA#",                 /* WCS column name key */
	"TCNAM#",                /* WCS column name key */
	"TCROT#",                /* WCS column rotation key */
	"TCRPX#",                /* WCS column reference pixel */
	"TCRVL#",                /* WCS column ref pixel value */
	"TCTYP#",                /* WCS column type key */
	"TCUNI#",                /* WCS column unit key */
	"TDBIN#",		/* sequenced keyword */
	"TDIM#",			/* sequenced keyword */
	"TDISP#",		/* sequenced keyword */
	"TDNULL#",		/* sequenced keyword */
	"TFIELDS",
	"TFORM#",		/* sequenced keyword */
	"THEAP#",
	"TLMAX#",		/* sequenced keyword */
	"TLMIN#",		/* sequenced keyword */
	"TNULL#",		/* sequenced keyword */
	"TP#_#",		/* WCS column rotation matrix key */
	"TSCAL#",		/* sequenced keyword */
	"TTYPE#",		/* sequenced keyword */
	"TUNIT#",		/* sequenced keyword */
	"TWCS#",                /* WCS column coordinate name key */
	"TZERO#",		/* sequenced keyword */
	"WCSNAME",              /* WCS coordinate name key */
	"XTENSION"
};


/***************************************************************/
/*                                                             */
/* Function:  ftp_key_private_fits                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*   Determine if keyword is private FITS keyword              */
/*                                                             */
/* Input parameters:                                           */
/*         const  char  *name                                  */
/*                                                             */
/* Output parameters:                                          */
/*         none                                                */
/*                                                             */
/* Returns:                                                    */
/*         dmBool - dmTRUE or dmFALSE                          */
/*                                                             */
/***************************************************************/
dmBool ftp_key_private_fits( const char *name )
{
  char buf[FITS_MAXLEN];
  char *ptr = NULL;

  long size = sizeof( FITS_PROHIBITED_KEYWORDS ) / UT_SZ_P;
  long k    = 0;
  long val  = 0;

  /* Optimize common case */
  if ( cs_is_blank( name ))
    return UT_TRUE;

  if ( cs_eq( name, "COMMENT" ) || cs_eq( name, "HISTORY" ))
    return UT_TRUE;

  cs_strip_digits( name, buf, FITS_MAXLEN );

  val = ar_match_sorted_c( buf, FITS_PROHIBITED_KEYWORDS, size );

  if ( val > 0 )
    return UT_TRUE;

  /* Special search for new WCS rules: remove last character */
  ptr = buf;
  k   = strlen( ptr );

  if ( k > 2 && isalpha( ptr[k-1] ) && ptr[k-2] == '#' ) 
  {
    ptr[k-1] = '\0';
    size     = sizeof( FITS_WCS_KEYWORDS ) / UT_SZ_P;

    if ( ar_match_sorted_c( ptr, FITS_WCS_KEYWORDS, size ) > 0 )
      return UT_TRUE;
  }

  return UT_FALSE;
}




