/*                                                                
**  Copyright (C) 2002-2012  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftrows.c                                                    */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmilib.h"
#include "ftkernelP.h"
#include "ftkernel.h"

#define XDEBUG 1



/***************************************************************/
/*                                                             */
/* Function:   ftColRead                                       */
/*                                                             */
/* Description:                                                */
/*   This is the main routine to get data - called by the      */
/*   kernel.                                                   */
/*                                                             */
/*  Read row1 to row2 of vector into buffer of size nrows.     */
/*  Vector cols are copied pair by pair.                       */
/*                                                             */
/*  Input parameters:                                          */
/*          dmDescriptor  *vector                              */
/*          long          *llpix                               */
/*          long          *urpix                               */
/*          long          *dpix                                */
/*          long           row1                                */
/*          long           row2                                */
/*                                                             */
/*  Output parameters:                                         */
/*          void          *data                                */
/*                                                             */
/*  Returns:                                                   */
/*          dmErrCode      status                              */
/*                                                             */
/***************************************************************/
dmErrCode ftColRead( dmDescriptor* vector, long* llpix, long* urpix,
                     long* dpix, long row1, long row2, void* data )
{
  dmDataType     type;
  dmDescriptor  *cpt_dd    = NULL;
  dmBlock       *block     = NULL; 
  dmDescriptor  *length_dd = NULL;   /* Buffer column for varlen */
                                     /* array row lengths        */

  ftCol    *ftcol   = NULL;
  ftBlock  *ftblock = NULL;
  fitsfile *fptr    = NULL;

  void *ptr      = NULL;

  char *iptr     = NULL; 
  char *data_ptr = NULL;
  char *optr     = NULL;
  char *name     = NULL;

  double tscal = 1.0;
  double tzero = 0.0;
  double crpix = 0.0;
  double crval = 0.0;

  long *axes = NULL;
  long *vs   = NULL;       /* Pixel vector start */
  long *ve   = NULL;       /* Pixel vector end */
  long *incr = NULL;       /* Step direction */
  long  vector_buffer_size = 0;
  long  vdim        = 0;
  long  cptNum      = 0;
  long  nelts       = 0;
  long  nbytes      = 0;   /* Bytes for a single row of a single component */
  long  raw_row     = 0;
  long  stop_row    = 0;
  long  nrows       = 0;
  long  item_offset = 0;   /* Loop iterator */
  long  size        = 0;   /* Size of a single scalar item */
  long  nvals       = 0;
  long  naxes       = 0; 
  long  flag        = 0;
  long  ftcolno     = 0;   /* FITS column number */
  long  drows       = 0;
  long  item_size   = 0;
  long  length      = 0;

  int  status       = UT_TRUE;
  int  is_scaled    = UT_FALSE;
  int  is_var       = UT_FALSE;  /* Is descriptor a variable-length array? */
  int  qalloc       = 0;
  int  is_vector    = 0; 
  int  is_special   = 0;
  int  is_image     = 0;
  int  null         = 0;
  int  ftstatus     = 0;
  int  ii           = 0;
  int  using_local_buffer = 0; 



  if ( !vector )
    return dmNOMOREROWS;

  name = dmi_descriptor_get_name( vector );
  vdim = dmi_descriptor_get_element_dim( vector );
  type = dmi_descriptor_get_datatype( vector );

  size = dmi_descriptor_get_bytes( vector );

  data_ptr = data;
  block = dmi_descriptor_get_block( vector );
  fptr = ftp_fptr( block );
  ftblock = (ftBlock*)dmi_block_get_kb( block );
  ftp_block_reset( block );

#if DEBUG
 printf( "VDIM = %ld\n", vdim );
#endif

  nrows = row2 - row1 + 1;

  is_image   = ( dmi_block_get_raw_type(block ) == dmIMAGE ); 
  is_vector  = ( vdim > 1 );
  is_special = ( type == DT_LOGICAL || type == DT_CHAR );

  // Code Change Bug 11332 October 28, 2008
  //  String vectors need to have enough buffer space allocated
  //  so use "size" , which is the number of bytes

  if ( type == DT_CHAR && !is_vector)
    item_size = UT_SZ_P;
  else
    item_size = size;
 
  
  /* If we have to reorder the data, allocate a suff large buffer to do so */
  using_local_buffer = is_vector || is_special;

  if ( using_local_buffer )
  {
    nvals = dmi_descriptor_get_nvals( vector );
    vector_buffer_size = nrows * item_size * nvals + 1;

    if ( ftblock->vector_buffer_size < vector_buffer_size )
    {
      ftblock->vector_buffer = (char*)realloc( ftblock->vector_buffer, vector_buffer_size );
      ftblock->vector_buffer_size = vector_buffer_size;
#if DEBUG
      printf( "ALLOC FVB %ld NV %ld\n", vector_buffer_size, nvals );
#endif
    }

    // Set data_ptr to the local buffer instead of the output parameter
    // buffer "data"

    data_ptr = ftblock->vector_buffer;
  }

 
  /* Is the data a scaled image? Make CFITSIO do the scaling */
  if ( is_image )
  {
    flag      = dmi_descriptor_get_flag( vector );
    is_scaled = ( flag & DM_FLAG_COORD );

    if ( type != dmUSHORT && type != dmULONG )
    {
      if ( is_scaled && vdim == 1 )
      { 
        dmi_coord_get_transform_d( vector, &crpix, &crval, &tscal, vdim );
        tzero = crval - tscal * crpix;
      }
  
      // Code Change  Bug 12880  January 10, 20120
      //  Set bscale for all
      fits_set_bscale( fptr, tscal, tzero, &ftstatus );
    }
  }
 

  /* Loop over components of the vector (often only 1) */
 
  for ( cptNum = 1; cptNum <= vdim; cptNum++)
  {
#if DEBUG
    printf( "FTCOL CPT \n" );
#endif

    cpt_dd = dmi_descriptor_get_cpt( vector, cptNum );
    length = dmi_descriptor_get_length( cpt_dd );
    nvals  = dmi_descriptor_get_nvals( cpt_dd );
    nbytes = size * nvals;

    if ( is_image )
    {
      ftcol   = NULL;
      ftcolno = 1;
    }
    else
    {
      ftcol = dmi_descriptor_get_kdd( cpt_dd );

      if ( !ftcol )
      {
        printf( "BAD FTCOL for %s\n", dmi_descriptor_get_name( cpt_dd ) );
        sprintf( dm_errbuf, "Descriptor %s is not a valid FITS column",
                 dmi_descriptor_get_name( cpt_dd ));
        dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
        return DM_ERR_GENERIC;
      }

      ftcolno = ftcol->no;
      is_var  = dmi_descriptor_is_varlen( cpt_dd );
    } 

    raw_row = row1;

    // 'ptr' is start of data buffer, either the output buffer 'data' or
    //  the local use buffer pointed to by 'data_ptr'
 
    ptr = data_ptr + (cptNum - 1) * nbytes * nrows;

     
    /* Q: how do we find the offsets for a given cpt dd? With a dmi routine? */
    ftstatus = 0;

    if ( is_image || llpix )
    {
      /* Need to specify subset - Do we want to do this for array columns too? */
      /* same read_subset call works for columns!                              */
      /* alloc vs and ve at descriptor time?                                   */
      /* However, better to have these as explicit arguments to the routine    */
      /* to make the interface clearer.                                        */
      /*                                                                       */
      /* If it's an image and we have defined a pixel range, just use it.      */
      /* If it's an image with no pixel range, we query the dimensions.        */
      /* If it's a table, we must extend the pixel range array to add the row  */
      /* range.                                                                */

      qalloc = 0;
      vs     = llpix;
      ve     = urpix;
      incr   = dpix;
      naxes  = dmi_descriptor_get_array_dimensions( cpt_dd, &axes );

      if ( !vs || !is_image )
      {
        qalloc = 1;
        vs   = calloc( naxes+1, UT_SZ_I );
        ve   = calloc( naxes+1, UT_SZ_I );
        incr = calloc( naxes+1, UT_SZ_I );

        if ( !llpix )
        {
          for ( ii = 0; ii < naxes; ii++ )
          {
            vs[ii] = 1; 
            ve[ii] = axes[ii];
            incr[ii] = 1;
          }
        }
        else
        {
          for ( ii = 0; ii < naxes; ii++ )
          {
            vs[ii]   = llpix[ii];
            ve[ii]   = urpix[ii];
            incr[ii] = dpix[ii];
          }
        }

        /* The final axis is the table row number. */
        vs[naxes]   = row1;
        ve[naxes]   = row2;
        incr[naxes] = 1;
      }
	 
      switch( type )
      {
        case DT_UINT1:
          fits_read_subset_byt( fptr, 1, (int)naxes, axes,  /* CFITSIO */
  			        vs, ve, incr, 0,
				(unsigned char*)ptr, &null, &ftstatus); 
	  break;

        case DT_UINT2:
          fits_read_subset_usht( fptr, 1, (int)naxes, axes,  /* CFITSIO */
				 vs, ve, incr, 0,
				 (unsigned short*)ptr, &null, &ftstatus);
	  break;

        case DT_INT2:
          fits_read_subset_sht( fptr, 1, (int)naxes, axes,  /* CFITSIO */
				vs, ve, incr, 0,
				(short*)ptr, &null, &ftstatus);
	     
	  break;

        case DT_INT4:
          fits_read_subset_lng( fptr, 1, (int)naxes, axes,  /* CFITSIO */
				vs, ve, incr, 0,
				(long*)ptr, &null, &ftstatus);
	  break;

        case DT_UINT4:
          fits_read_subset_ulng( fptr, 1, (int)naxes, axes,  /* CFITSIO */
				 vs, ve, incr, 0,
				 (unsigned long*)ptr, &null, &ftstatus);
	  break;

        case DT_REAL4:
          fits_read_subset_flt( fptr, 1, (int)naxes, axes,  /* CFITSIO */
				vs, ve, incr, 0, 
				(float*)ptr, &null, &ftstatus);
	  break;

        case DT_REAL8:
          fits_read_subset_dbl( fptr, 1, (int)naxes, axes,  /* CFITSIO */
				vs, ve, incr, 0,
				(double*)ptr, &null, &ftstatus);
	  break;

        default:
          sprintf( dm_errbuf, "Unknown data type reading image descriptor %s",
                   dmi_descriptor_get_name( cpt_dd ));
	  dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
	  break;
      }

      if ( qalloc )
      {
        free( incr );
        free( vs );
        free( ve );
      }

      free( axes );
    }
    else
    {
      // reading TABLE data

      nelts    = nrows * nvals;
      drows    = nrows;
      stop_row = row1;

      if ( is_var )
      {
        memset( ptr, 0, nrows * nvals * size );
        drows = 1;
        stop_row = row2;
	     
        length_dd = dmi_buffer_get_varlen_length_col( block, cpt_dd );
	     
        if ( !length_dd )
        {
          sprintf( dm_errbuf, "Error in buffering variable-length array descriptor %s; all row-length information will be discarded",
                   dmi_descriptor_get_name( cpt_dd ) );
          dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
		 
          /* Continue processing.  If length_dd is NULL, then calls to                   */
	     /* dmi_buffer_write_varlen_length() will fail harmlessly, and no row length */
          /* information will be saved.                                                  */
        }
      }

      if ( ftcol && ftcol->scaled )
        fits_set_tscale( fptr, ftcolno, 1.0, 0.0, &ftstatus );  /* CFITSIO */


      /* Usually only one iteration except for variable length arrays */

      for ( raw_row = row1; raw_row <= row2; raw_row += drows )
      {
        if ( is_var )
        {
          long idum = 0;

          fits_read_descript( fptr, ftcolno, raw_row, &nelts, &idum, &ftstatus );  /* CFITSIO */
          ptr = data_ptr + (cptNum-1) * nbytes * nrows + ( raw_row - row1 ) * nbytes;
		 
          /* Store number of elements in this variable-length array row */
          dmi_buffer_write_varlen_length( block, cpt_dd, length_dd, ptr, nelts );
        }

        switch( type )
        {
          case DT_REAL8:    
            fits_read_col_dbl( fptr, ftcolno, raw_row, 1, nelts, 0, (double*)ptr, &null, &ftstatus ); /* CFITSIO */
            break;

          case DT_REAL4:  
            fits_read_col_flt( fptr, ftcolno, raw_row, 1, nelts, 0, (float*)ptr, &null, &ftstatus ); /* CFITSIO */
            break;

          case DT_INT4:
            fits_read_col_lng( fptr, ftcolno, raw_row, 1, nelts, 0, (long*)ptr, &null, &ftstatus ); /* CFITSIO */ 
            break;

          case DT_INT2:
            fits_read_col_sht( fptr, ftcolno, raw_row, 1, nelts, 0, (short*)ptr, &null, &ftstatus ); /* CFITSIO */
            break;

          case DT_LOGICAL:
            fits_read_col_log( fptr, ftcolno, raw_row, 1, nelts, 0, (char*)ptr, &null, &ftstatus ); /* CFITSIO */

            for ( item_offset = 0; item_offset < nelts; item_offset++ )
              ((int*)data)[item_offset] = (int)*((char*)ptr+item_offset);

            break;

          case DT_BIT:
            fits_read_col_byt( fptr, ftcolno, raw_row, 1, nelts, 0, (unsigned char*)ptr, &null, &ftstatus );  /* CFITSIO */
            break;

          case DT_UINT1:
            fits_read_col_byt( fptr, ftcolno, raw_row, 1, nelts, 0, (unsigned char*)ptr, &null, &ftstatus ); /* CFITSIO */
            break;

          case DT_UINT2:
            fits_read_col_usht( fptr, ftcolno, raw_row, 1, nelts, 0, (unsigned short*)ptr, &null, &ftstatus ); /* CFITSIO */
            break;

          case DT_UINT4:
            fits_read_col_ulng( fptr, ftcolno, raw_row, 1, nelts, 0, (unsigned long*)ptr, &null, &ftstatus ); /* CFITSIO */
            break;

          case DT_CHAR:
            /* CFITSIO requires an array of pointers; we pass this to contiguous memory */

            // Code Change Bug 11322  October 28, 2008
            //  If this is a vector, fill the use local buffer, not output buffer 'data'
            //  because data will need to be shuffled later.
            //  If this is not a vector, use the output buffer 'data' directly

            for ( item_offset = 0; item_offset < nelts; item_offset++ )
            {
              if ( is_vector)
                ((char**)ptr)[item_offset] = (char*)data_ptr + ((length+1) * item_offset * UT_SZ_C)+((cptNum-1)*nrows*nbytes);
              else
                ((char**)ptr)[item_offset] = (char*)data + ((length+1) * item_offset * UT_SZ_C)+((cptNum-1)*nrows*nbytes);

	      /* clear memory to be filled in this read */
	      memset( ((char**)ptr)[item_offset], 0, (length+1) );
            }

            // Read the data into a buffer              
            fits_read_col_str( fptr, ftcolno, raw_row, 1, nelts, "", (char**)ptr, &null, &ftstatus ); /* CFITSIO */
            break;

          default:
            sprintf( dm_errbuf, "Unknown data type reading column descriptor %s",
                     dmi_descriptor_get_name( cpt_dd ));
            dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
            break;
        }
      } 
    }

    if ( is_vector )
    {
      /* If the column is variable-length, we've been changing ptr, so */
      /* restore it to its base value                                  */

      // Code Change Bug 11322 october 28, 2008
      // Need to reset ptr tto point to local buffer if type is char strings

      if ( is_var || (type == DT_CHAR) )
        ptr = data_ptr + (cptNum-1) * nbytes * nrows;

      /* Rearrange vectors */
      for ( raw_row = 0; raw_row < nrows; raw_row++ )
      {  
        iptr = (char*)ptr + nbytes *  raw_row;
        optr = (char*)data + nbytes * ( vdim * raw_row + (cptNum-1));
        memcpy( optr, iptr, nbytes );     
      }  /* End loop on row for vectors */

    } /* End if using buffer */
  } /* End loop on cpt */


  // Check for any errors coming from cfitsio functions
  if ( ftstatus != FITS_KERNEL_OK )
  {
    dmDataset* ds = dmi_block_get_ds( block );
    ftp_ds_status( ds, ftstatus, "[ftColRead]" );
    status = UT_FALSE;
  }

  // Set tscal and tzero   
  if ( is_scaled && type != dmUSHORT && type != dmULONG )
  {
    tzero = 0.0;
    tscal = 1.0;
    fits_set_bscale( fptr, tscal, tzero, &ftstatus );
  }
 
  return status;
}


/***************************************************************/
/*                                                             */
/* Function:   ftColWrite                                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*          dmDescriptor  *vector                              */
/*          long          *llpix                               */
/*          long          *urpix                               */
/*          long          *dpix                                */
/*          long           row1                                */
/*          long           row2                                */
/*                                                             */
/*  Output parameters:                                         */
/*          void          *data                                */
/*                                                             */
/*  Returns:                                                   */
/*          dmErrCode      status                              */
/*                                                             */
/***************************************************************/
dmErrCode ftColWrite( dmDescriptor* vector, long* llpix, long* urpix,
                      long* dpix, long row1, long row2, void* data )
{
  dmDataType    type;
  dmBlock      *block     = NULL; 
  dmDescriptor *cpt_dd    = NULL;
  dmDescriptor *length_dd = NULL; /* Buffer column for varlen array row lengths */
  ftBlock      *ftblock   = NULL;
  ftCol        *ftcol     = NULL;
  fitsfile     *fptr      = NULL; 

  void* ptr = NULL;

  char **sptr = NULL; 
  char *logical_vals = NULL;
  char *name         = NULL;
  char *iptr         = NULL; 
  char *data_ptr     = NULL;
  char *optr         = NULL;

  long *axes  = NULL;
  long *vs    = llpix;
  long *ve    = urpix;
  long *dv    = NULL;

  long  vector_buffer_size = 0;
  long  item_offset = 0;
  long  item_size   = 0;
  long  naxes   = 0;
  long  vdim    = 0;
  long  cptNum  = 0;
  long  nbytes  = 0;       /* Bytes for a single row of a single component */
  long  raw_row = 0;
  long  nitems  = 0;
  long  nrows   = 0;
  long  size    = 0;       /* Size of a single scalar item */
  long  nvals   = 0;
  long  length  = 0;
  long  ii      = 0;
  long  drows   = 0;
  long  ftcolno = 0;       /* FITS column number */

  int  using_local_buffer = 0;
  int  qalloc     = 0;
  int  is_vector  = 0;
  int  is_special = 0;
  int  ftstatus   = 0;
  int  is_image   = 0;
  int  is_var     = 0;     /* Is descriptor a variable-length array? */


  if ( !vector )
    return dmNOMOREROWS;

  dv = dpix;
  name = dmi_descriptor_get_name( vector );
 
  vdim = dmi_descriptor_get_element_dim( vector );
  type = dmi_descriptor_get_datatype( vector );

  /* Read unscaled descriptor then scale them */
 
  size     = dmi_descriptor_get_bytes( vector );
  length   = dmi_descriptor_get_length( vector );
  data_ptr = data;
  ptr      = data;
  block    = dmi_descriptor_get_block( vector );
 
#if DEBUG
  printf( "FTROWS %s %s %ld:%ld\n", dmi_block_get_name( block ), dmi_descriptor_get_name( vector ), row1, row2 );
#endif
 
  is_image = ( dmi_block_get_raw_type(block ) == dmIMAGE ); 
  fptr     = ftp_fptr( block );
  ftblock  = (ftBlock*)dmi_block_get_kb( block );
  ftp_block_reset( block );

#if DEBUG
 printf( "VDIM = %ld\n", vdim );
#endif

  nrows = row2 - row1 + 1;

  if ( nrows < 0 )
  {
    printf( "INTERNAL ERROR ROW RANGE = %ld:%ld\n", row1, row2 );
  }

  is_vector = ( vdim > 1 );
 
  naxes = dmi_descriptor_get_array_dimensions( vector, &axes );

  if ( is_image && !vs )
  {
    /* Check for empty image */
    if ( naxes == 0 )
      return dmSUCCESS;
     
    qalloc = 1;
    vs = calloc( naxes+1, UT_SZ_I );
    ve = calloc( naxes+1, UT_SZ_I );

    for ( ii = 0; ii < naxes; ii++ )
    {
      vs[ii] = 1; 
      ve[ii] = axes[ii];
    }

    /* The final axis is the table row number. Not needed right now */
    vs[naxes] = 1;
    ve[naxes] = 1;
  }
 
  if ( is_image )
  {
    switch( type )
    {
      case DT_UINT1:
        fits_write_subset_byt( fptr, 1, (int)naxes, axes,  /* CFITSIO */
				vs, ve, (unsigned char*)ptr, &ftstatus);
        break;

      case DT_UINT2:
        fits_write_subset_usht( fptr, 1, (int)naxes, axes,  /* CFITSIO */
				 vs, ve, (unsigned short*)ptr, &ftstatus);
	break;

      case DT_INT2:
        fits_write_subset_sht( fptr, 1, (int)naxes, axes,  /* CFITSIO */
				vs, ve, (short*)ptr, &ftstatus);
	break;

      case DT_INT4:
        fits_write_subset_lng( fptr, 1, (int)naxes, axes,  /* CFITSIO */
				vs, ve,	(long*)ptr,  &ftstatus);
	break;

      case DT_UINT4:
        fits_write_subset_ulng( fptr, 1, (int)naxes, axes,  /* CFITSIO */
				 vs, ve, (unsigned long*)ptr, &ftstatus);
	break;

      case DT_REAL4:
        fits_write_subset_flt( fptr, 1, (int)naxes, axes,  /* CFITSIO */
				vs, ve,	(float*)ptr,  &ftstatus);
	break;

      case DT_REAL8:
        fits_write_subset_dbl( fptr, 1, (int)naxes, axes,   /* CFITSIO */
				vs, ve,	(double*)ptr, &ftstatus);
	break;

      default:
        sprintf( dm_errbuf, "Unknown data type reading image descriptor %s",
                 dmi_descriptor_get_name( vector ));
        dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
        break;
    }

    if ( qalloc )
    {
      free( vs );
      free( ve );
    }
  }
  else
  {
    // TABLE data
 
    nvals      = dmi_descriptor_get_nvals( vector );
    is_special = ( type == DT_CHAR );

    // Code Change Bug 11322  October 23, 2008
    //  String vectors need to have enough buffer space allocated
    //  so use "size" , which is the number of bytes

    if ( type == DT_CHAR && !is_vector)
      item_size = UT_SZ_P;
    else
      item_size = size;
     
    using_local_buffer = is_vector || is_special;


    // Vector data needs to be shuffled in the buffer before being output
    //  Allocate a sufficiently large buffer to hold the data

    if ( using_local_buffer )
    {
      vector_buffer_size = nrows * item_size * nvals + 1;
	 
      if ( ftblock->vector_buffer_size < vector_buffer_size )
      {
        ftblock->vector_buffer = (char*)realloc( ftblock->vector_buffer, vector_buffer_size );
        ftblock->vector_buffer_size = vector_buffer_size;
      }

      // Set data_ptr to the local buffer instead of the output parameter
      // buffer "data"

      data_ptr = ftblock->vector_buffer;
    }


    // loop over each component
    for ( cptNum = 1; cptNum <= vdim; cptNum++ )
    {
#if DEBUG
      printf( "FTCOL CPT \n" );
#endif

      cpt_dd = dmi_descriptor_get_cpt( vector, cptNum );
      nvals  = dmi_descriptor_get_nvals( cpt_dd );
      nbytes = size * nvals;
      nitems = nvals * nrows;

      if ( nitems < 0 )
      {
        printf( "INTERNAL ERROR nvals = %ld nrows = %ld %s\n", nvals, nrows,
                dmi_descriptor_get_name( cpt_dd ) );
      }

      ftcol = dmi_descriptor_get_kdd( cpt_dd );

      if ( !ftcol )
      {
        sprintf( dm_errbuf, "Descriptor %s is not a valid FITS column",
                 dmi_descriptor_get_name( cpt_dd ));
        dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
        return DM_ERR_GENERIC;
      }

      is_var  = dmi_descriptor_is_varlen( cpt_dd );
      ftcolno = ftcol->no;

      // Set pointer to start of output data buffer
      ptr     = data_ptr + (cptNum-1) * nbytes * nrows;

      if ( !ptr )
      {
        sprintf( dm_errbuf, "No data buffer for column %s", dmi_descriptor_get_name( cpt_dd ));
        dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
        return DM_ERR_GENERIC;
      }

      /* Rearrange vectors */
      if ( is_vector )
      {
        for ( raw_row = 0; raw_row < nrows; raw_row++ )
        {  
          iptr = (char*)ptr + nbytes *  raw_row;
          optr = (char*)data + nbytes * ( vdim * raw_row + (cptNum-1));
          memcpy( iptr, optr, nbytes );
        }
      }

      ftstatus = 0;
      raw_row  = row1;
      drows    = nrows;

      if ( is_var )
      {
        drows = 1;
        length_dd = dmi_buffer_get_varlen_length_col( block, cpt_dd );

        if ( !length_dd )
        {
          sprintf( dm_errbuf, "Error in buffering variable-length array descriptor %s; all row-length information will be discarded", dmi_descriptor_get_name( cpt_dd ) );
          dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
		 
          /* Continue processing, taking care to ignore row-length info when */
          /* dmi_buffer_read_varlen_length() fails.                          */
        }
      }


      // loop over rows to get data elements

      for( raw_row = row1; raw_row <= row2; raw_row += drows )
      {
        if ( is_var )
        {
          /* Get number of items in this variable-length array row.  nitems < 0 */
          /* signifies an error, so only use it if it's non-negative.           */

          nitems = dmi_buffer_read_varlen_length( block, cpt_dd, length_dd, ptr );

          if ( nitems < 0 )
            nitems = nvals;
        }
	     

        switch( type )
        {
          case DT_REAL8:    
            fits_write_col_dbl( fptr, ftcolno, raw_row, 1, nitems, (double*)ptr,  &ftstatus );  /* CFITSIO */
	    break;

          case DT_REAL4:  
            fits_write_col_flt( fptr, ftcolno, raw_row, 1, nitems, (float*)ptr,  &ftstatus );/* CFITSIO */
	    break;

          case DT_INT4:
	    fits_write_col_lng( fptr, ftcolno, raw_row, 1, nitems, (long*)ptr, &ftstatus ); /* CFITSIO */
	    break;

          case DT_INT2:
            fits_write_col_sht( fptr, ftcolno, raw_row, 1, nitems, (short*)ptr, &ftstatus ); /* CFITSIO */
            break;

          case DT_UINT4:
            fits_write_col_ulng( fptr, ftcolno, raw_row, 1, nitems, (unsigned long*)ptr, &ftstatus ); /* CFITSIO */
            break;

          case DT_UINT2:
            fits_write_col_usht( fptr, ftcolno, raw_row, 1, nitems, (unsigned short*)ptr, &ftstatus ); /* CFITSIO */
            break;

          case DT_UINT1:
            fits_write_col_byt( fptr, ftcolno, raw_row, 1, nitems, (unsigned char*)ptr, &ftstatus ); /* CFITSIO */
            break;

          case DT_BIT:
            fits_write_col_byt( fptr, ftcolno, raw_row, 1, nitems, (unsigned char*)ptr, &ftstatus ); /* CFITSIO */
            break;

          case DT_LOGICAL:
	    logical_vals = (char*)malloc( nitems );

            for ( ii =0; ii < nitems; ii++ )
              logical_vals[ii] = ((dmBool*)ptr)[ii];

            fits_write_col_log( fptr, ftcolno, raw_row, 1, nitems, (char*)logical_vals, &ftstatus ); /* CFITSIO */
            free( logical_vals );
            break;

          case DT_CHAR:
            // Code Change  Bug 11322 October 23, 2008
            //   cfitsio expecting an array of pointers, that are pointing to place in
            //   data buffer to write data
            //   Allocate a local pointer array here

            sptr = (char**)malloc(nitems * UT_SZ_P);

            for ( ii = 0; ii < nitems; ii++)
            {
              item_offset = (length+1) * ii;

              // If this is a vector, data will be in the local use buffer
              // otherwise it is in the input buffer, 'data'

              if (is_vector)
                sptr[ii] = (char*)ptr + (item_offset * UT_SZ_C);
              else
                sptr[ii] = (char*)data + (item_offset * UT_SZ_C);
            }

	    /* MCD 20090805 */
	    /* CFITSIO will write a single space for empty strings unless you specify the Null.. */
	    /* Use default of an empty string, but allow for user defined.                       */
	    char strnul[dmMAXSTR] = "";
	    dmi_descriptor_get_null( cpt_dd, (void*)strnul );

            /* write data.. use call which includes null checks. */
	    // fits_write_col_str( fptr, ftcolno, raw_row, 1, nitems, (char**)sptr, &ftstatus );
	    fits_write_colnull_str( fptr, ftcolno, raw_row, 1, nitems, (char**)sptr, strnul, &ftstatus );

            // clean up
            free(sptr);
            break;

          default:
            printf( "BAD DT\n" );
            break;

        } // end switch

        if ( is_var )
          ptr = (char*)ptr + nbytes;

      } // end for (row ...
    } // end for (cptNum ...    
  } // end else - table data

  // one last clean-up chore
  if ( axes )
    free( axes );

 return dmSUCCESS;
}


