/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftscan.c                                                    */
/*                                                               */
/*   Functions to access datasets                                */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmilib.h"
#include "ftkernelP.h"

/***************************************************************/
/*                                                             */
/* Function:   ftp_raw_block_alloc                             */
/*                                                             */
/* Description:                                                */
/*          Alloc a new block. Set block type and name.        */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDataset  *ds   - dataset                       */
/*           const long   no   - block number                  */
/*            dmBlockTYpe type - block type                    */
/*           const char  *name - block name                    */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             long   blockno  - block number                  */
/*                                                             */
/***************************************************************/
long ftp_raw_block_alloc( dmDataset* ds, const long no, const dmBlockType type, 
                          const char* name )
{
  long blockno = 0;
 
  blockno = dmi_dataset_block_alloc( ds, no );
  dmi_dataset_set_block_type( ds, no, type );
  dmi_dataset_set_block_name( ds, no, name );
  return blockno;
}

/***************************************************************/
/*                                                             */
/* Function:   ftp_dataset_add_block                           */
/*                                                             */
/* Description:                                                */
/*          Add new block to dataset                           */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDataset  *ds   - dataset                       */
/*            long        hdu  - block number                  */
/*            int         ftHduType - hdu type                 */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void ftp_dataset_add_block( dmDataset* ds, long hdu, int ftHduType )
{
  ftDataset *ftds = (ftDataset*)dmi_dataset_get_kds( ds );

  if ( hdu > ftds->nmax )
  {
    ftds->nmax = 2 * ( hdu + 10 );
    ftds->hdutype = realloc( ftds->hdutype, ftds->nmax * UT_SZ_I );
  }

  if ( hdu > ftds->n )  
    ftds->n = hdu;

  ftds->hdutype[hdu-1] = ftHduType;
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_dataset_scan                                */
/*                                                             */
/* Description:                                                */
/*    At file open time, scan the file to look for all the     */
/*    blocks in it. After the scan, move back to the first HDU */
/*    Check the first HDU for the case NAXIS=0, defining a NULL*/
/*    PRIMARY HDU.                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDataset  *ds   - dataset                       */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*             void                                            */
/*                                                             */
/***************************************************************/
void ftp_dataset_scan( dmDataset* ds )
{
  ftDataset  *ftds = NULL;
  dmBlockType type;
  char       *name = NULL;

  long blockno = 0;
  long current = 1;
  long no      = 0;
  int ftstatus = FITS_KERNEL_OK;
  int fthdutype = 0;
  int naxis    = 0;
  int hdu      = 0;


  ftds = (ftDataset*)dmi_dataset_get_kds( ds );
  ftds->current = 1;

  while( ftstatus == FITS_KERNEL_OK )
  {
    hdu = ftds->current; 
    fits_movabs_hdu( ftds->fptr, hdu, &fthdutype, &ftstatus );  /* CFITSIO */

    if ( ftstatus == FITS_KERNEL_OK ) 
    {
      no = current;
      name    = ftp_read_block_name( ftds->fptr, ftds->current );
      type    = ftp_block_type( fthdutype );
      blockno = ftp_raw_block_alloc( ds, no, type, name );
      ftp_dataset_add_block( ds, no, fthdutype );
      free( name );
      ftds->current++;
      ++current;

      if ( hdu == 1 )
      {
        /* Check naxis for null primary HDU */
        fits_read_key( ftds->fptr, TINT, "NAXIS", &naxis, NULL, &ftstatus );  /* CFITSIO */
        ftds->null = ( naxis == 0 );
      }
    }
  }

  if( ftstatus != END_OF_FILE )
    ftp_ds_status( ds, ftstatus, "[ftp_dataset_scan]" );
  else 
  {
    ftds->current = 1;
    ftp_move_to_current_hdu( ds );
  }
}

