/*                                                                
**  Copyright (C) 2002-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftsubspace.c                                                */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmilib.h"
#include "ftkernel.h"
#include "ftkernelP.h"

/* For some reason this isn't being picked up */
#define fits_update_key_longstr ffukls

int ftp_tstart_read( dmBlock* block, double* tstartp, double* tstopp );

dmBool ftp_subspace_is_gti( char* tabname );

regRegion* dmi_mask_region_from_string( dmDescriptor* ssd, const char* dsref, const char* dsval );



/***************************************************************/
/*                                                             */
/* Function:   ftp_get_ftss                                    */
/*                                                             */
/* Description:                                                */
/*        Get the kernel-specific subspace col for a           */
/*            descriptor, creating it if needed.               */
/*            Manages the index of DSTYPn keywords.  If n is   */
/*            negative, increment the DSTYP index; otherwise   */
/*            use n as the index.                              */
/*                                                             */
/*  Input parameters:                                          */
/*                dmBlock      *block                          */
/*                dmDescriptor *dd                             */
/*                long          n                              */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                  ftSubspaceCol *ftss                        */
/*                                                             */
/***************************************************************/
ftSubspaceCol* ftp_get_ftss( dmBlock* block, dmDescriptor* dd, long n )
{
  ftSubspaceCol *ftss    = NULL;
  ftBlock       *ftblock = (ftBlock*)dmi_block_get_kb( block );

  ftss = dmi_descriptor_get_kdd( dd );

  if ( ftss && ftss->type != FTK_SUBSPACE )
  {
     sprintf( dm_errbuf, "INTERNAL ERROR FTK SUB %s %ld \n", dmi_descriptor_get_name( dd ), ftss->type );
     dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
  } 
  
  if ( !ftss )
  {
    ftss = ftp_ftss_alloc( ftblock, n );
     
    dmi_descriptor_set_kdd( dd, ftss );
  } 

  return ftss;
}



/***************************************************************/
/*                                                             */
/* Function:   ftp_ftss_free                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                ftSubspaceCol *ftss                          */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                 nothing                                     */
/*                                                             */
/***************************************************************/
void ftp_ftss_free( ftSubspaceCol* ftss )
{
  if ( !ftss )
    return;

  free( ftss );
}



/***************************************************************/
/*                                                             */
/* Function:   ftp_ftss_alloc                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                ftblock *ftblock                             */
/*                long     n                                   */
/*                                                             */
/*  Output parameters:                                         */
/*                none                                         */
/*                                                             */
/*  Returns:                                                   */
/*                nothing                                      */
/*                                                             */
/***************************************************************/
ftSubspaceCol* ftp_ftss_alloc( ftBlock* ftblock, long n )
{
  ftSubspaceCol *ftss = NULL;

  ftss = calloc( 1, sizeof( ftSubspaceCol ));
  ftss->type = FTK_SUBSPACE;

  if ( n < 0 )
  {
    ftss->index = ++ftblock->max_dstyp;  
  }
  else 
  {
    ftss->index = n;

    if ( n > ftblock->max_dstyp ) 
      ftblock->max_dstyp = n;
  }

  return ftss;
}



/***************************************************************/
/*                                                             */
/* Function:   ftSubspaceDescriptorWrite                       */
/*                                                             */
/* Description:                                                */
/*        This routine writes subspace keywords.  It does not  */
/*        write GTI tables.                                    */
/*                                                             */
/*  Input parameters:                                          */
/*                ftblock *ftblock                             */
/*                long     n                                   */
/*                                                             */
/*  Output parameters:                                         */
/*                none                                         */
/*                                                             */
/*  Returns:                                                   */
/*                nothing                                      */
/*                                                             */
/***************************************************************/
void ftSubspaceDescriptorWrite( dmDescriptor *dd, long cptno,
                                char *value, char *tabname )
{
  dmBlock    *block = NULL;
  dmuKeyList  keylist;
  dmDataType  type;
  dmDescriptorType dtype;
  
  ftBlock       *ftblock = NULL;
  ftSubspaceCol *ftss    = NULL;
  
  char keyname[CARD_SIZE];
  char tabval[TEXT_SIZE];
  char fullname[CARD_SIZE];
  char tvalue[2] = "I";

  char *desc     = " ";
  char *valdesc  = "DM Keyword: Descriptor value.";
  char *fdesc    = "DM Keyword: Descriptor datatype.";
  char *name     = NULL;
  char *valuep   = value;
  char *table_value = "TABLE";
  char *tvaluep  = NULL;
  char *unitp    = NULL;
  int   do_table = 0;
  
  
  if ( !dd )
    return;
  
  type =  dmi_descriptor_get_datatype( dd );
  dtype = dmi_descriptor_get_type( dd );

  if ( dtype != dmSUBSPACE )
  {
    sprintf( dm_errbuf, "Descriptor %s is not a subspace descriptor", dmi_descriptor_get_name( dd ));
    dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
    return;
  }

  do_table = !cs_is_blank( tabname );
  block    = (dmBlock*)dmi_descriptor_get_block( dd );
  ftblock  = (ftBlock*)dmi_block_get_kb( block );
  keylist  = ftblock->keylist;
  
  
  ftss = ftp_get_ftss( block, dd, -1 );

  if ( ftss->index == 0 )
  {
    /* We already had an ftss and its index is zero */
    /* No DSTYP keyword used? EG Default time */  
      
    if ( cptno == 1 )
      ftss->sum++;

    if ( do_table && ftss->sum > 1 )
      ftss->index = ++ftblock->max_dstyp;
    else
      return;
  } 
  
  
  /* Have we already written DSTYPn? */
  if ( !ftss->written )
  {
    sprintf( keyname, "DSTYP%ld", ftss->index );
    dmi_descriptor_get_name_with_cpts( dd, fullname, CARD_SIZE );
    name = fullname;
      
    ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &name, NULL, 
			     "DM Keyword: Descriptor name." );
    ftss->written = UT_TRUE;  
  } 

  unitp = NULL;
  
  /* 
   * The first subspace component has DSVALn, the second has 2DSVALn etc 
   * If it's a range, we do  DSVALn = 'min:max'
   * If it's a table, we do  DSVALn = 'TABLE'
   *                         DSREFn = 'GTI4:'  etc 
   */
  if ( cptno == 1 )
  {
    sprintf( keyname, "DSVAL%ld", ftss->index );
    desc = valdesc;

    if (( cs_is_blank( value ) && do_table ) || cs_eq( value, "TABLE" ))
    {
      valuep = table_value;
      desc = NULL;
    }
  }
  else
  {
    sprintf( keyname, "%ldDSVAL%ld", cptno, ftss->index );

   if ( cs_is_blank( value ) && do_table )
     valuep = NULL;
  }

  unitp = dmi_descriptor_get_unit( dd );

  if (( dm_global & DM_GLOBAL_OLD ) && ( unitp && *unitp ))
    desc = " ";

  if ( valuep )
    ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &valuep, unitp, desc ); 
  
  /* Write the data type and unit */
  if ( cptno == 1 )
  {
    tvalue[0] = ftp_datatype_to_bt_tform( type );  
    tvaluep = &tvalue[0];
    sprintf( keyname, "DSFORM%ld", ftss->index );  
    ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &tvaluep, NULL, fdesc ); 
    sprintf( keyname, "DSUNIT%ld", ftss->index );  
    tvaluep = dmi_descriptor_get_unit( dd );

    if ( !cs_is_blank( tvaluep ))
      ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &tvaluep, NULL, "DM Keyword: Descriptor unit."); 
    }
  
  
  if ( tabname && *tabname )
  {
    if ( cptno == 1 )
      sprintf( keyname, "DSREF%ld", ftss->index );
    else
      sprintf( keyname, "%ldDSREF%ld", cptno, ftss->index );
      
    /* HEASARC Grouping convention: add colon. Not widely adopted, may remove someday */
    sprintf( tabval, "%s", tabname );
    tvaluep = &tabval[0];
    ftp_key_write_generic( block, 0, FTK_STRING, DT_CHAR, keyname, &tvaluep, NULL, NULL ); 
  }
}


/***************************************************************/
/*                                                             */
/* Function:   ftSubspaceTableWrite                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDescriptor *dd                                */
/*             char         *tabname                           */
/*             void         *val1                              */
/*             void         *val2                              */
/*             long          nvals                             */
/*                                                             */
/*  Output parameters:                                         */
/*                none                                         */
/*                                                             */
/*  Returns:                                                   */
/*                nothing                                      */
/*                                                             */
/***************************************************************/
void ftSubspaceTableWrite( dmDescriptor* dd, char* tabname, void* val1, void* val2, long nvals )
{
  dmBlock      *parent_block = NULL; /* Block we are writing */ 
  dmDataset    *ds        = NULL;
  dmBlock      *block     = NULL;    /* New GTI block we create here */
  dmDescriptor *new_dd    = NULL;        
  dmDescriptor *cols[2];
  ftBlock      *ftblock   = NULL;
  fitsfile     *fptr      = NULL;
  
  char  checksum[80];
  char *name = NULL;
  
  long dim = 2;
  long i   = 0;
  long no  = 0;
  int  ftstatus    = 0;
  int  do_checksum = UT_FALSE;       /* Write CHECKSUM */

  
  /*
   * Keys to copy from parent header if present.
   * This breaks the DM paradigm a bit but is included to keep
   * SDS/Archives (Arnold) happy, so that the block has a minimal CXC header 
   * if the parent block also has a CXC (or HEASARC) header.
   */
  
#define NCXC 8
  int    keystatus[NCXC];
  char  *keyname[NCXC]   = { "MISSION", "TELESCOP", "INSTRUME", "MJDREF", "TSTART",
			     "TSTOP", "TIMESYS", "TIMEUNIT" };
  char   keytype[NCXC+1] = "cccdddcc";
  double keydval[NCXC];
  long   keyival[NCXC];
  char  *keyval[NCXC];
  char  *keycmt[NCXC];

  long   nkeys = NCXC;
  
  
  if ( !dd || !tabname || !(*tabname))
    return;

  parent_block = (dmBlock*)dmi_descriptor_get_block( dd );
  ds = dmi_block_get_ds( parent_block );
  
  if ( !dmi_dataset_mode( ds ))   /* WRITE TO FILE */
    return;

  dim = dmi_descriptor_get_element_dim( dd );
  
  ftp_block_reset( parent_block );
  fptr = ftp_fptr( parent_block );
  
  /* If the parent block has checksums, the new one will too */
  if ( fits_read_key_str( fptr, "CHECKSUM", checksum, NULL, &ftstatus ) != KEY_NO_EXIST ) /* CFITSIO */
    do_checksum = UT_TRUE;
  
  for ( i = 0; i < nkeys; i++ )
  {
    ftstatus = 0;
    keycmt[i] = malloc( FLEN_CARD * UT_SZ_C );
    switch( keytype[i] )
    {
      case 'c' : 
        keyval[i]= calloc( FLEN_CARD, UT_SZ_C );
	keystatus[i] = fits_read_key_str( fptr, keyname[i], keyval[i], keycmt[i], &ftstatus );/* CFITSIO */
	break;
      case 'i' :
	keystatus[i] = fits_read_key_lng( fptr, keyname[i], &keyival[i], keycmt[i], &ftstatus );    /* CFITSIO */
	break;
      case 'd' :
      default:
	keystatus[i] = fits_read_key_dbl( fptr, keyname[i], &keydval[i], keycmt[i], &ftstatus ); /* CFITSIO */
	break;
    }
  }

  name = ftp_subspace_table_name( ds, tabname ); 

  /* Is this name an existing block or not? */
  no = dmi_block_search( ds, name );

  if ( no > 0 )
  {
    /* Handle case of multiply defined block? This is the raw block; we 
       might be in syncing trouble if it's already open. We don't make it read its own subspace,too! */
    block = ftp_block_open_mode( ds, no, dmTABLE, UT_FALSE, NULL ); 
    ftblock = (ftBlock*)dmi_block_get_kb( block );

    if ( ftblock->ncols < 2 )
    {
      sprintf( dm_errbuf, "Subspace table %s in %s has less than two columns", name, dmi_dataset_get_name( ds ) );
      dmu_err_set( DM_ERR_SUBSPACE, dm_errbuf );
      return;
    }

    ftp_block_reset( block );

    /* Convention: GTI is first two scalar cols whatever they are */
    cols[0] = ftblock->cols[0];
    cols[1] = ftblock->cols[1];  
  }
  else
  {
    /* Create  block */
    no = dmi_dataset_get_n( ds ) + 1;
    block = ftBlockCreate( ds, no, name, dmTABLE, NULL, NULL );

    if ( !block ) 
    {
      dmu_err_set( DM_ERR_INTERNAL, "Failed to make subspace table in FITS kernel" );
      return;
    }
      
    ftp_block_reset( block );

    /* If we use the table mechanism for other than TIME, we'll pass the
     * column names (equiv to START, STOP) in as dd->eltNames. Recall that
     * subspaces are dmRANGE element type so dd->edim will be 2 * dd->dim
     * when all is working. At this time dd->edim/eltNames stuff is not implemented
     * properly.
     */
    /* Create the columns */
      
    new_dd = dmi_column_create_subtable( block, dd );
      
    
    /* 
     * In the old DM we make two scalar columns; in the new we support
     * making a single vector column with two components, which is cleaner.
     * Alas, this would change behaviour in certain cases: namely, if
     * a CIAO tool tries to read the GTI extensions directly.
     * The chances that this happens and would mess things up is too great,
     * and I regretfully retain the old behaviour by default.
     */
#if 1
    if ( dm_global & DM_GLOBAL_OLD )
#endif
    {
      cols[0] = dmi_descriptor_get_cpt( new_dd, 1 );
      cols[1] = dmi_descriptor_get_cpt( new_dd, 2 );
      ftColCreate( block, cols[0], 0, NULL );
      ftColCreate( block, cols[1], 0, NULL );
    }
    else
    {
      ftColCreate( block, new_dd, 0, NULL );
      cols[0] = dmi_descriptor_get_cpt( new_dd, 1 );
      cols[1] = dmi_descriptor_get_cpt( new_dd, 2 );
    }
  }
 
  /* Copy keys */
  for ( i = 0; i < nkeys; i++ )
  {
    ftstatus = 0;

    if ( !keystatus[i] )
    {
      if ( keytype[i] == 'c' )
      {
        fits_update_key_longstr( fptr, keyname[i], keyval[i], keycmt[i], &ftstatus ); /* CFITSIO */
        free( keyval[i] );
      }
      else
        if ( keytype[i] == 'i' )
          fits_update_key_lng( fptr, keyname[i], keyival[i], keycmt[i], &ftstatus );  /* CFITSIO */
        else
          if ( keytype[i] == 'd' )
            fits_update_key_dbl( fptr, keyname[i], keydval[i], FITSK_DOUBLE_DECIMALS, keycmt[i], &ftstatus ); /* CFITSIO */
    }
    else
    {
      if ( keytype[i] == 'c' ) 
        free( keyval[i] );
    }

    free( keycmt[i] );
  }

  if ( do_checksum )  /* Will be updated at block close */
    fits_write_chksum( fptr, &ftstatus ); /* CFITSIO */
  
  /* Write the actual data */
  
  ftColWrite( cols[0], NULL, NULL, NULL, 1, nvals, val1 );
  ftColWrite( cols[1], NULL, NULL, NULL, 1, nvals, val2 );
  
  /* Update DM number of rows - if zero at block close, that would mess up NAXIS2 */
  dmi_block_set_nrows( block, nvals );
  
  ftBlockFlush( block );
  ftBlockClose( block );

  /* We close and free the block descriptors */
  dmi_block_close_descriptors( block, ftDescriptorFree );
  free( name );
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_subspace_is_gti                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*             char  *tabname                                  */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*                none                                         */
/*                                                             */
/*  Returns:                                                   */
/*                dmBool  true-or-false                        */
/*                                                             */
/***************************************************************/
dmBool ftp_subspace_is_gti( char* tabname )
{
  return ( cs_uceq( tabname, "ALLGTI" )
	   || cs_uceq( tabname, "STDGTI" )
	   || cs_uceq_siz( tabname, "GTI", 3 ));
}



/***************************************************************/
/*                                                             */
/* Function:   ftp_subspace_table_name                         */
/*                                                             */
/* Description:                                                */
/*        If the subspace table has no name, assume it is one  */
/*        of the default names ALLGTI, STDGTI, or GTI.  Look   */
/*        for it in the data file.                             */
/*                                                             */
/*  Input parameters:                                          */
/*             dmDataset *dse                                  */
/*             char      *tabname                              */
/*                                                             */
/*  Output parameters:                                         */
/*                none                                         */
/*                                                             */
/*  Returns:                                                   */
/*                char  *name                                  */
/*                                                             */
/***************************************************************/
char *ftp_subspace_table_name( dmDataset *ds, char *tabname )
{
  long  no    = 0;
  char *local = NULL;
  char *name  = NULL;
  long  n     = 0;

  name = tabname;

  if ( !cs_is_blank( name )) 
    return cs_dup( name );

  n = dmi_dataset_get_n( ds );

  for ( no = 1; no <= n; no++ )
  {
    local = dmi_dataset_get_block_name( ds, no );

    if ( ftp_subspace_is_gti( local ))
    {
      name = cs_dup( local );
      return name;
    }
  }
  return cs_dup( name );
}




/***************************************************************/
/*                                                             */
/* Function:   ftp_subspace_open                               */
/*                                                             */
/* Description:                                                */
/*            Open subspace                                    */
/*            NOte:  The create key is a hack to let us know   */
/*            if we are trying to initialize a subspace on a   */
/*            block we have just created.                      */
/*                                                             */
/*  Input parameters:                                          */
/*                dmBlock  *block                              */
/*                int       create                             */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                   nothing                                   */
/*                                                             */
/***************************************************************/
void ftp_subspace_open( dmBlock* block, int create )
{
  dmDescriptor   *dd   = NULL;
  dmDescriptor  **list = NULL;
  dmSubspace     *subspace = NULL;
  dmuKey         *ftkey = NULL;
  dmDataset      *ds    = NULL;
  dmuKeyList      keylist;
  dmDataType      type;
  dmErrCode       status = dmSUCCESS;
  
  ftBlock        *ftblock = NULL;
  ftSubspaceCol  *ftss    = NULL;

  char   typekey[80]; 
  char   typename[80];
  char   compNumString[4] = {'\0', '\0', '\0', '\0' };
  char  *dsval   = NULL;
  char  *tabname = NULL;
  char  *keyName = NULL;
  char  *ssname  = NULL;
  char  *pKeyDSVAL = NULL;

  long componentNumber  = 0;
  long keyNo  = 0;
  long n      = 0;
  long onmax  = 0;      /* Temp for reallocation */
  long nmax   = 0;       /* Actual max entry in list */
  long lnmax  = 0;      /* Space allocated in list */
  long offset = 0;
  long index  = 0;
  long maxcpt = 1;

  int have_time        = UT_FALSE; /* Is TIME an explicit DSTYP? */
  int was_table        = UT_FALSE;   /* Was first cpt a table?  */
  int is_image         = 0;
  int find_default_gti = 0;
  int  ii         = 0;
  int  numLength  = 0;
  int  is_compNum = 1;
 
 
  is_image = dmi_block_get_raw_type( block ) == dmIMAGE;
  ds       = dmi_block_get_ds( block );
  ftblock  = (ftBlock*)dmi_block_get_kb( block );
  keylist  = ftblock->keylist;

  
  // Code Change  Bug 10136  November 7, 2006

  /* Pass 0: count subspace components */
  for ( keyNo = 0; keyNo < keylist->n; keyNo++ )
  {
    ftkey   = keylist->data[keyNo];
    keyName = ftkey->name;

    // Is this a  DSVAL keyword?
    pKeyDSVAL = strstr(keyName, "DSVAL");
 
    if (pKeyDSVAL)
    {
      // Check for preceeding subspace component number
      // Keyword format is <digits>DSVAL<digits>
      numLength = pKeyDSVAL - keyName;

      if ( numLength )
      {
        strncpy(compNumString, keyName, numLength );
        compNumString[numLength] = '\0';
        is_compNum = 1;

        // Verify that this is truly a number, no alpha chars
        for (ii = 0; ii < numLength; ii++)
          if ( ! isdigit(compNumString[ii] ))
          {
            is_compNum = 0;
            break;
	  }
        
 	if (is_compNum)
	{
          // Extract component number.  If number is larger than
          // current max component number, update max value

          componentNumber = (long) atoi(compNumString);

          if ( componentNumber > maxcpt )
            maxcpt = componentNumber;
        }
      }
    }
  }
  // end code change  Bug 10136


  // Limit is 99 for max components
  if (maxcpt > 99)
    maxcpt = 99;       
  
  /*
   *  Pass 1: Make a subspace descriptor for each raw column
   * If it's an image, we should do it for the axis groups...
   */
  dmi_block_make_default_subspace( block, maxcpt );
  subspace = dmi_block_get_subspace( block );
 
 /*
   *  Pass 2: Look for DSTYP keywords and assign them
   */
  for ( keyNo = 1; keyNo <= keylist->n; keyNo++ )
    {
      offset = keyNo - 1;
      ftkey = keylist->data[offset];
      keyName = ftkey->name;
      
      /* Rewrite this to add new descriptors to subspace.
	 Preinit subspace with range values */
      /* DSTYPn defines a subspace */

      if ( cs_eq_siz( keyName, "DSTYP", 5 ) && ftkey->datatype == DT_CHAR )
	{
	  n = atol( keyName + 5 );
	  /* Store the n of DSTYPn in a kernel structure */
	  ssname = dmu_key_value_c( ftkey );
	  /* Here should search subspace only */
	  /* Watch it: SKY(X,Y) needs to be parsed */
	  sprintf( typekey, "DSFORM%ld", n );
	  ftkey = dmu_keylist_read_value_c( keylist, typekey, typename, CARD_SIZE );
	  if ( ftkey )
	    {
	      type = ftp_bt_tform_to_datatype( *typename );
	    } else {
	      type = DT_REAL8;
	    }
	  dd = dmi_subspace_col_init( block, ssname, NULL, type );
	  if ( !dd ) /* Internal error */
	    return;
	  ftss = ftp_get_ftss( block, dd, n );
	  dmi_descriptor_set_kdd( dd, ftss );
	  
#if 0 /* ryan, have_time gets set only if we successfully read the subspace */
	  /* Special case: time */
	  if ( cs_uceq( ssname, "TIME" ))
	    have_time = UT_TRUE;
#endif
	  
	  /* Extend list for new entry. Entries may not be in any particular order
	   * in the header.
	   */
	  if ( lnmax < n )
	    {
	      onmax = lnmax;
	      lnmax = 2 * n + 10;
	      list = ut_realloc( list, onmax, lnmax, UT_SZ_P );
	    } 
	  if ( nmax < n )
	    nmax = n;
	  list[n-1] = dd;
	  
	}
    }
  ftblock->max_dstyp = nmax;
  
  /* Now we have a list of all the DSTYP values ordered by DSTYP index */
  /*
   *  Pass 3: Look for other DS keywords 
   */
  
  for ( index = 1; index <= nmax; index++ )
  {
    was_table = UT_FALSE;
    dd = list[index-1];
    
    if ( !dd ){ continue; } /* skip missing-empty ones (Bug 9916)*/
    
    for ( componentNumber = 1; componentNumber <= maxcpt; componentNumber++ )
    {
      dmi_block_set_subspace_cpt( block, componentNumber );
      ftp_read_dsval_key( keylist, index, componentNumber, was_table, &dsval, &tabname );
      
      if ( componentNumber == 1 && tabname )
        was_table = UT_TRUE;

      /* NOTE: Take care of loading GTI table later, below..        */
      /*   If we have a TIME subspace but no explicit keyword.      */
      /*   We grab the default GTI or TSTART/TSTOP values.          */
      /*                                                            */
      
      status = ftp_read_subspace_values( dd, componentNumber, dsval, tabname, create );
      if ( (status == dmSUCCESS) && cs_uceq( dmi_descriptor_get_name( dd ), "TIME" ))
        have_time = UT_TRUE;

      free( tabname );
    }
  } /* end list loop */
  
  free( list );
  componentNumber = 1;
  dmi_block_set_subspace_cpt( block, componentNumber );
  
  /* Remember to look for the default GTI!                                       */
  /*  Since no DSTYP for TIME if we get here, no ftss either, so must create it .*/ 
  /*  Null primary doesn't get one; neither does file without tstart/tstop       */
  /*  11/05/2003 - Null primary DOES get one if it is the only interesting block.*/
  /*  Code Change  Bug 10044 June 10, 2010                                                 */
  /*    Do not attach default GTI on newly created subspace.                     */
  /*    Do that later using key information                                      */
 
  find_default_gti = ( !have_time && maxcpt == 1 );

  if ( is_image && find_default_gti && dmi_block_is_null_primary( block ))
  {
    ds = dmi_block_get_ds( block );
    n = ftp_dataset_find_interesting_block_no( ds, dmUNKNOWNBLOCK );

    if ( n != dmi_block_get_no( block ))
      find_default_gti = UT_FALSE;
  }
  else
    find_default_gti = ( !is_image && find_default_gti );
  
  if ( !create)
  {
    if ( find_default_gti )
    {
      double tstart  = 0.0;
      char  *tabname = NULL;

      ftkey = dmu_keylist_read_typed_value( keylist, "TSTART", DT_REAL8, &tstart );

      if ( ftp_subspace_is_gti( dmi_block_get_name( block )))
      {
        /* A GTI block has itself as a GTI */
        tabname = cs_dup( dmi_block_get_name( block ));
      }
      else
	tabname = ftp_subspace_table_name( ds, NULL );

      if ( ( ftkey != NULL ) || ( tabname != NULL ) ) 
      {
	  double tstop = 0.0;

	  dd = dmi_subspace_col_init( block, "TIME", NULL, DT_REAL8 );
	  dmi_descriptor_set_unit( dd, "s" );
	  dmi_descriptor_clear_update( dd );  /* Don't try flushing this descriptor */
	  ftss = ftp_get_ftss( block, dd, 0 );
	  dmi_descriptor_set_kdd( dd, ftss );
	  componentNumber = 1;

	  if ( tabname != NULL )
	  {
	    ftp_read_subspace_values( dd, componentNumber, NULL, tabname, create );      
	    free( tabname );
	  }
	  else if ( ftkey != NULL )
	  {
	    ftkey = dmu_keylist_read_typed_value( keylist, "TSTOP", DT_REAL8, &tstop );

            if ( ftkey != NULL )      
	      dmi_subspace_col_init_cpt( dd, NULL, 1, DT_REAL8, &tstart, &tstop, 1 );
	  }
      }
    }
  }
}



/***************************************************************/
/*                                                             */
/* Function:   ftp_read_dsval_key                              */
/*                                                             */
/* Description:                                                */
/*   Note that the dsval is a pointer that is not to           */
/*   be deallocated by the caller.                             */
/*                                                             */
/* Input parameters:                                           */
/*   dmuKeyList  *keylist     - keyword list                   */
/*   long         index                                        */
/*   long         cptno                                        */
/*   int          was_ref     - flag 'reference' found         */
/*                                                             */
/* Output parameters:                                          */
/*   char       **dsvalp      - dsval keyword value string     */
/*   char       **refblock    - associated dsref value         */
/*                                                             */
/* Returns:                                                    */
/*   nothing                                                   */
/*                                                             */
/***************************************************************/
void ftp_read_dsval_key( dmuKeyList keylist, long index, long cptno, int was_ref, 
                         char** dsvalp, char** refblock )
{
  char   *dsval = NULL;
  char   *dsref = NULL;
  dmuKey *ftkey = NULL;
  char keyname[CARD_SIZE];

  int is_reference = UT_FALSE;    /* Is current cpt a table? */
  
  *refblock = NULL;

  if ( cptno ==1  )
    sprintf( keyname, "DSVAL%ld", index );
  else 
    sprintf( keyname, "%ldDSVAL%ld", cptno, index );

  ftkey = dmu_keylist_read_typed_value( keylist, keyname, DT_CHAR, &dsval );

  /* ---------------------------------------------------------------------- */
  /* iDSVALn = "TABLE" tells us to look for a DSREF keyword.                */
  /* If DSVALn = "TABLE" and iDSVALn does not exist, we inherit DSVALn.     */
  /*                                                                        */
  /* 20131104:                                                              */
  /*   iDSVALn contains "MASK" indicates need for DSREF keyword also.       */
  /* ---------------------------------------------------------------------- */

  if ( !dsval && cptno > 1 && was_ref )
    is_reference = was_ref;
  else if ( dsval && ( cs_uceq( dsval, "TABLE" ) || (cs_index( dsval, "MASK" ) > 0)) )
    is_reference = UT_TRUE;

  if ( cptno == 1 )
    was_ref = is_reference;
  
  if ( is_reference )
  {
    if ( cptno == 1 )
      sprintf( keyname, "DSREF%ld", index );
    else
      sprintf( keyname, "%ldDSREF%ld", cptno, index );

    ftkey = dmu_keylist_read_typed_value( keylist, keyname, DT_CHAR, &dsref );   

    if( ftkey )
    {
      if ( *dsref == ':' ) 
        ++dsref;
      *refblock = cs_dup( dsref );
    }

  } /* End if reference  */
  
  *dsvalp = dsval;
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_tstart_read                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock  *block                                  */
/*            double   *tstartpx                               */
/*            double   *tstopp                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            int  true-or-false                               */
/*                                                             */
/***************************************************************/
int ftp_tstart_read( dmBlock* block, double* tstartp, double* tstopp )
{
  dmuKeyList keylist;

  ftBlock *ftblock = NULL;
  dmuKey  *ftkey   = NULL;
  
  /* If a file has a TIME column, the TSTART/TSTOP were already handled in its
   * default range. But if there's no TIME column, we need to handle it here.
   */

  ftblock  = (ftBlock*)dmi_block_get_kb( block );
  keylist  = ftblock->keylist;    
  ftkey    = dmu_keylist_read_typed_value( keylist, "TSTART", DT_REAL8, tstartp );

  if ( ftkey )
    ftkey = dmu_keylist_read_typed_value( keylist, "TSTOP", DT_REAL8, tstopp );

  if ( ftkey )
    return UT_TRUE;

  return UT_FALSE;
}



/***************************************************************/
/*                                                             */
/* Function:   ftp_read_subspace_values                        */
/*                                                             */
/* Description:                                                */
/*           Read values into subspace. I think subspace_cptno */
/*           isn't needed because we have already called set   */
/*           current component.                                */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor *dd                                 */
/*            long          subspace_cptno                     */
/*            char         *dsval                              */
/*            char         *tabname                            */
/*            int           create                             */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode  error_code                            */
/*                                                             */
/***************************************************************/
dmErrCode ftp_read_subspace_values( dmDescriptor* dd, long subspace_cptno,
                             char* dsval, char* tabname, int create )
{
  dmDataset  *ds    = NULL;
  dmBlock    *block = NULL;
  regRegion  *local_region = NULL;
  char       *name  = NULL;
  char       *new_tabname = NULL;
  void       *value1 = NULL;
  void       *value2 = NULL;
  void       *v1     = NULL;
  void       *v2     = NULL;

  dmDataType type = DT_REAL8;
 
  long   vcpt    = 0;
  long   minvcpt = 1;
  long   maxvcpt = 1;
  long   nvals   = 0;
  long   dim     = 0;
  int    status  = 0;
  int    null       = UT_FALSE;
  int    is_alloced = UT_TRUE;
  int    incnull    = UT_FALSE;

  
  if ( subspace_cptno < 0 )
    return dmFAILURE;
  
  block = dmi_descriptor_get_block( dd );
  ds    = dmi_block_get_ds( block );
  dim   = dmi_descriptor_get_element_dim( dd );
  name  = dmi_descriptor_get_name( dd );
  
  if ( tabname )
  {
    if ( dsval && cs_index(dsval, "MASK") > 0 )
    {
      /* Case of a Region string containing MASK node.        */
      /* Region library cannot handle a MASK shape,           */
      /* so need to handle separately.                        */
   
      /* Ultimately, it would be great to just pass the string */
      /* along to dmi_parse_filter_string() below, but it just */
      /* uses regParse to manage the full region string..      */
      /* unless region parsing can invoke user defined shape   */
      /* create, this will not be possible.                    */
   
      local_region = dmi_mask_region_from_string( dd, tabname, dsval );
      if ( local_region == NULL )
      {
        return status;
      }

      /* Return kernel to starting block.  The above moves the */
      /* kernel to the mask block. We need to reset it so that */
      /* things above stay in sync.                            */
      /* NOTE: This is done within ftp_subspace_read_table()   */
      /*   for GTI and REGION.                                 */
      /*                                                       */
      ftp_block_reset( block );

    }
    else
    {
      /* Case of a table: GTI or REGION */
      
      /* Check if tabname is BLANK.. if so, find default GTI block and set. */
      new_tabname = ftp_subspace_table_name( ds, tabname );
      
      /* returns errcode */
      status = ftp_subspace_read_table( dd, new_tabname, create, &value1, &value2, 
          				&nvals, &local_region ); 
      if ( status != dmSUCCESS )
      {
        free( new_tabname );
        return status;
      }
    }
  }
  else
  {
    /* 
     * Case of a DSVAL filter string  - parse it.
     * We recognize the following cases:
     *  - String contains a "(". It must be a region.
     *  - String is "NULL" denoting the null subspace.
     *  - Otherwise, string must be a simple range list.
     */
    if ( dsval )
    {
      null = dmi_parse_filter_string( dsval, dd, &value1, &value2, &nvals, &local_region, &incnull );
      if ( null ) 
	 new_tabname = cs_dup( "NULL" );

      if ( !local_region ) 
	type = dmi_descriptor_get_datatype( dd );
    }
    else
      return dmFAILURE;
  }

  if ( new_tabname ) 
    name = new_tabname;
  else
    name = tabname;
  
  if ( local_region )
  {
    dmi_subspace_col_init_region( dd, name, local_region );
  } 
  else
  {
    if ( nvals > 0 || null ) 
    {
      maxvcpt = dim;

      for ( vcpt = minvcpt; vcpt <= maxvcpt; vcpt++ )
      {
	if ( dim == 2 ) 
        {
          v1 = ar_gen_ptr( value1, type, 0, vcpt-1 );
          v2 = ar_gen_ptr( value2, type, 0, vcpt-1 );
          /* Array mode */
        }
        else
        {
          v1 = value1;
          v2 = value2;
        }
	  
        dmi_subspace_col_init_cpt( dd, name, vcpt, type, v1, v2, nvals );
      }
    }
    else
      if ( incnull )
      {
        for ( vcpt = minvcpt; vcpt <= maxvcpt; vcpt++ )
	  dmi_subspace_col_init_cpt_incnull( dd, name, vcpt, type, incnull );
      }
  }

  if ( nvals > 0 && is_alloced )
  {
    free( value1 );
    free( value2 ); 
  }

  if ( new_tabname )
    free( new_tabname );
  
  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_block_read_region                           */
/*                                                             */
/* Description:                                                */
/*           FITS specific region reader.                      */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock  *block                                  */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            regRegion  *region                               */
/*                                                             */
/***************************************************************/
regRegion* ftp_block_read_region( dmBlock* block )
{
  dmDescriptor **cols = NULL;
  regRegion     *region = NULL;
  dmDataType    *types = NULL;

  char shape_name[FLEN_CARD] = "";
  char shape_type[FLEN_CARD] = "";

  void   *ptr      = NULL;
  double  r[2]     = { 0.0, 0.0 };
  double  angle[2] = { 0.0, 0.0 };
  double *pos      = NULL;
  double *x        = NULL;
  double *y        = NULL;

  long *dims  = NULL;
  long  n     = 0;
  long  i     = 0;
  long  ncols = 0;
  long  vdim  = 2;
  long  offset = 0;
  long  maxpoints = 0; /* Max number of points for position vector */

  int   include = 1;
  int   orFlag  = 0;
  int   start   = 1; 
  short oldCpt  = 0;
  short cpt     = 0;

  
  if ( !block )
    return NULL;
  
  /* Which columns are the interesting ones? */
  ncols = dmi_region_file_cols( block, &cols, &dims );

  if ( !ncols )
    return NULL;
  
  types = (dmDataType*)calloc( ncols, UT_SZ_I );

  for ( i = 0; i < ncols; i++ )
    types[i] = dmi_descriptor_get_datatype( cols[i] );
  
  maxpoints = dims[REG_POS_COL];
  ptr = (double*)calloc( vdim * maxpoints, UT_SZ_D );
  pos = (double*)calloc( vdim * maxpoints, UT_SZ_D );

  x = &pos[0];  
  y = &pos[maxpoints];

  oldCpt = 1;
  cpt = 1;  
  vdim = 2;
  region = regCreateRegion( NULL, NULL );
  
  /* 
   *  Now we have the columns.
   */
  n = dmi_block_get_nrows( block );

  for ( i = 0; i < n; i++ )
  {
    if ( cols[REG_POS_COL] )
    {
      if ( types[REG_POS_COL] == DT_REAL8 )
        ftColRead( cols[REG_POS_COL], NULL, NULL, NULL, i+1, i+1, pos );
      else
      {
        ftColRead( cols[REG_POS_COL], NULL, NULL, NULL, i+1, i+1, ptr );
        ar_mix_numeric_cast_array( ptr, types[REG_POS_COL], DT_REAL8, pos, vdim * maxpoints );
      }
    }
    if ( cols[REG_SHAPE_COL] )
      ftColRead( cols[REG_SHAPE_COL], NULL, NULL, NULL, i+1, i+1, shape_name );
      
    if ( cols[REG_R_COL] )
    {
      if ( types[REG_R_COL] == DT_REAL8 )
        ftColRead( cols[REG_R_COL], NULL, NULL, NULL, i+1, i+1, r );
      else
      {
        ftColRead( cols[REG_R_COL], NULL, NULL, NULL, i+1, i+1, ptr );
        ar_mix_numeric_cast_array( ptr, types[REG_R_COL], DT_REAL8, r, 2 );
      }
    }

    if ( cols[REG_ROT_COL] )
    {
      if ( types[REG_ROT_COL] == DT_REAL8 )
        ftColRead( cols[REG_ROT_COL], NULL, NULL, NULL, i+1, i+1, angle );
      else
      {
        ftColRead( cols[REG_ROT_COL], NULL, NULL, NULL, i+1, i+1, ptr );
        ar_mix_numeric_cast_array( ptr, types[REG_ROT_COL], DT_REAL8, angle, 2 );
      }
    }

    if ( cols[REG_CPT_COL] )
    {
      if ( types[REG_CPT_COL] == DT_INT2 )
        ftColRead( cols[REG_CPT_COL], NULL, NULL, NULL, i+1, i+1, &cpt );
      else
      {
        ftColRead( cols[REG_CPT_COL], NULL, NULL, NULL, i+1, i+1, ptr );
        ar_mix_numeric_cast( ptr, types[REG_CPT_COL], DT_INT2, &cpt );
      }
    }

    include = (shape_name[0] != '!' );
    offset = include ? 0 : 1;
    strcpy( shape_type, shape_name+offset );

    if ( start )
      start = 0;
    else 
      orFlag = ( cpt != oldCpt );

    oldCpt = cpt;
    regAppendShape( region, shape_type, include, orFlag, x, y, maxpoints, r, angle, UT_FALSE, UT_FALSE );
  }

  free( types );
  free( pos );
  free( cols );
  free( dims );
  free( ptr );

  return region;

}



/***************************************************************/
/*                                                             */
/* Function:   ftp_subspace_read_table                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmDescriptor *dd                                 */
/*            const char   *tabname                            */
/*            dmBool        create                             */
/*            void        **val1                               */
/*            void        **val2                               */
/*            long         *nvals                              */
/*            regRegion   **ss_region                          */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode   error_code                           */
/*                                                             */
/***************************************************************/
dmErrCode ftp_subspace_read_table( dmDescriptor* dd, const char* tabname,
                                   dmBool create, void** val1, void** val2,
                                   long* nvals, regRegion** ss_region )
{
  dmDataset *ds   = NULL;
  dmBlock   *parent_block = NULL;
  dmBlock   *ss_block     = NULL;
  dmDataType dd_type;
  dmDataType type1;
  dmDataType type2;
  dmErrCode status = dmFAILURE;
 
  void* vval1 = NULL;
  void* vval2 = NULL;
  
  long dim  = 0;
  long no   = 0;
  long size = 0;
 
  parent_block = dmi_descriptor_get_block( dd );
  ds   = dmi_block_get_ds( parent_block );
  dim = dmi_descriptor_get_element_dim( dd );

  *ss_region = NULL;
  *nvals = 0;

  ftp_block_reset( parent_block );
  
  /* A block should not have itself as a subspace.
   * This is arguable, but saves some problems...
   */
  if ( create && cs_eq( tabname, dmi_block_get_name( parent_block )))
  {
    return dmSUCCESS; /* Smile */
  }

  /* Is this name an existing block or not? */
  no = dmi_block_search( ds, tabname );

  if ( no > 0 )
  {
    /* 
     * We don't use ftBlockOpen here to avoid creating the subspace and
     * getting in an infinite loop.
     */
    ss_block = ftp_block_open( ds, no, UT_FALSE, NULL );
    ftp_block_init( ss_block, UT_FALSE );
    ftp_block_reset( ss_block );

    if ( dim == 2 )
    {
      /* try reading as a region first */
      *ss_region = ftp_block_read_region( ss_block );
    }

    if ( !*ss_region )
    {
      /* try reading as a GTI block */
      *nvals = ftp_block_read_gti( ss_block, val1, &type1, val2, &type2 );

      /* make sure we return range values of the same type as the col */
      dd_type = dmi_descriptor_get_datatype( dd );

      if ( dd_type != type1 )
      {
        size = ar_mix_type_size( dd_type );
        vval1 = calloc( *nvals, size );
        ar_mix_cast_array( *val1, type1, dd_type, vval1, *nvals );
        free( ar_gen_ptr( *val1, type1, 0, 0 ));
        *val1 = vval1;
      }

      if ( dd_type != type2 )
      {
        size = ar_mix_type_size( dd_type );
        vval2 = calloc( *nvals, size );
        ar_mix_cast_array( *val2, type2, dd_type, vval2, *nvals );
        free( ar_gen_ptr( *val2, type2, 0, 0 ));
        *val2 = vval2;
      }
    }

    ftp_block_close( ss_block );

    /* We close and free the block descriptors */
    dmi_block_close_descriptors( ss_block, ftDescriptorFree );
    ftp_block_reset( parent_block );

    if ( *nvals > 0 || *ss_region )
      status = dmSUCCESS;

  }
  else
    if ( tabname && !cs_is_blank( tabname ))
    {
      sprintf( dm_errbuf, "NO SUBSPACE TABLE %s\n", tabname );
      dmu_err_set( DM_ERR_SUBSPACE, dm_errbuf );
    }
  
  return status;
}



/***************************************************************/
/*                                                             */
/* Function:   ftp_block_read_gti                              */
/*                                                             */
/* Description:                                                */
/*          Returns number of intervals in the GTI             */
/*                                                             */
/*  Input parameters:                                          */
/*            dmBlock    *block                                */
/*            void      **vale                                 */
/*            dmDataType *type1                                */
/*            void       **val2                                */
/*            dmDataType  *type2                               */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            long  nrows                                      */
/*                                                             */
/***************************************************************/
long ftp_block_read_gti( dmBlock* block, void** val1, dmDataType *type1,
                        void** val2, dmDataType* type2 )
{
  ftBlock      *ftblock = NULL;
  dmDescriptor *cols[2];
  dmDataType    ltype1; /* local type1 */
  dmDataType    ltype2; /* local type2 */
  
  long nrows = 0;
  long size  = 0;
  
  
  ftblock = (ftBlock*)dmi_block_get_kb( block );

  /* Convention: GTI is first two scalar cols whatever they are */
  if ( ftblock->ncols < 2 )
  {
    return 0;
  }

  cols[0] = ftblock->cols[0];
  cols[1] = ftblock->cols[1];
  nrows   = dmi_block_get_nrows( block );
  ltype1  = dmi_descriptor_get_datatype( cols[0] );
  ltype2  = dmi_descriptor_get_datatype( cols[1] );
  
  if ( nrows > 0 && ar_mix_is_numeric( ltype1 ) && ar_mix_is_numeric( ltype2 ))
  {
    *type1 = ltype1;
    *type2 = ltype2;
    size = ar_mix_type_size( *type1 );
    *val1 = calloc( nrows, size );
    size = ar_mix_type_size( *type2 );
    *val2 = calloc( nrows, size );
    ftColRead( cols[0], NULL, NULL, NULL, 1, nrows, *val1 );
    ftColRead( cols[1], NULL, NULL, NULL, 1, nrows, *val2 );
  }
  else
    nrows = 0;
  
  return nrows;
}
