/*                                                                
**  Copyright (C) 2002-2003,2005,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   fttablib.c                                                  */
/*                                                               */
/*   Routines to access tables                                   */
/*                                                               */
/*   Everything is a table.  Treat image as a special case.      */
/*                                                               */
/*****************************************************************/

#include "dmulib.h"
#include "dmilib.h"
#include "ftkernelP.h"

dmuKey **ftp_colkeys_list_create( dmuKeyList keylist, long numCols);

/***************************************************************/
/*                                                             */
/* Function:   ftp_table_open                                  */
/*                                                             */
/* Description:                                                */
/*      Open table. Create a key list with pointers to each    */
/*      of the possible column define keyword descriptors.     */
/*      Use this limited keylist to get column information.    */
/*      This is done to speed up processing by reducing the    */
/*      number of times the whole keylist is scanned when      */
/*      searching for a key.                                   */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock  *block                                */
/*        const dmBool    create                               */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*              dmBool table_open_status                       */
/*                                                             */
/***************************************************************/
dmBool ftp_table_open( dmBlock *block, const dmBool create )
{
  dmDataset   *ds = NULL;
  dmBlockType  type;
  dmuKeyList   keylist;
  dmuKey     **ppColKeysList = NULL;
  dmuKey      *ftkey   = NULL;
  ftBlock     *ftblock = NULL;

  int ftstatus = FITS_KERNEL_OK;
  int is_image = 0; /* For image support */

  long  nmax = 20;
  long  row_bytes = 0;
  long  ncols = 0;
  long  colno = 0;
  long  dim = 0;
  long  nrows = 0;
  long  buffer_nrows = 1;
  long  default_buffer_nrows = 100;
  long  default_buffer_bytes = 86400; /* 30 FITS blocks */

  fitsfile* fptr = NULL;

  /* FITS standard parameters */
  long pcount = 0;
  long gcount = 0;  
  int  extend = 0;
  int  simple = 0;
  int  bitpix = 0;
  int  naxis  = 0;


  ds       = dmi_block_get_ds( block );
  type     = dmi_block_get_raw_type( block );
  is_image = ( type == dmIMAGE );
  fptr     = ftp_fptr( block );
  ftblock  = (ftBlock*)dmi_block_get_kb(block);


  /* Special cases:              */
  /*   IMAGE or BINTABLE/TABLE   */
  /*   (Random groups and other weird FITS cases not supported yet)*/       
  /*    Create or Read */
  /*    (obviously we don't want to try reading header keys if it is new..) */

  if ( is_image )
  {
    nmax  = 1;
    ncols = 1; 
    nrows = 1;
    dim   = 0;

    /* Calculate row_bytes here */
    if ( !create )
    {
      fits_read_imghdr( fptr, FITS_MAXDIM, &simple, &bitpix, &naxis,   /* CFITSIO */
                        ftblock->axes, &pcount, &gcount, &extend, &ftstatus );

      if ( !ftp_ds_status( ds, ftstatus, "[ftp_table_open]" ) )
        return UT_FALSE;

      dim = (long)naxis;
    }
  }
  else
    if ( create )
    {
      /* Initial allocation of space for cols */
      nmax = 20;
      dim = ncols;
    }
    else
    {
      keylist = ftblock->keylist;  
      ftkey   = dmu_keylist_read_typed_value( keylist, "TFIELDS", DT_INT4, (void*)&ncols );

      if ( ftkey )
        ftkey = dmu_keylist_read_typed_value( keylist, "NAXIS1", DT_INT4, (void*)&row_bytes );
      if ( ftkey )
        ftkey = dmu_keylist_read_typed_value( keylist, "NAXIS2", DT_INT4, (void*)&nrows );

      if ( !ftkey )
      {
        dmu_err_set( DM_ERR_INTERNAL, "Internal error parsing FITS header" );
        return UT_FALSE;
      }

      nmax = ncols;
      dim  = ncols;
    }

  if ( row_bytes == 0 )
    row_bytes = 1;

  ftblock->nmax  = nmax;
  ftblock->ncols = ncols;
  dmi_block_set_nrows( block, nrows );
  ftblock->bytes = (double)nrows * (double)row_bytes;

  /* Set the buffer size (no of rows); note that an image has only one row */
  if ( !is_image )
  {
    default_buffer_nrows = default_buffer_bytes / row_bytes;
    fits_get_rowsize( fptr, &buffer_nrows, &ftstatus );  /* CFITSIO */

    if ( buffer_nrows < default_buffer_nrows )
      buffer_nrows = default_buffer_nrows;
  }

  ftblock->buffer_nrows = buffer_nrows;

  if ( dmverbose > 0 )
    printf( "***** Read Table Cols\n" );

  ftblock->naxes = dim;
  ftblock->cols  = (dmDescriptor**)calloc( ftblock->nmax, UT_SZ_P );

  // Code Change  Bug 12317  May 2010
  // Create keylist of important descriptor keys.
  // This requires only one pass over ftblock->keylist.

  keylist = ftblock->keylist;

  if ( !keylist )
    return UT_FALSE;

  ppColKeysList = ftp_colkeys_list_create(keylist, ncols);
  // end code change

  for ( colno = 1; colno <= ncols; colno++ ) 
    ftblock->cols[colno-1] = ftp_col_read( block, colno, ppColKeysList );

  // Free generated column keys list. It is not needed
  // after column descriptors are generated.
  if ( ppColKeysList )
    free ( ppColKeysList );

  return UT_TRUE;
}

/***************************************************************/
/*                                                             */
/* Function:   ftp_vectorize                                   */
/*                                                             */
/* Description:                                                */
/*      group the individual FITS descriptors into DM vector   */
/*      descriptors.                                           */
/*                                                             */
/*  Input parameters:                                          */
/*              dmBlock  *block                                */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*              void                                           */
/*                                                             */
/***************************************************************/
void ftp_vectorize( dmBlock *block )
{
  ftBlock         *ftblock = NULL;
  dmuAliasKeyList  mlist   = NULL;    /* List of MTYPE/MFORM aliases */
  dmDescriptor   **rawcols = NULL;

  long nrcols = 0;
  long ncols  = 0;

  ftblock = (ftBlock*)dmi_block_get_kb( block );

  if ( !ftblock )
    return;
 
  ncols = ftblock->ncols;

  /*  Go through the MTYPE keywords and figure out which are */
  /*  columns and which are coord axes                       */

  mlist = dmu_alias_key_make_list( ftblock->keylist, "MTYPE", "MFORM", "METYP", "MENAM", 5 );  

  /*  Apply the MTYPE info to the WCS information            */
  /*  Handle cols at same time (since RA/DEC forces parent to be vector) */

  rawcols   = ftp_map_vectors( block, mlist, &nrcols );

  /* The raw cols are attached to the block */
  dmi_block_set_raw_cols( block, rawcols, nrcols );
  dmu_alias_key_list_free( mlist );
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_colkeys_list_create                         */
/*                                                             */
/* Description:                                                */
/*      create a sorted columns keylist from existing key list */
/*      recording important keys needed to build columns       */
/*                                                             */
/*      Code change  Bug 12317 May 27, 2010                    */
/*      This function was added for aspect performance         */
/*      enhancement                                            */
/*                                                             */
/*      Code Change Bug 12317 update  July 29, 2010            */
/*      Modification needed to var array 'columnKeys' so that  */
/*      other DM tools run properly. Matching of keyword       */
/*      names fails or confuses other tools                    */
/*                                                             */
/*  Input parameters:                                          */
/*              dmuKeyList *keylist                            */
/*              long        numCols                            */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*              void                                           */
/*                                                             */
/***************************************************************/
dmuKey **ftp_colkeys_list_create( dmuKeyList keylist, long numCols)
{
  dmuKey **ppColKeysList = NULL;
  dmuKey  *pKey   = NULL;

  char  keyName[FITS_MAXLEN];

  long  ii = 0;
  long  keyIndex      = 0;
  long  numColumnKeys = 14;
  long  numColKeys    = 12;
  long  columnNumber  = 0;

  // '#' in strings indicate that a digit appears in that place in the string

  char *columnKeys[14] = { "TFORM#", "TDIM#", "TTYPE#", "TUNIT#", "TSCAL#", "TZERO#", "TDISP#", "TLMIN#", "TLMAX#", "TDBIN#", "TNULL#", "TDNULL#", "TSTART", "TSTOP" };

 
  // if no columns, return immediately
  if ( numCols == 0 ){ return NULL; }
 
  ppColKeysList =  (dmuKey**) calloc( (numColKeys * numCols)+2, UT_SZ_P );
  memset( ppColKeysList, 0, ((numColKeys * numCols + 2 )*UT_SZ_P));


  // loop over all keys in keylist

  for ( ii = 0; ii < keylist->n; ii++)
  {
    pKey = keylist->data[ii];

    // Get keyword without column number attached
    cs_strip_digits(pKey->ucname, keyName, FITS_MAXLEN);

    // Check for match with keywords in columnKeys array. Return index
    // number if there is a match, otherwise 0
    keyIndex = ar_match_c(keyName, columnKeys, numColumnKeys);

    if ( keyIndex )
    {
      // First twelve keyword templates in columnKeys refer to columns
      //  Strip column number from key name to use for
      //  storing key index

      if ( --keyIndex < numColKeys )
      {
        long  len = 0;
        int   isDigit = 1;

        // how long is string?
        len = strlen(pKey->ucname);

        // start at end, look for first non-digit char
        while ( isDigit && len > 0)
        {
          if ((isDigit = isdigit(pKey->ucname[len-1])) )
            --len;
        }

        // get column number
        columnNumber = atol(&pKey->ucname[len]);
         
        // put key index into key list
        //     location is (col number - 1) * 12 + column key enum + 2
        ppColKeysList[(columnNumber-1)*numColKeys+keyIndex+2] = pKey;
      }
      else
      {
        // TSTART and TSTOP do not have column numbers
        // Put those indexes into first 2 locations of index list

	if ( keyIndex == (numColumnKeys - 1))
          ppColKeysList[1] = pKey;
        else
          ppColKeysList[0] = pKey;
      }                          
    } // end if key match
  } // end for loop

  return ( ppColKeysList );
}
