/*                                                                
**  Copyright (C) 2002-2013  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   fttform.c                                                   */
/*                                                               */
/*   Functions to process TFORM keywords                         */
/*                                                               */
/*  The binary table TFORM keyword is of the form   nCm          */
/*  where n is the array size, C is the datatype, and m is extra */
/*  info.                                                        */
/*                                                               */
/*  For string cols, the string length is m (default 1)          */
/*  and the array length is n/m.                                 */
/*                                                               */
/*  For BIT columns, the length is n and the array length is the */
/*  next multiple of 8.                                          */
/*                                                               */
/*  For ASCII tables, the array size is always 1 and the string  */
/*  length is  m.                                                */
/*                                                               */
/*  For variable length arrays, the max size of the array can be */
/*  returned.                                                    */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmilib.h"
#include "ftkernelP.h"



/***************************************************************/
/*                                                             */
/* Function:   ftp_compose_bt_tform                            */
/*                                                             */
/* Description:                                                */
/*            Encode the TFORM convention which tells FITS     */
/*            about the column properties (data type, size,    */
/*            etc).                                            */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDataType  type                            */
/*                 long        isize                           */
/*                 long        length                          */
/*                 dmBool      variable                        */
/*                                                             */
/*  Output parameters:                                         */
/*                 char       *tform                           */
/*                                                             */
/*  Returns:                                                   */
/*                 nothing                                     */
/*                                                             */
/***************************************************************/
void ftp_compose_bt_tform( dmDataType type, long isize, long length,
                           dmBool variable, char* tform )
{
  char multiplier[CARD_SIZE] = "";
  char suffix[CARD_SIZE]     = "";
  char letters[] = "XBIJEDLIJA";
  char letter    = ' ';

  long size     = isize;
  long type_id  = 0;
  long ntypes   = 10;
  long types[]  = { DT_BIT, DT_UINT1, DT_INT2, DT_INT4, DT_REAL4, DT_REAL8,
                    DT_LOGICAL, DT_UINT2, DT_UINT4, DT_CHAR };


  /* Force scalars to be of size 1 */
  if ( !variable && size == 0 )
    size = 1;

  if ( type == DT_BIT )
    sprintf( multiplier, "%ld", length );
  else
    if ( variable )
      cs_copy( "1P", multiplier );
    else
      if ( type == DT_CHAR ) 
      {
        sprintf( multiplier, "%ld", length * size );

        if ( size > 1 )
          sprintf( suffix, "%ld", length );
      }
      else
        sprintf( multiplier, "%ld", size );

  type_id = ar_match_i( (long)type, types, ntypes );

  if ( type_id > 0 )
    letter = letters[type_id-1];
  else
    dmu_err_set( DM_ERR_INTERNAL, "FTK: Bad datatype in TFORM" );

  sprintf( tform, "%s%c%s", multiplier, letter, suffix );

  if( variable && size > 0 )
  {
    sprintf( suffix, "(%ld)", size );
    strcat( tform, suffix );
  }
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_compose_at_tform                            */
/*                                                             */
/* Description:                                                */
/*            Encode the TFORM convention which tells FITS     */
/*            about the column properties (data type, size,    */
/*            etc).                                            */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDataType  type                            */
/*                 long        isize                           */
/*                 long        length                          */
/*                 char       *format                          */
/*                                                             */
/*  Output parameters:                                         */
/*                 char       *tform                           */
/*                                                             */
/*  Returns:                                                   */
/*                 nothing                                     */
/*                                                             */
/***************************************************************/
void ftp_compose_at_tform( dmDataType type, long isize, long length,
                           char* format, char* tform )
{
  long size    = isize;
  long type_id = 0;
  long ntypes  = 10;
  long types[] = { DT_BIT, DT_UINT1, DT_INT2, DT_INT4, DT_REAL4,
                   DT_REAL8, DT_LOGICAL, DT_UINT2, DT_UINT4, DT_CHAR };
  char *formats[] = { "I8",   "I4",     "I6",    "I10",   "F12.3",
                      "F12.3", "L1",      "I6",    "I10",    "A" };

  if ( size == 0 )
    size = 1;

  if ( format )
  {
    cs_copy( format, tform );
  } 
  else
    if ( type == DT_CHAR )
    {
      sprintf( tform, "A%ld", length * size );
    }
    else
    {
      type_id = ar_match_i( (long)type, types, ntypes );

      if ( type_id > 0 )
        cs_copy( formats[type_id-1], tform );
      else
        dmu_err_set( DM_ERR_INTERNAL, "FTK: Bad datatype in TFORM" );
    }
}



/***************************************************************/
/*                                                             */
/* Function:   ftp_parse_tform                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              const  char   *tform                           */
/*                     dmBool  ascii                           */
/*                                                             */
/*  Output parameters:                                         */
/*                     long   *arraysize                       */
/*                     long   *length                          */
/*                     dmBool *variable                        */
/*                                                             */
/*  Returns:                                                   */
/*                 dmDataType  type                            */
/*                                                             */
/***************************************************************/
dmDataType ftp_parse_tform( const char* tform, dmBool ascii, 
                            long* arraysize, long* length, 
                            dmBool* variable )
{
  char *tptr  = (char*)tform;
  char *ptr   = NULL;
  char *final = NULL;
  char  code  = '\0';

  long  trailer    = 1;
  long  multiplier = 1;

  dmDataType type;


  if ( !tform )
    return dmUNKNOWNTYPE;

  /* skip white space */
  while( isspace( *tptr ) )
    ++tptr;

  /* Strip off leading multiplier */
  if ( isdigit( *tptr ) )
    multiplier = strtol( tptr, &ptr, 10 );
  else
    ptr = (char*)tptr;  /* Discard const char* */

  *length    = 0;
  *arraysize = multiplier; 

  if ( !ptr ) 
    return dmUNKNOWNTYPE;

  /*    Parse the character code for the data type. */
  /*    Check for variable length columns.          */

  code = *ptr;

  if ( code )
    ++ptr;

  if ( ascii )
  {
    type = ftp_at_tform_to_datatype( code );
    *variable = UT_FALSE;
  }
  else
  {
    *variable = ( code == 'P' ); /* Variable length */

    if ( *variable )
    {
      code = *ptr++;

      if ( *ptr == '(' )
        ++ptr;
    }

    /* Provisional type - must still check for unsigned case */
    type = ftp_bt_tform_to_datatype( code );
  }

  if ( ptr && isdigit( *ptr ) ) 
    trailer = strtol( ptr, &final, 10 );
  else
    trailer = 0;

  /* Set the final array size and length.       */
  /*  String and bit columns are special cases. */

  if ( *variable )
  {
    if ( final && *final == ')' )
    {
      if ( trailer == 0 )
        trailer = 1;

      *arraysize = trailer;
     }
    else
    {
      *arraysize = 0;
    }
  } 
  else
    if ( type == DT_CHAR )
    {
      *length    = trailer;

      if ( !ascii )
      {
        if ( trailer > 0 )
        {
          *length    = trailer;
          *arraysize = multiplier/trailer; 
        }
        else
        {
          *arraysize = 1;
          *length    = multiplier;
        }
      }
    } 
    else
      if ( type == DT_BIT )
      {
        *length = multiplier;
        *arraysize =  ( multiplier - 1 ) / 8 + 1;
      }

  return type;
}




/***************************************************************/
/*                                                             */
/* Function:   ftp_at_tform_to_datatype                        */
/*                                                             */
/* Description:                                                */
/*         FITS ASCII table TFORM parsingg                     */
/*                                                             */
/*            Code Change  Bug 13511  March 6, 2013            */
/*            Added 'G' as DT_REAL8 for TSV                    */
/*                                                             */
/*  Input parameters:                                          */
/*              const  char    *c                              */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                 dmDataType  type                            */
/*                                                             */
/***************************************************************/
dmDataType ftp_at_tform_to_datatype( const char c )
{
  switch( c ) 
  {
    case 'A': return DT_CHAR;
    case 'D': return DT_REAL8;
    case 'E': return DT_REAL4;
    case 'F': return DT_REAL4;
    case 'G': return DT_REAL8; 
    case 'I': return DT_INT4;
    case 'L': return DT_LOGICAL;
    default:  return DT_UNK;
  }
}



/***************************************************************/
/*                                                             */
/* Function:   ftp_bt_tform_to_datatype                        */
/*                                                             */
/* Description:                                                */
/*         Convert BINTABLE TFORM character code to datatype   */
/*         We also support the 'U' and 'V' pseudocodes, used   */
/*         in DM key comments and in CFITSIO                   */
/*                                                             */
/*            Code Change  Bug 13511  March 6, 2013            */
/*            Added 'G' as DT_REAL8 for TSV                    */
/*                                                             */
/*  Input parameters:                                          */
/*              const  char    *c                              */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                 dmDataType  type                            */
/*                                                             */
/***************************************************************/
dmDataType ftp_bt_tform_to_datatype( const char c )
{
  switch( c )
  {
    case 'A':  return DT_CHAR;
    case 'B':  return DT_UINT1;
    case 'D':  return DT_REAL8;
    case 'E':  return DT_REAL4;
    case 'G':  return DT_REAL8; 
    case 'I':  return DT_INT2;
    case 'J':  return DT_INT4;
    case 'L':  return DT_LOGICAL;
    case 'U':  return DT_UINT2;
    case 'V':  return DT_UINT4;
    case 'X':  return DT_BIT;
    default:   return DT_UNK;
  }
}



/***************************************************************/
/*                                                             */
/* Function:   ftp_datatype_to_bt_tform                        */
/*                                                             */
/* Description:                                                */
/*         Map DM data type to extended TFORM code             */
/*                                                             */
/*  Input parameters:                                          */
/*              const  dmDataType  type                        */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                 char  code                                  */
/*                                                             */
/***************************************************************/
char ftp_datatype_to_bt_tform( const dmDataType type )
{
  switch( type )
  {
    case DT_BIT:     return 'X';
    case DT_UINT1:   return 'B';
    case DT_UINT2:   return 'U';
    case DT_UINT4:   return 'V';
    case DT_INT2:    return 'I';
    case DT_INT4:    return 'J';
    case DT_REAL4:   return 'E';
    case DT_REAL8:   return 'D';
    case DT_LOGICAL: return 'L';
    case DT_CHAR:    return 'A';
    default:         return '#';
  }
}
