/*                                                                
**  Copyright (C) 2002-2005,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftvecmap.c                                                  */
/*                                                               */
/*   Functions that map vectors                                  */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmilib.h"
#include "ftkernelP.h"


/***************************************************************/
/*                                                             */
/*  Function:   ftp_map_vectors                                */
/*                                                             */
/*  Description:                                               */
/*      Brave reader: this is arguably the most arcane part of */
/*      the code                                               */
/*                                                             */
/*      Map vector columns.                                    */ 
/*      NoFtAxes is the number of raw axes for an image, the   */
/*      number of raw columns for a table.                     */
/*      We allocate a Vectors array this big to start with.    */
/*      We iterate over the columns (axes) in numerical order. */
/*      First, we check if it is a pre-declared vector column  */
/*      (with MTYPE)                                           */
/*      Next,we check if it has a coord system which is a      */
/*      predeclared MTYPE.  Finally, we check for RA,Dec pairs */
/*      and treat them as implicitly defined vectors.          */
/*      For any of these cases, we deal with the other         */
/*      components of the column.                              */
/*      If none of these cases, we treat it as a scalar column.*/
/*                                                             */  
/*      Now, we mask off all the components we've used and     */
/*      ignore them when we reach them in the iteration.       */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmBLock        *block                       */
/*                 dmuAliasKeylist mlist                       */
/*                                                             */
/*  Output parameters:                                         */
/*                  long   *nrcols - number of raw columns     */
/*                                                             */
/*  Returns:                                                   */
/*                dmDescriptor **vector                        */
/*                                                             */
/***************************************************************/
dmDescriptor** ftp_map_vectors( dmBlock* block,
                                dmuAliasKeyList mlist, /* Mtype list */
                                long* nrcols )
{
   dmDescriptor **vectors = NULL;   /* The resulting list of vectorized cols */
   dmDescriptor  *vector  = NULL;   /* An individual vector */
   dmDescriptor **cols    = NULL;   /* The list of raw scalar cols */
   dmDescriptor  *dd      = NULL;   
   dmDescriptor  *parent  = NULL;
   ftBlock       *ftblock = NULL;

   long ftaxisno = 0;
   long nvectors = 0; 
   long ncols    = 0;


   ftblock = (ftBlock*)dmi_block_get_kb( block );
   cols = ftblock->cols;
   ncols = ftblock->ncols;
   vectors = calloc( ncols, UT_SZ_P );
   nvectors = 0; 


   for ( ftaxisno = 1; ftaxisno <= ncols; ftaxisno++ )
   {
     dd = cols[ftaxisno-1];
     dd = dmi_descriptor_get_scaled( dd );
     parent = dmi_descriptor_get_parent( dd );

     if ( !parent  )   /* Not parented yet */
     {
       vector = ftp_column_match( block, cols, ncols, ftaxisno, mlist );

       if ( vector )
       {
         if ( nvectors > ncols )
           vectors = realloc( vectors, (nvectors+1) * UT_SZ_P );
         vectors[nvectors++] = vector;
         dmi_descriptor_set_no( vector, nvectors );
       }
       /* Now look for axis groups on each column (e.g. for image) */
       ftp_axis_match( block, cols, ncols, ftaxisno, mlist );
     }
   } /* End loop on rawaxis */
   *nrcols = nvectors;
   
   dmi_make_unscaled_groups( vectors, nvectors );
   return vectors;
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_column_match                               */
/*                                                             */
/*  Description:                                               */
/*      We look for the column at a given raw number and see   */
/*      if it is a vector column. First we look to see if it   */
/*      is in the MTYPE/MFORM list, then we look to see if its */
/*      coordinate transforms are (in which case it inherits   */
/*      their vector-ness).                                    */
/*      The order of the FITS columns (as in MFORM) is also    */
/*      the order we store things internally. The remapping is */
/*      done with the 'mcpt_offset=' line below.               */
/*          Example (for the complicated METYP1 case which is  */
/*                   experimental)                             */
/*          MTYPE1 = VEC                                       */
/*          MFORM1 = XLO,YLO,ZLO,XHI,YHI,ZHI                   */
/*          METYP1 = R                                         */
/*          MENAM1 = X,Y,Z                                     */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmBLock        *block                       */
/*                 dmDescriptor  **cols                        */
/*                 long            ncols                       */
/*                 long            ftaxisno                    */
/*                 dmuAliasKeyList mlist                       */
/*                                                             */
/*  Output parameters:                                         */
/*                 none                                        */
/*                                                             */
/*  Returns:                                                   */
/*                dmDescriptor *vector                         */
/*                                                             */
/***************************************************************/
dmDescriptor* ftp_column_match( dmBlock* block, dmDescriptor** cols, long ncols, 
                               long ftaxisno, dmuAliasKeyList mlist )
{
  dmDescriptor *vector = NULL;
  dmDescriptor *transform_dd = NULL;
  dmDescriptor *transform_vector = NULL;
  dmDescriptor *dd = NULL;  

  char *name  = NULL;
  char *cname = NULL;
  long  coord_no = 0;
  long nc = 0;


  /* Get logical (unscaled) column */  
  if ( ftaxisno > ncols )
    return NULL;

  dd = cols[ftaxisno-1];


  /* Get physical (scaled) column */  
  dd = dmi_descriptor_get_scaled( dd );

  name = dmi_descriptor_get_name( dd );

  /* Look in MFORMn for the name of this column, and make a vector
     with the other names in that MFORMn */

  if ( mlist )
    vector = ftp_dd_match( block, name, mlist );

  /* Now look for transform vectors */
  nc = dmi_descriptor_get_no_coords( dd );

  for ( coord_no = 1; coord_no <= nc; coord_no++ )
  {
    transform_dd = dmi_descriptor_get_coord_no( dd, coord_no );
    cname = dmi_descriptor_get_name( transform_dd );
    /* transform_dd is a singlet transform. Find the other components of its vector by name */
    transform_vector = ftp_dd_match( block, cname, mlist );

    /* Search for partner in WCS rotation */
    if( !transform_vector )
      transform_vector = ftp_wcs_pair_match( block, transform_dd );

    /* Search for implicit vector transforms (FITS WCS convention) */
    if( !transform_vector )
      transform_vector = ftp_implicit_vector_transform( block, transform_dd );

    if ( transform_vector )
    {
      /* Now inherit col vector if needed */
      if ( !vector )
        vector = ftp_dd_inherit( block, transform_vector );

      if ( vector )
      {
        /* For subsequent transforms, make sure they are parented */
        dmi_descriptor_add_coord( vector, transform_vector );

        /* Break the links to the component transform parents */
        dmi_coord_detach_cpts( transform_vector );

        /* dmi_coord_detach_cpts() shifts the remaining coords forward in the list,
         * so we need to decrement nc and coordno so that we don't miss the
         * next one */
        --nc;
        --coord_no;
      }
    }

  /* It would be an error to have different dim transforms on a column.
     We should check for this but don't */
  }

  /*
   *  No vector found
   */
  if ( !vector )
     vector = dmi_make_vector( block, name, &dd, 1 );
   return vector;
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_axis_match                                 */
/*                                                             */
/*  Description:                                               */
/*      Now we look at each column which is an array (image),  */
/*      to see if any of the axes on this image are grouped    */
/*      (vectorized).  Note that grouping axes doesn't affect  */
/*      the vector dimension of the column itself.             */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmBLock        *block                       */
/*                 dmDescriptor  **cols                        */
/*                 long            ncols                       */
/*                 long            ftaxisno                    */
/*                 dmuAliasKeyList mlist                       */
/*                                                             */
/*  Output parameters:                                         */
/*                 none                                        */
/*                                                             */
/*  Returns:                                                   */
/*              void                                           */
/*                                                             */
/***************************************************************/
void ftp_axis_match( dmBlock* block, dmDescriptor** cols, long ncols, long ftaxisno,
    dmuAliasKeyList mlist )
{
  dmDescriptor  *vector  = NULL;
  dmDescriptor **vectors = NULL;
  dmDescriptor **ddlist  = NULL;
  dmDescriptor  *transform_dd = NULL;
  dmDescriptor  *transform_vector = NULL;
  dmDescriptor  *dd     = NULL;   
  dmDescriptor  *col    = NULL;
  dmDescriptor  *parent = NULL; 

  char *name  = NULL;
  char *cname = NULL;
  long  ng = 0;
  long  coord_no = 0;
  long  nc = 0;
  long  nwc = 0;
  long  group = 0;


  /* Get physical (scaled) column */  
  if  ( ftaxisno > ncols )
    return;

  col = cols[ftaxisno-1];
  col = dmi_descriptor_get_scaled( col );

  if ( !col )
  {
    sprintf( dm_errbuf, "No physical column on axis %ld", ftaxisno );
    dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
    return;
  }
  /* What if it's a vector column? */

  /* Find axis groups */
  nc = dmi_descriptor_get_no_groups( col );
  ddlist = calloc( nc, UT_SZ_P );
  vectors = calloc( nc, UT_SZ_P );

  for ( group = 1; group <= nc; group++ )
  {
    ddlist[group-1] = dmi_descriptor_get_group_no( col, group );
  }

  /* Now we have saved the raw groups we can reorder them */
  /* Maybe attach scalar groups to raw col? */
  for ( group = 1; group <= nc; group++ )
  {
    vector = NULL;
    dd = ddlist[group-1];  
    parent = dmi_descriptor_get_parent( dd );
  
    if ( dd && !parent )
    {
       name = dmi_descriptor_get_name( dd );

       if ( mlist )
         vector = ftp_dd_match( block, name, mlist );
       /* Now look for world transform vectors */
       nwc = dmi_descriptor_get_no_coords( dd );

       for ( coord_no = 1; coord_no <= nwc; coord_no++ )
       {
         transform_dd = dmi_descriptor_get_coord_no( dd, coord_no );
         cname = dmi_descriptor_get_name( transform_dd );
         transform_vector = ftp_dd_match( block, cname, mlist );
         /* Search for partner in WCS rotation */
         if( !transform_vector )
           transform_vector = ftp_wcs_pair_match( block, transform_dd );
         /* Search for implicit vector transforms (FITS WCS convention) */
         if( !transform_vector )
           transform_vector = ftp_implicit_vector_transform( block, transform_dd );
         /* transform_dd is a singlet transform. Find the other components of its vector by name */
         if ( transform_vector )
         {
           /* Now inherit col vector if needed */
           if ( !vector )
             vector = ftp_dd_inherit( block, transform_vector );

           if ( vector )
           {
             /* For subsequent transforms, make sure they are parented */
             dmi_descriptor_add_coord( vector, transform_vector );
	
             dmi_coord_detach_cpts( transform_vector );

             /* dmi_coord_detach_cpts() shifts the remaining coords forward in the list,
              * so we need to decrement nc and coordno so that we don't miss the
              * next one */
             --nwc;
             --coord_no;
           }
         }
       } /* End loop world coords */
 
      /*  No vector found   */   
      if ( !vector )
         vector = dmi_make_vector( block, name, &dd, 1 );

      vectors[ng++] = vector;
    } /* End no parent */
  } /* End group */

  free( ddlist );
  dmi_descriptor_set_groups( col, vectors, ng );
  free( vectors );
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_wcs_pair_match                             */
/*                                                             */
/*  Description:                                               */
/*      Check if transform_dd is associated with a partner     */
/*      column/axis via a WCS rotation matrix (i.e. PC/CD      */
/*      keywords).  If so, find the partner dd and create a    */
/*      vector.                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmBLock        *block                       */
/*                 dmDescriptor   *transform_dd                */
/*                                                             */
/*  Output parameters:                                         */
/*                 none                                        */
/*                                                             */
/*  Returns:                                                   */
/*              dmDescriptor *vector                           */
/*                                                             */
/***************************************************************/
dmDescriptor* ftp_wcs_pair_match( dmBlock* block, dmDescriptor* transform_dd )
{
  ftBlock *kb = (ftBlock*)dmi_block_get_kb( block );
  ftCoord *ftcoord = NULL;
  dmDescriptor *vector = NULL;
  dmDescriptor *cpts[2] = {NULL, NULL};

  ftcoord = dmi_descriptor_get_kdd( transform_dd );

  if ( kb && ftcoord && ftcoord->partner_name )     
  {
    /* The rotation has already been applied, so we just make the vector here */

    cpts[0] = transform_dd;
    cpts[1] = dmi_descriptor_list_search( (const dmDescriptor**)kb->cols, kb->ncols, ftcoord->partner_name );

    if ( !cpts[1] )
    {
      sprintf( dm_errbuf, "Cannot find descriptor %s, which is partnered with descriptor %s in a WCS rotation", ftcoord->partner_name, dmi_descriptor_get_name( transform_dd ) );
      dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
      return NULL;
    }

    vector = dmi_make_vector( block, NULL, cpts, 2 );
  }
 return vector;
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_implicit_vector_transform                  */
/*                                                             */
/*  Description:                                               */
/*      Here we mark off WCS pairs:  RA---TAN,DEC--TAN         */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmBLock        *block                       */
/*                 dmDescriptor   *transform_dd                */
/*                                                             */
/*  Output parameters:                                         */
/*                 none                                        */
/*                                                             */
/*  Returns:                                                   */
/*              dmDescriptor *vector                           */
/*                                                             */
/***************************************************************/
dmDescriptor* ftp_implicit_vector_transform( dmBlock* block, dmDescriptor* transform_dd )
{
  ftCoord *ftcoord = NULL;
  ftBlock *ftblock = NULL;
  dmDescriptor **cols    = NULL;
  dmDescriptor  *vector  = NULL;
  dmDescriptor  *cpts[2] = { NULL, NULL}; 
  char alt_name[CARD_SIZE] = " ";
  long ncols = 0;


  ftcoord = dmi_descriptor_get_kdd( transform_dd );

  /* We recognize that e.g. GLON--SIN is part of a pair */
  if ( ftcoord->composite )     
  {
    ftp_find_wcs_pair_name( (char*)ftcoord->ctype_name, alt_name );

   if ( alt_name[0] == '\0' ) 
     return NULL;

   ftblock = (ftBlock*)dmi_block_get_kb( block );
   cols = ftblock->cols;
   ncols= ftblock->ncols;

   /* Now look for the paired column */
   cpts[0] = transform_dd;
   cpts[1] = ftp_match_ctype_name( cols, ncols, alt_name );  

   if ( !cpts[1] ) 
     return NULL;

   /* We denote the LONG/NPOL pair with a TAN-P transform type */

   if ( cs_eq( alt_name, "NPOL" ))
   {
     dmi_coord_set_transform_type( cpts[0], "TAN-P" );
     dmi_coord_set_transform_type( cpts[1], "TAN-P" );
   }
 
   /* Map TANP here */

   vector = dmi_make_vector( block, NULL, cpts, 2 );
  }
  return vector;
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_match_ctype_name                           */
/*                                                             */
/*  Description:                                               */
/*     Search for a name in a list of columns and the things   */
/*     (coords, etc) hanging off those columns                 */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmDescriptor  **cols                        */
/*                 long            ncols                       */
/*                 char           *name                        */
/*                                                             */
/*  Output parameters:                                         */
/*                 none                                        */
/*                                                             */
/*  Returns:                                                   */
/*              dmDescriptor *dd - world or transform dd       */
/*                                                             */
/***************************************************************/
dmDescriptor* ftp_match_ctype_name( dmDescriptor** cols, long ncols, char* name )
{
  dmDescriptor *dd = NULL;
  dmDescriptor *transform_dd = NULL;
  dmDescriptor *world_dd = NULL;
  ftCoord      *ftcoord  = NULL; 

  long i  = 0;
  long j  = 0;
  long k  = 0;
  long nc = 0;
  long nw = 0;


  for ( i = 0; i < ncols; i++ )
  {
    /* Unscaled descriptor */
    dd = cols[i];
    /* Scaled descriptor */
    dd = dmi_descriptor_get_coord_no( dd, 1 );

    /* Look at each of its coords */
    nc = dmi_descriptor_get_no_coords( dd );

    for ( j = 1; j <= nc; j++ )
    {
      transform_dd = dmi_descriptor_get_coord_no( dd, j );
      ftcoord = (ftCoord*)dmi_descriptor_get_kdd( transform_dd );
      if ( cs_uceq( name, ftcoord->ctype_name ))
       return transform_dd;
    } /* End loop tfm */

    /* Look at each of its axis groups */
    nc = dmi_descriptor_get_no_groups( dd );

    for ( j = 1; j <= nc; j++ )
    {
      transform_dd = dmi_descriptor_get_group_no( dd, j );
      ftcoord = (ftCoord*)dmi_descriptor_get_kdd( transform_dd );
      if ( cs_uceq( name, ftcoord->ctype_name ))
        return transform_dd;

      /* For each axis group, look at the world coords */
      nw = dmi_descriptor_get_no_coords( transform_dd );
      for ( k = 1; k <= nw; k++ )
      {
        world_dd = dmi_descriptor_get_coord_no( transform_dd, k );
        ftcoord = (ftCoord*)dmi_descriptor_get_kdd( world_dd );
        if ( cs_uceq( name, ftcoord->ctype_name ))
          return world_dd;    
      } /* End loop world */
    } /* End loop tfm */
  } /* End loop cols */
  return NULL; 
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_dd_inherit                                 */
/*                                                             */
/*  Description:                                               */
/*      For the vector transform, we make a vector column      */
/*      consisting of the parents of each component of the     */
/*      vector transform                                       */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmBlock      *block                         */
/*                 dmDescriptor *transform_vector              */
/*                                                             */
/*  Output parameters:                                         */
/*                 none                                        */
/*                                                             */
/*  Returns:                                                   */
/*              dmDescriptor *vector                           */
/*                                                             */
/***************************************************************/
dmDescriptor* ftp_dd_inherit( dmBlock* block, dmDescriptor* transform_vector )
{
  dmDescriptor  *vector = NULL;
  dmDescriptor  *transform_cpt_dd = NULL; 
  dmDescriptor **cpt_dd = NULL;
  long cptno = 0;
  long dim   = 0;


  dim = dmi_descriptor_get_element_dim( transform_vector );
  cpt_dd = calloc( dim, UT_SZ_P );

  for  ( cptno = 1; cptno <= dim; cptno++ )
  {
    transform_cpt_dd = dmi_descriptor_get_cpt( transform_vector, cptno );
    cpt_dd[cptno-1]  = dmi_coord_get_parent( transform_cpt_dd );

    if ( !cpt_dd[cptno-1] )
      printf( "INTERNAL ERROR: LOOSE COORD CPT %ld %s\n", cptno, dmi_descriptor_get_name( transform_cpt_dd ));

    /* Don't allow a descriptor to be a component of more than one vector */
    if ( dmi_descriptor_get_flag( cpt_dd[cptno-1] ) & DM_FLAG_CPT )
    {
      /* FIXME: add a warning (when we support them) about this component already
       * being in a vector */
      free( cpt_dd );
      return NULL;
    }
  }
  vector = dmi_make_vector( block,  NULL, cpt_dd, dim );
  free( cpt_dd );
  return vector;
}


/***************************************************************/
/*                                                             */
/*  Function:   ftp_dd_match                                   */
/*                                                             */
/*  Description:                                               */
/*      This routine looks up a name in the MTYPE/MFORM list   */
/*      and makes the corresponding vector descriptor.         */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmBlock        *block                       */
/*                 char           *name                        */
/*                 dmuAliasKeyList mlist                       */
/*                                                             */
/*  Output parameters:                                         */
/*                 none                                        */
/*                                                             */
/*  Returns:                                                   */
/*              dmDescriptor *vector                           */
/*                                                             */
/***************************************************************/
dmDescriptor* ftp_dd_match( dmBlock* block, char* name,  dmuAliasKeyList mlist )
{
  ftBlock *ftblock = NULL;
  ftCol   *ftcol = NULL;
  dmDescriptor   *vector = NULL;
  dmDescriptor  **cpts  = NULL;
  dmDescriptor  **cols  = NULL;
  dmElementType   etype = dmVALUE; 

  char **enames = NULL;
  char  *desc   = NULL;
  char   cptname[CARD_SIZE];
  char   vname[CARD_SIZE];
  
  long mtype_offset = 0;
  long cpt_offset   = 0;      /* Offset of component within element */
  long mcpt_offset  = 0;      /* Offset of component name in MFORM list */
  long cptno        = 0;
  long dim          = 0;
  long ncols        = 0; 
  long maxlen = CARD_SIZE;


  if ( !mlist )
    return NULL;

  ftblock = (ftBlock*)dmi_block_get_kb( block );
  cols = ftblock->cols;
  ncols = ftblock->ncols;

  /* Check if this name appears in an MFORM (or MENAM) */
  mtype_offset = ar_alias_search( mlist->cpts, name );

  if( mtype_offset >= 0 )
  {
    /* We have a vector. Let's make up the remaining entries. */
    ar_alias_get_name( mlist->names, mtype_offset, vname, maxlen );
    dim= ar_alias_get_dim( mlist->cpts,  mtype_offset );
    cpts = calloc( dim, UT_SZ_P );

    for ( cpt_offset = 0; cpt_offset < dim; cpt_offset++ )
    {
       mcpt_offset = cpt_offset;
       cptno  = cpt_offset + 1;
       ar_alias_get_cpt( mlist->names, mtype_offset, mcpt_offset, cptname, maxlen );
       cpts[cpt_offset] =  dmi_descriptor_list_search( (const dmDescriptor**)cols, ncols, cptname );
       if ( !cpts[cpt_offset] )
       {
         sprintf( dm_errbuf, "Error: token %s in MTYPE list not found in descriptor list", cptname );
         dmu_err_set( DM_ERR_GENERIC, dm_errbuf );
         return NULL;
       }

       /* Don't allow a descriptor to be a component of more than one vector */
       if ( dmi_descriptor_get_flag( cpts[cpt_offset] ) & DM_FLAG_CPT )
       {
	 /* FIXME: add a warning (when we support them) about this component already
	  * being in a vector */
	 free( cpts );
	 return NULL;
       }

       ftcol = dmi_descriptor_get_kdd( cpts[cpt_offset]);

       if ( ftcol && ftcol->type == FTK_COL )
       {
         ftcol->mtype_index = mtype_offset + 1;
       }
    } /* End for elt */
    desc = mlist->desc[mtype_offset];
    /* Compatibility with old DM */
    desc = NULL;
    vector = dmi_make_element_vector( block, vname, cpts, dim, etype, desc );

    free( cpts );

    /* Handle element types */
    dim = ar_alias_get_dim( mlist->cpts, mtype_offset );
    ar_mix_get_elt( mlist->etypes, mtype_offset, NULL, &etype );
    enames = ar_alias_get_cpt_list( mlist->cpts, mtype_offset );
    dmi_descriptor_set_element_names( vector, etype, enames, dim );
  }
  return vector;
}



