/*                                                                
**  Copyright (C) 2002-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftwcs.c                                                     */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmi.h"
#include "dmilib.h"
#include "ftkernelP.h"

void ftp_rename_default_wcs( dmBlock* block );

enum special_offsets { FWCS_EQUI, FWCS_LONP, FWCS_LATP, FWCS_RADE };

/* If an input file uses a letter other than 'P' to indicate */
/* physical coordinate transforms, we store it here          */

static char alt_wcs_phys_sub = '\0';

/* Complex argument of x+iy */
#define ARG(x,y)  atan2((y), (x))

/* Floating-point tolerance used in rotation-angle calculation */
#define FTOL  1.0e-3


/***************************************************************/
/*                                                             */
/* Function:  ftp_coord_alloc                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*   Make a kernel coord structure                             */
/*                                                             */
/* Input parameters:                                           */
/*           none                                              */
/*                                                             */
/* Output parameters:                                          */
/*           none                                              */
/*                                                             */
/* Returns:                                                    */
/*             ftCoord  *ftcoord                               */
/*                                                             */
/***************************************************************/
ftCoord* ftp_coord_alloc( void )
{
  ftCoord *ftcoord = NULL;

  ftcoord = calloc( 1, sizeof( struct ftCoord_s ));
  ftcoord->type = FTK_COORD;

  return ftcoord;
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_coord_free                                   */
/*                                                             */
/* Description:                                                */
/*         Free coordinate descriptor                          */
/*                                                             */
/* Input parameters:                                           */
/*         ftCoord  *ftcoord                                   */
/*                                                             */
/* Output parameters:                                          */
/*         none                                                */
/*                                                             */
/* Returns:                                                    */
/*                void                                         */
/*                                                             */
/***************************************************************/
void ftp_coord_free( ftCoord *ftcoord )
{
  if ( ftcoord )
  {
    free( ftcoord->wname );
    free( ftcoord->ctype_name );
    free( ftcoord->pc_row );
    free( ftcoord->partner_name );
    free( ftcoord );
  }
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_remove_badly_named_transforms                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*   This is a hack to remove any coordinate transform that    */
/*   has the same name as a real column.                       */
/*   It's called in ftp_find_transforms() after all the        */
/*   transforms have been create.                              */
/*                                                             */
/* Input parameters:                                           */
/*       ftBlock  *ftblock                                     */
/*                                                             */
/* Output parameters:                                          */
/*       none                                                  */
/*                                                             */
/* Returns:                                                    */
/*       void                                                  */
/*                                                             */
/***************************************************************/
static void ftp_remove_badly_named_transforms( ftBlock *ftblock )
{
  dmDescriptor **unscaled_cols = ftblock->cols;
  dmDescriptor  *scaled_col    = NULL;
  dmDescriptor  *cmp_col       = NULL;
  dmDescriptor  *coord         = NULL;

  long ncols   = ftblock->ncols;
  long ncoords = 0;
  long nc;
  long ii, jj;

  for ( ii = 0; ii < ncols; ii++ )
  {
    scaled_col = dmi_descriptor_get_scaled( unscaled_cols[ii] );
    ncoords    = dmi_descriptor_get_no_coords( scaled_col );

    nc = 1;

    /* Loop over the coordinates on this scaled column, removing any */
    /* that have the same name as another scaled column.             */

    while ( nc <= ncoords )
    {
      coord = dmi_descriptor_get_coord_no( scaled_col, nc );

      for ( jj = 0; jj < ncols; jj++ )
      {
	cmp_col = dmi_descriptor_get_scaled( unscaled_cols[jj] );

	if ( dmi_descriptor_compare( cmp_col, coord, dmDESCNAME ) )
	{
	  dmi_descriptor_remove_coord( scaled_col, nc );

	  /* If scaled_col has more coordinates after the one we've just */
          /* removed, then coord ncoords has just become coord nc, so we */
	  /* have to do nc again.                                        */

	  if ( nc != ncoords )
	  {
	    --nc;
	    --ncoords;
	  }

	  /* Break out of the inner column (j) loop */
	  break;
	}
      }

      /* Move on to the next coord on scaled_col */
      ++nc;
    }
  }

}


/***************************************************************/
/*                                                             */
/* Function:  ftp_find_transforms                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*   Hunt through a FITS HDU header looking for WCS keywords,  */
/*   and setting up the scalar coord transform descriptors     */
/*   on each column, as needed.                                */
/*   Called at block open                                      */
/*                                                             */
/*                                                             */
/* Input parameters:                                           */
/*       dmBlock *block                                        */
/*                                                             */
/* Output parameters:                                          */
/*       none                                                  */
/*                                                             */
/* Returns:                                                    */
/*          void                                               */
/*                                                             */
/***************************************************************/
void ftp_find_transforms( dmBlock *block )
{
  ftBlock *ftblock = 0;

  dmDescriptor *dd = NULL;
  dmBlockType   block_type;
  dmuKeyList    keylist;
  dmuKey       *ftkey  = 0;

  char **wcs_keys = NULL;
  char  *key    = NULL;   /* Key name */
  char  *name   = NULL;   /* Transform axis name */
  char  *ctype  = NULL;   /* Value of CTYPE keyword */
  char  *sBlank = " ";
  char  *pBlank = NULL;
  char  *pRadesys = NULL;
  char  *pRade  = NULL;
  char   sub    = '\0';   /* Coord subscript (blank or A-Z) */
  char   cUnderscore = '_';

  double sval[3] = { 3 * 0.0 };
  double delt    = 0.0;
  double pix     = 0.0;
  double val     = 0.0;
  double theta   = 0.0;

  long keyno    = 0;    /* Key number */   
  long offset   = 0;    /* Offset  in keylist array */
  long colno    = 0;    /* Column number */

  int  found[4] = { 0, 0, 0, 0 };  /* Have we found EQUIn etc */
  int  is_image = 0;            /* Is block an image? */

  // FITS Standard Compliance   July 2010
  //  Added *CNA to tables below to cover TCNA and iCNA keywords

  char *ImageKeys[] = { "CTYPE", "CUNIT", "CRPIX", "CRVAL", "CDELT", "CROTA", "CNAME", "--------", "CD", "PC", "LTV", "LTM", "WCSNAME" };
  char *TableKeys[] = { "*CTYP", "*CUNI", "*CRPX", "*CRVL", "*CDLT", "*CROT", "*CNAM", "*CNA", "*C", "*P", "--------", "--------", "TWCS" };

  enum offsets { FWCS_TYPE, FWCS_UNIT, FWCS_PIX, FWCS_VAL, FWCS_DELT, FWCS_ROT, FWCS_NAME, FWCS_CNA, FWCS_CD, FWCS_PC, FWCS_LTV, FWCS_LTM, FWCS_WCSNAME };


  /* Erase the alternate physical-coord subscript obtained from */
  /* any previous block opens                                   */
  alt_wcs_phys_sub = '\0';

  block_type = dmi_block_get_raw_type( block );
  ftblock    = (ftBlock*)dmi_block_get_kb( block );
  keylist    = ftblock->keylist;
  is_image   = ( block_type == dmIMAGE );
  wcs_keys   = is_image ? &ImageKeys[0] : &TableKeys[0];


  /* Pass 1: Look for WCSNAME/TWCS keywords to find any alternate */
  /*  physical transform identifier                               */

  for ( keyno = 1; keyno <= keylist->n; keyno++ )
  {
    offset = keyno - 1;
    ftkey = keylist->data[offset];
    key = ftkey->name;

    if( ftp_wcsname_parse( key, wcs_keys[FWCS_WCSNAME], &colno, &sub, is_image ) )
    {
      if ( cs_uceq( dmu_key_value_c( ftkey ), "PHYSICAL" ) )
      {
	/* Store the alternate physical transform sub.  We can only use one of
	 * these, so quit loop. */
	alt_wcs_phys_sub = sub;
	break;
      }
    }
  }


  /*  Pass 2: Look for CTYPE/TCTYP keywords and assign them to transforms */

  for ( keyno = 1; keyno <= keylist->n; keyno++ )
  {
    offset = keyno - 1;
    ftkey  = keylist->data[offset];
    key    = ftkey->name;

    /* Check for TCTYPn and iCTYPn */

    if( ftp_coord_wcskey_match( key, wcs_keys[FWCS_TYPE] ))
    {
      /* Add a new axis */
      /* Initialize it */
      ctype = dmu_key_value_c( ftkey );
      dd = ftp_create_wcs( block, key, ctype );
    }
  }


  /* Pass 3: Look for other coord keywords                         */
  /*  LTV/LTM are used only if the given axis has no standard WCS  */
  /*  keywords defining its physical transform.  (This is what all */
  /*  the calls to ftp_coord_clear_ltm() are about.)               */

  for ( keyno = 1; keyno <= keylist->n; keyno++ )
  {
    offset = keyno-1;
    ftkey  = keylist->data[offset];
    key    = ftkey->name;

    if ( cs_eq( key, "EQUINOX" ) )
    {
      dmu_key_typed_value( ftkey, DT_REAL8, &sval[FWCS_EQUI] );
      found[FWCS_EQUI] = -1; 
    }
    else if ( cs_eq( key, "RADECSYS") || cs_eq( key, "RADESYS")) 
    {
      pRadesys = dmu_key_value_c( ftkey );
      found[FWCS_RADE] = -1;
    }
    else if ( cs_eq( key, "LONPOLE") || cs_eq( key, "LONGPOLE")) 
    {
      dmu_key_typed_value( ftkey, DT_REAL8,  &sval[FWCS_LONP] );
      found[FWCS_LONP] = -1;
    }
    else if ( cs_eq( key, "LATPOLE" )) 
    {
      dmu_key_typed_value( ftkey, DT_REAL8,  &sval[FWCS_LATP] );
      found[FWCS_LATP] = -1;
    }
    else if ( ( dd = ftp_coord_find_key( block, key, wcs_keys[FWCS_DELT] )) != NULL )
    {
      ftp_coord_clear_ltm( dd );
      dmu_key_typed_value( ftkey, DT_REAL8, &delt );
      dmi_coord_set_delt( dd, &delt, 1 );
    }
    else if ( ( dd = ftp_coord_find_key( block, key, wcs_keys[FWCS_PIX] )) != NULL )
    {
      ftp_coord_clear_ltm( dd );
      dmu_key_typed_value( ftkey, DT_REAL8, &pix );
      dmi_coord_set_pix( dd, &pix, 1 );
    }
    else if ( ( dd = ftp_coord_find_key( block, key, wcs_keys[FWCS_UNIT] )) != NULL )
    {
      name = dmu_key_value_c( ftkey );
      dmi_descriptor_set_unit( dd, name );
    }
    else if ( ( dd = ftp_coord_find_key( block, key, wcs_keys[FWCS_VAL] )) != NULL )
    {
      ftp_coord_clear_ltm( dd );
      dmu_key_typed_value( ftkey, DT_REAL8, &val );
      dmi_coord_set_val( dd, &val, 1 );
    }    
    else if ( ( dd = ftp_coord_find_key( block, key, wcs_keys[FWCS_NAME] )) != NULL )
    {
      /* This DM convention overrides dd->name but not ftcoord->ctype_name */
      name = dmu_key_value_c( ftkey );

      // FITS Standard Compliance   July 2010
      // CNAME and TCNAM value strings can have spaces in them. 
      // Substitute space with underscore if present in string
      //  If string is completely blank, skip this section

      if ( !cs_is_blank(name) )
      {
        do
        {
          pBlank = strpbrk(name, sBlank);
      
          if ( pBlank )
            *pBlank = cUnderscore;

        } while (pBlank != NULL); 
      }

      dmi_descriptor_set_name( dd, name );
    }
    else if ( ( dd = ftp_coord_find_key( block, key, wcs_keys[FWCS_CNA] )) != NULL )
    {
      /* This DM convention overrides dd->name but not ftcoord->ctype_name */
      name = dmu_key_value_c( ftkey );

      // Code Change  FITS Standard Compliance  July 2010
      // TCNA and iCNA values string can have spaces in them. 
      // Substitute space with underscore if present in string
      //  If string is completely blank, skip this section

      if ( !cs_is_blank(name) )
      {
        do
        {
          pBlank = strpbrk(name, sBlank);
      
          if ( pBlank )
            *pBlank = cUnderscore;

        } while (pBlank != NULL); 
      }

      dmi_descriptor_set_name( dd, name );

    }
    else if ( ( dd = ftp_coord_find_key( block, key, wcs_keys[FWCS_ROT] )) != NULL )
    {
      ftp_coord_clear_ltm( dd );
      dmu_key_typed_value( ftkey, DT_REAL8, &theta );
      dmi_coord_set_theta( dd, theta );     
    }
    else if ( ( dd = ftp_coord_find_key( block, key, "LONP" ) ) != NULL )
    {
      dmu_key_typed_value( ftkey, DT_REAL8, &val);
      dmi_coord_set_longpole( dd, val );
    }
    else if ( ( dd = ftp_coord_find_key( block, key, "LATP" ) ) != NULL )
    {
      dmu_key_typed_value( ftkey, DT_REAL8, &val);
      dmi_coord_set_latpole( dd, val );
    }
    else if ( ( dd = ftp_coord_find_key( block, key, "EQUI" ) ) != NULL )
    {  
      dmu_key_typed_value( ftkey, DT_REAL8, &val);
      dmi_coord_set_equinox( dd, val );
    }
    else if ( ( dd = ftp_coord_find_key( block, key, "RADE" ) ) != NULL )
    {  
      pRade = dmu_key_value_c( ftkey );
      dmi_coord_set_radesys( dd, pRade );
    }
    else if ( ( dd= ftp_coord_find_key( block, key, "WCSTY" ) ) != NULL )
    {
      name = dmu_key_value_c( ftkey );
      ftp_coord_set_wname( dd, name );
    }
    else if ( ( dd = ftp_coord_find_key( block, key, wcs_keys[FWCS_LTV] )) != NULL )
    {
      if ( ftp_coord_is_ltm( dd ) )
      {
	dmu_key_typed_value( ftkey, DT_REAL8, &pix );
	dmi_coord_set_pix( dd, &pix, 1 );
      }
    }
    else if ( ftp_coord_wcskey_match( key, wcs_keys[FWCS_LTM] ) )
    {
      /* Need to set both is_cd and is_ltm parameters to true */
      ftp_coord_set_rotmat_element( block, ftkey, 1, 1 );
    }
    else if ( ftp_coord_wcskey_match( key, wcs_keys[FWCS_PC] ) )
    {
      /* check PC/CD last, since they could match a lot of other keys */
      ftp_coord_set_rotmat_element( block, ftkey, 0, 0 );
    }
    else if ( ftp_coord_wcskey_match( key, wcs_keys[FWCS_CD] ) )
    {
      ftp_coord_set_rotmat_element( block, ftkey, 1, 0 );
    }

  }  /* End of loop on keys */

  ftp_apply_coord_params( block, sval, found, pRadesys );


  /* Rename default physical axis descriptors if possible */
  ftp_rename_default_wcs( block );
  

  /* Remove any transform with the same name as a column */
  ftp_remove_badly_named_transforms( ftblock );

  /* Apply any rotations specified by PC/CD keywords */
  ftp_coords_update_rotation( block );

  ftp_coords_update_physical( block );

}


/***************************************************************/
/*                                                             */
/* Function:  ftp_apply_coord_params                           */
/*                                                             */
/* Description:                                                */
/*    Apply LONP LATP etc to ALL 2D WCS coordinate systems     */
/*                                                             */
/* Input parameters:                                           */
/*          dmBlock    *block -                                */
/*          double     *val   -                                */
/*          int        *found -                                */
/*                                                             */
/* Output parameters:                                          */
/*          none                                               */
/*                                                             */
/* Returns:                                                    */
/*          void                                               */
/*                                                             */
/***************************************************************/
void ftp_apply_coord_params( dmBlock *block, double *val, int *found,
                             char *pRadesys )
{
  ftBlock       *ftblock = NULL;
  ftCoord       *ftcoord = NULL;

  dmDescriptor  *dd   = NULL;
  dmDescriptor  *col  = NULL;
  dmDescriptor  *phys = NULL;
  dmDescriptor **cols = NULL;

  dmCoordTransform *transform = NULL;

  dmBool  bIsEQorEC = dmFALSE;
  dmBool  bIsICRS   = dmTRUE;

  char   *pRade = NULL;
  char   *pRadeDefault = "ICRS";
  char    cDash = '-';

  double  dValue = 0.0;
  double  B1950  = 1950.0;
  double  J2000  = 2000.0;

  long ii    = 0;
  long jj    = 0;
  long kk    = 0;
  long nCoords  = 0;
  long nGroups  = 0;
  long nwCoords = 0;
  long ncols    = 0;


  ftblock = (ftBlock*)dmi_block_get_kb( block );
  cols    = ftblock->cols;
  ncols   = ftblock->ncols;

  /* Apply LONP LATP etc to ALL 2D WCS coordinate systems */

  for ( ii = 1; ii <= ncols; ii++ )
  {
    col = cols[ii-1];
    col = dmi_descriptor_get_scaled( col );
    nCoords = dmi_descriptor_get_no_coords( col );

    for ( jj = 1; jj <= nCoords; jj++ )
    {
      dd = dmi_descriptor_get_coord_no( col, jj );
      ftcoord = dmi_descriptor_get_kdd( dd );

      if ( ( strlen( ftcoord->ctype_name ) >= 8) &&
	   ( ftcoord->ctype_name[4] == cDash ) )
      {
        bIsICRS   = dmTRUE; // ICRS is default
        bIsEQorEC = ftp_use_equinox( ftcoord->ctype_name );
        transform = dd->aux;
 
        /* WCS key */

        // Apply global longpole if no column longpole value
        if ( (found[FWCS_LONP] < 0) && (dValue = dmi_coord_get_longpole( transform )) == 0.0 )
	  dmi_coord_set_longpole( dd, val[FWCS_LONP] );

        // Apply global latpole if no column latpole value
        if ( (found[FWCS_LATP] < 0) && (dValue = dmi_coord_get_latpole( transform )) == 0.0 )
          dmi_coord_set_latpole( dd, val[FWCS_LATP] );

        // Apply global RADESYS if no column RADE# value
        if ( (pRade = dmi_coord_get_radesys( transform )) == NULL)
	{
          // If RADESYS found, use it, otherwise default is "ICRS"

          if( (found[FWCS_RADE] < 0) )
	  {
            dmi_coord_set_radesys( dd, pRadesys );
            pRade = pRadesys;
          }
          else
	  {
            dmi_coord_set_radesys( dd, pRadeDefault );
            pRade = pRadeDefault;
          }
        }

        // EQUINOX applies only to equitorial & ecliptic coordinate systems

        if ( bIsEQorEC )
	{
          // Check first for global EQUINOX and no column EQUI#

          if ( (dValue = dmi_coord_get_equinox( transform)) == 0.0 )
	  {
            // Before assigning EQUINOX value, consider value of RADESYS
            // EQUINOX not used if RADESYS is ICRS
            // Pointer pRade gets value from call above. If NULL, then no global
            // RADESYS value was found and default frame is ICRS
            // Flag bIsICRS is set to dmTRUE at top  of loop

            if ( pRade ) 
              bIsICRS = (cs_uceq(pRade, "ICRS")) | (cs_uceq(pRade, "GAPPT"));
          
            // Apply global equinox if no column equinox value and RADESYS != ICRS

	    if ( !bIsICRS )
	    {
              // If global EQUINOX, use it

              if ( found[FWCS_EQUI] < 0 )
	        dmi_coord_set_equinox( dd, val[FWCS_EQUI] );
              else
	      {
                // use defaults

                if ( cs_uceq(pRade, "FK4") )
		  dmi_coord_set_equinox( dd, B1950 );
                else
                  dmi_coord_set_equinox( dd, J2000 );
              }
	    }
          }
        }
        transform = NULL;
      }
    }

    nGroups = dmi_descriptor_get_no_groups( col );

    for ( jj = 1; jj <= nGroups; jj++ )
    {
      phys = dmi_descriptor_get_group_no( col, jj );
      nwCoords = dmi_descriptor_get_no_coords( phys );

      for ( kk = 1; kk <= nwCoords; kk++ )
      {
        dd = dmi_descriptor_get_coord_no( phys, kk );
        ftcoord = dmi_descriptor_get_kdd( dd );

        /* A FITS WCS CTYPE must be at least 8 chars and have a dash in byte number 5 */

        if ( ( strlen( ftcoord->ctype_name ) >= 8) &&
	     ( ftcoord->ctype_name[4] == cDash ) )
        {
          bIsICRS   = dmTRUE;
          bIsEQorEC = ftp_use_equinox( ftcoord->ctype_name );
          transform = dd->aux;
 
          /* WCS key */

          // Apply global longpole if no column longpole value
          if ( (found[FWCS_LONP] < 0) && (dValue = dmi_coord_get_longpole( transform )) == 0.0 )
	    dmi_coord_set_longpole( dd, val[FWCS_LONP] );

          // Apply global latpole if no column latpole value
          if ( (found[FWCS_LATP] < 0) && (dValue = dmi_coord_get_latpole( transform )) == 0.0 )
	    dmi_coord_set_latpole( dd, val[FWCS_LATP] );

          // Apply global RADESYS if no column RADE# value
          if ( (pRade = dmi_coord_get_radesys( transform )) == NULL)
  	  {
            // If RADESYS found, use it, otherwise default is "ICRS"

            if( (found[FWCS_RADE] < 0) )
	    {
              dmi_coord_set_radesys( dd, pRadesys );
              pRade = pRadesys;
            }
            else
	    {
              dmi_coord_set_radesys( dd, "ICRS" );
              pRade = pRadeDefault;
            }
          }

          // EQUINOX applies only to equitorial & ecliptic coordinate systems
          if (bIsEQorEC)
	  {
            // Check first for global EQUINOX and no column EQUI#

            if ( (dValue = dmi_coord_get_equinox( transform)) == 0.0 )
	    {
              // Before assigning EQUINOX value, consider value of RADESYS
              // EQUINOX not used if RADESYS is ICRS
              // Pointer pRade gets value from call above. If NULL, then no global
              // RADESYS value was found and default frame is ICRS
              // Flag bIsICRS is set to dmTRUE at top  of loop

              if ( pRade ) 
                bIsICRS = (cs_uceq(pRade, "ICRS")) | (cs_uceq(pRade, "GAPPT"));

              // Apply global equinox if no column equinox value and RADESYS != ICRS

	      if ( !bIsICRS )
	      {
                // If global EQUINOX found, use it
                if ( found[FWCS_EQUI] < 0 )
	          dmi_coord_set_equinox( dd, val[FWCS_EQUI] );
                else
	        {
                  // use defaults

                  if ( cs_uceq(pRade, "FK4") )
		    dmi_coord_set_equinox( dd, B1950 );
                  else
                    dmi_coord_set_equinox( dd, J2000 );
                }
              }
            }
          }
          transform = NULL;
        }
      }
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_coord_set_wname                              */
/*                                                             */
/* Description:                                                */
/*     Internal info to handle the WCSTY keyword               */
/*                                                             */
/*                                                             */
/* Input parameters:                                           */
/*         dmDescriptor   *dd   -                              */
/*         const char     *name -                              */
/*                                                             */
/* Output parameters:                                          */
/*          none                                               */
/*                                                             */
/* Returns:                                                    */
/*          void                                               */
/*                                                             */
/***************************************************************/
void ftp_coord_set_wname( dmDescriptor *dd, const char *wname )
{
  ftCoord *ftcoord = NULL;

  ftcoord = dmi_descriptor_get_kdd( dd );

  if ( !ftcoord ) 
    return;

  if ( ftcoord->wname ) 
    free( ftcoord->wname );

  ftcoord->wname = cs_dup( wname );
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_coords_update_physical                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*   FITS uses logical-to-world where we use physical-to-world.*/
/*   For all transforms, update their physical-to-world        */
/*   values from their logical-to-world values.                */
/*   Find the matching physical transform                      */
/*      (same axis no, but sub = 'P' )                         */
/*   and use it for the update.                                */
/*   Physical tfms themselves do not get updated.              */
/*                                                             */
/*  Fill the physical-to-world transform given the             */
/*  logical-to-world and the associated logical-to-physical.   */
/*      W = wv + wd * ( P - wp )                               */
/*      P = pv + pd * ( L - pp )                               */
/*      W = lv + ld * ( L - lp )                               */
/* Then                                                        */
/*      L = pp + ( P - pv ) / pd                               */
/* so   W = lv + ld * ( pp + (P-pv)/pd - lp )                  */
/*        = lv + ( ld / pd ) * ( P - [ pv +  (lp-pp)*pd ] )    */
/* Hence                                                       */
/*      wv = lv                                                */
/*      wd = ld / pd                                           */
/*      wp = pv + ( lp - pp ) * pd                             */
/*                                                             */
/*  EG suppose phys-to-world is                                */
/*         W =  1000 + 2.0 * ( P - 4.0 )                       */
/*  and we bin and scale to generate a logical-to-phys of      */
/*         P =  20 + 4.0 * ( L - 10 )                          */
/*  Then                                                       */
/*         W = 1000 + 2.0 * ( (20-4) - 4 * 10 ) + 2 * 4 $      */
/*                                                             */
/* Input parameters:                                           */
/*        dmBlock  *block                                      */
/*                                                             */
/* Output parameters:                                          */
/*          none                                               */
/*                                                             */
/* Returns:                                                    */
/*          void                                               */
/*                                                             */
/***************************************************************/
void ftp_coords_update_physical( dmBlock *block )
{
  ftBlock *ftblock = NULL;

  dmDescriptor  *dd   = NULL;
  dmDescriptor  *col  = NULL;
  dmDescriptor  *phys = NULL;
  dmDescriptor **cols = NULL;

  long colno   = 0;
  long axisno  = 0;
  long naxes   = 0; 
  long n       = 0;
  long nw      = 0;
  long ncols   = 0;
  long groupno = 0;
  long coordno = 0;


  ftblock  = (ftBlock*)dmi_block_get_kb( block );
  cols     = ftblock->cols;
  ncols    = ftblock->ncols;

  /* Get all physical coords */
  for ( colno = 1; colno <= ncols; colno++ )
  {
    col = cols[colno-1];
    col = dmi_descriptor_get_scaled( col );
    n = dmi_descriptor_get_no_groups( col );

    for ( groupno = 1; groupno <= n; groupno++ )
    {
      phys = dmi_descriptor_get_group_no( col, groupno );
      nw = dmi_descriptor_get_no_coords( phys );

      for ( coordno = 1; coordno <= nw; coordno++ )
      {
	dd = dmi_descriptor_get_coord_no( phys, coordno );
	dmi_coord_convert_lw_to_pw( dd, phys );     
      }
    }

    if ( !dmi_descriptor_check_groups( col ))
    {
      /* We need to add some missing axis groups. Right now we'll assume */
      /* this is because all of them are missing - might get bitten if   */
      /* some CTYPE are present but not all.                             */

      naxes = dmi_descriptor_get_array_dim( col );
 
      for ( axisno = 1; axisno <= naxes; axisno++ )
      {
	phys = ftp_create_default_wcs( block, colno, axisno, 'P' );

	/* Don't flush unit transform on block close */
	phys->flag |= DM_FLAG_TEMP;  
	dmi_descriptor_set_axis( col, axisno, phys );
      }

      /* Rename default physical axis descriptors if possible */
      ftp_rename_default_wcs( block );
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_find_wcs_pair_name                           */
/*                                                             */
/* Description:                                                */
/*    Find the matching pair for a WCS name                    */
/*    We do this hardcoded to allow any  xLON, xLAT pair       */
/*                                                             */
/*    Code Change  Bug 11488  August 13, 2010                  */
/*                                                             */
/*    This function used only the axis name for matching and   */
/*    returned an corresponding axis name. However, a change   */
/*    in function ftp_create_wcs() now stores the entire CTYPE */
/*    composite value in element ftcoord->ctype_name, not just */
/*    the axis name. Therefore, in order to find matches, the  */
/*    composite name needs to be broken into its parts, axis   */
/*    name and transform function. Matches are made using only */
/*    the axis names. For a return value, a composite name     */
/*    needs to be formed using the transform function string   */
/*    that was on the input composite string and the matching  */
/*    axis name.                                               */
/*                                                             */
/*    CTYPE composite name always contains 8 characters and    */
/*    has the form:                                            */
/*        AAAA-TTT                                             */
/*     where                                                   */
/*                                                             */
/*        AAAA is the axis name, 1 to 4 alpha chars, padded    */
/*        to length 4 with dash character                      */
/*        - character is always in the 5th character positon   */
/*        TTT is 3 character transform function string         */
/*                                                             */
/*     NB:  Calling function only sends composite type strings */
/*          to this function, so no checking is done for that  */
/*          and always assumes string length of 8 chars and in */
/*          expected format.                                   */
/*                                                             */
/* Input parameters:                                           */
/*          const char  *name -    composite CTYPE name        */
/*                                                             */
/* Output parameters:                                          */
/*          const char  *alt_name -  matching composite CTYPE  */
/*                                   name                      */
/* Returns:                                                    */
/*          void                                               */
/*                                                             */
/***************************************************************/
void ftp_find_wcs_pair_name( const char *inName, char *alt_name )
{
  char axisName[5]  = { 5 * '\0' };
  char transFunc[5] = { 5 * '\0' };
  char namePair[10] = { 10 * '\0' };
  char cDash         = '-';
  char cNull         = '\0';

  int  KEY_NAME_LENGTH = 4;
  int  ii              = KEY_NAME_LENGTH - 1;


  // initialize alt_name to null string
  strcpy( alt_name, "" );

  // Break composite name into axis name and tranform function
  //  transform function string has dash as first char

  strncpy(transFunc, inName+KEY_NAME_LENGTH, KEY_NAME_LENGTH);
  strncpy(axisName, inName, KEY_NAME_LENGTH);

  // Remove any training dashes from axis name
  while ( axisName[ii] == cDash && ii >= 0 )
    axisName[ii--] = cNull;
 
  if ( axisName[0] == cNull ) 
  {
    // bad name
    return;
  }

  // Match name pairs

  if( !strcmp( axisName, "RA" ) )
    strcpy( alt_name, "DEC" );
  else if ( !strcmp( axisName, "DEC" ) )
    strcpy( alt_name, "RA" );
  else if ( !strcmp( axisName+1, "LON" ) )
    sprintf( alt_name, "%cLAT", axisName[0] );
  else if ( !strcmp( axisName+1, "LAT" ) )
    sprintf( alt_name, "%cLON", axisName[0] );
  else if ( !strcmp( axisName, "NPOL" ) )
    strcpy( alt_name, "LONG" );
  else if ( !strcmp( axisName, "LONG" ) )
    strcpy( alt_name, "NPOL" );

    //  Code change FITS Standard Complience  July 2010
    //    Add pairings abLN, abLT
  else if ( !strcmp( axisName+2, "LN" ) )
    sprintf( alt_name, "%c%cLT",axisName[0],axisName[1] );
  else if ( !strcmp( axisName+2, "LT" ) )
    sprintf( alt_name, "%c%cLN",axisName[0],axisName[1] );
  else
    // no match
    return;

  // Form CTYPE composite name for return value

  // Get axis name first, then its length
  strcpy(namePair, alt_name);
  ii = strlen(namePair);

  // fill in with dash character to length = 5
  while (ii < KEY_NAME_LENGTH )
    namePair[ii++] = cDash;

  // add null terminator for good measure
  namePair[ii] = cNull;

  // add on transform function string to end
  strcat(namePair,transFunc);

  // copy composite name into alt_name for return
  strcpy(alt_name, namePair);

  return;
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_create_wcs                                   */
/*                                                             */
/* Description:                                                */
/*   Create a WCS descriptor given a WCS keyword               */
/*                                                             */
/* Input parameters:                                           */
/*         dmBlock     *block                                  */
/*         const char  *key                                    */
/*         const cahr  *ctype                                  */
/*                                                             */
/* Output parameters:                                          */
/*         none                                                */
/*                                                             */
/* Returns:                                                    */
/*         dmDescriptor *dd                                    */
/*                                                             */
/***************************************************************/
dmDescriptor* ftp_create_wcs( dmBlock *block, const char *key, const char *ctype )
{
  ftCoord  *ftcoord = NULL;
  ftBlock  *ftblock = NULL;

  dmDescriptor      *dd    = NULL;
  dmDescriptor      *col   = NULL;
  dmCoordTransform  *coord = NULL;
  dmDataType         dtype = DT_REAL8;
  dmElementType      etype = dmVALUE;
  dmBool             composite = dmFALSE;
 
  char  *name       = NULL;
  char  *unit       = NULL;
  char  *desc       = NULL;
  char  *ctype_name = NULL;
  char  *transform  = NULL;
  char  *tmpstr     = NULL;
  char  *sUnderscore = "_";
  char   end[5]     = { 5 * '\0' };
  char   sub        = '\0';
  char   cNull      = '\0';

  double delt = 1.0;
  double pix  = 0.0;
  double val  = 0.0;
 
  long *axes = NULL;
  long  colno   = 0;
  long  no      = 0;
  long  dim     = 1;
  long  vdim    = 1;
  long  naxes   = 1;
  long  axisno  = 0;
  long  axisno2 = 0;
  
  int  new      = 0;
  int  is_image = 0;
  int  mm       = 0;
  int  nn       = 0;


  is_image  = ( dmi_block_get_raw_type( block ) == dmIMAGE ); 
  colno     = ftp_parse_wcs_key( key, &axisno, &axisno2, &sub, is_image );
  composite = ftp_ctype_parse( ctype, &ctype_name, &transform );

  if ( colno > 0 ) 
    no = colno;
  else
    no = axisno;

  name = cs_dup( ctype_name );
  
  if ( sub == alt_wcs_phys_sub )
    sub = 'P';
  
  /*
   *  For physical, Check if it already exists by looking through
   *  the list of columns and axes 
   */
  if ( sub == 'P' )
    dd = ftp_coord_match_col_axis( block, colno, axisno, sub );

  if ( dd )
  {
    new = 0;
    ftcoord = dmi_descriptor_get_kdd( dd );
    free( ftcoord->ctype_name );
    dmi_descriptor_set_name( dd, name );
  }
  else 
  {
    new = 1;

    ftblock = (ftBlock*)dmi_block_get_kb(block);

    /* Bug_10319: Set WCS descriptor size based on source descriptor */
    if ( !is_image )
    {
      col = dmi_descriptor_get_scaled( ftblock->cols[colno-1] );
      naxes = dmi_descriptor_get_array_dimensions( col, &axes );
    }

    /* check dd group and coord tree for an existing dd with the same name */
    if ( (dd = dmi_descriptor_search(ftblock->cols[colno-1], name)) != NULL)
    {
      /* get CTYPE number and append it to name */
      mm = 0;

      for (nn=5; nn<strlen(key);nn++)
	end[mm++] = key[nn];	  
      end[mm] = cNull;
	  
      tmpstr = cs_dup_cat(name, end, sUnderscore);
	
      free(name);
      free(ctype_name);

      ctype_name = tmpstr;
      name       = cs_dup(ctype_name);

    }

    ftcoord = ftp_coord_alloc();
    dd = dmi_descriptor_alloc( block, dmCOORD, no, name, dtype, unit, 
			       desc, vdim, etype, naxes, axes, DM_FLAG_COORD );  
    coord = dmi_coord_transform_alloc( dd, transform, &val, &pix, &delt, dim );
    dmi_descriptor_set_kdd( dd, ftcoord );
  }

  ftcoord->axis       = axisno;
  ftcoord->no         = colno;
  ftcoord->sub        = sub;
  ftcoord->composite  = composite;
  ftcoord->pc_row     = NULL;
  ftcoord->is_cd      = 0;
  ftcoord->is_ltm     = 1;
  ftcoord->partner_name = NULL;

  // Code Change  Bug 11488 August 3, 2010
  // Use value of CTYPE keyword for name
  ftcoord->ctype_name = cs_dup((char*)ctype);
  
  if ( new )
    ftp_attach_wcs( block, dd );

  free( name );
  free( transform );
  
  return dd;
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_ctype_parse                                  */
/*                                                             */
/* Description:                                                */
/*   Transforms are always linear except if they are of the    */
/*   form   NNNN-TTT, where NNNN is the axis name and TTT is   */
/*   the transform name.  The NNNN may have trailing dashes    */
/*   which we strip. Example: RA---TAN                         */
/*   Returns true if there is a composite '-' 2D projection    */
/*   transform                                                 */
/*   (later we will support non-linear transforms that are 1D) */
/*                                                             */
/*                                                             */
/* Input parameters:                                           */
/*          const char *ctype_in -                             */
/*                                                             */
/* Output parameters:                                          */
/*          char  **nameptr  -                                 */
/*          char **transform -                                 */
/*                                                             */
/* Returns:                                                    */
/*          dmBool  composite  - UT_TRUE or UT_FALSE           */
/*                                                             */
/***************************************************************/
dmBool ftp_ctype_parse( const char* ctype_in, char** nameptr, char** transform )
{
  dmBool composite  = UT_FALSE;
  
  const char *ctype = ctype_in;
  char       *name  = NULL;
  char        cDash = '-';
  char        cNull = '\0';

  long  clen = 0; 


  clen = strlen( ctype );
      
  /* Special case: TAN-P */
  if ( cs_eq( ctype_in, "NPOL-TAN" ))
  {
    *nameptr = cs_dup( "NPOL" );
    *transform = cs_dup( "TAN-P" );
    composite = UT_TRUE;
  }
  else if( clen >= 8 && ctype[4] == '-'  )
  {
    /* Have a NNNN-TTT type keyword */
    *transform = cs_dup( &ctype[5] );

    name = (char*)calloc( 5, UT_SZ_C );
    *nameptr = name;

    while ( *ctype != cDash ) 
    {
      *name = *ctype++;
      ++name;
    }
    *name = cNull;

    composite = UT_TRUE;      

  }
  else 
  {
    *nameptr = cs_dup( ctype );
    *transform = cs_dup( "LINEAR" );
  } 

  return composite;
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_wcsname_parse                                */
/*                                                             */
/* Description:                                                */
/*   Match and parse a keyword of the form 'WCSNAMEa' or       */
/*   'TWCSna', where 'a' is blank or A-Z and 'n' is a table    */
/*   column number in the range 1-999.  Return value is true   */
/*   if the keyword is matched and successfully parsed, false  */
/*   otherwise.  The values of 'a' and 'n' are returned in     */
/*   sub and colno, respectively.                              */
/*                                                             */
/*                                                             */
/* Input parameters:                                           */
/*          const char  *key                                   */
/*          const char  *keyPrefix                             */
/*          int          is_image                              */
/*                                                             */
/* Output parameters:                                          */
/*          long        *colno                                 */
/*          char        *sub                                   */
/*                                                             */
/* Returns:                                                    */
/*         dmBool - UT_TRUE or UT_FALSE                        */
/*                                                             */
/***************************************************************/
dmBool ftp_wcsname_parse( const char *key, const char *keyPrefix,
                          long *colno, char *sub, int is_image )
{
  char *tail = NULL;

  long  key_size    = 0;
  long  prefix_size = 0;

  prefix_size = strlen( keyPrefix );

  if ( !cs_eq_siz( key, keyPrefix, prefix_size ) )
    return UT_FALSE;

  key_size = strlen( key );
  *colno   = 0;
  *sub     = ' ';

  if ( is_image )
  {
    if ( key_size == prefix_size + 1 )
      *sub = key[key_size-1];
  }
  else
  {
    if ( key_size <= prefix_size )
      return UT_FALSE;

    *colno = strtol( &(key[prefix_size]), &tail, 10 );

    if ( tail && *tail )
      *sub = *tail;
  }

  return UT_TRUE;
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_parse_wcs_key                                */
/*                                                             */
/* Description:                                                */
/*   Parse a keyword of the form 'xxxxnnnc' or 'xxxxnn_mmc'    */
/*   Extract                                                   */
/*     + column number                                         */
/*     + the numbers nn and mm                                 */
/*     + the one-byte character suffix                         */
/*                                                             */
/*   Example:    CTYPE4B returns 1, 4, 0, B                    */
/*               CD4_5   returns 1, 4, 5, ' '                  */
/*                                                             */
/* Input parameters:                                           */
/*        const cahr  *name                                    */
/*        int          is_image                                */
/*                                                             */
/* Output parameters:                                          */
/*        long        *axisno                                  */
/*        long        *axisno2                                 */
/*        char        *sub                                     */
/*                                                             */
/* Returns:                                                    */
/*         long colno = column number                          */
/*                                                             */
/***************************************************************/
long ftp_parse_wcs_key( const char *name, long *axisno, long *axisno2,
                        char *sub, int is_image )
{
  char *ptr    = NULL;
  char *aptr   = NULL;

  long suffix_mod = 0;
  long prefix     = 0;
  long prefix2    = 0;
  long suffix     = 0;
  long suffix2    = 0;
  long colno      = 0;

  int is_ltm_ltv = 0;


  /* LTM/LTV always define a physical transform, so we need to set sub
   * to 'P' for them */

  if ( cs_eq_siz( name, "LTM", 3 ) || cs_eq_siz( name, "LTV", 3 ) )
    is_ltm_ltv = 1;

  *axisno2 = 0;

  /* Skip letters */     
  if ( isdigit( *name ))
  {
    prefix = *name - '0';
    ++name;
    
    if ( isdigit( *name ))
    {
      prefix2 = *name - '0';
      ++name;
    }
  } 
  else if ( *name == 'C' ) 
  {
    is_image = 1;
  }

  while( *name && !isdigit( *name ) )
    ++name;


  /* Read digit and trailing letter */

  if ( name && *name )
    suffix = strtol( name, &ptr, 10 );
  else 
    suffix = 0;

  if ( ptr && *ptr == '_' ) /*  CDi_j type keyword */
  {
    ++ptr;
    suffix2 = strtol( ptr, &aptr, 10 );
    ptr = aptr;
  }

  /* PCiiijjj */
  suffix_mod = is_image ? 100 : 1000;

  if ( suffix > suffix_mod )
  {
    suffix2 = suffix % suffix_mod;
    suffix  = suffix / suffix_mod; 
  }

  if ( is_ltm_ltv ) 
  {
    *sub = 'P';
  }
  else if ( ptr && *ptr ) 
  { 
    if ( *ptr == alt_wcs_phys_sub )
      *sub = 'P';
    else
      *sub = *ptr; 
  } 
  else 
  {
    *sub = ' ';
  }

  if ( is_image )
  {
    *axisno  = suffix;
    *axisno2 = suffix2;
    colno    = 1;
  }
  else 
  {
    *axisno = prefix;

    /* Interpretation of this argument varies with context */

    if ( prefix2 )
      *axisno2 = prefix2;
    else
      *axisno2 = suffix2;

    colno = suffix;
  }

   return colno;
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_create_default_wcs                           */
/*                                                             */
/* Description:                                                */
/*  Make a trivial WCS transform for given column,             */
/*  axis and subscript                                         */
/*                                                             */
/* Input parameters:                                           */
/*          dmBlock    *block                                  */
/*          const long  colno                                  */
/*          const long  axisno                                 */
/*          const char  sub                                    */
/*                                                             */
/* Output parameters:                                          */
/*          none                                               */
/*                                                             */
/* Returns:                                                    */
/*         dmDescriptor *dd                                    */
/*                                                             */
/***************************************************************/
dmDescriptor* ftp_create_default_wcs( dmBlock *block, const long colno,
                                      const long axisno, const char sub )
{
  ftBlock       *ftblock = (ftBlock*)dmi_block_get_kb( block );
  dmDescriptor  *dd      = NULL;
  dmBlockType    type;

  char nkey[10];
  char ctype[80];

  /* No CTYPE or TCTYP key present, so we invent one */
  /*   and add a new axis to the list                */

  /* Handle case of spuriously copied keywords eg TCROT5 where TFIELDS=3 */
  if ( colno > ftblock->ncols )
    return NULL;

  type = dmi_block_get_raw_type( block );

  if( type == dmTABLE )
  {
    if ( axisno > 0 )
    {
      sprintf( nkey, "%ldCTYP%ld%c", axisno, colno, sub );
      sprintf( ctype, "#COL%ld_AXIS%ld%c", colno, axisno, sub );
    }
    else 
    {
      sprintf( nkey, "TCTYP%ld%c", colno, sub );

      if ( dm_global & DM_GLOBAL_OLD )
      {
	char *name = NULL;

	dd = ftblock->cols[colno-1];
	dd = dmi_descriptor_get_scaled( dd );
	name = dmi_descriptor_get_name( dd );
	sprintf( ctype, "%s_COORD%c", name, sub );
      }
      else 
      {
	sprintf( ctype, "COORD%ld%c", colno, sub );
      }
    }
    dd = ftp_create_wcs( block, nkey, ctype );
  }
  else 
  {
    sprintf( nkey, "CTYPE%ld%c", axisno, sub );
    sprintf( ctype, "#AXIS%ld", axisno );

    dd = ftp_create_wcs( block, nkey, ctype );
  }

  return dd;
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_coord_match_col_axis                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*   Hunt for an existing coord descriptor with the            */
/*   given column, axis and subscript                          */
/*                                                             */
/*   If we find a match, we unset the DM_FLAG_TEMP flag bit    */
/*   in case the descriptor was automatically created          */
/*   by the DM.  (If we're looking for it with this function,  */
/*   then there's a keyword corresponding to it, and we want   */
/*   it to be flushed on block close.)                         */
/*                                                             */
/*                                                             */
/* Input parameters:                                           */
/*          const dmBlock   *block                             */
/*          const long       colno                             */
/*          const long       axisno                            */
/*          const char       sub                               */
/*                                                             */
/* Output parameters:                                          */
/*          none                                               */
/*                                                             */
/* Returns:                                                    */
/*          dmDescriptor coord_dd - coordinate descriptor      */
/*                                                             */
/***************************************************************/
dmDescriptor* ftp_coord_match_col_axis( const dmBlock *block,
              const long colno, const long axisno, const char sub )
{
  ftBlock* ftblock = NULL;
  ftCoord* ftcoord = NULL;

  dmDescriptor* dd  = NULL;
  dmDescriptor* col = NULL;
  dmDescriptor* coord_dd = NULL;

  long ncoord;
  long ngroup;
  long dim;

  long ii;
  long kk;
  
  ftblock = (ftBlock*)dmi_block_get_kb( block );
  
  /* Get unscaled and scaled column descriptors */
  if ( colno > 0 && colno <= ftblock->ncols )
    col = ftblock->cols[colno-1];

  if ( col )
    col = dmi_descriptor_get_scaled( col );

  if ( !col )
    return NULL;


  if ( axisno == 0 )
  {
    /* Find coord with matching number n and letter subscript c  (CTYPEnc)  */

    ncoord = dmi_descriptor_get_no_coords( col );

    for ( ii = 1; ii <= ncoord; ii++ )
    {
      dd = dmi_descriptor_get_coord_no( col, ii );
      ftcoord = dmi_descriptor_get_kdd( dd );
      if ( ftcoord && 
	   ( ftcoord->no == colno ) && 
	   ( ftcoord->sub == sub  ) &&
	   ( ftcoord->axis == axisno ) )
      {
	dd->flag &= ~DM_FLAG_TEMP;
	return dd; 
      }
    }
  }
  else 
  {
    /* For each axis group, check for matching axis number, 
       column number and subscript */

    ngroup = dmi_descriptor_get_no_groups( col );
    for ( ii = 1; ii <= ngroup; ii++ )
    {
      dd = dmi_descriptor_get_group_no( col, ii );
      if ( dd )
      {
	dim = dmi_descriptor_get_element_dim( dd );
	ftcoord = dmi_descriptor_get_kdd( dd );

	if ( ftcoord && 
	     ( ftcoord->no == colno ) && 
	     ( ftcoord->sub == sub  ) &&
	     ( ftcoord->axis == axisno ) )
	{
	  dd->flag &= ~DM_FLAG_TEMP;
	  return dd;
	}

	/* Also look on each world coord belonging to the axis group */
	ncoord = dmi_descriptor_get_no_coords( dd );

	for ( kk = 1; kk <= ncoord; kk++ )
	{
	  coord_dd = dmi_descriptor_get_coord_no( dd, kk );
	  ftcoord = dmi_descriptor_get_kdd( coord_dd );
	  dim = dmi_descriptor_get_element_dim( coord_dd );
	  
	  if ( ftcoord && 
	       ( ftcoord->no == colno ) && 
	       ( ftcoord->sub == sub  ) && 
	       ( ftcoord->axis == axisno ) )
	  {
	    coord_dd->flag &= ~DM_FLAG_TEMP;
	    return coord_dd;
	  }
	}
      }
    }
  }
  return NULL;
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_attach_wcs                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  This routine sets up the pointers between a coordinate     */
/*  descriptor and the descriptor it transforms.               */
/*  The arrows go both ways                                    */
/*  ("is a coord of" and "has a coord").                       */
/*                                                             */
/* Input parameters:                                           */
/*          dmBlock       *block                               */
/*          dmDescriptor  *dd                                  */
/*                                                             */
/* Output parameters:                                          */
/*          none                                               */
/*                                                             */
/* Returns:                                                    */
/*          void                                               */
/*                                                             */
/***************************************************************/
void ftp_attach_wcs( dmBlock *block, dmDescriptor *dd )
{
  ftCoord *ftcoord = NULL;
  ftBlock *ftblock = NULL;

  dmDescriptor  *col     = NULL;
  dmDescriptor  *raw_col = NULL;
  dmDescriptor  *phys    = NULL;

  long  colno   = 0;
  long  axisno  = 0;
  int   is_phys = 0;

  ftblock  = (ftBlock*)dmi_block_get_kb( block );
  ftcoord = dmi_descriptor_get_kdd( dd );

  /* World coord */
  colno = ftcoord->no;

  if ( colno == 0 )
    return;

  /* Get Unscaled and Scaled column descriptors */
  raw_col = ftblock->cols[colno-1];     
  col = dmi_descriptor_get_scaled( raw_col );

  /* Is this an axis group? */
  axisno  = ftcoord->axis;
  is_phys = ( ftcoord->sub == 'P' || ftcoord->sub == alt_wcs_phys_sub );

  /* Three cases:
   *   Coord system on the descriptor  (axisno=0)
   *   Physical axis group on the descriptor 
   *   World transform on one of the descriptor's physical axis groups 
   */

  if ( axisno == 0 )
    dmi_descriptor_add_coord( col, dd );
  else if ( is_phys )
    dmi_descriptor_set_axis( col, axisno, dd );
  else 
  {
    phys = dmi_descriptor_get_group_no( col, axisno );

    /* Create phys if needed. This is recursive, calling ftp_attach_wcs */
 
    if ( !phys )
    {
      phys = ftp_create_default_wcs( block, colno, axisno, 'P' );
      phys->flag |= DM_FLAG_TEMP;  /* Don't flush unit transform on block close */

      /* Special cases */

      if( cs_eq( dmi_descriptor_get_name( dd ), "RA" ))
	dmi_descriptor_set_name( phys, "X" );
      else if ( cs_eq( dmi_descriptor_get_name( dd ), "DEC" ))
	dmi_descriptor_set_name( phys, "Y" );

    }
    dmi_descriptor_add_coord( phys, dd );
  }
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_coord_find_key                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Parse a WCS key and return the descriptor for the          */
/*  corresponding transform.  Creating it if needed.           */
/*                                                             */
/* Input parameters:                                           */
/*      dmBlock      *block                                    */
/*      const char   *key                                      */
/*      const char   *keyPrefix                                */
/*                                                             */
/* Output parameters:                                          */
/*                                                             */
/*                                                             */
/* Returns:                                                    */
/*           dmDescriptor *dd                                  */
/*                                                             */
/***************************************************************/
dmDescriptor* ftp_coord_find_key( dmBlock *block, const char *key,
                                  const char *keyPrefix )
{
  dmDescriptor *dd = NULL;
  ftBlock      *ftblock = (ftBlock*)dmi_block_get_kb( block );

  char sub = ' ';

  long naxes   = 0;
  long axisno  = 0;
  long axisno2 = 0;
  long diag    = 0;
  long colno   = 0;
  int is_image = 0;


  if ( ftp_coord_wcskey_match( key, keyPrefix ))
  {
    is_image = dmi_block_get_raw_type( block ) == dmIMAGE;
    colno    = ftp_parse_wcs_key( key, &axisno, &axisno2, &sub, is_image );

    if ( colno == 0 && axisno == 0 )
      return dd;

    naxes = ftblock->naxes;

    /* Protect against bad things like LONP9 keyword making it into image header... */
    if ( is_image && axisno > naxes )
      return NULL;

    if ( sub == alt_wcs_phys_sub )
      sub = 'P';

    /* Do we already have a WCS with this axisno and sub? */
    diag = is_image ? axisno : colno;

    if ( axisno2 == 0 || axisno2 == diag )
      dd = ftp_coord_match_col_axis( block, colno, axisno, sub );
    
    /* Make a new WCS with this axisno and sub */
    if ( !dd )
      dd = ftp_create_default_wcs( block, colno, axisno, sub );  
  }
  
  return dd;
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_coord_wcskey_match                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Match a WCS key pattern.                                   */
/*     CDELT matches CDELTxxx  (any trailing chars)            */
/*     A leading * character is T or digit:                    */
/*       *CRPX  matches TCRPXn and 1CRPXn...                   */
/*     This supports the BINTABLE WCS keywords                 */
/*                                                             */
/* Input parameters:                                           */
/*      const  char  *key        - key name                    */
/*      const  char  *keyPrefix  - name to match               */
/*                                                             */
/* Output parameters:                                          */
/*       none                                                  */
/*                                                             */
/* Returns:                                                    */
/*         int match - length of string matching keyPrefix     */
/*                                                             */
/***************************************************************/
int ftp_coord_wcskey_match( const char *key, const char *keyPrefix )
{
  char c = '\0';
  long siz   = 0;
  int  match = 0;

  if ( !key || !key[0] )
    return 0;

  siz = strlen( keyPrefix );
  c   = key[0];

  if ( keyPrefix[0] == '*' )
  {
    match = ( isdigit( c ) || c == 'T' ) && 
      cs_eq_siz( key+1, keyPrefix+1, siz - 1 );
  }
  else
  {
    match = cs_eq_siz( key, keyPrefix, siz );
  }

  if ( match )
    match = (long)strlen( key ) > siz && !isalpha( key[siz] );

  return match;
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_coord_set_rotmat_element                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*   Take a FITS PC/CD rotation-matrix keyword and store its   */
/*   value with the coord descriptor associated with the       */
/*   keyword's row in the matrix (e.g. PC2_3 will get          */
/*   attached to column/axis 2).  For non-diagonal elements,   */
/*   also associate the coord with it's partner in the         */
/*   rotation (e.g. column/axis 3 in previous example).        */
/*                                                             */
/*   We also handle LTM keywords here, which are treated just  */
/*   like CD until we actually calculate the rotation angle.   */
/*                                                             */
/* Input parameters:                                           */
/*          dmBlock   *block                                   */
/*          dmuKey    *ftkey                                   */
/*          int        is_cs                                   */
/*          int        is_ltm                                  */
/*                                                             */
/* Output parameters:                                          */
/*          none                                               */
/*                                                             */
/* Returns:                                                    */
/*          void                                               */
/*                                                             */
/***************************************************************/
void ftp_coord_set_rotmat_element( dmBlock *block, dmuKey *ftkey,
                                   int is_cd, int is_ltm )
{
  ftBlock  *ftblock = (ftBlock*)dmi_block_get_kb( block );
  ftCoord  *ftcoord = NULL;
  ftCoord  *partner_ftcoord = NULL;

  dmDescriptor  *dd = NULL;
  dmDescriptor  *partner_dd = NULL;

  char  *key  = NULL;
  char  *name = NULL;
  char  *partner_name = NULL;
  char   sub = ' ';

  long  colno    = 0;
  long  axisno   = 0;
  long  axisno2  = 0;
  long  naxes    = 0;	 /* Number of columns/axes in table/image */
  long  pc_rowno = 0;
  long  pc_colno = 0;    /* i,j in PCi_j */
  long  partner_colno  = 0;
  long  partner_axisno = 0;
  int   is_image = 0;


  /* Quit if any input is NULL */
  if ( !block || !ftblock || !ftkey )
    return;

  /* Extract info from keyword */

  key      = ftkey->name;
  is_image = ( dmi_block_get_raw_type( block ) == dmIMAGE );
  colno    = ftp_parse_wcs_key( key, &axisno, &axisno2, &sub, is_image );
  pc_colno = axisno2;

  if ( sub == alt_wcs_phys_sub )
    sub = 'P';

  /* Interpretation of ftp_parse_wcs_key() results depends on */
  /* block/column type                                        */

  if ( is_image )
  {
    /* Image */
    naxes = ftblock->naxes;
    pc_rowno = axisno;
    partner_colno = colno;
    partner_axisno = axisno2;
  }
  else if ( axisno )
  {
    /* Image in table column */
    /* Currently, we don't support WCS transforms on these, so just quit */
 
   return;
  }
  else
  {
    /* Table */
    naxes = ftblock->ncols;
    pc_rowno = colno;
    partner_colno = axisno2;
    partner_axisno = axisno;
  }

  /* Validate keyword info */

  /* Same checks as in ftp_coord_find_key() */
  if ( ( colno == 0 && axisno == 0 ) || ( is_image && axisno > naxes ) )
    return;
  
  if ( pc_rowno < 1 || pc_rowno > naxes || pc_colno < 1 || pc_colno > naxes )
  {
    /* FIXME: enable this when we can do proper warnings */

    // sprintf( dm_errbuf, "Keyword %s specifies bad matrix indices and will be ignored.", key );
    // dmu_err_set( DM_WARN_BAD_ROT_KEY, dm_errbuf );

    return;
  }
  
  /* Get coord descriptor corresponding to row number for this keyword */
  /* in PC/CD matrix */
  
  dd = ftp_coord_match_col_axis( block, colno, axisno, sub );

  if ( !dd )
    dd = ftp_create_default_wcs( block, colno, axisno, sub );

  ftcoord = dmi_descriptor_get_kdd( dd );
  
  
  /* If this isn't an LTM keyword, discard any previous LTM/LTV keywords. */
  /* Otherwise, use it only if no standard WCS keys have been found.      */

  if ( !is_ltm )
    ftp_coord_clear_ltm( dd );
  else if ( !ftp_coord_is_ltm( dd ) )
    return;
  
  /* Deal with case where we have both PC and CD keywords;
  *    PC takes precedence  */

  if ( ftcoord->pc_row )
  {
    if ( is_cd && !ftcoord->is_cd )
    {
      /* This is a CD keyword, but we've already found PC keywords,*/
      /* so do nothing */

      return;
    }

    if ( !is_cd && ftcoord->is_cd )
    {
      /* We've previously found CD keywords, but this is a PC keyword, */
      /* so dump the CD matrix in favor of PC.  However, we must keep  */
      /* the partner specified by the CD keywords, since it's already  */
      /* associated with us.                                           */

      free( ftcoord->pc_row );
      ftcoord->pc_row = NULL;
      ftcoord->is_cd = 0;
    }
  }


  /* Allocate the matrix row (with default values) if necessary   */
  if ( !ftcoord->pc_row )
  {
    ftcoord->pc_row = (double*)calloc( 2, sizeof(double) );
    
    if ( !is_cd )
    {
      ftcoord->is_cd = 0;

      /* calloc() initializes all bits to zero, but for PC */
      /* diagonal elements should default to 1 */

      ftcoord->pc_row[0] = 1.0;  /* diagonal element is always pc_row[0] */
    }
    else
    {
      ftcoord->is_cd = 1;
    }
  }

  /* Store the matrix element and find any partner coord  */
  if ( pc_rowno == pc_colno )
  {
    /* Diagonal element; doesn't specify any partner, just store the value */

    dmu_key_typed_value( ftkey, DT_REAL8, &(ftcoord->pc_row[0]) );
  }
  else if ( abs( pc_rowno - pc_colno ) <= 1 )
  {
    /* Off-diagonal element; associates this axis with another, so mark them */
    /* as partners.  (Note that the above test ensures that paired axes are  */
    /* adjacent in the FITS file.  Is this definitely necessary? )           */

    /* Get partner coord, i.e. coord descriptor corresponding to column number for */
    /* this keyword in PC/CD matrix                                                */

    partner_dd = ftp_coord_match_col_axis( block, partner_colno, partner_axisno, sub );

    if ( !partner_dd )
      partner_dd = ftp_create_default_wcs( block, partner_colno, partner_axisno, sub );

    partner_ftcoord = dmi_descriptor_get_kdd( partner_dd );

    /* Get descriptor names */
    name = dmi_descriptor_get_name( dd );
    partner_name = dmi_descriptor_get_name( partner_dd );

    /* Verify that dd and partner_dd are either unpaired or already paired */
    /* with each other; quit if they aren't                                */

    if ( ftcoord->partner_name && !cs_eq( ftcoord->partner_name, partner_name ) )
    {
      /* FIXME: enable this when we can do proper warnings */
      // sprintf( dm_errbuf, "Descriptor %s is already paired with %s in a rotation "
      // "and cannot be paired with %s.  Keyword %s will be ignored.",
      // name, ftcoord->partner_name, partner_name, key );
      // dmu_err_set( DM_WARN_BAD_ROT_KEY, dm_errbuf );

      return;
    }
    else if ( partner_ftcoord->partner_name && !cs_eq( partner_ftcoord->partner_name, name ) )
    {
      /* FIXME: enable this when we can do proper warnings */
      // sprintf( dm_errbuf, "Descriptor %s is already paired with %s in a rotation "
      // "and cannot be paired with %s.  Keyword %s will be ignored.",
      // partner_name, partner_ftcoord->partner_name, name, key );
      // dmu_err_set( DM_WARN_BAD_ROT_KEY, dm_errbuf );

      return;
    }

    /* Store off-diagonal element */
    dmu_key_typed_value( ftkey, DT_REAL8, &(ftcoord->pc_row[1]) );

    /* Store partner info (if we haven't already) */

    if ( !ftcoord->partner_name )
    {
      ftcoord->composite = 1;
      ftcoord->partner_name = cs_dup( partner_name );
    }

    if ( !partner_ftcoord->partner_name )
    {
      partner_ftcoord->composite = 1;
      partner_ftcoord->partner_name = cs_dup( name );
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_coords_update_rotation                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*   Loop through all raw columns, their axis groups, and the  */
/*   coordinates on their axis groups, applying any rotation   */
/*   matrices that were read in via PC/CD keywords.  Every     */
/*   coord descriptor gets passed to ftp_coord_apply_rotmat(), */
/*   which does nothing is there's no rotation to apply.       */
/*   							       */
/*   This function must be called *BEFORE*                     */
/*     ftp_coords_update_physical().                           */
/*   						               */
/*   The descriptor search algorithm is taken from             */
/*     ftp_coord_match_col_axis().                             */
/*                                                             */
/* Input parameters:                                           */
/*          dmBlock   *block                                   */
/*                                                             */
/* Output parameters:                                          */
/*          none                                               */
/*                                                             */
/* Returns:                                                    */
/*          void                                               */
/*                                                             */
/***************************************************************/
void ftp_coords_update_rotation( dmBlock *block )
{
  ftBlock  *ftblock = NULL;

  dmDescriptor* dd       = NULL;
  dmDescriptor* col      = NULL;
  dmDescriptor* coord_dd = NULL;

  long  colno   = 0;  /* Column number              */
  long  ncoord  = 0;  /* Number of coords on column */
  long  ngroups = 0;  /* Number of axis groups      */
  long  ii      = 0;
  long  kk      = 0;

  ftblock = (ftBlock*)dmi_block_get_kb( block );

  if ( !ftblock )
    return;

  for ( colno = 1; colno <= ftblock->ncols; colno++ )
  {
    /* Get unscaled column */
    col = ftblock->cols[colno-1];

    /* Get scaled column */
    if ( col )
      col = dmi_descriptor_get_scaled( col );

    if ( !col )
      continue;

    /* Process all coords on this column */

    ncoord = dmi_descriptor_get_no_coords( col );
    for ( ii = 1; ii <= ncoord; ii++ )
    {
      dd = dmi_descriptor_get_coord_no( col, ii );
      ftp_coord_apply_rotmat( block, dd );
    }
 
   /* Process all axis groups on this column */

    ngroups = dmi_descriptor_get_no_groups( col );

    for ( ii = 1; ii <= ngroups; ii++ )
    {
      dd = dmi_descriptor_get_group_no( col, ii );

      if ( dd )
      {
	ftp_coord_apply_rotmat( block, dd );

	/* Process each world coord belonging to this axis group */
	ncoord = dmi_descriptor_get_no_coords( dd );

	for ( kk = 1; kk <= ncoord; kk++ )
	{
	  coord_dd = dmi_descriptor_get_coord_no( dd, kk );
	  ftp_coord_apply_rotmat( block, coord_dd );
	}
      }
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_coord_convert_ltm_to_cd                      */
/*                                                             */
/* Description:                                                */
/*   Invert the LTM matrix given by the input parameters to    */
/*   produce the corresponding CD matrix.                      */
/*   If all goes well, return 1.                               */
/*   If the matrix is not invertible, return 0.                */
/*                                                             */
/* Input parameters:                                           */
/*          double  *cd11                                      */
/*          double  *cd12                                      */
/*          double  *cd21                                      */
/*          double  *cd22                                      */
/*                                                             */
/* Output parameters:                                          */
/*          none                                               */
/*                                                             */
/* Returns:                                                    */
/*         static int  - 0 or 1                                */
/*                                                             */
/***************************************************************/
static int ftp_coord_convert_ltm_to_cd( double *cd11, double *cd12,
                                        double *cd21, double *cd22 )
{
 double ltm11   = *cd11;
 double ltm12   = *cd12;
 double ltm21   = *cd21;
 double ltm22   = *cd22;
 double det_ltm = ltm11*ltm22 - ltm12*ltm21;

 if ( fabs(det_ltm) < DBL_EPSILON )
  return 0;

 /* Standard 2x2 matrix inversion */
 *cd11 =  ltm22 / det_ltm;
 *cd12 = -ltm12 / det_ltm;
 *cd21 = -ltm21 / det_ltm;
 *cd22 =  ltm11 / det_ltm;

 return 1;
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_coord_apply_rotmat                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*   If descriptor dd1 has a rotation matrix (PC/CD keywords)  */
/*   associated with it, use it to extract and set the         */
/*   rotation angle and CDELT* values for it and its partner   */
/*   descriptor.  If the matrix defines more than just         */
/*   rotation (e.g. a general linear transformation), set an   */
/*   error condition and quit.                                 */
/*                                                             */
/* Input parameters:                                           */
/*          dmBlock       *block                               */
/*          dmDescriptor  *dd1                                 */
/*                                                             */
/* Output parameters:                                          */
/*                                                             */
/* Returns:                                                    */
/*                                                             */
/***************************************************************/
void ftp_coord_apply_rotmat( dmBlock *block, dmDescriptor *dd1 )
{

  ftBlock *kb       = NULL;       /* FITS kernel block desc. */
  ftCoord *ftcoord1 = NULL;       /* FITS kernel coord. desc.*/
  ftCoord *ftcoord2 = NULL;       /* FITS kernel coord. desc.*/

  dmDescriptor     *dd2  = NULL;  /* DM Descriptor           */
  dmCoordTransform *aux1 = NULL;
  dmCoordTransform *aux2 = NULL;

  double cd11 = 0.0;  /* rotation matrix elements */
  double cd12 = 0.0;
  double cd21 = 0.0;
  double cd22 = 0.0;

  double cdelt1 = 0.0;
  double cdelt2 = 0.0;
  double rho_a  = 0.0;
  double rho_b  = 0.0;
  double crota2 = 0.0;
  double crpix  = 0.0;
  double crval  = 0.0;  /* Not actually used in algorithm */
  double coscrota2 = 0.0;
  double sincrota2 = 0.0;

  int is_pc = 1;
  int ok    = 1;
  
  aux1     = (dmCoordTransform*)dmi_descriptor_get_aux( dd1 );
  ftcoord1 = dmi_descriptor_get_kdd( dd1 );
  
   /* Quit if any input args are NULL */
  if ( !block || !dd1 || !aux1 || !ftcoord1 )  
    return;
  
  if ( !ftcoord1->partner_name )
  {
    /* Coord doesn't have a partner, so there's no rotation.  However,    */
    /* if we read in a CD1_1, then use it to set CDELT1.  (This opens the */
    /* question of whether to believe CDELT1 or CD1_1 in the case where   */
    /* the file has both and they differ.  For now, just assume they're   */
    /* meant to be the same.)                                             */

    if ( ftcoord1->pc_row && ftcoord1->is_cd )
    {
      cdelt1 = ftcoord1->pc_row[0];
      
      if ( fabs(cdelt1) >= DBL_EPSILON )
      {
	/* If we have LTM1_1, convert to CD1_1 */

	if ( ftcoord1->is_ltm )
	  cdelt1 = 1.0 / cdelt1;
	
	dmi_coord_set_delt( dd1, &cdelt1, 1 );
      }
    }
    return;
  }
  
  /* Find partner coord; quit on failure */
  
  kb = (ftBlock*)dmi_block_get_kb( block );
  dd2 = dmi_descriptor_list_search( (const dmDescriptor**)kb->cols, 
				        kb->ncols, ftcoord1->partner_name );
  if ( !dd2 )
  {
    /* FIXME: enable this when we can do proper warnings */
    // sprintf( dm_errbuf, "Cannot find descriptor %s, which is partnered with 
    // descriptor %s in a WCS rotation", ftcoord1->partner_name, dmi_descriptor_get_name(dd1) );
    // dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );

    return;
  }
  
  aux2 = (dmCoordTransform*)dmi_descriptor_get_aux( dd2 );
  ftcoord2 = dmi_descriptor_get_kdd( dd2 );
  
  
  /* Quit if there's no rotation matrix to apply */
  if ( !ftcoord1->pc_row && !ftcoord2->pc_row  )  
    return;
  

  /* PC takes precedence over CD. */
  /*   only use CD if neither coord has PC keywords associated with it */
  if (!( ftcoord1->pc_row && !ftcoord1->is_cd ) &&
      !( ftcoord2->pc_row && !ftcoord2->is_cd ))
  {
    is_pc = 0;
  }
  
  /* CROTA takes precedence over CD */
  /*  so if we have both (but not PC), discard CD and quit */

  if (!is_pc &&
      ( ( aux1 && fabs( dmi_coord_get_theta(aux1) ) >= DBL_EPSILON ) ||
        ( aux2 && fabs( dmi_coord_get_theta(aux2) ) >= DBL_EPSILON ) )
      )
  {
    free( ftcoord1->pc_row );
    free( ftcoord2->pc_row );
    ftcoord1->pc_row = NULL;
    ftcoord2->pc_row = NULL;
    ftcoord1->is_cd = 0;
    ftcoord2->is_cd = 0;
    
    return;
  }
  

  /* If one coord has PC and the other has CD, discard the CD  */
  if ( ftcoord1->pc_row && !ftcoord1->is_cd )
  {
    if ( ftcoord2->pc_row && ftcoord2->is_cd )
    {
      free( ftcoord2->pc_row );
      ftcoord2->pc_row = NULL;
      ftcoord2->is_cd = 0;
    }
  }
  else if ( ftcoord2->pc_row && !ftcoord2->is_cd )
  {
    if ( ftcoord1->pc_row && ftcoord1->is_cd )
    {
      free( ftcoord1->pc_row );
      ftcoord1->pc_row = NULL;
      ftcoord1->is_cd = 0;
    }
  }
  
  /* If one coord has CD and the other has LTM, discard the LTM  */
  if ( !is_pc )
  {
    if ( ftcoord1->pc_row && !ftcoord1->is_ltm )
    {
      if ( ftcoord2->pc_row && ftcoord2->is_ltm )
      {
	free( ftcoord2->pc_row );
	ftcoord2->pc_row = NULL;
	ftcoord2->is_ltm = 0;
      }
    }
    else if ( ftcoord2->pc_row && !ftcoord2->is_ltm )
    {
      if ( ftcoord1->pc_row && ftcoord1->is_ltm )
      {
	free( ftcoord1->pc_row );
	ftcoord1->pc_row = NULL;
	ftcoord1->is_ltm = 0;
      }
    }
  }
  
  /* Get rotation matrix elements. */
  /*   use default values if none were specified  */
  
  /* FIXME: we're assuming that dd1,dd2 are in correct lat/lon order !! */
  if ( ftcoord1->pc_row )
  {
    cd11 = ftcoord1->pc_row[0];
    cd12 = ftcoord1->pc_row[1];
  }
  else if ( is_pc )
  {
    cd11 = 1.0;
  }
  
  if ( ftcoord2->pc_row )
  {
    /* Diagonal matrix element is always pc_row[0] */
    cd21 = ftcoord2->pc_row[1];
    cd22 = ftcoord2->pc_row[0];
  }
  else if ( is_pc )
  {
    cd22 = 1.0;
  }
  
  
  /* Calculate rotation angle and cdelts  */
  if ( is_pc )
  {
    /* cdelts default to 1.0 in ftp_create_wcs(), so we don't need to */
    /* worry about setting them here */
    
    dmi_coord_get_transform_d( dd1, &crpix, &crval, &cdelt1, 1 );
    dmi_coord_get_transform_d( dd2, &crpix, &crval, &cdelt2, 1 );
      
    cd11 *= cdelt1;
    cd12 *= cdelt1;
    cd21 *= cdelt2;
    cd22 *= cdelt2;

  }
  else if ( ftcoord1->is_ltm || ftcoord2->is_ltm )
  {
    /* If we have LTM, convert it to CD */
    ok = ftp_coord_convert_ltm_to_cd( &cd11, &cd12, &cd21, &cd22 );
  }
  
  if ( ok )
  {
    if ( fabs(cd21) < DBL_EPSILON )
      rho_a = 0.0;
    else if ( cd21 > 0.0 )
      rho_a = ARG( cd11, cd21 );
    else
      rho_a = ARG( -cd11, -cd21 );
      
    if ( fabs(cd12) < DBL_EPSILON )
      rho_b = 0.0;
    else if ( cd12 > 0.0 )
      rho_b = ARG( -cd22, cd12 );
    else
      rho_b = ARG( cd22, -cd12 );
      
    if ( fabs( rho_a - rho_b ) >= FTOL )
      ok = 0;
  }
  
  if ( ok )
  {
    crota2    = ( rho_a + rho_b ) / 2.0;
    coscrota2 = dmcos(crota2);
    sincrota2 = dmsin(crota2);

    /* Have these equations for cdelt1 and cdelt2 */
    /* 	    cdelt1 =  cd11 / cos(crota2)          */
    /* 	    cdelt1 =  cd21 / sin(crota2)          */
    /* 	    cdelt2 = -cd12 / sin(crota2)          */
    /* 	    cdelt2 =  cd22 / cos(crota2)          */
    /*                                            */
    /* so                                         */

    cdelt1 = sqrt( cd11*cd11 + cd21*cd21 );
    cdelt2 = sqrt( cd12*cd12 + cd22*cd22 );

    /* determine sign of deltas */
    /* Use product rather than quotient to avoid divisor problems */

    if ( fabs(coscrota2) >= DBL_EPSILON )
    {
      if ( cd11*coscrota2 < 0.0 ) 
	cdelt1 *= -1.0;

      if ( cd22*coscrota2 < 0.0 )
	cdelt2 *= -1.0;
    }
    else
    {
      /* Angle near +/-90 deg.. use sine versions */
      if ( cd21*sincrota2 < 0.0 )
	cdelt1 *= -1.0;

      if ( -cd12 / sincrota2 < 0.0 )
	cdelt2 *= -1.0;
    }
      
    if ( fabs(cdelt1) < DBL_EPSILON ) cdelt1 = 0.0;
    if ( fabs(cdelt2) < DBL_EPSILON ) cdelt2 = 0.0;
      
    /* If the other solution for CROTA2 gives CDELT1<0, CDELT2>0, */
    /* use it instead */

    if ( ( -cdelt1 < 0.0 ) && ( -cdelt2 > 0.0 ) )
    {
      crota2 -= M_PI;
      cdelt1 *= -1.0;
      cdelt2 *= -1.0;
    }
  }
  
#if 0
  if ( !ok )
  {
    /* FIXME: enable this when we can do proper warnings */

      sprintf( dm_errbuf, "WARNING: WCS linear transformation matrix "
      "(given by PC/CD/TP/TC keywords) on descriptors %s,%s does not "
      "define a simple rotation and will be ignored",
      dmi_descriptor_get_name(dd1), dmi_descriptor_get_name(dd2) );
      dmu_err_set( DM_WARN_BAD_ROT_KEY, dm_errbuf );
  }
#endif
  
  /* Store rotation angle (if it's non-zero) and cdelts */
  if ( ok )
  {
    /* Radians to degrees */
    crota2 = 180.0 * ( crota2 / M_PI );
      
    if ( fabs(crota2) >= DBL_EPSILON )
    {
      /* Set theta on both components, just to be safe */
      dmi_coord_set_theta( dd1, crota2 );
      dmi_coord_set_theta( dd2, crota2 );
    }
      
    dmi_coord_set_delt( dd1, &cdelt1, 1 );
    dmi_coord_set_delt( dd2, &cdelt2, 1 );
  }
  
  
  /* Free pc_row so that it isn't applied again     */
  /*  (e.g. when this function gets called for dd2) */

  if ( ftcoord1->pc_row )
  {
    free( ftcoord1->pc_row );
    ftcoord1->pc_row = NULL;
  }

  if ( ftcoord2->pc_row )
  {
    free( ftcoord2->pc_row );
    ftcoord2->pc_row = NULL;
  }
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_coord_is_ltm                                 */
/*                                                             */
/* Description:                                                */
/*   Test whether LTV/LTM keywords should be applied to        */
/*   coordinate descriptor dd.  Really, we're checking whether */
/*   we've called ftp_coord_clear_ltm() because we've found    */
/*   standard WCS keywords for dd (and therefore want to       */
/*   ignore LTV/LTM).                                          */
/*                                                             */
/* Input parameters:                                           */
/*          const dmDescriptor  *dd                            */
/*                                                             */
/* Output parameters:                                          */
/*          none                                               */
/*                                                             */
/* Returns:                                                    */
/*          int  ftcoord->is_ltm                               */
/*                                                             */
/***************************************************************/
int ftp_coord_is_ltm( const dmDescriptor *dd )
{
  ftCoord* ftcoord = dmi_descriptor_get_kdd( dd );

  if ( !ftcoord )
    return 0;

  return ftcoord->is_ltm;
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_coord_clear_ltm                              */
/*                                                             */
/* Description:                                                */
/*   To be called if any standard WCS keywords are found for   */
/*   coordinate descriptor dd so that we ignore any subsequent */
/*   LTV/LTM keywords for it.  The first time this function is */
/*   called for dd, it resets CDELT, CRPIX, CRVAL to their     */
/*   default values and erases any rotation matrix, so that    */
/*   none of the information obtained from previously-read     */
/*   LTV/LTM keywords gets used.  On subsequent calls,         */
/*   it does nothing.                                          */
/*                                                             */
/* Input parameters:                                           */
/*          dmDescriptor  *dd                                  */
/*                                                             */
/* Output parameters:                                          */
/*          none                                               */
/*                                                             */
/* Returns:                                                    */
/*          void                                               */
/*                                                             */
/***************************************************************/
void ftp_coord_clear_ltm( dmDescriptor *dd )
{
  ftCoord *ftcoord = dmi_descriptor_get_kdd( dd );

  double delt = 1.0;
  double pix  = 0.0;
  double val  = 0.0;

  if ( !ftcoord )
    return;

  if ( ftcoord->is_ltm )
  {
    ftcoord->is_ltm = 0;

    dmi_coord_set_delt( dd, &delt, 1 );
    dmi_coord_set_pix( dd, &pix, 1 );
    dmi_coord_set_val( dd, &val, 1 );

    if ( ftcoord->pc_row )
    {
      free( ftcoord->pc_row );
      ftcoord->pc_row = NULL;
    }
  }
}


/***************************************************************/
/*                                                             */
/* Function:  ftp_rename_default_wcs                           */
/*                                                             */
/* Description:                                                */
/*  Replaces names of default WCS descriptors to names with    */
/*  some meaning IF those names are not already being used     */
/*  (ie 'X','Y' for image axes 1 & 2 respectively)             */
/*                                                             */
/* Input parameters:                                           */
/*          dmBlock  *block                                    */
/*                                                             */
/* Output parameters:                                          */
/*          none                                               */
/*                                                             */
/* Returns:                                                    */
/*          void                                               */
/*                                                             */
/***************************************************************/
void ftp_rename_default_wcs( dmBlock *block )
{
  ftBlock      *ftblock = (ftBlock*)dmi_block_get_kb( block );

  dmBlockType   type;
  dmDescriptor *dd = NULL;
  dmDescriptor *td = NULL;

  char *default_axis_names[3] = { "X", "Y", "Z" };
  char  name[80]; 

  long  ii = 0;


  type = dmi_block_get_raw_type( block );

  if( type != dmTABLE )
  {
    /* Only want to replace the names for the first 3 axes.. */
    /* so look for those defaults specifically               */

    for ( ii = 0; ii < 3; ii++ )
    {
      /* Check for descriptor with default name */
      sprintf( name, "#AXIS%ld", ii+1 );
      dd = dmi_descriptor_search(ftblock->cols[0], name);

      if ( dd )
      {
	/* Check if there is a descriptor with the replacement name */
	cs_copy( default_axis_names[ii], name );
	td = dmi_descriptor_search(ftblock->cols[0], name);

	if ( !td )
	{
	  /* There is none.. so can rename */
	  dmi_descriptor_set_name( dd, default_axis_names[ii] );
	}
      }
    } /* end axis loop */
  } /* end type check */
}

/***************************************************************/
/*                                                             */
/* Function:   ftp_use_equinox                                 */
/*                                                             */
/* Description:                                                */
/*     Check for equitorial or ecliptic coordinate system      */
/*     Return dmTRUE or dmFALSE. Used to determine whether or  */
/*     not to apply EQUINOX value                              */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*           char       *ctype_name                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*          dmBool  dmTRUE or dmFALSE                         */
/*                                                             */
/***************************************************************/
dmBool ftp_use_equinox( char *ctype_name)
{
  char name[5];

  char *pStart = NULL;
  char *pEnd   = NULL;
  char  *sDash = "-";
  char  cNull  ='\0';

  long  len = 0;

  pStart = ctype_name;
  pEnd = strpbrk(pStart, sDash );
  len  = pEnd - pStart;

  strncpy( name, pStart, len);
  name[len] = cNull;
 
  // Is this a name of a know column
  if ( cs_eq(name, "RA" ))
       return dmTRUE;

  if (cs_eq(name, "DEC"))
    return dmTRUE;

  if (cs_eq(name, "ELON"))
    return dmTRUE;

  if (cs_eq(name, "ELAT"))
    return dmTRUE;

  if (cs_eq(name, "HLON"))
    return dmTRUE;

  if (cs_eq(name, "HLAT"))
    return dmTRUE;

  return dmFALSE;
  
}
