/*                                                                
**  Copyright (C) 2002-2003,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   ftwrite.c                                                   */
/*                                                               */
/*   Functions for writing FITS files                            */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmilib.h"
#include "ftkernelP.h"



/***************************************************************/
/*                                                             */
/* Function:   ftp_insert_key_generic                          */
/*                                                             */
/* Description:                                                */
/*              Write key to FITS file.  Called by             */
/*              ftp_key_write_generic() only if file is        */
/*              writable                                       */
/*                                                             */
/*  Input parameters:                                          */
/*              fitsfile    *fptr                              */
/*              long         pos                               */
/*              char        *name                              */
/*              dmDataType   type                              */
/*              const void  *value                             */
/*              char        *comment                           */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            int status                                       */
/*                                                             */
/***************************************************************/
int ftp_insert_key_generic( fitsfile* fptr, long pos, char* name,
                            dmDataType type, const void* value, char* comment )
{
  long   lval   = 0;
  double dval   = 0;
  char   *ptr   = NULL;
  char   *Blank = " ";

  long   k      = 0;
  char   card[FITS_MAXLEN] = " ";

  int    status = FITS_KERNEL_OK;

  /* Position HDU at pos */
  if ( pos > 0 )
  {
    fits_read_record( fptr, (int)(pos-1), card, &status );        /* CFITSIO */

    /* Hack to make sure you don't tread on CONTINUE */
    k = strlen( card );

    while ( k> 2 && card[k-1] == '\'' && card[k-2] == '&')
    {
      ++pos;

      fits_read_record( fptr, (int)(pos-1), card, &status );      /* CFITSIO */

      k = strlen( card );
    }
  }

  switch( type )
  {
    case DT_INT4:
      lval = *(long*)value;
      fits_insert_key_lng( fptr, name, lval, comment, &status );             /* CFITSIO */
      break;

    case DT_INT2:
      lval = (long) *(short*)value;
      fits_insert_key_lng( fptr, name, lval, comment, &status );            /* CFITSIO */
      break;

    case DT_UINT1:
      lval = (long) *(unsigned char*)value;
      fits_insert_key_lng( fptr, name, lval, comment, &status );            /* CFITSIO */
      break;

    case DT_UINT2:
      lval = (long) *(unsigned short*)value;
      fits_insert_key_lng( fptr, name, lval, comment, &status );            /* CFITSIO */
      break;

    case DT_REAL4:
      fits_insert_key_flt( fptr, name, *(float*)value, FITSK_FLOAT_DECIMALS,
                           comment, &status );                              /* CFITSIO */
      break;

    case DT_REAL8:
      fits_insert_key_dbl( fptr, name, *(double*)value, FITSK_DOUBLE_DECIMALS,
                           comment, &status );                              /* CFITSIO */
      break;

    case DT_LOGICAL:
      fits_insert_key_log( fptr, name, *(int*)value, comment, &status );    /* CFITSIO */
      break;

    case DT_UINT4:
      dval = (double)*(unsigned long*)value;
      fits_insert_key_fixdbl( fptr, name, dval, 0, comment, &status );      /* CFITSIO */
      break;

    case DT_CHAR:
      ptr = *(char**)value;
      if ( cs_is_blank( ptr ) ){ ptr = Blank; }

      if ( ftp_key_is_comment( name ) )
      {
        status = ftp_write_comment( fptr, name, ptr );
      }
      else if ( ptr && ( strlen( ptr ) > FITS_MAXSTRKEY ) )
      {
        /* Note there is no fits_insert_key_longstr routine      */
        /* Note also a bug in writing a null string with longstr */
	fits_write_key_longwarn( fptr, &status );                         /* CFITSIO */
	fits_write_key_longstr( fptr, name, ptr, comment, &status );      /* CFITSIO */

	if (status == FITSERR207)
	{
	  status = FITS_KERNEL_OK;
	  ftp_write_remove_invalid_chars( ptr );
	  ftp_write_remove_invalid_chars( comment );
	  fits_write_key_longstr( fptr, name, ptr, comment, &status );
	}
      }
      else
      {
	fits_insert_key_str( fptr, name, ptr, comment, &status );       /* CFITSIO */

	/* Code change  Bug 9522-A  June 12, 2006  */
	if ( status == FITSERR207 )
	{
	  status = FITS_KERNEL_OK;
	  ftp_write_remove_invalid_chars( ptr );
	  ftp_write_remove_invalid_chars(comment);
	  fits_insert_key_str( fptr, name, ptr, comment, &status );
	}

      }
      break;

    default:
      break;

  } // end switch

  return status;
}



/***************************************************************/
/*                                                             */
/* Function:   ftp_update_key_generic                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              fitsfile    *fptr                              */
/*              char        *name                              */
/*              dmDataType   type                              */
/*              const void  *value                             */
/*              char        *comment                           */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            int status                                       */
/*                                                             */
/***************************************************************/
int ftp_update_key_generic( fitsfile* fptr, char* name, dmDataType type,
                            const void* value, char* comment )
{
  int     status = FITS_KERNEL_OK;
  long    lval   = 0;
  double  dval   = 0;
  char  **pptr   = NULL;
  char   *ptr    = NULL;
  char   *Blank = " ";


  if ( !value )
  {
    sprintf( dm_errbuf, "No value while writing key %s", name );
    dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
    status = KEY_NO_EXIST; 
    return  status;
  }


  switch( type )
  {
    case DT_INT4:
      lval = *(long*)value;
      fits_update_key_lng( fptr, name, lval, comment, &status );             /* CFITSIO */
      break;

    case DT_INT2:
      lval = (long) *(short*)value;
      fits_update_key_lng( fptr, name, lval, comment, &status );            /* CFITSIO */
      break;

    case DT_UINT1:
      lval = (long) *(unsigned char*)value;
      fits_update_key_lng( fptr, name, lval, comment, &status );            /* CFITSIO */
      break;

    case DT_UINT2:
      lval = (long) *(unsigned short*)value;
      fits_update_key_lng( fptr, name, lval, comment, &status );            /* CFITSIO */
      break;

    case DT_REAL4:
      fits_update_key_flt( fptr, name, *(float*)value, FITSK_FLOAT_DECIMALS, comment, &status );            /* CFITSIO */
      break;

    case DT_REAL8:
      fits_update_key_dbl( fptr, name, *(double*)value, FITSK_DOUBLE_DECIMALS, comment, &status );            /* CFITSIO */
      break;

    case DT_LOGICAL:
      fits_update_key_log( fptr, name, *(int*)value, comment, &status );            /* CFITSIO */
      break;

    case DT_UINT4:
      dval = (double)*(unsigned long*)value;
      fits_update_key_fixdbl( fptr, name, dval, 0, comment, &status );            /* CFITSIO */
      break;

    case DT_CHAR:
      pptr = (char**)value;
      ptr  = *pptr;
      if ( !ptr || ( *ptr == '\0') ){ ptr = Blank; }

      if ( ftp_key_is_comment( name ) )
      {
        status = ftp_write_comment( fptr, name, ptr );
      }
      else
      {
	if ( ptr && ( strlen( ptr ) > FITS_MAXSTRKEY ) )
          fits_modify_key_longstr( fptr, name, ptr, comment, &status );        /* CFITSIO */
        else 
          /* Bug in CFITSIO for blank/null keys */
          fits_modify_key_str( fptr, name, ptr, comment, &status );            /* CFITSIO */

	if (status == FITSERR207 )
	{
	  status = FITS_KERNEL_OK;

	  ftp_write_remove_invalid_chars( ptr );
	  ftp_write_remove_invalid_chars( comment );

	  if ( ptr && ( strlen( ptr ) > FITS_MAXSTRKEY ) )
	    fits_modify_key_longstr( fptr, name, ptr, comment, &status );      /* CFITSIO */
	  else 
	    fits_modify_key_str( fptr, name, ptr, comment, &status );          /* CFITSIO */
	  
	}
	else if ( status == KEY_NO_EXIST )
        {
	  /* Key not already in file, add it */

	  status = FITS_KERNEL_OK;
	  if ( ptr && ( strlen( ptr ) > FITS_MAXSTRKEY ) )
	  {
	    fits_write_key_longwarn( fptr, &status );            /* CFITSIO */
	    fits_write_key_longstr( fptr, name, ptr, comment, &status );       /* CFITSIO */
	  }
	  else
	    fits_write_key_str( fptr, name, ptr, comment, &status );            /* CFITSIO */

          if (status == FITSERR207 )
	  {
            status = FITS_KERNEL_OK;

            /* Code change - Bug 9522  June 12, 2006             */
            /* If write fails, check for invalid chars in string */

            ftp_write_remove_invalid_chars( ptr );
            ftp_write_remove_invalid_chars( comment );

	    if ( ptr && ( strlen( ptr ) > FITS_MAXSTRKEY ) )
              fits_write_key_longstr( fptr, name, ptr, comment, &status );      /* CFITSIO */
            else 
              fits_write_key_str( fptr, name, ptr, comment, &status );          /* CFITSIO */
	  }
	}
      }
      break;

     default:
       break;
   }

 return  status;
}



/***************************************************************/
/*                                                             */
/* Function:   ftp_write_comment                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*              fitsfile    *fptr                              */
/*              const char  *tag                               */
/*              const char  *value                             */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*            int status                                       */
/*                                                             */
/***************************************************************/
int ftp_write_comment( fitsfile *fptr, const char *tag, const char *value )
{
  int  status  = FITS_KERNEL_OK;
  long pos     = 0;
  long len     = strlen( value );

  char card[CARD_SIZE+1];         // buffer is 81 chars long
  char cTag[9];


  // guarantee that tag is no more than 8 chars long
  cTag[0] = '\0';
  strncat(cTag, tag, 8);  

  while( len > 0 )
  {
    /* FITS_COMMENT_SIZE is 72, tag size is 8, card buffer size is 81 */

    sprintf( card, "%-8s", cTag );

    // strncat adds null to end of combined string
    // which should be 80 chars or fewer
    strncat( card, value+pos, FITS_COMMENT_SIZE );


    /* fits_write_record always writes at end of HDU - yuck */
    fits_write_record( fptr, card, &status );                   /* CFITSIO */


    // Code Change - Bug 9522  June 12, 2006
    // If the write fails with an error 207, it is because the comment
    // contains a character that the CFITSIO does not process.  The one
    // most likely is the tab character.  Scan the value for tab and
    // replace with space.

    //  Code Change = Bug 9522  March 24, 2008
    //  Removed copy to temp buffer. Do changes in buffer 'card'
    //  That buffer is already contains a segment from 'value'
    //  of appropriate length

    if ( status == FITSERR207 )
    {
      ftp_write_remove_invalid_chars( card );

      // try writing again and report what happens
      // clear status and then try to write again
      status = FITS_KERNEL_OK;

      fits_write_record( fptr, card, &status );
    }
    // end code change

    pos += FITS_COMMENT_SIZE;
    len -= FITS_COMMENT_SIZE;
  }


  return  status;
}


/***************************************************************/
/*                                                             */
/* Function:   ftp_write_remove_invalid_chars                  */
/*                                                             */
/* Description:                                                */
/*            CFITSIO write functions fail if the string       */
/*            contains tab characters (among others)  Replace  */
/*            tab characters with blanks and retry write.      */
/*                                                             */
/*            To check for more than one character, expand the */
/*            array of characters to replace and update the    */
/*            value of variable iNumOfChars, which is the      */
/*            number of chars to look for and replace.         */
/*                                                             */
/*  Input parameters:                                          */
/*              const char *cTmpValue - array with input string*/
/*                                                             */
/*  Output parameters:                                         */
/*              const char *cTmpValue - contains new string    */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void  ftp_write_remove_invalid_chars(  const char *cTmpValue )
{
  char *pTabChar    = NULL;
  char *pStartScan  = NULL;
  int   ii          = 0; 
  int   iNumOfChars = 1;
  char  cBlank      = ' ';

  char  cInvalidChars[1]  = {'\t'};


  // Only work with non-null pointers
  if ( !cTmpValue || ! *cTmpValue )
    return;


  // Loop through list of all invalid characters

  for (ii = 0; ii < iNumOfChars; ii++)
  {
    // start of string
    pStartScan = (char *) cTmpValue;

    do
    {
      // Scan for invalid characters.  If any, replace with blank

      pTabChar = strchr(pStartScan, cInvalidChars[ii]);

      if (pTabChar)
      {
        *pTabChar  = cBlank;
        pStartScan = pTabChar;
      }
    } while (pTabChar != NULL);
  }

  // all substitutions done
  return;
}


