/*                                                                
**  Copyright (C) 2002-2012  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmi.h                                                       */
/*                                                               */
/*   Definitions and structure definitions for datamodel         */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#ifndef DMI_H
#define DMI_H


#include "dmkernel.h"
#include "dmilib.h"
#include "cxcregion.h"
/* Internal types */
/* Phase 1 parse results */

struct dmParseList_s {
 char* dsname;
 char* block;
 char* outname;
 char* filter;
 char* select;
 char* binning;
 char* option;
 char* subspace;
 int compact;
 dmBlockType type;
};


struct dmRangeList_s {
 void* mins;
 void* maxes;
 long n;
 long nmax;
 int incnull;    /* Include or Exclude NULL value in the range?
		  * 0 => ignore NULL 
		  * 1 => include NULL
		  *-1 => exclude NULL  */

 dmBool sorted; /* Is the rangelist sorted? */

 /* Ryan, This flag is a hack.  It is just so we mark to intersect the 
  * first filter, and union the rest */
 int flag; /* Initialized for intersect? */
};


/* What type of aux field in descriptor? */
#define DM_AUX_COORD 1
#define DM_AUX_BIN   2
#define DM_AUX_SS    3


typedef struct dmBinAxis_s {
 dmDescriptor* dd;
 double** vals; 
 long type;   /* Linear, log or other binning? */
/* May support more complicated binning later */ 
} dmBinAxis;


struct dmBinSpec_s {
 int auxtype;  /* Common field for aux */
 long n;
 long* offsets;
 long* weights;
 long* vcpts;
 long* groups;
 long* axis_map;
 dmBinAxis* axis;
 dmBool compact;
 dmDescriptor* filtered_image;
 long* image_mins;
 long* image_maxes;
 double* min;
 double* max; 
 long* axes;
 double* step;
 long* start;
 int is_binned;
 dmDescriptor* weight_dd;
};


/*
 * The execute stack - used for compiling dynamically defined (DMTCALC-style) columns.
 *  Right now only coordinate columns are supported, e.g. [cols dec]
 */

struct dmExecuteFunc_s {
 long id;
 long* args;
 long nargs;
};
   
struct dmExecute_s {
 dmExecuteFunc* stack;
 long n;
 long nmax;
};

#define DM_EXEC_COORD  1
#define DM_EXEC_ROW    2

struct dmFilter_s {
 dmBool null;
 char* name; /* Table name */
 long dim;
 dmRangeList* ranges;   /* Range list for each component */
 regRegion* region;
 regRegion* logical_region;
 char* regstring;
 dmDataType type;
#if 0
 int flag;  /* Used in parsing; initialize in run time subspace */
#endif
};


struct dmSubspaceData_s {
 int auxtype;
 dmFilter** filters;
 long* mins;     /* Array section info */
 long* maxes;
 dmSubspace* subspace; 
};



struct dmCoordTransform_s {
 int   auxtype;  /* Common field: mark as coord transform */
 dmDescriptor *parent;  /* Transform source column */
 char  *type;         /* Transform type */
 char  *name;              /* Name */
 char  *radesys;
 dmBool physical;  /* Physical? */
 double val[2];
 double pix[2];
 double delt[2];
 double theta;
 double equinox;
 double longpole;
 double latpole;
 long   no;      /* Coord no; 0 if axis group */
 void  *wcs;   /* WCSLIB structure */
 dmBool linear;   /* Is it linear? */
 dmBool scaled;
};


struct dmDescriptor_s
{
 long no;        /* Number in type of descriptor */
 long bufno;     /* Buffer column number */
 char* name;                  /* Name */
 char* unit;                  /* Unit */
 char* desc;                  /* Description */
 char* disp;                   /* Display format */
 dmBlock* block;              /* Parent block */
 dmDescriptorType descType;   /* Descriptor type */
 dmDataType type;             /* Data type */
 dmElementType etype;         /* Element type */
 long length;                 /* Data object length */
 long size;                   /* Number of cells per element */
 long dim;                    /* Vector/Element dimension */
 long vdim;                   /* Vector dimension */
 long edim;                   /* Element dimension */
 char** eltNames;             /* Element names */
 char** cptNames;             /* Component names for each element */
 long cptno;        /* Component number in vector */
 long naxes;                  /* Array dimension */
 long* axes;                  /* Array dimensions */
 dmBool varlen;               /* Is array variable-length? */
 dmDescriptor** cpts;
 dmDescriptor* parent;        /* Parent descriptor if any */
 long ncoords;                /* Ncoords */
 dmDescriptor** coords;        /* Child descriptors if any */
 long ngroups;
 dmDescriptor** groups;         /* Axis groups */
 void* aux;                   /* Descriptor aux structure */
 void* kernel_dd;                   /* kernel descriptor */
 void* null;  /* Null value */
 void* bin;   /* Bin value */
 void* range;  /* Range value */
 long flag;  /* Detailed descriptor type bitmask */
 dmBool declared;   /* Image axes declared yet or not? */
 dmBool update;   /* Descriptor is dirty: use for WCS and DSS which get
                    cached and later flushed  by dmp_block_flush */
 dmBool ref;  /* Reference (unit coord): this field caches unit transforms */
};


struct dmSubspace_s
{
 dmDescriptor** cols;
 long ncols;
 long ncpts;
 long cpt;
 int invert;
 long maxcols;
 long maxcpts;
/* Members for aiding real-time filtering */
 dmRTSCache* cache;
};

/* Real-Time Subspace cache to cache filtering information */
struct dmRTSCache_s {
 long n_filt;              /* Number of filters in the subspace */
 dmDescriptor** ddlist;   /* List of filtered data descriptors */
 dmDescriptor** sub_ddlist; /* List of filtered subspace descriptors */
 dmFilter*** filter_list; /* List of filters in the subspace */
 long* col_offsets; /* subspace col offset to buffer col offset */
 dmBool* accept;    /* For each subspace component, is current row good? */
 dmBool null;       /* Is this subspace a NULL filter? */
};

struct dmIndexBlock_s
{
 char* name;
 dmDataset* ds;
 long no;
 dmBlockType type;
};

struct dmBlock_s 
{
 dmIndexBlock* index_block;
 dmkBlock* kernel_block;
 char* name;
 dmDataset* ds;
 long no;
 dmBlockType type;



 long nrows;   /* Raw nrows */
 long nvrows;   /* Virtual nrows */
 dmDescriptor** keys;
 long nkeys;
 long nkmax;
 dmDescriptor** rawcols;
 long nrcols;
 dmDescriptor** cols;
 long ncols;
 long keyno;          /* Current key position */
 dmSubspace* subspace;  /* Total subspace */
 dmBuffer* buffer;       /* Run time subspace stack */

 char** pref;  /* Preferred column ID (raw) */
 long npref;
 dmBool pref_default;   /* Pref cols are default - don't write them out */
 dmDescriptor* row_dd;
 long maxmeg;       /* Max no of megabytes to allocate */
 double region_bin;   /* Region area pixellation parameter */
 long region_nmax;  /* Region area pixellation parameter */
 int virtual;     /* Is a virtual block */
 dmBool unscaled;   /* Return unscaled descriptors even when scaled descriptors requested */
 dmBool ssupdate; /* Update subspace on filter */
 dmBool compact;  /* Update image size on filter */
 dmBool traversed;  /* Traversed or not filtered - true if we know nvrows */
 char* null;  /* Null value to use in images */
 dmDataType image_bin_type; /* Hint for binning type */
 dmBool update; /* Does the block need flushing?  */
};


struct dmDataset_s 
{
 char*  name;
 dmBool mode;       /* Writeable? */
 dmBlock* current;  /* Current block */
 long current_no;   /* Current block number */
 long nblocks;
 long maxblocks;
 dmIndexBlock** index_blocks;

 long nvblocks;
 long maxvblocks;
 dmBlock** blocks;

 dmKernel kernel;
 dmkDataset* kernel_ds;
};

typedef enum { DMBI_FILT, DMBI_FCALC, DMBI_FREAD, DMBI_READ, DMBI_BINNED, DMBI_OTHER, DMBI_N } DM_BUFFER_INDEX;


struct dmBuffer_s
{
 void* data; /* The actual data */
 void* vdata; /* Cache for filtered data */
 dmBool in_vdata; /* Used the vdata buffer */
 long* index; /* Index into good rows in raw buffer */
 long* contiguous; /* Index of contiguous good rows */
 long* contig_start; /* First row in this contiguous sequence */
 long nmax;   /* Max number of cols (allocated size of cols and flags array) */
 long ncols;  /* Used number of cols */
 dmDescriptor** cols;    /* This includes hidden cols */
 long* offsets;   /* Offsets for each col*/
 long* voffsets;  /* Offsets for each col when hidden cols are excluded */
 long* sizes;     /* Byte sizes for each col */
 long total;      /* Total bytes for buffer record */
 long imtotal;  /* As above plus space for image buffer if present in first row */
 long vtotal;  /* Total bytes for visible (external) buffer record */
 dmBool update;   /* Need to flush buffer? */
 dmBool resync;   /* Need to recalc buffer offsets? (new column) */
 dmBool col_ordered; /* Column or row ordered buffer? */
   /* Used in explicit buffer (not hidden) */
   /* Needed to calculate row */
   /* Needed to filter row */
 long default_size;
 long max_size;    /* Requested maximum number of rows in buffer */
 long kernel_rows;
 long nrows;       /* Number of rows in buffer */
 long rrow1;  /* Raw row - first in buffer */
 long rrow2;  /* Raw row - last in buffer */
 long rrow_seg1;  /* Raw row range for current segment */
 long rrow_seg2;
 long soffset;  /* Offset of srow1 in buffer - usually srow1 - rrow1 */
 long rrow_top;  /* Top slot in buffer rel to rrow1 */
 long rrow_end;   /* Max row allowed in buffer */
 long vrow;     /* Current row */
 long vrow1;    /* Virtual rows in buffer */
 long vrow2;     
 long vrow_go1;  /* Target virtual rows in buffer */
 long vrow_go2;
 long lastrow;   /* Last virtual row filled in buffer */
 dmBool filtered;  /* Is buffer filtered? */
 dmExecute execute1;  /* Eval stack needed prior to filtering */
 dmExecute execute2;  /* Eval stack for stuff not needed for filtering */

 dmSubspace* subspace;
 dmDescriptor* image;
 dmDescriptor*** col_index;
 long** offset_index; 
 long* n_index; /* number of columns in each "index" category */

 /*
  * Filter Workspace.
  * The following variables are kept in the buffer to optimize filtering.
  */
 long minrow; /* Offset in subspace of any row filter; -1 if none */
 long maxrow;
 dmBool empty;    /* If true, vrow1 is one before the beginning. */

 /* the function to perform the actual filtering */
 long (*filter_rows)( dmBuffer* buffer, void* data ); 
};

#endif
