/*                                                                
**  Copyright (C) 2002-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmulib.h                                                    */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#ifndef DMULIB_H
#define DMULIB_H

/* Trigger loading of nutils.h */
#ifndef DMLIB_INTERNAL
#define DMLIB_INTERNAL
#endif

#include "dmlib.h"
#include <float.h>
#include <math.h>

#define DM_NULLTYPE_BAD -1
#define DM_NULLTYPE_NONE 0
#define DM_NULLTYPE_NAN  1
#define DM_NULLTYPE_VAL  2


typedef long dmuKeyType;
typedef struct dmuKey_s dmuKey;

typedef struct dmuKeyList_s
{
  dmuKey** data;
  long n;
  long nmax;
} *dmuKeyList;


struct dmuKey_s {
 long type;    /* Reserved field */
 long index;   /* keylist index of value field */
 char* name;
 char* ucname; /* upper case version of name */
 GenericType value;
 char* desc;
 char* unit;  
 int mask;     /* Used for marking DM visible keys */
 long group;   /* Used for marking grouped keys */
 dmuKeyType keytype;
 dmDataType datatype;
};


/* The vector key object stores the key indices of a structured key
   for conventions analogous to the CTYPEn convention. */

typedef struct dmuVectorKey_s {
 dmuKey* name;
 dmuKey* val;
 dmuKey* form;
 dmuKey* unit;
} *dmuVectorKey;

/*  The AliasKeyList object supports the MTYPE, MFORM convention */

typedef struct dmuAliasKeyList_s {
 long n;
 char** desc;  /* Key desc index for MTYPE */
 AliasList names;
 AliasList cpts;
 MixArray etypes;
} *dmuAliasKeyList;

char* dmu_extract_filename( const char* vspec );

/* Key lists */
long dmu_keylist_n( const dmuKeyList keylist );
long dmu_keylist_check_index( const dmuKeyList keylist, const char* index_name, const char* name );
logical dmu_key_match( const char* name, const char* pattern );
void dmu_key_typed_value( dmuKey* ftkey, dmDataType type, void* value );
dmuKeyList dmu_keylist_alloc( void );
void dmu_keylist_free( dmuKeyList list );
void dmu_keylist_print( const dmuKeyList keylist,  void (*NativeTypeName)( dmuKeyType, char*, long ) );
long dmu_keyword_max_index( dmuKeyList keylist, char* pattern );
dmuKey* dmu_keylist_read_value_c( dmuKeyList keylist, char* name, char* value, long maxlen );
dmuKey* dmu_keylist_read_typed_value( dmuKeyList keylist, char* name, dmDataType type, void* value );
long dmu_keylist_check_index( const dmuKeyList keylist, const char* index_name, const char* name );
void dmu_key_free( dmuKey* ftkey );
dmuKey* dmu_key_alloc( void );
long dmu_key_insert( dmuKeyList keylist, dmuKey* ftkey, long posp );
void dmu_key_read_value_c(dmuKey *pKey, long maxLen, char *value);

dmuKey* dmu_key_find( dmuKeyList keylist, char* name );
void dmu_key_set( dmuKey* ftkey, const dmuKeyType fktype, const dmDataType datatype, const char* name, const void* value, const char* unit, const char* desc, const dmBool mask );
dmBool dmu_key_delete(  dmuKeyList keylist, dmuKey* ftkey );
char* dmu_key_value_c( dmuKey* ftkey );
void* dmu_key_value( dmuKey* ftkey );

void dmu_pixel_from_pixelno( long PixelNo, long* axes, long naxes, long* Pixel );
long dmu_pixelno_from_pixel( long* Pixel, long* axes, long naxes );
void dmu_vector_key_init( dmuVectorKey v );
MixArray dmu_vector_key_make_list( dmuKeyList keylist, char* name_tag, char* form_tag, char* unit_tag, char* val_tag, long tagsize );

dmuAliasKeyList dmu_alias_key_make_list( dmuKeyList keylist, char* name_tag, char* form_tag, char* etype_tag, char* ename_tag, long tagsize );
void dmu_alias_key_list_free( dmuAliasKeyList mlist );

/* Element types */
void dmu_etype_name( dmElementType etype, char* name, long maxlen );
long dmu_etype_dim( dmElementType etype );
dmElementType dmu_parse_etype_code( char* code );


void dmu_collapse_ranges_d( double** dmin, double** dmax, long* nvals );
void dmu_union_d( double xmin, double xmax, double* amin, double* amax, long* nvals );
void dmu_intersect_d( double xmin, double xmax, double* dmin, double* dmax, double* rmin, double* rmax,  long* nvals  );
void dmu_intersect_array_d( double* xmin, double* xmax, long n, double** dmin, double** dmax, long* nvals, long* nmax );

void dmu_block_type_name( const dmBlockType type, char* name );

char* dmu_parse_name_value( char* buf, char* name, long maxlen, char* op );
char*  dmu_parse_token( char* ptr, char sep, char** result, long maxlen );
char*  dmu_parse_paren_token( char* ptr, char lsep, char rsep, char** result, long maxlen );

dmErrCode dmu_err_set_code( long code, char* text, long fcode );
dmErrCode dmu_err_get( void );
dmErrCode dmu_err_set( long code, char* text );
void dmu_err_clear( void );
void dmu_err_disp_select_msgs( long lErrCode);
void dmuSetVerbose( long level );


long dmu_count_commas( const char* buf );
int dmu_lex_number( char* buf, double* dval );
char* dmu_make_range_string( void* mins, void* maxes, long n, int incnull, dmDataType type, dmBool cast, char* fmt );
void dmu_err_init( void );
void dmu_err_status( void );
void dmu_print_number( const double x, char* buf, const long maxlen );
char* dmu_split_token_chars( char* buf, char* name, long maxlen, char* op, char* opset );
char* dmu_split_token( char* buf, char* name, long maxlen, char* op );
char* dmu_string_delta( char* buf, long n );
void dmu_key_print( dmuKey* key, void (*NativeTypeName)( dmuKeyType, char*, long ), int opts, char* buf, long maxlen );
int dmu_bit_match( unsigned char* buf, long n, char* pattern );
dmBool dmu_bit_union( const char* bufA, const char* bufB, char* obuf );
dmBool dmu_bit_intersect( const char* bufA, const char* bufB, char* obuf );
int dmu_text_in_range( char* buf, char* start, char* end );
void dmu_format_number_d( const double val, char* buf );
void dmu_err_get_message( char* buf, long maxlen );
int dmu_parse_null( const char* val, dmDataType type, GenericType* null );
void dmu_map_bits( unsigned char* buf, long n, char* bits );
void dmu_print_pixel( long* pixel, long naxes, char* buf );
char* dmu_extract_filename_extension( const char* filename );
void dmu_key_dump( dmuKeyList list, char* tag );
dmDataType dmu_bt_tform_to_datatype( const char c );
void dx( dmuKeyList keylist ); /* debug */
#endif
