/*                                                                
**  Copyright (C) 2002-2004,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/**********************************************************************/
/*                                                                    */
/*  KERLIB                                                            */
/*                                                                    */
/*  This file has functions which are kernel related but don't access */
/*  the private kernel data directly                                  */
/*                                                                    */
/**********************************************************************/

#include "dmlibP.h"
#include "kertable.h"


/***************************************************************/
/*                                                             */
/* Function:   dmKernelSetCreate                               */
/*                                                             */
/* Description:                                                */
/*      Create kernel set                                      */
/*                                                             */
/*  Input parameters:                                          */
/*        const char *buf -                                    */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            char *buf                                        */
/*                                                             */
/***************************************************************/
char* dmKernelSetCreate( const char* buf )
{
  int status = 0;
  char rbuf[TEXT_SIZE];

  status = dmkKernelSet( buf, "W" );
  if ( status != dmSUCCESS )
    return NULL;
  else
  {
    dmKernelGetCreate( rbuf, TEXT_SIZE );
    return (char*)buf; /* Compat with old DM */
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmKernelSetCopy                                 */
/*                                                             */
/* Description:                                                */
/*      Set kernel as copy                                     */
/*                                                             */
/*  Input parameters:                                          */
/*        const char *buf -                                    */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            char *buf                                        */
/*                                                             */
/***************************************************************/
char* dmKernelSetCopy( const char* buf )
{
  int status = 0;
  char rbuf[TEXT_SIZE];

  status = dmkKernelSet( buf, "C" );
  if ( status != dmSUCCESS )
    return NULL;
  else
  {
    dmKernelGetCreate( rbuf, TEXT_SIZE );
    return (char*)buf;  /* Compat with old DM */
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmKernelSet                                     */
/*                                                             */
/* Description:                                                */
/*       Set kernel type                                       */
/*                                                             */
/*  Input parameters:                                          */
/*        const char *buf -                                    */
/*        const char *kerbuf                                   */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode dmSUCCESS or dmFAILURE                 */
/*                                                             */
/***************************************************************/
dmErrCode dmkKernelSet( const char* buf, const char* kerType )
{
  long nker  = 0;
  long kerId = 0;
  dmKernel kernel = NULL;

  nker = dmkGetNoKernels(); 
  /* Special case */
  if ( cs_uceq( buf, "DEFAULT" ))
  {
    kerId = 1;
    dmkSetTypedKernel( kerId, kerType );
    return dmSUCCESS;
  }

  for  ( kerId = 1; kerId <= nker; kerId++ )
  {
    kernel = dmkKernelNo( kerId );
    if ( cs_uceq( buf, kernel->name ) )
    {
      dmkSetTypedKernel( kerId, kerType );
      return dmSUCCESS;
    }
  }

  sprintf( dm_errbuf, "[dmKernelSetCreate]: Dont know kernel %s", buf );
  return dmu_err_set( DM_ERR_NO_KERNEL, dm_errbuf );
}


/***************************************************************/
/*                                                             */
/* Function:   dmKernelGetCreate                               */
/*                                                             */
/* Description:                                                */
/*      Return kernel name                                     */
/*                                                             */
/*  Input parameters:                                          */
/*        const long  maxlen -  buffer length                  */
/*                                                             */
/*  Output parameters:                                         */
/*              char *buf - name buffer                        */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode dmSUCCESS or dmFAILURE                 */
/*                                                             */
/***************************************************************/
dmErrCode dmKernelGetCreate( char* buf, const long maxlen )
{
  dmkGetKernelName( dmkGetTypedKernel("W"), buf, maxlen );
  return dmu_err_get();
}


/***************************************************************/
/*                                                             */
/* Function:   dmKernelGetCopy                                 */
/*                                                             */
/* Description:                                                */
/*     Return kernel name copy                                 */
/*                                                             */
/*  Input parameters:                                          */
/*                    none                                     */
/*                                                             */
/*  Output parameters:                                         */
/*              char *buf    - name buffer                     */
/*        const long  maxlen - buffer length                   */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode dmSUCCESS or dmFAILURE                 */
/*                                                             */
/***************************************************************/
dmErrCode dmKernelGetCopy( char* buf, const long maxlen )
{
  dmkGetKernelName( dmkGetTypedKernel("C"), buf, maxlen );
  return dmu_err_get();
}


/***************************************************************/
/*                                                             */
/* Function:   dmGetKernelName                                 */
/*                                                             */
/* Description:                                                */
/*   Get the name of a kernel by number                        */
/*                                                             */
/*  Input parameters:                                          */
/*        const long  kerId  -                                 */
/*        const long  maxlen -                                 */
/*                                                             */
/*  Output parameters:                                         */
/*              char *name -                                   */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmkGetKernelName( const long kerId, char* name, const long maxlen )
{
  dmKernel kernel = dmkKernelNo( kerId );
  if ( kernel )
    strncpy( name, kernel->name, maxlen );
  else
   strcpy( name, "" );
}


/***************************************************************/
/*                                                             */
/* Function:   dmk_init                                        */
/*                                                             */
/* Description:                                                */
/*    Call initialization for all kernels                      */
/*                                                             */
/*  Input parameters:                                          */
/*                 none                                        */
/*                                                             */
/*  Output parameters:                                         */
/*                 none                                        */
/*                                                             */
/*  Returns:                                                   */
/*            void                                             */
/*                                                             */
/***************************************************************/
void dmk_init( void )
{
  dmKernel kernel;
  long  kermax = dmkGetNoKernels();
  long  kerId = 0;

  for ( kerId = 1; kerId <= kermax; kerId++ )
  {
    kernel = dmkKernelNo( kerId );
    kernel->Init();
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmk_dataset_open                                */
/*                                                             */
/* Description:                                                */
/*   Find which kernel a dataset belongs to                    */
/*   Try the kernels in numerical order.                       */
/*   Return zero if no valid kernel is found.                  */
/*   The order of kernels is important! The first one which    */
/*   succeeds will be used.                                    */
/*                                                             */
/*  Input parameters:                                          */
/*          const char *dsname - dataset name                  */
/*          const char *hint   - dataset hints                 */
/*          const char *mode   - access mode                   */
/*                                                             */
/*  Output parameters:                                         */
/*                 none                                        */
/*                                                             */
/*  Returns:                                                   */
/*            dmDataset *dataset                               */
/*                                                             */
/***************************************************************/
dmDataset* dmk_dataset_open( const char* dsname  /* (i) Dataset name */,
			     const char* hint    /* (i) dataset hints */,
			     const char* mode    /* (i) Access mode */  )
{
  dmDataset *ds = NULL;
  dmKernel   kernel;
  dmErrCode  status = dmSUCCESS;
  long       kerId  = 0;
  long       kermin = 0;
  long       kermax = 0;

  /* For create or copy, get specified kernel. */
  /* Otherwise loop through all kernels.       */
  kerId = dmkGetTypedKernel( mode );

  if ( kerId > 0 )
  {
    kermin = kerId;
    kermax = kerId;
  }
  else
  {
    kermin = 1;
    kermax = dmkGetNoKernels();
  }

  for ( kerId = kermin; kerId <= kermax; kerId++ )
  {
    kernel = dmkKernelNo( kerId );
    ds = kernel->DatasetOpen( dsname, hint, mode );

    if ( ds )
    {
      ds->kernel = kernel;
      return ds;
    }
  }

  if ( !ds ) 
  {
    status = dmu_err_get();
    if ( status != DM_ERR_NO_FILE )
    {
      sprintf( dm_errbuf, "No kernel found for dataset %s", dsname );
      dmu_err_set( DM_ERR_NO_KERNEL, dm_errbuf );
    }
  }
  return NULL;
}

/***************************************************************/
/*                                                             */
/* Function:   dmKernelGetList                                 */
/*                                                             */
/* Description:                                                */
/*     Return list of kernels installed                        */
/*                                                             */
/*  Input parameters:                                          */
/*                none                                         */
/*                                                             */
/*  Output parameters:                                         */
/*                char ***kernelNames - list of kernels        */
/*                long   *numKer      - number of kernels      */
/*                                                             */
/*  Returns:                                                   */
/*            dmErrCode dmSUCCESS or dmFAILURE                 */
/*                                                             */
/***************************************************************/
dmErrCode dmKernelGetList( char*** kernelNames, long* numKer )
{
  char **names = NULL;
  long   nker  = 0;
  long   kerId = 0;
  dmKernel kernel;

  nker = dmkGetNoKernels();
  names = (char**)calloc( nker, UT_SZ_P );

  for ( kerId = 1; kerId <= nker; kerId++ )
  {
    kernel = dmkKernelNo( kerId );
    names[kerId-1] = cs_dup( kernel->name );
  }

  *kernelNames = names;
  *numKer = nker;
  return dmu_err_get();
}


/***************************************************************/
/*                                                             */
/* Function:   dmk_kernel_eq                                   */
/*                                                             */
/* Description:                                                */
/*    Determine if two datasets have same kernel               */
/*                                                             */
/*  Input parameters:                                          */
/*                dmDataset  *ds1                              */
/*                dmDataset  *ds2                              */
/*                                                             */
/*  Output parameters:                                         */
/*                none                                         */
/*                                                             */
/*  Returns:                                                   */
/*            int same_kernel   1=yes, 0=no                    */
/*                                                             */
/***************************************************************/
int dmk_kernel_eq( dmDataset* ds1, dmDataset* ds2 )
{
  int      same_kernel = 0;
  dmKernel kernel1;
  dmKernel kernel2;

  kernel1 = ds1->kernel;

  if ( ds2 )
    kernel2 = ds2->kernel;
  else
    kernel2 = dmkKernelNo( dmkGetTypedKernel( "C" ));

  same_kernel =  ( kernel1 == kernel2 || cs_eq( kernel1->name, kernel2->name ));
  return same_kernel;
}


/***************************************************************/
/*                                                             */
/* Function:   dmk_block_kernel                                */
/*                                                             */
/* Description:                                                */
/*    Get kernel of a block                                    */
/*                                                             */
/*  Input parameters:                                          */
/*                dmBlock    *block                            */
/*                                                             */
/*  Output parameters:                                         */
/*                none                                         */
/*                                                             */
/*  Returns:                                                   */
/*            dmKernel kernel                                  */
/*                                                             */
/***************************************************************/
dmKernel dmk_block_kernel( dmBlock* block )
{
  dmKernel   kernel = NULL;
  dmDataset *ds     = NULL;

  ds = block->ds;

  if ( ds )
    kernel = ds->kernel;
  return kernel;
}
