/*                                                                
**  Copyright (C) 2002-2004,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*
 *  kermap.h
 *
 *  This file contains the private data for the ASCDM2 kernels.
 *  To add a new function to the kernel interface:
 *  (1)  Code the kernel-specific function for each kernel 
 *       (e.g. ftAccessDataset) and put its prototype in the
 *       kernel interface file (e.g. ftkernel.h).
 *
 *  (2)  Add the name of the kernel-specific function to the
 *       initialization list in KernelTable[] below, this file.
 *
 *  (3)  Add the generic prototype for the function (e.g. AccessDataset)
 *       to the KernelStruct structure definition in kertable.h
 *
 *  To add a new kernel to the interface:
 *
 *  (1)  Code each function in the kernel table for the new kernel,
 *       and put their prototype in an analog to ftkernel.h
 *
 *  (2)  Add an extra wrapper to the #ifndef..#define... #endif in Part A below,
 *       to ensure that if no kernels are initially chosen, all of them
 *       get chosen.
 *
 *  (3)  Include the new kernel interface in Part B below,
 *           #ifdef MY_KERNEL
 *           #include "../mykernel/mykernel.h"
 *           #endif
 *
 *  (4)  Add the list of new kernel functions to Part C below.
 *       The order of the kernels in Part C determines the mapping
 *       of kernel ID to kernel, and thus the order in which 
 *       a dataset open tries each kernel when opening a file.
 * 
 */


#ifndef KERMAP_H
#define KERMAP_H


/*
 *  [Part A]
 *  If no kernels defined in configure, define all of them
 */
#ifndef FITS_KERNEL
#ifndef IRAF_KERNEL
#ifndef TEXT_KERNEL

#define FITS_KERNEL
#define IRAF_KERNEL
#define TEXT_KERNEL
#endif
#endif
#endif

#include "kertable.h"

/* 
 * [Part B] Include defined kernels
 */


#ifdef FITS_KERNEL
#include "../ftkernel/ftkernel.h"
#endif

#ifdef IRAF_KERNEL
#include "../iqkernel/iqkernel.h"
#endif

#ifdef TEXT_KERNEL
#include "../afkernel/afkernel.h"
#endif


/*
 *  [Part C] Kernel table initialization
 */

struct KernelStruct KernelTable[] = {

#ifdef FITS_KERNEL
 {
  "FITS",
  ftInit,
  ftDatasetAccess,
  ftDatasetOpen,
  ftDatasetCreate,
  ftDatasetClose,
  ftDatasetDelete,
  ftDatasetPrint,
  ftBlockCreate,
  ftBlockOpen, 
  ftBlockClose,
  ftBlockDelete,
  ftBlockFlush,
  ftImageSetAxes,
  ftBlockSetPref,
  ftBlockUpdateFilter,
  ftKeyWrite,
  ftKeyDelete,
  ftReadComment,
  ftWriteComment,
  ftColCreate,
  ftColSetRange,
  ftColSetNull,
  ftColSetBin,
  ftColRead,
  ftColWrite,
  NULL,
  NULL,
  ftColFlush,
  ftDescriptorCopy,
  ftDescriptorFree,
  ftSubspaceDescriptorWrite,
  ftSubspaceTableWrite,
  ftCoordWrite,
  ftBufferSize,
  ftSetHint,
  ftGetKernelKey,
  ftGetNoKernelKeys
 },
#endif

#ifdef IRAF_KERNEL
{
  "IRAF",
  iqInit,
  iqDatasetAccess,
  iqDatasetOpen,
  iqDatasetCreate,
  iqDatasetClose,
  iqDatasetDelete,
  iqDatasetPrint,
  iqBlockCreate,
  iqBlockOpen, 
  iqBlockClose,
  iqBlockDelete,
  iqBlockFlush,
  iqImageSetAxes,
  iqBlockSetPref,
  iqBlockUpdateFilter,
  iqKeyWrite,
  iqKeyDelete,
  iqReadComment,
  iqWriteComment,
  iqColCreate,
  iqColSetRange,
  iqColSetNull,
  iqColSetBin,
  iqColRead,
  iqColWrite,
  NULL,
  NULL,
  iqColFlush,
  NULL,
  iqDescriptorFree,
  iqSubspaceDescriptorWrite,
  iqSubspaceTableWrite,
  iqCoordWrite,
  iqBufferSize,
  iqSetHint,
  iqGetKernelKey,
  iqGetNoKernelKeys
},
#endif

#ifdef TEXT_KERNEL
{
  "TEXT",
  afInit,
  afDatasetAccess,
  afDatasetOpen,
  afDatasetCreate,
  afDatasetClose,
  afDatasetDelete,
  afDatasetPrint,
  afBlockCreate,
  afBlockOpen, 
  afBlockClose,
  afBlockDelete,
  afBlockFlush,
  afImageSetAxes,
  afBlockSetPref,
  afBlockUpdateFilter,
  afKeyWrite,
  afKeyDelete,
  afReadComment,
  afWriteComment,
  afColCreate,
  afColSetRange,
  afColSetNull,
  afColSetBin,
  NULL,
  afColWrite,
  afReadRows,
  afWriteRows,
  afColFlush,
  NULL,
  afDescriptorFree,
  afSubspaceDescriptorWrite,
  afSubspaceTableWrite,
  afCoordWrite,
  afBufferSize,
  afSetHint,
  afGetKernelKey,
  afGetNoKernelKeys
}
#endif
};

#define KernelN ((long)(sizeof(KernelTable) / sizeof( struct KernelStruct) ))



#endif  /* KERMAP_H */
