/*                                                                
**  Copyright (C) 1996,1999-2000,2003-2004,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   kertable.c                                                  */
/*                                                               */
/*   Functions that are kernel related and access the private    */
/*   kernel data directly                                        */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmulib.h"
#include "kertable.h"
#include "kermap.h"



/* We initialize all kernel pointers to be the first kernel in the table */

static long CurrentKernel  = 1;   /* Current kernel */
static long CreationKernel = 1;   /* Current creation kernel */
static long CopyKernel     = 1;   /* Current copy kernel */



/***************************************************************/
/*                                                             */
/* Function:   dmkGetNoKernels                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  none                                       */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                 long  KernelN                               */
/*                                                             */
/***************************************************************/
long dmkGetNoKernels( void )
{
  return KernelN;
}



/***************************************************************/
/*                                                             */
/* Function:   dmkKernelNo                                     */
/*                                                             */
/* Description:                                                */
/*            Look up a kernel by number                       */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  const long kerId                           */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                 dmKernel kernel                             */
/*                                                             */
/***************************************************************/
dmKernel dmkKernelNo( const long kerId )
{
  dmKernel kernel;

  if ( kerId <= 0 || kerId > KernelN )
    kernel = NULL;
  else
    kernel = &KernelTable[kerId-1];
 
  return kernel;
}



/***************************************************************/
/*                                                             */
/* Function:   dmkKernel                                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    none                                     */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                 dmKernel kernel                             */
/*                                                             */
/***************************************************************/
dmKernel dmkKernel( void )
{
  return dmkKernelNo( CurrentKernel );
}



/***************************************************************/
/*                                                             */
/* Function:   dmkSetKernel                                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  const long kerId                           */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                 nothing                                     */
/*                                                             */
/***************************************************************/
void dmkSetKernel( const long kerId )
{
  CurrentKernel = kerId;
}



/***************************************************************/
/*                                                             */
/* Function:   dmkGettypedKernel                               */
/*                                                             */
/* Description:                                                */
/*            Find out the name of the creation kernel         */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  const char *kerType                        */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                   long  kernel                              */
/*                                                             */
/***************************************************************/
long dmkGetTypedKernel( const char* kerType )
{
  if ( cs_uceq( kerType, "W" ) )
    return CreationKernel;
  else
    if ( cs_uceq( kerType, "C" ) )
      return CopyKernel;
    else
      return 0;
}



/***************************************************************/
/*                                                             */
/* Function:   dmkGetNamedKernel                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  const char *name                           */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                  long kerId                                 */
/*                                                             */
/***************************************************************/
long dmkGetNamedKernel( const char* name )
{
  dmKernel kernel;
  long     offset = 0;
  long     kerId  = 0;

  for ( offset = 0; offset < KernelN; offset++ )
  {
    kernel = &KernelTable[offset];

    if ( cs_uceq_siz( kernel->name, name, strlen( kernel->name )))
      kerId = offset+1;
  }

  return kerId;
}



/***************************************************************/
/*                                                             */
/* Function:   dmkSetTypedKerneln                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                  const long  kerId   -                      */
/*                  const char *kerType -                      */
/*                                                             */
/*  Output parameters:                                         */
/*                    none                                     */
/*                                                             */
/*  Returns:                                                   */
/*                    nothing                                  */
/*                                                             */
/***************************************************************/
void dmkSetTypedKernel( const long kerId, const char* kerType )
{
  if ( cs_uceq( kerType, "W" ) )
    CreationKernel = kerId;
  else
    if ( cs_uceq( kerType, "C" ) )
      CopyKernel = kerId;
}



/***************************************************************/
/*                                                             */
/* Function:   dmkKernelSetOption                              */
/*                                                             */
/* Description:                                                */
/*   Routine to set general Datamodel options                  */
/*   Assumes a single option input.                            */
/*                                                             */
/*   Continue to support null block for back compatibility.    */
/*                                                             */
/*  Input parameters:                                          */
/*    dmBlock    *block   - block pointer                      */
/*    const char *option  - input option string                */
/*                                                             */
/*  Output parameters:                                         */
/*    none                                                     */
/*                                                             */
/*  Returns:                                                   */
/*    dmErrCode                                                */
/*                                                             */
/***************************************************************/
dmErrCode dmkKernelSetOption( dmBlock* block, const char* option )
{
  char    *value = NULL;
  char    *ptr   = NULL;
  char     name[CARD_SIZE+1] = "";
  long     kerId  = 0;
  long     n      = 0;
  dmKernel kernel;

  dmErrCode status = dmSUCCESS;
  dmBool   matched = dmFALSE;

  ptr = name;

  /* Split token into name-value pair */
  value = dmu_parse_token( (char*)option, '=', &ptr, CARD_SIZE );

  if ( !block )
  {
    /* Get the number of kernels available */
    n = dmkGetNoKernels();

    /* Send option to each kernel for them to attempt to use */
    for (kerId = 1; kerId <= n;  kerId++)
    {
      kernel = dmkKernelNo( kerId );
      
      if ( kernel )
      {
	status = kernel->SetHint( block, name, value );
	if ( status == dmSUCCESS ){ matched = dmTRUE; }
      }
    }
  }
  else
  {
    kernel = dmk_block_kernel( block );
    status = kernel->SetHint( block, name, value );
    if ( status == dmSUCCESS ){ matched = dmTRUE; }
  }

  if ( !matched )
  {
    /* This option was not recognized by any of the available kernels.. */
    sprintf( dm_errbuf, "Unrecognized option %s  ignored.", option);
    dmu_err_set( DM_WARN_PARSE_OPTION, dm_errbuf );
    status = dmFAILURE;
  }
  else
  {
    status = dmSUCCESS;
  }

  /* Print out messages for unmatched options   */
  /* NOTE: This section may move up some levels */
  if (dmu_err_get() != dmSUCCESS )
  {
    /* The function below prints select error messages on the error stack */
   dmu_err_disp_select_msgs( DM_WARN_PARSE_OPTION );
  }

  /* clear error stack */
  dmu_err_clear();

  return status;
}


