/*                                                                
**  Copyright (C) 1997-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   dmbuffer.c                                                  */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmlibP.h"
#include <limits.h>

#define DEBUG 0
#define XDEBUG 1
#define CAST_SIZE 1024

void dmi_block_apply_image_filter( dmDescriptor* image , long* v_llpix, long* v_urpix );
void dmp_buffer_copy_image_data( dmDescriptor* dd1, dmDescriptor* dd2 );
long dmp_buffer_write( dmBlock* block, dmDescriptor* dd, long vrowno, long nvrows, void* values, long item_offset, long item_nbytes, long vrow_out_offset );
void dmp_image_read_segment( dmDescriptor* parent_image, long* tmin, long* tmax, dmDataType type, void* image_buffer, long image_buffer_size );
void dmp_filtered_image_read( dmDescriptor* image, dmDescriptor* raw_image, long* llpix, long* urpix, long* v_llpix, long* v_urpix, void* image_buffer, long image_buffer_size );
long dmp_buffer_read_seg( dmBlock* block, dmBuffer* buffer,  long vrowno );
int dmp_buffer_read_filter_cols( dmBlock* block, dmBuffer* buffer );
int dmp_buffer_read_main_cols( dmBlock* block, dmBuffer* buffer, long rrow_offset, long rrowno1, long rrowno2 );
void dmp_buffer_read_other_cols( dmBuffer* buffer, long rrow_offset, long rrowno1, long rrowno2 );
long dmp_buffer_set_data( dmBlock* block, dmDescriptor* dd, long vrowno, long nvrows, long item_offset, long item_nbytes, void* data );
long dmp_buffer_read( dmBlock* block, dmDescriptor* dd, long rowno, long nrows, void* values, long item1, long item2, long rowoff,  long* ll, long* ur, long* dp  );
void dmp_buffer_nullify_image_space_gen( void* data, void* nullval, dmDataType type, long* axes, long naxes, long* min_offsets, long* max_offsets );
void dmp_buffer_nullify_image_space_2d( void* data, void* nullval, dmDataType type, long* axes, long* min_offsets, long* max_offsets );

int  dmp_filtered_image_row_read(dmDescriptor *image, char *cast_buffer, long *llpix, long *urpix, long rrowno1, long rrowno2 );

long dmp_buffer_subarray_extract(dmDescriptor *dd, long *llpix, long *uprix,  char *inptr, char *outptr );

/**************************************************************************************
   We have two design options: a column-ordered buffer and a row-ordered buffer:
   we will support both? For now, col-ordered.
  
  Col-ordered 
   [
      [ PHA 1.................n] 
      [ X   1.................n]
      [ Y   1.................n]
   ]
 
   Row-ordered
   [
     1  PHA X Y..
     2  PHA X Y...
     ...
     n  PHA X Y...
   ]
FTK: Note that CFITSIO uses 40 2880-byte buffers.
We should read 100 kb at a time into the raw buffer.

Should the kernel raw read be row-oriented, col-oriented, or both?
 row-oriented: more work at kernel layer? more efficient?
 col-oriented: could implement by a row-oriented buffer fill followed by
               col-oriented flag checking. But still problems for doing GetRow.
 Perhaps there should be no 'values' arg in the kernel->ColRead, it should
 just be an instruction to make sure the raw buffer has the right values filled in.

 If the data format is col-oriented, and we are filtering with [cols foo],
 it may be much more efficient to read only [foo] to the buffer.
 
 If the data format is row-oriented, we don't want to parse the boring columns
 but we may want to move the useful columns to the buffer on a per-row basis.

 If it is row-oriented and not filtered, it's much better to do things on
 a per-row basis and work only row-oriented.

 We'll start with a col-oriented raw buffer.

   CFITSIO ->  RAW BUF  -> FILTERED VBUF

 I had considered using a single VBUF and arrays of flags for each filter
 in a stack of filters, but that won't work well. Consider the case
 of  evt.fits[ccd_id=*] in a file that is already ccd_id ordered.
 The first 10 v-rows of ccd_id=1 will be at a very different raw-row range from
 the first 10 v-rows of ccd_id=2, so we can't use the same raw-buffer. 
 Therefore we do need different buffers for them. But for cases where 
 we want to go back and forth in a buffer, it does make sense to have
 an explicitly separate v-buf and not just index the raw-buf. So I think
 we should have a single raw-buf and multiple separate v-bufs, with the
 v-buf avoided only in the case of an unfiltered file.

 OK, now we read some stuff from a column. Really to do this without 
 paging the FITS buffers, we have two choices:
  - 1: Fill the buffers inside the kernel
  - 2: Read only the number of rows suitable for the FITS buffer
 In either case for efficiency we must make 1 cfitsio call per 
 column per CFITSIO buffer.

 The complication is that the size of the CFITSIO buffer covers
 all open FITS files in the process and not just the current file.


 ****************************************************************************/


/***************************************************************/
/*                                                             */
/* Function:   dmp_buffer_flush                                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock  *block                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmp_buffer_flush( dmBlock* block )
{
  dmDescriptor *dd     = NULL;
  dmBuffer     *buffer = NULL; 
  dmDataset    *ds     = NULL;
  dmKernel      kernel;

  char *values      = NULL;
  long *llpix       = NULL;
  long *urpix       = NULL;
  long *dpix        = NULL; 
  long  col_offset  = 0;
  long  rrowno1     = 0;
  long  rrowno2     = 0;
  long  rrow_offset = 0;


 buffer = block->buffer;

 ds = block->ds;

 if ( !buffer ) return;
 if ( !buffer->update ) return;
 buffer->update = UT_FALSE;
/* Skip for read-only */
 if ( !dmi_dataset_mode( ds ) ) return;
 kernel = ds->kernel;

/*
 * If we are writeable, there is no filtering. We only need to flush up to
 * the last row written in the buffer.
 */
 rrowno1 = buffer->rrow1;
 rrowno2 = block->nrows < buffer->rrow2 ? block->nrows : buffer->rrow2;

#if DEBUG
printf( "FLU %ld:%ld\n", buffer->rrow1, buffer->lastrow );
#endif
 if ( buffer->lastrow > 0 )
 {

  if ( buffer->lastrow < buffer->rrow1 )
  {
   fprintf( stderr, "DM INTERNAL ERROR: BUFFER %ld:%ld BUT LASTROW=%ld\n", buffer->rrow1, buffer->rrow2, buffer->lastrow );
   exit(1);
  } 

  /* A kernel may be specialized for writing either columns or rows */
  if ( kernel->WriteRows )
  {
    /* NOTE: This is not a statement of preference.     */
    /* The Ascii kernel has both WriteRows and ColWrite */
    /* defined.. with WriteRows more efficient.         */
    /* The FITS kernel has only ColWrite defined        */
    kernel->WriteRows( block, buffer, rrowno1, rrowno2 );
  }
  else
  {
    if ( kernel->ColWrite )
    {
      for ( col_offset = 0; col_offset < buffer->ncols; col_offset++ )
      {
	dd = buffer->cols[col_offset];

	/* Skip unbuffered columns */
	if ( ( dd->flag & DM_FLAG_BINNED ) ||
	     ( dd->flag & DM_FLAG_UNBUF )  ||
	     ( dd->flag & DM_FLAG_AUX_COLUMN ) )
	  continue;
	  
	values = (char*)buffer->data + dmi_buffer_offset( buffer, col_offset, rrow_offset );

	/* Descriptors here are always raw cols */
	if ( dd && values )
	  kernel->ColWrite( dd, llpix, urpix, dpix, rrowno1, rrowno2, values );
      }
    }
  }

  /* Update nrows */
  if ( rrowno2 > block->nrows )
    block->nrows = rrowno2;
  if ( rrowno2 > block->nvrows)
    block->nvrows = rrowno2;

 }

 return;
}



/***************************************************************/
/*                                                             */
/* Function:   dmp_buffer_set_cast_data                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor dd -                        */
/*                    void        *val                         */
/*                    dmDataType   type                        */
/*                    long         vrowno                      */
/*                    long         nrows                       */
/*                    long         start                       */
/*                    long         nitems                      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long   nvrows_got                                */
/*                                                             */
/***************************************************************/
long dmp_buffer_set_cast_data( dmDescriptor* dd, void* val, dmDataType type, 
                               long vrowno, long nvrows, long start, long nitems )
{
 dmDescriptor *vector = dd;
 dmBlock      *block;
 dmDataType    dd_type;  /* Actual type of descriptor */

 char  cast_static[CAST_SIZE];   /* Avoid calloc for small cases */
 void *cast_buffer   = NULL;     /* Dynamic buffer for casting arrays */

 int   allocated     = 0;        /* Are we using the dynamic buffer? */
 long  size          = 0;        /* Bytes per value, descriptor internal */
 long  bytes_per_row = 0;        /* Number of bytes internal, per row */
 long  nvrows_got    = 0;        /* Number of filtered rows returned */
 long  nbytes        = 0;        /* Total number of bytes transferred in call */
 long  nvals         = 0;        /* Total number of values transferred */
 long  dim           = 0;        /* DM vector dimension */
 long  cptno         = 0;


 if ( nvrows == 0 || !dd ) return nvrows_got;

 
 if ( nitems <= 0 ) return nvrows_got;


 cptno = dd->cptno;
 if ( cptno == 0 )
  cptno = 1;


 block = (dmBlock*)dd->block;
 dd_type = dd->type;
 dim = dd->dim;
 size = dmi_descriptor_get_bytes( dd );

 allocated = 0;
 bytes_per_row = size * nitems * dim;
 nbytes = nvrows * bytes_per_row;
 if ( dd_type != type )
 {
  if ( nbytes > CAST_SIZE )
  {
   cast_buffer = calloc( nbytes, UT_SZ_C );
   allocated =1;
  }
  else 
  {
   cast_buffer = cast_static; 
   memset( cast_buffer, 0, nbytes );
  }
  nvals  = nvrows * nitems * dim;
  ar_mix_cast_array( val, type, dd_type, cast_buffer, nvals );
 }
 else
  cast_buffer = val;

 nvrows_got = dmp_buffer_set_data( block, vector, vrowno, nvrows, start, nitems, cast_buffer );
 if ( allocated )
   free( cast_buffer );

 return nvrows_got;
}


/* Return several rows of data from virtual col 
   filterno is the entry in the filter stack
   colno is the virt col no
   rowno is the start virt row
   nrwos is the number of rows returned
  item1:item2 = start:start+nitems-1 gives the array entry numbers

   Handle reading a component: set the byte offset and the size, and read
   the parent vector.

 */


/***************************************************************/
/*                                                             */
/* Function:   dmp_buffer_get_cast_data                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor dd -                        */
/*                    void        *val                         */
/*                    dmDataType   type                        */
/*                    long         vrowno                      */
/*                    long         nrows                       */
/*                    long         start_item                  */
/*                    long         nitems                      */
/*                    long        *llpix                       */
/*                    long        *urpix                       */
/*                    long        *dpix                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long   nvrows_got                                */
/*                                                             */
/***************************************************************/
long dmp_buffer_get_cast_data( dmDescriptor* dd, void* val, dmDataType type, 
                               long vrowno, long nvrows, long start_item, long nitems,
                               long* llpix, long* urpix, long* dpix )
{
  char cast_static[CAST_SIZE];   /* Avoid calloc for small cases */
  dmDataType     dd_type;           /* Actual type of descriptor */
  dmBlock       *block       = NULL;
  dmDescriptor  *vector      = NULL;
  void          *cast_buffer = NULL; /* Dynamic buffer for casting arrays */

  long size          = 0;  /* Bytes per value, descriptor internal */
  long bytes_per_row = 0;  /* Number of bytes internal, per row */
  long nvrows_got    = 0;  /* Number of filtered rows returned */
  long nbytes        = 0;  /* Total number of bytes transferred in call */
  long nvals         = 0;  /* Total number of values transferred */
  long dim           = 0;  /* DM vector dimension */
  long item_byte_offset = 0; 
  long cptno         = 0;
  long col_offset    = 0;

  int allocated      = 0;  /* Are we using the dynamic buffer? */


  if ( nvrows == 0 || nitems == 0 || !dd )
    return nvrows_got;

  block   = (dmBlock*)dd->block;
  dd_type = dd->type;
  dim     = dd->dim;
  cptno   = dd->cptno;
  col_offset = dd->bufno-1;

  /* Only vectorize if not in the buffer */
  if ( cptno > 0 && col_offset < 0 )
    vector = dd->parent;
  else
    vector = dd;

  if ( cptno == 0 )
    cptno = 1;

  size = dmi_descriptor_get_bytes( dd );
  allocated = 0;

  if ( cptno > 1 && dd->parent )
  {
    // Code change  Bug 9901 April 21, 2006
    // changed nitems to dd->size in calculation below

    /* need to check if parent is in buffer! */

    if ( dd->parent->bufno > 0 )
      item_byte_offset = ( start_item - 1 +  (cptno - 1 ) * dd->size )* size;
    else
      item_byte_offset = ( start_item - 1 ) * size;
  }
  else
     item_byte_offset = ( start_item - 1 ) * size;


  bytes_per_row = size * nitems * dim;

  nbytes = nvrows * bytes_per_row;


  if ( dd_type != type )
  {
    char* cptr = NULL;
    if ( nbytes > CAST_SIZE )
    {
      cptr = (char*)calloc( nbytes, UT_SZ_C );
      allocated =1;
    }
    else 
    {
      cptr = cast_static; 
      memset( cast_static, 0, nbytes );
    }

    /* MCD NOTE: string types need to be accessed as a char** so       */
    /* need to send address of memory start.. this probably does       */
    /* not work if getting more than one row (only cptr[0] has memory) */
    if ( dd_type == DT_CHAR )
    {
      cast_buffer = (void*)&cptr;
    }
    else
    {
      cast_buffer = cptr;
    }
  }
  else
    cast_buffer = val;

 nvrows_got = dmp_buffer_get_data( block, vector, vrowno, nvrows, item_byte_offset, bytes_per_row, llpix, urpix, dpix, cast_buffer );

 if ( dd_type != type && nvrows_got > 0  ) 
 {
   nbytes = nvrows_got * bytes_per_row;
   nvals  = nvrows_got * nitems * dim;
#if DEBUG
printf( "CAST %s %ld ITEMS FROM %p TO %p GOT %ld\n", dd->name, nbytes, cast_buffer, val,  nvrows_got );
#endif

   ar_mix_cast_array( cast_buffer, dd_type, type, val, nvals );

   if ( allocated )
     free( cast_buffer );
  }

  return nvrows_got;
}



/*
 * We don't know how many raw rows needed for n filtered rows.
 *  Two approaches:
 *   1) Ask for N raw rows of dd
        Get m filtered rows
        Copy m rows of dd to output
        Loop for new raw row buffer until have n filtered rows
     2) Ask for N raw rows of dd
        Get m filtered rows
        Copy m rows of all cols to filtered buffer
        Loop for new raw row buffer
        Copy m rows of filtered buffer dd to output
        Advantage: next column not paged
        Con: one extra copy
Is there an intermediate way?
      Ask for N raw rows of dd
      Get m filtered rows
      Is m >= n?
      If so, mark for direct copy
      Else, copy to filtered buffer
 This way, extra copy only when filtering is sparse.

 Also: cache index of filtered file in case we have to traverse it again.

 */

/***************************************************************/
/*                                                             */
/* Function:   dmp_buffer_get_data                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock      *block                      */
/*                    dmDescriptor *dd                         */
/*                    long          vrowno                     */
/*                    long          nvrows                     */
/*                    long          item_offset                */
/*                    long          item_nbytes                */
/*                    long         *llpix                      */
/*                    long         *ulpix                      */
/*                    long         *dpix                       */
/*                                                             */
/*  Output parameters:                                         */
/*                    void         *data                       */
/*                                                             */
/*  Returns:                                                   */
/*            long  nvrows_got                                 */
/*                                                             */
/***************************************************************/
long dmp_buffer_get_data( dmBlock* block, dmDescriptor* dd, long vrowno,
                          long nvrows, long item_offset, long item_nbytes,
                          long* llpix, long* urpix, long* dpix, void* data )
{
 long nvrows_got  = 0;       /* Total rows returned */
 long nvrows_left = nvrows;  /* Total rows yet to return */
 long nvrows_seg  = 1;       /* Rows returned in this segment; initialize to nonzero */
 long vrowno_seg  = 0;       /* Filtered row number at start of segment */
 long vrow_offset = 0;       /* Row offset into output buffer */



/* Here we need to loop until have nrows */
 vrowno_seg = vrowno;

 if ( dmverbose > 4 )
 {
  fprintf( stderr, "Buffer seek     VRN = %ld:%ld\n", vrowno, vrowno+nvrows - 1 );
 }
 while( nvrows_left > 0 && nvrows_seg > 0 )
 {
/* Here we need to copy the data to the output by vrow number */  

  nvrows_seg = dmp_buffer_read( block, dd, vrowno_seg, nvrows_left, data, item_offset, item_nbytes, vrow_offset,
    llpix, urpix, dpix );
  if ( nvrows_seg < 0 )
   return nvrows_seg;

  vrowno_seg  += nvrows_seg;
  nvrows_got  += nvrows_seg;
  nvrows_left -= nvrows_seg;
  vrow_offset = nvrows_got;  /* For scalar reads */
  if ( dmverbose > 4 )
  {
   fprintf( stderr, "Buffer segment: VRN = %ld:%ld Left=%ld\n", vrowno_seg-nvrows_seg, vrowno_seg - 1, nvrows_left );
  }
 }


#if DEBUG
printf( "DBGD %ld\n", nvrows_got );
#endif
 return nvrows_got;
}


/***************************************************************/
/*                                                             */
/* Function:   dmp_buffer_set_data                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock      *block                      */
/*                    dmDescriptor *dd                         */
/*                    long          vrowno                     */
/*                    long          nvrows                     */
/*                    long          start                      */
/*                    long          nitems                     */
/*                    void         *data                       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long  nvrows_got                                 */
/*                                                             */
/***************************************************************/
long dmp_buffer_set_data( dmBlock* block, dmDescriptor* dd, long vrowno,
                          long nvrows, long start, long nitems, void* data )
{
 long nvrows_got  = 0;
 long vrow_offset = 0;       /* Row offset into output buffer */
 long nvrows_left = nvrows;
 long nvrows_seg  = 1;
 long vrowno_seg  = vrowno;
 
 while( nvrows_left > 0 && nvrows_seg > 0 )
 {
/* Here we need to copy the data to the output by vrow number */  

  nvrows_seg = dmp_buffer_write( block, dd, vrowno, nvrows_left, data, start, nitems, vrow_offset );

  vrowno_seg  += nvrows_seg;
  nvrows_got  += nvrows_seg;
  nvrows_left -= nvrows_seg;
  vrow_offset = nvrows_got; 
  if ( dmverbose > 1 )
  {
   fprintf( stderr, "Buffer segment: VRN = %ld:%ld Left=%ld\n", vrowno_seg-nvrows_seg, vrowno_seg - 1, nvrows_left );
  }
 }

 return nvrows_got;
}




/***************************************************************/
/*                                                             */
/* Function:   dmp_buffer_write                                */
/*                                                             */
/* Description:                                                */
/*    Try to write some rows to the buffer from                */
/*      values[vrow_out_offset]                                */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock      *block                      */
/*                    dmDescriptor *dd                         */
/*                    long          vrowno                     */
/*                    long          nvrows                     */
/*                    void         *values                     */
/*                    long          start                      */
/*                    long          nitems                     */
/*                    long          vrow_out_offset            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long  nvrow_seg                                  */
/*                                                             */
/***************************************************************/
long dmp_buffer_write( dmBlock* block, dmDescriptor* dd, long vrowno,
                       long nvrows, void* values, long start,
                       long nitems, long vrow_out_offset )
{
 dmBuffer     *buffer = block->buffer;
 dmDataset    *ds     = NULL;
 dmDescriptor *length_dd;        /* Buffer column for varlen array row lengths */
 dmKernel      kernel;
 dmDataType    tsize;

 char* optr = NULL;
 char* iptr = NULL;

 long *llpix      = NULL;
 long *urpix      = NULL;
 long *dpix       = NULL;
 long  lastrow    = 0;
 long  col_offset = 0;          /* Field offset in buffer */
 long  nvrows_seg = 0;          /* Filtered rows returned in this segment */
 long  dvrows     = 0;          /* Contiguous filtered rows to copy  */
 long  dpix1      = 1;
 long  rrow_start_offset = 0;   /* Filtered row offset in buffer of requested first row */
 long  output_row_offset = 0;
 long  rrow_offset = 0;         /* Raw row offset in raw buffer */
 long  nbytes      = 0;         /* Total bytes returned in segment */
 long  size        = 0;         /* Bytes per row for descriptor */
 long  last        = 0;
 long  step        = 0; 
 long  piece       = 0;
 long  dim         = 0;
 long  pdim        = 1;
 long  item_offset = 0;
 long  item_nbytes = 0; 
 long  i           = 0;
 long  cptno       = 0;
 long  arsize      = 0;
 int   partial     = 0;
 int   is_var      = 0;         /* Is dd a variable-length array? */
 


 ds = block->ds;

/* Check for change of block */
 if ( ds->current != block )
 {
  if ( ds->current )
   dmp_buffer_flush( ds->current ); 
  dmi_dataset_set_current( ds, block );
 }

 col_offset = dd->bufno-1;
 if ( col_offset < 0 )
 {
   char tmp[1024];
   sprintf( dm_errbuf, "Error: descriptor %s not in buffer\n", dmi_descriptor_get_qualname( dd, tmp ));
   dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );

   if ( dmverbose > 4 )
   {
     printf( "DD ERROR %s NOT IN BUFFER\n", dmi_descriptor_get_qualname( dd, dm_errbuf ));
   }
   return -1;
 }

/* Get new buffer if needed */

 nvrows_seg = dmp_buffer_adjust( block, vrowno + vrow_out_offset, nvrows, UT_TRUE );

 if ( nvrows_seg == DM_BUF_BADROW )
 {
  nvrows_seg = 0;
  return nvrows_seg;
 }
 if ( nvrows_seg > nvrows )
  nvrows_seg = nvrows;

 dim = dd->dim;
 cptno = dd->cptno;
 if ( cptno > 0 && dd->parent )
  pdim = dd->parent->dim;
 if ( cptno == 0 )
  cptno = 1;
 arsize = dmi_descriptor_get_array_size( dd );
 if ( arsize == 0 )
  arsize = 1;
 size = buffer->sizes[col_offset];
 is_var = dd->varlen;
 
 tsize = dmi_descriptor_get_bytes( dd );
 item_offset = ( (start - 1) + (cptno - 1) * arsize ) * tsize;
 item_nbytes = tsize * nitems * dim;

/* The starting row number may not be at the beginning of the current buffer */
 rrow_start_offset = vrowno - buffer->rrow1;
 partial = ( item_nbytes < size  || item_offset > 0 || pdim > 1 || is_var );

 if ( is_var )
 {
  length_dd = dmi_buffer_get_varlen_length_col( block, dd );

  /* We can only write one variable-length array row at a time */
  if ( !length_dd || nvrows != 1 )
  {
   sprintf( dm_errbuf, "Error in buffering variable-length array descriptor %s; all row-length information will be discarded", dd->name );
   dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );

   /* Continue processing, but treat dd as a fixed-length array */
   is_var = 0;
  }
 }

 /* 
   Write all the rows (writing out is never filtered) except if writing an array subset 
   in which case the data is not contiguous 
 */

 for ( output_row_offset = 0; output_row_offset < nvrows_seg; output_row_offset += dvrows )
 {
   rrow_offset = output_row_offset + rrow_start_offset + vrow_out_offset;
#if DEBUG
printf( "VROWNO = %ld RRO = %ld VR=%ld:%ld RR=%ld:%ld\n", vrowno, rrow_offset, buffer->vrow1, buffer->vrow2, buffer->rrow1, buffer->rrow2 );
#endif

   if ( partial )
     dvrows = 1;
   else
     dvrows = nvrows_seg;

   /* Copy dvrows from data to vdata */
   /* Assumes col_order */

   nbytes = dvrows * item_nbytes;
   optr = (char*)values + item_nbytes * ( output_row_offset + vrow_out_offset );

   if ( dd->flag & DM_FLAG_UNBUF )
   {
     /* Unbuffered write here */

     kernel = ds->kernel;
     last = start + nitems - 1;
     if ( dd->naxes > 1 )
     {
       llpix = malloc( dd->naxes * UT_SZ_I ); 
       urpix = malloc( dd->naxes * UT_SZ_I ); 
       dpix = malloc( dd->naxes * UT_SZ_I ); 
       for ( i = 0; i < dd->naxes; i++ )
	 dpix[i] = 1;

       dmu_pixel_from_pixelno( start, dd->axes, dd->naxes, llpix );
       dmu_pixel_from_pixelno( last, dd->axes, dd->naxes, urpix );
     }
     else 
     {
       llpix = &start;
       urpix = &last;
       dpix = &dpix1;
     }

     /* A kernel may be specialized for writing either columns or rows */
     if ( kernel->ColWrite )
       kernel->ColWrite( dd, llpix, urpix, dpix, 1, 1, optr );

     if ( dd->naxes > 1 )
     {
       free( llpix );
       free( urpix );
       free( dpix );
     }
   }
   else
   {
     /* Buffered */
#if DEBUG
printf( "VALUES = %p\n", values );
printf( "COPY %ld BYTES FROM %p OFF = %d R =%d %d %d\n", nbytes, iptr, (char*)iptr-(char*)buffer->data, item_offset, col_offset, rrow_offset );

#endif
     iptr = (char*)buffer->data + dmi_buffer_offset( buffer, col_offset, rrow_offset ) + item_offset;
     if ( dd->type == DT_CHAR )
     {
       piece = dmi_descriptor_get_bytes( dd );
       optr = (char*)values + UT_SZ_P * ( output_row_offset + vrow_out_offset );
       i = 0;
       while ( nbytes )
       {
	 char* cptr;
	 long slen;
	 cptr = ((char**)optr)[i];
	 step = piece;
	 if ( step > nbytes ) step = nbytes;
	 slen = strlen( cptr ) + 1;
	 if ( slen < step ) step = slen;

	 /* clear memory to be filled in this read */
	 memset( iptr+i*piece, 0, piece );

	 memcpy( (char*)iptr + i * piece, cptr, step );
	 nbytes -= piece;
	 i++;
       }
     }
     else 
     {
       memcpy( iptr, optr, nbytes );

#if DEBUG
if ( dd->type == DT_BIT && size == 2 )
{
printf( "BIT %s VAL %d Size %ld Tsize %ld INB %ld DVR %ld NVR %ld START %ld NITEMS %ld NB %ld\n", 
  dd->name, *(unsigned char*)optr, size, tsize, item_nbytes, dvrows, nvrows, start, nitems, nbytes );
}
#endif

     }

     if ( is_var )
     {
       /* Store number of values written to this variable-length array row, taking into
	* account the specified starting offset into the array */
       dmi_buffer_write_varlen_length( block, dd, length_dd, (void*)iptr, nitems + start - 1 );
     }
   } /* end buffered write */

 } /* End row loop */

 buffer->update = UT_TRUE;
 dmi_block_set_update( block );

 /* Set current row */
 lastrow = vrowno + vrow_out_offset + nvrows_seg - 1; 

 if ( buffer->vrow1 == 0 )
  buffer->vrow1 = buffer->rrow1;

 buffer->lastrow = lastrow;
 if ( buffer->lastrow > block->nrows )
  block->nrows = buffer->lastrow;
 if ( buffer->lastrow > block->nvrows )
  block->nvrows = buffer->lastrow;
 if ( buffer->lastrow > buffer->rrow2 )
  buffer->rrow2 = buffer->lastrow;
 return nvrows_seg;
}



/***************************************************************/
/*                                                             */
/* Function:   dmp_buffer_read                                 */
/*                                                             */
/* Description:                                                */
/*    Read filtered rows from buffer                           */
/*    The llpix/urpix/dpix arguments are not currently used    */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock      *block                      */
/*                    dmDescriptor *dd                         */
/*                    long          vrowno                     */
/*                    long          nvrows                     */
/*                    void         *values                     */
/*                    long          item_offset                */
/*                    long          item_nbytes                */
/*                    long          vrow_out_buffer            */
/*                    long         *llpix                      */
/*                    long         *urpix                      */
/*                    long         *dpix                       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long  nvrow_seg                                  */
/*                                                             */
/***************************************************************/
long dmp_buffer_read( dmBlock* block, dmDescriptor* dd, long vrowno,
                      long nvrows, void* values, long item_offset,
                      long item_nbytes, long vrow_out_offset,
                      long* llpix, long* urpix, long* dpix )
{
 dmBuffer     *buffer = block->buffer;
 dmDataset    *ds     = NULL;
 dmDescriptor *image  = NULL;
 dmKernel      kernel;

 void *bdata = NULL;
 char *optr  = NULL;
 char *iptr  = NULL;

 long  col_offset        = 0;
 long  piece             = 0;
 long  step              = 0;
 long  i                 = 0;
 long  nvrows_seg        = 0;   /* Filtered rows returned in this segment */
 long  dvrows            = 0;   /* Contiguous filtered rows to copy  */
 long  contig_start      = 0;
 long  vrow_offset       = 0; 
 long  vrow_start_offset = 0;   /* Filtered row offset in buffer of requested first row */
 long  output_row_offset = 0;
 long  nleft             = 0;
 long  nbytes            = 0;   /* Total bytes returned in segment */
 long  size              = 0;   /* Bytes per row for descriptor */
 long  flag              = 0;
 long  n                 = 0;
 long  byte_size         = 0;
 int   partial           = 0;
 int   direct            = 0;


 col_offset = dd->bufno-1;

 if ( col_offset < 0 )
 {
   char tmp[1024];
   sprintf( dm_errbuf, "Error: descriptor %s not in buffer\n", dmi_descriptor_get_qualname( dd, tmp ));
   dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );

   if ( dmverbose > 4 )
   {
     sprintf( dm_errbuf, "buffer read: DD ERROR %s NOT IN BUFFER\n", dmi_descriptor_get_qualname( dd, tmp ));
     printf( "%s\n", dm_errbuf );
     dmi_buffer_print( block );
   }

   dmi_block_set_current_row( block, dmENDOFTABLE );
   return -1;
 }

/* Get new buffer if needed */
 nvrows_seg = dmp_buffer_adjust( block, vrowno, nvrows, UT_FALSE );

 if ( nvrows_seg == DM_BUF_BADROW )
 { 
  return 0;
 }
/* If buffer is read/write, available rows may be larger */
 if ( nvrows_seg > nvrows )
  nvrows_seg = nvrows;

 if ( dmverbose > 4 )
 {
/* dpix used to supress compiler warn */
  fprintf( stderr, "Buffer adjust to VRN %ld:%ld Raw NR=%ld NSEG=%ld DPIX = %p\n", vrowno, vrowno+nvrows_seg-1, block->nrows, nvrows_seg, (void*)dpix );
 }

 size = buffer->sizes[col_offset];

 vrow_start_offset = vrowno - buffer->vrow1;
 partial = ( item_nbytes < size  || item_offset > 0 );

 flag = dd->flag;
 bdata = buffer->data;

 if ( flag & DM_FLAG_UNBUF )
 {
  ds = block->ds;
  kernel = ds->kernel;
/* Direct read without buffering. Assumes no row filtering. But there may be image filter.
 * Suppose we have a 512x512 image and we are filtering 100:300,100:300 and reading a 
 * subset 1:10,1:10 of that. We need to translate that to reading subset  100:109,100:109
 * of the 512x512 image. So we have an image descriptor for the raw image and one for
 * the filtered image.
 *
 * Eventually we want to support this for any array table column:
 *   evt.fits[cols *,phas=phas[2:4,2:4]]
 * This is just saying that each entry in block->cols is a function of entries
 * in block->rawcols. When we make the kernel->colread calls in the buffer,
 * we'll then set llpix and urpix accordingly. 
 *
 * When we look at dd->axes in the kernel,
 * we'll see 512x512. When we look at dd->axes in the user interface, we'll get 201x201.
 * Hence they have to be different descriptors.
 *
 */

/* Assumes single row read from image here  - change if table dds ever unbuffered.
 *  Also, size needs to be recalculated if dd data type is ever a cast - make it the buffer size
 *  in #items 
 */

  /* allow the user to request data from the unscaled data descriptor */
  if ( 0 != ( n = dmi_descriptor_match( dd, buffer->cols, buffer->ncols ) ) )
   image = buffer->cols[n-1];
  else 
   image = block->rawcols[0];

  dmp_filtered_image_read( dd, image, llpix, urpix, NULL, NULL, values, size ); 
  nvrows_seg = 1;
 }
 else
 {
   /* Now get it in contiguous segments from the raw buffer */
   direct = !buffer->filtered || buffer->in_vdata;

   if ( buffer->in_vdata )
     bdata = buffer->vdata;

   /* direct and not partial is simplest and most common case,
    * make it first and make it fast. */

   if ( direct && !partial )
   {
     iptr = (char*)bdata + dmi_buffer_offset( buffer, col_offset, vrow_start_offset ) + item_offset;
     nbytes = nvrows_seg * item_nbytes;

     if ( dd->type != DT_CHAR )
     {
       optr = (char*)values + item_nbytes * vrow_out_offset;
       memcpy( optr, iptr, nbytes );
     }
     else
     {
       piece = dmi_descriptor_get_bytes( dd );
       optr = (char*)values + UT_SZ_P * ( vrow_out_offset );
       i = 0;
       step = piece;

       while ( nbytes )
       {
         if ( step > nbytes )
           step = nbytes;

         memcpy( ((char**)optr)[i], (char*)iptr + i * piece, step );
         nbytes -= piece;
         ++i;
       }
     }
     return nvrows_seg;
   }

  /* if we get here, this is either a partial or not direct read meaning
   * we can't just copy it all at once.
   * If it is a partial read, we need to loop over each row.
   * If it is not direct, we need to loop over each contiguous segment */

  /* loop over the segment rows */

  if ( partial )
    dvrows = 1L;

  for ( output_row_offset = 0; output_row_offset < nvrows_seg; output_row_offset += dvrows, iptr += size )
  {
    nleft = nvrows_seg - output_row_offset;
    vrow_offset = output_row_offset + vrow_start_offset;

    if ( direct )
      iptr = (char*)bdata + dmi_buffer_offset( buffer, col_offset, vrow_offset ) + item_offset;   
    else
      iptr = (char*)bdata + dmi_buffer_offset( buffer, col_offset, buffer->index[vrow_offset] ) + item_offset;

    if ( !partial )
    {
      contig_start = buffer->contig_start[vrow_offset];
      dvrows = buffer->contiguous[contig_start] - ( vrow_offset - contig_start );

      if ( dvrows > nleft ) 
        dvrows = nleft;
      else if ( dvrows == 0 ) 
      {
        long rrow_offset;

        if ( direct )
          rrow_offset = vrow_offset;
        else
          rrow_offset = buffer->index[vrow_offset];
        sprintf( dm_errbuf, "Sync error in buffer: nvrows_seg = %ld partial=%d vrow_offset=%ld rrow_offset=%ld nleft=%ld",
	         nvrows_seg, partial, vrow_offset, rrow_offset, nleft );
        dmu_err_set( DM_ERR_INTERNAL, dm_errbuf );
        return 0;
      }
    }
 
    nbytes = dvrows * item_nbytes;

    if ( dd->type != DT_CHAR )
    {
      optr = (char*)values + item_nbytes * ( output_row_offset + vrow_out_offset );
      if ( dd->naxes > 1 && (llpix && urpix) )
      {
        dmp_buffer_subarray_extract(dd, llpix, urpix, iptr, optr );
      }
      else
      {
        if ( !partial || dd->dim <= 1 || dd->size <= 1 )
          memcpy( optr, iptr, nbytes );
        else
        {
          /* Need to handle case where column is a vector of arrays and the 
	     user requested a subset of the array */

          byte_size = dmi_descriptor_get_bytes( dd );
          piece = item_nbytes  / dd->dim;
          step = dd->size * byte_size;
          i = 0;

          while ( nbytes > 0 )
          {
            if ( piece > nbytes )
              piece = nbytes;
            memcpy( (char*)optr + piece*i, (char*)iptr + i*step, piece );
            nbytes -= piece;
            ++i;
          }
        }
      }
    }
    else /* type == DT_CHAR */
    {
      piece = dmi_descriptor_get_bytes( dd );
      optr = (char*)values + UT_SZ_P * ( output_row_offset + vrow_out_offset );
      i = 0;
      step = piece;
    
      while ( nbytes )
      {
        if ( step > nbytes )
          step = nbytes;
     
        memcpy( ((char**)optr)[i], (char*)iptr + i * piece, step );
        nbytes -= piece;
        ++i;
      }     
    }
   } 
  }
  return nvrows_seg;
}




/***************************************************************/
/*                                                             */
/* Function:   dmp_copy_buffer_to_row                          */
/*                                                             */
/* Description:                                                */
/*   There's a difference between row and column based access. */
/*   String arrays are returns as arrays of pointers in column */
/*   based calls and directly in the buffer in row based       */
/*   access. Because of this and for efficiency, I duplicate   */
/*   the buffer read code for the two cases.  Note that the    */
/*   buffer adjust code is called in tablegetrow.              */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock    *block                        */
/*                    void       *row_buffer                   */
/*                    long        vrowno                       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmp_copy_buffer_to_row( dmBlock* block, void* row_buffer, long vrowno )
{
 dmBuffer      *buffer = block->buffer;
 dmDescriptor  *dd     = NULL;
 dmDescriptor **pdd    = NULL;
 dmDataset     *ds     = NULL;
 dmDescriptor  *image  = NULL;
 dmDescriptor **cols   = NULL;
 dmKernel       kernel;

 void  *llpix  = NULL;
 void  *urpix  = NULL;
 char  *optr   = NULL;
 char  *bdata  = NULL;
 char  *iptr   = NULL;
 char  *rowbuf = (char*)row_buffer;
 long  *psize  = NULL;
 long  *poff   = NULL;

 long col_offset  = 0;
 long vrow_offset = 0;
 long rrow_offset = 0;    /* Raw row offset in raw buffer */
 long size        = 0;    /* Bytes per row for descriptor */
 long flag        = 0;
 long n           = 0;
 long ncols       = 0;
 long nrows       = 0;
 int  direct      = 0;


/* In case we have an unbuffered read */
 ds = block->ds;
 kernel = ds->kernel;

/* Loop over cols */
 vrow_offset = vrowno - buffer->vrow1;

 bdata = (char*)buffer->data;
 if ( buffer->in_vdata )
  bdata = (char*)buffer->vdata;
 direct = !buffer->filtered || buffer->in_vdata;
 if ( direct )
  rrow_offset = vrow_offset;
 else
  rrow_offset = buffer->index[vrow_offset];

 nrows = buffer->nrows;
 ncols = buffer->ncols;
 cols = buffer->cols;
 /* need to make this loop as tight as possible, hence all the pointer
  * increments */
 poff = &buffer->voffsets[0];  /* We want visible columns only, so use voffsets */
 pdd = &cols[0];
 psize = &buffer->sizes[0];
 for ( col_offset = 0; col_offset < ncols; col_offset++,pdd++,psize++,poff++ )
 {
  optr = rowbuf + *poff;
  dd = *pdd;
  size = *psize;

  flag = dd->flag;
  /* skip invisible cols */
  if ( !( flag & DM_FLAG_VISIBLE ) ) continue;

  if ( flag & DM_FLAG_UNBUF )
  {
/* Direct read without buffering. See notes in code for col access. */
   if ( 0 != ( n = dmi_descriptor_match( dd, cols, ncols ) ) )
    image = cols[n-1];
   else 
    image = block->rawcols[0];
   dmp_filtered_image_read( dd, image, llpix, urpix, NULL, NULL, optr, size ); 
  } else {
/* Now get it in contiguous segments from the raw buffer */
   iptr = bdata + buffer->offsets[col_offset] * nrows + rrow_offset * size;
   /*   iptr = bdata + dmi_buffer_offset( buffer, col_offset, rrow_offset ); */
   memcpy( optr, iptr, size );
  }
 }
}




/***************************************************************/
/*                                                             */
/* Function:   dmp_buffer_adjust                               */
/*                                                             */
/* Description:                                                */
/*   Load given virtual row numbers.                           */
/*   Return smaller of actual and requested number of filtered */
/*   rows.                                                     */
/*                                                             */
/*   Developer Note:                                           */
/*     This function is called a LOT, and most of the time     */
/*     returns very early since the data is in the buffer.     */
/*     Therefore, initialize variables as late in the code as  */
/*     possible.                                               */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock    *block                        */
/*                    long        nvrows                       */
/*                    int         create                       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long  nvrows                                     */
/*                                                             */
/***************************************************************/
long dmp_buffer_adjust( dmBlock* block, long vrowno, long nvrows, int create )
{
 dmBuffer  *buffer = block->buffer;
 dmDataset *ds     = NULL;
 dmKernel   kernel;

 /* Local variables for determining next buffer segment   */
 /*   without modifying the current buffer settings       */
 long  next_rrow1 = 0;
 long  next_rrow2 = 0;
 long  next_vrow1 = 0;
 long  next_vrow2 = 0;

 long  dvrow1       = 0;
 long  last_vrow1   = 0;
 long  nvrows_total = 0;    /* Filtered rows in buffer */
 long  nvrows_got   = 0;    /* Filtered rows returned */
 long  nrrows_left  = 0;    /* Number of rows left free in raw buffer */

 /* The llpix urpix dpix arrays support image subset reading  */
/* - currently handled in the unbuffered reads and null here. */

 long kernel_buffer_nrows = 0;
 long nrrows_kseg = 0;         /* Number of raw rows read in kernel segment */
 long nvrows_kseg = 0;         /* Number of filtered rows read in kernel segment */
 long new_size    = 0;         /* size of new buffer if resync'ed */

 dmBool status       = dmFALSE;
 dmBool is_binned    = dmFALSE; 
 dmBool clearMemory  = dmFALSE;


/*
 *============================
 * Check if buffer is already full
 *============================
 */

 /* Set range of rows being requested. */
 buffer->vrow_go1 = vrowno;
 buffer->vrow_go2 = vrowno + nvrows -1;

 /* Check current buffer range to see if the current buffer covers part of the range already */
 if ( !buffer->resync && buffer->vrow1 <= buffer->vrow_go1 )  
 { 
   /* Buffer start is prior to first target row */

   /* NOTE: vrow2 is the upper virtual row number in the buffer. we have read 
    * up to the smaller of vrow2 and nvrows. */

  if ( buffer->vrow2 >= buffer->vrow_go2 )
  {
    /* Buffer fully covers requested range */
   /*   nvrows_got = nvrows;   faster... */
   return nvrows;
  }
  else if ( buffer->vrow2 >= buffer->vrow_go1 )  
  {
    /* Buffer covers part of range */
   /*   nvrows_got = buffer->vrow2 - buffer->vrow_go1 + 1; */
   return buffer->vrow2 - buffer->vrow_go1 + 1;
  }
 }

 /* If here,                                 */
 /* Buffer does NOT include the desired rows */

/*
 *============================
 * Flush and resync buffer
 *============================
 */


 if ( buffer->update )
  dmp_buffer_flush( block );


 ds = block->ds;
 kernel = ds->kernel;
 if( buffer->kernel_rows == 0 )
  buffer->kernel_rows = kernel->BufferSize( block );


/* Initialize minrow and maxrow; the buffer reset may change them if there is a row filter */
 if ( buffer->resync )
  dmi_virtual_buffer_reset( block, block->buffer );

 is_binned = ( buffer->n_index && buffer->n_index[DMBI_BINNED] > 0 );

/* Binned image has only one row */
 if ( buffer->vrow_go1 > 1 && is_binned )
  return DM_BUF_BADROW;

/* Ensure we have a buffer and it is big enough to hold nvrows */
 if ( buffer->resync || !buffer->data )
  dmi_buffer_resync( block, buffer, nvrows );
 else if ( buffer->nrows < nvrows )
 {
  /* Only resync if it is possible to get a larger buffer than
   * the one we already have. */
  new_size = dmi_buffer_eval_size( block, nvrows );
  if ( new_size > buffer->nrows )
   dmi_buffer_resync( block, buffer, nvrows );
 }
 
 /* Initialize next_ values */
 next_rrow1 = 0;
 next_rrow2 = 0;
 next_vrow1 = 0;
 next_vrow2 = 0;
 clearMemory = dmFALSE;


/*
 *================================
 * Seek to desired rows
 *================================
 * Now we care if the buffer is filtered or not. If not,
 * we can skip ahead (or back) to the desired row. If it's filtered,
 * we have to go row by row. */

 if ( buffer->filtered || is_binned )
 {
   clearMemory = dmTRUE;
   if ( (buffer->vrow_go1 < buffer->vrow1) || (buffer->vrow1 == 0 && buffer->vrow_go1 == 1 ))
   {
     /* Rewind */
     next_rrow1 = buffer->minrow;
     next_vrow2 = 0;
   } 
   else 
   {
     /* Desired segment is after current segment */
     next_rrow1 = buffer->rrow2 + 1;
     next_vrow2 = buffer->vrow2;
   }

   /* Set virtual row start */
   next_vrow1 = next_vrow2;

   /* Set raw row end; Check within file bounds */
   next_rrow2 = next_rrow1 + buffer->nrows - 1;
   if ( next_rrow2 > buffer->maxrow ) 
     next_rrow2 = buffer->maxrow;

 } 
 else 
 {
   /* Not filtered, so virtual rows are raw rows */

   /* Set raw row start */
   next_rrow1 = buffer->vrow_go1;

   /* Set raw row end; Check within file bounds */
   next_rrow2 = next_rrow1 + buffer->nrows - 1;
   if ( next_rrow2 > buffer->maxrow ) 
     next_rrow2 = buffer->maxrow;

   /* Set virtual row bounds */
   next_vrow1 = next_rrow1 - 1;
   next_vrow2 = next_rrow2;
 }

 /* Check for trouble conditions with n2ext segment settings */
 if ( next_rrow1 > buffer->maxrow )
 {
  /* need to handle special case of an empty binned image */
  if ( is_binned )
  {
   buffer->in_vdata = UT_TRUE;
   nvrows_got = 1;
   buffer->vrow1 = 1;
   buffer->vrow2 = 1;
   buffer->lastrow = 1;
   return nvrows_got;  
  }
  else
  {
   /* Desired range is beyond EOF */
   return DM_BUF_BADROW;
  }
 }
 else if ( next_rrow1 > block->nrows + 1) 
 { 
   /* Writing too far beyond EOF */
   return DM_BUF_BADROW;
 }
 else if ( ( next_rrow1 == block->nrows + 1 )  && !create )
 {
   /* Writing beyond EOF and cant create */
   return DM_BUF_BADROW;
 }

 /* OK to modify buffer variables */
 if ( clearMemory )
 {
   memset( buffer->contiguous, 0, UT_SZ_I * buffer->nrows );
   memset( buffer->index, 0, UT_SZ_I * buffer->nrows );
   memset( buffer->contig_start, 0, UT_SZ_I * buffer->nrows );
 }

 buffer->rrow1   = next_rrow1;
 buffer->rrow2   = next_rrow2;
 buffer->vrow1   = next_vrow1 + 1;
 buffer->vrow2   = next_vrow2;
 buffer->lastrow = next_vrow1;

 nrrows_left = next_rrow2 - next_rrow1 + 1;

 /* Check case of writing beyond current EOF, but OK */
 if ( ( next_rrow1 == block->nrows + 1 )  && create ) {
   return nrrows_left;
 }
 /*
  * We now have reserved the raw row buffer.
  * It will usually be size buffer->nrows, but it might be
  * shortened by the end of file 
  */


/* The top row to read may be less than the top of buffer if we are writing.
 * It is the smaller of max_rawrow, block->nrows and buffer row2
 * On the other hand for a filtered file we may increment rrow2 on the fly
 */
 buffer->rrow_top = buffer->rrow2; /* Was buffer->rrow2 */
 if ( buffer->rrow_top > block->nrows || is_binned )
  buffer->rrow_top = block->nrows;

/* Fill the raw buffer.
   We must keep looping until we have some rows in the buffer
   beyond vrow1, but mustn't exceed raw row = buffer->nrows.

   nrrows_left is incremented only if good rows in the desired range are read
   Thus, srowno1 and srowno2 get incremented even if buffer row1/row2 are not.

   Cases are:
    writing: endrow = buffer->rrow2
    reading: endrow = block->nrows
   Example:  block->nrows = 1000, buffer->nrows = 100
             rrow_end = 1000
             rrow_top = 100
             

  This value of endrow is different from the one used above to determine nrrows_left.



 */

 if ( create )
  buffer->rrow_end = buffer->rrow_top;
 else
  buffer->rrow_end = buffer->maxrow;

/*
 *   Ask the kernel how many rows we should read. 
 *   If the kernel doesn't know, read them all.
 */

 nrrows_left = buffer->rrow_top - buffer->rrow1 + 1;

 kernel_buffer_nrows = buffer->kernel_rows;
 if ( kernel_buffer_nrows <= 0 )
  kernel_buffer_nrows = nrrows_left;

/*
 *  This loop is here to optimize for buffers in kernel libraries.
 *  e.g. CFITSIO has kernel_buffer_nrows in memory, so you want to
 *  read all the columns for those rows before accessing any other rows.
 */

 nrrows_kseg = kernel_buffer_nrows;
 nvrows_kseg = 0;
 dvrow1 = 0;
 last_vrow1 = 0;
 nvrows_total = 0;
/*
 *  Put raw row pointer just before first row to read
 */
 buffer->soffset =0;
 buffer->rrow_seg1 = 0;
 buffer->rrow_seg2 = buffer->rrow1 - 1;
 while ( (is_binned ||  
         ( buffer->rrow1 + buffer->soffset <= buffer->rrow_top 
           && nrrows_left > 0 ) )
         && buffer->rrow_seg2 < buffer->rrow_end )
 {

/*
 * The lower end of the virtual buffer is incremented by the number of virtual
 * rows we read last time.
 */
  last_vrow1 = buffer->vrow1;
  buffer->vrow1 += dvrow1;

/* 
 *  Specify a chunk of raw rows to read from the kernel to the raw row buffer
 *
 *  The maximum number of slots left in the buffer is the raw rows remaining to read
 *  The maximum number of raw rows to read this segment is the number of slots in the buffer
 */

  if ( nrrows_left > buffer->rrow_end - buffer->rrow_seg2 )
    nrrows_left = buffer->rrow_end - buffer->rrow_seg2;
   
  if ( nrrows_kseg > nrrows_left ) nrrows_kseg = nrrows_left;

  buffer->rrow_seg1 = buffer->rrow_seg2 + 1;
  buffer->rrow_seg2 = buffer->rrow_seg2 + nrrows_kseg;

/* If we are reading a filtered file and don't get any rows,
 * we need to move the raw row buffer up
 * On the other hand if we find a few rows and need to append to the buffer,
 * the boundaries of the buffer don't change.
 * Warning: if we read a filtered chunk with no rows, we may end up
 * with raw rows not being contiguous - we don't bother storing that
 * useless kernel chunk in the buffer.
 */
  if (buffer->soffset == 0 )
  {
   buffer->rrow1 = buffer->rrow_seg1;
   buffer->rrow2 = buffer->rrow_seg1 + buffer->nrows - 1;
   if ( buffer->rrow2 > buffer->rrow_end )
    buffer->rrow2 = buffer->rrow_end;
   if ( buffer->rrow_top < buffer->rrow2 )
    buffer->rrow_top = buffer->rrow2;
  }
/* Read a segment with the given number of virtual rows */
  if ( dmverbose > 2 )
   dmi_buffer_status( block );
  nvrows_kseg = dmp_buffer_read_seg( block, buffer, buffer->vrow_go1 );
/* Keep this buffer only if it contains some filtered rows  (noop on first iter) */
  if ( is_binned ) 
  {
   nvrows_kseg = 0;
   buffer->vrow1 = 0;
   buffer->vrow2 = 0;
   buffer->lastrow = 0;
  }
/* We have some filtered rows and we are not seeking */
  if ( nvrows_total + nvrows_kseg > 0 &&  buffer->lastrow >= buffer->vrow_go1 )
  {
   nvrows_total += nvrows_kseg;
   buffer->soffset += nrrows_kseg;
   nrrows_left -= nrrows_kseg;
   dvrow1 = 0;
  } else {
/* vrowno1 > lastrow, we are seeking, increment vrow1 */
   dvrow1 = nvrows_kseg;
   buffer->soffset = 0; /* make sure soffset is 0! */
  }

/* Loop for next chunk of nleft */

 }

/*========================
 * Copy data to intermediate virtual buffer 
 *========================
 */

 buffer->in_vdata =  ( nvrows_total > 0 && nvrows_total < nvrows );
 if ( buffer->in_vdata )
 {
  status = dmi_buffer_copy_to_vdata( buffer, nvrows_total );
  if ( !status ) return 0;
 }
 if ( nvrows_total > nvrows )
  nvrows_got = nvrows;
 else
  nvrows_got = nvrows_total;

 if ( dmverbose > 2 )
  dmi_buffer_status( block );

 if ( is_binned )
 {
  buffer->in_vdata = UT_TRUE;
  nvrows_got = 1;
  buffer->vrow1 = 1;
  buffer->vrow2 = 1;
  buffer->lastrow = 1;

 } else if ( !buffer->filtered ) {

/*
 * Currently nvrows_got is the number of rows  we read. But
 * it may be that the buffer contains a bunch of empty rows that have
 * not yet been written and we want to count those too. 
 */
  if ( buffer->rrow2 > block->nrows )
   nvrows_got = buffer->rrow2 - buffer->rrow1 + 1;
 } else if ( buffer->vrow1 > buffer->vrow2 && nvrows_got > 0 ) {
  /* if we tried to read more virtual rows, but got none we still have
     the old rows in the buffer and need to back up vrow1 */
  buffer->vrow1 = last_vrow1;
 } else if ( nvrows_got == 0 ) {
  /* Make sure the virtual row numbers reflect that there are no
     virtual rows in the buffer */
  buffer->vrow1 = buffer->vrow2 + 1;
 }

 return nvrows_got;
}


/***************************************************************/
/*                                                             */
/* Function:   dmp_buffer_read_seg                             */
/*                                                             */
/* Description:                                                */
/*     Read a segment of data from the kernel, filter it,      */
/*     and see how many filtered rows we have.                 */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock    *block                        */
/*                    dmBuffer   *buffer                       */
/*                    long        vrowno                       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            long  nvrows_kseg                                */
/*                                                             */
/***************************************************************/
long dmp_buffer_read_seg( dmBlock* block, dmBuffer* buffer, long vrowno )
{
 dmDataset     *ds    = NULL;
 dmDescriptor  *image = NULL;

 long nvrows_kseg = 0;
 long nrrows_kseg = 0;
 int  eof         = UT_FALSE;
 int  seek        = UT_FALSE;
 int  is_binned   = UT_FALSE;
 int  copy        = UT_FALSE;
 int  is_image    = UT_FALSE;


/* The buffer segment is contiguous in raw rows; the whole buffer may not be */
 nrrows_kseg = buffer->rrow_seg2 - buffer->rrow_seg1 + 1;

 is_image = ( dmi_block_get_raw_type( block ) == dmIMAGE );
 if ( buffer->filtered )
 {
  eof = dmp_buffer_read_filter_cols( block, buffer );
  nvrows_kseg = dmi_filter_rows( buffer, buffer->data );
 }
 else
 {
  nvrows_kseg = nrrows_kseg;
 }

 if ( nvrows_kseg > 0 )
 {
/* End the buffer and set the current row to end of buffer */
  buffer->lastrow += nvrows_kseg;
  if ( buffer->lastrow > buffer->vrow2 )
   buffer->vrow2 = buffer->lastrow;

/* Check if we can skip columns we don't need to calculate filtering on */
  is_binned = ( buffer->n_index && buffer->n_index[DMBI_BINNED] > 0 );
  seek = !is_binned && ( buffer->vrow1 > vrowno || buffer->vrow2 < vrowno );
/* Have we completed scanning the file at least once? If so, block->nvrows is good */
  if ( !block->traversed && buffer->lastrow > block->nvrows && !is_binned )
     block->nvrows = buffer->lastrow;

/* At least some rows good - so read the rest of the columns for ALL these rows.
   Further optimization could be envisaged here, but at least we skip the waste
   when NONE of the rows are good.
 */
/* Don't bother if we are just seeking */

  if ( !seek )
  {
   eof = dmp_buffer_read_main_cols( block, buffer, buffer->soffset, buffer->rrow_seg1, buffer->rrow_seg2 );
   dmi_buffer_execute_stack( buffer, buffer->execute2, nvrows_kseg, buffer->filtered ); 
  }

  if ( buffer->image )
  {
   if ( is_image )
   {
    ds = block->ds;
    image = block->rawcols[0];
    copy = ( ds->mode ) && ( image->flag & DM_FLAG_COPY );
 /* For simply writing in dmcopy, don't need to read the (nonexistent) image data in adjust buffer 
  * If normal read, then we must read the image and filter it as we go.
  */
    if ( !copy ) 
     dmp_buffer_rebin_image( block, image, buffer->image, buffer->vdata );
   }
   else
   {
    dmi_buffer_calc_image( block, buffer->image, nvrows_kseg );
   }
  }
 }  /* end if q */   

 /* check if we've traversed all possible raw rows */
 if ( buffer->rrow_seg2 >= buffer->maxrow )
 {
   if ( buffer->image && !is_image )  /* For table binning only */
    dmi_image_filter_regions( buffer->image, buffer->image, NULL, NULL, NULL, buffer->vdata, buffer->subspace );
   block->traversed = UT_TRUE;
 }
 if ( eof )
   return 0;
 return nvrows_kseg;
}



/***************************************************************/
/*                                                             */
/* Function:   dmp_buffer_read_filter_cols                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock    *block                        */
/*                    dmBuffer   *buffer                       */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            int  eof                                         */
/*                                                             */
/***************************************************************/
int dmp_buffer_read_filter_cols( dmBlock* block, dmBuffer* buffer )
{
 dmDescriptor  *dd = NULL;
 dmDataset     *ds = NULL;
 dmKernel        kernel;

 void  *values = NULL;
 long  *llpix  = NULL;
 long  *urpix  = NULL;
 long  *dpix   = NULL;
 long   i      = 0;
 long   nrows  = 0;
 long   col_offset = 0;
 int    status = 0;
 int    eof    = UT_FALSE;


 ds = block->ds;
 kernel = ds->kernel;

 /* A kernel may be specialized to read columns or rows */
 if ( kernel->ColRead )
 {
/* Fill the buffer: get filter quantities for each column */
  for ( i = 0; i < buffer->n_index[DMBI_FILT]; i++ )
  {
   dd = buffer->col_index[DMBI_FILT][i];
/* Where is the offset in the row buffer for this column? */
   col_offset = buffer->offset_index[DMBI_FILT][i];
   values = (char*)buffer->data + dmi_buffer_offset( buffer, col_offset, buffer->soffset );    
   status = kernel->ColRead( dd, llpix, urpix, dpix, buffer->rrow_seg1, buffer->rrow_seg2, values ); 
   if ( !status ) eof = UT_TRUE;
  }

/* Now add the columns that are needed as operands for other filter
 * columns (i.e. virtual columns) */
  for ( i = 0; i < buffer->n_index[DMBI_FREAD]; i++ )
  {
   dd = buffer->col_index[DMBI_FREAD][i];
/* Where is the offset in the row buffer for this column? */
   col_offset = buffer->offset_index[DMBI_FREAD][i];
   values = (char*)buffer->data + dmi_buffer_offset( buffer, col_offset, buffer->soffset );    
   status = kernel->ColRead( dd, llpix, urpix, dpix, buffer->rrow_seg1, buffer->rrow_seg2, values ); 
   if ( !status ) eof = UT_TRUE;
  }
 }
 else
 {
  status = kernel->ReadRows( block, buffer, buffer->rrow_seg1, buffer->rrow_seg2 );
  if ( !status ) eof = UT_TRUE;
 }
/* Implement calculation of virtual cols here */
 nrows = buffer->rrow_seg2 - buffer->rrow_seg1 + 1;
 dmi_buffer_execute_stack( buffer, buffer->execute1, nrows, UT_FALSE ); 
 return eof;
}


/***************************************************************/
/*                                                             */
/* Function:   dmp_buffer_read_main_cols                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock  *block                          */
/*                    dmBuffer *buffer                         */
/*                    long      rrow_offset                    */
/*                    long      rrowno1                        */
/*                    long      rrowno2                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            int   eof                                        */
/*                                                             */
/***************************************************************/
int dmp_buffer_read_main_cols( dmBlock* block, dmBuffer* buffer, long rrow_offset, long rrowno1, long rrowno2 )
{
  dmDescriptor *dd = NULL;
  dmDataset    *ds = NULL;
  dmKernel      kernel;

  void  *values = NULL;
  long  *llpix  = NULL;
  long  *urpix  = NULL;
  long  *dpix   = NULL;
  long   col_offset = 0;
  long   i      = 0;
  int    status = 0;
  int    eof    = UT_FALSE;

  ds = block->ds;
  kernel = ds->kernel;

#if DEBUG
printf( "NREAD = %ld  VR %ld:%ld VRN %ld\n", buffer->n_index[DMBI_READ], buffer->vrow1, buffer->vrow2, (long)0 );
#endif

  if ( kernel->ColRead )
  { 
    for ( i = 0; i < buffer->n_index[DMBI_READ]; i++ )
    {
      dd = buffer->col_index[DMBI_READ][i];
      col_offset = buffer->offset_index[DMBI_READ][i];

#if DEBUG
printf( "READING %s INTO RAW BUFFER COL OFF %ld START OFF %ld\n", dd->name, col_offset, rrow_offset );
#endif

      values = (char*)buffer->data + dmi_buffer_offset( buffer, col_offset, rrow_offset );    
#if DEBUG
{
long nrrows_left = 0;
printf( "Read raw row %ld to %ld col %s, nleft = %ld offset =%ld %p\n",rrowno1, rrowno2, dd->name, nrrows_left, (long)((char*)values-(char*)buffer->data), values );
}
#endif

      /* Descriptors here are always raw cols */

      status = kernel->ColRead( dd, llpix, urpix, dpix, rrowno1, rrowno2, values );        

      if ( !status )
        eof = UT_TRUE;
    }
  }
  else
  {
    // Code Change  September 10, 2008  Bug 10800
    // Binning an image from an ascii table requires data be read into the
    // buffer using ReadRows function. In this case the block type is dmIMAGE,
    // and fails the standard check that looks only for a table. Add to this
    // conditional a check for a its original data type, which is stored in
    // index_block->type. That type should be dmTABLE in this case
  
    if ( !buffer->filtered &&
          (block->type != dmIMAGE || block->index_block->type != dmIMAGE)) 
    {
      /* if buffer is filtered and we prefer ColRead, we already read the rows */
      status = kernel->ReadRows( block, buffer, rrowno1, rrowno2 );

      if ( !status )
        eof = UT_TRUE;
    }
  }
  return eof;
}



/***************************************************************/
/*                                                             */
/* Function:   dmp_block_copy_buffer                           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock  *block1                         */
/*                                                             */
/*  Output parameters:                                         */
/*                    dmBlock  *block2                         */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmp_block_copy_buffer( dmBlock* block1, dmBlock* block2 )
{
 dmBuffer  *buffer1 = NULL;
 dmBuffer  *buffer2 = NULL;
 dmDataset *ds2     = NULL;
 dmKernel   kernel;

 void  *data1       = NULL;
 void  *data2       = NULL;
 long   vrowno      = 1;
 long   nvrows      = 0;
 long   nvrows_in   = 1;
 long   nvrows_seg  = 1;
 long   buffer_size = 0;


 if ( !block1 || !block2 || !block1->buffer || !block2->buffer )
 {
  dmu_err_set( DM_ERR_INTERNAL, "Problem in buffer copy" );
  return;
 }
/* Trigger buffer init */
 nvrows_seg = dmp_buffer_adjust( block1, vrowno, nvrows_in, UT_FALSE );

 buffer1 = block1->buffer;
 buffer_size = buffer1->nrows;
/* Synch output buffer size to input one */
 buffer2 = block2->buffer;
 buffer2->kernel_rows = buffer_size;

 nvrows_seg = dmp_buffer_adjust( block2, vrowno, nvrows_in, UT_TRUE );


 if ( buffer1->image || buffer1->in_vdata )
  data1 = buffer1->vdata;
 else 
  data1 = buffer1->data;
 data2 = buffer2->data;
 nvrows = 1;  /* One row is enough to trigger filling of buffer; will use default size */
 nvrows_seg = 1;

 while ( nvrows_seg > 0 )
 {
#if DEBUG
printf( "BUFFER ACOPY ROW %ld:%ld  WAS %ld:%ld BR2 = %ld:%ld\n", vrowno, vrowno+nvrows-1, buffer1->vrow1, buffer1->vrow2, buffer2->rrow1, buffer2->rrow2);
#endif
  nvrows_seg = dmp_buffer_adjust( block1, vrowno, nvrows, UT_FALSE );


/* Test for bad buffer (e.g. null image). If good, set full range of buf */
  if ( nvrows_seg > 0 )
   nvrows_seg = buffer1->vrow2 - buffer1->vrow1 + 1;

  if ( nvrows_seg > 0 )
  {
#if DEBUG
printf( "BUFFER BCOPY ROW %ld:%ld V %ld:%ld NIN %ld BR2 %ld:%ld DATA %p %p D25120=%ld\n", vrowno, vrowno+nvrows_seg-1, 
buffer1->vrow1, buffer1->vrow2, nvrows_in, buffer2->rrow1, buffer2->rrow2, data1, data2, 0);
#endif
   dmp_buffer_adjust( block2, vrowno, nvrows_seg, UT_TRUE );



   if ( block1->type == dmIMAGE )
   {
    dmDescriptor* dd1;
    dmDescriptor* dd2;
#if 0
    dd1 = dmi_descriptor_get_unscaled( block1->cols[0] );
    dd2 = dmi_descriptor_get_unscaled( block2->cols[0] );
#endif
    dd1 = block1->cols[0];
    dd2 = block2->cols[0];
    dmp_buffer_copy_image_data( dd1,dd2 );
   }
   else
    dmi_buffer_copy( buffer1, buffer2, data1, data2, nvrows_seg );

   buffer2->update = UT_TRUE;
   dmi_block_set_update( block2 );
   vrowno += nvrows_seg;
   buffer2->vrow1 = buffer1->vrow1;
   buffer2->vrow2 = buffer1->vrow2;
   buffer2->lastrow= buffer2->vrow2;
   block2->nrows = buffer2->lastrow;
   block2->nvrows = buffer2->lastrow;
  }
 }

 if ( buffer2->update )
  dmp_buffer_flush( block2 );

 ds2 = block2->ds;
 kernel = ds2->kernel;
 if ( kernel )
  kernel->BlockFlush( block2 ); 

 dmi_block_set_current_row( block2, buffer2->lastrow+1 );
}



/***************************************************************/
/*                                                             */
/* Function:   dmp_buffer_rebin_image                          */
/*                                                             */
/* Description:                                                */
/*     Currently, 'opt full' only works with 2D images because */
/*     of the way we offset the data buffers to position the   */
/*     valid part of the image correctly.  I believe we can    */
/*     make this code much prettier and allow for 1D-nD        */
/*     'opt full' if we use llpix and urpix to do this instead.*/
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock      *block                      */
/*                    dmDescriptor *parent_image               */
/*                    dmDescriptor *image                      */
/*                                                             */
/*  Output parameters:                                         */
/*                    void         *data                       */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmp_buffer_rebin_image( dmBlock* block, dmDescriptor* parent_image, dmDescriptor* image, void* data )
{
 dmBuffer   *buffer  = NULL;
 dmBinSpec  *binspec = NULL;
 dmDescriptor *logical_image = NULL;
 void *image_buffer  = NULL;
 void *image_ptr     = NULL;
 void *data_ptr      = NULL;

 double step_div = 1.0;

 long *faxes = NULL;
 long *llpix = NULL;
 long *urpix = NULL;
 long *tmin  = NULL;
 long *tmax  = NULL;
 long n      = 0;
 long i      = 0;
 long imrow  = 0;
 long width  = 1;
 long imrow0 = 0;
 long imrow1 = 0;
 long imrow2 = 0;
 long pnaxes = 0;
 long dx     = 0;
 int  rebin  = 0;
 long new_size       = 0;
 long default_size   = 1000000;
 int  is_buffered    = 0;
 int  force_buffered = 0;

 /* The size of the part of the image that passes the range filter.*/
 /* This is necessary to allow 'opt full' without UMRs.            */
 /* I believe we should be able to merge image_buffer_size and     */
 /* filtered_buffer_size into one.                                 */

 long image_buffer_size    = 0;
 long filtered_buffer_size = 0;



 buffer = block->buffer;
 logical_image = dmi_coord_get_parent( image );

 pnaxes =  parent_image->naxes;
 binspec = dmi_descriptor_get_aux( logical_image );
 if ( !binspec ) binspec = calloc( 1, sizeof( struct dmBinSpec_s ));
 if ( !binspec->offsets )
  dmi_bin_map_offsets( binspec, parent_image, image );
 llpix = calloc( pnaxes, UT_SZ_I );
 urpix = calloc( pnaxes, UT_SZ_I );
 faxes = calloc( pnaxes, UT_SZ_I );
 for ( i = 0; i < pnaxes; i++ )
 {
  llpix[i] = binspec->start[i];
  faxes[i] = binspec->image_maxes[i] - binspec->image_mins[i] + 1;
  urpix[i] = binspec->start[i] + faxes[i];
 }
 tmin = malloc( pnaxes * UT_SZ_I );
 tmax = malloc( pnaxes * UT_SZ_I );
 for ( i = 0; i < pnaxes; i++ )
 {
  tmin[i] = binspec->image_mins[i];
  tmax[i] = binspec->image_maxes[i];
 }
 if ( ( image->naxes <= 2 ) || block->compact )
  dmi_block_apply_image_filter( parent_image, tmin, tmax );
 force_buffered = UT_FALSE;
 for ( i = 0; i < pnaxes; i++ )
 {
  if ( tmin[i] != binspec->image_mins[i] || 
       tmax[i] != binspec->image_maxes[i] )
  {
   dx = ( tmin[i] - binspec->image_mins[i] ) % (long)binspec->step[i];
   if ( dx > 0 )
   {
    /* need to adjust if the filter lands inside of the new 
     * pixels */
    llpix[i] += dx;
    urpix[i] += dx;
   }
   force_buffered = UT_TRUE;
  }
 }

 image_buffer_size = dmi_calc_image_buffer_size( faxes, image->naxes, pnaxes, binspec, parent_image->type, default_size, force_buffered, &is_buffered );
 new_size = ar_mix_type_size( image->type );
 rebin =(  binspec->n > 0 );
 if ( rebin )
  image_buffer = malloc( image_buffer_size * new_size );
 else
  image_buffer = data;

 if ( is_buffered )
 {
  imrow1 = tmin[pnaxes-1];
  imrow2 = tmax[pnaxes-1];
  imrow0 = binspec->image_mins[pnaxes-1];
  n = pnaxes - 1;
  width = 1;
  filtered_buffer_size = 1;
  for ( i = 0; i < pnaxes-1; i++ )
  {
   width *= image->axes[i];
   filtered_buffer_size *= (tmax[i] - tmin[i]+1);
  }
  step_div *= binspec->step[pnaxes-1];
 } else {
  imrow1 = 1;
  imrow2 = 1;
  imrow0 = 1;
  n = pnaxes;
  filtered_buffer_size = image_buffer_size;
 }

/*
 *  If we are image filtering with [opt full], so that
 *  our subset goes into only part of the output array,
 *  we want to set the lower left corner of the data.
 *  This won't work if pnaxes > 2 since the subset will not
 *  be contiguous in the output array.
 *  To fix this, we should either just read the whole array
 *  and then apply the filter to set things to zero or NULL. 
 *  This is a bit wasteful though.
 *  Even as is, [opt full,null=-999] won't work on an image.
 *  
 *  However, the normal use when you filter an image to just
 *  get the subset should work fine even in 3D and more.
 *   dmcopy "cube[x=100:200,y=50:100,z=2:10]" subcube
 *  It's just
 *   dmcopy "cube[x=100:200,y=50:100,z=2:10][opt full]" subcube
 *  that doesn't work.
 */

 if ( rebin || pnaxes == 2 )
 {
  /* This offset dx exists if there is a filter smaller than the image
   * field and 'opt full' is set.   */
  dx = ( tmin[0] - binspec->image_mins[0] ) / binspec->step[0];
  if ( image->null )
  {
   /* Need to set empty space around image to null */
   long image_min_offsets_2d[2];
   long image_max_offsets_2d[2];
   long* image_min_offsets;
   long* image_max_offsets;
   if ( pnaxes == 2 )
   {
    image_min_offsets = &image_min_offsets_2d[0];
    image_max_offsets = &image_max_offsets_2d[0];
   }
   else
   {
    image_min_offsets = calloc( pnaxes, UT_SZ_I );
    image_max_offsets = calloc( pnaxes, UT_SZ_I );
   }
   for ( i = 0; i < pnaxes; i++ )
   {
    image_min_offsets[i] = ( tmin[i] - binspec->image_mins[i] ) / binspec->step[i];
    image_max_offsets[i] = ( tmax[i] - binspec->image_mins[i] ) / binspec->step[i];
    if ( binspec->start[i] > 0.0 )
    {
     image_min_offsets[i] += binspec->start[i];
     image_max_offsets[i] += binspec->start[i];
    }
   }
   /* nullifying is optimized for the 2D case.  Use the slower
    * method for the general case. */
   if ( pnaxes == 2 )
    dmp_buffer_nullify_image_space_2d( data, image->null, image->type, image->axes, image_min_offsets, image_max_offsets );
   else
   {
    dmp_buffer_nullify_image_space_gen( data, image->null, image->type, image->axes, pnaxes, image_min_offsets, image_max_offsets );  
    free( image_min_offsets );
    free( image_max_offsets );
   }
  }
 }
 else
  dx = 0;

/* Read a row from the image */  
 for ( imrow = imrow1; imrow <= imrow2; imrow++ )
 {  
  if ( is_buffered ) 
  {
   tmin[pnaxes-1] = imrow;
   tmax[pnaxes-1] = imrow;
  }
  
  data_ptr = (char*)data + ((((long)(((double)(imrow-imrow0) + 0.5) / step_div ) ) * width ) + dx ) * new_size;
  if ( rebin )
   image_ptr = image_buffer;
  else
   image_ptr = data_ptr;


/* Here tmin, tmax must be in unfiltered image logical units */
  dmp_filtered_image_read( image, parent_image, llpix, urpix, tmin, tmax, image_ptr, image_buffer_size );
  if ( rebin )
   dmi_image_rebin_segment( image, image_ptr, filtered_buffer_size, binspec->axis_map, llpix, urpix, binspec->step, pnaxes, data_ptr );

 }

 free( llpix );
 free( urpix );
 free( faxes );
 if ( rebin )
  free( image_buffer );
 free( tmin );
 free( tmax );
 dmi_binspec_free_cache( binspec );
}



/***************************************************************/
/*                                                             */
/* Function:   dmi_block_apply_image_filter                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor *image                      */
/*                    long         *v_llpix                    */
/*                    long         *v_urpix                    */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmi_block_apply_image_filter( dmDescriptor* image , long* v_llpix, long* v_urpix )
{
 dmBlock    *block  = NULL;
 dmSubspace *rts    = NULL;
 dmBuffer   *buffer = NULL;
 dmFilter  **filter_list = NULL;

 int compact = UT_TRUE; /* calculate the compact boundaries even though not in compact mode */
 int reset   = UT_FALSE;



 block = (dmBlock*)image->block;
 if ( block->compact ) return;
 if ( dmi_block_get_raw_type( block ) != dmIMAGE ) return;

 buffer = block->buffer;
 if ( buffer )
  rts = buffer->subspace;
 if ( !rts ) return;
/* Reset to filter boundaries for case of image with opt = full */
 filter_list = dmi_map_filters_to_image_groups( image, rts, compact );
 dmi_map_image_filter_bounds( image, filter_list, v_llpix, v_urpix, image->naxes, compact, reset );
 free( filter_list );
}



/***************************************************************/
/*                                                             */
/* Function:   dmp_filtered_image_read                         */
/*                                                             */
/* Description:                                                */
/*    Read entries from a filtered image.  llpix and urpix are */
/*    the logical pixel ranges for the filtered image.         */
/*    v_llpix and v_urpix are the logical ranges for the raw   */
/*    image.                                                   */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor *image                      */
/*                    dmDescriptor *raw_image                  */
/*                    long         *llpix                      */
/*                    long         *urpix                      */
/*                    long         *v_llpix                    */
/*                    long         *v_urpix                    */
/*                    void         *image_buffer               */
/*                    long          image_buffer_size          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmp_filtered_image_read( dmDescriptor* image, dmDescriptor* raw_image,
                              long* llpix, long* urpix, long* v_llpix,
                              long* v_urpix, void* image_buffer,
                              long image_buffer_size )
{
 dmBlock    *block  = NULL;
 dmDataset  *ds     = NULL;
 dmBuffer   *buffer = NULL;
 dmSubspace *rts    = NULL;
 dmKernel    kernel;
 dmDataType  type;

 char  *cast_buffer = NULL;
 long  *dpix        = NULL;
 long   byte_size   = 0;
 long   pnaxes      = 0;
 long   i           = 0;
 long   rrowno1     = 1;
 long   rrowno2     = 1;
 int    status      = 0;
 int    alloced     = 0;

 if ( !image )
   return;

 type = image->type;

 if ( rrowno1 !=1 || rrowno2 !=1 )
 {
  dmu_err_set( DM_ERR_INTERNAL, "Attempt to read beyond row 1 in image" );
  /* Carry on anyway in case the kernel can do something */
 }

 block  = (dmBlock*)image->block; 
 ds     = block->ds;
 buffer = block->buffer;
 kernel = ds->kernel;

 if ( buffer )
  rts = buffer->subspace;
 
 if ( block->type != dmIMAGE )
 {
/* 
 * Soon we may add support for opening a table column (eg PHA2 file) as an image.
 * In that case, however, we should use the table buffer for read and not come through
 * this code... I guess. 
 *  Imagine doing    
 *      dmcopy "pha2.fits[#row=2][bin counts]" pha1.img
 *  and having dm recognize that counts is an array column and should form the first
 *  axis of the image.
 */
  dmu_err_set( DM_ERR_INTERNAL, "Attempt to do an image read from non-image - not yet supported" );
 }

 if ( !v_llpix || !v_urpix )
 {
  if ( !llpix )
  {
   alloced = 1;
   v_llpix = calloc( image->naxes, UT_SZ_I );
   v_urpix = calloc( image->naxes, UT_SZ_I );
   for ( i = 0; i < image->naxes; i++ )
   {
    v_llpix[i] = 1;
    v_urpix[i] = image->axes[i];
   }  
  }
  else
  {
   v_llpix = llpix;
   v_urpix = urpix;
  }
 }

#if DEBUG
 dmverbose = 2;
#endif
 if ( dmverbose >= 2 )
 {
  char llbuf[80];
  char urbuf[80];
  dmu_print_pixel( llpix, image->naxes, llbuf );
  dmu_print_pixel( urpix, image->naxes, urbuf );
  printf( "Read image filtered pixel subset %s:%s\n", llbuf, urbuf ); 
  dmu_print_pixel( v_llpix, image->naxes, llbuf );
  dmu_print_pixel( v_urpix, image->naxes, urbuf );
  printf( "Read image raw pixel subset      %s:%s SIZE %ld\n", llbuf, urbuf, image_buffer_size );
 }

#if DEBUG
 dmverbose = 0;
#endif

 if ( raw_image->type != type )
 { 
#if DEBUG
printf( "CAST IMAGE\n" );
#endif
  byte_size = ar_mix_type_size( raw_image->type );
  cast_buffer = malloc( image_buffer_size * byte_size );   
 }
 else
 {
   cast_buffer = image_buffer;
 }

 pnaxes = raw_image->naxes;
 dpix = calloc( pnaxes, UT_SZ_I );

 for ( i =0; i <  pnaxes; i++ )
   dpix[i] = 1;

#if DEBUG
printf( "LOGICAL RANGE READ %ld,%ld : %ld,%ld BUFFER %p\n", v_llpix[0], v_llpix[1], v_urpix[0], v_urpix[1], image_buffer );
#endif


  /* Code change  - Bug 9023 - Decebmer 5, 2006                       */
  /*                                                                  */
  /* If a ColRead() functions exists for this kernel, use it.         */
  /* Otherwise, call the new function, dmp_filtered_image_row_read(), */
  /* instead to read image file data.                                 */

 if ( kernel->ColRead )
   status = kernel->ColRead( raw_image, v_llpix, v_urpix, dpix, rrowno1,
                             rrowno2, cast_buffer ); 
 else
 {
   status = dmp_filtered_image_row_read(raw_image, cast_buffer,
                                        v_llpix, v_urpix, rrowno1, rrowno2 );
 }


/* 
 * Care here: the rts region is in logical coords of the OLD image.
 * We must transfer this to the cast buffer
 * Note that the cast buffer always has datatype raw_image->type.
 *
 */
 if ( raw_image->naxes == 2 )
  dmi_image_filter_regions( image, raw_image, v_llpix, v_urpix, dpix, cast_buffer, rts );
 else
  dmi_image_filter_regions( image, raw_image, llpix, urpix, dpix, cast_buffer, rts );


/* Cast to destination type */
 if ( raw_image->type != type )
 {
  ar_mix_numeric_cast_array( cast_buffer, raw_image->type, type, image_buffer, image_buffer_size );
  free( cast_buffer );
 }


 if ( alloced )
 {
  free( v_llpix );
  free( v_urpix );
 }
 free( dpix );
}



/***************************************************************/
/*                                                             */
/* Function:   dmp_buffer_copy_image_data                      */
/*                                                             */
/* Description:                                                */
/*    Transfer image data values of source descriptor (dd1)    */
/*    to the destination descriptor (dd2)                      */
/*                                                             */
/* Input parameters:                                           */
/*    dmDescriptor *dd1   - source DM descriptor               */
/*    dmDescriptor *dd2   - destination DM descriptor          */
/*                                                             */
/* Output parameters:                                          */
/*                                                             */
/*                                                             */
/* Returns:                                                    */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmp_buffer_copy_image_data( dmDescriptor* dd1, dmDescriptor* dd2 )
{
 dmBlock   *block1 = NULL;  /* DM Block ptr (source)    */
 dmBuffer  *buffer = NULL;  /* source DM buffer         */
 char      *data   = NULL;  /* source image data values */

 long  nitems = 0;
 long  size   = 0;
 long  vrowno = 1;
 long  nvrows = 1;
 long  start  = 1;
 long  ngot   = 0;

 size   = dmi_descriptor_get_bytes( dd1 );
 nitems = dmi_descriptor_get_nvals( dd1 );
 block1 = (dmBlock*)dd1->block;
 buffer = block1->buffer;
 data = calloc( nitems, size );

 /*
  * --> Note by cstawarz, 08/09/2005 <--
  *
  * The code used to do this:
  *
  *   data = buffer->vdata;
  *   if ( !data )
  *   {
  *     data = calloc( nitems, size );
  *     buffer->vdata = data;
  *   }
  *   [data is not freed later]
  *
  * Not surprisingly, this lead to a memcpy to and from buffer->vdata down in
  * dmp_buffer_get_cast_data.  I wish I knew what whoever coded that was
  * thinking, because it seems nonsensical to me.  (Probably they were trying
  * to take advantage of some DM vaguery about when buffer->vdata is actually
  * used.)
  */

 /* NOTE:                                                                */
 /*   This function is geared to a column based kernel by providing the  */
 /*   data values for a particular column.  This method leads eventually */
 /*   to a kernel->ColWrite call.                                        */

 ngot = dmp_buffer_get_cast_data( dd1, data, dd1->type, vrowno, nvrows, start, nitems, NULL, NULL, NULL );
 ngot = dmp_buffer_set_cast_data( dd2, data, dd2->type, vrowno, nvrows, start, nitems );

 free( data );
 
}

/* 
 * These routines, not yet public, will allow users to
 * filter a row of data with a user defined filter.
 * Only works for tables at present.
 * In fact, probably doesn't work at all as it hasn't been debugged.
 * I just got far enough to convince myself that the structure of the
 * rest of the DM now supports what needs to be done (e.g. separation
 * of buffer calls from rest of block).
 * Example:
 *
 *
 * dmBuffer* buffer1;
 * dmBuffer* buffer2;
 *
 *  block = dmTableOpen( "my.fits[events]" );
 *  size = dmTableGetRowSize( block );
 *  data = malloc( size );
 *  nrows = dmTableGetNoRows( block );
 *  buffer1 = dmFilterCreate( block, "ccd_id=5,grade=0,2:4" );
 *  buffer2 = dmFilterCreate( block, "ccd_id=7,grade=2:6" );
 *  for ( i = 1; i<= nrows; i++ )
 *  {
 *    dmTableGetRow( block, data );
 *    ok1 = dmFilterRow( buffer1, data );
 *    ok2 = dmFilterRow( buffer2, data );
 *    ....
 *  }
 *  dmFilterFree( buffer1 );
 *  dmFilterFree( buffer2 );
 *  dmTableClose( block );
 */



/***************************************************************/
/*                                                             */
/* Function:   dmFilterFree                                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBuffer  *buffer                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmFilterFree( dmBuffer* buffer )
{
  dmi_buffer_free( buffer ); 
}



/***************************************************************/
/*                                                             */
/* Function:   dmFilterCreate                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock  *block                          */
/*                    char     *spec                           */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmBuffer  *buffer                                */
/*                                                             */
/***************************************************************/
dmBuffer* dmFilterCreate( dmBlock* block, char* spec )
{
 dmBuffer   *buffer = NULL;
 dmSubspace *rts    = NULL;

 rts    = dmp_block_parse_filter( block, spec );
 buffer = dmi_buffer_alloc();

 dmi_block_setup_runtime_filter( block, buffer, rts );
 dmi_virtual_buffer_reset( block, buffer );

 return buffer;
}



/***************************************************************/
/*                                                             */
/* Function:   dmFilterRow                                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBuffer *buffer                         */
/*                                                             */
/*  Output parameters:                                         */
/*                    void   *data                             */
/*                                                             */
/*  Returns:                                                   */
/*            dmBool   ok                                      */
/*                                                             */
/***************************************************************/
dmBool dmFilterRow( dmBuffer* buffer, void* data )
{
 long ngood = 0;
 dmBool ok;

 buffer->soffset = 0;
 buffer->rrow_seg1 = 1;
 buffer->rrow_seg2 = 1;
 ngood = dmi_filter_rows( buffer, data );
 ok = ( ngood > 0 );

 return ok;
}




/***************************************************************/
/*                                                             */
/* Function:   dmp_buffer_nullify_image_space_2d               */
/*                                                             */
/* Description:                                                */
/*    Nullify the space in the image that doesn't get touched  */
/*    by rebinning / filtering routines.                       */
/*    This is for 2D images only.                              */
/*                                                             */
/*    tmin/tmax describe a box.  Nullify everyting outside of  */
/*    the box.                                                 */
/*                                                             */
/*  Input parameters:                                          */
/*               void      *data                               */
/*               void      *nullval                            */
/*               dmDataType type                               */
/*               long      *image_axes                         */
/*               long      *min_offsets                        */
/*               long      *max_offsets                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmp_buffer_nullify_image_space_2d( void* data, void* nullval, dmDataType type, long* image_axes, long* min_offsets, long* max_offsets )
{
 void *data_ptr   = data;
 long  imrow      = 0;
 long  imcol      = 0;
 long  imrow0     = 0;
 long  imrow1     = 0;
 long  imrow2     = 0;
 long  imrow3     = 0;
 long  imcol0     = 0;
 long  imcol1     = 0;
 long  imcol2     = 0;
 long  imcol3     = 0;
 long  offset     = 0;
 long  middle     = 0;
 long  max_offset = 0;


 /* if there is no space around the edges, no need to do this */
 if ( min_offsets[0] == 0 && min_offsets[1] == 0 && max_offsets[0] == ( image_axes[0]-1 ) && max_offsets[1] == ( image_axes[1]-1 ))
  return;

 imcol0 = 0;
 imcol1 = min_offsets[0];
 imcol2 = max_offsets[0]+1;
 imcol3 = image_axes[0];
 imrow0 = 0;
 imrow1 = min_offsets[1];
 imrow2 = max_offsets[1]+1;
 imrow3 = image_axes[1];
 

 middle = imcol2 - imcol1;
 switch ( type )
 {
 case DT_UINT1:
  offset = 0;
  max_offset = imrow1 * image_axes[0];
  /* nullify the top */
  for ( offset = 0; offset < max_offset; offset++ )
  {
   ((unsigned char*)data_ptr)[offset] = *(unsigned char*)nullval;
  }

  for ( imrow = imrow1; imrow < imrow2; imrow++ )
  {
   /* nullify the left side */
   for ( imcol = 0; imcol < imcol1; imcol++, offset++ )
   {
    ((unsigned char*)data_ptr)[offset] = *(unsigned char*)nullval;
   }
   offset += middle;
   /* nullify the right side */
   for ( imcol = imcol2; imcol < imcol3; imcol++, offset++ )
   {
    ((unsigned char*)data_ptr)[offset] = *(unsigned char*)nullval;
   }
  }
  /* nullify the bottom */
  max_offset = image_axes[0] * image_axes[1];
  for ( ; offset < max_offset; offset++ )
  {
   ((unsigned char*)data_ptr)[offset] = *(unsigned char*)nullval;
  }
  break;
 case DT_UINT2:
  offset = 0;
  max_offset = imrow1 * image_axes[0];
  /* nullify the top */
  for ( offset = 0; offset < max_offset; offset++ )
  {
   ((unsigned short*)data_ptr)[offset] = *(unsigned short*)nullval;
  }
  for ( imrow = imrow1; imrow < imrow2; imrow++ )
  {
   /* nullify the left side */
   for ( imcol = 0; imcol < imcol1; imcol++ )
   {
    ((unsigned short*)data_ptr)[offset] = *(unsigned short*)nullval;
    offset++;
   }
   offset += middle;
   /* nullify the right side */
   for ( imcol = imcol2; imcol < imcol3; imcol++ )
   {
    ((unsigned short*)data_ptr)[offset] = *(unsigned short*)nullval;
    offset++;
   }
  }
  /* nullify the bottom */
  max_offset = image_axes[0] * image_axes[1];
  for ( ; offset < max_offset; offset++ )
  {
   ((unsigned short*)data_ptr)[offset] = *(unsigned short*)nullval;
  }
  break;
 case DT_UINT4:
  offset = 0;
  max_offset = imrow1 * image_axes[0];
  /* nullify the top */
  for ( offset = 0; offset < max_offset; offset++ )
  {
   ((unsigned long*)data_ptr)[offset] = *(unsigned long*)nullval;
  }
  for ( imrow = imrow1; imrow < imrow2; imrow++ )
  {
   /* nullify the left side */
   for ( imcol = 0; imcol < imcol1; imcol++ )
   {
    ((unsigned long*)data_ptr)[offset] = *(unsigned long*)nullval;
    offset++;
   }
   offset += middle;
   /* nullify the right side */
   for ( imcol = imcol2; imcol < imcol3; imcol++ )
   {
    ((unsigned long*)data_ptr)[offset] = *(unsigned long*)nullval;
    offset++;
   }
  }
  /* nullify the bottom */
  max_offset = image_axes[0] * image_axes[1];
  for ( ; offset < max_offset; offset++ )
  {
   ((unsigned long*)data_ptr)[offset] = *(unsigned long*)nullval;
  }
  break;
 case DT_INT2:
  offset = 0;
  max_offset = imrow1 * image_axes[0];
  /* nullify the top */
  for ( offset = 0; offset < max_offset; offset++ )
  {
   ((short*)data_ptr)[offset] = *(short*)nullval;
  }
  for ( imrow = imrow1; imrow < imrow2; imrow++ )
  {
   /* nullify the left side */
   for ( imcol = 0; imcol < imcol1; imcol++ )
   {
    ((short*)data_ptr)[offset] = *(short*)nullval;
    offset++;
   }
   offset += middle;
   /* nullify the right side */
   for ( imcol = imcol2; imcol < imcol3; imcol++ )
   {
    ((short*)data_ptr)[offset] = *(short*)nullval;
    offset++;
   }
  }
  /* nullify the bottom */
  max_offset = image_axes[0] * image_axes[1];
  for ( ; offset < max_offset; offset++ )
  {
   ((short*)data_ptr)[offset] = *(short*)nullval;
  }
  break;
 case DT_INT4:
  offset = 0;
  max_offset = imrow1 * image_axes[0];
  /* nullify the top */
  for ( offset = 0; offset < max_offset; offset++ )
  {
   ((long*)data_ptr)[offset] = *(long*)nullval;
  }
  for ( imrow = imrow1; imrow < imrow2; imrow++ )
  {
   /* nullify the left side */
   for ( imcol = 0; imcol < imcol1; imcol++ )
   {
    ((long*)data_ptr)[offset] = *(long*)nullval;
    offset++;
   }
   offset += middle;
   /* nullify the right side */
   for ( imcol = imcol2; imcol < imcol3; imcol++ )
   {
    ((long*)data_ptr)[offset] = *(long*)nullval;
    offset++;
   }
  }
  /* nullify the bottom */
  max_offset = image_axes[0] * image_axes[1];
  for ( ; offset < max_offset; offset++ )
  {
   ((long*)data_ptr)[offset] = *(long*)nullval;
  }
  break;
 case DT_REAL4:
  offset = 0;
  max_offset = imrow1 * image_axes[0];
  /* nullify the top */
  for ( offset = 0; offset < max_offset; offset++ )
  {
   ((float*)data_ptr)[offset] = *(float*)nullval;
  }
  for ( imrow = imrow1; imrow < imrow2; imrow++ )
  {
   /* nullify the left side */
   for ( imcol = 0; imcol < imcol1; imcol++ )
   {
    ((float*)data_ptr)[offset] = *(float*)nullval;
    offset++;
   }
   offset += middle;
   /* nullify the right side */
   for ( imcol = imcol2; imcol < imcol3; imcol++ )
   {
    ((float*)data_ptr)[offset] = *(float*)nullval;
    offset++;
   }
  }
  /* nullify the bottom */
  max_offset = image_axes[0] * image_axes[1];
  for ( ; offset < max_offset; offset++ )
  {
   ((float*)data_ptr)[offset] = *(float*)nullval;
  }
  break;
 case DT_REAL8:
  offset = 0;
  max_offset = imrow1 * image_axes[0];
  /* nullify the top */
  for ( offset = 0; offset < max_offset; offset++ )
  {
   ((double*)data_ptr)[offset] = *(double*)nullval;
  }
  for ( imrow = imrow1; imrow < imrow2; imrow++ )
  {
   /* nullify the left side */
   for ( imcol = 0; imcol < imcol1; imcol++ )
   {
    ((double*)data_ptr)[offset] = *(double*)nullval;
    offset++;
   }
   offset += middle;
   /* nullify the right side */
   for ( imcol = imcol2; imcol < imcol3; imcol++ )
   {
    ((double*)data_ptr)[offset] = *(double*)nullval;
    offset++;
   }
  }
  /* nullify the bottom */
  max_offset = image_axes[0] * image_axes[1];
  for ( ; offset < max_offset; offset++ )
  {
   ((double*)data_ptr)[offset] = *(double*)nullval;
  }
  break;
 default:
  break;
 }
}



/***************************************************************/
/*                                                             */
/* Function:   dmp_buffer_nullify_image_space_gen              */
/*                                                             */
/* Description:                                                */
/*    Nullify the space in the image that doesn't get touched  */
/*    by rebinning / filtering routines.                       */
/*                                                             */
/*     tmin/tmax describe a box.  Nullify everyting outside of */
/*     the box.                                                */
/*                                                             */
/*  Input parameters:                                          */
/*               void      *data                               */
/*               void      *nullval                            */
/*               dmDataType type                               */
/*               long      *image_axes                         */
/*               long      *min_offsets                        */
/*               long      *max_offsets                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmp_buffer_nullify_image_space_gen( void* data, void* nullval, dmDataType type, long* axes, long naxes, long* min_offsets, long* max_offsets )
{
 void   *data_ptr = data;
 short  *sptr   = NULL;
 float  *fptr   = NULL;
 double *dptr   = NULL;
 long   *lptr   = NULL;
 long   *weight = NULL;
 long   *opixel = NULL;
 long   *llpix  = NULL;
 long   *urpix  = NULL;

 unsigned char  *ucptr = NULL;
 unsigned short *usptr = NULL;
 unsigned long  *ulptr = NULL;

 double  dnull        = 0;
 float   fnull        = 0.0;
 long    lnull        = 0;
 long    axis_offset  = 0;
 long    offset       = 0;
 long    axis_weight  = 0;
 long    axis_offset2 = 0;
 long    offset2      = 0;
 long    pix          = 0;
 long    start_pix    = 0;
 long    end_pix      = 0;
 long    first_axis   = 0;
 long    last_axis    = 0;
 long    inc_axis     = 0;
 short   snull        = 0;

 unsigned long  ulnull  = 0;
 unsigned short usnull  = 0;
 unsigned char  ucnull  = '\0';

 dmBool isdefault = UT_TRUE;


 /* if there is no space around the edges, no need to do this */
 for ( axis_offset = 0; axis_offset < naxes; axis_offset++ )
 {
  if ( min_offsets[axis_offset] != 0 || max_offsets[axis_offset] != axes[axis_offset]-1 )
  {
   isdefault = UT_FALSE;
   break;
  }
 }
 if ( isdefault ) 
  return;


 weight = dmi_calc_image_weights( axes, naxes );
 opixel = malloc( naxes * UT_SZ_I );
 llpix = malloc( naxes * UT_SZ_I );
 urpix = malloc( naxes * UT_SZ_I );

/* loop over each axis' lower null range and upper null range.
 * Send a hyper-ray down each of the other dimensions to nullify
 * in each dimension.
 */
 switch ( type )
 {
 case DT_UINT1:
  ucptr = (unsigned char*)data_ptr;
  ucnull = *(unsigned char*)nullval;

  for ( axis_offset = 0; axis_offset < naxes; axis_offset++ )
  {
   axis_weight = weight[axis_offset];
   start_pix = 0;
   end_pix = min_offsets[axis_offset];
   last_axis = ( axis_offset == naxes-1 ) ? naxes-2 : naxes-1;
   first_axis = ( axis_offset == 0 ) ? 1 : 0;
   for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
   {
    llpix[axis_offset2] = min_offsets[axis_offset2];
    urpix[axis_offset2] = max_offsets[axis_offset2]+1;
   }
   for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
   {
    llpix[axis_offset2] = 0;
    urpix[axis_offset2] = axes[axis_offset2];
   }   

   /* loop over lower null range */
   for ( pix = start_pix; pix < end_pix; pix++ )
   {
    offset = pix * axis_weight;
    ucptr[offset] = ucnull;
    offset2 = offset;
    llpix[axis_offset] = pix;
    urpix[axis_offset] = pix+1;
    opixel[axis_offset] = pix;

    for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
    {
     opixel[axis_offset2] = llpix[axis_offset2];
     offset2 += llpix[axis_offset2] * weight[axis_offset2];
    }
    for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
     opixel[axis_offset2] = 0;

    inc_axis = 0;
    while( opixel[last_axis] < axes[last_axis] )
    {
     ucptr[offset2] = ucnull;
     opixel[first_axis]++;
     offset2 += weight[first_axis];
     inc_axis = first_axis;
     while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
     {
      offset2 += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );

      opixel[inc_axis] = llpix[inc_axis];
      inc_axis++;
      opixel[inc_axis]++;
     }
    }
   }

   start_pix = max_offsets[axis_offset]+1;
   end_pix = axes[axis_offset];
   /* loop over upper null range */
   for ( pix = start_pix; pix < end_pix; pix++ )
   {
    offset = pix * axis_weight;
    ucptr[offset] = ucnull;
    offset2 = offset;
    llpix[axis_offset] = pix;
    urpix[axis_offset] = pix+1;
    opixel[axis_offset] = pix;
    
    for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
    {
     opixel[axis_offset2] = llpix[axis_offset2];
     offset2 += llpix[axis_offset2] * weight[axis_offset2];
    }
    for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
     opixel[axis_offset2] = 0;

    inc_axis = 0;
    while( opixel[last_axis] < axes[last_axis] )
    {
     ucptr[offset2] = ucnull;
     opixel[first_axis]++;
     offset2 += weight[first_axis];
     inc_axis = first_axis;
     while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
     {
      offset2 += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );

      opixel[inc_axis] = llpix[inc_axis];
      inc_axis++;
      opixel[inc_axis]++;
     }
    }
   }
  } 

  break;
 case DT_UINT2:
  usptr = (unsigned short*)data_ptr;
  usnull = *(unsigned short*)nullval;

  for ( axis_offset = 0; axis_offset < naxes; axis_offset++ )
  {
   axis_weight = weight[axis_offset];
   start_pix = 0;
   end_pix = min_offsets[axis_offset];
   last_axis = ( axis_offset == naxes-1 ) ? naxes-2 : naxes-1;
   first_axis = ( axis_offset == 0 ) ? 1 : 0;
   for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
   {
    llpix[axis_offset2] = min_offsets[axis_offset2];
    urpix[axis_offset2] = max_offsets[axis_offset2]+1;
   }
   for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
   {
    llpix[axis_offset2] = 0;
    urpix[axis_offset2] = axes[axis_offset2];
   }   

   /* loop over lower null range */
   for ( pix = start_pix; pix < end_pix; pix++ )
   {
    offset = pix * axis_weight;
    usptr[offset] = usnull;
    offset2 = offset;
    llpix[axis_offset] = pix;
    urpix[axis_offset] = pix+1;    
    opixel[axis_offset] = pix;

    for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
    {
     opixel[axis_offset2] = llpix[axis_offset2];
     offset2 += llpix[axis_offset2] * weight[axis_offset2];
    }
    for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
     opixel[axis_offset2] = 0;

    inc_axis = 0;
    while( opixel[last_axis] < axes[last_axis] )
    {
     usptr[offset2] = usnull;
     opixel[first_axis]++;
     offset2 += weight[first_axis];
     inc_axis = first_axis;
     while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
     {
      offset2 += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );

      opixel[inc_axis] = llpix[inc_axis];
      inc_axis++;
      opixel[inc_axis]++;
     }
    }
   }

   start_pix = max_offsets[axis_offset]+1;
   end_pix = axes[axis_offset];
   /* loop over upper null range */
   for ( pix = start_pix; pix < end_pix; pix++ )
   {
    offset = pix * axis_weight;
    usptr[offset] = usnull;
    offset2 = offset;
    llpix[axis_offset] = pix;
    urpix[axis_offset] = pix+1;
    opixel[axis_offset] = pix;
    
    for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
    {
     opixel[axis_offset2] = llpix[axis_offset2];
     offset2 += llpix[axis_offset2] * weight[axis_offset2];
    }
    for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
     opixel[axis_offset2] = 0;

    inc_axis = 0;
    while( opixel[last_axis] < axes[last_axis] )
    {
     usptr[offset2] = usnull;
     opixel[first_axis]++;
     offset2 += weight[first_axis];
     inc_axis = first_axis;
     while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
     {
      offset2 += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );

      opixel[inc_axis] = llpix[inc_axis];
      inc_axis++;
      opixel[inc_axis]++;
     }
    }
   }
  } 

  break;
 case DT_UINT4:
  ulptr = (unsigned long*)data_ptr;
  ulnull = *(unsigned long*)nullval;

  for ( axis_offset = 0; axis_offset < naxes; axis_offset++ )
  {
   axis_weight = weight[axis_offset];
   start_pix = 0;
   end_pix = min_offsets[axis_offset];
   last_axis = ( axis_offset == naxes-1 ) ? naxes-2 : naxes-1;
   first_axis = ( axis_offset == 0 ) ? 1 : 0;
   for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
   {
    llpix[axis_offset2] = min_offsets[axis_offset2];
    urpix[axis_offset2] = max_offsets[axis_offset2]+1;
   }
   for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
   {
    llpix[axis_offset2] = 0;
    urpix[axis_offset2] = axes[axis_offset2];
   }   

   /* loop over lower null range */
   for ( pix = start_pix; pix < end_pix; pix++ )
   {
    offset = pix * axis_weight;
    ulptr[offset] = ulnull;
    offset2 = offset;
    llpix[axis_offset] = pix;
    urpix[axis_offset] = pix+1;
    opixel[axis_offset] = pix;
    
    for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
    {
     opixel[axis_offset2] = llpix[axis_offset2];
     offset2 += llpix[axis_offset2] * weight[axis_offset2];
    }
    for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
     opixel[axis_offset2] = 0;

    inc_axis = 0;
    while( opixel[last_axis] < axes[last_axis] )
    {
     ulptr[offset2] = ulnull;
     opixel[first_axis]++;
     offset2 += weight[first_axis];
     inc_axis = first_axis;
     while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
     {
      offset2 += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );

      opixel[inc_axis] = llpix[inc_axis];
      inc_axis++;
      opixel[inc_axis]++;
     }
    }
   }

   start_pix = max_offsets[axis_offset]+1;
   end_pix = axes[axis_offset];
   /* loop over upper null range */
   for ( pix = start_pix; pix < end_pix; pix++ )
   {
    offset = pix * axis_weight;
    ulptr[offset] = ulnull;
    offset2 = offset;
    llpix[axis_offset] = pix;
    urpix[axis_offset] = pix+1;
    opixel[axis_offset] = pix;
    
    for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
    {
     opixel[axis_offset2] = llpix[axis_offset2];
     offset2 += llpix[axis_offset2] * weight[axis_offset2];
    }
    for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
     opixel[axis_offset2] = 0;

    inc_axis = 0;
    while( opixel[last_axis] < axes[last_axis] )
    {
     ulptr[offset2] = ulnull;
     opixel[first_axis]++;
     offset2 += weight[first_axis];
     inc_axis = first_axis;
     while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
     {
      offset2 += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );

      opixel[inc_axis] = llpix[inc_axis];
      inc_axis++;
      opixel[inc_axis]++;
     }
    }
   }
  } 

  break;
 case DT_INT2:
  sptr = (short*)data_ptr;
  snull = *(short*)nullval;

  for ( axis_offset = 0; axis_offset < naxes; axis_offset++ )
  {
   axis_weight = weight[axis_offset];
   start_pix = 0;
   end_pix = min_offsets[axis_offset];
   last_axis = ( axis_offset == naxes-1 ) ? naxes-2 : naxes-1;
   first_axis = ( axis_offset == 0 ) ? 1 : 0;
   for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
   {
    llpix[axis_offset2] = min_offsets[axis_offset2];
    urpix[axis_offset2] = max_offsets[axis_offset2]+1;
   }
   for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
   {
    llpix[axis_offset2] = 0;
    urpix[axis_offset2] = axes[axis_offset2];
   }   

   /* loop over lower null range */
   for ( pix = start_pix; pix < end_pix; pix++ )
   {
    offset = pix * axis_weight;
    sptr[offset] = snull;
    offset2 = offset;
    llpix[axis_offset] = pix;
    urpix[axis_offset] = pix+1;
    opixel[axis_offset] = pix;
    
    for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
    {
     opixel[axis_offset2] = llpix[axis_offset2];
     offset2 += llpix[axis_offset2] * weight[axis_offset2];
    }
    for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
     opixel[axis_offset2] = 0;

    inc_axis = 0;
    while( opixel[last_axis] < axes[last_axis] )
    {
     sptr[offset2] = snull;
     opixel[first_axis]++;
     offset2 += weight[first_axis];
     inc_axis = first_axis;
     while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
     {
      offset2 += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );

      opixel[inc_axis] = llpix[inc_axis];
      inc_axis++;
      opixel[inc_axis]++;
     }
    }
   }

   start_pix = max_offsets[axis_offset]+1;
   end_pix = axes[axis_offset];
   /* loop over upper null range */
   for ( pix = start_pix; pix < end_pix; pix++ )
   {
    offset = pix * axis_weight;
    sptr[offset] = snull;
    offset2 = offset;
    llpix[axis_offset] = pix;
    urpix[axis_offset] = pix+1;
    opixel[axis_offset] = pix;
    
    for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
    {
     opixel[axis_offset2] = llpix[axis_offset2];
     offset2 += llpix[axis_offset2] * weight[axis_offset2];
    }
    for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
     opixel[axis_offset2] = 0;

    inc_axis = 0;
    while( opixel[last_axis] < axes[last_axis] )
    {
     sptr[offset2] = snull;
     opixel[first_axis]++;
     offset2 += weight[first_axis];
     inc_axis = first_axis;
     while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
     {
      offset2 += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );

      opixel[inc_axis] = llpix[inc_axis];
      inc_axis++;
      opixel[inc_axis]++;
     }
    }
   }
  } 

  break;
 case DT_INT4:
  lptr = (long*)data_ptr;
  lnull = *(long*)nullval;

  for ( axis_offset = 0; axis_offset < naxes; axis_offset++ )
  {
   axis_weight = weight[axis_offset];
   start_pix = 0;
   end_pix = min_offsets[axis_offset];
   last_axis = ( axis_offset == naxes-1 ) ? naxes-2 : naxes-1;
   first_axis = ( axis_offset == 0 ) ? 1 : 0;
   for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
   {
    llpix[axis_offset2] = min_offsets[axis_offset2];
    urpix[axis_offset2] = max_offsets[axis_offset2]+1;
   }
   for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
   {
    llpix[axis_offset2] = 0;
    urpix[axis_offset2] = axes[axis_offset2];
   }   

   /* loop over lower null range */
   for ( pix = start_pix; pix < end_pix; pix++ )
   {
    offset = pix * axis_weight;
    lptr[offset] = lnull;
    offset2 = offset;
    llpix[axis_offset] = pix;
    urpix[axis_offset] = pix+1;
    opixel[axis_offset] = pix;
    
    for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
    {
     opixel[axis_offset2] = llpix[axis_offset2];
     offset2 += llpix[axis_offset2] * weight[axis_offset2];
    }
    for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
     opixel[axis_offset2] = 0;

    inc_axis = 0;
    while( opixel[last_axis] < axes[last_axis] )
    {
     lptr[offset2] = lnull;
     opixel[first_axis]++;
     offset2 += weight[first_axis];
     inc_axis = first_axis;
     while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
     {
      offset2 += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );

      opixel[inc_axis] = llpix[inc_axis];
      inc_axis++;
      opixel[inc_axis]++;
     }
    }
   }

   start_pix = max_offsets[axis_offset]+1;
   end_pix = axes[axis_offset];
   /* loop over upper null range */
   for ( pix = start_pix; pix < end_pix; pix++ )
   {
    offset = pix * axis_weight;
    lptr[offset] = lnull;
    offset2 = offset;
    llpix[axis_offset] = pix;
    urpix[axis_offset] = pix+1;
    opixel[axis_offset] = pix;
    
    for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
    {
     opixel[axis_offset2] = llpix[axis_offset2];
     offset2 += llpix[axis_offset2] * weight[axis_offset2];
    }
    for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
     opixel[axis_offset2] = 0;

    inc_axis = 0;
    while( opixel[last_axis] < axes[last_axis] )
    {
     lptr[offset2] = lnull;
     opixel[first_axis]++;
     offset2 += weight[first_axis];
     inc_axis = first_axis;
     while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
     {
      offset2 += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );

      opixel[inc_axis] = llpix[inc_axis];
      inc_axis++;
      opixel[inc_axis]++;
     }
    }
   }
  } 

  break;
 case DT_REAL4:
  fptr = (float*)data_ptr;
  fnull = *(float*)nullval;

  for ( axis_offset = 0; axis_offset < naxes; axis_offset++ )
  {
   axis_weight = weight[axis_offset];
   start_pix = 0;
   end_pix = min_offsets[axis_offset];
   last_axis = ( axis_offset == naxes-1 ) ? naxes-2 : naxes-1;
   first_axis = ( axis_offset == 0 ) ? 1 : 0;
   for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
   {
    llpix[axis_offset2] = min_offsets[axis_offset2];
    urpix[axis_offset2] = max_offsets[axis_offset2]+1;
   }
   for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
   {
    llpix[axis_offset2] = 0;
    urpix[axis_offset2] = axes[axis_offset2];
   }   

   /* loop over lower null range */
   for ( pix = start_pix; pix < end_pix; pix++ )
   {
    offset = pix * axis_weight;
    fptr[offset] = fnull;
    offset2 = offset;
    llpix[axis_offset] = pix;
    urpix[axis_offset] = pix+1;
    opixel[axis_offset] = pix;
    
    for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
    {
     opixel[axis_offset2] = llpix[axis_offset2];
     offset2 += llpix[axis_offset2] * weight[axis_offset2];
    }
    for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
     opixel[axis_offset2] = 0;

    inc_axis = 0;
    while( opixel[last_axis] < axes[last_axis] )
    {
     fptr[offset2] = fnull;
     opixel[first_axis]++;
     offset2 += weight[first_axis];
     inc_axis = first_axis;
     while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
     {
      offset2 += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );

      opixel[inc_axis] = llpix[inc_axis];
      inc_axis++;
      opixel[inc_axis]++;
     }
    }
   }

   start_pix = max_offsets[axis_offset]+1;
   end_pix = axes[axis_offset];
   /* loop over upper null range */
   for ( pix = start_pix; pix < end_pix; pix++ )
   {
    offset = pix * axis_weight;
    fptr[offset] = fnull;
    offset2 = offset;
    llpix[axis_offset] = pix;
    urpix[axis_offset] = pix+1;
    opixel[axis_offset] = pix;
    
    for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
    {
     opixel[axis_offset2] = llpix[axis_offset2];
     offset2 += llpix[axis_offset2] * weight[axis_offset2];
    }
    for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
     opixel[axis_offset2] = 0;

    inc_axis = 0;
    while( opixel[last_axis] < axes[last_axis] )
    {
     fptr[offset2] = fnull;
     opixel[first_axis]++;
     offset2 += weight[first_axis];
     inc_axis = first_axis;
     while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
     {
      offset2 += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );

      opixel[inc_axis] = llpix[inc_axis];
      inc_axis++;
      opixel[inc_axis]++;
     }
    }
   }
  } 

  break;
 case DT_REAL8:
  dptr = (double*)data_ptr;
  dnull = *(double*)nullval;

  for ( axis_offset = 0; axis_offset < naxes; axis_offset++ )
  {
   axis_weight = weight[axis_offset];
   start_pix = 0;
   end_pix = min_offsets[axis_offset];
   last_axis = ( axis_offset == naxes-1 ) ? naxes-2 : naxes-1;
   first_axis = ( axis_offset == 0 ) ? 1 : 0;
   for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
   {
    llpix[axis_offset2] = min_offsets[axis_offset2];
    urpix[axis_offset2] = max_offsets[axis_offset2]+1;
   }
   for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
   {
    llpix[axis_offset2] = 0;
    urpix[axis_offset2] = axes[axis_offset2];
   }   

   /* loop over lower null range */
   for ( pix = start_pix; pix < end_pix; pix++ )
   {
    offset = pix * axis_weight;
    dptr[offset] = dnull;
    offset2 = offset;
    llpix[axis_offset] = pix;
    urpix[axis_offset] = pix+1;
    opixel[axis_offset] = pix;
    
    for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
    {
     opixel[axis_offset2] = llpix[axis_offset2];
     offset2 += llpix[axis_offset2] * weight[axis_offset2];
    }
    for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
     opixel[axis_offset2] = 0;

    inc_axis = 0;
    while( opixel[last_axis] < axes[last_axis] )
    {
     dptr[offset2] = dnull;
     opixel[first_axis]++;
     offset2 += weight[first_axis];
     inc_axis = first_axis;
     while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
     {
      offset2 += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );

      opixel[inc_axis] = llpix[inc_axis];
      inc_axis++;
      opixel[inc_axis]++;
     }
    }
   }

   start_pix = max_offsets[axis_offset]+1;
   end_pix = axes[axis_offset];
   /* loop over upper null range */
   for ( pix = start_pix; pix < end_pix; pix++ )
   {
    offset = pix * axis_weight;
    dptr[offset] = dnull;
    offset2 = offset;
    llpix[axis_offset] = pix;
    urpix[axis_offset] = pix+1;
    opixel[axis_offset] = pix;
    
    for ( axis_offset2 = 0; axis_offset2 < axis_offset; axis_offset2++ )
    {
     opixel[axis_offset2] = llpix[axis_offset2];
     offset2 += llpix[axis_offset2] * weight[axis_offset2];
    }
    for ( axis_offset2 = axis_offset+1; axis_offset2 < naxes; axis_offset2++ )
     opixel[axis_offset2] = 0;

    inc_axis = 0;
    while( opixel[last_axis] < axes[last_axis] )
    {
     dptr[offset2] = dnull;
     opixel[first_axis]++;
     offset2 += weight[first_axis];
     inc_axis = first_axis;
     while( opixel[inc_axis] >= urpix[inc_axis] && inc_axis < last_axis )
     {
      offset2 += weight[inc_axis] * ( axes[inc_axis] - urpix[inc_axis] + llpix[inc_axis] );

      opixel[inc_axis] = llpix[inc_axis];
      inc_axis++;
      opixel[inc_axis]++;
     }
    }
   }
  } 

  break;
 default:
  break;
 }
 free( opixel );
 free( weight );
 free( llpix );
 free( urpix );
}


/***************************************************************/
/*                                                             */
/* Function:   dmp_filtered_image_row_read                     */
/*                                                             */
/* Description:                                                */
/*    Called from dmp_filtered_image_read().                   */
/*    This routine reads the full IMAGE, when the full image   */
/*    is requested ( subarray defined by llpix/urpix is the    */
/*    same size as image ). When a subarray is requested,      */
/*    ( subarray size is smaller than full image size), a      */
/*    check is made to determine if the image had been read    */
/*    into the image block's buffer->data pointer is non-null. */
/*    If the buffer->data pointer is null, then the whole image*/
/*    is read and saved in that buffer. Then the  appropriate  */
/*    image pixels are extracted from the image buffer.        */
/*    The image or subimage is transferred to the output       */
/*    buffer, whose pointer is input parameter, data_buffer.   */
/*                                                             */
/*    Added for Bug 9023                                       */
/*                                                             */
/*    Modified, September 29, 2008  Enh 11292                  */
/*     Improve efficiency by not reading full image data with  */
/*     every pass through when returning a subarray data value.*/
/*                                                             */
/*    Modified, June 30, 2011  Bug 12273 side effect           */
/*      Bug 12273 added new function                           */ 
/*      dmp_buffer_subarray_extract(), that does much of the   */
/*      same work as this function when getting subarrays from */
/*      DTF files. This function was modified to set up values */
/*      and then call dmp_buffer_subarray_extract() to do the  */
/*      rest of the work, thus reducing duplicate code.        */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDescriptor *image      - DM image descriptor           */
/*    long         *llpix      - Subarray min. coord.          */
/*    long         *urpix      - Subarray max. coord.          */
/*    long          rrowno1    - row to read                   */
/*    long          rrowno2    -  s/b redundant with rrowno1   */
/*                                                             */
/*  Output parameters:                                         */
/*    char        *data_buffer                                 */
/*                                                             */
/*  Returns:                                                   */
/*            int status                                       */
/*                                                             */
/***************************************************************/
int  dmp_filtered_image_row_read(dmDescriptor *image,
				 char *data_buffer,
				 long *llpix, long *urpix,
				 long rrowno1, long rrowno2 )
{
  dmBlock    *block  = NULL;
  dmDataset  *ds     = NULL;
  dmBuffer   *buffer = NULL;
  dmKernel    kernel;

  dmBuffer  TmpBuffer;
  char     *pOutBuffer    = NULL;
  dmBool    FlagReadData  = FALSE;

  long     *bytesPerRow   = NULL;
  long      naxes         = 0;
  long      Length        = 0;
  long      Offset        = 0;
  long      rowOffset     = 0;
  long      SubArraySize  = 1;
  long      FullArraySize = 0;
  long      byte_size     = 0;
  int       IsSmaller     = 0; 
  int       status        = 0;
  int       ii            = 0;
 
  block  = (dmBlock*)image->block; 
  buffer = block->buffer;

  ds     = block->ds;
  kernel = ds->kernel;

  // Get size, in bytes, of each value.
  byte_size = ar_mix_type_size( image->type );

  // Does data_buffer hold the full image or just a cell section?

  // Determine the (sub)array size from bounding box.
  naxes = image->naxes;

  for (ii = 0; ii < naxes; ii++)
  {
    Length = urpix[ii] - llpix[ii] + 1;
    SubArraySize *= Length;
  }

  // Get full image size
  FullArraySize = dmGetCurrentArraySize(image);

  // If subarray size is same as array size, we are extracting the 
  // full image, use provided data buffer directly
  if (SubArraySize == FullArraySize)
  {
    IsSmaller = 0;
    TmpBuffer.data  = data_buffer;
  }
  else
  {
    // Extracting a subarray
    // Code Change  September 29, 2008  Enh 11292
    // Allocate enough space to do a complete read, only if the data
    // is not already present in the image block's data buffer

    IsSmaller = 1;

    if ( !buffer->data)
    {
      buffer->data = malloc ((FullArraySize + 1) * byte_size);
      FlagReadData = dmTRUE;
    }

    TmpBuffer.data = buffer->data;
  }

  TmpBuffer.cols    = buffer->cols;
  TmpBuffer.rrow1   = buffer->rrow1;
  TmpBuffer.ncols   = buffer->ncols;
  TmpBuffer.nrows   = 1;
  TmpBuffer.sizes   = buffer->sizes;
  TmpBuffer.nmax    = buffer->nmax;
  TmpBuffer.offsets = buffer->offsets;


  // Call afReadRows() to get data
  // Read full image only if requested or, if reading a subarray,
  //  read data only if not already in image block's data buffer
  //  ( buffer->data is null)

  if ( !IsSmaller || (IsSmaller && FlagReadData))
    status = kernel->ReadRows(block, &TmpBuffer, rrowno1, rrowno2);

  // If a subarray was requested, transfer data into buffer that 
  // was passed into this function

  if ( IsSmaller )
  {
   // Need to transfer desired data into output data buffer.
    // First determine offset into buffer

    pOutBuffer    = data_buffer;

    if ( naxes > 3 )
    {
      // This should work for naxis = 4; not bothering to
      //   verify for naxis > 4 at this time

      // Find offset to first cell
      // dmp_buffer_subarray_extracct expects input pointr at starting location

      bytesPerRow  = malloc( naxes *  UT_SZ_I);

      for ( ii = 0; ii < naxes; ii++)
      {
	if ( !ii )
          bytesPerRow[ii] = byte_size;
        else
          bytesPerRow[ii] = image->axes[ii-1] * bytesPerRow[ii-1];
      }

      for ( ii = 0; ii < naxes; ii++)
      {
        rowOffset += (((llpix[ii] - 1)) * bytesPerRow[ii]);
      }

      // Offset gets to the first desired data value to be transferred
      Offset    = rowOffset;

      free(bytesPerRow);
    }

    // Call dmp_buffer_subarray_extract() to do the rest of the work
    
    dmp_buffer_subarray_extract(image, llpix, urpix, ((char*)TmpBuffer.data + Offset), pOutBuffer);

  }
  return  status;
}


/***************************************************************/
/*                                                             */
/* Function:   dmp_buffer_subarray_extract                     */
/*                                                             */
/* Description:                                                */
/*    Called from dmp_buffer_read()                            */
/*    This routine reads a subarray of values held in the      */
/*    buffer and defined by llpix/urpix. Items read from the   */
/*    data buffer and returned in memory pointed to by optr.   */
/*                                                             */
/*    Added for Bug 12273                                      */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*    dmDescriptor *image      - DM image descriptor           */
/*    long         *llpix      - Subarray min. coord.          */
/*    long         *urpix      - Subarray max. coord.          */
/*    char         *inptr      - start of data buffer to read  */
/*    long          item_nbytes - number of items              */
/*                                                             */
/*  Output parameters:                                         */
/*    char        *optr                                        */
/*                                                             */
/*  Returns:                                                   */
/*            int status                                       */
/*                                                             */
/***************************************************************/
long dmp_buffer_subarray_extract( dmDescriptor *dd, long *llpix, long *urpix,
                             char *inptr, char *outptr )
{
  dmBool bUseThisCell = dmTRUE;

  void *pOutBuffer = outptr;
  void *pInput     = inptr;

  long *pAxes       = dd->axes; 
  long  naxes       = dd->naxes;
  long  startCell   = 0;
  long  endCell     = 0;
  long  bytesToRead = 0;
  long  byteSize    = 0;
  long  nItems      = 0;
  long  ii          = 0;
  long  jj          = 0;

  byteSize = dmi_descriptor_get_bytes( dd );

  switch (naxes )
  {
    case 0:
      break;

    case 1:
    case 2:
    case 3:
    {
      /* Most subarrays come from image of naxes = 2 or naxes = 3. */
      /* This section of code handles those cases                  */

      dmBlock *block   = dd->block;
      dmBuffer *buffer = block->buffer;

      long *bytesPerRow   = NULL;
      long  valuesPerRow  = 0;
      long  rowOffset     = 0;
      long  Offset        = 0;
      long  numberOfRows  = 0;
      long  numberOfPages = 0;
      long  pageOffset    = 0;
      long  COL           = 0;
      long  ROW           = 1;
      long  PAGE          = 2;
 

      // Offsets are calculated based on buffer start location

      pInput = (buffer->in_vdata ? buffer->vdata: buffer->data);

      // Set up byte offset array
      bytesPerRow = malloc( naxes * UT_SZ_I);

      for ( ii = 0; ii < naxes; ii++)
      {
	if ( !ii )
          bytesPerRow[ii] = byteSize;
        else
          bytesPerRow[ii] = pAxes[ii-1] * bytesPerRow[ii-1];
      }

      valuesPerRow = urpix[COL] - llpix[COL] + 1;
      bytesToRead  = valuesPerRow * byteSize;

      numberOfRows  = (naxes <= 1) ? 1 : (urpix[ROW]  - llpix[ROW]  + 1);
      numberOfPages = (naxes <= 2) ? 1 : (urpix[PAGE] - llpix[PAGE] + 1);

      // Read data values in clumps, starting with the first row
      // on the first "page", then the second row, etc. Next move onto 
      // additional "pages" if naxes > 2.

      for ( jj = 0; jj < numberOfPages; jj++)
      {
        // This gives the additional byte offset to get to the
        //  correct data page - necessary if naxes > 2

        if ( naxes > 2 )
          pageOffset = (((llpix[PAGE] - 1) + jj) * bytesPerRow[PAGE]);

        // Load data into the buffer one subarray "row" at a time

        for ( ii = 0; ii < numberOfRows; ii++)
	{
          // First get column offset value
          rowOffset = (llpix[COL] - 1) * bytesPerRow[COL];

          // Next get row offset value
          if ( naxes > 1)
            rowOffset += (((llpix[ROW] - 1) + ii) * bytesPerRow[ROW]); 
 
          // Offset gets to the first desired data value to be transferred
          Offset = rowOffset + pageOffset;
 
          // Copy "number of bytes to read" data values starting at the offset
          // location in the data buffer
 
          memcpy(pOutBuffer, (pInput + Offset), bytesToRead );

          // increment output pointer
          pOutBuffer += bytesToRead;
        }
      }

      free(bytesPerRow);
      break;
    }

    default:
    {
      /* NAXES > 3                   */
      /* Get and store single values */

      long *local_pix   = NULL;

      /* Set up local pix structure */
      local_pix = malloc( (naxes + 1) * UT_SZ_I);

      for ( ii = 0; ii < naxes; ii++ )
        local_pix[ii] = llpix[ii];

       /* Get number of items  */
       nItems = dmi_array_check_range( dd, llpix, urpix );

       /* First, get starting cell and ending cell numbers */

       startCell = dmu_pixelno_from_pixel(local_pix, pAxes, naxes);
       endCell   = dmu_pixelno_from_pixel( urpix, pAxes, naxes );

       /* Retrieve nItems individually */

      for (ii = 0; ii < nItems; ii++)
      {
        /* Get value */
        memcpy(pOutBuffer, pInput, byteSize);
 
        /* Adjust pointer to buffer for next row of data values */
        pOutBuffer += byteSize;

        /* Set up for next cell */
        bUseThisCell = dmFALSE;

        while ( !bUseThisCell && ( startCell < endCell) )
        {
          bUseThisCell = dmTRUE;

          /* Increment cell number */
          ++startCell;

          /* Increment input pointer */
          pInput += byteSize;

          /* Get the cell's pixel value. */
          dmu_pixel_from_pixelno( startCell, pAxes, naxes, local_pix);
 
          /* Verify that the next cell is in the subarray */
          for( jj= 0; jj < dd->naxes; jj++)
          {
            if ( local_pix[jj] > urpix[jj] || local_pix[jj] < llpix[jj] )
            {
               bUseThisCell = dmFALSE; 
               break;
            }
          }
        }  // end while
      } // end for ii loop
      free(local_pix);
      break;
    }
  }

  return dmSUCCESS;
}
