/*                                                                
**  Copyright (C) 2002-2003,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   miscdm.c                                                    */
/*                                                               */
/*   Miscellaneous functions for the DataModel library           */
/*                                                               */
/*                                                               */
/*****************************************************************/
#include "dmlibP.h"


static char* dm_optlist = NULL;
static long dmGInitialized = 0;   /* Static variable to spot initialization status */
static const char DM_VERSION[] = ASCDM_VERSION;



/***************************************************************/
/*                                                             */
/* Function: dmSetInternals                                    */
/*                                                             */
/* Description:                                                */
/*   Provide mechanism for user to set certain internal        */
/*   variable values.                                          */
/*                                                             */
/*  Input parameters:                                          */
/*     char* param   - tag for variable to set.                */
/*     void* value   - new value for field.. assumed to be     */ 
/*                      of the correct data type.              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*     int status  dmFAILURE/dmSUCCESS                         */
/*                                                             */
/***************************************************************/
int  dmSetInternals(char *param, void *lvalue)
{ 
  int retval = dmFAILURE;

  if ( !param )
    return retval;

  if ( cs_uceq( param, "buffersize" ))
  {
    dm_buffersize = *(long*)lvalue;
    retval = dmSUCCESS;
  }

  return retval;
}


/***************************************************************/
/*                                                             */
/* Function: dmGetVersion                                      */
/*                                                             */
/* Description:                                                */
/*   Write DataModel library version to provided string        */
/*                                                             */
/*  Input parameters:                                          */
/*     long maxlen   - maximum size of string                  */
/*                                                             */
/*  Output parameters:                                         */
/*     char* str     - String to receive version info          */
/*                                                             */
/*  Returns:                                                   */
/*     int dmSUCCESS                                           */
/*                                                             */
/***************************************************************/
int dmGetVersion(char *str, long maxlen)
{
  if ( str )
    cs_copy_siz( DM_VERSION, str, maxlen );

  return dmSUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function: dmGetError		                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*    dmErrCode  - Error status code                           */
/*                                                             */
/***************************************************************/
dmErrCode dmGetError( void )
{
  return dmu_err_get();
}

/***************************************************************/
/*                                                             */
/* Function: dmResetErrorState                                 */
/*                                                             */
/* Description:                                                */
/*   Clears error status codes                                 */
/*                                                             */
/*  Input parameters:                                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*                                                             */
/***************************************************************/
void dmResetErrorState( void )
{
  dmu_err_clear();
}

/***************************************************************/
/*                                                             */
/* Function: dmGetErrorMessage                                 */
/*                                                             */
/* Description:                                                */
/*   Write error message to provided string                    */
/*                                                             */
/*  Input parameters:                                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*                                                             */
/***************************************************************/
void dmGetErrorMessage( char* buf, long maxlen )
{
  dmu_err_get_message( buf, maxlen );
}

/***************************************************************/
/*                                                             */
/* Function: dmErrList                                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*                                                             */
/***************************************************************/
void dmErrList( void )
{
  dmu_err_status();
}

/***************************************************************/
/*                                                             */
/* Function: dmKernelSetOption                                 */
/*                                                             */
/* Description:                                                */
/*   Set or append input option(s) to DM internals             */
/*                                                             */
/*  Input parameters:                                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*                                                             */
/***************************************************************/
dmErrCode dmKernelSetOption( char* option )
{ 
  long ii;
  long slen;
  long nitems = 0;

  char **list = NULL;
  char  *item = NULL;

  dmErrCode status = dmSUCCESS;

  if ( cs_is_blank(option))
    return status;

  /* Split input options into individual tokens */
  list = cs_parse_comma_list( option, &nitems );

  for ( ii = 0; ii < nitems; ii++ )
  {
    item = list[ii];

    status = dmkKernelSetOption( NULL, item );
    if ( status == dmSUCCESS )
    {
      /* Add to DM internals option list */
      if ( dm_optlist )
      { 
	slen = strlen( dm_optlist ) + strlen( item ) + 2;
	dm_optlist = realloc( dm_optlist, slen );
	strcat( dm_optlist, "," );
	strcat( dm_optlist, item );
      }
      else 
      {
	dm_optlist = cs_dup( item );
      }
    }

    free( item );

  }  /* end for */

  free( list );


  return status;
}

/***************************************************************/
/*                                                             */
/* Function:  dmKernelGetOption                                */
/*                                                             */
/* Description:                                                */
/*   Returns current DM internal options list                  */
/*                                                             */
/*  Input parameters:                                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*                                                             */
/***************************************************************/
char* dmKernelGetOption( void )
{
  return dm_optlist;
}


/***************************************************************/
/*                                                             */
/* Function: dmInit                                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*                                                             */
/***************************************************************/
void dmInit( const int a, const void* b )
{
  /* Change this to stop keeping excessive back compatibility */
  long global_start = DM_GLOBAL_TSTART_RANGE | DM_GLOBAL_OLD;

  if ( !dmGInitialized )
  {  
    dmGInitialized++;
    dmp_init_func( (long)a, b );
  }

  dmverbose = 0;
  dm_global = global_start;
  dm_buffersize = dmBUFFER_ROWS_DEFAULT;    

}


/***************************************************************/
/*                                                             */
/* Function: dmp_init_func                                     */
/*                                                             */
/* Description:                                                */
/*   Used within dmInit                                        */
/*                                                             */
/*  Input parameters:                                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*                                                             */
/***************************************************************/
void dmp_init_func( const long a, const void* b )
{
  ut_init_jcmlib();
  dmu_err_init();

  /* Add IRAF error handling here */
  if ( a != 0 || b != NULL )
    dmu_err_set( DM_ERR_NULL, "Nontrivial arguments for dmInit not supported" );

  dmk_init();

  if ( dmverbose > 0 )
  {
    dmGetVersion( dm_errbuf, dmMAXSTR );
    fprintf( stderr, "CXC Data Model: DM V%s\n", dm_errbuf );
  }

  return;
}


/***************************************************************/
/*                                                             */
/* Function:   dmSetError                                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*                                                             */
/***************************************************************/
void   dmSetError( dmErrCode errCode, char* errStr )
{
  dmu_err_set( errCode, errStr );
}


/***************************************************************/
/*                                                             */
/* Function: dmSetVerbose                                      */
/*                                                             */
/* Description:                                                */
/*   Control error messages, default off                       */
/*   This is just a wrapper for the internal verbose routine.  */
/*                                                             */
/*  Input parameters:                                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*                                                             */
/***************************************************************/
void dmSetVerbose( const int value )
{
  dmuSetVerbose( value );
}


/***************************************************************/
/*                                                             */
/* Function:  dmBlockGetKernelKey                              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*                                                             */
/***************************************************************/
dmErrCode  dmBlockGetKernelKey( dmBlock* block, long keyno, char* buf, long maxlen )
{
  dmDataset* ds = NULL;
  dmKernel   kernel;

  cs_copy( " ", buf );

  if (block)
  {
    ds = block->ds;
    if ( ds )
      kernel = ds->kernel;

    if ( kernel )
      kernel->GetKernelKey( block, keyno, buf, maxlen );
  }

  return dmGetError();
}


/***************************************************************/
/*                                                             */
/* Function: dmBlockGetNoKernelKeys                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*  Input parameters:                                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*                                                             */
/***************************************************************/
long dmBlockGetNoKernelKeys( dmBlock* block )
{
  dmDataset* ds = NULL;
  dmKernel   kernel;
  long n = 0;

  if ( !block )
    return n;
  
  ds = block->ds;
  if ( ds )
    kernel = ds->kernel;

  if ( kernel )
    n = kernel->GetNoKernelKeys( block );

  return n;
}

/***************************************************************/
/*                                                             */
/* Function:  dmStrDup                                         */
/*                                                             */
/* Description:                                                */
/*   Returns a copy of the input string                        */
/*                                                             */
/*  Input parameters:                                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*                                                             */
/***************************************************************/
char* dmStrDup( char* a )
{
  return cs_dup(a);
}

/***************************************************************/
/*                                                             */
/* Function:  dmFileExists                                     */
/*                                                             */
/* Description:                                                */
/*   Check if specified file exists                            */
/*                                                             */
/*  Input parameters:                                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*                                                             */
/***************************************************************/
int dmFileExists( char* f )
{
  return fio_access( f );
}


/***************************************************************/
/*                                                             */
/* Function:  dmFree                                           */
/*                                                             */
/* Description:                                                */
/*   Simple free of DM allocated memory                        */
/*                                                             */
/*  Input parameters:                                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*                                                             */
/***************************************************************/
void dmFree( void* ptr )
{

  if (ptr)
    free( ptr );

}


/***************************************************************/
/*                                                             */
/* Function: dmFreeDescriptorArray                             */
/*                                                             */
/* Description:                                                */
/*    Obsolete                                                 */
/*                                                             */
/*  Input parameters:                                          */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*                                                             */
/***************************************************************/
void dmFreeDescriptorArray(dmDescriptor** ppDescriptors,long numDescriptors)
{
  /* Note that we do not actually free the individual elements */
  /* because the dmDescriptors are always stored in the dmBlock*/
  /* object, and will be freed at the time of the block close. */
  
  if (numDescriptors>0 && ppDescriptors != NULL)
    free(ppDescriptors);
}

