/*                                                                
**  Copyright (C) 2010  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/* *****************************************************************
 * FILE NAME:  cxcdm_dataset.c
 *
 * DEVELOPMENT: DataModel
 *
 * DESCRIPTION:  
 *   Provides a C-extension for Python to the DM functions 
 *   related to the dmDataset object.
 *
 *
 * REVISION HISTORY:
 *
 ***************************************************************** */
#include "Python.h"

/* ---------------------------------------------------------------------- */
/*                      *** dmDataset datatype ***                        */
/* ---------------------------------------------------------------------- */
static PyTypeObject dmDataset_Type = 
{
  PyObject_HEAD_INIT(NULL)
  0,                                    /*ob_size*/
  (char*)"dmDataset",                   /*tp_name*/
  sizeof(pyDataset),                    /*tp_basicsize*/
  0,                                    /*tp_itemsize*/
  (destructor)pyDataset_dealloc,        /*tp_dealloc*/
  0,                                    /*tp_print*/
  0,                                    /*tp_getattr*/
  0,                                    /*tp_setattr*/
  0,                                    /*tp_compare*/
  0,                                    /*tp_repr*/
  0,                                    /*tp_as_number*/
  0,                                    /*tp_as_sequence*/
  0,                                    /*tp_as_mapping*/
  0,                                    /*tp_hash */
  0,                                    /*tp_call*/
  0,                                    /*tp_str*/
  0,                                    /*tp_getattro*/
  0,                                    /*tp_setattro*/
  0,                                    /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT,                   /*tp_flags*/
  "dmDataset object",                   /* tp_doc */
  0,		                        /* tp_traverse */
  0,		                        /* tp_clear */
  0,		                        /* tp_richcompare */
  0,		                        /* tp_weaklistoffset */
  0,		                        /* tp_iter */
  0,		                        /* tp_iternext */
  dataset_methods,                      /* tp_methods */
  0,                                    /* tp_members */
  0,                                    /* tp_getset */
  0,                                    /* tp_base */
  0,                                    /* tp_dict */
  0,                                    /* tp_descr_get */
  0,                                    /* tp_descr_set */
  0,                                    /* tp_dictoffset */
  0,                                    /* tp_init */
  0,                                    /* tp_alloc */
  pyDataset_new,                        /* tp_new */

};

/* ---------------------------------------------------------------------- */
/* TypeFunc: pyDataset_new                                                */
/* ---------------------------------------------------------------------- */
static PyObject* pyDataset_new()
{
  pyDataset* self = PyObject_New( pyDataset, &dmDataset_Type);
  self->ds = NULL;

  return (PyObject*)self;
}

/* ---------------------------------------------------------------------- */
/* TypeFunc: pyDataset_dealloc                                            */
/* ---------------------------------------------------------------------- */
static void pyDataset_dealloc( pyDataset* self )
{
  CMMObject* node = NULL;
  dmDataset* ds   = self->ds;

  if ( cxcdm_verbose_level > 1 )
    printf("  VERBOSE: pyDataset_dealloc() called\n");

  /* ****   CXCDM MEMORY SEGMENT - BEGIN **** */
  node = cxcdm_memory_find( ds );
  if (node)
  {
    cxcdm_memory_release_holder( node, (PyObject*)self );
  }
  if ( cxcdm_verbose_level > 0 )
    cxcdm_memory_print();
  /* ****   CXCDM MEMORY SEGMENT - END   **** */

  PyObject_Del(self);
}

/* ---------------------------------------------------------------------- */
/*                     *** dmDataset functions ***                        */
/* ---------------------------------------------------------------------- */

/* ---------------------------------------------------------------------- */
/* Binding: _dmDatasetClose                                               */
/* ---------------------------------------------------------------------- */
static PyObject* _dmDatasetClose(PyObject *self, PyObject *args )
{
  char* funcName = "dmDatasetClose";

  PyObject* obj1   = NULL;

  dmErrCode status = dmFAILURE;

  /* Parse and verify input arguments */
  if (!PyArg_ParseTuple(args, "O!:dmDatasetClose", &dmDataset_Type, &obj1)){ return NULL;}

  /* Check underlying dmDataset pointer is available */
  if ( ((pyDataset*)obj1)->ds == NULL )
  {
    sprintf( errmsg, DMDS_EMPTY_PTR_MSG, funcName );
    PyErr_SetString(PyExc_ValueError, errmsg);
    return NULL;
  }
  else
  {
    status = dmDatasetClose( ((pyDataset*)obj1)->ds );
    if ( status != dmSUCCESS )
    {
      cxc_get_dm_errmsg( errmsg, dmMAXSTR, 0 );
      PyErr_SetString(PyExc_IOError, errmsg);
      return NULL;
    }

    /* Remove this object from memory management */
    cxcdm_memory_delete( (void*)((pyDataset*)obj1)->ds );
  }

  Py_INCREF(Py_None);
  return(Py_None);
}

/* ---------------------------------------------------------------------- */
/* Binding: _dmDatasetDestroy                                               */
/* ---------------------------------------------------------------------- */
static PyObject* _dmDatasetDestroy(PyObject *self, PyObject *args )
{
  char* funcName = "dmDatasetDestroy";

  char*      fname  = NULL;

  dmErrCode status = dmFAILURE;

  /* Parse and verify input arguments */
  if (!PyArg_ParseTuple(args, "s:dmDatasetDestroy", &fname)){ return NULL;}

  status = dmDatasetDestroy (fname);
  if ( status != dmSUCCESS )
  {
    cxc_get_dm_errmsg( errmsg, dmMAXSTR, 0 );
    PyErr_SetString(PyExc_IOError, errmsg);
    return NULL;
  }

  Py_INCREF(Py_None);
  return(Py_None);
}



/* ---------------------------------------------------------------------- */
/* Binding: _dmDatasetCreate                                               */
/* ---------------------------------------------------------------------- */
static PyObject* _dmDatasetCreate(PyObject *self, PyObject *args )
{
  char* funcName = "dmDatasetCreate";

  char*      fname  = NULL;
  dmDataset* _ds    = NULL;
  PyObject*  result = NULL;

  /* Parse and verify input arguments */
  if (!PyArg_ParseTuple(args, "s:dmDatasetCreate", &fname)){ return NULL;}


  _ds = dmDatasetCreate( fname );
  if ( _ds == NULL )
  {
    sprintf(errmsg, DMDS_CREATE_ERR_MSG ,funcName, fname);
    PyErr_SetString(PyExc_IOError, errmsg);
    return NULL;
  }

  result = pyDataset_new();
  ((pyDataset*)result)->ds = _ds;

  /* Register this object with memory management */
  cxcdm_memory_register( result );

  return result;
}

/* ---------------------------------------------------------------------- */
/* Binding: _dmDatasetGetBlockName                                        */
/* ---------------------------------------------------------------------- */
static PyObject* _dmDatasetGetBlockName(PyObject *self, PyObject *args )
{
  char* funcName = "dmDatasetGetBlockName";

  PyObject* obj1   = NULL;
  PyObject* result = NULL;

  dmErrCode status;

  long blockNo;
  char _name[dmMAXSTR+1];
  long maxlen = dmMAXSTR;

  /* Parse and verify input arguments */
  if (!PyArg_ParseTuple(args, "O!l:dmDatasetGetBlockName", &dmDataset_Type, &obj1, &blockNo)){ return NULL;}

  /* Check underlying dmBlock pointer is available */
  if ( ((pyDataset*)obj1)->ds == NULL )
  {
    sprintf( errmsg, DMDS_EMPTY_PTR_MSG, funcName );
    PyErr_SetString(PyExc_ValueError, errmsg);
  }
  else
  {
    status = dmDatasetGetBlockName( ((pyDataset*)obj1)->ds, blockNo, _name, maxlen );
    if ( status != dmSUCCESS )
    {
      cxc_get_dm_errmsg( errmsg, dmMAXSTR, 0 );
      PyErr_SetString(PyExc_RuntimeError, errmsg);
      return NULL;
    }
    result = PyString_FromString( _name );
  }

  return result;
}

/* ---------------------------------------------------------------------- */
/* Binding: _dmDatasetGetBlockType                                        */
/* ---------------------------------------------------------------------- */
static PyObject* _dmDatasetGetBlockType(PyObject *self, PyObject *args )
{
  char* funcName = "dmDatasetGetBlockType";

  PyObject* obj1   = NULL;
  PyObject* result = NULL;

  long blockNo;
  dmBlockType _res;

  /* Parse and verify input arguments */
  if (!PyArg_ParseTuple(args, "O!l:dmDatasetGetBlockType", &dmDataset_Type, &obj1, &blockNo)){ return NULL;}

  /* Check underlying dmDataset pointer is available */
  if ( ((pyDataset*)obj1)->ds == NULL )
  {
    sprintf( errmsg, DMDS_EMPTY_PTR_MSG, funcName );
    PyErr_SetString(PyExc_ValueError, errmsg);
  }
  else
  {
    _res = dmDatasetGetBlockType( ((pyDataset*)obj1)->ds, blockNo );

    result = pyBlockType_alloc();
    if ( pyBlockType_set( (pyBlockType*)result, (long)_res ) < 0 ) 
    {
      /* invalid type.. clear result and return error */
      Py_DECREF(result);
      return NULL;
    }
  }

  return result;
}

/* ---------------------------------------------------------------------- */
/* Binding: _dmDatasetGetCurrentBlockNo                                   */
/* ---------------------------------------------------------------------- */
static PyObject* _dmDatasetGetCurrentBlockNo(PyObject *self, PyObject *args )
{
  char* funcName = "dmDatasetGetCurrentBlockNo";

  PyObject* obj1   = NULL;
  PyObject* result = NULL;

  long _blkno;

  /* Parse and verify input arguments */
  if (!PyArg_ParseTuple(args, "O!:dmDatasetGetCurrentBlockNo", &dmDataset_Type, &obj1)){ return NULL;}

  /* Check underlying dmDataset pointer is available */
  if ( ((pyDataset*)obj1)->ds == NULL )
  {
    sprintf( errmsg, DMDS_EMPTY_PTR_MSG, funcName );
    PyErr_SetString(PyExc_ValueError, errmsg);
  }
  else
  {
    _blkno = dmDatasetGetCurrentBlockNo( ((pyDataset*)obj1)->ds );
    result = Py_BuildValue("l", _blkno );
  }

  return result;
}

/* ---------------------------------------------------------------------- */
/* Binding: _dmDatasetGetName                                             */
/* ---------------------------------------------------------------------- */
static PyObject* _dmDatasetGetName(PyObject *self, PyObject *args )
{
  char* funcName = "dmDatasetGetName";

  PyObject* obj1   = NULL;
  PyObject* result = NULL;

  dmErrCode status;

  char _fname[dmMAXSTR+1];
  long maxlen = dmMAXSTR;

  /* Parse and verify input arguments */
  if (!PyArg_ParseTuple(args, "O!:dmDatasetGetName", &dmDataset_Type, &obj1)){ return NULL;}

  /* Check underlying dmBlock pointer is available */
  if ( ((pyDataset*)obj1)->ds == NULL )
  {
    sprintf( errmsg, DMDS_EMPTY_PTR_MSG, funcName );
    PyErr_SetString(PyExc_ValueError, errmsg);
  }
  else
  {
    status = dmDatasetGetName( ((pyDataset*)obj1)->ds, _fname, maxlen );
    if ( status != dmSUCCESS )
    {
      cxc_get_dm_errmsg( errmsg, dmMAXSTR, 0 );
      PyErr_SetString(PyExc_RuntimeError, errmsg);
      return NULL;
    }
    result = PyString_FromString( _fname );
  }

  return result;
}

/* ---------------------------------------------------------------------- */
/* Binding: _dmDatasetGetNoBlocks                                         */
/* ---------------------------------------------------------------------- */
static PyObject* _dmDatasetGetNoBlocks(PyObject *self, PyObject *args )
{
  char* funcName = "dmDatasetGetNoBlocks";

  PyObject* obj1   = NULL;
  PyObject* result = NULL;

  long _nblocks;

  /* Parse and verify input arguments */
  if (!PyArg_ParseTuple(args, "O!:dmDatasetGetNoBlocks", &dmDataset_Type, &obj1)){ return NULL;}

  /* Check underlying dmBlock pointer is available */
  if ( ((pyDataset*)obj1)->ds == NULL )
  {
    sprintf( errmsg, DMDS_EMPTY_PTR_MSG, funcName );
    PyErr_SetString(PyExc_ValueError, errmsg);
  }
  else
  {
    _nblocks = dmDatasetGetNoBlocks( ((pyDataset*)obj1)->ds );
    result = Py_BuildValue("l", _nblocks );
  }

  return result;
}

/* ---------------------------------------------------------------------- */
/* Binding: _dmDatasetOpen                                                */
/* ---------------------------------------------------------------------- */
static PyObject* _dmDatasetOpen(PyObject *self, PyObject *args, PyObject *kwds )
{
  char* funcName = "dmDatasetOpen";

  PyObject*  obj1   = NULL;
  PyObject*  result = NULL;

  dmBool   update   = dmFALSE;
  char*      name   = NULL;
  dmDataset* _ds    = NULL;

  static char *kwlist[] =
    {"fname", "update", NULL};

  /* Parse and verify input arguments */
  if (!PyArg_ParseTupleAndKeywords(args, kwds, "s|O!:dmDatasetOpen", kwlist, &name, &PyBool_Type, &obj1)){ return NULL;}

  /* Process update argument for boolean value */
  if ( obj1 &&  PyObject_IsTrue(obj1) ){ update = dmTRUE; }

  if ( update )
    _ds = dmDatasetOpenUpdate( name );
  else
    _ds = dmDatasetOpen( name );

  if ( _ds == NULL )
  {
    sprintf(errmsg, DMDS_OPEN_ERR_MSG ,funcName, name);
    PyErr_SetString(PyExc_IOError, errmsg );
    return NULL;
  }

  result = pyDataset_new();
  ((pyDataset*)result)->ds = _ds;

  /* Register this object with memory management */
  cxcdm_memory_register( result );

  return result;
}

/* ---------------------------------------------------------------------- */
/* Binding: _dmDatasetWriteable                                           */
/* ---------------------------------------------------------------------- */
static PyObject* _dmDatasetWriteable(PyObject *self, PyObject *args )
{
  char* funcName = "dmDatasetWriteable";

  PyObject* obj1   = NULL;
  PyObject* result = NULL;

  dmBool _res;

  /* Parse and verify input arguments */
  if (!PyArg_ParseTuple(args, "O!:dmDatasetWriteable", &dmDataset_Type, &obj1)){ return NULL;}

  /* Check underlying dmBlock pointer is available */
  if ( ((pyDataset*)obj1)->ds == NULL )
  {
    sprintf( errmsg, DMDS_EMPTY_PTR_MSG, funcName );
    PyErr_SetString(PyExc_ValueError, errmsg);
  }
  else
  {
    _res = dmDatasetWriteable( ((pyDataset*)obj1)->ds );
    result = PyBool_FromLong( (long)_res );
  }

  return result;
}


/* ---------------------------------------------------------------------- */
/* Binding: _dmDatasetAccess                                              */
/* ---------------------------------------------------------------------- */
static PyObject* _dmDatasetAccess(PyObject *self, PyObject *args )
{
  char* funcName = "dmDatasetAccess";
  char *fname = NULL, *mode = NULL; 
  PyObject* result = NULL;
  dmBool _res;

  /* Parse and verify input arguments */
  if (!PyArg_ParseTuple(args, "ss:dmDatasetAccess", &fname, &mode)){ return NULL;}

  /* Check underlying dmBlock pointer is available */
  if (fname == NULL || (fname && strlen(fname) == 0))
  {
    sprintf( errmsg, DMDS_GENERIC_MSG, funcName, "could not access file- no filename specified." );
    PyErr_SetString(PyExc_ValueError, errmsg);
  }
  else if (mode && (strcmp(mode, "R") == 0 || strcmp(mode, "RW") == 0))
  {
    _res = dmDatasetAccess( fname, mode );
    result = PyBool_FromLong( (long)_res );
  }
  else
  {
    sprintf( errmsg, DMDS_GENERIC_MSG, funcName, "mode must be either R|RW." );
    PyErr_SetString(PyExc_ValueError, errmsg);
  }
  return result;
}

/* ---------------------------------------------------------------------- */
/* Binding: _dmDatasetDelete                                              */
/* ---------------------------------------------------------------------- */
static PyObject* _dmDatasetDelete(PyObject *self, PyObject *args )
{
  char* funcName = "dmDatasetDelete";

  PyObject* obj1   = NULL;

  /* Parse and verify input arguments */
  if (!PyArg_ParseTuple(args, "O!:dmDatasetDelete", &dmDataset_Type, &obj1)){ return NULL;}

  /* Check underlying dmDataset pointer is available */
  if ( ((pyDataset*)obj1)->ds == NULL )
  {
    sprintf( errmsg, DMDS_EMPTY_PTR_MSG, funcName );
    PyErr_SetString(PyExc_ValueError, errmsg);
    return NULL;
  }
  else
  {
    dmDatasetDelete(((pyDataset*)obj1)->ds); 
    /* Remove this object from memory management */
    cxcdm_memory_delete( (void*)((pyDataset*)obj1)->ds );
  }
  Py_INCREF(Py_None);
  return(Py_None);
}

