static PyObject*  _dmIsEmpty(PyObject *self, PyObject *args);
static PyObject*  _dmSetVerbose(PyObject *self, PyObject *args);
static PyObject* _dmTanPixToWorld(PyObject *self, PyObject *args, PyObject *kwds);
static PyObject* _dmTanWorldToPix(PyObject *self, PyObject *args, PyObject *kwds);

/* Error Messages */
#define DMMISC_INVALID_ARG_MSG     "%s() invalid argument type."
#define DMMISC_INVALID_ARGVAL_MSG  "%s() invalid argument value."


static PyObject*  _dmIsEmpty(PyObject *self, PyObject *args)
{
  char* funcName = "dmIsEmpty";

  PyObject* obj1   = NULL;
  PyObject* result = NULL;

  pyDataset*    _ds = NULL;
  pyBlock*      _bl = NULL;
  pyDescriptor* _dd = NULL;

  result = Py_False;

  /* Parse and verify input arguments */
  if (!PyArg_ParseTuple(args, "O:dmIsEmpty", &obj1)){ return NULL;}


  if( PyObject_TypeCheck( obj1, &dmDataset_Type) )
  {
    /* Object is a dmDataset struct */
    _ds = (pyDataset*)obj1;

    /* Check underlying dmDataset pointer is available */
    if ( _ds->ds == NULL )
    {
      Py_RETURN_TRUE;
    }
  }
  else if( PyObject_TypeCheck( obj1, &dmBlock_Type) )
  {
    /* Object is a dmBlock struct */
    _bl = (pyBlock*)obj1;

    /* Check underlying dmBlock pointer is available */
    if ( _bl->block == NULL )
    {
      Py_RETURN_TRUE;
    }
  }
  else if( PyObject_TypeCheck( obj1, &dmDescriptor_Type) )
  {
    /* Object is a dmDescriptor struct */
    _dd = (pyDescriptor*)obj1;

    /* Check underlying dmDescriptor pointer is available */
    if ( _dd->dd == NULL )
    {
      Py_RETURN_TRUE;
    }
  }
  else
  {
    sprintf( errmsg, DMMISC_INVALID_ARG_MSG, funcName );
    PyErr_SetString(PyExc_TypeError, errmsg );
    return NULL;
  }
  if (result == Py_False)
  {
    Py_RETURN_FALSE;
  }
  return result;
}


static PyObject*  _dmSetVerbose(PyObject *self, PyObject *args)
{
  long verbose = 0;

  /* Parse and verify input arguments */
  if (!PyArg_ParseTuple(args, "l:dmSetVerbose", &verbose)){ return NULL;}

  cxcdm_verbose_level = verbose;

  Py_RETURN_NONE;
}


/* ---------------------------------------------------------------------- */
/* Binding: _dmGetVersion                                                 */
/* ---------------------------------------------------------------------- */
static PyObject* _dmGetVersion(void)
{
  char val[dmMAXSTR];
  PyObject*  result  = NULL;

  if (dmGetVersion(val, dmMAXSTR) != dmSUCCESS)
  {
    cxc_get_dm_errmsg( errmsg, dmMAXSTR, 0 );
    PyErr_SetString(PyExc_RuntimeError, errmsg);
    return NULL;
  }
  result = Py_BuildValue("O", PyString_FromString(val));
  return result;
}

/* ---------------------------------------------------------------------- */
/* Binding: _dmGetErrorMessage                                            */
/* ---------------------------------------------------------------------- */
static PyObject* _dmGetErrorMessage(void)
{
  char val[dmMAXSTR];
  PyObject*  result  = NULL;

  val[0] = '\0';
  dmGetErrorMessage(val, dmMAXSTR);
  result = Py_BuildValue("O", PyString_FromString(val));
  return result;
}


/* ---------------------------------------------------------------------- */
/* Binding: _dmResetErrorState                                            */
/* ---------------------------------------------------------------------- */
static PyObject* _dmResetErrorState(void)
{
  PyObject*  result  = NULL;
  dmResetErrorState();
  Py_RETURN_NONE;
}

/* ---------------------------------------------------------------------- */
/* Binding: _dmTanPixToWorld                                              */
/* ---------------------------------------------------------------------- */
static PyObject* _dmTanPixToWorld(PyObject *self, PyObject *args, PyObject *kwds)
{
  char* funcName = "dmTanPixToWorld";

  PyArrayObject *obj1 = NULL, *obj2 = NULL, *obj3 = NULL, *obj4 = NULL;

  PyArrayObject* obj_pix = NULL;
  PyArrayObject* obj_asp = NULL;
  PyArrayObject* obj_crpix = NULL;
  PyArrayObject* obj_cdelt = NULL;
  PyArrayObject* obj_world = NULL;

  PyArray_Dims   dims = {NULL, 0};
  double *pix = NULL, *asp = NULL, *crpix = NULL, *cdelt = NULL, *world = NULL;

  /* Parse and verify input arguments */
  static char *kwlist[] = {"pix", "asp", "crpix", "cdelt", NULL};

  if (!PyArg_ParseTupleAndKeywords(args, kwds, "O&O&O&O&:dmTanPixToWorld",
                                   kwlist,
                                   &PyArray_Converter, &obj1,
                                   &PyArray_Converter, &obj2,
                                   &PyArray_Converter, &obj3,
                                   &PyArray_Converter, &obj4)){ return NULL;}

  /* validate input arrays */
  if ( obj1 != NULL )
  {
    /* This contains the pix array of doubles[2] */
    if ( PyArray_NDIM( obj1 ) > 1 )
    {
      sprintf( errmsg, "%s() pix must be 1D array.", funcName );
      PyErr_SetString(PyExc_ValueError, errmsg);
      goto fail;
    }
    else if (PyArray_SIZE( obj1 ) != 2)
    {
      sprintf( errmsg, "%s() pix is expected to be of length 2.", funcName );
      PyErr_SetString(PyExc_ValueError, errmsg);
      goto fail;
    }
    else
    {
      obj_pix = (PyArrayObject*)PyArray_Cast (obj1, NPY_DOUBLE);
      pix = PyArray_DATA (obj_pix);
    }
  }
  if ( obj2 != NULL )
  {
    /* This contains the asp array of doubles[3] (aspect ra, dec, roll) */
    if ( PyArray_NDIM( obj2 ) > 1 )
    {
      sprintf( errmsg, "%s() asp must be 1D array.", funcName );
      PyErr_SetString(PyExc_ValueError, errmsg);
      goto fail;
    }
    else if (PyArray_SIZE( obj2 ) != 3)
    {
      sprintf( errmsg, "%s() asp is expected to be of length 3.", funcName );
      PyErr_SetString(PyExc_ValueError, errmsg);
      goto fail;
    }
    else
    {
      obj_asp = (PyArrayObject*)PyArray_Cast (obj2, NPY_DOUBLE);
      asp = PyArray_DATA (obj_asp);
    }
  }
  if ( obj3 != NULL )
  {
    /* This contains the crpix array of doubles[2] */
    if ( PyArray_NDIM( obj3 ) > 1 )
    {
      sprintf( errmsg, "%s() crpix must be 1D array.", funcName );
      PyErr_SetString(PyExc_ValueError, errmsg);
      goto fail;
    }
    else if (PyArray_SIZE( obj3 ) != 2)
    {
      sprintf( errmsg, "%s() crpix is expected to be of length 2.", funcName );
      PyErr_SetString(PyExc_ValueError, errmsg);
      goto fail;
    }
    else
    {
      obj_crpix = (PyArrayObject*)PyArray_Cast (obj3, NPY_DOUBLE);
      crpix = PyArray_DATA (obj_crpix);
    }
  }
  if ( obj4 != NULL )
  {
    /* This contains the cdelt array of doubles[2] */
    if ( PyArray_NDIM( obj4 ) > 1 )
    {
      sprintf( errmsg, "%s() cdelt must be 1D array.", funcName );
      PyErr_SetString(PyExc_ValueError, errmsg);
      goto fail;
    }
    else if (PyArray_SIZE( obj4 ) != 2)
    {
      sprintf( errmsg, "%s() cdelt is expected to be of length 2.", funcName );
      PyErr_SetString(PyExc_ValueError, errmsg);
      goto fail;
    }
    else
    {
      obj_cdelt = (PyArrayObject*)PyArray_Cast (obj4, NPY_DOUBLE);
      cdelt = PyArray_DATA (obj_cdelt);
    }
  }
  dims.ptr = PyArray_DIMS( obj_pix );
  dims.len = PyArray_NDIM( obj_pix );
  obj_world = (PyArrayObject*) PyArray_ZEROS(dims.len, dims.ptr, cxc_get_py_type( dmDOUBLE ), 0 );

  world = PyArray_DATA(obj_world); 

  dmTanPixToWorld(pix, asp, crpix, cdelt, world);  

  if (obj_pix)  { Py_DECREF ( obj_pix );}
  if (obj_asp)  { Py_DECREF ( obj_asp ); }
  if (obj_crpix){ Py_DECREF ( obj_crpix ); }
  if (obj_cdelt){ Py_DECREF ( obj_cdelt ); }

  return PyArray_Return( obj_world );

 fail:
  if (obj_pix)  { Py_DECREF ( obj_pix );}
  if (obj_asp)  { Py_DECREF ( obj_asp ); }
  if (obj_crpix){ Py_DECREF ( obj_crpix ); }
  if (obj_cdelt){ Py_DECREF ( obj_cdelt ); }

  return NULL;
}
 
/* ---------------------------------------------------------------------- */
/* Binding: _dmTanWorldToPix                                              */
/* ---------------------------------------------------------------------- */
static PyObject* _dmTanWorldToPix(PyObject *self, PyObject *args, PyObject *kwds)
{
  char* funcName = "dmTanWorldToPix";

  PyArrayObject *obj1 = NULL, *obj2 = NULL, *obj3 = NULL, *obj4 = NULL;
  PyArrayObject* obj_world = NULL;
  PyArrayObject* obj_asp = NULL;
  PyArrayObject* obj_crpix = NULL;
  PyArrayObject* obj_cdelt = NULL;
  PyArrayObject* obj_pix = NULL;
  PyArray_Dims   dims = {NULL, 0};

  double  *world = NULL, *asp = NULL, *crpix = NULL, *cdelt = NULL, *pix = NULL;

  /* Parse and verify input arguments */
  static char *kwlist[] = {"world", "asp", "crpix", "cdelt", NULL};

  if (!PyArg_ParseTupleAndKeywords(args, kwds, "O&O&O&O&:dmTanWorldToPix",
                                   kwlist,
                                   &PyArray_Converter, &obj1,
                                   &PyArray_Converter, &obj2,
                                   &PyArray_Converter, &obj3,
                                   &PyArray_Converter, &obj4)){ return NULL;}

  /* validate input arrays */
  if ( obj1 != NULL )
  {
    /* This contains the world array of doubles[2] */
    if ( PyArray_NDIM( obj1 ) > 1 )
    {
      sprintf( errmsg, "%s() world must be 1D array.", funcName );
      PyErr_SetString(PyExc_ValueError, errmsg);
      goto fail;
    }
    else if (PyArray_SIZE( obj1 ) != 2)
    {
      sprintf( errmsg, "%s() world is expected to be of length 2.", funcName );
      PyErr_SetString(PyExc_ValueError, errmsg);
      goto fail;
    }
    else
    {
      obj_world = (PyArrayObject*)PyArray_Cast (obj1, NPY_DOUBLE);
      world = PyArray_DATA (obj_world);
    }
  }
  if ( obj2 != NULL )
  {
    /* This contains the asp array of doubles[3] (aspect ra, dec, roll) */
    if ( PyArray_NDIM( obj2 ) > 1 )
    {
      sprintf( errmsg, "%s() asp must be 1D array.", funcName );
      PyErr_SetString(PyExc_ValueError, errmsg);
      goto fail;
    }
    else if (PyArray_SIZE( obj2 ) != 3)
    {
      sprintf( errmsg, "%s() asp is expected to be of length 3.", funcName );
      PyErr_SetString(PyExc_ValueError, errmsg);
      goto fail;
    }
    else
    {
      obj_asp = (PyArrayObject*)PyArray_Cast (obj2, NPY_DOUBLE);
      asp = PyArray_DATA (obj_asp);
    }
  }
  if ( obj3 != NULL )
  {
    /* This contains the crpix array of doubles[2] */
    if ( PyArray_NDIM( obj3 ) > 1 )
    {
      sprintf( errmsg, "%s() crpix must be 1D array.", funcName );
      PyErr_SetString(PyExc_ValueError, errmsg);
      goto fail;
    }
    else if (PyArray_SIZE( obj3 ) != 2)
    {
      sprintf( errmsg, "%s() crpix is expected to be of length 2.", funcName );
      PyErr_SetString(PyExc_ValueError, errmsg);
      goto fail;
    }
    else
    {
      obj_crpix = (PyArrayObject*)PyArray_Cast (obj3, NPY_DOUBLE);
      crpix = PyArray_DATA (obj_crpix);
    }
  }
  if ( obj4 != NULL )
  {
    /* This contains the cdelt array of doubles[2] */
    if ( PyArray_NDIM( obj4 ) > 1 )
    {
      sprintf( errmsg, "%s() cdelt must be 1D array.", funcName );
      PyErr_SetString(PyExc_ValueError, errmsg);
      goto fail;
    }
    else if (PyArray_SIZE( obj4 ) != 2)
    {
      sprintf( errmsg, "%s() cdelt is expected to be of length 2.", funcName );
      PyErr_SetString(PyExc_ValueError, errmsg);
      goto fail;
    }
    else
    {
      obj_cdelt = (PyArrayObject*)PyArray_Cast (obj4, NPY_DOUBLE);
      cdelt = PyArray_DATA (obj_cdelt);
    }
  }
  dims.ptr = PyArray_DIMS( obj_world );
  dims.len = PyArray_NDIM( obj_world );
  obj_pix = (PyArrayObject*) PyArray_ZEROS(dims.len, dims.ptr, cxc_get_py_type( dmDOUBLE ), 0 );
  pix = PyArray_DATA(obj_pix); 

  dmTanWorldToPix(world, asp, crpix, cdelt, pix);  

  if (obj_world){ Py_DECREF ( obj_world );}
  if (obj_asp)  { Py_DECREF ( obj_asp ); }
  if (obj_crpix){ Py_DECREF ( obj_crpix ); }
  if (obj_cdelt){ Py_DECREF ( obj_cdelt ); }

  return PyArray_Return( obj_pix );

 fail:
  if (obj_world){ Py_DECREF ( obj_world );}
  if (obj_asp)  { Py_DECREF ( obj_asp ); }
  if (obj_crpix){ Py_DECREF ( obj_crpix ); }
  if (obj_cdelt){ Py_DECREF ( obj_cdelt ); }

  return NULL;
}
 



