/*                                                                
**  Copyright (C) 2010  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/* *****************************************************************
 * FILE NAME:  cxcdm_mm.c
 *
 * DEVELOPMENT: DataModel
 *
 * DESCRIPTION:  
 *
 * REVISION HISTORY:
 *
 ***************************************************************** */


/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_new_cmmobj                                          */
/* ---------------------------------------------------------------------- */
CMMObject* cxcdm_memory_new_cmmobj( void )
{
  CMMObject* res = calloc( 1, sizeof(CMMObject) );

  res->cptr      = NULL;
  res->type      = NULL;
  res->parent    = NULL;
  res->children  = NULL;
  res->holders   = NULL;
  res->refcnt    = 0;

  return res;
}

/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_new_cmmll                                           */
/* ---------------------------------------------------------------------- */
CMMLL* cxcdm_memory_new_cmmll( void )
{
  CMMLL* res = calloc( 1, sizeof(CMMLL) );

  res->cobj = NULL;
  res->pobj = NULL;
  res->next = NULL;

  return res;
}

/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_add                                                 */
/* ---------------------------------------------------------------------- */
void cxcdm_memory_add( CMMObject* node )
{
  CMMLL* item = NULL;

  if ( node == NULL )
    return;

  if ( cxcdm_verbose_level > 1 )
    printf("  VERBOSE: cxcdm_memory_add() - add %s node to master list.\n", node->type->tp_name);

  item = cxcdm_memory_new_cmmll();
  item->cobj = node;

  /* add the new item to the beginning of the master list */
  item->next = cxcdm_memory_master;
  cxcdm_memory_master = item;

}

/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_decref                                              */
/* ---------------------------------------------------------------------- */
void cxcdm_memory_decref( CMMObject* node )
{
  CMMObject* parent = NULL;

  if ( node == NULL )
    return;

  if ( node->parent )
    parent = node->parent->cobj;

  if ( cxcdm_verbose_level > 1 )
    printf("  VERBOSE: cxcdm_memory_decref() - decrement %s node reference count.\n", node->type->tp_name );

  /* decrement counter on this node */
  node->refcnt--;

  if ( ( node->refcnt < 1 ) && ( node->cptr != NULL ) )
  {
    if ( cxcdm_verbose_level > 1 )
      printf("  VERBOSE: cxcdm_memory_decref() - close %s object.\n", node->type->tp_name );
    
    /* The last python reference to the object is gone and it is still open. */
    /* close it out and remove from master list.                             */
    if ( node->type == &dmDataset_Type )
    {
      dmDatasetClose( (dmDataset*)node->cptr );
    }
    else if ( node->type == &dmBlock_Type )
    {
      dmBlockClose( (dmBlock*)node->cptr );
    }
    else if ( node->type == &dmDescriptor_Type )
    {
      /* nothing to close */
    }

    /* Free this node and its decendents */
    /* Note: if cptr is NULL, we are already in a remove thread, so  */
    /*       we only remove the node when cptr is alive.             */
    /* Note: the sequencing of actions should have the lowest branch */
    /*       reach zero first, so each 'tree' should only be 1 layer */
    /*       deep.  use 'tree' function to cut this from parent.     */
    cxcdm_memory_remove_tree( node );

  }

  /* also decrement counter on parent */
  if ( parent )
    cxcdm_memory_decref( parent );


  return;
}

/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_delete                                              */
/* ---------------------------------------------------------------------- */
void cxcdm_memory_delete( void* cptr )
{
  CMMObject* node = NULL;

  if ( cxcdm_verbose_level > 1 )
    printf("  VERBOSE: cxcdm_memory_delete() - Delete C ptr node.\n");

  node = cxcdm_memory_find( cptr );
  if ( node != NULL )
  {
    cxcdm_memory_remove_tree( node );
  }

  if ( cxcdm_verbose_level > 0 )
    cxcdm_memory_print();

  return;
}

/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_find                                                */
/* ---------------------------------------------------------------------- */
CMMObject* cxcdm_memory_find( void* cptr )
{
  CMMLL* this = cxcdm_memory_master;   /* set base of search */
  CMMObject* res = NULL;

  if ( cptr == NULL )
    return res;

  while ( this != NULL )
  {
    if ( this->cobj->cptr == cptr )
    {
      res = this->cobj;
      break;
    }
    this = this->next;
  }

  if ( cxcdm_verbose_level > 1 )
  {
    if ( res )
      printf("  VERBOSE: cxcdm_memory_find() - Found C ptr in %s node.\n", res->type->tp_name);
    else
      printf("  VERBOSE: cxcdm_memory_find() - Did not find C ptr.\n");
  }

  return res;
}

/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_incref                                              */
/* ---------------------------------------------------------------------- */
void cxcdm_memory_incref( CMMObject* node )
{
  if ( node == NULL )
    return;

  if ( cxcdm_verbose_level > 1 )
    printf("  VERBOSE: cxcdm_memory_incref() - increment %s node reference count.\n", node->type->tp_name );

  /* increment counter on this node */
  node->refcnt++;

  /* also increment counter on parent */
  if ( node->parent )
    cxcdm_memory_incref( node->parent->cobj );

  return;
}

/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_nullify                                             */
/* ---------------------------------------------------------------------- */
void cxcdm_memory_nullify( CMMObject* node )
{
  if ( node == NULL )
    return;

  if ( cxcdm_verbose_level > 1 )
    printf("  VERBOSE: cxcdm_memory_nullify() - set %s node cptr references to NULL.\n", node->type->tp_name );

  /* increment counter on this node */
  node->cptr = NULL;

  if ( node->holders != NULL )
  {
    cxcdm_memory_empty_holders( node );
  }

  return;
}

/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_print                                               */
/* ---------------------------------------------------------------------- */
void cxcdm_memory_print( void )
{
  CMMLL* this = NULL;
  CMMObject* node = NULL;
  CMMLL* item = NULL;
  long ii;
  long nparents;
  long nchildrn;
  long nholders;
  char type[15];

  printf("  CXCDM Memory Management\n");
  printf("   node     type       nParents   nChildren   nHolders   nDependents\n");
  this = cxcdm_memory_master;
  ii = 0;
  while ( this != NULL )
  {
    ii++;
    nparents = 0;
    nchildrn = 0;
    nholders = 0;

    node = this->cobj;

    item = node->parent;
    while ( item != NULL ){ nparents++; item = item->next;}

    item = node->children;
    while ( item != NULL ){ nchildrn++; item = item->next;}

    item = node->holders;
    while ( item != NULL ){ nholders++; item = item->next;}

    if ( node->type == &dmDataset_Type )
      sprintf(type, "dmDataset");
    else if ( node->type == &dmBlock_Type )
      sprintf(type, "dmBlock");
    else if ( node->type == &dmDescriptor_Type )
      sprintf(type, "dmDescriptor");

    printf("   %4ld %12s %5ld %14ld %11ld %11ld \n",ii, type, nparents, nchildrn, nholders, node->refcnt );

    this = this->next;
  }
  if ( ii == 0 )
    printf("    <empty> \n");

  return;
}

/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_register                                            */
/* ---------------------------------------------------------------------- */
void cxcdm_memory_register( PyObject* obj )
{
  void*  cptr = NULL;
  CMMObject* node = NULL;
  PyTypeObject* type = NULL;

  if ( PyObject_TypeCheck( obj, &dmDataset_Type) )
  {
    cptr = ((pyDataset*)obj)->ds;
    type = &dmDataset_Type;
  }
  else if ( PyObject_TypeCheck( obj, &dmBlock_Type) )
  {
    cptr = ((pyBlock*)obj)->block;
    type = &dmBlock_Type;
  }
  else if ( PyObject_TypeCheck( obj, &dmDescriptor_Type) )
  {
    cptr = ((pyDescriptor*)obj)->dd;
    type = &dmDescriptor_Type;
  }
  if ( type == NULL )
    return;

  if ( cxcdm_verbose_level > 1 )
    printf("  VERBOSE: cxcdm_memory_register() - register %s ptr\n", type->tp_name);

  node = cxcdm_memory_find( cptr );
  if ( node == NULL )
  {
    /* make a new node for this ptr */
    node = cxcdm_memory_new_cmmobj();
    node->cptr    = cptr;
    node->type    = type;

    /* link in any parent */
    cxcdm_memory_add_parent( node );

    /* add to master list */
    cxcdm_memory_add( node );
  }
  cxcdm_memory_add_holder( node, obj );

  if ( cxcdm_verbose_level > 0 )
    cxcdm_memory_print();

  return;
}

/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_remove                                              */
/* ---------------------------------------------------------------------- */
void cxcdm_memory_remove( CMMObject* node )
{
  CMMLL* this = NULL;

  if ( node == NULL )
    return;

  if ( cxcdm_verbose_level > 1 )
    printf("  VERBOSE: cxcdm_memory_remove() - remove %s node and decendents.\n", node->type->tp_name);

  if ( cxcdm_memory_unlink( node ) == 0 )
  {
    if ( cxcdm_verbose_level > 1 )
      printf("  VERBOSE: cxcdm_memory_remove() - already removed.\n");
    return;  /* already removed */
  }

  /* Set references to the cptr to NULL (in the node and in any  */
  /* python objects which may still be open.. so the user cannot */
  /* access free'd memory space.                                 */
  cxcdm_memory_nullify( node );

  /* Remove children of this node.. they are not valid if the parent is not */
  while ( node->children != NULL )
  {
    /* pop this child off list so it is no longer visible */
    this = node->children;
    node->children = this->next;

    /* remove the child node */
    cxcdm_memory_remove( this->cobj );
    free(this);
  }

  /* Release the holders list.. */
  cxcdm_memory_release_holders( node );

  /* Remove link to parent. */
  if ( node->parent != NULL )
  {
    free(node->parent);
    node->parent = NULL;
  }      

  /* free the node structure */
  if ( cxcdm_verbose_level > 1 )
    printf("  VERBOSE: cxcdm_memory_remove() - Free %s node.\n", node->type->tp_name);
  free(node);

  return;
}

/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_remove_tree                                         */
/* ---------------------------------------------------------------------- */
void cxcdm_memory_remove_tree( CMMObject* node )
{
  CMMObject* parent = NULL;

  if ( node == NULL )
    return;

  if ( node->parent )
    parent = node->parent->cobj;

  /* Remove this node from child list of parent. */
  if ( parent )
  {
    cxcdm_memory_release_child(parent, node );
  }

  /* Remove this node and all decendents */
  cxcdm_memory_remove( node );

  return;
}


/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_unlink                                              */
/* ---------------------------------------------------------------------- */
int cxcdm_memory_unlink( CMMObject* node )
{
  int     res = 0;
  CMMLL* prev = NULL;
  CMMLL* this = NULL;

  if ( node == NULL )
    return res;

  if ( cxcdm_verbose_level > 1 )
    printf("  VERBOSE: cxcdm_memory_unlink() - cut %s node from master list.\n", node->type->tp_name);

  this = cxcdm_memory_master;
  while ( this != NULL )
  {
    if ( this->cobj == node )
    {
      if (prev == NULL)
      {
	/* node is first on list.. */
	cxcdm_memory_master = this->next;
      }
      else
      {
	/* link previous node to next on list.. */
	prev->next = this->next;
      }
      break;
    }
    prev = this;
    this = this->next;
  }

  if ( this != NULL )
  {
    res = 1;

    /* free the list struct containing this node */
    free(this); 
  }

  return res;
}


/* ------------------------------------------------------------------------------ */
/* ------------------------------------------------------------------------------ */

/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_add_child                                           */
/* ---------------------------------------------------------------------- */
void cxcdm_memory_add_child( CMMObject* node, CMMObject* child )
{
  CMMLL* item;

  if ( ( node == NULL ) || ( child == NULL ) )
    return;

  if ( cxcdm_verbose_level > 1 )
    printf("  VERBOSE: cxcdm_memory_add_child() - add child to %s node.\n", node->type->tp_name);

  /* make new list element for the object */
  item = cxcdm_memory_new_cmmll();
  item->cobj = child;

  /* put new object at head of list..  */
  item->next = node->children;
  node->children = item;

  return;
}

/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_add_holder                                          */
/* ---------------------------------------------------------------------- */
void cxcdm_memory_add_holder( CMMObject* node, PyObject* obj )
{
  CMMLL*  item;

  if (( node == NULL )||( obj == NULL ))
    return;

  if ( cxcdm_verbose_level > 1 )
    printf("  VERBOSE: cxcdm_memory_add_holder() - add holder to %s node.\n", node->type->tp_name);

  item = cxcdm_memory_new_cmmll();
  item->pobj = obj;

  /* put new object at head of list..  */
  item->next    = node->holders;
  node->holders = item;

  cxcdm_memory_incref( node );

  return;
}


/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_add_parent                                          */
/* ---------------------------------------------------------------------- */
void cxcdm_memory_add_parent( CMMObject* node )
{
  CMMLL*  item;

  void* cptr         = NULL;
  CMMObject*  parent = NULL;
  PyTypeObject* type = NULL;

  if ( node == NULL )
    return;

  if ( cxcdm_verbose_level > 1 )
    printf("  VERBOSE: cxcdm_memory_add_parent() - add parent to %s node.\n", node->type->tp_name);

  if ( node->type == &dmBlock_Type )
  {
    cptr = (void*)dmBlockGetDataset( (dmBlock*)(node->cptr) );
    type = &dmDataset_Type;
  }
  else if ( node->type == &dmDescriptor_Type )
  {
    cptr = (void*)dmDescriptorGetBlock( (dmDescriptor*)(node->cptr) );
    type = &dmBlock_Type;
  }

  if (cptr == NULL)
  {
    /* object has no parent.. nothing to do */
    return;
  }

  /* object has a parent.. */
  parent = cxcdm_memory_find( cptr );
  if ( parent == NULL )
  {
    parent = cxcdm_memory_new_cmmobj();
    parent->cptr = cptr;
    parent->type = type;

    /* link in any parent.. recursive ???? */
    /* cxcdm_memory_add_parent( node );    */

    /* add to master list */
    cxcdm_memory_add( parent );
  }

  /* attach parent to this node there is not one already. */
  if ( node->parent == NULL )
  {
    item = cxcdm_memory_new_cmmll();
    item->cobj   = parent;
    item->next   = node->parent;
    node->parent = item;
  }

  /* Add this node to child list of parent */
  cxcdm_memory_add_child( parent, node );
    
  return;
}

/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_empty_holders                                       */
/* ---------------------------------------------------------------------- */
void cxcdm_memory_empty_holders( CMMObject* node )
{
  CMMLL*  this;
  PyObject* obj;

  if ( node == NULL )
    return;

  if ( cxcdm_verbose_level > 1 )
    printf("  VERBOSE: cxcdm_memory_empty_holders() - set C ptr within python objects to NULL.\n");

  this = node->holders;
  while ( this != NULL )
  {
    obj = this->pobj;
    if ( node->type == &dmDataset_Type )
    {
      ((pyDataset*)obj)->ds = NULL;
    }
    else if ( node->type == &dmBlock_Type )
    {
      ((pyBlock*)obj)->block = NULL;
    }
    else if ( node->type == &dmDescriptor_Type )
    {
      ((pyDescriptor*)obj)->dd = NULL;
    }
    this = this->next;
  }

  return;
}

/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_release_child                                       */
/* ---------------------------------------------------------------------- */
void cxcdm_memory_release_child( CMMObject* node, CMMObject* child )
{
  CMMLL* prev = NULL;
  CMMLL* this = NULL;

  if (( node == NULL )||( child == NULL) )
    return;

  if ( cxcdm_verbose_level > 1 )
    printf("  VERBOSE: cxcdm_memory_release_child() - remove child from children list of %s node.\n", node->type->tp_name);

  this = node->children;
  while ( this != NULL )
  {  
    if ( this->cobj == child )
    {
      if (prev == NULL)
      {
	/* node is first on list.. */
	node->children = this->next;
      }
      else
      {
	/* link previous node to next on list.. */
	prev->next = this->next;
      }
      break;
    }
    prev = this;
    this = this->next;
  }
  if ( this != NULL )
    free(this);  /* free the link struct */

  return;
}

/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_release_holders                                     */
/* ---------------------------------------------------------------------- */
void cxcdm_memory_release_holders( CMMObject* node )
{
  CMMLL* this;

  if ( node == NULL )
    return;

  if ( cxcdm_verbose_level > 1 )
    printf("  VERBOSE: cxcdm_memory_free_holders() - release %s node holder list.\n", node->type->tp_name);

  while ( node->holders != NULL )
  {
    this = node->holders;
    node->holders = this->next;
    free( this );

    cxcdm_memory_decref( node );
  }
  node->holders = NULL;

  return;
}

/* ---------------------------------------------------------------------- */
/* Func: cxcdm_memory_release_holder                                      */
/* ---------------------------------------------------------------------- */
void cxcdm_memory_release_holder( CMMObject* node, PyObject* obj )
{
  CMMLL* prev = NULL;
  CMMLL* this = NULL;

  if (( node == NULL )||( obj == NULL) )
    return;

  if ( cxcdm_verbose_level > 1 )
    printf("  VERBOSE: cxcdm_memory_release_holder() - remove python object from holder list of %s node.\n", node->type->tp_name);

  this = node->holders;
  while ( this != NULL )
  {  
    if ( this->pobj == obj )
    {
      if (prev == NULL)
      {
	/* node is first on list.. */
	node->holders = this->next;
      }
      else
      {
	/* link previous node to next on list.. */
	prev->next = this->next;
      }
      break;
    }
    prev = this;
    this = this->next;
  }

  if ( this != NULL )
  {
    cxcdm_memory_decref( node );

    /* free the list struct containing this node */
    free(this);
  }

  return;
}

