static void         cxc_get_dm_errmsg( char* errstr, long maxlen, int append );
static dmDataType   cxc_get_dm_type( int intype );
static int          cxc_get_py_type( dmDataType intype );
static void         cxc_get_dimarr( dmDescriptor* dd, PyArray_Dims* dims, long *nrows );
static void*        cxc_strings_to_carray( PyArrayObject* obj );


/* ---------------------------------------------------------------------- */
/*  cxc_get_dm_errmsg()                                                   */
/*    Helper function extract DM level error message.                     */
/*     - may process DM message for CXCDM purposes                        */
/*     - may append to existing input message (using append arg)          */
/*     - will reset the DM Error status when done.                        */
/* ---------------------------------------------------------------------- */
static void cxc_get_dm_errmsg( char* errstr, long maxlen, int append )
{
  char* cptr = NULL;
  char  tmpstr[dmMAXSTR+1];
  long  tmpmax = dmMAXSTR;
  long  nleft;

  if (append)
  {
    /* appending DM message to existing message.. add newline first. */
    /* start DM message after newline                                */
    tmpstr[0] = '\n';
    cptr = &tmpstr[1];
    tmpmax--;
  }
  else
  {
    /* clear input string memory     */
    memset( errstr, 0, maxlen );

    /* start DM message at beginning */
    cptr = &tmpstr[0];
  }
  
  /* how many chars left on input string? */
  nleft = maxlen - strlen(errstr);
  if ( nleft > 0 )
  {
    /* Get the DM error message */
    dmGetErrorMessage(cptr, tmpmax);

    /* Any editing of DM error message would go here.. using cptr */

    /* Put DM message on provided string */
    strncat( errstr, (const char*)tmpstr, nleft );
  }

  /* Clear the DM error message state*/
  dmResetErrorState();

  return;
}


/* ---------------------------------------------------------------------- */
/*  cxc_get_dm_type()                                                     */
/*    Helper function to convert from numpy data types to DM data types   */
/* ---------------------------------------------------------------------- */
static dmDataType cxc_get_dm_type( int intype )
{
    dmDataType retval;

    switch ( intype ){
    case NPY_VOID:
      retval = DT_VOID;
      break;
    case NPY_SHORT:
      retval = dmSHORT;
      break;
    case NPY_LONG:
      retval = dmLONG;
      break;
    case NPY_INT:
      retval = dmLONG;
      break;
    case NPY_FLOAT:
      retval = dmFLOAT;
      break;
    case NPY_DOUBLE:
      retval = dmDOUBLE;
      break;
    case NPY_STRING:
      retval = dmTEXT;
      break;
    case NPY_BYTE:
      retval = dmBYTE;
      break;
    case NPY_BOOL:
      retval = dmBOOL;
      break;
    case NPY_USHORT:
      retval = dmUSHORT;
      break;
    case NPY_ULONG:
      retval = dmULONG;
      break;
    case NPY_UBYTE:
      retval = dmBIT;
      break;
    default:
      retval = dmUNKNOWNTYPE;
      break;
    }
    return retval;
}

/* ---------------------------------------------------------------------- */
/*  cxc_get_py_type()                                                     */
/*    Helper function to convert from DM data types to numpy data types   */
/* ---------------------------------------------------------------------- */
static int cxc_get_py_type( dmDataType intype )
{
    int retval;

    switch ( intype ){
    case DT_VOID:
      retval = NPY_VOID;
      break;
    case dmSHORT:
      retval = NPY_SHORT;
      break;
    case dmLONG:
      retval = NPY_LONG;
      break;
    case dmFLOAT:
      retval = NPY_FLOAT;
      break;
    case dmDOUBLE:
      retval = NPY_DOUBLE;
      break;
    case dmTEXT:
      retval = NPY_STRING;
      break;
    case dmBYTE:
    case dmBIT:
      retval = NPY_UBYTE;
      break;
    case dmBOOL:
      retval = NPY_BOOL;
      break;
    case dmUSHORT:
      retval = NPY_USHORT;
      break;
    case dmULONG:
      retval = NPY_ULONG;
      break;
    default:
      retval = NPY_NOTYPE;
      break;
    }
    return retval;
}

/* ---------------------------------------------------------------------- */
/*  cxc_get_dimarr()                                                      */
/*    Helper function to convert DM descriptor dimarray to the            */
/*    Python PyArray_Dims structure.                                      */
/*    Scalars are returned with NULL array and 0 size.                    */
/* ---------------------------------------------------------------------- */
static void cxc_get_dimarr( dmDescriptor* dd, PyArray_Dims* dims, long *nrows )
{
  long  ii;
  long  cpt_ndim=0;
  long  ndim;
  long* dimarr = NULL;
  long* cpt_dimarr = NULL;
  intp* tmp    = NULL;
  long  vdim;
  int   index;
  

  if ( (dd == NULL) || (dims == NULL) ){ return; }

  /* Get number of dimensions in descriptor */
  ndim = dmGetArrayDim( dd );
  vdim = dmGetElementDim( dd );

  dims->len = 0;

  if ( nrows)
  {
    if ( *nrows < 1 )  *nrows = 1;

    /* Count the size of the dim array */
    if ( *nrows > 1 )
      dims->len++;
  }


  if ( vdim > 1 )
  {

    // This code assumes that all components will have the same shape
    // Therefore it is only necessary to check the first component.

    dims->len++;

    dmDescriptor *cpt = dmGetCpt(dd, 0);
    cpt_ndim = dmGetArrayDim( cpt );

    if (cpt_ndim > 0)
    {
      dmGetArrayDimensions( cpt, &cpt_dimarr);
    }

    dims->len += cpt_ndim;
  }
  else if (ndim > 0)
    dims->len += (int)ndim;

  // create dim array
  tmp    = calloc( dims->len, sizeof(intp) );
  index = 0;

  if (nrows && ( *nrows > 1 ))
  {
    tmp[index++] = *nrows;
  }

  if ( vdim > 1 )
  {
    tmp[index++] = vdim;
    for ( ii = cpt_ndim; ii--;)
       tmp[index++] = (intp)cpt_dimarr[ii];
    free (cpt_dimarr);
  }
  else if (ndim > 0)
  {
    ndim = dmGetArrayDimensions( dd, &dimarr );
    for ( ii = ndim; ii--; )
      tmp[index++] = (intp)dimarr[ii];
    free( dimarr );
  }

  dims->ptr = tmp;

}

/* ---------------------------------------------------------------------- */
/*  cxc_strings_to_carray()                                               */
/* ---------------------------------------------------------------------- */
static void* cxc_strings_to_carray( PyArrayObject* obj )
{
  void* result = NULL;

  long  ii;
  long  slen  = PyArray_ITEMSIZE( obj );
  long  nvals = PyArray_SIZE( obj );
  char *cptr  = (char*)PyArray_DATA( obj );

  char **tmp  = (char**)calloc( nvals, sizeof(char*) );
  char *buf   = (char*)calloc(nvals*(slen+1), sizeof(unsigned char) );

  tmp[0] = buf;
  strncpy(tmp[0], cptr, slen);

  for ( ii=1; ii < nvals; ii++)
  {
    cptr += slen;
    tmp[ii] = tmp[ii-1] + (slen+1);
    strncpy(tmp[ii], cptr, slen);
  }

  result = (void*)tmp;	

  return result;
}
