/*                                                                
**  Copyright (C) 2010  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/* *****************************************************************
 *
 * FILE NAME:  pyregion.c
 *
 * DEVELOPMENT: DataModel
 *
 * DESCRIPTION:  
 *   Provides a C-extension for Python to the Region functions
 *
 *
 * REVISION HISTORY:
 *
 ***************************************************************** */

#define REGMODULE
#include "pyregion.h"

char regmsg[1024];

staticforward PyTypeObject Region_Type;

#define pyRegion_Check(v) (Py_TYPE(v) == &Region_Type)

/* ---------------------------------------------------------------------- */
/* TypeFunc: pyRegion_new                                                 */
/* ---------------------------------------------------------------------- */
static PyObject* pyRegion_new(PyTypeObject *type,
  PyObject *args,
  PyObject *kwds)
{
  pyRegion* self = PyObject_New(pyRegion, &Region_Type);
  self->reg = NULL;

  return (PyObject*)self;
}

/* ---------------------------------------------------------------------- */
/* NewFunc: pyRegion_New                                                  */
/*   Allocate new pyRegion object and initialize with input               */
/*   C-level regReg structure.                                            */
/*   This function is made available for other C extensions (cxcdm)       */
/* ---------------------------------------------------------------------- */
static PyObject* pyRegion_New(regRegion* inreg)
{
  pyRegion* region = NULL;
  PyObject* result = NULL;

  region = PyObject_New(pyRegion, &Region_Type);
  if (region != NULL)
  {
    region->reg = inreg;
    result = Py_BuildValue("O", region);
  }

  return result;
}

/* ---------------------------------------------------------------------- */
/* TypeFunc: pyRegion_dealloc                                             */
/* ---------------------------------------------------------------------- */
static void pyRegion_dealloc(pyRegion* self)
{
  PyObject_Del(self);
}

/* ---------------------------------------------------------------------- */
/* Set up the region type methods                                         */
/* ---------------------------------------------------------------------- */
static PyMethodDef region_methods[] = { {NULL, NULL, 0, NULL} /* Sentinel - marks the end of this structure */

};/*end... */

/* ---------------------------------------------------------------------- */
/* Region Type                                                            */
/* ---------------------------------------------------------------------- */
static PyTypeObject Region_Type =
{
  PyObject_HEAD_INIT(NULL)
  0, /*ob_size*/
  (char*)"Region", /*tp_name*/
  sizeof(pyRegion), /*tp_basicsize*/
  0, /*tp_itemsize*/
  (destructor)pyRegion_dealloc, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_compare*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash */
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT, /*tp_flags*/
  "Region object", /* tp_doc */
  0, /* tp_traverse */
  0, /* tp_clear */
  0, /* tp_richcompare */
  0, /* tp_weaklistoffset */
  0, /* tp_iter */
  0, /* tp_iternext */
  region_methods, /* tp_methods */
  0, /* tp_members */
  0, /* tp_getset */
  0, /* tp_base */
  0, /* tp_dict */
  0, /* tp_descr_get */
  0, /* tp_descr_set */
  0, /* tp_dictoffset */
  0, /* tp_init */
  0, /* tp_alloc */
  pyRegion_new, /* tp_new */

};

/* ---------------------------------------------------------------------- */
/* Region Module                                                          */
/* ---------------------------------------------------------------------- */

/* Error Message Format Strings */
#define REGION_GENERIC_MSG        "%s() %s"
#define REGION_EMPTY_PTR_MSG      "%s() operation on empty Region pointer."
#define REGION_PARSE_MSG          "%s() Could not parse region string"
#define REGION_BAD_INPUT1_MSG     "%s() Could not create array from given parameters."
#define REGION_BAD_INPUT2_MSG     "%s() The xpos and ypos arrays are not the same length."
#define REGION_BAD_OUTPUT1_MSG    "%s() Could not create array object."
#define REGION_ALLOC_MEMORY_MSG   "%s() Could not allocate memory for %s."

/* Static declarations of Region module functions.. defined below.        */
static PyObject *_regPrintRegion(PyObject *self, PyObject *args);

/* ---------------------------------------------------------------------- */
/* Set up the module methods table                                        */
/* ---------------------------------------------------------------------- */
static PyMethodDef
    regionFuncs[] =
        {

          {"regArea", (PyCFunction)_regArea, METH_VARARGS,
           "Example:  area = regArea(regParse(\"circle(10,10,4)\"))"},
          {
           "regExtent",
           (PyCFunction)_regExtent,
           METH_VARARGS,
           "Example:  (x0,y0,x1,y1) = regExtent( regParse(\"circle(10,10,4)-rect(4,2,5,4)\") ) \n\
    print \"Bounding box is (%.1f,%.1f) to (%.1f,%.1f)\" % (x0, y0, x1, y1)"},
          {
           "regInsideRegion",
           (PyCFunction)_regInsideRegion,
           METH_VARARGS,
           "Example:  check = regInsideRegion( regParse(\"circle(10,10,4)\"), 12, 11 ) or \n\
     print regInsideRegion(regParse(\"circle(10,10,4)\"), numpy.array([ 10, 16, 12 ]), numpy.array([ 11, -4, 8 ]))"},
          {"regParse", (PyCFunction)_regParse, METH_VARARGS,
           "Example:  reg = regParse(\"circle(10,10,4)\")"},
          {"regPrintRegion", (PyCFunction)_regPrintRegion, METH_VARARGS,
           "Example:  regPrintRegion(regParse(\"circle(10,10,4)\"))"},
          {"regRegionString", (PyCFunction)_regRegionString, METH_VARARGS,
           "Example:  print regRegionString(regParse(\"circle(10,10,4)\"))"},

          {NULL, NULL, 0, NULL} /* Sentinel - marks the end of this structure */

        };/*end... */

/* ---------------------------------------------------------------------- */
/* InitFunc: init_region                                                  */
/* ---------------------------------------------------------------------- */
PyMODINIT_FUNC init_region(void)
{
  PyObject* m = NULL;

  static void* PyRegion_API[3];
  PyObject* c_api_object;

  if (PyType_Ready(&Region_Type) < 0)
    return;

  m = Py_InitModule((char*)"_region", regionFuncs);
  if (m == NULL)
    return;

  /* Must be present for NumPy.  Called first after above line. */
  import_array();

  /* initialize C API pointer array */
  PyRegion_API[0] = (void*)&Region_Type;
  PyRegion_API[1] = (void*)pyRegion_New;

  /* PyCapsule replaces PyCObject post python  v2.7??  */
  /* Create a Capsule for API pointers                 */
  /*  c_api_object = PyCapsule_New((void*)PyRegion_API, "region._C_API", NULL); */

  c_api_object = PyCObject_FromVoidPtr((void*)PyRegion_API, NULL);
  if (c_api_object != NULL)
    PyModule_AddObject(m, (char*)"C_API", c_api_object);

  Py_INCREF(&Region_Type);
  PyModule_AddObject(m, (char*)"Region", (PyObject *)&Region_Type);

  return;
}

/* ---------------------------------------------------------------------- */
/* Binding: _regArea                                                      */
/*   Calculate and return area inside a region.                           */
/* ---------------------------------------------------------------------- */
static PyObject* _regArea(PyObject *self, PyObject *args)
{
  char* funcName = "regArea";

  double xmin = -DBL_MAX; /* xmin coordinate */
  double ymin = -DBL_MAX; /* xmin coordinate*/
  double xmax = DBL_MAX; /* xmax coordinate*/
  double ymax = DBL_MAX; /* ymax coordinate*/
  double bin = 1; /* pixels */

  double xx[2]; /* 2d array to store xmin, xmax*/
  double yy[2]; /* 2d array to store ymin, ymax*/

  double area;

  PyObject* obj1 = NULL;
  PyObject* result = NULL;

  /* This will check for the correct number of args, and check to see if
   the input of xpos, and ypos were given as arrays, else they are doubles */
  if (PyTuple_Size(args) == 1)
  {
    /* Simple Case with 1 argument */
    if (!PyArg_ParseTuple(args, "O!:regArea", &Region_Type, &obj1))
      return NULL;

    /* Check underlying regRegion pointer is available */
    if (((pyRegion*)obj1)->reg == NULL)
    {
      sprintf(regmsg, REGION_EMPTY_PTR_MSG, funcName);
      PyErr_SetString(PyExc_ValueError, regmsg);
      return NULL;
    }

  }
  else
  {
    /* More complex case with 6 arguments */
    if (!PyArg_ParseTuple(args, "O!ddddd:regArea", &Region_Type, &obj1, &xmin,
                          &ymin, &xmax, &ymax, &bin))
      return NULL;

    /* Check underlying regRegion pointer is available */
    if (((pyRegion*)obj1)->reg == NULL)
    {
      sprintf(regmsg, REGION_EMPTY_PTR_MSG, funcName);
      PyErr_SetString(PyExc_ValueError, regmsg);
      return NULL;
    }

  }
  /* set array values */
  xx[0] = xmin;
  xx[1] = xmax;
  yy[0] = ymin;
  yy[1] = ymax;

  area = regArea(((pyRegion*)obj1)->reg, xx, yy, bin);
  result = Py_BuildValue("d", area);

  return result;
}

/* ---------------------------------------------------------------------- */
/* Binding: _regExtent                                                    */
/*   Determine the bounding box of a region.                              */
/* ---------------------------------------------------------------------- */
static PyObject *_regExtent(PyObject *self, PyObject *args)
{
  char* funcName = "regExtent";

  double xpos[2] = {0, 0}; /* Contains the returned xmin and xmax values */
  double ypos[2] = {0, 0}; /* Contains the returned ymin and ymax values */
  double fieldx[2] = {-DBL_MAX, DBL_MAX}; /* parameters for regExtent */
  double fieldy[2] = {-DBL_MAX, DBL_MAX}; /* parameters for regExtent */

  PyObject *obj1 = NULL;
  PyObject *result = NULL;

  /* Process arguments */
  if (!PyArg_ParseTuple(args, "O!:regExtent", &Region_Type, &obj1))
    return NULL;

  /* Check underlying regRegion pointer is available */
  if (((pyRegion*)obj1)->reg == NULL)
  {
    sprintf(regmsg, REGION_EMPTY_PTR_MSG, funcName);
    PyErr_SetString(PyExc_ValueError, regmsg);
    return NULL;
  }
  else
  {
    regExtent(((pyRegion*)obj1)->reg, fieldx, fieldy, xpos, ypos);
    result = Py_BuildValue("dddd", xpos[0], ypos[0], xpos[1], ypos[1]);
  }

  return result;
}

/* ---------------------------------------------------------------------- */
/* Binding: _regInsideRegion                                              */
/*   Queries a region to find out if a point (or an array of points)      */
/*   is inside the region or not.                                         */
/* ---------------------------------------------------------------------- */
static PyObject* _regInsideRegion(PyObject *self, PyObject *args)
{
  char* funcName = "regInsideRegion";

  double xpos; /* the x coordinate */
  double ypos; /* the y coordinate */
  int ii; /* loop variable */
  npy_intp dims[1]; /* the dimensions of the arrays */
  npy_intp typenum; /* the typenum */
  int numCoords = 0; /* the number of coordinates in the array*/
  short *c_value = NULL; /* the value in c style array */
  short *c_valueOut = NULL; /* the c style array of the output values */
  double c_xpos = 0; /* the array position of xpos */
  double c_ypos = 0; /* the array position of ypos */
  int stridex = 0; /* used to find next value in non-contiguous arrays */
  int stridey = 0; /* used to find next value in non-contiguous arrays */

  PyObject *obj1 = NULL; /* Python Object with Region_Type */
  PyObject *result = NULL; /* Python return value */

  PyArrayObject *py_xpos = NULL; /* Python Array Object to be converted to a C array for processing */
  PyArrayObject *py_ypos = NULL; /* Python Array Object to be converted to a C array for processing */
  PyArrayObject *py_valueArray = NULL; /* The result if xpos and ypos are arrays */

  /* This will check for the correct number of args, and check to see if
   the input of xpos, and ypos were given as arrays, else they are doubles */
  if ((PyTuple_Size(args) == 3)
      && (PyObject_TypeCheck(PyTuple_GET_ITEM(args, 1), &PyArray_Type))
      && (PyObject_TypeCheck(PyTuple_GET_ITEM(args, 2), &PyArray_Type)))
  {
    /* Case where the arguments of xpos and ypos are doubles arrays */
    if (!PyArg_ParseTuple(args, "O!O!O!:regInsideRegion", &Region_Type, &obj1,
                          &PyArray_Type, &py_xpos, &PyArray_Type, &py_ypos))
      return NULL;

    if ((obj1 == NULL) || (py_xpos == NULL) || (py_ypos == NULL))
    {
      sprintf(regmsg, REGION_BAD_INPUT1_MSG, funcName);
      PyErr_SetString(PyExc_Exception, regmsg);
      return NULL;
    }

    /* Make sure the arrays are of correct type */
    if ((PyArray_CanCastSafely(py_xpos->descr->type_num, NPY_DOUBLE))
        && (PyArray_CanCastSafely(py_ypos->descr->type_num, NPY_DOUBLE)))
    {
      /*cast to a double*/
      py_xpos = (PyArrayObject *)PyArray_Cast(py_xpos, NPY_DOUBLE);
      py_ypos = (PyArrayObject *)PyArray_Cast(py_ypos, NPY_DOUBLE);
    }

    /* Handles case if array is not contiguous in memory */
    stridex = PyArray_STRIDE(py_xpos, 0);
    stridey = PyArray_STRIDE(py_ypos, 0);
    char *arrx = PyArray_BYTES(py_xpos);
    char *arry = PyArray_BYTES(py_ypos);

    /* check that the xpos array and ypos array have same length */
    if ((py_xpos->dimensions[0]) == (py_ypos->dimensions[0]))
    {
      numCoords = py_xpos->dimensions[0];
    }
    else
    {
      sprintf(regmsg, REGION_BAD_INPUT2_MSG, funcName);
      PyErr_SetString(PyExc_Exception, regmsg);
      return NULL;
    }

    c_value = (short *)calloc(numCoords, sizeof(short));
    if (c_value == NULL)
    {
      sprintf(regmsg, REGION_ALLOC_MEMORY_MSG, funcName, (char*)"results array");
      PyErr_SetString(PyExc_MemoryError, regmsg);
      return NULL;
    }

    /* loop through all of the coordinates and put the return value in c_value */
    for (ii = 0; ii < numCoords; ii++)
    {
      c_xpos = *((double*)((void*)(arrx + stridex * ii)));
      c_ypos = *((double*)((void*)(arry + stridey * ii)));
      c_value[ii] = regInsideRegion(((pyRegion*)(obj1))->reg, c_xpos, c_ypos);
    }

    dims[0] = numCoords;
    typenum = NPY_SHORT;

    /* create the output array */
    py_valueArray = (PyArrayObject *)PyArray_SimpleNew(1, dims, typenum);
    if (py_valueArray == NULL)
    {
      sprintf(regmsg, REGION_BAD_OUTPUT1_MSG, funcName);
      PyErr_SetString(PyExc_Exception, regmsg);
      return NULL;
    }

    c_valueOut = SDATA(py_valueArray);

    for (ii = 0; ii < numCoords; ii++)
    {
      c_valueOut[ii] = c_value[ii]; /*grab the data from c_val and place in array for output */
    }/*end for... */

    free(c_value);/* free the allocated memory */

    result = Py_BuildValue("O", PyArray_Return(py_valueArray));

  }/*end if... check for arrayTypes for xpos, ypos */
  else
  {
    /* Case where the arguments of xpos and ypos are doubles */
    if (!PyArg_ParseTuple(args, "O!dd:regInsideRegion", &Region_Type, &obj1,
                          &xpos, &ypos))
      return NULL;

    /* Return  */
    result = Py_BuildValue("h", regInsideRegion(((pyRegion*)(obj1))->reg, xpos,
                                                ypos));
  }
  return result;
}

/* ---------------------------------------------------------------------- */
/* Binding: _regParse                                                     */
/*   Parse region string to Region object                                 */
/* ---------------------------------------------------------------------- */
static PyObject* _regParse(PyObject *self, PyObject *args)
{
  char* funcName = "regParse";

  char* buff = NULL;
  regRegion* _region = NULL;
  PyObject* result = NULL;

  /* Process arguments */
  if (!PyArg_ParseTuple(args, "s:regParse", &buff))
    return NULL;

  _region = dmRegParse(buff);
  if (!_region)
  {
    sprintf(regmsg, REGION_PARSE_MSG, funcName);
    PyErr_SetString(PyExc_ValueError, regmsg);
    return NULL;
  }
  else
  {
    result = pyRegion_New(_region);
  }

  return result;
}

/* ---------------------------------------------------------------------- */
/* Binding: _regPrintRegion                                               */
/*   Prints the various shapes and components within a region to stdout.  */
/* ---------------------------------------------------------------------- */
static PyObject* _regPrintRegion(PyObject *self, PyObject *args)
{
  char* funcName = "regPrintRegion";

  PyObject *obj1 = NULL;

  /* Process arguments */
  if (!PyArg_ParseTuple(args, "O!:regPrintRegion", &Region_Type, &obj1))
    return NULL;

  /* Check underlying regRegion pointer is available */
  if (((pyRegion*)obj1)->reg == NULL)
  {
    sprintf(regmsg, REGION_EMPTY_PTR_MSG, funcName);
    PyErr_SetString(PyExc_ValueError, regmsg);
    return NULL;
  }
  else
  {
    regPrintRegion(((pyRegion*)obj1)->reg);
  }

  Py_INCREF(Py_None);
  return (Py_None);
}

/* ---------------------------------------------------------------------- */
/* Binding: _regRegionString                                              */
/*   returns a string representation of the region.                       */
/* ---------------------------------------------------------------------- */
static PyObject* _regRegionString(PyObject *self, PyObject *args)
{
  char* funcName = "regRegionString";

  char *outstr = NULL;
  PyObject *obj1 = NULL;
  PyObject* result = NULL;

  /* Process arguments */
  if (!PyArg_ParseTuple(args, "O!:regRegionString", &Region_Type, &obj1))
    return NULL;

  /* Check underlying regRegion pointer is available */
  if (((pyRegion*)obj1)->reg == NULL)
  {
    sprintf(regmsg, REGION_EMPTY_PTR_MSG, funcName);
    PyErr_SetString(PyExc_ValueError, regmsg);
    return NULL;
  }
  else
  {
    outstr = regAllocComposeRegion(((pyRegion*)obj1)->reg);
    result = Py_BuildValue("s", outstr);
  }

  return result;
}
