/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

#include <stdio.h>
#include "utlib.h"

static void ieee_get( const char* buf, char* ptr, const long n, const long size );
static void ieee_put( const char* ptr, char* buf, const long n, const long size );



/* Promote structure buffer entry to be aligned as required 
 * total  the number of bytes up to end of data for entry n
 * size   size of one element of entry n+1
 * old_size size of one element of entry n
 * pad    extra bytes to put on end of entry n to get
 *       the beginning of entry n+1 on the right boundary
 */
long ut_sys_align( long total, long size, long old_size )
{
 long pad = 0;
 long align = 0;

 if ( old_size < size )
 {
  align = size;

#ifdef LINUX_X86
  if ( align > UT_SZ_I ) align = UT_SZ_I;
#endif
  pad = align - ( total % align );
  if ( pad == align ) pad = 0;
 }
 return pad;
}

void ieee_put_r( const float* data, char* buf, const long n )
{
 ieee_put( (char*)data, buf, n, (long)UT_SZ_R );
}
void ieee_put_d( const double* data, char* buf, const long n )
{
 ieee_put( (char*)data, buf, n, (long)UT_SZ_D );
}
void ieee_put_s( const short* data, char* buf, const long n )
{
 ieee_put( (char*)data, buf, n, (long)UT_SZ_S );
}
void ieee_put_i( const long* data, char* buf, const long n )
{
 ieee_put( (char*)data, buf, n, (long)UT_SZ_I );
}


void ieee_get_r( const char* buf, float* data, const long n )
{
 ieee_get( buf, (char*)data, n, UT_SZ_R );
}

void ieee_get_d( const char* buf, double* data, const long n )
{
 ieee_get( buf, (char*)data, n, UT_SZ_D );
}

void ieee_get_s( const char* buf, short* data, const long n )
{
 ieee_get( buf, (char*)data, n, UT_SZ_S );
}

void ieee_get_i( const char* buf, long* data, const long n )
{
 ieee_get( buf, (char*)data, n, UT_SZ_I );
}

static void ieee_get( const char* ibuf, char* ptr, const long nn, long size )
{
 char* buf = (char*)ibuf;
 long m;
 long n = nn;
 long nbytes;

 m = size;
 nbytes = size * n; 
#if BYTE_SWAPPED
 while( n-- ) {
  while ( m-- ) {
   *ptr = *(buf + m );
   ptr++;
  }
  m = size;
  buf += size;
 }
#else
 memcpy( ptr, buf, nbytes );
#endif
}

static void ieee_put( const char* iptr, char* buf, const long nn, const long size )
{
 char* ptr = (char*)iptr;
 long nbytes;
 long n = nn;
 long m;
 nbytes = n * size;
 m = size;
#if BYTE_SWAPPED
 while( n-- ) {
  while( m-- ) {
   *buf = *(ptr + m );
   buf++;
  }
  m = size;
  ptr += size;
 }
#else
 memcpy( buf, ptr, nbytes );
#endif
}



void bytes_split_i( const long value, long* bytes )
{
 long byte_no;
 for ( byte_no = 1; byte_no <= 4; byte_no++ ) {
  bytes[byte_no-1] = byte_get_i( value, byte_no );
 }
}

void bytes_split_r( const float value, long* bytes )
{
 long byte_no;
 for ( byte_no = 1; byte_no <= 4; byte_no++ ) {
  bytes[byte_no-1] = byte_get_r( value, byte_no );
 }
}

void bytes_split_d( const double value, long* bytes )
{
 long byte_no;
 for ( byte_no = 1; byte_no <= 8; byte_no++ ) {
  bytes[byte_no-1] = byte_get_d( value, byte_no );
 }
}

void bytes_split_s( const short value, long* bytes )
{
 long byte_no;
 for ( byte_no = 1; byte_no <= 2; byte_no++ ) {
  bytes[byte_no-1] = byte_get_s( value, byte_no );
 }
}

void bytes_split_c( const char* value, long* bytes, const long n_bytes )
{
 long byte_no;
 for ( byte_no = 1; byte_no <= n_bytes; byte_no++ ) {
  bytes[byte_no-1] = byte_get_c( value, byte_no );
 }
}


long bytes_combine_i( const long* bytes )
{
 long byte_no;
 long value = 0;
 for ( byte_no = 1; byte_no <= 4; byte_no++ ) {
  byte_set_i( &value, bytes[byte_no-1], byte_no );
 }
 return( value );
}

short bytes_combine_s( const long* bytes )
{
 long byte_no;
 short value = 0;
 for ( byte_no = 1; byte_no <= 2; byte_no++ ) {
  byte_set_s( &value, bytes[byte_no-1], byte_no );
 }
 return( value );
}


float bytes_combine_r( const long* bytes )
{
 long byte_no;
 float value = 0;
 for ( byte_no = 1; byte_no <= 4; byte_no++ ) {
  byte_set_r( &value, bytes[byte_no-1], byte_no );
 }
 return( value );
}


double bytes_combine_d( const long* bytes )
{
 long byte_no;
 double value = 0;
 for ( byte_no = 1; byte_no <= 8; byte_no++ ) {
  byte_set_d( &value, bytes[byte_no-1], byte_no );
 }
 return( value );
}


void bytes_combine_c( const long* bytes, char* value, const long n_bytes )
{
 long byte_no;

 for ( byte_no = 1; byte_no <= n_bytes; byte_no++ ) {
  byte_set_c( value, bytes[byte_no-1], byte_no );
 }
}



void byte_set_s( short* number, const long value, const long byte_no )
{
 short buffer;
 buffer = bit_shft_s( (short)value, UT_BYTE_BITS * ( byte_no - 1 ) );
 *number |= buffer;
}


void byte_set_i( long* number, const long value, const long byte_no )
{
 long buffer;
 buffer = bit_shft_i( value, UT_BYTE_BITS * ( byte_no - 1 ) );
 *number |= buffer;
}

void byte_set_r( float* number, const long value, const long byte_no )
{
 long* ptr;
 ptr = (long *)number;
 byte_set_i( ptr, value, byte_no );
}


void byte_set_d( double* number, const long value, const long byte_no )
{
 long* ptr;
 long segment;
 long ibyte;

 ptr = (long *)number;
 sys_byte_d( byte_no, &segment, &ibyte );
 byte_set_i( ptr + segment - 1, value, ibyte );
}

long byte_get_i( const long value, const long byte_no )
{
 long start_bit;
 long result;

 start_bit = ( byte_no - 1 ) * UT_BYTE_BITS;
 result = bit_unpack_i( value, start_bit, UT_BYTE_BITS );
 return( result ); 
}


long byte_get_r( const float value, const long byte_no )
{
 long* buffer;
 long result;
 buffer = (long *)(&value);
 result =  byte_get_i( *buffer, byte_no );
 return( result );
}

long byte_get_s( const short value, const long byte_no )
{
 long start_bit;
 long result;
 start_bit = ( byte_no - 1 ) * UT_BYTE_BITS;
 result = bit_unpack_s( value, start_bit, UT_BYTE_BITS );
 return( result ); 
}

long byte_get_d( const double value, const long byte_no )
{
 long* ptr;
 long segment;
 long ibyte;
 long result;
 
 ptr = (long *)(&value);
 sys_byte_d( byte_no, &segment, &ibyte );
 result = byte_get_i( *(ptr + segment - 1), ibyte );
 return( result );
}


long byte_get_c( const char* value, const long byte_no )
{
 uchar c;
 c = (uchar)value[byte_no-1];
 return( (long)c );
}

void byte_set_c( char* data, const long value, const long byte_no )
{
 data[byte_no-1] = (uchar)value;
}


void sys_byte_d( const long byte_no, long* segment, long* ibyte )
{
 if ( byte_no > 4 ) {
  *ibyte = byte_no - 4;
  *segment = UT_R8_WORD2;
 } else {
  *ibyte = byte_no;
  *segment = UT_R8_WORD1;
 }
}
