/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*======================================================================*/
/*
 * Math routines
 *
 * <math.h>:  	log  used in logbase_d
 * 		modf used in frac_d_std
 * <float.h>:	DBL_EPSILON used in frac_d
 */
/*======================================================================*/

#include <float.h>
#include "utlib.h"

double utm_range_d( double x, double xmin, double xmax )
{
 if ( x < xmin ) x = xmin;
 if ( x > xmax ) x = xmax;
 return x;
}


double utm_min_d( double x, double y )
{
 return( x < y ? x : y );
}

double utm_max_d( double x, double y )
{
 return( x > y ? x : y );
}

long utm_ceiling_d( const double x )
{
 long i;
 i = (long)x;
 if ( x > (double)i ) i++;
 return( i );
}

long utm_ceiling_r( const float x )
{
 long i;
 i = (long)x;
 if ( x > (float)i ) i++;
 return( i );
}

long utm_floor_d( const double x )
{
 long i;
 i = (long)(x+ 2 * DBL_EPSILON);
 if ( x < (double)i ) i--;
 return( i );
}

long utm_floor_r( const float x )
{
 long i;
 i = (long)(x + 2 * FLT_EPSILON);
 if ( x < (float)i ) i--;
 return( i );
}


/* The ANSI C modulo operator is unreliable (machine dept) for negative operands 
 and undefined for float operands */

long utm_modulo_i( const long x, const long base )
{
 long result;
 if ( x > 0 && base > 0 ) {
  result = x % base;
 } else {
  result = x - floor_d( (double)x / (double)base ) * base;
 }
 return( result );
}

short utm_modulo_s( const short x, const short base )
{
 short result;
 if ( x > 0 && base > 0 ) {
  result = x % base;
 } else {
  result = x - floor_d( (double)x / (double)base ) * base;
 }
 return( result );
}

float utm_modulo_r( const float x, const float base )
{
 float result;
 result = x - floor_r( x / base ) * base;
 if ( base > 0 && result > base - 10 * base * FLT_EPSILON ) result = 0.0;
 return( result );
}

double utm_modulo_d( const double x, const double base )
{
 double result;
 result = x - floor_d( x / base ) * base;
 if ( base > 0 && result > base - 10 * base * DBL_EPSILON ) result = 0.0;
 return( result );
}



/*----------------------------------------------------------------------*/
/*   divisible
 *
 *   Is a divisible by b? 
 *
 *   1996 Aug 29 JCM
 *   Args:
 *  	a		(i)	Operand
 *	b		(i)	Divisor
 */

logical utm_divisible( const long a, const long b )
{
 return (  ( a/b ) * b == a  );
}

/*----------------------------------------------------------------------*/
/*
 *   min_i
 *
 *   Return smaller of i and j
 *   1996 Aug 29 JCM
 *
 */
 
long utm_min_i( const long i, const long j )
{
 long k;
 k = i < j ? i : j;
 return( k );
}

/*----------------------------------------------------------------------*/
/*
 *   max_i
 *
 *   Return larger of i and j
 *   1996 Aug 29 JCM
 *
 */

long utm_max_i( const long i, const long j )
{
 long k;
 k = i > j ? i : j;
 return( k );
}


/*----------------------------------------------------------------------*/
/*
 *   rmod_d
 *
 *   Calculate quotient and remainder
 *   1996 Aug 29 JCM
 *   Args:
 *	x		(i)	Input value
 *	y		(o)	Remainder
 *	r		(i)	Divisor
 *	rmod_d		(o)	Quotient
 */

long utm_rmod_d( const double value, double* remainder, const double divisor )
{
 long quotient;
 quotient = floor_d( value / divisor );
 *remainder = value - divisor * quotient; 
 return( quotient );
}

/*----------------------------------------------------------------------*/
/*
 *   rmod_i
 *
 *   Calculate quotient and remainder
 *   1996 Aug 29 JCM
 */

long utm_rmod_i( const long value, long* remainder, const long divisor )
{
 long quotient;
 quotient = floor_d( (double)value / (double) divisor );
 *remainder = value - divisor * quotient;
 return( quotient );
}


long utm_rmod_r( const float value, float* remainder, const float divisor )
{
 long quotient;
 quotient = floor_r( value / divisor );
 *remainder = value - divisor * quotient; 
 return( quotient );
}

long utm_rmod_s( const short value, short* remainder, const short divisor )
{
 long quotient;
 quotient = floor_r( (float)value / (float) divisor );
 *remainder = value - divisor * quotient;
 return( quotient );
}


/*----------------------------------------------------------------------*/
/*
 *  frac_r
 *
 *  Split value into long and fractional parts.
 *  Uses FLT_EPSILON from <float.h>
 *  1996 Aug 29 JCM
 *
 */

void utm_frac_r( const float value, long* int_val, float* fraction )
{
 long long_part;

 long_part = floor_r( value );
 *fraction = value - long_part;
 if ( *fraction >= 1.0 - 2 * FLT_EPSILON ) 
 {
  *fraction = 0.0; long_part++;
 }
 *int_val = long_part;
}

/*----------------------------------------------------------------------*/
/*
 *  frac_d
 *
 *  Split value into long and fractional parts.
 *  Uses DBL_EPSILON from <float.h>
 *  1996 Aug 29 JCM
 *
 */

void utm_frac_d( const double value, long* int_val, double* fraction )
{
 long long_part;

 long_part = floor_d( value );
 *fraction = value - long_part;
 if ( *fraction > 1.0 - 2 * DBL_EPSILON ) { *fraction = 0.0; long_part++; }
 *int_val = long_part;
}

/*----------------------------------------------------------------------*/
/*
 *  logbase_i
 *
 *  Return log to base n of argument, rounded down to nearest long and
 *  at least unity. Used in sort algorithm.
 *  1996 Aug 30 JCM
 *  Args:
 * 	i		(i)	Argument
 *	n		(i)	Base
 */

long utm_logbase_i( const long i, const long n )
{
 double x;
 x = logbase_d(  (double)i , n ) + UT_FUNC_TOL;
 return( max_i( (long)x, 1 ));
}

/*----------------------------------------------------------------------*/
/*
 *  logbase_d
 *
 *  Return log to base n of argument
 *  Uses log from <math.h>
 *  1996 Aug 30 JCM
 *  Args:
 * 	x		(i)	Argument
 *	n		(i)	Base
 */

double utm_logbase_d( const double x, const long n )
{
 return( log( x )/log( (double)n ) );
}

float utm_logbase_r( const float x, const long n )
{
 double result;
 result = ( log((double)x )/log( (double)n ) );
 return( (float) result );
}

float utm_round_r( const float x, const long f )
{
 long tol;
 float y;
 float x_log;
 long x_exp;
 long x_rounded;
 float x_log_mantissa;

 tol = f - 1;
 if ( x == 0.0 ) {
  y = 0.0;
 } else {
  x_log = rlog10( rabs( x ) );
  x_exp = floor_d( x_log ) - tol;
  x_log_mantissa = x_log - x_exp;
  x_rounded = (long) ( pow( 10.0, x_log_mantissa ) + 0.5 );
  y = x_rounded * tpow_r( x_exp );
  if ( x < 0 ) y *= -1;
 }
 return( y );

}

double utm_round_d( const double x, const long f )
{
 long tol;
 double y;
 double x_log;
 long x_exp;
 long x_rounded;
 double x_log_mantissa;

 tol = f - 1;
 if ( x == 0.0 ) {
  y = 0.0;
 } else {
  x_log = log10( fabs( x ) );
  x_exp = floor_d( x_log ) - tol;
  x_log_mantissa = x_log - x_exp;
  x_rounded = (long) ( pow( 10.0, x_log_mantissa ) + 0.5 );
  y = x_rounded * tpow_d( x_exp );
  if ( x < 0 ) y *= -1;
 }
 return( y );
}

void utm_order_r( float* x, float* y)
{
 float tmp;
 if ( *y < *x ) { tmp = *x; *x = *y; *y = tmp; }
}

void utm_order_d( double* x, double* y )
{
 double tmp;
 if ( *y < *x ) { tmp = *x; *x = *y; *y = tmp; }
}


float utm_lint_r( const float x1, const float x2, const float yy1, const float yy2, const float x0 )
{
 float y;
 if ( x1 == x2 ) {
  y = ( yy1 + yy2 ) / 2.0;
 } else {
  y = ( yy1 * ( x2 - x0 ) + yy2 * ( x0 - x1 ) ) / ( x2 - x1 );
 }
 return( y );

}


double utm_lint_d( const double x1, const double x2, const double yy1, const  double yy2, const double x0 )
{
 double y;
 if ( x1 == x2 ) {
  y = ( yy1 + yy2 ) / 2.0;
 } else {
  y = ( yy1 * ( x2 - x0 ) + yy2 * ( x0 - x1 ) ) / ( x2 - x1 );
 }
 return( y );

}

float utm_rpow( const float x, const float n )
{
 return( (float)pow( (float)x, (float)n ) );
}

float utm_rlog10( const float x )
{
 return( (float)log10( (double) x ) );
}

float utm_rabs( const float x )
{
/* Not in math.h! */
 float y;
 y = x < 0.0 ? -x : x;
 return( y );
}

float utm_rsqrt( const float x )
{
 return( (float)sqrt( (double)x ) );
}


long utm_iplus( const logical q )
{
 return( q ? +1 : -1 );
}

long utm_nint_r( const float x )
{
 long i;
 float f;
 frac_r( x, &i, &f );
 if ( f >= 0.5 ) i++;
 return( i );
}

long utm_nint_d( const double x )
{
 long i;
 double f;
 frac_d( x, &i, &f );
 if ( f >= 0.5 ) i++;
 return( i );
}

