/*                                                                
**  Copyright (C) 2003,2005,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

#include "utlib.h"
#include "esc_types.h"
#include "jcmstate.h"

char** tpio_set( Tpio tpio, FioStack iofs );
void tpio_loc( Tpio tpio, FioStack iofs, long* level, long* pos );
TpioLoop* tpio_loop_push( Tpio tpio );
void tpio_level_check( Tpio tpio, long level );

Tpio tpio_get( void )
{
 return ut_state_tpio( NULL );
}

Tpio tpio_data_init( void )
{
 long i;
 Tpio tpio_data = calloc( 1, sizeof ( struct Tpio_s ));
 tpio_data->buffer = calloc( UT_IO_DEPTH, UT_SZ_P);
 for ( i = 0; i < UT_IO_DEPTH; i++ )
  tpio_data->buffer[i] = calloc( UT_TPIO_BUFLEN, UT_SZ_C );

 tpio_data->bptr = calloc( UT_IO_DEPTH, UT_SZ_P);
 tpio_data->cmd_mode = UT_FALSE;
 tpio_data->cmd = tpio_data->buffer[0];
 tpio_data->ptr = NULL;
 tpio_data->loop = calloc( UT_TPIO_MAX_LOOPS, sizeof( TpioLoop ));
 tpio_data->nloop = 0;
 tpio_data->maxloop = UT_TPIO_MAX_LOOPS;
 tpio_data->depth = UT_IO_DEPTH;
 tpio_tp_init( tpio_data );
 return tpio_data;
}

void tpio_level_check( Tpio tpio, long level )
{
 long lev;
 long lmax;
 if ( level >= tpio->depth )
 {
   lmax = level + 1;
   tpio->buffer = realloc( tpio->buffer, lmax * UT_SZ_P);
   tpio->bptr  = realloc( tpio->bptr, lmax * UT_SZ_P);
   for ( lev = tpio->depth; lev < lmax; lev++ )
    tpio->buffer[lev] = calloc( UT_TPIO_BUFLEN, UT_SZ_C);
   tpio->depth = lmax;
 }
}

void tpio_free( Tpio tpio )
{
 ar_free_cp( tpio->buffer, tpio->depth );
 free(tpio->bptr );
 free( tpio->loop ); 
 free( tpio );
}

/*
 *--------------------------------------------------------------------------------
 * Internal routine to give buffer status
 */

void tpio_loc( Tpio tpio, FioStack iofs, long* level, long* pos )
{
 if ( !tpio ) tpio = tpio_get();
 *level = fio_iostack_get_level( iofs, FIO_INPUT );
 tpio_level_check( tpio, *level );
 if ( *tpio->ptr == NULL ) {
  *pos = -1;
 } else {
  *pos = *tpio->ptr - tpio->buffer[*level] + 1;
 }
}


/*
 *--------------------------------------------------------------------------------
 * Clear one level of the buffer
 */
void tpio_cclear( Tpio tpio, long level )
{
 if ( !tpio ) tpio = tpio_get();
 tpio->bptr[level] = tpio->buffer[level];
 cs_zero( tpio->bptr[level] );
}


/*----------------------------------------------------------------------*/
/*
 * tpio_dbg
 * Uses <stdio.h>
 */

void tpio_dbg( void )
{
 tpio_tp_dbg( NULL );
}

void tpio_tp_dbg( Tpio tpio )
{
 if ( !tpio ) tpio = tpio_get();
 printf("TPIO BUFFER POINTER %p\n", (void*)(tpio->ptr) );
 if ( tpio->ptr != NULL ) printf("TPIO BUFFER HEAD    %p\n", *tpio->ptr );
}


/*----------------------------------------------------------------------*/
/*
 * tpio_init
 *
 * Zero the buffer and set the stack to bottom level
 * Uses csproto.h
 */

void tpio_tp_init( Tpio tpio )
{
 long level;

 if ( !tpio ) tpio = tpio_get();
 for ( level = 0; level < tpio->depth; level++ ) {
  tpio->bptr[level] = NULL;
  cs_zero( tpio->buffer[level] );
 }
 tpio->ptr = tpio->bptr;
}

/*----------------------------------------------------------------------*/
/* Internal buffer manipulation */
char** tpio_set( Tpio tpio, FioStack iofs )
{
 long level; 
 if ( !tpio ) tpio = tpio_get();
 if ( !iofs ) iofs = fio_get_stack();
 level = fio_iostack_get_level( iofs, FIO_INPUT );
 tpio_level_check( tpio, level );
 if ( tpio->bptr[level] == NULL ) tpio->bptr[level] = tpio->buffer[level];
 tpio->ptr = &(tpio->bptr[level]);
 return( tpio->ptr );
}
/*
 * tpio_getbuf
 *
 * For interactive use, prompt for new buffer and read it from the terminal
 * using iofs_read.
 * Strip comments (tabs and continuations are handled in iofs_read).
 * In non-interactive use, read buffer from input stream; keep trying
 * if you get a blank line or a line starting with comment as the first character. Keeps going until
 * it gets either end of file, or a line with data. The data returned to
 * the internal buffer can be blank if either 1) end of file is found or 2)
 * the line consists of whitespace followed by comment.
 *
 * Uses csproto.h
 * Uses escproto.h
 * Uses esc_types.h

 */


/*
 *--------------------------------------------------------------------------------
 */
logical tpio_getbuf( Tpio tpio, FioStack iofs, char* prompt )
{
  logical retry;
  character cmt_char, esc_char;
  logical status;
  FioFile file;
  char* buf;
  char** ptr;
  TextCard ebuf;

  if ( !iofs ) iofs = fio_get_stack();
  if ( !tpio ) tpio = tpio_get();
  ptr = tpio_set( tpio, iofs );
  buf = *ptr;
  retry = UT_TRUE;
  esc_char = esc_special_char( ESC_CHAR );
  cmt_char = esc_special_char( CMT_CHAR );
  while( retry || *buf == cmt_char ) {
   status = fio_iostack_read( iofs, buf, prompt, UT_TPIO_BUFLEN );
   if ( ut_dbg_level(UT_VERBOSE_FIO) ) {
    sprintf( ebuf, "TPIO new buffer:%s", buf );   
    ut_msg_d( UT_VERBOSE_FIO, ebuf );
   }
   retry = ( status && cs_is_blank( buf ));
   if ( retry )
   {
    file = fio_iostack_file( iofs, FIO_INPUT );
    retry  = !fio_file_interactive( file );
   }
  }

  cs_decmt( buf, esc_char, cmt_char );
  return( status );
}

/*----------------------------------------------------------------------*/
/*
 *
 * tpio_loop_check
 *
 * Test for \END if in loop.
 * Uses csproto.h, esc_types.h, escproto.h
 */

static logical tpio_loop_check( Tpio tpio, FioFileList list, char* buf )
{
 character end_string[5] = "\\END";

 if ( !tpio ) tpio = tpio_get();

 if ( tpio->nloop > 0 ) {
  if ( !list ) list = fio_get_list();
  if ( fio_list_any_eof( list, FIO_CHECK ))
  {
   end_string[0] = esc_special_char( ESC_CHAR );
   return ( !cs_eq( end_string, buf ) );
  }
 } 
 return( UT_FALSE );
}

/*----------------------------------------------------------------------*/
/*
 * tpio_next_name
 *
 * Get the next word in the buffer
 * Uses tokenproto.h
 */

static void tpio_next_name( Tpio tpio, char* buf )
{
  long siz;

 if ( !tpio ) tpio = tpio_get();
  siz = FN_SIZE;
  cs_get_token( tpio->ptr, buf, TMODE_WORD, siz );
}

/*----------------------------------------------------------------------*/
/*
 * tpio->cmdfile
 *
 * Open file and use as input 
 * Uses csproto.h
 */

static void tpio_cmdfile( Tpio tpio, FioFileList list, FioStack iofs, char* buf )
{
 Filename filename;
 if ( !tpio ) tpio = tpio_get();
 if( *buf == '\0' ) {
  tpio_next_name( tpio, filename );
 } else {
  cs_copy( buf, filename );
 }
 tpio_tp_in( tpio, list, iofs, filename );
}

/*----------------------------------------------------------------------*/
/*
 * tpio_logfile
 *
 * Uses csproto.h
 * 
 */

static void tpio_logfile( Tpio tpio, FioFileList list, FioStack iofs, char* buf )
{
 Filename filename;
 logical append, syserr;
 long lptr;

 char* ptr;
 lptr = 1;
 ptr = buf;
 append = (*ptr == '>' );
 if ( append ) ptr++;
 syserr = (*ptr == '&' );
 if ( syserr ) ptr++;
 fio_iostack_opop( iofs, list, FIO_POP_CLOSE );
 cs_copy( ptr, filename );
 if ( cs_is_blank( filename ) ) tpio_next_name( tpio, filename );
 tpio_out( filename, append, syserr );
}

/*----------------------------------------------------------------------*/
/*
 * tpio_esc
 * Uses csproto.h
 */

static logical tpio_esc( Tpio tpio, FioFileList list, FioStack iofs, char* buf )
{
 logical flag;
 TextCard ebuf;
 TextCard ubuf;
 long ival;
 char* ptr;
 TextWord token;
 TextWord token1;
 TextCard embuf;


 if ( !tpio ) tpio = tpio_get();

 cs_copy( " ", embuf );
 flag = UT_TRUE;
 cs_copy( buf, ebuf );
 cs_upper( ebuf );
 ptr = ebuf;
 if( buf[0] == '<' )
  tpio_cmdfile( tpio, list, iofs, &buf[1] );
 else if ( buf[0] == '>' )
  tpio_logfile( tpio, list, iofs, &buf[1] );
 else if ( cs_eq_siz( buf, "[*", 2 ) ) {
  ival = cs_read_i( cs_upto( &buf[2], ']', ubuf ) );
  ut_dbg_set( ival );
  sprintf( ut_tmp, "DEBUG SET %ld\n", (long)ival );
  ut_msg_d( ival, ut_tmp );
 } else if ( cs_eq( ptr, "SET" ) ) {
  ut_msg_d( UT_VERBOSE_FIO, "ESC <Reset special chars>" );
  cs_get_token( tpio->ptr, token, TMODE_WORD, WORD_SIZE );
  sprintf( ut_tmp, "ARG = %s", token );
  ut_msg_d( UT_VERBOSE_FIO, ut_tmp );
  if ( cs_eq( token, "?" ) ) {
   esc_status();
  } else {
   cs_get_token( tpio->ptr, token1, TMODE_WORD, WORD_SIZE );
   esc_reset( token, *token1 );
  }
 } else if ( cs_eq( ptr, "VER" ) ) {
  ut_ver_msg();
 } else if ( cs_eq( ptr, "VERBOSE" ) || cs_eq( ptr, "DEBUG") ) {
  cs_get_token( tpio->ptr, token, TMODE_WORD, WORD_SIZE );
  ival = cs_read_i( token );
  ut_dbg_set( ival );
  sprintf( ut_tmp, "DEBUG SET %ld\n", (long)ival );
  ut_msg_d( ival, ut_tmp );
 } else if ( cs_eq( ptr, "CLEAR" ) ) {
  Error* error = ut_error_state();
  ut_error_clear( error );
  ut_msg_d( UT_VERBOSE, "Cleared error state" );
 } else if ( cs_eq( ptr, "DO" ) ) {
  loop_do( tpio, iofs );
 } else if ( cs_eq( ptr, "END" ) ) {
  loop_end( tpio, iofs );
 } else if ( cs_eq( ptr, "ECHO" ) ) {
  loop_echo( tpio, iofs );
 } else if ( cs_eq( ptr, "EXIT" ) ) {
  fio_iostack_ipop( iofs, tpio, list, FIO_POP_CLOSE );
 } else if ( *ptr == '?' ) { 
  if ( cs_eq( ptr, "?IOFS" ) ) {
   fio_iostack_print_state(iofs);
  } else if ( cs_eq( ptr, "?FIO" ) ) {
   fio_list_print(list);
  } else if ( cs_eq( ptr, "?BIO" ) ){
   fio_basic_print_state();
  } else if ( cs_eq( ptr, "?PAGER" ) ){
   fio_pager_print_state( list );
  } else if ( cs_eq( ptr, "?PATH" ) ){
   fio_list_list_path( list );
  } else if ( cs_eq( ptr, "?TPIO" ) ) {
   tpio_print_state();
  } else if ( cs_eq( ptr, "?ESC" ) ) {
   esc_status();
  } else if ( cs_eq( ptr, "?ERR" ) ) {
   ut_error_status();
  } else {
   cs_copy( ptr, embuf );
  } 
 } else {
  cs_copy( ptr, embuf );
 }
 flag = !( cs_is_blank( buf ) || !cs_is_blank( embuf ) );
 if ( !flag ) {
  sprintf( ut_tmp, " [ESC] Unknown escape command-:%-s", embuf );
  ut_err_util( ut_tmp, UT_ERR_TPIO_ESC );
 }
 return( flag );
}

/*----------------------------------------------------------------------*/
/*
 * tpio_test_esc
 * Uses csproto.h escproto.h esc_types.h
 */

static logical tpio_test_esc( Tpio tpio, FioFileList list, FioStack iofs, char* buf )
{
 logical flag;
 character esc_char;
 esc_char = esc_special_char( ESC_CHAR );
 flag = tpio_loop_check( tpio, list, buf );
 if ( !flag ) {
  if ( buf[0] == esc_char ) {
   if ( buf[1] == esc_char ) { cs_copy( &buf[1], buf ); return( flag ); }
   flag = tpio_esc( tpio, list, iofs, &buf[1] );
  }
 }
 return( flag );
}


/*----------------------------------------------------------------------*/
/*
 * tpio_token
 *
 * Read new buffer if needed until token found. Handle escapes.
 * Pop all stacks if EOF found.
 * Uses tokenproto.h, ioproto.h
 */

void tpio_tp_token( Tpio tpio, FioFileList list, FioStack iofs, char* prompt, char** pptr, long mode, char* token, long siz )
{

 logical loop;
 logical status;
 TextCard tbuf;
 long level, pos;

 if ( pptr )
 {
  cs_get_token( pptr, token, mode, siz );
  return;
 }
 if ( !tpio ) tpio=tpio_get();
 if ( !list ) list = fio_get_list();
 if ( !iofs ) iofs = fio_get_stack();
 loop = UT_TRUE;
 status = UT_FALSE;
 while ( loop ) {
  if ( ut_dbg_level(UT_VERBOSE_FIO) ) {
   tpio_loc( tpio, iofs, &level, &pos );
   sprintf( tbuf, "TPIO TOKEN on Level %-6ld Pos %-6ld", (long)level, (long)pos );
   ut_msg_d( UT_VERBOSE_FIO, tbuf );
  }
  if ( *tpio->ptr == NULL ||  **tpio->ptr == '\0' ) 
  { 
    status = tpio_getbuf( tpio, iofs, prompt );
   if ( ut_dbg_level(UT_VERBOSE_FIO) ) {
    tpio_loc( tpio, iofs, &level, &pos );
    sprintf( tbuf, "TPIO new buffer on Level %-6ld Pos %-6ld", (long)level, 0L );
    ut_msg_d( UT_VERBOSE_FIO, tbuf );
   }
  }
  cs_get_token( tpio->ptr, token, mode, siz );
  loop = tpio_test_esc( tpio, list, iofs, token );
 }

 if ( ut_dbg_level(UT_VERBOSE_FIO) ) {
  sprintf( tbuf, "TPIO TOKEN =:%s: REM =:%s: %c", *tpio->ptr, token,
   status ? ' ' : '*' );
  ut_msg_d( UT_VERBOSE_FIO, tbuf );
 }
}

/*----------------------------------------------------------------------*/
/*
 * tpio_in
 * Uses ioproto.h fileproto.h
 * Resets TPIO input to the file named in the argument
 */

void tpio_in( char* buf )
{
 tpio_tp_in( NULL, NULL, NULL, buf );
}

void tpio_tp_in( Tpio tpio, FioFileList list, FioStack iofs, char* buf )
{

 FioFile file = NULL;
 long in = 0;
 TextCard tbuf;

 if ( !iofs ) iofs = fio_get_stack(); 
 if ( !list ) list = fio_get_list();
 if ( cs_eq( buf, "*" ) ) 
 {
  ut_msg_d( UT_VERBOSE_FIO , "Opened interactive command level" );
 } else {
  file = fio_list_open_file( list, buf, "SEQ R", 0 );
  if ( !file ) {
   sprintf( ut_tmp, "Failed to open command file %s", buf );
   ut_err_util( ut_tmp, UT_ERR_TPIO_OPEN_FAIL );
   fio_iostack_iclear( list, iofs );
  } else {
   in = fio_file_fid( file );
   sprintf( tbuf, "OPEN COMMAND FILE %s CHANNEL %ld", buf, (long)in );
   ut_msg_d( UT_VERBOSE_FIO, tbuf );
  }
 }
 fio_iostack_ipush( iofs, tpio, list, file );
}

/*----------------------------------------------------------------------*/
/*
 * tpio_out
 * Uses ioproto.h fileproto.h
 * Resets TPIO output to the file in the argument
 * If syserr is true, sends errors there too.
 * If append is true, appends output to an existing file.
 */

void tpio_out( char* buf, logical append, logical syserr )
{
 tpio_tp_out( NULL, NULL, buf, append, syserr );
}

void tpio_tp_out( FioFileList list, FioStack iofs, char* buf, logical append, logical syserr )
{
 long out = 0;
 FioFile file = NULL;

 if ( cs_eq( buf, "*" ) ) {
  file = NULL;
 } else if ( append ) {
  file = fio_list_open_file( list, buf, "SEQ A", 0 );
 } else {
  file = fio_list_open_file( list, buf, "SEQ W", 0 );
 }
 out = fio_file_fid( file );
 fio_iostack_opush( iofs, list, file );
 if ( syserr ) fio_stack_set_err( file );
}

/*
 *--------------------------------------------------------------------------------
 * Handles TPIO loop control
 */

TpioLoop* tpio_loop_push( Tpio tpio )
{
 TpioLoop* loop;
 if ( tpio->nloop >= tpio->maxloop ) 
 {
  tpio->maxloop *= 2;
  tpio->loop = realloc( tpio->loop, tpio->maxloop * sizeof ( TpioLoop ));
  sprintf( ut_tmp, "TPIO loop stack size increased to %ld", tpio->maxloop );
  ut_msg_d( UT_VERBOSE_FIO, ut_tmp );
 }

 tpio->nloop++;
 loop = &tpio->loop[tpio->nloop-1];
 return loop;
}

void loop_do( Tpio tpio, FioStack iofs )
{
 TextCard buf0;
 char* buf;
 FioFile file;
 TpioLoop* loop;

 buf = buf0;

 if ( !tpio ) tpio=tpio_get();

 cs_get_token( tpio->ptr, buf, TMODE_STRING, CARD_SIZE );
 loop = tpio_loop_push( tpio );
 cs_copy_token_esc( &buf, "=", " ", 0, loop->lvar, WORD_SIZE );
 if ( cs_eq( loop->lvar, "EOF" ) ) 
  {
   loop->value = 0;
   loop->step = 0;
   loop->maxval = 0;
  } else {
   loop->value = cs_get_i( &buf );
   loop->maxval  = cs_get_i( &buf );
   loop->step =  ( buf == NULL ) ? 1 : cs_get_i( &buf );
  }
 file = fio_iostack_file( iofs, FIO_INPUT );
 loop->addr = fio_file_get_lineno( file );
}

/*
 *--------------------------------------------------------------------------------
 * Handles loop iteration and end of loop condition
 */
void loop_end( Tpio tpio, FioStack iofs )
{
 logical finish;
 TpioLoop* loop;
 if ( !tpio ) tpio=tpio_get();

 if ( tpio->nloop == 0 ) {
  ut_err_util( "TPIO: Loop END without DO", UT_ERR_TPIO_LOOP );  
 } else {
  loop = &tpio->loop[tpio->nloop-1];
  if ( loop->step == 0 ) {
   finish = fio_any_eof( 1 );
  } else {
   loop->value += loop->step;
   if ( loop->step >= 0 ) {
    finish = ( loop->value > loop->maxval );
   } else {
    finish = ( loop->value < loop->maxval );
   }
  }
  if ( finish ) {
   loop->value = 0;
   tpio->nloop--;
  } else {
   loop_goto( iofs, loop->addr );
  }
 }
}

/*
 *--------------------------------------------------------------------------------
 * Go to a specified line in a sequential text file
 */

void loop_goto( FioStack iofs, long addr )
{
 long i;
 TextCard buf;
 FioFile file;

 fio_iostack_rewind_input( iofs );
 file = fio_iostack_file( iofs, FIO_INPUT );
 for ( i = 0; i < addr; i++ ) 
 {
  fio_read_line( file, buf, CARD_SIZE );
 }
}

/*
 *--------------------------------------------------------------------------------
 * Echo text substituting $I macros
 */

void loop_echo( Tpio tpio, FioStack iofs )
{

 character macro_char;
 TextBuf buf;
 TextWord old, new;
 TpioLoop* loop;
 long i;
 if ( !tpio ) tpio=tpio_get();

 cs_get_token( tpio->ptr, buf, TMODE_STRING, CARD_SIZE );
 macro_char = esc_special_char( MACRO_CHAR );
 if ( cs_index_char( buf, macro_char ) > 0 ) {
  for ( i = 0; i < tpio->nloop; i++ ) {
   *old = macro_char;
   loop = &tpio->loop[i];
   cs_copy( loop->lvar, old+1 );
   cs_write_i( loop->value, new, WORD_SIZE );
   cs_string_replace( buf, old, new );
  }
 }
 fio_iostack_msg( iofs, NULL, buf, UT_TRUE, UT_FALSE );
}

/*
 *--------------------------------------------------------------------------------
 * Display TPIO internal state
 */
void tpio_disp( long i )
{
 tpio_tp_disp( NULL, NULL, NULL, i );
}

void tpio_tp_disp( Tpio tpio, FioStack iofs, FioFileList list, long i )
{
 long idbg;

 character clev[8];

 if ( !iofs ) iofs = fio_get_stack();
 idbg = ut_dbg_get();
 if ( idbg >= i ) {
  sprintf( clev, "[*%ld]", (long)i );
  if ( idbg > UT_VERBOSE_FIO ) { 
   fio_list_print( list );
   fio_iostack_print_state( iofs );
  } else {
   fio_iostack_print_top( iofs, "I", clev );
  }
  tpio_tp_print_state( tpio, iofs, list );
 } 

}
/*
 *--------------------------------------------------------------------------------
 */
void tpio_print_state( void )
{
 tpio_tp_print_state( NULL, NULL, NULL );
}

void tpio_tp_print_state( Tpio tpio, FioStack iofs, FioFileList list  )
{
 long level;
 TextBuf buf;
 TextWord spacer;
 long min_lev;
 long lev;
 char* gptr;
 long pos;
 TextWord eflag;
 TextBuf disp;
 TextWord stat;

 (void) list;

 if ( !tpio ) tpio=tpio_get();

 level = fio_iostack_get_level( iofs, FIO_INPUT );
 tpio_level_check( tpio, level );
 fio_iostack_msg( iofs, NULL, " ", UT_TRUE, UT_FALSE );
 fio_iostack_dashline( iofs, 40 );
 fio_iostack_msg( iofs, NULL, " JCMLIB Command Processor (TPIO) Status ", UT_TRUE, UT_FALSE );
 fio_iostack_dashline( iofs, 40 );
 sprintf( buf, "  TPIO Stack Level: %ld", (long)level );
 fio_iostack_msg( iofs, NULL, buf, UT_TRUE, UT_FALSE );
 fio_iostack_msg( iofs, NULL, "  ------------------- ", UT_TRUE, UT_FALSE );
 fio_iostack_msg( iofs, NULL, "  | Level Pos Flag  | ", UT_TRUE, UT_FALSE );
 cs_copy( "  |                 | " , spacer );
 min_lev = tpio->cmd_mode ? 0 : 1;
 for ( lev = level; lev >= min_lev; lev-- ) {
  gptr = tpio->buffer[lev];
  if ( tpio->bptr[lev] == NULL ) {
   pos = -1;
   cs_copy( "Empty", eflag );
  } else {
   pos = tpio->bptr[lev] - tpio->buffer[lev];
   cs_copy( " ", eflag );
  }
  sprintf( stat, "  | %-6ld%-4ld%-6s| ", (long)lev, (long)pos, eflag );
  if ( pos == 0 ) {
   cs_copy( ":^", disp );
  } else if ( pos < 0 ) {
   cs_copy( "<", disp );
  } else if ( pos > 250 ) {
   cs_copy( ":?", disp );
  } else {
   sprintf( disp, ":%-250s", " " );
   disp[pos] = '^';
  }
  sprintf( buf, "%-21s:%s", stat, tpio->buffer[lev] );
  fio_iostack_msg( iofs, NULL, buf, UT_TRUE, UT_FALSE );
  sprintf( buf, "%-21s%s", spacer, disp );
  fio_iostack_msg( iofs, NULL, buf, UT_TRUE, UT_FALSE );
 } 
 cs_copy( "  ------------------ ", buf );
 fio_iostack_msg( iofs, NULL, buf, UT_TRUE, UT_FALSE );
 fio_iostack_msg( iofs, NULL, " ", UT_TRUE, UT_FALSE );
}

/*
 *--------------------------------------------------------------------------------
 * Top level TPIO special command routine
 */


void tpio_tp_cc( Tpio tpio, FioFileList list, FioStack iofs, char* prompt, char* buf, long siz )
{

 if ( !tpio ) tpio=tpio_get();

 if ( tpio->cmd_mode ) {
  tpio_tp_cc_cmd( tpio, list, iofs, prompt, &tpio->cmd_mode, buf, siz );
 }
 if ( !tpio->cmd_mode ) {
  tpio_c( prompt, buf, siz  );
  if ( ut_dbg_get() >= UT_VERBOSE_FIO ) {
   sprintf( ut_tmp, "TPIO_CC READ WORD %s", buf );
   ut_msg_d( UT_VERBOSE_FIO, ut_tmp );
  }
  tpio_tp_cc_start_check( tpio, buf, &tpio->cmd_mode );
  if ( tpio->cmd_mode ) tpio_tp_cc_cmd( tpio, list, iofs, prompt, &tpio->cmd_mode, buf, siz );
 }
}

/*
 *--------------------------------------------------------------------------------
 * Check if we are in a special mode
 */

void tpio_tp_cc_start_check( Tpio tpio, char* buf, logical* mode )
{
 TextCard ubuf;

 if ( !tpio ) tpio=tpio_get();

 if ( *buf == esc_special_char( ESC_CHAR ) ) {
  cs_copy( buf+1, ubuf );
  if ( cs_is_blank( ubuf ) || cs_eq( ubuf, "TABLE" ) ) {
   *mode = UT_TRUE;
   ut_msg_d( UT_VERBOSE_FIO, "TPIO_CC SET CMD TRUE" );
   cs_get_token( tpio->ptr, tpio->cmd, TMODE_LINE, TEXT_SIZE );
   tpio_cclear( tpio, 0 );
  }
 }
}

/*
 *--------------------------------------------------------------------------------
 */

void tpio_tp_cc_cmd( Tpio tpio, FioFileList list, FioStack iofs, char* prompt, logical* mode, char* buf, long siz )
{
 logical eot;
 logical found_esc;

 logical verbose;

 if ( !tpio ) tpio=tpio_get();
 verbose = ( ut_dbg_get() >= UT_VERBOSE_FIO );

 if ( cs_is_blank( tpio->cmd ) ) {
  *mode = UT_FALSE;
  if( verbose ) ut_msg_d( UT_VERBOSE_FIO, "TPIO_CC SET CMD FALSE (EMPTY)" );
 } else {
  tpio_tp_cc_check( tpio, list, iofs, prompt, &eot, &found_esc, buf, siz );
  if ( eot ) {
   *mode = UT_FALSE;
   if( verbose ) ut_msg_d( UT_VERBOSE_FIO, "TPIO_CC SET CMD FALSE" );
  }
  if ( *mode  && !found_esc ) {
   cs_get_token( tpio->ptr, buf, TMODE_WORD, siz );
   if ( verbose )
   {
    sprintf( ut_tmp, "TPIO_CC READ CMD %s", buf );
    ut_msg_d( UT_VERBOSE_FIO, ut_tmp );
   }
  }
 }
}

/*
 *--------------------------------------------------------------------------------
 */

void tpio_tp_cc_check( Tpio tpio, FioFileList list, FioStack iofs,
 char* prompt, logical* eot, logical* found_esc, char* buf, long siz )
{

 TextWord etoken;
 character esc;
 char* save_ptr;

 esc = esc_special_char( ESC_CHAR );
 *eot = UT_FALSE;
 *found_esc = UT_FALSE;
 if ( !tpio ) tpio=tpio_get();

 if ( !eot ) {
  save_ptr = *tpio->ptr;
  cs_get_token( tpio->ptr, etoken, TMODE_WORD, WORD_SIZE );
  if ( ut_dbg_level( UT_VERBOSE_FIO ))
  {
   sprintf( ut_tmp, "TPIO_CC CHECK %s", etoken );
   ut_msg_d( UT_VERBOSE_FIO, ut_tmp );
  }
  if ( *etoken == esc && *(etoken+1) != esc ) {
   cs_upper( etoken );
   tpio_tp_cc_esc( tpio, list, iofs, etoken + 1, prompt, eot, found_esc, buf, siz );
  }
 }
 if ( ! (*found_esc) ) {
  *tpio->ptr = save_ptr;
 }
}

/*
 *--------------------------------------------------------------------------------
 */

void tpio_tp_cc_esc( Tpio tpio, FioFileList list, FioStack iofs,
  char* ecmd, char* prompt, logical* eot, 
 logical* found_esc, char* buf, long siz )
{

 *(found_esc) = UT_TRUE;
 if ( !tpio ) tpio=tpio_get();

 if ( cs_is_blank( ecmd ) || cs_eq( ecmd, "ENDTABLE" ) ) {
  *eot = UT_TRUE;
  ut_msg_d( UT_VERBOSE_FIO, "TPIO_CC EOT FOUND" );
 } else if ( cs_eq( ecmd, "C" ) ) {
  tpio_tp_token( tpio, list, iofs, prompt, NULL, TMODE_WORD, buf, siz );
 } else if ( cs_eq( ecmd, "D" ) ) {
  cs_get_token( tpio->ptr, buf, TMODE_WORD, siz );
  sprintf( ut_tmp, "TPIO_CC SKIP CMD %s", buf );
  ut_msg_d( UT_VERBOSE_FIO, ut_tmp );
  cs_copy( " ", buf );
 } else {
  *found_esc = UT_FALSE;
 }

}

/*
 *--------------------------------------------------------------------------------
 */


void tpio_setbuf( char** arr, long n )
{
 tpio_tp_setbuf( NULL, NULL, arr, n );
}

void tpio_tp_setbuf( Tpio tpio, FioStack iofs, char** arr, long n )
{
 character sep;
 char* ptr;
 long i;
 long siz = UT_TPIO_BUFLEN;
 long le;
 if ( !tpio ) tpio = tpio_get();
 if ( !iofs ) iofs = fio_get_stack();
 sep = esc_special_char( STRING_SEP );
 ptr = *tpio_set( tpio, iofs );
 if (!cs_is_blank(ptr) ) {
  le = cs_ends( ptr );
  ptr += le;
  *ptr = sep;
  ptr++;
 }
 for ( i = 0; i < n; i++ ) { 
  cs_put_token( arr[i], &ptr, sep, &siz );
 }
}


/*
 *  Specialized command argument processing for PRECESS family
 * This still needs some work to make it a clone of the Fortran version.
 * 1997 Nov 27 JCM
 */
/*
 *--------------------------------------------------------------------------------
 */

void tpio_cmdargs( 
 char** args,         /* (i) Command line args */
 long nargs,       /* (i) Number of args */
 logical* interact,   /* (o) Interactive? */
 logical* opts        /* (o) Command line was not blank */
)
{
 TextBuf argline;
 TextCard clargs;  
 FioFile in;
 FioFile out;
 long in_fid;
 long out_fid;
 TextCard buf;
 char* ptr = NULL;
 Filename infile;
 Filename outfile;
 TextWord dargs;
 character sep[2];
 long id;
 FioFileList list;
 FioStack iofs;
 Tpio tpio;
 character esc;
 tpio = tpio_get();
 iofs = fio_get_stack();
 list = fio_get_list();


/* Reconstitute the original command line */
 sys_cmdline( args, nargs, argline, TEXT_SIZE );
 sep[0] = esc_special_char( ARG_SEP );
 sep[1] = '\0';
 esc = esc_special_char( ESC_CHAR );
 *interact = UT_TRUE;
 *opts = !cs_is_blank( argline );
 if ( *opts ) {

/* There are some command line arguments */

  ptr = argline;
/* Read the first token and check if it is DEBUG or DEBUGn */
  cs_copy_token_esc( &ptr, sep, " ", esc, clargs, CARD_SIZE );
  cs_copy_siz( clargs, dargs, 5 );
  cs_upper( dargs );
  if ( cs_eq( dargs, "DEBUG" ) ) {
   if ( !cs_is_blank( clargs+5 ) ) {
    id = cs_read_i( clargs+5 );
   } else {
    id = 10;
   }
   if ( ptr == NULL ) {
    cs_copy( " ", clargs );
   } else {
    cs_copy_token_esc( &ptr, sep, " ", esc, clargs, CARD_SIZE );
   }
  }
/* Get another token: this will be the input file, open it */
  *interact = ( ptr == NULL );
  if ( ptr != NULL ) {
   cs_copy_token_esc( &ptr, sep, " ", esc, infile, FN_SIZE );
   if ( !cs_is_blank( infile ) ) {
    in = fio_list_open_file( list, infile, "SEQ R", 0 );
    in_fid = fio_file_fid( in );
    fio_iostack_ipush( iofs, tpio, list, in );
    if ( in ) {
      sprintf( buf, "Opened input file %s", infile ); 
      fio_tmsg( buf );
     }
   }
  }
/* Get another token: this will be the output file,open it */
  if ( ptr != NULL ) {
   cs_copy_token_esc( &ptr, sep, " ", esc, outfile, FN_SIZE );
   if ( !cs_is_blank( outfile ) ) {
    out = fio_open_aw( outfile );
    out_fid = fio_file_fid( out );
    fio_iostack_opush( iofs, list, out );
    if ( out ) { 
      sprintf( buf, "Opened output file %s", outfile ); 
      fio_tmsg( buf ); 
     }
   } 
  }
/*  If you have arguments, put them in the TPIO buffer to be executed */
  if ( !cs_is_blank( clargs ) ) {
   ptr = clargs;

   tpio_tp_setbuf( tpio, iofs, &ptr, 1 );
  }
 }
 
}

