/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

#include "utlib.h"

static logical fio_set_mode( char* jmode, char* mode );

/*----------------------------------------------------------------------*/
/* 
 * fio_read_line
 *
 * Read a line from an ASCII file.
 * Uses  stdin and type FILE*  from <stdio.h>
 * Uses fgets from <stdio.h>
 * Uses fileproto.h: fio_file_fptr and fio_set_eof
 * Uses csproto.h: cs_null and cs_remove_nl from csproto.h
 */



logical fio_write_line( FioFile file, const char* buf )
{
 logical ios;
 if ( !file ) {
  ios = fio_tty_write( buf );
 } else if ( buf == NULL ) {
  if( fio_file_get_pos( file ) > 0 ) {
   ios = fio_put_line( file, " ", UT_TRUE, UT_TRUE );
  } else {
   ios = UT_TRUE;
  }
 } else {
  ios = fio_put_line( file, buf, UT_TRUE, UT_TRUE );
 }
 return( ios );
}

logical fio_write_text( FioFile file, char* buf )
{
 logical ios;
 if ( !file ) {
  ios = fio_tty_writel( buf );
 } else {
  ios = fio_put_line( file, buf, UT_FALSE, UT_FALSE );
 }
 return( ios );
}

long fio_terminate_buf( long pos, const char* buf, char* buf2, logical strip, logical nl, long siz )
{
 long i;
 if ( strip ) {
  i = min_i( cs_ends( buf ), siz-1 );
 } else {
  i = min_i( strlen( buf ), siz-1 );
 }
 cs_copy_siz( buf, buf2, i );
 if( nl ) {
  buf2[i++] = '\n';
  pos = 0;
 } else {
  pos += i;
 }
 buf2[i] = '\0';
 return( pos );
}

logical fio_read_line( FioFile file, char* line, long lmax )
{
 logical ios;
 if ( !file ) {
  ios = fio_tty_read( line, lmax );   
 } else {
  ios = fio_get_line( file, line, lmax );  
  fio_set_status_pos( file, ios, 0 );
 }
 return( ios );
}

logical fio_read_line_alloc( FioFile file, char** line, long* lmax )
{
 logical ios;
 if ( !file ) {
  ios = fio_tty_read( *line, *lmax );   
 } else {
  ios = fio_get_line_alloc( file, line, lmax, UT_TRUE );  
  fio_set_status_pos( file, ios, 0 );
 }
 return( ios );
}

FioFile fio_list_open_file( FioFileList list, const char* filename, char* mode, long rec_bytes )
{
 logical retry;
 logical loop;
 Filename fullname;
 FioFile file = NULL;
 FioFile tty_err;
 long tty_err_fid;

 
 if ( !list ) 
  list = fio_get_list();

 if ( cs_eq( mode, "RAW" ) || cs_eq( mode, "CIO RAW") )
  return fio_list_open_raw( list );

 if ( cs_is_blank( filename )  || cs_eq( filename, "*" ) ) {
  return NULL;  
 }

 if (! cs_is_blank( filename ) ) {
  cs_copy_siz( filename, fullname, FN_SIZE );
  retry = fio_list_get_retry( list );
  loop = UT_TRUE;
  while ( loop ) {
   file = fio_list_open_generic( list, fullname, mode, rec_bytes );
   loop = !file && retry;
   if ( loop ) {
    tty_err_fid = fio_list_tty_err( list );
    tty_err = fio_list_get_file( list, tty_err_fid );
    fio_write_text( tty_err, "OPEN RETRY: Try again, full filename (Q to quit): " );
    fio_get_line( NULL, fullname, FN_SIZE );
    if ( cs_eq( fullname, "Q" ) || cs_eq( fullname, "q" ) ) {
     loop = UT_FALSE;
     cs_copy( " ", fullname );
     sprintf( ut_tmp, "OPEN RETRY: Abandoning attempt to open file %s", filename );
     fio_write_line( tty_err, ut_tmp );
    }
   }
  }
 }
 return( file );
}

/*
 *  Here we convert from JCMLIB's file modes to stdio.h's file modes.
 */
static logical fio_set_mode( char* jmode, char* mode )
{ 
 static char* mode_list[] = { "SEQ R", "SEQ W", "SEQ A", "SIO R", "SIO W", "SIO RW",
                       "CIO R", "CIO W", "CIO RW" };
 static char* mode_val[]  = { "r", "w", "a", "r", "w", "r+", "cr", "cw", "crw" };
 static logical mode_exist[] = { UT_TRUE, UT_FALSE, UT_TRUE, UT_TRUE, UT_FALSE, UT_TRUE, UT_TRUE, 
                          UT_FALSE, UT_TRUE };
 long nmodes = 9;
 long mode_no;
 mode_no = ar_cmatch_c( jmode, mode_list, nmodes );
 if ( mode_no >= 0 ) {
  cs_copy( mode_val[mode_no], mode );
  return( mode_exist[mode_no] );
 } else {
  cs_copy( "x", mode );
  sprintf( "Unknown FIO mode %s", jmode );
  ut_err_util( ut_tmp, UT_ERR_FIO_BAD_MODE );
  return( UT_FALSE );
 }
}


FioFile fio_list_open_generic( FioFileList list, const char* filename, char* jmode, long rec_bytes )
{
 FioFile file = NULL;
 FILE *file_ptr; 
 long fid; 
 TextWord mode;
 Filename truefile;
 Filename pathfile;
 long k;

 cs_copy_siz( filename, pathfile, FN_SIZE );
 k = cs_ends( pathfile );
 pathfile[k] = '\0';
 if ( fio_set_mode( jmode, mode ) ) {
  fio_list_inquire_path( list, pathfile, truefile, FN_SIZE );
 } else {
  cs_copy( pathfile, truefile );
 }
 fid = 0;
 if ( cs_ss_char( "rwa", *mode ) ) {
  file_ptr = fopen( truefile, mode );
  file = fio_file_alloc( list, filename, jmode, UT_FILE_OPEN, file_ptr );
 } else if ( *mode == 'c' ) {
  file = fio_open_c( list, truefile, jmode, mode );
 } else {
  sprintf( ut_tmp, "FIO OPEN FILE: Bad mode %s %ld\n", mode, rec_bytes );
  ut_msg_d( UT_VERBOSE_FIO, ut_tmp );
 }
 if ( !file ) { 
  sprintf( ut_tmp, "FIO OPEN: Failed to open file %s", filename );
  ut_err_util( ut_tmp, UT_ERR_FIO_OPEN_FAIL );
 }
 return( file );
}


