/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*======================================================================*/
/*
 *   cs lib 
 *
 *   C char* routines to manipulate char*s
 *
 *   char*.h: strlen 		used in cs_remove_nl
 *
 */
/*======================================================================*/

#include "utlib.h"



/*----------------------------------------------------------------------*/
/*
 *   cs_remove_nl
 *
 *   Remove trailing \r and \n chars from a line.
 *
 *   1996 Aug 26 JCM
 *   Args:
 *	ptr		(i/o)  char* to remove trailing newlines from
 */

void cs_remove_nl( char* ptr )
{
 long length;

 length = strlen( ptr );
 if ( length ) {
  ptr += length - 1;
  while ( length && ( *ptr == '\n' || *ptr == '\r' ) ) { *ptr-- = '\0'; length--; }
 }
}


/*----------------------------------------------------------------------*/
/*
 *   cs_detab
 *
 *   Change tabs to spaces in a C char*
 *   1996 Aug 26 JCM
 *   Args:
 *	buf		(i/o)	char* to remove tabs from
 *
 */

void cs_detab( char* buf )
{
 while( *buf != '\0' ) {
  if ( *buf == '\t' ) *buf = ' ';
  buf++;
 }
}


/*----------------------------------------------------------------------*/
/*
 *   cs_decmt
 *
 *   Remove comments from a char* by placing a null at the first comment char.
 *   Goes through the char* changing occurrences of esc-cmt to cmt until
 *   a null is encountered or until a cmt without an esc is encountered.
 * 
 *   1996 Aug 26 JCM
 *   
 *   Args:
 *	buf		(i/o) 	char* to remove comments from
 *	esc		(i)	Escape char
 *	cmt		(i)	Comment char
 *
 */


void cs_decmt( char* buf, const char esc, const char cmt )
{
 while ( *buf != '\0' ) {
  if ( buf[0] == esc && buf[1] == cmt ) {
   cs_copy( buf+1, buf );   /* Move everything back one byte */
  }  
  else if ( *buf == cmt ) {
   *buf = '\0';
   return;
  }
  buf++;
 }
}


/*----------------------------------------------------------------------*/
/*   cs_compress_spaces
 *  
 *   Copy source char* to destination, while
 *   reducing multiple spaces to a single space.
 *   1996 Aug 26
 *
 *   Args:
 *	source		(i)	Source char*
 *	dest		(o)	Output char*
 */
void cs_compress_spaces( const char* isource, char* dest )
{
 char* source = (char*)isource;
 while ( ( *dest = *source++ ) != '\0' ) 
 {
  if ( *dest++ == ' ' ) 
  {
   while (*source == ' ' ) source++; 
  }
 }
}


/*----------------------------------------------------------------------*/
/*
 *+cs_upto
 * Copy source to dest up to first occurrence of char targ
 */
char* cs_upto( const char* isource, const char targ, char* cs_upto_buf )
{
 char* source = (char*)isource;
 char* ptr;
 ptr = cs_upto_buf;
 while ( *source != '\0' && *source != targ ) {
  *ptr = *source++;
  ptr++;
 }
 *ptr = '\0';
 return( cs_upto_buf );
}




void cs_string_replace( char* cbuf, const char* old, const char* new )
{
 long lold;
 long lnew;
 long lbuf;
 TextBuf buf;
 long i;
 char* iptr;
 char* optr;
 char* iend;
 lold = strlen( old );
 lnew = strlen( new );
 lbuf = strlen( cbuf );
 iend = cbuf + lbuf - 1;
 i = 1;
 iptr = cbuf;
 optr = buf;
 while ( i > 0 ) { 
  i = cs_index( iptr, old );  /* Location of OLD */
  if ( i > 0 ) {
   cs_get_ss( iptr, optr, 1, i-1 );
   iptr += i-1;
   optr += i-1;
   cs_put_ss( new, optr, 1, lnew );
   iptr += lold;
   optr += lnew;
   if ( iptr > iend ) i = 0;
  }
 }
 if ( iptr < iend ) {
  cs_copy( iptr, optr );
 } else {
  *optr = '\0';
 }
 cs_copy( buf, cbuf );
}


/*
 *   cs_match
 *
 *  Return array index of first equality match of char* in Fortran-style char* array
 *  The size argument does not include the terminating null.
 *  Return 0 if no match
 *  1996 Aug 26 JCM
 *  Args:
 *	buf		(i) Search char*
 *	array		(i) Pointer to first char in array
 *	n		(i) Size of array
 *      size		(i) Size of each array char*; Number of bytes to advance pointer less one
 */

long cs_fs_match( const char* buf, const char* iarray, const long n, const long size1 )
{
 char* array = (char*)iarray;
 long no;
 long size = size1;
 size++;  /* Allow for terminating null */

 for ( no = 1; no <= n; no++ )
  {
  if ( cs_eq( buf, array ) ) return( no );
  array += size;
   }
 return( 0 );
}

/* Version of strcmp supporting null char*s */
logical cs_strcmp( char* a, char* b )
{
 if ( !a )
 {
  if ( !b ) return 0;
  return 1;
 } else if ( !b ) {
  return -1;
 } else {
  return strcmp( a, b );
 } 
}
      
      

/*
 * Concatenate with separator
 */
char* cs_dup_cat( char* a, char* b, char* sep )
{
 long n;
 char* p;
 if ( !a )
  return cs_dup( b );
 if ( !b )
  return cs_dup( a );

 n = strlen(a)+strlen(b)+strlen(sep) +1;
 p = malloc(n);
 sprintf( p, "%s%s%s", a, sep, b );
 return p;
}

