/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

#include "utlib.h"


double utm_eexp( const double x )
{
 double result;
 if ( x > UT_MAX_EXP_ARG ) {
  result = ut_infinity_d( 1 );
 } else if ( x < -UT_MAX_EXP_ARG ) {
  result = 0.0;
 } else {
  result = exp( x );
 }
 return( result );
}

double utm_ecosh( const double x )
{
 double result;
 result = 0.5 * ( eexp( x ) + eexp( -x ) );
 return( result );
}

double utm_esinh( const double x )
{
 double result;
 if ( fabs( x ) < UT_MIN_SINH_ARG ) {
  result = x;
 } else {
  result = 0.5 * ( eexp( x ) - eexp( -x ) ); 
 }
 return( result );

}

double utm_asinh( const double x )
{
 double result;
 double ax;

 ax = fabs( x );
 if ( ax < UT_MIN_SINH_ARG ) {
  result = ax;
 } else if ( ax > UT_MAX_SINH_ARG ) { 
  result = log( 2.0 * ax );
 } else {
  result = log( ax + sqrt( 1.0 + ax * ax ) );
 }
 if ( x < 0.0 ) result *= -1;
 return( result );
}

double utm_atanh( double x )
{
 double f, y;
 f = ( 1.0 + x ) / ( 1.0 - x );
 y = 0.5 * log(  f );
 return( y );
}



double utm_acosh( double x )
{
 double eps;
 double result;
 eps = x - 1.0;
 
 if ( eps < 0.0 ) {
  result = 0.0;
 } else if ( eps < UT_MIN_SINH_ARG ) {
  result = sqrt( 2.0 * eps );
 } else if ( eps > UT_MAX_SINH_ARG ) {
  result = log( 2.0 * x );
 } else {
  result = log( x + sqrt ( x * x - 1.0 ) );
 }
 return( result );
}


double utm_bose_einstein( const double x )
{
 double result;
 if ( x > UT_MAX_EXP_ARG ) {
  result = 0.0;
 } else if ( x < -UT_MAX_EXP_ARG ) {
  result = -1.0;
 } else if ( fabs(x) < UT_MIN_EXP_ARG ) {
  if ( fabs(x) < UT_TINY_D ) {
   result = ut_infinity_d( 1 );
  } else {
   result = 1.0 / x;
  }
 } else {
  result = 1.0 / ( eexp(x) - 1.0 );
 }
 return( result );
}

double utm_dplanckn( const double x, const long n )
{
 double result;
 if ( n == 0 ) {
  result = bose_einstein( x );
 } else if ( x > UT_MAX_EXP_ARG ) {
  result = 0.0;
 } else if ( x < -UT_MAX_EXP_ARG ) {
  result = -1.0 * ipow_d( x, n );
 } else if ( fabs(x) < UT_MIN_EXP_ARG ) {

  if ( fabs( x ) < UT_TINY_D ) {
   if ( n == 1 ) {
    result = 1.0;
   } else if ( n < 0 ) {
    result = ut_infinity_d( 1 );
   } else {
    result = 0.0;
   }
  } else {
   result = ipow_d( x, (n-1) );
  }
 } else {
  result = ipow_d( x, n ) / ( eexp( x ) - 1.0 );
 }
 return( result );
}

/*
 *  Raise to long power
 *   ipow_d( x, 0 ) = 1
 *   ipow_d( x, 1 ) = x
 *   ipow_d( x, -n ) = 1 / x^n
 *   ipow_d( x, n ) = x^n
 */
double utm_ipow_d( const double x, const long m )
{
 long n = m;
 double result;
 result = 1.0;
 if ( n == 0 ) {
  return( result );
 } else if ( n == 1 ) {
  return ( x );
 } else if ( n < 0 ) {
  if ( x == 0.0 ) return( ut_infinity_d( 1 ) );
  while( n++ ) result *= x;
  return( 1.0/result );
 } else {
  if ( x == 0.0 ) return ( x );
  while( n-- ) result *= x;
  return( result );
 }
}

float utm_ipow_r( const float x, const long m )
{
 long n = m;
 float result;
 result = 1.0;
 if ( n == 0 ) {
  return( result );
 } else if ( n == 1 ) {
  return ( x );
 } else if ( n < 0 ) {
  if ( x == 0.0 ) return( ut_infinity_d( 1 ) );
  while( n++ ) result *= x;
  return( 1.0/result );
 } else {
  if ( x == 0.0 ) return ( x );
  while( n-- ) result *= x;
  return( result );
 }
}

/* Ten to the power */

double utm_tpow_d( const long m )
{
 long n = m;
 double result;
 result = 1.0;
 if ( n == 0 ) {
  return( result );
 } else if ( n < 0 ) {
  while( n++ ) result *= 10.0;
  return( 1.0/result );
 } else {
  while( n-- ) result *= 10.0;
  return( result );
 }
}

float utm_tpow_r( const long m )
{
 long n = m;
 float result;
 result = 1.0;
 if ( n == 0 ) {
  return( result );
 } else if ( n < 0 ) {
  while( n++ ) result *= 10.0;
  return( 1.0/result );
 } else {
  while( n-- ) result *= 10.0;
  return( result );
 }
}

double utm_dgauss( const double x, const double x0, const double sigma )
{
 double arg;
 double result;
 if ( sigma == 0.0 ) {
  result = 0.0;
 } else {
  arg = ( x - x0 ) / sigma;
  result = eexp( -0.5 *  ( arg * arg ) );
 }
 return( result );
}

