/*                                                                
**  Copyright (C) 2002-2003,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   error.c                                                     */
/*                                                               */
/*   Functions that set, display, or clear error messages        */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/


#include "utlib.h"
#include "jcmstate.h"
#include "uterr.h"


/***************************************************************/
/*                                                             */
/* Function:   ut_error_set                                    */
/*                                                             */
/* Description:                                                */
/*             Creates and initializes a new error             */
/*                                                             */
/*  Input parameters:                                          */
/*                  Error *in_error - pointer to error stack   */
/*                  char  *text     - error message            */
/*                  long   code     - error number             */
/*                  long   auxcode  - other error number       */
/*                                                             */
/*  Output parameters:                                         */
/*                  none                                       */
/*                                                             */
/*  Returns:                                                   */
/*                  nothing                                    */
/*                                                             */
/***************************************************************/
void ut_error_set( Error* in_error, char* text, long code, long auxcode )
{
  Error error = calloc( 1, sizeof( struct ErrorData_s ));

  if ( !in_error )
    in_error = ut_error_state(); 

  error->msg = cs_dup( text );
  error->code = code;
  error->auxcode = auxcode;
  error->next = *in_error;
  *in_error =  error;

  ut_msg_d( UT_VERBOSE, text );
}

/***************************************************************/
/*                                                             */
/* Function:   ut_error_code                                   */
/*                                                             */
/* Description:                                                */
/*             Return error code                               */
/*                                                             */
/*  Input parameters:                                          */
/*                  Error  error    - pointer to error posting */
/*                                                             */
/*  Output parameters:                                         */
/*                  none                                       */
/*                                                             */
/*  Returns:                                                   */
/*                  long  error_code                           */
/*                                                             */
/***************************************************************/
long ut_error_code( Error error )
{
  if ( error )
    return error->code;
  else
    return 0;
}


/***************************************************************/
/*                                                             */
/* Function:   ut_error_message                                */
/*                                                             */
/* Description:                                                */
/*             Get first error message                         */
/*                                                             */
/*  Input parameters:                                          */
/*                  Error *errorp   - pointer to error stack   */
/*                  long   maxlen   - buffer length            */
/*                                                             */
/*  Output parameters:                                         */
/*                  char  *buf      - buffer for message       */
/*                                                             */
/*  Returns:                                                   */
/*                  nothing                                    */
/*                                                             */
/***************************************************************/
void ut_error_message( Error* errorp, char* buf, long maxlen )
{
  Error error;

  if ( !errorp )
    errorp = ut_error_state();

  error = *errorp;

  if ( error )
    cs_copy_siz( error->msg, buf, maxlen );
  else
    cs_copy( " ", buf );
}


/***************************************************************/
/*                                                             */
/* Function:   ut_error_disp                                   */
/*                                                             */
/* Description:                                                */
/*             Write possible multiple lines to output error   */
/*             stream. If input parameters are NULL, uses      */
/*             default error state and class                   */
/*                                                             */
/*  Input parameters:                                          */
/*                  Error *errorp   - pointer to error stack   */
/*                  ErrorClasses list - error class            */
/*                                                             */
/*  Output parameters:                                         */
/*                  none                                       */
/*                                                             */
/*  Returns:                                                   */
/*                  nothing                                    */
/*                                                             */
/***************************************************************/
void ut_error_disp( Error* errorp, ErrorClasses list )
{
  TextBuf text;
  TextBuf code_text;
  Error error;

  if ( !list )
    list = ut_error_class();

  if ( !errorp )
    errorp = ut_error_state();

  error = *errorp;

  while( error )
  {
    ut_error_class_decode( list, error->code, code_text ); 
    sprintf( text, "%s %s", code_text, error->msg );  
    ut_err_msg( text );
    error = error->next;
  }
}


/***************************************************************/
/*                                                             */
/* Function:   ut_error_clear                                  */
/*                                                             */
/* Description:                                                */
/*             Remove all errors in the error list by freeing  */
/*             storage.  Set input pointer to NULL upon return.*/
/*                                                             */
/*  Input parameters:                                          */
/*                  Error *error_p - pointer to error listor   */
/*                                                             */
/*  Output parameters:                                         */
/*                  Error *error_p = NULL                      */
/*                                                             */
/*  Returns:                                                   */
/*                  nothing                                    */
/*                                                             */
/***************************************************************/
void ut_error_clear( Error* error_p )
{
  Error next;
  Error error;

  if ( !error_p )
    error_p = ut_error_state();

  error = *error_p;

  while( error )
  {
    next = error->next;
    free( error->msg );
    free( error );
    error = next;
  }

  *error_p = NULL;
}


/***************************************************************/
/*                                                             */
/* Function:   ut_error_disp_select_msgs                       */
/*                                                             */
/* Description:                                                */
/*             Display any error messages on error stack       */
/*             that have a particular error code               */
/*                                                             */
/*  Input parameters:                                          */
/*                  int  iErrorCode - error code to look for   */
/*                                                             */
/*  Output parameters:                                         */
/*                  none                                       */
/*                                                             */
/*  Returns:                                                   */
/*                  nothing                                    */
/*                                                             */
/***************************************************************/
void ut_error_disp_select_msgs(long lErrorCode)
{
  TextBuf  text;
  TextBuf  code_text;
  Error    error;
  Error   *errorp;
  ErrorClasses  list;

  list   = ut_error_class();
  errorp = ut_error_state();
  error  = *errorp;

  while( error )
  {
     ut_error_class_decode( list, error->code, code_text ); 

     if ( error->code == lErrorCode)
     {
       sprintf( text, "%s %s", code_text, error->msg );  
       ut_err_msg( text );
     }

     error = error->next;
  }
}
