/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


#include "utlib.h"
#include <stdio.h>

static logical strip = TRUE;

#define FIO_FB_DEFAULT_CHUNK    10000
#define FIO_FB_DEFAULT_MAXLINES 1000
static FileBuffer fio_file_buffer_current;  /* Current output buffer in use */

FileBuffer fio_file_buffer_alloc( long size, long chunk_size, long maxlines )
{
 FileBuffer buffer;

 if ( chunk_size == 0 ) chunk_size = FIO_FB_DEFAULT_CHUNK;
 if ( size == 0 ) size = chunk_size;
 if ( maxlines == 0 ) maxlines = FIO_FB_DEFAULT_MAXLINES;

 buffer = (FileBuffer)malloc( sizeof( struct FileBuffer_s ) );
 buffer->nmax = maxlines;
 buffer->line = (long*)malloc( maxlines * UT_SZ_I );
 buffer->chunk_size = chunk_size;
 buffer->size = size;
 buffer->data = (char*)malloc( size * UT_SZ_C );
 if ( !buffer->data ) {
  free( buffer );
  fio_tmsg( "Failed to allocate memory buffer" );
  return( NULL );
 }
/* Set first line to beginning of data */
 fio_file_buffer_clear( buffer );
 return( buffer );
}

/* Free the file buffer */
void fio_file_buffer_free( FileBuffer buffer )
{
 if ( !buffer ) return;
 free( buffer->data );
 free( buffer->line );
 free( buffer );
}

/* Clear the file buffer */
void fio_file_buffer_clear( FileBuffer buffer )
{
 buffer->n = 1;
 buffer->top = 0;
 buffer->line[0] = 0;
 buffer->pos = 1;
}

logical fio_file_buffer_print_line( FileBuffer buffer )
{
 TextBuf line;
 TextWord tmp;
 long pos;
 cs_blank( line, WORD_SIZE );
 if ( fio_file_buffer_read_line( buffer, line+6, TEXT_SIZE ) ) {
  pos = fio_file_buffer_pos( buffer );
  cs_write_i( pos, tmp, WORD_SIZE );
  cs_put_ss( tmp, line, 1, 6 );
  fio_msg( line );
  return( TRUE );
 } else {
  fio_msg( "EOF" );
  return( FALSE );
 }
}


/* Skip blank lines */
logical fio_file_buffer_skip_blanks( FileBuffer buffer )
{
 TextBuf line;
 logical loop = TRUE;
 logical blank = TRUE;
 while ( loop && blank ) {
  loop = fio_file_buffer_read_line( buffer, line, TEXT_SIZE );
  if ( loop ) blank = cs_is_blank( line );
 }
 fio_file_buffer_move( buffer, -1 ); 
 return( loop );
}

/* Print file buffer status */
void fio_file_buffer_status( FileBuffer buffer )
{
 TextCard buf;
 sprintf( buf, "Buffer size: %ld of %ld bytes %ld of %ld lines", buffer->top, buffer->size, buffer->n, buffer->nmax );
 fio_msg( buf );
}

/* Find all the lines in the file buffer */

void fio_file_buffer_find_lines( FileBuffer buffer )
{
 char* ptr;
 char* top;

 ptr = buffer->data;
 top = ptr + buffer->top;

 while ( ptr < top ) {
  if ( buffer->n >= buffer->nmax-1 ) 
  { 
   buffer->nmax = 2 * buffer->nmax + 10;
   buffer->line = realloc( buffer->line, buffer->nmax * UT_SZ_P ); 
  }
  if ( *ptr++ == '\n' ) {
   buffer->line[buffer->n++] = ptr - buffer->data;
  }
 }
 buffer->line[buffer->n] = buffer->top;
}

/* Return number of lines in file buffer */
long fio_file_buffer_lines( FileBuffer buffer )
{
 return( buffer->n );
}


/* copy bytes from buffer to line until line is full or all of line is taken.
 * Terminate line with null.
 * Increment buffer position.
 * Return false if already at last line.
 */
logical fio_file_buffer_read_line( FileBuffer buffer, char* line, long maxlen )
{
 char* ptr;
 char* top;
 char* buf;
 long n;
 if ( buffer->pos >= buffer->n ) {
  return( FALSE );
 } else {
  n = buffer->pos++;
  buf = line;
  ptr = buffer->data + buffer->line[n-1];
  top = buffer->data + buffer->line[n]-1;  /* Skip trailing newline */
   

  while( maxlen-- && ptr < top ) { *buf++ = *ptr++; }
/* Strip trailing blanks */
  if ( strip && *buf == ' ' ) {
   while( buf > line && *buf == ' ' ) buf--;
   buf++;
  }

  *buf = '\0';
  return( TRUE );
 }
}


/*
 *  Go to specified line
 */
logical fio_file_buffer_seek( FileBuffer buffer, long n )
{
 logical ok;
 buffer->pos = n;
 ok = ( buffer->pos <= buffer->n );
 return ok;
}

logical fio_file_buffer_move( FileBuffer buffer, long n )
{
 logical ok;
 buffer->pos += n;
 ok = ( buffer->pos <= buffer->n );
 return ok;
}

long fio_file_buffer_pos( FileBuffer buffer )
{
 return( buffer->pos );
}

void fio_file_buffer_write_line( FileBuffer buffer, char* line )
{
 fio_file_buffer_write_general_line( buffer, line, TRUE );
}

void fio_file_buffer_write_partial_line( FileBuffer buffer, char* line )
{
 fio_file_buffer_write_general_line( buffer, line, FALSE );
}


void fio_file_buffer_write_general_line( FileBuffer buffer, char* line, logical newline )
{
 char* ptr;
 char* buf;
 char* start;
 long n;
 if ( line == NULL ) return;
 ptr = line;
 n = strlen( line );
 if ( buffer->top + n > buffer->size )
 {
/* Need to enlarge buffer */
   buffer->size = ( buffer->size + buffer->top + n ) * 2 + 10;
   buffer->data = realloc( buffer->data, buffer->size * UT_SZ_C );
 }
 start = buffer->data + buffer->top; 
 buf = start;
 while( *ptr != '\0' ) { *buf = *ptr++; buf++; }
 if( newline ) { 
  if ( strip ) {  /* Strip trailing blanks */
   buf--;
   while( *buf == ' ' && buf > start ) buf--;
   buf++;
  }
  *buf = '\n'; buf++;
  buffer->n++;
  buffer->line[buffer->n] = buf - buffer->data;
 }
 buffer->top = buf - buffer->data;
 if ( buffer->n >= buffer->nmax-1 ) 
 {
  buffer->nmax = buffer->n * 2 + 10;
  buffer->line = realloc( buffer->line, buffer->nmax * UT_SZ_P );
 }
}


/*  
 * Read a file entirely into memory
 */ 
logical fio_file_buffer_read( char* filename, FileBuffer buffer )
{
 FioFile in; 
 logical loop;
 long n;
 char* ptr;

/*
 *  Read data into buffer
 */
 ut_msg_d( 1, "Read file into memory" );
 in = fio_open_sr( filename );
 if ( !in ) return( FALSE );
 loop = TRUE;
 while ( loop ) {
  ptr = buffer->data + buffer->top;
  if ( buffer->top + buffer->chunk_size > buffer->size ) 
  {
/* Need to enlarge buffer */
   buffer->size = ( buffer->size + buffer->top +  buffer->chunk_size ) * 2 + 10;
   buffer->data = realloc( buffer->data, buffer->size * UT_SZ_C );
  }
  loop = fio_read_sio( in, ptr, buffer->chunk_size, &n, UT_SZ_C );
  buffer->top += n;
 }
 fio_file_close( in );
 ut_msg_d( 1, "Find lines" );
 fio_file_buffer_find_lines( buffer );
 ut_msg_d( 1, "Completed buffer read" );
 return( TRUE );
}

/*
 * Write a file from memory to disk
 */

void fio_file_buffer_write( FileBuffer buffer, char* filename )
{
 FioFile file; 
 logical loop;
 long nw;
 long n;
 char* ptr;
 long nleft;
 char tbuf[1024];
/*
 *   Read data to disk from buffer
 */
 sprintf( tbuf, "Writing buffered file %s Records %ld", filename, buffer->top );
 ut_msg_d( 1, tbuf );

 file = fio_open_sw( filename );
 loop = TRUE;
 ptr = buffer->data;
 nleft = buffer->top;
 while ( nleft > 0 && loop ) {
  n = min_i( nleft, buffer->chunk_size );
  loop = fio_write_sio( file, ptr, n, &nw, UT_SZ_C );
  if ( !loop || n != nw ) 
   fio_tmsg( "fio_file_buffer_ write failed" );
  ptr  += nw;
  nleft -= nw;
 }
 fio_file_close( file );
 sprintf( tbuf, "Closed file %s", filename );
 ut_msg_d( 1, tbuf );
}




/* Configure msg to use the file buffer instead of the output stream */

void fio_file_buffer_output( FileBuffer buffer )
{
 if ( buffer ) {
  fio_file_buffer_clear( buffer );
  fio_file_buffer_current = buffer;
  fio_set_msg( fio_file_buffer_msg );
 } else {
  fio_set_msg( NULL );
 }
}

void fio_file_buffer_msg( char* buf, logical newline )
{
 fio_file_buffer_write_general_line( fio_file_buffer_current, buf, newline );
}



