/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include "utlib.h"
#include "fio.h"

long fio_file_get_lineno( FioFile file )
{
 return( file->line );
}

FILE* fio_file_get_fptr( FioFile file )
{
/*
 * This routine uses the fio_file structure definition.
 */
 if ( file )
  return( file->fptr );
 else 
  return( NULL );
}


/*----------------------------------------------------------------------*/
/*
 *  fio_test_eof
 *
 *  Test if file status to EOF
 *
 */
logical fio_file_test_eof( FioFile file )
{
 if ( file )
  return ( file->status == UT_FILE_EOF );
 else
  return ( UT_FALSE ); 
}

logical fio_file_mode_eof( FioFile file, long mode )
{
 return ( fio_file_status( file, mode ) == UT_FILE_EOF );
}

/*
 *--------------------------------------------------------------------------------
 */
/*
 *  Test whether the file exists.
 */
logical fio_inquire( const char* name )
{
 struct stat stat_var;
 return( stat( name, &stat_var ) == 0 );
}

/*
 *--------------------------------------------------------------------------------
 */
/*----------------------------------------------------------------------*/
/*
 *  fio_clear_eof
 *
 *  Set file status to Open
 *
 */


void fio_file_clear_eof( FioFile file )
{
/*
 * Set EOF status
 */
 FILE* fptr;
 if ( file ) {
  file->status = UT_FILE_OPEN;
  fptr = file->fptr;
  if ( fptr != NULL ) clearerr( fptr );
 }
}
 
void fio_file_set_pos( FioFile file, long pos )
{
 if ( file ) 
  file->pos = pos;
}

 
/*----------------------------------------------------------------------*/
/*
 * fio_get_name
  * Return the name of the file
 *
 */

void fio_file_get_name( FioFile file, char* name, const long maxlen )
{
/*
 * This routine uses the fio_file structure definition.
 */
 if ( file ) {
  cs_copy_siz( file->name, name, maxlen );
 } else {
  cs_copy( " ", name );
 }
}

/*
 *--------------------------------------------------------------------------------
 */

long fio_file_get_recl( FioFile file )
{
 if ( file ) {
  return( file->recl );
 } else {
  return( 0 );
 }
}
void fio_file_set_recl( FioFile file, long recl )
{
 if ( file ) 
  file->recl = recl;
}

long fio_file_fid( FioFile file )
{
 if ( file )
  return file->fid;
 else
  return 0;
}


logical fio_file_interactive( FioFile file )
{
 logical q;
 q = !file || cs_eq_siz( file->type, "- TTY", 5 );
 return q;
}

logical fio_file_close( FioFile file )
{
 TextWord type;
 long fd;
 logical status = UT_TRUE;
 int fst = 0;
 if ( file ) { 
  fio_file_get_type( file, type, WORD_SIZE );
  if ( cs_eq_siz( type, "CIO", 3 ) )
  {
   fd = fio_file_get_cfd( file );
   if ( cs_eq( type, "CIO RAW" ) ) 
    cio_clear_raw( file->list, fd );
   fst = close( fd );
  } else {
   fst = fclose( file->fptr );
  }
  fio_file_clear( file );
  status = ( fst == 0 );
 }
 return status;
}



void fio_file_flush( FioFile file )
{
 FILE* fptr = NULL;
 if ( file )
  fptr = file->fptr;

 if ( !fptr )
  fptr = stdout;

 fflush( fptr );
}

logical fio_get_line_alloc( FioFile file, char** bufp, long* sizp, logical dynamic )
{
 logical ios;
 FILE* fptr;
 char* buf = *bufp;
 char* ptr = buf;
 long siz = *sizp;
 long nmax = siz;
 logical loop = UT_TRUE;
 long status;
 char* fst;
 long n;

 if ( !file ) {
  fptr = stdin;
 } else {
  fptr = file->fptr;
 }

 while( loop )
 {
  fst = fgets( ptr, siz, fptr );
  if ( feof( fptr ))
   status = UT_FILE_EOF;
  else if ( ferror( fptr ) )
   status = UT_FILE_ERR;
  else if ( !fst )
   status = UT_FILE_ERR;
  else
   status = UT_FILE_OPEN;

  file->status = status; 
  ios = ( status == UT_FILE_OPEN );
  n = strlen(buf);
  loop = dynamic && ( ios && buf[n-1] != '\n' );
  if ( loop )
  {
   buf = realloc( buf, 2 * nmax );
   ptr = buf + nmax;
   siz = nmax;
   nmax *= 2;
   *bufp = buf;
   *sizp = nmax;
  } 
 }
 if ( ios ) {
   cs_remove_nl( buf );
 } else {
   cs_copy( " ", buf );
 }
 return( ios );
}

/*
 *  Read a line using fgets. If no error, remove trailing newline.
 *  If error, return is the blank buffer. Return value is TRUE if 
 *  no error.
 */
logical fio_get_line( FioFile file, char* buf, long siz )
{
 return fio_get_line_alloc( file, &buf, &siz, UT_FALSE );
}

logical fio_put_line( FioFile file, const char* buf, logical strip, logical nl )
{
 const long siz = 1023;
 char buf2[1024];
 char* ptr;
 logical ios;
 long pos;
 FILE* fptr;
 long clen;
 if( !file ) {
  fptr = stdout;
  pos = 0;
 } else {
  fptr = file->fptr; 
  pos = file->pos;
 }
 ios = UT_TRUE;
 ptr = (char*)buf;
 clen = cs_ends( buf );
 while( clen > siz && ios ) 
 {
  cs_copy_siz( ptr, buf2, siz );
  ios = ( fputs( buf2, fptr ) >= 0 );
  ptr  += siz;
  clen -= siz;
 }
 pos = fio_terminate_buf( pos, ptr, buf2, strip, nl, siz );
 ios = ( fputs( buf2, fptr ) >= 0 ); /* TRUE if good */
 if ( ios ) ios = ( ferror( fptr ) == 0 ) && ( feof( fptr ) == 0 );
 fio_set_status_pos( file, ios, pos ); 
 return( ios );
}

/*
 *--------------------------------------------------------------------------------
 */

void fio_file_set_type( FioFile file, char* type )
{
 if ( file ) 
  cs_copy( type, file->type );
}

/*
 *--------------------------------------------------------------------------------
 */
void fio_file_get_type( FioFile file, char* type, const long maxlen )
{

/*
 * This routine uses the fio_file structure definition.
 */

 if ( file ) {
  cs_copy_siz( file->type, type, maxlen );
 } else {
  cs_copy( " ", type );
 }
}

/*
 *--------------------------------------------------------------------------------
 */
long fio_file_get_pos( FioFile file )
{
 if ( file )
  return( file->pos );
 else 
  return( 0 );
}


/*
 * Convert a char* to a bit pattern for posix access function
 */

logical fio_posix_access( const char* name, const char* mode )
{
  logical retval;
  int xmode = 0;

  xmode |= ( strpbrk(mode, "r") != NULL ? R_OK 	: 0 );  /* Read perm */
  xmode |= ( strpbrk(mode, "w") != NULL ? W_OK 	: 0 );  /* Write perm */
  xmode |= ( strpbrk(mode, "x") != NULL ? X_OK	: 0 );  /* Execute perm */
  xmode |= ( strpbrk(mode, "f") != NULL ? F_OK 	: 0 );  /* Existence  */

  retval = access( name, xmode );
  return ( retval == 0 );
}	


long fio_stat_type( const char* name )
{
 struct stat stat_var;
 int status;
 mode_t mode;


 if ( !name ) return( FIO_NOFILE );

 status = stat( name, &stat_var );
 if ( status == 0 ) {
/*  Is it world accessible? */
  mode = stat_var.st_mode;
  if ( S_ISREG( mode ) || S_ISLNK( mode ) )
  {
   return FIO_FILE_STD;
  } else if ( S_ISDIR( mode )) {
   return FIO_FILE_DIR;
  } else if ( S_ISCHR( mode )) {
   return FIO_FILE_CDEV;
  } else if ( S_ISBLK( mode )) {
   return FIO_FILE_BDEV;
  } else if ( S_ISFIFO( mode )) {
   return FIO_FILE_FIFO;
#if 0
/* Not defined in POSIX... */
  } else if ( S_ISSOCK( mode )) {
   return FIO_FILE_SOCKET;
#endif
  }
 }
 return ( FIO_NOFILE );
}

/*
 *   Test whether you have access permission to the file.
 */
long fio_access( const char* name )
{
 struct stat stat_var;
 int status;
 uid_t uid, fuid;
 gid_t gid, fgid;
 mode_t mode;

 logical r, w;

 if ( !name ) return( FIO_NOFILE );

 status = stat( name, &stat_var );
 if ( status == 0 ) {
/*  Is it world accessible? */
  mode = stat_var.st_mode;

  r = ( mode & S_IROTH );
  w = ( mode & S_IWOTH );
  if ( !r || !w ) {
/* Is it group accessible for your group? */
   fgid = stat_var.st_gid;
   gid = getegid();
   if ( gid == fgid ) {
    r = r || ( mode & S_IRGRP );
    w = w || ( mode & S_IWGRP );
   }
   if ( !r || !w ) { 
/*  Is it accessible for your effective UID? */
     uid = geteuid();
     fuid= stat_var.st_uid;
     if ( uid == fuid ) {
      r =  r || ( mode & S_IRUSR );
      w =  w || ( mode & S_IWUSR );
     }   
   }
  }
/* OK, now we know whether you can read or write to it. */
  if ( r && w ) {
   return( FIO_READWRITE );
  } else if ( r ) {
   return( FIO_READONLY );
  } else {
   return( FIO_NOPERM );
  }
 } else {
  return( FIO_NOFILE );
 }
} 

