/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

#include "utlib.h"
#include "fio.h"
#include "fiolist.h"
#include "jcmstate.h"


/*
 */
void fio_list_free( FioFileList list )
{
 free( list->files );
 fio_list_free_path( list );
 fio_iostack_free( list->iofs );
 fio_pager_free( list->pager );
 free( list );
}

FioFile fio_file_init( FioFileList f, long fid )
{
 FioFile file;
  file = &f->files[fid-1];  
  file->list = f;
  file->fid = fid;
  cs_copy( " ", file->name );
  cs_copy( " ", file->type );
  file->status = UT_FILE_CLOSED;
  file->line = 0;
  file->pos = 0;
  file->recl = 0;
  file->stream = 0;
  file->fptr = (FILE*)NULL;
 return file;

}


FioFileList fio_list_init( long in, long out, long err )
{
 FioFileList fiolist = NULL;
 FioFile file;
 long fid;
 long nfiles = 3;
 fiolist = calloc( 1, sizeof( struct FioFileList_s ));
 if ( err== 0 ) err =FIO_DEFAULT_ERR;
 if ( out== 0 ) out= FIO_DEFAULT_OUT;
 if ( in == 0 ) in = FIO_DEFAULT_IN;
 if ( err > nfiles ) nfiles = err;
 if ( out > nfiles ) nfiles = out;
 if ( in  > nfiles ) nfiles = in;
 fiolist->nmax = UT_FILE_LIST_MAX;
 if ( fiolist->nmax < nfiles ) fiolist->nmax = nfiles;
 fiolist->files = calloc( fiolist->nmax, sizeof( FioFileObject ));
 fiolist->tty_err = err;
 fiolist->tty_out = out;
 fiolist->tty_in = in;
 fiolist->fio_nfiles = nfiles;
/* Initialize the first few files */
 for ( fid = 1; fid <= fiolist->fio_nfiles; fid++ )
 {
  file = fio_file_init( fiolist, fid );
 }
/* Initialize the properties of the standard files */
 file = fio_list_get_file( fiolist, fiolist->tty_in );
 fio_checkin( file, "stdin",  "- TTY IN -", UT_FILE_OPEN, stdin, 0 );
 file = fio_list_get_file( fiolist, fiolist->tty_out );
 fio_checkin( file, "stdout", "- TTY OUT-", UT_FILE_OPEN, stdout, 0 );
 file = fio_list_get_file( fiolist, fiolist->tty_err );
 fio_checkin( file, "stderr", "- TTY ERR-", UT_FILE_OPEN, stderr, 0 );

/* Initialize remaining attributes */
 fio_list_set_retry( fiolist, UT_FALSE );

 fiolist->stdin_is_tty = -1;  /* Postpone determining this till cio_raw is called */

 fiolist->pager = fio_pager_init();
 fiolist->iofs =  fio_iostack_init();
 fio_iostack_reinit( fiolist, fiolist->iofs );
 fio_list_set_path( fiolist, NULL );

 return fiolist;
}



/*
 *--------------------------------------------------------------------------------
 * RETRY
 *--------------------------------------------------------------------------------
 */
void fio_list_set_retry( FioFileList fiolist, logical q )
{
 if ( fiolist ) fiolist->retry = q;
} 

logical fio_list_get_retry( FioFileList fiolist )
{
 if ( fiolist )
  return fiolist->retry;
 return UT_FALSE;
}

/*
 *--------------------------------------------------------------------------------
 * Access to JCM FILE parameters
 *--------------------------------------------------------------------------------
 */
  

long fio_list_n( FioFileList f )
{
 return( f->fio_nfiles );
}

/*
 *  The value fid = 0 and file=NULL is used for stdin/stdout, and is valid data.
 */
FioFile fio_list_get_file( FioFileList f, long fid )
{
 /* This routine uses fio_file static storage */
 if ( fid > 0 ) {
  return( &f->files[fid-1] );
 } else {
  return( NULL );
 }
}




long fio_list_alloc_fid( FioFileList f )
{

 long fid;
 for ( fid = 1; fid <= f->fio_nfiles; fid++ )
 { 
  if ( fio_file_ptr( fid )->status == UT_FILE_CLOSED ) return( fid );
 }
 if ( f->fio_nfiles >= f->nmax ) {
  f->nmax = 2 * f->fio_nfiles;
  f->files = realloc( f->files, f->nmax * sizeof( FioFileObject ));
 }

 f->fio_nfiles++;
 fid = f->fio_nfiles; 
 return( fid );
}
/*----------------------------------------------------------------------*/

FioFile fio_list_alloc_file( FioFileList f )
{
 long fid;
 FioFile file;
 fid = fio_list_alloc_fid( f );
 file = &f->files[fid-1]; 
 file->fid = fid;
 file->list = f;
 return file;
}


logical fio_list_any_eof( FioFileList f, long mode )
{

 logical eof;
 long fid;
 FioFile file;
 eof = UT_FALSE;
 for ( fid = 1; fid <= f->fio_nfiles; fid++ ) {
  file = &f->files[fid-1];
  if ( file->status == UT_FILE_EOF ) {
   eof = UT_TRUE;
   if ( mode == 0 ) fio_file_clear_eof( file );
  }
 }
 return( eof );
}


long fio_list_tty_in( FioFileList f )
{
 if ( !f ) return 0;
 return f->tty_in;
}

long fio_list_tty_out( FioFileList f )
{
 if ( !f ) return 0;
 return f->tty_out;
}


long fio_list_tty_err( FioFileList f )
{
 if ( !f ) return 0;
 return f->tty_err;
}



