/*                                                                
**  Copyright (C) 2002-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   gtoken.c                                                    */
/*                                                               */
/*   Functions for parsing                                       */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "utlib.h"

#include "esc_types.h"


/*----------------------------------------------------------------------*/
/*
 *  cs_put_token
 *
 * JCMLIB library routine (TOKEN)
 * Copy from buf to *dest, appending the sep character or until dest is full up to siz.
 * Update the dest pointer to point to the end of the text
 */

void cs_put_token( const char* ibuf, char** dest, const char sep, long* siz )
{
 char* buf = (char*)ibuf;
 char* ptr = *dest;
 while ( *siz && *buf != '\0' ) 
 {
  *ptr = *buf++; ptr++; (*siz)--;
 }
 if ( *siz && sep )
 {
  *ptr = sep; ptr++;
 }
 *ptr = '\0';
 *dest = ptr;
}

/*----------------------------------------------------------------------*/
/*
 *
 * cs_get_token
 *
 * Uses: escproto.h: esc_get_char
 * tokenproto.h: cs_copy_token
 * csproto.h: cs_copy
 * iolib.h: tmsg
 */

void cs_get_token( char** buf, char* token, const long mode, const long siz )
{
 char  seps[5]=" ";
 char pars[2];

 if ( !buf || !(*buf) ) { *token = '\0'; return; }
 switch( mode ) {
  case TMODE_WORD: 
   pars[0] = esc_special_char( LEFT_PAREN );
   pars[1] = esc_special_char( RIGHT_PAREN );
   seps[0]= ' ';
   seps[1]= '\t';
   seps[2]= esc_special_char( WORD_SEP );
   seps[3]= esc_special_char( STRING_SEP );  
   seps[4] = 0;
   cs_copy_token_esc( buf, seps, pars, 0, token, siz );
   break;
  case TMODE_STRING: 
   pars[0] = esc_special_char( LEFT_PAREN );
   pars[1] = esc_special_char( RIGHT_PAREN );
   seps[0] = esc_special_char( STRING_SEP );
   seps[1] = 0;
   cs_copy_token_esc( buf, seps, pars, 0, token, siz );
   break;
  case TMODE_LINE:
   cs_copy( *buf, token );
   *buf = NULL;
   break;
  default:
   fio_tmsg("Bad token mode"); 
   break;
 }
}

/*----------------------------------------------------------------------*/
/*
 * cs_put
 *
 */

void cs_put_cs( const char* token, char** buf, long* siz )
{
 char sep;
 sep = esc_special_char( STRING_SEP );
 cs_put_token( token, buf, sep, siz );
}

/*----------------------------------------------------------------------*/
/*
 * cs_putw
 *
 */

void cs_put_c( const char* token, char** buf, long* siz )
{
 cs_put_token( token, buf, ' ', siz );
}

void cs_put_i( const long token, char** buf, long* siz )
{
 TextWord word;
 cs_put_token( cs_write_i(token, word, WORD_SIZE), buf, ' ', siz );
}

void cs_put_s( const short token, char** buf, long* siz )
{
 TextWord word;
 cs_put_token( cs_write_s(token, word, WORD_SIZE), buf, ' ', siz );
}

void cs_put_r( const float token, char** buf, long* siz )
{
 TextWord word;
 cs_put_token( cs_write_r(token, word, WORD_SIZE), buf, ' ', siz );
}

void cs_put_d( const double token, char** buf, long* siz )
{
 TextWord word;
 cs_put_token( cs_write_d(token, word, WORD_SIZE), buf, ' ', siz );
}



/*----------------------------------------------------------------------*/
/*
 * cs_puts
 *
 */

void cs_put_cn( const char* token, char** buf, long* siz )
{
 cs_put_token( token, buf, '\0', siz );
}

/*----------------------------------------------------------------------*/
/*
 * cs_get
 *
 */

void cs_get_cs( char** buf, char* token, const long siz )
{
 cs_get_token( buf, token, TMODE_STRING, siz );
}

/*----------------------------------------------------------------------*/
/*
 * cs_get_c
 *
 */

void cs_get_c( char** buf, char* token, const long siz )
{
 cs_get_token( buf, token, TMODE_WORD, siz );
}

/*----------------------------------------------------------------------*/
/*
 * cs_getl
 *
 */

void cs_get_cl( char** buf, char* token, const long siz )
{
 cs_get_token( buf, token, TMODE_LINE, siz );
}


long cs_get_i( char** buf )
{
 TextWord token;
 cs_get_c( buf, token, WORD_SIZE );
 return( cs_read_i( token ) );
}
short cs_get_s( char** buf )
{
 TextWord token;
 cs_get_c( buf, token, WORD_SIZE );
 return( cs_read_s( token ) );
}
logical cs_get_q( char** buf )
{
 TextWord token;
 cs_get_c( buf, token, WORD_SIZE );
 return( cs_read_q( token ) );
}
float cs_get_r( char** buf )
{
 TextWord token;
 cs_get_c( buf, token, WORD_SIZE );
 return( cs_read_r( token ) );
}
double cs_get_d( char** buf )
{
 TextWord token;
 cs_get_c( buf, token, WORD_SIZE );
 return( cs_read_d( token ) );
}

long cs_getd_i( char** buf, const long xdefault )
{
 TextWord token;
 cs_get_c( buf, token, WORD_SIZE );
 if ( cs_is_blank( token ) ) {
  return( xdefault );
 } else {
  return( cs_read_i( token ) );
 }
}
short cs_getd_s( char** buf, const short xdefault )
{
 TextWord token;
 cs_get_c( buf, token, WORD_SIZE );
 if( cs_is_blank( token ) ) {
  return( xdefault );
 } else {
  return( cs_read_s( token ) );
 }
}
float cs_getd_r( char** buf,  const float xdefault )
{
 TextWord token;
 cs_get_c( buf, token, WORD_SIZE );
 if ( cs_is_blank( token ) ) {
  return( xdefault ); 
 } else {
  return( cs_read_r( token ) );
 }
}
double cs_getd_d( char** buf, const double xdefault )
{
 TextWord token;
 cs_get_c( buf, token, WORD_SIZE );
 if ( cs_is_blank( token )) {
  return( xdefault );
 } else {
  return( cs_read_d( token ) );
 }
}



long cs_list_parse( char* buf, char** list, const long nsiz )
{
 long n;

 n = 0;
 while ( buf != NULL && !cs_is_blank( buf ) ) {
  cs_get_c( &buf, list[n++], nsiz );
 }
 return( n );
}


long cs_count_char( const char* buf, const char c )
{
 char* ptr = (char*)buf;
 long n = 0;
 while ( *ptr )
 {
  if ( *ptr++ == c ) n++;
 }
 return n;
}


void cs_compose_paren_list( char** items, const long n, char* buf, const long maxlen )
{
 long i; 

 (void) maxlen;

 /* Compose the output char* */
 buf[0] = '\0';
 if ( items[0])
 {
  strcat( buf, items[0] );
  if ( n > 0 )
   strcat( buf, "(" );
 }
 for ( i = 1; i<= n; i++ )
 {
  strcat( buf, items[i] );
  if ( i < n ) strcat( buf, "," );
 }
 if ( items[0] && n > 0 )
  strcat( buf, ")" );
}

/* Need to improve this to handle embedded commas*/
/* entries[0] to entries[*np] where [0] is the prefix  E0(E1,E2,...En-1) */
char** cs_parse_paren_list( const char* buffer, long* np )
{
 return cs_parse_parsep_list( buffer, "()", ',', np, UT_TRUE );
}



/***************************************************************/
/*                                                             */
/* Function:   cs_parse_comma_list                             */
/*                                                             */
/* Description:                                                */
/*             parse comma list, entries [0] to entries[*np-1] */
/*                                                             */
/*             Calls cs_parse_sequential_list() with parameter */
/*             of double quote char (") as quote               */
/*                                                             */
/*  Input parameters:                                          */
/*                   const char  *buffer                       */
/*                   long        *np                           */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            char  **result                                   */
/*                                                             */
/***************************************************************/
char** cs_parse_comma_list( const char* buffer, long* np )
{
  return cs_parse_sequential_list( buffer, "\"\'", ',', np );
}


/***************************************************************/
/*                                                             */
/* Function:   cs_parse_parcom_list                            */
/*                                                             */
/* Description:                                                */
/*             This is obsolete (??)                           */
/*                                                             */
/*             Calls cs_parse_parsep_list() to do work         */
/*                                                             */
/*  Input parameters:                                          */
/*                   const char  *buffer                       */
/*                   long        *np                           */
/*                   logical      mode                         */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*            char  **result                                   */
/*                                                             */
/***************************************************************/
char** cs_parse_parcom_list( const char* buffer, long* np, logical mode )
{
  return cs_parse_parsep_list( buffer, "()", ',', np, mode );
}


/***************************************************************/
/*                                                             */
/* Function:   cs_parse_parsep_list                            */
/*                                                             */
/* Description:                                                */
/*               parse parenthesis and separator list          */
/*                                                             */
/*  Input parameters:                                          */
/*                   const char  *buffer                       */
/*                   char        *pars                         */
/*                   char         sep                          */
/*                   logical      mode                         */
/*                                                             */
/*  Output parameters:                                         */
/*                   long        *np                           */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            char  **result                                   */
/*                                                             */
/***************************************************************/
char** cs_parse_parsep_list( const char* buffer, char* pars,
                             char sep, long* np, logical mode )
{
  long   n = 0;
  long   nmax   = 0; 
  long   extra  = 0;
  long   len    = 0;
  char  *ptr    = NULL; 
  char **result = NULL;
  char  *next   = NULL;
  char  *tail   = NULL;

  *np = 0;

  if ( !buffer || cs_is_blank( buffer ) )
    return NULL;

  if ( mode )
    extra = 1;

  /* number of values = number of separators + 1 */
  nmax = cs_count_char( buffer, sep ) + 1;


  /* allocate pointers for values + extra field. */
  result = calloc( nmax+extra, UT_SZ_P );

  /* locate openning bracket if any */ 
  if ( pars && pars[0] )
    ptr = strchr( buffer, pars[0] );

  if ( ptr && mode )
  {
    /* openning bracked located, and we want to keep the preceding value */
    len = ptr- buffer;
    result[0] = cs_dup_siz( buffer, len );
    ptr++;
  }
  else
  {
    /* either none found or don't care */
    /* set pointer to start of values.. either right after bracket char */
    /* or at the beginning of the buffer                                */
    if ( ptr == buffer )
    {
      ptr++;
    }
    else
    {
      if ( !ptr )
      {
        ptr = (char*)buffer;
      }
    }
  }

  /* if ending bracket specified, locate end of bracketed segment */
  if ( pars && pars[1] )
    tail = strchr( ptr, pars[1] );

  /* parse out the values */
  while( ptr )
  {
    if ( n >= nmax ) 
    {
      /* Found more than expected?? add more space */
      nmax = n+extra;
      result = realloc( result, nmax+1 * UT_SZ_P);
    }

    /* find next separator */
    next = strchr( ptr, sep );


    if ( next && ( !tail || ( tail && ( tail - next > 0 ))))
    {
      /* have a value.. add it to the results */
      result[n+extra] = cs_dup_siz( ptr, next-ptr );

      n++;
      ptr = ++next;
    }
    else
    {
      if ( tail )
      {
        if ( tail - ptr > 0 )
        {
	  /* the last value before bracket segment end */
          result[n+extra] = cs_dup_siz( ptr, tail-ptr );
          n++;
        }
        ptr = NULL;
      }
      else
      {
	/* the last value */
        result[n+extra] = cs_dup( ptr );
        n++;
        ptr = NULL;
      }
    }
  } // end while
   
  *np = n;

  return result;
}

/***************************************************************/
/*                                                             */
/* Function:   cs_parse_sequential_list                        */
/*                                                             */
/* Description:                                                */
/*   Parse list according to separator specified               */
/*   Ignores separators enclosed within quote segment,         */
/*   (char to use provided at quote arg), and escaped          */
/*   separators (preceded by backslash '\')                    */
/*                                                             */
/* Input parameters:                                           */
/*   const char  *buffer   = string to parse                   */
/*   char        quote     = quote char (' or " or <other>)    */
/*   char         sep      = char to split string on           */
/*                                                             */
/* Output parameters:                                          */
/*   long         *np      = number of parsed segments         */
/*                                                             */
/*  Returns:                                                   */
/*   char  **result        = list of parsed segments (copies)  */
/*                                                             */
/***************************************************************/
char** cs_parse_sequential_list( const char* buffer, char* quote,
                             char sep, long *np )
{
  long   n      = 0;
  long   nmax   = 0; 
  char  *ptr    = NULL;
  char  *last   = NULL; 
  char **result = NULL;
  char   qchar  = '\0';
  int    inquote = 0;
  int    escaped = 0;
  int    done    = 0;

  *np = 0;

  if ( !buffer || cs_is_blank( buffer ) )
    return NULL;

  /* Setup output list                               */
  /* max number of values = number of separators + 1 */
  nmax   = cs_count_char( buffer, sep ) + 1;
  result = calloc( nmax+1, UT_SZ_P );

  /* Set pointer to beginning of buffer */
  ptr  = (char *)buffer;
  last = ptr;

  /* parse out the values */
  while( !done )
  {
    if ( n >= nmax ) 
    {
      /* Found more than expected?? add more space */
      nmax = n;
      result = realloc( result, nmax+1 * UT_SZ_P);
    }

    if ( (*ptr != '\0') && strchr(quote, *ptr) && !escaped )
    {
      /* Found non-escaped quote.. */
      if ( !inquote )
      {
	/* found opening quote */
	inquote = 1;
	qchar = *ptr;
      }
      else if ( *ptr == qchar )
      {
	/* found matching quote */
	inquote = 0;
	qchar = '\0';
      }
    }
    else if ( ( *ptr == '\0' ) || (!inquote && (*ptr == sep) && !escaped ) )
    {
      /* end of string  or             */
      /* found a separator to parse on */
      result[n] = cs_dup_siz( last, (ptr -last) );
      ++n;

      if ( *ptr == '\0' )
	done = 1;

      /* set 'last' pointer to next token (skip sep) */
      last = ptr;
      last++;
    }

    /* Check if char is backslash (for escaping) */
    (*ptr == '\\')?(escaped = !escaped):(escaped = 0);

    if ( !done )
      ++ptr;

  } /* end while */
   
  *np = n;

  return result;
}

