/*                                                                
**  Copyright (C) 2002-2003,2005,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

#include "utlib.h"
#include "jcmstate.h"

/* IO stack for C JCMLIB */


FioStack fio_iostack_init( void )
{
 FioStack iofs;
 iofs = calloc( 1, sizeof( FioStackData ));
 iofs->input_stack = calloc( UT_IO_DEPTH, sizeof( FioFile ));
 iofs->output_stack = calloc( UT_IO_DEPTH, sizeof( FioFile ));
 iofs->depth = UT_IO_DEPTH;
 iofs->input = iofs->input_stack;
 iofs->output= iofs->output_stack;
 iofs->err_chan = NULL;
 iofs->ilevel = 1;
 iofs->olevel = 1;
 iofs->msgfunc = fio_std_msg;
 iofs->detab_flag = UT_TRUE;
 return iofs;
}

void fio_iostack_free( FioStack iofs )
{
 free( iofs->input_stack );
 free( iofs->output_stack );
 free( iofs );
}

/* Thread-safe versions */

FioFile fio_iostack_file( FioStack iofs, long mode )
{
 if ( !iofs) iofs = fio_get_stack();
 switch( mode )
 {
  case FIO_ERR:
   return iofs->err_chan;
  case FIO_INPUT:
   return *iofs->input;
  case FIO_OUTPUT:
   return *iofs->output;
  default:
   return NULL;
 }
}

void fio_iostack_realloc( FioStack iofs, long n )
{
  iofs->depth = n;
  iofs->input_stack  = realloc( iofs->input_stack, iofs->depth * sizeof( FioFile ));
  iofs->output_stack = realloc( iofs->input_stack, iofs->depth * sizeof( FioFile ));
  iofs->input = &iofs->input_stack[iofs->ilevel-1];
  iofs->output= &iofs->output_stack[iofs->olevel-1];
}

void fio_iostack_ipush( FioStack iofs, Tpio tpio, FioFileList list, FioFile file )
{

 if ( !iofs ) iofs = fio_get_stack();
 if ( iofs->ilevel >= iofs->depth ) 
  fio_iostack_realloc( iofs, 2 * iofs->depth );

 (void) tpio;

 iofs->ilevel++;
 iofs->input++;
 if ( file ) {

  fio_iostack_set( iofs, list, file, FIO_INPUT );
 }
}

void fio_iostack_opush( FioStack iofs, FioFileList list, FioFile file )
{
 if ( !iofs ) iofs = fio_get_stack();
 if ( iofs->olevel >= iofs->depth )
  fio_iostack_realloc( iofs, 2 * iofs->depth );

 iofs->olevel++;
 iofs->output++;

 if ( file ) 
  fio_iostack_set( iofs, list, file, FIO_OUTPUT );

}


void fio_iostack_ipop( FioStack iofs, Tpio tpio, FioFileList list, long mode )
{ 
 (void) tpio;

 if ( !iofs ) iofs = fio_get_stack();
 if ( mode >= 0 ) {
  FioFile file = fio_iostack_file( iofs, FIO_INPUT );
  if ( !fio_file_interactive( file ) ) fio_file_close( file );

 }
 iofs->ilevel--;
 iofs->input--;
 if ( iofs->ilevel <= 0 ) {
  fio_iostack_iclear( list, iofs );
 }
}

void fio_iostack_opop( FioStack iofs, FioFileList list, long mode )
{
 FioFile file;

 if ( !iofs ) iofs = fio_get_stack();

 if ( mode >= 0 ) {
  file = fio_iostack_file( iofs, FIO_OUTPUT );
  if ( !fio_file_interactive( file )) fio_file_close( file );
 }
 iofs->olevel--;
 iofs->output--;
 if ( iofs->olevel <= 0 ) {
  iofs->olevel = 1; 
  fio_iostack_set( iofs, list, NULL, FIO_OUTPUT );
  ut_msg_d( UT_VERBOSE, "Output to *" );
 }
}


long fio_iostack_get_level( FioStack iofs, long mode )
{
 if ( !iofs ) iofs = fio_get_stack();
 if ( mode == FIO_INPUT )
  return iofs->ilevel;
 else
  return iofs->olevel;
}

void fio_iostack_iclear( FioFileList list, FioStack iofs )
{
 if ( !iofs ) iofs = fio_get_stack();
 if ( !list ) list = fio_get_list();
 iofs->ilevel = 1;
 iofs->input = iofs->input_stack;
 fio_iostack_set( iofs, list, NULL, FIO_INPUT );
 ut_msg_d( UT_VERBOSE, "Input to *" ); 
}
/*----------------------------------------------------------------------*/
/*
 * io_init
 *
 * Reinitialize IO stacks
 *
 */

void fio_iostack_reinit( FioFileList list, FioStack iofs )
{
 if ( !iofs ) iofs = fio_get_stack();
 iofs->input = iofs->input_stack;
 iofs->output = iofs->output_stack;
 fio_iostack_set( iofs, list, NULL, FIO_INPUT );
 fio_iostack_set( iofs, list, NULL, FIO_OUTPUT );
 fio_iostack_set( iofs, list, NULL, FIO_ERR );
}


void fio_stack_reinit( void )
{ 
 fio_iostack_reinit( NULL, NULL );
}

void fio_iostack_set( FioStack iofs, FioFileList list, FioFile file, long mode )
{
 long fid;
 if ( !iofs ) iofs = fio_get_stack();
 if ( !list && !file ) list = fio_get_list();
 switch( mode )
 {
  case FIO_ERR:
    if ( !file )
    { 
     fid = fio_list_tty_err(list);
     file = fio_list_get_file( list, fid );
    }
    iofs->err_chan = file;
    break;
  case FIO_OUTPUT:
    if ( !file )
    { 
     fid = fio_list_tty_out(list);
     file = fio_list_get_file( list, fid );
    }
    *iofs->output = file;
    break;
  case FIO_INPUT:
    if ( !file )
    { 
     fid = fio_list_tty_in(list);
     file = fio_list_get_file( list, fid );
    }
    *iofs->input = file;
    break;
  default:
    break;
 }
}

void fio_basic_print_state( void )
{
 FioStack iofs = fio_get_stack();
 TextCard buf;
 FioFile file;
 long fid;
 fio_msg( " " );
 fio_dashline( 40 );
 fio_msg( " JCMLIB Basic IO (BIO) Status " );
 fio_dashline( 40 );
 file = *iofs->input;
 fid = file? file->fid : 0;
 sprintf( buf, "  Input Channel:  %ld", fid ); fio_msg( buf );
 file = *iofs->output;
 fid = file? file->fid : 0;
 sprintf( buf, "  Output Channel: %ld", fid ); fio_msg( buf );
 file = iofs->err_chan;
 fid = file? file->fid : 0;
 sprintf( buf, "  Error Channel:  %ld", fid ); fio_msg( buf );
}

/*----------------------------------------------------------------------*/
/*
 * input_test_eof
 * Uses fileproto.h fio_test_eof
 */

logical fio_input_test_eof( void )
{
 FioFile file = fio_iostack_file( NULL, FIO_INPUT );
 return ( fio_file_test_eof( file ));
}

/*----------------------------------------------------------------------*/
/*
 * msg
 *
 */

void fio_std_msg( char* buf, logical newline )
{
 FioFile file = fio_iostack_file( NULL, FIO_OUTPUT );
 if ( newline ) {
  fio_write_line( file, buf );
 } else {
  fio_write_text( file, buf );
 }
}

void fio_msg_endline( void )
{
 fio_msg( NULL );
}



/*----------------------------------------------------------------------*/
/*
 * tty_input
 *
 */
logical fio_tty_input( char* buf, long siz )
{
 return( fio_read_line( NULL, buf, siz ) );
} 

void fio_dashline( long n )
{
 fio_iostack_dashline( NULL, n );
}

void fio_iostack_dashline( FioStack iofs, long n )
{
 TextCard chunk;
 long m;
 long k;
 char* ptr;
 m = n;
 while ( m > CARD_SIZE ) {
  k = CARD_SIZE; 
  ptr = chunk; 
  while( k-- )
  { *ptr = '-'; ptr++; }
  *ptr = '\0';
  fio_msgl( chunk );
  fio_iostack_msg( iofs, NULL, chunk, UT_FALSE, UT_FALSE );
  m -= CARD_SIZE;
 } 
 k = m; 
 ptr = chunk; 
 while( k-- ) 
  { *ptr = '-'; ptr++; }
 *ptr = '\0';
 fio_iostack_msg( iofs, NULL, chunk, UT_TRUE, UT_FALSE );
}

void fio_iostack_get_entry( FioStack iofs, char* io_type, long level, char* buf, long* fid, char* status )
{

 logical inmode;
 long stack_top;
 long istat;
 FioFile file;

 if ( !iofs ) iofs = fio_get_stack(); 
 inmode = *io_type == 'I';
 cs_copy( " ", status ); 
 stack_top = inmode ? iofs->ilevel : iofs->olevel;
 if ( level <= 0 ) {
  cs_copy( ">> Invalid IO stack level <<", buf );
 } else if ( level > stack_top ) {
  cs_copy( ">> Beyond top of stack    <<", buf );
 } else {
  file = inmode?  iofs->input_stack[level-1] : iofs->output_stack[level-1];
  *fid = file ? file->fid : 0;   
  if ( file && *fid == fio_tty_in() ) {
   cs_copy( "(Interactive stream)", buf );
   cs_copy( "[TTY IN]", status );
  } else if ( file && *fid == fio_tty_out() ) {
   cs_copy( "(Interactive stream)", buf );
   cs_copy( "[TTY OUT]", status );
  } else if ( *fid == 0 ) { 
   cs_copy( "(No file)", buf );
  } else {
   fio_file_get_name( file, buf, FN_SIZE );
   istat = fio_file_status( file, FIO_CHECK );
   if ( istat == UT_FILE_EOF ) {
    cs_copy( "[EOF]", status );
   }  else if ( istat == UT_FILE_OPEN ) { 
    cs_copy( "[Open]", status );
   } else if ( istat == UT_FILE_CLOSED ) {
    cs_copy( "(No file)", status );
   } else {
    cs_copy( "(Bad status)", status ); 
   }

  }
 }
}


void fio_iostack_print_top( FioStack iofs, char* type, char* pfx )
{
 long stack_top;
 Filename filename;
 TextCard status;
 TextCard mbuf;
 TextBuf buf;
 long fid;

 if ( !iofs ) iofs = fio_get_stack();

 stack_top = fio_iostack_get_level( iofs, FIO_INPUT );
 fio_iostack_get_entry( iofs, type, stack_top, filename, &fid, status );
 if ( fid > 0 ) {
  sprintf( mbuf, " Unit %ld", (long)fid );
 } else {
  cs_copy( " ", mbuf );
 }
 sprintf( buf, "%s IO stream: %s%s%s", pfx, filename, status, mbuf );
 fio_msg( buf );
}


void fio_stack_print_state( void )
{
 FioStack iofs = fio_get_stack();
 fio_iostack_print_state( iofs );
}

void fio_iostack_print_state( FioStack iofs )
{

 long top;
 TextWord tmp;
 long iofs_in_ios;
 TextBuf buf;
 if ( !iofs ) iofs = fio_get_stack();
 top = fio_iostack_get_level( iofs, FIO_INPUT );
 if ( fio_stack_status(FIO_CHECK) == UT_FILE_EOF )  {
  iofs_in_ios = -1;
 } else {
  iofs_in_ios = 0;
 }
 if ( fio_stack_eof(FIO_CHECK) ) {
  cs_copy( "set", tmp );
 } else {
  cs_copy( "clear", tmp );
 }
 fio_msg( " " );
 fio_dashline( 40 );
 fio_msg( " JCMLIB IO File Stack (IOFS) Status" );
 fio_dashline( 40 );
 sprintf( buf, "  IOFS Input Stack depth: %-6ld IOS Code: %-6ld EOF Flag: %s", (long)top, (long)iofs_in_ios, tmp );
 fio_msg( buf );
 if ( top > 0 ) {
  fio_iostack_print_stack( iofs, "I", top );
 }
 top = iofs->olevel;
 sprintf( buf, "  IOFS Output Stack depth: %-6ld", (long)top );
 fio_msg( buf );
 if ( top > 0 ) {
  fio_iostack_print_stack( iofs, "O", (long)top );
 }
 fio_msg( " " );
}


void fio_iostack_print_stack( FioStack iofs, char* type, long top )
{
 long level;
 long fid;
 Filename filename;
 TextCard status;
 TextBuf buf;
 fio_msg( "    ------------------------------" );
 fio_msg( "    | Level    FID   Status      | File ");
 for ( level = top; level > 0; level-- ) {
  fio_iostack_get_entry( iofs, type, level, filename, &fid, status );
  sprintf( buf, "    | %-6ld   %-6ld%-12s| %s", (long)level, (long)fid, status, filename );
  fio_msg( buf );
 }
 fio_msg( "    ------------------------------" );
 
}

void fio_iostack_set_detab( FioStack f, logical q )
{
 f->detab_flag = q;
}

/*
 *  Internal routine
 */
void fio_iostack_set_input( FioStack iofs, Tpio tpio, FioFileList list )
{
 long level;
 long fid;
 Filename tbuf;
 FioFile file;
 TextCard ebuf;
 long lineno;
 
 if ( !iofs ) iofs = fio_get_stack();

 level = fio_iostack_get_level( iofs, FIO_INPUT );
 file = fio_iostack_input_entry( iofs, 0 );
 while ( level > 0 && fio_file_test_eof( file ) ) {
  ut_msg_d( UT_VERBOSE_FIO, "IOFS Input Pop" );
  fio_iostack_ipop( iofs, tpio, list, FIO_POP_CLOSE );
  level = fio_iostack_get_level( iofs, FIO_INPUT );
  file = fio_iostack_input_entry( iofs, 0 );
 }
 if ( ut_dbg_level(UT_VERBOSE_FIO) ) {
  fio_file_get_name( file, tbuf, FN_SIZE );
  lineno = fio_file_get_lineno( file );
  fid = file? file->fid : 0;
  sprintf( ebuf, "IOFS INPUT Level %-6ldFID %-6ldLine %-6ldFile %s", (long)level, (long)fid, (long)lineno, tbuf );
  ut_msg_d( UT_VERBOSE_FIO, ebuf );
 }
}


FioFile fio_iostack_input_entry( FioStack iofs, long level )
{
 long stklevel;
 if ( !iofs ) iofs = fio_get_stack();

 if ( level == 0 ) {
  stklevel = iofs->ilevel;
 } else {
  stklevel = level;
 }
 return( iofs->input_stack[stklevel-1] );
}


long fio_stack_status( long mode )
{
 FioFile file = fio_iostack_file( NULL, FIO_INPUT );
 return( fio_file_status( file, mode ) );
}

logical fio_stack_ok( long mode )
{
 FioFile file = fio_iostack_file( NULL, FIO_INPUT );
 return( fio_file_status( file, mode ) == UT_FILE_OPEN );
}

logical fio_stack_eof( long mode )
{
 FioFile file = fio_iostack_file( NULL, FIO_INPUT );
 return( fio_file_mode_eof( file, mode ) );
}


void fio_iostack_rewind_input( FioStack iofs )
{
 FioFile file;
 if ( !iofs ) iofs = fio_get_stack();
 file = fio_iostack_file( iofs, FIO_INPUT );
 fio_file_rewind( file );
}
