/*                                                                
**  Copyright (C) 2002-2003,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   polar.c                                                     */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/
#include "utlib.h"

static double polar_ang_d( const double*, const double );



/***************************************************************/
/*                                                             */
/* Function: polar_ang_d                                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
static double polar_ang_d( const double* p, const double r )
{
  double x;
  double re;

  /* Make sure you get an accurate off axis angle. If the off axis
     angle is small, safer to take the size of the normal component.
  */
  x = acos( p[2] / r );
  if ( fabs( x ) < 0.001 ) 
  {
    re = sqrt( p[0] * p[0] + p[1] * p[1] );
    x  = asin( re / r );
  }

  return( x/UT_DEG_RAD );
}

/***************************************************************/
/*                                                             */
/* Function: utm_cos                                           */
/*                                                             */
/* Description:                                                */
/*   Returns cosine of x                                       */
/*   This function switches to using sin( pi/2 - x ) when      */
/*   the angle approaches n*pi/2 where n=+/- 1,3,5,...         */
/*   where errors become large and results unpredictable       */
/*                                                             */
/***************************************************************/
double utm_cos( const double x )
{

#define ANGLE_TOL 1.0e-10  /* radians */

  double y;
  double tol;
  double retval;

  tol = cos(M_PI/2 - ANGLE_TOL);

  retval = cos( x );
  if ( fabs(retval) < tol )
  {
    /* angle is getting into the high error range.. use sin() */
    y = M_PI/2.0 - x;
    retval = sin(y);
  }

  return( retval );
}


/***************************************************************/
/*                                                             */
/* Function: utm_sin                                           */
/*                                                             */
/* Description:                                                */
/*   Returns sine of x                                         */
/*                                                             */
/***************************************************************/
double utm_sin( const double x )
{
  double retval;

  retval = sin(x);
  return( retval );
}


/***************************************************************/
/*                                                             */
/* Function: utm_dacosd                                        */
/*                                                             */
/* Description:                                                */
/*   Returns arccos (in degrees) of x                          */
/*                                                             */
/***************************************************************/
double utm_dacosd( const double x )
{
  double retval = acos( x )/UT_DEG_RAD;
  return( retval );
}


/***************************************************************/
/*                                                             */
/* Function: utm_dasind                                        */
/*                                                             */
/* Description:                                                */
/*   Returns arcsine (in degrees) of x                         */
/*                                                             */
/***************************************************************/
double utm_dasind( const double x )
{
  double retval = asin( x )/UT_DEG_RAD;
  return( retval );
}

/***************************************************************/
/*                                                             */
/* Function: utm_racosd                                        */
/*   Returns arccosine (in degrees) of x as a float            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
float utm_racosd( const float x )
{ 
  float utr;
  utr = (float)acos( (double)x );
  return( utr/UT_DEG_RAD );
}

/***************************************************************/
/*                                                             */
/* Function: utm_rasind                                        */
/*   Returns arcsine (in degrees) of x as a float              */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
float utm_rasind( const float x )
{
  float utr;
  utr = (float) asin( (double)x );
  return (utr/UT_DEG_RAD);
}


/***************************************************************/
/*                                                             */
/* Function: utm_dcosd                                         */
/*   Returns cosine of x. (x is in degrees)                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
double utm_dcosd( const double x )
{
  return( cos( UT_DEG_RAD * x  ));
}

/***************************************************************/
/*                                                             */
/* Function: utm_dsind                                         */
/*   Returns sine of x. (x is in degrees)                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
double utm_dsind( const double x )
{
  return( sin( UT_DEG_RAD * x  ));
}

/***************************************************************/
/*                                                             */
/* Function: utm_rcosd                                         */
/*   Returns cosine of x as a float. (x is in degrees)         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
float utm_rcosd( const float x )
{
  return( (float)cos( UT_DEG_RAD * (double)x  ));
}

/***************************************************************/
/*                                                             */
/* Function: utm_rsind                                         */
/*   Returns sine of x as a float. (x is in degrees)           */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
float utm_rsind( const float x )
{
  return( (float)sin( UT_DEG_RAD * (double)x ) ) ;
}

/***************************************************************/
/*                                                             */
/* Function: ar_vec_d                                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
void ar_vec_d( const double r, const double theta, const double phi, double* p )
{
  if ( theta == 0.0 || theta == 180.0 ) 
  {
    p[0] = 0.0;
    p[1] = 0.0;
  } 
  else 
  {
    p[0] = r * dcosd( phi ) * dsind( theta );
    p[1] = r * dsind( phi ) * dsind( theta );
  }
  p[2] = r * dcosd( theta );
}

/***************************************************************/
/*                                                             */
/* Function: ar_vec_r                                          */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
void ar_vec_r( const float r, const float theta, const float phi, float* p )
{
  if ( theta == 0.0 || theta == 180.0 ) 
  {
    p[0] = 0.0;
    p[1] = 0.0;
  } 
  else 
  {
    p[0] = r * rcosd( phi ) * rsind( theta );
    p[1] = r * rsind( phi ) * rsind( theta );
  }
  p[2] = r * rcosd( theta );
}

/***************************************************************/
/*                                                             */
/* Function: ar_polar_d                                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
void ar_polar_d( const double* p, double* r, double* theta, double* phi )
{
  *r = ar_vnorm_d( p );

  if ( *r > 0 ) 
  {
    *phi = argd_d( p[0], p[1] );
    *theta = polar_ang_d( p, *r );
  } 
  else 
  { 
    *phi =0.0;
    *theta = 0.0;
  }
}

/***************************************************************/
/*                                                             */
/* Function: ar_polar_r                                        */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
void ar_polar_r( const float* p, float* r, float* theta, float* phi )
{
  *r = ar_vnorm_r( p );

  if ( *r > 0.0 ) 
  {
    *phi = argd_r( p[0], p[1] );
    *theta = racosd( p[2]/(*r) );
  } 
  else 
  { 
    *phi =0.0;
    *theta = 0.0;
  }
}


/***************************************************************/
/*                                                             */
/* Function: ar_polarll_d                                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
double ar_polarll_d( const double* p, double* llong, double* lat )
{
  double r;
  r = ar_vnorm_d( p );
  
  if ( r > 0.0 ) 
  {
    *llong = argd_d( p[0], p[1] );
    *lat  = ut_constants.d.polar_lat - polar_ang_d( p, r );
  } 
  else 
  {
    *llong = 0.0;
    *lat = ut_constants.d.polar_lat;
  }
  return( r );
}

/***************************************************************/
/*                                                             */
/* Function: ar_polarll_r                                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
float ar_polarll_r( const float* p, float* llong, float* lat )
{
  float r;
  r = ar_vnorm_r( p );

  if ( r > 0.0 ) 
  {
    *llong = argd_r( p[0], p[1] );
    *lat  = ut_constants.r.polar_lat - racosd( p[2]/r );
  } 
  else 
  {
    *llong = 0.0;
    *lat = ut_constants.r.polar_lat;
  }
  return( r );
}


/***************************************************************/
/*                                                             */
/* Function: ar_unitll_d                                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
void ar_unitll_d( const double llong, const double lat, double* p )
{
  if ( fabs( lat ) == ut_constants.d.polar_lat ) 
  {
    p[0] = 0.0;
    p[1] = 0.0;
  } 
  else 
  {
    p[0] = dcosd( llong ) * dcosd( lat );
    p[1] = dsind( llong ) * dcosd( lat );
  }
  p[2] = dsind( lat );
}

/***************************************************************/
/*                                                             */
/* Function: ar_unitll_r                                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
void ar_unitll_r( const float llong, const float lat, float* p )
{
  if ( rabs( lat ) == ut_constants.r.polar_lat ) 
  {
    p[0] = 0.0;
    p[1] = 0.0;
  } 
  else 
  {
    p[0] = rcosd( llong ) * rcosd( lat );
    p[1] = rsind( llong ) * rcosd( lat );
  }
  p[2] = rsind( lat );
}


/***************************************************************/
/*                                                             */
/* Function: ar_polar6_r                                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
void ar_polar6_r( const float* p, float* r, float* theta, float* phi, float* rrdot, float* tdot, float* phidot )
{
  double dr, dtheta, dphi, drdot, dtdot, dphidot;
  double dp[6];
  long ii;

  for ( ii = 0; ii < 6; ii++ )
    dp[ii] = p[ii];

  ar_polar6_d( dp, &dr, &dtheta, &dphi, &drdot, &dtdot, &dphidot );
  *r = (float)dr;
  *theta = (float)dtheta;
  *phi = (float)dphi;
  *rrdot = (float) drdot;
  *tdot = (float)  dtdot;
  *phidot= (float) dphidot;
}

/***************************************************************/
/*                                                             */
/* Function: ar_polar6_d                                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
void ar_polar6_d( const double* p, double* r, double* theta, double* phi, double* rrdot, double* tdot, double* phidot )
{
  double e[3];
  double de[3];
  double enorm;
  double ec[3];
  double r0;

  ar_polar_d( p, &r0, theta, phi );
  *r = r0;

  if ( r0 > 0.0 ) 
  {
    ar_vcopy_d( p, e );
    ar_vcopy_d( p+3, de );
    *rrdot = ar_vdot_d( e, de ) / r0;
    e[2] = 0.0;
    de[2] = 0.0;
    enorm = ar_vnorm_d( e );
    ar_vcross_d( e, de, ec );
    if ( enorm > 0.0 ) 
    {
      *tdot = ( ( p[2] * ar_vdot_d( e, de ) - p[5] * ( enorm * enorm ) )
		/ ( enorm * r0 * r0 ) ) / UT_DEG_RAD;
      *phidot = (  ec[2] / ( enorm * enorm ) ) / UT_DEG_RAD;
    } 
    else 
    {
      *tdot = ( ar_vnorm_d( de ) / r0 ) / UT_DEG_RAD;
      if ( *theta > 90.0 ) *tdot *= -1.0;
      *phidot = 0;
    }
  } 
  else 
  {
    *rrdot = 0;
    *tdot = 0;
    *phidot = 0;
  }
}

/***************************************************************/
/*                                                             */
/* Function: ar_vec6_r                                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
void ar_vec6_r( const float r, const float theta, const float phi, const float rrdot, const float tdot, const float phidot, float* p )
{
  double dp[6];
  long ii;

  ar_vec6_d( (double)r, (double)theta, (double)phi, (double)rrdot, 
	                        (double)tdot, (double)phidot, dp );
  for ( ii = 0; ii < 6; ii++ )
    p[ii] = dp[ii];
}


/***************************************************************/
/*                                                             */
/* Function: ar_vec6_d                                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
void ar_vec6_d( const double r, const double theta, const double phi, const double rrdot, const double tdot, const double phidot, double* p )
{
  double tdotr, phidotr;
  double ca, sa, cb, sb;
  double w;

  ar_vec_d( r, theta, phi, p );
  tdotr = UT_DEG_RAD * tdot;
  phidotr = UT_DEG_RAD * phidot;

  ca = dcosd( phi );
  sa = dsind( phi );
  cb = dcosd( theta );
  sb = dsind( theta );
  w = r * tdotr * cb + rrdot * sb;

  p[3] = -p[1] * phidotr      + w * ca;
  p[4] =  p[0] * phidotr      + w * sa;
  p[5] = -r    * tdotr   * sb + rrdot * cb;
}

/***************************************************************/
/*                                                             */
/* Function: utm_spharc_r                                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
float utm_spharc_r( const float a1, const float b1, const float a2, const float b2 )
{
  return( (float)utm_spharc_d( (double)a1, (double)b1, (double)a2, (double)b2 ) );
}

/***************************************************************/
/*                                                             */
/* Function: utm_spharc_d                                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/***************************************************************/
double utm_spharc_d( const double a1, const double b1, const double a2, const double b2 )
{
  double v1[3];
  double v2[3];
  double w;
  double dw;
  double x, y;
  long  ii;

  ar_unitll_d( a1, b1, v1 );
  ar_unitll_d( a2, b2, v2 );
  w = 0.0;

  for ( ii = 0; ii < 3; ii++ ) 
  {
    dw = ( v1[ii] - v2[ii] ) / 2.0;
    w += dw * dw;
  }

  x = w < 1.0 ? sqrt( 1.0 - w ) : 0;
  y = sqrt( w );

  return( 2.0 * utm_argd_d( x, y ) );
}
