/*                                                                
**  Copyright (C) 2002-2004,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*****************************************************************/
/*                                                               */
/*   token.c                                                     */
/*                                                               */
/*   Token parsing routines                                      */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "utlib.h"


/***************************************************************/
/*                                                             */
/* Function:   check_pars                                      */
/*                                                             */
/* Description:                                                */
/*             local routine to check parenthesis level in     */
/*             char *                                          */
/*                                                             */
/*  Input parameters:                                          */
/*             const char     source                           */
/*             long          *paren_level                      */
/*             const logical  paren_use                        */
/*             const char     lpar                             */
/*             const char     rpar                             */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            static logical  status                           */
/*                                                             */
/***************************************************************/
static logical check_pars( const char source, long* paren_level,
                           const logical paren_use, const char lpar,
                           const char rpar )
{
  logical status;


  if ( !paren_use )
    return( UT_TRUE );

  status = UT_FALSE;

  if ( source == lpar ) 
    (*paren_level)++;
  else
    if ( source == rpar )
      (*paren_level)--;
    else
      status = (*paren_level <= 0 );

 return( status );
}


/***************************************************************/
/*                                                             */
/* Function:   cs_copy_token_esc                               */
/*                                                             */
/* Description:                                                */
/*             Copy from *buf to dest until you get to one of  */
/*             the seps characters lying outside of all        */
/*             parenthesis; or until dest is full up to siz.   */
/* 1996 Aug 26 JCM                                             */
/* 1997 Aug  5 JCM modified to return the separator found.     */
/*                                                             */
/*  Input parameters:                                          */
/*	       buf     Pointer to input buffer                 */ 
/*             seps    Token separator characters              */
/*             pars    Left and right paren chars              */
/*             siz     Max allowed size of output buffer       */
/*                                                             */
/*  Output parameters:                                         */
/*             dest    Output buffer                           */
/*                                                             */
/*  Returns:                                                   */
/*             char  found                                     */
/*                                                             */
/***************************************************************/
char cs_copy_token_esc( char** buf, const char* seps, const char* ppars,
                        const char esc, char* dest, const long gsiz )
{
  char    *pars        = (char*)ppars;
  long     siz         = gsiz;
  long     paren_level = 0;
  char     lpar        = '\0';
  char     rpar        = '\0';
  char    *source      = NULL;
  char     found       = 0;
  char     c           = '\0';
  logical  paren_use   = UT_FALSE;
  logical  loop        = UT_TRUE;
  logical  is_esc      = UT_FALSE;

  if ( !buf || !(*buf) )
  {
    cs_copy( " ", dest );
    return( 0 );
  }

  /* Advance to next non blank character.  */
  cs_advptr( buf );
  source = *buf;

  if ( *source != '\0' )
  {
    paren_level = 0;
    paren_use   = pars && !cs_eq_siz( pars, "  ", 2 );

    if ( pars )
    {
      lpar = pars[0];
      rpar = pars[1];
    }

    /* Copy source characters to destination */

    while ( *source != '\0' && loop )
    {
      c = *source;

      if ( check_pars ( c, &paren_level, paren_use, lpar, rpar ) )
      {
        loop = is_esc || !cs_ss_char( seps, c );

        if ( !loop )
          found = c;
      }

      is_esc = ( esc && esc == c );

      if ( !is_esc && loop && siz )
      {
        *dest++ = c;
        --siz;
      }

      ++source;

    }  /* end while */

    /* Advance source pointer to end of token */

    *buf = source;

    /* Advance to next non blank character.  */

    cs_advptr( buf );
  }  /* end if non null */

  if ( cs_is_blank( *buf ) )
  {
    *buf = NULL;
  }

  *dest = '\0'; 
  return( found );
}


/***************************************************************/
/*                                                             */
/* Function:   cs_parse_paren_token                            */
/*                                                             */
/* Description:                                                */
/*      Return the parsed token in *result and a pointer to the*/
/*      remainder of the input buffer as the function value.   */
/*      If the maxlen character is 0, or result is NULL on     */
/*      input, the result is allocated internally and must be  */
/*      freed by the user. Otherwise it is assumed that result */
/*      is preallocated and of size at least maxlen+1.         */
/*      The token is copied to a working buffer.               */
/*      For a fixed output, the working buffer is then copied  */
/*      to the output.  For an allocated output, the working   */
/*      buffer is copied to an allocated char*, or directly    */
/*      becomes the output if it is already allocated.         */
/*                                                             */
/*       The backslash character is used to escape separators. */
/*                                                             */
/*  Input parameters:                                          */
/*	       char  **pptr                                    */ 
/*             char    lsep                                    */
/*             char    rsep                                    */
/*             char   *result                                  */
/*             long    maxlen                                  */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*             char  *obuf                                     */
/*                                                             */
/***************************************************************/
char*  cs_parse_paren_token( char** pptr, char lsep, char rsep,
                             char* result, long maxlen )
{
  TextBuf   gbuf;           /* The default working buffer, for */
                            /* small char*s - saves a malloc   */

  char     *buf  = gbuf;            /* The working buffer */
  char     *iptr = *pptr;           /* The input buffer pointer */
  char     *optr = NULL;            /* The output buffer pointer */
  char     *obuf = result;          /* The output buffer */
  char      c    = '\0';
  char      esc  = '\\';
  char      lpar = '(';
  char      rpar = ')';

  logical   out_alloc  = UT_FALSE;  /* The output is a fixed buffer */
  logical   work_alloc = UT_FALSE;  /* The working buffer is an allocated buffer */
  logical   last_esc   = UT_FALSE;

  long      depth      = 1;         /* Separator depth */
  long      pdepth     = 0;         /* Paren depth */
  long      input_size = 0;         /* Input buffer size */
  long      size       = TEXT_SIZE;


  /* Check if output buffer is fixed */
  if ( maxlen && obuf ) 
  { 
    obuf[0] = UT_NULL_CHAR;
  }
  else
  {
    out_alloc = UT_TRUE;
    obuf = NULL;
  }

  /* Skip spaces */
  while( *iptr == ' ' )
    ++iptr;

  /* If the target char* is bigger than the standard working buffer, */
  /* use an allocated working buffer                                 */

  input_size = strlen( iptr ) + 2;

  if( input_size > TEXT_SIZE )
  {
    buf  = malloc( input_size * UT_SZ_C );
    size = input_size;
    work_alloc = 1;
  }

  optr = buf;


  /* Copy until we find a token delimeted by rsep.            */
  /*   Usually lsep = [  rsep = ]                             */
  /*   However, skip lsep and rsep pairs as well as () pairs. */
  /*                                                          */
  /*  Thus                                                    */
  /*      bbb]ccc                                             */
  /*  returns *result = bbb   value = ccc                     */
  /*      bbb[ccc]de]fg                                       */
  /*  returns *result = bbb[ccc]de  value = fg                */
  /*      bbb(a])cc]ee                                        */
  /*  returns *result = bbb(a])cc   value = ee                */


  if ( *iptr == lsep )
    ++iptr;


  // Code change  March 2, 2006
  // only increment pointer iptr at the bottom of the
  // loop, so that the pointer doesn't go past the 
  // found separator

  while ( depth > 0 && *iptr != UT_NULL_CHAR && size-- ) 
  {
    c = *iptr;

    if ( !last_esc )
    {
      if ( c == lpar )
        ++pdepth;

      if ( c == rpar )
        --pdepth;

      if ( pdepth <= 0 ) 
      {
        if ( c == rsep )
          --depth;

        if ( c == lsep )
          ++depth;
      }
    }

    last_esc = ( c == esc ) && !last_esc;

    if ( depth > 0 )
    {
      if ( !last_esc )
      {
        *optr = c;
        ++optr;
      }

      ++iptr;
    }

  }

  *optr = '\0';

  /* Remove trailing blanks */
  while ( optr > buf && *(--optr) == ' ' )
    *optr = '\0';

  /* Copy working buffer to output */
 
  if ( *buf ) 
  {
    if ( !out_alloc )
      cs_copy_siz( buf, obuf, maxlen );
    else
      if  ( work_alloc ) 
        obuf = buf;
      else 
        obuf = cs_dup( buf );
  }
  else
    if ( work_alloc )
    {
      free( buf );
    }

  // NOTE:  since code now breaks with pointer at separator,
  //        or NULL terminator, this code won't be executed.
  //        If the break is because "size == 0", then the
  //        pointer points to end of buffer and that 
  //        character should be a NULL.  If for some impossible
  //        reason it is not NULL, then incrementing iptr will
  //        cause it to point to some space beyond the buffer.

  /* Advance past trailing spaces in input buffer */
  if ( *iptr )
  {
    /* Skip spaces */
    while( *iptr == ' ' )
      ++iptr;
  }

  *pptr = iptr;

  return obuf;
}


/***************************************************************/
/*                                                             */
/* Function:   cs_parse_paren_token_validate                   */
/*                                                             */
/* Description:                                                */
/*       Perform the same task as cs_parse_paren_token(), but  */
/*       check that each opened separator or parenthesis is    */
/*       closed (in the correct order) before the end of the   */
/*       token.  If so, status is set to UT_TRUE, and the      */
/*       function returns normally.  If not, status is set to  */
/*       UT_FALSE, and the function returns NULL.              */
/*                                                             */
/*       NOTE:                                                 */
/*       The token is assumed to begin with the left separator,*/
/*       regardless of whether the first character is lsep.    */
/*                                                             */
/*       A maximum of sizeof(long) levels of separator nesting */
/*       (including parentheses) is allowed.  If this limit is */
/*       exceeded, status is set to UT_FALSE, and the function */
/*       returns NULL.                                         */
/*                                                             */
/*  Input parameters:                                          */
/*	       char  **pptr                                    */ 
/*             char    lsep                                    */
/*             char    rsep                                    */
/*             char   *result                                  */
/*             long    maxlen                                  */
/*                                                             */
/*  Output parameters:                                         */
/*             logical *status                                 */
/*                                                             */
/*  Returns:                                                   */
/*             char  *obuf                                     */
/*                                                             */
/***************************************************************/
char*  cs_parse_paren_token_validate( char** pptr, char lsep, char rsep,
                                      char* result, long maxlen, logical* status )
{
  TextBuf  gbuf;                /* The default working buffer, */
                                /* for small char*s - saves a malloc */

  char    *buf  = gbuf;     /* The working buffer */
  char    *iptr = *pptr;    /* The input buffer pointer */
  char    *optr = NULL;     /* The output buffer pointer */
  char    *obuf = result;   /* The output buffer */
  char    c     = '\0';
  char    esc   = '\\';
  char    lpar  = '(';
  char    rpar  = ')';

  logical out_alloc  = UT_FALSE;        /* The output is a fixed buffer */
  logical work_alloc = UT_FALSE;        /* The working buffer is an allocated buffer */
  logical last_esc   = UT_FALSE;
  long    sep_stack  = 1;                    /* Stack of open separators (lsep and '(') */
  long    depth      = 1;                    /* Separator depth */
  long    max_depth  = 8*sizeof(sep_stack);  /* Maximum separator depth */
  long    input_size = 0;                    /* Input buffer size */
  long    size        = TEXT_SIZE;


 /* Check if output buffer is fixed */
  if ( maxlen && obuf ) 
  { 
    obuf[0] = UT_NULL_CHAR;
  }
  else
  {
    out_alloc = UT_TRUE;
    obuf = NULL;
  }

  /* Skip spaces */
  while( *iptr == ' ' )
    ++iptr;

  /* If the target char* is bigger than the standard working buffer,*/
  /* use an allocated working buffer                                */

  input_size = strlen( iptr ) + 2;

  if( input_size > TEXT_SIZE )
  {
    buf = malloc( input_size * UT_SZ_C );
    size = input_size;
    work_alloc = 1;
  }

  optr = buf;


  /* Copy until all lsep and '(' characters (including      */
  /* the lsep that marks the start of the token) are closed */
  /* by a matching rsep or ')', or the separator depth      */
  /* exceeds max_depth, or we hit the end of the string.    */
  /*                                                        */
  /* depth counts the number of open separators, and the    */
  /* depth low-order bits of sep_stack denote the type      */
  /* of each separator: a 0 bit indicates an lpar, while a  */
  /* 1 bit indicates an lsep.                               */

  if ( *iptr == lsep )
    ++iptr;


  while ( depth > 0 && depth <= max_depth && *iptr != UT_NULL_CHAR && size-- ) 
  {
    c = *iptr;   

    if ( !last_esc )
    {
      if ( c == lpar )
      {
        sep_stack <<= 1;
        ++depth;
      }
      else
        if ( c == rpar )
        {
          /* Break if the last open separator was lsep (if should have been lpar) */
          if ( sep_stack & 1 )
            break;

          sep_stack >>= 1;
          --depth;
        }
        else
          if ( c == lsep )
          {
            sep_stack <<= 1;
            ++sep_stack;
            ++depth;
          }
          else
            if ( c == rsep )
            {
              /* Break if the last open separator was lpar (if should have been lsep) */
              if ( ~sep_stack & 1 )
                break;

              sep_stack >>= 1;
              --depth;
            }
    }

    last_esc = ( c == esc ) && !last_esc;

    if ( depth > 0)
    {
      if (!last_esc )
      {
        *optr = c;
        ++optr;
      }

      ++iptr;
    }
  }

  /* If we exited the loop with separators still open, signal an error */
  if ( depth > 0 )
  {
    *status = UT_FALSE;
    return NULL;
  }

  *optr = '\0';

  /* Remove trailing blanks */
  while ( optr > buf && *(--optr) == ' ' )
    *optr = '\0';

  /* Copy working buffer to output */
 
  if ( *buf ) 
  {
    if ( !out_alloc )
      cs_copy_siz( buf, obuf, maxlen );
    else
      if  ( work_alloc ) 
        obuf = buf;
      else 
        obuf = cs_dup( buf );
  }
  else
    if ( work_alloc )
    {
      free( buf );
    }


  // NOTE:  since code now breaks with pointer at separator,
  //        or NULL terminator, may cause problems down the
  //        line, since calling code may depend on being past
  //        the closing separator.  If current character is
  //        a right bracket ( ']' ), then increment the pointer.
  //        If execution got to this point, then there was a
  //        matched set of brackets.

  if ( *iptr == rsep )
    ++iptr;

  /* Advance past trailing spaces in input buffer */
  if ( *iptr )
  {
    /* Skip spaces */
    while( *iptr == ' ' )
      ++iptr;
  }

  *pptr = iptr;
  *status = UT_TRUE;

  return obuf;
}


/***************************************************************/
/*                                                             */
/* Function:   cs_parse_term_token                             */
/*                                                             */
/* Description:                                                */
/*            Calls cs_parse_paren_token() with some default   */
/*            parameters.                                      */
/*                                                             */
/*  Input parameters:                                          */
/*	       char  **pptr                                    */ 
/*             char    rsep                                    */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*             char  *obuf                                     */
/*                                                             */
/***************************************************************/
char*  cs_parse_term_token( char** pptr, char rsep )
{
  return  cs_parse_paren_token( pptr, UT_NULL_CHAR, rsep, NULL, 0 );
}


/***************************************************************/
/*                                                             */
/* Function:   cs_parse_assignment                             */
/*                                                             */
/* Description:                                                */
/*            Split char* into <name><op><remainder>           */
/*            where <op> is one of the characters in opset:    */
/*            !=, =, <, >                                      */
/*            Ignore everything within parenthesis, except     */
/*            if the ( character is part of the opset.         */
/*                                                             */
/*  Input parameters:                                          */
/*	       char  **pptr                                    */ 
/*             char   *name                                    */
/*             long    maxlen                                  */
/*             char   *op                                      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*             nothing                                         */
/*                                                             */
/***************************************************************/
void cs_parse_assignment( char** pptr, char* name, long maxlen,
                          char* op )
{
  char   opset[] = "!=<>";
  char  *iptr    = *pptr;
  char  *optr    = name;
  char  *op_ptr  = op;
  long   depth   = 0;
  long   pdepth  = 0;
  int    seek    = UT_TRUE;
  long   size    = maxlen;
  char   c       = '\0';
  int    doparens = UT_TRUE;    /* If ( is a valid opset character,*/
                                /*  don't check nested parens      */

  *op_ptr = '\0';
  *optr   = '\0';

  if ( strchr( opset, '(' ) )
    doparens = UT_FALSE;

  /* Skip leading blanks */
  while( *iptr == ' ' )
    ++iptr;

  while ( seek && *iptr != '\0' && size-- )
  {
    c = *iptr;
    if ( c == '('  && doparens )
      ++pdepth;

    if ( c == ')'  && doparens )
      --pdepth;

    if ( c == '['  )
      ++depth;

    if ( c == ']'  )
      --depth;

    if ( pdepth == 0 && depth == 0 )
    {
      if ( strchr( opset, c ))
      {
        if ( c == '!' && *(iptr+1) != '=' )  /* !> disallowed; ! must be part of != */
          seek = 1;
        else
          seek = 0;
      }
    }

    if ( seek )
      *optr++ = *iptr++; 
  }

  *optr = '\0';

  /* Delete trailing spaces */
  --optr;

  while( *optr == ' ' ) 
  {
    *optr = '\0';
    --optr;
  }

  if ( !*iptr )
  {
    *pptr = NULL;
    return;
  }

  /* Copy operator characters to op char*; allow !=, <=, ==, etc but disallow =! */ 
  /* since in chip=!circle(..) the ! belongs to the region not the operator      */
  if ( !seek )
  {
    *op_ptr = *iptr++;
    ++op_ptr;

    if ( strchr( opset, *iptr ) && *iptr!= '!' ) 
    {
      *op_ptr = *iptr++;
      ++op_ptr;
    }

    *op_ptr = '\0';

    /* Copy remaining characters to value char* by returning tptr */
  }

  /* Skip spaces */
  while( *iptr == ' ' )
    ++iptr;

  *pptr = iptr;

}


