/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

#include "utlib.h"
#include "unit.h"
/*
 *--------------------------------------------------------------------------------
 */

/*
 *   Here we parse a unit char* into a number of unit items, an array of base
 *   units, an array of exponents, and an array of SI prefix factors.
 */

void cs_unit_parse( const char* unit, UnitRegister* reg )
{
 TextWord unit_item;
 char* ptr;
 double exponent;
 TextWord base_token;
 long n;
 logical loop;

 n = 0;
 cs_copy( "10", reg->base[n] );
 reg->exp[n] = 0.0;
 reg->fac[n] = 0;
 loop = !cs_is_blank( unit );
 ptr = (char*)unit;
 while ( loop && ptr != NULL ) {
  cs_get_c( &ptr, unit_item, WORD_SIZE );
  unit_parse_exp( unit_item, base_token, &exponent );
  if ( n == 0 && cs_eq( base_token, "10" ) ) {
   reg->exp[n] = exponent;
  } else {
   n++;
   if ( n > reg->nmax )
    cs_unit_reg_enlarge( reg, 2 * n );
   cs_copy( base_token, reg->base[n] );
   reg->exp[n] = exponent;
   reg->fac[n] = 0;
  }
 }
 reg->n = n+1;
}

void cs_unit_reg_enlarge( UnitRegister* reg, long n )
{
 reg->base = (char**)realloc( reg->base, n * UT_SZ_P);
 reg->exp = (double*)realloc( reg->exp, n * UT_SZ_D);
 reg->fac = (long*)realloc( reg->fac, n * UT_SZ_I);
}

/*
 *--------------------------------------------------------------------------------
 */

void unit_parse_exp( const char* unit_item, char* base_unit, double* exp_unit )
{
 long j;
 long k;
 char* ptr;
 TextWord tmp;

 /* We look for a ^ character indicating the start of an exponent */

 j = cs_index_char( unit_item, '^' );
 if ( j > 0 ) {
  cs_copy_siz( unit_item, base_unit, j-1 );
  ptr = (char*)(unit_item + j); 

  /* Handle the case of TeX parens  x^{y} */

  if ( *ptr == '{' ) {
   k = cs_index_char( ptr, '}' );
   cs_copy_siz( ptr, tmp, k-1 );
   *exp_unit = cs_read_d( tmp );
  } else {

  /* or the case when we have a plain exponent   x^y */

   *exp_unit = cs_read_d( ptr );
  }
 } else {

 /* Default case is an exponent of 1 */

  cs_copy( unit_item, base_unit );
  *exp_unit = 1.0;
 }
 unit_parse_mantissa( base_unit, exp_unit );
}

/*
 *--------------------------------------------------------------------------------
 */

void unit_parse_mantissa( char* base_unit, double* exp_unit )
{
 logical ok;
 double log_value;
 long j,k;
 logical found_ten;
 TextWord number;
 double value;

 /* Check for a number that is not 10 */

 ok = cs_digit_char( *base_unit ) && !cs_eq( base_unit, "10" );
 log_value = 0;

 if ( ok ) {
  found_ten = UT_FALSE;
  k = cs_ends( base_unit );

  /* check for TeX  \times 10^2  syntax */

  j = cs_index( base_unit, "\\times10" );
  if ( j > 0 && j+7 == k ) {
   cs_copy_siz( base_unit, number, j-1 );
   found_ten = UT_TRUE;
  } else {
  /*  check for normal x10^2 syntax */
   j = cs_index( base_unit, "x10" );
   if ( j > 0 && j+2 == k ) {
    cs_copy_siz( base_unit, number, j-1 );
    found_ten = UT_TRUE;
   }
  }
/* if we have a leading number, make the base unit 10 and
 * take the log of the number to put in the base unit's exponent
 */
  if ( !found_ten ) {
   cs_copy( base_unit, number );
   if ( *exp_unit != 1.0 ) 
   {
     sprintf( ut_tmp, "Weird unit: %s", base_unit );
     ut_err_util( ut_tmp, UT_ERR_UNIT_PARSE );
   }
  }
  value = cs_read_d( number );
  if ( value > 0.0 ) log_value = log10( value );
  if ( found_ten ) {
   *exp_unit += log_value;
  } else {
   *exp_unit = log_value;
  }
  cs_copy( "10", base_unit );
 }

}



/*
 *--------------------------------------------------------------------------------
 */


void unit_compose( char** base_unit, double* exp_unit, long* si_fac, long n_uo, logical texmode, char* unit, long siz )
{
 long ten_exp_i;
 double ten_mantissa;
 char* ptr;
 long i;
 ptr = unit;
 cs_copy( " ", unit );
 if ( n_uo == 0 ) return;
 if ( cs_eq( base_unit[0], "10" ) && exp_unit[0] != 0.0 ) {
  frac_d( exp_unit[0], &ten_exp_i, &ten_mantissa );
  if ( ten_mantissa != 0.0 ) {
   cs_put_d( tpow_d( ten_mantissa ), &ptr, &siz );
   if ( texmode ) {
    cs_put_c( "\times", &ptr, &siz );
   } else {
    cs_put_c( "x", &ptr, &siz );
   }
  }
  unit_compitem( &ptr, base_unit[0], (double)ten_exp_i, 0, &siz );
  if ( !siz ) return;
 }

 for( i = 1; i < n_uo; i++ ) {
  if ( exp_unit[i] > 0.0 ) {
   unit_compitem( &ptr, base_unit[i], exp_unit[i], si_fac[i], &siz );
   if ( !siz ) return;
  }
 }
 for ( i = 1; i < n_uo; i++ ) {
  if ( exp_unit[i] < 0.0 ) {
   unit_compitem( &ptr, base_unit[i], exp_unit[i], si_fac[i], &siz );
   if ( !siz ) return;
  }
 }
}

/*
 *--------------------------------------------------------------------------------
 */



void unit_compitem( char** ptr, char* item, double exp_unit, long si_fac, long* siz )
{
 long k;
 double x;
 long n;
 char pfx[6];
 long ix;
 double rx;
 TextWord tmp;

 si_getpfx( si_fac, pfx, &k );
 if ( cs_is_blank( pfx ) ) { 
  cs_copy( item, tmp );
 } else {
  *tmp = *pfx;
  cs_copy( item, tmp+1 );
 }
 cs_put_c( tmp, ptr, siz );
 x = exp_unit;
 if ( x != 1.0 ) {
  frac_d( x, &ix, &rx );
  if ( rx == 0.0 ) {
   n = sprintf( *ptr, "^{%ld}", (long)ix );
  } else {
   n = sprintf( *ptr, "^{%s}", cs_write_d( x, tmp, WORD_SIZE ) );
  }
  *ptr += n;
  *siz -= n;
 }
 **ptr = ' ';
 (*ptr)++;
}

/*
 *--------------------------------------------------------------------------------
 */


void si_tmult( long fac1, double exponent1, long fac2, double exponent2, long* fac, double* mantissa )
{
 double result_factor;
 double exponent;

 result_factor = fac1 * exponent1 + fac2 * exponent2;
 exponent = exponent1 + exponent2;
 if ( exponent == 0.0 ) {
  *fac = 0;
  *mantissa = result_factor;
 } else {
  *fac = fac1;
  *mantissa = result_factor - exponent * fac1;
 }
}

/*
 *--------------------------------------------------------------------------------
  Assumes n1-arrays are sized n1+n2
 */

void unit_tmult( long* n1, char** base1, double* exponent1, long* si_fac1,
                 long* n2, char** base2, double* exponent2, long* si_fac2, long power )
{
 long i;
 long fac;
 long k;
 double mantissa;
 for ( i = 0; i < *n2; i++ ) {
  k = ar_cmatch_c( base2[i], base1, *n1 );
  if ( k >= 0 ) {
   si_tmult( si_fac1[k], exponent1[k], si_fac2[i], power * exponent2[i], 
     &fac, &mantissa );
   exponent1[k] += power * exponent2[i];
   si_fac1[k] = fac;
   exponent1[1] += mantissa;
  } else {
   (*n1)++;
   k = *n1;
   cs_copy( base2[i], base1[k] );
   exponent1[k] = power * exponent2[i];
   si_fac1[k] = si_fac2[i];
  } 
 }
}
