/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

#include "utlib.h"
#include "unit.h"


void unit_init_domain( UnitDomain* unit_domain_p );

void cs_unit_si_resolve( char** unit, long n, char** d_unit, long dn, long* fac )
{
 long i;
 long factor, pos;
 long j;
 TextWord temp;

 if ( n == 0 || dn == 0 ) return;
 for ( i = 0; i < n; i++ ) {
  si_check( unit[i], &factor, &pos );
  if ( pos > 1 ) {
   cs_copy( unit[i] + pos - 1, temp );
   j = ar_cmatch_c( temp, d_unit, dn );
   if ( j > 0 ) {
    cs_copy( temp, unit[i] );
    fac[i] += factor;
   }
  }
 }

}

void cs_unit_free( UnitDomain u )
{ 
 if ( !u ) return;
 ar_free_c( u->unit );
 ar_free_c( u->def );
 unit_free( u->ureg1 );
 unit_free( u->ureg2 );
 free( u );
}

UnitDomain cs_unit_init( const char* filename )
{
 FioFile file;
 long n;
 TextBuf line;
 char* ptr;
 UnitDomain unit_domain;
 long nmax = UT_UDF_MAX;

 unit_domain = calloc( 1, sizeof(UnitDomain));
 unit_domain->ureg1 = unit_alloc( UT_UNIT_MAX );
 unit_domain->ureg2 = unit_alloc( UT_UNIT_MAX ); 

 unit_domain->n = 0;
 if ( !filename || !(*filename) ) 
  return unit_domain;

 unit_domain->unit = ar_alloc_c( WORD_SIZE, nmax );
 unit_domain->def  = ar_alloc_c( WORD_SIZE, nmax );

 n = 0;
 file = fio_open_ar( filename );
 if ( file ) {
/* Fixed bug 1997 Mar 12, fio read result is logical */
  while ( fio_read_line( file, line, TEXT_SIZE )  ) {
   cs_detab( line );
   if ( !cs_is_blank( line ) ) {
    ptr = line;
    cs_get_c( &ptr, unit_domain->unit[n], WORD_SIZE );
    if ( ptr != NULL ) {
     cs_copy( ptr, unit_domain->def[n] );
    } else {
     cs_copy( " ", unit_domain->def[n] );
    }
    n++;
    if ( n >= nmax )
    {
     nmax = 2 * n;
     unit_domain->unit = (char**)realloc( unit_domain->unit, nmax * UT_SZ_P );
     unit_domain->def  = (char**)realloc( unit_domain->def, nmax * UT_SZ_P );
    }
   }
  }
 }
 unit_domain->n = n;
 fio_file_close( file );
 return unit_domain;
}




void cs_unit_dparse( UnitDomain unit_domain, const char* u1, char* u2, const long siz )
{
 if ( !unit_domain ) return;
 cs_unit_regparse( unit_domain, u1, unit_domain->ureg1 );
 unit_ddisp( unit_domain->ureg1 );
 unit_regcompose( unit_domain->ureg1, unit_domain->texmode, u2, siz );
}

void cs_unit_resolve( UnitDomain unit_domain, const char* u1, char* u2, const long siz )
{
 if ( !unit_domain ) return;
 cs_unit_regparse( unit_domain, u1, unit_domain->ureg1 );
 unit_ddisp( unit_domain->ureg1 );
 unit_regresolve( unit_domain );
 unit_regdedup( unit_domain->ureg1, unit_domain->ureg2 );
 unit_regcompose( unit_domain->ureg1, unit_domain->texmode, u2, siz );
}


void cs_unit_mult( UnitDomain unit_domain, const char* u1, const char* u2, char* u, const long siz )
{
 if ( !unit_domain ) return;
 cs_unit_regparse( unit_domain, u1, unit_domain->ureg1 );
 cs_unit_regparse( unit_domain, u2, unit_domain->ureg2 );
 unit_ddisp( unit_domain->ureg1 );
 unit_ddisp( unit_domain->ureg2 );
 unit_regmult( unit_domain->ureg1, unit_domain->ureg2, +1 );
 unit_ddisp( unit_domain->ureg1 );
 unit_regcompose( unit_domain->ureg1, unit_domain->texmode, u, siz );
}

void cs_unit_div( UnitDomain u, const char* u1, const char* u2, char* udiv, const long siz )
{
 if ( !u ) return;
 cs_unit_regparse( u, u1, u->ureg1 );
 cs_unit_regparse( u, u2, u->ureg2 );
 unit_ddisp( u->ureg1 );
 unit_ddisp( u->ureg2 );
 unit_regmult( u->ureg1, u->ureg2, -1 );
 unit_ddisp( u->ureg1 );
 unit_regcompose( u->ureg1, u->texmode, udiv, siz );
}



void cs_unit_exp( UnitDomain u, const char* unit1, const double power, char* unit2, const long siz )
{
 if ( !u ) return;
 cs_unit_regparse( u, unit1, u->ureg1 );
 unit_regexp( u->ureg1, power );
 unit_regcompose( u->ureg1, u->texmode, unit2, siz );

}


