/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

#include "utlib.h"
#include "unit.h"

/* The unit register routines, internal to the library */

UnitRegister* unit_alloc( const long n )
{
 UnitRegister* unit;
 unit = (UnitRegister *)malloc( sizeof( UnitRegister ) );
 unit->nmax = n;
 unit->n = 0;
 unit->base = ar_alloc_c( WORD_SIZE, n );
 unit->exp  = (double *)malloc( n * UT_SZ_D );
 unit->fac  = (long *)malloc( n * UT_SZ_I );
 return( unit );
}

void unit_free( UnitRegister* unit )
{
 free( unit->fac );
 free( unit->exp );
 ar_free_c( unit->base );
 free( unit );
}


void unit_regresolve( UnitDomain u )
{
 long i;
 long j;
 UnitRegister* reg1 = u->ureg1;
 UnitRegister* reg2 = u->ureg2;
 for ( i = 0; i < reg1->n; i++ ) {
  j = ar_cmatch_c( reg1->base[i], u->unit, u->n );
  if ( j >= 0 ) {
   if ( !cs_is_blank( u->def[j] ) ) {
    cs_unit_parse( u->def[j], reg2 );
    cs_unit_si_resolve( reg2->base, reg2->n, u->unit, u->n, reg2->fac );
    unit_regexp( reg2, reg1->exp[i] );
    reg2->exp[UT_UNIT_MANT_POS] += reg1->fac[i] * reg1->exp[i];
    reg1->exp[i] = 0.0;
    unit_ddisp( reg1 );
    unit_ddisp( reg2 );
    unit_regmult( reg1, reg2, +1 );
    unit_ddisp( reg1 );
   }
  }
 }

}

void unit_regdedup( UnitRegister* reg, UnitRegister* temp )
{
 long i;
 long j;
 long fac;
 double exponent;
 for ( i = 2; i < reg->n; i++ ) {
  j = ar_cmatch_c( reg->base[i], reg->base, i-1 );
  if ( j >= 0 ) {
   si_tmult( reg->fac[i], reg->exp[i], temp->fac[j], temp->exp[j], &fac, &exponent );
   reg->fac[j] = fac;
   reg->exp[UT_UNIT_MANT_POS] += exponent;
   reg->exp[j] += reg->exp[i];
   reg->exp[i] = 0.0;
  }
 }
 unit_ddisp( reg );
}

void unit_regcompose( UnitRegister* reg, logical texmode, char* unit, const long siz )
{
 unit_compose( reg->base, reg->exp, reg->fac, reg->n, texmode, unit, siz );
}

void unit_ddisp( UnitRegister* reg )
{
 long i;
 TextWord tmp;
 TextCard buf;
 if ( ut_dbg_level( UT_VERBOSE_UNIT ) ) {
  fio_dashline( 60 );
  for ( i = 0; i < reg->n; i++ ) {
   sprintf( buf, "%6ld %6ld %-10s %s", (long)i, (long)reg->fac[i], reg->base[i], 
                                   cs_write_d( reg->exp[i], tmp, WORD_SIZE ));
   fio_tmsg( buf );
  }
  fio_dashline( 60 );
 }
}

void cs_unit_regparse( UnitDomain u, const char* unit, UnitRegister* reg )
{
 cs_unit_parse( unit, reg );
 cs_unit_si_resolve( reg->base, reg->n, u->unit, u->n, reg->fac );
}

void unit_regexp( UnitRegister* reg, const double power )
{
 long i;
 for ( i = 0; i < reg->n; i++ ) {
  reg->exp[i] *= power;
 }
}


void unit_regmult( UnitRegister* reg1, UnitRegister* reg2, const long power2 )
{
 unit_tmult( &reg1->n, reg1->base, reg1->exp, reg1->fac,
             &reg2->n, reg2->base, reg2->exp, reg2->fac, power2 );
}


void cs_unit_texmode( UnitDomain u, const logical q )
{
 u->texmode = q;
}




