/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

/*
 * jcmerror.h   Internal error class
 *
 *   Handling Errors in JCMLIB
 *
   The most important thing is to set a detailed text error message,
   including such things as the name of a file and what the user
   can do to fix it.

   It is also sometimes important to return an long error code, 
   so that software can take different actions depending on the
   class of error.

   How can we maintain the error code namespace? One approach
   is to update the library each time a new error is required.
   A better approach is to provide a way to register new errors
   with the library. We could add an extra field for 'context',
   but it's easier to use a mask in the long error code itself.
   Most software subsystems have many fewer than 100 errors. 
   1000 is a safe range (you can always register more than one
   context per subsystem.)

   You want to be able to use the error system in two ways:
  - registering errors which will be used more than once
  - defining errors once on the fly.
   We'll use #defines in the .h file for the subsystem
   How to make this flexible?

 #define APE_READ_ERR 002
   ut_error_set( Error, APE_ERRS, "Read error", APE_READ_ERR, status );
   if ( ut_error_code( Error, APE_ERRS ) == APE_READ_ERR ) )...

or
   errors[APE_READ_ERR] = ut_register_err( APE_ERRS, ...)

   code = ut_error_code( APE_ERRS, APE_READ_ERR );

or
   ut_error_set( Error, APE_ERRS, "Read error", APE_ERRS | APE_READ_ERR, status );
   err = APE_ERRS | APE_READ_ERR;  ( Cool, this is easier to use )
   switch( code )
   {
     case ( APE_ERRS | APE_READ_ERR ) ...
   }
  Here, the APE_READ_ERR field is user-defined hard coded, and the APE_ERRS
   field is registered once per subsystem. A fair compromise...   


   We provide fields for:
    - the text message
    - the error code
    - the external error code, when another library (eg CFITSIO)
      is passing up info.

   We want to pass up a list of errors.

   This is a context in which a linked list is actually
   rather useful.

 *
 */


struct ErrorData_s {
 char* msg;
 long code;
 long auxcode;  
 struct ErrorData_s* next;
};

/* Advanced error stuff: internal UTILS errors */
struct ErrorGlobal_s {
 ErrorClasses error_classes;
 Error error;
 long user_mask;
 long ut_mask;
};

