/*                                                                
**  Copyright (C) 2002-2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */

#include "utlib.h"



void ar_vcopy_d( const double* p, double* r )
{
 long n = 3;
 while( n-- ) 
 {
  *r = *p++; r++;
 }
}

void ar_vzero_d( double* p )
{
 long n = 3;
 while( n-- )
 {  *p = 0.0; p++; }
}

void ar_vadd_d( const double* p, const double* q, double* r )
{
 long n=0;
 for ( n =0; n<3; n++ )
  r[n] = p[n] + q[n];
}

void ar_vsub_d( const double* p, const double* q, double* r )
{
 long n=0;
 for ( n =0; n<3; n++ )
  r[n] = p[n] - q[n];
}

void ar_vcmult_d( double* p, const double c )
{
 long n = 3;
 while( n-- )
 { *p *= c; p++; }
}

void ar_vdelta_d( const double* p, const double delta, const double* dp, double* r )
{
 long i;
 for ( i = 0; i < 3; i++ )
  r[i] = p[i] + delta * dp[i];
}


void ar_vcdiv_d( double* p, const double c )
{
 long n = 3;
 while( n-- )
 {  *p /= c; p++; }
}


double ar_vdot_d( const double* a, const double* b )
{
 return( a[0] * b[0] + a[1] * b[1] + a[2] * b[2] );
}

double ar_vnorm_d( const double* a )
{
 return( sqrt( a[0]*a[0] + a[1]*a[1] + a[2]*a[2] ) );
}

void ar_vcross_d( const double* a, const double* b, double* c )
{
 c[0] = a[1] * b[2] - a[2] * b[1];
 c[1] = a[2] * b[0] - a[0] * b[2];
 c[2] = a[0] * b[1] - a[1] * b[0];
}


double ar_vmaxmod_d( const double* r )
{
 double result = 0;
 double a;
 long n = 3;
 while ( n-- ) { a = fabs( *r++ ); result = ( a > result ) ? a : result; }
 return( result );
}

double ar_vunit_d( double* r )
{
 double r0;
 long n = 3;
 r0 = ar_vnorm_d( r );
 if ( r0 > 0.0 ) {
  while( n-- ) { *r /= r0; r++; }
 }
 return( r0 );
}

double ar_vcopy_unit_d( double* r, double* unit )
{
 double r0;
 long i;
 long n = 3;
 r0 = ar_vnorm_d( r );
 if ( r0 > 0.0 ) 
 {
  for ( i = 0; i < n; i++ )
   unit[i] = r[i] / r0;
 }
 else
 {
  ar_vzero_d( unit );
 }
 return( r0 );
}


float ar_vdot_r( const float* a, const float* b )
{
 return( a[0] * b[0] + a[1] * b[1] + a[2] * b[2] );
}

float ar_vnorm_r( const float* a )
{
 return( rsqrt( a[0]*a[0] + a[1]*a[1] + a[2]*a[2] ) );
}

void ar_vcross_r( const float* a, const float* b, float* c )
{
 c[0] = a[1] * b[2] - a[2] * b[1];
 c[1] = a[2] * b[0] - a[0] * b[2];
 c[2] = a[0] * b[1] - a[1] * b[0];
}


float ar_vmaxmod_r( const float* r )
{
 float result = 0;
 float a;
 long n = 3;
 while ( n-- ) { a = rabs( *r ); r++; result = ( a > result ) ? a : result; }
 return( result );
}

float ar_vunit_r( float* r )
{
 float r0;
 long n = 3;
 r0 = ar_vnorm_r( r );
 if ( r0 > 0.0 ) {
  while( n-- ) { *r /= r0; r++; }
 }
 return( r0 );
}




void ar_vcopy_r( const float* p, float* r )
{
 long n = 3;
 while( n-- ) { *r = *p++; r++; }
}

void ar_vzero_r( float* p )
{
 long n = 3;
 while( n-- ) { *p = 0.0; p++; }
}

void ar_vadd_r( const float* p, const float* q, float* r )
{
 long n=0;
 for ( n =0; n<3; n++ )
  r[n] = p[n] + q[n];

}

void ar_vsub_r( const float* p, const float* q, float* r )
{
 long n=0;
 for ( n =0; n<3; n++ )
  r[n] = p[n] - q[n];
}

void ar_vcmult_r( float* p, const float c )
{
 long n = 3;
 while( n-- ) {*p *= c; p++; }
}


/* Linear matrices */
void ar_lmat_prex_d( double* aa, const double* y, double* x )
{
 long i,j;
 long ji;
 for ( i = 0; i < 3; i++ ) {
  *x = 0.0;
  for( j = 0; j < 3; j++ ) {
   ji = 3 * j + i;
   *x += y[j] * aa[ji];
  }
  x++;
 }
}

void ar_lmat_postx_d( double* aa, const double* y, double* x )
{
 long i,j;
 long ij;
 for ( i = 0; i < 3; i++ ) {
  *x = 0.0;
  for( j = 0; j < 3; j++ ) {
   ij = i * 3 + j;
   *x += y[j] * aa[ij];
  }
  x++;
 }
}

void ar_mat_prex_d( double** aa, const double* y, double* x )
{
 long i,j;
 for ( i = 0; i < 3; i++ ) {
  *x = 0.0;
  for( j = 0; j < 3; j++ ) {
   *x += y[j] * aa[j][i];
  }
  x++;
 }
}

void ar_mat_prex_r( float** aa, const float* y, float* x )
{
 long i,j;
 for ( i = 0; i < 3; i++ ) {
  *x = 0.0;
  for( j = 0; j < 3; j++ ) {
   *x += y[j] * aa[j][i];
  }
  x++;
 }
}


void ar_mat_postx_d( double** aa, const double* y, double* x )
{
 long i,j;
 for ( i = 0; i < 3; i++ ) {
  *x = 0.0;
  for( j = 0; j < 3; j++ ) {
   *x += y[j] * aa[i][j];
  }
  x++;
 }
}

void ar_mat_postx_r( float** aa, const float* y, float* x )
{
 long i,j;
 for ( i = 0; i < 3; i++ ) {
  *x = 0.0;
  for( j = 0; j < 3; j++ ) {
   *x += y[j] * aa[i][j];
  }
  x++;
 }
}

void ar_mat_mult_d( double** a, double** b, double** c )
{
 long i,j,k;
 double w;
 for ( i = 0; i < 3; i++ ) {
  for( j = 0; j < 3; j++ ) {
   w = 0.0;
   for( k = 0; k < 3; k++ ) {
    w += a[i][k] * b[k][j];
   }
   c[i][j] = w;
  }
 }
}

void ar_lmat_mult_d( double* a, double* b, double* c )
{
 long i,j,k;
 double w;
 for ( i = 1; i <= 3; i++ ) {
  for( j = 1; j <= 3; j++ ) {
   w = 0.0;
   for( k = 1; k <= 3; k++ ) {
    w += ar_lmat33_d( a, i, k )  * ar_lmat33_d( b, k, j );
   }
   ar_lmat33_d( c, i, j ) = w;
  }
 }
}

void ar_mat_mult_r( float** a, float** b, float** c )
{
 long i,j,k;
 float w;
 for ( i = 0; i < 3; i++ ) {
  for( j = 0; j < 3; j++ ) {
   w = 0.0;
   for( k = 0; k < 3; k++ ) {
    w += a[i][k] * b[k][j];
   }
   c[i][j] = w;
  }
 }
}


/* Forward translate/rotate 3-vector */
void ar_transform_forward( const double* in, double* rotate, double* translate, double* out )
{
 long i,j;
 
 for ( i = 0; i < 3; i++ ) {
  out[i] = translate[i];  
  for ( j = 0; j < 3; j++ ) {
   out[i] += ar_lmat33_d( rotate, j+1, i+1 ) * in[j];
  }
 }
}

void ar_transform_backward( const double* in, double* rotate, double* translate, double* out )
{
 long i,j;
 for ( i = 0; i < 3; i++ ) {
  out[i] = 0.0;
  for ( j = 0; j < 3; j++ ) {
   out[i] += ar_lmat33_d( rotate, i+1, j+1 ) * ( in[j] - translate[j] );
  }
 }
}
 

void ar_euler_rotate( double* euler, double* rotate )
{
 long axes[3] = { 1, 2, 3 };
/* ZYZ */
 ar_euler_rotate_gen( euler, axes, rotate );
}

void ar_euler_rotate_alt( double* euler, double* rotate )
{
/* XYX */
 long axes[3] = { 3, 2, 1 };
 ar_euler_rotate_gen( euler, axes, rotate );
}

void ar_euler_rotate_gen( double* euler, long* axes, double* rotate )
{
 double cphi;
 double ctheta;
 double cpsi;
 double sphi;
 double stheta;
 double spsi;
 long a1,a2,a3;

 a1 =axes[0];
 a2 =axes[1];
 a3 =axes[2];

/* Accelerated case */
 if ( euler[0] == 270.0 && euler[2] == 90.0 ) 
 {
  cphi =  0.0;
  sphi = -1.0;
  ctheta=dcosd( euler[1] );
  stheta=dsind( euler[1] );
  cpsi  = 0.0;
  spsi  = 1.0;
 } else {
  cphi = dcosd( euler[0] );
  sphi = dsind( euler[0] );
  ctheta=dcosd( euler[1] );
  stheta=dsind( euler[1] );
  cpsi  =dcosd( euler[2] );
  spsi  =dsind( euler[2] );
 }
 ar_lmat33_d( rotate, a1, a1 ) =   cphi * ctheta * cpsi - sphi * spsi;
 ar_lmat33_d( rotate, a1, a2 ) =   sphi * ctheta * cpsi + cphi * spsi;
 ar_lmat33_d( rotate, a1, a3 ) =        - stheta * cpsi;
 ar_lmat33_d( rotate, a2, a1 ) = - cphi * ctheta * spsi - sphi * cpsi;
 ar_lmat33_d( rotate, a2, a2 ) = - sphi * ctheta * spsi + cphi * cpsi;
 ar_lmat33_d( rotate, a2, a3 ) =          stheta * spsi;
 ar_lmat33_d( rotate, a3, a1 ) =   cphi * stheta;
 ar_lmat33_d( rotate, a3, a2 ) =   sphi * stheta;
 ar_lmat33_d( rotate, a3, a3 ) =          ctheta;
}



void ar_euler_calc( double* r, double* euler )
{
 long axes[3] = { 1, 2, 3 };
 ar_euler_calc_gen( r, axes, euler );
}

void ar_euler_calc_gen( double* r, long* axes, double* euler )
{
 double phi, theta, psi;
 double s1, s2, s;
 double r31, r32, r13, r23;
 double x, y;
 long a1,a2,a3; 

 a1 = axes[0];
 a2 = axes[1];
 a3 = axes[2];
 r31 = ar_lmat33_d( r, a3, a1 );
 r32 = ar_lmat33_d( r, a3, a2 );
 r13 = ar_lmat33_d( r, a1, a3 );
 r23 = ar_lmat33_d( r, a2, a3 );
 s1 = sqrt( r31 * r31 + r32 * r32 );
 s2 = sqrt( r13 * r13 + r23 * r23 );
 s  = 0.5 * ( s1 + s2 );

 theta = argd_d(  ar_lmat33_d( r, a3, a3 ), s );
 if ( s1 > 0.0 && s2 > 0.0 )
 {
  x =  -ar_lmat33_d( r, a1, a3 ) / s2;
  y =   ar_lmat33_d( r, a2, a3 ) / s2;
  psi = argd_d( x, y );
  x =   ar_lmat33_d( r, a3, a1 ) / s1; 
  y =   ar_lmat33_d( r, a3, a2 ) / s1;
  phi = argd_d( x, y );
 } else {
  phi = argd_d( ar_lmat33_d( r, a1, a1 ), ar_lmat33_d( r, a1, a2 ) );
  psi = 0.0;
 }
 euler[0] = phi;
 euler[1] = theta;
 euler[2] = psi;

}

double utm_quat_unit( double* p )
{
 double norm;
 norm = sqrt( ar_dot_d( p, p, 4 ) );
 if ( norm > 0.0 )
 {
  ar_cmult_d( p, 4, 1.0/norm );
 }
 return norm;
}


double utm_quat_dot( double* p, double* q )
{
 return ar_dot_d( p, q, 4 );
}

/* Multiply quaternions */
void utm_quat_mult( double* p, double* q, double* r )
{
 r[0] = p[0] * q[0] - ar_dot_d( p+1, q+1, 3 );
 ar_vcross_d( p+1, q+1, r+1 );
 ar_vdelta_d( r+1,  p[0], q+1, r+1 );
 ar_vdelta_d( r+1,  q[0], p+1, r+1 );
}

void utm_quat_rot( double theta, double* n, double* p )
{
 double s;

 p[0] =  dcosd( 0.5 * theta );
 s = dsind( 0.5 * theta );
 p[1] =  s * n[0];
 p[2] =  s * n[1];
 p[3] =  s * n[2];
}

void utm_quat_to_matrix( double* q, double* rotate )
{
 ar_lmat33_d( rotate, 1, 1 ) =   q[0] * q[0] + q[1] * q[1] - q[2] * q[2] - q[3] * q[3];
 ar_lmat33_d( rotate, 1, 2 ) =   2 * ( -q[0] * q[3] + q[1] * q[2] );
 ar_lmat33_d( rotate, 1, 3 ) =   2 * (  q[0] * q[2] + q[1] * q[3] );
 ar_lmat33_d( rotate, 2, 1 ) =   2 * (  q[0] * q[3] + q[2] * q[1] );
 ar_lmat33_d( rotate, 2, 2 ) =   q[0] * q[0] - q[1] * q[1] + q[2] * q[2] - q[3] * q[3];
 ar_lmat33_d( rotate, 2, 3 ) =   2 * ( -q[0] * q[1] + q[2] * q[3] );
 ar_lmat33_d( rotate, 3, 1 ) =   2 * ( -q[0] * q[2] + q[3] * q[1] );
 ar_lmat33_d( rotate, 3, 2 ) =   2 * (  q[0] * q[1] + q[3] * q[2] );
 ar_lmat33_d( rotate, 3, 3 ) =  q[0] * q[0] - q[1] * q[1] - q[2] * q[2] + q[3] * q[3];
}


void utm_quat_from_matrix( double* rotate, double* q )
{
 double qq[4];
 long qi;
 double qx;
 double r11,r22,r33;
 r11 = ar_lmat33_d( rotate, 1, 1 );
 r22 = ar_lmat33_d( rotate, 2, 2 );
 r33 = ar_lmat33_d( rotate, 3, 3 );
 qi = 0;
 qq[0] = 1 + r11 + r22 + r33;
 qq[1] = 1 + r11 - r22 - r33;
 if ( qq[1] > qq[qi] ) qi = 1;
 qq[2] = 1 - r11 + r22 - r33;
 if ( qq[2] > qq[qi] ) qi = 2;
 qq[3] = 1 - r11 - r22 + r33;
 if ( qq[3] > qq[qi] ) qi = 3;

 qx = sqrt( qq[qi] / 4.0 );
 
 switch ( qi )
 {
  case 0:
   q[0] = qx;
   q[1] = 0.25 * ( ar_lmat33_d( rotate, 3, 2 ) - ar_lmat33_d( rotate, 2, 3 ) ) / qx;
   q[2] = 0.25 * ( ar_lmat33_d( rotate, 1, 3 ) - ar_lmat33_d( rotate, 3, 1 ) ) / qx;
   q[3] = 0.25 * ( ar_lmat33_d( rotate, 2, 1 ) - ar_lmat33_d( rotate, 1, 2 ) ) / qx;
   break;
  case 1:
   q[0] = 0.25 * ( ar_lmat33_d( rotate, 3, 2 ) - ar_lmat33_d( rotate, 2, 3 ) ) / qx;
   q[1] = qx;
   q[2] = 0.25 * ( ar_lmat33_d( rotate, 1, 2 ) + ar_lmat33_d( rotate, 2, 1 ) ) / qx;
   q[3] = 0.25 * ( ar_lmat33_d( rotate, 1, 3 ) + ar_lmat33_d( rotate, 3, 1 ) ) / qx;
   break;
  case 2:
   q[0] = 0.25 * ( ar_lmat33_d( rotate, 1, 3 ) - ar_lmat33_d( rotate, 3, 1 ) ) / qx;
   q[1] = 0.25 * ( ar_lmat33_d( rotate, 1, 2 ) + ar_lmat33_d( rotate, 2, 1 ) ) / qx;
   q[2] = qx;
   q[3] = 0.25 * ( ar_lmat33_d( rotate, 2, 3 ) + ar_lmat33_d( rotate, 3, 2 ) ) / qx;
   break;
  case 3:
   q[0] = 0.25 * ( ar_lmat33_d( rotate, 2, 1 ) - ar_lmat33_d( rotate, 1, 2 ) ) / qx;
   q[1] = 0.25 * ( ar_lmat33_d( rotate, 1, 3 ) + ar_lmat33_d( rotate, 3, 1 ) ) / qx;
   q[2] = 0.25 * ( ar_lmat33_d( rotate, 2, 3 ) + ar_lmat33_d( rotate, 3, 2 ) ) / qx;
   q[3] = qx;
   break;   
 }
}


void ar_lmatnn_prex( double* v1, double* r, double* v2, long n )
{
 long i, j;

 double w;

 for ( i =1; i <= n; i++ )
 {
  w = 0.0;
  for ( j = 1; j <= n; j++ )
   w += ar_lmatnn_d( r, n, i, j ) * v1[j-1];
  v2[i-1] = w;
 }
}

