/*                                                                
**  Copyright (C) 1999-2001,2003-2004,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/*****************************************************************/
/*                                                               */
/*   dmtcommon.c                                                 */
/*                                                               */
/*   Functions that are common between dmcopy and dmlist         */
/*   and were formerly in dmtlib                                 */
/*                                                               */
/*                                                               */
/*****************************************************************/

#if !defined(DMLOCAL)
#include "liberr.h"
#endif

#include "dmtlib.h"
#include "dmtcommon.h"


static FILE* dmlist_out = NULL;


/***************************************************************/
/*                                                             */
/* Function:   dmaFlagDSSBlocks                                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*   Return a flag array is_dss_block such that if             */
/*   is_dss_block[i] is true, then block i is a DSS block      */
/*   (e.g. a GTI or region) attached to main_block.  The array */
/*   must be freed by the caller.                              */
/*                                                             */
/*   Called from dmaProcessStacks()                            */
/*                                                             */
/*  Input parameters:                                          */
/*        dmDataset   *ds                                      */
/*        dmBlock     *main_block                              */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*        int  is_dss_block                                    */
/*                                                             */
/***************************************************************/
static int* dmaFlagDSSBlocks( dmDataset* ds, dmBlock* main_block )
{
  char table_name[DC_MAXSTR+1];
  char block_name[DC_MAXSTR+1];

  dmDescriptor  **ss_cols = NULL;

  int   *is_dss_block = 0;
  long   nblocks      = 0;
  long   ncols        = 0;
  long   ncpts        = 0;
  long   blockno      = 0;
  long   colno        = 0;
  long   cptno        = 0;

  nblocks = dmDatasetGetNoBlocks( ds );
  is_dss_block = (int*)calloc( nblocks, sizeof(int) );

  ss_cols = dmBlockGetSubspace( main_block, &ncols, &ncpts );

  for ( cptno = 1; cptno <= ncpts; cptno++ ) {
    /*
     * Loop over subspace components
     */

    dmBlockSetSubspaceCpt( main_block, cptno );

    for ( colno = 1; colno <= ncols; colno++ ) {
      /*
       * Loop over subspace columns
       */

      dmSubspaceColGetTableName( ss_cols[colno-1], table_name, DC_MAXSTR );

      if ( table_name[0] )
      {
	/*
	 * Find the first block whose name matches table_name and flag it
	 * as a DSS block.  (If there are multiple blocks with the name
	 * table_name, the DM will find the first one when it goes to open
	 * the subspace table by name, so this is the one we want to mark as
	 * part of the DSS.)
	 */

        for ( blockno = 1; blockno <= nblocks; blockno++ ) {
	  dmDatasetGetBlockName( ds, blockno, block_name, DC_MAXSTR );

	  if ( !strcmp( table_name, block_name ) ) {
	    is_dss_block[blockno-1] = 1;
	    break;
	  }
	}
      }
    }
  }

  return is_dss_block;
}


/***************************************************************/
/*                                                             */
/* Function:   dmaErrInit                                      */
/*                                                             */
/* Description:                                                */
/*   Initialize Error handling                                 */
/*                                                             */
/* Input parameters:                                           */
/*   char *tool  - Tool name tag                               */
/*                                                             */
/* Output parameters:                                          */
/*   none                                                      */
/*                                                             */
/* Returns:                                                    */
/*   dmSUCCESS                                                 */
/*                                                             */
/***************************************************************/
int dmaErrInit( char* tag )
{
#if !defined(DMLOCAL)
  init_error_lib( tag );
#endif
  return(dmSUCCESS);
}

/***************************************************************/
/*                                                             */
/* Function:   dmaErrExit                                      */
/*                                                             */
/* Description:                                                */
/*       Prints error message and exits program                */
/*                                                             */
/*       Called from dmaProcessStacks                          */
/*                   dmaPickNextOutputFile                     */
/*                   dmsPickNextOutputDataset                  */
/*                   dmcopy_clobber                            */
/*                                                             */
/*  Input parameters:                                          */
/*                  char  *buf  - error message                */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            int  1                                           */
/*                                                             */
/***************************************************************/
int dmaErrExit( char* buf )
{
#if defined(DMLOCAL)
  fprintf( stderr, "%s\n", buf );
#else
  err_msg( buf );
#endif
  exit(1);
}

/***************************************************************/
/*                                                             */
/* Function:   dmaErrPrint                                     */
/*                                                             */
/* Description:                                                */
/*   Prints error message                                      */
/*                                                             */
/* Input parameters:                                           */
/*   char *buf  - error message                                */
/*                                                             */
/* Output parameters:                                          */
/*   none                                                      */
/*                                                             */
/* Returns:                                                    */
/*   dmSUCCESS                                                 */
/*                                                             */
/***************************************************************/
int dmaErrPrint( char* buf )
{
#if defined(DMLOCAL)
  fprintf( stderr, "%s\n", buf );
#else
  err_msg( buf );
#endif
  return(dmSUCCESS);
}

/***************************************************************/
/*                                                             */
/* Function:   dmaGetNextOutputDataset                         */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*   Called from dmsProcessStacks                              */
/*                                                             */
/*  Input parameters:                                          */
/*        Stack       outstack                                 */
/*        dmBool      clobber                                  */
/*        int         debug                                    */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*        dmDataset *destDs                                    */
/*                                                             */
/***************************************************************/
dmDataset* dmaGetNextOutputDataset( char* outfile, dmBool clobber, int debug)
{
  dmDataset *destDs  = NULL;
  char       errbuf[DC_MAXSTR];


  if ( !outfile || cs_is_blank(outfile) ) 
    dmaErrExit( "No output file specified" );

  if ( debug >= 2 )
    printf( "Output file is %s\n", outfile );

  dmcopy_clobber( outfile, clobber, debug );
  destDs = dmDatasetCreate( outfile );

  if ( !destDs )
  {
    fprintf( stderr, "Couldn't create output dataset %s\n", outfile);
    dmGetErrorMessage( errbuf, DC_MAXSTR );
    dmaErrExit( errbuf );
  }

  return destDs;
}



/***************************************************************/
/*                                                             */
/* Function:   dmaGetNextOutputFile                            */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*   Called from dmsProcessStacks                              */
/*                                                             */
/*  Input parameters:                                          */
/*        Stack       outstack                                 */
/*        dmBool      clobber                                  */
/*        int         debug                                    */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*        FILE  *out_stream                                    */
/*                                                             */
/***************************************************************/
FILE* dmaGetNextOutputFile( char* outfile, dmBool clobber, int debug)
{
  FILE *out_stream = NULL;
  char  errbuf[DC_MAXSTR];


  if ( !clobber )
  {
    out_stream = fopen( outfile, "r" );

    if ( out_stream )
    {
      sprintf( errbuf, "%s exists and clobber is no", outfile );
      dmaErrExit( errbuf );
    }
  }

  /* Setup output */
  out_stream = stdout;

  if ( !cs_is_blank( outfile ) ) 
  {
    out_stream = fopen( outfile, "w" );

    if ( !out_stream )
    {
      sprintf( errbuf, "No write permission for output file %s", outfile );
      dmaErrExit( errbuf );
    }
    dmlist_io_out( out_stream );
  } 

  return out_stream;
}



/***************************************************************/
/*                                                             */
/* Function:   dmaProcessFile                                  */
/*                                                             */
/* Description:                                                */
/*   This is the general routine for processing files          */
/*                                                             */
/*   Called from n_dmlist() and n_dmcopy()                     */
/*                                                             */
/*  Input parameters:                                          */
/*        char        *inname      - input file name           */
/*        char        *outname     - ouptut file name          */
/*        char        *toolname    - tool name for HISTORY     */
/*        int          mode        - control flag              */
/*        int          debug       - verbose parameter 0-5     */
/*        BlockProcess func        - function to call          */
/*        void        *params      - struc ptr for function    */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*        int  exit-status                                     */
/*                                                             */
/***************************************************************/
int dmaProcessFile( 
  char* inname,            /* (i)  Name of input file  */
  char* outname,           /*      Name of output file */
  char* toolname,          /*      Tool Name for HISTORY record to add to output */
  int   mode,              /*      Control flags */
  int   debug,             /*      Verbose parameter 0-5 */
  BlockProcess func,       /*      Function to apply to each entry in the stack */
  void* params,            /*      Structure pointer to pass to BlockProcess function */
  long  fileNumber,
  long  numberInFiles  )

{
  FILE   *out_stream = stdout;

  dmCopyInfo   info;              /* Stack state info to pass to block process function */
  dmDataset   *srcDs     = NULL;
  dmDataset   *destDs    = NULL;
  dmBlock     *srcBlock  = NULL;
  dmBlock     *destBlock = NULL;
  dmBlock     *block     = NULL;

  char 	     blname[DC_MAXSTR];
  char 	     outkern[DC_MAXSTR];
  char 	     inkern[DC_MAXSTR];
  char 	     outblname[DC_MAXSTR];
  char       errStr[dmMAXSTR+1]     = ""; 

  char  *infile        = inname;
  char  *ckern         = NULL;

  long   first_block   = 0;
  long   last_block    = 0;
  long   mainBlockNo   = 0;
  long   outputBlockNo = 0;
  int    copySS        = 1;


  if( ! (cs_is_blank(infile)) )
  {
    if ( mode & DMA_MODE_FORCE_IMAGE )  /* Force image */
    {
      if ( debug >= 2 )
        printf( "Open %s as image\n", infile );

      srcBlock = dmImageOpen(infile);
    }
    else
    {
      if ( debug >= 2 )
        printf( "Open %s as table or image\n", infile );
      srcBlock = dmBlockOpen( NULL, infile);
    }

    dmGetErrorMessage( errStr,dmMAXSTR );    

    /* exit if open failed */

    if (!srcBlock)  
    {          
      fprintf( stderr, "\n" );
      fprintf( stderr, "Failed to open virtual file %s\n", infile );

      if ( *infile == '[' )
      {
        fprintf( stderr, "Perhaps there is an unwanted space before the [ in \"%s\"?\n", inname );
      }
      dmaErrExit( errStr );
    }
    else
      if ( dmGetError() != dmSUCCESS )
      {
        dmaErrPrint( errStr );
        dmResetErrorState();
      }

    if ( mode & DMA_DELETE_GTI )
    {
      dmDescriptor *dd = dmSubspaceColOpen( srcBlock, "TIME" );
      double t1 = 0.0;
      double t2 = 0.0;

      dmKeyRead_d( srcBlock, "TSTART", &t1 );
      dmKeyRead_d( srcBlock, "TSTOP", &t2 );
      dmSubspaceColSet_d( dd, &t1, &t2, 1 );
    }

    dmBlockGetName( srcBlock,outblname,DC_MAXSTR );

    if ( debug > 2 )
      printf( "Opened virtual block %s \n", outblname );


    srcDs = dmBlockGetDataset( srcBlock );

    dmDatasetGetKernel( srcDs, inkern, DC_MAXSTR );

    dmKernelGetCopy( outkern, DC_MAXSTR );

    if ( !strcmp( outkern, "DEFAULT" ))
      ckern = dmKernelSetCreate( inkern );
    else
      ckern = dmKernelSetCreate( outkern );


    if ( mode & DMA_MODE_TEXT )
    {
      out_stream = dmaGetNextOutputFile( outname, mode & DMA_MODE_CLOBBER, debug);
      dmKernelGetCreate( outkern, DC_MAXSTR );
    }
    else
    {
      destDs = dmaGetNextOutputDataset( outname, mode & DMA_MODE_CLOBBER, debug);
      dmDatasetGetKernel( destDs, outkern, DC_MAXSTR );
    }


    /* Now we map the input blocks to the output blocks.                */
    /*  In 'All' mode, each input block maps to the corresponding       */
    /*  output block.  Otherwise, in text mode we copy a single block   */
    /*  In copy mode we copy the primary header if needed, and we don't */         
    /*  copy the subspace of the first block (which is primary header). */

    first_block = 1;
    mainBlockNo = 1;

    if ( mode & DMA_MODE_COPY_ALL )
    {
      mainBlockNo = dmBlockGetNo( srcBlock);
      last_block = dmDatasetGetNoBlocks( srcDs );
    }
    else
    {
      if ( mode & DMA_MODE_TEXT)
      {
        last_block  = 1;
      }
      else
      {
        /* Get the next output file; text or dataset mode */
        if ( dmBlockGetType( srcBlock ) == dmTABLE && !strcmp( inkern, "FITS" ) )
        {
          copySS = 0;
 	  mainBlockNo = 2;

          if ( !strcmp( outkern, "IRAF" ) || !strcmp( outkern, "TEXT" ) || ( mode & DMA_MODE_COPY_BARE ) ) 
            first_block = 2;
        }
        last_block  = mainBlockNo;	  
      }
    }

    /* Start looping on blocks. Fill the info structure. */
    /* Each iteration needs to know the output details,  */
    /* the pointer to the main input block.              */

    if ( debug > 2 )
      printf("Do the copy: \n" );

    info.first    = 1;
    info.ds       = destDs;
    info.name     = outblname;
    info.block    = srcBlock;
    info.nfiles   = numberInFiles;
    info.toolname = toolname;
    info.input    = infile;
    info.is_dss_block = dmaFlagDSSBlocks( srcDs, srcBlock );

    for( outputBlockNo = first_block; outputBlockNo <= last_block; outputBlockNo++ )
    {
      destBlock = NULL;

      info.copy_subspace = copySS;
      info.main = ( outputBlockNo == mainBlockNo );

      if ( info.main )
        block = srcBlock;
      else
        block = dmDatasetMoveToBlock( srcDs, outputBlockNo );

      dmBlockGetName( block, blname, DC_MAXSTR );

      if ( debug > 2 ) 
        printf( "Operate on block %s\n", blname );

      destBlock = func( block, debug, &info, params );

      /* We need to keep the main block open as it may be accessed in later processing */
      if ( !info.main )
        dmBlockClose( block );

      /* Note that in text mode destBlock is always null */
      if ( destBlock )
        dmBlockClose( destBlock );

    }

    /* Close up shop */
    if ( destDs )
      dmDatasetClose( destDs );

    if ( out_stream != stdout )
      fclose( out_stream ); 	   

    dmTableClose( srcBlock );
    free( info.is_dss_block );
  } 

  return EXIT_SUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmcopy_clobber                                  */
/*                                                             */
/* Description:                                                */
/*   Clobber specified file                                    */
/*                                                             */
/*   Called from dmaGetNextOutputDataset()                     */
/*                                                             */
/*  Input parameters:                                          */
/*    char   *file            - name of file                   */
/*    dmBool clobber          - clobber control flag           */
/*    int    verbose          - verbosity level                */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*    none                                                     */
/*                                                             */
/***************************************************************/
void dmcopy_clobber( char* outfile, dmBool clobber, int verbose )
{
  char *ofile = cs_dup(outfile);

  /* remove output filename extension if one exists */
  strcpy(ofile, outfile );
  outfile= strchr( ofile, '[' );
  if (outfile)
    *outfile = '\0';
  outfile = ofile;
 
  /* Check for existence of output dataset */
  if ( dmFileExists( outfile ) != DMF_NOFILE )
  {
    if ( clobber )
    {
      if ( verbose )
        fprintf( stderr, "Clobbering old output file: %s\n", outfile);

      if ( dmDatasetDestroy( outfile ) != dmSUCCESS )
      {
        fprintf( stderr, "%s\n", "Can't clobber output file." );
        exit(1);        
      }
    }
    else 
    {
      fprintf( stderr, "Clobber set to no, and output file %s exists.\n", outfile );
      exit(1);
    }
  }
}


#if 0
/***************************************************************/
/*                                                             */
/* Function:   dmKernelSetOptionList                           */
/*                                                             */
/* Description:                                                */
/*   Send comma delimited list of options to kernel            */
/*                                                             */
/*   Called from n_dmlist() and n_dmcopy()                     */
/*                                                             */
/*  Input parameters:                                          */
/*    char   *opt             - option string                  */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*    none                                                     */
/*                                                             */
/***************************************************************/
void dmKernelSetOptionList( char* opt )
{
  long   ii      = 0;
  long   nopts   = 0;
  char **optlist = NULL;     /* Hold parsed array of input options */
  char  *option  = NULL;     /* Hold one option                    */


  /* Split option list into tokens */
  optlist = cs_parse_comma_list( opt, &nopts );

  for ( ii=0; ii<nopts; ii++)
  {
    option = optlist[ii];

    dmKernelSetOption( option );

    dmFree(option);
  }

  dmFree(optlist);
}
#endif



/***************************************************************/
/*                                                             */
/* Function:   dmlist_io_out                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*   Called from functions dmaGetNextOutputFile()              */
/*                         n_dmlist()                          */
/*                                                             */
/* Input parameters:                                           */
/*    FILE  *file                                              */
/*                                                             */
/* Output parameters:                                          */
/*                                                             */
/*                                                             */
/* Returns:                                                    */
/*                                                             */
/*                                                             */
/***************************************************************/
void dmlist_io_out( FILE* file )
{
 dmlist_out = file;
}


/***************************************************************/
/*                                                             */
/* Function:   dmlist_print_std                                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*   Called from functions dmlist_print()                      */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*       char  *buf                                            */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           none                                              */
/*                                                             */
/***************************************************************/
void dmlist_print_std( char* buf )
{
  fprintf( dmlist_out, "%s\n", buf );
}



/***************************************************************/
/*                                                             */
/* Function:   dmu_option_match                                */
/*                                                             */
/* Description:                                                */
/*    Match a string against an array of possibilities.        */
/*                                                             */
/* Input parameters:                                           */
/*    char  **optarr     - array of valid options              */
/*    int     nopts      - size of array                       */
/*    char   *option     - test string                         */
/*    long    minsize    - min # of chars to match             */
/*                                                             */
/* Output parameters:                                          */
/*    long    nmatch     - # of options matching               */
/*                                                             */
/* Returns:                                                    */
/*    int     ndx        - index of matching option            */
/*                                                             */
/***************************************************************/
int dmu_option_match( char** optarr, int nopts, char* tststr, long minsize, long* nmatch )
{
  int   jj     = 0;
  int   ndx    = -1;
  char *ptr    = NULL;
  char *name   = cs_dup( tststr );
  char *option = NULL;

  /* Isolate name portion from <name>=<value> type */
  if ( (ptr = strchr( name, '=' )) )
    *ptr = '\0';

  for ( jj = 0; jj < nopts; jj++ )
  {
    option = optarr[jj];
    
    if ( !strcmp( option, name ) )
    {
      /* exact match */
      *nmatch = 1;
      ndx     = jj;
      break;
    }

    if ( strlen(name) >= minsize )
    {
      if ( strstr( option, name ) == option )
      {
	*nmatch += 1;
	if ( ndx < 0 )
	  ndx = jj;  /* mark first match */
      }
    }
  }
  dmFree( name );

  return( ndx );
}



