/*                                                                
**  Copyright (C) 1999-2001,2003-2004,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/*****************************************************************/
/*                                                               */
/*   dmcopy_funcs.c                                              */
/*                                                               */
/*   Functions that are used by dmcopy and were formerlly in     */
/*   dmtlib                                                      */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmnutils.h"
#include "dmtlib.h"
#include "dmcopy_main.h"
#include "dmtcommon.h"

#if !defined(DMLOCAL)
#include "histlib.h"
#include "dmtprocess.h"
#endif


/***************************************************************/
/*                                                             */
/* Function:   n_dmcopy                                        */
/*                                                             */
/* Description:                                                */
/*   Setup and execute dmaProcess to perform copies of         */
/*   desired blocks                                            */
/*                                                             */
/*   Called from function main()                               */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         char   *inname                                      */
/*         char   *outname                                     */
/*         char   *optstr                                      */
/*         char   *kernelopt                                   */
/*         dmBool  clobber                                     */
/*         int     verbose                                     */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*    dmSUCCESS or Error code                                  */
/*                                                             */
/***************************************************************/
int n_dmcopy( char* inname, char* outname, char* optstr, char* kernelopt, 
	      dmBool clobber, int verbose )
{
  char  toolname[DC_MAXSTR]; /* Tool name "dmcopy"                        */

  char  outkern[DC_MAXSTR];  /* Hold kernel part of kernelopt param value */
  char  flavor[DC_MAXSTR];   /* Hold flavor part of kernelopt param value */
  char  errbuf[DC_MAXSTR];

  char  *cptr     = NULL;
  char **optlist  = NULL;    /* Hold parsed array of input options */
  char  *tmpstr   = NULL;    /* string to hold unmatched options   */
  char **toolopts = NULL;

  int    status   = dmSUCCESS;
  int    mode     = 0;
  int    ntoolopts = 4;
  long   ii       = 0;
  long   nopts    = 0;
  long   sizeOfUnsignedChar = 1;


  toolopts = ar_alloc_c( 10, ntoolopts );
  sprintf( toolopts[0], "ALL");
  sprintf( toolopts[1], "BARE");
  sprintf( toolopts[2], "IMAGE");
  sprintf( toolopts[3], "NOGTI");


  dmSetVerbose( verbose );

  if ( clobber )
    mode |= DMA_MODE_CLOBBER;

  /* Convert option list to upper case */
  cs_upper( optstr );

  /* Split option list into tokens */
  optlist = cs_parse_comma_list( optstr, &nopts );

  for ( ii=0; ii<nopts; ii++)
  {
    char* option = optlist[ii];

    int  ndx     = 0;  /* index of matching option                        */
    long nmatch  = 0;  /* number of matches found                         */
    long minsize = 3;  /* minimum size of option string for partial match */

    ndx = dmu_option_match( toolopts, ntoolopts, option, minsize, &nmatch );

    if ( nmatch == 1 )
    {
      switch ( ndx )
      {
        case 0:  /* ALL */
	{
	  mode |= DMA_MODE_COPY_ALL;
	  break;
	}
        case 1:  /* BARE */
	{
	  mode |= DMA_MODE_COPY_BARE;
	  break;
	}
        case 2:  /* IMAGE */
	{
	  mode |= DMA_MODE_FORCE_IMAGE;
	  break;
	}
        case 3:  /* NOGTI */
	{
	  mode |= DMA_DELETE_GTI;
	  break;
	}
        default:
	  break;

      } /* end switch     */
    }  /* end if matched */
    else
    {
      /* Not a matched option.. */
      if ( !tmpstr )
      {
	tmpstr = calloc(strlen(optstr), sizeOfUnsignedChar );
	strcpy( tmpstr, option );
      }
      else
      {
	strcat( tmpstr, ",");
	strcat( tmpstr, option);
      }
    }

  } /* end for        */

  /* Free memory */
  ar_free_cp( optlist, nopts );
  ar_free_c( toolopts );

  /* Transfer unused options to option string */
  if ( tmpstr )
  {
    strcpy( optstr, tmpstr );
    free( tmpstr );
  }
  else
  {
    *optstr = '\0';
  }

  /* Handle Kernel spec at 'kernel' parameter    */
  /* Use it to (re)set the default output kernel */
  /* NOTE: The kernel parameter is NOT the preferred method of */
  /*       specifying the kernel.                              */
  cs_upper( kernelopt );
  cs_split_string( kernelopt,  '/', outkern, flavor );

  dmKernelSetCopy( outkern );

  if ( dmGetError() != dmSUCCESS )
  {
    dmGetErrorMessage( errbuf, DC_MAXSTR );
    dmaErrExit(errbuf);
  }

  /* NOTE: if the kernel parameter is just a kernel spec, add     */
  /* it to the output filename rather than sending it to the      */
  /* kernel option list which is used for input as well as output.*/

  cptr = flavor;

  if ( !cs_ss( flavor, "=" ) && !cs_eq( kernelopt, "DEFAULT") )
  {
    /* add kernel spec if not already on the filename.*/
    if ( !cs_ss( outname, "kernel=" ) )
    {
      strcat(outname, "[opt kernel=");
      strcat(outname, kernelopt);
      strcat(outname, "]");
    }
    *cptr = '\0';
  }

  if ( !*cptr )
    cptr = optstr;

  /* Send kernel options to kernel */
  if ( cptr && *cptr )
    dmKernelSetOption( cptr );

  /* Create the toolname string */
  strcpy( toolname, "dmcopy" );

#if !defined( DMLOCAL)
  status = dmaProcessStacks( inname, outname, toolname, mode, verbose,
			     (BlockProcess)dmCopyBlock, NULL);
#else
  status =  dmaProcessFile( inname, outname, toolname, mode, verbose,
			(BlockProcess)dmCopyBlock, NULL, 1, 1 );
#endif

  return status;

}



/***************************************************************/
/*                                                             */
/* Function:   dmCopyBlock                                     */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*   Passed as parameter to dmaProcessStacks and               */
/*   dmaProcessFile from n_dmcopy()                            */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmBlock *block                                      */
/*         int      debug                                      */
/*         void    *vinfo                                      */
/*         void    *params                                     */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*         dmBlock *destBlock                                  */
/*                                                             */
/***************************************************************/
dmBlock* dmCopyBlock( dmBlock* block, int debug, void* vinfo, void* params )
{
  dmCopyInfo *info      = (dmCopyInfo*) vinfo;
  dmBlock    *destBlock = NULL;

  char 	blname[DC_MAXSTR];
  char  errbuf[DC_MAXSTR];

  long  blockno = 0;



  if ( info->main )
  {
    destBlock = dmBlockCreateGenCopy( info->ds,info->name, block, NULL ,dmTRUE,dmTRUE );

    if ( !destBlock )
    {
      dmGetErrorMessage( errbuf, DC_MAXSTR );
      dmaErrExit(errbuf);
    } 
    else
      if ( dmGetError() != dmSUCCESS )
      {
        dmGetErrorMessage( errbuf, DC_MAXSTR );
        dmaErrPrint( errbuf );
      }

#if !defined( DMLOCAL)
    /* Use DS History library to write history */
    put_param_hist_info( destBlock, info->toolname, NULL, 0 );
#endif

  }
  else
  {
    blockno = dmBlockGetNo( block );

    /* We won't copy DSS blocks because they'll be copied */
    /* with the block to which they're attached           */

    if ( !(info->is_dss_block[blockno-1]) )
    {          
      dmBlockGetName( block, blname, DC_MAXSTR );
      destBlock = dmBlockCreateGenCopy( info->ds,blname, block, NULL, dmTRUE, info->copy_subspace );	 
      info->copy_subspace =1;

      if ( !destBlock )
      {
        dmGetErrorMessage( errbuf, DC_MAXSTR );
        dmaErrExit(errbuf);
      }
    }
  }

  return destBlock;
}

