/*                                                                
**  Copyright (C) 1998-1999,2003,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/*****************************************************************/
/*                                                               */
/*   dmtcopy_ui.c                                                */
/*                                                               */
/*   Contains functions for getting command line parameters      */
/*                                                               */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include <stdio.h>
#include <readline/readline.h>
#include "dmnutils.h"
#include "dmtlib.h"
#include "dmtcommon.h"
#include "dmtcopy_ui.h"



/***************************************************************/
/*                                                             */
/* Function:   dmGetdmcopyInput                                */
/*                                                             */
/* Description:                                                */
/*   separates input line arguments                            */
/*                                                             */
/*   Called from main()                                        */
/*                                                             */
/*  Input parameters:                                          */
/*    int     argc                                             */
/*    char  **argv                                             */
/*    int     tool                                             */
/*                                                             */
/*  Output parameters:                                         */
/*    char   *inName                                           */
/*    char   *outName                                          */
/*    char   *inOptions                                        */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*    int  status                                              */
/*                                                             */
/***************************************************************/
int dmGetdmcopyInput( int argc, char **argv,  dmtools_params *params)
{
  int   ii        = 0;
  int   optLength = 0;
  int   inLength  = 0;
  int   status    = 0;
  int   ArgCount  = argc;
  int  *debug     = &params->verbose;

  char  prompt[DC_MAXSTR];
  char  options[DC_MAXSTR];
  char  temp_in[DC_MAXSTR];

  char *inName    = params->infile;
  char *inOptions = params->option;
  char *outName   = params->outfile;
  char *kernel    = params->kernel;
  char *optString = NULL;
  char *rdline    = NULL;
  char *tag       = NULL;
  char *prompt1   = "Input dataset/block specification: ";
  char *prompt2   = "Output dataset name: ";

  dmBool  got_inName   = dmFALSE;
  dmBool  got_outName  = dmFALSE;
  dmBool  got_option   = dmFALSE;
  dmBool  flag_inName  = dmFALSE;
  dmBool  flag_outName = dmFALSE;
  dmBool  flag_option  = dmFALSE;
  dmBool *bclobber     = &params->clobber;

  options[0]  = '\0';
  temp_in[0]  = '\0';
  optLength   = strlen(inOptions);


  // Check argument count in command line. If count is
  // less than 3, then information is missing and it will
  // have to entered through keyboard

#if 0 
  if (ArgCount < 3)
  {
    // set up to request keyboard input

    if ( dev_tty == NULL )
    {
      dev_tty = fopen("/dev/tty", "r" );
      rl_instream = dev_tty;
      rl_outstream = stderr;
    }
  }
#endif


  // Get input file name from keyboard input
  switch ( ArgCount )
  {
    case 1:
    {
      // Only the tool name entered in command
      // Need to prompt for input and output

      strcpy(prompt, prompt1);
      rdline = readline(prompt);

      if (rdline)
      {
        strncpy(inName, rdline, DC_MAXSTR);
        free (rdline);
      }

      strcpy(prompt, prompt2);
      rdline = readline(prompt);

      if (rdline)
      {
        strncpy(outName, rdline, DC_MAXSTR);
        free (rdline);
      }

      break;
    }

    case 2:
    case 3:
    case 4:
    default:
    {
      // tool name and 1 or more entries on command line
      // DEFAULT ASSUMTION: <tool> <inName> <outName> <options>
      //  read first param

      for (ii  = 1; ii < argc; ii++ )
      {
        if ( argv[ii] )
          strncpy(temp_in, argv[ii], DC_MAXSTR );

        if ( (optString = strstr(argv[ii], "=")) )
        {
          cs_upper(temp_in);

          // designated parameter - which one?
          getDesignatedParameter(temp_in, &flag_inName, &flag_outName, &flag_option);
	}

        if (optString && (flag_inName || flag_outName || flag_option) )
	{
          if ( flag_inName )
	  {
            strncpy(inName, optString+1, DC_MAXSTR);
            got_inName = dmTRUE;
            flag_inName = dmFALSE;
          }
          else
	  {
	    if (flag_outName )
	    {
              strncpy(outName, optString+1, DC_MAXSTR);
              got_outName = dmTRUE;
              flag_outName = dmFALSE;
            }
            else
	    {
              // assume option
              optString = strstr(temp_in, "OPT=");

              if (optString)
                inLength = strlen(optString+4);
              else
                inLength = strlen(temp_in);

	      if (inLength > 0 && ((inLength + optLength + 1) < DC_MAXSTR))
              {
                if (!optString)
                  strcat(options, temp_in);
                else
                  strcat(options, optString+4 );

                if (optLength > 0)
                  strcat(inOptions, ",");

                strcat(inOptions, options);

                optLength  = strlen(inOptions);
                options[0] = '\0';
              }
              got_option = dmTRUE;
              flag_option = dmFALSE;
            }
	  }
        }
        else
        {
          int param_switch = ii;

          if ( !got_inName)
	  {
            // first non-designated param is input file name
            param_switch = 1;
          }
          else
	  {
            // got input - next non-designated param is
            // either outfile or option

            if ( !got_outName )
              param_switch = 2;
            else
              param_switch = 3;  // default case too
          }

          if ( (optString = strstr(temp_in, "CLOB")) && 
               ((tag = strchr(temp_in, '+')) ||
                (tag = strchr(temp_in, '-'))) )
	    param_switch = 3;

          switch (param_switch)
	  {
	    case 1:
            // assume this is input file name
            strncpy(inName, argv[ii], DC_MAXSTR );
            got_inName = dmTRUE;
            break;

	    case 2:
            // assume this is output file name
            strncpy(outName, argv[ii], DC_MAXSTR );
            got_outName = dmTRUE;
            break;

	    case 3:
	    default:
	    {
              // assume this is option name
              inLength = strlen(argv[ii]);

	      if (inLength > 0 && ((inLength + optLength + 1) < DC_MAXSTR))
              {
                strcat(options, argv[ii] );

                if (optLength > 0)
                  strcat(inOptions, ",");

                strcat(inOptions, options);

                optLength  = strlen(inOptions);
                options[0] = '\0';
              }

              got_option = dmTRUE;
              break;
            }
          }// end switch (param_switch)
        }
      } // end for loop

      // All parameters read - prompt for any missing item
      if (!got_inName)
      {
        strcpy(prompt, prompt1);

        rdline = readline(prompt);

        if (rdline)
        {
          strncpy(inName, rdline, DC_MAXSTR);
          free (rdline);
        }
      }

      if (!got_outName)
      {
        strcpy(prompt, prompt2);

        rdline = readline(prompt);

        if (rdline)
        {
          strncpy(outName, rdline, DC_MAXSTR);
          free (rdline);
        }
      }

      break;
    } // end case 4/default

  } // end switch (ArgCount)


  cs_upper(inOptions);

  // Verify option information
  dmcopyCheckOptions(inOptions, kernel, bclobber, debug);

  // Verify input information
  status = dmValidatedmcopyInfo(inName, outName, inOptions);

  return status;
}


/***************************************************************/
/*                                                             */
/* Function:   dmValidatedmcopyInfo                            */
/*                                                             */
/* Description:                                                */
/*   validates information obtained from command line call     */
/*                                                             */
/*   Called from main()                                        */
/*                                                             */
/*  Input parameters:                                          */
/*    char   *inName                                           */
/*    char   *outName                                          */
/*    char   *inOptions                                        */
/*                                                             */
/*  Output parameters:                                         */
/*          none                                               */
/*                                                             */
/*  Returns:                                                   */
/*    int  status                                              */
/*                                                             */
/***************************************************************/
int dmValidatedmcopyInfo( char *inName, char *outName, char *inOptions)
{
  int status = 1;


  if ( cs_is_blank ( inName ))
    status = 0;
  else
  {
    // insure that it is single file name, not a list
    //  remove any leading or trailing blanks

    cs_trim(inName);

    if ( inName[0] == '@' )
      status = 0;
  }

  if ( cs_is_blank ( outName ))
    status = 0;
#if 0
  else
  {
    // insure that it is single file name, not a list
    //  remove any leading or trailing blanks

    cs_trim(outName);

    if ( outName[0] == '@' )
      status = 0;
  }
#endif

  return status;
}



/***************************************************************/
/*                                                             */
/* Function:   dmcopyCheckOptions                              */
/*                                                             */
/* Description:                                                */
/*   checks options and breaks aparts components               */
/*                                                             */
/*   Called from main()                                        */
/*                                                             */
/*  Input parameters:                                          */
/*    char   *inOptions                                        */
/*    int     tool                                             */
/*                                                             */
/*  Output parameters:                                         */
/*    char   *kernel                                           */
/*    dmBool *bclobber                                         */
/*    int    *debug                                            */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*         none                                                */
/*                                                             */
/***************************************************************/
void  dmcopyCheckOptions(char *inOptions, char *kernel,
                         dmBool *bclobber, int *debug)
{
  char *tag = NULL;
  int   inOptLen = 0;


  // Check for clobber
  tag = strstr(inOptions, "CLOB");

  if (tag)
  {
    checkClobberOption(tag, bclobber);
  }

  tag = strstr(inOptions, "KERNEL");

  if ( tag )
  {
    checkKernelOption(tag, kernel);
  }


   tag = strstr(inOptions, "VERBOSE");

   if (!tag)
     tag = strstr(inOptions, "VER");

  if ( tag )
  {
    checkVerboseOption(tag, debug);
  }


   // remove any excess blanks
   cs_trim(inOptions);
 
   inOptLen = strlen(inOptions);

   // remove any trailing comma
   if (inOptions[--inOptLen] == ',')
     inOptions[inOptLen] = '\0';   

}


/***************************************************************/
/*                                                             */
/* Function:   checkClobberOption                              */
/*                                                             */
/* Description:                                                */
/*   checks for clobber true                                   */
/*                                                             */
/*   Called from dmCheckCopyOptions()                          */
/*                                                             */
/*  Input parameters:                                          */
/*    char   *option                                           */
/*                                                             */
/*  Output parameters:                                         */
/*    dmBool *bclobber                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*         none                                                */
/*                                                             */
/***************************************************************/
void checkClobberOption(char *option, dmBool *bclobber)
{
  char *tag = NULL;
  char *comma = NULL;
  char  cBlank = ' ';

  int   ii  = 0;
  int   paramLen = strlen(option);
 

  if ( (tag = strstr(option, "CLOB+")) ||
       (tag = strstr(option, "CLOB=1")) ||
       (tag = strstr(option, "CLOBBER=1")) ||
       (tag = strstr(option, "CLOB=Y")) ||
       (tag = strstr(option, "CLOBBER=Y")) )
    *bclobber = dmTRUE;
  else
    *bclobber = dmFALSE;

  comma = strchr(option, ',');

  if ( comma )
    paramLen = comma - option +1;

  // remove from options string
  for ( ii = 0; ii < paramLen; ii++)
  {
    *(option+ii) = cBlank;
  }

}


/***************************************************************/
/*                                                             */
/* Function:   checkKernelOption                               */
/*                                                             */
/* Description:                                                */
/*   checks for kernel specification                           */
/*                                                             */
/*   Called from dmCheckCopyOptions()                          */
/*                                                             */
/*  Input parameters:                                          */
/*    char   *option                                           */
/*                                                             */
/*  Output parameters:                                         */
/*    char   *kernel                                           */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*         none                                                */
/*                                                             */
/***************************************************************/
void checkKernelOption(char *option, char *kernel)
{
  char *tag = NULL;
  char *comma = NULL;
  char  cBlank = ' ';
  int   paramLen = strlen(option);
  int ii = 0;
 
  if ( (tag = strstr(option, "KER=")) ||
       (tag = strstr(option, "KERNEL=")) )
  {
    tag = strstr(option, "=");

    comma = strstr(option, ",");

    if (comma)
      paramLen = comma - (tag+1);

    strncpy(kernel, (tag+1), paramLen);

    kernel[paramLen] = '\0';

    if ( comma )
      paramLen = (comma - option) + 1;

    // remove from options string
    for ( ii = 0; ii < paramLen; ii++)
    {
      *(option+ii) = cBlank;
    }
  }
}
