/*                                                                
**  Copyright (C) 1999-2000,2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/*****************************************************************/
/*                                                               */
/*   dmlist_blocks.c                                             */
/*                                                               */
/*   Functions that are used by dmlist and were formerlly in     */
/*   dmtlib                                                      */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmtlib.h"
#include "dmtcommon.h"
#include "dmlist.h"


/***************************************************************/
/*                                                             */
/* Function:   dmlistDatasetListBlocks                         */
/*                                                             */
/* Description:                                                */
/*   List all the blocks in the dataset                        */
/*                                                             */
/*   Called from function dmlistBlock()                        */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDataset *ds                                       */
/*         dmBlock   *main_block                               */
/*         long       main_block_no                            */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           int EXIT_SUCCESS                                  */
/*                                                             */
/***************************************************************/
int dmlistDatasetListBlocks( dmDataset* ds, dmBlock* main_block, long main_block_no )
{
  dmBlock *block = NULL;

  char name[DC_MAXSTR];
  char buf[DC_MAXSTR];

  long blockno = 0;
  long nblocks = 0;
  long nrows   = 0;
  int  opened  = 0;

  dmDatasetGetName( ds, name, DC_MAXSTR ); 
  sprintf( buf, "Dataset: %s", name );
  dmlistBanner( buf );

  nblocks = dmDatasetGetNoBlocks( ds );
         
  if ( nblocks == 0 )
  { 
    dmlist_print( "Empty dataset" );
  }
  else
  {
    sprintf( buf, "%10s %-29s %-12s %s", "Block", "Name", "Type", "Dimensions" );
    dmlist_print( buf );
    dmlistDashline();

    for ( blockno = 1; blockno <= nblocks; blockno++ )
    {
      if ( main_block && blockno == main_block_no ) 
      {
        block = main_block;
        opened = 0;
      }
      else
      {
        opened = 1;
        block = dmDatasetMoveToBlock( ds, blockno );
      }

      dmlistPrintBlockInfo( block, blockno, buf, &nrows );
      dmlist_print( buf );

      if ( opened )
        dmBlockClose( block );		
    }
  }

  return EXIT_SUCCESS;
}




/***************************************************************/
/*                                                             */
/* Function:   dmlistPrintBlockInfo                            */
/*                                                             */
/* Description:                                                */
/*   Return block information, type if image, rows and columns */
/*   if a table                                                */
/*                                                             */
/*   Called from function dmlistBlock()                        */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmBlock  *block                                     */
/*         long      blockno                                   */
/*         char     *buf                                       */
/*         long     *nrows                                     */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           int EXIT_SUCCESS                                  */
/*                                                             */
/***************************************************************/
int dmlistPrintBlockInfo( dmBlock* block, long blockno, char* buf, long* nrows )
{
  dmDescriptor *imageData = NULL;

  char name[DCSIZE+1];
  char typename[DCSIZE+1];
  char auxbuf[DCSIZE+1];

  char *tmpTypename = NULL;
  long  naxes = 0;
  long  ncols = 0;


  dmBlockGetName( block, name, DCSIZE );
  tmpTypename = dmBlockGetTypeStr( block );
  strcpy( typename, tmpTypename );
  free( tmpTypename );
  strcpy( auxbuf, " " );

  if ( cs_eq( typename, "Image" ) )
  {
    imageData = dmImageGetDataDescriptor( block );
    naxes = dmGetArrayDim( imageData );

    if ( naxes == 0 )
    {
      strcpy( typename, "Null" );
      strcpy( auxbuf, " " );
    }
    else
    {
      dmlistFormatType( imageData, auxbuf, DCSIZE );
    }

    *nrows = 1;
  }
  else
    if ( cs_eq( typename, "Table" ) )
    {
      ncols = dmTableGetNoCols( block );
      *nrows = dmTableGetNoRows( block );

      sprintf( auxbuf, "%4ld cols x %-8ld rows", ncols, *nrows );
    }
    else
    {
      strcpy( auxbuf, " " );
    }

  sprintf( buf, "Block %4ld: %-30s %-10s %s", blockno, name, typename, auxbuf );

  return EXIT_SUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   dmlistPrintCounts                               */
/*                                                             */
/* Description:                                                */
/*   Return the number of rows in a table                      */
/*                                                             */
/*   Called from function dmlistBlock()                        */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmBlock  *block                                     */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           int EXIT_SUCCESS                                  */
/*                                                             */
/***************************************************************/
int dmlistPrintCounts( dmBlock* block )
{
  char  name[DCSIZE+1];
  char  typename[DCSIZE+1];
  char  auxbuf[DCSIZE+1];

  char *tmpTypename = NULL;
  long  nrows       = 0;


  dmBlockGetName( block, name, DCSIZE );
  tmpTypename = dmBlockGetTypeStr( block );
  strcpy( typename, tmpTypename );
  free( tmpTypename );
  strcpy( auxbuf, " " );

  if ( cs_eq( typename, "Image" ) )
  {
    sprintf( auxbuf, "Not supported for images" );
  }
  else
    if ( cs_eq( typename, "Table" ) )
    {
      nrows = dmTableGetNoRows( block );
     sprintf( auxbuf, "%-8ld", nrows );
    }
    else
    {
      strcpy( auxbuf, " " );
    }

  dmlist_print( auxbuf );

  return EXIT_SUCCESS;
}





