/*                                                                
**  Copyright (C) 1999-2001,2003-2004,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/*****************************************************************/
/*                                                               */
/*   dmlist_col.c                                                */
/*                                                               */
/*   Functions used by dmlist to list column information. These  */
/*   functions were formerly in dmtlib.                          */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include <float.h>
#include "dmtlib.h"
#include "dmtcommon.h"
#include "dmlist.h"

#define PHYS  1
#define WORLD 0


static int  dmlistPrintColInfo( dmDescriptor* dd, char* buf, char* ptr );

int   dmlistGetNullString( dmDescriptor* dd, char* range );
int   dmlistPrintAxes( dmDescriptor* dd, long mode );
int   dmlistPrintTransform( dmDescriptor* dd, char* parent, char** cpts, char* buf, char* tbuf, long maxlen );
int   dmlistPrintCol( dmDescriptor* dd, long colno, int h, char* buf, long maxlen );
void  pair_paren_format( char* cpt1, char* cpt2, char* fmt );
void  pair_paren_print( char** pptr1, char** pptr2, char* cpt1, char* cpt2 );
void  pair_top_print( char** pptr1, char** pptr2, char* buf );
void  put_blanks( char** pptr, long n );


/***************************************************************/
/*                                                             */
/* Function:   dmlistBlockListCols                             */
/*                                                             */
/* Description:                                                */
/*   List columns                                              */
/*                                                             */
/*   Called from function dmListBlock()                        */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmBlock  *block                                     */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           int EXIT_SUCCESS                                  */
/*                                                             */
/***************************************************************/
int dmlistBlockListCols( dmBlock* block ) 
{
  dmDescriptor **cols = NULL;

  dmBool image = dmFALSE;
  dmBool start = dmTRUE;

  char  name[DCSIZE+1];
  char  typename[DCSIZE+1];
  char  buf[DC_MAXSTR];
  char  tbuf[20];
  char *ptr = NULL;
  char *tmp = NULL;

  long ncols   = 0;
  long colno   = 0;
  long ncoords = 0;
  long ngroups = 0;
  long mode    = 0;

  int have_nulls = 0;
  int ok         = 0;


  dmBlockGetName( block, name, DCSIZE );
  tmp = dmBlockGetTypeStr( block );
  strncpy( typename, tmp, DCSIZE );
  free( tmp );
  image = !strcmp( typename, "Image" );

  sprintf( buf, "Columns for %s Block %s", typename, name );
  dmlistBanner( buf );

  cols = dmTableOpenColumnList( block, &ncols );

  for ( colno = 1; colno <= ncols; colno++ )
  {
    ok = dmlistGetNullString( cols[colno-1], tbuf );
    if (ok )
      have_nulls = 1;
  }

  if ( have_nulls )
    dmlist_print( "ColNo  Name                 Unit        Type             Range            Null" );
  else
    dmlist_print( "ColNo  Name                 Unit        Type             Range" );


  for ( colno = 1; colno <= ncols; colno++ )
  {
    dmlistPrintCol( cols[colno-1], colno, have_nulls, buf, DC_MAXSTR );
    dmlist_print( buf );
  }

  if ( !image )
  {
    for ( colno = 1; colno <= ncols; colno++ )
    {
      ncoords = dmDescriptorGetNoCoords ( cols[colno-1] );
      if ( ncoords > 0 )
      {
        if ( start )
        {
          sprintf( buf, "World Coord Transforms for Columns in %s Block %s", typename, name );
          start = dmFALSE;
          dmlistBanner( buf );
          sprintf( buf, "ColNo    Name" );
          dmlist_print( buf );
        }
        sprintf( tbuf, "%ld: ", colno );
        ptr = buf;
        ptr += sprintf( ptr, "%-6s", tbuf );
        dmlistPrintColInfo( cols[colno-1], buf, ptr );
      }
    }
  }

  if (image)
  { 
    double crpix = 0.0;
    double crval = 0.0;
    double cdelt = 0.0;

    dmCoordGetTransform_d( cols[0], &crpix, &crval, &cdelt, 1 );

    if ( crpix != 0.0 || crval != 0.0 || cdelt != 1.0 )
    {
      sprintf( buf, "Image Rescaling for %s Block %s", typename, name );
      dmlistBanner( buf );
      sprintf( buf, "Scaled Data = %g + %g * Raw Data", crval, cdelt );
      dmlist_print( buf );
    }
  }

 
  sprintf( buf, "Physical Axis Transforms for %s Block %s", typename, name );
  start = dmTRUE;
  mode = PHYS;

  for ( colno = 1; colno <= ncols; colno++ )
  {
    ngroups = dmArrayGetNoAxisGroups( cols[colno-1] );

    if ( start && ngroups > 0 )
    {
       dmlistBanner( buf );
       start = dmFALSE;
       sprintf( buf, "Group# Axis# " );
       dmlist_print( buf );
    }

    dmlistPrintAxes( cols[colno-1], mode );
  }


  sprintf( buf, "World Coordinate Axis Transforms for %s Block %s", typename, name );
  start = dmTRUE;
  mode = WORLD;

  for ( colno = 1; colno <= ncols; colno++ )
  {
    ngroups = dmArrayGetNoAxisGroups( cols[colno-1] );

    if ( start && ngroups > 0 )
    {
      dmlistBanner( buf );
      start = dmFALSE;
      sprintf( buf, "Group# Axis# " );
      dmlist_print( buf );
    }

    dmlistPrintAxes( cols[colno-1], mode );
  }

  free( cols );

  return EXIT_SUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmlistBlockListRawCols                          */
/*                                                             */
/* Description:                                                */
/*   Test raw column list                                      */
/*                                                             */
/*   Called from function dmListBlock()                        */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmBlock  *block                                     */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           int dmSUCCESS                                     */
/*                                                             */
/***************************************************************/
int dmlistBlockListRawCols( dmBlock* block )
{
  dmDescriptor **list = NULL;

  char buf[DC_MAXSTR];
  char name[DCSIZE+1];

  long *cols  = 0;
  long *dims  = 0;
  long  ncols = 0;
  long  i     = 0;

  dmTableComponentList( block, &list, &dims, &cols, &ncols );
 
  if ( ncols > 0 )
  {
    strcpy( buf, "RawNo DMNo Dim  Name" );
    dmlist_print( buf );

    for ( i = 0; i < ncols; i++ )
    {
     dmGetName( list[i], name, DCSIZE );
     sprintf( buf, "%4ld  %4ld  %4ld  %s", i, cols[i], dims[i], name );
     dmlist_print( buf );
    }

    free( list );
    free( dims );
    free( cols );
  }

  return dmSUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmlistFormatType                                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*   Called from functions dmlistPrintCol()                    */
/*                         dmlistPrintBlockInfo()              */
/*                         dmlistPrintHeader()                 */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor *dd                                    */
/*         char         *buf                                   */
/*         long          maxlen                                */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void dmlistFormatType( dmDescriptor* dd, char* buf, long maxlen )
{
  dmDescriptorType t;

  char tmp[20];

  long  len     = 0;
  long  desclen = 0;
  long  naxes   = 0;
  long *axes    = NULL;

  t = dmDescriptorGetType( dd );
  dmaGetDataTypeName( dd, buf, maxlen );
  desclen = dmDescriptorGetLength( dd );

  len = strlen( buf );

  if ( desclen > 0 && t != dmKEY ) 
  {
    sprintf( tmp, "[%ld]", desclen );  
    strncpy( buf + len, tmp, maxlen - len );
  }

  naxes = dmGetArrayDim( dd );

  if ( naxes > 0 )
  {
    naxes = dmGetArrayDimensions( dd, &axes );
    dmlistFormatAxes( axes, naxes, buf + len, maxlen - len );
    free( axes );
  }
}



/***************************************************************/
/*                                                             */
/* Function:   dmlistGetNullString                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*   Called from functions dmlistPrintData()                   */
/*                         dmlistPrintArrayData()              */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor *dd                                    */
/*         char         *range                                 */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           int true or false                                 */
/*                                                             */
/***************************************************************/
int dmlistGetNullString( dmDescriptor* dd, char* range )
{
  char dataType[DCSIZE+1];
  char cmin[DCSIZE+1];
//  char cmax[DCSIZE+1];
  char dt = '\0';

  double dmin  = 0.0;
  long   idmin = 0;
  unsigned long umin = 0;
  int    ok = 0;


  strcpy( range, " " );
  dmaGetDataTypeName( dd, dataType, DCSIZE );
  dt = *dataType;  /* The first char of dataType gives the general type */

  strcpy( range, "-" );

  switch( dt )
  {
    case 'I':
      if ( dmDescriptorGetNull_l( dd, &idmin ) && idmin != 0 )
      {
        ok = 1;
        sprintf( range, "%ld", idmin );
      }
      break;

    case 'U':
      if ( dmDescriptorGetNull_ul( dd, &umin ) && umin != 0 )
      {
        ok = 1;
        sprintf( range, "%lu", umin );
      }
      break;
    case 'R':
      if ( dmDescriptorGetNull_d( dd, &dmin ) && dmin != 0.0 )
      {
        ok = 1;
        Print_d( dmin, cmin );
        sprintf( range, "%s", cmin );
      }
      break;
    default:
      strcpy( range, " " );
  }

  return ok;
}



/***************************************************************/
/*                                                             */
/* Function:   dmlistPrintAxes                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*   Called from function  dmlistBlockListCols()               */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor *dd                                    */
/*         long          mode                                  */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           int EXIT_SUCCESS                                  */
/*                                                             */
/***************************************************************/
int dmlistPrintAxes( dmDescriptor* dd, long mode ) 
{
  dmDescriptor *axis_dd = NULL;

  char  name[DCSIZE+1];
  char  axbuf[DCSIZE+1];
  char  tmpBuf[DC_MAXSTR];
  char  cpt1[DCSIZE+1];
  char  cpt2[DCSIZE+1];
  char *cpts[2]; 
  char *cname = NULL;
  char *ptr   = NULL;

  long ngroups = 0;
  long dim     = 0;
  long i       = 0;
  long axno    = 0;


  cpts[0] = cpt1;
  cpts[1] = cpt2;

  ngroups = dmArrayGetNoAxisGroups( dd );

  if ( ngroups > 0 )
  {
    for ( i = 1; i <= ngroups; i++ )
    {
      axis_dd= dmArrayGetAxisGroup( dd, i );
      dim = dmGetElementDim( axis_dd );

      if ( dim == 2 )
      {
        sprintf( cpt1, "#%ld", axno+1 );
        sprintf( cpt2, "#%ld", axno+2 );
        sprintf( name, "(%s,%s)", cpt1, cpt2 );
        cname = NULL;
        sprintf( axbuf, "%ld,%ld", axno+1, axno+2 );
      }
      else 
      {
        sprintf( axbuf, "%ld", axno+1 );
        sprintf( name, "#%s", axbuf );
        cname = name;
      }

      axno += dim;

      if ( mode == PHYS )
      {
        ptr = tmpBuf;
        ptr+= sprintf( ptr, "%4ld   %-6s ", i, axbuf );
        dmlistPrintTransform( axis_dd, cname, cpts, tmpBuf, ptr, DC_MAXSTR );
      }
      else
      {
        ptr = tmpBuf;
        ptr+= sprintf( ptr, "%4ld   %-6s ", i, axbuf );
        dmlistPrintColInfo( axis_dd, tmpBuf, ptr );
      }
    }
  }

  return  EXIT_SUCCESS;
}


/***************************************************************/
/*                                                             */
/* Function:   dmlistPrintCol                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*   Called from functions dmlistBlockListCols()               */
/*                         dmlistPrintAxes()                   */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor *dd                                    */
/*         long          colno                                 */
/*         int           do_null                               */
/*         char         *buf                                   */
/*         long          maxlen                                */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           int EXIT_SUCCESS                                  */
/*                                                             */
/***************************************************************/
int dmlistPrintCol( dmDescriptor* dd, long colno, int do_null, char* buf, long maxlen ) 
{
  char  name[DCSIZE+1];
  char  range[DCSIZE+1];
  char  desc[DCSIZE+1];
  char  aux[DCSIZE+1];
  char  unit[DCSIZE+1];
  char  tmp[2*DCSIZE+1];
  char  null[DCSIZE+1];
  char *rptr = NULL;

  strcpy( desc,"" );
  dmlistGetDescCompositeName( dd, name, DCSIZE );
  dmGetDesc( dd, desc, DCSIZE );
  dmlistGetRange( dd, range );

  rptr = range;

  while( *rptr == ' ' )
   ++rptr;

  dmGetUnit( dd, unit, DCSIZE );
  dmlistFormatType( dd, aux, DCSIZE ); 

  if ( do_null )
  {
    dmlistGetNullString( dd, null );
    sprintf( tmp, "%4ld   %-20s %-12s %-14s %-20s %-10s %s", colno, name, unit, aux, rptr, null, desc );
  }
  else
  {
    sprintf( tmp, "%4ld   %-20s %-12s %-14s %-20s %s", colno, name, unit, aux, rptr, desc );
  }

  strncpy( buf, tmp, maxlen );

  return  EXIT_SUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmlistPrintColInfo                              */
/*                                                             */
/* Description:                                                */
/*   Print the column info and then the coordinate info        */
/*                                                             */
/*   Called from functions dmlistBlockListCols()               */
/*                         dmlistPrintAxes()                   */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor *dd                                    */
/*         char         *tmpBuf                                */
/*         char         *ptr                                   */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           int EXIT_SUCCESS                                  */
/*                                                             */
/***************************************************************/
static int dmlistPrintColInfo( dmDescriptor* dd, char* tmpBuf, char* ptr )
{
  dmDescriptor *coord = NULL;

  char  name[DCSIZE+1];
  char  cpt1[DCSIZE+1] = "";
  char  cpt2[DCSIZE+1] = "";
  char *cpts[2];

  long ncoords = 0;
  long i       = 0;
  long axno    = 0;
  long dim     = 0;

  cpts[0] = cpt1;
  cpts[1] = cpt2;
  dmGetName( dd, name, DCSIZE );
  dim = dmGetElementDim( dd );

  if ( dim == 2 )
  {
    dmGetCptName( dd, 1, cpt1, DCSIZE );
    dmGetCptName( dd, 2, cpt2, DCSIZE );
  }

  ncoords = dmDescriptorGetNoCoords ( dd );

  for ( i = 1; i <= ncoords; i++ )
  {
    coord = dmDescriptorGetCoordNo( dd, i );
    axno  = dmCoordGetAxisGroupNo( coord );

    if ( axno == 0 )
      dmlistPrintTransform( coord, name, cpts, tmpBuf, ptr, DC_MAXSTR );
  }

  return EXIT_SUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmlistPrintTransform                            */
/*                                                             */
/* Description:                                                */
/*   Print the column transform information                    */
/*                                                             */
/*   Called from functions dmlistPrintColInfo()                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor *dd                                    */
/*         char         *parent                                */
/*         char        **cpts                                  */
/*         char         *buf                                   */
/*         char         *tbuf                                  */
/*         long          maxlen                                */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           int EXIT_SUCCESS                                  */
/*                                                             */
/***************************************************************/
int dmlistPrintTransform( dmDescriptor* dd, char* parent, char** cpts,
                          char* buf, char* tbuf, long maxlen )
{
  dmBool linear = dmTRUE;
  dmBool unity = dmFALSE;

  char  s_crpix[DCSIZE];
  char  s_crval[DCSIZE];
  char  s_cdelt[DCSIZE];
  char  t_crpix[DCSIZE];
  char  t_crval[DCSIZE];
  char  t_cdelt[DCSIZE];
  char  name[DCSIZE];
  char  tmp[DCSIZE];
  char  cpt1[DCSIZE];
  char  cpt2[DCSIZE];
  char  buf2[DC_MAXSTR];
  char  unit[DCSIZE];
  char  tunit[DCSIZE];
  char  ttype[DCSIZE];
//char  fmt[DCSIZE];
  char  thead[DCSIZE] = ""; 
  char  ttail[2] = "";
  char  spacer[2];
  char *ptr1 = NULL;
  char *ptr2 = NULL;

  double *params         = NULL;
  double  crpix[2]       = { 0.0, 0.0 };
  double  crval[2]       = { 0.0, 0.0 };
  double  cdelt[2]       = { 1.0, 1.0 };
  double  minus_crpix[2] = { 0.0, 0.0 };
  double  theta          = 0.0;

  long i       = 0;
  long sg      = 1;
  long nparams = 0;
  long dim     = 0;
//long sizes[10];
//long size1, size2;
//long offset  = 0;

  int  mode = 1;


  dmGetName( dd, name, DCSIZE );

  dim = dmGetElementDim( dd );
  dmCoordGetTransformType( dd, ttype, &nparams, DCSIZE );
  dmCoordGetTransform_d( dd, crpix, crval, cdelt, dim );
  params = dmCoordGetParams( dd, &nparams );

  if ( params )
    theta = params[0];

  free( params );
  dmGetUnit( dd, unit, DCSIZE );

  if ( *unit && *unit != ' ' )
    sprintf( tunit, "[%s]", unit );
  else
    strcpy( tunit, "" );

  linear = cs_uceq( ttype, "LINEAR" );

  if ( linear )
  {
    unity = ( crval[0] == crpix[0] && cdelt[0] == 1.0 );

    if ( dim == 2 ) 
    {
      unity &= ( crval[1] == crpix[1] && cdelt[1] == 1.0 );
      strcpy( thead, "+" );
    }
  }
  else
  {
    sprintf( thead, "+%s[", ttype );
    strcpy( ttail, "]" );  
  }
 
  if ( dim > 1 )
  {
    sg = 1;
    strcpy( spacer, "-" );
  }
  else
  {
    sg = -1;  
    strcpy( spacer, "" );
  }

  for ( i = 0; i < dim; i++ )
    minus_crpix[i] = sg * crpix[i];

  if ( dim == 2 ) 
  {
    ptr1 = tbuf;
    ptr2 = buf2;
    dmGetCptName( dd, 1, cpt1, DCSIZE );
    dmGetCptName( dd, 2, cpt2, DCSIZE );
    /* Blank out start of second line */
    put_blanks( &ptr2, (long)(tbuf-buf) );
    pair_top_print( &ptr1, &ptr2, name );
    pair_paren_print( &ptr1, &ptr2, cpt1, cpt2 );

    if ( unity )
    {
      strcpy( tmp, " = " ); 

      if ( parent )
        strcat( tmp, parent );
      pair_top_print( &ptr1, &ptr2, tmp );
      pair_paren_print( &ptr1, &ptr2, cpts[0], cpts[1] );
      sprintf( ptr1, " %s", tunit );
    }
    else
    {
      PrintSigned_d( crval[0], s_crval, DCSIZE );
      PrintSigned_d( minus_crpix[0], s_crpix, DCSIZE );
      PrintSigned_d( cdelt[0], s_cdelt, DCSIZE );
      PrintSigned_d( crval[1], t_crval, DCSIZE );
      PrintSigned_d( minus_crpix[1], t_crpix, DCSIZE );
      PrintSigned_d( cdelt[1], t_cdelt, DCSIZE );
      pair_top_print( &ptr1, &ptr2, " = " );
      pair_paren_print( &ptr1, &ptr2, s_crval, t_crval );
      sprintf( tmp, "%s %s", tunit, thead );
      pair_top_print( &ptr1, &ptr2, tmp );
      pair_paren_print( &ptr1, &ptr2, s_cdelt, t_cdelt );
      pair_top_print( &ptr1, &ptr2, "* " );    

      if ( fabs(theta) > DBL_EPSILON )
      {
        pair_top_print( &ptr1, &ptr2, "ROT(" );
        PrintSigned_d( theta, tmp, DCSIZE );
        pair_top_print( &ptr1, &ptr2, tmp );
        pair_top_print( &ptr1, &ptr2, " deg)* " );
      }

      *ptr1++ = '(';
      *ptr2++ = '(';

      if ( parent )
        pair_top_print( &ptr1, &ptr2, parent );

      pair_paren_print( &ptr1, &ptr2, cpts[0], cpts[1] );
      pair_top_print( &ptr1, &ptr2, spacer );    
      pair_paren_print( &ptr1, &ptr2, s_crpix, t_crpix );
      *ptr1++ = ')';
      *ptr2++ = ')';
      pair_top_print( &ptr1, &ptr2, ttail );    
    }

    dmlist_print( buf );
    dmlist_print( buf2 );
  }
  else
  {
    PrintSignedVector_d( minus_crpix, dim, 1, mode, NULL, s_crpix, DCSIZE );
    PrintSignedVector_d( crval, dim, 1, mode, NULL, s_crval, DCSIZE );
    PrintSignedVector_d( cdelt, dim, 1, mode, NULL, s_cdelt, DCSIZE );

    if ( unity )
    {
      sprintf( tbuf, "%-20s = %s %s", name, parent, tunit );
    }
    else
    {
      sprintf( tbuf, "%-20s = %s %s %s%s * (%s %s %s)%s", name, 
               s_crval, tunit, thead, s_cdelt, parent, spacer, s_crpix, ttail );
    }
    dmlist_print( buf );
  }
  return EXIT_SUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   pair_paren_format                               */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*   Called from function pair_paren_print()                   */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         char  *cpt1                                         */
/*         char  *cpt2                                         */
/*                                                             */
/*  Output parameters:                                         */
/*         char  *fmt                                          */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void pair_paren_format( char* cpt1, char* cpt2, char* fmt )
{
  long  size1 = 0;
  long  size2 = 0;
  long  size  = 0;

  size1 = strlen(cpt1);
  size2 = strlen(cpt2);
  size  = size1 > size2 ? size1:size2;

  sprintf( fmt, "(%%-%lds)", size );
}



/***************************************************************/
/*                                                             */
/* Function:   pair_paren_print                                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*   Called from function dmlistPrintTransform()               */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         char  *cpt1                                         */
/*         char  *cpt22                                        */
/*                                                             */
/*  Output parameters:                                         */
/*         char **pptr1                                        */
/*         char **pptr2                                        */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void pair_paren_print( char** pptr1, char** pptr2, char* cpt1, char* cpt2 )
{
  char fmt[DCSIZE];

  pair_paren_format( cpt1, cpt2, fmt );

  *pptr1 += sprintf( *pptr1, fmt, cpt1 );
  *pptr2 += sprintf( *pptr2, fmt, cpt2 );   
}



/***************************************************************/
/*                                                             */
/* Function:   pair_top_print                                  */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*   Called from function dmlistPrintTransform()               */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         char  **pptr1                                       */
/*         char  **pptr2                                       */
/*         char   *buf                                         */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void pair_top_print( char** pptr1, char** pptr2, char* buf )
{
  *pptr1 += sprintf( *pptr1, "%s", buf );
  put_blanks( pptr2, strlen( buf ));
}


/***************************************************************/
/*                                                             */
/* Function:   put_blanks                                      */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*   Called from functions pair_print_top()                    */
/*                         dmlistPrintTransform()              */
/*                                                             */
/*  Input parameters:                                          */
/*         char  **pptr                                        */
/*         long    n                                           */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           void                                              */
/*                                                             */
/***************************************************************/
void put_blanks( char** pptr, long n )
{
  char* ptr = *pptr;

  while( n-- )
    *ptr++ = ' ';

  *ptr  = '\0';
  *pptr = ptr;
}









