/*                                                                
**  Copyright (C) 1999-2001,2003-2004,2006-2008  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/*****************************************************************/
/*                                                               */
/*   dmlist_data.c                                               */
/*                                                               */
/*   Functions used by dmlist to list column data. These         */
/*   functions were formerly in dmtlib.                          */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmtlib.h"
#include "cxcregion.h"
#include "dmnutils.h"
#include "dmtcommon.h"
#include "dmlist.h"

void **dmuAllocVoids( long n );
void   dmlistGetNull( dmDescriptor* dd, void** pnull );


/***************************************************************/
/*                                                             */
/* Function:   dmlistGetNull                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*     Called from functions dmlistPrintArrayData()            */
/*                           dmlistPrintData()                 */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor *dd                         */
/*                    void        **pnull                      */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmlistGetNull( dmDescriptor* dd, void** pnull )
{
  dmDataType dt;

  void *null = *pnull;
  int   ok   = 0;

  if ( !null )
    return;

  dt = dmGetDataType( dd );

  switch( dt )
  {
    case dmLONG:
    case dmSHORT:
    case dmBYTE:
      ok = dmDescriptorGetNull_l( dd, null );  
      break;
    case dmULONG:
    case dmUSHORT:
      ok = dmDescriptorGetNull_ul( dd,  null );
      break;
    case dmDOUBLE:
    case dmFLOAT:
      ok =dmDescriptorGetNull_d( dd, null );
      break;
    default:
      break;
  }

  if ( !ok )
    *pnull = NULL;

  return;
}



/***************************************************************/
/*                                                             */
/* Function:   dmlistGetValueString                            */
/*                                                             */
/* Description:                                                */
/*   Print the value of a cell.                                */
/*   Arrays and vector descriptors are handled.                */
/*   However, vectored arrays are not, and only the first part */
/*   of a large array will be printed.                         */
/*    Generic element types are not handled.                   */
/*                                                             */
/*   Called from function dmlistPrintArrayData()               */
/*                        dmlistPrintData()                    */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmDescriptor *dd                         */
/*                    long          verbose                    */
/*                    dmDataType    type                       */
/*                    long          mincell                    */
/*                    long          maxcell                    */
/*                    void         *null                       */
/*                    char         *value                      */
/*                    long          maxlen                     */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmlistGetValueString( dmDescriptor* dd, long verbose,
     dmDataType type, long mincell, long maxcell, void* null,
     char* value, long maxlen )
{
  /* Data buffers */
  char  null_mode[] = "   ";
  char  vec_mode[]  = "(,)";
  char  array_mode[]= "[ ]";

  char **valarray = NULL;
  char  *mode1    = null_mode;
  char  *mode2    = null_mode;

  double *darr = NULL;
  double  dval = 0.0;

  unsigned long *uarr = NULL;
  unsigned long  uval = 0;

  long  *iarr       = NULL;
  long   i          = 0;
  long   dim        = 0;
  long   asize      = 0;
  long   nvals      = 0;
  long   a_mincell  = 0;
  long   a_maxcell  = 0;
  long   lval       = 0;
  long   orig_asize = 0;
  long   slen       = 0;
  long   disp_asize = 0;
  long   vlen       = 0;

  int q      = 0;
  int scalar = 1;
  int fancy  = 0;

  dmBool *barr = NULL;
  dmBool  bIsVarLen = dmFALSE;
  dmBool  bImageData = dmFALSE;

  if ( verbose > 0 )
  {
    fancy = 1;
    mode1 = vec_mode;
    mode2 = array_mode;
  }

  strcpy( value, " " );

  bIsVarLen  = dmDescriptorIsVarLength( dd );
  bImageData = (dmDescriptorGetType( dd ) == dmIMAGEDATA) ? dmTRUE: dmFALSE; 
  orig_asize = dmGetCurrentArraySize( dd );  /* dd might be variable-length array */
  dim        = dmGetElementDim( dd );
  asize      = orig_asize;
  slen       = dmDescriptorGetLength( dd );
  disp_asize = ( type == dmBIT )? slen : asize;

  if ( orig_asize > 1 || dim > 1 || bImageData )
    scalar = 0;

  if ( bIsVarLen )
  {
    scalar = 0;

    /* Make sure brackets (if originally requested) are printed, even */
    /* if the current row has 0 or 1 elements                         */

    if ( fancy )
      ++fancy;
  }

  if ( mincell <= maxcell )
  {

    a_mincell = mincell;
    a_maxcell = maxcell;

    if ( mincell < 1 )
      a_mincell = 1;

    // If start cell number is beyond current cells in row
    //  just return because that data is not present

    if ( a_mincell > disp_asize )
      return;

    // Cannot display beyond last cell
    if ( maxcell > disp_asize )
      a_maxcell = disp_asize;

    asize = a_maxcell - a_mincell + 1;
  }
  else
  {
    // Display all values for column, scalar or array
    a_mincell = 1;
    a_maxcell = disp_asize;
  }

  nvals = dim * asize;

  /* Don't try to print empty images (variable-length array rows might */
  /* also be empty, but we want to print empty brackets for those)     */

  if ( bImageData && (nvals < 1)  )
    return;


  switch( type )
  {
    case dmDOUBLE: 
    case dmFLOAT:

    if ( !scalar )
    {
      darr = (double*)malloc( nvals * sizeof( double ));
      if ( !darr )
       return;
      dmaGetArray_d( dd, darr, a_mincell, a_maxcell );   
      PrintVector_d( darr, dim, asize, fancy, null, value, maxlen );
      free( darr );
    }
    else
    { 
      dval = dmGetScalar_d( dd );
      dmuPrint_d( dval, null, 0, value, 0 );
    }
    break;

    case dmUSHORT: 
    case dmULONG:
      if ( !scalar )
      {
        uarr = (unsigned long*)malloc( nvals * sizeof( unsigned long ));
        dmaGetArray_ul( dd, uarr, a_mincell, a_maxcell );   
        PrintVector_ul( uarr, dim, asize, fancy, null,  value, maxlen );
        free( uarr );
      }
      else
      { 
        uval = dmGetScalar_ul( dd );
        PrintArrayItem_ul( &uval, 0, null, value );
      }
      break;

    case dmBIT:
      dmaGetBit( dd, a_mincell, a_maxcell, value, maxlen );
      break;

    case dmBYTE:
    case dmSHORT: 
    case dmLONG:
      if ( !scalar )
      {
        iarr = (long*)malloc( nvals * sizeof( long ));
        dmaGetArray_l( dd, iarr, a_mincell, a_maxcell );   
        PrintVector_l( iarr, dim, asize, fancy, null, value, maxlen );
        free( iarr );
      }
      else
      { 
        lval = dmGetScalar_l( dd );
        PrintArrayItem_l( &lval, 0, null, value );
      }
      break;

    case dmTEXT:
      if ( !scalar ) 
      {
        valarray = (char**)malloc( nvals * sizeof( char* ) );

        for ( i = 0; i < nvals; i++ )
          valarray[i] = (char*)calloc( nvals * ( slen + 1 ), sizeof( char ) );

        dmaGetArray_c( dd, valarray, a_mincell, a_maxcell );

        // Code Change  Bug 11332  November 5, 2008
        //   For text-simple columns, the default string size (slen) 
        //   is 1025 chars. Unfortunately, for a vector, the print program
        //   decides that it will run out of buffer space and doesn't
        //   print the vector data. So, find the greatest length of the
        //   string for this vector and put that into the PrintVector_c
        //   parameter line.

        for ( i = 0; i < nvals; i++)
          if ( vlen < strlen(valarray[i]))
	    vlen = strlen(valarray[i]);
        
        PrintVector_c( valarray, dim, asize, fancy, vlen, value, maxlen );

        for ( i = 0; i < nvals; i++ )
        {
          free( valarray[i] );  
        }

        free( valarray );
      }
      else
      {
        dmGetScalar_c( dd, value, maxlen );
        value[maxlen-1] = '\0';
      }
      break;

    // Code Change Bug 11058 October 9, 2008
    //   This case previously returned only a scalar value,
    //   the first value in the field, wheter it was a scalar
    //   or an array. It now works to return a correct value
    //   or set of values depending on column type
 
    case dmBOOL:
      if ( !scalar ) 
      {
        // This array is used to get boolean array values
        barr = (dmBool *)malloc( nvals * sizeof( dmBool ));

        // Read boolean array values
        dmaGetArray_q( dd, barr, a_mincell, a_maxcell );

	// use PrintVector_c to output char array of TRUE and FALSE 
        PrintVector_q( barr, dim, asize, fancy, 6, value, maxlen );

        // Clean up
        free( barr );
      }
      else
      {
        // Returns scalar value
         q = (int)dmGetScalar_q( dd );

	 // Map boolean value to either T or F for output
         if ( q )
           strcpy( value, "TRUE " );
         else
           strcpy( value, "FALSE" );
      }
      // end code change

      break;

    default:
      strcpy( value, "" );
  }
}



/***************************************************************/
/*                                                             */
/* Function:   dmlistPrintArrayData                            */
/*                                                             */
/* Description:                                                */
/*   Print the data section as array data                      */
/*                                                             */
/*  Called from function dmListBlock()                         */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock  *block                          */
/*                    long      nrows                          */
/*                    long      minrow                         */
/*                    long      maxrow                         */
/*                    long      mincell                        */
/*                    long      maxcell                        */
/*                    long      verbose                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            int  status                                      */
/*                                                             */
/***************************************************************/
int dmlistPrintArrayData( dmBlock* block, long nrows, long minrow,
                          long maxrow, long mincell, long maxcell, 
                          dmBool bRawOpt, long verbose )
{
  dmDescriptor **cols        = NULL;
  dmDescriptor **varlen_cols = NULL;
  dmDescriptor  *dd          = NULL;
  dmDataType     type;
  dmDataType     dtype;
  regRegion     *region      = NULL;

  void  *null  = NULL;
  void **nulls = NULL;

  char  value[DC_MAXSTR+1];
  char  unit[DCSIZE+1];
  char  name[DCSIZE+1];
  char  typename[DCSIZE+1];
  char  buf[DC_MAXSTR+1];
  char  tbuf[DC_MAXSTR+1];

  char **fmt       = NULL;  /* Format for writing the name of each column with */
  char  *ptr       = NULL;
  char  *eptr      = NULL;
  char  *Alloc_buf = NULL;
  char  *start_ptr = NULL;
  char  *tmp       = NULL;

  long  lpix[20];
  long *size         = NULL;  /* Expected number of bytes for each column */
  long *axes         = NULL;
  const long spare   = 80;
  long  ncols        = 0;
  long  varlen_ncols = 0;
  long  colno        = 0;
  long  naxes        = 0;
  long  retrow       = 0;
  long  use_naxes    = 0;
  long  row_no       = 0;
  long  cell         = 0;
  long  sum          = 0;
  long  buflen       = 0;
  long  left         = 0;
  long  maxsize      = 1;
  long  basesize     = 0;
  long  row_maxsize  = 1;
  long  row_maxcell  = 1;
  long  asize        = 0;
  int   all_rows     = 0;


  strcpy( unit, " ");
  dmBlockGetName( block, name, DCSIZE );

  tmp = dmBlockGetTypeStr( block );
  strncpy( typename, tmp, DCSIZE );
  free( tmp );

  if ( verbose > 0 )
  {
    sprintf( buf, "Data for %s Block %s", typename, name );
    dmlistBanner( buf );
  }

  if ( dmaBlockIsNull( block ) )
    return EXIT_SUCCESS; /* Null image */


  cols  = dmTableOpenColumnList( block, &ncols );
  size  = (long*)malloc( ncols * sizeof( long ) );
  fmt   =  ar_alloc_c( 20, ncols );

  // Code Change Bug 8426 April 8, 2011
  // If option is for "RAW" data, do not
  // set up nulls array for value substitutions

  if (! bRawOpt )
    nulls = dmuAllocVoids( ncols );


  /* Code Change  Bug 10676  August 24, 2007            */ 
  /*  changed comparison functin from strcmp to cs_uceq */

  if ( ncols > 1 && (cs_uceq( name, "REGION") ) )
  {
    region = dmTableReadRegion( block );

    if ( region )
    {
      regComposeRegion( region, buf, DC_MAXSTR );
      sprintf( tbuf, "Region Block: %s\n", buf );
      dmlist_print( tbuf );
    }
  }

  /*  Check how much space we need */
  maxsize = 0;
  sum     = 20; /* Starting left buffer */

  for ( colno = 0; colno < ncols; colno++ ) 
  {
    dd = cols[colno];
    size[colno] = dmlistGetColSize( dd );   /* Assumes arraysize =1 */
    asize = ( dmGetDataType( dd ) == dmBIT ) ? dmDescriptorGetLength( dd ) : dmGetArraySize( dd );

    // Code Change Bug 8426  April 1, 2011
    // Don't use null substitutions if RAW flag being used

    if ( nulls )
      dmlistGetNull( dd, &nulls[colno] );

    if ( asize > maxsize )
      maxsize = asize;

    sum += size[colno] + 1;
    naxes = dmGetArrayDim( dd );

    if ( naxes > 1 && !axes ) 
    { 
      /* Pick the first 2D or greater image */
      use_naxes = dmGetArrayDimensions( dd, &axes );
    }

    if ( dmDescriptorIsVarLength( dd ) )
    {
      if ( !varlen_cols )
        varlen_cols = (dmDescriptor**)malloc( ( ncols - colno ) *
          sizeof( dmDescriptor* ) );

      varlen_cols[varlen_ncols++] = dd;
    }
    else
      if ( asize > basesize )
      {
        /* basesize is maximum size of fixed-length columns */
        basesize = asize;
      }
  }

  if ( mincell > maxsize )
   return EXIT_SUCCESS;

  if ( mincell ==1 && maxcell == 0 )
   maxcell = maxsize;

  if ( maxcell > maxsize )
   maxcell = maxsize;


  /*  Take special action if it's huge  */
  if ( sum < DC_MAXSTR )
  {
    buflen = DC_MAXSTR;
    start_ptr = buf;
  }
  else
  {
    buflen = sum + DC_MAXSTR;
    Alloc_buf = (char*)malloc( buflen+1 );
    start_ptr = Alloc_buf;
  }

  ptr  = start_ptr;
  eptr = ptr + buflen - spare;

  if ( verbose > 0 )
    ptr += sprintf( ptr, "%-6s %-6s", "ROW","CELL" );
  else
    ptr += sprintf( ptr, "# " );

  for ( colno = 0; colno < ncols; colno++ )
  {
    dd = cols[colno];
    sprintf( fmt[colno], " %%-%lds",  size[colno] );
    dmlistGetDescCompositeName( dd, name, DCSIZE );

    if ( ptr < eptr )
    {
      ptr += sprintf( ptr, fmt[colno], name );
    }

    /* Right justify numeric cols */
    dtype = dmGetDataType( dd );

    if ( dtype != dmTEXT )
      sprintf( fmt[colno], " %%%lds",  size[colno] );
  }

  dmlist_print( start_ptr );

  if ( verbose > 0 )
    dmlist_print( " " );

  all_rows =  ( maxrow <= 0 );
  row_no = minrow;
  retrow = row_no;

  // Code change - Bug 9189 - January 5, 2007
  //
  //  Need to know if this is a null block (retrow = dmNOMOREROWS)
  //  so always call dmTableSetRow() except if this is image block

  if ( dmBlockGetType(block) != dmIMAGE )
    retrow = dmTableSetRow( block, row_no );

  if ( nrows <= 0 || retrow == dmNOMOREROWS )
    row_no = dmNOMOREROWS;

  row_maxcell = maxcell;

  while( row_no != dmNOMOREROWS  && ( all_rows || row_no <= maxrow ) )
  { 
    if ( varlen_cols )
    {
      /* If the data contains variable-length array columns, the number of */
      /* cells with data in the current row may be less than maxcell.      */
      /* Since we don't want to print a bunch of empty cells, we calculate */
      /* a row_maxcell for each row and use it rather than maxcell when    */
      /* looping over the cells.                                           */

      row_maxsize = basesize;
      row_maxcell = maxcell;

      for ( colno = 0; colno < varlen_ncols; colno++ ) 
      {
        asize = dmGetCurrentArraySize( varlen_cols[colno] );
        if ( asize > row_maxsize )
          row_maxsize = asize;
      }

      if ( row_maxcell > row_maxsize )
        row_maxcell = row_maxsize;
    }

    for ( cell = mincell; cell <= row_maxcell; cell++ )
    {
      ptr = start_ptr;
      if ( verbose > 0 ) 
      {
        if ( use_naxes > 0 )
        {
          long ii = 0;
          dmaPixelNoToPixel( cell, axes, use_naxes, lpix );  
          ptr += sprintf( ptr, "%6ld [ ", row_no );
          for ( ii = 0; ii < use_naxes; ii++ )
          ptr+= sprintf( ptr, "%5ld ", lpix[ii] );
          ptr += sprintf( ptr, "] " );      
        }
        else
        {
          ptr += sprintf( ptr, "%6ld %6ld", row_no, cell );
        }
      }

      for ( colno = 0; colno < ncols; colno++ ) 
      {
        dd = cols[colno];

        // Check for nulls array set up before using it
        if ( nulls )
          null = nulls[colno];

        type = dmGetDataType( dd );
        left = start_ptr + buflen - ptr;

        if ( left > DC_MAXSTR )
          left = DC_MAXSTR;

        /* Set verbose to 0 to avoid sq brackets */
        dmlistGetValueString( dd, 0, type, cell, cell, null, value, left );

        if ( ptr < eptr ) 
        {
          ptr +=  sprintf( ptr, fmt[colno], value );
        }
      }

      *ptr++ = '\0';
      dmlist_print( start_ptr );
    }

    row_no = dmTableNextRow( block );
  }

  if ( nulls )
    free( nulls );


  free( cols );
  free( varlen_cols );
  free( size );

  if (fmt)
    ar_free_c( fmt );

  if ( Alloc_buf )
    free( Alloc_buf );

  return EXIT_SUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmlistPrintData                                 */
/*                                                             */
/* Description:                                                */
/*   Print the data section                                    */
/*                                                             */
/*                                                             */
/*   Called from function dmListBlock()                        */
/*                                                             */
/*  Input parameters:                                          */
/*                    dmBlock  *block                          */
/*                    long      nrows                          */
/*                    long      minrow                         */
/*                    long      maxrow                         */
/*                    long      mincell                        */
/*                    long      maxcell                        */
/*                    long      verbose                        */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            int  status                                      */
/*                                                             */
/***************************************************************/
int dmlistPrintData( dmBlock* block, long nrows, long minrow,
                     long maxrow, long mincell, long maxcell, 
                     dmBool bRawOpt, long verbose )
{
  dmDescriptor **cols   = NULL;
  dmDescriptor  *dd     = NULL;
  dmDataType     type;
  dmDataType     dtype;
  regRegion     *region = NULL;

  void  *null  = NULL;
  void **nulls = NULL;

  char value[DC_MAXSTR+1];
  char unit[DCSIZE+1];
  char name[DCSIZE+1];
  char typename[DCSIZE+1];
  char buf[DC_MAXSTR+1];
  char errbuf[DC_MAXSTR];

  char **fmt       = NULL;  /* Format for writing the name of each column with */
  char  *ptr       = NULL;
  char  *eptr      = NULL;
  char  *Alloc_buf = NULL;
  char  *start_ptr = NULL;
  char  *tmp       = NULL;

  long *size       = NULL;  /* Expected number of bytes for each column */
  const long spare = 80;
  long  ncols      = 0;
  long  colno      = 0;
  long  retrow     = 0;
  long  row_no     = 0;
  long  sum        = 0;
  long  buflen     = 0;
  long  left       = 0;
  long  maxsize    = 0;
  long  asize      = 0;
  long  vsize      = 0; 

  long start_cell = 0;
  long end_cell   = 0;

  int truncated = 0;
  int all_rows  = 0;


  strcpy( unit, " ");
  dmBlockGetName( block, name, DCSIZE );

  tmp = dmBlockGetTypeStr( block );
  strncpy( typename, tmp, DCSIZE );
  free( tmp );

  if ( verbose > 0 )
  {
    sprintf( buf, "Data for %s Block %s", typename, name );
    dmlistBanner( buf );
  }

  if ( dmaBlockIsNull( block ) )
   return EXIT_SUCCESS; /* Null image */

  cols = dmTableOpenColumnList( block, &ncols );
  size = (long*)malloc( ncols * sizeof( long ) );

  // Code Change Bug 8426 April 8, 2011
  // If option is for "RAW" data, do not
  // set up nulls array for value substitutions

  if ( !bRawOpt )
    nulls = dmuAllocVoids( ncols );


  fmt =  ar_alloc_c( 20, ncols );

  if ( ncols > 1 && (cs_uceq( name, "REGION")) )
  {
    region = dmTableReadRegion( block );

    if ( region )
    {
      regComposeRegion( region, buf, DC_MAXSTR );
      printf( "Region Block: %s\n", buf );
    }
    regFree( region );
  }

  /* Check how much space we need.  Maxsize is max number */
  /*  of allowed array elements                           */

  maxsize = maxcell - mincell + 1;

  for ( colno = 0; colno < ncols; colno++ ) 
  {
    dd = cols[colno];

    /* buflen here = size of buffer needed for one data element */
    buflen = dmlistGetColSize( dd );

    // Code Change Bug 8426  April 1, 2011
    // Don't use null substitutions if RAW flag being used

    if ( nulls )
      dmlistGetNull( dd, &nulls[colno] );

    asize = dmGetArraySize( dd );

    if ( asize > maxsize && maxsize > 0 )
      asize = maxsize;

    size[colno] = asize * buflen;

    if ( size[colno] > DC_MAXSTR )
    {
      /* Code Change  Bug 10815  November 28, 2007         */
      /* Ascii format string columns are given default     */
      /* length of 1024, which is exactly the size of      */
      /* DC_MAXSTR. String length gets 2 extra char places */
      /* added to it in call to dmlistGetColSize()         */
      /* For all columns except string type, generate an   */
      /* error message if length > 1024                    */
      /* For string type columns, set the size of column to*/
      /* zero and let the print function handle size issue */

      /* reduce size to fit column in DC_MAXSTR */
      dmGetName( dd, name, DCSIZE );
      left = size[colno] - DC_MAXSTR;
      asize -= ceil((double)left / (double)buflen);

      if ( dmGetDataType(dd) != DT_CHAR )
      { 
        if ( asize )
        {
          size[colno] = asize * buflen;
	  sprintf(errbuf,"Warning: col %s has too many array elements and will be truncated.\n", name );
          dmaErrPrint( errbuf );
        }
        else
        {
          size[colno] = asize;
          sprintf(errbuf, "Warning: col %s length is too large and will be truncated.\n", name );
	  dmaErrPrint(errbuf);
        }

        truncated = 1;
      }
      else
        size[colno] = 0;
    }

    sum += size[colno] + 1;
  }

  if ( truncated )
  {
    dmaErrPrint( "\tPlease try opt=array.\n\n" );
  }

  /* Take special action if it's huge  */
  if ( sum < DC_MAXSTR )
  {
    buflen = DC_MAXSTR;
    Alloc_buf = (char*)malloc( buflen+spare );
    start_ptr = Alloc_buf;
  }
  else
  {
    buflen = sum + DC_MAXSTR;
    Alloc_buf = (char*)malloc( buflen+spare );
    start_ptr = Alloc_buf;
  }
 
  ptr  = start_ptr;
  eptr = ptr + buflen - spare;
 
  if ( verbose > 0 )
    ptr += sprintf( ptr, "%-6s", "ROW " );
  else
    ptr += sprintf( ptr, "# " );

  for ( colno = 0; colno < ncols; colno++ )
  {
    dd = cols[colno];
    sprintf( fmt[colno], " %%-%lds",  size[colno] );
    dmlistGetDescCompositeName( dd, name, DCSIZE );

    if ( ptr < eptr )
    {
      if ( colno < ncols-1 )
        ptr += sprintf( ptr, fmt[colno], name );
      else
        ptr += sprintf( ptr, " %-s", name );
    }

    /* Right justify numeric cols */
    dtype = dmGetDataType( dd );

    if ( dtype != dmTEXT )
      sprintf( fmt[colno], " %%%lds",  size[colno] );
  }

  dmlist_print( start_ptr );

  if ( verbose > 0 )
    dmlist_print( " " );

  all_rows =  ( maxrow <= 0 );
  row_no = minrow;
  retrow = row_no;


  // Code change - Bug 9189 - January 5, 2007
  //
  //  Need to know if this is a null block (retrow = dmNOMOREROWS)
  //  so always call dmTableSetRow() except if this is image block

  if ( dmBlockGetType(block) != dmIMAGE )
    retrow = dmTableSetRow( block, row_no );

  if ( nrows <= 0 || retrow == dmNOMOREROWS )
     row_no = dmNOMOREROWS;


  while( row_no != dmNOMOREROWS  && ( all_rows || row_no <= maxrow ) )
  { 
    ptr = start_ptr;

    if ( verbose > 0 ) 
      ptr += sprintf( ptr, "%6ld", row_no );

    for ( colno = 0; colno < ncols; colno++ ) 
    {
      dd   = cols[colno];
      type = dmGetDataType( dd );
      left = start_ptr + buflen - ptr;

      if ( left > DC_MAXSTR )
        left = DC_MAXSTR;

      if ( nulls )
        null = (void*)nulls[colno];

      /* ENH 11815                                                    */
      /* Determine if the column is a scalar or array                 */
      /* Print scalar columns even if cell range is given.            */
      /* NOTE: BIT columns are stored as bytes.. need to check length */
      ( type == dmBIT )?( asize = dmDescriptorGetLength(dd)):( asize = dmGetArraySize(dd));
      if (( asize == 1 ) && ( maxcell > 0 ))
      {
	/* Scalar column with cell range specified */
	start_cell = mincell;
	end_cell   = 0;
      }
      else
      {
	start_cell = mincell;
	end_cell   = maxcell;
      }
      dmlistGetValueString( dd, verbose, type, start_cell, end_cell, null, value, left );

      vsize = strlen( value );
 
      if ( vsize < size[colno] )
        vsize = size[colno];

      if ( ptr + vsize >= eptr ) 
      {
        long buf_offset = 0;

        buflen *= 2;
        buf_offset = ptr - start_ptr;
        Alloc_buf = (char*)realloc( Alloc_buf, buflen+spare );
        ptr = Alloc_buf + buf_offset;
        start_ptr = Alloc_buf;
        eptr = start_ptr + buflen;
      }

      ptr +=  sprintf( ptr, fmt[colno], value );
    }

    *ptr++ = '\0';

    dmlist_print( start_ptr );
    row_no = dmTableNextRow( block );

  } /* end while loop */

  if ( nulls )
    free( nulls );

  free( cols );
  free( size );

  if (fmt)
    ar_free_c( fmt );

  if ( Alloc_buf )
    free( Alloc_buf );

  return  EXIT_SUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmuAllocVoids                                   */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*    Called from functions dmlistPrintArrayData()             */
/*                          dmlistPrintData()                  */
/*                                                             */
/*  Input parameters:                                          */
/*                    long   n                                 */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void **dmuAllocVoids( long n )
{
  double *data = NULL;
  void  **vp   = NULL;
  long    i    = 0;

  if ( n <= 0 )
    return NULL;

  data = calloc( n, sizeof( double ));
  vp = calloc( n, sizeof( void* ));

  for ( i = 0; i < n; i++ )
    vp[i] = (void*)(data+i);

  return vp;
}

