/*                                                                
**  Copyright (C) 1999-2001,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


#include "dmtlib.h"
#include "dmtcommon.h"
#include "dmlist.h"
#include "dmtnan.h"

/*****************************************************************/
/*                                                               */
/*   dmlist_format.c                                             */
/*                                                               */
/*   Functions used by dmlist to format column information.      */
/*   These functions were formerly in dmtlib.                    */
/*                                                               */
/*                                                               */
/*****************************************************************/

/* Prototypes */
typedef long (*PrintItemClass)( void*, long, void*, char* );

static void PrintVector( void* data, long dim, long n, int fancy, 
    PrintItemClass PrintFunc, const long spare, void* null, char* value, long maxlen );

static void StringRemoveTrailZeroes( char* buf );

long PrintSignedArrayItem_d( double* data, long offset, void* null, char* buf );



/***************************************************************/
/*                                                             */
/* Function:  dmuPrint_d                                       */
/*                                                             */
/* Description:                                                */
/*   Generic routine to print a double precision value         */
/*   to a string, handling INF, NaN, Null etc.                 */
/*   Remove multiple trailing zeros and tailor the format      */
/*   to the size of the number.                                */
/*                                                             */
/*   Called from functions Print_d                             */
/*                         PrintSigned_d                       */
/*                         PrintArrayItem_d                    */
/*                         PrintSignedArrayItem_d              */
/*                         PrintSignedVector_d                 */
/*                         dmlistGetValueString                */
/*                                                             */
/* Input parameters:                                           */
/*   double  x       -  value to be tranferred                 */
/*   double *null    -  value which represents NUL             */
/*   int     vsigned -  flag indicating signed format desired  */
/*   int     maxlen  -  max chars to write into buf            */
/*                                                             */
/* Output parameters:                                          */
/*   char *buf       - output buffer (alloc'd by user)         */
/*                                                             */
/* Returns:                                                    */
/*   Number of characters written to buf                       */
/*   long nchars                                               */
/*                                                             */
/***************************************************************/
long dmuPrint_d( double x, double* null, int vsigned, char* buf, int maxlen )
{
  char   tbuf[DCSIZE];   /* temp buffer to hold value as string */
  char  *ptr    = NULL;  /* pointer into temp buf               */
  long   nchars = 0;     /* Number of chars written to buf      */
  double abs_x  = 0.0;   /* Absolute value of input x           */


  abs_x = fabs( x ); 
  
  if ( dmt_dINF( x ) )
  {
    nchars = sprintf( tbuf, "%s", "Inf" );
  }
  else if ( dmt_dNINF( x )) 
  {
    nchars = sprintf( tbuf, "%s", "-Inf" );
  }
  else if ( dmt_dNAN( x ) )
  {
    nchars = sprintf( tbuf, "%s", "NaN" );
  }
  else if ( null && x == *null )
  {
    nchars = sprintf( tbuf, "%s", "Null" );
  }
  else if ( abs_x < 1.0E-3 || abs_x > 1.0E9 ) 
  {
    if ( vsigned )
      nchars = sprintf( tbuf, "%+G", x );
    else 
      nchars = sprintf( tbuf, "%20.10G", x );
  }
  else 
  {
    if ( vsigned && abs_x < 1.0E6 ) 
    {
      sprintf( tbuf, "%+12.4f", x ); 
    }
    else if ( vsigned ) 
    {
      sprintf( tbuf, "%+10f", x ); 
    }
    else if ( abs_x < 1.0 ) 
    {
      sprintf( tbuf, "%20.14f", x );
    }
    else 
    {
      sprintf( tbuf, "%20.10f", x ); 
    }

    StringRemoveTrailZeroes( tbuf );
    nchars = strlen( tbuf );
  }

  /* Set pointer to beginning of string */
  ptr = tbuf;
 
  /* Delete leading spaces */
  if ( vsigned )
  {
    while( *ptr == ' ' ) 
    {
      ++ptr;
      --nchars;
    }
  }

  /* Transfer string to output buffer */
  strcpy( buf, ptr );

  return( nchars );
}


/***************************************************************/
/*                                                             */
/* Function:  Print_d                                          */
/*                                                             */
/* Description:                                                */
/*   Write a single double precision number to a string        */
/*                                                             */
/*   Called from functions dmlistGetNullString()               */
/*                         dmlistGetRange()                    */
/*                         dmlistSubspacePrint()               */
/*                                                             */
/* Input parameters:                                           */
/*   double x         - value to be tranferred                 */
/*                                                             */
/* Output parameters:                                          */
/*   char  *buf       - string rep. of value, alloc'd by user  */
/*                                                             */
/* Returns:                                                    */
/*        long " number of chars"                              */
/*                                                             */
/***************************************************************/
long Print_d( double x, char* buf )
{
  int vsigned = 0;
  int maxlen  = 0;

  return dmuPrint_d( x, NULL, vsigned, buf, maxlen );
}


/***************************************************************/
/*                                                             */
/* Function:   PrintSigned_d                                   */
/*                                                             */
/* Description:                                                */
/*   Write a single double precision number to a string        */
/*                                                             */
/*   Called from function dmlistPrintTransform()               */
/*                                                             */
/* Input parameters:                                           */
/*  double  x         - value to be tranferred                 */
/*  long    maxlen    - max chars to write to buf              */
/*                                                             */
/* Output parameters:                                          */
/*   char  *buf       - string rep. of value, alloc'd by user  */
/*                                                             */
/* Returns:                                                    */
/*        long " number of chars"                              */
/*                                                             */
/***************************************************************/
long PrintSigned_d( double x, char* buf, long maxlen )
{
  int vsigned = 1;

  return dmuPrint_d( x, NULL, vsigned, buf, maxlen );
}


/***************************************************************/
/*                                                             */
/* Function:  PrintArrayItem_d                                 */
/*                                                             */
/* Description:                                                */
/*   Write a single double precision value from an array       */
/*   to a string                                               */
/*                                                             */
/*   Passed as a parameter to function PrintVector() from      */
/*   function PrintVector_d                                    */
/*                                                             */
/* Input parameters:                                           */
/*   double  *data      - value to be tranferred               */
/*   long     offset    - index in array to transfer           */
/*   void    *null      - value which represents NULL          */
/*                                                             */
/* Output parameters:                                          */
/*   char    *buf       - string rep. of value, alloc'd by user */
/*                                                             */
/* Returns:                                                    */
/*        long " number of chars"                              */
/*                                                             */
/***************************************************************/
long PrintArrayItem_d( double* data, long offset, void* null, char* buf )
{
  return( dmuPrint_d( data[offset], null, 0, buf, DC_MAXSTR ) );
}


/***************************************************************/
/*                                                             */
/* Function:  PrintSignedArrayItem_d                           */
/*                                                             */
/* Description:                                                */
/*   Write a single double precision value from an array       */
/*   to a string                                               */
/*                                                             */
/*   Passed as a parameter to function PrintVector() from      */
/*   function PrintSignedVector_d                              */
/*                                                             */
/* Input parameters:                                           */
/*   double  *data      - value to be tranferred               */
/*   long     offset    - index in array to transfer           */
/*   void    *null      - value which represents NULL          */
/*                                                             */
/* Output parameters:                                          */
/*   char    *buf       - string rep. of value, alloc'd by user*/
/*                                                             */
/* Returns:                                                    */
/*        long "number of chars"                               */
/*                                                             */
/***************************************************************/
long PrintSignedArrayItem_d( double* data, long offset, void* null, char* buf )
{
  return ( dmuPrint_d( data[offset], null, 1, buf, DC_MAXSTR ));
}


/***************************************************************/
/*                                                             */
/* Function:  PrintArrayItem_c                                 */
/*                                                             */
/* Description:                                                */
/*   Write a single double precision value from an array       */
/*   to a string                                               */
/*                                                             */
/*   Passed as a parameter to function PrintVector() from      */
/*   function PrintVector_c                                    */
/*                                                             */
/* Input parameters:                                           */
/*   char   **data    - value to be tranferred                 */
/*   long     offset  - index in array to transfer             */
/*   void    *null    - value which represents NULL            */
/*                                                             */
/* Output parameters:                                          */
/*   char    *buf     - string rep. of value, alloc'd by user  */
/*                                                             */
/* Returns:                                                    */
/*        long "number of chars"                               */
/*                                                             */
/***************************************************************/
long PrintArrayItem_c( char** data, long offset, void* null, char* buf )
{
  return( sprintf( buf, "%s", data[offset] ) );
}


/***************************************************************/
/*                                                             */
/* Function:  PrintArrayItem_l                                 */
/*                                                             */
/* Description:                                                */
/*   Write a single double precision value from an array       */
/*   to a string                                               */
/*                                                             */
/*   Passed as a parameter to function PrintVector() from      */
/*   function PrintVector_l                                    */
/*                                                             */
/* Input parameters:                                           */
/*   long    *data    - value to be tranferred                 */
/*   long     offset  - index in array to transfer             */
/*   void    *null    - value which represents NULL            */
/*                                                             */
/* Output parameters:                                          */
/*   char    *buf     - string rep. of value, alloc'd by user  */
/*                                                             */
/* Returns:                                                    */
/*        long "number of chars"                               */
/*                                                             */
/***************************************************************/
long PrintArrayItem_l( long* data, long offset, void* null, char* buf )
{
  if ( null && data[offset] == *(long*)null )
    return( sprintf( buf, "Null" ) );
  else
    return( sprintf( buf, "%ld", data[offset] ) );
}


/***************************************************************/
/*                                                             */
/* Function:  PrintArrayItem_ul                                */
/*                                                             */
/* Description:                                                */
/*   Write a single double precision value from an array       */
/*   to a string                                               */
/*                                                             */
/*   Passed as a parameter to function PrintVector() from      */
/*   function PrintVector_ul                                   */
/*                                                             */
/* Input parameters:                                           */
/*   unsigned long *data   - value to be tranferred            */
/*   long           offset - index in array to transfer        */
/*   void          *null   - value which represents NUL        */
/*                                                             */
/* Output parameters:                                          */
/*   char    *buf     - string rep. of value, alloc'd by user  */
/*                                                             */
/* Returns:                                                    */
/*        long "number of chars"                               */
/*                                                             */
/***************************************************************/
long PrintArrayItem_ul( unsigned long* data, long offset, void* null, char* buf )
{
  if ( null && data[offset] == *(unsigned long*)null )
    return( sprintf( buf, "Null" ) );
  else
   return( sprintf( buf, "%lu", data[offset] ) );
}



/***************************************************************/
/*                                                             */
/* Function:  PrintVector_d                                    */
/*                                                             */
/* Description:                                                */
/*   Print a vector to a string as a separated list enclosed   */
/*   in square parens.  Style of separator is determined by    */
/*   the mode parameter                                        */
/*                                                             */
/*   Called from function dmlistGetValueString()               */
/*                                                             */
/* Input parameters:                                           */
/*   double *data   -  values to be tranferred                 */
/*   long    dim    -  vector dimentionality                   */
/*   long    n      -  vector length                           */
/*   int     mode   -  determines separators to use            */
/*   void   *null   -  value representing NULL                 */
/*   long    maxlen -  max number of chars to write into value */
/*                                                             */
/* Output parameters:                                          */
/*   char   *value  -  output string                           */
/*                                                             */
/* Returns:                                                    */
/*                                                             */
/*                                                             */
/***************************************************************/
void PrintVector_d( double* data, long dim, long n, int mode, void* null, char* value, long maxlen )
{
  const long spare = 20; /* max expected size of one element */

  PrintVector( data, dim, n, mode, (PrintItemClass)PrintArrayItem_d, spare, null, value, maxlen );
}

/***************************************************************/
/*                                                             */
/* Function:  PrintsignedVector_d                              */
/*                                                             */
/* Description:                                                */
/*   Print a vector to a string as a separated list enclosed   */
/*   in square parens.  Style of separator is determined by    */
/*   the mode parameter                                        */
/*                                                             */
/*   Called from function dmlistPrintTransform()               */
/*                                                             */
/* Input parameters:                                           */
/*   double *data   -  values to be tranferred                 */
/*   long    dim    -  vector dimentionality                   */
/*   long    n      -  vector length                           */
/*   int     mode   -  determines separators to use            */
/*   void   *null   -  value representing NULL                 */
/*   long    maxlen -  max number of chars to write into value */
/*                                                             */
/* Output parameters:                                          */
/*   char   *value  -  output string                           */
/*                                                             */
/* Returns:                                                    */
/*                                                             */
/*                                                             */
/***************************************************************/
void PrintSignedVector_d( double* data, long dim, long n, int mode, void* null, char* value, long maxlen )
{
  const long spare = 20; /* max expected size of one element */

  if  ( n * dim > 1 ) 
    PrintVector( data, dim, n, mode, (PrintItemClass)PrintSignedArrayItem_d, spare, null, value, maxlen );
  else
    dmuPrint_d( data[0], null, 1, value, maxlen );
}

/***************************************************************/
/*                                                             */
/* Function:  PrintVector_c                                    */
/*                                                             */
/* Description:                                                */
/*   Print a vector to a string as a separated list enclosed   */
/*   in square parens.  Style of separator is determined by    */
/*   the mode parameter                                        */
/*                                                             */
/*   Called from function dmlistGetValueString()               */
/*                                                             */
/* Input parameters:                                           */
/*   char  **data   -  values to be tranferred                 */
/*   long    dim    -  vector dimentionality                   */
/*   long    n      -  vector length                           */
/*   int     fancy  -  use fancy formatting?                   */
/*   long    spare  -   max expected size of one element       */
/*   long    maxlen -  max number of chars to write into value */
/*                                                             */
/* Output parameters:                                          */
/*   char   *value  -  output string                           */
/*                                                             */
/* Returns:                                                    */
/*                                                             */
/*                                                             */
/***************************************************************/
void PrintVector_c( char** data, long dim, long n, int fancy, long spare, char* value, long maxlen )
{
  PrintVector( data, dim, n, fancy, (PrintItemClass)PrintArrayItem_c, spare, NULL, value, maxlen );
}

/***************************************************************/
/*                                                             */
/* Function:  PrintVector_l                                    */
/*                                                             */
/* Description:                                                */
/*   Print a vector to a string as a separated list enclosed   */
/*   in square parens.  Style of separator is determined by    */
/*   the mode parameter                                        */
/*                                                             */
/*   Called from function dmlistGetValueString()               */
/*                                                             */
/* Input parameters:                                           */
/*   double *data   -  values to be tranferred                 */
/*   long    dim    -  vector dimentionality                   */
/*   long    n      -  vector length                           */
/*   int     mode   -  determines separators to use            */
/*   void   *null   -  value representing NULL                 */
/*   long    maxlen -  max number of chars to write into value */
/*                                                             */
/* Output parameters:                                          */
/*   char   *value  -  output string                           */
/*                                                             */
/* Returns:                                                    */
/*                                                             */
/*                                                             */
/***************************************************************/
void PrintVector_l( long* data, long dim, long n, int mode, void* null, char* value, long maxlen )
{
  const long spare = 10;  /*  max expected size of one element */

  PrintVector( data, dim, n, mode, (PrintItemClass)PrintArrayItem_l, spare, null, value, maxlen );
}

/***************************************************************/
/*                                                             */
/* Function:  PrintVector_ul                                   */
/*                                                             */
/* Description:                                                */
/*   Print a vector to a string as a separated list enclosed   */
/*   in square parens.  Style of separator is determined by    */
/*   the mode parameter                                        */
/*                                                             */
/*   Called from function dmlistGetValueString()               */
/*                                                             */
/* Input parameters:                                           */
/* unsigned long *data - values to be tranferred               */
/*   long    dim    -  vector dimentionality                   */
/*   long    n      -  vector length                           */
/*   int     mode   -  determines separators to use            */
/*   void   *null   -  value representing NULL                 */
/*   long    maxlen -  max number of chars to write into value */
/*                                                             */
/* Output parameters:                                          */
/*   char   *value  -  output string                           */
/*                                                             */
/* Returns:                                                    */
/*                                                             */
/*                                                             */
/***************************************************************/
void PrintVector_ul( unsigned long* data, long dim, long n, int mode, void* null, char* value, long maxlen )
{
  const long spare = 9;  /* max expected size of one element */

  PrintVector( data, dim, n, mode, (PrintItemClass)PrintArrayItem_ul, spare, null, value, maxlen );
}

/***************************************************************/
/*                                                             */
/* Function:  PrintVector_q                                    */
/*                                                             */
/* Description:                                                */
/*   Print a vector to a string as a separated list enclosed   */
/*   in square parens.  Style of separator is determined by    */
/*   the mode parameter                                        */
/*                                                             */
/*   Called from function dmlistGetValueString()               */
/*                                                             */
/*  This function was added for Bug 11058                      */
/*                                                             */
/* Input parameters:                                           */
/*   dmBool *data   - values to be tranferred                  */
/*   long    dim    -  vector dimentionality                   */
/*   long    nvals  -  vector length                           */
/*   int     fancy  -  fancy formatting?                       */
/*   long    spare  -  max length of any one value             */
/*   long    maxlen -  max number of chars to write into value */
/*                                                             */
/* Output parameters:                                          */
/*   char   *value  -  output string                           */
/*                                                             */
/* Returns:                                                    */
/*                                                             */
/*                                                             */
/***************************************************************/
void PrintVector_q( dmBool* data, long dim, long nvals, int fancy,
                    long spare, char *value, long maxlen)
{
  char **strArray  = NULL;
  dmBool boolValue = dmFALSE;
  long   ii         = 0;
  long   maxLength  = 6;

  // Array strArray is used to output TRUE or FALSE values
  // into output parameter array 'values'
  // The variable maxLength used below in calloc is size of the
  // string 'false' plus one char for null terminator and
  // is the max expected size of string values

  strArray = (char**)malloc( nvals * sizeof( char* ) );

   for ( ii = 0; ii < nvals; ii++ )
     strArray[ii] = (char*)calloc( maxLength, sizeof( char ) );

   // Map boolean values to either TRUE or FALSE in string array
   for ( ii = 0; ii < nvals; ii++)
   {
     boolValue = data[ii];

     if (boolValue)
       strcpy(strArray[ii], "TRUE ");
     else
       strcpy(strArray[ii], "FALSE");
   }

   PrintVector( strArray, dim, nvals, fancy, (PrintItemClass)PrintArrayItem_c,
                spare,  NULL, value, maxlen );

   for ( ii = 0; ii < nvals; ii++ )
     free (strArray[ii]);

   free(strArray);
}

/***************************************************************/
/*                                                             */
/* Function:  PrintVector                                      */
/*                                                             */
/* Description:                                                */
/*   Tranfer vector data to a formatted string                 */
/*                                                             */
/*   Called from function PrintVector_d()                      */
/*                                                             */
/* Input parameters:                                           */
/*  void * data       - values to be tranferred                */
/*  long   dim        - vector dimentionality                  */
/*  long   n          - vector length                          */
/*  int    fancy      - determines separators to ue            */
/*  PrintItemClass  PrintFunc  -  function to use to transfer  */
/*                                each value                   */
/*  const long  spare - max expected size of one element       */
/*  void  *null       - value representing NULL                */
/*  long   maxlen     - max number of chars to write into value*/
/*                                                             */
/* Output parameters:                                          */
/*   value            - output string                          */
/*                                                             */
/* Returns:                                                    */
/*   nothing                                                   */
/*                                                             */
/***************************************************************/
void PrintVector(
  void *data,                /* Data array                          */
  long dim,                  /* Vector dimensionality               */
  long n,                    /* Vector length                       */
  int fancy,                 /* Use fancy formatting?               */
  PrintItemClass PrintFunc,  /* Function to print one element       */
                             /*    returns the number of bytes to advance the pointer */
  const long spare,          /* Max expected size of one element    */
  void* null,
  char* value,               /* Result string                       */
  long maxlen                /* Max chars to write to output string */
 )
{
  long ii = 0;
  long jj = 0;               /* counters                            */
  long offset = 0;           /* offset to value in data array       */
  long estlen = 0;           /* Est. size of vector set as string   */
  int  array_pars = 0;

  char *ptr = NULL;          /* Current position in output string   */
  char lpar ='(',  rpar=')',  sep = ' ';
  char alpar='[', arpar=']', asep = ' ';


  /* make sure dim is non-zero */
  if ( dim == 0 )
    dim = 1;

  /* setup separators and array paren. flag */
  array_pars = ( n > 1 && fancy) || (fancy > 1 );
  if (fancy)
    { sep = ','; }
  if (fancy > 1 )
    { asep = ','; }

  /* Set pointer to beginning of value string */
  ptr = value;

  /* open array bracket. */
  if ( array_pars )
    *ptr++ = alpar;

  /* Determine expected size of each vector set */
  estlen = (dim*spare) + (dim-1);         /* values + vector element sep.  */

  if ( dim > 1 && fancy )
    estlen += 2;      /* vector brackets  */

  if ( n > 1 )
    ++estlen;        /* array separator */

  if ( array_pars )
    ++estlen;         /* closing array bracket */


  /* Loop over array elements */
  for ( ii = 0; ii < n; ii++ ) 
  {
    /* Check that enough room exists for next set */
    if ( (ptr - value + estlen) > maxlen )
      break;

    /* Add array separator */
    if ( ii > 0)
      *ptr++ = asep;

    /* Open vector bracket. */
    if ( dim > 1 && fancy ) 
      *ptr++ = lpar;

    /* Loop vector dims  */
    for ( jj = 0; jj < dim; jj++ )
    {
      /* Add vector element separator */
      if ( jj > 0 )
        *ptr++ = sep;

      /* Calculate offset into data array */
      offset = ii  + jj * n;

      /* Print value and increment current pointer */
      ptr += PrintFunc( data, offset, null, ptr );
    }

    /* close vector bracket */ 
    if ( dim > 1 && fancy )
      *ptr++ = rpar;
  }

  /* close array bracket. */
  if ( array_pars )
    *ptr++ = arpar;

  *ptr++ = '\0';
}


/***************************************************************/
/*                                                             */
/* Function: StringRemoveTrailZeroes                           */
/*                                                             */
/* Description:                                                */
/*   Remove trailing zeroes except for one                     */
/*                                                             */
/*   Called from function dmuPrint_d()                         */
/*                                                             */
/* Input parameters:                                           */
/*        char *buf                                            */
/*                                                             */
/* Output parameters:                                          */
/*                                                             */
/* Returns:                                                    */
/*          void                                               */
/*                                                             */
/***************************************************************/
static void StringRemoveTrailZeroes( char* buf )
{
  char* ptr1 = NULL;
  char* ptr2 = NULL;

  ptr1 = strchr( buf, '.' );

  if ( ptr1 ) 
  {
    ptr2 = strchr( ptr1, '\0' );
    --ptr2;

    while( ptr2 > ptr1 && *ptr2 == '0' )
      --ptr2;

    ++ptr2;

    if ( *ptr2 == '0' )
      ++ptr2;

    *ptr2 = '\0';
  }
}

