/*                                                                
**  Copyright (C) 1999-2000,2003,2006-2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/*****************************************************************/
/*                                                               */
/*   dmlist_funcs.c                                              */
/*                                                               */
/*   This is the callable version of the main program routine    */
/*   for tool dmlist.  This function was formerly in dmtlib.     */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmtlib.h"
#include "dmtcommon.h"
#include "dmlist.h"

#if !defined(DMLOCAL)
#include "dmtprocess.h"
#endif


/***************************************************************/
/*                                                             */
/* Function:   n_dmlist                                        */
/*                                                             */
/* Description:                                                */
/*   main function for tool dmlist                             */
/*                                                             */
/*   Called from function dmlist()                             */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*     char   *input                                           */
/*     char   *output                                          */
/*     char   *opt                                             */
/*     char   *rows                                            */
/*     char   *cells                                           */
/*     dmBool  clobber                                         */
/*     int     verbose                                         */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           int return value from dmaProcessStacks()          */
/*                                                             */
/***************************************************************/
int n_dmlist( char* input, char* output, char* opt, char* rows, char* cells, dmBool clobber, int verbose )
{
  dmListInfo info;                 /* DMLIST-specific info for block process */
  char       toolname[DC_MAXSTR];  /* Text string to write as history record */
  int        retval = 0;


  /* Initialize error library */
  dmInit(0,NULL);

  /* Initialize info structure */
  dmaListInfoInit( &info );

  if ( clobber )
    info.mode |= DMA_MODE_CLOBBER;

  /* Set dmlist info structure fields.                */
  /* NOTE: Will trim matched options from option spec */
  dmaListInfoSet( &info, rows, cells, opt );

  dmlist_io_out( stdout );

  /* Send remaining option spec to kernel layer */
  if ( opt && *opt )
    dmKernelSetOption( opt );	

  strcpy( toolname, "dmlist" );

#if !defined( DMLOCAL)
  retval = dmaProcessStacks( input, output, toolname, info.mode, verbose,
			     (BlockProcess)dmListBlock, (void*)&info);
#else
  retval = dmaProcessFile( input, output, toolname, info.mode, verbose,
		       (BlockProcess)dmListBlock, (void*)&info, 1, 1);
#endif
  return retval;
}


/***************************************************************/
/*                                                             */
/* Function:   dmListBlock                                     */
/*                                                             */
/* Description:                                                */
/*   Function passes as parameter to dmaProcessStacks          */
/*   from function n_dmcopy()                                  */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmBlock  *block                                     */
/*         int       dmverbose                                 */
/*         void     *vinfo                                     */
/*         void     *params                                    */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*  Returns:                                                   */
/*          dmBlock* NULL                                      */
/*                                                             */
/***************************************************************/
dmBlock* dmListBlock( dmBlock* block, int dmverbose, void* vinfo, void* params )
{
  dmCopyInfo *ginfo = (dmCopyInfo*)vinfo;
  dmListInfo *info  = (dmListInfo*)params;
  dmDataset  *ds    = NULL;
  dmBool      bRawOpt = dmFALSE;

  char  buf[DC_MAXSTR];
       
  long block_no = 0;
  long nrows    = 0;
  long main_no  = 0;
  int  data_verbose = 1;
  int  list_verbose = 1;
  int  pdata        = 0;

  // Code Change Bug 8426 April 8, 2011
  // If option is "RAW", set flag for use in
  // called functions

  if ( info->PrintRaw )
    bRawOpt = dmTRUE;

  if ( info->PrintClean )
    list_verbose = 0;

  block_no = dmBlockGetNo( block );

  if ( info->PrintBlocks && ginfo->first && list_verbose > 0 ) 
  {
    ds = dmBlockGetDataset( block );
    main_no = dmBlockGetNo( ginfo->block );
    dmlistDatasetListBlocks( ds, ginfo->block, main_no );
  }

  ginfo->first = 0;

  if ( dmverbose >=2 )
    fprintf( stderr, "DMLIST: Print Block Info\n" );

  pdata = ( info->PrintData || info->PrintArray );

  if ( pdata || dmverbose >= 1 )
  {
    dmlistPrintBlockInfo( block, block_no, buf, &nrows );

    if ( dmverbose >= 1 ) 
    {
      dmlistBanner( buf );
    }
  }

  if ( info->PrintCounts )
    dmlistPrintCounts( block );

  if ( info->PrintBins )
    dmlistPrintBins( block, ginfo->input, info->minrow, info->maxrow );

  if ( info->PrintHeader || info->PrintComment )  
  {
    if ( dmverbose >=3 )
     fprintf( stderr, "DMLIST: Print Header\n" );

    if ( info->PrintRaw )  
      dmlistPrintRawHeader( block, list_verbose );
    else        
      dmlistPrintHeader( block, list_verbose, info->PrintHeader, info->PrintComment );
  }

  if ( info->PrintCols ) 
  {
    if ( dmverbose >= 3 )
      fprintf( stderr, "DMLIST: Print Cols\n" );

    if ( info->PrintRaw )
      dmlistBlockListRawCols( block );
    else
      dmlistBlockListCols( block );
  }

  if ( dmverbose >= 3 )
    fprintf( stderr, "DMLIST: Print Subspace\n" );

  if ( info->PrintSubspace )
     dmlistBlockPrintSubspace( block );

  if ( dmverbose >= 3 )
    fprintf( stderr, "DMLIST: Print Data\n" );

  data_verbose = list_verbose;

  if ( info->PrintRaw )
    data_verbose = 0;  /* Back compat*/

  if ( info->PrintData )
    dmlistPrintData( block, nrows, info->minrow, info->maxrow, 
                     info->mincell, info->maxcell, bRawOpt, data_verbose );

  if ( info->PrintArray)
    dmlistPrintArrayData( block, nrows, info->minrow, info->maxrow, 
                          info->mincell, info->maxcell, bRawOpt, list_verbose );

  if ( dmverbose >= 3 )
    fprintf( stderr, "DMLIST: Iterate\n" );

 return NULL;
}

