/*                                                                
**  Copyright (C) 1999-2001,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


#include "dmtlib.h"
#include "dmtcommon.h"
#include "dmlist.h"

/*****************************************************************/
/*                                                               */
/*   dmlist_header.c                                             */
/*                                                               */
/*   Functions used by dmlist to print header information for a  */
/*   DM file. It assumes scalar keys. These functions were       */
/*   formerly in dmtlib.                                         */
/*                                                               */
/*****************************************************************/

/***************************************************************/
/*                                                             */
/* Function:   dmlistPrintHeader                               */
/*                                                             */
/* Description:                                                */
/*   Print header information                                  */
/*                                                             */
/*   Called from function dmListBlock()                        */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmBlock  *block                                     */
/*         int       verbose                                   */
/*         int       read_keys                                 */
/*         int       read_comments                             */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           int EXIT_SUCCESS                                  */
/*                                                             */
/***************************************************************/
int dmlistPrintHeader( dmBlock* block, int verbose, int read_keys, int read_comments )
{
  dmDescriptor *key = NULL;
  dmDataType    type;

  char  dataType[DCSIZE+1];
  char  value[DC_MAXSTR+1];
  char  name[DCSIZE+1];
  char  unit[DCSIZE+1];
  char  desc[DCSIZE+1];
  char  buf[2*DCSIZE+1];
  char  tmpBuf[DC_MAXSTR];
  char *comment = NULL;
  char *tag     = NULL;

  long nkeys = 0;
  long keyno = 0;
  long du    = 0; 
  long dv    = 0;


  dmBlockGetName( block, name, DCSIZE );

  if (read_keys)
    sprintf( buf, "Header keys for block %s", name );
  else
    if ( read_comments )
      sprintf( buf, "Comment keys for block %s", name );

  if (verbose >= 1 ) 
    dmlistBanner( buf );

  strcpy( unit, " ");
  nkeys = dmBlockGetNoKeys( block );
  dmBlockMoveToKeyNo( block, 0 );

  if ( read_comments )
    while( dmBlockReadComment( block, &tag, &comment ))
    {  
      sprintf( tmpBuf,"%-4s %-20s %s", " -- ", tag, comment );
      dmlist_print( tmpBuf );
      free( tag );
      free( comment );
    }

  for ( keyno = 1; keyno <= nkeys; keyno++ )
  {
    key = dmBlockGetKey( block, keyno );

    if ( key )
    {
      type = dmGetDataType( key );

      /* Get the name and unit */
      dmlistGetDescCompositeName( key, name, DCSIZE );
      dmGetUnit( key, unit, DCSIZE );
     /* Get the value and postfix the unit (if there is space) */
     dmlistGetValueString( key, verbose, type, 0, 0, NULL, value, DC_MAXSTR );

     if ( !cs_is_blank( unit ) )
     {
       dv = cs_ends( value );
       du = cs_ends( unit );

       if ( dv + du + 4 < DC_MAXSTR )
       {
         sprintf( value + dv, " [%s]", unit );
       }
     }

     /* Get the datatype */
     dmlistFormatType( key, dataType, DCSIZE );
     dmGetDesc( key, desc, DCSIZE );

     if ( verbose >= 1 ) 
       sprintf( tmpBuf, "%04ld %-20s %-30s %-12s %s", keyno, name, value, dataType, desc );
     else 
       sprintf( tmpBuf, "%-20s %-30s %s", name, value, desc );

     if ( read_keys )
       dmlist_print( tmpBuf );

     if ( read_comments )
       while( dmBlockReadComment( block, &tag, &comment ))
       {  
         sprintf( tmpBuf,"%-4s %-20s %s", " -- ", tag, comment );
         dmlist_print( tmpBuf );
         free( tag );
         free( comment );
       }
    }
  }
 
  return( EXIT_SUCCESS );
}


/***************************************************************/
/*                                                             */
/* Function:   dmlistPrintRawHeader                            */
/*                                                             */
/* Description:                                                */
/*   Print header information                                  */
/*                                                             */
/*   Called from function dmListBlock()                        */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmBlock  *block                                     */
/*         int       verbose                                   */
//*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           int EXIT_SUCCESS                                  */
/*                                                             */
/***************************************************************/
int dmlistPrintRawHeader( dmBlock* block, int verbose )
{
  char buf[2*DCSIZE+1];
  char name[DCSIZE+1];

  long nkeys  = 0;
  long keyno  = 0;
  long maxlen = 2 * DCSIZE + 1;


  dmBlockGetName( block, name, DCSIZE );
  sprintf( buf, "Raw Header keys for block %s", name );

  if ( verbose > 0 )
    dmlistBanner( buf );

  nkeys = dmBlockGetNoKernelKeys( block );

  for ( keyno = 1; keyno <= nkeys; keyno++ )
  {
    dmBlockGetKernelKey( block, keyno, buf, maxlen );

    if ( verbose > 0 ) 
      dmlist_print( buf );
    else
      dmlist_print( buf+25 );
  }

 return( EXIT_SUCCESS );
}


