/*                                                                
**  Copyright (C) 1999-2001,2003,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/*****************************************************************/
/*                                                               */
/*   dmlist_misc.c                                               */
/*                                                               */
/*   Functions that are used by dmlist. These functions were     */
/*   formerly in dmtlib                                          */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmtlib.h"
#include "dmtcommon.h"
#include "dmlist.h"
#include "dmulib.h"

 /* Global function pointer */
static void (*dmlist_print_function)(char*) = dmlist_print_std;

extern  FILE* dmlist_out;



/***************************************************************/
/*                                                             */
/* Function:   dmaBlockIsNull                                  */
/*                                                             */
/* Description:                                                */
/*   Return true if block is null image                        */
/*                                                             */
/*   Called from function dmlistPrintArrayData() and from      */
/*         dmlistPrintData()                                   */
/*                                                             */
/*  Input parameters:                                          */
/*         dmBlock  *block                                     */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           int dmTRUE or dmFALSE                             */
/*                                                             */
/***************************************************************/
int dmaBlockIsNull( dmBlock* block )
{
  long          naxes     = 9;
  dmDescriptor *imageData = NULL;

  if ( block && dmBlockGetType( block ) == dmIMAGE )
  { 
    imageData = dmImageGetDataDescriptor( block );
    naxes = dmGetArrayDim( imageData );
    return( naxes == 0 );
  }

  return( dmFALSE );
}


/***************************************************************/
/*                                                             */
/* Function:   dmaGetArray_c                                   */
/*                                                             */
/* Description:                                                */
/*   A version of dmGetArray that works for images as well as  */
/*   table columns. Does not support vector arrays.            */
/*                                                             */
/*   Called from function dmlistGetValueString()               */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor *dd                                    */
/*         long          mincell                               */
/*         long          maxcell                               */
/*                                                             */
/*  Output parameters:                                         */
/*         char        **iarr                                  */
/*                                                             */
/*  Returns:                                                   */
/*           nothing                                           */
/*                                                             */
/***************************************************************/
void dmaGetArray_c( dmDescriptor* dd, char** iarr, long mincell, long maxcell )
{
  dmDescriptorType t;  /* Are we table column or image? */
//  dmDataType       dt;

  char **ctmp = NULL;

  long i             = 0;
  long dim           = 0;
  long j             = 0;
  long nitems        = 0;
  long length        = 0;
  long offset        = 0;
  long buffer_offset = 0;
  long nelts         = 0;

  dim    = dmGetElementDim( dd );
  nitems = maxcell - mincell + 1;
  nelts  = maxcell * dim;
  t      = dmDescriptorGetType( dd );
  length = dmDescriptorGetLength( dd )+1;  /* need length for a string */

  if ( t == dmCOLUMN || t == dmCOORD ) 
  {
    if ( mincell > 1 )
    {
      ctmp = (char**)malloc( nelts * sizeof( char* ) );

      for ( j = 0; j < nelts; j++ )
      {
        ctmp[j] = calloc( length, sizeof(char) );
      }

      dmGetVectorArray_c( dd, ctmp, dim, maxcell );

      for ( j = 0; j < dim; j++ )
      {
        offset = j * nitems;
        buffer_offset = j * maxcell + mincell -1;

        for ( i = 0; i < nitems; i++ )
        {
          strcpy( iarr[i+offset], ctmp[buffer_offset+i] );
        }
      }

      for ( j = 0; j < nelts; j++ )
        free( ctmp[j] );
      free( ctmp );
    }
    else
    {
      dmGetVectorArray_c( dd, iarr, dim, maxcell );
    }
  }
  else
    if ( t == dmIMAGEDATA )
    {
      iarr[0] = NULL;
      dmu_err_set( DM_ERR_GENERIC, "dmaGetArray: dmTEXT unsupported for images." );

    }
    else
    {
      iarr[0] = NULL;
      printf( "No GetArray for this descriptor's datatype.\n" );
    }
}



/***************************************************************/
/*                                                             */
/* Function:   dmaGetArray_d                                   */
/*                                                             */
/* Description:                                                */
/*   A version of dmGetArray that works for images as well as  */
/*   table columns. For a table column, use the normal         */
/*   For a table column, use the normal dmGetArray(). For an   */
/*   image, use dmImageDataGetSubArray()                       */
/*                                                             */
/*   Called from function dmlistGetValueString()               */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor *dd                                    */
/*         long          mincell                               */
/*         long          maxcell                               */
/*                                                             */
/*  Output parameters:                                         */
/*         double       *iarr                                  */
/*                                                             */
/*  Returns:                                                   */
/*           nothing                                           */
/*                                                             */
/***************************************************************/
void dmaGetArray_d( dmDescriptor* dd, double* iarr, long mincell, long maxcell )
{
  dmDescriptorType t;  /* Are we table column or image? */
  dmDataType       dt;

  long  lpix[DMA_MAXDIM];
  long  upix[DMA_MAXDIM];
  long *axes          = NULL;
  long  naxes         = 0;
  long  dim           = 0;
  long  maxvals       = 0;
  long  i             = 0;
  long  j             = 0;
  long  offset        = 0;
  long  buffer_offset = 0;
  long  nitems        = 0;

  double *dtmp         = NULL;
  float  *float_buffer = NULL;



  t       = dmDescriptorGetType( dd );
  nitems  = maxcell - mincell + 1;
  dim     = dmGetElementDim( dd );
  maxvals = dmGetArraySize( dd );

  if ( t == dmCOLUMN || t == dmCOORD )
  {
    if ( mincell > 1 )
    {
      dtmp = (double*)malloc( maxcell * dim * sizeof( double ) );
      dmGetVectorArray_d( dd, dtmp, dim, maxcell );

      for ( j = 0; j < dim; j++ )
      {
        offset = j * nitems;
        buffer_offset = j * maxcell + mincell -1;

        for ( i = 0; i < nitems; i++ )
        {
          iarr[i+j] = dtmp[buffer_offset+i];
        }     
      }

      free( dtmp );
    }
    else
    {
      dmGetVectorArray_d( dd, iarr,  dim, maxcell );
    }
  }
  else
    if ( t == dmIMAGEDATA )
    {
      naxes = dmGetArrayDimensions( dd, &axes );
      dmaPixelNoToPixel( mincell, axes, naxes, lpix );  
      dmaPixelNoToPixel( maxcell, axes, naxes, upix );
      dt = dmGetDataType( dd );

      switch( dt )
      {
        case dmDOUBLE:
          dmImageDataGetSubArray_d( dd, lpix, upix, iarr );  
          break;
        case dmFLOAT:
          float_buffer = (float*)malloc( nitems * sizeof( float ) );

          if ( ! float_buffer )
	  {
            dmu_err_set(DM_ERR_GENERIC, "Out of memory");
            break;
          }

          dmImageDataGetSubArray_f( dd, lpix, upix, float_buffer );

          for ( i =0; i < nitems; i++ )
          {
            iarr[i] = (double)float_buffer[i];
          }

          free( float_buffer );
          break;
 
        default:
	{
          dmu_err_set( DM_ERR_GENERIC, "dmaGetArray: data type unsupported");
          break;
        }
      }

      free( axes );

    }
    else
    {
      printf( "No GetArray for this descriptor type\n" );
    }
}




/***************************************************************/
/*                                                             */
/* Function:   dmaGetArray_l                                   */
/*                                                             */
/* Description:                                                */
/*   A version of dmGetArray that works for images as well as  */
/*   table columns. For a table column, use the normal         */
/*   For a table column, use the normal dmGetArray(). For an   */
/*   image, use dmImageDataGetSubArray()                       */
/*                                                             */
/*   Called from function dmlistGetValueString()               */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor *dd                                    */
/*         long          mincell                               */
/*         long          maxcell                               */
/*                                                             */
/*  Output parameters:                                         */
/*         long         *iarr                                  */
/*                                                             */
/*  Returns:                                                   */
/*           nothing                                           */
/*                                                             */
/***************************************************************/
void dmaGetArray_l( dmDescriptor* dd, long* iarr, long mincell, long maxcell )
{
  dmDescriptorType t;  /* Are we table column or image? */
  dmDataType       dt;

  long  lpix[DMA_MAXDIM];
  long  upix[DMA_MAXDIM];
  long *axes          = NULL;
  long *ltmp          = NULL;
  long  naxes         = 0;
  long  i             = 0;
  long  dim           = 0;
  long  j             = 0;
  long  nitems        = 0;
  long  offset        = 0;
  long  buffer_offset = 0;

  short *short_buffer = NULL;

  unsigned char *byte_buffer = NULL;
 

  nitems = maxcell - mincell + 1;
  dim    = dmGetElementDim( dd );
  t      = dmDescriptorGetType( dd );

  if ( t == dmCOLUMN || t == dmCOORD )
  {
    if ( mincell > 1 )
    {
      ltmp = (long*)malloc( maxcell * dim * sizeof( long ) );
      dmGetVectorArray_l( dd, ltmp, dim, maxcell );

      for ( j = 0; j < dim; j++ )
      {
        offset = j * nitems;
        buffer_offset = j * maxcell + mincell -1;

        for ( i = 0; i < nitems; i++ )
        {
          iarr[i+j] = ltmp[buffer_offset+i];
        }     
      }

      free( ltmp );
    }
    else
    {
      dmGetVectorArray_l( dd, iarr, dim, maxcell );
    }
  }
  else
    if ( t == dmIMAGEDATA )
    {
      naxes = dmGetArrayDimensions( dd, &axes );
      dmaPixelNoToPixel( mincell, axes, naxes, lpix );  
      dmaPixelNoToPixel( maxcell, axes, naxes, upix );
      dt = dmGetDataType( dd );

      switch( dt )
      {
        case dmLONG:
          dmImageDataGetSubArray_l( dd, lpix, upix, iarr );  
          break;

        case dmSHORT:
          short_buffer = (short*)malloc( nitems * sizeof( short ) );

          if ( ! short_buffer )
          {
            dmu_err_set( DM_ERR_GENERIC, "Out of memory" );
	    break;
	  }
          dmImageDataGetSubArray_s( dd, lpix, upix, short_buffer );

          for ( i =0; i < nitems; i++ )
          {
            iarr[i] = (long)short_buffer[i];
          }

          free( short_buffer );
          break;

        case dmBYTE:
          byte_buffer = (unsigned char*)malloc( nitems * sizeof(unsigned char) );
          if ( ! byte_buffer )
          {
            dmu_err_set( DM_ERR_GENERIC, "Our of memory" );
            break;
          }

          dmImageDataGetSubArray_ub( dd, lpix, upix, byte_buffer );

          for ( i =0; i < nitems; i++ )
          {
            iarr[i] = (long)byte_buffer[i];
          }

          free( byte_buffer );   
          break;

        default:
          dmu_err_set(DM_ERR_GENERIC, "dmaGetArray: data type unsupported" );
          break;
      }

      free( axes );
    }
    else
    {
      printf( "No GetArray for this descriptor type\n" );
    }
}



/***************************************************************/
/*                                                             */
/* Function:   dmaGetArray_ul                                  */
/*                                                             */
/* Description:                                                */
/*   A version of dmGetArray that works for images as well as  */
/*   table columns. For a table column, use the normal         */
/*   For a table column, use the normal dmGetArray(). For an   */
/*   image, use dmImageDataGetSubArray()                       */
/*                                                             */
/*   Called from function dmlistGetValueString()               */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor  *dd                                   */
/*         long           mincell                              */
/*         long           maxcell                              */
/*                                                             */
/*  Output parameters:                                         */
/*         unsigned long *iarr                                 */
/*                                                             */
/*  Returns:                                                   */
/*           nothing                                           */
/*                                                             */
/***************************************************************/
void dmaGetArray_ul( dmDescriptor* dd, unsigned long* iarr, long mincell, long maxcell )
{
  dmDescriptorType t;  /* Are we table column or image? */
  dmDataType       dt;

  long  lpix[DMA_MAXDIM];
  long  upix[DMA_MAXDIM];
  long *axes          = NULL;
  long  naxes         = 0;
  long  i             = 0;
  long  n             = 0;
  long  j             = 0;
  long  dim           = 0;
  long  nitems        = 0;
  long  offset        = 0;
  long  buffer_offset = 0;

  unsigned long  *utmp = NULL;
  unsigned short *short_buffer = NULL;


  t      = dmDescriptorGetType( dd );
  nitems = maxcell - mincell + 1;
  dim    = dmGetElementDim( dd );

  if ( t == dmCOLUMN || t == dmCOORD )
  {
    if ( mincell > 1 )
    {
      utmp = (unsigned long*)malloc( dim * maxcell * sizeof( unsigned long ) );
      dmGetVectorArray_ul( dd, utmp, dim, maxcell );

      for ( j = 0; j < dim; j++ )
      {
        offset = j * nitems;
        buffer_offset = j * maxcell + mincell -1;

        for ( i = 0; i < nitems; i++ )
        {
          iarr[i+j] = utmp[buffer_offset+i];
        }     
      }
      free( utmp );
    }
    else
    {
      dmGetVectorArray_ul( dd, iarr, dim, maxcell );
    }
  }
  else
    if ( t == dmIMAGEDATA )
    {
      naxes = dmGetArrayDimensions( dd, &axes );
      n = maxcell - mincell + 1;
      dmaPixelNoToPixel( mincell, axes, naxes, lpix );  
      dmaPixelNoToPixel( maxcell, axes, naxes, upix );
      dt = dmGetDataType( dd );

      switch( dt )
      {
        case dmULONG:
          dmImageDataGetSubArray_ul( dd, lpix, upix, iarr );  
          break;

        case dmUSHORT:
          short_buffer = (unsigned short*)malloc( n * sizeof( unsigned short ) );

          if ( ! short_buffer )
          {
            dmu_err_set( DM_ERR_GENERIC, "Out of memory" );
            break;
          }

          dmImageDataGetSubArray_us( dd, lpix, upix, short_buffer );

          for ( i =0; i < n; i++ )
          {
            iarr[i] = (unsigned long)short_buffer[i];
          }
          free( short_buffer );
          break;

        default:
          dmu_err_set( DM_ERR_GENERIC, "dmaGetArray: data type unsupported" );
          break;
      }

      free( axes );
    }
    else
    {
      printf( "No GetArray for this descriptor type\n" );
    }
}

/***************************************************************/
/*                                                             */
/* Function:   dmaGetArray_q                                   */
/*                                                             */
/* Description:                                                */
/*   A version of dmGetArray                                   */
/*   For a table column, use the normal dmGetArray().          */
/*   Presently not designed for image files                    */
/*                                                             */
/*   Called from function dmlistGetValueString()               */
/*                                                             */
/*   This function added for Bug 11058                         */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor  *dd                                   */
/*         long           mincell                              */
/*         long           maxcell                              */
/*                                                             */
/*  Output parameters:                                         */
/*         dmBool *barr                                        */
/*                                                             */
/*  Returns:                                                   */
/*           nothing                                           */
/*                                                             */
/***************************************************************/
void dmaGetArray_q( dmDescriptor* dd, dmBool *barr, long mincell,
                    long maxcell )
{
  dmDescriptorType t;  /* Are we table column or image? */


  long  i             = 0;
  long  j             = 0;
  long  dim           = 0;
  long  nitems        = 0;
  long  offset        = 0;
  long  buffer_offset = 0;

  dmBool  *btmp = NULL;



  t      = dmDescriptorGetType( dd );
  nitems = maxcell - mincell + 1;
  dim    = dmGetElementDim( dd );

  if ( t == dmCOLUMN || t == dmCOORD )
  {
    if ( mincell > 1 )
    {
      btmp = (dmBool*)malloc( dim * maxcell * sizeof( dmBool ) );
      dmGetVectorArray_q( dd, btmp, dim, maxcell );

      for ( j = 0; j < dim; j++ )
      {
        offset = j * nitems;
        buffer_offset = j * maxcell + mincell -1;

        for ( i = 0; i < nitems; i++ )
        {
          barr[i+j] = btmp[buffer_offset+i];
        }     
      }
      free( btmp );
    }
    else
    {
      dmGetVectorArray_q( dd, barr, dim, maxcell );
    }
  }
  else
    if ( t == dmIMAGEDATA )
    {
      // do nothing
      ;
    }
    else
    {
      printf( "No GetArray for this descriptor type\n" );
    }
}


/***************************************************************/
/*                                                             */
/* Function:   dmaGetBit                                       */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*   Called from function dmlistGetValueString()               */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor  *dd                                   */
/*         long           mincell                              */
/*         long           maxcell                              */
/*         long           maxlen                               */
/*                                                             */
/*  Output parameters:                                         */
/*         char          *value                                */
/*                                                             */
/*  Returns:                                                   */
/*           nothing                                           */
/*                                                             */
/***************************************************************/
void dmaGetBit( dmDescriptor* dd, long mincell, long maxcell, char* value, long maxlen )
{
  long len     = 0;
  long nbytes  = 0;
  long byte_no = 0;
  long bit_no  = 0;
  long pos     = 0;
  long offset  = 0;

  unsigned char mask[BYTE_BITS] = { 1, 2, 4, 8, 16, 32, 64, 128 };
  unsigned char buffer[DMAMAXBITS];

  len = dmDescriptorGetLength( dd );

  if ( len == 0 )
    len = BYTE_BITS;

  nbytes = ( len - 1 ) / BYTE_BITS + 1;

  if ( nbytes > DMAMAXBITS || len > maxlen )
  {
    /* Unlikely */
    strcpy( value, "[Bits]" );
    return;
  }

  dmGetScalar_bit( dd, (void*)&buffer );

  for ( bit_no = 1; bit_no <= len; bit_no++ )
  {
    offset  = BYTE_BITS - 1 - (bit_no-1) % BYTE_BITS;
    byte_no = (bit_no-1)/BYTE_BITS;

    if ( bit_no >= mincell && bit_no <= maxcell )
    {
      value[pos++] = ( buffer[byte_no]  & mask[offset] ) ? '1' : '0';
    }
  }

  value[pos]  = '\0';
}



/***************************************************************/
/*                                                             */
/* Function:   dmaGetDataTypeName                              */
/*                                                             */
/* Description:                                                */
/*   Return the descriptor's data type code                    */
/*                                                             */
/*   Called from functions dmlistFormatType()                  */
/*                         dmlistGetNullString()               */
/*                         dmlistGetRange()                    */
/*                         dmlistBlockPrintSubspace()          */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor  *dd                                   */
/*         char          *name                                 */
/*         long           maxlen                               */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           int EXIT_FAILURE                                  */
/*                                                             */
/***************************************************************/
int dmaGetDataTypeName( dmDescriptor* dd, char* name, long maxlen )
{
  char *tmp = NULL;

  tmp = dmGetDataTypeName( dd );

  if ( tmp )
  {
    strncpy( name, tmp, maxlen );
    free( tmp ); 
  }
  else
    strncpy( name, "?", maxlen );

  return( EXIT_FAILURE );

}



/***************************************************************/
/*                                                             */
/* Function:   dmaListInfoInit                                 */
/*                                                             */
/* Description:                                                */
/*    Initialize Info list structure                           */
/*                                                             */
/* Input parameters:                                           */
/*    dmListInfo  *info                                        */
/*                                                             */
/* Output parameters:                                          */
/*                                                             */
/* Returns:                                                    */
/*    nothing                                                  */
/*                                                             */
/***************************************************************/
void dmaListInfoInit( dmListInfo* info )
{
  info->PrintAll      = dmFALSE;
  info->PrintAny      = dmFALSE;
  info->PrintArray    = dmFALSE;
  info->PrintBins     = dmFALSE;
  info->PrintBlocks   = dmFALSE;
  info->PrintClean    = dmFALSE;
  info->PrintCols     = dmFALSE;
  info->PrintComment  = dmFALSE;
  info->PrintCounts   = dmFALSE;
  info->PrintData     = dmFALSE;
  info->PrintFull     = dmFALSE;
  info->PrintHeader   = dmFALSE;
  info->PrintRaw      = dmFALSE;
  info->PrintSubspace = dmFALSE;
  info->minrow  = -1;
  info->maxrow  = -1;
  info->mincell = -1;
  info->maxcell = -1;

  info->mode = DMA_MODE_TEXT;
}



/***************************************************************/
/*                                                             */
/* Function:   dmaListInfoSet                                  */
/*                                                             */
/* Description:                                                */
/*             Parse options for dmlist                        */
/*                                                             */
/*  Input parameters:                                          */
/*             dmListInfo  *info                               */
/*             char        *rows                               */
/*             char        *cells                              */
/*             char        *opt                                */ 
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            nothing                                          */
/*                                                             */
/***************************************************************/
void dmaListInfoSet( dmListInfo* info, char* rows, char* cells, char* opt )
{
  long     ii;               /* Counter                            */ 
  long     nopts = 0;        /* Number of input options            */
  char **optlist = NULL;     /* Hold parsed array of input options */
  char  *tmpstr  = NULL;     /* string to hold unmatched options   */

  int  ntoolopts = 17;
  char **toolopts = NULL;
  dmBool matched = dmFALSE;

  char* option = NULL;
  int     ndx;         /* index of matching option                        */
  long nmatch;         /* number of matches found                         */
  long minsize;        /* minimum size of option string for partial match */
  long sizeOfUnsignedChar = 1;

  /* Define tool options array */
  toolopts = ar_alloc_c( 10, ntoolopts );
  sprintf( toolopts[0], "ALL");
  sprintf( toolopts[1], "ARRAY");
  sprintf( toolopts[2], "BINS");
  sprintf( toolopts[3], "BLOCKS");
  sprintf( toolopts[4], "CLEAN");
  sprintf( toolopts[5], "COLS");
  sprintf( toolopts[6], "COMMENTS");
  sprintf( toolopts[7], "COUNTS");
  sprintf( toolopts[8], "DATA");
  sprintf( toolopts[9], "FULL");
  sprintf( toolopts[10], "HEADER");
  sprintf( toolopts[11], "IMAGE");
  sprintf( toolopts[12], "KEYS");
  sprintf( toolopts[13], "RAW");
  sprintf( toolopts[14], "REG");
  sprintf( toolopts[15], "STRUCT");
  sprintf( toolopts[16], "SUBSPACE");

  if ( rows )
    dmlistParseRange( rows, &info->minrow, &info->maxrow );

  if ( info->minrow <= 0 ) 
  {
    fprintf( stderr, "Error: row number must be 1 or more: setting minrow to 1\n" );
    info->minrow = 1;
  }

  if ( cells )
    dmlistParseRange( cells, &info->mincell, &info->maxcell );	

  if ( info->mincell <= 0 ) 
  {
    fprintf( stderr, "Error: cell number must be 1 or more: setting mincell to 1\n" );
    info->mincell = 1;
  }

  /* Convert option list to upper case */
  cs_upper( opt );

  /* Split option list into tokens */
  optlist = cs_parse_comma_list( opt, &nopts );

  for (ii=0; ii<nopts; ii++)
  {
    option = optlist[ii];

    if ( option == NULL )
      { continue; } /* skip empty options */

    cs_advptr( &option );              /* skip leading white space */

    nmatch  = 0;           /* number of matches found                         */
    minsize = 3;           /* minimum size of option string for partial match */
    matched = dmFALSE;     /* reset flag                                      */

    ndx = dmu_option_match( toolopts, ntoolopts, option, minsize, &nmatch );

    if ( nmatch == 1 )
    {
      matched = dmTRUE;

      switch ( ndx )
      {
        case 0:  /* ALL */
	{
	  info->PrintAll = dmTRUE;
	  info->mode |= DMA_MODE_COPY_ALL;
	  break;
        }
        case 1:  /* ARRAY */
        {
	  info->PrintArray = dmTRUE;
	  break;
        }
        case 2:  /* BINS */
        {
	  info->PrintBins = dmTRUE;
	  break;
	}
        case 3:  /* BLOCKS */
	{
	  info->PrintBlocks = dmTRUE;	    
	  break;
	}
        case 4:  /* CLEAN */
        {
	  info->PrintClean = dmTRUE;
	  break;
	}
        case 5:  /* COLS */
        {
          info->PrintCols = dmTRUE;
	  break;
        }
        case 6:  /* COMMENTS */
        {
          /* Check that this is not a name=value option. */
          /* which belongs to the ASCII kernel            */
          if ( strchr( option, '=' ) )
          {
            matched = dmFALSE;
          }
          else
          {
            info->PrintComment = dmTRUE; 
          }
	  break;
        }
        case 7:  /* COUNTS */
	{
	  info->PrintCounts = dmTRUE;
	  break;
	}
        case 8:  /* DATA */
        {
	  info->PrintData = dmTRUE;
	  break;
	}
        case 9:  /* FULL */
	{
	  /* Check that this is not a name=value option. */
	  /* which belongs to the Block layer            */
	  if ( strchr( option, '=' ) )
	  {
	    matched = dmFALSE;
	  }
	  else
	  {
	    info->PrintBlocks  = dmTRUE;
	    info->PrintCols    = dmTRUE;
	    info->PrintComment = dmTRUE; 
	    info->PrintData    = dmTRUE;
	    info->PrintFull    = dmTRUE;
	    info->PrintHeader  = dmTRUE;
	    info->PrintSubspace = dmTRUE;
	  }
	  break;
	}
        case 10: /* HEADER */
        {
	  /* Check that this is not a name=value option. */
	  /* which belongs to the FITS kernel            */
	  if ( strchr( option, '=' ) )
	  {
	    matched = dmFALSE;
	  }
	  else
	  {
	    info->PrintHeader  = dmTRUE;
	    info->PrintComment = dmTRUE;
	  } 
	  break;
	}
        case 11: /* IMAGE */
	{
	  info->mode |= DMA_MODE_FORCE_IMAGE;
	  break;
	}
        case 12: /* KEYS */
	{
	  info->PrintHeader  = dmTRUE;
	  break;
	}
        case 13: /* RAW */
	{
	  info->PrintRaw = dmTRUE;
	  break;
	}
        case 15: /* STRUCT */
	{
	  info->PrintHeader = TRUE;
	  info->PrintCols   = TRUE;
	  info->PrintSubspace=TRUE;
	  break;
	}
        case 16: /* SUBSPACE */
	{
	  info->PrintSubspace = dmTRUE;
	  break;
	}
        case 14: /* REG removed 6/19/2008 */ 
        default:
	  break;

      } /* end switch      */
    }  /* end if one match */

    if ( !matched )
    {
      /* Not a matched option.. */
      if ( !tmpstr )
      {
        // Code Change  Bug 10152  November 13, 2007
	//   Need to allocate space the length of opt plus
        //   one char for the null pointer
        tmpstr = calloc( (strlen(opt) + 1), sizeOfUnsignedChar );
	strcpy( tmpstr, option );
      }
      else
      {
	strcat( tmpstr, ",");
	strcat( tmpstr, option);
      }
    }
  }   /* end for        */

  /* Default: */
  if ( cs_is_blank( opt ) ) 
    info->PrintBlocks = TRUE;

  info->PrintAny = ( info->PrintHeader | info->PrintCols | info->PrintSubspace | 
		     info->PrintData   | info->PrintArray );

  if ( info->PrintArray )
    info->PrintData = FALSE;

  /* Transfer unused options to option string */
  if ( tmpstr )
  {
    strcpy( opt, tmpstr );
    free( tmpstr );
  }
  else
  {
    *opt = '\0';
  }

  /* Free memory */
  ar_free_cp( optlist, nopts );
  ar_free_c( toolopts );
}



/***************************************************************/
/*                                                             */
/* Function:   dmaPixelNoToPixel                               */
/*                                                             */
/* Description:                                                */
/*  For an array, map linear pixel order to n-dim pixel vector */
/*                                                             */
/*   Called from functions dmGetArray_l()                      */
/*                         dmGetArray_d()                      */
/*                         dmGetArray_ul()                     */
/*                         dmlistPrintArrayData()              */
/*                                                             */
/*  Input parameters:                                          */
/*         long   pixelNo                                      */
/*         long  *axes                                         */
/*         long   naxes                                        */
/*                                                             */
/*  Output parameters:                                         */
/*         long  *pixel                                        */
/*                                                             */
/*  Returns:                                                   */
/*           none                                              */
/*                                                             */
/***************************************************************/
void dmaPixelNoToPixel( long pixelNo,    /*  (i) Pixel number in array storage order */
                        long* axes,      /*  (i) Array dimensions */
                        long naxes,      /*  (i) Array dimensionality */
                        long* pixel)     /*  (o) Pixel vector  */ 
{
  long axisNo = 0;   /* Axis number */
  long index  = 0;    /* Index in pixel array */
  long n      = 0;         
  long factor = 0;
  long rem    = 0;

  if ( naxes <= 1 )
  {
    pixel[0] = pixelNo;
  }
  else
  {
    n = pixelNo - 1;

    for ( axisNo = 1; axisNo < naxes; axisNo++ )
    {
      index = axisNo - 1;
      factor = n / axes[index];
      rem    = n - factor * axes[index]; 
      pixel[index] = rem + 1;
      n = factor;
    }

    index = naxes - 1;
    pixel[index] = n + 1;
  }
}


/***************************************************************/
/*                                                             */
/* Function:   dmlistBanner                                    */
/*                                                             */
/* Description:                                                */
/*   Outputs a blank line, a dashed line, information, a       */
/*   second dashed line, and a final blank line                */
/*                                                             */
/*   Called from functions  dmlistBlockListCols()              */
/*                          dmlistPrintHeader()                */
/*                          dmlistPrintRawHeader()             */
/*                          dmlistBlockPrintSubspace()         */
/*                                                             */
/*  Input parameters:                                          */
/*        char *buf                                            */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           none                                              */
/*                                                             */
/***************************************************************/
void dmlistBanner( char* buf )
{
  dmlistSkip();
  dmlistDashline();
  dmlist_print( buf );
  dmlistDashline();
  dmlistSkip();
}


/***************************************************************/
/*                                                             */
/* Function:   dmlistDashline                                  */
/*                                                             */
/* Description:                                                */
/*   Outputs a dashed line                                     */
/*                                                             */
/*   Called from function dmlistBanner()                       */
/*                                                             */
/*  Input parameters:                                          */
/*          none                                               */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           none                                              */
/*                                                             */
/***************************************************************/
void dmlistDashline( void ) 
{
 dmlist_print( "--------------------------------------------------------------------------------" );
}


/***************************************************************/
/*                                                             */
/* Function:   dmlistFormatAxes                                */
/*                                                             */
/* Description:                                                */
/*   Describe the dimensions of the image (e.g. "512x512" ) in */
/*   string                                                    */
/*                                                             */
/*   Called from function  dmlistFormatType()                  */
/*                                                             */
/*  Input parameters:                                          */
/*       long  *axes                                           */
/*       long   dim                                            */
/*       char  *sizebug                                        */
/*       long   maxlen                                         */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           none                                              */
/*                                                             */
/***************************************************************/

void dmlistFormatAxes( long* axes, long dim, char* sizebuf, long maxlen )
{
  char *ptr        = NULL;
  long  axis_no    = 0;
  const long spare = 6;

  ptr = sizebuf; 

  *ptr++ = '(';

  for ( axis_no = 1; axis_no < dim && ptr - sizebuf + spare < maxlen; axis_no++ )
  {
    ptr += sprintf( ptr, "%ldx", axes[axis_no-1] );
  }
 
  sprintf( ptr, "%ld)", axes[dim-1] );
}



/***************************************************************/
/*                                                             */
/* Function:   dmlistGetColSize                                */
/*                                                             */
/* Description:                                                */
/*   Determine number of bytes to use for width of column      */
/*   printing.                                                 */
/*                                                             */
/*   First we find the number of elements per cell, which is   */
/*   the product of the element dimension and the array size.  */
/*   This needs to be modified to handle generic element types.*/
/*                                                             */
/*   Then we allocate a size per element depending on the data */
/*   type. For integer sizes, we look at the maximum range to  */
/*   see if we can squeeze it into a smaller column.           */
/*   This should be modified to use dmGetDisp when that is     */
/*   implemented.                                              */
/*                                                             */
/*   Does not include array size                               */
/*                                                             */
/*   Called from functions  dmlistPrintArrayData()             */
/*                          dmlistPrintData()                  */
/*                                                             */
/*  Input parameters:                                          */
/*     dmDescriptor *dd                                        */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*          long  size                                         */
/*                                                             */
/***************************************************************/
long dmlistGetColSize( dmDescriptor* dd )
{
  dmDataType type;

  long imin  = 0;
  long imax  = 0;
  long ival  = 0;
  long size  = 0;
  long nvals = 0;


  nvals = dmGetElementDim( dd );

  if ( nvals == 0 )
    nvals = 1;

  type = dmGetDataType( dd ); 

  switch( type )
  {
    case dmDOUBLE: 
      size = 20;
      break;
    case dmFLOAT:
      /* currently float is printed just like double */
      size = 20;
      break;
    case dmLONG: case dmSHORT: case dmBYTE:
      size = 10;
      if ( dmDescriptorGetRange_l( dd, &imin, &imax ) )
      {
        ival = abs( imin );

        if ( imax > ival )
         ival = imax;
        if ( ival < 10000 )
          size = 6;  /* Replace by more sophisticated alg. later */
        if ( ival < 100 )
          size = 4;
      } 
      break;
    case dmBIT:
      ival = dmDescriptorGetLength( dd );
      if ( ival > 0 ) 
      {
        nvals = 1;
        size = ival + 2;
      }
      else
      {
        size = 10;
      }
      break;
    case dmTEXT:
      size = dmDescriptorGetLength( dd ) + 2;
      break;

    // Code Change  Bug 11058 October 9, 2008
    //  Added this case for dmBools
    case dmBOOL:
      size = 6;
      break;

    default:
      size = 10; /* Default */
      break;
   }

  size *= nvals;

  return( size );
}



/***************************************************************/
/*                                                             */
/* Function:   dmlistGetDescCompositeName                      */
/*                                                             */
/* Description:                                                */
/*   Get the composite name for a descriptor.                  */
/*   Normally this is just the descriptor name, but for        */
/*   a vector descriptor the name is followed by the component */
/*   names in parens, e.g. "SKY(X,Y)".                         */
/*                                                             */
/*   Called from functions dmlistPrintHeader                   */
/*                         dmlistPrintCol                      */
/*                         dmlistPrintArrayData                */
/*                         dmlistPrintData                     */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor *dd                                    */
/*         char         *name                                  */
/*         long          maxlen                                */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           int EXIT_SUCCESS                                  */
/*                                                             */
/***************************************************************/
int dmlistGetDescCompositeName( dmDescriptor* dd, char* name, long maxlen )
{
  char nbuf[DC_MAXSTR];
  char cbuf[DC_MAXSTR];
  char tmp[DCSIZE+1];

  long *axes  = NULL;
  long  dim   = 0; 
  long  i     = 0;
  long  len   = 0;
  long  naxes = 0;
  int   mode  = DMT_FORCE_ARRAY_PARS;


   dmGetName( dd, nbuf, DC_MAXSTR );
   dim = dmGetElementDim( dd );
   naxes = dmGetArrayDimensions( dd, &axes );
 
   if ( dim > 1 )
   {
     strcpy( cbuf, "" );

     for ( i = 1; i <= dim; i++ )
     {
       dmGetCptName( dd, i, tmp, DCSIZE );
       strcat( cbuf, tmp );

       if ( i < dim )
         strcat( cbuf, "," );
     }

     if ( !strcmp( nbuf, cbuf ) )
       strcpy( nbuf, "" );

     strcat( nbuf, "(" );
     strcat( nbuf, cbuf );
     strcat( nbuf, ")" );
   }

   if ( naxes > 0 )
   {
     len = strlen( nbuf );
     PrintVector_l( axes, 1, naxes, mode, NULL, nbuf + len, DC_MAXSTR - len );
   }

   strncpy( name, nbuf, maxlen );

  free( axes );

  return  EXIT_SUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmlistParseRange                                */
/*                                                             */
/* Description:                                                */
/*   Parse a string specifying a range, as follows:            */
/*      23:48        Return 23, 48    (min,max)                */
/*      23           Return 1, 23     (max)                    */
/*       0           Return 1, 0      (zero)                   */
/*      -1           Return 1, 0      (negative maps to zero)  */
/*     (blank)       Return 1, 0      (blank maps to zero)     */
/*   This is used for selecting the range of rows to print     */
/*                                                             */
/*   Called from function  dmaListInfoSet()                    */
/*                                                             */
/*  Input parameters:                                          */
/*       char  *buf                                            */
/*                                                             */
/*  Output parameters:                                         */
/*       long  *umin                                           */
/*       long  *umax                                           */
/*                                                             */
/*  Returns:                                                   */
/*           none                                              */
/*                                                             */
/***************************************************************/
void dmlistParseRange( char* buf, long* umin, long* umax )
{
  char tmp[DCSIZE];
  long i = 0;

  cs_upper( buf );

  if ( cs_is_blank( buf ) )
  {
    *umin = 1;
    *umax = 0;
  }
  else
    if ( !strcmp( buf, "all" ) )
    {
      *umin = 1;
      *umax = 0;
    }
    else
    {
      i = cs_index_char( buf, ':' );
      if ( i > 1 )
      {
        *umin = atol( cs_get_ss( buf, tmp, 1, i-1 ) );
        *umax = atol( buf+i );
      }
      else
      {
        *umin = 1;
        *umax = atol( buf + i );
       }
    }
}



/***************************************************************/
/*                                                             */
/* Function:   dmlistPrintBins                                 */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*                 dmBlock  *vblock                            */
/*                 char     *input                             */
/*                 long      minrow                            */
/*                 long      maxrow                            */
/*                                                             */
/*  Output parameters:                                         */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*            int 0                                            */
/*                                                             */
/***************************************************************/
int  dmlistPrintBins( dmBlock* vblock, char* input, long minrow, long maxrow )
{
  char    *ptr   = NULL;
  char    *bin   = NULL;
  char    *spec  = NULL;
  char    *name  = NULL;
  dmBlock *block = NULL;

  name = dmStrDup( input );
  ptr = strstr( name, "[bin " );

  if ( !ptr )
    return 0;

  *ptr= 0;
  bin = ptr + 5;
  ptr = strchr( bin, '=' );

  if ( ptr )
  {
    spec = ptr + 1;
    *ptr = 0;
    ptr = strchr( spec, ']' );
    *ptr = 0;
  }
  else
  {
    ptr = strchr( bin, ']' );
    *ptr = 0;
    spec = NULL;
  }
 
  block  = dmBlockOpen( NULL, name );
  dmBinningDisp( block, bin, spec, minrow, maxrow );

  return 0;
}



/***************************************************************/
/*                                                             */
/* Function:   dmlistSkip                                      */
/*                                                             */
/* Description:                                                */
/*   Outputs a blank.                                          */
/*                                                             */
/*   Called from function dmlistBanner()                       */
/*                                                             */
/*  Input parameters:                                          */
/*          none                                               */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           none                                              */
/*                                                             */
/***************************************************************/
void dmlistSkip( void )
{
  dmlist_print( " " );
}


/***************************************************************/
/*                                                             */
/* Function:   dmlist_print                                    */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*   Called from functions:                                    */
/*     dmlistSkip()                 dmlistPrintTransform()     */
/*     dmlistDashline()             dmlistPrintArrayData()     */
/*     dmlistBanner()               dmlistPrintData()          */
/*     dmlistPrintCounts()          dmlistPrintHeader()        */
/*     dmlistDataSetListBlocks()    dmlistPrintRawHeader()     */
/*     dmlistBlockListRawCols()     dmlistBlockPrintSubspace() */
/*     dmlistBlockListCols()        dmlistSubspacePrint()      */
/*                                                             */
/*  Input parameters:                                          */
/*       char  *buf                                            */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           none                                              */
/*                                                             */
/***************************************************************/
void dmlist_print( char* buf )
{
  dmlist_print_function( buf );
}




/***************************************************************/
/*                                                             */
/* Function:   dmTableComponentList                            */
/*                                                             */
/* Description:                                                */
/*   Return list of scalar column component                    */
/*                                                             */
/*   Called from function dmlistBlockListRawCols() in          */
/*     dmlist_col.c                                            */
/*                                                             */
/*  Input parameters:                                          */
/*         dmBlock        *table                               */
/*                                                             */
/*  Output parameters:                                         */
/*         dmDescriptor ***list                                */
/*         long          **dims                                */
/*         long          **colno                               */
/*         long           *ncols                               */
/*                                                             */
/*  Returns:                                                   */
/*          int status                                         */
/*                                                             */
/***************************************************************/
int dmTableComponentList( dmBlock* table, dmDescriptor*** list,
                          long** dims, long** colno, long* ncols )
{
  dmDescriptor **vlist = NULL;
  dmDescriptor **localList = NULL;

  long  nvcols = 0;
  long  n      = 0;
  long  dim    = 0;
  long  cpt    = 0;
  long  v      = 0;
  long  entry  = 0;
  long *localDims = NULL;
  long *localCol  = NULL;


  if ( !ncols || !colno || !dims || !list ) 
    return dmFAILURE;

  *ncols = 0;
  *colno = NULL;
  *dims  = NULL;
  *list  = NULL;
  vlist  = dmTableOpenColumnList( table, &nvcols );

  if ( !vlist )
    return dmFAILURE;

  for ( v = 0; v < nvcols; v++ )
  {
    dim = dmGetElementDim( vlist[v] );
    n += dim;
  }	

  *ncols = n;
 
  localList = malloc( n * sizeof( dmDescriptor* ) );
  if ( !localList )
    return dmFAILURE;

  localDims = malloc( n * sizeof( long ));
  if ( !localDims )
    return dmFAILURE;

  localCol= malloc( n * sizeof( long ));
  if ( !localCol )
    return dmFAILURE;

  for ( v = 0; v < nvcols; v++ )
  {
    dim = dmGetElementDim( vlist[v] );

    for ( cpt = 1; cpt <= dim; cpt++ )
    {
      localList[entry] = dmGetCpt( vlist[v], cpt );
      localDims[entry] = dim;
      localCol[entry]  = v+1;

      if ( ++entry > n )
        return dmFAILURE;
    }
  }

  free( vlist );

  *list  = localList;
  *dims  = localDims;
  *colno = localCol;
 
  return dmSUCCESS;
}
