/*                                                                
**  Copyright (C) 1999,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/*****************************************************************/
/*                                                               */
/*   dmlist_range.c                                              */
/*                                                               */
/*   Functions used by dmlist to extract range information.      */
/*   These functions were formerly in dmtlib.                    */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include "dmtlib.h"
#include "dmtcommon.h"
#include "dmlist.h"

static int dmlistRangeSpecialCases( dmDescriptor* dd, double* dmin, double* dmax );


/***************************************************************/
/*                                                             */
/* Function:   dmlistGetRange                                  */
/*                                                             */
/* Description:                                                */
/*   Make a string giving the ranage for a descriptor          */
/*                                                             */
/*   Called from function dmlistPrintCol()                     */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor *dd                                    */
/*                                                             */
/*  Output parameters:                                         */
/*         char         *range                                 */
/*                                                             */
/*  Returns:                                                   */
/*           int EXIT_SUCCESS                                  */
/*                                                             */
/***************************************************************/
int dmlistGetRange( dmDescriptor* dd, char* range )
{
  char dataType[DCSIZE+1];
  char cmin[DCSIZE+1];
  char cmax[DCSIZE+1];
  char dt = '\0';

  unsigned long umin = 0;
  unsigned long umax = 0;

  double dmin  = 0.0;
  double dmax  = 0.0;
  double max_reasonable = 1.0E200;

  long   idmin  = 0;
  long   idmax  = 0;
  int    ok     = 0;
  int    ok_min = 0;
  int    ok_max = 0;

  strcpy( range, " " );
  dmaGetDataTypeName( dd, dataType, DCSIZE );
  dt = *dataType;  /* The first char of dataType gives the general type */

  strcpy( range, "-" );

  switch( dt )
  {
    case 'I':
      if ( dmDescriptorGetRange_l( dd, &idmin, &idmax ) )
        sprintf( range, "%ld:%ld", idmin, idmax );
      break;

    case 'U':
      if ( dmDescriptorGetRange_ul( dd, &umin, &umax ) )
        sprintf( range, "%lu:%lu", umin, umax );
      break;

    case 'R':
      strcpy(cmin, "-Inf");
      strcpy(cmax, "+Inf");

      ok = dmDescriptorGetRange_d( dd, &dmin, &dmax );
      ok_min = (fabs( dmin ) < max_reasonable);
      ok_max = (fabs( dmax ) < max_reasonable);

      if ( !ok || !ok_min || !ok_max )
      {
        ok = dmlistRangeSpecialCases( dd, &dmin, &dmax );
      }

      if ( ok || ok_min )
        Print_d( dmin, cmin );

      if ( ok || ok_max )
        Print_d( dmax, cmax );

      sprintf( range, "%s:%s", cmin, cmax );
      break;

    default:
      strcpy( range, " " );
  }

  return EXIT_SUCCESS;
}
 

/***************************************************************/
/*                                                             */
/* Function:   dmlistRangeSpecialCases                         */
/*                                                             */
/* Description:                                                */
/*   This routine overwrites the default values of the range   */
/*   for the special cases of TIME and EQPOS columns.          */
/*   For TIME, we read the GTI and set the range to be the     */
/*   extreme ends of the GTI. For EQPOS, the range is -360 to  */
/*   360. (RA and Dec nominal acceptable ranges in degrees).   */
/*                                                             */
/*   Called from function dmlistGetRange()                     */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor *dd                                    */
/*                                                             */
/*  Output parameters:                                         */
/*         double  *dmin                                       */
/*         double  *dmax                                       */
/*                                                             */
/*  Returns:                                                   */
/*           int dmSUCCESS or dmFAILURE                         */
/*                                                             */
/***************************************************************/
static int dmlistRangeSpecialCases( dmDescriptor* dd, double* dmin, double* dmax )
{
  dmDescriptor *dsd   = NULL;
  dmBlock      *block = NULL;
  char name[DCSIZE+1];

  double *tmin = NULL;
  double *tmax = NULL;
  long    ngti = 0;


  dmGetName( dd, name, DCSIZE );

  if ( cs_eq( name, "TIME" ) )
  {
    block = dmDescriptorGetBlock( dd );
    dsd = dmSubspaceColRead_d( block, "TIME", &tmin, &tmax, &ngti );

    if ( ngti == 0 )
      return dmFAILURE;

    *dmin = tmin[0];
    *dmax = tmax[ngti-1];

    free( tmin );
    free( tmax );
    return dmSUCCESS;
  }
  else
    if ( cs_eq( name, "EQPOS" ) )
    {
      *dmin = -360.0;
      *dmax =  360.0;
      return dmSUCCESS;
    }

  return dmFAILURE;  /* not really a failure */
}


