/*                                                                
**  Copyright (C) 1999-2002,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/*****************************************************************/
/*                                                               */
/*   dmlist_ss.c                                                 */
/*                                                               */
/*   Functions used by dmlist to obtain subspace information.    */
/*   These functions were formerly in dmtlib.                    */
/*                                                               */
/*                                                               */
/*****************************************************************/
#include "dmtlib.h"
#include "dmtcommon.h"
#include "dmlist.h"


static int dmlistSubspacePrint( dmDescriptor* dd, long tab, char* buf, long maxlen, dmBool tprint, int f );

static int dmlist_ignore_ss( dmDescriptor* ddp );

void dmpRegionArea( dmDescriptor* dd, regRegion* region, double* rrarea, double* f, int v );


/***************************************************************/
/*                                                             */
/* Function:   dmlistBlockPrintSubspace                        */
/*                                                             */
/* Description:                                                */
/*   Summarize the data subspace                               */
/*                                                             */
/*   Called from function dmListBlock()                        */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmBlock  *block                                     */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           int EXIT_SUCCESS                                  */
/*                                                             */
/***************************************************************/
int dmlistBlockPrintSubspace( dmBlock* block )
{
  dmDescriptor *dd = NULL;
  dmBool tprint = dmFALSE;

  char name[DCSIZE+1];
  char dataType[DCSIZE+1];
  char buf[DC_MAXSTR];
  char tname[DCSIZE];

  long *flag  = NULL ;
  long  ncols = 0;
  long  colno = 0;
  long  tcol  = 0;
  long  tab   = 0;
  long  cptno = 0;
  long  ncpts = 0;
  long  tncols= 0;
//  int   s     = 0;

  dmBlockGetName( block, name, DCSIZE );
  ncols = dmBlockGetNoSubspaceCols( block );
  ncpts = dmBlockGetNoSubspaceCpts( block );
  flag = calloc( ncols, sizeof( long ));

  for ( cptno = 1; cptno <= ncpts; cptno++ ) 
  {
    dmBlockSetSubspaceCpt( block, cptno );

    for ( colno = 1; colno <= ncols; colno++ ) 
    {
      dd = dmBlockGetSubspaceColNo( block, colno );

      if ( !dmlist_ignore_ss( dd ) ) 
        flag[colno-1] = 1;    
    }
  }

  for ( colno = 1; colno <= ncols; colno++ ) 
  { 
    if ( flag[colno-1] ) tncols++;
  }
 
  sprintf( buf, "Data subspace for block %s: Components: %ld Descriptors: %ld ", name, ncpts, tncols );
  dmlistBanner( buf );

  for ( cptno = 1; cptno <= ncpts; cptno++ )
  {
    sprintf( buf, " --- Component %ld --- ", cptno );
    dmlist_print( buf );
    dmBlockSetSubspaceCpt( block, cptno );
    tcol = 0;

    for ( colno = 1; colno <= ncols; colno++ )
    {
      if ( !flag[colno-1] )
        continue;

      dd = dmBlockGetSubspaceColNo( block, colno );
      dmGetName( dd, name, DCSIZE );
    
      ++tcol;
      strcpy( tname, " " );
      dmSubspaceColGetTableName( dd, tname, DCSIZE );
      tprint = !cs_is_blank( tname );
      dmaGetDataTypeName( dd, dataType, DCSIZE );
      tab = sprintf( buf, "%4ld %-20s %-20s", tcol, name, dataType );

      if ( *dataType== 'I' || !strcmp( dataType, "Byte" ) )
      { 
        if ( tprint )
        {
          sprintf( buf+tab, "TABLE %s", tname );    
          dmlist_print( buf );
          cs_blank( buf, tab );
        }
        dmlistSubspacePrint( dd, tab, buf, DC_MAXSTR, tprint, 0 );
      }
      else
        if ( *dataType == 'R' )
        {
          if ( tprint )
          {
            sprintf( buf+tab, "TABLE %s", tname );    
            dmlist_print( buf );
            cs_blank( buf, tab );
          }  
          dmlistSubspacePrint( dd, tab, buf, DC_MAXSTR, tprint, 1 );
        }
        else
        {
          dmlist_print( buf );   
        }
    }
  }

  free( flag );
  return EXIT_SUCCESS;
}



/***************************************************************/
/*                                                             */
/* Function:   dmlistSubspacePrint                             */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*   Called from function dmlistBlockPrintSubspace()           */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor *ddp                                   */
/*         long          tab                                   */
/*         char         *buf                                   */
/*         long          maxlen                                */
/*         dmBool        tprint                                */
/*         int           ftype                                 */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           int 1                                             */
/*                                                             */
/***************************************************************/
static int dmlistSubspacePrint( dmDescriptor* ddp, long tab, char* buf, long maxlen, dmBool tprint, int ftype )
{
  dmDescriptor *dd = NULL;

  char  cmin[DCSIZE+1];
  char  cmax[DCSIZE+1];
  char  region[DC_MAXSTR+1] = "";
  char  cptname[DCSIZE];
//  char  name[DCSIZE];
  char *ptr  = NULL;
  char *sptr = NULL;
  char  c    = '\0';

  double *dval1 = NULL;
  double *dval2 = NULL;

  long *ival1 = NULL;
  long *ival2 = NULL;
  long  n     = 0;
//  long  size  = maxlen;
  long  nvals = 0;
  long  i     = 0;
  long  vcpt  = 0;
  long  dim   = 0;
  int   ok1   = 0;
  int   ok    = 0;


  ptr = buf + tab;

  sptr = ptr;
  dim = dmGetElementDim( ddp );
  ok1 = (dmSubspaceColGetRegionString( ddp, region, DC_MAXSTR ) == dmSUCCESS);

  if (ok1 && !cs_is_blank( region ))
  {
    regRegion *region_ptr = dmSubspaceColGetRegion( ddp );
    double     rrarea     = 0.0;
    double     farea      = 0.0;

    strcpy( ptr, region );
    dmlist_print( buf );  

    if ( region_ptr )
    {
      dmpRegionArea( ddp, region_ptr,  &rrarea, &farea, 0 );
      sprintf( ptr, "Field area = %g Region area = %g\n", farea, rrarea );
      dmlist_print( buf );
    }

    if ( dim == 1 && !strcmp( region, "NULL" ))
      return 1;
  } 

  for ( vcpt = 1; vcpt <= dim; vcpt++ )
  {
    dval1 = NULL;
    dval2 = NULL;
    ival1 = NULL;
    ival2 = NULL;
    ptr   = sptr;
    dd    = dmGetCpt( ddp, vcpt );

    dmGetName( dd, cptname, DCSIZE );

    if ( dim > 1 )
    {
      ptr = buf+tab-20;
      ptr += sprintf( ptr, "[%2ld] %-12s   ", vcpt, cptname );
    }

    if ( ftype )
      ok = dmSubspaceColGet_d( dd, &dval1, &dval2, &nvals );
    else
      ok = dmSubspaceColGet_l( dd, &ival1, &ival2, &nvals );

    if ( !ok || nvals == 0 )
    {
      if ( !cs_is_blank( region ))
      {
        n = strlen(region) + ptr - buf;

        if ( n > maxlen )
          strcpy( ptr, "Region" );
        else
          strcpy( ptr, region );

        dmlist_print( buf );
      }
    }
    else
      if ( nvals < 10 && !tprint )
      {
        for ( i = 0; i < nvals; i++ )
        {
          c = ( i < nvals-1 ) ? ',' : ' ' ;   /* Commas until last item */

          if ( ftype )
          {
            Print_d( dval1[i], cmin ); 
            Print_d( dval2[i], cmax );
            ptr += sprintf( ptr, "%s:%s%c", cmin, cmax, c );
          }
          else
          {
            ptr += sprintf( ptr, "%ld:%ld%c", ival1[i], ival2[i], c );
          }
        }
        dmlist_print( buf );
      }
      else
      {
        dmlist_print( buf );
        cs_blank( buf, tab );
        ptr = buf+tab;

        for ( i = 0; i < nvals; i++ )
        {
          if ( ftype )
          {
            Print_d( dval1[i], cmin ); 
            Print_d( dval2[i], cmax );
            sprintf( ptr, "%s:%s", cmin, cmax );
          }
          else
          {
            sprintf( ptr, "%ld:%ld", ival1[i], ival2[i] );
          }
          dmlist_print( buf );
        }
      }
 
    free( dval1 );
    free( dval2 );
    free( ival1 );
    free( ival2 );

    dval1 = NULL;
    dval2 = NULL;
    ival1 = NULL;
    ival2 = NULL;
  }

  return 1;
}



/***************************************************************/
/*                                                             */
/* Function:   dmlist_ignore_ss                                */
/*                                                             */
/* Description:                                                */
/*                                                             */
/*                                                             */
/*   Called from function dmlistBlockPrintSubspace()           */
/*                                                             */
/*                                                             */
/*  Input parameters:                                          */
/*         dmDescriptor *ddp                                   */
/*                                                             */
/*                                                             */
/*  Output parameters:                                         */
/*           none                                              */
/*                                                             */
/*  Returns:                                                   */
/*           int 0 or 1                                        */
/*                                                             */
/***************************************************************/
static int dmlist_ignore_ss( dmDescriptor* ddp )
{
  char region[DC_MAXSTR+1] = "";
  char cptname[DCSIZE];

  dmSubspaceColGetRegionString( ddp, region, DC_MAXSTR );
  dmGetName( ddp, cptname, DCSIZE );

  if ( *region && cs_uceq( cptname, "#row" ) && cs_uceq( region, "DEFAULT" ))
   return 1;

 return 0;
}
