/*                                                                
**  Copyright (C) 1998-2000,2007  Smithsonian Astrophysical Observatory 
*/                                                                

/*                                                                          */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 3 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/*  GNU General Public License for more details.                            */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License along */
/*  with this program; if not, write to the Free Software Foundation, Inc., */
/*  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.             */
/*                                                                          */


/*****************************************************************/
/*                                                               */
/*   dmtlist_ui.c                                                */
/*                                                               */
/*   Functions used for dmlist-lite command line/user interface  */
/*                                                               */
/*                                                               */
/*****************************************************************/

#include <stdio.h>
#include <readline/readline.h>
#include "dmnutils.h"
#include "dmtlib.h"
#include "dmtcommon.h"
#include "dmtlist_ui.h"


/***************************************************************/
/*                                                             */
/* Function:   dmGetdmlistInput                                */
/*                                                             */
/* Description:                                                */
/*   separates input line arguments                            */
/*                                                             */
/*   Called from main()                                        */
/*                                                             */
/*  Input parameters:                                          */
/*    int     argc                                             */
/*    char  **argv                                             */
/*    int     tool                                             */
/*                                                             */
/*  Output parameters:                                         */
/*    char   *inName                                           */
/*    char   *outName                                          */
/*    char   *inOptions                                        */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*    int  status                                              */
/*                                                             */
/***************************************************************/
int dmGetdmlistInput( int argc, char **argv, dmtools_params *params)
{
  int   ii = 0;
  int   optLength = 0;
  int   inLength  = 0;
  int   status    = 0;
  int   ArgCount  = argc;
  int  *debug     = &params->verbose;

  char  prompt[DC_MAXSTR];
  char  options[DC_MAXSTR];
  char  temp_in[DC_MAXSTR];

  char *inName    = params->infile;
  char *inOptions = params->option;
  char *outName   = params->outfile;
  char *rows      = params->rows;
  char *cells     = params->cells;
  char *optString = NULL;
  char *rdline    = NULL;
  char *prompt1   = "Input dataset/block specification: ";
  char *prompt3   = "Options: ";

  dmBool got_inName   = dmFALSE;
  dmBool got_outName  = dmFALSE;
  dmBool got_option   = dmFALSE;
  dmBool flag_inName  = dmFALSE;
  dmBool flag_outName = dmFALSE;
  dmBool flag_option  = dmFALSE;

  options[0]  = '\0';
  optLength   = strlen(inOptions);


  // Check argument count in command line. If count is
  // less than 3, then information is missing which will
  // have to be entered through keyboard
#if 0 
  if (ArgCount < 3)
  {
    // set up to request keyboard input

    if ( dev_tty == NULL )
    {
      dev_tty = fopen("/dev/tty", "r" );
      rl_instream = dev_tty;
      rl_outstream = stderr;
    }
  }
#endif

  // Get input file name from keyboard input
  switch ( ArgCount )
  {
    case 1:
    {
      // Only the tool name entered in command
      // Need to prompt for input and options

      strcpy(prompt, prompt1);
      rdline = readline(prompt);

      if (rdline)
      {
        strncpy(inName, rdline, DC_MAXSTR);
        free (rdline);
      }

      strcpy(prompt, prompt3);
      rdline = readline(prompt);

      if (rdline)
      {
        strncpy(inOptions, rdline, DC_MAXSTR);
        free(rdline);
      }

      break;
    }

    case 2:
    case 3:
    default:
    {
      // 2 or more entries on command line
      // DEFAULT ASSUMPTION: <tool> <inName> <opt or outname> < opt or outname>
      //  read first param

      for (ii  = 1; ii < argc; ii++ )
      {
        if ( argv[ii] )
          strncpy(temp_in, argv[ii], DC_MAXSTR );

        cs_upper(temp_in);

        if ((optString = strstr(argv[ii], "=")))
        {
          // designated parameter - which one?
          getDesignatedParameter(temp_in, &flag_inName, &flag_outName, &flag_option);
        }

        if ( optString && (flag_inName || flag_outName || flag_option))
	{
          if ( flag_inName )
	  {
            strncpy(inName, optString+1, DC_MAXSTR);
            got_inName = dmTRUE;
            flag_inName = dmFALSE;
          }
          else
	  {
	    if (flag_outName )
	    {
              strncpy(outName, optString+1, DC_MAXSTR);
              got_outName = dmTRUE;
              flag_outName = dmFALSE;
            }
            else
	    {
              // assume option
              optString = strstr(temp_in, "OPT=");

              if (optString)
                inLength = strlen(optString+4);
              else
                inLength = strlen(temp_in);

	      if (inLength > 0 && ((inLength + optLength + 1) < DC_MAXSTR))
              {
                if (!optString)
                  strcat(options, temp_in);
                else
                  strcat(options, optString+4 );

                if (optLength > 0)
                  strcat(inOptions, ",");

                strcat(inOptions, options);

                optLength  = strlen(inOptions);
                options[0] = '\0';
              }
              got_option = dmTRUE;
              flag_option = dmFALSE;
            }
	  }
        }
        else
        {
          int param_switch = ii;

          if ( !got_inName)
	  {
            // first non-designated param is input file name
            param_switch = 1;
          }
          else
	  {
            // got input - next non-designated param is
            // option
            param_switch = 3;  // default case too
          }

          switch (param_switch)
	  {
	    case 1:
            // assume this is input file name
            strncpy(inName, argv[ii], DC_MAXSTR );
            got_inName = dmTRUE;
            break;

	    case 2:
            // assume this is output file name
            strncpy(outName, argv[ii], DC_MAXSTR );
            got_outName = dmTRUE;
            break;

	    case 3:
	    default:
	    {
              // assume this is option name
              inLength = strlen(argv[ii]);

	      if (inLength > 0 && ((inLength + optLength + 1) < DC_MAXSTR))
              {
                strcat(options, argv[ii] );

                if (optLength > 0)
                  strcat(inOptions, ",");

                strcat(inOptions, options);

                optLength  = strlen(inOptions);
                options[0] = '\0';
              }

              got_option = dmTRUE;
              break;
            }
          }// end switch (param_switch)
        }  
      } // end for loop

      // All parameters read - prompt for any missing item
      if (!got_inName)
      {
        strcpy(prompt, prompt1);

        rdline = readline(prompt);

        if (rdline)
        {
          strncpy(inName, rdline, DC_MAXSTR);
          free (rdline);
        }
      }

      if (!got_option)
      {
        strcpy(prompt, prompt3);

        rdline = readline(prompt);

        if (rdline)
        {
          strncpy(inOptions, rdline, DC_MAXSTR);
          free (rdline);
        }
      }

      break;
    } // end case 3/default
   } // end switch


  // Verify option information
  dmlistCheckOptions(inOptions, rows, cells, debug);

  // Verify input information
  status = dmValidatedmlistInfo(inName, outName, inOptions);


  return status;
}


/***************************************************************/
/*                                                             */
/* Function:   dmValidatedmlistInfo                            */
/*                                                             */
/* Description:                                                */
/*   validates information obtained from command line call     */
/*                                                             */
/*   Called from main()                                        */
/*                                                             */
/*  Input parameters:                                          */
/*    int     argc                                             */
/*    char  **argv                                             */
/*    int     tool                                             */
/*                                                             */
/*  Output parameters:                                         */
/*    char   *inName                                           */
/*    char   *outName                                          */
/*    char   *inOptions                                        */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*    int  status                                              */
/*                                                             */
/***************************************************************/
int dmValidatedmlistInfo( char *inName, char *outName, char *inOptions)
{
  int    status    = 1;


  if ( cs_is_blank ( inName ))
    status = 0;
  else
  {
    // insure that it is single file name, not a list
    //  remove any leading or trailing blanks

    cs_trim(inName);

    if ( inName[0] == '@' )
      status = 0;
#if 0
    else
    {
      // Be sure that this is a real filename and not
      // an option or outfile string
      optString = strstr(inName,"=");

      if (optString)
        status = 0;
     }
#endif
  }

  return status;
}



/***************************************************************/
/*                                                             */
/* Function:   dmlistCheckOptions                              */
/*                                                             */
/* Description:                                                */
/*   separates options string into components                  */
/*                                                             */
/*   Called from main()                                        */
/*                                                             */
/*  Input parameters:                                          */
/*    char   *inOptions                                        */
/*                                                             */
/*  Output parameters:                                         */
/*    char   *rows                                             */
/*    char   *cells                                            */
/*    intr   *debug                                            */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*         none                                                */
/*                                                             */
/***************************************************************/
void  dmlistCheckOptions(char *inOptions, char *rows, char *cells, int *debug)
{
  char *tag = NULL;
  int   inOptLen = 0;


  tag = strstr(inOptions, "ROWS");

  if ( tag )
  {
    checkRowsOption(tag, rows);
  }

  tag = strstr(inOptions, "CELLS");

  if ( tag )
  {
    checkCellsOption(tag, cells);
  }

  tag = strstr(inOptions, "VERBOSE");

  if (!tag)
    tag = strstr(inOptions, "VER");

  if ( tag )
  {
    checkVerboseOption(tag, debug);
  }

   // remove any excess blanks
   cs_trim(inOptions);
 
   inOptLen = strlen(inOptions);

   // remove any trailing comma
   if (inOptions[--inOptLen] == ',')
     inOptions[inOptLen] = '\0';   
}


/***************************************************************/
/*                                                             */
/* Function:   checkRowsOption                                 */
/*                                                             */
/* Description:                                                */
/*   gets string info entered as rows option                   */
/*                                                             */
/*   Called from dmlistCheckOptions()                          */
/*                                                             */
/*  Input parameters:                                          */
/*    char   *option                                           */
/*                                                             */
/*  Output parameters:                                         */
/*    char   *rows                                             */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*         none                                                */
/*                                                             */
/***************************************************************/
void checkRowsOption(char *option, char *rows)
{
  char *tag = NULL;
  char *comma = NULL;
  char  cBlank = ' ';
  int   paramLen = strlen(option);
  int   ii = 0;
 

  if ((tag = strstr(option, "ROWS=")) )
  {
    tag = strstr(option, "=");

    comma = strstr(option, ",");

    if (comma)
      paramLen = comma - (tag+1);

    strncpy(rows, (tag+1), paramLen);
    *(rows+paramLen) = '\0';
  }
  else
    *rows = '\0';


  if ( comma )
    paramLen = comma - option + 1;

   // remove from options string
   for ( ii = 0; ii < paramLen; ii++)
   {
     *(option+ii) = cBlank;
   }
}


/***************************************************************/
/*                                                             */
/* Function:   checkCellsOption                                */
/*                                                             */
/* Description:                                                */
/*   gets string info entered as cells option                  */
/*                                                             */
/*   Called from dmlistCheckOptions()                          */
/*                                                             */
/*  Input parameters:                                          */
/*    char   *option                                           */
/*                                                             */
/*  Output parameters:                                         */
/*    char   *cells                                            */
/*                                                             */
/*                                                             */
/*  Returns:                                                   */
/*         none                                                */
/*                                                             */
/***************************************************************/
void checkCellsOption(char *option, char *cells)
{;
  char *tag = NULL;
  char *comma = NULL;
  char  cBlank = ' ';
  int   paramLen = strlen(option);
  int   ii = 0;
 
  if ( (tag = strstr(option, "CELLS=")) )
  {
    tag = strstr(option, "=");

    comma = strstr(option, ",");

    if (comma)
      paramLen = comma - (tag+1);

    strncpy(cells, (tag+1), paramLen);
    *(cells+paramLen) = '\0';
  }
  else
    *cells = '\0';

  if ( comma )
    paramLen = comma - option + 1;

   // remove from options string
   for ( ii = 0; ii < paramLen; ii++)
   {
     *(option+ii) = cBlank;
   }
}



