// 
//  Copyright (C) 2002,2003,2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include <sys/types.h>
#include <regex.h>

#include "AhelpDBTable.hh"
#include "AhelpUtils.hh"



/*
**
** Desc: Constructor
**
**
** Pre: inPath location of the table
**      inCreateFlag if 1 create table(delete old one)
**                      0 open existing one
**
** Post: file inPath is opened and internal data structures 
**        are initialized
**
**
*/


AhelpDBTable::AhelpDBTable(const AhelpStrings& inPath,int inCreateFlag) 
{
  AhelpStrings tmpType;


  rowCount=0;
  colCount=0;
  rowSize=0;
  sortedFlag=0;
  
  path = inPath;

  

  if ( inCreateFlag)
    {

      // create new table

      file.open(path.chars(),fstream::in | fstream::out|fstream::binary|fstream::trunc);
    }
  else
    {
      // open for reading

      file.open(path.chars(),fstream::in | fstream::binary);
    };

  
  if (! file.good() )
    {
      tmpType="Error:  AhelpDBTable::AhelpDBTable - failed to open file '";
      tmpType.append(inPath);
      tmpType.append("'");
      cerr<<tmpType.chars()<<endl;
      throw -1;
    };

  if (! inCreateFlag )
    {
      file >> tmpType;
      
      if ( file.good())
	{
	  // check file format
	  
	  if ( tmpType != FILEID)
	    {
	      file.close();
	      tmpType="Error:  AhelpDBTable::AhelpDBTable - incompatible format of file '";
	      tmpType.append(inPath);
	      tmpType.append("'");
	      cerr<<tmpType.chars()<<endl;
	      throw -1;
	    }
	  else
	    {
	      // read in header

	      file >> rowCount >>colCount>>rowSize>>sortedFlag;
	    };
	};
    };
      
  currRowNo=-1;
  currOffset=-1;
  cachedRowCount=0;
};




/*
**
** Desc: Destructor
**
** Pre: None
**
** Post: object is dealocated
**
**
*/

AhelpDBTable::~AhelpDBTable()
{


};




/*
**
** Desc: Retrieves a row from the table at offset inRowOffset
**
**
**
** Pre: inRowOffset is an offset of a row
**      outFieldList is list of Strings
**
**
**
**
**
** Post: columns of the row at offset inRowOffset are put into
**       outFieldList
**       Returns 1 on success
**               0 or negative on failure
**
**
*/


int AhelpDBTable::GetOffsetRow(long inRowOffset,AhelpStringsList& outFieldList)
{

  int status=MoveToOffset(inRowOffset);

  if ( status > 0 )
    {
      outFieldList = currFieldList;
    };

  return status;

};



/*
**
** Desc: Retrievs inRowNo row from table
**
**
** Pre: inRowNo is number of row to be retrieved
**      outFieldList empty list of Strings
**
** Post: Row inRowNo is retrieved and values of columns 
**       are put into outFieldList
**       Returns 1 on success
**               0 or negative on failure
**
**
*/


int AhelpDBTable::GetRow(long inRowNo,AhelpStringsList& outFieldList)
{
  int status = MoveToRow(inRowNo);

  if ( status > 0 )
    {
      outFieldList = currFieldList;
    };

  return status;
    
};




/*
**
** Desc: Retrieve cell from a table into outCell as Strings
**
**
**
** Pre: inRowNo row number
**      inColNo column number in inRowNo row
**      outCell Strings var
**
**
** Post: retrieves col inColNo from row inRowNo and puts its value
**       in outCell.
**       Returns 1 on success
**               0 or negative on failure
**
**
*/


int AhelpDBTable::GetCell(long inRowNo, int inColNo,AhelpStrings& outCell)
{

  if ( ( inRowNo < 0 ) || (inColNo >=colCount) || (inColNo < 0))
    return 0;

  int status=1;
  AhelpStrings tmpRow;

    
  if ( inRowNo != currRowNo )
    {
      status = MoveToRow(inRowNo);
    };

  if ( status  != 0 )
    {
      outCell=currFieldList.GetIndex(inColNo);
      status =1;
    };

  return status;
  

};




/*
**
** Desc: Retrieve column inColNo from a row at offset inRowOffset
**
**
** Pre: inRowOffset offset of row to be retrieved
**      inColNo number of column at row at offset inRowOffset
**      outCell StringVar
**
** Post: column inColNo at row at offset inRowOffset is retrieved
**       and its value put in outCell.
**       Returns 1 on success
**               0 or negative on failure
**
*/

int AhelpDBTable::GetOffsetCell(long inRowOffset, int inColNo,AhelpStrings& outCell)
{
  int status=1;
  AhelpStrings tmpRow;
  
  if ((inColNo >= colCount)|| (inColNo < 0))
    {
      return 0;
    };

    
  status = MoveToOffset(inRowOffset);

  if ( status  != 0 )
    {
      outCell=currFieldList.GetIndex(inColNo);
      status =1;
    };

  return status;
};


/*
**
** Desc: Finds offset of row that contains inValue in column inColNo
**
**
**
** Pre: inColNo column number
**      inValue value in column inCol
**         
**
**
**
** Post: Returns -1 on error
**               offset of row that has inValue in inColNo on success
**
**
**
*/


long AhelpDBTable::FindOffset(int inColNo,const AhelpStrings& inValue)
{
  
  if (Find(inColNo,inValue) >= 0 )
    {
      return currOffset;
    };

  return -1;

};


/*
**
** Desc: Finds row number that has inValue in column inColNo
**
**
**
** Pre: inColNo column number
**      inValue value in column inCol
**
**
** Post: Post: Returns -1 on error
**               offset of row that has inValue in inColNo on success
**       If table is sorted binary search is performed otherwise whole
**        table is searched
**
**
*/


long AhelpDBTable::Find(int inColNo,const AhelpStrings& inValue)
{


  long left=0,right;
  long indx;
  AhelpStrings tmpValue;
  
  if ( ( inColNo >= colCount) || (inColNo < 0))
    {
      return -1;
    };


  if (( rowSize != 0)  && (sortedFlag) && (inColNo==0))
    {

      // do binary search if table sorted

      right = rowCount-1;
      left  = 0;
      indx  = (left+right)/2;  
      
      while (left <= right)
	{
	  
	  if ( GetCell(indx,inColNo,tmpValue) <=0 )
	    {
	      return -1;
	    };
	  
	  if ( tmpValue > inValue )
	    {
	      
	      right = indx-1;
	    }
	  else if ( tmpValue < inValue )
	    {
	      left = indx+1;
	    }
	  else
	    {
	      while (indx > 0 )
		{
		  if ( GetCell(indx-1,inColNo,tmpValue) <= 0 )
		    {
		      return -1;
		    };
		  
		  if ( tmpValue == inValue)
		    {
		      indx--;
		    }
		  else
		    {
		      break;
		    };
		};
	      break;
	    };
	  
	  indx = (left+right)/2;
	  
	};
      
      if ( left > right)
	{
	  indx=-1;
	};
    }
  else
    {

      //go through all the rows if table not sorted

      for(indx= 0;indx<rowCount;indx++)
	{ 
	  if ( GetCell(indx,inColNo,tmpValue) <=0 )
	    {
	      return -1;
	    };

	  if ( tmpValue == inValue )
	    {
	      break;
	    };
	};

      if ( indx == rowCount )
	{
	  indx=-1;
	};
    };      

  return indx;
};


/*
**
** Desc: Append row to a table
**
**
**
** Pre: inFieldList list of values for corresponding columns
**
**
**
**
**
**
** Post: values from inFieldList are used to construct new row that is
**       appended to the table
**       Returns 1 on success
**               0 or negative on failure
**
**
**
*/


int AhelpDBTable::AddRow(AhelpStringsList& inFieldList)
{
  AhelpStrings tmpRow;

  if (( ! file.good()) || ( MakeRow(inFieldList,tmpRow) <=0 ))
    {
      AhelpPostWarning("AhelpDBTable::AddRow failed");
      return 0;
    };
  
  cachedRowCount++;
  rowSize=0;


  rows.Append(tmpRow);

  return 1;

};




/*
**
** Desc: Append new row to the table with constraint for row length
**
**
**
** Pre: inFieldList list of values for corresponding columns
**      inMaxRowSize max length of table row including field separators
**
**
**
**
**
** Post: Values from inFieldList are used to construct new row. If its
**       length is legal it is appended to the table
**       Returns 1 on success
**               0 or negative on failure
**
**
**
*/


int AhelpDBTable::AddRow(AhelpStringsList& inFieldList, int inMaxRowSize)
{
  AhelpStrings tmpRow;

  if (( ! file.good()) || ( MakeRow(inFieldList,tmpRow,inMaxRowSize) <= 0 ) || (inMaxRowSize <=0))
    {
      AhelpPostWarning("AhelpDBTable::AddRow failed");
      return 0;
    };

  if  ( rowCount != 0 )
    {
      if ( (inMaxRowSize+1) != rowSize)
	{
	  return 0;
	};
    }
  else
    {
      rowSize= inMaxRowSize+1;
    };

  cachedRowCount++;

  rows.Append(tmpRow);

  return 1;

};


/*
**
** Desc: Delete conten of the table
**
**
**
** Pre: None
**
**
**
**
**
**
** Post: Table is empty, table header is initialized
**       Returns 1 on success
**               0 or negative on failure
**
**
**
**
*/


int AhelpDBTable::Clear()
{
  file.close();
  file.open(path.chars(),fstream::in | fstream::out |fstream::trunc| fstream::binary);
  cachedRowCount=0;
  colCount=0;
  rowCount=0;
  rowSize=0;
  sortedFlag=0;
  currFieldList.RemoveAll();
  currRowNo=-1;
  currOffset=-1;
  WriteHeader(rowCount,colCount,rowSize,sortedFlag);
  rows.RemoveAll();


  if ( file.good() )
    return 1;

  AhelpPostWarning("AhelpDBTable::Clear failed");
  
  return 0;
};





/*
**
** Desc: Move internal pointer to row inRowNo
**
**
**
** Pre: inRowNo row number
**
**
**
**
**
**
** Post: Internal pointer is moved to point to row inRowNo
**       Returns 1 on success
**               0 or negative on failure
**
**
**
**
*/

int AhelpDBTable::MoveToRow(long inRowNo)
{
  if (( !file.good()) || (inRowNo < 0 ) || (inRowNo >= (rowCount+cachedRowCount)))
    {
      AhelpPostWarning("AhelpDBTable::MoveToRow failed");
      return 0;
    };

  if ( inRowNo == currRowNo)
    {
      return 1;
    };

  int status = 0;
  AhelpStrings tmpRow;

  // check if table consists of records of constant size: rowSize > 0

  if ( rowSize == 0)
    {
      // table has records of variable length
      
      if ( inRowNo < rowCount )
	{
	  char* tmpBuf = new char[MAXRECORDSIZE];
	  long indx=0;
	  currOffset=HEADERSIZE;
	  file.seekg(currOffset);

	  //Iterate through untill searched one is reached

	  while(file.getline(tmpBuf,MAXRECORDSIZE,RECORDSEPARATOR) )
	    {
	      if ( indx == inRowNo)
		{
		  tmpRow=tmpBuf;
		  currFieldList = tmpRow.Split(FIELDSEPARATOR);
		  currRowNo = inRowNo;
		  break;
		};
	      currOffset += strlen(tmpBuf)+1;
	      indx++;
	    };

	  if ( file.good() )
	    {
	      status = 1;
	    }
	  else
	    {
	      AhelpPostWarning("AhelpDBTable::MoveToRow getline failed");
	    };
	  
	  file.clear(); 
	  delete[] tmpBuf;
	}
      else
	{
	  // Constant size records only in this table
	  
	  if ( inRowNo < (rowCount+cachedRowCount))
	    {

	      //Move directrly to specified row

	      currFieldList = rows.GetIndex(inRowNo-rowCount).Split(FIELDSEPARATOR);
	      currRowNo=inRowNo;
	      file.seekg(0,fstream::end);

	      // calculate offset of current row: what's in the file and whats in the cache

	      currOffset=CalcCacheOffset(inRowNo-rowCount)+file.tellg();
	    };
	  
	  if ( file.good() )
	    {
	      status = 1;
	    }
	  else
	    {
	      AhelpPostWarning("AhelpDBTable::MoveToRow seekg failed");
	    };
	};
    }
  else
    {
      return MoveToOffset(HEADERSIZE+rowSize*inRowNo);
    };


  return status;

};



/*
**
** Desc: Move internal pointer to row at offset inRowOffset
**
**
**
** Pre: inRowOffset - row offset to move to
**
**
**
**
**
**
** Post: internal pointer points to inRowOffset(row at that offset
**       is loaded to cache)
**       Returns 1 on success
**               0 or negative on failure
**
**
**
*/


int AhelpDBTable::MoveToOffset(long inRowOffset)
{
  if ( (! file.good()) || (inRowOffset<HEADERSIZE))
    {
      AhelpPostWarning("AhelpDBTable::MoveToOffset failed");
      return 0;
    };

  if ( inRowOffset == currOffset )
    return 1;
  
  int status=0;
  long endOffset = GetEndOffset();


  if ( endOffset > inRowOffset)
    {
      // row not in the cache

      char* tmpBuf = new char[MAXRECORDSIZE];
      AhelpStrings tmpRow;
      
      // read it from file into cache

      file.seekg(inRowOffset);
      file.getline(tmpBuf,MAXRECORDSIZE,RECORDSEPARATOR) ;
      
      if ( file.good())
	{
	  status = 1;
	  if ( rowSize != 0)
	    {
	      currRowNo=(inRowOffset-HEADERSIZE)/rowSize;
	    }
	  else
	    {
	      currRowNo=-1;
	    };
	  currOffset=inRowOffset;
	  tmpRow=tmpBuf;
	  currFieldList = tmpRow.Split(FIELDSEPARATOR);
	}
      else
	{
	  AhelpPostWarning("AhelpDBTable::MoveToOffset getline failed");
	};
      
      file.clear();
      delete[] tmpBuf;
    }
  else if ( rows.GetLength() )
    {

      // row in the cache

      AhelpStringsListIterator rowIter(&rows);
      long tmpIndx=0;
      rowIter.First();
      
      // iterate throu cache to determine the actual offset

      while(!rowIter.IsDone())
	{
	  if ( endOffset == inRowOffset)
	    {
	      currOffset=inRowOffset;
	      currFieldList = rowIter.GetCurrent().Split(FIELDSEPARATOR);
	      currRowNo = rowCount+tmpIndx;
	      status = 1;
	      break;
	    };
	  endOffset += rowIter.GetCurrent().length()+1;
	  tmpIndx++;
	  rowIter.Next();
	};
    };

  return status;
 

};


/*
**
** Desc: Creates row string
**
**
**
** Pre: inFieldList list of values for corresponding columns of the row
**      outRow  formatted row
**      inMaxRowSize constraint on row length column separator including
**
**
**
**
** Post: Corresponding values from inFieldList are formatted into a string 
**       that represents row with columns separated by FIELDSEPARATOR. 
**       Resulting string is put into outRow. If resulting row string is 
**       shorter than inMaxRowSize,  then extra column of spaces is 
**       padded onto that string on the right.
**       Returns 1 on success
**               0 or negative on failure
**
**
*/


int AhelpDBTable::MakeRow(AhelpStringsList& inFieldList, AhelpStrings& outRow,int inMaxRowSize)
{
  if (( inMaxRowSize > MAXRECORDSIZE ) || (inMaxRowSize < 0 ))
    {
      return 0;
    };

  if ( colCount == 0 )
    {
      colCount = inFieldList.GetLength();
    }
  else if ( colCount != inFieldList.GetLength() )
    {
      return 0;
    };


  outRow.clear();

  AhelpStringsListIterator fieldIter(&((AhelpStringsList&)inFieldList));

  fieldIter.First();

  // build row

  while(!fieldIter.IsDone())
    {
      outRow.append(fieldIter.GetCurrent());

      fieldIter.Next();

      if (! fieldIter.IsDone())
	{
	  outRow.append(FIELDSEPARATOR);
	};
    };

  // pad with extra column if necessary

  if ( inMaxRowSize > 0 )
    {

      if ( outRow.length() > inMaxRowSize)
	{
	  return 0;
	}
      else if ( outRow.length() < inMaxRowSize)
	{
	  outRow.append(FIELDSEPARATOR);
	  outRow.padR(inMaxRowSize - outRow.length());
	};
    };

  return 1;
};



/*
**
** Desc: Writes table's header to the disk
**
**
**
** Pre: inRowCount number of rows in the table
**      inColCount number of columns in the table
**      inRowSize number of characters in the row(including row separator)
**      inSortedFlag 1 if table is sorted, 0 otherwise
**
**
**
** Post: Header is written to a file representing table.
**       Header size is HEADERSIZE. Header format is:
**       inRowCount inColCount inRowSize inSortedFlag
** 
**       Header is padded on the right to HEADERSIZE length
**       Returns 1 on success
**               0 or negative on failure
*/


int AhelpDBTable::WriteHeader(long inRowCount,int inColCount,int inRowSize,int inSortedFlag)

{
  AhelpStrings tmpHeader;
  

  tmpHeader = FILEID;
  tmpHeader.append(" ");
  tmpHeader.append(inRowCount);
  tmpHeader.append(" ");
  tmpHeader.append(inColCount);
  tmpHeader.append(" ");
  tmpHeader.append(inRowSize);
  tmpHeader.append(" ");
  tmpHeader.append(inSortedFlag);
  tmpHeader.padR(HEADERSIZE-tmpHeader.length());

  if (tmpHeader.length() != HEADERSIZE)
    return 0;

  file.seekp(0,fstream::beg);
  file<<tmpHeader.chars();
  file.flush();

  if ( ! file.good() )
    {
      AhelpPostWarning("AhelpDBTable::WriteHeader failed");
      return 0;
    };

  return 1;

};


/*
**
** Desc: Retrieves value of column inColNo from a row at inRowOffset
**       as 'long' type.
**
**
** Pre: inRowOffset offset of the beginning of a row
**      inColNo column number to be retrieved
**      outCell 
**
**
**
**
** Post: a value in column inColNo in row that starts at offset inRowOffset
**       is retrieved as 'long' type value and put into outCell
**       Returns 1 on success
**               0 or negative on failure
**
**
**
*/



int AhelpDBTable::GetOffsetCell(long inRowOffset, int inColNo,long& outCell)
{
  AhelpStrings tmpValue;

  if ( GetOffsetCell(inRowOffset, inColNo,tmpValue) > 0 )
    {
      outCell = atol(tmpValue.chars());
      return 1;
    };


  return 0;

};



/*
**
** Desc: Sorts the rows in the table as strings
**
**
**
** Pre: uniqueColList list of columns checked for uniqeness
**
**
** Post: the table is sorted as list of strings, the columns specified
**       in uniqueColList are checked for uniqueness and rows
**       with duplicates are removed
**       Returns 1 on success
**               0 or negative on failure
**
**
*/


int AhelpDBTable::Sort(AhelpList<int> uniqueColList)
{
  if (( rowCount + cachedRowCount ) <= 0 || (colCount<=0))
    {
      return 1;
    };
  
  // load whole table into memory

  Load();

  rows.Sort(TRUE);

  //remove duplicates

  if ( uniqueColList.GetLength())
    {
      
      AhelpStrings tmpValue,newKey,oldKey;
      AhelpStringsListIterator rowIter(&rows);
      int indx;
      


      rowIter.First();
      
      while(!rowIter.IsDone())
	{
	  newKey.clear();

	  for(indx = 0;indx<uniqueColList.GetLength();indx++)
	    {
	      if ( ( uniqueColList.GetIndex(indx) >= colCount) || (uniqueColList.GetIndex(indx)< 0 ))
		{
		  return 0;
		}
	      else if ( GetCell(rowIter.GetCurrent(),uniqueColList.GetIndex(indx),tmpValue) <=0 )
		{
		  return 0;
		}
	      else
		{
		  newKey.append(FIELDSEPARATOR);
		  newKey.append(tmpValue);
		};
	    };


	  if ( oldKey == newKey)
	    {
	      tmpValue=FIELDSEPARATOR;
	      newKey.Substitute(tmpValue," ",1);
	      //	      cerr<<"AhelpDBTable: removing  duplicate key '"<<newKey<<"' from "<<path<<endl;
	      rowIter.RemoveCurrent();
	      cachedRowCount--;
	    }
	  else
	    {
	      oldKey = newKey;
	    };

	  rowIter.Next();

	};
    };

  sortedFlag=1;

  return 1;
};


/*
**
** Desc: Sort the table
**
**
**
** Pre: None
**
**
**
**
**
**
** Post: Sorts the rows in the table as strings
**       Returns 1 on success
**               0 or negative on failure
**
**
**
**
*/


int AhelpDBTable::Sort()
{
  AhelpList<int> emptyList;
  return  Sort(emptyList);
};



/*
**
** Desc: Writes the table to file
**
**
**
** Pre: None
**
**
**
**
**
**
** Post: Rows from cache are written into file then table header is 
**       written too.
**       Returns 1 on success
**               0 or negative on failure
**
**
**
*/


int AhelpDBTable::Commit()
{

  while(rows.GetLength())
    {  
      file.seekp(0,fstream::end);
      file<<rows.GetFirst().chars()<<RECORDSEPARATOR;
      file.flush();
      if ( ! file.good() )
	{
	  AhelpPostWarning("AhelpDBTable::Commit failed");
	  return 0;
	};
      rows.RemoveTop();
    }; 

  WriteHeader(rowCount+cachedRowCount,colCount,rowSize,sortedFlag);
  rowCount += cachedRowCount;
  cachedRowCount=0;
  return 1;
};

/*
**
** Desc: Get offset of the end of table 
**
**
**
** Pre: None
**
**
**
**
**
**
** Post: Returns size of table in bytes on success,
**               -1 otherwise
**
**
**
**
*/


long AhelpDBTable::GetEndOffset()
{
  file.seekg(0,fstream::end);

  if ( ! file.good())
    {
      AhelpPostWarning("AhelpDBTable::GetEndOffset failed");
      return -1;
    };
  return file.tellg();

};


/*
**
** Desc: Calculates offset of a row without table header 
**
**
**
** Pre: inRowNo number of a row
**
**
**
** Post: Returns offset of the beginning of a row without header 
**                size included on success,
**               -1 otherwise.
**
**
**
**
*/


long  AhelpDBTable::CalcCacheOffset(long inRowNo)
{
  long tmpOffset=0;
  long tmpIndx=0;

  if ( inRowNo >= rows.GetLength())
    return -1;

  AhelpStringsListIterator rowIter(&rows);
  
  rowIter.First();
  
  while(!rowIter.IsDone())
    {
      if ( tmpIndx == inRowNo)
	{
	  break;
	};
      tmpOffset += rowIter.GetCurrent().length()+1;
      tmpIndx++;
      rowIter.Next();
    };

  return tmpOffset;
};


/*
**
** Desc: Loads table into cache
**
**
**
** Pre: None
**
**
** Post: Loads all the rows from the file into internal cache
**       Returns 1 on success
**               0 or negative on failure
**
**
**
**
*/


int AhelpDBTable::Load()
{
  long tmpSize = rows.GetLength();

  if ( rowCount > 0 )
    {
      char* tmpBuf = new char[MAXRECORDSIZE];
      file.seekg(HEADERSIZE);
      while(file.getline(tmpBuf,MAXRECORDSIZE,RECORDSEPARATOR) )
	{
	  if (! file.good() )
	    {
	      AhelpPostWarning("AhelpDBTable::Load failed");
	      return 0;
	    };

	  rows.Append(tmpBuf);
	  cachedRowCount++;
	  rowCount--;
	}; 
      
      for(long indx=0;indx<tmpSize;indx++)
	{
	  rows.Append(rows.GetFirst());
	  rows.RemoveTop();
	};
    };

  file.close();
  file.open(path.chars(),fstream::in | fstream::out |fstream::trunc| fstream::binary);
  rowCount=0;
  
  WriteHeader(rowCount,colCount,rowSize,sortedFlag);
  return 1;
};

int AhelpDBTable::GetCell(const AhelpStrings& inRow,int inColNo, AhelpStrings& outValue)
{
  if (( inColNo < 0 ) || (inColNo >=colCount))
    {
      return 0;
    };

  int startIndx=0,endIndx=0;

  if (( inColNo > 0 ) && (inColNo< (colCount-1)))
    {
      startIndx = inRow.FindChar(FIELDSEPARATOR,inColNo);
      endIndx = inRow.FindChar(FIELDSEPARATOR,inColNo+1);
    }
  else if ( inColNo == 0 )
    {
      startIndx=0;
      endIndx = inRow.FindChar(FIELDSEPARATOR,inColNo+1);
    }
  else 
    {
      startIndx=inRow.FindChar(FIELDSEPARATOR,inColNo);
      endIndx=inRow.length();
    };
    

  outValue = inRow.GetSubStr(startIndx,endIndx);

  return 1;

};


/*
**
** Desc: Calculates max record size and writes table into file
**       padding each record on the right to that max record size
**
**
** Pre: None
**
**
**
**
**
**
** Post: All the rows of table are compared and max lenth is remembered.
**       Then if necessary every record is padded with extra column of 
**       spaces to that max length and written to file.
**       Returns 1 on success
**               0 or negative on failure
**
**
**
*/


int AhelpDBTable::CommitWithMaxRecordWidth()
{

  int maxLength=0;
  int indx;
  AhelpStrings padding;

  if ( rowCount != 0 )
    {
      if ( Load() <=0 )
	{
	  return 0;
	};
    }; 

  // determine max length of record
  
  for(indx=0;indx<rows.GetLength();indx++)
    {
      if ( rows.GetIndex(indx).length() > maxLength )
	{  
	  maxLength = rows.GetIndex(indx).length();
	};
    };
  
  // commit the table into memory

  while(rows.GetLength())
    {  
      file.seekp(0,fstream::end);
      file<<rows.GetFirst().chars();
      if ( rows.GetFirst().length() < maxLength)
	{
	  padding.padR(maxLength-rows.GetFirst().length()-1);
	  file<<FIELDSEPARATOR<<padding;
	  padding.clear();
	};
      file<<RECORDSEPARATOR;

      file.flush();
      if ( ! file.good() )
	{
	  AhelpPostWarning("AhelpDBTable::CommitWithMaxRecordWidth failed");
	  return 0;
	};
      rows.RemoveTop();
    }; 

  rowSize = maxLength+1;

  WriteHeader(rowCount+cachedRowCount,colCount,rowSize,sortedFlag);
  rowCount += cachedRowCount;
  cachedRowCount=0;
  return 1;
};


/*
**
** Desc: Get the value of a column inColNo in row inRowNo as 'long' type
**
**
**
** Pre: inRowNo row number
**      inColNo column whose value is to be retrieved 
**      outCell long value
**
**
**
**
** Post: content of column inColNo in row inRowNo is interpreted as 'long'
**       and assigned to outCell
**       Returns 1 on success
**               0 or negative on failure
**
**
**
*/

int AhelpDBTable::GetCell(long inRowNo, int inColNo,long& outCell)
{


  AhelpStrings tmpValue;

  if ( GetCell(inRowNo, inColNo,tmpValue) > 0 )
    {
      outCell = atol(tmpValue.chars());
      return 1;
    };


  return 0;


};


/*
**
** Desc: Overloaded stream output operator
**
**
**
** Pre: ostr is output stream
**      inTable is AhelpDBTable
**
** Post: Prints content of inTable into ostr stream
**
**
**
**
**
*/

ostream& operator<<(ostream& ostr,AhelpDBTable& inTable)
{
  AhelpStringsList tmpRow;


  for(int indx=0;indx<inTable.GetLength();indx++)
    {
      if ( inTable.GetRow(indx,tmpRow) <= 0 )
	{
	  ostr<<"Failed to read table "<<inTable.path<<endl;
	  break;
	};

      AhelpStringsListIterator fieldIter(&tmpRow);
      ostr << indx<<" "<<inTable.currOffset<< " ";
      for(fieldIter.First();!fieldIter.IsDone();fieldIter.Next())
	{
	  ostr<< fieldIter.GetCurrent()<<"|";
	};
      ostr<<endl;
    };

  return ostr;
};

/*
**
** Desc:  Finds the rows that have value inValue in column inColNo 
**        in the table and retrievs values of column outColNo from these rows
**        into outList. The retrieved values are interpreted as long type.
**
** Pre: inColNo column number to be searched in every row
**      inValue value to be searched in inColNo ( it is interpreted as 
**               regular expression if expressed as '/expression/' or 
**                "/expression/")
**      outColNo column number whose values are to be retrieved
**      outList list of longs
**
**
**
** Post: Finds all rows that have inValue in inColNo column then it takes 
**       content of outColNo column from them, interprets it as long type
**       and adds it to outList.
**       Returns 1 on success
**               0 or negative on failure
**
**
*/


int  AhelpDBTable::Find(int inColNo,const AhelpStrings& inValue,int outColNo,AhelpList<long>& outList)
{
  AhelpStrings tmpString;
  long rowIndx;
  long tmpValue;
  regex_t  regExp;


  if ( (inColNo >=colCount) || (inColNo < 0))
    return 0;

  outList.RemoveAll();


  
  if ( (tmpString = GetRegularExpression(inValue) ) != "")
    {

      // we are searching using regular expression

      if ( regcomp(&regExp,tmpString.chars(),REG_EXTENDED|REG_NOSUB) != 0 )
	{
	  return 0;
	};

      rowIndx=0;

      // go through all the rows in table

      while(rowIndx <GetLength()) 
	{

	  if ( GetCell(rowIndx,inColNo,tmpString) <= 0 )
	    {
	      return 0;
	    }
	  else  if ( !regexec(&regExp,tmpString.chars(),0,NULL,0))
	    {
	      if ( GetCell(rowIndx,outColNo,tmpValue) <= 0 )
		{
		  return 0;
		};
	      
	      outList.Append(tmpValue);
	    };
	  rowIndx++;
	};
      
      regfree(&regExp);
    }
  else
    {
      // find the first row

      if ( (rowIndx=Find(inColNo,inValue)) < 0 )
	{
	  return 0;
	};

      // get all the rest

      do
	{
	  
	  if ( GetCell(rowIndx,outColNo,tmpValue) <= 0 )
	    {
	      return 0;
	    };
	  
	  outList.Append(tmpValue);
	  
	  rowIndx++;
	  
	  if ( rowIndx < GetLength())
	    {
	      if ( GetCell(rowIndx,inColNo,tmpString) <= 0 )
		{
		  return 0;
		}
	      else if ( tmpString != inValue)
		{
		  break;
		};
	    };
	  
	}while(rowIndx <GetLength()) ;
    };
  
  return 1;

};


/*
**
** Desc: Tries to interpret inString as regular expression of the form
**        '/expression/'
**
**
** Pre: inStrings a string
**
**
** Post: returns extracted regular expression without '/'.
**
**
**
**
**
*/

AhelpStrings AhelpDBTable::GetRegularExpression(const AhelpStrings& inString)
{
  AhelpStrings tmpString="";


  if ( ( inString.FindChar(REGULAREXPRBRACKET) == 0 ) && (inString.FindLastChar(REGULAREXPRBRACKET) == (inString.length() -1 )))
    {
      tmpString=inString.GetSubStr(1,inString.length() -2); 

    };
  return tmpString;

};


/*
**
** Desc: Retrieves specified columns in rows which start offsets are specified
**       in inRowOffsetList.  
**
**
** Pre: inRowOffsetList is a list of offsets of rows to be processed
**      inColList list of column numbers from which content is to be retrieved
**      outRowList list of strings
**
**
**
**
** Post: Columns specified in inColList are retrieved from rows specified by
**       starting offsets in inRowOffsetList.  Then columns are concatenated
**       into Strings (columns are separated by FIELDSEPARATOR) and added into
**       outRowList.
**       Returns 1 on success
**               0 or negative on failure
**
*/


int AhelpDBTable::GetOffsetCells(AhelpList<long>& inRowOffsetList,AhelpList<int>& inColList,AhelpStringsList& outRowList)
{
  long rowIndx;
  int colIndx;
  AhelpStrings tmpString;
  AhelpStrings tmpRow;

  if ( inColList.GetLength() == 0 )
    return 0;
  
  outRowList.RemoveAll();
  
  for(rowIndx=0;rowIndx<inRowOffsetList.GetLength();rowIndx++)
    {
	
      for(colIndx=0;colIndx<inColList.GetLength();colIndx++)
	{ 
	  if ( (inColList.GetIndex(colIndx) >= colCount ) || (inColList.GetIndex(colIndx)< 0 ))
	    {
	      outRowList.RemoveAll();
	      return 0;
	    };
	  
	  if ( GetOffsetCell(inRowOffsetList.GetIndex(rowIndx),inColList.GetIndex(colIndx),tmpString) <=0 )
	    {
	      return 0;
	    };
	  
	  tmpRow.append(tmpString);

	  if ( colIndx < (inColList.GetLength()-1) )
	    {
	      tmpRow.append(FIELDSEPARATOR);
	    };
	};

      outRowList.Append(tmpRow);
      tmpRow.clear();
    };

  return 1;
};


/*
**
** Desc: Retrieve columns specified in inColList into outRowList
**
**
**
** Pre: inColList list of column numbers to be retrieved
**      outRowList list of Strings
**
**
** Post: For each row int the table columns specified in inColList 
**       are retrieved concatenated into string with FIELDSEPARATOR
**       separating them then newly formed row is added to outRowList
**       Returns 1 on success
**               0 or negative on failure
**
**
*/


int AhelpDBTable::GetCells(AhelpList<int>& inColList,AhelpStringsList& outRowList){

  long rowIndx;
  int colIndx;
  AhelpStrings tmpString;
  AhelpStrings tmpRow;
  
  if ( inColList.GetLength() == 0 )
    return 0;
  
  outRowList.RemoveAll();

  
  for(rowIndx=0;rowIndx<GetLength();rowIndx++)
    {
	
      for(colIndx=0;colIndx<inColList.GetLength();colIndx++)
	{ 
	  if ( (inColList.GetIndex(colIndx) >= colCount ) || (inColList.GetIndex(colIndx)< 0 ))
	    {
	      outRowList.RemoveAll();
	      return 0;
	    };
	  
	  if ( GetCell(rowIndx,inColList.GetIndex(colIndx),tmpString) <=0 )
	    {
	      return 0;
	    };
	  
	  tmpRow.append(tmpString);

	  if ( colIndx < (inColList.GetLength()-1) )
	    {
	      tmpRow.append(FIELDSEPARATOR);
	    };
	};

      outRowList.Append(tmpRow);
      tmpRow.clear();
    };

  return 1;

};



const char AhelpDBTable::REGULAREXPRBRACKET = '/';
const char AhelpDBTable::RECORDSEPARATOR='\x1F';
const char AhelpDBTable::FIELDSEPARATOR='\x1E';
const int AhelpDBTable::MAXRECORDSIZE=1024;
const char* const AhelpDBTable::FILEID="AHELPDB";
const int AhelpDBTable::HEADERSIZE=strlen(AhelpDBTable::FILEID)+42+4;

