// 
//  Copyright (C) 2002,2003,2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "AhelpDoc.hh"
#include "XMLFParser.hh"
#include "AhelpDB.hh"
#include "AhelpUtils.hh"

/*
**
** DESC: Constructor
**
**
**
** PRE:
**
**
** POST:
**
**
**
*/


AhelpDoc::AhelpDoc()
{
  Reset();
};

/*
**
** DESC: Desctructor
**
**
**
** PRE:
**
**
** POST:
**
**
**
*/


AhelpDoc::~AhelpDoc()
{
};

/*
**
** DESC: Returns reference to a string that contains list of refkeywords
**
**
**
** PRE: None
**
**
** POST: Returns reference to internal data memeber that is a string 
**       containing list of reference keywords for a given context
**       or for all contexts if inContext is EMTPY_CONTEXT
**
*/


const AhelpStrings& AhelpDoc::Refkeywords(int inIndx)
{ 
  if ( ( refkeywords.GetLength() == 0 ) || ( inIndx >= refkeywords.GetLength())
       || (inIndx <0) )
    return AhelpDoc::EMPTY_STRING;
  
  
  return refkeywords.GetIndex(inIndx);
};


/*
**
** DESC: Returns reference to a string that contains list of seealso groups
**
**
**
** PRE: None
**
**
** POST: Returns reference to internal data memeber that is a string 
**       containing list of seealso groups
**
*/


const AhelpStrings& AhelpDoc::Seealsogroups(int inIndx)
{
  if ( ( seealsogroups.GetLength() == 0 ) || ( inIndx >= seealsogroups.GetLength())
       || (inIndx <0) )
    return AhelpDoc::EMPTY_STRING;
  
  
  return seealsogroups.GetIndex(inIndx);
  
};



/*
**
** DESC: Returns reference to a string that contains list of seealso groups
**
**
**
** PRE: None
**
**
** POST: Returns reference to internal data memeber that is a string 
**       containing list of seealso groups
**
*/


const AhelpStrings& AhelpDoc::Displayseealsogroups(int inIndx)
{
  if ( ( displayseealsogroups.GetLength() == 0 ) || ( inIndx >= displayseealsogroups.GetLength())
       || (inIndx <0) )
    return AhelpDoc::EMPTY_STRING;
  
  
  return displayseealsogroups.GetIndex(inIndx);
  
};


/*
**
** DESC: Returns reference to a string that is a key
**
**
**
** PRE: None
**
**
** POST: Returns reference to key string
**
**
**
*/


const AhelpStrings& AhelpDoc::Key()
{
  return key;
};


/*
**
** DESC: Returns reference to context string
**
**
**
** PRE: None 
**
**
** POST: Returns reference to context string
**
**
**
*/


const AhelpStrings& AhelpDoc::Context(int inIndx)
{ 
  if ( ( context.GetLength() == 0 ) || ( inIndx >= context.GetLength())
       || (inIndx <0) )
    return AhelpDoc::EMPTY_STRING;
  
  
  return context.GetIndex(inIndx);

};


/*
**
** DESC: Returns reference to list of addresses that were extracted 
**       from xml document's ADDRESS block
**
**
** PRE: None
**
**
** POST: Returns reference to list of addresses
**
**
**
*/


AhelpList<Address>& AhelpDoc::AddressList()
{ 
  return addressList;
};



/*
**
** DESC: Formats parsed xml document into a text string.  Internal 
**       representation of parsed document is converted to a text
**       block of specified width.
**
** PRE: outFormattedString is Strings
**
**
** POST: Internal data structures are traversed recursively and 
**       contained information is converted into formatted text.
**
**
*/


void AhelpDoc::Format(AhelpStrings& outFormattedString,const AhelpStrings& inFormatContextString)
{
  TextBlock tmpTextBlock; 
  int tmpWidth=outputWidth;
  int status=1;

  tmpTextBlock.Resize(tmpWidth);
  if ( children.GetLength())
    {

      formatContextString = inFormatContextString;

      
      if ( status )
	FormatCXCHELPTOPICS(children.GetFirst(),tmpWidth,tmpWidth*marginRatio,tmpTextBlock);
      
      children.RemoveAll();
      
    };
  
  tmpTextBlock.Str(outFormattedString);
  outFormattedString.prepend(data);
  data.clear();
  
};



/*
**
** DESC: Construct a header for the formatted document
**
**
**
** PRE: inKey key
**      inContext context
**      inWidth width of formatted document
**      outFormattedBlock TextBlock
**
** POST: A header is created using supplied information and put into 
**       outFormattedBlock.
**
**
*/

void AhelpDoc::FormatHeader(const AhelpStrings& inKey,const AhelpStrings& inContext, int inWidth,TextBlock& outFormattedBlock)
{
  AhelpStrings tmpHeader;
  int padLen=inWidth;

  outFormattedBlock.Clear();
  
  tmpHeader="SUBJECT(";
  tmpHeader.append(inKey);
  tmpHeader.append(")");

  padLen -= inKey.length()+inContext.length()+18;

  if (padLen< 0 )
    padLen=0;

  tmpHeader.padR(padLen);
  tmpHeader.append("CONTEXT(");
  tmpHeader.append(inContext);
  tmpHeader.append(")");

  outFormattedBlock.Set(inWidth,tmpHeader.chars());
  outFormattedBlock.Append();

};



/*
**
** DESC: Formats CXCHELPTOPICS block
**
**
**
** PRE: inElement DTDElement containing info for CXCHELPTOPICS block
**      inWidth - width of formatted block
**      inMarginWidth - margin width of formatted block
**      outFormattedBlock TextBLock
**
** POST: CXCHELPTOPICS information contained in inElement is processed
**       and formatted recursively.  Result is put in outFormattedBlock
**
**
*/


void AhelpDoc::FormatCXCHELPTOPICS(DTDElement& inElement, int inWidth,int inMarginWidth,TextBlock& outFormattedBlock)
{
  outFormattedBlock.Resize(inWidth);
  FormatENTRY(inElement.children.GetFirst(),inWidth,inMarginWidth,outFormattedBlock);
  inElement.children.RemoveAll();

};


/*
**
** DESC: Formats ENTRY block
**
**
**
** PRE: inElement DTDElement containing info for ENTRY block
**      inWidth - width of formatted block
**      inMarginWidth - margin width of formatted block
**      outFormattedBlock TextBLock
**
** POST: ENTRY information contained in inElement is processed
**       and formatted recursively.  Result is put in outFormattedBlock
**
**
**
*/


void AhelpDoc::FormatENTRY(DTDElement& inElement, int inWidth,int inMarginWidth,TextBlock& outFormattedBlock)
{ 
  int SEEALSOGENFlag=0;
  TextBlock tmpBlock;
  TextBlock tmpOutFormattedBlock;
  AhelpStrings tmpParamStr;

  AhelpListIterator<AhelpStrings> attIter(&(inElement.attributes));
  
  // Extract attributes of the ENTRY block

  for(attIter.First();!attIter.IsDone();attIter.Next())
    {
      if ( attIter.GetCurrent() == ATT_KEY)
	{
	  attIter.Next();
	  key=attIter.GetCurrent();
	  key.lowercase();
	}
      else if ( attIter.GetCurrent() == ATT_CONTEXT)
	{
	  attIter.Next();
	  AhelpStrings tmpStr=attIter.GetCurrent();
	  tmpStr.lowercase();
	  context=tmpStr.Split('|');

	  if (! formatContextString.length() )
	    {
	      formatContextString=context.GetIndex(0);
	    }
	  else
	    {
	      AhelpListIterator<AhelpStrings> contextIter(&context);
	      formatContextIndx=0;
	      for(contextIter.First();!contextIter.IsDone();contextIter.Next())
		{
		  if ( formatContextString == contextIter.GetCurrent() )
		    {
		      
		      break;
		    };
		  formatContextIndx++;
		};

	      if (contextIter.IsDone() )
		return;
	    }
	}
      else if ( attIter.GetCurrent() == ATT_REFKEYWORDS)
	{
	  attIter.Next();
	  AhelpStrings tmpStr=attIter.GetCurrent();
	  tmpStr.lowercase();
	  refkeywords=tmpStr.Split('|');
	  
	}
      else if ( attIter.GetCurrent() == ATT_SEEALSOGROUPS)
	{
	  attIter.Next();
	  AhelpStrings tmpStr=attIter.GetCurrent();
	  tmpStr.lowercase();
	  seealsogroups=tmpStr.Split('|');
	}
      else if ( attIter.GetCurrent() == ATT_DISPLAYSEEALSOGROUPS)
	{
	  attIter.Next();
	  AhelpStrings tmpStr=attIter.GetCurrent();
	  tmpStr.lowercase();
	  displayseealsogroups=tmpStr.Split('|');
	};
    };
  

  tmpOutFormattedBlock.Resize(inWidth);
  outFormattedBlock.Resize(inWidth);
  genSYNTAXFlag=1;

  // Generate header if requested

  if ( HEADERFlag)
    {
      FormatHeader(key,formatContextString,inWidth,outFormattedBlock);
    };
  

  // Format all block that are contained in ENTRY block
  
  formattedBlocksCount=0;

  FilterContext(inElement);

  while (inElement.children.GetLength() !=  0 )
    {
      DTDElement& currChild = inElement.children.GetFirst();

      if ( (currChild.name == TOK_SYNOPSIS ) && (SYNOPSISFlag))
	{      
	  formattedBlocksCount++;
	  FormatSYNOPSIS(currChild,"SYNOPSIS",inWidth,inMarginWidth,tmpBlock);
	  outFormattedBlock.Append(tmpBlock);
	  inElement.children.RemoveTop();
	  continue;
	}
      else if (( currChild.name == TOK_SYNTAX ) && (SYNTAXFlag) )
	{
	  formattedBlocksCount++;
	  if ( FormatSYNTAX(currChild,"SYNTAX",inWidth,inMarginWidth,tmpBlock) )
	    {
	      if ( ! tmpBlock.IsEmpty())
		{
		  tmpBlock.Prepend();
		};
	      genSYNTAXFlag=0;
	    }
	  else
	    {
	      genSYNTAXFlag=1;
	      inElement.children.RemoveTop(); 
	      continue;
	    };
	}
      else if (( currChild.name == TOK_ADDRESS ) && (ADDRESSFlag))
	{
	  formattedBlocksCount++;
	  FormatADDRESS(currChild);
	  inElement.children.RemoveTop();
	  continue;
	}
      else if (( currChild.name == TOK_DESC ) && (DESCFlag))
	{
	  formattedBlocksCount++;
	  FormatDESC(currChild,"DESCRIPTION",inWidth,inMarginWidth,tmpBlock);
	  if ( ! tmpBlock.IsEmpty())
	    {
	      tmpBlock.Prepend();
	    };
	}
      else if (( currChild.name == TOK_QEXAMPLELIST ) && (QEXAMPLELISTFlag))
	{
	  formattedBlocksCount++;
	  FormatQEXAMPLELIST(currChild,"EXAMPLES","EXAMPLE",inWidth,inMarginWidth,tmpBlock);
	  if ( ! tmpBlock.IsEmpty())
	    {
	      tmpBlock.Prepend();
	    };
	}
      else if (( currChild.name == TOK_PARAMLIST ) && (PARAMLISTFlag))
	{
	  if ( selectionPARAMName.isEmpty())
	    {
	      formattedBlocksCount++;
	      FormatPARAMLIST(currChild,"PARAMETERS","DETAILED PARAMETER DESCRIPTIONS","PARAM",inWidth, inMarginWidth,1,tmpBlock,tmpParamStr);
	    }
	  else
	    {
	      FormatSelectedPARAM(currChild,"PARAM",selectionPARAMName,inWidth, inMarginWidth,tmpBlock);
	    };
	  if ( ! tmpBlock.IsEmpty())
	    {
	      tmpBlock.Prepend();
	    };
	}
      else if (( currChild.name == TOK_ADESC ) && (ADESCFlag))
	{
	  if ( selectionADESCTitle.isEmpty())
	    {
	      formattedBlocksCount++;
	      FormatADESCList(inElement.children,inWidth,inMarginWidth,tmpBlock);
	    }
	  else
	    {
	      FormatSelectedADESC(inElement.children,selectionADESCTitle,inWidth,inMarginWidth,tmpBlock);
	    };
	  if ( ! tmpBlock.IsEmpty())
	    {
	      tmpBlock.Prepend();
	    };
	  tmpOutFormattedBlock.Append(tmpBlock);
	  continue;
	}
      else if (( currChild.name == TOK_BUGS ) && (BUGSFlag))
	{
	  formattedBlocksCount++;
	  FormatBUGS(currChild,"BUGS",inWidth,inMarginWidth,tmpBlock);
	  if ( ! tmpBlock.IsEmpty())
	    {
	      tmpBlock.Prepend();
	    };
	}
      else if (( ! SEEALSOGENFlag ) && ( SEEALSOFlag))
	{
	  formattedBlocksCount++;
	  FormatSEEALSO("SEE ALSO",inWidth,inMarginWidth,tmpBlock);
	  SEEALSOGENFlag=1;
	  if ( ! tmpBlock.IsEmpty())
	    {
	      tmpBlock.Prepend();
	    };
	  tmpOutFormattedBlock.Append(tmpBlock);
	  continue;
	}
      else if (( currChild.name == TOK_VERSION ) && (VERSIONFlag))
	{
	  formattedBlocksCount++;
	  FormatVERSION(currChild,"VERSION",inWidth,inMarginWidth,tmpBlock);
	  if ( ! tmpBlock.IsEmpty())
	    {
	      tmpBlock.Prepend();
	    };
	}
      else if (( currChild.name == TOK_LASTMODIFIED ) && (LASTMODIFIEDFlag))
	{
	  formattedBlocksCount++;
	  FormatLASTMODIFIED(currChild,"LAST MODIFIED",inWidth,inMarginWidth,tmpBlock);
	  if ( ! tmpBlock.IsEmpty())
	    {
	      tmpBlock.Prepend();
	    };
	}
      else
	{
	  inElement.children.RemoveTop();
	  continue;
	};

      inElement.children.RemoveTop();
      tmpOutFormattedBlock.Append(tmpBlock);
    };

  if ( genSYNTAXFlag && SYNTAXFlag && (!tmpParamStr.isEmpty()))
    {  
      FormatSYNTAX(tmpParamStr,"SYNTAX",inWidth,inMarginWidth,tmpBlock);
      tmpBlock.Prepend();
      outFormattedBlock.Append(tmpBlock);
    };
  outFormattedBlock.Append(tmpOutFormattedBlock);

};


/*
**
** DESC:  Extract Address info into internal data structures
**
**
**
** PRE: inElement DTDElement that contains info for ADDRESS block
**
**
** POST: Address info is put into internal data structures
**
**
**
*/


void AhelpDoc::FormatADDRESS(DTDElement& inElement)
{ 

  Address tmpAddress;
  
  AhelpListIterator<DTDElement> childIter(&(inElement.children));
  
  for(childIter.First();!childIter.IsDone();childIter.Next())
    {
      if ( childIter.GetCurrent().name == TOK_URL )
	{
	  FormatURL(childIter.GetCurrent());
	  //	  childIter.GetCurrent().data.Substitute(" ","",1);
	  tmpAddress.url = childIter.GetCurrent().data;
	}
	  else if ( childIter.GetCurrent().name == TOK_LABEL )
	    {
	      FormatLABEL(childIter.GetCurrent());
	      tmpAddress.label = childIter.GetCurrent().data;
	    };
    };

  if ( tmpAddress.url.length())
    addressList.Append(tmpAddress);
  inElement.children.RemoveAll();


};


/*
**
** DESC: Formats ADESC block that has title inTitle
**
**
**
** PRE: inElementList list of DTDElements that are ADESC blocks
**      inTitle title of searched block to be formatted
**      inWidth width of formatted block
**      inMargin margin width of formatted block
**      outFormattedBlock contains formatted block
**
** POST: list of ADESC blocks is searched and inElementList and the first
**       one with attribute title inTitle is formatted accordingly. Result
**       is put in outFormattedBlock
**
*/


void AhelpDoc::FormatSelectedADESC(AhelpList<DTDElement>& inElementList, AhelpStrings inTitle,int inWidth, int inMargin,TextBlock& outFormattedBlock)
{
  
  outFormattedBlock.Clear();
  outFormattedBlock.Resize(inWidth);
  AhelpStrings tmpTitle;
  int indx=1;
  AhelpStrings tmpIndx;
  bool numFlag=true;

  while( inElementList.GetLength() )
    {
      DTDElement& tmpElement = inElementList.GetFirst();
      if ( tmpElement.name != TOK_ADESC)
	{
	  break;
	};  

      GetAttribute(tmpElement,ATT_PARAGRAPHNUMBERING,tmpIndx);
      if (tmpIndx == "no")
	numFlag=false;
      else
	numFlag=true;

      tmpIndx.clear();
      tmpIndx.append(indx);

      if ( tmpElement.attributes.GetLength())
	{
	  GetAttribute(tmpElement,ATT_TITLE,tmpTitle);

	  
	  CompactString(tmpTitle);
	  CompactString(inTitle);
	  if ( tmpTitle == inTitle)
	    {
	      formattedBlocksCount++;
	      FormatADESC(tmpElement,tmpIndx.chars(),inWidth,inMargin,outFormattedBlock);
	      inElementList.RemoveTop();
	      break;
	    }
	  else
	    {
	      
	      FormatADESC(tmpElement,tmpIndx.chars(),inWidth,inMargin,outFormattedBlock);
	      if (! outFormattedBlock.IsEmpty())
		{
		  if ( numFlag )
		    indx++;
		};
	      outFormattedBlock.Clear();
	      
	    };
	};

      inElementList.RemoveTop();

    };  

};

/*
**
** DESC: Formats list of ADESC blocks
**
**
**
** PRE: inElementList list of ADESC blocks
**      inWidth width of formatted block
**      inMargin margin width of formatted block
**      outFormattedBlock formatted block
**
**
** POST: list inElementList is scanned and all ADESC blocks are formatted
**       and appended to outFormattedBlock
**
**
*/


void AhelpDoc::FormatADESCList(AhelpList<DTDElement>& inElementList, int inWidth, int inMargin,TextBlock& outFormattedBlock)
{
  
  int indx=1;
  AhelpStrings tmpIndx;
  TextBlock tmpBlock;
  bool numFlag=true;

  outFormattedBlock.Clear();
  outFormattedBlock.Resize(inWidth);

  tmpBlock.Resize(inWidth);
  while( inElementList.GetLength() )
    {
      DTDElement& tmpElement = inElementList.GetFirst();
      if ( tmpElement.name != TOK_ADESC)
	{
	  break;
	};

      GetAttribute(tmpElement,ATT_PARAGRAPHNUMBERING,tmpIndx);
      if (tmpIndx == "no")
	numFlag=false;
      else
	numFlag=true;

      tmpIndx.clear();
      tmpIndx.append(indx);
      
      FormatADESC(tmpElement,tmpIndx.chars(),inWidth,inMargin,tmpBlock);
      inElementList.RemoveTop();

      if ( ! tmpBlock.IsEmpty())
	{
	  if ( numFlag )
	    indx++;
	  outFormattedBlock.Append(tmpBlock);

	  if  ( inElementList.GetLength()  )
	    {
	      outFormattedBlock.Append();
	    };
	};
    };  
  
};

/*
**
** DESC: Format ADESC block
**
**
**
** PRE: inElement contains ADESC block info
**      inIndx label to be put in block header
**      inWidth width of formatted block
**      inMargin margin width in formatted block
**      outFormattedBlock formatted block
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
**
*/


void AhelpDoc::FormatADESC(DTDElement& inElement, const AhelpStrings& inIndx,int inWidth, int inMargin,TextBlock& outFormattedBlock)
{ 

  TextBlock tmpBlock;
  int titleFlag=0;
  AhelpStrings tmpString;
  AhelpStrings tmpIndx=inIndx;
  bool numFlag=true;
  int tmpVal;

  outFormattedBlock.Clear();

  if ( inElement.IsEmpty())
    return;

  GetAttribute(inElement,ATT_PARAGRAPHNUMBERING,tmpString);

  if (tmpString == "no")
    {
      tmpIndx="";
      tmpIndx.fill(inIndx.length());
      numFlag=false;
    };

  GetAttribute(inElement,ATT_TITLE,tmpString);


  outFormattedBlock.Resize(inWidth);


  if ( tmpString.length() )
    {
      TextBlock tmpTitle(inWidth-tmpIndx.length()-1,tmpString.chars());
      outFormattedBlock.Append(tmpIndx.chars());
      if ( numFlag)
	{
	  outFormattedBlock.Append(". ");
	}
      else
	{
	   outFormattedBlock.Append("  ");
	}
      outFormattedBlock.Paste(tmpIndx.length()+2,0,tmpTitle);
      outFormattedBlock.Append();
      titleFlag=1;
    };


  int indx=1;

  while(inElement.children.GetLength())
    {
      DTDElement& tmpElement=inElement.children.GetFirst();

      if (tmpElement.name == TOK_PARA )
	{
	  AhelpStrings tmpPARAIndx;
	  int tmpFlag;
	  tmpPARAIndx.append( indx);
	  if ( titleFlag)
	    {
	      tmpPARAIndx.prepend(".");
	      tmpPARAIndx.prepend(tmpIndx);
	    }
	  else
	    {
	      tmpPARAIndx.append(". ");
	    };


	  if ( !numFlag)
	    {
	      tmpVal=tmpPARAIndx.length();
	      tmpPARAIndx="";
	      tmpPARAIndx.fill(tmpVal);
	    };


	  if (FormatPARA(tmpElement,tmpPARAIndx,inWidth-2*inMargin,tmpBlock) )
	    indx++;
	}
      else if (tmpElement.name == TOK_TABLE )
	{
	  FormatTABLE(tmpElement,inWidth-2*inMargin,tmpBlock);
	}
      else if ( tmpElement.name == TOK_VERBATIM )
	{
	  FormatVERBATIM(tmpElement,tmpBlock);
	}
      else if ( tmpElement.name == TOK_LIST )
	{
	  FormatLIST(tmpElement,inWidth-2*inMargin,tmpBlock);
	};
      
      inElement.children.RemoveTop();
      
      if ( ! tmpBlock.IsEmpty())
	{
	  outFormattedBlock.Paste(inMargin,outFormattedBlock.Height(),tmpBlock);
	  
	  if (inElement.children.GetLength())
	    {
	      outFormattedBlock.Append();
	    };
	};
    };
};


/*
**
** DESC: Formats BUGS block
**
**
**
** PRE: inElement contains BUGS info
**      inHeader header for the formatted block
**      inWidth width of formatted block
**      inMargin margin width in formatted block
**      outFormattedBlock formatted block
**
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
**
*/


void AhelpDoc::FormatBUGS(DTDElement& inElement, const char* inHeader,int inWidth, int inMargin,TextBlock& outFormattedBlock)
{ 

  TextBlock tmpBlock;

  outFormattedBlock.Clear();
  if ( inElement.IsEmpty())
    return;

  outFormattedBlock.Resize(inWidth);

  if ( inHeader != 0 )
    {
      TextBlock tmpBlock(inWidth,inHeader);
      outFormattedBlock.Append(tmpBlock);
      outFormattedBlock.Append();
    };
  
  FilterContext(inElement);
  ConcatPCDATAandCONDTEXT(inElement);


  while(inElement.children.GetLength())
    {
      DTDElement& tmpElement=inElement.children.GetFirst();

      if ( tmpElement.name == TOK_PCDATA )
	{
	  FormatPCDATA(tmpElement); 
	  tmpBlock.Set(inWidth-2*inMargin,tmpElement.data.chars());
	}
      else if ( tmpElement.name == TOK_PARA )
	{
	  FormatPARA(tmpElement,"",inWidth-2*inMargin,tmpBlock);
	}
      else if ( tmpElement.name == TOK_TABLE )
	{
	  FormatTABLE(tmpElement,inWidth-2*inMargin,tmpBlock);
	}
      else if ( tmpElement.name == TOK_LIST )
	{
	  FormatLIST(tmpElement,inWidth-2*inMargin,tmpBlock);
	}
      else if ( tmpElement.name == TOK_VERBATIM )
	{
	  FormatVERBATIM(tmpElement,tmpBlock);
	};
      
      inElement.children.RemoveTop();
      
      if ( ! tmpBlock.IsEmpty())
	{
	  outFormattedBlock.Paste(inMargin,outFormattedBlock.Height(),tmpBlock);  
	  if (inElement.children.GetLength())
	    {
	      outFormattedBlock.Append();
	    };
	};
    };
};



/*
**
** DESC: Formats CAPTION block
**
**
**
** PRE: inElement contains CAPTION info
**      inMaxWidth max width of formatted block 
**      outFormattedBlock formatted block
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
**
*/


void AhelpDoc::FormatCAPTION(DTDElement& inElement,int inMaxWidth,TextBlock& outFormattedBlock)
{    

  outFormattedBlock.Clear();
  if ( inElement.IsEmpty())
    return;

 if ( inElement.children.GetLength() )
   {
     ConcatPCDATAandCONDTEXT(inElement);
     
     FormatPCDATA(inElement.children.GetFirst()); 
     
     if ( inElement.children.GetFirst().data.length() > inMaxWidth)
       {
	 outFormattedBlock.Set(inMaxWidth,inElement.children.GetFirst().data.chars());
       }
     else
       {
	 outFormattedBlock.Set(inElement.children.GetFirst().data.chars());
       };

      inElement.children.RemoveAll();
   };

};



/*
**
** DESC: Formats DATA block
**
**
**
** PRE: inElement contains block info
**
**
** POST: formatted block is put in inElement.data member as Strings
**
**
**
*/


void AhelpDoc::FormatDATA(DTDElement& inElement)
{   

  if ( inElement.children.GetLength() )
    {  
      ConcatPCDATAandCONDTEXT(inElement);
      FormatPCDATA(inElement.children.GetFirst());
      inElement.data = inElement.children.GetFirst().data;
      inElement.children.RemoveAll();
      
    }
  
};


/*
**
** DESC: Formats DESC block
**
**
**
** PRE: inElement contains DESC info
**      inHeader header for the formatted block
**      inWidth width of formatted block
**      inMargin margin width in formatted block
**      outFormattedBlock formatted block
**
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
**
*/

void AhelpDoc::FormatDESC(DTDElement& inElement, const char* inHeader,int inWidth, int inMargin,TextBlock& outFormattedBlock)
{ 

  TextBlock tmpBlock;

  outFormattedBlock.Clear();
  if ( inElement.IsEmpty())
    return;

  outFormattedBlock.Resize(inWidth);

  if ( inHeader != 0 )
    {   
      TextBlock tmpBlock(inWidth,inHeader);
      outFormattedBlock.Append(tmpBlock);
      outFormattedBlock.Append();
    };
  
  while(inElement.children.GetLength())
    {
      DTDElement& tmpElement=inElement.children.GetFirst();
      
      if ( tmpElement.name == TOK_PARA )
	{
	  FormatPARA(tmpElement,"",inWidth-2*inMargin,tmpBlock);
	}
      else if ( tmpElement.name == TOK_TABLE )
	{
	  FormatTABLE(tmpElement,inWidth-2*inMargin,tmpBlock);
	}
      else if ( tmpElement.name == TOK_LIST )
	{
	  FormatLIST(tmpElement,inWidth-2*inMargin,tmpBlock);
	}
      else if ( tmpElement.name == TOK_VERBATIM )
	{
	  FormatVERBATIM(tmpElement,tmpBlock);
	};

      
      inElement.children.RemoveTop();
      
      if ( ! tmpBlock.IsEmpty())
	{
	  outFormattedBlock.Paste(inMargin,outFormattedBlock.Height(),tmpBlock);  
	  if (inElement.children.GetLength())
	    {
	      outFormattedBlock.Append();
	    };
	};
    };
 
};


/*
**
** DESC: Formats EQUATION block
**
**
**
** PRE: inElement contains EQUATION info
**      inMaxWidth max width of formatted block 
**      outFormattedBlock formatted block
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
*/


void AhelpDoc::FormatEQUATION(DTDElement& inElement,int inMaxWidth, TextBlock& outFormattedBlock)
{ 

  outFormattedBlock.Clear();
  if ( inElement.IsEmpty())
    return;

  
  ConcatPCDATAandCONDTEXT(inElement);

  while(inElement.children.GetLength())
    {
      DTDElement& tmpElement=inElement.children.GetFirst();

      if ( tmpElement.name == TOK_PCDATA )
	{
	  FormatPCDATA(tmpElement);
	}
      else if ( tmpElement.name == TOK_PASSTHRU )
	{
	  inElement.children.RemoveTop();
	  continue;
	  //   FormatPASSTHRU(tmpElement);
	}
      else if ( tmpElement.name == TOK_XMLONLY )
	{
	  FormatXMLONLY(tmpElement);
	};
      inElement.data.append(" ");
      inElement.data.append(tmpElement.data);
      inElement.children.RemoveTop();

      
      inElement.children.RemoveTop();
    };
  
  FormatPCDATA(inElement);
  
  outFormattedBlock.Set(inElement.data.chars());
  
  if(outFormattedBlock.Width() > inMaxWidth)
    {
      outFormattedBlock.Resize(inMaxWidth);
    };

};



/*
**
** DESC: Format HREF block
**
**
**
** PRE: inElement contains info on HREF block
**      
**
** POST: HREF block is formatted and put into inElement.data member
**
**
**
*/


void AhelpDoc::FormatHREF(DTDElement& inElement)
{  

  AhelpStrings tmpString;


  GetAttribute(inElement,ATT_LINK,tmpString);

  if (inElement.children.GetLength())
    {

      ConcatPCDATAandCONDTEXT(inElement);
      
      FormatPCDATA(inElement.children.GetFirst());
      inElement.data = " ";
      inElement.data.append(inElement.children.GetFirst().data); 
      if ( tmpString.length())
	{
	  CompactString(tmpString);
	  if ( inElement.children.GetFirst().data != tmpString)
	    {
	      inElement.data.append(" [");
	      inElement.data.append(tmpString);
	      inElement.data.append("]");
	    };
	};
      inElement.data.append(" ");  
      
      inElement.children.RemoveAll();
    };
};


/*
**
** DESC: Formats ITEM block
**
**
**
** PRE: inElement contains ITEM info
**      inWidth width of formatted block 
**      outFormattedBlock formatted block
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
*/


void AhelpDoc::FormatITEM(DTDElement& inElement, int inWidth,TextBlock& outFormattedBlock)
{  
  outFormattedBlock.Clear();
  if ( inElement.IsEmpty())
    return;

  

  outFormattedBlock.Resize(inWidth);
  TextBlock tmpBlock;
  if ( inElement.children.GetLength() )
    {

      ConcatPCDATAandCONDTEXT(inElement);

      FormatPCDATA(inElement.children.GetFirst());
      tmpBlock.Set(inWidth-3,inElement.children.GetFirst().data.chars());
      outFormattedBlock.Append("o) ");
      outFormattedBlock.Paste(3,0,tmpBlock); 
      inElement.children.RemoveAll();
    };
};


/*
**
** DESC: Formats LABEL block
**
**
**
** PRE: inElement contains block info
**
**
** POST: formatted block is put in inElement.data member as Strings
**
**
**
*/


void AhelpDoc::FormatLABEL(DTDElement& inElement)
{ 

  if(inElement.children.GetLength())
    {      
      ConcatPCDATAandCONDTEXT(inElement);
      FormatPCDATA(inElement.children.GetFirst());
      inElement.data = inElement.children.GetFirst().data;
      inElement.children.RemoveAll();
    };
};


/*
**
** DESC: Formats LASTMODIFIED block
**
**
**
** PRE: inElement contains LASTMODIFIED info
**      inHeader header for the formatted block
**      inWidth width of formatted block
**      inMargin margin width in formatted block
**      outFormattedBlock formatted block
**
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
**
*/


void AhelpDoc::FormatLASTMODIFIED(DTDElement& inElement, const char* inHeader,int inWidth, int inMargin,TextBlock& outFormattedBlock)
{  
  outFormattedBlock.Clear();
  if ( inElement.IsEmpty())
    return;

  outFormattedBlock.Resize(inWidth);
  if ( inHeader != 0 )
    {   
      TextBlock tmpBlock(inWidth,inHeader);
      outFormattedBlock.Append(tmpBlock);
      outFormattedBlock.Append();
    };
  
  if ( inElement.children.GetLength() )
    {
      TextBlock tmpBlock;
      
      ConcatPCDATAandCONDTEXT(inElement);
      FormatPCDATA(inElement.children.GetFirst());

      tmpBlock.Set(inWidth-2*inMargin,inElement.children.GetFirst().data.chars());
      outFormattedBlock.Paste(inMargin,outFormattedBlock.Height(),tmpBlock);
      inElement.children.RemoveAll();
    };
};



/*
**
** DESC: Formats LINE block
**
**
**
** PRE: inElement contains LINE info
**      inWidth width of formatted block 
**      outFormattedBlock formatted block
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
**
*/


void AhelpDoc::FormatLINE(DTDElement& inElement, int inWidth,TextBlock& outFormattedBlock)
{ 
  outFormattedBlock.Clear();
  if ( inElement.IsEmpty())
    return;


  outFormattedBlock.Resize(inWidth);



  ConcatPCDATAandCONDTEXT(inElement);

  while(inElement.children.GetLength())
    {
      DTDElement& tmpElement=inElement.children.GetFirst();

      if ( tmpElement.name == TOK_PCDATA )
	{
	  FormatPCDATA(tmpElement);

	} 
      else if ( tmpElement.name == TOK_HREF )
	{
	  FormatHREF(tmpElement);
	};

      outFormattedBlock.Append(tmpElement.data.chars());  

      inElement.children.RemoveTop();

    };

};



/*
**
** DESC: Formats LIST block
**
**
**
** PRE: inElement contains LIST info
**      inWidth width of formatted block 
**      outFormattedBlock formatted block
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
**
*/

void AhelpDoc::FormatLIST(DTDElement& inElement, int inWidth,TextBlock& outFormattedBlock)
{ 
  outFormattedBlock.Clear();
  if ( inElement.IsEmpty())
    return;


  outFormattedBlock.Resize(inWidth);
  TextBlock tmpBlock;

  if ( inElement.children.GetLength())
    {
      if ( inElement.children.GetFirst().name == TOK_CAPTION )
	{
	  FormatCAPTION(inElement.children.GetFirst(),inWidth,tmpBlock);
	  int currWidth = tmpBlock.Width();
	  outFormattedBlock.Paste((inWidth-currWidth)/2,0,tmpBlock);
	  inElement.children.RemoveTop();
	};
    };
  
  outFormattedBlock.Append();

  while( inElement.children.GetLength())
    {
      DTDElement& tmpElement=inElement.children.GetFirst();
      FormatITEM(tmpElement,inWidth,tmpBlock);

      inElement.children.RemoveTop();

      if ( ! tmpBlock.IsEmpty())
	{
	  outFormattedBlock.Append(tmpBlock);
	  
	  if (inElement.children.GetLength())
	    {
	      outFormattedBlock.Append();
	    };
	};
    };
 
  inElement.children.RemoveAll();

};


/*
**
** DESC: Format PARA block
**
**
**
** PRE: inElement contains PARA block info
**      inIndx label to be put in block header
**      inWidth width of formatted block
**      inMargin margin width in formatted block
**      outFormattedBlock formatted block
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
**
*/


int AhelpDoc::FormatPARA(DTDElement& inElement,const AhelpStrings& inIndx,int inWidth,TextBlock& outFormattedBlock)
{  
  outFormattedBlock.Clear();
  if ( inElement.IsEmpty())
    return 0 ;


  outFormattedBlock.Resize(inWidth);
  int outTitleFlag=0;
  TextBlock tmpBlock;
  int initFlag=1;
  AhelpStrings tmpString;

  AhelpListIterator<AhelpStrings> attIter(&(inElement.attributes));


  GetAttribute(inElement,ATT_TITLE,tmpString);


  if ( tmpString.length() )
    {
      
      TextBlock tmpTitle(inWidth-inIndx.length()-1,tmpString.chars());
      if ( inIndx.length() > 0)
	{
	  outFormattedBlock.Append(inIndx.chars());
	  outFormattedBlock.Append(" ");
	  outFormattedBlock.Paste(inIndx.length()+1,0,tmpTitle);
	}
      else
	{
	  outFormattedBlock.Paste(0,0,tmpTitle);
	};
      
      outFormattedBlock.Append();
      outTitleFlag=1;
    };

  ConcatPCDATAandCONDTEXT(inElement);

  while(inElement.children.GetLength())
    {
      DTDElement& tmpElement=inElement.children.GetFirst();
      
      if ( tmpElement.name == TOK_PCDATA )
	{
	  if (( initFlag ) && (outTitleFlag))
	    {
	      outFormattedBlock.Append();
	    };
	  FormatPCDATA(tmpElement); 
	  outFormattedBlock.Append(tmpElement.data.chars());  
	}
      else if ( tmpElement.name == TOK_SYNTAX )
	{
	  FormatSYNTAX(tmpElement,0,inWidth,0,tmpBlock);
	  outFormattedBlock.Append(tmpBlock);
	}
      else if ( tmpElement.name == TOK_EQUATION )
	{
	  FormatEQUATION(tmpElement,inWidth,tmpBlock);
	  int curWidth = tmpBlock.Width();
	  tmpBlock.Append();
	  outFormattedBlock.Paste((inWidth - curWidth)/2,outFormattedBlock.Height(),tmpBlock);  
	}
      else if ( tmpElement.name == TOK_PASSTHRU )
	{
	  //	  FormatPASSTHRU(tmpElement);
	}
      else if ( tmpElement.name == TOK_XMLONLY )
	{
	  if (( initFlag )&& (outTitleFlag))
	    {
	      outFormattedBlock.Append();
	    };
	  FormatXMLONLY(tmpElement);
	  outFormattedBlock.Append(tmpElement.data.chars());
	}
      else if ( tmpElement.name == TOK_HREF )
	{
	  if (( initFlag )&& (outTitleFlag))
	    {
	      outFormattedBlock.Append();
	    };
	  FormatHREF(tmpElement);
	  outFormattedBlock.Append(tmpElement.data.chars());
	};

      inElement.children.RemoveTop();
      initFlag=0;
      
    };
 
  return outTitleFlag;
};



/*
**
** DESC: Formats single param info of PARAM block
**
**
**
** PRE: inElmenet contains info on single param
**      inHeadre header to be put in formatted block
**      inWidth width of formatted block
**      inMargin margin in formatted block
**      outAtts list of extracted attributes of PARAM block  
**      outParamString info later used to generate synopsis
**                     
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
**
*/


void AhelpDoc::FormatPARAMDetail(DTDElement& inElement, const char* inHeader,int inWidth, int inMargin,AhelpList<AhelpStrings>& outAtts,TextBlock& outFormattedBlock,AhelpStrings& outParamString)
{
  outFormattedBlock.Clear();
  if ( inElement.IsEmpty() )
    return;


  outFormattedBlock.Resize(inWidth);
  int attFlag=0;
  TextBlock tmpBlock;
  
  AhelpStrings tmpName,tmpType,tmpFiletype,tmpDef,tmpMin,tmpMax,tmpUnits,tmpReqd,tmpStacks,tmpAutoname;
  

  AhelpListIterator<AhelpStrings> attIter(&(inElement.attributes));

  for(attIter.First();!attIter.IsDone();attIter.Next())
    {

      if ( attIter.GetCurrent() == ATT_NAME)
	{
	  attIter.Next();
	  tmpName=attIter.GetCurrent();
	}
      else if ( attIter.GetCurrent() == ATT_TYPE)
	{
	  attIter.Next();
	  tmpType=attIter.GetCurrent();
	}
      else if ( attIter.GetCurrent() == ATT_FILETYPE)
	{
	  attIter.Next();
	  tmpFiletype=attIter.GetCurrent();
	}
      else if ( attIter.GetCurrent() == ATT_DEF)
	{
	  attIter.Next();
	  tmpDef=attIter.GetCurrent();
	}
      else if ( attIter.GetCurrent() == ATT_MIN)
	{
	  attIter.Next();
	  tmpMin=attIter.GetCurrent();
	}
      else if ( attIter.GetCurrent() == ATT_MAX)
	{
	  attIter.Next();
	  tmpMax=attIter.GetCurrent();
	}
      else if ( attIter.GetCurrent() == ATT_UNITS)
	{
	  attIter.Next();
	  tmpUnits=attIter.GetCurrent();
	}
      else if ( attIter.GetCurrent() == ATT_REQD)
	{
	  attIter.Next();
	  tmpReqd=attIter.GetCurrent();
	}
      else if ( attIter.GetCurrent() == ATT_STACKS)
	{
	  attIter.Next();
	  tmpStacks=attIter.GetCurrent();
	}
      else if ( attIter.GetCurrent() == ATT_AUTONAME)
	{
	  attIter.Next();
	  tmpAutoname=attIter.GetCurrent();
	}
      

    };

  outAtts.Append(tmpName);
  outAtts.Append(tmpType);
  outAtts.Append(tmpFiletype);
  outAtts.Append(tmpDef);
  outAtts.Append(tmpMin);
  outAtts.Append(tmpMax);
  outAtts.Append(tmpUnits);
  outAtts.Append(tmpReqd);
  outAtts.Append(tmpStacks);
  outAtts.Append(tmpAutoname);


  if ( inHeader != 0 )
    {
      TextBlock tmpHeader(inWidth,inHeader);
      if (! tmpName.isEmpty())
	{
	  tmpHeader.Append(" ");
	  tmpHeader.Append(tmpName.chars());
	};

      outFormattedBlock.Append(tmpHeader);
      outFormattedBlock.Append();
    };


  if(! tmpType.isEmpty())
    {
      outFormattedBlock.Paste(inMargin, outFormattedBlock.Height(),ATT_TYPE.chars());
      outFormattedBlock.Append("=");
      outFormattedBlock.Append(tmpType.chars());
    };
  
  if(! tmpFiletype.isEmpty() )
    {
      outFormattedBlock.Paste(inMargin, outFormattedBlock.Height(),ATT_FILETYPE.chars());
      outFormattedBlock.Append("=");
      outFormattedBlock.Append(tmpFiletype.chars());
    };

  if(! tmpDef.isEmpty() )
    {
      outFormattedBlock.Paste(inMargin, outFormattedBlock.Height(),ATT_DEF.chars());
      outFormattedBlock.Append("=");
      outFormattedBlock.Append(tmpDef.chars());
    };


  if(! tmpMin.isEmpty())
    {
      outFormattedBlock.Paste(inMargin, outFormattedBlock.Height(),ATT_MIN.chars());
      outFormattedBlock.Append("=");
      outFormattedBlock.Append(tmpMin.chars());
    };


  if(! tmpMax.isEmpty())
    {
      outFormattedBlock.Paste(inMargin, outFormattedBlock.Height(),ATT_MAX.chars());
      outFormattedBlock.Append("=");
      outFormattedBlock.Append(tmpMax.chars());
    };


  if(! tmpUnits.isEmpty())
    {
      outFormattedBlock.Paste(inMargin, outFormattedBlock.Height(),ATT_UNITS.chars());
      outFormattedBlock.Append("=");
      outFormattedBlock.Append(tmpUnits.chars());
    };


  if(! tmpReqd.isEmpty())
    {
      outFormattedBlock.Paste(inMargin, outFormattedBlock.Height(),ATT_REQD.chars());
      outFormattedBlock.Append("=");
      outFormattedBlock.Append(tmpReqd.chars());

    };

  if ( tmpReqd.FindStr("yes") >= 0 )
    {
      outParamString.append(" ");
      outParamString.append(tmpName);
    }
  else
    {
      outParamString.append(" [");
      outParamString.append(tmpName);
      outParamString.append("]");
    };
  
  if(! tmpStacks.isEmpty())
    {
      outFormattedBlock.Paste(inMargin, outFormattedBlock.Height(),ATT_STACKS.chars());
      outFormattedBlock.Append("=");
      outFormattedBlock.Append(tmpStacks.chars());
    };


  if(! tmpAutoname.isEmpty())
    {
      outFormattedBlock.Paste(inMargin, outFormattedBlock.Height(),ATT_AUTONAME.chars());
      outFormattedBlock.Append("=");
      outFormattedBlock.Append(tmpAutoname.chars());
    };

  outFormattedBlock.Append();


  

  while(inElement.children.GetLength())
    {
      DTDElement& tmpElement=inElement.children.GetFirst();

      if ( tmpElement.name == TOK_SYNOPSIS )
	{
	  FormatSYNOPSIS(tmpElement,0,inWidth-inMargin , 0,tmpBlock);
	}
      else if ( tmpElement.name == TOK_DESC )
	{
	  FormatDESC(tmpElement,0,inWidth-inMargin , 0,tmpBlock);
	};
      

      inElement.children.RemoveTop();

      if ( ! tmpBlock.IsEmpty())
	{
	  outFormattedBlock.Paste(inMargin,outFormattedBlock.Height(),tmpBlock);
	  
	  if (inElement.children.GetLength())
	    {
	      outFormattedBlock.Append();
	    };
	};
    };

};



/*
**
** DESC: Formats PARAMLIST block
**
**
**
** PRE: inElmenet contains info on single param
**      inHeadre header to be put in formatted block
**      inDetailHeader header for the whole detail block 
**      inDetailElemHeader header for single param info
**      inWidth width of formatted block
**      inMarginWidth margin in formatted block
**      detailFlag 1 creat detailed info 0 no detailed info
**      outFormattedBlock formatted block
**      outParamStr param info later used to generate synopsis
**                  
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
**
*/


void AhelpDoc::FormatPARAMLIST(DTDElement& inElement,const char* inHeader, const char* inDetailHeader,const char* inDetailElemHeader,int inWidth, int inMarginWidth, int detailFlag,TextBlock& outFormattedBlock,AhelpStrings& outParamStr)
{

  int indx=1;
  AhelpList<AhelpStrings> atts;
  TableData paramTblData(10);
  TextBlock tmpBlock;
  TextBlock tmpTitle;
  TextBlock tmpParamBlock;
  AhelpStrings tmpHeader;
  int detailMargin=4;
  int tmpWidth;

  atts.Append("NAME");
  atts.Append("TYPE");
  atts.Append("FILETYPE");
  atts.Append("DEF");
  atts.Append("MIN");
  atts.Append("MAX");
  atts.Append("UNITS");
  atts.Append("REQD");
  atts.Append("STACKS");
  atts.Append("AUTONAME");
  paramTblData.SetHeader(atts);
  atts.RemoveAll();


  outFormattedBlock.Clear();
  if ( inElement.IsEmpty())
    return;


  outFormattedBlock.Resize(inWidth);

  outParamStr.clear();
  tmpBlock.Resize(inWidth);
  
  while( inElement.children.GetLength() )
    {
      DTDElement& tmpElement = inElement.children.GetFirst();
      tmpHeader.clear();
      tmpHeader.append(indx++);
      tmpHeader.append(". ");
      tmpHeader.fill(detailMargin);
      tmpHeader.append(inDetailElemHeader);

      tmpParamBlock.Clear();
      FormatPARAMDetail(tmpElement,tmpHeader.chars(),inWidth-2*inMarginWidth,detailMargin,atts,tmpParamBlock,outParamStr);
      tmpBlock.Paste(inMarginWidth,tmpBlock.Height(),tmpParamBlock);
      paramTblData.AddRow(atts);
      atts.RemoveAll();
      inElement.children.RemoveTop();
      if ( inElement.children.GetLength() )
	{
	  tmpBlock.Append();
	};
    };

  TextBlockTable paramsTable(paramTblData,inWidth);
  int curWidth  = paramsTable.Width();



  if ( inHeader != 0 )
    {
      tmpTitle.Set(inWidth,inHeader);
      outFormattedBlock.Append(tmpTitle);
      outFormattedBlock.Append();
    };

  outFormattedBlock.Paste((inWidth - curWidth)/2,outFormattedBlock.Height(),paramsTable);

  if ( detailFlag )
    {
      outFormattedBlock.Append();
      if ( inDetailHeader != 0)
	{
	  tmpTitle.Set(inWidth,inDetailHeader);
	  outFormattedBlock.Append(tmpTitle);
	  outFormattedBlock.Append();
	};
      outFormattedBlock.Append(tmpBlock);
    };

};


/*
**
** DESC: Formats PARAM block that has name attribute inName
**
**
**
** PRE: inElement contains PARAMS blocks  info
**      inHeader header for the formatted block
**      inName name of searched attribute
**      inWidth width of formatted block
**      inMarginWidth margin width in formatted block
**      outFormattedBlock formatted block
**
**
**
** POST: Selected PARAM block is formatted and put into outFormattedBlock
**
**
*/


void AhelpDoc::FormatSelectedPARAM(DTDElement& inElement,const char* inHeader,AhelpStrings inName,int inWidth, int inMarginWidth, TextBlock& outFormattedBlock)
{

  AhelpStrings tmpName;
  AhelpList<AhelpStrings> atts;
  outFormattedBlock.Clear();
  outFormattedBlock.Resize(inWidth);
  AhelpStrings tmpHeader;
  AhelpStrings tmpString;


  while( inElement.children.GetLength() )
    {
      DTDElement& tmpElement = inElement.children.GetFirst();

      AhelpListIterator<AhelpStrings> attIter(&(tmpElement.attributes));


      GetAttribute(tmpElement,ATT_NAME,tmpName);

	  
      CompactString(tmpName);
      CompactString(inName);
      if ( tmpName == inName)
	{
	  formattedBlocksCount++;
	  FormatPARAMDetail(tmpElement,inHeader,inWidth-2*inMarginWidth,inMarginWidth,atts,outFormattedBlock,tmpString);
	  break;
	};
      atts.RemoveAll();
      inElement.children.RemoveTop();

    };
};


/*
**
** DESC: Formats PASSTHRU block
**
**
**
** PRE: inElement contains block info
**
**
** POST: formatted block is put in inElement.data member as Strings
**
**
**
*/



void AhelpDoc::FormatPASSTHRU(DTDElement& inElement)
{   
  
  if ( inElement.children.GetLength() )
    {
      inElement.data.clear();
      inElement.children.RemoveAll();
    };
};



/*
**
** DESC: Formats ROW block
**
**
**
** PRE: inElement contains block info
**
**
** POST: formatted components of the block are put in outRow list
**
**
**
*/


void AhelpDoc::FormatROW(DTDElement& inElement, AhelpList<AhelpStrings>& outRow)
{  
  TextBlock tmpBlock;
  outRow.RemoveAll();


  while(inElement.children.GetLength())
    {
      DTDElement& tmpElement=inElement.children.GetFirst();
      FormatDATA(tmpElement);
      outRow.Append(tmpElement.data);
      inElement.children.RemoveTop();
    };
  
};


/*
**
** DESC: Formats SEEALSO block
**
**
**
** PRE: inElement contains SEEALSO info
**      inHeader header for the formatted block
**      inWidth width of formatted block
**      inMargin margin width in formatted block
**      outFormattedBlock formatted block
**
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
**
*/

void AhelpDoc::FormatSEEALSO(const char* inHeader,int inWidth, int inMargin,TextBlock& outFormattedBlock)
{
  AhelpDB tmpDB(getenv("ASCDS_INSTALL"));
  AhelpStringsList tmpList,tmpList1;
  AhelpStringsList resultList,finalList;
  AhelpStrings tmpString;
  AhelpStrings tmpContext;
  AhelpStrings tmpSubjects;
  TableData tmpTableData(2);
  AhelpStringsList tmpRow;
  int seealsoCount=0;

  if (  formatContextIndx < displayseealsogroups.GetLength() )
    {
      CompactString(displayseealsogroups.GetIndex(formatContextIndx));
      tmpList=displayseealsogroups.GetIndex(formatContextIndx).Split();
    };

  if ( formatContextIndx < seealsogroups.GetLength() )
    {
      CompactString(seealsogroups.GetIndex(formatContextIndx));
      tmpList1=seealsogroups.GetIndex(formatContextIndx).Split();
    };


  AhelpStringsListIterator tmpIter(&tmpList1);
  for(tmpIter.First();!tmpIter.IsDone();tmpIter.Next())
    {
      tmpList.Append(tmpIter.GetCurrent());
    };

  tmpList1.RemoveAll();

  outFormattedBlock.Clear();
  tmpList.Sort(TRUE,TRUE);

  AhelpStringsListIterator elIter(&tmpList);
  for(elIter.First();!elIter.IsDone();elIter.Next())
    {
      if ( tmpDB.FindSEEALSO(elIter.GetCurrent(),resultList) > 0 )
	{
	  finalList.Append(resultList);
	};
    };
  

  finalList.Sort(TRUE,TRUE);

  if ( finalList.GetLength())
    {
      
      AhelpStringsListIterator elIter(&finalList);
      
      for(elIter.First();!elIter.IsDone();elIter.Next())
	{
	  AhelpStringsList tmpPair=elIter.GetCurrent().Split();

	  if (( tmpPair.GetFirst() == context.GetIndex(formatContextIndx)) && (tmpPair.GetLast() == key))
	    {
	      continue;
	    };

	  if ( tmpContext != tmpPair.GetFirst())
	    {
	      if (! tmpSubjects.isEmpty())
		{

		  tmpRow.Append(tmpContext);
		  tmpRow.Append(tmpSubjects);
		  tmpTableData.AddRow(tmpRow);
		  tmpRow.RemoveAll();

		  tmpSubjects.clear();
		};

	      tmpContext = tmpPair.GetFirst();
	    };
	  if (! tmpSubjects.isEmpty() )
	    {
	      tmpSubjects.append(", ");
	    };
	  tmpSubjects.append(tmpPair.GetLast());
	  seealsoCount++;
	};

      if ( seealsoCount)
	{
	  
	  tmpRow.Append("CONTEXT");
	  tmpRow.Append("SUBJECTS");
	  tmpTableData.SetHeader(tmpRow);
	  tmpRow.RemoveAll();
	  
	  outFormattedBlock.Resize(inWidth);
	  
	  if ( inHeader != 0)
	    {
	      outFormattedBlock.Set(inWidth,inHeader);
	      outFormattedBlock.Append();
	    };
	  
	  tmpRow.Append(tmpContext);
	  tmpRow.Append(tmpSubjects);
	  tmpTableData.AddRow(tmpRow);
	  tmpRow.RemoveAll();
	  
	  TextBlockTable tmpTable(tmpTableData,inWidth-2*inMargin,1,1);
	  outFormattedBlock.Paste(inMargin,outFormattedBlock.Height(),tmpTable);
	};
    };



};



/*
**
** DESC: Formats QEXAMPLELIST block
**
**
**
** PRE: inElement contains QEXAMPLELIST info
**      inHeader header for the formatted block
**      inExampleHeader haeder for each EXAMPLE block
**      inWidth width of formatted block
**      inMarginWidth margin width in formatted block
**      outFormattedBlock formatted block
**
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
**
*/


void AhelpDoc::FormatQEXAMPLELIST(DTDElement& inElement,const char* inHeader,const char* inExampleHeader,int inWidth, int inMarginWidth,TextBlock& outFormattedBlock)
{
  
  int indx=1;
  AhelpStrings tmpHeader;
  TextBlock tmpBlock;

  outFormattedBlock.Clear();
  if ( inElement.IsEmpty())
    return;

  
  outFormattedBlock.Resize(inWidth);

  if ( inHeader != 0)
    {
      outFormattedBlock.Set(inWidth,inHeader);
      outFormattedBlock.Append();
    };


  while( inElement.children.GetLength())
    {
      DTDElement& tmpElement = inElement.children.GetFirst();
      tmpHeader.clear();
      tmpHeader.append(inExampleHeader);
      tmpHeader.append(" ");
      tmpHeader.append(indx);
      
      tmpBlock.Clear();

      FormatQEXAMPLE(tmpElement,tmpHeader.chars(),inWidth-inMarginWidth*2,0,tmpBlock);

      inElement.children.RemoveTop();

      if ( ! tmpBlock.IsEmpty())
	{
	  indx++;
	  outFormattedBlock.Paste(inMarginWidth,outFormattedBlock.Height(),tmpBlock);
	  
	  if (inElement.children.GetLength())
	    {
	      outFormattedBlock.Append();
	    };
	};
    };
  
};


/*
**
** DESC: Formats QEXAMPLE block
**
**
**
** PRE: inElement contains QEXAMPLE info
**      inHeader header for the formatted block
**      inWidth width of formatted block
**      inMargin margin width in formatted block
**      outFormattedBlock formatted block
**
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
**
*/

void AhelpDoc::FormatQEXAMPLE(DTDElement& inElement, const char* inHeader,int inWidth, int inMargin,TextBlock& outFormattedBlock)
{  
  outFormattedBlock.Clear();
  if ( inElement.IsEmpty())
    return;

  outFormattedBlock.Resize(inWidth);
  TextBlock tmpBlock;

  if ( inHeader != 0 )
    {    
      TextBlock tmpBlock(inWidth,inHeader);
      outFormattedBlock.Append(tmpBlock);
      outFormattedBlock.Append();
    };


  while(inElement.children.GetLength())
    {
      DTDElement& tmpElement=inElement.children.GetFirst();

      tmpBlock.Clear();
      if ( tmpElement.name == TOK_SYNTAX )
	{
	  FormatSYNTAX(tmpElement,0,inWidth-2*inMargin,0,tmpBlock);
	}
      else if ( tmpElement.name == TOK_DESC )
	{
	  FormatDESC(tmpElement,0,inWidth-2*inMargin,0,tmpBlock);
	};

      inElement.children.RemoveTop();
       
      if ( ! tmpBlock.IsEmpty())
	{
	  outFormattedBlock.Paste(inMargin,outFormattedBlock.Height(),tmpBlock);
	  
	  if (inElement.children.GetLength())
	    {
	      outFormattedBlock.Append();
	    };
	};
    };
 
  inElement.children.RemoveAll();
 
};



/*
**
** DESC: Formats SYNOPSIS block
**
**
**
** PRE: inElement contains SYNOPSIS info
**      inHeader header for the formatted block
**      inWidth width of formatted block
**      inMargin margin width in formatted block
**      outFormattedBlock formatted block
**
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
**
*/

void AhelpDoc::FormatSYNOPSIS(DTDElement& inElement, const char* inHeader,int inWidth, int inMargin,TextBlock& outFormattedBlock)
{
  outFormattedBlock.Clear();
  if ( inElement.IsEmpty())
    return;


  outFormattedBlock.Resize(inWidth);
  if ( inHeader != 0 )
    { 
      TextBlock tmpBlock(inWidth,inHeader);
      outFormattedBlock.Append(tmpBlock);
      outFormattedBlock.Append();
    };
  
  if ( inElement.children.GetLength() )
    {
      TextBlock tmpBlock;

      ConcatPCDATAandCONDTEXT(inElement);
      FormatPCDATA(inElement.children.GetFirst());

      tmpBlock.Set(inWidth-2*inMargin,inElement.children.GetFirst().data.chars());
      outFormattedBlock.Paste(inMargin,outFormattedBlock.Height(),tmpBlock);
      inElement.children.RemoveAll();
    };
  
};



/*
**
** DESC: Formats SYNTAX block
**
**
**
** PRE: inElement contains SYNTAX info
**      inHeader header for the formatted block
**      inWidth width of formatted block
**      inMargin margin width in formatted block
**      outFormattedBlock formatted block
**
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
**
*/

int AhelpDoc::FormatSYNTAX(DTDElement& inElement, const char* inHeader,int inWidth, int inMargin,TextBlock& outFormattedBlock)
{   
  outFormattedBlock.Clear();

  if ( inElement.IsEmpty())
    return 0;

  outFormattedBlock.Resize(inWidth); 
  TextBlock tmpBlock;
  int status = 0;

  if ( inHeader != 0 )
    { 
      TextBlock tmpBlock(inWidth,inHeader);
      outFormattedBlock.Append(tmpBlock);
      outFormattedBlock.Append();
    };
  
  while(inElement.children.GetLength())
    {
      DTDElement& tmpElement=inElement.children.GetFirst();
      
      tmpBlock.Clear();
      FormatLINE(tmpElement,inWidth-2*inMargin,tmpBlock);
      
      if (! tmpBlock.IsEmpty())
	{
	  status=1;
	  
	};
      
      outFormattedBlock.Paste(inMargin,outFormattedBlock.Height(),tmpBlock); 

      inElement.children.RemoveTop();
    };

  return status;
  
};


/*
**
** DESC: Formats SYNTAX block
**
**
**
** PRE: inParamStr contains SYNTAX info
**      inHeader header for the formatted block
**      inWidth width of formatted block
**      inMargin margin width in formatted block
**      outFormattedBlock formatted block
**
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
**
*/


void AhelpDoc::FormatSYNTAX(const AhelpStrings& inParamStr, const char* inHeader,int inWidth, int inMargin,TextBlock& outFormattedBlock)
{
  TextBlock tmpBlock;
  int tmpRow;
  int tmpWidth;

  outFormattedBlock.Clear();
  if ( inParamStr.isEmpty())
    return;

  outFormattedBlock.Resize(inWidth); 
  
  if ( inHeader != 0 )
    { 
      TextBlock tmpBlock(inWidth,inHeader);
      outFormattedBlock.Append(tmpBlock);
      outFormattedBlock.Append();
    };

  if (! inParamStr.isEmpty())
    {
      tmpWidth = inWidth - key.length()-2*inMargin+1;
      if ( tmpWidth < 0 )
	tmpWidth = 0;
      tmpBlock.Clear();
      tmpBlock.Resize(tmpWidth);
      tmpBlock.Append(inParamStr.chars());
      tmpRow = outFormattedBlock.Height();
      outFormattedBlock.Paste(inMargin,tmpRow,key.chars());
      outFormattedBlock.Paste(inMargin+key.length()+1,tmpRow,tmpBlock);
    };


};


/*
**
** DESC: Formats TABLE block
**
**
**
** PRE: inElement contains TABLE block info
**      inWidth width of formatted block
**      outFormattedBlock formatted block
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
*/


void AhelpDoc::FormatTABLE(DTDElement& inElement, int inWidth,TextBlock& outFormattedBlock)
{ 
  outFormattedBlock.Clear();


  outFormattedBlock.Resize(inWidth);
  TextBlock tmpBlock;
  AhelpList<AhelpStrings> tmpRow;
  TableData* tableDataPtr=0;
  int currWidth;
  int headerInitFlag=0;

  while(inElement.children.GetLength())
    {
      DTDElement& tmpElement=inElement.children.GetFirst();
      
      if ( tmpElement.name == TOK_CAPTION )
	{
	  FormatCAPTION(tmpElement,inWidth,tmpBlock);
	  currWidth = tmpBlock.Width();
	  outFormattedBlock.Paste((inWidth-currWidth)/2,outFormattedBlock.Height(),tmpBlock);  
	}
      else if ( tmpElement.name == TOK_ROW )
	{
	  FormatROW(tmpElement,tmpRow);
	  if (! tableDataPtr )
	    {
	      tableDataPtr = new TableData(tmpRow.GetLength());
	    };
	  if ( !headerInitFlag )
	    {
	      tableDataPtr->SetHeader(tmpRow);
	      headerInitFlag=1;
	    }
	  else
	    {
	      tableDataPtr->AddRow(tmpRow);
	    };
	  tmpRow.RemoveAll();

	};
      
      inElement.children.RemoveTop();
      
    };
 
  if ( tableDataPtr)
    {
      TextBlockTable tmpTable(*tableDataPtr,inWidth);
      
      currWidth = tmpTable.Width();
      
      outFormattedBlock.Paste((inWidth-currWidth)/2,outFormattedBlock.Height(),tmpTable);
    };

  delete tableDataPtr;
};



/*
**
** DESC: Formats URL block
**
**
**
** PRE: inElement contains URL info
**
**
** POST: info in inElement is formatted and put into inElement.data member
**
**
**
*/


void AhelpDoc::FormatURL(DTDElement& inElement)
{

     
  if ( inElement.children.GetLength() )
    {
      ConcatPCDATAandCONDTEXT(inElement);
      FormatPCDATA(inElement.children.GetFirst());
      inElement.data=inElement.children.GetFirst().data;
      inElement.children.RemoveAll();
    };
};



/*
**
** DESC: Formats VERBATIM block
**
**
**
** PRE: inElement contains VERBATIM info
**      outFormattedBlock formatted block
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
**
*/


void AhelpDoc::FormatVERBATIM(DTDElement& inElement,TextBlock& outFormattedBlock)
{    

  if ( inElement.children.GetLength() )
    {
      ConcatPCDATAandCONDTEXT(inElement);
      outFormattedBlock = inElement.children.GetFirst().data.chars();
      inElement.children.RemoveAll();
    };
};



/*
**
** DESC: Formats VERSION block
**
**
**
** PRE: inElement contains VERSION info
**      inHeader header for the formatted block
**      inWidth width of formatted block
**      inMargin margin width in formatted block
**      outFormattedBlock formatted block
**
**
**
** POST: info in inElement is formatted and put into outFormattedBlock
**
**
**
*/

void AhelpDoc::FormatVERSION(DTDElement& inElement, const char* inHeader,int inWidth, int inMargin,TextBlock& outFormattedBlock)
{ 
  outFormattedBlock.Clear();
  if ( inElement.IsEmpty())
    return;


  outFormattedBlock.Resize(inWidth);
  if ( inHeader != 0 )
    {     
      TextBlock tmpBlock(inWidth,inHeader);
      outFormattedBlock.Append(tmpBlock);
      outFormattedBlock.Append();
    };
  
  if ( inElement.children.GetLength() )
    {
      TextBlock tmpBlock; 
      ConcatPCDATAandCONDTEXT(inElement);
      
      FormatPCDATA(inElement.children.GetFirst());

      tmpBlock.Set(inWidth-2*inMargin,inElement.children.GetFirst().data.chars());
      outFormattedBlock.Paste(inMargin,outFormattedBlock.Height(),tmpBlock); 
      inElement.children.RemoveAll();
    };

};



/*
**
** DESC: Formats XMLONLY block
**
**
**
** PRE: inElement contains XMLONLY info
**
**
** POST: info in inElement is formatted and put into inElement.data member
**
**
**
*/


void AhelpDoc::FormatXMLONLY(DTDElement& inElement)
{
 
  if ( inElement.children.GetLength() )
    {     
      ConcatPCDATAandCONDTEXT(inElement);
      FormatPCDATA(inElement.children.GetFirst());
      inElement.data = inElement.children.GetFirst().data;
      inElement.children.RemoveAll();
    };
};

/*
**
** DESC: Formats PCDATA 
**
**
**
** PRE: inElement contains PCDATA in data member
**
**
** POST: inElement.data is compacted
**
**
**
*/


void AhelpDoc::FormatPCDATA(DTDElement& inElement,int inRemoveLTWspaces)
{ 
  CompactString(inElement.data,inRemoveLTWspaces);
};




void AhelpDoc::FilterContext(DTDElement& inElement)
{

  AhelpListIterator<DTDElement> childIter(&(inElement.children));
  
  childIter.First();


  while(!childIter.IsDone())
    {
      if ( childIter.GetCurrent().name != TOK_PCDATA )
	{
	  if ( CheckContext(childIter.GetCurrent()) )
	    {
	      FilterContext( childIter.GetCurrent());
	    }
	  else
	    {
	      childIter.GetCurrent().children.RemoveAll();
	      childIter.RemoveCurrent();
	      continue;
	    }
	};
      childIter.Next();
    };
}

void AhelpDoc::ConcatPCDATAandCONDTEXT(DTDElement& inElement)
{

  AhelpListIterator<DTDElement> childIter(&(inElement.children));
  
  childIter.First();


  while(!childIter.IsDone())
    {
      if ( childIter.GetCurrent().name == TOK_CONDTEXT )
	{
	  childIter.GetCurrent().name = TOK_PCDATA;
	  childIter.GetCurrent().data=childIter.GetCurrent().children.GetFirst().data;
	  childIter.GetCurrent().children.RemoveAll();
	};
      childIter.Next();
    };

  for(childIter.First();!childIter.IsDone();childIter.Next())
    { 
      DTDElement& tmpElement=  childIter.GetCurrent();
      if ( tmpElement.name == TOK_PCDATA )
	{
	  childIter.Next();
	  
	  while (! childIter.IsDone() )
	    {
	      DTDElement& tmpElement1 = childIter.GetCurrent();
	      
	      if ( tmpElement1.name == TOK_PCDATA )
		{
		  tmpElement.data.append(tmpElement1.data);
		  childIter.RemoveCurrent();
		}
	      else
		break;
	      childIter.Next();
	    }
	  
	};
    };
};




/*
**
** DESC: Compact a String
**
**
**
** PRE: inString string to be compacted
**
**
** POST: white spaces and newlines are compacted and replaced with space
**
**
**
*/


void AhelpDoc::CompactString(AhelpStrings& inString,int inRemoveLTWspaces)
{
   inString.Substitute("\n"," ",1);
   inString.Substitute("\t"," ",1);
   inString.Substitute("\v"," ",1);
   inString.Substitute("\b"," ",1);
   inString.Substitute("\r"," ",1);
   inString.Substitute("\f"," ",1);
   inString.Substitute("\a"," ",1);
   if (inRemoveLTWspaces )
     {
       inString.remove_leading_white_space(1);
       inString.remove_trailing_white_space(1);
     };
   inString.pack_spaces(); 

};


bool AhelpDoc::CheckContext(const AhelpStrings& inContextAttributeString)
{
  
  AhelpStringsList tmpContext=inContextAttributeString.Split('|');
  
  AhelpListIterator<AhelpStrings> contextIter(&tmpContext);
  
  for(contextIter.First();!contextIter.IsDone();contextIter.Next())
    {
      if ( formatContextString == contextIter.GetCurrent() )
	{
	  return true;
	}
    };
  
  return false;


};


bool AhelpDoc::CheckContext(DTDElement& inElement)
{

  AhelpListIterator<AhelpStrings> attIter(&(inElement.attributes));
  
  
  for(attIter.First();!attIter.IsDone();attIter.Next())
    {
      if ( attIter.GetCurrent() == ATT_CONTEXT)
	{
	  attIter.Next();
	  AhelpStrings tmpStr=attIter.GetCurrent();
	  tmpStr.lowercase();

	  if ( CheckContext(tmpStr))
	    return true;
	  else
	    return false;
	}
    }
  
  return true;

}



int AhelpDoc::GetAttribute(DTDElement& inElement, const AhelpStrings& inAttributeString,AhelpStrings& outAttributeValueString)
{
  AhelpListIterator<AhelpStrings> attIter(&(inElement.attributes));
  
  outAttributeValueString="";

  for(attIter.First();!attIter.IsDone();attIter.Next())
    {
      if ( attIter.GetCurrent() == inAttributeString)
	{
	  attIter.Next();
	  outAttributeValueString=attIter.GetCurrent();
	  return 1;
	}
    };


  return 0;
};






/*
**
** DESC: Sets the blocks to be included in formatting of document
**
**
**
** PRE: Has to be called before Format is called
**       inHEADERFlag 
**       inSYNOPSISFlag,
**       inSYNTAXFlag,
**       inADDRESSFlag,
**       inDESCFlag,
**       inQEXAMPLELISTFlag,
**       inPARAMLISTFlag,
**       inPARAMName,
**       inADESCFlag,
**       inADESCTitle,
**       inBUGSFlag,
**       inSEEALSOFlag,
**       inVERSIONFlag,
**       inLASTMODIFIEDFlag 
**
**
** POST: Internal data members are initialized
**       
**
**
*/


void AhelpDoc::SetBlocks(int inHEADERFlag,
			 int inSYNOPSISFlag,
			 int inSYNTAXFlag,
			 int inADDRESSFlag,
			 int inDESCFlag,
			 int inQEXAMPLELISTFlag,
			 int inPARAMLISTFlag,
			 const AhelpStrings& inPARAMName,
			 int inADESCFlag,
			 const AhelpStrings& inADESCTitle,
			 int inBUGSFlag,
			 int inSEEALSOFlag,
			 int inVERSIONFlag,
			 int inLASTMODIFIEDFlag)
{
  HEADERFlag=inHEADERFlag;
  SYNOPSISFlag=inSYNOPSISFlag;
  SYNTAXFlag=inSYNTAXFlag;
  genSYNTAXFlag=0;
  ADDRESSFlag=inADDRESSFlag;
  DESCFlag=inDESCFlag;
  QEXAMPLELISTFlag=inQEXAMPLELISTFlag;
  PARAMLISTFlag=inPARAMLISTFlag;
  ADESCFlag=inADESCFlag;
  BUGSFlag=inBUGSFlag;
  SEEALSOFlag=inSEEALSOFlag;
  VERSIONFlag=inVERSIONFlag;
  LASTMODIFIEDFlag=inLASTMODIFIEDFlag;
  
  selectionPARAMName = inPARAMName;
  selectionADESCTitle= inADESCTitle;
};



/*
**
** DESC:  Sets width and margin for formatted document
**
**
**
** PRE: Has to be called before Format is called
**      inWidth width of formatted doc
**      inMarginRatio ratio of margin size to the width of doc
**
**
** POST: internal data is initialized
**
**
**
*/


void AhelpDoc::SetWidth(int inWidth,float inMarginRatio)
{
  outputWidth = inWidth;
  if ( outputWidth < 0 )
    {
      outputWidth = 0 ;
    };
  marginRatio = inMarginRatio;

};



/*
**
** DESC: Resets AhelpDoc
**
**
**
** PRE: None
**
**
** POST: Width and margin are reset, internal data is cleared
**
**
**
*/


void AhelpDoc::Reset()
{
  DTDElement::Reset();

  outputWidth = AhelpDoc::DEFAULTOUTPUTWIDTH;
  marginRatio = AhelpDoc::DEFAULTMARGINRATIO; 
  /* 
     HEADERFlag=1;
     SYNOPSISFlag=1;
  SYNTAXFlag=1;
  ADDRESSFlag=1;
  DESCFlag=1;
  QEXAMPLELISTFlag=1;
  PARAMLISTFlag=1;
  ADESCFlag=1;
  BUGSFlag=1;
  SEEALSOFlag=1;
  VERSIONFlag=1;
  LASTMODIFIEDFlag=1;

  selectionPARAMName.clear();
  selectionADESCTitle.clear();
  */
  key.clear();
  context.RemoveAll();
  refkeywords.RemoveAll();
  seealsogroups.RemoveAll();

  addressList.RemoveAll();

  formatContextString="";
  formatContextIndx=0;
};





const AhelpStrings AhelpDoc::TOK_CXCHELPTOPICS="CXCHELPTOPICS";
const AhelpStrings AhelpDoc::TOK_ENTRY="ENTRY";
const AhelpStrings AhelpDoc::TOK_ADDRESS="ADDRESS";
const AhelpStrings AhelpDoc::TOK_ADESC="ADESC";
const AhelpStrings AhelpDoc::TOK_BUGS="BUGS";
const AhelpStrings AhelpDoc::TOK_CAPTION="CAPTION";
const AhelpStrings AhelpDoc::TOK_DATA="DATA";
const AhelpStrings AhelpDoc::TOK_DESC="DESC";
const AhelpStrings AhelpDoc::TOK_EQUATION="EQUATION";
const AhelpStrings AhelpDoc::TOK_HREF="HREF";
const AhelpStrings AhelpDoc::TOK_ITEM="ITEM";
const AhelpStrings AhelpDoc::TOK_LABEL="LABEL";
const AhelpStrings AhelpDoc::TOK_LABELTEXT="LABELTEXT";
const AhelpStrings AhelpDoc::TOK_LASTMODIFIED="LASTMODIFIED";
const AhelpStrings AhelpDoc::TOK_LINE="LINE";
const AhelpStrings AhelpDoc::TOK_LIST="LIST";
const AhelpStrings AhelpDoc::TOK_PARA="PARA";
const AhelpStrings AhelpDoc::TOK_PARAM="PARAM";
const AhelpStrings AhelpDoc::TOK_PARAMLIST="PARAMLIST";
const AhelpStrings AhelpDoc::TOK_PASSTHRU="PASSTHRU";
const AhelpStrings AhelpDoc::TOK_ROW="ROW";
const AhelpStrings AhelpDoc::TOK_QEXAMPLE="QEXAMPLE";
const AhelpStrings AhelpDoc::TOK_QEXAMPLELIST="QEXAMPLELIST";
const AhelpStrings AhelpDoc::TOK_SEEALSO="SEEALSO";
const AhelpStrings AhelpDoc::TOK_SYNOPSIS="SYNOPSIS";
const AhelpStrings AhelpDoc::TOK_SYNTAX="SYNTAX";
const AhelpStrings AhelpDoc::TOK_TABLE="TABLE";
const AhelpStrings AhelpDoc::TOK_URL="URL";
const AhelpStrings AhelpDoc::TOK_VERBATIM="VERBATIM";
const AhelpStrings AhelpDoc::TOK_VERSION="VERSION";
const AhelpStrings AhelpDoc::TOK_XMLONLY="XMLONLY";
const AhelpStrings AhelpDoc::TOK_CONDTEXT="CONDTEXT";
const AhelpStrings AhelpDoc::ATT_KEY="key";
const AhelpStrings AhelpDoc::ATT_REFKEYWORDS="refkeywords";
const AhelpStrings AhelpDoc::ATT_CONTEXT="context";
const AhelpStrings AhelpDoc::ATT_SEEALSOGROUPS="seealsogroups";
const AhelpStrings AhelpDoc::ATT_DISPLAYSEEALSOGROUPS="displayseealsogroups";
const AhelpStrings AhelpDoc::ATT_NAME="name";
const AhelpStrings AhelpDoc::ATT_TYPE="type";
const AhelpStrings AhelpDoc::ATT_FILETYPE="filetype";
const AhelpStrings AhelpDoc::ATT_DEF="def";
const AhelpStrings AhelpDoc::ATT_MIN="min";
const AhelpStrings AhelpDoc::ATT_MAX="max";
const AhelpStrings AhelpDoc::ATT_UNITS="units";
const AhelpStrings AhelpDoc::ATT_REQD="reqd";
const AhelpStrings AhelpDoc::ATT_STACKS="stacks";
const AhelpStrings AhelpDoc::ATT_AUTONAME="autoname";
const AhelpStrings AhelpDoc::ATT_TITLE="title";
const AhelpStrings AhelpDoc::ATT_PARAGRAPHNUMBERING="paragraph_numbering";
const AhelpStrings AhelpDoc::ATT_LINK="link";
const AhelpStrings AhelpDoc::EMPTY_STRING="";
const int AhelpDoc::DEFAULTOUTPUTWIDTH=80;
const float AhelpDoc::DEFAULTMARGINRATIO=0.05;
