// 
//  Copyright (C) 2002,2003,2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include <cstdlib>
#include <termios.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <regex.h>
#include "AhelpInterface.hh"
#include "AhelpDB.hh"
#include "XMLFParser.hh"
#include "AhelpDoc.hh"
#include "AhelpUtils.hh"
#include "cxcds_version.h"




/*
** DESC: Constructor 
**
**
** PRE: 
**      
**
**
** POST: 
**
**
*/


AhelpInterface::AhelpInterface (const AhelpStrings& inPath)
{
  if ( inPath.isEmpty())
    {
      installPath = getenv("ASCDS_INSTALL");
    };

  defaultContext=getenv(AhelpInterface::ENVIRON_DEFCONTEXT_VARNAME);
};


/*
** DESC: Destructor
**
**
** PRE: 
**      
**
**
** POST: 
**
**
*/


AhelpInterface::~AhelpInterface()
{


};

  
/*
** DESC: Set default cotext
**
**
** PRE: 
**      
**
**
** POST: 
**
**
*/


void AhelpInterface::SetDefaultContext(const AhelpStrings& inContext)
{
  defaultContext=inContext;

};




/*
** DESC: Set default cotext
**
**
** PRE: 
**      
**
**
** POST: 
**
**
*/


int AhelpInterface::GetAllContexts( int inOutputWidth,AhelpStrings& outInfo)
{
  AhelpDB ahelpDB(installPath);
  AhelpStringsList tmpList;
  int status;


  if (inOutputWidth <=0)
    {
      struct winsize size;
      if ( ioctl(STDIN_FILENO, TIOCGWINSZ, (char *)&size) >= 0 )
	inOutputWidth = size.ws_col;
      if ( inOutputWidth <=0 )
	inOutputWidth=AhelpDoc::DEFAULTOUTPUTWIDTH;
      
    };


  
  outInfo.clear();

  status = ahelpDB.GetAllContexts(tmpList);

  CreateContextList(tmpList,inOutputWidth,outInfo);

  return status;
};


/*
** DESC: Gets help info from xml file 
**
**
** PRE: inPath path of xml file
**      inStyle style for formatting the help
**      inOutputWidth width of output page
**      outInfo formatted help text
**
**
** POST: inPath is parsed according to inStyle and inOutputWidth
**       and result is put inot outInfo
**
*/


int AhelpInterface::GetHelpFromFile(const AhelpStrings& inPath,
				   AhelpInterface::CXCHelpStyle inStyle,
				   int inOutputWidth,
				   AhelpStrings& outInfo)
{

  AhelpDB ahelpDB(installPath);
  XMLFParser  XMLParser;
  AhelpDoc tmpDoc;
  AhelpStringsList tmpOpts=inPath.Split();
  AhelpStrings tmpPath,tmpContext;


  if (!tmpOpts.GetLength())
    return 0;


  tmpPath = tmpOpts.GetIndex(0);

  if ( tmpOpts.GetLength()>1)
    {
      tmpContext=tmpOpts.GetIndex(1);
    }
  else
    {
      tmpContext=defaultContext;
    }

  outInfo.clear();

  if (inOutputWidth <=0)
    {
      struct winsize size;
      if ( ioctl(STDIN_FILENO, TIOCGWINSZ, (char *)&size) >= 0 )
	inOutputWidth = size.ws_col;
      if ( inOutputWidth <=0 )
	inOutputWidth=AhelpDoc::DEFAULTOUTPUTWIDTH;
    };



  if ( ! XMLParser.parse(tmpPath,ahelpDB.DTDPath(),true,tmpDoc) )
    {
      tmpDoc.Format(outInfo,tmpContext);
      return 0;
    };



  switch( inStyle )
    {
    case CXCHELP_SHORT:
      tmpDoc.SetBlocks(1,1,1,0,0,0,0,"",0,"",0,0,0,0);
      break;
    case CXCHELP_MED:
      tmpDoc.SetBlocks(1,1,1,0,0,1,0,"",0,"",0,0,0,0);
      break;
    default:
      tmpDoc.SetBlocks(1,1,1,1,1,1,1,"",1,"",1,1,1,1);
      break;
    }
  tmpDoc.SetWidth(inOutputWidth);
  tmpDoc.Format(outInfo,tmpContext);

  return 1;

};


/*
** DESC: Gets help from ahelp DB
**
**
** PRE: inKey subject( can be string or regular expression of the from
**                          /reg_exp/)
**      inContext context( can be string or regular expression of the from
**                          /reg_exp/)
**      inStyle style for formatting the help
**      inOutputWidth width of output page
**      outInfo formatted help text
**
**
** POST: ahelp DB is searched for help on inKey in context inContext
**       then info is formatted  according to inStyle and inOutputWidth
**       and result is put inot outInfo. 
**
*/

int AhelpInterface::GetHelp(AhelpStrings inKey, 
			   AhelpStrings inContext,
			   AhelpInterface::CXCHelpStyle inStyle,
			   int inOutputWidth,
			   AhelpStrings& outInfo)
{

  AhelpDB ahelpDB(installPath);
  AhelpStringsList tmpList;
  int localFlag=1;
  int status;

  inKey.lowercase();
  inContext.lowercase();

  outInfo.clear();

  if (inOutputWidth <=0)
    {
      struct winsize size;
      if ( ioctl(STDIN_FILENO, TIOCGWINSZ, (char *)&size) >= 0 )
	inOutputWidth = size.ws_col;
      if ( inOutputWidth <=0 )
	inOutputWidth=AhelpDoc::DEFAULTOUTPUTWIDTH;
      
    };


  if (( inKey.isEmpty()) && (inContext.isEmpty()))
    {
      inKey=HELPKEY;
      if ( defaultContext.isEmpty())
	{
	  inContext=HELPCONTEXT;
	}
      else
	{
	  inContext = defaultContext;
	};
    }
  else if (inContext.isEmpty())
    {
      inContext = defaultContext;
    };


  if (( inStyle != CXCHELP_WEB) && ( inStyle != CXCHELP_INTERNET ))
    {

      status = ahelpDB.FindXML(inKey,inContext,tmpList);

      if (( status == 0 ) && (!inContext.isEmpty()) && (! inKey.isEmpty()))
	{
	  status = ahelpDB.FindXML(inKey,"",tmpList);
	};
      

      if ( ( status == 0 ) && (! inKey.isEmpty()))
	{
	  status = ahelpDB.FindXML(inKey,tmpList);
	};

      if ( status == 1)
	{
	  status = GetHelpFromFile(tmpList.GetIndex(0),inStyle,inOutputWidth,outInfo);
	}
      else if ( status == 0 )
	{
	  if ( (status = ahelpDB.FindTEXT(inKey,tmpList)) == 1 )
	    {
	      status = ReadTEXT(tmpList.GetIndex(0),outInfo);
	    } 
	  else if ( status == 0 )
	    {
	      if ( !inKey.isEmpty())
		{
		  status = GetFHelp(inKey,outInfo);
		}
	      else
		{
		  status = GetFHelp(inContext,outInfo);
		};
	    };
	};
    }
  else
    {
      if ( inStyle == CXCHELP_INTERNET)
	{
	  localFlag=0;
	};

      status = ahelpDB.FindHTML(inKey,inContext,localFlag,tmpList);



      if (( status == 0 ) && (!inContext.isEmpty()) && (! inKey.isEmpty()))
	{
	  status = ahelpDB.FindHTML(inKey,"",localFlag,tmpList);
	};
      

      if ( ( status == 0 ) && (! inKey.isEmpty()))
	{
	  status = ahelpDB.FindHTML(inKey,localFlag,tmpList);
	};
      

      if ( status == 1)
	{
	  status = DisplayURL(tmpList.GetIndex(0));
	};

    };


  if ( status > 1 )
    {
      CreateSubjectList(tmpList,inOutputWidth,outInfo);
    }
  else if ( status == 0 )
    {
      outInfo="CXC Help: No ";
      if ( inStyle == CXCHELP_INTERNET)
	{
	  outInfo.append("internet ");
	}
      else if (inStyle == CXCHELP_WEB) 
	{
	  outInfo.append("local html ");
	};

      outInfo.append("info found for ");

      if ( !inKey.isEmpty())
	{
	  outInfo.append("subject '");
	  outInfo.append(inKey);
	  outInfo.append("' ");
	};
      if ( !inContext.isEmpty())
	{
	  outInfo.append("context '");
	  outInfo.append(inContext);
	  outInfo.append("' ");
	};
    };

  return status;

};



/*
** DESC: Gets help from ahelp DB
**
**
** PRE: inKeyword subject ( can be string or regular expression of the from
**                          /reg_exp/)
**      inStyle style for formatting the help
**      inOutputWidth width of output page
**      outInfo formatted help text
**
**
** POST: ahelp DB is searched for help on inKeyword
**       then info is formatted  according to inStyle and inOutputWidth
**       and result is put inot outInfo. 
**
**
*/

int AhelpInterface::GetHelp(AhelpStrings inKeyword, 
			    AhelpInterface::CXCHelpStyle inStyle,
			    int inOutputWidth,
			    AhelpStrings& outInfo)
{
  
  AhelpDB ahelpDB(installPath);
  AhelpStringsList tmpList;
  int status;

  inKeyword.lowercase();

  if (inOutputWidth <=0)
    {
      struct winsize size;
      if ( ioctl(STDIN_FILENO, TIOCGWINSZ, (char *)&size) >= 0 )
	inOutputWidth = size.ws_col;
      if ( inOutputWidth <=0 )
	inOutputWidth=AhelpDoc::DEFAULTOUTPUTWIDTH;
    };


  status = ahelpDB.FindXML(inKeyword,tmpList);

  
  if ( status == 1)
    {
      status = GetHelpFromFile(tmpList.GetIndex(0),inStyle,inOutputWidth,outInfo);

    }
  else if ( status > 1 )
    {
      CreateSubjectList(tmpList,inOutputWidth,outInfo);
    }
  else
    {
      if ( !inKeyword.isEmpty())
	{
	  status = GetFHelp(inKeyword,outInfo);
	};
      
      if ( status <= 0 )
	{
	  outInfo="CXC Help: No info found for ";
	  outInfo.append("keyword '");
	  outInfo.append(inKeyword);
	  outInfo.append("' ");
	};
    };

  return status;

};


/*
** DESC: Gets help from ahelp DB
**
**
** PRE: inKey subject
**      inContext context
**      inElementName name of block to be extracted
**      inAttributeName attribute value for block to be extracted
**      inOutputWidth width of output page
**      outInfo formatted help text
**
**
** POST: ahelp DB is searched for help on inKey in context inContext
**       then block inElementName is extracted whose attribute value 
**       is inAttributeName info is formatted  according to 
**       inOutputWidth and result is put inot outInfo
**
*/


int AhelpInterface::GetHelp(AhelpStrings inKey, 
			   AhelpStrings inContext,
			   const AhelpStrings& inElementName,
			   const AhelpStrings& inAttributeName,
			   int inOutputWidth,
			   AhelpStrings& outInfo)
{
  AhelpDB ahelpDB(installPath);
  XMLFParser  XMLParser;
  AhelpDoc tmpDoc;
  int status;
  AhelpStringsList tmpList;

  inKey.lowercase();
  inContext.lowercase();

  if (( inKey.isEmpty()) && (inContext.isEmpty()))
    {
      inKey=HELPKEY;
      if ( defaultContext.isEmpty())
	{
	  inContext=HELPCONTEXT;
	}
      else
	{
	  inContext = defaultContext;
	};
    }
  else if (inContext.isEmpty())
    {
      inContext = defaultContext;
    };

  if (inOutputWidth <=0)
    {
      struct winsize size;
      if ( ioctl(STDIN_FILENO, TIOCGWINSZ, (char *)&size) >= 0 )
	inOutputWidth = size.ws_col;
      if ( inOutputWidth <=0 )
	inOutputWidth=AhelpDoc::DEFAULTOUTPUTWIDTH;
    };

  outInfo.clear();

  if ( (status = ahelpDB.FindXML(inKey,inContext,tmpList))  <= 0 )
    {

      if (( status == 0 ) && (!inContext.isEmpty()) && (! inKey.isEmpty()))
	{
	  status = ahelpDB.FindXML(inKey,"",tmpList);
	};

      if ( ( status == 0 ) && (! inKey.isEmpty()))
	{
	  status = ahelpDB.FindXML(inKey,tmpList);
	};
      
      if ( status <=0 )
	{
	  outInfo="CXC Help: No info found for ";
	  if ( !inKey.isEmpty())
	    {
	      outInfo.append("subject '");
	      outInfo.append(inKey);
	      outInfo.append("' ");
	    };
	  if ( !inContext.isEmpty())
	    {
	      outInfo.append("context '");
	      outInfo.append(inContext);
	      outInfo.append("' ");
	    };
	  return 0 ;
	}
    };
  
  AhelpStringsList tmpOpts=tmpList.GetIndex(0).Split();
  AhelpStrings tmpPath,tmpContext;

  tmpPath = tmpOpts.GetIndex(0);
  
  tmpContext=tmpOpts.GetIndex(1);


  if ( status == 1)
    {
      
      if ( ! XMLParser.parse(tmpPath,ahelpDB.DTDPath(),true,tmpDoc) )
	{
	  tmpDoc.Format(outInfo,tmpContext);
	  return -1;
	};
  
      if ( inElementName == AhelpDoc::TOK_SYNOPSIS)
	{
	  tmpDoc.SetBlocks(1,1,0,0,0,0,0,"",0,"",0,0,0,0);
	}
      else if (inElementName == AhelpDoc::TOK_SYNTAX)
	{
	  tmpDoc.SetBlocks(1,0,1,0,0,0,0,"",0,"",0,0,0,0);
	}
      else if (inElementName == AhelpDoc::TOK_ADDRESS)
	{
	  tmpDoc.SetBlocks(1,0,0,1,0,0,0,"",0,"",0,0,0,0);
	}
      else if (inElementName == AhelpDoc::TOK_DESC)
	{
	  tmpDoc.SetBlocks(1,0,0,0,1,0,0,"",0,"",0,0,0,0);
	}
      else if (inElementName == AhelpDoc::TOK_QEXAMPLELIST)
	{
	  tmpDoc.SetBlocks(1,0,0,0,0,1,0,"",0,"",0,0,0,0);
	}
      else if ((inElementName == AhelpDoc::TOK_PARAM) || (inElementName == AhelpDoc::TOK_PARAMLIST))
	{
	  tmpDoc.SetBlocks(1,0,0,0,0,0,1,inAttributeName,0,"",0,0,0,0);
	}
      else if (inElementName == AhelpDoc::TOK_ADESC)
	{
	  tmpDoc.SetBlocks(1,0,0,0,0,0,0,"",1,inAttributeName,0,0,0,0);
	}
      else if (inElementName == AhelpDoc::TOK_BUGS)
	{
	  tmpDoc.SetBlocks(1,0,0,0,0,0,0,"",0,"",1,0,0,0);
	}
      else if (inElementName == AhelpDoc::TOK_SEEALSO)
	{
	  tmpDoc.SetBlocks(1,0,0,0,0,0,0,"",0,"",0,1,0,0);
	}
      else if (inElementName == AhelpDoc::TOK_VERSION)
	{
	  tmpDoc.SetBlocks(1,0,0,0,0,0,0,"",0,"",0,0,1,0);
	}
      else if (inElementName == AhelpDoc::TOK_LASTMODIFIED)
	{
	  tmpDoc.SetBlocks(1,0,0,0,0,0,0,"",0,"",0,0,0,1);
	}
      else
	{
	  outInfo="CXC Help: '";
	  outInfo.append(inElementName);
	  outInfo.append("' is not legal block name");
	  status = -1;
	};
    }
  else
    {
      CreateSubjectList(tmpList,inOutputWidth,outInfo);
      status = -1;
    };

  if ( status > 0 )
    {
      tmpDoc.SetWidth(inOutputWidth);
      tmpDoc.Format(outInfo,tmpContext);
      if ( tmpDoc.FormattedBlocksCount() != 1 )
	{
	  outInfo="CXC Help: block '";
	  outInfo.append(inElementName);
	  outInfo.append("'");
	  if ( !inAttributeName.isEmpty() )
	    {
	      outInfo.append(" with attribute value '");
	      outInfo.append(inAttributeName);
	      outInfo.append("'");
	    };
	  outInfo.append(" not found for '");
	  outInfo.append(inKey);
	  outInfo.append("' in context '");
	  outInfo.append(tmpContext);
	  outInfo.append("'.");
	  status=0;
	};
    };
  
  return status;
};



/*
** DESC: Displays specified url in browser 
**
**
** PRE: inUrl url to be displayed
**
**
** POST: Displays inUrl in browser
**       
**       
**
*/

int AhelpInterface::DisplayURL(const AhelpStrings& inUrl)
{

  
  AhelpStrings command;
  AhelpStrings tmpUrl=inUrl;
  int retries=0;
  AhelpStrings currBrowser=GetAhelpResource("HelpBrowser");
  AhelpStrings getFILETmpl=GetAhelpResource("HelpBrowserOpenFILECommand");
  AhelpStrings getURLTmpl=GetAhelpResource("HelpBrowserOpenURLCommand");
  int delay=0;


#ifdef __APPLE__
 
  if ( getFILETmpl.isEmpty())
    {
      getFILETmpl = "open -a Safari file://$ &";
    };
  
  if ( getURLTmpl.isEmpty())
    {
      getURLTmpl = "open -a Safari $ &";
    };

  currBrowser.clear();

#else

  if ( currBrowser.isEmpty() && getFILETmpl.isEmpty()  && getURLTmpl.isEmpty() )
    {
      currBrowser="netscape";
    };
  
  if ( getFILETmpl.isEmpty())
    {
      getFILETmpl = command = "netscape -remote 'openFile($)'";
    };
  
  if ( getURLTmpl.isEmpty())
    {
      getURLTmpl = command = "netscape -remote 'openURL($)'";
    };

#endif  
  



  
  tmpUrl.remove_leading_white_space(1);


  if (  ! currBrowser.isEmpty() )
    {

      if ( AhelpCheckForProcess(currBrowser) == FALSE )
	{
	  //      cout << "Launching help browser...";
	  command = currBrowser;
	  command.append(" &");
	  if ( system(command.chars()) != 0 )
	    {
	      cerr << "Failed to launch browser"<<endl;
	      return -1;
	    }
	  else
	    {
	      command = GetAhelpResource("HelpBrowserDelay");
	      if (! command.isEmpty())
		{
		  delay = atoi(command.chars());
		}
	      else
		{
		  delay = 1;
		};
	      
	      if ( delay < 0 )
		{
		  delay = 1;
		};
	      
	      sleep(delay);
	    };
	  //      cout <<"Done"<<endl;
	};
    };

  if  (tmpUrl.FindStr("http://",TRUE) == 0)
    {
      command = getURLTmpl;
    }
  else
    {
      command = getFILETmpl;

      if ( tmpUrl.FindChar('/') != 0 )
	{
	  char buf[PATH_MAX+1];
	  tmpUrl.prepend("/");
	  tmpUrl.prepend(getcwd(buf,PATH_MAX));
	};
    };
  command.Substitute("\\$","\n\n",1);
  command.Substitute("$",tmpUrl,1);
  command.Substitute("\n\n","$",1);

  command.append(" >/dev/null 2>&1");
  
  retries=0;


  while (retries < MAXTRIES )
    {
      if (system(command.chars()) == 0)
	{
	  //	  cout <<"Done"<<endl;
	  return 1;
	};

      sleep(1);
      retries++;
    }
  
  
  cerr << "Failed to display info"<<endl;
  return -1;
};
 

/*
** DESC: Reads content of inPath file into outInfo Strings
**
**
** PRE: inPath text file to be read
**      outInfo Strings var
**
**
** POST: inPath is read into outInfo
**
**
*/


int AhelpInterface::ReadTEXT(const AhelpStrings& inPath,AhelpStrings& outInfo)
{
  ifstream infile;
  char* inputBuf = new char[INPUTBUFSIZE];
  int status =0;

  infile.open(inPath.chars());
  
  if ( infile)
    {
      while( infile.getline(inputBuf,INPUTBUFSIZE) )
	{
	  outInfo.append(inputBuf);
	  outInfo.append('\n');
	};
      
      status = 1;

    };
  
  infile.close();
  delete[] inputBuf;
  return status;
};


/*
** DESC: Creates tabulated list of rows containig subjec,context,synopsis
**
**
** PRE: inList of rows containing FIELDSEPARATOR separated columns: subject,
**                context,synopsis
**
**
** POST: Table is formatted and put into outInfo as Strings.
**
**
*/

void AhelpInterface::CreateSubjectList(AhelpStringsList& inList,int inOutputWidth,AhelpStrings& outInfo)
{
  
  TableData tmpTableData(3);
  AhelpStringsList tmpRow;
  AhelpStrings tmpString;
  AhelpStrings tmpString2 = " ahelp ";
  TextBlock tmpBlock;

  tmpRow.Append("SUBJECT");
  tmpRow.Append("CONTEXT");
  tmpRow.Append("SYNOPSIS");
  tmpTableData.SetHeader(tmpRow);
  
  
  for(int indx=0;indx<inList.GetLength();indx++)
    {
      tmpRow = inList.GetIndex(indx).Split(AhelpDBTable::FIELDSEPARATOR);
      tmpTableData.AddRow(tmpRow);
      if ( indx == ( inList.GetLength()-1 ) )
	{
	  tmpString = tmpRow.GetIndex(1);
	  tmpString.append(" ");
	  tmpString.append(tmpRow.GetIndex(0));

	  /* For second example on how to use ahelp */
	  tmpString2.append(tmpRow.GetIndex(0));
	};
    };
  TextBlockTable tmpTable(tmpTableData,inOutputWidth,1,2);
  tmpTable.Append();

  tmpBlock.Set(inOutputWidth,"Use \"ahelp subject\" (\"ahelp context subject\" for non-unique cases) e.g.: ");
  tmpTable.Append(tmpBlock);

  tmpString.prepend(" ahelp ");

  /* First example (without context) */
  tmpBlock.Set(inOutputWidth,tmpString2.chars());
  tmpTable.Append(tmpBlock);

  /* Second example (with context) */
  tmpBlock.Set(inOutputWidth,tmpString.chars());
  tmpTable.Append(tmpBlock);
  tmpTable.Str(outInfo);
  
};








/*
** DESC: Creates tabulated list of contexts containig # and context
**
**
** PRE: inList of rows containing context
**
**
** POST: Table is formatted and put into outInfo as Strings.
**
**
*/

void AhelpInterface::CreateContextList(AhelpStringsList& inList,int inOutputWidth,AhelpStrings& outInfo)
{
  
  TableData tmpTableData(2);
  AhelpStringsList tmpRow;
  AhelpStrings tmpString;
  TextBlock tmpBlock;

  tmpRow.Append("#");
  tmpRow.Append("CONTEXT");
  tmpTableData.SetHeader(tmpRow);
  
  
  for(int indx=0;indx<inList.GetLength();indx++)
    {
      tmpString="";
      tmpString.append(indx+1);
      tmpRow.ReplaceIndex(0,tmpString);
      tmpRow.ReplaceIndex(1,inList.GetIndex(indx));
      tmpTableData.AddRow(tmpRow);
    };

  TextBlockTable tmpTable(tmpTableData,inOutputWidth,1,2);
  tmpTable.Append();


  tmpTable.Str(outInfo);
  
};














/*
** DESC: Gets help info from ftools' fhelp
**
**
** PRE: inKey subject
**      outInfo Strings var
**
**
** POST: If ftools are installed fhelp is run and if any info is found
**       it is put into outInfo
**
*/


int AhelpInterface::GetFHelp(const AhelpStrings& inKey,AhelpStrings& outInfo)
{

  return 0;

  // Disabled due to circular dependency of ahelp and fhelp

  AhelpStrings command;
  AhelpStrings file;
  char* tmpPtr;
  int status ;

  tmpPtr =  tempnam(GetAhelpWorkPath(), NULL);
  file=tmpPtr;
  if ( tmpPtr)
    {
      free(tmpPtr);
    };


  if ( inKey.isEmpty())
    {
      return 0;
    };

  command="( fhelp '";

  command.append(inKey);
  

  command.append("' > ");
  command.append(file);
  command.append(" ) 2>/dev/null");

  if ( system(command.chars() ) == 0 )
    {
      status = ReadTEXT(file,outInfo);
    }
  else
    {
      status=0;
    };

  unlink(file.chars());

  return status;
};


/*
** DESC: Displays inPath in netscape browser.  
**
**
** PRE: inPath - Strings object containing path to help file
**      
**
**
** POST: returns 1 on success
**               negative or 0 otherwise.
**       if inPath points to a file with extension .xml then it is
**          parsed and formatted before displaying in the browser.
*/


int AhelpInterface::DisplayHelpFromFile(const AhelpStrings& inPath)
{

  AhelpStrings tmpPath = inPath;
  int status;
  int tmpFileFlag=0;
  int len = tmpPath.length();
  



  tmpPath.remove_trailing_white_space(1);
  if ( tmpPath.FindStr(".xml") == (len-4))
    {
      AhelpStrings outInfo;
      if ( GetHelpFromFile(tmpPath,CXCHELP_LONG,100,outInfo) == 1 )
	{
	  AhelpStrings tmpFileName; 
	  char* tmpPtr;

	  tmpPtr =  tempnam(GetAhelpWorkPath(), NULL);
	  tmpFileName=tmpPtr;
	  if ( tmpPtr)
	    {
	      free(tmpPtr);
	    };

	  ofstream outFile(tmpFileName.chars());

	  if ( outFile)
	    {
	      outFile<<outInfo;
	      outFile.flush();
	      outFile.close();
	      tmpPath= tmpFileName;
	      tmpFileFlag=1;
	    }
	  else
	    {
	      return -1;
	    };
	};
    };

  status = DisplayURL(tmpPath);

  if ( tmpFileFlag)
    {
      unlink(tmpPath.chars());
    };

  return status;
  
};


void GlobalAhelpInterface::PrintSynopsis(AhelpStrings& outInfo)
{
  outInfo.append("\nahelp, version ");
  outInfo.append(CXCDS_VERSION_STRING);
  outInfo.append("\n");
  outInfo.append(CXCDS_COPYRIGHT_STRING);
  outInfo.append("\n\n");
  outInfo.append("Usage: ahelp [-n width ] [-s|-m|-l] [[-c context] [subject]]|[context subject]\n");
  outInfo.append("       ahelp [-n width ] -b block [-t attribute] [[-c context] [subject]]|[context subject] \n");
  outInfo.append("       ahelp [-n width ] -c \n");
  outInfo.append("       ahelp  -k keyword\n");
  outInfo.append("       ahelp [-x default_context] [-n width ] [-s|-m|-l] -f xmlfile\n");
  outInfo.append("       ahelp  -r [path]\n\n");
  outInfo.append("       subject, context and keyword may be strings or extended \n");
  outInfo.append("       regular expressions of the form: '/reg_exp/' or \"/reg_exp/\"\n\n");

};

bool GlobalAhelpInterface::IsOption(const AhelpStrings& inString)
{
  
  if (inString.isEmpty() )
    return false;
  
  if (( inString  ==  "-n" )
      || (inString == "-s")
      || (inString == "-m")
      || (inString == "-l")
      || (inString == "-b")
      || (inString == "-t")
      || (inString == "-c")
      || (inString == "-k")
      || (inString == "-r")
      || (inString == "-d")
      || (inString == "-e") 
      || (inString == "-x"))
    {
      return true;
    }
  
  return false;

};


int GlobalAhelpInterface::GetHelp(AhelpStringsList& inCommandOptions,AhelpStrings& outInfo)
{

  AhelpStrings tmpArgs[2];
  int argIndx;
  AhelpStrings option;
  int outputWidth=0;
  AhelpStrings blockName,attributeValue;
  AhelpInterface::CXCHelpStyle style=AhelpInterface::CXCHELP_LONG;
  AhelpStrings tmpString;
  AhelpStrings optionString;
  regex_t  regExp;
  AhelpDB* ahelpDB;
  bool widthFlag = false;
  bool styleFlag = false;
  bool contextFlag = false;
  bool blockFlag = false;
  bool attFlag = false;
  bool keywordFlag = false;
  bool fileFlag = false;
  bool reindexFlag = false;
  bool arg0Flag=false;
  bool arg1Flag=false;
  bool dumpFlag=false;
  bool browserOutputFlag=false;
  bool defaultContextFlag=false;
  int  status=1;
  AhelpStrings context;
  AhelpStrings defaultContext;
  int argCount;

  Init(); 

  argCount=inCommandOptions.GetLength();


  argIndx=0;

  while((argIndx < argCount) && (status>0))
    {

      status=0;
      option = inCommandOptions.GetIndex(argIndx);
      option.lowercase();


      if (( option == "-n") && (!widthFlag))
	{
	  optionString.append("n");
	  if ( (++argIndx) < argCount )
	    {
	      tmpString = inCommandOptions.GetIndex(argIndx);
	      
	      if ( (tmpString.isNumeric()) && (! IsOption(tmpString)))
		{
		  outputWidth = atoi(tmpString.chars());
		  
		  widthFlag=true;
		  status=1;
		}
	    };
	}
      else if ( (option == "-s") && (! styleFlag))
	{
	  
	  optionString.append("s");
	  style = AhelpInterface::CXCHELP_SHORT;
	  styleFlag=true;
	  status=1;
	}
      else if ((option == "-m") && (! styleFlag))
	{
	  optionString.append("m");
	  style = AhelpInterface::CXCHELP_MED;
	  styleFlag=true;
	  status=1;
	}
      else if ((option == "-l") && (! styleFlag))
	{
	  optionString.append("l");
	  style = AhelpInterface::CXCHELP_LONG;	
	  styleFlag=true;
	  status=1;
	}
      else if ((option == "-b") && (! blockFlag))
	{
	  optionString.append("b");
	  if ( (++argIndx) < argCount )
	    {
	      if ( ! IsOption(inCommandOptions.GetIndex(argIndx)))
		{
		  blockName = inCommandOptions.GetIndex(argIndx);
		  blockFlag=true;
		  status=1;
		}
	    };
	}
      else if  ((option == "-t") && (! attFlag))
	{
	  optionString.append("t");
	  if ( (++argIndx) < argCount )
	    {
	  
	      if ( ! IsOption(inCommandOptions.GetIndex(argIndx)))
		{
		  attributeValue = inCommandOptions.GetIndex(argIndx);
		  attFlag=true;
		  status=1;
		}
	    };
	}
      else if  ((option == "-c") && (! contextFlag))
	{

	  if ( (++argIndx) < argCount )
	    {
	      optionString.append("c");
	      
	      if ( ! IsOption(inCommandOptions.GetIndex(argIndx)))
		{
		  context = inCommandOptions.GetIndex(argIndx);
		}
	    }
	  else
	    {
	      optionString.append("C");
	    };

	  contextFlag=true;
	  status=1;
	  
	}
      else if  ((option == "-k") && (! keywordFlag))
	{
	  optionString.append("k");
	  
	  if ( (++argIndx) < argCount )
	    {
	      if ( ! IsOption(inCommandOptions.GetIndex(argIndx)))
		{
		  tmpArgs[0] = inCommandOptions.GetIndex(argIndx);
		  keywordFlag=true;
		  status=1;
		}
	    };
	}
      else if  (( option == "-r") && (! reindexFlag))
	{
	  optionString.append("r");
	  reindexFlag=true;
	  if ( (++argIndx) < argCount )
	    {

	      if ( ! IsOption(inCommandOptions.GetIndex(argIndx)))
		{
		  tmpArgs[0] = inCommandOptions.GetIndex(argIndx);
		  status=1;
		};
	    }
	  else
	    {
	      tmpArgs[0] = getenv("ASCDS_INSTALL");
	      status=1;
	    };

	}
      else if  ((option == "-f") && (! fileFlag))
	{

	  globalAhelpInterface->SetDefaultContext("");
	  optionString.append("f");
	  if ( (++argIndx) < argCount )
	    {
	      if ( ! IsOption(inCommandOptions.GetIndex(argIndx)))
		{
		  tmpArgs[0] = inCommandOptions.GetIndex(argIndx);
		  fileFlag=true;
		  status=1;
		}
	    };
	}
      else if  ((option == "-d") && (! dumpFlag))
	{
	  optionString.append("d");
	  if ( (++argIndx) < argCount )
	    {
	      if ( ! IsOption(inCommandOptions.GetIndex(argIndx)))
		{
		  tmpArgs[0] = inCommandOptions.GetIndex(argIndx);
		  dumpFlag=true;
		  status=1;
		}
	    };
	}
      else if ( (option == "-e") && (! browserOutputFlag))
	{
	  optionString.append("e");
	  browserOutputFlag=true;
	  status=1;
	}
      else if  ((option == "-x") && (! defaultContextFlag))
	{
	  optionString.append("x");
	  if ( (++argIndx) < argCount )
	    {
	      if ( ! IsOption(inCommandOptions.GetIndex(argIndx)))
		{
		  defaultContext = inCommandOptions.GetIndex(argIndx);
		  defaultContextFlag=true;
		  status=1;
		}
	    };
	}
      else if (( ! IsOption(option)) && (option.chars()[0] != '-'))
	{
	  optionString.append("a");

	  if ( ! arg0Flag )
	    {
	      tmpArgs[0]=option;
	      arg0Flag=true;
	      status=1;
	    }
	  else if ( (! arg1Flag) && (! contextFlag))
	    {
	      tmpArgs[1]=option;
	      arg1Flag=true;
	      status=1;;
	    };
	};
      
      argIndx++;
      
    };

  

  if ( status<=0 )
    {
      PrintSynopsis(outInfo);
      return 2;  
    };


  if ( contextFlag)
    {
      if ( arg0Flag)
	{
	  tmpArgs[1]=tmpArgs[0];
	};
      tmpArgs[0]=context;
    }
  else if ((arg0Flag ) && (!arg1Flag))
    {
      tmpArgs[1]=tmpArgs[0];
      tmpArgs[0].clear();
    };



  for(argIndx=0;argIndx<optionsCount;argIndx++)
    {
      if ( regcomp(&regExp,optionsTab[argIndx],REG_EXTENDED|REG_NOSUB) != 0 )
	{
	  cerr<<"main:  Failed to create reg expressions"<<endl;
	  regfree(&regExp);
	  return 2;
	}
      else
	{

	  if ( !regexec(&regExp,optionString.chars(),0,NULL,0))
	    {
	      regfree(&regExp);
	      break;
	    };
	  regfree(&regExp);
	};
    };


  if ( !defaultContext.isEmpty())
    {
      globalAhelpInterface->SetDefaultContext(defaultContext);
    };

  switch (argIndx)
    {
    case 0:
      status = globalAhelpInterface->GetHelp(tmpArgs[1],tmpArgs[0],style,outputWidth,outInfo);
      break;
    case 1:
      status = globalAhelpInterface->GetHelp(tmpArgs[1],tmpArgs[0],blockName,attributeValue,outputWidth,outInfo);
      break;
    case 2:
      status = globalAhelpInterface->GetHelp(tmpArgs[0],style,outputWidth,outInfo);
      break;
    case 3:
      if ( browserOutputFlag)
	{
	  status = globalAhelpInterface->DisplayHelpFromFile(tmpArgs[0]);
	}
      else
	{
	  status = globalAhelpInterface->GetHelpFromFile(tmpArgs[0],style,outputWidth,outInfo);
	};
      break;
    case 4:
      ahelpDB = new AhelpDB(tmpArgs[0]);
      status = ahelpDB->Reindex();
      delete ahelpDB;
      break;
    case 5:
      ahelpDB = new AhelpDB(getenv("ASCDS_INSTALL"));
      status = ahelpDB->DumpTable(tmpArgs[0]);
      delete ahelpDB;
      break;
    case 6:
      status = globalAhelpInterface->GetAllContexts(outputWidth,outInfo);
      break;
    default:
      PrintSynopsis(outInfo);
      return 2; 
      break;
    };

  return status;
};

const char* AhelpInterface::ENVIRON_DEFCONTEXT_VARNAME="AHELP_DEFCONTEXT";
const int AhelpInterface::MAXTRIES=50;
const int AhelpInterface::INPUTBUFSIZE=1024;
const char* AhelpInterface::HELPKEY="ahelp";
const char* AhelpInterface::HELPCONTEXT="ahelp";
AhelpInterface* GlobalAhelpInterface::globalAhelpInterface=0;
int GlobalAhelpInterface::optionsCount=7;
const char* const  GlobalAhelpInterface::optionsTab[]={ "^x?(n?(s|m|l)?((c?a?)|(a?a?)))*$", 
				   "^x?(n?(bt?)?((c?a?)|(a?a?)))*$",
				   "^x?n?kn?$",
				   "^(x?n?(s|m|l)?e?)*f(x?n?(s|m|l)?e?)*$",
				   "^x?r$",
                                   "^x?d$",
                                   "^n?Cn?$"};
