// 
//  Copyright (C) 2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#ifndef AHELPLIST_HH
#define AHELPLIST_HH

#include <iostream>
#include "Ahelpstd_types.h"

using namespace std;

template<class Type>
class AhelpIterator
{
public:
  // Goes to the first item in the list
  virtual void First() = 0;

  // Goes to the Next item in the list
  virtual void Next() = 0;
  
  // Return a true if done iterating the list
  virtual Boolean IsDone() = 0;

  // Returns the current data item
  virtual Type& GetCurrent() = 0;
};

class AhelpContainerObject
{
  friend ostream& operator<<(ostream&,AhelpContainerObject&);
  friend istream& operator>>(istream&,AhelpContainerObject&);
public:
  // Reads the container from the input stream
  virtual void Read(istream&);

  // Prints the container to the output stream
  virtual void Print(ostream&);
};



template< class Type >
class AhelpListItem 
{
public:
  // Constructor
  AhelpListItem();
  AhelpListItem(const Type &typ);
  
  // Destructor
  ~AhelpListItem();

  // Returns the data 
  Type& GetData() { return data; };

  void SetData(const Type &typ) { data = (Type)typ; };

  // Set the next item
  void SetNext(AhelpListItem< Type > *item) { next = item; };

  // Returns the next item
  AhelpListItem< Type >* GetNext() { return next; };

  // Set the prev item
  void SetPrev(AhelpListItem< Type > *item) { prev = item; };

  // Returns the prev item
  AhelpListItem< Type >* GetPrev() { return prev; };

protected:
  // Objects
  AhelpListItem< Type > *next;
  AhelpListItem< Type > *prev;
  Type data;
};

template < class Type >
class AhelpList;

template< class Type >
class AhelpListIterator : public AhelpIterator< Type >
{
public:
  // Constructor
  AhelpListIterator(AhelpList< Type >*);

  // Goes to the first item in the list
  virtual void First();

  // Goes to the last item in the list
  virtual void Last();

  // Goes to the Next item in the list
  virtual void Next();
  
  // Goes to the Previous item in the list
  virtual void Prev();

  // Return a true if done iterating the list
  virtual Boolean IsDone();

  // Return a true if first item is current item
  virtual Boolean IsFirst();

  // Return a true if last item is current item
  virtual Boolean IsLast();

  // Returns the current data item
  virtual Type& GetCurrent();

  // Replace the current data in the list
  virtual void  ReplaceCurrent(const Type &typ);

  // Removes current data in list - does not delete (need to check this)
  virtual void RemoveCurrentItem();

  // Removes current data in list
  virtual void RemoveCurrent();

  // Inserts the list item typ before the current List item
  virtual void InsertBeforeCurrent(const Type &typ);

  // Inserts the list item typ after the current List item
  virtual void InsertAfterCurrent(const Type &typ);

protected:
  // Objects
  AhelpList< Type > *list;
  AhelpListItem< Type > *current;
};

template< class Type >
class AhelpList : public AhelpContainerObject
{
  friend class AhelpListIterator<Type>;
public:
  // Constructor
  AhelpList();
  AhelpList(const AhelpList<Type> &list);

  // Destructor
  virtual ~AhelpList();

  // Returns the length of the list list
  int GetLength() { return this->length; };

  // Returns the data at index
  Type& GetIndex(int index);

  // Replaces the data at index
  void ReplaceIndex(int index, const Type &typ);

  // Returns the first item
  Type& GetFirst() { return first->GetData(); };

  // Returns the last item
  Type& GetLast() { return last->GetData(); };

  // Appends the data to the end of the list
  int Append(const Type &typ);
  int Append(const AhelpList< Type > &typ);

  // Prepends the data to the top of the list
  int Prepend(const Type &typ);
  
  // Removes data from the top of the list
  void RemoveTop();

  // Removes data from the bottom of the list
  void RemoveBottom();

  // Removes all of the data from the list
  void RemoveAll();

  // Removes the requested element, if it exists
  void Remove(const Type &element);

  // Operations
  virtual AhelpList< Type >& operator=(const AhelpList< Type > &list);

  // Reads the container from the input stream
  virtual void Read(istream&);

  // Prints the container to the output stream
  virtual void Print(ostream&);

protected:
  // Objects
  int length;
  AhelpListItem< Type > *first;
  AhelpListItem< Type > *last;
};

// ************************* Implementation ******************************

template < class Type >
AhelpListItem< Type >::AhelpListItem()
{
  next = NULL;
  prev = NULL;
}

template < class Type >
AhelpListItem< Type >::AhelpListItem(const Type &typ)
{
  next = NULL;
  prev = NULL;
  data = (Type)typ;
}

template < class Type >
AhelpListItem< Type >::~AhelpListItem() { }

template < class Type >
AhelpList< Type >::AhelpList() : AhelpContainerObject()
{
  first = NULL;
  last = NULL;
  length = 0;
}

template < class Type >
AhelpList< Type >::AhelpList(const AhelpList< Type > &list)
{
  first = NULL;
  last = NULL;
  length = 0;
  AhelpListIterator< Type > it((AhelpList< Type >*)&list);
  for(it.First();!it.IsDone();it.Next())
    Append(it.GetCurrent());
}

template < class Type >
AhelpList< Type >::~AhelpList()
{
  while(first)
    RemoveTop();
}

template < class Type >
Type& AhelpList< Type >::GetIndex(int index)
{
  AhelpListItem< Type > *temp = first;

  for(int count=0;count < index;count++)
    temp = temp->GetNext();
  return temp->GetData();
}

template < class Type >
void AhelpList< Type >::ReplaceIndex(int index, const Type &typ)
{
  AhelpListItem< Type > *temp = first;

  for(int count=0;count < index;count++)
    temp = temp->GetNext();
  temp->SetData(typ);
}

template < class Type >
int AhelpList< Type >::Append(const Type &typ)
{
  AhelpListItem< Type > *temp = new AhelpListItem< Type >(typ);
  if(temp)
    {
      if(first && last)
	{
	  last->SetNext(temp);
	  temp->SetPrev(last);
	  last = temp;
	}
      else
	first = last = temp;
      length++;
      return 1;
    }
  return 0;
}

template < class Type >
int AhelpList< Type >::Append(const AhelpList< Type > &typ)
{
  AhelpListIterator< Type > it((AhelpList< Type >*)&typ);
  for(it.First();!it.IsDone();it.Next())
    Append(it.GetCurrent());
  return 1;
}

template < class Type >
int AhelpList< Type >::Prepend(const Type &typ)
{
  AhelpListItem< Type > *temp = new AhelpListItem< Type >(typ);
  if(temp)
    {
      if(first && last)
	{
	  first->SetPrev(temp);
	  temp->SetNext(first);
	  first = temp;
	}
      else 
	first = last = temp;
      length++;
      return 1;
    }
  return 0;
}

template < class Type >
void AhelpList< Type >::Remove(const Type &element)
{

   // Remove the column from our selected columns list
   AhelpListIterator<Type> it(this);
   for (it.First(); !it.IsDone(); it.Next())
	if (it.GetCurrent() == element)
		it.RemoveCurrentItem();
}

template < class Type >
void AhelpList< Type >::RemoveTop()
{
  if(first)
    {
      AhelpListItem< Type > *temp = first;
      first = first->GetNext();
      delete temp;
      if(first)
	first->SetPrev(NULL);
      else
	last = NULL;
      length--;
    }
}

template < class Type >
void AhelpList< Type >::RemoveBottom()
{
  if(last)
    {
      AhelpListItem< Type > *temp = last;
      last = last->GetPrev();
      delete temp;
      if(last)
	last->SetNext(NULL);
      else
	first = NULL;
      length--;
    }  
}

template < class Type >
void AhelpList< Type >::RemoveAll()
{
  while(first)
    RemoveTop();
  first = last = NULL;
}

template < class Type >
AhelpList< Type >& AhelpList< Type >::operator=(const AhelpList< Type > &list)
{
  RemoveAll();
  AhelpListIterator< Type > it((AhelpList< Type >*)&list);
  for(it.First();!it.IsDone();it.Next())
    Append(it.GetCurrent());
  return *this;
}

template < class Type >
void AhelpList< Type >::Print(ostream &oss)
{
  AhelpListIterator< Type > it(this);
  for(it.First();!it.IsDone();it.Next())
    oss << it.GetCurrent() << endl;
}

template < class Type >
void AhelpList< Type >::Read(istream&)
{
  
}

template < class Type >
AhelpListIterator< Type >::AhelpListIterator(AhelpList< Type > *lst) : AhelpIterator< Type >()
{
  list = lst;
  First();
}

template < class Type >
void AhelpListIterator< Type >::First()
{
  current = list->first;
}

template < class Type >
void AhelpListIterator< Type >::Last()
{
  current = list->last;
}

template < class Type >
void AhelpListIterator< Type >::Next()
{
  if(current)    
    current = current->GetNext();
}

template < class Type >
void AhelpListIterator< Type >::Prev()
{
  if(current)    
    current = current->GetPrev();
}

template < class Type >
Boolean AhelpListIterator< Type >::IsDone()
{
  if(current)
    return FALSE;
  return TRUE;
}

template < class Type >
Boolean AhelpListIterator< Type >::IsFirst()
{
  if (IsDone())
    return FALSE;
  if(current->GetPrev())
    return FALSE;
  return TRUE;
}

template < class Type >
Boolean AhelpListIterator< Type >::IsLast()
{
  if (IsDone())
    return TRUE;
  if (current->GetNext())
    return FALSE;
  return TRUE;
}

template < class Type >
Type& AhelpListIterator< Type >::GetCurrent()
{
  return current->GetData();
}

template < class Type >
void AhelpListIterator< Type >::ReplaceCurrent(const Type &typ)
{
  current->SetData(typ);
}

template < class Type >
void AhelpListIterator< Type >::RemoveCurrentItem() {
    AhelpListItem< Type > *tmp2;
    AhelpListItem< Type > *tmp3; 

    if (!current) 
      return;

    tmp2 = current->GetPrev();
    tmp3 = current->GetNext();

    if (tmp2)
      tmp2->SetNext(tmp3);
    else
      list->first = tmp3;
     
    current = tmp2;

    if (tmp3)
      {
      tmp3->SetPrev(tmp2);

      // if previous is null, reset current
      if (!current)
        current = tmp3;
      }
    else
      list->last = current;

    list->length--;
}

template < class Type >
void AhelpListIterator< Type >::RemoveCurrent()
{
   
    AhelpListItem< Type > *tmp2;
    AhelpListItem< Type > *tmp3; 
    AhelpListItem< Type > *temp; 

    if (!current) 
      return;

    tmp2 = current->GetPrev();
    tmp3 = current->GetNext();

    // save pointer to current entry that we want to delete
    temp = current;

    if (tmp2)
      tmp2->SetNext(tmp3);
    else
      list->first = tmp3;
     
    current = tmp2;

    if (tmp3)
      {
      tmp3->SetPrev(tmp2);

      // if previous is null, reset current
      if (!current)
        current = tmp3;
      }
    else
      list->last = current;

    delete temp;
    list->length--;
}


template < class Type >
void AhelpListIterator< Type >::InsertBeforeCurrent(const Type &typ)
{
 
    // allocate a new AhelpListItem< Type >
    AhelpListItem< Type > *temp = new AhelpListItem< Type >(typ); 
  

    AhelpListItem< Type > *tmp_prev;
    AhelpListItem< Type > *tmp_next; 

    if (!current) 
      return;

    tmp_prev = current->GetPrev();
    tmp_next = current;


    if (tmp_prev){
      tmp_prev->SetNext(temp);
      temp->SetPrev(tmp_prev);
    }
    else  // was no previous, current is the 1st
      list->first = temp;
     
    
    // tmp_next was previous current so it must exist    
    tmp_next->SetPrev(temp);
    temp->SetNext(tmp_next);

    list->length++;

}

template < class Type >
void AhelpListIterator< Type >::InsertAfterCurrent(const Type &typ)
{
 
  // allocate a new AhelpListItem< Type >
  AhelpListItem< Type > *temp = new AhelpListItem< Type >((Type)typ); 
  
  
  AhelpListItem< Type > *tmp_prev;
  AhelpListItem< Type > *tmp_next; 
  
  if (!current) 
    return;
  
  tmp_prev = current;
  tmp_next = current->GetNext();
  
  // tmp_prev was the current element so it must exist
  tmp_prev->SetNext(temp);
  temp->SetPrev(tmp_prev);
  
  if (tmp_next){
    tmp_next->SetPrev(temp);
    temp->SetNext(tmp_next);
  } else
    list->last = temp;
  
  
  list->length++;
  
}

#endif
