// 
//  Copyright (C) 2007,2009  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#ifndef __AHELPSTRINGS_HH
#define __AHELPSTRINGS_HH

#include <iostream>

using namespace std;

#include "AhelpList.hh"

class AhelpStringsList;


class AhelpDataObject
{
  friend ostream& operator<<(ostream&,const AhelpDataObject&);
  friend istream& operator>>(istream&, AhelpDataObject&);
public:
  // Print the task to the output stream
  virtual void Print(ostream&) const {}; 

  // Read the task from the input stream
  virtual void Read(istream&)  {};
};


class AhelpStrings : public AhelpDataObject
{ 
public:

  AhelpStrings ();
  AhelpStrings (char);
  AhelpStrings (const char*);
  AhelpStrings (const AhelpStrings&);
  AhelpStrings (const AhelpStringsList&,char delim=' ');
  virtual ~AhelpStrings();

  static AhelpStrings emptyString();

  unsigned int length () const;

  int  count(char) const;
  void append(char);
  void append(const char*);
  void append(const AhelpStrings&); 
  void append(int);
  void append(long);
  void append(unsigned long);
  void append(double);
  void clear ();
  void lowercase ();
  void padL(int count, char padc = ' ');
  void padR(int count, char padc = ' ');
  void fill(int toLength, char padc = ' ');
  void prepend (int);
  void prepend (char);
  void prepend (const char*);
  void prepend (const AhelpStrings&);
  void uppercase ();
  void remove(char delim);
  void remove_rest(char delim );
  void remove_all_spaces();
  void remove_leading_white_space(int removeNewLines = 0); 
  void remove_trailing_white_space(int removeNewLines = 0);

  // packs multiple contiguous spaces into a single space accross string
  void pack_spaces();

  Boolean isNumeric() const;
  Boolean isSignedNumeric() const;
  Boolean isDecimal() const;
  Boolean isSignedDecimal() const;
  Boolean isEmpty() const;
  void chop();
  void Substitute(char newchar, char oldchar);
  void Substitute(const AhelpStrings &oldstr, const AhelpStrings &newstr, 
		  int allOccurences=0);
  void Join(const AhelpStringsList &list,char delim=' ');

  AhelpStringsList Split(const char delim=' ', int keepWhiteSpace = 0) const;


  // Splits one long string into many short ones.  Each short string
  // will be 'len' or less characters, however, we only divide at white
  // space, so most of the short strings will be less than 'len' chars long.
  // Also, in this context, tabs and newlines do not make sense, so they 
  // are removed.
  // 
  // When Splitting, commonly the client code will want the strings
  // to end in newlines, and possibly begin with specific characters 
  // (usually tabs).
  AhelpStringsList Split(int len, 
		    Boolean appendNewLine, 
		    AhelpStrings prependString);

  // Finds the first occurance of the character in the strings.
  // If there are no occurances, then a -1 is returned.
  int FindChar(char delim) const;
  int FindChar(char delim,int nth) const;
  int FindLastChar(char delim) const;
  
  // Finds the first occurance of the string str in the string object
  int FindStr(const char *str) const;
  int FindStr(const char *str, Boolean loosely) const;

  char* chars() const { return theString; } ;
  char* operator()() const;

  // Returns the substring that occurs between index a and b.
  // Returns a empty string if the indexes are invalid.
  AhelpStrings GetSubStr(int a,int b) const;

  // compares String against another string ignoring case and
  // preceeding/trailing whitespace- returns 0 if equivalent 
  int StrcmpLoose(AhelpStrings other, int disregardNewLines = 0) const;
  int StrcmpLoose(const char* other, int disregardNewLines = 0) const;

  AhelpStrings& operator = (char);
  AhelpStrings& operator = (const char*);
  AhelpStrings& operator = (const AhelpStrings&);
  AhelpStrings& operator = (const long val);
  AhelpStrings& operator = (const unsigned long val);
  AhelpStrings& operator = (const double val);
  int operator == (const AhelpStrings&) const;
  int operator == (const char*) const;
  int operator != (const AhelpStrings&) const;
  int operator != (const char*) const;
  int operator < (const AhelpStrings&) const;
  int operator < (const char*) const; 
  int operator > (const AhelpStrings&) const;
  int operator > (const char*) const;  
  int operator <= (const AhelpStrings&) const;
  int operator <= (const char*) const;  
  int operator >= (const AhelpStrings&) const; 
  int operator >= (const char*) const;

  void chopc(AhelpStrings& lastchar); // chop and returns last char as AhelpStrings

  virtual void Print(ostream&) const ;
  virtual void Read(istream&)  ;

protected:
  
  char* theString;
};


class AhelpStringsList : public AhelpList<AhelpStrings>
{
public:
  // Constructor
  AhelpStringsList(char dlim = ',');
  AhelpStringsList(const AhelpStringsList&);

  // Destructors
  ~AhelpStringsList();

  // Operator
  AhelpStringsList& operator=(const AhelpStringsList&);

  // converts all of the strings into a char**
  char** GetChars();

  Boolean Exists(char *);

  // Replace 1st string with 2nd string in list
  void Replace(char *,char *);

  // Sort the list, if order is True -> ascending
  void Sort(Boolean order,Boolean unique=FALSE);

  // Sort the list, if order is True -> ascending
  void Sort1(Boolean order,Boolean unique=FALSE);



  // Prints the contents if the String list to the output stream
  virtual void Print(ostream&);

  // Reads the contents of the String list from the input stream
  virtual void Read(istream&);
private:

  void MergeSort(AhelpStringsList& inList,Boolean order);
  void MergeLists(AhelpStringsList& inList1, AhelpStringsList& inList2, Boolean order,AhelpStringsList& outList);

  // Object
  // The character that will be printed out to seperate one
  // string in the list with the next.
  char delim;
};


class AhelpStringsListIterator : public AhelpListIterator<AhelpStrings>  
{ 
public:
  // Constructor
  AhelpStringsListIterator(AhelpStringsList*);
};


#endif



