// 
//  Copyright (C) 2002,2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include <fstream>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <stdlib.h>
#include <pwd.h>
#include <sys/param.h>
#include "AhelpUtils.hh"

static char workDirValue[FILENAME_MAX] = "";
static char resourceValue[257] = "";

const char REGULAREXPRBRACKET = '/';

/*
**
** Desc: Tries to interpret inString as regular expression of the form
**        '/expression/'
**
**
** Pre: inStrings a string
**
**
** Post: returns extracted regular expression without '/'.
**
**
**
**
**
*/

AhelpStrings GetRegularExpression(const AhelpStrings& inString)
{
  AhelpStrings tmpString="";

  if ( inString.length() > 1 )
    {
      if ( ( inString.FindChar(REGULAREXPRBRACKET) == 0 ) && (inString.FindLastChar(REGULAREXPRBRACKET) == (inString.length() -1 )))
	{
	  tmpString=inString.GetSubStr(1,inString.length() -2); 
	  
	};
    };
  return tmpString;

};

/*
**
** Desc: Post a given warning into std error
**       
**
**
** Pre: warn_txt - warning text, kind - prefix for warning
**
**
** Post: warning is printed to std error
**
**
**
**
**
*/

void AhelpPostWarning(const char *warn_txt, const char *kind)
{
        AhelpStrings msg = kind;
        msg.append(" ");
        msg.append(warn_txt);
        cerr << msg() << endl;
};



// ----------------------------------------------------------
// GetHomeDir()
//
// The routine GetHomeDir returns the home directory of the 
// owner of a running process. It attempts to read the HOME
// environment variable and returns the value as the home
// dir. If the environment variable is not set, the routine
// gets the user id of the process and then uses that info
// to retrieve the home directory from the password file.
// ----------------------------------------------------------
AhelpStrings AhelpGetHomeDir()
{
   AhelpStrings home = getenv("HOME");

   if (home.length() == 0)
   {
      struct passwd *userdata_p = getpwuid(getuid());

      if (userdata_p && userdata_p->pw_dir)
      {
         home = userdata_p->pw_dir;
      }
   }

   return home;
}





// Retrieve an X resource value, using the order:
//
//	$HOME/.CXCdefaults
//	$ASCDS_INSTALL/config/system/CXCdefaults
//
// We should eventually add code to query the X resource database, too.
//
// Returns a pointer to static memory, which will be overwritten each
// call, so user must copy out the value to preserve.

const char *GetAhelpResource(const char *resource)
{
   char filename[FILENAME_MAX] = "";
   AhelpStrings dir = AhelpGetHomeDir();

   // Does the defaults file exist in user's home directory?
   Ahelp_find_file(dir.chars(), ".CXCdefaults", R_OK, filename);

   if (!filename[0])
   {
      // check existing XENVIRONMENT var
      AhelpStrings xenv = getenv("XENVIRONMENT");
      if (xenv != "")
      {
         int pos = xenv.FindStr(".CXCdefaults");
         if (pos-- != -1)
         {
            dir = xenv.GetSubStr(0, pos);
            if (dir.length() == (xenv.length()- 12))
            {
               Ahelp_find_file(dir.chars(), ".CXCdefaults", R_OK, filename);
            }
         }
      }

      if (!filename[0])
      {
         // How about in the installation tree? 
         dir = getenv("ASCDS_INSTALL");
         if (dir != "")
         {
            dir.append("/config/system");
            Ahelp_find_file(dir.chars(), ".CXCdefaults", R_OK, filename);

            // look for defaults file w/o preceding '.'
            if (!filename[0])
            {
               Ahelp_find_file(dir.chars(), "CXCdefaults", R_OK, filename);
            }
         }
      }
   }

   // If the file was found, look within for the requested resource
   if (filename[0])
   {
      #define BUFFSIZE 1000
      char buf[BUFFSIZE+1] = "";
      strcpy(resourceValue,"");

      // Search the file, line by line
      ifstream infile(filename);
      while (infile.good() && !infile.eof() && !resourceValue[0])
      {
         char *substrPos;

         infile.getline(buf,BUFFSIZE);
         if ((substrPos = strstr(buf,resource)) != 0)
         {
            // possibly found resource - make sure its the
            // resource and doesn't just contain a substring
            // (eg. ImagerArgs when looking for Imager)
            AhelpStrings resourceName(substrPos);
            resourceName.remove_rest(':'); 
            resourceName.remove_trailing_white_space();
            if (resourceName == resource)
            {
               // Found the requested resource
               AhelpStrings tempStr(buf);
               tempStr.remove_leading_white_space();
               // Make sure it is not commented out.
               if (tempStr.FindChar('!') != 0)
               {
                  int posn = tempStr.FindChar(':');
                  AhelpStrings browser = tempStr.GetSubStr(posn+1, tempStr.length());
                  browser.remove_leading_white_space();
                  browser.remove_trailing_white_space();
                  strcpy(resourceValue, browser.chars());
               }
            }
         }
      }
   }

   if (resourceValue[0])
      return (const char*)resourceValue;
   else 
      return 0;
}



// ----------------------------------------------------------
// GetAhelpWorkPath()
//
// The routine GetCXCWorkPath returns the working directory that
// CIAO tools should utilize as tmp/scratch space. If the
// ASCDS_WORK_PATH environment variable is not set, the
// routine utilizes the value of ASCDS_TMP. If neither is set
// the path "/tmp" is utilized.
// ----------------------------------------------------------
const char *GetAhelpWorkPath()
{
   AhelpStrings  wdir = getenv("ASCDS_WORK_PATH");
 
   if (wdir.length() == 0)
   {
      wdir = getenv("ASCDS_TMP");

      if (wdir.length() == 0)
      {
         wdir = "/tmp";
      }
   }

   // if multiple paths provided, take first one
   wdir.remove_leading_white_space();
   wdir.remove_rest(';');
   wdir.remove_rest(':');
   wdir.remove_rest(' ');

   strcpy(workDirValue, wdir());
   return (const char*)workDirValue;
}




// ----------------------------------------------------------
// CheckForProcess()
//
//  routine to determine if a process is
//  already running on the current host. It does a ps (aux and
//  ef) to list out proceses and greps for the requested tool.
//  It does not consider parameters of other command line
//  args which may have been used on the running process. A 
//  value of TRUE is returned if a process is detected. 
//
//  If the checkuser flag is set to TRUE or is omitted, the
//  routine will search for the requested process being run 
//  by the current username (specified by the LOGNAME 
//  enviroment variable). If the flag is set to FALSE, the
//  routine will merely check whether the specified process
//  is currently running regardless of who owns it.
// ----------------------------------------------------------
Boolean AhelpCheckForProcess(AhelpStrings name, Boolean checkuser)
{
   char temp[256];
   AhelpStrings username;
   Boolean val = FALSE; 

   name.remove_trailing_white_space();

   if (checkuser)
   {
      username = AhelpGetUserName();

      sprintf(temp, 
         "ps -auxww 2>/dev/null | grep '%s' | grep '%s' | grep -v grep 1> /dev/null 2>&1", 
         name.chars(), username.chars());
   }
   else
   {
      sprintf(temp, 
         "ps -auxww 2>/dev/null | grep '%s'| grep -v grep 1> /dev/null 2>&1", 
         name.chars());
   }

   if (system(temp)) 
   {
      if (checkuser)
      {
         sprintf(temp, 
            "ps -ef 2>/dev/null | grep '%s' | grep '%s' | grep -v grep 1> /dev/null 2>&1", 
            name.chars(), username.chars());
      }
      else 
      {
         sprintf(temp, 
            "ps -ef 2>/dev/null | grep '%s' | grep -v grep 1> /dev/null 2>&1", 
            name.chars());
      }

      if (!system(temp))
      {
         // process already running 
         val = TRUE; 
      } 
   }
   else
   {
      // process already running 
      val = TRUE; 
   } 
 
   return (val); 
}

/* check for name in given path */
Boolean Ahelp_find_file(const char *env, const char *name, const int mode,
		  char *ret_path)
{

  return(Ahelp_find_file_delim(env,name,mode,ret_path,NULL));
}

Boolean Ahelp_find_file_delim(const char *env, const char *name, const int mode,
			char *ret_path, const char *delim)
{
  char    *path;
  char    *save_path = NULL;
  char    *ptr;
  char    buffer[MAXPATHLEN] = "";
  Boolean retval = FALSE;

  if (!name) 
    return retval;

  /* copy the PATH environment variable */
  if (env) {
    path = (char *)calloc(1,strlen(env)+1); 
    save_path = path; 
    strcpy(path,env);
  }
  else {
    path = ".";
  }
  
  /* parse through the path names, separated by colons and */
  /* check each path for the executable file */
  ptr = path;
  while (ptr)
    {
    if (!delim) {
      if ((ptr = strchr(path,':')) || (ptr = strchr(path,' ')) )
        *ptr = '\0';
    }
    else {
      if ((ptr = strchr(path,*delim)) )
        *ptr = '\0';
    }
    sprintf(buffer,"%s/%s",path,name);

    if (access(buffer,mode) == 0)
      {
      retval = TRUE;
      break;
      }
    path = ptr + 1;
    }

  if (save_path)
    free(save_path);
  if (retval && ret_path)
    {
      strcpy(ret_path, buffer);
    }
  
  return retval;
}

// ----------------------------------------------------------
// GetUserName()
//
// The routine GetUserName returns the username of the owner
// of a running process. It attempts to read the LOGNAME 
// environment variable and returns the value as the user
// name. If the environment variable is not set, the routine 
// gets the user id of the process and then uses that info
// to retrieve the username.   
// ----------------------------------------------------------
AhelpStrings AhelpGetUserName()
{
   AhelpStrings username = getenv("LOGNAME");
  
   if (username.length() == 0)
   {
      uid_t userid = getuid();
      struct passwd  *userdata_p;

      userdata_p = getpwuid(userid); 
      if (userdata_p)
      {
         username = userdata_p->pw_name; 
      }  
   }

   return username; 
}
