// 
//  Copyright (C) 2002,2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


 
#include "TextBlock.hh"
#include <string.h>


// default width of text a block
int TextBlock::DEFAULTWIDTH=80;

/*
** DESC: Constructs empty block of width inWidth
**
**
** PRE: inWidth - width of block to be constructed
**
**
**
** POST: Empty block of width inWidth is constructed
**
**
*/

TextBlock::TextBlock(int inWidth)
{
  width = inWidth;
  if ( width < 0 )
    {
      width =0;
    };

};


/*
** DESC: Copy constructor 
**
**
** PRE: inBlock is a TextBlock object
**
**
**
** POST: inBlock is copied to newly created TextBlock
**
**
*/

TextBlock::TextBlock(const TextBlock& inBlock)
{
  char* tmpPtr;
  width = inBlock.width;

  if ( ((TextBlock&)inBlock).lines.GetLength()> 0  )
    {
      AhelpListIterator<char*> lineIter(&(((TextBlock&)inBlock).lines));

      // Copy line by line
      
      for(lineIter.First();!lineIter.IsDone();lineIter.Next())
	{
	  tmpPtr = new char[strlen(lineIter.GetCurrent())+1];
	  strcpy(tmpPtr,lineIter.GetCurrent());
	  lines.Append(tmpPtr);
	};
    };
};


/*
** DESC: Constructs new block of width inWidth and copies
**       inStr to it.
**
**
** PRE: inWidth - width of instantiated block
**      inStr - pointer to a char string
**
**
** POST: TextBlock of width inWidth is created and
**       it contains copy of string inStr
**
**
*/

TextBlock::TextBlock(int inWidth, const char* const inStr)
{
  width = inWidth;
  if ( inWidth <  0 )
    width = 0;

  CreateLines(inStr,width,lines);

};


/*
** DESC: Copy constructor from string buffer
**
**
** PRE: inStr is a pointer to char string
**
**
**
** POST: TextBlock is created and initialized to
**       contain copy of inStr.  Width of instantiated TextBlock
**       is equal length of inStr
**
*/

TextBlock::TextBlock(const char* const inStr)
{
  width = CreateLines(inStr,lines);

};


/*
** DESC: Destructor
**
**
** PRE:
**
**
**
** POST: Instantiation of TextBlock is deleted
**
**
*/

TextBlock::~TextBlock()
{
  
  DeleteLines(lines);
};


/*
** DESC: Pastes content of inBlock at coordinates inCoordX 
**       and inCoordY(the same way images are pasted)
**
** PRE: inBlock is TextBlock
**      inCoordX is x coordinate where upper left corner of inBlock
**                     should be pasted
**      inCoordY is y coordinate where upper left corner of inBlock
**                     should be pasted
**
**
** POST: inBlock is pasted onto this block.  Height of this block
**        is automatically extended if necessary, x coord is not
**        automatically extended and inBlock is truncated if necessary.
**        If truncation occurs ">>" is pasted at the and of truncated lines
*/

void TextBlock::Paste(int inCoordX, int inCoordY, const TextBlock& inBlock)
{
  TextBlock* tmpBlock;

  if ( this == &inBlock)
    {
      tmpBlock = new TextBlock(inBlock);
    }
  else
    {
      tmpBlock = &(TextBlock&)inBlock;
    };

  
  if ( (tmpBlock->width > 0 ) && ( inCoordX < width)   )
    {

      if ( inCoordX < 0 )
	inCoordX = 0;

      if ( inCoordY < 0 )
	inCoordY = 0;

      char* tmpPtr;
      int indx;

      if ( inCoordY >= lines.GetLength())
	{
	  for( indx=0;indx< (inCoordY -lines.GetLength()+1);indx++)
	    {
	      tmpPtr= new char[1];
	      tmpPtr[0]=0;
	      lines.Append(tmpPtr);
	    };
	  
	};

      AhelpListIterator<char*> lineIter(&(tmpBlock->lines));
      indx=inCoordY;
      int tmpSize;
      tmpPtr=0;
      for(lineIter.First();!lineIter.IsDone();lineIter.Next())
	{
	  if (lines.GetLength() <= indx )
	    {
	      Paste(tmpPtr, width,lineIter.GetCurrent(),tmpBlock->Width(),inCoordX);
	      lines.Append(tmpPtr);
	      tmpPtr=0;
	    }
	  else
	    {
	      Paste(lines.GetIndex(indx),width,lineIter.GetCurrent(),tmpBlock->Width(),inCoordX);
	    };
	  indx++;
	};
    };  
  
  if ( this == &inBlock)
    {
      delete tmpBlock;
    }
};


/*
** DESC: Pastes a string int this block starting at coords inCoordX
**       and inCoordY
**
** PRE: inCoordX is x coordinate where pasting starts 
**      inCoordY is y coordinate where pasting starts
**      inStr is pointer to char string
**
** POST: string inStr is pasted into this block at coord inCoordX
**       inCoordY. if necessary this block's height is extended.
**       If inStr does not fit in one line its truncated
**       portion is pasted and ">>" is appended to the end of line
**       
*/

void TextBlock::Paste(int inCoordX, int inCoordY, const char* const  inStr)
{
  if ( ( inStr == 0 ) || (inCoordX< 0) || (inCoordX >=width) )
    return;

  if ( inStr[0] ==0)
    return;


  TextBlock tmpBlock(inStr);

  Paste(inCoordX,inCoordY,tmpBlock);

};


/*
** DESC: Paste character at inCoordX and inCoordY
**
**
** PRE: inCoordX is x coordinate where pasting starts 
**      inCoordY is y coordinate where pasting starts
**      inChar is a character
**
** POST:inChar is pasted into this block at coords
**       inCoordX, inCoordY.  If necessary this block's
**       height is extended automatically.
*/

void TextBlock::Paste(int inCoordX, int inCoordY, char inChar)
{
  char* tmpPtr = new char[2];
  tmpPtr[0] =inChar;
  tmpPtr[1]=0;
  Paste(inCoordX,inCoordY,tmpPtr);
  delete[] tmpPtr;

};

/*
** DESC: Clear an area of the block
**
**
** PRE: inCoordX - x coord of lower left corner of cleard area
**      inCoordY - y coord of lower left corner ... 
**      inWidth  - width of cleared area
**      inHeight - height of cleared area
** 
** POST: a square of dimensions inWidth , inHeight with lower left
**       corner at inCoordX, inCoordY is cleared
**
*/

void TextBlock::Clear(int inCoordX, int inCoordY, int inWidth, int inHeight)
{
  if (( inCoordX < 0 ) || (inCoordX >=width) || (inCoordY < 0 ) || (inCoordY >= lines.GetLength()))
    return;

  int indx,indx1;
  char* tmpPtr;
  int blockEndIndx= inCoordY+inHeight-1;
  int tmpLen;

  if ( blockEndIndx >= lines.GetLength())
    {
      blockEndIndx= lines.GetLength() -1;
    };


  for(indx =inCoordY;indx<=blockEndIndx;indx++)
    {
      tmpPtr = lines.GetIndex(indx);

      tmpLen = inWidth;
      
      if ((inCoordX+inWidth) > strlen(tmpPtr))
	{
	  tmpLen = strlen(tmpPtr) - inCoordX;
	};

      memset(tmpPtr+inCoordX,' ',tmpLen);

    };


};

/*
** DESC: Clear whole block 
**
**
** PRE: 
**
**
**
** POST: This block is cleared
**
**
*/

void TextBlock::Clear()
{
  DeleteLines(lines);
  //  width = 0;
};


/*
** DESC: Resize this block to inWidth
**
**
** PRE: inWidth - new width of this block
**
**
**
** POST: This block is resized to have inWidth.  Height is adjusted
**       accordingly. Contained text is reformatted
**
*/

void TextBlock::Resize(int inWidth)
{
  if (( width == inWidth) || ( inWidth < 0 ))
    return;

  if ( inWidth == 0 )
    {

      DeleteLines(lines);
      width=0;
      return;
    };

  //Put text content into one string

  char* tmpPtr = new char[width*(lines.GetLength()+1)];
  int indx;
  
  

  tmpPtr[0]=0;
  AhelpListIterator<char*> lineIter(&(lines));
  
  for(lineIter.First();!lineIter.IsDone();lineIter.Next())
    {
      strcat(tmpPtr,lineIter.GetCurrent());
      strcat(tmpPtr," ");
      delete[] lineIter.GetCurrent();
    };

  lines.RemoveAll();

  // Reformat it as block of width inWidth

  CreateLines(tmpPtr,inWidth,lines);
  width = inWidth;
  delete[] tmpPtr;

};


/*
** DESC: Paste inSrcStr into inDstStr starting at inDstIndx
**
**
** PRE: inDstSrc      reference to a pointer that will hold pasted string
**      inDstMaxWidth size of memory allocated for inDstSrc
**      inSrcStr      pointer to source string
**      inSrcMaxWidth number of chars to be pasted into dst string
**      inDstIndx     offset in dst string where source string should be
**                    pasted
**
** POST: inSrcStr is pasted into inDstSrc starting at position inDstIndx
**       if necessary inDstSrc is allocated.  If inSrcStr is longer than
**       space where pastins should be done only whatever fits in inDstSrc
**       is pasted. If inSrcMaxWidth is bigger than length of inSrcStr
**       extra space in inDstSrc is padded with spaces
**
*/

void TextBlock::Paste(char*& inDstStr, int inDstMaxWidth, char* inSrcStr,int inSrcMaxWidth,int inDstIndx)
{

  // allocate dst buffor if necessary

  if ( inDstStr == 0 )
    {
      inDstStr= new char[1];
      inDstStr[0]=0;
    };

  // check some constraints

  if (( inDstMaxWidth <=0 ) ||  (inSrcStr==0) || (inDstIndx <0) || (inDstIndx>= inDstMaxWidth) || ( inSrcMaxWidth <= 0))
    return;

  if ( inSrcStr[0] ==  0 )
    return;
  
  int srcLen = strlen(inSrcStr);
  int dstLen = 0;
  int srccpLen=srcLen, dstcpLen=dstLen,spaceCount=0,totalcpLen;
  char * tmpPtr;


  // check if src and dst are the same

  if ( inSrcStr == inDstStr)
    {
      inDstStr = new char[srcLen];
      strcpy(inDstStr,inSrcStr);
    };


  if ( inDstStr != 0 )
    {
	dstLen =strlen(inDstStr);
    };


  totalcpLen = inDstIndx+srcLen;
  if ( totalcpLen > inDstMaxWidth)
    {
      totalcpLen = inDstMaxWidth;
      srccpLen = totalcpLen - inDstIndx;
    };

  if ( totalcpLen == 0 )
    return ;

  // check if we can paste whole src string

  if ( totalcpLen > dstLen)
    {
      // are we gonna manage whole src to dst

      if ( inDstIndx < dstLen )
	{
	  dstcpLen = inDstIndx;
	}
      else
	{
	  dstcpLen = dstLen;
	};
      
      // how many spaces to pad

      if ( inDstIndx > dstLen)
	{
	  spaceCount = inDstIndx - dstLen;
	}
      else
	{
	  spaceCount = 0;
	};
	  
      
      // calc how many chars to copy and allocate tmp space
      
      srccpLen = totalcpLen - spaceCount - dstcpLen;
      tmpPtr=0;
      tmpPtr = new char[totalcpLen+1];
      
      

      if ( dstcpLen > 0  )
	{
	  strncpy(tmpPtr,inDstStr,dstcpLen);
	};
      
      memset(tmpPtr+dstcpLen,' ',spaceCount);

      if ( srccpLen < srcLen)
	{

	  // if can not paste whole buffer
	  // mark that the pasted content has been truncated 

	  if ( srccpLen > 2 )
	    {
	      strncpy(tmpPtr+dstcpLen+spaceCount,inSrcStr,srccpLen-2);
	      tmpPtr[totalcpLen-2]=0;
	      strcat(tmpPtr,">>");
	    };
	}
      else
	{
	  // copy the actuall src string

	  strncpy(tmpPtr+dstcpLen+spaceCount,inSrcStr,srccpLen);
	  tmpPtr[totalcpLen]=0;
	};


      //reallocate dst and copy tmp to dst
      
      if ( inDstStr != 0 )
	{
	    delete[] inDstStr;
	};

      inDstStr = tmpPtr;
    }
  else
    {
      // src can be copied in whole to dst

      strncpy(inDstStr+inDstIndx,inSrcStr,srccpLen);
      if (( inDstIndx+inSrcMaxWidth ) >= dstLen)
	{
	  spaceCount = dstLen - (inDstIndx+srccpLen);
	}
      else
	{
	  spaceCount = inSrcMaxWidth - srccpLen;
	};
      
      // pad with spaces 

      memset(inDstStr+inDstIndx+srccpLen,' ',spaceCount);
    };

};


/*
** DESC: Creates line components of TextBlock from input string
**
**
** PRE: inStr pointer to source string that is null terminated
**      inWidth width of lines to be created
**      outLines list of pointers to char that are created lines
**
**
** POST: inStr is split into null terminated strings of length 
**       no more than inWidth and put into outLines list.
**
**
**
*/

void TextBlock::CreateLines(const char* inStr,int inWidth, AhelpList<char*>& outLines)
{
  
  // Do some initial cleanup
  
  DeleteLines(outLines);

  if ((inWidth <=0 ) || (inStr == 0) )
    return;



  int srcIndx,desIndx=0;
  int startIndx=0,endIndx=0;
  int len = strlen(inStr);
  char* tmpPtr;

  tmpPtr= new char[len+1];

  // copy str to tmp buffor and remove multiple white spaces

  for( srcIndx = 0 ;srcIndx<len;srcIndx++)
    {
      if ( inStr[srcIndx] == '\n')
	{

	  // skip multiple white spaces

	  if (!(( ( srcIndx == 0 ) || (srcIndx == (len-1) ))
	      || (( srcIndx > 0 )  && (isspace(inStr[srcIndx-1]) ))
	      || (( srcIndx <(len-1) ) &&(isspace(inStr[srcIndx+1]))) 
	      ))
	    {
	      tmpPtr[desIndx++] =' ';
	    }
	}
      else
	{
	  if ( isspace(inStr[srcIndx]))
	    {
	      tmpPtr[desIndx++]=' ';
	    }
	  else
	    {
	      tmpPtr[desIndx++] =inStr[srcIndx];
	    };
	};
    };

  tmpPtr[desIndx] = 0;

  char* srcData = tmpPtr;
  len = desIndx;


  startIndx=0;

  // split the tmp string and put it into output list
  // try to split on white spaces if possible

  while( startIndx < len )
    {

      if ( (startIndx+inWidth) < len )
	{
	  endIndx = startIndx+inWidth-1;

	  
	  while(( ! isspace(srcData[endIndx]) ) && (endIndx>startIndx))
	    {
	      --endIndx;
	    };
	  

	  if ( endIndx < startIndx+inWidth*0.5)
	    {
	      endIndx = startIndx+inWidth-1;
	    };

	}
      else 
	{
	  endIndx = len-1;
	};
      
      tmpPtr = new char[endIndx-startIndx+2];
      strncpy(tmpPtr,srcData+startIndx,endIndx-startIndx+1);
      tmpPtr[endIndx-startIndx+1]=0;
      outLines.Append(tmpPtr);
      startIndx=endIndx+1;
    };

  delete[] srcData;

};


/*
** DESC: Splits inStr on new lines and puts strings into outLines
**
**
** PRE: inStr pointer to char that contains null terminated string
**      outLines list of null terminated strings
**
**
** POST: inStr is split into strings on every new line results are
**       put into outLines
**
**
*/

int TextBlock::CreateLines(const char* inStr, AhelpList<char*>& outLines)
{

  DeleteLines(outLines);
  
  
  int tmpWidth =  0;


  if ( inStr == 0)
    {
      return 0 ;
    }
  
  int srcIndx;
  int len = strlen(inStr);
  char* tmpPtr;
  int startIndx,endIndx;


  startIndx=0;
  for(srcIndx=0;srcIndx<len;srcIndx++)
    {
      // split on new line
      if ( inStr[srcIndx] == '\n')
	{
	  if ( tmpWidth < ( srcIndx-startIndx ))
	    {
	      tmpWidth = srcIndx-startIndx;
	    };
	  tmpPtr = new char[srcIndx-startIndx+1];
	  strncpy(tmpPtr,inStr+startIndx,srcIndx-startIndx);
	  tmpPtr[srcIndx-startIndx]=0;
	  outLines.Append(tmpPtr);
	  startIndx=srcIndx+1;
	};
    };

  // if no new lines found, use the whole src string

  if ( startIndx <= len)
    {
      tmpPtr = new char[srcIndx-startIndx+1];
      strncpy(tmpPtr,inStr+startIndx,srcIndx-startIndx);
      tmpPtr[srcIndx-startIndx]=0;
      outLines.Append(tmpPtr);
      len = srcIndx-startIndx;
      if ( tmpWidth < len)
	{
	  tmpWidth = len;
	};
    };

  return tmpWidth;

};


/*
** DESC: Deletes strings from outLines list
**
**
** PRE: outLines is a reference to a list of strings
**
**
**
** POST: memory for each entry in outLines is deallocated
**
**
*/

void TextBlock::DeleteLines(AhelpList<char*>& outLines)
{

  AhelpListIterator<char*> lineIter(&(outLines));
  
  for(lineIter.First();!lineIter.IsDone();lineIter.Next())
    {
      delete[] lineIter.GetCurrent();
    };

  outLines.RemoveAll();

};


/*
** DESC: Converts TextBlock into a String
**
**
** PRE: outString is output String
**
**
**
** POST: Content of this TextBlock is converted into a String class
**       After conversion content of this TextBlock is dealocated
**
**
*/

void TextBlock::Str(AhelpStrings& outString)
{


  AhelpListIterator<char*> lineIter(&(lines));
  outString = "";

  while( lines.GetLength() )
    {
      lineIter.First();
      outString.append(lineIter.GetCurrent());
      delete[] lineIter.GetCurrent();
      lineIter.RemoveCurrent();

      if (lines.GetLength())
	{
	  outString.append('\n');
	};
    };
};



/*
** DESC: Sets this TextBlock to contain inStr 
**
**
** PRE: inWidth width of the block to be set
**      inStr string to be copied to this block
**
**
** POST: This TextBlock is set to contain inStr that is split 
**        into lines of width inWidth
**
**
*/

void TextBlock::Set(int inWidth,const char* const inStr)
{

  DeleteLines(lines);

  width = inWidth;
  if ( inWidth <  0 )
    width = 0;

  CreateLines(inStr,width,lines);

};



/*
** DESC: Sets this TextBlock to contain inStr
**
**
** PRE: inStr input string
**
**
**
** POST: This block contains copies of inStr split on new lines.
**        Width of this block is set to the max of contained lines
**
*/

void TextBlock::Set(const char* const inStr)
{

  DeleteLines(lines);


  width =  CreateLines(inStr,lines);

};


/*
** DESC: Copies inBlock into this TextBlock
**
**
** PRE: inBlock
**
**
**
** POST: This block contains copy of inBlock
**
**
*/

void TextBlock::Set( const TextBlock& inBlock)
{
  if ( this == &inBlock)
    return;

  DeleteLines(lines);
  width = 0;

  Append(inBlock);


};



/*
** DESC: Checks if this block is empty
**
**
** PRE: None
**
**
**
** POST: Returns 0 if contains at least one non-empty line
**               1 if empty
**
*/

int TextBlock::IsEmpty()
{
  if (width == 0 )
    return 1;

  if ( width > 0 )
    {
      AhelpListIterator<char*> lineIter(&(lines));
      
      for(lineIter.First();!lineIter.IsDone();lineIter.Next())
	{
	  if ( strlen(lineIter.GetCurrent()) > 0 )
	    return 0;
	};
    };

  return 1;

};


/*
** DESC: Appends inBlock to this block
**
**
** PRE: None
**
**
**
** POST: inBlock is appended to this block
**
**
*/

void TextBlock::Append(const TextBlock& inBlock)
{
  TextBlock* tmpBlock;
  char* tmpPtr;
 
  if ( this == &inBlock)
    {
      tmpBlock = new TextBlock(inBlock);
    }
  else
    {
      tmpBlock = &(TextBlock&)inBlock;
    };
  


  AhelpListIterator<char*> lineIter(&(tmpBlock->lines));

  for(lineIter.First();!lineIter.IsDone();lineIter.Next())
    {
      if ( lineIter.GetCurrent() != 0 )
	{
	  if ( width < strlen(lineIter.GetCurrent() ) )
	    {
	      width = strlen(lineIter.GetCurrent() );
	    };
	  tmpPtr = new char[strlen(lineIter.GetCurrent())+1];
	  strcpy(tmpPtr,lineIter.GetCurrent());
	  lines.Append(tmpPtr);
	};
    };

  
  if ( this == &inBlock)
    {
      delete tmpBlock;
    };
};


/*
** DESC: Copies inBlock to this block
**
**
** PRE: none
**
**
**
** POST: inBlock is copied to this block
**
**
*/

TextBlock& TextBlock::operator = (const TextBlock& inBlock)
{
  if ( this == &inBlock)
    return *this;

  char* tmpPtr;
  width = inBlock.width;

  DeleteLines(lines);
  
  if ( ((TextBlock&)inBlock).lines.GetLength()> 0  )
    {
      AhelpListIterator<char*> lineIter(&(((TextBlock&)inBlock).lines));
      
      for(lineIter.First();!lineIter.IsDone();lineIter.Next())
	{
	  tmpPtr = new char[strlen(lineIter.GetCurrent())+1];
	  strcpy(tmpPtr,lineIter.GetCurrent());
	  lines.Append(tmpPtr);
	};
    };

  return *this;
};



/*
** DESC: Copies inStr to this block
**
**
** PRE: inStr points to null terminated string
**
**
**
** POST: This block contains copy of inStr split on new lines.
**       Width of this block is set to max width of contained lines
**
*/

TextBlock& TextBlock::operator = (const char* inStr)
{
  
  Set(inStr);
  return *this;

};


/*
** DESC: Removes multiple white spaces for this block
**
**
** PRE: None
**
**
**
** POST: All multiple white spaces are compacted to single one
**
**
*/

void TextBlock::Compact()
{

  char* tmpPtr = new char[width*lines.GetLength()+1];
  int indx;
  char* lastPtr;  
  char* tmpTok;

  tmpPtr[0]=0;
  AhelpListIterator<char*> lineIter(&(lines));
  
  for(lineIter.First();!lineIter.IsDone();lineIter.Next())
    {
      if ( (tmpTok=strtok(lineIter.GetCurrent()," ")) != 0)
	{
	  if ( strlen(tmpPtr) !=0 )
	    {
	      strcat(tmpPtr," ");
	    };
	  strcat(tmpPtr,tmpTok);

	  while( (tmpTok=strtok(0," ")) !=0)
	    {
	      strcat(tmpPtr," ");
	      strcat(tmpPtr,tmpTok);
	    };
	};
      delete[] lineIter.GetCurrent();

    };

  lines.RemoveAll();
  CreateLines(tmpPtr,width,lines);

  delete[] tmpPtr;
  
};


/*
** DESC: Appends empty lines to the end of this TextBlock 
**
**
** PRE:  None
**
**
**
** POST: Empty lines are appended to the end of this TextBlock
**
**
*/

void TextBlock::Append(int inEmptyLineCount)
{
  for( int indx = 0;indx<inEmptyLineCount;indx++)
    {
      char* tmpPtr = new char[1];
      tmpPtr[0]=0;
      lines.Append(tmpPtr);
    };
};


/*
** DESC: Recalculates the width of the block
**
**
** PRE:  None
**
**
**
** POST: Width of the longest line in this block is assigned to the width
**        of this block
**
**
*/

void TextBlock::Resize()
{
  int tmpWidth=0;

  AhelpListIterator<char*> lineIter(&(lines));
  
  for(lineIter.First();!lineIter.IsDone();lineIter.Next())
    {
      if ( tmpWidth < strlen( lineIter.GetCurrent()))
	tmpWidth = strlen( lineIter.GetCurrent());
    };

  width = tmpWidth;

};


/*
** DESC: Appends inStr to the end of this block
**
**
** PRE: inStr is null terminated string
**      resizeFlag is int
**
**
** POST: inStr is split into strings on new line and created
**         strings are appended to this block as lines. If resizeFlag
**         is set width of this block is recalculated for the max 
**         length of appended line.  If resizeFlag is not set inStr
**         is split to strings of length less than width of this block
**
**
*/

void TextBlock::Append(const char* const inStr,int resizeFlag)
{
  if ( inStr == 0 ) 
    return;

  if ( inStr[0] == 0 )
    return;

  char* tmpPtr;
  AhelpList<char*> tmpLines;
  int tmpWidth=0;

  if ( lines.GetLength() > 0 )
    {
      tmpPtr = new char[strlen(lines.GetLast()) + strlen(inStr) +1];
      strcpy(tmpPtr,lines.GetLast());
      strcat(tmpPtr,inStr);
      lines.RemoveBottom();
    }
  else
    {
      tmpPtr = new char[strlen(inStr)+1];
      strcpy(tmpPtr,inStr);
    };

  if ( resizeFlag)
    {
      tmpWidth=CreateLines(tmpPtr,tmpLines);
    }
  else
    {
      CreateLines(tmpPtr,width,tmpLines);
    };

  if ( tmpWidth > width)
    width = tmpWidth;

  lines.Append(tmpLines);

  tmpLines.RemoveAll();
  delete[] tmpPtr;

};


/*
** DESC: Return width of this block's inCoordY line. First line has
**        coord 0
**
** PRE:  None
**
**
**
** POST: see DESC
**
**
*/

int  TextBlock::Width(int inCoordY)
{
  if ( inCoordY >= lines.GetLength() )
    return 0;

  if ( lines.GetIndex(inCoordY) != 0 )
    {
      return strlen(lines.GetIndex(inCoordY));
    };

  return 0;
};


/*
** DESC: Prepends empty lines before this block
**
**
** PRE:  inEmptyLineCount number of empty lines to be prepended
**
**
**
** POST: inEmptyLineCount black lines are inserted at the beginnig of
**       this block.
**
**
*/

void TextBlock::Prepend(int inEmptyLineCount)
{
  for( int indx = 0;indx<inEmptyLineCount;indx++)
    {
      char* tmpPtr = new char[1];
      tmpPtr[0]=0;
      lines.Prepend(tmpPtr);
    };

};


/*
** DESC:  Output operator
**
**
** PRE:   ostr output stream
**        inBlock block to be output
**
**
** POST:  text content of inBlock is output to stream ostr
**
**
*/

ostream& operator<<(ostream& ostr,TextBlock& inBlock)
{
  AhelpListIterator<char*> lineIter(&(inBlock.lines));
  
  for(lineIter.First();!lineIter.IsDone();lineIter.Next())
    {
      ostr<< lineIter.GetCurrent()<<endl;
    };

  return ostr;
};

