// 
//  Copyright (C) 2002,2003,2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include "TextBlockTable.hh"



/*
** DESC:  Constructor for TableData
**
**
** PRE: inColCount number of columns in the table
**        
**
**
** POST:  Creates data object  that is used by TextBlockTable
**
**
*/

TableData::TableData(int inColCount)
{
  colCount=inColCount;
  for(int colIndx=0;colIndx<colCount;colIndx++)
    {
      colWidths.Append(0);
      headerColWidths.Append(0);
    }
};



/*
** DESC: Sets each columns header 
**
**
** PRE:  inHeader is a list of headers starting with most left column
**        
**
**
** POST:  Internal structures are initialized
**
**
*/

void TableData::SetHeader(AhelpList<AhelpStrings>& inHeader)
{
  
  if ( ! inHeader.GetLength() )
    {
      return;
    };

  for(int colIndx=0;colIndx<colCount;colIndx++)
    {
      if ( colIndx < inHeader.GetLength())
	{
	  header.Append(inHeader.GetIndex(colIndx));
	}
      else
	{
	  header.Append("");
	};
    }

  SetColumnWidths(inHeader,headerColWidths);
  
};




/*
** DESC:  Adds a row to internal structures
**
**
** PRE:   inRow is a list of values in each column
**        
**
**
** POST:  Internal data structures are updated
**
**
*/

void TableData::AddRow(AhelpList<AhelpStrings>& inRow)
{
  if ( !inRow.GetLength())
    {
      return;
    };

  AhelpList<AhelpStrings> tmpRow;
  
  for(int colIndx=0;colIndx<colCount;colIndx++)
    {
      if ( colIndx < inRow.GetLength())
	{
	  tmpRow.Append(inRow.GetIndex(colIndx));
	}
      else
	{
	  tmpRow.Append("");
	};
    }
  rows.Append(tmpRow);
  SetColumnWidths(inRow,colWidths);

  

};




/*
** DESC:  Sets column width 
**
**
** PRE:   inRow is a list with values for each column
**        outWidths is a list of width for each column
**
**
** POST:  inRow is compared with data already in the structure
**        and return list of column widths
**
**
*/

void TableData::SetColumnWidths(AhelpList<AhelpStrings>& inRow,AhelpList<int>& outWidths)
{
  int colIndx;
  
  int lastIndx = inRow.GetLength();
  
  if ( lastIndx > outWidths.GetLength())
    {
      lastIndx = outWidths.GetLength();
    };

  for(colIndx=0;colIndx<lastIndx;colIndx++)
    {
      if ( outWidths.GetIndex(colIndx) < inRow.GetIndex(colIndx).length())
	{
	  outWidths.GetIndex(colIndx) = inRow.GetIndex(colIndx).length();
	}
    };
};







/*
** DESC:  Allocates object
**
**
** PRE:   inTableData - object with data to be inserted into table
**        inMaxWidth - max width of the whole table (incl. frame)
**        inCompactFlag - set if empty columns should not be included 
**        inSqueezeColNo - which column should be squeezed first 
**                         if table is wider than inMaxWidth
**
**
** POST:  Table is created from inTableData.  If inCompactFlag is set
**        empty columns are excluded.  Table is squeezed accordingly
**        if inSqueezeColNo is set. 
**
**
*/

TextBlockTable::TextBlockTable(TableData& inTableData,int inMaxWidth,int inCompactFlag,int inSqueezeColNo)
{
  TextBlock tmpCellBlock;
  int colXCoord=0;
  int colYCoord=0;
  AhelpStrings rowSeparator;
  AhelpStrings colSeparator;
  int colIndx=0;
  int rowIndx;
  int tableWidth=0;
  int tableHeight=0;
  int maxRowHeight=0;
  AhelpList<int> colWidths;
  AhelpList<int> rowHeights;
  int currColWidth=0;
  int tmpTableWidth;
  int actualColCount=0;
  int tmpWidth;
  AhelpStrings tmpString;
  int changeFlag=0;


  if ( inTableData.rows.GetLength() == 0 )
    return;

  // Determine the actual widths of columns 

  for(colIndx=0;colIndx<inTableData.colCount;colIndx++)
    {
      if ((inCompactFlag) && (inTableData.colWidths.GetIndex(colIndx) == 0 ))
	{
	  currColWidth=0;
	}
      else 
	{
	  currColWidth=inTableData.headerColWidths.GetIndex(colIndx);
	}
      
      colWidths.Append(currColWidth);
      if (! (( inCompactFlag) && (currColWidth == 0 )))
	{
	  tableWidth+=currColWidth+3;
	  actualColCount++;
	};

    };

  tableWidth++;


  // Strech the table if possible to inMaxWidth

  while(tableWidth < inMaxWidth)
    {
      changeFlag=0;
      for(colIndx=0;colIndx<inTableData.colCount;colIndx++)
	{
	  if ((!((inCompactFlag) && (inTableData.colWidths.GetIndex(colIndx) == 0 )) )
	      && (inTableData.colWidths.GetIndex(colIndx) >colWidths.GetIndex(colIndx ) ))
	    {
	      ++( colWidths.GetIndex(colIndx ));
	      tableWidth++;
	      changeFlag=1;
	    };
	  if ( tableWidth >=inMaxWidth)
	    break;
	};

      if ( !changeFlag )
	break;
    };


  // If necessary squeeze given columns

  if (( tableWidth > inMaxWidth) && ( inSqueezeColNo >=0 ) && (inSqueezeColNo <colWidths.GetLength()))
    {
      if ( colWidths.GetIndex(inSqueezeColNo ) )
	{
	  currColWidth = colWidths.GetIndex(inSqueezeColNo ) - tableWidth+inMaxWidth;
	  if ( currColWidth <= 0 ) 
	    {
	      tableWidth = tableWidth - colWidths.GetIndex(inSqueezeColNo) +1;
	      colWidths.GetIndex(inSqueezeColNo)=1;
	    }
	  else
	    {
	      tableWidth = tableWidth - colWidths.GetIndex(inSqueezeColNo)+currColWidth;
	      colWidths.GetIndex(inSqueezeColNo) = currColWidth;
	    };
	};
    }; 
  


  // squeeze other columns if neccessary

  if ( tableWidth > inMaxWidth)
    {
      tmpTableWidth=0;

      for(colIndx=0;colIndx<colWidths.GetLength();colIndx++)
	{
	  if ( colWidths.GetIndex(colIndx ) )
	    {
	      currColWidth = (colWidths.GetIndex(colIndx)*(inMaxWidth-actualColCount*3-1))/(tableWidth-actualColCount*3-1);
	      if ( currColWidth == 0 ) 
		{
		  colWidths.GetIndex(colIndx)=1;
		}
	      else
		{
		  colWidths.GetIndex(colIndx) = currColWidth;
		};
	      tmpTableWidth +=colWidths.GetIndex(colIndx)+3;
	    };
	};

      tableWidth=tmpTableWidth+1;
    };
  
  Resize(tableWidth);  

  // Check if there is any non-empty cols

  actualColCount=0;
  for( colIndx=0;colIndx<inTableData.colCount;colIndx++)
    {
      if (( inCompactFlag) &&  ( colWidths.GetIndex(colIndx) == 0 ))
	{
	  continue;
	}
      else
	{
	  actualColCount++;
	};
    }

  if ( ! actualColCount)
    {
      return;
    };

  
  // Create header 

  colYCoord=1;
  if ( inTableData.header.GetLength())
    {
      colXCoord=2;
      for( colIndx=0;colIndx<inTableData.colCount;colIndx++)
	{
	  if (( inCompactFlag) &&  ( colWidths.GetIndex(colIndx) == 0 ))
	    {
	      continue;
	    };

	  tmpWidth= colWidths.GetIndex(colIndx) - inTableData.header.GetIndex(colIndx).length();

	  tmpString = inTableData.header.GetIndex(colIndx);

	  if ( tmpWidth > 1 )
	    {
	      tmpString.padR(tmpWidth/2);
	      tmpString.padL(tmpWidth/2);
	    };

	  tmpCellBlock.Set(colWidths.GetIndex(colIndx),tmpString.chars());
	  if ( tmpCellBlock.Height() > maxRowHeight)
	    {
	      maxRowHeight = tmpCellBlock.Height();
	    };

	  Paste(colXCoord,colYCoord,tmpCellBlock);
	  colXCoord +=colWidths.GetIndex(colIndx) + 3;
	};
  
      colYCoord += maxRowHeight+1;
      tableHeight += maxRowHeight+1;

      rowHeights.Append(maxRowHeight);
    };


  // Create rows

  if ( inTableData.rows.GetLength())
    {
      for(rowIndx=0;rowIndx<inTableData.rows.GetLength();rowIndx++)
	{
	  maxRowHeight=0;
	  colXCoord=2;
	  for( colIndx=0;colIndx<inTableData.colCount;colIndx++)
	    {
	      if (( inCompactFlag) &&  ( colWidths.GetIndex(colIndx) == 0 ))
		{
		  continue;
		};

	      tmpCellBlock.Set(colWidths.GetIndex(colIndx),inTableData.rows.GetIndex(rowIndx).GetIndex(colIndx).chars());

	      if ( tmpCellBlock.Height() > maxRowHeight)
		{
		  maxRowHeight = tmpCellBlock.Height();
		};
	      
	      Paste(colXCoord,colYCoord,tmpCellBlock);
	      colXCoord +=colWidths.GetIndex(colIndx) + 3;
	    };
	  
	  colYCoord +=maxRowHeight+1;
	  rowHeights.Append(maxRowHeight);
	  tableHeight += maxRowHeight+1;
	};
    };


  // Create the table frame

  tableHeight++;
  colYCoord=0;
  rowSeparator.fill(tableWidth,'-');

  for(rowIndx=0;rowIndx<rowHeights.GetLength();rowIndx++)
    {
      Paste(0,colYCoord,rowSeparator.chars());
      colYCoord += rowHeights.GetIndex(rowIndx)+1;
    };
  Paste(0,colYCoord,rowSeparator.chars());

  colXCoord=0;
  for( colIndx=0;colIndx<colWidths.GetLength();colIndx++)
    {
      if (( inCompactFlag) &&  ( colWidths.GetIndex(colIndx) == 0 ))
	{
	  continue;
	};
      for(rowIndx=1;rowIndx<(tableHeight-1);rowIndx++)
	{
	  Paste(colXCoord,rowIndx,'|');
	};
      colXCoord += colWidths.GetIndex(colIndx)+3;
    };


  for(rowIndx=1;rowIndx<(tableHeight-1);rowIndx++)
    {
      Paste(colXCoord,rowIndx,'|');
    };
  
};

