// 
//  Copyright (C) 2006-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

//---------------------------------------------------------------------------
//
//  FILE NAME:          LINEAR2DTransform.cc
//
//  DEVELOPMENT:        CXCDS
//
//  DESCRIPTION:
//
//---------------------------------------------------------------------------

#include "Transform.hh"


// **************************************************************
//  Method:          LINEAR2DTransform
//
//  Description:     Empty constructor.
//                   
//  Arguments:       none
//
//  Returns:         none
//
// **************************************************************
LINEAR2DTransform::LINEAR2DTransform() :
  matrix(NULL)
{
  clear();

  // Initialize Class Name string
  _strname = "LINEAR2DTransform";

  // Generate a new parameter list for this Transform object.
  generate_parameter_list();
}


// **************************************************************
//  Method:          LINEAR2DTransform
//
//  Description:     LINEAR2DTransform Constructor.
//                   
//  Arguments:       
//     string - the name of the transform
//
//  Returns:         none
//
// **************************************************************
LINEAR2DTransform::LINEAR2DTransform(string in_name)  :
  matrix(NULL)
{
  clear();

  // Initialize Class Name string
  _strname = "LINEAR2DTransform";

  // Initialize Transform Name
  name = in_name;

  // Generate a new parameter list for this Transform object.
  generate_parameter_list();
}


// **************************************************************
//  Method:          LINEAR2DTransform
//
//  Description:     LINEAR2DTransform Constructor.
//                   
//  Arguments:       
//     string - the name of the transform
//     list<TransParam*> - a list of Transform Parameters
//
//  Returns:         none
//
// **************************************************************
LINEAR2DTransform::LINEAR2DTransform(string in_name, list<TransParam*> in_params)  :
  matrix(NULL)
{
  clear();

  // Initialize Class Name string
  _strname = "LINEAR2DTransform";

  // Initialize Transform Name
  name = in_name;

  // Generate a new parameter list for this Transform object.
  generate_parameter_list();

  // Initialize Parameter values from input list..
  set_parameter_list( in_params );

}


// **************************************************************
//  Method:          LINEAR2DTransform
//
//  Description:     LINEAR2DTransform copy constructor
//
//  Arguments:
//     string - the name of the transform
//     LINEAR2DTransform& - the LINEAR2DTransform object to copy
//
//  Returns:         none
//
// **************************************************************
LINEAR2DTransform::LINEAR2DTransform(LINEAR2DTransform &init)  :
  matrix(NULL)
{

  if (this == &init) return;
  clear();
 
  // Initialize Class Name string
  _strname = "LINEAR2DTransform";

  name = init.get_name();

  list<TransParam*>in_params = init.get_parameter_list();

  if (!in_params.empty()) {

    list<TransParam*>::iterator iter;
    for (iter = in_params.begin(); iter != in_params.end(); iter++) {
      TransParam* par_copy = new TransParam( *(*iter) );
      params.push_back(par_copy);
    }

  } else {
    // Generate a new parameter list for this Transform object.
    generate_parameter_list();
  }

  calculate_transform_matrix();

}


// **************************************************************
//  Method:          ~LINEAR2DTransform
//
//  Description:     LINEAR2DTransform Desctructor.  
//
//  Arguments:       none
//
//  Returns:         none
//
// **************************************************************
LINEAR2DTransform::~LINEAR2DTransform() {
  clear();

  if (matrix) {
    for (int rr = 3; rr--; ) delete [] (matrix[rr]);  
    delete [] matrix;
  }
  matrix = NULL;
}


// **************************************************************
//  Method:          copy
//
//  Description:     Makes a copy of this LINEAR2DTransform and returns it.
//
//  Arguments:       none
//
//  Returns:         
//     LINEAR2DTransform* - a copy of this Transform
//
// **************************************************************
LINEAR2DTransform* LINEAR2DTransform::copy() 
{
  LINEAR2DTransform* trans = new LINEAR2DTransform( (LINEAR2DTransform&)*this );
  return trans;
}


// **************************************************************
//  Method:          get_transform_matrix
//
//  Description:     Returns the 3x3 transform matrix.
//
//  Arguments:       
//     long - 
//
//  Returns:
//     double** - 3x3 tranform matrix
//
//
// **************************************************************
double** LINEAR2DTransform::get_transform_matrix() {

  calculate_transform_matrix();
  return matrix;
}


// **************************************************************
//  Method:          calculate_transform_matrix
//
//  Description:     Calculates the 3x3 transform matrix.
//
//  Arguments:       none
//
//  Returns:         none
//
// **************************************************************
void LINEAR2DTransform::calculate_transform_matrix() {

  double* scale =    (double*)get_parameter_value("SCALE");
  double* rotation = (double*)get_parameter_value("ROTATION");
  double* offset =   (double*)get_parameter_value("OFFSET");

  if ( scale && rotation && offset ) {

    if ( matrix == NULL )
      if ((matrix = new double* [3]) != NULL) 
	for (int rr = 3; rr--; )  matrix[rr] = new double[3];  
  
    if ( matrix ) {

      matrix[0][0] = scale[0] * cos(*rotation);
      matrix[0][1] = scale[0] * sin(*rotation);
      matrix[0][2] = 0; 
      
      matrix[1][0] = -scale[1] * sin(*rotation);
      matrix[1][1] = scale[1] * cos(*rotation);
      matrix[1][2] = 0;
      
      matrix[2][0] = offset[0];
      matrix[2][1] = offset[1];
      matrix[2][2] = 1;
    }
  }
}


// **************************************************************
//  Method:          set_transform_matrix
//
//  Description:     
//     Determines and assigns values to Transform parameters
//     from the input 3x3 Transform matrix.
//
//     Matrix is defined according to FITS standards w.r.t. 
//     application and orientation.   X' = X*[M] 
//     
//       M  = | CD0cos(rho)   -CD1sin(rho)   0.0  |
//            | CD0sin(rho)    CD1cos(rho)   0.0  |
//            | Tx             Ty            1.0  |
//
//  Arguments:
//     double** - the transform matrix
//
//  Returns:
//
// **************************************************************
void LINEAR2DTransform::set_transform_matrix(double** in_matrix) {

  if (in_matrix == NULL) return;

  // allocate new memory for transform matrix
  if ( matrix == NULL ) {
    if ((matrix = new double* [3]) != NULL) 
      for (int rr = 3; rr--; )  matrix[rr] = new double[3];  
  }

  // copy in_matrix
  for (int ii = 0; ii < 3; ii++)
    for (int jj = 0; jj < 3; jj++)
      matrix[ii][jj] = in_matrix[ii][jj];

  double rho_a, rho_b;
  double tmpcrota, coscrota, sincrota;

  double rotation = 0;
  double offset[2];
  double scale[2];
  
  offset[0] = matrix[2][0];
  offset[1] = matrix[2][1];

  if (matrix[0][1] == 0)
    rho_a = 0.0;
  else if (matrix[0][1] > 0.0)
    rho_a = atan(matrix[0][1] / matrix[0][0]);
  else
    rho_a = atan(-matrix[0][1] / -matrix[0][0]);
  
  if (matrix[1][0] == 0)
    rho_b = 0.0;
  else if (matrix[1][0] > 0)
    rho_b = atan(matrix[1][0] / -matrix[1][1]);
  else
    rho_b = atan(-matrix[1][0] / matrix[1][1]);
  
  tmpcrota = (rho_a + rho_b) / 2.0;
  coscrota = cos(tmpcrota);
  sincrota = sin(tmpcrota);
  
  scale[0] = 1.0;
  scale[1] = 1.0;
  
  scale[0] = sqrt(matrix[0][0]*matrix[0][0] + matrix[0][1]*matrix[0][1]);
  scale[1] = sqrt(matrix[1][0]*matrix[1][0] + matrix[1][1]*matrix[1][1]);
  
  if (coscrota > 0.0) {
    if (matrix[0][0]*coscrota < 0.0)
      scale[0] = scale[0] * -1.0;
    if (matrix[1][1]*coscrota < 0.0)
      scale[1] = scale[1] * -1.0;
  } else {
    if (matrix[1][0]*sincrota < 0.0)
      scale[0] = scale[0] * -1.0;
    if (-matrix[0][1] / sincrota < 0.0)
      scale[1] = scale[1] * -1.0;
  }

  rotation = (tmpcrota*180.0)/M_PI;

  get_parameter("OFFSET")->set_value(&offset, dmDOUBLE, 2);
  get_parameter("SCALE")->set_value(&scale, dmDOUBLE, 2);
  get_parameter("ROTATION")->set_value(&rotation, dmDOUBLE, 1);

}


// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     void*            - interleaved data array
//     dmDataType       - array datatype
//     long             - number of value pairs.
//
//  Returns:
//     void*            - the transformed data array
//
// **************************************************************
void* LINEAR2DTransform::apply(void* in_array, dmDataType in_datatype, long nsets, long arrsize) 
{
  void*   optr = NULL;            // Pointer to output data

  switch (in_datatype) 
  {
    case dmDOUBLE:
      optr = (void*) apply((double*)in_array, nsets, arrsize );
      break;
    case dmSHORT: 
      optr = (void*) apply((short*)in_array, nsets, arrsize );
      break;
    case dmLONG:
      optr = (void*) apply((long*)in_array, nsets, arrsize );
      break;
    case dmFLOAT:
      optr = (void*) apply((float*)in_array, nsets, arrsize );
      break;
    case dmUSHORT:
      optr = (void*) apply((unsigned short*)in_array, nsets, arrsize );
      break;
    case dmULONG:
      optr = (void*) apply((unsigned long*)in_array, nsets, arrsize );
      break;
    case dmBOOL:
    case dmBIT:
    case dmBYTE:
    case dmTEXT:
    case dmUNKNOWNTYPE:
    default:
      optr = NULL;
      break;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     short*           - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     short*           - the transformed data array
//
// **************************************************************
short* LINEAR2DTransform::apply(short* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  short*  optr    = NULL;         // Pointer to output data
  dmDataType intype = dmSHORT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new short[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     long*            - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     long*            - the transformed data array
//
// **************************************************************
long* LINEAR2DTransform::apply(long* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  long*   optr    = NULL;         // Pointer to output data
  dmDataType intype = dmLONG;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new long[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     float*           - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     float*           - the transformed data array
//
// **************************************************************
float* LINEAR2DTransform::apply(float* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  float*   optr    = NULL;        // Pointer to output data
  dmDataType intype = dmFLOAT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new float[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     unsigned short*  - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     unsigned short*  - the transformed data array
//
// **************************************************************
unsigned short* LINEAR2DTransform::apply(unsigned short* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  unsigned short* optr = NULL;    // Pointer to output data
  dmDataType intype = dmUSHORT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new unsigned short[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     unsigned long*   - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     unsigned long*   - the transformed data array
//
// **************************************************************
unsigned long* LINEAR2DTransform::apply(unsigned long* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  unsigned long* optr = NULL;     // Pointer to output data
  dmDataType intype = dmULONG;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new unsigned long[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     double*          - input data array
//     long             - number of value pairs.
//
//  Returns:
//     double*          - the transformed data array
//
// **************************************************************
double* LINEAR2DTransform::apply(double* inarr, long nsets, long arrsize) {

  double* outarr = NULL;          // output data array.
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  outarr = new double[nvals];
  if ( !outarr ){ return NULL; }

  if (matrix == NULL) {
    calculate_transform_matrix();
    if (matrix == NULL) 
      return NULL;
  }

  long ndx = 0;
  long ndy = 0;

  for (int ii = 0; ii < nsets; ii++)
  {
    for (int jj = 0; jj < arrsize; jj++)
    {
      ndx = jj + (ii*2*arrsize);
      ndy = jj + (ii*2*arrsize) + arrsize;
	  
      outarr[ndx] = (matrix[0][0] * inarr[ndx]) + (matrix[1][0] * inarr[ndy]) + matrix[2][0];
      outarr[ndy] = (matrix[0][1] * inarr[ndx]) + (matrix[1][1] * inarr[ndy]) + matrix[2][1];

    }    
  }

  return outarr;
}

// **************************************************************
//  Method:          invert
//
//  Description:     Apply the inverse transform matrix to the 
//                   input data array.
//
//  Arguments:
//     void*            - interleaved data array
//     dmDataType       - array datatype
//     long             - number of value pairs.
//
//  Returns:
//     void*            - the transformed data array
//
// **************************************************************
void* LINEAR2DTransform::invert(void* in_array, dmDataType in_datatype, long nsets, long arrsize) 
{
  void*   optr = NULL;            // Pointer to output data

  switch (in_datatype) 
  {
    case dmDOUBLE:
      optr = (void*) invert((double*)in_array, nsets, arrsize );
      break;
    case dmSHORT: 
      optr = (void*) invert((short*)in_array, nsets, arrsize );
      break;
    case dmLONG:
      optr = (void*) invert((long*)in_array, nsets, arrsize );
      break;
    case dmFLOAT:
      optr = (void*) invert((float*)in_array, nsets, arrsize );
      break;
    case dmUSHORT:
      optr = (void*) invert((unsigned short*)in_array, nsets, arrsize );
      break;
    case dmULONG:
      optr = (void*) invert((unsigned long*)in_array, nsets, arrsize );
      break;
    case dmBOOL:
    case dmBIT:
    case dmBYTE:
    case dmTEXT:
    case dmUNKNOWNTYPE:
    default:
      optr = NULL;
      break;
  }

  return optr;
}

// **************************************************************
//  Method:          invert
//
//  Description:     Apply the inverse transform matrix to the 
//                   input data array.
//
//  Arguments:
//     short*           - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     short*           - the transformed data array
//
// **************************************************************
short* LINEAR2DTransform::invert(short* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  short*  optr    = NULL;         // Pointer to output data
  dmDataType intype = dmSHORT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Invert transform to double array..
  outdata = invert( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new short[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          invert
//
//  Description:     Apply the inverse transform matrix to the 
//                   input data array.
//
//  Arguments:
//     long*            - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     long*            - the transformed data array
//
// **************************************************************
long* LINEAR2DTransform::invert(long* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  long*   optr    = NULL;         // Pointer to output data
  dmDataType intype = dmLONG;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Invert transform to double array..
  outdata = invert( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new long[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          invert
//
//  Description:     Apply the inverse transform matrix to the 
//                   input data array.
//
//  Arguments:
//     float*           - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     float*           - the transformed data array
//
// **************************************************************
float* LINEAR2DTransform::invert(float* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  float*   optr    = NULL;        // Pointer to output data
  dmDataType intype = dmFLOAT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Invert transform to double array..
  outdata = invert( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new float[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          invert
//
//  Description:     Apply the inverse transform matrix to the 
//                   input data array.
//
//  Arguments:
//     unsigned short*  - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     unsigned short*  - the transformed data array
//
// **************************************************************
unsigned short* LINEAR2DTransform::invert(unsigned short* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  unsigned short* optr = NULL;    // Pointer to output data
  dmDataType intype = dmUSHORT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Invert transform to double array..
  outdata = invert( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new unsigned short[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          invert
//
//  Description:     Apply the inverse transform matrix to the 
//                   input data array.
//
//  Arguments:
//     unsigned long*   - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     unsigned long*   - the transformed data array
//
// **************************************************************
unsigned long* LINEAR2DTransform::invert(unsigned long* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  unsigned long* optr = NULL;     // Pointer to output data
  dmDataType intype = dmULONG;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Invert transform to double array..
  outdata = invert( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new unsigned long[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          invert
//
//  Description:     Apply the inverse transform matrix to the 
//                   input data array.
//
//  Arguments:
//     double*          - input data array
//     long             - number of value pairs.
//
//  Returns:
//     double*          - the transformed data array
//
// **************************************************************
double* LINEAR2DTransform::invert(double* inarr, long nsets, long arrsize) {

  double* outarr = NULL;          // output data array.
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  outarr = new double[nvals];
  if ( !outarr ){ return NULL; }

  double** inv_matrix = get_inverse_matrix();
  if (inv_matrix ==  NULL) {
    delete []outarr;
    return NULL;
  }

  long ndx = 0;
  long ndy = 0;

  for (int ii = 0; ii < nsets; ii++)
  {
    for (int jj = 0; jj < arrsize; jj++)
    {
      ndx = jj + (ii*2*arrsize);
      ndy = jj + (ii*2*arrsize) + arrsize;
	  
      outarr[ndx] = (inv_matrix[0][0] * inarr[ndx]) + (inv_matrix[1][0] * inarr[ndy]) + inv_matrix[2][0];
      outarr[ndy] = (inv_matrix[0][1] * inarr[ndx]) + (inv_matrix[1][1] * inarr[ndy]) + inv_matrix[2][1];
    }    
  }

  for (int rr = 3; rr--; ) delete []inv_matrix[rr];
  delete []inv_matrix;

  return outarr;
}


// **************************************************************
//  Method:          get_inverse_matrix
//
//  Description:     Calculates and returns the inverse matrix of 
//                   the 3x3 transform matrix.
//
//     If the transform matrix is defined as:
//     M  = | a   b   0.0  |
//          | c   d   0.0  |
//          | Tx  Ty  1.0  |
//
//
//     Then the inverse matrix is:
//     M^-1 =      1          |  d             -b            0.0 |
//            -----------  *  | -c              a            0.0 |
//            (a*d - b*c)     | (c*Ty-d*tx)   -(a*Ty-b*Tx)   1.0 |
//              
//
//  Arguments:       none
//
//  Returns:
//     double** - the 3x3 inverse transform matrix
//
// **************************************************************
double** LINEAR2DTransform::get_inverse_matrix() {

  double **inv_matrix = NULL;
  
  calculate_transform_matrix();

  if (matrix == NULL) return NULL;

  if ((inv_matrix = new double* [3]) != NULL) 
    for (int rr = 3; rr--; )  inv_matrix[rr] = new double[3];  

  double determinant = (matrix[0][0] * matrix[1][1]) - (matrix[0][1] * matrix[1][0]);

  // if determinant is zero, then matrix is not invertable
  if (determinant == 0)
  {
     if (inv_matrix) for (int rr = 3; rr--; ) delete []inv_matrix[rr];
     delete []inv_matrix; 
     return NULL;
  }

  inv_matrix[0][0] = matrix[1][1]/determinant;
  inv_matrix[0][1] = -matrix[0][1]/determinant;
  inv_matrix[0][2] = 0;

  inv_matrix[1][0] = -matrix[1][0]/determinant;
  inv_matrix[1][1] = matrix[0][0]/determinant;
  inv_matrix[1][2] = 0;

  inv_matrix[2][0] = ((matrix[1][0] * matrix[2][1]) - (matrix[1][1] * matrix[2][0])) / determinant;
  inv_matrix[2][1] = (-(matrix[0][0] * matrix[2][1]) - (matrix[0][1] * matrix[2][0])) / determinant;
  inv_matrix[2][2] = 1;

  return inv_matrix;

}


// **************************************************************
//  Method:          get_parameter_list
//
//  Description:     Returns the list of transform parameters.
//                   If list is empty, will generate a generic
//                   list of parameters to return.
//
//  Arguments:       none
//
//  Returns:
//     list<TransParam*> - list of transform parameters
//
// **************************************************************
list<TransParam*> LINEAR2DTransform::get_parameter_list() {
  if ( params.empty() )
    generate_parameter_list();

  return params;
}


// **************************************************************
//  Method:          generate_parameter_list
//
//  Description:     Creates a list of parameters needed
//                   for the LINEAR2DTransform.
//                   User must fill in value.
//
//  Arguments:       none
//
//  Returns:
//     Loads Transform params attribute.
//                   
// **************************************************************
void LINEAR2DTransform::generate_parameter_list() {
  TransParam*  ptr = NULL;

  ptr = new TransParam("SCALE", NULL, dmDOUBLE, 
		       "Scale", name, 2);
  params.push_back( ptr );

  ptr = new TransParam("ROTATION", NULL, dmDOUBLE, 
		       "Rotation Angle", name, 1);
  params.push_back( ptr );

  ptr = new TransParam("OFFSET", NULL, dmDOUBLE, 
		       "Origin Offset", name, 2);
  params.push_back( ptr );


}

