// 
//  Copyright (C) 2006-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

//---------------------------------------------------------------------------
//
//  FILE NAME:          LINEARTransform.cc
//
//  DEVELOPMENT:        CXCDS
//
//  DESCRIPTION:
//
//---------------------------------------------------------------------------

#include "Transform.hh"



// **************************************************************
//  Method:          LINEARTransform
//
//  Description:     Empty constructor.
//                   
//  Arguments:       none
//
//  Returns:         none
//
// **************************************************************
LINEARTransform::LINEARTransform()
{
  clear();

  // Initialize Class Name string
  _strname = "LINEARTransform";

  // Generate a new parameter list for this Transform object.
  generate_parameter_list();
}


// **************************************************************
//  Method:          LINEARTransform
//
//  Description:     LINEARTransform Constructor.
//                   
//  Arguments:       
//     string - the name of the transform
//
//  Returns:         none
//
// **************************************************************
LINEARTransform::LINEARTransform(string in_name)
{
  clear();

  // Initialize Class Name string
  _strname = "LINEARTransform";

  // Initialize Transform Name
  name = in_name;

  // Generate a new parameter list for this Transform object.
  generate_parameter_list();
}


// **************************************************************
//  Method:          LINEARTransform
//
//  Description:     LINEARTransform Constructor.
//                   
//  Arguments:       
//     string - the name of the transform
//     list<TransParam*> - a list of Transform Parameters
//
//  Returns:         none
//
// **************************************************************
LINEARTransform::LINEARTransform(string in_name, list<TransParam*> in_params)
{
  clear();

  // Initialize Class Name string
  _strname = "LINEARTransform";

  // Initialize Transform Name
  name = in_name;

  // Generate a new parameter list for this Transform object.
  generate_parameter_list();

  // Initialize Parameter values from input list..
  set_parameter_list( in_params );

}


// **************************************************************
//  Method:          LINEARTransform
//
//  Description:     LINEARTransform copy constructor
//
//  Arguments:
//     string - the name of the transform
//     LINEARTransform& - the LINEARTransform object to copy
//
//  Returns:         none
//
// **************************************************************
LINEARTransform::LINEARTransform(LINEARTransform &init)
{

  if (this == &init) return;
  clear();
 
  // Initialize Class Name string
  _strname = "LINEARTransform";

  name = init.get_name();

  list<TransParam*>in_params = init.get_parameter_list();

  if (!in_params.empty()) {

    list<TransParam*>::iterator iter;
    for (iter = in_params.begin(); iter != in_params.end(); iter++) {
      TransParam* par_copy = new TransParam( *(*iter) );
      params.push_back(par_copy);
    }

  } else {
    // Generate a new parameter list for this Transform object.
    generate_parameter_list();
  }

}


// **************************************************************
//  Method:          ~LINEARTransform
//
//  Description:     LINEARTransform Desctructor.  
//
//  Arguments:       none
//
//  Returns:         none
//
// **************************************************************
LINEARTransform::~LINEARTransform() {
  clear();
}


// **************************************************************
//  Method:          copy
//
//  Description:     Makes a copy of this LINEARTransform and returns it.
//
//  Arguments:       none
//
//  Returns:         
//     LINEARTransform* - a copy of this Transform
//
// **************************************************************
LINEARTransform* LINEARTransform::copy() 
{
  LINEARTransform* trans = new LINEARTransform( (LINEARTransform&)*this );
  return trans;
}


// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     void*            - interleaved data array
//     dmDataType       - array datatype
//     long             - number of value pairs.
//
//  Returns:
//     void*            - the transformed data array
//
// **************************************************************
void* LINEARTransform::apply(void* in_array, dmDataType in_datatype, long nsets, long arrsize) 
{
  void*   optr = NULL;            // Pointer to output data

  switch (in_datatype) 
  {
    case dmDOUBLE:
      optr = (void*) apply((double*)in_array, nsets, arrsize );
      break;
    case dmSHORT: 
      optr = (void*) apply((short*)in_array, nsets, arrsize );
      break;
    case dmLONG:
      optr = (void*) apply((long*)in_array, nsets, arrsize );
      break;
    case dmFLOAT:
      optr = (void*) apply((float*)in_array, nsets, arrsize );
      break;
    case dmUSHORT:
      optr = (void*) apply((unsigned short*)in_array, nsets, arrsize );
      break;
    case dmULONG:
      optr = (void*) apply((unsigned long*)in_array, nsets, arrsize );
      break;
    case dmBOOL:
    case dmBIT:
    case dmBYTE:
    case dmTEXT:
    case dmUNKNOWNTYPE:
    default:
      optr = NULL;
      break;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     short*           - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     short*           - the transformed data array
//
// **************************************************************
short* LINEARTransform::apply(short* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * arrsize;    // Total #vals at pointer.

  short*  optr    = NULL;         // Pointer to output data
  dmDataType intype = dmSHORT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new short[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     long*            - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     long*            - the transformed data array
//
// **************************************************************
long* LINEARTransform::apply(long* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * arrsize;    // Total #vals at pointer.

  long*   optr    = NULL;         // Pointer to output data
  dmDataType intype = dmLONG;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new long[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     float*           - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     float*           - the transformed data array
//
// **************************************************************
float* LINEARTransform::apply(float* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * arrsize;    // Total #vals at pointer.

  float*   optr    = NULL;        // Pointer to output data
  dmDataType intype = dmFLOAT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new float[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     unsigned short*  - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     unsigned short*  - the transformed data array
//
// **************************************************************
unsigned short* LINEARTransform::apply(unsigned short* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * arrsize;    // Total #vals at pointer.

  unsigned short* optr = NULL;    // Pointer to output data
  dmDataType intype = dmUSHORT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new unsigned short[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     unsigned long*   - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     unsigned long*   - the transformed data array
//
// **************************************************************
unsigned long* LINEARTransform::apply(unsigned long* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * arrsize;    // Total #vals at pointer.

  unsigned long* optr = NULL;     // Pointer to output data
  dmDataType intype = dmULONG;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new unsigned long[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     double*          - input data array
//     long             - number of value pairs.
//
//  Returns:
//     double*          - the transformed data array
//
// **************************************************************
double* LINEARTransform::apply(double* inarr, long nsets, long arrsize) {

  double* outarr = NULL;          // output data array.
  double scale  = *(double*)get_parameter_value("SCALE");
  double offset = *(double*)get_parameter_value("OFFSET");

  outarr = new double[nsets * arrsize];
  if ( !outarr ){ return NULL; }

  int ndx = 0;

  for (int ii = 0; ii < nsets; ii++) {      
    for (int jj = 0; jj < arrsize; jj++) {
      ndx = jj + (ii*arrsize);	  
      outarr[ndx] = offset + (scale * inarr[ndx]);
    }
  }

  return outarr;
}


// **************************************************************
//  Method:          invert
//
//  Description:     Apply the inverse transform matrix to the 
//                   input data array.
//
//  Arguments:
//     void*            - interleaved data array
//     dmDataType       - array datatype
//     long             - number of value pairs.
//
//  Returns:
//     void*            - the transformed data array
//
// **************************************************************
void* LINEARTransform::invert(void* in_array, dmDataType in_datatype, long nsets, long arrsize) 
{
  void*   optr = NULL;            // Pointer to output data

  switch (in_datatype) 
  {
    case dmDOUBLE:
      optr = (void*) invert((double*)in_array, nsets, arrsize );
      break;
    case dmSHORT: 
      optr = (void*) invert((short*)in_array, nsets, arrsize );
      break;
    case dmLONG:
      optr = (void*) invert((long*)in_array, nsets, arrsize );
      break;
    case dmFLOAT:
      optr = (void*) invert((float*)in_array, nsets, arrsize );
      break;
    case dmUSHORT:
      optr = (void*) invert((unsigned short*)in_array, nsets, arrsize );
      break;
    case dmULONG:
      optr = (void*) invert((unsigned long*)in_array, nsets, arrsize );
      break;
    case dmBOOL:
    case dmBIT:
    case dmBYTE:
    case dmTEXT:
    case dmUNKNOWNTYPE:
    default:
      optr = NULL;
      break;
  }

  return optr;
}

// **************************************************************
//  Method:          invert
//
//  Description:     Apply the inverse transform matrix to the 
//                   input data array.
//
//  Arguments:
//     short*           - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     short*           - the transformed data array
//
// **************************************************************
short* LINEARTransform::invert(short* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * arrsize;    // Total #vals at pointer.

  short*  optr    = NULL;         // Pointer to output data
  dmDataType intype = dmSHORT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Invert transform to double array..
  outdata = invert( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new short[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          invert
//
//  Description:     Apply the inverse transform matrix to the 
//                   input data array.
//
//  Arguments:
//     long*            - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     long*            - the transformed data array
//
// **************************************************************
long* LINEARTransform::invert(long* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * arrsize;    // Total #vals at pointer.

  long*   optr    = NULL;         // Pointer to output data
  dmDataType intype = dmLONG;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Invert transform to double array..
  outdata = invert( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new long[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          invert
//
//  Description:     Apply the inverse transform matrix to the 
//                   input data array.
//
//  Arguments:
//     float*           - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     float*           - the transformed data array
//
// **************************************************************
float* LINEARTransform::invert(float* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * arrsize;    // Total #vals at pointer.

  float*   optr    = NULL;        // Pointer to output data
  dmDataType intype = dmFLOAT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Invert transform to double array..
  outdata = invert( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new float[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          invert
//
//  Description:     Apply the inverse transform matrix to the 
//                   input data array.
//
//  Arguments:
//     unsigned short*  - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     unsigned short*  - the transformed data array
//
// **************************************************************
unsigned short* LINEARTransform::invert(unsigned short* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * arrsize;    // Total #vals at pointer.

  unsigned short* optr = NULL;    // Pointer to output data
  dmDataType intype = dmUSHORT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Invert transform to double array..
  outdata = invert( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new unsigned short[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          invert
//
//  Description:     Apply the inverse transform matrix to the 
//                   input data array.
//
//  Arguments:
//     unsigned long*   - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     unsigned long*   - the transformed data array
//
// **************************************************************
unsigned long* LINEARTransform::invert(unsigned long* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * arrsize;    // Total #vals at pointer.

  unsigned long* optr = NULL;     // Pointer to output data
  dmDataType intype = dmULONG;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Invert transform to double array..
  outdata = invert( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new unsigned long[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          invert
//
//  Description:     Apply the inverse transform matrix to the 
//                   input data array.
//
//  Arguments:
//     double*          - input data array
//     long             - number of value pairs.
//
//  Returns:
//     double*          - the transformed data array
//
// **************************************************************
double* LINEARTransform::invert(double* inarr, long nsets, long arrsize) {

  double* outarr = NULL;          // output data array.
  long    nvals   = nsets * arrsize;    // Total #vals at pointer.

  if ( arrsize < 1 ) return NULL;

  outarr = new double[nvals];
  if ( !outarr ){ return NULL; }

  double scale  = *(double*)get_parameter_value("SCALE");
  double offset = *(double*)get_parameter_value("OFFSET");

  long ndx = 0;

  for (int ii = 0; ii < nsets; ii++)
  {
    for (int jj = 0; jj < arrsize; jj++)
    {
      ndx = jj + (ii*arrsize) + arrsize - 1 ;
      outarr[ndx] = (inarr[ndx] - offset) / scale; 
    }    
  }

  return outarr;
}


// **************************************************************
//  Method:          get_parameter_list
//
//  Description:     Returns the list of transform parameters.
//                   If list is empty, will generate a generic
//                   list of parameters to return.
//
//  Arguments:       none
//
//  Returns:
//     list<TransParam*> - list of transform parameters
//
// **************************************************************
list<TransParam*> LINEARTransform::get_parameter_list()
{
  if ( params.empty() )
    generate_parameter_list();

  return params;
}

// **************************************************************
//  Method:          generate_parameter_list
//
//  Description:     Creates a list of parameters needed
//                   for the LINEARTransform.
//                   User must fill in value.
//
//  Arguments:       none
//
//  Returns:
//     Loads Transform params attribute.
//                   
// **************************************************************
void LINEARTransform::generate_parameter_list() {
  TransParam*  ptr = NULL;

  ptr = new TransParam("SCALE", NULL, dmDOUBLE, 
		       "Scale", name, 1);
  params.push_back( ptr );

  ptr = new TransParam("OFFSET", NULL, dmDOUBLE, 
		       "Origin Offset", name, 1);
  params.push_back( ptr );

}

