// 
//  Copyright (C) 2006-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

//---------------------------------------------------------------------------
//
//  FILE NAME:          SPLITTransform.cc
//
//  DEVELOPMENT:        CXCDS
//
//  DESCRIPTION:
//
//---------------------------------------------------------------------------

#include "Transform.hh"


// **************************************************************
//  Method:          SPLITTransform
//
//  Description:     Empty constructor.
//                   
//  Arguments:       none
//
//  Returns:         none
//
// **************************************************************
SPLITTransform::SPLITTransform()
{
  clear();

  // Initialize Class Name string
  _strname = "SPLITTransform";

  // Generate a new parameter list for this Transform object.
  generate_parameter_list();
}


// **************************************************************
//  Method:          SPLITTransform
//
//  Description:     SPLITTransform Constructor.
//                   
//  Arguments:       
//     string - the name of the transform
//
//  Returns:         none
//
// **************************************************************
SPLITTransform::SPLITTransform(string in_name)
{
  clear();

  // Initialize Class Name string
  _strname = "SPLITTransform";

  // Initialize Transform Name
  name = in_name;

  // Generate a new parameter list for this Transform object.
  generate_parameter_list();
}


// **************************************************************
//  Method:          SPLITTransform
//
//  Description:     SPLITTransform Constructor.
//                   
//  Arguments:       
//     string - the name of the transform
//     list<TransParam*> - a list of Transform Parameters
//
//  Returns:         none
//
// **************************************************************
SPLITTransform::SPLITTransform(string in_name, list<TransParam*> in_params)
{
  clear();

  // Initialize Class Name string
  _strname = "SPLITTransform";

  // Initialize Transform Name
  name = in_name;

  // Generate a new parameter list for this Transform object.
  generate_parameter_list();

  // Initialize Parameter values from input list..
  set_parameter_list( in_params );

}


// **************************************************************
//  Method:          SPLITTransform
//
//  Description:     SPLITTransform copy constructor
//
//  Arguments:
//     string - the name of the transform
//     SPLITTransform& - the SPLITTransform object to copy
//
//  Returns:         none
//
// **************************************************************
SPLITTransform::SPLITTransform(SPLITTransform &init)
{

  if (this == &init) return;
  clear();
 
  // Initialize Class Name string
  _strname = "SPLITTransform";

  name = init.get_name();

  list<TransParam*>in_params = init.get_parameter_list();

  if (!in_params.empty()) {

    list<TransParam*>::iterator iter;
    for (iter = in_params.begin(); iter != in_params.end(); iter++) {
      TransParam* par_copy = new TransParam( *(*iter) );
      params.push_back(par_copy);
    }

  } else {
    // Generate a new parameter list for this Transform object.
    generate_parameter_list();
  }

}


// **************************************************************
//  Method:          ~SPLITTransform
//
//  Description:     SPLITTransform Desctructor.  
//
//  Arguments:       none
//
//  Returns:         none
//
// **************************************************************
SPLITTransform::~SPLITTransform() {
  clear();
}


// **************************************************************
//  Method:          copy
//
//  Description:     Makes a copy of this SPLITTransform and returns it.
//
//  Arguments:       none
//
//  Returns:         
//     SPLITTransform* - a copy of this Transform
//
// **************************************************************
SPLITTransform* SPLITTransform::copy() 
{
  SPLITTransform* trans = new SPLITTransform( (SPLITTransform&)*this );
  return trans;
}


// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     void*            - interleaved data array
//     dmDataType       - array datatype
//     long             - number of value pairs.
//
//  Returns:
//     void*            - the transformed data array
//
// **************************************************************
void* SPLITTransform::apply(void* in_array, dmDataType in_datatype, long nsets, long arrsize) 
{
  void*   optr = NULL;            // Pointer to output data

  switch (in_datatype) 
  {
    case dmDOUBLE:
      optr = (void*) apply((double*)in_array, nsets, arrsize );
      break;
    case dmSHORT: 
      optr = (void*) apply((short*)in_array, nsets, arrsize );
      break;
    case dmLONG:
      optr = (void*) apply((long*)in_array, nsets, arrsize );
      break;
    case dmFLOAT:
      optr = (void*) apply((float*)in_array, nsets, arrsize );
      break;
    case dmUSHORT:
      optr = (void*) apply((unsigned short*)in_array, nsets, arrsize );
      break;
    case dmULONG:
      optr = (void*) apply((unsigned long*)in_array, nsets, arrsize );
      break;
    case dmBOOL:
    case dmBIT:
    case dmBYTE:
    case dmTEXT:
    case dmUNKNOWNTYPE:
    default:
      optr = NULL;
      break;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     short*           - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     short*           - the transformed data array
//
// **************************************************************
short* SPLITTransform::apply(short* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  short*  optr    = NULL;         // Pointer to output data
  dmDataType intype = dmSHORT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new short[nvals/2];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals/2 );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     long*            - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     long*            - the transformed data array
//
// **************************************************************
long* SPLITTransform::apply(long* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  long*   optr    = NULL;         // Pointer to output data
  dmDataType intype = dmLONG;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new long[nvals/2];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals/2 );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     float*           - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     float*           - the transformed data array
//
// **************************************************************
float* SPLITTransform::apply(float* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  float*   optr    = NULL;        // Pointer to output data
  dmDataType intype = dmFLOAT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new float[nvals/2];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals/2 );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     unsigned short*  - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     unsigned short*  - the transformed data array
//
// **************************************************************
unsigned short* SPLITTransform::apply(unsigned short* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  unsigned short* optr = NULL;    // Pointer to output data
  dmDataType intype = dmUSHORT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new unsigned short[nvals/2];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, optr, nvals/2 );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     unsigned long*   - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     unsigned long*   - the transformed data array
//
// **************************************************************
unsigned long* SPLITTransform::apply(unsigned long* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  unsigned long* optr = NULL;     // Pointer to output data
  dmDataType intype = dmULONG;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new unsigned long[nvals/2];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, optr, nvals/2 );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     Apply the transform to the input array data.
//
//  Arguments:
//     double*          - input data array
//     long             - number of value pairs.
//
//  Returns:
//     double*          - the transformed data array
//
// **************************************************************
double* SPLITTransform::apply(double* inarr, long nsets, long arrsize) {

  double* outarr = NULL;          // output data array.
  long    cptnum = *(short*)get_parameter_value("CPT_NUMBER");
  long    vdim   = *(long*)get_parameter_value("VDIM");

  if (( cptnum < 1 ) || ( arrsize < 1 ) || ( vdim < 2 )) return NULL;

  outarr = new double[nsets * arrsize];
  if ( !outarr ){ return NULL; }

  for (int jj = 0; jj < nsets; jj++) {
      
    for (int kk = 0; kk < arrsize; kk++) {
      outarr[(jj*arrsize)+kk] = inarr[((cptnum-1)*arrsize)+(jj*vdim*arrsize)+kk];
    }
  }

  return outarr;
}

// **************************************************************
//  Method:          get_parameter_list
//
//  Description:     Returns the list of transform parameters.
//                   If list is empty, will generate a generic
//                   list of parameters to return.
//
//  Arguments:       none
//
//  Returns:
//     list<TransParam*> - list of transform parameters
//
// **************************************************************
list<TransParam*> SPLITTransform::get_parameter_list()
{
  if ( params.empty() )
    generate_parameter_list();

  return params;
}

// **************************************************************
//  Method:          generate_parameter_list
//
//  Description:     Creates a list of parameters needed
//                   for the SPLITTransform.
//                   User must fill in value.
//
//  Arguments:       none
//
//  Returns:
//     Loads Transform params attribute.
//                   
// **************************************************************
void SPLITTransform::generate_parameter_list() {
  TransParam*  ptr = NULL;

  ptr = new TransParam("CPT_NUMBER", NULL, dmSHORT, 
		       "Component Number >= 1", name, 1);
  params.push_back( ptr );

  ptr = new TransParam("VDIM", NULL, dmLONG, 
		       "Number of Vector Components", name, 1);
  params.push_back( ptr );

}


