// 
//  Copyright (C) 2006-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

//---------------------------------------------------------------------------
//
//  FILE NAME:          TransParam.cc
//
//  DEVELOPMENT:        CXCDS
//
//  DESCRIPTION:
//
//---------------------------------------------------------------------------

#include "TransParam.hh"
#include "Transform.hh"

#include <stdio.h>

// **************************************************************
//  Method:          TransParam
//
//  Description:     TransParam empty constructor.
//
//  Arguments:       none
// 
//  Returns:         none
//
// **************************************************************
TransParam::TransParam() :
  value(NULL)
{
  clear();
}

// **************************************************************
//  Method:          TransParam
//
//  Description:     TransParam constructor.
//
//  Arguments:
//     string           - parameter name
//     void*            - parameter value
//     dmDataType       - parameter datatype
//     string           - parameter description
//     long             - number of parameter values
//
//  Returns:
//     none
//
// **************************************************************
TransParam::TransParam(string in_name, void* in_value, dmDataType in_datatype, 
		     string in_desc, long in_nvals) :
  value(NULL)
{
  name     = in_name;
  datatype = in_datatype;
  desc     = in_desc;
  parent   = "";
  nvals    = in_nvals;
  value    = NULL;

  set_value( in_value, datatype, nvals);
}

// **************************************************************
//  Method:          TransParam
//
//  Description:     TransParam constructor.
//
//  Arguments:
//     string           - parameter name
//     void*            - parameter value
//     dmDataType       - parameter datatype
//     string           - parameter description
//     string           - parameter parent (the transform name)
//     long             - number of parameter values
//
//  Returns:
//     none
//
// **************************************************************
TransParam::TransParam(string in_name, void* in_value, dmDataType in_datatype, 
		     string in_desc, string in_parent, long in_nvals) :
  value(NULL)
{
  name     = in_name;
  datatype = in_datatype;
  desc     = in_desc;
  parent   = in_parent;
  nvals    = in_nvals;
  value    = NULL;

  // value = in_value;
  set_value( in_value, datatype, nvals);
}


// **************************************************************
//  Method:          TransParam
//
//  Description:     TransParam copy constructor.
//
//  Arguments:
//     TransParam& - the TransParam object to copy
//
//  Returns:
//     none
//
// **************************************************************
TransParam::TransParam(TransParam& init) :
  value(NULL)
{
  if (this == &init) return;
  
  clear();

  name     = init.get_name();
  desc     = init.get_desc();
  parent   = init.get_parent();
  datatype = init.get_datatype();
  nvals    = init.get_nvals();

  set_value( init.get_value(), datatype, nvals );

  return;
}


// **************************************************************
//  Method:          ~TransParam
//
//  Description:     TransParam destructor.
//
//  Arguments:       none
//
//  Returns:         none
//
// **************************************************************
TransParam::~TransParam() 
{
  clear();
}


// **************************************************************
//  Method:          operator=
//
//  Description:     TransParam overloaded assignment operator.
//
//  Arguments:
//     TransParam&       - parameter object to copy
//
//  Returns:
//     TransParam        - updated parameter object
//
// **************************************************************
TransParam& TransParam::operator=(TransParam& init)
{
  if (this == &init) return *this;
  
  clear();

  // MCD TEMP: QUESTION..
  // Should this redefine the parameter?.. 

  name     = init.get_name();
  desc     = init.get_desc();
  parent   = init.get_parent();
  datatype = init.get_datatype();
  nvals    = init.get_nvals();

  set_value( init.get_value(), datatype, nvals );

  return *this;
}


// **************************************************************
//  Method:          clear
//
//  Description:     Resets all TransParam attributes.
//
//  Arguments:       none
//
//  Returns:         none
//
// **************************************************************
void TransParam::clear() 
{
  name.clear();

  if ( value != NULL )
  {
    switch (datatype)
    {
    case dmBOOL:
      delete [] (int*)value;
      break;
    case dmSHORT:
      delete [] (short*)value;
      break;
    case dmLONG:
      delete [] (long*)value;
      break;
    case dmFLOAT:
      delete [] (float*)value;
      break;
    case dmDOUBLE:
      delete [] (double*)value;
      break;
    case dmUSHORT:
      delete [] (unsigned short*)value;
      break;
    case dmULONG:
      delete [] (unsigned long*)value;
      break;
    case dmBYTE:
      delete [] (unsigned char*)value;
      break;
    case dmTEXT:
      delete [] (string*)value;
    default:
      break;
    }
  }

  value = NULL;
  nvals = 0;

  datatype = dmUNKNOWNTYPE;
  desc.clear();
  parent.clear();

}

// **************************************************************
//  Method:          get_name
//
//  Description:     Returns value of name field.
//
//  Arguments:       none
//
//  Returns:
//     string           - name of parameter
//
// **************************************************************
string TransParam::get_name()
{
  return name;
}


// **************************************************************
//  Method:          set_name
//
//  Description:     Assigns value to name field.
//
//  Arguments:       
//     string           - name of parameter
//
//  Returns:         none
//
// **************************************************************
void TransParam::set_name(string in_name)
{
  name = in_name;
}


// **************************************************************
//  Method:          get_value
//
//  Description:     Returns value of value field.
//
//  Arguments:       none
//
//  Returns:
//     void*            - parameter value(s)
//
// **************************************************************
void* TransParam::get_value()
{
  return value;
}


// **************************************************************
//  Method:          set_value
//
//  Description:     Copies input value to value field.
//      Enforces integrity of parameter definition and 
//      casts input values as needed.
// 
//  Arguments:
//     void*            - parameter value
//     dmDataType       - parameter datatype
//     long             - number of values at pointer
//
//  Returns:         none
//
// **************************************************************
void TransParam::set_value(void* in_value, dmDataType in_datatype, long in_nvals)
{
  // Need values
  if ( in_nvals == 0 ){ return; }

  // Need same # of values.
  if( in_nvals != nvals ){ return; }

  // No need to transfer into itself
  if ( in_value == value ){ return; }

  if ( in_value )
  {

    // Be sure there is memory allocated at value pointer
    if ( !value )
    {
      switch (datatype)
      {
      case dmBOOL:
	value = (void*) new int[nvals];
	break;
      case dmSHORT:
	value = (void*) new short[nvals];
	break;
      case dmLONG:
	value = (void*) new long[nvals];
	break;
      case dmFLOAT:
	value = (void*) new float[nvals];
	break;
      case dmDOUBLE:
	value = (void*) new double[nvals];
	break;
      case dmUSHORT:
	value = (void*) new unsigned short[nvals];
	break;
      case dmULONG:
	value = (void*) new unsigned long[nvals];
	break;
      case dmBYTE:
	value = (void*) new unsigned char[nvals];
	break;
      case dmTEXT:
	value = (void*) new string[nvals];
	break;
      default:
	break;
      }
    }

    // Transfer value.. converting datatype as necessary.

    if (datatype != dmTEXT)
      Transform::cast_array_numeric( in_value, in_datatype, datatype, value, nvals );
    else 
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((string*)value)[ii] = ((char**)in_value)[ii];
      }
    }
  }
}


// **************************************************************
//  Method:          set_value - overloads
//
//  Description:     Call generic set_value()
//
//  Arguments:
//     <dtype>*         - parameter value
//     long             - number of values at pointer
//
//  Returns:         none
//
// **************************************************************
void TransParam::set_value(bool* in_value, long in_nvals)
{
  // call generic routine.
  set_value( (void*)in_value, dmBOOL, in_nvals );
}

void TransParam::set_value(short* in_value, long in_nvals)
{
  // call generic routine.
  set_value( (void*)in_value, dmSHORT, in_nvals );
}

void TransParam::set_value(long* in_value, long in_nvals)
{
  // call generic routine.
  set_value( (void*)in_value, dmLONG, in_nvals );
}

void TransParam::set_value(float* in_value, long in_nvals)
{
  // call generic routine.
  set_value( (void*)in_value, dmFLOAT, in_nvals );
}

void TransParam::set_value(double* in_value, long in_nvals)
{
  // call generic routine.
  set_value( (void*)in_value, dmDOUBLE, in_nvals );
}

void TransParam::set_value(unsigned short* in_value, long in_nvals)
{
  // call generic routine.
  set_value( (void*)in_value, dmUSHORT, in_nvals );
}

void TransParam::set_value(unsigned long* in_value, long in_nvals)
{
  // call generic routine.
  set_value( (void*)in_value, dmULONG, in_nvals );
}

void TransParam::set_value(unsigned char* in_value, long in_nvals)
{
  // call generic routine.
  set_value( (void*)in_value, dmBYTE, in_nvals );
}

void TransParam::set_value(string* in_value, long in_nvals)
{
  // call generic routine.
  set_value( (void*)in_value, dmTEXT, in_nvals );
}


// **************************************************************
//  Method:          get_desc
//
//  Description:     Returns value of desc field.
//
//  Arguments:       none
//
//  Returns:
//     string           - parameter description
//
// **************************************************************
string TransParam::get_desc()
{
  return desc;
}


// **************************************************************
//  Method:          set_desc
//
//  Description:     Assigns value to description field.
//
//  Arguments:
//     string           - parameter description
//
//  Returns:         none 
//
// **************************************************************
void TransParam::set_desc(string in_desc)
{
  desc = in_desc;
}


// **************************************************************
//  Method:          get_parent
//
//  Description:     Returns the value of the parent field.
//
//  Arguments:       none
//
//  Returns:
//     string - name of the parameter's parent (i.e. the 
//              transform name)
//
// **************************************************************
string TransParam::get_parent()
{
  return parent;
}

// **************************************************************
//  Method:          set_parent
//
//  Description:     Assigns value to the parent field.
//
//  Arguments:
//     string           - name of the parameter's parent 
//                        (i.e. the transform name)
//
//  Returns:         none
//
// **************************************************************
void TransParam::set_parent(string in_parent)
{
  parent = in_parent;
}


// **************************************************************
//  Method:          get_datatype
//
//  Description:     Returns the value of the datatype field.
//
//  Arguments:       none
//
//  Returns:
//     dmDataType       - the parameter value's datatype
//
// **************************************************************
dmDataType TransParam::get_datatype() 
{
  return datatype;
}


// **************************************************************
//  Method:          get_nvals
//
//  Description:     Returns the value of the nvals field.
//
//  Arguments:       none
//
//  Returns:
//     long             - the number of parameter values
//
// **************************************************************
long TransParam::get_nvals()
{
  return nvals;
}


// **************************************************************
//  Method:          print
//
//  Description:     Prints the TransParam attributes and their 
//                   values.
//
//  Arguments:       none
//
//  Returns:
//     string           - formatted string describing Parameter
//
// **************************************************************
string TransParam::print() 
{

  char retstr[BUFFER_LENGTH];
  retstr[0] = '\0';

  char tmpstr[MAX_LENGTH];
  tmpstr[0] = '\0';

  char tmp2str[MAX_LENGTH];
  tmp2str[0] = '\0';

  strcat(retstr, (char*)("Name: " + name).c_str());

  switch (datatype) {
      case dmBOOL    :
        if ( value ) {
          sprintf(tmpstr, "\tType: dmBOOL\tValue: %i   \t", (int)((bool*)value)[0]);

	  for (int ii = 1; ii < nvals; ii++) {
	    sprintf(tmp2str, ", %i", (int)((bool*)value)[ii]);
	    strcat(tmpstr, tmp2str);
	  }
	  strcat(tmpstr, "\t   ");
        } else
          sprintf(tmpstr, "\tType: dmBOOL\tValue: NULL   ");
        break;

      case dmSHORT   :
        if ( value ) {
          sprintf(tmpstr, "\tType: dmSHORT\tValue: %i", ((short*)value)[0]);
	  
	  for (int ii = 1; ii < nvals; ii++) {
	    sprintf(tmp2str, ", %i", ((short*)value)[ii]);
	    strcat(tmpstr, tmp2str);
	  }
	  strcat(tmpstr, "\t   ");

        } else
          sprintf(tmpstr, "\tType: dmSHORT\tValue: NULL   ");
        break;
      case dmUSHORT  :
        if ( value ) {
          sprintf(tmpstr, "\tType: dmUSHORT\tValue: %u   \t", ((unsigned short*)value)[0]);
	  
	  for (int ii = 1; ii < nvals; ii++) {
	    sprintf(tmp2str, ", %u", ((unsigned short*)value)[ii]);
	    strcat(tmpstr, tmp2str);
	  }
	  strcat(tmpstr, "\t   ");

        } else
          sprintf(tmpstr, "\tType: dmUSHORT\tValue: NULL   ");
        break;
      case dmLONG    :
        if ( value ) {
          sprintf(tmpstr, "\tType: dmLONG\tValue: %li", ((long*)value)[0]);

	  for (int ii = 1; ii < nvals; ii++) {
	    sprintf(tmp2str, ", %li", ((long*)value)[ii]);
	    strcat(tmpstr, tmp2str);
	  }
	  strcat(tmpstr, "\t   ");

        } else
          sprintf(tmpstr, "\tType: dmLONG\tValue: NULL   ");
        break;
      case dmULONG   :
        if ( value ) {
          sprintf(tmpstr, "\tType: dmULONG\tValue: %lu   \t", ((unsigned long*)value)[0]);

	  for (int ii = 1; ii < nvals; ii++) {
	    sprintf(tmp2str, ", %lu", ((unsigned long*)value)[ii]);
	    strcat(tmpstr, tmp2str);
	  }
	  strcat(tmpstr, "\t   ");

        } else
          sprintf(tmpstr, "\tType: dmULONG\tValue: NULL   ");
        break;
      case dmFLOAT   :
        if ( value ) {
          sprintf(tmpstr, "\tType: dmFLOAT\tValue: %f", ((float*)value)[0]);
	  
	  for (int ii = 1; ii < nvals; ii++) {
	    sprintf(tmp2str, ", %f", ((float*)value)[ii]);
	    strcat(tmpstr, tmp2str);
	  }
	  strcat(tmpstr, "\t   ");
	  
	} else
          sprintf(tmpstr, "\tType: dmFLOAT\tValue: NULL   ");
        break;
      case dmDOUBLE  :
        if ( value ) {
          sprintf(tmpstr, "\tType: dmDOUBLE\tValue: %lf", ((double*)value)[0]);

	  for (int ii = 1; ii < nvals; ii++) {
	    sprintf(tmp2str, ", %lf", ((double*)value)[ii]);
	    strcat(tmpstr, tmp2str);
	  }
	  strcat(tmpstr, "\t   ");

        } else
          sprintf(tmpstr, "\tType: dmDOUBLE\tValue: NULL   ");
        break;
      case dmBYTE    :
      case dmBIT     :
        if ( value ) {
          sprintf(tmpstr, "\tType: dmBYTE\tValue: %u   \t", ((unsigned char*)value)[0]);

	  for (int ii = 1; ii < nvals; ii++) {
	    sprintf(tmp2str, ", %u", ((unsigned char*)value)[ii]);
	    strcat(tmpstr, tmp2str);
	  }
	  strcat(tmpstr, "\t   ");

        } else
          sprintf(tmpstr, "\tType: dmBYTE\tValue: NULL   ");
        break;
      case dmTEXT    :
        if ( value ) {
          sprintf(tmpstr, "\tType: dmCHAR\tValue: %s   ", ((string*)value)[0].c_str());

	  for (int ii = 1; ii < nvals; ii++) {
	    sprintf(tmp2str, ", %s", ((string*)value)[ii].c_str());
	    strcat(tmpstr, tmp2str);
	  }
	  strcat(tmpstr, "\t   ");

	}  else
          sprintf(tmpstr, "\tType: dmCHAR\tValue: NULL   ");
        break;
      default :
        sprintf(tmpstr, "\tType: dmUNKOWNTYPE\tValue: NULL   ");
        break;
  }

  strcat(retstr, tmpstr);
  strcat(retstr, (char*)("\tDesc: " + desc).c_str());
  strcat(retstr, (char*)("\tParent: " + parent).c_str());

  return retstr;


}

