// 
//  Copyright (C) 2006-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

//---------------------------------------------------------------------------
//
//  FILE NAME:          Transform.cc
//
//  DEVELOPMENT:        CXCDS
//
//  DESCRIPTION:
//
//---------------------------------------------------------------------------


#include "Transform.hh"
#include "wcs.h"


// **************************************************************
//  Method:          Transform
//
//  Description:     Transform empty constructor.
//
//  Arguments:       none
//
//  Returns:         Transform object with initialized values
//                   including Parameter list.
//
// **************************************************************
Transform::Transform() {
  clear();

  // Initialize Class Name string
  _strname = "Transform";

  // Generate a new parameter list for this Transform object.
  // NOTE: none for base class
}


// **************************************************************
//  Method:          Transform
//
//  Description:     Transform constructor.
//
//  Arguments:       
//     string - name of transform
//
//  Returns:         Transform object with initialized values
//                   including Parameter list.
//
// **************************************************************
Transform::Transform(string in_name)
{
  clear();

  // Initialize Class Name string
  _strname = "Transform";

  // Initialize Transform Name
  name = in_name;

  // Generate a new parameter list for this Transform object.
  // NOTE: none for base class

}


// **************************************************************
//  Method:          Transform
//
//  Description:     Transform constructor.
//
//  Arguments:       
//     string           - name of transform
//     list<TransParam*> - list of transform parameters
//
//  Returns:         Transform object with initialized values
//                   including Parameter list.
//
// **************************************************************
Transform::Transform(string in_name, list<TransParam*> in_params) 
{
  clear();

  // Initialize Class Name string
  _strname = "Transform";

  // Initialize Transform Name
  name = in_name;

  // Generate a new parameter list for this Transform object.
  // NOTE: none for base class

  // Initialize Parameter values from input list..
  // NOTE: no parameters to match.. therefore nothing to do

}


// **************************************************************
//  Method:          Transform
//
//  Description:     Transform copy constructor
//
//  Arguments:
//     string        - name of transform
//     Transform&    - the Transform object to copy
//
//  Returns:         none
//
// **************************************************************
Transform::Transform(Transform &init) 
{

  if (this == &init) return;
  clear();

  // Initialize Class Name string
  _strname = "Transform";

  name = init.get_name();

  list<TransParam*> in_params = init.get_parameter_list();

  if ( !in_params.empty() ) {
    list<TransParam*>::iterator iter;
    for (iter = in_params.begin(); iter != in_params.end(); iter++) {
      TransParam* par_copy = new TransParam( *(*iter) );
      params.push_back(par_copy);
    }
  }

}


// **************************************************************
//  Method:          Transform
//
//  Description:     Transform desctructor.
//
//  Arguments:       none
//
//  Returns:         none
//
// **************************************************************
Transform::~Transform()
{
  clear();
}


// **************************************************************
//  Method:          clear
//
//  Description:     Resets Transform attributes.
//
//  Arguments:       none
//
//  Returns:         none
//
// **************************************************************
void Transform::clear()
{
  if ( !(params.empty()) ) {
    list<TransParam*>::reverse_iterator par_iter;
    for (par_iter = params.rbegin(); par_iter != params.rend(); par_iter++) {
      delete *par_iter;
    }
    params.clear();
  }
  
  name.clear();
}


// **************************************************************
//  Method:          copy
//
//  Description:     Makes a copy of this Transform and returns it.
//
//  Arguments:       none
//
//  Returns:         
//     Transform* - a copy of this Transform
//
// **************************************************************
Transform* Transform::copy() 
{
  Transform* trans = new Transform( (Transform&)*this );
  return trans;
}


// **************************************************************
//  Method:          get_className
//
//  Description:     Returns the string representation of the 
//                   Transform Class name.
//
//  Arguments:       none
//
//  Returns:         
//     string - Transform class name
//
// **************************************************************
string Transform::get_className() 
{
  return _strname;
}

// **************************************************************
//  Method:          get_name
//
//  Description:     Returns value of the name attribute.
//
//  Arguments:       none
//
//  Returns:         
//     string - name of transform
//
// **************************************************************
string Transform::get_name() {
  return name;
}

// **************************************************************
//  Method:          set_name
//
//  Description:     Assigns a value to the name attribute.
//                   Updates parent field of contained Parameters
//
//  Arguments:       
//     string           - name of transform
//
//  Returns:         none
//
// **************************************************************
void Transform::set_name(string in_name) {

  // set name attribute
  name = in_name;

  // update parent field of contained Parameters
  if ( !(params.empty()) ) {
    list<TransParam*>::iterator iter;
    for (iter = params.begin(); iter != params.end(); iter++) 
    {
      (*iter)->set_parent( name );
    }
  }

}

// **************************************************************
//  Method:          apply
//
//  Description:     Not yet implemented for generic Transform.
//
//  Arguments:       
//     void*            - data array to apply the transform to
//     dmDataType       - array datatype
//
//  Returns:         none
//
// **************************************************************
void* Transform::apply(void* in_array, dmDataType in_datatype, long nsets, long arrsize) {
  return NULL;
}


// **************************************************************
//  Method:          invert
//
//  Description:     Not yet implemented for generic Transform. 
//
//  Arguments:
//     void*            - data array to invert
//     dmDataType       - array datatype
//
//  Returns:         none
//
// **************************************************************
void* Transform::invert(void* in_array, dmDataType in_datatype, long nsets, long arrsize)
{
  return NULL;
}


// **************************************************************
//  Method:          get_transform_matrix
//
//  Description:     Not yet implemented for generic Transform. 
//
//  Arguments:
//     const string& - transform name
//
//  Returns:
//     double** - the transform matrix
//
// **************************************************************
double** Transform::get_transform_matrix() 
{
  return NULL;
}

// **************************************************************
//  Method:          set_transform_matrix
//
//  Description:     Not yet implemented for generic Transform. 
//
//  Arguments:
//     double** - the transform matrix
//
//  Returns:
//
// **************************************************************
void Transform::set_transform_matrix(double** matrix) 
{

}


// **************************************************************
//  Method:          get_parameter
//
//  Description:     Searches the parameter list by name and 
//                   returns a pointer to that parameter.  If
//                   not found, returns NULL.
//
//  Arguments:       
//     string           - parameter name
//
//  Returns:         
//     TransParam*       - pointer to the parameter or NULL
//
// **************************************************************
TransParam* Transform::get_parameter(string in_name)
{

  list< TransParam* >::iterator iter; 
  TransParam* retpar = NULL;

  for ( iter = params.begin(); iter != params.end(); iter++) 
  {
    if ( (*iter)->get_name() == in_name ){
      retpar = (*iter);
      break;
    }
  }

  return retpar;
}


// **************************************************************
//  Method:          set_parameter
//
//  Description:     Replaces already existing parameter.
//
//  Arguments:       none
//
//  Returns:         none
//
// **************************************************************
dmErrCode Transform::set_parameter(TransParam* in_param)
{
  string in_name = in_param->get_name();
  TransParam* par = get_parameter(in_name);

  if (par) 
  {
    if (par != in_param)
    {
      // Matched parameter... set value.
      par->set_value( in_param->get_value(),
		      in_param->get_datatype(),
		      in_param->get_nvals()     );
    }
    return dmSUCCESS;
  }

  return dmFAILURE;
}


// **************************************************************
//  Method:          get_parameter_list
//
//  Description:     Returns the transform parameter list.
//
//  Arguments:       none
//
//  Returns:         
//     list<TransParam*> - the transform parameter list
//
// **************************************************************
list<TransParam*> Transform::get_parameter_list()
{
  return params;
}


// **************************************************************
//  Method:          set_parameter_list
//
//  Description:
//     Updates values for Transform parameters which
//     match those in the input TransParam List.
//
//  NOTE:
//     A 'match' is determined using the Parameter name
//     attribute only.
//
//  Arguments:       
//     list<TransParam*> - transform parameter list
//
//  Returns:         none
//
// **************************************************************
void Transform::set_parameter_list(list<TransParam*> in_params)
{

  if ( !(in_params.empty()) && !(params.empty()) ) 
  {
    list<TransParam*>::iterator inIter;
    TransParam* myparam;

    // Search input parameter list for matches to this Transform.
    for ( inIter = in_params.begin(); inIter != in_params.end(); inIter++) 
    {
      myparam = get_parameter( (*inIter)->get_name() );

      if ( myparam != NULL )
      {
	// Matched parameter names... set value.
	myparam->set_value( (*inIter)->get_value(),
		 	    (*inIter)->get_datatype(),
			    (*inIter)->get_nvals()     );
      }
    }  // end inParam loop
  }

}


// **************************************************************
//  Method:          get_parameter_value
//
//  Description:     Finds the value of transform parameter and
//                   returns it.
//
//  Arguments:       
//     string           - parameter name
//
//  Returns:         
//     void*            - parameter value
//
// **************************************************************
void* Transform::get_parameter_value(string in_name)
{
  TransParam* par = get_parameter(in_name);

  if (par)
    return par->get_value();

  return NULL;
}


// **************************************************************
//  Method:          print_parameter_list
//
//  Description:     Prints the list of transform parameters.
//
//  Arguments:       none
//
//  Returns:         none
//
// **************************************************************
string Transform::print_parameter_list() {

  if (params.empty()) {
    return "Parameter list is empty.";
  }

  list< TransParam* >::iterator par_iter; 
  string retstr = "";

  for (par_iter = params.begin(); par_iter != params.end(); par_iter++) {
    retstr += (*par_iter)->print() + "\n";
  }

  return retstr;
}

// ******************************************************************************
// CUT...
// **************************************************************
//  Method:          cast_array_numeric
//
//  Description:     Converts array to specified datatype.
//    Memory for output values is assumed to be allocated
//    and large enough to hold all output values.
//
//  Arguments:
//     void*            - input data array
//     dmDataType       - input datatype
//     dmDataType       - output datatype
//     void*            - output array destination
//     long             - number of values.
//
//  Returns:         none
//
// **************************************************************
void Transform::cast_array_numeric( void* inarr,
				    dmDataType intype,
				    dmDataType outtype, 
				    void* outarr,
				    long  nvals) 
{

  // NOTE: In down casting.. have overflow to check!!!

  switch( intype ) 
  {
  case dmSHORT:
  {
    switch( outtype )
    {
    case dmSHORT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((short*)outarr)[ii] = (short)(((short*)inarr)[ii]);
      }
    }
    break;
    case dmLONG:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((long*)outarr)[ii] = (long)(((short*)inarr)[ii]);
      }
    }
    break;
    case dmFLOAT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((float*)outarr)[ii] = (float)(((short*)inarr)[ii]);
      }
    }
    break;
    case dmDOUBLE:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((double*)outarr)[ii] = (double)(((short*)inarr)[ii]);
      }
    }
    break;
    case dmUSHORT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned short*)outarr)[ii] = (unsigned short)(((short*)inarr)[ii]);
      }
    }
    break;
    case dmULONG:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned long*)outarr)[ii] = (unsigned long)(((short*)inarr)[ii]);
      }
    }
    break;
    case dmBOOL:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((int*)outarr)[ii] = (int)(((short*)inarr)[ii]);
      }
    }
    break;
    case dmBIT:
    case dmBYTE:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned char*)outarr)[ii] = (unsigned char)(((short*)inarr)[ii]);
      }
    }
    break;
    default:
      break;
    }
  }  
  break;
  case dmLONG:
  {
    switch( outtype )
    {
    case dmSHORT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((short*)outarr)[ii] = (short)(((long*)inarr)[ii]);
      }
    }
    break;
    case dmLONG:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((long*)outarr)[ii] = (long)(((long*)inarr)[ii]);
      }
    }
    break;
    case dmFLOAT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((float*)outarr)[ii] = (float)(((long*)inarr)[ii]);
      }
    }
    break;
    case dmDOUBLE:
    {
      for (int ii = 0; ii < nvals; ii++) 
      {
	((double*)outarr)[ii] = (double)(((long*)inarr)[ii]);
      }
    }
    break;
    case dmUSHORT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned short*)outarr)[ii] = (unsigned short)(((long*)inarr)[ii]);
      }
    }
    break;
    case dmULONG:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned long*)outarr)[ii] = (unsigned long)(((long*)inarr)[ii]);
      }
    }
    break;
    case dmBOOL:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((int*)outarr)[ii] = (int)(((long*)inarr)[ii]);
      }
    }
    break;
    case dmBIT:
    case dmBYTE:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned char*)outarr)[ii] = (unsigned char)(((long*)inarr)[ii]);
      }
    }
    break;
    default:
      break;
    }
  }
  break;
  case dmFLOAT:
  {
    switch( outtype )
    {
    case dmSHORT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((short*)outarr)[ii] = (short)(((float*)inarr)[ii]);
      }
    }
    break;
    case dmLONG:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((long*)outarr)[ii] = (long)(((float*)inarr)[ii]);
      }
    }
    break;
    case dmFLOAT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((float*)outarr)[ii] = (float)(((float*)inarr)[ii]);
      }
    }
    break;
    case dmDOUBLE:
    {
      for (int ii = 0; ii < nvals; ii++) 
      {
	((double*)outarr)[ii] = (double)(((float*)inarr)[ii]);
      }
    }
    break;
    case dmUSHORT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned short*)outarr)[ii] = (unsigned short)(((float*)inarr)[ii]);
      }
    }
    break;
    case dmULONG:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned long*)outarr)[ii] = (unsigned long)(((float*)inarr)[ii]);
      }
    }
    break;
    case dmBOOL:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((int*)outarr)[ii] = (int)(((float*)inarr)[ii]);
      }
    }
    break;
    case dmBIT:
    case dmBYTE:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned char*)outarr)[ii] = (unsigned char)(((float*)inarr)[ii]);
      }
    }
    break;
    default:
      break;
    }
  }
  break;
  case dmDOUBLE:
  {
    switch( outtype )
    {
      case dmSHORT:
      {
	for (int ii = 0; ii < nvals; ii++)
	  ((short*)outarr)[ii] = (short)((double*)inarr)[ii];
      }
      break;
      case dmLONG:
      {
	for (int ii = 0; ii < nvals; ii++)
	  ((long*)outarr)[ii] = (long)((double*)inarr)[ii];
      }
      break;
      case dmFLOAT:
      {
	for (int ii = 0; ii < nvals; ii++)
	  ((float*)outarr)[ii] = (float)((double*)inarr)[ii];
      }
      break;
      case dmDOUBLE:
      {
	for (int ii = 0; ii < nvals; ii++)
	  ((double*)outarr)[ii] = (double)((double*)inarr)[ii];
      }
      break;
      case dmUSHORT:
      {
	for (int ii = 0; ii < nvals; ii++)
	  ((unsigned short*)outarr)[ii] = (unsigned short)((double*)inarr)[ii];
      }
      break;
      case dmULONG:
      {
	for (int ii = 0; ii < nvals; ii++)
	  ((unsigned long*)outarr)[ii] = (unsigned long)((double*)inarr)[ii];
      }
      break;
      case dmBOOL:
      {
	for (int ii = 0; ii < nvals; ii++)
	  ((int*)outarr)[ii] = (int)((int*)inarr)[ii];
      }
      break;
      case dmBYTE:
      case dmBIT:
      {
	for (int ii = 0; ii < nvals; ii++)
	  ((unsigned char*)outarr)[ii] = (unsigned char)((double*)inarr)[ii];
      }
      break;
      default:
        outarr = NULL;
        break;
    }
  }
  break;
  case dmUSHORT:
  {
    switch( outtype )
    {
    case dmSHORT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((short*)outarr)[ii] = (short)(((unsigned short*)inarr)[ii]);
      }
    }
    break;
    case dmLONG:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((long*)outarr)[ii] = (long)(((unsigned short*)inarr)[ii]);
      }
    }
    break;
    case dmFLOAT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((float*)outarr)[ii] = (float)(((unsigned short*)inarr)[ii]);
      }
    }
    break;
    case dmDOUBLE:
    {
      for (int ii = 0; ii < nvals; ii++) 
      {
	((double*)outarr)[ii] = (double)(((unsigned short*)inarr)[ii]);
      }
    }
    case dmUSHORT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned short*)outarr)[ii] = (unsigned short)(((unsigned short*)inarr)[ii]);
      }
    }
    break;
    case dmULONG:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned long*)outarr)[ii] = (unsigned long)(((unsigned short*)inarr)[ii]);
      }
    }
    break;
    case dmBOOL:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((int*)outarr)[ii] = (int)(((unsigned short*)inarr)[ii]);
      }
    }
    case dmBIT:
    case dmBYTE:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned char*)outarr)[ii] = (unsigned char)(((unsigned short*)inarr)[ii]);
      }
    }
    default:
      break;
    }
  }
  break;
  case dmULONG:
  {
    switch( outtype )
    {
    case dmSHORT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((short*)outarr)[ii] = (short)(((unsigned long*)inarr)[ii]);
      }
    }
    case dmLONG:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((long*)outarr)[ii] = (long)(((unsigned long*)inarr)[ii]);
      }
    }
    break;
    case dmFLOAT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((float*)outarr)[ii] = (float)(((unsigned long*)inarr)[ii]);
      }
    }
    break;
    case dmDOUBLE:
    {
      for (int ii = 0; ii < nvals; ii++) 
      {
	((double*)outarr)[ii] = (double)(((unsigned long*)inarr)[ii]);
      }
    }
    break;
    case dmUSHORT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned short*)outarr)[ii] = (unsigned short)(((unsigned long*)inarr)[ii]);
      }
    }
    break;
    case dmULONG:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned long*)outarr)[ii] = (unsigned long)(((unsigned long*)inarr)[ii]);
      }
    }
    break;
    case dmBOOL:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((int*)outarr)[ii] = (int)(((unsigned long*)inarr)[ii]);
      }
    }
    break;
    case dmBIT:
    case dmBYTE:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned char*)outarr)[ii] = (unsigned char)(((unsigned long*)inarr)[ii]);
      }
    }
    break;
    default:
      break;
    }
  }
  break;
  case dmBOOL:
  {
    switch( outtype )
    {
    case dmSHORT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((short*)outarr)[ii] = (short)(((int*)inarr)[ii]);
      }
    }
    break;
    case dmLONG:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((long*)outarr)[ii] = (long)(((int*)inarr)[ii]);
      }
    }
    break;
    case dmFLOAT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((float*)outarr)[ii] = (float)(((int*)inarr)[ii]);
      }
    }
    break;
    case dmDOUBLE:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((double*)outarr)[ii] = (double)(((int*)inarr)[ii]);
      }
    }
    break;
    case dmUSHORT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned short*)outarr)[ii] = (unsigned short)(((int*)inarr)[ii]);
      }
    }
    break;
    case dmULONG:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned long*)outarr)[ii] = (unsigned long)(((int*)inarr)[ii]);
      }
    }
    break;
    case dmBOOL:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((int*)outarr)[ii] = (int)(((int*)inarr)[ii]);
      }
    }
    break;
    case dmBIT:
    case dmBYTE:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned char*)outarr)[ii] = (unsigned char)(((int*)inarr)[ii]);
      }
    }
    break;
    default:
      break;
    }
  }
  break;
  case dmBIT:
  case dmBYTE:
  {
    switch( outtype )
    {
    case dmSHORT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((short*)outarr)[ii] = (short)(((unsigned char*)inarr)[ii]);
      }
    }
    break;
    case dmLONG:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((long*)outarr)[ii] = (long)(((unsigned char*)inarr)[ii]);
      }
    }
    break;
    case dmFLOAT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((float*)outarr)[ii] = (float)(((unsigned char*)inarr)[ii]);
      }
    }
    break;
    case dmDOUBLE:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((double*)outarr)[ii] = (double)(((unsigned char*)inarr)[ii]);
      }
    }
    break;
    case dmUSHORT:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned short*)outarr)[ii] = (unsigned short)(((unsigned char*)inarr)[ii]);
      }
    }
    break;
    case dmULONG:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned long*)outarr)[ii] = (unsigned long)(((unsigned char*)inarr)[ii]);
      }
    }
    break;
    case dmBOOL:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((int*)outarr)[ii] = (int)(((unsigned char*)inarr)[ii]);
      }
    }
    break;
    case dmBIT:
    case dmBYTE:
    {
      for (int ii = 0; ii < nvals; ii++)
      {
	((unsigned char*)outarr)[ii] = (unsigned char)(((unsigned char*)inarr)[ii]);
      }
    }
    break;
    default:
      break;
    }
  }
  break;
  default:
    break;
  }

    
}
// END CUT...



// ******************************************************************************

// TO BE DELETED AFTER TRANSITION


WCSTANTransform::WCSTANTransform() {
  wcstrans = new WCSTransform();
  _strname = "WCSTANTransform";
  params = wcstrans->params;
}

WCSTANTransform::WCSTANTransform(string in_name) {
  wcstrans = new WCSTransform(in_name);
  _strname = "WCSTANTransform";
  name = in_name;
  params = wcstrans->params;
}

WCSTANTransform::WCSTANTransform(string in_name, list<TransParam*> in_params){
  wcstrans =  new WCSTransform(in_name, in_params);
  _strname = "WCSTANTransform";
  name = in_name;
  params = wcstrans->params;
}

WCSTANTransform::WCSTANTransform(WCSTANTransform &init) {
  wcstrans = new WCSTransform( *(init.wcstrans) );
  _strname = "WCSTANTransform";
  params = wcstrans->params;
}

WCSTANTransform::~WCSTANTransform() {
  name.clear();
  _strname.clear();
  params.clear();
  delete wcstrans;
}

WCSTANTransform* WCSTANTransform::copy(){
  WCSTANTransform* trans = new WCSTANTransform( (WCSTANTransform&)(*this) );
  return trans;
}

void* WCSTANTransform::apply(void* in_array, dmDataType in_datatype, long nsets, long arrsize) {
  return wcstrans->apply(in_array, in_datatype, nsets, arrsize);
}

short* WCSTANTransform::apply(short* in_array, long nsets, long arrsize) {
  return wcstrans->apply(in_array, nsets, arrsize);
}

long* WCSTANTransform::apply(long* in_array, long nsets, long arrsize) {
  return wcstrans->apply(in_array, nsets, arrsize);
}

float* WCSTANTransform::apply(float* in_array, long nsets, long arrsize) {
  return wcstrans->apply(in_array, nsets, arrsize);
}

double* WCSTANTransform::apply(double* in_array, long nsets, long arrsize) {
  return wcstrans->apply(in_array, nsets, arrsize);
}

unsigned short* WCSTANTransform::apply(unsigned short* in_array, long nsets, long arrsize) {
  return wcstrans->apply(in_array, nsets, arrsize);
}

unsigned long*  WCSTANTransform::apply(unsigned long* in_array, long nsets, long arrsize) {
  return wcstrans->apply(in_array, nsets, arrsize);
}

void* WCSTANTransform::invert(void* in_array, dmDataType in_datatype, long nsets, long arrsize) {
  return wcstrans->invert(in_array, in_datatype, nsets, arrsize);
}

short* WCSTANTransform::invert(short* in_array, long nsets, long arrsize) {
  return wcstrans->invert(in_array, nsets, arrsize);
}

long* WCSTANTransform::invert(long* in_array, long nsets, long arrsize) {
  return wcstrans->invert(in_array, nsets, arrsize);
}

float* WCSTANTransform::invert(float* in_array, long nsets, long arrsize) {
  return wcstrans->invert(in_array, nsets, arrsize);
}

double* WCSTANTransform::invert(double* in_array, long nsets, long arrsize) {
  return wcstrans->invert(in_array, nsets, arrsize);
}

unsigned short* WCSTANTransform::invert(unsigned short* in_array, long nsets, long arrsize) {
  return wcstrans->invert(in_array, nsets, arrsize);
}

unsigned long*  WCSTANTransform::invert(unsigned long* in_array, long nsets, long arrsize) {
  return wcstrans->invert(in_array, nsets, arrsize);
}

double** WCSTANTransform::get_transform_matrix(){
  return wcstrans->get_transform_matrix();
}

void WCSTANTransform::set_transform_matrix(double** matrix) {
  wcstrans->set_transform_matrix(matrix);
  TransParam* ctype_param = get_parameter("CTYPE");
  string ctype[2] = {"RA---TAN", "DEC--TAN"};
  ctype_param->set_value( (void*)ctype, dmTEXT, 2 );

}

list<TransParam*> WCSTANTransform::get_parameter_list() {
  return wcstrans->get_parameter_list();
}


// ******************************************************************************

