// 
//  Copyright (C) 2006-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

//---------------------------------------------------------------------------
//
//  FILE NAME:          Transform.hh
//
//  DEVELOPMENT:        CXCDS
//
//  DESCRIPTION:
//
//---------------------------------------------------------------------------

#ifndef _TRANSFORM_HH_
#define _TRANSFORM_HH_

#include <typeinfo>
#include <iostream>
#include <string>
#include <list>
#include <math.h>
#include <cstdlib>
#include <cstring>

#ifndef _TRANSPARAM_HH_
#include "TransParam.hh"
#endif

using namespace std;

class Transform 
{
  friend class CrateData;
  friend class TransParam;

public:
  Transform();
  Transform(string in_name);
  Transform(string in_name, list<TransParam*> in_params);
  Transform(Transform &init);
  virtual ~Transform();

  string get_className();
  string get_name();
  void   set_name(string in_name);

  virtual list<TransParam*> get_parameter_list();
  TransParam* get_parameter(string in_name);
  void* get_parameter_value(string in_name);

  void set_parameter_list(list<TransParam*> in_params);
  dmErrCode set_parameter(TransParam* in_param);

  string print_parameter_list();

  virtual void* apply(void* in_array, dmDataType in_datatype, long nsets, long arrsize=1);
  virtual void* invert(void* in_array, dmDataType in_datatype, long nsets, long arrsize=1);

  virtual double** get_transform_matrix();
  virtual void set_transform_matrix(double** matrix);

  void clear();

  virtual Transform* copy();

protected:
  string _strname;  // String representation of class name.
  string name;
  list<TransParam*> params;

  static void cast_array_numeric( void* inarr, dmDataType intype, 
				  dmDataType outtype, void* outarr, long nvals);

};

class WCSTransform : public Transform
{
  friend class WCSTANTransform;
public:
  WCSTransform();
  WCSTransform(string in_name);
  WCSTransform(string in_name, list<TransParam*> in_params);
  WCSTransform(WCSTransform &init);
  ~WCSTransform();

  WCSTransform* copy();

  list<TransParam*> get_parameter_list();

  void*           apply(void* in_array, dmDataType in_datatype, long nsets, long arrsize=1);
  short*          apply(short*  in_array, long nsets, long arrsize=1);
  long*           apply(long*   in_array, long nsets, long arrsize=1);
  float*          apply(float*  in_array, long nsets, long arrsize=1);
  double*         apply(double* in_array, long nsets, long arrsize=1);
  unsigned short* apply(unsigned short* in_array, long nsets, long arrsize=1);
  unsigned long*  apply(unsigned long*  in_array, long nsets, long arrsize=1);

  void*           invert(void* in_array, dmDataType in_datatype, long nsets, long arrsize=1);
  short*          invert(short*  in_array, long nsets, long arrsize=1);
  long*           invert(long*   in_array, long nsets, long arrsize=1);
  float*          invert(float*  in_array, long nsets, long arrsize=1);
  double*         invert(double* in_array, long nsets, long arrsize=1);
  unsigned short* invert(unsigned short* in_array, long nsets, long arrsize=1);
  unsigned long*  invert(unsigned long*  in_array, long nsets, long arrsize=1);

  double** get_transform_matrix();
  void     set_transform_matrix(double** matrix);

protected:
  void generate_parameter_list();

};


class WCSTANTransform : public Transform 
{

public:
  WCSTANTransform();
  WCSTANTransform(string in_name);
  WCSTANTransform(string in_name, list<TransParam*> in_params);
  WCSTANTransform(WCSTANTransform &init);
  ~WCSTANTransform();

  WCSTANTransform* copy();

  list<TransParam*> get_parameter_list();

  void*           apply(void* in_array, dmDataType in_datatype, long nsets, long arrsize=1);
  short*          apply(short*  in_array, long nsets, long arrsize=1);
  long*           apply(long*   in_array, long nsets, long arrsize=1);
  float*          apply(float*  in_array, long nsets, long arrsize=1);
  double*         apply(double* in_array, long nsets, long arrsize=1);
  unsigned short* apply(unsigned short* in_array, long nsets, long arrsize=1);
  unsigned long*  apply(unsigned long*  in_array, long nsets, long arrsize=1);

  void*           invert(void* in_array, dmDataType in_datatype, long nsets, long arrsize=1);
  short*          invert(short*  in_array, long nsets, long arrsize=1);
  long*           invert(long*   in_array, long nsets, long arrsize=1);
  float*          invert(float*  in_array, long nsets, long arrsize=1);
  double*         invert(double* in_array, long nsets, long arrsize=1);
  unsigned short* invert(unsigned short* in_array, long nsets, long arrsize=1);
  unsigned long*  invert(unsigned long*  in_array, long nsets, long arrsize=1);

  double** get_transform_matrix();
  void     set_transform_matrix(double** matrix);

protected:

private:
  WCSTransform*  wcstrans;

};


class LINEAR2DTransform : public Transform
{

public:
  LINEAR2DTransform();
  LINEAR2DTransform(string in_name);
  LINEAR2DTransform(string in_name, list<TransParam*> in_params);
  LINEAR2DTransform(LINEAR2DTransform &init);
  ~LINEAR2DTransform();

  LINEAR2DTransform* copy();

  list<TransParam*> get_parameter_list();

  void*           apply(void* in_array, dmDataType in_datatype, long nsets, long arrsize=1);
  short*          apply(short*  in_array, long nsets, long arrsize=1);
  long*           apply(long*   in_array, long nsets, long arrsize=1);
  float*          apply(float*  in_array, long nsets, long arrsize=1);
  double*         apply(double* in_array, long nsets, long arrsize=1);
  unsigned short* apply(unsigned short* in_array, long nsets, long arrsize=1);
  unsigned long*  apply(unsigned long*  in_array, long nsets, long arrsize=1);

  void*           invert(void* in_array, dmDataType in_datatype, long nsets, long arrsize=1);
  short*          invert(short*  in_array, long nsets, long arrsize=1);
  long*           invert(long*   in_array, long nsets, long arrsize=1);
  float*          invert(float*  in_array, long nsets, long arrsize=1);
  double*         invert(double* in_array, long nsets, long arrsize=1);
  unsigned short* invert(unsigned short* in_array, long nsets, long arrsize=1);
  unsigned long*  invert(unsigned long*  in_array, long nsets, long arrsize=1);

  double** get_transform_matrix();
  void     set_transform_matrix(double** matrix);
  double** get_inverse_matrix();

protected:
  double** matrix;
  void calculate_transform_matrix();

private:
  void generate_parameter_list();


};


class SPLITTransform : public Transform
{

public:
  SPLITTransform();
  SPLITTransform(string in_name);
  SPLITTransform(string in_name, list<TransParam*> in_params);
  SPLITTransform(SPLITTransform &init);
  ~SPLITTransform();

  SPLITTransform* copy();

  list<TransParam*> get_parameter_list();

  void*           apply(void* in_array, dmDataType in_datatype, long nsets, long arrsize=1);
  short*          apply(short*  in_array, long nsets, long arrsize=1);
  long*           apply(long*   in_array, long nsets, long arrsize=1);
  float*          apply(float*  in_array, long nsets, long arrsize=1);
  double*         apply(double* in_array, long nsets, long arrsize=1);
  unsigned short* apply(unsigned short* in_array, long nsets, long arrsize=1);
  unsigned long*  apply(unsigned long*  in_array, long nsets, long arrsize=1);

private:
  void generate_parameter_list();

};


class LINEARTransform : public Transform
{

public:
  LINEARTransform();
  LINEARTransform(string in_name);
  LINEARTransform(string in_name, list<TransParam*> in_params);
  LINEARTransform(LINEARTransform &init);
  ~LINEARTransform();

  LINEARTransform* copy();

  list<TransParam*> get_parameter_list();

  void*           apply(void* in_array, dmDataType in_datatype, long nsets, long arrsize=1);
  short*          apply(short*  in_array, long nsets, long arrsize=1);
  long*           apply(long*   in_array, long nsets, long arrsize=1);
  float*          apply(float*  in_array, long nsets, long arrsize=1);
  double*         apply(double* in_array, long nsets, long arrsize=1);
  unsigned short* apply(unsigned short* in_array, long nsets, long arrsize=1);
  unsigned long*  apply(unsigned long*  in_array, long nsets, long arrsize=1);

  void*           invert(void* in_array, dmDataType in_datatype, long nsets, long arrsize=1);
  short*          invert(short*  in_array, long nsets, long arrsize=1);
  long*           invert(long*   in_array, long nsets, long arrsize=1);
  float*          invert(float*  in_array, long nsets, long arrsize=1);
  double*         invert(double* in_array, long nsets, long arrsize=1);
  unsigned short* invert(unsigned short* in_array, long nsets, long arrsize=1);
  unsigned long*  invert(unsigned long*  in_array, long nsets, long arrsize=1);


private:
  void generate_parameter_list();

};


#endif
