// 
//  Copyright (C) 2006-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//

//---------------------------------------------------------------------------
//
//  FILE NAME:          WCSTransform.cc
//
//  DEVELOPMENT:        CXCDS
//
//  DESCRIPTION:
//
//---------------------------------------------------------------------------

#include "Transform.hh"
#include "wcs.h"


// **************************************************************
//  Method:          WCSTransform
//
//  Description:     WCSTransform empty constructor
//
//  Arguments:       none
//
//  Returns:         Transform object with initialized values
//                   including Parameter list.
//
// **************************************************************
WCSTransform::WCSTransform() {
  clear();

  // Initialize Class Name string
  _strname = "WCSTransform";

  // Generate a new parameter list for this Transform object.
  generate_parameter_list();
}


// **************************************************************
//  Method:          WCSTransform
//
//  Description:     WCSTransform constructor
//
//  Arguments:       
//     string           - name of transform
//
//  Returns:         Transform object with initialized values
//                   including Parameter list.
//
// **************************************************************
WCSTransform::WCSTransform(string in_name)
{
  clear();

  // Initialize Class Name string
  _strname = "WCSTransform";

  // Initialize Transform Name
  name = in_name;

  // Generate a new parameter list for this Transform object.
  generate_parameter_list();

}


// **************************************************************
//  Method:          WCSTransform
//
//  Description:     WCSTransform constructor
//
//  Arguments:       
//     string           - name of transform
//     list<TransParam*> - list of transform parameters
//
//  Returns:         Transform object with initialized values
//                   including Parameter list.
//
// **************************************************************
WCSTransform::WCSTransform(string in_name, list<TransParam*> in_params)
{
  clear();

  // Initialize Class Name string
  _strname = "WCSTransform";

  // Initialize Transform Name
  name = in_name;

  // Generate a new parameter list for this Transform object.
  generate_parameter_list();

  // Initialize Parameter values from input list..
  set_parameter_list( in_params );

}


// **************************************************************
//  Method:          WCSTransform
//
//  Description:     WCSTransform copy constructor
//
//  Arguments: 
//     string           - name of transform
//     WCSTransform& - the WCSTransform object to copy
//
//  Returns:         none
//
// **************************************************************
WCSTransform::WCSTransform(WCSTransform &init) 
{
 
  if (this == &init) return;
  clear();

  // Initialize Class Name string
  _strname = "WCSTransform";

  // Initialize Transform Name
  name = init.get_name();

  // Initialize parameter list
  list<TransParam*>in_params = init.get_parameter_list();
  
  if (!in_params.empty()) {
    
    list<TransParam*>::iterator iter;
    for (iter = in_params.begin(); iter != in_params.end(); iter++) {
      TransParam* par_copy = new TransParam( *(*iter) );
      params.push_back(par_copy);
    }
    
  } else {
    // Generate a new parameter list for this Transform object.
    generate_parameter_list();
  }

}


// **************************************************************
//  Method:          ~WCSTransform
//
//  Description:     WCSTransform destructor.
//
//  Arguments:       none
//
//  Returns:         none
//
// **************************************************************
WCSTransform::~WCSTransform() 
{
  clear();
}


// **************************************************************
//  Method:          copy
//
//  Description:     Makes a copy of this WCSTransform and returns it.
//
//  Arguments:       none
//
//  Returns:         
//     WCSTransform* - a copy of this Transform
//
// **************************************************************
WCSTransform* WCSTransform::copy() 
{
  WCSTransform* trans = new WCSTransform( (WCSTransform&)*this );
  return trans;
}


// **************************************************************
//  Method:          get_transform_matrix
//
//  Description:     
//     Returns 3x3 Transform matrix.
//     Matrix is defined according to FITS standards w.r.t. 
//     application and orientation.   X' = X*[M] 
//     
//       M  = | CD0cos(rho)   -CD1sin(rho)   0.0  |
//            | CD0sin(rho)    CD1cos(rho)   0.0  |
//            | Tx             Ty            1.0  |
//
//       X  = [ x, y, 1]
//
//  Arguments:
//     const string& - transform name
//
//  Returns:
//     double** - the transform matrix
//
// **************************************************************
double** WCSTransform::get_transform_matrix()
{
  double** matrix = NULL;
  
  if ((matrix = new double* [3]) != NULL) 
    for (int rr = 3; rr--; )  matrix[rr] = new double[3];  
  
  double* crpix = (double*)get_parameter_value("CRPIX");
  double* crval = (double*)get_parameter_value("CRVAL");
  double* cdelt = (double*)get_parameter_value("CDELT");
  double* crota = (double*)get_parameter_value("CROTA");
  
  
  if ( matrix && crpix && crval && cdelt ) 
   {
     // initialize as unit matrix
     matrix[0][0] = 1.0;
     matrix[1][1] = 1.0;
     matrix[2][2] = 1.0; 
     
     matrix[0][1] = 0.0;
     matrix[0][2] = 0.0;
     matrix[1][0] = 0.0; 
     matrix[1][2] = 0.0;
     matrix[2][0] = 0.0;
     matrix[2][1] = 0.0;

     double rho = (*crota)*(M_PI/180.0);
     if ( fabs(rho) < FLT_EPSILON )
     {
       // no rotation
       matrix[0][0] = cdelt[0];
       matrix[2][0] = (-crpix[0] * cdelt[0]) + crval[0];
     
       if ( (get_parameter("CRPIX")->get_nvals() == 2) &&
	    (get_parameter("CRVAL")->get_nvals() == 2) &&
	    (get_parameter("CDELT")->get_nvals() == 2) )
       {
	 matrix[1][1] = cdelt[1];
	 matrix[2][1] = (-crpix[1] * cdelt[1]) + crval[1];
       }

     }
     else
     {
       // include rotation components
	  // Result of Mscale * Mrot
	  matrix[0][0] = cdelt[0] * cos(rho);
	  matrix[1][0] = cdelt[0] * sin(rho);
	  matrix[2][0] = crval[0] - 
                         cdelt[0]*crpix[0]*cos(rho) - 
                         cdelt[0]*crpix[1]*sin(rho);
	  matrix[0][1] = -cdelt[1] * sin(rho);
	  matrix[1][1] =  cdelt[1] * cos(rho);
	  matrix[2][1] = crval[1] + 
                         cdelt[1]*crpix[0]*sin(rho) - 
                         cdelt[1]*crpix[1]*cos(rho);
     }
     
   }
   
   return matrix; 
}

// **************************************************************
//  Method:          set_transform_matrix
//
//  Description:     
//     Determines and assigns values to Transform parameters
//     from the input 3x3 Transform matrix.
//
//     Matrix is defined according to FITS standards w.r.t. 
//     application and orientation.   X' = X*[M] 
//     
//       M  = | CD0cos(rho)   -CD1sin(rho)   0.0  |
//            | CD0sin(rho)    CD1cos(rho)   0.0  |
//            | Tx             Ty            1.0  |
//
//  Arguments:
//     double** - the transform matrix
//
//  Returns:
//
// **************************************************************
void WCSTransform::set_transform_matrix(double** in_matrix) 
{
  if (in_matrix == NULL) return;

  double** matrix = in_matrix;

  double rho_a, rho_b;
  double tmpcrota, coscrota, sincrota;

  double crota = 0;
  double crpix[2];
  double crval[2];
  double cdelt[2];

  if (matrix[1][0] == 0)
    rho_a = 0.0;
  else if (matrix[1][0] > 0.0)
    rho_a = atan(matrix[1][0] / matrix[0][0]);
  else
    rho_a = atan(-matrix[1][0] / -matrix[0][0]);

  if (matrix[0][1] == 0)
    rho_b = 0.0;
  else if (matrix[0][1] > 0)
    rho_b = atan(matrix[0][1] / -matrix[1][1]);
  else
    rho_b = atan(-matrix[0][1] / matrix[1][1]);

  tmpcrota = (rho_a + rho_b) / 2.0;
  coscrota = cos(tmpcrota);
  sincrota = sin(tmpcrota);

  cdelt[0] = 1.0;
  cdelt[1] = 1.0;
  
  cdelt[0] = sqrt(matrix[0][0]*matrix[0][0] + matrix[1][0]*matrix[1][0]);
  cdelt[1] = sqrt(matrix[0][1]*matrix[0][1] + matrix[1][1]*matrix[1][1]);

  if (coscrota > 0.0) {
    if (matrix[0][0]*coscrota < 0.0)
      cdelt[0] = cdelt[0] * -1.0;
    if (matrix[1][1]*coscrota < 0.0)
      cdelt[1] = cdelt[1] * -1.0;
  } else {
    if (matrix[0][1]*sincrota < 0.0)
      cdelt[0] = cdelt[0] * -1.0;
    if (-matrix[1][0] / sincrota < 0.0)
      cdelt[1] = cdelt[1] * -1.0;
  }


  // CRPIX & CRVAL cannot be distinguished by the matrix alone,
  // check for one or the other being set.. if not, then use 
  // a default CRPIX = ( 0.5, 0.5 )
  //
  double* tmpval = NULL;
  tmpval = (double*)get_parameter("CRVAL")->get_value();
  if (tmpval) 
  {
    crval[0] = tmpval[0];
    crval[1] = tmpval[1];

    crpix[0] = 1.0;
    crpix[1] = 1.0;
    crpix[0] = ((crval[0]-matrix[2][0])*coscrota)/cdelt[0] - ((crval[1]-matrix[2][1])*sincrota)/cdelt[1];
    crpix[1] = ((crval[1]-matrix[2][1])*coscrota)/cdelt[1] + ((crval[0]-matrix[2][0])*sincrota)/cdelt[0];
  }
  else
  {
    tmpval = NULL;
    tmpval = (double*)get_parameter("CRPIX")->get_value();
    if (!tmpval )
    {
      crpix[0] = 0.5;
      crpix[1] = 0.5;
    }
    else
    {
      crpix[0] = tmpval[0];
      crpix[1] = tmpval[1];
    }
    crval[0] = matrix[2][0] + (crpix[0]*coscrota + crpix[1]*sincrota)*cdelt[0];
    crval[1] = matrix[2][1] + (crpix[1]*coscrota - crpix[0]*sincrota)*cdelt[1];
  }

  crota = (tmpcrota*180.0)/M_PI;

  get_parameter("CRPIX")->set_value(&crpix, dmDOUBLE, 2);
  get_parameter("CRVAL")->set_value(&crval, dmDOUBLE, 2);
  get_parameter("CDELT")->set_value(&cdelt, dmDOUBLE, 2);
  get_parameter("CROTA")->set_value(&crota, dmDOUBLE, 1);

}


// **************************************************************
//  Method:          apply
//
//  Description:     pix to world transform
//
//  Arguments:
//     void*            - interleaved data array
//     dmDataType       - array datatype
//     long             - number of value pairs.
//
//  Returns:
//     void*            - the transformed data array
//
// **************************************************************
void* WCSTransform::apply(void* in_array, dmDataType in_datatype, long nsets, long arrsize) 
{
  void*   optr = NULL;            // Pointer to output data

  switch (in_datatype) 
  {
    case dmDOUBLE:
      optr = (void*) apply((double*)in_array, nsets, arrsize );
      break;
    case dmSHORT: 
      optr = (void*) apply((short*)in_array, nsets, arrsize );
      break;
    case dmLONG:
      optr = (void*) apply((long*)in_array, nsets, arrsize );
      break;
    case dmFLOAT:
      optr = (void*) apply((float*)in_array, nsets, arrsize );
      break;
    case dmUSHORT:
      optr = (void*) apply((unsigned short*)in_array, nsets, arrsize );
      break;
    case dmULONG:
      optr = (void*) apply((unsigned long*)in_array, nsets, arrsize );
      break;
    case dmBOOL:
    case dmBIT:
    case dmBYTE:
    case dmTEXT:
    case dmUNKNOWNTYPE:
    default:
      optr = NULL;
      break;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     pix to world transform
//
//  Arguments:
//     short*           - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     short*           - the transformed data array
//
// **************************************************************
short* WCSTransform::apply(short* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  short*  optr    = NULL;         // Pointer to output data
  dmDataType intype = dmSHORT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new short[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     pix to world transform
//
//  Arguments:
//     long*            - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     long*            - the transformed data array
//
// **************************************************************
long* WCSTransform::apply(long* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  long*   optr    = NULL;         // Pointer to output data
  dmDataType intype = dmLONG;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new long[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     pix to world transform
//
//  Arguments:
//     float*           - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     float*           - the transformed data array
//
// **************************************************************
float* WCSTransform::apply(float* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  float*   optr    = NULL;        // Pointer to output data
  dmDataType intype = dmFLOAT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new float[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     pix to world transform
//
//  Arguments:
//     unsigned short*  - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     unsigned short*  - the transformed data array
//
// **************************************************************
unsigned short* WCSTransform::apply(unsigned short* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  unsigned short* optr = NULL;    // Pointer to output data
  dmDataType intype = dmUSHORT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new unsigned short[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     pix to world transform
//
//  Arguments:
//     unsigned long*   - interleaved data array
//     long             - number of value pairs.
//
//  Returns:
//     unsigned long*   - the transformed data array
//
// **************************************************************
unsigned long* WCSTransform::apply(unsigned long* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  unsigned long* optr = NULL;     // Pointer to output data
  dmDataType intype = dmULONG;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = apply( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new unsigned long[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          apply
//
//  Description:     pix to world transform
//
//  Arguments:
//     double*          - input data array
//     long             - number of value pairs.
//
//  Returns:
//     double*          - the transformed data array
//
// **************************************************************
double* WCSTransform::apply(double* inarr, long nsets, long arrsize) 
{

  double* outarr = NULL;          // output data array.
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.
  char   ct1[80] = "";
  char   ct2[80] = "";

  // Extract Transform paramater info for WCSLIB call.
  // NOTE: Parameter class should ensure datatype integrity
  //       so we should be OK assuming the types are unchanged
  //       from the original settings.

  double* crpix = (double*)get_parameter_value("CRPIX");
  double* crval = (double*)get_parameter_value("CRVAL");
  double* cdelt = (double*)get_parameter_value("CDELT");
  double* crota = (double*)get_parameter_value("CROTA");
  short* equinox = (short*)get_parameter_value("EQUINOX");
  double* epoch = (double*)get_parameter_value("EPOCH");
  string* tctype= (string*)get_parameter_value("CTYPE");

  if ( !crpix || !crval || !cdelt || !crota || !equinox || !epoch )
    return NULL;

  if ( tctype == NULL ) {   // set default
    strcpy(ct1, "RA---TAN");
    strcpy(ct2, "DEC--TAN");
  } else {
    strcpy(ct1, (char*)tctype[0].c_str());
    strcpy(ct2, (char*)tctype[1].c_str());
  }

  outarr = new double[nvals];
  if ( !outarr ){ return NULL; }
 
  // Initialize WCS
  WorldCoor* wcs = wcskinit( 2,           // number of values per set
			     nsets,       // number of value sets.
			     ct1,         // Tranform flavor
			     ct2,         // Tranform flavor
			     crpix[0],
			     crpix[1],
			     crval[0],
			     crval[1],
			     NULL,
			     cdelt[0],
			     cdelt[1],
			     *crota,
			     *equinox,
			     *epoch
			     );


  if (( tctype != NULL ) && ( tctype[0] == "LINEAR" ))
  {
    // The above will initialize wcs->sysout to "FK5" which will then cause
    // pix2wcs() to bound check the xpos output value and adjust it by +/-
    // 360.0 if the value is either positive or negative.  To avoid this,
    // the sysout value must be 'WCS_LINEAR'
    // This behaviour change occurs for WCSSUBS-3.7.0.. probably since 3.5.7
    // I do not see an init routine that will not do this..
    wcs->sysout = WCS_LINEAR;
  }
  else if (( tctype != NULL ) && ( tctype[0] == "LONG-TAN" ))
  {
    // LONGPOLE and LATPOLE should be populated from the header keywords
    // LONP# and LATP#, where # is associated with the transform.
    // Hardcode for now, but should modify to do better (add parameters?)
    wcs->longpole = 270.;
    wcs->latpole  =   0.;
    if ( wcs->longpole != 0.0 )
      wcs->cel.ref[2] = wcs->longpole;
    if ( wcs->latpole != 0.0 )
      wcs->cel.ref[3] = wcs->latpole;
  }  

  
  long ndx = 0;
  long ndy = 0;

  for (int ii = 0; ii < nsets; ii++)
  {
    for (int jj = 0; jj < arrsize; jj++)
    {
      ndx = jj + (ii*2*arrsize);
      ndy = jj + (ii*2*arrsize) + arrsize;
	  
      // Convert to world coords 
      pix2wcs(wcs, inarr[ndx], inarr[ndy], &outarr[ndx], &outarr[ndy]);      
    }    
  }
  wcsfree(wcs);

  return outarr;
}


// **************************************************************
//  Method:          invert
//
//  Description:     world to pix transform
//
//  Arguments:       
//     void*            - interleaved data array
//     dmDataType       - array datatype
//
//  Returns:
//     void*            - the inverted data
//
// **************************************************************
void* WCSTransform::invert(void* in_array, dmDataType in_datatype, long nsets, long arrsize)
{

  void*   optr = NULL;            // Pointer to output data

  switch (in_datatype) 
  {
    case dmDOUBLE:
      optr = (void*) invert((double*)in_array, nsets, arrsize );
      break;
    case dmSHORT: 
      optr = (void*) invert((short*)in_array, nsets, arrsize );
      break;
    case dmLONG:
      optr = (void*) invert((long*)in_array, nsets, arrsize );
      break;
    case dmFLOAT:
      optr = (void*) invert((float*)in_array, nsets, arrsize );
      break;
    case dmUSHORT:
      optr = (void*) invert((unsigned short*)in_array, nsets, arrsize );
      break;
    case dmULONG:
      optr = (void*) invert((unsigned long*)in_array, nsets, arrsize );
      break;
    case dmBOOL:
    case dmBIT:
    case dmBYTE:
    case dmTEXT:
    case dmUNKNOWNTYPE:
    default:
      optr = NULL;
      break;
  }

  return optr;


  return NULL;
}

// **************************************************************
//  Method:          invert
//
//  Description:     world to pix transform
//
//  Arguments:       
//     short*           - input data array
//     long             - number of value pairs.
//
//  Returns:
//     short*           - the transformed data
//
// **************************************************************
short* WCSTransform::invert(short* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  short*  optr    = NULL;         // Pointer to output data
  dmDataType intype = dmSHORT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = invert( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new short[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          invert
//
//  Description:     world to pix transform
//
//  Arguments:       
//     long*            - input data array
//     long             - number of value pairs.
//
//  Returns:
//     long*            - the transformed data
//
// **************************************************************
long* WCSTransform::invert(long* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  long*  optr     = NULL;         // Pointer to output data
  dmDataType intype = dmLONG;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = invert( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new long[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          invert
//
//  Description:     world to pix transform
//
//  Arguments:       
//     float*           - input data array
//     long             - number of value pairs.
//
//  Returns:
//     float*           - the transformed data
//
// **************************************************************
float* WCSTransform::invert(float* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  float*  optr    = NULL;         // Pointer to output data
  dmDataType intype = dmFLOAT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = invert( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new float[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          invert
//
//  Description:     world to pix transform
//
//  Arguments:       
//     unsigned short*  - input data array
//     long             - number of value pairs.
//
//  Returns:
//     unsigned short*  - the transformed data
//
// **************************************************************
unsigned short* WCSTransform::invert(unsigned short* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  unsigned short* optr = NULL;    // Pointer to output data
  dmDataType intype = dmUSHORT;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = invert( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new unsigned short[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          invert
//
//  Description:     world to pix transform
//
//  Arguments:       
//     unsigned long*   - input data array
//     long             - number of value pairs.
//
//  Returns:
//     unsigned long*   - the transformed data
//
// **************************************************************
unsigned long* WCSTransform::invert(unsigned long* inarr, long nsets, long arrsize) 
{
  double* indata  = NULL;         // Data array to send into wcs
  double* outdata = NULL;         // Data array output from wcs
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.

  unsigned long* optr = NULL;    // Pointer to output data
  dmDataType intype = dmULONG;

  // Allocate memory for input values as double for WCS
  indata = new double[nvals];

  // Convert input array to double.. as required.
  cast_array_numeric( inarr, intype, dmDOUBLE, (void*)indata, nvals );

  // Apply transform to double array..
  outdata = invert( indata, nsets, arrsize );

  // Free up memory for input data as double..
  delete [] indata;

  // set pointer to output values.
  optr = new unsigned long[nvals];

  if ( optr != NULL )
  {
    cast_array_numeric( (void*)outdata, dmDOUBLE, intype, (void*)optr, nvals );
    delete [] outdata;
  }

  return optr;
}

// **************************************************************
//  Method:          invert
//
//  Description:     world to pix transform
//
//  Arguments:       
//     double*          - input data array
//     long             - number of value pairs.
//
//  Returns:
//     double*          - the transformed data
//
// **************************************************************
double* WCSTransform::invert(double* inarr, long nsets, long arrsize)
{

  double* outarr = NULL;          // output data array.
  long    nvals   = nsets * 2 * arrsize;    // Total #vals at pointer.
  char   ct1[80] = "";
  char   ct2[80] = "";


  // Extract Transform paramater info for WCSLIB call.
  // NOTE: Parameter class should ensure datatype integrity
  //       so we should be OK assuming the types are unchanged
  //       from the original settings.
  double* crpix = (double*)get_parameter_value("CRPIX");
  double* crval = (double*)get_parameter_value("CRVAL");
  double* cdelt = (double*)get_parameter_value("CDELT");
  double* crota = (double*)get_parameter_value("CROTA");
  short* equinox = (short*)get_parameter_value("EQUINOX");
  double* epoch = (double*)get_parameter_value("EPOCH");
  string* tctype= (string*)get_parameter_value("CTYPE");

  if ( !crpix || !crval || !cdelt || !crota || !equinox || !epoch )
    return NULL;

  outarr = new double[nvals];
  if ( !outarr ){ return NULL; }

  if ( tctype == NULL ) {   // set default
    strcpy(ct1, "RA---TAN");
    strcpy(ct2, "DEC--TAN");
  } else {
    strcpy(ct1, (char*)tctype[0].c_str());
    strcpy(ct2, (char*)tctype[1].c_str());
  }

  // Initialize WCS
  WorldCoor* wcs = wcskinit( 2,           // number of values per set
			     nsets,       // number of value sets.
			     ct1,         // Tranform flavor
			     ct2,         // Tranform flavor
			     crpix[0],
			     crpix[1],
			     crval[0],
			     crval[1],
			     NULL,
			     cdelt[0],
			     cdelt[1],
			     *crota,
			     *equinox,
			     *epoch
			     );
  

  int scale; 
  long ndx = 0;
  long ndy = 0;

  for (int ii = 0; ii < nsets; ii++) 
  {
    for (int jj = 0; jj < arrsize; jj++)
    {
      ndx = jj + (ii*2*arrsize);
      ndy = jj + (ii*2*arrsize) + arrsize;
	  
      // Convert to pixel coords 
      wcs2pix(wcs, inarr[ndx], inarr[ndy], &outarr[ndx], &outarr[ndy], &scale);
    }
  }

  wcsfree(wcs);

  return outarr;
}


// **************************************************************
//  Method:          get_parameter_list
//
//  Description:     Returns the list of transform parameters.
//                   If list is empty, will generate a generic
//                   list of parameters to return.
//
//  Arguments:       none
//
//  Returns:
//     list<TransParam*> - list of transform parameters
//
// **************************************************************
list<TransParam*> WCSTransform::get_parameter_list()
{
  if ( params.empty() ) 
    generate_parameter_list();

  return params;
}

// **************************************************************
//  Method:          generate_parameter_list
//
//  Description:     Creates a list of parameters needed
//                   for the WCS Transform.
//                   User must fill in value.
//
//  Arguments:       none
//
//  Returns:
//     Loads Transform params attribute.
//
// **************************************************************
void WCSTransform::generate_parameter_list()
{
  short    equinox = 2000;
  double   epoch   = 2000.0;
  double   crota   = 0.0;
  TransParam*  ptr = NULL;

  ptr = new TransParam("CRPIX", NULL, dmDOUBLE, 
		      "Reference Point Pixel Coordinates", name, 2);
  params.push_back( ptr );

  ptr = new TransParam("CRVAL", NULL, dmDOUBLE, 
		      "Center Coordinate in decimal degrees", name, 2);
  params.push_back( ptr );

  ptr = new TransParam("CDELT", NULL, dmDOUBLE, 
		      "Transform Scale in degrees/pixel", name, 2);
  params.push_back( ptr );

  ptr = new TransParam("CROTA", &crota, dmDOUBLE, 
		      "Rotation Angle", name, 1);
  params.push_back( ptr );

  ptr = new TransParam("EQUINOX", (void*)&equinox , dmSHORT, 
		      "Equinox of Coordinates", name, 1);
  params.push_back( ptr );

  ptr = new TransParam("EPOCH", (void*)&epoch, dmDOUBLE, 
		      "Epoch of Coordinates", name, 1);
  params.push_back( ptr );

  ptr = new TransParam("CTYPE", NULL, dmTEXT, 
		      "Coordinate Projection Type", name, 2);
  params.push_back( ptr );


}
