/* ************************************************************** */
/*  File: pytransform.i                                           */
/*                                                                */
/*  Description:                                                  */
/*    Template file for Python wrapper generation                 */
/*    of Transform library.                                       */
/*                                                                */
/* ************************************************************** */
%module pytransform
%include "std_string.i"
%include "std_list.i"
%include "typemaps.i"
%{
  #include <numpy/noprefix.h>
  #include "Transform.hh"
  #include "TransParam.hh"

  static int get_array_type(void* ptr)   { return NPY_VOID; }
  static int get_array_type(short* ptr)  { return NPY_SHORT;  }
  static int get_array_type(long* ptr)   { return NPY_LONG;   }
  static int get_array_type(float* ptr)  { return NPY_FLOAT;  }
  static int get_array_type(double* ptr) { return NPY_DOUBLE; }
  static int get_array_type(string* ptr) { return NPY_STRING; }
%}


/* Rename all print() functions to show() */
%rename(show) print;
 

/* ====================================================================== */
/*  Ownership Management                                                  */
/* ====================================================================== */
/* Transfer ownership of memory for new allocations */
/* %newobject <method name>;*/
%newobject apply;
%newobject invert;
%newobject get_transform_matrix;

/* Release ownership of memory for assignment of pointers */
/* %delobject <method name>;*/
%delobject get_parameter_value;
%delobject get_value;

/* Wrapper for generic apply covers the datatypes */
%ignore *::apply(void* in_array, dmDataType in_datatype, long nsets);
%ignore *::apply(double* in_array, long nsets, long arrsize=1);
%ignore *::apply(float* in_array, long nsets, long arrsize=1);
%ignore *::apply(long* in_array, long nsets, long arrsize=1);
%ignore *::apply(short* in_array, long nsets, long arrsize=1);
%ignore *::apply(unsigned long* in_array, long nsets, long arrsize=1);
%ignore *::apply(unsigned short* in_array, long nsets, long arrsize=1);
%ignore *::invert(void* in_array, dmDataType in_datatype, long nsets);
%ignore *::invert(double* in_array, long nsets, long arrsize=1);
%ignore *::invert(float* in_array, long nsets, long arrsize=1);
%ignore *::invert(long* in_array, long nsets, long arrsize=1);
%ignore *::invert(short* in_array, long nsets, long arrsize=1);
%ignore *::invert(unsigned long* in_array, long nsets, long arrsize=1);
%ignore *::invert(unsigned short* in_array, long nsets, long arrsize=1);


/* Typecheck maps.. havent quite figured these out */
%typemap(typecheck,precedence=SWIG_TYPECHECK_VOIDPTR) void *
{
  $1 = 1;  /* Pass it along */
}
%typemap(typecheck,precedence=SWIG_TYPECHECK_POINTER) list<TransParam*>
{
  $1 = ( PyList_Check($input) ) ? 1 : 0;
}


/* ====================================================================== */
/*  Input typemap for set_value (TRANSPARAM)                              */
/* ====================================================================== */
%typemap(in) (void* in_value, dmDataType in_datatype, long in_nvals=1){

  PyObject *inarr = NULL;
  $3 = 0;
  $2 = dmUNKNOWNTYPE;
  $1 = NULL;

  if ($input == NULL ){ SWIG_fail; }

  // Convert input object to an array
  inarr = PyArray_FromAny( $input, NULL, 0, MAX_DIMS, NPY_CARRAY , NULL);
  if ( inarr != NULL )
  {
    $2 = get_dmType( PyArray_TYPE(inarr) );
    $3 = PyArray_SIZE(inarr);

    if ( $2 == dmTEXT )
    {
       //Pull out the null terminated string.
       char *tptr  = (char*)PyArray_DATA( inarr );
       int itemsize = PyArray_ITEMSIZE(inarr);

       int nsets = $3;

       if ( nsets > 1) {
           char **tmp  = (char**)calloc( nsets, sizeof(char*) );
           char *buf   = (char*)calloc(nsets*(itemsize+1), sizeof(unsigned char) );

           tmp[0] = buf;
           strncpy(tmp[0], tptr, itemsize);
	   
           for (int ii=1; ii < nsets; ii++)
           {
               tptr += itemsize;
               tmp[ii] = tmp[ii-1] + (itemsize+1);
               strncpy(tmp[ii], tptr, itemsize);
           }

           $1 = (void*)tmp;     

       } else if ( nsets == 1 ) {

           // Pull out the null terminated string.
           char* keytmp = (char*) realloc(tptr, itemsize+1);
           keytmp[itemsize] = '\0';

           $1 = (void*)keytmp;
       }

    }
    else
    {
       // Set data directly 
       $1 = (void*)PyArray_DATA( inarr ) ;
    }
  }
}

/* ====================================================================== */
/*  Input typemap for set_transform_matrix (TRANSFORM)                    */
/* ====================================================================== */
%typemap(in) (double **matrix) {
  double ** mat = (double **) malloc(3*sizeof(double*));
  double* array_buffer = (double *)PyArray_DATA($input);

  int ii,jj;  
  for ( ii=0; ii<3; ii++) {
    mat[ii] = (double*) malloc(3*sizeof(double));  

    for ( jj=0; jj<3; jj++) {
      mat[ii][jj] = array_buffer[jj+(ii*3)];
    }
  }

  $1 = mat;
}


/* ====================================================================== */
/*  Output typemap for Parameter values (TRANSPARAM)                      */
/* ====================================================================== */
%typemap(out) void* get_value{
  if (NULL == $1) {
    Py_INCREF(Py_None);
    $result = Py_None;
  }
  else
  {
    void* dptr  = $1;
    PyArrayObject *tmparr = 0;
    int   pytype;
    int   nvals;
    int   ndim;
    intp  dims[1];
    int   itemsize = 0;  /* size of each value in Bytes, 0=default  */
                         /* non-zero for strings = length           */

    nvals   = arg1->get_nvals();
    ( nvals > 1 )?( ndim = 1 ):( ndim = 0 );
    dims[0] = nvals;

    pytype  = get_pyType( arg1->get_datatype() );

    if ( pytype == NPY_STRING)
    {     
        itemsize = strlen(((char**)$1)[0]) + 1;  

	char* buf = (char*)calloc( nvals * (itemsize), sizeof(char));
	char** tmpdata = (char**)calloc(nvals, sizeof(char*));

	tmpdata[0] = buf; 
    	strncpy(tmpdata[0], ((char**)$1)[0], itemsize);

    	for (int ii=1; ii < nvals; ii++){
      	  tmpdata[ii] = tmpdata[ii-1] + itemsize;
      	  strncpy(tmpdata[ii], ((char**)$1)[ii], itemsize);
    	}

        dptr = (void*)buf;
    }

    tmparr = (PyArrayObject*) PyArray_New(&PyArray_Type, ndim, dims, pytype, NULL, dptr, itemsize, NPY_CARRAY, NULL );

    if (NULL == tmparr) SWIG_fail;
    $result = PyArray_Return( tmparr );

  }
}

/* ====================================================================== */
/*  Output typemap for Parameter value (TRANSFORM)                        */
/* ====================================================================== */
%typemap(out) void* get_parameter_value{
  if (NULL == $1) {
    Py_INCREF(Py_None);
    $result = Py_None;
  }
  else
  {
    void* dptr  = $1;
    PyArrayObject *tmparr  = 0;
    TransParam    *myParam = arg1->get_parameter(arg2);
    int   pytype;
    int   nvals;
    int   ndim;
    intp  dims[1];
    int   itemsize = 0;  /* size of each value in Bytes, 0=default  */
                         /* non-zero for strings = length           */

    if ( myParam != NULL )
    {
      nvals   = myParam->get_nvals();
      ( nvals > 1 )?( ndim = 1 ):( ndim = 0 );
      dims[0] = nvals;
      pytype  = get_pyType( myParam->get_datatype() );
 
      if ( pytype == NPY_STRING)
      {     
        itemsize = strlen(((char**)$1)[0]) + 1;  

	char* buf = (char*)calloc( nvals * (itemsize), sizeof(char));
	char** tmpdata = (char**)calloc(nvals, sizeof(char*));

	tmpdata[0] = buf; 
    	strncpy(tmpdata[0], ((char**)$1)[0], itemsize);

    	for (int ii=1; ii < nvals; ii++){
      	  tmpdata[ii] = tmpdata[ii-1] + itemsize;
      	  strncpy(tmpdata[ii], ((char**)$1)[ii], itemsize);
    	}

        dptr = (void*)buf;
      }
    }

    tmparr = (PyArrayObject*) PyArray_New(&PyArray_Type, ndim, dims, pytype, NULL, dptr, itemsize, NPY_CARRAY, NULL );

    if (NULL == tmparr) SWIG_fail;
    $result = PyArray_Return( tmparr );

  }
}

/* ====================================================================== */
/*  Input typemap for apply method (TRANSFORM)                            */
/* ====================================================================== */
%typemap(in) (void* in_array, dmDataType in_datatype, long nsets, long arrsize=1) {

  PyObject *inarr = NULL;
  PyObject *inarr_flat = NULL;

  $4 = 1;
  $3 = 1;
  $2 = dmUNKNOWNTYPE;
  $1 = NULL;

  if ($input == NULL ){ SWIG_fail; }


  // Convert input object to an array
  inarr = PyArray_FromAny( $input, NULL, 1, MAX_DIMS, NPY_CARRAY, NULL);

  if ( inarr != NULL )
  {
    long ndim = PyArray_NDIM( inarr );

    if ( (( arg1->get_className().compare(string("LINEAR2DTransform")) == 0) || ( arg1->get_className().compare(string("WCSTANTransform")) == 0) || ( arg1->get_className().compare(string("WCSTransform")) == 0) )&& (ndim < 2) ) {
      PyErr_SetString(PyExc_ValueError,"Input must be a 2D array.");
      return NULL;
    }

    if (ndim > 2)
        $4 = (long) PyArray_DIM( inarr, 2 );
    else if ( ( arg1->get_className().compare(string("LINEARTransform")) == 0) && (ndim == 2) )
        $4 = (long) PyArray_DIM( inarr, 1 );

    $3 = (long) PyArray_DIM( inarr, 0 );  // inarr = [nsets, nvalsPerSet, arrSize ]
    $2 = get_dmType( PyArray_TYPE(inarr) );

    // make a 1D copy of the input array so that data format is
    // compatible with what the Transform.apply() function is expecting
    inarr_flat = (PyObject*)PyArray_Flatten( (PyArrayObject*)inarr, (NPY_ORDER)0 );
    $1 = (void*)PyArray_DATA( inarr_flat ) ;

  } else {
    
    PyErr_SetString(PyExc_ValueError,"Input must be an array.");
    return NULL;
  }

}

/* ====================================================================== */
/*  output typemap for apply method (TRANSFORM)                           */
/* ====================================================================== */
%typemap(out) void* apply{

  if (NULL == $1) {
    printf(">>> PyArray is NONE\n");

    Py_INCREF(Py_None);
    $result = Py_None;
  }
  else
  {

    PyArrayObject *tmparr  = 0;
    int   ndim;
    intp* dims;
    int   itemsize = 0;  /* size of each value in Bytes, 0=default  */
                         /* non-zero for strings = length           */

    int cols   = 2;                 /* Need something better here */

    if (( arg1->get_className().compare(string("SPLITTransform")) == 0)  || ( arg1->get_className().compare(string("LINEARTransform")) == 0) )
      cols   = 1;

    int nsets  = arg4;              /* Use variable input to apply() */
    int pytype = get_pyType(arg3);  /* Use variable input to apply() */
    
    int arrsize = arg5;

    if (arrsize > 1) 
    {
      if (cols == 1 ) 
      {
        ndim = 2;                     /* 2D array [nsets,arrsize] */
        dims = new intp[ndim];
        dims[0] = nsets;
        dims[1] = arrsize;
      }
      else
      {
        ndim = 3;                     /* 3D array [nsets,ncpts,arrsize] */
        dims = new intp[ndim];
        dims[0] = nsets;
        dims[1] = cols;
        dims[2] = arrsize;
      }

    } 
    else 
    {
      if ( cols > 1 )
      {
        ndim = 2;                     /* 2D array [nsets,cols] */
        dims = new intp[ndim];
        dims[0] = nsets;
        dims[1] = cols;

      } 
      else
      {
	ndim = 1;                   /* 1D array [nsets] */
	dims = new intp[ndim];
	dims[0] = nsets;
      }
    }
    tmparr = (PyArrayObject*) PyArray_New(&PyArray_Type, ndim, dims, pytype, NULL, $1, itemsize, NPY_CARRAY, NULL );


    if (NULL == tmparr) SWIG_fail;
    $result = PyArray_Return( tmparr );
  }
}
%apply void* apply {
  void* invert
};

/* ====================================================================== */
/*  Input typemap for Parameter List (TRANSFORM)                          */
/* ====================================================================== */
%typemap(in) std::list<TransParam*> in_params{

  PyObject* pyparam = 0;
  void *tmp2 = 0 ;
  int   res2 = 0 ;
  int     ii = 0;

  TransParam* item = NULL;
  list< TransParam* > inpars;

  int nelem = PyList_Size( $input );

  if ( nelem == 0 ){ SWIG_fail; }
  for ( ii = 0; ii<nelem; ii++ )
  {
    pyparam = PyList_GetItem( $input, ii );
    res2 = SWIG_ConvertPtr( pyparam, &tmp2,SWIGTYPE_p_TransParam, SWIG_POINTER_DISOWN |  0 );
    if ( !SWIG_IsOK(res2)){ SWIG_fail; }

    item = reinterpret_cast<TransParam * >(tmp2);
    inpars.push_back( item );    
  }
  $1 = inpars;
}

/* ====================================================================== */
/*  Output typemap for Parameter List (TRANSFORM)                         */
/* ====================================================================== */
%typemap(out) list<TransParam*> get_parameter_list{
   PyObject* outlist = 0;
   PyObject* pyparam = 0;

   if ( $1.empty() )
   {
     Py_INCREF(Py_None);
     $result = Py_None;
   }
   else
   {
     list< TransParam* >::iterator iter; 
     TransParam* item = NULL;
     int nelem = $1.size();
     int ii    = 0;

     // Generate Python list object
     outlist = PyList_New( nelem );

     // Generate Python object for each Parameter and add to the output list
     for (iter = $1.begin(); iter != $1.end(); iter++) 
     {
       item = (*iter);
       if ( item )
       {
         pyparam = SWIG_NewPointerObj(SWIG_as_voidptr(item), SWIGTYPE_p_TransParam, 0 |  0 );
         PyList_SetItem( outlist, ii, pyparam );
         ii++;
       }
     }

     if (NULL == outlist) SWIG_fail;
     $result = (PyObject *)outlist;
   }
}




/* ====================================================================== */
/* Typemap for get_transform_matrix method..                              */
/*   Hard coded as a 3x3 array                                            */
/* ====================================================================== */
%typemap(out) double** WCSTransform::get_transform_matrix {
  if (NULL == $1) {
    Py_INCREF(Py_None);
    $result = Py_None;
  } else {
    intp dims[2];
    dims[0] = intp(3);
    dims[1] = intp(3);

    int ii,jj;

    /* Need to ensure that the matrix is stored in a  contiguous  */
    /* chunk of memory.. which it currently is not..              */

    double * tarr = new double[9];
    for ( ii=0; ii<3; ii++)
       for ( jj=0; jj<3; jj++)
          tarr[jj+ii*3] = $1[ii][jj];

    /* Free the memory allocated in the C library                 */
    for ( ii=0; ii<3; ii++)
       delete []$1[ii];
    delete []$1;
    $1 = NULL;

    /* Convert to a python array object                           */
    if (NULL == ($result =
                 PyArray_SimpleNewFromData(2, dims, get_array_type(tarr), tarr)))
    {
      delete [] tarr;
      return NULL;
    }
  }
}

%apply double** WCSTransform::get_transform_matrix  {
  double** WCSTANTransform::get_transform_matrix 
};


%typemap(out) double** Transform::get_transform_matrix {
  if (NULL == $1) {
    Py_INCREF(Py_None);
    $result = Py_None;
  } else {
    intp dims[2];
    dims[0] = intp(3);
    dims[1] = intp(3);

    int ii,jj;

    /* Need to ensure that the matrix is stored in a  contiguous  */
    /* chunk of memory.. which it currently is not..              */

    double * tarr = new double[9];
    for ( ii=0; ii<3; ii++)
       for ( jj=0; jj<3; jj++)
          tarr[jj+ii*3] = $1[ii][jj];

    /* DO NOT Free the memory allocated in the C library          */
    /* Implemented the way WCSTAN will migrate to too.. keeps an  */
    /* internal pointer to the matrix                             */


    /* TEMPORARY - should be removed when 			  */
    /*             WCSTransform::get_transform_matrix()	  */
    /*             typemap(out) is removed			  */

    /* For WCSTANTransform when stored as a generic Transform     */
    if ((arg1->get_className() == "WCSTANTransform")  || (arg1->get_className() == "WCSTransform")){
      /* Free the memory allocated in the C library               */
      for ( ii=0; ii<3; ii++)
         delete []$1[ii];
      delete []$1;
      $1 = NULL;
    }

    /* Convert to a python array object                           */
    if (NULL == ($result =
                 PyArray_SimpleNewFromData(2, dims, get_array_type(tarr), tarr)))
    {
      delete [] tarr;
      return NULL;
    }
  }
}
%apply double** Transform::get_transform_matrix{
  double** LINEAR2DTransform::get_transform_matrix
};




/* List functions to ignore in Python */
/* Have a constructor defined in Python layer TransParam( name, data, descr. ) */
/* which is equivalent to all of these                                        */
%ignore TransParam(string in_name, void* in_value, dmDataType in_datatype, 
	            string in_desc, long in_nvals=1);
%ignore TransParam(string in_name, void* in_value, dmDataType in_datatype,
	            string in_desc);
%ignore TransParam(string in_name, void* in_value, dmDataType in_datatype,
                    string in_desc, string in_parent, long nvals);
%ignore TransParam(string in_name, void* in_value, dmDataType in_datatype,
                    string in_desc, string in_parent);
%ignore operator= ;


/* From the python side, all of these accept just an input array,  */
/* the typemap takes care of the other argument(s).. so ingore the */
/* defaulted functions in the python wrappers                      */
%ignore TransParam::set_value(bool* in_value, long in_nvals=1);
%ignore TransParam::set_value(short* in_value, long in_nvals=1);
%ignore TransParam::set_value(long* in_value, long in_nvals=1);
%ignore TransParam::set_value(float* in_value, long in_nvals=1);
%ignore TransParam::set_value(double* in_value, long in_nvals=1);
%ignore TransParam::set_value(unsigned short* in_value, long in_nvals=1);
%ignore TransParam::set_value(unsigned long* in_value, long in_nvals=1);
%ignore TransParam::set_value(unsigned char* in_value, long in_nvals=1);
%ignore TransParam::set_value(char** in_value, long in_nvals=1);
%ignore TransParam::set_value(void* in_value, dmDataType in_datatype);
%ignore TransParam::set_value(bool* in_value);
%ignore TransParam::set_value(short* in_value);
%ignore TransParam::set_value(long* in_value);
%ignore TransParam::set_value(float* in_value);
%ignore TransParam::set_value(double* in_value);
%ignore TransParam::set_value(unsigned short* in_value);
%ignore TransParam::set_value(unsigned long* in_value);
%ignore TransParam::set_value(unsigned char* in_value);
%ignore TransParam::set_value(char** in_value);


%ignore Transform::clear;



/* ====================================================================== */
/* Make dmErrCode defs available to SWIG/pytransform                      */
/* ====================================================================== */
typedef int dmErrCode;


%inline
%{

  /* ---------------------------------------------------------------------- */
  /*  get_pyType()                                                          */
  /*    Helper function to convert from DM data types to numpy data types   */
  /* ---------------------------------------------------------------------- */
  int get_pyType( dmDataType intype )
  {
    int retval;

    switch ( intype ){
    case dmSHORT:
      retval = NPY_SHORT;
      break;
    case dmLONG:
      retval = NPY_LONG;
      break;
    case dmFLOAT:
      retval = NPY_FLOAT;
      break;
    case dmDOUBLE:
      retval = NPY_DOUBLE;
      break;
    case dmTEXT:
      retval = NPY_STRING;
      break;
    case dmBYTE:
      retval = NPY_BYTE;
      break;
    case dmBOOL:
      retval = NPY_BOOL;
      break;
    case dmUSHORT:
      retval = NPY_USHORT;
      break;
    case dmULONG:
      retval = NPY_ULONG;
      break;
    default:
      retval = NPY_NOTYPE;
      break;
    }
    return retval;
  }

  /* ---------------------------------------------------------------------- */
  /*  get_dmType()                                                          */
  /*    Helper function to convert from numpy data types to DM data types   */
  /* ---------------------------------------------------------------------- */
  dmDataType get_dmType( int intype )
  {
    dmDataType retval;

    switch ( intype ){
    case NPY_VOID:
      retval = dmUNKNOWNTYPE;
      break;
    case NPY_SHORT:
      retval = dmSHORT;
      break;
    case NPY_LONG:
      retval = dmLONG;
      break;
    case NPY_FLOAT:
      retval = dmFLOAT;
      break;
    case NPY_DOUBLE:
      retval = dmDOUBLE;
      break;
    case NPY_STRING:
      retval = dmTEXT;
      break;
    case NPY_BYTE:
      retval = dmBYTE;
      break;
    case NPY_BOOL:
      retval = dmBOOL;
      break;
    case NPY_USHORT:
      retval = dmUSHORT;
      break;
    case NPY_ULONG:
      retval = dmULONG;
      break;
    default:
      retval = dmUNKNOWNTYPE;
      break;
    }
    return retval;
  }
  /* ---------------------------------------------------------------------- */
  /*  get_dmType_str()                                                      */
  /*    Helper function to return the string name of the DM datatype        */
  /*    which corresponds to the input numpy array type.                    */
  /* ---------------------------------------------------------------------- */
  char* get_dmType_str( int intype )
  {
    char* retval = (char*)malloc(100*sizeof(char));

    switch ( intype ){
    case NPY_VOID:
      strcpy(retval,"VOID");
      break;
    case NPY_SHORT:
      strcpy(retval,"dmSHORT");
      break;
    case NPY_LONG:
      strcpy(retval,"dmLONG");
      break;
    case NPY_FLOAT:
      strcpy(retval,"dmFLOAT");
      break;
    case NPY_DOUBLE:
      strcpy(retval,"dmDOUBLE");
      break;
    case NPY_STRING:
      strcpy(retval,"dmTEXT");
      break;
    case NPY_BYTE:
      strcpy(retval,"dmBYTE");
      break;
    case NPY_BOOL:
      strcpy(retval,"dmBOOL");
      break;
    case NPY_USHORT:
      strcpy(retval,"dmUSHORT");
      break;
    case NPY_ULONG:
      strcpy(retval,"dmULONG");
      break;
    default:
      strcpy(retval,"dmUNKNOWNTYPE");
      break;
    }
    return retval;
  }
%}

%extend TransParam {

  /* ---------------------------------------------------------------------- */
  /*  Constructor( name, value(s), description )                            */
  /*    Defines Parameter object based on input value.                      */
  /* ---------------------------------------------------------------------- */
  TransParam( PyObject* name, PyObject* indata, PyObject* description )
  {
     PyObject*  inarr = NULL;

     void*      vptr = NULL;
     int        pytype;
     dmDataType dtype;
     long       nvals;

     TransParam* tParam;

     if ( indata != NULL )
     {
       inarr = PyArray_FromAny( indata, NULL, 0, MAX_DIMS, NPY_CARRAY , NULL);
     }

     if ( inarr == NULL )
     {
       dtype = dmUNKNOWNTYPE;
       nvals = 0;
       vptr  = NULL;
     }
     else
     {
       /* Get datatype and nvals from array object */
       pytype = PyArray_TYPE(inarr);
       dtype  = get_dmType( pytype );
       nvals  = PyArray_SIZE(inarr);

       vptr  = PyArray_DATA(inarr);
     }

     /* Generate a temporary Parameter object with input definition */
     tParam = new TransParam( PyString_AsString( name ),
		             vptr,
			     dtype,
			     PyString_AsString( description ),
			     "",
			     nvals);

     return( tParam );
  }

}


%include "Transform.hh"     
%include "TransParam.hh"     

%init %{
  import_array();
%}
