// 
//  Copyright (C) 2006-2007  Smithsonian Astrophysical Observatory
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//


#include <iostream>
#include <string>
#include <cstdlib>

#include "Transform.hh"
 
using namespace std;

int main(int argc, char* argv[], char* env[] )
{
  int ii=0;
  string indir = getenv("TESTIN") + string("/test_transform/");
  string outdir = getenv("TESTOUT") + string("/transform/");

  // Create a new WCSTANTransform
  cout << "Create a new WCSTANTransform" << endl;
  WCSTANTransform* wcstrans = new WCSTANTransform("EQPOS");

  cout << "Transform Class Name == " << wcstrans->get_className() << endl;

  // Print generated parameter list
  cout << "Generate parameter list w/ transform name 'EQPOS':" << endl;

  list<TransParam*> params = wcstrans->get_parameter_list();
  list<TransParam*>::iterator p_iter;
  for (p_iter = params.begin(); p_iter != params.end(); p_iter++)
    cout << (*p_iter)->print() << endl;
  cout << endl;


  // Update Parameter list
  TransParam* crpix_param = wcstrans->get_parameter("CRPIX");
  TransParam* crval_param = wcstrans->get_parameter("CRVAL");
  TransParam* cdelt_param = wcstrans->get_parameter("CDELT");
  TransParam* ctype_param = wcstrans->get_parameter("CTYPE");

  // Update CRPIX parameter 
  double* crpix = new double[2];
  crpix[0] = 359.98373514;
  crpix[1] = -12.69992072;

  TransParam* new_crpix_param = new TransParam("CRPIX", (void*)crpix, dmDOUBLE,
					       crpix_param->get_desc(), 2);
  wcstrans->set_parameter(new_crpix_param);

  crpix_param->set_parent((string)"eqposssss");
  cout << "Update CRPIX parent:" << endl;
  cout << "   Name: " << crpix_param->get_name() << endl;
  cout << "   Value: " << ((double*)crpix_param->get_value())[0] << ", " 
       << ((double*)crpix_param->get_value())[1] << endl;
  cout << "   Datatype: " << crpix_param->get_datatype() << endl;
  cout << "   Description: " << crpix_param->get_desc() << endl;
  cout << "   Parent: " << crpix_param->get_parent() << endl;
  cout << "   Nvals: " << crpix_param->get_nvals() << endl;
  cout << endl;

  // Update CRVAL parameter 
  double* crval = new double[2];
  crval[0] = 4096.50000000;
  crval[1] = 4096.50000000;

  crval_param->set_value((void*)crval, dmDOUBLE, 2);

  // Update CDELT parameter 
  double* cdelt = new double[2];
  cdelt[0] = -0.136667;
  cdelt[1] = 0.136667;

  cdelt_param->set_value(cdelt, 2);

  //cout << "Update CTYPE" << endl;
  string ctype[2] = {"LINEAR", "LINEAR"};
  ctype_param->set_value( (void*)ctype, dmTEXT, 2 );

  // Print updated parameter list
  cout << "Updated parameter list w/ parents changed to 'eqpos': " << endl;
  wcstrans->set_name((string)"eqpos");

  cout << wcstrans->print_parameter_list() << endl;

  // Get WCSTAN transform matrix
  cout << "Get WCSTAN transform matrix: " << endl;
  double** matrix = wcstrans->get_transform_matrix();
  for (int mm = 0; mm < 3 ; mm++)
    cout << "[" << mm << "][0]: " << matrix[mm][0] << "   [" << mm << "][1]: " << matrix[mm][1] 
	 << "   [" << mm << "][2]: " << matrix[mm][2]  << endl;
  cout << endl;


  cout << "Test WCS Apply and Invert Transform functions using shorts" << endl;
  // Create data array of logical pairs along the diagonal - dmSHORT
  // e.g. (0,0), (1,1), (2,2), etc.
  short* data_s = new short[200];
  int nsets = 100;

  for(ii=0; ii < nsets; ii++) {
    data_s[ii*2] = (ii+1)*5;
    data_s[ii*2+1] = (ii+1)*5;
  }

  // Apply WCSTAN transform to data_s
  short* wcstrans_data_s = (short*)wcstrans->apply((void*)data_s, dmSHORT, nsets);

  // Invert applied data_s
  short* wcsinvs_data_s = ((WCSTANTransform*)wcstrans)->invert(wcstrans_data_s, nsets);

  // Print results from apply and inverse transforms
  // cout << "Results of transforming data_s array and then applying the inverse:" << endl;
  for(ii=0; ii < nsets; ii++) {
    cout << "Data_S: (" << data_s[ii*2] << "," << data_s[ii*2+1] <<")   \tApply: (" 
	 << wcstrans_data_s[ii*2] << "," << wcstrans_data_s[ii*2+1] << ")   \tInverse: (" 
	 << wcsinvs_data_s[ii*2] << "," << wcsinvs_data_s[ii*2+1] << ")" << endl;
  }

  //delete [] data_s;
  delete [] wcstrans_data_s;
  delete [] wcsinvs_data_s;

  cout << endl << "Test WCS Apply and Invert Transform functions using doubles" << endl;
  // Create data_d array of logical pairs along the diagonal - dmDOUBLE
  // e.g. (0,0), (1,1), (2,2), etc.
  double* data_d = new double[200];

  for(ii=0; ii < nsets; ii++) {
    data_d[ii*2] = (ii+1)*5;
    data_d[ii*2+1] = (ii+1)*5;
  }

  // Apply WCSTAN transform to data_d
  double* wcstrans_data_d = ((WCSTANTransform*)wcstrans)->apply(data_d, nsets);

  // Invert applied data_d
  double* wcsinvs_data_d = (double*)wcstrans->invert((void*)wcstrans_data_d, dmDOUBLE,  nsets);

  // Print results from apply and inverse transforms
  //cout << "Results of transforming data_d array and then applying the inverse:" << endl;
  for(ii=0; ii < nsets; ii++) {
    cout.setf(ios_base::fixed,ios_base::floatfield);
    cout << "Data_D: (" << data_d[ii*2] << "," << data_d[ii*2+1] <<")   \tApply: (" 
	 << wcstrans_data_d[ii*2] << "," << wcstrans_data_d[ii*2+1] << ")   \tInverse: (" 
	 << wcsinvs_data_d[ii*2] << "," << wcsinvs_data_d[ii*2+1] << ")" << endl;
    cout.unsetf(ios_base::floatfield);
  }

  delete [] wcstrans_data_d;
  delete [] wcsinvs_data_d;

  cout << endl << "Create a new WCSTANTransform" << endl;
  WCSTANTransform* wcs2 = new WCSTANTransform("WCS2");
  wcs2->get_parameter("CRVAL")->set_value((void*)crval, dmDOUBLE, 2);
  cout << wcs2->print_parameter_list() << endl;

  cout << "Set parameters from input transform matrix" << endl;
  wcs2->set_transform_matrix(matrix);
  cout << wcs2->print_parameter_list() << endl;


  cout << "Test LINEAR2D transform" << endl;
  cout << "Create a new LINEAR2D transform" << endl << endl;

  Transform* lin2d = new LINEAR2DTransform("LIN1");

  cout << "Transform Class Name == " << lin2d->get_className() << endl;
  cout << "Print LINEAR2D parameter list: " << endl;
  cout <<  lin2d->print_parameter_list();


  cout << endl << "Copy the linear2d transform" << endl;
  Transform* lin2d_copy = lin2d->copy();

  lin2d_copy->set_name("LIN1_COPY");

  cout << "COPY - Transform Class Name == " << lin2d_copy->get_className() << endl;
  cout << "COPY - Transform Name == " << lin2d_copy->get_name() << endl;

  cout << "COPY - Print LINEAR2D parameter list: " << endl;
  cout << lin2d_copy->print_parameter_list();

  cout << endl << "Set the transform matrix using matrix from WCSTransform" << endl;
  lin2d->set_transform_matrix(matrix);

  cout << endl << "Print updated LINEAR2D parameter list: " << endl;
  cout << lin2d->print_parameter_list();

  cout << endl << "COPY - Print LINEAR2D parameter list: " << endl;
  cout << lin2d_copy->print_parameter_list();

  cout << "Delete copy of lin2d transform" << endl;
  delete lin2d_copy;

  cout << endl << "Get/Calculate the transform matrix" << endl;
  double ** lin2d_matrix = lin2d->get_transform_matrix();
  for (int mm = 0; mm < 3 ; mm++)
    cout << "[" << mm << "][0]: " << lin2d_matrix[mm][0] << "   [" << mm << "][1]: " << lin2d_matrix[mm][1] 
	 << "   [" << mm << "][2]: " << lin2d_matrix[mm][2]  << endl;
  cout << endl;

  cout << "Test LINEAR2D Apply and Invert Transform functions using shorts" << endl;

  // Apply LINEAR2D transform to data_s
  short* lin2dtrans_data_s = (short*)lin2d->apply((void*)data_s, dmSHORT, nsets);

  // Invert applied data_s
  short* lin2dinvs_data_s = ((LINEAR2DTransform*)lin2d)->invert(lin2dtrans_data_s, nsets);

  // Print results from apply and inverse transforms
  // cout << "Results of transforming data_s array and then applying the inverse:" << endl;
  for(ii=0; ii < nsets; ii++) {
    cout << "Data_S: (" << data_s[ii*2] << "," << data_s[ii*2+1] <<")   \tApply: (" 
	 << lin2dtrans_data_s[ii*2] << "," << lin2dtrans_data_s[ii*2+1] << ")   \tInverse: (" 
	 << lin2dinvs_data_s[ii*2] << "," << lin2dinvs_data_s[ii*2+1] << ")" << endl;
  }

  cout << endl;
  delete [] data_s;
  delete [] lin2dtrans_data_s;
  delete [] lin2dinvs_data_s;

  cout << "Test LINEAR2D Apply and Invert Transform functions using doubless" << endl;

  // Apply LINEAR2D transform to data_d
  double* lin2dtrans_data_d = ((LINEAR2DTransform*)lin2d)->apply(data_d, nsets);

  // Invert applied data_d
  double* lin2dinvs_data_d = (double*)lin2d->invert((void*)lin2dtrans_data_d, dmDOUBLE, nsets);

  // Print results from apply and inverse transforms
  //cout << "Results of transforming data_d array and then applying the inverse:" << endl;
  for(ii=0; ii < nsets; ii++) {
    cout.setf(ios_base::fixed,ios_base::floatfield);
    cout << "Data_D: (" << data_d[ii*2] << "," << data_d[ii*2+1] <<")   \tApply: (" 
	 << lin2dtrans_data_d[ii*2] << "," << lin2dtrans_data_d[ii*2+1] << ")   \tInverse: (" 
	 << lin2dinvs_data_d[ii*2] << "," << lin2dinvs_data_d[ii*2+1] << ")" << endl;
    cout.unsetf(ios_base::floatfield);
  }

  delete [] data_d;
  delete [] lin2dtrans_data_d;
  delete [] lin2dinvs_data_d;

  exit(0);

};
